/*
 * Copyright (C) 2008, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#if ENABLE(ASSEMBLER)

#include "AssemblerData.h"
#include "ExecutableAllocator.h"
#include "JITCompilationEffort.h"
#include "JSGlobalData.h"
#include "stdint.h"
#include <string.h>
#include <wtf/Assertions.h>
#include <wtf/FastMalloc.h>
#include <wtf/StdLibExtras.h>
#include <wtf/UnalignedAccess.h>

namespace JSC {

class CodeBlock;

struct AssemblerLabel {
    inline AssemblerLabel() { setOffset(std::numeric_limits<uint32_t>::max()); }
    inline AssemblerLabel(const AssemblerLabel& other) { setOffset(other.offset()); }
    inline AssemblerLabel(AssemblerLabel&& other) { setOffset(other.offset()); }
    inline explicit AssemblerLabel(uint32_t offset) { setOffset(offset); }

    AssemblerLabel& operator=(const AssemblerLabel& other) { setOffset(other.offset()); return *this; }
    AssemblerLabel& operator=(AssemblerLabel&& other) { setOffset(other.offset()); return *this; }

    bool isSet() const { return (offset() != std::numeric_limits<uint32_t>::max()); }

    inline AssemblerLabel labelAtOffset(int offset) const
    {
        return AssemblerLabel(this->offset() + offset);
    }

    bool operator==(const AssemblerLabel& other) const { return offset() == other.offset(); }

    inline uint32_t offset() const
    {
#if CPU(ARM64E)
        return static_cast<uint32_t>(untagInt(m_offset, bitwise_cast<PtrTag>(this)));
#else
        return m_offset;
#endif
    }

private:
    inline void setOffset(uint32_t offset)
    {
#if CPU(ARM64E)
        m_offset = tagInt(static_cast<uint64_t>(offset), bitwise_cast<PtrTag>(this));
#else
        m_offset = offset;
#endif
    }

#if CPU(ARM64E)
    uint64_t m_offset;
#else
    uint32_t m_offset;
#endif
};

class AssemblerBuffer {
public:
	AssemblerBuffer()
	: m_index(0)
	{
	}

	ALWAYS_INLINE bool isAvailable(unsigned space)
	{
        return m_index + space <= m_storage.capacity();
	}

	ALWAYS_INLINE void ensureSpace(unsigned space)
	{
		if (UNLIKELY(!isAvailable(space)))
			grow(space);
	}

	ALWAYS_INLINE bool isAligned(int alignment) const
	{
		return !(m_index & (alignment - 1));
	}

#if !CPU(ARM_THUMB2)
	ALWAYS_INLINE void putByteUnchecked(int8_t value) { putIntegralUnchecked(value); }
	ALWAYS_INLINE void putByte(int8_t value) { putIntegral(value); }
#endif
	ALWAYS_INLINE void putShortUnchecked(int16_t value) { putIntegralUnchecked(value); }
	ALWAYS_INLINE void putShort(int16_t value) { putIntegral(value); }
#if !CPU(ARM_THUMB2)
	ALWAYS_INLINE void putIntUnchecked(int32_t value) { putIntegralUnchecked(value); }
	ALWAYS_INLINE void putInt(int32_t value) { putIntegral(value); }
	ALWAYS_INLINE void putInt64Unchecked(int64_t value) { putIntegralUnchecked(value); }
	ALWAYS_INLINE void putInt64(int64_t value) { putIntegral(value); }
#endif

	void* data() const
	{
        return reinterpret_cast<void*>(m_storage.buffer());
	}

	size_t codeSize() const
	{
		return m_index;
	}

    void setCodeSize(size_t index)
    {
        // Warning: Only use this if you know exactly what you are doing.
        // For example, say you want 40 bytes of nops, it's ok to grow
        // and then fill 40 bytes of nops using bigger instructions.
        m_index = index;
        ASSERT(m_index <= m_storage.capacity());
    }

	AssemblerLabel label() const
	{
		return AssemblerLabel(m_index);
	}

	unsigned debugOffset() { return m_index; }

    AssemblerData&& releaseAssemblerData() { return WTFMove(m_storage); }

    // LocalWriter is a trick to keep the storage buffer and the index
    // in memory while issuing multiple Stores.
    // It is created in a block scope and its attribute can stay live
    // between writes.
    //
    // LocalWriter *CANNOT* be mixed with other types of access to AssemblerBuffer.
    // AssemblerBuffer cannot be used until its LocalWriter goes out of scope.
    class LocalWriter {
    public:
        LocalWriter(AssemblerBuffer& buffer, unsigned requiredSpace)
            : m_buffer(buffer)
        {
            buffer.ensureSpace(requiredSpace);
            m_storageBuffer = buffer.m_storage.buffer();
            m_index = buffer.m_index;
#if !defined(NDEBUG)
            m_initialIndex = m_index;
            m_requiredSpace = requiredSpace;
#endif
        }

        ~LocalWriter()
        {
            ASSERT(m_index - m_initialIndex <= m_requiredSpace);
            ASSERT(m_buffer.m_index == m_initialIndex);
            ASSERT(m_storageBuffer == m_buffer.m_storage.buffer());
            m_buffer.m_index = m_index;
        }

        void putByteUnchecked(int8_t value) { putIntegralUnchecked(value); }
        void putShortUnchecked(int16_t value) { putIntegralUnchecked(value); }
        void putIntUnchecked(int32_t value) { putIntegralUnchecked(value); }
        void putInt64Unchecked(int64_t value) { putIntegralUnchecked(value); }
    private:
        template<typename IntegralType>
        void putIntegralUnchecked(IntegralType value)
        {
            ASSERT(m_index + sizeof(IntegralType) <= m_buffer.m_storage.capacity());
            WTF::unalignedStore<IntegralType>(m_storageBuffer + m_index, value);
            m_index += sizeof(IntegralType);
        }
        AssemblerBuffer& m_buffer;
        char* m_storageBuffer;
        unsigned m_index;
#if !defined(NDEBUG)
        unsigned m_initialIndex;
        unsigned m_requiredSpace;
#endif
    };

protected:
    template<typename IntegralType>
    ALWAYS_INLINE void putIntegral(IntegralType value)
    {
        unsigned nextIndex = m_index + sizeof(IntegralType);
        if (UNLIKELY(nextIndex > m_storage.capacity()))
        	grow();
        putIntegralUnchecked<IntegralType>(value);
    }

    template<typename IntegralType>
    ALWAYS_INLINE void putIntegralUnchecked(IntegralType value)
    {
        ASSERT(isAvailable(sizeof(IntegralType)));
#if CPU(ARM_THUMB2)
        // only push short
        *reinterpret_cast_ptr<IntegralType*>(m_storage.buffer() + m_index) = value;
#else
        WTF::unalignedStore<IntegralType>(m_storage.buffer() + m_index, value);
#endif
        m_index += sizeof(IntegralType);
    }

	void append(const char* data, int size)
	{
		if (UNLIKELY(!isAvailable(size)))
			grow(size);

        memcpy(m_storage.buffer() + m_index, data, size);
		m_index += size;
	}

	void grow(int extraCapacity = 0)
	{
        m_storage.grow(extraCapacity);
	}

private:
    friend class LocalWriter;

    unsigned m_index;
    AssemblerData m_storage;
};

} // namespace JSC

#endif // ENABLE(ASSEMBLER)
