/*
 * Copyright (C) 2008, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <wtf/Assertions.h>
#include <wtf/FastAllocBase.h>
#include <wtf/Noncopyable.h>

#if ENABLE(ASSEMBLER)

namespace JSC {

class JSGlobalData;
typedef JSGlobalData VM;

class AssemblerData final {
    WTF_MAKE_NONCOPYABLE(AssemblerData);WTF_MAKE_FAST_ALLOCATED;
    static constexpr size_t InlineCapacity = 4096;
public:
    AssemblerData()
		: m_buffer(m_inlineBuffer)
		, m_capacity(InlineCapacity)
	{
	}

    AssemblerData(AssemblerData&& other)
    {
        if (other.isInlineBuffer()) {
            ASSERT(other.m_capacity == InlineCapacity);
            memcpy(m_inlineBuffer, other.m_inlineBuffer, InlineCapacity);
            m_buffer = m_inlineBuffer;
        } else
            m_buffer = other.m_buffer;
        m_capacity = other.m_capacity;

        other.m_buffer = other.m_inlineBuffer;
        other.m_capacity = InlineCapacity;
    }

    AssemblerData& operator=(AssemblerData&& other)
    {
        if (m_buffer && !isInlineBuffer())
            fastFree(m_buffer);

        if (other.isInlineBuffer()) {
            ASSERT(other.m_capacity == InlineCapacity);
            memcpy(m_inlineBuffer, other.m_inlineBuffer, InlineCapacity);
            m_buffer = m_inlineBuffer;
        } else
            m_buffer = other.m_buffer;
        m_capacity = other.m_capacity;

        other.m_buffer = other.m_inlineBuffer;
        other.m_capacity = InlineCapacity;
        return *this;
    }

    ~AssemblerData()
    {
		clear();
	}

	void clear()
	{
		if (m_buffer && !isInlineBuffer()) {
			fastFree(m_buffer);
			m_capacity = InlineCapacity;
			m_buffer = m_inlineBuffer;
		}
	}

    char* buffer() const { return m_buffer; }

    unsigned capacity() const { return m_capacity; }

    void grow(unsigned extraCapacity = 0);

private:
    bool isInlineBuffer() const { return m_buffer == m_inlineBuffer; }
    char* m_buffer;
    char m_inlineBuffer[InlineCapacity];
    unsigned m_capacity;
};

} // namespace JSC

#endif // ENABLE(ASSEMBLER)
