/*
 * Copyright (C) 2016 Yusuke Suzuki <utatane.tea@gmail.com>
 * Copyright (C) 2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "BytecodeGeneratorification.h"

#include "BytecodeGenerator.h"
#include "BytecodeLivenessAnalysisInlines.h"
#include "BytecodeRewriter.h"
#include "BytecodeUseDef.h"
#include "IdentifierInlines.h"
#include "InterpreterInlines.h"
#include "JSCInlines.h"
#include "JSCJSValueInlines.h"
#include "JSGeneratorFunction.h"
#include "StrongInlines.h"
#include "CodeBlock.h"
#include <wtf/Optional.h>

namespace JSC {

namespace {
struct YieldData {
	YieldData() : point(0), argument(0) { }
    size_t point;
    int argument;
    FastBitVector liveness;
};

class BytecodeGeneratorification {
public:
    typedef Vector<YieldData> Yields;

    struct GeneratorFrameData {
    	size_t m_point;
        int m_dst;
        int m_scope;
        int m_symbolTable;
        int m_initialValue;
    };

    BytecodeGeneratorification(BytecodeGenerator& bytecodeGenerator, CodeBlock* codeBlock, Vector<Instruction>& instructions, SymbolTable* generatorFrameSymbolTable, int generatorFrameSymbolTableIndex)
        : m_bytecodeGenerator(bytecodeGenerator)
    	, m_enterPoint(0)
    	, m_graph(codeBlock, instructions)
        , m_generatorFrameSymbolTable(codeBlock->vm(), generatorFrameSymbolTable)
        , m_generatorFrameSymbolTableIndex(generatorFrameSymbolTableIndex)
    {
    	for(int blockIndex = 0; blockIndex < m_graph.size(); ++blockIndex) {
    		BytecodeBasicBlock* block = m_graph.at(blockIndex);
    		for(int offsetIndex = 0; offsetIndex < block->offsets().size(); ++offsetIndex) {
    			unsigned bytecodeOffset = block->offsets()[offsetIndex];
                const Instruction* pc = &m_graph.instructions()[bytecodeOffset];
                switch (pc->u.opcode) {
                case op_enter: {
                    m_enterPoint = bytecodeOffset;
                    break;
                }

                case op_yield: {
                    unsigned liveCalleeLocalsIndex = pc[2].u.unsignedValue;
                    if (liveCalleeLocalsIndex >= m_yields.size())
                        m_yields.resize(liveCalleeLocalsIndex + 1);
                    YieldData& data = m_yields[liveCalleeLocalsIndex];
                    data.point = bytecodeOffset;
                    data.argument = pc[3].u.operand;
                    break;
                }

                case op_create_generator_frame_environment: {
                    GeneratorFrameData data;
                    data.m_point = bytecodeOffset;
                    data.m_dst = pc[1].u.operand;
                    data.m_scope = pc[2].u.operand;
                    data.m_symbolTable = pc[3].u.operand;
                    data.m_initialValue = pc[4].u.operand;
                    m_generatorFrameData = data;
                    break;
                }

                default:
                    break;
                }
            }
        }
    }

    struct Storage {
        Identifier identifier;
        unsigned identifierIndex;
        ScopeOffset scopeOffset;
    };

    void run();

    BytecodeGraph<CodeBlock, Vector<Instruction>>& graph() { return m_graph; }

    const Yields& yields() const
    {
        return m_yields;
    }

    Yields& yields()
    {
        return m_yields;
    }

    unsigned enterPoint() const
    {
        return m_enterPoint;
    }

    Optional<GeneratorFrameData> generatorFrameData() const
    {
        return m_generatorFrameData;
    }

    int generatorFrameSymbolTableIndex() const { return m_generatorFrameSymbolTableIndex; }

    Storage storageForGeneratorLocal(unsigned index)
    {
        // We assign a symbol to a register. There is one-on-one corresponding between a register and a symbol.
        // By doing so, we allocate the specific storage to save the given register.
        // This allow us not to save all the live registers even if the registers are not overwritten from the previous resuming time.
        // It means that, the register can be retrieved even if the immediate previous op_save does not save it.

        if (m_storages.size() <= index)
            m_storages.resize(index + 1);
        if (Optional<Storage> storage = m_storages[index])
            return *storage;

        CodeBlock* codeBlock = m_graph.codeBlock();
        Identifier identifier = Identifier::fromUid(PrivateName());
        unsigned identifierIndex = codeBlock->numberOfIdentifiers();
        codeBlock->addIdentifier(identifier);
        ScopeOffset scopeOffset = m_generatorFrameSymbolTable->takeNextScopeOffset();
        m_generatorFrameSymbolTable->set(identifier.impl(), SymbolTableEntry(VarOffset(scopeOffset)));

        Storage storage = {
            identifier,
            identifierIndex,
            scopeOffset
        };
        m_storages[index] = storage;
        return storage;
    }

private:
    BytecodeGenerator& m_bytecodeGenerator;
    unsigned m_enterPoint;
    Optional<GeneratorFrameData> m_generatorFrameData;
    BytecodeGraph<CodeBlock, Vector<Instruction>> m_graph;
    Vector<Optional<Storage>> m_storages;
    Yields m_yields;
    Strong<SymbolTable> m_generatorFrameSymbolTable;
    int m_generatorFrameSymbolTableIndex;
};

class GeneratorLivenessAnalysis : public BytecodeLivenessPropagation {
public:
    GeneratorLivenessAnalysis(BytecodeGeneratorification& generatorification)
        : m_generatorification(generatorification)
    {
    }

    void run()
    {
        // Perform modified liveness analysis to determine which locals are live at the merge points.
        // This produces the conservative results for the question, "which variables should be saved and resumed?".

        runLivenessFixpoint(m_generatorification.graph());

        for(int i = 0; i < m_generatorification.yields().size(); ++i) {
        	YieldData& data = m_generatorification.yields()[i];
            data.liveness = getLivenessInfoAtBytecodeOffset(m_generatorification.graph(), data.point + opcodeLengths[op_yield]);
        }
    }

private:
    BytecodeGeneratorification& m_generatorification;
};

void BytecodeGeneratorification::run()
{
    // We calculate the liveness at each merge point. This gives us the information which registers should be saved and resumed conservatively.

    {
        GeneratorLivenessAnalysis pass(*this);
        pass.run();
    }

    CodeBlock* codeBlock = m_graph.codeBlock();
    BytecodeRewriter rewriter(m_graph);

    // Setup the global switch for the generator.
    {
        unsigned nextToEnterPoint = enterPoint() + opcodeLengths[op_enter];
        unsigned switchTableIndex = m_graph.codeBlock()->numberOfSwitchJumpTables();
        VirtualRegister_ state(virtualRegisterForArgument(static_cast<int32_t>(JSGeneratorFunction::GeneratorArgument::State)));
        auto& jumpTable = m_graph.codeBlock()->addSwitchJumpTable();
        jumpTable.min = 0;
        jumpTable.branchOffsets.resize(m_yields.size() + 1);
        jumpTable.branchOffsets.fill(0);
        jumpTable.add(0, nextToEnterPoint);
        for (unsigned i = 0; i < m_yields.size(); ++i)
            jumpTable.add(i + 1, m_yields[i].point);

        struct FragmentFunctor
        {
        	FragmentFunctor(unsigned nextToEnterPoint, unsigned switchTableIndex, VirtualRegister_ state)
        	: nextToEnterPoint(nextToEnterPoint)
        	, switchTableIndex(switchTableIndex)
        	, state(state)
        	{ }
			void operator()(BytecodeRewriter::Fragment& fragment) {
				fragment.appendInstruction(op_switch_imm, switchTableIndex, nextToEnterPoint, state.offset());
			}
			unsigned nextToEnterPoint;
			unsigned switchTableIndex;
			VirtualRegister_ state;
        };
        FragmentFunctor fragmentFunctor(nextToEnterPoint, switchTableIndex, state);
        rewriter.insertFragmentBefore(nextToEnterPoint, fragmentFunctor);
    }

    for(int i = 0; i < m_yields.size(); ++i) {
    	const YieldData& data = m_yields[i];
        VirtualRegister_ scope(virtualRegisterForArgument(static_cast<int32_t>(JSGeneratorFunction::GeneratorArgument::Frame)));

        // Emit save sequence.
        {
        	struct SaveFragmentFunctor {
        		SaveFragmentFunctor(BytecodeGenerator& bytecodeGenerator, BytecodeGeneratorification& bytecodeGeneratorification, const YieldData& data, VirtualRegister_ scope)
        		: bytecodeGenerator(bytecodeGenerator)
        		, bytecodeGeneratorification(bytecodeGeneratorification)
        		, data(data)
        		, scope(scope)
        		{ }
        		void operator()(BytecodeRewriter::Fragment& fragment) const {
					data.liveness.forEachSetBit([&] (size_t index) {
						VirtualRegister_ operand(virtualRegisterForLocal(index));
						Storage storage = bytecodeGeneratorification.storageForGeneratorLocal(index);

						fragment.appendInstruction(
							op_put_to_scope,
							scope.offset(), // scope
							storage.identifierIndex, // identifier
							operand.offset(), // value
							GetPutInfo(DoNotThrowIfNotFound, ResolvedClosureVar, InitializationMode::NotInitialization, bytecodeGenerator.ecmaMode()).operand(), // info
							bytecodeGeneratorification.generatorFrameSymbolTableIndex(), // symbol table constant index
							storage.scopeOffset.offset() // scope offset
						);
					});

					// Insert op_ret just after save sequence.
					fragment.appendInstruction(op_ret, data.argument);
				}
        		BytecodeGenerator& bytecodeGenerator;
        		BytecodeGeneratorification& bytecodeGeneratorification;
        		const YieldData& data;
        		VirtualRegister_ scope;
        	};
        	SaveFragmentFunctor fragmentFunctor(m_bytecodeGenerator, *this, data, scope);
        	rewriter.insertFragmentBefore(data.point, fragmentFunctor);
        }

        {
        	// Emit resume sequence.
        	struct ResumeFragmentFunctor {
        		ResumeFragmentFunctor(BytecodeGenerator& bytecodeGenerator, BytecodeGeneratorification& bytecodeGeneratorification, CodeBlock* codeBlock, const YieldData& data, VirtualRegister_ scope)
        		: bytecodeGenerator(bytecodeGenerator)
        		, bytecodeGeneratorification(bytecodeGeneratorification)
        		, codeBlock(codeBlock)
        		, data(data)
        		, scope(scope)
        		{ }
        		void operator()(BytecodeRewriter::Fragment& fragment) {
    				data.liveness.forEachSetBit([&] (size_t index) {
						VirtualRegister_ operand(virtualRegisterForLocal(index));
						Storage storage = bytecodeGeneratorification.storageForGeneratorLocal(index);

						UnlinkedValueProfile profile = codeBlock->addValueProfile();
						fragment.appendInstruction(
							op_get_from_scope,
							operand.offset(), // dst
							scope.offset(), // scope
							storage.identifierIndex, // identifier
							GetPutInfo(DoNotThrowIfNotFound, ResolvedClosureVar, InitializationMode::NotInitialization, bytecodeGenerator.ecmaMode()).operand(), // info
							0, // local scope depth
							storage.scopeOffset.offset(), // scope offset
							profile // profile
						);
					});
				}
        		BytecodeGenerator& bytecodeGenerator;
        		BytecodeGeneratorification& bytecodeGeneratorification;
        		CodeBlock* codeBlock;
        		const YieldData& data;
        		VirtualRegister_ scope;
        	};
        	ResumeFragmentFunctor fragmentFunctor(m_bytecodeGenerator, *this, codeBlock, data, scope);
			rewriter.insertFragmentAfter(data.point, fragmentFunctor);
        }

        // Clip the unnecessary bytecodes.
        rewriter.removeBytecode(data.point);
    }

    if (m_generatorFrameData) {
    	class Functor {
    	public:
    		Functor(Strong<SymbolTable>& generatorFrameSymbolTable, Optional<GeneratorFrameData>& generatorFrameData)
			: m_generatorFrameSymbolTable(generatorFrameSymbolTable)
    		, m_generatorFrameData(generatorFrameData)
			{
			}

    		void operator()(BytecodeRewriter::Fragment& fragment) {
                if (!m_generatorFrameSymbolTable->scopeSize()) {
                    // This will cause us to put jsUndefined() into the generator frame's scope value.
                    fragment.appendInstruction(op_mov, m_generatorFrameData->m_dst, m_generatorFrameData->m_initialValue);
                } else
                    fragment.appendInstruction(op_create_lexical_environment, m_generatorFrameData->m_dst, m_generatorFrameData->m_scope, m_generatorFrameData->m_symbolTable, m_generatorFrameData->m_initialValue);
            }
    	private:
    		Strong<SymbolTable>& m_generatorFrameSymbolTable;
    		Optional<GeneratorFrameData>& m_generatorFrameData;
    	};
    	Functor functor(m_generatorFrameSymbolTable, m_generatorFrameData);
        rewriter.insertFragmentAfter(m_generatorFrameData->m_point, functor);
        rewriter.removeBytecode(m_generatorFrameData->m_point);
    }

    rewriter.execute();
}
} // Anonymous

void performGeneratorification(BytecodeGenerator& bytecodeGenerator, CodeBlock* codeBlock, Vector<Instruction>& instructions, SymbolTable* generatorFrameSymbolTable, int generatorFrameSymbolTableIndex)
{
    BytecodeGeneratorification pass(bytecodeGenerator, codeBlock, instructions, generatorFrameSymbolTable, generatorFrameSymbolTableIndex);
    pass.run();
}

} // namespace JSC
