/*
 * Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "MacroAssembler.h"
#include "Opcode.h"
#include "PutByIdFlags.h"
#include "PropertySlot.h"
#include "PrivateFieldPutKind.h"
#include "SpecialPointer.h"
#include "SymbolTable.h"
#include "Structure.h"
#include "StructureChain.h"
#include "ToThisStatus.h"
#include <wtf/VectorTraits.h>

namespace JSC {

class ArrayAllocationProfile;
class ArrayProfile;
class ObjectAllocationProfile;
class RegisterID;
class WatchpointSet;
struct LLIntCallLinkInfo;
struct ValueProfile;

struct Instruction {
	ALWAYS_INLINE Instruction()
	{
        u.pointer = nullptr;
	}

	ALWAYS_INLINE Instruction(Opcode opcode)
	{
#if !ENABLE(COMPUTED_GOTO_OPCODES)
		// We have to initialize one of the pointer members to ensure that
		// the entire struct is initialized, when opcode is not a pointer.
        u.pointer = nullptr;
#endif
		u.opcode = opcode;
	}

	ALWAYS_INLINE Instruction(int operand)
	{
		// We have to initialize one of the pointer members to ensure that
		// the entire struct is initialized in 64-bit.
        u.pointer = nullptr;
		u.operand = operand;
	}
    ALWAYS_INLINE Instruction(unsigned unsignedValue)
    {
        // We have to initialize one of the pointer members to ensure that
        // the entire struct is initialized in 64-bit.
        u.pointer = nullptr;
        u.unsignedValue = unsignedValue;
    }

    ALWAYS_INLINE Instruction(ECMAMode ecmaMode)
    {
        u.pointer = nullptr;
        u.operand = ecmaMode.value();
    }

    ALWAYS_INLINE Instruction(PutByIdFlags flags)
    {
        u.pointer = nullptr;
        u.operand = PutByIdFlags::convert(flags);
    }

    ALWAYS_INLINE Instruction(PrivateFieldPutKind flags)
    {
        u.pointer = nullptr;
        u.operand = flags.value();
    }

    ALWAYS_INLINE Instruction(JSGlobalData& globalData, JSCell* owner, Structure* structure)
	{
        u.pointer = nullptr;
		u.structure.set(globalData, owner, structure);
	}
    ALWAYS_INLINE Instruction(JSGlobalData& globalData, JSCell* owner, StructureChain* structureChain)
	{
        u.pointer = nullptr;
		u.structureChain.set(globalData, owner, structureChain);
	}
    ALWAYS_INLINE Instruction(JSGlobalData& globalData, JSCell* owner, JSCell* jsCell)
	{
        u.pointer = nullptr;
		u.jsCell.set(globalData, owner, jsCell);
	}

    ALWAYS_INLINE Instruction(PropertySlot::GetValueFunc getterFunc) { u.getterFunc = getterFunc; }

    ALWAYS_INLINE Instruction(LLIntCallLinkInfo* callLinkInfo) { u.callLinkInfo = callLinkInfo; }
    ALWAYS_INLINE Instruction(ValueProfile* profile) { u.profile = profile; }
    ALWAYS_INLINE Instruction(ArrayProfile* profile) { u.arrayProfile = profile; }
    ALWAYS_INLINE Instruction(ArrayAllocationProfile* profile) { u.arrayAllocationProfile = profile; }
	ALWAYS_INLINE Instruction(ObjectAllocationProfile* profile) { u.objectAllocationProfile = profile; }
	ALWAYS_INLINE Instruction(WriteBarrier<Unknown>* variablePointer) { u.variablePointer = variablePointer; }
	ALWAYS_INLINE Instruction(Special::Pointer pointer) { u.specialPointer = pointer; }
    ALWAYS_INLINE Instruction(UniquedStringImpl* uid) { u.uid = uid; }
    ALWAYS_INLINE Instruction(bool* predicatePointer) { u.predicatePointer = predicatePointer; }

#if !ENABLE(COMPUTED_GOTO_OPCODES)
    ALWAYS_INLINE Instruction(void* pointer) { u.pointer = pointer; }
#endif

    ALWAYS_INLINE Instruction(WatchpointSet* watchpointSet) { u.watchpointSet = watchpointSet; }

	union {
		Opcode opcode;
		int operand;
		unsigned unsignedValue;
		WriteBarrierBase<Structure> structure;
        WriteBarrierBase<SymbolTable> symbolTable;
		WriteBarrierBase<StructureChain> structureChain;
		WriteBarrierBase<JSCell> jsCell;
		WriteBarrier<Unknown>* variablePointer;
		Special::Pointer specialPointer;
		PropertySlot::GetValueFunc getterFunc;
		LLIntCallLinkInfo* callLinkInfo;
		UniquedStringImpl* uid;
		ValueProfile* profile;
		ArrayProfile* arrayProfile;
        ArrayAllocationProfile* arrayAllocationProfile;
        ObjectAllocationProfile* objectAllocationProfile;
        WatchpointSet* watchpointSet;
		void* pointer;
		bool* predicatePointer;
        ToThisStatus toThisStatus;
	} u;

private:
    // Avoid accidental use
    Instruction(RegisterID*) = delete;
    Instruction(RegisterID&) = delete;
	Instruction(StructureChain*) = delete;
	Instruction(Structure*) = delete;
};

} // namespace JSC

namespace WTF {

template<> struct VectorTraits<JSC::Instruction> : VectorTraitsBase<true, JSC::Instruction> { };

} // namespace WTF
