/*
 * Copyright (C) 2015 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "ObjectPropertyCondition.h"

#include "JSCInlines.h"
// #include "TrackedReferences.h"

namespace JSC {

void ObjectPropertyCondition::dumpInContext(PrintStream& out, DumpContext* context) const
{
#ifndef NDEBUG
    if (!*this) {
        out.print("<invalid>");
        return;
    }
    
    out.print("<", inContext(JSValue(m_object), context), ": ", inContext(m_condition, context), ">");
#endif
}

void ObjectPropertyCondition::dump(PrintStream& out) const
{
    dumpInContext(out, 0);
}

bool ObjectPropertyCondition::structureEnsuresValidityAssumingImpurePropertyWatchpoint(VM& vm) const
{
	if(!m_object)
		return false;

    return m_condition.isStillValidAssumingImpurePropertyWatchpoint(vm, m_object->structure(), 0);
}

bool ObjectPropertyCondition::validityRequiresImpurePropertyWatchpoint(Structure* structure) const
{
    return m_condition.validityRequiresImpurePropertyWatchpoint(structure);
}

bool ObjectPropertyCondition::validityRequiresImpurePropertyWatchpoint() const
{
	if(!m_object)
		return false;

    return m_condition.validityRequiresImpurePropertyWatchpoint(m_object->structure());
}

bool ObjectPropertyCondition::isStillValidAssumingImpurePropertyWatchpoint(VM& vm, Structure* structure) const
{
    return m_condition.isStillValidAssumingImpurePropertyWatchpoint(vm, structure, m_object);
}

bool ObjectPropertyCondition::isStillValidAssumingImpurePropertyWatchpoint(VM& vm) const
{
	if(!m_object)
		return false;

    return isStillValidAssumingImpurePropertyWatchpoint(vm, m_object->structure());
}

bool ObjectPropertyCondition::isStillValid(VM& vm, Structure* structure) const
{
    return m_condition.isStillValid(vm, structure, m_object);
}

bool ObjectPropertyCondition::isStillValid(VM& vm) const
{
	if(!m_object)
		return false;

    return isStillValid(vm, m_object->structure());
}

bool ObjectPropertyCondition::structureEnsuresValidity(VM& vm, Structure* structure) const
{
    return m_condition.isStillValid(vm, structure);
}

bool ObjectPropertyCondition::structureEnsuresValidity(VM& vm) const
{
	if(!m_object)
		return false;

    return structureEnsuresValidity(vm, m_object->structure());
}

bool ObjectPropertyCondition::isWatchableAssumingImpurePropertyWatchpoint(
	VM& vm, Structure* structure, PropertyCondition::WatchabilityEffort effort) const
{
    return m_condition.isWatchableAssumingImpurePropertyWatchpoint(vm, structure, m_object, effort);
}

bool ObjectPropertyCondition::isWatchableAssumingImpurePropertyWatchpoint(
	VM& vm, PropertyCondition::WatchabilityEffort effort) const
{
	if(!m_object)
		return false;

    return isWatchableAssumingImpurePropertyWatchpoint(vm, m_object->structure(), effort);
}

bool ObjectPropertyCondition::isWatchable(
	VM& vm, Structure* structure, PropertyCondition::WatchabilityEffort effort) const
{
    return m_condition.isWatchable(vm, structure, m_object, effort);
}

bool ObjectPropertyCondition::isWatchable(VM& vm, PropertyCondition::WatchabilityEffort effort) const
{
	if(!m_object)
		return false;

    return isWatchable(vm, m_object->structure(), effort);
}

bool ObjectPropertyCondition::isStillLive() const
{
    if (!*this)
        return false;

    bool isStillLive = true;
    forEachDependentCell([&](JSCell* cell) {
        isStillLive &= Heap::isMarked(cell);
    });
    return isStillLive;
}

#if 0
void ObjectPropertyCondition::validateReferences(const TrackedReferences& tracked) const
{
    if (!*this)
        return;
    
    tracked.check(m_object);
    m_condition.validateReferences(tracked);
}
#endif

ObjectPropertyCondition ObjectPropertyCondition::attemptToMakeEquivalenceWithoutBarrier() const
{
    PropertyCondition result = condition().attemptToMakeEquivalenceWithoutBarrier(object());
    if (!result)
        return ObjectPropertyCondition();
    return ObjectPropertyCondition(object(), result);
}

} // namespace JSC

