/*
 * Copyright (C) 2015-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "ObjectPropertyConditionSet.h"

#include "JSCInlines.h"
#include <wtf/ListDump.h>

namespace JSC {

ObjectPropertyCondition ObjectPropertyConditionSet::forObject(JSObject* object) const
{
	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it)
			if (it->object() == object)
				return *it;
    }
    return ObjectPropertyCondition();
}

ObjectPropertyCondition ObjectPropertyConditionSet::forConditionKind(
    PropertyCondition::Kind kind) const
{
	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it)
			if (it->kind() == kind)
				return *it;
    }
    return ObjectPropertyCondition();
}

unsigned ObjectPropertyConditionSet::numberOfConditionsWithKind(PropertyCondition::Kind kind) const
{
    unsigned result = 0;
	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it)
			if (it->kind() == kind)
				result++;
    }
    return result;
}

bool ObjectPropertyConditionSet::hasOneSlotBaseCondition() const
{
    bool sawBase = false;
	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it) {
			const ObjectPropertyCondition& condition = *it;
			switch (condition.kind()) {
			case PropertyCondition::Presence:
			case PropertyCondition::Equivalence:
			case PropertyCondition::HasStaticProperty:
				if (sawBase)
					return false;
				sawBase = true;
				break;
			default:
				break;
			}
        }
    }

    return sawBase;
}

ObjectPropertyCondition ObjectPropertyConditionSet::slotBaseCondition() const
{
    ObjectPropertyCondition result;
    unsigned numFound = 0;
	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it) {
			const ObjectPropertyCondition& condition = *it;
	        if (condition.kind() == PropertyCondition::Presence
				|| condition.kind() == PropertyCondition::Equivalence
	            || condition.kind() == PropertyCondition::HasStaticProperty) {
				result = *it;
				numFound++;
			}
		}
    }
    RELEASE_ASSERT(numFound == 1);
    return result;
}

ObjectPropertyConditionSet ObjectPropertyConditionSet::mergedWith(
    const ObjectPropertyConditionSet& other) const
{
    if (!isValid() || !other.isValid())
        return invalid();

    Vector<ObjectPropertyCondition, 16> result;
    
    if (!isEmpty())
        result.appendVector(m_data->vector);

    if(other.m_data) {
		for(iterator newIter = other.m_data->vector.begin(); newIter != other.m_data->vector.end(); ++newIter) {
			const ObjectPropertyCondition& newCondition = *newIter;
			bool foundMatch = false;
			if(this->m_data) {
				for(iterator existingIter = this->m_data->vector.begin(); existingIter != this->m_data->vector.end(); ++existingIter) {
					const ObjectPropertyCondition& existingCondition = *existingIter;
					if (newCondition == existingCondition) {
						foundMatch = true;
						continue;
					}
					if (!newCondition.isCompatibleWith(existingCondition))
						return invalid();
				}
			}
			if (!foundMatch)
				result.append(newCondition);
		}
    }

    return create(result);
}

bool ObjectPropertyConditionSet::structuresEnsureValidity(VM& vm) const
{
    if (!isValid())
        return false;

	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it)
			if (!it->structureEnsuresValidity(vm))
				return false;
    }
    return true;
}

bool ObjectPropertyConditionSet::structuresEnsureValidityAssumingImpurePropertyWatchpoint(VM& vm) const
{
    if (!isValid())
        return false;

	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it)
			if (!it->structureEnsuresValidityAssumingImpurePropertyWatchpoint(vm))
				return false;
    }
    return true;
}

bool ObjectPropertyConditionSet::needImpurePropertyWatchpoint() const
{
	if(Data* data = m_data.get()) {
		for(iterator it = data->vector.begin(); it != data->vector.end(); ++it)
			if (it->validityRequiresImpurePropertyWatchpoint())
				return true;
    }
    return false;
}

bool ObjectPropertyConditionSet::areStillLive() const
{
    bool stillLive = true;
    forEachDependentCell([&](JSCell* cell) {
        stillLive &= Heap::isMarked(cell);
    });
    return stillLive;
}

void ObjectPropertyConditionSet::dumpInContext(PrintStream& out, DumpContext* context) const
{
#ifndef NDEBUG
    if (!isValid()) {
        out.print("<invalid>");
        return;
    }
    
    out.print("[");
    // if (m_data)
    //    out.print(listDumpInContext(m_data->vector, context));
    out.print("]");
#endif
}

void ObjectPropertyConditionSet::dump(PrintStream& out) const
{
    dumpInContext(out, 0);
}

namespace {

namespace ObjectPropertyConditionSetInternal {
static const bool verbose = false;
}

static ObjectPropertyCondition generateCondition(
    VM& vm, JSCell* owner, JSObject* object, UniquedStringImpl* uid, PropertyCondition::Kind conditionKind)
{
    Structure* structure = object->structure();
    if (ObjectPropertyConditionSetInternal::verbose)
        dataLog("Creating condition ", conditionKind, " for ", pointerDump(structure), "\n");

    ObjectPropertyCondition result;
    switch (conditionKind) {
    case PropertyCondition::Presence: {
        unsigned attributes;
        PropertyOffset offset = structure->getConcurrently(vm, uid, attributes);
        if (offset == invalidOffset)
            return ObjectPropertyCondition();
        result = ObjectPropertyCondition::presence(vm, owner, object, uid, offset, attributes);
        break;
    }
    case PropertyCondition::Absence: {
        result = ObjectPropertyCondition::absence(
            vm, owner, object, uid, object->structure()->storedPrototypeObject());
        break;
    }
    case PropertyCondition::AbsenceOfSetEffect: {
        result = ObjectPropertyCondition::absenceOfSetEffect(
            vm, owner, object, uid, object->structure()->storedPrototypeObject());
        break;
    }
    case PropertyCondition::Equivalence: {
        unsigned attributes;
        PropertyOffset offset = structure->getConcurrently(vm, uid, attributes);
        if (offset == invalidOffset)
            return ObjectPropertyCondition();
        JSValue value = object->getDirect(offset);
        result = ObjectPropertyCondition::equivalence(vm, owner, object, uid, value);
        break;
    }
    case PropertyCondition::HasStaticProperty: {
        auto entry = object->findPropertyHashEntry(vm, uid);
        if (!entry)
            return ObjectPropertyCondition();
        result = ObjectPropertyCondition::hasStaticProperty(vm, owner, object, uid);
        break;
    }
    default:
        RELEASE_ASSERT_NOT_REACHED();
        return ObjectPropertyCondition();
    }

    if (!result.isStillValidAssumingImpurePropertyWatchpoint(vm)) {
        if (ObjectPropertyConditionSetInternal::verbose)
            dataLog("Failed to create condition: ", result, "\n");
        return ObjectPropertyCondition();
    }

    if (ObjectPropertyConditionSetInternal::verbose)
        dataLog("New condition: ", result, "\n");
    return result;
}

template<typename Functor>
static ObjectPropertyConditionSet generateConditions(
    VM& vm, JSGlobalObject* globalObject, Structure* structure, JSObject* prototype, const Functor& functor)
{
    Vector<ObjectPropertyCondition, 8> conditions;
    
    for (;;) {
        if (ObjectPropertyConditionSetInternal::verbose)
            dataLog("Considering structure: ", pointerDump(structure), "\n");
        
        if (structure->isProxy()) {
            if (ObjectPropertyConditionSetInternal::verbose)
                dataLog("It's a proxy, so invalid.\n");
            return ObjectPropertyConditionSet::invalid();
        }
        
        JSValue value = structure->prototypeForLookup(globalObject);
        
        if (value.isNull()) {
            if (!prototype) {
                if (ObjectPropertyConditionSetInternal::verbose)
                    dataLog("Reached end of prototype chain as expected, done.\n");
                break;
            }
            if (ObjectPropertyConditionSetInternal::verbose)
                dataLog("Unexpectedly reached end of prototype chain, so invalid.\n");
            return ObjectPropertyConditionSet::invalid();
        }
        
        JSObject* object = jsCast<JSObject*>(value);
        structure = object->structure(vm);
        
        if (structure->isDictionary()) {
            if (ObjectPropertyConditionSetInternal::verbose)
                dataLog("Cannot cache dictionary.\n");
            return ObjectPropertyConditionSet::invalid();
        }

        if (!functor(conditions, object)) {
            if (ObjectPropertyConditionSetInternal::verbose)
                dataLog("Functor failed, invalid.\n");
            return ObjectPropertyConditionSet::invalid();
        }
        
        if (object == prototype) {
            if (ObjectPropertyConditionSetInternal::verbose)
                dataLog("Reached desired prototype, done.\n");
            break;
        }
    }

    if (ObjectPropertyConditionSetInternal::verbose)
        dataLog("Returning conditions: ", listDump(conditions), "\n");
    return ObjectPropertyConditionSet::create(conditions);
}

} // anonymous namespace

ObjectPropertyConditionSet generateConditionsForPropertyMiss(
    VM& vm, JSCell* owner, ExecState* exec, Structure* headStructure, UniquedStringImpl* uid)
{
    return generateConditions(
        vm, exec->lexicalGlobalObject(), headStructure, 0,
		[&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) -> bool {
			ObjectPropertyCondition result =
					generateCondition(vm, owner, object, uid, PropertyCondition::Absence);
				if (!result)
					return false;
				conditions.append(result);
				return true;
		});
}

ObjectPropertyConditionSet generateConditionsForPropertySetterMiss(
    VM& vm, JSCell* owner, ExecState* exec, Structure* headStructure, UniquedStringImpl* uid)
{
    return generateConditions(
        vm, exec->lexicalGlobalObject(), headStructure, 0,
		[&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) -> bool {
			ObjectPropertyCondition result =
				generateCondition(vm, owner, object, uid, PropertyCondition::AbsenceOfSetEffect);
			if (!result)
				return false;
			conditions.append(result);
			return true;
		});
}

ObjectPropertyConditionSet generateConditionsForPrototypePropertyHit(
    VM& vm, JSCell* owner, ExecState* exec, Structure* headStructure, JSObject* prototype,
    UniquedStringImpl* uid)
{
    return generateConditions(
        vm, exec->lexicalGlobalObject(), headStructure, prototype,
		[&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) -> bool {
	        PropertyCondition::Kind kind =
	            object == prototype ? PropertyCondition::Presence : PropertyCondition::Absence;
	        ObjectPropertyCondition result =
	            generateCondition(vm, owner, object, uid, kind);
	        if (!result)
	            return false;
	        conditions.append(result);
	        return true;
		});
}

ObjectPropertyConditionSet generateConditionsForPrototypePropertyHitCustom(
    VM& vm, JSCell* owner, ExecState* exec, Structure* headStructure, JSObject* prototype,
    UniquedStringImpl* uid, unsigned attributes)
{
    return generateConditions(
        vm, exec->lexicalGlobalObject(), headStructure, prototype,
		[&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) ->bool {
		auto kind = PropertyCondition::Absence;
		if (object == prototype) {
			Structure* structure = object->structure(vm);
			PropertyOffset offset = structure->get(vm, uid);
			if (isValidOffset(offset)) {
				// When we reify custom accessors, we wrap them in a JSFunction that we shove
				// inside a GetterSetter. So, once we've reified a custom accessor, we will
				// no longer see it as a "custom" accessor/value. Hence, if our property access actually
				// notices a custom, it must be a CustomGetterSetterType cell or something
				// in the static property table. Custom values get reified into CustomGetterSetters.
				JSValue value = object->getDirect(offset);

				if (!value.isCell() || value.asCell()->structure()->typeInfo().type() != CustomGetterSetterType) {
					// The value could have just got changed to some other type, so check if it's still
					// a custom getter setter.
					return false;
				}

				kind = PropertyCondition::Equivalence;
			} else if (structure->findPropertyHashEntry(uid))
				kind = PropertyCondition::HasStaticProperty;
			else if (attributes & DontDelete) {
				// This can't change, so we can blindly cache it.
				return true;
			} else {
				// This means we materialized a custom out of thin air and it's not DontDelete (i.e, it can be
				// redefined). This is curious. We don't actually need to crash here. We could blindly cache
				// the function. Or we could blindly not cache it. However, we don't actually do this in WebKit
				// right now, so it's reasonable to decide what to do later (or to warn people of forgetting DoneDelete.)
				ASSERT_NOT_REACHED();
				return false;
			}
		}
		ObjectPropertyCondition result = generateCondition(vm, owner, object, uid, kind);
		if (!result)
			return false;
		conditions.append(result);
		return true;
	});
}

ObjectPropertyConditionSet generateConditionsForInstanceOf(
    VM& vm, JSCell* owner, ExecState* exec, Structure* headStructure, JSObject* prototype,
    bool shouldHit)
{
    bool didHit = false;
    if (ObjectPropertyConditionSetInternal::verbose)
        dataLog("Searching for prototype ", JSValue(prototype), " starting with structure ", RawPointer(headStructure), " with shouldHit = ", shouldHit, "\n");
    ObjectPropertyConditionSet result = generateConditions(
        vm, exec->lexicalGlobalObject(), headStructure, shouldHit ? prototype : 0,
        [&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) -> bool {
            if (ObjectPropertyConditionSetInternal::verbose)
                dataLog("Encountered object: ", RawPointer(object), "\n");
            if (object == prototype) {
                RELEASE_ASSERT(shouldHit);
                didHit = true;
                return true;
            }

            Structure* structure = object->structure(vm);
            if (structure->hasPolyProto())
                return false;
            conditions.append(
                ObjectPropertyCondition::hasPrototype(
                    vm, owner, object, structure->storedPrototypeObject()));
            return true;
        });
    if (result.isValid()) {
        if (ObjectPropertyConditionSetInternal::verbose)
            dataLog("didHit = ", didHit, ", shouldHit = ", shouldHit, "\n");
        RELEASE_ASSERT(didHit == shouldHit);
    }
    return result;
}

ObjectPropertyConditionSet generateConditionsForPrototypeEquivalenceConcurrently(
    VM& vm, JSGlobalObject* globalObject, Structure* headStructure, JSObject* prototype, UniquedStringImpl* uid)
{
	return generateConditions(vm, globalObject, headStructure, prototype,
		[&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) -> bool {
            PropertyCondition::Kind kind =
                object == prototype ? PropertyCondition::Equivalence : PropertyCondition::Absence;
            ObjectPropertyCondition result = generateCondition(vm, 0, object, uid, kind);
            if (!result)
                return false;
            conditions.append(result);
            return true;
        });
}

ObjectPropertyConditionSet generateConditionsForPropertyMissConcurrently(
    VM& vm, JSGlobalObject* globalObject, Structure* headStructure, UniquedStringImpl* uid)
{
    return generateConditions(
        vm, globalObject, headStructure, 0,
		[&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) -> bool {
			ObjectPropertyCondition result = generateCondition(vm, 0, object, uid, PropertyCondition::Absence);
			if (!result)
				return false;
			conditions.append(result);
			return true;
		});
}

ObjectPropertyConditionSet generateConditionsForPropertySetterMissConcurrently(
    VM& vm, JSGlobalObject* globalObject, Structure* headStructure, UniquedStringImpl* uid)
{
    return generateConditions(
        vm, globalObject, headStructure, 0,
		[&] (Vector<ObjectPropertyCondition, 8>& conditions, JSObject* object) -> bool {
	        ObjectPropertyCondition result =
	            generateCondition(vm, 0, object, uid, PropertyCondition::AbsenceOfSetEffect);
	        if (!result)
	            return false;
	        conditions.append(result);
	        return true;
		});
}

ObjectPropertyCondition generateConditionForSelfEquivalence(
    VM& vm, JSCell* owner, JSObject* object, UniquedStringImpl* uid)
{
    return generateCondition(vm, owner, object, uid, PropertyCondition::Equivalence);
}

// Current might be null. Structure can't be null.
static Optional<PrototypeChainCachingStatus> prepareChainForCaching(JSGlobalObject* globalObject, JSCell* current, Structure* structure, JSObject* target)
{
    ASSERT(structure);
    VM& vm = globalObject->vm();

    bool found = false;
    bool usesPolyProto = false;
    bool flattenedDictionary = false;

    while (true) {
        if (structure->isDictionary()) {
            if (!current)
                return Nullopt;

            ASSERT(structure->isObject());
            if (structure->hasBeenFlattenedBefore())
                return Nullopt;

            structure->flattenDictionaryStructure(vm, asObject(current));
            flattenedDictionary = true;
        }

        if (!structure->propertyAccessesAreCacheable())
            return Nullopt;

        if (structure->isProxy())
            return Nullopt;

        if (current && current == target) {
            found = true;
            break;
        }

        // We only have poly proto if we need to access our prototype via
        // the poly proto protocol. If the slot base is the only poly proto
        // thing in the chain, and we have a cache hit on it, then we're not
        // poly proto.
        JSValue prototype;
        if (structure->hasPolyProto()) {
            if (!current)
                return Nullopt;
            usesPolyProto = true;
            prototype = structure->prototypeForLookup(globalObject/*, current*/);
        } else
            prototype = structure->prototypeForLookup(globalObject);

        if (prototype.isNull())
            break;
        current = asObject(prototype);
        structure = current->structure(vm);
    }

    if (!found && !!target)
        return Nullopt;

    PrototypeChainCachingStatus result;
    result.usesPolyProto = usesPolyProto;
    result.flattenedDictionary = flattenedDictionary;

    return result;
}

Optional<PrototypeChainCachingStatus> prepareChainForCaching(JSGlobalObject* globalObject, JSCell* base, JSObject* target)
{
    return prepareChainForCaching(globalObject, base, base->structure(globalObject->vm()), target);
}

Optional<PrototypeChainCachingStatus> prepareChainForCaching(JSGlobalObject* globalObject, JSCell* base, const PropertySlot& slot)
{
    JSObject* target = slot.isUnset() ? 0 : slot.slotBase();
    return prepareChainForCaching(globalObject, base, target);
}

Optional<PrototypeChainCachingStatus> prepareChainForCaching(JSGlobalObject* globalObject, Structure* baseStructure, JSObject* target)
{
    return prepareChainForCaching(globalObject, 0, baseStructure, target);
}

} // namespace JSC

