/*
 * Copyright (C) 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "PutByIdStatus.h"

#include "CCallHelpers.h"
#include "CodeBlock.h"
#include "ComplexGetStatus.h"
#include "GetterSetterAccessCase.h"
#include "ICStatusUtils.h"
#include "LowLevelInterpreter.h"
#include "Operations.h"
#include "PolymorphicAccess.h"
#include "Structure.h"
#include "StructureChain.h"
#include <wtf/ListDump.h>

namespace JSC {

bool PutByIdStatus::appendVariant(const PutByIdVariant& variant)
{
	return appendICStatusVariant(m_variants, variant);
}

bool PutByIdStatus::hasExitSite(CodeBlock* profiledBlock, unsigned bytecodeIndex)
{
#if ENABLE(DFG_JIT)
    return profiledBlock->hasExitSite(DFG::FrequentExitSite(bytecodeIndex, BadCache))
        || profiledBlock->hasExitSite(DFG::FrequentExitSite(bytecodeIndex, BadWeakConstantCache));
#else
    return false;
#endif
}

PutByIdStatus PutByIdStatus::computeFromLLInt(CodeBlock* profiledBlock, unsigned bytecodeIndex, UniquedStringImpl* uid)
{
    UNUSED_PARAM(profiledBlock);
    UNUSED_PARAM(bytecodeIndex);
    UNUSED_PARAM(uid);
#if ENABLE(LLINT)
    Instruction* instruction = profiledBlock->instructions().begin() + bytecodeIndex;

    // We are not yet using `computeFromLLInt` in any place for `put_private_name`.
    // We can add support for it if this is required in future changes, since we have
    // IC implemented for this operation on LLInt.
    ASSERT(!instruction->is<OpPutPrivateName>());

    Structure* structure = instruction[4].u.structure.get();
    if (!structure)
        return PutByIdStatus(NoInformation);
    
    if (instruction[0].u.opcode == llint_op_put_by_id
        || instruction[0].u.opcode == llint_op_put_by_id_out_of_line) {
        PropertyOffset offset = structure->getConcurrently(*profiledBlock->globalData(), uid);
        if (!isValidOffset(offset))
            return PutByIdStatus(NoInformation);

        return PutByIdVariant::replace(structure, offset);
    }
    
    ASSERT(structure->transitionWatchpointSetHasBeenInvalidated());

    Structure* newStructure = instruction[6].u.structure.get();
    
    PropertyOffset offset = newStructure->getConcurrently(*profiledBlock->globalData(), uid);
    if (!isValidOffset(offset))
        return PutByIdStatus(NoInformation);

    ObjectPropertyConditionSet conditionSet;
    if (!(bytecode.m_flags.isDirect())) {
        conditionSet =
            generateConditionsForPropertySetterMissConcurrently(
                *profiledBlock->vm(), profiledBlock->globalObject(), structure, uid);
        if (!conditionSet.isValid())
            return PutByIdStatus(NoInformation);
    }

    return PutByIdVariant::transition(structure, newStructure, conditionSet, offset);
#else
    return PutByIdStatus(NoInformation);
#endif
}

PutByIdStatus PutByIdStatus::computeFor(CodeBlock* profiledBlock, StubInfoMap& map, unsigned bytecodeIndex, UniquedStringImpl* uid)
{
    UNUSED_PARAM(profiledBlock);
    UNUSED_PARAM(bytecodeIndex);
    UNUSED_PARAM(uid);
    if (hasExitSite(profiledBlock, bytecodeIndex))
        return PutByIdStatus(TakesSlowPath);

    StructureStubInfo* stubInfo = map.get(CodeOrigin(bytecodeIndex));
    PutByIdStatus result = computeForStubInfo(
		profiledBlock, stubInfo, uid,
		CallLinkStatus::computeExitSiteData(profiledBlock, bytecodeIndex));
    if (!result)
        return computeFromLLInt(profiledBlock, bytecodeIndex, uid);

    return result;
}

PutByIdStatus PutByIdStatus::computeForStubInfo(CodeBlock* baselineBlock, StructureStubInfo* stubInfo, CodeOrigin codeOrigin, UniquedStringImpl* uid)
{
    return computeForStubInfo(
        baselineBlock, stubInfo, uid,
        CallLinkStatus::computeExitSiteData(baselineBlock, codeOrigin.bytecodeIndex));
}

PutByIdStatus PutByIdStatus::computeForStubInfo(
	CodeBlock* profiledBlock, StructureStubInfo* stubInfo,
	UniquedStringImpl* uid, CallLinkStatus::ExitSiteData callExitSiteData)
{
	if (!stubInfo || !stubInfo->everConsidered)
        return PutByIdStatus();

    if (stubInfo->tookSlowPath)
        return PutByIdStatus(TakesSlowPath);

    switch (stubInfo->cacheType) {
    case CacheType::Unset:
    	// This means that we attempted to cache but failed for some reason.
        return PutByIdStatus(TakesSlowPath);

    case CacheType::PutByIdReplace: {
        PropertyOffset offset =
        	stubInfo->inlineAccessBaseStructure->getConcurrently(
        		profiledBlock->globalData(), uid);
        if (isValidOffset(offset)) {
            return PutByIdVariant::replace(
                stubInfo->inlineAccessBaseStructure.get(), offset);
        }
        return PutByIdStatus(TakesSlowPath);
    }

    case CacheType::Stub: {
        PolymorphicAccess* list = stubInfo->m_stub.get();

        PutByIdStatus result;
        result.m_state = Simple;

        State slowPathState = TakesSlowPath;
        for (unsigned i = 0; i < list->size(); ++i) {
            const AccessCase& access = list->at(i);
            if (access.doesCalls())
                slowPathState = MakesCalls;
        }
        
        for (unsigned i = 0; i < list->size(); ++i) {
            const AccessCase& access = list->at(i);
            if (access.viaProxy())
                return PutByIdStatus(slowPathState);

            PutByIdVariant variant;

            switch (access.type()) {
            case AccessCase::Replace: {
                Structure* structure = access.structure();
                PropertyOffset offset = structure->getConcurrently(profiledBlock->vm(), uid);
                if (!isValidOffset(offset))
					return PutByIdStatus(slowPathState);
				variant = PutByIdVariant::replace(structure, offset);
				break;
            }

            case AccessCase::Transition: {
                PropertyOffset offset =
                    access.newStructure()->getConcurrently(profiledBlock->vm(), uid);
                if (!isValidOffset(offset))
                    return PutByIdStatus(slowPathState);
                ObjectPropertyConditionSet conditionSet = access.conditionSet();
                if (!conditionSet.structuresEnsureValidity(profiledBlock->vm()))
                    return PutByIdStatus(slowPathState);
                variant = PutByIdVariant::transition(
                    access.structure(), access.newStructure(), conditionSet, offset);
                break;
            }

            case AccessCase::Setter: {
#if ENABLE(FTL_JIT)
                Structure* structure = access.structure();

                ComplexGetStatus complexGetStatus = ComplexGetStatus::computeFor(
					profiledBlock->vm(), structure, access.conditionSet(), uid);

                switch (complexGetStatus.kind()) {
                case ComplexGetStatus::ShouldSkip:
                    continue;

                case ComplexGetStatus::TakesSlowPath:
                    return PutByIdStatus(slowPathState);

                case ComplexGetStatus::Inlineable: {
                    OwnPtr<CallLinkStatus> callLinkStatus =
                        adoptPtr(new CallLinkStatus();
                    if (CallLinkInfo* callLinkInfo = access.as<GetterSetterAccessCase>().callLinkInfo()) {
                        *callLinkStatus = CallLinkStatus::computeFor(
                            locker, profiledBlock, *callLinkInfo, callExitSiteData);
                    }

                    variant = PutByIdVariant::setter(
                        structure, complexGetStatus.offset(),
                        complexGetStatus.conditionSet(),
						callLinkStatus.release());
                } }
                break;
#else
                return PutByIdStatus(slowPathState);
#endif
            }

            case AccessCase::CustomValueSetter:
            case AccessCase::CustomAccessorSetter:
                return PutByIdStatus(MakesCalls);

            default:
                return PutByIdStatus(slowPathState);
            }

            if (!result.appendVariant(variant))
                return PutByIdStatus(slowPathState);
        }

        return result;
    }

    default:
        // FIXME: We should handle polymorphic PutById. We probably have some interesting things
        // we could do about it.
        return PutByIdStatus(TakesSlowPath);
    }
}

PutByIdStatus PutByIdStatus::computeFor(CodeBlock* baselineBlock, CodeBlock* dfgBlock, StubInfoMap& baselineMap, StubInfoMap& dfgMap, CodeOrigin codeOrigin, UniquedStringImpl* uid)
{
    if (dfgBlock) {
        CallLinkStatus::ExitSiteData exitSiteData;
        {
            //ConcurrentJITLocker locker(baselineBlock->m_lock);
            if (hasExitSite(baselineBlock, codeOrigin.bytecodeIndex))
                return PutByIdStatus(TakesSlowPath);
            exitSiteData = CallLinkStatus::computeExitSiteData(
                baselineBlock, codeOrigin.bytecodeIndex);
        }

        PutByIdStatus result;
        {
            //ConcurrentJITLocker locker(dfgBlock->m_lock);
            result = computeForStubInfo(
                dfgBlock, dfgMap.get(codeOrigin), uid, exitSiteData);
        }

        if (result.isSet())
            return result;
    }

    return computeFor(baselineBlock, baselineMap, codeOrigin.bytecodeIndex, uid);
}

PutByIdStatus PutByIdStatus::computeFor(JSGlobalData& globalData, JSGlobalObject* globalObject, const StructureSet& set, UniquedStringImpl* uid, bool isDirect, PrivateFieldPutKind privateFieldPutKind)
{
    if (parseIndex(*uid))
        return PutByIdStatus(TakesSlowPath);

    if (set.isEmpty())
        return PutByIdStatus();

    PutByIdStatus result;
    result.m_state = Simple;
    for (unsigned i = 0; i < set.size(); ++i) {
        Structure* structure = set[i];

		if (structure->typeInfo().overridesGetOwnPropertySlot() && structure->typeInfo().type() != GlobalObjectType)
			return PutByIdStatus(TakesSlowPath);

		if (!structure->propertyAccessesAreCacheable())
			return PutByIdStatus(TakesSlowPath);

		unsigned attributes;
		PropertyOffset offset = structure->getConcurrently(globalData, uid, attributes);
		if (isValidOffset(offset)) {
            // We can't have a valid offset for structures on `PutPrivateNameById` define mode
            // since it means we are redefining a private field. In such case, we need to take
            // slow path to throw exception.
            if (privateFieldPutKind.isDefine())
                return PutByIdStatus(TakesSlowPath);

			if (attributes & CustomAccessorOrValue)
				return PutByIdStatus(MakesCalls);

			if (attributes & (Accessor | ReadOnly))
				return PutByIdStatus(TakesSlowPath);

            if (isDirect && attributes) {
#if 0
                Structure* existingTransition = Structure::attributeChangeTransitionToExistingStructureConcurrently(structure, identifier.uid(), 0, offset);
                if (!existingTransition)
                    return PutByStatus(LikelyTakesSlowPath);
                bool didAppend = result.appendVariant(PutByVariant::transition(identifier, structure, existingTransition, { }, offset));
                if (!didAppend)
                    return PutByStatus(LikelyTakesSlowPath);
                continue;
#else
				return PutByIdStatus(TakesSlowPath);
#endif
            }

			RefPtr<WatchpointSet> replaceSet = structure->propertyReplacementWatchpointSet(offset);
			if (!replaceSet || replaceSet->isStillValid()) {
				// When this executes, it'll create, and fire, this replacement watchpoint set.
				// That means that  this has probably never executed or that something fishy is
				// going on. Also, we cannot create or fire the watchpoint set from the concurrent
				// JIT thread, so even if we wanted to do this, we'd need to have a lazy thingy.
				// So, better leave this alone and take slow path.
				return PutByIdStatus(TakesSlowPath);
			}

			if (!result.appendVariant(PutByIdVariant::replace(structure, offset)))
				return PutByIdStatus(TakesSlowPath);
			continue;
		}

        // We can have a case with PutPrivateNameById in set mode and it
        // should never cause a structure transition because it means we are
        // trying to store in a not installed private field. We need to take
        // slow path to throw excpetion if it ever gets executed.
        if (privateFieldPutKind.isSet())
            return PutByIdStatus(TakesSlowPath);

		// Our hypothesis is that we're doing a transition. Before we prove that this is really
		// true, we want to do some sanity checks.

		// Don't cache put transitions on dictionaries.
		if (structure->isDictionary())
			return PutByIdStatus(TakesSlowPath);

		// If the structure corresponds to something that isn't an object, then give up, since
		// we don't want to be adding properties to strings.
		if (!structure->typeInfo().isObject())
			return PutByIdStatus(TakesSlowPath);

		ObjectPropertyConditionSet conditionSet;
		if (!isDirect) {
            ASSERT(privateFieldPutKind.isNone());
			conditionSet = generateConditionsForPropertySetterMissConcurrently(
				globalObject->vm(), globalObject, structure, uid);
			if (!conditionSet.isValid())
				return PutByIdStatus(TakesSlowPath);
		}

		// We only optimize if there is already a structure that the transition is cached to.
		Structure* transition = Structure::addPropertyTransitionToExistingStructure(structure, uid, 0, offset);
		if (!transition)
			return PutByIdStatus(TakesSlowPath);
		ASSERT(isValidOffset(offset));

        bool didAppend = result.appendVariant(
            PutByIdVariant::transition(structure, transition, conditionSet, offset));
        if (!didAppend)
            return PutByIdStatus(TakesSlowPath);
    }

    return result;
}

bool PutByIdStatus::makesCalls() const
{
    if (m_state == MakesCalls)
        return true;

    if (m_state != Simple)
        return false;

    for (unsigned i = m_variants.size(); i--;) {
        if (m_variants[i].makesCalls())
            return true;
    }

    return false;
}

void PutByIdStatus::dump(PrintStream& out) const
{
#ifndef NDEBUG
    switch (m_state) {
    case NoInformation:
        out.print("(NoInformation)");
        return;

    case Simple:
        out.print("(", listDump(m_variants), ")");
        return;

    case TakesSlowPath:
        out.print("(TakesSlowPath)");
        return;
    case MakesCalls:
        out.print("(MakesCalls)");
        return;
    }

    RELEASE_ASSERT_NOT_REACHED();
#endif
}

} // namespace JSC

