/*
 * Copyright (C) 2008, 2014-2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "StructureStubInfo.h"

#include "CCallHelpers.h"
#include "JSObject.h"
#include "PolymorphicAccess.h"
#include "JSScope.h"
#include "Repatch.h"
#include "StructureInlines.h"

namespace JSC {

namespace StructureStubInfoInternal {
static const bool verbose = false;
}

StructureStubInfo::StructureStubInfo(AccessType accessType)
	: accessType(accessType)
	, cacheType(CacheType::Unset)
	, countdown(1) // For a totally clear stub, we'll patch it after the initial count down.
	, repatchCount(0)
	, numberOfCoolDowns(0)
	, bufferingCountdown(Options::repatchBufferingCountdown())
	, resetByGC(false)
	, tookSlowPath(false)
	, everConsidered(false)
	, prototypeIsKnownObject(false)
	, byIdSelfOffset(0)
	, callSiteIndex(UINT_MAX)
{
}

StructureStubInfo::~StructureStubInfo()
{
}

void StructureStubInfo::initGetByIdSelf(CodeBlock* codeBlock, Structure* inlineAccessBaseStructure, PropertyOffset offset)
{
    cacheType = CacheType::GetByIdSelf;

    this->inlineAccessBaseStructure.set(codeBlock->vm(), codeBlock->ownerExecutable(), inlineAccessBaseStructure);
    byIdSelfOffset = offset;
}

void StructureStubInfo::initArrayLength()
{
    cacheType = CacheType::ArrayLength;
}

void StructureStubInfo::initStringLength()
{
    cacheType = CacheType::StringLength;
}

void StructureStubInfo::initPutByIdReplace(CodeBlock* codeBlock, Structure* inlineAccessBaseStructure, PropertyOffset offset)
{
    cacheType = CacheType::PutByIdReplace;

    this->inlineAccessBaseStructure.set(codeBlock->vm(), codeBlock->ownerExecutable(), inlineAccessBaseStructure);
    byIdSelfOffset = offset;
}

void StructureStubInfo::deref(CodeBlock* codeBlock)
{
    switch (cacheType) {
    case CacheType::Stub:
    	if(m_stub && m_stub->isExecutingCall()) {
    		WTF::logMessage(ASCIILiteral("StructureStubInfo::deref m_stub is still executing call."));

    		class DeleteStub final : public VMIdleCallback
    		{
    		public:
    			DeleteStub(PassOwnPtr<PolymorphicAccess> stub) : m_stub(stub) { }
    			virtual void operator()(VM&, JSGlobalObject*)
    			{
    				m_stub = nullptr;
    			}
    		private:
    			OwnPtr<PolymorphicAccess> m_stub;
    		};
    		codeBlock->vm().whenIdle(adoptPtr(new DeleteStub(m_stub.release())));
    	}
    	else
    		m_stub = nullptr;
        return;
    case CacheType::Unset:
    case CacheType::GetByIdSelf:
    case CacheType::PutByIdReplace:
    case CacheType::ArrayLength:
    case CacheType::StringLength:
        return;
    }
}

#if 0
void StructureStubInfo::aboutToDie()
{
    switch (cacheType) {
    case CacheType::Stub:
    	m_stub->aboutToDie();
        return;
    case CacheType::Unset:
    case CacheType::GetByIdSelf:
    case CacheType::PutByIdReplace:
    case CacheType::ArrayLength:
    case CacheType::StringLength:
        return;
    }

    RELEASE_ASSERT_NOT_REACHED();
}
#endif

AccessGenerationResult StructureStubInfo::addAccessCase(
    CodeBlock* codeBlock, ECMAMode ecmaMode, const Identifier& ident, PassOwnPtr<AccessCase> accessCase)
{
    VM& vm = codeBlock->vm();

    //if (StructureStubInfoInternal::verbose)
    //    dataLog("Adding access case: ", accessCase, "\n");

    if (!accessCase)
        return AccessGenerationResult(AccessGenerationResult::GaveUp);

    AccessGenerationResult result;

    if (cacheType == CacheType::Stub) {
        result = m_stub->addCase(vm, codeBlock, *this, ident, accessCase);

        //if (StructureStubInfoInternal::verbose)
        //    dataLog("Had stub, result: ", result, "\n");

        if (!result.buffered()) {
            bufferedStructures.clear();
            return result;
        }
    } else {
		OwnPtr<PolymorphicAccess> access = adoptPtr(new PolymorphicAccess());

		Vector<OwnPtr<AccessCase>, 2> accessCases;

		if(OwnPtr<AccessCase> previousCase = AccessCase::fromStructureStubInfo(vm, codeBlock->ownerExecutable(), *this))
			accessCases.append(previousCase.release());

		accessCases.append(accessCase);

		AccessGenerationResult result = access->addCases(vm, codeBlock, *this, ident, WTFMove(accessCases));

        //if (StructureStubInfoInternal::verbose)
        //    dataLog("Created stub, result: ", result, "\n");

        if (!result.buffered()) {
            bufferedStructures.clear();
            return result;
        }

        cacheType = CacheType::Stub;
        m_stub = access.release();
    }

    RELEASE_ASSERT(!result.generatedSomeCode());

    // If we didn't buffer any cases then bail. If this made no changes then we'll just try again
    // subject to cool-down.
    if (!result.buffered()) {
        //if (StructureStubInfoInternal::verbose)
        //    dataLog("Didn't buffer anything, bailing.\n");
        bufferedStructures.clear();
        return result;
    }

    // The buffering countdown tells us if we should be repatching now.
    if (bufferingCountdown) {
        //if (StructureStubInfoInternal::verbose)
        //    dataLog("Countdown is too high: ", bufferingCountdown, ".\n");
        return result;
    }

    // Forget the buffered structures so that all future attempts to cache get fully handled by the
    // PolymorphicAccess.
    bufferedStructures.clear();

    result = m_stub->regenerate(vm, codeBlock, ecmaMode, *this, ident);

    //if (StructureStubInfoInternal::verbose)
    //    dataLog("Regeneration result: ", result, "\n");

    RELEASE_ASSERT(!result.buffered());

    if (!result.generatedSomeCode())
        return result;

    // When we first transition to becoming a Stub, we might still be running the inline
    // access code. That's because when we first transition to becoming a Stub, we may
    // be buffered, and we have not yet generated any code. Once the Stub finally generates
    // code, we're no longer running the inline access code, so we can then clear out
    // inlineAccessBaseStructure. The reason we don't clear inlineAccessBaseStructure while
    // we're buffered is because we rely on it to reset during GC if inlineAccessBaseStructure
    // is collected.
    inlineAccessBaseStructure.clear();

    // If we generated some code then we don't want to attempt to repatch in the future until we
    // gather enough cases.
    bufferingCountdown = Options::repatchBufferingCountdown();
    return result;
}

void StructureStubInfo::reset(CodeBlock* codeBlock)
{
    bufferedStructures.clear();
    inlineAccessBaseStructure.clear();

    if (cacheType == CacheType::Unset)
        return;

#if 0
    if (Options::verboseOSR()) {
        // This can be called from GC destructor calls, so we don't try to do a full dump
        // of the CodeBlock.
        dataLog("Clearing structure cache (kind ", static_cast<int>(accessType), ") in ", RawPointer(codeBlock), ".\n");
    }
#endif

    switch (accessType) {
    case AccessType::TryGet:
        resetGetByID(codeBlock, *this, GetByIDKind::Try);
        break;
    case AccessType::GetById:
        resetGetByID(codeBlock, *this, GetByIDKind::Normal);
        break;
    case AccessType::GetByIdWithThis:
        resetGetByID(codeBlock, *this, GetByIDKind::WithThis);
        break;
    case AccessType::GetDirect:
        resetGetByID(codeBlock, *this, GetByIDKind::Direct);
        break;
    case AccessType::GetPrivateName:
        resetGetByID(codeBlock, *this, GetByIDKind::PrivateName);
        break;
    case AccessType::GetPrivateNameById:
        resetGetByID(codeBlock, *this, GetByIDKind::PrivateNameById);
        break;
    case AccessType::Put:
        resetPutByID(codeBlock, *this);
        break;
    case AccessType::In:
        resetIn(*this);
        break;
    case AccessType::InstanceOf:
        resetInstanceOf(*this);
        break;
    }

    deref(codeBlock);
    cacheType = CacheType::Unset;
}

void StructureStubInfo::visitWeakReferences(CodeBlock* codeBlock)
{
    VM& vm = codeBlock->vm();

    struct Functor
    {
    	bool operator()(Structure* structure) {
            return Heap::isMarked(structure);
        }
    };
    Functor functor;
    bufferedStructures.genericFilter(functor);

    bool isValid = true;
    if (inlineAccessBaseStructure)
        isValid &= vm.heap.isMarked(inlineAccessBaseStructure.get());
    if (cacheType == CacheType::Stub)
        isValid &= m_stub->visitWeak(vm);

    if (isValid)
        return;

    reset(codeBlock);
    resetByGC = true;
}

bool StructureStubInfo::propagateTransitions(SlotVisitor& visitor)
{
    bool result = true;

    if (inlineAccessBaseStructure)
    	result &= inlineAccessBaseStructure->markIfCheap(visitor);

    if (cacheType == CacheType::Stub)
    	result &= m_stub->propagateTransitions(visitor);

    return result;
}

} // namespace JSC
