/*
 * Copyright (C) 2008-2019 Apple Inc. All rights reserved.
 * Copyright (C) 2008 Cameron Zwarich <cwzwarich@uwaterloo.ca>
 * Copyright (C) 2012 Igalia, S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "CodeBlock.h"
#include <wtf/HashTraits.h>
#include "JSFixedArray.h"
#include "Instruction.h"
#include "Label.h"
#include "LabelScope.h"
#include "Interpreter.h"
#include "JSAsyncGeneratorFunction.h"
#include "JSBigInt.h"
#include "JSGeneratorFunction.h"
#include "JSTemplateObjectDescriptor.h"
#include "ParserError.h"
#include "RegisterID.h"
#include "SymbolTable.h"
#include "Debugger.h"
#include "Nodes.h"
#include "Operands.h"
#include "StaticPropertyAnalyzer.h"
#include <wtf/CheckedArithmetic.h>
#include <wtf/FastMallocMonitor.h>
#include <wtf/HashFunctions.h>
#include <wtf/Optional.h>
#include <wtf/PassRefPtr.h>
#include <wtf/SegmentedVector.h>
#include <wtf/SetForScope.h>
#include <wtf/Vector.h>

namespace JSC {

class ExecState;
class Identifier;
class IndexedForInContext;
class StructureForInContext;
class Label;
class JSScope;

enum ExpectedFunction {
	NoExpectedFunction,
	ExpectObjectConstructor,
	ExpectArrayConstructor
};

enum class EmitAwait { Yes, No };

enum class ThisResolutionType { Local, Scoped };

class CallArguments {
public:
	CallArguments(BytecodeGenerator&, ArgumentsNode*, unsigned additionalArguments = 0);

	RegisterID* thisRegister() { return m_argv[0].get(); }
	RegisterID* argumentRegister(unsigned i) { return m_argv[i + 1].get(); }
	unsigned registerOffset() { return -m_argv.last()->index() + CallFrame::offsetFor(argumentCountIncludingThis()); }
	unsigned argumentCountIncludingThis() { return m_argv.size(); }
	RegisterID* profileHookRegister() { return m_profileHookRegister.get(); }
	ArgumentsNode* argumentsNode() { return m_argumentsNode; }

private:
	void newArgument(BytecodeGenerator&);

	RefPtr<RegisterID> m_profileHookRegister;
	ArgumentsNode* m_argumentsNode;
	Vector<RefPtr<RegisterID>, 8> m_argv;
};

class Variable {
public:
    enum VariableKind { NormalVariable, SpecialVariable };

    Variable()
        : m_offset()
        , m_local(0)
        , m_attributes(0)
        , m_kind(NormalVariable)
    	, m_symbolTableConstantIndex(0) // This is meaningless here for this kind of Variable.
    	, m_isLexicallyScoped(false)
    {
    }

    Variable(const Identifier& ident)
        : m_ident(ident)
        , m_local(0)
        , m_attributes(0)
        , m_kind(NormalVariable) // This is somewhat meaningless here for this kind of Variable.
    	, m_symbolTableConstantIndex(0) // This is meaningless here for this kind of Variable.
    	, m_isLexicallyScoped(false)
	{
	}

    Variable(const Identifier& ident, VarOffset offset, RegisterID* local, unsigned attributes, VariableKind kind, int symbolTableConstantIndex, bool isLexicallyScoped)
        : m_ident(ident)
        , m_offset(offset)
        , m_local(local)
        , m_attributes(attributes)
        , m_kind(kind)
    	, m_symbolTableConstantIndex(symbolTableConstantIndex)
    	, m_isLexicallyScoped(isLexicallyScoped)
    {
    }

    ALWAYS_INLINE ~Variable() { }

    // If it's unset, then it is a non-locally-scoped variable. If it is set, then it could be
    // a stack variable, a scoped variable in a local scope, or a variable captured in the
    // direct arguments object.
    bool isResolved() const { return !!m_offset; }
    int symbolTableConstantIndex() const { ASSERT(isResolved() && !isSpecial()); return m_symbolTableConstantIndex; }

    const Identifier& ident() const { return m_ident; }

    VarOffset offset() const { return m_offset; }
    bool isLocal() const { return m_offset.isStack(); }
    RegisterID* local() const { return m_local; }

    bool isReadOnly() const { return m_attributes & ReadOnly; }
    bool isSpecial() const { return m_kind != NormalVariable; }
    bool isConst() const { return isReadOnly() && m_isLexicallyScoped; }
    void setIsReadOnly() { m_attributes |= ReadOnly; }

    bool operator==(const Variable& other) const
    {
        return m_ident == other.m_ident
            && m_offset == other.m_offset
            && m_local == other.m_local
            && m_attributes == other.m_attributes
            && m_kind == other.m_kind
            && m_symbolTableConstantIndex == other.m_symbolTableConstantIndex
            && m_isLexicallyScoped == other.m_isLexicallyScoped;
    }

private:
    Identifier m_ident;
    VarOffset m_offset;
    RegisterID* m_local;
    unsigned m_attributes;
    VariableKind m_kind;
    int m_symbolTableConstantIndex;
    bool m_isLexicallyScoped;
};

// https://tc39.github.io/ecma262/#sec-completion-record-specification-type
//
// For the Break and Continue cases, instead of using the Break and Continue enum values
// below, we use the unique jumpID of the break and continue statement as the encoding
// for the CompletionType value. emitFinallyCompletion() uses this jumpID value later
// to determine the appropriate jump target to jump to after executing the relevant finally
// blocks. The jumpID is computed as:
//     jumpID = bytecodeOffset (of the break/continue node) + CompletionType::NumberOfTypes.
// Hence, there won't be any collision between jumpIDs and CompletionType enums.
enum class CompletionType : int {
    Normal,
    Throw,
    Return,
    NumberOfTypes
};

inline CompletionType bytecodeOffsetToJumpID(unsigned offset)
{
    int jumpIDAsInt = offset + static_cast<int>(CompletionType::NumberOfTypes);
    ASSERT(jumpIDAsInt >= static_cast<int>(CompletionType::NumberOfTypes));
    return static_cast<CompletionType>(jumpIDAsInt);
}

struct FinallyJump {
    FinallyJump(CompletionType jumpID, int targetLexicalScopeIndex, Label* targetLabel)
        : jumpID(jumpID)
        , targetLexicalScopeIndex(targetLexicalScopeIndex)
        , targetLabel(targetLabel)
    { }

    CompletionType jumpID;
    int targetLexicalScopeIndex;
    RefPtr<Label> targetLabel;
};

class FinallyContext {
public:
    FinallyContext() : m_outerContext(0), m_finallyLabel(0), m_handlesReturns(false) { }
    FinallyContext(BytecodeGenerator&, Label* finallyLabel);

    FinallyContext* outerContext() const { return m_outerContext; }
    Label* finallyLabel() const { return m_finallyLabel; }

    RegisterID* completionTypeRegister() const { return m_completionRecord.typeRegister.get(); }
    RegisterID* completionValueRegister() const { return m_completionRecord.valueRegister.get(); }

    uint32_t numberOfBreaksOrContinues() const { return m_numberOfBreaksOrContinues.unsafeGet(); }
    void incNumberOfBreaksOrContinues() { m_numberOfBreaksOrContinues++; }

    bool handlesReturns() const { return m_handlesReturns; }
    void setHandlesReturns() { m_handlesReturns = true; }

    void registerJump(CompletionType jumpID, int lexicalScopeIndex, Label* targetLabel)
    {
        m_jumps.append(FinallyJump(jumpID, lexicalScopeIndex, targetLabel));
    }

    size_t numberOfJumps() const { return m_jumps.size(); }
    FinallyJump& jumps(size_t i) { return m_jumps[i]; }

private:
    FinallyContext* m_outerContext;
    Label* m_finallyLabel;
    Checked<uint32_t, WTF::CrashOnOverflow> m_numberOfBreaksOrContinues;
    bool m_handlesReturns;
    Vector<FinallyJump> m_jumps;
    struct {
        RefPtr<RegisterID> typeRegister;
        RefPtr<RegisterID> valueRegister;
    } m_completionRecord;
};

struct ControlFlowScope {
    typedef uint8_t Type;
    enum {
        Label,
        Finally
    };
    ControlFlowScope(Type type, int lexicalScopeIndex, FinallyContext* finallyContext = 0)
        : type(type)
        , lexicalScopeIndex(lexicalScopeIndex)
        , finallyContext(finallyContext)
    { }

    bool isLabelScope() const { return type == Label; }
    bool isFinallyScope() const { return type == Finally; }

    Type type;
    int lexicalScopeIndex;
    FinallyContext* finallyContext;
};

class ForInContext : public RefCounted<ForInContext> {
    WTF_MAKE_FAST_ALLOCATED;
    WTF_MAKE_NONCOPYABLE(ForInContext);
public:
    virtual ~ForInContext()
    {
    }

    bool isValid() const { return m_isValid; }
    void invalidate() { m_isValid = false; }

    enum class Type : uint8_t {
        IndexedForIn,
        StructureForIn
    };

    Type type() const { return m_type; }
    bool isIndexedForInContext() const { return m_type == Type::IndexedForIn; }
    bool isStructureForInContext() const { return m_type == Type::StructureForIn; }

    IndexedForInContext& asIndexedForInContext()
    {
        ASSERT(isIndexedForInContext());
        return *reinterpret_cast<IndexedForInContext*>(this);
    }

    StructureForInContext& asStructureForInContext()
    {
        ASSERT(isStructureForInContext());
        return *reinterpret_cast<StructureForInContext*>(this);
    }

    RegisterID* local() const { return m_localRegister.get(); }

protected:
    ForInContext(RegisterID* localRegister, Type type, unsigned bodyBytecodeStartOffset)
		: m_localRegister(localRegister)
		, m_isValid(true)
        , m_type(type)
		, m_bodyBytecodeStartOffset(bodyBytecodeStartOffset)
    { }

    unsigned bodyBytecodeStartOffset() const { return m_bodyBytecodeStartOffset; }

    void finalize(BytecodeGenerator&, CodeBlock*, unsigned bodyBytecodeEndOffset);

private:
    RefPtr<RegisterID> m_localRegister;
    bool m_isValid;
    Type m_type;
    unsigned m_bodyBytecodeStartOffset;
};

class StructureForInContext : public ForInContext {
    typedef ForInContext Base;
public:
    StructureForInContext(RegisterID* localRegister, RegisterID* indexRegister, RegisterID* propertyRegister, RegisterID* enumeratorRegister, unsigned bodyBytecodeStartOffset)
        : ForInContext(localRegister, Type::StructureForIn, bodyBytecodeStartOffset)
        , m_indexRegister(indexRegister)
        , m_propertyRegister(propertyRegister)
        , m_enumeratorRegister(enumeratorRegister)
    {
    }

    RegisterID* index() const { return m_indexRegister.get(); }
    RegisterID* property() const { return m_propertyRegister.get(); }
    RegisterID* enumerator() const { return m_enumeratorRegister.get(); }

    struct GetInstData
	{
    	GetInstData(unsigned instIndex, int propertyRegIndex, ArrayProfile* arrayProfile, UnlinkedValueProfile valueProfile)
    	: instIndex(instIndex)
    	, propertyRegIndex(propertyRegIndex)
    	, arrayProfile(arrayProfile)
    	, valueProfile(valueProfile)
    	{
    	}
    	unsigned instIndex;
    	int propertyRegIndex;
    	ArrayProfile* arrayProfile;
    	UnlinkedValueProfile valueProfile;
	};
    void addGetInst(unsigned instIndex, int propertyRegIndex, ArrayProfile* arrayProfile, UnlinkedValueProfile valueProfile)
    {
        m_getInsts.append(GetInstData(instIndex, propertyRegIndex, arrayProfile, valueProfile));
    }

    void finalize(BytecodeGenerator&, CodeBlock*, unsigned bodyBytecodeEndOffset);

private:
    RefPtr<RegisterID> m_indexRegister;
    RefPtr<RegisterID> m_propertyRegister;
    RefPtr<RegisterID> m_enumeratorRegister;
    Vector<GetInstData> m_getInsts;
};

class IndexedForInContext : public ForInContext {
    typedef ForInContext Base;
public:
    IndexedForInContext(RegisterID* localRegister, RegisterID* indexRegister, unsigned bodyBytecodeStartOffset)
        : ForInContext(localRegister, Type::IndexedForIn, bodyBytecodeStartOffset)
        , m_indexRegister(indexRegister)
    {
    }

    RegisterID* index() const { return m_indexRegister.get(); }

    void finalize(BytecodeGenerator&, CodeBlock*, unsigned bodyBytecodeEndOffset);
    void addGetInst(unsigned instIndex, int propertyIndex) { m_getInsts.append(std::make_pair(instIndex, propertyIndex)); }

private:
    RefPtr<RegisterID> m_indexRegister;
    Vector<std::pair<unsigned, int>> m_getInsts;
};

struct TryData {
	RefPtr<Label> target;
    HandlerType handlerType;
};

struct TryContext {
	RefPtr<Label> start;
	TryData* tryData;
};

enum CaptureMode {
    NotCaptured,
    IsCaptured
};

struct TryRange {
    RefPtr<Label> start;
    RefPtr<Label> end;
    TryData* tryData;
};

enum ProfileTypeBytecodeFlag {
    ProfileTypeBytecodePutToScope,
    ProfileTypeBytecodeGetFromScope,
    ProfileTypeBytecodePutToLocalScope,
    ProfileTypeBytecodeGetFromLocalScope,
    ProfileTypeBytecodeHasGlobalID,
    ProfileTypeBytecodeDoesNotHaveGlobalID,
    ProfileTypeBytecodeFunctionArgument,
    ProfileTypeBytecodeFunctionReturnStatement
};

class BytecodeGenerator {
	WTF_MAKE_FAST_ALLOCATED;
    WTF_MAKE_NONCOPYABLE(BytecodeGenerator);

    friend class FinallyContext;
	friend class Label;
    friend class StrictModeScope;
public:
	typedef DeclarationStacks::FunctionStack FunctionStack;

	JS_EXPORT_PRIVATE static void setDumpsGeneratedCode(bool dumpsGeneratedCode);
	static bool dumpsGeneratedCode();

	BytecodeGenerator(ProgramNode*, JSGlobalObject*, ProgramCodeBlock*, const RefPtr<TDZEnvironmentLink>&, const PrivateNameEnvironment*, const Vector<Identifier>*, ECMAMode);
	BytecodeGenerator(FunctionNode*, JSGlobalObject*, CodeBlock*, const RefPtr<TDZEnvironmentLink>&, const PrivateNameEnvironment*, const Vector<Identifier>*, ECMAMode);
	BytecodeGenerator(EvalNode*, JSGlobalObject*, EvalCodeBlock*, const RefPtr<TDZEnvironmentLink>&, const PrivateNameEnvironment*, const Vector<Identifier>*, ECMAMode);
	BytecodeGenerator(ModuleProgramNode*, JSGlobalObject*, ModuleProgramCodeBlock*, const RefPtr<TDZEnvironmentLink>&, const PrivateNameEnvironment*, const Vector<Identifier>*, ECMAMode);

	~BytecodeGenerator();

	JSGlobalData& vm() const { return *m_vm; }
	ParserArena& parserArena() const { return m_scopeNode->parserArena(); }
	const CommonIdentifiers& propertyNames() const { return *m_vm->propertyNames; }

	bool isConstructor() const { return m_codeBlock->isConstructor(); }
	DerivedContextType derivedContextType() const { return m_derivedContextType; }
	bool usesArrowFunction() const { return m_scopeNode->usesArrowFunction(); }
	bool needsToUpdateArrowFunctionContext() const { return m_needsToUpdateArrowFunctionContext; }
	bool usesEval() const { return m_scopeNode->usesEval(); }
	bool usesThis() const { return m_scopeNode->usesThis(); }
	PrivateBrandRequirement privateBrandRequirement() const { return m_codeBlock->privateBrandRequirement(); }
	ConstructorKind constructorKind() const { return m_codeBlock->constructorKind(); }
    SuperBinding superBinding() const { return m_codeBlock->superBinding(); }
    JSParserScriptMode scriptMode() const { return m_codeBlock->scriptMode(); }
    NeedsClassFieldInitializer needsClassFieldInitializer() const { return m_codeBlock->needsClassFieldInitializer(); }

	ParserError generate(ExecState*, JSScope*);

	bool isArgumentNumber(const Identifier&, int);

    Variable variable(const Identifier&, ThisResolutionType = ThisResolutionType::Local);

    enum ExistingVariableMode { VerifyExisting, IgnoreExisting };
    void createVariable(const Identifier&, VarKind, SymbolTable*, ExistingVariableMode = VerifyExisting); // Creates the variable, or asserts that the already-created variable is sufficiently compatible.

	// Returns the register storing "this"
	RegisterID* thisRegister() { return &m_thisRegister; }
    RegisterID* argumentsRegister() { return m_argumentsRegister; }
    RegisterID* newTarget()
	{
        ASSERT(m_newTargetRegister);
        return m_newTargetRegister;
	}

    RegisterID* scopeRegister() { return m_scopeRegister; }

    RegisterID* generatorRegister() { return m_generatorRegister; }

    RegisterID* promiseCapabilityRegister() { return m_promiseCapabilityRegister; }

	// Returns the next available temporary register. Registers returned by
	// newTemporary require a modified form of reference counting: any
	// register with a refcount of 0 is considered "available", meaning that
	// the next instruction may overwrite it.
	RegisterID* newTemporary();

	// The same as newTemporary(), but this function returns "suggestion" if
	// "suggestion" is a temporary. This function is helpful in situations
	// where you've put "suggestion" in a RefPtr, but you'd like to allow
	// the next instruction to overwrite it anyway.
	RegisterID* newTemporaryOr(RegisterID* suggestion) { return suggestion->isTemporary() ? suggestion : newTemporary(); }

	// Functions for handling of dst register

	RegisterID* ignoredResult() { return &m_ignoredResultRegister; }

    // This will be allocated in the temporary region of registers, but it will
    // not be marked as a temporary. This will ensure that finalDestination() does
    // not overwrite a block scope variable that it mistakes as a temporary. These
    // registers can be (and are) reclaimed when the lexical scope they belong to
    // is no longer on the symbol table stack.
    RegisterID* newBlockScopeVariable();

	// Returns a place to write intermediate values of an operation
	// which reuses dst if it is safe to do so.
	RegisterID* tempDestination(RegisterID* dst)
	{
		return (dst && dst != ignoredResult() && dst->isTemporary()) ? dst : newTemporary();
	}

	// Returns the place to write the final output of an operation.
	ALWAYS_INLINE RegisterID* finalDestination(RegisterID* originalDst, RegisterID* tempDst = nullptr)
	{
		if (originalDst && originalDst != ignoredResult())
			return originalDst;
		ASSERT(tempDst != ignoredResult());
		if (tempDst && tempDst->isTemporary())
			return tempDst;
		return newTemporary();
	}

	RegisterID* destinationForAssignResult(RegisterID* dst)
	{
		if (dst && dst != ignoredResult())
			return dst->isTemporary() ? dst : newTemporary();
		return 0;
	}

	// Moves src to dst if dst is not null and is different from src, otherwise just returns src.
	ALWAYS_INLINE RegisterID* move(RegisterID* dst, RegisterID* src)
	{
		return dst == ignoredResult() ? 0 : (dst && dst != src) ? emitMove(dst, src) : src;
	}

	PassRefPtr<LabelScope> newLabelScope(LabelScope::Type, const Identifier* = nullptr);
	Ref<Label> newLabel();
	Ref<Label> newEmittedLabel();

	ALWAYS_INLINE void emitNode(RegisterID* dst, StatementNode* n)
    {
        SetForScope<bool> tailPositionPoisoner(m_inTailPosition, false);
        return emitNodeInTailPosition(dst, n);
    }

    void emitNodeInTailPosition(RegisterID* dst, StatementNode* n)
    {
        // Node::emitCode assumes that dst, if provided, is either a local or a referenced temporary.
        ASSERT(!dst || dst == ignoredResult() || !dst->isTemporary() || dst->refCount());
        if (LIKELY(m_stack.isSafeToRecurse()))
        	n->emitBytecode(*this, dst);
        else
            emitThrowExpressionTooDeepException();
    }

    ALWAYS_INLINE void emitNode(StatementNode* n)
    {
        emitNode(0, n);
    }

    void emitNodeInTailPosition(StatementNode* n)
    {
        emitNodeInTailPosition(0, n);
    }

    ALWAYS_INLINE RegisterID* emitNode(RegisterID* dst, ExpressionNode* n)
    {
        SetForScope<bool> tailPositionPoisoner(m_inTailPosition, false);
        return emitNodeInTailPosition(dst, n);
    }

    RegisterID* emitNodeInTailPosition(RegisterID* dst, ExpressionNode* n)
    {
        // Node::emitCode assumes that dst, if provided, is either a local or a referenced temporary.
        ASSERT(!dst || dst == ignoredResult() || !dst->isTemporary() || dst->refCount());
        if (LIKELY(m_stack.isSafeToRecurse()))
        	return n->emitBytecode(*this, dst);
        return emitThrowExpressionTooDeepException();
    }

    ALWAYS_INLINE RegisterID* emitNode(ExpressionNode* n)
    {
        return emitNode(nullptr, n);
    }

    RegisterID* emitNodeInTailPosition(ExpressionNode* n)
    {
        return emitNodeInTailPosition(nullptr, n);
    }

    RegisterID* emitDefineClassElements(PropertyListNode* n, RegisterID* constructor, RegisterID* prototype, Vector<JSTextPosition>* instanceFieldLocations, Vector<JSTextPosition>* staticFieldLocations)
    {
        ASSERT(constructor->refCount() && prototype->refCount());
        if (UNLIKELY(!m_stack.isSafeToRecurse()))
            return emitThrowExpressionTooDeepException();
        if (UNLIKELY(n->needsDebugHook()))
            emitDebugHook(n);
        return n->emitBytecode(*this, constructor, prototype, instanceFieldLocations, staticFieldLocations);
    }

    RegisterID* emitNodeForProperty(RegisterID* dst, ExpressionNode* node)
    {
        if (node->isString()) {
            if (Optional<uint32_t> index = parseIndex(static_cast<StringNode*>(node)->value()))
                return emitLoad(dst, jsNumber(index.value()));
        }
        return emitNode(dst, node);
    }

    RegisterID* emitNodeForProperty(ExpressionNode* n)
    {
        return emitNodeForProperty(nullptr, n);
    }

	void emitNodeInConditionContext(ExpressionNode* n, Label* trueTarget, Label* falseTarget, FallThroughMode fallThroughMode)
	{
        if (!m_stack.isSafeToRecurse()) {
            emitThrowExpressionTooDeepException();
            return;
        }

        n->emitBytecodeInConditionContext(*this, trueTarget, falseTarget, fallThroughMode);
    }

    void emitExpressionInfo(const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
    {
        ASSERT(divot && divotStart && divotEnd);
        ASSERT(divot.offset >= divotStart.offset);
        ASSERT(divotEnd.offset >= divot.offset);

        // Don't emit expression info if the data could cause us to crash later.
        // In this case we'll just use the wrong info for an error message, not crash.
        if (!divot || !divotStart || !divotEnd)
            return;

        if (m_isBuiltinFunction)
            return;

        if (lowMemory())
        	return;

        int sourceOffset = m_scopeNode->source().startOffset();
        unsigned firstLine = m_scopeNode->source().firstLine();

        int divotOffset = divot.offset - sourceOffset;
        int startOffset = divot.offset - divotStart.offset;
        int endOffset = divotEnd.offset - divot.offset;

        unsigned line = divot.line;
        ASSERT(line >= firstLine);
        line -= firstLine;

        int lineStart = divot.lineStartOffset;
        if (lineStart > sourceOffset)
            lineStart -= sourceOffset;
        else
            lineStart = 0;
        ASSERT(divotOffset >= lineStart);
        unsigned column = divotOffset - lineStart;

        unsigned instructionOffset = instructions().size();
        m_codeBlock->addExpressionInfo(instructionOffset, divotOffset, startOffset, endOffset, line, column);
    }

	ALWAYS_INLINE bool leftHandSideNeedsCopy(bool rightHasAssignments, bool rightIsPure)
	{
		return (m_codeType != FunctionCode || rightHasAssignments) && !rightIsPure;
	}

	ALWAYS_INLINE PassRefPtr<RegisterID> emitNodeForLeftHandSide(ExpressionNode* n, bool rightHasAssignments, bool rightIsPure)
	{
		if (leftHandSideNeedsCopy(rightHasAssignments, rightIsPure)) {
			PassRefPtr<RegisterID> dst = newTemporary();
			emitNode(dst.get(), n);
			return dst;
		}

		return emitNode(0, n);
	}

    ALWAYS_INLINE RefPtr<RegisterID> emitNodeForLeftHandSideForProperty(ExpressionNode* n, bool rightHasAssignments, bool rightIsPure)
    {
        if (leftHandSideNeedsCopy(rightHasAssignments, rightIsPure)) {
            RefPtr<RegisterID> dst = newTemporary();
            emitNodeForProperty(dst.get(), n);
            return dst;
        }

        return emitNodeForProperty(n);
    }

    void hoistSloppyModeFunctionIfNecessary(FunctionMetadataNode*);

    RegisterID* emitLoadArrowFunctionLexicalEnvironment(const Identifier&);
    RegisterID* ensureThis();
    void emitLoadThisFromArrowFunctionLexicalEnvironment();
    RegisterID* emitLoadNewTargetFromArrowFunctionLexicalEnvironment();

    unsigned addConstantIndex();
	RegisterID* emitLoad(RegisterID* dst, bool);
	RegisterID* emitLoad(RegisterID* dst, const Identifier&);
	RegisterID* emitLoad(RegisterID* dst, JSValue, SourceCodeRepresentation = SourceCodeRepresentation::Other);
	RegisterID* emitLoad(RegisterID* dst, IdentifierSet&& excludedList);
	RegisterID* emitLoadGlobalObject(RegisterID* dst);

	RegisterID* emitUnaryOp(OpcodeID, RegisterID* dst, RegisterID* src, OperandTypes = OperandTypes());
	RegisterID* emitUnaryOpProfiled(OpcodeID, RegisterID* dst, RegisterID* src);
	RegisterID* emitBinaryOp(OpcodeID, RegisterID* dst, RegisterID* src1, RegisterID* src2, OperandTypes = OperandTypes());
	RegisterID* emitEqualityOp(OpcodeID, RegisterID* dst, RegisterID* src1, RegisterID* src2);
	bool emitEqualityOpImpl(RegisterID* dst, RegisterID* src1, RegisterID* src2);
	RegisterID* emitUnaryNoDstOp(OpcodeID, RegisterID* src);

    RegisterID* emitCreateThis(RegisterID* dst);
    RegisterID* emitInstanceFieldInitializationIfNeeded(RegisterID* dst, RegisterID* constructor, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
    void emitTDZCheck(RegisterID* target);
    bool needsTDZCheck(const Variable&);
	void emitTDZCheckIfNecessary(const Variable&, RegisterID* target, RegisterID* scope);
	void liftTDZCheckIfPossible(const Variable&);
	RegisterID* emitNewObject(RegisterID* dst);
	RegisterID* emitNewArray(RegisterID* dst, ElementNode*, unsigned length); // stops at first elision
	RegisterID* emitNewArrayWithSpread(RegisterID* dst, ElementNode*);
	RegisterID* emitNewArrayWithSize(RegisterID* dst, RegisterID* length);
	RegisterID* emitNewArrayBuffer(RegisterID* dst, JSFixedArray*);

	RegisterID* emitNewFunction(RegisterID* dst, FunctionMetadataNode*);
	RegisterID* emitNewFunctionExpression(RegisterID* dst, FuncExprNode*);
	RegisterID* emitNewDefaultConstructor(RegisterID* dst, ConstructorKind, const Identifier& name, const Identifier& ecmaName, const SourceCode& classSource, NeedsClassFieldInitializer, PrivateBrandRequirement);
	RegisterID* emitNewClassFieldInitializerFunction(RegisterID* dst, PassOwnPtr<Vector<JSTextPosition>> classFieldLocations, bool isDerived);
	RegisterID* emitNewArrowFunctionExpression(RegisterID*, ArrowFuncExprNode*);
	RegisterID* emitNewMethodDefinition(RegisterID* dst, MethodDefinitionNode*);
	RegisterID* emitNewRegExp(RegisterID* dst, RegExp*);

    bool shouldSetFunctionName(ExpressionNode*);
    void emitSetFunctionName(RegisterID* value, RegisterID* name);
    void emitSetFunctionName(RegisterID* value, const Identifier&);

	RegisterID* emitMoveLinkTimeConstant(RegisterID* dst, LinkTimeConstant);
    RegisterID* moveEmptyValue(RegisterID* dst);

	RegisterID* emitToNumber(RegisterID* dst, RegisterID* src) { return emitUnaryOpProfiled(op_to_number, dst, src); }
    RegisterID* emitToNumeric(RegisterID* dst, RegisterID* src) { return emitUnaryOpProfiled(op_to_numeric, dst, src); }
	RegisterID* emitToString(RegisterID* dst, RegisterID* src) { return emitUnaryOp(op_to_string, dst, src); }
	RegisterID* emitToObject(RegisterID* dst, RegisterID* src, const Identifier& message);
	RegisterID* emitToThis(RegisterID* srcDst);
	RegisterID* emitInc(RegisterID* srcDst);
	RegisterID* emitDec(RegisterID* srcDst);

	RegisterID* emitOverridesHasInstance(RegisterID* dst, RegisterID* constructor, RegisterID* hasInstanceValue);
	RegisterID* emitInstanceOf(RegisterID* dst, RegisterID* value, RegisterID* basePrototype);
	RegisterID* emitInstanceOfCustom(RegisterID* dst, RegisterID* value, RegisterID* constructor, RegisterID* hasInstanceValue);
	RegisterID* emitTypeOf(RegisterID* dst, RegisterID* src) { return emitUnaryOp(op_typeof, dst, src); }
	RegisterID* emitIn(RegisterID* dst, RegisterID* property, RegisterID* base);

	RegisterID* emitTryGetById(RegisterID* dst, RegisterID* base, const Identifier& property);
	RegisterID* emitGetById(RegisterID* dst, RegisterID* base, const Identifier& property);
	RegisterID* emitGetById(RegisterID* dst, RegisterID* base, RegisterID* thisVal, const Identifier& property);
	RegisterID* emitDirectGetById(RegisterID* dst, RegisterID* base, const Identifier& property);
	RegisterID* emitPutById(RegisterID* base, const Identifier& property, RegisterID* value);
	RegisterID* emitPutById(RegisterID* base, RegisterID* thisValue, const Identifier& property, RegisterID* value);
	RegisterID* emitDirectPutById(RegisterID* base, const Identifier& property, RegisterID* value);
	RegisterID* emitDeleteById(RegisterID* dst, RegisterID* base, const Identifier&);
	RegisterID* emitGetByVal(RegisterID* dst, RegisterID* base, RegisterID* property);
	RegisterID* emitGetByVal(RegisterID* dst, RegisterID* base, RegisterID* thisValue, RegisterID* property);
	RegisterID* emitGetPrototypeOf(RegisterID* dst, RegisterID* value);
	RegisterID* emitDirectSetPrototypeOf(RegisterID* base, RegisterID* prototype);
	RegisterID* emitPutByVal(RegisterID* base, RegisterID* property, RegisterID* value);
	RegisterID* emitPutByVal(RegisterID* base, RegisterID* thisValue, RegisterID* property, RegisterID* value);
	RegisterID* emitDirectPutByVal(RegisterID* base, RegisterID* property, RegisterID* value);
	RegisterID* emitDeleteByVal(RegisterID* dst, RegisterID* base, RegisterID* property);

	RegisterID* emitDefinePrivateField(RegisterID* base, RegisterID* property, RegisterID* value);
	RegisterID* emitPrivateFieldPut(RegisterID* base, RegisterID* property, RegisterID* value);
	RegisterID* emitGetPrivateName(RegisterID* dst, RegisterID* base, RegisterID* property);
	RegisterID* emitHasPrivateName(RegisterID* dst, RegisterID* base, RegisterID* property);

	void emitCreatePrivateBrand(const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
	void emitInstallPrivateBrand(RegisterID* target);
    void emitInstallPrivateClassBrand(RegisterID* target);

    RegisterID* emitGetPrivateBrand(RegisterID* dst, RegisterID* scope, bool isStatic);
    RegisterID* emitHasPrivateBrand(RegisterID* dst, RegisterID* base, RegisterID* brand, bool isStatic);
	void emitCheckPrivateBrand(RegisterID* base, RegisterID* brand, bool isStatic);

    RegisterID* emitIdWithProfile(RegisterID* src, SpeculatedType profile);
    void emitUnreachable();

	void emitPutGetterById(RegisterID* base, const Identifier& property, unsigned propertyDescriptorOptions, RegisterID* getter);
	void emitPutSetterById(RegisterID* base, const Identifier& property, unsigned propertyDescriptorOptions, RegisterID* setter);
	void emitPutGetterSetter(RegisterID* base, const Identifier& property, unsigned attributes, RegisterID* getter, RegisterID* setter);
	void emitPutGetterByVal(RegisterID* base, RegisterID* property, unsigned propertyDescriptorOptions, RegisterID* getter);
	void emitPutSetterByVal(RegisterID* base, RegisterID* property, unsigned propertyDescriptorOptions, RegisterID* setter);

    // Initialize object with generator fields (@generatorThis, @generatorNext, @generatorState, @generatorFrame)
    void emitPutGeneratorFields(RegisterID* nextFunction);

    void emitPutAsyncGeneratorFields(RegisterID* nextFunction);

    RegisterID* emitGetArgument(RegisterID* dst, int32_t index);

	ExpectedFunction expectedFunctionForIdentifier(const Identifier&);
	RegisterID* emitCall(RegisterID* dst, RegisterID* func, ExpectedFunction, CallArguments&, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
	RegisterID* emitCallInTailPosition(RegisterID* dst, RegisterID* func, ExpectedFunction, CallArguments&, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
	RegisterID* emitCallEval(RegisterID* dst, RegisterID* func, CallArguments&, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
	RegisterID* emitCallVarargs(RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
	RegisterID* emitCallVarargsInTailPosition(RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);

    enum PropertyDescriptorOption {
        PropertyConfigurable = 1,
        PropertyWritable     = 1 << 1,
        PropertyEnumerable   = 1 << 2,
    };
    void emitCallDefineProperty(RegisterID* newObj, RegisterID* propertyNameRegister,
        RegisterID* valueRegister, RegisterID* getterRegister, RegisterID* setterRegister, unsigned options, const JSTextPosition&);

	class EnumerationFunctor
	{
	public:
		virtual ~EnumerationFunctor() { }
		virtual void operator()(BytecodeGenerator&, RegisterID*) = 0;
	};
	void emitEnumeration(ThrowableExpressionData* enumerationNode, ExpressionNode* subjectNode, EnumerationFunctor& callBack, ForOfNode* = 0, RegisterID* forLoopSymbolTable = 0);

	RegisterID* emitGetTemplateObject(RegisterID* dst, TaggedTemplateNode*);
    RegisterID* emitGetGlobalPrivate(RegisterID* dst, const Identifier& property);

    RegisterID* emitReturn(RegisterID* src);
	RegisterID* emitEnd(RegisterID* src) { return emitUnaryNoDstOp(op_end, src); }

	RegisterID* emitConstruct(RegisterID* dst, RegisterID* func, RegisterID* lazyThis, ExpectedFunction, CallArguments&, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
	RegisterID* emitStrcat(RegisterID* dst, RegisterID* src, int count);
	void emitToPrimitive(RegisterID* dst, RegisterID* src);
    RegisterID* emitToPropertyKey(RegisterID* dst, RegisterID* src);

    ResolveType resolveType();
    RegisterID* emitResolveConstantLocal(RegisterID* dst, const Variable&);
    RegisterID* emitResolveScope(RegisterID* dst, const Variable&);
    RegisterID* emitGetFromScope(RegisterID* dst, RegisterID* scope, const Variable&, ResolveMode);
    RegisterID* emitPutToScope(RegisterID* scope, const Variable&, RegisterID* value, ResolveMode, InitializationMode);
    RegisterID* emitPutToScopeDynamic(RegisterID* scope, const Identifier&, RegisterID* value, ResolveMode, InitializationMode);

    RegisterID* emitResolveScopeForHoistingFuncDeclInEval(RegisterID* dst, const Identifier&);

    RegisterID* initializeVariable(const Variable&, RegisterID* value);

	void emitLabel(Label&);
	void emitLabel(Label* lable) { emitLabel(*lable); }
	void emitLoopHint();
	PassRefPtr<Label> emitJump(Label* target);
	void emitJumpIfTrue(RegisterID* cond, Label* target);
	void emitJumpIfFalse(RegisterID* cond, Label* target);
	PassRefPtr<Label> emitJumpIfNotFunctionCall(RegisterID* cond, Label* target);
	PassRefPtr<Label> emitJumpIfNotFunctionApply(RegisterID* cond, Label* target);

    RegisterID* emitHasEnumerableIndexedProperty(RegisterID* dst, RegisterID* base, RegisterID* propertyName);
    RegisterID* emitHasEnumerableStructureProperty(RegisterID* dst, RegisterID* base, RegisterID* propertyName, RegisterID* enumerator);
    RegisterID* emitHasEnumerableProperty(RegisterID* dst, RegisterID* base, RegisterID* propertyName);
    RegisterID* emitGetPropertyEnumerator(RegisterID* dst, RegisterID* base);
	RegisterID* emitGetEnumerableLength(RegisterID* dst, RegisterID* base);
    RegisterID* emitGetStructurePropertyEnumerator(RegisterID* dst, RegisterID* base, RegisterID* length);
    RegisterID* emitGetGenericPropertyEnumerator(RegisterID* dst, RegisterID* base, RegisterID* length, RegisterID* structureEnumerator);
    RegisterID* emitEnumeratorStructurePropertyName(RegisterID* dst, RegisterID* enumerator, RegisterID* index);
	RegisterID* emitEnumeratorGenericPropertyName(RegisterID* dst, RegisterID* enumerator, RegisterID* index);
	RegisterID* emitToIndexString(RegisterID* dst, RegisterID* index);

    RegisterID* emitIsCellWithType(RegisterID* dst, RegisterID* src, JSType);
    RegisterID* emitIsJSArray(RegisterID* dst, RegisterID* src) { return emitIsCellWithType(dst, src, ArrayType); }
    RegisterID* emitIsProxyObject(RegisterID* dst, RegisterID* src) { return emitIsCellWithType(dst, src, ProxyObjectType); }
    RegisterID* emitIsRegExpObject(RegisterID* dst, RegisterID* src) { return emitIsCellWithType(dst, src, RegExpObjectType); }
	RegisterID* emitIsMap(RegisterID* dst, RegisterID* src) { return emitIsCellWithType(dst, src, JSMapType); }
	RegisterID* emitIsSet(RegisterID* dst, RegisterID* src) { return emitIsCellWithType(dst, src, JSSetType); }
    RegisterID* emitIsObject(RegisterID* dst, RegisterID* src);
    RegisterID* emitIsConstructor(RegisterID* dst, RegisterID* src);
    RegisterID* emitIsCallable(RegisterID* dst, RegisterID* src);
    RegisterID* emitIsNumber(RegisterID* dst, RegisterID* src);
    RegisterID* emitIsNull(RegisterID* dst, RegisterID* src) { return emitEqualityOp(op_stricteq, dst, src, emitLoad(0, jsNull())); }
    RegisterID* emitIsUndefined(RegisterID* dst, RegisterID* src) { return emitEqualityOp(op_stricteq, dst, src, emitLoad(0, jsUndefined())); }
    RegisterID* emitIsUndefinedOrNull(RegisterID* dst, RegisterID* src);
    RegisterID* emitIsEmpty(RegisterID* dst, RegisterID* src);
    RegisterID* emitIsDerivedArray(RegisterID* dst, RegisterID* src) { return emitIsCellWithType(dst, src, DerivedArrayType); }
	void emitRequireObjectCoercible(RegisterID* value, const String& error);

    RegisterID* emitIteratorNext(RegisterID* dst, RegisterID* nextMethod, RegisterID* iterator, const ThrowableExpressionData* node, JSC::EmitAwait = JSC::EmitAwait::No);
    RegisterID* emitIteratorNextWithValue(RegisterID* dst, RegisterID* nextMethod, RegisterID* iterator, RegisterID* value, const ThrowableExpressionData* node);
	void emitIteratorClose(RegisterID* iterator, const ThrowableExpressionData* node, EmitAwait = EmitAwait::No);

    RegisterID* emitRestParameter(RegisterID* result, unsigned numParametersToSkip);

    bool emitReadOnlyExceptionIfNeeded(const Variable&);

    // Start a try block. 'start' must have been emitted.
    TryData* pushTry(Label* start, Label* handlerLabel, HandlerType);
    // End a try block. 'end' must have been emitted.
    void popTry(TryData*, Label* end);

    void emitOutOfLineCatchHandler(RegisterID* thrownValueRegister, RegisterID* completionTypeRegister, TryData*);
    void emitOutOfLineFinallyHandler(RegisterID* exceptionRegister, RegisterID* completionTypeRegister, TryData*);

    void pushClassHeadLexicalScope(VariableEnvironment&);
    void popClassHeadLexicalScope(VariableEnvironment&);

private:
    static const int CurrentLexicalScopeIndex = -2;
    static const int OutermostLexicalScopeIndex = -1;

    int currentLexicalScopeIndex() const
    {
        int size = static_cast<int>(m_lexicalScopeStack.size());
        ASSERT(static_cast<size_t>(size) == m_lexicalScopeStack.size());
        ASSERT(size >= 0);
        if (!size)
            return OutermostLexicalScopeIndex;
        return size - 1;
    }

    void emitOutOfLineExceptionHandler(RegisterID* exceptionRegister, RegisterID* thrownValueRegister, RegisterID* completionTypeRegister, TryData*);

public:
    enum class ScopeType : uint8_t { CatchScope, CatchScopeWithSimpleParameter, LetConstScope, FunctionNameScope, ClassScope };

    void restoreScopeRegister();
    void restoreScopeRegister(int lexicalScopeIndex);

    int labelScopeDepthToLexicalScopeIndex(int labelScopeDepth);

	void emitThrow(RegisterID* exc)
	{
		m_usesExceptions = true;
		emitUnaryNoDstOp(op_throw, exc);
	}

    void emitThrowStaticError(ErrorType, RegisterID*);
    void emitThrowStaticError(ErrorType, const Identifier& message);
	void emitThrowReferenceError(const String& message);
    void emitThrowTypeError(const String& message);
    void emitThrowTypeError(const Identifier& message);
    void emitThrowRangeError(const Identifier& message);
    void emitThrowOutOfMemoryError();

    void emitPushCatchScope(VariableEnvironment&, ScopeType);
    void emitPopCatchScope(VariableEnvironment&);

    RegisterID* emitGetIterator(RegisterID*, ThrowableExpressionData*);
    RegisterID* emitGetAsyncIterator(RegisterID*, ThrowableExpressionData*);

    void emitGetScope();
    RegisterID* emitPushWithScope(RegisterID* objectScope);
    void emitPopWithScope();
    void emitPutThisToArrowFunctionContextScope();
	void emitPutNewTargetToArrowFunctionContextScope();
	void emitPutDerivedConstructorToArrowFunctionContextScope();
	RegisterID* emitLoadDerivedConstructorFromArrowFunctionLexicalEnvironment();
    RegisterID* emitLoadDerivedConstructor();
	void emitPushFunctionNameScope(const Identifier& property, RegisterID* value, bool isCaptured);

    void emitDebugHook(DebugHookID, const JSTextPosition&);
	void emitDebugHook(DebugHookID, unsigned line, unsigned charOffset, unsigned lineStart);
	void emitDebugHook(StatementNode*);
	void emitDebugHook(ExpressionNode*);
	void emitWillLeaveCallFrameDebugHook();

	RegisterID* emitLoad(RegisterID* dst, CompletionType type) { return emitLoad(dst, jsNumber(static_cast<int32_t>(type))); }
	RegisterID* emitLoad(RegisterID* dst, JSGeneratorFunction::GeneratorResumeMode mode) { return emitLoad(dst, jsNumber(static_cast<int32_t>(mode))); }

    bool emitJumpViaFinallyIfNeeded(int targetLabelScopeDepth, Label* jumpTarget);
    bool emitReturnViaFinallyIfNeeded(RegisterID* returnRegister);
    void emitFinallyCompletion(FinallyContext&, Label* normalCompletionLabel);

public:
    void pushFinallyControlFlowScope(FinallyContext&);
    void popFinallyControlFlowScope();

    void pushOptionalChainTarget();
    void popOptionalChainTarget();
    void popOptionalChainTarget(RegisterID* dst, bool isDelete);
    void emitOptionalCheck(RegisterID* src);

    void pushIndexedForInScope(RegisterID* local, RegisterID* index);
    void popIndexedForInScope(RegisterID* local);
    void pushStructureForInScope(RegisterID* local, RegisterID* index, RegisterID* property, RegisterID* enumerator);
    void popStructureForInScope(RegisterID* local);

    LabelScope* breakTarget(const Identifier&);
    LabelScope* continueTarget(const Identifier&);

	void beginSwitch(RegisterID*, SwitchInfo::SwitchType);
	void endSwitch(uint32_t clauseCount, RefPtr<Label>*, ExpressionNode**, Label* defaultLabel, int32_t min, int32_t range);

    void emitYieldPoint(RegisterID*, JSAsyncGeneratorFunction::AsyncGeneratorSuspendReason);

    void emitGeneratorStateLabel();
    void emitGeneratorStateChange(int32_t state);
    RegisterID* emitYield(RegisterID* argument);
	RegisterID* emitAwait(RegisterID* dst, RegisterID* src);
	RegisterID* emitAwait(RegisterID* srcDst) { return emitAwait(srcDst, srcDst); }
	RegisterID* emitDelegateYield(RegisterID* argument, ThrowableExpressionData*);
    RegisterID* generatorStateRegister() { return &m_parameters[static_cast<int32_t>(JSGeneratorFunction::GeneratorArgument::State)]; }
	RegisterID* generatorValueRegister() { return &m_parameters[static_cast<int32_t>(JSGeneratorFunction::GeneratorArgument::Value)]; }
	RegisterID* generatorResumeModeRegister() { return &m_parameters[static_cast<int32_t>(JSGeneratorFunction::GeneratorArgument::ResumeMode)]; }
	RegisterID* generatorFrameRegister() { return &m_parameters[static_cast<int32_t>(JSGeneratorFunction::GeneratorArgument::Frame)]; }

	CodeType codeType() const { return m_codeType; }

	bool shouldEmitProfileHooks() { return m_shouldEmitProfileHooks; }

    bool shouldBeConcernedWithCompletionValue() const { return m_codeType != FunctionCode; }

    bool shouldEmitDebugHooks() const { return m_shouldEmitDebugHooks; }

    ECMAMode ecmaMode() const { return m_ecmaMode; }

    SourceParseMode parseMode() const { return m_codeBlock->parseMode(); }

    bool isBuiltinFunction() const { return m_isBuiltinFunction; }

    OpcodeID lastOpcodeID() const { return m_lastOpcodeID; }

    ArrayProfile* newArrayProfile();

    bool isDerivedConstructorContext() const { return m_derivedContextType == DerivedContextType::DerivedConstructorContext; }
    bool isDerivedClassContext() const { return m_derivedContextType == DerivedContextType::DerivedMethodContext; }
    bool isArrowFunction() const { return m_codeBlock->isArrowFunction(); }

    enum class TDZCheckOptimization { Optimize, DoNotOptimize };
    enum class NestedScopeType { IsNested, IsNotNested };
private:
    enum class TDZRequirement { UnderTDZ, NotUnderTDZ };
	enum class ScopeRegisterType { Var, Block };
	void pushLexicalScopeInternal(VariableEnvironment&, TDZCheckOptimization, NestedScopeType, RegisterID** constantSymbolTableResult, TDZRequirement, ScopeType, ScopeRegisterType);
	void initializeBlockScopedFunctions(VariableEnvironment&, FunctionStack&, RegisterID* constantSymbolTable);
	template<typename LookUpVarKindFunctor>
	bool instantiateLexicalVariables(const VariableEnvironment&, ScopeType, SymbolTable*, ScopeRegisterType, LookUpVarKindFunctor);
	void emitPrefillStackTDZVariables(const VariableEnvironment&, SymbolTable*);
	void popLexicalScopeInternal(VariableEnvironment&);
    RegisterID* emitGetParentScope(RegisterID* dst, RegisterID* scope);
    void emitNewFunctionExpressionCommon(RegisterID*, FunctionMetadataNode*);

    bool isNewTargetUsedInInnerArrowFunction();
    bool isArgumentsUsedInInnerArrowFunction();

    void emitToThis() { emitToThis(&m_thisRegister); }

public:
    bool isSuperUsedInInnerArrowFunction();
    bool isSuperCallUsedInInnerArrowFunction();
    bool isThisUsedInInnerArrowFunction();
    void pushLexicalScope(VariableEnvironmentNode*, ScopeType, TDZCheckOptimization, NestedScopeType = NestedScopeType::IsNotNested, RegisterID** constantSymbolTableResult = 0, bool shouldInitializeBlockScopedFunctions = true);
    void pushClassLexicalScope(VariableEnvironmentNode*);
	void popLexicalScope(VariableEnvironmentNode*);
    void prepareLexicalScopeForNextForLoopIteration(VariableEnvironmentNode*, RegisterID* loopSymbolTable);
    int labelScopeDepth() const;

private:
    void reclaimFreeRegisters();
    Variable variableForLocalEntry(const Identifier&, const SymbolTableEntry&, int symbolTableConstantIndex, bool isLexicallyScoped);

	void emitOpcode(OpcodeID);
    UnlinkedArrayAllocationProfile newArrayAllocationProfile();
    UnlinkedObjectAllocationProfile newObjectAllocationProfile();
    ALWAYS_INLINE UnlinkedValueProfile emitProfiledOpcode(OpcodeID);
	int kill(RegisterID* dst)
	{
		int index = dst->index();
		m_staticPropertyAnalyzer.kill(index);
		return index;
	}

	void retrieveLastBinaryOp(int& dstIndex, int& src1Index, int& src2Index);
	void retrieveLastUnaryOp(int& dstIndex, int& srcIndex);
	ALWAYS_INLINE void rewindBinaryOp();
	ALWAYS_INLINE void rewindUnaryOp();

    void allocateAndEmitScope();

    using BigIntMapEntry = std::tuple<UniquedStringImpl*, uint8_t, bool>;

	typedef HashMap<double, JSValue> NumberMap;
	typedef HashMap<UniquedStringImpl*, JSString*, IdentifierRepHash> IdentifierStringMap;
	using IdentifierBigIntMap = HashMap<BigIntMapEntry, JSValue>;
    typedef HashSet<RefPtr<TemplateObjectDescriptor>> TemplateObjectDescriptorSet;
    typedef HashMap<uint64_t, JSTemplateObjectDescriptor*, WTF::IntHash<uint64_t>, WTF::UnsignedWithZeroKeyHashTraits<uint64_t>> TemplateDescriptorMap;

	// Helper for emitCall() and emitConstruct(). This works because the set of
	// expected functions have identical behavior for both call and construct
	// (i.e. "Object()" is identical to "new Object()").
	ExpectedFunction emitExpectedFunctionSnippet(RegisterID* dst, RegisterID* func, ExpectedFunction, CallArguments&, Label* done);

	RegisterID* emitCall(OpcodeID, RegisterID* dst, RegisterID* func, ExpectedFunction, CallArguments&, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);

	RegisterID* emitCallIterator(RegisterID* iterator, RegisterID* argument, ThrowableExpressionData*);
	RegisterID* newRegister();

    // Adds an anonymous local var slot. To give this slot a name, add it to symbolTable().
	RegisterID* addVar()
	{
		++m_codeBlock->m_numVars;
        RegisterID* result = newRegister();
        ASSERT(VirtualRegister_(result->index()).toLocal() == m_codeBlock->m_numVars - 1);
        result->ref(); // We should never free this slot.
        return result;
	}

	void addParameter(const Identifier&, int parameterIndex);

    // Initializes the stack form the parameter; does nothing for the symbol table.
    RegisterID* initializeNextParameter();
    UniquedStringImpl* visibleNameForParameter(DestructuringPatternNode*);

    RegisterID& registerFor(VirtualRegister_ reg)
    {
        if (reg.isLocal())
            return m_calleeLocals[reg.toLocal()];

        if (reg.offset() == JSStack::Callee)
            return m_calleeRegister;

        ASSERT(m_parameters.size());
        return m_parameters[reg.toArgument()];
    }

	RegisterID& registerFor(int index)
	{
		return registerFor(VirtualRegister_(index));
	}

	unsigned addConstant(const Identifier&);
	RegisterID* addConstantValue(JSValue, SourceCodeRepresentation = SourceCodeRepresentation::Other);
	RegisterID* addConstantEmptyValue();

    FunctionExecutable* makeFunction(FunctionMetadataNode* metadata)
    {
        DerivedContextType newDerivedContextType = DerivedContextType::None;

        NeedsClassFieldInitializer needsClassFieldInitializer = metadata->isConstructorAndNeedsClassFieldInitializer() ? NeedsClassFieldInitializer::Yes : NeedsClassFieldInitializer::No;
        PrivateBrandRequirement privateBrandRequirement = metadata->privateBrandRequirement();
		if (SourceParseModeSet(SourceParseMode::ArrowFunctionMode, SourceParseMode::AsyncArrowFunctionMode, SourceParseMode::AsyncArrowFunctionBodyMode).contains(metadata->parseMode())) {
        	if (constructorKind() == ConstructorKind::Extends || isDerivedConstructorContext()) {
                newDerivedContextType = DerivedContextType::DerivedConstructorContext;
                needsClassFieldInitializer = m_codeBlock->needsClassFieldInitializer();
                privateBrandRequirement = m_codeBlock->privateBrandRequirement();
			}
            else if (m_codeBlock->isClassContext() || isDerivedClassContext())
                newDerivedContextType = DerivedContextType::DerivedMethodContext;
        }

        auto variablesUnderTDZ = getVariablesUnderTDZ();
        Optional<PrivateNameEnvironment> parentPrivateNameEnvironment = getAvailablePrivateAccessNames();
        OwnPtr<Vector<Identifier>> generatorOrAsyncWrapperFunctionParameterNames;

        // FIXME: These flags, ParserModes and propagation to XXXCodeBlocks should be reorganized.
        // https://bugs.webkit.org/show_bug.cgi?id=151547
        SourceParseMode parseMode = metadata->parseMode();
        ConstructAbility constructAbility = constructAbilityForParseMode(parseMode);
        if (parseMode == SourceParseMode::MethodMode && metadata->constructorKind() != ConstructorKind::None)
            constructAbility = ConstructAbility::CanConstruct;

        if (isGeneratorOrAsyncFunctionWrapperParseMode(m_codeBlock->parseMode()) && isGeneratorOrAsyncFunctionBodyParseMode(parseMode))
            generatorOrAsyncWrapperFunctionParameterNames = getParameterNames();

        ScriptExecutable* topLevelExecutable = m_codeBlock->ownerExecutable()->topLevelExecutable();
        return FunctionExecutable::create(*m_vm, topLevelExecutable, metadata, metadata->startColumn(), m_isBuiltinFunction, constructAbility, scriptMode(), variablesUnderTDZ, WTFMove(parentPrivateNameEnvironment), generatorOrAsyncWrapperFunctionParameterNames.release(), newDerivedContextType, needsClassFieldInitializer, privateBrandRequirement, NoIntrinsic);
    }

    RefPtr<TDZEnvironmentLink> getVariablesUnderTDZ();
    Optional<PrivateNameEnvironment> getAvailablePrivateAccessNames();
    PassOwnPtr<Vector<Identifier>> getParameterNames() const;

	RegisterID* emitConstructVarargs(RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);
	RegisterID* emitCallVarargs(OpcodeID, RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd);

    void initializeParameters(FunctionParameters&);
    void initializeVarLexicalEnvironment(int symbolTableConstantIndex, SymbolTable* functionSymbolTable, bool hasCapturedVariables);
    template<typename CapturesFunctor>
    void initializeDefaultParameterValuesAndSetupFunctionScopeStack(FunctionParameters&, bool isSimpleParameterList, FunctionNode*, SymbolTable*, int symbolTableConstantIndex, CapturesFunctor& captures, bool shouldCreateArgumentsVariableInParameterScope);
    void initializeArrowFunctionContextScopeIfNeeded(SymbolTable* functionSymbolTable = 0, bool canReuseLexicalEnvironment = false);
    bool needsDerivedConstructorInArrowFunctionLexicalEnvironment();

    enum class TDZNecessityLevel {
        NotNeeded,
        Optimize,
        DoNotOptimize
    };
    typedef HashMap<RefPtr<UniquedStringImpl>, TDZNecessityLevel, IdentifierRepHash> TDZMap;

    RegisterID* emitMove(RegisterID* dst, RegisterID* src);

public:
    void disablePeepholeOptimization() { m_lastOpcodeID = op_end; }
private:
    bool canDoPeepholeOptimization() const { return m_lastOpcodeID != op_end; }

public:
    JSString* addStringConstant(const Identifier&);
    JSValue addBigIntConstant(const Identifier&, uint8_t radix, bool sign);
    RegisterID* addTemplateObjectConstant(RefPtr<TemplateObjectDescriptor>, int);

	Vector<Instruction>& instructions() { return m_instructions; }

	RegisterID* emitThrowExpressionTooDeepException();

    using TDZStackEntry = std::pair<TDZMap, RefPtr<TDZEnvironmentLink>>;

    class PreservedTDZStack {
    private:
        Vector<TDZStackEntry> m_preservedTDZStack;
        friend class BytecodeGenerator;
    };

    void preserveTDZStack(PreservedTDZStack&);
    void restoreTDZStack(const PreservedTDZStack&);

    PrivateNameEntry getPrivateTraits(const Identifier&);

    void pushPrivateAccessNames(const PrivateNameEnvironment*);
    void popPrivateAccessNames();

    bool lowMemory() const { return m_monitorMallocFailedPreventer.lowMemory(); }

private:
    MonitorMallocFailedPreventer m_monitorMallocFailedPreventer;

	Vector<Instruction> m_instructions;

	bool m_shouldEmitDebugHooks;
	bool m_shouldEmitProfileHooks;
	bool m_shouldEmitRichSourceInfo;

    struct LexicalScopeStackEntry {
        SymbolTable* m_symbolTable;
        RegisterID* m_scope;
        bool m_isWithScope;
        int m_symbolTableConstantIndex;
    };
    Vector<LexicalScopeStackEntry> m_lexicalScopeStack;

    const Vector<Identifier>* m_generatorOrAsyncWrapperFunctionParameterNames { nullptr };
    RefPtr<TDZEnvironmentLink> m_cachedParentTDZ;
    Vector<TDZStackEntry> m_TDZStack;
    Vector<PrivateNameEnvironment> m_privateNamesStack;
    Optional<size_t> m_varScopeLexicalScopeStackIndex;
    void pushTDZVariables(const VariableEnvironment&, TDZCheckOptimization, TDZRequirement);

	ScopeNode* m_scopeNode { nullptr };
	FunctionNode* m_functionNode { nullptr };
	JSGlobalObject* m_globalObject { nullptr };
	CodeBlock* m_codeBlock { nullptr };

	// Some of these objects keep pointers to one another. They are arranged
	// to ensure a sane destruction order that avoids references to freed memory.
	HashSet<RefPtr<UniquedStringImpl>, IdentifierRepHash> m_functions;
	RegisterID m_ignoredResultRegister;
	RegisterID m_thisRegister;
	RegisterID m_calleeRegister;
    RegisterID* m_scopeRegister { nullptr };
    RegisterID* m_argumentsRegister { nullptr };
	RegisterID* m_lexicalEnvironmentRegister { nullptr };
	RegisterID* m_generatorRegister { nullptr };
	RegisterID* m_emptyValueRegister { nullptr };
    RegisterID* m_globalObjectRegister { nullptr };
    RegisterID* m_newTargetRegister { nullptr };
    RegisterID* m_isDerivedConstuctor { nullptr };
    FixedArray<Optional<unsigned>, LinkTimeConstantCount> m_linkTimeConstantCodeBlockConstantIndexes;
    FixedArray<RegisterID*, LinkTimeConstantCount> m_linkTimeConstantRegisters;
    RegisterID* m_arrowFunctionContextLexicalEnvironmentRegister { nullptr };
    RegisterID* m_promiseCapabilityRegister { nullptr };

    FinallyContext* m_currentFinallyContext { nullptr };

	SegmentedVector<RegisterID, 32> m_constantPoolRegisters;
	SegmentedVector<RegisterID, 64> m_calleeLocals;
	SegmentedVector<RegisterID, 16> m_parameters;
	SegmentedVector<Label, 32> m_labels;
    SegmentedVector<LabelScope, 32> m_labelScopes;
	unsigned m_finallyDepth { 0 };
	int m_localScopeDepth { 0 };
	CodeType m_codeType;

    int localScopeDepth() const;
    void pushLocalControlFlowScope();
    void popLocalControlFlowScope();

    // FIXME: Restore overflow checking with UnsafeVectorOverflow once SegmentVector supports it.
    // https://bugs.webkit.org/show_bug.cgi?id=165980
    SegmentedVector<ControlFlowScope, 16> m_controlFlowScopeStack;
	Vector<SwitchInfo> m_switchContextStack;
    Vector<RefPtr<ForInContext>> m_forInContextStack;
	Vector<TryContext> m_tryContextStack;
    unsigned m_yieldPoints { 0 };

    Strong<SymbolTable> m_generatorFrameSymbolTable;
    int m_generatorFrameSymbolTableIndex { 0 };

	enum FunctionVariableType : uint8_t { NormalFunctionVariable, TopLevelFunctionVariable };
	Vector<std::pair<FunctionMetadataNode*, FunctionVariableType>> m_functionsToInitialize;
	bool m_needToInitializeArguments { false };
	RestParameterNode* m_restParameter { nullptr };

	Vector<TryRange> m_tryRanges;
	SegmentedVector<TryData, 8> m_tryData;

    Vector<RefPtr<Label>> m_optionalChainTargetStack;

	int m_nextConstantOffset { 0 };

	// Constant pool
	IdentifierMap m_identifierMap;

    typedef HashMap<EncodedJSValueWithRepresentation, unsigned, EncodedJSValueWithRepresentationHash, EncodedJSValueWithRepresentationHashTraits> JSValueMap;
	JSValueMap m_jsValueMap;
	IdentifierStringMap m_stringMap;
    IdentifierBigIntMap m_bigIntMap;
    TemplateObjectDescriptorSet m_templateObjectDescriptorSet;
    TemplateDescriptorMap m_templateDescriptorMap;

    StaticPropertyAnalyzer m_staticPropertyAnalyzer;

	JSGlobalData* m_vm;

	OpcodeID m_lastOpcodeID { op_end };
#ifndef NDEBUG
	size_t m_lastOpcodePosition { 0 };
#endif

	StackBounds m_stack;

    bool m_usesExceptions { false };
    bool m_expressionTooDeep { false };
    bool m_isBuiltinFunction { false };
    bool m_usesNonStrictEval { false };
    bool m_inTailPosition { false };
    bool m_needsToUpdateArrowFunctionContext { false };
    ECMAMode m_ecmaMode;
    DerivedContextType m_derivedContextType { DerivedContextType::None };

    unsigned m_arrayAllocationProfileCount { 0 };
    unsigned m_objectAllocationProfileCount { 0 };

    Identifier m_functionBodyIdent;

    struct CatchEntry {
        TryData* tryData;
        VirtualRegister_ exceptionRegister;
        VirtualRegister_ thrownValueRegister;
        VirtualRegister_ completionTypeRegister;
    };
    Vector<CatchEntry> m_exceptionHandlersToEmit;
};

class StrictModeScope : private SetForScope<ECMAMode> {
public:
    StrictModeScope(BytecodeGenerator& generator)
        : SetForScope(generator.m_ecmaMode, ECMAMode::strict())
    {
    }
};

ALWAYS_INLINE void BytecodeGenerator::emitOpcode(OpcodeID opcodeID)
{
#ifndef NDEBUG
    size_t opcodePosition = instructions().size();
    ASSERT(opcodePosition - m_lastOpcodePosition == opcodeLengths[m_lastOpcodeID] || m_lastOpcodeID == op_end);
    m_lastOpcodePosition = opcodePosition;
#endif

    unsigned opLength = opcodeLengths[opcodeID];
    if(UNLIKELY(instructions().capacity() < (instructions().size() + opLength)))
    	instructions().expandCapacity(instructions().size() * 2 + opLength);

#if ENABLE(COMPUTED_GOTO_OPCODES)
    instructions().uncheckedAppend(vm().interpreter->getOpcode(opcodeID));
#else
    instructions().uncheckedAppend(opcodeID);
#endif
    m_lastOpcodeID = opcodeID;
}

ALWAYS_INLINE RegisterID* BytecodeGenerator::emitUnaryNoDstOp(OpcodeID opcodeID, RegisterID* src)
{
    emitOpcode(opcodeID);
    instructions().uncheckedAppend(src->index());
    return src;
}

}
