/*
 * Copyright (C) 2011, 2012, 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "DFGDriver.h"

#include "JSObject.h"
#include "JSString.h"

#if ENABLE(DFG_JIT)

#include "CodeBlock.h"
#include "DFGArgumentsEliminationPhase.h"
#include "DFGBackwardsPropagationPhase.h"
#include "DFGByteCodeParser.h"
#include "DFGCFAPhase.h"
#include "DFGCFGSimplificationPhase.h"
#include "DFGCPSRethreadingPhase.h"
#include "DFGCSEPhase.h"
#include "DFGCleanUpPhase.h"
#include "DFGConstantFoldingPhase.h"
#include "DFGDCEPhase.h"
#include "DFGFixupPhase.h"
#include "DFGIntegerCheckCombiningPhase.h"
#include "DFGInvalidationPointInjectionPhase.h"
#include "DFGJITCompiler.h"
#include "DFGLiveCatchVariablePreservationPhase.h"
#include "DFGLoopPreHeaderCreationPhase.h"
#include "DFGPhantomInsertionPhase.h"
#include "DFGPredictionInjectionPhase.h"
#include "DFGPredictionPropagationPhase.h"
#include "DFGStaticExecutionCountEstimationPhase.h"
#include "DFGStackLayoutPhase.h"
#include "DFGStrengthReductionPhase.h"
#include "DFGTypeCheckHoistingPhase.h"
#include "DFGUnificationPhase.h"
#include "DFGValidate.h"
#include "DFGVirtualRegisterAllocationPhase.h"
#include "DFGWatchpointCollectionPhase.h"
#include "ExecutableAllocator.h"
#include "Operations.h"
#include "Options.h"
#include "SamplingTool.h"
#include <wtf/CompilationThread.h>
#include <wtf/FastMallocMonitor.h>
#include <wtf/MemoryMonitor.h>

namespace JSC { namespace DFG {

static unsigned numCompilations;

unsigned getNumCompilations()
{
    return numCompilations;
}

static CompilationResult compileImpl(ExecState* exec, CodeBlock* codeBlock, unsigned osrEntryBytecodeIndex)
{
    SamplingRegion samplingRegion("DFG Compilation (Driver)");
    CompilationScope compilationScope;
    
    numCompilations++;
    
    ASSERT(codeBlock);
    ASSERT(codeBlock->alternative());
    ASSERT(codeBlock->alternative()->getJITType() == JITCode::BaselineJIT);
    
    ASSERT(osrEntryBytecodeIndex != UINT_MAX);

    if (!MacroAssembler::supportsFloatingPoint())
        return CompilationFailed;

    if (logCompilationChanges())
        dataLog("DFG compiling ", *codeBlock, ", number of instructions = ", codeBlock->instructionCount(), "\n");

	MonitorMallocFailedPreventer monitorMallocFailedPreventer;

    VM& vm = codeBlock->vm();

#if 0
    // Make sure that any stubs that the DFG is going to use are initialized. We want to
    // make sure that all JIT code generation does finalization on the main thread.
    vm.getCTIStub(callToJavaScript);
    vm.getCTIStub(throwNotCaught);
#if ENABLE(FTL_JIT)
    vm.getCTIStub(osrExitGenerationThunkGenerator);
#endif
    vm.getCTIStub(throwExceptionFromCallSlowPathGenerator);
    vm.getCTIStub(linkCallThunkGenerator);
    vm.getCTIStub(linkPolymorphicCallThunkGenerator);
#endif
    
    // Derive our set of must-handle values. The compilation must be at least conservative
    // enough to allow for OSR entry with these values.
    unsigned numVarsWithValues;
    if (osrEntryBytecodeIndex)
        numVarsWithValues = codeBlock->m_numCalleeLocals;
    else
        numVarsWithValues = 0;
    Operands<JSValue> mustHandleValues(codeBlock->numParameters(), numVarsWithValues);
    for (size_t i = 0; i < mustHandleValues.size(); ++i) {
        int operand = mustHandleValues.operandForIndex(i);
        mustHandleValues[i] = exec->uncheckedR(operand).jsValue();
    }
    
    OwnPtr<Graph> dfgPtr = Graph::create(vm, codeBlock, osrEntryBytecodeIndex, mustHandleValues);
    if(UNLIKELY(!dfgPtr))
        return CompilationFailed;

    Graph& dfg = *dfgPtr;
    if (!parse(dfg))
        return CompilationFailed;
    
    // By this point the DFG bytecode parser will have potentially mutated various tables
    // in the CodeBlock. This is a good time to perform an early shrink, which is more
    // powerful than a late one. It's safe to do so because we haven't generated any code
    // that references any of the tables directly, yet.
    codeBlock->shrinkToFit(CodeBlock::EarlyShrink);

    if (validationEnabled())
        validate(dfg);

    performCPSRethreading(dfg);
    performUnification(dfg);
    performPredictionInjection(dfg);
    
    if (validationEnabled())
        validate(dfg);

#if ENABLE(FTL_JIT)
    performStaticExecutionCountEstimation(dfg);
#endif
    
    performPredictionPropagation(dfg);
    performFixup(dfg);
    performInvalidationPointInjection(dfg);
    performTypeCheckHoisting(dfg);
    
    dfg.m_fixpointState = FixpointNotConverged;

    // For now we're back to avoiding a fixpoint. Note that we've ping-ponged on this decision
	// many times. For maximum throughput, it's best to fixpoint. But the throughput benefit is
	// small and not likely to show up in FTL anyway. On the other hand, not fixpointing means
	// that the compiler compiles more quickly. We want the third tier to compile quickly, which
	// not fixpointing accomplishes; and the fourth tier shouldn't need a fixpoint.

	//if (validationEnabled())
	//    validate(dfg);

    performBackwardsPropagation(dfg);
	performStrengthReduction(dfg);
	performCPSRethreading(dfg);
	performCFA(dfg);
	performConstantFolding(dfg);
	bool changed = false;
	changed |= performCFGSimplification(dfg);
	changed |= performLocalCSE(dfg);

	if (validationEnabled())
		validate(dfg);

	performCPSRethreading(dfg);
#if 0
    if (!isFTL(mode)) {
        // Only run this if we're not FTLing, because currently for a LoadVarargs that is forwardable and
        // in a non-varargs inlined call frame, this will generate ForwardVarargs while the FTL
        // ArgumentsEliminationPhase will create a sequence of GetStack+PutStacks. The GetStack+PutStack
        // sequence then gets sunk, eliminating anything that looks like an escape for subsequent phases,
        // while the ForwardVarargs doesn't get simplified until later (or not at all) and looks like an
        // escape for all of the arguments. This then disables object allocation sinking.
        //
        // So, for now, we just disable this phase for the FTL.
        //
        // If we wanted to enable it, we'd have to do any of the following:
        // - Enable ForwardVarargs->GetStack+PutStack strength reduction, and have that run before
        //   PutStack sinking and object allocation sinking.
        // - Make VarargsForwarding emit a GetLocal+SetLocal sequence, that we can later turn into
        //   GetStack+PutStack.
        //
        // But, it's not super valuable to enable those optimizations, since the FTL
        // ArgumentsEliminationPhase does everything that this phase does, and it doesn't introduce this
        // pathology.

        changed |= performVarargsForwarding(dfg); // Do this after CFG simplification and CPS rethreading.
    }
#endif
	if (changed) {
		performCFA(dfg);
		performConstantFolding(dfg);
	}

    // If we're doing validation, then run some analyses, to give them an opportunity
    // to self-validate. Now is as good a time as any to do this.
    if (validationEnabled()) {
#if ENABLE(FTL_JIT)
        dfg.m_dominators.computeIfNecessary(dfg);
        dfg.m_naturalLoops.computeIfNecessary(dfg);
#endif
    }

#if ENABLE(FTL_JIT)
    if (Options::useExperimentalFTL()
        && compileMode == CompileFunction
        && FTL::canCompile(dfg)) {

        performCriticalEdgeBreaking(dfg);
        performLoopPreHeaderCreation(dfg);
        performCPSRethreading(dfg);
        performSSAConversion(dfg);
        performLivenessAnalysis(dfg);
        performCFA(dfg);
        performLICM(dfg);
        performGlobalCSE(dfg);
        performDCE(dfg); // We rely on this to convert dead SetLocals into the appropriate hint, and to kill dead code that won't be recognized as dead by LLVM.
        performStackLayout(dfg);
        performLivenessAnalysis(dfg);
        performFlushLivenessAnalysis(dfg);
        performOSRAvailabilityAnalysis(dfg);

        dumpAndVerifyGraph(dfg, "Graph just before FTL lowering:");

        // FIXME: Support OSR entry.
        // https://bugs.webkit.org/show_bug.cgi?id=113625

        FTL::State state(dfg);
        FTL::lowerDFGToLLVM(state);
        return FTL::compile(state, jitCode, *jitCodeWithArityCheck);
    }
#endif // ENABLE(FTL_JIT)

    dfg.m_fixpointState = FixpointConverged;

    performCleanUp(dfg);
    performCPSRethreading(dfg);
    performDCE(dfg);
	performPhantomInsertion(dfg);
    performStackLayout(dfg);
    performVirtualRegisterAllocation(dfg);
    performWatchpointCollection(dfg);

    GraphDumpMode modeForFinalValidate = DumpGraph;
    if (verboseCompilationEnabled()) {
        dataLogF("Graph after optimization:\n");
        dfg.dump();
        modeForFinalValidate = DontDumpGraph;
    }
    if (validationEnabled())
        validate(dfg, modeForFinalValidate);

    if (monitorMallocFailedPreventer.lowMemory())
        return CompilationFailed;
    
    JITCompiler dataFlowJIT(dfg);
    bool result;
    RefPtr<JSC::JITCode> jitCode;
    MacroAssemblerCodePtr jitCodeWithArityCheck;
    if (codeBlock->codeType() == FunctionCode) {
        if (!dataFlowJIT.compileFunction())
            return CompilationFailed;
        compilationScope.leaveEarly();
        result = dataFlowJIT.linkFunction(jitCode, jitCodeWithArityCheck);
    } else {
        if (!dataFlowJIT.compile())
            return CompilationFailed;
        compilationScope.leaveEarly();
        result = dataFlowJIT.link(jitCode);
    }

    if(jitCode) {
    	codeBlock->setJITCode(jitCode);

		// Since Plan::reallyAdd could fire watchpoints (see ArrayBufferViewWatchpointAdaptor::add), it is possible that the current CodeBlock is now invalidated.
		if (!jitCode->dfgCommon()->isStillValid) {
			//CODEBLOCK_LOG_EVENT(m_codeBlock, "dfgFinalize", ("invalidated"));
			return CompilationFailed;
		}
    }

    if (monitorMallocFailedPreventer.lowMemory())
        return CompilationFailed;

    return result ? CompilationSuccessful : CompilationFailed;
}

CompilationResult tryCompile(ExecState* exec, CodeBlock* codeBlock, unsigned bytecodeIndex)
{
    if (WTF::isLowMemory() || WTF::isUnderMemoryPressure())
        return CompilationFailed;

    double compileStartTime = WTF::monotonicallyIncreasingTime();

	CompilationResult result = compileImpl(exec, codeBlock, bytecodeIndex);

    double compileEndTime = WTF::monotonicallyIncreasingTime();
    codeBlock->vm().jitCompileLength += (compileEndTime - compileStartTime);

    return result;
}

} } // namespace JSC::DFG

#endif // ENABLE(DFG_JIT)

