/*
 * Copyright (C) 2015 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "DFGLiveCatchVariablePreservationPhase.h"

#if ENABLE(DFG_JIT)

#include "DFGBasicBlockInlines.h"
#include "DFGGraph.h"
#include "DFGInsertionSet.h"
#include "DFGPhase.h"
#include "FullBytecodeLiveness.h"
#include "JSCInlines.h"

namespace JSC { namespace DFG {

class LiveCatchVariablePreservationPhase {
public:
	LiveCatchVariablePreservationPhase(Graph& graph)
		: m_graph(graph)
    {
    }

    bool run()
    {
        if (!m_graph.m_hasExceptionHandlers)
            return false;

        m_graph.determineReachability();

        //DFG_ASSERT(m_graph, 0, m_graph.m_form == LoadStore);

        InsertionSet insertionSet(m_graph);
        for (BasicBlock* block : m_graph.blocksInNaturalOrder()) {
			if (block->isReachable) {
				handleBlock(block, insertionSet);
				insertionSet.execute(block);
			}
        }

        m_graph.clearReachability();
        return true;
    }

    void handleBlock(BasicBlock* block, InsertionSet& insertionSet)
    {
        FastBitVector liveAtCatchHead;
        liveAtCatchHead.resize(m_graph.block(0)->variablesAtTail.numberOfLocals());

        auto willCatchException = [&] (CodeOrigin origin) -> bool {
            unsigned bytecodeIndexToCheck = origin.bytecodeIndex;

            while (1) {
                InlineCallFrame* inlineCallFrame = origin.inlineCallFrame;
                CodeBlock* codeBlock = m_graph.baselineCodeBlockFor(inlineCallFrame);
                if (HandlerInfo* handler = codeBlock->handlerForBytecodeOffset(bytecodeIndexToCheck)) {
                	liveAtCatchHead.clearAll();

                    unsigned catchBytecodeIndex = handler->target;
                    m_graph.forAllLocalsLiveInBytecode(CodeOrigin(catchBytecodeIndex, inlineCallFrame), [&] (int operand) {
                		liveAtCatchHead[VirtualRegister_(operand).toLocal()] = true;
                    });
                    return true;
                }

                if (!inlineCallFrame)
                    return false;

                bytecodeIndexToCheck = inlineCallFrame->directCaller.bytecodeIndex;
                origin = inlineCallFrame->directCaller;
            }

            return false;
        };

        // Because precise jump targets ensures that the start of a "try" block is its
        // own basic block, we will never have two "try" statements in the same DFG
        // basic block. Therefore, checking the first node in the block is sufficient 
        // to checking if we're in a try block.
        if (!willCatchException(block->at(0)->origin.semantic))
            return;

        Operands<VariableAccessData*> currentBlockAccessData(block->variablesAtTail.numberOfArguments(), block->variablesAtTail.numberOfLocals(), 0);

        {
            for (unsigned i = 0; i < block->size(); i++) {
                Node* node = block->at(i);
                bool isPrimordialSetArgument = node->op() == SetArgument && VirtualRegister_(node->local()).isArgument() && node == m_graph.m_arguments[VirtualRegister_(node->local()).toArgument()];
                InlineCallFrame* inlineCallFrame = node->origin.semantic.inlineCallFrame;

                if (node->op() == SetLocal || (node->op() == SetArgument && !isPrimordialSetArgument)) {
                    VirtualRegister_ operand = VirtualRegister_(node->local());
                    if ((operand.isLocal() && liveAtCatchHead[operand.toLocal()]) || operand.isArgument()) {
                    	ASSERT(isValidFlushLocation(block, nodeIndex, operand));

                        VariableAccessData* variableAccessData = currentBlockAccessData.operand(operand.offset());
                        if (!variableAccessData)
                        	variableAccessData = newVariableAccessData(static_cast<VirtualRegister>(operand.offset()));

                        insertionSet.insertNode(i, SpecNone, 
							Flush, node->origin, OpInfo(variableAccessData));
                    }
                }

                if (node->accessesStack(m_graph))
                    currentBlockAccessData.operand(node->local()) = node->variableAccessData();
            }
        }

        // Insert Flush for everything at the end of the block.
        {
            NodeOrigin origin = block->at(block->size() - 1)->origin;
            auto flush = [&] (int operand) {
				if ((VirtualRegister_(operand).isLocal() && liveAtCatchHead[VirtualRegister_(operand).toLocal()])
					|| VirtualRegister_(operand).isArgument()) {
					VariableAccessData* accessData = currentBlockAccessData.operand(operand);
					if (!accessData)
						accessData = newVariableAccessData(static_cast<VirtualRegister>(operand));

					currentBlockAccessData.operand(operand) = accessData;

					insertionSet.insertNode(block->size(), SpecNone,
						Flush, origin, OpInfo(accessData));
				}
			};
            for (unsigned local = 0; local < block->variablesAtTail.numberOfLocals(); local++)
            	flush(virtualRegisterForLocal(local));
            flush(CallFrame::thisArgumentOffset());
        }
    }

    VariableAccessData* newVariableAccessData(VirtualRegister operand)
    {
        ASSERT(!operand.isConstant());
        
        m_graph.m_variableAccessData.append(VariableAccessData(operand));
        return &m_graph.m_variableAccessData.last();
    }

    Graph& m_graph;
};

bool performLiveCatchVariablePreservationPhase(Graph& graph)
{
    SamplingRegion samplingRegion("DFG Live Catch Variables Preservation Phase");
    return runPhase<LiveCatchVariablePreservationPhase>(graph);
}

} } // namespace JSC::DFG

#endif // ENABLE(DFG_JIT)
