/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "CodeBlockSet.h"

#include "CodeBlock.h"
#include "Heap.h"
#include "SlotVisitor.h"

namespace JSC {

namespace CodeBlockSetInternal {
static const bool verbose = false;
}

CodeBlockSet::CodeBlockSet() { }

CodeBlockSet::~CodeBlockSet()
{
    HashSet<CodeBlock*>::iterator iter = m_set.begin();
    HashSet<CodeBlock*>::iterator end = m_set.end();
    for (; iter != end; ++iter)
        (*iter)->deref();
}

void CodeBlockSet::add(PassRefPtr<CodeBlock> codeBlock)
{
    bool isNewEntry = m_set.add(codeBlock.leakRef()).isNewEntry;
    ASSERT_UNUSED(isNewEntry, isNewEntry);
}

void CodeBlockSet::clearMarks()
{
    HashSet<CodeBlock*>::iterator end = m_set.end();
    for (auto iter = m_set.begin(); iter != end; ++iter) {
        CodeBlock* codeBlock = *iter;
        codeBlock->m_mayBeExecuting = false;
        codeBlock->m_didVisitAggregate = false;
#if ENABLE(PARALLEL_GC)
        codeBlock->m_visitAggregateHasBeenCalled = false;
#endif
    }
}

void CodeBlockSet::deleteUnmarkedAndUnreferenced()
{
    // This needs to be a fixpoint because code blocks that are unmarked may
    // refer to each other. For example, a DFG code block that is owned by
    // the GC may refer to an FTL for-entry code block that is also owned by
    // the GC.
    if (CodeBlockSetInternal::verbose)
        dataLog("Fixpointing over unmarked, set size = ", m_set.size(), "...\n");
    for (;;) {
        class Functor
        {
        public:
        	bool operator()(HashSet<CodeBlock*>::ValueType& iterator) const
        	{
                CodeBlock& codeBlock = *iterator;
                if (!codeBlock.hasOneRef())
                    return false;
                if (codeBlock.m_mayBeExecuting)
                    return false;
                codeBlock.deref();
                return true;
        	}
        };
        Functor functor;

        bool didRemove = m_set.removeIf(functor);
        if(!didRemove)
        	return;
    }
}

void CodeBlockSet::traceMarked(SlotVisitor& visitor)
{
    if (CodeBlockSetInternal::verbose)
        dataLog("Tracing ", m_set.size(), " code blocks.\n");
    HashSet<CodeBlock*>::iterator end = m_set.end();
    for (auto iter = m_set.begin(); iter != end; ++iter) {
        CodeBlock* codeBlock = *iter;
        if (!codeBlock->m_mayBeExecuting)
            continue;
        if(codeBlock->m_didVisitAggregate)
        	continue;
        if(codeBlock->m_gcCollected) {
        	WTF::logMessage(ASCIILiteral("CodeBlock is GC collected."));
        	continue;
        }

        codeBlock->visitAggregate(visitor);
    }
}

void CodeBlockSet::traceGCDeleted()
{
    if (CodeBlockSetInternal::verbose)
        dataLog("Tracing ", m_set.size(), " code blocks.\n");
    HashSet<CodeBlock*>::iterator end = m_set.end();
    for (auto iter = m_set.begin(); iter != end; ++iter) {
        CodeBlock* codeBlock = *iter;
        if(codeBlock->m_didVisitAggregate)
        	continue;

        codeBlock->unlinkIncomingCalls();
        codeBlock->m_gcCollected = true;
    }
}

} // namespace JSC

