/*
 * Copyright (C) 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "CopyVisitor.h"

#include "CopiedBlockInlines.h"
#include "CopyVisitorInlineMethods.h"
#include "CopyWorkList.h"
#include "JSCell.h"
#include "JSObject.h"
#include "JSScope.h"
#include "Operations.h"
#include <wtf/Threading.h>

namespace JSC {

CopyVisitor::CopyVisitor(Heap& heap)
    : m_heap(heap)
{
}

inline void CopyVisitor::visitItem(CopyWorklistItem item)
{
    if (item.token() == ButterflyCopyToken) {
        JSObject::copyBackingStore(item.cell(), *this, ButterflyCopyToken);
        return;
    }

    item.cell()->methodTable()->copyBackingStore(item.cell(), *this, item.token());
}

void CopyVisitor::copyFromShared()
{
    size_t next, end;
    m_heap.getNextBlocksToCopy(next, end);
    while (next < end) {
        for (; next < end; ++next) {
            CopiedBlock* block = m_heap.m_blocksToCopy[next];
            if (!block->hasWorkList())
                continue;

            auto& workList = block->workList();
            for (auto it = workList.begin(); it != workList.end(); ++it)
            	visitItem(*it);

            ASSERT(!block->liveBytes());
            m_heap.m_storageSpace.recycleEvacuatedBlock(block);
        }
        m_heap.getNextBlocksToCopy(next, end);
    }
    ASSERT(next == end);
}

void* CopyVisitor::allocateNewSpace(size_t bytes)
{
#if DEBUG_JSC_COPIED_SPACE
    if(!bytes || !is8ByteAligned(reinterpret_cast<void*>(bytes)))
    	CRASH();
#endif

    void* result = nullptr; // Compilers don't realize that this will be assigned.
    if (LIKELY(m_copiedAllocator.tryAllocate(bytes, &result)))
        return result;

    result = allocateNewSpaceSlow(bytes);
    ASSERT(result);
    return result;
}

void* CopyVisitor::allocateNewSpaceSlow(size_t bytes)
{
    CopiedBlock* newBlock = m_heap.m_storageSpace.allocateBlockForCopyingPhase();
    m_heap.m_storageSpace.doneFillingBlock(m_copiedAllocator.resetCurrentBlock());
    m_copiedAllocator.setCurrentBlock(newBlock);

    void* result = nullptr;
    CheckedBoolean didSucceed = m_copiedAllocator.tryAllocate(bytes, &result);
    ASSERT(didSucceed);
    return result;
}

} // namespace JSC
