#include "config.h"
#include "SlotVisitor.h"
#include "SlotVisitorInlines.h"

#include "ConservativeRoots.h"
#include "CopiedBlockInlines.h"
#include "CopiedSpace.h"
#include "CopiedSpaceInlineMethods.h"
#include "GCThread.h"
#include "JSArray.h"
#include "JSDestructibleObject.h"
#include "JSGlobalData.h"
#include "JSGlobalObject.h"
#include "JSObject.h"
#include "JSScope.h"
#include "JSStringInlines.h"
#include "Operations.h"
#include <wtf/HashSet.h>
#include <wtf/StackStats.h>

namespace JSC {

#if ENABLE(GC_VALIDATION)
static void validate(JSCell* cell)
{
    if (!cell) {
        dataLog("cell is NULL\n");
        RELEASE_ASSERT_NOT_REACHED();
    }

    if (!cell->structure()) {
        dataLog("cell at %p has a null structure\n" , cell);
        RELEASE_ASSERT_NOT_REACHED();
    }

    // Both the cell's structure, and the cell's structure's structure should be the Structure Structure.
    // I hate this sentence.
    if (cell->structure()->structure()->JSCell::classInfo() != cell->structure()->JSCell::classInfo()) {
        const char* parentClassName = 0;
        const char* ourClassName = 0;
        if (cell->structure()->structure() && cell->structure()->structure()->JSCell::classInfo())
            parentClassName = cell->structure()->structure()->JSCell::classInfo()->className;
        if (cell->structure()->JSCell::classInfo())
            ourClassName = cell->structure()->JSCell::classInfo()->className;
        dataLog("parent structure (%p <%s>) of cell at %p doesn't match cell's structure (%p <%s>)\n",
                cell->structure()->structure(), parentClassName, cell, cell->structure(), ourClassName);
        RELEASE_ASSERT_NOT_REACHED();
    }

    // Make sure we can walk the ClassInfo chain
    const ClassInfo* info = cell->classInfo();
    do { } while ((info = info->parentClass));
}
#else
static void validate(JSCell*)
{
}
#endif

SlotVisitor::SlotVisitor(Heap& heap)
: m_stack()
, m_bytesVisited(0)
, m_bytesCopied(0)
, m_visitCount(0)
, m_isInParallelMode(false)
, m_heap(heap)
#if !ASSERT_DISABLED
, m_isCheckingForDefaultMarkViolation(false)
, m_isDraining(false)
#endif
{
}

SlotVisitor::~SlotVisitor()
{
    ASSERT(m_stack.isEmpty());
}

void SlotVisitor::reset()
{
    m_bytesVisited = 0;
    m_bytesCopied = 0;
    m_visitCount = 0;
    ASSERT(m_stack.isEmpty());
#if ENABLE(PARALLEL_GC)
    ASSERT(m_opaqueRoots.isEmpty()); // Should have merged by now.
#else
    m_opaqueRoots.clear();
#endif
}

void SlotVisitor::append(ConservativeRoots& conservativeRoots)
{
    JSCell** roots = conservativeRoots.roots();
    size_t size = conservativeRoots.size();
    for (size_t i = 0; i < size; ++i)
    	setMarkedAndAppendToMarkStack(roots[i]);
}

static void visitChildren(SlotVisitor& visitor, const JSCell* cell)
{
    ASSERT(Heap::isMarked(cell));

    switch(cell->structure()->typeInfo().type()) {
    case StringType:
        JSString::visitChildren(const_cast<JSCell*>(cell), visitor);
        return;

    case FinalObjectType:
		JSFinalObject::visitChildren(const_cast<JSCell*>(cell), visitor);
        return;

    case ArrayType:
        JSArray::visitChildren(const_cast<JSCell*>(cell), visitor);
        return;

    default:
    	cell->methodTable()->visitChildren(const_cast<JSCell*>(cell), visitor);
    	return;
    }
}

void SlotVisitor::donateKnownParallel()
{
    // NOTE: Because we re-try often, we can afford to be conservative, and
    // assume that donating is not profitable.

    // Avoid locking when a thread reaches a dead end in the object graph.
    if (m_stack.size() < 2)
        return;

    // If there's already some shared work queued up, be conservative and assume
    // that donating more is not profitable.
    if (m_heap.m_sharedMarkStack.size())
        return;

    // If we're contending on the lock, be conservative and assume that another
    // thread is already donating.
    MutexTryLocker locker(m_heap.m_markingLock);
    if (!locker.locked())
        return;

    // Otherwise, assume that a thread will go idle soon, and donate.
    m_stack.donateSomeCellsTo(m_heap.m_sharedMarkStack);

    if (m_heap.m_numberOfActiveParallelMarkers < Options::numberOfGCMarkers())
        m_heap.m_markingCondition.broadcast();
}

void SlotVisitor::drain()
{
    ASSERT(m_isInParallelMode);
   
#if ENABLE(PARALLEL_GC)
    if (Options::numberOfGCMarkers() > 1) {
        while (!m_stack.isEmpty()) {
            m_stack.refill();
            for (unsigned countdown = Options::minimumNumberOfScansBetweenRebalance(); m_stack.canRemoveLast() && countdown--;)
                visitChildren(*this, m_stack.removeLast());
            donateKnownParallel();
        }
        
        mergeOpaqueRootsIfNecessary();
        return;
    }
#endif
    
    while (!m_stack.isEmpty()) {
        m_stack.refill();
        while (m_stack.canRemoveLast())
            visitChildren(*this, m_stack.removeLast());
    }
}

void SlotVisitor::drainFromShared(SharedDrainMode sharedDrainMode)
{
    ASSERT(m_isInParallelMode);
    
    ASSERT(Options::numberOfGCMarkers());
    
    bool shouldBeParallel;

#if ENABLE(PARALLEL_GC)
    shouldBeParallel = Options::numberOfGCMarkers() > 1;
#else
    ASSERT(Options::numberOfGCMarkers() == 1);
    shouldBeParallel = false;
#endif
    
    if (!shouldBeParallel) {
        // This call should be a no-op.
        ASSERT_UNUSED(sharedDrainMode, sharedDrainMode == MasterDrain);
        ASSERT(m_stack.isEmpty());
        ASSERT(m_heap.m_sharedMarkStack.isEmpty());
        return;
    }
    
#if ENABLE(PARALLEL_GC)
    {
        MutexLocker locker(m_heap.m_markingLock);
        m_heap.m_numberOfActiveParallelMarkers++;
    }
    while (true) {
        {
            MutexLocker locker(m_heap.m_markingLock);
            m_heap.m_numberOfActiveParallelMarkers--;

            // How we wait differs depending on drain mode.
            if (sharedDrainMode == MasterDrain) {
                // Wait until either termination is reached, or until there is some work
                // for us to do.
                while (true) {
                    // Did we reach termination?
                    if (!m_heap.m_numberOfActiveParallelMarkers && m_heap.m_sharedMarkStack.isEmpty()) {
                        // Let any sleeping slaves know it's time for return;
                        m_heap.m_markingCondition.broadcast();
                        return;
                    }
                    
                    // Is there work to be done?
                    if (!m_heap.m_sharedMarkStack.isEmpty())
                        break;
                    
                    // Otherwise wait.
                    m_heap.m_markingCondition.wait(m_heap.m_markingLock);
                }
            } else {
                ASSERT(sharedDrainMode == SlaveDrain);
                
                // Did we detect termination? If so, let the master know.
                if (!m_heap.m_numberOfActiveParallelMarkers && m_heap.m_sharedMarkStack.isEmpty())
                    m_heap.m_markingCondition.broadcast();
                
                while (m_heap.m_sharedMarkStack.isEmpty() && !m_heap.m_parallelMarkersShouldExit)
				   m_heap.m_markingCondition.wait(m_heap.m_markingLock);

			   // Is the current phase done? If so, return from this function.
			   if (m_heap.m_parallelMarkersShouldExit)
				   return;
		   }
           
            size_t idleThreadCount = Options::numberOfGCMarkers() - m_heap.m_numberOfActiveParallelMarkers;
            m_stack.stealSomeCellsFrom(m_heap.m_sharedMarkStack, idleThreadCount);
            m_heap.m_numberOfActiveParallelMarkers++;
        }
        
        drain();
    }
#endif
}


#if ENABLE(PARALLEL_GC)
inline void SlotVisitor::mergeOpaqueRootsIfNecessary()
{
    if (m_opaqueRoots.size())
    	mergeOpaqueRoots();
}

inline void SlotVisitor::mergeOpaqueRootsIfProfitable()
{
    if (static_cast<unsigned>(m_opaqueRoots.size()) < Options::opaqueRootMergeThreshold())
        return;
    mergeOpaqueRoots();
}
#endif

inline void SlotVisitor::donate()
{
    ASSERT(m_isInParallelMode);
    if (Options::numberOfGCMarkers() == 1)
        return;

    donateKnownParallel();
}

void SlotVisitor::donateAndDrain()
{
    donate();
    drain();
}

void SlotVisitor::copyLater(JSCell* owner, CopyToken token, void* ptr, size_t bytes)
{
    ASSERT(bytes);

    m_bytesCopied += bytes;
    CopiedBlock* block = CopiedSpace::blockFor(ptr);
    if (block->isOversize()) {
    	m_heap.m_storageSpace.pin(block);
        return;
    }

    if (block->isPinned())
        return;

    block->reportLiveBytes(owner, token, bytes);
}

#if ENABLE(PARALLEL_GC)
void SlotVisitor::mergeOpaqueRoots()
{
    ASSERT(!m_opaqueRoots.isEmpty()); // Should only be called when opaque roots are non-empty.
    {
        MutexLocker locker(m_heap.m_opaqueRootsLock);
        auto end = m_opaqueRoots.end();
        for (auto iter = m_opaqueRoots.begin(); iter != end; ++iter)
            m_heap.m_opaqueRoots.add(*iter);
    }
    m_opaqueRoots.clear();
}
#endif

void SlotVisitor::harvestWeakReferences()
{
    for (WeakReferenceHarvester* current = m_heap.m_weakReferenceHarvesters.head(); current; current = current->next())
        current->visitWeakReferences(*this);
}

void SlotVisitor::finalizeUnconditionalFinalizers()
{
    while (m_heap.m_unconditionalFinalizers.hasNext())
    	m_heap.m_unconditionalFinalizers.removeNext()->finalizeUnconditionally(*m_heap.m_vm);
}

} // namespace JSC
