/*
 * Copyright (C) 2009, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef Weak_h
#define Weak_h

#include "PassWeak.h"
#include <wtf/Assertions.h>
#include <wtf/HashMap.h>

namespace JSC {

// This is a free function rather than a Weak<T> member function so we can put it in Weak.cpp.
JS_EXPORT_PRIVATE void weakClearSlowCase(WeakImpl*&);

template<typename T> class Weak : public WeakImplAccessor<Weak<T>, T> {
    WTF_MAKE_NONCOPYABLE(Weak);
public:
    friend class WeakImplAccessor<Weak<T>, T>;
    typedef typename WeakImplAccessor<Weak<T>, T>::GetType GetType;

    Weak() = default;

    explicit Weak(std::nullptr_t)
	{
	}
    explicit Weak(GetType, WeakHandleOwner* = nullptr, void* context = nullptr);

    bool isHashTableDeletedValue() const;
    Weak(WTF::HashTableDeletedValueType);
    bool isHashTableEmptyValue() const { return !m_impl; }

    Weak(Weak&&);
    template<typename U> Weak(const PassWeak<U>&);

    ~Weak()
    {
        clear();
    }

    Weak& operator=(Weak&&);
    void swap(Weak&);
    Weak& operator=(const PassWeak<T>&);
    
    bool operator!() const;
    T* operator->() const;
    T& operator*() const;
    T* get() const;

    bool was(T*) const;

    explicit operator bool() const;

    PassWeak<T> release();
    WeakImpl* leakImpl() WARN_UNUSED_RETURN;
    WeakImpl* unsafeImpl() const { return m_impl; }
    void clear()
    {
        if (!m_impl)
            return;
        weakClearSlowCase(m_impl);
    }
    
private:
    static WeakImpl* hashTableDeletedValue();

    WeakImpl* m_impl { nullptr };
};

template<typename T> ALWAYS_INLINE Weak<T>::Weak(typename Weak<T>::GetType getType, WeakHandleOwner* weakOwner, void* context)
    : m_impl(getType ? WeakSet::allocate(getType, weakOwner, context) : 0)
{
}

template<typename T> ALWAYS_INLINE bool Weak<T>::isHashTableDeletedValue() const
{
    return m_impl == hashTableDeletedValue();
}

template<typename T> ALWAYS_INLINE Weak<T>::Weak(WTF::HashTableDeletedValueType)
    : m_impl(hashTableDeletedValue())
{
}

template<typename T> template<typename U>  ALWAYS_INLINE Weak<T>::Weak(const PassWeak<U>& other)
    : m_impl(other.leakImpl())
{
}

template<class T> ALWAYS_INLINE void swap(Weak<T>& a, Weak<T>& b)
{
    a.swap(b);
}

template<typename T> inline Weak<T>::Weak(Weak&& other)
    : m_impl(other.leakImpl())
{
}

template<typename T> ALWAYS_INLINE void Weak<T>::swap(Weak& other)
{
    std::swap(m_impl, other.m_impl);
}

template<typename T> ALWAYS_INLINE Weak<T>& Weak<T>::operator=(const PassWeak<T>& o)
{
    clear();
    m_impl = o.leakImpl();
    return *this;
}

template<typename T> inline auto Weak<T>::operator=(Weak&& other) -> Weak&
{
    Weak weak = WTFMove(other);
    swap(weak);
    return *this;
}

template<typename T> inline bool Weak<T>::was(T* other) const
{
    return static_cast<T*>(m_impl->jsValue().asCell()) == other;
}

template<typename T> ALWAYS_INLINE bool Weak<T>::operator!() const
{
    return !m_impl || !m_impl->jsValue() || m_impl->state() != WeakImpl::Live;
}

template<typename T> ALWAYS_INLINE Weak<T>::operator bool() const
{
    return !!*this;
}

template<typename T> ALWAYS_INLINE PassWeak<T> Weak<T>::release()
{
    PassWeak<T> tmp = adoptWeak<T>(m_impl);
    m_impl = 0;
    return tmp;
}

template<typename T> inline WeakImpl* Weak<T>::leakImpl()
{
    WeakImpl* impl = m_impl;
    m_impl = 0;
    return impl;
}
    
template<typename T> ALWAYS_INLINE WeakImpl* Weak<T>::hashTableDeletedValue()
{
    return reinterpret_cast<WeakImpl*>(-1);
}

template <typename T> inline bool operator==(const Weak<T>& lhs, const Weak<T>& rhs)
{
    return lhs.get() == rhs.get();
}

// This function helps avoid modifying a weak table while holding an iterator into it. (Object allocation
// can run a finalizer that modifies the table. We avoid that by requiring a pre-constructed object as our value.)
template<typename Map, typename Key, typename Value> ALWAYS_INLINE void weakAdd(Map& map, const Key& key, Value value)
{
    ASSERT(!map.get(key));
    map.set(key, value); // The table may still have a zombie for value.
}

template<typename Map, typename Key, typename Value> ALWAYS_INLINE void weakRemove(Map& map, const Key& key, Value value)
{
    typename Map::iterator it = map.find(key);
    ASSERT_UNUSED(value, value);
    ASSERT(it != map.end());
    ASSERT(it->second.was(value));
    ASSERT(!it->second);
    map.remove(it);
}

template<typename T> ALWAYS_INLINE void weakClear(Weak<T>& weak, typename Weak<T>::GetType value)
{
    ASSERT_UNUSED(value, value);
    ASSERT(weak.was(value));
    ASSERT(!weak);
    weak.clear();
}

} // namespace JSC

namespace WTF {

template<typename T> struct VectorTraits<JSC::Weak<T> > : SimpleClassVectorTraits {
    static const bool canCompareWithMemcmp = false;
};

template<typename T> struct HashTraits<JSC::Weak<T> > : SimpleClassHashTraits<JSC::Weak<T> > {
    typedef JSC::Weak<T> StorageType;

    typedef std::nullptr_t EmptyValueType;
    static ALWAYS_INLINE EmptyValueType emptyValue() { return nullptr; }

    static constexpr bool hasIsEmptyValueFunction = true;
    static bool isEmptyValue(const JSC::Weak<T>& value) { return value.isHashTableEmptyValue(); }

    typedef JSC::PassWeak<T> PassInType;
    static ALWAYS_INLINE void store(PassInType value, StorageType& storage) { storage = value; }

    typedef JSC::PassWeak<T> PassOutType;
    static ALWAYS_INLINE PassOutType passOut(StorageType& value) { return value.release(); }
    static ALWAYS_INLINE PassOutType passOut(EmptyValueType) { return PassOutType(); }

    typedef typename StorageType::GetType PeekType;
    static ALWAYS_INLINE PeekType peek(const StorageType& value) { return value.get(); }
    static ALWAYS_INLINE PeekType peek(EmptyValueType) { return PeekType(); }
};

}

#endif // Weak_h
