/*
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "CallFrame.h"

#include "CallFrameInlines.h"
#include "CodeBlock.h"
#include "Interpreter.h"
#include "JSLexicalEnvironment.h"
#include "Operations.h"
#include "VMEntryScope.h"

namespace JSC {

#ifndef NDEBUG
JSStack* CallFrame::registerFile()
{
    return &interpreter()->registerFile();
}
#endif

unsigned CallFrame::bytecodeOffset()
{
    if (!codeBlock())
        return 0;
#if ENABLE(DFG_JIT)
    if (callSiteBitsAreCodeOriginIndex()) {
        ASSERT(codeBlock());
        CodeOrigin codeOrigin = this->codeOrigin();
        for (InlineCallFrame* inlineCallFrame = codeOrigin.inlineCallFrame; inlineCallFrame;) {
            codeOrigin = inlineCallFrame->directCaller;
            inlineCallFrame = codeOrigin.inlineCallFrame;
        }
        return codeOrigin.bytecodeIndex;
    }
#endif
    ASSERT(callSiteBitsAreBytecodeOffset());
    return callSiteBitsAsBytecodeOffset();
}

CodeOrigin CallFrame::codeOrigin()
{
    if (!codeBlock())
        return CodeOrigin(0);
#if ENABLE(DFG_JIT)
    if (callSiteBitsAreCodeOriginIndex()) {
        CallSiteIndex index = callSiteIndex();
        ASSERT(codeBlock()->canGetCodeOrigin(index));
        return codeBlock()->codeOrigin(index);
    }
#endif
    return CodeOrigin(callSiteBitsAsBytecodeOffset());
}

Register* CallFrame::frameExtentInternal()
{
    CodeBlock* codeBlock = this->codeBlock();
    ASSERT(codeBlock);
    return registers() + virtualRegisterForLocal(codeBlock->frameRegisterCount());
}

JSGlobalObject* CallFrame::vmEntryGlobalObject()
{
    if (this == lexicalGlobalObject()->globalExec())
        return lexicalGlobalObject();

    // For any ExecState that's not a globalExec, the
    // dynamic global object must be set since code is running
    ASSERT(vm().entryScope);
    return vm().entryScope->globalObject();
}

static inline bool exchange(bool& obj, bool new_value)
{
	bool old_value = obj;
    obj = new_value;
    return old_value;
}

SourceOrigin CallFrame::callerSourceOrigin()
{
    SourceOrigin sourceOrigin;
    bool haveSkippedFirstFrame = false;
    class Functor
    {
    public:
    	Functor(SourceOrigin& sourceOrigin, bool& haveSkippedFirstFrame) : sourceOrigin(sourceOrigin), haveSkippedFirstFrame(haveSkippedFirstFrame) { }
    	StackVisitor::Status operator()(StackVisitor& visitor) {
			if (!exchange(haveSkippedFirstFrame, true))
				return StackVisitor::Status::Continue;

			switch (visitor->codeType()) {
			case StackVisitor::Frame::CodeType::Function:
				// Skip the builtin functions since they should not pass the source origin to the dynamic code generation calls.
				// Consider the following code.
				//
				// [ "42 + 44" ].forEach(eval);
				//
				// In the above case, the eval function will be interpreted as the indirect call to eval inside forEach function.
				// At that time, the generated eval code should have the source origin to the original caller of the forEach function
				// instead of the source origin of the forEach function.
				if (static_cast<FunctionExecutable*>(visitor->codeBlock()->ownerScriptExecutable())->isBuiltinFunction())
					return StackVisitor::Status::Continue;
				FALLTHROUGH;

			case StackVisitor::Frame::CodeType::Eval:
			case StackVisitor::Frame::CodeType::Module:
			case StackVisitor::Frame::CodeType::Global:
				sourceOrigin = visitor->codeBlock()->ownerScriptExecutable()->sourceOrigin();
				return StackVisitor::Status::Done;

			case StackVisitor::Frame::CodeType::Native:
				return StackVisitor::Status::Continue;

			case StackVisitor::Frame::CodeType::Wasm:
				// FIXME: Should return the source origin for WASM.
				return StackVisitor::Status::Done;
			}

			RELEASE_ASSERT_NOT_REACHED();
			return StackVisitor::Status::Done;
		}
    private:
        SourceOrigin& sourceOrigin;
    	bool& haveSkippedFirstFrame;
    };
    Functor functor(sourceOrigin, haveSkippedFirstFrame);
    StackVisitor::visit(this, functor);
    return sourceOrigin;
}

}
