/*
 * Copyright (C) 2008, 2009, 2010, 2012 Apple Inc. All rights reserved.
 * Copyright (C) 2008 Cameron Zwarich <cwzwarich@uwaterloo.ca>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "Interpreter.h"

#include "BatchedTransitionOptimizer.h"
#include "CallFrameClosure.h"
#include "CallFrameInlines.h"
#include "CopiedSpaceInlineMethods.h"
#include "CodeBlock.h"
#include "Heap.h"
#include "Debugger.h"
#include "DebuggerCallFrame.h"
#include "DirectArguments.h"
#include "ErrorInstance.h"
#include "EvalCodeCache.h"
#include "Exception.h"
#include "ExceptionHelpers.h"
#include "GetterSetter.h"
#include "HostCallReturnValue.h"
#include "InterpreterInlines.h"
#include "JITCodeInlines.h"
#include "JSFixedArray.h"
#include "JSLexicalEnvironment.h"
#include "JSModuleEnvironment.h"
#include "JSArray.h"
#include "JSBoundFunction.h"
#include "JSStackInlines.h"
#include "JSString.h"
#include "JSWithScope.h"
#include "LiteralParser.h"
#include "ObjectPrototype.h"
#include "Operations.h"
#include "Parser.h"
#include "RegExpObject.h"
#include "RegExpPrototype.h"
#include "Register.h"
#include "SamplingTool.h"
#include "ScopedArguments.h"
#include "StackVisitor.h"
#include "SourceProvider.h"
#include "StrictEvalActivation.h"
#include "StrongInlines.h"
#include "StackFrame.h"
#include "Symbol.h"
#include "VMEntryScope.h"
#include <limits.h>
#include <wtf/StackStats.h>
#include <wtf/StringPrintStream.h>
#include <wtf/Threading.h>
#include <wtf/WTFThreadData.h>
#include <wtf/text/StringBuilder.h>

#if ENABLE(JIT)
#include "JIT.h"
#include "JITInlineMethods.h"
#endif

#define WTF_USE_GCC_COMPUTED_GOTO_WORKAROUND (ENABLE(LLINT) && !defined(__llvm__))

using namespace std;

namespace JSC {

Interpreter::ErrorHandlingMode::ErrorHandlingMode(ExecState *exec)
    : m_interpreter(*exec->vm().interpreter)
{
    if (!m_interpreter.m_errorHandlingModeReentry)
        m_interpreter.registerFile().enableErrorStackReserve();
    m_interpreter.m_errorHandlingModeReentry++;
}

Interpreter::ErrorHandlingMode::~ErrorHandlingMode()
{
    m_interpreter.m_errorHandlingModeReentry--;
    ASSERT(m_interpreter.m_errorHandlingModeReentry >= 0);
    if (!m_interpreter.m_errorHandlingModeReentry)
        m_interpreter.registerFile().disableErrorStackReserve();
}


// The Interpreter::StackPolicy class is used to compute a stack capacity
// requirement to ensure that we have enough room on the native stack for:
// 1. the max cumulative stack used by the interpreter and all code
//    paths sub of it up till leaf functions.
// 2. the max cumulative stack used by the interpreter before it reaches
//    the next checkpoint (execute...() function) in the interpreter.
//
// The interpreter can be run on different threads and hence, different
// native stacks (with different sizes) before exiting out of the first
// frame. Hence, the required capacity needs to be re-computed on every
// entry into the interpreter.
//
// Currently the requiredStack is computed based on a policy. See comments
// in StackPolicy::StackPolicy() for details.

Interpreter::StackPolicy::StackPolicy(Interpreter& interpreter, const StackBounds& stack)
    : m_interpreter(interpreter)
{
    const size_t size = stack.size();

    // We have two separate stack limits, one for regular JS execution, and one
    // for when we're handling errors. We need the error stack to be smaller
    // otherwise there would obviously not be any stack left to execute JS in when
    // there's a stack overflow.
    //
    // These sizes were derived from the stack usage of a number of sites when
    // layout occurs when we've already consumed most of the C stack.
    // const size_t requiredStack = 256 * KB;
    // const size_t errorModeRequiredStack = 64 * KB;

    size_t requiredCapacity = m_interpreter.m_errorHandlingModeReentry ? Options::reservedZoneSize() : Options::softReservedZoneSize();

    RELEASE_ASSERT(size > requiredCapacity);

    m_requiredCapacity = requiredCapacity;
}

JSValue eval(CallFrame* callFrame, ECMAMode ecmaMode)
{
    if (!callFrame->argumentCount())
        return jsUndefined();

    JSValue program = callFrame->argument(0);
    if (!program.isString())
        return program;
    
    JSGlobalData& vm = callFrame->vm();
    TopCallFrameSetter topCallFrame(vm, callFrame);
    JSGlobalObject* globalObject = callFrame->lexicalGlobalObject();
    if (!globalObject->evalEnabled()) {
        vm.throwException(callFrame, createEvalError(callFrame, globalObject->evalDisabledErrorMessage()));
        return jsUndefined();
    }
    String programSource = asString(program)->value(callFrame);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    
    CallFrame* callerFrame = callFrame->callerFrame();
    CallSiteIndex callerCallSiteIndex = callerFrame->callSiteIndex();
    CodeBlock* callerCodeBlock = callerFrame->codeBlock();
    JSScope* callerScopeChain = callerFrame->uncheckedR(callerCodeBlock->scopeRegister().offset()).Register::scope();

    bool isArrowFunctionContext = callerCodeBlock->isArrowFunction() || callerCodeBlock->isArrowFunctionContext();

    DerivedContextType derivedContextType = callerCodeBlock->derivedContextType();
    if (!isArrowFunctionContext && callerCodeBlock->isClassContext()) {
        derivedContextType = callerCodeBlock->isConstructor()
            ? DerivedContextType::DerivedConstructorContext
            : DerivedContextType::DerivedMethodContext;
    }

    EvalContextType evalContextType;
    if (callerCodeBlock->parseMode() == SourceParseMode::ClassFieldInitializerMode)
        evalContextType = EvalContextType::InstanceFieldEvalContext;
    else if (isFunctionParseMode(callerCodeBlock->parseMode()))
        evalContextType = EvalContextType::FunctionEvalContext;
    else if (callerCodeBlock->codeType() == EvalCode)
        evalContextType = callerCodeBlock->evalContextType();
    else
        evalContextType = EvalContextType::None;

    auto& evalCodeCache = callerCodeBlock->evalCodeCache();
    EvalExecutable* eval = evalCodeCache.tryGet(programSource, callerCallSiteIndex);
	if (!eval) {
        if (!ecmaMode.isStrict()) {
            if (programSource.is8Bit()) {
                LiteralParser<LChar> preparser(callFrame, programSource.characters8(), programSource.length(), NonStrictJSON);
                if (JSValue parsedObject = preparser.tryLiteralParse())
                    return parsedObject;
            } else {
                LiteralParser<UChar> preparser(callFrame, programSource.characters16(), programSource.length(), NonStrictJSON);
                if (JSValue parsedObject = preparser.tryLiteralParse())
                    return parsedObject;                
            }
            RETURN_IF_EXCEPTION(vm, jsUndefined());
        }

        TDZEnvironment variablesUnderTDZ;
        PrivateNameEnvironment privateNameEnvironment;
		JSScope::collectVariablesUnderTDZ(callerScopeChain, variablesUnderTDZ, privateNameEnvironment);
		eval = EvalExecutable::create(vm, makeSource(programSource, callerCodeBlock->source()->sourceOrigin()), ecmaMode.isStrict(), derivedContextType, callerCodeBlock->needsClassFieldInitializer(), callerCodeBlock->privateBrandRequirement(), isArrowFunctionContext, evalContextType, &variablesUnderTDZ, &privateNameEnvironment);
		if (!eval)
            return jsUndefined();

		evalCodeCache.set(vm, callerCodeBlock->ownerExecutable(), programSource, callerCallSiteIndex, eval);
    }

    JSValue thisValue = callerFrame->thisValue();
    Interpreter* interpreter = vm.interpreter;
    return interpreter->execute(eval, callFrame, thisValue, callerScopeChain);
}

static unsigned sizeOfVarargs(VM& vm, CallFrame* callFrame, JSValue arguments, uint32_t firstVarArgOffset)
{
    if (UNLIKELY(!arguments.isCell())) {
        if (arguments.isUndefinedOrNull())
            return 0;

        vm.throwException(callFrame, createInvalidFunctionApplyParameterError(callFrame, arguments));
        return 0;
    }

    JSCell* cell = arguments.asCell();
    unsigned length;
    switch(cell->structure()->typeInfo().type()) {
    case DirectArgumentsType:
        length = jsCast<DirectArguments*>(arguments)->length(callFrame);
        break;
    case ScopedArgumentsType:
        length = jsCast<ScopedArguments*>(arguments)->length(callFrame);
        break;
    case JSFixedArrayType:
        length = jsCast<JSFixedArray*>(cell)->size();
        break;
    case StringType:
    case SymbolType:
    case HeapBigIntType:
        vm.throwException(callFrame, createInvalidFunctionApplyParameterError(callFrame,  arguments));
        return 0;
    default:
        ASSERT(arguments.isObject());
        if(!cell->isObject())
        	return 0;
        length = clampToUnsigned(toLength(vm, callFrame, jsCast<JSObject*>(arguments)));
        break;
    }
    RETURN_IF_EXCEPTION(vm, 0);

    if (length >= firstVarArgOffset)
    	length -= firstVarArgOffset;
	else
		length = 0;

    return length;
}

CallFrame* sizeAndAllocFrameForVarargs(CallFrame* callFrame, JSStack* registerFile, JSValue arguments, int firstFreeRegister, uint32_t firstVarArgOffset, uint32_t& length)
{
    VM& vm = callFrame->vm();

    length = sizeOfVarargs(vm, callFrame, arguments, firstVarArgOffset);
    RETURN_IF_EXCEPTION(vm, 0);

    CallFrame* newCallFrame = CallFrame::create(callFrame->registers() + firstFreeRegister - CallFrame::offsetFor(length + 1));
    if (length > maxArguments || !registerFile->grow(newCallFrame->registers())) {
        vm.throwException(callFrame, createStackOverflowError(callFrame));
        return 0;
    }

    return newCallFrame;
}

void loadVarargs(CallFrame* callFrame, CallFrame* newCallFrame, JSValue thisValue, JSValue arguments, uint32_t firstVarArgOffset, uint32_t length)
{
    VM& vm = callFrame->vm();

    newCallFrame->setArgumentCountIncludingThis(length + 1);
    newCallFrame->setThisValue(thisValue);

    if (!arguments || !length)
    	return;

    JSCell* cell = arguments.asCell();

    switch(cell->structure()->typeInfo().type()) {
    case DirectArgumentsType:
        jsCast<DirectArguments*>(cell)->copyToArguments(callFrame, newCallFrame, length, firstVarArgOffset);
        return;
    case  ScopedArgumentsType:
        jsCast<ScopedArguments*>(cell)->copyToArguments(callFrame, newCallFrame, length, firstVarArgOffset);
        return;
    case JSFixedArrayType:
        jsCast<JSFixedArray*>(cell)->copyToArguments(callFrame, newCallFrame, length, firstVarArgOffset);
        return;
    default: {
        ASSERT(arguments.isObject());
        JSObject* object = jsCast<JSObject*>(cell);
    	if (isJSArray(object)) {
    		asArray(object)->copyToArguments(callFrame, newCallFrame, length, firstVarArgOffset);
    		return;
    	}
        unsigned i;
        for (i = 0; i < length && object->canGetIndexQuickly(i + firstVarArgOffset); ++i)
        	newCallFrame->setArgument(i, object->getIndexQuickly(i + firstVarArgOffset));
        for (; i < length; ++i) {
            JSValue value = object->get(vm, callFrame, i + firstVarArgOffset);
	        RETURN_IF_EXCEPTION(vm, void());
			newCallFrame->setArgument(i, value);
		}
		return;
    }
    }
}

static constexpr size_t defaultCapacity = 128 * 1024; // 1MB, each entry is 8 bytes

Interpreter::Interpreter(JSGlobalData& vm)
: m_sampleEntryDepth(0)
, m_vm(vm)
, m_errorHandlingModeReentry(0)
, m_registerFile(vm, defaultCapacity)
#if !ASSERT_DISABLED
, m_initialized(false)
#endif
{
}

Interpreter::~Interpreter()
{
}

void Interpreter::initialize(LLInt::Data* llintData, bool canUseJIT)
{
    UNUSED_PARAM(llintData);
    UNUSED_PARAM(canUseJIT);

#if ENABLE(COMPUTED_GOTO_OPCODES) && ENABLE(LLINT)
    m_opcodeTable = LLInt::opcodeMap();
    for (int i = 0; i < numOpcodeIDs; ++i)
        m_opcodeIDTable.add(m_opcodeTable[i], static_cast<OpcodeID>(i));
#endif

#if !ASSERT_DISABLED
    m_initialized = true;
#endif

#if ENABLE(OPCODE_SAMPLING)
    enableSampler();
#endif
}

#ifdef NDEBUG

void Interpreter::dumpCallFrame(CallFrame*)
{
}

#else

void Interpreter::dumpCallFrame(CallFrame* callFrame)
{
    callFrame->codeBlock()->dump(callFrame);
    dumpRegisters(callFrame);
}

class DumpRegisterFunctor {
public:
    DumpRegisterFunctor(const Register*& it)
        : m_hasSkippedFirstFrame(false)
        , m_it(it)
    {
    }

    StackVisitor::Status operator()(StackVisitor& visitor)
    {
        if (!m_hasSkippedFirstFrame) {
            m_hasSkippedFirstFrame = true;
            return StackVisitor::Continue;
        }

        unsigned line = 0;
        unsigned unusedColumn = 0;
        visitor->computeLineAndColumn(line, unusedColumn);
        dataLogF("[ReturnVPC]                | %10p | %d (line %d)\n", m_it, visitor->bytecodeOffset(), line);
        --m_it;
        return StackVisitor::Done;
    }

private:
    bool m_hasSkippedFirstFrame;
    const Register*& m_it;
};

void Interpreter::dumpRegisters(CallFrame* callFrame)
{
    dataLog("Register frame: \n\n");
    dataLog("-----------------------------------------------------------------------------\n");
    dataLog("            use            |   address  |                value               \n");
    dataLog("-----------------------------------------------------------------------------\n");

    CodeBlock* codeBlock = callFrame->codeBlock();
    const Register* it;
    const Register* end;

    it = callFrame->registers() + JSStack::ThisArgument + callFrame->argumentCount();
    end = callFrame->registers() + JSStack::ThisArgument - 1;
    while (it > end) {
        JSValue v = it->jsValue();
        int registerNumber = it - callFrame->registers();
        String name = codeBlock->nameForRegister(registerNumber);
        dataLogF("[r% 3d %14s]      | %10p | %-16s 0x%lld \n", registerNumber, name.ascii().data(), it, toCString(v).data(), (long long)JSValue::encode(v));
		--it;
    }
    
    dataLog("-----------------------------------------------------------------------------\n");
    dataLog("[argumentCountIncludingThis]            | %10p | %lu \n", it, (unsigned long) callFrame->argumentCount());
    --it;
    dataLog("[CallerFrame]              | %10p | %p \n", it, callFrame->callerFrame());
    --it;
    dataLog("[Callee]                   | %10p | %p \n", it, callFrame->callee());
    --it;
    // FIXME: Remove the next decrement when the ScopeChain slot is removed from the call header
    --it;
#if ENABLE(JIT)
    AbstractPC pc = callFrame->abstractReturnPC(callFrame->vm());
    if (pc.hasJITReturnAddress())
        dataLog("[ReturnJITPC]              | %10p | %p \n", it, pc.jitReturnAddress().value());
#endif

    DumpRegisterFunctor functor(it);
    callFrame->iterate(functor);

    dataLog("[CodeBlock]                | %10p | %p \n", it, callFrame->codeBlock());
    --it;
    dataLog("-----------------------------------------------------------------------------\n");

    end = it - codeBlock->m_numVars;
    if (it != end) {
        do {
            JSValue v = it->jsValue();
            int registerNumber = it - callFrame->registers();
            String name = codeBlock->nameForRegister(registerNumber);
            dataLogF("[r% 3d %14s]      | %10p | %-16s 0x%lld \n", registerNumber, name.ascii().data(), it, toCString(v).data(), (long long)JSValue::encode(v));
			--it;
        } while (it != end);
    }
    dataLog("-----------------------------------------------------------------------------\n");

    end = it - codeBlock->m_numCalleeLocals + codeBlock->m_numVars;
    if (it != end) {
        do {
            JSValue v = (*it).jsValue();
            int registerNumber = it - callFrame->registers();
            dataLogF("[r% 3d]                     | %10p | %-16s 0x%lld \n", registerNumber, it, toCString(v).data(), (long long)JSValue::encode(v));
            --it;
        } while (it != end);
    }
    dataLog("-----------------------------------------------------------------------------\n");
}

#endif

bool Interpreter::isOpcode(Opcode opcode)
{
#if ENABLE(COMPUTED_GOTO_OPCODES)
#if !ENABLE(LLINT)
    return opcode >= 0 && static_cast<OpcodeID>(bitwise_cast<uintptr_t>(opcode)) <= op_end;
#else
    return opcode != HashTraits<Opcode>::emptyValue()
        && !HashTraits<Opcode>::isDeletedValue(opcode)
        && m_opcodeIDTable.contains(opcode);
#endif
#else
    return opcode >= 0 && opcode <= op_end;
#endif
}

StackFrame::StackFrame()
    : m_codeType(GlobalCode)
	, m_bytecodeOffset(UINT_MAX)
	, m_isWasmFrame(false)
	, m_line(0)
	, m_column(0)
{ }

StackFrame::StackFrame(const VM& vm, JSCell* owner, JSObject* callee)
	: m_callee(const_cast<VM&>(vm), owner, callee)
	, m_codeType(GlobalCode)
	, m_bytecodeOffset(UINT_MAX)
	, m_isWasmFrame(false)
	, m_line(0)
	, m_column(0)
{
}

StackFrame::StackFrame(const VM& vm, JSCell* owner, JSObject* callee, CodeBlock* codeBlock, unsigned bytecodeOffset)
	: m_callee(const_cast<VM&>(vm), owner, callee)
	, m_executable(const_cast<VM&>(vm), owner, codeBlock->ownerExecutable())
	, m_codeType(codeBlock->codeType())
	, m_bytecodeOffset(bytecodeOffset)
	, m_isWasmFrame(false)
	, m_line(0)
	, m_column(0)
{
    int divot = 0;
    int unusedStartOffset = 0;
    int unusedEndOffset = 0;
    codeBlock->baselineVersion()->expressionRangeForBytecodeOffset(bytecodeOffset, divot, unusedStartOffset, unusedEndOffset, m_line, m_column);
}

StackFrame::~StackFrame()
{
}

String StackFrame::sourceURL() const
{
    if (m_isWasmFrame)
        return ASCIILiteral("[wasm code]");

    if (!m_executable) {
        return ASCIILiteral("[native code]");
    }

    const String& sourceURL = m_executable->sourceURL();
    if (!sourceURL.isNull())
        return sourceURL;
    return emptyString();
}

String StackFrame::functionName(VM& vm) const
{
    if (m_executable) {
        switch (m_codeType) {
        case EvalCode:
            return ASCIILiteral("eval code");
        case ModuleCode:
            return ASCIILiteral("module code");
        case FunctionCode:
            break;
        case GlobalCode:
            return ASCIILiteral("global code");
        default:
            ASSERT_NOT_REACHED();
        }
    }
    String name;
    if (m_callee)
        name = getCalculatedDisplayName(vm, m_callee.get());
    return name.isNull() ? emptyString() : name;
}

void StackFrame::computeLineAndColumn(unsigned& line, unsigned& column) const
{
	line = m_line;
	column = m_column;
}

String StackFrame::toString(VM& vm) const
{
    StringBuilder traceBuild;
    String functionName = this->functionName(vm);
    String sourceURL = this->sourceURL();
    traceBuild.append(functionName);
    if (!sourceURL.isEmpty()) {
        if (!functionName.isEmpty())
            traceBuild.append('@');
        traceBuild.append(sourceURL);
        if (m_executable) {
            unsigned line;
            unsigned column;
            computeLineAndColumn(line, column);

            traceBuild.append(':');
            traceBuild.appendNumber(line);
            traceBuild.append(':');
            traceBuild.appendNumber(column);
        }
    }
    return traceBuild.toString().impl();
}

void StackFrame::visitChildren(SlotVisitor& visitor)
{
    // FIXME: We should do something here about Wasm::IndexOrName.
    // https://bugs.webkit.org/show_bug.cgi?id=176644

    if (m_callee)
        visitor.append(m_callee);
    if (m_executable)
        visitor.append(m_executable);
}

class GetStackTraceFunctor {
public:
    GetStackTraceFunctor(VM& vm, JSCell* owner, Vector<StackFrame>& results, size_t framesToSkip, size_t capacity)
        : m_vm(vm)
		, m_owner(owner)
        , m_results(results)
        , m_framesToSkip(framesToSkip)
        , m_remainingCapacityForFrameCapture(capacity)
    {
        m_results.reserveInitialCapacity(capacity);
    }

    StackVisitor::Status operator()(StackVisitor& visitor)
    {
        if (m_framesToSkip > 0) {
            m_framesToSkip--;
            return StackVisitor::Continue;
        }

        if (m_remainingCapacityForFrameCapture) {
			if (visitor->isJSFrame() && !visitor->codeBlock()->isBuiltinFunction())
				m_results.constructAndAppend(m_vm, m_owner, visitor->callee(), visitor->codeBlock(), visitor->bytecodeOffset());
			else
				m_results.constructAndAppend(m_vm, m_owner, visitor->callee());

			m_remainingCapacityForFrameCapture--;
			return StackVisitor::Continue;
		}
		return StackVisitor::Done;
	}

private:
	VM& m_vm;
    JSCell* m_owner;
	Vector<StackFrame>& m_results;
    mutable size_t m_framesToSkip;
	size_t m_remainingCapacityForFrameCapture;
};

void Interpreter::getStackTrace(JSCell* owner, Vector<StackFrame>& results, size_t framesToSkip, size_t maxStackSize)
{
	if(WTF::isLowMemory())
		return;

    //DisallowGC disallowGC;
	VM& vm = m_vm;
	DeferGCForAWhile deferGC(vm.heap);
	ASSERT(!vm.topCallFrame->isVMEntrySentinel());
	CallFrame* callFrame = vm.topCallFrame;
	if (!callFrame)
		return;

    size_t framesCount = 0;
    struct FramesCountFunctor {
    	FramesCountFunctor(size_t& framesCount) : framesCount(framesCount) { }
        StackVisitor::Status operator()(StackVisitor& visitor)
        {
            framesCount++;
            return StackVisitor::Continue;
        }
        size_t& framesCount;
    };
    FramesCountFunctor framesCountFunctor(framesCount);
    callFrame->iterate(framesCountFunctor);
    if (framesCount <= framesToSkip)
        return;

    framesCount -= framesToSkip;
    framesCount = std::min(maxStackSize, framesCount);

    GetStackTraceFunctor functor(vm, owner, results, framesToSkip, framesCount);
    callFrame->iterate(functor);
    ASSERT(results.size() == results.capacity());
}

String Interpreter::stackTraceAsString(VM& vm, const Vector<StackFrame>& stackTrace)
{
    // FIXME: JSStringJoiner could be more efficient than StringBuilder here.
    StringBuilder builder;
    for (unsigned i = 0; i < stackTrace.size(); i++) {
        builder.append(String(stackTrace[i].toString(vm)));
        if (i != stackTrace.size() - 1)
            builder.append('\n');
    }
    return builder.toString();
}

ALWAYS_INLINE static HandlerInfo* findExceptionHandler(StackVisitor& visitor, CodeBlock* codeBlock, RequiredHandler requiredHandler)
{
    ASSERT(codeBlock);
#if ENABLE(DFG_JIT)
    ASSERT(!visitor->isInlinedFrame());
#endif

    CallFrame* callFrame = visitor->callFrame();
    unsigned exceptionHandlerIndex;
    if (JITCode::isOptimizingJIT(codeBlock->jitType()))
        exceptionHandlerIndex = callFrame->callSiteIndex().bits();
    else
        exceptionHandlerIndex = callFrame->bytecodeOffset();

    return codeBlock->handlerForIndex(exceptionHandlerIndex, requiredHandler);
}

namespace {
class GetCatchHandlerFunctor {
public:
	GetCatchHandlerFunctor() = default;

    HandlerInfo* handler() { return m_handler; }

    StackVisitor::Status operator()(StackVisitor& visitor)
    {
        visitor.unwindToMachineCodeBlockFrame();

        CodeBlock* codeBlock = visitor->codeBlock();
        if (!codeBlock)
            return StackVisitor::Continue;

        m_handler = findExceptionHandler(visitor, codeBlock, RequiredHandler::CatchHandler);
        if (m_handler)
            return StackVisitor::Done;

        return StackVisitor::Continue;
    }

private:
    HandlerInfo* m_handler { nullptr };
};

ALWAYS_INLINE static void notifyDebuggerOfUnwinding(VM& vm, CallFrame* callFrame)
{
    if (Debugger* debugger = callFrame->vmEntryGlobalObject()->debugger()) {
    	SuspendExceptionScope scope(&callFrame->vm());
        if (jsDynamicCast<JSFunction*>(callFrame->callee()))
            debugger->returnEvent(callFrame);
        else
            debugger->didExecuteProgram(callFrame);
        ASSERT(!callFrame->hadException());
    }
}

class UnwindFunctor final {
public:
    UnwindFunctor(VM& vm, CallFrame*& callFrame, bool isTermination, bool isOutOfMemory, CodeBlock*& codeBlock, HandlerInfo*& handler)
        : m_vm(vm)
		, m_callFrame(callFrame)
        , m_isTermination(isTermination)
		, m_isOutOfMemory(isOutOfMemory)
        , m_codeBlock(codeBlock)
        , m_handler(handler)
    {
    }

    StackVisitor::Status operator()(StackVisitor& visitor)
    {
        visitor.unwindToMachineCodeBlockFrame();
        m_callFrame = visitor->callFrame();
        m_codeBlock = visitor->codeBlock();

        m_handler = 0;
        if (!m_isTermination && !WTF::isLowMemory() && !m_isOutOfMemory) {
            if (m_codeBlock) {
                m_handler = findExceptionHandler(visitor, m_codeBlock, RequiredHandler::AnyHandler);
                if (m_handler)
                    return StackVisitor::Done;
            }
        }

        notifyDebuggerOfUnwinding(m_vm, m_callFrame);

		bool shouldStopUnwinding = m_callFrame->callerFrame()->isVMEntrySentinel();
		if (shouldStopUnwinding) {
			if (LegacyProfiler* profiler = m_vm.enabledProfiler())
				profiler->exceptionUnwind(m_callFrame);
			return StackVisitor::Done;
		}

		return StackVisitor::Continue;
	}

private:
    VM& m_vm;
	CallFrame*& m_callFrame;
	bool m_isTermination : 1;
	bool m_isOutOfMemory : 1;
	CodeBlock*& m_codeBlock;
	HandlerInfo*& m_handler;
};
}

NEVER_INLINE HandlerInfo* Interpreter::unwind(CallFrame*& callFrame, Exception* exception)
{
    if (callFrame->isVMEntrySentinel()) {
        // This happens when we throw stack overflow in a function that is called
        // directly from callToJavaScript.  Stack overflow throws the exception in the
        // context of the caller.  In that case the caller is the sentinel frame.  The
        // right thing to do is to pretend that the exception is uncaught so that we
        // go to the uncaught exception handler, which returns through callToJavaScript.
        return nullptr;
    }

    CodeBlock* codeBlock = callFrame->codeBlock();

    JSValue exceptionValue = exception->value();
    ASSERT(!exceptionValue.isEmpty());
    ASSERT(!exceptionValue.isCell() || exceptionValue.asCell());
    // This shouldn't be possible (hence the assertions), but we're already in the slowest of
    // slow cases, so let's harden against it anyway to be safe.
    if (exceptionValue.isEmpty() || (exceptionValue.isCell() && !exceptionValue.asCell()))
        exceptionValue = jsNull();

    //ASSERT_UNUSED(scope, scope.exception() && (!Options::exceptionStackTraceLimit() || scope.exception()->stack().size()));

    // Calculate an exception handler vPC, unwinding call frames as necessary.
    HandlerInfo* handler = nullptr;
    UnwindFunctor functor(codeBlock->vm(), callFrame, isTerminatedExecutionException(exception), isOutOfMemoryException(exception), codeBlock, handler);
    callFrame->iterate(functor);
    if (!handler)
        return nullptr;

    return handler;
}

void Interpreter::notifyDebuggerOfExceptionToBeThrown(CallFrame* callFrame, Exception* exception)
{
	Debugger* debugger = callFrame->vmEntryGlobalObject()->debugger();
	if (debugger && !exception->didNotifyInspectorOfThrow()) {
		// This code assumes that if the debugger is enabled then there is no inlining.
		// If that assumption turns out to be false then we'll ignore the inlined call
		// frames.
		// https://bugs.webkit.org/show_bug.cgi?id=121754

		bool isTermination = isTerminatedExecutionException(exception);

		bool hasCatchHandler;
		if (isTermination || WTF::isLowMemory() || isOutOfMemoryException(exception))
			hasCatchHandler = false;
		else {
			GetCatchHandlerFunctor functor;
			callFrame->iterate(functor);
			HandlerInfo* handler = functor.handler();
			ASSERT(!handler || handler->isCatchHandler());
			hasCatchHandler = !!handler;
		}

		debugger->exception(callFrame, exception->value(), hasCatchHandler);
	}
    exception->setDidNotifyInspectorOfThrow();
}

JSValue Interpreter::execute(ProgramExecutable* program, CallFrame* callFrame, JSObject* thisObj)
{
    SamplingScope samplingScope(this);

    JSScope* scope = thisObj->globalObject()->globalScope();
    JSGlobalData& vm = scope->vm();

    ASSERT(!vm->exception());
    ASSERT(!vm.isCollectorBusy());

    if (vm.isCollectorBusy())
        return jsUndefined();

    VMEntryScope entryScope(vm, scope->globalObject());

    StackStats::CheckPoint stackCheckPoint;
    const StackBounds& nativeStack = wtfThreadData().stack();
    StackPolicy policy(*this, nativeStack);
    if (!nativeStack.isSafeToRecurse(policy.requiredCapacity()))
        return throwStackOverflowError(callFrame);

    if(!canGrowHeap(vm, callFrame)) {
    	throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    // First check if the "program" is actually just a JSON object. If so,
    // we'll handle the JSON object here. Else, we'll handle real JS code
    // below at failedJSONP.
    Vector<JSONPData> JSONPData;
    bool parseResult;
    auto programSource = program->source().view();
    if (programSource.isNull())
        return jsUndefined();
    if (programSource.is8Bit()) {
        LiteralParser<LChar> literalParser(callFrame, programSource.characters8(), programSource.length(), JSONP);
        parseResult = literalParser.tryJSONPParse(JSONPData, scope->globalObject()->globalObjectMethodTable()->supportsRichSourceInfo(scope->globalObject()));
    } else {
        LiteralParser<UChar> literalParser(callFrame, programSource.characters16(), programSource.length(), JSONP);
        parseResult = literalParser.tryJSONPParse(JSONPData, scope->globalObject()->globalObjectMethodTable()->supportsRichSourceInfo(scope->globalObject()));
    }

    RETURN_IF_EXCEPTION(vm, jsUndefined());
    if (parseResult) {
        JSGlobalObject* globalObject = scope->globalObject();
        JSValue result;
        for (unsigned entry = 0; entry < JSONPData.size(); entry++) {
            Vector<JSONPPathEntry> JSONPPath;
            JSONPPath.swap(JSONPData[entry].m_path);
            JSValue JSONPValue = JSONPData[entry].m_value.get();
            if (JSONPPath.size() == 1 && JSONPPath[0].m_type == JSONPPathEntryTypeDeclareVar) {
            	if (UNLIKELY(!globalObject->isStructureExtensible()))
					goto failedJSONP;
				globalObject->createGlobalVarBinding<BindingCreationContext::Global>(callFrame, JSONPPath[0].m_pathEntryName);
				RETURN_IF_EXCEPTION(vm, jsUndefined());
				PutPropertySlot slot(globalObject);
				globalObject->methodTable()->put(globalObject, callFrame, JSONPPath[0].m_pathEntryName, JSONPValue, slot);
				RETURN_IF_EXCEPTION(vm, jsUndefined());
				result = jsUndefined();
				continue;
            }
            JSValue baseObject(globalObject);
            for (unsigned i = 0; i < JSONPPath.size() - 1; i++) {
                ASSERT(JSONPPath[i].m_type != JSONPPathEntryTypeDeclareVar);
                switch (JSONPPath[i].m_type) {
                case JSONPPathEntryTypeDot: {
                	if (i == 0) {
						RELEASE_ASSERT(baseObject == globalObject);

						class DoGet
						{
						public:
							DoGet(VM& vm, CallFrame* callFrame, Vector<JSONPPathEntry>& JSONPPath, unsigned i)
							: vm(vm)
							, callFrame(callFrame)
							, JSONPPath(JSONPPath)
							, i(i)
							{
							}
							JSValue operator()(JSSegmentedVariableObject* scope) {
								PropertySlot slot(scope, PropertySlot::InternalMethodType::Get);
								if (scope->getPropertySlot(vm, callFrame, JSONPPath[i].m_pathEntryName, slot))
									return slot.getValue(callFrame, JSONPPath[i].m_pathEntryName);
								return JSValue();
							};
						private:
							VM& vm;
							CallFrame* callFrame;
							Vector<JSONPPathEntry>& JSONPPath;
							const unsigned i;
						};

						DoGet doGet(vm, callFrame, JSONPPath, i);

						JSValue result = doGet(globalObject->globalLexicalEnvironment());
						RETURN_IF_EXCEPTION(vm, jsUndefined());
						if (result) {
							baseObject = result;
							continue;
						}

						result = doGet(globalObject);
						RETURN_IF_EXCEPTION(vm, jsUndefined());
						if (result) {
							baseObject = result;
							continue;
						}

						if (entry)
							return vm.throwException(callFrame, createUndefinedVariableError(callFrame, JSONPPath[i].m_pathEntryName));
						goto failedJSONP;
					}

					baseObject = baseObject.get(vm, callFrame, JSONPPath[i].m_pathEntryName);
					RETURN_IF_EXCEPTION(vm, jsUndefined());
                    continue;
                }
                case JSONPPathEntryTypeLookup: {
                    baseObject = baseObject.get(vm, callFrame, static_cast<unsigned>(JSONPPath[i].m_pathIndex));
    				RETURN_IF_EXCEPTION(vm, jsUndefined());
                    continue;
                }
                default:
                    ASSERT_NOT_REACHED();
                    return jsUndefined();
                }
            }

            if (JSONPPath.size() == 1 && JSONPPath.last().m_type != JSONPPathEntryTypeLookup) {
                RELEASE_ASSERT(baseObject == globalObject);
                JSGlobalLexicalEnvironment* scope = globalObject->globalLexicalEnvironment();
                if (scope->hasProperty(callFrame, JSONPPath.last().m_pathEntryName))
                    baseObject = scope;
                RETURN_IF_EXCEPTION(vm, jsUndefined());
            }

            PutPropertySlot slot(baseObject);
            switch (JSONPPath.last().m_type) {
            case JSONPPathEntryTypeCall: {
                JSValue function = baseObject.get(vm, callFrame, JSONPPath.last().m_pathEntryName);
				RETURN_IF_EXCEPTION(vm, jsUndefined());
                CallData callData = getCallData(vm, function);
                if (callData.type == CallData::Type::None)
                    return vm.throwException(callFrame, createNotAFunctionError(callFrame, function));
                MarkedArgumentBuffer jsonArg;
                jsonArg.append(JSONPValue);
                JSValue thisValue = JSONPPath.size() == 1 ? jsUndefined(): baseObject;
                JSONPValue = JSC::call(callFrame, function, callData, thisValue, jsonArg);
				RETURN_IF_EXCEPTION(vm, jsUndefined());
                break;
            }
            case JSONPPathEntryTypeDot: {
                baseObject.put(callFrame, JSONPPath.last().m_pathEntryName, JSONPValue, slot);
				RETURN_IF_EXCEPTION(vm, jsUndefined());
                break;
            }
            case JSONPPathEntryTypeLookup: {
                baseObject.putByIndex(callFrame, JSONPPath.last().m_pathIndex, JSONPValue, slot.isStrictMode());
				RETURN_IF_EXCEPTION(vm, jsUndefined());
                break;
            }
            default:
                ASSERT_NOT_REACHED();
				return jsUndefined();
            }
            result = JSONPValue;
        }
        return result;
    }
failedJSONP:
	// If we get here, then we have already proven that the script is not a JSON
	// object.

	// Compile source to bytecode if necessary:

	ProgramCodeBlock* codeBlock;
	{
		CodeBlock* tempCodeBlock;
		Exception* error = program->prepareForExecution<ProgramExecutable>(callFrame, 0, scope, CodeForCall, tempCodeBlock);
		if (UNLIKELY(error))
			return vm.throwException(callFrame, error);
		codeBlock = jsCast<ProgramCodeBlock*>(tempCodeBlock);
	}

#if ENABLE(JIT)
    if(!program->generatedJITCode())
        // Mixing Interpreter + JIT is not supported.
    	return throwOutOfMemoryError(callFrame);
#endif

    if (UNLIKELY(vm.watchdog && vm.watchdog->didFire(callFrame))) {
    	throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    if (scope->structure()->isUncacheableDictionary())
        scope->flattenDictionaryObject(vm);

    // Push the call frame for this invocation:
    ASSERT(codeBlock->numParameters() == 1); // 1 parameter for 'this'.
    CallFrame* newCallFrame = m_registerFile.pushFrame(callFrame, codeBlock, 1, JSCallee::create(vm, scope->globalObject(), scope));
    if (UNLIKELY(!newCallFrame))
        return checkedReturn(throwStackOverflowError(callFrame));

    // Set the arguments for the callee:
    newCallFrame->setThisValue(thisObj);

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->willExecute(callFrame, program->sourceURL(), program->lineNo());

    // Execute the code:
    JSValue result;
    {
        SamplingTool::CallRecord callRecord(m_sampler.get());
        Watchdog::Scope watchdogScope(vm.watchdog.get());

        result = program->generatedJITCode()->execute(&m_registerFile, newCallFrame, &vm);
    }

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->didExecute(callFrame, program->sourceURL(), program->lineNo());

    m_registerFile.popFrame(newCallFrame);

    return checkedReturn(result);
}

JSValue Interpreter::executeCall(VM& vm, CallFrame* callFrame, JSObject* function, const CallData& callData, JSValue thisValue, const ArgList& args)
{
    ASSERT(!callFrame->hadException());
    ASSERT(!vm.isCollectorBusy());
    if (vm.isCollectorBusy())
        return jsNull();

    StackStats::CheckPoint stackCheckPoint;
    const StackBounds& nativeStack = wtfThreadData().stack();
    StackPolicy policy(*this, nativeStack);
    if (!nativeStack.isSafeToRecurse(policy.requiredCapacity()))
        return throwStackOverflowError(callFrame);

    bool isJSCall = (callData.type == CallData::Type::JS);
    JSScope* scope = 0;
    CodeBlock* newCodeBlock;
    size_t argsCount = 1 + args.size(); // implicit "this" parameter

    JSGlobalObject* globalObject;

    if (isJSCall) {
        scope = callData.js.scope;
        globalObject = scope->globalObject();
    } else {
        ASSERT(callData.type == CallData::Type::Native);
        globalObject = function->globalObject();
    }

    VMEntryScope entryScope(vm, globalObject);

    if(!canGrowHeap(vm, callFrame)) {
    	throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    if (isJSCall) {
        // Compile the callee:
    	Exception* compileError = callData.js.functionExecutable->prepareForExecution<FunctionExecutable>(callFrame, jsCast<JSFunction*>(function), scope, CodeForCall, newCodeBlock);
        if (UNLIKELY(!!compileError))
            return checkedReturn(vm.throwException(callFrame, compileError));
        if(!newCodeBlock)
        	return checkedReturn(createOutOfMemoryError(callFrame));
        newCodeBlock->m_shouldAlwaysBeInlined = false;
    } else
        newCodeBlock = 0;

    if (UNLIKELY(vm.watchdog && vm.watchdog->didFire(callFrame))) {
    	throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    CallFrame* newCallFrame = m_registerFile.pushFrame(callFrame, newCodeBlock, argsCount, function);
    if (UNLIKELY(!newCallFrame))
        return checkedReturn(throwStackOverflowError(callFrame));

    // Set the arguments for the callee:
    newCallFrame->setThisValue(thisValue);
    for (size_t i = 0; i < args.size(); ++i)
        newCallFrame->setArgument(i, args.at(i));

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->willExecute(callFrame, function);

    JSValue result;
    {
        SamplingTool::CallRecord callRecord(m_sampler.get(), !isJSCall);
        Watchdog::Scope watchdogScope(vm.watchdog.get());

        // Execute the code:
        if (isJSCall) {
#if ENABLE(LLINT_C_LOOP)
            result = LLInt::CLoop::execute(newCallFrame, llint_function_for_call_prologue);
#elif ENABLE(JIT)
            result = callData.js.functionExecutable->generatedJITCodeForCall()->execute(&m_registerFile, newCallFrame, &vm);
#endif // ENABLE(JIT)
        } else
            result = JSValue::decode(callData.native.function(newCallFrame));
    }

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->didExecute(callFrame, function);

    m_registerFile.popFrame(newCallFrame);
    return checkedReturn(result);
}

JSObject* Interpreter::executeConstruct(VM& vm, CallFrame* callFrame, JSObject* constructor, const CallData& constructData, const ArgList& args, JSValue newTarget)
{
    ASSERT(!callFrame->hadException());
    ASSERT(!vm.isCollectorBusy());
    // We throw in this case because we have to return something "valid" but we're
    // already in an invalid state.
    if (UNLIKELY(vm.isCollectorBusy())) {
        throwStackOverflowError(callFrame);
        return 0;
    }

    StackStats::CheckPoint stackCheckPoint;
    const StackBounds& nativeStack = wtfThreadData().stack();
    StackPolicy policy(*this, nativeStack);
    if (UNLIKELY(!nativeStack.isSafeToRecurse(policy.requiredCapacity()))) {
        throwStackOverflowError(callFrame);
        return 0;
    }

    bool isJSConstruct = (constructData.type == CallData::Type::JS);
    JSScope* scope = 0;
    CodeBlock* newCodeBlock;
    size_t argsCount = 1 + args.size(); // implicit "this" parameter

    JSGlobalObject* globalObject;

    if (isJSConstruct) {
        scope = constructData.js.scope;
        globalObject = scope->globalObject();
    } else {
        ASSERT(constructData.type == CallData::Type::Native);
        globalObject = constructor->globalObject();
    }

    VMEntryScope entryScope(vm, globalObject);

    if(!canGrowHeap(vm, callFrame)) {
    	throwTerminatedExecutionException(callFrame);
    	return 0;
    }

    if (isJSConstruct) {
        // Compile the callee:
    	Exception* compileError = constructData.js.functionExecutable->prepareForExecution<FunctionExecutable>(callFrame, jsCast<JSFunction*>(constructor), scope, CodeForConstruct, newCodeBlock);
        if (UNLIKELY(!!compileError)) {
            vm.throwException(callFrame, compileError);
            return 0;
        }
        if(!newCodeBlock) {
        	vm.throwException(callFrame, createOutOfMemoryError(callFrame));
			return 0;
        }
        newCodeBlock->m_shouldAlwaysBeInlined = false;
    } else
        newCodeBlock = 0;

    if (UNLIKELY(vm.watchdog && vm.watchdog->didFire(callFrame))) {
    	throwTerminatedExecutionException(callFrame);
    	return 0;
    }

    CallFrame* newCallFrame = m_registerFile.pushFrame(callFrame, newCodeBlock, argsCount, constructor);
    if (UNLIKELY(!newCallFrame)) {
        throwStackOverflowError(callFrame);
    	return 0;
    }

    // Set the arguments for the callee:
    newCallFrame->setThisValue(newTarget);
    for (size_t i = 0; i < args.size(); ++i)
        newCallFrame->setArgument(i, args.at(i));

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->willExecute(callFrame, constructor);

    JSValue result;
    {
        SamplingTool::CallRecord callRecord(m_sampler.get(), !isJSConstruct);
        Watchdog::Scope watchdogScope(vm.watchdog.get());

        // Execute the code.
        if (isJSConstruct) {
#if ENABLE(LLINT_C_LOOP)
            result = LLInt::CLoop::execute(newCallFrame, llint_function_for_construct_prologue);
#elif ENABLE(JIT)
            result = constructData.js.functionExecutable->generatedJITCodeForConstruct()->execute(&m_registerFile, newCallFrame, &vm);
#endif // ENABLE(JIT)
        } else {
            result = JSValue::decode(constructData.native.function(newCallFrame));
            if (!vm.exception()) {
				ASSERT_WITH_MESSAGE(result.isObject(), "Host constructor returned non object.");
				if (!result.isObject()) {
					throwTypeError(newCallFrame, vm);
			    	return 0;
			    }
			}
        }
    }

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->didExecute(callFrame, constructor);

    m_registerFile.popFrame(newCallFrame);

    RETURN_IF_EXCEPTION(vm, 0);
    if(result.isObject())
    	return checkedReturn(asObject(result));
	throwTypeError(callFrame, vm);
	return 0;
}

CallFrameClosure Interpreter::prepareForRepeatCall(FunctionExecutable* functionExecutable, CallFrame* callFrame, JSFunction* function, int argumentCountIncludingThis, JSScope* scope)
{
    JSGlobalData& vm = scope->vm();
    ASSERT(!vm->exception());
    
    if (vm.isCollectorBusy())
        return CallFrameClosure();

    StackStats::CheckPoint stackCheckPoint;
    const StackBounds& nativeStack = wtfThreadData().stack();
    StackPolicy policy(*this, nativeStack);
    if (!nativeStack.isSafeToRecurse(policy.requiredCapacity())) {
        throwStackOverflowError(callFrame);
        return CallFrameClosure();
    }

    if(!canGrowHeap(vm, callFrame)){
    	throwTerminatedExecutionException(callFrame);
        return CallFrameClosure();
    }

    // Compile the callee:
    CodeBlock* newCodeBlock = 0;
    Exception* error = functionExecutable->prepareForExecution<FunctionExecutable>(callFrame, function, scope, CodeForCall, newCodeBlock);
    if (error) {
    	vm.throwException(callFrame, error);
        return CallFrameClosure();
    }
    if(!newCodeBlock){
        throwOutOfMemoryError(callFrame);
        return CallFrameClosure();
    }

    newCodeBlock->m_shouldAlwaysBeInlined = false;

    size_t argsCount = argumentCountIncludingThis;

    CallFrame* newCallFrame = m_registerFile.pushFrame(callFrame, newCodeBlock, argsCount, function);
    if (UNLIKELY(!newCallFrame)) {
        throwStackOverflowError(callFrame);
        return CallFrameClosure();
    }

    // Return the successful closure:
    CallFrameClosure result = { callFrame, newCallFrame, function, functionExecutable, &vm, scope, newCodeBlock->numParameters(), argumentCountIncludingThis };
    return result;
}

void Interpreter::endRepeatCall(CallFrameClosure& closure)
{
    m_registerFile.popFrame(closure.newCallFrame);
}

JSValue Interpreter::execute(EvalExecutable* eval, CallFrame* callFrame, JSValue thisValue, JSScope* scope)
{
    JSGlobalData& vm = scope->vm();
    auto throwScope = DECLARE_THROW_SCOPE(vm);
    SamplingScope samplingScope(this);

    ASSERT(scope->vm() == &callFrame->vm());
    ASSERT(!vm->exception());
    ASSERT(!vm.isCollectorBusy());
    if (vm.isCollectorBusy())
        return jsNull();

    VMEntryScope entryScope(vm, scope->globalObject());

    if(!canGrowHeap(vm, callFrame)) {
    	throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    StackStats::CheckPoint stackCheckPoint;
    const StackBounds& nativeStack = wtfThreadData().stack();
    StackPolicy policy(*this, nativeStack);
    if (!nativeStack.isSafeToRecurse(policy.requiredCapacity()))
        return throwStackOverflowError(callFrame);

    // Compile the callee:
    EvalCodeBlock* codeBlock;
    JSScope* newScope = scope;
    {
        CodeBlock* tempCodeBlock;

        double startTime = WTF::monotonicallyIncreasingTime();

        Exception* compileError = eval->prepareForExecution<EvalExecutable>(callFrame, nullptr, newScope, CodeForCall, tempCodeBlock);

        vm.evalPrepareForExecutionLength += (WTF::monotonicallyIncreasingTime() - startTime);

		if (UNLIKELY(!!compileError))
			return vm.throwException(callFrame, compileError);
		codeBlock = jsCast<EvalCodeBlock*>(tempCodeBlock);
	}

    JSScope* variableObject;
    if (newScope != scope) {
    	scope = newScope; // already created StrictEvalActivation
        variableObject = scope;
    } else {
		for (JSScope* node = scope; ; node = node->next()) {
			ASSERT(node);
            if (node->isGlobalObject()) {
				variableObject = node;
				break;
			}
            if (node->isJSLexicalEnvironment()) {
                JSLexicalEnvironment* lexicalEnvironment = jsCast<JSLexicalEnvironment*>(node);
                if (lexicalEnvironment->symbolTable()->scopeType() == SymbolTable::ScopeType::VarScope) {
					variableObject = node;
					break;
				}
            }
        }
    }

#if ENABLE(JIT)
    if (!eval->generatedJITCode())
        // Mixing Interpreter + JIT is not supported.
    	return throwOutOfMemoryError(callFrame);
#endif

    unsigned numVariables = eval->numVariables();
    unsigned numTopLevelFunctionDecls = eval->numTopLevelFunctionDecls();
    unsigned numFunctionHoistingCandidates = eval->numFunctionHoistingCandidates();

    if (variableObject->structure()->isUncacheableDictionary())
        variableObject->flattenDictionaryObject(vm);

    if (numVariables || numTopLevelFunctionDecls || numFunctionHoistingCandidates) {
        // Scope for BatchedTransitionOptimizer
        BatchedTransitionOptimizer optimizer(vm, variableObject);
        if (variableObject->next() && !eval->isInStrictContext())
        	variableObject->globalObject()->varInjectionWatchpoint()->fireAll(vm);

        if (!eval->isInStrictContext()) {
            for (unsigned i = 0; i < numVariables; ++i) {
                const Identifier& ident = codeBlock->variable(i);
                JSValue resolvedScope = JSScope::resolveScopeForHoistingFuncDeclInEval(callFrame, scope, ident);
                RETURN_IF_EXCEPTION(throwScope, throwScope.exception());
                if (resolvedScope.isUndefined())
                    return throwSyntaxError(callFrame, tryMakeString("Can't create duplicate variable in eval: '", StringView(ident.impl()), '\''));
            }

            for (unsigned i = 0; i < numTopLevelFunctionDecls; ++i) {
                FunctionExecutable* function = codeBlock->functionDecl(i);
                JSValue resolvedScope = JSScope::resolveScopeForHoistingFuncDeclInEval(callFrame, scope, function->name());
                RETURN_IF_EXCEPTION(throwScope, throwScope.exception());
                if (resolvedScope.isUndefined())
                    return throwSyntaxError(callFrame, tryMakeString("Can't create duplicate variable in eval: '", StringView(function->name().impl()), '\''));
            }
        }

        bool isGlobalVariableEnvironment = variableObject->isGlobalObject();
        if (isGlobalVariableEnvironment) {
            for (unsigned i = 0; i < numTopLevelFunctionDecls; ++i) {
                FunctionExecutable* function = codeBlock->functionDecl(i);
                bool canDeclare = jsCast<JSGlobalObject*>(variableObject)->canDeclareGlobalFunction(callFrame, function->name());
                throwScope.assertNoExceptionExceptTermination();
                if (!canDeclare)
                    return throwException(callFrame, throwScope, createErrorForInvalidGlobalFunctionDeclaration(callFrame, function->name()));
            }

            if (!variableObject->isStructureExtensible()) {
                for (unsigned i = 0; i < numVariables; ++i) {
                    const Identifier& ident = codeBlock->variable(i);
                    bool canDeclare = jsCast<JSGlobalObject*>(variableObject)->canDeclareGlobalVar(callFrame, ident);
                    throwScope.assertNoExceptionExceptTermination();
                    if (!canDeclare)
                        return throwException(callFrame, throwScope, createErrorForInvalidGlobalVarDeclaration(callFrame, ident));
                }
            }
        }

        auto ensureBindingExists = [&](const Identifier& ident) {
            bool hasProperty = variableObject->hasOwnProperty(callFrame, ident);
            throwScope.assertNoExceptionExceptTermination();
            if (!hasProperty) {
                bool shouldThrow = true;
                PutPropertySlot slot(variableObject, shouldThrow);
                variableObject->methodTable()->put(variableObject, callFrame, ident, jsUndefined(), slot);
                throwScope.assertNoExceptionExceptTermination();
            }
        };

        if (!eval->isInStrictContext()) {
			for (unsigned i = 0; i < numFunctionHoistingCandidates; ++i) {
				const Identifier& ident = codeBlock->functionHoistingCandidate(i);
				JSValue resolvedScope = JSScope::resolveScopeForHoistingFuncDeclInEval(callFrame, scope, ident);
				RETURN_IF_EXCEPTION(vm, checkedReturn(vm.exception()));
				if (!resolvedScope.isUndefined()) {
                    if (isGlobalVariableEnvironment) {
                        bool canDeclare = jsCast<JSGlobalObject*>(variableObject)->canDeclareGlobalVar(callFrame, ident);
                        throwScope.assertNoExceptionExceptTermination();
                        if (canDeclare) {
                            jsCast<JSGlobalObject*>(variableObject)->createGlobalVarBinding<BindingCreationContext::Eval>(callFrame, ident);
                            throwScope.assertNoExceptionExceptTermination();
                        }
                    } else
                        ensureBindingExists(ident);
                }
            }
        }

        for (unsigned i = 0; i < numTopLevelFunctionDecls; ++i) {
            FunctionExecutable* function = codeBlock->functionDecl(i);
            if (isGlobalVariableEnvironment) {
                jsCast<JSGlobalObject*>(variableObject)->createGlobalFunctionBinding<BindingCreationContext::Eval>(callFrame, function->name());
                throwScope.assertNoExceptionExceptTermination();
            } else
                ensureBindingExists(function->name());
        }

        for (unsigned i = 0; i < numVariables; ++i) {
            const Identifier& ident = codeBlock->variable(i);
            if (isGlobalVariableEnvironment) {
                jsCast<JSGlobalObject*>(variableObject)->createGlobalVarBinding<BindingCreationContext::Eval>(callFrame, ident);
                throwScope.assertNoExceptionExceptTermination();
            } else
                ensureBindingExists(ident);
        }
    }

    if (UNLIKELY(vm.watchdog && vm.watchdog->didFire(callFrame))) {
    	throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    // Push the frame:
    ASSERT(codeBlock->numParameters() == 1); // 1 parameter for 'this'.
    CallFrame* newCallFrame = m_registerFile.pushFrame(callFrame, codeBlock, 1, JSCallee::create(vm, scope->globalObject(), scope));
    if (UNLIKELY(!newCallFrame))
        return checkedReturn(throwStackOverflowError(callFrame));

    // Set the arguments for the callee:
    newCallFrame->setThisValue(thisValue);

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->willExecute(callFrame, eval->sourceURL(), eval->lineNo());

    // Execute the code:
    JSValue result;
    {
        SamplingTool::CallRecord callRecord(m_sampler.get());
        Watchdog::Scope watchdogScope(vm.watchdog.get());

        ++vm.evalCount;

#if ENABLE(LLINT_C_LOOP)
        result = LLInt::CLoop::execute(newCallFrame, llint_eval_prologue);
#elif ENABLE(JIT)
        result = eval->generatedJITCode()->execute(&m_registerFile, newCallFrame, &vm);
#endif // ENABLE(JIT)
    }

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->didExecute(callFrame, eval->sourceURL(), eval->lineNo());

    m_registerFile.popFrame(newCallFrame);
    return checkedReturn(result);
}

JSValue Interpreter::execute(ModuleProgramExecutable* executable, CallFrame* callFrame, JSModuleEnvironment* scope)
{
    VM& vm = scope->vm();
    SamplingScope samplingScope(this);

    ASSERT(scope->vm() == &callFrame->vm());
    ASSERT(!vm.exception());
    ASSERT(!vm.isCollectorBusy());
    RELEASE_ASSERT(vm.currentThreadIsHoldingAPILock());
    if (vm.isCollectorBusy())
        return jsNull();

    StackStats::CheckPoint stackCheckPoint;
    const StackBounds& nativeStack = wtfThreadData().stack();
    StackPolicy policy(*this, nativeStack);
    if (!nativeStack.isSafeToRecurse(policy.requiredCapacity()))
        return throwStackOverflowError(callFrame);

    VMEntryScope entryScope(vm, scope->globalObject());

    if(!canGrowHeap(vm, callFrame)) {
    	throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    ModuleProgramCodeBlock* codeBlock;
    {
		CodeBlock* tempCodeBlock;
		JSScope* tempScope = scope;
		Exception* compileError = executable->prepareForExecution<ModuleProgramExecutable>(callFrame, 0, tempScope, CodeForCall, tempCodeBlock);
		if (UNLIKELY(!!compileError))
			return checkedReturn(vm.throwException(callFrame, compileError));
		codeBlock = jsCast<ModuleProgramCodeBlock*>(tempCodeBlock);
    }

#if ENABLE(JIT)
    if(!executable->generatedJITCode())
        // Mixing Interpreter + JIT is not supported.
    	return throwOutOfMemoryError(callFrame);
#endif

    if (UNLIKELY(vm.watchdog && vm.watchdog->didFire(callFrame))) {
        throwTerminatedExecutionException(callFrame);
        return jsUndefined();
    }

    if (scope->structure()->isUncacheableDictionary())
        scope->flattenDictionaryObject(vm);

    // Push the frame:
    ASSERT(codeBlock->numParameters() == 1); // 1 parameter for 'this'.
    CallFrame* newCallFrame = m_registerFile.pushFrame(callFrame, codeBlock, 1, JSCallee::create(vm, scope->globalObject(), scope));
    if (UNLIKELY(!newCallFrame))
        return checkedReturn(throwStackOverflowError(callFrame));

    // The |this| of the module is always `undefined`.
    // http://www.ecma-international.org/ecma-262/6.0/#sec-module-environment-records-hasthisbinding
    // http://www.ecma-international.org/ecma-262/6.0/#sec-module-environment-records-getthisbinding
    newCallFrame->setThisValue(jsUndefined());

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->willExecute(callFrame, executable->sourceURL(), executable->lineNo());

    // Execute the code:
    JSValue result;
    {
        SamplingTool::CallRecord callRecord(m_sampler.get());
        result = executable->generatedJITCode()->execute(&m_registerFile, newCallFrame, &vm);
    }

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->didExecute(callFrame, executable->sourceURL(), executable->lineNo());

    return checkedReturn(result);
}

NEVER_INLINE void Interpreter::debug(CallFrame* callFrame, DebugHookID debugHookID)
{
    Debugger* debugger = callFrame->vmEntryGlobalObject()->debugger();
    if (!debugger)
        return;

    ASSERT(callFrame->codeBlock()->hasDebuggerRequests());
    ASSERT(!callFrame->hadException());

    switch (debugHookID) {
        case DidEnterCallFrame:
            debugger->callEvent(callFrame);
            break;
        case WillLeaveCallFrame:
            debugger->returnEvent(callFrame);
            break;
        case WillExecuteStatement:
            debugger->atStatement(callFrame);
            break;
        case WillExecuteProgram:
            debugger->willExecuteProgram(callFrame);
            break;
        case DidExecuteProgram:
            debugger->didExecuteProgram(callFrame);
            break;
        case DidReachBreakpoint:
            debugger->didReachBreakpoint(callFrame);
            break;
    }
    ASSERT(!callFrame->hadException());
}

void Interpreter::enableSampler()
{
#if ENABLE(OPCODE_SAMPLING)
    if (!m_sampler) {
        m_sampler = adoptPtr(new SamplingTool(this));
        m_sampler->setup();
    }
#endif
}
void Interpreter::dumpSampleData(ExecState* exec)
{
#if ENABLE(OPCODE_SAMPLING)
    if (m_sampler)
        m_sampler->dump(exec);
#else
    UNUSED_PARAM(exec);
#endif
}
void Interpreter::startSampling()
{
#if ENABLE(SAMPLING_THREAD)
    if (!m_sampleEntryDepth)
        SamplingThread::start();

    m_sampleEntryDepth++;
#endif
}
void Interpreter::stopSampling()
{
#if ENABLE(SAMPLING_THREAD)
    m_sampleEntryDepth--;
    if (!m_sampleEntryDepth)
        SamplingThread::stop();
#endif
}

} // namespace JSC
