/*
 * Copyright (C) 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2012 Research In Motion Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ArgList.h"
#include "CodeType.h"
#include "Heap.h"
#include "JSCell.h"
#include "JSValue.h"
#include "JSObject.h"
#include "LLIntData.h"
#include "Opcode.h"
#include "JSStack.h"
#include "SourceProvider.h"
#include <wtf/HashMap.h>
#include <wtf/UnalignedAccess.h>
#include <wtf/text/StringBuilder.h>

namespace JSC {

class CodeBlock;
class EvalExecutable;
class ExecutableBase;
class FunctionExecutable;
class JSFunction;
class JSGlobalData;
class JSGlobalObject;
class JSModuleEnvironment;
class JSModuleRecord;
class LLIntOffsetsExtractor;
class ProgramExecutable;
class ModuleProgramExecutable;
class Register;
class JSScope;
class SamplingTool;
class ScriptExecutable;
class SourceProvider;
class StackFrame;
class StackVisitor;
struct CallFrameClosure;
struct HandlerInfo;
struct Instruction;

typedef JSGlobalData VM;

enum DebugHookID {
	WillExecuteProgram,
	DidExecuteProgram,
	DidEnterCallFrame,
	DidReachBreakpoint,
	WillLeaveCallFrame,
	WillExecuteStatement
};

enum StackFrameCodeType {
	StackFrameGlobalCode,
	StackFrameEvalCode,
	StackFrameModuleCode,
	StackFrameFunctionCode,
	StackFrameNativeCode
};

class SuspendExceptionScope {
public:
	ALWAYS_INLINE SuspendExceptionScope(JSGlobalData* vm)
		: m_vm(vm)
	{
		oldException = vm->exception();
		vm->clearException();
	}
	ALWAYS_INLINE ~SuspendExceptionScope()
	{
		m_vm->restorePreviousException(oldException);
	}
private:
	Exception* oldException;
	JSGlobalData* m_vm;
};

class TopCallFrameSetter {
public:
	ALWAYS_INLINE TopCallFrameSetter(JSGlobalData& currentVM, CallFrame* callFrame)
		: globalData(currentVM)
		, oldCallFrame(currentVM.topCallFrame)
	{
		ASSERT(!callFrame->isVMEntrySentinel());
		currentVM.topCallFrame = callFrame;
	}

	ALWAYS_INLINE ~TopCallFrameSetter()
	{
		ASSERT(!oldCallFrame->isVMEntrySentinel());
		globalData.topCallFrame = oldCallFrame;
	}
private:
	JSGlobalData& globalData;
	CallFrame* oldCallFrame;
};

class NativeCallFrameTracer {
public:
	ALWAYS_INLINE NativeCallFrameTracer(JSGlobalData& vm, CallFrame* callFrame)
	{
		ASSERT(callFrame);
		ASSERT(!callFrame->isVMEntrySentinel());
		vm.topCallFrame = callFrame;
	}
	ALWAYS_INLINE NativeCallFrameTracer(JSGlobalData* vm, CallFrame* callFrame)
	{
		ASSERT(vm);
		ASSERT(callFrame);
		ASSERT(!callFrame->isVMEntrySentinel());
		vm->topCallFrame = callFrame;
	}

	enum VMEntrySentinelOKTag { VMEntrySentinelOK };
	ALWAYS_INLINE NativeCallFrameTracer(VM* vm, CallFrame* callFrame, VMEntrySentinelOKTag)
	{
		ASSERT(vm);
		ASSERT(callFrame);
		if (!callFrame->isVMEntrySentinel())
			vm->topCallFrame = callFrame;
	}
};

class Interpreter {
	WTF_MAKE_FAST_ALLOCATED;
	friend class CachedCall;
	friend class LLIntOffsetsExtractor;
	friend class JIT;
	friend class JSGlobalData;

public:
	class ErrorHandlingMode {
	public:
		JS_EXPORT_PRIVATE ErrorHandlingMode(ExecState*);
		JS_EXPORT_PRIVATE ~ErrorHandlingMode();
	private:
		Interpreter& m_interpreter;
	};

	Interpreter(JSGlobalData &);
	~Interpreter();

	void initialize(LLInt::Data*, bool canUseJIT);

	JSStack& registerFile() { return m_registerFile; }
	JSStack& stack() { return m_registerFile; }

	Opcode getOpcode(OpcodeID id)
	{
		ASSERT(m_initialized);
#if ENABLE(COMPUTED_GOTO_OPCODES)
		return m_opcodeTable[id];
#else
		return id;
#endif
	}

	static OpcodeID getOpcodeID(Opcode opcode)
	{
		ASSERT(m_initialized);
#if ENABLE(COMPUTED_GOTO_OPCODES) && ENABLE(LLINT)
#if USE(LLINT_EMBEDDED_OPCODE_ID)
    // The OpcodeID is embedded in the int32_t word preceding the location of
    // the LLInt code for the opcode (see the EMBED_OPCODE_ID_IF_NEEDED macro
    // in LowLevelInterpreter.cpp).
    auto codePtr = MacroAssemblerCodePtr<BytecodePtrTag>::createFromExecutableAddress(opcode);
    int32_t* opcodeIDAddress = codePtr.dataLocation<int32_t*>() - 1;
    OpcodeID opcodeID = static_cast<OpcodeID>(WTF::unalignedLoad<int32_t>(opcodeIDAddress));
    ASSERT(opcodeID < NUMBER_OF_BYTECODE_IDS);
    return opcodeID;
#else
		ASSERT(isOpcode(opcode));
		return m_opcodeIDTable.get(opcode);
#endif
#else
		return opcode;
#endif
	}

	static OpcodeID getOpcodeID(const Instruction&);
	//OpcodeID getOpcodeID(const UnlinkedInstruction&);

	bool isOpcode(Opcode);

	JSValue execute(ProgramExecutable*, CallFrame*, JSObject* thisObj);
	JSValue executeCall(JSGlobalData&, CallFrame*, JSObject* function, const CallData&, JSValue thisValue, const ArgList&);
	JSObject* executeConstruct(VM&, CallFrame*, JSObject* function, const CallData&, const ArgList&, JSValue newTarget);
	JSValue execute(EvalExecutable*, CallFrame*, JSValue thisValue, JSScope*);
	JSValue execute(ModuleProgramExecutable*, CallFrame*, JSModuleEnvironment*);

	void getArgumentsData(CallFrame*, JSFunction*&, ptrdiff_t& firstParameterIndex, Register*& argv, int& argc);

	SamplingTool* sampler() { return m_sampler.get(); }

	NEVER_INLINE HandlerInfo* unwind(CallFrame*&, Exception*);
    void notifyDebuggerOfExceptionToBeThrown(CallFrame*, Exception*);
	NEVER_INLINE void debug(CallFrame*, DebugHookID);
	static String stackTraceAsString(VM&, const Vector<StackFrame>&);

	void getStackTrace(JSCell* owner, Vector<StackFrame>& results, size_t framesToSkip = 0, size_t maxStackSize = std::numeric_limits<size_t>::max());

	void dumpSampleData(ExecState* exec);
	void startSampling();
	void stopSampling();

	JS_EXPORT_PRIVATE void dumpCallFrame(CallFrame*);

	class StackPolicy {
	public:
		StackPolicy(Interpreter&, const StackBounds&);
		inline size_t requiredCapacity() { return m_requiredCapacity; }

	private:
		Interpreter& m_interpreter;
		size_t m_requiredCapacity;
	};

private:

	enum ExecutionFlag { Normal, InitializeAndReturn };

    static JSValue checkedReturn(JSValue returnValue)
    {
        ASSERT(returnValue);
        return returnValue;
    }

    static JSObject* checkedReturn(JSObject* returnValue)
    {
        ASSERT(returnValue);
        return returnValue;
    }

	CallFrameClosure prepareForRepeatCall(FunctionExecutable*, CallFrame*, JSFunction*, int argumentCountIncludingThis, JSScope*);
	void endRepeatCall(CallFrameClosure&);
	JSValue execute(CallFrameClosure&);

	JSValue privateExecute(ExecutionFlag, JSStack*, CallFrame*);

	void dumpRegisters(CallFrame*);

	void enableSampler();
	int m_sampleEntryDepth;
	OwnPtr<SamplingTool> m_sampler;

	JSGlobalData& m_vm;
	JSStack m_registerFile;
	int m_errorHandlingModeReentry;

#if ENABLE(COMPUTED_GOTO_OPCODES) && ENABLE(LLINT)
	Opcode* m_opcodeTable; // Maps OpcodeID => Opcode for compiling
	HashMap<Opcode, OpcodeID> m_opcodeIDTable; // Maps Opcode => OpcodeID for decompiling
#endif

#if !ASSERT_DISABLED
	bool m_initialized;
#endif
};

JSValue eval(CallFrame*, ECMAMode);
static const unsigned maxArguments = 0x10000;
CallFrame* sizeAndAllocFrameForVarargs(CallFrame*, JSStack*, JSValue, int, uint32_t firstVarArgOffset, uint32_t& length);
void loadVarargs(CallFrame*, CallFrame*, JSValue, JSValue, uint32_t firstVarArgOffset, uint32_t length);

ALWAYS_INLINE JSValue call(VM& vm, ExecState* exec, JSValue functionObject, const CallData& callData, JSValue thisValue, const ArgList& args)
{
	ASSERT(callType == CallType::JS || callType == CallType::Host);
	return vm.interpreter->executeCall(vm, exec, asObject(functionObject), callData, thisValue, args);
}

} // namespace JSC
