/*
 * Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"

#include "CodeBlock.h"
#include "ExecutableAllocator.h"
#include "JSGlobalData.h"
#include "JSValueInlines.h"

#if ENABLE(EXECUTABLE_ALLOCATOR_DEMAND)
#include "CodeProfiling.h"
#include "JITCodeAllocationInfo.h"
#include <wtf/DoublyLinkedList.h>
#include <wtf/HashSet.h>
#include <wtf/MetaAllocator.h>
#include <wtf/NeverDestroyed.h>
#if ENABLE(ASSEMBLER_WX_EXCLUSIVE)
#include <wtf/PassOwnPtr.h>
#endif
#include <wtf/ThreadingPrimitives.h>
#include <wtf/VMTags.h>
#endif
#include <wtf/MemoryMonitor.h>
#include <mutex>

#if ENABLE(ASSEMBLER)

using namespace WTF;

namespace JSC {

#if ENABLE(EXECUTABLE_ALLOCATOR_DEMAND)

static const unsigned jitAllocationGranule = 32; // round up all allocations to cache line size

static const unsigned int DefaultJITCodeCap = (32 * MB);

static unsigned int sJITCodeCap = DefaultJITCodeCap;

static int sTotalReservedBytes;
static int sPeakReservedBytes;

static int targetRequestedNumberOfPages(int numPages)
{
	unsigned jitCodeMinimumBlockPages = Options::jitCodeMinimumBlockPages();
	if(numPages < jitCodeMinimumBlockPages) {
		size_t newNumPages = (numPages + jitCodeMinimumBlockPages - 1) / jitCodeMinimumBlockPages * jitCodeMinimumBlockPages;
		if(newNumPages > numPages)
			numPages = newNumPages;
	}

	return numPages;
}

static NEVER_INLINE SpinLock& getExecutableAllocatorLock()
{
    static LazyNeverDestroyed<SpinLock> lock;
    static std::once_flag onceFlag;
    std::call_once(
        onceFlag,
        [] {
            lock.construct();
        });

    return lock;
}

class DemandExecutableAllocator FINAL : public MetaAllocator {
	WTF_MAKE_FAST_ALLOCATED;

	friend class DoublyLinkedList<DemandExecutableAllocator>;

	struct Reservation
	{
		Reservation() = default;
		Reservation(void* p, size_t len)
		: base(p)
		, length(len)
		{
		}

		void* base { nullptr };
		size_t length { 0 };
	};

    void setPrev(DemandExecutableAllocator* prev) { m_prev = prev; }
    void setNext(DemandExecutableAllocator* next) { m_next = next; }

public:
    DemandExecutableAllocator* prev() const { return m_prev; }
    DemandExecutableAllocator* next() const { return m_next; }

    DemandExecutableAllocator(ExecutableAllocator::CodeType codeType, bool isSmallAllocator, JITCompilationEffort effort, DoublyLinkedList<DemandExecutableAllocator>& allocatorsList)
		: MetaAllocator(getExecutableAllocatorLock(), jitAllocationGranule)
		, m_codeType(codeType)
		, m_isSmallAllocator(isSmallAllocator)
		, m_compilationEffort(effort)
    	, m_allocatorsList(allocatorsList)
    {
    	allocatorsList.append(this);
    }

    DemandExecutableAllocator(void *start, int len, ExecutableAllocator::CodeType codeType, bool isSmallAllocator, JITCompilationEffort effort, DoublyLinkedList<DemandExecutableAllocator>& allocatorsList)
		: MetaAllocator(getExecutableAllocatorLock(), jitAllocationGranule)
		, m_fixedStart(start)
		, m_fixedLen(len)
		, m_codeType(codeType)
		, m_isSmallAllocator(isSmallAllocator)
		, m_compilationEffort(effort)
		, m_allocatorsList(allocatorsList)
    {
    	allocatorsList.append(this);
    }
    
    virtual ~DemandExecutableAllocator()
    {
		m_allocatorsList.remove(this);

    	sTotalReservedBytes -= bytesReserved();

    	if(!m_fixedStart)
    		fastAlignedFree(m_reservation.base, m_reservation.length, true);
    }

    ExecutableAllocator::CodeType codeType() const { return m_codeType; }
    bool isSmallAllocator() const { return m_isSmallAllocator; }
    size_t bytesReservedByAllAllocators()
    {
        return bytesReserved();
    }

    size_t bytesCommittedByAllocactors()
    {
        return bytesCommitted();
    }

    int fixedLen() const { return m_fixedLen; }

    ALWAYS_INLINE PassRefPtr<MetaAllocatorHandle> allocate(const SpinLockHolder& holder, size_t sizeInBytes, void* ownerUID)
	{
    	return MetaAllocator::allocate(holder, sizeInBytes, ownerUID);
	}

#if ENABLE(META_ALLOCATOR_PROFILE)
    static void dumpProfileFromAllAllocators()
    {
        MutexLocker lock(allocatorsMutex());
        for (HashSet<DemandExecutableAllocator*>::const_iterator allocator = allocators().begin(); allocator != allocators().end(); ++allocator)
            (*allocator)->dumpProfile();
    }
#endif

protected:
    virtual void* allocateNewSpace(size_t& numPages) FINAL
    {
    	if(m_reservation.base)
    		return nullptr;

        void* start;
        size_t len;
        size_t pageSize = WTF::pageSize();

    	if(m_fixedStart) {
    		start = m_fixedStart;
    		len = m_fixedLen;
    		numPages = m_fixedLen / pageSize;
    	} else {
    		if (m_isSmallAllocator)
    			numPages = targetRequestedNumberOfPages(numPages);

			len = numPages * pageSize;
			if (m_compilationEffort == JITCompilationCanFail) {
				if (!tryFastPageAlignedMalloc(pageSize, len, true).getValue(start))
					return nullptr;
			} else {
				start = fastPageAlignedMalloc(pageSize, len, true);
			}
    	}

		sTotalReservedBytes += len;

		if(sTotalReservedBytes > sPeakReservedBytes)
			sPeakReservedBytes = sTotalReservedBytes;

		m_reservation = Reservation(start, len);

        return start;
    }
    
    virtual void notifyNeedPage(void* page) FINAL
    {
    }
    
    virtual void notifyPageIsFree(void* page) FINAL
    {
    }

private:
    DoublyLinkedList<DemandExecutableAllocator>& m_allocatorsList;
    void* m_fixedStart { nullptr };
    int m_fixedLen { 0 };
    Reservation m_reservation;
    DemandExecutableAllocator* m_next { nullptr };
    DemandExecutableAllocator* m_prev { nullptr };
    ExecutableAllocator::CodeType m_codeType;
    bool m_isSmallAllocator;
    JITCompilationEffort m_compilationEffort;
};

#if ENABLE(ASSEMBLER_WX_EXCLUSIVE)
void ExecutableAllocator::initializeAllocator()
{
}
#else
static DoublyLinkedList<DemandExecutableAllocator> gFixedAllocators;
static DoublyLinkedList<DemandExecutableAllocator> gGenericSmallAllocators;
static DoublyLinkedList<DemandExecutableAllocator> gGenericLargeAllocators;
static DoublyLinkedList<DemandExecutableAllocator> gSpecialAllocators;

void ExecutableAllocator::initializeAllocator()
{
}
#endif

ExecutableAllocator::ExecutableAllocator(JSGlobalData&)
#if ENABLE(ASSEMBLER_WX_EXCLUSIVE)
    : m_allocator(adoptPtr(new  DemandExecutableAllocator()))
#endif
{
    ASSERT(allocator());
}

ExecutableAllocator::~ExecutableAllocator() = default;

bool ExecutableAllocator::isValid() const
{
    return true;
}

bool ExecutableAllocator::underMemoryPressure()
{
#if CPU(ARM64)
	int totalBytesAllocated = 0;

	for(auto* cur = gFixedAllocators.head(); cur; cur = cur->next())
		totalBytesAllocated += cur->bytesAllocated();

	// use half space
	return totalBytesAllocated >= (sJITCodeCap / 2);
#else
    return sTotalReservedBytes > sJITCodeCap;
#endif
}

double ExecutableAllocator::memoryPressureMultiplier(size_t addedMemoryUsage)
{
    UNUSED_PARAM(addedMemoryUsage);
    return 1.0f;
}

static void addJITCodeMemNoLock(void *ptr, int size);

PassRefPtr<ExecutableMemoryHandle> ExecutableAllocator::allocate(size_t sizeInBytes, void* ownerUID, JITCompilationEffort effort, ExecutableAllocator::CodeType codeType)
{
	if (WTF::isLowMemory()) {
	    if (effort == JITCompilationCanFail)
	    	return nullptr;
	}

	SpinLockHolder holder(getExecutableAllocatorLock());

#if CPU(ARM64)
	// must use fixed pool
	if (!gFixedAllocators.head()) {
		size_t fixedExecutableMemoryPoolSize = 128 * MB;
		if (void* ptr = mmap(NULL, fixedExecutableMemoryPoolSize, PROT_READ | PROT_WRITE | PROT_EXEC, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0))
			addJITCodeMemNoLock(ptr, fixedExecutableMemoryPoolSize);
	}
#endif

    for(auto* cur = gFixedAllocators.head(); cur; cur = cur->next()) {
		if(RefPtr<ExecutableMemoryHandle> result = cur->DemandExecutableAllocator::allocate(holder, sizeInBytes, ownerUID))
			return result.release();
	}

#if CPU(ARM64)
    if (effort == JITCompilationCanFail)
    	return nullptr;
    else {
    	WTF::logMessage(ASCIILiteral("JIT code: no space left."));
    	CRASH();
    	return nullptr;
    }
#endif

	if(codeType == ExecutableAllocator::CodeType::Special) {
		for(auto* cur = gSpecialAllocators.head(); cur; cur = cur->next()) {
			if(RefPtr<ExecutableMemoryHandle> result = cur->DemandExecutableAllocator::allocate(holder, sizeInBytes, ownerUID))
				return result.release();
		}

		DemandExecutableAllocator* newAllocator = new DemandExecutableAllocator(codeType, true, effort, gSpecialAllocators);
		return newAllocator->DemandExecutableAllocator::allocate(holder, sizeInBytes, ownerUID);
	}

	unsigned jitCodeMinimumBlockBytes = Options::jitCodeMinimumBlockPages() * pageSize();
	sizeInBytes = (sizeInBytes + jitAllocationGranule - 1) & ~(jitAllocationGranule - 1); // round up

	if(sizeInBytes <= (jitCodeMinimumBlockBytes / 2)) {
		for(auto* cur = gGenericSmallAllocators.head(); cur; cur = cur->next()) {
			if((cur->bytesReserved() - cur->bytesAllocated()) < sizeInBytes)
				continue;

			if(RefPtr<ExecutableMemoryHandle> result = cur->DemandExecutableAllocator::allocate(holder, sizeInBytes, ownerUID))
				return result.release();
		}

		DemandExecutableAllocator* newAllocator = new DemandExecutableAllocator(codeType, true, effort, gGenericSmallAllocators);
		return newAllocator->DemandExecutableAllocator::allocate(holder, sizeInBytes, ownerUID);
	} else {
		for(auto* cur = gGenericLargeAllocators.head(); cur; cur = cur->next()) {
			if((cur->bytesReserved() - cur->bytesAllocated()) < sizeInBytes)
				continue;

			if(RefPtr<ExecutableMemoryHandle> result = cur->DemandExecutableAllocator::allocate(holder, sizeInBytes, ownerUID))
				return result.release();
		}

		DemandExecutableAllocator* newAllocator = new DemandExecutableAllocator(codeType, false, effort, gGenericLargeAllocators);
		return newAllocator->DemandExecutableAllocator::allocate(holder, sizeInBytes, ownerUID);
	}
}

size_t ExecutableAllocator::committedByteCount()
{
	SpinLockHolder holder(getExecutableAllocatorLock());

	size_t committed = 0;

    for(auto* cur = gFixedAllocators.head(); cur; cur = cur->next())
		committed += cur->bytesCommitted();

    for(auto* cur = gGenericSmallAllocators.head(); cur; cur = cur->next())
		committed += cur->bytesCommitted();

    for(auto* cur = gGenericLargeAllocators.head(); cur; cur = cur->next())
		committed += cur->bytesCommitted();

    for(auto* cur = gSpecialAllocators.head(); cur; cur = cur->next())
		committed += cur->bytesCommitted();

    return committed;
}

DLL_PUBLIC void setJITCodeMemoryLimit(unsigned int cap)
{
	SpinLockHolder holder(getExecutableAllocatorLock());

	if(cap > 0)
		sJITCodeCap = cap;
	else
		sJITCodeCap = DefaultJITCodeCap;

	sPeakReservedBytes = 0;

	int fixedLen = 0;

	for(auto* cur = gFixedAllocators.head(); cur; cur = cur->next()) {
		sPeakReservedBytes += cur->bytesReserved();
		fixedLen += cur->fixedLen();
	}

	if(sJITCodeCap < fixedLen)
		sJITCodeCap = fixedLen;

	for(auto* cur = gGenericSmallAllocators.head(); cur; cur = cur->next())
		sPeakReservedBytes += cur->bytesReserved();

	for(auto* cur = gGenericLargeAllocators.head(); cur; cur = cur->next())
		sPeakReservedBytes += cur->bytesReserved();

	for(auto* cur = gSpecialAllocators.head(); cur; cur = cur->next())
		sPeakReservedBytes += cur->bytesReserved();
}

static void addJITCodeMemNoLock(void *ptr, int size)
{
	// Never destroyed
	new DemandExecutableAllocator(ptr, size, ExecutableAllocator::CodeType::None, false, JITCompilationCanFail, gFixedAllocators);

	int fixedLen = 0;

	for(auto* cur = gFixedAllocators.head(); cur; cur = cur->next())
		fixedLen += cur->fixedLen();

	if(sJITCodeCap < fixedLen)
		sJITCodeCap = fixedLen;
}

DLL_PUBLIC void addJITCodeMem(void *ptr, int size)
{
	SpinLockHolder holder(getExecutableAllocatorLock());
	addJITCodeMemNoLock(ptr, size);
}

DLL_PUBLIC int jitCodeCapBytes()
{
	return sJITCodeCap;
}

DLL_PUBLIC int jitCodeReservedBytes()
{
	return sTotalReservedBytes;
}

DLL_PUBLIC int jitCodePeakReservedBytes()
{
	return sPeakReservedBytes;
}

DLL_PUBLIC JITCodeAllocationInfo jitCodeAllocationInfo()
{
	JITCodeAllocationInfo info;
	memset(&info, 0, sizeof(info));

	SpinLockHolder holder(getExecutableAllocatorLock());

    for(DemandExecutableAllocator* cur = gFixedAllocators.head(); cur; cur = cur->next()) {
		info.allocated += cur->bytesAllocated();
		info.reserved  += cur->bytesReserved();
		info.committed += cur->bytesCommitted();
		++info.fixedBlocks;
	}

	for(DemandExecutableAllocator* cur = gGenericSmallAllocators.head(); cur; cur = cur->next()) {
		info.allocated += cur->bytesAllocated();
		info.reserved  += cur->bytesReserved();
		info.committed += cur->bytesCommitted();
		++info.smallBlocks;
	}

	for(DemandExecutableAllocator* cur = gGenericLargeAllocators.head(); cur; cur = cur->next()) {
		info.allocated += cur->bytesAllocated();
		info.reserved  += cur->bytesReserved();
		info.committed += cur->bytesCommitted();
		++info.largeBlocks;
	}

	for(DemandExecutableAllocator* cur = gSpecialAllocators.head(); cur; cur = cur->next()) {
		info.allocated += cur->bytesAllocated();
		info.reserved  += cur->bytesReserved();
		info.committed += cur->bytesCommitted();
		++info.specialBlocks;
	}

	return info;
}

#if ENABLE(META_ALLOCATOR_PROFILE)
void ExecutableAllocator::dumpProfile()
{
    DemandExecutableAllocator::dumpProfileFromAllAllocators();
}
#endif

#endif // ENABLE(EXECUTABLE_ALLOCATOR_DEMAND)

#if ENABLE(ASSEMBLER_WX_EXCLUSIVE)

#if OS(WINDOWS)
#error "ASSEMBLER_WX_EXCLUSIVE not yet suported on this platform."
#endif

void ExecutableAllocator::reprotectRegion(void* start, size_t size, ProtectionSetting setting)
{
    size_t pageSize = WTF::pageSize();

    // Calculate the start of the page containing this region,
    // and account for this extra memory within size.
    intptr_t startPtr = reinterpret_cast<intptr_t>(start);
    intptr_t pageStartPtr = startPtr & ~(pageSize - 1);
    void* pageStart = reinterpret_cast<void*>(pageStartPtr);
    size += (startPtr - pageStartPtr);

    // Round size up
    size += (pageSize - 1);
    size &= ~(pageSize - 1);

    mprotect(pageStart, size, (setting == Writable) ? PROTECTION_FLAGS_RW : PROTECTION_FLAGS_RX);
}

#endif

}

namespace WTF
{

MetaAllocatorHandle::~MetaAllocatorHandle()
{
    ASSERT(m_allocator);

	SpinLockHolder holder(JSC::getExecutableAllocatorLock());

    m_allocator->release(holder, this);

	auto* demandExecutableAllocator = static_cast<JSC::DemandExecutableAllocator*>(m_allocator);
	if(demandExecutableAllocator->bytesAllocated() == 0)
		delete demandExecutableAllocator;
}

}
#endif // HAVE(ASSEMBLER)
