/*
 * Copyright (C) 2008, 2009, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"

// #if ENABLE(JIT)
#include "JIT.h"

// This probably does not belong here; adding here for now as a quick Windows build fix.
//#if ENABLE(ASSEMBLER) && CPU(X86) && !OS(MAC_OS_X)
//#include "MacroAssembler.h"
//JSC::MacroAssemblerX86Common::SSE2CheckState JSC::MacroAssemblerX86Common::s_sse2CheckState = NotCheckedSSE2;
//#endif

#include "CodeBlock.h"
#include "DFGCapabilities.h"
#include "Interpreter.h"
#include "JITInlineMethods.h"
#include "JITOperations.h"
#include "JSArray.h"
#include "JSFunction.h"
#include "LinkBuffer.h"
#include "Operations.h"
#include "ResultType.h"
#include "SamplingTool.h"
#include "SlowPathCall.h"
#include <wtf/CryptographicallyRandomNumber.h>
#include <wtf/FastMallocMonitor.h>
#include <wtf/MemoryMonitor.h>


namespace JSC {

void ctiPatchCallByReturnAddress(ReturnAddressPtr returnAddress, FunctionPtr newCalleeFunction)
{
    MacroAssembler::repatchCall(
        CodeLocationCall(MacroAssemblerCodePtr(returnAddress)),
        newCalleeFunction);
}

JIT::JIT(JSGlobalData* globalData, CodeBlock* codeBlock)
    : JSInterfaceJIT(globalData, codeBlock)
	, m_interpreter(globalData->interpreter)
    , m_labels(codeBlock ? codeBlock->numberOfInstructions() : 0)
    , m_bytecodeOffset((unsigned)-1)
	, m_getByIdIndex(UINT_MAX)
	, m_getByIdWithThisIndex(UINT_MAX)
	, m_putByIdIndex(UINT_MAX)
	, m_instanceOfIndex(UINT_MAX)
	, m_byValInstructionIndex(UINT_MAX)
	, m_callLinkInfoIndex(UINT_MAX)
    , m_canBeOptimized(false)
	, m_canBeOptimizedOrInlined(false)
    , m_shouldEmitProfiling(false)
	, m_enableGetByIdIC(Options::maxGetByIdICSize())
	, m_enableByValIC(Options::maxByValICSize())
	, m_forceTryGetByIdICFailure(Options::forceTryGetByIdICFailure())
	, m_enablePutByIdIC(Options::maxPutByIdICSize())
	, m_enableInstanceOfIC(Options::enableInstanceOfIC())
{
	auto disableICs = [&] () {
		m_enableGetByIdIC = false;
		m_enableByValIC = false;
		m_enablePutByIdIC = false;
		m_enableInstanceOfIC = false;

		m_getByIds.shrinkCapacity(0);
		m_putByIds.shrinkCapacity(0);
	};

	if (UNLIKELY(WTF::isUnderMemoryPressure()))
		disableICs();

	if(codeBlock) {
	    if(m_codeBlock->getByIdCount() > Options::maxGetByIdICSize())
	    	m_enableGetByIdIC = false;

		if(m_enableGetByIdIC) {
			if (!m_getByIds.tryReserveInitialCapacity(codeBlock->getByIdCount()))
				disableICs();
		}

	    if(m_codeBlock->byValCount() > Options::maxByValICSize())
	    	m_enableByValIC = false;

	    // some programs have huge amount of PutById, compiling takes long time.
	    if(m_codeBlock->putByIdCount() > Options::maxPutByIdICSize())
	    	m_enablePutByIdIC = false;

	    if (m_enablePutByIdIC) {
			if (!m_putByIds.tryReserveInitialCapacity(codeBlock->putByIdCount()))
				disableICs();
	    }

	    if (!m_calls.tryReserveInitialCapacity(codeBlock->getByIdCount() + codeBlock->putByIdCount() + codeBlock->newObjectCount()))
			disableICs();

	    if (!m_exceptionChecks.tryReserveInitialCapacity(codeBlock->getByIdCount() + codeBlock->putByIdCount() + codeBlock->newObjectCount()))
			disableICs();
	}

#if ENABLE_JIT_PROFILE
	memset(&m_opCounts, 0x00, sizeof(m_opCounts));
	memset(&m_slowOpCounts, 0x00, sizeof(m_slowOpCounts));

	memset(&m_opCodeSizes, 0x00, sizeof(m_opCodeSizes));
	memset(&m_slowCodeSizes, 0x00, sizeof(m_slowCodeSizes));
#endif
}

JIT::~JIT() = default;

#if ENABLE(DFG_JIT)
void JIT::emitEnterOptimizationCheck()
{
    if (!canBeOptimized())
        return;

    JumpList skipOptimize;
    
    skipOptimize.append(branchAdd32(Signed, TrustedImm32(Options::executionCounterIncrementForEntry()), AbsoluteAddress(m_codeBlock->addressOfJITExecuteCounter())));
    ASSERT(!m_bytecodeOffset);
    callOperation(operationOptimize, m_bytecodeOffset);
    skipOptimize.append(branchTestPtr(Zero, returnValueGPR));
    jump(returnValueGPR);
    skipOptimize.link(this);
}
#endif

void JIT::emitNotifyWrite(WatchpointSet* set)
{
    if (!set || set->state() == IsInvalidated)
        return;

    addSlowCase(branch8(NotEqual, AbsoluteAddress(set->addressOfState()), TrustedImm32(IsInvalidated)));
}

void JIT::emitNotifyWrite(GPRReg pointerToSet)
{
    addSlowCase(branch8(NotEqual, Address(pointerToSet, WatchpointSet::offsetOfState()), TrustedImm32(IsInvalidated)));
}

void JIT::emitVarReadOnlyCheck(ResolveType resolveType)
{
    if (resolveType == GlobalVar || resolveType == GlobalVarWithVarInjectionChecks)
        addSlowCase(branch8(Equal, AbsoluteAddress(m_codeBlock->globalObject()->varReadOnlyWatchpoint()->addressOfState()), TrustedImm32(IsInvalidated)));
}

void JIT::assertStackPointerOffset()
{
#if 0
    if (ASSERT_DISABLED)
        return;

    addPtr(TrustedImm32(stackPointerOffsetFor(m_codeBlock) * sizeof(Register)), callFrameRegister, regT0);
    Jump ok = branchPtr(Equal, regT0, stackPointerRegister);
    breakpoint();
    ok.link(this);
#endif
}

#define NEXT_OPCODE(name) \
    m_bytecodeOffset += OPCODE_LENGTH(name); \
    break;

#if ENABLE_JIT_PROFILE
#define DEFINE_SLOW_OP(name) \
    case op_##name: { \
    	++m_opCounts[op_##name]; \
        unsigned startInstructionSize = assembler().buffer().codeSize(); \
        JITSlowPathCall slowPathCall(this, currentInstruction, slow_path_##name); \
        slowPathCall.call(); \
        m_opCodeSizes[op_##name] += (assembler().buffer().codeSize() - startInstructionSize); \
        NEXT_OPCODE(op_##name); \
    }

#define DEFINE_OP(name) \
    case name: { \
    	++m_opCounts[name]; \
        unsigned startInstructionSize = assembler().buffer().codeSize(); \
        emit_##name(currentInstruction); \
        m_opCodeSizes[name] += (assembler().buffer().codeSize() - startInstructionSize); \
        NEXT_OPCODE(name); \
    }

#define DEFINE_SLOWCASE_OP(name) \
    case name: { \
    	++m_slowOpCounts[name]; \
        unsigned startInstructionSize = assembler().buffer().codeSize(); \
        emitSlow_##name(currentInstruction, iter); \
        m_slowCodeSizes[name] += (assembler().buffer().codeSize() - startInstructionSize); \
        NEXT_OPCODE(name); \
    }
#else
#define DEFINE_SLOW_OP(name) \
    case op_##name: { \
        JITSlowPathCall slowPathCall(this, currentInstruction, slow_path_##name); \
        slowPathCall.call(); \
        NEXT_OPCODE(op_##name); \
    }

#define DEFINE_OP(name) \
    case name: { \
        emit_##name(currentInstruction); \
    	if (effort == JITCompilationCanFail && monitorMallocFailedPreventer.lowMemory()) \
    		return false; \
        NEXT_OPCODE(name); \
    }

#define DEFINE_SLOWCASE_OP(name) \
    case name: { \
        emitSlow_##name(currentInstruction, iter); \
        NEXT_OPCODE(name); \
    }
#endif

bool JIT::privateCompileMainPass(JITCompilationEffort effort)
{
	MonitorMallocFailedPreventer monitorMallocFailedPreventer;

    Instruction* instructionsBegin = m_codeBlock->instructions().begin();
    unsigned instructionCount = m_codeBlock->instructions().size();

    m_callLinkInfoIndex = 0;

    for (m_bytecodeOffset = 0; m_bytecodeOffset < instructionCount; ) {
        Instruction* currentInstruction = instructionsBegin + m_bytecodeOffset;
        ASSERT_WITH_MESSAGE(m_interpreter->isOpcode(currentInstruction->u.opcode), "privateCompileMainPass gone bad @ %d", m_bytecodeOffset);

#if ENABLE(OPCODE_SAMPLING)
        if (m_bytecodeOffset > 0) // Avoid the overhead of sampling op_enter twice.
            sampleInstruction(currentInstruction);
#endif

        m_labels[m_bytecodeOffset] = label();

#if ENABLE(JIT_VERBOSE)
        dataLog("Old JIT emitting code for bc#%u at offset 0x%lx.\n", m_bytecodeOffset, (long)debugOffset());
#endif

        if (Options::eagerlyUpdateTopCallFrame())
            updateTopCallFrame();

        switch (m_interpreter->getOpcodeID(currentInstruction->u.opcode)) {
        //DEFINE_SLOW_OP(assert)
        DEFINE_SLOW_OP(in)
		DEFINE_SLOW_OP(has_private_name)
		DEFINE_SLOW_OP(has_private_brand)
		DEFINE_SLOW_OP(less)
		DEFINE_SLOW_OP(lesseq)
		DEFINE_SLOW_OP(greater)
		DEFINE_SLOW_OP(greatereq)
        DEFINE_SLOW_OP(is_constructor)
		DEFINE_SLOW_OP(is_callable)
		DEFINE_SLOW_OP(typeof)
		DEFINE_SLOW_OP(typeof_is_object)
        DEFINE_SLOW_OP(typeof_is_function)
        DEFINE_SLOW_OP(new_array_buffer)
        DEFINE_SLOW_OP(strcat)
        DEFINE_SLOW_OP(push_with_scope)
        DEFINE_SLOW_OP(create_lexical_environment)
        DEFINE_SLOW_OP(get_by_val_with_this)
        DEFINE_SLOW_OP(put_by_id_with_this)
        DEFINE_SLOW_OP(put_by_val_with_this)
        //DEFINE_SLOW_OP(resolve_scope_for_hoisting_func_decl_in_eval)
        //DEFINE_SLOW_OP(define_data_property)
        //DEFINE_SLOW_OP(define_accessor_property)
        //DEFINE_SLOW_OP(unreachable)
        DEFINE_SLOW_OP(throw_static_error)
        //DEFINE_SLOW_OP(new_array_with_spread)
        //DEFINE_SLOW_OP(spread)
        DEFINE_SLOW_OP(get_enumerable_length)
        DEFINE_SLOW_OP(has_enumerable_property)
        DEFINE_SLOW_OP(get_property_enumerator)
        DEFINE_SLOW_OP(to_index_string)
        DEFINE_SLOW_OP(create_direct_arguments)
        DEFINE_SLOW_OP(create_scoped_arguments)
        DEFINE_SLOW_OP(create_cloned_arguments)
        //DEFINE_SLOW_OP(create_rest)
		DEFINE_SLOW_OP(copy_rest)
        DEFINE_SLOW_OP(pow)

        DEFINE_OP(op_add)
        DEFINE_OP(op_bitnot)
        DEFINE_OP(op_bitand)
        DEFINE_OP(op_bitor)
        DEFINE_OP(op_bitxor)
        DEFINE_OP(op_call)
        DEFINE_OP(op_call_eval)
        DEFINE_OP(op_call_varargs)
        DEFINE_OP(op_construct_varargs)
        DEFINE_OP(op_catch)
        DEFINE_OP(op_construct)
        DEFINE_OP(op_create_this)
        DEFINE_OP(op_to_this)
        DEFINE_OP(op_get_argument)
        DEFINE_OP(op_argument_count)
        DEFINE_OP(op_get_rest_length)
        DEFINE_OP(op_check_tdz)
        DEFINE_OP(op_identity_with_profile)
        DEFINE_OP(op_unreachable)
        DEFINE_OP(op_debug)
        DEFINE_OP(op_del_by_id)
        DEFINE_OP(op_del_by_val)
        DEFINE_OP(op_div)
        DEFINE_OP(op_end)
        DEFINE_OP(op_enter)
        DEFINE_OP(op_get_scope)
        DEFINE_OP(op_eq)
        DEFINE_OP(op_eq_null)
        DEFINE_OP(op_below)
        DEFINE_OP(op_beloweq)
        DEFINE_OP(op_try_get_by_id)
        case op_get_array_length:
        DEFINE_OP(op_get_by_id)
        DEFINE_OP(op_get_by_id_with_this)
        DEFINE_OP(op_get_by_id_direct)
        DEFINE_OP(op_get_by_val)
        DEFINE_OP(op_get_private_name)
        DEFINE_OP(op_set_private_brand)
        DEFINE_OP(op_check_private_brand)
        DEFINE_OP(op_get_prototype_of)
        DEFINE_OP(op_overrides_has_instance)
        DEFINE_OP(op_instanceof)
        DEFINE_OP(op_instanceof_custom)
        DEFINE_OP(op_is_empty)
        DEFINE_OP(op_typeof_is_undefined)
        DEFINE_OP(op_is_undefined_or_null)
        DEFINE_OP(op_is_boolean)
        DEFINE_OP(op_is_number)
        DEFINE_OP(op_is_big_int)
        DEFINE_OP(op_is_object)
        DEFINE_OP(op_is_cell_with_type)
        DEFINE_OP(op_jeq_null)
        DEFINE_OP(op_jfalse)
        DEFINE_OP(op_jmp)
        DEFINE_OP(op_jneq_null)
        DEFINE_OP(op_jundefined_or_null)
        DEFINE_OP(op_jnundefined_or_null)
        DEFINE_OP(op_jneq_ptr)
        DEFINE_OP(op_jless)
        DEFINE_OP(op_jlesseq)
        DEFINE_OP(op_jgreater)
        DEFINE_OP(op_jgreatereq)
        DEFINE_OP(op_jnless)
        DEFINE_OP(op_jnlesseq)
        DEFINE_OP(op_jngreater)
        DEFINE_OP(op_jngreatereq)
        DEFINE_OP(op_jeq)
        DEFINE_OP(op_jneq)
        DEFINE_OP(op_jstricteq)
        DEFINE_OP(op_jnstricteq)
        DEFINE_OP(op_jbelow)
        DEFINE_OP(op_jbeloweq)
        DEFINE_OP(op_jtrue)
        DEFINE_OP(op_loop_hint)
        DEFINE_OP(op_nop)
        DEFINE_OP(op_lshift)
        DEFINE_OP(op_mod)
        DEFINE_OP(op_mov)
        DEFINE_OP(op_mul)
        DEFINE_OP(op_negate)
        DEFINE_OP(op_neq)
        DEFINE_OP(op_neq_null)
        DEFINE_OP(op_new_array)
        DEFINE_OP(op_new_array_with_size)
        DEFINE_OP(op_new_array_with_spread)
        DEFINE_OP(op_spread)
        DEFINE_OP(op_new_func)
        DEFINE_OP(op_new_func_exp)
        DEFINE_OP(op_new_generator_func)
        DEFINE_OP(op_new_generator_func_exp)
        DEFINE_OP(op_new_async_func)
        DEFINE_OP(op_new_async_func_exp)
        DEFINE_OP(op_new_async_generator_func)
        DEFINE_OP(op_new_async_generator_func_exp)
        DEFINE_OP(op_new_object)
        DEFINE_OP(op_new_regexp)
        DEFINE_OP(op_not)
        DEFINE_OP(op_nstricteq)
        DEFINE_OP(op_dec)
        DEFINE_OP(op_inc)
        DEFINE_OP(op_profile_did_call)
        DEFINE_OP(op_profile_will_call)
        DEFINE_OP(op_get_parent_scope)
        DEFINE_OP(op_put_by_id)
        case op_put_by_val_direct:
        DEFINE_OP(op_put_by_val)
        DEFINE_OP(op_put_private_name)
        DEFINE_OP(op_put_getter_by_id)
        DEFINE_OP(op_put_setter_by_id)
        DEFINE_OP(op_put_getter_setter)
        DEFINE_OP(op_put_getter_by_val)
        DEFINE_OP(op_put_setter_by_val)
        DEFINE_OP(op_define_data_property)
        DEFINE_OP(op_define_accessor_property)
        DEFINE_OP(op_to_property_key)

        DEFINE_OP(op_ret)
        DEFINE_OP(op_rshift)
        DEFINE_OP(op_unsigned)
        DEFINE_OP(op_urshift)
        DEFINE_OP(op_set_function_name)
        DEFINE_OP(op_stricteq)
        DEFINE_OP(op_sub)
        DEFINE_OP(op_switch_char)
        DEFINE_OP(op_switch_imm)
        DEFINE_OP(op_switch_string)
        DEFINE_OP(op_throw)
        DEFINE_OP(op_to_number)
        DEFINE_OP(op_to_numeric)
        DEFINE_OP(op_to_string)
        DEFINE_OP(op_to_object)
        DEFINE_OP(op_to_primitive)

        DEFINE_OP(op_resolve_scope)
        DEFINE_OP(op_resolve_scope_for_hoisting_func_decl_in_eval)
        DEFINE_OP(op_get_from_scope)
        DEFINE_OP(op_put_to_scope)
        DEFINE_OP(op_get_from_arguments)
        DEFINE_OP(op_put_to_arguments)

        DEFINE_OP(op_has_enumerable_structure_property)
        DEFINE_OP(op_has_enumerable_indexed_property)
        DEFINE_OP(op_get_direct_pname)
        DEFINE_OP(op_enumerator_structure_pname)
        DEFINE_OP(op_enumerator_generic_pname)

        default:
            RELEASE_ASSERT_NOT_REACHED();
        }
    }

    ASSERT(m_callLinkInfoIndex == m_callCompilationInfo.size());

#ifndef NDEBUG
    // Reset this, in order to guard its use with ASSERTs.
    m_bytecodeOffset = (unsigned)-1;
#endif

    return true;
}

void JIT::privateCompileLinkPass()
{
    unsigned jmpTableCount = m_jmpTable.size();
    for (unsigned i = 0; i < jmpTableCount; ++i) {
    	JumpTable& jumpTable = m_jmpTable[i];
    	jumpTable.from.linkTo(m_labels[m_jmpTable[i].toBytecodeOffset], this);
    }
    m_jmpTable.clear();
}

void JIT::privateCompileSlowCases()
{
    Instruction* instructionsBegin = m_codeBlock->instructions().begin();

    m_getByIdIndex = 0;
    m_getByIdWithThisIndex = 0;
    m_putByIdIndex = 0;
    m_instanceOfIndex = 0;
    m_byValInstructionIndex = 0;
    m_callLinkInfoIndex = 0;
    
    // Use this to assert that slow-path code associates new profiling sites with existing
    // ValueProfiles rather than creating new ones. This ensures that for a given instruction
    // (say, get_by_id) we get combined statistics for both the fast-path executions of that
    // instructions and the slow-path executions. Furthermore, if the slow-path code created
    // new ValueProfiles then the ValueProfiles would no longer be sorted by bytecode offset,
    // which would break the invariant necessary to use CodeBlock::valueProfileForBytecodeOffset().
    unsigned numberOfValueProfiles = m_codeBlock->numberOfValueProfiles();

    for (Vector<SlowCaseEntry>::iterator iter = m_slowCases.begin(); iter != m_slowCases.end();) {
        m_bytecodeOffset = iter->to;
#ifndef NDEBUG
        unsigned firstTo = m_bytecodeOffset;
#endif
        Instruction* currentInstruction = instructionsBegin + m_bytecodeOffset;
        
#if ENABLE(JIT_VERBOSE)
        dataLog("Old JIT emitting slow code for bc#%u at offset 0x%lx.\n", m_bytecodeOffset, (long)debugOffset());
#endif

        switch (m_interpreter->getOpcodeID(currentInstruction->u.opcode)) {
        DEFINE_SLOWCASE_OP(op_add)
		DEFINE_SLOWCASE_OP(op_bitnot)
        DEFINE_SLOWCASE_OP(op_bitand)
        DEFINE_SLOWCASE_OP(op_bitor)
        DEFINE_SLOWCASE_OP(op_bitxor)
        DEFINE_SLOWCASE_OP(op_call)
        DEFINE_SLOWCASE_OP(op_call_eval)
        DEFINE_SLOWCASE_OP(op_call_varargs)
        DEFINE_SLOWCASE_OP(op_construct_varargs)
        DEFINE_SLOWCASE_OP(op_construct)
        DEFINE_SLOWCASE_OP(op_to_this)
        DEFINE_SLOWCASE_OP(op_check_tdz)
        DEFINE_SLOWCASE_OP(op_create_this)
        DEFINE_SLOWCASE_OP(op_div)
        DEFINE_SLOWCASE_OP(op_eq)
        DEFINE_SLOWCASE_OP(op_try_get_by_id)
        case op_get_array_length:
        DEFINE_SLOWCASE_OP(op_get_by_id)
        DEFINE_SLOWCASE_OP(op_get_by_id_with_this)
        DEFINE_SLOWCASE_OP(op_get_by_id_direct)
        DEFINE_SLOWCASE_OP(op_get_by_val)
        DEFINE_SLOWCASE_OP(op_get_private_name)
        DEFINE_SLOWCASE_OP(op_set_private_brand)
        DEFINE_SLOWCASE_OP(op_check_private_brand)
        DEFINE_SLOWCASE_OP(op_instanceof)
        DEFINE_SLOWCASE_OP(op_instanceof_custom)
        DEFINE_SLOWCASE_OP(op_jless)
        DEFINE_SLOWCASE_OP(op_jlesseq)
        DEFINE_SLOWCASE_OP(op_jgreater)
        DEFINE_SLOWCASE_OP(op_jgreatereq)
        DEFINE_SLOWCASE_OP(op_jnless)
        DEFINE_SLOWCASE_OP(op_jnlesseq)
        DEFINE_SLOWCASE_OP(op_jngreater)
        DEFINE_SLOWCASE_OP(op_jngreatereq)
        DEFINE_SLOWCASE_OP(op_jeq)
        DEFINE_SLOWCASE_OP(op_jneq)
        DEFINE_SLOWCASE_OP(op_jstricteq)
        DEFINE_SLOWCASE_OP(op_jnstricteq)
        DEFINE_SLOWCASE_OP(op_loop_hint)
        DEFINE_SLOWCASE_OP(op_lshift)
        DEFINE_SLOWCASE_OP(op_mod)
        DEFINE_SLOWCASE_OP(op_mul)
        DEFINE_SLOWCASE_OP(op_negate)
        DEFINE_SLOWCASE_OP(op_neq)
        DEFINE_SLOWCASE_OP(op_new_object)
        DEFINE_SLOWCASE_OP(op_not)
        DEFINE_SLOWCASE_OP(op_nstricteq)
        DEFINE_SLOWCASE_OP(op_dec)
        DEFINE_SLOWCASE_OP(op_inc)
        DEFINE_SLOWCASE_OP(op_put_by_id)
        case op_put_by_val_direct:
        DEFINE_SLOWCASE_OP(op_put_by_val)
        DEFINE_SLOWCASE_OP(op_put_private_name)
        DEFINE_SLOWCASE_OP(op_rshift)
        DEFINE_SLOWCASE_OP(op_unsigned)
        DEFINE_SLOWCASE_OP(op_urshift)
        DEFINE_SLOWCASE_OP(op_stricteq)
        DEFINE_SLOWCASE_OP(op_sub)
        DEFINE_SLOWCASE_OP(op_to_number)
        DEFINE_SLOWCASE_OP(op_to_numeric)
        DEFINE_SLOWCASE_OP(op_to_string)
        DEFINE_SLOWCASE_OP(op_to_object)
        DEFINE_SLOWCASE_OP(op_to_primitive)
        DEFINE_SLOWCASE_OP(op_has_enumerable_indexed_property)
        DEFINE_SLOWCASE_OP(op_has_enumerable_structure_property)
        DEFINE_SLOWCASE_OP(op_get_direct_pname)
        DEFINE_SLOWCASE_OP(op_get_prototype_of)

        DEFINE_SLOWCASE_OP(op_resolve_scope)
        DEFINE_SLOWCASE_OP(op_get_from_scope)
        DEFINE_SLOWCASE_OP(op_put_to_scope)
        DEFINE_SLOWCASE_OP(op_to_property_key)
        default:
            ASSERT_NOT_REACHED();
        }

        ASSERT_WITH_MESSAGE(iter == m_slowCases.end() || firstTo != iter->to,"Not enough jumps linked in slow case codegen.");
        ASSERT_WITH_MESSAGE(firstTo == (iter - 1)->to, "Too many jumps linked in slow case codegen.");
        
        emitJumpSlowToHot(jump(), 0);
    }

    ASSERT(m_getByIdIndex == m_getByIds.size());
    ASSERT(m_getByIdWithThisIndex == m_getByIdsWithThis.size());
    ASSERT(m_putByIdIndex == m_putByIds.size());
    RELEASE_ASSERT(m_instanceOfIndex == m_instanceOfs.size());
    ASSERT(m_callLinkInfoIndex == m_callCompilationInfo.size());
    ASSERT(numberOfValueProfiles == m_codeBlock->numberOfValueProfiles());

#ifndef NDEBUG
    // Reset this, in order to guard its use with ASSERTs.
    m_bytecodeOffset = (unsigned)-1;
#endif
}

CompilationResult JIT::privateCompile(JITCompilationEffort effort)
{
#if ENABLE(JIT_VERBOSE_OSR)
    printf("Compiling JIT code!\n");
#endif
    
    double compileStartTime = WTF::monotonicallyIncreasingTime();

    DFG::CapabilityLevel level = m_codeBlock->capabilityLevel();
    switch (level) {
    case DFG::CannotCompile:
        m_canBeOptimized = false;
        m_canBeOptimizedOrInlined = false;
        m_shouldEmitProfiling = false;
        m_codeBlock->m_shouldAlwaysBeInlined = false;
        break;
    case DFG::CanCompile:
    case DFG::CanCompileAndInline:
        m_canBeOptimized = true;
        m_canBeOptimizedOrInlined = true;
        m_shouldEmitProfiling = true;
        break;
    default:
        ASSERT_NOT_REACHED();
        break;
    }

    switch (m_codeBlock->codeType()) {
    case GlobalCode:
    case ModuleCode:
    case EvalCode:
        m_codeBlock->m_shouldAlwaysBeInlined = false;
        break;
    case FunctionCode:
        // We could have already set it to false because we detected an uninlineable call.
        // Don't override that observation.
        m_codeBlock->m_shouldAlwaysBeInlined &= canInline(level) && DFG::mightInlineFunction(m_codeBlock);
        break;
    }

    Label entryLabel(this);

    // Just add a little bit of randomness to the codegen
    if (random() & 1)
    	nop();

    preserveReturnAddressAfterCall(regT2);
    emitPutReturnPCToCallFrameHeader(regT2);
    emitPutImmediateToCallFrameHeader(m_codeBlock, JSStack::CodeBlock);

    Label beginLabel(this);

    sampleCodeBlock(m_codeBlock);
#if ENABLE(OPCODE_SAMPLING)
    sampleInstruction(m_codeBlock->instructions().begin());
#endif

    Jump registerFileCheck;
    if (m_codeBlock->codeType() == FunctionCode) {
#if ENABLE(DFG_JIT)
#if DFG_ENABLE(SUCCESS_STATS)
        static SamplingCounter counter("orignalJIT");
        emitCount(counter);
#endif
#endif

        ASSERT(m_bytecodeOffset == (unsigned)-1);
        if (shouldEmitProfiling()) {
            for (int argument = 0; argument < m_codeBlock->numParameters(); ++argument) {
                // If this is a constructor, then we want to put in a dummy profiling site (to
                // keep things consistent) but we don't actually want to record the dummy value.
                if (m_codeBlock->m_isConstructor && !argument)
                    continue;
                int offset = CallFrame::argumentOffsetIncludingThis(argument) * static_cast<int>(sizeof(Register));
#if USE(JSVALUE64)
                JSValueRegs resultRegs = JSValueRegs(regT0);
                load64(Address(callFrameRegister, offset), resultRegs.payloadGPR());
#elif USE(JSVALUE32_64)
                JSValueRegs resultRegs = JSValueRegs(regT1, regT0);
                load32(Address(callFrameRegister, offset + OBJECT_OFFSETOF(JSValue, u.asBits.payload)), resultRegs.payloadGPR());
                load32(Address(callFrameRegister, offset + OBJECT_OFFSETOF(JSValue, u.asBits.tag)), resultRegs.tagGPR());
#endif
                emitValueProfilingSite(m_codeBlock->valueProfileForArgument(argument), resultRegs);
            }
        }

		addPtr(TrustedImm32(virtualRegisterForLocal(frameRegisterCountFor(m_codeBlock)) * sizeof(Register)), callFrameRegister, regT1);
		registerFileCheck = branchPtr(Above, AbsoluteAddress(m_vm->interpreter->registerFile().addressOfEnd()), regT1);
	}

    Label functionBody = label();
    
    if (!privateCompileMainPass(effort))
        return CompilationFailed;
    privateCompileLinkPass();
    privateCompileSlowCases();

    Label arityCheck;
    if (m_codeBlock->codeType() == FunctionCode) {
    	registerFileCheck.link(this);
    	m_bytecodeOffset = 0;
    	callOperationWithCallFrameRollbackOnException(operationStackCheck, m_codeBlock);
#ifndef NDEBUG
    	m_bytecodeOffset = (unsigned)-1; // Reset this, in order to guard its use with ASSERTs.
#endif
		jump(functionBody);

	    // If the number of parameters is 1, we never require arity fixup.
	    bool requiresArityFixup = m_codeBlock->m_numParameters != 1;
		if(requiresArityFixup) {
			arityCheck = label();
			m_codeBlock->m_shouldAlwaysBeInlined = false;
			preserveReturnAddressAfterCall(regT2);
			emitPutReturnPCToCallFrameHeader(regT2);
			emitPutImmediateToCallFrameHeader(m_codeBlock, JSStack::CodeBlock);

			load32(payloadFor(JSStack::argumentCountIncludingThis), regT1);
			branch32(AboveOrEqual, regT1, TrustedImm32(m_codeBlock->m_numParameters)).linkTo(beginLabel, this);

			m_bytecodeOffset = 0;

			callOperationWithCallFrameRollbackOnException(m_codeBlock->m_isConstructor ? operationConstructArityCheck : operationCallArityCheck);
			branchTest32(Zero, returnValueGPR).linkTo(beginLabel, this);
	        move(returnValueGPR, GPRInfo::argumentGPR0);
			emitNakedCall(m_vm->getCTIStub(arityFixupGenerator).code());

#if !ASSERT_DISABLED
			m_bytecodeOffset = (unsigned)-1; // Reset this, in order to guard its use with ASSERTs.
#endif

			jump(beginLabel);
		} else
			arityCheck = entryLabel; // Never require arity fixup.
    }

    ASSERT(m_jmpTable.isEmpty());

    privateCompileExceptionHandlers();

    LinkBuffer patchBuffer(*this, m_codeBlock, effort, ExecutableAllocator::CodeType::Baseline);
    if (patchBuffer.didFailToAllocate())
        return CompilationFailed;

    // Translate vPC offsets into addresses in JIT generated code, for switch tables.
    for (unsigned i = 0; i < m_switches.size(); ++i) {
        SwitchRecord record = m_switches[i];
        unsigned bytecodeOffset = record.bytecodeOffset;

        if (record.type != SwitchRecord::String) {
            ASSERT(record.type == SwitchRecord::Immediate || record.type == SwitchRecord::Character); 
            ASSERT(record.jumpTable.simpleJumpTable->branchOffsets.size() == record.jumpTable.simpleJumpTable->ctiOffsets.size());

            record.jumpTable.simpleJumpTable->ctiDefault = patchBuffer.locationOf(m_labels[bytecodeOffset + record.defaultOffset]);

            for (unsigned j = 0; j < record.jumpTable.simpleJumpTable->branchOffsets.size(); ++j) {
                unsigned offset = record.jumpTable.simpleJumpTable->branchOffsets[j];
                record.jumpTable.simpleJumpTable->ctiOffsets[j] = offset ? patchBuffer.locationOf(m_labels[bytecodeOffset + offset]) : record.jumpTable.simpleJumpTable->ctiDefault;
            }
        } else {
            ASSERT(record.type == SwitchRecord::String);

            record.jumpTable.stringJumpTable->ctiDefault = patchBuffer.locationOf(m_labels[bytecodeOffset + record.defaultOffset]);

            auto end = record.jumpTable.stringJumpTable->offsetTable.end();
            for (auto it = record.jumpTable.stringJumpTable->offsetTable.begin(); it != end; ++it) {
                unsigned offset = it->second.branchOffset;
                it->second.ctiOffset = offset ? patchBuffer.locationOf(m_labels[bytecodeOffset + offset]) : record.jumpTable.stringJumpTable->ctiDefault;
            }
        }
    }

    for (size_t i = 0; i < m_codeBlock->numberOfExceptionHandlers(); ++i) {
        HandlerInfo& handler = m_codeBlock->exceptionHandler(i);
        handler.nativeCode = patchBuffer.locationOf(m_labels[handler.target]);
    }

    for (auto iter = m_calls.begin(); iter != m_calls.end(); ++iter) {
        if (iter->to)
            patchBuffer.link(iter->from, FunctionPtr(iter->to));
    }

    finalizeInlineCaches(m_getByIds, patchBuffer);
    finalizeInlineCaches(m_getByIdsWithThis, patchBuffer);
    finalizeInlineCaches(m_putByIds, patchBuffer);
    finalizeInlineCaches(m_instanceOfs, patchBuffer);

    if (m_byValCompilationInfo.size()) {
        CodeLocationLabel exceptionHandler = patchBuffer.locationOf(m_exceptionHandler);

		for(int i = 0; i < m_byValCompilationInfo.size(); ++i) {
			const ByValCompilationInfo& byValCompilationInfo = m_byValCompilationInfo[i];
			PatchableJump patchableNotIndexJump = byValCompilationInfo.notIndexJump;
			CodeLocationJump notIndexJump = CodeLocationJump();
			if (Jump(patchableNotIndexJump).isSet())
				notIndexJump = CodeLocationJump(patchBuffer.locationOf(patchableNotIndexJump));

            PatchableJump patchableBadTypeJump = byValCompilationInfo.badTypeJump;
            auto badTypeJump = CodeLocationJump();
            if (Jump(patchableBadTypeJump).isSet())
                badTypeJump = CodeLocationJump(patchBuffer.locationOf(byValCompilationInfo.badTypeJump));

			CodeLocationLabel doneTarget = patchBuffer.locationOf(byValCompilationInfo.doneTarget);
			CodeLocationLabel nextHotPathTarget = patchBuffer.locationOf(byValCompilationInfo.nextHotPathTarget);
			CodeLocationLabel slowPathTarget = patchBuffer.locationOf(byValCompilationInfo.slowPathTarget);
			CodeLocationCall returnAddress = patchBuffer.locationOf(byValCompilationInfo.returnAddress);

			*byValCompilationInfo.byValInfo = ByValInfo(
				byValCompilationInfo.bytecodeIndex,
				notIndexJump,
				badTypeJump,
				exceptionHandler,
				byValCompilationInfo.arrayMode,
				byValCompilationInfo.arrayProfile,
				doneTarget,
				nextHotPathTarget,
				slowPathTarget);
		}
    }

    for (unsigned i = 0; i < m_callCompilationInfo.size(); ++i) {
        CallCompilationInfo& compilationInfo = m_callCompilationInfo[i];
        CallLinkInfo& info = *compilationInfo.callLinkInfo;
        info.setCallLocations(patchBuffer.locationOfNearCall(compilationInfo.callReturnLocation),
        	patchBuffer.locationOf(compilationInfo.hotPathBegin),
			patchBuffer.locationOfNearCall(compilationInfo.hotPathOther));
    }

#if ENABLE(DFG_JIT) || ENABLE(LLINT)
    if (canBeOptimizedOrInlined()
#if ENABLE(LLINT)
        || true
#endif
        ) {
        JITCodeMap jitCodeMap;
        for (unsigned bytecodeOffset = 0; bytecodeOffset < m_labels.size(); ++bytecodeOffset) {
            if (m_labels[bytecodeOffset].isSet())
            	jitCodeMap.append(bytecodeOffset, patchBuffer.locationOf(m_labels[bytecodeOffset]));
        }
        jitCodeMap.finish();
        m_codeBlock->setJITCodeMap(WTFMove(jitCodeMap));
    }
#endif

    MacroAssemblerCodePtr withArityCheck;
    if (m_codeBlock->codeType() == FunctionCode)
        withArityCheck = patchBuffer.locationOf(arityCheck);

    CodeRef result = patchBuffer.finalizeCodeWithoutDisassembly();
    
    m_vm->machineCodeBytesPerBytecodeWordForBaselineJIT.add(
        static_cast<double>(result.size()) /
        static_cast<double>(m_codeBlock->instructions().size()));
    
    if(Options::enableDebugDFGJIT() && codeBlock()){
		unsigned jitCodeMinimumBlockBytes = Options::jitCodeMinimumBlockPages() * pageSize();
		if (codeBlock()->getByIdCount() > 100 || m_getByIdsWithThis.size() > 100 || codeBlock()->putByIdCount() > 100 || result.size() > jitCodeMinimumBlockBytes / 2
			|| codeBlock()->byValCount() > 100 || m_instanceOfs.size() > 100 || m_callCompilationInfo.size() > 100) {
			fprintf(stderr, "JIT instructions %i, code size %i, get-by-id %i, get-by-id-with-this %i, put-by-id %i, by-val %i, instance-of %i, call %i\n",
				static_cast<int>(m_codeBlock->instructions().size()), static_cast<int>(result.size()),
				codeBlock()->getByIdCount(), static_cast<int>(m_getByIdsWithThis.size()), codeBlock()->putByIdCount(),
				codeBlock()->byValCount(), static_cast<int>(m_instanceOfs.size()), static_cast<int>(m_callCompilationInfo.size()));
		}
    }

    m_codeBlock->shrinkToFit(CodeBlock::LateShrink);
    m_codeBlock->setJITCode(
        adoptRef(new DirectJITCode(result, withArityCheck, JITCode::BaselineJIT)));
    
#if ENABLE(JIT_VERBOSE)
    dataLog("JIT generated code for %p at [%p, %p).\n", m_codeBlock, result.executableMemory()->start(), result.executableMemory()->end());
#endif

    double compileEndTime = WTF::monotonicallyIncreasingTime();
    m_vm->jitCompileLength += (compileEndTime - compileStartTime);

#if ENABLE_JIT_PROFILE
    if(m_codeBlock) {
    	unsigned ms = (unsigned)((compileEndTime - compileStartTime) * 1000);
    	if(!m_enablePutByIdIC || ms >= 1) {
    		int totalOpCodeSize = 0;

    		WTF::logMessage(String("BaseLine JIT ") + String(codeTypeToString(m_codeBlock->codeType())) + String(" compiling takes ") + String::number(ms) + String("ms"));
			WTF::logMessage(String("  Instructions ") + String::number(m_codeBlock->instructionCount()));
			for(int i = 0; i < numOpcodeIDs; ++i) {
    			if(m_opCounts[i]) {
    				WTF::logMessage(String("       ") + String(opcodeNames[i]) + String(" ") + String::number(m_opCounts[i]) +
    					String(" code size ") + String::number(m_opCodeSizes[i]));

    				totalOpCodeSize += m_opCodeSizes[i];
    			}
    			if(m_slowOpCounts[i]) {
    				WTF::logMessage(String("  Slow ") + String(opcodeNames[i]) + String(" ") + String::number(m_slowOpCounts[i]) +
						String(" code size ") + String::number(m_slowCodeSizes[i]));

    				totalOpCodeSize += m_slowCodeSizes[i];
    			}
    		}

			WTF::logMessage(String(" numCalleeLocals ") + String::number(m_codeBlock->m_numCalleeLocals) + String(" numParameters ") + String::number(m_codeBlock->m_numParameters));
    	}
    }
#endif

    return CompilationSuccessful;
}

void JIT::privateCompileExceptionHandlers()
{
    if (m_exceptionChecks.empty() && m_exceptionChecksWithCallFrameRollback.empty())
        return;

    Jump doLookup;

    if (!m_exceptionChecksWithCallFrameRollback.empty()) {
        m_exceptionChecksWithCallFrameRollback.link(this);
        emitGetCallerFrameFromCallFrameHeaderPtr(GPRInfo::argumentGPR1);
        doLookup = jump();
    }

    if (!m_exceptionChecks.empty() || m_byValCompilationInfo.size()) {
        m_exceptionHandler = label();
    	m_exceptionChecks.link(this);
    }

    // lookupExceptionHandler is passed two arguments, the VM and the exec (the CallFrame*).
    move(GPRInfo::callFrameRegister, GPRInfo::argumentGPR1);

    if (doLookup.isSet())
        doLookup.link(this);

    move(TrustedImmPtr(vm()), GPRInfo::argumentGPR0);

#if CPU(X86)
    // FIXME: should use the call abstraction, but this is currently in the SpeculativeJIT layer!
    poke(GPRInfo::argumentGPR0);
    poke(GPRInfo::argumentGPR1, 1);
#endif
    m_calls.append(CallRecord(call(), (unsigned)-1, FunctionPtr(lookupExceptionHandler).value()));
    jumpToExceptionHandler();
}

} // namespace JSC

// #endif // ENABLE(JIT)
