/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "JITOperations.h"

#include "ArithProfile.h"
#include "ArrayConstructor.h"
#include "CallFrameInlines.h"
#include "CommonSlowPaths.h"
#include "DFGDriver.h"
#include "DFGOSREntry.h"
#include "DirectArguments.h"
#include "Error.h"
#include "GetterSetter.h"
#include "HostCallReturnValue.h"
#include "JIT.h"
#include "JITInlineMethods.h"
#include "JITOperationWrappers.h"
#include "JITToDFGDeferredCompilationCallback.h"
#include "JSAsyncFunction.h"
#include "JSAsyncGeneratorFunction.h"
#include "JSGeneratorFunction.h"
#include "JSGlobalObjectFunctions.h"
#include "JSLexicalEnvironment.h"
#include "JSPropertyNameEnumerator.h"
#include "JSWithScope.h"
#include "ObjectConstructor.h"
#include "Operations.h"
#include "Options.h"
#include "PropertyName.h"
#include "Repatch.h"
#include "ScopedArguments.h"
#include "Symbol.h"
#include "ThrowScope.h"
#include <wtf/CurrentTime.h>
#include <wtf/MemoryMonitor.h>
#include <wtf/PlatformRunState.h>
#include <wtf/text/StringBuilder.h>

namespace JSC {

#if COMPILER(MSVC)
void * _ReturnAddress(void);
#pragma intrinsic(_ReturnAddress)

#define OUR_RETURN_ADDRESS _ReturnAddress()
#else
#define OUR_RETURN_ADDRESS __builtin_return_address(0)
#endif

#if ENABLE(OPCODE_SAMPLING)
#define CTI_SAMPLER vm->interpreter->sampler()
#else
#define CTI_SAMPLER 0
#endif

void JIT_OPERATION operationStackCheck(ExecState* exec, CodeBlock* codeBlock)
{
    // We pass in our own code block, because the callframe hasn't been populated.
    VM& vm = codeBlock->vm();
    CallFrame* callerFrame = exec->callerFrameSkippingVMEntrySentinel();

    if (LIKELY(vm.interpreter->stack().grow(&exec->registers()[virtualRegisterForLocal(codeBlock->frameRegisterCount())])))
    	return;

	NativeCallFrameTracer tracer(vm, callerFrame);
	vm.throwException(callerFrame, createStackOverflowError(callerFrame));
}

int32_t JIT_OPERATION operationCallArityCheck(ExecState* exec)
{
    VM& vm = exec->vm();

    int32_t missingArgCount = CommonSlowPaths::arityCheckFor(exec, vm, vm.interpreter->stack(), CodeForCall);
    if (missingArgCount < 0) {
        CallFrame* callerFrame = exec->callerFrameSkippingVMEntrySentinel();
        NativeCallFrameTracer tracer(vm, callerFrame);
        vm.throwException(callerFrame, createStackOverflowError(callerFrame));
    }

    return missingArgCount;
}

int32_t JIT_OPERATION operationConstructArityCheck(ExecState* exec)
{
    VM& vm = exec->vm();

    int32_t missingArgCount = CommonSlowPaths::arityCheckFor(exec, vm, vm.interpreter->stack(), CodeForConstruct);
    if (missingArgCount < 0) {
        CallFrame* callerFrame = exec->callerFrameSkippingVMEntrySentinel();
        NativeCallFrameTracer tracer(vm, callerFrame);
        vm.throwException(callerFrame, createStackOverflowError(callerFrame));
    }

    return missingArgCount;
}

EncodedJSValue JIT_OPERATION operationTryGetById(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Identifier ident = Identifier::fromUid(vm, uid);
    stubInfo->tookSlowPath = true;

    JSValue baseValue = JSValue::decode(base);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::VMInquiry);
    baseValue.getPropertySlot(vm, exec, ident, slot);
    return JSValue::encode(slot.getPureResult());
}

EncodedJSValue JIT_OPERATION operationTryGetByIdGeneric(ExecState* exec, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Identifier ident = Identifier::fromUid(vm, uid);

    JSValue baseValue = JSValue::decode(base);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::VMInquiry);
    baseValue.getPropertySlot(vm, exec, ident, slot);
    return JSValue::encode(slot.getPureResult());
}

EncodedJSValue JIT_OPERATION operationTryGetByIdOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Identifier ident = Identifier::fromUid(vm, uid);

    JSValue baseValue = JSValue::decode(base);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::VMInquiry);

    baseValue.getPropertySlot(vm, exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (stubInfo->considerCaching(baseValue.structureOrNull()) && !slot.isTaintedByOpaqueObject() && (slot.isCacheableValue() || slot.isCacheableGetter() || slot.isUnset()))
        repatchGetByID(exec, baseValue, ident, slot, *stubInfo, GetByIDKind::Try);

    return JSValue::encode(slot.getPureResult());
}

EncodedJSValue JIT_OPERATION operationGetByIdDirect(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    //auto scope = DECLARE_THROW_SCOPE(vm);
    Identifier ident = Identifier::fromUid(vm, uid);
    if (stubInfo)
    	stubInfo->tookSlowPath = true;

    JSValue baseValue = JSValue::decode(base);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::GetOwnProperty);

    bool found = baseValue.getOwnPropertySlot(exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    //scope.release();
    return JSValue::encode(found ? slot.getValue(exec, ident) : jsUndefined());
}

EncodedJSValue JIT_OPERATION operationGetByIdDirectGeneric(ExecState* exec, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    //auto scope = DECLARE_THROW_SCOPE(vm);
    Identifier ident = Identifier::fromUid(vm, uid);

    JSValue baseValue = JSValue::decode(base);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::GetOwnProperty);

    bool found = baseValue.getOwnPropertySlot(exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    //scope.release();
    return JSValue::encode(found ? slot.getValue(exec, ident) : jsUndefined());
}

EncodedJSValue JIT_OPERATION operationGetByIdDirectOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    //auto scope = DECLARE_THROW_SCOPE(vm);
    Identifier ident = Identifier::fromUid(vm, uid);

    JSValue baseValue = JSValue::decode(base);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::GetOwnProperty);

    bool found = baseValue.getOwnPropertySlot(exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (stubInfo->considerCaching(baseValue.structureOrNull()))
        repatchGetByID(exec, baseValue, ident, slot, *stubInfo, GetByIDKind::Direct);

    //scope.release();
    return JSValue::encode(found ? slot.getValue(exec, ident) : jsUndefined());
}

EncodedJSValue JIT_OPERATION operationGetById(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    if (stubInfo)
    	stubInfo->tookSlowPath = true;
    
    JSValue baseValue = JSValue::decode(base);
    Identifier ident = Identifier::fromUid(vm, *uid);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::Get);
    return JSValue::encode(baseValue.get(vm, exec, ident, slot));
}

EncodedJSValue JIT_OPERATION operationGetByIdGeneric(ExecState* exec, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue baseValue = JSValue::decode(base);
    Identifier ident = Identifier::fromUid(vm, *uid);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::Get);
    return JSValue::encode(baseValue.get(vm, exec, ident, slot));
}

EncodedJSValue JIT_OPERATION operationGetByIdNoIC(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[2].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[3].u.operand);

    PropertySlot slot(base, PropertySlot::InternalMethodType::Get);
    return JSValue::encode(base.get(vm, exec, ident, slot));
}

EncodedJSValue JIT_OPERATION operationGetByIdNoICNotIndex(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[2].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[3].u.operand);

    PropertySlot slot(base, PropertySlot::InternalMethodType::Get);
    bool hasSlot = base.getPropertySlot<false>(vm, exec, ident, slot);

    return JSValue::encode(!hasSlot ? jsUndefined() : slot.getValue(exec, ident));
}

EncodedJSValue JIT_OPERATION operationGetByIdOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    Identifier ident = Identifier::fromUid(vm, *uid);

	JSValue baseValue = JSValue::decode(base);
	PropertySlot slot(baseValue, PropertySlot::InternalMethodType::Get);

    bool hasResult = baseValue.getPropertySlot(vm, exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (stubInfo->considerCaching(baseValue.structureOrNull()))
		repatchGetByID(exec, baseValue, ident, slot, *stubInfo, GetByIDKind::Normal);

    return JSValue::encode(hasResult? slot.getValue(exec, ident) : jsUndefined());
}

EncodedJSValue JIT_OPERATION operationGetByIdWithThis(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, EncodedJSValue thisEncoded, UniquedStringImpl* uid)
{
    //SuperSamplerScope superSamplerScope(false);

    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Identifier ident = Identifier::fromUid(vm, *uid);

    stubInfo->tookSlowPath = true;

    JSValue baseValue = JSValue::decode(base);
    JSValue thisValue = JSValue::decode(thisEncoded);
    PropertySlot slot(thisValue, PropertySlot::InternalMethodType::Get);

    return JSValue::encode(baseValue.get(vm, exec, ident, slot));
}

EncodedJSValue JIT_OPERATION operationGetByIdWithThisGeneric(ExecState* exec, EncodedJSValue base, EncodedJSValue thisEncoded, UniquedStringImpl* uid)
{
    //SuperSamplerScope superSamplerScope(false);

    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Identifier ident = Identifier::fromUid(vm, *uid);

    JSValue baseValue = JSValue::decode(base);
    JSValue thisValue = JSValue::decode(thisEncoded);
    PropertySlot slot(thisValue, PropertySlot::InternalMethodType::Get);

    return JSValue::encode(baseValue.get(vm, exec, ident, slot));
}

EncodedJSValue JIT_OPERATION operationGetByIdWithThisOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, EncodedJSValue thisEncoded, UniquedStringImpl* uid)
{
    //SuperSamplerScope superSamplerScope(false);

    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Identifier ident = Identifier::fromUid(vm, *uid);

    JSValue baseValue = JSValue::decode(base);
    JSValue thisValue = JSValue::decode(thisEncoded);
    //LOG_IC((ICEvent::OperationGetByIdWithThisOptimize, baseValue.classInfoOrNull(*vm), ident));

    PropertySlot slot(thisValue, PropertySlot::InternalMethodType::Get);
    bool hasResult = baseValue.getPropertySlot(vm, exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (stubInfo->considerCaching(baseValue.structureOrNull()))
    	repatchGetByID(exec, baseValue, ident, slot, *stubInfo, GetByIDKind::WithThis);

    return JSValue::encode(hasResult? slot.getValue(exec, ident) : jsUndefined());
}

EncodedJSValue JIT_OPERATION operationInOptimize(ExecState* exec, StructureStubInfo* stubInfo, JSCell* base, UniquedStringImpl* key)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    if (!base->isObject()) {
    	vm.throwException(exec, createInvalidInParameterError(exec, base));
        return JSValue::encode(jsUndefined());
    }

    AccessType accessType = static_cast<AccessType>(stubInfo->accessType);

    Identifier ident = Identifier::fromUid(vm, key);
    PropertySlot slot(base, PropertySlot::InternalMethodType::HasProperty);
    bool result = asObject(base)->getPropertySlot(vm, exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    RELEASE_ASSERT(accessType == stubInfo->accessType);

	if (stubInfo->considerCaching(asObject(base)->structure()))
		repatchIn(exec, base, ident, result, slot, *stubInfo);

    return JSValue::encode(jsBoolean(result));
}

EncodedJSValue JIT_OPERATION operationIn(ExecState* exec, StructureStubInfo* stubInfo, JSCell* base, UniquedStringImpl* key)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    stubInfo->tookSlowPath = true;

    if (!base->isObject()) {
    	vm.throwException(exec, createInvalidInParameterError(exec, base));
        return JSValue::encode(jsUndefined());
    }

    Identifier ident = Identifier::fromUid(vm, key);
    return JSValue::encode(jsBoolean(asObject(base)->hasProperty(exec, ident)));
}

EncodedJSValue JIT_OPERATION operationGenericIn(ExecState* exec, JSCell* base, EncodedJSValue key)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return JSValue::encode(jsBoolean(CommonSlowPaths::opIn(exec, vm, base, JSValue::decode(key))));
}

EncodedJSValue JIT_OPERATION operationHasPrivateName(ExecState* exec, JSCell* base, EncodedJSValue key)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    auto scope = DECLARE_THROW_SCOPE(vm);
    if (!base->isObject()) {
        throwException(exec, scope, createInvalidInParameterError(exec, base));
        return encodedJSUndefined();
    }

    JSValue propertyValue = JSValue::decode(key);
    ASSERT(propertyValue.isSymbol());
    auto property = propertyValue.toPropertyKey(vm, exec);
    EXCEPTION_ASSERT(!scope.exception());

    return JSValue::encode(jsBoolean(asObject(base)->hasPrivateField(vm, exec, property)));
}

EncodedJSValue JIT_OPERATION operationHasPrivateBrand(ExecState* exec, JSCell* base, EncodedJSValue brand)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    auto scope = DECLARE_THROW_SCOPE(vm);
    if (!base->isObject()) {
        throwException(exec, scope, createInvalidInParameterError(exec, base));
        return encodedJSUndefined();
    }

    return JSValue::encode(jsBoolean(asObject(base)->hasPrivateBrand(vm, exec, JSValue::decode(brand))));
}

EncodedJSValue JIT_OPERATION operationCallGetter(ExecState* exec, JSCell* base, JSCell* getterSetter)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return JSValue::encode(jsCast<GetterSetter*>(getterSetter)->callGetter(exec, base));
}

void JIT_OPERATION operationPutByIdStrict(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    stubInfo->tookSlowPath = true;

    Identifier ident = Identifier::fromUid(vm, *uid);
    PutPropertySlot slot(JSValue::decode(encodedBase), true, exec->codeBlock()->putByIdContext());
    JSValue::decode(encodedBase).putInline(exec, vm, ident, JSValue::decode(encodedValue), slot);
}

void JIT_OPERATION operationPutByIdNonStrict(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    stubInfo->tookSlowPath = true;

    Identifier ident = Identifier::fromUid(vm, *uid);
    PutPropertySlot slot(JSValue::decode(encodedBase), false, exec->codeBlock()->putByIdContext());
    JSValue::decode(encodedBase).putInline(exec, vm, ident, JSValue::decode(encodedValue), slot);
}

void JIT_OPERATION operationPutByIdDirectStrict(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    stubInfo->tookSlowPath = true;

    JSValue baseValue = JSValue::decode(encodedBase);
    Identifier ident = Identifier::fromUid(vm, *uid);
    PutPropertySlot slot(baseValue, true, exec->codeBlock()->putByIdContext());
    CommonSlowPaths::putDirectWithReify(vm, exec, asObject(baseValue), ident, JSValue::decode(encodedValue), slot);
}

void JIT_OPERATION operationPutByIdDirectNonStrict(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    stubInfo->tookSlowPath = true;

    JSValue baseValue = JSValue::decode(encodedBase);
    Identifier ident = Identifier::fromUid(vm, *uid);
    PutPropertySlot slot(baseValue, false, exec->codeBlock()->putByIdContext());
    CommonSlowPaths::putDirectWithReify(vm, exec, asObject(baseValue), ident, JSValue::decode(encodedValue), slot);
}

void JIT_OPERATION operationPutByIdStrictGeneric(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[1].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[2].u.operand);
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    PutPropertySlot slot(base, true, codeBlock->putByIdContext());
    base.putInline(exec, vm, ident, value, slot);
}

void JIT_OPERATION operationPutByIdNonStrictGeneric(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[1].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[2].u.operand);
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    PutPropertySlot slot(base, false, codeBlock->putByIdContext());
    base.putInline(exec, vm, ident, value, slot);
}

void JIT_OPERATION operationPutByIdDirectStrictGeneric(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[1].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[2].u.operand);
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    PutPropertySlot slot(base, true, codeBlock->putByIdContext());
    CommonSlowPaths::putDirectWithReify(vm, exec, asObject(base), ident, value, slot);
}

void JIT_OPERATION operationPutByIdDirectNonStrictGeneric(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[1].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[2].u.operand);
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    PutPropertySlot slot(base, false, codeBlock->putByIdContext());
    CommonSlowPaths::putDirectWithReify(vm, exec, asObject(base), ident, value, slot);
}

void JIT_OPERATION operationPutByIdStrictOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    Identifier ident = Identifier::fromUid(vm, *uid);
    AccessType accessType = static_cast<AccessType>(stubInfo->accessType);

    JSValue value = JSValue::decode(encodedValue);
    JSValue baseValue = JSValue::decode(encodedBase);
    PutPropertySlot slot(baseValue, true, exec->codeBlock()->putByIdContext());

    Structure* structure = CommonSlowPaths::originalStructureBeforePut(vm, baseValue);
    baseValue.putInline(exec, vm, ident, value, slot);
    RETURN_IF_EXCEPTION(vm, void());
    
    if (accessType != static_cast<AccessType>(stubInfo->accessType))
        return;

    if (stubInfo->considerCaching(structure))
        repatchPutByID(exec, baseValue, structure, ident, slot, *stubInfo, PutKind::NotDirect);
}

void JIT_OPERATION operationPutByIdNonStrictOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    Identifier ident = Identifier::fromUid(vm, *uid);
    AccessType accessType = static_cast<AccessType>(stubInfo->accessType);

    JSValue value = JSValue::decode(encodedValue);
    JSValue baseValue = JSValue::decode(encodedBase);
    PutPropertySlot slot(baseValue, false, exec->codeBlock()->putByIdContext());

    Structure* structure = CommonSlowPaths::originalStructureBeforePut(vm, baseValue);
    baseValue.putInline(exec, vm, ident, value, slot);
    RETURN_IF_EXCEPTION(vm, void());

    if (accessType != static_cast<AccessType>(stubInfo->accessType))
        return;

    if (stubInfo->considerCaching(structure))
        repatchPutByID(exec, baseValue, structure, ident, slot, *stubInfo, PutKind::NotDirect);
}

void JIT_OPERATION operationPutByIdDirectStrictOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    Identifier ident = Identifier::fromUid(vm, *uid);
    AccessType accessType = static_cast<AccessType>(stubInfo->accessType);

    JSValue value = JSValue::decode(encodedValue);
    JSObject* baseObject = asObject(JSValue::decode(encodedBase));
    PutPropertySlot slot(baseObject, true, exec->codeBlock()->putByIdContext());
    Structure* structure = 0;
    CommonSlowPaths::putDirectWithReify(vm, exec, baseObject, ident, value, slot, &structure);
    RETURN_IF_EXCEPTION(vm, void());

    if (accessType != static_cast<AccessType>(stubInfo->accessType))
        return;

    if (stubInfo->considerCaching(structure))
        repatchPutByID(exec, baseObject, structure, ident, slot, *stubInfo, PutKind::Direct);
}

void JIT_OPERATION operationPutByIdDirectNonStrictOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    Identifier ident = Identifier::fromUid(vm, *uid);
    AccessType accessType = static_cast<AccessType>(stubInfo->accessType);

    JSValue value = JSValue::decode(encodedValue);
    JSObject* baseObject = asObject(JSValue::decode(encodedBase));
    PutPropertySlot slot(baseObject, false, exec->codeBlock()->putByIdContext());
    Structure* structure = nullptr;
    CommonSlowPaths::putDirectWithReify(vm, exec, baseObject, ident, value, slot, &structure);
    RETURN_IF_EXCEPTION(vm, void());

    if (accessType != static_cast<AccessType>(stubInfo->accessType))
        return;

    if (stubInfo->considerCaching(structure))
        repatchPutByID(exec, baseObject, structure, ident, slot, *stubInfo, PutKind::Direct);
}

#if 0
} // extern "C"

using PutPrivateFieldCallback = std::function<void(VM&, CodeBlock*, Structure*, PutPropertySlot&, const Identifier&)>;
ALWAYS_INLINE static void setPrivateField(VM& vm, ExecState* exec, JSValue baseValue, UniquedStringImpl* uid, JSValue value, const PutPrivateFieldCallback& callback = PutPrivateFieldCallback())
{
    auto scope = DECLARE_THROW_SCOPE(vm);

    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(globalObject);
    RETURN_IF_EXCEPTION(scope, void());
    CodeBlock* codeBlock = exec->codeBlock();
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->setPrivateField(vm, exec, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());

    if (callback)
        callback(vm, codeBlock, oldStructure, putSlot, ident);
}

ALWAYS_INLINE static void definePrivateField(VM& vm, ExecState* exec, JSValue baseValue, UniquedStringImpl* uid, JSValue value, const PutPrivateFieldCallback& callback = PutPrivateFieldCallback())
{
    auto scope = DECLARE_THROW_SCOPE(vm);

    Identifier ident = Identifier::fromUid(vm, uid);
    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(globalObject);
    RETURN_IF_EXCEPTION(scope, void());
    CodeBlock* codeBlock = exec->codeBlock();
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->definePrivateField(globalObject, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());

    if (callback)
        callback(vm, codeBlock, oldStructure, putSlot, ident);
}

extern "C" {
#endif

void JIT_OPERATION operationPutByIdDefinePrivateFieldStrict(ExecState* exec, StructureStubInfo*, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    Identifier ident = Identifier::fromUid(vm, *uid);
    JSValue value = JSValue::decode(encodedValue);
    JSValue baseValue = JSValue::decode(encodedBase);

    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());
    CodeBlock* codeBlock = exec->codeBlock();
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->definePrivateField(vm, exec, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());
}

void JIT_OPERATION operationPutByIdDefinePrivateFieldStrictOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    Identifier ident = Identifier::fromUid(vm, *uid);
    AccessType accessType = static_cast<AccessType>(stubInfo->accessType);
    JSValue value = JSValue::decode(encodedValue);
    JSValue baseValue = JSValue::decode(encodedBase);

    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());

    CodeBlock* codeBlock = exec->codeBlock();
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->definePrivateField(vm, exec, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());

    //LOG_IC((ICEvent::OperationPutByIdDefinePrivateFieldFieldStrictOptimize, baseObject->classInfo(vm), ident, putSlot.base() == baseObject));

	ASSERT_UNUSED(accessType, accessType == static_cast<AccessType>(stubInfo->accessType));

	if (stubInfo->considerCaching(oldStructure))
		repatchPutByID(exec, baseObject, oldStructure, ident, putSlot, *stubInfo, PutKind::DirectPrivateFieldDefine);
}

void JIT_OPERATION operationPutByIdDefinePrivateFieldStrictGeneric(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue baseValue = exec->r(pc[1].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[2].u.operand);
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->definePrivateField(vm, exec, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());
}

void JIT_OPERATION operationPutByIdSetPrivateFieldStrict(ExecState* exec, StructureStubInfo*, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    Identifier ident = Identifier::fromUid(vm, *uid);
    JSValue value = JSValue::decode(encodedValue);
    JSValue baseValue = JSValue::decode(encodedBase);

    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());
    CodeBlock* codeBlock = exec->codeBlock();
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->setPrivateField(vm, exec, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());
}

void JIT_OPERATION operationPutByIdSetPrivateFieldStrictOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    Identifier ident = Identifier::fromUid(vm, *uid);
    AccessType accessType = static_cast<AccessType>(stubInfo->accessType);
    JSValue value = JSValue::decode(encodedValue);
    JSValue baseValue = JSValue::decode(encodedBase);

    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());
    CodeBlock* codeBlock = exec->codeBlock();
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->setPrivateField(vm, exec, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());

    //LOG_IC((ICEvent::OperationPutByIdPutPrivateFieldFieldStrictOptimize, baseObject->classInfo(vm), ident, putSlot.base() == baseObject));

	ASSERT_UNUSED(accessType, accessType == static_cast<AccessType>(stubInfo->accessType));

	if (stubInfo->considerCaching(oldStructure))
		repatchPutByID(exec, baseObject, oldStructure, ident, putSlot, *stubInfo, PutKind::DirectPrivateFieldSet);
}

void JIT_OPERATION operationPutByIdSetPrivateFieldStrictGeneric(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue baseValue = exec->r(pc[1].u.operand).jsValue();
    const Identifier& ident = codeBlock->identifier(pc[2].u.operand);
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    ASSERT(ident.isPrivateName());

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());
    Structure* oldStructure = baseObject->structure(vm);

    PutPropertySlot putSlot(baseObject, true, codeBlock->putByIdContext());
    baseObject->setPrivateField(vm, exec, ident, value, putSlot);
    RETURN_IF_EXCEPTION(scope, void());
}

ALWAYS_INLINE static bool isStringOrSymbol(JSValue value)
{
    return value.isString() || value.isSymbol();
}

static inline bool isCacheableIdentifierCell(JSCell* cell)
{
    if (cell->isSymbol())
        return true;
    if (!cell->isString())
        return false;
    JSString* string = jsCast<JSString*>(cell);
    if (const StringImpl* impl = string->tryGetValueImpl())
        return impl->isAtomic();
    return false;
}

static inline bool isCacheableIdentifierCell(JSValue value)
{
    if (!value.isCell())
        return false;
    return isCacheableIdentifierCell(value.asCell());
}

static ALWAYS_INLINE void putByVal(CallFrame* callFrame, VM& vm, JSValue baseValue, JSValue subscript, JSValue value, ByValInfo* byValInfo, ECMAMode ecmaMode)
{
	ThrowScope scope(vm);
    if (Optional<uint32_t> index = subscript.tryGetAsUint32Index()) {
    	if (byValInfo)
    		byValInfo->tookSlowPath = true;
        uint32_t i = *index;
        if (baseValue.isObject()) {
            JSObject* object = asObject(baseValue);
            if (object->canSetIndexQuickly(i, value)) {
                object->setIndexQuickly(vm, i, value);
                return;
            }

        	if (byValInfo)
        		byValInfo->arrayProfile->setOutOfBounds();
			scope.release();
			object->methodTable()->putByIndex(object, callFrame, i, value, ecmaMode.isStrict());
			return;
        }

        scope.release();
        baseValue.putByIndex(callFrame, i, value, ecmaMode.isStrict());
        return;
    }
    if (byValInfo && subscript.isNumber()) {
        byValInfo->tookSlowPath = true;
        if (baseValue.isObject())
        	byValInfo->arrayProfile->setOutOfBounds();
    }

	Identifier property = subscript.toPropertyKey(vm, callFrame);
    // Don't put to an object if toString threw an exception.
    RETURN_IF_EXCEPTION(scope, void());

    if (byValInfo && byValInfo->stubInfo && (!isStringOrSymbol(subscript) || byValInfo->cachedId != property))
        byValInfo->tookSlowPath = true;

    scope.release();
	PutPropertySlot slot(baseValue, ecmaMode.isStrict());
	baseValue.putInline(callFrame, vm, property, value, slot);
}

static ALWAYS_INLINE void directPutByVal(CallFrame* callFrame, VM& vm, JSObject* baseObject, JSValue subscript, JSValue value, ByValInfo* byValInfo, ECMAMode ecmaMode)
{
    if (Optional<uint32_t> maybeIndex = subscript.tryGetAsUint32Index()) {
        // Despite its name, JSValue::isUInt32 will return true only for positive boxed int32_t; all those values are valid array indices.
    	if (byValInfo)
    		byValInfo->tookSlowPath = true;
        uint32_t index = *maybeIndex;

        switch (baseObject->structure()->indexingType()) {
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            if (index < baseObject->butterfly()->vectorLength())
                break;
            // FALLTHROUGH;
        default:
        	if (byValInfo)
        		byValInfo->arrayProfile->setOutOfBounds();
            break;
        }

		baseObject->putDirectIndex(callFrame, index, value, 0, ecmaMode.isStrict() ? PutDirectIndexShouldThrow : PutDirectIndexShouldNotThrow);
		return;
	}

    // Don't put to an object if toString threw an exception.
    Identifier property = subscript.toPropertyKey(vm, callFrame);
    RETURN_IF_EXCEPTION(vm, void());

    if (Optional<uint32_t> index = parseIndex(property)) {
    	if (byValInfo)
    		byValInfo->tookSlowPath = true;
        baseObject->putDirectIndex(callFrame, index.value(), value, 0, ecmaMode.isStrict() ? PutDirectIndexShouldThrow : PutDirectIndexShouldNotThrow);
        return;
    }

    if (byValInfo && byValInfo->stubInfo && (!isStringOrSymbol(subscript) || byValInfo->cachedId != property))
        byValInfo->tookSlowPath = true;

	PutPropertySlot slot(baseObject, ecmaMode.isStrict());
    CommonSlowPaths::putDirectWithReify(vm, callFrame, baseObject, property, value, slot);
}

enum class OptimizationResult {
    NotOptimized,
    SeenOnce,
    Optimized,
    GiveUp,
};

static OptimizationResult tryPutByValOptimize(VM& vm, ExecState* exec, JSValue baseValue, JSValue subscript, ByValInfo* byValInfo, ECMAMode ecmaMode, ReturnAddressPtr returnAddress)
{
    // See if it's worth optimizing at all.
    OptimizationResult optimizationResult = OptimizationResult::NotOptimized;

    if (baseValue.isObject() && subscript.isInt32()) {
        JSObject* object = asObject(baseValue);

        ASSERT(exec->locationAsBytecodeOffset());
        ASSERT(!byValInfo->stubRoutine);

        Structure* structure = object->structure(vm);
        if (hasOptimizableIndexing(structure)) {
            // Attempt to optimize.
            JITArrayMode arrayMode = jitArrayModeForStructure(structure);
            if (jitArrayModePermitsPut(arrayMode) && arrayMode != byValInfo->arrayMode) {
                CodeBlock* codeBlock = exec->codeBlock();
            	// ConcurrentJITLocker locker(codeBlock->m_lock);
            	byValInfo->arrayProfile->computeUpdatedPrediction(codeBlock, structure);

                JIT::compilePutByVal(&vm, codeBlock, byValInfo, returnAddress, arrayMode, ecmaMode);
                optimizationResult = OptimizationResult::Optimized;
			}
        }

        // If we failed to patch and we have some object that intercepts indexed get, then don't even wait until 10 times.
        if (optimizationResult != OptimizationResult::Optimized && object->structure(vm)->typeInfo().interceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero())
            optimizationResult = OptimizationResult::GiveUp;
    }

    if (baseValue.isObject() && isStringOrSymbol(subscript)) {
        const Identifier propertyName = subscript.toPropertyKey(vm, exec);
        RETURN_IF_EXCEPTION(vm, OptimizationResult::GiveUp);
        if (subscript.isSymbol() || !parseIndex(propertyName)) {
            ASSERT(exec->locationAsBytecodeOffset());
            ASSERT(!byValInfo->stubRoutine);
            if (byValInfo->seen) {
                if (byValInfo->cachedId == propertyName) {
                	JIT::compilePutByValWithCachedId(&vm, exec->codeBlock(), byValInfo, returnAddress, PutKind::NotDirect, propertyName, ecmaMode);
                	optimizationResult = OptimizationResult::Optimized;
                } else {
					// Seem like a generic property access site.
					optimizationResult = OptimizationResult::GiveUp;
				}
			} else {
                CodeBlock* codeBlock = exec->codeBlock();
                //ConcurrentJITLocker locker(codeBlock->m_lock);
				byValInfo->seen = true;
				byValInfo->cachedId = propertyName;
                if (subscript.isSymbol())
                    byValInfo->cachedSymbol.set(vm, codeBlock->ownerExecutable(), asSymbol(subscript));
				optimizationResult = OptimizationResult::SeenOnce;
			}
		}
	}

	if (optimizationResult != OptimizationResult::Optimized && optimizationResult != OptimizationResult::SeenOnce) {
        // If we take slow path more than 10 times without patching then make sure we
        // never make that mistake again. For cases where we see non-index-intercepting
        // objects, this gives 10 iterations worth of opportunity for us to observe
        // that the put_by_val may be polymorphic. We count up slowPathCount even if
        // the result is GiveUp.
        if (++byValInfo->slowPathCount >= 10)
            optimizationResult = OptimizationResult::GiveUp;
    }

    return optimizationResult;
}

static void JIT_OPERATION operationPutByValOptimize(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo, ECMAMode ecmaMode, ReturnAddressPtr returnAddress)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue subscript = JSValue::decode(encodedSubscript);
    JSValue value = JSValue::decode(encodedValue);
    OptimizationResult result = tryPutByValOptimize(vm, exec, baseValue, subscript, byValInfo, ecmaMode, returnAddress);
    RETURN_IF_EXCEPTION(vm, void());
    if (result == OptimizationResult::GiveUp) {
        // Don't ever try to optimize.
        byValInfo->tookSlowPath = true;
        ctiPatchCallByReturnAddress(returnAddress, FunctionPtr(ecmaMode.isStrict() ? operationPutByValGenericStrict : operationPutByValGenericSloppy));
    }
    putByVal(exec, vm, baseValue, subscript, value, byValInfo, ecmaMode);
}

void JIT_OPERATION operationPutByValOptimizeSloppy(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationPutByValOptimize(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::sloppy(), ReturnAddressPtr(OUR_RETURN_ADDRESS));
}

void JIT_OPERATION operationPutByValOptimizeStrict(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationPutByValOptimize(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::strict(), ReturnAddressPtr(OUR_RETURN_ADDRESS));
}

static OptimizationResult tryDirectPutByValOptimize(VM& vm, ExecState* exec, JSObject* object, JSValue subscript, ByValInfo* byValInfo, ECMAMode ecmaMode, ReturnAddressPtr returnAddress)
{
    // See if it's worth optimizing at all.
    OptimizationResult optimizationResult = OptimizationResult::NotOptimized;

    if (subscript.isInt32()) {
        ASSERT(exec->locationAsBytecodeOffset());
        ASSERT(!byValInfo->stubRoutine);

        Structure* structure = object->structure(vm);
        if (hasOptimizableIndexing(structure)) {
            // Attempt to optimize.
            JITArrayMode arrayMode = jitArrayModeForStructure(structure);
            if (jitArrayModePermitsPutDirect(arrayMode) && arrayMode != byValInfo->arrayMode) {
                CodeBlock* codeBlock = exec->codeBlock();
            	// ConcurrentJITLocker locker(codeBlock->m_lock);
            	byValInfo->arrayProfile->computeUpdatedPrediction(codeBlock, structure);

                JIT::compileDirectPutByVal(&vm, codeBlock, byValInfo, returnAddress, arrayMode, ecmaMode);
                optimizationResult = OptimizationResult::Optimized;
			}
		}

		// If we failed to patch and we have some object that intercepts indexed get, then don't even wait until 10 times.
		if (optimizationResult != OptimizationResult::Optimized && object->structure(vm)->typeInfo().interceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero())
			optimizationResult = OptimizationResult::GiveUp;
	} else if (isStringOrSymbol(subscript)) {
		const Identifier propertyName = subscript.toPropertyKey(vm, exec);
        RETURN_IF_EXCEPTION(vm, OptimizationResult::GiveUp);
        if (subscript.isSymbol() || !parseIndex(propertyName)) {
			ASSERT(exec->locationAsBytecodeOffset());
			ASSERT(!byValInfo->stubRoutine);
            if (byValInfo->seen) {
                if (byValInfo->cachedId == propertyName) {
                	JIT::compilePutByValWithCachedId(&vm, exec->codeBlock(), byValInfo, returnAddress, PutKind::Direct, propertyName, ecmaMode);
                	optimizationResult = OptimizationResult::Optimized;
                } else {
                    // Seem like a generic property access site.
                    optimizationResult = OptimizationResult::GiveUp;
                }
            } else {
                CodeBlock* codeBlock = exec->codeBlock();
                //ConcurrentJITLocker locker(codeBlock->m_lock);
                byValInfo->seen = true;
                byValInfo->cachedId = propertyName;
                if (subscript.isSymbol())
                    byValInfo->cachedSymbol.set(vm, codeBlock->ownerExecutable(), asSymbol(subscript));
                optimizationResult = OptimizationResult::SeenOnce;
            }
        }
    }

    if (optimizationResult != OptimizationResult::Optimized && optimizationResult != OptimizationResult::SeenOnce) {
		// If we take slow path more than 10 times without patching then make sure we
		// never make that mistake again. For cases where we see non-index-intercepting
		// objects, this gives 10 iterations worth of opportunity for us to observe
		// that the get_by_val may be polymorphic. We count up slowPathCount even if
		// the result is GiveUp.
		if (++byValInfo->slowPathCount >= 10)
			optimizationResult = OptimizationResult::GiveUp;
	}

	return optimizationResult;
}

static void JIT_OPERATION operationDirectPutByValOptimize(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo, ECMAMode ecmaMode, ReturnAddressPtr returnAddress)
{
	VM& vm = exec->vm();
	NativeCallFrameTracer tracer(&vm, exec);

	JSValue baseValue = JSValue::decode(encodedBaseValue);
	JSValue subscript = JSValue::decode(encodedSubscript);
	JSValue value = JSValue::decode(encodedValue);
	RELEASE_ASSERT(baseValue.isObject());
	JSObject* object = asObject(baseValue);
	OptimizationResult result = tryDirectPutByValOptimize(vm, exec, object, subscript, byValInfo, ecmaMode, returnAddress);
	RETURN_IF_EXCEPTION(vm, void());
	if (result == OptimizationResult::GiveUp) {
		// Don't ever try to optimize.
		byValInfo->tookSlowPath = true;
		ctiPatchCallByReturnAddress(returnAddress, FunctionPtr(ecmaMode.isStrict() ? operationDirectPutByValGenericStrict : operationDirectPutByValGenericSloppy));
	}

	directPutByVal(exec, vm, object, subscript, value, byValInfo, ecmaMode);
}

void JIT_OPERATION operationDirectPutByValOptimizeSloppy(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationDirectPutByValOptimize(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::sloppy(), ReturnAddressPtr(OUR_RETURN_ADDRESS));
}

void JIT_OPERATION operationDirectPutByValOptimizeStrict(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationDirectPutByValOptimize(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::strict(), ReturnAddressPtr(OUR_RETURN_ADDRESS));
}

static void JIT_OPERATION operationPutByValGeneric(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo, ECMAMode ecmaMode)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue subscript = JSValue::decode(encodedSubscript);
    JSValue value = JSValue::decode(encodedValue);

    putByVal(exec, vm, baseValue, subscript, value, byValInfo, ecmaMode);
}

void JIT_OPERATION operationPutByValGenericSloppy(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationPutByValGeneric(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::sloppy());
}

void JIT_OPERATION operationPutByValGenericStrict(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationPutByValGeneric(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::strict());
}

void JIT_OPERATION operationDirectPutByVal(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[1].u.operand).jsValue();
    JSValue subscript = exec->r(pc[2].u.operand).jsValue();
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    RELEASE_ASSERT(baseValue.isObject());
    ECMAMode ecmaMode = ECMAMode::fromByte(pc[4].u.operand);
    directPutByVal(exec, vm, asObject(base), subscript, value, nullptr, ecmaMode);
}

void JIT_OPERATION operationPutByVal(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[1].u.operand).jsValue();
    JSValue subscript = exec->r(pc[2].u.operand).jsValue();
    JSValue value = exec->r(pc[3].u.operand).jsValue();

    ECMAMode ecmaMode = ECMAMode::fromByte(pc[4].u.operand);
    putByVal(exec, vm, base, subscript, value, nullptr, ecmaMode);
}

static void JIT_OPERATION operationDirectPutByValGeneric(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo, ECMAMode ecmaMode)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue subscript = JSValue::decode(encodedSubscript);
    JSValue value = JSValue::decode(encodedValue);
    RELEASE_ASSERT(baseValue.isObject());
    directPutByVal(exec, vm, asObject(baseValue), subscript, value, byValInfo, ecmaMode);
}

void JIT_OPERATION operationDirectPutByValGenericSloppy(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationDirectPutByValGeneric(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::sloppy());
}

void JIT_OPERATION operationDirectPutByValGenericStrict(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo)
{
	operationDirectPutByValGeneric(exec, encodedBaseValue, encodedSubscript, encodedValue, byValInfo, ECMAMode::strict());
}

void JIT_OPERATION operationSetPrivateBrandOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedBaseValue, EncodedJSValue encodedBrand)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue brand = JSValue::decode(encodedBrand);

    ASSERT(baseValue.isObject());
    ASSERT(brand.isSymbol());

    JSObject* baseObject = asObject(baseValue);
    Structure* oldStructure = baseObject->structure(vm);
    baseObject->setPrivateBrand(vm, exec, brand);
    RETURN_IF_EXCEPTION(scope, void());

#if 0
    CodeBlock* codeBlock = callFrame->codeBlock();
    if (CacheableIdentifier::isCacheableIdentifierCell(brand)) {
        CacheableIdentifier identifier = CacheableIdentifier::createFromCell(brand.asCell());
        if (stubInfo->considerCachingBy(vm, codeBlock, baseObject->structure(vm), identifier))
            repatchSetPrivateBrand(globalObject, codeBlock, baseObject, oldStructure, identifier, *stubInfo);
    }
#endif
}

void JIT_OPERATION operationSetPrivateBrandGeneric(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedBaseValue, EncodedJSValue encodedBrand)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue brand = JSValue::decode(encodedBrand);

    if (stubInfo)
        stubInfo->tookSlowPath = true;

    ASSERT(baseValue.isObject());
    ASSERT(brand.isSymbol());

    JSObject* baseObject = asObject(baseValue);
    baseObject->setPrivateBrand(vm, exec, brand);
    RETURN_IF_EXCEPTION(scope, void());
}

void JIT_OPERATION operationCheckPrivateBrandOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedBaseValue, EncodedJSValue encodedBrand)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue brand = JSValue::decode(encodedBrand);

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());

    ASSERT(brand.isSymbol());

    baseObject->checkPrivateBrand(vm, exec, brand);
    RETURN_IF_EXCEPTION(scope, void());

#if 0
    CodeBlock* codeBlock = callFrame->codeBlock();
    if (CacheableIdentifier::isCacheableIdentifierCell(brand)) {
        CacheableIdentifier identifier = CacheableIdentifier::createFromCell(brand.asCell());
        if (stubInfo->considerCachingBy(vm, codeBlock, baseObject->structure(vm), identifier))
            repatchCheckPrivateBrand(globalObject, codeBlock, baseObject, identifier, *stubInfo);
    }
#endif
}

void JIT_OPERATION operationCheckPrivateBrandGeneric(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedBaseValue, EncodedJSValue encodedBrand)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue brand = JSValue::decode(encodedBrand);

    if (stubInfo)
    	stubInfo->tookSlowPath = true;

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());

    ASSERT(brand.isSymbol());

    baseObject->checkPrivateBrand(vm, exec, brand);
    RETURN_IF_EXCEPTION(scope, void());
}

void JIT_OPERATION operationPutPrivateNameOptimize(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo, int32_t putKindValue)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    CodeBlock* codeBlock = exec->codeBlock();

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue subscript = JSValue::decode(encodedSubscript);
    JSValue value = JSValue::decode(encodedValue);
    PrivateFieldPutKind putKind = PrivateFieldPutKind::fromByte(static_cast<uint8_t>(putKindValue));

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());

    auto propertyName = subscript.toPropertyKey(vm, exec);
    EXCEPTION_ASSERT(!scope.exception());

    OptimizationResult optimizationResult = OptimizationResult::NotOptimized;

    if (baseValue.isObject() && isCacheableIdentifierCell(subscript)) {
        ASSERT(subscript.isSymbol());
        ASSERT(callFrame->bytecodeIndex() != BytecodeIndex(0));
        ASSERT(!byValInfo->stubRoutine);
        if (byValInfo->seen) {
            if (byValInfo->cachedId == propertyName) {
                JIT::compilePutPrivateNameWithCachedId(&vm, codeBlock, byValInfo, ReturnAddressPtr(OUR_RETURN_ADDRESS), byValInfo->cachedId);
                optimizationResult = OptimizationResult::Optimized;
            } else {
                // Seem like a generic property access site.
                optimizationResult = OptimizationResult::GiveUp;
            }
        } else {
            {
                //ConcurrentJSLocker locker(codeBlock->m_lock);
                byValInfo->seen = true;
                byValInfo->cachedId = propertyName;
                optimizationResult = OptimizationResult::SeenOnce;
            }
            vm.heap.writeBarrier(codeBlock->ownerExecutable(), subscript.asCell());
        }
    }

    if (optimizationResult != OptimizationResult::Optimized && optimizationResult != OptimizationResult::SeenOnce) {
        // If we take slow path more than 10 times without patching then make sure we
        // never make that mistake again. This gives 10 iterations worth of opportunity
        // for us to observe that the put_private_name may be polymorphic.
        // We count up slowPathCount even if the result is GiveUp.
        if (++byValInfo->slowPathCount >= 10)
            optimizationResult = OptimizationResult::GiveUp;
    }

    if (optimizationResult == OptimizationResult::GiveUp) {
        // Don't ever try to optimize.
        byValInfo->tookSlowPath = true;
        ctiPatchCallByReturnAddress(ReturnAddressPtr(OUR_RETURN_ADDRESS), FunctionPtr(operationPutPrivateNameGeneric));
    }

    scope.release();

    // Private fields can only be accessed within class lexical scope
    // and class methods are always in strict mode
    const bool isStrictMode = true;
    PutPropertySlot slot(baseObject, isStrictMode);
    if (putKind.isDefine())
        baseObject->definePrivateField(vm, exec, propertyName, value, slot);
    else
        baseObject->setPrivateField(vm, exec, propertyName, value, slot);
}

// We need to match the signature of operationPutPrivateNameOptimize
void JIT_OPERATION operationPutPrivateNameGeneric(ExecState* exec, EncodedJSValue encodedBaseValue, EncodedJSValue encodedSubscript, EncodedJSValue encodedValue, ByValInfo* byValInfo, int32_t putKindValue)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue baseValue = JSValue::decode(encodedBaseValue);
    JSValue subscript = JSValue::decode(encodedSubscript);
    JSValue value = JSValue::decode(encodedValue);
    PrivateFieldPutKind privateFieldPutKind = PrivateFieldPutKind::fromByte(static_cast<uint8_t>(putKindValue));

    JSObject* baseObject = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, void());

    auto propertyName = subscript.toPropertyKey(vm, exec);
    EXCEPTION_ASSERT(!scope.exception());

    scope.release();

    // Private fields can only be accessed within class lexical scope
    // and class methods are always in strict mode
    const bool isStrictMode = true;
    PutPropertySlot slot(baseObject, isStrictMode);
    if (privateFieldPutKind.isDefine())
        baseObject->definePrivateField(vm, exec, propertyName, value, slot);
    else
        baseObject->setPrivateField(vm, exec, propertyName, value, slot);

    if (byValInfo)
        byValInfo->tookSlowPath = true;
}

void JIT_OPERATION operationPutPrivateNameGenericDFG(ExecState* exec, EncodedJSValue encodedBaseValue, JSCell* subscriptCell, EncodedJSValue encodedValue, int32_t putKindValue)
{
	operationPutPrivateNameGeneric(exec, encodedBaseValue, JSValue::encode(JSValue(subscriptCell)), encodedValue, nullptr, putKindValue);
}

static EncodedJSValue JIT_OPERATION operationCallEval(ExecState* execCallee, ECMAMode ecmaMode, ReturnAddressPtr returnAddress)
{
    //UNUSED_PARAM(exec);

    execCallee->setReturnPC(static_cast<Instruction*>(returnAddress.value()));
    execCallee->setCodeBlock(0);

    if (!isHostFunction(execCallee->calleeAsValue(), globalFuncEval))
        return JSValue::encode(JSValue());

    VM& vm = execCallee->vm();
    JSValue result = eval(execCallee, ecmaMode);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationCallEvalSloppy(ExecState* execCallee)
{
	return operationCallEval(execCallee, ECMAMode::sloppy(), ReturnAddressPtr(OUR_RETURN_ADDRESS));
}

EncodedJSValue JIT_OPERATION operationCallEvalStrict(ExecState* execCallee)
{
	return operationCallEval(execCallee, ECMAMode::strict(), ReturnAddressPtr(OUR_RETURN_ADDRESS));
}

static void* handleHostCall(ExecState* execCallee, JSValue callee, CodeSpecializationKind kind)
{
    ExecState* exec = execCallee->callerFrame();
    auto& vm = exec->vm();

    execCallee->setCodeBlock(0);

    if (kind == CodeForCall) {
        CallData callData = getCallData(vm, callee);
    
        ASSERT(callData.type != CallData::Type::JS);
    
        if (LIKELY(callData.type == CallData::Type::Native)) {
        	if(callee.isObject()) {
				NativeCallFrameTracer tracer(vm, execCallee);
				execCallee->setCallee(asObject(callee));
				vm.hostCallReturnValue = JSValue::decode(callData.native.function(execCallee));
				if (UNLIKELY(vm.exception()))
					return vm.getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress();

				return reinterpret_cast<void*>(getHostCallReturnValue);
        	}
        }

        ASSERT(callData.type == CallData::Type::None);
        vm.throwException(exec, createNotAFunctionError(exec, callee));
        return vm.getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress();
    }

    ASSERT(kind == CodeForConstruct);
    
    auto constructData = getConstructData(vm, callee);
    
    ASSERT(constructData.type != CallData::Type::JS);
    
    if (LIKELY(constructData.type == CallData::Type::Native)) {
    	if(callee.isObject()) {
			NativeCallFrameTracer tracer(vm, execCallee);
			execCallee->setCallee(asObject(callee));
			vm.hostCallReturnValue = JSValue::decode(constructData.native.function(execCallee));
			if (UNLIKELY(vm.exception()))
				return vm.getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress();

			return reinterpret_cast<void*>(getHostCallReturnValue);
    	}
    }

    ASSERT(constructData.type == CallData::Type::None);
    vm.throwException(exec, createNotAConstructorError(exec, callee));
    return vm.getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress();
}

char* JIT_OPERATION operationLinkCall(ExecState* execCallee, CallLinkInfo* callLinkInfo)
{
    ExecState* exec = execCallee->callerFrame();
    VM* vm = &exec->vm();
    CodeSpecializationKind kind = callLinkInfo->specializationKind();
    NativeCallFrameTracer tracer(vm, exec);
    
    JSValue calleeAsValue = execCallee->calleeAsValue();
    JSCell* calleeAsFunctionCell = getJSFunction(calleeAsValue);
    if (!calleeAsFunctionCell)
        return reinterpret_cast<char*>(handleHostCall(execCallee, calleeAsValue, kind));

    JSFunction* callee = jsCast<JSFunction*>(calleeAsFunctionCell);
    JSScope* scope = callee->scopeUnchecked();
    ExecutableBase* executable = callee->executable();

    MacroAssemblerCodePtr codePtr;
    CodeBlock* codeBlock = 0;
    if (executable->isHostFunction()){
        codePtr = executable->entrypointFor(*vm, kind, MustCheckArity, callLinkInfo->registerPreservationMode());
    } else {
        FunctionExecutable* functionExecutable = static_cast<FunctionExecutable*>(executable);

        if (!isCall(kind) && functionExecutable->constructAbility() == ConstructAbility::CannotConstruct) {
            vm->throwException(exec, createNotAConstructorError(exec, callee));
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
        }

    	if(UNLIKELY(WTF::isLowMemory())) {
            vm->throwException(exec, createOutOfMemoryError(exec, nullptr));
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
    	}

        CodeBlock** codeBlockSlot = execCallee->addressOfCodeBlock();
        Exception* error = functionExecutable->prepareForExecution<FunctionExecutable>(execCallee, callee, scope, kind, *codeBlockSlot);
        if (UNLIKELY(error)) {
            vm->throwException(exec, error);
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
        }
        codeBlock = *codeBlockSlot;
        if(!codeBlock){
            vm->throwException(exec, createOutOfMemoryError(exec));
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
        }
        ArityCheckMode arity;
        if (execCallee->argumentCountIncludingThis() < static_cast<size_t>(codeBlock->numParameters()) || callLinkInfo->callType() == CallLinkInfo::CallVarargs || callLinkInfo->callType() == CallLinkInfo::ConstructVarargs)
        	arity = MustCheckArity;
		else
			arity = ArityCheckNotRequired;
		codePtr = functionExecutable->entrypointFor(*vm, kind, arity, callLinkInfo->registerPreservationMode());
	}
    if (!callLinkInfo->seenOnce())
        callLinkInfo->setSeen();
    else
        linkFor(execCallee, *callLinkInfo, codeBlock, callee, codePtr);

    return reinterpret_cast<char*>(codePtr.executableAddress());
}

inline char* virtualForWithFunction(
    ExecState* execCallee, CallLinkInfo* callLinkInfo, JSCell*& calleeAsFunctionCell)
{
    ExecState* exec = execCallee->callerFrame();
    VM* vm = &exec->vm();
    CodeSpecializationKind kind = callLinkInfo->specializationKind();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue calleeAsValue = execCallee->calleeAsValue();
    calleeAsFunctionCell = getJSFunction(calleeAsValue);
    if (UNLIKELY(!calleeAsFunctionCell))
        return reinterpret_cast<char*>(handleHostCall(execCallee, calleeAsValue, kind));
    
    JSFunction* function = jsCast<JSFunction*>(calleeAsFunctionCell);
    JSScope* scope = function->scopeUnchecked();
    ExecutableBase* executable = function->executable();
    if (UNLIKELY(!executable->hasJITCodeFor(kind))) {
        FunctionExecutable* functionExecutable = static_cast<FunctionExecutable*>(executable);

        if (!isCall(kind) && functionExecutable->constructAbility() == ConstructAbility::CannotConstruct) {
            vm->throwException(exec, createNotAConstructorError(exec, function));
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
        }

    	if(UNLIKELY(WTF::isLowMemory())) {
            vm->throwException(exec, createOutOfMemoryError(exec, nullptr));
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
    	}

        CodeBlock** codeBlockSlot = execCallee->addressOfCodeBlock();
        Exception* error = functionExecutable->prepareForExecution<FunctionExecutable>(execCallee, function, scope, kind, *codeBlockSlot);
        if (UNLIKELY(error)) {
        	vm->throwException(exec, error);
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
        }
        CodeBlock* codeBlock = *codeBlockSlot;
        if(!codeBlock){
            vm->throwException(exec, createOutOfMemoryError(exec));
            return reinterpret_cast<char*>(vm->getCTIStub(throwExceptionFromCallSlowPathGenerator).code().executableAddress());
        }
    }
    return reinterpret_cast<char*>(executable->entrypointFor(
        *vm, kind, MustCheckArity, callLinkInfo->registerPreservationMode()).executableAddress());
}

char* JIT_OPERATION operationLinkPolymorphicCall(ExecState* execCallee, CallLinkInfo* callLinkInfo)
{
    ASSERT(callLinkInfo->specializationKind() == CodeForCall);
    JSCell* calleeAsFunctionCell;
    char* result = virtualForWithFunction(execCallee, callLinkInfo, calleeAsFunctionCell);

    linkPolymorphicCall(execCallee, *callLinkInfo, CallVariant(calleeAsFunctionCell));

    return result;
}

char* JIT_OPERATION operationVirtualCall(ExecState* execCallee, CallLinkInfo* callLinkInfo)
{
    JSCell* calleeAsFunctionCellIgnored;
    return virtualForWithFunction(execCallee, callLinkInfo, calleeAsFunctionCellIgnored);
}

size_t JIT_OPERATION operationCompareLess(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return jsLess<true>(exec, JSValue::decode(encodedOp1), JSValue::decode(encodedOp2));
}

size_t JIT_OPERATION operationCompareLessEq(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return jsLessEq<true>(exec, JSValue::decode(encodedOp1), JSValue::decode(encodedOp2));
}

size_t JIT_OPERATION operationCompareGreater(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return jsLess<false>(exec, JSValue::decode(encodedOp2), JSValue::decode(encodedOp1));
}

size_t JIT_OPERATION operationCompareGreaterEq(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return jsLessEq<false>(exec, JSValue::decode(encodedOp2), JSValue::decode(encodedOp1));
}

size_t JIT_OPERATION operationCompareEq(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return JSValue::equalSlowCaseInline(exec, JSValue::decode(encodedOp1), JSValue::decode(encodedOp2));
}

size_t JIT_OPERATION operationCompareStrictEq(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    JSGlobalData* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue src1 = JSValue::decode(encodedOp1);
    JSValue src2 = JSValue::decode(encodedOp2);

    return JSValue::strictEqual(exec, src1, src2);
}

#if USE(JSVALUE64)
EncodedJSValue JIT_OPERATION operationCompareStringEq(ExecState* exec, JSCell* left, JSCell* right)
#else
size_t JIT_OPERATION operationCompareStringEq(ExecState* exec, JSCell* left, JSCell* right)
#endif
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    bool result = asString(left)->equal(exec, asString(right));
#if USE(JSVALUE64)
    return JSValue::encode(jsBoolean(result));
#else
    return result;
#endif
}

#if USE(BIGINT32)
size_t JIT_OPERATION operationCompareEqHeapBigIntToInt32(JSGlobalObject* globalObject, JSCell* heapBigInt, int32_t smallInt)
{
    VM& vm = globalObject->vm();
    CallFrame* callFrame = DECLARE_CALL_FRAME(vm);
    JITOperationPrologueCallFrameTracer tracer(vm, callFrame);

    ASSERT(heapBigInt->isHeapBigInt());

    return static_cast<JSBigInt*>(heapBigInt)->equalsToInt32(smallInt);
}
#endif

EncodedJSValue JIT_OPERATION operationNewArrayWithProfile(ExecState* exec, ArrayAllocationProfile* profile, const JSValue* values, int size)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    if (auto* jsArray = constructArrayNegativeIndexed(vm, exec, profile, values, size))
    	return JSValue::encode(jsArray);

	if (!vm.exception())
		throwOutOfMemoryError(vm, exec);
	return encodedJSUndefined();
}

EncodedJSValue JIT_OPERATION operationNewArrayWithSizeAndProfile(ExecState* exec, ArrayAllocationProfile* profile, EncodedJSValue size)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    JSValue sizeValue = JSValue::decode(size);

    if (auto* jsArray = constructArrayWithSizeQuirk(vm, exec, profile, exec->lexicalGlobalObject(), sizeValue))
    	return JSValue::encode(jsArray);

	if (!vm.exception())
		throwOutOfMemoryError(vm, exec);
	return encodedJSUndefined();
}

template<typename FunctionType>
static EncodedJSValue operationNewFunctionCommon(ExecState* exec, JSScope* scope, JSCell* functionExecutable, bool isInvalidated)
{
    ASSERT(functionExecutable->inherits(FunctionExecutable::info()));
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    if (isInvalidated)
        return JSValue::encode(FunctionType::createWithInvalidatedReallocationWatchpoint(vm, static_cast<FunctionExecutable*>(functionExecutable), scope));
    return JSValue::encode(FunctionType::create(vm, static_cast<FunctionExecutable*>(functionExecutable), scope));
}

EncodedJSValue JIT_OPERATION operationNewFunction(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSFunction>(exec, scope, functionExecutable, false);
}

EncodedJSValue JIT_OPERATION operationNewFunctionWithInvalidatedReallocationWatchpoint(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSFunction>(exec, scope, functionExecutable, true);
}

EncodedJSValue JIT_OPERATION operationNewGeneratorFunction(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSGeneratorFunction>(exec, scope, functionExecutable, false);
}

EncodedJSValue JIT_OPERATION operationNewGeneratorFunctionWithInvalidatedReallocationWatchpoint(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSGeneratorFunction>(exec, scope, functionExecutable, true);
}

EncodedJSValue JIT_OPERATION operationNewAsyncFunction(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSAsyncFunction>(exec, scope, functionExecutable, false);
}

EncodedJSValue JIT_OPERATION operationNewAsyncFunctionWithInvalidatedReallocationWatchpoint(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSAsyncFunction>(exec, scope, functionExecutable, true);
}

EncodedJSValue JIT_OPERATION operationNewAsyncGeneratorFunction(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSAsyncGeneratorFunction>(exec, scope, functionExecutable, false);
}

EncodedJSValue JIT_OPERATION operationNewAsyncGeneratorFunctionWithInvalidatedReallocationWatchpoint(ExecState* exec, JSScope* scope, JSCell* functionExecutable)
{
    return operationNewFunctionCommon<JSAsyncGeneratorFunction>(exec, scope, functionExecutable, true);
}

void JIT_OPERATION operationSetFunctionName(ExecState* exec, JSCell* funcCell, EncodedJSValue encodedName)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSFunction* func = jsCast<JSFunction*>(funcCell);
    JSValue name = JSValue::decode(encodedName);
    func->setFunctionName(exec, name);
}

JSCell* JIT_OPERATION operationNewObject(ExecState* exec, Structure* structure)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return constructEmptyObject(vm, structure);
}

JSCell* JIT_OPERATION operationNewRegexp(ExecState* exec, JSCell* regexpPtr)
{
    JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    VM& vm = globalObject->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    RegExp* regexp = static_cast<RegExp*>(regexpPtr);
    ASSERT(regexp->isValid());
    return RegExpObject::create(vm, globalObject->regExpStructure(), regexp);
}

enum class AccessorType {
    Getter,
    Setter
};

static void putAccessorByVal(ExecState* exec, VM& vm, JSObject* base, JSValue subscript, int32_t attribute, JSObject* accessor, AccessorType accessorType)
{
    auto propertyKey = subscript.toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, void());

    if (accessorType == AccessorType::Getter)
        base->putGetter(exec, propertyKey, accessor, attribute);
    else
        base->putSetter(exec, propertyKey, accessor, attribute);
}

void JIT_OPERATION operationPutGetterById(ExecState* exec, JSCell* object, UniquedStringImpl* uid, int32_t options, JSCell* getter)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    ASSERT(object && object->isObject());
    JSObject* baseObj = object->getObject();

    ASSERT(getter->isObject());
    baseObj->putGetter(exec, uid, getter, options);
}

void JIT_OPERATION operationPutSetterById(ExecState* exec, JSCell* object, UniquedStringImpl* uid, int32_t options, JSCell* setter)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    ASSERT(object && object->isObject());
    JSObject* baseObj = object->getObject();

    ASSERT(setter->isObject());
    baseObj->putSetter(exec, uid, setter, options);
}

void JIT_OPERATION operationPutGetterByVal(ExecState* exec, JSCell* base, EncodedJSValue encodedSubscript, int32_t attribute, JSCell* getter)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    putAccessorByVal(exec, vm, asObject(base), JSValue::decode(encodedSubscript), attribute, asObject(getter), AccessorType::Getter);
}

void JIT_OPERATION operationPutSetterByVal(ExecState* exec, JSCell* base, EncodedJSValue encodedSubscript, int32_t attribute, JSCell* setter)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    putAccessorByVal(exec, vm, asObject(base), JSValue::decode(encodedSubscript), attribute, asObject(setter), AccessorType::Setter);
}

#if USE(JSVALUE64)
void JIT_OPERATION operationPutGetterSetter(ExecState* exec, JSCell* object, UniquedStringImpl* uid, int32_t attribute, EncodedJSValue encodedGetterValue, EncodedJSValue encodedSetterValue)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    ASSERT(object && object->isObject());
    JSObject* baseObject = asObject(object);

    JSValue getter = JSValue::decode(encodedGetterValue);
    JSValue setter = JSValue::decode(encodedSetterValue);
    ASSERT(getter.isObject() || setter.isObject());
    GetterSetter* accessor = GetterSetter::create(vm, exec->lexicalGlobalObject(), getter, setter);
    CommonSlowPaths::putDirectAccessorWithReify(vm, exec, baseObject, uid, accessor, attribute);
}

#else
void JIT_OPERATION operationPutGetterSetter(ExecState* exec, JSCell* object, UniquedStringImpl* uid, int32_t attribute, JSCell* getterCell, JSCell* setterCell)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    ASSERT(object && object->isObject());
    JSObject* baseObject = asObject(object);

    ASSERT(getterCell || setterCell);
    JSObject* getter = getterCell ? getterCell->getObject() : 0;
    JSObject* setter = setterCell ? setterCell->getObject() : 0;
    GetterSetter* accessor = GetterSetter::create(vm, exec->lexicalGlobalObject(), getter, setter);
    CommonSlowPaths::putDirectAccessorWithReify(vm, exec, baseObject, uid, accessor, attribute);
}
#endif

EncodedJSValue JIT_OPERATION operationGetArgumentsLength(ExecState* exec, int32_t argumentsRegister)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    // Here we can assume that the argumernts were created. Because otherwise the JIT code would
    // have not made this call.
    Identifier ident = Identifier::fromString(vm, "length");
    JSValue baseValue = exec->uncheckedR(argumentsRegister).jsValue();
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::Get);
    return JSValue::encode(baseValue.get(vm, exec, ident, slot));
}

static ALWAYS_INLINE JSValue getByVal(ExecState* exec, VM& vm, JSValue baseValue, JSValue subscript, ByValInfo* byValInfo, ReturnAddressPtr returnAddress)
{
    if (LIKELY(baseValue.isCell() && subscript.isString())) {
        Structure& structure = *baseValue.asCell()->structure(vm);
        if (JSCell::canUseFastGetOwnProperty(structure)) {
            RefPtr<AtomicStringImpl> existingAtomicString = asString(subscript)->toExistingAtomicString(exec);
            RETURN_IF_EXCEPTION(vm, jsUndefined());
			if (existingAtomicString) {
                if (JSValue result = baseValue.asCell()->fastGetOwnProperty(vm, structure, existingAtomicString.get())) {
                    ASSERT(exec->locationAsBytecodeOffset());
                    if (byValInfo && byValInfo->stubInfo && byValInfo->cachedId.impl() != existingAtomicString)
                        byValInfo->tookSlowPath = true;
                    return result;
                }
            }
        }
    }

    if (Optional<int32_t> index = subscript.tryGetAsInt32()) {
        ASSERT(exec->locationAsBytecodeOffset());
    	if(byValInfo)
    		byValInfo->tookSlowPath = true;

        int32_t i = *index;
        if (isJSString(baseValue)) {
            if (i >= 0 && asString(baseValue)->canGetIndex(i)) {
            	if(byValInfo)
            		ctiPatchCallByReturnAddress(returnAddress, FunctionPtr(operationGetByValString));
				return asString(baseValue)->getIndex(vm, exec, i);
			}
            if(byValInfo)
            	byValInfo->arrayProfile->setOutOfBounds();
        } else if (baseValue.isObject()) {
            JSObject* object = asObject(baseValue);
            if (object->canGetIndexQuickly(static_cast<uint32_t>(i)))
                return object->getIndexQuickly(i);

            bool skipMarkingOutOfBounds = false;

            if (object->structure()->indexingType() == ArrayWithContiguous && i >= 0 && static_cast<uint32_t>(i) < object->butterfly()->publicLength()) {
                // FIXME: expand this to ArrayStorage, Int32, and maybe Double:
                // https://bugs.webkit.org/show_bug.cgi?id=182940
                auto* globalObject = object->globalObject();
                skipMarkingOutOfBounds = globalObject->isOriginalArrayStructure(object->structure()) && globalObject->arrayPrototypeChainIsSane();
            }

            if (byValInfo && !skipMarkingOutOfBounds && !CommonSlowPaths::canAccessArgumentIndexQuickly(*object, i)) {
                // FIXME: This will make us think that in-bounds typed array accesses are actually
                // out-of-bounds.
                // https://bugs.webkit.org/show_bug.cgi?id=149886
            	byValInfo->arrayProfile->setOutOfBounds();
            }
        }

        if (i >= 0)
        	return baseValue.get(vm, exec, static_cast<uint32_t>(i));
    } else if (byValInfo && subscript.isNumber() && baseValue.isCell() && byValInfo->arrayProfile)
    	byValInfo->arrayProfile->setOutOfBounds();

    baseValue.requireObjectCoercible(vm, exec);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    Identifier property = subscript.toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    ASSERT(exec->locationAsBytecodeOffset());
    if (byValInfo && byValInfo->stubInfo && (!isStringOrSymbol(subscript) || byValInfo->cachedId != property))
        byValInfo->tookSlowPath = true;

    return baseValue.get(vm, exec, property);
}

static ALWAYS_INLINE EncodedJSValue getByValWithIndex(ExecState* exec, JSCell* base, uint32_t index)
{
    JSGlobalData& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    if (base->isObject()) {
        JSObject* object = asObject(base);
        if (object->canGetIndexQuickly(index))
            return JSValue::encode(object->getIndexQuickly(index));
    }

    if (isJSString(base) && asString(base)->canGetIndex(index))
        return JSValue::encode(asString(base)->getIndex(vm, exec, index));

    return JSValue::encode(JSValue(base).get(vm, exec, index));
}

static OptimizationResult tryGetByValOptimize(VM& vm, ExecState* exec, JSValue baseValue, JSValue subscript, ByValInfo* byValInfo, ReturnAddressPtr returnAddress)
{
    // See if it's worth optimizing at all.
    OptimizationResult optimizationResult = OptimizationResult::NotOptimized;

    if (baseValue.isObject() && subscript.isInt32()) {
        JSObject* object = asObject(baseValue);

        ASSERT(exec->locationAsBytecodeOffset());
        ASSERT(!byValInfo->stubRoutine);

        if (hasOptimizableIndexing(object->structure())) {
            // Attempt to optimize.
            Structure* structure = object->structure(vm);
            JITArrayMode arrayMode = jitArrayModeForStructure(structure);
            switch (arrayMode) {
        	// FIXME: Optimize BigInt64Array / BigUint64Array in IC
        	// https://bugs.webkit.org/show_bug.cgi?id=221183
        	case JITBigInt64Array:
        	case JITBigUint64Array:
        		optimizationResult = OptimizationResult::GiveUp;
        		break;
        	default:
				if (arrayMode != byValInfo->arrayMode) {
					// If we reached this case, we got an interesting array mode we did not expect when we compiled.
					// Let's update the profile to do better next time.
					CodeBlock* codeBlock = exec->codeBlock();
					// ConcurrentJITLocker locker(codeBlock->m_lock);
					byValInfo->arrayProfile->computeUpdatedPrediction(codeBlock, structure);

					JIT::compileGetByVal(&vm, exec->codeBlock(), byValInfo, returnAddress, arrayMode);
					optimizationResult = OptimizationResult::Optimized;
				}
				break;
            }
        }

        // If we failed to patch and we have some object that intercepts indexed get, then don't even wait until 10 times.
        if (optimizationResult != OptimizationResult::Optimized && object->structure(vm)->typeInfo().interceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero())
            optimizationResult = OptimizationResult::GiveUp;
    }

    if (baseValue.isObject() && isStringOrSymbol(subscript)) {
        const Identifier propertyName = subscript.toPropertyKey(vm, exec);
        RETURN_IF_EXCEPTION(vm, OptimizationResult::GiveUp);
        if (subscript.isSymbol() || !parseIndex(propertyName)) {
            ASSERT(exec->locationAsBytecodeOffset());
            ASSERT(!byValInfo->stubRoutine);
            if (byValInfo->seen) {
                if (byValInfo->cachedId == propertyName) {
                	JIT::compileGetByValWithCachedId(&vm, exec->codeBlock(), byValInfo, returnAddress, propertyName);
                	optimizationResult = OptimizationResult::Optimized;
				} else {
					// Seem like a generic property access site.
					optimizationResult = OptimizationResult::GiveUp;
				}
			} else {
                CodeBlock* codeBlock = exec->codeBlock();
                //ConcurrentJITLocker locker(codeBlock->m_lock);
                byValInfo->seen = true;
                byValInfo->cachedId = propertyName;
                if (subscript.isSymbol())
                    byValInfo->cachedSymbol.set(vm, codeBlock->ownerExecutable(), asSymbol(subscript));
                optimizationResult = OptimizationResult::SeenOnce;
			}
        }
    }

    if (optimizationResult != OptimizationResult::Optimized && optimizationResult != OptimizationResult::SeenOnce) {
        // If we take slow path more than 10 times without patching then make sure we
        // never make that mistake again. For cases where we see non-index-intercepting
        // objects, this gives 10 iterations worth of opportunity for us to observe
        // that the put_by_val may be polymorphic. We count up slowPathCount even if
        // the result is GiveUp.
        if (++byValInfo->slowPathCount >= 10)
            optimizationResult = OptimizationResult::GiveUp;
    }

    return optimizationResult;
}

EncodedJSValue JIT_OPERATION operationGetByValNoIC(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[2].u.operand).jsValue();
    JSValue subscript = exec->r(pc[3].u.operand).jsValue();

    JSValue result = getByVal(exec, vm, base, subscript, nullptr, ReturnAddressPtr());
    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationGetByValGeneric(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedSubscript, ByValInfo* byValInfo)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue subscript = JSValue::decode(encodedSubscript);

    JSValue result = getByVal(exec, vm, baseValue, subscript, byValInfo, ReturnAddressPtr(OUR_RETURN_ADDRESS));
    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationGetByValOptimize(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedSubscript, ByValInfo* byValInfo)
{
    // See if it's worth optimizing at all.
    OptimizationResult optimizationResult = OptimizationResult::NotOptimized;

    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue subscript = JSValue::decode(encodedSubscript);
    ReturnAddressPtr returnAddress = ReturnAddressPtr(OUR_RETURN_ADDRESS);
    OptimizationResult result = tryGetByValOptimize(vm, exec, baseValue, subscript, byValInfo, returnAddress);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	if (result == OptimizationResult::GiveUp) {
        // Don't ever try to optimize.
        byValInfo->tookSlowPath = true;
        ctiPatchCallByReturnAddress(returnAddress, FunctionPtr(operationGetByValGeneric));
    }

    return JSValue::encode(getByVal(exec, vm, baseValue, subscript, byValInfo, returnAddress));
}

EncodedJSValue JIT_OPERATION operationGetByVal(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedProperty)
{
    JSGlobalData& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue property = JSValue::decode(encodedProperty);

    if (LIKELY(baseValue.isCell())) {
        JSCell* base = baseValue.asCell();

		if (Optional<uint32_t> index = property.tryGetAsUint32Index())
			return getByValWithIndex(exec, base, *index);

		if (property.isString()) {
			if(JSCell::canUseFastGetOwnProperty(base->structure())) {
				RefPtr<AtomicStringImpl> existingAtomicString = asString(property)->toExistingAtomicString(exec);
				RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
				if (existingAtomicString) {
					if (JSValue result = base->fastGetOwnProperty(vm, base->structure(), existingAtomicString.get()))
						return JSValue::encode(result);
				}
			}
		}
    }

    baseValue.requireObjectCoercible(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    Identifier propertyName = property.toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(baseValue.get(vm, exec, propertyName));
}

ALWAYS_INLINE static JSValue getPrivateName(VM& vm, ExecState* exec, JSValue baseValue, JSValue fieldNameValue)
{
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSObject* base = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, JSValue());
    auto fieldName = fieldNameValue.toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(scope, JSValue());

    PropertySlot slot(base, PropertySlot::InternalMethodType::GetOwnProperty);
    base->getPrivateField(vm, exec, fieldName, slot);
    RETURN_IF_EXCEPTION(scope, JSValue());

    return slot.getValue(exec, fieldName);
}

ALWAYS_INLINE static JSValue getPrivateName(VM& vm, ExecState* exec, JSValue baseValue, const Identifier& fieldName)
{
    ASSERT(fieldName.isPrivateName());
    UNUSED_PARAM(exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    baseValue.requireObjectCoercible(vm, exec);
    RETURN_IF_EXCEPTION(scope, JSValue());

    JSObject* base = baseValue.toObject(exec);
    RETURN_IF_EXCEPTION(scope, JSValue());

    PropertySlot slot(base, PropertySlot::InternalMethodType::GetOwnProperty);
    base->getPrivateField(vm, exec, fieldName, slot);
    RETURN_IF_EXCEPTION(scope, JSValue());

    return slot.getValue(exec, fieldName);
}

EncodedJSValue JIT_OPERATION operationGetPrivateNameOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedBase, EncodedJSValue encodedFieldName)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue fieldNameValue = JSValue::decode(encodedFieldName);
    ASSERT(CacheableIdentifier::isCacheableIdentifierCell(fieldNameValue));

    if (baseValue.isObject()) {
        const Identifier fieldName = fieldNameValue.toPropertyKey(vm, exec);
        EXCEPTION_ASSERT(!scope.exception());
        ASSERT(fieldName.isSymbol());

        JSObject* base = jsCast<JSObject*>(baseValue.asCell());

        PropertySlot slot(base, PropertySlot::InternalMethodType::GetOwnProperty);
        base->getPrivateField(vm, exec, fieldName, slot);
        RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

        //LOG_IC((ICEvent::OperationGetPrivateNameOptimize, baseValue.classInfoOrNull(vm), fieldName, true));

        if (stubInfo && stubInfo->considerCaching(baseValue.structureOrNull()))
            repatchGetByID(exec, baseValue, fieldName, slot, *stubInfo, GetByIDKind::PrivateName);
        return JSValue::encode(slot.getValue(exec, fieldName));
    }

    scope.release();
    return JSValue::encode(getPrivateName(vm, exec, baseValue, fieldNameValue));
}

EncodedJSValue JIT_OPERATION operationGetPrivateName(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedBase, EncodedJSValue encodedFieldName)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue fieldNameValue = JSValue::decode(encodedFieldName);

    if (stubInfo)
    	stubInfo->tookSlowPath = true;

    return JSValue::encode(getPrivateName(vm, exec, baseValue, fieldNameValue));
}

EncodedJSValue JIT_OPERATION operationGetPrivateNameById(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    if (stubInfo)
    	stubInfo->tookSlowPath = true;

    JSValue baseValue = JSValue::decode(base);
    Identifier fieldName = Identifier::fromUid(vm, uid);

    JSValue result = getPrivateName(vm, exec, baseValue, fieldName);

    //LOG_IC((ICEvent::OperationGetPrivateNameById, baseValue.classInfoOrNull(vm), fieldName, true));

    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationGetPrivateNameByIdOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue baseValue = JSValue::decode(base);
    auto fieldName = Identifier::fromUid(vm, uid);

    if (baseValue.isObject()) {
        JSObject* base = jsCast<JSObject*>(baseValue.asCell());

        PropertySlot slot(base, PropertySlot::InternalMethodType::GetOwnProperty);
        base->getPrivateField(vm, exec, fieldName, slot);
        RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

        //LOG_IC((ICEvent::OperationGetPrivateNameOptimize, baseValue.classInfoOrNull(vm), fieldName, true));

        CodeBlock* codeBlock = exec->codeBlock();
        if (stubInfo && stubInfo->considerCaching(baseValue.structureOrNull()))
            repatchGetByID(exec, baseValue, fieldName, slot, *stubInfo, GetByIDKind::PrivateNameById);
        return JSValue::encode(slot.getValue(exec, fieldName));
    }

    return JSValue::encode(getPrivateName(vm, exec, baseValue, fieldName));
}

EncodedJSValue JIT_OPERATION operationGetPrivateNameByIdGeneric(ExecState* exec, EncodedJSValue base, UniquedStringImpl* uid)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue baseValue = JSValue::decode(base);
    Identifier fieldName = Identifier::fromUid(vm, uid);

    JSValue result = getPrivateName(vm, exec, baseValue, fieldName);

    //LOG_IC((ICEvent::OperationGetPrivateNameByIdGeneric, baseValue.classInfoOrNull(vm), fieldName, true));

    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationHasIndexedPropertyDefault(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedSubscript, ByValInfo* byValInfo)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue subscript = JSValue::decode(encodedSubscript);

    ASSERT(baseValue.isObject());
    ASSERT(subscript.isUInt32AsAnyInt());

    JSObject* object = asObject(baseValue);
    bool didOptimize = false;

    ASSERT(exec->locationAsBytecodeOffset());
    ASSERT(!byValInfo->stubRoutine);

    if (hasOptimizableIndexing(object->structure(vm))) {
        // Attempt to optimize.
        JITArrayMode arrayMode = jitArrayModeForStructure(object->structure(vm));
        if (arrayMode != byValInfo->arrayMode) {
            JIT::compileHasIndexedProperty(&vm, exec->codeBlock(), byValInfo, ReturnAddressPtr(OUR_RETURN_ADDRESS), arrayMode);
            didOptimize = true;
        }
    }

    if (!didOptimize) {
        // If we take slow path more than 10 times without patching then make sure we
        // never make that mistake again. Or, if we failed to patch and we have some object
        // that intercepts indexed get, then don't even wait until 10 times. For cases
        // where we see non-index-intercepting objects, this gives 10 iterations worth of
        // opportunity for us to observe that the get_by_val may be polymorphic.
        if (++byValInfo->slowPathCount >= 10
            || object->structure(vm)->typeInfo().interceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero()) {
            // Don't ever try to optimize.
            ctiPatchCallByReturnAddress(ReturnAddressPtr(OUR_RETURN_ADDRESS), FunctionPtr(operationHasIndexedPropertyGeneric));
        }
    }

    uint32_t index = subscript.asUInt32AsAnyInt();
    if (object->canGetIndexQuickly(index))
        return JSValue::encode(JSValue(JSValue::JSTrue));

    if (!CommonSlowPaths::canAccessArgumentIndexQuickly(*object, index))
    	byValInfo->arrayProfile->setOutOfBounds();
    return JSValue::encode(jsBoolean(object->hasEnumerableProperty(vm, exec, index)));
}

EncodedJSValue JIT_OPERATION operationHasIndexedPropertyGeneric(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedSubscript, ByValInfo* byValInfo)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue subscript = JSValue::decode(encodedSubscript);

    ASSERT(baseValue.isObject());
    ASSERT(subscript.isUInt32AsAnyInt());

    JSObject* object = asObject(baseValue);
    uint32_t index = subscript.asUInt32AsAnyInt();
    if (object->canGetIndexQuickly(index))
        return JSValue::encode(JSValue(JSValue::JSTrue));

    if (!CommonSlowPaths::canAccessArgumentIndexQuickly(*object, index))
    	byValInfo->arrayProfile->setOutOfBounds();
    return JSValue::encode(jsBoolean(object->hasEnumerableProperty(vm, exec, index)));
}

EncodedJSValue JIT_OPERATION operationHasIndexedPropertyNoIC(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue base = exec->r(pc[2].u.operand).jsValue();
    JSValue subscript = exec->r(pc[3].u.operand).jsValue();

    ASSERT(base.isObject());
    ASSERT(subscript.isUInt32AsAnyInt());

    JSObject* object = asObject(base);
    uint32_t index = subscript.asUInt32AsAnyInt();
    if (object->canGetIndexQuickly(index))
        return JSValue::encode(JSValue(JSValue::JSTrue));

    return JSValue::encode(jsBoolean(object->hasEnumerableProperty(vm, exec, index)));
}

EncodedJSValue JIT_OPERATION operationGetByValString(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedSubscript, ByValInfo* byValInfo)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue baseValue = JSValue::decode(encodedBase);
    JSValue subscript = JSValue::decode(encodedSubscript);

    JSValue result;
    if (LIKELY(subscript.isUInt32())) {
        uint32_t i = subscript.asUInt32();
        if (isJSString(baseValue) && asString(baseValue)->canGetIndex(i))
            result = asString(baseValue)->getIndex(vm, exec, i);
        else {
            result = baseValue.get(vm, exec, i);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            if (!isJSString(baseValue)) {
                ASSERT(exec->locationAsBytecodeOffset());
				ctiPatchCallByReturnAddress(ReturnAddressPtr(OUR_RETURN_ADDRESS), FunctionPtr(byValInfo->stubRoutine ? operationGetByValGeneric : operationGetByValOptimize));
			}
        }
    } else {
        baseValue.requireObjectCoercible(vm, exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        Identifier property = subscript.toPropertyKey(vm, exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        result = baseValue.get(vm, exec, property);
    }

    return JSValue::encode(result);
}

void JIT_OPERATION operationDebug(ExecState* exec, int32_t debugHookID)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    vm.interpreter->debug(exec, static_cast<DebugHookID>(debugHookID));
}

#if ENABLE(DFG_JIT)
char* JIT_OPERATION operationOptimize(ExecState* exec, int32_t bytecodeIndex)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    // Defer GC for a while so that it doesn't run between when we enter into this
    // slow path and when we figure out the state of our code block. This prevents
    // a number of awkward reentrancy scenarios, including:
    //
    // - The optimized version of our code block being jettisoned by GC right after
    //   we concluded that we wanted to use it, but have not planted it into the JS
    //   stack yet.
    //
    // - An optimized version of our code block being installed just as we decided
    //   that it wasn't ready yet.
    //
    // Note that jettisoning won't happen if we already initiated OSR, because in
    // that case we would have already planted the optimized code block into the JS
    // stack.
    DeferGCForAWhile deferGC(vm.heap);

    CodeBlock* codeBlock = exec->codeBlock();
    if (codeBlock->jitType() != JITCode::BaselineJIT) {
        //dataLog("Unexpected code block in Baseline->DFG tier-up: ", *codeBlock, "\n");
        RELEASE_ASSERT_NOT_REACHED();
        return 0;
    }

    if (bytecodeIndex) {
        // If we're attempting to OSR from a loop, assume that this should be
        // separately optimized.
        codeBlock->m_shouldAlwaysBeInlined = false;
    }

#if ENABLE(JIT_VERBOSE_OSR)
    dataLog("%p: Entered optimize with bytecodeIndex = %u, executeCounter = %s, reoptimizationRetryCounter = %u, optimizationDelayCounter = %u, exitCounter = ", codeBlock, bytecodeIndex, codeBlock->jitExecuteCounter().status(), codeBlock->reoptimizationRetryCounter(), codeBlock->optimizationDelayCounter());
    if (codeBlock->hasOptimizedReplacement())
        dataLog("%u", codeBlock->replacement()->osrExitCounter());
    else
        dataLog("N/A");
    dataLog("\n");
#endif

    if (!codeBlock->checkIfOptimizationThresholdReached()) {
        codeBlock->updateAllPredictions();
        return 0;
    }

    if (codeBlock->m_shouldAlwaysBeInlined) {
        codeBlock->updateAllPredictions();
        codeBlock->optimizeAfterWarmUp();

        // should optimize next time
        codeBlock->m_shouldAlwaysBeInlined = false;

#if ENABLE(JIT_VERBOSE_OSR)
        if (Options::verboseOSR())
            dataLog("Choosing not to optimize ", *codeBlock, " yet, because m_shouldAlwaysBeInlined == true.\n");
#endif
        return 0;
    }

    if (codeBlock->hasOptimizedReplacement()) {
#if ENABLE(JIT_VERBOSE_OSR)
        dataLog("Considering OSR into %p(%p).\n", codeBlock, codeBlock->replacement());
#endif
        // If we have an optimized replacement, then it must be the case that we entered
        // cti_optimize from a loop. That's because if there's an optimized replacement,
        // then all calls to this function will be relinked to the replacement and so
        // the prologue OSR will never fire.

        // This is an interesting threshold check. Consider that a function OSR exits
        // in the middle of a loop, while having a relatively low exit count. The exit
        // will reset the execution counter to some target threshold, meaning that this
        // code won't be reached until that loop heats up for >=1000 executions. But then
        // we do a second check here, to see if we should either reoptimize, or just
        // attempt OSR entry. Hence it might even be correct for
        // shouldReoptimizeFromLoopNow() to always return true. But we make it do some
        // additional checking anyway, to reduce the amount of recompilation thrashing.
        if (codeBlock->replacement()->shouldReoptimizeFromLoopNow()) {
#if ENABLE(JIT_VERBOSE_OSR)
            dataLog("Triggering reoptimization of %p(%p) (in loop).\n", codeBlock, codeBlock->replacement());
#endif
            codeBlock->replacement()->jettison(CountReoptimization);
            return nullptr;
        }
    } else {
        if (!codeBlock->shouldOptimizeNow()) {
#if ENABLE(JIT_VERBOSE_OSR)
            dataLog("Delaying optimization for %p (in loop) because of insufficient profiling.\n", codeBlock);
#endif
            return nullptr;
        }

        if(WTF::platformRunState() != WTF::Platform_State_Running) {
    		codeBlock->dontOptimizeAnytimeSoon();
        	return nullptr;
        }

    	if(ExecutableAllocator::underMemoryPressure() || WTF::isLowMemory() || WTF::isUnderMemoryPressure()) {
    		codeBlock->dontOptimizeAnytimeSoon();
    		return nullptr;
    	}

		if(codeBlock->isConstructor()) {
			if(!codeBlock->ownerExecutable()->canDFGCompileConstruct()) {
	    		codeBlock->dontOptimizeAnytimeSoon();
	    		return nullptr;
			}
		} else {
			if(!codeBlock->ownerExecutable()->canDFGCompileCall()) {
	    		codeBlock->dontOptimizeAnytimeSoon();
	    		return nullptr;
			}
		}

        RefPtr<CodeBlock> newCodeBlock = codeBlock->newReplacement();
        double startTime = WTF::monotonicallyIncreasingTime();
        CompilationResult result = DFG::compile(exec, newCodeBlock.get(), bytecodeIndex);
        double endTime = WTF::monotonicallyIncreasingTime();

        if((endTime - startTime) > Options::jitCodesOptimizationTimeLimit()) {
    		codeBlock->dontOptimizeAnytimeSoon();

    		if(codeBlock->isConstructor())
    			codeBlock->ownerExecutable()->disableDFGCompileConstruct();
    		else
    			codeBlock->ownerExecutable()->disableDFGCompileCall();
        }

        if (result == CompilationSuccessful)
        	newCodeBlock->ownerExecutable()->installCode(newCodeBlock.get());

        if (result != CompilationSuccessful){
    		codeBlock->dontOptimizeAnytimeSoon();
            return nullptr;
        }
    }

    CodeBlock* optimizedCodeBlock = codeBlock->replacement();
    ASSERT(JITCode::isOptimizingJIT(optimizedCodeBlock->getJITType()));

	if (optimizedCodeBlock->getJITType() == JITCode::FTLJIT) {
		// FTL JIT doesn't support OSR entry yet.
		// https://bugs.webkit.org/show_bug.cgi?id=113625

		// Don't attempt OSR entry again.
		codeBlock->dontOptimizeAnytimeSoon();
		return nullptr;
	}

    if (void* address = DFG::prepareOSREntry(exec, optimizedCodeBlock, bytecodeIndex)) {
    	if (Options::showDFGDisassembly()) {
			//dataLog(
			//	"Performing OSR from code block %p to code block %p, address %p to %p.\n",
			//	codeBlock, optimizedCodeBlock, (STUB_RETURN_ADDRESS).value(), address);
		}
#if ENABLE(JIT_VERBOSE_OSR)
        dataLog("Optimizing %p succeeded, performing OSR after a delay of %u.\n", codeBlock, codeBlock->optimizationDelayCounter());
#endif

        codeBlock->optimizeSoon();
        return static_cast<char*>(address);
    }

#if ENABLE(JIT_VERBOSE_OSR)
    dataLog("Optimizing %p succeeded, OSR failed, after a delay of %u.\n", codeBlock, codeBlock->optimizationDelayCounter());
#endif

    // Count the OSR failure as a speculation failure. If this happens a lot, then
    // reoptimize.
    optimizedCodeBlock->countOSRExit();

#if ENABLE(JIT_VERBOSE_OSR)
    dataLog("Encountered OSR failure into %p(%p).\n", codeBlock, codeBlock->replacement());
#endif

    // We are a lot more conservative about triggering reoptimization after OSR failure than
    // before it. If we enter the optimize_from_loop trigger with a bucket full of fail
    // already, then we really would like to reoptimize immediately. But this case covers
    // something else: there weren't many (or any) speculation failures before, but we just
    // failed to enter the speculative code because some variable had the wrong value or
    // because the OSR code decided for any spurious reason that it did not want to OSR
    // right now. So, we only trigger reoptimization only upon the more conservative (non-loop)
    // reoptimization trigger.
    if (optimizedCodeBlock->shouldReoptimizeNow()) {
#if ENABLE(JIT_VERBOSE_OSR)
        dataLog("Triggering reoptimization of %p(%p) (after OSR fail).\n", codeBlock, codeBlock->replacement());
#endif
        optimizedCodeBlock->jettison(CountReoptimization);
        return 0;
    }

    // OSR failed this time, but it might succeed next time! Let the code run a bit
    // longer and then try again.
    codeBlock->optimizeAfterWarmUp();

    return 0;
}
#endif

void JIT_OPERATION operationPopScope(ExecState* exec, int32_t scopeReg)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSScope* scope = exec->uncheckedR(scopeReg).Register::scope();
    exec->uncheckedR(scopeReg) = scope->next();
}

void JIT_OPERATION operationProfileDidCall(ExecState* exec, EncodedJSValue encodedValue)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->didExecute(exec, JSValue::decode(encodedValue));
}

void JIT_OPERATION operationProfileWillCall(ExecState* exec, EncodedJSValue encodedValue)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    if (LegacyProfiler* profiler = vm.enabledProfiler())
        profiler->willExecute(exec, JSValue::decode(encodedValue));
}

int32_t JIT_OPERATION operationInstanceOfCustom(ExecState* exec, EncodedJSValue encodedValue, JSObject* constructor, EncodedJSValue encodedHasInstance)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue value = JSValue::decode(encodedValue);
    JSValue hasInstanceValue = JSValue::decode(encodedHasInstance);

    ASSERT(hasInstanceValue != exec->lexicalGlobalObject()->functionProtoHasInstanceSymbolFunction() || !constructor->structure()->typeInfo().implementsDefaultHasInstance());

    if (constructor->hasInstance(exec, value, hasInstanceValue))
        return 1;
    return 0;
}

EncodedJSValue JIT_OPERATION operationInstanceOfCustomGeneric(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue value = exec->r(pc[2].u.operand).jsValue();
    JSValue constructorValue = exec->r(pc[3].u.operand).jsValue();
    JSValue hasInstanceValue = exec->r(pc[4].u.operand).jsValue();

    JSObject* constructor = asObject(constructorValue);

    ASSERT(hasInstanceValue != exec->lexicalGlobalObject()->functionProtoHasInstanceSymbolFunction() || !constructor->structure()->typeInfo().implementsDefaultHasInstance());

    return JSValue::encode(jsBoolean(constructor->hasInstance(exec, value, hasInstanceValue)));
}

static size_t JIT_OPERATION operationDeleteById(ExecState* exec, EncodedJSValue encodedBase, UniquedStringImpl* uid, ECMAMode ecmaMode)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSObject* baseObj = JSValue::decode(encodedBase).toObject(exec);
    RETURN_IF_EXCEPTION(vm, false);
    if (!baseObj)
        return false;
    bool couldDelete = baseObj->methodTable()->deleteProperty(baseObj, exec, Identifier::fromUid(vm, uid));
    RETURN_IF_EXCEPTION(vm, false);
    if (!couldDelete && ecmaMode.isStrict())
        vm.throwException(exec, createTypeError(exec, "Unable to delete property."));
    return couldDelete;
}

size_t JIT_OPERATION operationDeleteByIdSloppy(ExecState* exec, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
	return operationDeleteById(exec, encodedBase, uid, ECMAMode::sloppy());
}

size_t JIT_OPERATION operationDeleteByIdStrict(ExecState* exec, EncodedJSValue encodedBase, UniquedStringImpl* uid)
{
	return operationDeleteById(exec, encodedBase, uid, ECMAMode::strict());
}

EncodedJSValue JIT_OPERATION operationDeleteByIdJSResultSloppy(ExecState* exec, EncodedJSValue base, UniquedStringImpl* uid)
{
    return JSValue::encode(jsBoolean(operationDeleteById(exec, base, uid, ECMAMode::sloppy())));
}

EncodedJSValue JIT_OPERATION operationDeleteByIdJSResultStrict(ExecState* exec, EncodedJSValue base, UniquedStringImpl* uid)
{
    return JSValue::encode(jsBoolean(operationDeleteById(exec, base, uid, ECMAMode::strict())));
}

static size_t JIT_OPERATION operationDeleteByVal(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedKey, ECMAMode ecmaMode)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSObject* baseObj = JSValue::decode(encodedBase).toObject(exec);
    RETURN_IF_EXCEPTION(vm, false);
    JSValue key = JSValue::decode(encodedKey);
    if (!baseObj)
        return false;

    bool couldDelete;
    uint32_t index;
    if (key.getUInt32(index))
        couldDelete = baseObj->methodTable()->deletePropertyByIndex(baseObj, exec, index);
    else {
        Identifier property = key.toPropertyKey(vm, exec);
        RETURN_IF_EXCEPTION(vm, false);
        couldDelete = baseObj->methodTable()->deleteProperty(baseObj, exec, property);
    }
    RETURN_IF_EXCEPTION(vm, false);
    if (!couldDelete && ecmaMode.isStrict())
        vm.throwException(exec, createTypeError(exec, ASCIILiteral("Unable to delete property.")));
    return couldDelete;
}

size_t JIT_OPERATION operationDeleteByValSloppy(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedKey)
{
	return operationDeleteByVal(exec, encodedBase, encodedKey, ECMAMode::sloppy());
}

size_t JIT_OPERATION operationDeleteByValStrict(ExecState* exec, EncodedJSValue encodedBase, EncodedJSValue encodedKey)
{
	return operationDeleteByVal(exec, encodedBase, encodedKey, ECMAMode::strict());
}

EncodedJSValue JIT_OPERATION operationDeleteByValJSResultSloppy(ExecState* exec, EncodedJSValue base,  EncodedJSValue key)
{
    return JSValue::encode(jsBoolean(operationDeleteByVal(exec, base, key, ECMAMode::sloppy())));
}

EncodedJSValue JIT_OPERATION operationDeleteByValJSResultStrict(ExecState* exec, EncodedJSValue base,  EncodedJSValue key)
{
    return JSValue::encode(jsBoolean(operationDeleteByVal(exec, base, key, ECMAMode::strict())));
}

JSCell* JIT_OPERATION operationPushWithScope(ExecState* exec, JSCell* currentScopeCell, EncodedJSValue objectValue)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    ThrowScope scope(vm);

    JSObject* object = JSValue::decode(objectValue).toObject(exec);
    RETURN_IF_EXCEPTION(scope, 0);

    JSScope* currentScope = jsCast<JSScope*>(currentScopeCell);

    return JSWithScope::create(vm, exec->lexicalGlobalObject(), currentScope, object);
}

JSCell* JIT_OPERATION operationPushWithScopeObject(ExecState* exec, JSCell* currentScopeCell, JSObject* object)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSScope* currentScope = jsCast<JSScope*>(currentScopeCell);
    return JSWithScope::create(vm, exec->lexicalGlobalObject(), currentScope, object);
}

EncodedJSValue JIT_OPERATION operationInstanceOfNoIC(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    JSValue value = exec->r(pc[2].u.operand).jsValue();
    JSValue proto = exec->r(pc[3].u.operand).jsValue();

    bool result = JSObject::defaultHasInstance(exec, value, proto);
    return JSValue::encode(jsBoolean(result));
}

EncodedJSValue JIT_OPERATION operationInstanceOfGeneric(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedProto)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue value = JSValue::decode(encodedValue);
    JSValue proto = JSValue::decode(encodedProto);

    stubInfo->tookSlowPath = true;

    bool result = JSObject::defaultHasInstance(exec, value, proto);
    return JSValue::encode(jsBoolean(result));
}

EncodedJSValue JIT_OPERATION operationInstanceOfOptimize(ExecState* exec, StructureStubInfo* stubInfo, EncodedJSValue encodedValue, EncodedJSValue encodedProto)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    //auto scope = DECLARE_THROW_SCOPE(vm);
    JSValue value = JSValue::decode(encodedValue);
    JSValue proto = JSValue::decode(encodedProto);

    bool result = JSObject::defaultHasInstance(exec, value, proto);
    RETURN_IF_EXCEPTION(vm, JSValue::encode(jsUndefined()));

    if (stubInfo->considerCaching(value.structureOrNull()))
        repatchInstanceOf(exec, value, proto, *stubInfo, result);

    return JSValue::encode(jsBoolean(result));
}

CallFrame* JIT_OPERATION operationSizeAndAllocFrameForVarargs(ExecState* exec, EncodedJSValue encodedArguments, int32_t firstFreeRegister, int32_t firstVarArgOffset)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSStack* stack = &vm.interpreter->stack();
    JSValue arguments = JSValue::decode(encodedArguments);
    CallFrame* newCallFrame = sizeAndAllocFrameForVarargs(exec, stack, arguments, firstFreeRegister, firstVarArgOffset, vm.varargsLength);
    return newCallFrame;
}

CallFrame* JIT_OPERATION operationLoadVarargs(ExecState* exec, CallFrame* newCallFrame, EncodedJSValue encodedThis, EncodedJSValue encodedArguments, int32_t firstVarArgOffset)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue thisValue = JSValue::decode(encodedThis);
    JSValue arguments = JSValue::decode(encodedArguments);
    loadVarargs(exec, newCallFrame, thisValue, arguments, firstVarArgOffset, vm.varargsLength);
    return newCallFrame;
}

char* JIT_OPERATION operationSwitchCharWithUnknownKeyType(ExecState* exec, EncodedJSValue encodedKey, size_t tableIndex)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    //auto throwScope = DECLARE_THROW_SCOPE(vm);
    JSValue key = JSValue::decode(encodedKey);
    CodeBlock* codeBlock = exec->codeBlock();

    SimpleJumpTable& jumpTable = codeBlock->switchJumpTable(tableIndex);
    void* result = jumpTable.ctiDefault.executableAddress();

    if (key.isString()) {
        JSString* string = asString(key);
        if (string->length() == 1) {
            String value = string->value(exec);
            RETURN_IF_EXCEPTION(vm, 0);
            result = jumpTable.ctiForValue(value[0]).executableAddress();
        }
    }

    return reinterpret_cast<char*>(result);
}

char* JIT_OPERATION operationSwitchImmWithUnknownKeyType(ExecState* exec, EncodedJSValue encodedKey, size_t tableIndex)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue key = JSValue::decode(encodedKey);
    CodeBlock* codeBlock = exec->codeBlock();

    SimpleJumpTable& jumpTable = codeBlock->switchJumpTable(tableIndex);
    void* result;
    if (key.isInt32())
        result = jumpTable.ctiForValue(key.asInt32()).executableAddress();
    else if (key.isDouble() && key.asDouble() == static_cast<int32_t>(key.asDouble()))
        result = jumpTable.ctiForValue(static_cast<int32_t>(key.asDouble())).executableAddress();
    else
        result = jumpTable.ctiDefault.executableAddress();
    return reinterpret_cast<char*>(result);
}

char* JIT_OPERATION operationSwitchStringWithUnknownKeyType(ExecState* exec, EncodedJSValue encodedKey, size_t tableIndex)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSValue key = JSValue::decode(encodedKey);
    CodeBlock* codeBlock = exec->codeBlock();

    void* result;
    StringJumpTable& jumpTable = codeBlock->stringSwitchJumpTable(tableIndex);

    if (key.isString()) {
        auto* value = asString(key)->value(exec).impl();

        RETURN_IF_EXCEPTION(vm, nullptr);

        result = jumpTable.ctiForValue(value).executableAddress();
    } else
        result = jumpTable.ctiDefault.executableAddress();

    return reinterpret_cast<char*>(result);
}

EncodedJSValue JIT_OPERATION operationGetFromScope(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    CodeBlock* codeBlock = exec->codeBlock();
    Instruction* pc = bytecodePC;

    const Identifier& ident = codeBlock->identifier(pc[3].u.operand);
    JSObject* scope = jsCast<JSObject*>(exec->uncheckedR(pc[2].u.operand).jsValue());
    GetPutInfo getPutInfo(pc[4].u.operand);

    // ModuleVar is always converted to ClosureVar for get_from_scope.
    ASSERT(getPutInfo.resolveType() != ModuleVar);

    PropertySlot slot(scope, PropertySlot::InternalMethodType::Get);
    bool found = scope->getPropertySlot(vm, exec, ident, slot);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (!found) {
        if (getPutInfo.resolveMode() == ThrowIfNotFound)
            vm.throwException(exec, createUndefinedVariableError(exec, ident));
        return JSValue::encode(jsUndefined());
    }

    JSValue result = JSValue();
    if (scope->isGlobalLexicalEnvironment()) {
        // When we can't statically prove we need a TDZ check, we must perform the check on the slow path.
        result = slot.getValue(exec, ident);
        if (UNLIKELY(result == jsTDZValue())) {
            vm.throwException(exec, createTDZError(exec));
            return JSValue::encode(jsUndefined());
        }
    }

    CommonSlowPaths::tryCacheGetFromScopeGlobal(exec, vm, codeBlock, pc, scope, slot, ident);

    if (!result)
        result = slot.getValue(exec, ident);
    return JSValue::encode(result);
}

void JIT_OPERATION operationPutToScope(ExecState* exec, Instruction* bytecodePC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    Instruction* pc = bytecodePC;

    CodeBlock* codeBlock = exec->codeBlock();
    const Identifier& ident = codeBlock->identifier(pc[2].u.operand);
    JSObject* scope = jsCast<JSObject*>(exec->uncheckedR(pc[1].u.operand).jsValue());
    JSValue value = exec->r(pc[3].u.operand).jsValue();
    GetPutInfo getPutInfo = GetPutInfo(pc[4].u.operand);

    // ModuleVar does not keep the scope register value alive in DFG.
    ASSERT(getPutInfo.resolveType() != ModuleVar);

	if (getPutInfo.resolveType() == ResolvedClosureVar) {
        JSLexicalEnvironment* environment = jsCast<JSLexicalEnvironment*>(scope);
		environment->variableAt(ScopeOffset(pc[6].u.operand)).set(vm, environment, value);
		if (WatchpointSet* set = pc[5].u.watchpointSet)
			set->touch(vm, "Executed op_put_scope<ResolvedClosureVar>");
		return;
    }

    bool hasProperty = scope->hasProperty(exec, ident);
    RETURN_IF_EXCEPTION(vm, void());
    if (hasProperty
		&& scope->isGlobalLexicalEnvironment()
        && !isInitialization(getPutInfo.initializationMode())) {
        // When we can't statically prove we need a TDZ check, we must perform the check on the slow path.
        PropertySlot slot(scope, PropertySlot::InternalMethodType::Get);
        JSGlobalLexicalEnvironment::getOwnPropertySlot(scope, exec, ident, slot);
        if (UNLIKELY(slot.getValue(exec, ident) == jsTDZValue())) {
            vm.throwException(exec, createTDZError(exec));
            return;
        }
    }

    if (getPutInfo.resolveMode() == ThrowIfNotFound && !hasProperty) {
        vm.throwException(exec, createUndefinedVariableError(exec, ident));
        return;
    }

    PutPropertySlot slot(scope, getPutInfo.ecmaMode().isStrict(), PutPropertySlot::UnknownContext, isInitialization(getPutInfo.initializationMode()));
    scope->putInline(exec, vm, ident, value, slot);

    RETURN_IF_EXCEPTION(vm, void());

    CommonSlowPaths::tryCachePutToScopeGlobal(exec, vm, codeBlock, pc, scope, getPutInfo, slot, ident);
}

void JIT_OPERATION operationThrow(ExecState* exec, EncodedJSValue encodedExceptionValue)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue exceptionValue = JSValue::decode(encodedExceptionValue);
    vm->throwException(exec, exceptionValue);

    // Results stored out-of-band in vm.targetMachinePCForThrow & vm.callFrameForCatch
    genericUnwind(vm, exec);
}

char* JIT_OPERATION operationReallocateButterflyToHavePropertyStorageWithInitialCapacity(ExecState* exec, JSObject* object)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    ASSERT(!object->structure()->outOfLineCapacity());
    DeferGC deferGC(vm.heap);
    Butterfly* result = object->allocateMoreOutOfLineStorage(vm, deferGC, 0, initialOutOfLineCapacity);
    object->setButterfly(vm, result);
    return reinterpret_cast<char*>(result);
}

char* JIT_OPERATION operationReallocateButterflyToGrowPropertyStorage(ExecState* exec, JSObject* object, size_t newSize)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    DeferGC deferGC(vm.heap);
    Butterfly* result = object->allocateMoreOutOfLineStorage(vm, deferGC, object->structure()->outOfLineCapacity(), newSize);
    object->setButterfly(vm, result);
    return reinterpret_cast<char*>(result);
}

void JIT_OPERATION operationHandleWatchdogTimer(ExecState* exec)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    if (UNLIKELY(vm.watchdog->didFire(exec)))
        vm.throwException(exec, createTerminatedExecutionException(vm));
}

void JIT_OPERATION lookupExceptionHandler(VM* vm, ExecState* exec)
{
    NativeCallFrameTracer tracer(vm, exec, NativeCallFrameTracer::VMEntrySentinelOK);
    genericUnwind(vm, exec);
    ASSERT(vm->targetMachinePCForThrow);
}

void JIT_OPERATION operationVMHandleException(ExecState* exec)
{
    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    ASSERT(!exec->isVMEntrySentinel());
    genericUnwind(vm, exec);
}

// This function "should" just take the ExecState*, but doing so would make it more difficult
// to call from exception check sites. So, unlike all of our other functions, we allow
// ourselves to play some gnarly ABI tricks just to simplify the calling convention. This is
// particularly safe here since this is never called on the critical path - it's only for
// testing.
void JIT_OPERATION operationExceptionFuzz(ExecState* exec)
{
    ASSERT(Options::enableExceptionFuzz());

    VM* vm = &exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    // This probably "just works" for GCC also, but I haven't tried.
}

JSCell* JIT_OPERATION operationGetPropertyEnumeratorCell(ExecState* exec, JSCell* cell)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSObject* base = cell->toObject(exec, exec->lexicalGlobalObject());
    RETURN_IF_EXCEPTION(vm, nullptr);

    return propertyNameEnumerator(exec, base);
}

JSCell* JIT_OPERATION operationGetPropertyEnumerator(ExecState* exec, EncodedJSValue encodedBase)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue base = JSValue::decode(encodedBase);
    if (base.isUndefinedOrNull())
        return JSPropertyNameEnumerator::create(vm);

    JSObject* baseObject = base.toObject(exec);
    RETURN_IF_EXCEPTION(vm, nullptr);

    return propertyNameEnumerator(exec, baseObject);
}

EncodedJSValue JIT_OPERATION operationNextEnumeratorPname(ExecState* exec, JSCell* enumeratorCell, int32_t index)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    JSPropertyNameEnumerator* enumerator = jsCast<JSPropertyNameEnumerator*>(enumeratorCell);
    JSString* propertyName = enumerator->propertyNameAtIndex(index);
    return JSValue::encode(propertyName ? propertyName : jsNull());
}

JSCell* JIT_OPERATION operationToIndexString(ExecState* exec, int32_t index)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);
    return jsString(vm, Identifier::from(vm, index).string());
}

ALWAYS_INLINE static EncodedJSValue unprofiledAdd(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    return JSValue::encode(jsAdd(vm, exec, op1, op2));
}

ALWAYS_INLINE static EncodedJSValue profiledAdd(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, ArithProfile& arithProfile)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    arithProfile.observeLHSAndRHS(op1, op2);

    JSValue result = jsAdd(vm, exec, op1, op2);
    arithProfile.observeResult(result);

    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationValueAdd(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    return unprofiledAdd(exec, encodedOp1, encodedOp2);
}

EncodedJSValue JIT_OPERATION operationValueAddProfiled(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, ArithProfile* arithProfile)
{
    ASSERT(arithProfile);
    return profiledAdd(exec, encodedOp1, encodedOp2, *arithProfile);
}

EncodedJSValue JIT_OPERATION operationValueAddProfiledOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITAddIC* addIC)
{
    VM&vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    ArithProfile* arithProfile = addIC->arithProfile();
    ASSERT(arithProfile);
    arithProfile->observeLHSAndRHS(op1, op2);
    auto nonOptimizeVariant = operationValueAddProfiledNoOptimize;
    addIC->generateOutOfLine(vm, exec->codeBlock(), nonOptimizeVariant);

    JSValue result = jsAdd(vm, exec, op1, op2);
    arithProfile->observeResult(result);

    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationValueAddProfiledNoOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITAddIC* addIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    ArithProfile* arithProfile = addIC->arithProfile();
    ASSERT(arithProfile);
    return profiledAdd(exec, encodedOp1, encodedOp2, *arithProfile);
}

EncodedJSValue JIT_OPERATION operationValueAddOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITAddIC* addIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    auto nonOptimizeVariant = operationValueAddNoOptimize;
    if (ArithProfile* arithProfile = addIC->arithProfile())
        arithProfile->observeLHSAndRHS(op1, op2);
    addIC->generateOutOfLine(vm, exec->codeBlock(), nonOptimizeVariant);

#if ENABLE(MATH_IC_STATS)
    exec->codeBlock()->dumpMathICStats();
#endif

    return JSValue::encode(jsAdd(vm, exec, op1, op2));
}

EncodedJSValue JIT_OPERATION operationValueAddNoOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITAddIC*)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    JSValue result = jsAdd(vm, exec, op1, op2);

    return JSValue::encode(result);
}

ALWAYS_INLINE static EncodedJSValue unprofiledMul(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    return JSValue::encode(jsMul(exec, op1, op2));
}

ALWAYS_INLINE static EncodedJSValue profiledMul(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, ArithProfile& arithProfile, bool shouldObserveLHSAndRHSTypes = true)
{
    VM& vm = exec->vm();
    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    if (shouldObserveLHSAndRHSTypes)
        arithProfile.observeLHSAndRHS(op1, op2);

    JSValue result = jsMul(exec, op1, op2);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    arithProfile.observeResult(result);
    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationValueMul(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return unprofiledMul(exec, encodedOp1, encodedOp2);
}

EncodedJSValue JIT_OPERATION operationValueMulNoOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITMulIC*)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return unprofiledMul(exec, encodedOp1, encodedOp2);
}

EncodedJSValue JIT_OPERATION operationValueMulOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITMulIC* mulIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    auto nonOptimizeVariant = operationValueMulNoOptimize;
    if (ArithProfile* arithProfile = mulIC->arithProfile())
		arithProfile->observeLHSAndRHS(JSValue::decode(encodedOp1), JSValue::decode(encodedOp2));
	mulIC->generateOutOfLine(vm, exec->codeBlock(), nonOptimizeVariant);

#if ENABLE(MATH_IC_STATS)
    exec->codeBlock()->dumpMathICStats();
#endif

    return unprofiledMul(exec, encodedOp1, encodedOp2);
}

EncodedJSValue JIT_OPERATION operationValueMulProfiled(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, ArithProfile* arithProfile)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    ASSERT(arithProfile);
    return profiledMul(exec, encodedOp1, encodedOp2, *arithProfile);
}

EncodedJSValue JIT_OPERATION operationValueMulProfiledOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITMulIC* mulIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    ArithProfile* arithProfile = mulIC->arithProfile();
    ASSERT(arithProfile);
    arithProfile->observeLHSAndRHS(JSValue::decode(encodedOp1), JSValue::decode(encodedOp2));
    auto nonOptimizeVariant = operationValueMulProfiledNoOptimize;
    mulIC->generateOutOfLine(vm, exec->codeBlock(), nonOptimizeVariant);

    return profiledMul(exec, encodedOp1, encodedOp2, *arithProfile, false);
}

EncodedJSValue JIT_OPERATION operationValueMulProfiledNoOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITMulIC* mulIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    ArithProfile* arithProfile = mulIC->arithProfile();
    ASSERT(arithProfile);
    return profiledMul(exec, encodedOp1, encodedOp2, *arithProfile);
}

ALWAYS_INLINE static EncodedJSValue unprofiledNegate(ExecState* exec, EncodedJSValue encodedOperand)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue operand = JSValue::decode(encodedOperand);

    JSValue primValue = operand.toPrimitive(exec, PreferNumber);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

#if USE(BIGINT32)
    if (primValue.isBigInt32()) {
        int32_t value = primValue.bigInt32AsInt32();
        if (value != INT_MIN)
            return JSValue::encode(JSValue(JSValue::JSBigInt32, -value));
        primValue = JSBigInt::createFrom(vm, value);
    }
#endif
    if (primValue.isHeapBigInt())
        return JSValue::encode(JSBigInt::unaryMinus(vm, primValue.asHeapBigInt()));

    double number = primValue.toNumber(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    return JSValue::encode(jsNumber(-number));
}

ALWAYS_INLINE static EncodedJSValue profiledNegate(ExecState* exec, EncodedJSValue encodedOperand, ArithProfile& arithProfile)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue operand = JSValue::decode(encodedOperand);
    arithProfile.observeLHS(operand);

    JSValue primValue = operand.toPrimitive(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

#if USE(BIGINT32)
    if (primValue.isBigInt32()) {
        int32_t value = primValue.bigInt32AsInt32();
        if (value != INT_MIN) {
            auto result = JSValue(JSValue::JSBigInt32, -value);
            arithProfile->observeResult(result);
            return JSValue::encode(result);
        }
        primValue = JSBigInt::createFrom(vm, value);
    }
#endif
    if (primValue.isHeapBigInt()) {
        JSBigInt* result = JSBigInt::unaryMinus(vm, primValue.asHeapBigInt());
        arithProfile.observeResult(result);
        return JSValue::encode(result);
    }

    double number = primValue.toNumber(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    JSValue result = jsNumber(-number);
    arithProfile.observeResult(result);
    return JSValue::encode(result);
}

// FIXME: it would be better to call those operationValueNegate, since the operand can be a BigInt
EncodedJSValue JIT_OPERATION operationArithNegate(ExecState* exec, EncodedJSValue operand)
{
    return unprofiledNegate(exec, operand);
}

// FIXME: it would be better to call those operationValueNegate, since the operand can be a BigInt
EncodedJSValue JIT_OPERATION operationArithNegateProfiled(ExecState* exec, EncodedJSValue operand, ArithProfile* arithProfile)
{
    ASSERT(arithProfile);
    return profiledNegate(exec, operand, *arithProfile);
}

EncodedJSValue JIT_OPERATION operationArithNegateProfiledOptimize(ExecState* exec, EncodedJSValue encodedOperand, JITNegIC* negIC)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue operand = JSValue::decode(encodedOperand);

    ArithProfile* arithProfile = negIC->arithProfile();
    ASSERT(arithProfile);
    arithProfile->observeLHS(operand);
    negIC->generateOutOfLine(vm, exec->codeBlock(), operationArithNegateProfiled);

#if ENABLE(MATH_IC_STATS)
    exec->codeBlock()->dumpMathICStats();
#endif

    JSValue primValue = operand.toPrimitive(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

#if USE(BIGINT32)
    if (primValue.isBigInt32()) {
        int32_t value = primValue.bigInt32AsInt32();
        if (value != INT_MIN) {
            auto result = JSValue(JSValue::JSBigInt32, -value);
            arithProfile->observeResult(result);
            return JSValue::encode(result);
        }
        primValue = JSBigInt::createFrom(vm, value);
    }
#endif
    if (primValue.isHeapBigInt()) {
        JSBigInt* result = JSBigInt::unaryMinus(vm, primValue.asHeapBigInt());
        arithProfile->observeResult(result);
        return JSValue::encode(result);
    }

    double number = primValue.toNumber(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    JSValue result = jsNumber(-number);
    arithProfile->observeResult(result);
    return JSValue::encode(result);
}

// FIXME: it would be better to call those operationValueNegate, since the operand can be a BigInt
EncodedJSValue JIT_OPERATION operationArithNegateOptimize(ExecState* exec, EncodedJSValue encodedOperand, JITNegIC* negIC)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    NativeCallFrameTracer tracer(&vm, exec);

    JSValue operand = JSValue::decode(encodedOperand);

    if (ArithProfile* arithProfile = negIC->arithProfile())
    	arithProfile->observeLHS(operand);
    negIC->generateOutOfLine(vm, exec->codeBlock(), operationArithNegate);

#if ENABLE(MATH_IC_STATS)
    exec->codeBlock()->dumpMathICStats();
#endif

    JSValue primValue = operand.toPrimitive(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

#if USE(BIGINT32)
    // FIXME: why does this function profile the argument but not the result?
    if (primValue.isBigInt32()) {
        int32_t value = primValue.bigInt32AsInt32();
        if (value != INT_MIN)
            return JSValue::encode(JSValue(JSValue::JSBigInt32, -value));
        primValue = JSBigInt::createFrom(vm, value);
    }
#endif
    if (primValue.isHeapBigInt())
        return JSValue::encode(JSBigInt::unaryMinus(vm, primValue.asHeapBigInt()));

    double number = primValue.toNumber(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    return JSValue::encode(jsNumber(-number));
}

ALWAYS_INLINE static EncodedJSValue unprofiledSub(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    return JSValue::encode(jsSub(exec, op1, op2));
}

ALWAYS_INLINE static EncodedJSValue profiledSub(VM& vm, ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, ArithProfile& arithProfile, bool shouldObserveLHSAndRHSTypes = true)
{
    JSValue op1 = JSValue::decode(encodedOp1);
    JSValue op2 = JSValue::decode(encodedOp2);

    if (shouldObserveLHSAndRHSTypes)
        arithProfile.observeLHSAndRHS(op1, op2);

    JSValue result = jsSub(exec, op1, op2);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    arithProfile.observeResult(result);
    return JSValue::encode(result);
}

EncodedJSValue JIT_OPERATION operationValueSub(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);
    return unprofiledSub(exec, encodedOp1, encodedOp2);
}

EncodedJSValue JIT_OPERATION operationValueSubProfiled(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, ArithProfile* arithProfile)
{
    ASSERT(arithProfile);

    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return profiledSub(vm, exec, encodedOp1, encodedOp2, *arithProfile);
}

EncodedJSValue JIT_OPERATION operationValueSubOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITSubIC* subIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    auto nonOptimizeVariant = operationValueSubNoOptimize;
    if (ArithProfile* arithProfile = subIC->arithProfile())
        arithProfile->observeLHSAndRHS(JSValue::decode(encodedOp1), JSValue::decode(encodedOp2));
    subIC->generateOutOfLine(vm, exec->codeBlock(), nonOptimizeVariant);

#if ENABLE(MATH_IC_STATS)
    exec->codeBlock()->dumpMathICStats();
#endif

    return unprofiledSub(exec, encodedOp1, encodedOp2);
}

EncodedJSValue JIT_OPERATION operationValueSubNoOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITSubIC*)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    return unprofiledSub(exec, encodedOp1, encodedOp2);
}

EncodedJSValue JIT_OPERATION operationValueSubProfiledOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITSubIC* subIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    ArithProfile* arithProfile = subIC->arithProfile();
    ASSERT(arithProfile);
    arithProfile->observeLHSAndRHS(JSValue::decode(encodedOp1), JSValue::decode(encodedOp2));
    auto nonOptimizeVariant = operationValueSubProfiledNoOptimize;
    subIC->generateOutOfLine(vm, exec->codeBlock(), nonOptimizeVariant);

#if ENABLE(MATH_IC_STATS)
    exec->codeBlock()->dumpMathICStats();
#endif

    return profiledSub(vm, exec, encodedOp1, encodedOp2, *arithProfile, false);
}

EncodedJSValue JIT_OPERATION operationValueSubProfiledNoOptimize(ExecState* exec, EncodedJSValue encodedOp1, EncodedJSValue encodedOp2, JITSubIC* subIC)
{
    VM& vm = exec->vm();
    NativeCallFrameTracer tracer(vm, exec);

    ArithProfile* arithProfile = subIC->arithProfile();
    ASSERT(arithProfile);
    return profiledSub(vm, exec, encodedOp1, encodedOp2, *arithProfile);
}

int32_t JIT_OPERATION operationCheckIfExceptionIsUncatchableAndNotifyProfiler(ExecState* exec, VM* vm)
{
    NativeCallFrameTracer tracer(vm, exec);
    RELEASE_ASSERT(!!vm.exception());

    if (LegacyProfiler* profiler = vm->enabledProfiler())
        profiler->exceptionUnwind(exec);

    if (isTerminatedExecutionException(vm->exception())) {
        genericUnwind(vm, exec);
        return 1;
    } else
        return 0;
}

// Note: getHostCallReturnValueWithExecState() needs to be placed before the
// definition of getHostCallReturnValue() below because the Windows build
// requires it.

extern "C" EncodedJSValue HOST_CALL_RETURN_VALUE_OPTION getHostCallReturnValueWithExecState(ExecState* exec)
{
    if (!exec)
        return JSValue::encode(JSValue());
    return JSValue::encode(exec->vm().hostCallReturnValue);
}

#if COMPILER(GCC)

#if CPU(X86_64)
asm (
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
	"mov 0(%rbp), %rbp\n" // CallerFrameAndPC::callerFrame
	"mov %rbp, %rdi\n"
    "jmp " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "\n"
);
#elif CPU(X86)
asm (
".text" "\n" \
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
	"mov 0(%ebp), %ebp\n" // CallerFrameAndPC::callerFrame
	"mov %ebp, 4(%esp)\n"
    "jmp " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "\n"
);
#elif CPU(ARM_THUMB2)
#if ENABLE(ARM_THUMB)
asm (
".text" "\n"
".align 2" "\n"
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
".thumb" "\n"
".thumb_func " THUMB_FUNC_PARAM(getHostCallReturnValue) "\n"
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
	"ldr r7, [r7, #0]" "\n" // CallerFrameAndPC::callerFrame
	"mov r0, r7" "\n"
    "b " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "\n"
);
#else
asm (
".arm" "\n"
".text" "\n"
".align 4" "\n"
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
	"ldr r7, [r7, #0]" "\n" // CallerFrameAndPC::callerFrame
	"mov r0, r7" "\n"
    "b " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "\n"
);
#endif
#elif CPU(ARM_TRADITIONAL)
asm (
".text" "\n"
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
INLINE_ARM_FUNCTION(getHostCallReturnValue)
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
	"ldr r11, [r11, #0]" "\n" // CallerFrameAndPC::callerFrame
	"mov r0, r11" "\n"
    "b " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "\n"
);
#elif CPU(ARM64)
asm (
".text" "\n"
".align 2" "\n"
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
	"ldur x29, [x29, #0]" "\n"
	"mov x0, x29" "\n"
	"b " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "\n"
);
#elif CPU(MIPS)
asm(
".text" "\n"
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
    LOAD_FUNCTION_TO_T9(getHostCallReturnValueWithExecState)
    "lw $fp, 0($fp)" "\n" // CallerFrameAndPC::callerFrame
    "move $a0, $fp" "\n"
    "b " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "\n"
);
#elif COMPILER(GCC) && CPU(SH4)
asm(
".text" "\n"
".globl " SYMBOL_STRING(getHostCallReturnValue) "\n"
HIDE_SYMBOL(getHostCallReturnValue) "\n"
SYMBOL_STRING(getHostCallReturnValue) ":" "\n"
	"mov.l @r14, r14" "\n" // CallerFrameAndPC::callerFrame
    "mov r14, r4" "\n"
    "mov.l 2f, " SH4_SCRATCH_REGISTER "\n"
    "braf " SH4_SCRATCH_REGISTER "\n"
    "nop" "\n"
    "1: .balign 4" "\n"
    "2: .long " LOCAL_REFERENCE(getHostCallReturnValueWithExecState) "-1b\n"
);
#endif

#endif // COMPILER(GCC)

} // namespace JSC

