/*
 * Copyright (C) 2008, 2009 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"

// #if ENABLE(JIT)
#if USE(JSVALUE32_64)
#include "JIT.h"

#include "CodeBlock.h"
#include "DirectArguments.h"
#include "GCAwareJITStubRoutine.h"
#include "Interpreter.h"
#include "JITInlineMethods.h"
#include "JSArray.h"
#include "JSFunction.h"
#include "JSEnvironmentRecord.h"
#include "LinkBuffer.h"
#include "OpcodeInlines.h"
#include "Operations.h"
#include "PrivateFieldPutKind.h"
#include "ResultType.h"
#include "SamplingTool.h"
#include "SlowPathCall.h"

#ifndef NDEBUG
#include <stdio.h>
#endif

namespace JSC {

void JIT::emit_op_put_getter_by_id(Instruction* currentInstruction)
{
    int base = currentInstruction[1].u.operand;
    int property = currentInstruction[2].u.operand;
    int options = currentInstruction[3].u.operand;
    int getter = currentInstruction[4].u.operand;

    emitLoadPayload(base, regT1);
    emitLoadPayload(getter, regT3);
    callOperation(operationPutGetterById, regT1, m_codeBlock->identifier(property).impl(), options, regT3);
}

void JIT::emit_op_put_setter_by_id(Instruction* currentInstruction)
{
    int base = currentInstruction[1].u.operand;
    int property = currentInstruction[2].u.operand;
    unsigned options = currentInstruction[3].u.operand;
    int setter = currentInstruction[4].u.operand;

    emitLoadPayload(base, regT1);
    emitLoadPayload(setter, regT3);
    callOperation(operationPutSetterById, regT1, m_codeBlock->identifier(property).impl(), options, regT3);
}

void JIT::emit_op_put_getter_setter(Instruction* currentInstruction)
{
	int base = currentInstruction[1].u.operand;
	int property = currentInstruction[2].u.operand;
    unsigned attribute = currentInstruction[3].u.operand;
    int getter = currentInstruction[4].u.operand;
    int setter = currentInstruction[5].u.operand;

    emitLoadPayload(base, regT1);
    emitLoadPayload(getter, regT3);
    emitLoadPayload(setter, regT4);
    callOperation(operationPutGetterSetter, regT1, m_codeBlock->identifier(property).impl(), attribute, regT3, regT4);
}

void JIT::emit_op_put_getter_by_val(Instruction* currentInstruction)
{
    int base = currentInstruction[1].u.operand;
    int property = currentInstruction[2].u.operand;
    int32_t attributes = currentInstruction[3].u.operand;
    int getter = currentInstruction[4].u.operand;

    emitLoadPayload(base, regT2);
    emitLoad(property, regT1, regT0);
    emitLoadPayload(getter, regT3);
    callOperation(operationPutGetterByVal, regT2, regT1, regT0, attributes, regT3);
}

void JIT::emit_op_put_setter_by_val(Instruction* currentInstruction)
{
    int base = currentInstruction[1].u.operand;
    int property = currentInstruction[2].u.operand;
    int32_t attributes = currentInstruction[3].u.operand;
    int getter = currentInstruction[4].u.operand;

    emitLoadPayload(base, regT2);
    emitLoad(property, regT1, regT0);
    emitLoadPayload(getter, regT3);
    callOperation(operationPutSetterByVal, regT2, regT1, regT0, attributes, regT3);
}

void JIT::emit_op_del_by_id(Instruction* currentInstruction)
{
	int dst = currentInstruction[1].u.operand;
	int base = currentInstruction[2].u.operand;
	int property = currentInstruction[3].u.operand;
	ECMAMode ecmaMode = ECMAMode::fromByte(currentInstruction[4].u.operand);

	emitLoad(base, regT1, regT0);
	callOperation(ecmaMode.isStrict() ? operationDeleteByIdJSResultStrict : operationDeleteByIdJSResultSloppy, dst, regT1, regT0, m_codeBlock->identifier(property).impl());
}

void JIT::emit_op_del_by_val(Instruction* currentInstruction)
{
    int dst = currentInstruction[1].u.operand;
    int base = currentInstruction[2].u.operand;
    int property = currentInstruction[3].u.operand;
	ECMAMode ecmaMode = ECMAMode::fromByte(currentInstruction[4].u.operand);
    emitLoad2(base, regT1, regT0, property, regT3, regT2);
    callOperation(ecmaMode.isStrict() ? operationDeleteByValJSResultStrict : operationDeleteByValJSResultSloppy, dst, regT1, regT0, regT3, regT2);
}

void JIT::emit_op_get_by_val(Instruction* currentInstruction)
{
	int dst = currentInstruction[1].u.operand;

	if(!m_enableByValIC) {
        callOperation(operationGetByValNoIC, dst, currentInstruction);
		return;
	}

	int base = currentInstruction[2].u.operand;
	int property = currentInstruction[3].u.operand;
    ArrayProfile* profile = arrayProfileFor<OpGetByValShape>(currentInstruction);
    ByValInfo* byValInfo = m_codeBlock->addByValInfo();
    
    emitLoad2(base, regT1, regT0, property, regT3, regT2);
    
    emitJumpSlowCaseIfNotJSCell(base, regT1);
    PatchableJump notIndex = patchableBranch32WithConstant(NotEqual, regT3, JSValue::Int32Tag);
    addSlowCase(notIndex);
    loadPtr(regT0, JSCell::structureOffset(), regT1);
    emitArrayProfilingSite(regT1, profile);
    and32(TrustedImm32(IndexingShapeMask), regT1);

    PatchableJump badType;
    JumpList slowCases;

    JITArrayMode mode = chooseArrayMode(profile);
    switch (mode) {
    case JITInt32:
        slowCases = emitInt32GetByVal(currentInstruction, badType);
        break;
    case JITDouble:
        slowCases = emitDoubleGetByVal(currentInstruction, badType);
        break;
    case JITContiguous:
        slowCases = emitContiguousGetByVal(currentInstruction, badType);
        break;
    case JITArrayStorage:
        slowCases = emitArrayStorageGetByVal(currentInstruction, badType);
        break;
    default:
        RELEASE_ASSERT_NOT_REACHED();
    }

    addSlowCase(badType);
    addSlowCase(slowCases);

    Label done = label();

#if !ASSERT_DISABLED
    Jump resultOK = branch32(NotEqual, regT1, TrustedImm32(JSValue::EmptyValueTag));
    breakpoint();
    resultOK.link(this);
#endif

    JSValueRegs resultRegs = JSValueRegs(regT1, regT0);

    emitValueProfilingSite(resultRegs);
    emitStore(dst, regT1, regT0);

    Label nextHotPath = label();

    m_byValCompilationInfo.constructAndAppend(byValInfo, m_bytecodeOffset, notIndex, badType, mode, profile, done, nextHotPath);
}

JIT::JumpList JIT::emitContiguousLoad(Instruction*, PatchableJump& badType, IndexingType expectedShape)
{
    JumpList slowCases;

    badType = patchableBranch32(NotEqual, regT1, TrustedImm32(expectedShape));
    loadPtr(regT0, JSObject::butterflyOffset(), regT3);
    slowCases.append(branch32(AboveOrEqual, regT2, Address(regT3, Butterfly::offsetOfPublicLength())));
    load32(BaseIndex(regT3, regT2, TimesEight, OBJECT_OFFSETOF(JSValue, u.asBits.tag)), regT1); // tag
    load32(BaseIndex(regT3, regT2, TimesEight, OBJECT_OFFSETOF(JSValue, u.asBits.payload)), regT0); // payload
    slowCases.append(branch32WithConstant(Equal, regT1, JSValue::EmptyValueTag));

    return slowCases;
}

JIT::JumpList JIT::emitDoubleLoad(Instruction*, PatchableJump& badType)
{
    JumpList slowCases;

    badType = patchableBranch32WithConstant(NotEqual, regT1, DoubleShape);
    loadPtr(regT0, JSObject::butterflyOffset(), regT3);
    slowCases.append(branch32(AboveOrEqual, regT2, Address(regT3, Butterfly::offsetOfPublicLength())));
    loadDouble(BaseIndex(regT3, regT2, TimesEight), fpRegT0);
    slowCases.append(branchIfNaN(fpRegT0));

    return slowCases;
}

JIT::JumpList JIT::emitArrayStorageLoad(Instruction*, PatchableJump& badType)
{
    JumpList slowCases;

    add32(TrustedImm32(-ArrayStorageShape), regT1, regT3);
    badType = patchableBranch32WithConstant(Above, regT3, SlowPutArrayStorageShape - ArrayStorageShape);
    loadPtr(regT0, JSObject::butterflyOffset(), regT3);
    slowCases.append(branch32(AboveOrEqual, regT2, Address(regT3, ArrayStorage::vectorLengthOffset())));
    load32(BaseIndex(regT3, regT2, TimesEight, ArrayStorage::vectorOffset() + OBJECT_OFFSETOF(JSValue, u.asBits.tag)), regT1); // tag
    load32(BaseIndex(regT3, regT2, TimesEight, ArrayStorage::vectorOffset() + OBJECT_OFFSETOF(JSValue, u.asBits.payload)), regT0); // payload
    slowCases.append(branch32WithConstant(Equal, regT1, JSValue::EmptyValueTag));

    return slowCases;
}

JITGetByIdGenerator JIT::emitGetByValWithCachedId(ByValInfo* byValInfo, Instruction* currentInstruction, const Identifier& propertyName, Jump& fastDoneCase, Jump& slowDoneCase, JumpList& slowCases)
{
    int dst = currentInstruction[1].u.operand;

    // base: tag(regT1), payload(regT0)
    // property: tag(regT3), payload(regT2)
    // scratch: regT4

    slowCases.append(branch32WithConstant(NotEqual, regT3, JSValue::CellTag));
    emitByValIdentifierCheck(byValInfo, regT2, regT4, propertyName, slowCases);

    JITGetByIdGenerator gen(
        m_codeBlock, CodeOrigin(m_bytecodeOffset), CallSiteIndex(currentInstruction), RegisterSet::stubUnavailableRegisters(),
		propertyName.impl(), JSValueRegs::payloadOnly(regT0), JSValueRegs(regT1, regT0), DontSpill, AccessType::GetById);
    gen.generateFastPath(*this);

    fastDoneCase = jump();

    Label coldPathBegin = label();
    gen.slowPathJump().link(this);

    Call call = callOperation(WithProfile, operationGetByIdOptimize, dst, gen.stubInfo(), regT1, regT0, propertyName.impl());
    gen.reportSlowPathCall(coldPathBegin, call);
    slowDoneCase = jump();

    return gen;
}

void JIT::emitSlow_op_get_by_val(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
	if(!m_enableByValIC)
		return;

	int dst = currentInstruction[1].u.operand;
	int base = currentInstruction[2].u.operand;
	int property = currentInstruction[3].u.operand;
    ByValInfo* byValInfo = m_byValCompilationInfo[m_byValInstructionIndex].byValInfo;

    linkSlowCaseIfNotJSCell(iter, base); // base cell check
    linkSlowCase(iter); // property int32 check

    Jump nonCell = jump();
    linkSlowCase(iter); // base array check
    Jump notString = branchPtr(NotEqual, Address(regT0, JSCell::structureOffset()), TrustedImmPtr(m_vm->stringStructure.get()));
    emitNakedCall(m_vm->getCTIStub(stringGetByValGenerator).code());
    Jump failed = branchTestPtr(Zero, regT0);
    emitStoreCell(dst, regT0);
    emitJumpSlowToHot(jump(), OPCODE_LENGTH(op_get_by_val));
    failed.link(this);
    notString.link(this);
    nonCell.link(this);

    linkSlowCase(iter); // vector length check
    linkSlowCase(iter); // empty value

    Label slowPath = label();

    emitLoad(base, regT1, regT0);
    emitLoad(property, regT3, regT2);
    Call call = callOperation(operationGetByValOptimize, dst, regT1, regT0, regT3, regT2, byValInfo);

    m_byValCompilationInfo[m_byValInstructionIndex].slowPathTarget = slowPath;
    m_byValCompilationInfo[m_byValInstructionIndex].returnAddress = call;
    m_byValInstructionIndex++;

    emitValueProfilingSite(JSValueRegs(regT1, regT0));
}

void JIT::emit_op_get_private_name(Instruction* currentInstruction)
{
#if 1
	int dst = currentInstruction[1].u.operand;
	int base = currentInstruction[2].u.operand;
	int property = currentInstruction[3].u.operand;

    emitLoad2(base, regT1, regT0, property, regT3, regT2);
    Call call = callOperation(operationGetPrivateNameOptimize, dst, reinterpret_cast<StructureStubInfo*>(0), regT1, regT0, regT3, regT2);
#else
    auto bytecode = currentInstruction->as<OpGetPrivateName>();
    VirtualRegister dst = bytecode.m_dst;
    VirtualRegister base = bytecode.m_base;
    VirtualRegister property = bytecode.m_property;
    auto baseGPR = JSValueRegs::payloadOnly(regT0);
    auto propertyGPR = JSValueRegs(regT3, regT2);

    emitLoad2(base, regT1, regT0, property, regT3, regT2);

    emitJumpSlowCaseIfNotJSCell(base, regT1);

    JITGetByValGenerator gen(
        m_codeBlock, CodeOrigin(m_bytecodeIndex), CallSiteIndex(m_bytecodeIndex), AccessType::GetPrivateName,
        RegisterSet::stubUnavailableRegisters(), baseGPR, propertyGPR, JSValueRegs(regT1, regT0));
    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());
    m_getByVals.append(gen);
#endif

    emitValueProfilingSite(JSValueRegs(regT1, regT0));
    emitStore(dst, regT1, regT0);
}

void JIT::emitSlow_op_get_private_name(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
#if 0
    ASSERT(hasAnySlowCases(iter));
    auto bytecode = currentInstruction->as<OpGetPrivateName>();
    VirtualRegister dst = bytecode.m_dst;

    linkAllSlowCases(iter);

    JITGetByValGenerator& gen = m_getByVals[m_getByValIndex];
    ++m_getByValIndex;
    Label coldPathBegin = label();

    auto baseGPR = JSValueRegs(regT1, regT0);
    auto propertyGPR = JSValueRegs(regT3, regT2);
    Call call = callOperationWithProfile(bytecode.metadata(m_codeBlock), operationGetPrivateNameOptimize, dst, TrustedImmPtr(m_codeBlock->globalObject()), gen.stubInfo(), baseGPR, propertyGPR);
    gen.reportSlowPathCall(coldPathBegin, call);
#endif
}

void JIT::emit_op_put_private_name(Instruction* currentInstruction)
{
	int base = currentInstruction[1].u.operand;
	int property = currentInstruction[2].u.operand;
    ByValInfo* byValInfo = m_codeBlock->addByValInfo();

    emitLoad2(base, regT1, regT0, property, regT3, regT2);

    emitJumpSlowCaseIfNotJSCell(base, regT1);
    PatchableJump fastPathJmp = patchableJump();
    addSlowCase(fastPathJmp);

    Label done = label();

    m_byValCompilationInfo.append(ByValCompilationInfo(byValInfo, m_bytecodeOffset, fastPathJmp, done, done));
}

void JIT::emitSlow_op_put_private_name(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
	int base = currentInstruction[1].u.operand;
	int property = currentInstruction[2].u.operand;
    int value = currentInstruction[3].u.operand;

    ByValInfo* byValInfo = m_byValCompilationInfo[m_byValInstructionIndex].byValInfo;
    PrivateFieldPutKind putKind = PrivateFieldPutKind::fromByte(static_cast<uint8_t>(currentInstruction[4].u.operand));

    linkAllSlowCases(iter);
    Label slowPath = label();

    JSValueRegs baseRegs(regT1, regT0);
    JSValueRegs propertyRegs(regT3, regT2);
    JSValueRegs valueRegs(regT5, regT4);

    emitLoad(base, baseRegs.tagGPR(), baseRegs.payloadGPR());
    emitLoad(property, propertyRegs.tagGPR(), propertyRegs.payloadGPR());
    emitLoad(value, valueRegs.tagGPR(), valueRegs.payloadGPR());
    Call call = callOperation(operationPutPrivateNameOptimize, baseRegs, propertyRegs, valueRegs, byValInfo, putKind.value());

    m_byValCompilationInfo[m_byValInstructionIndex].slowPathTarget = slowPath;
    m_byValCompilationInfo[m_byValInstructionIndex].returnAddress = call;
    m_byValInstructionIndex++;
}

void JIT::emit_op_set_private_brand(const Instruction* currentInstruction)
{
	int base = currentInstruction[1].u.operand;
	int brand = currentInstruction[2].u.operand;
    JSValueRegs baseRegs(regT1, regT0);
    JSValueRegs brandRegs(regT3, regT2);
    emitLoad(base, baseRegs.tagGPR(), baseRegs.payloadGPR());
    emitLoad(brand, brandRegs.tagGPR(), brandRegs.payloadGPR());

#if 0
    emitJumpSlowCaseIfNotJSCell(base, baseRegs.tagGPR());

    JITPrivateBrandAccessGenerator gen(
        m_codeBlock, CodeOrigin(m_bytecodeIndex), CallSiteIndex(m_bytecodeIndex), AccessType::SetPrivateBrand, RegisterSet::stubUnavailableRegisters(),
        baseRegs, brandRegs);
    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());
    m_privateBrandAccesses.append(gen);

    // We should emit write-barrier at the end of sequence since write-barrier clobbers registers.
    // IC can write new Structure without write-barrier if a base is cell.
    // FIXME: Use UnconditionalWriteBarrier in Baseline effectively to reduce code size.
    // https://bugs.webkit.org/show_bug.cgi?id=209395
    emitWriteBarrier(base, ShouldFilterBase);
#else
    callOperation(operationSetPrivateBrandGeneric, nullptr, baseRegs, brandRegs);
#endif
}

void JIT::emitSlow_op_set_private_brand(const Instruction*, Vector<SlowCaseEntry>::iterator& iter)
{
#if 0
    JSValueRegs baseRegs(regT1, regT0);
    JSValueRegs brandRegs(regT3, regT2);

    linkAllSlowCases(iter);

    JITPrivateBrandAccessGenerator& gen = m_privateBrandAccesses[m_privateBrandAccessIndex];
    ++m_privateBrandAccessIndex;
    Label coldPathBegin = label();
    Call call = callOperation(operationSetPrivateBrandOptimize, TrustedImmPtr(m_codeBlock->globalObject()), gen.stubInfo(), baseRegs, brandRegs);
    gen.reportSlowPathCall(coldPathBegin, call);
#endif
}

void JIT::emit_op_check_private_brand(const Instruction* currentInstruction)
{
	int base = currentInstruction[1].u.operand;
	int brand = currentInstruction[2].u.operand;
    JSValueRegs baseRegs(regT1, regT0);
    JSValueRegs brandRegs(regT3, regT2);
    emitLoad(base, baseRegs.tagGPR(), baseRegs.payloadGPR());
    emitLoad(brand, brandRegs.tagGPR(), brandRegs.payloadGPR());

#if 0
    emitJumpSlowCaseIfNotJSCell(base, baseRegs.tagGPR());

    JITPrivateBrandAccessGenerator gen(
        m_codeBlock, CodeOrigin(m_bytecodeIndex), CallSiteIndex(m_bytecodeIndex), AccessType::CheckPrivateBrand, RegisterSet::stubUnavailableRegisters(),
        baseRegs, brandRegs);
    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());
    m_privateBrandAccesses.append(gen);
#else
    callOperation(operationCheckPrivateBrandGeneric, nullptr, baseRegs, brandRegs);
#endif
}

void JIT::emitSlow_op_check_private_brand(const Instruction*, Vector<SlowCaseEntry>::iterator& iter)
{
#if 0
    JSValueRegs baseRegs(regT1, regT0);
    JSValueRegs brandRegs(regT3, regT2);

    linkAllSlowCases(iter);

    JITPrivateBrandAccessGenerator& gen = m_privateBrandAccesses[m_privateBrandAccessIndex];
    ++m_privateBrandAccessIndex;
    Label coldPathBegin = label();
    Call call = callOperation(operationCheckPrivateBrandOptimize, TrustedImmPtr(m_codeBlock->globalObject()), gen.stubInfo(), baseRegs, brandRegs);
    gen.reportSlowPathCall(coldPathBegin, call);
#endif
}

void JIT::emit_op_put_by_val(Instruction* currentInstruction)
{
	if(!m_enableByValIC) {
	    bool isDirect = m_interpreter->getOpcodeID(currentInstruction->u.opcode) == op_put_by_val_direct;
        callOperation(isDirect ? operationDirectPutByVal : operationPutByVal, currentInstruction);
		return;
	}

	int base = currentInstruction[1].u.operand;
	int property = currentInstruction[2].u.operand;
    ArrayProfile* profile = arrayProfileFor<OpPutByValShape>(currentInstruction);
    ByValInfo* byValInfo = m_codeBlock->addByValInfo();
    
    emitLoad2(base, regT1, regT0, property, regT3, regT2);
    
    emitJumpSlowCaseIfNotJSCell(base, regT1);
    PatchableJump notIndex = patchableBranch32WithConstant(NotEqual, regT3, JSValue::Int32Tag);
    addSlowCase(notIndex);
    loadPtr(regT0, JSCell::structureOffset(), regT1);
    emitArrayProfilingSite(regT1, profile);

    PatchableJump badType;
    JumpList slowCases;

    and32(TrustedImm32(IndexingShapeMask), regT1);

    JITArrayMode mode = chooseArrayMode(profile);
    switch (mode) {
    case JITInt32:
        slowCases = emitInt32PutByVal(currentInstruction, badType);
        break;
    case JITDouble:
        slowCases = emitDoublePutByVal(currentInstruction, badType);
        break;
    case JITContiguous:
        slowCases = emitContiguousPutByVal(currentInstruction, badType);
        break;
    case JITArrayStorage:
        slowCases = emitArrayStoragePutByVal(currentInstruction, badType);
        break;
    default:
        RELEASE_ASSERT_NOT_REACHED();
        break;
    }

    addSlowCase(badType);
    addSlowCase(slowCases);

    Label done = label();

    m_byValCompilationInfo.constructAndAppend(byValInfo, m_bytecodeOffset, notIndex, badType, mode, profile, done, done);
}

JIT::JumpList JIT::emitGenericContiguousPutByVal(Instruction* currentInstruction, PatchableJump& badType, IndexingType indexingShape)
{
	int value = currentInstruction[3].u.operand;
    ArrayProfile* profile = arrayProfileFor<OpPutByValShape>(currentInstruction);

    JumpList slowCases;

    badType = patchableBranch32WithConstant(NotEqual, regT1, ContiguousShape);

    loadPtr(regT0, JSObject::butterflyOffset(), regT3);
    Jump outOfBounds = branch32(AboveOrEqual, regT2, Address(regT3, Butterfly::offsetOfPublicLength()));

    Label storeResult = label();
    emitLoad(value, regT1, regT0);
    switch (indexingShape) {
    case Int32Shape:
        slowCases.append(branch32WithConstant(NotEqual, regT1, JSValue::Int32Tag));
        // Fall through.
    case ContiguousShape:
		store32(regT0, BaseIndex(regT3, regT2, TimesEight, OBJECT_OFFSETOF(JSValue, u.asBits.payload)));
		store32(regT1, BaseIndex(regT3, regT2, TimesEight, OBJECT_OFFSETOF(JSValue, u.asBits.tag)));
		break;
	case DoubleShape: {
		Jump notInt = branch32WithConstant(NotEqual, regT1, JSValue::Int32Tag);
		convertInt32ToDouble(regT0, fpRegT0);
		Jump ready = jump();
		notInt.link(this);
		moveIntsToDouble(regT0, regT1, fpRegT0, fpRegT1);
        slowCases.append(branchIfNaN(fpRegT0));
		ready.link(this);
		storeDouble(fpRegT0, BaseIndex(regT3, regT2, TimesEight));
		break;
	}
	default:
		RELEASE_ASSERT_NOT_REACHED();
		break;
	}

    Jump done = jump();

    outOfBounds.link(this);
    slowCases.append(branch32(AboveOrEqual, regT2, Address(regT3, Butterfly::offsetOfVectorLength())));

    emitArrayProfileStoreToHoleSpecialCase(profile);

    add32(TrustedImm32(1), regT2, regT1);
    store32(regT1, Address(regT3, Butterfly::offsetOfPublicLength()));
    jump().linkTo(storeResult, this);

    done.link(this);

    emitWriteBarrier(regT0, regT1, regT1, regT3, UnconditionalWriteBarrier, WriteBarrierForPropertyAccess);
    
    return slowCases;
}

JIT::JumpList JIT::emitArrayStoragePutByVal(Instruction* currentInstruction, PatchableJump& badType)
{
	int value = currentInstruction[3].u.operand;
    ArrayProfile* profile = arrayProfileFor<OpPutByValShape>(currentInstruction);

    JumpList slowCases;

    badType = patchableBranch32WithConstant(NotEqual, regT1, ArrayStorageShape);

    loadPtr(regT0, JSObject::butterflyOffset(), regT3);
    slowCases.append(branch32(AboveOrEqual, regT2, Address(regT3, ArrayStorage::vectorLengthOffset())));

    Jump empty = branch32WithConstant(Equal, BaseIndex(regT3, regT2, TimesEight, ArrayStorage::vectorOffset() + OBJECT_OFFSETOF(JSValue, u.asBits.tag)), JSValue::EmptyValueTag);
    
    Label storeResult(this);
    emitLoad(value, regT1, regT0);
    store32(regT0, BaseIndex(regT3, regT2, TimesEight, ArrayStorage::vectorOffset() + OBJECT_OFFSETOF(JSValue, u.asBits.payload))); // payload
    store32(regT1, BaseIndex(regT3, regT2, TimesEight, ArrayStorage::vectorOffset() + OBJECT_OFFSETOF(JSValue, u.asBits.tag))); // tag
    Jump end = jump();
    
    empty.link(this);
    emitArrayProfileStoreToHoleSpecialCase(profile);
    add32(TrustedImm32(1), Address(regT3, OBJECT_OFFSETOF(ArrayStorage, m_numValuesInVector)));
    branch32(Below, regT2, Address(regT3, ArrayStorage::lengthOffset())).linkTo(storeResult, this);
    
    add32(TrustedImm32(1), regT2, regT0);
    store32(regT0, Address(regT3, ArrayStorage::lengthOffset()));
    jump().linkTo(storeResult, this);
    
    end.link(this);

    emitWriteBarrier(regT0, regT1, regT1, regT3, UnconditionalWriteBarrier, WriteBarrierForPropertyAccess);

    return slowCases;
}

JITPutByIdGenerator JIT::emitPutByValWithCachedId(ByValInfo* byValInfo, Instruction* currentInstruction, PutKind putKind, const Identifier& propertyName, JumpList& doneCases, JumpList& slowCases)
{
    // base: tag(regT1), payload(regT0)
    // property: tag(regT3), payload(regT2)

	auto opcodeID = m_interpreter->getOpcodeID(currentInstruction->u.opcode);

    int base = currentInstruction[1].u.operand;
    int value = currentInstruction[3].u.operand;
	ECMAMode ecmaMode = ECMAMode::fromByte(currentInstruction[4].u.operand);

    slowCases.append(branch32WithConstant(NotEqual, regT3, JSValue::CellTag));
    emitByValIdentifierCheck(byValInfo, regT2, regT2, propertyName, slowCases);

    // Write barrier breaks the registers. So after issuing the write barrier,
    // reload the registers.
    emitWriteBarrier(regT0, regT1, regT1, regT3, UnconditionalWriteBarrier, WriteBarrierForPropertyAccess);
    emitLoadPayload(base, regT0);
    emitLoad(value, regT3, regT2);

    JITPutByIdGenerator gen(
        m_codeBlock, CodeOrigin(m_bytecodeOffset), CallSiteIndex(currentInstruction), RegisterSet::stubUnavailableRegisters(),
        JSValueRegs::payloadOnly(regT0), JSValueRegs(regT3, regT2),
		regT1, DontSpill, ecmaMode, putKind);

    gen.generateFastPath(*this);
    doneCases.append(jump());

    Label coldPathBegin = label();
    gen.slowPathJump().link(this);

    // JITPutByIdGenerator only preserve the value and the base's payload, we have to reload the tag.
    emitLoadTag(base, regT1);

    Call call = callOperation(gen.slowPathFunction(), gen.stubInfo(), regT3, regT2, regT1, regT0, propertyName.impl());
    gen.reportSlowPathCall(coldPathBegin, call);
    doneCases.append(jump());

    return gen;
}

void JIT::emitSlow_op_put_by_val(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
	if(!m_enableByValIC)
		return;

	auto opcodeID = m_interpreter->getOpcodeID(currentInstruction->u.opcode);

	int base = currentInstruction[1].u.operand;
	int property = currentInstruction[2].u.operand;
	int value = currentInstruction[3].u.operand;
	ECMAMode ecmaMode = ECMAMode::fromByte(currentInstruction[4].u.operand);
    ByValInfo* byValInfo = m_byValCompilationInfo[m_byValInstructionIndex].byValInfo;

    linkAllSlowCases(iter);
    Label slowPath = label();

    bool isDirect = opcodeID == op_put_by_val_direct;

#if CPU(X86)
    // FIXME: We only have 5 temp registers, but need 6 to make this call, therefore we materialize
    // our own call. When we finish moving JSC to the C call stack, we'll get another register so
    // we can use the normal case.
    resetCallArguments();
    addCallArgument(GPRInfo::callFrameRegister);
    emitLoad(base, regT0, regT1);
    addCallArgument(regT1);
    addCallArgument(regT0);
    emitLoad(property, regT0, regT1);
    addCallArgument(regT1);
    addCallArgument(regT0);
    emitLoad(value, regT0, regT1);
    addCallArgument(regT1);
    addCallArgument(regT0);
    addCallArgument(TrustedImmPtr(byValInfo));
    Call call = appendCallWithExceptionCheck(isDirect ? operationDirectPutByValOptimize : operationPutByValOptimize);
#else
    // The register selection below is chosen to reduce register swapping on ARM.
    // Swapping shouldn't happen on other platforms.
    emitLoad(base, regT2, regT1);
    emitLoad(property, regT3, regT0);
    emitLoad(value, regT5, regT4);
    Call call = callOperation(
    	(isDirect ?
    		(ecmaMode.isStrict() ? operationDirectPutByValOptimizeStrict : operationDirectPutByValOptimizeSloppy) :
    		(ecmaMode.isStrict() ? operationPutByValOptimizeStrict : operationPutByValOptimizeSloppy)),
    	regT2, regT1, regT3, regT0, regT5, regT4, byValInfo);
#endif

    m_byValCompilationInfo[m_byValInstructionIndex].slowPathTarget = slowPath;
    m_byValCompilationInfo[m_byValInstructionIndex].returnAddress = call;
    m_byValInstructionIndex++;
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emit_op_try_get_by_id(Instruction* currentInstruction)
{
    int dst = currentInstruction[1].u.operand;
    int base = currentInstruction[2].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[3].u.operand));

    emitLoad(base, regT1, regT0);

    if(m_forceTryGetByIdICFailure) {
    	callOperation(WithProfile, operationTryGetByIdGeneric, dst, regT1, regT0, ident->impl());
    	return;
    }

    // Sunspider is much slower
    emitJumpSlowCaseIfNotJSCell(base, regT1);

    JSValueRegs resultRegs = JSValueRegs(regT1, regT0);

    m_getByIds.constructAndAppend(
        m_codeBlock, CodeOrigin(m_bytecodeOffset), CallSiteIndex(currentInstruction), RegisterSet::stubUnavailableRegisters(),
		ident->impl(), JSValueRegs::payloadOnly(regT0), resultRegs, DontSpill, AccessType::TryGet);

    JITGetByIdGenerator& gen = m_getByIds.last();

    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());
    m_getByIds.append(gen);

    emitValueProfilingSite(resultRegs);
    emitStore(dst, resultRegs.tagGPR(), resultRegs.payloadGPR());
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emitSlow_op_try_get_by_id(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    if(m_forceTryGetByIdICFailure)
    	return;

    linkAllSlowCases(iter);

    int resultVReg = currentInstruction[1].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[3].u.operand));

    JITGetByIdGenerator& gen = m_getByIds[m_getByIdIndex++];

    Label coldPathBegin = label();

    Call call = callOperation(WithProfile, operationTryGetByIdOptimize, resultVReg, gen.stubInfo(), regT1, regT0, ident->impl());

    gen.reportSlowPathCall(coldPathBegin, call);
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emit_op_get_by_id_direct(Instruction* currentInstruction)
{
    int dst = currentInstruction[1].u.operand;
    int base = currentInstruction[2].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[3].u.operand));

    emitLoad(base, regT1, regT0);
    emitJumpSlowCaseIfNotJSCell(base, regT1);

    JSValueRegs resultRegs = JSValueRegs(regT1, regT0);

    JITGetByIdGenerator gen(
        m_codeBlock, CodeOrigin(m_bytecodeOffset), CallSiteIndex(currentInstruction), RegisterSet::stubUnavailableRegisters(),
        ident->impl(), JSValueRegs::payloadOnly(regT0), resultRegs, DontSpill, AccessType::GetDirect);
    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());
    m_getByIds.append(gen);

    emitValueProfilingSite(resultRegs);
    emitStore(dst, resultRegs.tagGPR(), resultRegs.payloadGPR());
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emitSlow_op_get_by_id_direct(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    linkAllSlowCases(iter);

    int resultVReg = currentInstruction[1].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[3].u.operand));

    JITGetByIdGenerator& gen = m_getByIds[m_getByIdIndex++];

    Label coldPathBegin = label();

    Call call = callOperation(WithProfile, operationGetByIdDirectOptimize, resultVReg, gen.stubInfo(), regT1, regT0, ident->impl());

    gen.reportSlowPathCall(coldPathBegin, call);
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emit_op_get_by_id(Instruction* currentInstruction)
{
    int dst = currentInstruction[1].u.operand;
    Identifier* ident = &(m_codeBlock->identifier(currentInstruction[3].u.operand));

    if(!m_enableGetByIdIC) {
    	if(parseIndex(*ident))
    		callOperation(operationGetByIdNoIC, dst, currentInstruction);
    	else
    		callOperation(operationGetByIdNoICNotIndex, dst, currentInstruction);
    	return;
    }

    int base = currentInstruction[2].u.operand;

#if ENABLE_JIT_PROFILE
    static unsigned sGetByIdInstructionSize = 0;
    unsigned startInstructionSize = assembler().buffer().codeSize();
#endif
    
    emitLoad(base, regT1, regT0);

    emitJumpSlowCaseIfNotJSCell(base, regT1);

    if (*ident == m_vm->propertyNames->length && shouldEmitProfiling()) {
        loadPtr(regT0, JSCell::structureOffset(), regT2);
        emitArrayProfilingSiteForBytecodeIndex(regT2, m_bytecodeOffset);
    }

    JSValueRegs resultRegs = JSValueRegs(regT1, regT0);
    m_getByIds.constructAndAppend(
        m_codeBlock, CodeOrigin(m_bytecodeOffset), CallSiteIndex(currentInstruction), RegisterSet::stubUnavailableRegisters(),
		ident->impl(), JSValueRegs::payloadOnly(regT0), resultRegs, DontSpill, AccessType::GetById);

    JITGetByIdGenerator& gen = m_getByIds.last();
    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());

    emitValueProfilingSite(resultRegs);
    emitStore(dst, resultRegs.tagGPR(), resultRegs.payloadGPR());

#if ENABLE_JIT_PROFILE
    if(!sGetByIdInstructionSize) {
    	sGetByIdInstructionSize = assembler().buffer().codeSize() - startInstructionSize;
    	WTF::logMessage(String("get_by_id JIT code size ") + String::number(sGetByIdInstructionSize));
    }
#endif
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emitSlow_op_get_by_id(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    if(!m_enableGetByIdIC)
    	return;

#if ENABLE_JIT_PROFILE
    static unsigned sGetByIdSlowInstructionSize = 0;
    unsigned startInstructionSize = assembler().buffer().codeSize();
#endif

    linkAllSlowCases(iter);

    int resultVReg = currentInstruction[1].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[3].u.operand));

    JITGetByIdGenerator& gen = m_getByIds[m_getByIdIndex++];

    Label coldPathBegin = label();
    
    Call call = callOperation(WithProfile, operationGetByIdOptimize, resultVReg, gen.stubInfo(), regT1, regT0, ident->impl());

    gen.reportSlowPathCall(coldPathBegin, call);

#if ENABLE_JIT_PROFILE
    if(!sGetByIdSlowInstructionSize) {
    	sGetByIdSlowInstructionSize = assembler().buffer().codeSize() - startInstructionSize;
    	WTF::logMessage(String("get_by_id slow JIT code size ") + String::number(sGetByIdSlowInstructionSize));
    }
#endif
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emit_op_get_by_id_with_this(Instruction* currentInstruction)
{
    int dst = currentInstruction[1].u.operand;
    int base = currentInstruction[2].u.operand;
    int thisVReg = currentInstruction[3].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[4].u.operand));

    emitLoad(base, regT1, regT0);
    emitLoad(thisVReg, regT4, regT3);
    emitJumpSlowCaseIfNotJSCell(base, regT1);
    emitJumpSlowCaseIfNotJSCell(thisVReg, regT4);

    JSValueRegs resultRegs = JSValueRegs(regT1, regT0);

    m_getByIdsWithThis.constructAndAppend(
        m_codeBlock, CodeOrigin(m_bytecodeOffset), CallSiteIndex(currentInstruction), RegisterSet::stubUnavailableRegisters(),
        ident->impl(), resultRegs, JSValueRegs::payloadOnly(regT0), JSValueRegs(regT4, regT3), DontSpill, AccessType::GetByIdWithThis);

    JITGetByIdWithThisGenerator& gen = m_getByIdsWithThis.last();
    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());

    emitValueProfilingSite(resultRegs);
    emitStore(dst, resultRegs.tagGPR(), resultRegs.payloadGPR());
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emitSlow_op_get_by_id_with_this(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    linkAllSlowCases(iter);

    int resultVReg = currentInstruction[1].u.operand;
    int baseVReg = currentInstruction[2].u.operand;
    int thisVReg = currentInstruction[3].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[4].u.operand));

    JITGetByIdWithThisGenerator& gen = m_getByIdsWithThis[m_getByIdWithThisIndex++];

    Label coldPathBegin = label();

    Call call = callOperation(WithProfile, operationGetByIdWithThisOptimize, resultVReg, gen.stubInfo(), regT1, regT0, regT4, regT3, ident->impl());

    gen.reportSlowPathCall(coldPathBegin, call);
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emit_op_put_by_id(Instruction* currentInstruction)
{
    // In order to be able to patch both the Structure, and the object offset, we store one pointer,
    // to just after the arguments have been loaded into registers 'hotPathBegin', and we generate code
    // such that the Structure & offset are always at the same distance from this.
	PutByIdFlags putByIdFlags = PutByIdFlags::convert(currentInstruction[OP_PUT_BY_ID_LENGTH - 1].u.operand);
    int direct = putByIdFlags.isDirect();
	ECMAMode ecmaMode = putByIdFlags.ecmaMode();

#if ENABLE_JIT_PROFILE
    static unsigned sPutByIdInstructionSize = 0;
    static unsigned sPutByIdDisabledInstructionSize = 0;
    unsigned startInstructionSize = assembler().buffer().codeSize();
#endif

    if(!m_enablePutByIdIC) {
    	auto operation = appropriateGenericPutByIdFunction(ecmaMode, direct ? PutKind::Direct : PutKind::NotDirect);
    	callOperation(operation, currentInstruction);

#if ENABLE_JIT_PROFILE
        if(!sPutByIdDisabledInstructionSize) {
        	sPutByIdDisabledInstructionSize = assembler().buffer().codeSize() - startInstructionSize;
        	WTF::logMessage(String("put_by_id disalbed JIT code size ") + String::number(sPutByIdDisabledInstructionSize));
        }
#endif

    	return;
    }

    int base = currentInstruction[1].u.operand;
    int value = currentInstruction[3].u.operand;

    emitLoad2(base, regT1, regT0, value, regT3, regT2);

    emitJumpSlowCaseIfNotJSCell(base, regT1);

    emitWriteBarrier(regT0, regT1, regT2, regT3, ShouldFilterImmediates, WriteBarrierForPropertyAccess);

    auto T1 = regT1;
    
    m_putByIds.constructAndAppend(
        m_codeBlock, CodeOrigin(m_bytecodeOffset), CallSiteIndex(currentInstruction), RegisterSet::stubUnavailableRegisters(),
        JSValueRegs::payloadOnly(regT0), JSValueRegs(regT3, regT2),
		T1, DontSpill, ecmaMode, direct ? PutKind::Direct : PutKind::NotDirect);
    
    JITPutByIdGenerator& gen = m_putByIds.last();
    gen.generateFastPath(*this);
    addSlowCase(gen.slowPathJump());

#if ENABLE_JIT_PROFILE
    if(!sPutByIdInstructionSize) {
    	sPutByIdInstructionSize = assembler().buffer().codeSize() - startInstructionSize;
    	WTF::logMessage(String("put_by_id JIT code size ") + String::number(sPutByIdInstructionSize));
    }
#endif
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emitSlow_op_put_by_id(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    if(!m_enablePutByIdIC)
    	return;

#if ENABLE_JIT_PROFILE
    static unsigned sPutByIdSlowInstructionSize = 0;
    unsigned startInstructionSize = assembler().buffer().codeSize();
#endif

    linkAllSlowCases(iter);

    int base = currentInstruction[1].u.operand;
    const Identifier* ident = &(m_codeBlock->identifier(currentInstruction[2].u.operand));

    Label coldPathBegin(this);

    // JITPutByIdGenerator only preserve the value and the base's payload, we have to reload the tag.
    emitLoadTag(base, regT1);

    JITPutByIdGenerator& gen = m_putByIds[m_putByIdIndex++];

    Call call = callOperation(
        gen.slowPathFunction(), gen.stubInfo(), regT3, regT2, regT1, regT0, ident->impl());
    
    gen.reportSlowPathCall(coldPathBegin, call);

#if ENABLE_JIT_PROFILE
    if(!sPutByIdSlowInstructionSize) {
    	sPutByIdSlowInstructionSize = assembler().buffer().codeSize() - startInstructionSize;
    	WTF::logMessage(String("put_by_id slow JIT code size ") + String::number(sPutByIdSlowInstructionSize));
    }
#endif
}

void JIT::emitVarInjectionCheck(bool needsVarInjectionChecks)
{
    if (!needsVarInjectionChecks)
        return;
    addSlowCase(branch8(Equal, AbsoluteAddress(m_codeBlock->globalObject()->varInjectionWatchpoint()->addressOfState()), TrustedImm32(IsInvalidated)));
}

void JIT::emitResolveClosure(int dst, int scope, bool needsVarInjectionChecks, unsigned depth)
{
    emitVarInjectionCheck(needsVarInjectionChecks);
    move(TrustedImm32(JSValue::CellTag), regT1);
    emitLoadPayload(scope, regT0);
    for (unsigned i = 0; i < depth; ++i)
        loadPtr(regT0, JSScope::offsetOfNext(), regT0);
    emitStore(dst, regT1, regT0);
}

void JIT::emit_op_resolve_scope_for_hoisting_func_decl_in_eval(Instruction* currentInstruction)
{
    JITSlowPathCall slowPathCall(this, currentInstruction, slow_path_resolve_scope_for_hoisting_func_decl_in_eval);
    slowPathCall.call();
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emit_op_resolve_scope(Instruction* currentInstruction)
{
	int dst = currentInstruction[1].u.operand;
    int scope = currentInstruction[2].u.operand;
    ResolveType resolveType = static_cast<ResolveType>(currentInstruction[4].u.operand);
    unsigned depth = currentInstruction[5].u.operand;

    auto emitCode = [&] (ResolveType resolveType) {
			switch (resolveType) {
	        case GlobalProperty:
	        case GlobalPropertyWithVarInjectionChecks:{
	            JSScope* constantScope = JSScope::constantScopeForCodeBlock(resolveType, m_codeBlock);
	            RELEASE_ASSERT(constantScope);
	            emitVarInjectionCheck(needsVarInjectionChecks(resolveType));
	            load32(&currentInstruction[5].u.operand, regT1);
	            addSlowCase(branch32(NotEqual, AbsoluteAddress(m_codeBlock->globalObject()->addressOfGlobalLexicalBindingEpoch()), regT1));
	            move(TrustedImm32(JSValue::CellTag), regT1);
	            move(TrustedImmPtr(constantScope), regT0);
	            emitStore(dst, regT1, regT0);
	            break;
	        }

	        case GlobalVar:
	        case GlobalVarWithVarInjectionChecks:
	        case GlobalLexicalVar:
	        case GlobalLexicalVarWithVarInjectionChecks: {
				JSScope* constantScope = JSScope::constantScopeForCodeBlock(resolveType, codeBlock());
				RELEASE_ASSERT(constantScope);
				emitVarInjectionCheck(needsVarInjectionChecks(resolveType));
				move(TrustedImm32(JSValue::CellTag), regT1);
				move(TrustedImmPtr(constantScope), regT0);
				emitStore(dst, regT1, regT0);
				break;
			}
			case ClosureVar:
			case ClosureVarWithVarInjectionChecks:
				emitResolveClosure(dst, scope, needsVarInjectionChecks(resolveType), depth);
				break;
		    case ModuleVar:
		    	move(TrustedImm32(JSValue::CellTag), regT1);
		    	move(TrustedImmPtr(currentInstruction[6].u.jsCell.get()), regT0);
		    	emitStore(dst, regT1, regT0);
		        break;
			case Dynamic:
				addSlowCase(jump());
				break;
			case ResolvedClosureVar:
			case UnresolvedProperty:
			case UnresolvedPropertyWithVarInjectionChecks:
				RELEASE_ASSERT_NOT_REACHED();
			}
    };
    switch (resolveType) {
    case GlobalProperty:
    case GlobalPropertyWithVarInjectionChecks: {
        JumpList skipToEnd;
        load32(&currentInstruction[4].u.operand, regT0);

        Jump notGlobalProperty = branch32(NotEqual, regT0, TrustedImm32(resolveType));
        emitCode(resolveType);
        skipToEnd.append(jump());

        notGlobalProperty.link(this);
        emitCode(needsVarInjectionChecks(resolveType) ? GlobalLexicalVarWithVarInjectionChecks : GlobalLexicalVar);

        skipToEnd.link(this);
        break;
    }
    case UnresolvedProperty:
    case UnresolvedPropertyWithVarInjectionChecks: {
        JumpList skipToEnd;
        load32(&currentInstruction[4], regT0);

        Jump notGlobalProperty = branch32WithConstant(NotEqual, regT0, GlobalProperty);
        emitCode(GlobalProperty);
        skipToEnd.append(jump());
        notGlobalProperty.link(this);

        Jump notGlobalPropertyWithVarInjections = branch32WithConstant(NotEqual, regT0, GlobalPropertyWithVarInjectionChecks);
        emitCode(GlobalPropertyWithVarInjectionChecks);
        skipToEnd.append(jump());
        notGlobalPropertyWithVarInjections.link(this);

        Jump notGlobalLexicalVar = branch32WithConstant(NotEqual, regT0, GlobalLexicalVar);
        emitCode(GlobalLexicalVar);
        skipToEnd.append(jump());
        notGlobalLexicalVar.link(this);

        Jump notGlobalLexicalVarWithVarInjections = branch32WithConstant(NotEqual, regT0, GlobalLexicalVarWithVarInjectionChecks);
        emitCode(GlobalLexicalVarWithVarInjectionChecks);
        skipToEnd.append(jump());
        notGlobalLexicalVarWithVarInjections.link(this);

        addSlowCase(jump());
        skipToEnd.link(this);
        break;
    }

    default:
        emitCode(resolveType);
        break;
    }
}

#if ENABLE(JSC_ALL_IN_ONE)
ALWAYS_INLINE
#endif
void JIT::emitSlow_op_resolve_scope(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    linkAllSlowCases(iter);

    JITSlowPathCall slowPathCall(this, currentInstruction, slow_path_resolve_scope);
    slowPathCall.call();
}

void JIT::emitLoadWithStructureCheck(int scope, Structure** structureSlot)
{
    emitLoad(scope, regT1, regT0);
    loadPtr(structureSlot, regT2);
    addSlowCase(branchPtr(NotEqual, Address(regT0, JSCell::structureOffset()), regT2));
}

void JIT::emitGetVarFromPointer(JSValue* operand, GPRReg tag, GPRReg payload)
{
    uintptr_t rawAddress = bitwise_cast<uintptr_t>(operand);
    load32(bitwise_cast<void*>(rawAddress + TagOffset), tag);
    load32(bitwise_cast<void*>(rawAddress + PayloadOffset), payload);
}
void JIT::emitGetVarFromIndirectPointer(JSValue** operand, GPRReg tag, GPRReg payload)
{
    loadPtr(operand, payload);
    load32(Address(payload, TagOffset), tag);
    load32(Address(payload, PayloadOffset), payload);
}

void JIT::emitGetClosureVar(int scope, uintptr_t operand)
{
    emitLoad(scope, regT1, regT0);
    load32(Address(regT0, JSEnvironmentRecord::offsetOfVariables() + operand * sizeof(Register) + TagOffset), regT1);
    load32(Address(regT0, JSEnvironmentRecord::offsetOfVariables() + operand * sizeof(Register) + PayloadOffset), regT0);
}

void JIT::emit_op_get_from_scope(Instruction* currentInstruction)
{
	int dst = currentInstruction[1].u.operand;
	int scope = currentInstruction[2].u.operand;
    ResolveType resolveType = GetPutInfo(currentInstruction[4].u.operand).resolveType();
    Structure** structureSlot = currentInstruction[5].u.structure.slot();
    uintptr_t* operandSlot = reinterpret_cast<uintptr_t*>(&currentInstruction[6].u.pointer);

    class EmitCode {
    public:
    	EmitCode(JIT& jit, int scope, Structure** structureSlot, uintptr_t* operandSlot)
    	: m_jit(jit)
    	, scope(scope)
		, structureSlot(structureSlot)
    	, operandSlot(operandSlot)
		{
		}
    	void operator()(ResolveType resolveType, bool indirectLoadForOperand) {
			switch (resolveType) {
			case GlobalProperty:
			case GlobalPropertyWithVarInjectionChecks: {
				m_jit.emitLoadWithStructureCheck(scope, structureSlot); // Structure check covers var injection since we don't cache structures for anything but the GlobalObject. Additionally, resolve_scope handles checking for the var injection.
				GPRReg base = regT2;
				GPRReg resultTag = regT1;
				GPRReg resultPayload = regT0;
				GPRReg offset = regT3;

				m_jit.move(regT0, base);
				m_jit.load32(operandSlot, offset);
#if !ASSERT_DISABLED
				if (!ASSERT_DISABLED) {
					Jump isOutOfLine = m_jit.branch32(GreaterThanOrEqual, offset, TrustedImm32(firstOutOfLineOffset));
					m_jit.abortWithReason(JITOffsetIsNotOutOfLine);
					isOutOfLine.link(&m_jit);
				}
#endif
				m_jit.loadPtr(base, JSObject::butterflyOffset(), base);
				m_jit.neg32(offset);
				m_jit.load32(BaseIndex(base, offset, TimesEight, OBJECT_OFFSETOF(JSValue, u.asBits.payload) + (firstOutOfLineOffset - 2) * sizeof(EncodedJSValue)), resultPayload);
				m_jit.load32(BaseIndex(base, offset, TimesEight, OBJECT_OFFSETOF(JSValue, u.asBits.tag) + (firstOutOfLineOffset - 2) * sizeof(EncodedJSValue)), resultTag);
				break;
			}
			case GlobalVar:
			case GlobalVarWithVarInjectionChecks:
			case GlobalLexicalVar:
			case GlobalLexicalVarWithVarInjectionChecks:
				m_jit.emitVarInjectionCheck(needsVarInjectionChecks(resolveType));
				if (indirectLoadForOperand)
					m_jit.emitGetVarFromIndirectPointer(bitwise_cast<JSValue**>(operandSlot), regT1, regT0);
				else
					m_jit.emitGetVarFromPointer(bitwise_cast<JSValue*>(*operandSlot), regT1, regT0);
				if (resolveType == GlobalLexicalVar || resolveType == GlobalLexicalVarWithVarInjectionChecks) // TDZ check.
					m_jit.addSlowCase(m_jit.branch32WithConstant(Equal, regT1, JSValue::EmptyValueTag));
				break;
			case ClosureVar:
			case ClosureVarWithVarInjectionChecks:
				m_jit.emitVarInjectionCheck(needsVarInjectionChecks(resolveType));
				m_jit.emitGetClosureVar(scope, *operandSlot);
				break;
			case Dynamic:
				m_jit.addSlowCase(m_jit.jump());
				break;
		    case ModuleVar:
			case ResolvedClosureVar:
			case UnresolvedProperty:
			case UnresolvedPropertyWithVarInjectionChecks:
				RELEASE_ASSERT_NOT_REACHED();
			}
    	}
    private:
    	JIT& m_jit;
    	const int scope;
    	Structure** structureSlot;
    	uintptr_t* operandSlot;
    };

    EmitCode emitCode(*this, scope, structureSlot, operandSlot);
    switch (resolveType) {
    case GlobalProperty:
    case GlobalPropertyWithVarInjectionChecks: {
        JumpList skipToEnd;
        load32(&currentInstruction[4], regT0);
        and32(TrustedImm32(GetPutInfo::typeBits), regT0); // Load ResolveType into T0

        Jump isNotGlobalProperty = branch32(NotEqual, regT0, TrustedImm32(resolveType));
        emitCode(resolveType, false);
        skipToEnd.append(jump());

        isNotGlobalProperty.link(this);
        emitCode(needsVarInjectionChecks(resolveType) ? GlobalLexicalVarWithVarInjectionChecks : GlobalLexicalVar, true);
        skipToEnd.link(this);
        break;
    }
    case UnresolvedProperty:
    case UnresolvedPropertyWithVarInjectionChecks: {
        JumpList skipToEnd;
        load32(&currentInstruction[4], regT0);
        and32(TrustedImm32(GetPutInfo::typeBits), regT0); // Load ResolveType into T0

        Jump isGlobalProperty = branch32WithConstant(Equal, regT0, GlobalProperty);
        Jump notGlobalPropertyWithVarInjections = branch32WithConstant(NotEqual, regT0, GlobalPropertyWithVarInjectionChecks);
        isGlobalProperty.link(this);
        emitCode(GlobalProperty, false);
        skipToEnd.append(jump());
        notGlobalPropertyWithVarInjections.link(this);

        Jump notGlobalLexicalVar = branch32WithConstant(NotEqual, regT0, GlobalLexicalVar);
        emitCode(GlobalLexicalVar, true);
        skipToEnd.append(jump());
        notGlobalLexicalVar.link(this);

        Jump notGlobalLexicalVarWithVarInjections = branch32WithConstant(NotEqual, regT0, GlobalLexicalVarWithVarInjectionChecks);
        emitCode(GlobalLexicalVarWithVarInjectionChecks, true);
        skipToEnd.append(jump());
        notGlobalLexicalVarWithVarInjections.link(this);

        addSlowCase(jump());

        skipToEnd.link(this);
        break;
    }

    default:
        emitCode(resolveType, false);
        break;
    }
    emitValueProfilingSite(JSValueRegs(regT1, regT0));
    emitStore(dst, regT1, regT0);
}

void JIT::emitSlow_op_get_from_scope(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    linkAllSlowCases(iter);

	int dst = currentInstruction[1].u.operand;
    callOperation(WithProfile, operationGetFromScope, dst, currentInstruction);
}

void JIT::emitPutGlobalVariable(JSValue* operand, int value, WatchpointSet* set)
{
    emitLoad(value, regT1, regT0);
    emitNotifyWrite(set);
    uintptr_t rawAddress = bitwise_cast<uintptr_t>(operand);
    store32(regT1, bitwise_cast<void*>(rawAddress + TagOffset));
    store32(regT0, bitwise_cast<void*>(rawAddress + PayloadOffset));
}

void JIT::emitPutGlobalVariableIndirect(JSValue** addressOfOperand, int value, WatchpointSet** indirectWatchpointSet)
{
    emitLoad(value, regT1, regT0);
    loadPtr(indirectWatchpointSet, regT2);
    emitNotifyWrite(regT2);
    loadPtr(addressOfOperand, regT2);
    store32(regT1, Address(regT2, TagOffset));
    store32(regT0, Address(regT2, PayloadOffset));
}

void JIT::emitPutClosureVar(int scope, uintptr_t operand, int value, WatchpointSet* set)
{
    emitLoad(value, regT3, regT2);
    emitLoad(scope, regT1, regT0);
    emitNotifyWrite(set);
    store32(regT3, Address(regT0, JSEnvironmentRecord::offsetOfVariables() + operand * sizeof(Register) + TagOffset));
	store32(regT2, Address(regT0, JSEnvironmentRecord::offsetOfVariables() + operand * sizeof(Register) + PayloadOffset));
}

void JIT::emit_op_put_to_scope(Instruction* currentInstruction)
{
	int scope = currentInstruction[1].u.operand;
	int value = currentInstruction[3].u.operand;
    GetPutInfo getPutInfo = GetPutInfo(currentInstruction[4].u.operand);
    ResolveType resolveType = getPutInfo.resolveType();
    Structure** structureSlot = currentInstruction[5].u.structure.slot();
    uintptr_t* operandSlot = reinterpret_cast<uintptr_t*>(&currentInstruction[6].u.pointer);

    auto emitCode = [&] (ResolveType resolveType, bool indirectLoadForOperand) {
		switch (resolveType) {
		case GlobalProperty:
		case GlobalPropertyWithVarInjectionChecks: {
			emitLoadWithStructureCheck(scope, structureSlot); // Structure check covers var injection since we don't cache structures for anything but the GlobalObject. Additionally, resolve_scope handles checking for the var injection.
			loadPtr(regT0, JSObject::butterflyOffset(), regT0);
			loadPtr(operandSlot, regT1);
			negPtr(regT1);
			emitLoad(value, regT3, regT2);
			store32(regT3, BaseIndex(regT0, regT1, TimesEight, (firstOutOfLineOffset - 2) * sizeof(EncodedJSValue) + OBJECT_OFFSETOF(EncodedValueDescriptor, asBits.tag)));
			store32(regT2, BaseIndex(regT0, regT1, TimesEight, (firstOutOfLineOffset - 2) * sizeof(EncodedJSValue) + OBJECT_OFFSETOF(EncodedValueDescriptor, asBits.payload)));
			break;
		}
		case GlobalVar:
		case GlobalVarWithVarInjectionChecks:
		case GlobalLexicalVar:
		case GlobalLexicalVarWithVarInjectionChecks: {
			JSScope* constantScope = JSScope::constantScopeForCodeBlock(resolveType, codeBlock());
			RELEASE_ASSERT(constantScope);
			//emitWriteBarrier(constantScope, value, ShouldFilterValue);
			emitVarInjectionCheck(needsVarInjectionChecks(resolveType));
            emitVarReadOnlyCheck(resolveType);
			if (!isInitialization(getPutInfo.initializationMode()) && (resolveType == GlobalLexicalVar || resolveType == GlobalLexicalVarWithVarInjectionChecks)) {
				// We need to do a TDZ check here because we can't always prove we need to emit TDZ checks statically.
				if (indirectLoadForOperand)
					emitGetVarFromIndirectPointer(bitwise_cast<JSValue**>(operandSlot), regT1, regT0);
				else
					emitGetVarFromPointer(bitwise_cast<JSValue*>(*operandSlot), regT1, regT0);
				addSlowCase(branch32WithConstant(Equal, regT1, JSValue::EmptyValueTag));
			}
			if (indirectLoadForOperand)
				emitPutGlobalVariableIndirect(bitwise_cast<JSValue**>(operandSlot), value, bitwise_cast<WatchpointSet**>(&currentInstruction[5]));
			else
				emitPutGlobalVariable(bitwise_cast<JSValue*>(*operandSlot), value, currentInstruction[5].u.watchpointSet);
			break;
		}
		case ResolvedClosureVar:
		case ClosureVar:
		case ClosureVarWithVarInjectionChecks:
			emitVarInjectionCheck(needsVarInjectionChecks(resolveType));
			emitPutClosureVar(scope, *operandSlot, value, currentInstruction[5].u.watchpointSet);
			break;
		case ModuleVar:
		case Dynamic:
			addSlowCase(jump());
			break;
		case UnresolvedProperty:
		case UnresolvedPropertyWithVarInjectionChecks:
			RELEASE_ASSERT_NOT_REACHED();
		}
	};

    switch (resolveType) {
    case GlobalProperty:
    case GlobalPropertyWithVarInjectionChecks: {
        JumpList skipToEnd;
        load32(&currentInstruction[4], regT0);
        and32(TrustedImm32(GetPutInfo::typeBits), regT0); // Load ResolveType into T0

        Jump isGlobalProperty = branch32(Equal, regT0, TrustedImm32(resolveType));
        Jump isGlobalLexicalVar = branch32(Equal, regT0, TrustedImm32(needsVarInjectionChecks(resolveType) ? GlobalLexicalVarWithVarInjectionChecks : GlobalLexicalVar));
        addSlowCase(jump()); // Dynamic, it can happen if we attempt to put a value to already-initialized const binding.

        isGlobalLexicalVar.link(this);
        emitCode(needsVarInjectionChecks(resolveType) ? GlobalLexicalVarWithVarInjectionChecks : GlobalLexicalVar, true);
        skipToEnd.append(jump());

        isGlobalProperty.link(this);
        emitCode(resolveType, false);
        skipToEnd.link(this);
        break;
    }
    case UnresolvedProperty:
    case UnresolvedPropertyWithVarInjectionChecks: {
        JumpList skipToEnd;
        load32(&currentInstruction[4], regT0);
        and32(TrustedImm32(GetPutInfo::typeBits), regT0); // Load ResolveType into T0

        Jump isGlobalProperty = branch32WithConstant(Equal, regT0, GlobalProperty);
        Jump notGlobalPropertyWithVarInjections = branch32WithConstant(NotEqual, regT0, GlobalPropertyWithVarInjectionChecks);
        isGlobalProperty.link(this);
        emitCode(GlobalProperty, false);
        skipToEnd.append(jump());
        notGlobalPropertyWithVarInjections.link(this);

        Jump notGlobalLexicalVar = branch32WithConstant(NotEqual, regT0, GlobalLexicalVar);
        emitCode(GlobalLexicalVar, true);
        skipToEnd.append(jump());
        notGlobalLexicalVar.link(this);

        Jump notGlobalLexicalVarWithVarInjections = branch32WithConstant(NotEqual, regT0, GlobalLexicalVarWithVarInjectionChecks);
        emitCode(GlobalLexicalVarWithVarInjectionChecks, true);
        skipToEnd.append(jump());
        notGlobalLexicalVarWithVarInjections.link(this);

        addSlowCase(jump());

        skipToEnd.link(this);
        break;
    }

    default:
        emitCode(resolveType, false);
        break;
    }
}

void JIT::emitSlow_op_put_to_scope(Instruction* currentInstruction, Vector<SlowCaseEntry>::iterator& iter)
{
    linkAllSlowCases(iter);

    GetPutInfo getPutInfo = GetPutInfo(currentInstruction[4].u.operand);
    ResolveType resolveType = getPutInfo.resolveType();
    if (resolveType == ModuleVar) {
        JITSlowPathCall slowPathCall(this, currentInstruction, slow_path_throw_strict_mode_readonly_property_write_error);
        slowPathCall.call();
    } else
    	callOperation(operationPutToScope, currentInstruction);
}

void JIT::emit_op_get_from_arguments(Instruction* currentInstruction)
{
    int dst = currentInstruction[1].u.operand;
    int arguments = currentInstruction[2].u.operand;
    int index = currentInstruction[3].u.operand;

    JSValueRegs resutlRegs = JSValueRegs(regT1, regT0);

    emitLoadPayload(arguments, regT0);
    load32(Address(regT0, DirectArguments::storageOffset() + index * sizeof(WriteBarrier<Unknown>) + TagOffset), resutlRegs.tagGPR());
    load32(Address(regT0, DirectArguments::storageOffset() + index * sizeof(WriteBarrier<Unknown>) + PayloadOffset), resutlRegs.payloadGPR());
    emitValueProfilingSite(resutlRegs);
    emitStore(dst, resutlRegs.tagGPR(), resutlRegs.payloadGPR());
}

void JIT::emit_op_put_to_arguments(Instruction* currentInstruction)
{
    int arguments = currentInstruction[1].u.operand;
    int index = currentInstruction[2].u.operand;
    int value = currentInstruction[3].u.operand;

    //if (Heap::isWriteBarrierEnabled())
    //	emitWriteBarrier(arguments, value, ShouldFilterImmediates);

    emitLoadPayload(arguments, regT0);
    emitLoad(value, regT1, regT2);
    store32(regT1, Address(regT0, DirectArguments::storageOffset() + index * sizeof(WriteBarrier<Unknown>) + TagOffset));
    store32(regT2, Address(regT0, DirectArguments::storageOffset() + index * sizeof(WriteBarrier<Unknown>) + PayloadOffset));
}

} // namespace JSC

#endif // USE(JSVALUE32_64)
// #endif // ENABLE(JIT)
