/*
 * Copyright (C) 2009, 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "ArrayBufferSharingMode.h"
#include "GCIncomingRefCounted.h"
#include "Watchpoint.h"
#include "Weak.h"
#include <wtf/PassRefPtr.h>
#include <wtf/StdLibExtras.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace JSC {

#define MAX_ARRAY_BUFFER_SIZE 0x7fffffffu

class ArrayBuffer;
class ArrayBufferView;
class JSArrayBuffer;

class ArrayBufferContents {
    WTF_MAKE_NONCOPYABLE(ArrayBufferContents);
public:
    ALWAYS_INLINE ArrayBufferContents()
        : m_shared(false)
    	, m_data(0)
        , m_sizeInBytes(0)
    { }

    inline ~ArrayBufferContents();
    
    void* data() { return m_data; }
    unsigned sizeInBytes() { return m_sizeInBytes; }

    bool isShared() const { return m_shared; }

private:
    ALWAYS_INLINE ArrayBufferContents(void* data, unsigned sizeInBytes)
        : m_shared(false)
    	, m_data(data)
        , m_sizeInBytes(sizeInBytes)
    { }

    friend class ArrayBuffer;

    enum InitializationPolicy {
        ZeroInitialize,
        DontInitialize
    };

    static inline void tryAllocate(unsigned numElements, unsigned elementByteSize, InitializationPolicy, ArrayBufferContents&);
    ALWAYS_INLINE void transferTo(ArrayBufferContents& other)
    {
        ASSERT(!other.m_data);
        other.m_data = m_data;
        other.m_sizeInBytes = m_sizeInBytes;
        m_data = 0;
        m_sizeInBytes = 0;
    }

    void copyTo(ArrayBufferContents& other)
    {
        ASSERT(!other.m_data);
        ArrayBufferContents::tryAllocate(m_sizeInBytes, sizeof(char), ArrayBufferContents::DontInitialize, other);
        if (!other.m_data)
            return;
        memcpy(other.m_data, m_data, m_sizeInBytes);
        other.m_sizeInBytes = m_sizeInBytes;
    }

    bool m_shared;
    void* m_data;
    unsigned m_sizeInBytes;
};

class ArrayBuffer : public GCIncomingRefCounted<ArrayBuffer> {
public:
    static inline PassRefPtr<ArrayBuffer> create(unsigned numElements, unsigned elementByteSize);
    static inline PassRefPtr<ArrayBuffer> create(const void* source, unsigned byteLength);
    static inline Ref<ArrayBuffer> create(ArrayBufferContents&);
    static inline Ref<ArrayBuffer> createAdopted(const void* data, unsigned byteLength);

    static inline PassRefPtr<ArrayBuffer> tryCreate(ArrayBuffer*);
    static inline PassRefPtr<ArrayBuffer> tryCreate(const void* source, unsigned byteLength);
    static inline PassRefPtr<ArrayBuffer> tryCreate(unsigned numElements, unsigned elementByteSize);

    // Only for use by Uint8ClampedArray::createUninitialized.
    static inline PassRefPtr<ArrayBuffer> createUninitialized(unsigned numElements, unsigned elementByteSize);

    inline void* data();
    inline const void* data() const;
    inline unsigned byteLength() const;

    inline bool isShared() const;
    
    ArrayBufferSharingMode sharingMode() const
    {
    	return isShared() ? ArrayBufferSharingMode::Shared : ArrayBufferSharingMode::Default;
    }

    inline size_t gcSizeEstimateInBytes() const;

    inline PassRefPtr<ArrayBuffer> slice(double begin, double end) const;
    inline PassRefPtr<ArrayBuffer> slice(double begin) const;
    JS_EXPORT_PRIVATE PassRefPtr<ArrayBuffer> sliceWithClampedIndex(unsigned begin, unsigned end) const;

    inline void pin();
    inline void unpin();
    inline void pinAndLock();
    inline bool isLocked();

    void makeWasmMemory();
    inline bool isWasmMemory();

    JS_EXPORT_PRIVATE bool transferTo(VM&, ArrayBufferContents&);

    void neuter(VM&);
    ALWAYS_INLINE bool isDetached() const { return !m_contents.m_data; }
    ALWAYS_INLINE bool isNeutered() const { return !m_contents.m_data; }
    InlineWatchpointSet& neuteringWatchpointSet() { return m_neuteringWatchpointSet; }

    static ptrdiff_t offsetOfData() { return OBJECT_OFFSETOF(ArrayBuffer, m_contents) + OBJECT_OFFSETOF(ArrayBufferContents, m_data); }

    ~ArrayBuffer() { }

private:
    static inline PassRefPtr<ArrayBuffer> create(unsigned numElements, unsigned elementByteSize, ArrayBufferContents::InitializationPolicy);
    static inline PassRefPtr<ArrayBuffer> tryCreate(unsigned numElements, unsigned elementByteSize, ArrayBufferContents::InitializationPolicy);

    inline ArrayBuffer(ArrayBufferContents&);
    inline unsigned clampIndex(double index) const;
    static inline unsigned clampValue(double x, unsigned left, unsigned right);

    void notifyNeutering(VM&);

    ArrayBufferContents m_contents;
    InlineWatchpointSet m_neuteringWatchpointSet;
public:
    Weak<JSArrayBuffer> m_wrapper;
private:
    unsigned m_pinCount : 30;
    bool m_isWasmMemory : 1;
    // m_locked == true means that some API user fetched m_contents directly from a TypedArray object,
    // the buffer is backed by a WebAssembly.Memory, or is a SharedArrayBuffer.
    bool m_locked : 1;
};

ALWAYS_INLINE PassRefPtr<ArrayBuffer> ArrayBuffer::create(unsigned numElements, unsigned elementByteSize)
{
    return create(numElements, elementByteSize, ArrayBufferContents::ZeroInitialize);
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::tryCreate(unsigned numElements, unsigned elementByteSize)
{
    return tryCreate(numElements, elementByteSize, ArrayBufferContents::ZeroInitialize);
}

ALWAYS_INLINE PassRefPtr<ArrayBuffer> ArrayBuffer::tryCreate(ArrayBuffer* other)
{
    return ArrayBuffer::tryCreate(other->data(), other->byteLength());
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::create(const void* source, unsigned byteLength)
{
    ArrayBufferContents contents;
    ArrayBufferContents::tryAllocate(byteLength, 1, ArrayBufferContents::ZeroInitialize, contents);
    if (!contents.m_data)
        return nullptr;
    RefPtr<ArrayBuffer> buffer = adoptRef(new ArrayBuffer(contents));
    ASSERT(!byteLength || source);
    memcpy(buffer->data(), source, byteLength);
    return buffer.release();
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::tryCreate(const void* source, unsigned byteLength)
{
    ArrayBufferContents contents;
    ArrayBufferContents::tryAllocate(byteLength, 1, ArrayBufferContents::ZeroInitialize, contents);
    if (!contents.m_data)
        return nullptr;
    RefPtr<ArrayBuffer> buffer = adoptRef(new ArrayBuffer(contents));
    ASSERT(!byteLength || source);
    memcpy(buffer->data(), source, byteLength);
    return buffer.release();
}

ALWAYS_INLINE Ref<ArrayBuffer> ArrayBuffer::create(ArrayBufferContents& contents)
{
    return adoptRef(*new ArrayBuffer(contents));
}

ALWAYS_INLINE Ref<ArrayBuffer> ArrayBuffer::createAdopted(const void* data, unsigned byteLength)
{
    ArrayBufferContents contents(const_cast<void*>(data), byteLength);
    return create(contents);
}

ALWAYS_INLINE PassRefPtr<ArrayBuffer> ArrayBuffer::createUninitialized(unsigned numElements, unsigned elementByteSize)
{
    return create(numElements, elementByteSize, ArrayBufferContents::DontInitialize);
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::create(unsigned numElements, unsigned elementByteSize, ArrayBufferContents::InitializationPolicy policy)
{
    ArrayBufferContents contents;
    ArrayBufferContents::tryAllocate(numElements, elementByteSize, policy, contents);
    if (!contents.m_data)
        return nullptr;
    return adoptRef(new ArrayBuffer(contents));
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::tryCreate(unsigned numElements, unsigned elementByteSize, ArrayBufferContents::InitializationPolicy policy)
{
    ArrayBufferContents contents;
    ArrayBufferContents::tryAllocate(numElements, elementByteSize, policy, contents);
    if (!contents.m_data)
        return nullptr;
    return adoptRef(new ArrayBuffer(contents));
}

ALWAYS_INLINE ArrayBuffer::ArrayBuffer(ArrayBufferContents& contents)
    : m_neuteringWatchpointSet(IsWatched)
	, m_pinCount(0)
	, m_isWasmMemory(false)
	, m_locked(false)
{
    contents.transferTo(m_contents);
}

ALWAYS_INLINE void* ArrayBuffer::data()
{
    return m_contents.m_data;
}

ALWAYS_INLINE const void* ArrayBuffer::data() const
{
    return m_contents.m_data;
}

ALWAYS_INLINE unsigned ArrayBuffer::byteLength() const
{
    return m_contents.m_sizeInBytes;
}

ALWAYS_INLINE bool ArrayBuffer::isShared() const
{
    return m_contents.isShared();
}

ALWAYS_INLINE size_t ArrayBuffer::gcSizeEstimateInBytes() const
{
    return sizeof(ArrayBuffer) + static_cast<size_t>(byteLength());
}

inline unsigned ArrayBuffer::clampValue(double x, unsigned left, unsigned right)
{
    ASSERT(left <= right);
    if (x < left)
        x = left;
    if (right < x)
        x = right;
    return x;
}

inline unsigned ArrayBuffer::clampIndex(double index) const
{
    unsigned currentLength = byteLength();
    if (index < 0)
        index = currentLength + index;
    return clampValue(index, 0, currentLength);
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::slice(double begin, double end) const
{
    return sliceWithClampedIndex(clampIndex(begin), clampIndex(end));
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::slice(double begin) const
{
    return sliceWithClampedIndex(clampIndex(begin), byteLength());
}

inline PassRefPtr<ArrayBuffer> ArrayBuffer::sliceWithClampedIndex(unsigned begin, unsigned end) const
{
    unsigned size = begin <= end ? end - begin : 0;
    RefPtr<ArrayBuffer> result = ArrayBuffer::tryCreate(static_cast<const char*>(data()) + begin, size);
#if ENABLE(SHARED_ARRAY_BUFFER)
    if (result)
		result->setSharingMode(sharingMode());
#endif
    return result.release();
}

ALWAYS_INLINE void ArrayBuffer::pin()
{
    m_pinCount++;
}

ALWAYS_INLINE void ArrayBuffer::unpin()
{
    m_pinCount--;
}

ALWAYS_INLINE void ArrayBuffer::pinAndLock()
{
    m_locked = true;
}

ALWAYS_INLINE bool ArrayBuffer::isLocked()
{
    return m_locked;
}

ALWAYS_INLINE bool ArrayBuffer::isWasmMemory()
{
    return m_isWasmMemory;
}

JS_EXPORT_PRIVATE ASCIILiteral errorMesasgeForTransfer(ArrayBuffer*);

inline void ArrayBufferContents::tryAllocate(unsigned numElements, unsigned elementByteSize, ArrayBufferContents::InitializationPolicy policy, ArrayBufferContents& result)
{
    CheckedSize sizeInBytes = numElements;
    sizeInBytes *= elementByteSize;
    if (sizeInBytes.hasOverflowed() || sizeInBytes.unsafeGet() > MAX_ARRAY_BUFFER_SIZE) {
        result.m_data = nullptr;
        return;
    }

    size_t allocationSize = sizeInBytes.unsafeGet();
    // Make sure malloc actually allocates something, but not too much. We use null to mean that the buffer is detached.
    if (!allocationSize)
        allocationSize = sizeof(EncodedJSValue);

    WTF::tryFastMalloc(allocationSize).getValue(result.m_data);
    if (!result.m_data)
		return;

    if (policy == ZeroInitialize)
        memset(result.m_data, 0, allocationSize);

    result.m_sizeInBytes = sizeInBytes.unsafeGet();
}

ALWAYS_INLINE ArrayBufferContents::~ArrayBufferContents()
{
	if (m_data)
		WTF::fastFree(m_data);
}

} // namespace JSC

using JSC::ArrayBuffer;
