/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2007, 2008, 2009, 2011 Apple Inc. All rights reserved.
 *  Copyright (C) 2003 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2006 Alexey Proskuryakov (ap@nypop.com)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
 *  USA
 *
 */

#include "config.h"
#include "ArrayPrototype.h"

#include "AdaptiveInferredPropertyValueWatchpointBase.h"
#include "ArrayConstructor.h"
#include "BuiltinNames.h"
#include "ButterflyInlineMethods.h"
#include "CachedCall.h"
#include "CodeBlock.h"
#include "CopiedSpaceInlineMethods.h"
#include "Error.h"
#include "GetterSetter.h"
#include "Interpreter.h"
#include "JSArrayInlines.h"
#include "JSArrayIterator.h"
#include "JSCBuiltins.h"
#include "JSStringJoiner.h"
#include "Lookup.h"
#include "ObjectConstructor.h"
#include "ObjectPrototype.h"
#include "Operations.h"
#include "StringRecursionChecker.h"
#include <algorithm>
#include <wtf/Assertions.h>
#include <wtf/HashSet.h>
#include <wtf/FastMallocMonitor.h>
#include <wtf/MemoryMonitor.h>

namespace JSC {

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncToLocaleString(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncJoin(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncPop(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncPush(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncReverse(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncShift(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncSlice(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncSplice(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncUnShift(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncIndexOf(ExecState*);
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncLastIndexOf(ExecState*);

// ------------------------------ ArrayPrototype ----------------------------

const ClassInfo ArrayPrototype::s_info = {"Array", &JSArray::s_info, 0, CREATE_METHOD_TABLE(ArrayPrototype)};

ArrayPrototype* ArrayPrototype::create(VM& vm, JSGlobalObject* globalObject, Structure* structure)
{
	ArrayPrototype* prototype = new (NotNull, allocateCell<ArrayPrototype>(vm.heap)) ArrayPrototype(vm, structure);
    prototype->finishCreation(vm, globalObject);
    return prototype;
}

// ECMA 15.4.4
inline ArrayPrototype::ArrayPrototype(VM& vm, Structure* structure)
    : JSArray(vm, structure, 0)
{
}

inline void ArrayPrototype::finishCreation(VM& vm, JSGlobalObject* globalObject)
{
    Base::finishCreation(vm);
    ASSERT(inherits(info()));
    vm.prototypeMap.addPrototype(this);

    putDirectWithoutTransition(vm, vm.propertyNames->toString, globalObject->arrayProtoToStringFunction(), DontEnum);
	putDirectWithoutTransition(vm, vm.propertyNames->builtinNames().valuesPublicName(), globalObject->arrayProtoValuesFunction(), DontEnum);
    putDirectWithoutTransition(vm, vm.propertyNames->iteratorSymbol, globalObject->arrayProtoValuesFunction(), DontEnum);

    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->toLocaleString, arrayProtoFuncToLocaleString, DontEnum, 0);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION("concat", arrayPrototypeConcatCodeGenerator, DontEnum);
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION("fill", arrayPrototypeFillCodeGenerator, DontEnum);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->join, arrayProtoFuncJoin, DontEnum, 1);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION("pop", arrayProtoFuncPop, DontEnum, 0, ArrayPopIntrinsic);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().pushPublicName(), arrayProtoFuncPush, DontEnum, 1, ArrayPushIntrinsic);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().pushPrivateName(), arrayProtoFuncPush, DontEnum | DontDelete | ReadOnly, 1, ArrayPushIntrinsic);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("reverse", arrayProtoFuncReverse, DontEnum, 0);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().shiftPublicName(), arrayProtoFuncShift, DontEnum, 0);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().shiftPrivateName(), arrayProtoFuncShift, DontEnum | DontDelete | ReadOnly, 0);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->slice, arrayProtoFuncSlice, DontEnum, 2, ArraySliceIntrinsic);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().sortPublicName(), arrayPrototypeSortCodeGenerator, DontEnum);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("splice", arrayProtoFuncSplice, DontEnum, 2);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("unshift", arrayProtoFuncUnShift, DontEnum, 1);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().everyPublicName(), arrayPrototypeEveryCodeGenerator, DontEnum);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().forEachPublicName(), arrayPrototypeForEachCodeGenerator, DontEnum);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().somePublicName(), arrayPrototypeSomeCodeGenerator, DontEnum);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->indexOf, arrayProtoFuncIndexOf, DontEnum, 1, ArrayIndexOfIntrinsic);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->lastIndexOf, arrayProtoFuncLastIndexOf, DontEnum, 1);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().filterPublicName(), arrayPrototypeFilterCodeGenerator, DontEnum);
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().flatPublicName(), arrayPrototypeFlatCodeGenerator, DontEnum);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().flatMapPublicName(), arrayPrototypeFlatMapCodeGenerator, DontEnum);
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().reducePublicName(), arrayPrototypeReduceCodeGenerator, DontEnum);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().reduceRightPublicName(), arrayPrototypeReduceRightCodeGenerator, DontEnum);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->map, arrayPrototypeMapCodeGenerator, DontEnum);
    JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().entriesPublicName(), arrayPrototypeEntriesCodeGenerator, DontEnum);
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().keysPublicName(), arrayPrototypeKeysCodeGenerator, DontEnum);
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().findPublicName(), arrayPrototypeFindCodeGenerator, DontEnum);
	if (Options::useArrayFindLastMethod())
		JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().findLastPublicName(), arrayPrototypeFindLastCodeGenerator, static_cast<unsigned>(PropertyAttribute::DontEnum));
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().findIndexPublicName(), arrayPrototypeFindIndexCodeGenerator, DontEnum);
	if (Options::useArrayFindLastMethod())
		JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().findLastIndexPublicName(), arrayPrototypeFindLastIndexCodeGenerator, static_cast<unsigned>(PropertyAttribute::DontEnum));
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().includesPublicName(), arrayPrototypeIncludesCodeGenerator, DontEnum);
	JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().copyWithinPublicName(), arrayPrototypeCopyWithinCodeGenerator, DontEnum);

    if (Options::useAtMethod())
        JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().atPublicName(), arrayPrototypeAtCodeGenerator, static_cast<unsigned>(PropertyAttribute::DontEnum));

    putDirectWithoutTransition(vm, vm.propertyNames->builtinNames().entriesPrivateName(), getDirect(vm, vm.propertyNames->builtinNames().entriesPublicName()), ReadOnly);
    putDirectWithoutTransition(vm, vm.propertyNames->builtinNames().forEachPrivateName(), getDirect(vm, vm.propertyNames->builtinNames().forEachPublicName()), ReadOnly);
    putDirectWithoutTransition(vm, vm.propertyNames->builtinNames().keysPrivateName(), getDirect(vm, vm.propertyNames->builtinNames().keysPublicName()), ReadOnly);
    putDirectWithoutTransition(vm, vm.propertyNames->builtinNames().valuesPrivateName(), globalObject->arrayProtoValuesFunction(), ReadOnly);

    {
        JSObject* unscopables = constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure());
        static const ASCIILiteral unscopableNames[] = {
			ASCIILiteral("copyWithin"),
			ASCIILiteral("entries"),
			ASCIILiteral("fill"),
			ASCIILiteral("find"),
			ASCIILiteral("findIndex"),
			ASCIILiteral("flat"),
			ASCIILiteral("flatMap"),
			ASCIILiteral("includes"),
			ASCIILiteral("keys"),
			ASCIILiteral("values")
        };
        if (Options::useArrayFindLastMethod()) {
            unscopables->putDirect(vm, vm.propertyNames->builtinNames().findLastPublicName(), jsBoolean(true));
            unscopables->putDirect(vm, vm.propertyNames->builtinNames().findLastIndexPublicName(), jsBoolean(true));
        }
        if (Options::useAtMethod())
			unscopables->putDirect(vm, vm.propertyNames->builtinNames().atPublicName(), jsBoolean(true));
        for(int i = 0; i < (sizeof(unscopableNames) / sizeof(ASCIILiteral)); ++i)
            unscopables->putDirect(vm, Identifier::fromString(vm, unscopableNames[i]), jsBoolean(true));
        putDirectWithoutTransition(vm, vm.propertyNames->unscopablesSymbol, unscopables, DontEnum | ReadOnly);
    }
}

// ------------------------------ Array Functions ----------------------------

static ALWAYS_INLINE JSValue getProperty(VM& vm, ExecState* exec, JSObject* object, uint64_t index)
{
    if (JSValue result = object->tryGetIndexQuickly(index))
        return result;
    // We want to perform get and has in the same operation.
    // We can only do so when this behavior is not observable. The
    // only time it is observable is when we encounter an opaque objects (ProxyObject and JSModuleNamespaceObject)
	// somewhere in the prototype chain.
    PropertySlot slot(object, PropertySlot::InternalMethodType::HasProperty);
    bool hasProperty = object->getPropertySlot(vm, exec, index, slot);
	ASSERT(!scope.exception() || !hasProperty);
	if (!hasProperty)
		return JSValue();
    if (UNLIKELY(slot.isTaintedByOpaqueObject()))
        return object->get(vm, exec, index);
    return slot.getValue(vm, exec, index);
}

static ALWAYS_INLINE void setLength(ExecState* exec, VM& vm, JSObject* obj, uint64_t value)
{
    auto scope = DECLARE_THROW_SCOPE(vm);
    static const bool throwException = true;
    if (LIKELY(isJSArray(obj))) {
        if (UNLIKELY(value > UINT32_MAX)) {
            throwRangeError(exec, vm, ASCIILiteral("Invalid array length"));
            return;
        }
        scope.release();
        jsCast<JSArray*>(obj)->setLength(exec, static_cast<uint32_t>(value), throwException);
        return;
	}
	scope.release();
	PutPropertySlot slot(obj, throwException);
	obj->methodTable(vm)->put(obj, exec, vm.propertyNames->length, jsNumber(value), slot);
}

namespace ArrayPrototypeInternal {
static bool verbose = false;
}

static ALWAYS_INLINE bool speciesWatchpointIsValid(ExecState* exec, JSObject* thisObject)
{
    JSGlobalObject* globalObject = thisObject->globalObject();
    ArrayPrototype* arrayPrototype = globalObject->arrayPrototype();

    if (globalObject->arraySpeciesWatchpointSet().stateOnJSThread() == ClearWatchpoint) {
        //dataLogLnIf(ArrayPrototypeInternal::verbose, "Initializing Array species watchpoints for Array.prototype: ", pointerDump(arrayPrototype), " with structure: ", pointerDump(arrayPrototype->structure(vm)), "\nand Array: ", pointerDump(globalObject->arrayConstructor()), " with structure: ", pointerDump(globalObject->arrayConstructor()->structure(vm)));
        globalObject->tryInstallArraySpeciesWatchpoint(exec);
        ASSERT(globalObject->arraySpeciesWatchpointSet().stateOnJSThread() != ClearWatchpoint);
    }

    return !thisObject->hasCustomProperties()
        && arrayPrototype == thisObject->getPrototypeDirect()
        && globalObject->arraySpeciesWatchpointSet().stateOnJSThread() == IsWatched;
}

enum class SpeciesConstructResult {
    FastPath,
    Exception,
    CreatedObject
};

static ALWAYS_INLINE std::pair<SpeciesConstructResult, JSObject*> speciesConstructArray(VM& vm, ExecState* exec, JSObject* thisObject, uint64_t length)
{
    auto exceptionResult = [] () {
        return std::make_pair(SpeciesConstructResult::Exception, static_cast<JSObject*>(0));
    };

    // ECMA 9.4.2.3: https://tc39.github.io/ecma262/#sec-arrayspeciescreate
    JSValue constructor = jsUndefined();
    bool thisIsArray = isArray(exec, thisObject);
    RETURN_IF_EXCEPTION(vm, std::make_pair(SpeciesConstructResult::Exception, static_cast<JSObject*>(0)));
    if (LIKELY(thisIsArray)) {
    	// Fast path in the normal case where the user has not set an own constructor and the Array.prototype.constructor is normal.
		// We need prototype check for subclasses of Array, which are Array objects but have a different prototype by default.
    	bool isValid = speciesWatchpointIsValid(exec, thisObject);
		if (LIKELY(isValid))
            return std::make_pair(SpeciesConstructResult::FastPath, static_cast<JSObject*>(0));

        constructor = thisObject->get(vm, exec, vm.propertyNames->constructor);
        RETURN_IF_EXCEPTION(vm, std::make_pair(SpeciesConstructResult::Exception, static_cast<JSObject*>(0)));
        if (constructor.isConstructor()) {
            JSObject* constructorObject = jsCast<JSObject*>(constructor);
            bool isArrayConstructorFromAnotherRealm = exec->lexicalGlobalObject() != constructorObject->globalObject()
                && constructorObject->inherits<ArrayConstructor>(vm);
            if (isArrayConstructorFromAnotherRealm)
                return std::make_pair(SpeciesConstructResult::FastPath, static_cast<JSObject*>(0));
        }
        if (constructor.isObject()) {
            constructor = constructor.get(vm, exec, vm.propertyNames->speciesSymbol);
            RETURN_IF_EXCEPTION(vm, std::make_pair(SpeciesConstructResult::Exception, static_cast<JSObject*>(0)));
            if (constructor.isNull())
                return std::make_pair(SpeciesConstructResult::FastPath, static_cast<JSObject*>(0));
        }
    } else {
        // If isArray is false, return ? ArrayCreate(length).
        return std::make_pair(SpeciesConstructResult::FastPath, static_cast<JSObject*>(0));
    }

    if (constructor.isUndefined())
        return std::make_pair(SpeciesConstructResult::FastPath, static_cast<JSObject*>(0));

    MarkedArgumentBuffer args;
    args.append(jsNumber(length));
    JSObject* newObject = construct(exec, constructor, args, ASCIILiteral("Species construction did not get a valid constructor"));
    RETURN_IF_EXCEPTION(vm, std::make_pair(SpeciesConstructResult::Exception, static_cast<JSObject*>(0)));
    return std::make_pair(SpeciesConstructResult::CreatedObject, newObject);
}

EncodedJSValue JSC_HOST_CALL arrayProtoFuncSpeciesCreate(ExecState* exec)
{
    VM& vm = exec->vm();
    ThrowScope scope(vm);
    JSObject* object = asObject(exec->uncheckedArgument(0));
    uint64_t length = static_cast<uint64_t>(exec->uncheckedArgument(1).asNumber());

    std::pair<SpeciesConstructResult, JSObject*> speciesResult = speciesConstructArray(vm, exec, object, length);
    EXCEPTION_ASSERT(!!scope.exception() == (speciesResult.first == SpeciesConstructResult::Exception));
    if (UNLIKELY(speciesResult.first == SpeciesConstructResult::Exception))
        return encodedJSUndefined();
    if (speciesResult.first == SpeciesConstructResult::CreatedObject)
        return JSValue::encode(speciesResult.second);

    if (length > std::numeric_limits<unsigned>::max()) {
        throwRangeError(exec, vm, ASCIILiteral("Array size is not a small enough positive integer."));
        return encodedJSUndefined();
    }

    RELEASE_AND_RETURN(scope, JSValue::encode(constructEmptyArray(exec, 0, static_cast<unsigned>(length))));
}

static inline uint64_t argumentClampedIndexFromStartOrEnd(ExecState* exec, JSValue value, uint64_t length, uint64_t undefinedValue = 0)
{
    if (value.isUndefined())
        return undefinedValue;

    double indexDouble = value.toIntegerOrInfinity(exec);
    if (indexDouble < 0) {
        indexDouble += length;
        return indexDouble < 0 ? 0 : static_cast<uint64_t>(indexDouble);
    }
    return indexDouble > length ? length : static_cast<uint64_t>(indexDouble);
}

// The shift/unshift function implement the shift/unshift behaviour required
// by the corresponding array prototype methods, and by splice. In both cases,
// the methods are operating an an array or array like object.
//
//  header  currentCount  (remainder)
// [------][------------][-----------]
//  header  resultCount  (remainder)
// [------][-----------][-----------]
//
// The set of properties in the range 'header' must be unchanged. The set of
// properties in the range 'remainder' (where remainder = length - header -
// currentCount) will be shifted to the left or right as appropriate; in the
// case of shift this must be removing values, in the case of unshift this
// must be introducing new values.

template<JSArray::ShiftCountMode shiftCountMode>
static void shift(ExecState* exec, JSObject* thisObj, uint64_t header, uint64_t currentCount, uint64_t resultCount, uint64_t length)
{
    ASSERT(currentCount > resultCount);
    uint64_t count = currentCount - resultCount;

    ASSERT(header <= length);
    ASSERT(currentCount <= (length - header));

    if (isJSArray(thisObj)) {
        JSArray* array = asArray(thisObj);
        uint32_t header32 = header;
        ASSERT(header32 == header);
        if (array->length() == length && array->shiftCount<shiftCountMode>(exec, header32, count))
            return;
        header = header32;
    }

    VM& vm = exec->vm();
    for (uint64_t k = header; k < length - currentCount; ++k) {
    	uint64_t from = k + currentCount;
    	uint64_t to = k + resultCount;
        JSValue value = getProperty(vm, exec, thisObj, from);
        RETURN_IF_EXCEPTION(vm, void());
		if(value) {
            thisObj->putByIndexInline(exec, vm, to, value, true);
            RETURN_IF_EXCEPTION(vm, void());
        } else {
        	bool success = thisObj->deleteProperty(exec, to);
            RETURN_IF_EXCEPTION(vm, void());
        	if (!success) {
        		throwTypeError(exec, vm, "Unable to delete property.");
        		return;
        	}
        }
    }
    for (uint64_t k = length; k > length - count; --k) {
    	bool success = thisObj->deleteProperty(exec, k - 1);
        RETURN_IF_EXCEPTION(vm, void());
        if (UNLIKELY(!success)) {
            throwTypeError(exec, vm, "Unable to delete property.");
            return;
        }
    }
}

static void unshift(ExecState* exec, JSObject* thisObj, uint64_t header, uint64_t currentCount, uint64_t resultCount, uint64_t length)
{
    ASSERT(header <= maxSafeInteger());
    ASSERT(currentCount <= maxSafeInteger());
    ASSERT(resultCount <= maxSafeInteger());
    ASSERT(length <= maxSafeInteger());

    VM& vm = exec->vm();

    ASSERT(resultCount > currentCount);
    uint64_t count = resultCount - currentCount;

    ASSERT(header <= length);
    ASSERT(currentCount <= (length - header));

    if (isJSArray(thisObj)) {
        // Spec says if we would produce an array of this size, we must throw a range error.
        if (count + length > std::numeric_limits<uint32_t>::max()) {
            throwRangeError(exec, vm, LengthExceededTheMaximumArrayLengthError);
            return;
        }

        JSArray* array = asArray(thisObj);
        if (array->length() == length) {
        	bool handled = array->unshiftCount(exec, static_cast<uint32_t>(header), static_cast<uint32_t>(count));
            RETURN_IF_EXCEPTION(vm, void());
        	ASSERT(!scope.exception() || handled);
			if (handled)
				return;
		}
    }

    for (uint64_t k = length - currentCount; k > header; --k) {
    	uint64_t from = k + currentCount - 1;
    	uint64_t to = k + resultCount - 1;
        JSValue value = getProperty(vm, exec, thisObj, from);
        RETURN_IF_EXCEPTION(vm, void());
		if(value) {
            thisObj->putByIndexInline(exec, vm, to, value, true);
            RETURN_IF_EXCEPTION(vm, void());
        } else {
        	bool success = thisObj->deleteProperty(exec, to);
            RETURN_IF_EXCEPTION(vm, void());
            if (UNLIKELY(!success)) {
				throwTypeError(exec, vm, "Unable to delete property.");
				return;
            }
        }
    }
}

static bool canUseFastJoin(const JSObject*);
static JSValue fastJoin(ExecState&, VM& vm, JSObject*, const StringView&, unsigned);

static inline bool canUseDefaultArrayJoinForToString(VM& vm, JSObject* thisObject)
{
    JSGlobalObject* globalObject = thisObject->globalObject();

    if (globalObject->arrayJoinWatchpointSet().stateOnJSThread() != IsWatched)
        return false;

    Structure* structure = thisObject->structure(vm);

    // This is the fast case. Many arrays will be an original array.
    // We are doing very simple check here. If we do more complicated checks like looking into getDirect "join" of thisObject,
    // it would be possible that just looking into "join" function will show the same performance.
    return globalObject->isOriginalArrayStructure(structure);
}

EncodedJSValue JSC_HOST_CALL arrayProtoFuncToString(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    JSValue thisValue = exec->hostThisValue().toThis(exec, ECMAMode::strict());

    // 1. Let array be the result of calling ToObject on the this value.
    JSObject* thisObject = thisValue.toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (!canUseDefaultArrayJoinForToString(vm, thisObject)) {
		// 2. Let func be the result of calling the [[Get]] internal method of array with argument "join".
		JSValue function = thisObject->get(vm, exec, vm.propertyNames->join);
		RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        // 3. If IsCallable(func) is false, then let func be the standard built-in method Object.prototype.toString (15.2.4.2).
        auto callData = getCallData(vm, function);
        if (UNLIKELY(callData.type == CallData::Type::None))
            return JSValue::encode(objectPrototypeToString(exec, thisObject));

		// 4. Return the result of calling the [[Call]] internal method of func providing array as the this value and an empty arguments list.
		if (!isJSArray(thisObject) || callData.type != CallData::Type::Native || callData.native.function != arrayProtoFuncJoin) {
			return JSValue::encode(call(vm, exec, function, callData, thisObject, *vm.emptyList));
		}
    }

    ASSERT(isJSArray(thisValue));
    JSArray* thisArray = asArray(thisValue);
    
    unsigned length = thisArray->length();

    StringRecursionChecker checker(exec, vm, thisArray);
    ASSERT(!scope.exception() || checker.earlyReturnValue());
    if (JSValue earlyReturnValue = checker.earlyReturnValue())
        return JSValue::encode(earlyReturnValue);

    if (LIKELY(canUseFastJoin(thisArray))) {
        const LChar comma = ',';

        JSValue result = fastJoin(*exec, vm, thisArray, StringView(&comma, 1), length);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        return JSValue::encode(result);
    }

    JSStringJoiner joiner(*exec, ',', length);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    for (unsigned i = 0; i < length; ++i) {
        JSValue element = thisArray->tryGetIndexQuickly(i);
        if (!element) {
            element = thisArray->get(vm, exec, i);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }
        joiner.append(*exec, vm, element);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    return JSValue::encode(joiner.join(*exec));
}

static JSString* toLocaleString(VM& vm, ExecState* exec, JSValue value, JSValue locales, JSValue options)
{
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue toLocaleStringMethod = value.get(vm, exec, vm.propertyNames->toLocaleString);
    RETURN_IF_EXCEPTION(scope, nullptr);

    auto callData = getCallData(vm, toLocaleStringMethod);
    if (callData.type == CallData::Type::None) {
        throwTypeError(exec, vm, "toLocaleString is not callable");
        return nullptr;
    }

    MarkedArgumentBuffer arguments;
    arguments.append(locales);
    arguments.append(options);
    ASSERT(!arguments.hasOverflowed());

    JSValue result = call(vm, exec, toLocaleStringMethod, callData, value, arguments);
    RETURN_IF_EXCEPTION(scope, nullptr);

    scope.release();
    return result.toString(exec);
}

// https://tc39.es/ecma402/#sup-array.prototype.tolocalestring
static EncodedJSValue JSC_HOST_CALL arrayProtoFuncToLocaleString(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    JSValue thisValue = exec->thisValue().toThis(exec, ECMAMode::strict());

    JSValue locales = exec->argument(0);
    JSValue options = exec->argument(1);

    // 1. Let array be ? ToObject(this value).
    JSObject* thisObject = thisValue.toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    StringRecursionChecker checker(exec, vm, thisObject);
    ASSERT(!scope.exception() || checker.earlyReturnValue());
    if (JSValue earlyReturnValue = checker.earlyReturnValue())
        return JSValue::encode(earlyReturnValue);

    // 2. Let len be ? ToLength(? Get(array, "length")).
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObject));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // 3. Let separator be the String value for the list-separator String appropriate for
    // the host environment's current locale (this is derived in an implementation-defined way).
    const LChar comma = ',';
    JSString* separator = jsSingleCharacterString(vm, comma);

    // 4. Let R be the empty String.
    if (!length)
        return JSValue::encode(jsEmptyString(vm));

    // 5. Let k be 0.
    JSValue element0 = thisObject->getIndex(vm, exec, 0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // 6. Repeat, while k < len,
    // 6.a. If k > 0, then
    // 6.a.i. Set R to the string-concatenation of R and separator.

    JSString* r = nullptr;
    if (element0.isUndefinedOrNull())
        r = jsEmptyString(vm);
    else {
        r = toLocaleString(vm, exec, element0, locales, options);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    // 8. Let k be 1.
    // 9. Repeat, while k < len
    // 9.e Increase k by 1..
    for (uint64_t k = 1; k < length; ++k) {
        // 6.b. Let nextElement be ? Get(array, ! ToString(k)).
        JSValue element = thisObject->getIndex(vm, exec, k);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        // c. If nextElement is not undefined or null, then
        JSString* next = nullptr;
        if (element.isUndefinedOrNull())
            next = jsEmptyString(vm);
        else {
            // i. Let S be ? ToString(? Invoke(nextElement, "toLocaleString",  locales, options )).
            // ii. Set R to the string-concatenation of R and S.
            next = toLocaleString(vm, exec, element, locales, options);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }

        // d. Increase k by 1.
        r = jsString(vm, exec, r, separator, next);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    // 7. Return R.
    return JSValue::encode(r);
}

static inline bool isHole(double value)
{
    return std::isnan(value);
}

static inline bool isHole(const WriteBarrier<Unknown>& value)
{
    return !value;
}

template<typename T> static inline bool containsHole(T* data, unsigned length)
{
    for (unsigned i = 0; i < length; ++i) {
        if (isHole(data[i]))
            return true;
    }
    return false;
}

static inline bool holesMustForwardToPrototype(VM& vm, JSObject* object)
{
    return object->structure(vm)->holesMustForwardToPrototype(vm);
}

static JSValue slowJoin(ExecState& exec, JSObject* thisObject, JSString* separator, uint64_t length)
{
    VM& vm = exec.vm();

    // 5. If len is zero, return the empty String.
    if (!length)
        return jsEmptyString(vm);

    // 6. Let element0 be Get(O, "0").
    JSValue element0 = thisObject->getIndex(vm, &exec, 0);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    // 7. If element0 is undefined or null, let R be the empty String; otherwise, let R be ? ToString(element0).
    JSString* r = 0;
    if (element0.isUndefinedOrNull())
        r = jsEmptyString(vm);
    else
        r = element0.toString(&exec);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    if(UNLIKELY(WTF::isLowMemory())) {
        throwOutOfMemoryError(&exec);
        return jsUndefined();
    }

    // 8. Let k be 1.
    // 9. Repeat, while k < len
    // 9.e Increase k by 1..
    MonitorMallocFailedPreventer preventer;
    for (uint64_t k = 1; k < length; ++k) {
        // b. Let element be ? Get(O, ! ToString(k)).
        JSValue element = thisObject->getIndex(vm, &exec, k);
        RETURN_IF_EXCEPTION(vm, jsUndefined());

        // c. If element is undefined or null, let next be the empty String; otherwise, let next be ? ToString(element).
        JSString* next = 0;
        if (element.isUndefinedOrNull()) {
            if (!separator->length())
                continue;
            next = jsEmptyString(vm);
        } else
            next = element.toString(&exec);
        RETURN_IF_EXCEPTION(vm, jsUndefined());

        // a. Let S be the String value produced by concatenating R and sep.
        // d. Let R be a String value produced by concatenating S and next.
        r = jsString(vm, &exec, r, separator, next);
        RETURN_IF_EXCEPTION(vm, jsUndefined());

		if(preventer.lowMemory()){
            throwOutOfMemoryError(vm, &exec);
            return jsUndefined();
        }
    }
    // 10. Return R.
    return r;
}

static inline bool canUseFastJoin(const JSObject* thisObject)
{
    switch (thisObject->structure()->indexingType()) {
    case ALL_CONTIGUOUS_INDEXING_TYPES:
    case ALL_INT32_INDEXING_TYPES:
    case ALL_DOUBLE_INDEXING_TYPES:
    case ALL_UNDECIDED_INDEXING_TYPES:
        return true;
    default:
        break;
    }
    return false;
}

static inline JSValue fastJoin(ExecState& state, VM& vm, JSObject* thisObject, const StringView& separator, unsigned length)
{
    switch (thisObject->structure()->indexingType()) {
    case ALL_INT32_INDEXING_TYPES: {
        auto& butterfly = *thisObject->butterfly();
        if (UNLIKELY(length > butterfly.publicLength()))
            break;
        JSStringJoiner joiner(state, separator, length);
        RETURN_IF_EXCEPTION(vm, jsUndefined());
        auto data = butterfly.contiguous().data();
        bool holesKnownToBeOK = false;
        for (unsigned i = 0; i < length; ++i) {
            JSValue value = data[i].get();
            if (LIKELY(value))
                joiner.appendNumber(vm, value.asInt32());
            else {
                if (!holesKnownToBeOK) {
                    if (holesMustForwardToPrototype(vm, thisObject))
                        goto generalCase;
                    holesKnownToBeOK = true;
                }
                joiner.appendEmptyString();
            }
        }
        return joiner.join(state);
    }
    case ALL_CONTIGUOUS_INDEXING_TYPES: {
        auto& butterfly = *thisObject->butterfly();
        if (UNLIKELY(length > butterfly.publicLength()))
            break;
        JSStringJoiner joiner(state, separator, length);
        RETURN_IF_EXCEPTION(vm, jsUndefined());
        auto data = butterfly.contiguous().data();
        bool holesKnownToBeOK = false;
        for (unsigned i = 0; i < length; ++i) {
            if (JSValue value = data[i].get()) {
                if (!joiner.appendWithoutSideEffects(state, vm, value))
                    goto generalCase;
                RETURN_IF_EXCEPTION(vm, jsUndefined());
            } else {
                if (!holesKnownToBeOK) {
                    if (holesMustForwardToPrototype(vm, thisObject))
                        goto generalCase;
                    holesKnownToBeOK = true;
                }
                joiner.appendEmptyString();
            }
        }
        return joiner.join(state);
    }
    case ALL_DOUBLE_INDEXING_TYPES: {
        auto& butterfly = *thisObject->butterfly();
        if (UNLIKELY(length > butterfly.publicLength()))
            break;
        JSStringJoiner joiner(state, separator, length);
        RETURN_IF_EXCEPTION(vm, jsUndefined());
        auto data = butterfly.contiguousDouble().data();
        bool holesKnownToBeOK = false;
        for (unsigned i = 0; i < length; ++i) {
            double value = data[i];
            if (LIKELY(!isHole(value)))
                joiner.appendNumber(vm, value);
            else {
                if (!holesKnownToBeOK) {
                    if (thisObject->structure(vm)->holesMustForwardToPrototype(vm))
                        goto generalCase;
                    holesKnownToBeOK = true;
                }
                joiner.appendEmptyString();
            }
        }
        return joiner.join(state);
    }
    case ALL_UNDECIDED_INDEXING_TYPES: {
        if (length && holesMustForwardToPrototype(vm, thisObject))
            goto generalCase;
        switch (separator.length()) {
        case 0:
            return jsEmptyString(vm);
        case 1: {
            if (length <= 1)
                return jsEmptyString(vm);
            if (separator.is8Bit())
                return repeatCharacter(vm, &state, separator.characters8()[0], length - 1);
            return repeatCharacter(vm, &state, separator.characters16()[0], length - 1);
        }
        default:
			JSString* result = jsEmptyString(vm);
			if (length <= 1)
				return result;

			JSString* operand = jsString(vm, separator.toString());
	        RETURN_IF_EXCEPTION(vm, jsUndefined());
			unsigned count = length - 1;
			for (;;) {
				if (count & 1) {
					result = jsString(vm, &state, result, operand);
			        RETURN_IF_EXCEPTION(vm, jsUndefined());
				}
				count >>= 1;
				if (!count)
					return result;
				operand = jsString(vm, &state, operand, operand);
		        RETURN_IF_EXCEPTION(vm, jsUndefined());
			}
		}
	}
    }

generalCase:
    JSStringJoiner joiner(state, separator, length);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    for (unsigned i = 0; i < length; ++i) {
        JSValue element = thisObject->getIndex(vm, &state, i);
        RETURN_IF_EXCEPTION(vm, jsUndefined());
        joiner.append(state, vm, element);
        RETURN_IF_EXCEPTION(vm, jsUndefined());
    }
    return joiner.join(state);
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncJoin(ExecState* exec)
{
    VM& vm = exec->vm();

    // 1. Let O be ? ToObject(this value).
    JSObject* thisObject = exec->thisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObject);
	if (UNLIKELY(!thisObject))
		return encodedJSUndefined();

    StringRecursionChecker checker(exec, vm, thisObject);
    ASSERT(!scope.exception() || checker.earlyReturnValue());
    if (JSValue earlyReturnValue = checker.earlyReturnValue())
        return JSValue::encode(earlyReturnValue);

    // 2. Let len be ? ToLength(? Get(O, "length")).
    double length = toLength(vm, exec, thisObject);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // 3. If separator is undefined, let separator be the single-element String ",".
    JSValue separatorValue = exec->argument(0);
    if (separatorValue.isUndefined()) {
        const LChar comma = ',';

        if (UNLIKELY(length > std::numeric_limits<unsigned>::max() || !canUseFastJoin(thisObject))) {
            uint64_t length64 = static_cast<uint64_t>(length);
            ASSERT(static_cast<double>(length64) == length);
            JSString* jsSeparator = jsSingleCharacterString(vm, comma);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

            return JSValue::encode(slowJoin(*exec, thisObject, jsSeparator, length64));
        }

        unsigned unsignedLength = static_cast<unsigned>(length);
        ASSERT(static_cast<double>(unsignedLength) == length);
        return JSValue::encode(fastJoin(*exec, vm, thisObject, StringView(&comma, 1), unsignedLength));
    }

    // 4. Let sep be ? ToString(separator).
    JSString* jsSeparator = separatorValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (UNLIKELY(length > std::numeric_limits<unsigned>::max() || !canUseFastJoin(thisObject))) {
        uint64_t length64 = static_cast<uint64_t>(length);
        ASSERT(static_cast<double>(length64) == length);
        return JSValue::encode(slowJoin(*exec, thisObject, jsSeparator, length64));
    }

    StringViewWithUnderlyingString viewWithString = jsSeparator->viewWithUnderlyingString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    return JSValue::encode(fastJoin(*exec, vm, thisObject, viewWithString.view, length));
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncPop(ExecState* exec)
{
    VM& vm = exec->vm();
    JSValue thisValue = exec->hostThisValue().toThis(exec, ECMAMode::strict());

    if (isJSArray(thisValue))
        return JSValue::encode(asArray(thisValue)->pop(exec));

    JSObject* thisObj = thisValue.toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
    if (UNLIKELY(!thisObj))
        return encodedJSUndefined();
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObj));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue result;
    if (length == 0) {
        setLength(exec, vm, thisObj, length);
        return JSValue::encode(jsUndefined());
    }

    //static_assert(MAX_ARRAY_INDEX + 1 > MAX_ARRAY_INDEX);
    uint64_t index = length - 1;
	result = thisObj->get(vm, exec, index);
	RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	bool success = thisObj->methodTable()->deletePropertyByIndex(thisObj, exec, index);
	RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	if (UNLIKELY(!success)) {
		throwTypeError(exec, vm, "Unable to delete property.");
		return encodedJSUndefined();
	}

    setLength(exec, vm, thisObj, index);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncPush(ExecState* exec)
{
    VM& vm = exec->vm();
    JSValue thisValue = exec->hostThisValue().toThis(exec, ECMAMode::strict());

    if (LIKELY(isJSArray(thisValue) && exec->argumentCount() == 1)) {
        JSArray* array = asArray(thisValue);
        array->pushInline(vm, exec, exec->uncheckedArgument(0));
        return JSValue::encode(jsNumber(array->length()));
    }

    JSObject* thisObj = thisValue.toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
    if (UNLIKELY(!thisObj))
        return encodedJSUndefined();
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObj));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    unsigned argCount = exec->argumentCount();

    if (UNLIKELY(length + argCount > static_cast<uint64_t>(maxSafeInteger())))
        return throwVMTypeError(exec, vm, ASCIILiteral("push cannot produce an array of length larger than (2 ** 53) - 1"));

    for (unsigned n = 0; n < argCount; n++) {
    	thisObj->putByIndexInline(exec, vm, static_cast<uint64_t>(length + n), exec->uncheckedArgument(n), true);
		RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }
    uint64_t newLength = length + argCount;
    setLength(exec, vm, thisObj, newLength);
    return JSValue::encode(jsNumber(newLength));
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncReverse(ExecState* exec)
{
	VM& vm = exec->vm();

    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObject);
	if (UNLIKELY(!thisObject))
		return encodedJSUndefined();

    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObject));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    switch (thisObject->structure()->indexingType()) {
    case ALL_CONTIGUOUS_INDEXING_TYPES:
    case ALL_INT32_INDEXING_TYPES: {
        auto& butterfly = *thisObject->butterfly();
        if (length > butterfly.publicLength())
            break;
        auto data = butterfly.contiguous().data();
        if (containsHole(data, length) && thisObject->structure(vm)->holesMustForwardToPrototype(vm))
            break;
        std::reverse(data, data + length);
        return JSValue::encode(thisObject);
    }
    case ALL_DOUBLE_INDEXING_TYPES: {
        auto& butterfly = *thisObject->butterfly();
        if (length > butterfly.publicLength())
            break;
        auto data = butterfly.contiguousDouble().data();
        if (containsHole(data, length) && thisObject->structure(vm)->holesMustForwardToPrototype(vm))
            break;
        std::reverse(data, data + length);
        return JSValue::encode(thisObject);
    }
    case ALL_ARRAY_STORAGE_INDEXING_TYPES: {
        auto& storage = *thisObject->butterfly()->arrayStorage();
        if (length > storage.vectorLength())
            break;
        if (storage.hasHoles() && thisObject->structure(vm)->holesMustForwardToPrototype(vm))
            break;
        auto data = storage.vector().data();
        std::reverse(data, data + length);
        return JSValue::encode(thisObject);
    }
    }

    uint64_t middle = length / 2;
    for (uint64_t lower = 0; lower < middle; lower++) {
    	uint64_t upper = length - lower - 1;
        bool lowerExists = thisObject->hasProperty(exec, lower);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        JSValue lowerValue;
        if (lowerExists) {
            lowerValue = thisObject->get(vm, exec, lower);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }

        bool upperExists = thisObject->hasProperty(exec, upper);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        JSValue upperValue;
        if (upperExists) {
            upperValue = thisObject->get(vm, exec, upper);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }

        if (!lowerExists && !upperExists) {
            // Spec says to do nothing when neither lower nor upper exist.
            continue;
        }

        if (upperExists) {
            thisObject->putByIndexInline(exec, vm, lower, upperValue, true);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        } else {
        	bool success = thisObject->deleteProperty(exec, lower);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            if (UNLIKELY(!success)) {
                throwTypeError(exec, vm, ASCIILiteral("Unable to delete property."));
                return encodedJSUndefined();
            }
        }

        if (lowerExists) {
            thisObject->putByIndexInline(exec, vm, upper, lowerValue, true);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        } else {
        	bool success = thisObject->deleteProperty(exec, upper);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            if (UNLIKELY(!success)) {
                throwTypeError(exec, vm, ASCIILiteral("Unable to delete property."));
                return encodedJSUndefined();
            }
        }
    }
    return JSValue::encode(thisObject);
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncShift(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* thisObj = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
	if (UNLIKELY(!thisObj))
		return encodedJSUndefined();
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObj));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (length == 0) {
    	setLength(exec, vm, thisObj, length);
        return JSValue::encode(jsUndefined());
    }

    JSValue result = thisObj->getIndex(vm, exec, 0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	shift<JSArray::ShiftCountForShift>(exec, thisObj, 0, 1, 0, length);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	setLength(exec, vm, thisObj, length - 1);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncSlice(ExecState* exec)
{
    // https://tc39.github.io/ecma262/#sec-array.prototype.slice
	VM& vm = exec->vm();
    JSObject* thisObj = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
	if (UNLIKELY(!thisObj))
		return encodedJSUndefined();
	uint64_t length = toLength(vm, exec, thisObj);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    uint64_t begin = argumentClampedIndexFromStartOrEnd(exec, exec->argument(0), length);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    uint64_t end = argumentClampedIndexFromStartOrEnd(exec, exec->argument(1), length, length);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (end < begin)
        end = begin;

    std::pair<SpeciesConstructResult, JSObject*> speciesResult = speciesConstructArray(vm, exec, thisObj, end - begin);
    // We can only get an exception if we call some user function.
    ASSERT(!!scope.exception() == (speciesResult.first == SpeciesConstructResult::Exception));
    if (UNLIKELY(speciesResult.first == SpeciesConstructResult::Exception))
    	 return encodedJSUndefined();

    bool okToDoFastPath = speciesResult.first == SpeciesConstructResult::FastPath && isJSArray(thisObj) && length == toLength(vm, exec, thisObj);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	if (LIKELY(okToDoFastPath)) {
		if (JSArray* result = asArray(thisObj)->fastSlice(*exec, begin, end - begin))
            return JSValue::encode(result);
    }

    JSObject* result;
    if (speciesResult.first == SpeciesConstructResult::CreatedObject)
        result = speciesResult.second;
    else {
        if (UNLIKELY(end - begin > std::numeric_limits<uint32_t>::max())) {
            throwRangeError(exec, vm, LengthExceededTheMaximumArrayLengthError);
            return encodedJSUndefined();
        }
        result = constructEmptyArray(exec, 0, static_cast<uint32_t>(end - begin));
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    // Document that we need to keep the source array alive until after anything
    // that can GC (e.g. allocating the result array).
    thisObj->use();

    uint64_t n = 0;
    for (uint64_t k = begin; k < end; k++, n++) {
        JSValue v = getProperty(vm, exec, thisObj, k);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (v) {
        	result->putDirectIndex(exec, n, v, 0, PutDirectIndexShouldThrow);
        	RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }
    }
    setLength(exec, vm, result, n);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncSplice(ExecState* exec)
{
    // 15.4.4.12

    VM& vm = exec->vm();

    JSObject* thisObj = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
	if (UNLIKELY(!thisObj))
		return encodedJSUndefined();
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObj));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    if (!exec->argumentCount()) {
        std::pair<SpeciesConstructResult, JSObject*> speciesResult = speciesConstructArray(vm, exec, thisObj, 0);
        ASSERT(!!scope.exception() == (speciesResult.first == SpeciesConstructResult::Exception));
		if (UNLIKELY(speciesResult.first == SpeciesConstructResult::Exception))
			return encodedJSUndefined();

        JSObject* result;
        if (speciesResult.first == SpeciesConstructResult::CreatedObject)
            result = speciesResult.second;
        else {
            result = constructEmptyArray(exec, 0);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }

        setLength(exec, vm, result, 0);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        setLength(exec, vm, thisObj, length);
        return JSValue::encode(result);
    }

    uint64_t actualStart = argumentClampedIndexFromStartOrEnd(exec, exec->argument(0), length);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    uint64_t actualDeleteCount = length - actualStart;
    if (exec->argumentCount() > 1) {
        double deleteCount = exec->uncheckedArgument(1).toIntegerOrInfinity(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (deleteCount < 0)
            actualDeleteCount = 0;
        else if (deleteCount > length - actualStart)
            actualDeleteCount = length - actualStart;
        else
            actualDeleteCount = static_cast<uint64_t>(deleteCount);
    }
    unsigned itemCount = std::max<int>(exec->argumentCount() - 2, 0);
    if (UNLIKELY(length - actualDeleteCount + itemCount > static_cast<uint64_t>(maxSafeInteger())))
        return throwVMTypeError(exec, vm, ASCIILiteral("Splice cannot produce an array of length larger than (2 ** 53) - 1"));

    std::pair<SpeciesConstructResult, JSObject*> speciesResult = speciesConstructArray(vm, exec, thisObj, actualDeleteCount);
    ASSERT(!!scope.exception() == (speciesResult.first == SpeciesConstructResult::Exception));
    if (speciesResult.first == SpeciesConstructResult::Exception)
        return JSValue::encode(jsUndefined());

    JSObject* result = 0;
    bool okToDoFastPath = speciesResult.first == SpeciesConstructResult::FastPath && isJSArray(thisObj) && length == toLength(vm, exec, thisObj);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	if (LIKELY(okToDoFastPath))
        result = asArray(thisObj)->fastSlice(*exec, actualStart, actualDeleteCount);

    if (!result) {
        DeferGC deferGC(vm.heap);

        if (speciesResult.first == SpeciesConstructResult::CreatedObject)
            result = speciesResult.second;
        else {
            if (UNLIKELY(actualDeleteCount > std::numeric_limits<uint32_t>::max())) {
                throwRangeError(exec, vm, LengthExceededTheMaximumArrayLengthError);
                return encodedJSUndefined();
            }
            result = JSArray::tryCreate(vm, exec, exec->lexicalGlobalObject()->arrayStructureForIndexingTypeDuringAllocation(ArrayWithUndecided), static_cast<uint32_t>(actualDeleteCount));
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }

        for (uint64_t k = 0; k < actualDeleteCount; ++k) {
            JSValue v = getProperty(vm, exec, thisObj, k + actualStart);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            if (UNLIKELY(!v))
                continue;
            result->putDirectIndex(exec, k, v, 0, PutDirectIndexShouldThrow);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }
        setLength(exec, vm, result, actualDeleteCount);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    if (itemCount < actualDeleteCount) {
        shift<JSArray::ShiftCountForSplice>(exec, thisObj, actualStart, actualDeleteCount, itemCount, length);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    } else if (itemCount > actualDeleteCount) {
        unshift(exec, thisObj, actualStart, actualDeleteCount, itemCount, length);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }
    for (unsigned k = 0; k < itemCount; ++k) {
        thisObj->putByIndexInline(exec, vm, k + actualStart, exec->uncheckedArgument(k + 2), true);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    setLength(exec, vm, thisObj, length - actualDeleteCount + itemCount);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncUnShift(ExecState* exec)
{
    // 15.4.4.13
	VM& vm = exec->vm();
    JSObject* thisObj = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
	if (UNLIKELY(!thisObj))
		return encodedJSUndefined();
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObj));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    unsigned nrArgs = exec->argumentCount();
    if (nrArgs) {
    	if (UNLIKELY(length + nrArgs > static_cast<uint64_t>(maxSafeInteger())))
			return throwVMTypeError(exec, vm, ASCIILiteral("unshift cannot produce an array of length larger than (2 ** 53) - 1"));
		unshift(exec, thisObj, 0, 0, nrArgs, length);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }
    for (unsigned k = 0; k < nrArgs; ++k) {
        thisObj->putByIndexInline(exec, vm, k, exec->uncheckedArgument(k), true);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }
    uint64_t newLength = length + nrArgs;
    setLength(exec, vm, thisObj, newLength);
    return JSValue::encode(jsNumber(newLength));
}

enum class IndexOfDirection { Forward, Backward };
template<IndexOfDirection direction>
ALWAYS_INLINE JSValue fastIndexOf(ExecState* exec, VM& vm, JSArray* array, uint64_t length64, JSValue searchElement, uint64_t index64)
{
	bool canDoFastPath = array->canDoFastIndexedAccess(vm)
		&& array->getArrayLength() == length64 // The effects in getting `index` could have changed the length of this array.
		&& static_cast<uint32_t>(index64) == index64;
	if (!canDoFastPath)
        return JSValue();

    uint32_t length = static_cast<uint32_t>(length64);
    uint32_t index = static_cast<uint32_t>(index64);

	switch (array->structure()->indexingType()) {
	case ALL_INT32_INDEXING_TYPES: {
		if (!searchElement.isNumber())
			return jsNumber(-1);
		JSValue searchInt32;
		if (searchElement.isInt32())
			searchInt32 = searchElement;
		else {
			double searchNumber = searchElement.asNumber();
			if (!canBeInt32(searchNumber))
				return jsNumber(-1);
			searchInt32 = jsNumber(static_cast<int32_t>(searchNumber));
		}
		auto& butterfly = *array->butterfly();
		auto data = butterfly.contiguous().data();
        if (direction == IndexOfDirection::Forward) {
			for (; index < length; ++index) {
				// Array#indexOf uses `===` semantics (not HashMap isEqual semantics).
				// And the hole never matches against Int32 value.
				if (searchInt32 == data[index].get())
					return jsNumber(index);
			}
		} else {
			do {
				ASSERT(index < length);
				// Array#lastIndexOf uses `===` semantics (not HashMap isEqual semantics).
				// And the hole never matches against Int32 value.
				if (searchInt32 == data[index].get())
					return jsNumber(index);
			} while (index--);
		}
		return jsNumber(-1);
	}
	case ALL_CONTIGUOUS_INDEXING_TYPES: {
		auto& butterfly = *array->butterfly();
		auto data = butterfly.contiguous().data();

        if (direction == IndexOfDirection::Forward) {
			for (; index < length; ++index) {
				JSValue value = data[index].get();
				if (!value)
					continue;
				bool isEqual = JSValue::strictEqual(exec, searchElement, value);
				RETURN_IF_EXCEPTION(vm, jsUndefined());
				if (isEqual)
                    return jsNumber(index);
            }
        } else {
            do {
                ASSERT(index < length);
                JSValue value = data[index].get();
                if (!value)
                    continue;
                bool isEqual = JSValue::strictEqual(exec, searchElement, value);
                RETURN_IF_EXCEPTION(vm, jsUndefined());
                if (isEqual)
                    return jsNumber(index);
            } while (index--);
        }
        return jsNumber(-1);
	}
	case ALL_DOUBLE_INDEXING_TYPES: {
		if (!searchElement.isNumber())
			return jsNumber(-1);
		double searchNumber = searchElement.asNumber();
		auto& butterfly = *array->butterfly();
		auto data = butterfly.contiguousDouble().data();
        if (direction == IndexOfDirection::Forward) {
			for (; index < length; ++index) {
				// Array#indexOf uses `===` semantics (not HashMap isEqual semantics).
				// And the hole never matches since it is NaN.
				if (data[index] == searchNumber)
                    return jsNumber(index);
            }
        } else {
            do {
                ASSERT(index < length);
                // Array#lastIndexOf uses `===` semantics (not HashMap isEqual semantics).
                // And the hole never matches since it is NaN.
                if (data[index] == searchNumber)
                    return jsNumber(index);
            } while (index--);
        }
        return jsNumber(-1);
    }
    default:
        return JSValue();
    }
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncIndexOf(ExecState* exec)
{
	VM& vm = exec->vm();

	// 15.4.4.14
    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
	if (UNLIKELY(!thisObject))
		return encodedJSUndefined();
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObject));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    uint64_t index = argumentClampedIndexFromStartOrEnd(exec, exec->argument(1), length);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    JSValue searchElement = exec->argument(0);

	if (isJSArray(thisObject)) {
		JSValue result = fastIndexOf<IndexOfDirection::Forward>(exec, vm, asArray(thisObject), length, searchElement, index);
	    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (result)
			return JSValue::encode(result);
	}

    for (; index < length; ++index) {
        JSValue e = getProperty(vm, exec, thisObject, index);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (!e)
            continue;
        bool isEqual = JSValue::strictEqual(exec, searchElement, e);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
		if (isEqual)
            return JSValue::encode(jsNumber(index));
    }

    return JSValue::encode(jsNumber(-1));
}

static EncodedJSValue JSC_HOST_CALL arrayProtoFuncLastIndexOf(ExecState* exec)
{
    VM& vm = exec->vm();
    // 15.4.4.15
    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
	if (UNLIKELY(!thisObject))
		return encodedJSUndefined();
    uint64_t length = static_cast<uint64_t>(toLength(vm, exec, thisObject));
    if (UNLIKELY(vm.exception()) || !length)
        return JSValue::encode(jsNumber(-1));

    uint64_t index = length - 1;
    if (exec->argumentCount() >= 2) {
        JSValue fromValue = exec->uncheckedArgument(1);
        double fromDouble = fromValue.toIntegerOrInfinity(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (fromDouble < 0) {
            fromDouble += length;
            if (fromDouble < 0)
                return JSValue::encode(jsNumber(-1));
        }
        if (fromDouble < length)
            index = static_cast<uint64_t>(fromDouble);
    }

    JSValue searchElement = exec->argument(0);

    if (isJSArray(thisObject)) {
        JSValue result = fastIndexOf<IndexOfDirection::Backward>(exec, vm, asArray(thisObject), length, searchElement, index);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (result)
            return JSValue::encode(result);
    }

    do {
        ASSERT(index < length);
        JSValue e = getProperty(vm, exec, thisObject, index);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (!e)
            continue;
        bool isEqual = JSValue::strictEqual(exec, searchElement, e);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (isEqual)
            return JSValue::encode(jsNumber(index));
    } while (index--);

    return JSValue::encode(jsNumber(-1));
}

static bool moveElements(ExecState* exec, VM& vm, JSArray* target, unsigned targetOffset, JSArray* source, unsigned sourceLength)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    if (LIKELY(!hasAnyArrayStorage(source->structure()->indexingType()) && !source->structure()->holesMustForwardToPrototype(vm))) {
        for (unsigned i = 0; i < sourceLength; ++i) {
            JSValue value = source->tryGetIndexQuickly(i);
            if (value) {
                target->putDirectIndex(exec, targetOffset + i, value, 0, PutDirectIndexShouldThrow);
                RETURN_IF_EXCEPTION(vm, false);
            }
        }
    } else {
        for (unsigned i = 0; i < sourceLength; ++i) {
            JSValue value = getProperty(vm, exec, source, i);
            RETURN_IF_EXCEPTION(vm, false);
            if (value) {
                target->putDirectIndex(exec, targetOffset + i, value, 0, PutDirectIndexShouldThrow);
                RETURN_IF_EXCEPTION(vm, false);
            }
        }
    }
    return true;
}

static JSValue concatAppendOne(ExecState* exec, VM& vm, JSArray* first, JSValue second)
{
    auto scope = DECLARE_THROW_SCOPE(vm);

    ASSERT(!isJSArray(second));
    ASSERT(!shouldUseSlowPut(first->indexingType()));
    Butterfly* firstButterfly = first->butterfly();
    unsigned firstArraySize = firstButterfly->publicLength();

    Checked<unsigned, RecordOverflow> checkedResultSize = firstArraySize;
    checkedResultSize += 1;
    if (UNLIKELY(checkedResultSize.hasOverflowed())) {
        throwOutOfMemoryError(exec);
        return JSValue();
    }

    // This code still can see concat-spreadable items. It is ProxyObject and DerivedArray.
    // We should not use `isArray` check here since it is side-effectful for ProxyObject. Let's just bail out if it is ProxyObject or DerivedArray.
    if (second.isObject()) {
        JSType type = asObject(second)->structure()->typeInfo().type();
        if (UNLIKELY(type == ProxyObjectType || type == DerivedArrayType))
            return jsNull();
    }

    unsigned resultSize = checkedResultSize.unsafeGet();
    IndexingType type = first->mergeIndexingTypeForCopying(indexingTypeForValue(second) | IsArray);
    if (type == NonArray)
        type = first->structure()->indexingType();

    Structure* resultStructure = exec->lexicalGlobalObject()->arrayStructureForIndexingTypeDuringAllocation(type);
    JSArray* result = JSArray::tryCreate(vm, exec, resultStructure, resultSize);
    RETURN_IF_EXCEPTION(vm, JSValue());

    bool success = result->appendMemcpy(exec, vm, 0, first);
    RETURN_IF_EXCEPTION(vm, JSValue());
    if (!success) {
        bool success = moveElements(exec, vm, result, 0, first, firstArraySize);
        ASSERT(!scope.exception() == success);
        if (UNLIKELY(!success))
            return JSValue();
    }

    scope.release();
    result->putDirectIndex(exec, firstArraySize, second);

    return result;

}

namespace {
template<typename T>
inline void clearElement(T& element)
{
    element.clear();
}

template<>
inline void clearElement(double& element)
{
    element = PNaN;
}
}

template<typename T>
static ALWAYS_INLINE void copyElements(T* buffer, unsigned offset, void* source, unsigned sourceSize, IndexingType sourceType)
{
    if (sourceType != ArrayWithUndecided) {
        memcpy(buffer + offset, source, sizeof(JSValue) * sourceSize);
        return;
    }

    for (unsigned i = sourceSize; i--;)
        clearElement<T>(buffer[i + offset]);
};

EncodedJSValue JSC_HOST_CALL arrayProtoPrivateFuncConcatMemcpy(ExecState* exec)
{
    ASSERT(exec->argumentCount() == 2);
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSArray* firstArray = jsCast<JSArray*>(exec->uncheckedArgument(0));

    // This code assumes that neither array has set Symbol.isConcatSpreadable. If the first array
    // has indexed accessors then one of those accessors might change the value of Symbol.isConcatSpreadable
    // on the second argument.
    if (UNLIKELY(shouldUseSlowPut(firstArray->structure()->indexingType())))
        return JSValue::encode(jsNull());

    // We need to check the species constructor here since checking it in the JS wrapper is too expensive for the non-optimizing tiers.
    bool isValid = speciesWatchpointIsValid(exec, firstArray);
    if (UNLIKELY(!isValid))
        return JSValue::encode(jsNull());

    JSValue second = exec->uncheckedArgument(1);
    if (!isJSArray(second))
    	RELEASE_AND_RETURN(scope, JSValue::encode(concatAppendOne(exec, vm, firstArray, second)));

    JSArray* secondArray = jsCast<JSArray*>(second);

    Butterfly* firstButterfly = firstArray->butterfly();
    Butterfly* secondButterfly = secondArray->butterfly();

    unsigned firstArraySize = firstButterfly->publicLength();
    unsigned secondArraySize = secondButterfly->publicLength();

    Checked<unsigned, RecordOverflow> checkedResultSize = firstArraySize;
    checkedResultSize += secondArraySize;

    if (UNLIKELY(checkedResultSize.hasOverflowed())) {
        throwOutOfMemoryError(exec);
        return encodedJSUndefined();
    }

    unsigned resultSize = checkedResultSize.unsafeGet();
    IndexingType firstType = firstArray->structure()->indexingType();
    IndexingType secondType = secondArray->structure()->indexingType();
    IndexingType type = firstArray->mergeIndexingTypeForCopying(secondType);
    if (type == NonArray || !firstArray->canFastCopy(vm, secondArray) || resultSize >= MIN_SPARSE_ARRAY_INDEX) {
        JSArray* result = constructEmptyArray(exec, 0, resultSize);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        bool success = moveElements(exec, vm, result, 0, firstArray, firstArraySize);
        ASSERT(!scope.exception() == success);
        if (UNLIKELY(!success))
            return encodedJSUndefined();
        success = moveElements(exec, vm, result, firstArraySize, secondArray, secondArraySize);
        ASSERT(!scope.exception() == success);
        if (UNLIKELY(!success))
            return encodedJSUndefined();

        return JSValue::encode(result);
    }

    JSGlobalObject* lexicalGlobalObject = exec->lexicalGlobalObject();
    Structure* resultStructure = lexicalGlobalObject->arrayStructureForIndexingTypeDuringAllocation(type);
    if (UNLIKELY(hasAnyArrayStorage(resultStructure->indexingType())))
        return JSValue::encode(jsNull());

    ASSERT(!lexicalGlobalObject->isHavingABadTime());
    ObjectInitializationScope initializationScope(vm);
    DeferGC deferGC(vm.heap);
	JSArray* result = JSArray::tryCreateUninitializedRestricted(initializationScope, deferGC, resultStructure, resultSize);
	if (UNLIKELY(!result)) {
		if (!vm.exception())
			throwOutOfMemoryError(vm, exec);
        return encodedJSUndefined();
    }

    if (type == ArrayWithDouble) {
        double* buffer = result->butterfly()->contiguousDouble().data();
        copyElements(buffer, 0, firstButterfly->contiguousDouble().data(), firstArraySize, firstType);
        copyElements(buffer, firstArraySize, secondButterfly->contiguousDouble().data(), secondArraySize, secondType);

    } else if (type != ArrayWithUndecided) {
        WriteBarrier<Unknown>* buffer = result->butterfly()->contiguous().data();
        copyElements(buffer, 0, firstButterfly->contiguous().data(), firstArraySize, firstType);
        copyElements(buffer, firstArraySize, secondButterfly->contiguous().data(), secondArraySize, secondType);
    }

    result->butterfly()->setPublicLength(resultSize);
    return JSValue::encode(result);
}

EncodedJSValue JSC_HOST_CALL arrayProtoPrivateFuncAppendMemcpy(ExecState* exec)
{
    ASSERT(exec->argumentCount() == 3);

    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);
    JSArray* resultArray = jsCast<JSArray*>(exec->uncheckedArgument(0));
    JSArray* otherArray = jsCast<JSArray*>(exec->uncheckedArgument(1));
    JSValue startValue = exec->uncheckedArgument(2);
    ASSERT(startValue.isUInt32AsAnyInt());
    unsigned startIndex = startValue.asUInt32AsAnyInt();
    bool success = resultArray->appendMemcpy(exec, vm, startIndex, otherArray);
    EXCEPTION_ASSERT(!scope.exception() || !success);
	if (success)
		return JSValue::encode(jsUndefined());
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    moveElements(exec, vm, resultArray, startIndex, otherArray, otherArray->length());
    return JSValue::encode(jsUndefined());
}

} // namespace JSC
