/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "ErrorInstance.h"
#include "ErrorType.h"
#include "Exception.h"
#include "InternalFunction.h"
#include "Interpreter.h"
#include "JSObject.h"
#include "RuntimeType.h"
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <stdint.h>

namespace JSC {

class ExecState;
class JSGlobalData;
class JSGlobalObject;
class JSObject;
class SourceCode;
class Structure;

typedef JSGlobalData VM;

// ExecState wrappers.
ErrorInstance* createError(ExecState*, const String&, ErrorInstance::SourceAppender);
ErrorInstance* createEvalError(ExecState*, const String&, ErrorInstance::SourceAppender);
ErrorInstance* createRangeError(ExecState*, const String&, ErrorInstance::SourceAppender);
ErrorInstance* createReferenceError(ExecState*, const String&, ErrorInstance::SourceAppender);
ErrorInstance* createSyntaxError(ExecState*, const String&, ErrorInstance::SourceAppender);
ErrorInstance* createTypeError(ExecState*, const String&, ErrorInstance::SourceAppender, RuntimeType);
ErrorInstance* createNotEnoughArgumentsError(ExecState*, ErrorInstance::SourceAppender);
ErrorInstance* createURIError(ExecState*, const String&, ErrorInstance::SourceAppender);
ErrorInstance* createOutOfMemoryError(ExecState*, ErrorInstance::SourceAppender = nullptr);

JS_EXPORT_PRIVATE JSObject* createError(ExecState*, const String&);
JS_EXPORT_PRIVATE JSObject* createEvalError(ExecState*, const String&);
JS_EXPORT_PRIVATE JSObject* createRangeError(ExecState*, const String&);
JS_EXPORT_PRIVATE JSObject* createReferenceError(ExecState*, const String&);
JS_EXPORT_PRIVATE JSObject* createSyntaxError(ExecState*, const String&);
JS_EXPORT_PRIVATE JSObject* createTypeError(ExecState*);
JS_EXPORT_PRIVATE JSObject* createTypeError(ExecState*, const ASCIILiteral&);
template<unsigned charactersCount>
ALWAYS_INLINE JSObject* createTypeError(ExecState* exec, const char (&message)[charactersCount])
{
	return createTypeError(exec, ASCIILiteral(message));
}
JS_EXPORT_PRIVATE JSObject* createTypeError(ExecState*, const String&);
JS_EXPORT_PRIVATE JSObject* createNotEnoughArgumentsError(ExecState*);
JS_EXPORT_PRIVATE JSObject* createURIError(ExecState*, const String&);

JS_EXPORT_PRIVATE JSObject* createOutOfMemoryError(ExecState*, const String&);

JS_EXPORT_PRIVATE JSObject* createError(ExecState*, ErrorType, const String&);

PassOwnPtr<Vector<StackFrame>> getStackTrace(ExecState*, VM&, JSObject*, bool useCurrentFrame);
void getBytecodeOffset(ExecState*, VM&, Vector<StackFrame>*, CallFrame*&, unsigned& bytecodeOffset);
bool getLineColumnAndSource(Vector<StackFrame>* stackTrace, unsigned& line, unsigned& column, String& sourceURL);
bool hasErrorInfo(ExecState*, JSObject* error);
JS_EXPORT_PRIVATE void addErrorInfo(ExecState*, JSObject*, bool);
JSObject* addErrorInfo(ExecState*, JSObject* error, int line, const SourceCode&);

// Methods to throw Errors.

// Convenience wrappers, create an throw an exception with a default message.
JS_EXPORT_PRIVATE Exception* throwConstructorCannotBeCalledAsFunctionTypeError(ExecState*, const char* constructorName);
JS_EXPORT_PRIVATE Exception* throwTypeError(ExecState*, VM&);
JS_EXPORT_PRIVATE Exception* throwTypeError(ExecState*);
Exception* throwTypeError(ExecState* exec, VM& vm, const String& errorMessage);
template<unsigned charactersCount>
ALWAYS_INLINE Exception* throwTypeError(ExecState* exec, VM& vm, const char (&errorMessage)[charactersCount])
{
	return throwTypeError(exec, vm, ASCIILiteral(errorMessage));
}

JS_EXPORT_PRIVATE Exception* throwSyntaxError(ExecState*);
JS_EXPORT_PRIVATE Exception* throwSyntaxError(ExecState*, const String& errorMessage);
Exception* throwRangeError(ExecState*, JSGlobalData&, const ASCIILiteral&);
Exception* throwRangeError(ExecState*, JSGlobalData&, const String&);

// Convenience wrappers, wrap result as an EncodedJSValue.
inline void throwVMError(ExecState* exec, VM& vm, Exception* exception) { vm.throwException(exec, exception); }
inline EncodedJSValue throwVMError(ExecState* exec, VM& vm, JSValue error) { return JSValue::encode(vm.throwException(exec, error)); }
inline EncodedJSValue throwVMTypeError(ExecState* exec, VM& vm, const String& errorMessage)
{
	return JSValue::encode(throwTypeError(exec, vm, errorMessage));
}
template<unsigned charactersCount>
inline EncodedJSValue throwVMTypeError(ExecState* exec, VM& vm, const char (&errorMessage)[charactersCount])
{
	return throwVMTypeError(exec, vm, ASCIILiteral(errorMessage));
}
inline EncodedJSValue throwVMTypeError(ExecState* exec, VM& vm) { return JSValue::encode(throwTypeError(exec, vm)); }
inline EncodedJSValue throwVMRangeError(ExecState* state, VM& vm, const ASCIILiteral& errorMessage)
{
	return JSValue::encode(throwRangeError(state, vm, errorMessage));
}
inline EncodedJSValue throwVMRangeError(ExecState* state, VM& vm, const String& errorMessage)
{
	return JSValue::encode(throwRangeError(state, vm, errorMessage));
}
EncodedJSValue throwVMNotEnoughArgumentsError(ExecState* exec);

} // namespace JSC
