/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#pragma once

#include "Interpreter.h"
#include "JSDestructibleObject.h"
#include "RuntimeType.h"
#include "SourceProvider.h"
#include <wtf/PassOwnPtr.h>
#include <wtf/OwnPtr.h>
#include <wtf/Vector.h>

namespace JSC {

class ErrorInstance : public JSDestructibleObject {
public:
	typedef JSDestructibleObject Base;
	const static unsigned StructureFlags = Base::StructureFlags | OverridesGetOwnPropertySlot | OverridesGetOwnSpecialPropertyNames | OverridesPut | GetOwnPropertySlotIsImpureForPropertyAbsence;

    static const bool needsDestruction = true;
    static void destroy(JSCell*);

    enum SourceTextWhereErrorOccurred { FoundExactSource, FoundApproximateSource };
    typedef String (*SourceAppender) (const String& originalMessage, const String& sourceText, RuntimeType, SourceTextWhereErrorOccurred);

    DECLARE_EXPORT_INFO;

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype)
	{
		return Structure::create(globalData, globalObject, prototype, TypeInfo(ErrorInstanceType, StructureFlags), info());
	}

    static ErrorInstance* create(ExecState* exec, JSGlobalData& vm, Structure* structure, const String& message, SourceAppender appender = 0, RuntimeType type = TypeNothing, bool useCurrentFrame = true)
    {
		ErrorInstance* instance = new (NotNull, allocateCell<ErrorInstance>(vm.heap)) ErrorInstance(vm, structure);
		instance->finishCreation(vm, exec, message, appender, type, useCurrentFrame);
		return instance;
	}

	static ErrorInstance* create(ExecState* exec, Structure* structure, JSValue message, SourceAppender appender = 0, RuntimeType type = TypeNothing, bool useCurrentFrame = true);

    bool hasSourceAppender() const { return !!m_sourceAppender; }
	SourceAppender sourceAppender() const { return m_sourceAppender; }
	void setSourceAppender(SourceAppender appender) { m_sourceAppender = appender; }
	void clearSourceAppender() { m_sourceAppender = 0; }
	void setRuntimeTypeForCause(RuntimeType type) { m_runtimeTypeForCause = type; }
	RuntimeType runtimeTypeForCause() const { return m_runtimeTypeForCause; }
	void clearRuntimeTypeForCause() { m_runtimeTypeForCause = TypeNothing; }

    void setStackOverflowError() { m_stackOverflowError = true; }
    bool isStackOverflowError() const { return m_stackOverflowError; }
    void setOutOfMemoryError() { m_outOfMemoryError = true; }
    bool isOutOfMemoryError() const { return m_outOfMemoryError; }

    JS_EXPORT_PRIVATE String sanitizedToString(ExecState*);

    Vector<StackFrame>* stackTrace() { return m_stackTrace.get(); }

    bool materializeErrorInfoIfNeeded(VM&);
    bool materializeErrorInfoIfNeeded(VM&, PropertyName);

    void finalizeUnconditionally(VM&);

protected:
	explicit ErrorInstance(JSGlobalData&, Structure*);
	~ErrorInstance();

	void finishCreation(JSGlobalData& globalData, ExecState*, const String& message, SourceAppender = nullptr, RuntimeType = TypeNothing, bool useCurrentFrame = true);

    static void visitChildren(JSCell*, SlotVisitor&);

    static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);
    static void getOwnSpecialPropertyNames(JSObject*, ExecState*, PropertyNameArray&, DontEnumPropertiesMode);
    static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool shouldThrow);
    static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);
    static bool deleteProperty(JSCell*, ExecState*, PropertyName);

    void computeErrorInfo(VM&);

    OwnPtr<Vector<StackFrame>> m_stackTrace;
    unsigned m_line;
    unsigned m_column;
    String m_sourceURL;
    String m_stackString;
    SourceAppender m_sourceAppender;
    RuntimeType m_runtimeTypeForCause;
    bool m_stackOverflowError : 1;
    bool m_outOfMemoryError : 1;
    bool m_errorInfoMaterialized : 1;

    class StackTraceCleaner final : public UnconditionalFinalizer {
    public:
    	StackTraceCleaner(ErrorInstance* target)
            : m_target(target)
        {
        }
    private:
        virtual void finalizeUnconditionally(VM&) override;
        ErrorInstance* m_target;
    };
    StackTraceCleaner m_stackTraceCleaner;
};

template<>
ALWAYS_INLINE ErrorInstance* jsDynamicCast(JSCell* from)
{
	if(from->structure()->typeInfo().type() == JSType::ErrorInstanceType)
		return static_cast<ErrorInstance*>(from);
	return nullptr;
}

} // namespace JSC
