/*
 * Copyright (C) 2008-2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ExceptionHelpers.h"

#include "CodeBlock.h"
#include "CallFrame.h"
#include "ErrorInstance.h"
#include "Exception.h"
#include "JSGlobalObjectFunctions.h"
#include "JSObject.h"
#include "Interpreter.h"
#include "Operations.h"
#include "ThrowScope.h"

namespace JSC {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(TerminatedExecutionError);

const ClassInfo TerminatedExecutionError::s_info = { "TerminatedExecutionError", &Base::s_info, nullptr, CREATE_METHOD_TABLE(TerminatedExecutionError) };

JSObject* createTerminatedExecutionException(JSGlobalData& globalData)
{
    return TerminatedExecutionError::create(globalData);
}

bool isTerminatedExecutionException(Exception* exception)
{
    if (!exception->value().isObject())
        return false;

    return exception->value().inherits(TerminatedExecutionError::info());
}

bool isOutOfMemoryException(Exception* exception)
{
    if (auto* error = jsDynamicCast<ErrorInstance*>(exception->value()))
        return error->isOutOfMemoryError();

    return false;
}

JSObject* createStackOverflowError(ExecState* exec)
{
    return createRangeError(exec, ASCIILiteral("Maximum call stack size exceeded."));
}

JSObject* createUndefinedVariableError(ExecState* exec, const Identifier& ident)
{
    if (ident.isPrivateName()) {
        String message(tryMakeString("Can't find private variable: PrivateSymbol.", ident.string()));
        return createReferenceError(exec, message);
    }
    String message(tryMakeString("Can't find variable: ", ident.ustring()));
    return createReferenceError(exec, message);
}

String errorDescriptionForValue(ExecState* exec, JSValue v)
{
    VM& vm = exec->vm();
    if (v.isString()) {
        String string = asString(v)->value(exec);
        if (!string)
            return string;
        return tryMakeString('"', string, '"');
    }
    if (v.isSymbol())
        return asSymbol(v)->descriptiveString();
    if (v.isObject()) {
        JSObject* object = asObject(v);
        if (object->isCallable())
            return vm.smallStrings.functionString()->value(exec);
        return object->methodTable()->className(object);
    }
    return v.toWTFString(exec);
}

static StringView clampErrorMessage(const String& originalMessage)
{
    // Hopefully this is sufficiently long. Note, this is the length of the string not the number of bytes used.
    constexpr unsigned maxLength = 2 * KB;
    return StringView(originalMessage).substring(0, maxLength);
}

static String defaultApproximateSourceError(const String& originalMessage, const String& sourceText)
{
    return tryMakeString(clampErrorMessage(originalMessage), " (near '...", sourceText, "...')");
}

static String defaultSourceAppender(const String& originalMessage, const String& sourceText, RuntimeType, ErrorInstance::SourceTextWhereErrorOccurred occurrence)
{
    if (occurrence == ErrorInstance::FoundApproximateSource)
        return defaultApproximateSourceError(originalMessage, sourceText);

    ASSERT(occurrence == ErrorInstance::FoundExactSource);
    return tryMakeString(clampErrorMessage(originalMessage), " (evaluating '", sourceText, "')");
}

static String functionCallBase(const String& sourceText)
{
    // This function retrieves the 'foo.bar' substring from 'foo.bar(baz)'.
    // FIXME: This function has simple processing of /* */ style comments.
    // It doesn't properly handle embedded comments of string literals that contain
    // parenthesis or comment constructs, e.g. foo.bar("/abc\)*/").
    // https://bugs.webkit.org/show_bug.cgi?id=146304

    unsigned sourceLength = sourceText.length();
    unsigned idx = sourceLength - 1;
    if (sourceLength < 2 || sourceText[idx] != ')') {
        // For function calls that have many new lines in between their open parenthesis
        // and their closing parenthesis, the text range passed into the message appender
        // will not inlcude the text in between these parentheses, it will just be the desired
        // text that precedes the parentheses.
        return String();
    }

    unsigned parenStack = 1;
    bool isInMultiLineComment = false;
    idx -= 1;
    // Note that we're scanning text right to left instead of the more common left to right,
    // so syntax detection is backwards.
    while (parenStack && idx) {
        UChar curChar = sourceText[idx];
        if (isInMultiLineComment) {
            if (curChar == '*' && sourceText[idx - 1] == '/') {
                isInMultiLineComment = false;
                --idx;
            }
        } else if (curChar == '(')
            --parenStack;
        else if (curChar == ')')
            ++parenStack;
        else if (curChar == '/' && sourceText[idx - 1] == '*') {
            isInMultiLineComment = true;
            --idx;
        }

        if (idx)
            --idx;
    }

    if (parenStack) {
        // As noted in the FIXME at the top of this function, there are bugs
        // in the above string processing. This algorithm is mostly best effort
        // and it works for most JS text in practice. However, if we determine
        // that the algorithm failed, we should just return the empty value.
        return String();
    }

    // Don't display the ?. of an optional call.
    if (idx > 1 && sourceText[idx] == '.' && sourceText[idx - 1] == '?')
        idx -= 2;

    return sourceText.left(idx + 1);
}

static String notAFunctionSourceAppender(const String& originalMessage, const String& sourceText, RuntimeType type, ErrorInstance::SourceTextWhereErrorOccurred occurrence)
{
    ASSERT(type != TypeFunction);

    if (occurrence == ErrorInstance::FoundApproximateSource)
        return defaultApproximateSourceError(originalMessage, sourceText);

    ASSERT(occurrence == ErrorInstance::FoundExactSource);
    size_t notAFunctionIndex = originalMessage.reverseFind(ASCIILiteral("is not a function"));
    RELEASE_ASSERT(notAFunctionIndex != notFound);
    String displayValue;
    if (originalMessage.is8Bit())
        displayValue = String(originalMessage.characters8(), notAFunctionIndex - 1);
    else
        displayValue = String(originalMessage.characters16(), notAFunctionIndex - 1);

    String base = functionCallBase(sourceText);
    if (!base)
		return defaultApproximateSourceError(originalMessage, sourceText);
    StringBuilder builder(StringBuilder::OverflowHandler::RecordOverflow);
    builder.append(base);
    builder.appendLiteral(" is not a function. (In '");
    builder.append(sourceText);
    builder.appendLiteral("', '");
    builder.append(base);
    builder.appendLiteral("' is ");
    if (type == TypeSymbol)
        builder.appendLiteral("a Symbol");
    else {
		if (type == TypeObject)
			builder.appendLiteral("an instance of ");
		builder.append(displayValue);
    }
    builder.append(')');

    if (builder.hasOverflowed())
        return ASCIILiteral("object is not a function.");

    return builder.toString();
}

static String invalidParameterInSourceAppender(const String& originalMessage, const String& sourceText, RuntimeType type, ErrorInstance::SourceTextWhereErrorOccurred occurrence)
{
    ASSERT_UNUSED(type, type != TypeObject);

    if (occurrence == ErrorInstance::FoundApproximateSource)
        return defaultApproximateSourceError(originalMessage, sourceText);

    ASSERT(occurrence == ErrorInstance::FoundExactSource);
    size_t inIndex = sourceText.reverseFind(ASCIILiteral("in"));
    if (inIndex == notFound) {
		// This should basically never happen, since JS code must use the literal
		// text "in" for the `in` operation. However, if we fail to find "in"
		// for any reason, just fail gracefully.
		return originalMessage;
	}
    if (sourceText.find(ASCIILiteral("in")) != inIndex)
        return tryMakeString(originalMessage, " (evaluating '", sourceText, "')");

    static const unsigned inLength = 2;
    String rightHandSide = sourceText.substring(inIndex + inLength);
    return tryMakeString(rightHandSide, " is not an Object. (evaluating '", sourceText, "')");
}

static inline String invalidParameterInstanceofSourceAppender(const String& content, const String& originalMessage, const String& sourceText, RuntimeType, ErrorInstance::SourceTextWhereErrorOccurred occurrence)
{
    if (occurrence == ErrorInstance::FoundApproximateSource)
        return defaultApproximateSourceError(originalMessage, sourceText);

    ASSERT(occurrence == ErrorInstance::FoundExactSource);
    size_t instanceofIndex = sourceText.reverseFind(ASCIILiteral("instanceof"));
    // This can happen when Symbol.hasInstance function is directly called.
    if (instanceofIndex == notFound)
        return originalMessage;

    if (sourceText.find(ASCIILiteral("instanceof")) != instanceofIndex)
        return tryMakeString(originalMessage, " (evaluating '", sourceText, "')");

    static const unsigned instanceofLength = 10;
    String rightHandSide = sourceText.substring(instanceofIndex + instanceofLength);
    return tryMakeString(rightHandSide, content, ". (evaluating '", sourceText, "')");
}

static String invalidParameterInstanceofNotFunctionSourceAppender(const String& originalMessage, const String& sourceText, RuntimeType runtimeType, ErrorInstance::SourceTextWhereErrorOccurred occurrence)
{
	return invalidParameterInstanceofSourceAppender(WTF::ASCIILiteral(" is not a function"), originalMessage, sourceText, runtimeType, occurrence);
}

static String invalidParameterInstanceofhasInstanceValueNotFunctionSourceAppender(const String& originalMessage, const String& sourceText, RuntimeType runtimeType, ErrorInstance::SourceTextWhereErrorOccurred occurrence)
{
	return invalidParameterInstanceofSourceAppender(WTF::ASCIILiteral("[Symbol.hasInstance] is not a function, undefined, or null"), originalMessage, sourceText, runtimeType, occurrence);
}

JSObject* createError(ExecState* exec, JSValue value, const String& message, ErrorInstance::SourceAppender appender)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_CATCH_SCOPE(vm);

    String valueDescription = errorDescriptionForValue(exec, value);
    if (vm.exception() || !valueDescription) {
        // When we see an exception, we're not returning immediately because
        // we're in a CatchScope, i.e. no exceptions are thrown past this scope.
        // We're using a CatchScope because the contract for createError() is
        // that it only creates an error object; it doesn't throw it.
        vm.clearException();
        return createOutOfMemoryError(exec);
    }
    String errorMessage = tryMakeString(valueDescription, ' ', message);
    if (!errorMessage)
        return createOutOfMemoryError(exec);
    //ASSERT_UNUSED(scope, !scope.exception());
    JSObject* exception = createTypeError(exec, errorMessage, appender, runtimeTypeForValue(vm, value));
    ASSERT(exception->isErrorInstance());
    return exception;
}

JSObject* createInvalidFunctionApplyParameterError(ExecState* exec, JSValue value)
{
    VM& vm = exec->vm();
    JSObject* exception = createTypeError(exec, ASCIILiteral("second argument to Function.prototype.apply must be an Array-like object"), defaultSourceAppender, runtimeTypeForValue(vm, value));
    ASSERT(exception->isErrorInstance());
    return exception;
}

JSObject* createInvalidInParameterError(ExecState* exec, JSValue value)
{
    return createError(exec, value, ASCIILiteral("is not an Object."), invalidParameterInSourceAppender);
}

JSObject* createInvalidInstanceofParameterErrorNotFunction(ExecState* exec, JSValue value)
{
    return createError(exec, value, ASCIILiteral(" is not a function"), invalidParameterInstanceofNotFunctionSourceAppender);
}

JSObject* createInvalidInstanceofParameterErrorhasInstanceValueNotFunction(ExecState* exec, JSValue value)
{
    return createError(exec, value, ASCIILiteral("[Symbol.hasInstance] is not a function, undefined, or null"), invalidParameterInstanceofhasInstanceValueNotFunctionSourceAppender);
}

JSObject* createNotAConstructorError(ExecState* exec, JSValue value)
{
    return createError(exec, value, ASCIILiteral("is not a constructor"), defaultSourceAppender);
}

JSObject* createNotAFunctionError(ExecState* exec, JSValue value)
{
    return createError(exec, value, ASCIILiteral("is not a function"), notAFunctionSourceAppender);
}

JSObject* createNotAnObjectError(ExecState* exec, JSValue value)
{
    return createError(exec, value, ASCIILiteral("is not an object"), defaultSourceAppender);
}

JSObject* createErrorForDuplicateGlobalVariableDeclaration(ExecState* exec, UniquedStringImpl* key)
{
    return createSyntaxError(exec, tryMakeString("Can't create duplicate variable: '", StringView(key), '\''));
}

JSObject* createErrorForInvalidGlobalFunctionDeclaration(ExecState* exec, const Identifier& ident)
{
    return createTypeError(exec, makeString("Can't declare global function '", ident.string(), "': property must be either configurable or both writable and enumerable"));
}

JSObject* createErrorForInvalidGlobalVarDeclaration(ExecState* exec, const Identifier& ident)
{
    return createTypeError(exec, makeString("Can't declare global variable '", ident.string(), "': global object must be extensible"));
}

JSObject* createTDZError(ExecState* exec)
{
    return createReferenceError(exec, ASCIILiteral("Cannot access uninitialized variable."));
}

JSObject* createInvalidPrivateNameError(ExecState* exec)
{
    return createTypeError(exec, ASCIILiteral("Cannot access invalid private field"), defaultSourceAppender, TypeNothing);
}

JSObject* createRedefinedPrivateNameError(ExecState* exec)
{
    return createTypeError(exec, ASCIILiteral("Cannot redefine existing private field"), defaultSourceAppender, TypeNothing);
}

JSObject* createPrivateMethodAccessError(ExecState* exec)
{
    return createTypeError(exec, ASCIILiteral("Cannot access private method or acessor"), defaultSourceAppender, TypeNothing);
}

JSObject* createReinstallPrivateMethodError(ExecState* exec)
{
    return createTypeError(exec, ASCIILiteral("Cannot install same private methods on object more than once"), defaultSourceAppender, TypeNothing);
}

Exception* throwOutOfMemoryError(JSGlobalData& vm, ExecState* exec, const String& message)
{
    return vm.throwException(exec, createOutOfMemoryError(exec, message));
}

Exception* throwOutOfMemoryError(JSGlobalData& vm, ExecState* exec)
{
    return vm.throwException(exec, createOutOfMemoryError(exec));
}

Exception* throwOutOfMemoryError(ExecState* exec)
{
    return throwOutOfMemoryError(exec->vm(), exec);
}

Exception* throwStackOverflowError(ExecState* exec)
{
    Interpreter::ErrorHandlingMode mode(exec);
    return exec->vm().throwException(exec, createStackOverflowError(exec));
}

Exception* throwTerminatedExecutionException(ExecState* exec, JSGlobalData& vm)
{
    return vm.throwException(exec, createTerminatedExecutionException(vm));
}

Exception* throwTerminatedExecutionException(ExecState* exec)
{
	return throwTerminatedExecutionException(exec, exec->vm());
}

} // namespace JSC
