/*
 * Copyright (C) 2009, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "Executable.h"

#include "BytecodeGenerator.h"
#include "CodeBlock.h"
#include "DFGDriver.h"
#include "DeferGC.h"
#include "Exception.h"
#include "JIT.h"
#include "JITInlineMethods.h"
#include "JSTemplateObjectDescriptor.h"
#include "Operations.h"
#include "Parser.h"
#include "StrictEvalActivation.h"
#include <wtf/CurrentTime.h>
#include <wtf/MemoryMonitor.h>
#include <wtf/Vector.h>
#include <wtf/text/StringBuilder.h>

namespace JSC {

void ScriptExecutable::installCode(CodeBlock* genericCodeBlock)
{
    RELEASE_ASSERT(genericCodeBlock->ownerExecutable() == this);

    JSGlobalData& vm = genericCodeBlock->vm();

    ASSERT(vm.heap.isDeferred());

    CodeSpecializationKind kind = genericCodeBlock->specializationKind();

    RefPtr<CodeBlock> oldCodeBlock;

    switch (kind) {
    case CodeForCall:
        m_jitCodeForCall = genericCodeBlock->jitCode();
        m_jitCodeForCallWithArityCheck = MacroAssemblerCodePtr();
        m_jitCodeForCallWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
        m_numParametersForCall = genericCodeBlock->numParameters();
        break;
    case CodeForConstruct:
        m_jitCodeForConstruct = genericCodeBlock->jitCode();
        m_jitCodeForConstructWithArityCheck = MacroAssemblerCodePtr();
        m_jitCodeForConstructWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
        m_numParametersForConstruct = genericCodeBlock->numParameters();
        break;
    }

    switch (genericCodeBlock->codeType()) {
    case GlobalCode: {
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        ProgramCodeBlock* codeBlock = static_cast<ProgramCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_programCodeBlock;
        executable->m_programCodeBlock = codeBlock;
        break;
    }

    case ModuleCode: {
        ModuleProgramExecutable* executable = jsCast<ModuleProgramExecutable*>(this);
        ModuleProgramCodeBlock* codeBlock = static_cast<ModuleProgramCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_moduleProgramCodeBlock;
        executable->m_moduleProgramCodeBlock = codeBlock;
        break;
    }

    case EvalCode: {
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        EvalCodeBlock* codeBlock = static_cast<EvalCodeBlock*>(genericCodeBlock);

        ASSERT(kind == CodeForCall);

        oldCodeBlock = executable->m_evalCodeBlock;
        executable->m_evalCodeBlock = codeBlock;
        break;
    }

    case FunctionCode: {
        FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
        FunctionCodeBlock* codeBlock = static_cast<FunctionCodeBlock*>(genericCodeBlock);

        switch (kind) {
        case CodeForCall:
            oldCodeBlock = executable->m_codeBlockForCall;
            executable->m_codeBlockForCall = codeBlock;
            break;
        case CodeForConstruct:
            oldCodeBlock = executable->m_codeBlockForConstruct;
            executable->m_codeBlockForConstruct = codeBlock;
            break;
        }
        break;
    }
    }

    if (oldCodeBlock)
        oldCodeBlock->unlinkIncomingCalls();

    m_codeCompiledTime = WTF::monotonicallyIncreasingTime();
}

PassRefPtr<CodeBlock> ScriptExecutable::newCodeBlockFor(
	ExecState* exec, CodeSpecializationKind kind, JSFunction* jsFunction, JSScope*& scope, Exception*& exception)
{
    VM& vm = scope->vm();
    JSGlobalObject* globalObject = scope->globalObject();

    ASSERT(vm->heap.isDeferred());

    if (classInfo() == EvalExecutable::info()) {
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_evalCodeBlock);
        RELEASE_ASSERT(!jsFunction);

        if (!globalObject->evalEnabled()) {
        	exception = vm.throwException(exec, createEvalError(exec, globalObject->evalDisabledErrorMessage()));
        	return nullptr;
        }
        ParserError error;
        JSParserStrictMode strictMode = executable->isInStrictContext() ? JSParserStrictMode::Strict : JSParserStrictMode::NotStrict;
        DerivedContextType derivedContextType = executable->derivedContextType();
        EvalContextType evalContextType = executable->evalContextType();
        PrivateNameEnvironment privateNameEnvironment = executable->privateNameEnvironment();

        OwnPtr<EvalNode> evalNode = parse<EvalNode>(
        	&vm, m_source, Identifier(), JSParserBuiltinMode::NotBuiltin, strictMode, JSParserScriptMode::Classic, SourceParseMode::ProgramMode, FunctionMode::None, SuperBinding::NotNeeded, error, ConstructorKind::None, derivedContextType, evalContextType, &privateNameEnvironment);

        if (globalObject->hasDebugger() && !EvalNode::scopeIsFunction)
        	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!evalNode) {
            exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
            return nullptr;
        }

    	if(UNLIKELY(WTF::isLowMemory())) {
            exception = vm.throwException(exec, createOutOfMemoryError(exec));
            return nullptr;
    	}

        unsigned arrowContextFeature = executable->isArrowFunctionContext() ? ArrowFunctionContextFeature : 0;
		recordParse(evalNode->features() | arrowContextFeature, evalNode->hasCapturedVariables(), evalNode->lineNo(), evalNode->lastLine(), startColumn());

        RefPtr<EvalCodeBlock> evalCodeBlock = adoptRef(new EvalCodeBlock(executable, executable->executableInfo(), globalObject, source().provider()));

		if (executable->isInStrictContext())
			scope = StrictEvalActivation::create(exec, scope);

	    RefPtr<TDZEnvironmentLink> parentVariablesUnderTDZ;
	    if (!executable->variablesUnderTDZ().isEmpty())
	    	parentVariablesUnderTDZ = TDZEnvironmentLink::create(vm.m_compactVariableMap->get(executable->variablesUnderTDZ()), nullptr);
		ECMAMode ecmaMode = executable->isInStrictContext() ? ECMAMode::strict() : ECMAMode::sloppy();
        BytecodeGenerator generator(evalNode.get(), globalObject, evalCodeBlock.get(), parentVariablesUnderTDZ, &privateNameEnvironment, nullptr, ecmaMode);

        error = generator.generate(exec, scope);
        if (error.isValid()) {
            exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
        	return nullptr;
        }

        return evalCodeBlock.release();
    }

    if (classInfo() == ProgramExecutable::info()) {
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_programCodeBlock);
        RELEASE_ASSERT(!jsFunction);

        ParserError error;
        JSParserStrictMode strictMode = executable->isInStrictContext() ? JSParserStrictMode::Strict : JSParserStrictMode::NotStrict;
        DerivedContextType derivedContextType = executable->derivedContextType();
        OwnPtr<ProgramNode> programNode = parse<ProgramNode>(
        	&vm, m_source, Identifier(), JSParserBuiltinMode::NotBuiltin, strictMode, JSParserScriptMode::Classic, SourceParseMode::ProgramMode, FunctionMode::None, SuperBinding::NotNeeded, error, ConstructorKind::None, derivedContextType);

        if (globalObject->hasDebugger() && !ProgramNode::scopeIsFunction)
        	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!programNode) {
            exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
            return nullptr;
        }

    	if(UNLIKELY(WTF::isLowMemory())) {
            exception = vm.throwException(exec, createOutOfMemoryError(exec));
            return nullptr;
    	}

        unsigned arrowContextFeature = executable->isArrowFunctionContext() ? ArrowFunctionContextFeature : 0;
        recordParse(programNode->features() | arrowContextFeature, programNode->hasCapturedVariables(), programNode->lineNo(), programNode->lastLine(), startColumn());

        executable->initializeGlobalProperties(vm, programNode.get(), exec, scope);

        RefPtr<ProgramCodeBlock> programCodeBlock = adoptRef(new ProgramCodeBlock(executable, GlobalCode, executable->executableInfo(), scope->globalObject(), source().provider(), source().startColumn()));
        ECMAMode ecmaMode = executable->isInStrictContext() ? ECMAMode::strict() : ECMAMode::sloppy();
        BytecodeGenerator generator(programNode.get(), scope->globalObject(), programCodeBlock.get(), nullptr, nullptr, nullptr, ecmaMode);

        error = generator.generate(exec, scope);
        if (error.isValid()) {
            exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
        	return nullptr;
        }

        return programCodeBlock.release();
    }

    if (classInfo() == ModuleProgramExecutable::info()) {
        ModuleProgramExecutable* executable = jsCast<ModuleProgramExecutable*>(this);
        RELEASE_ASSERT(kind == CodeForCall);
        RELEASE_ASSERT(!executable->m_moduleProgramCodeBlock);
        RELEASE_ASSERT(!jsFunction);

        ParserError error;
        JSParserStrictMode strictMode = JSParserStrictMode::Strict;
        DerivedContextType derivedContextType = executable->derivedContextType();
        OwnPtr<ModuleProgramNode> moduleProgramNode = parse<ModuleProgramNode>(
        	&vm, m_source, Identifier(), JSParserBuiltinMode::NotBuiltin, strictMode, JSParserScriptMode::Module, SourceParseMode::ModuleEvaluateMode, FunctionMode::None, SuperBinding::NotNeeded, error, ConstructorKind::None, derivedContextType);

        if (globalObject->hasDebugger() && !ModuleProgramNode::scopeIsFunction)
        	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

        if (!moduleProgramNode) {
            exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
            return nullptr;
        }

    	if(UNLIKELY(WTF::isLowMemory())) {
            exception = vm.throwException(exec, createOutOfMemoryError(exec));
            return nullptr;
    	}

        unsigned arrowContextFeature = executable->isArrowFunctionContext() ? ArrowFunctionContextFeature : 0;
		recordParse(moduleProgramNode->features() | arrowContextFeature, moduleProgramNode->hasCapturedVariables(), moduleProgramNode->lineNo(), moduleProgramNode->lastLine(), startColumn());

        ECMAMode ecmaMode = ECMAMode::strict();
        RefPtr<ModuleProgramCodeBlock> moduleProgramCodeBlock = adoptRef(new ModuleProgramCodeBlock(executable, executable->executableInfo(), scope->globalObject(), source().provider(), source().startColumn()));
        BytecodeGenerator generator(moduleProgramNode.get(), scope->globalObject(), moduleProgramCodeBlock.get(), nullptr, nullptr, nullptr, ecmaMode);

        error = generator.generate(exec, scope);
        if (error.isValid()) {
            exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
        	return nullptr;
        }

        executable->m_moduleEnvironmentSymbolTable.set(vm, executable, jsCast<SymbolTable*>(moduleProgramCodeBlock->constantRegister(moduleProgramCodeBlock->moduleEnvironmentSymbolTableConstantRegisterOffset()).get())->cloneScopePart(vm));

        return moduleProgramCodeBlock.release();
    }

    RELEASE_ASSERT(classInfo() == FunctionExecutable::info());
    RELEASE_ASSERT(jsFunction);
    FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
    RELEASE_ASSERT(!executable->codeBlockFor(kind));

    exception = nullptr;
    ParserError error;
    JSParserBuiltinMode builtinMode = executable->isBuiltinFunction() ? JSParserBuiltinMode::Builtin : JSParserBuiltinMode::NotBuiltin;
    JSParserStrictMode strictMode = executable->isInStrictContext() ? JSParserStrictMode::Strict : JSParserStrictMode::NotStrict;
    ASSERT(isFunctionParseMode(executable->parseMode()));
    const Vector<Identifier>* generatorOrAsyncWrapperFunctionParameterNames = executable->generatorOrAsyncWrapperFunctionParameterNames();
    Vector<JSTextPosition>* classFieldLocations = executable->classFieldLocations();
    OwnPtr<FunctionNode> function = parse<FunctionNode>(
        &vm, m_source, executable->name(), builtinMode, strictMode, executable->scriptMode(), executable->parseMode(), executable->functionMode(), executable->superBinding(), error, executable->constructorKind(), executable->derivedContextType(), EvalContextType::None, nullptr, classFieldLocations);

    if (globalObject->hasDebugger() && !FunctionNode::scopeIsFunction)
    	globalObject->debugger()->sourceParsed(exec, m_source.provider(), error.line(), error.message());

    if (!function) {
        exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
        return nullptr;
    }

	if(UNLIKELY(WTF::isLowMemory())) {
        exception = vm.throwException(exec, createOutOfMemoryError(exec));
        return nullptr;
	}

    function->finishParsing(executable->m_name, executable->functionMode());
    recordParse(function->features(), function->hasCapturedVariables(), function->lineNo(), function->lastLine(), startColumn());

    bool usesEval = function->features() & EvalFeature;
	bool isClassContext = executable->superBinding() == SuperBinding::Needed || executable->parseMode() == SourceParseMode::ClassFieldInitializerMode;

    SourceProvider* provider = source().provider();
    unsigned sourceOffset = source().startOffset();
    unsigned startColumn = source().startColumn();

    RefPtr<FunctionCodeBlock> functionCodeBlock = adoptRef(new FunctionCodeBlock(executable, FunctionCode, ExecutableInfo(usesEval, kind == CodeForConstruct, executable->privateBrandRequirement(), executable->isBuiltinFunction(), executable->constructorKind(), executable->scriptMode(), executable->superBinding(), executable->parseMode(), executable->derivedContextType(), executable->needsClassFieldInitializer(), false, isClassContext, EvalContextType::FunctionEvalContext), scope->globalObject(), provider, sourceOffset, startColumn));

    auto parentScopeTDZVariables = executable->parentScopeTDZVariables();
    PrivateNameEnvironment parentPrivateNameEnvironment = executable->parentPrivateNameEnvironment();
	ECMAMode ecmaMode = function->features() & StrictModeFeature ? ECMAMode::strict() : ECMAMode::sloppy();
	BytecodeGenerator generator(function.get(), scope->globalObject(), functionCodeBlock.get(), parentScopeTDZVariables, &parentPrivateNameEnvironment, generatorOrAsyncWrapperFunctionParameterNames, ecmaMode);
    error = generator.generate(exec, scope);
    if (error.isValid()) {
        exception = vm.throwException(exec, error.toErrorObject(globalObject, m_source));
		return nullptr;
    }

	return functionCodeBlock.release();
}

PassRefPtr<CodeBlock> ScriptExecutable::newReplacementCodeBlockFor(
    CodeSpecializationKind kind)
{
    if (classInfo() == EvalExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        EvalExecutable* executable = jsCast<EvalExecutable*>(this);
        RefPtr<EvalCodeBlock> result = adoptRef(new EvalCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_evalCodeBlock));
        result->setAlternative(executable->m_evalCodeBlock);
        return result;
    }

    if (classInfo() == ProgramExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        ProgramExecutable* executable = jsCast<ProgramExecutable*>(this);
        RefPtr<ProgramCodeBlock> result = adoptRef(new ProgramCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_programCodeBlock));
        result->setAlternative(executable->m_programCodeBlock);
        return result;
    }

    if (classInfo() == ModuleProgramExecutable::info()) {
        RELEASE_ASSERT(kind == CodeForCall);
        ModuleProgramExecutable* executable = jsCast<ModuleProgramExecutable*>(this);
        RefPtr<ModuleProgramCodeBlock> result = adoptRef(new ModuleProgramCodeBlock(
            CodeBlock::CopyParsedBlock, *executable->m_moduleProgramCodeBlock));
        result->setAlternative(executable->m_moduleProgramCodeBlock);
        return result;
    }

    RELEASE_ASSERT(classInfo() == FunctionExecutable::info());
    FunctionExecutable* executable = jsCast<FunctionExecutable*>(this);
    RefPtr<FunctionCodeBlock> result = adoptRef(new FunctionCodeBlock(
        CodeBlock::CopyParsedBlock, *executable->codeBlockFor(kind)));
    result->setAlternative(executable->codeBlockFor(kind));
    return result;
}

Exception* ScriptExecutable::prepareForExecutionImpl(ExecState* exec, JSFunction* jsFunction, JSScope*& scope, CodeSpecializationKind kind, CodeBlock*& resultCodeBlock)
{
    VM& vm = exec->vm();
    DeferGCForAWhile deferGC(vm.heap);

    Exception* exception = nullptr;
    RefPtr<CodeBlock> codeBlock = newCodeBlockFor(exec, kind, jsFunction, scope, exception);
    resultCodeBlock = codeBlock.get();
    if (!codeBlock) {
        RELEASE_ASSERT(exception);
        if(!exception)
        	exception = vm.throwException(exec, createOutOfMemoryError(exec));
        return exception;
    }

	if(UNLIKELY(WTF::isLowMemory())) {
        if(!exception)
        	exception = vm.throwException(exec, createOutOfMemoryError(exec));
        return exception;
	}

#if ENABLE(LLINT)
    LLInt::setEntrypoint(vm, codeBlock.get());
#else
    CompilationResult result = JIT::compile(&vm, codeBlock.get(), JITCompilationCanFail);
    RELEASE_ASSERT(result == CompilationSuccessful);
#endif

    if(result == CompilationSuccessful) {
    	installCode(codeBlock.get());
    	return nullptr;
    }

    return vm.throwException(exec, createOutOfMemoryError(exec));
}

ScriptExecutable* ScriptExecutable::topLevelExecutable()
{
    switch (structure()->typeInfo().type()) {
    case FunctionExecutableType:
        return jsCast<FunctionExecutable*>(this)->topLevelExecutable();
    default:
        return this;
    }
}

JSArray* ScriptExecutable::createTemplateObject(ExecState* exec, JSTemplateObjectDescriptor* descriptor)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    TemplateObjectMap& templateObjectMap = ensureTemplateObjectMap(vm);
	//auto locker = holdLock(cellLock());
	auto result = templateObjectMap.add(descriptor->endOffset(), WriteBarrier<JSArray>());
    if (JSArray* array = result.iterator->second.get())
        return array;
    JSArray* templateObject = descriptor->createTemplateObject(exec);
    RETURN_IF_EXCEPTION(scope, nullptr);
    result.iterator->second.set(vm, this, templateObject);
    return templateObject;
}

auto ScriptExecutable::ensureTemplateObjectMapImpl(OwnPtr<TemplateObjectMap>& dest) -> TemplateObjectMap&
{
    if (dest)
        return *dest;
    OwnPtr<TemplateObjectMap> result = adoptPtr(new TemplateObjectMap());
    WTF::storeStoreFence();
    dest = result.release();
    return *dest;
}

auto ScriptExecutable::ensureTemplateObjectMap(VM& vm) -> TemplateObjectMap&
{
    switch (structure()->typeInfo().type()) {
    case FunctionExecutableType:
        return static_cast<FunctionExecutable*>(this)->ensureTemplateObjectMap(vm);
    case EvalExecutableType:
        return static_cast<EvalExecutable*>(this)->ensureTemplateObjectMap(vm);
    case ProgramExecutableType:
        return static_cast<ProgramExecutable*>(this)->ensureTemplateObjectMap(vm);
    case ModuleProgramExecutableType:
    default:
        ASSERT(type() == ModuleProgramExecutableType);
        return static_cast<ModuleProgramExecutable*>(this)->ensureTemplateObjectMap(vm);
    }
}

const ClassInfo ExecutableBase::s_info = { "Executable", nullptr, nullptr, CREATE_METHOD_TABLE(ExecutableBase) };

void ExecutableBase::destroy(JSCell* cell)
{
    static_cast<ExecutableBase*>(cell)->ExecutableBase::~ExecutableBase();
}

void ExecutableBase::clearCode()
{
#if ENABLE(JIT)
    m_jitCodeForCall.clear();
    m_jitCodeForConstruct.clear();
    m_jitCodeForCallWithArityCheck = MacroAssemblerCodePtr();
    m_jitCodeForConstructWithArityCheck = MacroAssemblerCodePtr();
    m_jitCodeForCallWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
    m_jitCodeForConstructWithArityCheckAndPreserveRegs = MacroAssemblerCodePtr();
#endif
    m_numParametersForCall = NUM_PARAMETERS_NOT_COMPILED;
    m_numParametersForConstruct = NUM_PARAMETERS_NOT_COMPILED;
}

const ClassInfo NativeExecutable::s_info = { "NativeExecutable", &ExecutableBase::s_info, 0, CREATE_METHOD_TABLE(NativeExecutable) };

void NativeExecutable::destroy(JSCell* cell)
{
    static_cast<NativeExecutable*>(cell)->NativeExecutable::~NativeExecutable();
}

const ClassInfo ScriptExecutable::s_info = { "ScriptExecutable", &ExecutableBase::s_info, 0, CREATE_METHOD_TABLE(ScriptExecutable) };

ScriptExecutable::ScriptExecutable(Structure* structure, JSGlobalData& globalData, const SourceCode& source, bool isInStrictContext, DerivedContextType derivedContextType, bool isInArrowFunctionContext, Intrinsic intrinsic, EvalContextType evalContextType)
: ExecutableBase(globalData, structure, NUM_PARAMETERS_NOT_COMPILED, intrinsic)
, m_features(isInStrictContext ? StrictModeFeature : 0)
, m_didTryToEnterInLoop(false)
, m_hasCapturedVariables(false)
, m_neverInline(false)
, m_neverOptimize(false)
, m_isArrowFunctionContext(isInArrowFunctionContext)
, m_derivedContextType(static_cast<unsigned>(derivedContextType))
, m_evalContextType(static_cast<unsigned>(evalContextType))
, m_firstLine(-1)
, m_lastLine(-1)
, m_startColumn(UINT_MAX)
, m_source(source)
{
}

void ScriptExecutable::destroy(JSCell* cell)
{
    static_cast<ScriptExecutable*>(cell)->ScriptExecutable::~ScriptExecutable();
}

const ClassInfo EvalExecutable::s_info = { "EvalExecutable", &ScriptExecutable::s_info, nullptr, CREATE_METHOD_TABLE(EvalExecutable) };

EvalExecutable::EvalExecutable(JSGlobalData& vm, const SourceCode& source, bool inStrictContext, DerivedContextType derivedContextType, NeedsClassFieldInitializer needsClassFieldInitializer, PrivateBrandRequirement privateBrandRequirement, bool isArrowFunctionContext, EvalContextType evalContextType, const TDZEnvironment* variablesUnderTDZ, const PrivateNameEnvironment* privateNameEnvironment)
    : ScriptExecutable(vm.evalExecutableStructure.get(), vm, source, inStrictContext, derivedContextType, isArrowFunctionContext, NoIntrinsic, evalContextType)
	, m_needsClassFieldInitializer(static_cast<unsigned>(needsClassFieldInitializer))
	, m_privateBrandRequirement(static_cast<unsigned>(privateBrandRequirement))
{
	ASSERT((needsClassFieldInitializer == NeedsClassFieldInitializer::No && privateBrandRequirement == PrivateBrandRequirement::None) || derivedContextType == DerivedContextType::DerivedConstructorContext);
	if (variablesUnderTDZ)
		m_variablesUnderTDZ = *variablesUnderTDZ;
    if (privateNameEnvironment)
    	m_privateNameEnvironment = *privateNameEnvironment;
}

void EvalExecutable::destroy(JSCell* cell)
{
    static_cast<EvalExecutable*>(cell)->EvalExecutable::~EvalExecutable();
}

unsigned EvalExecutable::numVariables()
{
	return m_evalCodeBlock->numVariables();
}

unsigned EvalExecutable::numFunctionHoistingCandidates()
{
	return m_evalCodeBlock->numFunctionHoistingCandidates();
}

unsigned EvalExecutable::numTopLevelFunctionDecls()
{
	return m_evalCodeBlock->numberOfFunctionDecls();
}

const ClassInfo ProgramExecutable::s_info = { "ProgramExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(ProgramExecutable) };

ProgramExecutable::ProgramExecutable(JSGlobalData& vm, const SourceCode& source)
    : ScriptExecutable(vm.programExecutableStructure.get(), vm, source, false, DerivedContextType::None, false, NoIntrinsic, EvalContextType::None)
{
}

void ProgramExecutable::destroy(JSCell* cell)
{
    static_cast<ProgramExecutable*>(cell)->ProgramExecutable::~ProgramExecutable();
}

const ClassInfo ModuleProgramExecutable::s_info = { "ModuleProgramExecutable", &ScriptExecutable::s_info, 0, CREATE_METHOD_TABLE(ModuleProgramExecutable) };

ModuleProgramExecutable::ModuleProgramExecutable(JSGlobalData& vm, const SourceCode& source)
    : ScriptExecutable(vm.moduleProgramExecutableStructure.get(), vm, source, false, DerivedContextType::None, false, NoIntrinsic, EvalContextType::None)
{
#if 0
    m_typeProfilingStartOffset = 0;
    m_typeProfilingEndOffset = source.length() - 1;
    if (exec->vm().typeProfiler() || exec->vm().controlFlowProfiler())
        exec->vm().functionHasExecutedCache()->insertUnexecutedRange(sourceID(), m_typeProfilingStartOffset, m_typeProfilingEndOffset);
#endif
}

ModuleProgramExecutable* ModuleProgramExecutable::create(JSGlobalData& vm, const SourceCode& source)
{
    ModuleProgramExecutable* executable = new (NotNull, allocateCell<ModuleProgramExecutable>(vm.heap)) ModuleProgramExecutable(vm, source);
    executable->finishCreation(vm);
    return executable;
}

void ModuleProgramExecutable::destroy(JSCell* cell)
{
    static_cast<ModuleProgramExecutable*>(cell)->ModuleProgramExecutable::~ModuleProgramExecutable();
}

const ClassInfo FunctionExecutable::s_info = { "FunctionExecutable", &ScriptExecutable::s_info, nullptr, CREATE_METHOD_TABLE(FunctionExecutable) };

FunctionExecutable::FunctionExecutable(JSGlobalData& vm, FunctionMetadataNode* node, unsigned startColumn, bool isBuiltinFunction, ConstructAbility constructAbility, JSParserScriptMode scriptMode, RefPtr<TDZEnvironmentLink> parentScopeTDZVariables, Optional<PrivateNameEnvironment> parentPrivateNameEnvironment, PassOwnPtr<Vector<Identifier>> generatorOrAsyncWrapperFunctionParameterNames, DerivedContextType derivedContextType, NeedsClassFieldInitializer needsClassFieldInitializer, PrivateBrandRequirement privateBrandRequirement, Intrinsic intrinsic)
    : ScriptExecutable(vm.functionExecutableStructure.get(), vm, node->source(), node->isInStrictContext(), derivedContextType, false, intrinsic, EvalContextType::None)
    , m_classSource(node->classSource())
	, m_parentScopeTDZVariables(parentScopeTDZVariables)
	, m_parameterCount(node->parameterCount())
	, m_sourceParseMode(node->parseMode())
	, m_scriptMode(scriptMode)
	, m_functionMode(static_cast<unsigned>(node->functionMode()))
	, m_isBuiltinFunction(isBuiltinFunction)
	, m_constructorKind(static_cast<unsigned>(node->constructorKind()))
	, m_constructAbility(static_cast<unsigned>(constructAbility))
	, m_superBinding(static_cast<unsigned>(node->superBinding()))
	, m_needsClassFieldInitializer(static_cast<unsigned>(needsClassFieldInitializer))
	, m_privateBrandRequirement(static_cast<unsigned>(privateBrandRequirement))
    , m_name(node->ident())
	, m_ecmaName(node->ecmaName())
	, m_generatorOrAsyncWrapperFunctionParameterNames(generatorOrAsyncWrapperFunctionParameterNames)
{
	m_firstLine = node->lineNo();
	m_lastLine = node->lastLine();
	m_startColumn = startColumn;
    m_parametersStartOffset = node->parametersStart();

    // Make sure these bitfields are adequately wide.
    ASSERT(m_constructAbility == static_cast<unsigned>(constructAbility));
    ASSERT(m_constructorKind == static_cast<unsigned>(node->constructorKind()));
    ASSERT(m_functionMode == static_cast<unsigned>(node->functionMode()));
    ASSERT(m_superBinding == static_cast<unsigned>(node->superBinding()));
    ASSERT(m_derivedContextType == static_cast<unsigned>(derivedContextType));
    ASSERT(m_privateBrandRequirement == static_cast<unsigned>(privateBrandRequirement));
    ASSERT(!m_needsClassFieldInitializer || (isClassConstructorFunction() || derivedContextType == DerivedContextType::DerivedConstructorContext));
    if (parentPrivateNameEnvironment)
		m_parentPrivateNameEnvironment = WTFMove(*parentPrivateNameEnvironment);
}

FunctionCodeBlock* FunctionExecutable::codeBlockFor(CodeSpecializationKind kind)
{
	if (kind == CodeForCall)
		return m_codeBlockForCall.get();
	ASSERT(kind == CodeForConstruct);
	return m_codeBlockForConstruct.get();
}

void FunctionExecutable::finishCreation(VM& vm, ScriptExecutable* topLevelExecutable)
{
	Base::finishCreation(vm);
    m_topLevelExecutable.set(vm, this, topLevelExecutable ? topLevelExecutable : this);
    m_singletonFunction.set(vm, this, InferredValue::create(vm));
}

void FunctionExecutable::destroy(JSCell* cell)
{
    static_cast<FunctionExecutable*>(cell)->FunctionExecutable::~FunctionExecutable();
}

inline const char* samplingDescription(JITCode::JITType jitType)
{
    switch (jitType) {
    case JITCode::InterpreterThunk:
        return "Interpreter Compilation (TOTAL)";
    case JITCode::BaselineJIT:
        return "Baseline Compilation (TOTAL)";
    case JITCode::DFGJIT:
        return "DFG Compilation (TOTAL)";
    case JITCode::FTLJIT:
        return "FTL Compilation (TOTAL)";
    default:
        ASSERT_NOT_REACHED();
        return nullptr;
    }
}

EvalExecutable* EvalExecutable::create(JSGlobalData& vm, const SourceCode& source, bool isInStrictContext, DerivedContextType derivedContextType, NeedsClassFieldInitializer needsClassFieldInitializer, PrivateBrandRequirement privateBrandRequirement, bool isArrowFunctionContext, EvalContextType evalContextType, const TDZEnvironment* variablesUnderTDZ, const PrivateNameEnvironment* privateNameEnvironment)
{
	EvalExecutable* executable = new (NotNull, allocateCell<EvalExecutable>(vm.heap)) EvalExecutable(vm, source, isInStrictContext, derivedContextType, needsClassFieldInitializer, privateBrandRequirement, isArrowFunctionContext, evalContextType, variablesUnderTDZ, privateNameEnvironment);
	executable->finishCreation(vm);
	return executable;
}

auto EvalExecutable::ensureTemplateObjectMap(VM&) -> TemplateObjectMap&
{
    return ensureTemplateObjectMapImpl(m_templateObjectMap);
}

void EvalExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    EvalExecutable* thisObject = jsCast<EvalExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    if (thisObject->m_evalCodeBlock)
        thisObject->m_evalCodeBlock->visitAggregate(visitor);
    if (TemplateObjectMap* map = thisObject->m_templateObjectMap.get()) {
        //auto locker = holdLock(thisObject->cellLock());
    	for(auto it = map->begin(); it != map->end(); ++it)
            visitor.append(it->second);
    }
}

void EvalExecutable::clearCode()
{
    m_evalCodeBlock.clear();
    Base::clearCode();
}

// http://www.ecma-international.org/ecma-262/6.0/index.html#sec-hasrestrictedglobalproperty
enum class GlobalPropertyLookUpStatus {
    NotFound,
    Configurable,
    NonConfigurable,
};
static GlobalPropertyLookUpStatus hasRestrictedGlobalProperty(ExecState* exec, JSGlobalObject* globalObject, PropertyName propertyName)
{
    PropertyDescriptor descriptor;
    if (!globalObject->getOwnPropertyDescriptor(exec, propertyName, descriptor))
        return GlobalPropertyLookUpStatus::NotFound;
    if (descriptor.configurable())
        return GlobalPropertyLookUpStatus::Configurable;
    return GlobalPropertyLookUpStatus::NonConfigurable;
}

JSObject* ProgramExecutable::initializeGlobalProperties(VM& vm, ProgramNode* programNode, CallFrame* callFrame, JSScope* scope)
{
    ThrowScope throwScope(vm);
    RELEASE_ASSERT(scope);
    JSGlobalObject* globalObject = scope->globalObject();
    RELEASE_ASSERT(globalObject);
    ASSERT(&globalObject->vm() == &vm);

    JSValue nextPrototype = globalObject->getPrototypeDirect();
    while (nextPrototype && nextPrototype.isObject()) {
        if (UNLIKELY(asObject(nextPrototype)->structure()->typeInfo().type() == ProxyObjectType))
            return createTypeError(callFrame, ASCIILiteral("Proxy is not allowed in the global prototype chain."));
        nextPrototype = asObject(nextPrototype)->getPrototypeDirect();
    }

    JSGlobalLexicalEnvironment* globalLexicalEnvironment = globalObject->globalLexicalEnvironment();
    bool hasGlobalLexicalDeclarations = !globalLexicalEnvironment->isEmpty();
	const VariableEnvironment& variableDeclarations = programNode->varDeclarations();
    const VariableEnvironment& lexicalDeclarations = programNode->lexicalVariables();
    size_t numberOfFunctions = programNode->numberOfFunctionDecls();
	// The ES6 spec says that no vars/global properties/let/const can be duplicated in the global scope.
    // This carried out section 15.1.8 of the ES6 spec: http://www.ecma-international.org/ecma-262/6.0/index.html#sec-globaldeclarationinstantiation
    {
        // Check for intersection of "var" and "let"/"const"/"class"
        // Check if any new "let"/"const"/"class" will shadow any pre-existing global property names (with configurable = false), or "var"/"let"/"const" variables.
        // It's an error to introduce a shadow.
        for (auto& entry : lexicalDeclarations) {
            if (hasGlobalLexicalDeclarations) {
                bool hasProperty = globalLexicalEnvironment->hasProperty(callFrame, entry.first.get());
                throwScope.assertNoExceptionExceptTermination();
                if (hasProperty) {
                    if (UNLIKELY(entry.second.isConst() && !vm.globalConstRedeclarationShouldThrow() && !isInStrictContext())) {
                        // We only allow "const" duplicate declarations under this setting.
                        // For example, we don't allow "let" variables to be overridden by "const" variables.
                        if (globalLexicalEnvironment->isConstVariable(entry.first.get()))
                            continue;
                    }
                    return createErrorForDuplicateGlobalVariableDeclaration(callFrame, entry.first.get());
                }
            }

            // The ES6 spec says that RestrictedGlobalProperty can't be shadowed.
        	GlobalPropertyLookUpStatus status = hasRestrictedGlobalProperty(callFrame, globalObject, entry.first.get());
        	RETURN_IF_EXCEPTION(throwScope, 0);
            switch (status) {
            case GlobalPropertyLookUpStatus::NonConfigurable:
				return createSyntaxError(callFrame, tryMakeString("Can't create duplicate variable that shadows a global property: '", String(entry.first.get()), "'"));

            case GlobalPropertyLookUpStatus::Configurable:
				// Lexical bindings can shadow global properties if the given property's attribute is configurable.
				// https://tc39.github.io/ecma262/#sec-globaldeclarationinstantiation step 5-c, `hasRestrictedGlobal` becomes false
				// However we may emit GlobalProperty look up in bytecodes already and it may cache the value for the global scope.
            	// To make it invalid,
				// 1. In LLInt and Baseline, we bump the global lexical binding epoch and it works.
				// 3. In DFG and FTL, we watch the watchpoint and jettison once it is fired.
				break;
			case GlobalPropertyLookUpStatus::NotFound:
				break;
			}
        }

        // Check if any new "var"s will shadow any previous "let"/"const"/"class" names.
        // It's an error to introduce a shadow.
        if (hasGlobalLexicalDeclarations) {
            for (auto& entry : variableDeclarations) {
                if (entry.second.isSloppyModeHoistedFunction())
                    continue;
                bool hasProperty = globalLexicalEnvironment->hasProperty(callFrame, entry.first.get());
                throwScope.assertNoExceptionExceptTermination();
                if (hasProperty)
                    return createErrorForDuplicateGlobalVariableDeclaration(callFrame, entry.first.get());
            }
        }

        for (size_t i = 0; i < numberOfFunctions; ++i) {
            auto* functionDecl = programNode->functionDecl(i);
            ASSERT(!functionDecl->name().isEmpty());
            bool canDeclare = globalObject->canDeclareGlobalFunction(callFrame, functionDecl->ident());
            throwScope.assertNoExceptionExceptTermination();
            if (!canDeclare)
                return createErrorForInvalidGlobalFunctionDeclaration(callFrame, functionDecl->ident());
        }

        if (!globalObject->isStructureExtensible()) {
            for (auto& entry : variableDeclarations) {
                if (entry.second.isFunction() || entry.second.isSloppyModeHoistedFunction())
                    continue;
                ASSERT(entry.value.isVar());
                const Identifier& ident = Identifier::fromUid(vm, entry.first.get());
                bool canDeclare = globalObject->canDeclareGlobalVar(callFrame, ident);
                throwScope.assertNoExceptionExceptTermination();
                if (!canDeclare)
                    return createErrorForInvalidGlobalVarDeclaration(callFrame, ident);
            }
        }
    }

    BatchedTransitionOptimizer optimizer(vm, globalObject);

    // https://tc39.es/ecma262/#sec-web-compat-globaldeclarationinstantiation (excluding last step)
    if (!isInStrictContext()) {
        for (auto& entry : variableDeclarations) {
            if (!entry.second.isSloppyModeHoistedFunction())
                continue;

            const Identifier& ident = Identifier::fromUid(vm, entry.first.get());

            if (hasGlobalLexicalDeclarations) {
                bool hasProperty = globalLexicalEnvironment->hasProperty(callFrame, ident);
                throwScope.assertNoExceptionExceptTermination();
                if (hasProperty)
                    continue;
            }

            bool canDeclare = globalObject->canDeclareGlobalVar(callFrame, ident);
            throwScope.assertNoExceptionExceptTermination();
            if (!canDeclare)
                continue;

            globalObject->createGlobalVarBinding<BindingCreationContext::Global>(callFrame, ident);
            throwScope.assertNoExceptionExceptTermination();
        }
    }

    for (size_t i = 0; i < numberOfFunctions; ++i) {
        auto* functionDecl = programNode->functionDecl(i);
        ASSERT(!functionDecl->name().isEmpty());
        globalObject->createGlobalFunctionBinding<BindingCreationContext::Global>(callFrame, functionDecl->ident());
        throwScope.assertNoExceptionExceptTermination();
    }

    for (auto& entry : variableDeclarations) {
        if (entry.second.isFunction() || entry.second.isSloppyModeHoistedFunction())
            continue;
        ASSERT(entry.second.isVar());
        globalObject->createGlobalVarBinding<BindingCreationContext::Global>(callFrame, Identifier::fromUid(vm, entry.first.get()));
        throwScope.assertNoExceptionExceptTermination();
    }

    {
        SymbolTable* symbolTable = globalLexicalEnvironment->symbolTable();
        //ConcurrentJITLocker locker(symbolTable->m_lock);
        for (auto it = lexicalDeclarations.begin(); it != lexicalDeclarations.end(); ++it) {
            if (UNLIKELY(it->second.isConst() && !vm.globalConstRedeclarationShouldThrow() && !isInStrictContext())) {
                if (symbolTable->contains(it->first.get()))
                    continue;
            }
            ScopeOffset offset = symbolTable->takeNextScopeOffset();
            SymbolTableEntry newEntry(VarOffset(offset), it->second.isConst() ? ReadOnly : 0);
            newEntry.prepareToWatch();
            symbolTable->add(it->first.get(), newEntry);

            ScopeOffset offsetForAssert = globalLexicalEnvironment->addVariables(1, jsTDZValue());
            RELEASE_ASSERT(offsetForAssert == offset);
        }
    }
    if (lexicalDeclarations.totalSize()) {
#if ENABLE(DFG_JIT)
        for (auto it = lexicalDeclarations.begin(); it != lexicalDeclarations.end(); ++it) {
            // If WatchpointSet exists, just fire it. Since DFG WatchpointSet addition is also done on the main thread, we can sync them.
            // So that we do not create WatchpointSet here. DFG will create if necessary on the main thread.
            // And it will only create not-invalidated watchpoint set if the global lexical environment binding doesn't exist, which is why this code works.
            if (auto* watchpointSet = globalObject->getReferencedPropertyWatchpointSet(it->first.get()))
                watchpointSet->fireAll(vm, "Lexical binding shadows an existing global property");
        }
#endif
        globalObject->bumpGlobalLexicalBindingEpoch(vm);
    }
    return nullptr;
}

auto ProgramExecutable::ensureTemplateObjectMap(VM&) -> TemplateObjectMap&
{
    return ensureTemplateObjectMapImpl(m_templateObjectMap);
}

void ProgramExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    ProgramExecutable* thisObject = jsCast<ProgramExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    if (thisObject->m_programCodeBlock)
        thisObject->m_programCodeBlock->visitAggregate(visitor);
    if (TemplateObjectMap* map = thisObject->m_templateObjectMap.get()) {
        //auto locker = holdLock(thisObject->cellLock());
        for (auto it = map->begin(); it != map->end(); ++it)
            visitor.append(it->second);
    }
}

void ProgramExecutable::clearCode()
{
    m_programCodeBlock.clear();
    Base::clearCode();
}

auto ModuleProgramExecutable::ensureTemplateObjectMap(VM&) -> TemplateObjectMap&
{
    return ensureTemplateObjectMapImpl(m_templateObjectMap);
}

void ModuleProgramExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    ModuleProgramExecutable* thisObject = jsCast<ModuleProgramExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    visitor.append(&thisObject->m_moduleEnvironmentSymbolTable);
    if (thisObject->m_moduleProgramCodeBlock)
        thisObject->m_moduleProgramCodeBlock->visitAggregate(visitor);
    if (TemplateObjectMap* map = thisObject->m_templateObjectMap.get()) {
        //auto locker = holdLock(thisObject->cellLock());
    	for(auto it = map->begin(); it != map->end(); ++it)
            visitor.append(it->second);
    }
}

void ModuleProgramExecutable::clearCode()
{
    m_moduleProgramCodeBlock.clear();
    m_moduleEnvironmentSymbolTable.clear();
    Base::clearCode();
}

FunctionCodeBlock* FunctionExecutable::baselineCodeBlockFor(CodeSpecializationKind kind)
{
    FunctionCodeBlock* result;
    if (kind == CodeForCall)
        result = m_codeBlockForCall.get();
    else {
        ASSERT(kind == CodeForConstruct);
        result = m_codeBlockForConstruct.get();
    }
    if (!result)
        return nullptr;
    return static_cast<FunctionCodeBlock*>(result->baselineAlternative());
}

void FunctionExecutable::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    FunctionExecutable* thisObject = jsCast<FunctionExecutable*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    visitor.append(thisObject->m_topLevelExecutable);
    if (thisObject->m_codeBlockForCall)
        thisObject->m_codeBlockForCall->visitAggregate(visitor);
    if (thisObject->m_codeBlockForConstruct)
        thisObject->m_codeBlockForConstruct->visitAggregate(visitor);
    visitor.append(&thisObject->m_singletonFunction);
    if (TemplateObjectMap* map = thisObject->m_templateObjectMap.get()) {
        //auto locker = holdLock(thisObject->cellLock());
        for (auto it = map->begin(); it != map->end(); ++it)
            visitor.append(it->second);
    }
}

void FunctionExecutable::clearCode()
{
    m_codeBlockForCall.clear();
    m_codeBlockForConstruct.clear();
    Base::clearCode();
}

// FIXME: There's no need to add the function's name to the key here. It's already in the source code.
FunctionExecutable* FunctionExecutable::fromGlobalCode(const Identifier& name, ExecState* exec, Debugger* debugger, const SourceCode& source, JSObject** exception)
{
	auto& vm = exec->vm();
    auto* lexicalGlobalObject = exec->lexicalGlobalObject();
    ParserError error;
    OwnPtr<ProgramNode> program = parse<ProgramNode>(
    	&vm, source, Identifier(), JSParserBuiltinMode::NotBuiltin,
		JSParserStrictMode::NotStrict, JSParserScriptMode::Classic, SourceParseMode::ProgramMode, FunctionMode::None, SuperBinding::NotNeeded, error);

    if (lexicalGlobalObject->hasDebugger() && !ProgramNode::scopeIsFunction)
    	lexicalGlobalObject->debugger()->sourceParsed(exec, source.provider(), error.line(), error.message());

    if (!program) {
        *exception = error.toErrorObject(lexicalGlobalObject, source);
        return nullptr;
    }

    // This function assumes an input string that would result in a single function declaration.
    StatementNode* statement = program->singleStatement();
    if (UNLIKELY(!statement)) {
		JSToken token;
		error = ParserError(ParserError::SyntaxError, ParserError::SyntaxErrorIrrecoverable, token, ASCIILiteral("Parser error"), -1);
		return nullptr;
	}
    ASSERT(statement->isBlock());

    StatementNode* funcDecl = static_cast<BlockNode*>(statement)->singleStatement();
    if (UNLIKELY(!funcDecl)) {
		JSToken token;
		error = ParserError(ParserError::SyntaxError, ParserError::SyntaxErrorIrrecoverable, token, ASCIILiteral("Parser error"), -1);
		return nullptr;
	}
    ASSERT(funcDecl->isFuncDeclNode());

    FunctionMetadataNode* metadata = static_cast<FuncDeclNode*>(funcDecl)->metadata();
    ASSERT(metadata);
    if (!metadata)
        return nullptr;

    metadata->overrideName(name);
    unsigned startColumn = source.startColumn();
    // The Function constructor only has access to global variables, so no variables will be under TDZ.
	ConstructAbility constructAbility = constructAbilityForParseMode(metadata->parseMode());
	FunctionExecutable* functionExecutable = FunctionExecutable::create(vm, nullptr, metadata, startColumn, false, constructAbility, JSParserScriptMode::Classic, nullptr, Nullopt, nullptr, DerivedContextType::None, NeedsClassFieldInitializer::No, PrivateBrandRequirement::None, NoIntrinsic);

    return functionExecutable;
}

auto FunctionExecutable::ensureTemplateObjectMap(VM&) -> TemplateObjectMap&
{
    return ensureTemplateObjectMapImpl(m_templateObjectMap);
}

}
