/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003-2019 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "FunctionConstructor.h"

#include "Completion.h"
#include "Debugger.h"
#include "ExceptionHelpers.h"
#include "FunctionPrototype.h"
#include "JSAsyncFunction.h"
#include "JSAsyncGeneratorFunction.h"
#include "JSFunction.h"
#include "JSGeneratorFunction.h"
#include "JSGlobalObject.h"
#include "JSString.h"
#include "Lexer.h"
#include "Nodes.h"
#include "Operations.h"
#include "Parser.h"
#include <wtf/text/StringBuilder.h>

namespace JSC {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(FunctionConstructor);

const ClassInfo FunctionConstructor::s_info = { "Function", &Base::s_info, 0, CREATE_METHOD_TABLE(FunctionConstructor) };

static EncodedJSValue JSC_HOST_CALL constructWithFunctionConstructor(ExecState* exec)
{
    ArgList args(exec);
    if (auto* result = constructFunction(exec, asInternalFunction(exec->callee())->globalObject(), args, FunctionConstructionMode::Function, exec->newTarget()))
    	return JSValue::encode(result);
    return encodedJSUndefined();
}

// ECMA 15.3.1 The Function Constructor Called as a Function
static EncodedJSValue JSC_HOST_CALL callFunctionConstructor(ExecState* exec)
{
    ArgList args(exec);
    if (auto* result = constructFunction(exec, asInternalFunction(exec->callee())->globalObject(), args))
    	return JSValue::encode(result);
    return encodedJSUndefined();
}

FunctionConstructor::FunctionConstructor(VM& vm, Structure* structure)
    : InternalFunction(vm, structure, callFunctionConstructor, constructWithFunctionConstructor)
{
}

void FunctionConstructor::finishCreation(VM& vm, FunctionPrototype* functionPrototype)
{
    Base::finishCreation(vm, String::fromLiteral(functionPrototype->classInfo()->className));
    putDirectWithoutTransition(vm, vm.propertyNames->prototype, functionPrototype, DontEnum | DontDelete | ReadOnly);
    putDirectWithoutTransition(vm, vm.propertyNames->length, jsNumber(1), ReadOnly | DontEnum);
}

// ECMA 15.3.2 The Function Constructor
JSObject* constructFunction(ExecState* exec, JSGlobalObject* globalObject, const ArgList& args, const Identifier& functionName, const SourceOrigin& sourceOrigin, const String& sourceURL, const TextPosition& position, FunctionConstructionMode functionConstructionMode, JSValue newTarget)
{
    VM& vm = globalObject->vm();
    if (UNLIKELY(!globalObject->evalEnabled())) {
        vm.throwException(exec, createEvalError(exec, globalObject->evalDisabledErrorMessage()));
        return nullptr;
    }
    return constructFunctionSkippingEvalEnabledCheck(exec, globalObject, args, functionName, sourceOrigin, sourceURL, position, functionConstructionMode, newTarget);
}

JSObject* constructFunctionSkippingEvalEnabledCheck(
	ExecState* exec, JSGlobalObject* globalObject, const ArgList& args,
	const Identifier& functionName, const SourceOrigin& sourceOrigin, const String& sourceURL,
	const TextPosition& position, FunctionConstructionMode functionConstructionMode, JSValue newTarget)
{
    VM& vm = exec->vm();

    const char* prefix = 0;
    switch (functionConstructionMode) {
    case FunctionConstructionMode::Function:
        prefix = "function ";
        break;
    case FunctionConstructionMode::Generator:
        prefix = "function *";
        break;
    case FunctionConstructionMode::Async:
        prefix = "async function ";
        break;
    case FunctionConstructionMode::AsyncGenerator:
        prefix = "async function*";
        break;
    }

    auto checkBody = [&] (const String& body) {
        // The spec mandates that the body parses a valid function body independent
        // of the parameters.
        String program = tryMakeString("(", prefix, "(){\n", body, "\n})");
        if (UNLIKELY(!program)) {
            throwOutOfMemoryError(vm, exec);
            return;
        }
        SourceCode source = makeSource(program, sourceOrigin, sourceURL, position);
        JSValue exception;
        checkSyntax(exec, source, &exception);
        if (exception) {
            vm.throwException(exec, exception);
            return;
        }
    };

    // How we stringify functions is sometimes important for web compatibility.
    // See https://bugs.webkit.org/show_bug.cgi?id=24350.
    String program;
    if (args.isEmpty()) {
        program = tryMakeString("{", prefix, functionName.string(), "() {\n\n}}");
        if (UNLIKELY(!program)) {
            throwOutOfMemoryError(vm, exec);
            return nullptr;
        }
    } else if (args.size() == 1) {
		auto body = args.at(0).toWTFString(exec);
        checkBody(body);
        RETURN_IF_EXCEPTION(vm, 0);
        program = tryMakeString("{", prefix, functionName.string(), "() {\n", body, "\n}}");
        if (UNLIKELY(!program)) {
            throwOutOfMemoryError(vm, exec);
            return nullptr;
        }
	} else {
        StringBuilder builder(StringBuilder::OverflowHandler::RecordOverflow);
        builder.append('{');
        builder.append(prefix);
        builder.append(functionName.string());
        builder.append('(');
        StringBuilder parameterBuilder(StringBuilder::OverflowHandler::RecordOverflow);
        JSString* jsString = args.at(0).toString(exec);
        RETURN_IF_EXCEPTION(vm, 0);
        auto viewWithString = jsString->viewWithUnderlyingString(exec);
        RETURN_IF_EXCEPTION(vm, 0);
        parameterBuilder.append(viewWithString.view);
        for (size_t i = 1; !parameterBuilder.hasOverflowed() && i < args.size() - 1; i++) {
        	parameterBuilder.appendLiteral(", ");
        	JSString* jsString = args.at(i).toString(exec);
            RETURN_IF_EXCEPTION(vm, 0);
            auto viewWithString = jsString->viewWithUnderlyingString(exec);
            RETURN_IF_EXCEPTION(vm, 0);
            parameterBuilder.append(viewWithString.view);
        }
        if (parameterBuilder.hasOverflowed()) {
            throwOutOfMemoryError(vm, exec);
            return 0;
        }
        String body = args.at(args.size() - 1).toWTFString(exec);
        RETURN_IF_EXCEPTION(vm, 0);

        {
            // The spec mandates that the parameters parse as a valid parameter list
            // independent of the function body.
            String program = tryMakeString("(", prefix, "(", parameterBuilder.toString(), "){\n\n})");
            if (UNLIKELY(!program)) {
                throwOutOfMemoryError(vm, exec);
                return 0;
            }
            SourceCode source = makeSource(program, sourceOrigin, sourceURL, position);
            JSValue exception;
            checkSyntax(exec, source, &exception);
            if (exception) {
                vm.throwException(exec, exception);
                return 0;
            }
        }
        RETURN_IF_EXCEPTION(vm, 0);

        builder.append(parameterBuilder);
        builder.appendLiteral(") {\n");
        checkBody(body);
        RETURN_IF_EXCEPTION(vm, 0);
        builder.append(body);
        builder.appendLiteral("\n}}");
        if (builder.hasOverflowed()) {
            throwOutOfMemoryError(vm, exec);
            return 0;
        }
        program = builder.toString();
    }

    SourceCode source = makeSource(program, sourceOrigin, sourceURL, position);
    JSObject* exception = 0;
    FunctionExecutable* function = FunctionExecutable::fromGlobalCode(functionName, exec, exec->vmEntryGlobalObject()->debugger(), source, &exception);
    if (!function) {
        ASSERT(exception);
        if(!exception)
        	exception = createError(exec, ASCIILiteral("Function constructor parsing error"));
        vm.throwException(exec, exception);
        return nullptr;
    }

    JSGlobalObject* structureGlobalObject = globalObject;
    bool needsSubclassStructure = newTarget && newTarget != globalObject->functionConstructor();
    if (needsSubclassStructure) {
		structureGlobalObject = getFunctionRealm(exec, globalObject, asObject(newTarget));
		RETURN_IF_EXCEPTION(vm, nullptr);
	}
    Structure* structure = nullptr;
    switch (functionConstructionMode) {
    case FunctionConstructionMode::Function:
        structure = JSFunction::selectStructureForNewFuncExp(structureGlobalObject, function);
        break;
    case FunctionConstructionMode::Generator:
        structure = structureGlobalObject->generatorFunctionStructure();
        break;
    case FunctionConstructionMode::Async:
        structure = structureGlobalObject->asyncFunctionStructure();
        break;
    case FunctionConstructionMode::AsyncGenerator:
        structure = structureGlobalObject->asyncGeneratorFunctionStructure();
        break;
    }

    if (needsSubclassStructure) {
        structure = InternalFunction::createSubclassStructure(exec, globalObject, asObject(newTarget), structure);
        RETURN_IF_EXCEPTION(vm, nullptr);
    }

    switch (functionConstructionMode) {
    case FunctionConstructionMode::Function:
        return JSFunction::create(vm, function, globalObject->globalScope(), structure);
    case FunctionConstructionMode::Generator:
        return JSGeneratorFunction::create(vm, function, globalObject->globalScope(), structure);
    case FunctionConstructionMode::Async:
        return JSAsyncFunction::create(vm, function, globalObject->globalScope(), structure);
    case FunctionConstructionMode::AsyncGenerator:
        return JSAsyncGeneratorFunction::create(vm, function, globalObject->globalScope(), structure);
    }

    ASSERT_NOT_REACHED();
    return 0;
}

// ECMA 15.3.2 The Function Constructor
JSObject* constructFunction(ExecState* exec, JSGlobalObject* globalObject, const ArgList& args, FunctionConstructionMode functionConstructionMode, JSValue newTarget)
{
    return constructFunction(exec, globalObject, args, globalObject->vm().propertyNames->anonymous, exec->callerSourceOrigin(), String(), TextPosition::minimumPosition(), functionConstructionMode, newTarget);
}

} // namespace JSC
