/*
 * Copyright (C) 2014 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "CallFrame.h"
#include "Identifier.h"
#include "Symbol.h"

namespace JSC  {

inline Identifier::Identifier(JSGlobalData& vm, AtomicStringImpl& string)
    : m_string(string)
{
#ifndef NDEBUG
    checkCurrentAtomicStringTable(exec);
    if (string)
        ASSERT_WITH_MESSAGE(!string->length() || string->isSymbol() || AtomicStringImpl::isInAtomicStringTable(string), "The atomic string comes from an other thread!");
#else
    UNUSED_PARAM(vm);
#endif
}

inline Identifier::Identifier(JSGlobalData& vm, AtomicStringImpl* string)
    : m_string(string)
{
#ifndef NDEBUG
    checkCurrentAtomicStringTable(exec);
    if (string)
        ASSERT_WITH_MESSAGE(!string->length() || string->isSymbol() || AtomicStringImpl::isInAtomicStringTable(string), "The atomic string comes from an other thread!");
#else
    UNUSED_PARAM(vm);
#endif
}

inline Identifier::Identifier(JSGlobalData& vm, const AtomicString& string)
    : m_string(string.string())
{
#ifndef NDEBUG
    checkCurrentAtomicStringTable(exec);
    if (!string.isNull())
        ASSERT_WITH_MESSAGE(!string.length() || string.impl()->isSymbol() || AtomicStringImpl::isInAtomicStringTable(string.impl()), "The atomic string comes from an other thread!");
#else
    UNUSED_PARAM(vm);
#endif
}

inline PassRefPtr<StringImpl> Identifier::add(JSGlobalData& vm, StringImpl* r)
{
#ifndef NDEBUG
    checkCurrentIdentifierTable(vm);
#endif
    return AtomicString::addWithStringTableProvider(vm, r);
}

inline PassRefPtr<StringImpl> Identifier::add(JSGlobalData& vm, StringImpl& r)
{
#ifndef NDEBUG
    checkCurrentIdentifierTable(vm);
#endif
    return AtomicString::addWithStringTableProvider(vm, r);
}

// Only to be used with string literals.
inline PassRefPtr<StringImpl> Identifier::add(JSGlobalData& vm, const ASCIILiteral& characters)
{
    ASSERT(characters);
    ASSERT(characters[0]);
    return AtomicString::addFromLiteralDataWithStringTableProvider(vm, characters.characters(), characters.length());
}

ALWAYS_INLINE Identifier Identifier::fromUid(VM& vm, UniquedStringImpl* uid)
{
    if (UNLIKELY(!uid))
        return Identifier();
    if (!uid->isSymbol()) {
    	ASSERT(uid->isAtomic());
        return Identifier(vm, static_cast<AtomicStringImpl&>(*uid));
    }
    return static_cast<SymbolImpl&>(*uid);
}

inline Identifier Identifier::fromUid(VM& vm, UniquedStringImpl& uid)
{
    if (!uid.isSymbol()) {
    	ASSERT(uid.isAtomic());
        return Identifier(vm, static_cast<AtomicStringImpl&>(uid));
    }
    return static_cast<SymbolImpl&>(uid);
}

inline Identifier Identifier::fromUid(SymbolImpl& symbol)
{
    return symbol;
}

inline Identifier Identifier::fromUid(const PrivateName& name)
{
    return *name.uid();
}

inline Identifier Identifier::fromString(VM& vm, const ASCIILiteral& characters)
{
    return Identifier(vm, characters);
}

inline Identifier Identifier::fromString(VM& vm, const LChar* s, int length)
{
    return Identifier(vm, s, length);
}

inline Identifier Identifier::fromString(VM& vm, const UChar* s, int length)
{
    return Identifier(vm, s, length);
}

inline Identifier Identifier::fromString(VM& vm, const String& string)
{
    return Identifier(vm, string.impl());
}

inline Identifier Identifier::fromString(VM& vm, AtomicStringImpl* atomicString)
{
    return Identifier(vm, atomicString);
}

inline Identifier Identifier::fromString(VM& vm, const AtomicString& atomicString)
{
    return Identifier(vm, atomicString);
}

#if 0
inline Identifier Identifier::fromString(VM& vm, SymbolImpl* symbolImpl)
{
    return Identifier(vm, symbolImpl);
}
#endif

inline Identifier Identifier::from(JSGlobalData& globalData, unsigned value)
{
	return Identifier(globalData, globalData.numericStrings.add(value));
}

inline Identifier Identifier::from(JSGlobalData& globalData, int value)
{
	return Identifier(globalData, globalData.numericStrings.add(value));
}

inline Identifier Identifier::from(JSGlobalData& globalData, double value)
{
	return Identifier(globalData, globalData.numericStrings.add(value));
}

inline JSValue identifierToJSValue(VM& vm, const Identifier& identifier)
{
    if (identifier.isSymbol())
        return Symbol::create(vm, static_cast<SymbolImpl&>(*identifier.impl()));
    return jsString(vm, identifier.impl());
}

inline JSValue identifierToSafePublicJSValue(VM& vm, const Identifier& identifier)
{
    if (identifier.isSymbol() && !identifier.isPrivateName())
        return Symbol::create(vm, static_cast<SymbolImpl&>(*identifier.impl()));
    return jsString(vm, identifier.impl());
}

} // namespace JSC
