/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *  Copyright (C) 2007 Cameron Zwarich (cwzwarich@uwaterloo.ca)
 *  Copyright (C) 2007 Maks Orlovich
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "CodeSpecializationKind.h"
#include "Identifier.h"
#include "JSDestructibleObject.h"

namespace JSC {

class FunctionPrototype;

class InternalFunction : public JSDestructibleObject {
public:
	typedef JSDestructibleObject Base;
    static const unsigned StructureFlags = Base::StructureFlags | ImplementsHasInstance | ImplementsDefaultHasInstance | OverridesGetCallData;

	DECLARE_EXPORT_INFO;

	JS_EXPORT_PRIVATE static void visitChildren(JSCell*, SlotVisitor&);

	JS_EXPORT_PRIVATE const String& name();
	const String displayName(JSGlobalData&);
	const String calculatedDisplayName(JSGlobalData&);

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(globalData, globalObject, proto, TypeInfo(InternalFunctionType, StructureFlags), info());
	}

    JS_EXPORT_PRIVATE static Structure* createSubclassStructure(ExecState* exec, JSGlobalObject*, JSObject* newTarget, Structure*);
    JS_EXPORT_PRIVATE static InternalFunction* createFunctionThatMasqueradesAsUndefined(JSGlobalData&, JSGlobalObject*, int length, const String& name, NativeFunction);

    NativeFunction nativeFunctionFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return m_functionForCall;
        ASSERT(kind == CodeForConstruct);
        return m_functionForConstruct;
    }

    static ptrdiff_t offsetOfNativeFunctionFor(CodeSpecializationKind kind)
    {
        if (kind == CodeForCall)
            return OBJECT_OFFSETOF(InternalFunction, m_functionForCall);
        ASSERT(kind == CodeForConstruct);
        return OBJECT_OFFSETOF(InternalFunction, m_functionForConstruct);
    }

    static ptrdiff_t offsetOfGlobalObject()
    {
        return OBJECT_OFFSETOF(InternalFunction, m_globalObject);
    }

    JSGlobalObject* globalObject() const { return m_globalObject.get(); }

protected:
	InternalFunction(JSGlobalData&, Structure*, NativeFunction functionForCall, NativeFunction functionForConstruct = nullptr);

	JS_EXPORT_PRIVATE void finishCreation(JSGlobalData&, const String& name);
	JS_EXPORT_PRIVATE void finishCreation(VM&);

    JS_EXPORT_PRIVATE static CallData getConstructData(JSCell*);
    JS_EXPORT_PRIVATE static CallData getCallData(JSCell*);

    NativeFunction m_functionForCall;
    NativeFunction m_functionForConstruct;
	WriteBarrier<JSString> m_originalName;
    WriteBarrier<JSGlobalObject> m_globalObject;
};

InternalFunction* asInternalFunction(JSValue);

inline InternalFunction* asInternalFunction(JSValue value)
{
	ASSERT(asObject(value)->inherits(InternalFunction::info()));
	return static_cast<InternalFunction*>(asObject(value));
}

template<>
ALWAYS_INLINE InternalFunction* jsDynamicCast(JSCell* from)
{
	auto type = from->structure()->typeInfo().type();
	if(type >= JSType::InternalFunctionType && type <= JSType::NullSetterFunctionType)
		return static_cast<InternalFunction*>(from);
	return nullptr;
}

JS_EXPORT_PRIVATE JSGlobalObject* getFunctionRealm(ExecState* exec, JSGlobalObject*, JSObject*);

#define JSC_GET_DERIVED_STRUCTURE(vm, exec, structureMemberFunctionName, newTarget, constructor) \
    ((newTarget) == (constructor) \
        ? globalObject->structureMemberFunctionName() \
        : ([&]() -> Structure* { \
            auto scope = DECLARE_THROW_SCOPE((vm)); \
            auto* functionGlobalObject = getFunctionRealm(exec, globalObject, (newTarget)); \
            RETURN_IF_EXCEPTION(scope, nullptr); \
            RELEASE_AND_RETURN(scope, InternalFunction::createSubclassStructure(exec, globalObject, (newTarget), functionGlobalObject->structureMemberFunctionName())); \
        }()))

} // namespace JSC
