/*
 * Copyright (C) 2015 Andy VanWagoner (thetalecrafter@gmail.com)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "IntlDateTimeFormat.h"

#if ENABLE(INTL)

#include "DateInstance.h"
#include "Error.h"
#include "IntlDateTimeFormatConstructor.h"
#include "IntlObject.h"
#include "JSBoundFunction.h"
#include "JSCJSValueInlines.h"
#include "JSCellInlines.h"
#include "JSObjectInlines.h"
#include "ObjectConstructor.h"
#include "SlotVisitorInlines.h"
#include "StructureInlines.h"
#include <unicode/ucal.h>
#include <unicode/udat.h>
#include <unicode/udatpg.h>
#include <unicode/uenum.h>
#include <wtf/unicode/icu/ICUHelpers.h>

namespace WTF {

template <>
ALWAYS_INLINE void deleteOwnedPtr<UDateFormat>(UDateFormat* dateFormat)
{
    if (dateFormat)
        udat_close(dateFormat);
}

template <>
ALWAYS_INLINE void deleteOwnedPtr<UFieldPositionIterator>(UFieldPositionIterator* iterator)
{
    if (iterator)
        ufieldpositer_close(iterator);
}

}

namespace JSC {

static const double minECMAScriptTime = -8.64E15;

const ClassInfo IntlDateTimeFormat::s_info = { "Object", &Base::s_info, 0, CREATE_METHOD_TABLE(IntlDateTimeFormat) };

static const char* const relevantExtensionKeys[3] = { "ca", "nu", "hc" };
static const size_t indexOfExtensionKeyCa = 0;
static const size_t indexOfExtensionKeyNu = 1;
static const size_t indexOfExtensionKeyHc = 2;

IntlDateTimeFormat* IntlDateTimeFormat::create(VM& vm, Structure* structure)
{
    IntlDateTimeFormat* format = new (NotNull, allocateCell<IntlDateTimeFormat>(vm.heap)) IntlDateTimeFormat(vm, structure);
    format->finishCreation(vm);
    return format;
}

Structure* IntlDateTimeFormat::createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
{
    return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
}

IntlDateTimeFormat::IntlDateTimeFormat(VM& vm, Structure* structure)
    : Base(vm, structure)
	, m_initializedDateTimeFormat(false)
	, m_locale(ASCIILiteral("en"))
	, m_calendar(ASCIILiteral("gregorian"))
	, m_numberingSystem(ASCIILiteral("latn"))
	, m_timeZone(ASCIILiteral("UTC"))
	, m_weekday(Weekday::None)
	, m_era(Era::None)
	, m_year(Year::None)
	, m_month(Month::None)
	, m_day(Day::None)
	, m_hour(Hour::None)
	, m_minute(Minute::None)
	, m_second(Second::None)
	, m_timeZoneName(TimeZoneName::None)
{
}

IntlDateTimeFormat::~IntlDateTimeFormat() = default;

void IntlDateTimeFormat::finishCreation(VM& vm)
{
    Base::finishCreation(vm);
    ASSERT(inherits(info()));
}

void IntlDateTimeFormat::destroy(JSCell* cell)
{
    static_cast<IntlDateTimeFormat*>(cell)->IntlDateTimeFormat::~IntlDateTimeFormat();
}

void IntlDateTimeFormat::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    IntlDateTimeFormat* thisObject = jsCast<IntlDateTimeFormat*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());

    Base::visitChildren(thisObject, visitor);

    visitor.append(&thisObject->m_boundFormat);
}

void IntlDateTimeFormat::setBoundFormat(VM& vm, JSBoundFunction* format)
{
    m_boundFormat.set(vm, this, format);
}

static ALWAYS_INLINE bool isUTCEquivalent(StringView timeZone)
{
    return timeZone == "Etc/UTC" || timeZone == "Etc/GMT";
}

// https://tc39.es/ecma402/#sec-defaulttimezone
static String defaultTimeZone()
{
    UErrorCode status = U_ZERO_ERROR;
    String canonical;

    Vector<UChar, 32> buffer(32);
    auto bufferLength = ucal_getDefaultTimeZone(buffer.data(), buffer.size(), &status);
    if (needsToGrowToProduceBuffer(status)) {
        status = U_ZERO_ERROR;
        buffer.grow(bufferLength);
        ucal_getDefaultTimeZone(buffer.data(), bufferLength, &status);
    }
    if (U_SUCCESS(status)) {
        status = U_ZERO_ERROR;
        Vector<UChar, 32> canonicalBuffer(32);
        auto canonicalLength = ucal_getCanonicalTimeZoneID(buffer.data(), bufferLength, canonicalBuffer.data(), canonicalBuffer.size(), 0, &status);
        if (needsToGrowToProduceBuffer(status)) {
            status = U_ZERO_ERROR;
            canonicalBuffer.grow(canonicalLength);
            ucal_getCanonicalTimeZoneID(buffer.data(), bufferLength, canonicalBuffer.data(), canonicalLength, 0, &status);
        }
        if (U_SUCCESS(status))
            canonical = String(canonicalBuffer.data(), canonicalLength);
    }

    if (canonical.isNull() || isUTCEquivalent(canonical))
    	return ASCIILiteral("UTC");

    return canonical;
}

static String canonicalizeTimeZoneName(const String& timeZoneName)
{
    // 6.4.1 IsValidTimeZoneName (timeZone)
    // The abstract operation returns true if timeZone, converted to upper case as described in 6.1, is equal to one of the Zone or Link names of the IANA Time Zone Database, converted to upper case as described in 6.1. It returns false otherwise.
    UErrorCode status = U_ZERO_ERROR;
    UEnumeration* timeZones = ucal_openTimeZones(&status);
    if (!U_SUCCESS(status))
		return ASCIILiteral("UTC");

    String canonical;
    do {
        status = U_ZERO_ERROR;
        int32_t ianaTimeZoneLength;
        // Time zone names are respresented as UChar[] in all related ICU apis.
        const UChar* ianaTimeZone = uenum_unext(timeZones, &ianaTimeZoneLength, &status);
        if (!U_SUCCESS(status))
        	break;

        // End of enumeration.
        if (!ianaTimeZone)
            break;

        StringView ianaTimeZoneView(ianaTimeZone, ianaTimeZoneLength);
        if (!equalIgnoringASCIICase(timeZoneName, ianaTimeZoneView))
            continue;

        // Found a match, now canonicalize.
        // 6.4.2 CanonicalizeTimeZoneName (timeZone) (ECMA-402 2.0)
        // 1. Let ianaTimeZone be the Zone or Link name of the IANA Time Zone Database such that timeZone, converted to upper case as described in 6.1, is equal to ianaTimeZone, converted to upper case as described in 6.1.
        // 2. If ianaTimeZone is a Link name, then let ianaTimeZone be the corresponding Zone name as specified in the backward file of the IANA Time Zone Database.

        Vector<UChar, 32> buffer(ianaTimeZoneLength);
        status = U_ZERO_ERROR;
        auto canonicalLength = ucal_getCanonicalTimeZoneID(ianaTimeZone, ianaTimeZoneLength, buffer.data(), ianaTimeZoneLength, 0, &status);
        if (needsToGrowToProduceBuffer(status)) {
            buffer.grow(canonicalLength);
            status = U_ZERO_ERROR;
            ucal_getCanonicalTimeZoneID(ianaTimeZone, ianaTimeZoneLength, buffer.data(), canonicalLength, 0, &status);
        }
        if (!U_SUCCESS(status))
        	break;
        canonical = String(buffer.data(), canonicalLength);
    } while (canonical.isNull());
    uenum_close(timeZones);

    // 3. If ianaTimeZone is "Etc/UTC" or "Etc/GMT", then return "UTC".
    if (isUTCEquivalent(canonical))
		return ASCIILiteral("UTC");

    // 4. Return ianaTimeZone.
    return canonical;
}

static Vector<String> localeData(const String& locale, size_t keyIndex)
{
    Vector<String> keyLocaleData;
    switch (keyIndex) {
    case indexOfExtensionKeyCa: {
        UErrorCode status = U_ZERO_ERROR;
        UEnumeration* calendars = ucal_getKeywordValuesForLocale("calendar", locale.utf8().data(), false, &status);
        if (!U_SUCCESS(status))
        	break;

        int32_t nameLength;
        while (const char* availableName = uenum_next(calendars, &nameLength, &status)) {
        	if (!U_SUCCESS(status))
				break;

            String calendar = String(availableName, nameLength);
            keyLocaleData.append(calendar);
            // Ensure aliases used in language tag are allowed.
            if (calendar == "gregorian")
                keyLocaleData.append(ASCIILiteral("gregory"));
            else if (calendar == "islamic-civil")
                keyLocaleData.append(ASCIILiteral("islamicc"));
            else if (calendar == "ethiopic-amete-alem")
                keyLocaleData.append(ASCIILiteral("ethioaa"));
        }
        uenum_close(calendars);
        break;
    }
    case indexOfExtensionKeyNu:
        keyLocaleData = numberingSystemsForLocale(locale);
        break;
    case indexOfExtensionKeyHc:
        // Null default so we know to use 'j' in pattern.
        keyLocaleData.append(String());
        keyLocaleData.append(ASCIILiteral("h11"));
        keyLocaleData.append(ASCIILiteral("h12"));
        keyLocaleData.append(ASCIILiteral("h23"));
        keyLocaleData.append(ASCIILiteral("h24"));
        break;
    default:
        ASSERT_NOT_REACHED();
    }
    return keyLocaleData;
}

static JSObject* toDateTimeOptionsAnyDate(ExecState& exec, JSValue originalOptions)
{
    // 12.1.1 ToDateTimeOptions abstract operation (ECMA-402 2.0)
    JSGlobalObject* globalObject = exec.lexicalGlobalObject();
    VM& vm = globalObject->vm();

    // 1. If options is undefined, then let options be null, else let options be ToObject(options).
    // 2. ReturnIfAbrupt(options).
    // 3. Let options be ObjectCreate(options).
    JSObject* options;
    if (originalOptions.isUndefined())
        options = constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure());
    else {
        JSObject* originalToObject = originalOptions.toObject(&exec);
        RETURN_IF_EXCEPTION(vm, 0);
        options = constructEmptyObject(globalObject, originalToObject);
    }

    // 4. Let needDefaults be true.
    bool needDefaults = true;

    // 5. If required is "date" or "any",
    // Always "any".

	// a. For each of the property names "weekday", "year", "month", "day":
	// i. Let prop be the property name.
	// ii. Let value be Get(options, prop).
	// iii. ReturnIfAbrupt(value).
	// iv. If value is not undefined, then let needDefaults be false.
	JSValue weekday = options->get(vm, &exec, vm.propertyNames->weekday);
    RETURN_IF_EXCEPTION(vm, 0);
	if (!weekday.isUndefined())
		needDefaults = false;

	JSValue year = options->get(vm, &exec, vm.propertyNames->year);
    RETURN_IF_EXCEPTION(vm, 0);
	if (!year.isUndefined())
		needDefaults = false;

	JSValue month = options->get(vm, &exec, vm.propertyNames->month);
    RETURN_IF_EXCEPTION(vm, 0);
	if (!month.isUndefined())
		needDefaults = false;

	JSValue day = options->get(vm, &exec, vm.propertyNames->day);
    RETURN_IF_EXCEPTION(vm, 0);
	if (!day.isUndefined())
		needDefaults = false;

    // 6. If required is "time" or "any",
    // Always "any".

	// a. For each of the property names "hour", "minute", "second":
	// i. Let prop be the property name.
	// ii. Let value be Get(options, prop).
	// iii. ReturnIfAbrupt(value).
	// iv. If value is not undefined, then let needDefaults be false.
	JSValue hour = options->get(vm, &exec, vm.propertyNames->hour);
    RETURN_IF_EXCEPTION(vm, 0);
	if (!hour.isUndefined())
		needDefaults = false;

	JSValue minute = options->get(vm, &exec, vm.propertyNames->minute);
    RETURN_IF_EXCEPTION(vm, 0);
	if (!minute.isUndefined())
		needDefaults = false;

	JSValue second = options->get(vm, &exec, vm.propertyNames->second);
    RETURN_IF_EXCEPTION(vm, 0);
	if (!second.isUndefined())
		needDefaults = false;

    // 7. If needDefaults is true and defaults is either "date" or "all", then
    // Defaults is always "date".
    if (needDefaults) {
        // a. For each of the property names "year", "month", "day":
        // i. Let status be CreateDatePropertyOrThrow(options, prop, "numeric").
        // ii. ReturnIfAbrupt(status).
        JSString* numeric = jsNontrivialString(vm, ASCIILiteral("numeric"));

        options->putDirect(vm, vm.propertyNames->year, numeric);
        RETURN_IF_EXCEPTION(vm, 0);

        options->putDirect(vm, vm.propertyNames->month, numeric);
        RETURN_IF_EXCEPTION(vm, 0);

        options->putDirect(vm, vm.propertyNames->day, numeric);
        RETURN_IF_EXCEPTION(vm, 0);
    }

    // 8. If needDefaults is true and defaults is either "time" or "all", then
    // Defaults is always "date". Ignore this branch.

    // 9. Return options.
    return options;
}

void IntlDateTimeFormat::setFormatsFromPattern(const StringView& pattern)
{
    // Get all symbols from the pattern, and set format fields accordingly.
    // http://unicode.org/reports/tr35/tr35-dates.html#Date_Field_Symbol_Table
    unsigned length = pattern.length();
    for (unsigned i = 0; i < length; ++i) {
        UChar currentCharacter = pattern[i];
        if (!isASCIIAlpha(currentCharacter))
            continue;

        unsigned count = 1;
        while (i + 1 < length && pattern[i + 1] == currentCharacter) {
            ++count;
            ++i;
        }

        // If hourCycle was null, this sets it to the locale default.
        if (m_hourCycle.isNull()) {
            if (currentCharacter == 'h')
                m_hourCycle = ASCIILiteral("h12");
            else if (currentCharacter == 'H')
                m_hourCycle = ASCIILiteral("h23");
            else if (currentCharacter == 'k')
                m_hourCycle = ASCIILiteral("h24");
            else if (currentCharacter == 'K')
                m_hourCycle = ASCIILiteral("h11");
        }

        switch (currentCharacter) {
        case 'G':
            if (count <= 3)
                m_era = Era::Short;
            else if (count == 4)
                m_era = Era::Long;
            else if (count == 5)
                m_era = Era::Narrow;
            break;
        case 'y':
            if (count == 1)
                m_year = Year::Numeric;
            else if (count == 2)
                m_year = Year::TwoDigit;
            break;
        case 'M':
        case 'L':
            if (count == 1)
                m_month = Month::Numeric;
            else if (count == 2)
                m_month = Month::TwoDigit;
            else if (count == 3)
                m_month = Month::Short;
            else if (count == 4)
                m_month = Month::Long;
            else if (count == 5)
                m_month = Month::Narrow;
            break;
        case 'E':
        case 'e':
        case 'c':
            if (count <= 3)
                m_weekday = Weekday::Short;
            else if (count == 4)
                m_weekday = Weekday::Long;
            else if (count == 5)
                m_weekday = Weekday::Narrow;
            break;
        case 'd':
            if (count == 1)
                m_day = Day::Numeric;
            else if (count == 2)
                m_day = Day::TwoDigit;
            break;
        case 'h':
        case 'H':
        case 'k':
        case 'K':
            if (count == 1)
                m_hour = Hour::Numeric;
            else if (count == 2)
                m_hour = Hour::TwoDigit;
            break;
        case 'm':
            if (count == 1)
                m_minute = Minute::Numeric;
            else if (count == 2)
                m_minute = Minute::TwoDigit;
            break;
        case 's':
            if (count == 1)
                m_second = Second::Numeric;
            else if (count == 2)
                m_second = Second::TwoDigit;
            break;
        case 'z':
        case 'v':
        case 'V':
            if (count == 1)
                m_timeZoneName = TimeZoneName::Short;
            else if (count == 4)
                m_timeZoneName = TimeZoneName::Long;
            break;
        }
    }
}

void IntlDateTimeFormat::initializeDateTimeFormat(ExecState& exec, JSValue locales, JSValue originalOptions)
{
	VM& vm = exec.vm();
    ThrowScope scope(vm);

    // 12.1.1 InitializeDateTimeFormat (dateTimeFormat, locales, options) (ECMA-402)
    // https://tc39.github.io/ecma402/#sec-initializedatetimeformat

    Vector<String> requestedLocales = canonicalizeLocaleList(&exec, locales);
    RETURN_IF_EXCEPTION(vm, void());

    JSObject* options = toDateTimeOptionsAnyDate(exec, originalOptions);
    RETURN_IF_EXCEPTION(vm, void());

    HashMap<String, String> opt;

    String localeMatcher = intlStringOption(&exec, options, vm.propertyNames->localeMatcher, HashSet<String>(ASCIILiteral("lookup"), ASCIILiteral("best fit")), ASCIILiteral("localeMatcher must be either \"lookup\" or \"best fit\""), ASCIILiteral("best fit"));
    RETURN_IF_EXCEPTION(vm, void());
    opt.add(vm.propertyNames->localeMatcher.string(), localeMatcher);

    String calendar = intlStringOption(&exec, options, vm.propertyNames->calendar, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!calendar.isNull()) {
        if (!isUnicodeLocaleIdentifierType(calendar)) {
            throwRangeError(&exec, vm, ASCIILiteral("calendar is not a well-formed calendar value"));
            return;
        }
        opt.add(ASCIILiteral("ca"), calendar);
    }

    String numberingSystem = intlStringOption(&exec, options, vm.propertyNames->numberingSystem, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!numberingSystem.isNull()) {
        if (!isUnicodeLocaleIdentifierType(numberingSystem)) {
            throwRangeError(&exec, vm, ASCIILiteral("numberingSystem is not a well-formed numbering system value"));
            return;
        }
        opt.add(ASCIILiteral("nu"), numberingSystem);
    }

    bool isHour12Undefined;
    bool hour12 = intlBooleanOption(&exec, options, vm.propertyNames->hour12, isHour12Undefined);
    RETURN_IF_EXCEPTION(vm, void());

    String hourCycle = intlStringOption(&exec, options, vm.propertyNames->hourCycle, HashSet<String>(ASCIILiteral("h11"), ASCIILiteral("h12"), ASCIILiteral("h23"), ASCIILiteral("h24")), ASCIILiteral("hourCycle must be \"h11\", \"h12\", \"h23\", or \"h24\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (isHour12Undefined) {
        // Set hour12 here to simplify hour logic later.
        hour12 = (hourCycle == "h11" || hourCycle == "h12");
        if (!hourCycle.isNull())
            opt.add(ASCIILiteral("hc"), hourCycle);
    } else
        opt.add(ASCIILiteral("hc"), String());

    auto& availableLocales = intlDateTimeFormatAvailableLocales();
    HashMap<String, String> resolved = resolveLocale(exec, availableLocales, requestedLocales, opt, relevantExtensionKeys, WTF_ARRAY_LENGTH(relevantExtensionKeys), localeData);

    m_locale = resolved.get(vm.propertyNames->locale.string());
    if (m_locale.isEmpty()) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to initialize DateTimeFormat due to invalid locale"));
        return;
    }
    m_calendar = resolved.get(ASCIILiteral("ca"));
    if (m_calendar == "gregorian")
        m_calendar = ASCIILiteral("gregory");
    else if (m_calendar == "islamicc")
        m_calendar = ASCIILiteral("islamic-civil");
    else if (m_calendar == "ethioaa")
        m_calendar = ASCIILiteral("ethiopic-amete-alem");

    m_hourCycle = resolved.get(ASCIILiteral("hc"));
    m_numberingSystem = resolved.get(ASCIILiteral("nu"));
    CString dataLocaleWithExtensions = makeString(resolved.get(ASCIILiteral("dataLocale")), "-u-ca-", m_calendar, "-nu-", m_numberingSystem).utf8();

    JSValue tzValue = options->get(vm, &exec, vm.propertyNames->timeZone);
    RETURN_IF_EXCEPTION(vm, void());

    String tz;
    if (!tzValue.isUndefined()) {
        String originalTz = tzValue.toWTFString(&exec);
        RETURN_IF_EXCEPTION(vm, void());
        tz = canonicalizeTimeZoneName(originalTz);
        if (tz.isNull()) {
            throwRangeError(&exec, vm, String::format("invalid time zone: %s", originalTz.utf8().data()));
            return;
        }
    } else
        tz = defaultTimeZone();
    m_timeZone = tz;

    StringBuilder skeletonBuilder;

    HashSet<String> narrowShortLong(ASCIILiteral("narrow"), ASCIILiteral("short"), ASCIILiteral("long"));
    HashSet<String> twoDigitNumeric(ASCIILiteral("2-digit"), ASCIILiteral("numeric"));
    HashSet<String> twoDigitNumericNarrowShortLong(ASCIILiteral("2-digit"), ASCIILiteral("numeric"), ASCIILiteral("narrow"), ASCIILiteral("short"), ASCIILiteral("long"));
    HashSet<String> shortLong(ASCIILiteral("short"), ASCIILiteral("long"));

    String weekday = intlStringOption(&exec, options, vm.propertyNames->weekday, narrowShortLong, ASCIILiteral("weekday must be \"narrow\", \"short\", or \"long\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!weekday.isNull()) {
        if (weekday == "narrow")
            skeletonBuilder.appendLiteral("EEEEE");
        else if (weekday == "short")
            skeletonBuilder.appendLiteral("EEE");
        else if (weekday == "long")
            skeletonBuilder.appendLiteral("EEEE");
    }

    String era = intlStringOption(&exec, options, vm.propertyNames->era, narrowShortLong, ASCIILiteral("era must be \"narrow\", \"short\", or \"long\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!era.isNull()) {
        if (era == "narrow")
            skeletonBuilder.appendLiteral("GGGGG");
        else if (era == "short")
            skeletonBuilder.appendLiteral("GGG");
        else if (era == "long")
            skeletonBuilder.appendLiteral("GGGG");
    }

    String year = intlStringOption(&exec, options, vm.propertyNames->year, twoDigitNumeric, ASCIILiteral("year must be \"2-digit\" or \"numeric\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!year.isNull()) {
        if (year == "2-digit")
            skeletonBuilder.appendLiteral("yy");
        else if (year == "numeric")
            skeletonBuilder.append('y');
    }

    String month = intlStringOption(&exec, options, vm.propertyNames->month, twoDigitNumericNarrowShortLong, ASCIILiteral("month must be \"2-digit\", \"numeric\", \"narrow\", \"short\", or \"long\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!month.isNull()) {
        if (month == "2-digit")
            skeletonBuilder.appendLiteral("MM");
        else if (month == "numeric")
            skeletonBuilder.append('M');
        else if (month == "narrow")
            skeletonBuilder.appendLiteral("MMMMM");
        else if (month == "short")
            skeletonBuilder.appendLiteral("MMM");
        else if (month == "long")
            skeletonBuilder.appendLiteral("MMMM");
    }

    String day = intlStringOption(&exec, options, vm.propertyNames->day, twoDigitNumeric, ASCIILiteral("day must be \"2-digit\" or \"numeric\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!day.isNull()) {
        if (day == "2-digit")
            skeletonBuilder.appendLiteral("dd");
        else if (day == "numeric")
            skeletonBuilder.append('d');
    }

    String hour = intlStringOption(&exec, options, vm.propertyNames->hour, twoDigitNumeric, ASCIILiteral("hour must be \"2-digit\" or \"numeric\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (hour == "2-digit") {
        if (isHour12Undefined && m_hourCycle.isNull())
            skeletonBuilder.appendLiteral("jj");
        else if (hour12)
            skeletonBuilder.appendLiteral("hh");
        else
            skeletonBuilder.appendLiteral("HH");
    } else if (hour == "numeric") {
        if (isHour12Undefined && m_hourCycle.isNull())
            skeletonBuilder.append('j');
        else if (hour12)
            skeletonBuilder.append('h');
        else
            skeletonBuilder.append('H');
    } else
        m_hourCycle = String();

    String minute = intlStringOption(&exec, options, vm.propertyNames->minute, twoDigitNumeric, ASCIILiteral("minute must be \"2-digit\" or \"numeric\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!minute.isNull()) {
        if (minute == "2-digit")
            skeletonBuilder.appendLiteral("mm");
        else if (minute == "numeric")
            skeletonBuilder.append('m');
    }

    String second = intlStringOption(&exec, options, vm.propertyNames->second, twoDigitNumeric, ASCIILiteral("second must be \"2-digit\" or \"numeric\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!second.isNull()) {
        if (second == "2-digit")
            skeletonBuilder.appendLiteral("ss");
        else if (second == "numeric")
            skeletonBuilder.append('s');
    }

    String timeZoneName = intlStringOption(&exec, options, vm.propertyNames->timeZoneName, shortLong, ASCIILiteral("timeZoneName must be \"short\" or \"long\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(vm, void());
    if (!timeZoneName.isNull()) {
        if (timeZoneName == "short")
            skeletonBuilder.append('z');
        else if (timeZoneName == "long")
            skeletonBuilder.appendLiteral("zzzz");
    }

    intlStringOption(&exec, options, vm.propertyNames->formatMatcher, HashSet<String>(ASCIILiteral("basic"), ASCIILiteral("best fit")), ASCIILiteral("formatMatcher must be either \"basic\" or \"best fit\""), ASCIILiteral("best fit"));
    RETURN_IF_EXCEPTION(vm, void());

    // Always use ICU date format generator, rather than our own pattern list and matcher.
    UErrorCode status = U_ZERO_ERROR;
    UDateTimePatternGenerator* generator = udatpg_open(dataLocaleWithExtensions.data(), &status);
    if (U_FAILURE(status)) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to initialize DateTimeFormat"));
        return;
    }

    String skeleton = skeletonBuilder.toString();
    StringView skeletonView(skeleton);
    Vector<UChar, 32> patternBuffer(32);
    status = U_ZERO_ERROR;
    auto patternLength = udatpg_getBestPatternWithOptions(generator, skeletonView.upconvertedCharacters().get(), skeletonView.length(), UDATPG_MATCH_HOUR_FIELD_LENGTH, patternBuffer.data(), patternBuffer.size(), &status);
    if (needsToGrowToProduceBuffer(status)) {
        status = U_ZERO_ERROR;
        patternBuffer.grow(patternLength);
        udatpg_getBestPattern(generator, skeletonView.upconvertedCharacters().get(), skeletonView.length(), patternBuffer.data(), patternLength, &status);
    }
    udatpg_close(generator);
    if (U_FAILURE(status)) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to initialize DateTimeFormat"));
        return;
    }

    // Enforce our hourCycle, replacing hour characters in pattern.
    if (!m_hourCycle.isNull()) {
        UChar hour = 'H';
        if (m_hourCycle == "h11")
            hour = 'K';
        else if (m_hourCycle == "h12")
            hour = 'h';
        else if (m_hourCycle == "h24")
            hour = 'k';

        bool isEscaped = false;
        bool hasHour = false;
        for (auto i = 0; i < patternLength; ++i) {
            UChar c = patternBuffer[i];
            if (c == '\'')
                isEscaped = !isEscaped;
            else if (!isEscaped && (c == 'h' || c == 'H' || c == 'k' || c == 'K')) {
                patternBuffer[i] = hour;
                hasHour = true;
            }
        }
        if (!hasHour)
            m_hourCycle = String();
    }

    StringView pattern(patternBuffer.data(), patternLength);
    setFormatsFromPattern(pattern);

    status = U_ZERO_ERROR;
    StringView timeZoneView(m_timeZone);
    m_dateFormat = adoptPtr(udat_open(UDAT_PATTERN, UDAT_PATTERN, dataLocaleWithExtensions.data(), timeZoneView.upconvertedCharacters().get(), timeZoneView.length(), pattern.upconvertedCharacters(), pattern.length(), &status));
    if (U_FAILURE(status)) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to initialize DateTimeFormat"));
        return;
    }

    // Gregorian calendar should be used from the beginning of ECMAScript time.
    // Failure here means unsupported calendar, and can safely be ignored.
    UCalendar* cal = const_cast<UCalendar*>(udat_getCalendar(m_dateFormat.get()));
    ucal_setGregorianChange(cal, minECMAScriptTime, &status);

    m_initializedDateTimeFormat = true;
}

ASCIILiteral IntlDateTimeFormat::weekdayString(Weekday weekday)
{
    switch (weekday) {
    case Weekday::Narrow:
        return ASCIILiteral("narrow");
    case Weekday::Short:
        return ASCIILiteral("short");
    case Weekday::Long:
        return ASCIILiteral("long");
    case Weekday::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::eraString(Era era)
{
    switch (era) {
    case Era::Narrow:
        return ASCIILiteral("narrow");
    case Era::Short:
        return ASCIILiteral("short");
    case Era::Long:
        return ASCIILiteral("long");
    case Era::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::yearString(Year year)
{
    switch (year) {
    case Year::TwoDigit:
        return ASCIILiteral("2-digit");
    case Year::Numeric:
        return ASCIILiteral("numeric");
    case Year::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::monthString(Month month)
{
    switch (month) {
    case Month::TwoDigit:
        return ASCIILiteral("2-digit");
    case Month::Numeric:
        return ASCIILiteral("numeric");
    case Month::Narrow:
        return ASCIILiteral("narrow");
    case Month::Short:
        return ASCIILiteral("short");
    case Month::Long:
        return ASCIILiteral("long");
    case Month::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::dayString(Day day)
{
    switch (day) {
    case Day::TwoDigit:
        return ASCIILiteral("2-digit");
    case Day::Numeric:
        return ASCIILiteral("numeric");
    case Day::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::hourString(Hour hour)
{
    switch (hour) {
    case Hour::TwoDigit:
        return ASCIILiteral("2-digit");
    case Hour::Numeric:
        return ASCIILiteral("numeric");
    case Hour::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::minuteString(Minute minute)
{
    switch (minute) {
    case Minute::TwoDigit:
        return ASCIILiteral("2-digit");
    case Minute::Numeric:
        return ASCIILiteral("numeric");
    case Minute::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::secondString(Second second)
{
    switch (second) {
    case Second::TwoDigit:
        return ASCIILiteral("2-digit");
    case Second::Numeric:
        return ASCIILiteral("numeric");
    case Second::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

ASCIILiteral IntlDateTimeFormat::timeZoneNameString(TimeZoneName timeZoneName)
{
    switch (timeZoneName) {
    case TimeZoneName::Short:
        return ASCIILiteral("short");
    case TimeZoneName::Long:
        return ASCIILiteral("long");
    case TimeZoneName::None:
        ASSERT_NOT_REACHED();
        return ASCIILiteral();
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

JSObject* IntlDateTimeFormat::resolvedOptions(ExecState& exec)
{
    // 12.3.5 Intl.DateTimeFormat.prototype.resolvedOptions() (ECMA-402 2.0)
    // The function returns a new object whose properties and attributes are set as if constructed by an object literal assigning to each of the following properties the value of the corresponding internal slot of this DateTimeFormat object (see 12.4): locale, calendar, numberingSystem, timeZone, hour12, weekday, era, year, month, day, hour, minute, second, and timeZoneName. Properties whose corresponding internal slots are not present are not assigned.
    // Note: In this version of the ECMAScript 2015 Internationalization API, the timeZone property will be the name of the default time zone if no timeZone property was provided in the options object provided to the Intl.DateTimeFormat constructor. The previous version left the timeZone property undefined in this case.
    if (!m_initializedDateTimeFormat) {
        initializeDateTimeFormat(exec, jsUndefined(), jsUndefined());
        ASSERT(!exec.hadException());
    }

    JSGlobalObject* globalObject = exec.lexicalGlobalObject();
    VM& vm = globalObject->vm();
    JSObject* options = constructEmptyObject(globalObject);
    options->putDirect(vm, vm.propertyNames->locale, jsNontrivialString(vm, m_locale));
    options->putDirect(vm, vm.propertyNames->calendar, jsNontrivialString(vm, m_calendar));
    options->putDirect(vm, vm.propertyNames->numberingSystem, jsNontrivialString(vm, m_numberingSystem));
    options->putDirect(vm, vm.propertyNames->timeZone, jsNontrivialString(vm, m_timeZone));

    if (!m_hourCycle.isNull()) {
        options->putDirect(vm, vm.propertyNames->hourCycle, jsNontrivialString(vm, m_hourCycle));
        options->putDirect(vm, vm.propertyNames->hour12, jsBoolean(m_hourCycle == "h11" || m_hourCycle == "h12"));
    }

    if (m_weekday != Weekday::None)
        options->putDirect(vm, vm.propertyNames->weekday, jsNontrivialString(vm, weekdayString(m_weekday)));

    if (m_era != Era::None)
        options->putDirect(vm, vm.propertyNames->era, jsNontrivialString(vm, eraString(m_era)));

    if (m_year != Year::None)
        options->putDirect(vm, vm.propertyNames->year, jsNontrivialString(vm, yearString(m_year)));

    if (m_month != Month::None)
        options->putDirect(vm, vm.propertyNames->month, jsNontrivialString(vm, monthString(m_month)));

    if (m_day != Day::None)
        options->putDirect(vm, vm.propertyNames->day, jsNontrivialString(vm, dayString(m_day)));

    if (m_hour != Hour::None)
        options->putDirect(vm, vm.propertyNames->hour, jsNontrivialString(vm, hourString(m_hour)));

    if (m_minute != Minute::None)
        options->putDirect(vm, vm.propertyNames->minute, jsNontrivialString(vm, minuteString(m_minute)));

    if (m_second != Second::None)
        options->putDirect(vm, vm.propertyNames->second, jsNontrivialString(vm, secondString(m_second)));

    if (m_timeZoneName != TimeZoneName::None)
        options->putDirect(vm, vm.propertyNames->timeZoneName, jsNontrivialString(vm, timeZoneNameString(m_timeZoneName)));

    return options;
}

JSValue IntlDateTimeFormat::format(ExecState& exec, double value)
{
	VM& vm = exec.vm();

    // 12.3.4 FormatDateTime abstract operation (ECMA-402 2.0)
    if (!m_initializedDateTimeFormat) {
        initializeDateTimeFormat(exec, jsUndefined(), jsUndefined());
        ASSERT(!exec.hadException());
    }

    // 1. If x is not a finite Number, then throw a RangeError exception.
    if (!std::isfinite(value))
        return throwRangeError(&exec, vm, ASCIILiteral("date value is not finite in DateTimeFormat format()"));

    // Delegate remaining steps to ICU.
    UErrorCode status = U_ZERO_ERROR;
    Vector<UChar, 32> result(32);
    auto resultLength = udat_format(m_dateFormat.get(), value, result.data(), result.size(), 0, &status);
    if (needsToGrowToProduceBuffer(status)) {
        status = U_ZERO_ERROR;
        result.grow(resultLength);
        udat_format(m_dateFormat.get(), value, result.data(), resultLength, 0, &status);
    }
    if (U_FAILURE(status))
        return throwTypeError(&exec, vm, ASCIILiteral("failed to format date value"));

    return jsString(vm, String(result.data(), resultLength));
}

static ASCIILiteral partTypeString(UDateFormatField field)
{
    switch (field) {
    case UDAT_ERA_FIELD:
        return ASCIILiteral("era");
    case UDAT_YEAR_FIELD:
    case UDAT_EXTENDED_YEAR_FIELD:
        return ASCIILiteral("year");
    case UDAT_YEAR_NAME_FIELD:
        return ASCIILiteral("yearName");
    case UDAT_MONTH_FIELD:
    case UDAT_STANDALONE_MONTH_FIELD:
        return ASCIILiteral("month");
    case UDAT_DATE_FIELD:
        return ASCIILiteral("day");
    case UDAT_HOUR_OF_DAY1_FIELD:
    case UDAT_HOUR_OF_DAY0_FIELD:
    case UDAT_HOUR1_FIELD:
    case UDAT_HOUR0_FIELD:
        return ASCIILiteral("hour");
    case UDAT_MINUTE_FIELD:
        return ASCIILiteral("minute");
    case UDAT_SECOND_FIELD:
    case UDAT_FRACTIONAL_SECOND_FIELD:
        return ASCIILiteral("second");
    case UDAT_DAY_OF_WEEK_FIELD:
    case UDAT_DOW_LOCAL_FIELD:
    case UDAT_STANDALONE_DAY_FIELD:
        return ASCIILiteral("weekday");
    case UDAT_AM_PM_FIELD:
    case UDAT_AM_PM_MIDNIGHT_NOON_FIELD:
    case UDAT_FLEXIBLE_DAY_PERIOD_FIELD:
        return ASCIILiteral("dayPeriod");
    case UDAT_TIMEZONE_FIELD:
    case UDAT_TIMEZONE_RFC_FIELD:
    case UDAT_TIMEZONE_GENERIC_FIELD:
    case UDAT_TIMEZONE_SPECIAL_FIELD:
    case UDAT_TIMEZONE_LOCALIZED_GMT_OFFSET_FIELD:
    case UDAT_TIMEZONE_ISO_FIELD:
    case UDAT_TIMEZONE_ISO_LOCAL_FIELD:
        return ASCIILiteral("timeZoneName");
    case UDAT_RELATED_YEAR_FIELD:
        return ASCIILiteral("relatedYear");
    // These should not show up because there is no way to specify them in DateTimeFormat options.
	// If they do, they don't fit well into any of known part types, so consider it an "unknown".
	case UDAT_DAY_OF_YEAR_FIELD:
    case UDAT_DAY_OF_WEEK_IN_MONTH_FIELD:
    case UDAT_WEEK_OF_YEAR_FIELD:
    case UDAT_WEEK_OF_MONTH_FIELD:
    case UDAT_YEAR_WOY_FIELD:
    case UDAT_JULIAN_DAY_FIELD:
    case UDAT_MILLISECONDS_IN_DAY_FIELD:
    case UDAT_QUARTER_FIELD:
    case UDAT_STANDALONE_QUARTER_FIELD:
    case UDAT_TIME_SEPARATOR_FIELD:
#if U_ICU_VERSION_MAJOR_NUM < 58
    case UDAT_FIELD_COUNT:
#endif
        return ASCIILiteral("unknown");
    }
    // Any newer additions to the UDateFormatField enum should just be considered an "unknown" part.
    return ASCIILiteral("unknown");
}

JSValue IntlDateTimeFormat::formatToParts(ExecState& exec, double value)
{
    JSGlobalObject* globalObject = exec.lexicalGlobalObject();
    VM& vm = globalObject->vm();
    ThrowScope scope(vm);

    // 12.1.8 FormatDateTimeToParts (ECMA-402 4.0)
    // https://tc39.github.io/ecma402/#sec-formatdatetimetoparts

    if (!std::isfinite(value))
        return throwRangeError(&exec, vm, ASCIILiteral("date value is not finite in DateTimeFormat formatToParts()"));

    UErrorCode status = U_ZERO_ERROR;
    OwnPtr<UFieldPositionIterator> fields = adoptPtr(ufieldpositer_open(&status));
    if (U_FAILURE(status))
        return throwTypeError(&exec, vm, ASCIILiteral("failed to open field position iterator"));

    status = U_ZERO_ERROR;
    Vector<UChar, 32> result(32);
    auto resultLength = udat_formatForFields(m_dateFormat.get(), value, result.data(), result.size(), fields.get(), &status);
    if (status == U_BUFFER_OVERFLOW_ERROR) {
        status = U_ZERO_ERROR;
        result.grow(resultLength);
        udat_formatForFields(m_dateFormat.get(), value, result.data(), resultLength, fields.get(), &status);
    }
    if (U_FAILURE(status))
        return throwTypeError(&exec, vm, ASCIILiteral("failed to format date value"));

    JSArray* parts = JSArray::tryCreate(vm, &exec, globalObject->arrayStructureForIndexingTypeDuringAllocation(ArrayWithContiguous), 0);
    RETURN_IF_EXCEPTION(scope, jsUndefined());

    auto resultString = String(result.data(), resultLength);
    auto literalString = jsNontrivialString(vm, ASCIILiteral("literal"));

    int32_t previousEndIndex = 0;
    int32_t beginIndex = 0;
    int32_t endIndex = 0;
    while (previousEndIndex < resultLength) {
        auto fieldType = ufieldpositer_next(fields.get(), &beginIndex, &endIndex);
        if (fieldType < 0)
            beginIndex = endIndex = resultLength;

        if (previousEndIndex < beginIndex) {
            auto value = jsString(vm, resultString.substring(previousEndIndex, beginIndex - previousEndIndex));
            JSObject* part = constructEmptyObject(globalObject);
            part->putDirect(vm, vm.propertyNames->type, literalString);
            part->putDirect(vm, vm.propertyNames->value, value);
            parts->push(&exec, part);
            RETURN_IF_EXCEPTION(scope, jsUndefined());
        }
        previousEndIndex = endIndex;

        if (fieldType >= 0) {
            JSString* type = jsString(vm, partTypeString(UDateFormatField(fieldType)));
            auto value = jsString(vm, resultString.substring(beginIndex, endIndex - beginIndex));
            JSObject* part = constructEmptyObject(globalObject);
            part->putDirect(vm, vm.propertyNames->type, type);
            part->putDirect(vm, vm.propertyNames->value, value);
            parts->push(&exec, part);
            RETURN_IF_EXCEPTION(scope, jsUndefined());
        }
    }
    return parts;
}

} // namespace JSC

#endif // ENABLE(INTL)
