/*
 * Copyright (C) 2020 Sony Interactive Entertainment Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "IntlLocale.h"

#include "Error.h"
#include "IntlObject.h"
#include "JSCInlines.h"
#include "ObjectConstructor.h"
#include <unicode/uloc.h>
#include <wtf/unicode/icu/ICUHelpers.h>

namespace JSC {

const ClassInfo IntlLocale::s_info = { "Object", &Base::s_info, nullptr, CREATE_METHOD_TABLE(IntlLocale) };

IntlLocale* IntlLocale::create(VM& vm, Structure* structure)
{
    auto* object = new (NotNull, allocateCell<IntlLocale>(vm.heap)) IntlLocale(vm, structure);
    object->finishCreation(vm);
    return object;
}

Structure* IntlLocale::createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
{
    return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
}

IntlLocale::IntlLocale(VM& vm, Structure* structure)
    : Base(vm, structure)
{
}

void IntlLocale::finishCreation(VM& vm)
{
    Base::finishCreation(vm);
    ASSERT(inherits(vm, info()));
}

void IntlLocale::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    auto* thisObject = jsCast<IntlLocale*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());

    Base::visitChildren(thisObject, visitor);
}

class LocaleIDBuilder final {
public:
    bool initialize(const String&);
    CString toCanonical();

    void overrideLanguageScriptRegion(StringView language, StringView script, StringView region);
    void setKeywordValue(ASCIILiteral key, StringView value);

private:
    Vector<char, 32> m_buffer;
};

bool LocaleIDBuilder::initialize(const String& tag)
{
    if (!isStructurallyValidLanguageTag(tag))
        return false;
    ASSERT(tag.isAllASCII());
    m_buffer = localeIDBufferForLanguageTag(tag.ascii());
    return m_buffer.size();
}

CString LocaleIDBuilder::toCanonical()
{
    ASSERT(m_buffer.size());

    UErrorCode status = U_ZERO_ERROR;
    Vector<char, 32> result(32);
    auto resultLength = uloc_canonicalize(m_buffer.data(), result.data(), result.size(), &status);
    if (needsToGrowToProduceBuffer(status)) {
        result.grow(resultLength);
        status = U_ZERO_ERROR;
        uloc_canonicalize(m_buffer.data(), result.data(), resultLength, &status);
    }
    if (U_FAILURE(status))
        return CString();

    return CString(result.data(), resultLength);
}

// Because ICU's C API doesn't have set[Language|Script|Region] functions...
void LocaleIDBuilder::overrideLanguageScriptRegion(StringView language, StringView script, StringView region)
{
    unsigned length = strlen(m_buffer.data());
    ASSERT(length);

    StringView localeIDView { m_buffer.data(), length };

    auto endOfLanguageScriptRegionVariant = localeIDView.find(ULOC_KEYWORD_SEPARATOR);
    if (endOfLanguageScriptRegionVariant == notFound)
        endOfLanguageScriptRegionVariant = length;

    Vector<StringView> subtags;
    for (auto subtag : localeIDView.left(endOfLanguageScriptRegionVariant).splitAllowingEmptyEntries('_'))
        subtags.append(subtag);

    if (!language.isNull())
        subtags[0] = language;

    bool hasScript = subtags.size() > 1 && subtags[1].length() == 4;
    if (!script.isNull()) {
        if (hasScript)
            subtags[1] = script;
        else {
            subtags.insert(1, script);
            hasScript = true;
        }
    }

    if (!region.isNull()) {
        size_t index = hasScript ? 2 : 1;
        bool hasRegion = subtags.size() > index && subtags[index].length() < 4;
        if (hasRegion)
            subtags[index] = region;
        else
            subtags.insert(index, region);
    }

    Vector<char, 32> buffer;
    bool hasAppended = false;
    for (auto subtag : subtags) {
        if (hasAppended)
            buffer.append('_');
        else
            hasAppended = true;

        ASSERT(subtag.isAllASCII());
        if (subtag.is8Bit())
            buffer.append(subtag.characters8(), subtag.length());
        else
            buffer.append(subtag.characters16(), subtag.length());
    }

    if (endOfLanguageScriptRegionVariant != length) {
        auto rest = localeIDView.right(length - endOfLanguageScriptRegionVariant);

        ASSERT(rest.isAllASCII());
        if (rest.is8Bit())
            buffer.append(rest.characters8(), rest.length());
        else
            buffer.append(rest.characters16(), rest.length());
    }

    buffer.append('\0');
    m_buffer.swap(buffer);
}

void LocaleIDBuilder::setKeywordValue(ASCIILiteral key, StringView value)
{
    ASSERT(m_buffer.size());

    ASSERT(value.isAllASCII());
    Vector<char, 32> rawValue(value.length() + 1);
    if (value.is8Bit())
        StringImpl::copyChars(reinterpret_cast<LChar*>(rawValue.data()), value.characters8(), value.length());
    else
        StringImpl::copyChars(reinterpret_cast<LChar*>(rawValue.data()), value.characters16(), value.length());
    rawValue[value.length()] = '\0';

    UErrorCode status = U_ZERO_ERROR;
    auto length = uloc_setKeywordValue(key.characters(), rawValue.data(), m_buffer.data(), m_buffer.size(), &status);
    // uloc_setKeywordValue does not set U_STRING_NOT_TERMINATED_WARNING.
    if (needsToGrowToProduceBuffer(status)) {
        m_buffer.grow(length + 1);
        status = U_ZERO_ERROR;
        uloc_setKeywordValue(key.characters(), rawValue.data(), m_buffer.data(), length + 1, &status);
    }
    ASSERT(U_SUCCESS(status));
}

String IntlLocale::keywordValue(ASCIILiteral key, bool isBoolean) const
{
    UErrorCode status = U_ZERO_ERROR;
    Vector<char, 32> buffer(32);
    auto bufferLength = uloc_getKeywordValue(m_localeID.data(), key.characters(), buffer.data(), buffer.size(), &status);
    if (needsToGrowToProduceCString(status)) {
        buffer.grow(bufferLength + 1);
        status = U_ZERO_ERROR;
        uloc_getKeywordValue(m_localeID.data(), key.characters(), buffer.data(), bufferLength + 1, &status);
    }
    if (!U_SUCCESS(status))
    	return emptyString();
    if (isBoolean)
        return String::fromLatin1(buffer.data());
    const char* value = uloc_toUnicodeLocaleType(key.characters(), buffer.data());
    if (!value)
        return String();
    auto result = String::fromLatin1(value);
    if (result == "true")
        return emptyString();
    return result;
}

// https://tc39.es/ecma402/#sec-Intl.Locale
void IntlLocale::initializeLocale(ExecState* exec, JSValue tagValue, JSValue optionsValue)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    String tag = tagValue.inherits<IntlLocale>(vm) ? jsCast<IntlLocale*>(tagValue)->toString() : tagValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(scope, void());
    initializeLocale(exec, tag, optionsValue);
}

// https://tc39.es/ecma402/#sec-Intl.Locale
void IntlLocale::initializeLocale(ExecState* exec, const String& tag, JSValue optionsValue)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue options = optionsValue;
    if (!optionsValue.isUndefined()) {
        options = optionsValue.toObject(exec);
        RETURN_IF_EXCEPTION(scope, void());
    }

    LocaleIDBuilder localeID;
    if (!localeID.initialize(tag)) {
        throwRangeError(exec, vm, ASCIILiteral("invalid language tag"));
        return;
    }

    String language = intlStringOption(exec, options, vm.propertyNames->language, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!language.isNull() && !isUnicodeLanguageSubtag(language)) {
        throwRangeError(exec, vm, ASCIILiteral("language is not a well-formed language value"));
        return;
    }

    String script = intlStringOption(exec, options, vm.propertyNames->script, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!script.isNull() && !isUnicodeScriptSubtag(script)) {
        throwRangeError(exec, vm, ASCIILiteral("script is not a well-formed script value"));
        return;
    }

    String region = intlStringOption(exec, options, vm.propertyNames->region, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!region.isNull() && !isUnicodeRegionSubtag(region)) {
        throwRangeError(exec, vm, ASCIILiteral("region is not a well-formed region value"));
        return;
    }

    if (!language.isNull() || !script.isNull() || !region.isNull())
        localeID.overrideLanguageScriptRegion(language, script, region);

    String calendar = intlStringOption(exec, options, vm.propertyNames->calendar, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!calendar.isNull()) {
        if (!isUnicodeLocaleIdentifierType(calendar)) {
            throwRangeError(exec, vm, ASCIILiteral("calendar is not a well-formed calendar value"));
            return;
        }
        localeID.setKeywordValue(ASCIILiteral("calendar"), calendar);
    }

    String collation = intlStringOption(exec, options, vm.propertyNames->collation, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!collation.isNull()) {
        if (!isUnicodeLocaleIdentifierType(collation)) {
            throwRangeError(exec, vm, ASCIILiteral("collation is not a well-formed collation value"));
            return;
        }
        localeID.setKeywordValue(ASCIILiteral("collation"), collation);
    }

    String hourCycle = intlStringOption(exec, options, vm.propertyNames->hourCycle, HashSet<String>({ ASCIILiteral("h11"), ASCIILiteral("h12"), ASCIILiteral("h23"), ASCIILiteral("h24") }), ASCIILiteral("hourCycle must be \"h11\", \"h12\", \"h23\", or \"h24\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!hourCycle.isNull())
        localeID.setKeywordValue(ASCIILiteral("hours"), hourCycle);

    String caseFirst = intlStringOption(exec, options, vm.propertyNames->caseFirst, HashSet<String>({ ASCIILiteral("upper"), ASCIILiteral("lower"), ASCIILiteral("false") }), ASCIILiteral("caseFirst must be either \"upper\", \"lower\", or \"false\""), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!caseFirst.isNull())
        localeID.setKeywordValue(ASCIILiteral("colcasefirst"), caseFirst);

    bool usesFallback;
    auto numeric = intlBooleanOption(exec, options, vm.propertyNames->numeric, usesFallback);
    RETURN_IF_EXCEPTION(scope, void());
    if (!usesFallback)
        localeID.setKeywordValue(ASCIILiteral("colnumeric"), numeric == true ? "yes" : "no");

    String numberingSystem = intlStringOption(exec, options, vm.propertyNames->numberingSystem, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!numberingSystem.isNull()) {
        if (!isUnicodeLocaleIdentifierType(numberingSystem)) {
            throwRangeError(exec, vm, ASCIILiteral("numberingSystem is not a well-formed numbering system value"));
            return;
        }
        localeID.setKeywordValue(ASCIILiteral("numbers"), numberingSystem);
    }

    m_localeID = localeID.toCanonical();
    if (m_localeID.isNull()) {
        throwTypeError(exec, vm, "failed to initialize Locale");
        return;
    }
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.maximize
const String& IntlLocale::maximal()
{
    if (m_maximal.isNull()) {
        UErrorCode status = U_ZERO_ERROR;
        Vector<char, 32> buffer(32);
        auto bufferLength = uloc_addLikelySubtags(m_localeID.data(), buffer.data(), buffer.size(), &status);
        if (needsToGrowToProduceCString(status)) {
            buffer.grow(bufferLength + 1);
            status = U_ZERO_ERROR;
            uloc_addLikelySubtags(m_localeID.data(), buffer.data(), bufferLength + 1, &status);
        }
        ASSERT(U_SUCCESS(status));

        m_maximal = languageTagForLocaleID(buffer.data());
    }
    return m_maximal;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.minimize
const String& IntlLocale::minimal()
{
    if (m_minimal.isNull()) {
        UErrorCode status = U_ZERO_ERROR;
        Vector<char, 32> buffer(32);
        auto bufferLength = uloc_minimizeSubtags(m_localeID.data(), buffer.data(), buffer.size(), &status);
        if (needsToGrowToProduceCString(status)) {
            buffer.grow(bufferLength + 1);
            status = U_ZERO_ERROR;
            uloc_minimizeSubtags(m_localeID.data(), buffer.data(), bufferLength + 1, &status);
        }
        ASSERT(U_SUCCESS(status));

        m_minimal = languageTagForLocaleID(buffer.data());
    }
    return m_minimal;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.toString
const String& IntlLocale::toString()
{
    if (m_fullString.isNull())
        m_fullString = languageTagForLocaleID(m_localeID.data());
    return m_fullString;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.baseName
const String& IntlLocale::baseName()
{
    if (m_baseName.isNull()) {
        UErrorCode status = U_ZERO_ERROR;
        Vector<char, 32> buffer(32);
        auto bufferLength = uloc_getBaseName(m_localeID.data(), buffer.data(), buffer.size(), &status);
        if (needsToGrowToProduceCString(status)) {
            buffer.grow(bufferLength + 1);
            status = U_ZERO_ERROR;
            uloc_getBaseName(m_localeID.data(), buffer.data(), bufferLength + 1, &status);
        }
        ASSERT(U_SUCCESS(status));

        m_baseName = languageTagForLocaleID(buffer.data());
    }
    return m_baseName;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.language
const String& IntlLocale::language()
{
    if (m_language.isNull()) {
        UErrorCode status = U_ZERO_ERROR;
        Vector<char, 8> buffer(8);
        auto bufferLength = uloc_getLanguage(m_localeID.data(), buffer.data(), buffer.size(), &status);
        if (needsToGrowToProduceBuffer(status)) {
            buffer.grow(bufferLength);
            status = U_ZERO_ERROR;
            uloc_getLanguage(m_localeID.data(), buffer.data(), bufferLength, &status);
        }
        ASSERT(U_SUCCESS(status));

        m_language = String(buffer.data(), bufferLength);
    }
    return m_language;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.script
const String& IntlLocale::script()
{
    if (m_script.isNull()) {
        UErrorCode status = U_ZERO_ERROR;
        Vector<char, 4> buffer(4);
        auto bufferLength = uloc_getScript(m_localeID.data(), buffer.data(), buffer.size(), &status);
        if (needsToGrowToProduceBuffer(status)) {
            buffer.grow(bufferLength);
            status = U_ZERO_ERROR;
            uloc_getScript(m_localeID.data(), buffer.data(), bufferLength, &status);
        }
        ASSERT(U_SUCCESS(status));

        m_script = String(buffer.data(), bufferLength);
    }
    return m_script;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.region
const String& IntlLocale::region()
{
    if (m_region.isNull()) {
        UErrorCode status = U_ZERO_ERROR;
        Vector<char, 3> buffer(3);
        auto bufferLength = uloc_getCountry(m_localeID.data(), buffer.data(), buffer.size(), &status);
        if (needsToGrowToProduceBuffer(status)) {
            buffer.grow(bufferLength);
            status = U_ZERO_ERROR;
            uloc_getCountry(m_localeID.data(), buffer.data(), bufferLength, &status);
        }
        ASSERT(U_SUCCESS(status));

        m_region = String(buffer.data(), bufferLength);
    }
    return m_region;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.calendar
const String& IntlLocale::calendar()
{
    if (m_calendar.isNull())
        m_calendar = keywordValue(ASCIILiteral("calendar"));
    return m_calendar;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.caseFirst
const String& IntlLocale::caseFirst()
{
    if (m_caseFirst.isNull())
        m_caseFirst = keywordValue(ASCIILiteral("colcasefirst"));
    return m_caseFirst;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.collation
const String& IntlLocale::collation()
{
    if (m_collation.isNull())
        m_collation = keywordValue(ASCIILiteral("collation"));
    return m_collation;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.hourCycle
const String& IntlLocale::hourCycle()
{
    if (m_hourCycle.isNull())
        m_hourCycle = keywordValue(ASCIILiteral("hours"));
    return m_hourCycle;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.numberingSystem
const String& IntlLocale::numberingSystem()
{
    if (m_numberingSystem.isNull())
        m_numberingSystem = keywordValue(ASCIILiteral("numbers"));
    return m_numberingSystem;
}

// https://tc39.es/ecma402/#sec-Intl.Locale.prototype.numeric
TriState IntlLocale::numeric()
{
    constexpr bool isBoolean = true;
    if (m_numeric == TriState::MixedTriState)
        m_numeric = triState(keywordValue(ASCIILiteral("colnumeric"), isBoolean) == "yes");
    return m_numeric;
}

} // namespace JSC
