/*
 * Copyright (C) 2020 Sony Interactive Entertainment Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "IntlRelativeTimeFormat.h"

#include "Error.h"
#include "IntlNumberFormat.h"
#include "IntlObject.h"
#include "JSCInlines.h"
#include "ObjectConstructor.h"
#include <wtf/unicode/icu/ICUHelpers.h>

namespace WTF {

template <>
ALWAYS_INLINE void deleteOwnedPtr<URelativeDateTimeFormatter>(URelativeDateTimeFormatter* relativeDateTimeFormatter)
{
    if (relativeDateTimeFormatter)
        ureldatefmt_close(relativeDateTimeFormatter);
}

template <>
ALWAYS_INLINE void deleteOwnedPtr<UNumberFormat>(UNumberFormat* numberFormat)
{
    if (numberFormat)
        unum_close(numberFormat);
}

template <>
ALWAYS_INLINE void deleteOwnedPtr<UFieldPositionIterator>(UFieldPositionIterator* iterator)
{
	if (iterator)
		ufieldpositer_close(iterator);
}

}

namespace JSC {

const ClassInfo IntlRelativeTimeFormat::s_info = { "Object", &Base::s_info, nullptr, CREATE_METHOD_TABLE(IntlRelativeTimeFormat) };

constexpr const char* relevantExtensionKeys[1] = { "nu" };

IntlRelativeTimeFormat* IntlRelativeTimeFormat::create(VM& vm, Structure* structure)
{
    auto* format = new (NotNull, allocateCell<IntlRelativeTimeFormat>(vm.heap)) IntlRelativeTimeFormat(vm, structure);
    format->finishCreation(vm);
    return format;
}

Structure* IntlRelativeTimeFormat::createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
{
    return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
}

IntlRelativeTimeFormat::IntlRelativeTimeFormat(VM& vm, Structure* structure)
    : Base(vm, structure)
{
}

IntlRelativeTimeFormat::~IntlRelativeTimeFormat() = default;

void IntlRelativeTimeFormat::finishCreation(VM& vm)
{
    Base::finishCreation(vm);
    ASSERT(inherits(vm, info()));
}

void IntlRelativeTimeFormat::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    auto* thisObject = jsCast<IntlRelativeTimeFormat*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());

    Base::visitChildren(thisObject, visitor);
}

Vector<String> IntlRelativeTimeFormat::localeData(const String& locale, size_t keyIndex)
{
    // The index of the extension key "nu" in relevantExtensionKeys is 0.
    ASSERT_UNUSED(keyIndex, !keyIndex);
    return numberingSystemsForLocale(locale);
}

// https://tc39.es/ecma402/#sec-InitializeRelativeTimeFormat
void IntlRelativeTimeFormat::initializeRelativeTimeFormat(ExecState& exec, JSValue locales, JSValue optionsValue)
{
    JSGlobalObject* globalObject = exec.lexicalGlobalObject();
    VM& vm = globalObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    Vector<String> requestedLocales = canonicalizeLocaleList(&exec, locales);
    RETURN_IF_EXCEPTION(scope, void());

    JSObject* options;
    if (optionsValue.isUndefined())
        options = constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure());
    else {
        options = optionsValue.toObject(&exec);
        RETURN_IF_EXCEPTION(scope, void());
    }

    HashMap<String, String> opt;
    String localeMatcher = intlStringOption(&exec, options, vm.propertyNames->localeMatcher, HashSet<String>({ ASCIILiteral("lookup"), ASCIILiteral("best fit") }), ASCIILiteral("localeMatcher must be either \"lookup\" or \"best fit\""), ASCIILiteral("best fit"));
    RETURN_IF_EXCEPTION(scope, void());
    opt.add(vm.propertyNames->localeMatcher.string(), localeMatcher);

    String numberingSystem = intlStringOption(&exec, options, vm.propertyNames->numberingSystem, HashSet<String>(), ASCIILiteral(), ASCIILiteral());
    RETURN_IF_EXCEPTION(scope, void());
    if (!numberingSystem.isNull()) {
        if (!isUnicodeLocaleIdentifierType(numberingSystem)) {
            throwRangeError(&exec, vm, ASCIILiteral("numberingSystem is not a well-formed numbering system value"));
            return;
        }
        opt.add(ASCIILiteral("nu"), numberingSystem);
    }

    const HashSet<String>& availableLocales = intlRelativeTimeFormatAvailableLocales();
    HashMap<String, String> resolved = resolveLocale(exec, availableLocales, requestedLocales, opt, relevantExtensionKeys, WTF_ARRAY_LENGTH(relevantExtensionKeys), localeData);
    m_locale = resolved.get(vm.propertyNames->locale.string());
    if (m_locale.isEmpty()) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to initialize RelativeTimeFormat due to invalid locale"));
        return;
    }

    m_numberingSystem = resolved.get(ASCIILiteral("nu"));
    CString dataLocaleWithExtensions = makeString(resolved.get(ASCIILiteral("dataLocale")), "-u-nu-", m_numberingSystem).utf8();

    String style = intlStringOption(&exec, options, vm.propertyNames->style, HashSet<String>({ ASCIILiteral("long"), ASCIILiteral("short"), ASCIILiteral("narrow") }), ASCIILiteral("style must be either \"long\", \"short\", or \"narrow\""), ASCIILiteral("long"));
    RETURN_IF_EXCEPTION(scope, void());
    if (style == "long")
        m_style = UDAT_STYLE_LONG;
    else if (style == "short")
        m_style = UDAT_STYLE_SHORT;
    else if (style == "narrow")
        m_style = UDAT_STYLE_NARROW;
    else
        ASSERT_NOT_REACHED();

    String numeric = intlStringOption(&exec, options, vm.propertyNames->numeric, HashSet<String>({ ASCIILiteral("always"), ASCIILiteral("auto") }), ASCIILiteral("numeric must be either \"always\" or \"auto\""), ASCIILiteral("always"));
    RETURN_IF_EXCEPTION(scope, void());
    m_numeric = (numeric == "always");

    UErrorCode status = U_ZERO_ERROR;
    m_numberFormat = adoptPtr(unum_open(UNUM_DECIMAL, nullptr, 0, dataLocaleWithExtensions.data(), nullptr, &status));
    if (UNLIKELY(U_FAILURE(status))) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to initialize RelativeTimeFormat"));
        return;
    }

    // Align to IntlNumberFormat's default.
    unum_setAttribute(m_numberFormat.get(), UNUM_MIN_INTEGER_DIGITS, 1);
    unum_setAttribute(m_numberFormat.get(), UNUM_MIN_FRACTION_DIGITS, 0);
    unum_setAttribute(m_numberFormat.get(), UNUM_MAX_FRACTION_DIGITS, 3);
    unum_setAttribute(m_numberFormat.get(), UNUM_GROUPING_USED, true);

    // Grouping attributes have hidden -2 option which makes grouping rules locale-sensitive.
    // While this is supported so long, this is not explicitly exposed as APIs.
    // After ICU 68, it is now exposed as UNUM_MINIMUM_GROUPING_DIGITS_AUTO. Before ICU 68, we can directly use -2.
    // https://unicode-org.atlassian.net/browse/ICU-21109
    // https://github.com/unicode-org/icu/commit/e7bd5b1cefa47a043a9714e21eb9946dd54d593f
    //
    // These options are tested in test262/test/intl402/RelativeTimeFormat/prototype/format/pl-pl-style-long.js etc.
    // e.g. https://github.com/tc39/test262/commit/79c1818a6812a2a6c47e3e3c56ba9f2b3eaff4d5
    constexpr int32_t useLocaleDefault = -2;
    unum_setAttribute(m_numberFormat.get(), UNUM_GROUPING_SIZE, useLocaleDefault);
    unum_setAttribute(m_numberFormat.get(), UNUM_SECONDARY_GROUPING_SIZE, useLocaleDefault);
    unum_setAttribute(m_numberFormat.get(), UNUM_MINIMUM_GROUPING_DIGITS, useLocaleDefault);

    UNumberFormat* clonedNumberFormat = unum_clone(m_numberFormat.get(), &status);
    if (UNLIKELY(U_FAILURE(status))) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to initialize RelativeTimeFormat"));
        return;
    }

    m_relativeDateTimeFormatter = adoptPtr(ureldatefmt_open(dataLocaleWithExtensions.data(), clonedNumberFormat, m_style, UDISPCTX_CAPITALIZATION_FOR_STANDALONE, &status));
    if (UNLIKELY(U_FAILURE(status))) {
        throwTypeError(&exec, vm, "failed to initialize RelativeTimeFormat");
        return;
    }
}

static ASCIILiteral styleString(UDateRelativeDateTimeFormatterStyle style)
{
    switch (style) {
    case UDAT_STYLE_LONG:
        return ASCIILiteral("long");
    case UDAT_STYLE_SHORT:
        return ASCIILiteral("short");
    case UDAT_STYLE_NARROW:
        return ASCIILiteral("narrow");
    }
    ASSERT_NOT_REACHED();
    return ASCIILiteral();
}

static ASCIILiteral numericString(bool numeric)
{
    return numeric ? ASCIILiteral("always") : ASCIILiteral("auto");
}

// https://tc39.es/ecma402/#sec-intl.relativetimeformat.prototype.resolvedoptions
JSObject* IntlRelativeTimeFormat::resolvedOptions(ExecState& exec)
{
    JSGlobalObject* globalObject = exec.lexicalGlobalObject();
    VM& vm = globalObject->vm();
    JSObject* options = constructEmptyObject(globalObject);
    options->putDirect(vm, vm.propertyNames->locale, jsNontrivialString(vm, m_locale));
    options->putDirect(vm, vm.propertyNames->style, jsNontrivialString(vm, styleString(m_style)));
    options->putDirect(vm, vm.propertyNames->numeric, jsNontrivialString(vm, numericString(m_numeric)));
    options->putDirect(vm, vm.propertyNames->numberingSystem, jsNontrivialString(vm, m_numberingSystem));
    return options;
}

static StringView singularUnit(StringView unit)
{
    // Plurals are allowed, but thankfully they're all just a simple -s.
    return unit.endsWith("s") ? unit.left(unit.length() - 1) : unit;
}

// https://tc39.es/ecma402/#sec-singularrelativetimeunit
static Optional<URelativeDateTimeUnit> relativeTimeUnitType(StringView unit)
{
    StringView singular = singularUnit(unit);

    if (singular == "second")
        return UDAT_REL_UNIT_SECOND;
    if (singular == "minute")
        return UDAT_REL_UNIT_MINUTE;
    if (singular == "hour")
        return UDAT_REL_UNIT_HOUR;
    if (singular == "day")
        return UDAT_REL_UNIT_DAY;
    if (singular == "week")
        return UDAT_REL_UNIT_WEEK;
    if (singular == "month")
        return UDAT_REL_UNIT_MONTH;
    if (singular == "quarter")
        return UDAT_REL_UNIT_QUARTER;
    if (singular == "year")
        return UDAT_REL_UNIT_YEAR;

    return WTF::Nullopt;
}

String IntlRelativeTimeFormat::formatInternal(ExecState& exec, double value, StringView unit)
{
    ASSERT(m_relativeDateTimeFormatter);

    VM& vm = exec.vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    if (!std::isfinite(value)) {
        throwRangeError(&exec, vm, ASCIILiteral("number argument must be finite"));
        return String();
    }

    auto unitType = relativeTimeUnitType(unit);
    if (!unitType) {
        throwRangeError(&exec, vm, ASCIILiteral("unit argument is not a recognized unit type"));
        return String();
    }

    auto formatRelativeTime = m_numeric ? ureldatefmt_formatNumeric : ureldatefmt_format;

    UErrorCode status = U_ZERO_ERROR;
    Vector<UChar, 32> result(32);
    auto resultLength = formatRelativeTime(m_relativeDateTimeFormatter.get(), value, unitType.value(), result.data(), result.size(), &status);
    if (status == U_BUFFER_OVERFLOW_ERROR) {
        status = U_ZERO_ERROR;
        result.grow(resultLength);
        formatRelativeTime(m_relativeDateTimeFormatter.get(), value, unitType.value(), result.data(), resultLength, &status);
    }
    if (UNLIKELY(U_FAILURE(status))) {
        throwTypeError(&exec, vm, ASCIILiteral("failed to format relative time"));
        return String();
    }

    return String(result.data(), resultLength);
}

// https://tc39.es/ecma402/#sec-FormatRelativeTime
JSValue IntlRelativeTimeFormat::format(ExecState& exec, double value, StringView unit)
{
    VM& vm = exec.vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    String result = formatInternal(exec, value, unit);
    RETURN_IF_EXCEPTION(scope, jsUndefined());

    return jsString(vm, result);
}

// https://tc39.es/ecma402/#sec-FormatRelativeTimeToParts
JSValue IntlRelativeTimeFormat::formatToParts(ExecState& exec, double value, StringView unit)
{
    JSGlobalObject* globalObject = exec.lexicalGlobalObject();
    VM& vm = globalObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    String formattedRelativeTime = formatInternal(exec, value, unit);
    RETURN_IF_EXCEPTION(scope, jsUndefined());

    UErrorCode status = U_ZERO_ERROR;
    OwnPtr<UFieldPositionIterator> iterator = adoptPtr(ufieldpositer_open(&status));
    if (!U_SUCCESS(status)) {
		throwOutOfMemoryError(&exec);
		return jsUndefined();
	}

    double absValue = std::abs(value);

    Vector<UChar, 32> buffer(32);
    auto numberLength = unum_formatDoubleForFields(m_numberFormat.get(), absValue, buffer.data(), buffer.size(), iterator.get(), &status);
    if (needsToGrowToProduceBuffer(status)) {
        status = U_ZERO_ERROR;
        buffer.grow(numberLength);
        unum_formatDoubleForFields(m_numberFormat.get(), absValue, buffer.data(), numberLength, iterator.get(), &status);
    }
    if (U_FAILURE(status))
        return throwTypeError(&exec, vm, ASCIILiteral("failed to format relative time"));

    auto formattedNumber = String(buffer.data(), numberLength);

    JSArray* parts = JSArray::tryCreate(vm, &exec, globalObject->arrayStructureForIndexingTypeDuringAllocation(ArrayWithContiguous), 0);
    RETURN_IF_EXCEPTION(scope, jsUndefined());

    JSString* literalString = jsNontrivialString(vm, ASCIILiteral("literal"));

    // We only have one input number, so our relative time will have at most one numeric substring,
    // but we need to list all of the numeric parts separately.
    size_t numberEnd = 0;
    size_t numberStart = formattedRelativeTime.find(formattedNumber);
    if (numberStart != notFound) {
        numberEnd = numberStart + numberLength;

        // Add initial literal if there is one.
        if (numberStart) {
            JSObject* part = constructEmptyObject(globalObject);
            part->putDirect(vm, vm.propertyNames->type, literalString);
            part->putDirect(vm, vm.propertyNames->value, jsSubstring(vm, formattedRelativeTime, 0, numberStart));
            parts->push(&exec, part);
            RETURN_IF_EXCEPTION(scope, jsUndefined());
        }

        IntlNumberFormat::formatToPartsInternal(exec, absValue, formattedNumber, iterator.get(), parts, jsString(vm, singularUnit(unit).toString()));
        RETURN_IF_EXCEPTION(scope, jsUndefined());
    }

    // Add final literal if there is one.
    auto stringLength = formattedRelativeTime.length();
    if (numberEnd != stringLength) {
        JSObject* part = constructEmptyObject(globalObject);
        part->putDirect(vm, vm.propertyNames->type, literalString);
        part->putDirect(vm, vm.propertyNames->value, jsSubstring(vm, formattedRelativeTime, numberEnd, stringLength - numberEnd));
        parts->push(&exec, part);
        RETURN_IF_EXCEPTION(scope, jsUndefined());
    }

    return parts;
}

} // namespace JSC
