/*
 * Copyright (C) 2015 Yusuke Suzuki <utatane.tea@gmail.com>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "IteratorOperations.h"

#include "Error.h"
#include "JSCInlines.h"
#include "ObjectConstructor.h"

using namespace WTF;

namespace JSC {

JSValue iteratorNext(VM& vm, ExecState* exec, IterationRecord iterationRecord, JSValue argument)
{
    JSValue iterator = iterationRecord.iterator;
    JSValue nextFunction = iterationRecord.nextMethod;

    CallData nextFunctionCallData = getCallData(vm, nextFunction);
    if (nextFunctionCallData.type == CallData::Type::None)
        return throwTypeError(exec, vm);

    MarkedArgumentBuffer nextFunctionArguments;
    if (!argument.isEmpty())
        nextFunctionArguments.append(argument);
    JSValue result = call(vm, exec, nextFunction, nextFunctionCallData, iterator, nextFunctionArguments);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    if (!result.isObject())
        return throwTypeError(exec, vm, ASCIILiteral("Iterator result interface is not an object."));

    return result;
}

JSValue iteratorValue(VM& vm, ExecState* exec, JSValue iterResult)
{
    return iterResult.get(vm, exec, vm.propertyNames->value);
}

bool iteratorComplete(VM& vm, ExecState* exec, JSValue iterResult)
{
    JSValue done = iterResult.get(vm, exec, vm.propertyNames->done);
    RETURN_IF_EXCEPTION(vm, true);
    return done.toBoolean(exec);
}

JSValue iteratorStep(VM& vm, ExecState* exec, IterationRecord iterationRecord)
{
    JSValue result = iteratorNext(vm, exec, iterationRecord);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    bool done = iteratorComplete(vm, exec, result);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    if (done)
        return jsBoolean(false);
    return result;
}

void iteratorClose(VM& vm, ExecState* exec, IterationRecord iterationRecord)
{
    Exception* exception = nullptr;
    if (UNLIKELY(vm.exception())) {
        exception = vm.exception();
        vm.clearException();
    }

    JSValue returnFunction = iterationRecord.iterator.get(vm, exec, vm.propertyNames->returnKeyword);
    if (UNLIKELY(vm.exception()) || returnFunction.isUndefinedOrNull()) {
        if (UNLIKELY(exception))
            vm.throwException(exec, exception);
        return;
    }

    CallData returnFunctionCallData = getCallData(vm, returnFunction);
    if (returnFunctionCallData.type == CallData::Type::None) {
        if (exception)
            vm.throwException(exec, exception);
        else
            throwTypeError(exec, vm);
        return;
    }

    MarkedArgumentBuffer returnFunctionArguments;
    JSValue innerResult = call(vm, exec, returnFunction, returnFunctionCallData, iterationRecord.iterator, returnFunctionArguments);

    if (exception) {
        vm.throwException(exec, exception);
        return;
    }

    RETURN_IF_EXCEPTION(vm, void());

    if (!innerResult.isObject()) {
        throwTypeError(exec, vm, ASCIILiteral("Iterator result interface is not an object."));
        return;
    }
}

static const PropertyOffset valuePropertyOffset = 0;
static const PropertyOffset donePropertyOffset = 1;

Structure* createIteratorResultObjectStructure(VM& vm, JSGlobalObject& globalObject)
{
    Structure* iteratorResultStructure = globalObject.structureCache().emptyObjectStructureForPrototype(&globalObject, globalObject.objectPrototype(), JSFinalObject::defaultInlineCapacity());
    PropertyOffset offset;
    iteratorResultStructure = Structure::addPropertyTransition(vm, iteratorResultStructure, vm.propertyNames->value, 0, offset);
    RELEASE_ASSERT(offset == valuePropertyOffset);
    iteratorResultStructure = Structure::addPropertyTransition(vm, iteratorResultStructure, vm.propertyNames->done, 0, offset);
    RELEASE_ASSERT(offset == donePropertyOffset);
    return iteratorResultStructure;
}

JSObject* createIteratorResultObject(VM& vm, ExecState* exec, JSValue value, bool done)
{
    JSObject* resultObject = constructEmptyObject(vm, exec->lexicalGlobalObject()->iteratorResultObjectStructure());
    resultObject->putDirect(vm, valuePropertyOffset, value);
    resultObject->putDirect(vm, donePropertyOffset, jsBoolean(done));
    return resultObject;
}

JSValue iteratorMethod(VM& vm, ExecState& state, JSObject* object)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    CallData callData;
    JSValue method = object->getMethod(&state, callData, vm.propertyNames->iteratorSymbol, ASCIILiteral("Symbol.iterator property should be callable"));
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    return method;
}

IterationRecord iteratorForIterable(VM& vm, ExecState& state, JSObject* object, JSValue iteratorMethod)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    CallData iteratorMethodCallData = getCallData(vm, iteratorMethod);
    if (iteratorMethodCallData.type == CallData::Type::None) {
        throwTypeError(&state, vm);
        return IterationRecord();
    }

    ArgList iteratorMethodArguments;
    JSValue iterator = call(vm, &state, iteratorMethod, iteratorMethodCallData, object, iteratorMethodArguments);
    RETURN_IF_EXCEPTION(vm, IterationRecord());

    if (!iterator.isObject()) {
        throwTypeError(&state, vm);
        return IterationRecord();
    }

    JSValue nextMethod = iterator.getObject()->get(vm, &state, vm.propertyNames->next);
    RETURN_IF_EXCEPTION(vm, IterationRecord());

    return IterationRecord(iterator, nextMethod);
}

IterationRecord iteratorForIterable(VM& vm, ExecState& state, JSValue iterable)
{
    JSValue iteratorFunction = iterable.get(vm, &state, vm.propertyNames->iteratorSymbol);
    RETURN_IF_EXCEPTION(vm, IterationRecord());

    CallData iteratorFunctionCallData = getCallData(vm, iteratorFunction);
    if (iteratorFunctionCallData.type == CallData::Type::None) {
        throwTypeError(&state, vm);
        return IterationRecord();
    }

    ArgList iteratorFunctionArguments;
    JSValue iterator = call(vm, &state, iteratorFunction, iteratorFunctionCallData, iterable, iteratorFunctionArguments);
    RETURN_IF_EXCEPTION(vm, IterationRecord());

    if (!iterator.isObject()) {
        throwTypeError(&state, vm);
        return IterationRecord();
    }

    JSValue nextMethod = iterator.getObject()->get(vm, &state, vm.propertyNames->next);
    RETURN_IF_EXCEPTION(vm, IterationRecord());

    return IterationRecord(iterator, nextMethod);
}

} // namespace JSC
