/*
 * Copyright (C) 2015 Yusuke Suzuki <utatane.tea@gmail.com>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef IteratorOperations_h
#define IteratorOperations_h

#include "JSCJSValue.h"
#include "JSObject.h"

namespace JSC {

struct IterationRecord {
	IterationRecord() { }
	IterationRecord(JSValue iterator, JSValue nextMethod)
		: iterator(iterator)
		, nextMethod(nextMethod)
	{
	}
    JSValue iterator;
    JSValue nextMethod;
};

JSValue iteratorNext(VM& vm, ExecState*, IterationRecord, JSValue argument = JSValue());
JS_EXPORT_PRIVATE JSValue iteratorValue(VM& vm, ExecState*, JSValue iterResult);
bool iteratorComplete(VM& vm, ExecState*, JSValue iterResult);
JS_EXPORT_PRIVATE JSValue iteratorStep(VM& vm, ExecState*, IterationRecord);
JS_EXPORT_PRIVATE void iteratorClose(VM& vm, ExecState*, IterationRecord);
JS_EXPORT_PRIVATE JSObject* createIteratorResultObject(VM&, ExecState*, JSValue, bool done);

Structure* createIteratorResultObjectStructure(VM&, JSGlobalObject&);

JS_EXPORT_PRIVATE JSValue iteratorMethod(VM&, ExecState&, JSObject*);
JS_EXPORT_PRIVATE IterationRecord iteratorForIterable(VM&, ExecState&, JSObject*, JSValue iteratorMethod);
JS_EXPORT_PRIVATE IterationRecord iteratorForIterable(VM&, ExecState&, JSValue iterable);

JS_EXPORT_PRIVATE JSValue iteratorMethod(VM&, ExecState&, JSObject*);

template <typename CallBackType>
void forEachInIterable(VM& vm, ExecState& exec, JSValue iterable, CallBackType& callback)
{
    IterationRecord iterationRecord = iteratorForIterable(vm, exec, iterable);
    RETURN_IF_EXCEPTION(vm, void());
    while (true) {
        JSValue next = iteratorStep(vm, &exec, iterationRecord);
        if (UNLIKELY(vm.exception()) || next.isFalse())
            return;

        JSValue nextValue = iteratorValue(vm, &exec, next);
        RETURN_IF_EXCEPTION(vm, void());

        callback(vm, &exec, nextValue);
        if (UNLIKELY(vm.exception())) {
            iteratorClose(vm, &exec, iterationRecord);
            return;
        }
    }
}

template<typename CallBackType>
void forEachInIterable(VM& vm, ExecState& state, JSObject* iterable, JSValue iteratorMethod, CallBackType& callback)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    auto iterationRecord = iteratorForIterable(vm, state, iterable, iteratorMethod);
    RETURN_IF_EXCEPTION(vm, void());
    while (true) {
        JSValue next = iteratorStep(vm, &state, iterationRecord);
        if (UNLIKELY(vm.exception()) || next.isFalse())
            return;

        JSValue nextValue = iteratorValue(vm, &state, next);
        RETURN_IF_EXCEPTION(vm, void());

        callback(vm, state, nextValue);
        if (UNLIKELY(vm.exception())) {
            //scope.release();
            iteratorClose(vm, &state, iterationRecord);
            return;
        }
    }
}

}

#endif // !defined(IteratorOperations_h)
