/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *  Copyright (C) 2003 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2006 Alexey Proskuryakov (ap@nypop.com)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "JSArrayInlines.h"

#include "ArrayPrototype.h"
#include "ButterflyInlineMethods.h"
#include "CopiedSpace.h"
#include "CopiedSpaceInlineMethods.h"
#include "CachedCall.h"
#include "Error.h"
#include "Executable.h"
#include "GetterSetter.h"
#include "IndexingHeaderInlineMethods.h"
#include "PropertyNameArray.h"
#include "Reject.h"
#include <wtf/Assertions.h>
#include <wtf/OwnPtr.h>
#include <Operations.h>

using namespace std;
using namespace WTF;

namespace JSC {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSArray);

const ASCIILiteral LengthExceededTheMaximumArrayLengthError("Length exceeded the maximum array length");

const ClassInfo JSArray::s_info = {"Array", &JSNonFinalObject::s_info, nullptr, CREATE_METHOD_TABLE(JSArray)};

JSArray* JSArray::tryCreateUninitializedRestricted(ObjectInitializationScope& scope, DeferGC& deferGC, Structure* structure, unsigned initialLength)
{
    VM& vm = scope.vm();

	unsigned vectorLength = std::max(BASE_ARRAY_STORAGE_VECTOR_LEN, initialLength);
	if (UNLIKELY(vectorLength > MAX_STORAGE_VECTOR_LENGTH))
		return nullptr;

    JSGlobalObject* globalObject = structure->globalObject();
    bool createUninitialized = globalObject->isOriginalArrayStructure(structure);
    unsigned outOfLineStorage = structure->outOfLineCapacity();

	Butterfly* butterfly;
    IndexingType indexingType = structure->indexingType();
	if (LIKELY(!hasAnyArrayStorage(indexingType))) {
		ASSERT(
			hasUndecided(indexingType)
			|| hasInt32(indexingType)
			|| hasDouble(indexingType)
			|| hasContiguous(indexingType));

		void* temp;
		if (UNLIKELY(!vm.heap.tryAllocateStorage(Butterfly::totalSize(0, outOfLineStorage, true, vectorLength * sizeof(EncodedJSValue)), &temp)))
			return nullptr;
		butterfly = Butterfly::fromBase(temp, 0, outOfLineStorage);
		butterfly->setVectorLength(vectorLength);
		butterfly->setPublicLength(initialLength);
        unsigned i = (createUninitialized ? initialLength : 0);
        if (hasDouble(indexingType)) {
            for (; i < vectorLength; ++i)
                butterfly->contiguousDouble()[i] = PNaN;
        } else {
            for (; i < vectorLength; ++i)
                butterfly->contiguous()[i].clear();
        }
	} else {
        ASSERT(
            indexingType == ArrayWithSlowPutArrayStorage
            || indexingType == ArrayWithArrayStorage);
        static const unsigned indexBias = 0;
		void* temp;
		if (UNLIKELY(!vm.heap.tryAllocateStorage(Butterfly::totalSize(indexBias, outOfLineStorage, true, ArrayStorage::sizeFor(vectorLength)), &temp)))
			return nullptr;
		butterfly = Butterfly::fromBase(temp, indexBias, outOfLineStorage);
		*butterfly->indexingHeader() = indexingHeaderForArrayStorage(initialLength, vectorLength);
		ArrayStorage* storage = butterfly->arrayStorage();
		storage->m_indexBias = indexBias;
		storage->m_sparseMap.clear();
		storage->m_numValuesInVector = initialLength;
        unsigned i = (createUninitialized ? initialLength : 0);
        for (; i < vectorLength; ++i)
            storage->m_vector[i].clear();
	}

	JSArray* result = createWithButterfly(vm, deferGC, structure, butterfly);
    scope.notifyAllocated(result, createUninitialized);
	return result;
}

void JSArray::eagerlyInitializeButterfly(JSArray* array, unsigned initialLength)
{
    Structure* structure = array->structure();
    IndexingType indexingType = structure->indexingType();
    Butterfly* butterfly = array->butterfly();

    // This function only serves as a companion to tryCreateUninitializedRestricted()
    // in the event that we really can't defer initialization of the butterfly after all.
    // tryCreateUninitializedRestricted() already initialized the elements between
    // initialLength and vector length. We just need to do 0 - initialLength.
    // ObjectInitializationScope::notifyInitialized() will verify that all elements are
    // initialized.
    if (LIKELY(!hasAnyArrayStorage(indexingType))) {
        if (hasDouble(indexingType)) {
            for (unsigned i = 0; i < initialLength; ++i)
                butterfly->contiguousDouble()[i] = PNaN;
        } else {
            for (unsigned i = 0; i < initialLength; ++i)
                butterfly->contiguous()[i].clear();
        }
    } else {
        ArrayStorage* storage = butterfly->arrayStorage();
        for (unsigned i = 0; i < initialLength; ++i)
            storage->m_vector[i].clear();
    }
    //scope.notifyInitialized(array);
}

void JSArray::setLengthWritable(ExecState* exec, bool writable)
{
    ASSERT(isLengthWritable() || !writable);
    if (!isLengthWritable() || writable)
        return;

    enterDictionaryIndexingMode(exec->globalData());

    SparseArrayValueMap* map = arrayStorage()->m_sparseMap.get();
    ASSERT(map);
    map->setLengthIsReadOnly();
}

// https://tc39.es/ecma262/#sec-array-exotic-objects-defineownproperty-p-desc
bool JSArray::defineOwnProperty(JSObject* object, ExecState* exec, PropertyName propertyName, const PropertyDescriptor& descriptor, bool throwException)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSArray* array = jsCast<JSArray*>(object);

    // 2. If P is "length", then
    // https://tc39.es/ecma262/#sec-arraysetlength
    if (propertyName == vm.propertyNames->length) {
        // FIXME: Nothing prevents this from being called on a RuntimeArray, and the length function will always return 0 in that case.
        unsigned newLength = array->length();
        if (descriptor.value()) {
            newLength = descriptor.value().toUInt32(exec);
            RETURN_IF_EXCEPTION(scope, false);
            double valueAsNumber = descriptor.value().toNumber(exec);
            RETURN_IF_EXCEPTION(scope, false);
            if (valueAsNumber != static_cast<double>(newLength)) {
                throwRangeError(exec, vm, ASCIILiteral("Invalid array length"));
                return false;
            }
        }

        // OrdinaryDefineOwnProperty (https://tc39.es/ecma262/#sec-validateandapplypropertydescriptor) at steps 1.a, 11.a, and 15 is now performed:
        // 4. If current.[[Configurable]] is false, then
        // 4.a. If Desc.[[Configurable]] is present and its value is true, return false.
        if (descriptor.configurablePresent() && descriptor.configurable())
            return typeError(exec, vm, throwException, UnconfigurablePropertyChangeConfigurabilityError);
        // 4.b. If Desc.[[Enumerable]] is present and SameValue(Desc.[[Enumerable]], current.[[Enumerable]]) is false, return false.
        if (descriptor.enumerablePresent() && descriptor.enumerable())
            return typeError(exec, vm, throwException, UnconfigurablePropertyChangeEnumerabilityError);
        // 6. Else if SameValue(IsDataDescriptor(current), IsDataDescriptor(Desc)) is false, then
        // 6.a. If current.[[Configurable]] is false, return false.
        if (descriptor.isAccessorDescriptor())
            return typeError(exec, vm, throwException, UnconfigurablePropertyChangeAccessMechanismError);
        // 7. Else if IsDataDescriptor(current) and IsDataDescriptor(Desc) are both true, then
        // 7.a. If current.[[Configurable]] is false and current.[[Writable]] is false, then
        if (!array->isLengthWritable()) {
            // 7.a.i. If Desc.[[Writable]] is present and Desc.[[Writable]] is true, return false.
            // This check is unaffected by steps 13-14 of ArraySetLength as they change non-writable descriptors only.
            if (descriptor.writablePresent() && descriptor.writable())
                return typeError(exec, vm, throwException, UnconfigurablePropertyChangeWritabilityError);
            // 7.a.ii. If Desc.[[Value]] is present and SameValue(Desc.[[Value]], current.[[Value]]) is false, return false.
            // This check also covers step 12 of ArraySetLength, which is only reachable if newLen < oldLen.
            if (newLength != array->length())
                return typeError(exec, vm, throwException, ReadonlyPropertyChangeError);
        }

        // setLength() clears indices >= newLength and sets correct "length" value if [[Delete]] fails (step 17.b.i)
        bool success = true;
		if (newLength != array->length()) {
			success = array->setLength(exec, newLength, throwException);
			EXCEPTION_ASSERT(!scope.exception() || !success);
		}
        if (descriptor.writablePresent())
            array->setLengthWritable(exec, descriptor.writable());
        return success;
    }

    // 4. Else if P is an array index (15.4), then
    // a. Let index be ToUint32(P).
    if (auto optionalIndex = parseIndex(propertyName)) {
        // b. Reject if index >= oldLen and oldLenDesc.[[Writable]] is false.
    	uint32_t index = optionalIndex.value();
    	// FIXME: Nothing prevents this from being called on a RuntimeArray, and the length function will always return 0 in that case.
        if (index >= array->length() && !array->isLengthWritable())
            return reject(exec, vm, throwException, "Attempting to define numeric property on array with non-writable length property.");
        // c. Let succeeded be the result of calling the default [[DefineOwnProperty]] internal method (8.12.9) on A passing P, Desc, and false as arguments.
        // d. Reject if succeeded is false.
        // e. If index >= oldLen
        // e.i. Set oldLenDesc.[[Value]] to index + 1.
        // e.ii. Call the default [[DefineOwnProperty]] internal method (8.12.9) on A passing "length", oldLenDesc, and false as arguments. This call will always return true.
        // f. Return true.
        return array->defineOwnIndexedProperty(exec, index, descriptor, throwException);
    }

    return array->JSObject::defineOwnNonIndexProperty(exec, propertyName, descriptor, throwException);
}

bool JSArray::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
	VM& vm = exec->vm();
    JSArray* thisObject = jsCast<JSArray*>(object);
    if (propertyName == vm.propertyNames->length) {
    	unsigned attributes = thisObject->isLengthWritable() ? DontDelete | DontEnum : DontDelete | DontEnum | ReadOnly;
		slot.setValue(thisObject, attributes, jsNumber(thisObject->length()));
        return true;
    }

    return JSObject::getOwnPropertySlot(thisObject, exec, propertyName, slot);
}

// https://tc39.es/ecma262/#sec-array-exotic-objects-defineownproperty-p-desc
bool JSArray::put(JSCell* cell, ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    JSArray* thisObject = jsCast<JSArray*>(cell);

    if (propertyName == vm.propertyNames->length) {
        if (!thisObject->isLengthWritable()) {
            if (slot.isStrictMode())
                throwTypeError(exec, vm, ASCIILiteral("Array length is not writable"));
            return false;
        }

        if (UNLIKELY(slot.thisValue() != thisObject))
            RELEASE_AND_RETURN(scope, JSObject::definePropertyOnReceiver(exec, propertyName, value, slot));

        unsigned newLength = value.toUInt32(exec);
        RETURN_IF_EXCEPTION(vm, false);
        double valueAsNumber = value.toNumber(exec);
        RETURN_IF_EXCEPTION(vm, false);
        if (valueAsNumber != static_cast<double>(newLength)) {
        	vm.throwException(exec, createRangeError(exec, ASCIILiteral("Invalid array length")));
            return false;
        }
        return thisObject->setLength(exec, newLength, slot.isStrictMode());
    }

    return JSObject::put(thisObject, exec, propertyName, value, slot);
}

bool JSArray::deleteProperty(JSCell* cell, ExecState* exec, PropertyName propertyName)
{
	VM& vm = exec->vm();
    JSArray* thisObject = jsCast<JSArray*>(cell);
    
    if (propertyName == vm.propertyNames->length)
        return false;

    return JSObject::deleteProperty(thisObject, exec, propertyName);
}

static int compareKeysForQSort(const void* a, const void* b)
{
    unsigned da = *static_cast<const unsigned*>(a);
    unsigned db = *static_cast<const unsigned*>(b);
    return (da > db) - (da < db);
}

void JSArray::getOwnSpecialPropertyNames(JSObject* object, ExecState* exec, PropertyNameArray& propertyNames, DontEnumPropertiesMode mode)
{
	VM& vm = exec->vm();
    if (mode == DontEnumPropertiesMode::Include)
        propertyNames.add(vm.propertyNames->length);
}

// This method makes room in the vector, but leaves the new space for count slots uncleared.
bool JSArray::unshiftCountSlowCase(JSGlobalData& vm, bool addToFront, unsigned count)
{
    ArrayStorage* storage = ensureArrayStorage(vm);
    Butterfly* butterfly = storage->butterfly();
    unsigned propertyCapacity = structure()->outOfLineCapacity();
    unsigned propertySize = structure()->outOfLineSize();

    // If not, we should have handled this on the fast path.
    ASSERT(!addToFront || count > storage->m_indexBias);

    // Step 1:
    // Gather 4 key metrics:
    //  * usedVectorLength - how many entries are currently in the vector (conservative estimate - fewer may be in use in sparse vectors).
    //  * requiredVectorLength - how many entries are will there be in the vector, after allocating space for 'count' more.
    //  * currentCapacity - what is the current size of the vector, including any pre-capacity.
    //  * desiredCapacity - how large should we like to grow the vector to - based on 2x requiredVectorLength.

    unsigned length = storage->length();
    unsigned usedVectorLength = min(storage->vectorLength(), length);
    ASSERT(usedVectorLength <= MAX_STORAGE_VECTOR_LENGTH);
    // Check that required vector length is possible, in an overflow-safe fashion.
    if (count > MAX_STORAGE_VECTOR_LENGTH - usedVectorLength)
        return false;
    unsigned requiredVectorLength = usedVectorLength + count;
    ASSERT(requiredVectorLength <= MAX_STORAGE_VECTOR_LENGTH);
    // The sum of m_vectorLength and m_indexBias will never exceed MAX_STORAGE_VECTOR_LENGTH.
    ASSERT(storage->vectorLength() <= MAX_STORAGE_VECTOR_LENGTH && (MAX_STORAGE_VECTOR_LENGTH - storage->vectorLength()) >= storage->m_indexBias);
	unsigned currentCapacity = storage->vectorLength() + storage->m_indexBias;
	// The calculation of desiredCapacity won't overflow, due to the range of MAX_STORAGE_VECTOR_LENGTH.
    unsigned desiredCapacity = min(MAX_STORAGE_VECTOR_LENGTH, max(BASE_ARRAY_STORAGE_VECTOR_LEN, requiredVectorLength) << 1);

    // Step 2:
    // We're either going to choose to allocate a new ArrayStorage, or we're going to reuse the existing one.

    DeferGC deferGC(vm.heap);
    void* newAllocBase = 0;
    unsigned newStorageCapacity;
    bool allocatedNewStorage;
    // If the current storage array is sufficiently large (but not too large!) then just keep using it.
    if (currentCapacity > desiredCapacity && isDenseEnoughForVector(currentCapacity, requiredVectorLength)) {
    	newAllocBase = butterfly->base(structure());
		newStorageCapacity = currentCapacity;
        allocatedNewStorage = false;
	} else {
        const unsigned preCapacity = 0;
        Butterfly* newButterfly = Butterfly::tryCreateUninitialized(vm, this, preCapacity, propertyCapacity, true, ArrayStorage::sizeFor(desiredCapacity));
        if (!newButterfly)
            return false;
        newAllocBase = newButterfly->base(preCapacity, propertyCapacity);
        newStorageCapacity = desiredCapacity;
        allocatedNewStorage = true;
    }

    // Step 3:
    // Work out where we're going to move things to.

    // Determine how much of the vector to use as pre-capacity, and how much as post-capacity.
    // If we're adding to the end, we'll add all the new space to the end.
    // If the vector had no free post-capacity (length >= m_vectorLength), don't give it any.
    // If it did, we calculate the amount that will remain based on an atomic decay - leave the
    // vector with half the post-capacity it had previously.
    unsigned postCapacity = 0;
    if (!addToFront)
		postCapacity = newStorageCapacity - requiredVectorLength;
	else if (length < storage->vectorLength()) {
		// Atomic decay, + the post-capacity cannot be greater than what is available.
		postCapacity = min((storage->vectorLength() - length) >> 1, newStorageCapacity - requiredVectorLength);
		// If we're moving contents within the same allocation, the post-capacity is being reduced.
		ASSERT(newAllocBase != butterfly->base(structure()) || postCapacity < storage->vectorLength() - length);
	}

	unsigned newVectorLength = requiredVectorLength + postCapacity;
    RELEASE_ASSERT(newVectorLength <= MAX_STORAGE_VECTOR_LENGTH);
	unsigned preCapacity = newStorageCapacity - newVectorLength;

	Butterfly* newButterfly = Butterfly::fromBase(newAllocBase, preCapacity, propertyCapacity);

    if (addToFront) {
        ASSERT(count + usedVectorLength <= newVectorLength);
        memmove(newButterfly->arrayStorage()->m_vector + count, storage->m_vector, sizeof(JSValue) * usedVectorLength);
        memmove(newButterfly->propertyStorage() - propertySize, butterfly->propertyStorage() - propertySize, sizeof(JSValue) * propertySize + sizeof(IndexingHeader) + ArrayStorage::sizeFor(0));

        // We don't need to zero the pre-capacity for the concurrent GC because it is not available to use as property storage.
        memset(newButterfly->base(0, propertyCapacity), 0, (propertyCapacity - propertySize) * sizeof(JSValue));

        if (allocatedNewStorage) {
            // We will set the vectorLength to newVectorLength. We populated requiredVectorLength
            // (usedVectorLength + count), which is less. Clear the difference.
            for (unsigned i = requiredVectorLength; i < newVectorLength; ++i)
                newButterfly->arrayStorage()->m_vector[i].clear();
        }
    } else if ((newAllocBase != butterfly->base(structure())) || (preCapacity != storage->m_indexBias)) {
        memmove(newButterfly->propertyStorage() - propertyCapacity, butterfly->propertyStorage() - propertyCapacity, sizeof(JSValue) * propertyCapacity + sizeof(IndexingHeader) + ArrayStorage::sizeFor(0));
        memmove(newButterfly->arrayStorage()->m_vector, storage->m_vector, sizeof(JSValue) * usedVectorLength);

        for (unsigned i = requiredVectorLength; i < newVectorLength; i++)
        	newButterfly->arrayStorage()->m_vector[i].clear();
    }

	newButterfly->arrayStorage()->setVectorLength(newVectorLength);
	newButterfly->arrayStorage()->m_indexBias = preCapacity;

	setButterfly(vm, newButterfly);

	return true;
}

bool JSArray::setLengthWithArrayStorage(ExecState* exec, unsigned newLength, bool throwException, ArrayStorage* storage)
{
	VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    unsigned length = storage->length();

    // If the length is read only then we enter sparse mode, so should enter the following 'if'.
    ASSERT(isLengthWritable() || storage->m_sparseMap);

    if (SparseArrayValueMap* map = storage->m_sparseMap.get()) {
        // Fail if the length is not writable.
        if (map->lengthIsReadOnly())
            return reject(exec, vm, throwException, ReadonlyPropertyWriteError);

        if (newLength < length) {
            // Copy any keys we might be interested in into a vector.
            Vector<unsigned> keys;
            keys.reserveInitialCapacity(min(map->size(), static_cast<size_t>(length - newLength)));
            SparseArrayValueMap::const_iterator end = map->end();
            for (SparseArrayValueMap::const_iterator it = map->begin(); it != end; ++it) {
                unsigned index = static_cast<unsigned>(it->first);
                if (index < length && index >= newLength)
                    keys.append(index);
            }

            // Check if the array is in sparse mode. If so there may be non-configurable
            // properties, so we have to perform deletion with caution, if not we can
            // delete values in any order.
            if (map->sparseMode()) {
                qsort(keys.begin(), keys.size(), sizeof(unsigned), compareKeysForQSort);
                unsigned i = keys.size();
                while (i) {
                    unsigned index = keys[--i];
                    SparseArrayValueMap::iterator it = map->find(index);
                    ASSERT(it != map->notFound());
                    if (it->second.attributes & DontDelete) {
                        storage->setLength(index + 1);
                        return reject(exec, vm, throwException, "Unable to delete property.");
                    }
                    map->remove(it);
                }
            } else {
                for (unsigned i = 0; i < keys.size(); ++i)
                    map->remove(keys[i]);
                if (map->isEmpty())
                    deallocateSparseIndexMap();
            }
        }
    }

    if (newLength < length) {
        // Delete properties from the vector.
        unsigned usedVectorLength = min(length, storage->vectorLength());
        for (unsigned i = newLength; i < usedVectorLength; ++i) {
            WriteBarrier<Unknown>& valueSlot = storage->m_vector[i];
            bool hadValue = !!valueSlot;
            valueSlot.clear();
            storage->m_numValuesInVector -= hadValue;
        }
    }

    storage->setLength(newLength);

    return true;
}

bool JSArray::appendMemcpy(ExecState* exec, VM& vm, unsigned startIndex, JSC::JSArray* otherArray)
{
    if (!canFastCopy(vm, otherArray))
        return false;

    IndexingType type = structure()->indexingType();
    IndexingType otherType = otherArray->structure()->indexingType();
    IndexingType copyType = mergeIndexingTypeForCopying(otherType);
    if (type == ArrayWithUndecided && copyType != NonArray) {
        if (copyType == ArrayWithInt32)
            convertUndecidedToInt32(vm);
        else if (copyType == ArrayWithDouble)
            convertUndecidedToDouble(vm);
        else if (copyType == ArrayWithContiguous)
            convertUndecidedToContiguous(vm);
        else {
            ASSERT(copyType == ArrayWithUndecided);
            return true;
        }
    } else if (type != copyType)
        return false;

    unsigned otherLength = otherArray->length();
    Checked<unsigned, RecordOverflow> checkedNewLength = startIndex;
    checkedNewLength += otherLength;

    unsigned newLength;
    if (checkedNewLength.safeGet(newLength) == CheckedState::DidOverflow) {
        vm.throwException(exec, createRangeError(exec, LengthExceededTheMaximumArrayLengthError));
        return false;
    }

    if (newLength >= MIN_SPARSE_ARRAY_INDEX)
        return false;

    if (!ensureLength(vm, newLength)) {
        throwOutOfMemoryError(exec);
        return false;
    }
    ASSERT(copyType == indexingType());

    if (UNLIKELY(otherType == ArrayWithUndecided)) {
        auto* butterfly = this->butterfly();
        if (type == ArrayWithDouble) {
            for (unsigned i = startIndex; i < newLength; ++i)
                butterfly->contiguousDouble()[i] = PNaN;
        } else {
            for (unsigned i = startIndex; i < newLength; ++i)
                butterfly->contiguousInt32()[i].setWithoutWriteBarrier(JSValue());
        }
    } else if (type == ArrayWithDouble)
        memcpy(butterfly()->contiguousDouble().data() + startIndex, otherArray->butterfly()->contiguousDouble().data(), sizeof(JSValue) * otherLength);
    else
        memcpy(butterfly()->contiguous().data() + startIndex, otherArray->butterfly()->contiguous().data(), sizeof(JSValue) * otherLength);

    return true;
}

bool JSArray::setLength(ExecState* exec, unsigned newLength, bool throwException)
{
    switch (structure()->indexingType()) {
    case ArrayClass:
        if (!newLength)
            return true;
        if (newLength >= MIN_SPARSE_ARRAY_INDEX) {
            return setLengthWithArrayStorage(
                exec, newLength, throwException,
				ensureArrayStorage(exec->globalData()));
        }
        createInitialUndecided(exec->globalData(), newLength);
        return true;

    case ArrayWithUndecided:
    case ArrayWithInt32:
    case ArrayWithDouble:
    case ArrayWithContiguous: {
        if (newLength == m_butterfly->publicLength())
            return true;
        if (newLength > MAX_STORAGE_VECTOR_LENGTH // This check ensures that we can do fast push.
            || (newLength >= MIN_SPARSE_ARRAY_INDEX
                && !isDenseEnoughForVector(newLength, countElements()))) {
            return setLengthWithArrayStorage(
                exec, newLength, throwException,
                ensureArrayStorage(exec->globalData()));
        }
        if (newLength > m_butterfly->publicLength()) {
            if(!ensureLength(exec->globalData(), newLength)) {
                throwOutOfMemoryError(exec);
                return false;
            }
            return true;
        }

        unsigned lengthToClear = m_butterfly->publicLength() - newLength;
        unsigned costToAllocateNewButterfly = 64; // a heuristic.
        if (lengthToClear > newLength && lengthToClear > costToAllocateNewButterfly) {
            reallocateAndShrinkButterfly(exec->vm(), newLength);
            return true;
        }

        if (structure()->indexingType() == ArrayWithDouble) {
            for (unsigned i = m_butterfly->publicLength(); i-- > newLength;)
                m_butterfly->contiguousDouble()[i] = PNaN;
        } else {
            for (unsigned i = m_butterfly->publicLength(); i-- > newLength;)
                m_butterfly->contiguous()[i].clear();
        }
        m_butterfly->setPublicLength(newLength);
        return true;
    }

    case ArrayWithArrayStorage:
    case ArrayWithSlowPutArrayStorage:
        return setLengthWithArrayStorage(exec, newLength, throwException, arrayStorage());

    default:
        RELEASE_ASSERT_NOT_REACHED();
        return false;
    }
}

JSValue JSArray::pop(ExecState* exec)
{
    VM& vm = this->vm();

    switch (structure()->indexingType()) {
    case ArrayClass:
        return jsUndefined();

    case ArrayWithUndecided:
        if (!m_butterfly->publicLength())
            return jsUndefined();
        // We have nothing but holes. So, drop down to the slow version.
        break;

    case ArrayWithInt32:
    case ArrayWithContiguous: {
        unsigned length = m_butterfly->publicLength();

        if (!length--)
            return jsUndefined();

        ASSERT(length < m_butterfly->vectorLength());
        JSValue value = m_butterfly->contiguous()[length].get();
        if (value) {
            m_butterfly->contiguous()[length].clear();
            m_butterfly->setPublicLength(length);
            return value;
        }
        break;
    }

    case ArrayWithDouble: {
        unsigned length = m_butterfly->publicLength();

        if (!length--)
            return jsUndefined();

        ASSERT(length < m_butterfly->vectorLength());
        double value = m_butterfly->contiguousDouble()[length];
        if (value == value) {
            m_butterfly->contiguousDouble()[length] = PNaN;
            m_butterfly->setPublicLength(length);
            return JSValue(JSValue::EncodeAsDouble, value);
        }
        break;
    }

    case ARRAY_WITH_ARRAY_STORAGE_INDEXING_TYPES: {
        ArrayStorage* storage = m_butterfly->arrayStorage();

        unsigned length = storage->length();
		if (!length) {
			if (!isLengthWritable())
				throwTypeError(exec, vm, ReadonlyPropertyWriteError);
			return jsUndefined();
		}

		unsigned index = length - 1;
        if (index < storage->vectorLength()) {
			WriteBarrier<Unknown>& valueSlot = storage->m_vector[index];
			if (valueSlot) {
				--storage->m_numValuesInVector;
				JSValue element = valueSlot.get();
				valueSlot.clear();

				ASSERT(isLengthWritable());
                storage->setLength(index);
				return element;
			}
		}
        break;
    }

    default:
        RELEASE_ASSERT_NOT_REACHED();
        return JSValue();
    }

    unsigned index = getArrayLength() - 1;
	// Let element be the result of calling the [[Get]] internal method of O with argument indx.
	JSValue element = get(vm, exec, index);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
	// Call the [[Delete]] internal method of O with arguments indx and true.
	if (!deletePropertyByIndex(this, exec, index)) {
		throwTypeError(exec, vm, "Unable to delete property.");
		return jsUndefined();
	}
	// Call the [[Put]] internal method of O with arguments "length", indx, and true.
	setLength(exec, index, true);
	// Return element.
	return element;
}

// Push & putIndex are almost identical, with two small differences.
//  - we always are writing beyond the current array bounds, so it is always necessary to update m_length & m_numValuesInVector.
//  - pushing to an array of length 2^32-1 stores the property, but throws a range error.
NEVER_INLINE void JSArray::push(ExecState* exec, JSValue value)
{
    VM& vm = exec->vm();
    pushInline(vm, exec, value);
}

JSArray* JSArray::fastSlice(ExecState& exec, unsigned startIndex, unsigned count)
{
	IndexingType arrayType = structure()->indexingType();
    switch (arrayType) {
    case ArrayWithDouble:
    case ArrayWithInt32:
    case ArrayWithContiguous: {
        VM& vm = exec.vm();
        if (count >= MIN_SPARSE_ARRAY_INDEX || structure()->holesMustForwardToPrototype(vm))
            return nullptr;

        JSGlobalObject* lexicalGlobalObject = exec.lexicalGlobalObject();
        Structure* resultStructure = lexicalGlobalObject->arrayStructureForIndexingTypeDuringAllocation(arrayType);
        if (UNLIKELY(hasAnyArrayStorage(resultStructure->indexingType())))
            return nullptr;

        DeferGC deferGC(vm.heap);

        ASSERT(!lexicalGlobalObject->isHavingABadTime());
        ObjectInitializationScope scope(vm);
        JSArray* resultArray = JSArray::tryCreateUninitializedRestricted(scope, deferGC, resultStructure, count);
        if (UNLIKELY(!resultArray))
        	return nullptr;

        Butterfly* resultButterfly = resultArray->butterfly();
        if (arrayType == ArrayWithDouble)
            memcpy(resultButterfly->contiguousDouble().data(), m_butterfly->contiguousDouble().data() + startIndex, sizeof(JSValue) * count);
        else
            memcpy(resultButterfly->contiguous().data(), m_butterfly->contiguous().data() + startIndex, sizeof(JSValue) * count);
        resultButterfly->setPublicLength(count);

        return resultArray;
    }
    default:
        return nullptr;
    }
}

bool JSArray::shiftCountWithArrayStorage(JSGlobalData& vm, unsigned startIndex, unsigned count, ArrayStorage* storage)
{
    unsigned oldLength = storage->length();
    ASSERT(count <= oldLength);
    
    // If the array contains holes or is otherwise in an abnormal state,
    // use the generic algorithm in ArrayPrototype.
    if (storage->hasHoles()
		|| hasSparseMap()
		|| shouldUseSlowPut(this->structure()->indexingType())) {
		return false;
	}

    if (!oldLength)
        return true;

    unsigned length = oldLength - count;
    
    storage->m_numValuesInVector -= count;
    storage->setLength(length);
    
    unsigned vectorLength = storage->vectorLength();
    if (!vectorLength)
        return true;

    if (startIndex >= vectorLength)
        return true;

    DeferGC deferGC(vm.heap);
    //auto locker = holdLock(*this);

    if (startIndex + count > vectorLength)
        count = vectorLength - startIndex;

    unsigned usedVectorLength = min(vectorLength, oldLength);

    unsigned numElementsBeforeShiftRegion = startIndex;
    unsigned firstIndexAfterShiftRegion = startIndex + count;
    unsigned numElementsAfterShiftRegion = usedVectorLength - firstIndexAfterShiftRegion;
    ASSERT(numElementsBeforeShiftRegion + count + numElementsAfterShiftRegion == usedVectorLength);

    // The point of this comparison seems to be to minimize the amount of elements that have to
    // be moved during a shift operation.
    if (numElementsBeforeShiftRegion < numElementsAfterShiftRegion) {
        // The number of elements before the shift region is less than the number of elements
        // after the shift region, so we move the elements before to the right.
        if (numElementsBeforeShiftRegion) {
            ASSERT(count + startIndex <= vectorLength);
			memmove(
				storage->m_vector + count,
				storage->m_vector,
				sizeof(JSValue) * startIndex);
        }
        // Adjust the Butterfly and the index bias. We only need to do this here because we're changing
        // the start of the Butterfly, which needs to point at the first indexed property in the used
        // portion of the vector.
        Butterfly* butterfly = m_butterfly->shift(structure(), count);
        setButterfly(vm, butterfly);
        storage = butterfly->arrayStorage();
        storage->m_indexBias += count;

        // Since we're consuming part of the vector by moving its beginning to the left,
        // we need to modify the vector length appropriately.
        storage->setVectorLength(vectorLength - count);
    } else {
        // The number of elements before the shift region is greater than or equal to the number
        // of elements after the shift region, so we move the elements after the shift region to the left.
		memmove(
			storage->m_vector + startIndex,
			storage->m_vector + firstIndexAfterShiftRegion,
			sizeof(JSValue) * numElementsAfterShiftRegion);

        // Clear the slots of the elements we just moved.
        unsigned startOfEmptyVectorTail = usedVectorLength - count;
        for (unsigned i = startOfEmptyVectorTail; i < usedVectorLength; ++i)
            storage->m_vector[i].clear();
        // We don't modify the index bias or the Butterfly pointer in this case because we're not changing
        // the start of the Butterfly, which needs to point at the first indexed property in the used
        // portion of the vector. We also don't modify the vector length because we're not actually changing
        // its length; we're just using less of it.
    }
    return true;
}

bool JSArray::shiftCountWithAnyIndexingType(ExecState* exec, unsigned& startIndex, unsigned count, unsigned shiftThreshold)
{
    VM& vm = exec->vm();
    ASSERT(count > 0);

    switch (structure()->indexingType()) {
    case ArrayClass:
        return true;

    case ArrayWithUndecided:
        // Don't handle this because it's confusing and it shouldn't come up.
        return false;

    case ArrayWithInt32:
    case ArrayWithContiguous: {
        unsigned oldLength = m_butterfly->publicLength();
        ASSERT(count <= oldLength);

        // We may have to walk the entire array to do the shift. We're willing to do
        // so only if it's not horribly slow.
        if (oldLength - (startIndex + count) >= MIN_SPARSE_ARRAY_INDEX || oldLength > shiftThreshold)
            return shiftCountWithArrayStorage(vm, startIndex, count, ensureArrayStorage(vm));

        // Storing to a hole is fine since we're still having a good time. But reading from a hole
        // is totally not fine, since we might have to read from the proto chain.
        // We have to check for holes before we start moving things around so that we don't get halfway
        // through shifting and then realize we should have been in ArrayStorage mode.
        unsigned end = oldLength - count;
        unsigned moveCount = end - startIndex;
		if (moveCount) {
		   if (UNLIKELY(this->structure()->holesMustForwardToPrototype(vm))) {
			   for (unsigned i = startIndex; i < end; ++i) {
				   JSValue v = m_butterfly->contiguous().at(i + count).get();
				   if (UNLIKELY(!v)) {
					   startIndex = i;
					   return shiftCountWithArrayStorage(vm, startIndex, count, ensureArrayStorage(vm));
				   }
				   m_butterfly->contiguous().at(i).setWithoutWriteBarrier(v);
			   }
			} else {
				memmove(m_butterfly->contiguous().data() + startIndex,
					m_butterfly->contiguous().data() + startIndex + count,
					sizeof(JSValue) * moveCount);
			}
		}

        for (unsigned i = end; i < oldLength; ++i)
            m_butterfly->contiguous()[i].clear();

        m_butterfly->setPublicLength(oldLength - count);
        return true;
    }

    case ArrayWithDouble: {
        unsigned oldLength = m_butterfly->publicLength();
        ASSERT(count <= oldLength);

        // We may have to walk the entire array to do the shift. We're willing to do
        // so only if it's not horribly slow.
        if (oldLength - (startIndex + count) >= MIN_SPARSE_ARRAY_INDEX || oldLength > shiftThreshold)
            return shiftCountWithArrayStorage(vm, startIndex, count, ensureArrayStorage(vm));

        // Storing to a hole is fine since we're still having a good time. But reading from a hole
        // is totally not fine, since we might have to read from the proto chain.
        // We have to check for holes before we start moving things around so that we don't get halfway
        // through shifting and then realize we should have been in ArrayStorage mode.
        unsigned end = oldLength - count;
        unsigned moveCount = end - startIndex;
        if (moveCount) {
            if (UNLIKELY(this->structure()->holesMustForwardToPrototype(vm))) {
                for (unsigned i = startIndex; i < end; ++i) {
                    double v = m_butterfly->contiguousDouble().at(i + count);
                    if (UNLIKELY(v != v)) {
                        startIndex = i;
                        return shiftCountWithArrayStorage(vm, startIndex, count, ensureArrayStorage(vm));
                    }
                    m_butterfly->contiguousDouble().at(i) = v;
                }
			} else {
				memmove(m_butterfly->contiguousDouble().data() + startIndex,
					m_butterfly->contiguousDouble().data() + startIndex + count,
					sizeof(JSValue) * moveCount);
			}
        }
        for (unsigned i = end; i < oldLength; ++i)
            m_butterfly->contiguousDouble()[i] = PNaN;

        m_butterfly->setPublicLength(oldLength - count);
        return true;
    }

    case ArrayWithArrayStorage:
    case ArrayWithSlowPutArrayStorage:
        return shiftCountWithArrayStorage(vm, startIndex, count, arrayStorage());

    default:
        RELEASE_ASSERT_NOT_REACHED();
        return false;
    }
}

// Returns true if the unshift can be handled, false to fallback.    
bool JSArray::unshiftCountWithArrayStorage(ExecState* exec, unsigned startIndex, unsigned count, ArrayStorage* storage)
{
    unsigned length = storage->length();

    ASSERT(startIndex <= length);

    // If the array contains holes or is otherwise in an abnormal state,
    // use the generic algorithm in ArrayPrototype.
    if (storage->hasHoles()  || storage->inSparseMode() || shouldUseSlowPut(structure()->indexingType()))
    	return false;

    VM& vm = exec->vm();

    bool moveFront = !startIndex || startIndex < length / 2;

    unsigned vectorLength = storage->vectorLength();

    // Need to have GC deferred around the unshiftCountSlowCase(), since that leaves the butterfly in
    // a weird state: some parts of it will be left uninitialized, which we will fill in here.
    DeferGC deferGC(vm.heap);

    if (moveFront && storage->m_indexBias >= count) {
    	setButterfly(vm, storage->butterfly()->unshift(structure(), count));
        storage = m_butterfly->arrayStorage();
        storage->m_indexBias -= count;
        storage->setVectorLength(vectorLength + count);
    } else if (!moveFront && vectorLength - length >= count)
        storage = storage->butterfly()->arrayStorage();
    else if (unshiftCountSlowCase(vm, moveFront, count))
        storage = arrayStorage();
    else {
        throwOutOfMemoryError(exec);
        return true;
    }

    WriteBarrier<Unknown>* vector = storage->m_vector;

    if (startIndex) {
        if (moveFront)
            memmove(vector, vector + count, startIndex * sizeof(JSValue));
        else if (length - startIndex)
            memmove(vector + startIndex + count, vector + startIndex, (length - startIndex) * sizeof(JSValue));
    }

    for (unsigned i = 0; i < count; i++)
        vector[i + startIndex].clear();
    return true;
}

bool JSArray::unshiftCountWithAnyIndexingType(ExecState* exec, unsigned startIndex, unsigned count)
{
    VM& vm = exec->vm();

    switch (structure()->indexingType()) {
    case ArrayClass:
    case ArrayWithUndecided:
        // We could handle this. But it shouldn't ever come up, so we won't.
        return false;

    case ArrayWithInt32:
    case ArrayWithContiguous: {
        unsigned oldLength = m_butterfly->publicLength();

        // We may have to walk the entire array to do the unshift. We're willing to do so
        // only if it's not horribly slow.
        unsigned moveCount = oldLength - startIndex;
		if (moveCount >= MIN_SPARSE_ARRAY_INDEX)
            return unshiftCountWithArrayStorage(exec, startIndex, count, ensureArrayStorage(vm));

        Checked<unsigned, RecordOverflow> checkedLength(oldLength);
        checkedLength += count;
        unsigned newLength;
        if (CheckedState::DidOverflow == checkedLength.safeGet(newLength)) {
            throwOutOfMemoryError(exec);
            return true;
        }
        if (newLength > MAX_STORAGE_VECTOR_LENGTH)
            return false;

        // FIXME: If we create a new butterfly, we should move elements at the same time.
        if (!ensureLength(vm, newLength)) {
            throwOutOfMemoryError(exec);
            return true;
        }

        // We have to check for holes before we start moving things around so that we don't get halfway
        // through shifting and then realize we should have been in ArrayStorage mode.
        if (moveCount) {
            if (UNLIKELY(this->structure()->holesMustForwardToPrototype(vm))) {
				for (unsigned i = oldLength; i-- > startIndex;) {
					JSValue v = m_butterfly->contiguous()[i].get();
					if (UNLIKELY(!v))
						return unshiftCountWithArrayStorage(exec, startIndex, count, ensureArrayStorage(vm));
				}
            }

            memmove(m_butterfly->contiguous().data() + startIndex + count, m_butterfly->contiguous().data() + startIndex, moveCount * sizeof(EncodedJSValue));
        }

        // NOTE: we're leaving being garbage in the part of the array that we shifted out
        // of. This is fine because the caller is required to store over that area, and
        // in contiguous mode storing into a hole is guaranteed to behave exactly the same
        // as storing over an existing element.

        return true;
    }

    case ArrayWithDouble: {
        unsigned oldLength = m_butterfly->publicLength();

        // We may have to walk the entire array to do the unshift. We're willing to do so
        // only if it's not horribly slow.
        unsigned moveCount = oldLength - startIndex;
		if (moveCount >= MIN_SPARSE_ARRAY_INDEX)
            return unshiftCountWithArrayStorage(exec, startIndex, count, ensureArrayStorage(vm));

        Checked<unsigned, RecordOverflow> checkedLength(oldLength);
        checkedLength += count;
        unsigned newLength;
        if (CheckedState::DidOverflow == checkedLength.safeGet(newLength)) {
            throwOutOfMemoryError(exec);
            return true;
        }
        if (newLength > MAX_STORAGE_VECTOR_LENGTH)
            return false;

        // FIXME: If we create a new butterfly, we should move elements at the same time.
        if (!ensureLength(vm, newLength)) {
            throwOutOfMemoryError(exec);
            return true;
        }

        // We have to check for holes before we start moving things around so that we don't get halfway
        // through shifting and then realize we should have been in ArrayStorage mode.
        if (moveCount) {
			if (UNLIKELY(this->structure()->holesMustForwardToPrototype(vm))) {
				for (unsigned i = oldLength; i-- > startIndex;) {
					double v = m_butterfly->contiguousDouble()[i];
					if (UNLIKELY(v != v))
						return unshiftCountWithArrayStorage(exec, startIndex, count, ensureArrayStorage(vm));
				}
			}

			memmove(m_butterfly->contiguousDouble().data() + startIndex + count, m_butterfly->contiguousDouble().data() + startIndex, moveCount * sizeof(double));
		}

        // NOTE: we're leaving being garbage in the part of the array that we shifted out
        // of. This is fine because the caller is required to store over that area, and
        // in contiguous mode storing into a hole is guaranteed to behave exactly the same
        // as storing over an existing element.

        return true;
    }

    case ArrayWithArrayStorage:
    case ArrayWithSlowPutArrayStorage:
        return unshiftCountWithArrayStorage(exec, startIndex, count, arrayStorage());

    default:
        RELEASE_ASSERT_NOT_REACHED();
        return false;
    }
}

void JSArray::fillArgList(ExecState* exec, MarkedArgumentBuffer& args)
{
    unsigned i = 0;
    unsigned vectorEnd;
    WriteBarrier<Unknown>* vector;

    switch (structure()->indexingType()) {
    case ArrayClass:
        return;

    case ArrayWithUndecided: {
        vector = 0;
        vectorEnd = 0;
        break;
    }

    case ArrayWithInt32:
    case ArrayWithContiguous: {
        vectorEnd = m_butterfly->publicLength();
        vector = m_butterfly->contiguous().data();
        break;
    }

    case ArrayWithDouble: {
        vector = 0;
        vectorEnd = 0;
        for (; i < m_butterfly->publicLength(); ++i) {
            ASSERT(i < butterfly()->vectorLength());
            double v = butterfly()->contiguousDouble()[i];
            if (v != v)
                break;
            args.append(JSValue(JSValue::EncodeAsDouble, v));
        }
        break;
    }

    case ARRAY_WITH_ARRAY_STORAGE_INDEXING_TYPES: {
        ArrayStorage* storage = m_butterfly->arrayStorage();

        vector = storage->m_vector;
        vectorEnd = min(storage->length(), storage->vectorLength());
        break;
    }

    default:
        RELEASE_ASSERT_NOT_REACHED();
        vector = 0;
        vectorEnd = 0;
        break;
    }

    for (; i < vectorEnd; ++i) {
        WriteBarrier<Unknown>& v = vector[i];
        if (!v)
            break;
        args.append(v.get());
    }

    // FIXME: What prevents this from being called with a RuntimeArray? The length function will always return 0 in that case.
    VM& vm = exec->vm();
    for (; i < length(); ++i)
        args.append(get(vm, exec, i));
}

void JSArray::copyToArguments(ExecState* exec, CallFrame* callFrame, uint32_t copyLength, int32_t firstVarArgOffset)
{
    VM& vm = exec->vm();

    unsigned i = firstVarArgOffset;
    WriteBarrier<Unknown>* vector;
    unsigned vectorEnd;
    unsigned length = copyLength + firstVarArgOffset;

    // FIXME: What prevents this from being called with a RuntimeArray? The length function will always return 0 in that case.
    ASSERT(length == this->length());

    switch (structure()->indexingType()) {
    case ArrayClass:
        return;

    case ArrayWithUndecided: {
        vector = 0;
        vectorEnd = 0;
        break;
    }

    case ArrayWithInt32:
    case ArrayWithContiguous: {
        vector = m_butterfly->contiguous().data();
        vectorEnd = m_butterfly->publicLength();
        break;
    }

    case ArrayWithDouble: {
        vector = 0;
        vectorEnd = 0;
        for (; i < m_butterfly->publicLength(); ++i) {
            double v = m_butterfly->contiguousDouble()[i];
            if (v != v)
                break;
            callFrame->setArgument(i - firstVarArgOffset, JSValue(JSValue::EncodeAsDouble, v));
        }
        break;
    }

    case ARRAY_WITH_ARRAY_STORAGE_INDEXING_TYPES: {
        ArrayStorage* storage = m_butterfly->arrayStorage();
        vector = storage->m_vector;
        vectorEnd = min(length, storage->vectorLength());
        break;
    }

    default:
        RELEASE_ASSERT_NOT_REACHED();
        vector = 0;
        vectorEnd = 0;
        break;
    }

    for (; i < vectorEnd; ++i) {
        WriteBarrier<Unknown>& v = vector[i];
        if (!v)
            break;
        callFrame->setArgument(i - firstVarArgOffset, v.get());
    }

    for (; i < length; ++i) {
        callFrame->setArgument(i - firstVarArgOffset, get(vm, exec, i));
        RETURN_IF_EXCEPTION(vm, void());
    }
}

static inline JSArray* constructArray(ExecState* exec, ObjectInitializationScope& scope, DeferGC& deferGC, Structure* arrayStructure, unsigned length)
{
    JSArray* array = JSArray::tryCreateUninitializedRestricted(scope, deferGC, arrayStructure, length);

    // FIXME: we should probably throw an out of memory error here, but
    // when making this change we should check that all clients of this
    // function will correctly handle an exception being thrown from here.
    // https://bugs.webkit.org/show_bug.cgi?id=169786
    RELEASE_ASSERT(array);
	if (UNLIKELY(!array)) {
		throwOutOfMemoryError(exec);
		return nullptr;
	}

    // FIXME: We only need this for subclasses of Array because we might need to allocate a new structure to change
    // indexing types while initializing. If this triggered a GC then we might scan our currently uninitialized
    // array and crash. https://bugs.webkit.org/show_bug.cgi?id=186811
    if (!arrayStructure->globalObject()->isOriginalArrayStructure(arrayStructure))
        JSArray::eagerlyInitializeButterfly(array, length);

    return array;
}

JSArray* constructArray(VM& vm, ExecState* exec, Structure* arrayStructure, const ArgList& values)
{
    DeferGC deferGC(vm.heap);

	unsigned length = values.size();
    ObjectInitializationScope scope(vm);

    JSArray* array = constructArray(exec, scope, deferGC, arrayStructure, length);
    RETURN_IF_EXCEPTION(vm, nullptr);

	// FIXME: we should probably throw an out of memory error here, but
	// when making this change we should check that all clients of this
	// function will correctly handle an exception being thrown from here.
    // https://bugs.webkit.org/show_bug.cgi?id=169786
    RELEASE_ASSERT(array);

	for (unsigned i = 0; i < length; ++i)
		array->initializeIndex(vm, i, values.at(i));
	return array;
}

JSArray* constructArray(VM& vm, ExecState* exec, Structure* arrayStructure, const JSValue* values, unsigned length)
{
    DeferGC deferGC(vm.heap);
    ObjectInitializationScope scope(vm);

    JSArray* array = constructArray(exec, scope, deferGC, arrayStructure, length);
    RETURN_IF_EXCEPTION(vm, nullptr);

	// FIXME: we should probably throw an out of memory error here, but
	// when making this change we should check that all clients of this
	// function will correctly handle an exception being thrown from here.
	// https://bugs.webkit.org/show_bug.cgi?id=169786
    RELEASE_ASSERT(array);

	for (unsigned i = 0; i < length; ++i)
		array->initializeIndex(vm, i, values[i]);
	return array;
}

JSArray* constructArrayNegativeIndexed(VM& vm, ExecState* exec, Structure* arrayStructure, const JSValue* values, unsigned length)
{
    DeferGC deferGC(vm.heap);
    ObjectInitializationScope scope(vm);

    JSArray* array = constructArray(exec, scope, deferGC, arrayStructure, length);
    RETURN_IF_EXCEPTION(vm, nullptr);

    // FIXME: we should probably throw an out of memory error here, but
    // when making this change we should check that all clients of this
    // function will correctly handle an exception being thrown from here.
    // https://bugs.webkit.org/show_bug.cgi?id=169786
    RELEASE_ASSERT(array);

    for (int i = 0; i < static_cast<int>(length); ++i)
        array->initializeIndex(vm, i, values[-i]);
    return array;
}

bool JSArray::isIteratorProtocolFastAndNonObservable()
{
    JSGlobalObject* globalObject = this->globalObject();
    if (!globalObject->isArrayPrototypeIteratorProtocolFastAndNonObservable())
        return false;

    Structure* structure = this->structure();
    // This is the fast case. Many arrays will be an original array.
    if (globalObject->isOriginalArrayStructure(structure))
        return true;

    if (structure->mayInterceptIndexedAccesses())
        return false;

    if (structure->storedPrototype() != globalObject->arrayPrototype())
        return false;

    if (getDirectOffset(globalObject->vm(), globalObject->vm().propertyNames->iteratorSymbol) != invalidOffset)
        return false;

    return true;
}

} // namespace JSC
