/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#pragma once

#include "ArrayConventions.h"
#include "ButterflyInlineMethods.h"
#include "JSCellInlines.h"
#include "Exception.h"
#include "JSObject.h"

namespace JSC {

class JSArray;
class LLIntOffsetsExtractor;

extern const ASCIILiteral LengthExceededTheMaximumArrayLengthError;

class JSArray : public JSNonFinalObject {
	friend class LLIntOffsetsExtractor;
	friend class Walker;
	friend class JIT;

public:
	typedef JSNonFinalObject Base;
	static const unsigned StructureFlags = Base::StructureFlags | OverridesGetOwnPropertySlot | OverridesGetOwnSpecialPropertyNames | OverridesPut;

    static size_t allocationSize(size_t inlineCapacity)
    {
        ASSERT_UNUSED(inlineCapacity, !inlineCapacity);
        return sizeof(JSArray);
    }

protected:
	explicit JSArray(JSGlobalData& globalData, Structure* structure, Butterfly* butterfly)
		: JSNonFinalObject(globalData, structure, butterfly)
	{
	}

public:
    static JSArray* tryCreate(JSGlobalData&, ExecState* exec, Structure*, unsigned initialLength = 0);
	static JSArray* tryCreate(JSGlobalData&, ExecState* exec, Structure*, unsigned initialLength, unsigned vectorLengthHint);
    static JSArray* createWithButterfly(JSGlobalData&, DeferGC&, Structure*, Butterfly*);

    // tryCreateUninitializedRestricted is used for fast construction of arrays whose size and
    // contents are known at time of creation. This is a restricted API for careful use only in
    // performance critical code paths. If you don't have a good reason to use it, you probably
    // shouldn't use it. Instead, you should go with
    //   - JSArray::tryCreate() or JSArray::create() instead of tryCreateUninitializedRestricted(), and
    //   - putDirectIndex() instead of initializeIndex().
    //
    // Clients of this interface must:
	//   - null-check the result (indicating out of memory, or otherwise unable to allocate vector).
	//   - call 'initializeIndex' for all properties in sequence, for 0 <= i < initialLength.
    //   - Provide a valid GCDefferalContext* if they might garbage collect when initializing properties,
    //     otherwise the caller can provide a null GCDefferalContext*.
    //   - Provide a local stack instance of ObjectInitializationScope at the call site.
	static JSArray* tryCreateUninitializedRestricted(ObjectInitializationScope& scope, DeferGC&, Structure*, unsigned initialLength);

    static void eagerlyInitializeButterfly(JSArray*, unsigned initialLength);

	JS_EXPORT_PRIVATE static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool throwException);

	static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);

	DECLARE_EXPORT_INFO

	// OK if we know this is a JSArray, but not if it could be an object of a derived class; for RuntimeArray this always returns 0.
	unsigned length() const { return getArrayLength(); }

	// OK to use on new arrays, but not if it might be a RegExpMatchArray or RuntimeArray.
	bool setLength(ExecState*, unsigned, bool throwException = false);

    void pushInline(VM&, ExecState*, JSValue);
	void push(ExecState*, JSValue);
	JSValue pop(ExecState*);

	JSArray* fastSlice(ExecState&, unsigned startIndex, unsigned count);

    bool canFastCopy(VM&, JSArray* otherArray);
    bool canDoFastIndexedAccess(VM&);
    // This function returns NonArray if the indexing types are not compatable for copying.
    IndexingType mergeIndexingTypeForCopying(IndexingType other);
    bool appendMemcpy(ExecState*, VM&, unsigned startIndex, JSArray* otherArray);

	enum ShiftCountMode {
		// This form of shift hints that we're doing queueing. With this assumption in hand,
		// we convert to ArrayStorage, which has queue optimizations.
		ShiftCountForShift,

		// This form of shift hints that we're just doing care and feeding on an array that
		// is probably typically used for ordinary accesses. With this assumption in hand,
		// we try to preserve whatever indexing type it has already.
		ShiftCountForSplice
	};

	template<ShiftCountMode shiftCountMode>
	bool shiftCount(ExecState* exec, unsigned& startIndex, unsigned count)
	{
        constexpr unsigned shiftThreashold = 128;
        //UNUSED_VARIABLE(shiftThreashold);
        if (shiftCountMode == ShiftCountForShift)
            return shiftCountWithAnyIndexingType(exec, startIndex, count, shiftThreashold);
        else if (shiftCountMode == ShiftCountForSplice)
            return shiftCountWithAnyIndexingType(exec, startIndex, count, UINT32_MAX);
        RELEASE_ASSERT_NOT_REACHED();
        return false;
	}

	bool unshiftCount(ExecState* exec, unsigned startIndex, unsigned count)
	{
		return unshiftCountWithAnyIndexingType(exec, startIndex, count);
	}

	void fillArgList(ExecState*, MarkedArgumentBuffer&);
	void copyToArguments(ExecState*, CallFrame*, uint32_t copyLength, int32_t firstVarArgOffset);

    bool isIteratorProtocolFastAndNonObservable();

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype, IndexingType indexingType)
	{
		return Structure::create(globalData, globalObject, prototype, TypeInfo(ArrayType, StructureFlags), info(), indexingType);
	}

protected:
    void finishCreation(JSGlobalData& vm)
    {
        Base::finishCreation(vm);
        ASSERT_WITH_MESSAGE(type() == ArrayType || type() == DerivedArrayType, "Instance inheriting JSArray should have either ArrayType or DerivedArrayType");
    }

	static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);

	static bool deleteProperty(JSCell*, ExecState*, PropertyName);
	JS_EXPORT_PRIVATE static void getOwnSpecialPropertyNames(JSObject*, ExecState*, PropertyNameArray&, DontEnumPropertiesMode);

private:
	bool isLengthWritable()
	{
		ArrayStorage* storage = arrayStorageOrNull();
		if (!storage)
			return true;
		SparseArrayValueMap* map = storage->m_sparseMap.get();
		return !map || !map->lengthIsReadOnly();
	}

	bool shiftCountWithAnyIndexingType(ExecState*, unsigned& startIndex, unsigned count, unsigned shiftArrayStorageThreshold);
	bool shiftCountWithArrayStorage(JSGlobalData&, unsigned startIndex, unsigned count, ArrayStorage*);

	bool unshiftCountWithAnyIndexingType(ExecState*, unsigned startIndex, unsigned count);
	bool unshiftCountWithArrayStorage(ExecState*, unsigned startIndex, unsigned count, ArrayStorage*);
	bool unshiftCountSlowCase(JSGlobalData&, bool, unsigned);

	bool setLengthWithArrayStorage(ExecState*, unsigned newLength, bool throwException, ArrayStorage*);
	void setLengthWritable(ExecState*, bool writable);
};

inline Butterfly* tryCreateArrayButterfly(VM& vm, JSCell* intendedOwner, Structure* structure, unsigned initialLength)
{
    unsigned outOfLineStorage = structure->outOfLineCapacity();

    Butterfly* butterfly = Butterfly::tryCreate(
        vm, intendedOwner, 0, outOfLineStorage, true, baseIndexingHeaderForArrayStorage(initialLength),
        ArrayStorage::sizeFor(BASE_ARRAY_STORAGE_VECTOR_LEN));
    if (UNLIKELY(!butterfly))
        return nullptr;
    ArrayStorage* storage = butterfly->arrayStorage();
    storage->m_sparseMap.clear();
    storage->m_indexBias = 0;
    storage->m_numValuesInVector = 0;
    return butterfly;
}

inline JSArray* JSArray::tryCreate(JSGlobalData& vm, ExecState* exec, Structure* structure, unsigned initialLength, unsigned vectorLengthHint)
{
    ASSERT(vectorLengthHint >= initialLength);
    unsigned outOfLineStorage = structure->outOfLineCapacity();

	// DO not GC until bufferfly is attached.
    DeferGC deferGC(vm.heap);

	Butterfly* butterfly;
    IndexingType indexingType = structure->indexingType();
	if (LIKELY(!hasAnyArrayStorage(indexingType))) {
		ASSERT(
			hasUndecided(indexingType)
			|| hasInt32(indexingType)
			|| hasDouble(indexingType)
			|| hasContiguous(indexingType));

        if (UNLIKELY(vectorLengthHint > MAX_STORAGE_VECTOR_LENGTH)) {
    		throwOutOfMemoryError(vm, exec);
            return nullptr;
        }

		IndexingHeader header;
		unsigned vectorLength = std::max(vectorLengthHint, BASE_ARRAY_STORAGE_VECTOR_LEN);
		header.setVectorLength(vectorLength);
		header.setPublicLength(initialLength);
		butterfly = Butterfly::create(vm, nullptr, 0, outOfLineStorage, true, header, vectorLength * sizeof(EncodedJSValue));
		ASSERT(initialLength < MIN_ARRAY_STORAGE_CONSTRUCTION_LENGTH);
		if (hasDouble(indexingType)) {
			for (unsigned i = 0; i < vectorLength; ++i)
				butterfly->contiguousDouble()[i] = PNaN;
		}
	} else {
		ASSERT(
			indexingType == ArrayWithSlowPutArrayStorage
			|| indexingType == ArrayWithArrayStorage);
		butterfly = tryCreateArrayButterfly(vm, nullptr, structure, initialLength);
        if (UNLIKELY(!butterfly)) {
    		throwOutOfMemoryError(vm, exec);
            return nullptr;
        }
	}
    return createWithButterfly(vm, deferGC, structure, butterfly);
}

inline JSArray* JSArray::tryCreate(JSGlobalData& vm, ExecState* exec, Structure* structure, unsigned initialLength)
{
    return tryCreate(vm, exec, structure, initialLength, initialLength);
}

inline JSArray* JSArray::createWithButterfly(JSGlobalData& vm, DeferGC&, Structure* structure, Butterfly* butterfly)
{
	// DO not GC until bufferfly is attached.
    JSArray* array = new (NotNull, allocateCell<JSArray>(vm.heap)) JSArray(vm, structure, butterfly);
    array->finishCreation(vm);
    return array;
}

JSArray* asArray(JSValue);

ALWAYS_INLINE JSArray* asArray(JSCell* cell)
{
	ASSERT(cell->inherits(JSArray::info()));
	return jsCast<JSArray*>(cell);
}

ALWAYS_INLINE JSArray* asArray(JSValue value)
{
	return asArray(value.asCell());
}

ALWAYS_INLINE bool isJSArray(const JSCell* cell)
{
    ASSERT((cell->classInfo() == JSArray::info()) == (cell->structure()->typeInfo().type() == ArrayType));
    return cell->structure()->typeInfo().type() == ArrayType;
}

ALWAYS_INLINE bool isJSArray(JSValue v) { return v.isCell() && isJSArray(v.asCell()); }

template<>
ALWAYS_INLINE JSArray* jsDynamicCast(JSCell* from)
{
	auto type = from->structure()->typeInfo().type();
	if(type >= JSType::ArrayType && type <= JSType::DerivedArrayType)
		return static_cast<JSArray*>(from);
	return nullptr;
}

JSArray* constructArray(VM&, ExecState* exec, Structure*, const ArgList& values);
JSArray* constructArray(VM&, ExecState* exec, Structure*, const JSValue* values, unsigned length);
JSArray* constructArrayNegativeIndexed(VM&, ExecState* exec, Structure*, const JSValue* values, unsigned length);

ALWAYS_INLINE double toLength(VM& vm, ExecState* exec, JSObject* obj)
{
    if (LIKELY(isJSArray(obj)))
        return jsCast<JSArray*>(obj)->length();

    JSValue lengthValue = obj->get(vm, exec, vm.propertyNames->length);
    RETURN_IF_EXCEPTION(vm, PNaN);
    return lengthValue.toLength(exec);
}

} // namespace JSC
