/*
 * Copyright (C) 2012-2020 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "CallFrame.h"
#include "DeferGC.h"
#include "Handle.h"
#include "JSBigInt.h"
#include "JSCell.h"
#include "JSDestructibleObject.h"
#include "JSObject.h"
#include "JSString.h"
#include "ExceptionScope.h"
#include "MarkedBlock.h"
#include "Structure.h"
#include "Symbol.h"
#include "WeakInlines.h"

namespace JSC {

ALWAYS_INLINE JSCell::JSCell(CreatingEarlyCellTag)
{
}

ALWAYS_INLINE JSCell::JSCell(JSGlobalData& globalData, Structure* structure)
: m_structure(globalData, this, structure)
{
    // Note that in the constructor initializer list above, we are only using values
    // inside structure but not necessarily the structure pointer itself. All these
    // values are contained inside Structure::m_blob. Note also that this constructor
    // is an inline function. Hence, the compiler may choose to pre-compute the address
    // of structure->m_blob and discard the structure pointer itself. There's a chance
    // that the GC may run while allocating this cell. In the event that the structure
    // is newly instantiated just before calling this constructor, there may not be any
    // other references to it. As a result, the structure may get collected before this
    // cell is even constructed. To avoid this possibility, we need to ensure that the
    // structure pointer is still alive at this point.
    ensureStillAliveHere(structure);
}

ALWAYS_INLINE void JSCell::finishCreation(JSGlobalData& globalData)
{
#if ENABLE(GC_VALIDATION)
	ASSERT(globalData.isInitializingObject());
	globalData.setInitializingObjectClass(0);
#else
	UNUSED_PARAM(globalData);
#endif
	ASSERT(m_structure);
}

ALWAYS_INLINE void JSCell::finishCreation(JSGlobalData& globalData, Structure* structure, CreatingEarlyCellTag)
{
#if ENABLE(GC_VALIDATION)
	ASSERT(globalData.isInitializingObject());
	globalData.setInitializingObjectClass(0);
	if (structure)
#endif
		m_structure.setEarlyValue(globalData, this, structure);
	// Very first set of allocations won't have a real structure.
	ASSERT(m_structure || !globalData.structureStructure);
}

ALWAYS_INLINE Structure* JSCell::structure() const
{
	return m_structure.get();
}

ALWAYS_INLINE Structure* JSCell::structure(JSGlobalData&) const
{
	return m_structure.get();
}

ALWAYS_INLINE void JSCell::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
	MARK_LOG_PARENT(visitor, cell);

	visitor.append(&cell->m_structure);
}

ALWAYS_INLINE JSGlobalData& JSCell::vm() const
{
    return MarkedBlock::blockFor(this)->vm();
}

ALWAYS_INLINE VM& ExecState::vm() const
{
    ASSERT(callee());
    ASSERT(callee()->vm());
    return calleeAsValue().asCell()->vm();
}

ALWAYS_INLINE JSGlobalData& ExecState::globalData() const
{
	return vm();
}

template<typename T>
ALWAYS_INLINE void* allocateCell(Heap& heap, size_t size)
{
	ASSERT(size >= sizeof(T));
	JSCell* result = 0;
	if (T::needsDestruction && (T::StructureFlags & StructureIsImmortal))
		result = static_cast<JSCell*>(heap.allocateWithImmortalStructureDestructor(size));
	else if (T::needsDestruction)
		result = static_cast<JSCell*>(heap.allocateWithNormalDestructor(size));
	else
		result = static_cast<JSCell*>(heap.allocateWithoutDestructor(size));
#if ENABLE(GC_VALIDATION)
	ASSERT(!heap.globalData()->isInitializingObject());
	heap.globalData()->setInitializingObjectClass(T::info());
#endif
	result->clearStructure();
	return result;
}

template<typename T>
ALWAYS_INLINE void* allocateCell(Heap& heap)
{
	return allocateCell<T>(heap, sizeof(T));
}

ALWAYS_INLINE bool isZapped(const JSCell* cell)
{
	return cell->isZapped();
}

ALWAYS_INLINE bool JSCell::isObject() const
{
	return m_structure->isObject();
}

ALWAYS_INLINE bool JSCell::isString() const
{
	return m_structure->typeInfo().type() == StringType;
}

inline bool JSCell::isHeapBigInt() const
{
    return m_structure->typeInfo().type() == HeapBigIntType;
}

ALWAYS_INLINE bool JSCell::isSymbol() const
{
    return m_structure->typeInfo().type() == SymbolType;
}

ALWAYS_INLINE bool JSCell::isGetterSetter() const
{
	return m_structure->typeInfo().type() == GetterSetterType;
}

ALWAYS_INLINE bool JSCell::isCustomGetterSetter() const
{
    return m_structure->typeInfo().type() == CustomGetterSetterType;
}

ALWAYS_INLINE bool JSCell::isProxy() const
{
	JSType type = structure()->typeInfo().type();
    return type == ImpureProxyType || type == PureForwardingProxyType || type == ProxyObjectType;
}

ALWAYS_INLINE bool JSCell::isCallable()
{
    if (structure()->typeInfo().type() == JSFunctionType)
        return true;
    if (structure()->typeInfo().overridesGetCallData())
        return methodTable()->getCallData(this).type != CallData::Type::None;
    return false;
}

ALWAYS_INLINE bool JSCell::isCallable(VM&)
{
	return isCallable();
}

inline bool JSCell::isConstructor()
{
    return methodTable()->getConstructData(this).type != CallData::Type::None;
}

inline bool JSCell::isConstructor(VM&)
{
	return isConstructor();
}

ALWAYS_INLINE bool JSCell::isAPIValueWrapper() const
{
	return m_structure->typeInfo().type() == APIValueWrapperType;
}

ALWAYS_INLINE void JSCell::setStructure(JSGlobalData& globalData, Structure* structure)
{
	ASSERT(structure->classInfo() == m_structure->classInfo());
	ASSERT(!m_structure
		   || m_structure->transitionWatchpointSetHasBeenInvalidated()
		   || m_structure.get() == structure);
	m_structure.set(globalData, this, structure);
}

ALWAYS_INLINE const MethodTable* JSCell::methodTable() const
{
    Structure* structure = this->structure();
#if 0
    if (Structure* rootStructure = structure->structure())
        RELEASE_ASSERT(rootStructure == rootStructure->structure());
#endif
    return &structure->classInfo()->methodTable;
}

ALWAYS_INLINE const MethodTable* JSCell::methodTable(VM& vm) const
{
    Structure* structure = this->structure(vm);
#if 0
    if (Structure* rootStructure = structure->structure(vm))
        RELEASE_ASSERT(rootStructure == rootStructure->structure(vm));
#endif
    return &structure->classInfo()->methodTable;
}

ALWAYS_INLINE bool JSCell::inherits(const ClassInfo* info) const
{
    return classInfo()->isSubClassOf(info);
}

template<typename T>
ALWAYS_INLINE bool JSCell::inherits() const
{
    return classInfo()->isSubClassOf(T::info());
}

template<typename T>
ALWAYS_INLINE bool JSCell::inherits(VM&) const
{
    return classInfo()->isSubClassOf(T::info());
}

ALWAYS_INLINE JSValue JSCell::fastGetOwnProperty(JSGlobalData& vm, Structure& structure, PropertyName name)
{
    ASSERT(canUseFastGetOwnProperty(vm));
    PropertyOffset offset = structure.get(vm, name);
	if (offset != invalidOffset)
		return asObject(this)->locationForOffset(offset)->get();
	return JSValue();
}

ALWAYS_INLINE JSValue JSCell::fastGetOwnProperty(JSGlobalData& vm, Structure* structure, PropertyName name)
{
	return fastGetOwnProperty(vm, *structure, name);
}

ALWAYS_INLINE bool JSCell::canUseFastGetOwnProperty(const Structure& structure)
{
    return !structure.hasAnyKindOfGetterSetterProperties()
    	&& !structure.typeInfo().overridesGetOwnPropertySlot();
}

ALWAYS_INLINE bool JSCell::canUseFastGetOwnProperty(const Structure* structure)
{
	return canUseFastGetOwnProperty(*structure);
}

ALWAYS_INLINE const ClassInfo* JSCell::classInfo() const
{
    return structure()->classInfo();
}

inline bool JSCell::toBoolean(ExecState* exec) const
{
	if (isString())
		return static_cast<const JSString*>(this)->toBoolean();
    if (isHeapBigInt())
        return static_cast<const JSBigInt*>(this)->toBoolean();
	return !structure()->masqueradesAsUndefined(exec->lexicalGlobalObject());
}

inline TriState JSCell::pureToBoolean() const
{
	if (isString())
    	return static_cast<const JSString*>(this)->toBoolean() ? TrueTriState : FalseTriState;
	if (isHeapBigInt())
		return static_cast<const JSBigInt*>(this)->toBoolean() ? TrueTriState : FalseTriState;
	if (isSymbol())
        return TrueTriState;
	return MixedTriState;
}

inline JSObject* JSCell::toObject(ExecState* exec, JSGlobalObject* globalObject) const
{
    if (isObject())
        return jsCast<JSObject*>(const_cast<JSCell*>(this));
    return toObjectSlow(exec, globalObject);
}

ALWAYS_INLINE bool JSCell::putInline(ExecState* exec, VM& vm, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    auto putMethod = methodTable()->put;
    if (LIKELY(putMethod == JSObject::put))
        return JSObject::putInlineForJSObject(asObject(this), exec, vm, propertyName, value, slot);
    return putMethod(this, exec, propertyName, value, slot);
}

} // namespace JSC
