/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *  Copyright (C) 2007 Cameron Zwarich (cwzwarich@uwaterloo.ca)
 *  Copyright (C) 2007 Maks Orlovich
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "FunctionRareData.h"
#include "InternalFunction.h"
#include "JSCallee.h"
#include "JSScope.h"
#include "Watchpoint.h"

namespace JSC {

class ExecutableBase;
class FunctionExecutable;
class FunctionPrototype;
class JSLexicalEnvironment;
class JSGlobalObject;
class LLIntOffsetsExtractor;
class NativeExecutable;
class SourceCode;
class InternalFunction;
namespace DFG {
class SpeculativeJIT;
class JITCompiler;
}

JS_EXPORT_PRIVATE EncodedJSValue JSC_HOST_CALL callHostFunctionAsConstructor(ExecState*);

JS_EXPORT_PRIVATE String getCalculatedDisplayName(JSGlobalData&, JSObject*);

class JSFunction : public JSCallee {
	friend class JIT;
	friend class DFG::SpeculativeJIT;
	friend class DFG::JITCompiler;
	friend class JSGlobalData;
    friend class InternalFunction;

public:
	typedef JSCallee Base;
	const static unsigned StructureFlags = Base::StructureFlags | OverridesGetOwnPropertySlot | ImplementsHasInstance | OverridesGetOwnSpecialPropertyNames | OverridesGetCallData | OverridesPut;

    static size_t allocationSize(size_t inlineCapacity)
    {
        ASSERT_UNUSED(inlineCapacity, !inlineCapacity);
        return sizeof(JSFunction);
    }

    static Structure* selectStructureForNewFuncExp(JSGlobalObject*, FunctionExecutable*);

	JS_EXPORT_PRIVATE static JSFunction* create(JSGlobalData&, JSGlobalObject*, int length, const String& name, NativeFunction, Intrinsic = NoIntrinsic, NativeFunction nativeConstructor = callHostFunctionAsConstructor);

	static JSFunction* createWithInvalidatedReallocationWatchpoint(JSGlobalData&, FunctionExecutable*, JSScope*);

    static JSFunction* create(JSGlobalData&, FunctionExecutable*, JSScope*);
    static JSFunction* create(JSGlobalData&, FunctionExecutable*, JSScope*, Structure*);

	JS_EXPORT_PRIVATE String name(JSGlobalData&);
	JS_EXPORT_PRIVATE String displayName(JSGlobalData&);
	const String calculatedDisplayName(JSGlobalData&);

	ExecutableBase* executable() const { return m_executable.get(); }

    // To call any of these methods include JSFunctionInlines.h
	inline bool isHostFunction() const;
	FunctionExecutable* jsExecutable() const;
    Intrinsic intrinsic() const;

	JS_EXPORT_PRIVATE const SourceCode* sourceCode() const;

    DECLARE_INFO;

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype)
	{
		ASSERT(globalObject);
		return Structure::create(globalData, globalObject, prototype, TypeInfo(JSFunctionType, StructureFlags), info());
	}

	NativeFunction nativeFunction();
	NativeFunction nativeConstructor();

	static CallData getConstructData(JSCell*);
	static CallData getCallData(JSCell*);

	static inline ptrdiff_t offsetOfExecutable()
	{
		return OBJECT_OFFSETOF(JSFunction, m_executable);
	}

    static inline ptrdiff_t offsetOfRareData()
    {
        return OBJECT_OFFSETOF(JSFunction, m_rareData);
    }

    FunctionRareData* rareData(VM& vm)
    {
        if (UNLIKELY(!m_rareData))
            return allocateRareData(vm);
        return m_rareData.get();
    }

    FunctionRareData* ensureRareDataAndObjectAllocationProfile(ExecState*, unsigned inlineCapacity);

    FunctionRareData* rareData()
    {
        FunctionRareData* rareData = m_rareData.get();

        // The JS thread may be concurrently creating the rare data
        // If we see it, we want to ensure it has been properly created
        WTF::loadLoadFence();

        return rareData;
    }

    bool isHostOrBuiltinFunction() const;
    bool isBuiltinFunction() const;
    JS_EXPORT_PRIVATE bool isHostFunctionNonInline() const;
    bool isClassConstructorFunction() const;

    // Returns the __proto__ for the |this| value if this JSFunction were to be constructed.
    JSObject* prototypeForConstruction(VM&, ExecState*);

    bool canUseAllocationProfiles(VM&);

    void setFunctionName(ExecState*, JSValue name);

    enum class PropertyStatus {
        Eager,
        Lazy,
        Reified,
    };
    PropertyStatus reifyLazyPropertyIfNeeded(VM&, ExecState*, PropertyName);

    bool mayHaveNonReifiedPrototype();

protected:
	JS_EXPORT_PRIVATE JSFunction(JSGlobalData&, NativeExecutable*, JSGlobalObject*, Structure*);
    JSFunction(JSGlobalData&, FunctionExecutable*, JSScope*, Structure*);

	void finishCreation(JSGlobalData&, NativeExecutable*, int length, const String& name);
    void finishCreation(VM&);

	static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);
	static void getOwnSpecialPropertyNames(JSObject*, ExecState*, PropertyNameArray&, DontEnumPropertiesMode);
	static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool shouldThrow);

	static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);

	static bool deleteProperty(JSCell*, ExecState*, PropertyName);

	static void visitChildren(JSCell*, SlotVisitor&);

private:
    static JSFunction* createImpl(JSGlobalData& vm, FunctionExecutable* executable, JSScope* scope, Structure* structure)
    {
        JSFunction* function = new (NotNull, allocateCell<JSFunction>(vm.heap)) JSFunction(vm, executable, scope, structure);
        ASSERT(function->structure()->globalObject());
        function->finishCreation(vm);
        return function;
    }

    FunctionRareData* allocateRareData(JSGlobalData&);
    FunctionRareData* allocateAndInitializeRareData(ExecState*, size_t inlineCapacity);
    FunctionRareData* initializeRareData(ExecState*, size_t inlineCapacity);

    bool hasReifiedLength() const;
    bool hasReifiedName() const;
    void reifyLength(VM&);
    void reifyName(VM&, ExecState*);
    void reifyName(VM&, ExecState*, String name);

    static bool isLazy(PropertyStatus property) { return property == PropertyStatus::Lazy || property == PropertyStatus::Reified; }
    static bool isReified(PropertyStatus property) { return property == PropertyStatus::Reified; }

    PropertyStatus reifyLazyPropertyForHostOrBuiltinIfNeeded(VM&, ExecState*, PropertyName);
    PropertyStatus reifyLazyPrototypeIfNeeded(VM&, ExecState*, PropertyName);
	PropertyStatus reifyLazyLengthIfNeeded(VM&, ExecState*, PropertyName);
    PropertyStatus reifyLazyNameIfNeeded(VM&, ExecState*, PropertyName);
    PropertyStatus reifyLazyBoundNameIfNeeded(VM&, ExecState*, PropertyName);

#if ASSERT_DISABLED
    void assertTypeInfoFlagInvariants() { }
#else
    void assertTypeInfoFlagInvariants();
#endif

	friend class LLIntOffsetsExtractor;

	WriteBarrier<ExecutableBase> m_executable;
    WriteBarrier<FunctionRareData> m_rareData;
};

class JSStrictFunction FINAL : public JSFunction {
public:
    typedef JSFunction Base;

    DECLARE_EXPORT_INFO;

    static constexpr unsigned StructureFlags = Base::StructureFlags;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        ASSERT(globalObject);
        return Structure::create(vm, globalObject, prototype, TypeInfo(JSFunctionType, StructureFlags), info());
    }
};
static_assert(sizeof(JSStrictFunction) == sizeof(JSFunction), "Allocated in JSFunction IsoSubspace");

class JSSloppyFunction FINAL : public JSFunction {
public:
    typedef JSFunction Base;

    DECLARE_EXPORT_INFO;

    static constexpr unsigned StructureFlags = Base::StructureFlags;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        ASSERT(globalObject);
        return Structure::create(vm, globalObject, prototype, TypeInfo(JSFunctionType, StructureFlags), info());
    }
};
static_assert(sizeof(JSSloppyFunction) == sizeof(JSFunction), "Allocated in JSFunction IsoSubspace");

class JSArrowFunction FINAL : public JSFunction {
public:
    typedef JSFunction Base;

    DECLARE_EXPORT_INFO;

    static constexpr unsigned StructureFlags = Base::StructureFlags;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        ASSERT(globalObject);
        return Structure::create(vm, globalObject, prototype, TypeInfo(JSFunctionType, StructureFlags), info());
    }
};
static_assert(sizeof(JSArrowFunction) == sizeof(JSFunction), "Allocated in JSFunction IsoSubspace");

template<>
ALWAYS_INLINE JSFunction* jsDynamicCast(JSCell* from)
{
	if(from->structure()->typeInfo().type() == JSType::JSFunctionType)
		return static_cast<JSFunction*>(from);
	return 0;
}

template<>
ALWAYS_INLINE JSFunction* jsDynamicCast(JSValue from)
{
	if(from.isCell())
		return jsDynamicCast<JSFunction*>(from.asCell());
	return 0;
}

template<>
ALWAYS_INLINE bool JSCell::inherits<JSFunction>() const
{
	return this->structure()->typeInfo().type() == JSType::JSFunctionType;
}

template<>
ALWAYS_INLINE bool JSCell::inherits<JSFunction>(VM&) const
{
	return this->structure()->typeInfo().type() == JSType::JSFunctionType;
}

} // namespace JSC
