/*
 * Copyright (C) 2008, 2011 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "JSGlobalData.h"

#include "ArgList.h"
#include "AssemblerBuffer.h"
#include "BytecodeIntrinsicRegistry.h"
#include "CallFrameInlines.h"
#include "CodeBlock.h"
#include "CommonIdentifiers.h"
#include "ConservativeRoots.h"
#include "CustomGetterSetter.h"
#include "DFGLongLivedState.h"
#include "DebuggerActivation.h"
#include "ErrorInstance.h"
#include "Exception.h"
#include "FunctionConstructor.h"
#include "GCActivityCallback.h"
#include "GetterSetter.h"
#include "HasOwnPropertyCache.h"
#include "Heap.h"
#include "HeapIterationScope.h"
#include "HostCallReturnValue.h"
#include "Identifier.h"
#include "IncrementalSweeper.h"
#include "Interpreter.h"
#include "JSLexicalEnvironment.h"
#include "JSAPIValueWrapper.h"
#include "JSArray.h"
#include "JSBigInt.h"
#include "JSFixedArray.h"
#include "JSFunction.h"
#include "JSGlobalObjectFunctions.h"
#include "JSInternalPromiseDeferred.h"
#include "JSLock.h"
#include "JSMap.h"
#include "JSMapIterator.h"
#include "JSPromiseDeferred.h"
#include "JSPropertyNameEnumerator.h"
#include "JSScriptFetcher.h"
#include "JSScriptFetchParameters.h"
#include "JSSet.h"
#include "JSSetIterator.h"
#include "JSSourceCode.h"
#include "JSStackInlines.h"
#include "JSTemplateObjectDescriptor.h"
#include "JSWithScope.h"
#include "Lexer.h"
#include "Lookup.h"
#include "MapData.h"
#include "NativeStdFunctionCell.h"
#include "Nodes.h"
#include "Parser.h"
#include "PropertyMapHashTable.h"
#include "RegExpCache.h"
#include "RegExpObject.h"
#include "SourceProviderCache.h"
#include "SimpleTypedArrayController.h"
#include "StrictEvalActivation.h"
#include "StrongInlines.h"
#include "StructureInlines.h"
#include "VariableEnvironment.h"
#include "VMEntryScope.h"
#include "WeakGCMapInlines.h"
#include "WeakMapData.h"
#include <wtf/CurrentTime.h>
#include <wtf/BumpPointerAllocator.h>
#include <wtf/PlatformRunState.h>
#include <wtf/RetainPtr.h>
#include <wtf/Threading.h>
#include <wtf/WTFThreadData.h>
#include <wtf/text/SymbolRegistry.h>

#if ENABLE(REGEXP_TRACING)
#include "RegExp.h"
#endif

#if USE(CF)
#include <CoreFoundation/CoreFoundation.h>
#endif

using namespace WTF;

namespace JSC {

JSGlobalData::JSGlobalData(GlobalDataType globalDataType, HeapType heapType)
    : heap(this, heapType)
    , globalDataType(globalDataType)
    , topCallFrame(CallFrame::noCaller())
    , identifierTable(globalDataType == Default ? wtfThreadData().currentIdentifierTable() : createIdentifierTable())
	, m_privateSymbolRegistry(WTF::SymbolRegistry::Type::PrivateSymbol)
    , propertyNames(nullptr)
    , emptyList(new ArgList)
	, prototypeMap(*this)
	, stringCache(*this)
	, symbolImplToSymbolMap(*this)
    , interpreter(nullptr)
    , jsArrayClassInfo(JSArray::info())
    , jsFinalObjectClassInfo(JSFinalObject::info())
#if ENABLE(JIT)
    , sizeOfLastScratchBuffer(0)
#endif
    , entryScope(nullptr)
	, executionStartTime(monotonicallyIncreasingTime())
	, jitCompileLength(0)
	, parseLength(0)
	, evalPrepareForExecutionLength(0)
	, evalCount(0)
	, m_stackLimit(0)
	, stringRecursionCheckFirstObject(nullptr)
    , m_enabledProfiler(0)
    , m_regExpCache(new RegExpCache(this))
	, m_regExpAllocator(adoptPtr(new BumpPointerAllocator()))
	, m_compactVariableMap(adoptRef(*new CompactTDZEnvironmentMap))
#if ENABLE(REGEXP_TRACING)
    , m_rtTraceList(new RTTraceList())
#endif
#ifndef NDEBUG
    , exclusiveThread(0)
#endif
#if ENABLE(ASSEMBLER)
    , m_canUseAssembler(true)
#endif
#if ENABLE(GC_VALIDATION)
    , m_initializingObjectClass(0)
#endif
	, m_lastStackTop(0)
	, m_exception(0)
	, m_lastException(0)
	, m_globalConstRedeclarationShouldThrow(true)
    , m_deletePropertyMode(DeletePropertyMode::Default)
	, m_lastTimePurgeWatchpointSetForImpureProperty(WTF::monotonicallyIncreasingTime())
{
    interpreter = new Interpreter(*this);

    // Need to be careful to keep everything consistent here
    JSLockHolder lock(this);
    IdentifierTable* existingEntryIdentifierTable = wtfThreadData().setCurrentIdentifierTable(identifierTable);
    structureStructure.set(*this, Structure::createStructure(*this));
    structureRareDataStructure.set(*this, StructureRareData::createStructure(*this, 0, jsNull()));
    stringStructure.set(*this, JSString::createStructure(*this, 0, jsNull()));

    smallStrings.initializeCommonStrings(*this);

    propertyNames = new CommonIdentifiers(*this);
    debuggerActivationStructure.set(*this, DebuggerActivation::createStructure(*this, nullptr, jsNull()));
    terminatedExecutionErrorStructure.set(*this, TerminatedExecutionError::createStructure(*this, nullptr, jsNull()));
    propertyNameEnumeratorStructure.set(*this, JSPropertyNameEnumerator::createStructure(*this, nullptr, jsNull()));
    getterSetterStructure.set(*this, GetterSetter::createStructure(*this, nullptr, jsNull()));
	customGetterSetterStructure.set(*this, CustomGetterSetter::createStructure(*this, nullptr, jsNull()));
    scopedArgumentsTableStructure.set(*this, ScopedArgumentsTable::createStructure(*this, nullptr, jsNull()));
	apiWrapperStructure.set(*this, JSAPIValueWrapper::createStructure(*this, nullptr, jsNull()));
    nativeExecutableStructure.set(*this, NativeExecutable::createStructure(*this, nullptr, jsNull()));
    evalExecutableStructure.set(*this, EvalExecutable::createStructure(*this, nullptr, jsNull()));
    programExecutableStructure.set(*this, ProgramExecutable::createStructure(*this, nullptr, jsNull()));
    functionExecutableStructure.set(*this, FunctionExecutable::createStructure(*this, nullptr, jsNull()));
    moduleProgramExecutableStructure.set(*this, ModuleProgramExecutable::createStructure(*this, nullptr, jsNull()));
	exceptionStructure.set(*this, Exception::createStructure(*this, nullptr, jsNull()));
    regExpStructure.set(*this, RegExp::createStructure(*this, nullptr, jsNull()));
    symbolStructure.set(*this, Symbol::createStructure(*this, nullptr, jsNull()));
	sharedSymbolTableStructure.set(*this, SymbolTable::createStructure(*this, nullptr, jsNull()));
    fixedArrayStructure.set(*this, JSFixedArray::createStructure(*this, nullptr, jsNull()));
    sourceCodeStructure.set(*this, JSSourceCode::createStructure(*this, nullptr, jsNull()));
	scriptFetcherStructure.set(*this, JSScriptFetcher::createStructure(*this, nullptr, jsNull()));
	scriptFetchParametersStructure.set(*this, JSScriptFetchParameters::createStructure(*this, nullptr, jsNull()));
	structureChainStructure.set(*this, StructureChain::createStructure(*this, nullptr, jsNull()));
    sparseArrayValueMapStructure.set(*this, SparseArrayValueMap::createStructure(*this, nullptr, jsNull()));
    templateObjectDescriptorStructure.set(*this, JSTemplateObjectDescriptor::createStructure(*this, nullptr, jsNull()));
	propertyTableStructure.set(*this, PropertyTable::createStructure(*this, nullptr, jsNull()));
	weakMapDataStructure.set(*this, WeakMapData::createStructure(*this, nullptr, jsNull()));
    functionRareDataStructure.set(*this, FunctionRareData::createStructure(*this, nullptr, jsNull()));
	inferredValueStructure.set(*this, InferredValue::createStructure(*this, nullptr, jsNull()));
    promiseDeferredStructure.set(*this, JSPromiseDeferred::createStructure(*this, nullptr, jsNull()));
    internalPromiseDeferredStructure.set(*this, JSInternalPromiseDeferred::createStructure(*this, nullptr, jsNull()));
    hashMapBucketSetStructure.set(*this, HashMapBucket<HashMapBucketDataKey>::createStructure(*this, nullptr, jsNull()));
    hashMapBucketMapStructure.set(*this, HashMapBucket<HashMapBucketDataKeyValue>::createStructure(*this, nullptr, jsNull()));
    setIteratorStructure.set(*this, JSSetIterator::createStructure(*this, 0, jsNull()));
	mapIteratorStructure.set(*this, JSMapIterator::createStructure(*this, 0, jsNull()));

	sentinelSetBucket.set(*this, JSSet::BucketType::createSentinel(*this));
	sentinelMapBucket.set(*this, JSMap::BucketType::createSentinel(*this));

	bigIntStructure.set(*this, JSBigInt::createStructure(*this, nullptr, jsNull()));

	nativeStdFunctionCellStructure.set(*this, NativeStdFunctionCell::createStructure(*this, nullptr, jsNull()));

	heapBigIntConstantOne.set(*this, JSBigInt::createFrom(*this, 1));

    wtfThreadData().setCurrentIdentifierTable(existingEntryIdentifierTable);

#if ENABLE(JIT)
    jitStubs = adoptPtr(new JITThunks());

    callJavaScriptJITFunction = reinterpret_cast<CallJavaScriptJITFunction>(getCTIStub(callToJavaScript).code().executableAddress());
#endif
    
    interpreter->initialize(&llintData, this->canUseJIT());
    
    initializeHostCallReturnValue(); // This is needed to convince the linker not to drop host call return support.

    heap.notifyIsSafeToCollect();
    
    llintData.performAssertions(*this);

    callFrameForCatch = nullptr;

#if ENABLE(DFG_JIT)
    if (canUseJIT())
        m_dfgState = adoptPtr(new DFG::LongLivedState());
#endif

    // Initialize this last, as a free way of asserting that VM initialization itself
    // won't use this.
    m_typedArrayController = adoptRef(new SimpleTypedArrayController());

    m_bytecodeIntrinsicRegistry = adoptPtr(new BytecodeIntrinsicRegistry(*this));
}

JSGlobalData::~JSGlobalData()
{
    ASSERT(!m_apiLock.currentThreadIsHoldingLock());
    smallStrings.setIsInitialized(false);
    heap.didStartVMShutdown();

    delete interpreter;
#ifndef NDEBUG
    interpreter = reinterpret_cast<Interpreter*>(0xbbadbeef);
#endif

    delete emptyList;

    delete propertyNames;
    if (globalDataType != Default)
        deleteIdentifierTable(identifierTable);

    delete clientData;
    delete m_regExpCache;

#if ENABLE(YARR_JIT_ALL_PARENS_EXPRESSIONS)
    if (m_regExpPatternContexBuffer)
        delete[] m_regExpPatternContexBuffer;
#endif

#if ENABLE(REGEXP_TRACING)
    delete m_rtTraceList;
#endif

#if ENABLE(DFG_JIT)
    for (unsigned i = 0; i < m_scratchBuffers.size(); ++i)
        fastFree(m_scratchBuffers[i]);
#endif
}

PassRefPtr<JSGlobalData> JSGlobalData::createContextGroup(HeapType heapType)
{
    return adoptRef(new JSGlobalData(APIContextGroup, heapType));
}

PassRefPtr<JSGlobalData> JSGlobalData::create(HeapType heapType)
{
    return adoptRef(new JSGlobalData(Default, heapType));
}

PassRefPtr<JSGlobalData> JSGlobalData::createLeaked(HeapType heapType)
{
    return create(heapType);
}

#if ENABLE(JAVA_BRIDGE) || OS(DARWIN)
bool JSGlobalData::sharedInstanceExists()
{
    return sharedInstanceInternal();
}

JSGlobalData& JSGlobalData::sharedInstance()
{
    GlobalJSLock globalLock;
    JSGlobalData*& instance = sharedInstanceInternal();
    if (!instance) {
        instance = adoptRef(new JSGlobalData(APIShared, SmallHeap)).leakRef();
        instance->makeUsableFromMultipleThreads();
    }
    return *instance;
}

JSGlobalData*& JSGlobalData::sharedInstanceInternal()
{
    static JSGlobalData* sharedInstance;
    return sharedInstance;
}
#endif

#if ENABLE(JIT)
static ThunkGenerator thunkGeneratorForIntrinsic(Intrinsic intrinsic)
{
    switch (intrinsic) {
    case CharCodeAtIntrinsic:
        return charCodeAtThunkGenerator;
    case CharAtIntrinsic:
        return charAtThunkGenerator;
    case Clz32Intrinsic:
        return clz32ThunkGenerator;
    case StringPrototypeCodePointAtIntrinsic:
        return stringPrototypeCodePointAtThunkGenerator;
    case FromCharCodeIntrinsic:
        return fromCharCodeThunkGenerator;
    case SqrtIntrinsic:
        return sqrtThunkGenerator;
    case AbsIntrinsic:
        return absThunkGenerator;
    case FloorIntrinsic:
        return floorThunkGenerator;
    case CeilIntrinsic:
        return ceilThunkGenerator;
    case TruncIntrinsic:
        return truncThunkGenerator;
    case RoundIntrinsic:
        return roundThunkGenerator;
    case ExpIntrinsic:
        return expThunkGenerator;
    case LogIntrinsic:
        return logThunkGenerator;
    case IMulIntrinsic:
        return imulThunkGenerator;
    case RandomIntrinsic:
        return randomThunkGenerator;
    default:
        return nullptr;
    }
}

NativeExecutable* JSGlobalData::getHostFunction(NativeFunction function, NativeFunction constructor, const String& name)
{
    return getHostFunction(function, NoIntrinsic, constructor, name);
}

NativeExecutable* JSGlobalData::getHostFunction(NativeFunction function, Intrinsic intrinsic, NativeFunction constructor, const String& name)
{
	return jitStubs->hostFunctionStub(
		this, function, constructor,
		intrinsic != NoIntrinsic ? thunkGeneratorForIntrinsic(intrinsic) : nullptr,
		intrinsic, name);
}
#else // !ENABLE(JIT)
NativeExecutable* JSGlobalData::getHostFunction(NativeFunction function, NativeFunction constructor, const String& name)
{
    return NativeExecutable::create(*this, function, constructor, name);
}
#endif // !ENABLE(JIT)

void JSGlobalData::resetDateCache()
{
    utcTimeOffsetCache.reset();
    localTimeOffsetCache.reset();
    cachedDateString = String();
    cachedDateStringValue = std::numeric_limits<double>::quiet_NaN();
    dateInstanceCache.reset();
}

void JSGlobalData::startSampling()
{
    interpreter->startSampling();
}

void JSGlobalData::stopSampling()
{
    interpreter->stopSampling();
}

void JSGlobalData::prepareToDeleteCode()
{
#if 0 && ENABLE(DFG_JIT)
    for (unsigned i = DFG::numberOfWorklists(); i--;) {
        if (DFG::Worklist* worklist = DFG::worklistForIndexOrNull(i))
            worklist->completeAllPlansForVM(*this);
    }
#endif // ENABLE(DFG_JIT)
}

void JSGlobalData::whenIdle(PassOwnPtr<VMIdleCallback> callback)
{
    if (!entryScope) {
    	(*callback)(*this, nullptr);
        return;
    }

    entryScope->addDidPopListener(callback);
}

void JSGlobalData::deleteAllCode()
{
    prepareToDeleteCode();
    //m_codeCache->clear();
    m_regExpCache->deleteAllCode();
    heap.deleteCodeBlocks(true);
}

void JSGlobalData::dumpSampleData(ExecState* exec)
{
    interpreter->dumpSampleData(exec);
#if ENABLE(ASSEMBLER)
    ExecutableAllocator::dumpProfile();
#endif
}

SourceProviderCache* JSGlobalData::addSourceProviderCache(SourceProvider* sourceProvider)
{
    auto addResult = sourceProviderCacheMap.add(sourceProvider, nullptr);
    if (addResult.isNewEntry) {
    	if (RefPtr<JSC::SourceProviderCache> cache = sourceProvider->getSourceProviderCache())
    		addResult.iterator->second = cache.release();
    	else {
    		addResult.iterator->second = SourceProviderCache::create();
    		sourceProvider->setSourceProviderCache(addResult.iterator->second.get());
    	}
    }
    return addResult.iterator->second.get();
}

void JSGlobalData::clearSourceProviderCaches()
{
	for (auto& cacheEntry : sourceProviderCacheMap)
		cacheEntry.first->setSourceProviderCache(cacheEntry.second.release());

    sourceProviderCacheMap.clear();
}

#if ENABLE(JIT)
void JSGlobalData::gatherConservativeRoots(ConservativeRoots& conservativeRoots, bool)
{
    for (size_t i = 0; i < m_scratchBuffers.size(); i++) {
        ScratchBuffer* scratchBuffer = m_scratchBuffers[i];
    	if (scratchBuffer->activeLength()) {
            void* bufferStart = scratchBuffer->dataBuffer();
            conservativeRoots.add(bufferStart, static_cast<void*>(static_cast<char*>(bufferStart) + scratchBuffer->activeLength()));
        }
    }
}
#endif

#if ENABLE(YARR_JIT_ALL_PARENS_EXPRESSIONS)
char* VM::acquireRegExpPatternContexBuffer()
{
    ASSERT(!m_regExpPatternContextLock.isLocked());

    m_regExpPatternContextLock.lock();
    if (!m_regExpPatternContexBuffer)
        m_regExpPatternContexBuffer = new char[VM::patternContextBufferSize];
    return m_regExpPatternContexBuffer;
}

void VM::releaseRegExpPatternContexBuffer()
{
    ASSERT(m_regExpPatternContextLock.isLocked());

    m_regExpPatternContextLock.unlock();
}
#endif

#if ENABLE(REGEXP_TRACING)
void JSGlobalData::addRegExpToTrace(RegExp* regExp)
{
    m_rtTraceList->add(regExp);
}

void JSGlobalData::dumpRegExpTrace()
{
    // The first RegExp object is ignored.  It is create by the RegExpPrototype ctor and not used.
    RTTraceList::iterator iter = ++m_rtTraceList->begin();
    
    if (iter != m_rtTraceList->end()) {
        dataLog("\nRegExp Tracing\n");
        dataLog("                                                            match()    matches\n");
        dataLog("Regular Expression                          JIT Address      calls      found\n");
        dataLog("----------------------------------------+----------------+----------+----------\n");
    
        unsigned reCount = 0;
    
        for (; iter != m_rtTraceList->end(); ++iter, ++reCount)
            (*iter)->printTraceData();

        dataLog("%d Regular Expressions\n", reCount);
    }
    
    m_rtTraceList->clear();
}
#else
void JSGlobalData::dumpRegExpTrace()
{
}
#endif

WatchpointSet* JSGlobalData::ensureWatchpointSetForImpureProperty(const Identifier& propertyName)
{
	auto result = m_impurePropertyWatchpointSets.add(propertyName.string(), nullptr);
    if (result.isNewEntry)
        result.iterator->second = WatchpointSet::create(IsWatched);
    return result.iterator->second.get();
}

void JSGlobalData::addImpureProperty(const String& propertyName)
{
    if (propertyName.isEmpty())
        return;
    if (RefPtr<WatchpointSet> watchpointSet = m_impurePropertyWatchpointSets.take(propertyName))
        watchpointSet->fireAll(*this);
}

void JSGlobalData::purgeWatchpointSetForImpurePropertyIfNecessary()
{
	if (m_impurePropertyWatchpointSets.isEmpty())
		return;

	double now = WTF::monotonicallyIncreasingTime();

	if ((now - m_lastTimePurgeWatchpointSetForImpureProperty) > 60 || WTF::platformRunState() != WTF::Platform_State_Running) {
		HashMap<String, RefPtr<WatchpointSet>> impurePropertyWatchpointSets = WTFMove(m_impurePropertyWatchpointSets);

		WTF::logMessage(tryMakeString("VM purgeWatchpointSetForImpurePropertyIfNecessary total ", String::number(impurePropertyWatchpointSets.size())));

		while (!impurePropertyWatchpointSets.isEmpty()) {
			if (RefPtr<WatchpointSet> watchpointSet = impurePropertyWatchpointSets.take(impurePropertyWatchpointSets.begin()->first))
				watchpointSet->fireAll(*this);
		}

		m_lastTimePurgeWatchpointSetForImpureProperty = now;
	}
}

Exception* JSGlobalData::throwException(ExecState* exec, Exception* exception)
{
#if 0
    if (Options::breakOnThrow()) {
        dataLog("In call frame ", RawPointer(exec), " for code block ", *exec->codeBlock(), "\n");
        CRASH();
    }
#endif

    ASSERT(exec == topCallFrame || exec == exec->lexicalGlobalObject()->globalExec() || exec == exec->vmEntryGlobalObject()->globalExec());

    interpreter->notifyDebuggerOfExceptionToBeThrown(exec, exception);

#if 0
    if (!m_exception && exception->value()) {
    	WTF::logMessage(exception->value().toWTFString(exec));
    }
#endif

    setException(exception);
    return exception;
}

Exception* JSGlobalData::throwException(ExecState* exec, JSValue thrownValue)
{
    Exception* exception = jsDynamicCast<Exception*>(thrownValue);
    if (!exception)
        exception = Exception::create(*this, thrownValue);

    return throwException(exec, exception);
}

Exception* JSGlobalData::throwException(ExecState* exec, JSObject* error)
{
    return throwException(exec, JSValue(error));
}

void JSGlobalData::throwOOMException(Exception* exception)
{
    setException(exception);
}

void sanitizeStackForVM(JSGlobalData& vm)
{
	JSStack& stack = vm.interpreter->stack();

    ASSERT(stack.getTopOfStack() <= stack.getBaseOfStack());

    if (vm.lastStackTop() < stack.getTopOfStack()) {
        char* begin = reinterpret_cast<char*>(vm.lastStackTop());
        char* end = reinterpret_cast<char*>(stack.getTopOfStack());
        memset(begin, 0, end - begin);
    }

    vm.setLastStackTop(stack.getTopOfStack());
}

ScratchBuffer* JSGlobalData::scratchBufferForSize(size_t size)
{
	if (!size)
		return 0;

	if (size > sizeOfLastScratchBuffer) {
		// Protect against a N^2 memory usage pathology by ensuring
		// that at worst, we get a geometric series, meaning that the
		// total memory usage is somewhere around
		// max(scratch buffer size) * 4.
		sizeOfLastScratchBuffer = size * 2;

		m_scratchBuffers.append(ScratchBuffer::create(sizeOfLastScratchBuffer));
	}

	ScratchBuffer* result = m_scratchBuffers.last();
	return result;
}

void JSGlobalData::clearScratchBuffers()
{
    //auto lock = holdLock(m_scratchBufferLock);
	for(int i = 0; i < m_scratchBuffers.size(); ++i)
		m_scratchBuffers[i]->setActiveLength(0);
}

} // namespace JSC
