/*
 * Copyright (C) 2008, 2009 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "AssemblerData.h"
#include "DateInstanceCache.h"
#include "Heap.h"
#include "Intrinsic.h"
#include "JITThunks.h"
#include "JSLock.h"
#include "JSValue.h"
#include "LLIntData.h"
#include "NumericStrings.h"
#include "PrivateName.h"
#include "PrototypeMap.h"
#include "SmallStrings.h"
#include "SourceCode.h"
#include "Strong.h"
#include "ThunkGenerators.h"
#include "TypedArrayController.h"
#include "Watchdog.h"
#include "Watchpoint.h"
#include <wtf/DateMath.h>
#include <wtf/Forward.h>
#include <wtf/HashMap.h>
#include <wtf/HashSet.h>
#include <wtf/OwnPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/RefPtr.h>
#include <wtf/SimpleStats.h>
#include <wtf/ThreadSafeRefCounted.h>
#include <wtf/ThreadSpecific.h>
#include <wtf/WeakRandom.h>
#include <wtf/WTFThreadData.h>
#include <wtf/text/SymbolRegistry.h>
#include <wtf/text/WTFString.h>
#if ENABLE(REGEXP_TRACING)
#include <wtf/ListHashSet.h>
#endif

namespace WTF {
class BumpPointerAllocator;
}
using WTF::BumpPointerAllocator;

namespace JSC {

class AssemblerData;
class BytecodeIntrinsicRegistry;
class CodeBlock;
class CommonIdentifiers;
class CompactTDZEnvironmentMap;
class CustomGetterSetter;
class ExecState;
class Exception;
class HandleStack;
class HasOwnPropertyCache;
class Identifier;
class Interpreter;
class JSBigInt;
class JSGlobalData;
class JSGlobalObject;
class JSObject;
class LLIntOffsetsExtractor;
class NativeExecutable;
class LegacyProfiler;
class RegExpCache;
class SourceProvider;
class SourceProviderCache;
struct StackFrame;
class Structure;
#if ENABLE(REGEXP_TRACING)
class RegExp;
#endif
class VMEntryScope;
class Symbol;
class Watchpoint;
class WatchpointSet;

#if ENABLE(DFG_JIT)
namespace DFG {
class LongLivedState;
}
#endif // ENABLE(DFG_JIT)

struct HashTable;
struct Instruction;

struct LocalTimeOffsetCache {
    LocalTimeOffsetCache()
        : start(0.0)
        , end(-1.0)
        , increment(0.0)
    {
    }

    void reset()
    {
        offset = LocalTimeOffset();
        start = 0.0;
        end = -1.0;
        increment = 0.0;
    }

    LocalTimeOffset offset;
    double start;
    double end;
    double increment;
};

#if ENABLE(JIT)
class ConservativeRoots;

#if COMPILER(MSVC)
#pragma warning(push)
#pragma warning(disable: 4200) // Disable "zero-sized array in struct/union" warning
#endif
struct ScratchBuffer {
	ScratchBuffer()
	{
        u.m_activeLength = 0;
	}

	static ScratchBuffer* create(size_t size)
	{
		ScratchBuffer* result = new (fastMalloc(ScratchBuffer::allocationSize(size))) ScratchBuffer;
		return result;
	}

    static ScratchBuffer* fromData(void* buffer)
    {
        return bitwise_cast<ScratchBuffer*>(static_cast<char*>(buffer) - OBJECT_OFFSETOF(ScratchBuffer, m_buffer));
    }

	static size_t allocationSize(size_t bufferSize) { return sizeof(ScratchBuffer) + bufferSize; }
	void setActiveLength(size_t activeLength) { u.m_activeLength = activeLength; }
	size_t activeLength() const { return u.m_activeLength; };
    size_t* addressOfActiveLength() { return &u.m_activeLength; };
	void* dataBuffer() { return m_buffer; }

    union {
        size_t m_activeLength;
        double pad; // Make sure m_buffer is double aligned.
    } u;
#if CPU(MIPS) && (defined WTF_MIPS_ARCH_REV && WTF_MIPS_ARCH_REV == 2)
	void* m_buffer[0] __attribute__((aligned(8)));
#else
	void* m_buffer[0];
#endif
};
#if COMPILER(MSVC)
#pragma warning(pop)
#endif
#endif

typedef JSGlobalData VM;

class VMIdleCallback
{
	WTF_MAKE_FAST_ALLOCATED;
public:
	virtual ~VMIdleCallback() { }

	virtual void operator()(VM&, JSGlobalObject*) = 0;

protected:
	VMIdleCallback() { }
};

class ActiveScratchBufferScope {
public:
    ActiveScratchBufferScope(ScratchBuffer*, size_t activeScratchBufferSizeInJSValues);
    ~ActiveScratchBufferScope();

private:
    ScratchBuffer* m_scratchBuffer;
};

class JSGlobalData : public ThreadSafeRefCounted<JSGlobalData> {
public:
	// WebCore has a one-to-one mapping of threads to JSGlobalDatas;
	// either create() or createLeaked() should only be called once
	// on a thread, this is the 'default' JSGlobalData (it uses the
	// thread's default string uniquing table from wtfThreadData).
	// API contexts created using the new context group aware interface
	// create APIContextGroup objects which require less locking of JSC
	// than the old singleton APIShared JSGlobalData created for use by
	// the original API.
	enum GlobalDataType { Default, APIContextGroup, APIShared };

    struct ClientData {
        JS_EXPORT_PRIVATE virtual ~ClientData() { };

        JS_EXPORT_PRIVATE virtual String overrideSourceURL(const StackFrame&, const String& originalSourceURL) const = 0;
    };

	bool isSharedInstance() { return globalDataType == APIShared; }
	bool usingAPI() { return globalDataType != Default; }
#if ENABLE(JAVA_BRIDGE) || OS(DARWIN)
	static bool sharedInstanceExists();
	JS_EXPORT_PRIVATE static JSGlobalData& sharedInstance();
#endif

	JS_EXPORT_PRIVATE static PassRefPtr<JSGlobalData> create(HeapType = SmallHeap);
	JS_EXPORT_PRIVATE static PassRefPtr<JSGlobalData> createLeaked(HeapType = SmallHeap);
	static PassRefPtr<JSGlobalData> createContextGroup(HeapType = SmallHeap);
	JS_EXPORT_PRIVATE ~JSGlobalData();

#if !PLATFORM(ESPIAL)
	void makeUsableFromMultipleThreads() { heap.machineThreads().makeUsableFromMultipleThreads(); }
#endif

private:
	JSLock m_apiLock;

public:
	// The heap should be just after executableAllocator and before other members to ensure that it's
	// destructed after all the objects that reference it.
	Heap heap;

#if ENABLE(DFG_JIT)
        OwnPtr<DFG::LongLivedState> m_dfgState;
#endif // ENABLE(DFG_JIT)

	GlobalDataType globalDataType;
	ClientData* clientData { nullptr };
	ExecState* topCallFrame;
	OwnPtr<Watchdog> watchdog;

	Strong<Structure> structureStructure;
    Strong<Structure> structureRareDataStructure;
	Strong<Structure> debuggerActivationStructure;
	Strong<Structure> terminatedExecutionErrorStructure;
	Strong<Structure> stringStructure;
	Strong<Structure> propertyNameIteratorStructure;
    Strong<Structure> propertyNameEnumeratorStructure;
    Strong<Structure> getterSetterStructure;
    Strong<Structure> customGetterSetterStructure;
    Strong<Structure> scopedArgumentsTableStructure;
	Strong<Structure> apiWrapperStructure;
	Strong<Structure> nativeExecutableStructure;
	Strong<Structure> evalExecutableStructure;
	Strong<Structure> programExecutableStructure;
	Strong<Structure> functionExecutableStructure;
    Strong<Structure> moduleProgramExecutableStructure;
	Strong<Structure> regExpStructure;
    Strong<Structure> symbolStructure;
	Strong<Structure> sharedSymbolTableStructure;
    Strong<Structure> fixedArrayStructure;
    Strong<Structure> sourceCodeStructure;
    Strong<Structure> scriptFetcherStructure;
    Strong<Structure> scriptFetchParametersStructure;
	Strong<Structure> structureChainStructure;
	Strong<Structure> sparseArrayValueMapStructure;
    Strong<Structure> templateObjectDescriptorStructure;
    Strong<Structure> propertyTableStructure;
    Strong<Structure> weakMapDataStructure;
    Strong<Structure> functionRareDataStructure;
    Strong<Structure> exceptionStructure;
    Strong<Structure> inferredValueStructure;
    Strong<Structure> promiseDeferredStructure;
    Strong<Structure> internalPromiseDeferredStructure;
    Strong<Structure> nativeStdFunctionCellStructure;
    Strong<Structure> hashMapBucketSetStructure;
    Strong<Structure> hashMapBucketMapStructure;
    Strong<Structure> setIteratorStructure;
    Strong<Structure> mapIteratorStructure;
    Strong<Structure> bigIntStructure;

    Strong<JSCell> emptyPropertyNameEnumerator;
    Strong<JSCell> sentinelSetBucket;
    Strong<JSCell> sentinelMapBucket;

    JSCell* currentlyDestructingCallbackObject;
    const ClassInfo* currentlyDestructingCallbackObjectClassInfo;

	IdentifierTable* identifierTable;
    WTF::SymbolRegistry m_symbolRegistry;
    WTF::SymbolRegistry m_privateSymbolRegistry;
	CommonIdentifiers* propertyNames;
    const ArgList* emptyList;
    SmallStrings smallStrings;
	NumericStrings numericStrings;
	DateInstanceCache dateInstanceCache;
	WTF::SimpleStats machineCodeBytesPerBytecodeWordForBaselineJIT;

    IdentifierTable& atomicStringTable() const { return *identifierTable; }
    WTF::SymbolRegistry& symbolRegistry() { return m_symbolRegistry; }
    WTF::SymbolRegistry& privateSymbolRegistry() { return m_privateSymbolRegistry; }

    WeakGCMap<SymbolImpl*, Symbol, PtrHash<SymbolImpl*>> symbolImplToSymbolMap;

    Strong<JSBigInt> heapBigIntConstantOne;

    enum class DeletePropertyMode {
        // Default behaviour of deleteProperty, matching the spec.
        Default,
        // This setting causes deleteProperty to force deletion of all
        // properties including those that are non-configurable (DontDelete).
        IgnoreConfigurable
    };

    DeletePropertyMode deletePropertyMode()
    {
        return m_deletePropertyMode;
    }

    class DeletePropertyModeScope {
    public:
        DeletePropertyModeScope(JSGlobalData& vm, DeletePropertyMode mode)
            : m_vm(vm)
            , m_previousMode(vm.m_deletePropertyMode)
        {
            m_vm.m_deletePropertyMode = mode;
        }

        ~DeletePropertyModeScope()
        {
            m_vm.m_deletePropertyMode = m_previousMode;
        }

    private:
        JSGlobalData& m_vm;
        DeletePropertyMode m_previousMode;
    };

	LegacyProfiler* enabledProfiler()
	{
		return m_enabledProfiler;
	}

#if ENABLE(JIT) && ENABLE(LLINT)
	bool canUseJIT() { return m_canUseJIT; }
#elif ENABLE(JIT)
	bool canUseJIT() { return true; } // jit only
#else
	bool canUseJIT() { return false; } // interpreter only
#endif

    SourceProviderCache* addSourceProviderCache(SourceProvider*);
    void clearSourceProviderCaches();

	PrototypeMap prototypeMap;

	typedef WeakGCMap<StringImpl*, JSString, PtrHash<StringImpl*>> JSStringCache;
	JSStringCache stringCache;
    Strong<JSString> lastCachedString;

    typedef HashMap<RefPtr<SourceProvider>, RefPtr<SourceProviderCache> > SourceProviderCacheMap;
    SourceProviderCacheMap sourceProviderCacheMap;
	Interpreter* interpreter;
#if ENABLE(JIT)
	OwnPtr<JITThunks> jitStubs;
	MacroAssemblerCodeRef getCTIStub(ThunkGenerator generator)
	{
		return jitStubs->ctiStub(this, generator);
	}

    typedef EncodedJSValue(*CallJavaScriptJITFunction)(void *, ExecState*);

    CallJavaScriptJITFunction callJavaScriptJITFunction;
#endif
	NativeExecutable* getHostFunction(NativeFunction, NativeFunction constructor, const String& name);
	NativeExecutable* getHostFunction(NativeFunction, Intrinsic intrinsic, NativeFunction constructor, const String& name);

    static ptrdiff_t exceptionOffset()
    {
        return OBJECT_OFFSETOF(JSGlobalData, m_exception);
    }

    static ptrdiff_t callFrameForCatchOffset()
    {
        return OBJECT_OFFSETOF(JSGlobalData, callFrameForCatch);
    }

    static ptrdiff_t targetMachinePCForThrowOffset()
    {
        return OBJECT_OFFSETOF(JSGlobalData, targetMachinePCForThrow);
    }

    void restorePreviousException(Exception* exception) { setException(exception); }

    void clearException() { m_exception = 0; }
    void clearLastException() { m_lastException = 0; }

    ExecState** addressOfCallFrameForCatch() { return &callFrameForCatch; }

    Exception* exception() const { return m_exception; }
    JSCell** addressOfException() { return reinterpret_cast<JSCell**>(&m_exception); }

    Exception* lastException() const { return m_lastException; }
    JSCell** addressOfLastException() { return reinterpret_cast<JSCell**>(&m_lastException); }

    JS_EXPORT_PRIVATE Exception* throwException(ExecState*, Exception*);
    JS_EXPORT_PRIVATE Exception* throwException(ExecState*, JSValue);
    JS_EXPORT_PRIVATE Exception* throwException(ExecState*, JSObject*);

    JS_EXPORT_PRIVATE void throwOOMException(Exception*);

    void** addressOfLastStackTop() { return &m_lastStackTop; }
    void* lastStackTop() { return m_lastStackTop; }
    void setLastStackTop(void* lastStackTop) { m_lastStackTop = lastStackTop; }

	const ClassInfo* const jsArrayClassInfo;
	const ClassInfo* const jsFinalObjectClassInfo;

	LLInt::Data llintData;

	JSValue hostCallReturnValue;
    unsigned varargsLength;
    ExecState* newCallFrameReturnValue;
	ExecState* callFrameForCatch;
	void* targetMachinePCForThrow;
	Instruction* targetInterpreterPCForThrow;
#if ENABLE(JIT)
	uint32_t osrExitIndex;
	void* osrExitJumpDestination;
	Vector<ScratchBuffer*> m_scratchBuffers;
	size_t sizeOfLastScratchBuffer;

    // - You can call scratchBufferForSize from any thread.
    // - You can only set the ScratchBuffer's activeLength from the main thread.
    // - You can only write to entries in the ScratchBuffer from the main thread.
	ScratchBuffer* scratchBufferForSize(size_t size);
    void clearScratchBuffers();

	void gatherConservativeRoots(ConservativeRoots&, bool noStackObjects = false);
#endif

    VMEntryScope* entryScope;

	double executionStartTime;
	double jitCompileLength;
	double parseLength;
	double evalPrepareForExecutionLength;
	unsigned evalCount;

    void* m_stackLimit;

    JSObject* stringRecursionCheckFirstObject;;
	HashSet<JSObject*> stringRecursionCheckVisitedObjects;

    LocalTimeOffsetCache utcTimeOffsetCache;
    LocalTimeOffsetCache localTimeOffsetCache;

	String cachedDateString;
	double cachedDateStringValue;

	LegacyProfiler* m_enabledProfiler;
    RefPtr<TypedArrayController> m_typedArrayController;
	RegExpCache* m_regExpCache;
	OwnPtr<BumpPointerAllocator> m_regExpAllocator;

#if ENABLE(YARR_JIT_ALL_PARENS_EXPRESSIONS)
    static constexpr size_t patternContextBufferSize = 8192; // Space allocated to save nested parenthesis context
    char* m_regExpPatternContexBuffer { nullptr };
    Lock m_regExpPatternContextLock;
    char* acquireRegExpPatternContexBuffer();
    void releaseRegExpPatternContexBuffer();
#else
    static constexpr size_t patternContextBufferSize = 0; // Space allocated to save nested parenthesis context
#endif

	OwnPtr<HasOwnPropertyCache> m_hasOwnPropertyCache;
    ALWAYS_INLINE HasOwnPropertyCache* hasOwnPropertyCache() { return m_hasOwnPropertyCache.get(); }
    HasOwnPropertyCache* ensureHasOwnPropertyCache();

#if ENABLE(REGEXP_TRACING)
	typedef ListHashSet<RefPtr<RegExp> > RTTraceList;
	RTTraceList* m_rtTraceList;
#endif

#ifndef NDEBUG
	ThreadIdentifier exclusiveThread;
#endif

    Ref<CompactTDZEnvironmentMap> m_compactVariableMap;

	JS_EXPORT_PRIVATE void resetDateCache();

	JS_EXPORT_PRIVATE void startSampling();
	JS_EXPORT_PRIVATE void stopSampling();
	JS_EXPORT_PRIVATE void dumpSampleData(ExecState* exec);
	RegExpCache* regExpCache() { return m_regExpCache; }
#if ENABLE(REGEXP_TRACING)
	void addRegExpToTrace(PassRefPtr<RegExp> regExp);
#endif
	JS_EXPORT_PRIVATE void dumpRegExpTrace();

	bool isCollectorBusy() const { return heap.isBusy(); }

#if ENABLE(GC_VALIDATION)
	bool isInitializingObject() const;
	void setInitializingObjectClass(const ClassInfo*);
#endif

	JSLock& apiLock() { return m_apiLock; }

    void prepareToDeleteCode();

    void whenIdle(PassOwnPtr<VMIdleCallback>);

    JS_EXPORT_PRIVATE void deleteAllCode();

    WatchpointSet* ensureWatchpointSetForImpureProperty(const Identifier&);

    // FIXME: Use AtomicString once it got merged with Identifier.
    JS_EXPORT_PRIVATE void addImpureProperty(const String&);

    void purgeWatchpointSetForImpurePropertyIfNecessary();

    JS_EXPORT_PRIVATE void setGlobalConstRedeclarationShouldThrow(bool globalConstRedeclarationThrow) { m_globalConstRedeclarationShouldThrow = globalConstRedeclarationThrow; }
    ALWAYS_INLINE bool globalConstRedeclarationShouldThrow() const { return m_globalConstRedeclarationShouldThrow; }

    BytecodeIntrinsicRegistry& bytecodeIntrinsicRegistry() { return *m_bytecodeIntrinsicRegistry; }

private:
	friend class LLIntOffsetsExtractor;
    friend class ClearExceptionScope;

	JSGlobalData(GlobalDataType, HeapType);
#if ENABLE(JAVA_BRIDGE) || OS(DARWIN)
	static JSGlobalData*& sharedInstanceInternal();
#endif
	void createNativeThunk();

    void setException(Exception* exception)
    {
        m_exception = exception;
        m_lastException = exception;
    }

#if ENABLE(ASSEMBLER)
	bool m_canUseAssembler;
#endif
#if ENABLE(GC_VALIDATION)
	const ClassInfo* m_initializingObjectClass;
#endif
    void* m_lastStackTop;
    Exception* m_exception;
    Exception* m_lastException;
    bool m_globalConstRedeclarationShouldThrow;
    DeletePropertyMode m_deletePropertyMode;

    HashMap<String, RefPtr<WatchpointSet>> m_impurePropertyWatchpointSets;
    double m_lastTimePurgeWatchpointSetForImpureProperty;

    OwnPtr<BytecodeIntrinsicRegistry> m_bytecodeIntrinsicRegistry;
};

void sanitizeStackForVM(JSGlobalData&);

#if ENABLE(GC_VALIDATION)
inline bool JSGlobalData::isInitializingObject() const
{
	return !!m_initializingObjectClass;
}

inline void JSGlobalData::setInitializingObjectClass(const ClassInfo* initializingObjectClass)
{
	m_initializingObjectClass = initializingObjectClass;
}
#endif

inline Heap* MarkedBlock::heap() const
{
	return &m_vm->heap;
}

inline ActiveScratchBufferScope::ActiveScratchBufferScope(ScratchBuffer* buffer, size_t activeScratchBufferSizeInJSValues)
    : m_scratchBuffer(buffer)
{
    // Tell GC mark phase how much of the scratch buffer is active during the call operation this scope is used in.
    if (m_scratchBuffer)
        m_scratchBuffer->setActiveLength(activeScratchBufferSizeInJSValues * sizeof(EncodedJSValue));
}

inline ActiveScratchBufferScope::~ActiveScratchBufferScope()
{
    // Tell the GC that we're not using the scratch buffer anymore.
    if (m_scratchBuffer)
        m_scratchBuffer->setActiveLength(0);
}

} // namespace JSC
