/*
 * Copyright (C) 2007, 2008, 2009 Apple Inc. All rights reserved.
 * Copyright (C) 2008 Cameron Zwarich (cwzwarich@uwaterloo.ca)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "JSGlobalObject.h"

#include "AggregateError.h"
#include "AggregateErrorConstructor.h"
#include "AggregateErrorPrototype.h"
#include "ArrayConstructor.h"
#include "ArrayIteratorPrototype.h"
#include "ArrayPrototype.h"
#include "AsyncFromSyncIteratorPrototype.h"
#include "AsyncFunctionConstructor.h"
#include "AsyncFunctionPrototype.h"
#include "AsyncGeneratorFunctionConstructor.h"
#include "AsyncGeneratorFunctionPrototype.h"
#include "AsyncGeneratorPrototype.h"
#include "AsyncIteratorPrototype.h"
#include "BigIntConstructor.h"
#include "BigIntObject.h"
#include "BigIntPrototype.h"
#include "BooleanConstructor.h"
#include "BooleanPrototype.h"
#include "BuiltinExecutables.h"
#include "BuiltinNames.h"
#include "ClonedArguments.h"
#include "CodeBlock.h"
#include "CodeBlockSet.h"
#include "Completion.h"
#include "ConsoleObject.h"
#include "DateConstructor.h"
#include "DatePrototype.h"
#include "Debugger.h"
#include "DirectArguments.h"
#include "Error.h"
#include "ErrorConstructor.h"
#include "ErrorPrototype.h"
#include "FunctionConstructor.h"
#include "FunctionPrototype.h"
#include "GeneratorFunctionConstructor.h"
#include "GeneratorFunctionPrototype.h"
#include "GeneratorPrototype.h"
#include "GetterSetter.h"
#include "HeapIterationScope.h"
#include "Interpreter.h"
#include "IteratorPrototype.h"
#include "IterationStatus.h"
#include "JSLexicalEnvironment.h"
#include "JSArrayBuffer.h"
#include "JSArrayBufferConstructor.h"
#include "JSArrayBufferPrototype.h"
#include "JSArrayIterator.h"
#include "JSAsyncFunction.h"
#include "JSAsyncGeneratorFunction.h"
#include "JSBigInt.h"
#include "JSBoundFunction.h"
#if !PLATFORM(ESPIAL)
#include "JSCallbackConstructor.h"
#include "JSCallbackFunction.h"
#include "JSCallbackObject.h"
#endif
#include "JSCustomGetterFunction.h"
#include "JSCustomSetterFunction.h"
#include "JSDataView.h"
#include "JSDataViewPrototype.h"
#include "JSFunction.h"
#include "JSGeneratorFunction.h"
#include "JSGenericTypedArrayViewConstructorInlines.h"
#include "JSGenericTypedArrayViewInlines.h"
#include "JSGenericTypedArrayViewPrototypeInlines.h"
#include "JSGlobalObjectFunctions.h"
#include "JSInternalPromise.h"
#include "JSInternalPromiseConstructor.h"
#include "JSInternalPromisePrototype.h"
#include "JSLock.h"
#include "JSMap.h"
#include "JSMicrotask.h"
#include "JSModuleEnvironment.h"
#include "JSModuleLoader.h"
#include "JSModuleNamespaceObject.h"
#include "JSModuleRecord.h"
#include "JSNativeStdFunction.h"
#include "JSONObject.h"
#include "JSSet.h"
#include "JSPromise.h"
#include "JSPromiseConstructor.h"
#include "JSPromisePrototype.h"
#include "JSStringIterator.h"
#include "JSTypedArrayConstructors.h"
#include "JSTypedArrayPrototypes.h"
#include "JSTypedArrayViewConstructor.h"
#include "JSTypedArrayViewPrototype.h"
#include "JSTypedArrays.h"
#include "JSValueInlines.h"
#include "JSWeakMap.h"
#include "JSWeakSet.h"
#include "JSWithScope.h"
#include "Lookup.h"
#include "MapConstructor.h"
#include "MapIteratorPrototype.h"
#include "MapPrototype.h"
#include "MathObject.h"
#include "Microtask.h"
#include "NativeErrorConstructor.h"
#include "NativeErrorPrototype.h"
#include "NullGetterFunction.h"
#include "NullSetterFunction.h"
#include "NumberConstructor.h"
#include "NumberPrototype.h"
#include "ObjectConstructor.h"
#include "ObjectPropertyConditionSet.h"
#include "ObjectPrototype.h"
#include "Operations.h"
#include "Profiler.h"
#include "ProxyConstructor.h"
#include "ProxyObject.h"
#include "ProxyRevoke.h"
#include "ReflectObject.h"
#include "RegExpConstructor.h"
#include "RegExpMatchesArray.h"
#include "RegExpObject.h"
#include "RegExpPrototype.h"
#include "RegExpStringIteratorPrototype.h"
#include "ScopedArguments.h"
#include "SetConstructor.h"
#include "SetIteratorPrototype.h"
#include "SetPrototype.h"
#include "StrictEvalActivation.h"
#include "StringConstructor.h"
#include "StringIteratorPrototype.h"
#include "StringPrototype.h"
#include "Symbol.h"
#include "SymbolConstructor.h"
#include "SymbolPrototype.h"
#include "VariableWriteFireDetail.h"
#include "WeakGCMapInlines.h"
#include "WeakMapConstructor.h"
#include "WeakMapPrototype.h"
#include "WeakSetConstructor.h"
#include "WeakSetPrototype.h"
#include <wtf/CurrentTime.h>
#include <wtf/Deque.h>
#include <wtf/Vector.h>

#if ENABLE(INTL)
#include "IntlCollator.h"
#include "IntlObject.h"
#include <unicode/ucol.h>
#include <unicode/udat.h>
#include <unicode/unum.h>
#endif // ENABLE(INTL)

#include "JSGlobalObject.lut.h"

namespace JSC {

const ClassInfo JSGlobalObject::s_info = { "GlobalObject", &Base::s_info, &globalObjectTable, CREATE_METHOD_TABLE(JSGlobalObject) };

const GlobalObjectMethodTable JSGlobalObject::s_globalObjectMethodTable = {
	&supportsProfiling,
	&supportsRichSourceInfo,
	&shouldInterruptScript,
	&javaScriptExperimentsEnabled,
    0, // queueTaskToEventLoop
    0, // moduleLoaderImportModule
    0, // moduleLoaderResolve
    0, // moduleLoaderFetch
    0, // moduleLoaderCreateImportMetaProperties
    0, // moduleLoaderEvaluate
    0, // promiseRejectionTracker
    0, // defaultLanguage
};

/* Source for JSGlobalObject.lut.h
@begin globalObjectTable
  isNaN                 JSBuiltin                       DontEnum|Function 1
  isFinite              JSBuiltin                       DontEnum|Function 1
  escape                globalFuncEscape                DontEnum|Function 1
  unescape              globalFuncUnescape              DontEnum|Function 1
  decodeURI             globalFuncDecodeURI             DontEnum|Function 1
  decodeURIComponent    globalFuncDecodeURIComponent    DontEnum|Function 1
  encodeURI             globalFuncEncodeURI             DontEnum|Function 1
  encodeURIComponent    globalFuncEncodeURIComponent    DontEnum|Function 1
@end
*/

static EncodedJSValue JSC_HOST_CALL enqueueJob(ExecState* exec)
{
    VM& vm = exec->vm();
    JSGlobalObject* globalObject = exec->lexicalGlobalObject();

    JSValue job = exec->argument(0);
    JSValue arguments = exec->argument(1);
    ASSERT(arguments.inherits(JSArray::info()));

    globalObject->queueMicrotask(createJSMicrotask(vm, job, jsCast<JSArray*>(arguments)));

    return JSValue::encode(jsUndefined());
}

JSGlobalObject::JSGlobalObject(JSGlobalData& vm, Structure* structure, const GlobalObjectMethodTable* globalObjectMethodTable)
	: Base(vm, structure, nullptr)
	, m_functionNameOffset(0)
	, m_structureCache(vm)
	, m_debugger(nullptr)
	, m_vm(vm)
	, m_masqueradesAsUndefinedWatchpoint(WatchpointSet::create(IsWatched))
	, m_havingABadTimeWatchpoint(WatchpointSet::create(IsWatched))
	, m_varInjectionWatchpoint(WatchpointSet::create(IsWatched))
	, m_varReadOnlyWatchpoint(WatchpointSet::create(IsWatched))
	, m_weakRandom(Options::forceWeakRandomSeed() ? Options::forcedWeakRandomSeed() : static_cast<unsigned>(randomNumber() * (std::numeric_limits<unsigned>::max() + 1.0)))
	, m_arrayIteratorProtocolWatchpointSet(IsWatched)
	, m_mapIteratorProtocolWatchpoint(IsWatched)
	, m_setIteratorProtocolWatchpoint(IsWatched)
	, m_stringIteratorProtocolWatchpoint(IsWatched)
	, m_mapSetWatchpoint(IsWatched)
	, m_setAddWatchpoint(IsWatched)
	, m_arraySpeciesWatchpointSet(ClearWatchpoint)
	, m_arrayJoinWatchpointSet(IsWatched)
	, m_numberToStringWatchpointSet(IsWatched)
	, m_structureCacheClearedWatchpoint(IsWatched)
	, m_arrayBufferSpeciesWatchpointSet(ClearWatchpoint)
	, m_sharedArrayBufferSpeciesWatchpointSet(ClearWatchpoint)
	, m_evalEnabled(true)
	, m_globalLexicalBindingEpoch(1)
	, m_experimentsEnabled(true)
	, m_consoleClient(nullptr)
	, m_customGetterFunctionMap(vm)
	, m_customSetterFunctionMap(vm)
	, m_globalObjectMethodTable(globalObjectMethodTable ? globalObjectMethodTable : &s_globalObjectMethodTable)
	, m_needsSiteSpecificQuirks(false)
{
	m_builtinExecutables = BuiltinExecutables::create(vm);
}

JSGlobalObject::~JSGlobalObject()
{
    if (m_debugger)
        m_debugger->detach(this);

    if (LegacyProfiler* profiler = globalData().enabledProfiler())
        profiler->stopProfiling(this);
}

void JSGlobalObject::destroy(JSCell* cell)
{
    static_cast<JSGlobalObject*>(cell)->JSGlobalObject::~JSGlobalObject();
}

void JSGlobalObject::setGlobalThis(JSGlobalData& vm, JSObject* globalThis)
{
    m_globalThis.set(vm, this, globalThis);
    putDirect(vm, vm.propertyNames->globalThis, m_globalThis.get(), DontEnum);
}

static GetterSetter* getGetterById(VM& vm, ExecState* exec, JSObject* base, const Identifier& ident)
{
    JSValue baseValue = JSValue(base);
    PropertySlot slot(baseValue, PropertySlot::InternalMethodType::VMInquiry);
    baseValue.getPropertySlot(vm, exec, ident, slot);
    return jsCast<GetterSetter*>(slot.getPureResult());
}

void JSGlobalObject::init(JSGlobalData& vm)
{
    ASSERT(vm.apiLock().currentThreadIsHoldingLock());

    convertToDictionary(vm);

    JSGlobalObject::globalExec()->init(nullptr, nullptr, CallFrame::noCaller(), 0, 0);

    m_debugger = nullptr;

#if ENABLE(REMOTE_INSPECTOR)
    m_inspectorController = std::make_unique<Inspector::JSGlobalObjectInspectorController>(*this);
    m_inspectorDebuggable = std::make_unique<JSGlobalObjectDebuggable>(*this);
    m_inspectorDebuggable->init();
    m_inspectorDebuggable->setRemoteDebuggingAllowed(true);
    m_consoleClient = m_inspectorController->consoleClient();
#endif

    reset(getPrototypeDirect());
}

bool JSGlobalObject::put(JSCell* cell, ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    JSGlobalObject* thisObject = jsCast<JSGlobalObject*>(cell);
    VM& vm = thisObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    ASSERT(!Heap::heap(value) || Heap::heap(value) == Heap::heap(thisObject));

    if (UNLIKELY(isThisValueAltered(slot, thisObject))) {
        SymbolTableEntry::Fast entry = thisObject->symbolTable()->get(propertyName.uid());
        if (!entry.isNull()) {
            if (entry.isReadOnly())
                return typeError(exec, vm, slot.isStrictMode(), ReadonlyPropertyWriteError);
            RELEASE_AND_RETURN(scope, JSObject::definePropertyOnReceiver(exec, propertyName, value, slot));
        }
        RELEASE_AND_RETURN(scope, Base::put(thisObject, exec, propertyName, value, slot));
    }
    bool shouldThrowReadOnlyError = slot.isStrictMode();
    bool ignoreReadOnlyErrors = false;
    bool putResult = false;
    if (symbolTablePutTouchWatchpointSet(thisObject, exec, propertyName, value, shouldThrowReadOnlyError, ignoreReadOnlyErrors, putResult))
        return putResult;
    RELEASE_AND_RETURN(scope, Base::put(thisObject, exec, propertyName, value, slot));
}

bool JSGlobalObject::defineOwnProperty(JSObject* object, ExecState* exec, PropertyName propertyName, const PropertyDescriptor& descriptor, bool shouldThrow)
{
    JSGlobalObject* thisObject = jsCast<JSGlobalObject*>(object);
    VM& vm = thisObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    SymbolTableEntry entry;
    PropertyDescriptor currentDescriptor;
    if (symbolTableGet(thisObject, propertyName, entry, currentDescriptor)) {
        bool isExtensible = false; // ignored since current descriptor is present
        bool isCurrentDefined = true;
        bool isCompatibleDescriptor = validateAndApplyPropertyDescriptor(exec, nullptr, propertyName, isExtensible, descriptor, isCurrentDefined, currentDescriptor, shouldThrow);
        RETURN_IF_EXCEPTION(scope, false);
		if (!isCompatibleDescriptor)
            return false;

        if (descriptor.value()) {
            bool ignoreReadOnlyErrors = true;
            bool putResult = false;
            if (symbolTablePutTouchWatchpointSet(thisObject, exec, propertyName, descriptor.value(), shouldThrow, ignoreReadOnlyErrors, putResult))
                ASSERT(putResult);
            scope.assertNoException();
        }
        if (descriptor.writablePresent() && !descriptor.writable() && !entry.isReadOnly()) {
        	entry.setReadOnly();
			thisObject->symbolTable()->set(propertyName.uid(), entry);
			thisObject->varReadOnlyWatchpoint()->fireAll(vm, "GlobalVar was redefined as ReadOnly");
        }
        return true;
    }

    scope.release();
    return Base::defineOwnProperty(thisObject, exec, propertyName, descriptor, shouldThrow);
}

// https://tc39.es/ecma262/#sec-candeclareglobalfunction
bool JSGlobalObject::canDeclareGlobalFunction(ExecState* exec, const Identifier& ident)
{
    auto scope = DECLARE_THROW_SCOPE(vm());

    PropertySlot slot(this, PropertySlot::InternalMethodType::GetOwnProperty);
    bool hasProperty = getOwnPropertySlot(this, exec, ident, slot);
    //scope.assertNoExceptionExceptTermination();
    if (LIKELY(!hasProperty))
        return isStructureExtensible();

    bool isConfigurable = !(slot.attributes() & PropertyAttribute::DontDelete);
    if (isConfigurable)
        return true;
    bool isDataDescriptor = !(slot.attributes() & (PropertyAttribute::Accessor | PropertyAttribute::CustomAccessor));
    bool isWritableAndEnumerable = !(slot.attributes() & (PropertyAttribute::ReadOnly | PropertyAttribute::DontEnum));
    return isDataDescriptor && isWritableAndEnumerable;
}

// https://tc39.es/ecma262/#sec-createglobalfunctionbinding
template<BindingCreationContext context>
void JSGlobalObject::createGlobalFunctionBinding(ExecState* exec, const Identifier& ident)
{
    VM& vm = this->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    PropertySlot slot(this, PropertySlot::InternalMethodType::GetOwnProperty);
    bool hasProperty = getOwnPropertySlot(this, exec, ident, slot);
    //scope.assertNoExceptionExceptTermination();
    if (UNLIKELY(hasProperty)) {
        if (slot.attributes() & PropertyAttribute::DontDelete) {
            ASSERT(!(slot.attributes() & PropertyAttribute::ReadOnly));
            // Nothing to do here: there is either a symbol table entry or non-configurable writable property
            // on the structure that will be updated with real function by put_to_scope.
        } else {
            unsigned newAttributes = 0;
            if (context == BindingCreationContext::Global)
                newAttributes |= PropertyAttribute::DontDelete;
            putDirect(vm, ident, jsUndefined(), newAttributes);
        }
    } else {
        ASSERT(isStructureExtensible());
        if (context == BindingCreationContext::Global)
            addSymbolTableEntry(ident);
        else
            putDirect(vm, ident, jsUndefined());
    }
}

template void JSGlobalObject::createGlobalFunctionBinding<BindingCreationContext::Global>(ExecState*, const Identifier&);
template void JSGlobalObject::createGlobalFunctionBinding<BindingCreationContext::Eval>(ExecState*, const Identifier&);

void JSGlobalObject::addSymbolTableEntry(const Identifier& ident)
{
    //ConcurrentJSLocker locker(symbolTable()->m_lock);
    ASSERT(!symbolTable()->contains(locker, ident.impl()));

    ScopeOffset offset = symbolTable()->takeNextScopeOffset();
    SymbolTableEntry newEntry(VarOffset(offset), 0);
	newEntry.prepareToWatch();
    symbolTable()->add(ident.impl(), WTFMove(newEntry));

    ScopeOffset offsetForAssert = addVariables(1, jsUndefined());
    RELEASE_ASSERT(offsetForAssert == offset);
}

static inline JSObject* lastInPrototypeChain(JSObject* object)
{
    JSObject* o = object;
    while (o->getPrototypeDirect().isObject())
        o = asObject(o->getPrototypeDirect());
    return o;
}

static EncodedJSValue JSC_HOST_CALL functionFalse1(ExecState*) { return JSValue::encode(jsBoolean(false)); }
static EncodedJSValue JSC_HOST_CALL functionFalse2(ExecState*) { return JSValue::encode(jsBoolean(false)); }

static EncodedJSValue JSC_HOST_CALL functionUndefined1(ExecState*) { return JSValue::encode(jsUndefined()); }
static EncodedJSValue JSC_HOST_CALL functionUndefined2(ExecState*) { return JSValue::encode(jsUndefined()); }

static EncodedJSValue JSC_HOST_CALL functionCreateProxy(ExecState* exec)
{
    JSLockHolder lock(exec);
    JSValue target = exec->argument(0);
    if (!target.isObject())
        return JSValue::encode(jsUndefined());
    JSObject* jsTarget = asObject(target.asCell());
    Structure* structure = JSProxy::createStructure(exec->vm(), exec->lexicalGlobalObject(), jsTarget->getPrototypeDirect(), ImpureProxyType);
    JSProxy* proxy = JSProxy::create(exec->vm(), structure, jsTarget);
    return JSValue::encode(proxy);
}

static EncodedJSValue JSC_HOST_CALL makeBoundFunction(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    JSGlobalObject* globalObject = exec->lexicalGlobalObject();

    JSObject* target = asObject(exec->uncheckedArgument(0));
    JSValue boundThis = exec->uncheckedArgument(1);
    JSValue boundArgs = exec->uncheckedArgument(2);
    JSValue lengthValue = exec->uncheckedArgument(3);
    JSString* nameString = asString(exec->uncheckedArgument(4));

    ASSERT(lengthValue.isInt32AsAnyInt());
    int32_t length = lengthValue.asInt32AsAnyInt();

    String name = nameString->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    return JSValue::encode(JSBoundFunction::create(vm, exec, globalObject, target, boundThis, boundArgs.isCell() ? jsCast<JSArray*>(boundArgs) : 0, length, name));
}

static EncodedJSValue JSC_HOST_CALL hasOwnLengthProperty(ExecState* exec)
{
    VM& vm = exec->vm();
    JSObject* target = asObject(exec->uncheckedArgument(0));
    return JSValue::encode(jsBoolean(target->hasOwnProperty(exec, vm.propertyNames->length)));
}

// FIXME: use a bytecode or intrinsic for creating a private symbol.
// https://bugs.webkit.org/show_bug.cgi?id=212782
static EncodedJSValue JSC_HOST_CALL createPrivateSymbol(ExecState* exec)
{
    VM& vm = exec->vm();
    return JSValue::encode(Symbol::create(vm, PrivateSymbolImpl::createNullSymbol().get()));
}

namespace
{
class Masquerader : public JSNonFinalObject {
public:
    Masquerader(VM& vm, Structure* structure)
        : Base(vm, structure)
    {
    }

    typedef JSNonFinalObject Base;
    static const unsigned StructureFlags = Base::StructureFlags | JSC::MasqueradesAsUndefined;

    static Masquerader* create(VM& vm, JSGlobalObject* globalObject)
    {
        globalObject->masqueradesAsUndefinedWatchpoint()->fireAll(vm);
        Structure* structure = createStructure(vm, globalObject, jsNull());
        Masquerader* result = new (NotNull, allocateCell<Masquerader>(vm.heap, sizeof(Masquerader))) Masquerader(vm, structure);
        result->finishCreation(vm);
        return result;
    }

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

    DECLARE_INFO;
};

class ImpureGetter : public JSNonFinalObject {
public:
    ImpureGetter(VM& vm, Structure* structure)
        : Base(vm, structure)
    {
    }

    DECLARE_INFO;
    typedef JSNonFinalObject Base;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

    static ImpureGetter* create(VM& vm, Structure* structure, JSObject* delegate)
    {
        ImpureGetter* getter = new (NotNull, allocateCell<ImpureGetter>(vm.heap, sizeof(ImpureGetter))) ImpureGetter(vm, structure);
        getter->finishCreation(vm, delegate);
        return getter;
    }

    void finishCreation(VM& vm, JSObject* delegate)
    {
        Base::finishCreation(vm);
        if (delegate)
            m_delegate.set(vm, this, delegate);
    }

    static const unsigned StructureFlags = JSC::GetOwnPropertySlotIsImpure | JSC::OverridesGetOwnPropertySlot | Base::StructureFlags;

    static bool getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName name, PropertySlot& slot)
    {
        ImpureGetter* thisObject = jsCast<ImpureGetter*>(object);

        if (thisObject->m_delegate) {
        	VM& vm = exec->vm();
        	if(thisObject->m_delegate->getPropertySlot(vm, exec, name, slot))
        		return true;

        	RETURN_IF_EXCEPTION(vm, false);
        }

        return Base::getOwnPropertySlot(object, exec, name, slot);
    }

    static void visitChildren(JSCell* cell, SlotVisitor& visitor)
    {
        Base::visitChildren(cell, visitor);
        ImpureGetter* thisObject = jsCast<ImpureGetter*>(cell);
        visitor.append(&thisObject->m_delegate);
    }

    void setDelegate(VM& vm, JSObject* delegate)
    {
        m_delegate.set(vm, this, delegate);
    }

private:
    WriteBarrier<JSObject> m_delegate;
};

static EncodedJSValue testStaticAccessorGetter(ExecState* exec, EncodedJSValue thisValue, PropertyName)
{
    //DollarVMAssertScope assertScope;
    VM& vm = exec->vm();

    JSObject* thisObject = jsDynamicCast<JSObject*>(JSValue::decode(thisValue));
    if (!thisObject)
		return throwVMTypeError(exec, vm);

    if (JSValue result = thisObject->getDirect(vm, PropertyName(Identifier::fromString(vm, "testField"))))
        return JSValue::encode(result);
    return JSValue::encode(jsUndefined());
}

static bool testStaticAccessorPutter(ExecState* exec, EncodedJSValue thisValue, EncodedJSValue value, PropertyName)
{
    //DollarVMAssertScope assertScope;
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue receiver = JSValue::decode(thisValue);
    JSObject* thisObject = receiver.isObject() ? asObject(receiver) : receiver.synthesizePrototype(exec);
    RETURN_IF_EXCEPTION(scope, false);
    RELEASE_ASSERT(thisObject);

    return thisObject->putDirect(vm, PropertyName(Identifier::fromString(vm, "testField")), JSValue::decode(value));
}

static const struct CompactHashIndex staticCustomAccessorTableIndex[9] = {
    { -1, -1 },
    { -1, -1 },
    { 2, -1 },
    { -1, -1 },
    { 0, 8 },
    { -1, -1 },
    { -1, -1 },
    { -1, -1 },
    { 1, -1 },
};

static const struct HashTableValue staticCustomAccessorTableValues[3] = {
    { ASCIILiteral("testStaticAccessor"), static_cast<unsigned>(PropertyAttribute::CustomAccessor), NoIntrinsic, { (intptr_t)static_cast<PropertySlot::GetValueFunc>(testStaticAccessorGetter), (intptr_t)static_cast<PutPropertySlot::PutValueFunc>(testStaticAccessorPutter) } },
    { ASCIILiteral("testStaticAccessorDontEnum"), PropertyAttribute::CustomAccessor | PropertyAttribute::DontEnum, NoIntrinsic, { (intptr_t)static_cast<PropertySlot::GetValueFunc>(testStaticAccessorGetter), (intptr_t)static_cast<PutPropertySlot::PutValueFunc>(testStaticAccessorPutter) } },
    { ASCIILiteral("testStaticAccessorReadOnly"), PropertyAttribute::CustomAccessor | PropertyAttribute::DontEnum | PropertyAttribute::ReadOnly, NoIntrinsic, { (intptr_t)static_cast<PropertySlot::GetValueFunc>(testStaticAccessorGetter), 0 } },
};

static const struct HashTable staticCustomAccessorTable =
    { 3, 7, PropertyAttribute::ReadOnly | PropertyAttribute::DontEnum | PropertyAttribute::CustomAccessor, staticCustomAccessorTableValues, staticCustomAccessorTableIndex };

class StaticCustomAccessor : public JSNonFinalObject {
    typedef JSNonFinalObject Base;
public:
    StaticCustomAccessor(VM& vm, Structure* structure)
        : Base(vm, structure)
    {
        //DollarVMAssertScope assertScope;
    }

    DECLARE_INFO;

    static constexpr unsigned StructureFlags = Base::StructureFlags | HasStaticPropertyTable | OverridesGetOwnPropertySlot;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        //DollarVMAssertScope assertScope;
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

    static StaticCustomAccessor* create(VM& vm, Structure* structure)
    {
        //DollarVMAssertScope assertScope;
        StaticCustomAccessor* accessor = new (NotNull, allocateCell<StaticCustomAccessor>(vm.heap)) StaticCustomAccessor(vm, structure);
        accessor->finishCreation(vm);
        return accessor;
    }

    static bool getOwnPropertySlot(JSObject* thisObject, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
    {
        if (String(propertyName.uid()) == "thinAirCustomGetter") {
            slot.setCacheableCustom(thisObject, DontDelete | ReadOnly | DontEnum | CustomAccessor, testStaticAccessorGetter);
            return true;
        }
        return JSNonFinalObject::getOwnPropertySlot(thisObject, exec, propertyName, slot);
    }
};

class StaticDontDeleteDontEnum : public JSNonFinalObject {
public:
    using Base = JSNonFinalObject;
    static constexpr unsigned StructureFlags = Base::StructureFlags | HasStaticPropertyTable;

    StaticDontDeleteDontEnum(VM& vm, Structure* structure)
        : Base(vm, structure)
    {
        //DollarVMAssertScope assertScope;
    }

    DECLARE_INFO;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        //DollarVMAssertScope assertScope;
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

    static StaticDontDeleteDontEnum* create(VM& vm, Structure* structure)
    {
        //DollarVMAssertScope assertScope;
        StaticDontDeleteDontEnum* result = new (NotNull, allocateCell<StaticDontDeleteDontEnum>(vm.heap)) StaticDontDeleteDontEnum(vm, structure);
        result->finishCreation(vm);
        return result;
    }
};

static EncodedJSValue JIT_OPERATION staticDontDeleteDontEnumMethod(ExecState*)
{
    //DollarVMAssertScope assertScope;
    return encodedJSUndefined();
}

static const struct CompactHashIndex staticDontDeleteDontEnumTableIndex[8] = {
    { 0, -1 },
    { 1, -1 },
    { -1, -1 },
    { -1, -1 },
    { -1, -1 },
    { 2, -1 },
    { -1, -1 },
    { -1, -1 },
};

static const struct HashTableValue staticDontDeleteDontEnumTableValues[3] = {
   { ASCIILiteral("dontEnum"), static_cast<unsigned>(PropertyAttribute::Function|PropertyAttribute::DontEnum), NoIntrinsic, { (intptr_t)(staticDontDeleteDontEnumMethod), (intptr_t)(0) } },
   { ASCIILiteral("dontDelete"), static_cast<unsigned>(PropertyAttribute::Function|PropertyAttribute::DontDelete), NoIntrinsic, { (intptr_t)(staticDontDeleteDontEnumMethod), (intptr_t)(0) } },
   { ASCIILiteral("dontDeleteDontEnum"), static_cast<unsigned>(PropertyAttribute::Function|PropertyAttribute::DontDelete|PropertyAttribute::DontEnum), NoIntrinsic, { (intptr_t)(staticDontDeleteDontEnumMethod), (intptr_t)(0) } },
};

static const struct HashTable staticDontDeleteDontEnumTable =
    { 3, 7, PropertyAttribute::DontEnum | PropertyAttribute::DontDelete, staticDontDeleteDontEnumTableValues, staticDontDeleteDontEnumTableIndex };

const ClassInfo StaticDontDeleteDontEnum::s_info = { "StaticDontDeleteDontEnum", &Base::s_info, &staticDontDeleteDontEnumTable, CREATE_METHOD_TABLE(StaticDontDeleteDontEnum) };

static EncodedJSValue JIT_OPERATION testStaticValueGetter(ExecState*, EncodedJSValue, PropertyName)
{
    //DollarVMAssertScope assertScope;
    return JSValue::encode(jsUndefined());
}

static bool JIT_OPERATION testStaticValuePutter(ExecState* exec, EncodedJSValue thisValue, EncodedJSValue value, PropertyName)
{
    //DollarVMAssertScope assertScope;
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSObject* thisObject = jsDynamicCast<JSObject*>(JSValue::decode(thisValue));
    if (!thisObject) {
        throwVMTypeError(exec, vm);
        return false;
    }

    return thisObject->putDirect(vm, PropertyName(Identifier::fromString(vm, "testStaticValue")), JSValue::decode(value));
}

static bool JIT_OPERATION testStaticValuePutterSetFlag(ExecState* exec, EncodedJSValue thisValue, EncodedJSValue, PropertyName)
{
    //DollarVMAssertScope assertScope;
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSObject* thisObject = jsDynamicCast<JSObject*>(JSValue::decode(thisValue));
    if (!thisObject) {
        throwVMTypeError(exec, vm);
        return false;
    }

    return thisObject->putDirect(vm, PropertyName(Identifier::fromString(vm, "testStaticValueSetterCalled")), jsBoolean(true));
}

static const struct CompactHashIndex staticCustomValueTableIndex[8] = {
    { 1, -1 },
    { -1, -1 },
    { -1, -1 },
    { -1, -1 },
    { 3, -1 },
    { 2, -1 },
    { 0, -1 },
    { -1, -1 },
};

static const struct HashTableValue staticCustomValueTableValues[4] = {
    { ASCIILiteral("testStaticValue"), static_cast<unsigned>(PropertyAttribute::CustomValue), NoIntrinsic, { (intptr_t)static_cast<GetValueFunc>(testStaticValueGetter), (intptr_t)static_cast<PutValueFunc>(testStaticValuePutter) } },
    { ASCIILiteral("testStaticValueNoSetter"), PropertyAttribute::CustomValue | PropertyAttribute::DontEnum, NoIntrinsic, { (intptr_t)static_cast<GetValueFunc>(testStaticValueGetter), 0 } },
    { ASCIILiteral("testStaticValueReadOnly"), PropertyAttribute::CustomValue | PropertyAttribute::DontEnum | PropertyAttribute::ReadOnly, NoIntrinsic, { (intptr_t)static_cast<GetValueFunc>(testStaticValueGetter), 0 } },
    { ASCIILiteral("testStaticValueSetFlag"), static_cast<unsigned>(PropertyAttribute::CustomValue), NoIntrinsic, { (intptr_t)static_cast<GetValueFunc>(testStaticValueGetter), (intptr_t)static_cast<PutValueFunc>(testStaticValuePutterSetFlag) } },
};

static const struct HashTable staticCustomValueTable =
    { 4, 7, PropertyAttribute::ReadOnly | PropertyAttribute::DontEnum | PropertyAttribute::CustomValue, staticCustomValueTableValues, staticCustomValueTableIndex };

class StaticCustomValue : public JSNonFinalObject {
    using Base = JSNonFinalObject;
public:
    StaticCustomValue(VM& vm, Structure* structure)
        : Base(vm, structure)
    {
        //DollarVMAssertScope assertScope;
    }

    DECLARE_INFO;

    static constexpr unsigned StructureFlags = Base::StructureFlags | HasStaticPropertyTable;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        //DollarVMAssertScope assertScope;
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

    static StaticCustomValue* create(VM& vm, Structure* structure)
    {
        //DollarVMAssertScope assertScope;
        StaticCustomValue* accessor = new (NotNull, allocateCell<StaticCustomValue>(vm.heap)) StaticCustomValue(vm, structure);
        accessor->finishCreation(vm);
        return accessor;
    }
};
const ClassInfo StaticCustomValue::s_info = { "StaticCustomValue", &Base::s_info, &staticCustomValueTable, CREATE_METHOD_TABLE(StaticCustomValue) };

static EncodedJSValue JIT_OPERATION functionCreateStaticCustomValue(ExecState* exec)
{
    //DollarVMAssertScope assertScope;
	auto* globalObject = exec->lexicalGlobalObject();
    VM& vm = globalObject->vm();
    JSLockHolder lock(vm);
    Structure* structure = StaticCustomValue::createStructure(vm, globalObject, jsNull());
    auto* result = StaticCustomValue::create(vm, structure);
    return JSValue::encode(result);
}

#if 0
class ObjectDoingSideEffectPutWithoutCorrectSlotStatus : public JSNonFinalObject {
    using Base = JSNonFinalObject;
    static constexpr unsigned StructureFlags = Base::StructureFlags | OverridesPut;
public:
    template<typename CellType, SubspaceAccess>
    static CompleteSubspace* subspaceFor(VM& vm)
    {
        return &vm.cellSpace;
    }

    ObjectDoingSideEffectPutWithoutCorrectSlotStatus(VM& vm, Structure* structure)
        : Base(vm, structure)
    {
        DollarVMAssertScope assertScope;
    }

    DECLARE_INFO;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        DollarVMAssertScope assertScope;
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

    static ObjectDoingSideEffectPutWithoutCorrectSlotStatus* create(VM& vm, Structure* structure)
    {
        DollarVMAssertScope assertScope;
        ObjectDoingSideEffectPutWithoutCorrectSlotStatus* accessor = new (NotNull, allocateCell<ObjectDoingSideEffectPutWithoutCorrectSlotStatus>(vm.heap)) ObjectDoingSideEffectPutWithoutCorrectSlotStatus(vm, structure);
        accessor->finishCreation(vm);
        return accessor;
    }

    static bool put(JSCell* cell, JSGlobalObject* globalObject, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
    {
        DollarVMAssertScope assertScope;
        auto* thisObject = jsCast<ObjectDoingSideEffectPutWithoutCorrectSlotStatus*>(cell);
        auto throwScope = DECLARE_THROW_SCOPE(globalObject->vm());
        auto* string = value.toString(globalObject);
        RETURN_IF_EXCEPTION(throwScope, false);
        RELEASE_AND_RETURN(throwScope, Base::put(thisObject, globalObject, propertyName, string, slot));
    }
};
#endif

class CustomGetter : public JSNonFinalObject {
public:
    CustomGetter(VM& vm, Structure* structure)
        : Base(vm, structure)
    {
    }

    DECLARE_INFO;
    typedef JSNonFinalObject Base;
    static const unsigned StructureFlags = Base::StructureFlags | JSC::OverridesGetOwnPropertySlot;

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
    }

    static CustomGetter* create(VM& vm, Structure* structure)
    {
        CustomGetter* getter = new (NotNull, allocateCell<CustomGetter>(vm.heap)) CustomGetter(vm, structure);
        getter->finishCreation(vm);
        return getter;
    }

    static bool getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
    {
    	VM& vm = exec->vm();
        CustomGetter* thisObject = jsCast<CustomGetter*>(object);
        if (propertyName == PropertyName(Identifier::fromString(vm, "customGetter"))) {
            slot.setCacheableCustom(thisObject, DontDelete | ReadOnly | DontEnum, customGetter);
            return true;
        }

        if (propertyName == PropertyName(Identifier::fromString(vm, "customGetterAccessor"))) {
            slot.setCacheableCustom(thisObject, DontDelete | ReadOnly | DontEnum | CustomAccessor, customGetterAcessor);
            return true;
        }

        return JSObject::getOwnPropertySlot(thisObject, exec, propertyName, slot);
    }

private:
    static EncodedJSValue customGetter(ExecState* exec, EncodedJSValue thisValue, PropertyName)
    {
    	VM& vm = exec->vm();
        CustomGetter* thisObject = jsDynamicCast<CustomGetter*>(JSValue::decode(thisValue));
        if (!thisObject)
            return throwVMTypeError(exec, vm);
        bool shouldThrow = thisObject->get(vm, exec, PropertyName(Identifier::fromString(vm, "shouldThrow"))).toBoolean(exec);
        if (shouldThrow)
            return throwVMTypeError(exec, vm);
        return JSValue::encode(jsNumber(100));
    }

    static EncodedJSValue customGetterAcessor(ExecState* exec, EncodedJSValue thisValue, PropertyName)
    {
        VM& vm = exec->vm();
        //auto scope = DECLARE_THROW_SCOPE(vm);

        JSObject* thisObject = jsDynamicCast<JSObject*>(JSValue::decode(thisValue));
        if (!thisObject)
            return throwVMTypeError(exec, vm);
        bool shouldThrow = thisObject->get(vm, exec, PropertyName(Identifier::fromString(vm, "shouldThrow"))).toBoolean(exec);
        if (shouldThrow)
            return throwVMTypeError(exec, vm);
        return JSValue::encode(jsNumber(100));
    }
};

class RuntimeArray : public JSArray {
public:
    typedef JSArray Base;
    static const unsigned StructureFlags = Base::StructureFlags | OverridesGetOwnPropertySlot | InterceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero;

    static RuntimeArray* create(ExecState* exec)
    {
        VM& vm = exec->vm();
        JSGlobalObject* globalObject = exec->lexicalGlobalObject();
        Structure* structure = createStructure(vm, globalObject, createPrototype(vm, globalObject));
        RuntimeArray* runtimeArray = new (NotNull, allocateCell<RuntimeArray>(vm.heap)) RuntimeArray(exec, structure);
        runtimeArray->finishCreation(exec);
        vm.heap.addFinalizer(runtimeArray, destroy);
        return runtimeArray;
    }

    ~RuntimeArray() { }

    static void destroy(JSCell* cell)
    {
        static_cast<RuntimeArray*>(cell)->RuntimeArray::~RuntimeArray();
    }

	// We don't need a destructor because we use a finalizer instead.
    static const bool needsDestruction = false;

    static bool getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
    {
        RuntimeArray* thisObject = jsCast<RuntimeArray*>(object);
        if (propertyName == exec->propertyNames().length) {
            slot.setCacheableCustom(thisObject, DontDelete | ReadOnly | DontEnum, lengthGetter);
            return true;
        }

        Optional<uint32_t> index = parseIndex(propertyName);
        if (index && index.value() < thisObject->getLength()) {
            slot.setValue(thisObject, DontDelete | DontEnum, jsNumber(thisObject->m_vector[index.value()]));
            return true;
        }

        return JSObject::getOwnPropertySlot(thisObject, exec, propertyName, slot);
    }

    static bool getOwnPropertySlotByIndex(JSObject* object, ExecState* exec, unsigned index, PropertySlot& slot)
    {
        RuntimeArray* thisObject = jsCast<RuntimeArray*>(object);
        if (index < thisObject->getLength()) {
            slot.setValue(thisObject, DontDelete | DontEnum, jsNumber(thisObject->m_vector[index]));
            return true;
        }

        return JSObject::getOwnPropertySlotByIndex(thisObject, exec, index, slot);
    }

    static NO_RETURN_DUE_TO_CRASH bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&)
    {
        RELEASE_ASSERT_NOT_REACHED();
        return false;
    }

    static NO_RETURN_DUE_TO_CRASH bool deleteProperty(JSCell*, ExecState*, PropertyName)
    {
        RELEASE_ASSERT_NOT_REACHED();
        return false;
    }

    unsigned getLength() const { return m_vector.size(); }

    DECLARE_INFO;

    static ArrayPrototype* createPrototype(VM&, JSGlobalObject* globalObject)
    {
        return globalObject->arrayPrototype();
    }

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject, JSValue prototype)
    {
        return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info(), ArrayClass);
    }

protected:
    void finishCreation(ExecState* exec)
    {
        Base::finishCreation(exec->vm());
        ASSERT(inherits(info()));

        for (size_t i = 0; i < exec->argumentCount(); i++)
            m_vector.append(exec->argument(i).toInt32(exec));
    }

private:
    RuntimeArray(ExecState* exec, Structure* structure)
        : JSArray(exec->vm(), structure, 0)
    {
    }

    static EncodedJSValue lengthGetter(ExecState* exec, EncodedJSValue thisValue, PropertyName)
    {
        VM& vm = exec->vm();
        RuntimeArray* thisObject = jsDynamicCast<RuntimeArray*>(JSValue::decode(thisValue));
        if (!thisObject)
            return throwVMTypeError(exec, vm);
        return JSValue::encode(jsNumber(thisObject->getLength()));
    }

    Vector<int> m_vector;
};

const ClassInfo Masquerader::s_info = { "Masquerader", &Base::s_info, nullptr, CREATE_METHOD_TABLE(Masquerader) };
const ClassInfo ImpureGetter::s_info = { "ImpureGetter", &Base::s_info, nullptr, CREATE_METHOD_TABLE(ImpureGetter) };
const ClassInfo StaticCustomAccessor::s_info = { "StaticCustomAccessor", &Base::s_info, &staticCustomAccessorTable, CREATE_METHOD_TABLE(StaticCustomAccessor) };
const ClassInfo CustomGetter::s_info = { "CustomGetter", &Base::s_info, nullptr, CREATE_METHOD_TABLE(CustomGetter) };
const ClassInfo RuntimeArray::s_info = { "RuntimeArray", &Base::s_info, nullptr, CREATE_METHOD_TABLE(RuntimeArray) };

static bool customFunctionSetter(ExecState* exec, EncodedJSValue, EncodedJSValue encodedValue, PropertyName)
{
    //DollarVMAssertScope assertScope;
    VM& vm = exec->vm();

    JSValue value = JSValue::decode(encodedValue);
    JSFunction* function = jsDynamicCast<JSFunction*>(value);
    if (!function)
        return false;

    CallData callData = getCallData(vm, function);
    MarkedArgumentBuffer args;
    call(vm, exec, function, callData, jsUndefined(), args);
    return true;
}

class JSTestCustomGetterSetter : public JSNonFinalObject {
public:
    typedef JSNonFinalObject Base;
    static const unsigned StructureFlags = Base::StructureFlags;

    JSTestCustomGetterSetter(VM& vm, Structure* structure)
        : Base(vm, structure)
    { }

    static JSTestCustomGetterSetter* create(VM& vm, JSGlobalObject*, Structure* structure)
    {
        JSTestCustomGetterSetter* result = new (NotNull, allocateCell<JSTestCustomGetterSetter>(vm.heap)) JSTestCustomGetterSetter(vm, structure);
        result->finishCreation(vm);
        return result;
    }

    void finishCreation(VM& vm);

    static Structure* createStructure(VM& vm, JSGlobalObject* globalObject)
    {
        return Structure::create(vm, globalObject, globalObject->objectPrototype(), TypeInfo(ObjectType, StructureFlags), info());
    }

    DECLARE_INFO;
};


static EncodedJSValue customGetAccessor(ExecState*, EncodedJSValue thisValue, PropertyName)
{
    // Passed |this|
    return thisValue;
}

static EncodedJSValue customGetValue(ExecState* exec, EncodedJSValue slotValue, PropertyName)
{
    RELEASE_ASSERT(JSValue::decode(slotValue).inherits(exec->vm(), JSTestCustomGetterSetter::info()));
    // Passed property holder.
    return slotValue;
}

static EncodedJSValue customGetAccessorGlobalObject(ExecState* exec, EncodedJSValue, PropertyName)
{
    return JSValue::encode(exec->lexicalGlobalObject());
}

static EncodedJSValue customGetValueGlobalObject(ExecState* exec, EncodedJSValue, PropertyName)
{
    return JSValue::encode(exec->lexicalGlobalObject());
}

static bool customSetAccessor(ExecState* exec, EncodedJSValue thisObject, EncodedJSValue encodedValue, PropertyName)
{
    VM& vm = exec->vm();

    JSValue value = JSValue::decode(encodedValue);
    if (!value.isObject())
        return false;

    JSObject* object = asObject(value);
    PutPropertySlot slot(object);
    object->put(object, exec, Identifier::fromString(vm, "result"), JSValue::decode(thisObject), slot);

    return true;
}

static bool customSetValue(ExecState* exec, EncodedJSValue slotValue, EncodedJSValue encodedValue, PropertyName)
{
    VM& vm = exec->vm();

    RELEASE_ASSERT(JSValue::decode(slotValue).inherits(exec->vm(), JSTestCustomGetterSetter::info()));

    JSValue value = JSValue::decode(encodedValue);
    if (!value.isObject())
        return false;

    JSObject* object = asObject(value);
    PutPropertySlot slot(object);
    object->put(object, exec, Identifier::fromString(vm, "result"), JSValue::decode(slotValue), slot);

    return true;
}

void JSTestCustomGetterSetter::finishCreation(VM& vm)
{
    Base::finishCreation(vm);

    putDirectCustomAccessor(vm, Identifier::fromString(vm, "customValue"),
        CustomGetterSetter::create(vm, customGetValue, customSetValue), 0);
    putDirectCustomAccessor(vm, Identifier::fromString(vm, "customAccessor"),
        CustomGetterSetter::create(vm, customGetAccessor, customSetAccessor), static_cast<unsigned>(CustomAccessor));
    putDirectCustomAccessor(vm, Identifier::fromString(vm, "customValueGlobalObject"),
		CustomGetterSetter::create(vm, customGetValueGlobalObject, nullptr), 0);
	putDirectCustomAccessor(vm, Identifier::fromString(vm, "customAccessorGlobalObject"),
		CustomGetterSetter::create(vm, customGetAccessorGlobalObject, nullptr), PropertyAttribute::ReadOnly | PropertyAttribute::CustomAccessor);

    putDirectCustomAccessor(vm, Identifier::fromString(vm, "customFunction"),
        CustomGetterSetter::create(vm, customGetAccessor, customFunctionSetter), static_cast<unsigned>(CustomAccessor));
}

const ClassInfo JSTestCustomGetterSetter::s_info = { "JSTestCustomGetterSetter", &Base::s_info, 0, CREATE_METHOD_TABLE(JSTestCustomGetterSetter) };

static EncodedJSValue JSC_HOST_CALL functionLoadGetterFromGetterSetter(ExecState* exec)
{
    VM& vm = exec->vm();
    RELEASE_ASSERT(exec->argumentCount() >= 1);
    GetterSetter* getterSetter = jsDynamicCast<GetterSetter*>(exec->argument(0));
    RELEASE_ASSERT(getterSetter);
    JSObject* getter = getterSetter->getter();
    RELEASE_ASSERT(getter);
    return JSValue::encode(getter);
}

static EncodedJSValue JSC_HOST_CALL functionCreateCustomTestGetterSetter(ExecState* exec)
{
    VM& vm = exec->vm();
    JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    return JSValue::encode(JSTestCustomGetterSetter::create(vm, globalObject, JSTestCustomGetterSetter::createStructure(vm, globalObject)));
}

static EncodedJSValue JSC_HOST_CALL functionFullGC(ExecState* exec)
{
	VM& vm = exec->vm();
    JSLockHolder lock(vm);
    vm.heap.collect(false, true);
    return JSValue::encode(jsNumber(vm.heap.size()));
}

static FunctionExecutable* getExecutableForFunction(JSValue value)
{
    JSFunction* theFunction = jsDynamicCast<JSFunction*>(value);
    if (!theFunction)
        return 0;

    FunctionExecutable* executable = jsDynamicCast<FunctionExecutable*>(theFunction->executable());
    return executable;
}

static EncodedJSValue JSC_HOST_CALL functionNeverInlineFunction(ExecState* exec)
{
    //DollarVMAssertScope assertScope;
    if (exec->argumentCount() < 1)
        return JSValue::encode(jsUndefined());

    JSValue theFunctionValue = exec->uncheckedArgument(0);

    if (FunctionExecutable* executable = getExecutableForFunction(theFunctionValue))
        executable->setNeverInline(true);

    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL functionNoDFG(ExecState* exec)
{
    //DollarVMAssertScope assertScope;
    if (exec->argumentCount() < 1)
        return JSValue::encode(jsUndefined());

    JSValue theFunctionValue = exec->uncheckedArgument(0);

    if (FunctionExecutable* executable = getExecutableForFunction(theFunctionValue))
        executable->setNeverOptimize(true);

    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL functionEffectful42(ExecState*)
{
    return JSValue::encode(jsNumber(42));
}

static EncodedJSValue JSC_HOST_CALL functionMakeMasquerader(ExecState* exec)
{
    return JSValue::encode(Masquerader::create(exec->vm(), exec->lexicalGlobalObject()));
}

static EncodedJSValue JSC_HOST_CALL functionCreateImpureGetter(ExecState* exec)
{
    JSLockHolder lock(exec);
    JSValue target = exec->argument(0);
    JSObject* delegate = 0;
    if (target.isObject())
        delegate = asObject(target.asCell());
    Structure* structure = ImpureGetter::createStructure(exec->vm(), exec->lexicalGlobalObject(), jsNull());
    ImpureGetter* result = ImpureGetter::create(exec->vm(), structure, delegate);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL functionCreateCustomGetterObject(ExecState*exec)
{
	VM& vm = exec->vm();
    JSLockHolder lock(vm);
    Structure* structure = CustomGetter::createStructure(vm, exec->lexicalGlobalObject(), jsNull());
    CustomGetter* result = CustomGetter::create(vm, structure);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL functionCreateRuntimeArray(ExecState* exec)
{
	VM& vm = exec->vm();
    JSLockHolder lock(vm);
    RuntimeArray* array = RuntimeArray::create(exec);
    return JSValue::encode(array);
}

static EncodedJSValue JSC_HOST_CALL functionCreateStaticCustomAccessor(ExecState* exec)
{
    //DollarVMAssertScope assertScope;
    VM& vm = exec->vm();
    JSLockHolder lock(vm);
    Structure* structure = StaticCustomAccessor::createStructure(vm, exec->lexicalGlobalObject(), jsNull());
    auto* result = StaticCustomAccessor::create(vm, structure);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL functionCreateStaticDontDeleteDontEnum(ExecState* exec)
{
    //DollarVMAssertScope assertScope;
	auto* globalObject = exec->lexicalGlobalObject();
    VM& vm = globalObject->vm();
    JSLockHolder lock(vm);
    Structure* structure = StaticDontDeleteDontEnum::createStructure(vm, globalObject, jsNull());
    auto* result = StaticDontDeleteDontEnum::create(vm, structure);
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL functionSetImpureGetterDelegate(ExecState* exec)
{
    JSLockHolder lock(exec);
    JSValue base = exec->argument(0);
    if (!base.isObject())
        return JSValue::encode(jsUndefined());
    JSValue delegate = exec->argument(1);
    if (!delegate.isObject())
        return JSValue::encode(jsUndefined());
    ImpureGetter* impureGetter = jsCast<ImpureGetter*>(asObject(base.asCell()));
    impureGetter->setDelegate(exec->vm(), asObject(delegate.asCell()));
    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL functionToUncacheableDictionary(ExecState* exec)
{
    //DollarVMAssertScope assertScope;
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSObject* object = jsDynamicCast<JSObject*>(exec->argument(0));
    if (!object)
        return throwVMTypeError(exec, vm, "Expected first argument to be an object");
    object->convertToUncacheableDictionary(vm);
    return JSValue::encode(object);
}

static CodeBlock* getSomeBaselineCodeBlockForFunction(JSValue theFunctionValue)
{
    FunctionExecutable* executable = getExecutableForFunction(theFunctionValue);
    if (!executable)
        return 0;

    CodeBlock* baselineCodeBlock = executable->baselineCodeBlockFor(CodeForCall);

    if (!baselineCodeBlock)
        baselineCodeBlock = executable->baselineCodeBlockFor(CodeForConstruct);

    return baselineCodeBlock;
}

static EncodedJSValue JSC_HOST_CALL numberOfDFGCompiles(ExecState* exec)
{
    bool pretendToHaveManyCompiles = false;
#if ENABLE(DFG_JIT)
    if (!Options::useJIT() || !Options::useDFGJIT())
        pretendToHaveManyCompiles = true;
#else
    pretendToHaveManyCompiles = true;
#endif

    if (CodeBlock* baselineCodeBlock = getSomeBaselineCodeBlockForFunction(exec->argument(0))) {
        if (pretendToHaveManyCompiles)
            return JSValue::encode(jsNumber(1000000.0));

        return JSValue::encode(jsNumber(baselineCodeBlock->numberOfDFGCompiles()));
    }

    return JSValue::encode(jsNumber(0));
}

static EncodedJSValue JSC_HOST_CALL functionReoptimizationRetryCount(ExecState* exec)
{
    if (exec->argumentCount() < 1)
        return JSValue::encode(jsUndefined());

    CodeBlock* block = getSomeBaselineCodeBlockForFunction(exec->argument(0));
    if (!block)
        return JSValue::encode(jsNumber(0));

    return JSValue::encode(jsNumber(block->reoptimizationRetryCounter()));
}

static EncodedJSValue JSC_HOST_CALL functionTransferArrayBuffer(ExecState* exec)
{
	VM& vm = exec->vm();

    if (exec->argumentCount() < 1)
        return JSValue::encode(exec->vm().throwException(exec, createError(exec, ASCIILiteral("Not enough arguments"))));

    JSArrayBuffer* buffer = jsDynamicCast<JSArrayBuffer*>(exec->argument(0));
    if (!buffer)
        return JSValue::encode(vm.throwException(exec, createError(exec, ASCIILiteral("Expected an array buffer"))));

    ArrayBufferContents dummyContents;
    buffer->impl()->transferTo(vm, dummyContents);

    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL functionHasCustomProperties(ExecState* exec)
{
    JSValue value = exec->argument(0);
    if (value.isObject())
        return JSValue::encode(jsBoolean(asObject(value)->hasCustomProperties()));
    return JSValue::encode(jsBoolean(false));
}

static EncodedJSValue JSC_HOST_CALL functionCreateBuiltin(ExecState* exec)
{
    if (exec->argumentCount() < 1 || !exec->argument(0).isString())
        return JSValue::encode(jsUndefined());

    VM& vm = exec->vm();
    String functionText = exec->argument(0).toWTFString(exec);
	RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSFunction* func = JSFunction::create(vm, createBuiltinExecutable(vm, makeSource(functionText, SourceOrigin()), Identifier::fromString(vm, "foo"), NoIntrinsic, ConstructAbility::CannotConstruct), exec->lexicalGlobalObject());

    return JSValue::encode(func);
}

static EncodedJSValue JSC_HOST_CALL functionEnsureArrayStorage(ExecState* exec)
{
    VM& vm = exec->vm();
    for (unsigned i = 0; i < exec->argumentCount(); ++i) {
        if (JSObject* object = jsDynamicCast<JSObject*>(exec->argument(i)))
            object->ensureArrayStorage(vm);
    }
    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL functionToCacheableDictionary(ExecState* exec)
{
    VM& vm = exec->vm();
	if (JSObject* object = jsDynamicCast<JSObject*>(exec->argument(0)))
		object->convertToDictionary(vm);
    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL functionDescribe(ExecState* exec)
{
    VM& vm = exec->vm();
    if (exec->argumentCount() < 1)
        return JSValue::encode(jsUndefined());
    return JSValue::encode(jsString(vm, toString(exec->argument(0))));
}

// Creates a new global object.
// Usage: $vm.createGlobalObject()
static EncodedJSValue JSC_HOST_CALL functionCreateGlobalObject(ExecState* exec)
{
    VM& vm = exec->vm();
    JSLockHolder lock(vm);
    JSGlobalObject* globalObject = JSGlobalObject::create(vm, JSGlobalObject::createStructure(vm, jsNull()));
    return JSValue::encode(globalObject);
}

static EncodedJSValue JSC_HOST_CALL functionRunString(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    String source = exec->argument(0).toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSGlobalObject* globalObject = JSGlobalObject::create(vm, JSGlobalObject::createStructure(vm, jsNull()));

    JSArray* array = constructEmptyArray(globalObject->globalExec(), 0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    for (unsigned i = 1; i < exec->argumentCount(); ++i) {
        array->putDirectIndex(globalObject->globalExec(), i - 1, exec->uncheckedArgument(i));
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }
    globalObject->putDirect(vm, Identifier::fromString(vm, "arguments"), array);

    NakedPtr<Exception> exception;
    evaluate(globalObject->globalExec(), makeSource(source, exec->callerSourceOrigin()), JSValue(), exception);

    if (exception) {
        vm.throwException(globalObject->globalExec(), exception);
        return JSValue::encode(jsUndefined());
    }

    return JSValue::encode(globalObject);
}

static EncodedJSValue JSC_HOST_CALL functionAsDoubleNumber(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);
    double num = exec->argument(0).toNumber(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(jsDoubleNumber(num));
}

static EncodedJSValue JSC_HOST_CALL functionGlobalObjectCount(ExecState* exec)
{
    return JSValue::encode(jsNumber(exec->vm().heap.globalObjectCount()));
}

EncodedJSValue JSC_HOST_CALL functionEvalScript(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    String sourceCode = exec->argument(0).toWTFString(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

    JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    if (!globalObject)
        return JSValue::encode(throwException(exec, scope, createError(exec, ASCIILiteral("Expected global to point to a global object"))));

    NakedPtr<Exception> evaluationException;
    JSValue result = evaluate(globalObject->globalExec(), makeSource(sourceCode, exec->callerSourceOrigin()), JSValue(), evaluationException);
    if (evaluationException)
        throwException(exec, scope, evaluationException);
    return JSValue::encode(result);
}

// This function is not generally very helpful in 64-bit code as the tag and payload
// share a register. But in 32-bit JITed code the tag may not be checked if an
// optimization removes type checking requirements, such as in ===.
static EncodedJSValue JSC_HOST_CALL functionAddressOf(ExecState* exec)
{
    JSValue value = exec->argument(0);
    if (!value.isCell())
        return JSValue::encode(jsUndefined());
    // Need to cast to uint64_t so bitwise_cast will play along.
    uint64_t asNumber = reinterpret_cast<uint64_t>(value.asCell());
    EncodedJSValue returnValue = JSValue::encode(jsNumber(bitwise_cast<double>(asNumber)));
    return returnValue;
}

static EncodedJSValue JSC_HOST_CALL functionGlobalObjectForObject(ExecState* exec)
{
    JSValue value = exec->argument(0);
    RELEASE_ASSERT(value.isObject());
    JSGlobalObject* globalObject = jsCast<JSObject*>(value)->globalObject();
    RELEASE_ASSERT(globalObject);
    return JSValue::encode(globalObject);
}

static bool testCustomSetterImpl(ExecState* exec, JSGlobalObject* thisObject, EncodedJSValue encodedValue, const ASCIILiteral& propertyName)
{
    VM& vm = exec->vm();

    Identifier identifier = Identifier::fromString(vm, propertyName);
    thisObject->putDirect(vm, identifier, JSValue::decode(encodedValue), DontEnum);

    return true;
}

static EncodedJSValue dummyCustomGetValue(ExecState*, EncodedJSValue, PropertyName)
{
    return encodedJSUndefined();
}

static bool testCustomAccessorSetter(ExecState* exec, EncodedJSValue thisValue, EncodedJSValue encodedValue, PropertyName)
{
    VM& vm = exec->vm();
    RELEASE_ASSERT(JSValue::decode(thisValue).isCell());
    JSCell* thisCell = JSValue::decode(thisValue).asCell();
    RELEASE_ASSERT(thisCell->type() == PureForwardingProxyType);
    JSGlobalObject* thisObject = jsDynamicCast<JSGlobalObject*>(jsCast<JSProxy*>(thisCell)->target());
    RELEASE_ASSERT(thisObject);
    return testCustomSetterImpl(exec, thisObject, encodedValue, ASCIILiteral("_testCustomAccessorSetter"));
}

static bool testCustomValueSetter(ExecState* exec, EncodedJSValue thisValue, EncodedJSValue encodedValue, PropertyName)
{
    VM& vm = exec->vm();
    RELEASE_ASSERT(JSValue::decode(thisValue).isCell());
    JSCell* thisCell = JSValue::decode(thisValue).asCell();
    JSGlobalObject* thisObject = jsDynamicCast<JSGlobalObject*>(thisCell);
    RELEASE_ASSERT(thisObject);
    return testCustomSetterImpl(exec, thisObject, encodedValue, ASCIILiteral("_testCustomValueSetter"));
}

static EncodedJSValue JSC_HOST_CALL functionCreateNonRopeNonAtomString(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    String source = exec->argument(0).toWTFString(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

    if (source.isEmpty() || source.length() == 1)
        return throwVMTypeError(exec, vm, ASCIILiteral("empty / one-length string can be always atom string"));

    if (source.impl()->isAtomic()) {
        if (source.is8Bit())
            source = StringImpl::create(source.characters8(), source.length());
        else
            source = StringImpl::create(source.characters16(), source.length());
    }

    RELEASE_ASSERT(!source.impl()->isAtom());

    return JSValue::encode(jsString(vm, source));
}

}

ALWAYS_INLINE void JSGlobalObject::reset(JSValue prototype)
{
    auto* exec = JSGlobalObject::globalExec();
    auto& vm = this->vm();
    auto* propertyNames = vm.propertyNames;
    const auto& builtinNames = propertyNames->builtinNames();

    m_functionPrototype.set(vm, this, FunctionPrototype::create(vm, FunctionPrototype::createStructure(vm, this, jsNull()))); // The real prototype will be set once ObjectPrototype is created.
    m_calleeStructure.set(vm, this, JSCallee::createStructure(vm, this, jsNull()));

    m_globalLexicalEnvironment.set(vm, this, JSGlobalLexicalEnvironment::create(vm, JSGlobalLexicalEnvironment::createStructure(vm, this), this));
    // Need to create the callee structure (above) before creating the callee.
    m_globalCallee.set(vm, this, JSCallee::create(vm, this, globalScope()));
    exec->setCallee(m_globalCallee.get());

    m_hostFunctionStructure.set(vm, this, JSFunction::createStructure(vm, this, m_functionPrototype.get()));

    m_builtinFunctions.strictFunctionStructure.set(vm, this, JSStrictFunction::createStructure(vm, this, m_functionPrototype.get()));
    m_builtinFunctions.sloppyFunctionStructure.set(vm, this, JSSloppyFunction::createStructure(vm, this, m_functionPrototype.get()));
    m_builtinFunctions.arrowFunctionStructure.set(vm, this, JSArrowFunction::createStructure(vm, this, m_functionPrototype.get()));
    m_ordinaryFunctions.strictFunctionStructure.set(vm, this, JSStrictFunction::createStructure(vm, this, m_functionPrototype.get()));
    m_ordinaryFunctions.sloppyFunctionStructure.set(vm, this, JSSloppyFunction::createStructure(vm, this, m_functionPrototype.get()));
    m_ordinaryFunctions.arrowFunctionStructure.set(vm, this, JSArrowFunction::createStructure(vm, this, m_functionPrototype.get()));

    m_customGetterFunctionStructure.set(vm, this, JSCustomGetterFunction::createStructure(vm, this, m_functionPrototype.get()));
    m_customSetterFunctionStructure.set(vm, this, JSCustomSetterFunction::createStructure(vm, this, m_functionPrototype.get()));
    m_boundFunctionStructure.set(vm, this, JSBoundFunction::createStructure(vm, this, m_functionPrototype.get()));
	m_nativeStdFunctionStructure.set(vm, this, JSNativeStdFunction::createStructure(vm, this, m_functionPrototype.get()));
	m_internalFunctionStructure.set(vm, this, InternalFunction::createStructure(vm, this, m_functionPrototype.get()));
    JSFunction* callFunction = 0;
    JSFunction* applyFunction = 0;
    JSFunction* hasInstanceSymbolFunction = 0;
	m_functionPrototype->addFunctionProperties(exec, this, &callFunction, &applyFunction, &hasInstanceSymbolFunction);
	m_callFunction.set(vm, this, callFunction);
    m_applyFunction.set(vm, this, applyFunction);
    m_nullGetterFunction.set(vm, this, NullGetterFunction::create(vm, NullGetterFunction::createStructure(vm, this, m_functionPrototype.get())));
    Structure* nullSetterFunctionStructure = NullSetterFunction::createStructure(vm, this, m_functionPrototype.get());
	m_nullSetterFunction.set(vm, this, NullSetterFunction::create(vm, nullSetterFunctionStructure, ECMAMode::sloppy()));
	m_nullSetterStrictFunction.set(vm, this, NullSetterFunction::create(vm, nullSetterFunctionStructure, ECMAMode::strict()));
	m_objectProtoToStringFunction.set(vm, this, JSFunction::create(vm, this, 0, propertyNames->toString.string(), objectProtoFuncToString, ObjectToStringIntrinsic));
	m_arrayProtoToStringFunction.set(vm, this, JSFunction::create(vm, this, 0, propertyNames->toString.string(), arrayProtoFuncToString));
    m_arrayProtoValuesFunction.set(vm, this, JSFunction::create(vm, arrayPrototypeValuesCodeGenerator(this), this));
    m_initializePromiseFunction.set(vm, this, JSFunction::create(vm, operationsPromiseInitializePromiseCodeGenerator(this), this));
    m_iteratorProtocolFunction.set(vm, this, JSFunction::create(vm, iteratorHelpersPerformIterationCodeGenerator(this), this));
    m_promiseResolveFunction.set(vm, this, JSFunction::create(vm, promiseConstructorResolveCodeGenerator(this), this));
	m_newPromiseCapabilityFunction.set(vm, this, JSFunction::create(vm, operationsPromiseNewPromiseCapabilityCodeGenerator(this), this));
    m_functionProtoHasInstanceSymbolFunction.set(vm, this, hasInstanceSymbolFunction);
	m_objectPrototype.set(vm, this, ObjectPrototype::create(vm, this, ObjectPrototype::createStructure(vm, this, jsNull())));
    GetterSetter* protoAccessor = GetterSetter::create(vm, this,
    	JSFunction::create(vm, this, 0, makeString("get ", propertyNames->underscoreProto.string()), globalFuncProtoGetter, UnderscoreProtoIntrinsic),
		JSFunction::create(vm, this, 0, makeString("set ", propertyNames->underscoreProto.string()), globalFuncProtoSetter));
    m_objectPrototype->putDirectAccessor(exec, propertyNames->underscoreProto, protoAccessor, Accessor | DontEnum);
    m_functionPrototype->structure()->setPrototypeWithoutTransition(vm, m_objectPrototype.get());
    m_objectStructureForObjectConstructor.set(vm, this, m_structureCache.emptyObjectStructureForPrototype(this, m_objectPrototype.get(), JSFinalObject::defaultInlineCapacity()));
    m_objectProtoValueOfFunction.set(vm, this, jsCast<JSFunction*>(objectPrototype()->getDirect(vm, propertyNames->valueOf)));

    JSFunction* thrower = JSFunction::create(vm, this, 0, emptyString(), globalFuncThrowTypeErrorArgumentsCalleeAndCaller);
    thrower->freeze(vm);
    m_throwTypeErrorArgumentsCalleeGetterSetter.set(vm, this, GetterSetter::create(vm, this, thrower, thrower));

    JSTypedArrayViewPrototype* typedArrayProto = JSTypedArrayViewPrototype::create(vm, this, JSTypedArrayViewPrototype::createStructure(vm, this, m_objectPrototype.get()));

    m_typedArrays[toIndex(TypeInt8)].prototype.set(vm, this, JSInt8ArrayPrototype::create(vm, this, JSInt8ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeInt16)].prototype.set(vm, this, JSInt16ArrayPrototype::create(vm, this, JSInt16ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeInt32)].prototype.set(vm, this, JSInt32ArrayPrototype::create(vm, this, JSInt32ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeUint8)].prototype.set(vm, this, JSUint8ArrayPrototype::create(vm, this, JSUint8ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeUint8Clamped)].prototype.set(vm, this, JSUint8ClampedArrayPrototype::create(vm, this, JSUint8ClampedArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeUint16)].prototype.set(vm, this, JSUint16ArrayPrototype::create(vm, this, JSUint16ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeUint32)].prototype.set(vm, this, JSUint32ArrayPrototype::create(vm, this, JSUint32ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeFloat32)].prototype.set(vm, this, JSFloat32ArrayPrototype::create(vm, this, JSFloat32ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeFloat64)].prototype.set(vm, this, JSFloat64ArrayPrototype::create(vm, this, JSFloat64ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeBigInt64)].prototype.set(vm, this, JSBigInt64ArrayPrototype::create(vm, this, JSBigInt64ArrayPrototype::createStructure(vm, this, typedArrayProto)));
    m_typedArrays[toIndex(TypeBigUint64)].prototype.set(vm, this, JSBigUint64ArrayPrototype::create(vm, this, JSBigUint64ArrayPrototype::createStructure(vm, this, typedArrayProto)));
	m_typedArrays[toIndex(TypeDataView)].prototype.set(vm, this, JSDataViewPrototype::create(vm, JSDataViewPrototype::createStructure(vm, this, m_objectPrototype.get())));

    m_typedArrays[toIndex(TypeInt8)].structure.set(vm, this, JSInt8Array::createStructure(vm, this, m_typedArrays[toIndex(TypeInt8)].prototype.get()));
    m_typedArrays[toIndex(TypeInt16)].structure.set(vm, this, JSInt16Array::createStructure(vm, this, m_typedArrays[toIndex(TypeInt16)].prototype.get()));
    m_typedArrays[toIndex(TypeInt32)].structure.set(vm, this, JSInt32Array::createStructure(vm, this, m_typedArrays[toIndex(TypeInt32)].prototype.get()));
    m_typedArrays[toIndex(TypeUint8)].structure.set(vm, this, JSUint8Array::createStructure(vm, this, m_typedArrays[toIndex(TypeUint8)].prototype.get()));
    m_typedArrays[toIndex(TypeUint8Clamped)].structure.set(vm, this, JSUint8ClampedArray::createStructure(vm, this, m_typedArrays[toIndex(TypeUint8Clamped)].prototype.get()));
    m_typedArrays[toIndex(TypeUint16)].structure.set(vm, this, JSUint16Array::createStructure(vm, this, m_typedArrays[toIndex(TypeUint16)].prototype.get()));
    m_typedArrays[toIndex(TypeUint32)].structure.set(vm, this, JSUint32Array::createStructure(vm, this, m_typedArrays[toIndex(TypeUint32)].prototype.get()));
    m_typedArrays[toIndex(TypeFloat32)].structure.set(vm, this, JSFloat32Array::createStructure(vm, this, m_typedArrays[toIndex(TypeFloat32)].prototype.get()));
    m_typedArrays[toIndex(TypeFloat64)].structure.set(vm, this, JSFloat64Array::createStructure(vm, this, m_typedArrays[toIndex(TypeFloat64)].prototype.get()));
    m_typedArrays[toIndex(TypeBigInt64)].structure.set(vm, this, JSBigInt64Array::createStructure(vm, this, m_typedArrays[toIndex(TypeBigInt64)].prototype.get()));
    m_typedArrays[toIndex(TypeBigUint64)].structure.set(vm, this, JSBigUint64Array::createStructure(vm, this, m_typedArrays[toIndex(TypeBigUint64)].prototype.get()));
	m_typedArrays[toIndex(TypeDataView)].structure.set(vm, this, JSDataView::createStructure(vm, this, m_typedArrays[toIndex(TypeDataView)].prototype.get()));

    m_lexicalEnvironmentStructure.set(vm, this, JSLexicalEnvironment::createStructure(vm, this, jsNull()));
    m_moduleEnvironmentStructure.set(vm, this, JSModuleEnvironment::createStructure(vm, this));
	m_strictEvalActivationStructure.set(vm, this, StrictEvalActivation::createStructure(vm, this, jsNull()));
    m_withScopeStructure.set(vm, this, JSWithScope::createStructure(vm, this, jsNull()));

    m_nullPrototypeObjectStructure.set(vm, this, JSFinalObject::createStructure(vm, this, jsNull(), JSFinalObject::defaultInlineCapacity()));

#if !PLATFORM(ESPIAL)
    m_callbackFunctionStructure.set(vm, this, JSCallbackFunction::createStructure(vm, this, m_functionPrototype.get()));
#endif
    m_directArgumentsStructure.set(vm, this, DirectArguments::createStructure(vm, this, m_objectPrototype.get()));
    m_scopedArgumentsStructure.set(vm, this, ScopedArguments::createStructure(vm, this, m_objectPrototype.get()));
    m_clonedArgumentsStructure.set(vm, this, ClonedArguments::createStructure(vm, this, m_objectPrototype.get()));
#if !PLATFORM(ESPIAL)
   	m_callbackConstructorStructure.set(vm, this, JSCallbackConstructor::createStructure(vm, this, m_objectPrototype.get()));
    m_callbackObjectStructure.set(vm, this, JSCallbackObject<JSDestructibleObject>::createStructure(vm, this, m_objectPrototype.get()));
#endif

    m_arrayPrototype.set(vm, this, ArrayPrototype::create(vm, this, ArrayPrototype::createStructure(vm, this, m_objectPrototype.get())));

    m_originalArrayStructureForIndexingShape[UndecidedShape >> IndexingShapeShift].set(vm, this, JSArray::createStructure(vm, this, m_arrayPrototype.get(), ArrayWithUndecided));
	m_originalArrayStructureForIndexingShape[Int32Shape >> IndexingShapeShift].set(vm, this, JSArray::createStructure(vm, this, m_arrayPrototype.get(), ArrayWithInt32));
	m_originalArrayStructureForIndexingShape[DoubleShape >> IndexingShapeShift].set(vm, this, JSArray::createStructure(vm, this, m_arrayPrototype.get(), ArrayWithDouble));
	m_originalArrayStructureForIndexingShape[ContiguousShape >> IndexingShapeShift].set(vm, this, JSArray::createStructure(vm, this, m_arrayPrototype.get(), ArrayWithContiguous));
	m_originalArrayStructureForIndexingShape[ArrayStorageShape >> IndexingShapeShift].set(vm, this, JSArray::createStructure(vm, this, m_arrayPrototype.get(), ArrayWithArrayStorage));
	m_originalArrayStructureForIndexingShape[SlowPutArrayStorageShape >> IndexingShapeShift].set(vm, this, JSArray::createStructure(vm, this, m_arrayPrototype.get(), ArrayWithSlowPutArrayStorage));
	for (unsigned i = 0; i < NumberOfIndexingShapes; ++i)
		m_arrayStructureForIndexingShapeDuringAllocation[i] = m_originalArrayStructureForIndexingShape[i];
    
    m_regExpPrototype.set(vm, this, RegExpPrototype::create(vm, this, RegExpPrototype::createStructure(vm, this, m_objectPrototype.get())));
    m_regExpStructure.set(vm, this, RegExpObject::createStructure(vm, this, m_regExpPrototype.get()));
    m_regExpMatchesArrayStructure.set(vm, this, createRegExpMatchesArrayStructure(vm, this));
    m_regExpMatchesArrayWithIndicesStructure.set(vm, this, createRegExpMatchesArrayWithIndicesStructure(vm, this));
	m_regExpMatchesIndicesArrayStructure.set(vm, this, createRegExpMatchesIndicesArrayStructure(vm, this));

    m_moduleRecordStructure.set(vm, this, JSModuleRecord::createStructure(vm, this, jsNull()));
    m_moduleNamespaceObjectStructure.set(vm, this, JSModuleNamespaceObject::createStructure(vm, this, jsNull()));

    {
		bool isCallable = false;
		m_proxyObjectStructure.set(vm, this, ProxyObject::createStructure(vm, this, jsNull(), isCallable));
		isCallable = true;
		m_callableProxyObjectStructure.set(vm, this, ProxyObject::createStructure(vm, this, jsNull(), isCallable));
	}
    m_proxyRevokeStructure.set(vm, this, ProxyRevoke::createStructure(vm, this, m_functionPrototype.get()));

#if ENABLE(SHARED_ARRAY_BUFFER)
    if (Options::useSharedArrayBuffer()) {
		m_sharedArrayBufferStructure.initLater(
			[] (LazyClassStructure::Initializer& init) {
				init.setPrototype(JSArrayBufferPrototype::create(init.vm, init.global, JSArrayBufferPrototype::createStructure(init.vm, init.global, init.global->m_objectPrototype.get()), ArrayBufferSharingMode::Shared));
				init.setStructure(JSArrayBuffer::createStructure(init.vm, init.global, init.prototype));
				init.setConstructor(JSSharedArrayBufferConstructor::create(init.vm, JSSharedArrayBufferConstructor::createStructure(init.vm, init.global, init.global->m_functionPrototype.get()), jsCast<JSArrayBufferPrototype*>(init.prototype), init.global->m_speciesGetterSetter.get()));
			});
	}
#endif

    m_iteratorPrototype.set(vm, this, IteratorPrototype::create(vm, this, IteratorPrototype::createStructure(vm, this, m_objectPrototype.get())));
    m_asyncIteratorPrototype.set(vm, this, AsyncIteratorPrototype::create(vm, this, AsyncIteratorPrototype::createStructure(vm, this, m_objectPrototype.get())));

    m_generatorPrototype.set(vm, this, GeneratorPrototype::create(vm, this, GeneratorPrototype::createStructure(vm, this, m_iteratorPrototype.get())));
    m_asyncGeneratorPrototype.set(vm, this, AsyncGeneratorPrototype::create(vm, this, AsyncGeneratorPrototype::createStructure(vm, this, m_asyncIteratorPrototype.get())));

    auto* arrayIteratorPrototype = ArrayIteratorPrototype::create(vm, this, ArrayIteratorPrototype::createStructure(vm, this, m_iteratorPrototype.get()));
    m_arrayIteratorPrototype.set(vm, this, arrayIteratorPrototype);
    m_arrayIteratorStructure.set(vm, this, JSArrayIterator::createStructure(vm, this, arrayIteratorPrototype));

#define CREATE_PROTOTYPE_FOR_SIMPLE_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    m_ ## lowerName ## Prototype.set(vm, this, capitalName##Prototype::create(vm, this, capitalName##Prototype::createStructure(vm, this, m_objectPrototype.get()))); \
    m_ ## properName ## Structure.set(vm, this, instanceType::createStructure(vm, this, m_ ## lowerName ## Prototype.get()));

    FOR_EACH_SIMPLE_BUILTIN_TYPE(CREATE_PROTOTYPE_FOR_SIMPLE_TYPE)
    FOR_EACH_BUILTIN_DERIVED_ITERATOR_TYPE(CREATE_PROTOTYPE_FOR_SIMPLE_TYPE)

    if (UNLIKELY(Options::useBigInt())) {
        FOR_BIG_INT_BUILTIN_TYPE_WITH_CONSTRUCTOR(CREATE_PROTOTYPE_FOR_SIMPLE_TYPE)
    }

#undef CREATE_PROTOTYPE_FOR_SIMPLE_TYPE

#define CREATE_PROTOTYPE_FOR_DERIVED_ITERATOR_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
	m_ ## lowerName ## Prototype.set(vm, this, capitalName##Prototype::create(vm, this, capitalName##Prototype::createStructure(vm, this, m_iteratorPrototype.get()))); \
	m_ ## properName ## Structure.set(vm, this, instanceType::createStructure(vm, this, m_ ## lowerName ## Prototype.get()));

    FOR_EACH_BUILTIN_DERIVED_ITERATOR_TYPE(CREATE_PROTOTYPE_FOR_DERIVED_ITERATOR_TYPE)
#undef CREATE_PROTOTYPE_FOR_DERIVED_ITERATOR_TYPE

    m_parseIntFunction.set(vm, this, JSFunction::create(vm, this, 2, propertyNames->parseInt.string(), globalFuncParseInt, ParseIntIntrinsic));
    putDirectWithoutTransition(vm, propertyNames->parseInt, m_parseIntFunction.get(), DontEnum);
    m_parseFloatFunction.set(vm, this, JSFunction::create(vm, this, 1, propertyNames->parseFloat.string(), globalFuncParseFloat, NoIntrinsic));
	putDirectWithoutTransition(vm, propertyNames->parseFloat, m_parseFloatFunction.get(), DontEnum);

     // Constructors

    m_speciesGetterSetter.set(vm, this, GetterSetter::create(vm, this, JSFunction::create(vm, globalOperationsSpeciesGetterCodeGenerator(this), this), 0));

    ObjectConstructor* objectConstructor = ObjectConstructor::create(vm, this, ObjectConstructor::createStructure(vm, this, m_functionPrototype.get()), m_objectPrototype.get());
    m_objectConstructor.set(vm, this, objectConstructor);

    FunctionConstructor* functionConstructor = FunctionConstructor::create(vm, FunctionConstructor::createStructure(vm, this, m_functionPrototype.get()), m_functionPrototype.get());
    m_functionConstructor.set(vm, this, functionConstructor);

    ArrayConstructor* arrayConstructor = ArrayConstructor::create(vm, this, ArrayConstructor::createStructure(vm, this, m_functionPrototype.get()), m_arrayPrototype.get(), speciesGetterSetter());
    m_arrayConstructor.set(vm, this, arrayConstructor);

    m_regExpConstructor.set(vm, this, RegExpConstructor::create(vm, RegExpConstructor::createStructure(vm, this, m_functionPrototype.get()), m_regExpPrototype.get(), speciesGetterSetter()));

#if ENABLE(SHARED_ARRAY_BUFFER)
    AtomicsObject* atomicsObject = nullptr;
    if (Options::useSharedArrayBuffer())
        atomicsObject = AtomicsObject::create(vm, this, AtomicsObject::createStructure(vm, this, m_objectPrototype.get()));
#endif

#define CREATE_CONSTRUCTOR_FOR_SIMPLE_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    capitalName ## Constructor* lowerName ## Constructor = capitalName ## Constructor::create(vm, capitalName ## Constructor::createStructure(vm, this, m_functionPrototype.get()), m_ ## lowerName ## Prototype.get(), speciesGetterSetter()); \
    m_ ## lowerName ## Prototype->putDirectWithoutTransition(vm, propertyNames->constructor, lowerName ## Constructor, DontEnum); \

    FOR_EACH_SIMPLE_BUILTIN_TYPE(CREATE_CONSTRUCTOR_FOR_SIMPLE_TYPE)
    BigIntConstructor* bigIntConstructor = nullptr;
    if (UNLIKELY(Options::useBigInt())) {
        bigIntConstructor = BigIntConstructor::create(vm, BigIntConstructor::createStructure(vm, this, m_functionPrototype.get()), m_bigIntPrototype.get(), m_speciesGetterSetter.get());
        m_bigIntPrototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, bigIntConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
    }

#undef CREATE_CONSTRUCTOR_FOR_SIMPLE_TYPE

    m_dateConstructor.set(vm, this, dateConstructor);
    m_errorConstructor.set(vm, this, errorConstructor);
    m_promiseConstructor.set(vm, this, promiseConstructor);
	m_internalPromiseConstructor.set(vm, this, internalPromiseConstructor);
	m_arrayBufferConstructor.set(vm, this, arrayBufferConstructor);

	NativeErrorConstructorBase* typeErrorConstructor = nullptr;

    {
    	NativeErrorPrototype* prototype = NativeErrorPrototype::create(vm, NativeErrorPrototype::createStructure(vm, this, m_errorPrototype.get()), errorTypeName(ErrorType::EvalError));
    	m_evalErrorStructure.set(vm, this, ErrorInstance::createStructure(vm, this, prototype));
        auto* evalErrorConstructor = NativeErrorConstructor<ErrorType::EvalError>::create(vm, NativeErrorConstructor<ErrorType::EvalError>::createStructure(vm, this, m_errorConstructor.get()), prototype);
        prototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, evalErrorConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
        putDirectWithoutTransition(vm, propertyNames->EvalError, evalErrorConstructor, DontEnum);
    }
    {
    	NativeErrorPrototype* prototype = NativeErrorPrototype::create(vm, NativeErrorPrototype::createStructure(vm, this, m_errorPrototype.get()), errorTypeName(ErrorType::RangeError));
    	m_rangeErrorStructure.set(vm, this, ErrorInstance::createStructure(vm, this, prototype));
        auto* rangeErrorConstructor = NativeErrorConstructor<ErrorType::RangeError>::create(vm, NativeErrorConstructor<ErrorType::RangeError>::createStructure(vm, this, m_errorConstructor.get()), prototype);
        prototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, rangeErrorConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
		putDirectWithoutTransition(vm, propertyNames->RangeError, rangeErrorConstructor, DontEnum);
    }
    {
    	NativeErrorPrototype* prototype = NativeErrorPrototype::create(vm, NativeErrorPrototype::createStructure(vm, this, m_errorPrototype.get()), errorTypeName(ErrorType::ReferenceError));
    	m_referenceErrorStructure.set(vm, this, ErrorInstance::createStructure(vm, this, prototype));
    	auto* referenceErrorConstructor = NativeErrorConstructor<ErrorType::ReferenceError>::create(vm, NativeErrorConstructor<ErrorType::ReferenceError>::createStructure(vm, this, m_errorConstructor.get()), prototype);
    	prototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, referenceErrorConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
		putDirectWithoutTransition(vm, propertyNames->ReferenceError, referenceErrorConstructor, DontEnum);
    }
    {
    	NativeErrorPrototype* prototype = NativeErrorPrototype::create(vm, NativeErrorPrototype::createStructure(vm, this, m_errorPrototype.get()), errorTypeName(ErrorType::SyntaxError));
    	m_syntaxErrorStructure.set(vm, this, ErrorInstance::createStructure(vm, this, prototype));
    	auto* syntaxErrorConstructor = NativeErrorConstructor<ErrorType::SyntaxError>::create(vm, NativeErrorConstructor<ErrorType::SyntaxError>::createStructure(vm, this, m_errorConstructor.get()), prototype);
    	prototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, syntaxErrorConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
		putDirectWithoutTransition(vm, propertyNames->SyntaxError, syntaxErrorConstructor, DontEnum);
    }
    {
    	NativeErrorPrototype* prototype = NativeErrorPrototype::create(vm, NativeErrorPrototype::createStructure(vm, this, m_errorPrototype.get()), errorTypeName(ErrorType::TypeError));
    	m_typeErrorStructure.set(vm, this, ErrorInstance::createStructure(vm, this, prototype));
    	typeErrorConstructor = NativeErrorConstructor<ErrorType::TypeError>::create(vm, NativeErrorConstructor<ErrorType::TypeError>::createStructure(vm, this, m_errorConstructor.get()), prototype);
    	prototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, typeErrorConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
		putDirectWithoutTransition(vm, propertyNames->TypeError, typeErrorConstructor, DontEnum);
    }
    {
    	NativeErrorPrototype* prototype = NativeErrorPrototype::create(vm, NativeErrorPrototype::createStructure(vm, this, m_errorPrototype.get()), errorTypeName(ErrorType::URIError));
    	m_URIErrorStructure.set(vm, this, ErrorInstance::createStructure(vm, this, prototype));
    	auto* URIErrorConstructor = NativeErrorConstructor<ErrorType::URIError>::create(vm, NativeErrorConstructor<ErrorType::URIError>::createStructure(vm, this, m_errorConstructor.get()), prototype);
    	prototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, URIErrorConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
		putDirectWithoutTransition(vm, propertyNames->URIError, URIErrorConstructor, DontEnum);
    }

    {
    	auto* prototype = AggregateErrorPrototype::create(vm, AggregateErrorPrototype::createStructure(vm, this, m_errorPrototype.get()));
    	m_aggregateErrorStructure.set(vm, this, AggregateError::createStructure(vm, this, prototype));
    	auto* aggregateErrorConstructor = AggregateErrorConstructor::create(vm, AggregateErrorConstructor::createStructure(vm, this, m_errorConstructor.get()), prototype);
    	prototype->putDirectWithoutTransition(vm, vm.propertyNames->constructor, aggregateErrorConstructor, static_cast<unsigned>(PropertyAttribute::DontEnum));
		putDirectWithoutTransition(vm, propertyNames->AggregateError, aggregateErrorConstructor, DontEnum);
    }

    m_generatorFunctionPrototype.set(vm, this, GeneratorFunctionPrototype::create(vm, GeneratorFunctionPrototype::createStructure(vm, this, m_functionPrototype.get())));
    GeneratorFunctionConstructor* generatorFunctionConstructor = GeneratorFunctionConstructor::create(vm, GeneratorFunctionConstructor::createStructure(vm, this, functionConstructor), m_generatorFunctionPrototype.get());
    m_generatorFunctionPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, generatorFunctionConstructor, DontEnum | ReadOnly);
    m_generatorFunctionStructure.set(vm, this, JSGeneratorFunction::createStructure(vm, this, m_generatorFunctionPrototype.get()));

    m_generatorPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, m_generatorFunctionPrototype.get(), DontEnum | ReadOnly);
    m_generatorFunctionPrototype->putDirectWithoutTransition(vm, propertyNames->prototype, m_generatorPrototype.get(), DontEnum | ReadOnly);

    m_asyncFunctionPrototype.set(vm, this, AsyncFunctionPrototype::create(vm, AsyncFunctionPrototype::createStructure(vm, this, m_functionPrototype.get())));
    AsyncFunctionConstructor* asyncFunctionConstructor = AsyncFunctionConstructor::create(vm, AsyncFunctionConstructor::createStructure(vm, this, functionConstructor), m_asyncFunctionPrototype.get());
    m_asyncFunctionPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, asyncFunctionConstructor, DontEnum);
    m_asyncFunctionStructure.set(vm, this, JSAsyncFunction::createStructure(vm, this, m_asyncFunctionPrototype.get()));

    m_asyncGeneratorFunctionPrototype.set(vm, this, AsyncGeneratorFunctionPrototype::create(vm, AsyncGeneratorFunctionPrototype::createStructure(vm, this, m_functionPrototype.get())));
    AsyncGeneratorFunctionConstructor* asyncGeneratorFunctionConstructor = AsyncGeneratorFunctionConstructor::create(vm, AsyncGeneratorFunctionConstructor::createStructure(vm, this, functionConstructor), m_asyncGeneratorFunctionPrototype.get());
    m_asyncGeneratorFunctionPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, asyncGeneratorFunctionConstructor, DontEnum | ReadOnly);
    m_asyncGeneratorFunctionStructure.set(vm, this, JSAsyncGeneratorFunction::createStructure(vm, this, m_asyncGeneratorFunctionPrototype.get()));

    m_asyncGeneratorPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, m_asyncGeneratorFunctionPrototype.get(), DontEnum | ReadOnly);
    m_asyncGeneratorFunctionPrototype->putDirectWithoutTransition(vm, propertyNames->prototype, m_asyncGeneratorPrototype.get(), DontEnum | ReadOnly);

    m_objectPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, objectConstructor, DontEnum);
    m_functionPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, functionConstructor, DontEnum);
    m_arrayPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, arrayConstructor, DontEnum);
	m_regExpPrototype->putDirectWithoutTransition(vm, propertyNames->constructor, m_regExpConstructor.get(), DontEnum);

    putDirectWithoutTransition(vm, propertyNames->Object, objectConstructor, DontEnum);
    putDirectWithoutTransition(vm, propertyNames->Function, functionConstructor, DontEnum);
    putDirectWithoutTransition(vm, propertyNames->Array, arrayConstructor, DontEnum);
    putDirectWithoutTransition(vm, propertyNames->RegExp, m_regExpConstructor.get(), DontEnum);

    putDirectWithoutTransition(vm, propertyNames->Proxy, ProxyConstructor::create(vm, ProxyConstructor::createStructure(vm, this, m_functionPrototype.get())), DontEnum);

#if ENABLE(SHARED_ARRAY_BUFFER)
    if (Options::useSharedArrayBuffer()) {
        putDirectWithoutTransition(vm, vm.propertyNames->SharedArrayBuffer, m_sharedArrayBufferStructure.constructor(this), static_cast<unsigned>(PropertyAttribute::DontEnum));
        putDirectWithoutTransition(vm, vm.propertyNames->Atomics, atomicsObject, static_cast<unsigned>(PropertyAttribute::DontEnum));
    }
#endif

#define PUT_CONSTRUCTOR_FOR_SIMPLE_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    putDirectWithoutTransition(vm, propertyNames->jsName, lowerName ## Constructor, DontEnum); \

    FOR_EACH_SIMPLE_BUILTIN_TYPE(PUT_CONSTRUCTOR_FOR_SIMPLE_TYPE)
    if (UNLIKELY(Options::useBigInt())) {
		FOR_BIG_INT_BUILTIN_TYPE_WITH_CONSTRUCTOR(PUT_CONSTRUCTOR_FOR_SIMPLE_TYPE)
    }

#undef PUT_CONSTRUCTOR_FOR_SIMPLE_TYPE

    m_iteratorResultObjectStructure.set(vm, this, createIteratorResultObjectStructure(vm, *this));

    m_evalFunction.set(vm, this, JSFunction::create(vm, this, 1, propertyNames->eval.ustring(), globalFuncEval));
    putDirectWithoutTransition(vm, propertyNames->eval, m_evalFunction.get(), DontEnum);

#if ENABLE(INTL)
    IntlObject* intl = IntlObject::create(vm, this, IntlObject::createStructure(vm, this, m_objectPrototype.get()));
    putDirectWithoutTransition(vm, propertyNames->Intl, intl, DontEnum);
    m_intlObject.set(vm, this, intl);
#endif // ENABLE(INTL)
    putDirectWithoutTransition(vm, propertyNames->JSON, JSONObject::create(vm, JSONObject::createStructure(vm, this, m_objectPrototype.get())), DontEnum);
    ReflectObject* reflectObject = ReflectObject::create(vm, this, ReflectObject::createStructure(vm, this, m_objectPrototype.get()));
	putDirectWithoutTransition(vm, propertyNames->Reflect, reflectObject, DontEnum);

    putDirectWithoutTransition(vm, vm.propertyNames->console, ConsoleObject::create(vm, this, ConsoleObject::createStructure(vm, this, m_objectPrototype.get())), DontEnum);

    JSTypedArrayViewConstructor* typedArraySuperConstructor = JSTypedArrayViewConstructor::create(vm, this, JSTypedArrayViewConstructor::createStructure(vm, this, m_functionPrototype.get()), typedArrayProto, speciesGetterSetter());
    typedArrayProto->putDirectWithoutTransition(vm, propertyNames->constructor, typedArraySuperConstructor, DontEnum);

    m_typedArrays[toIndex(TypeInt8)].constructor.set(vm , this, JSInt8ArrayConstructor::create(vm, this, JSInt8ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeInt8)].prototype.get(), ASCIILiteral("Int8Array")));
    m_typedArrays[toIndex(TypeInt16)].constructor.set(vm, this, JSInt16ArrayConstructor::create(vm, this, JSInt16ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeInt16)].prototype.get(), ASCIILiteral("Int16Array")));
    m_typedArrays[toIndex(TypeInt32)].constructor.set(vm, this, JSInt32ArrayConstructor::create(vm, this, JSInt32ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeInt32)].prototype.get(), ASCIILiteral("Int32Array")));
    m_typedArrays[toIndex(TypeUint8)].constructor.set(vm, this, JSUint8ArrayConstructor::create(vm, this, JSUint8ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeUint8)].prototype.get(), ASCIILiteral("Uint8Array")));
    m_typedArrays[toIndex(TypeUint8Clamped)].constructor.set(vm, this, JSUint8ClampedArrayConstructor::create(vm, this, JSUint8ClampedArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeUint8Clamped)].prototype.get(), ASCIILiteral("Uint8ClampedArray")));
    m_typedArrays[toIndex(TypeUint16)].constructor.set(vm, this, JSUint16ArrayConstructor::create(vm, this, JSUint16ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeUint16)].prototype.get(), ASCIILiteral("Uint16Array")));
    m_typedArrays[toIndex(TypeUint32)].constructor.set(vm, this, JSUint32ArrayConstructor::create(vm, this, JSUint32ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeUint32)].prototype.get(), ASCIILiteral("Uint32Array")));
    m_typedArrays[toIndex(TypeFloat32)].constructor.set(vm, this, JSFloat32ArrayConstructor::create(vm, this, JSFloat32ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeFloat32)].prototype.get(), ASCIILiteral("Float32Array")));
    m_typedArrays[toIndex(TypeFloat64)].constructor.set(vm, this, JSFloat64ArrayConstructor::create(vm, this, JSFloat64ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeFloat64)].prototype.get(), ASCIILiteral("Float64Array")));
    m_typedArrays[toIndex(TypeBigInt64)].constructor.set(vm, this, JSBigInt64ArrayConstructor::create(vm, this, JSBigInt64ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeBigInt64)].prototype.get(), ASCIILiteral("BigInt64Array")));
    m_typedArrays[toIndex(TypeBigUint64)].constructor.set(vm, this, JSBigUint64ArrayConstructor::create(vm, this, JSBigUint64ArrayConstructor::createStructure(vm, this, typedArraySuperConstructor), m_typedArrays[toIndex(TypeBigUint64)].prototype.get(), ASCIILiteral("BigUint64Array")));
	m_typedArrays[toIndex(TypeDataView)].constructor.set(vm, this, JSDataViewConstructor::create(vm, this, JSDataViewConstructor::createStructure(vm, this, m_functionPrototype.get()), m_typedArrays[toIndex(TypeDataView)].prototype.get(), ASCIILiteral("DataView")));

    for (unsigned typedArrayIndex = NumberOfTypedArrayTypes; typedArrayIndex--;) {
        m_typedArrays[typedArrayIndex].prototype->putDirectWithoutTransition(vm, propertyNames->constructor, m_typedArrays[typedArrayIndex].constructor.get(), DontEnum);
        putDirectWithoutTransition(vm, Identifier::fromString(vm, m_typedArrays[typedArrayIndex].constructor.get()->name()), m_typedArrays[typedArrayIndex].constructor.get(), DontEnum);
    }

    putDirectWithoutTransition(vm, propertyNames->builtinNames().Int8ArrayPrivateName(), m_typedArrays[toIndex(TypeInt8)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Int16ArrayPrivateName(), m_typedArrays[toIndex(TypeInt16)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Int32ArrayPrivateName(), m_typedArrays[toIndex(TypeInt32)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Uint8ArrayPrivateName(), m_typedArrays[toIndex(TypeUint8)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Uint8ClampedArrayPrivateName(), m_typedArrays[toIndex(TypeUint8Clamped)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Uint16ArrayPrivateName(), m_typedArrays[toIndex(TypeUint16)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Uint32ArrayPrivateName(), m_typedArrays[toIndex(TypeUint32)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Float32ArrayPrivateName(), m_typedArrays[toIndex(TypeFloat32)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().Float64ArrayPrivateName(), m_typedArrays[toIndex(TypeFloat64)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().BigInt64ArrayPrivateName(), m_typedArrays[toIndex(TypeBigInt64)].constructor.get(), DontEnum);
    putDirectWithoutTransition(vm, propertyNames->builtinNames().BigUint64ArrayPrivateName(), m_typedArrays[toIndex(TypeBigUint64)].constructor.get(), DontEnum);

    m_moduleLoader.set(vm, this, JSModuleLoader::create(globalExec(), vm, this, JSModuleLoader::createStructure(vm, this, jsNull())));
    if (Options::exposeInternalModuleLoader())
        putDirectWithoutTransition(vm, propertyNames->Loader, m_moduleLoader.get(), DontEnum);

    JSFunction* privateFuncAbs = JSFunction::create(vm, this, 0, String(), mathProtoFuncAbs, AbsIntrinsic);
    JSFunction* privateFuncFloor = JSFunction::create(vm, this, 0, String(), mathProtoFuncFloor, FloorIntrinsic);
    JSFunction* privateFuncIsFinite = JSFunction::create(vm, globalObjectIsFiniteCodeGenerator(this), this);
    JSFunction* privateFuncIsNaN = JSFunction::create(vm, globalObjectIsNaNCodeGenerator(this), this);
	JSFunction* privateFuncTrunc = JSFunction::create(vm, this, 0, String(), mathProtoFuncTrunc, TruncIntrinsic);

    JSFunction* privateFuncToLength = JSFunction::create(vm, globalOperationsToLengthCodeGenerator(this), this);
	JSFunction* privateFuncToIntegerOrInfinity = JSFunction::create(vm, globalOperationsToIntegerOrInfinityCodeGenerator(this), this);
	JSFunction* privateFuncTypedArrayLength = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncLength);
	JSFunction* privateFuncTypedArrayGetOriginalConstructor = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncGetOriginalConstructor);
	JSFunction* privateFuncTypedArraySort = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncSort);
	JSFunction* privateFuncTypedArrayContentType = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncContentType);
	JSFunction* privateFuncIsTypedArrayView = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncIsTypedArrayView, IsTypedArrayViewIntrinsic);
	JSFunction* privateFuncIsBigIntTypedArrayView = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncIsBigIntTypedArrayView);
	JSFunction* privateFuncIsNeutered = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncIsNeutered);
	JSFunction* privateFuncTypedArraySubarrayCreate = JSFunction::create(vm, this, 0, String(), typedArrayViewPrivateFuncSubarrayCreate);
	JSFunction* privateFuncIsBoundFunction = JSFunction::create(vm, this, 0, String(), isBoundFunction);
	JSFunction* privateFuncHasInstanceBoundFunction = JSFunction::create(vm, this, 0, String(), hasInstanceBoundFunction);
	JSFunction* privateFuncInstanceOf = JSFunction::create(vm, this, 0, String(), objectPrivateFuncInstanceOf);
	JSFunction* privateFuncThisTimeValue = JSFunction::create(vm, this, 0, String(), dateProtoFuncGetTime);
#if ENABLE(INTL)
    JSFunction* privateFuncDateTimeFormat = JSFunction::create(vm, this, 0, String(), globalFuncDateTimeFormat);
#endif
    JSFunction* privateFuncConcatMemcpy = JSFunction::create(vm, this, 0, String(), arrayProtoPrivateFuncConcatMemcpy);
	JSFunction* privateFuncAppendMemcpy = JSFunction::create(vm, this, 0, String(), arrayProtoPrivateFuncAppendMemcpy);
	JSFunction* privateFuncConcatSlowPath = JSFunction::create(vm, arrayPrototypeConcatSlowPathCodeGenerator(this), this);
	JSFunction* privateFuncIsArraySlow = JSFunction::create(vm, this, 0, String(), arrayConstructorPrivateFuncIsArraySlow);
	JSFunction* privateFuncToBigInt = JSFunction::create(vm, this, 0, String(), globalFuncToBigInt);

	JSFunction* privateFuncMapBucketHead = JSFunction::create(vm, this, 0, String(), mapPrivateFuncMapBucketHead, JSMapBucketHeadIntrinsic);
	JSFunction* privateFuncMapBucketNext = JSFunction::create(vm, this, 0, String(), mapPrivateFuncMapBucketNext, JSMapBucketNextIntrinsic);
	JSFunction* privateFuncMapBucketKey = JSFunction::create(vm, this, 0, String(), mapPrivateFuncMapBucketKey, JSMapBucketKeyIntrinsic);
	JSFunction* privateFuncMapBucketValue = JSFunction::create(vm, this, 0, String(), mapPrivateFuncMapBucketValue, JSMapBucketValueIntrinsic);
	JSFunction* privateFuncSetBucketHead = JSFunction::create(vm, this, 0, String(), setPrivateFuncSetBucketHead, JSSetBucketHeadIntrinsic);
	JSFunction* privateFuncSetBucketNext = JSFunction::create(vm, this, 0, String(), setPrivateFuncSetBucketNext, JSSetBucketNextIntrinsic);
	JSFunction* privateFuncSetBucketKey = JSFunction::create(vm, this, 0, String(), setPrivateFuncSetBucketKey, JSSetBucketKeyIntrinsic);

	JSFunction* privateFuncCreateArrayIterator = JSFunction::create(vm, arrayPrototypeArrayIteratorConstructorCodeGenerator(this), this);
    privateFuncCreateArrayIterator->putDirect(vm, propertyNames->prototype, arrayIteratorPrototype);
    JSFunction* privateFuncArrayIteratorValueNext = JSFunction::create(vm, arrayIteratorPrototypeArrayIteratorValueNextCodeGenerator(this), this);
    JSFunction* privateFuncArrayIteratorKeyNext = JSFunction::create(vm, arrayIteratorPrototypeArrayIteratorKeyNextCodeGenerator(this), this);
    JSFunction* privateFuncArrayIteratorKeyValueNext = JSFunction::create(vm, arrayIteratorPrototypeArrayIteratorKeyValueNextCodeGenerator(this), this);

    JSObject* asyncFromSyncIteratorPrototype = AsyncFromSyncIteratorPrototype::create(vm, this, AsyncFromSyncIteratorPrototype::createStructure(vm, this, m_iteratorPrototype.get()));
    JSFunction* privateFuncAsyncFromSyncIterator = JSFunction::create(vm, asyncFromSyncIteratorPrototypeAsyncFromSyncIteratorConstructorCodeGenerator(this), this);
    privateFuncAsyncFromSyncIterator->putDirect(vm, propertyNames->prototype, asyncFromSyncIteratorPrototype);

    JSObject* mapIteratorPrototype = MapIteratorPrototype::create(vm, this, MapIteratorPrototype::createStructure(vm, this, m_iteratorPrototype.get()));
    JSFunction* createMapIteratorPrivateFunction = JSFunction::create(vm, mapPrototypeMapIteratorConstructorCodeGenerator(this), this);
	createMapIteratorPrivateFunction->putDirect(vm, vm.propertyNames->prototype, mapIteratorPrototype);

    JSObject* setIteratorPrototype = SetIteratorPrototype::create(vm, this, SetIteratorPrototype::createStructure(vm, this, m_iteratorPrototype.get()));
    JSFunction* createSetIteratorPrivateFunction = JSFunction::create(vm, setPrototypeSetIteratorConstructorCodeGenerator(this), this);
	createSetIteratorPrivateFunction->putDirect(vm, vm.propertyNames->prototype, setIteratorPrototype);

    JSObject* regExpStringIteratorPrototype = RegExpStringIteratorPrototype::create(vm, this, RegExpStringIteratorPrototype::createStructure(vm, this, m_iteratorPrototype.get()));
    JSFunction* privateFuncRegExpStringIterator = JSFunction::create(vm, regExpPrototypeRegExpStringIteratorConstructorCodeGenerator(this), this);
    privateFuncRegExpStringIterator->putDirect(vm, vm.propertyNames->prototype, regExpStringIteratorPrototype);

    GetterSetter* regExpProtoFlagsGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->flags);
    GetterSetter* regExpProtoHasIndicesGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->hasIndices);
	GetterSetter* regExpProtoGlobalGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->global);
    m_regExpProtoGlobalGetter.set(vm, this, regExpProtoGlobalGetter);
    GetterSetter* regExpProtoIgnoreCaseGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->ignoreCase);
    GetterSetter* regExpProtoMultilineGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->multiline);
    GetterSetter* regExpProtoSourceGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->source);
    GetterSetter* regExpProtoStickyGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->sticky);
    GetterSetter* regExpProtoUnicodeGetter = getGetterById(vm, exec, m_regExpPrototype.get(), propertyNames->unicode);
    GetterSetter* regExpProtoDotAllGetter = getGetterById(vm ,exec, m_regExpPrototype.get(), propertyNames->dotAll);
	m_regExpProtoUnicodeGetter.set(vm, this, regExpProtoUnicodeGetter);
    JSObject* builtinRegExpExec = asObject(m_regExpPrototype->getDirect(vm, propertyNames->exec).asCell());
    m_regExpProtoExec.set(vm, this, builtinRegExpExec);
    JSObject* regExpSymbolReplace = asObject(m_regExpPrototype->getDirect(vm, propertyNames->replaceSymbol).asCell());
    m_regExpProtoSymbolReplace.set(vm, this, regExpSymbolReplace);

    JSFunction* privateFuncsetPrototypeDirect = JSFunction::create(vm, this, 1, String(), globalFuncSetPrototypeDirect, NoIntrinsic);
    m_setPrototypeDirectFunction.set(vm, this, privateFuncsetPrototypeDirect);

    JSFunction* copyDataPropertiesPrivateFunction = JSFunction::create(vm, this, 2, String(), globalFuncCopyDataProperties);
    m_linkTimeConstants[static_cast<unsigned>(LinkTimeConstant::copyDataProperties)].set(m_vm, this, copyDataPropertiesPrivateFunction);

    {
    	DeferGC deferGC(vm.heap);
		GlobalPropertyInfo staticGlobals[] = {
			GlobalPropertyInfo(propertyNames->Math, MathObject::create(vm, this, MathObject::createStructure(vm, this, m_objectPrototype.get())), DontEnum | DontDelete),
			GlobalPropertyInfo(propertyNames->NaN, jsNaN(), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(propertyNames->Infinity, jsNumber(std::numeric_limits<double>::infinity()), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(propertyNames->undefinedKeyword, jsUndefined(), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.ObjectPrivateName(), objectConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.enqueueJobPrivateName(), JSFunction::create(vm, this, 0, ASCIILiteral("enqueueJob"), enqueueJob), DontEnum | DontDelete | ReadOnly),
			// FIXME: Offer @makeTypeError function instead of exposing @TypeError here.
			// https://bugs.webkit.org/show_bug.cgi?id=193858
			GlobalPropertyInfo(builtinNames.TypeErrorPrivateName(), typeErrorConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArrayLengthPrivateName(), privateFuncTypedArrayLength, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArrayGetOriginalConstructorPrivateName(), privateFuncTypedArrayGetOriginalConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArraySortPrivateName(), privateFuncTypedArraySort, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArrayContentTypePrivateName(), privateFuncTypedArrayContentType, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isTypedArrayViewPrivateName(), privateFuncIsTypedArrayView, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isBigIntTypedArrayViewPrivateName(), privateFuncIsBigIntTypedArrayView, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isNeuteredPrivateName(), privateFuncIsNeutered, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArraySubarrayCreatePrivateName(), privateFuncTypedArraySubarrayCreate, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isBoundFunctionPrivateName(), privateFuncIsBoundFunction, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.hasInstanceBoundFunctionPrivateName(), privateFuncHasInstanceBoundFunction, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.instanceOfPrivateName(), privateFuncInstanceOf, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.StringPrivateName(), stringConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.ArrayPrivateName(), arrayConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.NumberPrivateName(), numberConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.RegExpPrivateName(), m_regExpConstructor.get(), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.absPrivateName(), privateFuncAbs, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.floorPrivateName(), privateFuncFloor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.minPrivateName(), JSFunction::create(vm, this, 0, String(), mathProtoFuncMin, MinIntrinsic), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.truncPrivateName(), privateFuncTrunc, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isFinitePrivateName(), privateFuncIsFinite, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isNaNPrivateName(), privateFuncIsNaN, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.PromisePrivateName(), promiseConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.ReflectPrivateName(), reflectObject, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.InternalPromisePrivateName(), internalPromiseConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.toBigIntPrivateName(), privateFuncToBigInt, DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.arraySpeciesCreatePrivateName(), JSFunction::create(vm, this, 2, String(), arrayProtoFuncSpeciesCreate), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isArrayPrivateName(), arrayConstructor->getDirect(vm, propertyNames->isArray), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.concatMemcpyPrivateName(), privateFuncConcatMemcpy, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.appendMemcpyPrivateName(), privateFuncAppendMemcpy, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.concatSlowPathPrivateName(), privateFuncConcatSlowPath, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isArraySlowPrivateName(), privateFuncIsArraySlow, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.symbolIteratorPrivateName(), Symbol::create(vm, static_cast<SymbolImpl&>(*propertyNames->iteratorSymbol.impl())), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.repeatCharacterPrivateName(), JSFunction::create(vm, this, 2, String(), stringProtoFuncRepeatCharacter), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.repeatSlowPathPrivateName(), JSFunction::create(vm, stringPrototypeRepeatSlowPathCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.repeatCharactersSlowPathPrivateName(), JSFunction::create(vm, stringPrototypeRepeatCharactersSlowPathCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.arrayIteratorValueNextPrivateName(), privateFuncArrayIteratorValueNext, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.arrayIteratorKeyNextPrivateName(), privateFuncArrayIteratorKeyNext, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.arrayIteratorKeyValueNextPrivateName(), privateFuncArrayIteratorKeyValueNext, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.ArrayIteratorPrivateName(), privateFuncCreateArrayIterator, DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.sortStringComparatorPrivateName(), JSFunction::create(vm, arrayPrototypeSortStringComparatorCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.sortCompactPrivateName(), JSFunction::create(vm, arrayPrototypeSortCompactCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.sortCommitPrivateName(), JSFunction::create(vm, arrayPrototypeSortCommitCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.sortMergePrivateName(), JSFunction::create(vm, arrayPrototypeSortMergeCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.sortMergeSortPrivateName(), JSFunction::create(vm, arrayPrototypeSortMergeSortCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.sortBucketSortPrivateName(), JSFunction::create(vm, arrayPrototypeSortBucketSortCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.createAsyncFromSyncIteratorPrivateName(), JSFunction::create(vm, asyncFromSyncIteratorPrototypeCreateAsyncFromSyncIteratorCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.AsyncFromSyncIteratorPrivateName(), privateFuncAsyncFromSyncIterator, DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.RegExpStringIteratorPrivateName(), privateFuncRegExpStringIterator, DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.asyncGeneratorQueueIsEmptyPrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorQueueIsEmptyCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorQueueEnqueuePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorQueueEnqueueCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorQueueDequeuePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorQueueDequeueCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorDequeuePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorDequeueCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isExecutionStatePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeIsExecutionStateCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isSuspendYieldStatePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeIsSuspendYieldStateCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorRejectPrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorRejectCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorResolvePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorResolveCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorYieldPrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorYieldCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.awaitValuePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAwaitValueCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.doAsyncGeneratorBodyCallPrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeDoAsyncGeneratorBodyCallCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorResumeNextPrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorResumeNextCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.asyncGeneratorEnqueuePrivateName(), JSFunction::create(vm, asyncGeneratorPrototypeAsyncGeneratorEnqueueCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.toLengthPrivateName(), privateFuncToLength, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.toIntegerOrInfinityPrivateName(), privateFuncToIntegerOrInfinity, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isDictionaryPrivateName(), JSFunction::create(vm, globalOperationsIsDictionaryCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isPromisePrivateName(), JSFunction::create(vm, operationsPromiseIsPromiseCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.newPromiseReactionPrivateName(), JSFunction::create(vm, operationsPromiseNewPromiseReactionCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.newPromiseCapabilityPrivateName(), m_newPromiseCapabilityFunction.get(), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.promiseResolvePrivateName(), JSFunction::create(vm, operationsPromisePromiseResolveCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.newHandledRejectedPromisePrivateName(), JSFunction::create(vm, operationsPromiseNewHandledRejectedPromiseCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.triggerPromiseReactionsPrivateName(), JSFunction::create(vm, operationsPromiseTriggerPromiseReactionsCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.rejectPromisePrivateName(), JSFunction::create(vm, operationsPromiseRejectPromiseCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.fulfillPromisePrivateName(), JSFunction::create(vm, operationsPromiseFulfillPromiseCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.createResolvingFunctionsPrivateName(), JSFunction::create(vm, operationsPromiseCreateResolvingFunctionsCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.promiseReactionJobPrivateName(), JSFunction::create(vm, operationsPromisePromiseReactionJobCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.promiseResolveThenableJobPrivateName(), JSFunction::create(vm, operationsPromisePromiseResolveThenableJobCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArrayElementComparePrivateName(), JSFunction::create(vm, typedArrayPrototypeTypedArrayElementCompareCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArrayMergePrivateName(), JSFunction::create(vm, typedArrayPrototypeTypedArrayMergeCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArrayMergeSortPrivateName(), JSFunction::create(vm, typedArrayPrototypeTypedArrayMergeSortCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.typedArraySpeciesConstructorPrivateName(), JSFunction::create(vm, typedArrayPrototypeTypedArraySpeciesConstructorCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.hostPromiseRejectionTrackerPrivateName(), JSFunction::create(vm, this, 2, String(), globalFuncHostPromiseRejectionTracker), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.MapPrivateName(), mapConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.SetPrivateName(), setConstructor, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.generatorResumePrivateName(), JSFunction::create(vm, generatorPrototypeGeneratorResumeCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.thisTimeValuePrivateName(), privateFuncThisTimeValue, DontEnum | DontDelete | ReadOnly),
#if ENABLE(INTL)
			GlobalPropertyInfo(builtinNames.dateTimeFormatPrivateName(), privateFuncDateTimeFormat, PropertyAttribute::DontEnum | PropertyAttribute::DontDelete | PropertyAttribute::ReadOnly),
#endif // ENABLE(INTL)

			GlobalPropertyInfo(builtinNames.sameValuePrivateName(), JSFunction::create(vm, this, 2, String(), objectConstructorIs, ObjectIsIntrinsic), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.setPrototypeDirectPrivateName(), privateFuncsetPrototypeDirect, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.speciesConstructorPrivateName(), JSFunction::create(vm, globalOperationsSpeciesConstructorCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.copyDataPropertiesPrivateName(), copyDataPropertiesPrivateFunction, DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.regExpProtoFlagsGetterPrivateName(), regExpProtoFlagsGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoHasIndicesGetterPrivateName(), regExpProtoHasIndicesGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoGlobalGetterPrivateName(), regExpProtoGlobalGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoIgnoreCaseGetterPrivateName(), regExpProtoIgnoreCaseGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoMultilineGetterPrivateName(), regExpProtoMultilineGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoSourceGetterPrivateName(), regExpProtoSourceGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoStickyGetterPrivateName(), regExpProtoStickyGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoUnicodeGetterPrivateName(), regExpProtoUnicodeGetter, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpProtoDotAllGetterPrivateName(), regExpProtoDotAllGetter, DontEnum | DontDelete | ReadOnly),

			// RegExp.prototype helpers.
			GlobalPropertyInfo(builtinNames.regExpBuiltinExecPrivateName(), builtinRegExpExec, DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpCreatePrivateName(), JSFunction::create(vm, this, 2, String(), esSpecRegExpCreate, NoIntrinsic), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.isRegExpPrivateName(), JSFunction::create(vm, this, 1, String(), esSpecIsRegExp, NoIntrinsic), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.hasObservableSideEffectsForRegExpMatchPrivateName(), JSFunction::create(vm, regExpPrototypeHasObservableSideEffectsForRegExpMatchCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.hasObservableSideEffectsForRegExpSplitPrivateName(), JSFunction::create(vm, regExpPrototypeHasObservableSideEffectsForRegExpSplitCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.advanceStringIndexPrivateName(), JSFunction::create(vm, regExpPrototypeAdvanceStringIndexCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpExecPrivateName(), JSFunction::create(vm, regExpPrototypeRegExpExecCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.matchSlowPrivateName(), JSFunction::create(vm, regExpPrototypeMatchSlowCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpMatchFastPrivateName(), JSFunction::create(vm, this, 1, String(), regExpProtoFuncMatchFast, RegExpMatchFastIntrinsic), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpSearchFastPrivateName(), JSFunction::create(vm, this, 1, String(), regExpProtoFuncSearchFast), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpSplitFastPrivateName(), JSFunction::create(vm, this, 2, String(), regExpProtoFuncSplitFast), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpPrototypeSymbolMatchPrivateName(), m_regExpPrototype->getDirect(vm, propertyNames->matchSymbol), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpPrototypeSymbolReplacePrivateName(), m_regExpPrototype->getDirect(vm, propertyNames->replaceSymbol), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.regExpTestFastPrivateName(), JSFunction::create(vm, this, 1, String(), regExpProtoFuncTestFast, RegExpTestFastIntrinsic), DontEnum | DontDelete | ReadOnly),

			// String.prototype helpers.
			GlobalPropertyInfo(builtinNames.hasObservableSideEffectsForStringReplacePrivateName(), JSFunction::create(vm, stringPrototypeHasObservableSideEffectsForStringReplaceCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.createHTMLPrivateName(), JSFunction::create(vm, stringPrototypeCreateHTMLCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.stringConcatSlowPathPrivateName(), JSFunction::create(vm, stringPrototypeStringConcatSlowPathCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.stringIncludesInternalPrivateName(), JSFunction::create(vm, this, 1, String(), builtinStringIncludesInternal), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.stringIndexOfInternalPrivateName(), JSFunction::create(vm, this, 1, String(), builtinStringIndexOfInternal), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.stringSplitFastPrivateName(), JSFunction::create(vm, this, 2, String(), stringProtoFuncSplitFast), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.stringSubstringInternalPrivateName(), JSFunction::create(vm, this, 2, String(), builtinStringSubstringInternal), DontEnum | DontDelete | ReadOnly),

			// Array.prototype helpers
			GlobalPropertyInfo(builtinNames.flatIntoArrayPrivateName(), JSFunction::create(vm, arrayPrototypeFlatIntoArrayCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.flatIntoArrayWithCallbackPrivateName(), JSFunction::create(vm, arrayPrototypeFlatIntoArrayWithCallbackCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),

			// Function prototype helpers.
			GlobalPropertyInfo(builtinNames.makeBoundFunctionPrivateName(), JSFunction::create(vm, this, 5, String(), makeBoundFunction), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.hasOwnLengthPropertyPrivateName(), JSFunction::create(vm, this, 1, String(), hasOwnLengthProperty), DontEnum | DontDelete | ReadOnly),

	        // Map and Set helpers.
	        GlobalPropertyInfo(builtinNames.mapBucketHeadPrivateName(), privateFuncMapBucketHead, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.mapBucketNextPrivateName(), privateFuncMapBucketNext, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.mapBucketKeyPrivateName(), privateFuncMapBucketKey, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.mapBucketValuePrivateName(), privateFuncMapBucketValue, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.mapIteratorNextPrivateName(), JSFunction::create(vm, mapIteratorPrototypeMapIteratorNextCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.setBucketHeadPrivateName(), privateFuncSetBucketHead, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.setBucketNextPrivateName(), privateFuncSetBucketNext, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.setBucketKeyPrivateName(), privateFuncSetBucketKey, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.MapIteratorPrivateName(), createMapIteratorPrivateFunction, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.SetIteratorPrivateName(), createSetIteratorPrivateFunction, DontEnum | DontDelete | ReadOnly),
	        GlobalPropertyInfo(builtinNames.setIteratorNextPrivateName(), JSFunction::create(vm, setIteratorPrototypeSetIteratorNextCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.asyncFunctionResumePrivateName(), JSFunction::create(vm, asyncFunctionPrototypeAsyncFunctionResumeCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),

			GlobalPropertyInfo(builtinNames.setStateToMaxPrivateName(), JSFunction::create(vm, moduleLoaderSetStateToMaxCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
			GlobalPropertyInfo(builtinNames.newRegistryEntryPrivateName(), JSFunction::create(vm, moduleLoaderNewRegistryEntryCodeGenerator(this), this), DontEnum | DontDelete | ReadOnly),
		};
		addStaticGlobals(staticGlobals, WTF_ARRAY_LENGTH(staticGlobals));
    }

    m_specialPointers[Special::CallFunction] = m_callFunction.get();
    m_specialPointers[Special::ApplyFunction] = m_applyFunction.get();
    m_specialPointers[Special::ObjectConstructor] = objectConstructor;
    m_specialPointers[Special::ArrayConstructor] = arrayConstructor;

    if(Options::enableDebugDFGJIT()){
        putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "DFGTrue"), 0, functionFalse1, DFGTrueIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "OSRExit"), 0, functionUndefined1, OSRExitIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "isFinalTier"), 0, functionFalse2, IsFinalTierIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "numberOfDFGCompiles"), 1, numberOfDFGCompiles, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "reoptimizationRetryCount"), 1, functionReoptimizationRetryCount, NoIntrinsic, DontEnum);

    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "fullGC"), 0, functionFullGC, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "noInline"), 1, functionNeverInlineFunction, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "noDFG"), 1, functionNoDFG, NoIntrinsic, DontEnum);
		putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "effectful42"), 0, functionEffectful42, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "makeMasquerader"), 0, functionMakeMasquerader, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createImpureGetter"), 1, functionCreateImpureGetter, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createRuntimeArray"), 0, functionCreateRuntimeArray, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createCustomGetterObject"), 0, functionCreateCustomGetterObject, NoIntrinsic, DontEnum);
		putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "setImpureGetterDelegate"), 2, functionSetImpureGetterDelegate, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createStaticCustomAccessor"), 0, functionCreateStaticCustomAccessor, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "transferArrayBuffer"), 1, functionTransferArrayBuffer, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createProxy"), 1, functionCreateProxy, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "hasCustomProperties"), 1, functionHasCustomProperties, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createBuiltin"), 1, functionCreateBuiltin, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "ensureArrayStorage"), 0, functionEnsureArrayStorage, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createGlobalObject"), 0, functionCreateGlobalObject, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "runString"), 1, functionRunString, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "asDoubleNumber"), 1, functionAsDoubleNumber, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "globalObjectCount"), 0, functionGlobalObjectCount, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "evalScript"), 1, functionEvalScript, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "toUncacheableDictionary"), 1, functionToUncacheableDictionary, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "addressOf"), 1, functionAddressOf, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "globalObjectForObject"), 1, functionGlobalObjectForObject, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createNonRopeNonAtomString"), 1, functionCreateNonRopeNonAtomString, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "toCacheableDictionary"), 1, functionToCacheableDictionary, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "describe"), 1, functionDescribe, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createStaticDontDeleteDontEnum"), 0, functionCreateStaticDontDeleteDontEnum, NoIntrinsic, DontEnum);

    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "loadGetterFromGetterSetter"), 1, functionLoadGetterFromGetterSetter, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createCustomTestGetterSetter"), 0, functionCreateCustomTestGetterSetter, NoIntrinsic, DontEnum);
    	putDirectNativeFunctionWithoutTransition(vm, this, Identifier::fromString(vm, "createStaticCustomValue"), 0, functionCreateStaticCustomValue, NoIntrinsic, DontEnum);

    	{
			CustomGetterSetter* custom = CustomGetterSetter::create(vm, dummyCustomGetValue, testCustomAccessorSetter);
			Identifier identifier = Identifier::fromString(vm, "testCustomAccessorSetter");
			this->putDirectCustomAccessor(vm, identifier, custom, PropertyAttribute::DontEnum | PropertyAttribute::DontDelete | PropertyAttribute::CustomAccessor);
		}

		{
			CustomGetterSetter* custom = CustomGetterSetter::create(vm, dummyCustomGetValue, testCustomValueSetter);
			Identifier identifier = Identifier::fromString(vm, "testCustomValueSetter");
			this->putDirectCustomAccessor(vm, identifier, custom, PropertyAttribute::DontEnum | PropertyAttribute::DontDelete | PropertyAttribute::CustomValue);
		}
	}

    {
        auto setupAdaptiveWatchpoint = [&] (JSObject* base, const Identifier& ident) {
			// Performing these gets should not throw.
			PropertySlot slot(base, PropertySlot::InternalMethodType::Get);
			bool result = base->getOwnPropertySlot(base, exec, ident, slot);
			ASSERT_UNUSED(result, result);
			//catchScope.assertNoException();
			RELEASE_ASSERT(slot.isCacheableValue());
			JSValue functionValue = slot.getValue(exec, ident);
			//catchScope.assertNoException();
			ASSERT(jsDynamicCast<JSFunction*>(vm, functionValue));

			ObjectPropertyCondition condition = generateConditionForSelfEquivalence(vm, 0, base, ident.impl());
			RELEASE_ASSERT(condition.requiredValue() == functionValue);

			bool isWatchable = condition.isWatchable(vm, PropertyCondition::EnsureWatchability);
			RELEASE_ASSERT(isWatchable); // We allow this to install the necessary watchpoints.

			return condition;
		};

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(arrayIteratorPrototype, m_vm.propertyNames->next);
            m_arrayIteratorPrototypeNext = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_arrayIteratorProtocolWatchpointSet));
            m_arrayIteratorPrototypeNext->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(this->arrayPrototype(), m_vm.propertyNames->iteratorSymbol);
            m_arrayPrototypeSymbolIteratorWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_arrayIteratorProtocolWatchpointSet));
            m_arrayPrototypeSymbolIteratorWatchpoint->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(mapIteratorPrototype, m_vm.propertyNames->next);
            m_mapIteratorPrototypeNextWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_mapIteratorProtocolWatchpoint));
            m_mapIteratorPrototypeNextWatchpoint->install(vm);
        }
        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(m_mapPrototype.get(), m_vm.propertyNames->iteratorSymbol);
            m_mapPrototypeSymbolIteratorWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_mapIteratorProtocolWatchpoint));
            m_mapPrototypeSymbolIteratorWatchpoint->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(setIteratorPrototype, m_vm.propertyNames->next);
            m_setIteratorPrototypeNextWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_setIteratorProtocolWatchpoint));
            m_setIteratorPrototypeNextWatchpoint->install(vm);
        }
        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(m_setPrototype.get(), m_vm.propertyNames->iteratorSymbol);
            m_setPrototypeSymbolIteratorWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_setIteratorProtocolWatchpoint));
            m_setPrototypeSymbolIteratorWatchpoint->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(m_stringIteratorPrototype.get(), m_vm.propertyNames->next);
            m_stringIteratorPrototypeNextWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_stringIteratorProtocolWatchpoint));
            m_stringIteratorPrototypeNextWatchpoint->install(vm);
        }
        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(m_stringPrototype.get(), m_vm.propertyNames->iteratorSymbol);
            m_stringPrototypeSymbolIteratorWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_stringIteratorProtocolWatchpoint));
            m_stringPrototypeSymbolIteratorWatchpoint->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(m_mapPrototype.get(), m_vm.propertyNames->set);
            m_mapPrototypeSetWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_mapSetWatchpoint));
            m_mapPrototypeSetWatchpoint->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(m_setPrototype.get(), m_vm.propertyNames->add);
            m_setPrototypeAddWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_setAddWatchpoint));
            m_setPrototypeAddWatchpoint->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(this->arrayPrototype(), m_vm.propertyNames->join);
            m_arrayPrototypeJoinWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_arrayJoinWatchpointSet));
            m_arrayPrototypeJoinWatchpoint->install(vm);
        }

        {
            ObjectPropertyCondition condition = setupAdaptiveWatchpoint(numberPrototype(), propertyNames->toString);
            m_numberPrototypeToStringWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, condition, m_numberToStringWatchpointSet));
            m_numberPrototypeToStringWatchpoint->install(vm);
            m_numberProtoToStringFunction.set(vm, this, jsCast<JSFunction*>(numberPrototype()->getDirect(vm, propertyNames->toString)));
        }
    }

    resetPrototype(vm, prototype);
}

JSCell* JSGlobalObject::jsCellForLinkTimeConstant(LinkTimeConstant type)
{
    unsigned index = static_cast<unsigned>(type);
    ASSERT(index < LinkTimeConstantCount);
    if (!m_linkTimeConstants[index].get()) {
    	switch (type) {
    	case LinkTimeConstant::ThrowTypeErrorFunction:
    	    m_linkTimeConstants[index].set(m_vm, this, JSFunction::create(m_vm, this, 0, String(), globalFuncThrowTypeError));
    		break;
    	case LinkTimeConstant::ImportModule:
    		m_linkTimeConstants[index].set(m_vm, this, JSFunction::create(m_vm, this, 0, String(), globalFuncImportModule));
    		break;
    	case LinkTimeConstant::CreatePrivateSymbol:
    	    m_linkTimeConstants[index].set(m_vm, this, JSFunction::create(m_vm, this, 0, String(), createPrivateSymbol));
    		break;
    	case LinkTimeConstant::copyDataProperties:
    		// Already set
    		break;
    	default:
    		break;
    	}
    }
    return m_linkTimeConstants[index].get();
}

// Private namespace for helpers for JSGlobalObject::haveABadTime()
namespace {

class GlobalObjectDependencyFinder final : public MarkedBlock::VoidFunctor {
public:
    GlobalObjectDependencyFinder(VM& vm)
        : m_vm(vm)
    { }

    IterationStatus operator()(JSCell* cell) const;

    void addDependency(JSGlobalObject* key, JSGlobalObject* dependent);
    HashSet<JSGlobalObject*>* dependentsFor(JSGlobalObject* key);

private:
    void visit(JSObject*);

    VM& m_vm;
    HashMap<JSGlobalObject*, HashSet<JSGlobalObject*>> m_dependencies;
};

inline void GlobalObjectDependencyFinder::addDependency(JSGlobalObject* key, JSGlobalObject* dependent)
{
    auto keyResult = m_dependencies.add(key, HashSet<JSGlobalObject*>());
    keyResult.iterator->second.add(dependent);
}

inline HashSet<JSGlobalObject*>* GlobalObjectDependencyFinder::dependentsFor(JSGlobalObject* key)
{
    auto iterator = m_dependencies.find(key);
    if (iterator == m_dependencies.end())
        return 0;
    return &iterator->second;
}

inline void GlobalObjectDependencyFinder::visit(JSObject* object)
{
    VM& vm = m_vm;

    if (!vm.prototypeMap.isPrototype(object))
        return;

    JSObject* current = object;
    JSGlobalObject* objectGlobalObject = object->globalObject();
    do {
        JSValue prototypeValue = current->getPrototypeDirect();
        if (prototypeValue.isNull())
            return;
        current = asObject(prototypeValue);

        JSGlobalObject* protoGlobalObject = current->globalObject();
        if (protoGlobalObject != objectGlobalObject)
            addDependency(protoGlobalObject, objectGlobalObject);
    } while (true);
}

IterationStatus GlobalObjectDependencyFinder::operator()(JSCell* cell) const
{
    if (cell->isObject()) {
        // FIXME: This const_cast exists because this isn't a C++ lambda.
        // https://bugs.webkit.org/show_bug.cgi?id=159644
        const_cast<GlobalObjectDependencyFinder*>(this)->visit(jsCast<JSObject*>(static_cast<JSCell*>(cell)));
    }
    return IterationStatus::Continue;
}

enum class BadTimeFinderMode {
    SingleGlobal,
    MultipleGlobals
};

template<BadTimeFinderMode mode>
class ObjectsWithBrokenIndexingFinder : public MarkedBlock::VoidFunctor {
public:
    ObjectsWithBrokenIndexingFinder(VM&, Vector<JSObject*>&, JSGlobalObject*);
    ObjectsWithBrokenIndexingFinder(VM&, Vector<JSObject*>&, HashSet<JSGlobalObject*>&);

    bool needsMultiGlobalsScan() const { return m_needsMultiGlobalsScan; }
    IterationStatus operator()(JSCell* cell) const;

private:
    IterationStatus visit(JSObject*);

    VM& m_vm;
    Vector<JSObject*>& m_foundObjects;
    JSGlobalObject* m_globalObject; // Only used for SingleBadTimeGlobal mode.
    HashSet<JSGlobalObject*>* m_globalObjects; // Only used for BadTimeGlobalGraph mode;
    bool m_needsMultiGlobalsScan;
};

template<>
ObjectsWithBrokenIndexingFinder<BadTimeFinderMode::SingleGlobal>::ObjectsWithBrokenIndexingFinder(
    VM& vm, Vector<JSObject*>& foundObjects, JSGlobalObject* globalObject)
    : m_vm(vm)
    , m_foundObjects(foundObjects)
    , m_globalObject(globalObject)
	, m_globalObjects(0)
	, m_needsMultiGlobalsScan(false)
{
}

template<>
ObjectsWithBrokenIndexingFinder<BadTimeFinderMode::MultipleGlobals>::ObjectsWithBrokenIndexingFinder(
    VM& vm, Vector<JSObject*>& foundObjects, HashSet<JSGlobalObject*>& globalObjects)
    : m_vm(vm)
    , m_foundObjects(foundObjects)
    , m_globalObject(0)
    , m_globalObjects(&globalObjects)
	, m_needsMultiGlobalsScan(false)
{
}

inline bool hasBrokenIndexing(IndexingType type)
{
    return type && !hasSlowPutArrayStorage(type);
}

inline bool hasBrokenIndexing(JSObject* object)
{
    IndexingType type = object->structure()->indexingType();
    return hasBrokenIndexing(type);
}

template<BadTimeFinderMode mode>
inline IterationStatus ObjectsWithBrokenIndexingFinder<mode>::visit(JSObject* object)
{
    VM& vm = m_vm;

    // We only want to have a bad time in the affected global object, not in the entire
    // VM. But we have to be careful, since there may be objects that claim to belong to
    // a different global object that have prototypes from our global object.
    auto isInAffectedGlobalObject = [&] (JSObject* object) {
		JSGlobalObject* objectGlobalObject = nullptr;
		bool objectMayBePrototype = false;

		if (mode == BadTimeFinderMode::SingleGlobal) {
			objectGlobalObject = object->globalObject();
			if (objectGlobalObject == m_globalObject)
				return true;

			objectMayBePrototype = objectGlobalObject->vm().prototypeMap.isPrototype(object);
		}

		for (JSObject* current = object; ;) {
			JSGlobalObject* currentGlobalObject = current->globalObject();
			if (mode == BadTimeFinderMode::SingleGlobal) {
				if (objectMayBePrototype && currentGlobalObject != objectGlobalObject)
					m_needsMultiGlobalsScan = true;
				if (currentGlobalObject == m_globalObject)
					return true;
			} else {
				if (m_globalObjects->contains(currentGlobalObject))
					return true;
			}

			JSValue prototypeValue = current->getPrototypeDirect();
			if (prototypeValue.isNull())
				return false;
			current = asObject(prototypeValue);
		}
		RELEASE_ASSERT_NOT_REACHED();
		return true;
    };

    auto checkStructureHasRelevantGlobalObject = [&](Structure* structure) -> bool {
		if (hasBrokenIndexing(structure->indexingType())) {
			 bool isRelevantGlobalObject =
				(mode == BadTimeFinderMode::SingleGlobal
					? m_globalObject == structure->globalObject()
					: m_globalObjects->contains(structure->globalObject()))
				|| (structure->hasMonoProto() && !structure->storedPrototype().isNull() && isInAffectedGlobalObject(asObject(structure->storedPrototype())));
			 return isRelevantGlobalObject;
		 }
		 return false;
	 };

	if (object->inherits<JSFunction>()) {
		JSFunction* function = jsCast<JSFunction*>(object);
		if (FunctionRareData* rareData = function->rareData()) {
			// We only use this to cache JSFinalObjects. They do not start off with a broken indexing type.
			ASSERT(!(rareData->objectAllocationStructure() && hasBrokenIndexing(rareData->objectAllocationStructure()->indexingType())));

			if (Structure* structure = rareData->internalFunctionAllocationStructure()) {
				bool isRelevantGlobalObject = checkStructureHasRelevantGlobalObject(structure);

				if (mode == BadTimeFinderMode::SingleGlobal && m_needsMultiGlobalsScan)
					return IterationStatus::Done; // Bailing early and let the MultipleGlobals path handle everything.
				if (isRelevantGlobalObject)
					rareData->clearInternalFunctionAllocationProfile();
			}
		}
	}

    if (object->inherits<JSGlobalObject>()) {
        JSGlobalObject* globalObject = jsCast<JSGlobalObject*>(object);
        // If this globalObject is already having a bad time, then structures in its StructureCache
        // does not affect on this new JSGlobalObject's haveABadTime since they are already slow mode.
        if (!globalObject->isHavingABadTime()) {
            VM& vm = globalObject->vm();
            ASSERT(vm.heap.isDeferred());
            bool willClear = false;
            globalObject->structureCache().forEach([&](Structure* structure) {
                bool isRelevantGlobalObject = checkStructureHasRelevantGlobalObject(structure);
                if (mode == BadTimeFinderMode::SingleGlobal && m_needsMultiGlobalsScan)
                    return IterationStatus::Done;
                if (isRelevantGlobalObject)
                    willClear = true;
                return IterationStatus::Continue;
            });
            if (mode == BadTimeFinderMode::SingleGlobal && m_needsMultiGlobalsScan)
                return IterationStatus::Done; // Bailing early and let the MultipleGlobals path handle everything.

            // StructureCache contains Structures which is no longer valid after relevant JSGlobalObject's haveABadTime.
            // We do not make such a JSGlobalObject status haveABadTime since still its own objects are intact.
            if (willClear)
                globalObject->clearStructureCache(vm);
        }
    }

    // Run this filter first, since it's cheap, and ought to filter out a lot of objects.
    if (!hasBrokenIndexing(object))
        return IterationStatus::Continue;

    if (isInAffectedGlobalObject(object))
        m_foundObjects.append(object);

    if (mode == BadTimeFinderMode::SingleGlobal && m_needsMultiGlobalsScan)
        return IterationStatus::Done; // Bailing early and let the MultipleGlobals path handle everything.

    return IterationStatus::Continue;
}

template<BadTimeFinderMode mode>
IterationStatus ObjectsWithBrokenIndexingFinder<mode>::operator()(JSCell* cell) const
{
    if (cell->isObject()) {
        // FIXME: This const_cast exists because this isn't a C++ lambda.
        // https://bugs.webkit.org/show_bug.cgi?id=159644
        return const_cast<ObjectsWithBrokenIndexingFinder*>(this)->visit(jsCast<JSObject*>(cell));
    }
    return IterationStatus::Continue;
}

} // end private namespace for helpers for JSGlobalObject::haveABadTime()

void JSGlobalObject::fireWatchpointAndMakeAllArrayStructuresSlowPut(VM& vm)
{
    if (isHavingABadTime())
        return;

    // This must happen first, because the compiler thread may race with haveABadTime.
    // Let R_BT, W_BT <- Read/Fire the watchpoint, R_SC, W_SC <- Read/clear the structure cache.
    // The possible interleavings are:
    // R_BT, R_SC, W_SC, W_BT: Compiler thread installs a watchpoint, and the code is discarded.
    // R_BT, W_SC, R_SC, W_BT: ^ Same
    // R_BT, W_SC, W_BT, W_SC: ^ Same
    // W_SC, R_BT, R_SC, W_BT: ^ Same
    // W_SC, R_BT, W_BT, R_SC: ^ Same
    // W_SC, W_BT, R_BT, R_SC: No watchpoint is installed, but we could not see old structures from the cache.
    clearStructureCache(vm);

    // Make sure that all JSArray allocations that load the appropriate structure from
    // this object now load a structure that uses SlowPut.
    for (unsigned i = 0; i < NumberOfIndexingShapes; ++i)
		m_arrayStructureForIndexingShapeDuringAllocation[i].set(vm, this, originalArrayStructureForIndexingType(ArrayWithSlowPutArrayStorage));

    // Same for any special array structures.
    Structure* slowPutStructure;
    slowPutStructure = createRegExpMatchesArraySlowPutStructure(vm, this);
    m_regExpMatchesArrayStructure.set(vm, this, slowPutStructure);
    slowPutStructure = createRegExpMatchesArrayWithIndicesSlowPutStructure(vm, this);
    m_regExpMatchesArrayWithIndicesStructure.set(vm, this, slowPutStructure);
    slowPutStructure = createRegExpMatchesIndicesArraySlowPutStructure(vm, this);
    m_regExpMatchesIndicesArrayStructure.set(vm, this, slowPutStructure);
    slowPutStructure = ClonedArguments::createSlowPutStructure(vm, this, m_objectPrototype.get());
    m_clonedArgumentsStructure.set(vm, this, slowPutStructure);

    // Make sure that all allocations or indexed storage transitions that are inlining
    // the assumption that it's safe to transition to a non-SlowPut array storage don't
    // do so anymore.
    // Note: we are deliberately firing the watchpoint here at the end only after
    // making all the array structures SlowPut. This ensures that the concurrent
    // JIT threads will always get the SlowPut versions of the structures if
    // isHavingABadTime() returns true. The concurrent JIT relies on this.
    m_havingABadTimeWatchpoint->fireAll(vm, "Having a bad time");
    ASSERT(isHavingABadTime()); // The watchpoint is what tells us that we're having a bad time.
};

void JSGlobalObject::clearStructureCache(VM& vm)
{
    m_structureCache.clear(); // We may be caching array structures in here.
    m_structureCacheClearedWatchpoint.fireAll(vm/*, "Clearing StructureCache"*/);
}

void JSGlobalObject::haveABadTime(VM& vm)
{
    ASSERT(&vm == &this->vm());

    if (isHavingABadTime())
        return;

    DeferGC deferGC(vm.heap);

    // Consider the following objects and prototype chains:
    //    O (of global G1) -> A (of global G1)
    //    B (of global G2) where G2 has a bad time
    //
    // If we set B as the prototype of A, G1 will need to have a bad time.
    // See comments in Structure::mayInterceptIndexedAccesses() for why.
    //
    // Now, consider the following objects and prototype chains:
    //    O1 (of global G1) -> A1 (of global G1) -> B1 (of global G2)
    //    O2 (of global G2) -> A2 (of global G2)
    //    B2 (of global G3) where G3 has a bad time.
    //
    // G1 and G2 does not have a bad time, but G3 already has a bad time.
    // If we set B2 as the prototype of A2, then G2 needs to have a bad time.
    // Note that by induction, G1 also now needs to have a bad time because of
    // O1 -> A1 -> B1.
    //
    // We describe this as global G1 being affected by global G2, and G2 by G3.
    // Similarly, we say that G1 is dependent on G2, and G2 on G3.
    // Hence, when G3 has a bad time, we need to ensure that all globals that
    // are transitively dependent on it also have a bad time (G2 and G1 in this
    // example).
    //
    // Apart from clearing the VM structure cache above, there are 2 more things
    // that we have to do when globals have a bad time:
    // 1. For each affected global:
    //    a. Fire its HaveABadTime watchpoint.
    //    b. Convert all of its array structures to SlowPutArrayStorage.
    // 2. Make sure that all affected objects  switch to the slow kind of
    //    indexed storage. An object is considered to be affected if it has
    //    indexed storage and has a prototype object which may have indexed
    //    accessors. If the prototype object belongs to a global having a bad
    //    time, then the prototype object is considered to possibly have indexed
    //    accessors. See comments in Structure::mayInterceptIndexedAccesses()
    //    for details.
    //
    // Note: step 1 must be completed before step 2 because step 2 relies on
    // the HaveABadTime watchpoint having already been fired on all affected
    // globals.
    //
    // In the common case, only this global will start having a bad time here,
    // and no other globals are affected by it. So, we first proceed on this assumption
    // with a simpler ObjectsWithBrokenIndexingFinder scan to find heap objects
    // affected by this global that need to be converted to SlowPutArrayStorage.
    // We'll also have the finder check for the presence of other global objects
    // depending on this one.
    //
    // If we do discover other globals depending on this one, we'll abort this
    // first ObjectsWithBrokenIndexingFinder scan because it will be insufficient
    // to find all affected objects that need to be converted to SlowPutArrayStorage.
    // It also does not make dependent globals have a bad time. Instead, we'll
    // take a more comprehensive approach of first creating a dependency graph
    // between globals, and then using that graph to determine all affected
    // globals and objects. With that, we can make all affected globals have a
    // bad time, and convert all affected objects to SlowPutArrayStorage.

    fireWatchpointAndMakeAllArrayStructuresSlowPut(vm); // Step 1 above.

    Vector<JSObject*> foundObjects;
    ObjectsWithBrokenIndexingFinder<BadTimeFinderMode::SingleGlobal> finder(vm, foundObjects, this);
    {
        HeapIterationScope iterationScope(vm.heap);
        vm.heap.objectSpace().forEachLiveCell(iterationScope, finder); // Attempt step 2 above.
    }

    if (finder.needsMultiGlobalsScan()) {
        foundObjects.clear();

        // Find all globals that will also have a bad time as a side effect of
        // this global having a bad time.
        GlobalObjectDependencyFinder dependencies(vm);
        {
            HeapIterationScope iterationScope(vm.heap);
            vm.heap.objectSpace().forEachLiveCell(iterationScope, dependencies);
        }

        HashSet<JSGlobalObject*> globalsHavingABadTime;
        Deque<JSGlobalObject*> globals;

        globals.append(this);
        while (!globals.isEmpty()) {
            JSGlobalObject* global = globals.takeFirst();
            global->fireWatchpointAndMakeAllArrayStructuresSlowPut(vm); // Step 1 above.
            auto result = globalsHavingABadTime.add(global);
            if (result.isNewEntry) {
                if (HashSet<JSGlobalObject*>* dependents = dependencies.dependentsFor(global)) {
                	for(auto iter = (*dependents).begin(); iter != (*dependents).end(); ++iter)
                        globals.append(*iter);
                }
            }
        }

        ObjectsWithBrokenIndexingFinder<BadTimeFinderMode::MultipleGlobals> finder(vm, foundObjects, globalsHavingABadTime);
        {
            HeapIterationScope iterationScope(vm.heap);
            vm.heap.objectSpace().forEachLiveCell(iterationScope, finder); // Step 2 above.
        }
    }

    while (!foundObjects.isEmpty()) {
        JSObject* object = asObject(foundObjects.last());
        foundObjects.removeLast();
        ASSERT(hasBrokenIndexing(object));
        object->switchToSlowPutArrayStorage(vm);
    }
}

bool JSGlobalObject::objectPrototypeIsSane() const
{
    return !hasIndexedProperties(m_objectPrototype->structure()->indexingType())
        && m_objectPrototype->getPrototypeDirect().isNull();
}

bool JSGlobalObject::arrayPrototypeChainIsSane() const
{
    return !hasIndexedProperties(m_arrayPrototype->structure()->indexingType())
        && m_arrayPrototype->getPrototypeDirect() == m_objectPrototype.get()
        && objectPrototypeIsSane();
}

bool JSGlobalObject::stringPrototypeChainIsSane() const
{
	return !hasIndexedProperties(m_stringPrototype->structure()->indexingType())
		&& m_stringPrototype->getPrototypeDirect() == m_objectPrototype.get()
		&& objectPrototypeIsSane();
}

// Set prototype, and also insert the object prototype at the end of the chain.
void JSGlobalObject::resetPrototype(JSGlobalData& globalData, JSValue prototype)
{
	setPrototypeDirect(globalData, prototype);

    JSObject* oldLastInPrototypeChain = lastInPrototypeChain(this);
    JSObject* objectPrototype = m_objectPrototype.get();
    if (oldLastInPrototypeChain != objectPrototype)
        oldLastInPrototypeChain->setPrototypeDirect(globalData, objectPrototype);
}

void JSGlobalObject::visitChildren(JSCell* cell, SlotVisitor& visitor)
{ 
    JSGlobalObject* thisObject = jsCast<JSGlobalObject*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);

    visitor.append(thisObject->m_globalThis);

    visitor.append(thisObject->m_globalLexicalEnvironment);
    visitor.append(thisObject->m_globalCallee);
    visitor.append(thisObject->m_regExpConstructor);
    visitor.append(thisObject->m_dateConstructor);
    visitor.append(thisObject->m_errorConstructor);
    visitor.append(thisObject->m_evalErrorStructure);
    visitor.append(thisObject->m_rangeErrorStructure);
    visitor.append(thisObject->m_referenceErrorStructure);
    visitor.append(thisObject->m_syntaxErrorStructure);
    visitor.append(thisObject->m_typeErrorStructure);
    visitor.append(thisObject->m_URIErrorStructure);
    visitor.append(thisObject->m_aggregateErrorStructure);
    visitor.append(thisObject->m_objectConstructor);
    visitor.append(thisObject->m_arrayConstructor);
    visitor.append(thisObject->m_functionConstructor);
    visitor.append(thisObject->m_promiseConstructor);
    visitor.append(thisObject->m_internalPromiseConstructor);
    visitor.append(thisObject->m_arrayBufferConstructor);
    visitor.append(thisObject->m_proxyObjectStructure);
    visitor.append(thisObject->m_callableProxyObjectStructure);
    visitor.append(thisObject->m_proxyRevokeStructure);
    visitor.append(thisObject->m_sharedArrayBufferStructure);

    visitor.append(thisObject->m_parseIntFunction);
    visitor.append(thisObject->m_parseFloatFunction);

#if ENABLE(INTL)
    visitor.append(thisObject->m_intlObject);
    visitor.append(thisObject->m_defaultCollator);
#endif

    visitor.append(thisObject->m_nullGetterFunction);
    visitor.append(thisObject->m_nullSetterFunction);
    visitor.append(thisObject->m_nullSetterStrictFunction);

    visitor.append(thisObject->m_evalFunction);
    visitor.append(thisObject->m_callFunction);
    visitor.append(thisObject->m_applyFunction);
    visitor.append(thisObject->m_objectProtoToStringFunction);
    visitor.append(thisObject->m_arrayProtoToStringFunction);
    visitor.append(thisObject->m_arrayProtoValuesFunction);
    visitor.append(thisObject->m_initializePromiseFunction);
    visitor.append(thisObject->m_iteratorProtocolFunction);
    visitor.append(thisObject->m_promiseResolveFunction);
    visitor.append(thisObject->m_numberProtoToStringFunction);
    visitor.append(thisObject->m_objectProtoValueOfFunction);
    visitor.append(thisObject->m_newPromiseCapabilityFunction);
    visitor.append(thisObject->m_functionProtoHasInstanceSymbolFunction);
    visitor.append(thisObject->m_setPrototypeDirectFunction);
    visitor.append(thisObject->m_regExpProtoExec);
    visitor.append(thisObject->m_regExpProtoSymbolReplace);
    visitor.append(thisObject->m_regExpProtoGlobalGetter);
    visitor.append(thisObject->m_regExpProtoUnicodeGetter);
    visitor.append(thisObject->m_throwTypeErrorArgumentsCalleeGetterSetter);

    visitor.append(thisObject->m_moduleLoader);

    visitor.append(thisObject->m_objectPrototype);
    visitor.append(thisObject->m_functionPrototype);
    visitor.append(thisObject->m_arrayPrototype);
    visitor.append(thisObject->m_regExpPrototype);
    visitor.append(thisObject->m_iteratorPrototype);
    visitor.append(thisObject->m_asyncIteratorPrototype);
    visitor.append(thisObject->m_generatorFunctionPrototype);
    visitor.append(thisObject->m_generatorPrototype);
    visitor.append(thisObject->m_arrayIteratorPrototype);
    visitor.append(thisObject->m_asyncFunctionPrototype);
    visitor.append(thisObject->m_asyncGeneratorPrototype);
    visitor.append(thisObject->m_asyncGeneratorFunctionPrototype);

    visitor.append(thisObject->m_withScopeStructure);
    visitor.append(thisObject->m_strictEvalActivationStructure);
    visitor.append(thisObject->m_lexicalEnvironmentStructure);
    visitor.append(thisObject->m_moduleEnvironmentStructure);
    visitor.append(thisObject->m_directArgumentsStructure);
    visitor.append(thisObject->m_scopedArgumentsStructure);
    visitor.append(thisObject->m_clonedArgumentsStructure);
    visitor.append(thisObject->m_objectStructureForObjectConstructor);

    for (unsigned i = 0; i < NumberOfIndexingShapes; ++i)
        visitor.append(thisObject->m_originalArrayStructureForIndexingShape[i]);
    for (unsigned i = 0; i < NumberOfIndexingShapes; ++i)
        visitor.append(thisObject->m_arrayStructureForIndexingShapeDuringAllocation[i]);

#if !PLATFORM(ESPIAL)
    visitor.append(thisObject->m_callbackConstructorStructure);
    visitor.append(thisObject->m_callbackFunctionStructure);
    visitor.append(thisObject->m_callbackObjectStructure);
#endif
    visitor.append(thisObject->m_nullPrototypeObjectStructure);
    visitor.append(thisObject->m_calleeStructure);

    visitor.append(thisObject->m_hostFunctionStructure);
    visitor.append(thisObject->m_builtinFunctions.arrowFunctionStructure);
    visitor.append(thisObject->m_builtinFunctions.sloppyFunctionStructure);
    visitor.append(thisObject->m_builtinFunctions.strictFunctionStructure);
    visitor.append(thisObject->m_ordinaryFunctions.arrowFunctionStructure);
    visitor.append(thisObject->m_ordinaryFunctions.sloppyFunctionStructure);
    visitor.append(thisObject->m_ordinaryFunctions.strictFunctionStructure);

    visitor.append(thisObject->m_boundFunctionStructure);
    visitor.append(thisObject->m_nativeStdFunctionStructure);
    visitor.append(thisObject->m_customGetterFunctionStructure);
    visitor.append(thisObject->m_customSetterFunctionStructure);
    visitor.append(thisObject->m_bigIntObjectStructure);
    visitor.append(thisObject->m_regExpStructure);
    visitor.append(thisObject->m_asyncFunctionStructure);
    visitor.append(thisObject->m_asyncGeneratorFunctionStructure);
    visitor.append(thisObject->m_generatorFunctionStructure);
    visitor.append(thisObject->m_internalFunctionStructure);
    visitor.append(thisObject->m_arrayIteratorStructure);
    visitor.append(thisObject->m_moduleRecordStructure);
    visitor.append(thisObject->m_moduleNamespaceObjectStructure);
    visitor.append(thisObject->m_iteratorResultObjectStructure);
    visitor.append(thisObject->m_regExpMatchesArrayStructure);
    visitor.append(thisObject->m_regExpMatchesArrayWithIndicesStructure);
    visitor.append(thisObject->m_regExpMatchesIndicesArrayStructure);

    for (int i = 0; i < LinkTimeConstantCount; ++i)
    	visitor.append(thisObject->m_linkTimeConstants[i]);

#define VISIT_SIMPLE_TYPE(CapitalName, lowerName, properName, instanceType, jsName) \
    visitor.append(thisObject->m_ ## lowerName ## Prototype); \
    visitor.append(thisObject->m_ ## properName ## Structure); \

    FOR_EACH_SIMPLE_BUILTIN_TYPE(VISIT_SIMPLE_TYPE)
    FOR_EACH_BUILTIN_DERIVED_ITERATOR_TYPE(VISIT_SIMPLE_TYPE)

    if (Options::useBigInt())
        FOR_BIG_INT_BUILTIN_TYPE_WITH_CONSTRUCTOR(VISIT_SIMPLE_TYPE)

#undef VISIT_SIMPLE_TYPE

    visitor.append(thisObject->m_speciesGetterSetter);

    for (unsigned i = NumberOfTypedArrayTypes; i--;) {
        visitor.append(thisObject->m_typedArrays[i].prototype);
        visitor.append(thisObject->m_typedArrays[i].constructor);
        visitor.append(thisObject->m_typedArrays[i].structure);
    }
}

void JSGlobalObject::addStaticGlobals(GlobalPropertyInfo* globals, int count)
{
    ScopeOffset startOffset = addVariables(count, jsUndefined());

    for (int i = 0; i < count; ++i) {
        GlobalPropertyInfo& global = globals[i];
        // This `configurable = false` is necessary condition for static globals,
        // otherwise lexical bindings can change the result of GlobalVar queries too.
        // We won't be able to declare a global lexical variable with the sanem name to
        // the static globals because configurable = false.
        ASSERT(global.attributes & DontDelete);

        WatchpointSet* watchpointSet = nullptr;
        WriteBarrierBase<Unknown>* variable = nullptr;
        {
            //ConcurrentJITLocker locker(symbolTable()->m_lock);
        	ScopeOffset offset = symbolTable()->takeNextScopeOffset();
            RELEASE_ASSERT(offset = startOffset + i);
            SymbolTableEntry newEntry(VarOffset(offset), global.attributes);
            if(!global.identifier.isPrivateName()) {
            	newEntry.prepareToWatch();
            	watchpointSet = newEntry.watchpointSet();
            }
            symbolTable()->add(global.identifier.impl(), WTFMove(newEntry));
            variable = &variableAt(offset);
        }
        symbolTablePutTouchWatchpointSet(vm(), this, global.identifier, global.value, variable, watchpointSet);
    }
}

bool JSGlobalObject::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    if (Base::getOwnPropertySlot(object, exec, propertyName, slot))
        return true;
    return symbolTableGet(jsCast<JSGlobalObject*>(object), propertyName, slot);
}

void JSGlobalObject::clearRareData(JSCell* cell)
{
    jsCast<JSGlobalObject*>(cell)->m_rareData.clear();
}

void JSGlobalObject::tryInstallSpeciesWatchpoint(ExecState* exec, JSObject* prototype, JSObject* constructor, OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>>& constructorWatchpoint, OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>>& speciesWatchpoint, InlineWatchpointSet& speciesWatchpointSet)
{
    RELEASE_ASSERT(!constructorWatchpoint);
    RELEASE_ASSERT(!speciesWatchpoint);

    VM& vm = this->vm();
    ThrowScope scope(vm);

    // First we need to make sure that the %prototype%.constructor property points to a %constructor%
    // and that %constructor%[Symbol.species] is the primordial GetterSetter.

    // We only initialize once so flattening the structures does not have any real cost.
    Structure* prototypeStructure = prototype->structure(vm);
    if (prototypeStructure->isDictionary())
        prototypeStructure = prototypeStructure->flattenDictionaryStructure(vm, prototype);
    RELEASE_ASSERT(!prototypeStructure->isDictionary());

    auto invalidateWatchpoint = [&] {
    	speciesWatchpointSet.invalidate(vm, StringFireDetail("Was not able to set up species watchpoint."));
    };

    PropertySlot constructorSlot(prototype, PropertySlot::InternalMethodType::VMInquiry);
    prototype->getOwnPropertySlot(prototype, exec, vm.propertyNames->constructor, constructorSlot);
    scope.assertNoException();
    if (constructorSlot.slotBase() != prototype
        || !constructorSlot.isCacheableValue()
        || constructorSlot.getValue(exec, vm.propertyNames->constructor) != constructor) {
        invalidateWatchpoint();
        return;
    }

    Structure* constructorStructure = constructor->structure(vm);
    if (constructorStructure->isDictionary())
        constructorStructure = constructorStructure->flattenDictionaryStructure(vm, constructor);

    PropertySlot speciesSlot(constructor, PropertySlot::InternalMethodType::VMInquiry);
    constructor->getOwnPropertySlot(constructor, exec, vm.propertyNames->speciesSymbol, speciesSlot);
    scope.assertNoException();
    if (speciesSlot.slotBase() != constructor
        || !speciesSlot.isCacheableGetter()
        || speciesSlot.getterSetter() != speciesGetterSetter()) {
        invalidateWatchpoint();
        return;
    }

    // Now we need to setup the watchpoints to make sure these conditions remain valid.
    prototypeStructure->startWatchingPropertyForReplacements(vm, constructorSlot.cachedOffset());
    constructorStructure->startWatchingPropertyForReplacements(vm, speciesSlot.cachedOffset());

    ObjectPropertyCondition constructorCondition = ObjectPropertyCondition::equivalence(vm, prototype, prototype, vm.propertyNames->constructor.impl(), constructor);
    ObjectPropertyCondition speciesCondition = ObjectPropertyCondition::equivalence(vm, prototype, constructor, vm.propertyNames->speciesSymbol.impl(), speciesGetterSetter());

    if (!constructorCondition.isWatchable(vm) || !speciesCondition.isWatchable(vm)) {
        invalidateWatchpoint();
        return;
    }

    // We only watch this from the DFG, and the DFG makes sure to only start watching if the watchpoint is in the IsWatched state.
    RELEASE_ASSERT(!speciesWatchpointSet.isBeingWatched());
    speciesWatchpointSet.touch(vm/*, "Set up array species watchpoint."*/);

    constructorWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, constructorCondition, speciesWatchpointSet));
    constructorWatchpoint->install(vm);

    speciesWatchpoint = adoptPtr(new ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>(this, speciesCondition, speciesWatchpointSet));
    speciesWatchpoint->install(vm);
}

void JSGlobalObject::tryInstallArraySpeciesWatchpoint(ExecState* exec)
{
    RELEASE_ASSERT(!m_arrayPrototypeConstructorWatchpoint);
    RELEASE_ASSERT(!m_arrayConstructorSpeciesWatchpoint);

    tryInstallSpeciesWatchpoint(exec, arrayPrototype(), arrayConstructor(), m_arrayPrototypeConstructorWatchpoint, m_arrayConstructorSpeciesWatchpoint, m_arraySpeciesWatchpointSet);
}

void JSGlobalObject::tryInstallArrayBufferSpeciesWatchpoint(ExecState* exec, ArrayBufferSharingMode sharingMode)
{
    //static_assert(static_cast<unsigned>(ArrayBufferSharingMode::Default) == 0);
    //static_assert(static_cast<unsigned>(ArrayBufferSharingMode::Shared) == 1);
    unsigned index = static_cast<unsigned>(sharingMode);
    tryInstallSpeciesWatchpoint(exec, arrayBufferPrototype(sharingMode), arrayBufferConstructor(sharingMode), m_arrayBufferPrototypeConstructorWatchpoints[index], m_arrayBufferConstructorSpeciesWatchpoints[index], arrayBufferSpeciesWatchpointSet(sharingMode));
}

void slowValidateCell(JSGlobalObject* globalObject)
{
    if (!globalObject->isGlobalObject())
        RELEASE_ASSERT_NOT_REACHED();
    ASSERT_GC_OBJECT_INHERITS(globalObject, JSGlobalObject::info());
}

# if ENABLE(INTL)
IntlCollator* JSGlobalObject::defaultCollator(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    if (m_defaultCollator)
        return m_defaultCollator.get();

    IntlCollator* collator = IntlCollator::create(vm, intlObject()->collatorStructure());
    collator->initializeCollator(*exec, jsUndefined(), jsUndefined());
    RETURN_IF_EXCEPTION(vm, nullptr);
    m_defaultCollator.set(vm, this, collator);
    return collator;
}

Structure* JSGlobalObject::collatorStructure() const { return intlObject()->collatorStructure(); }
Structure* JSGlobalObject::dateTimeFormatStructure() const { return intlObject()->dateTimeFormatStructure(); }
Structure* JSGlobalObject::numberFormatStructure() const { return intlObject()->numberFormatStructure(); }
Structure* JSGlobalObject::localeStructure() const { return intlObject()->localeStructure(); }
Structure* JSGlobalObject::pluralRulesStructure() const { return intlObject()->pluralRulesStructure(); }
Structure* JSGlobalObject::relativeTimeFormatStructure() const { return intlObject()->relativeTimeFormatStructure(); }

#endif // ENABLE(INTL)

void JSGlobalObject::bumpGlobalLexicalBindingEpoch(VM& vm)
{
    if (++m_globalLexicalBindingEpoch == Options::thresholdForGlobalLexicalBindingEpoch()) {
        // Since the epoch overflows, we should rewrite all the CodeBlock to adjust to the newly started generation.
        m_globalLexicalBindingEpoch = 1;
		auto& codeBlockSet = vm.heap.codeBlockSet().m_set;
		for (auto iter = codeBlockSet.begin(); iter != codeBlockSet.end(); ++iter) {
			CodeBlock* codeBlock = *iter;
			if (codeBlock->globalObject() != this)
				continue;
			codeBlock->notifyLexicalBindingUpdate();
		}
	}
}

void JSGlobalObject::queueMicrotask(PassRefPtr<Microtask> task)
{
    if (globalObjectMethodTable()->queueTaskToEventLoop)
        globalObjectMethodTable()->queueTaskToEventLoop(this, task);
    else
        WTFLogAlways("ERROR: Event loop not supported.");
}

#if ENABLE(DFG_JIT)
WatchpointSet* JSGlobalObject::getReferencedPropertyWatchpointSet(UniquedStringImpl* uid)
{
    return m_referencedGlobalPropertyWatchpointSets.get(uid);
}

WatchpointSet& JSGlobalObject::ensureReferencedPropertyWatchpointSet(UniquedStringImpl* uid)
{
    return *m_referencedGlobalPropertyWatchpointSets.ensure(uid, [] {
        return WatchpointSet::create(IsWatched);
    }).iterator->second.get();
}
#endif

void JSGlobalObject::setConsoleClient(ConsoleClient* consoleClient)
{
    m_consoleClient = makeWeakPtr(consoleClient);
}

Structure* JSGlobalObject::arrayStructureForIndexingTypeDuringAllocation(ExecState* exec, IndexingType indexingType, JSValue newTarget) const
{
    JSGlobalObject* globalObject = exec->lexicalGlobalObject();

    VM& vm = globalObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    if (!newTarget || newTarget == globalObject->arrayConstructor())
        return globalObject->arrayStructureForIndexingTypeDuringAllocation(indexingType);
    auto* functionGlobalObject = getFunctionRealm(exec, globalObject, asObject(newTarget));
    RETURN_IF_EXCEPTION(scope, nullptr);
    RELEASE_AND_RETURN(scope, InternalFunction::createSubclassStructure(exec, globalObject, asObject(newTarget), functionGlobalObject->arrayStructureForIndexingTypeDuringAllocation(indexingType)));
}

} // namespace JSC
