/*
 *  Copyright (C) 2007 Eric Seidel <eric@webkit.org>
 *  Copyright (C) 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "ArrayAllocationProfile.h"
#include "ArrayBufferSharingMode.h"
#include "BigIntPrototype.h"
#include "BooleanPrototype.h"
#include "ConsoleClient.h"
#include "ErrorType.h"
#include "InternalFunction.h"
#include "JSArray.h"
#include "JSArrayBufferConstructor.h"
#if !PLATFORM(ESPIAL)
#include "JSClassRef.h"
#endif
#include "JSGlobalLexicalEnvironment.h"
#include "JSProxy.h"
#include "JSSegmentedVariableObject.h"
#if !PLATFORM(ESPIAL)
#include "JSWeakObjectMapRefInternal.h"
#endif
#include "NumberPrototype.h"
#include "ObjectPropertyChangeAdaptiveWatchpoint.h"
#include "SpecialPointer.h"
#include "StringPrototype.h"
#include "StructureCache.h"
#include "StructureChain.h"
#include "StructureRareDataInlines.h"
#include "SymbolPrototype.h"
#include "VM.h"
#include "VariableEnvironment.h"
#include "Watchpoint.h"
#include <wtf/HashSet.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RandomNumber.h>
#include <wtf/WeakPtr.h>

struct OpaqueJSClass;
struct OpaqueJSClassContextData;

namespace JSC {

class AggregateErrorConstructor;
class AggregateErrorPrototype;
class ArrayConstructor;
class ArrayPrototype;
class ArrayIteratorPrototype;
class AsyncIteratorPrototype;
class AsyncFunctionPrototype;
class AsyncGeneratorPrototype;
class AsyncGeneratorFunctionPrototype;
class BooleanPrototype;
class BuiltinExecutables;
class ConsoleClient;
class Debugger;
class ErrorConstructor;
class ErrorPrototype;
class FunctionConstructor;
class FunctionPrototype;
class GeneratorPrototype;
class GeneratorFunctionPrototype;
class GetterSetter;
class GlobalCodeBlock;
class IntlObject;
class IntlCollator;
class JSArrayBufferConstructor;
class JSCustomGetterFunction;
class JSCustomSetterFunction;
class JSPromise;
class JSPromiseConstructor;
class JSPromisePrototype;
class JSInternalPromise;
class JSModuleLoader;
class JSModuleRecord;
class LLIntOffsetsExtractor;
class MapPrototype;
class Microtask;
class ModuleLoader;
class NativeErrorConstructorBase;
class ObjectConstructor;
class ProgramCodeBlock;
class RegExpConstructor;
class RegExpPrototype;
class JSStack;
class NullGetterFunction;
class NullSetterFunction;
class SetPrototype;

struct ActivationStackNode;
struct HashTable;

enum class BindingCreationContext : bool { Global, Eval };

#define DEFINE_STANDARD_BUILTIN(macro, upperName, lowerName) macro(upperName, lowerName, lowerName, JS ## upperName, upperName)

#define FOR_EACH_SIMPLE_BUILTIN_TYPE_WITH_CONSTRUCTOR(macro) \
    macro(Set, set, set, JSSet, Set) \
    macro(Map, map, map, JSMap, Map) \
    macro(Date, date, date, DateInstance, Date) \
    macro(String, string, stringObject, StringObject, String) \
    macro(Symbol, symbol, symbolObject, SymbolObject, Symbol) \
    macro(Boolean, boolean, booleanObject, BooleanObject, Boolean) \
    macro(Number, number, numberObject, NumberObject, Number) \
    macro(Error, error, error, ErrorInstance, Error) \
    macro(JSPromise, promise, promise, JSPromise, Promise) \
    macro(JSArrayBuffer, arrayBuffer, arrayBuffer, JSArrayBuffer, ArrayBuffer) \
    DEFINE_STANDARD_BUILTIN(macro, WeakMap, weakMap) \
    DEFINE_STANDARD_BUILTIN(macro, WeakSet, weakSet) \

#define FOR_BIG_INT_BUILTIN_TYPE_WITH_CONSTRUCTOR(macro) \
    macro(BigInt, bigInt, bigIntObject, BigIntObject, BigInt)

#define FOR_EACH_BUILTIN_DERIVED_ITERATOR_TYPE(macro) \
    macro(StringIterator, stringIterator, stringIterator, JSStringIterator, StringIterator) \

#define FOR_EACH_SIMPLE_BUILTIN_TYPE(macro) \
    FOR_EACH_SIMPLE_BUILTIN_TYPE_WITH_CONSTRUCTOR(macro) \
    macro(JSInternalPromise, internalPromise, internalPromise, JSInternalPromise, InternalPromise) \

#define DECLARE_SIMPLE_BUILTIN_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    class JS ## capitalName; \
    class capitalName ## Prototype; \
    class capitalName ## Constructor;

class IteratorPrototype;
FOR_EACH_SIMPLE_BUILTIN_TYPE(DECLARE_SIMPLE_BUILTIN_TYPE)
FOR_BIG_INT_BUILTIN_TYPE_WITH_CONSTRUCTOR(DECLARE_SIMPLE_BUILTIN_TYPE)
FOR_EACH_BUILTIN_DERIVED_ITERATOR_TYPE(DECLARE_SIMPLE_BUILTIN_TYPE)

#undef DECLARE_SIMPLE_BUILTIN_TYPE

enum class JSPromiseRejectionOperation : unsigned {
    Reject, // When a promise is rejected without any handlers.
    Handle, // When a handler is added to a rejected promise for the first time.
};

struct GlobalObjectMethodTable {
	typedef bool (*SupportsProfilingFunctionPtr)(const JSGlobalObject*);
	SupportsProfilingFunctionPtr supportsProfiling;

	typedef bool (*SupportsRichSourceInfoFunctionPtr)(const JSGlobalObject*);
	SupportsRichSourceInfoFunctionPtr supportsRichSourceInfo;

	typedef bool (*ShouldInterruptScriptFunctionPtr)(const JSGlobalObject*);
	ShouldInterruptScriptFunctionPtr shouldInterruptScript;

	typedef bool (*JavaScriptExperimentsEnabledFunctionPtr)(const JSGlobalObject*);
	JavaScriptExperimentsEnabledFunctionPtr javaScriptExperimentsEnabled;

    typedef void (*QueueTaskToEventLoopFunctionPtr)(const JSGlobalObject*, PassRefPtr<Microtask>);
    QueueTaskToEventLoopFunctionPtr queueTaskToEventLoop;

    typedef JSInternalPromise* (*ModuleLoaderImportModulePtr)(JSGlobalObject*, ExecState*, JSModuleLoader*, JSString*,  JSValue, const SourceOrigin&);
    ModuleLoaderImportModulePtr moduleLoaderImportModule;

    typedef Identifier (*ModuleLoaderResolvePtr)(JSGlobalObject*, ExecState*, JSModuleLoader*, JSValue, JSValue, JSValue);
    ModuleLoaderResolvePtr moduleLoaderResolve;

    typedef JSInternalPromise* (*ModuleLoaderFetchPtr)(JSGlobalObject*, ExecState*, JSModuleLoader*, JSValue,  JSValue, JSValue);
    ModuleLoaderFetchPtr moduleLoaderFetch;

    typedef JSObject* (*ModuleLoaderCreateImportMetaPropertiesPtr)(JSGlobalObject*, ExecState*, JSModuleLoader*, JSValue, JSModuleRecord*, JSValue);
    ModuleLoaderCreateImportMetaPropertiesPtr moduleLoaderCreateImportMetaProperties;

    typedef JSValue (*ModuleLoaderEvaluatePtr)(JSGlobalObject*, ExecState*, JSModuleLoader*, JSValue, JSValue, JSValue);
    ModuleLoaderEvaluatePtr moduleLoaderEvaluate;

    typedef void (*PromiseRejectionTrackerPtr)(JSGlobalObject*, ExecState*, JSPromise*, JSPromiseRejectionOperation);
    PromiseRejectionTrackerPtr promiseRejectionTracker;

    typedef String (*DefaultLanguageFunctionPtr)();
    DefaultLanguageFunctionPtr defaultLanguage;
};

class JSGlobalObject : public JSSegmentedVariableObject {
private:
#if !PLATFORM(ESPIAL)
	typedef HashSet<RefPtr<OpaqueJSWeakObjectMap> > WeakMapSet;
	typedef HashMap<OpaqueJSClass*, OwnPtr<OpaqueJSClassContextData> > OpaqueJSClassDataMap;
#endif

	struct JSGlobalObjectRareData {
		WTF_MAKE_FAST_ALLOCATED;
	public:
		JSGlobalObjectRareData()
			: profileGroup(0)
		{
		}

#if !PLATFORM(ESPIAL)
		WeakMapSet weakMaps;
#endif
		unsigned profileGroup;

#if !PLATFORM(ESPIAL)
        OpaqueJSClassDataMap opaqueJSClassData;
#endif
	};

// Our hashtable code-generator tries to access these properties, so we make them public.
// However, we'd like it better if they could be protected.
public:
    Register m_globalCallFrame[JSStack::CallFrameHeaderSize];

	WriteBarrier<JSObject> m_globalThis;

    WriteBarrier<JSGlobalLexicalEnvironment> m_globalLexicalEnvironment;
    WriteBarrier<JSObject> m_globalCallee;
	WriteBarrier<RegExpConstructor> m_regExpConstructor;

	WriteBarrier<DateConstructor> m_dateConstructor;
	WriteBarrier<ErrorConstructor> m_errorConstructor;
	WriteBarrier<Structure> m_evalErrorStructure;
	WriteBarrier<Structure> m_rangeErrorStructure;
	WriteBarrier<Structure> m_referenceErrorStructure;
	WriteBarrier<Structure> m_syntaxErrorStructure;
	WriteBarrier<Structure> m_typeErrorStructure;
	WriteBarrier<Structure> m_URIErrorStructure;
	WriteBarrier<Structure> m_aggregateErrorStructure;

    WriteBarrier<ObjectConstructor> m_objectConstructor;
    WriteBarrier<ArrayConstructor> m_arrayConstructor;
    WriteBarrier<FunctionConstructor> m_functionConstructor;
    WriteBarrier<JSPromiseConstructor> m_promiseConstructor;
    WriteBarrier<JSInternalPromiseConstructor> m_internalPromiseConstructor;
    WriteBarrier<JSArrayBufferConstructor> m_arrayBufferConstructor;
    WriteBarrier<Structure> m_proxyObjectStructure;
    WriteBarrier<Structure> m_callableProxyObjectStructure;
    WriteBarrier<Structure> m_proxyRevokeStructure;
    WriteBarrier<Structure> m_sharedArrayBufferStructure;

    WriteBarrier<JSFunction> m_parseIntFunction;
    WriteBarrier<JSFunction> m_parseFloatFunction;

#if ENABLE(INTL)
    WriteBarrier<IntlObject> m_intlObject;
    WriteBarrier<IntlCollator> m_defaultCollator;
#endif
    WriteBarrier<NullGetterFunction> m_nullGetterFunction;
    WriteBarrier<NullSetterFunction> m_nullSetterFunction;
    WriteBarrier<NullSetterFunction> m_nullSetterStrictFunction;

	WriteBarrier<JSFunction> m_evalFunction;
	WriteBarrier<JSFunction> m_callFunction;
	WriteBarrier<JSFunction> m_applyFunction;
    WriteBarrier<JSFunction> m_objectProtoToStringFunction;
    WriteBarrier<JSFunction> m_arrayProtoToStringFunction;
    WriteBarrier<JSFunction> m_arrayProtoValuesFunction;
    WriteBarrier<JSFunction> m_initializePromiseFunction;
    WriteBarrier<JSFunction> m_iteratorProtocolFunction;
    WriteBarrier<JSFunction> m_promiseResolveFunction;
    WriteBarrier<JSFunction> m_numberProtoToStringFunction;
    WriteBarrier<JSFunction> m_objectProtoValueOfFunction;
    WriteBarrier<JSFunction> m_newPromiseCapabilityFunction;
    WriteBarrier<JSFunction> m_functionProtoHasInstanceSymbolFunction;
    WriteBarrier<JSFunction> m_setPrototypeDirectFunction;
    WriteBarrier<JSObject> m_regExpProtoExec;
    WriteBarrier<JSObject> m_regExpProtoSymbolReplace;
    WriteBarrier<GetterSetter> m_regExpProtoGlobalGetter;
    WriteBarrier<GetterSetter> m_regExpProtoUnicodeGetter;
    WriteBarrier<GetterSetter> m_throwTypeErrorArgumentsCalleeGetterSetter;

    WriteBarrier<JSModuleLoader> m_moduleLoader;

	WriteBarrier<ObjectPrototype> m_objectPrototype;
	WriteBarrier<FunctionPrototype> m_functionPrototype;
	WriteBarrier<ArrayPrototype> m_arrayPrototype;
	WriteBarrier<RegExpPrototype> m_regExpPrototype;
    WriteBarrier<IteratorPrototype> m_iteratorPrototype;
    WriteBarrier<AsyncIteratorPrototype> m_asyncIteratorPrototype;
    WriteBarrier<GeneratorFunctionPrototype> m_generatorFunctionPrototype;
    WriteBarrier<GeneratorPrototype> m_generatorPrototype;
    WriteBarrier<AsyncFunctionPrototype> m_asyncFunctionPrototype;
    WriteBarrier<AsyncGeneratorPrototype> m_asyncGeneratorPrototype;
    WriteBarrier<AsyncGeneratorFunctionPrototype> m_asyncGeneratorFunctionPrototype;
    WriteBarrier<ArrayIteratorPrototype> m_arrayIteratorPrototype;

	WriteBarrier<Structure> m_withScopeStructure;
	WriteBarrier<Structure> m_strictEvalActivationStructure;
	WriteBarrier<Structure> m_lexicalEnvironmentStructure;
    WriteBarrier<Structure> m_moduleEnvironmentStructure;
    WriteBarrier<Structure> m_directArgumentsStructure;
    WriteBarrier<Structure> m_scopedArgumentsStructure;
    WriteBarrier<Structure> m_clonedArgumentsStructure;

    WriteBarrier<Structure> m_objectStructureForObjectConstructor;

	// Lists the actual structures used for having these particular indexing shapes.
	WriteBarrier<Structure> m_originalArrayStructureForIndexingShape[NumberOfIndexingShapes];
	// Lists the structures we should use during allocation for these particular indexing shapes.
	// These structures will differ from the originals list above when we are having a bad time.
	WriteBarrier<Structure> m_arrayStructureForIndexingShapeDuringAllocation[NumberOfIndexingShapes];

#if !PLATFORM(ESPIAL)
	WriteBarrier<Structure> m_callbackConstructorStructure;
	WriteBarrier<Structure> m_callbackFunctionStructure;
	WriteBarrier<Structure> m_callbackObjectStructure;
#endif
	WriteBarrier<Structure> m_nullPrototypeObjectStructure;
    WriteBarrier<Structure> m_calleeStructure;

    WriteBarrier<Structure> m_hostFunctionStructure;

    struct FunctionStructures {
        WriteBarrier<Structure> arrowFunctionStructure;
        WriteBarrier<Structure> sloppyFunctionStructure;
        WriteBarrier<Structure> strictFunctionStructure;
    };
    FunctionStructures m_builtinFunctions;
    FunctionStructures m_ordinaryFunctions;

	WriteBarrier<Structure> m_boundFunctionStructure;
    WriteBarrier<Structure> m_customGetterFunctionStructure;
    WriteBarrier<Structure> m_customSetterFunctionStructure;
    WriteBarrier<Structure> m_nativeStdFunctionStructure;
	PropertyOffset m_functionNameOffset;
	WriteBarrier<Structure> m_regExpStructure;
	WriteBarrier<Structure> m_asyncFunctionStructure;
    WriteBarrier<Structure> m_asyncGeneratorFunctionStructure;
    WriteBarrier<Structure> m_generatorFunctionStructure;
	WriteBarrier<Structure> m_internalFunctionStructure;
    WriteBarrier<Structure> m_arrayIteratorStructure;
    WriteBarrier<Structure> m_moduleRecordStructure;
    WriteBarrier<Structure> m_moduleNamespaceObjectStructure;
    WriteBarrier<Structure> m_iteratorResultObjectStructure;
    WriteBarrier<Structure> m_regExpMatchesArrayStructure;
    WriteBarrier<Structure> m_regExpMatchesArrayWithIndicesStructure;
    WriteBarrier<Structure> m_regExpMatchesIndicesArrayStructure;

#define DEFINE_STORAGE_FOR_SIMPLE_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    WriteBarrier<capitalName ## Prototype> m_ ## lowerName ## Prototype; \
    WriteBarrier<Structure> m_ ## properName ## Structure;

    FOR_EACH_SIMPLE_BUILTIN_TYPE(DEFINE_STORAGE_FOR_SIMPLE_TYPE)
    FOR_BIG_INT_BUILTIN_TYPE_WITH_CONSTRUCTOR(DEFINE_STORAGE_FOR_SIMPLE_TYPE)
    FOR_EACH_BUILTIN_DERIVED_ITERATOR_TYPE(DEFINE_STORAGE_FOR_SIMPLE_TYPE)

#undef DEFINE_STORAGE_FOR_SIMPLE_TYPE

    WriteBarrier<GetterSetter> m_speciesGetterSetter;

    struct TypedArrayData {
        WriteBarrier<JSObject> prototype;
        WriteBarrier<InternalFunction> constructor;
        WriteBarrier<Structure> structure;
    };

    FixedArray<TypedArrayData, NumberOfTypedArrayTypes> m_typedArrays;

    JSCell* m_specialPointers[Special::TableSize]; // Special pointers used by the LLInt and JIT.
    WriteBarrier<JSCell> m_linkTimeConstants[LinkTimeConstantCount];

    StructureCache m_structureCache;

	Debugger* m_debugger;

    VM& m_vm;

	RefPtr<WatchpointSet> m_masqueradesAsUndefinedWatchpoint;
	RefPtr<WatchpointSet> m_havingABadTimeWatchpoint;
    RefPtr<WatchpointSet> m_varInjectionWatchpoint;
    RefPtr<WatchpointSet> m_varReadOnlyWatchpoint;

	OwnPtr<JSGlobalObjectRareData> m_rareData;

    OwnPtr<BuiltinExecutables> m_builtinExecutables;

	WeakRandom m_weakRandom;

	InlineWatchpointSet& arrayIteratorProtocolWatchpointSet() { return m_arrayIteratorProtocolWatchpointSet; }
	const InlineWatchpointSet& arrayIteratorProtocolWatchpointSet() const { return m_arrayIteratorProtocolWatchpointSet; }
	InlineWatchpointSet& mapIteratorProtocolWatchpoint() { return m_mapIteratorProtocolWatchpoint; }
	const InlineWatchpointSet& mapIteratorProtocolWatchpoint() const { return m_mapIteratorProtocolWatchpoint; }
    InlineWatchpointSet& setIteratorProtocolWatchpoint() { return m_setIteratorProtocolWatchpoint; }
    const InlineWatchpointSet& setIteratorProtocolWatchpoint() const { return m_setIteratorProtocolWatchpoint; }
    InlineWatchpointSet& stringIteratorProtocolWatchpoint() { return m_stringIteratorProtocolWatchpoint; }
    const InlineWatchpointSet& stringIteratorProtocolWatchpoint() const { return m_stringIteratorProtocolWatchpoint; }
    InlineWatchpointSet& mapSetWatchpoint() { return m_mapSetWatchpoint; }
    const InlineWatchpointSet& mapSetWatchpoint() const { return m_mapSetWatchpoint; }
    InlineWatchpointSet& setAddWatchpoint() { return m_setAddWatchpoint; }
    const InlineWatchpointSet& setAddWatchpoint() const { return m_setAddWatchpoint; }
	InlineWatchpointSet& arraySpeciesWatchpointSet() { return m_arraySpeciesWatchpointSet; }
	const InlineWatchpointSet& arraySpeciesWatchpointSet() const { return m_arraySpeciesWatchpointSet; }
    InlineWatchpointSet& arrayJoinWatchpointSet() { return m_arrayJoinWatchpointSet; }
    const InlineWatchpointSet& arrayJoinWatchpointSet() const { return m_arrayJoinWatchpointSet; }
	InlineWatchpointSet& numberToStringWatchpointSet() { return m_numberToStringWatchpointSet; }
	const InlineWatchpointSet& numberToStringWatchpointSet() const
	{
		return m_numberToStringWatchpointSet;
	}
	InlineWatchpointSet& structureCacheClearedWatchpoint() { return m_structureCacheClearedWatchpoint; }
	InlineWatchpointSet& arrayBufferSpeciesWatchpointSet(ArrayBufferSharingMode sharingMode)
    {
        switch (sharingMode) {
        case ArrayBufferSharingMode::Default:
            return m_arrayBufferSpeciesWatchpointSet;
#if ENABLE(SHARED_ARRAY_BUFFER)
        case ArrayBufferSharingMode::Shared:
            return m_sharedArrayBufferSpeciesWatchpointSet;
#endif
        }
        RELEASE_ASSERT_NOT_REACHED();
        return m_arrayBufferSpeciesWatchpointSet;
    }
	// If this hasn't been invalidated, it means the array iterator protocol
	// is not observable to user code yet.
	InlineWatchpointSet m_arrayIteratorProtocolWatchpointSet;
    InlineWatchpointSet m_mapIteratorProtocolWatchpoint;
    InlineWatchpointSet m_setIteratorProtocolWatchpoint;
    InlineWatchpointSet m_stringIteratorProtocolWatchpoint;
    InlineWatchpointSet m_mapSetWatchpoint;
    InlineWatchpointSet m_setAddWatchpoint;
    InlineWatchpointSet m_arraySpeciesWatchpointSet;
    InlineWatchpointSet m_arrayJoinWatchpointSet;
    InlineWatchpointSet m_numberToStringWatchpointSet;
    InlineWatchpointSet m_structureCacheClearedWatchpoint;
    InlineWatchpointSet m_arrayBufferSpeciesWatchpointSet;
    InlineWatchpointSet m_sharedArrayBufferSpeciesWatchpointSet;

    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_arrayConstructorSpeciesWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_arrayPrototypeConstructorWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_arrayPrototypeSymbolIteratorWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_arrayPrototypeJoinWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_arrayIteratorPrototypeNext;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_mapPrototypeSymbolIteratorWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_mapIteratorPrototypeNextWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_setPrototypeSymbolIteratorWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_setIteratorPrototypeNextWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_stringPrototypeSymbolIteratorWatchpoint;
	OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_stringIteratorPrototypeNextWatchpoint;
	OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_mapPrototypeSetWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_setPrototypeAddWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_numberPrototypeToStringWatchpoint;
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_arrayBufferConstructorSpeciesWatchpoints[2];
    OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>> m_arrayBufferPrototypeConstructorWatchpoints[2];

    bool isArrayPrototypeIteratorProtocolFastAndNonObservable() const;
    bool isMapPrototypeIteratorProtocolFastAndNonObservable() const;
    bool isSetPrototypeIteratorProtocolFastAndNonObservable() const;
    bool isStringPrototypeIteratorProtocolFastAndNonObservable() const;
    bool isMapPrototypeSetFastAndNonObservable() const;
    bool isSetPrototypeAddFastAndNonObservable() const;

    WeakGCMap<GetValueFunc, JSCustomGetterFunction>& customGetterFunctionMap() { return m_customGetterFunctionMap; }
    WeakGCMap<PutValueFunc, JSCustomSetterFunction>& customSetterFunctionMap() { return m_customSetterFunctionMap; }

#if ENABLE(DFG_JIT)
    typedef HashMap<RefPtr<UniquedStringImpl>, RefPtr<WatchpointSet>, IdentifierRepHash> ReferencedGlobalPropertyWatchpointSets;
    ReferencedGlobalPropertyWatchpointSets m_referencedGlobalPropertyWatchpointSets;
#endif

	bool m_evalEnabled;
	unsigned m_globalLexicalBindingEpoch;
    String m_evalDisabledErrorMessage;
	bool m_experimentsEnabled;
    WeakPtr<ConsoleClient> m_consoleClient;
    WeakGCMap<GetValueFunc, JSCustomGetterFunction> m_customGetterFunctionMap;
    WeakGCMap<PutValueFunc, JSCustomSetterFunction> m_customSetterFunctionMap;

	static JS_EXPORTDATA const GlobalObjectMethodTable s_globalObjectMethodTable;
	const GlobalObjectMethodTable* m_globalObjectMethodTable;

	void createRareDataIfNeeded()
	{
		if (m_rareData)
			return;
		m_rareData = adoptPtr(new JSGlobalObjectRareData);
	}

public:
	typedef JSSegmentedVariableObject Base;
	static const unsigned StructureFlags = Base::StructureFlags | HasStaticPropertyTable | OverridesGetOwnPropertySlot | OverridesPut | IsImmutablePrototypeExoticObject;

	static JSGlobalObject* create(JSGlobalData& globalData, Structure* structure)
	{
		JSGlobalObject* globalObject = new (NotNull, allocateCell<JSGlobalObject>(globalData.heap)) JSGlobalObject(globalData, structure);
		globalObject->finishCreation(globalData);
		globalData.heap.addFinalizer(globalObject, destroy);
		return globalObject;
	}

	DECLARE_EXPORT_INFO;

    bool hasDebugger() const { return m_debugger; }
    bool hasProfiler() const { return globalObjectMethodTable()->supportsProfiling(this); }

#if ENABLE(DFG_JIT)
    WatchpointSet* getReferencedPropertyWatchpointSet(UniquedStringImpl*);
    WatchpointSet& ensureReferencedPropertyWatchpointSet(UniquedStringImpl*);
#endif

protected:
	JS_EXPORT_PRIVATE explicit JSGlobalObject(JSGlobalData&, Structure*, const GlobalObjectMethodTable* = 0);

	void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		structure()->setGlobalObject(vm, this);
		m_experimentsEnabled = m_globalObjectMethodTable->javaScriptExperimentsEnabled(this);
		init(vm);
		setGlobalThis(vm, JSProxy::create(vm, JSProxy::createStructure(vm, this, getPrototypeDirect(), PureForwardingProxyType), this));
	}

	void finishCreation(JSGlobalData& vm, JSObject* thisValue)
	{
		Base::finishCreation(vm);
		structure()->setGlobalObject(vm, this);
		m_experimentsEnabled = m_globalObjectMethodTable->javaScriptExperimentsEnabled(this);
		init(vm);
        setGlobalThis(vm, thisValue);
	}

    void addSymbolTableEntry(const Identifier&);

public:
	JS_EXPORT_PRIVATE ~JSGlobalObject();
	JS_EXPORT_PRIVATE static void destroy(JSCell*);
	// We don't need a destructor because we use a finalizer instead.
	static const bool needsDestruction = false;

	JS_EXPORT_PRIVATE static void visitChildren(JSCell*, SlotVisitor&);

	JS_EXPORT_PRIVATE static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);
	JS_EXPORT_PRIVATE static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);
	JS_EXPORT_PRIVATE static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool shouldThrow);

    bool canDeclareGlobalFunction(ExecState*, const Identifier&);
    template<BindingCreationContext> void createGlobalFunctionBinding(ExecState*, const Identifier&);

    inline bool canDeclareGlobalVar(ExecState*, const Identifier&);
    template<BindingCreationContext> inline void createGlobalVarBinding(ExecState*, const Identifier&);

    JSScope* globalScope() { return m_globalLexicalEnvironment.get(); }
    JSGlobalLexicalEnvironment* globalLexicalEnvironment() { return m_globalLexicalEnvironment.get(); }

	// The following accessors return pristine values, even if a script
	// replaces the global object's associated property.

    GetterSetter* speciesGetterSetter() const { return m_speciesGetterSetter.get(); }

	RegExpConstructor* regExpConstructor() const { return m_regExpConstructor.get(); }

	DateConstructor* dateConstructor() const { return m_dateConstructor.get(); }
	ErrorConstructor* errorConstructor() const { return m_errorConstructor.get(); }
	ArrayConstructor* arrayConstructor() const { return m_arrayConstructor.get(); }
	ObjectConstructor* objectConstructor() const { return m_objectConstructor.get(); }
	FunctionConstructor* functionConstructor() const { return m_functionConstructor.get(); }
	JSPromiseConstructor* promiseConstructor() const { return m_promiseConstructor.get(); }
	JSInternalPromiseConstructor* internalPromiseConstructor() const { return m_internalPromiseConstructor.get(); }

    JSFunction* parseIntFunction() const { return m_parseIntFunction.get(); }
    JSFunction* parseFloatFunction() const { return m_parseFloatFunction.get(); }

#if ENABLE(INTL)
    IntlObject* intlObject() const { return m_intlObject.get(); }
    IntlCollator* defaultCollator(ExecState*);

    Structure* collatorStructure() const;
    Structure* dateTimeFormatStructure() const;
    Structure* numberFormatStructure() const;
    Structure* localeStructure() const;
    Structure* pluralRulesStructure() const;
    Structure* relativeTimeFormatStructure() const;
#endif

    NullGetterFunction* nullGetterFunction() const { return m_nullGetterFunction.get(); }
    NullSetterFunction* nullSetterFunction() const { return m_nullSetterFunction.get(); }
    NullSetterFunction* nullSetterStrictFunction() const { return m_nullSetterStrictFunction.get(); }

	JSFunction* evalFunction() const { return m_evalFunction.get(); }
	JSFunction* callFunction() const { return m_callFunction.get(); }
	JSFunction* applyFunction() const { return m_applyFunction.get(); }
	JSFunction* objectProtoToStringFunction() const { return m_objectProtoToStringFunction.get(); }
	JSFunction* arrayProtoToStringFunction() const { return m_arrayProtoToStringFunction.get(); }
	JSFunction* arrayProtoValuesFunction() const { return m_arrayProtoValuesFunction.get(); }
	JSFunction* initializePromiseFunction() const { return m_initializePromiseFunction.get(); }
	JSFunction* iteratorProtocolFunction() const { return m_iteratorProtocolFunction.get(); }
	JSFunction* promiseResolveFunction() const { return m_promiseResolveFunction.get(); }
	JSFunction* numberProtoToStringFunction() const { return m_numberProtoToStringFunction.get(); }
	JSFunction* objectProtoValueOfFunction() const { return m_objectProtoValueOfFunction.get(); }
	JSFunction* newPromiseCapabilityFunction() const { return m_newPromiseCapabilityFunction.get(); }
	JSFunction* functionProtoHasInstanceSymbolFunction() const { return m_functionProtoHasInstanceSymbolFunction.get(); }
	JSFunction* setPrototypeDirectFunction() const { return m_setPrototypeDirectFunction.get(); }
	JSObject* regExpProtoExecFunction() const { return m_regExpProtoExec.get(); }
	JSObject* regExpProtoSymbolReplaceFunction() const { return m_regExpProtoSymbolReplace.get(); }
	GetterSetter* regExpProtoGlobalGetter() const { return m_regExpProtoGlobalGetter.get(); }
	GetterSetter* regExpProtoUnicodeGetter() const { return m_regExpProtoUnicodeGetter.get(); }
	GetterSetter* throwTypeErrorArgumentsCalleeGetterSetter() { return m_throwTypeErrorArgumentsCalleeGetterSetter.get(); }

    JSModuleLoader* moduleLoader() const { return m_moduleLoader.get(); }

	ObjectPrototype* objectPrototype() const { return m_objectPrototype.get(); }
	FunctionPrototype* functionPrototype() const { return m_functionPrototype.get(); }
	ArrayPrototype* arrayPrototype() const { return m_arrayPrototype.get(); }
	BooleanPrototype* booleanPrototype() const { return m_booleanPrototype.get(); }
	StringPrototype* stringPrototype() const { return m_stringPrototype.get(); }
    SymbolPrototype* symbolPrototype() const { return m_symbolPrototype.get(); }
	NumberPrototype* numberPrototype() const { return m_numberPrototype.get(); }
    BigIntPrototype* bigIntPrototype() const { return m_bigIntPrototype.get(); }
	DatePrototype* datePrototype() const { return m_datePrototype.get(); }
	RegExpPrototype* regExpPrototype() const { return m_regExpPrototype.get(); }
	ErrorPrototype* errorPrototype() const { return m_errorPrototype.get(); }
	IteratorPrototype* iteratorPrototype() const { return m_iteratorPrototype.get(); }
	AsyncIteratorPrototype* asyncIteratorPrototype() const { return m_asyncIteratorPrototype.get(); }
	GeneratorFunctionPrototype* generatorFunctionPrototype() const { return m_generatorFunctionPrototype.get(); }
	GeneratorPrototype* generatorPrototype() const { return m_generatorPrototype.get(); }
	AsyncGeneratorPrototype* asyncGeneratorPrototype() const { return m_asyncGeneratorPrototype.get(); }
	AsyncGeneratorFunctionPrototype* asyncGeneratorFunctionPrototype() const { return m_asyncGeneratorFunctionPrototype.get(); }

	Structure* lazyAsyncFunctionStructure()
    {
        return m_asyncFunctionStructure.get();
    }
    const Structure* lazyAsyncFunctionStructure() const
    {
        return m_asyncFunctionStructure.get();
    }
    AsyncFunctionPrototype* asyncFunctionPrototype() const { return m_asyncFunctionPrototype.get(); }
    AsyncFunctionPrototype* asyncFunctionPrototypeConcurrently() const { return m_asyncFunctionPrototype.get(); }
    ArrayIteratorPrototype* arrayIteratorPrototype() const { return m_arrayIteratorPrototype.get(); }
    MapPrototype* mapPrototype() const { return m_mapPrototype.get(); }
    // Workaround for the name conflict between JSCell::setPrototype.
    SetPrototype* jsSetPrototype() const { return m_setPrototype.get(); }

	Structure* withScopeStructure() const { return m_withScopeStructure.get(); }
	Structure* strictEvalActivationStructure() const { return m_strictEvalActivationStructure.get(); }
	Structure* activationStructure() const { return m_lexicalEnvironmentStructure.get(); }
	Structure* moduleEnvironmentStructure() const { return m_moduleEnvironmentStructure.get(); }
	Structure* directArgumentsStructure() const { return m_directArgumentsStructure.get(); }
	Structure* scopedArgumentsStructure() const { return m_scopedArgumentsStructure.get(); }
	Structure* clonedArgumentsStructure() const { return m_clonedArgumentsStructure.get(); }
	Structure* objectStructureForObjectConstructor() const { return m_objectStructureForObjectConstructor.get(); }
	ALWAYS_INLINE Structure* originalArrayStructureForIndexingType(IndexingType indexingType) const
	{
		ASSERT(indexingType & IsArray);
		return m_originalArrayStructureForIndexingShape[(indexingType & IndexingShapeMask) >> IndexingShapeShift].get();
	}
	ALWAYS_INLINE Structure* arrayStructureForIndexingTypeDuringAllocation(IndexingType indexingType) const
	{
		ASSERT(indexingType & IsArray);
		return m_arrayStructureForIndexingShapeDuringAllocation[(indexingType & IndexingShapeMask) >> IndexingShapeShift].get();
	}
	Structure* arrayStructureForIndexingTypeDuringAllocation(ExecState* exec, IndexingType indexingType, JSValue newTarget) const;
	ALWAYS_INLINE Structure* arrayStructureForProfileDuringAllocation(ExecState* exec, ArrayAllocationProfile* profile, JSValue newTarget) const
    {
        return arrayStructureForIndexingTypeDuringAllocation(exec, ArrayAllocationProfile::selectIndexingTypeFor(profile), newTarget);
    }

    ALWAYS_INLINE bool isOriginalArrayStructure(Structure* structure) const
	{
		return originalArrayStructureForIndexingType(structure->indexingType() | IsArray) == structure;
	}

	Structure* booleanObjectStructure() const { return m_booleanObjectStructure.get(); }
#if !PLATFORM(ESPIAL)
	Structure* callbackConstructorStructure() const { return m_callbackConstructorStructure.get(); }
	Structure* callbackFunctionStructure() const { return m_callbackFunctionStructure.get(); }
	Structure* callbackObjectStructure() const { return m_callbackObjectStructure.get(); }
#endif
	Structure* dateStructure() const { return m_dateStructure.get(); }
	Structure* nullPrototypeObjectStructure() const { return m_nullPrototypeObjectStructure.get(); }
	Structure* errorStructure() const { return m_errorStructure.get(); }
	Structure* errorStructure(ErrorType errorType) const
	{
		switch (errorType) {
	    case ErrorType::Error:
	        return errorStructure();
	    case ErrorType::EvalError:
	        return m_evalErrorStructure.get();
	    case ErrorType::RangeError:
	        return m_rangeErrorStructure.get();
	    case ErrorType::ReferenceError:
	        return m_referenceErrorStructure.get();
	    case ErrorType::SyntaxError:
	        return m_syntaxErrorStructure.get();
	    case ErrorType::TypeError:
	        return m_typeErrorStructure.get();
	    case ErrorType::URIError:
	        return m_URIErrorStructure.get();
		case ErrorType::AggregateError:
			return m_aggregateErrorStructure.get();
		}
		ASSERT_NOT_REACHED();
		return nullptr;
	}
    template<ErrorType errorType> Structure* errorStructureWithErrorType() const { return errorStructure(errorType); }

    Structure* calleeStructure() const { return m_calleeStructure.get(); }
    Structure* hostFunctionStructure() const { return m_hostFunctionStructure.get(); }

    Structure* arrowFunctionStructure(bool isBuiltin) const
    {
        if (isBuiltin)
            return m_builtinFunctions.arrowFunctionStructure.get();
        return m_ordinaryFunctions.arrowFunctionStructure.get();
    }
    Structure* sloppyFunctionStructure(bool isBuiltin) const
    {
        if (isBuiltin)
            return m_builtinFunctions.sloppyFunctionStructure.get();
        return m_ordinaryFunctions.sloppyFunctionStructure.get();
    }
    Structure* strictFunctionStructure(bool isBuiltin) const
    {
        if (isBuiltin)
            return m_builtinFunctions.strictFunctionStructure.get();
        return m_ordinaryFunctions.strictFunctionStructure.get();
    }

	Structure* boundFunctionStructure() const { return m_boundFunctionStructure.get(); }
	Structure* customGetterFunctionStructure() const { return m_customGetterFunctionStructure.get(); }
	Structure* customSetterFunctionStructure() const { return m_customSetterFunctionStructure.get(); }
	Structure* nativeStdFunctionStructure() const { return m_nativeStdFunctionStructure.get(); }
	PropertyOffset functionNameOffset() const { return m_functionNameOffset; }
	Structure* numberObjectStructure() const { return m_numberObjectStructure.get(); }
	Structure* internalFunctionStructure() const { return m_internalFunctionStructure.get(); }
	Structure* mapStructure() const { return m_mapStructure.get(); }
	Structure* regExpStructure() const { return m_regExpStructure.get(); }
	Structure* generatorFunctionStructure() const { return m_generatorFunctionStructure.get(); }
    Structure* asyncFunctionStructure() const { return m_asyncFunctionStructure.get(); }
    Structure* asyncFunctionStructureConcurrently() const { return m_asyncFunctionStructure.get(); }
    Structure* asyncGeneratorFunctionStructure() const { return m_asyncGeneratorFunctionStructure.get(); }
	Structure* stringObjectStructure() const { return m_stringObjectStructure.get(); }
	Structure* arrayIteratorStructure() const { return m_arrayIteratorStructure.get(); }
	Structure* symbolObjectStructure() const { return m_symbolObjectStructure.get(); }
	Structure* bigIntObjectStructure() const { return m_bigIntObjectStructure.get(); }
	Structure* iteratorResultObjectStructure() const { return m_iteratorResultObjectStructure.get(); }
	Structure* regExpMatchesArrayStructure() const { return m_regExpMatchesArrayStructure.get(); }
	Structure* regExpMatchesArrayWithIndicesStructure() const { return m_regExpMatchesArrayWithIndicesStructure.get(); }
	Structure* regExpMatchesIndicesArrayStructure() const { return m_regExpMatchesIndicesArrayStructure.get(); }
	Structure* proxyObjectStructure() const { return m_proxyObjectStructure.get(); }
	Structure* callableProxyObjectStructure() const { return m_callableProxyObjectStructure.get(); }
    Structure* proxyRevokeStructure() const { return m_proxyRevokeStructure.get(); }
	Structure* moduleRecordStructure() const { return m_moduleRecordStructure.get(); }
	Structure* moduleNamespaceObjectStructure() const { return m_moduleNamespaceObjectStructure.get(); }
	Structure* restParameterStructure() const { return arrayStructureForIndexingTypeDuringAllocation(ArrayWithContiguous); }

    JS_EXPORT_PRIVATE void setConsoleClient(ConsoleClient* consoleClient);
    ConsoleClient* consoleClient() const { return m_consoleClient.get(); }

    StructureCache& structureCache() { return m_structureCache; }

    void bumpGlobalLexicalBindingEpoch(VM&);
    unsigned globalLexicalBindingEpoch() const { return m_globalLexicalBindingEpoch; }
    static ptrdiff_t globalLexicalBindingEpochOffset() { return OBJECT_OFFSETOF(JSGlobalObject, m_globalLexicalBindingEpoch); }
    unsigned* addressOfGlobalLexicalBindingEpoch() { return &m_globalLexicalBindingEpoch; }

    JSArrayBufferPrototype* arrayBufferPrototype(ArrayBufferSharingMode sharingMode = ArrayBufferSharingMode::Default) const
    {
        switch (sharingMode) {
        case ArrayBufferSharingMode::Default:
        	return m_arrayBufferPrototype.get();
        case ArrayBufferSharingMode::Shared:
#if ENABLE(SHARED_ARRAY_BUFFER)
            return m_sharedArrayBufferStructure.prototype(this);
#else
            return nullptr;
#endif
        }
        RELEASE_ASSERT_NOT_REACHED();
        return nullptr;
    }
    Structure* arrayBufferStructure(ArrayBufferSharingMode sharingMode) const
    {
        switch (sharingMode) {
        case ArrayBufferSharingMode::Default:
            return m_arrayBufferStructure.get();
        case ArrayBufferSharingMode::Shared:
#if ENABLE(SHARED_ARRAY_BUFFER)
            return m_sharedArrayBufferStructure.get(this);
#else
            return nullptr;
#endif
        }
        RELEASE_ASSERT_NOT_REACHED();
        return nullptr;
    }
    template<ArrayBufferSharingMode sharingMode> Structure* arrayBufferStructureWithSharingMode() const { return arrayBufferStructure(sharingMode); }
    JSObject* arrayBufferConstructor(ArrayBufferSharingMode sharingMode) const
    {
        switch (sharingMode) {
        case ArrayBufferSharingMode::Default:
            return m_arrayBufferConstructor.get();
        case ArrayBufferSharingMode::Shared:
#if ENABLE(SHARED_ARRAY_BUFFER)
            return m_sharedArrayBufferStructure.constructor(this);
#else
            return nullptr;
#endif
        }
        RELEASE_ASSERT_NOT_REACHED();
        return nullptr;
    }

#define DEFINE_ACCESSORS_FOR_SIMPLE_TYPE(capitalName, lowerName, properName, instanceType, jsName) \
    Structure* properName ## Structure() { return m_ ## properName ## Structure.get(); }

    FOR_EACH_SIMPLE_BUILTIN_TYPE(DEFINE_ACCESSORS_FOR_SIMPLE_TYPE)
    FOR_BIG_INT_BUILTIN_TYPE_WITH_CONSTRUCTOR(DEFINE_ACCESSORS_FOR_SIMPLE_TYPE)
    FOR_EACH_BUILTIN_DERIVED_ITERATOR_TYPE(DEFINE_ACCESSORS_FOR_SIMPLE_TYPE)

#undef DEFINE_ACCESSORS_FOR_SIMPLE_TYPE

    Structure* typedArrayStructure(TypedArrayType type) const
    {
        return m_typedArrays[toIndex(type)].structure.get();
    }

    bool isOriginalTypedArrayStructure(Structure* structure)
    {
        TypedArrayType type = structure->classInfo()->typedArrayStorageType;
        if (type == NotTypedArray)
            return false;
        return typedArrayStructure(type) == structure;
    }
    template<TypedArrayType type> Structure* typedArrayStructureWithTypedArrayType() const { return typedArrayStructure(type); }

    JSObject* typedArrayConstructor(TypedArrayType type) const
    {
        return m_typedArrays[toIndex(type)].constructor.get();
    }

    JSCell* actualPointerFor(Special::Pointer pointer)
    {
        ASSERT(pointer < Special::TableSize);
        return m_specialPointers[pointer];
    }
    JSCell* jsCellForLinkTimeConstant(LinkTimeConstant type);

	WatchpointSet* masqueradesAsUndefinedWatchpoint() { return m_masqueradesAsUndefinedWatchpoint.get(); }
	WatchpointSet* havingABadTimeWatchpoint() { return m_havingABadTimeWatchpoint.get(); }
	WatchpointSet* varInjectionWatchpoint() { return m_varInjectionWatchpoint.get(); }
	WatchpointSet* varReadOnlyWatchpoint() { return m_varReadOnlyWatchpoint.get(); }

	bool isHavingABadTime() const
	{
		return m_havingABadTimeWatchpoint->hasBeenInvalidated();
	}

	void haveABadTime(JSGlobalData&);
    void clearStructureCache(VM&);

    bool objectPrototypeIsSane() const;
    bool arrayPrototypeChainIsSane() const;
    bool stringPrototypeChainIsSane() const;

	void setProfileGroup(unsigned value) { createRareDataIfNeeded(); m_rareData->profileGroup = value; }
	unsigned profileGroup() const
	{
		if (!m_rareData)
			return 0;
		return m_rareData->profileGroup;
	}

	Debugger* debugger() const { return m_debugger; }
	void setDebugger(Debugger* debugger) { m_debugger = debugger; }

	const GlobalObjectMethodTable* globalObjectMethodTable() const { return m_globalObjectMethodTable; }

	static bool supportsProfiling(const JSGlobalObject*) { return false; }
	static bool supportsRichSourceInfo(const JSGlobalObject*) { return true; }

	JS_EXPORT_PRIVATE ExecState* globalExec();

	static bool shouldInterruptScript(const JSGlobalObject*) { return true; }
	static bool javaScriptExperimentsEnabled(const JSGlobalObject*) { return false; }

    void queueMicrotask(PassRefPtr<Microtask>);

    bool evalEnabled() const { return m_evalEnabled; }
    const String& evalDisabledErrorMessage() const { return m_evalDisabledErrorMessage; }
    void setEvalEnabled(bool enabled, const String& errorMessage = String())
    {
        m_evalEnabled = enabled;
        m_evalDisabledErrorMessage = errorMessage;
    }

	void resetPrototype(JSGlobalData&, JSValue prototype);

	JSGlobalData& globalData() const { return m_vm; }
	JSGlobalData& vm() const { return m_vm; }
	JSObject* globalThis() const;
    WriteBarrier<JSObject>* addressOfGlobalThis() { return &m_globalThis; }

	static Structure* createStructure(JSGlobalData& globalData, JSValue prototype)
	{
		Structure* result = Structure::create(globalData, 0, prototype, TypeInfo(GlobalObjectType, StructureFlags), info());
		result->setTransitionWatchpointIsLikelyToBeFired(true);
		return result;
	}

#if !PLATFORM(ESPIAL)
	void registerWeakMap(OpaqueJSWeakObjectMap* map)
	{
		createRareDataIfNeeded();
		m_rareData->weakMaps.add(map);
	}

	void unregisterWeakMap(OpaqueJSWeakObjectMap* map)
	{
		if (m_rareData)
			m_rareData->weakMaps.remove(map);
	}

    OpaqueJSClassDataMap& opaqueJSClassData()
    {
        createRareDataIfNeeded();
        return m_rareData->opaqueJSClassData;
    }
#endif

    static ptrdiff_t weakRandomOffset() { return OBJECT_OFFSETOF(JSGlobalObject, m_weakRandom); }
	double weakRandomNumber() { return m_weakRandom.get(); }
	unsigned weakRandomInteger() { return m_weakRandom.getUint32(); }

    BuiltinExecutables* builtinExecutables() { return m_builtinExecutables.get(); }

    bool needsSiteSpecificQuirks() const { return m_needsSiteSpecificQuirks; }

    void tryInstallArraySpeciesWatchpoint(ExecState*);
    void tryInstallArrayBufferSpeciesWatchpoint(ExecState*, ArrayBufferSharingMode);

protected:
    void tryInstallSpeciesWatchpoint(ExecState*, JSObject* prototype, JSObject* constructor, OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>>& constructorWatchpoint, OwnPtr<ObjectPropertyChangeAdaptiveWatchpoint<InlineWatchpointSet>>& speciesWatchpoint, InlineWatchpointSet& speciesWatchpointSet);

	struct GlobalPropertyInfo {
		GlobalPropertyInfo(const Identifier& i, JSValue v, unsigned a)
			: identifier(i)
			, attributes(a)
			, value(v)
		{
		}

		const Identifier identifier;
		const unsigned attributes;
		JSValue value;
	};
	JS_EXPORT_PRIVATE void addStaticGlobals(GlobalPropertyInfo*, int count);

    void setNeedsSiteSpecificQuirks(bool needQuirks) { m_needsSiteSpecificQuirks = needQuirks; }

private:
	friend class LLIntOffsetsExtractor;

    void fireWatchpointAndMakeAllArrayStructuresSlowPut(VM&);
    void setGlobalThis(VM&, JSObject* globalThis);

	// FIXME: Fold reset into init.
	JS_EXPORT_PRIVATE void init(JSGlobalData&);
	void reset(JSValue prototype);

	JS_EXPORT_PRIVATE static void clearRareData(JSCell*);

    bool m_needsSiteSpecificQuirks;
};

template<>
ALWAYS_INLINE JSGlobalObject* jsDynamicCast(JSCell* from)
{
	if(from->structure()->typeInfo().type() == JSType::GlobalObjectType)
		return static_cast<JSGlobalObject*>(from);
	return 0;
}

JSGlobalObject* asGlobalObject(JSValue);

ALWAYS_INLINE JSGlobalObject* asGlobalObject(JSValue value)
{
	ASSERT(asObject(value)->isGlobalObject());
	return jsCast<JSGlobalObject*>(asObject(value));
}

inline JSArray* constructEmptyArray(ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, unsigned initialLength = 0, JSValue newTarget = JSValue())
{
	VM& vm = globalObject->vm();
    Structure* structure;
    if (initialLength >= MIN_ARRAY_STORAGE_CONSTRUCTION_LENGTH)
        structure = globalObject->arrayStructureForIndexingTypeDuringAllocation(exec, ArrayWithArrayStorage, newTarget);
    else
        structure = globalObject->arrayStructureForProfileDuringAllocation(exec, profile, newTarget);
	RETURN_IF_EXCEPTION(vm, nullptr);

	auto* jsArray = JSArray::tryCreate(vm, exec,  structure, initialLength);
	RETURN_IF_EXCEPTION(vm, nullptr);
	return ArrayAllocationProfile::updateLastAllocationFor(profile, jsArray);
}

inline JSArray* constructEmptyArray(ExecState* exec, ArrayAllocationProfile* profile, unsigned initialLength = 0, JSValue newTarget = JSValue())
{
    return constructEmptyArray(exec, profile, exec->lexicalGlobalObject(), initialLength, newTarget);
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, const ArgList& values, JSValue newTarget = JSValue())
{
	VM& vm = globalObject->vm();
	Structure* structure = globalObject->arrayStructureForProfileDuringAllocation(exec, profile, newTarget);
	RETURN_IF_EXCEPTION(vm, nullptr);

	auto* jsArray = constructArray(vm, exec, structure, values);
	RETURN_IF_EXCEPTION(vm, nullptr);
	return ArrayAllocationProfile::updateLastAllocationFor(profile, jsArray);
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, const ArgList& values, JSValue newTarget = JSValue())
{
    return constructArray(exec, profile, exec->lexicalGlobalObject(), values, newTarget);
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, const JSValue* values, unsigned length, JSValue newTarget = JSValue())
{
	VM& vm = globalObject->vm();
	Structure* structure = globalObject->arrayStructureForProfileDuringAllocation(exec, profile, newTarget);
	RETURN_IF_EXCEPTION(vm, nullptr);

	auto* jsArray = constructArray(vm, exec, structure, values, length);
	RETURN_IF_EXCEPTION(vm, nullptr);
	return ArrayAllocationProfile::updateLastAllocationFor(profile, jsArray);
}

inline JSArray* constructArray(ExecState* exec, ArrayAllocationProfile* profile, const JSValue* values, unsigned length, JSValue newTarget = JSValue())
{
    return constructArray(exec, profile, exec->lexicalGlobalObject(), values, length, newTarget);
}

inline JSArray* constructArrayNegativeIndexed(VM& vm, ExecState* exec, ArrayAllocationProfile* profile, JSGlobalObject* globalObject, const JSValue* values, unsigned length, JSValue newTarget = JSValue())
{
	Structure* structure = globalObject->arrayStructureForProfileDuringAllocation(exec, profile, newTarget);
	RETURN_IF_EXCEPTION(vm, nullptr);

	auto* jsArray = constructArrayNegativeIndexed(vm, exec, structure, values, length);
	RETURN_IF_EXCEPTION(vm, nullptr);
	return ArrayAllocationProfile::updateLastAllocationFor(profile, jsArray);
}

inline JSArray* constructArrayNegativeIndexed(VM& vm, ExecState* exec, ArrayAllocationProfile* profile, const JSValue* values, unsigned length, JSValue newTarget = JSValue())
{
    return constructArrayNegativeIndexed(vm, exec, profile, exec->lexicalGlobalObject(), values, length, newTarget);
}

ALWAYS_INLINE JSObject* ExecState::globalThisValue() const
{
    return lexicalGlobalObject()->globalThis();
}

ALWAYS_INLINE JSObject* JSScope::globalThis()
{
	return globalObject()->globalThis();
}

ALWAYS_INLINE ExecState* JSGlobalObject::globalExec()
{
    return CallFrame::create(m_globalCallFrame);
}

ALWAYS_INLINE JSObject* JSGlobalObject::globalThis() const
{
	return m_globalThis.get();
}

ALWAYS_INLINE bool JSGlobalObject::isArrayPrototypeIteratorProtocolFastAndNonObservable() const
{
    // We're fast if we don't have any indexed properties on the prototype.
    // We're non-observable if the iteration protocol hasn't changed.
    //
    // Note: it only makes sense to call this from the main thread. If you're
    // trying to prove this behavior on the compiler thread, you'll want to
    // carefully set up watchpoints to have correct ordering while JS code is
    // executing concurrently.

    return arrayIteratorProtocolWatchpointSet().isStillValid() && !isHavingABadTime() && arrayPrototypeChainIsSane();
}

// We're non-observable if the iteration protocol hasn't changed.
//
// Note: it only makes sense to call this from the main thread. If you're
// trying to prove this behavior on the compiler thread, you'll want to
// carefully set up watchpoints to have correct ordering while JS code is
// executing concurrently.
ALWAYS_INLINE bool JSGlobalObject::isMapPrototypeIteratorProtocolFastAndNonObservable() const
{
    return mapIteratorProtocolWatchpoint().isStillValid();
}

ALWAYS_INLINE bool JSGlobalObject::isSetPrototypeIteratorProtocolFastAndNonObservable() const
{
    return setIteratorProtocolWatchpoint().isStillValid();
}

ALWAYS_INLINE bool JSGlobalObject::isStringPrototypeIteratorProtocolFastAndNonObservable() const
{
    return stringIteratorProtocolWatchpoint().isStillValid();
}

ALWAYS_INLINE bool JSGlobalObject::isMapPrototypeSetFastAndNonObservable() const
{
    return mapSetWatchpoint().isStillValid();
}

ALWAYS_INLINE bool JSGlobalObject::isSetPrototypeAddFastAndNonObservable() const
{
    return setAddWatchpoint().isStillValid();
}

// https://tc39.es/ecma262/#sec-candeclareglobalvar
inline bool JSGlobalObject::canDeclareGlobalVar(ExecState* exec, const Identifier& ident)
{
    if (LIKELY(isStructureExtensible()))
        return true;

    PropertySlot slot(this, PropertySlot::InternalMethodType::GetOwnProperty);
    return getOwnPropertySlot(this, exec, ident, slot);
}

// https://tc39.es/ecma262/#sec-createglobalvarbinding
template<BindingCreationContext context>
inline void JSGlobalObject::createGlobalVarBinding(ExecState* exec, const Identifier& ident)
{
    VM& vm = this->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    PropertySlot slot(this, PropertySlot::InternalMethodType::GetOwnProperty);
    bool hasProperty = getOwnPropertySlot(this, exec, ident, slot);
    //scope.assertNoExceptionExceptTermination();
    if (UNLIKELY(hasProperty))
        return;

    ASSERT(isStructureExtensible());
    if (context == BindingCreationContext::Global)
        addSymbolTableEntry(ident);
    else
        putDirect(vm, ident, jsUndefined());
}

} // namespace JSC
