/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "JSMicrotask.h"

#include "Error.h"
#include "JSCJSValueInlines.h"
#include "JSCellInlines.h"
#include "JSGlobalObject.h"
#include "JSObjectInlines.h"
#include "Microtask.h"
#include "SlotVisitorInlines.h"
#include "StrongInlines.h"

namespace JSC {

class JSMicrotask FINAL : public Microtask {
public:
	JSMicrotask(VM& vm, JSValue job, JSArray* arguments)
    {
        m_job.set(vm, job);
        m_arguments.set(vm, arguments);
    }

    JSMicrotask(VM& vm, JSValue job)
    {
        m_job.set(vm, job);
    }

    virtual bool isJSJob() const { return true; }

private:
    virtual void run(ExecState*) OVERRIDE;

    Strong<Unknown> m_job;
    Strong<JSArray> m_arguments;
};

PassRefPtr<Microtask> createJSMicrotask(VM& vm, JSValue job)
{
    return adoptRef(new JSMicrotask(vm, job));
}

PassRefPtr<Microtask> createJSMicrotask(VM& vm, JSValue job, JSArray* arguments)
{
    return adoptRef(new JSMicrotask(vm, job, arguments));
}

void JSMicrotask::run(ExecState* exec)
{
    VM& vm = exec->vm();
    CallData handlerCallData = getCallData(vm, m_job.get());
    ASSERT(handlerCallData.type != CallData::Type::None);

    MarkedArgumentBuffer handlerArguments;
    if (m_arguments) {
		for (unsigned index = 0, length = m_arguments->length(); index < length; ++index) {
			JSValue arg = m_arguments->JSArray::get(vm, exec, index);
			if(UNLIKELY(vm.exception())) {
				vm.clearException();
				return handlerArguments.overflowCheckNotNeeded();
			}
			handlerArguments.append(arg);
		}
    }
    call(exec, m_job.get(), handlerCallData, jsUndefined(), handlerArguments);
    vm.clearException();
}

} // namespace JSC
