/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003-2022 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "ArgList.h"
#include "ArrayConventions.h"
#include "ArrayStorage.h"
#include "Butterfly.h"
#include "ClassInfo.h"
#include "CommonIdentifiers.h"
#include "CallFrame.h"
#include "CustomGetterSetter.h"
#include "DeferGC.h"
#include "ExceptionScope.h"
#include "JSCell.h"
#include "MathCommon.h"
#include "ObjectInitializationScope.h"
#include "PropertySlot.h"
#include "PropertyStorage.h"
#include "PutDirectIndexMode.h"
#include "PutPropertySlot.h"
#include "RuntimeType.h"
#include "CustomGetterSetter.h"
#include "Structure.h"
#include "JSGlobalData.h"
#include "JSString.h"
#include "SlotVisitorInlines.h"
#include "SparseArrayValueMap.h"
#include <wtf/StdLibExtras.h>
#include <wtf/text/ASCIILiteral.h>

namespace JSC {

ALWAYS_INLINE JSCell* getJSFunction(JSValue value)
{
	if (value.isCell() && (value.asCell()->structure()->typeInfo().type() == JSFunctionType))
		return value.asCell();
	return nullptr;
}

class Exception;
class GetterSetter;
class InternalFunction;
class JSFunction;
class LLIntOffsetsExtractor;
class MarkedBlock;
class PropertyDescriptor;
class PropertyNameArray;
class Structure;
struct HashTable;
struct HashTableValue;

JS_EXPORT_PRIVATE bool isTerminatedExecutionException(VM&, Exception*);
JS_EXPORT_PRIVATE Exception* throwTypeError(ExecState*, VM&, const String&);
extern JS_EXPORTDATA const ASCIILiteral ReadonlyPropertyWriteError;
extern JS_EXPORTDATA const ASCIILiteral ReadonlyPropertyChangeError;
extern JS_EXPORTDATA const ASCIILiteral UnconfigurablePropertyChangeAccessMechanismError;
extern JS_EXPORTDATA const ASCIILiteral UnconfigurablePropertyChangeConfigurabilityError;
extern JS_EXPORTDATA const ASCIILiteral UnconfigurablePropertyChangeEnumerabilityError;
extern JS_EXPORTDATA const ASCIILiteral UnconfigurablePropertyChangeWritabilityError;
extern JS_EXPORTDATA const ASCIILiteral PrototypeValueCanOnlyBeAnObjectOrNullTypeError;

class JSFinalObject;

extern JS_EXPORT_PRIVATE const ASCIILiteral NonExtensibleObjectPropertyDefineError;
extern JS_EXPORT_PRIVATE const ASCIILiteral typedArrayBufferHasBeenDetachedErrorMessage;

#if !ASSERT_DISABLED
#define JS_EXPORT_PRIVATE_IF_ASSERT_ENABLED JS_EXPORT_PRIVATE
#else
#define JS_EXPORT_PRIVATE_IF_ASSERT_ENABLED
#endif

class JSObject : public JSCell {
	friend class BatchedTransitionOptimizer;
	friend class JIT;
	friend class JSCell;
	friend class JSFinalObject;
	friend class MarkedBlock;
	JS_EXPORT_PRIVATE friend bool setUpStaticFunctionSlot(VM&, const HashTableValue*, JSObject*, PropertyName, PropertySlot&);

    enum PutMode : uint8_t {
        PutModePut,
        PutModeDefineOwnProperty,
    };

public:
	typedef JSCell Base;

	JS_EXPORT_PRIVATE static void visitChildren(JSCell*, SlotVisitor&);
	JS_EXPORT_PRIVATE static void copyBackingStore(JSCell*, CopyVisitor&, CopyToken);

	JS_EXPORT_PRIVATE static String className(const JSObject*);
    JS_EXPORT_PRIVATE static String calculatedClassName(JSObject*);

    // This function is what Object.prototype.toString() will use to get the name of
    // an object when using Symbol.toStringTag fails. For the most part there is no
    // difference between this and className(). The main use case is for new JS language
    // objects to set the default tag to "Object".
    JS_EXPORT_PRIVATE static String toStringName(const JSObject*, ExecState*);

    // This is the fully virtual [[GetPrototypeOf]] internal function defined
    // in the ECMAScript 6 specification. Use this when doing a [[GetPrototypeOf]]
    // operation as dictated in the specification.
    JSValue getPrototype(VM&, ExecState*);
    JS_EXPORT_PRIVATE static JSValue getPrototype(JSObject*, ExecState*);
    // This gets the prototype directly off of the structure. This does not do
    // dynamic dispatch on the getPrototype method table method. It is not valid
    // to use this when performing a [[GetPrototypeOf]] operation in the specification.
    // It is valid to use though when you know that you want to directly get it
    // without consulting the method table. This is akin to getting the [[Prototype]]
    // internal field directly as described in the specification.
    JSValue getPrototypeDirect() const;

    // This sets the prototype without checking for cycles and without
    // doing dynamic dispatch on [[SetPrototypeOf]] operation in the specification.
    // It is not valid to use this when performing a [[SetPrototypeOf]] operation in
    // the specification. It is valid to use though when you know that you want to directly
    // set it without consulting the method table and when you definitely won't
    // introduce a cycle in the prototype chain. This is akin to setting the
    // [[Prototype]] internal field directly as described in the specification.
    JS_EXPORT_PRIVATE void setPrototypeDirect(VM&, JSValue prototype);
private:
    // This is OrdinarySetPrototypeOf in the specification. Section 9.1.2.1
    // https://tc39.github.io/ecma262/#sec-ordinarysetprototypeof
    JS_EXPORT_PRIVATE bool setPrototypeWithCycleCheck(VM&, ExecState*, JSValue prototype, bool shouldThrowIfCantSet);
public:
    // This is the fully virtual [[SetPrototypeOf]] internal function defined
    // in the ECMAScript 6 specification. Use this when doing a [[SetPrototypeOf]]
    // operation as dictated in the specification.
    bool setPrototype(VM&, ExecState*, JSValue prototype, bool shouldThrowIfCantSet = false);
	JS_EXPORT_PRIVATE static bool setPrototype(JSObject*, ExecState*, JSValue prototype, bool shouldThrowIfCantSet);

    inline bool mayInterceptIndexedAccesses() const;
    inline bool mayInterceptIndexedAccesses(VM&) const;

	JSValue get(JSGlobalData&, ExecState*, PropertyName) const;
	JSValue get(JSGlobalData&, ExecState*, unsigned propertyName) const;
    JSValue get(JSGlobalData&, ExecState*, uint64_t propertyName) const;

    template<bool maybeIndex = true> ALWAYS_INLINE bool getPropertySlot(JSGlobalData&, ExecState*, const PropertyName&, PropertySlot&);
	bool getPropertySlot(JSGlobalData&, ExecState*, unsigned propertyName, PropertySlot&);
    bool getPropertySlot(JSGlobalData&, ExecState*, uint64_t propertyName, PropertySlot&);
	//template<typename CallbackWhenNoException> typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type getPropertySlot(ExecState*, PropertyName, CallbackWhenNoException) const;
	//template<typename CallbackWhenNoException> typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type getPropertySlot(ExecState*, PropertyName, PropertySlot&, CallbackWhenNoException) const;

private:
    static bool getOwnPropertySlotImpl(JSObject*, ExecState*, const PropertyName&, PropertySlot&);
public:
    JS_EXPORT_PRIVATE_IF_ASSERT_ENABLED static bool getOwnPropertySlot(JSObject*, ExecState*, PropertyName, PropertySlot&);

	JS_EXPORT_PRIVATE static bool getOwnPropertySlotByIndex(JSObject*, ExecState*, unsigned propertyName, PropertySlot&);
	bool getOwnPropertySlotInline(JSGlobalData&, ExecState*, const PropertyName&, PropertySlot&);
	bool getOwnPropertyDescriptor(ExecState*, PropertyName, PropertyDescriptor&);

    static bool getPrivateFieldSlot(JSObject*, JSGlobalData&, PropertyName, PropertySlot&);
    inline bool hasPrivateField(JSGlobalData&, ExecState*, PropertyName);
    inline bool getPrivateField(JSGlobalData&, ExecState*, PropertyName, PropertySlot&);
    inline void setPrivateField(JSGlobalData&, ExecState*, PropertyName, JSValue, PutPropertySlot&);
    inline void definePrivateField(JSGlobalData&, ExecState*, PropertyName, JSValue, PutPropertySlot&);
    inline bool hasPrivateBrand(JSGlobalData&, ExecState*, JSValue brand);
    inline void checkPrivateBrand(JSGlobalData&, ExecState*, JSValue brand);
	inline void setPrivateBrand(JSGlobalData&, ExecState*, JSValue brand);

	unsigned getArrayLength() const
	{
		if (!hasIndexedProperties(structure()->indexingType()))
			return 0;
		return m_butterfly->publicLength();
	}

	unsigned getVectorLength()
	{
		if (!hasIndexedProperties(structure()->indexingType()))
			return 0;
		return m_butterfly->vectorLength();
	}

    static bool putInlineForJSObject(JSCell*, ExecState*, VM&, const PropertyName&, JSValue, PutPropertySlot&);

	JS_EXPORT_PRIVATE static bool put(JSCell*, ExecState*, PropertyName, JSValue, PutPropertySlot&);
    static bool mightBeSpecialProperty(VM&, JSType, UniquedStringImpl*);
	JS_EXPORT_PRIVATE NEVER_INLINE static bool definePropertyOnReceiver(ExecState*, PropertyName, JSValue, PutPropertySlot&);
	// putByIndex assumes that the receiver is this JSCell object.
	JS_EXPORT_PRIVATE static bool putByIndex(JSCell*, ExecState*, unsigned propertyName, JSValue, bool shouldThrow);

    // This performs the ECMAScript Set() operation.
	ALWAYS_INLINE bool putByIndexInline(ExecState* exec, VM& vm, unsigned propertyName, JSValue value, bool shouldThrow)
    {
        if (canSetIndexQuickly(propertyName, value)) {
            setIndexQuickly(vm, propertyName, value);
            return true;
        }
        return methodTable()->putByIndex(this, exec, propertyName, value, shouldThrow);
    }

    ALWAYS_INLINE bool putByIndexInline(ExecState* exec, VM& vm, uint64_t propertyName, JSValue value, bool shouldThrow)
    {
        if (LIKELY(propertyName <= MAX_ARRAY_INDEX))
            return putByIndexInline(exec, vm, static_cast<uint32_t>(propertyName), value, shouldThrow);

        ASSERT(propertyName <= maxSafeInteger());
        PutPropertySlot slot(this, shouldThrow);
        return methodTable(vm)->put(this, exec, Identifier::from(vm, propertyName), value, slot);
    }

    // This is similar to the putDirect* methods:
    //  - the prototype chain is not consulted
    //  - accessors are not called.
    //  - it will ignore extensibility and read-only properties if PutDirectIndexLikePutDirect is passed as the mode (the default).
    // This method creates a property with attributes writable, enumerable and configurable all set to true if attributes is zero,
    // otherwise, it creates a property with the provided attributes. Semantically, this is performing defineOwnProperty.
    bool putDirectIndex(ExecState* exec, unsigned propertyName, JSValue value, unsigned attributes, PutDirectIndexMode mode)
    {
        ASSERT(!value.isCustomGetterSetter());
        class CanSetIndexQuicklyForPutDirect
        {
        public:
        	bool operator()(JSObject* object, unsigned propertyName) const {
				switch (object->structure()->indexingType()) {
				case ALL_BLANK_INDEXING_TYPES:
				case ALL_UNDECIDED_INDEXING_TYPES:
					return false;
				case ALL_INT32_INDEXING_TYPES:
				case ALL_DOUBLE_INDEXING_TYPES:
				case ALL_CONTIGUOUS_INDEXING_TYPES:
				case ALL_ARRAY_STORAGE_INDEXING_TYPES:
					return propertyName < object->m_butterfly->vectorLength();
				default:
					RELEASE_ASSERT_NOT_REACHED();
					return false;
				}
        	}
        };
        CanSetIndexQuicklyForPutDirect canSetIndexQuicklyForPutDirect;

        if (!attributes && canSetIndexQuicklyForPutDirect(this, propertyName)) {
            setIndexQuickly(exec->vm(), propertyName, value);
            return true;
        }
        return putDirectIndexSlowOrBeyondVectorLength(exec, propertyName, value, attributes, mode);
    }
    // This is semantically equivalent to performing defineOwnProperty(propertyName, {configurable:true, writable:true, enumerable:true, value:value}).
	bool putDirectIndex(ExecState* exec, unsigned propertyName, JSValue value)
    {
        return putDirectIndex(exec, propertyName, value, 0, PutDirectIndexLikePutDirect);
    }

    ALWAYS_INLINE bool putDirectIndex(ExecState* exec, uint64_t propertyName, JSValue value, unsigned attributes, PutDirectIndexMode mode)
    {
        if (LIKELY(propertyName <= MAX_ARRAY_INDEX))
            return putDirectIndex(exec, static_cast<uint32_t>(propertyName), value, attributes, mode);
        VM& vm = exec->vm();
        return putDirect(vm, Identifier::from(vm, propertyName), value, attributes);
    }

	// A generally non-throwing version of putDirect and putDirectIndex.
	// However, it's only guaranteed to not throw based on what the receiver is.
	// For example, if the receiver is a ProxyObject, this is not guaranteed, since
	// it may call into arbitrary JS code. It's the responsibility of the user of
	// this API to ensure that the receiver object is a well known type if they
	// want to ensure that this won't throw an exception.
	bool putDirectMayBeIndex(ExecState*, VM&, PropertyName, JSValue);

    bool hasIndexingHeader() const
    {
        return structure()->hasIndexingHeader(this);
    }

    bool canGetIndexQuicklyForTypedArray(unsigned) const;
    JSValue getIndexQuicklyForTypedArray(unsigned) const;

    ALWAYS_INLINE bool canGetIndexQuickly(unsigned i) const
	{
		switch (structure()->indexingType()) {
		case ALL_BLANK_INDEXING_TYPES:
            return canGetIndexQuicklyForTypedArray(i);
		case ALL_UNDECIDED_INDEXING_TYPES:
			return false;
		case ALL_INT32_INDEXING_TYPES:
		case ALL_CONTIGUOUS_INDEXING_TYPES:
			return i < m_butterfly->vectorLength() && m_butterfly->contiguous()[i];
		case ALL_DOUBLE_INDEXING_TYPES: {
			if (i >= m_butterfly->vectorLength())
				return false;
			double value = m_butterfly->contiguousDouble()[i];
			if (value != value)
				return false;
			return true;
		}
		case ALL_ARRAY_STORAGE_INDEXING_TYPES:
			return i < m_butterfly->arrayStorage()->vectorLength() && m_butterfly->arrayStorage()->m_vector[i];
		default:
			ASSERT_NOT_REACHED();
			return false;
		}
	}

    bool canGetIndexQuickly(uint64_t i) const
    {
        ASSERT(i <= maxSafeInteger());
        if (LIKELY(i <= MAX_ARRAY_INDEX))
            return canGetIndexQuickly(static_cast<uint32_t>(i));
        return false;
    }

    ALWAYS_INLINE JSValue getIndexQuickly(unsigned i)
    {
        switch (structure()->indexingType()) {
        case ALL_INT32_INDEXING_TYPES:
            return jsNumber(m_butterfly->contiguous()[i].get().asInt32());
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            return m_butterfly->contiguous()[i].get();
        case ALL_DOUBLE_INDEXING_TYPES:
        	return JSValue(JSValue::EncodeAsDouble, m_butterfly->contiguousDouble()[i]);
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return m_butterfly->arrayStorage()->m_vector[i].get();
        case ALL_BLANK_INDEXING_TYPES:
            return getIndexQuicklyForTypedArray(i);
        default:
            ASSERT_NOT_REACHED();
            return JSValue();
        }
    }

    ALWAYS_INLINE JSValue tryGetIndexQuickly(unsigned i) const
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
            if (canGetIndexQuicklyForTypedArray(i))
                return getIndexQuicklyForTypedArray(i);
            break;
        case ALL_UNDECIDED_INDEXING_TYPES:
            break;
        case ALL_INT32_INDEXING_TYPES:
            if (i < m_butterfly->publicLength()) {
                JSValue result = m_butterfly->contiguous()[i].get();
                ASSERT(result.isInt32() || !result);
                return result;
            }
            break;
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            if (i < m_butterfly->publicLength())
                return m_butterfly->contiguous()[i].get();
            break;
        case ALL_DOUBLE_INDEXING_TYPES: {
            if (i >= m_butterfly->publicLength())
                break;
            double result = m_butterfly->contiguousDouble()[i];
            if (result != result)
                break;
            return JSValue(JSValue::EncodeAsDouble, result);
        }
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            if (i < m_butterfly->arrayStorage()->vectorLength())
                return m_butterfly->arrayStorage()->m_vector[i].get();
            break;
        default:
            ASSERT_NOT_REACHED();
            break;
        }
        return JSValue();
    }

    JSValue tryGetIndexQuickly(uint64_t i) const
    {
        ASSERT(i <= maxSafeInteger());
        if (LIKELY(i <= MAX_ARRAY_INDEX))
            return tryGetIndexQuickly(static_cast<uint32_t>(i));
        return JSValue();
    }

    JSValue getDirectIndex(VM& vm, ExecState* exec, unsigned i)
    {
        if (JSValue result = tryGetIndexQuickly(i))
            return result;
        PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
        if (methodTable()->getOwnPropertySlotByIndex(this, exec, i, slot))
            return slot.getValue(vm, exec, i);
        return JSValue();
    }

    JSValue getIndex(VM& vm, ExecState* exec, unsigned i) const
    {
        if (JSValue result = tryGetIndexQuickly(i))
            return result;
        return get(vm, exec, i);
    }

    bool canSetIndexQuicklyForTypedArray(unsigned, JSValue) const;
    void setIndexQuicklyForTypedArray(unsigned, JSValue);

    ALWAYS_INLINE bool canSetIndexQuickly(unsigned i, JSValue value)
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
            return canSetIndexQuicklyForTypedArray(i, value);
        case ALL_UNDECIDED_INDEXING_TYPES:
            return false;
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
        case NonArrayWithArrayStorage:
        case ArrayWithArrayStorage:
            return i < m_butterfly->vectorLength();
        case NonArrayWithSlowPutArrayStorage:
        case ArrayWithSlowPutArrayStorage:
            return i < m_butterfly->arrayStorage()->vectorLength()
                && !!m_butterfly->arrayStorage()->m_vector[i];
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    }

    ALWAYS_INLINE void setIndexQuickly(JSGlobalData& vm, unsigned i, JSValue v)
    {
        switch (structure()->indexingType()) {
        case ALL_INT32_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isInt32()) {
                convertInt32ToDoubleOrContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            // Fall through to contiguous case.
        }
        case ALL_CONTIGUOUS_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->vectorLength());
            m_butterfly->contiguous()[i].set(vm, this, v);
            if (i >= m_butterfly->publicLength())
                m_butterfly->setPublicLength(i + 1);
            break;
        }
        case ALL_DOUBLE_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isNumber()) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            double value = v.asNumber();
            if (value != value) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            m_butterfly->contiguousDouble()[i] = value;
            if (i >= m_butterfly->publicLength())
                m_butterfly->setPublicLength(i + 1);
            break;
        }
        case ALL_ARRAY_STORAGE_INDEXING_TYPES: {
            ArrayStorage* storage = m_butterfly->arrayStorage();
            WriteBarrier<Unknown>& x = storage->m_vector[i];
            JSValue old = x.get();
            x.set(vm, this, v);
            if (!old) {
                ++storage->m_numValuesInVector;
                if (i >= storage->length())
                    storage->setLength(i + 1);
            }
            break;
        }
        case ALL_BLANK_INDEXING_TYPES:
            setIndexQuicklyForTypedArray(i, v);
            break;
        default:
            ASSERT_NOT_REACHED();
        }
    }

    void initializeIndex(JSGlobalData& vm, unsigned i, JSValue v)
    {
        initializeIndex(vm, i, v, structure()->indexingType());
    }

    void initializeIndex(JSGlobalData& vm, unsigned i, JSValue v, IndexingType indexingType)
    {
        switch (indexingType) {
        case ALL_UNDECIDED_INDEXING_TYPES: {
            setIndexQuicklyToUndecided(vm, i, v);
            break;
        }
        case ALL_INT32_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->publicLength());
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isInt32()) {
                convertInt32ToDoubleOrContiguousWhilePerformingSetIndex(vm, i, v);
                break;
            }
            // Fall through.
        }
        case ALL_CONTIGUOUS_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->publicLength());
            ASSERT(i < m_butterfly->vectorLength());
            m_butterfly->contiguous()[i].set(vm, this, v);
            break;
        }
        case ALL_DOUBLE_INDEXING_TYPES: {
            ASSERT(i < m_butterfly->publicLength());
            ASSERT(i < m_butterfly->vectorLength());
            if (!v.isNumber()) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            double value = v.asNumber();
            if (value != value) {
                convertDoubleToContiguousWhilePerformingSetIndex(vm, i, v);
                return;
            }
            m_butterfly->contiguousDouble()[i] = value;
            break;
        }
        case ALL_ARRAY_STORAGE_INDEXING_TYPES: {
            ArrayStorage* storage = m_butterfly->arrayStorage();
            ASSERT(i < storage->length());
            ASSERT(i < storage->m_numValuesInVector);
            storage->m_vector[i].set(vm, this, v);
            break;
        }
        default:
            ASSERT_NOT_REACHED();
        }
    }

    bool hasSparseMap()
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
        case ALL_UNDECIDED_INDEXING_TYPES:
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            return false;
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return !!m_butterfly->arrayStorage()->m_sparseMap;
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    }

    bool inSparseIndexingMode()
    {
        switch (structure()->indexingType()) {
        case ALL_BLANK_INDEXING_TYPES:
        case ALL_UNDECIDED_INDEXING_TYPES:
        case ALL_INT32_INDEXING_TYPES:
        case ALL_DOUBLE_INDEXING_TYPES:
        case ALL_CONTIGUOUS_INDEXING_TYPES:
            return false;
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return m_butterfly->arrayStorage()->inSparseMode();
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    }

    void enterDictionaryIndexingMode(JSGlobalData&);

	// putDirect is effectively an unchecked vesion of 'defineOwnProperty':
	//  - the prototype chain is not consulted
	//  - accessors are not called.
	//  - attributes will be respected (after the call the property will exist with the given attributes)
    //  - the property name is assumed to not be an index.
	bool putDirect(JSGlobalData&, PropertyName, JSValue, unsigned attributes = 0);
	bool putDirect(VM&, PropertyName, JSValue, unsigned attributes, PutPropertySlot&);
	bool putDirect(JSGlobalData&, PropertyName, JSValue, PutPropertySlot&);
	bool putDirectWithoutTransition(JSGlobalData&, PropertyName, JSValue, unsigned attributes = 0);
	bool putDirectNonIndexAccessor(VM&, PropertyName, GetterSetter*, unsigned attributes);
	void putDirectNonIndexAccessorWithoutTransition(VM&, PropertyName, GetterSetter*, unsigned attributes);
	bool putDirectAccessor(ExecState*, PropertyName, GetterSetter*, unsigned attributes);
	JS_EXPORT_PRIVATE bool putDirectCustomAccessor(VM&, PropertyName, JSValue, unsigned attributes);
	void putDirectCustomGetterSetterWithoutTransition(VM&, PropertyName, JSValue, unsigned attributes);

	bool putGetter(ExecState*, PropertyName, JSValue, unsigned attributes);
	bool putSetter(ExecState*, PropertyName, JSValue, unsigned attributes);

	JS_EXPORT_PRIVATE bool hasProperty(ExecState*, PropertyName) const;
	JS_EXPORT_PRIVATE bool hasProperty(ExecState*, unsigned propertyName) const;
    bool hasProperty(ExecState*, uint64_t propertyName) const;
	bool hasEnumerableProperty(JSGlobalData&, ExecState*, PropertyName) const;
	bool hasEnumerableProperty(JSGlobalData&, ExecState*, unsigned propertyName) const;
	bool hasOwnProperty(ExecState*, PropertyName, PropertySlot&) const;
	bool hasOwnProperty(ExecState*, PropertyName) const;
    bool hasOwnProperty(ExecState*, unsigned) const;

	JS_EXPORT_PRIVATE static bool deleteProperty(JSCell*, ExecState*, PropertyName);
	JS_EXPORT_PRIVATE static bool deletePropertyByIndex(JSCell*, ExecState*, unsigned propertyName);
	bool deleteProperty(ExecState*, PropertyName);
	bool deleteProperty(ExecState*, uint32_t propertyName);
	bool deleteProperty(ExecState*, uint64_t propertyName);

	JSValue ordinaryToPrimitive(ExecState*, PreferredPrimitiveType) const;

	JS_EXPORT_PRIVATE bool hasInstance(ExecState*, JSValue value, JSValue hasInstanceValue);
	bool hasInstance(ExecState*, JSValue);
    static bool defaultHasInstance(ExecState*, JSValue, JSValue prototypeProperty);

    static constexpr unsigned maximumPrototypeChainDepth = 40000;
    JS_EXPORT_PRIVATE void getPropertyNames(ExecState*, PropertyNameArray&, DontEnumPropertiesMode);
    JS_EXPORT_PRIVATE static void getOwnPropertyNames(JSObject*, ExecState*, PropertyNameArray&, DontEnumPropertiesMode);
    JS_EXPORT_PRIVATE static void getOwnSpecialPropertyNames(JSObject*, ExecState*, PropertyNameArray&, DontEnumPropertiesMode);
    JS_EXPORT_PRIVATE void getOwnIndexedPropertyNames(ExecState*, PropertyNameArray&, DontEnumPropertiesMode);
    JS_EXPORT_PRIVATE void getOwnNonIndexPropertyNames(ExecState*, PropertyNameArray&, DontEnumPropertiesMode);
    void getNonReifiedStaticPropertyNames(VM&, PropertyNameArray&, DontEnumPropertiesMode);

    JS_EXPORT_PRIVATE static uint32_t getEnumerableLength(ExecState*, JSObject*);

	JSValue toPrimitive(ExecState*, PreferredPrimitiveType = NoPreference) const;
	JS_EXPORT_PRIVATE double toNumber(ExecState*) const;
	JS_EXPORT_PRIVATE JSString* toString(ExecState*) const;

	// This get function only looks at the property map.
	ALWAYS_INLINE JSValue getDirect(JSGlobalData& vm, PropertyName propertyName) const
	{
		PropertyOffset offset = structure()->get(vm, propertyName);
		checkOffset(offset, structure()->inlineCapacity());
		return offset != invalidOffset ? getDirect(offset) : JSValue();
	}

	ALWAYS_INLINE JSValue getDirect(VM& vm, PropertyName propertyName, unsigned& attributes) const
    {
        PropertyOffset offset = structure()->get(vm, propertyName, attributes);
        checkOffset(offset, structure()->inlineCapacity());
        return offset != invalidOffset ? getDirect(offset) : JSValue();
    }

	ALWAYS_INLINE PropertyOffset getDirectOffset(JSGlobalData& vm, PropertyName propertyName)
	{
		PropertyOffset offset = structure()->get(vm, propertyName);
		checkOffset(offset, structure()->inlineCapacity());
		return offset;
	}

	ALWAYS_INLINE PropertyOffset getDirectOffset(VM& vm, PropertyName propertyName, unsigned& attributes)
    {
        PropertyOffset offset = structure()->get(vm, propertyName, attributes);
        checkOffset(offset, structure()->inlineCapacity());
        return offset;
    }

	ALWAYS_INLINE bool hasInlineStorage() const { return structure()->hasInlineStorage(); }
	ALWAYS_INLINE ConstPropertyStorage inlineStorageUnsafe() const
	{
		return bitwise_cast<ConstPropertyStorage>(this + 1);
	}
	ALWAYS_INLINE PropertyStorage inlineStorageUnsafe()
	{
		return bitwise_cast<PropertyStorage>(this + 1);
	}
	ALWAYS_INLINE ConstPropertyStorage inlineStorage() const
	{
		ASSERT(hasInlineStorage());
		return inlineStorageUnsafe();
	}
	ALWAYS_INLINE PropertyStorage inlineStorage()
	{
		ASSERT(hasInlineStorage());
		return inlineStorageUnsafe();
	}

    const Butterfly* butterfly() const { return m_butterfly; }
    Butterfly* butterfly() { return m_butterfly; }

    ALWAYS_INLINE ConstPropertyStorage outOfLineStorage() const { return m_butterfly->propertyStorage(); }
    ALWAYS_INLINE PropertyStorage outOfLineStorage() { return m_butterfly->propertyStorage(); }

    ALWAYS_INLINE const WriteBarrierBase<Unknown>* locationForOffset(PropertyOffset offset) const
	{
		if (isInlineOffset(offset))
			return &inlineStorage()[offsetInInlineStorage(offset)];
		return &outOfLineStorage()[offsetInOutOfLineStorage(offset)];
	}

    ALWAYS_INLINE WriteBarrierBase<Unknown>* locationForOffset(PropertyOffset offset)
	{
		if (isInlineOffset(offset))
			return &inlineStorage()[offsetInInlineStorage(offset)];
		return &outOfLineStorage()[offsetInOutOfLineStorage(offset)];
	}

    PropertyOffset offsetForLocation(WriteBarrierBase<Unknown>* location) const
    {
        PropertyOffset result;
        size_t offsetInInlineStorage = location - inlineStorageUnsafe();
        if (offsetInInlineStorage < static_cast<size_t>(firstOutOfLineOffset))
            result = offsetInInlineStorage;
        else
            result = outOfLineStorage() - location + (firstOutOfLineOffset - 1);
        validateOffset(result, structure()->inlineCapacity());
        return result;
    }

	void transitionTo(JSGlobalData&, Structure*);

	ALWAYS_INLINE bool hasCustomProperties() { return structure()->didTransition(); }

    bool hasNonReifiedStaticProperties(VM& vm)
    {
        return TypeInfo::hasStaticPropertyTable(structure(vm)->typeInfo().inlineTypeFlags()) && !structure(vm)->staticPropertiesReified();
    }

	// putOwnDataProperty has 'put' like semantics, however this method:
	//  - assumes the object contains no own getter/setter properties.
	//  - provides no special handling for __proto__
	//  - does not walk the prototype chain (to check for accessors or non-writable properties).
	// This is used by JSLexicalEnvironment.
	bool putOwnDataProperty(JSGlobalData&, PropertyName, JSValue, PutPropertySlot&);
	bool putOwnDataPropertyMayBeIndex(ExecState*, PropertyName, JSValue, PutPropertySlot&);
private:
    void validatePutOwnDataProperty(VM&, PropertyName, JSValue);
public:

	// Fast access to known property offsets.
	ALWAYS_INLINE JSValue getDirect(PropertyOffset offset) const { return locationForOffset(offset)->get(); }
	void putDirect(JSGlobalData& vm, PropertyOffset offset, JSValue value) { locationForOffset(offset)->set(vm, this, value); }
	void putDirectUndefined(PropertyOffset offset) { locationForOffset(offset)->setUndefined(); }

	JS_EXPORT_PRIVATE bool putDirectNativeIntrinsicGetter(VM&, JSGlobalObject*, const Identifier&, NativeFunction, Intrinsic, unsigned attributes);
	JS_EXPORT_PRIVATE void putDirectNativeIntrinsicGetterWithoutTransition(VM&, JSGlobalObject*, const Identifier&, NativeFunction, Intrinsic, unsigned attributes);
	bool putDirectNativeFunction(VM&, JSGlobalObject*, PropertyName, unsigned functionLength, NativeFunction, Intrinsic, unsigned attributes);
	bool putDirectNativeFunction(VM&, JSGlobalObject*, const Identifier&, unsigned functionLength, NativeFunction, Intrinsic, unsigned attributes);
	bool putDirectNativeFunctionWithoutTransition(VM&, JSGlobalObject*, PropertyName, unsigned functionLength, NativeFunction, Intrinsic, unsigned attributes);
	bool putDirectNativeFunctionWithoutTransition(VM&, JSGlobalObject*, const Identifier&, unsigned functionLength, NativeFunction, Intrinsic, unsigned attributes);

    JSFunction* putDirectBuiltinFunction(VM&, JSGlobalObject*, PropertyName, FunctionExecutable*, unsigned attributes);
    JSFunction* putDirectBuiltinFunction(VM&, JSGlobalObject*, const Identifier&, FunctionExecutable*, unsigned attributes);
    JSFunction* putDirectBuiltinFunctionWithoutTransition(VM&, JSGlobalObject*, const Identifier&, FunctionExecutable*, unsigned attributes);

	JS_EXPORT_PRIVATE static bool defineOwnProperty(JSObject*, ExecState*, PropertyName, const PropertyDescriptor&, bool shouldThrow);
	bool createDataProperty(ExecState*, PropertyName, JSValue, bool shouldThrow);

    bool isEnvironmentRecord() const;
	bool isGlobalObject() const;
    bool isJSLexicalEnvironment() const;
    bool isGlobalLexicalEnvironment() const;
    bool isStrictEvalActivation() const;
    bool isWithScope() const;

    bool isErrorInstance() const;

	void seal(JSGlobalData&);
	void freeze(JSGlobalData&);
    JS_EXPORT_PRIVATE static bool preventExtensions(JSObject*, ExecState*);
    JS_EXPORT_PRIVATE static bool isExtensible(JSObject*, ExecState*);
	bool isSealed(JSGlobalData& vm) { return structure()->isSealed(vm); }
	bool isFrozen(JSGlobalData& vm) { return structure()->isFrozen(vm); }

    bool anyObjectInChainMayInterceptIndexedAccesses(VM&) const;
    bool needsSlowPutIndexing(VM&) const;

private:
    TransitionKind suggestedArrayStorageTransition(VM&) const;
public:
    // You should only call isStructureExtensible() when:
    // - Performing this check in a way that isn't described in the specification
    //   as calling the virtual [[IsExtensible]] trap.
    // - When you're guaranteed that object->methodTable()->isExtensible isn't
    //   overridden.
    ALWAYS_INLINE bool isStructureExtensible() { return structure()->isStructureExtensible(); }
    // You should call this when performing [[IsExtensible]] trap in a place
    // that is described in the specification. This performs the fully virtual
    // [[IsExtensible]] trap.
    bool isExtensible(ExecState*);
	bool indexingShouldBeSparse()
	{
		return !isStructureExtensible()
			|| structure()->typeInfo().interceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero();
	}

	bool staticPropertiesReified() { return structure()->staticPropertiesReified(); }
    void reifyAllStaticProperties(JSGlobalData&);

    JS_EXPORT_PRIVATE Butterfly* allocateMoreOutOfLineStorage(JSGlobalData&, DeferGC&, size_t oldSize, size_t newSize);

    void setButterfly(VM&, Butterfly*); // Always change the butterfly before changing the structure!
	void setStructure(JSGlobalData&, Structure*);

    JS_EXPORT_PRIVATE void convertToDictionary(VM&);
    JS_EXPORT_PRIVATE void convertToUncacheableDictionary(VM&);

	void flattenDictionaryObject(JSGlobalData& vm)
	{
		structure()->flattenDictionaryStructure(vm, this);
	}

    void shiftButterflyAfterFlattening(JSGlobalData&, size_t outOfLineCapacityBefore, size_t outOfLineCapacityAfter);

	ALWAYS_INLINE JSGlobalObject* globalObject() const
	{
		ASSERT(structure()->globalObject());
		ASSERT(!isGlobalObject() || ((JSObject*)structure()->globalObject()) == this);
		return structure()->globalObject();
	}

	ALWAYS_INLINE JSGlobalObject* globalObject(VM& vm) const
	{
		return globalObject();
	}

    // Does everything possible to return the global object. If it encounters an object
    // that does not have a global object, it returns 0 instead (for example
    // JSNotAnObject).

    void switchToSlowPutArrayStorage(JSGlobalData&);

    // The receiver is the prototype in this case. The following:
    //
    // asObject(foo->structure()->storedPrototype())->attemptToInterceptPutByIndexOnHoleForPrototype(...)
    //
    // is equivalent to:
    //
    // foo->attemptToInterceptPutByIndexOnHole(...);
    bool attemptToInterceptPutByIndexOnHoleForPrototype(ExecState*, JSGlobalData&, JSValue thisValue, unsigned propertyName, JSValue, bool shouldThrow, bool& putResult);

    // Returns 0 if int32 storage cannot be created - either because
    // indexing should be sparse, we're having a bad time, or because
    // we already have a more general form of storage (double,
    // contiguous, array storage).
    ContiguousJSValues ensureInt32(JSGlobalData& vm)
    {
        if (LIKELY(hasInt32(structure()->indexingType())))
            return m_butterfly->contiguousInt32();

        return ensureInt32Slow(vm);
    }

    // Returns 0 if double storage cannot be created - either because
    // indexing should be sparse, we're having a bad time, or because
    // we already have a more general form of storage (contiguous,
    // or array storage).
    ContiguousDoubles ensureDouble(JSGlobalData& vm)
    {
        if (LIKELY(hasDouble(structure()->indexingType())))
            return m_butterfly->contiguousDouble();

        return ensureDoubleSlow(vm);
    }

    // Returns 0 if contiguous storage cannot be created - either because
    // indexing should be sparse or because we're having a bad time.
    ContiguousJSValues ensureContiguous(JSGlobalData& vm)
    {
        if (LIKELY(hasContiguous(structure()->indexingType())))
            return m_butterfly->contiguous();

        return ensureContiguousSlow(vm);
    }

    // Ensure that the object is in a mode where it has array storage. Use
    // this if you're about to perform actions that would have required the
    // object to be converted to have array storage, if it didn't have it
    // already.
    ArrayStorage* ensureArrayStorage(JSGlobalData& vm)
    {
        if (LIKELY(hasAnyArrayStorage(structure()->indexingType())))
            return m_butterfly->arrayStorage();

        return ensureArrayStorageSlow(vm);
    }

	static size_t offsetOfInlineStorage();

    static ptrdiff_t butterflyOffset()
    {
        return OBJECT_OFFSETOF(JSObject, m_butterfly);
    }

    void* butterflyAddress()
    {
        return &m_butterfly;
    }

    JSValue getMethod(ExecState* exec, CallData& callData, const Identifier& ident, const String& errorMessage);

    bool canPerformFastPutInline(JSGlobalData&, const PropertyName&);
    bool canPerformFastPutInlineExcludingProto(JSGlobalData&);

    Optional<Structure::PropertyHashEntry> findPropertyHashEntry(VM&, const PropertyName&) const;

    DECLARE_INFO;

protected:
	ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		ASSERT(inherits(info()));
		ASSERT(!structure()->outOfLineCapacity());
		ASSERT(structure()->isEmpty());
		ASSERT(getPrototypeDirect().isNull() || Heap::heap(this) == Heap::heap(getPrototypeDirect()));
		ASSERT(structure()->isObject());
		ASSERT(classInfo());
	}

	static ALWAYS_INLINE Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue prototype)
	{
		return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
	}

	// To instantiate objects you likely want JSFinalObject, below.
	// To create derived types you likely want JSNonFinalObject, below.
	JSObject(JSGlobalData&, Structure*, Butterfly* = nullptr);

    void visitButterfly(SlotVisitor&, Butterfly*);
    void copyButterfly(CopyVisitor&, Butterfly*);

    // Call this if you know that the object is in a mode where it has array
    // storage. This will assert otherwise.
    ALWAYS_INLINE ArrayStorage* arrayStorage()
    {
        ASSERT(hasAnyArrayStorage(structure()->indexingType()));
        return m_butterfly->arrayStorage();
    }

    // Call this if you want to predicate some actions on whether or not the
    // object is in a mode where it has array storage.
    ArrayStorage* arrayStorageOrNull()
    {
        switch (structure()->indexingType()) {
        case ALL_ARRAY_STORAGE_INDEXING_TYPES:
            return m_butterfly->arrayStorage();

        default:
            return 0;
        }
    }

    size_t butterflyTotalSize();
    size_t butterflyPreCapacity();

    Butterfly* createInitialUndecided(JSGlobalData&, unsigned length);
    ContiguousJSValues createInitialInt32(JSGlobalData&, unsigned length);
    ContiguousDoubles createInitialDouble(JSGlobalData&, unsigned length);
    ContiguousJSValues createInitialContiguous(JSGlobalData&, unsigned length);

    void convertUndecidedForValue(JSGlobalData&, JSValue);
    void createInitialForValueAndSet(VM&, unsigned index, JSValue);
    void convertInt32ForValue(JSGlobalData&, JSValue);

    static Butterfly* createArrayStorageButterfly(JSGlobalData&, JSCell* intendedOwner, Structure*, unsigned length, unsigned vectorLength, Butterfly* oldButterfly = 0);
    ArrayStorage* createArrayStorage(JSGlobalData&, unsigned length, unsigned vectorLength);
    ArrayStorage* createInitialArrayStorage(JSGlobalData&);

    ContiguousJSValues convertUndecidedToInt32(JSGlobalData&);
    ContiguousDoubles convertUndecidedToDouble(JSGlobalData&);
    ContiguousJSValues convertUndecidedToContiguous(JSGlobalData&);
    ArrayStorage* convertUndecidedToArrayStorage(JSGlobalData&, TransitionKind);
    ArrayStorage* convertUndecidedToArrayStorage(JSGlobalData&);

    ContiguousDoubles convertInt32ToDouble(JSGlobalData&);
    ContiguousJSValues convertInt32ToContiguous(JSGlobalData&);
    ArrayStorage* convertInt32ToArrayStorage(JSGlobalData&, TransitionKind);
    ArrayStorage* convertInt32ToArrayStorage(JSGlobalData&);

    ContiguousJSValues convertDoubleToContiguous(JSGlobalData&);
    ArrayStorage* convertDoubleToArrayStorage(JSGlobalData&, TransitionKind);
    ArrayStorage* convertDoubleToArrayStorage(JSGlobalData&);

	ArrayStorage* convertContiguousToArrayStorage(JSGlobalData&, TransitionKind);
	ArrayStorage* convertContiguousToArrayStorage(JSGlobalData&);

    ArrayStorage* ensureArrayStorageExistsAndEnterDictionaryIndexingMode(JSGlobalData&);

    bool defineOwnNonIndexProperty(ExecState*, PropertyName, const PropertyDescriptor&, bool throwException);

    template<IndexingType indexingShape>
    bool putByIndexBeyondVectorLengthWithoutAttributes(ExecState*, unsigned propertyName, JSValue);
    bool putByIndexBeyondVectorLengthWithArrayStorage(ExecState*, unsigned propertyName, JSValue, bool shouldThrow, ArrayStorage*);

    bool increaseVectorLength(ExecState*, JSGlobalData&, unsigned newLength);
    void deallocateSparseIndexMap();
    bool defineOwnIndexedProperty(ExecState*, unsigned, const PropertyDescriptor&, bool throwException);
    SparseArrayValueMap* allocateSparseIndexMap(JSGlobalData&);

    void notifyPresenceOfIndexedAccessors(JSGlobalData&);

    bool attemptToInterceptPutByIndexOnHole(ExecState*, VM&, unsigned index, JSValue, bool shouldThrow, bool& putResult);

    // Call this if you want setIndexQuickly to succeed and you're sure that
    // the array is contiguous.
    bool ensureLength(JSGlobalData& vm, unsigned length)
	{
    	RELEASE_ASSERT(length <= MAX_STORAGE_VECTOR_LENGTH);
		ASSERT(hasContiguous(structure()->indexingType()) || hasInt32(structure()->indexingType()) || hasDouble(structure()->indexingType()) || hasUndecided(structure()->indexingType()));

        if (m_butterfly->vectorLength() < length) {
            if (!ensureLengthSlow(vm, length))
                return false;
        }

        if (m_butterfly->publicLength() < length)
            m_butterfly->setPublicLength(length);
        return true;
    }

    // Call this if you want to shrink the butterfly backing store, and you're
    // sure that the array is contiguous.
    void reallocateAndShrinkButterfly(JSGlobalData&, unsigned length);

    template<IndexingType indexingShape>
    unsigned countElements(Butterfly*);

    // This is relevant to undecided, int32, double, and contiguous.
    unsigned countElements();

private:
	friend class LLIntOffsetsExtractor;

	// Nobody should ever ask any of these questions on something already known to be a JSObject.
	using JSCell::isAPIValueWrapper;
	using JSCell::isGetterSetter;
	void getObject();
	void getString(ExecState* exec);
	void isObject();
	void isString();

    Butterfly* createInitialIndexedStorage(JSGlobalData&, unsigned length, size_t elementSize);

    ArrayStorage* enterDictionaryIndexingModeWhenArrayStorageAlreadyExists(JSGlobalData&, ArrayStorage*);

	template<PutMode>
	ASCIILiteral putDirectInternal(JSGlobalData&, const PropertyName&, JSValue, unsigned attr, PutPropertySlot&);

    JS_EXPORT_PRIVATE NEVER_INLINE bool putInlineSlow(ExecState*, const PropertyName&, JSValue, PutPropertySlot&);
    JS_EXPORT_PRIVATE NEVER_INLINE bool putInlineFastReplacingStaticPropertyIfNeeded(ExecState*, PropertyName, JSValue, PutPropertySlot&);
	bool putInlineFast(ExecState*, PropertyName, JSValue, PutPropertySlot&);

    bool getNonIndexPropertySlot(VM&, ExecState*, const PropertyName&, PropertySlot&);
    ALWAYS_INLINE bool getOwnNonIndexPropertySlot(JSGlobalData&, Structure&, const PropertyName&, PropertySlot&);
	JS_EXPORT_PRIVATE void fillGetterPropertySlot(PropertySlot&, JSValue, unsigned, PropertyOffset);
	bool fillCustomGetterPropertySlot(PropertySlot&, JSValue, unsigned, Structure&);

    JS_EXPORT_PRIVATE bool getOwnStaticPropertySlot(VM&, const PropertyName&, PropertySlot&);

	bool putIndexedDescriptor(ExecState*, SparseArrayEntry*, const PropertyDescriptor&, PropertyDescriptor& old);

	bool putByIndexBeyondVectorLength(ExecState*, unsigned propertyName, JSValue, bool shouldThrow);
    bool putDirectIndexBeyondVectorLengthWithArrayStorage(ExecState*, unsigned propertyName, JSValue, unsigned attributes, PutDirectIndexMode, ArrayStorage*);
    JS_EXPORT_PRIVATE bool putDirectIndexSlowOrBeyondVectorLength(ExecState*, unsigned propertyName, JSValue, unsigned attributes, PutDirectIndexMode);

    unsigned getNewVectorLength(unsigned currentVectorLength, unsigned currentLength, unsigned desiredLength);
    unsigned getNewVectorLength(unsigned desiredLength);

    ArrayStorage* constructConvertedArrayStorageWithoutCopyingElements(JSGlobalData&, unsigned neededLength);

    JS_EXPORT_PRIVATE void setIndexQuicklyToUndecided(JSGlobalData&, unsigned index, JSValue);
    JS_EXPORT_PRIVATE void convertInt32ToDoubleOrContiguousWhilePerformingSetIndex(JSGlobalData&, unsigned index, JSValue);
    JS_EXPORT_PRIVATE void convertDoubleToContiguousWhilePerformingSetIndex(JSGlobalData&, unsigned index, JSValue);

    bool ensureLengthSlow(JSGlobalData&, unsigned length);

    ContiguousJSValues ensureInt32Slow(JSGlobalData&);
    ContiguousDoubles ensureDoubleSlow(JSGlobalData&);
    ContiguousJSValues ensureContiguousSlow(JSGlobalData&);
    ArrayStorage* ensureArrayStorageSlow(JSGlobalData&);

    PropertyOffset prepareToPutDirectWithoutTransition(VM&, DeferGC&, PropertyName, unsigned attributes);

protected:
    Butterfly* m_butterfly;
};

// JSNonFinalObject is a type of JSObject that has some internal storage,
// but also preserves some space in the collector cell for additional
// data members in derived types.
class JSNonFinalObject : public JSObject {
	friend class JSObject;

public:
	typedef JSObject Base;

	static ALWAYS_INLINE Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue prototype)
	{
		return Structure::create(vm, globalObject, prototype, TypeInfo(ObjectType, StructureFlags), info());
	}

protected:
    explicit ALWAYS_INLINE JSNonFinalObject(JSGlobalData& vm, Structure* structure, Butterfly* butterfly = 0)
        : JSObject(vm, structure, butterfly)
    {
    }

    ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		ASSERT(!this->structure()->hasInlineStorage());
		ASSERT(classInfo());
	}
};

class JSFinalObject;

// JSFinalObject is a type of JSObject that contains sufficent internal
// storage to fully make use of the colloctor cell containing it.
class JSFinalObject final : public JSObject {
	friend class JSObject;

public:
	typedef JSObject Base;
	static const unsigned StructureFlags = JSObject::StructureFlags;

    static size_t allocationSize(size_t inlineCapacity)
    {
        return sizeof(JSObject) + inlineCapacity * sizeof(WriteBarrierBase<Unknown>);
    }

    static inline const TypeInfo typeInfo() { return TypeInfo(FinalObjectType, StructureFlags); }
    static const IndexingType defaultIndexingType = NonArray;

    static const unsigned defaultSize = 64;
    static inline unsigned defaultInlineCapacity()
    {
        return (defaultSize - allocationSize(0)) / sizeof(WriteBarrier<Unknown>);
    }

    static const unsigned maxSize = 512;
    static inline unsigned maxInlineCapacity()
    {
        return (maxSize - allocationSize(0)) / sizeof(WriteBarrier<Unknown>);
    }

    static JSFinalObject* create(VM&, Structure*);
    static JSFinalObject* createWithButterfly(VM&, DeferGC&, Structure*, Butterfly*);
	static ALWAYS_INLINE Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue prototype, unsigned inlineCapacity)
	{
		return Structure::create(vm, globalObject, prototype, typeInfo(), info(), defaultIndexingType, inlineCapacity);
	}

	JS_EXPORT_PRIVATE static void visitChildren(JSCell*, SlotVisitor&);

    DECLARE_INFO;

protected:
	void visitChildrenCommon(SlotVisitor&);

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
		ASSERT(structure()->totalStorageCapacity() == structure()->inlineCapacity());
		ASSERT(classInfo());
	}

private:
	friend class LLIntOffsetsExtractor;

	explicit ALWAYS_INLINE JSFinalObject(JSGlobalData& vm, Structure* structure, Butterfly* butterfly)
		: JSObject(vm, structure, butterfly)
	{
        memset(inlineStorageUnsafe(), 0, structure->inlineCapacity() * sizeof(EncodedJSValue));
    }
};

template<>
ALWAYS_INLINE JSObject* jsDynamicCast(JSCell* from)
{
	if(from->structure()->typeInfo().isObject())
		return static_cast<JSObject*>(from);
	return 0;
}

template<>
ALWAYS_INLINE JSObject* jsDynamicCast(JSValue from)
{
	if (LIKELY(from.isCell()))
		return jsDynamicCast<JSObject*>(from.asCell());
	return 0;
}

template<>
ALWAYS_INLINE JSFinalObject* jsDynamicCast(JSCell* from)
{
	if(from->structure()->typeInfo().isFinalObject())
		return static_cast<JSFinalObject*>(from);
	return 0;
}

template<>
ALWAYS_INLINE JSFinalObject* jsDynamicCast(JSValue from)
{
	if (LIKELY(from.isCell()))
		return jsDynamicCast<JSFinalObject*>(from.asCell());
	return 0;
}

JS_EXPORT_PRIVATE EncodedJSValue JSC_HOST_CALL objectPrivateFuncInstanceOf(ExecState*);

inline JSFinalObject* JSFinalObject::create(VM& vm, Structure* structure)
{
    JSFinalObject* finalObject = new (
        NotNull,
        allocateCell<JSFinalObject>(
            vm.heap,
            allocationSize(structure->inlineCapacity())
        )
    ) JSFinalObject(vm, structure, nullptr);
    finalObject->finishCreation(vm);
    return finalObject;
}

inline JSFinalObject* JSFinalObject::createWithButterfly(VM& vm, DeferGC&, Structure* structure, Butterfly* butterfly)
{
	// DO not GC until bufferfly is attached.
    JSFinalObject* finalObject = new (
        NotNull,
        allocateCell<JSFinalObject>(
            vm.heap,
            allocationSize(structure->inlineCapacity())
        )
    ) JSFinalObject(vm, structure, butterfly);
    finalObject->finishCreation(vm);
    return finalObject;
}

ALWAYS_INLINE size_t JSObject::offsetOfInlineStorage()
{
    return sizeof(JSObject);
}

ALWAYS_INLINE bool JSObject::isGlobalObject() const
{
    return structure()->typeInfo().type() == GlobalObjectType;
}

inline bool JSObject::isJSLexicalEnvironment() const
{
	JSType type = structure()->typeInfo().type();
    return type == LexicalEnvironmentType || type == ModuleEnvironmentType;
}

inline bool JSObject::isGlobalLexicalEnvironment() const
{
    return structure()->typeInfo().type() == GlobalLexicalEnvironmentType;
}

inline bool JSObject::isStrictEvalActivation() const
{
	JSType type = structure()->typeInfo().type();
    return type == StrictEvalActivationType;
}

inline bool JSObject::isWithScope() const
{
	JSType type = structure()->typeInfo().type();
    return type == WithScopeType;
}

inline bool JSObject::isEnvironmentRecord() const
{
	JSType type = structure()->typeInfo().type();
    bool result = GlobalObjectType <= type && type <= StrictEvalActivationType;
    ASSERT((isGlobalObject() || isJSLexicalEnvironment() || isGlobalLexicalEnvironment() || isStrictEvalActivation()) == result);
    return result;
}

ALWAYS_INLINE bool JSObject::isErrorInstance() const
{
    return structure()->typeInfo().type() == ErrorInstanceType;
}

ALWAYS_INLINE void JSObject::setStructure(VM& vm, Structure* structure)
{
    ASSERT(structure);
    ASSERT(!m_butterfly == !(structure->outOfLineCapacity() || structure->hasIndexingHeader(this)));
    JSCell::setStructure(vm, structure);
}

inline void JSObject::setButterfly(VM& vm, Butterfly* butterfly)
{
    m_butterfly = butterfly;
}

ALWAYS_INLINE CallData getCallData(VM& vm, JSCell* cell)
{
	CallData result = cell->methodTable(vm)->getCallData(cell);
    ASSERT(result.type == CallData::Type::None || cell->isValidCallee());
    return result;
}

ALWAYS_INLINE CallData getCallData(VM& vm, JSValue value)
{
    if (!value.isCell())
        return CallData();
    return getCallData(vm, value.asCell());
}

ALWAYS_INLINE CallData getConstructData(VM& vm, JSCell* cell)
{
	CallData result = cell->methodTable(vm)->getConstructData(cell);
    ASSERT(result.type == CallData::Type::None || cell->isValidCallee());
    return result;
}

ALWAYS_INLINE CallData getConstructData(VM& vm, JSValue value)
{
    if (!value.isCell())
        return CallData();
    return getConstructData(vm, value.asCell());
}

ALWAYS_INLINE JSObject* asObject(JSCell* cell)
{
    ASSERT(cell->isObject());
    return jsCast<JSObject*>(cell);
}

ALWAYS_INLINE JSObject* asObject(JSValue value)
{
    return asObject(value.asCell());
}

ALWAYS_INLINE JSObject::JSObject(JSGlobalData& vm, Structure* structure, Butterfly* butterfly)
    : JSCell(vm, structure)
    , m_butterfly(butterfly)
{
}

ALWAYS_INLINE JSValue JSObject::getPrototypeDirect() const
{
    return structure()->storedPrototype();
}

ALWAYS_INLINE JSValue JSObject::getPrototype(VM& vm, ExecState* exec)
{
    if (LIKELY(!structure(vm)->typeInfo().overridesGetPrototype()))
        return getPrototypeDirect();
    return methodTable(vm)->getPrototype(this, exec);
}

// It is safe to call this method with a PropertyName that is actually an index,
// but if so will always return false (doesn't search index storage).
ALWAYS_INLINE bool JSObject::getOwnNonIndexPropertySlot(JSGlobalData& vm, Structure& structure, const PropertyName& propertyName, PropertySlot& slot)
{
    unsigned attributes;
	PropertyOffset offset = structure.get(vm, propertyName, attributes);
    if (!isValidOffset(offset)) {
        if (!TypeInfo::hasStaticPropertyTable(this->structure(vm)->typeInfo().inlineTypeFlags()))
            return false;
        if (staticPropertiesReified())
            return false;
        return getOwnStaticPropertySlot(vm, propertyName, slot);
    }

    // getPropertySlot relies on this method never returning index properties!
    ASSERT(!parseIndex(propertyName));

	JSValue value = getDirect(offset);
	if (value.isCell()) {
		ASSERT(value);
		JSCell* cell = value.asCell();
		if (UNLIKELY(!cell))
			return false;
		JSType type = cell->structure()->typeInfo().type();
		switch (type) {
		case GetterSetterType:
			fillGetterPropertySlot(slot, value, attributes, offset);
			return true;
		case CustomGetterSetterType:
			if (fillCustomGetterPropertySlot(slot, value, attributes, structure))
				return true;
			return false;
		default:
			break;
		}
	}

	slot.setValue(this, attributes, value, offset);
	return true;
}

// It may seem crazy to inline a function this large, especially a virtual function,
// but it makes a big difference to property lookup that derived classes can inline their
// base class call to this.
ALWAYS_INLINE bool JSObject::getOwnPropertySlotImpl(JSObject* object, ExecState* exec, const PropertyName& propertyName, PropertySlot& slot)
{
    VM& vm = exec->vm();
    Structure& structure = *object->structure(vm);
    if(object->getOwnNonIndexPropertySlot(vm, structure, propertyName, slot))
    	return true;
    if (Optional<uint32_t> index = parseIndex(propertyName))
        return getOwnPropertySlotByIndex(object, exec, index.value(), slot);
    return false;
}

#if ASSERT_DISABLED
ALWAYS_INLINE bool JSObject::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    return getOwnPropertySlotImpl(object, exec, propertyName, slot);
}
#endif

ALWAYS_INLINE bool JSObject::getNonIndexPropertySlot(VM& vm, ExecState* exec, const PropertyName& propertyName, PropertySlot& slot)
{
    // This method only supports non-index PropertyNames.
    ASSERT(!parseIndex(propertyName));

    JSObject* object = this;
    while (true) {
        Structure* structure = object->structure();
        if (LIKELY(!structure->typeInfo().overridesGetOwnPropertySlot())) {
            if (object->getOwnNonIndexPropertySlot(vm, *structure, propertyName, slot))
                return true;
        } else {
        	bool hasSlot = structure->classInfo()->methodTable.getOwnPropertySlot(object, exec, propertyName, slot);
        	RETURN_IF_EXCEPTION(vm, false);
			if (hasSlot)
				return true;
            if (UNLIKELY(slot.isVMInquiry() && slot.isTaintedByOpaqueObject()))
                return false;
            if (structure->typeInfo().type() == ProxyObjectType && slot.internalMethodType() == PropertySlot::InternalMethodType::HasProperty)
                return false;
            if (isTypedArrayType(structure->typeInfo().type()) && isCanonicalNumericIndexString(propertyName))
                return false;
        }
        JSValue prototype;
        if (LIKELY(!structure->typeInfo().overridesGetPrototype() || slot.internalMethodType() == PropertySlot::InternalMethodType::VMInquiry))
            prototype = object->getPrototypeDirect();
        else {
            prototype = object->getPrototype(vm, exec);
        	RETURN_IF_EXCEPTION(vm, false);
        }
        if (!prototype.isObject())
            return false;
        object = asObject(prototype);
    }
}

// It may seem crazy to inline a function this large but it makes a big difference
// since this is function very hot in variable lookup
template<bool maybeIndex>
ALWAYS_INLINE bool JSObject::getPropertySlot(JSGlobalData& vm, ExecState* exec, const PropertyName& propertyName, PropertySlot& slot)
{
    JSObject* object = this;
    while (true) {
        Structure& structure = *object->structure(vm);
        if (UNLIKELY(structure.typeInfo().overridesGetOwnPropertySlot())) {
            // If propertyName is an index then we may have missed it (as this loop is using
            // getOwnNonIndexPropertySlot), so we cannot safely call the overridden getOwnPropertySlot
            // (lest we return a property from a prototype that is shadowed). Check now for an index,
            // if so we need to start afresh from this object.
        	if (maybeIndex) {
        		if (Optional<uint32_t> index = parseIndex(propertyName))
        			return getPropertySlot(vm, exec, index.value(), slot);
        	}
            // Safe to continue searching from current position; call getNonIndexPropertySlot to avoid
            // parsing the int again.
            return object->getNonIndexPropertySlot(vm, exec, propertyName, slot);
        }
        ASSERT(object->structure()->typeInfo().type() != ProxyObjectType);
        if (object->getOwnNonIndexPropertySlot(vm, structure, propertyName, slot))
            return true;
        JSValue prototype = structure.storedPrototype();
        if (!prototype.isObject())
        	break;
        object = asObject(prototype);
    }

    if (maybeIndex) {
    	if (Optional<uint32_t> index = parseIndex(propertyName))
    		return getPropertySlot(vm, exec, index.value(), slot);
    }
    return false;
}

ALWAYS_INLINE JSValue JSObject::get(VM& vm, ExecState* exec, PropertyName propertyName) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
    bool hasProperty = const_cast<JSObject*>(this)->getPropertySlot(vm, exec, propertyName, slot);

    EXCEPTION_ASSERT(!scope.exception() || isTerminatedExecutionException(vm, scope.exception()) || !hasProperty);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    if (hasProperty)
        return slot.getValue(exec, propertyName);
    
    return jsUndefined();
}

ALWAYS_INLINE JSValue JSObject::get(VM& vm, ExecState* exec, unsigned propertyName) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
    bool hasProperty = const_cast<JSObject*>(this)->getPropertySlot(vm, exec, propertyName, slot);

    EXCEPTION_ASSERT(!scope.exception() || isTerminatedExecutionException(vm, scope.exception()) || !hasProperty);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    if (hasProperty)
        return slot.getValue(vm, exec, propertyName);

    return jsUndefined();
}

#if 0
template<typename CallbackWhenNoException>
ALWAYS_INLINE typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type JSObject::getPropertySlot(ExecState* exec, PropertyName propertyName, CallbackWhenNoException callback) const
{
    PropertySlot slot(this, PropertySlot::InternalMethodType::Get);
    return getPropertySlot(exec, propertyName, slot, callback);
}

template<typename CallbackWhenNoException>
ALWAYS_INLINE typename std::result_of<CallbackWhenNoException(bool, PropertySlot&)>::type JSObject::getPropertySlot(ExecState* exec, PropertyName propertyName, PropertySlot& slot, CallbackWhenNoException callback) const
{
	VM& vm = exec->vm();
    bool found = const_cast<JSObject*>(this)->getPropertySlot(exec, propertyName, slot);
    if (UNLIKELY(vm.exception()))
        return { };
    return callback(found, slot);
}
#endif

inline void JSObject::validatePutOwnDataProperty(VM& vm, PropertyName propertyName, JSValue value)
{
#if !ASSERT_DISABLED
    ASSERT(value);
    ASSERT(!Heap::heap(value) || Heap::heap(value) == Heap::heap(this));
    unsigned attributes;
    PropertyOffset offset = structure(vm)->get(vm, propertyName, attributes);
    if (isValidOffset(offset))
        ASSERT(!(attributes & (PropertyAttribute::Accessor | PropertyAttribute::CustomAccessor | PropertyAttribute::ReadOnly)));
    else if (TypeInfo::hasStaticPropertyTable(inlineTypeFlags())) {
        if (auto entry = findPropertyHashEntry(vm, propertyName))
            ASSERT(!(entry->value->attributes() & (PropertyAttribute::Accessor | PropertyAttribute::CustomAccessor | PropertyAttribute::ReadOnly)));
    }
#else
    UNUSED_PARAM(vm);
    UNUSED_PARAM(propertyName);
    UNUSED_PARAM(value);
#endif
}

inline bool JSObject::putOwnDataProperty(JSGlobalData& vm, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    validatePutOwnDataProperty(vm, propertyName, value);
    return putDirectInternal<PutModePut>(vm, propertyName, value, 0, slot).isNull();
}

inline bool JSObject::putOwnDataPropertyMayBeIndex(ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    VM& vm = exec->vm();
    validatePutOwnDataProperty(vm, propertyName, value);
    if (Optional<uint32_t> index = parseIndex(propertyName))
        return putDirectIndex(exec, index.value(), value, 0, PutDirectIndexLikePutDirect);

    return putDirectInternal<PutModePut>(vm, propertyName, value, 0, slot).isNull();
}

inline bool JSObject::putDirect(JSGlobalData& vm, PropertyName propertyName, JSValue value, unsigned attributes)
{
    ASSERT(!value.isGetterSetter() && !(attributes & Accessor));
    ASSERT(!value.isCustomGetterSetter() && !(attributes & PropertyAttribute::CustomAccessorOrValue));
    PutPropertySlot slot(this);
    return putDirectInternal<PutModeDefineOwnProperty>(vm, propertyName, value, attributes, slot).isNull();
}

inline bool JSObject::putDirect(VM& vm, PropertyName propertyName, JSValue value, unsigned attributes, PutPropertySlot& slot)
{
    ASSERT(!value.isGetterSetter());
    ASSERT(!value.isCustomGetterSetter());
    return putDirectInternal<PutModeDefineOwnProperty>(vm, propertyName, value, attributes, slot).isNull();
}

inline bool JSObject::putDirect(JSGlobalData& vm, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    ASSERT(!value.isGetterSetter());
    ASSERT(!value.isCustomGetterSetter());
    return putDirectInternal<PutModeDefineOwnProperty>(vm, propertyName, value, 0, slot).isNull();
}

ALWAYS_INLINE JSObject* Register::object() const
{
	return asObject(jsValue());
}

ALWAYS_INLINE Register& Register::operator=(JSObject* object)
{
    u.value = JSValue::encode(JSValue(object));
    return *this;
}

ALWAYS_INLINE size_t offsetInButterfly(PropertyOffset offset)
{
    return offsetInOutOfLineStorage(offset) + Butterfly::indexOfPropertyStorage();
}

ALWAYS_INLINE size_t JSObject::butterflyPreCapacity()
{
    if (UNLIKELY(hasIndexingHeader()))
        return butterfly()->indexingHeader()->preCapacity(structure());
    return 0;
}

ALWAYS_INLINE size_t JSObject::butterflyTotalSize()
{
    Structure* structure = this->structure();
    Butterfly* butterfly = this->butterfly();
    size_t preCapacity;
    size_t indexingPayloadSizeInBytes;
    bool hasIndexingHeader = this->hasIndexingHeader();

    if (UNLIKELY(hasIndexingHeader)) {
        preCapacity = butterfly->indexingHeader()->preCapacity(structure);
        indexingPayloadSizeInBytes = butterfly->indexingHeader()->indexingPayloadSizeInBytes(structure);
    } else {
        preCapacity = 0;
        indexingPayloadSizeInBytes = 0;
    }

    return Butterfly::totalSize(preCapacity, structure->outOfLineCapacity(), hasIndexingHeader, indexingPayloadSizeInBytes);
}

inline int indexRelativeToBase(PropertyOffset offset)
{
    if (isOutOfLineOffset(offset))
        return offsetInOutOfLineStorage(offset) + Butterfly::indexOfPropertyStorage();
    ASSERT(!(JSObject::offsetOfInlineStorage() % sizeof(EncodedJSValue)));
    return JSObject::offsetOfInlineStorage() / sizeof(EncodedJSValue) + offsetInInlineStorage(offset);
}

inline int offsetRelativeToBase(PropertyOffset offset)
{
    if (isOutOfLineOffset(offset))
        return offsetInOutOfLineStorage(offset) * sizeof(EncodedJSValue) + Butterfly::offsetOfPropertyStorage();
    return JSObject::offsetOfInlineStorage() + offsetInInlineStorage(offset) * sizeof(EncodedJSValue);
}

// Returns the maximum offset (away from zero) a load instruction will encode.
inline size_t maxOffsetRelativeToBase(PropertyOffset offset)
{
    ptrdiff_t addressOffset = offsetRelativeToBase(offset);
#if USE(JSVALUE32_64)
    if (addressOffset >= 0)
        return static_cast<size_t>(addressOffset) + OBJECT_OFFSETOF(EncodedValueDescriptor, asBits.tag);
#endif
    return static_cast<size_t>(addressOffset);
}

COMPILE_ASSERT(!(sizeof(JSObject) % sizeof(WriteBarrierBase<Unknown>)), JSObject_inline_storage_has_correct_alignment);

template<unsigned charactersCount>
ALWAYS_INLINE Identifier makeIdentifier(VM& vm, const char (&characters)[charactersCount])
{
    return Identifier::fromString(vm, ASCIILiteral(characters));
}

ALWAYS_INLINE Identifier makeIdentifier(JSGlobalData& vm, const Identifier& name)
{
    return name;
}

bool validateAndApplyPropertyDescriptor(ExecState*, JSObject*, PropertyName, bool isExtensible,
    const PropertyDescriptor& descriptor, bool isCurrentDefined, const PropertyDescriptor& current, bool throwException);

JS_EXPORT_PRIVATE NEVER_INLINE bool ordinarySetSlow(ExecState*, JSObject*, PropertyName, JSValue, JSValue receiver, bool shouldThrow);

// Helper for defining native functions, if you're not using a static hash table.
// Use this macro from within finishCreation() methods in prototypes. This assumes
// you've defined variables called exec, globalObject, and vm, and they
// have the expected meanings.
#define JSC_NATIVE_INTRINSIC_FUNCTION(jsName, cppName, attributes, length, intrinsic) \
    putDirectNativeFunction(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (length), cppName, \
        (intrinsic), (attributes))

#define JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(jsName, cppName, attributes, length, intrinsic) \
    putDirectNativeFunctionWithoutTransition(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (length), cppName, \
        (intrinsic), (attributes))

// As above, but this assumes that the function you're defining doesn't have an
// intrinsic.
#define JSC_NATIVE_FUNCTION(jsName, cppName, attributes, length) \
    JSC_NATIVE_INTRINSIC_FUNCTION(jsName, cppName, (attributes), (length), NoIntrinsic)

#define JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(jsName, cppName, attributes, length) \
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(jsName, cppName, (attributes), (length), NoIntrinsic)

// Identical helpers but for builtins. Note that currently, we don't support builtins that are
// also intrinsics, but we probably will do that eventually.
#define JSC_BUILTIN_FUNCTION(jsName, generatorName, attributes) \
    putDirectBuiltinFunction(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (generatorName)(globalObject), (attributes))

#define JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(jsName, generatorName, attributes) \
    putDirectBuiltinFunctionWithoutTransition(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (generatorName)(globalObject), (attributes))

#define JSC_TO_STRING_TAG_WITHOUT_TRANSITION() \
    putDirectWithoutTransition(vm, vm.propertyNames->toStringTagSymbol, \
        jsNontrivialString(vm, String::fromLiteral(info()->className)), DontEnum | ReadOnly)

// Helper for defining native getters on properties.
#define JSC_NATIVE_INTRINSIC_GETTER(jsName, cppName, attributes, intrinsic)  \
    putDirectNativeIntrinsicGetter(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (cppName), \
        (intrinsic), ((attributes) | Accessor))

#define JSC_NATIVE_INTRINSIC_GETTER_WITHOUT_TRANSITION(jsName, cppName, attributes, intrinsic)  \
    putDirectNativeIntrinsicGetterWithoutTransition(\
        vm, globalObject, makeIdentifier(vm, (jsName)), (cppName), \
        (intrinsic), ((attributes) | Accessor))

#define JSC_NATIVE_GETTER(jsName, cppName, attributes) \
    JSC_NATIVE_INTRINSIC_GETTER((jsName), (cppName), (attributes), NoIntrinsic)

#define JSC_NATIVE_GETTER_WITHOUT_TRANSITION(jsName, cppName, attributes) \
    JSC_NATIVE_INTRINSIC_GETTER_WITHOUT_TRANSITION((jsName), (cppName), (attributes), NoIntrinsic)

} // namespace JSC
