/*
 *  Copyright (C) 1999-2002 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2004, 2007, 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "JSString.h"

#include "JSGlobalObject.h"
#include "JSGlobalObjectFunctions.h"
#include "JSObject.h"
#include "Operations.h"
#include "StringObject.h"
#include "StringPrototype.h"
#include "JSStringInlines.h"
#include "StrongInlines.h"

namespace JSC {
    
static const unsigned substringFromRopeCutoff = 4;

const ClassInfo JSString::s_info = { "string", nullptr, nullptr, CREATE_METHOD_TABLE(JSString) };

template<>
void JSRopeString::RopeBuilder<RecordOverflow>::expand()
{
    RELEASE_ASSERT(!this->hasOverflowed());
    ASSERT(m_index == JSRopeString::s_maxInternalRopeLength);
    JSString* jsString = m_jsString;
    m_jsString = jsStringBuilder(&m_vm);
    m_index = 0;
    append(jsString);
}

void JSString::destroy(JSCell* cell)
{
    static_cast<JSString*>(cell)->JSString::~JSString();
}

bool JSString::equalSlowCase(ExecState* exec, JSString* other) const
{
	VM& vm = exec->vm();

    if (length() != other->length())
        return false;

    String str1 = value(exec);
    RETURN_IF_EXCEPTION(vm, false);
    String str2 = other->value(exec);
    RETURN_IF_EXCEPTION(vm, false);
    return WTF::equal(*str1.impl(), *str2.impl());
}

void JSString::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    JSString* thisObject = asString(cell);
    Base::visitChildren(thisObject, visitor);

    MARK_LOG_MESSAGE1("[%u]: ", thisObject->length());

#if ENABLE(OBJECT_MARK_LOGGING)
    if (!thisObject->isRope()) {
        WTF::StringImpl* ourImpl = thisObject->m_value.impl();
        if (ourImpl->is8Bit())
            MARK_LOG_MESSAGE1("[8 %p]", ourImpl->characters8());
        else
            MARK_LOG_MESSAGE1("[16 %p]", ourImpl->characters16());
    } else
        MARK_LOG_MESSAGE0("[rope]: ");
#endif

    if (thisObject->isRope())
        static_cast<JSRopeString*>(thisObject)->visitFibers(visitor);
    else
    	visitor.reportExtraMemoryVisited(cell, thisObject->m_value.impl()->costDuringGC());
}

ALWAYS_INLINE void JSRopeString::visitFibers(SlotVisitor& visitor)
{
    if (isSubstring()) {
        visitor.append(&substringBase());
        return;
    }
    for (size_t i = 0; i < s_maxInternalRopeLength; ++i)
    	visitor.append(fiber(i));
}

static const unsigned maxLengthForOnStackResolve = 512;

template<typename CharacterType>
void JSRopeString::resolveRopeInternalNoSubstring(CharacterType* buffer) const
{
    for (size_t i = 0; i < s_maxInternalRopeLength && fiber(i); ++i) {
        if (fiber(i)->isRope()) {
            resolveRopeSlowCase(buffer);
            return;
        }
    }

    CharacterType* position = buffer;
    for (size_t i = 0; i < s_maxInternalRopeLength && fiber(i); ++i) {
        StringImpl* string = fiber(i)->m_value.impl();
        unsigned length = string->length();
        if (string->is8Bit())
            StringImpl::copyChars(position, string->characters8(), length);
        else
            StringImpl::copyChars(position, string->characters16(), length);
        position += length;
    }
    ASSERT((buffer + length()) == position);
}

void JSRopeString::resolveRopeToAtomicString(ExecState* exec) const
{
    VM& vm = exec->vm();

    if (length() > maxLengthForOnStackResolve) {
        constexpr bool reportAllocation = true;
        resolveRopeInternal<reportAllocation>(exec);
        RETURN_IF_EXCEPTION(vm, void());
        m_value = AtomicString(m_value);
        setIs8Bit(m_value.impl()->is8Bit());
        return;
    }

    AtomicString atomString;
	if (!isSubstring()) {
        if (is8Bit()) {
            LChar buffer[maxLengthForOnStackResolve];
            resolveRopeInternalNoSubstring(buffer);
            atomString = AtomicString(buffer, length());
        } else {
            UChar buffer[maxLengthForOnStackResolve];
            resolveRopeInternalNoSubstring(buffer);
            atomString = AtomicString(buffer, length());
        }
    } else
        atomString = StringView(substringBase()->m_value).substring(substringOffset(), length()).toAtomicString();

    // If we resolved a string that didn't previously exist, notify the heap that we've grown.
    if (atomString.impl()->hasOneRef()) {
    	// GC crash when called from DFG JIT operationPutByValCellStringNonStrict
    	DeferGCForAWhile deferGC(vm.heap);
        vm.heap.reportExtraMemoryAllocated(atomString.impl()->cost());
    }

	m_value = atomString;
	setIs8Bit(m_value.impl()->is8Bit());
    clearFibers();
}

void JSRopeString::clearFibers() const
{
    for (size_t i = 0; i < s_maxInternalRopeLength; ++i)
        u[i].number = 0;
}

PassRefPtr<AtomicStringImpl> JSRopeString::resolveRopeToExistingAtomicString(ExecState* exec) const
{
    VM& vm = exec->vm();

    if (length() > maxLengthForOnStackResolve) {
        constexpr bool reportAllocation = true;
        resolveRopeInternal<reportAllocation>(exec);
        RETURN_IF_EXCEPTION(vm, nullptr);
        if (RefPtr<AtomicStringImpl> existingAtomicString = AtomicString::find(m_value.impl())) {
            m_value = existingAtomicString.get();
            setIs8Bit(m_value.impl()->is8Bit());
            clearFibers();
            return existingAtomicString.release();
        }
        return nullptr;
    }

    RefPtr<AtomicStringImpl> existingAtomString;
	if (!isSubstring()) {
		if (is8Bit()) {
			LChar buffer[maxLengthForOnStackResolve];
			resolveRopeInternalNoSubstring(buffer);
			existingAtomString = AtomicString::lookup(buffer, length());
		} else {
			UChar buffer[maxLengthForOnStackResolve];
			resolveRopeInternalNoSubstring(buffer);
			existingAtomString = AtomicString::lookup(buffer, length());
		}
    } else
        existingAtomString = StringView(substringBase()->m_value).substring(substringOffset(), length()).toExistingAtomicString();

    if (existingAtomString) {
    	m_value = existingAtomString.get();
		setIs8Bit(m_value.impl()->is8Bit());
		clearFibers();
    }
    return existingAtomString;
}

template<bool reportAllocation>
inline void JSRopeString::resolveRopeInternal(ExecState* nullOrExecForOOM) const
{
    ASSERT(isRope());

    if (isSubstring()) {
        ASSERT(!substringBase()->isRope());
        m_value = substringBase()->m_value.substringSharingImpl(substringOffset(), length());
        substringBase().clear();
        return;
    }

    if (is8Bit()) {
        LChar* buffer;
        if (RefPtr<StringImpl> newImpl = StringImpl::tryCreateUninitialized(length(), buffer)) {
        	if (reportAllocation)
        		Heap::heap(this)->reportExtraMemoryAllocated(newImpl->cost());
            m_value = newImpl.release();
        } else {
            outOfMemory(nullOrExecForOOM);
            m_value = emptyString();
            return;
        }

        resolveRopeInternalNoSubstring(buffer);
        clearFibers();

        return;
    }

    UChar* buffer;
    if (RefPtr<StringImpl> newImpl = StringImpl::tryCreateUninitialized(length(), buffer)) {
    	if (reportAllocation)
    		Heap::heap(this)->reportExtraMemoryAllocated(newImpl->cost());
        m_value = newImpl.release();
    } else {
        outOfMemory(nullOrExecForOOM);
        m_value = emptyString();
        return;
    }

    resolveRopeInternalNoSubstring(buffer);
    clearFibers();
    ASSERT(!isRope());
}

void JSRopeString::resolveRope(ExecState* nullOrExecForOOM) const
{
    constexpr bool reportAllocation = true;
    resolveRopeInternal<reportAllocation>(nullOrExecForOOM);
}

void JSRopeString::resolveRopeWithoutGC() const
{
    constexpr bool reportAllocation = false;
    resolveRopeInternal<reportAllocation>(nullptr);
}

// Overview: These functions convert a JSString from holding a string in rope form
// down to a simple String representation.  It does so by building up the string
// backwards, since we want to avoid recursion, we expect that the tree structure
// representing the rope is likely imbalanced with more nodes down the left side
// (since appending to the string is likely more common) - and as such resolving
// in this fashion should minimize work queue size.  (If we built the queue forwards
// we would likely have to place all of the constituent StringImpls into the
// Vector before performing any concatenation, but by working backwards we likely
// only fill the queue with the number of substrings at any given level in a
// rope-of-ropes.)    
template<typename CharacterType>
void JSRopeString::resolveRopeSlowCase(CharacterType* buffer) const
{
	CharacterType* position = buffer + length(); // We will be working backwards over the rope.
    Vector<JSString*, 32> workQueue; // These strings are kept alive by the parent rope, so using a Vector is OK.

    for (size_t i = 0; i < s_maxInternalRopeLength && fiber(i); ++i)
        workQueue.append(fiber(i).get());

    while (!workQueue.isEmpty()) {
        JSString* currentFiber = workQueue.last();
        workQueue.removeLast();

        if (currentFiber->isRope()) {
            JSRopeString* currentFiberAsRope = static_cast<JSRopeString*>(currentFiber);
            if (currentFiberAsRope->isSubstring()) {
				ASSERT(!currentFiberAsRope->substringBase()->isRope());
				StringImpl* string = static_cast<StringImpl*>(
					currentFiberAsRope->substringBase()->m_value.impl());
				unsigned offset = currentFiberAsRope->substringOffset();
				unsigned length = currentFiberAsRope->length();
				position -= length;
				if (string->is8Bit())
					StringImpl::copyChars(position, string->characters8() + offset, length);
				else
					StringImpl::copyChars(position, string->characters16() + offset, length);
				continue;
			}
			for (size_t i = 0; i < s_maxInternalRopeLength && currentFiberAsRope->fiber(i); ++i)
				workQueue.append(currentFiberAsRope->fiber(i).get());
			continue;
		}

        StringImpl* string = static_cast<StringImpl*>(currentFiber->m_value.impl());
        unsigned length = string->length();
        position -= length;
        if(string->is8Bit())
            StringImpl::copyChars(position, string->characters8(), length);
        else
        	StringImpl::copyChars(position, string->characters16(), length);
    }

    ASSERT(buffer == position);
}

void JSRopeString::outOfMemory(ExecState* nullOrExecForOOM) const
{
	clearFibers();
    ASSERT(isRope());
    ASSERT(m_value.isNull());
    if (nullOrExecForOOM) {
        VM& vm = nullOrExecForOOM->vm();
        //auto scope = DECLARE_THROW_SCOPE(vm);
        throwOutOfMemoryError(vm, nullOrExecForOOM);
    }
}

double JSString::toNumber(ExecState* exec) const
{
    VM& vm = exec->vm();
    StringView view = unsafeView(exec);
    RETURN_IF_EXCEPTION(vm, 0);
    return jsToNumber(view);
}

inline StringObject* StringObject::create(VM& vm, JSGlobalObject* globalObject, JSString* string)
{
    StringObject* object = new (NotNull, allocateCell<StringObject>(vm.heap)) StringObject(vm, globalObject->stringObjectStructure());
    object->finishCreation(vm, string);
    return object;
}

JSObject* JSString::toObject(ExecState*, JSGlobalObject* globalObject) const
{
    return StringObject::create(globalObject->vm(), globalObject, const_cast<JSString*>(this));
}

bool JSString::getStringPropertyDescriptor(ExecState* exec, PropertyName propertyName, PropertyDescriptor& descriptor)
{
	VM& vm = exec->vm();
    if (propertyName == vm.propertyNames->length) {
        descriptor.setDescriptor(jsNumber(length()), DontEnum | DontDelete | ReadOnly);
        return true;
    }

    Optional<uint32_t> index = parseIndex(propertyName);
    if (index && index.value() < length()) {
        descriptor.setDescriptor(getIndex(vm, exec, index.value()), DontDelete | ReadOnly);
        return true;
    }
    
    return false;
}

inline JSString* jsStringWithCacheSlowCase(VM& vm, StringImpl& stringImpl)
{
    if (JSString* string = vm.stringCache.get(&stringImpl))
        return string;

    JSString* string = jsString(vm, String(&stringImpl));
    vm.lastCachedString.set(vm, string);
    vm.stringCache.set(&stringImpl, string);
    return string;
}

JSString* jsStringWithCache(VM& vm, const String& s)
{
    StringImpl* stringImpl = s.impl();
    if (UNLIKELY(!stringImpl || !stringImpl->length()))
        return jsEmptyString(&vm);

    if (stringImpl->length() == 1) {
        UChar singleCharacter = (*stringImpl)[0u];
        if (singleCharacter <= maxSingleCharacterString)
            return vm.smallStrings.singleCharacterString(static_cast<unsigned char>(singleCharacter));
    }

    if (JSString* lastCachedString = vm.lastCachedString.get()) {
        if (lastCachedString->tryGetValueImpl() == stringImpl)
            return lastCachedString;
    }

    return jsStringWithCacheSlowCase(vm, *stringImpl);
}

} // namespace JSC
