/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "CallFrame.h"
#include "CatchScope.h"
#include "CommonIdentifiers.h"
#include "Identifier.h"
#include "PropertyDescriptor.h"
#include "PropertySlot.h"
#include "Structure.h"
#include "ThrowScope.h"
#include <wtf/CheckedArithmetic.h>
#include <wtf/text/StringView.h>

namespace JSC {

class JSString;
class JSRopeString;
class LLIntOffsetsExtractor;

JSString* jsEmptyString(JSGlobalData&);
JSString* jsEmptyString(JSGlobalData*);
JSString* jsString(JSGlobalData&, const String&); // returns empty string if passed null string

JSString* jsSingleCharacterString(JSGlobalData&, UChar);
JSString* jsSingleCharacterString(JSGlobalData*, UChar);
JSString* jsSubstring(JSGlobalData&, const String&, unsigned offset, unsigned length);
JSString* jsSubstring(JSGlobalData*, const String&, unsigned offset, unsigned length);

// Non-trivial strings are two or more characters long.
// These functions are faster than just calling jsString.
JSString* jsNontrivialString(JSGlobalData&, const String&);
JSString* jsNontrivialString(JSGlobalData*, const String&);
JSString* jsNontrivialString(JSGlobalData&, PassRefPtr<StringImpl>);
JSString* jsNontrivialString(JSGlobalData*, PassRefPtr<StringImpl>);

// Should be used for strings that are owned by an object that will
// likely outlive the JSValue this makes, such as the parse tree or a
// DOM object that contains a String
JSString* jsOwnedString(JSGlobalData&, const String&);
JSString* jsOwnedString(JSGlobalData*, const String&);

JSRopeString* jsStringBuilder(JSGlobalData&);
JSRopeString* jsStringBuilder(JSGlobalData*);

bool isJSString(JSValue);
JSString* asString(JSValue);

class JSString : public JSCell {
public:
	friend class JIT;
	friend class JSGlobalData;
	friend class SpecializedThunkJIT;
	friend class JSRopeString;
	friend class MarkStack;
	friend class SlotVisitor;
	friend struct ThunkHelpers;

	typedef JSCell Base;
    // Do we really need OverridesGetOwnPropertySlot?
    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=212956
    // Do we really need InterceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero?
    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=212958
    static const unsigned StructureFlags = Base::StructureFlags | OverridesGetOwnPropertySlot | InterceptsGetOwnPropertySlotByIndexEvenWhenLengthIsNotZero | StructureIsImmortal | OverridesToThis;

	static const bool needsDestruction = true;
	static void destroy(JSCell*);

    // We employ overflow checks in many places with the assumption that MaxLength
    // is INT_MAX. Hence, it cannot be changed into another length value without
    // breaking all the bounds and overflow checks that assume this.
    static unsigned MaxLength() { return std::numeric_limits<int32_t>::max(); }
    //static_assert(MaxLength() == String::MaxLength(), "");

private:
	ALWAYS_INLINE JSString(JSGlobalData& vm, PassRefPtr<StringImpl> value)
		: JSCell(vm, vm.stringStructure.get())
		, m_flags(0)
		, m_value(value)
	{
	}

	ALWAYS_INLINE JSString(JSGlobalData& vm)
		: JSCell(vm, vm.stringStructure.get())
		, m_flags(0)
	{
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm, unsigned length)
	{
		ASSERT(!m_value.isNull());
		Base::finishCreation(vm);
        setLength(length);
		setIs8Bit(m_value.impl()->is8Bit());
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm, unsigned length, size_t cost)
	{
		ASSERT(!m_value.isNull());
		Base::finishCreation(vm);
        setLength(length);
		setIs8Bit(m_value.impl()->is8Bit());
		vm.heap.reportExtraMemoryAllocated(cost);
	}

protected:
	ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		Base::finishCreation(vm);
        setLength(0);
		setIs8Bit(true);
	}

public:
	static ALWAYS_INLINE JSString* create(JSGlobalData& vm, PassRefPtr<StringImpl> value)
	{
		ASSERT(value);
		unsigned length = value->length();
        RELEASE_ASSERT(length >= 0);
		size_t cost = value->cost();
		JSString* newString = new (NotNull, allocateCell<JSString>(vm.heap)) JSString(vm, value);
		newString->finishCreation(vm, length, cost);
		return newString;
	}
	static ALWAYS_INLINE JSString* createHasOtherOwner(JSGlobalData& vm, PassRefPtr<StringImpl> value)
	{
		ASSERT(value);
		unsigned length = value->length();
		JSString* newString = new (NotNull, allocateCell<JSString>(vm.heap)) JSString(vm, value);
		newString->finishCreation(vm, length);
		return newString;
	}

    Identifier toIdentifier(VM&, ExecState*) const;
    AtomicString toAtomicString(ExecState*) const;
    PassRefPtr<AtomicStringImpl> toExistingAtomicString(ExecState*) const;

    StringViewWithUnderlyingString viewWithUnderlyingString(ExecState*) const;

    inline bool equal(ExecState*, JSString* other) const;
	const String& value(ExecState*) const;
	const String& value(ExecState&) const;
	inline const String& tryGetValue() const;
	const String& tryGetValueWithoutGC() const;
    const StringImpl* tryGetValueImpl() const;
    ALWAYS_INLINE unsigned length() const { return m_length; }

	JSValue toPrimitive(ExecState*, PreferredPrimitiveType) const;
	JS_EXPORT_PRIVATE bool toBoolean() const { return !!length(); }
	JSObject* toObject(ExecState*, JSGlobalObject*) const;
	double toNumber(ExecState*) const;

	template<bool maybeIndex = true> ALWAYS_INLINE bool getStringPropertySlot(JSGlobalData&, ExecState*, const PropertyName&, PropertySlot&);
	bool getStringPropertySlot(JSGlobalData&, ExecState*, unsigned propertyName, PropertySlot&);
	bool getStringPropertyDescriptor(ExecState*, PropertyName, PropertyDescriptor&);

	ALWAYS_INLINE bool canGetIndex(unsigned i) { return i < length(); }
	JSString* getIndex(VM&, ExecState*, unsigned);

	static Structure* createStructure(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue proto)
	{
		return Structure::create(vm, globalObject, proto, TypeInfo(StringType, StructureFlags), info());
	}

	static size_t offsetOfLength() { return OBJECT_OFFSETOF(JSString, m_length); }
	static size_t offsetOfFlags() { return OBJECT_OFFSETOF(JSString, m_flags); }
	static size_t offsetOfValue() { return OBJECT_OFFSETOF(JSString, m_value); }

	DECLARE_EXPORT_INFO;

	static void visitChildren(JSCell*, SlotVisitor&);

    enum {
        Is8Bit = 1u
    };

    bool isRope() const { return m_value.isNull(); }
	ALWAYS_INLINE bool is8Bit() const { return m_flags & Is8Bit; }

protected:
    friend class JSValue;

    JS_EXPORT_PRIVATE bool equalSlowCase(ExecState*, JSString* other) const;
    bool isSubstring() const;
	ALWAYS_INLINE void setIs8Bit(bool flag) const
	{
		if (flag)
			m_flags |= Is8Bit;
		else
			m_flags &= ~Is8Bit;
	}

    ALWAYS_INLINE void setLength(unsigned length)
    {
        m_length = length;
    }

private:
    mutable unsigned m_flags { 0 };

	// A string is represented either by a String or a rope of fibers.
	unsigned m_length { 0 };
	mutable String m_value;

	friend class LLIntOffsetsExtractor;

	ALWAYS_INLINE String& string() { ASSERT(!isRope()); return m_value; }
    StringView unsafeView(ExecState*) const;
};

class JSRopeString : public JSString {
	friend class JSString;

	friend JSRopeString* jsStringBuilder(JSGlobalData*);

public:
    template <class OverflowHandler = CrashOnOverflow>
    class RopeBuilder : public OverflowHandler {
	public:
		ALWAYS_INLINE RopeBuilder(JSGlobalData& vm)
		: m_vm(vm)
		, m_jsString(jsStringBuilder(&vm))
		, m_index(0)
		{
		}

		ALWAYS_INLINE bool append(JSString* jsString)
		{
            if (UNLIKELY(this->hasOverflowed()))
                return false;
			if (m_index == JSRopeString::s_maxInternalRopeLength)
				expand();
            //static_assert(JSString::MaxLength() == std::numeric_limits<int32_t>::max(), "");
            auto sum = checkedSum<int32_t>(m_jsString->length(), jsString->length());
            if (sum.hasOverflowed()) {
            	this->overflowed();
                return false;
            }
            //ASSERT(static_cast<unsigned>(sum.unsafeGet()) <= MaxLength);
			m_jsString->append(m_vm, m_index++, jsString);
            return true;
		}

		ALWAYS_INLINE JSRopeString* release()
		{
            RELEASE_ASSERT(!this->hasOverflowed());
			JSRopeString* tmp = m_jsString;
			m_jsString = 0;
			return tmp;
		}

		ALWAYS_INLINE unsigned length() const
        {
            ASSERT(!this->hasOverflowed());
            return m_jsString->length();
        }

	private:
		void expand();

		JSGlobalData& m_vm;
		JSRopeString* m_jsString;
		size_t m_index;
	};

private:
	ALWAYS_INLINE JSRopeString(JSGlobalData& vm)
		: JSString(vm)
	{
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm, JSString* s1, JSString* s2)
	{
		Base::finishCreation(vm);
        ASSERT(!sumOverflows<int32_t>(s1->length(), s2->length()));
        setLength(s1->length() + s2->length());
		setIs8Bit(s1->is8Bit() && s2->is8Bit());
		setIsSubstring(false);
        fiber(0).set(vm, this, s1);
        fiber(1).set(vm, this, s2);
        fiber(2).clear();
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm, JSString* s1, JSString* s2, JSString* s3)
	{
		Base::finishCreation(vm);
        ASSERT(!sumOverflows<int32_t>(s1->length(), s2->length(), s3->length()));
        setLength(s1->length() + s2->length() + s3->length());
		setIs8Bit(s1->is8Bit() && s2->is8Bit() &&  s3->is8Bit());
		setIsSubstring(false);
		fiber(0).set(vm, this, s1);
		fiber(1).set(vm, this, s2);
		fiber(2).set(vm, this, s3);
	}

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm, ExecState* exec, JSString* base, unsigned offset, unsigned length)
	{
        Base::finishCreation(vm);
        ASSERT(!sumOverflows<int32_t>(offset, length));
        ASSERT(offset + length <= base->length());
        setLength(length);
        setIs8Bit(base->is8Bit());
        setIsSubstring(true);
        if (base->isSubstring()) {
            JSRopeString* baseRope = jsCast<JSRopeString*>(base);
            substringBase().set(vm, this, baseRope->substringBase().get());
            substringOffset() = baseRope->substringOffset() + offset;
        } else {
            substringBase().set(vm, this, base);
            substringOffset() = offset;

            // For now, let's not allow substrings with a rope base.
            // Resolve non-substring rope bases so we don't have to deal with it.
            // FIXME: Evaluate if this would be worth adding more branches.
            if (base->isRope())
            	jsCast<JSRopeString*>(base)->resolveRope(exec);
        }
	}

    ALWAYS_INLINE void finishCreationSubstringOfResolved(VM& vm, JSString* base, unsigned offset, unsigned length)
    {
        Base::finishCreation(vm);
        ASSERT(!sumOverflows<int32_t>(offset, length));
        ASSERT(offset + length <= base->length());
        setLength(length);
        setIs8Bit(base->is8Bit());
        setIsSubstring(true);
        substringBase().set(vm, this, base);
        substringOffset() = offset;
    }

	ALWAYS_INLINE void finishCreation(JSGlobalData& vm)
	{
		JSString::finishCreation(vm);
		setIsSubstring(false);
		fiber(0).clear();
		fiber(1).clear();
		fiber(2).clear();
	}

	void append(JSGlobalData& vm, size_t index, JSString* jsString)
	{
		fiber(index).set(vm, this, jsString);
        setLength(length() + jsString->length());
		setIs8Bit(is8Bit() && jsString->is8Bit());
	}

public:
	static ALWAYS_INLINE JSRopeString* createNull(JSGlobalData& vm)
	{
		JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(vm.heap)) JSRopeString(vm);
		newString->finishCreation(vm);
		return newString;
	}

	ALWAYS_INLINE static JSString* create(JSGlobalData& vm, ExecState* exec, JSString* base, unsigned offset, unsigned length)
    {
        JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(vm.heap)) JSRopeString(vm);
        newString->finishCreation(vm, exec, base, offset, length);
        return newString;
    }

    ALWAYS_INLINE static JSString* createSubstringOfResolved(VM& vm, JSString* base, unsigned offset, unsigned length)
    {
        JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(vm.heap)) JSRopeString(vm);
        newString->finishCreationSubstringOfResolved(vm, base, offset, length);
        return newString;
    }

	void visitFibers(SlotVisitor&);

    static ptrdiff_t offsetOfFibers() { return OBJECT_OFFSETOF(JSRopeString, u); }

    static const unsigned s_maxInternalRopeLength = 3;

    static JSString* create(VM& vm, JSString* s1, JSString* s2)
    {
        JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(vm.heap)) JSRopeString(vm);
        newString->finishCreation(vm, s1, s2);
        return newString;
    }
    static JSString* create(VM& vm, JSString* s1, JSString* s2, JSString* s3)
    {
        JSRopeString* newString = new (NotNull, allocateCell<JSRopeString>(vm.heap)) JSRopeString(vm);
        newString->finishCreation(vm, s1, s2, s3);
        return newString;
    }

private:
	friend JSValue jsStringFromRegisterArray(ExecState*, Register*, unsigned);
	friend JSValue jsStringFromArguments(ExecState*, JSValue);

    // If nullOrExecForOOM is null, resolveRope() will be do nothing in the event of an OOM error.
    // The rope value will remain a null string in that case.
    JS_EXPORT_PRIVATE void resolveRope(ExecState* nullOrExecForOOM) const;
    JS_EXPORT_PRIVATE void resolveRopeWithoutGC() const;
	JS_EXPORT_PRIVATE void resolveRopeToAtomicString(ExecState*) const;
	JS_EXPORT_PRIVATE PassRefPtr<AtomicStringImpl> resolveRopeToExistingAtomicString(ExecState*) const;
	template<typename CharacterType> NEVER_INLINE void resolveRopeSlowCase(CharacterType*) const;
	template<typename CharacterType> void resolveRopeInternalNoSubstring(CharacterType*) const;
	void outOfMemory(ExecState* nullOrExecForOOM) const;
    void clearFibers() const;
    StringView unsafeView(ExecState*) const;
    StringViewWithUnderlyingString viewWithUnderlyingString(ExecState*) const;

    template<bool reportAllocation>
    inline void resolveRopeInternal(ExecState* nullOrExecForOOM) const;

    ALWAYS_INLINE WriteBarrierBase<JSString>& fiber(unsigned i) const
    {
        ASSERT(!isSubstring());
        ASSERT(i < s_maxInternalRopeLength);
        return u[i].string;
    }

    ALWAYS_INLINE WriteBarrierBase<JSString>& substringBase() const
    {
        return u[1].string;
    }

    ALWAYS_INLINE uintptr_t& substringOffset() const
    {
        return u[2].number;
    }

    static uintptr_t notSubstringSentinel()
    {
        return 0;
    }

    static uintptr_t substringSentinel()
    {
        return 1;
    }

    ALWAYS_INLINE bool isSubstring() const
    {
        return u[0].number == substringSentinel();
    }

    ALWAYS_INLINE void setIsSubstring(bool isSubstring)
    {
        u[0].number = isSubstring ? substringSentinel() : notSubstringSentinel();
    }

    mutable union {
        uintptr_t number;
        WriteBarrierBase<JSString> string;
    } u[s_maxInternalRopeLength];

    friend JSString* jsString(JSGlobalData&, ExecState*, JSString*, JSString*);
    friend JSString* jsString(JSGlobalData&, ExecState*, const String&, JSString*);
    friend JSString* jsString(JSGlobalData&, ExecState*, JSString*, const String&);
    friend JSString* jsString(JSGlobalData&, ExecState*, JSString*, JSString*, JSString*);
    friend JSString* jsString(JSGlobalData&, ExecState*, const String&, const String&);
    friend JSString* jsString(JSGlobalData&, ExecState*, const String&, const String&, const String&);
};

template<>
ALWAYS_INLINE JSString* jsDynamicCast(JSCell* from)
{
	if(from->structure()->typeInfo().type() == JSType::StringType)
		return static_cast<JSString*>(from);
	return 0;
}

ALWAYS_INLINE const StringImpl* JSString::tryGetValueImpl() const
{
    return m_value.impl();
}

ALWAYS_INLINE JSString* asString(JSValue value)
{
	ASSERT(value.asCell()->isString());
	return jsCast<JSString*>(value.asCell());
}

ALWAYS_INLINE JSString* jsEmptyString(JSGlobalData& vm)
{
	return vm.smallStrings.emptyString();
}

ALWAYS_INLINE JSString* jsEmptyString(JSGlobalData* vm)
{
	return jsEmptyString(*vm);
}

ALWAYS_INLINE JSString* jsSingleCharacterString(JSGlobalData& vm, UChar c)
{
	if (c <= maxSingleCharacterString)
		return vm.smallStrings.singleCharacterString(c);
	return JSString::create(vm, String(&c, 1).impl());
}

ALWAYS_INLINE JSString* jsSingleCharacterString(JSGlobalData* vm, UChar c)
{
	return jsSingleCharacterString(*vm, c);
}

ALWAYS_INLINE JSString* jsNontrivialString(JSGlobalData& vm, const String& s)
{
	ASSERT(s.length() > 1);
	return JSString::create(vm, s.impl());
}

ALWAYS_INLINE JSString* jsNontrivialString(JSGlobalData& vm, const ASCIILiteral& s)
{
	ASSERT(s.length() > 1);
	return JSString::create(vm, StringImpl::create(s));
}

ALWAYS_INLINE JSString* jsNontrivialString(JSGlobalData* vm, const String& s)
{
	return jsNontrivialString(*vm, s);
}

ALWAYS_INLINE JSString* jsNontrivialString(JSGlobalData& vm, PassRefPtr<StringImpl> impl)
{
	ASSERT(impl->length() > 1);
	return JSString::create(vm, impl);
}

ALWAYS_INLINE JSString* jsNontrivialString(JSGlobalData* vm, PassRefPtr<StringImpl> impl)
{
	return jsNontrivialString(*vm, impl);
}

ALWAYS_INLINE Identifier JSString::toIdentifier(VM& vm, ExecState* exec) const
{
    return Identifier::fromString(vm, toAtomicString(exec));
}

ALWAYS_INLINE AtomicString JSString::toAtomicString(ExecState* exec) const
{
    if (isRope())
        static_cast<const JSRopeString*>(this)->resolveRopeToAtomicString(exec);
    return AtomicString(m_value);
}

ALWAYS_INLINE PassRefPtr<AtomicStringImpl> JSString::toExistingAtomicString(ExecState* exec) const
{
    if (isRope())
        return static_cast<const JSRopeString*>(this)->resolveRopeToExistingAtomicString(exec);
    return AtomicString::find(*m_value.impl());
}

ALWAYS_INLINE const String& JSString::value(ExecState* nullOrExecForOOM) const
{
	if (isRope())
		static_cast<const JSRopeString*>(this)->resolveRope(nullOrExecForOOM);
	return m_value;
}

ALWAYS_INLINE const String& JSString::value(ExecState& exec) const
{
	return value(&exec);
}

ALWAYS_INLINE const String& JSString::tryGetValue() const
{
    if (isRope()) {
        // Pass nullptr for the ExecState so that resolveRope does not throw in the event of an OOM error.
        static_cast<const JSRopeString*>(this)->resolveRope(nullptr);
    }
	return m_value;
}

ALWAYS_INLINE const String& JSString::tryGetValueWithoutGC() const
{
    if (isRope()) {
        // Pass nullptr for the JSGlobalObject so that resolveRope does not throw in the event of an OOM error.
        static_cast<const JSRopeString*>(this)->resolveRopeWithoutGC();
    }
	return m_value;
}

ALWAYS_INLINE JSString* JSString::getIndex(VM& vm, ExecState* exec, unsigned i)
{
    ThrowScope scope(vm);
	ASSERT(canGetIndex(i));
    StringView view = unsafeView(exec);
    RETURN_IF_EXCEPTION(scope, vm.smallStrings.emptyString());
    return jsSingleCharacterString(vm, view[i]);
}

ALWAYS_INLINE JSString* jsString(JSGlobalData& vm, const String& s)
{
	int size = s.length();
	if (!size)
		return vm.smallStrings.emptyString();
	if (size == 1) {
        UChar c = s.uncheckedCharacterAt(0);
		if (c <= maxSingleCharacterString)
			return vm.smallStrings.singleCharacterString(c);
	}
	return JSString::create(vm, s.impl());
}

ALWAYS_INLINE JSString* jsSubstring(JSGlobalData& vm, ExecState* exec, JSString* s, unsigned offset, unsigned length)
{
	ASSERT(offset <= static_cast<unsigned>(s->length()));
	ASSERT(length <= static_cast<unsigned>(s->length()));
	ASSERT(offset + length <= static_cast<unsigned>(s->length()));
	if (!length)
		return vm.smallStrings.emptyString();
    if (!offset && length == s->length())
        return s;
    return JSRopeString::create(vm, exec, s, offset, length);
}

inline JSString* jsSubstringOfResolved(VM& vm, JSString* s, unsigned offset, unsigned length)
{
    ASSERT(offset <= static_cast<unsigned>(s->length()));
    ASSERT(length <= static_cast<unsigned>(s->length()));
    ASSERT(offset + length <= static_cast<unsigned>(s->length()));
    if (!length)
        return vm.smallStrings.emptyString();
    if (!offset && length == s->length())
        return s;
    return JSRopeString::createSubstringOfResolved(vm, s, offset, length);
}

ALWAYS_INLINE JSString* jsSubstring(JSGlobalData& vm, const String& s, unsigned offset, unsigned length)
{
	ASSERT(offset <= static_cast<unsigned>(s.length()));
	ASSERT(length <= static_cast<unsigned>(s.length()));
	ASSERT(offset + length <= static_cast<unsigned>(s.length()));
	if (!length)
		return vm.smallStrings.emptyString();
	if (length == 1) {
        UChar c = s.characterAt(offset);
		if (c <= maxSingleCharacterString)
			return vm.smallStrings.singleCharacterString(c);
	}
	return JSString::createHasOtherOwner(vm, StringImpl::create(s.impl(), offset, length));
}

ALWAYS_INLINE JSString* jsSubstring(JSGlobalData* vm, const String& s, unsigned offset, unsigned length)
{
	return jsSubstring(*vm, s, offset, length);
}

ALWAYS_INLINE JSString* jsOwnedString(JSGlobalData& vm, const String& s)
{
	int size = s.length();
	if (!size)
		return vm.smallStrings.emptyString();
	if (size == 1) {
        UChar c = s.uncheckedCharacterAt(0);
		if (c <= maxSingleCharacterString)
			return vm.smallStrings.singleCharacterString(c);
	}
	return JSString::createHasOtherOwner(vm, s.impl());
}

ALWAYS_INLINE JSString* jsOwnedString(JSGlobalData* vm, const String& s)
{
	return jsOwnedString(*vm, s);
}

ALWAYS_INLINE JSRopeString* jsStringBuilder(JSGlobalData& vm)
{
	return JSRopeString::createNull(vm);
}

ALWAYS_INLINE JSRopeString* jsStringBuilder(JSGlobalData* vm)
{
	return JSRopeString::createNull(*vm);
}

JS_EXPORT_PRIVATE JSString* jsStringWithCacheSlowCase(VM&, StringImpl&);

JSString* jsStringWithCache(VM& vm, const String& s);

ALWAYS_INLINE JSString* jsStringWithCache(ExecState* state, const String& s)
{
	return jsStringWithCache(state->vm(), s);
}

ALWAYS_INLINE JSString* jsStringWithCache(ExecState& state, const String& s)
{
	return jsStringWithCache(state.vm(), s);
}

template<bool maybeIndex>
ALWAYS_INLINE bool JSString::getStringPropertySlot(JSGlobalData& vm, ExecState* exec, const PropertyName& propertyName, PropertySlot& slot)
{
	ThrowScope scope(vm);

	if (propertyName == vm.propertyNames->length) {
		slot.setValue(this, DontEnum | DontDelete | ReadOnly, jsNumber(length()));
		return true;
	}

	if(maybeIndex) {
		Optional<uint32_t> index = parseIndex(propertyName);
		if (index && index.value() < length()) {
			JSString* value = getIndex(vm, exec, index.value());
			RETURN_IF_EXCEPTION(scope, false);
			slot.setValue(this, DontDelete | ReadOnly, value);
			return true;
		}
	}

	return false;
}

ALWAYS_INLINE bool JSString::getStringPropertySlot(JSGlobalData& vm, ExecState* exec, unsigned propertyName, PropertySlot& slot)
{
	ThrowScope scope(vm);

	if (propertyName < length()) {
		JSString* value = getIndex(vm, exec, propertyName);
		RETURN_IF_EXCEPTION(scope, false);
		slot.setValue(this, DontDelete | ReadOnly, value);
		return true;
	}

	return false;
}

ALWAYS_INLINE bool isJSString(const JSCell* cell)
{
	return cell->structure()->typeInfo().type() == StringType;
}

ALWAYS_INLINE bool isJSString(JSValue v) { return v.isCell() && isJSString(v.asCell()); }

ALWAYS_INLINE JSValue JSString::toPrimitive(ExecState*, PreferredPrimitiveType) const
{
	return const_cast<JSString*>(this);
}

ALWAYS_INLINE StringView JSRopeString::unsafeView(ExecState* state) const
{
    if (isSubstring()) {
        if (is8Bit())
            return StringView(substringBase()->m_value.characters8() + substringOffset(), length());
        return StringView(substringBase()->m_value.characters16() + substringOffset(), length());
    }
    resolveRope(state);
    return m_value;
}

ALWAYS_INLINE StringViewWithUnderlyingString JSRopeString::viewWithUnderlyingString(ExecState* state) const
{
    if (isSubstring()) {
        String& base = substringBase()->m_value;
        if (is8Bit())
            return StringViewWithUnderlyingString(StringView(base.characters8() + substringOffset(), length()), base);
        return StringViewWithUnderlyingString(StringView(base.characters16() + substringOffset(), length()), base);
    }
    resolveRope(state);
    return StringViewWithUnderlyingString(m_value, m_value);
}

ALWAYS_INLINE StringView JSString::unsafeView(ExecState* exec) const
{
    if (isRope())
        return static_cast<const JSRopeString*>(this)->unsafeView(exec);
    return m_value;
}

ALWAYS_INLINE StringViewWithUnderlyingString JSString::viewWithUnderlyingString(ExecState* exec) const
{
    if (isRope())
        return static_cast<const JSRopeString&>(*this).viewWithUnderlyingString(exec);
    return StringViewWithUnderlyingString(m_value, m_value);
}

ALWAYS_INLINE bool JSString::isSubstring() const
{
    return isRope() && static_cast<const JSRopeString*>(this)->isSubstring();
}

// --- JSValue inlines ----------------------------

inline bool JSValue::toBoolean(ExecState* exec) const
{
	if (isInt32())
		return asInt32();
	if (isDouble())
		return asDouble() > 0.0 || asDouble() < 0.0; // false for NaN
	if (isCell())
		return asCell()->toBoolean(exec);
#if USE(BIGINT32)
    if (isBigInt32())
        return !!bigInt32AsInt32();
#endif
	return isTrue(); // false, null, and undefined all convert to false.
}

inline JSString* JSValue::toString(ExecState* exec) const
{
	if (isString())
        return asString(asCell());
    bool returnEmptyStringOnError = true;
    return toStringSlowCase(exec, returnEmptyStringOnError);
}

inline JSString* JSValue::toStringOrNull(ExecState* exec) const
{
    if (isString())
        return asString(asCell());
    bool returnEmptyStringOnError = false;
    return toStringSlowCase(exec, returnEmptyStringOnError);
}

inline String JSValue::toWTFString(ExecState* exec) const
{
	if (isString())
		return static_cast<JSString*>(asCell())->value(exec);
	return toWTFStringSlowCase(exec);
}

} // namespace JSC
