/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#pragma once

#include "BatchedTransitionOptimizer.h"
#include "CallFrame.h"
#include "CustomGetterSetter.h"
#include "IdentifierInlines.h"
#include "Intrinsic.h"
#include "JSGlobalObject.h"
#include "PropertySlot.h"
#include "PutPropertySlot.h"
#include "Reject.h"
#include <stdio.h>
#include <wtf/Assertions.h>
#include <wtf/text/ASCIILiteral.h>

namespace JSC {

struct CompactHashIndex {
    const int16_t value;
    const int16_t next;
};

typedef PropertySlot::GetValueFunc GetFunction;
typedef PutPropertySlot::PutValueFunc PutFunction;
typedef FunctionExecutable* (*BuiltinGenerator)(JSGlobalObject*);

// Hash table generated by the create_hash_table script.
struct HashTableValue {
	ASCIILiteral m_key; // property name
    unsigned m_attributes : 24; // JSObject attributes
    unsigned m_intrinsic : 8;
    union ValueStorage {
        constexpr ValueStorage(intptr_t value1, intptr_t value2)
            : value1(value1)
            , value2(value2)
        { }
        constexpr ValueStorage(long long constant)
            : constant(constant)
        { }

        struct {
            intptr_t value1;
            intptr_t value2;
        };
        long long constant;
    } m_values;

	unsigned attributes() const { return m_attributes; }

	Intrinsic intrinsic() const { ASSERT(m_attributes & Function); return static_cast<Intrinsic>(m_intrinsic); }

	BuiltinGenerator builtinGenerator() const { ASSERT(m_attributes & Builtin); return reinterpret_cast<BuiltinGenerator>(m_values.value1); }
	NativeFunction function() const { ASSERT(m_attributes & Function); return reinterpret_cast<NativeFunction>(m_values.value1); }
	unsigned char functionLength() const { ASSERT(m_attributes & Function); return static_cast<unsigned char>(m_values.value2); }

    GetFunction propertyGetter() const { ASSERT(!(m_attributes & BuiltinOrFunctionOrAccessorOrConstant)); return reinterpret_cast<GetFunction>(m_values.value1); }
    PutFunction propertyPutter() const { ASSERT(!(m_attributes & BuiltinOrFunctionOrAccessorOrConstant)); return reinterpret_cast<PutFunction>(m_values.value2); }

    NativeFunction accessorGetter() const { ASSERT(m_attributes & Accessor); return reinterpret_cast<NativeFunction>(m_values.value1); }
    NativeFunction accessorSetter() const { ASSERT(m_attributes & Accessor); return reinterpret_cast<NativeFunction>(m_values.value2); }
    BuiltinGenerator builtinAccessorGetterGenerator() const;
	BuiltinGenerator builtinAccessorSetterGenerator() const;

    long long constantInteger() const { ASSERT(m_attributes & ConstantInteger); return m_values.constant; }

	intptr_t lexerValue() const { ASSERT(!m_attributes); return m_values.value1; }
};

struct HashTable {
    int numberOfValues;
    int indexMask;
    uint8_t seenPropertyAttributes;

	const HashTableValue* values; // Fixed values generated by script.
    const CompactHashIndex* index;

	// Find an entry in the table, and return the entry.
	ALWAYS_INLINE const HashTableValue* entry(const PropertyName& propertyName) const
	{
        if (propertyName.isSymbol())
            return 0;

        auto* uid = propertyName.uid();
        if (!uid)
            return 0;

		int indexEntry = IdentifierRepHash::hash(uid) & indexMask;
		int valueIndex = index[indexEntry].value;
		if (valueIndex == -1)
			return 0;

		while (true) {
			if (WTF::equal(*uid, values[valueIndex].m_key))
				return &values[valueIndex];

			indexEntry = index[indexEntry].next;
			if (indexEntry == -1)
				return 0;
			valueIndex = index[indexEntry].value;
			ASSERT(valueIndex != -1);
		};
	}

	class ConstIterator {
	public:
		ALWAYS_INLINE ConstIterator(const HashTable* table, int position)
			: m_table(table)
			, m_position(position)
		{
			skipInvalidKeys();
		}

		ALWAYS_INLINE const HashTableValue* value() const
        {
            return &m_table->values[m_position];
        }

		ALWAYS_INLINE const HashTableValue& operator*() const { return *value(); }

		ALWAYS_INLINE const ASCIILiteral& key() const
        {
            return m_table->values[m_position].m_key;
        }

		ALWAYS_INLINE const HashTableValue* operator->() const
        {
            return value();
        }

		ALWAYS_INLINE bool operator!=(const ConstIterator& other) const
		{
			ASSERT(m_table == other.m_table);
			return m_position != other.m_position;
		}

		ALWAYS_INLINE ConstIterator& operator++()
		{
			ASSERT(m_position < m_table->numberOfValues);
			++m_position;
			skipInvalidKeys();
			return *this;
		}

	private:
		ALWAYS_INLINE void skipInvalidKeys()
		{
            ASSERT(m_position <= m_table->numberOfValues);
            while (m_position < m_table->numberOfValues && !m_table->values[m_position].m_key)
                ++m_position;
            ASSERT(m_position <= m_table->numberOfValues);
		}

		const HashTable* m_table;
		int m_position;
	};

	ALWAYS_INLINE ConstIterator begin() const
	{
		return ConstIterator(this, 0);
	}
	ALWAYS_INLINE ConstIterator end() const
	{
		return ConstIterator(this, numberOfValues);
	}
};

JS_EXPORT_PRIVATE bool setUpStaticFunctionSlot(VM&, const HashTableValue*, JSObject* thisObject, PropertyName, PropertySlot&);
JS_EXPORT_PRIVATE void reifyStaticAccessor(VM&, const HashTableValue&, JSObject& thisObject, PropertyName);

inline BuiltinGenerator HashTableValue::builtinAccessorGetterGenerator() const
{
    ASSERT(m_attributes & Accessor);
    ASSERT(m_attributes & Builtin);
    return reinterpret_cast<BuiltinGenerator>(m_values.value1);
}

inline BuiltinGenerator HashTableValue::builtinAccessorSetterGenerator() const
{
    ASSERT(m_attributes & Accessor);
    ASSERT(m_attributes & Builtin);
    return reinterpret_cast<BuiltinGenerator>(m_values.value2);
}

inline bool getStaticPropertySlotFromTable(VM& vm, const HashTable& table, JSObject* thisObject, const PropertyName& propertyName, PropertySlot& slot)
{
    if (thisObject->staticPropertiesReified())
        return false;

    auto* entry = table.entry(propertyName);
    if (!entry)
        return false;

    if (entry->attributes() & BuiltinOrFunctionOrAccessor)
        return setUpStaticFunctionSlot(vm, entry, thisObject, propertyName, slot);

    if (entry->attributes() & ConstantInteger) {
        slot.setValue(thisObject, attributesForStructure(entry->attributes()), jsNumber(entry->constantInteger()));
        return true;
    }

#if 0
    if (entry->attributes() & PropertyAttribute::DOMJITAttribute) {
        ASSERT_WITH_MESSAGE(entry->attributes() & PropertyAttribute::ReadOnly, "DOMJITAttribute supports readonly attributes currently.");
        const DOMJIT::GetterSetter* domJIT = entry->domJIT();
        slot.setCacheableCustom(thisObject, attributesForStructure(entry->attributes()), domJIT->getter(), nullptr, DOMAttributeAnnotation { classInfo, domJIT });
        return true;
    }

    if (entry->attributes() & PropertyAttribute::DOMAttribute) {
        slot.setCacheableCustom(thisObject, attributesForStructure(entry->attributes()), entry->propertyGetter(), entry->propertyPutter(), DOMAttributeAnnotation { classInfo, nullptr });
        return true;
    }
#endif

    slot.setCacheableCustom(thisObject, attributesForStructure(entry->attributes()), entry->propertyGetter(), entry->propertyPutter());
    return true;
}

inline void reifyStaticProperty(JSGlobalData& vm, PropertyName propertyName, const HashTableValue& value, JSObject& thisObject)
{
	if (value.attributes() & Builtin) {
		if (value.attributes() & Accessor)
        	reifyStaticAccessor(vm, value, thisObject, propertyName);
        else
        	thisObject.putDirectBuiltinFunction(vm, thisObject.globalObject(), propertyName, value.builtinGenerator()(thisObject.globalObject()), attributesForStructure(value.attributes()));
        return;
	}

	if (value.attributes() & Function) {
		thisObject.putDirectNativeFunction(
			vm, thisObject.globalObject(), propertyName, value.functionLength(),
			value.function(), value.intrinsic(), attributesForStructure(value.attributes()));
		return;
	}

	if (value.attributes() & ConstantInteger) {
		thisObject.putDirect(vm, propertyName, jsNumber(value.constantInteger()), attributesForStructure(value.attributes()));
		return;
	}

	if (value.attributes() & Accessor) {
        reifyStaticAccessor(vm, value, thisObject, propertyName);
		return;
	}

	CustomGetterSetter* customGetterSetter = CustomGetterSetter::create(vm, value.propertyGetter(), value.propertyPutter());
	thisObject.putDirectCustomAccessor(vm, propertyName, customGetterSetter, attributesForStructure(value.attributes()));
}

template<unsigned numberOfValues>
inline void reifyStaticProperties(JSGlobalData& vm, const HashTableValue (&values)[numberOfValues], JSObject& thisObject)
{
    BatchedTransitionOptimizer transitionOptimizer(vm, &thisObject);
    for (int i = 0; i < numberOfValues; ++i) {
    	const HashTableValue& value = values[i];
        if (!value.m_key)
            continue;
        Identifier key = Identifier::fromString(vm, ASCIILiteral(value.m_key));
        reifyStaticProperty(vm, key, value, thisObject);
    }
}

} // namespace JSC
