/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "ObjectConstructor.h"

#include "ButterflyInlineMethods.h"
#include "CallFrameInlines.h"
#include "CopiedSpaceInlineMethods.h"
#include "Error.h"
#include "ExceptionHelpers.h"
#include "JSFunction.h"
#include "JSArray.h"
#include "JSGlobalObject.h"
#include "JSGlobalObjectFunctions.h"
#include "Lookup.h"
#include "ObjectPrototype.h"
#include "Operations.h"
#include "PropertyDescriptor.h"
#include "PropertyNameArray.h"
#include "StackVisitor.h"
#include "Symbol.h"
#include "ThrowScope.h"
#include <wtf/MemoryMonitor.h>

namespace JSC {

EncodedJSValue JSC_HOST_CALL objectConstructorAssign(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorEntries(ExecState*);
EncodedJSValue JSC_HOST_CALL objectConstructorValues(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorGetPrototypeOf(ExecState*);
EncodedJSValue JSC_HOST_CALL objectConstructorSetPrototypeOf(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperty(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperties(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertyNames(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorCreate(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorSeal(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorFreeze(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorPreventExtensions(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorIsSealed(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorIsFrozen(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorIsExtensible(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectConstructorHasOwn(ExecState*);

}

#include "ObjectConstructor.lut.h"

namespace JSC {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(ObjectConstructor);

const ClassInfo ObjectConstructor::s_info = { "Function", &InternalFunction::s_info, &objectConstructorTable, CREATE_METHOD_TABLE(ObjectConstructor) };

/* Source for ObjectConstructor.lut.h
@begin objectConstructorTable
  getPrototypeOf            objectConstructorGetPrototypeOf             DontEnum|Function 1 ObjectGetPrototypeOfIntrinsic
  setPrototypeOf            objectConstructorSetPrototypeOf             DontEnum|Function 2
  getOwnPropertyDescriptor  objectConstructorGetOwnPropertyDescriptor   DontEnum|Function 2
  getOwnPropertyDescriptors objectConstructorGetOwnPropertyDescriptors  DontEnum|Function 1
  getOwnPropertyNames       objectConstructorGetOwnPropertyNames        DontEnum|Function 1
  getOwnPropertySymbols     objectConstructorGetOwnPropertySymbols      DontEnum|Function 1
  keys                      objectConstructorKeys                       DontEnum|Function 1
  defineProperty            objectConstructorDefineProperty             DontEnum|Function 3
  defineProperties          objectConstructorDefineProperties           DontEnum|Function 2
  create                    objectConstructorCreate                     DontEnum|Function 2 ObjectCreateIntrinsic
  seal                      objectConstructorSeal                       DontEnum|Function 1
  freeze                    objectConstructorFreeze                     DontEnum|Function 1
  preventExtensions         objectConstructorPreventExtensions          DontEnum|Function 1
  isSealed                  objectConstructorIsSealed                   DontEnum|Function 1
  isFrozen                  objectConstructorIsFrozen                   DontEnum|Function 1
  isExtensible              objectConstructorIsExtensible               DontEnum|Function 1
  is                        objectConstructorIs                         DontEnum|Function 2 ObjectIsIntrinsic
  assign                    objectConstructorAssign                     DontEnum|Function 2
  values                    objectConstructorValues                     DontEnum|Function 1
  entries                   objectConstructorEntries                    DontEnum|Function 1
  fromEntries               JSBuiltin                                   DontEnum|Function 1
@end
*/

static EncodedJSValue JSC_HOST_CALL callObjectConstructor(ExecState*);
static EncodedJSValue JSC_HOST_CALL constructWithObjectConstructor(ExecState*);

ObjectConstructor::ObjectConstructor(VM& vm, Structure* structure)
    : InternalFunction(vm, structure, callObjectConstructor, constructWithObjectConstructor)
{
}

void ObjectConstructor::finishCreation(VM& vm, JSGlobalObject* globalObject, ObjectPrototype* objectPrototype)
{
    Base::finishCreation(vm, String::fromLiteral(objectPrototype->classInfo()->className));
    // ECMA 15.2.3.1
    putDirectWithoutTransition(vm, vm.propertyNames->prototype, objectPrototype, DontEnum | DontDelete | ReadOnly);
    putDirectWithoutTransition(vm, vm.propertyNames->length, jsNumber(1), ReadOnly | DontEnum);

    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().createPrivateName(), objectConstructorCreate, DontEnum, 2);
	JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().definePropertyPrivateName(), objectConstructorDefineProperty, DontEnum, 3);
	if (Options::useHasOwn())
		JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->hasOwn, objectConstructorHasOwn, static_cast<unsigned>(PropertyAttribute::DontEnum), 2);
}

// ES 19.1.1.1 Object([value])
static ALWAYS_INLINE JSObject* constructObject(ExecState* exec, JSValue newTarget)
{
	JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    VM& vm = globalObject->vm();
    ObjectConstructor* objectConstructor = jsCast<ObjectConstructor*>(exec->callee());
    auto scope = DECLARE_THROW_SCOPE(vm);

    // We need to check newTarget condition in this caller side instead of InternalFunction::createSubclassStructure side.
    // Since if we found this condition is met, we should not fall into the type conversion in the step 3.

    // 1. If NewTarget is neither undefined nor the active function, then
    if (newTarget && newTarget != objectConstructor) {
        // a. Return ? OrdinaryCreateFromConstructor(NewTarget, "%ObjectPrototype%").
    	JSGlobalObject* functionGlobalObject = getFunctionRealm(exec, globalObject, asObject(newTarget));
		RETURN_IF_EXCEPTION(scope, nullptr);
		Structure* baseStructure = functionGlobalObject->objectStructureForObjectConstructor();
		Structure* objectStructure = InternalFunction::createSubclassStructure(exec, globalObject, asObject(newTarget), baseStructure);
		RETURN_IF_EXCEPTION(vm, nullptr);
        return constructEmptyObject(vm, objectStructure);
    }

    // 2. If value is null, undefined or not supplied, return ObjectCreate(%ObjectPrototype%).
    JSValue argument = exec->argument(0);
	if (argument.isUndefinedOrNull())
        return constructEmptyObject(vm, globalObject->objectStructureForObjectConstructor());

    // 3. Return ToObject(value).
    return argument.toObject(exec, globalObject);
}

static EncodedJSValue JSC_HOST_CALL constructWithObjectConstructor(ExecState* exec)
{
    return JSValue::encode(constructObject(exec, exec->newTarget()));
}

static EncodedJSValue JSC_HOST_CALL callObjectConstructor(ExecState* exec)
{
    return JSValue::encode(constructObject(exec, JSValue()));
}

EncodedJSValue JSC_HOST_CALL objectConstructorGetPrototypeOf(ExecState* exec)
{
    VM& vm = exec->vm();
    return JSValue::encode(exec->argument(0).getPrototype(vm, exec));
}

EncodedJSValue JSC_HOST_CALL objectConstructorSetPrototypeOf(ExecState* exec)
{
    VM& vm = exec->vm();

    JSValue objectValue = exec->argument(0);
    if (objectValue.isUndefinedOrNull())
        return throwVMTypeError(exec, vm, ASCIILiteral("Cannot set prototype of undefined or null"));

    JSValue protoValue = exec->argument(1);
    if (!protoValue.isObject() && !protoValue.isNull())
        return throwVMTypeError(exec, vm, PrototypeValueCanOnlyBeAnObjectOrNullTypeError);

    JSObject* object = objectValue.toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    bool shouldThrowIfCantSet = true;
    bool didSetPrototype = object->setPrototype(vm, exec, protoValue, shouldThrowIfCantSet);
    ASSERT_UNUSED(didSetPrototype, vm.exception() || didSetPrototype);
    return JSValue::encode(objectValue);
}

JSValue objectConstructorGetOwnPropertyDescriptor(ExecState* exec, JSObject* object, const Identifier& propertyName)
{
	VM& vm = exec->vm();
    PropertyDescriptor descriptor;
    if (!object->getOwnPropertyDescriptor(exec, propertyName, descriptor))
        return jsUndefined();
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    JSObject* result = constructObjectFromPropertyDescriptor(exec, descriptor);
    if (!result)
        return jsUndefined();
    return result;
}

JSValue objectConstructorGetOwnPropertyDescriptors(ExecState* exec, JSObject* object)
{
    JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    VM& vm = globalObject->vm();
    PropertyNameArray properties(vm, PropertyNameMode::StringsAndSymbols, PrivateSymbolMode::Exclude);
    object->methodTable()->getOwnPropertyNames(object, exec, properties, DontEnumPropertiesMode::Include);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    JSObject* descriptors = constructEmptyObject(globalObject);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    size_t numProperties = properties.size();
    for (size_t i = 0; i < numProperties; i++) {
    	PropertyDescriptor descriptor;
		bool didGetDescriptor = object->getOwnPropertyDescriptor(exec, properties[i], descriptor);
	    RETURN_IF_EXCEPTION(vm, jsUndefined());

        if (!didGetDescriptor)
            continue;

        JSObject* fromDescriptor = constructObjectFromPropertyDescriptor(exec, descriptor);
        if (!fromDescriptor)
            return jsUndefined();

        PutPropertySlot slot(descriptors);
        descriptors->putOwnDataPropertyMayBeIndex(exec, properties[i], fromDescriptor, slot);
        //scope.assertNoExceptionExceptTermination();
		RETURN_IF_EXCEPTION(vm, jsUndefined());
    }

    return descriptors;
}

EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertyDescriptor(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    Identifier propertyName = exec->argument(1).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(objectConstructorGetOwnPropertyDescriptor(exec, object, propertyName));
}

EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertyDescriptors(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(objectConstructorGetOwnPropertyDescriptors(exec, object));
}

// FIXME: Use the enumeration cache.
EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertyNames(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    JSArray* keys = ownPropertyKeys(exec, object, PropertyNameMode::Strings, DontEnumPropertiesMode::Include);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(keys);
}

// FIXME: Use the enumeration cache.
EncodedJSValue JSC_HOST_CALL objectConstructorGetOwnPropertySymbols(ExecState* exec)
{
    VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    JSArray* keys = ownPropertyKeys(exec, object, PropertyNameMode::Symbols, DontEnumPropertiesMode::Include);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(keys);
}

// FIXME: Use the enumeration cache.
EncodedJSValue JSC_HOST_CALL objectConstructorKeys(ExecState* exec)
{
    VM& vm = exec->vm();
    JSObject* object = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    JSArray* keys = ownPropertyKeys(exec, object, PropertyNameMode::Strings, DontEnumPropertiesMode::Exclude);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(keys);
}

EncodedJSValue JSC_HOST_CALL objectConstructorAssign(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue targetValue = exec->argument(0);
    if (targetValue.isUndefinedOrNull())
        return throwVMTypeError(exec, vm, ASCIILiteral("Object.assign requires that input parameter not be null or undefined"));
    JSObject* target = targetValue.toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // FIXME: Extend this for non JSFinalObject. For example, we would like to use this fast path for function objects too.
    // https://bugs.webkit.org/show_bug.cgi?id=185358
    bool targetCanPerformFastPut = jsDynamicCast<JSFinalObject*>(target) && target->canPerformFastPutInlineExcludingProto(vm);

    Vector<RefPtr<UniquedStringImpl>, 8> properties;
    MarkedArgumentBuffer values;
    unsigned argsCount = exec->argumentCount();
    for (unsigned i = 1; i < argsCount; ++i) {
        JSValue sourceValue = exec->uncheckedArgument(i);
        if (sourceValue.isUndefinedOrNull())
            continue;
        JSObject* source = sourceValue.toObject(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        if (targetCanPerformFastPut) {
            if (!source->staticPropertiesReified()) {
                source->reifyAllStaticProperties(vm);
                RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            }

			if (source->structure(vm)->canPerformFastPropertyEnumeration()) {
				// |source| Structure does not have any getters. And target can perform fast put.
				// So enumerating properties and putting properties are non observable.

				// FIXME: It doesn't seem like we should have to do this in two phases, but
				// we're running into crashes where it appears that source is transitioning
				// under us, and even ends up in a state where it has a null butterfly. My
				// leading hypothesis here is that we fire some value replacement watchpoint
				// that ends up transitioning the structure underneath us.
				// https://bugs.webkit.org/show_bug.cgi?id=187837

				// Do not clear since Vector::clear shrinks the backing store.
				properties.resize(0);
				values.clear();
				source->structure(vm)->forEachProperty(vm, [&] (const PropertyMapEntry& entry) -> bool {
					if (entry.attributes & DontEnum)
						return true;

					PropertyName propertyName(entry.key);
					if (propertyName.isPrivateName())
						return true;

					properties.append(entry.key);
					values.appendWithCrashOnOverflow(source->getDirect(entry.offset));

					return true;
				});

                for (size_t i = 0; i < properties.size(); ++i) {
                    // FIXME: We could put properties in a batching manner to accelerate Object.assign more.
                    // https://bugs.webkit.org/show_bug.cgi?id=185358
                    PutPropertySlot putPropertySlot(target, true);
                    target->putOwnDataProperty(vm, properties[i].get(), values.at(i), putPropertySlot);
                }
                continue;
            }
        }

        // [[GetOwnPropertyNames]], [[Get]] etc. could modify target object and invalidate this assumption.
        // For example, [[Get]] of source object could configure setter to target object. So disable the fast path.
        targetCanPerformFastPut = false;

        PropertyNameArray properties(vm, PropertyNameMode::StringsAndSymbols, PrivateSymbolMode::Exclude);
        source->methodTable(vm)->getOwnPropertyNames(source, exec, properties, DontEnumPropertiesMode::Include);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        unsigned numProperties = properties.size();
        for (unsigned j = 0; j < numProperties; j++) {
            const auto& propertyName = properties[j];
            ASSERT(!propertyName.isPrivateName());

			PropertySlot slot(source, PropertySlot::InternalMethodType::GetOwnProperty);
			bool hasProperty = source->methodTable(vm)->getOwnPropertySlot(source, exec, propertyName, slot);
			RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
			if (!hasProperty)
				continue;
			if (slot.attributes() & DontEnum)
				continue;

			JSValue value;
			if (LIKELY(!slot.isTaintedByOpaqueObject()))
				value = slot.getValue(exec, propertyName);
			else
				value = source->get(vm, exec, propertyName);
			RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

			PutPropertySlot putPropertySlot(target, true);
			target->putInline(exec, vm, propertyName, value, putPropertySlot);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        }
    }
    return JSValue::encode(target);
}

EncodedJSValue JSC_HOST_CALL objectConstructorEntries(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue targetValue = exec->argument(0);
    if (targetValue.isUndefinedOrNull())
        return throwVMTypeError(exec, vm, "Object.entries requires that input parameter not be null or undefined");
    JSObject* target = targetValue.toObject(exec);
	RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSArray* entries = constructEmptyArray(exec, nullptr);
	RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    PropertyNameArray properties(vm, PropertyNameMode::Strings, PrivateSymbolMode::Exclude);
    target->methodTable(vm)->getOwnPropertyNames(target, exec, properties, DontEnumPropertiesMode::Include);
	RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    unsigned index = 0;
    auto append = [&] (ExecState* exec, PropertyName propertyName) {
        PropertySlot slot(target, PropertySlot::InternalMethodType::GetOwnProperty);
        bool hasProperty = target->methodTable(vm)->getOwnPropertySlot(target, exec, propertyName, slot);
        RETURN_IF_EXCEPTION(scope, void());
        if (!hasProperty)
            return;
        if (slot.attributes() & DontEnum)
            return;

        JSValue value;
        if (LIKELY(!slot.isTaintedByOpaqueObject()))
            value = slot.getValue(exec, propertyName);
        else
            value = target->get(vm, exec, propertyName);
        RETURN_IF_EXCEPTION(scope, void());

        JSString* key = jsOwnedString(vm, propertyName.uid());
        JSArray* entry = nullptr;
        {
            ObjectInitializationScope initializationScope(vm);
            DeferGC deferGC(vm.heap);
            JSGlobalObject* lexicalGlobalObject = exec->lexicalGlobalObject();
            if ((entry = JSArray::tryCreateUninitializedRestricted(initializationScope, deferGC, lexicalGlobalObject->arrayStructureForIndexingTypeDuringAllocation(ArrayWithContiguous), 2))) {
                entry->initializeIndex(vm, 0, key);
                entry->initializeIndex(vm, 1, value);
            }
        }
        if (!entry) {
            throwOutOfMemoryError(exec);
            return;
        }
        entries->putDirectIndex(exec, index++, entry);
    };

    for (const auto& propertyName : properties) {
        append(exec, propertyName);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    return JSValue::encode(entries);
}

EncodedJSValue JSC_HOST_CALL objectConstructorValues(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue targetValue = exec->argument(0);
    if (targetValue.isUndefinedOrNull())
        return throwVMTypeError(exec, vm, ASCIILiteral("Object.values requires that input parameter not be null or undefined"));
    JSObject* target = targetValue.toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSArray* values = constructEmptyArray(exec, 0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    PropertyNameArray properties(vm, PropertyNameMode::Strings, PrivateSymbolMode::Exclude);
    target->methodTable(vm)->getOwnPropertyNames(target, exec, properties, DontEnumPropertiesMode::Include);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    unsigned index = 0;
    auto append = [&] (ExecState* exec, PropertyName propertyName) {
		PropertySlot slot(target, PropertySlot::InternalMethodType::GetOwnProperty);
		bool hasProperty = target->methodTable(vm)->getOwnPropertySlot(target, exec, propertyName, slot);
		RETURN_IF_EXCEPTION(vm, void());
		if (!hasProperty)
			return;
		if (slot.attributes() & DontEnum)
			return;

		JSValue value;
		if (LIKELY(!slot.isTaintedByOpaqueObject()))
			value = slot.getValue(exec, propertyName);
		else
			value = target->get(vm, exec, propertyName);
		RETURN_IF_EXCEPTION(vm, void());

		values->putDirectIndex(exec, index++, value);
    };

    for (const auto& propertyName : properties) {
        append(exec, propertyName);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    return JSValue::encode(values);
}

// https://tc39.github.io/ecma262/#sec-topropertydescriptor
bool toPropertyDescriptor(ExecState* exec, JSValue in, PropertyDescriptor& desc)
{
    ASSERT(desc.isEmpty());
	VM& vm = exec->vm();
    if (!in.isObject()) {
        vm.throwException(exec, createTypeError(exec, "Property description must be an object."));
        return false;
    }
    JSObject* description = asObject(in);

    bool hasProperty = description->hasProperty(exec, vm.propertyNames->enumerable);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(vm, exec, vm.propertyNames->enumerable);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setEnumerable(value.toBoolean(exec));
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->configurable);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(vm, exec, vm.propertyNames->configurable);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setConfigurable(value.toBoolean(exec));
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->value);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(vm, exec, vm.propertyNames->value);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setValue(value);
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->writable);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue value = description->get(vm, exec, vm.propertyNames->writable);
        RETURN_IF_EXCEPTION(vm, false);
        desc.setWritable(value.toBoolean(exec));
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->get);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue get = description->get(vm, exec, vm.propertyNames->get);
        RETURN_IF_EXCEPTION(vm, false);
        if (!get.isUndefined() && !get.isCallable()) {
			vm.throwException(exec, createTypeError(exec, "Getter must be a function."));
			return false;
        }
        desc.setGetter(get);
    } else
        RETURN_IF_EXCEPTION(vm, false);

    hasProperty = description->hasProperty(exec, vm.propertyNames->set);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        JSValue set = description->get(vm, exec, vm.propertyNames->set);
        RETURN_IF_EXCEPTION(vm, false);
        if (!set.isUndefined() && !set.isCallable(vm)) {
			vm.throwException(exec, createTypeError(exec, "Setter must be a function."));
			return false;
        }
        desc.setSetter(set);
    } else
        RETURN_IF_EXCEPTION(vm, false);

    if (!desc.isAccessorDescriptor())
        return true;

    if (desc.value()) {
        vm.throwException(exec, createTypeError(exec, "Invalid property.  'value' present on property with getter or setter."));
        return false;
    }

    if (desc.writablePresent()) {
        vm.throwException(exec, createTypeError(exec, "Invalid property.  'writable' present on property with getter or setter."));
        return false;
    }
    return true;
}

EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperty(ExecState* exec)
{
	VM& vm = exec->vm();
	if(UNLIKELY(WTF::isLowMemory()))
		return JSValue::encode(vm.throwException(exec, createOutOfMemoryError(exec)));
    if (UNLIKELY(!exec->argument(0).isObject()))
        return JSValue::encode(vm.throwException(exec, createTypeError(exec, "Properties can only be defined on Objects.")));
    JSObject* O = asObject(exec->argument(0));
    Identifier propertyName = exec->argument(1).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    PropertyDescriptor descriptor;
    if (UNLIKELY(!toPropertyDescriptor(exec, exec->argument(2), descriptor)))
        return JSValue::encode(jsNull());
    ASSERT((descriptor.attributes() & Accessor) || (!descriptor.isAccessorDescriptor()));
    ASSERT(!vm.exception());
    O->methodTable()->defineOwnProperty(O, exec, propertyName, descriptor, true);
    return JSValue::encode(O);
}

static JSValue defineProperties(ExecState* exec, VM& vm, JSObject* object, JSObject* properties)
{
	ThrowScope scope(vm);

    PropertyNameArray propertyNames(vm, PropertyNameMode::StringsAndSymbols, PrivateSymbolMode::Exclude);
    asObject(properties)->methodTable()->getOwnPropertyNames(asObject(properties), exec, propertyNames, DontEnumPropertiesMode::Exclude);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    size_t numProperties = propertyNames.size();
    Vector<PropertyDescriptor> descriptors;
    MarkedArgumentBuffer markBuffer;
#define RETURN_IF_EXCEPTION_CLEARING_OVERFLOW(value) do { \
    if (scope.exception()) { \
        markBuffer.overflowCheckNotNeeded(); \
        return value; \
    } \
} while (false)
    for (size_t i = 0; i < numProperties; i++) {
        JSValue prop = properties->get(vm, exec, propertyNames[i]);
        RETURN_IF_EXCEPTION_CLEARING_OVERFLOW(jsUndefined());
        PropertyDescriptor descriptor;
        toPropertyDescriptor(exec, prop, descriptor);
        RETURN_IF_EXCEPTION_CLEARING_OVERFLOW(jsUndefined());
        descriptors.append(descriptor);
        // Ensure we mark all the values that we're accumulating
        if (descriptor.isDataDescriptor() && descriptor.value())
            markBuffer.append(descriptor.value());
        if (descriptor.isAccessorDescriptor()) {
            if (descriptor.getter())
                markBuffer.append(descriptor.getter());
            if (descriptor.setter())
                markBuffer.append(descriptor.setter());
        }
    }
    RELEASE_ASSERT(!markBuffer.hasOverflowed());
#undef RETURN_IF_EXCEPTION_CLEARING_OVERFLOW
    for (size_t i = 0; i < numProperties; i++) {
    	Identifier propertyName = propertyNames[i];
		ASSERT(!propertyName.isPrivateName());
		object->methodTable(vm)->defineOwnProperty(object, exec, propertyName, descriptors[i], true);
		RETURN_IF_EXCEPTION(vm, jsUndefined());
    }
    return object;
}

EncodedJSValue JSC_HOST_CALL objectConstructorDefineProperties(ExecState* exec)
{
	VM& vm = exec->vm();
    if (!exec->argument(0).isObject())
        return throwVMError(exec, vm, createTypeError(exec, "Properties can only be defined on Objects."));
    JSObject* targetObj = asObject(exec->argument(0));
	JSObject* props = exec->argument(1).toObject(exec);
    ASSERT(!!scope.exception() == !props);
    if (UNLIKELY(!props))
        return encodedJSUndefined();
	return JSValue::encode(defineProperties(exec, vm, targetObj, props));
}

EncodedJSValue JSC_HOST_CALL objectConstructorCreate(ExecState* exec)
{
    JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    VM& vm = globalObject->vm();
    JSValue proto = exec->argument(0);
    if (!proto.isObject() && !proto.isNull())
        return throwVMError(exec, vm, createTypeError(exec, "Object prototype may only be an Object or null."));
    JSObject* newObject = proto.isObject()
		? constructEmptyObject(globalObject, asObject(proto))
		: constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure());
   if (exec->argument(1).isUndefined())
        return JSValue::encode(newObject);
   JSObject* properties = exec->uncheckedArgument(1).toObject(exec);
   RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

   return JSValue::encode(defineProperties(exec, vm, newObject, properties));
}

enum class IntegrityLevel {
    Sealed,
    Frozen
};

template<IntegrityLevel level>
bool setIntegrityLevel(ExecState* exec, VM& vm, JSObject* object)
{
    // See https://tc39.github.io/ecma262/#sec-setintegritylevel.
	ThrowScope scope(vm);

    bool success = object->methodTable(vm)->preventExtensions(object, exec);
    RETURN_IF_EXCEPTION(scope, false);
    if (UNLIKELY(!success))
        return false;

    PropertyNameArray properties(vm, PropertyNameMode::StringsAndSymbols, PrivateSymbolMode::Exclude);
    object->methodTable(vm)->getOwnPropertyNames(object, exec, properties, DontEnumPropertiesMode::Include);
    RETURN_IF_EXCEPTION(scope, false);

    PropertyNameArray::const_iterator end = properties.end();
    for (PropertyNameArray::const_iterator iter = properties.begin(); iter != end; ++iter) {
        const Identifier& propertyName = *iter;
        ASSERT(!propertyName.isPrivateName());

        PropertyDescriptor desc;
        if (level == IntegrityLevel::Sealed)
            desc.setConfigurable(false);
        else {
            PropertyDescriptor currentDesc;
            bool hasPropertyDescriptor = object->getOwnPropertyDescriptor(exec, propertyName, currentDesc);
            RETURN_IF_EXCEPTION(scope, false);
            if (!hasPropertyDescriptor)
                continue;

            if (!currentDesc.isAccessorDescriptor())
                desc.setWritable(false);

            desc.setConfigurable(false);
        }

        object->methodTable(vm)->defineOwnProperty(object, exec, propertyName, desc, true);
        RETURN_IF_EXCEPTION(scope, false);
    }
    return true;
}

template<IntegrityLevel level>
bool testIntegrityLevel(ExecState* exec, VM& vm, JSObject* object)
{
	ThrowScope scope(vm);

    // 1. Assert: Type(O) is Object.
    // 2. Assert: level is either "sealed" or "frozen".

    // 3. Let status be ?IsExtensible(O).
    bool status = object->isExtensible(exec);
    RETURN_IF_EXCEPTION(scope, false);

    // 4. If status is true, return false.
    if (status)
        return false;

    // 6. Let keys be ? O.[[OwnPropertyKeys]]().
    PropertyNameArray keys(vm, PropertyNameMode::StringsAndSymbols, PrivateSymbolMode::Exclude);
    object->methodTable(vm)->getOwnPropertyNames(object, exec, keys, DontEnumPropertiesMode::Include);
    RETURN_IF_EXCEPTION(scope, false);

    // 7. For each element k of keys, do
    PropertyNameArray::const_iterator end = keys.end();
    for (PropertyNameArray::const_iterator iter = keys.begin(); iter != end; ++iter) {
        const Identifier& propertyName = *iter;
        ASSERT(!propertyName.isPrivateName());

        // a. Let currentDesc be ? O.[[GetOwnProperty]](k)
        PropertyDescriptor desc;
        bool didGetDescriptor = object->getOwnPropertyDescriptor(exec, propertyName, desc);
        RETURN_IF_EXCEPTION(scope, false);

        // b. If currentDesc is not undefined, then
        if (!didGetDescriptor)
            continue;

        // i. If currentDesc.[[Configurable]] is true, return false.
        if (desc.configurable())
            return false;

        // ii. If level is "frozen" and IsDataDescriptor(currentDesc) is true, then
        // 1. If currentDesc.[[Writable]] is true, return false.
        if (level == IntegrityLevel::Frozen && desc.isDataDescriptor() && desc.writable())
            return false;
    }

    return true;
}

JSObject* objectConstructorSeal(ExecState* exec, JSObject* object)
{
    VM& vm = exec->vm();
    ThrowScope scope(vm);

    if (jsDynamicCast<JSFinalObject*>(object) && !hasIndexedProperties(object->structure()->indexingType())) {
        object->seal(vm);
        return object;
    }

    bool success = setIntegrityLevel<IntegrityLevel::Sealed>(exec, vm, object);
    RETURN_IF_EXCEPTION(scope, 0);
    if (UNLIKELY(!success)) {
        throwTypeError(exec, vm, ASCIILiteral("Unable to prevent extension in Object.seal"));
        return 0;
    }

    return object;
}

EncodedJSValue JSC_HOST_CALL objectConstructorSeal(ExecState* exec)
{
    VM& vm = exec->vm();
    ThrowScope scope(vm);

    // 1. If Type(O) is not Object, return O.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(obj);
    JSObject* object = asObject(obj);

    if (object->structure()->typeInfo().isFinalObject() && !hasIndexedProperties(object->structure()->indexingType())) {
        object->seal(vm);
        return JSValue::encode(obj);
    }

    bool success = setIntegrityLevel<IntegrityLevel::Sealed>(exec, vm, object);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (UNLIKELY(!success)) {
        throwTypeError(exec, vm, ASCIILiteral("Unable to prevent extension in Object.seal"));
        return encodedJSUndefined();
    }

    return JSValue::encode(obj);
}

JSObject* objectConstructorFreeze(ExecState* exec, JSObject* object)
{
	VM& vm = exec->vm();
    ThrowScope scope(vm);

    if (object->structure()->typeInfo().isFinalObject() && !hasIndexedProperties(object->structure()->indexingType())) {
        object->freeze(vm);
        return object;
    }

    bool success = setIntegrityLevel<IntegrityLevel::Frozen>(exec, vm, object);
    RETURN_IF_EXCEPTION(scope, 0);
    if (UNLIKELY(!success)) {
        throwTypeError(exec, vm, ASCIILiteral("Unable to prevent extension in Object.freeze"));
        return 0;
    }

    return object;
}

EncodedJSValue JSC_HOST_CALL objectConstructorFreeze(ExecState* exec)
{
	VM& vm = exec->vm();
    // 1. If Type(O) is not Object, return O.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(obj);
    JSObject* result = objectConstructorFreeze(exec, asObject(obj));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(result);
}

EncodedJSValue JSC_HOST_CALL objectConstructorPreventExtensions(ExecState* exec)
{
	VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue argument = exec->argument(0);
    if (!argument.isObject())
        return JSValue::encode(argument);
    JSObject* object = asObject(argument);
    bool status = object->methodTable(vm)->preventExtensions(object, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
	if (UNLIKELY(!status))
		return throwVMTypeError(exec, vm, ASCIILiteral("Unable to prevent extension in Object.preventExtensions"));
    return JSValue::encode(object);
}

EncodedJSValue JSC_HOST_CALL objectConstructorIsSealed(ExecState* exec)
{
	VM& vm = exec->vm();

    // 1. If Type(O) is not Object, return true.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(jsBoolean(true));
    JSObject* object = asObject(obj);

    // Quick check for final objects.
    if (object->structure()->typeInfo().isFinalObject() && !hasIndexedProperties(object->structure()->indexingType()))
        return JSValue::encode(jsBoolean(object->isSealed(vm)));

    // 2. Return ? TestIntegrityLevel(O, "sealed").
    return JSValue::encode(jsBoolean(testIntegrityLevel<IntegrityLevel::Sealed>(exec, vm, object)));
}

EncodedJSValue JSC_HOST_CALL objectConstructorIsFrozen(ExecState* exec)
{
	VM& vm = exec->vm();

    // 1. If Type(O) is not Object, return true.
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(jsBoolean(true));
    JSObject* object = asObject(obj);

    // Quick check for final objects.
    if (object->structure()->typeInfo().isFinalObject() && !hasIndexedProperties(object->structure()->indexingType()))
        return JSValue::encode(jsBoolean(object->isFrozen(vm)));

    // 2. Return ? TestIntegrityLevel(O, "frozen").
    return JSValue::encode(jsBoolean(testIntegrityLevel<IntegrityLevel::Frozen>(exec, vm, object)));
}

EncodedJSValue JSC_HOST_CALL objectConstructorIsExtensible(ExecState* exec)
{
    VM& vm = exec->vm();
    JSValue obj = exec->argument(0);
    if (!obj.isObject())
        return JSValue::encode(jsBoolean(false));
    JSObject* object = asObject(obj);
    bool isExtensible = object->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(jsBoolean(isExtensible));
}

EncodedJSValue JSC_HOST_CALL objectConstructorIs(ExecState* exec)
{
    return JSValue::encode(jsBoolean(sameValue(exec, exec->argument(0), exec->argument(1))));
}

// FIXME: Use the enumeration cache.
JSArray* ownPropertyKeys(ExecState* exec, JSObject* object, PropertyNameMode propertyNameMode, DontEnumPropertiesMode dontEnumPropertiesMode)
{
    VM& vm = exec->vm();
    PropertyNameArray properties(vm, propertyNameMode, PrivateSymbolMode::Exclude);
    object->methodTable(vm)->getOwnPropertyNames(object, exec, properties, dontEnumPropertiesMode);
    RETURN_IF_EXCEPTION(vm, 0);

    if (propertyNameMode != PropertyNameMode::StringsAndSymbols) {
        ASSERT(propertyNameMode == PropertyNameMode::Strings || propertyNameMode == PropertyNameMode::Symbols);
        if (properties.size() < MIN_SPARSE_ARRAY_INDEX) {
			auto* globalObject = exec->lexicalGlobalObject();
			if (LIKELY(!globalObject->isHavingABadTime())) {
				size_t numProperties = properties.size();
                // FIXME: We should probably be calling tryCreate here:
                // https://bugs.webkit.org/show_bug.cgi?id=221984
				JSArray* keys = JSArray::tryCreate(vm, exec, globalObject->originalArrayStructureForIndexingType(ArrayWithContiguous), numProperties);
			    RETURN_IF_EXCEPTION(vm, nullptr);
				WriteBarrier<Unknown>* buffer = keys->butterfly()->contiguous().data();
				for (size_t i = 0; i < numProperties; i++) {
					const Identifier& identifier = properties[i];
                    if (propertyNameMode == PropertyNameMode::Strings) {
                    	ASSERT(!identifier.isSymbol());
                    	buffer[i].set(vm, keys, jsOwnedString(&vm, identifier.string()));
                    } else {
						ASSERT(identifier.isSymbol());
						buffer[i].set(vm, keys, Symbol::create(vm, static_cast<SymbolImpl&>(*identifier.impl())));
					}
				}
				return keys;
			}
        }
    }

    JSArray* keys = constructEmptyArray(exec, 0);
    RETURN_IF_EXCEPTION(vm, nullptr);

    unsigned index = 0;
    auto pushDirect = [&] (ExecState* exec, JSArray* array, JSValue value) {
        array->putDirectIndex(exec, index++, value);
    };

    switch (propertyNameMode) {
    case PropertyNameMode::Strings: {
        size_t numProperties = properties.size();
        for (size_t i = 0; i < numProperties; i++) {
            const Identifier& identifier = properties[i];
            ASSERT(!identifier.isSymbol());
            pushDirect(exec, keys, jsOwnedString(vm, identifier.string()));
			RETURN_IF_EXCEPTION(vm, nullptr);
        }
        break;
    }

    case PropertyNameMode::Symbols: {
        size_t numProperties = properties.size();
        for (size_t i = 0; i < numProperties; i++) {
            const Identifier& identifier = properties[i];
            ASSERT(identifier.isSymbol());
            ASSERT(!identifier.isPrivateName());
            pushDirect(exec, keys, Symbol::create(vm, static_cast<SymbolImpl&>(*identifier.impl())));
			RETURN_IF_EXCEPTION(vm, nullptr);
        }
        break;
    }

    case PropertyNameMode::StringsAndSymbols: {
        size_t numProperties = properties.size();
        for (size_t i = 0; i < numProperties; i++) {
            const Identifier& identifier = properties[i];
            if (identifier.isSymbol()) {
            	ASSERT(!identifier.isPrivateName());
            	pushDirect(exec, keys, Symbol::create(vm, static_cast<SymbolImpl&>(*identifier.impl())));
            }else
            	pushDirect(exec, keys, jsOwnedString(vm, identifier.string()));
            RETURN_IF_EXCEPTION(vm, nullptr);
        }
        break;
    }
    }

    return keys;
}

EncodedJSValue JSC_HOST_CALL objectConstructorHasOwn(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    JSObject* base = exec->argument(0).toObject(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    auto propertyName = exec->argument(1).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    scope.release();
    return JSValue::encode(jsBoolean(objectPrototypeHasOwnProperty(exec, base, propertyName)));
}

} // namespace JSC
