/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2008, 2011 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "ObjectPrototype.h"

#include "Error.h"
#include "GetterSetter.h"
#include "HasOwnPropertyCache.h"
#include "JSFunction.h"
#include "JSGlobalObject.h"
#include "JSString.h"
#include "Operations.h"
#include "PropertySlot.h"
#include "StructureInlines.h"
#include "StructureRareDataInlines.h"
#include "ThrowScope.h"

namespace JSC {

static EncodedJSValue JSC_HOST_CALL objectProtoFuncValueOf(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncHasOwnProperty(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncIsPrototypeOf(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncDefineGetter(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncDefineSetter(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncLookupGetter(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncLookupSetter(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncPropertyIsEnumerable(ExecState*);
static EncodedJSValue JSC_HOST_CALL objectProtoFuncToLocaleString(ExecState*);

ASSERT_HAS_TRIVIAL_DESTRUCTOR(ObjectPrototype);

const ClassInfo ObjectPrototype::s_info = { "Object", &JSNonFinalObject::s_info, nullptr, CREATE_METHOD_TABLE(ObjectPrototype) };

ObjectPrototype::ObjectPrototype(VM& vm, Structure* stucture)
    : JSNonFinalObject(vm, stucture)
{
}

void ObjectPrototype::finishCreation(VM& vm, JSGlobalObject* globalObject)
{
    Base::finishCreation(vm);
    ASSERT(inherits(info()));
    vm.prototypeMap.addPrototype(this);

    putDirectWithoutTransition(vm, vm.propertyNames->toString, globalObject->objectProtoToStringFunction(), static_cast<unsigned>(PropertyAttribute::DontEnum));
	JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->toLocaleString, objectProtoFuncToLocaleString, DontEnum, 0);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->valueOf, objectProtoFuncValueOf, DontEnum, 0);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->hasOwnProperty, objectProtoFuncHasOwnProperty, DontEnum, 1, HasOwnPropertyIntrinsic);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->propertyIsEnumerable, objectProtoFuncPropertyIsEnumerable, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->isPrototypeOf, objectProtoFuncIsPrototypeOf, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->__defineGetter__, objectProtoFuncDefineGetter, DontEnum, 2);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->__defineSetter__, objectProtoFuncDefineSetter, DontEnum, 2);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->__lookupGetter__, objectProtoFuncLookupGetter, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->__lookupSetter__, objectProtoFuncLookupSetter, DontEnum, 1);
}

ObjectPrototype* ObjectPrototype::create(VM& vm, JSGlobalObject* globalObject, Structure* structure)
{
    ObjectPrototype* prototype = new (NotNull, allocateCell<ObjectPrototype>(vm.heap)) ObjectPrototype(vm, structure);
    prototype->finishCreation(vm, globalObject);
    return prototype;
}

// ------------------------------ Functions --------------------------------

static EncodedJSValue JSC_HOST_CALL objectProtoFuncValueOf(ExecState* exec)
{
    JSValue thisValue = exec->hostThisValue().toThis(exec, ECMAMode::strict());
    JSObject* valueObj = thisValue.toObject(exec);
    if (UNLIKELY(!valueObj))
        return encodedJSUndefined();
    return JSValue::encode(valueObj);
}

bool objectPrototypeHasOwnProperty(ExecState* exec, JSObject* thisObject, const Identifier& propertyName)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    Structure* structure = thisObject->structure(vm);
    HasOwnPropertyCache* hasOwnPropertyCache = vm.ensureHasOwnPropertyCache();
    if (Optional<bool> result = hasOwnPropertyCache->get(structure, propertyName)) {
        ASSERT(*result == thisObject->hasOwnProperty(globalObject, propertyName) || vm.hasPendingTerminationException());
        scope.assertNoException();
        return *result;
    }

    PropertySlot slot(thisObject, PropertySlot::InternalMethodType::GetOwnProperty);
    bool result = thisObject->hasOwnProperty(exec, propertyName, slot);
    RETURN_IF_EXCEPTION(scope, false);

    hasOwnPropertyCache->tryAdd(vm, slot, thisObject, propertyName, result);
    return result;
}

static EncodedJSValue JSC_HOST_CALL objectProtoFuncHasOwnProperty(ExecState* exec)
{
	VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue base = exec->thisValue();
    auto propertyName = exec->argument(0).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    JSObject* thisObject = base.toThis(exec, ECMAMode::strict()).toObject(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    scope.release();
    return JSValue::encode(jsBoolean(objectPrototypeHasOwnProperty(exec, thisObject, propertyName)));
}

static EncodedJSValue JSC_HOST_CALL objectProtoFuncIsPrototypeOf(ExecState* exec)
{
    VM& vm = exec->vm();
	ThrowScope scope(vm);

    if (!exec->argument(0).isObject())
        return JSValue::encode(jsBoolean(false));

    JSValue thisValue = exec->hostThisValue().toThis(exec, ECMAMode::strict());
    JSObject* thisObj = thisValue.toObject(exec);
    ASSERT(!!scope.exception() == !thisObj);
    if (UNLIKELY(!thisObj))
        return encodedJSUndefined();

    JSValue v = asObject(exec->argument(0))->getPrototype(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    while (true) {
        if (!v.isObject())
            return JSValue::encode(jsBoolean(false));
        if (v == thisObj)
            return JSValue::encode(jsBoolean(true));
        v = asObject(v)->getPrototype(vm, exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }
}

static EncodedJSValue JSC_HOST_CALL objectProtoFuncDefineGetter(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue get = exec->argument(1);
    if (!get.isCallable())
        return throwVMError(exec, vm, createTypeError(exec, "invalid getter usage"));

    Identifier propertyName = exec->argument(0).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    PropertyDescriptor descriptor;
    descriptor.setGetter(get);
    descriptor.setEnumerable(true);
    descriptor.setConfigurable(true);

    bool shouldThrow = true;
    thisObject->methodTable()->defineOwnProperty(thisObject, exec, propertyName, descriptor, shouldThrow);

    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL objectProtoFuncDefineSetter(ExecState* exec)
{
	VM& vm = exec->vm();

    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue set = exec->argument(1);
    if (!set.isCallable())
        return throwVMError(exec, vm, createTypeError(exec, "invalid setter usage"));

    Identifier propertyName = exec->argument(0).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    PropertyDescriptor descriptor;
    descriptor.setSetter(set);
    descriptor.setEnumerable(true);
    descriptor.setConfigurable(true);

    bool shouldThrow = true;
    thisObject->methodTable()->defineOwnProperty(thisObject, exec, propertyName, descriptor, shouldThrow);

    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL objectProtoFuncLookupGetter(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    if (!thisObject)
        return JSValue::encode(jsUndefined());

    Identifier propertyName = exec->argument(0).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    PropertySlot slot(thisObject, PropertySlot::InternalMethodType::GetOwnProperty);
    bool hasProperty = thisObject->getPropertySlot(vm, exec, propertyName, slot);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        if(slot.isAccessor()){
            GetterSetter* getterSetter = slot.getterSetter();
            return getterSetter->isGetterNull() ? JSValue::encode(jsUndefined()) : JSValue::encode(getterSetter->getter());
        }
        if (slot.attributes() & CustomAccessor) {
            PropertyDescriptor descriptor;
            ASSERT(slot.slotBase());
            if (slot.slotBase()->getOwnPropertyDescriptor(exec, propertyName, descriptor))
                return descriptor.getterPresent() ? JSValue::encode(descriptor.getter()) : JSValue::encode(jsUndefined());
        }
    }

    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL objectProtoFuncLookupSetter(ExecState* exec)
{
	VM& vm = exec->vm();
    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    if (!thisObject)
        return JSValue::encode(jsUndefined());

    Identifier propertyName = exec->argument(0).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    PropertySlot slot(thisObject, PropertySlot::InternalMethodType::GetOwnProperty);
    bool hasProperty = thisObject->getPropertySlot(vm, exec, propertyName, slot);
    ASSERT(!scope.exception() || !hasProperty);
    if (hasProperty) {
        if(slot.isAccessor()){
            GetterSetter* getterSetter = slot.getterSetter();
            return getterSetter->isSetterNull() ? JSValue::encode(jsUndefined()) : JSValue::encode(getterSetter->setter());
        }
        if (slot.attributes() & CustomAccessor) {
            PropertyDescriptor descriptor;
            ASSERT(slot.slotBase());
            if (slot.slotBase()->getOwnPropertyDescriptor(exec, propertyName, descriptor))
                return descriptor.setterPresent() ? JSValue::encode(descriptor.setter()) : JSValue::encode(jsUndefined());
        }
    }

    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL objectProtoFuncPropertyIsEnumerable(ExecState* exec)
{
	VM& vm = exec->vm();
    Identifier propertyName = exec->argument(0).toPropertyKey(vm, exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSObject* thisObject = exec->hostThisValue().toThis(exec, ECMAMode::strict()).toObject(exec);
    if (!thisObject)
        return JSValue::encode(jsUndefined());
    PropertyDescriptor descriptor;
    bool enumerable = thisObject->getOwnPropertyDescriptor(exec, propertyName, descriptor) && descriptor.enumerable();
    return JSValue::encode(jsBoolean(enumerable));
}

// 15.2.4.3 Object.prototype.toLocaleString()
static EncodedJSValue JSC_HOST_CALL objectProtoFuncToLocaleString(ExecState* exec)
{
    VM& vm = exec->vm();
	ThrowScope scope(vm);

    // 1. Let V be the this value.
    JSValue thisValue = exec->thisValue();

    // 2. Invoke(V, "toString")

    // Let O be the result of calling ToObject passing the this value as the argument.
    JSObject* object = thisValue.toThis(exec, ECMAMode::strict()).toObject(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // Let toString be the O.[[Get]]("toString", V)
    PropertySlot slot(thisValue, PropertySlot::InternalMethodType::Get);
    bool hasProperty = object->getPropertySlot(vm, exec, vm.propertyNames->toString, slot);
    JSValue toString = hasProperty ? slot.getValue(exec, vm.propertyNames->toString) : jsUndefined();
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // 3. If IsCallable(toString) is false, throw a TypeError exception.
    CallData callData = getCallData(vm, toString);
    if (callData.type == CallData::Type::None)
        return throwVMTypeError(exec, vm);

    // 4. Return the result of calling the [[Call]] internal method of toString passing the this value and no arguments.
    return JSValue::encode(call(vm, exec, toString, callData, thisValue, *vm.emptyList));
}

JSValue objectPrototypeToString(ExecState* exec, JSValue thisValue)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    if (thisValue.isUndefined())
        return vm.smallStrings.undefinedObjectString();
    if (thisValue.isNull())
        return vm.smallStrings.nullObjectString();
    JSObject* thisObject = thisValue.toObject(exec);
    scope.assertNoException();

    JSValue result = thisObject->structure(vm)->cachedSpecialProperty(CachedSpecialPropertyKey::ToStringTag);
    if (result)
        return result;

	PropertyName toStringTagSymbol = vm.propertyNames->toStringTagSymbol;
	PropertySlot slot(thisObject, PropertySlot::InternalMethodType::Get);
	bool found = thisObject->getPropertySlot(vm, exec, toStringTagSymbol, slot);
    RETURN_IF_EXCEPTION(vm, JSValue());
	if (found) {
		JSValue stringTag = slot.getValue(exec, toStringTagSymbol);
	    RETURN_IF_EXCEPTION(vm, JSValue());
		if (stringTag.isString()) {
			JSRopeString::RopeBuilder<RecordOverflow> ropeBuilder(vm);
			ropeBuilder.append(vm.smallStrings.objectStringStart());
			ropeBuilder.append(jsCast<JSString*>(stringTag));
			ropeBuilder.append(vm.smallStrings.singleCharacterString(']'));
			if (ropeBuilder.hasOverflowed()) {
				throwOutOfMemoryError(exec);
				return JSValue();
			}
			result = ropeBuilder.release();

			thisObject->structure(vm)->cacheSpecialProperty(exec, vm, result, CachedSpecialPropertyKey::ToStringTag, slot);
			return result;
		}
	}

    String tag = thisObject->methodTable(vm)->toStringName(thisObject, exec);
    RETURN_IF_EXCEPTION(vm, JSValue());
    String newString = WTF::tryMakeString("[object ", tag, "]");
    if (!newString) {
		throwOutOfMemoryError(exec);
		return vm.smallStrings.emptyString();
    }

	result = jsNontrivialString(vm, newString);
	thisObject->structure(vm)->cacheSpecialProperty(exec, vm, result, CachedSpecialPropertyKey::ToStringTag, slot);

	return result;
}

EncodedJSValue JSC_HOST_CALL objectProtoFuncToString(ExecState* exec)
{
    JSValue thisValue = exec->thisValue().toThis(exec, ECMAMode::strict());
    return JSValue::encode(objectPrototypeToString(exec, thisValue));
}

} // namespace JSC
