/*
 * Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "JSGlobalObject.h"
#include "Operations.h"

#include "Error.h"
#include "JSBigInt.h"
#include "JSObject.h"
#include "JSString.h"
#include <math.h>
#include <stdio.h>
#include <wtf/MathExtras.h>

namespace JSC {

bool JSValue::equalSlowCase(ExecState* exec, JSValue v1, JSValue v2)
{
    return equalSlowCaseInline(exec, v1, v2);
}

NEVER_INLINE JSValue jsAddSlowCase(JSGlobalData& vm, CallFrame* callFrame, JSValue v1, JSValue v2)
{
    auto scope = DECLARE_THROW_SCOPE(vm);
    JSValue p1 = v1.toPrimitive(callFrame);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    JSValue p2 = v2.toPrimitive(callFrame);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    if (p1.isString()) {
        if (p2.isCell()) {
			JSString* p2String = p2.toString(callFrame);
			RETURN_IF_EXCEPTION(vm, jsUndefined());
			return jsString(vm, callFrame, asString(p1), p2String);
		}
        String p2String = p2.toWTFString(callFrame);
        RETURN_IF_EXCEPTION(vm, jsUndefined());
        return jsString(vm, callFrame, asString(p1), p2String);
    }

    if (p2.isString()) {
        if (p1.isCell()) {
			JSString* p1String = p1.toString(callFrame);
			RETURN_IF_EXCEPTION(vm, jsUndefined());
			return jsString(vm, callFrame, p1String, asString(p2));
        }
        String p1String = p1.toWTFString(callFrame);
        RETURN_IF_EXCEPTION(vm, jsUndefined());
        return jsString(vm, callFrame, p1String, asString(p2));
    }

    auto doubleOp = [] (double left, double right) -> double {
        return left + right;
    };
    auto bigInt32Op = [] (int32_t left, int32_t right) -> JSValue {
#if USE(BIGINT32)
        CheckedInt32 result = left;
        result += right;
        if (UNLIKELY(result.hasOverflowed()))
            return JSValue();
        return JSValue(JSValue::JSBigInt32, result.unsafeGet());
#else
        UNUSED_PARAM(left);
        UNUSED_PARAM(right);
        RELEASE_ASSERT_NOT_REACHED();
        return JSValue();
#endif
    };
    RELEASE_AND_RETURN(scope, arithmeticBinaryOp(callFrame, p1, p2, JSBigInt::add, doubleOp, bigInt32Op, ASCIILiteral("Invalid mix of BigInt and other type in addition.")));
}

JSValue jsTypeStringForValue(JSGlobalData& vm, JSGlobalObject* globalObject, JSValue v)
{
    if (v.isUndefined())
        return vm.smallStrings.undefinedString();
    if (v.isBoolean())
        return vm.smallStrings.booleanString();
    if (v.isNumber())
        return vm.smallStrings.numberString();
    if (v.isString())
        return vm.smallStrings.stringString();
    if (v.isSymbol())
        return vm.smallStrings.symbolString();
    if (v.isBigInt())
        return vm.smallStrings.bigintString();
    if (v.isObject()) {
        JSObject* object = asObject(v);
        // Return "undefined" for objects that should be treated
        // as null when doing comparisons.
    	if (object->structure()->masqueradesAsUndefined(globalObject))
            return vm.smallStrings.undefinedString();
        if (object->isCallable(vm))
            return vm.smallStrings.functionString();
    }
    return vm.smallStrings.objectString();
}

JSValue jsTypeStringForValue(CallFrame* callFrame, JSValue v)
{
    return jsTypeStringForValue(callFrame->vm(), callFrame->lexicalGlobalObject(), v);
}

bool jsTypeofIsObject(CallFrame* callFrame, JSValue v)
{
    VM& vm = callFrame->vm();
    if (!v.isCell())
        return v.isNull();

    JSType type = v.asCell()->structure()->typeInfo().type();
    if (type == StringType || type == SymbolType || type == HeapBigIntType)
        return false;
    if (type >= ObjectType) {
    	if (asObject(v)->structure()->masqueradesAsUndefined(callFrame->lexicalGlobalObject()))
            return false;
        JSObject* object = asObject(v);
        if (object->isCallable(vm))
            return false;
    }
    return true;
}

size_t normalizePrototypeChain(CallFrame* callFrame, Structure* structure)
{
	size_t count = 0;
	while (1) {
		if (structure->isProxy())
			return InvalidPrototypeChain;

        JSValue v = structure->prototypeForLookup(callFrame);
		if (v.isNull())
			return count;

        JSCell* base = v.asCell();
        structure = base->structure();
        if (structure->isDictionary()) {
            if (structure->hasBeenFlattenedBefore())
                return InvalidPrototypeChain;
            VM& vm = callFrame->vm();
            structure->flattenDictionaryStructure(vm, asObject(base));
        }

		++count;
	}

	return count;
}

} // namespace JSC
