/*
 * Copyright (C) 2011, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "JSExportMacros.h"
#include <wtf/StdLibExtras.h>
#include <stdint.h>
#include <stdio.h>

namespace JSC {

// How do JSC VM options work?
// ===========================
// The JSC_OPTIONS() macro below defines a list of all JSC options in use,
// along with their types and default values. The options values are actually
// realized as an array of Options::Entry elements.
//
//     Options::initialize() will initialize the array of options values with
// the defaults specified in JSC_OPTIONS() below. After that, the values can
// be programmatically read and written to using an accessor method with the
// same name as the option. For example, the option "useJIT" can be read and
// set like so:
//
//     bool jitIsOn = Options::useJIT();  // Get the option value.
//     Options::useJIT() = false;         // Sets the option value.
//
//     If you want to tweak any of these values programmatically for testing
// purposes, you can do so in Options::initialize() after the default values
// are set.
//
//     Alternatively, you can enable RUN_TIME_HEURISTICS which will allow you
// to override the default values by specifying environment variables of the
// form: JSC_<name of JSC option>.
//
// Note: Options::initialize() tries to ensure some sanity on the option values
// which are set by doing some range checks, and value corrections. These
// checks are done after the option values are set. If you alter the option
// values after the sanity checks (for your own testing), then you're liable to
// ensure that the new values set are sane and reasonable for your own run.

#define JSC_OPTIONS(v) \
    v(unsigned, useJIT,    1) \
    v(unsigned, useDFGJIT, 1) \
    v(unsigned, useRegExpJIT, 1/*, Normal, "allows the RegExp JIT to be used if true"*/) \
	\
	v(unsigned, assemblerDataInitalCapacity, 16 * KB) \
    \
    v(unsigned, maxPerThreadStackUsage, 4 * MB/*, Normal, "Max allowed stack usage by the VM"*/) \
    v(unsigned, softReservedZoneSize, 128 * KB/*, Normal, "A buffer greater than reservedZoneSize that reserves space for stringifying exceptions."*/) \
    v(unsigned, reservedZoneSize, 64 * KB/*, Normal, "The amount of stack space we guarantee to our clients (and to interal VM code that does not call out to clients)."*/) \
    \
    v(bool, forceDFGCodeBlockLiveness, false) \
    v(bool, forceICFailure, false) \
    v(unsigned, maxGetByIdICSize, 500) \
    v(unsigned, maxByValICSize, 50) \
    v(bool, enableInstanceOfIC, true) \
    v(bool, forceTryGetByIdICFailure, true) \
    v(unsigned, maxAccessVariantListSize, 8) \
    v(unsigned, maxPutByIdICSize, 500) \
	\
    v(unsigned, repatchCountForCoolDown, 8) \
    v(unsigned, initialCoolDownCount, 20) \
    v(unsigned, repatchBufferingCountdown, 8) \
    \
    v(bool, validateBytecode, false) \
	\
    v(bool, enableTailCalls, true) \
    \
    /* showDisassembly implies showDFGDisassembly. */ \
    v(bool, showDisassembly, false) \
    v(bool, showDFGDisassembly, false) \
    v(bool, showAllDFGNodes, false) \
    v(bool, dumpBytecodeAtDFGTime, false) \
    v(bool, dumpGraphAfterParsing, false) \
    v(bool, dumpGraphAtEachPhase, false) \
    v(bool, dumpBytecodeLivenessResults, false) \
    v(bool, verboseCompilation, false) \
    v(bool, logCompilationChanges, false) \
    v(bool, printEachOSRExit, false) \
    v(bool, validateGraph, false) \
    v(bool, validateGraphAtEachPhase, false) \
    v(bool, eagerlyUpdateTopCallFrame, false) \
    v(bool, scribbleFreeCells, false) \
    \
    v(bool, enableOSREntryToDFG, true) \
    \
	v(bool, enablePolymorphicCallInlining, false) \
    v(unsigned, maxPolymorphicCallVariantListSize, 8) \
    v(unsigned, maxPolymorphicCallVariantListSizeForTopTier, 5) \
    v(unsigned, maxPolymorphicCallVariantsForInlining, 5) \
    v(unsigned, frequentCallThreshold, 2) \
    v(double, minimumCallToKnownRate, 0.51) \
    v(bool, optimizeNativeCalls, false) \
    v(bool, enableObjectAllocationSinking, true) \
    v(unsigned, maxDFGNodesInBasicBlockForPreciseAnalysis, 20000) \
    \
    v(bool, enableProfiler, false) \
    \
    v(bool, enableDebugDFGJIT, false) \
    \
    v(bool, enableArchitectureSpecificOptimizations, true) \
    \
    v(unsigned, maximumOptimizationCandidateInstructionCount, 100000) \
    \
    v(unsigned, maximumFunctionForCallInlineCandidateInstructionCount, 190) \
    v(unsigned, maximumFunctionForClosureCallInlineCandidateInstructionCount, 100) \
    v(unsigned, maximumFunctionForConstructInlineCandidateInstructionCount, 100) \
    \
    v(unsigned, maximumFTLCandidateInstructionCount, 20000) \
    \
    /* Depth of inline stack, so 1 = no inlining, 2 = one level, etc. */ \
    v(unsigned, maximumInliningDepth, 5) \
    v(unsigned, maximumInliningRecursion, 2) \
    v(unsigned, maximumInliningDepthForMustInline, 7) \
    v(unsigned, maximumInliningRecursionForMustInline, 3) \
    \
    v(bool, enablePolyvariantCallInlining, true) \
    \
    /* Maximum size of a caller for enabling inlining. This is purely to protect us */\
    /* from super long compiles that take a lot of memory. */\
    v(unsigned, maximumInliningCallerSize, 10000) \
    \
    v(unsigned, maximumBinaryStringSwitchCaseLength, 50) \
    v(unsigned, maximumBinaryStringSwitchTotalLength, 2000) \
    \
    v(int32, thresholdForJITAfterWarmUp, 100) \
    v(int32, thresholdForJITSoon, 100) \
    \
    v(int32, thresholdForOptimizeAfterWarmUp, 1000) \
    v(int32, thresholdForOptimizeAfterLongWarmUp, 1000) \
    v(int32, thresholdForOptimizeSoon, 1000) \
    v(int32, executionCounterIncrementForLoop, 1) \
    v(int32, executionCounterIncrementForEntry, 15) \
    \
    v(int32, evalThresholdMultiplier, 10) \
    \
    v(bool, randomizeExecutionCountsBetweenCheckpoints, false) \
    v(int32, maximumExecutionCountsBetweenCheckpoints, 1000) \
    \
    v(unsigned, likelyToTakeSlowCaseMinimumCount, 20) \
    v(unsigned, couldTakeSlowCaseMinimumCount, 10) \
    \
    v(unsigned, osrExitCountForReoptimization, 100) \
    v(unsigned, osrExitCountForReoptimizationFromLoop, 5) \
    \
    v(unsigned, reoptimizationRetryCounterMax, 0)  \
    v(unsigned, reoptimizationRetryCounterStep, 1) \
    \
    v(unsigned, minimumOptimizationDelay, 1) \
    v(unsigned, maximumOptimizationDelay, 5) \
    v(double, desiredProfileLivenessRate, 0.75) \
    v(double, desiredProfileFullnessRate, 0.35) \
    \
    v(double, doubleVoteRatioForDoubleFormat, 2) \
    v(double, structureCheckVoteRatioForHoisting, 1) \
    v(double, checkArrayVoteRatioForHoisting, 1) \
    \
    v(unsigned, minimumNumberOfScansBetweenRebalance, 100) \
    v(unsigned, numberOfGCMarkers, computeNumberOfGCMarkers(7)) \
    v(unsigned, opaqueRootMergeThreshold, 1000) \
	\
    v(double, minHeapUtilization, 0.75) \
    v(int32, minHeapIncrement, 16 * 1024 * 1024) \
    v(double, minCopiedBlockUtilization, 0.95) \
    \
    v(bool, forceWeakRandomSeed, false) \
    v(unsigned, forcedWeakRandomSeed, 0) \
    \
    v(bool, useZombieMode, false) \
    v(bool, objectsAreImmortal, false) \
    v(bool, showHeapStatistics, false) \
    \
    v(bool, logGC, false) \
    \
    v(unsigned, exceptionStackTraceLimit, 100/*, Normal, "Stack trace limit for internal Exception object"*/) \
    v(unsigned, defaultErrorStackTraceLimit, 100/*, Normal, "The default value for Error.stackTraceLimit"*/) \
    v(bool, enableExceptionFuzz, false) \
    v(unsigned, fireExceptionFuzzAt, 0) \
	\
    v(bool, shouldEmitDebugHooks, false) \
	\
	v(unsigned, jitCodeMinimumBlockPages, 32) \
	v(int32, jitCodesDeletionInterval, 0) \
    v(unsigned, jitCodesDeletionGCLengthThreshold, 150 /* in ms */) \
	v(double, jitCodesDeletionPercentage, 0.001) \
	\
    v(unsigned, largeHeapSize, 16 * 1024 * 1024) \
	v(unsigned, forceGCOnEveryAllocation, 0) \
	\
	v(double, jitCodesOptimizationTimeLimit, 0.25) \
	\
    v(bool, enableES6Symbol, true) \
    v(bool, enableES6ClassSyntax, true) \
	\
    v(bool, enableDFGInline, true) \
	\
    v(bool, enableFunctionDotArguments, true) \
	\
    v(unsigned, maximumEvalCacheableSourceLength, 256) \
	\
    v(bool, dumpModuleRecord, false) \
    v(bool, dumpModuleLoadingState, false) \
    v(bool, exposeInternalModuleLoader, false) \
    v(unsigned, thresholdForGlobalLexicalBindingEpoch, UINT_MAX/*, Normal, "Threshold for global lexical binding epoch. If the epoch reaches to this value, CodeBlock metadata for scope operations will be revised globally. It needs to be greater than 1."*/) \
	\
    v(bool, useSourceProviderCache, true/*, Normal, "If false, the parser will not use the source provider cache. It's good to verify everything works when this is false. Because the cache is so successful, it can mask bugs."*/) \
	\
	v(bool, useAsyncIterator, true/*, Normal, "Allow to use Async Iterator in JS."*/) \
	v(bool, useObjectRestSpread, true/*, Normal, "If true, we will enable Object Rest/Spread feature."*/) \
	v(bool, useBigInt, true/*, Normal, "If true, we will enable BigInt support."*/) \
	v(bool, useIntlNumberFormatToParts, true/*, Normal, "If true, we will enable Intl.NumberFormat.prototype.formatToParts"*/) \
	v(bool, useIntlPluralRules, true/*, Normal, "If true, we will enable Intl.PluralRules."*/) \
	v(bool, useLogicalAssignmentOperators, true/*, Normal, "Enable support for \?\?=, ||=, and &&= logical assignment operators."*/) \
	\
	v(bool, usePrivateMethods, true/*, Normal, "If true, the parser will understand private methods inside classes."*/) \
	v(bool, usePrivateIn, false/*, Normal, "If true, the parser will understand private member existence checks with the `in` operator."*/) \
	\
	v(bool, useAccessInlining, true) \
	v(bool, useAtMethod, true/*, Normal, "Expose the at() method on Array and %TypedArray%."*/) \
	v(bool, useArrayFindLastMethod, true/*, Normal, "Expose the findLast() and findLastIndex() methods on Array and %TypedArray%."*/) \
	v(bool, useHasOwn, true/*, Normal, "Expose the Object.hasOwn method"*/) \
	v(bool, useTopLevelAwait, false/*, Normal, "allow the await keyword at the top level of a module."*/) \


class Options {
public:
    // This typedef is to allow us to eliminate the '_' in the field name in
    // union inside Entry. This is needed to keep the style checker happy.
    typedef int32_t int32;

    // Declare the option IDs:
    enum OptionID {
#define FOR_EACH_OPTION(type_, name_, defaultValue_) \
        OPT_##name_,
        JSC_OPTIONS(FOR_EACH_OPTION)
#undef FOR_EACH_OPTION
        numberOfOptions
    };


    static void initialize();

    // Parses a single command line option in the format "<optionName>=<value>"
    // (no spaces allowed) and set the specified option if appropriate.
    JS_EXPORT_PRIVATE static bool setOption(const char* arg);
    JS_EXPORT_PRIVATE static void dumpAllOptions(FILE* stream = stdout);
    static void dumpOption(OptionID id, FILE* stream = stdout, const char* header = "", const char* footer = "");

    static void recomputeDependentOptions();

    // Declare accessors for each option:
#define FOR_EACH_OPTION(type_, name_, defaultValue_) \
    ALWAYS_INLINE static type_& name_() { return s_options[OPT_##name_].u.type_##Val; }

    JSC_OPTIONS(FOR_EACH_OPTION)
#undef FOR_EACH_OPTION

private:
    enum EntryType {
        boolType,
        unsignedType,
        doubleType,
        int32Type
    };

    // For storing for an option value:
    struct Entry {
        union {
            bool boolVal;
            unsigned unsignedVal;
            double doubleVal;
            int32 int32Val;
        } u;
    };

    // For storing constant meta data about each option:
    struct EntryInfo {
        const char* name;
        EntryType type;
    };

    Options();

    // Declare the options:
#define FOR_EACH_OPTION(type_, name_, defaultValue_) \
    type_ m_##name_;
    JSC_OPTIONS(FOR_EACH_OPTION)
#undef FOR_EACH_OPTION

    // Declare the singleton instance of the options store:
    JS_EXPORTDATA static Entry s_options[numberOfOptions];
    static const EntryInfo s_optionsInfo[numberOfOptions];
};

} // namespace JSC
