/*
 * Copyright (C) 2016 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ProxyObject.h"

#include "ArrayConstructor.h"
#include "Error.h"
#include "IdentifierInlines.h"
#include "JSCJSValueInlines.h"
#include "JSCellInlines.h"
#include "JSObjectInlines.h"
#include "ObjectConstructor.h"
#include "Options.h"
#include "StackEntryScope.h"
#include "SlotVisitorInlines.h"
#include "StructureInlines.h"
#include <wtf/NoTailCalls.h>

// Note that we use NO_TAIL_CALLS() throughout this file because we rely on the machine stack
// growing larger for throwing OOM errors for when we have an effectively cyclic prototype chain.

namespace JSC {

// STATIC_ASSERT_IS_TRIVIALLY_DESTRUCTIBLE(ProxyObject);

const ClassInfo ProxyObject::s_info = { "ProxyObject", &Base::s_info, 0, CREATE_METHOD_TABLE(ProxyObject) };

ProxyObject::ProxyObject(VM& vm, Structure* structure)
    : Base(vm, structure)
	, m_isCallable(false)
	, m_isConstructible(false)
{
}

String ProxyObject::toStringName(const JSObject* object, ExecState* exec)
{
    VM& vm = exec->vm();
    ThrowScope scope(vm);
    const ProxyObject* proxy = jsCast<const ProxyObject*>(object);
    while (proxy) {
        const JSObject* target = proxy->target();
        bool targetIsArray = isArray(exec, target);
        if (UNLIKELY(scope.exception()))
            break;
        if (targetIsArray) {
            scope.release();
            return target->classInfo()->methodTable.toStringName(target, exec);
        }

        proxy = jsDynamicCast<const ProxyObject*>(target);
    }
    return ASCIILiteral("Object");
}

Structure* ProxyObject::structureForTarget(JSGlobalObject* globalObject, JSValue target)
{
	VM& vm = globalObject->vm();
	return target.isCallable(vm) ? globalObject->callableProxyObjectStructure() : globalObject->proxyObjectStructure();
}

void ProxyObject::finishCreation(VM& vm, ExecState* exec, JSValue target, JSValue handler)
{
    Base::finishCreation(vm);
    if (!target.isObject()) {
        throwTypeError(exec, vm, ASCIILiteral("A Proxy's 'target' should be an Object"));
        return;
    }
    if (!handler.isObject()) {
        throwTypeError(exec, vm, ASCIILiteral("A Proxy's 'handler' should be an Object"));
        return;
    }

    JSObject* targetAsObject = jsCast<JSObject*>(target);

    m_isCallable = targetAsObject->isCallable(vm);
    if (m_isCallable) {
		TypeInfo info = structure(vm)->typeInfo();
		RELEASE_ASSERT(info.implementsHasInstance() && info.implementsDefaultHasInstance());
	}

    m_isConstructible = targetAsObject->methodTable(vm)->getConstructData(jsCast<JSObject*>(target)).type != CallData::Type::None;

    m_target.set(vm, this, targetAsObject);
    m_handler.set(vm, this, handler);
}

static ASCIILiteral s_proxyAlreadyRevokedErrorMessage("Proxy has already been revoked. No more operations are allowed to be performed on it");

static JSValue performProxyGet(ExecState* exec, ProxyObject* proxyObject, JSValue receiver, PropertyName propertyName)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return jsUndefined();
    }

    JSObject* target = proxyObject->target();

    auto performDefaultGet = [&] () {
		PropertySlot slot(receiver, PropertySlot::InternalMethodType::Get);
		if (target->getPropertySlot(vm, exec, propertyName, slot))
			return slot.getValue(exec, propertyName);
		return jsUndefined();
    };

	if (propertyName.isPrivateName())
		return jsUndefined();

    JSValue handlerValue = proxyObject->handler();
    if (handlerValue.isNull())
        return throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue getHandler = handler->getMethod(exec, callData, vm.propertyNames->get, ASCIILiteral("'get' property of a Proxy's handler object should be callable"));
	RETURN_IF_EXCEPTION(vm, jsUndefined());

    if (getHandler.isUndefined())
    	return performDefaultGet();

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(identifierToSafePublicJSValue(vm, Identifier::fromUid(vm, propertyName.uid())));
    arguments.append(receiver.toThis(exec, ECMAMode::strict()));
	JSValue trapResult = call(exec, getHandler, callData, handler, arguments);
	RETURN_IF_EXCEPTION(vm, jsUndefined());

    PropertyDescriptor descriptor;
    bool result = target->getOwnPropertyDescriptor(exec, propertyName, descriptor);
    EXCEPTION_ASSERT(!scope.exception() || !result);
    if (result) {
        if (descriptor.isDataDescriptor() && !descriptor.configurable() && !descriptor.writable()) {
        	bool isSame = sameValue(exec, descriptor.value(), trapResult);
			RETURN_IF_EXCEPTION(vm, jsUndefined());
			if (!isSame)
                return throwTypeError(exec, vm, ASCIILiteral("Proxy handler's 'get' result of a non-configurable and non-writable property should be the same value as the target's property"));
        } else if (descriptor.isAccessorDescriptor() && !descriptor.configurable() && descriptor.getter().isUndefined()) {
            if (!trapResult.isUndefined())
                return throwTypeError(exec, vm, ASCIILiteral("Proxy handler's 'get' result of a non-configurable accessor property without a getter should be undefined"));
        }
    }

	RETURN_IF_EXCEPTION(vm, jsUndefined());

    return trapResult;
}

bool ProxyObject::performGet(ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    JSValue result = performProxyGet(exec, this, slot.thisValue(), propertyName);
    RETURN_IF_EXCEPTION(vm, false);
    unsigned ignoredAttributes = 0;
    slot.setValue(this, ignoredAttributes, result);
    return true;
}

// https://tc39.es/ecma262/#sec-completepropertydescriptor
static void completePropertyDescriptor(PropertyDescriptor& desc)
{
    if (desc.isAccessorDescriptor()) {
        if (!desc.getter())
            desc.setGetter(jsUndefined());
        if (!desc.setter())
            desc.setSetter(jsUndefined());
    } else {
        if (!desc.value())
            desc.setValue(jsUndefined());
        if (!desc.writablePresent())
            desc.setWritable(false);
    }
    if (!desc.enumerablePresent())
        desc.setEnumerable(false);
    if (!desc.configurablePresent())
        desc.setConfigurable(false);
}

bool ProxyObject::performInternalMethodGetOwnProperty(ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }
    JSObject* target = this->target();

    if (propertyName.isPrivateName())
        return false;

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue getOwnPropertyDescriptorMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "getOwnPropertyDescriptor"), ASCIILiteral("'getOwnPropertyDescriptor' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);
    if (getOwnPropertyDescriptorMethod.isUndefined())
        return target->methodTable(vm)->getOwnPropertySlot(target, exec, propertyName, slot);

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(identifierToSafePublicJSValue(vm, Identifier::fromUid(vm, propertyName.uid())));
    JSValue trapResult = call(exec, getOwnPropertyDescriptorMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);

    if (!trapResult.isUndefined() && !trapResult.isObject()) {
        throwTypeError(exec, vm, ASCIILiteral("result of 'getOwnPropertyDescriptor' call should either be an Object or undefined"));
        return false;
    }

    PropertyDescriptor targetPropertyDescriptor;
    bool isTargetPropertyDescriptorDefined = target->getOwnPropertyDescriptor(exec, propertyName, targetPropertyDescriptor);
    RETURN_IF_EXCEPTION(vm, false);

    if (trapResult.isUndefined()) {
        if (!isTargetPropertyDescriptorDefined)
            return false;
        if (!targetPropertyDescriptor.configurable()) {
            throwTypeError(exec, vm, ASCIILiteral("When the result of 'getOwnPropertyDescriptor' is undefined the target must be configurable"));
            return false;
        }
        // FIXME: this doesn't work if 'target' is another Proxy. We don't have isExtensible implemented in a way that fits w/ Proxys.
        // https://bugs.webkit.org/show_bug.cgi?id=154375
        bool isExtensible = target->isExtensible(exec);
        RETURN_IF_EXCEPTION(vm, false);
        if (!isExtensible) {
            // FIXME: Come up with a test for this error. I'm not sure how to because
            // Object.seal(o) will make all fields [[Configurable]] false.
            // https://bugs.webkit.org/show_bug.cgi?id=154376
            throwTypeError(exec, vm, ASCIILiteral("When 'getOwnPropertyDescriptor' returns undefined, the 'target' of a Proxy should be extensible"));
            return false;
        }

        return false;
    }

    bool isExtensible = target->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, false);
    PropertyDescriptor trapResultAsDescriptor;
    toPropertyDescriptor(exec, trapResult, trapResultAsDescriptor);
    RETURN_IF_EXCEPTION(vm, false);
    completePropertyDescriptor(trapResultAsDescriptor);
    bool throwException = false;
    bool valid = validateAndApplyPropertyDescriptor(exec, 0, propertyName, isExtensible,
        trapResultAsDescriptor, isTargetPropertyDescriptorDefined, targetPropertyDescriptor, throwException);
    RETURN_IF_EXCEPTION(vm, false);
    if (!valid) {
        throwTypeError(exec, vm, ASCIILiteral("Result from 'getOwnPropertyDescriptor' fails the IsCompatiblePropertyDescriptor test"));
        return false;
    }

    if (!trapResultAsDescriptor.configurable()) {
        if (!isTargetPropertyDescriptorDefined || targetPropertyDescriptor.configurable()) {
            throwTypeError(exec, vm, ASCIILiteral("Result from 'getOwnPropertyDescriptor' can't be non-configurable when the 'target' doesn't have it as an own property or if it is a configurable own property on 'target'"));
            return false;
        }
        if (trapResultAsDescriptor.writablePresent() && !trapResultAsDescriptor.writable() && targetPropertyDescriptor.writable()) {
            throwVMTypeError(exec, vm, ASCIILiteral("Result from 'getOwnPropertyDescriptor' can't be non-configurable and non-writable when the target's property is writable"));
            return false;
        }
    }

    if (trapResultAsDescriptor.isAccessorDescriptor()) {
        GetterSetter* getterSetter = trapResultAsDescriptor.slowGetterSetter(exec);
        RETURN_IF_EXCEPTION(vm, false);
        slot.setGetterSlot(this, trapResultAsDescriptor.attributes(), getterSetter);
    } else if (trapResultAsDescriptor.isDataDescriptor() && !trapResultAsDescriptor.value().isEmpty())
        slot.setValue(this, trapResultAsDescriptor.attributes(), trapResultAsDescriptor.value());
    else
        slot.setValue(this, trapResultAsDescriptor.attributes(), jsUndefined()); // We use undefined because it's the default value in object properties.

    return true;
}

bool ProxyObject::performHasProperty(ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    JSObject* target = this->target();
    slot.setValue(this, None, jsUndefined()); // Nobody should rely on our value, but be safe and protect against any bad actors reading our value.

    auto performDefaultHasProperty = [&] {
	   return target->getPropertySlot(vm, exec, propertyName, slot);
   };

    if (propertyName.isPrivateName())
        return false;

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
    	throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
    	return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue hasMethod = handler->getMethod(exec, callData, vm.propertyNames->has, ASCIILiteral("'has' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);
    if (hasMethod.isUndefined())
    	return performDefaultHasProperty();

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(identifierToSafePublicJSValue(vm, Identifier::fromUid(vm, propertyName.uid())));
    JSValue trapResult = call(exec, hasMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);

    bool trapResultAsBool = trapResult.toBoolean(exec);
    RETURN_IF_EXCEPTION(vm, false);

    if (!trapResultAsBool) {
        PropertyDescriptor descriptor;
        bool isPropertyDescriptorDefined = target->getOwnPropertyDescriptor(exec, propertyName, descriptor);
        RETURN_IF_EXCEPTION(vm, false);
        if (isPropertyDescriptorDefined) {
            if (!descriptor.configurable()) {
                throwTypeError(exec, vm, ASCIILiteral("Proxy 'has' must return 'true' for non-configurable properties"));
                return false;
            }
            bool isExtensible = target->isExtensible(exec);
            RETURN_IF_EXCEPTION(vm, false);
            if (!isExtensible) {
                throwTypeError(exec, vm, ASCIILiteral("Proxy 'has' must return 'true' for a non-extensible 'target' object with a configurable property"));
                return false;
            }
        }
    }

    return trapResultAsBool;
}

bool ProxyObject::getOwnPropertySlotCommon(ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    VM& vm = exec->vm();

    slot.disableCaching();
	slot.setIsTaintedByOpaqueObject();

    if (slot.isVMInquiry()) {
        slot.setValue(this, static_cast<unsigned>(JSC::None), jsUndefined());
        return false;
    }

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    switch (slot.internalMethodType()) {
    case PropertySlot::InternalMethodType::Get:
        return performGet(exec, propertyName, slot);
    case PropertySlot::InternalMethodType::GetOwnProperty:
        return performInternalMethodGetOwnProperty(exec, propertyName, slot);
    case PropertySlot::InternalMethodType::HasProperty:
		return performHasProperty(exec, propertyName, slot);
	default:
        return false;
    }

    RELEASE_ASSERT_NOT_REACHED();
    return false;
}

bool ProxyObject::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    ProxyObject* thisObject = jsCast<ProxyObject*>(object);
    return thisObject->getOwnPropertySlotCommon(exec, propertyName, slot);
}

bool ProxyObject::getOwnPropertySlotByIndex(JSObject* object, ExecState* exec, unsigned propertyName, PropertySlot& slot)
{
	VM& vm = exec->vm();
    ProxyObject* thisObject = jsCast<ProxyObject*>(object);
    Identifier ident = Identifier::from(vm, propertyName);
    return thisObject->getOwnPropertySlotCommon(exec, ident, slot);
}

template <typename PerformDefaultPutFunction>
bool ProxyObject::performPut(ExecState* exec, JSValue putValue, JSValue thisValue, PropertyName propertyName, PerformDefaultPutFunction performDefaultPut, bool shouldThrow)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    if (propertyName.isPrivateName())
        return false;

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue setMethod = handler->getMethod(exec, callData, vm.propertyNames->set, ASCIILiteral("'set' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);
    JSObject* target = this->target();
    if (setMethod.isUndefined())
    	return performDefaultPut();

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(identifierToSafePublicJSValue(vm, Identifier::fromUid(vm, propertyName.uid())));
    arguments.append(putValue);
    arguments.append(thisValue.toThis(exec, ECMAMode::strict()));
    JSValue trapResult = call(exec, setMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);
    bool trapResultAsBool = trapResult.toBoolean(exec);
    RETURN_IF_EXCEPTION(vm, false);
    if (!trapResultAsBool) {
        if (shouldThrow)
            throwTypeError(exec, vm, tryMakeString("Proxy object's 'set' trap returned falsy value for property '", String(propertyName.uid()), "'"));
        return false;
    }

    PropertyDescriptor descriptor;
    bool result = target->getOwnPropertyDescriptor(exec, propertyName, descriptor);
    EXCEPTION_ASSERT(!scope.exception() || !result);
    if (result) {
        if (descriptor.isDataDescriptor() && !descriptor.configurable() && !descriptor.writable()) {
        	bool isSame = sameValue(exec, descriptor.value(), putValue);
			RETURN_IF_EXCEPTION(vm, false);
			if (!isSame) {
                throwTypeError(exec, vm, ASCIILiteral("Proxy handler's 'set' on a non-configurable and non-writable property on 'target' should either return false or be the same value already on the 'target'"));
                return false;
            }
        } else if (descriptor.isAccessorDescriptor() && !descriptor.configurable() && descriptor.setter().isUndefined()) {
            throwTypeError(exec, vm, ASCIILiteral("Proxy handler's 'set' method on a non-configurable accessor property without a setter should return false"));
            return false;
        }
    }
    return true;
}

bool ProxyObject::put(JSCell* cell, ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    VM& vm = exec->vm();
    slot.disableCaching();
    slot.setIsTaintedByOpaqueObject();

    ProxyObject* thisObject = jsCast<ProxyObject*>(cell);
    class PerformDefaultPut
    {
    public:
    	PerformDefaultPut(ProxyObject* thisObject, ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
    	: thisObject(thisObject)
    	, exec(exec)
    	, propertyName(propertyName)
    	, value(value)
    	, slot(slot)
    	{
    	}

    	bool operator()() {
            JSObject* target = jsCast<JSObject*>(thisObject->target());
            return target->methodTable()->put(target, exec, propertyName, value, slot);
    	};

    private:
    	ProxyObject* thisObject;
    	ExecState* exec;
    	PropertyName propertyName;
    	JSValue value;
    	PutPropertySlot& slot;
    };
    PerformDefaultPut performDefaultPut(thisObject, exec, propertyName, value, slot);
    return thisObject->performPut(exec, value, slot.thisValue(), propertyName, performDefaultPut, slot.isStrictMode());
}

bool ProxyObject::putByIndexCommon(ExecState* exec, JSValue thisValue, unsigned propertyName, JSValue putValue, bool shouldThrow)
{
    VM& vm = exec->vm();
    Identifier ident = Identifier::from(vm, propertyName);
    RETURN_IF_EXCEPTION(vm, false);
    class PerformDefaultPutByIndex
    {
    public:
    	PerformDefaultPutByIndex(ProxyObject* thisObject, ExecState* exec, JSValue thisValue, PropertyName propertyName, JSValue putValue, bool shouldThrow)
    	: thisObject(thisObject)
    	, exec(exec)
    	, thisValue(thisValue)
    	, propertyName(propertyName)
    	, putValue(putValue)
    	, shouldThrow(shouldThrow)
    	{
    	}

    	bool operator()() {
            JSObject* target = thisObject->target();
            bool isStrictMode = shouldThrow;
            PutPropertySlot slot(thisValue, isStrictMode); // We must preserve the "this" target of the putByIndex.
            return target->methodTable()->put(target, exec, propertyName, putValue, slot);
    	};

    private:
    	ProxyObject* thisObject;
    	ExecState* exec;
    	JSValue thisValue;
    	PropertyName propertyName;
    	JSValue putValue;
    	bool shouldThrow;
    };
    PerformDefaultPutByIndex performDefaultPut(this, exec, thisValue, ident, putValue, shouldThrow);
    return performPut(exec, putValue, thisValue, ident, performDefaultPut, shouldThrow);
}

bool ProxyObject::putByIndex(JSCell* cell, ExecState* exec, unsigned propertyName, JSValue value, bool shouldThrow)
{
    ProxyObject* thisObject = jsCast<ProxyObject*>(cell);
    return thisObject->putByIndexCommon(exec, thisObject, propertyName, value, shouldThrow);
}

static EncodedJSValue JSC_HOST_CALL performProxyCall(ExecState* exec)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    ProxyObject* proxy = jsCast<ProxyObject*>(exec->callee());

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return encodedJSUndefined();
    }

    JSValue handlerValue = proxy->handler();
    if (handlerValue.isNull())
        return throwVMTypeError(exec, vm, s_proxyAlreadyRevokedErrorMessage);

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue applyMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "apply"), ASCIILiteral("'apply' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    JSObject* target = proxy->target();
    if (applyMethod.isUndefined()) {
        auto callData = getCallData(vm, target);
        RELEASE_ASSERT(callType != CallType::None);
        return JSValue::encode(call(exec, target, callData, exec->thisValue(), ArgList(exec)));
    }

    JSArray* argArray = constructArray(exec, nullptr, ArgList(exec));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(exec->thisValue().toThis(exec, ECMAMode::strict()));
    arguments.append(argArray);
    return JSValue::encode(call(exec, applyMethod, callData, handler, arguments));
}

CallData ProxyObject::getCallData(JSCell* cell)
{
    CallData callData;
    ProxyObject* proxy = jsCast<ProxyObject*>(cell);
    if (proxy->m_isCallable) {
        callData.type = CallData::Type::Native;
        callData.native.function = performProxyCall;
    }
    return callData;
}

static EncodedJSValue JSC_HOST_CALL performProxyConstruct(ExecState* exec)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    ProxyObject* proxy = jsCast<ProxyObject*>(exec->callee());

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return encodedJSUndefined();
    }

    JSValue handlerValue = proxy->handler();
    if (handlerValue.isNull())
        return throwVMTypeError(exec, vm, s_proxyAlreadyRevokedErrorMessage);

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue constructMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "construct"), ASCIILiteral("'construct' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    JSObject* target = proxy->target();
    if (constructMethod.isUndefined()) {
        auto constructData = getConstructData(vm, target);
        RELEASE_ASSERT(constructData.type != CallData::Type::None);
        if (auto* result = construct(exec, target, constructData, ArgList(exec), exec->newTarget()))
        	return JSValue::encode(result);
		return encodedJSUndefined();
    }

    JSArray* argArray = constructArray(exec, nullptr, ArgList(exec));
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(argArray);
    arguments.append(exec->newTarget());
    JSValue result = call(exec, constructMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (!result.isObject())
        return throwVMTypeError(exec, vm, ASCIILiteral("Result from Proxy handler's 'construct' method should be an object"));
    return JSValue::encode(result);
}

CallData ProxyObject::getConstructData(JSCell* cell)
{
    CallData constructData;
    ProxyObject* proxy = jsCast<ProxyObject*>(cell);
    if (proxy->m_isConstructible) {
        constructData.type = CallData::Type::Native;
        constructData.native.function = performProxyConstruct;
    }
    return constructData;
}

template <typename DefaultDeleteFunction>
bool ProxyObject::performDelete(ExecState* exec, PropertyName propertyName, DefaultDeleteFunction performDefaultDelete)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    if (propertyName.isPrivateName())
        return false;

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue deletePropertyMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "deleteProperty"), ASCIILiteral("'deleteProperty' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);
    JSObject* target = this->target();
    if (deletePropertyMethod.isUndefined())
        return performDefaultDelete();

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(identifierToSafePublicJSValue(vm, Identifier::fromUid(vm, propertyName.uid())));
    JSValue trapResult = call(exec, deletePropertyMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);

    bool trapResultAsBool = trapResult.toBoolean(exec);
    RETURN_IF_EXCEPTION(vm, false);

    if (!trapResultAsBool)
        return false;

    PropertyDescriptor descriptor;
    if (target->getOwnPropertyDescriptor(exec, propertyName, descriptor)) {
        if (!descriptor.configurable()) {
            throwTypeError(exec, vm, ASCIILiteral("Proxy handler's 'deleteProperty' method should return false when the target's property is not configurable"));
            return false;
        }
        bool targetIsExtensible = target->isExtensible(exec);
        RETURN_IF_EXCEPTION(vm, false);
        if (!targetIsExtensible) {
            throwVMTypeError(exec, vm, ASCIILiteral("Proxy handler's 'deleteProperty' method should return false when the target has property and is not extensible"));
            return false;
        }
    }

    RETURN_IF_EXCEPTION(vm, false);

    return true;
}

bool ProxyObject::deleteProperty(JSCell* cell, ExecState* exec, PropertyName propertyName)
{
    ProxyObject* thisObject = jsCast<ProxyObject*>(cell);
    class PerformDefaultDelete
    {
    public:
    	PerformDefaultDelete(ProxyObject* thisObject, ExecState* exec, PropertyName propertyName)
    	: thisObject(thisObject)
    	, exec(exec)
    	, propertyName(propertyName)
    	{
    	}

    	bool operator()() {
    	    JSObject* target = thisObject->target();
    	    return target->methodTable(exec->vm())->deleteProperty(target, exec, propertyName);
    	};

    private:
    	ProxyObject* thisObject;
    	ExecState* exec;
    	PropertyName propertyName;
    };
    PerformDefaultDelete performDefaultDelete(thisObject, exec, propertyName);
    return thisObject->performDelete(exec, propertyName, performDefaultDelete);
}

bool ProxyObject::deletePropertyByIndex(JSCell* cell, ExecState* exec, unsigned propertyName)
{
	VM& vm = exec->vm();
    ProxyObject* thisObject = jsCast<ProxyObject*>(cell);
    Identifier ident = Identifier::from(vm, propertyName);
    class PerformDefaultDeleteByIndex
    {
    public:
    	PerformDefaultDeleteByIndex(ProxyObject* thisObject, ExecState* exec, unsigned propertyName)
    	: thisObject(thisObject)
    	, exec(exec)
    	, propertyName(propertyName)
    	{
    	}

    	bool operator()() {
            JSObject* target = thisObject->target();
            return target->methodTable(exec->vm())->deletePropertyByIndex(target, exec, propertyName);
    	};

    private:
    	ProxyObject* thisObject;
    	ExecState* exec;
    	unsigned propertyName;
    };
	PerformDefaultDeleteByIndex performDefaultDelete(thisObject, exec, propertyName);
    return thisObject->performDelete(exec, ident, performDefaultDelete);
}

bool ProxyObject::performPreventExtensions(ExecState* exec)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue preventExtensionsMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "preventExtensions"), ASCIILiteral("'preventExtensions' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);
    JSObject* target = this->target();
    if (preventExtensionsMethod.isUndefined())
        return target->methodTable(vm)->preventExtensions(target, exec);

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    JSValue trapResult = call(exec, preventExtensionsMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);

    bool trapResultAsBool = trapResult.toBoolean(exec);
    RETURN_IF_EXCEPTION(vm, false);

    if (trapResultAsBool) {
        bool targetIsExtensible = target->isExtensible(exec);
        RETURN_IF_EXCEPTION(vm, false);
        if (targetIsExtensible) {
            throwTypeError(exec, vm, ASCIILiteral("Proxy's 'preventExtensions' trap returned true even though its target is extensible. It should have returned false"));
            return false;
        }
    }

    return trapResultAsBool;
}

bool ProxyObject::preventExtensions(JSObject* object, ExecState* exec)
{
    return jsCast<ProxyObject*>(object)->performPreventExtensions(exec);
}

bool ProxyObject::performIsExtensible(ExecState* exec)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue isExtensibleMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "isExtensible"), ASCIILiteral("'isExtensible' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);

    JSObject* target = this->target();
    if (isExtensibleMethod.isUndefined())
        return target->isExtensible(exec);

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    JSValue trapResult = call(exec, isExtensibleMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);

    bool trapResultAsBool = trapResult.toBoolean(exec);
    RETURN_IF_EXCEPTION(vm, false);

    bool isTargetExtensible = target->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, false);

    if (trapResultAsBool != isTargetExtensible) {
        if (isTargetExtensible) {
            ASSERT(!trapResultAsBool);
            throwTypeError(exec, vm, ASCIILiteral("Proxy object's 'isExtensible' trap returned false when the target is extensible. It should have returned true"));
        } else {
            ASSERT(!isTargetExtensible);
            ASSERT(trapResultAsBool);
            throwTypeError(exec, vm, ASCIILiteral("Proxy object's 'isExtensible' trap returned true when the target is non-extensible. It should have returned false"));
        }
    }

    return trapResultAsBool;
}

bool ProxyObject::isExtensible(JSObject* object, ExecState* exec)
{
    return jsCast<ProxyObject*>(object)->performIsExtensible(exec);
}

bool ProxyObject::performDefineOwnProperty(ExecState* exec, PropertyName propertyName, const PropertyDescriptor& descriptor, bool shouldThrow)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    JSObject* target = this->target();
    //auto performDefaultDefineOwnProperty = [&] {
    //    return target->methodTable(vm)->defineOwnProperty(target, exec, propertyName, descriptor, shouldThrow);
    //};

    if (propertyName.isPrivateName())
        return false;

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue definePropertyMethod = handler->getMethod(exec, callData, vm.propertyNames->defineProperty, ASCIILiteral("'defineProperty' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);

    if (definePropertyMethod.isUndefined())
        return target->methodTable(vm)->defineOwnProperty(target, exec, propertyName, descriptor, shouldThrow);

    JSObject* descriptorObject = constructObjectFromPropertyDescriptor(exec, descriptor);
    RETURN_IF_EXCEPTION(vm, false);

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(identifierToSafePublicJSValue(vm, Identifier::fromUid(vm, propertyName.uid())));
    arguments.append(descriptorObject);
    JSValue trapResult = call(exec, definePropertyMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);

    bool trapResultAsBool = trapResult.toBoolean(exec);
    RETURN_IF_EXCEPTION(vm, false);

    if (!trapResultAsBool) {
        if (shouldThrow)
            throwTypeError(exec, vm, tryMakeString("Proxy's 'defineProperty' trap returned falsy value for property '", String(propertyName.uid()), "'"));
        return false;
    }

    PropertyDescriptor targetDescriptor;
    bool isTargetDescriptorDefined = target->getOwnPropertyDescriptor(exec, propertyName, targetDescriptor);
    RETURN_IF_EXCEPTION(vm, false);

    bool targetIsExtensible = target->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, false);
    bool settingConfigurableToFalse = descriptor.configurablePresent() && !descriptor.configurable();

    if (!isTargetDescriptorDefined) {
        if (!targetIsExtensible) {
            throwTypeError(exec, vm, ASCIILiteral("Proxy's 'defineProperty' trap returned true even though getOwnPropertyDescriptor of the Proxy's target returned undefined and the target is non-extensible"));
            return false;
        }
        if (settingConfigurableToFalse) {
            throwTypeError(exec, vm, ASCIILiteral("Proxy's 'defineProperty' trap returned true for a non-configurable field even though getOwnPropertyDescriptor of the Proxy's target returned undefined"));
            return false;
        }

        return true;
    }

    ASSERT(isTargetDescriptorDefined);
    bool isCurrentDefined = isTargetDescriptorDefined;
    const PropertyDescriptor& current = targetDescriptor;
    bool throwException = false;
    bool isCompatibleDescriptor = validateAndApplyPropertyDescriptor(exec, 0, propertyName, targetIsExtensible, descriptor, isCurrentDefined, current, throwException);
    RETURN_IF_EXCEPTION(vm, false);
    if (!isCompatibleDescriptor) {
        throwTypeError(exec, vm, ASCIILiteral("Proxy's 'defineProperty' trap did not define a property on its target that is compatible with the trap's input descriptor"));
        return false;
    }
    if (settingConfigurableToFalse && targetDescriptor.configurable()) {
        throwTypeError(exec, vm, ASCIILiteral("Proxy's 'defineProperty' trap did not define a non-configurable property on its target even though the input descriptor to the trap said it must do so"));
        return false;
    }
    if (targetDescriptor.isDataDescriptor() && !targetDescriptor.configurable() && targetDescriptor.writable()) {
        if (descriptor.writablePresent() && !descriptor.writable()) {
            throwTypeError(exec, vm, ASCIILiteral("Proxy's 'defineProperty' trap returned true for a non-writable input descriptor when the target's property is non-configurable and writable"));
            return false;
        }
    }

    return true;
}

bool ProxyObject::defineOwnProperty(JSObject* object, ExecState* exec, PropertyName propertyName, const PropertyDescriptor& descriptor, bool shouldThrow)
{
    ProxyObject* thisObject = jsCast<ProxyObject*>(object);
    return thisObject->performDefineOwnProperty(exec, propertyName, descriptor, shouldThrow);
}

void ProxyObject::performGetOwnPropertyNames(ExecState* exec, PropertyNameArray& propertyNames)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return;
    }

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue ownKeysMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "ownKeys"), ASCIILiteral("'ownKeys' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, void());
    JSObject* target = this->target();
    if (ownKeysMethod.isUndefined()) {
        target->methodTable(exec->vm())->getOwnPropertyNames(target, exec, propertyNames, DontEnumPropertiesMode::Include);
        return;
    }

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    JSValue trapResult = call(exec, ownKeysMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, void());

    HashSet<UniquedStringImpl*> uncheckedResultKeys;
    {
		RuntimeTypeMask resultFilter = 0;
		switch (propertyNames.propertyNameMode()) {
		case PropertyNameMode::Symbols:
			resultFilter = TypeSymbol;
			break;
		case PropertyNameMode::Strings:
			resultFilter = TypeString;
			break;
		case PropertyNameMode::StringsAndSymbols:
			resultFilter = TypeSymbol | TypeString;
			break;
		}
		ASSERT(resultFilter);

        auto addPropName = [&] (JSValue value, RuntimeType type) -> bool {
			static const bool doExitEarly = true;
			static const bool dontExitEarly = false;

			Identifier ident = value.toPropertyKey(vm, exec);
			RETURN_IF_EXCEPTION(vm, doExitEarly);

            if (!uncheckedResultKeys.add(ident.impl()).isNewEntry) {
				throwTypeError(exec, vm, "Proxy handler's 'ownKeys' trap result must not contain any duplicate names");
				return doExitEarly;
			}

            if (type & resultFilter)
                propertyNames.add(ident.impl());
			return dontExitEarly;
		};

		RuntimeTypeMask dontThrowAnExceptionTypeFilter = TypeString | TypeSymbol;
		createListFromArrayLike(exec, trapResult, dontThrowAnExceptionTypeFilter, ASCIILiteral("Proxy handler's 'ownKeys' method must return an object"), ASCIILiteral("Proxy handler's 'ownKeys' method must return an array-like object containing only Strings and Symbols"), addPropName);
		RETURN_IF_EXCEPTION(vm, void());
    }

    bool targetIsExensible = target->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, void());

    PropertyNameArray targetKeys(vm, PropertyNameMode::StringsAndSymbols, PrivateSymbolMode::Exclude);
	target->methodTable(vm)->getOwnPropertyNames(target, exec, targetKeys, DontEnumPropertiesMode::Include);
    RETURN_IF_EXCEPTION(vm, void());
    Vector<UniquedStringImpl*> targetConfigurableKeys;
    Vector<UniquedStringImpl*> targetNonConfigurableKeys;
    for(auto iter = targetKeys.begin(); iter != targetKeys.end(); ++iter) {
    	const Identifier& ident = *iter;
        PropertyDescriptor descriptor;
        bool isPropertyDefined = target->getOwnPropertyDescriptor(exec, ident, descriptor);
        RETURN_IF_EXCEPTION(vm, void());
        if (isPropertyDefined && !descriptor.configurable())
            targetNonConfigurableKeys.append(ident.impl());
        else
            targetConfigurableKeys.append(ident.impl());
    }

    enum ContainedIn { IsContainedIn, IsNotContainedIn };
    auto removeIfContainedInUncheckedResultKeys = [&] (UniquedStringImpl* impl) -> ContainedIn {
		auto iter = uncheckedResultKeys.find(impl);
		if (iter == uncheckedResultKeys.end())
			return IsNotContainedIn;

		uncheckedResultKeys.remove(iter);
		return IsContainedIn;
	};
    for(int i = 0; i < targetNonConfigurableKeys.size(); ++i) {
    	UniquedStringImpl* impl = targetNonConfigurableKeys[i];
        if (removeIfContainedInUncheckedResultKeys(impl) == IsNotContainedIn) {
            throwTypeError(exec, vm, tryMakeString("Proxy object's 'target' has the non-configurable property '", String(impl), "' that was not in the result from the 'ownKeys' trap"));
            return;
        }
    }

    if (!targetIsExensible) {
		for(int i = 0; i < targetConfigurableKeys.size(); ++i) {
			UniquedStringImpl* impl = targetConfigurableKeys[i];
			if (removeIfContainedInUncheckedResultKeys(impl) == IsNotContainedIn) {
				throwTypeError(exec, vm, tryMakeString("Proxy object's non-extensible 'target' has configurable property '", String(impl), "' that was not in the result from the 'ownKeys' trap"));
				return;
			}
		}

		if (uncheckedResultKeys.size()) {
			throwTypeError(exec, vm, ASCIILiteral("Proxy handler's 'ownKeys' method returned a key that was not present in its non-extensible target"));
			return;
		}
    }

}

void ProxyObject::performGetOwnEnumerablePropertyNames(ExecState* exec, PropertyNameArray& propertyNames)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    PropertyNameArray unfilteredNames(vm, propertyNames.propertyNameMode(), propertyNames.privateSymbolMode());
    performGetOwnPropertyNames(exec, unfilteredNames);
    RETURN_IF_EXCEPTION(vm, void());
    // Filtering DontEnum properties is observable in proxies and must occur after the invariant checks pass.
    for(auto iter = unfilteredNames.begin(); iter != unfilteredNames.end(); ++iter) {
		const Identifier& propertyName = *iter;
		PropertySlot slot(this, PropertySlot::InternalMethodType::GetOwnProperty);
		auto isPropertyDefined = getOwnPropertySlotCommon(exec, propertyName, slot);
	    RETURN_IF_EXCEPTION(vm, void());
		if (!isPropertyDefined)
			continue;
		if (slot.attributes() & DontEnum)
			continue;
		propertyNames.add(propertyName.impl());
	}
}

void ProxyObject::getOwnPropertyNames(JSObject* object, ExecState* exec, PropertyNameArray& propertyNameArray, DontEnumPropertiesMode mode)
{
    ProxyObject* thisObject = jsCast<ProxyObject*>(object);
    if (mode == DontEnumPropertiesMode::Include)
        thisObject->performGetOwnPropertyNames(exec, propertyNameArray);
    else
        thisObject->performGetOwnEnumerablePropertyNames(exec, propertyNameArray);
}

bool ProxyObject::performSetPrototype(ExecState* exec, JSValue prototype, bool shouldThrowIfCantSet)
{
    NO_TAIL_CALLS();

    ASSERT(prototype.isObject() || prototype.isNull());

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return false;
    }

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return false;
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue setPrototypeOfMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "setPrototypeOf"), ASCIILiteral("'setPrototypeOf' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, false);

    JSObject* target = this->target();
    if (setPrototypeOfMethod.isUndefined())
        return target->setPrototype(vm, exec, prototype, shouldThrowIfCantSet);

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    arguments.append(prototype);
    JSValue trapResult = call(exec, setPrototypeOfMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, false);

    bool trapResultAsBool = trapResult.toBoolean(exec);
    RETURN_IF_EXCEPTION(vm, false);

    if (!trapResultAsBool) {
        if (shouldThrowIfCantSet)
            throwTypeError(exec, vm, ASCIILiteral("Proxy 'setPrototypeOf' returned false indicating it could not set the prototype value. The operation was expected to succeed"));
        return false;
    }

    bool targetIsExtensible = target->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, false);
    if (targetIsExtensible)
        return true;

    JSValue targetPrototype = target->getPrototype(vm, exec);
    RETURN_IF_EXCEPTION(vm, false);
    bool isSame = sameValue(exec, prototype, targetPrototype);
    RETURN_IF_EXCEPTION(vm, false);
    if (!isSame) {
        throwTypeError(exec, vm, ASCIILiteral("Proxy 'setPrototypeOf' trap returned true when its target is non-extensible and the new prototype value is not the same as the current prototype value. It should have returned false"));
        return false;
    }

    return true;
}

bool ProxyObject::setPrototype(JSObject* object, ExecState* exec, JSValue prototype, bool shouldThrowIfCantSet)
{
    return jsCast<ProxyObject*>(object)->performSetPrototype(exec, prototype, shouldThrowIfCantSet);
}

JSValue ProxyObject::performGetPrototype(ExecState* exec)
{
    NO_TAIL_CALLS();

    VM& vm = exec->vm();

    StackEntryScope stackEntryScope(vm, Options::maxPerThreadStackUsage());
    if (UNLIKELY(!stackEntryScope.isSafeToRecurse())) {
        throwStackOverflowError(exec);
        return jsUndefined();
    }

    JSValue handlerValue = this->handler();
    if (handlerValue.isNull()) {
        throwTypeError(exec, vm, String(s_proxyAlreadyRevokedErrorMessage));
        return jsUndefined();
    }

    JSObject* handler = jsCast<JSObject*>(handlerValue);
    CallData callData;
    JSValue getPrototypeOfMethod = handler->getMethod(exec, callData, makeIdentifier(vm, "getPrototypeOf"), ASCIILiteral("'getPrototypeOf' property of a Proxy's handler should be callable"));
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    JSObject* target = this->target();
    if (getPrototypeOfMethod.isUndefined())
        return target->getPrototype(vm, exec);

    MarkedArgumentBuffer arguments;
    arguments.append(target);
    JSValue trapResult = call(exec, getPrototypeOfMethod, callData, handler, arguments);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    if (!trapResult.isObject() && !trapResult.isNull()) {
        throwTypeError(exec, vm, ASCIILiteral("Proxy handler's 'getPrototypeOf' trap should either return an object or null"));
        return jsUndefined();
    }

    bool targetIsExtensible = target->isExtensible(exec);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    if (targetIsExtensible)
        return trapResult;

    JSValue targetPrototype = target->getPrototype(vm, exec);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    bool isSame = sameValue(exec, targetPrototype, trapResult);
    RETURN_IF_EXCEPTION(vm, jsUndefined());
    if (!isSame) {
        throwTypeError(exec, vm, ASCIILiteral("Proxy's 'getPrototypeOf' trap for a non-extensible target should return the same value as the target's prototype"));
        return jsUndefined();
    }

    return trapResult;
}

JSValue ProxyObject::getPrototype(JSObject* object, ExecState* exec)
{
    return jsCast<ProxyObject*>(object)->performGetPrototype(exec);
}

void ProxyObject::revoke(VM& vm)
{
    // This should only ever be called once and we should strictly transition from Object to null.
    RELEASE_ASSERT(!m_handler.get().isNull() && m_handler.get().isObject());
    m_handler.set(vm, this, jsNull());
}

bool ProxyObject::isRevoked() const
{
    return handler().isNull();
}

void ProxyObject::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
    ProxyObject* thisObject = jsCast<ProxyObject*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);

    visitor.append(&thisObject->m_target);
    visitor.append(&thisObject->m_handler);
}

} // namespace JSC
