/*
 *  Copyright (C) 1999-2000 Harri Porten (porten@kde.org)
 *  Copyright (C) 2007, 2008, 2009 Apple Inc. All rights reserved.
 *  Copyright (C) 2009 Torch Mobile, Inc.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef RegExp_h
#define RegExp_h

#include "ExecutableAllocator.h"
#include "MatchResult.h"
#include "RegExpKey.h"
#include "Structure.h"
#include "yarr/Yarr.h"
#include "yarr/YarrInterpreter.h"
#include <wtf/Forward.h>
#include <wtf/RefCounted.h>
#include <wtf/text/WTFString.h>

#if ENABLE(YARR_JIT)
#include "yarr/YarrJIT.h"
#endif

namespace WTF {
class PrintStream;
}

namespace JSC {

struct RegExpRepresentation;
class JSGlobalData;

class RegExp : public JSCell {
public:
	typedef JSCell Base;
	static const unsigned StructureFlags = StructureIsImmortal | Base::StructureFlags;

	JS_EXPORT_PRIVATE static RegExp* create(JSGlobalData&, const String& pattern, OptionSet<Yarr::Flags>);
	static const bool needsDestruction = true;
	static void destroy(JSCell*);
	void dump(WTF::PrintStream&) const;
	JS_EXPORT_PRIVATE static void dumpToStream(const JSCell*, WTF::PrintStream&);

    bool global() const { return m_flags.contains(Yarr::Flags::Global); }
    bool ignoreCase() const { return m_flags.contains(Yarr::Flags::IgnoreCase); }
    bool multiline() const { return m_flags.contains(Yarr::Flags::Multiline); }
    bool sticky() const { return m_flags.contains(Yarr::Flags::Sticky); }
    bool globalOrSticky() const { return global() || sticky(); }
    bool unicode() const { return m_flags.contains(Yarr::Flags::Unicode); }
    bool dotAll() const { return m_flags.contains(Yarr::Flags::DotAll); }
    bool hasIndices() const { return m_flags.contains(Yarr::Flags::HasIndices); }

	const String& pattern() const { return m_patternString; }

    bool isValid() const { return !Yarr::hasError(m_constructionErrorCode); }
    const char* errorMessage() const { return Yarr::errorMessage(m_constructionErrorCode); }
    JSObject* errorToThrow(ExecState* exec) const { return Yarr::errorToThrow(exec, m_constructionErrorCode); }
    void reset()
	{
		m_state = NotCompiled;
		m_constructionErrorCode = Yarr::ErrorCode::NoError;
	}

	JS_EXPORT_PRIVATE int match(JSGlobalData&, const String&, unsigned startOffset, Vector<int, 32>& ovector);

    // Returns false if we couldn't run the regular expression for any reason.
    bool matchConcurrently(VM&, const String&, unsigned startOffset, int& position, Vector<int, 32>& ovector);

	MatchResult match(JSGlobalData&, const String&, unsigned startOffset);

    bool matchConcurrently(VM&, const String&, unsigned startOffset, MatchResult&);

	unsigned numSubpatterns() const { return m_numSubpatterns; }

    bool hasNamedCaptures() const
    {
        return m_rareData && !m_rareData->m_captureGroupNames.isEmpty();
    }

    String getCaptureGroupName(unsigned i) const
    {
        if (!i || !m_rareData || m_rareData->m_captureGroupNames.size() <= i)
            return String();
        ASSERT(m_rareData);
        return m_rareData->m_captureGroupNames[i];
    }

    unsigned subpatternForName(const String& groupName) const
    {
        if (!m_rareData)
            return 0;
        auto it = m_rareData->m_namedGroupToParenIndex.find(groupName);
        if (it == m_rareData->m_namedGroupToParenIndex.end())
            return 0;
        return it->second;
    }

	// Call these versions of the match functions if you're desperate for performance.
	int matchInline(VM&, const String&, unsigned startOffset, Vector<int, 32>& ovector);
	MatchResult matchInline(VM&, const String&, unsigned startOffset);

	bool hasCode() const
	{
        return m_state == JITCode || m_state == ByteCode;
	}

    bool hasCodeFor(Yarr::CharSize);
    bool hasMatchOnlyCodeFor(Yarr::CharSize);

	void deleteCode();

#if ENABLE(REGEXP_TRACING)
	void printTraceData();
#endif

	static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype)
	{
		return Structure::create(globalData, globalObject, prototype, TypeInfo(CellType, StructureFlags), info());
	}

	DECLARE_INFO;

	RegExpKey key() { return RegExpKey(m_flags, m_patternString); }

protected:
	void finishCreation(JSGlobalData&);

private:
	friend class RegExpCache;
	RegExp(JSGlobalData&, const String&, OptionSet<Yarr::Flags>);

	static RegExp* createWithoutCaching(JSGlobalData&, const String&, OptionSet<Yarr::Flags>);

    enum RegExpState : uint8_t {
		ParseError,
		JITCode,
		ByteCode,
		NotCompiled
	};

    void byteCodeCompileIfNecessary(JSGlobalData&);

	void compile(JSGlobalData&, Yarr::CharSize);
	void compileIfNecessary(JSGlobalData&, Yarr::CharSize);

	void compileMatchOnly(JSGlobalData&, Yarr::CharSize);
	void compileIfNecessaryMatchOnly(JSGlobalData&, Yarr::CharSize);

#if ENABLE(YARR_JIT_DEBUG)
	void matchCompareWithInterpreter(const String&, int startOffset, int* offsetVector, int jitResult);
#endif

#if ENABLE(YARR_JIT)
    Yarr::YarrCodeBlock& ensureRegExpJITCode()
    {
        if (!m_regExpJITCode)
            m_regExpJITCode = adoptPtr(new Yarr::YarrCodeBlock());
        return *m_regExpJITCode.get();
    }
#endif

    struct RareData {
        WTF_MAKE_FAST_ALLOCATED;
    public:
        Vector<String> m_captureGroupNames;
        HashMap<String, unsigned> m_namedGroupToParenIndex;
    };

	String m_patternString;
    RegExpState m_state;
    OptionSet<Yarr::Flags> m_flags;
	Yarr::ErrorCode m_constructionErrorCode;
	unsigned m_numSubpatterns;
	OwnPtr<Yarr::BytecodePattern> m_regExpBytecode;
#if ENABLE(YARR_JIT)
    OwnPtr<Yarr::YarrCodeBlock> m_regExpJITCode;
#endif
    OwnPtr<RareData> m_rareData;
#if ENABLE(REGEXP_TRACING)
	unsigned m_rtMatchCallCount;
	unsigned m_rtMatchFoundCount;
#endif
	//ConcurrentJITLock m_lock;
};

} // namespace JSC

namespace WTF {

class PrintStream;
inline void printInternal(PrintStream& out, JSC::RegExp& regExp)
{
	JSC::RegExp::dumpToStream(&regExp, out);
}

} // namespace WTF

#endif // RegExp_h
