/*
 *  Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#pragma once

#include "ButterflyInlineMethods.h"
#include "JSArray.h"
#include "JSCInlines.h"
#include "JSGlobalObject.h"
#include "ObjectConstructor.h"
#include "ObjectInitializationScope.h"
#include "RegExpInlines.h"
#include "RegExpObject.h"

namespace JSC {

static const PropertyOffset RegExpMatchesArrayIndexPropertyOffset = 100;
static const PropertyOffset RegExpMatchesArrayInputPropertyOffset = 101;
static const PropertyOffset RegExpMatchesArrayGroupsPropertyOffset = 102;
static const PropertyOffset RegExpMatchesArrayIndicesPropertyOffset = 103;
static const PropertyOffset RegExpMatchesIndicesGroupsPropertyOffset = 100;

inline JSArray* tryCreateUninitializedRegExpMatchesArray(VM& vm, DeferGC& deferGC, Structure* structure, unsigned initialLength)
{
    unsigned vectorLength = std::max(BASE_ARRAY_STORAGE_VECTOR_LEN, initialLength);
    if (vectorLength > MAX_STORAGE_VECTOR_LENGTH)
        return nullptr;

    JSGlobalObject* globalObject = structure->globalObject();
    bool createUninitialized = globalObject->isOriginalArrayStructure(structure);
    void* temp;
    if (UNLIKELY(!vm.heap.tryAllocateStorage(Butterfly::totalSize(0, structure->outOfLineCapacity(), true, vectorLength * sizeof(EncodedJSValue)), &temp)))
        return nullptr;
    Butterfly* butterfly = Butterfly::fromBase(temp, 0, structure->outOfLineCapacity());
    butterfly->setVectorLength(vectorLength);
    butterfly->setPublicLength(initialLength);

    unsigned i = (createUninitialized ? initialLength : 0);
    for (; i < vectorLength; ++i)
        butterfly->contiguous()[i].clear();

    JSArray* result = JSArray::createWithButterfly(vm, deferGC, structure, butterfly);
    //scope.notifyAllocated(result, createUninitialized);
	return result;
}

inline JSArray* createRegExpMatchesArray(
    VM& vm, ExecState* exec, JSGlobalObject* globalObject, JSString* input, const String& inputValue,
    RegExp* regExp, unsigned startOffset, MatchResult& result)
{
    SamplingRegion samplingRegion("createRegExpMatchesArray");

    Vector<int, 32> subpatternResults;
    int position = regExp->matchInline(vm, inputValue, startOffset, subpatternResults);
    if (position == -1) {
        result = MatchResult::failed();
        return nullptr;
    }

    result.start = position;
    result.end = subpatternResults[1];

    JSArray* array;
    JSArray* indicesArray = nullptr;

    // FIXME: This should handle array allocation errors gracefully.
    // https://bugs.webkit.org/show_bug.cgi?id=155144

    unsigned numSubpatterns = regExp->numSubpatterns();
    bool hasNamedCaptures = regExp->hasNamedCaptures();
    bool createIndices = regExp->hasIndices();
    JSObject* groups = hasNamedCaptures ? constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure()) : nullptr;
    Structure* matchStructure = createIndices ? globalObject->regExpMatchesArrayWithIndicesStructure() : globalObject->regExpMatchesArrayStructure();

    JSObject* indicesGroups = createIndices && hasNamedCaptures ? constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure()) : nullptr;

    auto setProperties = [&] () {
        array->putDirect(vm, RegExpMatchesArrayIndexPropertyOffset, jsNumber(result.start));
        array->putDirect(vm, RegExpMatchesArrayInputPropertyOffset, input);
        array->putDirect(vm, RegExpMatchesArrayGroupsPropertyOffset, hasNamedCaptures ? groups : jsUndefined());

		ASSERT(!array->butterfly()->indexingHeader()->preCapacity(matchStructure));
		auto capacity = matchStructure->outOfLineCapacity();
		auto size = matchStructure->outOfLineSize();
		memset(array->butterfly()->base(0, capacity), 0, (capacity - size) * sizeof(JSValue));

        if (createIndices) {
            array->putDirect(vm, RegExpMatchesArrayIndicesPropertyOffset, indicesArray);

            Structure* indicesStructure = globalObject->regExpMatchesIndicesArrayStructure();

            indicesArray->putDirect(vm, RegExpMatchesIndicesGroupsPropertyOffset, indicesGroups ? indicesGroups : jsUndefined());

            ASSERT(!indicesArray->butterfly()->indexingHeader()->preCapacity(indicesStructure));
            auto indicesCapacity = indicesStructure->outOfLineCapacity();
            auto indicesSize = indicesStructure->outOfLineSize();
            memset(static_cast<JSValue*>(indicesArray->butterfly()->base(0, indicesCapacity)), 0, (indicesCapacity - indicesSize) * sizeof(JSValue));
        }
	};

    auto createIndexArrayOrNull = [&] (DeferGC& deferralContext, int start, int end) -> JSValue {
        ObjectInitializationScope scope(vm);
        JSArray* result = JSArray::tryCreateUninitializedRestricted(scope, deferralContext, globalObject->arrayStructureForIndexingTypeDuringAllocation(ArrayWithContiguous), 2);
        if (result) {
        	result->initializeIndex(vm, 0, jsNumber(start));
        	result->initializeIndex(vm, 1, jsNumber(end));
            return result;
        }
		throwOutOfMemoryError(vm, exec);
        return jsNull();
    };

	if (UNLIKELY(globalObject->isHavingABadTime())) {
		DeferGC deferralContext(vm.heap);
        ObjectInitializationScope matchesArrayScope(vm);
        ObjectInitializationScope indicesArrayScope(vm);
        array = JSArray::tryCreateUninitializedRestricted(matchesArrayScope, deferralContext, matchStructure, numSubpatterns + 1);

        // FIXME: we should probably throw an out of memory error here, but
		// when making this change we should check that all clients of this
		// function will correctly handle an exception being thrown from here.
		// https://bugs.webkit.org/show_bug.cgi?id=169786
		RELEASE_ASSERT(array);
        if(UNLIKELY(!array)) {
    		throwOutOfMemoryError(vm, exec);
            result = MatchResult::failed();
            return nullptr;
        }

        if (createIndices) {
            indicesArray = JSArray::tryCreateUninitializedRestricted(indicesArrayScope, deferralContext, globalObject->regExpMatchesIndicesArrayStructure(), numSubpatterns + 1);
            if(UNLIKELY(!indicesArray)) {
        		throwOutOfMemoryError(vm, exec);
                result = MatchResult::failed();
                return nullptr;
            }
        }

        setProperties();

        array->initializeIndex(vm, 0, jsSubstringOfResolved(vm, input, result.start, result.end - result.start));

		for (unsigned i = 1; i <= numSubpatterns; ++i) {
			int start = subpatternResults[2 * i];
			JSValue value;
			if (start >= 0)
				value = JSRopeString::createSubstringOfResolved(vm, input, start, subpatternResults[2 * i + 1] - start);
			else
				value = jsUndefined();
			array->initializeIndex(vm, i, value);
		}

        if (createIndices) {
            for (unsigned i = 0; i <= numSubpatterns; ++i) {
                int start = subpatternResults[2 * i];
                if (start >= 0)
                    indicesArray->initializeIndex(vm, i, createIndexArrayOrNull(deferralContext, start, subpatternResults[2 * i + 1]));
                else
                    indicesArray->initializeIndex(vm, i, jsUndefined());
            }
        }
    } else {
        DeferGC deferralContext(vm.heap);
        ObjectInitializationScope matchesArrayScope(vm);
        ObjectInitializationScope indicesArrayScope(vm);
        array = tryCreateUninitializedRegExpMatchesArray(vm, deferralContext, matchStructure, numSubpatterns + 1);

        // FIXME: we should probably throw an out of memory error here, but
        // when making this change we should check that all clients of this
        // function will correctly handle an exception being thrown from here.
        // https://bugs.webkit.org/show_bug.cgi?id=169786
        if(UNLIKELY(!array)) {
    		throwOutOfMemoryError(vm, exec);
			result = MatchResult::failed();
			return nullptr;
		}

        if (createIndices) {
            indicesArray = tryCreateUninitializedRegExpMatchesArray(vm, deferralContext, globalObject->regExpMatchesIndicesArrayStructure(), numSubpatterns + 1);
            if(UNLIKELY(!indicesArray)) {
        		throwOutOfMemoryError(vm, exec);
				result = MatchResult::failed();
				return nullptr;
			}
        }

        setProperties();

        array->initializeIndex(vm, 0, jsSubstringOfResolved(vm, input, result.start, result.end - result.start), ArrayWithContiguous);

		for (unsigned i = 1; i <= numSubpatterns; ++i) {
			int start = subpatternResults[2 * i];
			JSValue value;
			if (start >= 0)
				value = JSRopeString::createSubstringOfResolved(vm, input, start, subpatternResults[2 * i + 1] - start);
			else
				value = jsUndefined();
			array->initializeIndex(vm, i, value, ArrayWithContiguous);
        }

        if (createIndices) {
            for (unsigned i = 0; i <= numSubpatterns; ++i) {
                int start = subpatternResults[2 * i];
                if (start >= 0)
                    indicesArray->initializeIndex(vm, i, createIndexArrayOrNull(deferralContext, start, subpatternResults[2 * i + 1]));
                else
                    indicesArray->initializeIndex(vm, i, jsUndefined());
            }
        }
    }

    // Now the object is safe to scan by GC.

	// We initialize the groups and indices objects late as they could allocate, which with the current API could cause
	// allocations.
    if (hasNamedCaptures) {
		for (unsigned i = 1; i <= numSubpatterns; ++i) {
			String groupName = regExp->getCaptureGroupName(i);
			if (!groupName.isEmpty()) {
				groups->putDirect(vm, Identifier::fromString(vm, groupName), array->getIndexQuickly(i));
				if (createIndices)
					indicesGroups->putDirect(vm, Identifier::fromString(vm, groupName), indicesArray->getIndexQuickly(i));
			}
		}
    }
    return array;
}

inline JSArray* createRegExpMatchesArray(VM& vm, ExecState* exec, JSGlobalObject* globalObject, JSString* string, RegExp* regExp, unsigned startOffset)
{
    //auto scope = DECLARE_THROW_SCOPE(vm);

    MatchResult ignoredResult;
    String input = string->value(exec);
    RETURN_IF_EXCEPTION(vm, nullptr);

    return createRegExpMatchesArray(vm, exec, globalObject, string, input, regExp, startOffset, ignoredResult);
}
JSArray* createEmptyRegExpMatchesArray(ExecState* exec, JSGlobalObject*, JSString*, RegExp*);
Structure* createRegExpMatchesArrayStructure(JSGlobalData&, JSGlobalObject*);
Structure* createRegExpMatchesArrayWithIndicesStructure(VM&, JSGlobalObject*);
Structure* createRegExpMatchesIndicesArrayStructure(VM&, JSGlobalObject*);
Structure* createRegExpMatchesArraySlowPutStructure(JSGlobalData&, JSGlobalObject*);
Structure* createRegExpMatchesArrayWithIndicesSlowPutStructure(VM&, JSGlobalObject*);
Structure* createRegExpMatchesIndicesArraySlowPutStructure(VM&, JSGlobalObject*);

}
