/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *  Copyright (C) 2009 Torch Mobile, Inc.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"
#include "StringPrototype.h"
#include "StringPrototypeInlines.h"

#include "ButterflyInlineMethods.h"
#include "CachedCall.h"
#include "CopiedSpaceInlineMethods.h"
#include "Error.h"
#include "Executable.h"
#include "InterpreterInlines.h"
#include "IntlCollator.h"
#include "IntlObjectInlines.h"
#include "JSGlobalObjectFunctions.h"
#include "JITCodeInlines.h"
#include "JSArray.h"
#include "JSFunction.h"
#include "JSStringIterator.h"
#include "Lookup.h"
#include "ObjectConstructor.h"
#include "ObjectPrototype.h"
#include "Operations.h"
#include "ParseInt.h"
#include "PropertyNameArray.h"
#include "RegExpCache.h"
#include "RegExpConstructor.h"
#include "RegExpMatchesArray.h"
#include "RegExpObject.h"
#include "JITOperations.h"
#include <unicode/uconfig.h>
#include <unicode/unorm2.h>
#include <unicode/ustring.h>
#include <wtf/ASCIICType.h>
#include <wtf/MathExtras.h>
#include <wtf/text/StringView.h>
#include <wtf/unicode/CharacterNames.h>
#include <wtf/unicode/Collator.h>
#include <wtf/unicode/icu/ICUHelpers.h>

using namespace WTF;

namespace JSC {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(StringPrototype);

static EncodedJSValue JSC_HOST_CALL stringProtoFuncToString(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncCharAt(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncCharCodeAt(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncCodePointAt(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncIndexOf(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncLastIndexOf(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncReplaceUsingRegExp(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncReplaceUsingStringSearch(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncReplaceAllUsingStringSearch(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncSlice(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncSubstr(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncSubstring(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncToLowerCase(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncToUpperCase(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncLocaleCompare(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncToLocaleLowerCase(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncToLocaleUpperCase(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncTrim(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncTrimStart(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncTrimEnd(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncStartsWith(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncEndsWith(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncIncludes(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncNormalize(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncIterator(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncIsWellFormed(ExecState*);
static EncodedJSValue JSC_HOST_CALL stringProtoFuncToWellFormed(ExecState*);

}

#include "StringPrototype.lut.h"

namespace JSC {

const ClassInfo StringPrototype::s_info = { "String", &StringObject::s_info, &stringPrototypeTable, CREATE_METHOD_TABLE(StringPrototype) };

/* Source for StringConstructor.lut.h
@begin stringPrototypeTable
    concat     JSBuiltin    DontEnum|Function 1
    match      JSBuiltin    DontEnum|Function 1
    matchAll   JSBuiltin    DontEnum|Function 1
    padStart   JSBuiltin    DontEnum|Function 1
    padEnd     JSBuiltin    DontEnum|Function 1
    repeat     JSBuiltin    DontEnum|Function 1
    replace    JSBuiltin    DontEnum|Function 2
    replaceAll JSBuiltin    DontEnum|Function 2
    search     JSBuiltin    DontEnum|Function 1
    split      JSBuiltin    DontEnum|Function 1
    anchor     JSBuiltin    DontEnum|Function 1
    big        JSBuiltin    DontEnum|Function 0
    bold       JSBuiltin    DontEnum|Function 0
    blink      JSBuiltin    DontEnum|Function 0
    fixed      JSBuiltin    DontEnum|Function 0
    fontcolor  JSBuiltin    DontEnum|Function 1
    fontsize   JSBuiltin    DontEnum|Function 1
    italics    JSBuiltin    DontEnum|Function 0
    link       JSBuiltin    DontEnum|Function 1
    small      JSBuiltin    DontEnum|Function 0
    strike     JSBuiltin    DontEnum|Function 0
    sub        JSBuiltin    DontEnum|Function 0
    sup        JSBuiltin    DontEnum|Function 0
@end
*/

// ECMA 15.5.4
StringPrototype::StringPrototype(VM& vm, Structure* structure)
    : StringObject(vm, structure)
{
}

void StringPrototype::finishCreation(VM& vm, JSGlobalObject* globalObject)
{
    Base::finishCreation(vm, jsEmptyString(vm));
    ASSERT(inherits(info()));

    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->toString, stringProtoFuncToString, DontEnum, 0, StringPrototypeValueOfIntrinsic);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->valueOf, stringProtoFuncToString, DontEnum, 0, StringPrototypeValueOfIntrinsic);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION("charAt", stringProtoFuncCharAt, DontEnum, 1, CharAtIntrinsic);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION("charCodeAt", stringProtoFuncCharCodeAt, DontEnum, 1, CharCodeAtIntrinsic);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION("codePointAt", stringProtoFuncCodePointAt, DontEnum, 1, StringPrototypeCodePointAtIntrinsic);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("indexOf", stringProtoFuncIndexOf, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("lastIndexOf", stringProtoFuncLastIndexOf, DontEnum, 1);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().replaceUsingRegExpPrivateName(), stringProtoFuncReplaceUsingRegExp, DontEnum, 2, StringPrototypeReplaceRegExpIntrinsic);
	JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().replaceUsingStringSearchPrivateName(), stringProtoFuncReplaceUsingStringSearch, DontEnum, 2);
	JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().replaceAllUsingStringSearchPrivateName(), stringProtoFuncReplaceAllUsingStringSearch, DontEnum, 2);
	JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION("slice", stringProtoFuncSlice, DontEnum, 2, StringPrototypeSliceIntrinsic);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("substr", stringProtoFuncSubstr, DontEnum, 2);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("substring", stringProtoFuncSubstring, DontEnum, 2);
    JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION("toLowerCase", stringProtoFuncToLowerCase, DontEnum, 0, StringPrototypeToLowerCaseIntrinsic);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("toUpperCase", stringProtoFuncToUpperCase, DontEnum, 0);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("localeCompare", stringProtoFuncLocaleCompare, static_cast<unsigned>(PropertyAttribute::DontEnum), 1);
#if ENABLE(INTL)
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("toLocaleLowerCase", stringProtoFuncToLocaleLowerCase, DontEnum, 0);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("toLocaleUpperCase", stringProtoFuncToLocaleUpperCase, DontEnum, 0);
#else
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("toLocaleLowerCase", stringProtoFuncToLowerCase, DontEnum, 0);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("toLocaleUpperCase", stringProtoFuncToUpperCase, DontEnum, 0);
#endif
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("trim", stringProtoFuncTrim, DontEnum, 0);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("startsWith", stringProtoFuncStartsWith, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("endsWith", stringProtoFuncEndsWith, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("includes", stringProtoFuncIncludes, DontEnum, 1);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION("normalize", stringProtoFuncNormalize, DontEnum, 0);
	JSC_NATIVE_INTRINSIC_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().charCodeAtPrivateName(), stringProtoFuncCharCodeAt, DontEnum, 1, CharCodeAtIntrinsic);

    if (Options::useAtMethod())
        JSC_BUILTIN_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->builtinNames().atPublicName(), stringPrototypeAtCodeGenerator, static_cast<unsigned>(PropertyAttribute::DontEnum));

    JSFunction* trimStartFunction = JSFunction::create(vm, globalObject, 0, ASCIILiteral("trimStart"), stringProtoFuncTrimStart, NoIntrinsic);
    JSFunction* trimEndFunction = JSFunction::create(vm, globalObject, 0, ASCIILiteral("trimEnd"), stringProtoFuncTrimEnd, NoIntrinsic);
    putDirectWithoutTransition(vm, Identifier::fromString(vm, "trimStart"), trimStartFunction, static_cast<unsigned>(DontEnum));
    putDirectWithoutTransition(vm, Identifier::fromString(vm, "trimLeft"), trimStartFunction, static_cast<unsigned>(DontEnum));
    putDirectWithoutTransition(vm, Identifier::fromString(vm, "trimEnd"), trimEndFunction, static_cast<unsigned>(DontEnum));
    putDirectWithoutTransition(vm, Identifier::fromString(vm, "trimRight"), trimEndFunction, static_cast<unsigned>(DontEnum));

    JSFunction* iteratorFunction = JSFunction::create(vm, globalObject, 0, ASCIILiteral("[Symbol.iterator]"), stringProtoFuncIterator, NoIntrinsic);
    putDirectWithoutTransition(vm, vm.propertyNames->iteratorSymbol, iteratorFunction, DontEnum);

    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->isWellFormed, stringProtoFuncIsWellFormed, static_cast<unsigned>(PropertyAttribute::DontEnum), 0/*, ImplementationVisibility::Public*/);
    JSC_NATIVE_FUNCTION_WITHOUT_TRANSITION(vm.propertyNames->toWellFormed, stringProtoFuncToWellFormed, static_cast<unsigned>(PropertyAttribute::DontEnum), 0/*, ImplementationVisibility::Public*/);

    // The constructor will be added later, after StringConstructor has been built
}

StringPrototype* StringPrototype::create(VM& vm, JSGlobalObject* globalObject, Structure* structure)
{
    StringPrototype* prototype = new (NotNull, allocateCell<StringPrototype>(vm.heap)) StringPrototype(vm, structure);
    prototype->finishCreation(vm, globalObject);
    return prototype;
}

// ------------------------------ Functions --------------------------

static NEVER_INLINE void substituteBackreferencesSlow(WTF::StringBuilder& result, StringView replacement, StringView source, const int* ovector, RegExp* reg, size_t i)
{
    bool hasNamedCaptures = reg && reg->hasNamedCaptures();
    int offset = 0;
    do {
        if (i + 1 == replacement.length())
            break;

        UChar ref = replacement[i + 1];
        if (ref == '$') {
            // "$$" -> "$"
            ++i;
            result.append(replacement.substring(offset, i - offset));
            offset = i + 1;
            continue;
        }

        int backrefStart;
        int backrefLength;
        int advance = 0;
        if (ref == '&') {
            backrefStart = ovector[0];
            backrefLength = ovector[1] - backrefStart;
        } else if (ref == '`') {
            backrefStart = 0;
            backrefLength = ovector[0];
        } else if (ref == '\'') {
            backrefStart = ovector[1];
            backrefLength = source.length() - backrefStart;
        } else if (reg && ref == '<') {
            // Named back reference
            if (!hasNamedCaptures)
                continue;

            size_t closingBracket = replacement.find('>', i + 2);
            if (closingBracket == WTF::notFound)
                continue;

            unsigned nameLength = closingBracket - i - 2;
            unsigned backrefIndex = reg->subpatternForName(replacement.substring(i + 2, nameLength).toString());

            if (!backrefIndex || backrefIndex > reg->numSubpatterns()) {
                backrefStart = 0;
                backrefLength = 0;
            } else {
                backrefStart = ovector[2 * backrefIndex];
                backrefLength = ovector[2 * backrefIndex + 1] - backrefStart;
            }
            advance = nameLength + 1;
        } else if (reg && ref >= '0' && ref <= '9') {
            // 1- and 2-digit back references are allowed
            unsigned backrefIndex = ref - '0';
            if (backrefIndex > reg->numSubpatterns())
                continue;
            if (replacement.length() > i + 2) {
                ref = replacement[i + 2];
                if (ref >= '0' && ref <= '9') {
                    backrefIndex = 10 * backrefIndex + ref - '0';
                    if (backrefIndex > reg->numSubpatterns())
                        backrefIndex = backrefIndex / 10;   // Fall back to the 1-digit reference
                    else
                        advance = 1;
                }
            }
            if (!backrefIndex)
                continue;
            backrefStart = ovector[2 * backrefIndex];
            backrefLength = ovector[2 * backrefIndex + 1] - backrefStart;
        } else
            continue;

        if (i - offset)
        	result.append(replacement.substring(offset, i - offset));
        i += 1 + advance;
        offset = i + 1;
        if (backrefStart >= 0)
        	result.append(source.substring(backrefStart, backrefLength));
    } while ((i = replacement.find('$', i + 1)) != notFound);

    if (replacement.length() - offset)
    	result.append(replacement.substring(offset));
}

inline void substituteBackreferencesInline(WTF::StringBuilder& result, const String& replacement, StringView source, const int* ovector, RegExp* reg)
{
    size_t i = replacement.find('$');
    if (UNLIKELY(i != notFound))
        return substituteBackreferencesSlow(result, replacement, source, ovector, reg, i);

    result.append(replacement);
}

void substituteBackreferences(WTF::StringBuilder& result, const String& replacement, StringView source, const int* ovector, RegExp* reg)
{
    substituteBackreferencesInline(result, replacement, source, ovector, reg);
}

namespace {
struct StringRange {
	ALWAYS_INLINE StringRange(int pos, int len)
        : position(pos)
        , length(len)
    {
    }

	ALWAYS_INLINE StringRange()
    : position(0)
    , length(0)
    {
    }

    int position;
    int length;
};
}

static ALWAYS_INLINE JSValue jsSpliceSubstrings(VM& vm, ExecState* exec, JSString* sourceVal, const String& source, const StringRange* substringRanges, int rangeCount)
{
    if (rangeCount == 1) {
        int sourceSize = source.length();
        int position = substringRanges[0].position;
        int length = substringRanges[0].length;
        if (position <= 0 && length >= sourceSize)
            return sourceVal;
        // We could call String::substringSharingImpl(), but this would result in redundant checks.
        return jsString(vm, StringImpl::create(source.impl(), std::max(0, position), std::min(sourceSize, length)));
    }

    int totalLength = 0;
    for (int i = 0; i < rangeCount; i++)
        totalLength += substringRanges[i].length;

    if (!totalLength)
        return jsEmptyString(vm);

    if (source.is8Bit()) {
        LChar* buffer;
        const LChar* sourceData = source.characters8();
        RefPtr<StringImpl> impl = StringImpl::tryCreateUninitialized(totalLength, buffer);
        if (!impl)
            return throwOutOfMemoryError(exec);

        int bufferPos = 0;
        for (int i = 0; i < rangeCount; i++) {
            if (int srcLen = substringRanges[i].length) {
                StringImpl::copyChars(buffer + bufferPos, sourceData + substringRanges[i].position, srcLen);
                bufferPos += srcLen;
            }
        }

        return jsString(vm, impl.release());
    }

    UChar* buffer;
    const UChar* sourceData = source.characters16();

    RefPtr<StringImpl> impl = StringImpl::tryCreateUninitialized(totalLength, buffer);
    if (!impl)
        return throwOutOfMemoryError(exec);

    int bufferPos = 0;
    for (int i = 0; i < rangeCount; i++) {
        if (int srcLen = substringRanges[i].length) {
            StringImpl::copyChars(buffer + bufferPos, sourceData + substringRanges[i].position, srcLen);
            bufferPos += srcLen;
        }
    }

    return jsString(vm, impl.release());
}

static ALWAYS_INLINE JSValue jsSpliceSubstringsWithSeparators(VM& vm, ExecState* exec, JSString* sourceVal, const String& source, const StringRange* substringRanges, int rangeCount, const String* separators, int separatorCount)
{
    if (rangeCount == 1 && separatorCount == 0) {
        int sourceSize = source.length();
        int position = substringRanges[0].position;
        int length = substringRanges[0].length;
        if (position <= 0 && length >= sourceSize)
            return sourceVal;
        // We could call String::substringSharingImpl(), but this would result in redundant checks.
        return jsString(vm, StringImpl::create(source.impl(), std::max(0, position), std::min(sourceSize, length)));
    }

    if (rangeCount == 2 && separatorCount == 1) {
        String leftPart(StringImpl::createSubstringSharingImpl(*source.impl(), substringRanges[0].position, substringRanges[0].length));
        String rightPart(StringImpl::createSubstringSharingImpl(*source.impl(), substringRanges[1].position, substringRanges[1].length));
        return jsString(vm, exec, leftPart, separators[0], rightPart);
    }

    Checked<int, RecordOverflow> totalLength = 0;
    bool allSeparators8Bit = true;
    for (int i = 0; i < rangeCount; i++)
        totalLength += substringRanges[i].length;
    for (int i = 0; i < separatorCount; i++) {
        totalLength += separators[i].length();
        if (separators[i].length() && !separators[i].is8Bit())
        	allSeparators8Bit = false;
    }
    if (totalLength.hasOverflowed())
        return throwOutOfMemoryError(exec);

    if (!totalLength)
        return jsEmptyString(vm);

    if (source.is8Bit() && allSeparators8Bit) {
        LChar* buffer;
        const LChar* sourceData = source.characters8();

        RefPtr<StringImpl> impl = StringImpl::tryCreateUninitialized(totalLength.unsafeGet(), buffer);
        if (!impl)
            return throwOutOfMemoryError(exec);

        int maxCount = std::max(rangeCount, separatorCount);
        int bufferPos = 0;
        for (int i = 0; i < maxCount; i++) {
            if (i < rangeCount) {
                if (int srcLen = substringRanges[i].length) {
                    StringImpl::copyChars(buffer + bufferPos, sourceData + substringRanges[i].position, srcLen);
                    bufferPos += srcLen;
                }
            }
            if (i < separatorCount) {
                if (int sepLen = separators[i].length()) {
                    StringImpl::copyChars(buffer + bufferPos, separators[i].characters8(), sepLen);
                    bufferPos += sepLen;
                }
            }
        }        

        return jsString(vm, impl.release());
    }

    UChar* buffer;
    RefPtr<StringImpl> impl = StringImpl::tryCreateUninitialized(totalLength.unsafeGet(), buffer);
    if (!impl)
        return throwOutOfMemoryError(exec);

    int maxCount = std::max(rangeCount, separatorCount);
    int bufferPos = 0;
    for (int i = 0; i < maxCount; i++) {
        if (i < rangeCount) {
            if (int srcLen = substringRanges[i].length) {
                if (source.is8Bit())
                    StringImpl::copyChars(buffer + bufferPos, source.characters8() + substringRanges[i].position, srcLen);
                else
                    StringImpl::copyChars(buffer + bufferPos, source.characters16() + substringRanges[i].position, srcLen);
                bufferPos += srcLen;
            }
        }
        if (i < separatorCount) {
            if (int sepLen = separators[i].length()) {
                if (separators[i].is8Bit())
                    StringImpl::copyChars(buffer + bufferPos, separators[i].characters8(), sepLen);
                else
                    StringImpl::copyChars(buffer + bufferPos, separators[i].characters16(), sepLen);
                bufferPos += sepLen;
            }
        }
    }

    return jsString(vm, impl.release());
}

#define OUT_OF_MEMORY(exec__) \
    do { \
        throwOutOfMemoryError(exec__); \
        return encodedJSUndefined(); \
    } while (false)

static ALWAYS_INLINE EncodedJSValue removeUsingRegExpSearch(VM& vm, ExecState* exec, JSString* string, const String& source, RegExp* regExp)
{
    size_t lastIndex = 0;
    unsigned startPosition = 0;

    Vector<StringRange, 16> sourceRanges;
    RegExpConstructor* regExpConstructor = exec->lexicalGlobalObject()->regExpConstructor();
    unsigned sourceLen = source.length();

    while (true) {
        MatchResult result = regExpConstructor->performMatch(vm, regExp, string, source, startPosition);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (!result)
            break;

        if (lastIndex < result.start) {
            if (UNLIKELY(!sourceRanges.tryConstructAndAppend(lastIndex, result.start - lastIndex)))
                OUT_OF_MEMORY(exec);
        }
        lastIndex = result.end;
        startPosition = lastIndex;

        // special case of empty match
        if (result.empty()) {
            startPosition++;
            if (startPosition > sourceLen)
                break;
        }
    }

    if (!lastIndex)
        return JSValue::encode(string);

    if (static_cast<unsigned>(lastIndex) < sourceLen) {
        if (UNLIKELY(!sourceRanges.tryConstructAndAppend(lastIndex, sourceLen - lastIndex)))
            OUT_OF_MEMORY(exec);
    }
    return JSValue::encode(jsSpliceSubstrings(vm, exec, string, source, sourceRanges.data(), sourceRanges.size()));
}

static ALWAYS_INLINE EncodedJSValue replaceUsingRegExpSearch(
	VM& vm, ExecState* exec, JSString* string, JSValue searchValue, CallData& callData,
    String& replacementString, JSValue replaceValue)
{
    const String& source = string->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    unsigned sourceLen = source.length();
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    RegExpObject* regExpObject = jsCast<RegExpObject*>(searchValue);
    RegExp* regExp = regExpObject->regExp();
    bool global = regExp->global();
    bool hasNamedCaptures = regExp->hasNamedCaptures();

    if (global) {
        // ES5.1 15.5.4.10 step 8.a.
        regExpObject->setLastIndex(exec, vm, 0);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        if (callData.type == CallData::Type::None && !replacementString.length())
            return removeUsingRegExpSearch(vm, exec, string, source, regExp);
    }

    // FIXME: This is wrong because we may be called directly from the FTL.
    // https://bugs.webkit.org/show_bug.cgi?id=154874
    RegExpConstructor* regExpConstructor = exec->lexicalGlobalObject()->regExpConstructor();

    size_t lastIndex = 0;
    unsigned startPosition = 0;

    Vector<StringRange, 16> sourceRanges;
    Vector<String, 16> replacements;

    // This is either a loop (if global is set) or a one-way (if not).
    if (global && callData.type == CallData::Type::JS) {
        // regExp->numSubpatterns() + 1 for pattern args, + 2 for match start and string
    	int argCount = regExp->numSubpatterns() + 1 + 2;
		if (hasNamedCaptures)
			++argCount;
        JSFunction* func = jsCast<JSFunction*>(replaceValue);
        CachedCall cachedCall(exec, vm, func, argCount);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        while (true) {
			int* ovector;
			MatchResult result = regExpConstructor->performMatch(vm, regExp, string, source, startPosition, &ovector);
	        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
			if (!result)
				break;

            if (UNLIKELY(!sourceRanges.tryConstructAndAppend(lastIndex, result.start - lastIndex)))
                OUT_OF_MEMORY(exec);

            JSGlobalObject* globalObject = exec->lexicalGlobalObject();
            JSObject* groups = hasNamedCaptures ? constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure()) : nullptr;

			unsigned i = 0;

			for (; i < regExp->numSubpatterns() + 1; ++i) {
				int matchStart = ovector[i * 2];
				int matchLen = ovector[i * 2 + 1] - matchStart;

                JSValue patternValue;

				if (matchStart < 0)
                    patternValue = jsUndefined();
				else
					patternValue = jsSubstring(&vm, source, matchStart, matchLen);

				cachedCall.setArgument(i, patternValue);

                if (i && hasNamedCaptures) {
                    String groupName = regExp->getCaptureGroupName(i);
                    if (!groupName.isEmpty())
                        groups->putDirect(vm, Identifier::fromString(vm, groupName), patternValue);
                }
			}

			cachedCall.setArgument(i++, jsNumber(result.start));
			cachedCall.setArgument(i++, string);
            if (hasNamedCaptures)
                cachedCall.setArgument(i++, groups);

			cachedCall.setThis(jsUndefined());
			JSValue jsResult = cachedCall.call();
	        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
			replacements.append(jsResult.toWTFString(exec));
	        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

			lastIndex = result.end;
			startPosition = lastIndex;

			// special case of empty match
			if (result.empty()) {
				startPosition++;
				if (startPosition > sourceLen)
					break;
                if (U16_IS_LEAD(source[startPosition - 1]) && U16_IS_TRAIL(source[startPosition])) {
                    startPosition++;
                    if (startPosition > sourceLen)
                        break;
                }
			}
		}
    } else {
        do {
            int* ovector;
            MatchResult result = regExpConstructor->performMatch(vm, regExp, string, source, startPosition, &ovector);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            if (!result)
                break;

            if (callData.type != CallData::Type::None) {
                if (UNLIKELY(!sourceRanges.tryConstructAndAppend(lastIndex, result.start - lastIndex)))
                    OUT_OF_MEMORY(exec);

                MarkedArgumentBuffer args;
                JSGlobalObject* globalObject = exec->lexicalGlobalObject();
                JSObject* groups = hasNamedCaptures ? constructEmptyObject(vm, globalObject->nullPrototypeObjectStructure()) : nullptr;

                for (unsigned i = 0; i < regExp->numSubpatterns() + 1; ++i) {
                    int matchStart = ovector[i * 2];
                    int matchLen = ovector[i * 2 + 1] - matchStart;

                    JSValue patternValue;

                    if (matchStart < 0)
                        patternValue = jsUndefined();
                    else {
                        patternValue = jsSubstring(vm, source, matchStart, matchLen);
                        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
                    }

                    args.append(patternValue);

                    if (i && hasNamedCaptures) {
                        String groupName = regExp->getCaptureGroupName(i);
                        if (!groupName.isEmpty())
                            groups->putDirect(vm, Identifier::fromString(vm, groupName), patternValue);
                    }
                }

                args.append(jsNumber(result.start));
                args.append(string);
                if (hasNamedCaptures)
                    args.append(groups);

                JSValue replacement = call(vm, exec, replaceValue, callData, jsUndefined(), args);
                RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
                String replacementString = replacement.toWTFString(exec);
                RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
                replacements.append(replacementString);
                RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            } else {
                int replLen = replacementString.length();
                if (lastIndex < result.start || replLen) {
                    if (UNLIKELY(!sourceRanges.tryConstructAndAppend(lastIndex, result.start - lastIndex)))
                        OUT_OF_MEMORY(exec);

                    if (replLen) {
                    	WTF::StringBuilder replacement(WTF::StringBuilder::OverflowHandler::RecordOverflow);
                        substituteBackreferences(replacement, replacementString, source, ovector, regExp);
                        if (UNLIKELY(replacement.hasOverflowed())) {
                        	throwOutOfMemoryError(vm, exec);
                        	return encodedJSUndefined();
                        }
                        replacements.append(replacement.toString());
                    } else
                        replacements.append(String());
                }
            }

            lastIndex = result.end;
            startPosition = lastIndex;

            // special case of empty match
            if (result.empty()) {
                startPosition++;
                if (startPosition > sourceLen)
                    break;
                if (U16_IS_LEAD(source[startPosition - 1]) && U16_IS_TRAIL(source[startPosition])) {
                    startPosition++;
                    if (startPosition > sourceLen)
                        break;
                }
            }
        } while (global);
    }

    if (!lastIndex && replacements.isEmpty())
        return JSValue::encode(string);

    if (static_cast<unsigned>(lastIndex) < sourceLen) {
        if (UNLIKELY(!sourceRanges.tryConstructAndAppend(lastIndex, sourceLen - lastIndex)))
            OUT_OF_MEMORY(exec);
    }
    return JSValue::encode(jsSpliceSubstringsWithSeparators(vm, exec, string, source, sourceRanges.data(), sourceRanges.size(), replacements.data(), replacements.size()));
}

EncodedJSValue JIT_OPERATION operationStringProtoFuncReplaceRegExpEmptyStr(
    ExecState* exec, JSString* thisValue, RegExpObject* searchValue)
{
    JSGlobalData& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    RegExp* regExp = searchValue->regExp();
    if (regExp->global()) {
        // ES5.1 15.5.4.10 step 8.a.
        searchValue->setLastIndex(exec, vm, 0);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        const String& source = thisValue->value(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        return removeUsingRegExpSearch(vm, exec, thisValue, source, regExp);
    }

    CallData callData;
    String replacementString = emptyString();
    return replaceUsingRegExpSearch(
        vm, exec, thisValue, searchValue, callData, replacementString, JSValue());
}

EncodedJSValue JIT_OPERATION operationStringProtoFuncReplaceRegExpString(
    ExecState* exec, JSString* thisValue, RegExpObject* searchValue, JSString* replaceString)
{
    JSGlobalData& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    CallData callData;
    String replacementString = replaceString->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return replaceUsingRegExpSearch(
        vm, exec, thisValue, searchValue, callData, replacementString, replaceString);
}

static ALWAYS_INLINE EncodedJSValue replaceUsingRegExpSearch(VM& vm, ExecState* exec, JSString* string, JSValue searchValue, JSValue replaceValue)
{
    String replacementString;
    CallData callData = getCallData(vm, replaceValue);
    if (callData.type == CallData::Type::None) {
        replacementString = replaceValue.toWTFString(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    return replaceUsingRegExpSearch(
        vm, exec, string, searchValue, callData, replacementString, replaceValue);
}

enum class ReplaceMode : bool { Single, Global };

static ALWAYS_INLINE EncodedJSValue replaceUsingStringSearch(VM& vm, ExecState* exec, JSString* jsString, JSValue searchValue, JSValue replaceValue, ReplaceMode mode)
{
    const String& string = jsString->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    String searchString = searchValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    CallData callData = getCallData(vm, replaceValue);
    Optional<CachedCall> cachedCall;
    String replaceString;
    if (callData.type == CallData::Type::None) {
        replaceString = replaceValue.toWTFString(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    } else if (callData.type == CallData::Type::JS) {
        cachedCall.emplace(exec, vm, jsCast<JSFunction*>(replaceValue), 3);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        cachedCall->setThis(jsUndefined());
    }

    size_t matchStart = string.find(searchString);
    if (matchStart == notFound)
        return JSValue::encode(jsString);

    size_t endOfLastMatch = 0;
    size_t searchStringLength = searchString.length();
    Vector<StringRange, 16> sourceRanges;
    Vector<String, 16> replacements;
    do {
        if (callData.type != CallData::Type::None) {
            JSValue replacement;
			if (cachedCall) {
				auto* substring = jsSubstring(vm, string, matchStart, searchStringLength);
				RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
				cachedCall->setArgument(0, substring);
				cachedCall->setArgument(1, jsNumber(matchStart));
				cachedCall->setArgument(2, jsString);
				ASSERT(!cachedCall->hasOverflowedArguments());
				replacement = cachedCall->call();
            } else {
                MarkedArgumentBuffer args;
                auto* substring = jsSubstring(vm, string, matchStart, searchString.impl()->length());
    			RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
                args.append(substring);
                args.append(jsNumber(matchStart));
                args.append(jsString);
                ASSERT(!args.hasOverflowed());
                replacement = call(exec, replaceValue, callData, jsUndefined(), args);
            }
			RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
			replaceString = replacement.toWTFString(exec);
			RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
		}

        if (UNLIKELY(!sourceRanges.tryConstructAndAppend(endOfLastMatch, matchStart - endOfLastMatch))) {
			throwOutOfMemoryError(vm, exec);
			return encodedJSUndefined();
        }

        size_t matchEnd = matchStart + searchStringLength;
        if (callData.type != CallData::Type::None)
            replacements.append(replaceString);
        else {
        	WTF::StringBuilder replacement(WTF::StringBuilder::OverflowHandler::RecordOverflow);
            int ovector[2] = { static_cast<int>(matchStart),  static_cast<int>(matchEnd) };
            substituteBackreferences(replacement, replaceString, string, ovector, nullptr);
            if (UNLIKELY(replacement.hasOverflowed())) {
    			throwOutOfMemoryError(vm, exec);
    			return encodedJSUndefined();
            }
            replacements.append(replacement.toString());
        }

        endOfLastMatch = matchEnd;
        if (mode == ReplaceMode::Single)
            break;
        matchStart = string.find(searchString, !searchStringLength ? endOfLastMatch + 1 : endOfLastMatch);
    } while (matchStart != notFound);

    if (UNLIKELY(!sourceRanges.tryConstructAndAppend(endOfLastMatch, string.length() - endOfLastMatch))) {
		throwOutOfMemoryError(vm, exec);
		return encodedJSUndefined();
    }
    return JSValue::encode(jsSpliceSubstringsWithSeparators(vm, exec, jsString, string, sourceRanges.data(), sourceRanges.size(), replacements.data(), replacements.size()));
}

static inline bool checkObjectCoercible(JSValue thisValue)
{
    if (LIKELY(thisValue.isString()))
        return true;

    if (thisValue.isUndefinedOrNull())
        return false;

    if (thisValue.isObject() && asObject(thisValue)->isEnvironmentRecord())
        return false;

    return true;
}

EncodedJSValue JSC_HOST_CALL stringProtoFuncRepeatCharacter(ExecState* exec)
{
	VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    // For a string which length is single, instead of creating ropes,
    // allocating a sequential buffer and fill with the repeated string for efficiency.
    ASSERT(exec->argumentCount() == 2);

    ASSERT(exec->uncheckedArgument(0).isString());
    JSString* string = asString(exec->uncheckedArgument(0));
    ASSERT(string->length() == 1);

    JSValue repeatCountValue = exec->uncheckedArgument(1);
    RELEASE_ASSERT(repeatCountValue.isNumber());
    int32_t repeatCount;
	double value = repeatCountValue.asNumber();
	if (value > JSString::MaxLength())
		return JSValue::encode(throwOutOfMemoryError(vm, exec));
	repeatCount = static_cast<int32_t>(value);
    ASSERT(repeatCount >= 0);
    ASSERT(!repeatCountValue.isDouble() || repeatCountValue.asDouble() == repeatCount);

    auto viewWithString = string->viewWithUnderlyingString(exec);
    StringView view = viewWithString.view;
    ASSERT(view.length() == 1 && !vm.exception());
    UChar character = view[0];
    if (!(character & ~0xff))
        return JSValue::encode(repeatCharacter(vm, exec, static_cast<LChar>(character), repeatCount));
    return JSValue::encode(repeatCharacter(vm, exec, character, repeatCount));
}

ALWAYS_INLINE static EncodedJSValue replace(
    ExecState* exec, JSString* string, JSValue searchValue, JSValue replaceValue)
{
	VM& vm = exec->vm();
    if (searchValue.inherits(RegExpObject::info()))
        return replaceUsingRegExpSearch(vm, exec, string, searchValue, replaceValue);
    return replaceUsingStringSearch(vm, exec, string, searchValue, replaceValue, ReplaceMode::Single);
}

ALWAYS_INLINE static EncodedJSValue replace(
    ExecState* exec, JSValue thisValue, JSValue searchValue, JSValue replaceValue)
{
	VM& vm = exec->vm();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    JSString* string = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return replace(exec, string, searchValue, replaceValue);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncReplaceUsingRegExp(ExecState* exec)
{
	VM& vm = exec->vm();
    JSString* string = exec->thisValue().toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue searchValue = exec->argument(0);
    if (!searchValue.inherits(RegExpObject::info()))
        return JSValue::encode(jsUndefined());

    return replaceUsingRegExpSearch(vm, exec, string, searchValue, exec->argument(1));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncReplaceUsingStringSearch(ExecState* exec)
{
	VM& vm = exec->vm();
    JSString* string = exec->thisValue().toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    return replaceUsingStringSearch(vm, exec, string, exec->argument(0), exec->argument(1), ReplaceMode::Single);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncReplaceAllUsingStringSearch(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSString* string = exec->thisValue().toString(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

    return replaceUsingStringSearch(vm, exec, string, exec->argument(0), exec->argument(1), ReplaceMode::Global);
}

EncodedJSValue JIT_OPERATION operationStringProtoFuncReplaceGeneric(
    ExecState* exec, EncodedJSValue thisValue, EncodedJSValue searchValue,
    EncodedJSValue replaceValue)
{
    JSGlobalData& vm = exec->vm();
    NativeCallFrameTracer tracer(&vm, exec);

    return replace(
        exec, JSValue::decode(thisValue), JSValue::decode(searchValue),
        JSValue::decode(replaceValue));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncToString(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    // Also used for valueOf.

    if (thisValue.isString())
        return JSValue::encode(thisValue);

    auto* stringObject = jsDynamicCast<StringObject*>(thisValue);
    if (stringObject)
        return JSValue::encode(stringObject->internalValue());

    return throwVMTypeError(exec, vm);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncCharAt(ExecState* exec)
{
    VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    auto* thisString = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    auto viewWithString = thisString->viewWithUnderlyingString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    StringView view = viewWithString.view;
    JSValue a0 = exec->argument(0);
    if (LIKELY(a0.isUInt32())) {
        uint32_t i = a0.asUInt32();
        if (LIKELY(i < view.length()))
            return JSValue::encode(jsSingleCharacterString(vm, view[i]));
        return JSValue::encode(jsEmptyString(vm));
    }
    double dpos = a0.toIntegerOrInfinity(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (dpos >= 0 && dpos < view.length())
        return JSValue::encode(jsSingleCharacterString(vm, view[static_cast<unsigned>(dpos)]));
    return JSValue::encode(jsEmptyString(vm));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncCharCodeAt(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    auto* thisString = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    auto viewWithString = thisString->viewWithUnderlyingString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    StringView view = viewWithString.view;
    JSValue a0 = exec->argument(0);
    if (a0.isUInt32()) {
        uint32_t i = a0.asUInt32();
        if (i < view.length())
            return JSValue::encode(jsNumber(view[i]));
        return JSValue::encode(jsNaN());
    }
    double dpos = a0.toIntegerOrInfinity(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (dpos >= 0 && dpos < view.length())
        return JSValue::encode(jsNumber(view[static_cast<int>(dpos)]));
    return JSValue::encode(jsNaN());
}

static inline UChar32 codePointAt(const String& string, unsigned position, unsigned length)
{
    RELEASE_ASSERT(position < length);
    if (string.is8Bit())
        return string.characters8()[position];
    UChar32 character;
    U16_NEXT(string.characters16(), position, length, character);
    return character;
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncCodePointAt(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    String string = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    unsigned length = string.length();

    JSValue argument0 = exec->argument(0);
    if (argument0.isUInt32()) {
        unsigned position = argument0.asUInt32();
        if (position < length)
            return JSValue::encode(jsNumber(codePointAt(string, position, length)));
        return JSValue::encode(jsUndefined());
    }

    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    double doublePosition = argument0.toIntegerOrInfinity(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (doublePosition >= 0 && doublePosition < length)
        return JSValue::encode(jsNumber(codePointAt(string, static_cast<unsigned>(doublePosition), length)));
    return JSValue::encode(jsUndefined());
}

static EncodedJSValue JSC_HOST_CALL stringIndexOfImpl(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    JSValue a0 = exec->argument(0);
    JSValue a1 = exec->argument(1);

    JSString* thisJSString = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    JSString* otherJSString = a0.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    unsigned pos = 0;
    if (!a1.isUndefined()) {
        int len = thisJSString->length();
        RELEASE_ASSERT(len >= 0);
        if (a1.isUInt32())
            pos = std::min<uint32_t>(a1.asUInt32(), len);
        else {
            double dpos = a1.toIntegerOrInfinity(exec);
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            if (dpos < 0)
                dpos = 0;
            else if (dpos > len)
                dpos = len;
            pos = static_cast<unsigned>(dpos);
        }
    }

    if (thisJSString->length() < otherJSString->length() + pos)
        return JSValue::encode(jsNumber(-1));

    StringViewWithUnderlyingString thisViewWithString = thisJSString->viewWithUnderlyingString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    StringViewWithUnderlyingString otherViewWithString = otherJSString->viewWithUnderlyingString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    size_t result = thisViewWithString.view.find(otherViewWithString.view, pos);
    if (result == notFound)
        return JSValue::encode(jsNumber(-1));
    return JSValue::encode(jsNumber(result));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncIndexOf(CallFrame* callFrame)
{
    return stringIndexOfImpl(callFrame);
}

EncodedJSValue JSC_HOST_CALL builtinStringIndexOfInternal(CallFrame* callFrame)
{
    ASSERT(callFrame->thisValue().isString());
    ASSERT(callFrame->argument(0).isString());
    ASSERT(callFrame->argument(1).isNumber() || callFrame->argument(1).isUndefined());
    return stringIndexOfImpl(callFrame);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncLastIndexOf(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    JSValue a0 = exec->argument(0);
    JSValue a1 = exec->argument(1);

    JSString* thisJSString = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    unsigned len = thisJSString->length();
    JSString* otherJSString = a0.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    double dpos = a1.toIntegerPreserveNaN(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    unsigned startPosition;
    if (dpos < 0)
        startPosition = 0;
    else if (!(dpos <= len)) // true for NaN
        startPosition = len;
    else
        startPosition = static_cast<unsigned>(dpos);

    if (len < otherJSString->length())
        return JSValue::encode(jsNumber(-1));

    String thisString = thisJSString->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    String otherString = otherJSString->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    size_t result;
	if (!startPosition)
		result = thisString.startsWith(otherString) ? 0 : notFound;
	else
		result = thisString.reverseFind(otherString, startPosition);
    if (result == notFound)
        return JSValue::encode(jsNumber(-1));
    return JSValue::encode(jsNumber(result));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncSlice(ExecState* exec)
{
    VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    JSString* string = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue a0 = exec->argument(0);
    JSValue a1 = exec->argument(1);

    int length = string->length();
    RELEASE_ASSERT(length >= 0);

    // The arg processing is very much like ArrayProtoFunc::Slice
    double start = a0.toIntegerOrInfinity(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    double end = a1.isUndefined() ? length : a1.toIntegerOrInfinity(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(stringSlice(exec, vm, string, length, start, end));
}

// Return true in case of early return (resultLength got to limitLength).
template<typename CharacterType>
static ALWAYS_INLINE bool splitStringByOneCharacterImpl(ExecState* exec, JSArray* result, JSValue originalValue, const String& input, StringImpl* string, UChar separatorCharacter, size_t& position, unsigned& resultLength, unsigned limitLength)
{
	VM& vm = exec->vm();

    // 12. Let q = p.
    size_t matchPosition;
    const CharacterType* characters = string->getCharacters<CharacterType>();
    // 13. Repeat, while q != s
    //   a. Call SplitMatch(S, q, R) and let z be its MatchResult result.
    //   b. If z is failure, then let q = q+1.
    //   c. Else, z is not failure
    while ((matchPosition = WTF::find(characters, string->length(), separatorCharacter, position)) != notFound) {
        // 1. Let T be a String value equal to the substring of S consisting of the characters at positions p (inclusive)
        //    through q (exclusive).
        // 2. Call the [[DefineOwnProperty]] internal method of A with arguments ToString(lengthA),
        //    Property Descriptor {[[Value]]: T, [[Writable]]: true, [[Enumerable]]: true, [[Configurable]]: true}, and false.
        result->putDirectIndex(exec, resultLength, jsSubstring(vm, exec, originalValue, input, position, matchPosition - position));
        RETURN_IF_EXCEPTION(vm, false);
        // 3. Increment lengthA by 1.
        // 4. If lengthA == lim, return A.
        if (++resultLength == limitLength)
            return true;

        // 5. Let p = e.
        // 8. Let q = p.
        position = matchPosition + 1;
    }
    return false;
}

// ES 21.1.3.17 String.prototype.split(separator, limit)
EncodedJSValue JSC_HOST_CALL stringProtoFuncSplitFast(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    ASSERT(checkObjectCoercible(thisValue));

    // 3. Let S be the result of calling ToString, giving it the this value as its argument.
    // 7. Let s be the number of characters in S.
    String input = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    ASSERT(!input.isNull());

    // 4. Let A be a new array created as if by the expression new Array()
    //    where Array is the standard built-in constructor with that name.
    JSArray* result = constructEmptyArray(exec, 0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // 5. Let lengthA be 0.
    unsigned resultLength = 0;

    // 6. If limit is undefined, let lim = 2^32-1; else let lim = ToUint32(limit).
    JSValue limitValue = exec->uncheckedArgument(1);
    unsigned limit = limitValue.isUndefined() ? 0xFFFFFFFFu : limitValue.toUInt32(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // 8. Let p = 0.
    size_t position = 0;

    // 9. If separator is a RegExp object (its [[Class]] is "RegExp"), let R = separator;
    //    otherwise let R = ToString(separator).
    JSValue separatorValue = exec->uncheckedArgument(0);
    { // FIXME: Keeping this indentation here to minimize the diff. Will unindent and remove this later.
        String separator = separatorValue.toWTFString(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        // 10. If lim == 0, return A.
        if (!limit)
            return JSValue::encode(result);

        // 11. If separator is undefined, then
        if (separatorValue.isUndefined()) {
            // a.  Call the [[DefineOwnProperty]] internal method of A with arguments "0",
            result->putDirectIndex(exec, 0, jsStringWithReuse(vm, exec, thisValue, input));
            // b.  Return A.
            return JSValue::encode(result);
        }

        // 12. If s == 0, then
        if (input.isEmpty()) {
            // a. Let z be SplitMatch(S, 0, R) where S is input, R is separator.
            // b. If z is not false, return A.
            // c. Call CreateDataProperty(A, "0", S).
            // d. Return A.
            if (!separator.isEmpty())
                result->putDirectIndex(exec, 0, jsStringWithReuse(vm, exec, thisValue, input));
            return JSValue::encode(result);
        }

        // Optimized case for splitting on the empty string.
        if (separator.isEmpty()) {
            limit = std::min(limit, input.length());
            // Zero limt/input length handled in steps 9/11 respectively, above.
            ASSERT(limit);

            do {
                result->putDirectIndex(exec, position, jsSingleCharacterString(vm, input[position]));
                RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
            } while (++position < limit);

            return JSValue::encode(result);
        }

        // 3 cases:
        // -separator length == 1, 8 bits
        // -separator length == 1, 16 bits
        // -separator length > 1
        StringImpl* stringImpl = input.impl();
        StringImpl* separatorImpl = separator.impl();
        size_t separatorLength = separatorImpl->length();

        if (separatorLength == 1) {
            UChar separatorCharacter;
            if (separatorImpl->is8Bit())
                separatorCharacter = separatorImpl->characters8()[0];
            else
                separatorCharacter = separatorImpl->characters16()[0];

            if (stringImpl->is8Bit()) {
                if (splitStringByOneCharacterImpl<LChar>(exec, result, thisValue, input, stringImpl, separatorCharacter, position, resultLength, limit))
                    return JSValue::encode(result);
            } else {
                if (splitStringByOneCharacterImpl<UChar>(exec, result, thisValue, input, stringImpl, separatorCharacter, position, resultLength, limit))
                    return JSValue::encode(result);
            }
            RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        } else {
            // 13. Let q = p.
            size_t matchPosition;
            // 14. Repeat, while q != s
            //   a. let e be SplitMatch(S, q, R).
            //   b. If e is failure, then let q = q+1.
            //   c. Else, e is an integer index <= s.
            while ((matchPosition = stringImpl->find(separatorImpl, position)) != notFound) {
                // 1. Let T be a String value equal to the substring of S consisting of the characters at positions p (inclusive)
                //    through q (exclusive).
                // 2. Call CreateDataProperty(A, ToString(lengthA), T).
                result->putDirectIndex(exec, resultLength, jsSubstring(vm, exec, thisValue, input, position, matchPosition - position));
                RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
                // 3. Increment lengthA by 1.
                // 4. If lengthA == lim, return A.
                if (++resultLength == limit)
                    return JSValue::encode(result);

                // 5. Let p = e.
                // 6. Let q = p.
                position = matchPosition + separator.length();
            }
        }
    } // FIXME: Keeping this indentation here to minimize the diff. Will unindent and remove this later.

    // 15. Let T be a String value equal to the substring of S consisting of the characters at positions p (inclusive)
    //     through s (exclusive).
    // 16. Call CreateDataProperty(A, ToString(lengthA), T).
    result->putDirectIndex(exec, resultLength++, jsSubstring(vm, exec, thisValue, input, position, input.length() - position));

    // 17. Return A.
    return JSValue::encode(result);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncSubstr(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    unsigned len;
    JSString* jsString = 0;
    String uString;
    if (thisValue.isString()) {
        jsString = asString(thisValue);
        len = jsString->length();
    } else {
        uString = thisValue.toWTFString(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        len = uString.length();
    }

    JSValue a0 = exec->argument(0);
    JSValue a1 = exec->argument(1);

    double start = a0.toIntegerOrInfinity(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    double length = a1.isUndefined() ? len : a1.toIntegerOrInfinity(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (start >= len || length <= 0)
        return JSValue::encode(jsEmptyString(vm));
    if (start < 0) {
        start += len;
        if (start < 0)
            start = 0;
    }
    if (start + length > len)
        length = len - start;
    unsigned substringStart = static_cast<unsigned>(start);
    unsigned substringLength = static_cast<unsigned>(length);
    if (jsString)
        return JSValue::encode(jsSubstring(vm, exec, jsString, substringStart, substringLength));
    return JSValue::encode(jsSubstring(vm, uString, substringStart, substringLength));
}

static EncodedJSValue stringSubstringImpl(CallFrame* exec)
{
    VM& vm = exec->vm();

    SamplingRegion samplingRegion("Doing substringing");
    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    JSString* jsString = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue a0 = exec->argument(0);
    JSValue a1 = exec->argument(1);
    int len = jsString->length();
    RELEASE_ASSERT(len >= 0);

    double start = a0.toNumber(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    double end;
    if (!(start >= 0)) // check for negative values or NaN
        start = 0;
    else if (start > len)
        start = len;
    if (a1.isUndefined())
        end = len;
    else { 
        end = a1.toNumber(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
        if (!(end >= 0)) // check for negative values or NaN
            end = 0;
        else if (end > len)
            end = len;
    }
    if (start > end) {
        double temp = end;
        end = start;
        start = temp;
    }
    unsigned substringStart = static_cast<unsigned>(start);
    unsigned substringLength = static_cast<unsigned>(end) - substringStart;
    return JSValue::encode(jsSubstring(vm, exec, jsString, substringStart, substringLength));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncSubstring(CallFrame* callFrame)
{
    return stringSubstringImpl(callFrame);
}

EncodedJSValue JSC_HOST_CALL builtinStringSubstringInternal(CallFrame* callFrame)
{
    ASSERT(callFrame->thisValue().isString());
    ASSERT(callFrame->argument(0).isNumber());
    ASSERT(callFrame->argument(1).isNumber() || callFrame->argument(1).isUndefined());
    return stringSubstringImpl(callFrame);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncToLowerCase(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    JSString* sVal = thisValue.toString(exec);
    const String& s = sVal->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    String lowercasedString = s.convertToLowercaseWithoutLocale();
    if (lowercasedString.impl() == s.impl())
        return JSValue::encode(sVal);
    return JSValue::encode(jsString(vm, lowercasedString));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncToUpperCase(ExecState* exec)
{
	VM& vm = exec->vm();
    JSValue thisValue = exec->hostThisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    JSString* sVal = thisValue.toString(exec);
    const String& s = sVal->value(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    String uppercasedString = s.convertToUppercaseWithoutLocale();
    if (uppercasedString.impl() == s.impl())
        return JSValue::encode(sVal);
    return JSValue::encode(jsString(vm, uppercasedString));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncLocaleCompare(ExecState* exec)
{
    // 13.1.1 String.prototype.localeCompare (that [, locales [, options ]]) (ECMA-402 2.0)
    // http://ecma-international.org/publications/standards/Ecma-402.htm

    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    // 1. Let O be RequireObjectCoercible(this value).
    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm, "String.prototype.localeCompare requires that |this| not be null or undefined");

    // 2. Let S be ToString(O).
    // 3. ReturnIfAbrupt(S).
    String string = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

    // 4. Let That be ToString(that).
    // 5. ReturnIfAbrupt(That).
    JSValue thatValue = exec->argument(0);
    String that = thatValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

#if ENABLE(INTL)
    JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    JSValue locales = exec->argument(1);
    JSValue options = exec->argument(2);
    IntlCollator* collator = nullptr;
    if (locales.isUndefined() && options.isUndefined()) {
        collator = globalObject->defaultCollator(exec);
        RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    } else {
        collator = IntlCollator::create(vm, globalObject->intlObject()->collatorStructure());
        collator->initializeCollator(*exec, locales, options);
        RETURN_IF_EXCEPTION(scope, encodedJSUndefined());
    }
    RELEASE_AND_RETURN(scope, JSValue::encode(collator->compareStrings(*exec, string, that)));
#else
    return JSValue::encode(jsNumber(Collator().collate(string, that)));
#endif
}


template<typename StringType>
static constexpr uint32_t computeTwoCharacters16Code(const StringType& string)
{
    return static_cast<uint16_t>(string.characterAt(0)) | (static_cast<uint32_t>(static_cast<uint16_t>(string.characterAt(1))) << 16);
}

#if ENABLE(INTL)
static EncodedJSValue toLocaleCase(ExecState* state, int32_t (*convertCase)(UChar*, int32_t, const UChar*, int32_t, const char*, UErrorCode*))
{
    VM& vm = state->vm();

    // 1. Let O be RequireObjectCoercible(this value).
    JSValue thisValue = state->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(state, vm);

    // 2. Let S be ToString(O).
    JSString* sVal = thisValue.toString(state);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    const String& s = sVal->value(state);

    // 3. ReturnIfAbrupt(S).
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue localeValue = state->argument(0);

    // Optimization for empty strings.
    if (s.isEmpty() && localeValue.isUndefined())
        return JSValue::encode(sVal);

    // 4. Let requestedLocales be CanonicalizeLocaleList(locales).
    Vector<String> requestedLocales = canonicalizeLocaleList(state, localeValue);

    // 5. ReturnIfAbrupt(requestedLocales).
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    // 6. Let len be the number of elements in requestedLocales.
    size_t len = requestedLocales.size();

    // 7. If len > 0, then
    // a. Let requestedLocale be the first element of requestedLocales.
    // 8. Else
    // a. Let requestedLocale be DefaultLocale().
    String requestedLocale = len > 0 ? requestedLocales.first() : defaultLocale(*state);

    // 9. Let noExtensionsLocale be the String value that is requestedLocale with all Unicode locale extension sequences (6.2.1) removed.
    String noExtensionsLocale = removeUnicodeLocaleExtension(requestedLocale);

    // 10. Let availableLocales be a List with the language tags of the languages for which the Unicode character database contains language sensitive case mappings.
	// Note 1: As of Unicode 5.1, the availableLocales list contains the elements "az", "el", "lt", and "tr".
	// 11. Let locale be BestAvailableLocale(availableLocales, noExtensionsLocale).
    String locale = bestAvailableLocale(noExtensionsLocale, [](const String& candidate) {
        if (candidate.length() != 2)
            return false;
        switch (computeTwoCharacters16Code(candidate)) {
        case static_cast<uint16_t>('a') | (static_cast<uint32_t>(static_cast<uint16_t>('z')) << 16):
        case static_cast<uint16_t>('e') | (static_cast<uint32_t>(static_cast<uint16_t>('l')) << 16):
        case static_cast<uint16_t>('l') | (static_cast<uint32_t>(static_cast<uint16_t>('t')) << 16):
        case static_cast<uint16_t>('t') | (static_cast<uint32_t>(static_cast<uint16_t>('r')) << 16):
            return true;
        default:
            return false;
        }
    });

    // 12. If locale is undefined, let locale be "und".
    if (locale.isNull())
        locale = ASCIILiteral("und");

    CString utf8LocaleBuffer = locale.utf8();
    const StringView view(s);
    const int32_t viewLength = view.length();

    // Delegate the following steps to icu u_strToLower or u_strToUpper.
    // 13. Let cpList be a List containing in order the code points of S as defined in ES2015, 6.1.4, starting at the first element of S.
    // 14. For each code point c in cpList, if the Unicode Character Database provides a lower(/upper) case equivalent of c that is either language insensitive or for the language locale, then replace c in cpList with that/those equivalent code point(s).
    // 15. Let cuList be a new List.
    // 16. For each code point c in cpList, in order, append to cuList the elements of the UTF-16 Encoding (defined in ES2015, 6.1.4) of c.
    // 17. Let L be a String whose elements are, in order, the elements of cuList.

    // Most strings lower/upper case will be the same size as original, so try that first.
    UErrorCode error = U_ZERO_ERROR;
    Vector<UChar> buffer(viewLength);
    String lower;
    const int32_t resultLength = convertCase(buffer.data(), viewLength, view.upconvertedCharacters().get(), viewLength, utf8LocaleBuffer.data(), &error);
    if (U_SUCCESS(error))
        lower = String(buffer.data(), resultLength);
    else if (needsToGrowToProduceBuffer(error)) {
        // Converted case needs more space than original. Try again.
        UErrorCode error = U_ZERO_ERROR;
        Vector<UChar> buffer(resultLength);
        convertCase(buffer.data(), resultLength, view.upconvertedCharacters().get(), viewLength, utf8LocaleBuffer.data(), &error);
        if (U_FAILURE(error))
            return throwVMTypeError(state, vm, String::fromLatin1(u_errorName(error)));
        lower = String(buffer.data(), resultLength);
    } else
        return throwVMTypeError(state, vm, String::fromLatin1(u_errorName(error)));

    // 18. Return L.
    return JSValue::encode(jsString(vm, lower));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncToLocaleLowerCase(ExecState* state)
{
    // 13.1.2 String.prototype.toLocaleLowerCase ([locales])
    // http://ecma-international.org/publications/standards/Ecma-402.htm
    return toLocaleCase(state, u_strToLower);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncToLocaleUpperCase(ExecState* state)
{
    // 13.1.3 String.prototype.toLocaleUpperCase ([locales])
    // http://ecma-international.org/publications/standards/Ecma-402.htm
    // This function interprets a string value as a sequence of code points, as described in ES2015, 6.1.4. This function behaves in exactly the same way as String.prototype.toLocaleLowerCase, except that characters are mapped to their uppercase equivalents as specified in the Unicode character database.
    return toLocaleCase(state, u_strToUpper);
}
#endif // ENABLE(INTL)

enum {
    TrimStart = 1,
    TrimEnd = 2
};

static inline JSValue trimString(ExecState* exec, JSValue thisValue, int trimKind)
{
    VM& vm = exec->vm();

    if (!checkObjectCoercible(thisValue))
        return throwTypeError(exec, vm);
    String str = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    unsigned left = 0;
    if (trimKind & TrimStart) {
        while (left < str.length() && isStrWhiteSpace(str[left]))
            left++;
    }
    unsigned right = str.length();
    if (trimKind & TrimEnd) {
        while (right > left && isStrWhiteSpace(str[right - 1]))
            right--;
    }

    // Don't gc allocate a new string if we don't have to.
    if (left == 0 && right == str.length() && thisValue.isString())
        return thisValue;

    return jsString(vm, str.substringSharingImpl(left, right - left));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncTrim(ExecState* exec)
{
    JSValue thisValue = exec->hostThisValue();
    return JSValue::encode(trimString(exec, thisValue, TrimStart | TrimEnd));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncTrimStart(ExecState* exec)
{
    JSValue thisValue = exec->hostThisValue();
    return JSValue::encode(trimString(exec, thisValue, TrimStart));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncTrimEnd(ExecState* exec)
{
    JSValue thisValue = exec->hostThisValue();
    return JSValue::encode(trimString(exec, thisValue, TrimEnd));
}

static inline unsigned clampAndTruncateToUnsigned(double value, unsigned min, unsigned max)
{
    if (value < min)
        return min;
    if (value > max)
        return max;
    return static_cast<unsigned>(value);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncStartsWith(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    String stringToSearchIn = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue a0 = exec->argument(0);
    bool isRegularExpression = isRegExp(vm, exec, a0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (isRegularExpression)
        return throwVMTypeError(exec, vm, ASCIILiteral("Argument to String.prototype.startsWith can not be a RegExp."));

    String searchString = a0.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue positionArg = exec->argument(1);
    unsigned length = stringToSearchIn.length();
    unsigned start;
    if (positionArg.isInt32())
        start = std::min(clampTo<unsigned>(positionArg.asInt32()), length);
    else {
        start = clampAndTruncateToUnsigned(positionArg.toIntegerOrInfinity(exec), 0, length);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    return JSValue::encode(jsBoolean(stringToSearchIn.hasInfixStartingAt(searchString, start)));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncEndsWith(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    String stringToSearchIn = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue a0 = exec->argument(0);
    bool isRegularExpression = isRegExp(vm, exec, a0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (isRegularExpression)
        return throwVMTypeError(exec, vm, ASCIILiteral("Argument to String.prototype.endsWith can not be a RegExp."));

    String searchString = a0.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue endPositionArg = exec->argument(1);
    unsigned length = stringToSearchIn.length();
    unsigned end;
    if (endPositionArg.isUndefined())
        end = length;
    else if (endPositionArg.isInt32())
        end = std::min(clampTo<unsigned>(endPositionArg.asInt32()), length);
    else {
        end = clampAndTruncateToUnsigned(endPositionArg.toIntegerOrInfinity(exec), 0, length);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    return JSValue::encode(jsBoolean(stringToSearchIn.hasInfixEndingAt(searchString, end)));
}

static EncodedJSValue JSC_HOST_CALL stringIncludesImpl(VM& vm, ExecState* exec, String stringToSearchIn, String searchString, JSValue positionArg)
{
    auto length = stringToSearchIn.length();
    unsigned start;
    if (positionArg.isInt32())
        start = std::min(clampTo<unsigned>(positionArg.asInt32()), length);
    else {
        start = clampAndTruncateToUnsigned(positionArg.toIntegerOrInfinity(exec), 0, length);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    }

    return JSValue::encode(jsBoolean(stringToSearchIn.contains(searchString, true, start)));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncIncludes(ExecState* exec)
{
	VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    String stringToSearchIn = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue a0 = exec->argument(0);
    bool isRegularExpression = isRegExp(vm, exec, a0);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    if (isRegularExpression)
        return throwVMTypeError(exec, vm, ASCIILiteral("Argument to String.prototype.includes can not be a RegExp."));

    String searchString = a0.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue positionArg = exec->argument(1);

    return stringIncludesImpl(vm, exec, stringToSearchIn, searchString, positionArg);
}

EncodedJSValue JSC_HOST_CALL builtinStringIncludesInternal(ExecState* exec)
{
    VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    ASSERT(checkObjectCoercible(thisValue));

    String stringToSearchIn = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue a0 = exec->uncheckedArgument(0);
    String searchString = a0.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    JSValue positionArg = exec->argument(1);

    return stringIncludesImpl(vm, exec, stringToSearchIn, searchString, positionArg);
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncIterator(ExecState* exec)
{
    VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    JSString* string = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());
    return JSValue::encode(JSStringIterator::create(exec, exec->callee()->globalObject()->stringIteratorStructure(), string));
}

enum class NormalizationForm { NFC, NFD, NFKC, NFKD };

static const UNormalizer2* normalizer(NormalizationForm form)
{
    UErrorCode status = U_ZERO_ERROR;
    const UNormalizer2* normalizer = 0;
    switch (form) {
    case NormalizationForm::NFC:
        normalizer = unorm2_getNFCInstance(&status);
        break;
    case NormalizationForm::NFD:
        normalizer = unorm2_getNFDInstance(&status);
        break;
    case NormalizationForm::NFKC:
        normalizer = unorm2_getNFKCInstance(&status);
        break;
    case NormalizationForm::NFKD:
        normalizer = unorm2_getNFKDInstance(&status);
        break;
    }
    ASSERT(normalizer);
    ASSERT(U_SUCCESS(status));
    return normalizer;
}

static JSValue normalize(ExecState* exec, JSString* string, NormalizationForm form)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    auto viewWithString = string->viewWithUnderlyingString(exec);
    RETURN_IF_EXCEPTION(vm, jsUndefined());

    StringView view = viewWithString.view;
    // Latin-1 characters (U+0000..U+00FF) are left unaffected by NFC.
    // ASCII characters (U+0000..U+007F) are left unaffected by all of the Normalization Forms
    // https://unicode.org/reports/tr15/#Description_Norm
    if (view.is8Bit() && (form == NormalizationForm::NFC || charactersAreAllASCII(view.characters8(), view.length())))
        return string;

    const UNormalizer2* normalizer = JSC::normalizer(form);

    // Since ICU does not offer functions that can perform normalization or check for
    // normalization with input that is Latin-1, we need to upconvert to UTF-16 at this point.
    auto characters = view.upconvertedCharacters();

    UErrorCode status = U_ZERO_ERROR;
    UBool isNormalized = unorm2_isNormalized(normalizer, characters, view.length(), &status);
    ASSERT(U_SUCCESS(status));
    if (isNormalized)
        return string;

    int32_t normalizedStringLength = unorm2_normalize(normalizer, characters, view.length(), 0, 0, &status);
    ASSERT(needsToGrowToProduceBuffer(status));

    UChar* buffer;
    RefPtr<StringImpl> result = StringImpl::tryCreateUninitialized(normalizedStringLength, buffer);
    if (!result)
        return throwOutOfMemoryError(vm, exec);

    status = U_ZERO_ERROR;
    unorm2_normalize(normalizer, characters, view.length(), buffer, normalizedStringLength, &status);
    if (!U_SUCCESS(status))
        return throwOutOfMemoryError(vm, exec);

    return jsString(vm, result.releaseNonNull());
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncNormalize(ExecState* exec)
{
    VM& vm = exec->vm();

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);
    JSString* string = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    auto form = NormalizationForm::NFC;
    JSValue formValue = exec->argument(0);
    if (!formValue.isUndefined()) {
        String formString = formValue.toWTFString(exec);
        RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

        if (formString == "NFC")
            form = NormalizationForm::NFC;
        else if (formString == "NFD")
            form = NormalizationForm::NFD;
        else if (formString == "NFKC")
            form = NormalizationForm::NFKC;
        else if (formString == "NFKD")
            form = NormalizationForm::NFKD;
        else
            return throwVMRangeError(exec, vm, ASCIILiteral("argument does not match any normalization form"));
    }

    return JSValue::encode(normalize(exec, string, form));
}

static inline Optional<unsigned> illFormedIndex(const UChar* characters, unsigned length)
{
    for (unsigned index = 0; index < length; ++index) {
        UChar character = characters[index];
        if (!U16_IS_SURROGATE(character))
            continue;

        if (U16_IS_SURROGATE_TRAIL(character))
            return index;

        ASSERT(U16_IS_SURROGATE_LEAD(character));
        if ((index + 1) == length)
            return index;
        UChar nextCharacter = characters[index + 1];

        if (!U16_IS_SURROGATE(nextCharacter))
            return index;

        if (!U16_IS_SURROGATE_TRAIL(nextCharacter))
            return index;

        ++index; // Increment additionally.
    }
    return Nullopt;
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncIsWellFormed(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    // Latin-1 characters do not have surrogates.
    if (thisValue.isString() && asString(thisValue)->is8Bit())
        return JSValue::encode(jsBoolean(true));

    String string = thisValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

    if (string.is8Bit())
        return JSValue::encode(jsBoolean(true));
    return JSValue::encode(jsBoolean(!illFormedIndex(string.characters16(), string.length())));
}

static EncodedJSValue JSC_HOST_CALL stringProtoFuncToWellFormed(ExecState* exec)
{
    VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue thisValue = exec->thisValue();
    if (!checkObjectCoercible(thisValue))
        return throwVMTypeError(exec, vm);

    // Latin-1 characters do not have surrogates.
    if (thisValue.isString() && asString(thisValue)->is8Bit())
        return JSValue::encode(thisValue);

    JSString* stringValue = thisValue.toString(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

    if (stringValue->is8Bit())
        return JSValue::encode(stringValue);

    String string = stringValue->value(exec);
    RETURN_IF_EXCEPTION(scope, encodedJSUndefined());

    if (string.is8Bit())
        return JSValue::encode(stringValue);

    const UChar* characters = string.characters16();
    unsigned length = string.length();
    auto firstIllFormedIndex = illFormedIndex(characters, length);
    if (!firstIllFormedIndex)
        return JSValue::encode(stringValue);

    Vector<UChar> buffer;
    buffer.reserveInitialCapacity(length);
    buffer.append(characters, firstIllFormedIndex.value());
    for (unsigned index = firstIllFormedIndex.value(); index < length; ++index) {
        UChar character = characters[index];

        if (!U16_IS_SURROGATE(character)) {
            buffer.append(character);
            continue;
        }

        if (U16_IS_SURROGATE_TRAIL(character)) {
            buffer.append(replacementCharacter);
            continue;
        }

        ASSERT(U16_IS_SURROGATE_LEAD(character));
        if ((index + 1) == length) {
            buffer.append(replacementCharacter);
            continue;
        }
        UChar nextCharacter = characters[index + 1];

        if (!U16_IS_SURROGATE(nextCharacter)) {
            buffer.append(replacementCharacter);
            continue;
        }

        if (!U16_IS_SURROGATE_TRAIL(nextCharacter)) {
            buffer.append(replacementCharacter);
            continue;
        }

        buffer.append(character);
        buffer.append(nextCharacter);
        index += 1;
    }

    return JSValue::encode(jsString(vm, String::adopt(buffer)));
}


} // namespace JSC

namespace WTF {

template <> struct VectorTraits<JSC::StringRange> : SimpleClassVectorTraits { };

} // namespace WTF
