/*
 * Copyright (C) 2008, 2009, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "ClassInfo.h"
#include "DeferGC.h"
#include "IndexingType.h"
#include "JSCell.h"
#include "JSType.h"
#include "JSTypeInfo.h"
#include "JSValue.h"
#include "PropertyName.h"
#include "PropertyNameArray.h"
#include "PropertyOffset.h"
#include "Protect.h"
#include "StructureRareData.h"
#include "PutPropertySlot.h"
#include "StructureTransitionTable.h"
#include "TinyBloomFilter.h"
#include "Watchpoint.h"
#include "Weak.h"
#include <wtf/Atomics.h>
#include <wtf/CompilationThread.h>
#include <wtf/ForbidHeapAllocation.h>
#include <wtf/Optional.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>

#ifndef NDEBUG
#define DO_PROPERTYMAP_CONSTENCY_CHECK 0
#else
#define DO_PROPERTYMAP_CONSTENCY_CHECK 0
#endif

namespace WTF {

class UniquedStringImpl;

} // namespace WTF

namespace JSC {

class LLIntOffsetsExtractor;
class PropertyNameArray;
class PropertyNameArrayData;
class PropertyTable;
class StructureChain;
class SlotVisitor;
class JSString;
struct DumpContext;
struct HashTable;
struct HashTableValue;

// The out-of-line property storage capacity to use when first allocating out-of-line
// storage. Note that all objects start out without having any out-of-line storage;
// this comes into play only on the first property store that exhausts inline storage.
static const unsigned initialOutOfLineCapacity = 4;

// The factor by which to grow out-of-line storage when it is exhausted, after the
// initial allocation.
static const unsigned outOfLineGrowthFactor = 2;

struct PropertyMapEntry {
	UniquedStringImpl* key;
    PropertyOffset offset;
    uint8_t attributes;

    PropertyMapEntry()
        : key(0)
        , offset(invalidOffset)
        , attributes(0)
    {
    }

    PropertyMapEntry(UniquedStringImpl* key, PropertyOffset offset, unsigned attributes)
        : key(key)
        , offset(offset)
        , attributes(attributes)
    {
        ASSERT(this->attributes == attributes);
    }
};

class StructureFireDetail final : public FireDetail {
public:
    StructureFireDetail(const Structure* structure)
        : m_structure(structure)
    {
    }

    virtual void dump(PrintStream& out) const override;

private:
    const Structure* m_structure;
};

class DeferredStructureTransitionWatchpointFire final : public DeferredWatchpointFire {
	WTF_FORBID_HEAP_ALLOCATION;
    WTF_MAKE_NONCOPYABLE(DeferredStructureTransitionWatchpointFire);
public:
    DeferredStructureTransitionWatchpointFire(VM& vm, Structure* structure)
        : DeferredWatchpointFire()
        , m_vm(vm)
        , m_structure(structure)
    {
    }

    ~DeferredStructureTransitionWatchpointFire()
    {
        if (watchpointsToFire().state() == IsWatched)
            fireAllSlow();
    }

    const Structure* structure() const { return m_structure; }

private:
    JS_EXPORT_PRIVATE void fireAllSlow();

    VM& m_vm;
    const Structure* m_structure;
};

class Structure : public JSCell {
public:
	friend class StructureTransitionTable;

	typedef JSCell Base;
    static const unsigned StructureFlags = Base::StructureFlags | StructureIsImmortal;

	static Structure* create(JSGlobalData&, JSGlobalObject*, JSValue prototype, const TypeInfo&, const ClassInfo*, IndexingType = NonArray, unsigned inlineCapacity = 0);

    ~Structure();

    JS_EXPORT_PRIVATE static bool isValidPrototype(JSValue);

protected:
    void finishCreation(JSGlobalData& vm, const Structure* previous, DeferredStructureTransitionWatchpointFire* deferred)
    {
        this->finishCreation(vm);
        if (previous->hasRareData()) {
            const StructureRareData* previousRareData = previous->rareData();
#if 0
            if (previousRareData->hasSharedPolyProtoWatchpoint()) {
                ensureRareData(vm);
                rareData()->setSharedPolyProtoWatchpoint(previousRareData->copySharedPolyProtoWatchpoint());
            }
#endif
        }
        previous->fireStructureTransitionWatchpoint(deferred);
    }

private:
	void finishCreation(JSGlobalData& globalData)
	{
		Base::finishCreation(globalData);
        ASSERT(m_prototype.get().isEmpty() || isValidPrototype(m_prototype.get()));
	}

	void finishCreation(JSGlobalData& globalData, CreatingEarlyCellTag)
	{
		Base::finishCreation(globalData, this, CreatingEarlyCell);
		ASSERT(m_prototype);
		ASSERT(m_prototype.isNull());
		ASSERT(!globalData.structureStructure);
	}

    void validateFlags();

public:
    bool isProxy() const
    {
        JSType type = typeInfo().type();
        return type == ImpureProxyType || type == PureForwardingProxyType || type == ProxyObjectType;
    }

	static void dumpStatistics();

	JS_EXPORT_PRIVATE static Structure* addPropertyTransition(JSGlobalData&, Structure*, PropertyName, unsigned attributes, PropertyOffset&, PutPropertySlot::Context = PutPropertySlot::UnknownContext, DeferredStructureTransitionWatchpointFire* = 0);
	JS_EXPORT_PRIVATE static Structure* addNewPropertyTransition(JSGlobalData&, Structure*, PropertyName, unsigned attributes, PropertyOffset&, PutPropertySlot::Context, DeferredStructureTransitionWatchpointFire*);
	JS_EXPORT_PRIVATE static Structure* addPropertyTransitionToExistingStructure(Structure*, UniquedStringImpl* uid, unsigned attributes, PropertyOffset&);
	static Structure* removePropertyTransition(JSGlobalData&, Structure*, PropertyName, PropertyOffset&, DeferredStructureTransitionWatchpointFire* = 0);
	JS_EXPORT_PRIVATE static Structure* changePrototypeTransition(JSGlobalData&, Structure*, JSValue prototype, DeferredStructureTransitionWatchpointFire&);
	JS_EXPORT_PRIVATE static Structure* attributeChangeTransition(JSGlobalData&, Structure*, PropertyName, unsigned attributes, DeferredStructureTransitionWatchpointFire*);
	static Structure* toCacheableDictionaryTransition(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire*);
	static Structure* toUncacheableDictionaryTransition(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire*);
	static Structure* sealTransition(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire* = nullptr);
	static Structure* freezeTransition(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire* = nullptr);
	static Structure* preventExtensionsTransition(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire* = nullptr);
	static Structure* nonPropertyTransition(JSGlobalData&, Structure*, TransitionKind, DeferredStructureTransitionWatchpointFire*);
    static Structure* setBrandTransitionFromExistingStructureConcurrently(Structure*, UniquedStringImpl*);
	static Structure* setBrandTransition(VM&, Structure*, Symbol* brand, DeferredStructureTransitionWatchpointFire* = nullptr);

	bool isSealed(JSGlobalData&);
	bool isFrozen(JSGlobalData&);
	ALWAYS_INLINE bool isStructureExtensible() const { return !didPreventExtensions(); }
	bool putWillGrowOutOfLineStorage();
	JS_EXPORT_PRIVATE size_t suggestedNewOutOfLineStorageCapacity();

	Structure* flattenDictionaryStructure(JSGlobalData&, JSObject*);

	static const bool needsDestruction = true;
	static void destroy(JSCell*);

	// These should be used with caution.
	JS_EXPORT_PRIVATE PropertyOffset addPropertyWithoutTransition(JSGlobalData&, PropertyName, unsigned attributes);
	PropertyOffset removePropertyWithoutTransition(JSGlobalData&, PropertyName);
	void setPrototypeWithoutTransition(JSGlobalData& globalData, JSValue prototype)
	{
	    ASSERT(isValidPrototype(prototype));
		m_prototype.set(globalData, this, prototype);
	}

	ALWAYS_INLINE bool isDictionary() const { return dictionaryKind() != NoneDictionaryKind; }
	ALWAYS_INLINE bool isUncacheableDictionary() const { return dictionaryKind() == UncachedDictionaryKind; }

    bool prototypeQueriesAreCacheable() const
    {
        return !typeInfo().prohibitsPropertyCaching();
    }

	ALWAYS_INLINE bool propertyAccessesAreCacheable() const
	{
		return dictionaryKind() != UncachedDictionaryKind
	            && prototypeQueriesAreCacheable()
				&& !(typeInfo().getOwnPropertySlotIsImpure() && !typeInfo().newImpurePropertyFiresWatchpoints());
    }

    bool propertyAccessesAreCacheableForAbsence() const
    {
        return !typeInfo().getOwnPropertySlotIsImpureForPropertyAbsence();
    }

    bool needImpurePropertyWatchpoint() const
    {
        return propertyAccessesAreCacheable()
            && typeInfo().getOwnPropertySlotIsImpure()
            && typeInfo().newImpurePropertyFiresWatchpoints();
    }

    bool isImmutablePrototypeExoticObject() const
    {
        return typeInfo().isImmutablePrototypeExoticObject();
    }

    // We use SlowPath in GetByIdStatus for structures that may get new impure properties later to prevent
    // DFG from inlining property accesses since structures don't transition when a new impure property appears.
    bool takesSlowPathInDFGForImpureProperty() const
    {
        return typeInfo().getOwnPropertySlotIsImpure();
    }

    bool hasNonReifiedStaticProperties() const
    {
        return typeInfo().hasStaticPropertyTable() && !staticPropertiesReified();
    }

	// Type accessors.
	const TypeInfo& typeInfo() const { ASSERT(structure()->classInfo() == info()); return m_typeInfo; }
	bool isObject() const { return typeInfo().isObject(); } // can not use JSCell m_type

	ALWAYS_INLINE IndexingType indexingType() const { return m_indexingType & AllArrayTypes; }
	IndexingType indexingTypeIncludingHistory() const { return m_indexingType; }

    inline bool mayInterceptIndexedAccesses() const;

    bool holesMustForwardToPrototype(VM&) const;

	ALWAYS_INLINE JSGlobalObject* globalObject() const { return m_globalObject.get(); }

    // NOTE: This method should only be called during the creation of structures, since the global
    // object of a structure is presumed to be immutable in a bunch of places.
	void setGlobalObject(JSGlobalData& globalData, JSGlobalObject* globalObject) { m_globalObject.set(globalData, this, globalObject); }

    bool hasMonoProto() const
    {
        return !m_prototype.get().isInt32();
    }
    bool hasPolyProto() const
    {
        return !hasMonoProto();
    }

	ALWAYS_INLINE JSValue storedPrototype() const { return m_prototype.get(); }
    JSObject* storedPrototypeObject() const;
    Structure* storedPrototypeStructure() const;
	JSValue prototypeForLookup(ExecState*) const;
	JSValue prototypeForLookup(JSGlobalObject*) const;
	JSValue prototypeForLookup(CodeBlock*) const;
    StructureChain* prototypeChain(JSGlobalData&, JSGlobalObject*) const;
	StructureChain* prototypeChain(ExecState*) const;
	static void visitChildren(JSCell*, SlotVisitor&);

    // A Structure is cheap to mark during GC if doing so would only add a small and bounded amount
    // to our heap footprint. For example, if the structure refers to a global object that is not
    // yet marked, then as far as we know, the decision to mark this Structure would lead to a large
    // increase in footprint because no other object refers to that global object. This method
    // returns true if all user-controlled (and hence unbounded in size) objects referenced from the
    // Structure are already marked.
    bool isCheapDuringGC();

    // Returns true if this structure is now marked.
    bool markIfCheap(SlotVisitor&);

    bool hasRareData() const
    {
        return isRareData(m_previousOrRareData.get());
    }

	Structure* previousID() const
	{
		ASSERT(structure()->classInfo() == info());
        // This is so written because it's used concurrently. We only load from m_previousOrRareData
        // once, and this load is guaranteed atomic.
        JSCell* cell = m_previousOrRareData.get();
        if (isRareData(cell))
            return static_cast<StructureRareData*>(cell)->previousID();
        return static_cast<Structure*>(cell);
    }
	bool transitivelyTransitionedFrom(Structure* structureToFind);

    unsigned outOfLineCapacity() const
    {
        ASSERT(checkOffsetConsistency());

        unsigned outOfLineSize = this->outOfLineSize();

        if (!outOfLineSize)
            return 0;

        if (outOfLineSize <= initialOutOfLineCapacity)
            return initialOutOfLineCapacity;

        ASSERT(outOfLineSize > initialOutOfLineCapacity);
        COMPILE_ASSERT(outOfLineGrowthFactor == 2, outOfLineGrowthFactor_is_two);
        return WTF::roundUpToPowerOfTwo(outOfLineSize);
    }
	unsigned outOfLineSize() const
	{
        ASSERT(checkOffsetConsistency());
        ASSERT(structure()->classInfo() == info());

		return numberOfOutOfLineSlotsForLastOffset(m_offset);
	}
	bool hasInlineStorage() const
	{
		return !!m_inlineCapacity;
	}
	unsigned inlineCapacity() const
	{
		return m_inlineCapacity;
	}
	unsigned inlineSize() const
	{
        return std::min<unsigned>(m_offset + 1, m_inlineCapacity);
	}
	unsigned totalStorageSize() const
	{
		return numberOfSlotsForLastOffset(m_offset, m_inlineCapacity);
	}
	unsigned totalStorageCapacity() const
	{
		ASSERT(structure()->classInfo() == info());
		return outOfLineCapacity() + inlineCapacity();
	}

	bool isValidOffset(PropertyOffset offset) const
	{
        return JSC::isValidOffset(offset)
            && offset <= m_offset
            && (offset < m_inlineCapacity || offset >= firstOutOfLineOffset);
    }

    bool hijacksIndexingHeader() const
    {
        return isTypedView(m_classInfo->typedArrayStorageType);
    }

    bool couldHaveIndexingHeader() const
    {
        return hasIndexedProperties(indexingType())
            || hijacksIndexingHeader();
    }

    bool hasIndexingHeader(const JSCell*) const;

	bool masqueradesAsUndefined(JSGlobalObject* lexicalGlobalObject);

	PropertyOffset get(JSGlobalData&, const PropertyName&);
	PropertyOffset get(JSGlobalData&, UniquedStringImpl* uid);
	JS_EXPORT_PRIVATE PropertyOffset get(JSGlobalData&, UniquedStringImpl*, unsigned& attributes);
	JS_EXPORT_PRIVATE PropertyOffset get(JSGlobalData&, const PropertyName&, unsigned& attributes);

    bool canPerformFastPropertyEnumeration() const;

    template<typename Functor>
    void forEachProperty(VM&, const Functor&);

    // This is a somewhat internalish method. It will call your functor while possibly holding the
    // Structure's lock. There is no guarantee whether the lock is held or not in any particular
    // call. So, you have to assume the worst. Also, the functor returns true if it wishes for you
    // to continue or false if it's done.
    template<typename Functor>
    void forEachPropertyConcurrently(const Functor&);

    PropertyOffset getConcurrently(VM&, UniquedStringImpl* uid);
    PropertyOffset getConcurrently(VM&, UniquedStringImpl* uid, unsigned& attributes);

    Vector<PropertyMapEntry> getPropertiesConcurrently();

    void setHasAnyKindOfGetterSetterPropertiesWithProtoCheck(bool is__proto__)
    {
    	setHasAnyKindOfGetterSetterProperties(true);
        if (!is__proto__)
            setHasReadOnlyOrGetterSetterPropertiesExcludingProto(true);
    }

    void setContainsReadOnlyProperties() { setHasReadOnlyOrGetterSetterPropertiesExcludingProto(true); }

	bool isEmpty() const
	{
        ASSERT(checkOffsetConsistency());
		return !JSC::isValidOffset(m_offset);
	}

    void setCachedPropertyNameEnumerator(VM&, JSPropertyNameEnumerator*);
    JSPropertyNameEnumerator* cachedPropertyNameEnumerator() const;
    bool canCachePropertyNameEnumerator() const;
    bool canAccessPropertiesQuicklyForEnumeration() const;

#if 0
    void setCachedOwnKeys(VM&, JSImmutableButterfly*);
    JSImmutableButterfly* cachedOwnKeys() const;
    JSImmutableButterfly* cachedOwnKeysIgnoringSentinel() const;
#endif
    bool canCacheOwnPropertyNames() const;

	void getPropertyNamesFromStructure(JSGlobalData&, PropertyNameArray&, DontEnumPropertiesMode);

    JSValue cachedSpecialProperty(CachedSpecialPropertyKey key)
    {
        if (!hasRareData())
            return JSValue();
        return rareData()->cachedSpecialProperty(key);
    }
    void cacheSpecialProperty(ExecState*, VM&, JSValue, CachedSpecialPropertyKey, const PropertySlot&);

	const ClassInfo* classInfo() const { return m_classInfo; }

	static ptrdiff_t prototypeOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_prototype);
	}

	static ptrdiff_t globalObjectOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_globalObject);
	}

	static ptrdiff_t typeInfoFlagsOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_typeInfo) + TypeInfo::flagsOffset();
	}

	static ptrdiff_t typeInfoTypeOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_typeInfo) + TypeInfo::typeOffset();
	}

    static ptrdiff_t outOfLineTypeFlagsOffset()
    {
		return OBJECT_OFFSETOF(Structure, m_typeInfo) + TypeInfo::outOfLineTypeFlagsOffset();
    }

	static ptrdiff_t classInfoOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_classInfo);
	}

	static ptrdiff_t indexingTypeOffset()
	{
		return OBJECT_OFFSETOF(Structure, m_indexingType);
	}

    static ptrdiff_t propertyTableUnsafeOffset()
    {
        return OBJECT_OFFSETOF(Structure, m_propertyTableUnsafe);
    }

    static ptrdiff_t inlineCapacityOffset()
    {
        return OBJECT_OFFSETOF(Structure, m_inlineCapacity);
    }

	static Structure* createStructure(JSGlobalData&);

	bool transitionWatchpointSetHasBeenInvalidated() const
	{
		return m_transitionWatchpointSet.hasBeenInvalidated();
	}

	bool transitionWatchpointSetIsStillValid() const
	{
		return m_transitionWatchpointSet.isStillValid();
	}

    bool dfgShouldWatchIfPossible() const
    {
        // FIXME: We would like to not watch things that are unprofitable to watch, like
        // dictionaries. Unfortunately, we can't do such things: a dictionary could get flattened,
        // in which case it will start to appear watchable and so the DFG will think that it is
        // watching it. We should come up with a comprehensive story for not watching things that
        // aren't profitable to watch.
        // https://bugs.webkit.org/show_bug.cgi?id=133625

    	// - We don't watch Structures that either decided not to be watched, or whose predecessors
    	//   decided not to be watched. This happens either when a transition is fired while being
    	//   watched.
    	if (transitionWatchpointIsLikelyToBeFired())
    		return false;

        // - Don't watch Structures that had been dictionaries.
        if (hasBeenDictionary())
            return false;

        return true;
    }

    bool dfgShouldWatch() const
    {
        return dfgShouldWatchIfPossible() && transitionWatchpointSetIsStillValid();
    }

	void addTransitionWatchpoint(Watchpoint* watchpoint) const
	{
		ASSERT(transitionWatchpointSetIsStillValid());
		m_transitionWatchpointSet.add(watchpoint);
	}

    void didTransitionFromThisStructureWithoutFiringWatchpoint() const;
    void fireStructureTransitionWatchpoint(DeferredStructureTransitionWatchpointFire*) const;

    InlineWatchpointSet& transitionWatchpointSet() const
    {
        return m_transitionWatchpointSet;
    }

    WatchpointSet* ensurePropertyReplacementWatchpointSet(VM&, PropertyOffset);
    void startWatchingPropertyForReplacements(VM& vm, PropertyOffset offset)
    {
        ensurePropertyReplacementWatchpointSet(vm, offset);
    }
    void startWatchingPropertyForReplacements(VM&, PropertyName);
    RefPtr<WatchpointSet> propertyReplacementWatchpointSet(PropertyOffset);
    void didReplaceProperty(VM&, PropertyOffset);
    void didCachePropertyReplacement(VM&, PropertyOffset);

    void startWatchingInternalPropertiesIfNecessary(VM& vm)
    {
        if (LIKELY(didWatchInternalProperties()))
            return;
        startWatchingInternalProperties(vm);
    }

    void startWatchingInternalPropertiesIfNecessaryForEntireChain(VM& vm)
    {
        for (Structure* structure = this; structure; structure = structure->storedPrototypeStructure())
            structure->startWatchingInternalPropertiesIfNecessary(vm);
    }

    void dump(PrintStream&) const;
    void dumpInContext(PrintStream&, DumpContext*) const;
    void dumpBrief(PrintStream&, const CString&) const;

    static void dumpContextHeader(PrintStream&);

    struct PropertyHashEntry {
    	PropertyHashEntry(const HashTable* table, const HashTableValue* value) : table(table), value(value) { }
        const HashTable* table;
        const HashTableValue* value;
    };
    Optional<PropertyHashEntry> findPropertyHashEntry(const PropertyName&) const;

    DECLARE_EXPORT_INFO;

private:
	enum DictionaryKind {
		NoneDictionaryKind = 0,
		CachedDictionaryKind = 1,
		UncachedDictionaryKind = 2
	};
public:
#define DEFINE_BITFIELD(type, lowerName, upperName, width, offset) \
    static const uint32_t s_##lowerName##Shift = offset;\
    static const uint32_t s_##lowerName##Mask = ((1 << (width - 1)) | ((1 << (width - 1)) - 1));\
    static constexpr uint32_t s_bitWidthOf##upperName = width;\
    ALWAYS_INLINE type lowerName() const { return static_cast<type>((m_bitField >> offset) & s_##lowerName##Mask); }\
    ALWAYS_INLINE void set##upperName(type newValue) \
    {\
        m_bitField &= ~(s_##lowerName##Mask << offset);\
        m_bitField |= (static_cast<uint32_t>(newValue) & s_##lowerName##Mask) << offset;\
    }

    DEFINE_BITFIELD(DictionaryKind, dictionaryKind, DictionaryKind, 2, 0);
    DEFINE_BITFIELD(bool, isPinnedPropertyTable, IsPinnedPropertyTable, 1, 2);
    DEFINE_BITFIELD(bool, hasAnyKindOfGetterSetterProperties, HasAnyKindOfGetterSetterProperties, 1, 3);
    DEFINE_BITFIELD(bool, hasReadOnlyOrGetterSetterPropertiesExcludingProto, HasReadOnlyOrGetterSetterPropertiesExcludingProto, 1, 4);
    DEFINE_BITFIELD(bool, isQuickPropertyAccessAllowedForEnumeration, IsQuickPropertyAccessAllowedForEnumeration, 1, 5);
    DEFINE_BITFIELD(TransitionPropertyAttributes, transitionPropertyAttributes, TransitionPropertyAttributes, 8, 6);
	DEFINE_BITFIELD(TransitionKind, transitionKind, TransitionKind, 6, 14);
    DEFINE_BITFIELD(bool, didPreventExtensions, DidPreventExtensions, 1, 20);
    DEFINE_BITFIELD(bool, didTransition, DidTransition, 1, 21);
    DEFINE_BITFIELD(bool, staticPropertiesReified, StaticPropertiesReified, 1, 22);
    DEFINE_BITFIELD(bool, hasBeenFlattenedBefore, HasBeenFlattenedBefore, 1, 23);
    DEFINE_BITFIELD(bool, isBrandedStructure, IsBrandedStructure, 1, 24);
    DEFINE_BITFIELD(bool, didWatchInternalProperties, DidWatchInternalProperties, 1, 25);
    DEFINE_BITFIELD(bool, transitionWatchpointIsLikelyToBeFired, TransitionWatchpointIsLikelyToBeFired, 1, 26);
	DEFINE_BITFIELD(bool, hasBeenDictionary, HasBeenDictionary, 1, 27);
	DEFINE_BITFIELD(bool, hasUnderscoreProtoPropertyExcludingOriginalProto, HasUnderscoreProtoPropertyExcludingOriginalProto, 1, 28);

    static_assert(s_bitWidthOfTransitionPropertyAttributes <= sizeof(TransitionPropertyAttributes) * 8, "");
    static_assert(s_bitWidthOfTransitionKind <= sizeof(TransitionKind) * 8, "");

    bool isPropertyAdditionTransition() const { return transitionKind() == TransitionKind::PropertyAddition; }
    bool isPropertyDeletionTransition() const { return transitionKind() == TransitionKind::PropertyDeletion; }

protected:
	Structure(JSGlobalData&, Structure*);

private:
	friend class LLIntOffsetsExtractor;

	JS_EXPORT_PRIVATE Structure(JSGlobalData&, JSGlobalObject*, JSValue prototype, const TypeInfo&, const ClassInfo*, IndexingType, unsigned inlineCapacity);
	explicit Structure(JSGlobalData&);

	static Structure* create(JSGlobalData&, Structure*, DeferredStructureTransitionWatchpointFire*);

    JS_EXPORT_PRIVATE static Structure* nonPropertyTransitionSlow(VM&, Structure*, TransitionKind, DeferredStructureTransitionWatchpointFire*);

    // This function does the both didTransitionFromThisStructureWithoutFiringWatchpoint and fireStructureTransitionWatchpoint.
    void didTransitionFromThisStructure(DeferredStructureTransitionWatchpointFire*) const;

	static Structure* setBrandTransitionFromExistingStructureImpl(Structure*, UniquedStringImpl*);

    void findStructuresAndMapForMaterialization(Vector<Structure*, 8>& structures, PropertyTable*&);

	static Structure* toDictionaryTransition(JSGlobalData&, Structure*, DictionaryKind, DeferredStructureTransitionWatchpointFire* = nullptr);

	PropertyOffset add(JSGlobalData&, PropertyName, unsigned attributes);
	PropertyOffset remove(JSGlobalData&, PropertyName);

	void createPropertyMap(JSGlobalData&, unsigned keyCount = 0);
	void checkConsistency();

    // This may grab the lock, or not. Do not call when holding the Structure's lock.
    ALWAYS_INLINE PropertyTable* ensurePropertyTableIfNotEmpty(VM& vm)
    {
        if (PropertyTable* result = m_propertyTableUnsafe.get())
            return result;
        if (!previousID())
            return 0;
        return materializePropertyTable(vm);
    }

    // This may grab the lock, or not. Do not call when holding the Structure's lock.
    ALWAYS_INLINE PropertyTable* ensurePropertyTable(VM& vm)
    {
        if (PropertyTable* result = m_propertyTableUnsafe.get())
            return result;
        return materializePropertyTable(vm);
    }

    PropertyTable* propertyTableOrNull() const
    {
        return m_propertyTableUnsafe.get();
    }

    // This will grab the lock. Do not call when holding the Structure's lock.
    JS_EXPORT_PRIVATE PropertyTable* materializePropertyTable(VM&, bool setPropertyTable = true);

    void setPropertyTable(VM& vm, PropertyTable* table);

	PropertyTable* takePropertyTableOrCloneIfPinned(JSGlobalData&);
	PropertyTable* copyPropertyTableForPinning(JSGlobalData&);

	WriteBarrier<PropertyTable>& propertyTable();

	ALWAYS_INLINE void materializePropertyMapIfNecessaryForPinning(JSGlobalData& vm)
	{
		ASSERT(structure()->classInfo() == info());
        checkOffsetConsistency();
		if (!propertyTable())
			materializePropertyTable(vm);
	}

    void setPreviousID(JSGlobalData& globalData, Structure* transition, Structure* structure)
    {
        if (hasRareData())
            rareData()->setPreviousID(globalData, transition, structure);
        else
            m_previousOrRareData.set(globalData, transition, structure);
    }

    void clearPreviousID()
    {
        if (hasRareData())
            rareData()->clearPreviousID();
        else
            m_previousOrRareData.clear();
    }

	int transitionCount() const
	{
		// Since the number of transitions is always the same as m_offset, we keep the size of Structure down by not storing both.
		return numberOfSlotsForLastOffset(m_offset, m_inlineCapacity);
	}

	bool isValid(JSGlobalObject*, StructureChain* cachedPrototypeChain) const;
	bool isValid(ExecState*, StructureChain* cachedPrototypeChain) const;

	void pin();
    void pinForCaching();

    bool isRareData(JSCell* cell) const
    {
        return cell && cell->structure() != structure();
    }

    StructureRareData* rareData() const
    {
        ASSERT(hasRareData());
        return static_cast<StructureRareData*>(m_previousOrRareData.get());
    }

    StructureRareData* tryRareData()
    {
        JSCell* value = m_previousOrRareData.get();
        WTF::dependentLoadLoadFence();
        if (isRareData(value))
            return static_cast<StructureRareData*>(value);
        return nullptr;
    }

    bool checkOffsetConsistency() const;

    void allocateRareData(JSGlobalData&);

    void startWatchingInternalProperties(VM&);

	static const int s_maxTransitionLength = 64;
	static const int s_maxTransitionLengthForNonEvalPutById = 512;

	WriteBarrier<JSGlobalObject> m_globalObject;
	WriteBarrier<Unknown> m_prototype;
	mutable WriteBarrier<StructureChain> m_cachedPrototypeChain;

    WriteBarrier<JSCell> m_previousOrRareData;

	RefPtr<UniquedStringImpl> m_transitionPropertyName;

	const ClassInfo* m_classInfo;

	StructureTransitionTable m_transitionTable;

    // Should be accessed through propertyTable(). During GC, it may be set to 0 by another thread.
    WriteBarrier<PropertyTable> m_propertyTableUnsafe;

	mutable InlineWatchpointSet m_transitionWatchpointSet;

	COMPILE_ASSERT(firstOutOfLineOffset < 256, firstOutOfLineOffset_fits);

	// m_offset does not account for anonymous slots
	PropertyOffset m_offset;

    TypeInfo m_typeInfo;
    IndexingType m_indexingType { 0 };

    uint8_t m_inlineCapacity { 0 };

    uint32_t m_bitField { 0 };

    TinyBloomFilter m_seenProperties;
};

#if ASSERT_DISABLED
ALWAYS_INLINE bool Structure::checkOffsetConsistency() const
{
	return true;
}
#endif

#if !DO_PROPERTYMAP_CONSTENCY_CHECK
inline void Structure::checkConsistency()
{
    checkOffsetConsistency();
}
#endif

} // namespace JSC
