/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "ArrayPrototype.h"
#include "BrandedStructure.h"
#include "JSArrayBufferView.h"
#include "JSCJSValueInlines.h"
#include "JSGlobalObject.h"
#include "ObjectPrototype.h"
#include "PropertyMapHashTable.h"
#include "Structure.h"

namespace JSC {

inline bool Structure::mayInterceptIndexedAccesses() const
{
    if (indexingTypeIncludingHistory() & MayHaveIndexedAccessors)
        return true;

    // Consider a scenario where object O (of global G1)'s prototype is set to A
    // (of global G2), and G2 is already having a bad time. If an object B with
    // indexed accessors is then set as the prototype of A:
    //      O -> A -> B
    // Then, O should be converted to SlowPutArrayStorage (because it now has an
    // object with indexed accessors in its prototype chain). But it won't be
    // converted because this conversion is done by JSGlobalObject::haveAbadTime(),
    // but G2 is already having a bad time. We solve this by conservatively
    // treating A as potentially having indexed accessors if its global is already
    // having a bad time. Hence, when A is set as O's prototype, O will be
    // converted to SlowPutArrayStorage.

    JSGlobalObject* globalObject = this->globalObject();
    if (!globalObject)
        return false;
    return globalObject->isHavingABadTime();
}

inline JSObject* Structure::storedPrototypeObject() const
{
    JSValue value = m_prototype.get();
    if (value.isNull())
        return nullptr;
    return asObject(value);
}

inline Structure* Structure::storedPrototypeStructure() const
{
    JSObject* object = storedPrototypeObject();
    if (!object)
        return nullptr;
    return object->structure();
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& vm, UniquedStringImpl* uid)
{
    ASSERT(!isCompilationThread());
	ASSERT(structure()->classInfo() == info());
    PropertyTable* table = ensurePropertyTableIfNotEmpty(vm);
    if (!table)
		return invalidOffset;

	PropertyMapEntry* entry = table->get(uid);
	return entry ? entry->offset : invalidOffset;
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& globalData, const PropertyName& propertyName)
{
    return get(globalData, propertyName.uid());
}

template<typename Functor>
void Structure::forEachProperty(VM& vm, const Functor& functor)
{
    if (PropertyTable* table = ensurePropertyTableIfNotEmpty(vm)) {
    	for(auto iter = table->begin(); iter != table->end(); ++iter) {
            if (!functor(*iter))
                return;
    	}
        ensureStillAliveHere(table);
    }
}

ALWAYS_INLINE PropertyOffset Structure::getConcurrently(VM& vm, UniquedStringImpl* uid)
{
    unsigned attributesIgnored;
    return getConcurrently(vm, uid, attributesIgnored);
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& vm, UniquedStringImpl* uid, unsigned& attributes)
{
    ASSERT(!isCompilationThread());
    ASSERT(structure()->classInfo() == info());

    if (m_seenProperties.ruleOut(bitwise_cast<uintptr_t>(uid)))
        return invalidOffset;

    PropertyTable* table = ensurePropertyTableIfNotEmpty(vm);
    if (!table)
        return invalidOffset;

    PropertyMapEntry* entry = table->get(uid);
    if (!entry)
        return invalidOffset;

    attributes = entry->attributes;
    return entry->offset;
}

template<typename Functor>
void Structure::forEachPropertyConcurrently(const Functor& functor)
{
    Vector<Structure*, 8> structures;
    PropertyTable* table;

    findStructuresAndMapForMaterialization(structures, table);

    if (table) {
    	for(auto it = table->begin(); it != table->end(); ++it) {
    		PropertyMapEntry& entry = *it;
            if (!functor(entry)) {
                //structure->m_lock.unlock();
                return;
            }
        }
        //structure->m_lock.unlock();
    }

    for (unsigned i = structures.size(); i--;) {
    	Structure* structure = structures[i];
        if (!structure->m_transitionPropertyName)
            continue;

        switch (structure->transitionKind()) {
        case TransitionKind::PropertyAddition:
        case TransitionKind::PropertyAttributeChange:
            break;
        case TransitionKind::PropertyDeletion:
        case TransitionKind::SetBrand:
            continue;
        default:
            ASSERT_NOT_REACHED();
            break;
        }

        if (!functor(PropertyMapEntry(structure->m_transitionPropertyName.get(), structure->m_offset, structure->transitionPropertyAttributes())))
            return;
    }
}

ALWAYS_INLINE PropertyOffset Structure::get(JSGlobalData& globalData, const PropertyName& propertyName, unsigned& attributes)
{
    return get(globalData, propertyName.uid(), attributes);
}

inline bool Structure::hasIndexingHeader(const JSCell* cell) const
{
    if (hasIndexedProperties(indexingType()))
        return true;

    if (!isTypedView(typedArrayTypeForType(typeInfo().type())))
        return false;

    return jsCast<const JSArrayBufferView*>(cell)->mode() == WastefulTypedArray;
}

ALWAYS_INLINE bool Structure::masqueradesAsUndefined(JSGlobalObject* lexicalGlobalObject)
{
	return typeInfo().masqueradesAsUndefined() && globalObject() == lexicalGlobalObject;
}

ALWAYS_INLINE bool Structure::transitivelyTransitionedFrom(Structure* structureToFind)
{
	for (Structure* current = this; current; current = current->previousID()) {
		if (current == structureToFind)
			return true;
	}
	return false;
}

#if 0
inline void Structure::setCachedOwnKeys(VM& vm, JSImmutableButterfly* ownKeys)
{
    ensureRareData(vm)->setCachedOwnKeys(vm, ownKeys);
}

inline JSImmutableButterfly* Structure::cachedOwnKeys() const
{
    if (!hasRareData())
        return nullptr;
    return rareData()->cachedOwnKeys();
}

inline JSImmutableButterfly* Structure::cachedOwnKeysIgnoringSentinel() const
{
    if (!hasRareData())
        return nullptr;
    return rareData()->cachedOwnKeysIgnoringSentinel();
}
#endif

inline bool Structure::canCacheOwnPropertyNames() const
{
    if (isDictionary())
        return false;
    if (hasIndexedProperties(indexingType()))
        return false;
    if (typeInfo().overridesAnyFormOfGetOwnPropertyNames())
        return false;
    return true;
}

ALWAYS_INLINE JSValue prototypeForLookupPrimitiveImpl(JSGlobalObject* globalObject, const Structure* structure)
{
    ASSERT(!structure->isObject());

    if (structure->typeInfo().type() == StringType)
        return globalObject->stringPrototype();

    if (structure->typeInfo().type() == HeapBigIntType)
        return globalObject->bigIntPrototype();

    ASSERT(structure->typeInfo().type() == SymbolType);
    return globalObject->symbolPrototype();
}

ALWAYS_INLINE JSValue Structure::prototypeForLookup(JSGlobalObject* globalObject) const
{
	if (typeInfo().isObject())
		return m_prototype.get();
	if(typeInfo().type() == StringType)
		return globalObject->stringPrototype();

    if (typeInfo().type() == HeapBigIntType)
        return globalObject->bigIntPrototype();

    if (typeInfo().type() == SymbolType)
        return globalObject->symbolPrototype();
	return jsNull();
}

ALWAYS_INLINE JSValue Structure::prototypeForLookup(ExecState* exec) const
{
	return prototypeForLookup(exec->lexicalGlobalObject());
}

ALWAYS_INLINE StructureChain* Structure::prototypeChain(JSGlobalData& vm, JSGlobalObject* globalObject) const
{
    // We cache our prototype chain so our clients can share it.
    if (!isValid(globalObject, m_cachedPrototypeChain.get())) {
        JSValue prototype = prototypeForLookup(globalObject);
        m_cachedPrototypeChain.set(vm, this, StructureChain::create(vm, prototype.isNull() ? 0 : asObject(prototype)->structure()));
    }
    return m_cachedPrototypeChain.get();
}

ALWAYS_INLINE StructureChain* Structure::prototypeChain(ExecState* exec) const
{
    return prototypeChain(exec->vm(), exec->lexicalGlobalObject());
}

inline bool Structure::isValid(JSGlobalObject* globalObject, StructureChain* cachedPrototypeChain) const
{
	if (!cachedPrototypeChain)
		return false;

	JSValue prototype = prototypeForLookup(globalObject);
	WriteBarrier<Structure>* cachedStructure = cachedPrototypeChain->head();
	while(*cachedStructure && !prototype.isNull()) {
		if (asObject(prototype)->structure() != cachedStructure->get())
			return false;
		++cachedStructure;
		prototype = asObject(prototype)->getPrototypeDirect();
	}
	return prototype.isNull() && !*cachedStructure;
}

ALWAYS_INLINE bool Structure::isValid(ExecState* exec, StructureChain* cachedPrototypeChain) const
{
    return isValid(exec->lexicalGlobalObject(), cachedPrototypeChain);
}

ALWAYS_INLINE bool Structure::putWillGrowOutOfLineStorage()
{
    checkOffsetConsistency();

    ASSERT(outOfLineCapacity() >= outOfLineSize());

    if (!propertyTable()) {
        unsigned currentSize = numberOfOutOfLineSlotsForLastOffset(m_offset);
        ASSERT(outOfLineCapacity() >= currentSize);
        return currentSize == outOfLineCapacity();
    }

    ASSERT(totalStorageCapacity() >= propertyTable()->propertyStorageSize());
    if (propertyTable()->hasDeletedOffset())
        return false;

    ASSERT(totalStorageCapacity() >= propertyTable()->size());
    return propertyTable()->size() == totalStorageCapacity();
}

ALWAYS_INLINE WriteBarrier<PropertyTable>& Structure::propertyTable()
{
    ASSERT(!globalObject() || !globalObject()->vm().heap.isBusy());
    return m_propertyTableUnsafe;
}

inline void Structure::didReplaceProperty(VM& vm, PropertyOffset offset)
{
    if (LIKELY(!hasRareData()))
        return;
    WatchpointSet* set = rareData()->m_replacementWatchpointSets.get(offset);
    if (LIKELY(!set))
        return;
    set->fireAll(vm, "Property did get replaced");
}

inline RefPtr<WatchpointSet> Structure::propertyReplacementWatchpointSet(PropertyOffset offset)
{
    //ConcurrentJSLocker locker(m_lock);
    StructureRareData* rareData = tryRareData();
    if (!rareData)
        return nullptr;
    return rareData->m_replacementWatchpointSets.get(offset);
}

#if !ASSERT_DISABLED
ALWAYS_INLINE bool Structure::checkOffsetConsistency() const
{
    PropertyTable* propertyTable = m_propertyTableUnsafe.get();

    if (!propertyTable) {
        ASSERT(!isPinnedPropertyTable());
        return true;
    }

    // We cannot reliably assert things about the property table in the concurrent
    // compilation thread. It is possible for the table to be stolen and then have
    // things added to it, which leads to the offsets being all messed up. We could
    // get around this by grabbing a lock here, but I think that would be overkill.
    if (isCompilationThread())
        return true;
    unsigned totalSize = propertyTable->propertyStorageSize();
    unsigned inlineOverflowAccordingToTotalSize = totalSize < m_inlineCapacity ? 0 : totalSize - m_inlineCapacity;

    auto fail = [&] (const char* description) {
        dataLog("Detected offset inconsistency: ", description, "!\n");
        dataLog("this = ", RawPointer(this), "\n");
        dataLog("m_offset = ", m_offset, "\n");
        dataLog("m_inlineCapacity = ", m_inlineCapacity, "\n");
        dataLog("propertyTable = ", RawPointer(propertyTable), "\n");
        dataLog("numberOfSlotsForLastOffset = ", numberOfSlotsForLastOffset(m_offset, m_inlineCapacity), "\n");
        dataLog("totalSize = ", totalSize, "\n");
        dataLog("inlineOverflowAccordingToTotalSize = ", inlineOverflowAccordingToTotalSize, "\n");
        dataLog("numberOfOutOfLineSlotsForLastOffset = ", numberOfOutOfLineSlotsForLastOffset(m_offset), "\n");
        UNREACHABLE_FOR_PLATFORM();
    };

    if (numberOfSlotsForLastOffset(m_offset, m_inlineCapacity) != totalSize)
        fail("numberOfSlotsForLastOffset doesn't match totalSize");
    if (inlineOverflowAccordingToTotalSize != numberOfOutOfLineSlotsForLastOffset(m_offset))
        fail("inlineOverflowAccordingToTotalSize doesn't match numberOfOutOfLineSlotsForLastOffset");

    return true;
}
#endif

ALWAYS_INLINE size_t nextOutOfLineStorageCapacity(size_t currentCapacity)
{
    if (!currentCapacity)
        return initialOutOfLineCapacity;
    return currentCapacity * outOfLineGrowthFactor;
}

inline PropertyOffset Structure::add(JSGlobalData& vm, PropertyName propertyName, unsigned attributes)
{
    //GCSafeConcurrentJITLocker locker(m_lock, vm.heap);

    setDidWatchInternalProperties(false);

    ASSERT(!JSC::isValidOffset(get(vm, propertyName)));

    checkConsistency();
    if (attributes & DontEnum || propertyName.isSymbol())
        setIsQuickPropertyAccessAllowedForEnumeration(false);
    if (propertyName == vm.propertyNames->underscoreProto)
        setHasUnderscoreProtoPropertyExcludingOriginalProto(true);

    auto* rep = propertyName.uid();

    if (!propertyTable())
        createPropertyMap(vm);

    m_seenProperties.add(bitwise_cast<uintptr_t>(rep));

    PropertyOffset newOffset = propertyTable()->nextOffset(m_inlineCapacity);

    propertyTable()->add(vm, PropertyMapEntry(rep, newOffset, attributes), m_offset, PropertyTable::PropertyOffsetMayChange);

    checkConsistency();
    return newOffset;
}

inline PropertyOffset Structure::remove(VM& vm, PropertyName propertyName)
{
    //GCSafeConcurrentJSLocker locker(m_lock, vm.heap);
    checkConsistency();

    auto* rep = propertyName.uid();

    if (!propertyTable())
        return invalidOffset;

    PropertyTable::find_iterator position = propertyTable()->find(rep);
    if (!position.first)
        return invalidOffset;

    PropertyOffset offset = position.first->offset;

    propertyTable()->remove(vm, position);
    propertyTable()->addDeletedOffset(offset);

    checkConsistency();
    return offset;
}

ALWAYS_INLINE PropertyOffset Structure::addPropertyWithoutTransition(JSGlobalData& vm, PropertyName propertyName, unsigned attributes)
{
    DeferGC deferGC(vm.heap); // JetStream 1.1 box2d failure when stressing GC with DFG JIT
    materializePropertyMapIfNecessaryForPinning(vm);

    pin();

    return add(vm, propertyName, attributes);
}

inline PropertyOffset Structure::removePropertyWithoutTransition(JSGlobalData& vm, PropertyName propertyName)
{
    ASSERT(isUncacheableDictionary());

    DeferGC deferGC(vm.heap); // JetStream 1.1 box2d failure when stressing GC with DFG JIT
    materializePropertyMapIfNecessaryForPinning(vm);

    pin();
    return remove(vm, propertyName);
}

ALWAYS_INLINE size_t Structure::suggestedNewOutOfLineStorageCapacity()
{
    return nextOutOfLineStorageCapacity(outOfLineCapacity());
}

inline void Structure::cacheSpecialProperty(ExecState* exec, VM& vm, JSValue value, CachedSpecialPropertyKey key, const PropertySlot& slot)
{
    if (!hasRareData())
        allocateRareData(vm);
    rareData()->cacheSpecialProperty(exec, vm, this, value, key, slot);
}

ALWAYS_INLINE void Structure::setPropertyTable(VM& vm, PropertyTable* table)
{
    m_propertyTableUnsafe.setMayBeNull(vm, this, table);
}

ALWAYS_INLINE bool Structure::isCheapDuringGC()
{
    // FIXME: We could make this even safer by returning false if this structure's property table
    // has any large property names.
    // https://bugs.webkit.org/show_bug.cgi?id=157334

	if(typeInfo().type() > LastJSCObjectType) // DOM node is not cheap
		return false;

    return (!m_globalObject || Heap::isMarked(m_globalObject.get()))
        && (!storedPrototypeObject() || Heap::isMarked(storedPrototypeObject()));
}

ALWAYS_INLINE bool Structure::markIfCheap(SlotVisitor& visitor)
{
    if (!isCheapDuringGC())
        return Heap::isMarked(this);

    visitor.appendUnbarrieredReadOnlyPointer(this);
    return true;
}

ALWAYS_INLINE Structure* Structure::nonPropertyTransition(JSGlobalData& vm, Structure* structure, TransitionKind transitionKind, DeferredStructureTransitionWatchpointFire* deferred)
{
    IndexingType indexingType = newIndexingType(structure->indexingTypeIncludingHistory(), transitionKind);

    if (changesIndexingType(transitionKind)) {
		if (JSGlobalObject* globalObject = structure->m_globalObject.get()) {
			if (globalObject->isOriginalArrayStructure(structure)) {
				Structure* result = globalObject->originalArrayStructureForIndexingType(indexingType);
				if (result->indexingTypeIncludingHistory() == indexingType) {
					structure->didTransitionFromThisStructure(deferred);
					return result;
				}
			}
		}
    }

    return nonPropertyTransitionSlow(vm, structure, transitionKind, deferred);
}

ALWAYS_INLINE bool Structure::canPerformFastPropertyEnumeration() const
{
    if (typeInfo().overridesGetOwnPropertySlot())
        return false;
    if (typeInfo().overridesAnyFormOfGetOwnPropertyNames())
        return false;
    // FIXME: Indexed properties can be handled.
    // https://bugs.webkit.org/show_bug.cgi?id=185358
    if (hasIndexedProperties(indexingType()))
        return false;
    if (hasAnyKindOfGetterSetterProperties())
        return false;
    if (hasReadOnlyOrGetterSetterPropertiesExcludingProto())
        return false;
    if (isUncacheableDictionary())
        return false;
    // Cannot perform fast [[Put]] to |target| if the property names of the |source| contain "__proto__".
    if (hasUnderscoreProtoPropertyExcludingOriginalProto())
        return false;
    return true;
}

} // namespace JSC
