/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "VMEntryScope.h"

#include "AssemblerBuffer.h"
#include "Debugger.h"
#include "JSGlobalObject.h"
#include "JSObjectInlines.h"
#include "VM.h"
#include <wtf/StackBounds.h>
#include <wtf/text/StringBuffer.h>
#include <wtf/text/WTFString.h>
#include <stdio.h>

namespace JSC {

VMEntryScope::VMEntryScope(VM& vm, JSGlobalObject* globalObject)
    : m_vm(vm)
    , m_globalObject(globalObject)
{
    if (!vm.entryScope) {
#if ENABLE(ASSEMBLER)
        if (vm.heap.shouldDeleteCodeBlocks())
            vm.heap.deleteCodeBlocks(false);
#endif
        vm.entryScope = this;

        // Reset the date cache between JS invocations to force the VM to
        // observe time xone changes.
        vm.resetDateCache();

        vm.executionStartTime = monotonicallyIncreasingTime();
        vm.jitCompileLength = 0;
        vm.parseLength = 0;
    }

    vm.clearLastException();
}

void VMEntryScope::addDidPopListener(PassOwnPtr<VMIdleCallback> listener)
{
    m_didPopListeners.append(listener);
}

VMEntryScope::~VMEntryScope()
{
	int jitCompileLengthMS = (int)(m_vm.jitCompileLength * 1000);
	if(jitCompileLengthMS > 10)
		WTF::logMessage(tryMakeString("JIT compile takes ", String::number(jitCompileLengthMS), "ms"));

	int parseLengthMS = (int)(m_vm.parseLength * 1000);
	if(parseLengthMS > 10)
		WTF::logMessage(tryMakeString("JIT parse takes ", String::number(parseLengthMS), "ms"));

	m_vm.jitCompileLength = 0;
	m_vm.parseLength = 0;

    if (m_vm.entryScope != this)
        return;

    m_vm.entryScope = nullptr;

	Vector<OwnPtr<VMIdleCallback>, 0, CrashOnOverflow, 1> didPopListeners;

	didPopListeners.swap(m_didPopListeners);

	while(!didPopListeners.isEmpty()) {
		(*didPopListeners[0])(m_vm, m_globalObject);
		didPopListeners.remove(0);
	}

    m_vm.clearScratchBuffers();

    m_vm.purgeWatchpointSetForImpurePropertyIfNecessary();

	if(ExecutableAllocator::underMemoryPressure())
        m_vm.heap.deleteCodeBlocks(false);

	double now = monotonicallyIncreasingTime();
	int ms = (int)((now - m_vm.executionStartTime) * 1000);
	if(ms > 15) {
		char outBuffer[256];
		int evalPrepareForExecutionLengthMS = (int)(m_vm.evalPrepareForExecutionLength * 1000);
		int totalWritten = snprintf(outBuffer, sizeof(outBuffer), "JS execution takes %ims  [eval count %i, prepare for execution takes %ims] at %f", ms, m_vm.evalCount, evalPrepareForExecutionLengthMS, currentTime());
		if (totalWritten > 0 && totalWritten < sizeof(outBuffer))
			WTF::logMessage(String(outBuffer, totalWritten));
	}

	m_vm.evalPrepareForExecutionLength = 0;
	m_vm.evalCount = 0;
}

} // namespace JSC
