/*
 * Copyright (C) 2013 Apple, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "WeakMapConstructor.h"

#include "Error.h"
#include "IteratorOperations.h"
#include "JSCJSValueInlines.h"
#include "JSCellInlines.h"
#include "JSGlobalObject.h"
#include "JSObjectInlines.h"
#include "JSWeakMap.h"
#include "StructureInlines.h"
#include "WeakMapPrototype.h"

namespace JSC {

const ClassInfo WeakMapConstructor::s_info = { "Function", &Base::s_info, 0, CREATE_METHOD_TABLE(WeakMapConstructor) };

void WeakMapConstructor::finishCreation(VM& vm, WeakMapPrototype* prototype)
{
    Base::finishCreation(vm, String::fromLiteral(prototype->classInfo()->className));
    putDirectWithoutTransition(vm, vm.propertyNames->prototype, prototype, DontEnum | DontDelete | ReadOnly);
    putDirectWithoutTransition(vm, vm.propertyNames->length, jsNumber(0), DontEnum | ReadOnly);
}

static EncodedJSValue JSC_HOST_CALL callWeakMap(ExecState*);
static EncodedJSValue JSC_HOST_CALL constructWeakMap(ExecState*);

WeakMapConstructor::WeakMapConstructor(VM& vm, Structure* structure)
    : Base(vm, structure, callWeakMap, constructWeakMap)
{
}

static EncodedJSValue JSC_HOST_CALL callWeakMap(ExecState* exec)
{
    return JSValue::encode(throwConstructorCannotBeCalledAsFunctionTypeError(exec, "WeakMap"));
}

static EncodedJSValue JSC_HOST_CALL constructWeakMap(ExecState* exec)
{
	JSGlobalObject* globalObject = exec->lexicalGlobalObject();
    VM& vm = globalObject->vm();

    JSObject* newTarget = asObject(exec->newTarget());
    Structure* weakMapStructure = JSC_GET_DERIVED_STRUCTURE(vm, exec, weakMapStructure, newTarget, exec->jsCallee());
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

	JSWeakMap* weakMap = JSWeakMap::create(exec, weakMapStructure);
    JSValue iterable = exec->argument(0);
    if (iterable.isUndefinedOrNull())
        return JSValue::encode(weakMap);

    JSValue adderFunction = weakMap->JSObject::get(vm, exec, vm.propertyNames->set);
    RETURN_IF_EXCEPTION(vm, encodedJSUndefined());

    CallData adderFunctionCallData = getCallData(vm, adderFunction);
    if (adderFunctionCallData.type == CallData::Type::None)
        return JSValue::encode(throwTypeError(exec, vm));

    struct Functor
    {
    	Functor(JSValue adderFunction, const CallData& adderFunctionCallData, JSWeakMap* weakMap)
    		: adderFunction(adderFunction)
    		, adderFunctionCallData(adderFunctionCallData)
    		, weakMap(weakMap)
    	{
    	}
		void operator()(VM& vm, ExecState* exec, JSValue nextItem) {
			if (!nextItem.isObject()) {
				throwTypeError(exec, vm);
	            return;
			}

			JSValue key = nextItem.get(vm, exec, static_cast<unsigned>(0));
		    RETURN_IF_EXCEPTION(vm, void());

			JSValue value = nextItem.get(vm, exec, static_cast<unsigned>(1));
		    RETURN_IF_EXCEPTION(vm, void());

			MarkedArgumentBuffer arguments;
			arguments.append(key);
			arguments.append(value);
			call(exec, adderFunction, adderFunctionCallData, weakMap, arguments);
		}
		JSValue adderFunction;
		const CallData& adderFunctionCallData;
		JSWeakMap* weakMap;
    };
    Functor functor(adderFunction, adderFunctionCallData, weakMap);
    forEachInIterable(vm, *exec, iterable, functor);

    return JSValue::encode(weakMap);
}

}
