/*
 * Copyright (C) 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <wtf/Assertions.h>
#include <wtf/FastAllocBase.h>

namespace WTF {

template <typename T, size_t Size> class FixedArray {
	WTF_MAKE_FAST_ALLOCATED;
public:
	typedef T* iterator;
	typedef const T* const_iterator;

	FixedArray() = default;
	explicit FixedArray(const T& value)
	{
		fill(value);
	}
	void fill(const T& value)
	{
		for(int i = 0; i < Size; ++i)
			m_data[i] = value;
	}

	ALWAYS_INLINE T& operator[](size_t i)
    {
        ASSERT(i < Size);
        return m_data[i];
    }

	ALWAYS_INLINE const T& operator[](size_t i) const
    {
        ASSERT(i < Size);
        return m_data[i];
    }

	ALWAYS_INLINE T* data() { return m_data; }
    const T* data() const { return m_data; }
	ALWAYS_INLINE size_t size() const { return Size; }

	ALWAYS_INLINE iterator begin() { return data(); }
	ALWAYS_INLINE iterator end() { return begin() + Size; }
	ALWAYS_INLINE const_iterator begin() const { return data(); }
	ALWAYS_INLINE const_iterator end() const { return begin() + Size; }

    bool operator==(const FixedArray& other) const
    {
		for(int i = 0; i < Size; ++i)
			if(m_data[i] != other.m_data[i])
				return false;
		return true;
    }

    bool operator!=(const FixedArray& other) const
    {
        return !(*this == other);
    }

    bool operator<(const FixedArray& other) const
    {
		for(int i = 0; i < Size; ++i) {
			if(m_data[i] < other.m_data[i])
				continue;
			return false;
		}
		return true;
    }

private:
    T m_data[Size];
};

} // namespace WTF

using WTF::FixedArray;
