/*
 * Copyright (C) 2005, 2006, 2007, 2008, 2011 Apple Inc. All rights reserved.
 * Copyright (C) 2011, Benjamin Poulain <ikipou@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include <wtf/HashSet.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>

namespace WTF {

// ListHashSet: Just like HashSet, this class provides a Set
// interface - a collection of unique objects with O(1) insertion,
// removal and test for containership. However, it also has an
// order - iterating it will always give back values in the order
// in which they are added.

// Unlike iteration of most WTF Hash data structures, iteration is
// guaranteed safe against mutation of the ListHashSet, except for
// removal of the item currently pointed to by a given iterator.

template<typename Value, typename HashFunctions> class ListHashSet;

template<typename ValueArg, typename HashArg> class ListHashSetIterator;
template<typename ValueArg, typename HashArg> class ListHashSetConstIterator;
template<typename ValueArg, typename HashArg> class ListHashSetReverseIterator;
template<typename ValueArg, typename HashArg> class ListHashSetConstReverseIterator;

template<typename ValueArg> struct ListHashSetNode;

template<typename HashArg> struct ListHashSetNodeHashFunctions;
template<typename HashArg> struct ListHashSetTranslator;

template<typename ValueArg, typename HashArg = typename DefaultHash<ValueArg>::Hash> class ListHashSet {
	WTF_MAKE_FAST_ALLOCATED;
private:
	typedef ListHashSetNode<ValueArg> Node;

	typedef HashTraits<Node*> NodeTraits;
	typedef ListHashSetNodeHashFunctions<HashArg> NodeHash;
	typedef ListHashSetTranslator<HashArg> BaseTranslator;

	typedef HashTable<Node*, Node*, IdentityExtractor, NodeHash, NodeTraits, NodeTraits> ImplType;
	typedef HashTableIterator<Node*, Node*, IdentityExtractor, NodeHash, NodeTraits, NodeTraits> ImplTypeIterator;
	typedef HashTableConstIterator<Node*, Node*, IdentityExtractor, NodeHash, NodeTraits, NodeTraits> ImplTypeConstIterator;

	typedef HashArg HashFunctions;

public:
	typedef ValueArg ValueType;

	typedef ListHashSetIterator<ValueType, HashArg> iterator;
	typedef ListHashSetConstIterator<ValueType, HashArg> const_iterator;
	friend class ListHashSetConstIterator<ValueType, HashArg>;

	typedef ListHashSetReverseIterator<ValueType, HashArg> reverse_iterator;
	typedef ListHashSetConstReverseIterator<ValueType, HashArg> const_reverse_iterator;
	friend class ListHashSetConstReverseIterator<ValueType, HashArg>;

	typedef HashTableAddResult<iterator> AddResult;

	ListHashSet();
	ListHashSet(const ListHashSet&);
    ListHashSet(ListHashSet&&);
	ListHashSet& operator=(const ListHashSet&);
    ListHashSet& operator=(ListHashSet&&);
	~ListHashSet();

	void swap(ListHashSet&);

	int size() const;
	int capacity() const;
	bool isEmpty() const;

	iterator begin();
	iterator end();
	const_iterator begin() const;
	const_iterator end() const;

    iterator random() { return makeIterator(m_impl.random()); }
    const_iterator random() const { return makeIterator(m_impl.random()); }

	reverse_iterator rbegin();
	reverse_iterator rend();
	const_reverse_iterator rbegin() const;
	const_reverse_iterator rend() const;

	ValueType& first();
	const ValueType& first() const;
    void removeFirst();
    ValueType takeFirst();

	ValueType& last();
	const ValueType& last() const;
    void removeLast();
    ValueType takeLast();

	iterator find(const ValueType&);
	const_iterator find(const ValueType&) const;
	bool contains(const ValueType&) const;

	// An alternate version of find() that finds the object by hashing and comparing
	// with some other type, to avoid the cost of type conversion.
	// The HashTranslator interface is defined in HashSet.
	// FIXME: We should reverse the order of the template arguments so that callers
	// can just pass the translator let the compiler deduce T.
	template<typename T, typename HashTranslator> iterator find(const T&);
	template<typename T, typename HashTranslator> const_iterator find(const T&) const;
	template<typename T, typename HashTranslator> bool contains(const T&) const;

	// The return value of add is a pair of an iterator to the new value's location,
	// and a bool that is true if an new entry was added.
	AddResult add(const ValueType&);
    AddResult add(ValueType&&);

    // Add the value to the end of the collection. If the value was already in
    // the list, it is moved to the end.
    AddResult appendOrMoveToLast(const ValueType&);
    AddResult appendOrMoveToLast(ValueType&&);

    // Add the value to the beginning of the collection. If the value was already in
    // the list, it is moved to the beginning.
    AddResult prependOrMoveToFirst(const ValueType&);
    AddResult prependOrMoveToFirst(ValueType&&);

	AddResult insertBefore(const ValueType& beforeValue, const ValueType& newValue);
    AddResult insertBefore(const ValueType& beforeValue, ValueType&& newValue);
	AddResult insertBefore(iterator, const ValueType&);
    AddResult insertBefore(iterator, ValueType&&);

	bool remove(const ValueType&);
	bool remove(iterator);
	void clear();

private:
    void unlink(Node*);
	void unlinkAndDelete(Node*);
	void appendNode(Node*);
    void prependNode(Node*);
	void insertNodeBefore(Node* beforeNode, Node* newNode);
	void deleteAllNodes();

	iterator makeIterator(Node*);
	const_iterator makeConstIterator(Node*) const;
	reverse_iterator makeReverseIterator(Node*);
	const_reverse_iterator makeConstReverseIterator(Node*) const;

	ImplType m_impl;
	Node* m_head;
	Node* m_tail;
};

template<typename ValueArg> struct ListHashSetNode {
    WTF_MAKE_FAST_ALLOCATED;
public:
    template<typename T>
    ListHashSetNode(T&& value)
    	: m_value(std::forward<T>(value))
		, m_prev(0)
		, m_next(0)
	{
	}

	ValueArg m_value;
	ListHashSetNode* m_prev;
	ListHashSetNode* m_next;
};

template<typename HashArg> struct ListHashSetNodeHashFunctions {
	template<typename T> static ALWAYS_INLINE unsigned hash(const T& key) { return HashArg::hash(key->m_value); }
	template<typename T> static ALWAYS_INLINE bool equal(const T& a, const T& b) { return HashArg::equal(a->m_value, b->m_value); }
	static const bool safeToCompareToEmptyOrDeleted = false;
};

template<typename ValueArg, typename HashArg> class ListHashSetIterator {
private:
	typedef ListHashSet<ValueArg, HashArg> ListHashSetType;
	typedef ListHashSetIterator<ValueArg, HashArg> iterator;
	typedef ListHashSetConstIterator<ValueArg, HashArg> const_iterator;
	typedef ListHashSetNode<ValueArg> Node;
	typedef ValueArg ValueType;
	typedef ValueType& ReferenceType;
	typedef ValueType* PointerType;

	friend class ListHashSet<ValueArg, HashArg>;

	ALWAYS_INLINE ListHashSetIterator(const ListHashSetType* set, Node* position) : m_iterator(set, position) { }

public:
	ALWAYS_INLINE ListHashSetIterator() { }

	// default copy, assignment and destructor are OK

	ALWAYS_INLINE PointerType get() const { return const_cast<PointerType>(m_iterator.get()); }
	ALWAYS_INLINE ReferenceType operator*() const { return *get(); }
	ALWAYS_INLINE PointerType operator->() const { return get(); }

	ALWAYS_INLINE iterator& operator++() { ++m_iterator; return *this; }

	// postfix ++ intentionally omitted

	ALWAYS_INLINE iterator& operator--() { --m_iterator; return *this; }

	// postfix -- intentionally omitted

	// Comparison.
	ALWAYS_INLINE bool operator==(const iterator& other) const { return m_iterator == other.m_iterator; }
	ALWAYS_INLINE bool operator!=(const iterator& other) const { return m_iterator != other.m_iterator; }

	ALWAYS_INLINE operator const_iterator() const { return m_iterator; }

private:
	ALWAYS_INLINE Node* node() { return m_iterator.node(); }

	const_iterator m_iterator;
};

template<typename ValueArg, typename HashArg> class ListHashSetConstIterator {
private:
	typedef ListHashSet<ValueArg, HashArg> ListHashSetType;
	typedef ListHashSetIterator<ValueArg, HashArg> iterator;
	typedef ListHashSetConstIterator<ValueArg, HashArg> const_iterator;
	typedef ListHashSetNode<ValueArg> Node;
	typedef ValueArg ValueType;
	typedef const ValueType& ReferenceType;
	typedef const ValueType* PointerType;

	friend class ListHashSet<ValueArg, HashArg>;
	friend class ListHashSetIterator<ValueArg, HashArg>;

	ALWAYS_INLINE ListHashSetConstIterator(const ListHashSetType* set, Node* position)
		: m_set(set)
		, m_position(position)
	{
	}

public:
	ALWAYS_INLINE ListHashSetConstIterator()
	{
	}

	ALWAYS_INLINE PointerType get() const
	{
		return &m_position->m_value;
	}
	ALWAYS_INLINE ReferenceType operator*() const { return *get(); }
	ALWAYS_INLINE PointerType operator->() const { return get(); }

	ALWAYS_INLINE const_iterator& operator++()
	{
		ASSERT(m_position);
		m_position = m_position->m_next;
		return *this;
	}

	// postfix ++ intentionally omitted

	ALWAYS_INLINE const_iterator& operator--()
	{
		ASSERT(m_position != m_set->m_head);
		if (!m_position)
			m_position = m_set->m_tail;
		else
			m_position = m_position->m_prev;
		return *this;
	}

	// postfix -- intentionally omitted

	// Comparison.
	ALWAYS_INLINE bool operator==(const const_iterator& other) const
	{
		return m_position == other.m_position;
	}
	ALWAYS_INLINE bool operator!=(const const_iterator& other) const
	{
		return m_position != other.m_position;
	}

private:
	ALWAYS_INLINE Node* node() { return m_position; }

	const ListHashSetType* m_set;
	Node* m_position;
};

template<typename ValueArg, typename HashArg> class ListHashSetReverseIterator {
private:
	typedef ListHashSet<ValueArg, HashArg> ListHashSetType;
	typedef ListHashSetReverseIterator<ValueArg, HashArg> reverse_iterator;
	typedef ListHashSetConstReverseIterator<ValueArg, HashArg> const_reverse_iterator;
	typedef ListHashSetNode<ValueArg> Node;
	typedef ValueArg ValueType;
	typedef ValueType& ReferenceType;
	typedef ValueType* PointerType;

	friend class ListHashSet<ValueArg, HashArg>;

	ALWAYS_INLINE ListHashSetReverseIterator(const ListHashSetType* set, Node* position) : m_iterator(set, position) { }

public:
	ALWAYS_INLINE ListHashSetReverseIterator() { }

	// default copy, assignment and destructor are OK

	ALWAYS_INLINE PointerType get() const { return const_cast<PointerType>(m_iterator.get()); }
	ALWAYS_INLINE ReferenceType operator*() const { return *get(); }
	ALWAYS_INLINE PointerType operator->() const { return get(); }

	ALWAYS_INLINE reverse_iterator& operator++() { ++m_iterator; return *this; }

	// postfix ++ intentionally omitted

	ALWAYS_INLINE reverse_iterator& operator--() { --m_iterator; return *this; }

	// postfix -- intentionally omitted

	// Comparison.
	ALWAYS_INLINE bool operator==(const reverse_iterator& other) const { return m_iterator == other.m_iterator; }
	ALWAYS_INLINE bool operator!=(const reverse_iterator& other) const { return m_iterator != other.m_iterator; }

	ALWAYS_INLINE operator const_reverse_iterator() const { return m_iterator; }

private:
	ALWAYS_INLINE Node* node() { return m_iterator.node(); }

	const_reverse_iterator m_iterator;
};

template<typename ValueArg, typename HashArg> class ListHashSetConstReverseIterator {
private:
	typedef ListHashSet<ValueArg, HashArg> ListHashSetType;
	typedef ListHashSetReverseIterator<ValueArg, HashArg> reverse_iterator;
	typedef ListHashSetConstReverseIterator<ValueArg, HashArg> const_reverse_iterator;
	typedef ListHashSetNode<ValueArg> Node;
	typedef ValueArg ValueType;
	typedef const ValueType& ReferenceType;
	typedef const ValueType* PointerType;

	friend class ListHashSet<ValueArg, HashArg>;
	friend class ListHashSetReverseIterator<ValueArg, HashArg>;

	ALWAYS_INLINE ListHashSetConstReverseIterator(const ListHashSetType* set, Node* position)
		: m_set(set)
		, m_position(position)
	{
	}

public:
	ALWAYS_INLINE ListHashSetConstReverseIterator()
	{
	}

	ALWAYS_INLINE PointerType get() const
	{
		return &m_position->m_value;
	}
	ALWAYS_INLINE ReferenceType operator*() const { return *get(); }
	ALWAYS_INLINE PointerType operator->() const { return get(); }

	ALWAYS_INLINE const_reverse_iterator& operator++()
	{
		ASSERT(m_position != 0);
		m_position = m_position->m_prev;
		return *this;
	}

	// postfix ++ intentionally omitted

	ALWAYS_INLINE const_reverse_iterator& operator--()
	{
		ASSERT(m_position != m_set->m_tail);
		if (!m_position)
			m_position = m_set->m_head;
		else
			m_position = m_position->m_next;
		return *this;
	}

	// postfix -- intentionally omitted

	// Comparison.
	ALWAYS_INLINE bool operator==(const const_reverse_iterator& other) const
	{
		return m_position == other.m_position;
	}
	ALWAYS_INLINE bool operator!=(const const_reverse_iterator& other) const
	{
		return m_position != other.m_position;
	}

private:
	ALWAYS_INLINE Node* node() { return m_position; }

	const ListHashSetType* m_set;
	Node* m_position;
};

template<typename HashFunctions>
struct ListHashSetTranslator {
	template<typename T> static ALWAYS_INLINE unsigned hash(const T& key) { return HashFunctions::hash(key); }
	template<typename T, typename U> static ALWAYS_INLINE bool equal(const T& a, const U& b) { return HashFunctions::equal(a->m_value, b); }
	template<typename T, typename U, typename V> static ALWAYS_INLINE void translate(T*& location, U&& key, V&&)
	{
		location = new T(std::forward<U>(key));
	}
};

template<typename T, typename U>
ALWAYS_INLINE ListHashSet<T, U>::ListHashSet()
	: m_head(0)
	, m_tail(0)
{
}

template<typename T, typename U>
inline ListHashSet<T, U>::ListHashSet(const ListHashSet& other)
	: m_head(0)
	, m_tail(0)
{
	const_iterator end = other.end();
	for (const_iterator it = other.begin(); it != end; ++it)
		add(*it);
}

template<typename T, typename U>
inline ListHashSet<T, U>& ListHashSet<T, U>::operator=(const ListHashSet& other)
{
	ListHashSet tmp(other);
	swap(tmp);
	return *this;
}

template<typename T, typename U>
inline ListHashSet<T, U>::ListHashSet(ListHashSet&& other)
    : m_impl(WTFMove(other.m_impl))
    , m_head(other.m_head)
    , m_tail(other.m_tail)
{
	other.m_head = nullptr;
	other.m_tail = nullptr;
}

template<typename T, typename U>
inline ListHashSet<T, U>& ListHashSet<T, U>::operator=(ListHashSet&& other)
{
    ListHashSet movedSet(WTFMove(other));
    swap(movedSet);
    return *this;
}

template<typename T, typename U>
inline void ListHashSet<T, U>::swap(ListHashSet& other)
{
	m_impl.swap(other.m_impl);
	std::swap(m_head, other.m_head);
	std::swap(m_tail, other.m_tail);
}

template<typename T, typename U>
inline ListHashSet<T, U>::~ListHashSet()
{
	deleteAllNodes();
}

template<typename T, typename U>
ALWAYS_INLINE int ListHashSet<T, U>::size() const
{
	return m_impl.size();
}

template<typename T, typename U>
ALWAYS_INLINE int ListHashSet<T, U>::capacity() const
{
	return m_impl.capacity();
}

template<typename T, typename U>
ALWAYS_INLINE bool ListHashSet<T, U>::isEmpty() const
{
	return m_impl.isEmpty();
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::iterator ListHashSet<T, U>::begin()
{
	return makeIterator(m_head);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::iterator ListHashSet<T, U>::end()
{
	return makeIterator(0);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::const_iterator ListHashSet<T, U>::begin() const
{
	return makeConstIterator(m_head);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::const_iterator ListHashSet<T, U>::end() const
{
	return makeConstIterator(0);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::reverse_iterator ListHashSet<T, U>::rbegin()
{
	return makeReverseIterator(m_tail);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::reverse_iterator ListHashSet<T, U>::rend()
{
	return makeReverseIterator(0);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::const_reverse_iterator ListHashSet<T, U>::rbegin() const
{
	return makeConstReverseIterator(m_tail);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::const_reverse_iterator ListHashSet<T, U>::rend() const
{
	return makeConstReverseIterator(0);
}

template<typename T, typename U>
ALWAYS_INLINE T& ListHashSet<T, U>::first()
{
	ASSERT(!isEmpty());
	return m_head->m_value;
}

template<typename T, typename U>
inline void ListHashSet<T, U>::removeFirst()
{
    takeFirst();
}

template<typename T, typename U>
inline T ListHashSet<T, U>::takeFirst()
{
    ASSERT(!isEmpty());
    auto it = m_impl.find(m_head);

    T result = WTFMove((*it)->m_value);
    m_impl.remove(it);
    unlinkAndDelete(m_head);

    return result;
}

template<typename T, typename U>
ALWAYS_INLINE const T& ListHashSet<T, U>::first() const
{
	ASSERT(!isEmpty());
	return m_head->m_value;
}

template<typename T, typename U>
ALWAYS_INLINE T& ListHashSet<T, U>::last()
{
	ASSERT(!isEmpty());
	return m_tail->m_value;
}

template<typename T, typename U>
ALWAYS_INLINE const T& ListHashSet<T, U>::last() const
{
	ASSERT(!isEmpty());
	return m_tail->m_value;
}

template<typename T, typename U>
inline void ListHashSet<T, U>::removeLast()
{
    takeLast();
}

template<typename T, typename U>
inline T ListHashSet<T, U>::takeLast()
{
    ASSERT(!isEmpty());
    auto it = m_impl.find(m_tail);

    T result = WTFMove((*it)->m_value);
    m_impl.remove(it);
    unlinkAndDelete(m_tail);

    return result;
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::iterator ListHashSet<T, U>::find(const ValueType& value)
{
	ImplTypeIterator it = m_impl.template find<BaseTranslator>(value);
	if (it == m_impl.end())
		return end();
	return makeIterator(*it);
}

template<typename T, typename U>
ALWAYS_INLINE typename ListHashSet<T, U>::const_iterator ListHashSet<T, U>::find(const ValueType& value) const
{
	ImplTypeConstIterator it = m_impl.template find<BaseTranslator>(value);
	if (it == m_impl.end())
		return end();
	return makeConstIterator(*it);
}

template<typename Translator>
struct ListHashSetTranslatorAdapter {
	template<typename T> static ALWAYS_INLINE unsigned hash(const T& key) { return Translator::hash(key); }
	template<typename T, typename U> static ALWAYS_INLINE bool equal(const T& a, const U& b) { return Translator::equal(a->m_value, b); }
};

template<typename ValueType, typename U>
template<typename T, typename HashTranslator>
ALWAYS_INLINE typename ListHashSet<ValueType, U>::iterator ListHashSet<ValueType, U>::find(const T& value)
{
	ImplTypeConstIterator it = m_impl.template find<ListHashSetTranslatorAdapter<HashTranslator> >(value);
	if (it == m_impl.end())
		return end();
	return makeIterator(*it);
}

template<typename ValueType, typename U>
template<typename T, typename HashTranslator>
ALWAYS_INLINE typename ListHashSet<ValueType, U>::const_iterator ListHashSet<ValueType, U>::find(const T& value) const
{
	ImplTypeConstIterator it = m_impl.template find<ListHashSetTranslatorAdapter<HashTranslator> >(value);
	if (it == m_impl.end())
		return end();
	return makeConstIterator(*it);
}

template<typename ValueType, typename U>
template<typename T, typename HashTranslator>
ALWAYS_INLINE bool ListHashSet<ValueType, U>::contains(const T& value) const
{
	return m_impl.template contains<ListHashSetTranslatorAdapter<HashTranslator> >(value);
}

template<typename T, typename U>
ALWAYS_INLINE bool ListHashSet<T, U>::contains(const ValueType& value) const
{
	return m_impl.template contains<BaseTranslator>(value);
}

template<typename T, typename U>
typename ListHashSet<T, U>::AddResult ListHashSet<T, U>::add(const ValueType &value)
{
	typename ImplType::AddResult result = m_impl.template add<BaseTranslator>(value, 0);
	if (result.isNewEntry)
		appendNode(*result.iterator);
	return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
auto ListHashSet<T, U>::add(ValueType&& value) -> AddResult
{
    auto result = m_impl.template add<BaseTranslator>(WTFMove(value), 0);
    if (result.isNewEntry)
        appendNode(*result.iterator);
    return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
typename ListHashSet<T, U>::AddResult ListHashSet<T, U>::appendOrMoveToLast(const ValueType &value)
{
    typename ImplType::AddResult result = m_impl.template add<BaseTranslator>(value, 0);
    Node* node = *result.iterator;
    if (!result.isNewEntry)
        unlink(node);
    appendNode(node);
    return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
auto ListHashSet<T, U>::appendOrMoveToLast(ValueType&& value) -> AddResult
{
    auto result = m_impl.template add<BaseTranslator>(WTFMove(value), 0);
    Node* node = *result.iterator;
    if (!result.isNewEntry)
        unlink(node);
    appendNode(node);

    return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
typename ListHashSet<T, U>::AddResult ListHashSet<T, U>::prependOrMoveToFirst(const ValueType &value)
{
    typename ImplType::AddResult result = m_impl.template add<BaseTranslator>(value, 0);
    Node* node = *result.iterator;
    if (!result.isNewEntry)
        unlink(node);
    prependNode(node);
    return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
auto ListHashSet<T, U>::prependOrMoveToFirst(ValueType&& value) -> AddResult
{
    auto result = m_impl.template add<BaseTranslator>(WTFMove(value), 0);
    Node* node = *result.iterator;
    if (!result.isNewEntry)
        unlink(node);
    prependNode(node);

    return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
auto ListHashSet<T, U>::insertBefore(const ValueType& beforeValue, const ValueType& newValue) -> AddResult
{
    return insertBefore(find(beforeValue), newValue);
}

template<typename T, typename U>
auto ListHashSet<T, U>::insertBefore(const ValueType& beforeValue, ValueType&& newValue) -> AddResult
{
    return insertBefore(find(beforeValue), WTFMove(newValue));
}

template<typename T, typename U>
typename ListHashSet<T, U>::AddResult ListHashSet<T, U>::insertBefore(iterator it, const ValueType& newValue)
{
	typename ImplType::AddResult result = m_impl.template add<BaseTranslator>(newValue, 0);
	if (result.isNewEntry)
		insertNodeBefore(it.node(), *result.iterator);
	return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
auto ListHashSet<T, U>::insertBefore(iterator it, ValueType&& newValue) -> AddResult
{
    auto result = m_impl.template add<BaseTranslator>(WTFMove(newValue), 0);
    if (result.isNewEntry)
        insertNodeBefore(it.node(), *result.iterator);
    return AddResult(makeIterator(*result.iterator), result.isNewEntry);
}

template<typename T, typename U>
inline bool ListHashSet<T, U>::remove(iterator it)
{
	if (it == end())
		return false;
	m_impl.remove(it.node());
	unlinkAndDelete(it.node());
	return true;
}

template<typename T, typename U>
ALWAYS_INLINE bool ListHashSet<T, U>::remove(const ValueType& value)
{
	return remove(find(value));
}

template<typename T, typename U>
inline void ListHashSet<T, U>::clear()
{
	deleteAllNodes();
	m_impl.clear();
	m_head = 0;
	m_tail = 0;
}

template<typename T, typename U>
void ListHashSet<T, U>::unlink(Node* node)
{
	if (!node->m_prev) {
		ASSERT(node == m_head);
		m_head = node->m_next;
	} else {
		ASSERT(node != m_head);
		node->m_prev->m_next = node->m_next;
	}

	if (!node->m_next) {
		ASSERT(node == m_tail);
		m_tail = node->m_prev;
	} else {
		ASSERT(node != m_tail);
		node->m_next->m_prev = node->m_prev;
	}
}

template<typename T, typename U>
void ListHashSet<T, U>::unlinkAndDelete(Node* node)
{
    unlink(node);
    delete node;
}

template<typename T, typename U>
void ListHashSet<T, U>::appendNode(Node* node)
{
	node->m_prev = m_tail;
	node->m_next = 0;

	if (m_tail) {
		ASSERT(m_head);
		m_tail->m_next = node;
	} else {
		ASSERT(!m_head);
		m_head = node;
	}

	m_tail = node;
}

template<typename T, typename U>
void ListHashSet<T, U>::prependNode(Node* node)
{
    node->m_prev = 0;
    node->m_next = m_head;

    if (m_head)
        m_head->m_prev = node;
    else
        m_tail = node;

    m_head = node;
}

template<typename T, typename U>
void ListHashSet<T, U>::insertNodeBefore(Node* beforeNode, Node* newNode)
{
	if (!beforeNode)
		return appendNode(newNode);

	newNode->m_next = beforeNode;
	newNode->m_prev = beforeNode->m_prev;
	if (beforeNode->m_prev)
		beforeNode->m_prev->m_next = newNode;
	beforeNode->m_prev = newNode;

	if (!newNode->m_prev)
		m_head = newNode;
}

template<typename T, typename U>
void ListHashSet<T, U>::deleteAllNodes()
{
	if (!m_head)
		return;

	for (Node* node = m_head, *next = m_head->m_next; node; node = next, next = node ? node->m_next : 0)
	    delete node;
}

template<typename T, typename U>
ALWAYS_INLINE ListHashSetReverseIterator<T, U> ListHashSet<T, U>::makeReverseIterator(Node* position)
{
	return ListHashSetReverseIterator<T, U>(this, position);
}

template<typename T, typename U>
ALWAYS_INLINE ListHashSetConstReverseIterator<T, U> ListHashSet<T, U>::makeConstReverseIterator(Node* position) const
{
	return ListHashSetConstReverseIterator<T, U>(this, position);
}

template<typename T, typename U>
ALWAYS_INLINE ListHashSetIterator<T, U> ListHashSet<T, U>::makeIterator(Node* position)
{
	return ListHashSetIterator<T, U>(this, position);
}

template<typename T, typename U>
ALWAYS_INLINE ListHashSetConstIterator<T, U> ListHashSet<T, U>::makeConstIterator(Node* position) const
{
	return ListHashSetConstIterator<T, U>(this, position);
}

} // namespace WTF

using WTF::ListHashSet;
