/*
 * Copyright (C) 2010 Espial Group Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Espial Group Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <atomic>
#include <stdlib.h>

namespace JSC
{
class ExecState;
class JSGlobalObject;
}

namespace WTF
{

extern std::atomic<unsigned> gLowMemory;
extern std::atomic<unsigned> gUnderMemoryPressure;

ALWAYS_INLINE bool isLowMemory()
{
	return gLowMemory.load(std::memory_order_relaxed);
}

ALWAYS_INLINE bool isUnderMemoryPressure()
{
	return gUnderMemoryPressure.load(std::memory_order_relaxed);
}

void setLowMemory(bool low);

unsigned underMemoryPressure();

class MmapMonitorBase
{
public:
	virtual ~MmapMonitorBase() = default;

	static MmapMonitorBase* instance();

	static void setInstance(MmapMonitorBase*);

	virtual void mmapStat(void* p, unsigned size) = 0;
	virtual void munmapStat(void* p, unsigned size) = 0;
	
	ALWAYS_INLINE unsigned allocatedSize() const { return m_allocatedBytes.load(std::memory_order_relaxed); }

	ALWAYS_INLINE unsigned capSize() const { return m_capBytes; }

	ALWAYS_INLINE unsigned peakSize() const { return m_peakBytes; }
	
	ALWAYS_INLINE bool canGrow(JSC::ExecState* exec, JSC::JSGlobalObject* jsGlobalObject, unsigned size)
	{
		unsigned allocatedSize = this->allocatedSize();
		return !m_capBytes ||
			(allocatedSize + size) <= m_capBytes ||
			jsMmapChecker(exec, jsGlobalObject, allocatedSize);
	}

	ALWAYS_INLINE bool canGrowFastCheck(unsigned size) const
	{
		return !m_capBytes || (allocatedSize() + size) <= m_capBytes;
	}

	ALWAYS_INLINE void setCapBytes(unsigned capBytes)
	{
		m_capBytes = capBytes;
	}
	
	ALWAYS_INLINE void resetPeakSize() { m_peakBytes = 0; }

	virtual bool shouldDeferGC(size_t& bytesAllocatedLimit) = 0;
	
	virtual void didGC(double lastGCLength, int codeBlocks, size_t sizeAfterLastCollect, size_t bytesAllocatedLimit, int opaqueRootCount) = 0;
	virtual void didDeleteAllCodeBlocks(bool all, int codeBlocks, int deletedBytes, int targetDeleteBytes) = 0;

	virtual bool jsMmapChecker(JSC::ExecState*, JSC::JSGlobalObject* jsGlobalObject, unsigned allocatedSize) = 0;

protected:
	MmapMonitorBase()
	: m_allocatedBytes(0)
	, m_peakBytes(0)
	, m_capBytes(0)
	{
	}

	std::atomic<unsigned> m_allocatedBytes;
	unsigned m_peakBytes;
	unsigned m_capBytes;
};

}

