/*
 *  Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include <wtf/GetPtr.h>
#include <wtf/Ref.h>
#include <wtf/AlwaysInline.h>
#include <wtf/NullPtr.h>

namespace WTF {

template<typename T> class RefPtr;
template<typename T> class PassRefPtr;
template<typename T> ALWAYS_INLINE PassRefPtr<T> adoptRef(T*);

#if !(PLATFORM(QT) && CPU(ARM))
#define REF_DEREF_INLINE ALWAYS_INLINE
#else
// Older version of gcc used by Harmattan SDK fails to build with ALWAYS_INLINE.
// See https://bugs.webkit.org/show_bug.cgi?id=37253 for details.
#define REF_DEREF_INLINE inline
#endif

template<typename T> ALWAYS_INLINE void refIfNotNull(T* ptr)
{
	if (LIKELY(ptr != 0))
		ptr->ref();
}

template<typename T> ALWAYS_INLINE void derefIfNotNull(T* ptr)
{
	if (LIKELY(ptr != 0))
		ptr->deref();
}

#undef REF_DEREF_INLINE

template<typename T> class PassRefPtr {
public:
	typedef T ValueType;
	typedef ValueType* PtrType;

	ALWAYS_INLINE PassRefPtr() : m_ptr(0) { }
	ALWAYS_INLINE PassRefPtr(std::nullptr_t) : m_ptr(0) { }
	ALWAYS_INLINE PassRefPtr(T& object) : m_ptr(&object) { object.ref(); }
	ALWAYS_INLINE PassRefPtr(T* ptr) : m_ptr(ptr) { refIfNotNull(ptr); }
	// It somewhat breaks the type system to allow transfer of ownership out of
	// a const PassRefPtr. However, it makes it much easier to work with PassRefPtr
	// temporaries, and we don't have a need to use real const PassRefPtrs anyway.
	ALWAYS_INLINE PassRefPtr(const PassRefPtr& o) : m_ptr(o.leakRef()) { }
	template<typename U> ALWAYS_INLINE PassRefPtr(const PassRefPtr<U>& o) : m_ptr(o.leakRef()) { }

	ALWAYS_INLINE PassRefPtr(PassRefPtr&& o) : m_ptr(o.leakRef()) { }
	template<typename U> PassRefPtr(PassRefPtr<U>&& o) : m_ptr(o.leakRef()) { }

	ALWAYS_INLINE ~PassRefPtr() { clear(); }

	template<typename U> PassRefPtr(const RefPtr<U>&);
	template<typename U> PassRefPtr(RefPtr<U>&& o) : m_ptr(o.leakRef()) { }

	template<typename U> PassRefPtr(Ref<U>&& reference) : m_ptr(&reference.leakRef()) { }

	ALWAYS_INLINE void clear()
	{
		T* ptr = m_ptr;
		m_ptr = 0;
		derefIfNotNull(ptr);
	}

	ALWAYS_INLINE Ref<T> releaseNonNull() { Ref<T> tmp = adoptRef(*m_ptr); m_ptr = nullptr; return tmp; }

	ALWAYS_INLINE T* get() const { return m_ptr; }

	ALWAYS_INLINE T* leakRef() const WARN_UNUSED_RETURN;

	ALWAYS_INLINE T& operator*() const { return *m_ptr; }
	ALWAYS_INLINE T* operator->() const { return m_ptr; }

	ALWAYS_INLINE bool operator!() const { return !m_ptr; }

	ALWAYS_INLINE explicit operator bool() const { return !!m_ptr; }

	ALWAYS_INLINE PassRefPtr& operator=(const PassRefPtr&) { COMPILE_ASSERT(!sizeof(T*), PassRefPtr_should_never_be_assigned_to); return *this; }

	friend PassRefPtr adoptRef<T>(T*);

private:
	// adopting constructor
	enum AdoptTag { Adopt };
	PassRefPtr(T* ptr, AdoptTag) : m_ptr(ptr) { }

	mutable T* m_ptr;
};

template<typename T> template<typename U> ALWAYS_INLINE PassRefPtr<T>::PassRefPtr(const RefPtr<U>& o)
	: m_ptr(o.get())
{
	T* ptr = m_ptr;
	refIfNotNull(ptr);
}

template<typename T> ALWAYS_INLINE T* PassRefPtr<T>::leakRef() const
{
	T* ptr = m_ptr;
	m_ptr = 0;
	return ptr;
}

template<typename T, typename U> ALWAYS_INLINE bool operator==(const PassRefPtr<T>& a, const PassRefPtr<U>& b)
{
	return a.get() == b.get();
}

template<typename T, typename U> ALWAYS_INLINE bool operator==(const PassRefPtr<T>& a, const RefPtr<U>& b)
{
	return a.get() == b.get();
}

template<typename T, typename U> ALWAYS_INLINE bool operator==(const RefPtr<T>& a, const PassRefPtr<U>& b)
{
	return a.get() == b.get();
}

template<typename T, typename U> ALWAYS_INLINE bool operator==(const PassRefPtr<T>& a, U* b)
{
	return a.get() == b;
}

template<typename T> ALWAYS_INLINE bool operator==(const PassRefPtr<T>& a, std::nullptr_t)
{
	return a.get() == 0;
}

template<typename T, typename U> ALWAYS_INLINE bool operator==(T* a, const PassRefPtr<U>& b)
{
	return a == b.get();
}

template<typename T, typename U> ALWAYS_INLINE bool operator!=(const PassRefPtr<T>& a, const PassRefPtr<U>& b)
{
	return a.get() != b.get();
}

template<typename T, typename U> ALWAYS_INLINE bool operator!=(const PassRefPtr<T>& a, const RefPtr<U>& b)
{
	return a.get() != b.get();
}

template<typename T, typename U> ALWAYS_INLINE bool operator!=(const RefPtr<T>& a, const PassRefPtr<U>& b)
{
	return a.get() != b.get();
}

template<typename T, typename U> ALWAYS_INLINE bool operator!=(const PassRefPtr<T>& a, U* b)
{
	return a.get() != b;
}

template<typename T> ALWAYS_INLINE bool operator!=(const PassRefPtr<T>& a, std::nullptr_t)
{
	return a.get() != 0;
}

template<typename T, typename U> ALWAYS_INLINE bool operator!=(T* a, const PassRefPtr<U>& b)
{
	return a != b.get();
}

template<typename T> ALWAYS_INLINE PassRefPtr<T> adoptRef(T* p)
{
	adopted(p);
	return PassRefPtr<T>(p, PassRefPtr<T>::Adopt);
}

template<typename T, typename U> ALWAYS_INLINE PassRefPtr<T> static_pointer_cast(const PassRefPtr<U>& p)
{
	return adoptRef(static_cast<T*>(p.leakRef()));
}

template<typename T, typename U> ALWAYS_INLINE PassRefPtr<T> const_pointer_cast(const PassRefPtr<U>& p)
{
	return adoptRef(const_cast<T*>(p.leakRef()));
}

template <typename T> struct IsSmartPtr<PassRefPtr<T>> {
	static const bool value = true;
};

} // namespace WTF

using WTF::PassRefPtr;
using WTF::adoptRef;
using WTF::static_pointer_cast;
using WTF::const_pointer_cast;
