/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include <wtf/Assertions.h>
#include <wtf/GetPtr.h>
#include <wtf/Noncopyable.h>
#include <wtf/StdLibExtras.h>
#include <algorithm>
#include <utility>

namespace WTF {

inline void adopted(const void*) { }

template<typename T> class Ref;
template<typename T> Ref<T> adoptRef(T&);

template<typename T> class Ref {
public:
    ALWAYS_INLINE ~Ref()
    {
        if (m_ptr) {
        	m_ptr->deref();
        	m_ptr = nullptr;
        }
    }

    ALWAYS_INLINE Ref(T& object)
        : m_ptr(&object)
    {
        m_ptr->ref();
    }

    ALWAYS_INLINE Ref(const Ref& other)
        : m_ptr(other.ptr())
    {
        m_ptr->ref();
    }

    template<typename X> ALWAYS_INLINE Ref(const Ref<X>& other)
        : m_ptr(other.ptr())
    {
        m_ptr->ref();
    }

    ALWAYS_INLINE Ref(Ref&& other)
        : m_ptr(&other.leakRef())
    {
        ASSERT(m_ptr);
    }

    template<typename U>
    ALWAYS_INLINE Ref(Ref<U>&& other)
        : m_ptr(&other.leakRef())
    {
        ASSERT(m_ptr);
    }

    ALWAYS_INLINE Ref& operator=(T&);
    ALWAYS_INLINE Ref& operator=(Ref&&);
    template<typename U> ALWAYS_INLINE Ref& operator=(Ref<U>&&);

    ALWAYS_INLINE Ref& operator=(const Ref&);
    template<typename X> ALWAYS_INLINE Ref& operator=(const Ref<X>&);

    ALWAYS_INLINE void swap(Ref&);

    // Hash table deleted/empty values, which are only constructed and never copied or destroyed.
    ALWAYS_INLINE Ref(HashTableDeletedValueType) : m_ptr(hashTableDeletedValue()) { }
    ALWAYS_INLINE bool isHashTableDeletedValue() const { return m_ptr == hashTableDeletedValue(); }
    static ALWAYS_INLINE T* hashTableDeletedValue() { return reinterpret_cast<T*>(-1); }

    ALWAYS_INLINE Ref(HashTableEmptyValueType) : m_ptr(hashTableEmptyValue()) { }
    ALWAYS_INLINE bool isHashTableEmptyValue() const { return m_ptr == hashTableEmptyValue(); }
    static ALWAYS_INLINE T* hashTableEmptyValue() { return nullptr; }

    const T* ptrAllowingHashTableEmptyValue() const { ASSERT(m_ptr || isHashTableEmptyValue()); return m_ptr; }
    T* ptrAllowingHashTableEmptyValue() { ASSERT(m_ptr || isHashTableEmptyValue()); return m_ptr; }

    ALWAYS_INLINE void assignToHashTableEmptyValue(Ref&& reference)
    {
        ASSERT(m_ptr == hashTableEmptyValue());
        m_ptr = &reference.leakRef();
        ASSERT(m_ptr);
    }

    ALWAYS_INLINE T* operator->() const { ASSERT(m_ptr); return m_ptr; }
    ALWAYS_INLINE T* ptr() const { ASSERT(m_ptr); return m_ptr; }
    ALWAYS_INLINE T& get() const { ASSERT(m_ptr); return *m_ptr; }
    ALWAYS_INLINE operator T&() const { ASSERT(m_ptr); return *m_ptr; }
    ALWAYS_INLINE bool operator!() const { ASSERT(m_ptr); return !*m_ptr; }

    template<typename U> ALWAYS_INLINE Ref<T> replace(Ref<U>&&) WARN_UNUSED_RETURN;

    ALWAYS_INLINE Ref copyRef() const WARN_UNUSED_RETURN
    {
        return Ref(*m_ptr);
    }

    ALWAYS_INLINE T& leakRef() WARN_UNUSED_RETURN
    {
        ASSERT(m_ptr);
        T* movedPtr = m_ptr;
        m_ptr = nullptr;
        return *movedPtr;
    }

private:
    friend Ref adoptRef<T>(T&);

    enum AdoptTag { Adopt };
    ALWAYS_INLINE Ref(T& object, AdoptTag)
        : m_ptr(&object)
    {
    }

    T* m_ptr;
};

template<typename T> void swap(Ref<T>&, Ref<T>&);

template<typename T> ALWAYS_INLINE Ref<T>& Ref<T>::operator=(T& reference)
{
    Ref copiedReference = reference;
    swap(copiedReference);
    return *this;
}

template<typename T> ALWAYS_INLINE Ref<T>& Ref<T>::operator=(Ref&& reference)
{
    Ref movedReference = WTFMove(reference);
    swap(movedReference);
    return *this;
}

template<typename T> template<typename U> ALWAYS_INLINE Ref<T>& Ref<T>::operator=(Ref<U>&& reference)
{
    Ref movedReference = WTFMove(reference);
    swap(movedReference);
    return *this;
}

template<typename T>
ALWAYS_INLINE Ref<T>& Ref<T>::operator=(const Ref& reference)
{
#if ASAN_ENABLED
    if (__asan_address_is_poisoned(this))
        __asan_unpoison_memory_region(this, sizeof(*this));
#endif
    Ref copiedReference = reference;
    swap(copiedReference);
    return *this;
}

template<typename T>
template<typename X>
ALWAYS_INLINE Ref<T>& Ref<T>::operator=(const Ref<X>& reference)
{
#if ASAN_ENABLED
    if (__asan_address_is_poisoned(this))
        __asan_unpoison_memory_region(this, sizeof(*this));
#endif
    Ref copiedReference = reference;
    swap(copiedReference);
    return *this;
}

template<typename T> ALWAYS_INLINE void Ref<T>::swap(Ref& other)
{
    std::swap(m_ptr, other.m_ptr);
}

template<typename T> ALWAYS_INLINE void swap(Ref<T>& a, Ref<T>& b)
{
    a.swap(b);
}

template<typename T> template<typename U> ALWAYS_INLINE Ref<T> Ref<T>::replace(Ref<U>&& reference)
{
    auto oldReference = adoptRef(*m_ptr);
    m_ptr = &reference.leakRef();
    return oldReference;
}

template<typename T>
ALWAYS_INLINE Ref<T> adoptRef(T& reference)
{
    adopted(&reference);
    return Ref<T>(reference, Ref<T>::Adopt);

}

template <typename T>
struct IsSmartPtr<Ref<T>> {
    static const bool value = true;
};

template <typename T>
struct GetPtrHelper<Ref<T>> {
    typedef T* PtrType;
    static T* getPtr(const Ref<T>& p) { return const_cast<T*>(p.ptr()); }
};

template<typename T, typename V>
ALWAYS_INLINE bool operator==(const Ref<T>& a, const Ref<V>& b)
{
    return a.ptr() == b.ptr();
}

template<typename T, typename V>
ALWAYS_INLINE bool operator==(const Ref<T>& a, V& b)
{
    return a.ptr() == &b;
}

template<typename T, typename U>
ALWAYS_INLINE bool operator==(T& a, const Ref<U>& b)
{
    return &a == b.ptr();
}

template<typename T, typename V>
ALWAYS_INLINE bool operator!=(const Ref<T>& a, const Ref<V>& b)
{
    return a.ptr() != b.ptr();
}

template<typename T, typename V>
ALWAYS_INLINE bool operator!=(const Ref<T>& a, V& b)
{
    return a.ptr() != &b;
}

template<typename T, typename U>
ALWAYS_INLINE bool operator!=(T& a, const Ref<U>& b)
{
    return &a != b.ptr();
}

template<typename T, typename X>
inline Ref<T> static_reference_cast(Ref<X>&& reference)
{
    return adoptRef(static_cast<T&>(reference.leakRef()));
}

template<typename T, typename X>
ALWAYS_INLINE Ref<T> static_reference_cast(const Ref<X>& reference)
{
    return static_reference_cast<T>(reference.copyRef());
}

} // namespace WTF

using WTF::Ref;
using WTF::static_reference_cast;
