/*
 *  Copyright (C) 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include <wtf/Alignment.h>
#include <wtf/CheckedArithmetic.h>
#include <wtf/FastAllocBase.h>
#include <wtf/FastMallocMonitor.h>
#include <wtf/Forward.h>
#include <wtf/Noncopyable.h>
#include <wtf/NotFound.h>
#include <wtf/StdLibExtras.h>
#include <wtf/UnusedParam.h>
#include <wtf/ValueCheck.h>
#include <wtf/VectorTraits.h>
#include <limits>
#include <utility>

namespace WTF {

template <bool needsDestruction, typename T>
struct VectorDestructor;

template<typename T>
struct VectorDestructor<false, T>
{
	static void destruct(T*, T*) {}
};

template<typename T>
struct VectorDestructor<true, T>
{
	static void destruct(T* begin, T* end)
	{
		for (T* cur = begin; cur != end; ++cur)
			cur->~T();
	}
};

template <bool needsInitialization, bool canInitializeWithMemset, typename T>
struct VectorInitializer;

template<bool ignore, typename T>
struct VectorInitializer<false, ignore, T>
{
	static void initialize(T*, T*) {}
};

template<typename T>
struct VectorInitializer<true, false, T>
{
	static void initialize(T* begin, T* end)
	{
		for (T* cur = begin; cur != end; ++cur)
			new (NotNull, cur) T;
	}
};

template<typename T>
struct VectorInitializer<true, true, T>
{
	static ALWAYS_INLINE void initialize(T* begin, T* end)
	{
		memset(begin, 0, reinterpret_cast<char*>(end) - reinterpret_cast<char*>(begin));
	}
};

template <bool canMoveWithMemcpy, typename T>
struct VectorMover;

template<typename T>
struct VectorMover<false, T>
{

    static void move(T* src, T* srcEnd, T* dst)
    {
        while (src != srcEnd) {
            new (NotNull, dst) T(WTFMove(*src));
            src->~T();
            ++dst;
            ++src;
        }
    }
    static void moveOverlapping(T* src, T* srcEnd, T* dst)
    {
        if (src > dst)
            move(src, srcEnd, dst);
        else {
            T* dstEnd = dst + (srcEnd - src);
            while (src != srcEnd) {
                --srcEnd;
                --dstEnd;
                new (NotNull, dstEnd) T(WTFMove(*srcEnd));
                srcEnd->~T();
            }
        }
    }
};

template<typename T>
struct VectorMover<true, T>
{
	static ALWAYS_INLINE void move(T* src, T* srcEnd, T* dst)
	{
		memcpy(dst, src, reinterpret_cast<char*>(srcEnd) - reinterpret_cast<char*>(src));
	}
	static ALWAYS_INLINE void moveOverlapping(T* src, T* srcEnd, T* dst)
	{
		memmove(dst, src, reinterpret_cast<char*>(srcEnd) - reinterpret_cast<char*>(src));
	}
};

template <bool canCopyWithMemcpy, typename T>
struct VectorCopier;

template<typename T>
struct VectorCopier<false, T>
{
	template<typename U>
	static void uninitializedCopy(const T* src, const T* srcEnd, U* dst)
	{
		while (src != srcEnd) {
			new (NotNull, dst) U(*src);
			++dst;
			++src;
		}
	}
};

template<typename T>
struct VectorCopier<true, T>
{
	static ALWAYS_INLINE void uninitializedCopy(const T* src, const T* srcEnd, T* dst)
	{
		memcpy(dst, src, reinterpret_cast<const char*>(srcEnd) - reinterpret_cast<const char*>(src));
	}
	template<typename U>
	static ALWAYS_INLINE void uninitializedCopy(const T* src, const T* srcEnd, U* dst)
	{
		VectorCopier<false, T>::uninitializedCopy(src, srcEnd, dst);
	}
};

template <bool canFillWithMemset, typename T>
struct VectorFiller;

template<typename T>
struct VectorFiller<false, T>
{
	static void uninitializedFill(T* dst, T* dstEnd, const T& val)
	{
		while (dst != dstEnd) {
			new (NotNull, dst) T(val);
			++dst;
		}
	}
};

template<typename T>
struct VectorFiller<true, T>
{
	static ALWAYS_INLINE void uninitializedFill(T* dst, T* dstEnd, const T& val)
	{
		ASSERT(sizeof(T) == sizeof(char));
		memset(dst, val, dstEnd - dst);
	}
};

template<bool canCompareWithMemcmp, typename T>
struct VectorComparer;

template<typename T>
struct VectorComparer<false, T>
{
	static bool compare(const T* a, const T* b, size_t size)
	{
		for (size_t i = 0; i < size; ++i)
			if (!(a[i] == b[i]))
				return false;
		return true;
	}
};

template<typename T>
struct VectorComparer<true, T>
{
	static ALWAYS_INLINE bool compare(const T* a, const T* b, size_t size)
	{
		return memcmp(a, b, sizeof(T) * size) == 0;
	}
};

template<typename T>
struct VectorTypeOperations
{
	static ALWAYS_INLINE void destruct(T* begin, T* end)
	{
		VectorDestructor<VectorTraits<T>::needsDestruction, T>::destruct(begin, end);
	}

	static ALWAYS_INLINE void initialize(T* begin, T* end)
	{
		VectorInitializer<VectorTraits<T>::needsInitialization, VectorTraits<T>::canInitializeWithMemset, T>::initialize(begin, end);
	}

	static ALWAYS_INLINE void move(T* src, T* srcEnd, T* dst)
	{
		VectorMover<VectorTraits<T>::canMoveWithMemcpy, T>::move(src, srcEnd, dst);
	}

	static ALWAYS_INLINE void moveOverlapping(T* src, T* srcEnd, T* dst)
	{
		VectorMover<VectorTraits<T>::canMoveWithMemcpy, T>::moveOverlapping(src, srcEnd, dst);
	}

	static ALWAYS_INLINE void uninitializedCopy(const T* src, const T* srcEnd, T* dst)
	{
		VectorCopier<VectorTraits<T>::canCopyWithMemcpy, T>::uninitializedCopy(src, srcEnd, dst);
	}

	static ALWAYS_INLINE void uninitializedFill(T* dst, T* dstEnd, const T& val)
	{
		VectorFiller<VectorTraits<T>::canFillWithMemset, T>::uninitializedFill(dst, dstEnd, val);
	}

	static ALWAYS_INLINE bool compare(const T* a, const T* b, size_t size)
	{
		return VectorComparer<VectorTraits<T>::canCompareWithMemcmp, T>::compare(a, b, size);
	}
};

template<typename T>
class VectorBufferBase {
	WTF_MAKE_NONCOPYABLE(VectorBufferBase);
public:
	ALWAYS_INLINE void allocateBuffer(size_t newCapacity)
	{
		ASSERT(newCapacity);
		//if (UNLIKELY(newCapacity > std::numeric_limits<unsigned>::max() / sizeof(T)))
		//    CRASH();
		size_t sizeToAllocate = fastMallocGoodSize(newCapacity * sizeof(T));
		m_capacity = sizeToAllocate / sizeof(T);
		m_buffer = static_cast<T*>(fastMalloc(sizeToAllocate));
	}

	ALWAYS_INLINE bool tryAllocateBuffer(size_t newCapacity)
	{
		ASSERT(newCapacity);
		//if (UNLIKELY(newCapacity > std::numeric_limits<unsigned>::max() / sizeof(T)))
		//    return false;

		size_t sizeToAllocate = fastMallocGoodSize(newCapacity * sizeof(T));
		T* newBuffer;
		if (tryFastMalloc(sizeToAllocate).getValue(newBuffer)) {
			m_capacity = sizeToAllocate / sizeof(T);
			m_buffer = newBuffer;
			return true;
		}
		return false;
	}

	bool shouldReallocateBuffer(size_t newCapacity) const
	{
		return VectorTraits<T>::canMoveWithMemcpy && m_capacity && newCapacity;
	}

	void reallocateBuffer(size_t newCapacity)
	{
		ASSERT(shouldReallocateBuffer(newCapacity));
		//if (newCapacity > std::numeric_limits<size_t>::max() / sizeof(T))
		//    CRASH();
		size_t sizeToAllocate = fastMallocGoodSize(newCapacity * sizeof(T));
		m_capacity = sizeToAllocate / sizeof(T);
		m_buffer = static_cast<T*>(fastRealloc(m_buffer, sizeToAllocate));
	}

	ALWAYS_INLINE void deallocateBuffer(T* bufferToDeallocate)
	{
		if (!bufferToDeallocate)
			return;

		if (m_buffer == bufferToDeallocate) {
			m_buffer = 0;
			m_capacity = 0;
		}

		fastFree(bufferToDeallocate);
	}

	T* buffer() { return m_buffer; }
	const T* buffer() const { return m_buffer; }
	static ptrdiff_t bufferMemoryOffset() { return OBJECT_OFFSETOF(VectorBufferBase, m_buffer); }
	size_t capacity() const { return m_capacity; }

	ALWAYS_INLINE T* releaseBuffer()
	{
		T* buffer = m_buffer;
		m_buffer = 0;
		m_capacity = 0;
		return buffer;
	}

protected:
	ALWAYS_INLINE VectorBufferBase()
		: m_buffer(0)
		, m_capacity(0)
		, m_size(0)
	{
	}

	ALWAYS_INLINE VectorBufferBase(T* buffer, size_t capacity, size_t size)
		: m_buffer(buffer)
		, m_capacity(capacity)
		, m_size(size)
	{
	}

	ALWAYS_INLINE ~VectorBufferBase()
	{
		// FIXME: It would be nice to find a way to ASSERT that m_buffer hasn't leaked here.
	}

	T* m_buffer;
	unsigned m_capacity;
	unsigned m_size; // Only used by the Vector subclass, but placed here to avoid padding the struct.
};

template<typename T, size_t inlineCapacity>
class VectorBuffer;

template<typename T>
class VectorBuffer<T, 0> : private VectorBufferBase<T> {
private:
	typedef VectorBufferBase<T> Base;
public:
	ALWAYS_INLINE VectorBuffer()
	{
	}

	ALWAYS_INLINE VectorBuffer(size_t capacity, size_t size = 0)
	{
		m_size = size;
		// Calling malloc(0) might take a lock and may actually do an
		// allocation on some systems.
		if (capacity)
			allocateBuffer(capacity);
	}

	ALWAYS_INLINE ~VectorBuffer()
	{
		deallocateBuffer(buffer());
	}

	void swap(VectorBuffer<T, 0>& other, size_t, size_t)
	{
		std::swap(m_buffer, other.m_buffer);
		std::swap(m_capacity, other.m_capacity);
	}

	ALWAYS_INLINE void restoreInlineBufferIfNeeded() { }

	using Base::allocateBuffer;
	using Base::tryAllocateBuffer;
	using Base::shouldReallocateBuffer;
	using Base::reallocateBuffer;
	using Base::deallocateBuffer;

	using Base::buffer;
	using Base::capacity;
	using Base::bufferMemoryOffset;

	using Base::releaseBuffer;

protected:
	using Base::m_size;

private:
	using Base::m_buffer;
	using Base::m_capacity;
};

template<typename T, size_t inlineCapacity>
class VectorBuffer : private VectorBufferBase<T> {
	WTF_MAKE_NONCOPYABLE(VectorBuffer);
private:
	typedef VectorBufferBase<T> Base;
public:
	ALWAYS_INLINE VectorBuffer()
		: Base(inlineBuffer(), inlineCapacity, 0)
	{
	}

	ALWAYS_INLINE VectorBuffer(size_t capacity, size_t size = 0)
		: Base(inlineBuffer(), inlineCapacity, size)
	{
		if (capacity > inlineCapacity)
			Base::allocateBuffer(capacity);
	}

	ALWAYS_INLINE ~VectorBuffer()
	{
		deallocateBuffer(buffer());
	}

	ALWAYS_INLINE void allocateBuffer(size_t newCapacity)
	{
		// FIXME: This should ASSERT(!m_buffer) to catch misuse/leaks.
		if (newCapacity > inlineCapacity)
			Base::allocateBuffer(newCapacity);
		else {
			m_buffer = inlineBuffer();
			m_capacity = inlineCapacity;
		}
	}

	bool tryAllocateBuffer(size_t newCapacity)
	{
		if (newCapacity > inlineCapacity)
			return Base::tryAllocateBuffer(newCapacity);
		m_buffer = inlineBuffer();
		m_capacity = inlineCapacity;
		return true;
	}

	ALWAYS_INLINE void deallocateBuffer(T* bufferToDeallocate)
	{
		if (bufferToDeallocate == inlineBuffer())
			return;
		Base::deallocateBuffer(bufferToDeallocate);
	}

	bool shouldReallocateBuffer(unsigned newCapacity) const
	{
		// We cannot reallocate the inline buffer.
		return Base::shouldReallocateBuffer(newCapacity) && std::min(m_capacity, newCapacity) > inlineCapacity;
	}

	void reallocateBuffer(size_t newCapacity)
	{
		ASSERT(shouldReallocateBuffer(newCapacity));
		Base::reallocateBuffer(newCapacity);
	}

	void swap(VectorBuffer& other, size_t mySize, size_t otherSize)
	{
		if (buffer() == inlineBuffer() && other.buffer() == other.inlineBuffer()) {
			swapInlineBuffer(other, mySize, otherSize);
			std::swap(m_capacity, other.m_capacity);
		} else if (buffer() == inlineBuffer()) {
			m_buffer = other.m_buffer;
			other.m_buffer = other.inlineBuffer();
			swapInlineBuffer(other, mySize, 0);
			std::swap(m_capacity, other.m_capacity);
		} else if (other.buffer() == other.inlineBuffer()) {
			other.m_buffer = m_buffer;
			m_buffer = inlineBuffer();
			swapInlineBuffer(other, 0, otherSize);
			std::swap(m_capacity, other.m_capacity);
		} else {
			std::swap(m_buffer, other.m_buffer);
			std::swap(m_capacity, other.m_capacity);
		}
	}

	ALWAYS_INLINE void restoreInlineBufferIfNeeded()
	{
		if (m_buffer)
			return;
		m_buffer = inlineBuffer();
		m_capacity = inlineCapacity;
	}

	using Base::buffer;
	using Base::capacity;
	using Base::bufferMemoryOffset;

	ALWAYS_INLINE T* releaseBuffer()
	{
		if (buffer() == inlineBuffer())
			return 0;
		return Base::releaseBuffer();
	}

protected:
	using Base::m_size;

private:
	using Base::m_buffer;
	using Base::m_capacity;

	void swapInlineBuffer(VectorBuffer& other, size_t mySize, size_t otherSize)
	{
		// FIXME: We could make swap part of VectorTypeOperations
		// https://bugs.webkit.org/show_bug.cgi?id=128863
		swapInlineBuffers(inlineBuffer(), other.inlineBuffer(), mySize, otherSize);
	}

	static void swapInlineBuffers(T* left, T* right, size_t leftSize, size_t rightSize)
	{
		if (left == right)
			return;

		ASSERT(leftSize <= inlineCapacity);
		ASSERT(rightSize <= inlineCapacity);

		size_t swapBound = std::min(leftSize, rightSize);
		for (unsigned i = 0; i < swapBound; ++i)
			std::swap(left[i], right[i]);
		VectorTypeOperations<T>::move(left + swapBound, left + leftSize, right + swapBound);
		VectorTypeOperations<T>::move(right + swapBound, right + rightSize, left + swapBound);
	}

	T* inlineBuffer() { return reinterpret_cast_ptr<T*>(m_inlineBuffer); }
	const T* inlineBuffer() const { return reinterpret_cast_ptr<const T*>(m_inlineBuffer); }

	typename std::aligned_storage<sizeof(T), std::alignment_of<T>::value>::type m_inlineBuffer[inlineCapacity];
};

struct UnsafeVectorOverflow {
    static NO_RETURN_DUE_TO_ASSERT void overflowed()
    {
        ASSERT_NOT_REACHED();
    }
};

enum class VectorConstructorFlag { OneItem };

// Template default values are in Forward.h.
template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
class Vector : private VectorBuffer<T, inlineCapacity> {
	WTF_MAKE_FAST_ALLOCATED;
private:
	typedef VectorBuffer<T, inlineCapacity> Base;
	typedef VectorTypeOperations<T> TypeOperations;

public:
	typedef T ValueType;

	typedef T* iterator;
	typedef const T* const_iterator;
	typedef std::reverse_iterator<iterator> reverse_iterator;
	typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

	ALWAYS_INLINE Vector()
	{
	}

	ALWAYS_INLINE explicit Vector(size_t size)
		: Base(size, size)
	{
		if (begin())
			TypeOperations::initialize(begin(), end());
	}

    Vector(WTF::HashTableDeletedValueType)
        : Base(0, std::numeric_limits<unsigned>::max())
    {
    }

	// can not always inline
	~Vector()
	{
		if (m_size)
			shrink(0);
	}

    Vector(std::initializer_list<T> initializerList)
    {
        reserveInitialCapacity(initializerList.size());

        //asanSetInitialBufferSizeTo(initializerList.size());

        for (const auto& element : initializerList)
            uncheckedAppend(element);
    }

    explicit Vector(const ValueType& arg, VectorConstructorFlag)
    {
    	reserveInitialCapacity(1);
    	uncheckedAppend(arg);
    }

    explicit Vector(const ValueType& arg1, const ValueType& arg2, VectorConstructorFlag)
    {
    	reserveInitialCapacity(2);
    	uncheckedAppend(arg1);
    	uncheckedAppend(arg2);
    }
    explicit Vector(const ValueType& arg1, const ValueType& arg2, const ValueType& arg3)
    {
    	reserveInitialCapacity(3);
    	uncheckedAppend(arg1);
    	uncheckedAppend(arg2);
    	uncheckedAppend(arg3);
    }
    explicit Vector(const ValueType& arg1, const ValueType& arg2, const ValueType& arg3, const ValueType& arg4)
    {
    	reserveInitialCapacity(4);
    	uncheckedAppend(arg1);
    	uncheckedAppend(arg2);
    	uncheckedAppend(arg3);
    	uncheckedAppend(arg4);
    }
    explicit Vector(const ValueType& arg1, const ValueType& arg2, const ValueType& arg3, const ValueType& arg4, const ValueType& arg5)
    {
    	reserveInitialCapacity(5);
    	uncheckedAppend(arg1);
    	uncheckedAppend(arg2);
    	uncheckedAppend(arg3);
    	uncheckedAppend(arg4);
    	uncheckedAppend(arg5);
    }
    explicit Vector(const ValueType& arg1, const ValueType& arg2, const ValueType& arg3, const ValueType& arg4, const ValueType& arg5, const ValueType& arg6)
    {
    	reserveInitialCapacity(6);
    	uncheckedAppend(arg1);
    	uncheckedAppend(arg2);
    	uncheckedAppend(arg3);
    	uncheckedAppend(arg4);
    	uncheckedAppend(arg5);
    	uncheckedAppend(arg6);
    }

    explicit Vector(ValueType&& arg, VectorConstructorFlag)
    {
    	reserveInitialCapacity(1);
    	uncheckedAppend(std::forward<ValueType>(arg));
    }
    explicit Vector(ValueType&& arg1, ValueType&& arg2, VectorConstructorFlag)
    {
    	reserveInitialCapacity(2);
    	uncheckedAppend(std::forward<ValueType>(arg1));
    	uncheckedAppend(std::forward<ValueType>(arg2));
    }

	Vector(const Vector&);
	template<size_t otherCapacity>
	explicit Vector(const Vector<T, otherCapacity, OverflowHandler, minCapacity>&);

	Vector& operator=(const Vector&);
	template<size_t otherCapacity>
	Vector& operator=(const Vector<T, otherCapacity, OverflowHandler, minCapacity>&);

	Vector(Vector&&);
	Vector& operator=(Vector&&);

	ALWAYS_INLINE size_t size() const { return m_size; }
	static ptrdiff_t sizeMemoryOffset() { return OBJECT_OFFSETOF(Vector, m_size); }
	ALWAYS_INLINE size_t capacity() const { return Base::capacity(); }
	ALWAYS_INLINE bool isEmpty() const { return !size(); }

	// std::vector
	ALWAYS_INLINE bool empty() const { return isEmpty(); }

	ALWAYS_INLINE T& at(size_t i)
	{
		ASSERT(i < size());
		return Base::buffer()[i];
	}
	ALWAYS_INLINE const T& at(size_t i) const
	{
		ASSERT(i < size());
		return Base::buffer()[i];
	}

	ALWAYS_INLINE T& operator[](size_t i) { return at(i); }
	ALWAYS_INLINE const T& operator[](size_t i) const { return at(i); }

	ALWAYS_INLINE T* data() { return Base::buffer(); }
	ALWAYS_INLINE const T* data() const { return Base::buffer(); }
	static ptrdiff_t dataMemoryOffset() { return Base::bufferMemoryOffset(); }

	ALWAYS_INLINE iterator begin() { return data(); }
	ALWAYS_INLINE iterator end() { return begin() + m_size; }
	ALWAYS_INLINE const_iterator begin() const { return data(); }
	ALWAYS_INLINE const_iterator end() const { return begin() + m_size; }

	ALWAYS_INLINE reverse_iterator rbegin() { return reverse_iterator(end()); }
	ALWAYS_INLINE reverse_iterator rend() { return reverse_iterator(begin()); }
	ALWAYS_INLINE const_reverse_iterator rbegin() const { return const_reverse_iterator(end()); }
	ALWAYS_INLINE const_reverse_iterator rend() const { return const_reverse_iterator(begin()); }

	ALWAYS_INLINE T& first() { return at(0); }
	ALWAYS_INLINE const T& first() const { return at(0); }
	ALWAYS_INLINE T& last() { return at(size() - 1); }
	ALWAYS_INLINE const T& last() const { return at(size() - 1); }

	// std::vector
	ALWAYS_INLINE T& back() { return last(); }
	ALWAYS_INLINE const T& back() const { return last(); }

	ALWAYS_INLINE T takeLast()
	{
		T result = WTFMove(last());
		removeLast();
		return result;
	}

	template<typename U> bool contains(const U&) const;
	template<typename U> size_t find(const U&) const;
	template<typename MatchFunction> size_t findMatching(const MatchFunction&) const;
	template<typename U> size_t reverseFind(const U&) const;

    template<typename U> bool appendIfNotContains(const U&);

	void shrink(size_t size);
	void grow(size_t size);
	void resize(size_t size);
	void resizeToFit(size_t size);
	void reserveCapacity(size_t newCapacity);
	bool tryReserveCapacity(size_t newCapacity);
	void reserveInitialCapacity(size_t initialCapacity);
	bool tryReserveInitialCapacity(size_t initialCapacity);
	void shrinkCapacity(size_t newCapacity);
	void shrinkToFit()
	{
		unsigned capacityToAllocate = fastMallocGoodSize(size() * sizeof(T)) / sizeof(T);
		if(capacityToAllocate < capacity())
			shrinkCapacity(size());
	}

	// std::vector
	void reserve(size_t newCapacity) { reserveCapacity(newCapacity); }

	ALWAYS_INLINE void clear() { shrinkCapacity(0); }

	ALWAYS_INLINE void constructAndAppend();
	template<typename Arg1>
	ALWAYS_INLINE void constructAndAppend(const Arg1&);
	template<typename Arg1, typename Arg2>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&);
	template<typename Arg1, typename Arg2, typename Arg3>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&, const Arg9&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&, const Arg9&, const Arg10&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10, typename Arg11>
	ALWAYS_INLINE void constructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&, const Arg9&, const Arg10&, const Arg11&);

	ALWAYS_INLINE bool tryConstructAndAppend();
	template<typename Arg1>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&);
	template<typename Arg1, typename Arg2>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&);
	template<typename Arg1, typename Arg2, typename Arg3>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&, const Arg9&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&, const Arg9&, const Arg10&);
	template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10, typename Arg11>
	ALWAYS_INLINE bool tryConstructAndAppend(const Arg1&, const Arg2&, const Arg3&, const Arg4&, const Arg5&, const Arg6&, const Arg7&, const Arg8&, const Arg9&, const Arg10&, const Arg11&);

	template<typename U> void append(const U*, size_t);
	template<typename U> void append(U&&);
	template<typename U> void uncheckedAppend(U&& val);
	template<typename U, size_t otherCapacity> void appendVector(const Vector<U, otherCapacity, OverflowHandler, minCapacity>&);
	template<typename U, size_t otherCapacity> void appendVector(Vector<U, otherCapacity>&&);
	template<typename U> bool tryAppend(const U*, size_t);

	// std::vector
	template<typename U> void push_back(const U* data, size_t dataSize) { append(data, dataSize); }
	template<typename U> void push_back(U&& value) { append(std::forward<U>(value)); }

	template<typename U> void insert(size_t position, const U*, size_t);
    template<typename U> void insert(size_t position, U&&);
	template<typename U, size_t c> void insertVector(size_t position, const Vector<U, c, OverflowHandler, minCapacity>&);

	void remove(size_t position);
	void remove(size_t position, size_t length);
	template<typename U> bool removeFirst(const U&);
	template<typename MatchFunction> bool removeFirstMatching(const MatchFunction&, size_t startIndex = 0);
	template<typename U> unsigned removeAll(const U&);
	template<typename MatchFunction> unsigned removeAllMatching(const MatchFunction&, size_t startIndex = 0);

	ALWAYS_INLINE void removeLast()
	{
		ASSERT(!isEmpty());
		shrink(size() - 1);
	}

	Vector(size_t size, const T& val)
		: Base(size, size)
	{
		if (begin())
			TypeOperations::uninitializedFill(begin(), end(), val);
	}

	void fill(const T&, size_t);
	void fill(const T& val) { fill(val, size()); }

	template<typename Iterator> void appendRange(Iterator start, Iterator end);

	T* releaseBuffer();

	void swap(Vector<T, inlineCapacity, OverflowHandler, minCapacity>& other)
	{
		Base::swap(other, m_size, other.m_size);
		std::swap(m_size, other.m_size);
	}

	void reverse();

    template<typename R, typename MapFunction> Vector<R> map(MapFunction) const;

	void checkConsistency();

	void expandCapacity(size_t newMinCapacity);

    bool isHashTableDeletedValue() const { return m_size == std::numeric_limits<unsigned>::max(); }

private:
	const T* expandCapacity(size_t newMinCapacity, const T*);
	bool tryExpandCapacity(size_t newMinCapacity);
	const T* tryExpandCapacity(size_t newMinCapacity, const T*);
	template<typename U> U* expandCapacity(size_t newMinCapacity, U*);

	using Base::m_size;
	using Base::buffer;
	using Base::capacity;
	using Base::swap;
	using Base::allocateBuffer;
	using Base::deallocateBuffer;
	using Base::tryAllocateBuffer;
	using Base::shouldReallocateBuffer;
	using Base::reallocateBuffer;
	using Base::restoreInlineBufferIfNeeded;
	using Base::releaseBuffer;
};

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
Vector<T, inlineCapacity, OverflowHandler, minCapacity>::Vector(const Vector& other)
	: Base(other.size(), other.size())
{
	if (begin())
		TypeOperations::uninitializedCopy(other.begin(), other.end(), begin());
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<size_t otherCapacity>
Vector<T, inlineCapacity, OverflowHandler, minCapacity>::Vector(const Vector<T, otherCapacity, OverflowHandler, minCapacity>& other)
	: Base(other.size(), other.size())
{
	if (begin())
		TypeOperations::uninitializedCopy(other.begin(), other.end(), begin());
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
Vector<T, inlineCapacity, OverflowHandler, minCapacity>& Vector<T, inlineCapacity, OverflowHandler, minCapacity>::operator=(const Vector<T, inlineCapacity, OverflowHandler, minCapacity>& other)
{
	if (&other == this)
		return *this;

	if (size() > other.size())
		shrink(other.size());
	else if (other.size() > capacity()) {
		clear();
		reserveCapacity(other.size());
		ASSERT(begin());
	}

// Works around an assert in VS2010. See https://connect.microsoft.com/VisualStudio/feedback/details/558044/std-copy-should-not-check-dest-when-first-last
#if COMPILER(MSVC) && defined(_ITERATOR_DEBUG_LEVEL) && _ITERATOR_DEBUG_LEVEL
	if (!begin())
		return *this;
#endif

	std::copy(other.begin(), other.begin() + size(), begin());
	TypeOperations::uninitializedCopy(other.begin() + size(), other.end(), end());
	m_size = other.size();

	return *this;
}

ALWAYS_INLINE bool typelessPointersAreEqual(const void* a, const void* b) { return a == b; }

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<size_t otherCapacity>
Vector<T, inlineCapacity, OverflowHandler, minCapacity>& Vector<T, inlineCapacity, OverflowHandler, minCapacity>::operator=(const Vector<T, otherCapacity, OverflowHandler, minCapacity>& other)
{
	// If the inline capacities match, we should call the more specific
	// template.  If the inline capacities don't match, the two objects
	// shouldn't be allocated the same address.
	ASSERT(!typelessPointersAreEqual(&other, this));

	if (size() > other.size())
		shrink(other.size());
	else if (other.size() > capacity()) {
		clear();
		reserveCapacity(other.size());
		ASSERT(begin());
	}

// Works around an assert in VS2010. See https://connect.microsoft.com/VisualStudio/feedback/details/558044/std-copy-should-not-check-dest-when-first-last
#if COMPILER(MSVC) && defined(_ITERATOR_DEBUG_LEVEL) && _ITERATOR_DEBUG_LEVEL
	if (!begin())
		return *this;
#endif

	std::copy(other.begin(), other.begin() + size(), begin());
	TypeOperations::uninitializedCopy(other.begin() + size(), other.end(), end());
	m_size = other.size();

	return *this;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline Vector<T, inlineCapacity, OverflowHandler, minCapacity>::Vector(Vector<T, inlineCapacity, OverflowHandler, minCapacity>&& other)
{
	swap(other);
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline Vector<T, inlineCapacity, OverflowHandler, minCapacity>& Vector<T, inlineCapacity, OverflowHandler, minCapacity>::operator=(Vector<T, inlineCapacity, OverflowHandler, minCapacity>&& other)
{
	swap(other);
	return *this;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::contains(const U& value) const
{
	return find(value) != notFound;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename MatchFunction>
size_t Vector<T, inlineCapacity, OverflowHandler, minCapacity>::findMatching(const MatchFunction& matches) const
{
	for (size_t i = 0; i < size(); ++i) {
		if (matches(at(i)))
			return i;
	}
	return notFound;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
size_t Vector<T, inlineCapacity, OverflowHandler, minCapacity>::find(const U& value) const
{
	class MatchFunction
	{
	public:
		MatchFunction(const U& value) : value(value) { }
		bool operator()(const T& item) const { return item == value; }
	private:
		const U& value;
	};
	MatchFunction matchFunction(value);
	return findMatching(matchFunction);
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
size_t Vector<T, inlineCapacity, OverflowHandler, minCapacity>::reverseFind(const U& value) const
{
	for (size_t i = 1; i <= size(); ++i) {
		const size_t index = size() - i;
		if (at(index) == value)
			return index;
	}
	return notFound;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::appendIfNotContains(const U& value)
{
    if (contains(value))
        return false;
    append(value);
    return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::fill(const T& val, size_t newSize)
{
	if (size() > newSize)
		shrink(newSize);
	else if (newSize > capacity()) {
		clear();
		reserveCapacity(newSize);
		ASSERT(begin());
	}

	std::fill(begin(), end(), val);
	TypeOperations::uninitializedFill(end(), begin() + newSize, val);
	m_size = newSize;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Iterator>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::appendRange(Iterator start, Iterator end)
{
	for (Iterator it = start; it != end; ++it)
		append(*it);
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
NEVER_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::expandCapacity(size_t newMinCapacity)
{
	reserveCapacity(std::max(newMinCapacity, std::max(static_cast<size_t>(minCapacity), capacity() + capacity() / 4 + 1)));
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
NEVER_INLINE const T* Vector<T, inlineCapacity, OverflowHandler, minCapacity>::expandCapacity(size_t newMinCapacity, const T* ptr)
{
	if (ptr < begin() || ptr >= end()) {
		expandCapacity(newMinCapacity);
		return ptr;
	}
	size_t index = ptr - begin();
	expandCapacity(newMinCapacity);
	return begin() + index;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
NEVER_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryExpandCapacity(size_t newMinCapacity)
{
	return tryReserveCapacity(std::max(newMinCapacity, std::max(static_cast<size_t>(minCapacity), capacity() + capacity() / 4 + 1)));
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
NEVER_INLINE const T* Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryExpandCapacity(size_t newMinCapacity, const T* ptr)
{
	if (ptr < begin() || ptr >= end()) {
		if (!tryExpandCapacity(newMinCapacity))
			return nullptr;
		return ptr;
	}
	size_t index = ptr - begin();
	if (!tryExpandCapacity(newMinCapacity))
		return nullptr;
	return begin() + index;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity> template<typename U>
ALWAYS_INLINE U* Vector<T, inlineCapacity, OverflowHandler, minCapacity>::expandCapacity(size_t newMinCapacity, U* ptr)
{
	expandCapacity(newMinCapacity);
	return ptr;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::resize(size_t size)
{
	if (size <= m_size)
		TypeOperations::destruct(begin() + size, end());
	else {
		if (size > capacity())
			expandCapacity(size);
		if (begin())
			TypeOperations::initialize(end(), begin() + size);
	}

	m_size = size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::resizeToFit(size_t size)
{
	reserveCapacity(size);
	resize(size);
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::shrink(size_t size)
{
	ASSERT(size <= m_size);
	TypeOperations::destruct(begin() + size, end());
	m_size = size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::grow(size_t size)
{
	ASSERT(size >= m_size);
	if (size > capacity())
		expandCapacity(size);
	if (begin())
		TypeOperations::initialize(end(), begin() + size);
	m_size = size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::reserveCapacity(size_t newCapacity)
{
	if (newCapacity <= capacity())
		return;
	T* oldBuffer = begin();
	T* oldEnd = end();
	Base::allocateBuffer(newCapacity);
	ASSERT(begin());
	TypeOperations::move(oldBuffer, oldEnd, begin());
	Base::deallocateBuffer(oldBuffer);
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryReserveCapacity(size_t newCapacity)
{
	if (newCapacity <= capacity())
		return true;
	T* oldBuffer = begin();
	T* oldEnd = end();
	if (!Base::tryAllocateBuffer(newCapacity))
		return false;
	ASSERT(begin());
	TypeOperations::move(oldBuffer, oldEnd, begin());
	Base::deallocateBuffer(oldBuffer);
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::reserveInitialCapacity(size_t initialCapacity)
{
	ASSERT(!m_size);
	ASSERT(capacity() == inlineCapacity);
	if (initialCapacity > inlineCapacity)
		Base::allocateBuffer(initialCapacity);
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryReserveInitialCapacity(size_t initialCapacity)
{
	ASSERT(!m_size);
	ASSERT(capacity() == inlineCapacity);
	if (initialCapacity > inlineCapacity)
		if (!Base::tryAllocateBuffer(initialCapacity))
			return false;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::shrinkCapacity(size_t newCapacity)
{
	if (newCapacity >= capacity())
		return;

	if (newCapacity < size())
		shrink(newCapacity);

	T* oldBuffer = begin();
	if (newCapacity > 0) {
		if (Base::shouldReallocateBuffer(newCapacity)) {
			Base::reallocateBuffer(newCapacity);
			return;
		}

		MonitorMallocFailedPreventer monitorMallocFailedPreventer;

		T* oldEnd = end();
		Base::allocateBuffer(newCapacity);
		if (begin() != oldBuffer)
			TypeOperations::move(oldBuffer, oldEnd, begin());
	}

	Base::deallocateBuffer(oldBuffer);
	Base::restoreInlineBufferIfNeeded();
}

// Templatizing these is better than just letting the conversion happen implicitly,
// because for instance it allows a PassRefPtr to be appended to a RefPtr vector
// without refcount thrash.
template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::append(const U* data, size_t dataSize)
{
	size_t newSize = m_size + dataSize;
	if (UNLIKELY(newSize > capacity())) {
		data = expandCapacity(newSize, data);
		ASSERT(begin());
	}
	//if (newSize < m_size)
	//    CRASH();
	T* dest = end();
	VectorCopier<VectorTraits<T>::canCopyWithMemcpy, U>::uninitializedCopy(data, &data[dataSize], dest);
	m_size = newSize;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryAppend(const U* data, size_t dataSize)
{
	size_t newSize = m_size + dataSize;
	if (UNLIKELY(newSize > capacity())) {
		data = tryExpandCapacity(newSize, data);
		if (!data)
			return false;
		ASSERT(begin());
	}
	if (UNLIKELY(newSize < m_size))
		return false;
	T* dest = end();
	VectorCopier<VectorTraits<T>::canCopyWithMemcpy, U>::uninitializedCopy(data, &data[dataSize], dest);
	m_size = newSize;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity> template<typename U>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::append(U&& value)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	new (NotNull, end()) T(std::forward<U>(value));
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend()
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T();
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8, const Arg9& arg9)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8, const Arg9& arg9, const Arg10& arg10)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10, typename Arg11>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::constructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8, const Arg9& arg9, const Arg10& arg10, const Arg11& arg11)
{
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11);
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend()
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T();
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8, const Arg9& arg9)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8, const Arg9& arg9, const Arg10& arg10)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10);
	++m_size;
	return true;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename Arg1, typename Arg2, typename Arg3, typename Arg4, typename Arg5, typename Arg6, typename Arg7, typename Arg8, typename Arg9, typename Arg10, typename Arg11>
ALWAYS_INLINE bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::tryConstructAndAppend(const Arg1& arg1, const Arg2& arg2, const Arg3& arg3, const Arg4& arg4, const Arg5& arg5, const Arg6& arg6, const Arg7& arg7, const Arg8& arg8, const Arg9& arg9, const Arg10& arg10, const Arg11& arg11)
{
	if (UNLIKELY(size() == capacity())) {
	    if (UNLIKELY(!tryExpandCapacity(size() + 1)))
	        return false;
	}
	//asanBufferSizeWillChangeTo(m_size + 1);
	new (NotNull, end()) T(arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11);
	++m_size;
	return true;
}

// This version of append saves a branch in the case where you know that the
// vector's capacity is large enough for the append to succeed.

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity> template<typename U>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::uncheckedAppend(U&& value)
{
	ASSERT(size() < capacity());
	new (NotNull, end()) T(std::forward<U>(value));
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity> template<typename U, size_t otherCapacity>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::appendVector(const Vector<U, otherCapacity, OverflowHandler, minCapacity>& val)
{
	append(val.begin(), val.size());
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U, size_t otherCapacity>
inline void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::appendVector(Vector<U, otherCapacity>&& val)
{
    size_t newSize = m_size + val.size();
    if (newSize > capacity())
        expandCapacity(newSize);
    for (auto& item : val)
        uncheckedAppend(WTFMove(item));
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity> template<typename U>
void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::insert(size_t position, const U* data, size_t dataSize)
{
	ASSERT(position <= size());
	size_t newSize = m_size + dataSize;
	if (UNLIKELY(newSize > capacity())) {
		data = expandCapacity(newSize, data);
		ASSERT(begin());
	}
	//if (newSize < m_size)
	//    CRASH();
	T* spot = begin() + position;
	TypeOperations::moveOverlapping(spot, end(), spot + dataSize);
	VectorCopier<VectorTraits<T>::canCopyWithMemcpy, U>::uninitializedCopy(data, &data[dataSize], spot);
	m_size = newSize;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity> template<typename U>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::insert(size_t position, U&& value)
{
	ASSERT(position <= size());
	if (UNLIKELY(size() == capacity())) {
		expandCapacity(size() + 1);
		ASSERT(begin());
	}
	T* spot = begin() + position;
	TypeOperations::moveOverlapping(spot, end(), spot + 1);
    new (NotNull, spot) T(std::forward<U>(value));
	++m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity> template<typename U, size_t c>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::insertVector(size_t position, const Vector<U, c, OverflowHandler, minCapacity>& val)
{
	insert(position, val.begin(), val.size());
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
ALWAYS_INLINE void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::remove(size_t position)
{
	ASSERT(position < size());
	T* spot = begin() + position;
	spot->~T();
	TypeOperations::moveOverlapping(spot + 1, end(), spot);
	--m_size;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::remove(size_t position, size_t length)
{
	ASSERT(position <= size());
	ASSERT(position + length <= size());
	T* beginSpot = begin() + position;
	T* endSpot = beginSpot + length;
	TypeOperations::destruct(beginSpot, endSpot);
	TypeOperations::moveOverlapping(endSpot, end(), beginSpot);
	m_size -= length;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
inline bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::removeFirst(const U& value)
{
	return removeFirstMatching( [&] (const T& current) {
		return current == value;
	});
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename MatchFunction>
inline bool Vector<T, inlineCapacity, OverflowHandler, minCapacity>::removeFirstMatching(const MatchFunction& matches, size_t startIndex)
{
	for (size_t i = startIndex; i < size(); ++i) {
		if (matches(at(i))) {
			remove(i);
			return true;
		}
	}
	return false;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename U>
inline unsigned Vector<T, inlineCapacity, OverflowHandler, minCapacity>::removeAll(const U& value)
{
	return removeAllMatching( [&] (const T& current) {
		return current == value;
	});
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename MatchFunction>
inline unsigned Vector<T, inlineCapacity, OverflowHandler, minCapacity>::removeAllMatching(const MatchFunction& matches, size_t startIndex)
{
	iterator holeBegin = end();
	iterator holeEnd = end();
	unsigned matchCount = 0;
	for (auto it = begin() + startIndex, itEnd = end(); it < itEnd; ++it) {
		if (matches(*it)) {
			if (holeBegin == end())
				holeBegin = it;
			else if (holeEnd != it) {
				TypeOperations::moveOverlapping(holeEnd, it, holeBegin);
				holeBegin += it - holeEnd;
			}
			holeEnd = it + 1;
			it->~T();
			++matchCount;
		}
	}
	if (holeEnd != end())
		TypeOperations::moveOverlapping(holeEnd, end(), holeBegin);
	//asanBufferSizeWillChangeTo(m_size - matchCount);
	m_size -= matchCount;
	return matchCount;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::reverse()
{
	for (size_t i = 0; i < m_size / 2; ++i)
		std::swap(at(i), at(m_size - 1 - i));
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
template<typename R, typename MapFunction>
inline Vector<R> Vector<T, inlineCapacity, OverflowHandler, minCapacity>::map(MapFunction mapFunction) const
{
    Vector<R> result;
    result.reserveInitialCapacity(size());
    for (size_t i = 0; i < size(); ++i)
        result.uncheckedAppend(mapFunction(at(i)));
    return result;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline T* Vector<T, inlineCapacity, OverflowHandler, minCapacity>::releaseBuffer()
{
	T* buffer = Base::releaseBuffer();
	if (inlineCapacity && !buffer && m_size) {
		// If the vector had some data, but no buffer to release,
		// that means it was using the inline buffer. In that case,
		// we create a brand new buffer so the caller always gets one.
		size_t bytes = m_size * sizeof(T);
		buffer = static_cast<T*>(fastMalloc(bytes));
		memcpy(buffer, data(), bytes);
	}
	m_size = 0;
	return buffer;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline void Vector<T, inlineCapacity, OverflowHandler, minCapacity>::checkConsistency()
{
#if !ASSERT_DISABLED
	for (size_t i = 0; i < size(); ++i)
		ValueCheck<T>::checkConsistency(at(i));
#endif
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
void deleteAllValues(const Vector<T, inlineCapacity, OverflowHandler, minCapacity>& collection)
{
	typedef typename Vector<T, inlineCapacity, OverflowHandler, minCapacity>::const_iterator iterator;
	iterator end = collection.end();
	for (iterator it = collection.begin(); it != end; ++it)
		delete *it;
}

template<typename T, size_t inlineCapacity, typename OverflowHandler, size_t minCapacity>
inline void swap(Vector<T, inlineCapacity, OverflowHandler, minCapacity>& a, Vector<T, inlineCapacity, OverflowHandler, minCapacity>& b)
{
	a.swap(b);
}

template<typename T, size_t inlineCapacityA, typename OverflowHandlerA, size_t minCapacityA, size_t inlineCapacityB, typename OverflowHandlerB, size_t minCapacityB>
bool operator==(const Vector<T, inlineCapacityA, OverflowHandlerA, minCapacityA>& a, const Vector<T, inlineCapacityB, OverflowHandlerB, minCapacityB>& b)
{
	if (a.size() != b.size())
		return false;

	return VectorTypeOperations<T>::compare(a.data(), b.data(), a.size());
}

template<typename T, size_t inlineCapacityA, typename OverflowHandlerA, size_t minCapacityA, size_t inlineCapacityB, typename OverflowHandlerB, size_t minCapacityB>
inline bool operator!=(const Vector<T, inlineCapacityA, OverflowHandlerA, minCapacityA>& a, const Vector<T, inlineCapacityB, OverflowHandlerB, minCapacityB>& b)
{
	return !(a == b);
}

#if !ASSERT_DISABLED
    template<typename T> struct ValueCheck<Vector<T> > {
        typedef Vector<T> TraitType;
        static void checkConsistency(const Vector<T>& v)
        {
            v.checkConsistency();
        }
    };
#endif

} // namespace WTF

using WTF::Vector;
using WTF::VectorConstructorFlag;
