
#pragma once

#include <wtf/GetPtr.h>
#include <wtf/UnusedParam.h>
#include <wtf/text/SuperFastHash.h>

namespace WTF {

class ASCIILiteral final {
public:
    static ASCIILiteral fromLiteralUnsafe(const char* string)
    {
        return ASCIILiteral { string, strlen(string) };
    }

	constexpr ASCIILiteral() = default;

	template<unsigned characterCount>
	explicit constexpr ALWAYS_INLINE ASCIILiteral(const char (&characters)[characterCount])
    	: m_characters(characters)
    	, m_len(characterCount - 1)
    {
        static_assert(characterCount > 0, "ASCIILiteral not empty");
        static_assert(characterCount < 10000, "ASCIILiteral too long");
    }

    static constexpr ASCIILiteral null()
    {
        return ASCIILiteral();
    }

    // Hash table deleted/empty values, which are only constructed and never copied or destroyed.
    ALWAYS_INLINE ASCIILiteral(HashTableDeletedValueType) : m_characters(hashTableDeletedValue()) { }
    ALWAYS_INLINE bool isHashTableDeletedValue() const { return m_characters == hashTableDeletedValue(); }
    static ALWAYS_INLINE const char* hashTableDeletedValue() { return reinterpret_cast<const char*>(-1); }

    unsigned hash() const;
    constexpr bool isNull() const { return !m_characters; }

    constexpr explicit operator bool() const { return !!m_characters; }
    constexpr bool operator!() const { return !m_characters; }

    constexpr operator const char*() const { return m_characters; }

    constexpr const char* characters() const { return m_characters; }
    constexpr unsigned length() const { return m_len; }

    constexpr char characterAt(unsigned index) const { return m_characters[index]; }

private:
	explicit ASCIILiteral(const char* characters, unsigned len)
    	: m_characters(characters)
    	, m_len(len)
    {
    }

    const char* m_characters { nullptr };
    unsigned m_len { 0 };
};

inline bool operator==(const ASCIILiteral& a, const char* b)
{
    if (!a || !b)
        return a.characters() == b;
    return !strcmp(a.characters(), b);
}

inline bool operator==(const ASCIILiteral& a, const ASCIILiteral& b)
{
    if (!a || !b)
        return a.characters() == b.characters();
    return !strcmp(a.characters(), b.characters());
}

inline unsigned ASCIILiteral::hash() const
{
    if (isNull())
        return 0;
    SuperFastHash hasher;
    hasher.addCharacters(characters(), length());
    return hasher.hash();
}

struct ASCIILiteralHash {
    static unsigned hash(const ASCIILiteral& literal) { return literal.hash(); }
    static bool equal(const ASCIILiteral& a, const ASCIILiteral& b) { return a == b; }
    static constexpr bool safeToCompareToEmptyOrDeleted = false;
};

template<typename T> struct DefaultHash;
template<> struct DefaultHash<ASCIILiteral>
{
    typedef ASCIILiteralHash Hash;
};

inline namespace StringLiterals {

inline ASCIILiteral operator"" _s(const char* characters, size_t n)
{
#if ASSERT_ENABLED
    for (size_t i = 0; i < n; ++i)
        ASSERT_UNDER_CONSTEXPR_CONTEXT(isASCII(characters[i]));
#else
    UNUSED_PARAM(n);
#endif
    return ASCIILiteral::fromLiteralUnsafe(characters);
}

} // namespace StringLiterals

} // namespace WTF

using WTF::ASCIILiteral;
using namespace WTF::StringLiterals;
