/*
 * Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Patrick Gansterer <paroga@paroga.com>
 * Copyright (C) 2012 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"

#include "AtomicString.h"

#include "AtomicStringTable.h"
#include "IntegerToStringConversion.h"
#include "dtoa.h"
#include "StringHash.h"
#include <wtf/HashSet.h>
#include <wtf/Threading.h>
#include <wtf/WTFThreadData.h>
#include <wtf/text/StringCommon.h>
#include <wtf/unicode/UTF8.h>

namespace WTF {

using namespace Unicode;

COMPILE_ASSERT(sizeof(AtomicString) == sizeof(String), atomic_string_and_string_must_be_same_size);

#if USE(WEB_THREAD)

class AtomicStringTableLocker : public SpinLockHolder {
    WTF_MAKE_NONCOPYABLE(AtomicStringTableLocker);

    static StaticSpinLock s_stringTableLock;
public:
    AtomicStringTableLocker()
        : SpinLockHolder(&s_stringTableLock)
    {
    }
};

StaticSpinLock AtomicStringTableLocker::s_stringTableLock;

#else

class AtomicStringTableLocker {
    WTF_MAKE_NONCOPYABLE(AtomicStringTableLocker);
public:
    AtomicStringTableLocker() { }
};

#endif // USE(WEB_THREAD)

typedef AtomicStringTable::StringTableImpl StringTableImpl;

static ALWAYS_INLINE StringTableImpl& stringTable()
{
    return wtfThreadData().atomicStringTable()->table();
}

template<typename T, typename HashTranslator>
static inline PassRefPtr<StringImpl> addToStringTable(AtomicStringTableLocker&, StringTableImpl& atomicStringTable, const T& value)
{
    auto addResult = atomicStringTable.fastAdd<HashTranslator>(value);

    // If the string is newly-translated, then we need to adopt it.
    // The boolean in the pair tells us if that is so.
    return addResult.isNewEntry ? adoptRef(*addResult.iterator) : *addResult.iterator;
}

template<typename T, typename HashTranslator>
static inline PassRefPtr<StringImpl> addToStringTable(const T& value)
{
    AtomicStringTableLocker locker;
    return addToStringTable<T, HashTranslator>(locker, stringTable(), value);
}

struct CStringTranslator {
    static unsigned hash(const LChar* characters)
    {
        return StringHasher::computeHashAndMaskTop8Bits(characters);
    }

    static inline bool equal(StringImpl* str, const LChar* characters)
    {
        return WTF::equal(str, characters);
    }

    static void translate(StringImpl*& location, const LChar* const& characters, unsigned hash)
    {
        location = &StringImpl::createFromCString(reinterpret_cast<const char*>(characters)).leakRef();
        location->setHash(hash);
        location->setIsAtomic(true);
    }
};

DLL_PUBLIC PassRefPtr<StringImpl> AtomicString::add(const LChar* characters)
{
    if (!characters)
        return nullptr;
    if (!*characters)
        return StringImpl::empty();

    return addToStringTable<const LChar*, CStringTranslator>(characters);
}

typedef HashTranslatorCharBuffer<UChar> UCharBuffer;
struct UCharBufferTranslator {
    static unsigned hash(const UCharBuffer& buf)
    {
        return buf.hash;
    }

    static ALWAYS_INLINE  bool equal(StringImpl* const& str, const UCharBuffer& buf)
    {
    	ASSERT(str);
    	ASSERT(buf.characters);
    	ASSERT(buf.length);
        if (str->length() != buf.length)
            return false;
        if (str->is8Bit())
            return WTF::equal(str->characters8(), buf.characters, buf.length);
        return WTF::equal(str->characters16(), buf.characters, buf.length);
    }

    static ALWAYS_INLINE void translate(StringImpl*& location, const UCharBuffer& buf, unsigned hash)
    {
        location = create8BitIfPossible(buf.characters, buf.length).leakRef();
        location->setHash(hash);
        location->setIsAtomic(true);
    }

private:
    static ALWAYS_INLINE PassRefPtr<StringImpl> createInternal(const UChar* characters, unsigned length)
    {
    	UChar* data;
        RefPtr<StringImpl> string = StringImpl::createUninitializedInternalNonEmpty(length, data);
        memcpy(data, characters, length * sizeof(UChar));
        return string.release();
    }

    static ALWAYS_INLINE PassRefPtr<StringImpl> create8BitIfPossible(const UChar* characters, unsigned length)
    {
        if (!isLatin1(characters[0]))
            return createInternal(characters, length);

        LChar* data;
        RefPtr<StringImpl> string = StringImpl::createUninitializedInternalNonEmpty(length, data);

        for (size_t i = 0; i < length; ++i) {
        	UChar character = characters[i];
            if (!isLatin1(character))
                return createInternal(characters, length);
            data[i] = static_cast<LChar>(character);
        }

        return string.release();
    }

};

template<typename CharacterType>
struct HashAndCharacters {
    unsigned hash;
    const CharacterType* characters;
    unsigned length;
};

template<typename CharacterType>
struct HashAndCharactersTranslator {
    static unsigned hash(const HashAndCharacters<CharacterType>& buffer)
    {
        ASSERT(buffer.hash == StringHasher::computeHashAndMaskTop8Bits(buffer.characters, buffer.length));
        return buffer.hash;
    }

    static bool equal(StringImpl* const& string, const HashAndCharacters<CharacterType>& buffer)
    {
        return WTF::equal(string, buffer.characters, buffer.length);
    }

    static void translate(StringImpl*& location, const HashAndCharacters<CharacterType>& buffer, unsigned hash)
    {
        location = &StringImpl::create(buffer.characters, buffer.length).leakRef();
        location->setHash(hash);
        location->setIsAtomic(true);
    }
};

struct HashAndUTF8Characters {
    unsigned hash;
    const char* characters;
    unsigned length;
    unsigned utf16Length;
};

struct HashAndUTF8CharactersTranslator {
    static unsigned hash(const HashAndUTF8Characters& buffer)
    {
        return buffer.hash;
    }

    static bool equal(StringImpl* const& string, const HashAndUTF8Characters& buffer)
    {
        if (buffer.utf16Length != string->length())
            return false;

        // If buffer contains only ASCII characters UTF-8 and UTF16 length are the same.
        if (buffer.utf16Length != buffer.length) {
            if (string->is8Bit())
                return equalLatin1WithUTF8(string->characters8(), buffer.characters, buffer.characters + buffer.length);

            return equalUTF16WithUTF8(string->characters16(), buffer.characters, buffer.characters + buffer.length);
        }

        if (string->is8Bit()) {
            const LChar* stringCharacters = string->characters8();

			for (unsigned i = 0; i < buffer.length; ++i) {
				ASSERT(isASCII(buffer.characters[i]));
				if (stringCharacters[i] != buffer.characters[i])
					return false;
			}

			return true;
        }

        const UChar* stringCharacters = string->characters16();

        for (unsigned i = 0; i < buffer.length; ++i) {
            ASSERT(isASCII(buffer.characters[i]));
            if (stringCharacters[i] != buffer.characters[i])
                return false;
        }

        return true;
    }

    static void translate(StringImpl*& location, const HashAndUTF8Characters& buffer, unsigned hash)
    {
        UChar* target;
        Ref<StringImpl> newString = StringImpl::createUninitialized(buffer.utf16Length, target);

        bool isAllASCII;
        const char* source = buffer.characters;
        if (!convertUTF8ToUTF16(source, source + buffer.length, &target, target + buffer.utf16Length, &isAllASCII))
            ASSERT_NOT_REACHED();

        if (isAllASCII)
            newString = StringImpl::create(buffer.characters, buffer.length);

        location = &newString.leakRef();
        location->setHash(hash);
        location->setIsAtomic(true);
    }
};

PassRefPtr<StringImpl> AtomicString::add(const UChar* characters, unsigned length)
{
    if (UNLIKELY(!characters))
        return nullptr;

    if (UNLIKELY(!length))
        return StringImpl::empty();
    
    UCharBuffer buffer(characters, length);
    return addToStringTable<UCharBuffer, UCharBufferTranslator>(buffer);
}

PassRefPtr<StringImpl> AtomicString::addSlowCaseNotEmpty(AtomicStringTable& stringTable, const UChar* characters, unsigned length)
{
    ASSERT(characters && length);
    AtomicStringTableLocker locker;
    UCharBuffer buffer(characters, length);
    return addToStringTable<UCharBuffer, UCharBufferTranslator>(locker, stringTable.table(), buffer);
}

PassRefPtr<StringImpl> AtomicString::addSlowCase(AtomicStringTable& stringTable, const UChar* characters, unsigned length)
{
    if (UNLIKELY(!characters))
        return nullptr;

    if (UNLIKELY(!length))
        return StringImpl::empty();

    AtomicStringTableLocker locker;
    UCharBuffer buffer(characters, length);
    return addToStringTable<UCharBuffer, UCharBufferTranslator>(locker, stringTable.table(), buffer);
}

PassRefPtr<StringImpl> AtomicString::add(const UChar* characters, unsigned length, unsigned existingHash)
{
    ASSERT(characters);
    ASSERT(existingHash);

    if (UNLIKELY(!length))
        return StringImpl::empty();

    HashAndCharacters<UChar> buffer = { existingHash, characters, length };
    return addToStringTable<HashAndCharacters<UChar>, HashAndCharactersTranslator<UChar> >(buffer);
}

PassRefPtr<StringImpl> AtomicString::add(const UChar* characters)
{
    if (!characters)
        return nullptr;

    int length = 0;
    while (characters[length] != UChar(0))
        length++;

    if (!length)
        return StringImpl::empty();

    UCharBuffer buffer(characters, static_cast<unsigned int>(length));
    return addToStringTable<UCharBuffer, UCharBufferTranslator>(buffer);
}

struct SubstringLocation {
    StringImpl* baseString;
    unsigned start;
    unsigned length;
};

struct SubstringTranslator {
    static void translate(StringImpl*& location, const SubstringLocation& buffer, unsigned hash)
    {
        location = &StringImpl::createSubstringSharingImpl(buffer.baseString, buffer.start, buffer.length).leakRef();
        location->setHash(hash);
        location->setIsAtomic(true);
    }
};

struct SubstringTranslator8 : SubstringTranslator {
    static unsigned hash(const SubstringLocation& buffer)
    {
        return StringHasher::computeHashAndMaskTop8Bits(buffer.baseString->characters8() + buffer.start, buffer.length);
    }

    static bool equal(StringImpl* const& string, const SubstringLocation& buffer)
    {
        return WTF::equal(string, buffer.baseString->characters8() + buffer.start, buffer.length);
    }
};

struct SubstringTranslator16 : SubstringTranslator {
    static unsigned hash(const SubstringLocation& buffer)
    {
        return StringHasher::computeHashAndMaskTop8Bits(buffer.baseString->characters16() + buffer.start, buffer.length);
    }

    static bool equal(StringImpl* const& string, const SubstringLocation& buffer)
    {
        return WTF::equal(string, buffer.baseString->characters16() + buffer.start, buffer.length);
    }
};

PassRefPtr<StringImpl> AtomicString::add(StringImpl* baseString, unsigned start, unsigned length)
{
    if (!baseString)
        return nullptr;

    if (!length || start >= baseString->length())
        return StringImpl::empty();

    unsigned maxLength = baseString->length() - start;
    if (length >= maxLength) {
        if (!start)
            return add(baseString);
        length = maxLength;
    }

    SubstringLocation buffer = { baseString, start, length };
    if (baseString->is8Bit())
        return addToStringTable<SubstringLocation, SubstringTranslator8>(buffer);
    return addToStringTable<SubstringLocation, SubstringTranslator16>(buffer);
}

typedef HashTranslatorCharBuffer<LChar> LCharBuffer;
struct LCharBufferTranslator {
    static unsigned hash(const LCharBuffer& buf)
    {
        return buf.hash;
    }

    static ALWAYS_INLINE bool equal(StringImpl* const& str, const LCharBuffer& buf)
    {
    	ASSERT(str);
    	ASSERT(buf.characters);
    	ASSERT(buf.length);
        if (str->length() != buf.length)
            return false;

        if (str->is8Bit())
            return WTF::equal(str->characters8(), buf.characters, buf.length);
        return WTF::equal(str->characters16(), buf.characters, buf.length);
    }

    static ALWAYS_INLINE void translate(StringImpl*& location, const LCharBuffer& buf, unsigned hash)
    {
        LChar* data;
        RefPtr<StringImpl> string = StringImpl::createUninitializedInternalNonEmpty(buf.length, data);
        memcpy(data, buf.characters, buf.length * sizeof(LChar));
        location = string.leakRef();
        location->setHash(hash);
        location->setIsAtomic(true);
    }
};

template<typename CharType>
struct BufferFromStaticDataTranslator {
    typedef HashTranslatorCharBuffer<CharType> Buffer;
    static unsigned hash(const Buffer& buf)
    {
        return buf.hash;
    }

    static bool equal(StringImpl* const& str, const Buffer& buf)
    {
        return WTF::equal(str, buf.characters, buf.length);
    }

    static void translate(StringImpl*& location, const Buffer& buf, unsigned hash)
    {
        location = &StringImpl::createWithoutCopying(buf.characters, buf.length).leakRef();
        location->setHash(hash);
        location->setIsAtomic(true);
    }
};

PassRefPtr<StringImpl> AtomicString::add(const LChar* characters, unsigned length)
{
    if (UNLIKELY(!characters))
        return nullptr;

    if (UNLIKELY(!length))
        return StringImpl::empty();

    LCharBuffer buffer(characters, length);
    return addToStringTable<LCharBuffer, LCharBufferTranslator>(buffer);
}

PassRefPtr<StringImpl> AtomicString::addSlowCaseNotEmpty(AtomicStringTable& stringTable, const LChar* characters, unsigned length)
{
    ASSERT(characters && length);
    AtomicStringTableLocker locker;
    LCharBuffer buffer(characters, length);
    return addToStringTable<LCharBuffer, LCharBufferTranslator>(locker, stringTable.table(), buffer);
}

PassRefPtr<StringImpl> AtomicString::addSlowCase(AtomicStringTable& stringTable, const LChar* characters, unsigned length)
{
    if (UNLIKELY(!characters))
        return nullptr;

    if (UNLIKELY(!length))
        return StringImpl::empty();

    AtomicStringTableLocker locker;
    LCharBuffer buffer(characters, length);
    return addToStringTable<LCharBuffer, LCharBufferTranslator>(locker, stringTable.table(), buffer);
}

DLL_PUBLIC PassRefPtr<StringImpl> AtomicString::addFromLiteralData(const char* characters, unsigned length)
{
    ASSERT(characters);
    ASSERT(length);

    LCharBuffer buffer(reinterpret_cast<const LChar*>(characters), length);
    return addToStringTable<LCharBuffer, BufferFromStaticDataTranslator<LChar>>(buffer);
}

PassRefPtr<StringImpl> AtomicString::addFromLiteralDataSlowCase(AtomicStringTable& stringTable, const char* characters, unsigned length)
{
    ASSERT(characters);
    ASSERT(length);

    AtomicStringTableLocker locker;
    LCharBuffer buffer(reinterpret_cast<const LChar*>(characters), length);
    return addToStringTable<LCharBuffer, BufferFromStaticDataTranslator<LChar>>(locker, stringTable.table(), buffer);
}

static inline PassRefPtr<StringImpl> addSymbol(AtomicStringTableLocker& locker, StringTableImpl& atomicStringTable, StringImpl& base)
{
    ASSERT(base.length());
    ASSERT(base.isSymbol());

    SubstringLocation buffer = { &base, 0, base.length() };
    if (base.is8Bit())
        return addToStringTable<SubstringLocation, SubstringTranslator8>(locker, atomicStringTable, buffer);
    return addToStringTable<SubstringLocation, SubstringTranslator16>(locker, atomicStringTable, buffer);
}

static inline PassRefPtr<StringImpl> addSymbol(StringImpl& base)
{
    AtomicStringTableLocker locker;
    return addSymbol(locker, stringTable(), base);
}

static PassRefPtr<StringImpl> addStatic(AtomicStringTableLocker& locker, StringTableImpl& atomicStringTable, StringImpl& base)
{
    ASSERT(base.length());
    ASSERT(base.isStatic());

    if (base.is8Bit()) {
        LCharBuffer buffer(base.characters8(), base.length(), base.hash());
        return addToStringTable<LCharBuffer, BufferFromStaticDataTranslator<LChar>>(locker, atomicStringTable, buffer);
    }
    UCharBuffer buffer(base.characters16(), base.length(), base.hash());
    return addToStringTable<UCharBuffer, BufferFromStaticDataTranslator<UChar>>(locker, atomicStringTable, buffer);
}

static inline PassRefPtr<StringImpl> addStatic(StringImpl& base)
{
    AtomicStringTableLocker locker;
    return addStatic(locker, stringTable(), base);
}

DLL_PUBLIC PassRefPtr<StringImpl> AtomicString::addSlowCase(StringImpl* string)
{
    // This check is necessary for null symbols.
    // Their length is zero, but they are not AtomicStringImpl.
    if (!string->length())
        return StringImpl::empty();

    if (string->isStatic())
        return addStatic(*string);

    if (string->isSymbol())
        return addSymbol(*string);

    ASSERT_WITH_MESSAGE(!string.isAtomic(), "AtomicString should not hit the slow case if the string is already atomic.");

    StringTableImpl::AddResult addResult = stringTable().add(string);
    if (addResult.isNewEntry) {
        ASSERT(*addResult.iterator == string);
        string->setIsAtomic(true);
    }
    return *addResult.iterator;
}

PassRefPtr<StringImpl> AtomicString::addSlowCase(AtomicStringTable& stringTable, StringImpl* string)
{
    // This check is necessary for null symbols.
    // Their length is zero, but they are not AtomicStringImpl.
    if (!string->length())
        return StringImpl::empty();

    if (string->isStatic()) {
        AtomicStringTableLocker locker;
        return addStatic(locker, stringTable.table(), *string);
    }

    if (string->isSymbol()) {
        AtomicStringTableLocker locker;
        return addSymbol(locker, stringTable.table(), *string);
    }

    ASSERT_WITH_MESSAGE(!string->isAtomic(), "AtomicString should not hit the slow case if the string is already atomic.");

    AtomicStringTableLocker locker;
    StringTableImpl::AddResult addResult = stringTable.table().add(string);

    if (addResult.isNewEntry) {
        ASSERT(*addResult.iterator == string);
        string->setIsAtomic(true);
    }

    return *addResult.iterator;
}

void AtomicString::remove(StringImpl* string)
{
    ASSERT(string->isAtomic());
    AtomicStringTableLocker locker;
    auto& atomicStringTable = stringTable();
    auto iterator = atomicStringTable.find(string);
    ASSERT_WITH_MESSAGE(iterator != atomicStringTable.end(), "The string being removed is atomic in the string table of an other thread!");
    ASSERT(string == *iterator);
    atomicStringTable.remove(iterator);
}

AtomicString AtomicString::lower() const
{
    // Note: This is a hot function in the Dromaeo benchmark.
    StringImpl* impl = this->impl();
    if (UNLIKELY(!impl))
        return *this;

    RefPtr<StringImpl> lowercasedString = impl->lower();
    if (LIKELY(lowercasedString == impl))
        return *this;

    AtomicString result;
    result.m_string = addSlowCase(lowercasedString.get());
    return result;
}

template<AtomicString::CaseConvertType type>
ALWAYS_INLINE AtomicString AtomicString::convertASCIICase() const
{
    StringImpl* impl = this->impl();
    if (UNLIKELY(!impl))
        return nullAtom;

    // Convert short strings without allocating a new StringImpl, since
    // there's a good chance these strings are already in the atomic
    // string table and so no memory allocation will be required.
    unsigned length;
    const unsigned localBufferSize = 100;
    if (impl->is8Bit() && (length = impl->length()) <= localBufferSize) {
        const LChar* characters = impl->characters8();
        unsigned failingIndex;
        for (unsigned i = 0; i < length; ++i) {
        	if (type == CaseConvertType::Lower ? UNLIKELY(isASCIIUpper(characters[i])) : LIKELY(isASCIILower(characters[i]))) {
				failingIndex = i;
                goto SlowPath;
            }
        }
        return *this;
SlowPath:
        LChar localBuffer[localBufferSize];
        for (unsigned i = 0; i < failingIndex; ++i)
            localBuffer[i] = characters[i];
        for (unsigned i = failingIndex; i < length; ++i)
            localBuffer[i] = type == CaseConvertType::Lower ? toASCIILower(characters[i]) : toASCIIUpper(characters[i]);
        return AtomicString(localBuffer, length);
    }

    RefPtr<StringImpl> convertedString = type == CaseConvertType::Lower ? impl->convertToASCIILowercase() : impl->convertToASCIIUppercase();
    if (LIKELY(convertedString.get() == impl))
        return *this;

    AtomicString result;
    result.m_string = addSlowCase(convertedString.get());
    return result;
}

AtomicString AtomicString::convertToASCIILowercase() const
{
    return convertASCIICase<CaseConvertType::Lower>();
}

AtomicString AtomicString::convertToASCIIUppercase() const
{
    return convertASCIICase<CaseConvertType::Upper>();
}

AtomicString AtomicString::number(int number)
{
    return numberToStringSigned<AtomicString>(number);
}

AtomicString AtomicString::number(unsigned number)
{
    return numberToStringUnsigned<AtomicString>(number);
}

AtomicString AtomicString::number(unsigned long number)
{
    return numberToStringUnsigned<AtomicString>(number);
}

AtomicString AtomicString::number(unsigned long long number)
{
    return numberToStringUnsigned<AtomicString>(number);
}

AtomicString AtomicString::number(float number)
{
    NumberToStringBuffer buffer;
    return AtomicString(numberToString(number, buffer));
}

AtomicString AtomicString::number(double number)
{
    NumberToStringBuffer buffer;
    return AtomicString(numberToString(number, buffer));
}

//RefPtr<AtomicStringImpl> AtomicStringImpl::lookUpSlowCase(StringImpl& string);
PassRefPtr<AtomicStringImpl> AtomicString::findSlowCase(StringImpl& string)
{
    ASSERT_WITH_MESSAGE(!string.isAtomic(), "AtomicStringImpls should return from the fast case.");

    if (!string.length())
        return static_cast<AtomicStringImpl*>(StringImpl::empty());

    AtomicStringTableLocker locker;
    auto& atomicStringTable = stringTable();
    auto iterator = atomicStringTable.inlineFind(&string);
    if (iterator != atomicStringTable.end())
        return static_cast<AtomicStringImpl*>(*iterator);
    return nullptr;
}

AtomicString AtomicString::fromUTF8Internal(const char* start, const char* end)
{
    ASSERT(start);

    // Caller needs to handle empty string.
    ASSERT(!end || end > start);
    ASSERT(end || start[0]);

    HashAndUTF8Characters buffer;
    buffer.characters = start;
    buffer.hash = calculateStringHashAndLengthFromUTF8MaskingTop8Bits(start, end ? end : start + strlen(start), buffer.length, buffer.utf16Length);

    if (!buffer.hash)
        return nullAtom;

    AtomicString atomicString;
    atomicString.m_string = addToStringTable<HashAndUTF8Characters, HashAndUTF8CharactersTranslator>(buffer);
    return atomicString;
}

PassRefPtr<AtomicStringImpl> AtomicString::findInternal(const LChar* characters, unsigned length)
{
    if (UNLIKELY(!characters))
        return nullptr;

    if (UNLIKELY(!length))
        return nullptr;

    AtomicStringTableLocker locker;
    StringTableImpl& table = stringTable();

    LCharBuffer buffer(characters, length);
    auto iterator = table.find<LCharBufferTranslator>(buffer);
    if (iterator != table.end())
        return static_cast<AtomicStringImpl*>(*iterator);
    return nullptr;
}

PassRefPtr<AtomicStringImpl> AtomicString::findInternal(const UChar* characters, unsigned length)
{
    if (UNLIKELY(!characters))
        return nullptr;

    if (UNLIKELY(!length))
        return nullptr;

    AtomicStringTableLocker locker;
    StringTableImpl& table = stringTable();

    UCharBuffer buffer(characters, length);
    auto iterator = table.find<UCharBufferTranslator>(buffer);
    if (iterator != table.end())
        return static_cast<AtomicStringImpl*>(*iterator);
    return nullptr;
}

#ifndef NDEBUG
void AtomicString::show() const
{
    m_string.show();
}
#endif

} // namespace WTF
