/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FetchBody.h"

#if ENABLE(FETCH_API)

#include "Dictionary.h"
#include "Document.h"
#include "FetchBodyOwner.h"
#include "FetchBodySource.h"
#include "FetchHeaders.h"
#include "FormData.h"
#include "HTTPHeaderValues.h"
#include "HTTPParsers.h"
#include "JSBlob.h"
#include "JSDOMFormData.h"
#include "JSDOMPromiseDeferred.h"
#include "JSURLSearchParams.h"
#include "ReadableStream.h"
#include "ReadableStreamSource.h"
#include "TextResourceDecoder.h"
#include "WebKitBlobBuilder.h"
#include "runtime/ArrayBuffer.h"
#include "runtime/ArrayBufferView.h"
#include "runtime/Error.h"

namespace WebCore {

FetchBody::FetchBody(PassRefPtr<Blob> blob)
    : m_blob(blob)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(PassRefPtr<FormData> formData)
    : m_formData(formData)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(const String& text)
    : m_text(text)
	, m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::~FetchBody()
{
}

FetchBody::FetchBody(PassRefPtr<ArrayBuffer> data)
    : m_arrayBuffer(data)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(PassRefPtr<ArrayBufferView> dataView)
    : m_dataView(dataView)
    , m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(PassRefPtr<URLSearchParams> url)
    : m_URLSearchParams(url)
	, m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(PassRefPtr<ReadableStream> readableStream)
    : m_readableStream(readableStream)
	, m_consumer(FetchBodyConsumer::Type::None)
{
}

FetchBody::FetchBody(const FetchBodyConsumer& consumer)
    : m_consumer(consumer)
{
}

PassOwnPtr<FetchBody> FetchBody::extract(JSC::ExecState& state, ScriptExecutionContext& context, JSC::JSValue value, String& contentType, Optional<Exception>& exception)
{
	ASSERT(scriptExecutionContext());
    if (value.isObject()) {
    	if(asObject(value)->inherits(JSBlob::info())) {
    		Blob* blob = toBlob(value);
    	    if (!blob->type().isEmpty())
    	    	contentType = blob->type();
    		return adoptPtr(new FetchBody(blob));
    	}
		if (asObject(value)->inherits(JSC::JSArrayBufferView::info())) {
			RefPtr<JSC::ArrayBufferView> buffer = toArrayBufferView(value);
		    ASSERT(buffer);
		    return adoptPtr(new FetchBody(buffer.release()));
		}
		if (asObject(value)->inherits(JSC::JSArrayBuffer::info())) {
			auto* buffer = toArrayBuffer(value);
		    ASSERT(buffer);
		    return adoptPtr(new FetchBody(buffer));
		}
		if (asObject(value)->inherits(JSDOMFormData::info())) {
			FormDataList* domFormData = toDOMFormData(value);
			RefPtr<FormData> formData = FormData::createMultiPart(*domFormData, domFormData->encoding(), &static_cast<Document&>(context));
			contentType = makeString("multipart/form-data;boundary=", formData->boundary().data());
			return adoptPtr(new FetchBody(formData.release()));
		}
		if (asObject(value)->inherits(JSURLSearchParams::info())) {
		    contentType = HTTPHeaderValues::formURLEncodedContentType();
		    return adoptPtr(new FetchBody(toURLSearchParams(value)));
		}
		if (asObject(value)->inherits(JSReadableStream::info())) {
			JSReadableStream* jsReadableStream = JSC::jsCast<JSReadableStream*>(value);
			RefPtr<ReadableStream> stream = ReadableStream::create(*jsReadableStream->globalObject(), *jsReadableStream);
	        if (stream->isDisturbed()) {
	        	exception = Exception(TypeError, "Input body is disturbed.");
	        	return nullptr;
	        }
	        if (stream->isLocked()) {
	        	exception = Exception(TypeError, "Input body is locked.");
	        	return nullptr;
	        }

		    return adoptPtr(new FetchBody(stream.release()));
		}
    }

    String body = valueToUSVString(&state, value);
    if(state.vm().exception())
    	return nullptr;
	contentType = HTTPHeaderValues::textPlainContentType();
	return adoptPtr(new FetchBody(body));
}

void FetchBody::arrayBuffer(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    m_consumer.setType(FetchBodyConsumer::Type::ArrayBuffer);
    consume(owner, promise);
}

void FetchBody::blob(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise, const String& contentType)
{
    m_consumer.setType(FetchBodyConsumer::Type::Blob);
    m_consumer.setContentType(Blob::normalizedContentType(extractMIMETypeFromMediaType(contentType)));
    consume(owner, promise);
}

void FetchBody::json(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    if (isText()) {
        fulfillPromiseWithJSON(promise, m_text);
        return;
    }
    m_consumer.setType(FetchBodyConsumer::Type::JSON);
    consume(owner, promise);
}

void FetchBody::text(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    if (isText()) {
        promise->resolve(m_text);
        return;
    }
    m_consumer.setType(FetchBodyConsumer::Type::Text);
    consume(owner, promise);
}

void FetchBody::formData(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    m_consumer.setType(FetchBodyConsumer::Type::FormData);
    consume(owner, promise);
}

void FetchBody::consumeOnceLoadingFinished(FetchBodyConsumer::Type type, PassRefPtr<DeferredPromise> promise, const String& contentType)
{
    m_consumer.setType(type);
    m_consumer.setConsumePromise(promise);
    if (type == FetchBodyConsumer::Type::Blob)
        m_consumer.setContentType(Blob::normalizedContentType(extractMIMETypeFromMediaType(contentType)));
}

void FetchBody::consume(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    if (isArrayBuffer()) {
        consumeArrayBuffer(owner, promise);
        return;
    }
    if (isArrayBufferView()) {
        consumeArrayBufferView(owner, promise);
        return;
    }
    if (isText()) {
        consumeText(owner, promise, textBody());
        return;
    }
    if(isURLSearchParams()) {
        consumeText(owner, promise, urlSearchParamsBody().toString());
        return;
    }
    if (isBlob()) {
        consumeBlob(owner, promise);
        return;
    }
    if (isFormData()) {
        consumeFormData(owner, promise);
        return;
    }

    m_consumer.resolve(promise, owner.contentType(), m_readableStream.get());
}

#if ENABLE(STREAMS_API)
void FetchBody::consumeAsStream(FetchBodyOwner& owner, FetchBodySource& source)
{
    bool closeStream = false;
    if (isArrayBuffer()) {
        ASSERT(m_arrayBuffer);
        closeStream = source.enqueue(ArrayBuffer::tryCreate(m_arrayBuffer->data(), m_arrayBuffer->byteLength()));
        m_arrayBuffer = nullptr;
    } else if (isArrayBufferView()) {
        ASSERT(m_dataView);
        closeStream = source.enqueue(ArrayBuffer::tryCreate(m_dataView->baseAddress(), m_dataView->byteLength()));
        m_dataView = nullptr;
    } else if (isText()) {
        auto data = UTF8Encoding().encode(textBody(), EntitiesForUnencodables);
        closeStream = source.enqueue(ArrayBuffer::tryCreate(data.data(), data.size()));
        m_text = String();
    } else if(isURLSearchParams()) {
    	auto data = UTF8Encoding().encode(urlSearchParamsBody().toString(), EntitiesForUnencodables);
    	closeStream = source.enqueue(ArrayBuffer::tryCreate(data.data(), data.size()));
        m_URLSearchParams = nullptr;
    } else if (isBlob()) {
        ASSERT(m_blob);
        owner.loadBlob(*m_blob, nullptr);
        m_blob = nullptr;
    } else if (isFormData())
        source.error(Exception(NOT_SUPPORTED_ERR, "Not implemented"));
    else if (m_consumer.hasData())
        closeStream = source.enqueue(m_consumer.takeAsArrayBuffer());
    else
        closeStream = true;

    if (closeStream)
        source.close();
}
#endif

void FetchBody::consumeArrayBuffer(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    ASSERT(m_arrayBuffer);
	m_consumer.resolveWithData(promise, owner.contentType(), static_cast<const uint8_t*>(m_arrayBuffer->data()), m_arrayBuffer->byteLength());
	m_arrayBuffer = nullptr;
}

void FetchBody::consumeArrayBufferView(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    ASSERT(m_dataView);
    m_consumer.resolveWithData(promise, owner.contentType(), static_cast<const uint8_t*>(m_dataView->baseAddress()), m_dataView->byteLength());
    m_dataView = nullptr;
}

void FetchBody::consumeText(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise, const String& text)
{
    auto data = UTF8Encoding().encode(text, EntitiesForUnencodables);
	m_consumer.resolveWithData(promise, owner.contentType(), reinterpret_cast<const uint8_t*>(data.data()), data.size());
	m_text = String();
}

void FetchBody::consumeBlob(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    ASSERT(m_blob);

    m_consumer.setConsumePromise(promise);
    owner.loadBlob(*m_blob, &m_consumer);
    m_blob = nullptr;
}

void FetchBody::consumeFormData(FetchBodyOwner& owner, PassRefPtr<DeferredPromise> promise)
{
    if (RefPtr<SharedBuffer> sharedBuffer = formDataBody().asSharedBuffer()) {
        m_consumer.resolveWithData(promise, owner.contentType(), reinterpret_cast<const uint8_t*>(sharedBuffer->data()), sharedBuffer->size());
        m_formData = nullptr;
    } else {
        // FIXME: If the form data contains blobs, load them like we do other blobs.
        // That will fix the last WPT test in response-consume.html.
        promise->reject<ExceptionCode>(NOT_SUPPORTED_ERR);
    }
}

void FetchBody::loadingFailed(const Exception& exception)
{
    m_consumer.loadingFailed(exception);
}

void FetchBody::loadingSucceeded(const String& contentType)
{
    m_consumer.loadingSucceeded(contentType);
}

RefPtr<FormData> FetchBody::bodyAsFormData(ScriptExecutionContext& context) const
{
    if (isText())
        return FormData::create(UTF8Encoding().encode(m_text, EntitiesForUnencodables));
    if(isURLSearchParams())
        return FormData::create(UTF8Encoding().encode(urlSearchParamsBody().toString(), EntitiesForUnencodables));
    if (isBlob()) {
		ASSERT(m_blob);
        RefPtr<FormData> body = FormData::create();
        body->appendBlob(m_blob->url());
        return body;
    }
    if (isArrayBuffer())
        return FormData::create(m_arrayBuffer->data(), m_arrayBuffer->byteLength());
    if (isArrayBufferView())
        return FormData::create(m_dataView->baseAddress(), m_dataView->byteLength());
    if (isFormData()) {
        ASSERT(m_formData);
        ASSERT(!context.isWorkerGlobalScope());
        auto body = m_formData;
        body->generateFiles(static_cast<Document*>(&context));
        return body;
    }
	ASSERT_NOT_REACHED();
	return nullptr;
}

void FetchBody::setAsFormData(PassRefPtr<FormData> data)
{
	m_blob = nullptr;
	m_formData = nullptr;
	m_arrayBuffer = nullptr;
	m_dataView = nullptr;
	m_URLSearchParams = nullptr;
	m_text = String();

	m_formData = data;
}

#if 0
FetchBody::TakenData FetchBody::take()
{
    if (m_consumer.hasData()) {
        auto buffer = m_consumer.takeData();
        if (!buffer)
            return nullptr;
        return buffer.releaseNonNull();
    }

    if (isBlob()) {
        auto body = FormData::create();
        body->appendBlob(blobBody().url());
        return WTFMove(body);
    }

    if (isFormData())
        return formDataBody();

    if (isText()) {
        auto data = UTF8Encoding().encode(textBody(), EntitiesForUnencodables);
        return SharedBuffer::create(data.data(), data.length());
    }

    if (isURLSearchParams()) {
        auto data = UTF8Encoding().encode(urlSearchParamsBody().toString(), EntitiesForUnencodables);
        return SharedBuffer::create(data.data(), data.length());
    }

    if (isArrayBuffer())
        return SharedBuffer::create(reinterpret_cast<const char*>(arrayBufferBody().data()), arrayBufferBody().byteLength());
    if (isArrayBufferView())
        return SharedBuffer::create(reinterpret_cast<const uint8_t*>(arrayBufferViewBody().baseAddress()), arrayBufferViewBody().byteLength());

    return nullptr;
}
#endif

PassOwnPtr<FetchBody> FetchBody::clone()
{
    OwnPtr<FetchBody> clone = adoptPtr(new FetchBody(m_consumer));

    if (isArrayBuffer())
        clone->m_arrayBuffer = m_arrayBuffer;
    else if (isArrayBufferView())
        clone->m_dataView = m_dataView;
    else if (isBlob())
        clone->m_blob = m_blob;
    else if (isFormData())
        clone->m_formData = m_formData;
    else if (isText())
        clone->m_text = m_text;
    else if(isURLSearchParams())
        clone->m_URLSearchParams = m_URLSearchParams;

    if (m_readableStream) {
        if (auto clones = m_readableStream->tee()) {
        	m_readableStream = clones->first.release();
        	clone->m_readableStream = clones->second.release();
        }
    }
    return clone.release();
}

}

#endif // ENABLE(FETCH_API)
