/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(FETCH_API)

#include "Blob.h"
#include "DOMFormData.h"
#include "Exception.h"
#include "FetchBodyConsumer.h"
#include "FetchLoader.h"
#include "FormData.h"
#include "ReadableStream.h"
#include "URLSearchParams.h"
#include <wtf/FastAllocBase.h>
#include <wtf/Optional.h>
#include <wtf/PassOwnPtr.h>

namespace JSC {
class ArrayBuffer;
class ArrayBufferView;
};

namespace WebCore {

class DOMFormData;
class DeferredPromise;
class FetchBodyOwner;
class FetchHeaders;
class FetchBodySource;
class ScriptExecutionContext;

class FetchBody {
	WTF_MAKE_FAST_ALLOCATED;
public:
    void arrayBuffer(FetchBodyOwner&, PassRefPtr<DeferredPromise>);
    void blob(FetchBodyOwner&, PassRefPtr<DeferredPromise>, const String&);
    void json(FetchBodyOwner&, PassRefPtr<DeferredPromise>);
    void text(FetchBodyOwner&, PassRefPtr<DeferredPromise>);
    void formData(FetchBodyOwner&, PassRefPtr<DeferredPromise> promise);

#if ENABLE(STREAMS_API)
    void consumeAsStream(FetchBodyOwner&, FetchBodySource&);
#endif

    bool isBlob() const { return m_blob.get(); }
    bool isFormData() const { return m_formData.get(); }
    bool isArrayBuffer() const { return m_arrayBuffer.get(); }
    bool isArrayBufferView() const { return m_dataView.get(); }
    bool isURLSearchParams() const { return m_URLSearchParams.get(); }
    bool isText() const { return !m_text.isNull(); }
    bool hasReadableStream() const { return !!m_readableStream; }

    //using BindingDataType = Variant<RefPtr<Blob>, RefPtr<ArrayBufferView>, RefPtr<ArrayBuffer>, RefPtr<DOMFormData>, RefPtr<URLSearchParams>, RefPtr<ReadableStream>, String>;
    static PassOwnPtr<FetchBody> extract(JSC::ExecState&, ScriptExecutionContext&, JSC::JSValue value, String&, Optional<Exception>&);

    FetchBody() : m_consumer(FetchBodyConsumer::Type::None) { }

    ~FetchBody();

    void loadingFailed(const Exception&);
    void loadingSucceeded(const String& contentType);

    RefPtr<FormData> bodyAsFormData(ScriptExecutionContext&) const;

    //using TakenData = Variant<std::nullptr_t, Ref<FormData>, Ref<SharedBuffer>>;
    //TakenData take();

    void setAsFormData(PassRefPtr<FormData>);
    FetchBodyConsumer& consumer() { return m_consumer; }

    void consumeOnceLoadingFinished(FetchBodyConsumer::Type, PassRefPtr<DeferredPromise>, const String&);
    void cleanConsumer() { m_consumer.clean(); }

	PassOwnPtr<FetchBody> clone();
    const ReadableStream* readableStream() const { return m_readableStream.get(); }
    ReadableStream* readableStream() { return m_readableStream.get(); }
    void setReadableStream(PassRefPtr<ReadableStream> stream)
    {
        ASSERT(!m_readableStream);
        m_readableStream = stream;
    }

private:
	explicit FetchBody(PassRefPtr<Blob>);
	explicit FetchBody(PassRefPtr<JSC::ArrayBuffer>);
	explicit FetchBody(PassRefPtr<JSC::ArrayBufferView>);
	explicit FetchBody(PassRefPtr<FormData>);
	explicit FetchBody(const String&);
	explicit FetchBody(PassRefPtr<URLSearchParams>);
	explicit FetchBody(PassRefPtr<ReadableStream>);
	explicit FetchBody(const FetchBodyConsumer& consumer);

    void consume(FetchBodyOwner&, PassRefPtr<DeferredPromise>);

    void consumeArrayBuffer(FetchBodyOwner&, PassRefPtr<DeferredPromise>);
    void consumeArrayBufferView(FetchBodyOwner&, PassRefPtr<DeferredPromise>);
    void consumeText(FetchBodyOwner&, PassRefPtr<DeferredPromise>, const String&);
    void consumeBlob(FetchBodyOwner&, PassRefPtr<DeferredPromise>);
    void consumeFormData(FetchBodyOwner&, PassRefPtr<DeferredPromise>);

    const Blob& blobBody() const { return *m_blob; }
    FormData& formDataBody() { return *m_formData; }
    const FormData& formDataBody() const { return *m_formData; }
    const JSC::ArrayBuffer& arrayBufferBody() const { return *m_arrayBuffer; }
    const JSC::ArrayBufferView& arrayBufferViewBody() const { return *m_dataView; }
    String& textBody() { return m_text; }
    const String& textBody() const { return m_text; }
    const URLSearchParams& urlSearchParamsBody() const { return *m_URLSearchParams; }

    // std::experimental::variant<std::nullptr_t, Ref<const Blob>, Ref<FormData>, Ref<const ArrayBuffer>, Ref<const ArrayBufferView>, Ref<const URLSearchParams>, String> m_data { nullptr };
    RefPtr<Blob> m_blob;
    RefPtr<FormData> m_formData;
    RefPtr<JSC::ArrayBuffer> m_arrayBuffer;
    RefPtr<JSC::ArrayBufferView> m_dataView;
    RefPtr<URLSearchParams> m_URLSearchParams;
    String m_text;

    FetchBodyConsumer m_consumer;
    RefPtr<ReadableStream> m_readableStream;
};

} // namespace WebCore

#endif // ENABLE(FETCH_API)
