/*
 * Copyright (C) 2016 Apple Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FetchBodyConsumer.h"

#if ENABLE(FETCH_API)

#include "DOMFormData.h"
#include "HTTPHeaderField.h"
#include "HTTPParsers.h"
#include "JSBlob.h"
#include "JSDOMFormData.h"
#include "JSDOMPromiseDeferred.h"
#include "ReadableStreamChunk.h"
#include "TextResourceDecoder.h"
#include "URLParser.h"
#include <runtime/ArrayBuffer.h>
#include <wtf/StringExtras.h>

namespace WebCore {

static inline PassRefPtr<Blob> blobFromData(ScriptExecutionContext* context, const unsigned char* data, unsigned length, const String& contentType)
{
    return Blob::create(context, data, length, Blob::normalizedContentType(contentType));
}

// https://mimesniff.spec.whatwg.org/#http-quoted-string-token-code-point
static bool isHTTPQuotedStringTokenCodePoint(UChar c)
{
    return c == 0x09
        || (c >= 0x20 && c <= 0x7E)
        || (c >= 0x80 && c <= 0xFF);
}

struct MimeType {
    String type;
    String subtype;
    HashMap<String, String> parameters;
};

static HashMap<String, String> parseParameters(StringView input, size_t position)
{
    HashMap<String, String> parameters;
    while (position < input.length()) {
        while (position < input.length() && RFC7230::isWhitespace(input[position]))
            position++;
        size_t nameBegin = position;
        while (position < input.length() && input[position] != '=' && input[position] != ';')
            position++;
        if (position >= input.length())
            break;
        if (input[position] == ';') {
            position++;
            continue;
        }
        StringView parameterName = input.substring(nameBegin, position - nameBegin);
        position++;
        if (position >= input.length())
            break;

        StringView parameterValue;
        if (position < input.length() && input[position] == '"') {
            position++;
            size_t valueBegin = position;
            while (position < input.length() && input[position] != '"')
                position++;
            parameterValue = input.substring(valueBegin, position - valueBegin);
            position++;
        } else {
            size_t valueBegin = position;
            while (position < input.length() && input[position] != ';')
                position++;
            parameterValue = stripLeadingAndTrailingHTTPSpaces(input.substring(valueBegin, position - valueBegin));
        }

        if (parameterName.length()
            && isValidHTTPToken(parameterName)
            && parameterValue.isAllSpecialCharacters<isHTTPQuotedStringTokenCodePoint>()) {
            String nameString = parameterName.toString();
            if (!parameters.contains(nameString))
                parameters.set(nameString, parameterValue.toString());
        }
    }
    return parameters;
}

// https://mimesniff.spec.whatwg.org/#parsing-a-mime-type
static Optional<MimeType> parseMIMEType(const String& contentType)
{
    String input = stripLeadingAndTrailingHTTPSpaces(contentType);
    size_t slashIndex = input.find('/');
    if (slashIndex == notFound)
        return WTF::Nullopt;

    String type = input.substring(0, slashIndex);
    if (!type.length() || !isValidHTTPToken(type))
        return WTF::Nullopt;

    size_t semicolonIndex = input.find(';', slashIndex);
    String subtype = stripLeadingAndTrailingHTTPSpaces(input.substring(slashIndex + 1, semicolonIndex - slashIndex - 1));
    if (!subtype.length() || !isValidHTTPToken(subtype))
        return WTF::Nullopt;

    return MimeType { WTFMove(type), WTFMove(subtype), parseParameters(StringView(input), semicolonIndex + 1) };
}

// https://fetch.spec.whatwg.org/#concept-body-package-data
RefPtr<DOMFormData> FetchBodyConsumer::packageFormData(ScriptExecutionContext* context, const String& contentType, const char* data, size_t length)
{
    auto parseMultipartPart = [] (ScriptExecutionContext* context, const char* part, size_t partLength, DOMFormData& form) -> bool {
        const char* headerEnd = static_cast<const char*>(memmem(part, partLength, "\r\n\r\n", 4));
        if (!headerEnd)
            return false;
        const char* headerBegin = part;
        size_t headerLength = headerEnd - headerBegin;

        const char* bodyBegin = headerEnd + strlen("\r\n\r\n");
        size_t bodyLength = partLength - (bodyBegin - headerBegin);

        String header = String::fromUTF8(headerBegin, headerLength);

        ASCIILiteral contentDispositionCharacters("Content-Disposition:");
        size_t contentDispositionBegin = header.find(contentDispositionCharacters);
        if (contentDispositionBegin == notFound)
            return false;
        size_t contentDispositionEnd = header.find("\r\n", contentDispositionBegin);
        size_t contentDispositionParametersBegin = header.find(';', contentDispositionBegin + contentDispositionCharacters.length());
        if (contentDispositionParametersBegin != notFound)
            contentDispositionParametersBegin++;

        auto parameters = parseParameters(StringView(header).substring(contentDispositionParametersBegin, contentDispositionEnd - contentDispositionParametersBegin), 0);
        String name = parameters.get("name");
        if (!name)
            return false;
        String filename = parameters.get("filename");
        if (!filename)
            form.append(name, String::fromUTF8(bodyBegin, bodyLength));
        else {
            String contentType = "text/plain";

            ASCIILiteral contentTypeCharacters("Content-Type:");
            size_t contentTypePrefixLength = contentTypeCharacters.length();
            size_t contentTypeBegin = header.find(contentTypeCharacters);
            if (contentTypeBegin != notFound) {
                size_t contentTypeEnd = header.find("\r\n", contentTypeBegin);
                contentType = stripLeadingAndTrailingHTTPSpaces(header.substring(contentTypeBegin + contentTypePrefixLength, contentTypeEnd - contentTypeBegin - contentTypePrefixLength));
            }

            form.append(name, Blob::create(context, SharedBuffer::create(bodyBegin, bodyLength).get(), Blob::normalizedContentType(contentType)), filename);
        }
        return true;
    };

    auto parseMultipartBoundary = [] (const Optional<MimeType>& mimeType) -> Optional<String> {
        if (!mimeType)
            return WTF::Nullopt;
        if (equalIgnoringASCIICase(mimeType->type, "multipart") && equalIgnoringASCIICase(mimeType->subtype, "form-data")) {
            auto iterator = mimeType->parameters.find("boundary");
            if (iterator != mimeType->parameters.end())
                return iterator->second;
        }
        return WTF::Nullopt;
    };

    RefPtr<DOMFormData> form = DOMFormData::create(UTF8Encoding());
    auto mimeType = parseMIMEType(contentType);
    if (auto multipartBoundary = parseMultipartBoundary(mimeType)) {
        String boundaryWithDashes = makeString("--", *multipartBoundary);
        CString boundary = boundaryWithDashes.utf8();
        size_t boundaryLength = boundary.length();

        const char* currentBoundary = static_cast<const char*>(memmem(data, length, boundary.data(), boundaryLength));
        if (!currentBoundary)
            return nullptr;
        const char* nextBoundary = static_cast<const char*>(memmem(currentBoundary + boundaryLength, length - (currentBoundary + boundaryLength - data), boundary.data(), boundaryLength));
        if (!nextBoundary)
            return nullptr;
        while (nextBoundary) {
            parseMultipartPart(context, currentBoundary + boundaryLength, nextBoundary - currentBoundary - boundaryLength - strlen("\r\n"), *form);
            currentBoundary = nextBoundary;
            nextBoundary = static_cast<const char*>(memmem(nextBoundary + boundaryLength, length - (nextBoundary + boundaryLength - data), boundary.data(), boundaryLength));
        }
    } else if (mimeType && equalIgnoringASCIICase(mimeType->type, "application") && equalIgnoringASCIICase(mimeType->subtype, "x-www-form-urlencoded")) {
        auto dataString = String::fromUTF8(data, length);
        for (auto& pair : URLParser::parseURLEncodedForm(dataString))
            form->append(pair.first, pair.second);
    } else
        return nullptr;

    return form;
}

static inline bool shouldPrependBOM(const unsigned char* data, unsigned length)
{
    if (length < 3)
        return true;
    return data[0] != 0xef || data[1] != 0xbb || data[2] != 0xbf;
}

static String textFromUTF8(const unsigned char* data, unsigned length)
{
    auto decoder = TextResourceDecoder::create("text/plain", "UTF-8");
    if (shouldPrependBOM(data, length))
        decoder->decode("\xef\xbb\xbf", 3);
    return decoder->decodeAndFlush(reinterpret_cast<const char*>(data), length);
}

static void resolveWithTypeAndData(PassRefPtr<DeferredPromise> promise, FetchBodyConsumer::Type type, const String& contentType, const unsigned char* data, unsigned length)
{
	ScriptExecutionContext* context = promise->scriptExecutionContext();

    switch (type) {
    case FetchBodyConsumer::Type::ArrayBuffer:
        fulfillPromiseWithArrayBuffer(promise, JSC::ArrayBuffer::tryCreate(data, length).get());
        return;
    case FetchBodyConsumer::Type::Blob:
        promise->resolve(blobFromData(context, data, length, contentType).get());
        return;
    case FetchBodyConsumer::Type::JSON:
        fulfillPromiseWithJSON(promise, textFromUTF8(data, length));
        return;
    case FetchBodyConsumer::Type::Text:
        promise->resolve(textFromUTF8(data, length));
        return;
    case FetchBodyConsumer::Type::FormData:
        if (auto formData = FetchBodyConsumer::packageFormData(context, contentType, reinterpret_cast<const char*>(data), length))
            promise->resolve(formData.get());
        else
            promise->reject(TypeError);
        return;
    case FetchBodyConsumer::Type::None:
        ASSERT_NOT_REACHED();
        return;
    }
}

void FetchBodyConsumer::resolveWithData(PassRefPtr<DeferredPromise> promise, const String& contentType, const unsigned char* data, unsigned length)
{
    resolveWithTypeAndData(promise, m_type, contentType, data, length);
}

#if 0
void FetchBodyConsumer::resolveWithFormData(Ref<DeferredPromise>&& promise, const String& contentType, const FormData& formData, ScriptExecutionContext* context)
{
    if (auto sharedBuffer = formData.asSharedBuffer()) {
        resolveWithData(WTFMove(promise), contentType, sharedBuffer->makeContiguous()->data(), sharedBuffer->size());
        return;
    }

    if (!context)
        return;

    m_formDataConsumer = makeUnique<FormDataConsumer>(formData, *context, [this, promise = WTFMove(promise), contentType, builder = SharedBufferBuilder { }](auto&& result) mutable {
        if (result.hasException()) {
            auto protectedPromise = WTFMove(promise);
            protectedPromise->reject(result.releaseException());
            return;
        }

        auto& value = result.returnValue();
        if (value.empty()) {
        	auto protectedPromise = WTFMove(promise);
            auto buffer = builder.takeAsContiguous();
            resolveWithData(WTFMove(protectedPromise), contentType, buffer->data(), buffer->size());
            return;
        }

        builder.append(value);
    });
}

void FetchBodyConsumer::consumeFormDataAsStream(const FormData& formData, FetchBodySource& source, ScriptExecutionContext* context)
{
    if (auto sharedBuffer = formData.asSharedBuffer()) {
        if (source.enqueue(ArrayBuffer::tryCreate(sharedBuffer->makeContiguous()->data(), sharedBuffer->size())))
            source.close();
        return;
    }

    if (!context)
        return;

    m_formDataConsumer = makeUnique<FormDataConsumer>(formData, *context, [this, source = Ref { source }](auto&& result) {
        if (result.hasException()) {
            source->error(result.releaseException());
            return;
        }

        auto& value = result.returnValue();
        if (value.empty()) {
            source->close();
            return;
        }

        if (!source->enqueue(ArrayBuffer::tryCreate(value.data(), value.size())))
            m_formDataConsumer->cancel();
    });
}
#endif

void FetchBodyConsumer::extract(ReadableStream& stream, PassOwnPtr<ReadableStreamToSharedBufferSink::Callback> callback)
{
    ASSERT(!m_sink);
    m_sink = ReadableStreamToSharedBufferSink::create(callback);
    m_sink->pipeFrom(stream);
}

void FetchBodyConsumer::clean()
{
    m_buffer = nullptr;
    m_consumePromise = nullptr;
    if (m_sink) {
        m_sink->clearCallback();
        return;
    }
}

void FetchBodyConsumer::resolve(PassRefPtr<DeferredPromise> promise, const String& contentType, ReadableStream* stream)
{
    if (stream) {
        ASSERT(!m_sink);
        class ResolveCallback final : public ReadableStreamToSharedBufferSink::Callback {
        public:
        	ResolveCallback(PassRefPtr<DeferredPromise> promise, Type type, const String& contentType)
				: promise(promise)
				, type(type)
				, contentType(contentType)
				, data(SharedBuffer::create())
			{
			}
    		virtual void operator()(ReadableStreamChunk* chunk, ExceptionCode ec, const String& message) override
    		{
    			if (ec) {
    				if (RefPtr<DeferredPromise> protectedPromise = promise.release())
    					protectedPromise->reject(ec, message);
					return;
				}

                if (!chunk) {
    				if (RefPtr<DeferredPromise> protectedPromise = promise.release())
    					resolveWithTypeAndData(protectedPromise.release(), type, contentType, reinterpret_cast<const unsigned char*>(data->data()), data->size());
                	return;
                }

                if (!data->tryAppend(reinterpret_cast<const char*>(chunk->data), chunk->size)) {
    				if (RefPtr<DeferredPromise> protectedPromise = promise.release())
    					protectedPromise->reject(OperationError, "Out of memory");
					return;
				}
    		}

        private:
    	    RefPtr<DeferredPromise> promise;
    	    Type type;
    	    String contentType;
    	    Ref<SharedBuffer> data;
        };
        m_sink = ReadableStreamToSharedBufferSink::create(adoptPtr(new ResolveCallback(promise, m_type, contentType)));
        m_sink->pipeFrom(*stream);
        return;
    }

    if (m_isLoading) {
        m_consumePromise = promise;
        return;
    }

    ScriptExecutionContext* context = promise->scriptExecutionContext();

    ASSERT(m_type != Type::None);
    switch (m_type) {
    case Type::ArrayBuffer:
        fulfillPromiseWithArrayBuffer(promise, takeAsArrayBuffer().get());
        return;
    case Type::Blob:
        promise->resolve(takeAsBlob(context).get());
        return;
    case Type::JSON:
        fulfillPromiseWithJSON(promise, takeAsText());
        return;
    case Type::Text:
        promise->resolve(takeAsText());
        return;
    case FetchBodyConsumer::Type::FormData: {
    	RefPtr<SharedBuffer> buffer = takeData();
        if (RefPtr<DOMFormData> formData = packageFormData(context, contentType, buffer ? buffer->data() : nullptr, buffer ? buffer->size() : 0))
            promise->resolve(formData.get());
        else
            promise->reject(TypeError);
        return;
    }
    case Type::None:
        ASSERT_NOT_REACHED();
        return;
    }
}

void FetchBodyConsumer::append(const char* data, unsigned size)
{
    if (m_source) {
        m_source->enqueue(ArrayBuffer::tryCreate(data, size));
        return;
    }
    if (!m_buffer) {
        m_buffer = SharedBuffer::create(data, size);
        return;
    }
    m_buffer->append(data, size);
}

void FetchBodyConsumer::append(const unsigned char* data, unsigned size)
{
    append(reinterpret_cast<const char*>(data), size);
}

PassRefPtr<SharedBuffer> FetchBodyConsumer::takeData()
{
    return m_buffer.release();
}

PassRefPtr<JSC::ArrayBuffer> FetchBodyConsumer::takeAsArrayBuffer()
{
    if (!m_buffer || m_buffer->isEmpty())
        return JSC::ArrayBuffer::tryCreate(reinterpret_cast<const void*>(0), 0);

    int size = m_buffer->size();
    RefPtr<JSC::ArrayBuffer> arrayBuffer = JSC::ArrayBuffer::createAdopted(m_buffer->releaseBuffer(), size);
    m_buffer = 0;
    return arrayBuffer.release();
}

PassRefPtr<Blob> FetchBodyConsumer::takeAsBlob(ScriptExecutionContext* context)
{
    if (!m_buffer)
        return Blob::create(context, Vector<uint8_t>(), Blob::normalizedContentType(m_contentType));

    // FIXME: We should try to move m_buffer to Blob without doing extra copy.
    return blobFromData(context, reinterpret_cast<const unsigned char*>(m_buffer->data()), m_buffer->size(), m_contentType);
}

String FetchBodyConsumer::takeAsText()
{
	// FIXME: We could probably text decode on the fly as soon as m_type is set to JSON or Text.
	if (!m_buffer)
        return String();

	auto text = textFromUTF8(reinterpret_cast<const unsigned char*>(m_buffer->data()), m_buffer->size());
	m_buffer = 0;
    return text;
}

void FetchBodyConsumer::setConsumePromise(PassRefPtr<DeferredPromise> promise)
{
    ASSERT(!m_consumePromise);
    m_consumePromise = promise;
}

void FetchBodyConsumer::setSource(PassRefPtr<FetchBodySource> source)
{
    m_source = source;
    if (m_buffer) {
        int size = m_buffer->size();
        m_source->enqueue(JSC::ArrayBuffer::createAdopted(m_buffer->releaseBuffer(), size));
        m_buffer = nullptr;
    }
}

void FetchBodyConsumer::loadingFailed(const Exception& exception)
{
    m_isLoading = false;
    if (m_consumePromise) {
        m_consumePromise->reject(exception);
        m_consumePromise = nullptr;
    }
    if (m_source) {
        m_source->error(exception);
        m_source = nullptr;
    }
}

void FetchBodyConsumer::loadingSucceeded(const String& contentType)
{
    m_isLoading = false;

    if (m_consumePromise)
        resolve(m_consumePromise.releaseNonNull(), contentType, nullptr);
    if (m_source) {
		m_source->close();
		m_source = nullptr;
	}
}

} // namespace WebCore

#endif // ENABLE(FETCH_API)
