/*
 * Copyright (C) 2016 Canon Inc.
 * Copyright (C) 2020 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FetchBodyOwner.h"

#if ENABLE(FETCH_API)

#include "ExceptionCode.h"
#include "FetchLoader.h"
#include "FetchBodySource.h"
#include "HTTPParsers.h"
#include "JSBlob.h"
#include "JSDOMFormData.h"
#include "JSDOMPromiseDeferred.h"
#include "ReadableStream.h"
#include "ResourceError.h"
#include "ResourceResponse.h"

namespace WebCore {

#if 0
void FetchBodyOwner::runNetworkTaskWhenPossible(Function<void()>&& task)
{
    auto* context = scriptExecutionContext();
    if (!context || !scriptExecutionContext()->activeDOMObjectsAreSuspended())
        return task();

    context->eventLoop().queueTask(TaskSource::Networking, [pendingActivity = makePendingActivity(*this), task = WTFMove(task)] {
        task();
    });
}
#endif

FetchBodyOwner::FetchBodyOwner(ScriptExecutionContext* context, PassOwnPtr<FetchBody> body, PassRefPtr<FetchHeaders> headers)
    : ActiveDOMObject(context)
    , m_body(body)
	, m_isDisturbed(false)
	, m_headers(headers)
	, m_isBodyOpaque(false)
{
    suspendIfNeeded();
}

FetchBodyOwner::~FetchBodyOwner()
{
    if (m_readableStreamSource)
        m_readableStreamSource->detach();
}

void FetchBodyOwner::stop()
{
	if (m_body)
		m_body->cleanConsumer();

    if (m_blobLoader) {
        bool isUniqueReference = hasOneRef();
        if (m_blobLoader->loader)
            m_blobLoader->loader->stop();
        // After that point, 'this' may be destroyed, since unsetPendingActivity should have been called.
        ASSERT_UNUSED(isUniqueReference, isUniqueReference || !m_blobLoader);
    }
}

bool FetchBodyOwner::isDisturbed() const
{
    if (isBodyNull())
        return false;

    if (m_isDisturbed)
        return true;

#if ENABLE(STREAMS_API)
    if (body().readableStream())
        return body().readableStream()->isDisturbed();
#endif

    return false;
}

bool FetchBodyOwner::isDisturbedOrLocked() const
{
    if (isBodyNull())
        return false;

    if (m_isDisturbed)
        return true;

#if ENABLE(STREAMS_API)
    if (body().readableStream())
        return body().readableStream()->isDisturbed() || body().readableStream()->isLocked();
#endif

    return false;
}

void FetchBodyOwner::arrayBuffer(PassRefPtr<DeferredPromise> promise)
{
    if (auto exception = loadingException()) {
        promise->reject(*exception);
        return;
    }

    if (isBodyNullOrOpaque()) {
        fulfillPromiseWithArrayBuffer(promise, JSC::ArrayBuffer::tryCreate(reinterpret_cast<const void*>(0), 0).get());
        return;
    }
    if (isDisturbedOrLocked()) {
        promise->reject<ExceptionCode>(TypeError);
        return;
    }
    m_isDisturbed = true;
    m_body->arrayBuffer(*this, promise);
}

void FetchBodyOwner::blob(PassRefPtr<DeferredPromise> promise)
{
    if (auto exception = loadingException()) {
        promise->reject(*exception);
        return;
    }

    if (isBodyNullOrOpaque()) {
    	ScriptExecutionContext* context = promise->scriptExecutionContext();
    	Ref<Blob> blob = Blob::create(context, Vector<uint8_t>(), Blob::normalizedContentType(extractMIMETypeFromMediaType(m_contentType)));
    	promise->resolve(blob.ptr());
		return;
    }
    if (isDisturbedOrLocked()) {
        promise->reject<ExceptionCode>(TypeError);
        return;
    }
    m_isDisturbed = true;
    m_body->blob(*this, promise, m_contentType);
}

void FetchBodyOwner::cloneBody(FetchBodyOwner& owner)
{
    m_loadingError = owner.m_loadingError;

    m_contentType = owner.m_contentType;
    if (owner.isBodyNull())
        return;
    m_body = owner.m_body->clone();
}

void FetchBodyOwner::updateContentType()
{
    String contentType = m_headers->fastGet(HTTPHeaderName::ContentType);
    if (!contentType.isNull()) {
        m_contentType = contentType;
        return;
    }
    if (!m_contentType.isNull())
        m_headers->fastSet(HTTPHeaderName::ContentType, m_contentType);
}

void FetchBodyOwner::consumeOnceLoadingFinished(FetchBodyConsumer::Type type, PassRefPtr<DeferredPromise> promise)
{
    if (isDisturbedOrLocked()) {
        promise->reject<ExceptionCode>(TypeError);
        return;
    }
    m_isDisturbed = true;
    m_body->consumeOnceLoadingFinished(type, promise, m_contentType);
}

void FetchBodyOwner::formData(PassRefPtr<DeferredPromise> promise)
{
    if (auto exception = loadingException()) {
        promise->reject(*exception);
        return;
    }

    if (isDisturbedOrLocked()) {
        promise->reject(Exception(TypeError, "Body is disturbed or locked"));
        return;
    }

    if (isBodyNullOrOpaque()) {
        if (isBodyNull()) {
            // If the content-type is 'application/x-www-form-urlencoded', a body is not required and we should package an empty byte sequence as per the specification.
            if (RefPtr<DOMFormData> formData = FetchBodyConsumer::packageFormData(promise->scriptExecutionContext(), m_contentType, nullptr, 0)) {
                promise->resolve(formData.get());
                return;
            }
        }

        promise->reject(TypeError);
        return;
    }

    m_isDisturbed = true;
    m_body->formData(*this, promise);
}

void FetchBodyOwner::json(PassRefPtr<DeferredPromise> promise)
{
    if (auto exception = loadingException()) {
        promise->reject(*exception);
        return;
    }

    if (isBodyNullOrOpaque()) {
        promise->reject<ExceptionCode>(SYNTAX_ERR);
        return;
    }
    if (isDisturbedOrLocked()) {
        promise->reject<ExceptionCode>(TypeError);
        return;
    }
    m_isDisturbed = true;
    m_body->json(*this, promise);
}

void FetchBodyOwner::text(PassRefPtr<DeferredPromise> promise)
{
    if (auto exception = loadingException()) {
        promise->reject(*exception);
        return;
    }

    if (isBodyNullOrOpaque()) {
        promise->resolve(String());
        return;
    }
    if (isDisturbedOrLocked()) {
        promise->reject<ExceptionCode>(TypeError);
        return;
    }
    m_isDisturbed = true;
    m_body->text(*this, promise);
}

void FetchBodyOwner::loadBlob(Blob& blob, FetchBodyConsumer* consumer)
{
    // Can only be called once for a body instance.
    ASSERT(!m_blobLoader);
    ASSERT(!isBodyNull());

    if (!scriptExecutionContext()) {
        m_body->loadingFailed(Exception(TypeError, "Blob loading failed"));
        return;
    }

    m_blobLoader = adoptPtr(new BlobLoader(*this));
    m_blobLoader->loader = adoptPtr(new FetchLoader(*m_blobLoader, consumer));

    m_blobLoader->loader->start(*scriptExecutionContext(), blob);
    if (!m_blobLoader->loader->isStarted()) {
        m_body->loadingFailed(Exception(TypeError, "Blob loading failed"));
        m_blobLoader.clear();
        return;
    }
}

void FetchBodyOwner::finishBlobLoading()
{
    ASSERT(m_blobLoader);

    m_blobLoader.clear();
}

void FetchBodyOwner::blobLoadingSucceeded()
{
    ASSERT(!isBodyNull());
#if ENABLE(STREAMS_API)
    if (m_readableStreamSource) {
        m_readableStreamSource->close();
        m_readableStreamSource.clear();
    }
#endif

    if (m_body)
    	m_body->loadingSucceeded(contentType());
    finishBlobLoading();
}

void FetchBodyOwner::blobLoadingFailed()
{
    ASSERT(!isBodyNull());
#if ENABLE(STREAMS_API)
    if (m_readableStreamSource) {
        if (!m_readableStreamSource->isCancelling())
            m_readableStreamSource->error(Exception(TypeError, "Blob loading failed"));
        m_readableStreamSource.clear();
    }
#endif

    if (m_body)
    	m_body->loadingFailed(Exception(TypeError, "Blob loading failed"));
    finishBlobLoading();
}

void FetchBodyOwner::blobChunk(const char* data, size_t size)
{
    ASSERT(data);
#if ENABLE(STREAMS_API)
    ASSERT(m_readableStreamSource);
    if (!m_readableStreamSource || !m_readableStreamSource->enqueue(ArrayBuffer::tryCreate(data, size)))
        stop();
#else
    UNUSED_PARAM(data);
    UNUSED_PARAM(size);
#endif
}

FetchBodyOwner::BlobLoader::BlobLoader(FetchBodyOwner& owner)
    : owner(owner)
{
}

void FetchBodyOwner::BlobLoader::didReceiveResponse(const ResourceResponse& response)
{
    if (response.httpStatusCode() != 200)
        didFail(ResourceError());
}

void FetchBodyOwner::BlobLoader::didFail(const ResourceError& error)
{
    // didFail might be called within FetchLoader::start call.
    if (loader->isStarted())
        owner.blobLoadingFailed();
}

RefPtr<ReadableStream> FetchBodyOwner::readableStream(JSC::ExecState* state)
{
    if (isBodyNullOrOpaque())
        return nullptr;

    if (!m_body->hasReadableStream()) {
        createReadableStream(*state);
        if (state->vm().exception())
        	return nullptr;
    }

    return m_body->readableStream();
}

void FetchBodyOwner::createReadableStream(JSC::ExecState& state)
{
	ASSERT(!m_readableStreamSource);
	if (isDisturbed()) {
		RefPtr<ReadableStream> stream = ReadableStream::create(&state, nullptr);
		if (!stream)
			return;
		m_body->setReadableStream(stream.release());
		m_body->readableStream()->lock();
		return;
	}

	m_readableStreamSource = adoptRef(new FetchBodySource(*this));
	RefPtr<ReadableStream> stream = ReadableStream::create(&state, m_readableStreamSource);
	if (!stream) {
        m_readableStreamSource = nullptr;
		return;
	}
	m_body->setReadableStream(stream);
}

void FetchBodyOwner::consumeBodyAsStream()
{
    ASSERT(m_readableStreamSource);

    if (auto exception = loadingException()) {
        m_readableStreamSource->error(*exception);
        return;
    }

    body().consumeAsStream(*this, *m_readableStreamSource);
    if (!m_readableStreamSource->isPulling())
        m_readableStreamSource = nullptr;
}

ResourceError FetchBodyOwner::loadingError() const
{
	if(m_loadingError)
		return *m_loadingError;
	if(m_exception)
		return ResourceError(errorDomainWebKitInternal, 0, String(), m_exception->message());
	return ResourceError();
}

Optional<Exception> FetchBodyOwner::loadingException() const
{
	if(m_loadingError)
		return Exception(TypeError, m_loadingError->localizedDescription().isEmpty() ? "Loading failed" : m_loadingError->localizedDescription());
	if(m_exception)
		return m_exception;
	return Nullopt;
}

bool FetchBodyOwner::virtualHasPendingActivity() const
{
    return !!m_blobLoader;
}

void FetchBodyOwner::contextDestroyed()
{
    ASSERT(m_scriptExecutionContext);
    ActiveDOMObject::contextDestroyed();

    if (m_blobLoader) {
    	WTF::logMessage("FetchBodyOwner::contextDestroyed(): blob loader is not null.");
    	blobLoadingFailed();
    }
}

bool FetchBodyOwner::hasLoadingError() const
{
	if(m_loadingError)
		return true;
	if(m_exception)
		return true;
	return false;
}

void FetchBodyOwner::setLoadingError(const Exception& exception)
{
    if (hasLoadingError())
        return;

    m_exception = exception;
}

void FetchBodyOwner::setLoadingError(const ResourceError& error)
{
    if (hasLoadingError())
        return;

    m_loadingError = error;
}

} // namespace WebCore

#endif // ENABLE(FETCH_API)
