/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(FETCH_API)

#include "ActiveDOMObject.h"
#include "FetchBody.h"
#include "FetchBodySource.h"
#include "FetchHeaders.h"
#include "FetchLoader.h"
#include "FetchLoaderClient.h"
#include "ResourceError.h"
#include <wtf/Optional.h>
#include <wtf/OwnPtr.h>
#include <wtf/WeakPtr.h>

namespace WebCore {

class FetchBodyOwner : public RefCounted<FetchBodyOwner>, public ActiveDOMObject, public CanMakeWeakPtr<FetchBodyOwner> {
public:
    FetchBodyOwner(ScriptExecutionContext*, PassOwnPtr<FetchBody>, PassRefPtr<FetchHeaders>);
    virtual ~FetchBodyOwner();

    bool bodyUsed() const { return isDisturbed(); }
    void arrayBuffer(PassRefPtr<DeferredPromise>);
    void blob(PassRefPtr<DeferredPromise>);
    void formData(PassRefPtr<DeferredPromise>);
    void json(PassRefPtr<DeferredPromise>);
    void text(PassRefPtr<DeferredPromise>);

    bool isDisturbed() const;
    bool isDisturbedOrLocked() const;

    void loadBlob(Blob&, FetchBodyConsumer*);

	bool isActive() const { return !!m_blobLoader; }

	RefPtr<ReadableStream> readableStream(JSC::ExecState*);
    bool hasReadableStreamBody() const { return m_body && m_body->hasReadableStream(); }

#if ENABLE(STREAMS_API)
    virtual void consumeBodyAsStream();
    virtual void feedStream() { }
    virtual void cancel() { }
#endif

    bool hasLoadingError() const;
    ResourceError loadingError() const;
    Optional<Exception> loadingException() const;

    const String& contentType() const { return m_contentType; }

protected:
    const FetchBody& body() const { return *m_body; }
    FetchBody& body() { return *m_body; }
    bool isBodyNull() const { return !m_body; }
    bool isBodyNullOrOpaque() const { return !m_body || m_isBodyOpaque; }
    void cloneBody(FetchBodyOwner&);

    void updateContentType();
    void consumeOnceLoadingFinished(FetchBodyConsumer::Type, PassRefPtr<DeferredPromise>);

    void setBody(PassOwnPtr<FetchBody> body) { m_body = body; }
    void createReadableStream(JSC::ExecState&);

    // ActiveDOMObject API
    virtual void stop() OVERRIDE;

    void setDisturbed() { m_isDisturbed = true; }

    void setBodyAsOpaque() { m_isBodyOpaque = true; }
    bool isBodyOpaque() const { return m_isBodyOpaque; }

    void setLoadingError(const Exception&);
    void setLoadingError(const ResourceError&);

private:
	// Blob loading routines
	void blobChunk(const char*, size_t);
	void blobLoadingSucceeded();
	void blobLoadingFailed();
	void finishBlobLoading();

    // ActiveDOMObject API
    bool virtualHasPendingActivity() const final;
	virtual void contextDestroyed() final;

	struct BlobLoader FINAL : public FetchLoaderClient {
		BlobLoader(FetchBodyOwner&);

		// FetchLoaderClient API
		virtual void didReceiveResponse(const ResourceResponse&) FINAL;
		void didReceiveData(const char* data, size_t size) FINAL { owner.blobChunk(data, size); }
		virtual void didFail(const ResourceError&) FINAL;
		virtual void didSucceed() FINAL { owner.blobLoadingSucceeded(); }

		FetchBodyOwner& owner;
		OwnPtr<FetchLoader> loader;
	};

protected:
    OwnPtr<FetchBody> m_body;
    String m_contentType;
    bool m_isDisturbed;
#if ENABLE(STREAMS_API)
    RefPtr<FetchBodySource> m_readableStreamSource;
#endif
    RefPtr<FetchHeaders> m_headers;

private:
	OwnPtr<BlobLoader> m_blobLoader;
	bool m_isBodyOpaque;

    Optional<ResourceError> m_loadingError;
    Optional<Exception> m_exception;

};

} // namespace WebCore

#endif // ENABLE(FETCH_API)
