/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FetchHeaders.h"

#if ENABLE(FETCH_API)

#include "ExceptionCode.h"
#include "HTTPParsers.h"
#include "JSDOMConvertRecord.h"
#include "JSFetchHeaders.h"
#include <wtf/text/StringView.h>

namespace WebCore {

static bool canWriteHeader(const String& name, const String& value, FetchHeaders::Guard guard, ExceptionCode& ec)
{
    if (!isValidHTTPToken(name) || !isValidHTTPHeaderValue(value)) {
        ec = TypeError;
        return false;
    }
    if (guard == FetchHeaders::Guard::Immutable) {
        ec = TypeError;
        return false;
    }
    if (guard == FetchHeaders::Guard::Request && isForbiddenHeaderName(name))
        return false;
    if (guard == FetchHeaders::Guard::RequestNoCors && !isSimpleHeader(name, value))
        return false;
    if (guard == FetchHeaders::Guard::Response && isForbiddenResponseHeaderName(name))
        return false;
    return true;
}

static void appendToHeaderMap(const String& name, const String& value, HTTPHeaderMap& headers, FetchHeaders::Guard guard, ExceptionCode& ec)
{
    String normalizedValue = stripLeadingAndTrailingHTTPSpaces(value);
    if (!canWriteHeader(name, normalizedValue, guard, ec))
        return;
    headers.add(name, normalizedValue);
}

static void appendToHeaderMap(const HTTPHeaderMap::const_iterator& iter, HTTPHeaderMap& headers, FetchHeaders::Guard guard, ExceptionCode& ec)
{
    auto canWriteResult = canWriteHeader(iter->first, iter->second, guard, ec);
    if (ec)
        return;
	headers.add(iter->first, iter->second);
}

// https://fetch.spec.whatwg.org/#concept-headers-fill
static void fillHeaderMap(JSC::ExecState& state, HTTPHeaderMap& headers, JSC::JSValue headersInit, FetchHeaders::Guard guard, ExceptionCode& ec)
{
	JSC::VM& vm = state.vm();

	if(!headersInit.isObject()) {
		JSC::throwTypeError(&state, vm, "Header is not object");
		ec = TypeError;
		return;
	}

	JSC::JSObject* object = JSC::asObject(headersInit);

	if(JSC::isJSArray(object)) {
		JSC::JSArray* array = asArray(object);
		unsigned length = array->length();

		for (unsigned i = 0; i < length; ++i) {
			Vector<String> header = toNativeArray<String>(state, array->getIndex(vm, &state, i));
		    RETURN_IF_EXCEPTION(vm, void());
            if (header.size() != 2) {
                JSC::throwTypeError(&state, vm, "Header sub-sequence must contain exactly two items");
        		ec = TypeError;
        		return;
            }
            appendToHeaderMap(header[0], header[1], headers, guard, ec);
            if (ec)
                return;
		}
	}
	else {
		auto record = ConverterIDLRecord<String, String, Detail::IdentifierConverterIDLByteString>::convert(state, headersInit);
	    RETURN_IF_EXCEPTION(vm, void());

	    for(int i = 0; i < record.size(); ++i) {
	    	auto& header = record[i];
			appendToHeaderMap(header.first, header.second, headers, guard, ec);
			if (UNLIKELY(ec))
				return;
		}
	}
}

PassRefPtr<FetchHeaders> FetchHeaders::create(JSC::ExecState& state, JSC::JSValue headersInit, ExceptionCode& ec)
{
	HTTPHeaderMap headers;

    fillHeaderMap(state, headers, headersInit, Guard::None, ec);
    if(ec)
    	return nullptr;

	return adoptRef(new FetchHeaders(Guard::None, headers));
}

void FetchHeaders::fill(JSC::ExecState& state, JSC::JSValue headerInit, ExceptionCode& ec)
{
    fillHeaderMap(state, m_headers, headerInit, m_guard, ec);
}

void FetchHeaders::fill(const FetchHeaders* otherHeaders, ExceptionCode& ec)
{
	if(!otherHeaders)
		return;

	for(auto iter = otherHeaders->m_headers.begin(); iter != otherHeaders->m_headers.end(); ++iter) {
        appendToHeaderMap(iter, m_headers, m_guard, ec);
        if (ec)
            return;
    }
}

void FetchHeaders::append(const String& name, const String& value, ExceptionCode& ec)
{
    return appendToHeaderMap(name, value, m_headers, m_guard, ec);
}

void FetchHeaders::remove(const String& name, ExceptionCode& ec)
{
    if (!canWriteHeader(name, String(), m_guard, ec))
        return;
    m_headers.remove(name);
}

String FetchHeaders::get(const String& name, ExceptionCode& ec) const
{
    if (!isValidHTTPToken(name)) {
        ec = TypeError;
        return String();
    }
    return m_headers.get(name);
}

bool FetchHeaders::has(const String& name, ExceptionCode& ec) const
{
    if (!isValidHTTPToken(name)) {
        ec = TypeError;
        return false;
    }
    return m_headers.contains(name);
}

void FetchHeaders::set(const String& name, const String& value, ExceptionCode& ec)
{
    String normalizedValue = stripLeadingAndTrailingHTTPSpaces(value);
    if (!canWriteHeader(name, normalizedValue, m_guard, ec))
        return;
    m_headers.set(name, normalizedValue);
}

void FetchHeaders::filterAndFill(const HTTPHeaderMap& headers, Guard guard)
{
    ExceptionCode ec;
    for(HTTPHeaderMap::const_iterator it = headers.begin(); it != headers.end(); ++it) {
        if (canWriteHeader(it->first, it->second, m_guard, ec))
			m_headers.add(it->first, it->second);
    }
}

Optional<WTF::KeyValuePair<String, String>> FetchHeaders::Iterator::next()
{
    while (m_currentIndex < m_keys.size()) {
        String key = m_keys[m_currentIndex++];
        String value = m_headers->m_headers.get(key);
        if (!value.isNull())
            return WTF::KeyValuePair<String, String>(key, value);
    }
    m_keys.clear();
    return Nullopt;
}

FetchHeaders::Iterator::Iterator(FetchHeaders& headers)
    : m_headers(&headers)
	, m_currentIndex(0)
{
    m_keys.reserveInitialCapacity(headers.m_headers.size());
    for(auto it = headers.m_headers.begin(); it != headers.m_headers.end(); ++it)
        m_keys.uncheckedAppend(it->first.convertToASCIILowercase());

    std::sort(m_keys.begin(), m_keys.end(), WTF::codePointCompareLessThan);
}

} // namespace WebCore

#endif // ENABLE(FETCH_API)
