/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FetchRequest.h"

#if ENABLE(FETCH_API)

#include "ExceptionCode.h"
#include "HTTPParsers.h"
#include "JSAbortSignal.h"
#include "JSBlob.h"
#include "JSDOMBinding.h"
#include "JSDOMFormData.h"
#include "JSFetchRequest.h"
#include "ReadableStream.h"
#include "JSURLSearchParams.h"
#include "ScriptExecutionContext.h"
#include "SecurityOrigin.h"
#include <runtime/Error.h>
#include <runtime/JSArrayBuffer.h>
#include <runtime/JSArrayBufferView.h>
#include <runtime/JSString.h>

namespace WebCore {

static Optional<Exception> setMethod(ResourceRequest& request, const String& initMethod)
{
    if (!isValidHTTPToken(initMethod))
        return Exception(TypeError, ASCIILiteral("Method is not a valid HTTP token."));

    String method = initMethod.convertToASCIIUppercase();
    if (method == "CONNECT" || method == "TRACE" || method == "TRACK")
        return Exception(TypeError, ASCIILiteral("Method is forbidden."));

    request.setHTTPMethod((method == "DELETE" || method == "GET" || method == "HEAD" || method == "OPTIONS" || method == "POST" || method == "PUT") ? method : initMethod);

    return Nullopt;
}

static Optional<Exception> setReferrer(FetchRequest::InternalRequest& request, ScriptExecutionContext& context, const String& referrer)
{
    if (referrer.isEmpty()) {
        request.referrer = ASCIILiteral("no-referrer");
        return Nullopt;
    }

    KURL referrerURL = context.completeURL(referrer, ScriptExecutionContext::ForceUTF8::Yes);
    if (!referrerURL.isValid())
        return Exception(TypeError, ASCIILiteral("Referrer is not a valid URL."));

    if (referrerURL.protocolIsAbout() && referrerURL.path() == "client") {
        request.referrer = ASCIILiteral("client");
        return Nullopt;
    }

    if (!(context.securityOrigin() && context.securityOrigin()->canRequest(referrerURL)))
        return Exception(TypeError, ASCIILiteral("Referrer is not same-origin."));

    request.referrer = referrerURL.string();
    return Nullopt;
}

static Optional<Exception> buildOptions(JSC::ExecState& state, FetchRequest::InternalRequest& request, ScriptExecutionContext& context, const FetchRequest::Init& init)
{
    if (!init.window.isUndefinedOrNull() && !init.window.isEmpty())
		return Exception(TypeError, ASCIILiteral("Window can only be null."));

    if (!init.referrer.isNull()) {
        if (auto exception = setReferrer(request, context, init.referrer))
            return exception;
    }

    if (init.referrerPolicy)
        request.options.referrerPolicy = init.referrerPolicy.value();

    if (init.mode)
        request.options.mode = init.mode.value();
    if (request.options.mode == FetchOptions::Mode::Navigate)
        return Exception(TypeError, ASCIILiteral("Request constructor does not accept navigate fetch mode."));

    if (init.credentials)
        request.options.credentials = init.credentials.value();

    if (init.cache)
        request.options.cache = init.cache.value();
    if (request.options.cache == FetchOptions::Cache::OnlyIfCached && request.options.mode != FetchOptions::Mode::SameOrigin)
    	return Exception(TypeError, ASCIILiteral("only-if-cached cache option requires fetch mode to be same-origin."));

    if (init.redirect)
        request.options.redirect = init.redirect.value();

    if (!init.integrity.isNull())
        request.options.integrity = init.integrity;

    if (!init.method.isNull()) {
        if (auto exception = setMethod(request.request, init.method))
            return exception;
    }

    return Nullopt;
}

static bool methodCanHaveBody(const FetchRequest::InternalRequest& internalRequest)
{
    return internalRequest.request.httpMethod() != "GET" && internalRequest.request.httpMethod() != "HEAD";
}

Optional<Exception> FetchRequest::initializeOptions(JSC::ExecState& state, const Init& init)
{
    ASSERT(scriptExecutionContext());

    auto exception = buildOptions(state, m_internalRequest, *scriptExecutionContext(), init);
    if(exception)
    	return exception;

    if (m_internalRequest.options.mode == FetchOptions::Mode::NoCors) {
        const String& method = m_internalRequest.request.httpMethod();
        if (method != "GET" && method != "POST" && method != "HEAD")
            return Exception(TypeError, ASCIILiteral("Method must be GET, POST or HEAD in no-cors mode."));
        if (!m_internalRequest.options.integrity.isEmpty())
            return Exception(TypeError, ASCIILiteral("There cannot be an integrity in no-cors mode."));
        m_headers->setGuard(FetchHeaders::Guard::RequestNoCors);
    }
    return Nullopt;
}

Optional<Exception> FetchRequest::initializeWith(JSC::ExecState& state, const String& url, const Init& init)
{
    ASSERT(scriptExecutionContext());

    KURL requestURL = scriptExecutionContext()->completeURL(url, ScriptExecutionContext::ForceUTF8::Yes);
    if (!requestURL.isValid() || !requestURL.user().isEmpty() || !requestURL.pass().isEmpty())
    	return Exception(TypeError, ASCIILiteral("URL is not valid or contains user credentials."));

    m_internalRequest.options.mode = FetchOptions::Mode::Cors;
    m_internalRequest.options.credentials = FetchOptions::Credentials::Omit;
    m_internalRequest.referrer = ASCIILiteral("client");
    m_internalRequest.request.setURL(requestURL);

    auto exception = initializeOptions(state, init);
    if (exception)
        return exception;

    if (init.signal) {
        if (auto* signal = JSAbortSignal::toWrapped(init.signal))
            m_signal->signalFollow(*signal);
        else if (!init.signal.isUndefinedOrNull())  {
            ASCIILiteral consoleMessage("FetchRequestInit.signal should be undefined, null or an AbortSignal object.");
            scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Warning, consoleMessage);
        }
    }

    if (!init.headers.isUndefinedOrNull() && !init.headers.isEmpty()) {
    	ExceptionCode ec;
        m_headers->fill(state, init.headers, ec);
        if (ec)
            return Exception(ec);
    }

    if (!init.body.isUndefinedOrNull() && !init.body.isEmpty()) {
    	exception = setBody(state, init.body);
        if (exception)
            return exception;
    }

    updateContentType();
    return Nullopt;
}

Optional<Exception> FetchRequest::initializeWith(JSC::ExecState& state, FetchRequest& input, const Init& init)
{
    if (input.isDisturbedOrLocked())
        return Exception(TypeError, ASCIILiteral("Request input is disturbed or locked."));

    m_internalRequest = input.m_internalRequest;

    auto exception = initializeOptions(state, init);
    if (exception)
        return exception;

    if (init.signal && !init.signal.isUndefined()) {
        if (auto* signal = JSAbortSignal::toWrapped(init.signal))
            m_signal->signalFollow(*signal);
        else if (!init.signal.isNull()) {
            ASCIILiteral consoleMessage("FetchRequestInit.signal should be undefined, null or an AbortSignal object.");
            scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Warning, consoleMessage);
        }

    } else
        m_signal->signalFollow(*input.m_signal);

	ExceptionCode ec = 0;
    if (!init.headers.isUndefinedOrNull() && !init.headers.isEmpty())
        m_headers->fill(state, init.headers, ec);
    else
        m_headers->fill(input.headers(), ec);
    if (ec)
        return Exception(ec);

    if (!init.body.isUndefinedOrNull() && !init.body.isEmpty())
    	exception = setBody(state, init.body);
    else
    	exception = setBody(input);
    if (exception)
        return exception;

    updateContentType();
    return Nullopt;
}

Optional<Exception> FetchRequest::setBody(JSC::ExecState& state, JSC::JSValue value)
{
	if (!methodCanHaveBody(m_internalRequest))
		return Exception(TypeError, tryMakeString("Request has method '", m_internalRequest.request.httpMethod(), "' and cannot have a body"));

	Optional<Exception> exception;
	OwnPtr<FetchBody> body = FetchBody::extract(state, *scriptExecutionContext(), value, m_contentType, exception);
	if (exception)
		return exception;

	m_body = body.release();
	return Nullopt;
}

Optional<Exception> FetchRequest::setBody(FetchRequest& request)
{
    if (request.isDisturbedOrLocked())
        return Exception(TypeError, "Request input is disturbed or locked.");

    if (!request.isBodyNull()) {
        if (!methodCanHaveBody(m_internalRequest))
    		return Exception(TypeError, tryMakeString("Request has method '", m_internalRequest.request.httpMethod(), "' and cannot have a body"));

        // FIXME: If body has a readable stream, we should pipe it to this new body stream.
        m_body = request.m_body.release();
        request.setDisturbed();
    }
    return Nullopt;
}

PassRefPtr<FetchRequest> FetchRequest::create(JSC::ExecState& state, ScriptExecutionContext* context, JSC::JSValue input, const Init& init, ExceptionCode& ec)
{
	RefPtr<FetchRequest> request = adoptRef(new FetchRequest(context, nullptr, FetchHeaders::create(FetchHeaders::Guard::Request), InternalRequest()));

	if(JSFetchRequest* jsFetchRequest = JSC::jsDynamicCast<JSFetchRequest*>(input)) {
		Optional<Exception> exception = request->initializeWith(state, jsFetchRequest->wrapped(), init);
		if (UNLIKELY(exception)) {
			setDOMException(&state, exception->code(), exception->message());
			ec = exception->code();
			return nullptr;
		}
	} else {
		String url = input.toWTFString(&state);
		if (state.exception()) {
			setDOMException(&state, TypeError, ASCIILiteral("Request input is not string."));
			ec = TypeError;
			return nullptr;
		}

		Optional<Exception> exception = request->initializeWith(state, url, init);
		if (UNLIKELY(exception)) {
			setDOMException(&state, exception->code(), exception->message());
			ec = exception->code();
			return nullptr;
		}
	}

    return request.release();
}

String FetchRequest::referrer() const
{
    if (m_internalRequest.referrer == "no-referrer")
        return String();
    if (m_internalRequest.referrer == "client")
        return ASCIILiteral("about:client");
    return m_internalRequest.referrer;
}

const String& FetchRequest::url() const
{
    if (m_requestURL.isNull())
        m_requestURL = m_internalRequest.request.url();
    return m_requestURL;
}

ResourceRequest FetchRequest::internalRequest() const
{
    ASSERT(scriptExecutionContext());

    ResourceRequest request = m_internalRequest.request;
    request.setHTTPHeaderFields(m_headers->internalHeaders());

    if (!isBodyNull())
    	request.setHTTPBody(body().bodyAsFormData(*scriptExecutionContext()));

    return request;
}

RefPtr<FetchRequest> FetchRequest::clone(ScriptExecutionContext* context, ExceptionCode& ec)
{
    if (isDisturbedOrLocked()) {
        ec = TypeError;
        return nullptr;
    }

    RefPtr<FetchRequest> clone = adoptRef(new FetchRequest(context, PassOwnPtr<FetchBody>(), FetchHeaders::create(*m_headers), FetchRequest::InternalRequest(m_internalRequest)));
    clone->cloneBody(*this);
    clone->m_signal->signalFollow(*m_signal);
    return clone.release();
}

const char* FetchRequest::activeDOMObjectName() const
{
    return "Request";
}

} // namespace WebCore

#endif // ENABLE(FETCH_API)
