/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(FETCH_API)

#include "AbortSignal.h"
#include "FetchBodyOwner.h"
#include "FetchHeaders.h"
#include "FetchOptions.h"
#include "FetchRequestInit.h"
#include "ResourceRequest.h"

namespace WebCore {

class Dictionary;
class ScriptExecutionContext;

typedef int ExceptionCode;

class FetchRequest : public FetchBodyOwner {
public:
    using Init = FetchRequestInit;

    using Cache = FetchOptions::Cache;
    using Credentials = FetchOptions::Credentials;
    using Destination = FetchOptions::Destination;
    using Mode = FetchOptions::Mode;
    using Redirect = FetchOptions::Redirect;
    using Type = FetchOptions::Type;

    static PassRefPtr<FetchRequest> create(JSC::ExecState&, ScriptExecutionContext* context, JSC::JSValue, const Init&, ExceptionCode&);

    const String& method() const { return m_internalRequest.request.httpMethod(); }
    const String& url() const;
    FetchHeaders* headers() const { return m_headers.get(); }

    Type type() const;
    Destination destination() const;
    String referrer() const;
    FetchReferrerPolicy referrerPolicy() const;
    Mode mode() const;
    Credentials credentials() const;
    Cache cache() const;
    Redirect redirect() const;
    const String& integrity() const { return m_internalRequest.options.integrity; }
    AbortSignal* signal() const { return m_signal.get(); }

    RefPtr<FetchRequest> clone(ScriptExecutionContext*, ExceptionCode&);

    struct InternalRequest {
        ResourceRequest request;
        FetchOptions options;
        String referrer;
    };

    const FetchOptions& fetchOptions() const { return m_internalRequest.options; }
    ResourceRequest internalRequest() const;
    const String& internalRequestReferrer() const { return m_internalRequest.referrer; }

private:
    FetchRequest(ScriptExecutionContext*, PassOwnPtr<FetchBody>, RefPtr<FetchHeaders>, const InternalRequest&);

    Optional<Exception> initializeOptions(JSC::ExecState&, const Init&);

    Optional<Exception> initializeWith(JSC::ExecState&, FetchRequest&, const Init&);
    Optional<Exception> initializeWith(JSC::ExecState&, const String&, const Init&);
    Optional<Exception>  setBody(FetchRequest& request);
    Optional<Exception>  setBody(JSC::ExecState&, JSC::JSValue);

    // ActiveDOMObject API.
    virtual const char* activeDOMObjectName() const FINAL;

    InternalRequest m_internalRequest;
    mutable String m_requestURL;
    RefPtr<AbortSignal> m_signal;
};

inline FetchRequest::FetchRequest(ScriptExecutionContext* context, PassOwnPtr<FetchBody> body, RefPtr<FetchHeaders> headers, const InternalRequest& internalRequest)
	: FetchBodyOwner(context, body, headers)
    , m_internalRequest(internalRequest)
	, m_signal(AbortSignal::create(context))
{
}

inline FetchRequest::Cache FetchRequest::cache() const
{
    return m_internalRequest.options.cache;
}

inline FetchRequest::Credentials FetchRequest::credentials() const
{
    return m_internalRequest.options.credentials;
}

inline FetchRequest::Destination FetchRequest::destination() const
{
    return m_internalRequest.options.destination;
}

inline FetchRequest::Mode FetchRequest::mode() const
{
    return m_internalRequest.options.mode;
}

inline FetchRequest::Redirect FetchRequest::redirect() const
{
    return m_internalRequest.options.redirect;
}

inline FetchReferrerPolicy FetchRequest::referrerPolicy() const
{
    return m_internalRequest.options.referrerPolicy;
}

inline FetchRequest::Type FetchRequest::type() const
{
    return m_internalRequest.options.type;
}

} // namespace WebCore

#endif // ENABLE(FETCH_API)
