/*
 * Copyright (C) 2016 Canon Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted, provided that the following conditions
 * are required to be met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Canon Inc. nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CANON INC. AND ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(FETCH_API)

#include "FetchBodyOwner.h"
#include "ReadableStreamSink.h"
#include "ResourceResponse.h"
#include <runtime/TypedArrays.h>
#include <wtf/ForbidHeapAllocation.h>
#include <wtf/Ref.h>
#include <wtf/WeakPtr.h>

namespace JSC {
class ArrayBuffer;
class ExecState;
class JSValue;
};

namespace WebCore {

class AbortSignal;
class DeferredPromise;
class Dictionary;
class FetchRequest;
struct ReadableStreamChunk;
class ReadableStreamSource;

typedef int ExceptionCode;

typedef ResourceResponse::Type FetchResponseType;

class FetchResponse final : public FetchBodyOwner {
public:
	struct Init {
	    WTF_FORBID_HEAP_ALLOCATION;
	public:
		unsigned short status { 200 };
		String statusText { ASCIILiteral("OK") };
		JSC::JSValue headers;
	};

	typedef ResourceResponse::Type Type;

    static Ref<FetchResponse> create(ScriptExecutionContext*, PassOwnPtr<FetchBody>, PassRefPtr<FetchHeaders>, const ResourceResponse&);

    static PassRefPtr<FetchResponse> create(JSC::ExecState&, ScriptExecutionContext*, JSC::JSValue, const Init&, ExceptionCode&);
    static Ref<FetchResponse> error(ScriptExecutionContext*);
    static PassRefPtr<FetchResponse> redirect(ScriptExecutionContext*, const String&, int, ExceptionCode&);

    static void fetch(ScriptExecutionContext&, FetchRequest&, RefPtr<DeferredPromise>);

#if ENABLE(STREAMS_API)
    void consumeChunk(PassRefPtr<JSC::Uint8Array>);
    void finishConsumingStream(PassRefPtr<DeferredPromise>);
#endif

    Type type() const { return filteredResponse().type(); }
    const String& url() const;
    bool redirected() const { return filteredResponse().isRedirected(); }
    int status() const { return filteredResponse().httpStatusCode(); }
    bool ok() const { return filteredResponse().isSuccessful(); }
    const String& statusText() const { return filteredResponse().httpStatusText(); }

    FetchHeaders* headers() const { return m_headers.get(); }
    PassRefPtr<FetchResponse> clone(ExceptionCode&);

#if ENABLE(STREAMS_API)
    virtual void consumeBodyAsStream() FINAL;
    virtual void feedStream() FINAL;
    virtual void cancel() FINAL;
#endif

    //using ResponseData = Variant<std::nullptr_t, Ref<FormData>, Ref<SharedBuffer>>;
    //ResponseData consumeBody();
    void setBodyData(std::nullptr_t);
    void setBodyData(PassRefPtr<FormData>);
    void setBodyData(PassRefPtr<SharedBuffer>);
    bool isLoading() const { return !!m_bodyLoader; }

    //using ConsumeDataCallback = WTF::Function<void(ExceptionOr<RefPtr<SharedBuffer>>&&)>;
    //void consumeBodyWhenLoaded(RefPtr<SharedBuffer>, ExceptionCode&);

    //using ConsumeDataByChunkCallback = WTF::Function<void(ExceptionOr<ReadableStreamChunk*>&&)>;
    //void consumeBodyFromReadableStream(ConsumeDataByChunkCallback&&);

    ResourceResponse resourceResponse() const;

    void runAbortSteps();

private:
    FetchResponse(ScriptExecutionContext*, PassOwnPtr<FetchBody>, PassRefPtr<FetchHeaders>, const ResourceResponse&);

    // ActiveDOMObject API
    virtual void stop() FINAL;
    virtual const char* activeDOMObjectName() const FINAL;
    virtual bool canSuspendForDocumentSuspension() const FINAL;

    const ResourceResponse& filteredResponse() const;

#if ENABLE(STREAMS_API)
    void closeStream();
#endif

    void addAbortSteps(PassRefPtr<AbortSignal>);

    class BodyLoader FINAL : public FetchLoaderClient {
        WTF_MAKE_FAST_ALLOCATED;
    public:
        BodyLoader(FetchResponse&, RefPtr<DeferredPromise>);
        ~BodyLoader();

        bool start(ScriptExecutionContext&, const FetchRequest&);
        void stop();

#if ENABLE(STREAMS_API)
        RefPtr<SharedBuffer> startStreaming();
#endif

        //NotificationCallback takeNotificationCallback() { return WTFMove(m_responseCallback); }
        //ConsumeDataByChunkCallback takeConsumeDataCallback() { return WTFMove(m_consumeDataCallback); }
        PassRefPtr<DeferredPromise> takepromise() { return m_promise.release(); }

    private:
        // FetchLoaderClient API
        virtual void didSucceed() FINAL;
        virtual void didFail(const ResourceError&) FINAL;
        virtual void didReceiveResponse(const ResourceResponse&) FINAL;
        void didReceiveData(const char*, size_t) FINAL;

        FetchResponse& m_response;
        //NotificationCallback m_responseCallback;
        RefPtr<DeferredPromise> m_promise;
        OwnPtr<FetchLoader> m_loader;
        Ref<PendingActivity<FetchResponse>> m_pendingActivity;
    };

    mutable Optional<ResourceResponse> m_filteredResponse;
    ResourceResponse m_internalResponse;
    OwnPtr<BodyLoader> m_bodyLoader;
    mutable String m_responseURL;
    RefPtr<AbortSignal> m_abortSignal;
};

} // namespace WebCore

#endif // ENABLE(FETCH_API)
