/*
 * Copyright (C) 2013 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MediaSource_h
#define MediaSource_h

#if ENABLE(MEDIA_SOURCE)

#include "ActiveDOMObject.h"
#include "EventTarget.h"
#include "GenericEventQueue.h"
#include "HTMLMediaElement.h"
#include "HTMLMediaSource.h"
#include "MediaSourcePrivate.h"
#include "ScriptWrappable.h"
#include "SourceBuffer.h"
#include "SourceBufferList.h"
#include "URLRegistry.h"
#include <wtf/PassOwnPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>
#include <wtf/WeakPtr.h>

namespace WebCore {

class MediaSource FINAL : public RefCounted<MediaSource>, public HTMLMediaSource, public ActiveDOMObject, public EventTargetWithInlineData {
public:
    static const AtomicString& openKeyword();
    static const AtomicString& closedKeyword();
    static const AtomicString& endedKeyword();

    static PassRefPtr<MediaSource> create(ScriptExecutionContext*);
    virtual ~MediaSource();

    void addedToRegistry();
    void removedFromRegistry();
    void openIfInEndedState();
    bool isOpen() const;

    // HTMLMediaSource
    virtual bool attachToElement(HTMLMediaElement&);
    virtual void detachFromElement(HTMLMediaElement&);
    virtual void setPrivateAndOpen(PassRefPtr<MediaSourcePrivate>) OVERRIDE;
    virtual bool isClosed() const OVERRIDE;
    virtual MediaTime duration() const OVERRIDE;
    virtual PassOwnPtr<PlatformTimeRanges> buffered() const OVERRIDE;
    virtual void refHTMLMediaSource() OVERRIDE { ref(); }
    virtual void derefHTMLMediaSource() OVERRIDE { deref(); }
    virtual PassOwnPtr<PlatformTimeRanges> seekable();
	void setLiveSeekableRange(double start, double end, ExceptionCode&);
	void clearLiveSeekableRange(ExceptionCode&);

    void setDuration(double, ExceptionCode&);
    void setDurationInternal(const MediaTime& duration);
    const AtomicString& readyState() const { return m_readyState; }
    void setReadyState(const AtomicString&);
    void endOfStream(const AtomicString& error, ExceptionCode&);

    HTMLMediaElement* mediaElement() const { return m_mediaElement.get(); }

    // MediaSource.idl methods
    SourceBufferList* sourceBuffers() { return m_sourceBuffers.get(); }
    SourceBufferList* activeSourceBuffers() { return m_activeSourceBuffers.get(); }
    SourceBuffer* addSourceBuffer(const String& type, ExceptionCode&);
    void removeSourceBuffer(SourceBuffer*, ExceptionCode&);
    static bool isTypeSupported(const String& type);

    // EventTarget interface
    virtual ScriptExecutionContext* scriptExecutionContext() const OVERRIDE;
    virtual void refEventTarget() OVERRIDE { ref(); }
    virtual void derefEventTarget() OVERRIDE { deref(); }

    // URLRegistrable interface
    virtual URLRegistry& registry() const OVERRIDE;

    void monitorSourceBuffers() { }

    // EventTarget interface
    virtual EventTargetInterface eventTargetInterface() const OVERRIDE;

    using RefCounted<MediaSource>::ref;
    using RefCounted<MediaSource>::deref;

private:
    explicit MediaSource(ScriptExecutionContext*);

    // ActiveDOMObject interface
    virtual void suspend(ReasonForSuspension) FINAL;
    virtual void resume() FINAL;
    virtual void stop() OVERRIDE;
    virtual bool canSuspendForDocumentSuspension() const FINAL;
    virtual const char* activeDOMObjectName() const OVERRIDE;
    virtual bool virtualHasPendingActivity() const OVERRIDE;

    RefPtr<SourceBufferPrivate> createSourceBufferPrivate(const String& type, const MediaSourcePrivate::CodecsArray&, ExceptionCode&);
    void scheduleEvent(const AtomicString& eventName);
    GenericEventQueue& asyncEventQueue() { return m_asyncEventQueue; }

    void onReadyStateChange(const AtomicString&, const AtomicString&);
    Vector<PlatformTimeRanges> activeRanges() const;

    RefPtr<MediaSourcePrivate> m_private;
    OwnPtr<PlatformTimeRanges> m_liveSeekable;
    WeakPtr<HTMLMediaElement> m_mediaElement;
    RefPtr<SourceBufferList> m_sourceBuffers;
    RefPtr<SourceBufferList> m_activeSourceBuffers;
    AtomicString m_readyState;
    GenericEventQueue m_asyncEventQueue;
};

} // namespace WebCore

#endif

#endif
