/*
 * Copyright (C) 2015 Ericsson AB. All rights reserved.
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of Ericsson nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(WEB_RTC)

#include "JSDOMPromiseDeferred.h"
#include "JSDOMPromise.h"
#include "RTCRtpParameters.h"
#include "RTCSignalingState.h"

namespace WebCore {

class MediaStream;
class MediaStreamTrack;
class PeerConnectionBackend;
class RTCDataChannelHandler;
class RTCIceCandidate;
class RTCPeerConnection;
class RTCRtpReceiver;
class RTCRtpSender;
class RTCSessionDescription;
class RTCStatsReport;

struct MediaEndpointConfiguration;
struct RTCAnswerOptions;
struct RTCDataChannelInit;
struct RTCOfferOptions;

namespace PeerConnection {
typedef RefPtr<DeferredPromise> SessionDescriptionPromise;
typedef RefPtr<DeferredPromise> VoidPromise;
typedef RefPtr<DeferredPromise> StatsPromise;
}

typedef PassOwnPtr<PeerConnectionBackend> (*CreatePeerConnectionBackend)(RTCPeerConnection&);

// FIXME: What is the value of this abstract class? There is only one concrete class derived from it.
class PeerConnectionBackend {
public:
    static CreatePeerConnectionBackend create;

    PeerConnectionBackend(RTCPeerConnection& peerConnection)
    	: m_peerConnection(peerConnection)
    	, m_shouldFilterICECandidates(true)
    	, m_negotiationNeeded(false)
    {
    }
    virtual ~PeerConnectionBackend() { }

    void createOffer(const RTCOfferOptions&, PeerConnection::SessionDescriptionPromise);
    void createAnswer(const RTCAnswerOptions&, PeerConnection::SessionDescriptionPromise);
    void setLocalDescription(RTCSessionDescription&, PeerConnection::VoidPromise);
    void setRemoteDescription(RTCSessionDescription&, PeerConnection::VoidPromise);
    void addIceCandidate(RTCIceCandidate&, PeerConnection::VoidPromise);

    virtual PassOwnPtr<RTCDataChannelHandler> createDataChannelHandler(const String&, const RTCDataChannelInit&) = 0;

    void stop();

    virtual RefPtr<RTCSessionDescription> localDescription() const = 0;
    virtual RefPtr<RTCSessionDescription> currentLocalDescription() const = 0;
    virtual RefPtr<RTCSessionDescription> pendingLocalDescription() const = 0;

    virtual RefPtr<RTCSessionDescription> remoteDescription() const = 0;
    virtual RefPtr<RTCSessionDescription> currentRemoteDescription() const = 0;
    virtual RefPtr<RTCSessionDescription> pendingRemoteDescription() const = 0;

    virtual void setConfiguration(MediaEndpointConfiguration&&) = 0;

    virtual void getStats(MediaStreamTrack*, RefPtr<DeferredPromise>&&) = 0;

    virtual Vector<RefPtr<MediaStream>> getRemoteStreams() const = 0;

    virtual RefPtr<RTCRtpReceiver> createReceiver(const String& transceiverMid, const String& trackKind, const String& trackId) = 0;
	virtual void replaceTrack(RTCRtpSender&, Ref<MediaStreamTrack>&&, PeerConnection::VoidPromise) = 0;
	virtual void notifyAddedTrack(RTCRtpSender&) { }
    virtual void notifyRemovedTrack(RTCRtpSender&) { }

    virtual RTCRtpParameters getParameters(RTCRtpSender&) const { return RTCRtpParameters(); }

    void markAsNeedingNegotiation();
    bool isNegotiationNeeded() const { return m_negotiationNeeded; };
    void clearNegotiationNeededState() { m_negotiationNeeded = false; };

    virtual void emulatePlatformEvent(const String& action) = 0;

    void newICECandidate(const String& sdp, const String& mid);
    void disableICECandidateFiltering();
    void enableICECandidateFiltering();

protected:
    void fireICECandidateEvent(PassRefPtr<RTCIceCandidate>);
    void doneGatheringCandidates();

    void updateSignalingState(RTCSignalingState);

    void createOfferSucceeded(const String&);
    void createOfferFailed(const Exception&);

    void createAnswerSucceeded(const String&);
    void createAnswerFailed(const Exception&);

    void setLocalDescriptionSucceeded();
    void setLocalDescriptionFailed(const Exception&);

    void setRemoteDescriptionSucceeded();
    void setRemoteDescriptionFailed(const Exception&);

    void addIceCandidateSucceeded();
    void addIceCandidateFailed(const Exception&);

private:
    virtual void doCreateOffer(const RTCOfferOptions&) = 0;
    virtual void doCreateAnswer(const RTCAnswerOptions&) = 0;
    virtual void doSetLocalDescription(RTCSessionDescription&) = 0;
    virtual void doSetRemoteDescription(RTCSessionDescription&) = 0;
    virtual void doAddIceCandidate(RTCIceCandidate&) = 0;
    virtual void doStop() = 0;

protected:
    RTCPeerConnection& m_peerConnection;

private:
    Optional<PeerConnection::SessionDescriptionPromise> m_offerAnswerPromise;
    Optional<PeerConnection::VoidPromise> m_setDescriptionPromise;
    Optional<PeerConnection::VoidPromise> m_addIceCandidatePromise;

    bool m_shouldFilterICECandidates;
    struct PendingICECandidate {
        // Fields described in https://www.w3.org/TR/webrtc/#idl-def-rtcicecandidateinit.
        String sdp;
        String mid;
    };
    Vector<PendingICECandidate> m_pendingICECandidates;

    bool m_negotiationNeeded;
};

} // namespace WebCore

#endif // ENABLE(WEB_RTC)

