/*
 * Copyright (C) 2012 Google Inc. All rights reserved.
 * Copyright (C) 2013 Nokia Corporation and/or its subsidiary(-ies).
 * Copyright (C) 2015 Ericsson AB. All rights reserved.
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer
 *    in the documentation and/or other materials provided with the
 *    distribution.
 * 3. Neither the name of Google Inc. nor the names of its contributors
 *    may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef RTCPeerConnection_h
#define RTCPeerConnection_h

#if ENABLE(WEB_RTC)

#include "ActiveDOMObject.h"
#include "EventTarget.h"
#include "ExceptionBase.h"
#include "MediaStream.h"
#include "RTCConfiguration.h"
#include "RTCDataChannel.h"
#include "RTCIceConnectionState.h"
#include "RTCIceGatheringState.h"
#include "RTCPeerConnectionState.h"
#include "RTCRtpTransceiver.h"
#include "RTCSignalingState.h"
#include "RTCRtpTransceiverInit.h"
#include "RTCIceCandidate.h"
#include "RTCPeerConnectionHandlerClient.h"
#include "Timer.h"

namespace WebCore {

class MediaStream;
class MediaStreamTrack;
class PeerConnectionBackend;
class RTCController;
class RTCIceCandidate;
class RTCPeerConnectionErrorCallback;
class RTCSessionDescription;
class RTCStatsCallback;

struct RTCAnswerOptions;
struct RTCOfferOptions;
struct RTCRtpParameters;

class RTCPeerConnection FINAL : public RefCounted<RTCPeerConnection>, public RTCRtpSender::Backend, public EventTargetWithInlineData, public ActiveDOMObject {
public:
    static PassRefPtr<RTCPeerConnection> create(ScriptExecutionContext*);
    virtual ~RTCPeerConnection();

    typedef RTCDataChannelInit DataChannelInit;

    void initializeWith(Document*, const RTCConfiguration&, ExceptionCode&);

    const Vector<RefPtr<RTCRtpSender>>& getSenders() const { return m_transceiverSet->senders(); }
    const Vector<RefPtr<RTCRtpReceiver>>& getReceivers() const { return m_transceiverSet->receivers(); }
    const Vector<RefPtr<RTCRtpTransceiver>>& getTransceivers() const { return m_transceiverSet->list(); }

    // Part of legacy MediaStream-based API (mostly implemented as JS built-ins)
    Vector<RefPtr<MediaStream>> getRemoteStreams() const { return m_backend->getRemoteStreams(); }

    PassRefPtr<RTCRtpSender> addTrack(MediaStreamTrack&, const Vector<RefPtr<MediaStream>>&, ExceptionCode&);
    void removeTrack(RTCRtpSender&, ExceptionCode&);

    RefPtr<RTCRtpTransceiver> addTransceiver(MediaStreamTrack&, const RTCRtpTransceiverInit&, ExceptionCode&);
    RefPtr<RTCRtpTransceiver> addTransceiver(const String& kind, const RTCRtpTransceiverInit&, ExceptionCode&);

    void queuedCreateOffer(const RTCOfferOptions&, PeerConnection::SessionDescriptionPromise);
    void queuedCreateAnswer(const RTCAnswerOptions&, PeerConnection::SessionDescriptionPromise);

    void queuedSetLocalDescription(RTCSessionDescription&, PeerConnection::VoidPromise);
    RefPtr<RTCSessionDescription> localDescription() const;
    RefPtr<RTCSessionDescription> currentLocalDescription() const;
    RefPtr<RTCSessionDescription> pendingLocalDescription() const;

    void queuedSetRemoteDescription(RTCSessionDescription&, PeerConnection::VoidPromise);
    RefPtr<RTCSessionDescription> remoteDescription() const;
    RefPtr<RTCSessionDescription> currentRemoteDescription() const;
    RefPtr<RTCSessionDescription> pendingRemoteDescription() const;

    void queuedAddIceCandidate(RTCIceCandidate&, PeerConnection::VoidPromise);

    RTCSignalingState signalingState() const { return m_signalingState; }
    RTCIceGatheringState iceGatheringState() const { return m_iceGatheringState; }
    RTCIceConnectionState iceConnectionState() const { return m_iceConnectionState; }
    RTCPeerConnectionState connectionState() const { return m_connectionState; }

    const RTCConfiguration& getConfiguration() const { return m_configuration; }
    void setConfiguration(const RTCConfiguration&, ExceptionCode&);

    void getStats(MediaStreamTrack*, RefPtr<DeferredPromise>);
    void getStats(RefPtr<DeferredPromise>);

    PassRefPtr<RTCDataChannel> createDataChannel(ScriptExecutionContext*, const String& label, const RTCDataChannelInit&, ExceptionCode&);

    void close();

    // EventTarget
    virtual EventTargetInterface eventTargetInterface() const OVERRIDE;
    virtual ScriptExecutionContext* scriptExecutionContext() const OVERRIDE;

    using RefCounted::ref;
    using RefCounted::deref;

    // Used for testing with a mock
    void emulatePlatformEvent(const String& action);

    // API used by PeerConnectionBackend and relatives
    void addTransceiver(PassRefPtr<RTCRtpTransceiver>);
    void setSignalingState(RTCSignalingState);
    void updateIceGatheringState(RTCIceGatheringState);
    void updateIceConnectionState(RTCIceConnectionState);

    void scheduleNegotiationNeededEvent();

    RTCRtpSender::Backend& senderBackend() { return *this; }
    void fireEvent(Event&);

    void disableICECandidateFiltering() { m_backend->disableICECandidateFiltering(); }
    void enableICECandidateFiltering() { m_backend->enableICECandidateFiltering(); }

    void enqueueReplaceTrackTask(RTCRtpSender&, Ref<MediaStreamTrack>&&, RefPtr<DeferredPromise>&&);

private:
    RTCPeerConnection(ScriptExecutionContext*);

    RefPtr<RTCRtpTransceiver> completeAddTransceiver(RefPtr<RTCRtpTransceiver>, const RTCRtpTransceiverInit&);

    RTCController& rtcController();
    void registerToController();
    void unregisterFromController();

    // EventTarget implementation.
    virtual void refEventTarget() { ref(); }
    virtual void derefEventTarget() { deref(); }

    // ActiveDOMObject
    virtual void stop() OVERRIDE;
    virtual const char* activeDOMObjectName() const OVERRIDE;
    virtual bool canSuspendForDocumentSuspension() const;

    // FIXME: We might want PeerConnectionBackend to be the Backend
    // RTCRtpSender::Backend
    void replaceTrack(RTCRtpSender&, RefPtr<MediaStreamTrack>&&, RefPtr<DeferredPromise>&&) FINAL;
    RTCRtpParameters getParameters(RTCRtpSender&) const FINAL;

    void updateConnectionState();
    bool doClose();
    void doStop();

    bool m_isStopped { false };
    RTCSignalingState m_signalingState;
    RTCIceGatheringState m_iceGatheringState;
    RTCIceConnectionState m_iceConnectionState;
    RTCPeerConnectionState m_connectionState { RTCPeerConnectionState::New };

    OwnPtr<RtpTransceiverSet> m_transceiverSet;

    OwnPtr<PeerConnectionBackend> m_backend;

    RTCConfiguration m_configuration;

    class CloseTask;
    class UpdateIceGatheringStateTask;
    class UpdateIceConnectionStateTask;
    class ScheduleNegotiationNeededEventTask;

    friend class CloseTask;
    friend class UpdateIceGatheringStateTask;
    friend class UpdateIceConnectionStateTask;
    friend class ScheduleNegotiationNeededEventTask;
};

} // namespace WebCore

#endif // ENABLE(WEB_RTC)

#endif // RTCPeerConnection_h
