/*
 * Copyright (C) 2012 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(WEB_RTC)

#include "JSDOMMapLike.h"
#include "RTCStatsType.h"
#include <wtf/Optional.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>

namespace WebCore {

struct RTCStats {
	RTCStats() : timestamp(0), type(RTCStatsType::Codec) { }
    unsigned long long timestamp;
    RTCStatsType type;
    String id;
};

struct RTCRTPStreamStats : RTCStats {
	RTCRTPStreamStats() { }
	uint32_t ssrc { 0 };
    String associateStatsId;
    bool isRemote { false };
    String mediaType;

    String mediaTrackId;
    String transportId;
    String codecId;
    uint32_t firCount { 0 };
    uint32_t pliCount { 0 };
    uint32_t nackCount { 0 };
    uint32_t sliCount { 0 };
    unsigned long long qpSum { 0 };
};

struct RTCInboundRTPStreamStats : RTCRTPStreamStats {
	RTCInboundRTPStreamStats() { type = RTCStatsType::InboundRtp; }

	uint32_t packetsReceived { 0 };
    unsigned long long bytesReceived { 0 };
    uint32_t packetsLost { 0 };
    double jitter { 0 };
    double fractionLost { 0 };
    uint32_t packetsDiscarded { 0 };
    uint32_t packetsRepaired { 0 };
    uint32_t burstPacketsLost { 0 };
    uint32_t burstPacketsDiscarded { 0 };
    uint32_t burstLossCount { 0 };
    uint32_t burstDiscardCount { 0 };
    double burstLossRate { 0 };
    double burstDiscardRate { 0 };
    double gapLossRate { 0 };
    double gapDiscardRate { 0 };
    uint32_t framesDecoded { 0 };
};

struct RTCOutboundRTPStreamStats : RTCRTPStreamStats {
	RTCOutboundRTPStreamStats() { type = RTCStatsType::OutboundRtp; }

	uint32_t packetsSent { 0 };
    unsigned long long bytesSent { 0 };
    double targetBitrate { 0 };
    uint32_t framesEncoded { 0 };
};

struct RTCMediaStreamTrackStats : RTCStats {
	RTCMediaStreamTrackStats() { type = RTCStatsType::Track; }

    String trackIdentifier;
    bool remoteSource { false };
    bool ended { false };
    bool detached { false };
    uint32_t frameWidth { 0 };
    uint32_t frameHeight { 0};
    double framesPerSecond { 0 };
    uint32_t framesSent { 0 };
    uint32_t framesReceived { 0 };
    uint32_t framesDecoded { 0 };
    uint32_t framesDropped { 0 };
    uint32_t framesCorrupted { 0 };
    uint32_t partialFramesLost { 0 };
    uint32_t fullFramesLost { 0 };
    double audioLevel { 0 };
    double echoReturnLoss { 0 };
    double echoReturnLossEnhancement { 0 };
};

struct RTCDataChannelStats : RTCStats {
	RTCDataChannelStats() { type = RTCStatsType::DataChannel; }

    String label;
    String protocol;
    long datachannelid { 0 };
    String state;
    uint32_t messagesSent { 0 };
    unsigned long long bytesSent { 0 };
    uint32_t messagesReceived { 0 };
    unsigned long long bytesReceived { 0 };
};

class RTCStatsReport : public RefCounted<RTCStatsReport> {
public:
	using RTCRTPStreamStats = WebCore::RTCRTPStreamStats;
	using InboundRTPStreamStats = RTCInboundRTPStreamStats;
	using OutboundRTPStreamStats = RTCOutboundRTPStreamStats;
	using MediaStreamTrackStats = RTCMediaStreamTrackStats;
	using DataChannelStats = RTCDataChannelStats;
	using Stats = RTCStats;
	using Type = RTCStatsType;

    static PassRefPtr<RTCStatsReport> create() { return adoptRef(new RTCStatsReport()); }

    void synchronizeBackingMap(Ref<DOMMapLike>&& mapLike) { m_mapLike = WTFMove(mapLike); }
    DOMMapLike* backingMap() { return m_mapLike.get(); }

    template<typename Value> void addStats(typename Value::ParameterType&& value) { m_mapLike->set<String, Value>(value.id, std::forward<typename Value::ParameterType>(value)); }

private:
    RTCStatsReport() = default;

private:
    RefPtr<DOMMapLike> m_mapLike;
};

} // namespace WebCore

#endif // ENABLE(WEB_RTC)
