/*
 * Copyright (C) 2011 Google Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(WEB_SOCKETS)

#include "WebSocket.h"

#include "Blob.h"
#include "BlobData.h"
#include "CloseEvent.h"
#include "ContentSecurityPolicy.h"
#include "DOMWindow.h"
#include "Document.h"
#include "Event.h"
#include "EventException.h"
#include "EventListener.h"
#include "EventNames.h"
#include "ExceptionCode.h"
#include "Frame.h"
#include "FrameLoader.h"
#include "Logging.h"
#include "MessageEvent.h"
#include "ScriptCallStack.h"
#include "ScriptDisallowedScope.h"
#include "ScriptExecutionContext.h"
#include "SecurityOrigin.h"
#include "ThreadableWebSocketChannel.h"
#include "WebSocketChannel.h"
#include <runtime/ArrayBuffer.h>
#include <runtime/ArrayBufferView.h>
#include <wtf/HashSet.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/StdLibExtras.h>
#include <wtf/text/CString.h>
#include <wtf/text/StringBuilder.h>
#include <wtf/text/WTFString.h>

using namespace std;

namespace WebCore {

const size_t maxReasonSizeInBytes = 123;

static inline bool isValidProtocolCharacter(UChar character)
{
    // Hybi-10 says "(Subprotocol string must consist of) characters in the range U+0021 to U+007E not including
    // separator characters as defined in [RFC2616]."
    const UChar minimumProtocolCharacter = '!'; // U+0021.
    const UChar maximumProtocolCharacter = '~'; // U+007E.
    return character >= minimumProtocolCharacter && character <= maximumProtocolCharacter
        && character != '"' && character != '(' && character != ')' && character != ',' && character != '/'
        && !(character >= ':' && character <= '@') // U+003A - U+0040 (':', ';', '<', '=', '>', '?', '@').
        && !(character >= '[' && character <= ']') // U+005B - U+005D ('[', '\\', ']').
        && character != '{' && character != '}';
}

static bool isValidProtocolString(const String& protocol)
{
    if (protocol.isEmpty())
        return false;
    for (size_t i = 0; i < protocol.length(); ++i) {
        if (!isValidProtocolCharacter(protocol[i]))
            return false;
    }
    return true;
}

static String encodeProtocolString(const String& protocol)
{
    StringBuilder builder;
    for (size_t i = 0; i < protocol.length(); i++) {
        if (protocol[i] < 0x20 || protocol[i] > 0x7E)
            builder.append(String::format("\\u%04X", protocol[i]));
        else if (protocol[i] == 0x5c)
            builder.append("\\\\");
        else
            builder.append(protocol[i]);
    }
    return builder.toString();
}

static String joinStrings(const Vector<String>& strings, const char* separator)
{
    StringBuilder builder;
    for (size_t i = 0; i < strings.size(); ++i) {
        if (i)
            builder.append(separator);
        builder.append(strings[i]);
    }
    return builder.toString();
}

unsigned long WebSocket::saturateAdd(unsigned long a, unsigned long b)
{
    if (numeric_limits<unsigned long>::max() - a < b)
        return numeric_limits<unsigned long>::max();
    return a + b;
}

static bool webSocketsAvailable = true;

void WebSocket::setIsAvailable(bool available)
{
    webSocketsAvailable = available;
}

bool WebSocket::isAvailable()
{
    return webSocketsAvailable;
}

ASCIILiteral WebSocket::subProtocolSeperator()
{
    return ASCIILiteral(", ");
}

WebSocket::WebSocket(ScriptExecutionContext* context)
: ActiveDOMObject(context)
, m_binaryType(BinaryTypeBlob)
, m_subprotocol(emptyString())
, m_extensions(emptyString())
{
    //fprintf(stderr, "\t==>WebSocket::WebSocket\n");
}

// Espial PS
WebSocket::WebSocket(ScriptExecutionContext* context, WebSocketClient* client)
: ActiveDOMObject(context)
, m_binaryType(BinaryTypeArrayBuffer)
, m_subprotocol(emptyString())
, m_extensions(emptyString())
, m_wsclient(client)
{
    //fprintf(stderr, "\t==>WebSocket::WebSocket\n");
    if (m_wsclient)
    	m_wsclient->didCreate();
}

DLL_PUBLIC WebSocket::~WebSocket()
{
    //fprintf(stderr, "\t==>WebSocket::~WebSocket\n");
    if (m_channel)
        m_channel->disconnect();
    if (m_wsclient) {
        m_wsclient->didDestroy();
        delete m_wsclient;
    }
}

PassRefPtr<WebSocket> WebSocket::create(ScriptExecutionContext* context)
{
    RefPtr<WebSocket> webSocket(adoptRef(new WebSocket(context)));
    webSocket->suspendIfNeeded();
    return webSocket.release();
}

// Espial PS
DLL_PUBLIC PassRefPtr<WebSocket> WebSocket::create(ScriptExecutionContext* context, WebSocketClient* client)
{
    RefPtr<WebSocket> webSocket(adoptRef(new WebSocket(context, client)));
    return webSocket.release();
}

DLL_PUBLIC void WebSocket::connect(const String& url, ExceptionCode& ec)
{
    Vector<String> protocols;
    connect(url, protocols, ec);
}
void WebSocket::connect(const char* url, int& ec)
{
    String url2 = String::fromLatin1(url);
    connect(url2, ec);
}

void WebSocket::connect(const String& url, const String& protocol, ExceptionCode& ec)
{
    Vector<String> protocols;
    protocols.append(protocol);
    connect(url, protocols, ec);
}
void WebSocket::connect(const char* url, const char* protocol, int& ec)
{
    String url2 = String::fromLatin1(url);
    String protocol2 = String::fromLatin1(protocol);
    connect(url2, protocol2, ec);
}

void WebSocket::connect(const char* url, const char* protocols[], int sz, int& ec)
{
    String url2 = String::fromLatin1(url);

    Vector<String> protocols2;
    protocols2.reserveInitialCapacity(sz);
    for (int i = 0; i < sz; ++i)
    	protocols2.append(String::fromLatin1(protocols[i]));

    connect(url2, protocols2, ec);
}

class WebSocket::FailAsynchronouslyKeepingObjectAliveTask final : public ActiveDOMObject::KeepingObjectAliveTask<WebSocket>
{
public:
	FailAsynchronouslyKeepingObjectAliveTask(WebSocket& thisObject) : KeepingObjectAliveTask<WebSocket>(thisObject) { }
private:
    virtual void operator()() final
    {
        // We must block this connection. Instead of throwing an exception, we indicate this
        // using the error event. But since this code executes as part of the WebSocket's
        // constructor, we have to wait until the constructor has completed before firing the
        // event; otherwise, users can't connect to the event.

    	m_protectedObject->dispatchErrorEventIfNeeded();
    	m_protectedObject->stop();
    }
};

void WebSocket::failAsynchronously()
{
    queueTaskKeepingObjectAlive(*this, TaskSource::WebSocket, adoptPtr(new FailAsynchronouslyKeepingObjectAliveTask(*this)));
}

DLL_PUBLIC void WebSocket::connect(const String& url, const Vector<String>& protocols, ExceptionCode& ec)
{
    LOG(Network, "WebSocket %p connect to %s", this, url.utf8().data());
    m_url = KURL(KURL(), url);

    if (!m_url.isValid()) {
        scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("Invalid url for WebSocket ", m_url.string()));
        m_state = CLOSED;
        ec = SYNTAX_ERR;
        return;
    }

    if (!m_url.protocolIs("ws") && !m_url.protocolIs("wss")) {
        scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("Wrong url scheme for WebSocket ", m_url.string()));
        m_state = CLOSED;
        ec = SYNTAX_ERR;
        return;
    }
    if (m_url.hasFragmentIdentifier()) {
        scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("URL has fragment component ", m_url.string()));
        m_state = CLOSED;
        ec = SYNTAX_ERR;
        return;
    }

    scriptExecutionContext()->contentSecurityPolicy()->upgradeInsecureRequestIfNeeded(m_url, ContentSecurityPolicy::InsecureRequestType::Load);

    if (!portAllowed(m_url)) {
        String message;
        if (m_url.port())
            message = tryMakeString("WebSocket port ", String::number(m_url.port().value()), " blocked");
        else
            message = ASCIILiteral("WebSocket without port blocked");
        scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, message);
        failAsynchronously();
        return;
    }

    // FIXME: Convert this to check the isolated world's Content Security Policy once webkit.org/b/104520 is solved.
    if (!scriptExecutionContext()->contentSecurityPolicy()->allowConnectToSource(m_url, scriptExecutionContext()->shouldBypassMainWorldContentSecurityPolicy())) {
        m_state = CLOSED;

        // FIXME: Should this be throwing an exception?
        ec = SECURITY_ERR;
        return;
    }

    m_channel = ThreadableWebSocketChannel::create(scriptExecutionContext(), this);

    // FIXME: There is a disagreement about restriction of subprotocols between WebSocket API and hybi-10 protocol
    // draft. The former simply says "only characters in the range U+0021 to U+007E are allowed," while the latter
    // imposes a stricter rule: "the elements MUST be non-empty strings with characters as defined in [RFC2616],
    // and MUST all be unique strings."
    //
    // Here, we throw SYNTAX_ERR if the given protocols do not meet the latter criteria. This behavior does not
    // comply with WebSocket API specification, but it seems to be the only reasonable way to handle this conflict.
    for (auto& protocol : protocols) {
        if (!isValidProtocolString(protocol)) {
            scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("Wrong protocol for WebSocket '", encodeProtocolString(protocol), "'"));
            m_state = CLOSED;
            ec = SYNTAX_ERR;
            return;
        }
    }
    HashSet<String> visited;
    for (auto& protocol : protocols) {
        if (visited.contains(protocol)) {
            scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("WebSocket protocols contain duplicates: '", encodeProtocolString(protocol), "'"));
            m_state = CLOSED;
            ec = SYNTAX_ERR;
            return;
        }
        visited.add(protocol);
    }

    if ((*scriptExecutionContext()).isDocument()) {
        Document& document = toDocument(*scriptExecutionContext());
        RefPtr<Frame> frame = document.frame();
        // FIXME: make the mixed content check equivalent to the non-document mixed content check currently in WorkerThreadableWebSocketChannel::Bridge::connect()
        // In particular we need to match the error messaging in the console and the inspector instrumentation. See WebSocketChannel::fail.
		if (!frame || !MixedContentChecker::frameAndAncestorsCanRunInsecureContent(*frame, *document.securityOrigin(), m_url)) {
			failAsynchronously();
            return;
        }
    }

    String protocolString;
    if (!protocols.isEmpty())
        protocolString = joinStrings(protocols, subProtocolSeperator());

    m_channel->connect(m_url, protocolString);

    m_pendingActivity = makePendingActivity(*this);
}

DLL_PUBLIC void WebSocket::send(const String& message, ExceptionCode& ec)
{
    LOG(Network, "WebSocket %p send %s", this, message.utf8().data());
    if (m_state == CONNECTING) {
        ec = INVALID_STATE_ERR;
        return;
    }
    // No exception is raised if the connection was once established but has subsequently been closed.
    if (m_state == CLOSING || m_state == CLOSED) {
        size_t payloadSize = message.utf8(StrictConversionReplacingUnpairedSurrogatesWithFFFD).length();
        m_bufferedAmountAfterClose = saturateAdd(m_bufferedAmountAfterClose, payloadSize);
        return;
    }
    ASSERT(m_channel);
    auto result = m_channel->send(message);
    if (result == ThreadableWebSocketChannel::SendResult::InvalidMessage) {
        scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("Websocket message contains invalid character(s)."));
        ec = SYNTAX_ERR;
        return;
    } else if (result == ThreadableWebSocketChannel::SendResult::Success) {
        size_t payloadSize = message.utf8(StrictConversionReplacingUnpairedSurrogatesWithFFFD).length();
        m_bufferedAmount = saturateAdd(m_bufferedAmount, payloadSize);
    }
}

void WebSocket::sendMessage(const char* message, int sz, int& ec)
{
    String msg = String::fromUTF8(message, sz);
    send(msg, ec);
}

DLL_PUBLIC bool WebSocket::sendPing(const char* data, int length)
{
    if (m_state == CONNECTING)
        return false;
    // No exception is raised if the connection was once established but has subsequently been closed.
    if (m_state == CLOSING || m_state == CLOSED) {
        size_t payloadSize = length;
        m_bufferedAmountAfterClose = saturateAdd(m_bufferedAmountAfterClose, payloadSize);
        m_bufferedAmountAfterClose = saturateAdd(m_bufferedAmountAfterClose, getFramingOverhead(payloadSize));
        return false;
    }
    ASSERT(m_channel);
    auto result = m_channel->sendPing(data, length);
    if (result == ThreadableWebSocketChannel::SendResult::Success)
        m_bufferedAmount = saturateAdd(m_bufferedAmount, static_cast<unsigned long>(length));
    return result == ThreadableWebSocketChannel::SendResult::Success;
}

void WebSocket::didReceivePong(PassOwnPtr<Vector<char> > pong)
{
    // TODO:: Notify upwards
    //if (m_didRecPongCB && m_wsIndex != -1) m_didRecPongCB(m_esp_context, m_wsIndex, pong->data(), pong->size());
    if (m_wsclient) m_wsclient->didReceivePong(pong);
}

void WebSocket::didReceivePing(PassOwnPtr<Vector<char> > ping)
{
    // TODO:: Notify upwards
    //if (m_didRecPingCB && m_wsIndex != -1) m_didRecPingCB(m_esp_context, m_wsIndex, ping->data(), ping->size());
    if (m_wsclient) m_wsclient->didReceivePing(ping);
}

void WebSocket::sendBinary(const char* buf, int sz, int& ec)
{
    if (RefPtr<JSC::ArrayBuffer> buffer = JSC::ArrayBuffer::tryCreate(buf, sz))
    	send(buffer.get(), ec);
}

DLL_PUBLIC void WebSocket::send(JSC::ArrayBuffer* binaryData, ExceptionCode& ec)
{
    LOG(Network, "WebSocket %p send arraybuffer %p", this, binaryData);
    ASSERT(binaryData);
    if (m_state == CONNECTING) {
        ec = INVALID_STATE_ERR;
        return;
    }
    if (m_state == CLOSING || m_state == CLOSED) {
        unsigned payloadSize = binaryData->byteLength();
        m_bufferedAmountAfterClose = saturateAdd(m_bufferedAmountAfterClose, payloadSize);
        return;
    }
    ASSERT(m_channel);
    auto result = m_channel->send(*binaryData, 0, binaryData->byteLength());
    if (result == ThreadableWebSocketChannel::SendResult::Success)
    	m_bufferedAmount = saturateAdd(m_bufferedAmount, binaryData->byteLength());
}

void WebSocket::send(JSC::ArrayBufferView* arrayBufferView, ExceptionCode& ec)
{
    LOG(Network, "WebSocket %p send arraybufferview %p", this, arrayBufferView);
    ASSERT(arrayBufferView);
    if (m_state == CONNECTING) {
        ec = INVALID_STATE_ERR;
        return;
    }
    if (m_state == CLOSING || m_state == CLOSED) {
        unsigned payloadSize = arrayBufferView->byteLength();
        m_bufferedAmountAfterClose = saturateAdd(m_bufferedAmountAfterClose, payloadSize);
        return;
    }
    ASSERT(m_channel);
    RefPtr<JSC::ArrayBuffer> arrayBuffer(arrayBufferView->buffer());
    auto result = m_channel->send(*arrayBuffer, arrayBufferView->byteOffset(), arrayBufferView->byteLength());
    if (result == ThreadableWebSocketChannel::SendResult::Success)
    	m_bufferedAmount = saturateAdd(m_bufferedAmount, arrayBufferView->byteLength());
}

void WebSocket::send(Blob* binaryData, ExceptionCode& ec)
{
    LOG(Network, "WebSocket %p send blob %s", this, binaryData->url().string().utf8().data());
    ASSERT(binaryData);
    if (m_state == CONNECTING) {
        ec = INVALID_STATE_ERR;
        return;
    }
    if (m_state == CLOSING || m_state == CLOSED) {
        unsigned long payloadSize = static_cast<unsigned long>(binaryData->size());
        m_bufferedAmountAfterClose = saturateAdd(m_bufferedAmountAfterClose, payloadSize);
        return;
    }
    ASSERT(m_channel);
    auto result = m_channel->send(*binaryData);
    if (result == ThreadableWebSocketChannel::SendResult::Success)
		m_bufferedAmount = saturateAdd(m_bufferedAmount, static_cast<unsigned long>(binaryData->size()));
}

DLL_PUBLIC void WebSocket::closeIgnoreException()
{
    // Yunfeng: refer to http://tools.ietf.org/html/rfc6455#section-7.4, Defined status codes
    ExceptionCode ec;
    close(1000, String(), ec);
}

void WebSocket::close(int code, const char* reason, int sz, int& ec)
{
    String rsn = String::fromUTF8(reason, sz);
    close(code, rsn, ec);
}

void WebSocket::close(int code, const String& reason, ExceptionCode& ec)
{
    if (code == WebSocketChannel::CloseEventCodeNotSpecified)
        LOG(Network, "WebSocket %p close without code and reason", this);
    else {
        LOG(Network, "WebSocket %p close with code = %d, reason = %s", this, code, reason.utf8().data());
        if (!(code == WebSocketChannel::CloseEventCodeNormalClosure || (WebSocketChannel::CloseEventCodeMinimumUserDefined <= code && code <= WebSocketChannel::CloseEventCodeMaximumUserDefined))) {
            ec = INVALID_ACCESS_ERR;
            return;
        }
        CString utf8 = reason.utf8(StrictConversionReplacingUnpairedSurrogatesWithFFFD);
        if (utf8.length() > maxReasonSizeInBytes) {
            scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("WebSocket close message is too long."));
            ec = SYNTAX_ERR;
            return;
        }
    }

    if (m_state == CLOSING || m_state == CLOSED)
        return;
    if (m_state == CONNECTING) {
        m_state = CLOSING;
        if (m_channel)
        	m_channel->fail(ASCIILiteral("WebSocket is closed before the connection is established."));
        return;
    }
    m_state = CLOSING;
    if (m_channel)
        m_channel->close(code, reason);
}

unsigned long WebSocket::bufferedAmount() const
{
    return saturateAdd(m_bufferedAmount, m_bufferedAmountAfterClose);
}

String WebSocket::binaryType() const
{
    switch (m_binaryType) {
    case BinaryTypeBlob:
        return "blob";
    case BinaryTypeArrayBuffer:
        return "arraybuffer";
    }
    ASSERT_NOT_REACHED();
    return String();
}

void WebSocket::setBinaryType(const String& binaryType, ExceptionCode& ec)
{
    if (binaryType == "blob") {
        m_binaryType = BinaryTypeBlob;
        return;
    }
    if (binaryType == "arraybuffer") {
        m_binaryType = BinaryTypeArrayBuffer;
        return;
    }
    ec = SYNTAX_ERR;
    scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("'", binaryType, "' is not a valid value for binaryType; binaryType remains unchanged."));
}

EventTargetInterface WebSocket::eventTargetInterface() const
{
    return WebSocketEventTargetInterfaceType;
}

ScriptExecutionContext* WebSocket::scriptExecutionContext() const
{
    return ActiveDOMObject::scriptExecutionContext();
}

void WebSocket::contextDestroyed()
{
    LOG(Network, "WebSocket %p scriptExecutionContext destroyed", this);
    ASSERT(!m_channel);
    ASSERT(m_state == CLOSED);
    ActiveDOMObject::contextDestroyed();
}

bool WebSocket::canSuspendForDocumentSuspension() const
{
    return true;
}

void WebSocket::suspend(ReasonForSuspension reason)
{
	ScriptDisallowedScope assertNoEventDispatch;

    if (!m_channel)
        return;

	if (reason == ActiveDOMObject::PageCache) {
		// This will cause didClose() to be called.
		m_channel->fail("WebSocket is closed due to suspension.");
		return;
	}

	m_channel->suspend();
}

void WebSocket::resume()
{
	ScriptDisallowedScope assertNoEventDispatch;

    if (m_channel)
        m_channel->resume();
}

void WebSocket::stop()
{
    if (m_channel)
        m_channel->disconnect();
    m_channel = nullptr;
    m_state = CLOSED;
    ActiveDOMObject::stop();
    m_pendingActivity = nullptr;
}

inline void WebSocket::didConnectTaskFunction(const DidConnectKeepingObjectAliveTask&)
{
    if (m_state == CLOSED)
        return;
	if (m_state != CONNECTING) {
		didClose(0, ClosingHandshakeIncomplete, WebSocketChannel::CloseEventCodeAbnormalClosure, emptyString());
		return;
	}
	ASSERT(scriptExecutionContext());
	m_state = OPEN;
	m_subprotocol = m_channel->subprotocol();
	m_extensions = m_channel->extensions();
	dispatchEvent(Event::create(eventNames().openEvent, false, false));

    if (m_wsclient)
    	m_wsclient->didConnect(m_subprotocol, m_extensions);
}

class WebSocket::DidConnectKeepingObjectAliveTask final : public ActiveDOMObject::KeepingObjectAliveTask<WebSocket>
{
public:
	DidConnectKeepingObjectAliveTask(WebSocket& thisObject) : KeepingObjectAliveTask<WebSocket>(thisObject) { }
private:
    virtual void operator()() final
    {
    	m_protectedObject->didConnectTaskFunction(*this);
    }
};

void WebSocket::didConnect()
{
    LOG(Network, "WebSocket %p didConnect", this);
    LOG(stderr, "\t==>    WebSocket %p didConnect\n", this);
    queueTaskKeepingObjectAlive(*this, TaskSource::WebSocket, adoptPtr(new DidConnectKeepingObjectAliveTask(*this)));
}

inline void WebSocket::didReceiveMessageTaskFunction(const DidReceiveMessageKeepingObjectAliveTask&, const String& msg)
{
	if (m_state != OPEN)
		return;
	ASSERT(scriptExecutionContext());
	dispatchEvent(MessageEvent::create(msg, SecurityOrigin::create(m_url)->toString()));

	if (m_wsclient)
		m_wsclient->didReceiveMessage(msg);
}

class WebSocket::DidReceiveMessageKeepingObjectAliveTask final : public ActiveDOMObject::KeepingObjectAliveTask<WebSocket>
{
public:
	DidReceiveMessageKeepingObjectAliveTask(WebSocket& thisObject, const String& msg)
		: KeepingObjectAliveTask<WebSocket>(thisObject)
		, msg(msg)
	{ }
private:
    virtual void operator()() final
    {
    	m_protectedObject->didReceiveMessageTaskFunction(*this, msg);
    }

    String msg;
};

void WebSocket::didReceiveMessage(const String& msg)
{
    LOG(Network, "WebSocket %p didReceiveMessage %s", this, msg.utf8().data());
    queueTaskKeepingObjectAlive(*this, TaskSource::WebSocket, adoptPtr(new DidReceiveMessageKeepingObjectAliveTask(*this, msg)));
}

inline void WebSocket::didReceiveBinaryDataTaskFunction(const DidReceiveBinaryDataKeepingObjectAliveTask&, PassOwnPtr<Vector<char>> binaryData)
{
	if (m_state != OPEN)
		return;
	switch (m_binaryType) {
	case BinaryTypeBlob: {
		size_t size = binaryData->size();
		RefPtr<RawData> rawData = RawData::create();
		binaryData->swap(*rawData->mutableData());
		OwnPtr<BlobData> blobData = BlobData::create();
		blobData->appendData(rawData.release(), 0, BlobDataItem::toEndOfFile);
		RefPtr<Blob> blob = Blob::create(scriptExecutionContext(), blobData.release(), size);
		dispatchEvent(MessageEvent::create(blob.release(), SecurityOrigin::create(m_url)->toString()));
		break;
	}

	case BinaryTypeArrayBuffer:
		dispatchEvent(MessageEvent::create(JSC::ArrayBuffer::tryCreate(binaryData->data(), binaryData->size()), SecurityOrigin::create(m_url)->toString()));

		if (m_wsclient)
			m_wsclient->didReceiveBinaryData(binaryData);
		break;
	}
}

class WebSocket::DidReceiveBinaryDataKeepingObjectAliveTask final : public ActiveDOMObject::KeepingObjectAliveTask<WebSocket>
{
public:
	DidReceiveBinaryDataKeepingObjectAliveTask(WebSocket& thisObject, PassOwnPtr<Vector<char>> binaryData)
		: KeepingObjectAliveTask<WebSocket>(thisObject)
		, binaryData(binaryData)
	{ }
private:
    virtual void operator()() final
    {
    	m_protectedObject->didReceiveBinaryDataTaskFunction(*this, binaryData.release());
    }

    OwnPtr<Vector<char>> binaryData;
};

void WebSocket::didReceiveBinaryData(PassOwnPtr<Vector<char>> binaryData)
{
	queueTaskKeepingObjectAlive(*this, TaskSource::WebSocket, adoptPtr(new DidReceiveBinaryDataKeepingObjectAliveTask(*this, binaryData)));
}

inline void WebSocket::didReceiveMessageErrorTaskFunction(const DidReceiveMessageErrorKeepingObjectAliveTask&)
{
    if (m_state == CLOSED)
        return;
	m_state = CLOSED;
	ASSERT(scriptExecutionContext());
	// FIXME: As per https://html.spec.whatwg.org/multipage/web-sockets.html#feedback-from-the-protocol:concept-websocket-closed, we should synchronously fire a close event.
	dispatchErrorEventIfNeeded();
}

class WebSocket::DidReceiveMessageErrorKeepingObjectAliveTask final : public ActiveDOMObject::KeepingObjectAliveTask<WebSocket>
{
public:
	DidReceiveMessageErrorKeepingObjectAliveTask(WebSocket& thisObject) : KeepingObjectAliveTask<WebSocket>(thisObject) { }
private:
    virtual void operator()() final
    {
    	m_protectedObject->didReceiveMessageErrorTaskFunction(*this);
    }
};

void WebSocket::didReceiveMessageError()
{
    LOG(Network, "WebSocket %p didReceiveErrorMessage", this);
    queueTaskKeepingObjectAlive(*this, TaskSource::WebSocket, adoptPtr(new DidReceiveMessageErrorKeepingObjectAliveTask(*this)));
}

void WebSocket::didUpdateBufferedAmount(unsigned long bufferedAmount)
{
    LOG(Network, "WebSocket %p didUpdateBufferedAmount %lu", this, bufferedAmount);
    if (m_state == CLOSED)
        return;
    m_bufferedAmount = bufferedAmount;
}

inline void WebSocket::didStartClosingHandshakeTaskFunction(const DidStartClosingHandshakeKeepingObjectAliveTask&)
{
    if (m_state == CLOSED)
        return;
	m_state = CLOSING;

	if (m_wsclient)
		m_wsclient->didStartClosingHandshake();
}

class WebSocket::DidStartClosingHandshakeKeepingObjectAliveTask final : public ActiveDOMObject::KeepingObjectAliveTask<WebSocket>
{
public:
	DidStartClosingHandshakeKeepingObjectAliveTask(WebSocket& thisObject) : KeepingObjectAliveTask<WebSocket>(thisObject) { }
private:
    virtual void operator()() final
    {
    	m_protectedObject->didStartClosingHandshakeTaskFunction(*this);
    }
};

void WebSocket::didStartClosingHandshake()
{
    LOG(Network, "WebSocket %p didStartClosingHandshake", this);
    queueTaskKeepingObjectAlive(*this, TaskSource::WebSocket, adoptPtr(new DidStartClosingHandshakeKeepingObjectAliveTask(*this)));
}

inline void WebSocket::didCloseTaskFunction(const DidCloseKeepingObjectAliveTask&, unsigned long unhandledBufferedAmount, ClosingHandshakeCompletionStatus closingHandshakeCompletion, unsigned short code, const String& reason)
{
	if (!m_channel)
		return;

	bool wasClean = m_state == CLOSING && !unhandledBufferedAmount && closingHandshakeCompletion == ClosingHandshakeComplete && code != WebSocketChannel::CloseEventCodeAbnormalClosure;
	m_state = CLOSED;
	m_bufferedAmount = unhandledBufferedAmount;
	ASSERT(scriptExecutionContext());

	dispatchEvent(CloseEvent::create(wasClean, code, reason));

	if (m_channel) {
		m_channel->disconnect();
		m_channel = nullptr;
	}
	m_pendingActivity = nullptr;

	if (m_wsclient)
		m_wsclient->didClose(unhandledBufferedAmount, closingHandshakeCompletion == ClosingHandshakeIncomplete? false:true, code, reason);
}

class WebSocket::DidCloseKeepingObjectAliveTask final : public ActiveDOMObject::KeepingObjectAliveTask<WebSocket>
{
public:
	DidCloseKeepingObjectAliveTask(WebSocket& thisObject, unsigned long unhandledBufferedAmount, ClosingHandshakeCompletionStatus closingHandshakeCompletion, unsigned short code, const String& reason)
		: KeepingObjectAliveTask<WebSocket>(thisObject)
		, unhandledBufferedAmount(unhandledBufferedAmount)
		, closingHandshakeCompletion(closingHandshakeCompletion)
		, code(code)
		, reason(reason)
	{ }
private:
    virtual void operator()() final
    {
    	m_protectedObject->didCloseTaskFunction(*this, unhandledBufferedAmount, closingHandshakeCompletion, code, reason);
    }
    unsigned long unhandledBufferedAmount;
    ClosingHandshakeCompletionStatus closingHandshakeCompletion;
    unsigned short code;
    String reason;
};

void WebSocket::didClose(unsigned long unhandledBufferedAmount, ClosingHandshakeCompletionStatus closingHandshakeCompletion, unsigned short code, const String& reason)
{
    LOG(Network, "WebSocket %p didClose", this);
    queueTaskKeepingObjectAlive(*this, TaskSource::WebSocket, adoptPtr(new DidCloseKeepingObjectAliveTask(*this, unhandledBufferedAmount, closingHandshakeCompletion, code, reason)));
}

void WebSocket::didUpgradeURL()
{
    ASSERT(m_url.protocolIs("ws"));
    m_url.setProtocol("wss");
}

size_t WebSocket::getFramingOverhead(size_t payloadSize)
{
    static const size_t hybiBaseFramingOverhead = 2; // Every frame has at least two-byte header.
    static const size_t hybiMaskingKeyLength = 4; // Every frame from client must have masking key.
    static const size_t minimumPayloadSizeWithTwoByteExtendedPayloadLength = 126;
    static const size_t minimumPayloadSizeWithEightByteExtendedPayloadLength = 0x10000;
    size_t overhead = hybiBaseFramingOverhead + hybiMaskingKeyLength;
    if (payloadSize >= minimumPayloadSizeWithEightByteExtendedPayloadLength)
        overhead += 8;
    else if (payloadSize >= minimumPayloadSizeWithTwoByteExtendedPayloadLength)
        overhead += 2;
    return overhead;
}

void WebSocket::dispatchErrorEventIfNeeded()
{
    if (m_dispatchedErrorEvent)
        return;

    m_dispatchedErrorEvent = true;
    dispatchEvent(Event::create(eventNames().errorEvent, false, false));
}

}  // namespace WebCore

#endif
