/*
 * Copyright (C) 2011 Google Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(WEB_SOCKETS)

#include "ActiveDOMObject.h"
#include "EventListener.h"
#include "EventNames.h"
#include "EventTarget.h"
#include "KURL.h"
#include "WebSocketChannel.h"
#include "WebSocketChannelClient.h"
#include "WebSocketClient.h"
#include <wtf/Forward.h>
#include <wtf/OwnPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/text/AtomicStringHash.h>

namespace WebCore {

class Blob;
class CloseEvent;
class ThreadableWebSocketChannel;

class WebSocket FINAL : public RefCounted<WebSocket>, public EventTargetWithInlineData, public ActiveDOMObject, public WebSocketChannelClient {
public:
    static void setIsAvailable(bool);
    static bool isAvailable();
    static ASCIILiteral subProtocolSeperator();
    static unsigned long saturateAdd(unsigned long a, unsigned long b);

    static PassRefPtr<WebSocket> create(ScriptExecutionContext*);
    static PassRefPtr<WebSocket> create(ScriptExecutionContext*, WebSocketClient* client);
    virtual ~WebSocket();

    enum State {
        CONNECTING = 0,
        OPEN = 1,
        CLOSING = 2,
        CLOSED = 3
    };

    // Espail PS
    void setCloseTimeout(int secs) {m_closeTimeout = secs;};
    void connect(const String& url, ExceptionCode&);
	// Espial PS
    void connect(const char* url, int&);
    void connect(const String& url, const String& protocol, ExceptionCode&);
	// Espial PS
    void connect(const char* url, const char* protocol, int&);
    void connect(const String& url, const Vector<String>& protocols, ExceptionCode&);
	// Espial PS
    void connect(const char* url, const char* protocols[], int sz, int&);

    void send(const String& message, ExceptionCode&);
    void send(JSC::ArrayBuffer*, ExceptionCode&);
    void send(JSC::ArrayBufferView*, ExceptionCode&);
    void send(Blob*, ExceptionCode&);

	// Espial PS
    void sendMessage(const char* message, int sz, int&);
    void sendBinary(const char* buf, int sz, int&);
    bool sendPing(const char* data, int length);

    void close(int code, const String& reason, ExceptionCode&);
    void close(ExceptionCode& ec) { close(WebSocketChannel::CloseEventCodeNotSpecified, String(), ec); }
    void close(int code, ExceptionCode& ec) { close(code, String(), ec); }

    void closeIgnoreException();
    void close(int code, const char* reason, int sz, int&);

    const KURL& url() const;
    State readyState() const;
    unsigned long bufferedAmount() const;

    const String& protocol() const;
    const String& extensions() const;

    String binaryType() const;
    void setBinaryType(const String&, ExceptionCode&);

    // EventTarget functions.
    virtual EventTargetInterface eventTargetInterface() const OVERRIDE;
    virtual ScriptExecutionContext* scriptExecutionContext() const OVERRIDE;

    // ActiveDOMObject functions.
    virtual void contextDestroyed() OVERRIDE;
    virtual bool canSuspendForDocumentSuspension() const OVERRIDE;
    virtual void suspend(ReasonForSuspension) OVERRIDE;
    virtual void resume() OVERRIDE;
    virtual void stop() OVERRIDE;
    virtual const char* activeDOMObjectName() const OVERRIDE { return "WebSocket"; }

    using RefCounted<WebSocket>::ref;
    using RefCounted<WebSocket>::deref;

    // WebSocketChannelClient functions.
    virtual void didConnect() FINAL;
    virtual void didReceiveMessage(const String& message) FINAL;
    virtual void didReceiveBinaryData(PassOwnPtr<Vector<char> >) FINAL;
    virtual void didReceiveMessageError() FINAL;
    virtual void didUpdateBufferedAmount(unsigned long bufferedAmount) FINAL;
    virtual void didStartClosingHandshake() FINAL;
    virtual void didClose(unsigned long unhandledBufferedAmount, ClosingHandshakeCompletionStatus, unsigned short code, const String& reason) FINAL;
    virtual void didUpgradeURL() FINAL;

	// Espial PS
    virtual void didReceivePing(PassOwnPtr<Vector<char> >) override;
    virtual void didReceivePong(PassOwnPtr<Vector<char> >) override;

private:
    WebSocket(ScriptExecutionContext*);
	// Espial PS
    WebSocket(ScriptExecutionContext* context, WebSocketClient* client);

    void dispatchErrorEventIfNeeded();

    virtual void refEventTarget() override { ref(); }
    virtual void derefEventTarget() override { deref(); }

    size_t getFramingOverhead(size_t payloadSize);

    void failAsynchronously();

    enum BinaryType {
        BinaryTypeBlob,
        BinaryTypeArrayBuffer
    };

    RefPtr<ThreadableWebSocketChannel> m_channel;

    State m_state { CONNECTING };
    KURL m_url;
    unsigned long m_bufferedAmount { 0 };
    unsigned long m_bufferedAmountAfterClose { 0 };
    BinaryType m_binaryType;
    String m_subprotocol;
    String m_extensions;

	// Espial PS
    WebSocketClient* m_wsclient { nullptr };
	// Espial PS
    int m_closeTimeout { 0 };

    bool m_dispatchedErrorEvent { false };
    RefPtr<PendingActivity<WebSocket>> m_pendingActivity;

    class FailAsynchronouslyKeepingObjectAliveTask;
    friend class FailAsynchronouslyKeepingObjectAliveTask;

    class DidConnectKeepingObjectAliveTask;
    friend class DidConnectKeepingObjectAliveTask;
    void didConnectTaskFunction(const DidConnectKeepingObjectAliveTask&);

    class DidReceiveMessageKeepingObjectAliveTask;
    friend class DidReceiveMessageKeepingObjectAliveTask;
    void didReceiveMessageTaskFunction(const DidReceiveMessageKeepingObjectAliveTask&, const String& msg);

    class DidReceiveBinaryDataKeepingObjectAliveTask;
    friend class DidReceiveBinaryDataKeepingObjectAliveTask;
    void didReceiveBinaryDataTaskFunction(const DidReceiveBinaryDataKeepingObjectAliveTask&, PassOwnPtr<Vector<char>> binaryData);

    class DidReceiveMessageErrorKeepingObjectAliveTask;
    friend class DidReceiveMessageErrorKeepingObjectAliveTask;
    void didReceiveMessageErrorTaskFunction(const DidReceiveMessageErrorKeepingObjectAliveTask&);

    class DidStartClosingHandshakeKeepingObjectAliveTask;
    friend class DidStartClosingHandshakeKeepingObjectAliveTask;
    void didStartClosingHandshakeTaskFunction(const DidStartClosingHandshakeKeepingObjectAliveTask&);

    class DidCloseKeepingObjectAliveTask;
    friend class DidCloseKeepingObjectAliveTask;
    void didCloseTaskFunction(const DidCloseKeepingObjectAliveTask&, unsigned long unhandledBufferedAmount, ClosingHandshakeCompletionStatus closingHandshakeCompletion, unsigned short code, const String& reason);
};

inline const KURL& WebSocket::url() const
{
    return m_url;
}

inline WebSocket::State WebSocket::readyState() const
{
    return m_state;
}

inline const String& WebSocket::protocol() const
{
    return m_subprotocol;
}

inline const String& WebSocket::extensions() const
{
    return m_extensions;
}

} // namespace WebCore

#endif // ENABLE(WEB_SOCKETS)
