/*
 * Copyright (C) 2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "AnimationEffect.h"
#include "AnimationEffectPhase.h"
#include "Styleable.h"
#include "WebAnimation.h"
#include <wtf/PassRefPtr.h>
#include <wtf/WeakPtr.h>

namespace WebCore {

class Animation;
class AnimationEventBase;
class Element;
class RenderStyle;

class DeclarativeAnimation : public WebAnimation {
public:
    virtual ~DeclarativeAnimation();

    virtual bool isDeclarativeAnimation() const FINAL { return true; }

    Optional<Styleable> owningElement() const;
    const Animation& backingAnimation() const { return *m_backingAnimation; }
    void setBackingAnimation(const Animation&);
    void cancelFromStyle();

    Optional<double> bindingsStartTime() const FINAL;
    virtual Optional<double> bindingsCurrentTime() const FINAL;
    virtual WebAnimation::PlayState bindingsPlayState() const FINAL;
    WebAnimation::ReplaceState bindingsReplaceState() const FINAL;
    virtual bool bindingsPending() const;
    virtual ReadyPromise& bindingsReady();
    virtual FinishedPromise& bindingsFinished();
    virtual void bindingsPlay(ExceptionCode&);
    virtual void bindingsPause(ExceptionCode&);

    virtual void setTimeline(PassRefPtr<AnimationTimeline>) FINAL;
    virtual void cancel() FINAL;

    virtual void tick() OVERRIDE;

    virtual bool canHaveGlobalPosition() FINAL;

    void flushPendingStyleChanges() const;

protected:
    DeclarativeAnimation(const Styleable&, const Animation&);

    virtual void initialize(const RenderStyle* oldStyle, const RenderStyle& newStyle, const Style::ResolutionContext&);
    virtual void syncPropertiesWithBackingAnimation();
    // elapsedTime is the animation's current time at the time the event is added and is exposed through the DOM API, timelineTime is the animations'
	// timeline current time and is not exposed through the DOM API but used by the DocumentTimeline for sorting events before dispatch.
	virtual Ref<AnimationEventBase> createEvent(const AtomicString& eventType, double elapsedTime, const String& pseudoId, Optional<Seconds> timelineTime) = 0;
	void invalidateDOMEvents(Seconds elapsedTime = Seconds(0));

private:
    void disassociateFromOwningElement();
    AnimationEffectPhase phaseWithoutEffect() const;
    void enqueueDOMEvent(const AtomicString&, Seconds);

    WeakPtr<Element> m_owningElement;
    PseudoId m_owningPseudoId;
    RefPtr<Animation> m_backingAnimation;
    bool m_wasPending;
	AnimationEffectPhase m_previousPhase;
	double m_previousIteration;

};

} // namespace WebCore

//SPECIALIZE_TYPE_TRAITS_WEB_ANIMATION(DeclarativeAnimation, isDeclarativeAnimation())
