/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "AnimationTimeline.h"
#include "DocumentTimelineOptions.h"
#include "GenericTaskQueue.h"
#include "ReducedResolutionSeconds.h"
#include "Timer.h"
#include <wtf/PassRefPtr.h>
#include <wtf/WeakPtr.h>

namespace WebCore {

class AnimationEventBase;
class DocumentTimelinesController;
class RenderBoxModelObject;
class RenderElement;

class DocumentTimeline FINAL : public AnimationTimeline
{
public:
    static PassRefPtr<DocumentTimeline> create(Document&);
    static PassRefPtr<DocumentTimeline> create(Document*, Optional<DocumentTimelineOptions>);
    virtual ~DocumentTimeline();

    virtual bool isDocumentTimeline() const FINAL { return true; }

    Document* document() const { return m_document.get(); }

    virtual Optional<Seconds> currentTime() OVERRIDE;

    virtual void animationTimingDidChange(WebAnimation&) OVERRIDE;
    virtual void removeAnimation(WebAnimation&) OVERRIDE;
    void transitionDidComplete(RefPtr<CSSTransition>);

    PassOwnPtr<RenderStyle> animatedStyleForRenderer(RenderElement& renderer);
#if USE(ACCELERATED_COMPOSITING) && !PLATFORM(ESPIAL)
    void animationAcceleratedRunningStateDidChange(WebAnimation&);
    void applyPendingAcceleratedAnimations();
#endif
    bool runningAnimationsForRendererAreAllAccelerated(const RenderBoxModelObject&) const;
	void detachFromDocument();

    void enqueueAnimationEvent(AnimationEventBase&);

    enum class ShouldUpdateAnimationsAndSendEvents : uint8_t { Yes, No };
    ShouldUpdateAnimationsAndSendEvents documentWillUpdateAnimationsAndSendEvents(ReducedResolutionSeconds);
    void removeReplacedAnimations();
    AnimationEvents prepareForPendingAnimationEventsDispatch();
    void documentDidUpdateAnimationsAndSendEvents();

    void updateThrottlingState();
    Seconds animationInterval() const;

    void maybeClearCachedCurrentTime();

    void vmIdleCallback();

private:
    DocumentTimeline(Document&, Seconds);

    DocumentTimelinesController* controller() const;
    ReducedResolutionSeconds liveCurrentTime() const;
    void cacheCurrentTime(ReducedResolutionSeconds);
    void scheduleInvalidationTaskIfNeeded();
    void scheduleAnimationResolution();
    void clearTickScheduleTimer();
    void internalUpdateAnimationsAndSendEvents();
    void scheduleNextTick();
    bool animationCanBeRemoved(WebAnimation&);
    bool shouldRunUpdateAnimationsAndSendEventsIgnoringSuspensionState() const;

    WeakPtr<Document> m_document;
    Seconds m_originTime;
    Optional<Seconds> m_cachedCurrentTime;
    GenericTaskQueue<DocumentTimeline> m_currentTimeClearingTaskQueue;
    HashSet<RefPtr<WebAnimation>> m_acceleratedAnimationsPendingRunningStateChange;
    AnimationEvents m_pendingAnimationEvents;
    Timer<DocumentTimeline> m_tickScheduleTimer;
    bool m_isSuspended;
    bool m_waitingOnVMIdle;
    bool m_animationResolutionScheduled;
    bool m_shouldScheduleAnimationResolutionForNewPendingEvents;
};

} // namespace WebCore

// SPECIALIZE_TYPE_TRAITS_ANIMATION_TIMELINE(DocumentTimeline, isDocumentTimeline())
