/*
 * Copyright (C) 2017-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ActiveDOMObject.h"
#include "AnimationPlayState.h"
#include "AnimationReplaceState.h"
#include "DOMPromiseProxy.h"
#include "EventTarget.h"
#include "ExceptionCode.h"
#include "JSDOMPromiseDeferred.h"
#include "Styleable.h"
#include "WebAnimationTypes.h"
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/PassRefPtr.h>
#include <wtf/Ref.h>
#include <wtf/RefCounted.h>
#include <wtf/RefPtr.h>
#include <wtf/Seconds.h>
#include <wtf/WeakPtr.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class AnimationEffect;
class AnimationEventBase;
class AnimationTimeline;
class Document;
class RenderStyle;

namespace Style {
struct ResolutionContext;
}

class WebAnimation : public RefCounted<WebAnimation>, public EventTargetWithInlineData, public ActiveDOMObject {
public:
    static PassRefPtr<WebAnimation> create(Document*, AnimationEffect*);
    static PassRefPtr<WebAnimation> create(Document*, AnimationEffect*, AnimationTimeline*);
    virtual ~WebAnimation();

    static HashSet<WebAnimation*>& instances();

    virtual bool isDeclarativeAnimation() const { return false; }
    virtual bool isCSSAnimation() const { return false; }
    virtual bool isCSSTransition() const { return false; }

    const String& id() const { return m_id; }
    void setId(const String& id) { m_id = id; }

    AnimationEffect* bindingsEffect() const { return effect(); }
    virtual void setBindingsEffect(PassRefPtr<AnimationEffect>);
    AnimationEffect* effect() const { return m_effect.get(); }
    void setEffect(PassRefPtr<AnimationEffect>);
    AnimationTimeline* timeline() const;
    virtual void setTimeline(PassRefPtr<AnimationTimeline>);

    Optional<Seconds> currentTime(Optional<Seconds> = WTF::Nullopt) const;
    void setCurrentTime(Optional<Seconds>, ExceptionCode&);

    double playbackRate() const { return m_playbackRate + 0; }
    void setPlaybackRate(double);

    typedef AnimationPlayState PlayState;
    PlayState playState() const;

    typedef AnimationReplaceState ReplaceState;
    ReplaceState replaceState() const { return m_replaceState; }
    void setReplaceState(ReplaceState replaceState) { m_replaceState = replaceState; }

    bool pending() const { return hasPendingPauseTask() || hasPendingPlayTask(); }

    typedef DOMPromiseProxyWithResolveCallback<WebAnimation> ReadyPromise;
    ReadyPromise& ready() { return *m_readyPromise; }

    typedef DOMPromiseProxyWithResolveCallback<WebAnimation> FinishedPromise;
    FinishedPromise& finished() { return *m_finishedPromise; }

    virtual void cancel();
    void finish(ExceptionCode&);
    void play(ExceptionCode&);
    void updatePlaybackRate(double);
    void pause(ExceptionCode&);
    virtual void bindingsReverse(ExceptionCode&);
    void reverse(ExceptionCode&);
    void persist();
    void commitStyles(ExceptionCode&);

    virtual Optional<double> bindingsStartTime() const;
    virtual void setBindingsStartTime(Optional<double>);
    Optional<Seconds> startTime() const { return m_startTime; }
    void setStartTime(Optional<Seconds>);
    virtual Optional<double> bindingsCurrentTime() const;
    virtual void setBindingsCurrentTime(Optional<double>, ExceptionCode&);
    virtual PlayState bindingsPlayState() const { return playState(); }
    virtual ReplaceState bindingsReplaceState() const { return replaceState(); }
    virtual bool bindingsPending() const { return pending(); }
    virtual ReadyPromise& bindingsReady() { return ready(); }
    virtual FinishedPromise& bindingsFinished() { return finished(); }
    virtual void bindingsPlay(ExceptionCode& ec) { return play(ec); }
    virtual void bindingsPause(ExceptionCode& ec) { return pause(ec); }

    bool needsTick() const;
    virtual void tick();
    Seconds timeToNextTick() const;
    virtual void resolve(RenderStyle& targetStyle, const Style::ResolutionContext&, Optional<Seconds> = WTF::Nullopt);
    void effectTargetDidChange(const Optional<Styleable>& previousTarget, const Optional<Styleable>& newTarget);
#if USE(ACCELERATED_COMPOSITING) && !PLATFORM(ESPIAL)
    void acceleratedStateDidChange();
    void applyPendingAcceleratedActions();
#endif
    void willChangeRenderer();

    bool isRunningAccelerated() const;
    bool isRelevant() const { return m_isRelevant; }
    void updateRelevance();
    void effectTimingDidChange();
    void suspendEffectInvalidation();
    void unsuspendEffectInvalidation();
    void setSuspended(bool);
    bool isSuspended() const { return m_isSuspended; }
    bool isReplaceable() const;
    void remove();
    void enqueueAnimationPlaybackEvent(const AtomicString&, Optional<Seconds>, Optional<Seconds>);

    uint64_t globalPosition() const { return m_globalPosition; }
    void setGlobalPosition(uint64_t globalPosition) { m_globalPosition = globalPosition; }

    virtual bool canHaveGlobalPosition() { return true; }

    // ContextDestructionObserver.
    virtual ScriptExecutionContext* scriptExecutionContext() const FINAL { return ActiveDOMObject::scriptExecutionContext(); }
    virtual void contextDestroyed() FINAL;

    String description();

    using RefCounted::ref;
	using RefCounted::deref;

    void performMicrotask();

    bool hasPendingPauseTask() const { return m_timeToRunPendingPauseTask != TimeToRunPendingTask::NotScheduled; }
    bool hasPendingPlayTask() const { return m_timeToRunPendingPlayTask != TimeToRunPendingTask::NotScheduled; }

    void runPendingPauseTask();
    void runPendingPlayTask();

protected:
	explicit WebAnimation(Document&);

    void enqueueAnimationEvent(Ref<AnimationEventBase>&&);

private:
    enum class DidSeek { Yes, No };
    enum class SynchronouslyNotify { Yes, No };
    enum class Silently : uint8_t { Yes, No };
    enum class RespectHoldTime { Yes, No };
    enum class AutoRewind { Yes, No };
    enum class TimeToRunPendingTask { NotScheduled, ASAP, WhenReady };

    void timingDidChange(DidSeek, SynchronouslyNotify, Silently = Silently::No);
    void updateFinishedState(DidSeek, SynchronouslyNotify);
	Seconds effectEndTime() const;
    WebAnimation& readyPromiseResolve();
    WebAnimation& finishedPromiseResolve();
    Optional<Seconds> currentTime(RespectHoldTime, Optional<Seconds> = WTF::Nullopt) const;
    void silentlySetCurrentTime(Optional<Seconds>, ExceptionCode&);
    void finishNotificationSteps();
    void play(AutoRewind, ExceptionCode&);
    void resetPendingTasks();
    void setEffectInternal(PassRefPtr<AnimationEffect>, bool = false);
    void setTimelineInternal(PassRefPtr<AnimationTimeline>);
    bool isEffectInvalidationSuspended() { return m_suspendCount; }
    bool computeRelevance();
    void invalidateEffect();
    double effectivePlaybackRate() const;
    void applyPendingPlaybackRate();

    String m_id;
    RefPtr<AnimationEffect> m_effect;
    WeakPtr<AnimationTimeline> m_timeline;
    Optional<Seconds> m_previousCurrentTime;
    Optional<Seconds> m_startTime;
    Optional<Seconds> m_holdTime;
    int m_suspendCount;
    double m_playbackRate;
    Optional<double> m_pendingPlaybackRate;
    bool m_isSuspended;
    bool m_finishNotificationStepsMicrotaskPending;
    bool m_isRelevant;
    bool m_shouldSkipUpdatingFinishedStateWhenResolving;
    bool m_hasScheduledEventsDuringTick;
    OwnPtr<ReadyPromise> m_readyPromise;
    OwnPtr<FinishedPromise> m_finishedPromise;
    TimeToRunPendingTask m_timeToRunPendingPlayTask;
    TimeToRunPendingTask m_timeToRunPendingPauseTask;
    ReplaceState m_replaceState;
    uint64_t m_globalPosition;

    // ActiveDOMObject.
    virtual const char* activeDOMObjectName() const FINAL;
    virtual void suspend(ReasonForSuspension) FINAL;
    virtual void resume() FINAL;
    virtual void stop() FINAL;
    virtual bool virtualHasPendingActivity() const final;

    // EventTarget
    virtual EventTargetInterface eventTargetInterface() const FINAL { return WebAnimationEventTargetInterfaceType; }
    virtual void refEventTarget() FINAL { ref(); }
    virtual void derefEventTarget() FINAL { deref(); }
};

inline bool isCSSAnimation(const WebAnimation* animation)
{
	return animation && animation->isCSSAnimation();
}

#if 0
#define SPECIALIZE_TYPE_TRAITS_WEB_ANIMATION(ToValueTypeName, predicate) \
SPECIALIZE_TYPE_TRAITS_BEGIN(WebCore::ToValueTypeName) \
static bool isType(const WebCore::WebAnimation& value) { return value.predicate; } \
SPECIALIZE_TYPE_TRAITS_END()
#endif

} // namespace WebCore
