/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 *  Copyright (C) 2007 Samuel Weinig <sam@webkit.org>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#if !HAVE(BINDINGS_ALL_IN_ONE)
#include "JSDOMBinding.h"

#include "CachedScript.h"
#include "DOMStringList.h"
#include "ExceptionCode.h"
#include "ExceptionCodeDescription.h"
#include "ExceptionHeaders.h"
#include "ExceptionInterfaces.h"
#include "Frame.h"
#include "HTMLFrameElementBase.h"
#include "HTMLParserIdioms.h"
#include "JSDOMConvert.h"
#include "JSDOMWindowCustom.h"
#include "JSExceptionBase.h"
#include "SecurityOrigin.h"
#include "ScriptCallStack.h"
#include "ScriptCallStackFactory.h"
#include "SourceProvider.h"
#include <bytecode/CodeBlock.h>
#include <interpreter/Interpreter.h>
#include <interpreter/StackVisitor.h>
#include <runtime/DateInstance.h>
#include <runtime/Error.h>
#include <runtime/ErrorInstance.h>
#include <runtime/Exception.h>
#include <runtime/ExceptionHelpers.h>
#include <runtime/JSFunction.h>
#include <runtime/ThrowScope.h>
#include <wtf/MathExtras.h>
#include <wtf/unicode/CharacterNames.h>

#if ENABLE(RMAI)
#include "JSMOMException.h"
#include "MOMException.h"
#endif

using namespace JSC;

namespace WebCore {

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSDOMConstructorBase);
ASSERT_HAS_TRIVIAL_DESTRUCTOR(DOMConstructorWithDocument);

void addImpureProperty(const AtomicString& propertyName)
{
    JSDOMWindow::commonJSGlobalData()->addImpureProperty(propertyName);
}

JSValue jsOwnedStringOrNull(ExecState* exec, const String& s)
{
    if (s.isNull())
        return jsNull();
    return jsOwnedString(exec->vm(), s);
}

JSValue jsStringOrUndefined(ExecState* exec, const String& s)
{
    if (s.isNull())
        return jsUndefined();
    return jsStringWithCache(exec, s);
}

JSValue jsString(ExecState* exec, const KURL& url)
{
    return jsStringWithCache(exec, url.string());
}

JSValue jsStringOrUndefined(ExecState* exec, const KURL& url)
{
    if (url.isNull())
        return jsUndefined();
    return jsString(exec->vm(), url.string());
}

String valueToStringWithNullCheck(ExecState* exec, JSValue value)
{
    if (value.isNull())
        return String();
    return value.toWTFString(exec);
}

String valueToStringTreatingNullAsEmptyString(ExecState* exec, JSValue value)
{
    if (value.isNull())
        return emptyString();
    return value.toWTFString(exec);
}

String valueToStringWithUndefinedOrNullCheck(ExecState* exec, JSValue value)
{
    if (value.isUndefinedOrNull())
        return String();
    return value.toWTFString(exec);
}

static inline String stringToByteString(ExecState& state, JSC::ThrowScope& scope, String&& string)
{
    if (!string.containsOnlyLatin1()) {
        throwTypeError(&state, scope.vm());
        return String();
    }

    return string;
}

String identifierToByteString(ExecState& state, const Identifier& identifier)
{
    VM& vm = state.vm();
    ThrowScope scope(vm);

    String string = identifier.string();
    return stringToByteString(state, scope, WTFMove(string));
}

String valueToByteString(ExecState& state, JSValue value)
{
    VM& vm = state.vm();
    ThrowScope scope(vm);

    String string = value.toWTFString(&state);
    RETURN_IF_EXCEPTION(vm, String());

    return stringToByteString(state, scope, WTFMove(string));
}

String valueToByteStringTreatingNullAsEmptyString(ExecState* exec, JSValue value)
{
    return value.isNull() ? emptyString() : valueToByteString(exec, value);
}

String valueToByteStringWithUndefinedOrNullCheck(ExecState* exec, JSValue value)
{
    return value.isUndefinedOrNull() ? String() : valueToByteString(exec, value);
}

static inline String stringToUSVString(String&& string)
{
	// Fast path for the case where there are no unpaired surrogates.
    if (!hasUnpairedSurrogate(string))
        return WTFMove(string);

    // Slow path: http://heycam.github.io/webidl/#dfn-obtain-unicode
    // Replaces unpaired surrogates with the replacememnt character.
    StringBuilder result;
    result.reserveCapacity(string.length());
    StringView view(string);
    for (auto codePoint : view.codePoints()) {
        if (U_IS_SURROGATE(codePoint))
            result.append(replacementCharacter);
        else
            result.appendCharacter(codePoint);
    }
    return result.toString();
}

String identifierToUSVString(ExecState&, const Identifier& identifier)
{
    String string = identifier.string();
    return stringToUSVString(WTFMove(string));
}

String valueToUSVString(ExecState& state, JSValue value)
{
    VM& vm = state.vm();
    ThrowScope scope(vm);

    String string = value.toWTFString(&state);
    RETURN_IF_EXCEPTION(scope, String());

    return stringToUSVString(WTFMove(string));
}

String valueToUSVStringTreatingNullAsEmptyString(ExecState* exec, JSValue value)
{
    return value.isNull() ? emptyString() : valueToUSVString(exec, value);
}

String valueToUSVStringWithUndefinedOrNullCheck(ExecState* exec, JSValue value)
{
    return value.isUndefinedOrNull() ? String() : valueToUSVString(exec, value);
}

JSValue jsDate(ExecState* exec, double value)
{
    return DateInstance::create(exec->vm(), exec->lexicalGlobalObject()->dateStructure(), value);
}

JSValue jsDateOrNull(ExecState* exec, double value)
{
    // For nullable Date types, we use NaN as null value.
    if (std::isnan(value))
        return jsNull();
    return DateInstance::create(exec->vm(), exec->lexicalGlobalObject()->dateStructure(), value);
}

double valueToDate(ExecState* exec, JSValue value)
{
    if (value.isNumber())
        return value.asNumber();
    if (!value.inherits(DateInstance::info()))
        return std::numeric_limits<double>::quiet_NaN();
    return static_cast<DateInstance*>(value.toObject(exec))->internalNumber();
}

JSC::JSValue jsArrayOrNull(JSC::ExecState* exec, JSDOMGlobalObject* globalObject, PassRefPtr<DOMStringList> stringList)
{
    JSC::MarkedArgumentBuffer list;
    if (stringList) {
        for (unsigned i = 0; i < stringList->length(); ++i)
            list.append(jsStringWithCache(exec, stringList->item(i)));
    }
    if (auto* result = JSC::constructArray(exec, 0, globalObject, list))
    	return result;
    return JSC::jsNull();
}

void callFunctionWithCurrentArguments(JSC::ExecState& state, JSC::JSObject& thisObject, JSC::JSFunction& function)
{
	auto& vm = state.vm();
    auto callData = JSC::getCallData(vm, &function);
    ASSERT(callType != CallType::None);

    JSC::MarkedArgumentBuffer arguments;
    for (unsigned i = 0; i < state.argumentCount(); ++i)
        arguments.append(state.uncheckedArgument(i));
    JSC::call(&state, &function, callData, &thisObject, arguments);
}

void DOMConstructorJSBuiltinObject::visitChildren(JSC::JSCell* cell, JSC::SlotVisitor& visitor)
{
    auto* thisObject = jsCast<DOMConstructorJSBuiltinObject*>(cell);
    ASSERT_GC_OBJECT_INHERITS(thisObject, info());
    Base::visitChildren(thisObject, visitor);
    visitor.append(&thisObject->m_initializeFunction);
}

EncodedJSValue JSC_HOST_CALL callThrowTypeErrorForJSDOMConstructor(ExecState* exec)
{
    throwTypeError(exec, exec->vm(), ASCIILiteral("Constructor requires 'new' operator"));
    return JSValue::encode(jsNull());
}

EncodedJSValue JSC_HOST_CALL callThrowTypeErrorForJSDOMConstructorNotConstructable(ExecState* exec)
{
    JSC::VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    JSC::throwTypeError(exec, vm, "Illegal constructor");
    return JSC::JSValue::encode(JSC::jsNull());
}

String JSDOMConstructorBase::className(const JSObject*)
{
	return ASCIILiteral("Function");
}

String JSDOMConstructorBase::toStringName(const JSObject* object, JSC::ExecState* exec)
{
    VM& vm = exec->vm();
    const ClassInfo* info = object->classInfo();
    ASSERT(info);
    return info->methodTable.className(object);
}

NEVER_INLINE void reportException(ExecState* exec, JSValue exceptionValue, CachedScript* cachedScript)
{
    RELEASE_ASSERT(exec->vm().currentThreadIsHoldingAPILock());
    JSC::Exception* exception = jsDynamicCast<JSC::Exception*>(exceptionValue);
    if (!exception) {
        exception = exec->lastException();
        if (!exception)
            exception = JSC::Exception::create(exec->vm(), exceptionValue, JSC::Exception::DoNotCaptureStack);
    }

    reportException(exec, exception, cachedScript);
}

NEVER_INLINE String retrieveErrorMessage(ExecState& state, VM& vm, JSValue exception, JSGlobalData& catchScope)
{
    if (auto* exceptionBase = toExceptionBase(exception))
        return exceptionBase->toString();

    // FIXME: <http://webkit.org/b/115087> Web Inspector: WebCore::reportException should not evaluate JavaScript handling exceptions
    // If this is a custom exception object, call toString on it to try and get a nice string representation for the exception.
    String errorMessage;
    if (auto* error = jsDynamicCast<ErrorInstance*>(exception))
        errorMessage = error->sanitizedToString(&state);
    else
        errorMessage = exception.toWTFString(&state);

    // We need to clear any new exception that may be thrown in the toString() call above.
    // reportException() is not supposed to be making new exceptions.
    catchScope.clearException();
    vm.clearLastException();
    return errorMessage;
}

NEVER_INLINE void reportException(ExecState* exec, JSC::Exception* exception, CachedScript* cachedScript)
{
    if (isTerminatedExecutionException(exception))
        return;

    VM& vm = exec->vm();

    Interpreter::ErrorHandlingMode mode(exec);

    RefPtr<ScriptCallStack> callStack(createScriptCallStackFromException(exec, exception, ScriptCallStack::maxCallStackSizeToCapture));
    vm.clearException();
    vm.clearLastException();

    JSDOMGlobalObject* globalObject = jsCast<JSDOMGlobalObject*>(exec->lexicalGlobalObject());
    if (JSDOMWindow* window = jsDynamicCast<JSDOMWindow*>(globalObject)) {
        if (!window->wrapped().isCurrentlyDisplayedInFrame())
            return;
    }

    int lineNumber = 0;
    int columnNumber = 0;
    String exceptionSourceURL;
    if (const ScriptCallFrame* callFrame = callStack->firstNonNativeCallFrame()) {
        lineNumber = callFrame->lineNumber();
        columnNumber = callFrame->columnNumber();
        exceptionSourceURL = callFrame->sourceURL();
    }

    String errorMessage = retrieveErrorMessage(*exec, vm, exception->value(), vm);

    if(ScriptExecutionContext* scriptExecutionContext = globalObject->scriptExecutionContext())
    	scriptExecutionContext->reportException(errorMessage, lineNumber, columnNumber, exceptionSourceURL, exception, callStack->size() ? callStack : nullptr, cachedScript);
}

NEVER_INLINE void reportCurrentException(ExecState* exec)
{
	JSC::Exception* exception = exec->exception();
    exec->clearException();
    reportException(exec, exception);
}

#define TRY_TO_CREATE_EXCEPTION(interfaceName) \
    case interfaceName##Type: \
        errorObject = toJS(exec, globalObject, interfaceName::create(description)); \
        break;

static JSValue createDOMException(ExecState* exec, ExceptionCode ec, const String* message)
{
    if (!ec)
        return jsUndefined();

    // FIXME: Handle other WebIDL exception types.
    if (ec == TypeError) {
        if (!message || message->isEmpty())
        	return createTypeError(exec, String("Type error", String::ConstructFromLiteral));
        return createTypeError(exec, *message);
    }

    if (ec == RangeError) {
        if (!message || message->isEmpty())
            return createRangeError(exec, ASCIILiteral("Bad value"));
        return createRangeError(exec, *message);
    }

    // FIXME: All callers to setDOMException need to pass in the right global object
    // for now, we're going to assume the lexicalGlobalObject. Which is wrong in cases like this:
    // frames[0].document.createElement(null, null); // throws an exception which should have the subframes prototypes.
    JSDOMGlobalObject* globalObject = deprecatedGlobalObjectForPrototype(exec);

    ExceptionCodeDescription description(ec);

    CString messageCString;
    if (message)
        messageCString = message->utf8();
    if (message && !message->isEmpty()) {
        // It is safe to do this because the char* contents of the CString are copied into a new WTF::String before the CString is destroyed.
        description.description = messageCString.data();
    }

    JSValue errorObject;
    switch (description.type) {
        DOM_EXCEPTION_INTERFACES_FOR_EACH(TRY_TO_CREATE_EXCEPTION)
    }

done:
    ASSERT(errorObject);
    addErrorInfo(exec, asObject(errorObject), true);
    return errorObject;
}

NEVER_INLINE JSValue createDOMException(ExecState* exec, ExceptionCode ec, const String& message)
{
	return createDOMException(exec, ec, &message);
}

NEVER_INLINE JSValue createDOMException(ExecState* exec, ExceptionCode ec)
{
    return createDOMException(exec, ec, nullptr);
}

NEVER_INLINE void setDOMException(ExecState* exec, ExceptionCode ec, const ASCIILiteral& message)
{
	setDOMException(exec, ec, String(message));
}

NEVER_INLINE void setDOMException(ExecState* exec, ExceptionCode ec, const String& message)
{
    VM& vm = exec->vm();
    ThrowScope scope(vm);

    if (!ec || scope.exception())
        return;

    throwException(exec, scope, createDOMException(exec, ec, message));
}

NEVER_INLINE void setDOMException(ExecState* exec, ExceptionCode ec)
{
	VM& vm = exec->vm();
    RETURN_IF_EXCEPTION(vm, void());

    if (UNLIKELY(ec))
    	vm.throwException(exec, createDOMException(exec, ec));
}

#undef TRY_TO_CREATE_EXCEPTION

bool hasIteratorMethod(JSC::ExecState& state, JSC::JSObject* object)
{
    auto& vm = state.vm();
    CallData callData;
    JSValue applyMethod = object->getMethod(&state, callData, vm.propertyNames->iteratorSymbol, ASCIILiteral("Symbol.iterator property should be callable"));
    RETURN_IF_EXCEPTION(vm, false);

    return !applyMethod.isUndefined();
}

bool hasIteratorMethod(JSC::ExecState& state, JSC::JSValue value)
{
    if (!value.isObject())
        return false;

    return hasIteratorMethod(state, JSC::asObject(value));
}

bool shouldAllowAccessToNode(ExecState* exec, Node* node)
{
    return BindingSecurity::shouldAllowAccessToNode(exec, node);
}

bool shouldAllowAccessToFrame(ExecState* exec, Frame* target, SecurityReportingOption reportingOption)
{
    return BindingSecurity::shouldAllowAccessToFrame(exec, target, reportingOption);
}

bool shouldAllowAccessToFrame(ExecState* exec, Frame* frame, String& message)
{
    if (!frame)
        return false;
    if (BindingSecurity::shouldAllowAccessToFrame(exec, frame, DoNotReportSecurityError))
        return true;
    // FIXME: The following line of code should move somewhere that it can be shared with immediatelyReportUnsafeAccessTo.
    message = frame->document()->domWindow()->crossDomainAccessErrorMessage(activeDOMWindow(exec));
    return false;
}

bool shouldAllowAccessToDOMWindow(ExecState* exec, DOMWindow* target, String& message)
{
    if (!target)
        return false;
    if (BindingSecurity::shouldAllowAccessToDOMWindow(exec, *target, DoNotReportSecurityError))
        return true;
    // FIXME: The following line of code should move somewhere that it can be shared with immediatelyReportUnsafeAccessTo.
    message = target->crossDomainAccessErrorMessage(activeDOMWindow(exec));
    return false;
}

bool shouldAllowAccessToDOMWindow(ExecState* exec, DOMWindow& target, String& message)
{
    if (BindingSecurity::shouldAllowAccessToDOMWindow(exec, target, DoNotReportSecurityError))
        return true;
    // FIXME: The following line of code should move somewhere that it can be shared with immediatelyReportUnsafeAccessTo.
    message = target.crossDomainAccessErrorMessage(activeDOMWindow(exec));
    return false;
}

void printErrorMessageForFrame(Frame* frame, const String& message)
{
    if (!frame)
        return;
    if(!frame->document())
    	return;
    frame->document()->domWindow()->printErrorMessage(message);
}

Structure* getCachedDOMStructure(JSDOMGlobalObject* globalObject, const ClassInfo* classInfo)
{
    JSDOMStructureMap& structures = globalObject->structures();
    return structures.get(classInfo).get();
}

Structure* cacheDOMStructure(JSDOMGlobalObject* globalObject, Structure* structure, const ClassInfo* classInfo)
{
    JSDOMStructureMap& structures = globalObject->structures();
    ASSERT(!structures.contains(classInfo));
    return structures.set(classInfo, WriteBarrier<Structure>(globalObject->globalData(), globalObject, structure)).iterator->second.get();
}

static EncodedJSValue throwTypeErrorInternal(JSC::ExecState* state, const String& errorMessage)
{
    return throwVMError(state, state->vm(), createTypeError(state, errorMessage));
}

static EncodedJSValue throwTypeErrorInternal(JSC::ExecState& state, const String& errorMessage)
{
    return throwTypeErrorInternal(&state, errorMessage);
}

static void appendArgumentMustBe(StringBuilder& builder, unsigned argumentIndex, const char* argumentName, const char* interfaceName, const char* functionName)
{
    builder.appendLiteral("Argument ");
    builder.appendNumber(argumentIndex + 1);
    builder.appendLiteral(" ('");
    builder.append(argumentName);
    builder.appendLiteral("') to ");
    if (!functionName) {
        builder.appendLiteral("the ");
        builder.append(interfaceName);
        builder.appendLiteral(" constructor");
    } else {
        builder.append(interfaceName);
        builder.append('.');
        builder.append(functionName);
    }
    builder.appendLiteral(" must be ");
}

NEVER_INLINE JSC::EncodedJSValue reportDeprecatedGetterError(JSC::ExecState& state, const char* interfaceName, const char* attributeName)
{
	ScriptExecutionContext& context = *jsCast<JSDOMGlobalObject*>(state.lexicalGlobalObject())->scriptExecutionContext();
    context.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("Deprecated attempt to access property '", attributeName, "' on a non-", interfaceName, " object."));
    return JSValue::encode(jsUndefined());
}

NEVER_INLINE void reportDeprecatedSetterError(JSC::ExecState& state, const char* interfaceName, const char* attributeName)
{
	ScriptExecutionContext& context = *jsCast<JSDOMGlobalObject*>(state.lexicalGlobalObject())->scriptExecutionContext();
    context.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, tryMakeString("Deprecated attempt to set property '", attributeName, "' on a non-", interfaceName, " object."));
}

NEVER_INLINE void throwSecurityError(JSC::ExecState& state, const String& message)
{
    ASSERT(!state.hadException());
    state.vm().throwException(&state, createDOMException(&state, SECURITY_ERR, message));
}

NEVER_INLINE JSC::EncodedJSValue throwArgumentMustBeEnumError(JSC::ExecState& state, unsigned argumentIndex, const char* argumentName, const char* functionInterfaceName, const char* functionName, const char* expectedValues)
{
    StringBuilder builder;
    appendArgumentMustBe(builder, argumentIndex, argumentName, functionInterfaceName, functionName);
    builder.appendLiteral("one of: ");
    builder.append(expectedValues);
    return throwTypeErrorInternal(state, builder.toString());
}

NEVER_INLINE JSC::EncodedJSValue throwArgumentMustBeFunctionError(JSC::ExecState& state, unsigned argumentIndex, const char* argumentName, const char* interfaceName, const char* functionName)
{
    StringBuilder builder;
    appendArgumentMustBe(builder, argumentIndex, argumentName, interfaceName, functionName);
    builder.appendLiteral("a function");
    return throwTypeErrorInternal(state, builder.toString());
}

NEVER_INLINE JSC::EncodedJSValue throwArgumentTypeError(JSC::ExecState& state, unsigned argumentIndex, const char* argumentName, const char* functionInterfaceName, const char* functionName, const char* expectedType)
{
    StringBuilder builder;
    appendArgumentMustBe(builder, argumentIndex, argumentName, functionInterfaceName, functionName);
    builder.appendLiteral("an instance of ");
    builder.append(expectedType);
    return throwTypeErrorInternal(state, builder.toString());
}

NEVER_INLINE void throwArrayElementTypeError(JSC::ExecState& state)
{
    throwTypeErrorInternal(state, "Invalid Array element type");
}

NEVER_INLINE void throwAttributeTypeError(JSC::ExecState& state, const char* interfaceName, const char* attributeName, const char* expectedType)
{
    throwTypeErrorInternal(state, tryMakeString("The ", interfaceName, '.', attributeName, " attribute must be an instance of ", expectedType));
}

NEVER_INLINE JSC::EncodedJSValue throwConstructorScriptExecutionContextUnavailableError(JSC::ExecState& state, const char* interfaceName)
{
    return throwVMError(&state, state.vm(), createReferenceError(&state, tryMakeString(interfaceName, " constructor associated execution context is unavailable")));
}

NEVER_INLINE void throwSequenceTypeError(JSC::ExecState& state)
{
    throwTypeErrorInternal(state, ASCIILiteral("Value is not a sequence"));
}

NEVER_INLINE void throwNonFiniteTypeError(ExecState& state)
{
    throwTypeErrorInternal(state, ASCIILiteral("The provided value is non-finite"));
}

NEVER_INLINE String makeGetterTypeErrorMessage(const char* interfaceName, const char* attributeName)
{
    return tryMakeString("The ", interfaceName, '.', attributeName, " getter can only be used on instances of ", interfaceName);
}

#if ENABLE(BINDINGS_ALL_IN_ONE)
static
#endif
NEVER_INLINE JSC::EncodedJSValue throwGetterTypeError(JSC::ExecState& state, const char* interfaceName, const char* attributeName)
{
    return throwVMTypeError(&state, state.vm(), makeGetterTypeErrorMessage(interfaceName, attributeName));
}

#if ENABLE(BINDINGS_ALL_IN_ONE)
static
#endif
NEVER_INLINE bool throwSetterTypeError(JSC::ExecState& state, const char* interfaceName, const char* attributeName)
{
    throwTypeErrorInternal(&state, tryMakeString("The ", interfaceName, '.', attributeName, " setter can only be used on instances of ", interfaceName));
    return false;
}

NEVER_INLINE String makeThisTypeErrorMessage(const char* interfaceName, const char* functionName)
{
    return tryMakeString("Can only call ", interfaceName, '.', functionName, " on instances of ", interfaceName);
}

String makeUnsupportedIndexedSetterErrorMessage(const char* interfaceName)
{
    return tryMakeString("Failed to set an indexed property on ", interfaceName, ": Indexed property setter is not supported.");
}

#if ENABLE(BINDINGS_ALL_IN_ONE)
static
#endif
NEVER_INLINE EncodedJSValue throwThisTypeError(JSC::ExecState& state, const char* interfaceName, const char* functionName)
{
    return throwTypeErrorInternal(&state, makeThisTypeErrorMessage(interfaceName, functionName));
}

static const int32_t kMaxInt32 = 0x7fffffff;
static const int32_t kMinInt32 = -kMaxInt32 - 1;
static const uint32_t kMaxUInt32 = 0xffffffffU;
static const int64_t kJSMaxInteger = 0x20000000000000LL - 1; // 2^53 - 1, largest integer exactly representable in ECMAScript.

static String rangeErrorString(double value, double min, double max)
{
    return tryMakeString("Value ", String::numberToStringECMAScript(value), " is outside the range [", String::numberToStringECMAScript(min), ", ", String::numberToStringECMAScript(max), "]");
}

static double enforceRange(ExecState& state, VM& vm, double x, double minimum, double maximum)
{
    if (std::isnan(x) || std::isinf(x)) {
        vm.throwException(&state, createTypeError(&state, rangeErrorString(x, minimum, maximum)));
        return 0;
    }
    x = trunc(x);
    if (x < minimum || x > maximum) {
        vm.throwException(&state, createTypeError(&state, rangeErrorString(x, minimum, maximum)));
        return 0;
    }
    return x;
}

template <typename T>
struct IntTypeLimits {
};

template <>
struct IntTypeLimits<int8_t> {
    static const int8_t minValue = -128;
    static const int8_t maxValue = 127;
    static const unsigned numberOfValues = 256; // 2^8
};

template <>
struct IntTypeLimits<uint8_t> {
    static const uint8_t maxValue = 255;
    static const unsigned numberOfValues = 256; // 2^8
};

template <>
struct IntTypeLimits<int16_t> {
    static const short minValue = -32768;
    static const short maxValue = 32767;
    static const unsigned numberOfValues = 65536; // 2^16
};

template <>
struct IntTypeLimits<uint16_t> {
    static const unsigned short maxValue = 65535;
    static const unsigned numberOfValues = 65536; // 2^16
};

template <typename T>
static inline T toSmallerInt(ExecState& state, JSValue value, IntegerConversionConfiguration configuration)
{
	VM& vm = state.vm();

    //static_assert(std::is_signed<T>::value && std::is_integral<T>::value, "Should only be used for signed integral types");

    typedef IntTypeLimits<T> LimitsTrait;
    // Fast path if the value is already a 32-bit signed integer in the right range.
    if (value.isInt32()) {
        int32_t d = value.asInt32();
        if (d >= LimitsTrait::minValue && d <= LimitsTrait::maxValue)
            return static_cast<T>(d);
        switch (configuration) {
        case NormalConversion:
            break;
        case EnforceRange:
            throwTypeError(&state, vm);
            return 0;
        case Clamp:
            return d < LimitsTrait::minValue ? LimitsTrait::minValue : LimitsTrait::maxValue;
        }
        d %= LimitsTrait::numberOfValues;
        return static_cast<T>(d > LimitsTrait::maxValue ? d - LimitsTrait::numberOfValues : d);
    }

    double x = value.toNumber(&state);
    if (UNLIKELY(vm.exception()))
        return 0;

    switch (configuration) {
    case NormalConversion:
        break;
    case EnforceRange:
        return enforceRange(state, vm, x, LimitsTrait::minValue, LimitsTrait::maxValue);
    case Clamp:
        return std::isnan(x) ? 0 : clampTo<T>(x);
    }

    if (std::isnan(x) || std::isinf(x) || !x)
        return 0;

    x = x < 0 ? -floor(fabs(x)) : floor(fabs(x));
    x = fmod(x, LimitsTrait::numberOfValues);

    return static_cast<T>(x > LimitsTrait::maxValue ? x - LimitsTrait::numberOfValues : x);
}

template <typename T>
static inline T toSmallerUInt(ExecState& state, JSValue value, IntegerConversionConfiguration configuration)
{
	VM& vm = state.vm();

    //static_assert(std::is_unsigned<T>::value && std::is_integral<T>::value, "Should only be used for unsigned integral types");

    typedef IntTypeLimits<T> LimitsTrait;
    // Fast path if the value is already a 32-bit unsigned integer in the right range.
    if (value.isUInt32()) {
        uint32_t d = value.asUInt32();
        if (d <= LimitsTrait::maxValue)
            return static_cast<T>(d);
        switch (configuration) {
        case NormalConversion:
            return static_cast<T>(d);
        case EnforceRange:
            throwTypeError(&state, vm);
            return 0;
        case Clamp:
            return LimitsTrait::maxValue;
        }
    }

    double x = value.toNumber(&state);
    if (UNLIKELY(vm.exception()))
        return 0;

    switch (configuration) {
    case NormalConversion:
        break;
    case EnforceRange:
        return enforceRange(state, vm, x, 0, LimitsTrait::maxValue);
    case Clamp:
        return std::isnan(x) ? 0 : clampTo<T>(x);
    }

    if (std::isnan(x) || std::isinf(x) || !x)
        return 0;

    x = x < 0 ? -floor(fabs(x)) : floor(fabs(x));
    return static_cast<T>(fmod(x, LimitsTrait::numberOfValues));
}

int8_t toInt8EnforceRange(JSC::ExecState& state, JSValue value)
{
    return toSmallerInt<int8_t>(state, value, EnforceRange);
}

uint8_t toUInt8EnforceRange(JSC::ExecState& state, JSValue value)
{
    return toSmallerUInt<uint8_t>(state, value, EnforceRange);
}

int8_t toInt8Clamp(JSC::ExecState& state, JSValue value)
{
    return toSmallerInt<int8_t>(state, value, Clamp);
}

uint8_t toUInt8Clamp(JSC::ExecState& state, JSValue value)
{
    return toSmallerUInt<uint8_t>(state, value, Clamp);
}

// http://www.w3.org/TR/WebIDL/#es-byte
int8_t toInt8(ExecState& state, JSValue value)
{
    return toSmallerInt<int8_t>(state, value, NormalConversion);
}

// http://www.w3.org/TR/WebIDL/#es-octet
uint8_t toUInt8(ExecState& state, JSValue value)
{
    return toSmallerUInt<uint8_t>(state, value, NormalConversion);
}

int16_t toInt16EnforceRange(ExecState& state, JSValue value)
{
    return toSmallerInt<int16_t>(state, value, EnforceRange);
}

uint16_t toUInt16EnforceRange(ExecState& state, JSValue value)
{
    return toSmallerUInt<uint16_t>(state, value, EnforceRange);
}

int16_t toInt16Clamp(ExecState& state, JSValue value)
{
    return toSmallerInt<int16_t>(state, value, Clamp);
}

uint16_t toUInt16Clamp(ExecState& state, JSValue value)
{
    return toSmallerUInt<uint16_t>(state, value, Clamp);
}

// http://www.w3.org/TR/WebIDL/#es-short
int16_t toInt16(ExecState& state, JSValue value)
{
    return toSmallerInt<int16_t>(state, value, NormalConversion);
}

// http://www.w3.org/TR/WebIDL/#es-unsigned-short
uint16_t toUInt16(ExecState& state, JSValue value)
{
    return toSmallerUInt<uint16_t>(state, value, NormalConversion);
}

// http://www.w3.org/TR/WebIDL/#es-long
int32_t toInt32EnforceRange(ExecState& state, JSValue value)
{
    if (value.isInt32())
        return value.asInt32();

    VM& vm = state.vm();
    double x = value.toNumber(&state);
    if (UNLIKELY(vm.exception()))
        return 0;
    return enforceRange(state, vm, x, kMinInt32, kMaxInt32);
}

int32_t toInt32Clamp(ExecState& state, JSValue value)
{
    if (value.isInt32())
        return value.asInt32();

    double x = value.toNumber(&state);
    return std::isnan(x) ? 0 : clampTo<int32_t>(x);
}

uint32_t toUInt32Clamp(ExecState& state, JSValue value)
{
    if (value.isUInt32())
        return value.asUInt32();

    double x = value.toNumber(&state);
    return std::isnan(x) ? 0 : clampTo<uint32_t>(x);
}

// http://www.w3.org/TR/WebIDL/#es-unsigned-long
uint32_t toUInt32EnforceRange(ExecState& state, JSValue value)
{
    if (value.isUInt32())
        return value.asUInt32();

    VM& vm = state.vm();
    double x = value.toNumber(&state);
    if (UNLIKELY(vm.exception()))
        return 0;
    return enforceRange(state, vm, x, 0, kMaxUInt32);
}

int64_t toInt64EnforceRange(ExecState& state, JSC::JSValue value)
{
    VM& vm = state.vm();
    double x = value.toNumber(&state);
    if (UNLIKELY(vm.exception()))
        return 0;
    return enforceRange(state, vm, x, -kJSMaxInteger, kJSMaxInteger);
}

uint64_t toUInt64EnforceRange(ExecState& state, JSC::JSValue value)
{
    VM& vm = state.vm();
    double x = value.toNumber(&state);
    if (UNLIKELY(vm.exception()))
        return 0;
    return enforceRange(state, vm, x, 0, kJSMaxInteger);
}

int64_t toInt64Clamp(ExecState& state, JSC::JSValue value)
{
    double x = value.toNumber(&state);
    return std::isnan(x) ? 0 : static_cast<int64_t>(std::min<double>(std::max<double>(x, -kJSMaxInteger), kJSMaxInteger));
}

uint64_t toUInt64Clamp(ExecState& state, JSC::JSValue value)
{
    double x = value.toNumber(&state);
    return std::isnan(x) ? 0 : static_cast<uint64_t>(std::min<double>(std::max<double>(x, 0), kJSMaxInteger));
}

// http://www.w3.org/TR/WebIDL/#es-long-long
int64_t toInt64(ExecState& state, JSValue value)
{
    double x = value.toNumber(&state);

    // Map NaNs and +/-Infinity to 0; convert finite values modulo 2^64.
    unsigned long long n;
    doubleToInteger(x, n);
    return n;
}

// http://www.w3.org/TR/WebIDL/#es-unsigned-long-long
uint64_t toUInt64(ExecState& state, JSValue value)
{
    double x = value.toNumber(&state);

    // Map NaNs and +/-Infinity to 0; convert finite values modulo 2^64.
    unsigned long long n;
    doubleToInteger(x, n);
    return n;
}

DOMWindow& incumbentDOMWindow(ExecState& state)
{
    return asJSDOMWindow(&callerGlobalObject(state))->wrapped();
}

Document* responsibleDocument(ExecState& state)
{
    CallerFunctor functor;
    state.iterate(functor);
    auto* callerFrame = functor.callerFrame();
    if (!callerFrame)
        return nullptr;
    return asJSDOMWindow(callerFrame->lexicalGlobalObject())->wrapped().document();
}

static bool canAccessDocument(BindingState* state, Document& targetDocument, SecurityReportingOption reportingOption = ReportSecurityError)
{
    DOMWindow* active = activeDOMWindow(state);
    if(!active)
    	return false;

    if (active->securityOrigin()->canAccess(targetDocument.securityOrigin()))
        return true;

    switch (reportingOption) {
    case ThrowSecurityError:
        throwSecurityError(*state, targetDocument.domWindow()->crossDomainAccessErrorMessage(active));
        break;
    case ReportSecurityError:
        immediatelyReportUnsafeAccessTo(state, targetDocument);
        break;
	case DoNotReportSecurityError:
		break;
	}

    return false;
}

static bool canAccessDocument(BindingState* state, Document* targetDocument, SecurityReportingOption reportingOption = ReportSecurityError)
{
    if (!targetDocument)
        return false;

    return canAccessDocument(state, *targetDocument, reportingOption);
}

bool BindingSecurity::shouldAllowAccessToDOMWindow(BindingState* state, DOMWindow* target, SecurityReportingOption reportingOption)
{
    return target && canAccessDocument(state, target->document(), reportingOption);
}

bool BindingSecurity::shouldAllowAccessToDOMWindow(BindingState* state, DOMWindow& target, SecurityReportingOption reportingOption)
{
    return canAccessDocument(state, target.document(), reportingOption);
}

bool BindingSecurity::shouldAllowAccessToFrame(BindingState* state, Frame* target, SecurityReportingOption reportingOption)
{
    return target && canAccessDocument(state, target->document(), reportingOption);
}

bool BindingSecurity::shouldAllowAccessToNode(BindingState* state, Node* target)
{
    return target && canAccessDocument(state, target->document());
}

bool BindingSecurity::shouldAllowAccessToNode(BindingState* state, Node& target)
{
    return canAccessDocument(state, target.document());
}

bool BindingSecurity::allowSettingFrameSrcToJavascriptUrl(BindingState* state, HTMLFrameElementBase* frame, const String& value)
{
    return !protocolIsJavaScript(stripLeadingAndTrailingHTMLSpaces(value)) || canAccessDocument(state, frame->contentDocument());
}

} // namespace WebCore

#endif //!HAVE(BINDINGS_ALL_IN_ONE)
