/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "JSDOMPromiseDeferred.h"

#include "ExceptionCode.h"
#include "JSDOMError.h"
#include "JSDOMWindow.h"
#include <builtins/BuiltinNames.h>
#include <heap/Strong.h>
#include <runtime/Exception.h>
#include <runtime/JSONObject.h>
#include <runtime/JSPromiseConstructor.h>

using namespace JSC;

namespace WebCore {

JSC::JSValue DeferredPromise::promise() const
{
    if(deferred())
    	return deferred()->promise();
    return jsUndefined();
}

class DeferredPromise::TaskFunctionImpl final : public EventLoop::TaskFunction {
public:
	TaskFunctionImpl(JSGlobalObject& lexicalGlobalObject, DeferredPromise& deferredPromise, JSValue function, JSValue resolution)
		: deferredPromise(deferredPromise)
		, strongFunction(lexicalGlobalObject.vm(), function)
		, strongResolution(lexicalGlobalObject.vm(), resolution)
	{
	}
    virtual void operator()() final
    {
    	if (deferredPromise->shouldIgnoreRequestToFulfill())
    		return;

        JSC::JSGlobalObject* lexicalGlobalObject = deferredPromise->globalObject();
        JSC::JSLockHolder locker(lexicalGlobalObject->vm());
		deferredPromise->callFunction(*lexicalGlobalObject->globalExec(), strongFunction.get(), strongResolution.get());
    }
private:
    Ref<DeferredPromise> deferredPromise;
    JSC::Strong<JSC::Unknown, ShouldStrongDestructorGrabLock::Yes> strongFunction;
    JSC::Strong<JSC::Unknown, ShouldStrongDestructorGrabLock::Yes> strongResolution;
};

void DeferredPromise::callFunction(ExecState& exec, JSValue function, JSValue resolution)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    if (activeDOMObjectsAreSuspended() || !ScriptDisallowedScope::isScriptAllowedInMainThread()) {
        ASSERT(!activeDOMObjectsAreSuspended() || scriptExecutionContext()->eventLoop().isSuspended());
		scriptExecutionContext()->eventLoop().queueTask(TaskSource::Networking, adoptPtr(new TaskFunctionImpl(*exec.lexicalGlobalObject(), *this, function, resolution)));
        return;
    }

    VM& vm = exec.vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);

    auto callData = JSC::getCallData(vm, function);
    ASSERT(callType != CallData::Type::None);

    MarkedArgumentBuffer arguments;
    arguments.append(resolution);

    call(&exec, function, callData, jsUndefined(), arguments);

    // DeferredPromise should only be used by internal implementations that are well behaved.
    // In practice, the only exception we should ever see here is the TerminatedExecutionException.
    //ASSERT_UNUSED(scope, !scope.exception() || isTerminatedExecutionException(vm, scope.exception()));

    if (m_mode == Mode::ClearPromiseOnResolve)
    	clear();
}

#if 0
void DeferredPromise::whenSettled(Function<void()>&& callback)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    if (activeDOMObjectsAreSuspended()) {
        scriptExecutionContext()->eventLoop().queueTask(TaskSource::Networking, *scriptExecutionContext(), [this, protectedThis = makeRef(*this), callback = WTFMove(callback)]() mutable {
            whenSettled(WTFMove(callback));
        });
        return;
    }

    DOMPromise::whenPromiseIsSettled(globalObject(), deferred(), WTFMove(callback));
}
#endif

void DeferredPromise::reject(ExceptionCode ec, const String& message)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    Ref<DeferredPromise> protectedThis(*this);
    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* state = globalObject()->globalExec();
    JSC::JSLockHolder locker(state);
    reject(*state, createDOMException(state, ec, message));
}

void rejectPromiseWithExceptionIfAny(JSC::ExecState& state, JSDOMGlobalObject& globalObject, JSPromiseDeferred& promiseDeferred)
{
    if (!state.hadException())
        return;

    JSValue error = state.exception()->value();
    state.clearException();

    DeferredPromise::create(&globalObject, &promiseDeferred)->reject(error);
}

PassRefPtr<DeferredPromise> createDeferredPromise(JSC::ExecState& state, JSDOMGlobalObject* globalObject)
{
    JSC::JSPromiseDeferred* deferred = JSC::JSPromiseDeferred::tryCreate(&state, globalObject);
    // deferred can only be null in workers.
    if(!deferred)
    	return nullptr;
    return DeferredPromise::create(globalObject, deferred);
}

JSC::EncodedJSValue createRejectedPromiseWithTypeError(JSC::ExecState& state, const String& errorMessage)
{
    ASSERT(state.lexicalGlobalObject());
    JSGlobalObject* globalObject = state.lexicalGlobalObject();
    VM& vm = globalObject->vm();

    JSPromiseConstructor* promiseConstructor = globalObject->promiseConstructor();
    JSValue rejectFunction = promiseConstructor->get(vm, &state, vm.propertyNames->builtinNames().rejectPrivateName());
    JSObject* rejectionValue = createTypeError(&state, errorMessage);

    CallData callData = JSC::getCallData(vm, rejectFunction);
    ASSERT(callType != CallType::None);

    MarkedArgumentBuffer arguments;
    arguments.append(rejectionValue);

    return JSValue::encode(call(&state, rejectFunction, callData, promiseConstructor, arguments));
}

static inline JSC::JSValue parseAsJSON(JSC::ExecState* state, const String& data)
{
    JSC::JSLockHolder lock(state);
    return JSC::JSONParse(state, data);
}

void fulfillPromiseWithJSON(RefPtr<DeferredPromise> promise, const String& data)
{
    JSC::JSValue value = parseAsJSON(promise->globalObject()->globalExec(), data);
    if (!value)
        promise->reject<ExceptionCode>(SYNTAX_ERR);
    else
        promise->resolve(value);
}

void fulfillPromiseWithArrayBuffer(RefPtr<DeferredPromise> promise, ArrayBuffer* arrayBuffer)
{
    if (!arrayBuffer) {
        promise->reject<JSValue>(createOutOfMemoryError(promise->globalObject()->globalExec()));
        return;
    }
    promise->resolve(arrayBuffer);
}

void fulfillPromiseWithArrayBuffer(RefPtr<DeferredPromise> promise, const void* data, size_t length)
{
    fulfillPromiseWithArrayBuffer(promise, ArrayBuffer::tryCreate(data, length).get());
}

}
