/*
 * Copyright (C) 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "EventLoop.h"
#include "JSDOMBinding.h"
#include "JSDOMGuardedObject.h"
#include "ScriptExecutionContext.h"
#include <heap/StrongInlines.h>
#include <runtime/JSPromiseDeferred.h>
#include <runtime/Symbol.h>

namespace WebCore {

class JSDOMWindow;

class DeferredPromise : public DOMGuarded<JSC::JSPromiseDeferred> {
public:
    enum class Mode {
        ClearPromiseOnResolve,
        RetainPromiseOnResolve
    };

    static PassRefPtr<DeferredPromise> tryCreate(JSC::ExecState& state, JSDOMGlobalObject& globalObject, Mode mode = Mode::ClearPromiseOnResolve)
    {
        JSC::JSPromiseDeferred* promiseDeferred = JSC::JSPromiseDeferred::tryCreate(&state, &globalObject);
        if (!promiseDeferred)
            return nullptr;
        return adoptRef(new DeferredPromise(&globalObject, promiseDeferred, mode));
    }

    static Ref<DeferredPromise> create(JSDOMGlobalObject* globalObject, JSC::JSPromiseDeferred* deferred, Mode mode = Mode::ClearPromiseOnResolve)
    {
        return adoptRef(*new DeferredPromise(globalObject, deferred, mode));
    }

    template<class ResolveResultType>
    void resolve(ResolveResultType*);

    template<class ResolveResultType>
    void resolve(const ResolveResultType&);

    template<class RejectResultType>
    void reject(const RejectResultType&);

    void reject(ExceptionCode, const String& message = String());

    void resolve()
    {
        if (shouldIgnoreRequestToFulfill())
            return;

        ASSERT(deferred());
        ASSERT(globalObject());
        JSC::ExecState* exec = globalObject()->globalExec();
        JSC::JSLockHolder locker(exec);
        resolve(*exec, JSC::jsUndefined());
    }

    template<typename Callback, typename Value>
    void resolveWithCallback(Callback callback, Value value)
    {
        if (shouldIgnoreRequestToFulfill())
            return;

        ASSERT(deferred());
        ASSERT(globalObject());
        JSC::ExecState* exec = globalObject()->globalExec();
        JSC::JSLockHolder locker(exec);
        resolve(*exec, callback(*exec, *globalObject(), std::forward<Value>(value)));
    }

    JSC::JSValue promise() const;

private:
    DeferredPromise(JSDOMGlobalObject* globalObject, JSC::JSPromiseDeferred* deferred, Mode mode)
		: DOMGuarded<JSC::JSPromiseDeferred>(*globalObject, *deferred)
		, m_mode(mode)
	{
	}

    bool shouldIgnoreRequestToFulfill() const { return isEmpty(); }

    JSC::JSPromiseDeferred* deferred() const { return guarded(); }

    void callFunction(JSC::ExecState&, JSC::JSValue function, JSC::JSValue resolution);
    void resolve(JSC::ExecState& state, JSC::JSValue resolution) { callFunction(state, deferred()->resolve(), resolution); }
    void reject(JSC::ExecState& state, JSC::JSValue resolution) { callFunction(state, deferred()->reject(), resolution); }

    class TaskFunctionImpl;
    friend class TaskFunctionImpl;

    Mode m_mode;
};

PassRefPtr<DeferredPromise> createDeferredPromise(JSC::ExecState&, JSDOMGlobalObject*);

void fulfillPromiseWithJSON(RefPtr<DeferredPromise>, const String&);
void fulfillPromiseWithArrayBuffer(RefPtr<DeferredPromise>, JSC::ArrayBuffer*);
void fulfillPromiseWithArrayBuffer(RefPtr<DeferredPromise>, const void*, size_t);
void rejectPromiseWithExceptionIfAny(JSC::ExecState&, JSDOMGlobalObject&, JSC::JSPromiseDeferred&);
JSC::EncodedJSValue createRejectedPromiseWithTypeError(JSC::ExecState&, const String&);

inline JSC::JSValue callPromiseFunction(JSC::ExecState& state, JSC::EncodedJSValue promiseFunction(JSC::ExecState*, RefPtr<DeferredPromise>))
{
    JSDOMGlobalObject& globalObject = *JSC::jsCast<JSDOMGlobalObject*>(state.lexicalGlobalObject());
    JSC::JSPromiseDeferred* promiseDeferred = JSC::JSPromiseDeferred::tryCreate(&state, &globalObject);

    // promiseDeferred can be null when terminating a Worker abruptly.
    if (!promiseDeferred)
        return JSC::jsUndefined();

    promiseFunction(&state, DeferredPromise::create(&globalObject, promiseDeferred));

    rejectPromiseWithExceptionIfAny(state, globalObject, *promiseDeferred);
    ASSERT(!state.hadException());
    return promiseDeferred->promise();
}

template <typename Value, typename Error>
class DOMPromiseDeferred {
public:
	DOMPromiseDeferred() { }
    DOMPromiseDeferred(RefPtr<DeferredPromise> wrapper) : m_promiseDeferred(wrapper) { }
    DOMPromiseDeferred(const DOMPromiseDeferred& promise) : m_promiseDeferred(promise.m_promiseDeferred) { }
    DOMPromiseDeferred(DOMPromiseDeferred&& promise) : m_promiseDeferred(WTFMove(promise.m_promiseDeferred)) { }

    void resolve(Value* value) { m_promiseDeferred->resolve<Value>(value); }
    void resolve(const Value& value) { m_promiseDeferred->resolve<Value>(value); }
    void reject(const Error& error) { m_promiseDeferred->reject<Error>(error); }

    RefPtr<DeferredPromise> promiseDeferred() const { return m_promiseDeferred; }
    JSC::JSValue promise() const { return m_promiseDeferred->promise(); };

private:
    RefPtr<DeferredPromise> m_promiseDeferred;
};

#if 0
template<class ResolveResultType>
inline void DeferredPromise::resolveWithValue(ResolveResultType&& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, toJS(exec, globalObject(), std::forward<ResolveResultType>(result)));
}

template<class ResolveResultType>
inline void DeferredPromise::resolveWithNewlyCreated(Ref<ResolveResultType>&& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, toJSNewlyCreated(exec, globalObject(), WTFMove(result)));
}

template<class RejectResultType>
inline void DeferredPromise::rejectWithValue(RejectResultType&& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, toJS(exec, globalObject(), std::forward<RejectResultType>(result)));
}
#endif

template<class ResolveResultType>
inline void DeferredPromise::resolve(ResolveResultType* result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, toJS(exec, globalObject(), result));
}

template<class ResolveResultType>
inline void DeferredPromise::resolve(const ResolveResultType& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, toJS(exec, globalObject(), &result));
}

template<class RejectResultType>
inline void DeferredPromise::reject(const RejectResultType& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, toJS(exec, globalObject(), result));
}

template<>
inline void DeferredPromise::reject(const std::nullptr_t&)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, JSC::jsNull());
}

template<>
inline void DeferredPromise::reject(const JSC::JSValue& value)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, value);
}

template<>
inline void DeferredPromise::reject<ExceptionCode>(const ExceptionCode& ec)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, createDOMException(exec, ec));
}

template<>
inline void DeferredPromise::reject<Exception>(const Exception& exception)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, createDOMException(exec, exception.code(), exception.message()));
}

template<>
inline void DeferredPromise::resolve<String>(const String& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, jsString(globalObject()->vm(), result));
}

template<>
inline void DeferredPromise::resolve<bool>(const bool& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, JSC::jsBoolean(result));
}

template<>
inline void DeferredPromise::resolve<JSC::JSValue>(const JSC::JSValue& value)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, value);
}

template<>
inline void DeferredPromise::resolve<Vector<unsigned char> >(const Vector<unsigned char>& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    RefPtr<ArrayBuffer> buffer = ArrayBuffer::tryCreate(result.data(), result.size());
    resolve(*exec, toJS(exec, globalObject(), buffer.get()));
}

template<>
inline void DeferredPromise::resolve(const std::nullptr_t&)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, JSC::jsUndefined());
}

template<>
inline void DeferredPromise::reject<String>(const String& result)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, jsString(globalObject()->vm(), result));
}

template<>
inline void DeferredPromise::resolve(const JSC::PrivateName& privateName)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    resolve(*exec, JSC::JSValue(JSC::Symbol::create(globalObject()->vm(), *privateName.uid())));
}

template<>
inline void DeferredPromise::reject(const JSC::PrivateName& privateName)
{
    if (shouldIgnoreRequestToFulfill())
        return;

    ASSERT(deferred());
    ASSERT(globalObject());
    JSC::ExecState* exec = globalObject()->globalExec();
    JSC::JSLockHolder locker(exec);
    reject(*exec, JSC::JSValue(JSC::Symbol::create(globalObject()->vm(), *privateName.uid())));
}

}
