/*
 * Copyright (C) 2007, 2008, 2009, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2011 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#if !HAVE(BINDINGS_ALL_IN_ONE)
#include "JSDOMWindowCustom.h"

#include "Frame.h"
#include "FrameLoaderClient.h"
#include "HTMLCollection.h"
#include "HTMLDocument.h"
#include "HTMLFrameOwnerElement.h"
#include "History.h"
#include "JSArrayBuffer.h"
#include "JSDataView.h"
#include "JSEvent.h"
#include "JSEventListener.h"
#include "JSEventSource.h"
#include "JSFloat32Array.h"
#include "JSFloat64Array.h"
#include "JSHTMLAudioElement.h"
#include "JSHTMLCollection.h"
#include "JSHTMLOptionElement.h"
#include "JSHistory.h"
#include "JSInt16Array.h"
#include "JSInt32Array.h"
#include "JSInt8Array.h"
#include "JSLocation.h"
#include "JSMessageChannel.h"
#include "JSMessagePortCustom.h"
#include "JSUint16Array.h"
#include "JSUint32Array.h"
#include "JSUint8Array.h"
#include "JSUint8ClampedArray.h"
#include "JSWebKitCSSMatrix.h"
#include "JSWebKitPoint.h"
#include "JSXMLHttpRequest.h"
#include "JSXSLTProcessor.h"
#include "Location.h"
#include "MediaPlayer.h"
#include "Page.h"
#include "RuntimeEnabledFeatures.h"
#include "ScheduledAction.h"
#include "Settings.h"
#include "SharedWorkerRepository.h"
#include <runtime/GetterSetter.h>
#include <runtime/JSCustomGetterFunction.h>
#include <runtime/JSCustomSetterFunction.h>
#include <runtime/JSFunction.h>
#include <runtime/JSMicrotask.h>
#include <runtime/Microtask.h>
#include <wtf/PlatformRunState.h>

#if ENABLE(LEGACY_ENCRYPTED_MEDIA_V2)
#include "JSWebKitMediaKeys.h"
#endif

#if ENABLE(MEDIA_SOURCE)
#include "JSMediaSource.h"
#include "JSSourceBuffer.h"
#include "JSSourceBufferList.h"
#endif

#if ENABLE(WORKERS)
#include "JSWorker.h"
#endif

#if ENABLE(SHARED_WORKERS)
#include "JSSharedWorker.h"
#endif

#if ENABLE(WEB_AUDIO)
#include "JSAudioContext.h"
#endif

#if ENABLE(WEB_SOCKETS)
#include "JSWebSocket.h"
#endif

#if ENABLE(RMAI)
#include "JSMOM.h"
#include "MOM.h"
#include "JSMOMGuideSearch.h"
#include "JSMOMChannelList.h"
#include "JSMOMChannel.h"
#include "JSMOMProgramme.h"
#endif

#if PLATFORM(COCOA)
#include "VersionChecks.h"
#endif

using namespace JSC;

namespace WebCore {

EncodedJSValue JSC_HOST_CALL jsDOMWindowInstanceFunctionShowModalDialog(ExecState*);

void JSDOMWindow::visitAdditionalChildren(SlotVisitor& visitor)
{
    if (Frame* frame = wrapped().frame())
        visitor.addOpaqueRoot(frame);

    visitor.addOpaqueRoot(&wrapped());

    // Normally JSEventTargetCustom.cpp's JSEventTarget::visitAdditionalChildren() would call this. But
    // even though DOMWindow is an EventTarget, JSDOMWindow does not subclass JSEventTarget, so we need
    // to do this here.
    wrapped().visitJSEventListeners(visitor);
}

#if ENABLE(USER_MESSAGE_HANDLERS)
static EncodedJSValue jsDOMWindowWebKit(ExecState* exec, EncodedJSValue thisValue, PropertyName)
{
    JSDOMWindow* castedThis = toJSDOMWindow(JSValue::decode(thisValue));
    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, castedThis->wrapped()))
        return JSValue::encode(jsUndefined());
    return JSValue::encode(toJS(exec, castedThis->globalObject(), castedThis->wrapped().webkitNamespace()));
}
#endif

static ALWAYS_INLINE bool allowsLegacyExpandoIndexedProperties()
{
#if PLATFORM(COCOA)
    // Given that WindowProxy is the default |this| value for sloppy mode functions, it's hard to prove
    // that older iOS and macOS apps don't accidentally depend on this behavior, so we keep it for them.
    static bool requiresQuirk = !linkedOnOrAfter(SDKVersion::FirstWithoutExpandoIndexedPropertiesOnWindow);
    return requiresQuirk;
#else
    return false;
#endif
}

static bool jsDOMWindowGetOwnPropertySlotRestrictedAccess(JSDOMWindow* thisObject, Frame* frame, ExecState* exec, PropertyName propertyName, PropertySlot& slot, String& errorMessage)
{
	auto* lexicalGlobalObject = exec->lexicalGlobalObject();
    VM& vm = lexicalGlobalObject->vm();

    // We don't want any properties other than "close" and "closed" on a frameless window
    // (i.e. one whose page got closed, or whose iframe got removed).
    // FIXME: This handling for frameless windows duplicates similar behaviour for cross-origin
    // access below; we should try to find a way to merge the two.
    if (!frame) {
        if (propertyName == vm.propertyNames->closed) {
            slot.setCustom(thisObject, ReadOnly | DontDelete | DontEnum, jsDOMWindowClosed);
            return true;
        }
        if (propertyName == vm.propertyNames->close) {
            slot.setCustom(thisObject, ReadOnly | DontDelete | DontEnum, nonCachingStaticFunctionGetter<jsDOMWindowInstanceFunctionClose, 0>);
            return true;
        }

        // FIXME: We should have a message here that explains why the property access/function call was
        // not allowed.
        slot.setUndefined();
        return true;
    }

    // https://html.spec.whatwg.org/#crossorigingetownpropertyhelper-(-o,-p-)
    if (propertyName == vm.propertyNames->toStringTagSymbol || propertyName == vm.propertyNames->hasInstanceSymbol || propertyName == vm.propertyNames->isConcatSpreadableSymbol) {
    	slot.setValue(thisObject, ReadOnly | DontEnum, jsUndefined());
		return true;
    }

    // These are the functions we allow access to cross-origin (DoNotCheckSecurity in IDL).
    // Always provide the original function, on a fresh uncached function object.
    if (propertyName == vm.propertyNames->blur) {
        slot.setCustom(thisObject, ReadOnly | DontEnum, nonCachingStaticFunctionGetter<jsDOMWindowInstanceFunctionBlur, 0>);
        return true;
    }
    if (propertyName == vm.propertyNames->close) {
        slot.setCustom(thisObject, ReadOnly | DontEnum, nonCachingStaticFunctionGetter<jsDOMWindowInstanceFunctionClose, 0>);
        return true;
    }
    if (propertyName == vm.propertyNames->focus) {
        slot.setCustom(thisObject, ReadOnly | DontEnum, nonCachingStaticFunctionGetter<jsDOMWindowInstanceFunctionFocus, 0>);
        return true;
    }
    if (propertyName == vm.propertyNames->postMessage) {
        slot.setCustom(thisObject, ReadOnly | DontEnum, nonCachingStaticFunctionGetter<jsDOMWindowInstanceFunctionPostMessage, 2>);
        return true;
    }

    // When accessing cross-origin known Window properties, we always use the original property getter,
    // even if the property was removed / redefined. As of early 2016, this matches Firefox and Chrome's
    // behavior.
    if (auto* entry = JSDOMWindow::info()->staticPropHashTable->entry(propertyName)) {
        // Only allow access to these specific properties.
        if (propertyName == vm.propertyNames->location
            || propertyName == vm.propertyNames->closed
            || propertyName == vm.propertyNames->length
            || propertyName == vm.propertyNames->self
            || propertyName == vm.propertyNames->window
            || propertyName == vm.propertyNames->frames
            || propertyName == vm.propertyNames->opener
            || propertyName == vm.propertyNames->parent
            || propertyName == vm.propertyNames->top) {
            auto* getter = JSCustomGetterFunction::create(vm, lexicalGlobalObject, propertyName, entry->propertyGetter());
            auto* setter = propertyName == vm.propertyNames->location ? JSCustomSetterFunction::create(vm, lexicalGlobalObject, propertyName, entry->propertyPutter()) : nullptr;
            auto* getterSetter = GetterSetter::create(vm, lexicalGlobalObject, getter, setter);
            slot.setGetterSlot(thisObject, PropertyAttribute::Accessor | PropertyAttribute::DontEnum, getterSetter);
            return true;
        }

        // For any other entries in the static property table, deny access. (Early return also prevents
        // named getter from returning frames with matching names - this seems a little questionable, see
        // FIXME comment on prototype search below.)
        throwSecurityError(*exec, errorMessage);
        slot.setUndefined();
        return false;
    }

    // Check for child frames by name before built-in properties to match Mozilla. This does
    // not match IE, but some sites end up naming frames things that conflict with window
    // properties that are in Moz but not IE. Since we have some of these, we have to do it
    // the Moz way.
    if (auto* scopedChild = frame->tree()->scopedChild(propertyNameToAtomicString(propertyName))) {
        slot.setValue(thisObject, ReadOnly | DontDelete | DontEnum, toJS(exec, scopedChild->document()->domWindow()));
        return true;
    }

    if (handleCommonCrossOriginProperties(thisObject, vm, propertyName, slot))
        return true;

    throwSecurityError(*exec, errorMessage);
    slot.setUndefined();
    return false;
}

// https://html.spec.whatwg.org/#crossorigingetownpropertyhelper-(-o,-p-)
bool handleCommonCrossOriginProperties(JSObject* thisObject, VM& vm, PropertyName propertyName, PropertySlot& slot)
{
    auto& propertyNames = vm.propertyNames;
    if (propertyName == propertyNames->builtinNames().thenPublicName() || propertyName == propertyNames->toStringTagSymbol || propertyName == propertyNames->hasInstanceSymbol || propertyName == propertyNames->isConcatSpreadableSymbol) {
        slot.setValue(thisObject, JSC::ReadOnly | JSC::DontEnum, jsUndefined());
        return true;
    }
    return false;
}

// Property access sequence is:
// (1) indexed properties,
// (2) regular own properties,
// (3) named properties (in fact, these shouldn't be on the window, should be on the NPO).
bool JSDOMWindow::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
    // (1) First, indexed properties.
    // Hand off all indexed access to getOwnPropertySlotByIndex, which supports the indexed getter.
    if (Optional<unsigned> index = parseIndex(propertyName))
        return getOwnPropertySlotByIndex(object, exec, index.value(), slot);

    auto* thisObject = jsCast<JSDOMWindow*>(object);
    auto* frame = thisObject->wrapped().frame();

    // Hand off all cross-domain/frameless access to jsDOMWindowGetOwnPropertySlotRestrictedAccess.
    String errorMessage;
    if (!frame || !shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), errorMessage))
        return jsDOMWindowGetOwnPropertySlotRestrictedAccess(thisObject, frame, exec, propertyName, slot, errorMessage);

    if (UNLIKELY(!thisObject->m_windowCloseWatchpoints))
        thisObject->m_windowCloseWatchpoints = WatchpointSet::create(thisObject->wrapped().frame() ? IsWatched : IsInvalidated);
    // We use m_windowCloseWatchpoints to clear any inline caches once the frame is cleared.
    // This is sound because DOMWindow can be associated with at most one frame in its lifetime.
    if (thisObject->m_windowCloseWatchpoints->isStillValidOnJSThread())
        slot.setWatchpointSet(*thisObject->m_windowCloseWatchpoints);

    VM& vm = thisObject->vm();

    // (2) Regular own properties.
    PropertySlot slotCopy = slot;
    if (Base::getOwnPropertySlot(thisObject, exec, propertyName, slot)) {
        // Detect when we're getting the property 'showModalDialog', this is disabled, and has its original value.
        bool isShowModalDialogAndShouldHide = propertyName == vm.propertyNames->showModalDialog
            && !DOMWindow::canShowModalDialog(frame)
            && slot.isValue() && isHostFunction(slot.getValue(exec, propertyName), jsDOMWindowInstanceFunctionShowModalDialog);
        // Unless we're in the showModalDialog special case, we're done.
        if (!isShowModalDialogAndShouldHide)
            return true;
        slot = slotCopy;
    }

#if ENABLE(USER_MESSAGE_HANDLERS)
    if (propertyName == vm.propertyNames->webkit && thisObject->wrapped().shouldHaveWebKitNamespaceForWorld(thisObject->world())) {
        slot.setCacheableCustom(thisObject, DontDelete | ReadOnly, jsDOMWindowWebKit);
        return true;
    }
#endif

    return false;
}

bool JSDOMWindow::getOwnPropertySlotByIndex(JSObject* object, ExecState* exec, unsigned index, PropertySlot& slot)
{
    auto* thisObject = jsCast<JSDOMWindow*>(object);
    auto& window = thisObject->wrapped();
    auto* frame = window.frame();

    // Indexed getters take precendence over regular properties, so caching would be invalid.
    slot.disableCaching();

    // These are also allowed cross-orgin, so come before the access check.
    if (frame && index < frame->tree()->scopedChildCount()) {
        slot.setValue(thisObject, ReadOnly, toJS(exec, frame->tree()->scopedChild(index)->document()->domWindow()));
        return true;
    }

    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, window, ThrowSecurityError))
        return false;
    if (allowsLegacyExpandoIndexedProperties())
    	return Base::getOwnPropertySlotByIndex(thisObject, exec, index, slot);
    return false;
}

bool JSDOMWindow::put(JSCell* cell, ExecState* exec, PropertyName propertyName, JSValue value, PutPropertySlot& slot)
{
    JSDOMWindow* thisObject = jsCast<JSDOMWindow*>(cell);

	VM& vm = thisObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

#if ENABLE(BML)
    if(thisObject->wrapped().frame()->document()->isBMLDocument())
    	return JSGlobalObject::put(thisObject, exec, propertyName, value, slot);
#endif

    if (propertyName != vm.propertyNames->location) {
    	if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), ThrowSecurityError))
			return false;
	}

    if (parseIndex(propertyName) && !allowsLegacyExpandoIndexedProperties())
		return typeError(exec, vm, slot.isStrictMode(), makeUnsupportedIndexedSetterErrorMessage("Window"));
	return Base::put(thisObject, exec, propertyName, value, slot);
}

bool JSDOMWindow::putByIndex(JSCell* cell, ExecState* exec, unsigned index, JSValue value, bool shouldThrow)
{
    JSDOMWindow* thisObject = jsCast<JSDOMWindow*>(cell);
	VM& vm = thisObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), ThrowSecurityError))
        return false;

    if (allowsLegacyExpandoIndexedProperties()) {
        if (auto* document = thisObject->wrapped().document())
            document->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Warning, ASCIILiteral("Adding expando indexed properties to 'window' was a non-standard behavior that is now removed."));
        RELEASE_AND_RETURN(scope, Base::putByIndex(thisObject, exec, index, value, shouldThrow));
    }
    return typeError(exec, vm, shouldThrow, makeUnsupportedIndexedSetterErrorMessage("Window"));
}

bool JSDOMWindow::deleteProperty(JSCell* cell, ExecState* exec, PropertyName propertyName)
{
    VM& vm = exec->vm();
    auto* thisObject = jsCast<JSDOMWindow*>(cell);
    auto scope = DECLARE_THROW_SCOPE(vm);

    String errorMessage;
    if (!shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), errorMessage)) {
        throwSecurityError(*exec, errorMessage);
        return false;
    }
    if (auto index = parseIndex(propertyName)) {
        if (!allowsLegacyExpandoIndexedProperties())
            return index.value() >= thisObject->wrapped().length();
    }
    return Base::deleteProperty(thisObject, exec, propertyName);
}

bool JSDOMWindow::deletePropertyByIndex(JSCell* cell, ExecState* exec, unsigned propertyName)
{
    JSDOMWindow* thisObject = jsCast<JSDOMWindow*>(cell);
    // Only allow deleting properties by frames in the same origin.
    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), ThrowSecurityError))
		return false;
    if (isIndex(propertyName) && !allowsLegacyExpandoIndexedProperties())
        return propertyName >= thisObject->wrapped().length();
    return Base::deletePropertyByIndex(thisObject, exec, propertyName);
}

// https://html.spec.whatwg.org/#crossoriginproperties-(-o-)

template <CrossOriginObject objectType>
static void addCrossOriginPropertyNames(VM& vm, PropertyNameArray& propertyNames)
{
    switch (objectType) {
    case CrossOriginObject::Location: {
        static const unsigned total = 2;
        const Identifier* properties[total] = { &vm.propertyNames->href, &vm.propertyNames->replace };
        for(int i = 0; i < total; ++i)
            propertyNames.add(*properties[i]);
        break;
    }
    case CrossOriginObject::Window: {
		static const unsigned total = 13;
		const Identifier* properties[total] = {
			&vm.propertyNames->blur, &vm.propertyNames->close, &vm.propertyNames->closed,
			&vm.propertyNames->focus, &vm.propertyNames->frames, &vm.propertyNames->length,
			&vm.propertyNames->location, &vm.propertyNames->opener, &vm.propertyNames->parent,
			&vm.propertyNames->postMessage, &vm.propertyNames->self, &vm.propertyNames->top,
			&vm.propertyNames->window
		};
		for(int i = 0; i < total; ++i)
			propertyNames.add(*properties[i]);
		break;
	}
	}
}

// https://html.spec.whatwg.org/#crossoriginownpropertykeys-(-o-)
template <CrossOriginObject objectType>
void addCrossOriginOwnPropertyNames(VM& vm, JSC::PropertyNameArray& propertyNames)
{
    addCrossOriginPropertyNames<objectType>(vm, propertyNames);

	static const unsigned total = 4;
    static const Identifier* const properties[total] = {
        &vm.propertyNames->builtinNames().thenPublicName(), &vm.propertyNames->toStringTagSymbol, &vm.propertyNames->hasInstanceSymbol, &vm.propertyNames->isConcatSpreadableSymbol
    };

	for(int i = 0; i < total; ++i)
		propertyNames.add(*properties[i]);

}
template void addCrossOriginOwnPropertyNames<CrossOriginObject::Window>(VM&, JSC::PropertyNameArray&);
template void addCrossOriginOwnPropertyNames<CrossOriginObject::Location>(VM&, JSC::PropertyNameArray&);

static void addScopedChildrenIndexes(VM& vm, DOMWindow& window, PropertyNameArray& propertyNames)
{
    auto* document = window.document();
    if (!document)
        return;

    auto* frame = document->frame();
    if (!frame)
        return;

    unsigned scopedChildCount = frame->tree()->scopedChildCount();
    for (unsigned i = 0; i < scopedChildCount; ++i)
        propertyNames.add(Identifier::from(vm, i));
}

// https://html.spec.whatwg.org/#windowproxy-ownpropertykeys
void JSDOMWindow::getOwnPropertyNames(JSObject* object, ExecState* exec, PropertyNameArray& propertyNames, DontEnumPropertiesMode mode)
{
    JSDOMWindow* thisObject = jsCast<JSDOMWindow*>(object);

    VM& vm = thisObject->vm();

	addScopedChildrenIndexes(vm, thisObject->wrapped(), propertyNames);

    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), DoNotReportSecurityError)) {
    	if (mode == DontEnumPropertiesMode::Include)
			addCrossOriginOwnPropertyNames<CrossOriginObject::Window>(vm, propertyNames);
        return;
    }
    Base::getOwnPropertyNames(thisObject, exec, propertyNames, mode);
}

bool JSDOMWindow::defineOwnProperty(JSC::JSObject* object, JSC::ExecState* exec, JSC::PropertyName propertyName, const JSC::PropertyDescriptor& descriptor, bool shouldThrow)
{
    JSDOMWindow* thisObject = jsCast<JSDOMWindow*>(object);
    VM& vm = thisObject->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    // Only allow defining properties in this way by frames in the same origin, as it allows setters to be introduced.
    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), ThrowSecurityError))
		return false;
    if (parseIndex(propertyName) && !allowsLegacyExpandoIndexedProperties())
        return typeError(exec, vm, shouldThrow, makeUnsupportedIndexedSetterErrorMessage("Window"));
    scope.release();

    auto* builtinNames = vm.propertyNames;
    if (propertyName == builtinNames->document || propertyName == builtinNames->window)
        return JSObject::defineOwnProperty(thisObject, exec, propertyName, descriptor, shouldThrow);

    return Base::defineOwnProperty(thisObject, exec, propertyName, descriptor, shouldThrow);
}

JSValue JSDOMWindow::getPrototype(JSObject* object, ExecState* exec)
{
    JSDOMWindow* thisObject = jsCast<JSDOMWindow*>(object);
    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), DoNotReportSecurityError))
        return jsNull();

    return Base::getPrototype(object, exec);
}

bool JSDOMWindow::preventExtensions(JSObject*, ExecState*)
{
    return false;
}

String JSDOMWindow::toStringName(const JSObject* object, ExecState* exec)
{
    auto* thisObject = jsCast<const JSDOMWindow*>(object);
    if (!BindingSecurity::shouldAllowAccessToDOMWindow(exec, thisObject->wrapped(), DoNotReportSecurityError))
        return ASCIILiteral("Object");
    return ASCIILiteral("Window");
}

// Custom Attributes

JSValue JSDOMWindow::event(ExecState* exec) const
{
    Event* event = currentEvent();
    if (!event)
        return jsUndefined();
    return toJS(exec, const_cast<JSDOMWindow*>(this), event);
}

#if ENABLE(WEB_AUDIO)
JSValue JSDOMWindow::webkitAudioContext(ExecState* exec) const
{
	if(Frame* frame = wrapped().frame())
		if (!frame->settings().webAudioEnabled())
		    return jsUndefined();

	return getDOMConstructor<JSAudioContextConstructor>(vm(), this);
}
#endif

#if ENABLE(RMAI)
JSValue JSDOMWindow::mom(ExecState* exec) const
{
    DOMWindow& imp = static_cast<DOMWindow&>(wrapped());
    if(MOM* mom = imp.mom())
        return toJS(exec, const_cast<JSDOMWindow*>(this), mom);
    return jsUndefined();
}
#endif

// Custom functions

class DialogHandler {
public:
    explicit DialogHandler(ExecState* exec)
        : m_exec(exec)
    {
    }

    void dialogCreated(DOMWindow*);
    JSValue returnValue() const;

private:
    ExecState* m_exec;
    RefPtr<Frame> m_frame;
};

inline void DialogHandler::dialogCreated(DOMWindow* dialog)
{
    m_frame = dialog->frame();
    // FIXME: This looks like a leak between the normal world and an isolated
    //        world if dialogArguments comes from an isolated world.
    VM& vm = m_exec->vm();
    JSDOMWindow* globalObject = toJSDOMWindow(m_frame.get(), normalWorld(vm));
    if (JSValue dialogArguments = m_exec->argument(1))
        globalObject->putDirect(vm, Identifier::fromString(vm, "dialogArguments"), dialogArguments);
}

inline JSValue DialogHandler::returnValue() const
{
    VM& vm = m_exec->vm();
    JSDOMWindow* globalObject = toJSDOMWindow(m_frame.get(), normalWorld(vm));
    if (!globalObject)
        return jsUndefined();
    Identifier identifier = Identifier::fromString(vm, "returnValue");
    PropertySlot slot(globalObject, PropertySlot::InternalMethodType::Get);
    if (!JSGlobalObject::getOwnPropertySlot(globalObject, m_exec, identifier, slot))
        return jsUndefined();
    return slot.getValue(m_exec, identifier);
}

static void setUpDialog(DOMWindow* dialog, void* handler)
{
    static_cast<DialogHandler*>(handler)->dialogCreated(dialog);
}

JSValue JSDOMWindow::showModalDialog(ExecState* exec)
{
    String urlString = valueToStringWithUndefinedOrNullCheck(exec, exec->argument(0));
    if (exec->hadException())
        return jsUndefined();
    String dialogFeaturesString = valueToStringWithUndefinedOrNullCheck(exec, exec->argument(2));
    if (exec->hadException())
        return jsUndefined();

    DialogHandler handler(exec);

    wrapped().showModalDialog(urlString, dialogFeaturesString, activeDOMWindow(exec), firstDOMWindow(exec), setUpDialog, &handler);

    return handler.returnValue();
}

JSValue JSDOMWindow::queueMicrotask(ExecState* state)
{
    VM& vm = state->vm();
    ThrowScope scope(vm);

    if (UNLIKELY(state->argumentCount() < 1))
        return throwException(state, scope, createNotEnoughArgumentsError(state));

    JSValue functionValue = state->uncheckedArgument(0);
    if (UNLIKELY(!functionValue.isCallable(vm)))
        return JSValue::decode(throwArgumentMustBeFunctionError(*state, 0, "callback", "Window", "queueMicrotask"));

    scope.release();
    Base::queueMicrotask(JSC::createJSMicrotask(vm, functionValue));
    return jsUndefined();
}

JSValue JSDOMWindow::postMessage(ExecState* exec)
{
    JSC::VM& vm = exec->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    // void postMessage(any message, USVString targetOrigin, optional sequence<object> transfer = [])
    // void postMessage(any message, optional WindowPostMessageOptions options);
    if (UNLIKELY(exec->argumentCount() < 1))
        return throwException(exec, scope, createNotEnoughArgumentsError(exec));

    Vector<RefPtr<MessagePort>> messagePorts;
    ArrayBufferArray arrayBuffers;

    JSValue jsTargetOrigin;
    JSValue jsTransfer;

    if (exec->argumentCount() > 1) {
    	JSValue jsArg1 = exec->uncheckedArgument(1);
    	if (jsArg1.isObject()) {
    		jsTargetOrigin = asObject(jsArg1)->get(vm, exec, Identifier::fromString(vm, "targetOrigin"));
    		RETURN_IF_EXCEPTION(scope, jsUndefined());

    		jsTransfer = asObject(jsArg1)->get(vm, exec, Identifier::fromString(vm, "transfer"));
    		RETURN_IF_EXCEPTION(scope, jsUndefined());
    	} else {
    		jsTargetOrigin = jsArg1;

    	    if (exec->argumentCount() > 2)
    	    	jsTransfer = exec->uncheckedArgument(2);
    	}
    }

	if (jsTransfer) {
		extractTransferables(exec, jsTransfer, messagePorts, arrayBuffers);
		RETURN_IF_EXCEPTION(scope, jsUndefined());
	}

	RefPtr<SerializedScriptValue> message = SerializedScriptValue::create(exec, exec->uncheckedArgument(0), messagePorts, arrayBuffers);
    RETURN_IF_EXCEPTION(scope, jsUndefined());

    String targetOrigin;
    if (jsTargetOrigin) {
    	targetOrigin = valueToUSVStringWithUndefinedOrNullCheck(exec, jsTargetOrigin);
    	RETURN_IF_EXCEPTION(scope, jsUndefined());
    }

    ExceptionCode ec = 0;
    wrapped().postMessage(exec, &incumbentDOMWindow(exec), message.release(), &messagePorts, targetOrigin, ec);
    if (UNLIKELY(ec))
    	setDOMException(exec, ec);

    return jsUndefined();
}

JSValue JSDOMWindow::setTimeout(ExecState* exec)
{
    ContentSecurityPolicy* contentSecurityPolicy = wrapped().document() ? wrapped().document()->contentSecurityPolicy() : nullptr;
    RefPtr<ScheduledAction> action = ScheduledAction::create(exec, globalObject()->world(), contentSecurityPolicy);
    if (exec->hadException())
        return jsUndefined();

    if (!action)
        return jsNumber(0);

    int delay = exec->argument(1).toInt32(exec);

    ExceptionCode ec = 0;
    int result = wrapped().setTimeout(action.release(), delay, ec);
    if (UNLIKELY(ec))
    	setDOMException(exec, ec);

    return jsNumber(result);
}

JSValue JSDOMWindow::setInterval(ExecState* exec)
{
    ContentSecurityPolicy* contentSecurityPolicy = wrapped().document() ? wrapped().document()->contentSecurityPolicy() : 0;
    RefPtr<ScheduledAction> action = ScheduledAction::create(exec, globalObject()->world(), contentSecurityPolicy);
    if (exec->hadException())
        return jsUndefined();
    int delay = exec->argument(1).toInt32(exec);

    if (!action)
        return jsNumber(0);

    ExceptionCode ec = 0;
    int result = wrapped().setInterval(action.release(), delay, ec);
    if (UNLIKELY(ec))
    	setDOMException(exec, ec);

    return jsNumber(result);
}

DOMWindow* toDOMWindow(JSValue value)
{
    if (!value.isObject())
        return 0;
    JSObject* object = asObject(value);
    if (object->inherits(JSDOMWindow::info()))
        return &jsCast<JSDOMWindow*>(object)->wrapped();
    if (object->inherits(JSDOMWindowShell::info()))
        return &jsCast<JSDOMWindowShell*>(object)->wrapped();
    return 0;
}

DOMWindow* JSDOMWindow::toWrapped(JSC::JSValue value)
{
	return toDOMWindow(value);
}

void JSDOMWindow::setOpener(JSC::ExecState* state, JSC::JSValue value)
{
    if (!BindingSecurity::shouldAllowAccessToDOMWindow(state, wrapped(), ReportSecurityError))
        return;

    if (value.isNull()) {
        wrapped().disownOpener();
        return;
    }

    bool shouldThrow = true;
    createDataProperty(state, Identifier::fromString(state->vm(), "opener"), value, shouldThrow);
}

#if 0
void JSDOMWindow::setOpenDatabase(JSC::ExecState* state, JSC::JSValue value)
{
    if (!BindingSecurity::shouldAllowAccessToDOMWindow(state, wrapped(), ThrowSecurityError))
        return;

    bool shouldThrow = true;
    createDataProperty(state, Identifier::fromString(state->vm(), "openDatabase"), value, shouldThrow);
}
#endif

#if ENABLE(CEHTML)
JSValue JSDOMWindow::escapeBeyondTopMost(ExecState* exec)
{
    String urlString = valueToStringWithUndefinedOrNullCheck(exec, exec->argument(0));
    if (exec->hadException())
        return jsUndefined();

    bool result = wrapped().escapeBeyondTopMost(urlString, activeDOMWindow(exec), firstDOMWindow(exec));
    return jsBoolean(result);
}

JSValue JSDOMWindow::reload(ExecState* exec)
{
    wrapped().reload(activeDOMWindow(exec));
    return jsUndefined();
}

JSValue JSDOMWindow::replace(ExecState* exec)
{
    String urlString = valueToStringWithUndefinedOrNullCheck(exec, exec->argument(0));
    if (exec->hadException())
        return jsUndefined();

    wrapped().replace(urlString, activeDOMWindow(exec), firstDOMWindow(exec));
    return jsUndefined();
}

JSValue JSDOMWindow::setTimer(ExecState* exec)
{
    ContentSecurityPolicy* contentSecurityPolicy = wrapped().document() ? wrapped().document()->contentSecurityPolicy() : 0;
    RefPtr<ScheduledAction> action = ScheduledAction::create(exec, globalObject()->world(), contentSecurityPolicy);
    if (exec->hadException())
        return jsUndefined();

    if (!action)
        return jsNumber(0);

    int delay = exec->argument(1).toInt32(exec);

    ExceptionCode ec = 0;
    int result = wrapped().setTimer(action.release(), delay, ec);
    if (UNLIKELY(ec))
    	setDOMException(exec, ec);

    return jsNumber(result);
}

#endif

} // namespace WebCore
#endif
