/*
 * Copyright (C) 2011 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "AbortSignal.h"
#include "AnimationReplaceState.h"
#include "CanvasRenderingContext2D.h"
#include "CompositeOperation.h"
#include "CompositeOperationOrAuto.h"
#include "DOMWindow.h"
#include "FetchOptions.h"
#include "FetchResponse.h"
#include "FillMode.h"
#include "FontFace.h"
#include "FontFaceSet.h"
#include "IDBAny.h"
#include "ImageBitmapOptions.h"
#include "IterationCompositeOperation.h"
#include "MessagePort.h"
#include "MediaDeviceInfo.h"
#include "MediaKeyMessageType.h"
#include "MediaKeysRequirement.h"
#include "MediaKeySessionType.h"
#include "MediaKeyStatus.h"
#include "MediaSourceSettings.h"
#include "MediaStreamTrack.h"
#include "PlaybackDirection.h"
#include "PredefinedColorSpace.h"
#include "RTCBundlePolicy.h"
#include "RTCConfiguration.h"
#include "RTCDataChannelState.h"
#include "RTCIceConnectionState.h"
#include "RTCIceGatheringState.h"
#include "RTCIceTransportPolicy.h"
#include "RTCPeerConnectionState.h"
#include "RTCRtpParameters.h"
#include "RTCSessionDescription.h"
#include "RTCStatsType.h"
#include "RTCRtpTransceiver.h"
#include "ShadowRoot.h"
#include "SourceBufferAppendMode.h"
#include "XMLHttpRequest.h"
#include "WebAnimation.h"
#include <heap/Strong.h>
#include <heap/StrongInlines.h>
#include <interpreter/CallFrame.h>
#include <runtime/JSGlobalObject.h>
#include <runtime/Operations.h>
#include <runtime/Uint8Array.h>
#include <wtf/Forward.h>
#include <wtf/Optional.h>

namespace WebCore {

class ArrayValue;
class DataTransfer;
class Dictionary;
class EventTarget;
class FetchHeaders;
class MediaStream;
class MediaStreamTrack;
class RTCRtpReceiver;
class RTCRtpTransceiver;
class Node;
class ScriptValue;
class SerializedScriptValue;
class SpeechRecognitionResultList;
class Storage;
class TrackBase;
class WebKitMediaKeyError;

class JSDictionary {
public:
    JSDictionary(JSC::ExecState* exec, JSC::JSObject* initializerObject)
    : m_exec(exec)
	{
		if (exec && initializerObject)
			m_initializerObject = JSC::Strong<JSC::JSObject>(exec->globalData(), initializerObject);
	}

    // Returns false if any exceptions were thrown, regardless of whether the property was found.
    template <typename Result>
    bool tryGetProperty(const ASCIILiteral& propertyName, Result&) const;
    template <typename T, typename Result>
    bool tryGetProperty(const ASCIILiteral& propertyName, T* context, void (*setter)(T* context, const Result&)) const;

    // Returns true if the property was found in the dictionary, and the value could be converted to the desired type.
    template <typename Result>
    bool get(const ASCIILiteral& propertyName, Result&) const;
    template <typename Result>
    bool get(const JSC::Identifier& propertyName, Result&) const;
    bool getWithUndefinedOrNullCheck(const ASCIILiteral& propertyName, String& value) const;

    JSC::ExecState* execState() const { return m_exec; }
    JSC::JSObject* initializerObject() const { return m_initializerObject.get(); }
    bool isValid() const { return m_exec && m_initializerObject; }

private:
    template <typename Result>
    struct IdentitySetter {
        static void identitySetter(Result* context, const Result& result)
        {
            *context = result;
        }
    };

    enum class GetPropertyResult {
        ExceptionThrown,
        NoPropertyFound,
        PropertyFound
    };
    
    template <typename T, typename Result>
    GetPropertyResult tryGetPropertyAndResult(JSC::VM&, const JSC::Identifier& propertyName, T* context, void (*setter)(T* context, const Result&)) const;
    GetPropertyResult tryGetProperty(JSC::VM&, const JSC::Identifier& propertyName, JSC::JSValue&) const;

    template <typename T>
    static void convertValue(JSC::ExecState* execState, JSC::JSValue value, Optional<T>& result)
    {
        if (value.isUndefinedOrNull()) {
            result = Nullopt;
            return;
        }

        T actualResult;
        convertValue(execState, value, actualResult);
        result = actualResult;
    }

    static void convertValue(JSC::ExecState*, JSC::JSValue, bool& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, int& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, unsigned& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, unsigned short& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, unsigned long& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, unsigned long long& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, double& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, Optional<double>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, CanvasWindingRule& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, Dictionary& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, String& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, ScriptValue& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, Vector<String>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<SerializedScriptValue>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<DOMWindow>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<EventTarget>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<Node>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<Storage>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, Vector<RefPtr<MessagePort>>& result);
#if ENABLE(VIDEO_TRACK)
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<TrackBase>& result);
#endif
#if ENABLE(MUTATION_OBSERVERS) || ENABLE(WEB_INTENTS)
    static void convertValue(JSC::ExecState*, JSC::JSValue, HashSet<AtomicString>& result);
#endif
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<JSC::Uint8Array>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<JSC::ArrayBuffer>& result);
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<WebKitMediaKeyError>& result);
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    static void convertValue(JSC::ExecState*, JSC::JSValue, MediaKeyMessageType& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, MediaKeysRequirement& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, MediaKeySessionType& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, MediaKeyStatus& result);
#endif
#if ENABLE(MEDIA_STREAM)
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<MediaStream>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<MediaStreamTrack>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, Vector<RefPtr<MediaStream>>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, MediaStreamTrackState&);
    static void convertValue(JSC::ExecState*, JSC::JSValue, MediaDeviceInfo::Kind& result);
#endif
#if ENABLE(WEB_RTC)
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<RTCRtpReceiver>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<RTCRtpTransceiver>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCIceConnectionState& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCIceGatheringState& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCSignalingState& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCIceTransportPolicy& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCBundlePolicy& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCPeerConnectionState& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCSdpType& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCStatsType& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCDegradationPreference& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCDtxStatus& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCPriorityType& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCDataChannelState& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RTCRtpTransceiverDirection& result);
#endif // ENABLE(WEB_RTC)
#if ENABLE(FETCH_API)
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<FetchHeaders>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchOptions::Type& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchOptions::Destination& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchOptions::Mode& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchOptions::Credentials& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchOptions::Cache& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchOptions::Redirect& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchReferrerPolicy& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FetchResponseType& result);
#endif
#if ENABLE(MEDIA_SOURCE)
    static void convertValue(JSC::ExecState*, JSC::JSValue, SourceBufferAppendMode& result);
#endif
#if ENABLE(SCRIPTED_SPEECH)
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<SpeechRecognitionResultList>& result);
#endif
    static void convertValue(JSC::ExecState*, JSC::JSValue, ArrayValue& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<DataTransfer>& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, JSC::JSFunction*&);
#if ENABLE(INDEXED_DATABASE)
    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<IDBAny>& result);
#endif
    static void convertValue(JSC::ExecState*, JSC::JSValue, FontFaceLoadStatus& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FontFaceSetLoadStatus& result);

    static void convertValue(JSC::ExecState*, JSC::JSValue, ShadowRootMode& result);

    static void convertValue(JSC::ExecState*, JSC::JSValue, RefPtr<AbortSignal>& result);

    static void convertValue(JSC::ExecState*, JSC::JSValue, AnimationPlayState& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, AnimationReplaceState& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, CompositeOperation& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, CompositeOperationOrAuto& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, FillMode& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, PlaybackDirection& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue, IterationCompositeOperation& result);

    static void convertValue(JSC::ExecState*, JSC::JSValue, XMLHttpRequest::ResponseType& result);

    static void convertValue(JSC::ExecState*, JSC::JSValue value, ImageBitmapOptions::Orientation& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue value, ImageBitmapOptions::PremultiplyAlpha& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue value, ImageBitmapOptions::ColorSpaceConversion& result);
    static void convertValue(JSC::ExecState*, JSC::JSValue value, ImageBitmapOptions::ResizeQuality& result);

    static void convertValue(JSC::ExecState*, JSC::JSValue value, PredefinedColorSpace& result);

    static void convertValue(JSC::ExecState*, JSC::JSValue value, JSC::JSValue& result)
    {
    	result = value;
    }

    JSC::ExecState* m_exec;
    JSC::Strong<JSC::JSObject> m_initializerObject;
};

template <typename T, typename Result>
bool JSDictionary::tryGetProperty(const ASCIILiteral& propertyName, T* context, void (*setter)(T* context, const Result&)) const
{
    auto& vm = m_exec->vm();
    JSC::Identifier identifier = JSC::Identifier::fromString(vm, propertyName);
    return tryGetPropertyAndResult(vm, identifier, context, setter) != GetPropertyResult::ExceptionThrown;
}

template <typename Result>
bool JSDictionary::tryGetProperty(const ASCIILiteral& propertyName, Result& finalResult) const
{
    auto& vm = m_exec->vm();
    JSC::Identifier identifier = JSC::Identifier::fromString(vm, propertyName);
    return tryGetPropertyAndResult(vm, identifier, &finalResult, IdentitySetter<Result>::identitySetter) != GetPropertyResult::ExceptionThrown;
}

template <typename Result>
bool JSDictionary::get(const ASCIILiteral& propertyName, Result& finalResult) const
{
    auto& vm = m_exec->vm();
    JSC::Identifier identifier = JSC::Identifier::fromString(vm, propertyName);
    return tryGetPropertyAndResult(vm, identifier, &finalResult, IdentitySetter<Result>::identitySetter) == GetPropertyResult::PropertyFound;
}

template <typename Result>
bool JSDictionary::get(const JSC::Identifier& identifier, Result& finalResult) const
{
    auto& vm = m_exec->vm();
    return tryGetPropertyAndResult(vm, identifier, &finalResult, IdentitySetter<Result>::identitySetter) == GetPropertyResult::PropertyFound;
}

template <>
inline bool JSDictionary::get(const ASCIILiteral& propertyName, JSC::JSValue& finalResult) const
{
    auto& vm = m_exec->vm();
    JSC::Identifier identifier = JSC::Identifier::fromString(vm, propertyName);
    return tryGetProperty(vm, identifier, finalResult) == GetPropertyResult::PropertyFound;
}

template <typename T, typename Result>
JSDictionary::GetPropertyResult JSDictionary::tryGetPropertyAndResult(JSC::VM& vm, const JSC::Identifier& identifier, T* context, void (*setter)(T* context, const Result&)) const
{
    JSC::JSValue value;
    GetPropertyResult getPropertyResult = tryGetProperty(vm, identifier, value);
    switch (getPropertyResult) {
    case GetPropertyResult::ExceptionThrown:
        return getPropertyResult;
    case GetPropertyResult::PropertyFound: {
        Result result;
        convertValue(m_exec, value, result);

        if (UNLIKELY(vm.exception()))
            return GetPropertyResult::ExceptionThrown;

        setter(context, result);
        break;
    }
    case GetPropertyResult::NoPropertyFound:
        break;
    }

    return getPropertyResult;
}

} // namespace WebCore
