/*
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003-2017 Apple Inc. All Rights Reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "JSEventListener.h"

#include "BeforeUnloadEvent.h"
#include "CommonVM.h"
#include "ContentSecurityPolicy.h"
#include "Event.h"
#include "Frame.h"
#include "HTMLElement.h"
#include "InspectorCounters.h"
#include "JSDOMConvert.h"
#include "JSDocument.h"
#include "JSEvent.h"
#include "JSEventTarget.h"
#include "JSExecState.h"
#if ENABLE(NETSCAPE_PLUGIN_API)
#include "c_instance.h"
#include "CRuntimeObject.h"
#endif
#include "WorkerGlobalScope.h"
#include <runtime/ExceptionHelpers.h>
#include <runtime/JSLock.h>
#include <runtime/VMEntryScope.h>
#include <wtf/RefCountedLeakCounter.h>

using namespace JSC;

namespace WebCore {

JSEventListener::JSEventListener(JSObject* function, JSObject* wrapper, bool isAttribute, DOMWrapperWorld* isolatedWorld)
    : EventListener(JSEventListenerType, isAttribute)
    , m_wrapper(wrapper)
    , m_isolatedWorld(*isolatedWorld)
{
    auto* vm = m_isolatedWorld->vm();
    if (function) {
        ASSERT(wrapper);
        m_jsFunction = JSC::Weak<JSC::JSObject>(function);
        vm->heap.writeBarrier(wrapper, function);
        m_isInitialized = true;
    }

    if (vm != commonVMOrNull())
    	static_cast<JSVMClientData*>(vm->clientData)->addClient(*this);

#if ENABLE(INSPECTOR)
    ThreadLocalInspectorCounters::current().incrementCounter(ThreadLocalInspectorCounters::JSEventListenerCounter);
#endif
}

PassRefPtr<JSEventListener> JSEventListener::create(JSC::JSObject* listener, JSC::JSObject* wrapper, bool isAttribute, DOMWrapperWorld* world)
{
	return adoptRef(new JSEventListener(listener, wrapper, isAttribute, world));
}

PassRefPtr<JSEventListener> JSEventListener::create(JSC::JSValue listener, JSC::JSObject& wrapper, bool isAttribute, DOMWrapperWorld* world)
{
    if (UNLIKELY(!listener.isObject()))
        return nullptr;

    return adoptRef(new JSEventListener(JSC::asObject(listener), &wrapper, isAttribute, world));
}

JSEventListener::~JSEventListener()
{
#if ENABLE(INSPECTOR)
    ThreadLocalInspectorCounters::current().decrementCounter(ThreadLocalInspectorCounters::JSEventListenerCounter);
#endif
}

JSObject* JSEventListener::initializeJSFunction(ScriptExecutionContext*) const
{
    return nullptr;
}

void JSEventListener::visitJSFunction(SlotVisitor& visitor)
{
    // If m_wrapper is null, we are not keeping m_jsFunction alive.
    if (!m_wrapper)
        return;

    visitor.appendUnbarrieredWeak(&m_jsFunction);
}

static void handleBeforeUnloadEventReturnValue(BeforeUnloadEvent& event, const String& returnValue)
{
    if (returnValue.isNull())
        return;

    event.preventDefault();
    if (event.returnValue().isEmpty())
        event.setReturnValue(returnValue);
}

void JSEventListener::handleEvent(ScriptExecutionContext* scriptExecutionContext, Event* event)
{
    ASSERT(scriptExecutionContext);
    if (!scriptExecutionContext || scriptExecutionContext->isJSExecutionForbidden())
        return;

    JSLockHolder lock(scriptExecutionContext->globalData());

    JSObject* jsFunction = ensureJSFunction(scriptExecutionContext);
    if (!jsFunction)
        return;

    if (UNLIKELY(!m_isolatedWorld))
        return;

    auto* globalObject = toJSDOMGlobalObject(scriptExecutionContext, m_isolatedWorld.get());
    if (!globalObject)
        return;

    if (scriptExecutionContext->isDocument()) {
        JSDOMWindow* window = jsCast<JSDOMWindow*>(globalObject);
        if (!window->wrapped().isCurrentlyDisplayedInFrame())
			return;
        if (wasCreatedFromMarkup() && !scriptExecutionContext->contentSecurityPolicy()->allowInlineEventHandlers(sourceURL(), sourcePosition().m_line))
            return;
		// FIXME: Is this check needed for other contexts?
		ScriptController* script = window->wrapped().frame()->script();
        if (!script->canExecuteScripts(AboutToExecuteScript) || script->isPaused())
            return;
    }

    ExecState* exec = globalObject->globalExec();
    JSValue handleEventFunction = jsFunction;

    VM& vm = globalObject->vm();

    auto callData = JSC::getCallData(vm, handleEventFunction);
    // If jsFunction is not actually a function, see if it implements the EventListener interface and use that
    if (callData.type == CallData::Type::None) {
        handleEventFunction = jsFunction->get(vm, exec, Identifier::fromString(vm, "handleEvent"));
        callData = JSC::getCallData(vm, handleEventFunction);
    }

    if (callData.type == CallData::Type::None)
        return;

	RefPtr<JSEventListener> protect(this);

	MarkedArgumentBuffer args;
	args.append(toJS(exec, globalObject, event));

	RefPtr<Event> savedEvent;
	auto* jsFunctionWindow = jsDynamicCast<JSDOMWindow*>(jsFunction->globalObject());
	if (jsFunctionWindow) {
		savedEvent = jsFunctionWindow->currentEvent();

		// window.event should not be set when the target is inside a shadow tree, as per the DOM specification.
		if (!event->currentTargetIsInShadowTree())
			jsFunctionWindow->setCurrentEvent(event);
	}

	VMEntryScope entryScope(vm, vm.entryScope ? vm.entryScope->globalObject() : globalObject);

	InspectorInstrumentationCookie cookie = JSExecState::instrumentFunctionCall(scriptExecutionContext, callData);

	JSValue thisValue = handleEventFunction == jsFunction ? toJS(exec, globalObject, event->currentTarget()) : jsFunction;
	NakedPtr<JSC::Exception> exception;
	JSValue retval = JSExecState::call(exec, handleEventFunction, callData, thisValue, args, exception);

	InspectorInstrumentation::didCallFunction(cookie);

	if (jsFunctionWindow)
		jsFunctionWindow->setCurrentEvent(savedEvent.get());

	auto handleExceptionIfNeeded = [&] () -> bool {
#if ENABLE(WORKERS)
		if (scriptExecutionContext->isWorkerGlobalScope()) {
			WorkerScriptController* scriptController = static_cast<WorkerGlobalScope&>(*scriptExecutionContext).script();
			bool terminatorCausedException = (vm.exception() && isTerminatedExecutionException(exec->exception()));
			if (terminatorCausedException || scriptController->isTerminatingExecution())
				scriptController->forbidExecution();
		}
#endif

		if (exception) {
			event->target()->uncaughtExceptionInEventHandler();
			reportException(exec, exception);
			return true;
		}
		return false;
	};

	if (handleExceptionIfNeeded())
		return;

    if (!m_isAttribute) {
        // This is an EventListener and there is therefore no need for any return value handling.
        return;
    }

    // Do return value handling for event handlers (https://html.spec.whatwg.org/#the-event-handler-processing-algorithm).

    if (event->type() == eventNames().beforeunloadEvent) {
        // This is a OnBeforeUnloadEventHandler, and therefore the return value must be coerced into a String.
        if (event->isBeforeUnloadEvent()) {
        	String resultStr = retval.isUndefinedOrNull() ? String() : retval.toWTFString(exec);
        	if (UNLIKELY(vm.exception())) {
				exception = vm.exception();
				if (handleExceptionIfNeeded())
					return;
			}
			handleBeforeUnloadEventReturnValue(*toBeforeUnloadEvent(event), resultStr);
		}
		return;
	}

#if ENABLE(RMAI)
	if (retval.isBoolean() && !retval.asBoolean()) {
		event->preventDefault();
		event->stopPropagation();
	}
#else  // ! ENABLE(RMAI)
	if (retval.isFalse())
		event->preventDefault();
#endif // ! ENABLE(RMAI)
}

bool JSEventListener::operator==(const EventListener& listener) const
{
	if (const JSEventListener* jsEventListener = JSEventListener::cast(&listener)){
		if(m_isAttribute != jsEventListener->m_isAttribute)
			return false;

		if(m_jsFunction == jsEventListener->m_jsFunction)
			return true;

		if(!m_jsFunction || !jsEventListener->m_jsFunction)
			return false;

#if ENABLE(NETSCAPE_PLUGIN_API)
		// do not compare wrapper
		if(Bindings::CRuntimeObject* thisRuntimeObject = JSC::jsDynamicCast<Bindings::CRuntimeObject*>(m_jsFunction.get()))
			return thisRuntimeObject->equals(JSC::jsDynamicCast<Bindings::CRuntimeObject*>(jsEventListener->m_jsFunction.get()));
#endif
	}
    return false;
}

void JSEventListener::willDestroyVM()
{
    m_jsFunction.clear();
    m_wrapper.clear();
    m_isInitialized = false;
    m_isolatedWorld = nullptr;
}

static inline JSC::JSValue eventHandlerAttribute(EventListener* abstractListener, ScriptExecutionContext& context)
{
    if (!abstractListener)
        return jsNull();

    auto* listener = JSEventListener::cast(abstractListener);
    if (!listener)
        return jsNull();

    auto* function = listener->ensureJSFunction(&context);
    if (!function)
        return jsNull();

    return function;
}

static inline RefPtr<JSEventListener> createEventListenerForEventHandlerAttribute(JSC::ExecState* state, JSC::JSValue listener, JSC::JSObject& wrapper)
{
    if (!listener.isObject())
        return nullptr;
    return JSEventListener::create(asObject(listener), &wrapper, true, currentWorld(state));
}

JSC::JSValue eventHandlerAttribute(EventTarget& target, const AtomicString& eventType, DOMWrapperWorld* isolatedWorld)
{
    auto* context = target.scriptExecutionContext();
    if (!context)
        return jsNull();
    return eventHandlerAttribute(target.attributeEventListener(eventType, isolatedWorld), *context);
}

void setEventHandlerAttribute(JSC::ExecState* state, JSC::JSObject& wrapper, EventTarget& target, const AtomicString& eventType, JSC::JSValue value)
{
    target.setAttributeEventListener(eventType, createEventListenerForEventHandlerAttribute(state, value, wrapper), currentWorld(state));
}

JSC::JSValue windowEventHandlerAttribute(HTMLElement& element, const AtomicString& eventType, DOMWrapperWorld* isolatedWorld)
{
    auto& document = element.document();
    return eventHandlerAttribute(document.getWindowAttributeEventListener(eventType, isolatedWorld), document);
}

void setWindowEventHandlerAttribute(JSC::ExecState* state, JSC::JSObject& wrapper, HTMLElement& element, const AtomicString& eventType, JSC::JSValue value)
{
    ASSERT(wrapper.globalObject());
    element.document().setWindowAttributeEventListener(eventType, createEventListenerForEventHandlerAttribute(state, value, *wrapper.globalObject()), currentWorld(state));
}

JSC::JSValue windowEventHandlerAttribute(DOMWindow& window, const AtomicString& eventType, DOMWrapperWorld* isolatedWorld)
{
    return eventHandlerAttribute(window, eventType, isolatedWorld);
}

void setWindowEventHandlerAttribute(JSC::ExecState* state, JSC::JSObject& wrapper, DOMWindow& window, const AtomicString& eventType, JSC::JSValue value)
{
    setEventHandlerAttribute(state, wrapper, window, eventType, value);
}

JSC::JSValue documentEventHandlerAttribute(HTMLElement& element, const AtomicString& eventType, DOMWrapperWorld* isolatedWorld)
{
    auto& document = element.document();
    return eventHandlerAttribute(document.attributeEventListener(eventType, isolatedWorld), document);
}

void setDocumentEventHandlerAttribute(JSC::ExecState* state, JSC::JSObject& wrapper, HTMLElement& element, const AtomicString& eventType, JSC::JSValue value)
{
    ASSERT(wrapper.globalObject());
    auto& document = element.document();
    auto* documentWrapper = jsDocumentCast(toJS(state, JSC::jsCast<JSDOMGlobalObject*>(wrapper.globalObject()), &document));
    ASSERT(documentWrapper);
    document.setAttributeEventListener(eventType, createEventListenerForEventHandlerAttribute(state, value, *documentWrapper), currentWorld(state));
}

JSC::JSValue documentEventHandlerAttribute(Document& document, const AtomicString& eventType, DOMWrapperWorld* isolatedWorld)
{
    return eventHandlerAttribute(document, eventType, isolatedWorld);
}

void setDocumentEventHandlerAttribute(JSC::ExecState* state, JSC::JSObject& wrapper, Document& document, const AtomicString& eventType, JSC::JSValue value)
{
    setEventHandlerAttribute(state, wrapper, document, eventType, value);
}

} // namespace WebCore
