/*
 *  Copyright (C) 2001 Peter Kelly (pmk@post.com)
 *  Copyright (C) 2003, 2008, 2009 Apple Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#pragma once

#include "EventListener.h"
#include "JSDOMWindow.h"
#include "JSEventTargetCustom.h"
#include "WebCoreJSClientData.h"
#include <heap/StrongInlines.h>
#include <heap/Weak.h>
#include <wtf/text/TextPosition.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class DOMWindow;
class EventTarget;
class HTMLElement;
class JSDOMGlobalObject;

class JSEventListener : public EventListener, public JSVMClientDataClient {
public:
	static PassRefPtr<JSEventListener> create(JSC::JSObject* listener, JSC::JSObject* wrapper, bool isAttribute, DOMWrapperWorld* world);
    static PassRefPtr<JSEventListener> create(JSC::JSValue listener, JSC::JSObject& wrapper, bool isAttribute, DOMWrapperWorld*);

	static const JSEventListener* cast(const EventListener* listener)
	{
		return listener->type() == JSEventListenerType
			? static_cast<const JSEventListener*>(listener)
			: nullptr;
	}

	virtual ~JSEventListener();

	virtual bool operator==(const EventListener& other) const final;

	JSC::JSObject* ensureJSFunction(ScriptExecutionContext*) const;
	JSC::JSObject* ensureJSFunction(ScriptExecutionContext&) const;

	ALWAYS_INLINE DOMWrapperWorld* isolatedWorld() const { return m_isolatedWorld.get(); }

    virtual JSC::JSObject* jsFunction() const FINAL { return m_jsFunction.get(); }
    virtual JSC::JSObject* wrapper() const FINAL { return m_wrapper.get(); }

    virtual String sourceURL() const { return String(); }
    virtual TextPosition sourcePosition() const { return TextPosition::minimumPosition(); }

private:
	virtual JSC::JSObject* initializeJSFunction(ScriptExecutionContext*) const;
	virtual void visitJSFunction(JSC::SlotVisitor&) final;

    // JSVMClientDataClient
    virtual void willDestroyVM() final;

protected:
	JSEventListener(JSC::JSObject* function, JSC::JSObject* wrapper, bool isAttribute, DOMWrapperWorld*);
	virtual void handleEvent(ScriptExecutionContext*, Event*) override;
	void setWrapperWhenInitializingJSFunction(JSC::VM&, JSC::JSObject* wrapper) const { m_wrapper = JSC::Weak<JSC::JSObject>(wrapper); }

private:
	mutable JSC::Weak<JSC::JSObject> m_jsFunction;
	mutable JSC::Weak<JSC::JSObject> m_wrapper;

	RefPtr<DOMWrapperWorld> m_isolatedWorld;
};

// For "onxxx" attributes that automatically set up JavaScript event listeners.
JSC::JSValue eventHandlerAttribute(EventTarget&, const AtomicString& eventType, DOMWrapperWorld*);
void setEventHandlerAttribute(JSC::ExecState*, JSC::JSObject&, EventTarget&, const AtomicString& eventType, JSC::JSValue);

// Like the functions above, but for attributes that forward event handlers to the window object rather than setting them on the target.
JSC::JSValue windowEventHandlerAttribute(HTMLElement&, const AtomicString& eventType, DOMWrapperWorld*);
void setWindowEventHandlerAttribute(JSC::ExecState*, JSC::JSObject&, HTMLElement&, const AtomicString& eventType, JSC::JSValue);
JSC::JSValue windowEventHandlerAttribute(DOMWindow&, const AtomicString& eventType, DOMWrapperWorld*);
void setWindowEventHandlerAttribute(JSC::ExecState*, JSC::JSObject&, DOMWindow&, const AtomicString& eventType, JSC::JSValue);

// Like the functions above, but for attributes that forward event handlers to the document rather than setting them on the target.
JSC::JSValue documentEventHandlerAttribute(HTMLElement&, const AtomicString& eventType, DOMWrapperWorld*);
void setDocumentEventHandlerAttribute(JSC::ExecState*, JSC::JSObject&, HTMLElement&, const AtomicString& eventType, JSC::JSValue);
JSC::JSValue documentEventHandlerAttribute(Document&, const AtomicString& eventType, DOMWrapperWorld*);
void setDocumentEventHandlerAttribute(JSC::ExecState*, JSC::JSObject&, Document&, const AtomicString& eventType, JSC::JSValue);

bool forwardsEventListeners(JSC::JSObject* wrapper);

inline JSC::JSObject* JSEventListener::ensureJSFunction(ScriptExecutionContext* scriptExecutionContext) const
{
	// initializeJSFunction can trigger code that deletes this event listener
	// before we're done. It should always return 0 in this case.
    auto* vm = m_isolatedWorld->vm();
	RefPtr<JSEventListener> protect(const_cast<JSEventListener*>(this));
	JSC::Strong<JSC::JSObject> wrapper(*vm, m_wrapper.get());

    if (!m_isInitialized) {
        ASSERT(!m_jsFunction);
        auto* function = initializeJSFunction(scriptExecutionContext);
        if (function) {
            m_jsFunction = JSC::Weak<JSC::JSObject>(function);
            // When JSFunction is initialized, initializeJSFunction must ensure that m_wrapper should be initialized too.
            ASSERT(m_wrapper);
            vm->heap.writeBarrier(m_wrapper.get(), function);
            m_isInitialized = true;
        }
    }

    // m_wrapper and m_jsFunction are Weak<>. nullptr of these fields do not mean that this event-listener is not initialized yet.
    // If this is initialized once, m_isInitialized should be true, and then m_wrapper and m_jsFunction must be alive. m_wrapper's
    // liveness should be kept correctly by using ActiveDOMObject, output-constraints, etc. And m_jsFunction must be alive if m_wrapper
    // is alive since JSEventListener marks m_jsFunction in JSEventListener::visitJSFunction if m_wrapper is alive.
    // If the event-listener is not initialized yet, we should skip invoking this event-listener.
    if (!m_isInitialized)
        return nullptr;

    ASSERT(m_wrapper);
    ASSERT(m_jsFunction);
    // Ensure m_jsFunction is live JSObject as a quick sanity check (while it is already ensured by Weak<>). If this fails, this is possibly JSC GC side's bug.
    ASSERT(static_cast<JSC::JSCell*>(m_jsFunction.get())->isObject());

	return m_jsFunction.get();
}

inline JSC::JSObject* JSEventListener::ensureJSFunction(ScriptExecutionContext& scriptExecutionContext) const
{
	return ensureJSFunction(&scriptExecutionContext);
}

} // namespace WebCore
