/*
 *  Copyright (C) 2015 Canon Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "JSReadableStreamPrivateConstructors.h"

#if ENABLE(STREAMS_API)

#include "JSDOMBinding.h"
#include "JSDOMConstructor.h"
#include "JSReadableByteStreamController.h"
#include "JSReadableStream.h"
#include "JSReadableStreamBYOBReader.h"
#include "JSReadableStreamBYOBRequest.h"
#include "JSReadableStreamDefaultController.h"
#include "JSReadableStreamDefaultReader.h"
#include "ReadableByteStreamInternalsBuiltins.h"
#include "ReadableStreamInternalsBuiltins.h"
#include "WebCoreJSClientData.h"

using namespace JSC;

namespace WebCore {

enum class ReaderType {
    Byob,
    Default,
};

template <ReaderType type>
EncodedJSValue JSC_HOST_CALL constructJSReadableStreamReaderGeneric(ExecState*);

// Public JS ReadableStreamDefaultController constructor callback.
template<> EncodedJSValue JSC_HOST_CALL JSReadableStreamDefaultControllerConstructor::construct(ExecState* state)
{
	VM& vm = state->vm();
    return throwVMError(state, vm, createTypeError(state, ASCIILiteral("ReadableStreamDefaultController constructor should not be called directly")));
}

// Public JS ReadableStreamDefaultReader and ReadableStreamBYOBReader callbacks.
template<ReaderType type>
EncodedJSValue JSC_HOST_CALL constructJSReadableStreamReaderGeneric(ExecState* exec)
{
    VM& vm = exec->vm();
    auto& clientData = *static_cast<JSVMClientData*>(vm.clientData);
    auto& globalObject = *JSC::jsCast<JSDOMGlobalObject*>(exec->lexicalGlobalObject());

    JSC::JSObject* constructor;
    if (type == ReaderType::Byob)
        constructor = JSC::asObject(globalObject.get(vm, exec, clientData.builtinNames().ReadableStreamBYOBReaderPrivateName()));
    else
        constructor = JSC::asObject(globalObject.get(vm, exec, clientData.builtinNames().ReadableStreamDefaultReaderPrivateName()));

    auto constructData = JSC::getConstructData(vm, constructor);
    ASSERT(constructData.type != CallData::Type::None);

    MarkedArgumentBuffer args;
    args.append(exec->argument(0));
    return JSValue::encode(JSC::construct(exec, constructor, constructData, args));
}

template<> EncodedJSValue JSC_HOST_CALL JSReadableStreamDefaultReaderConstructor::construct(ExecState* exec)
{
    return constructJSReadableStreamReaderGeneric<ReaderType::Default>(exec);
}

template<> EncodedJSValue JSC_HOST_CALL JSReadableStreamBYOBReaderConstructor::construct(ExecState* exec)
{
    return constructJSReadableStreamReaderGeneric<ReaderType::Byob>(exec);
}

// Public JS ReadableByteStreamController and ReadableStreamBYOBRequest constructor callbacks.
template<> EncodedJSValue JSC_HOST_CALL JSReadableByteStreamControllerConstructor::construct(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);
    return throwVMTypeError(exec, vm, ASCIILiteral("ReadableByteStreamController constructor should not be called directly"));
}

template<> EncodedJSValue JSC_HOST_CALL JSReadableStreamBYOBRequestConstructor::construct(ExecState* exec)
{
    VM& vm = exec->vm();
    //auto scope = DECLARE_THROW_SCOPE(vm);
    return throwVMTypeError(exec, vm, ASCIILiteral("ReadableStreamBYOBRequest constructor should not be called directly"));
}

// Private JS ReadableStreamDefaultReader and ReadableStreamDefaultController constructors.
typedef JSBuiltinConstructor<JSReadableStreamDefaultReader> JSBuiltinReadableStreamDefaultReaderPrivateConstructor;
typedef JSBuiltinConstructor<JSReadableStreamDefaultController> JSBuiltinReadableStreamDefaultControllerPrivateConstructor;
// Private JS ReadableByteStreamController, ReadableStreamBYOBReader and ReadableStreamBYOBRequest constructors.
typedef JSBuiltinConstructor<JSReadableByteStreamController> JSBuiltinReadableByteStreamControllerPrivateConstructor;
typedef JSBuiltinConstructor<JSReadableStreamBYOBReader> JSBuiltinReadableStreamBYOBReaderPrivateConstructor;
typedef JSBuiltinConstructor<JSReadableStreamBYOBRequest> JSBuiltinReadableStreamBYOBRequestPrivateConstructor;

// STATIC_ASSERT_IS_TRIVIALLY_DESTRUCTIBLE(JSBuiltinReadableStreamReaderPrivateConstructor);
// STATIC_ASSERT_IS_TRIVIALLY_DESTRUCTIBLE(JSBuiltinReadableStreamControllerPrivateConstructor);
// STATIC_ASSERT_IS_TRIVIALLY_DESTRUCTIBLE(JSBuiltinReadableByteStreamControllerPrivateConstructor);
//STATIC_ASSERT_IS_TRIVIALLY_DESTRUCTIBLE(JSBuiltinReadableStreamBYOBReaderPrivateConstructor);
//STATIC_ASSERT_IS_TRIVIALLY_DESTRUCTIBLE(JSBuiltinReadableStreamBYOBRequestPrivateConstructor);

template<> const ClassInfo JSBuiltinReadableStreamDefaultReaderPrivateConstructor::s_info = { "ReadableStreamDefaultReaderPrivateConstructor", &Base::s_info, 0, CREATE_METHOD_TABLE(JSBuiltinReadableStreamDefaultReaderPrivateConstructor) };
template<> const ClassInfo JSBuiltinReadableStreamDefaultControllerPrivateConstructor::s_info = { "ReadableStreamDefaultControllerPrivateConstructor", &Base::s_info, 0, CREATE_METHOD_TABLE(JSBuiltinReadableStreamDefaultControllerPrivateConstructor) };
template<> const ClassInfo JSBuiltinReadableByteStreamControllerPrivateConstructor::s_info = { "ReadableByteStreamControllerPrivateConstructor", &Base::s_info, 0, CREATE_METHOD_TABLE(JSBuiltinReadableByteStreamControllerPrivateConstructor) };
template<> const ClassInfo JSBuiltinReadableStreamBYOBReaderPrivateConstructor::s_info = { "ReadableStreamBYOBReaderPrivateConstructor", &Base::s_info, 0, CREATE_METHOD_TABLE(JSBuiltinReadableStreamBYOBReaderPrivateConstructor) };
template<> const ClassInfo JSBuiltinReadableStreamBYOBRequestPrivateConstructor::s_info = { "ReadableStreamBYOBRequestPrivateConstructor", &Base::s_info, 0, CREATE_METHOD_TABLE(JSBuiltinReadableStreamBYOBRequestPrivateConstructor) };

template<> FunctionExecutable* JSBuiltinReadableStreamDefaultReaderPrivateConstructor::initializeExecutable(JSDOMGlobalObject& globalObject)
{
    return readableStreamInternalsPrivateInitializeReadableStreamDefaultReaderCodeGenerator(&globalObject);
}

template<> FunctionExecutable* JSBuiltinReadableStreamDefaultControllerPrivateConstructor::initializeExecutable(JSDOMGlobalObject& globalObject)
{
    return readableStreamInternalsPrivateInitializeReadableStreamDefaultControllerCodeGenerator(&globalObject);
}

template<> FunctionExecutable* JSBuiltinReadableByteStreamControllerPrivateConstructor::initializeExecutable(JSDOMGlobalObject& globalObject)
{
    return readableByteStreamInternalsPrivateInitializeReadableByteStreamControllerCodeGenerator(&globalObject);
}

template<> FunctionExecutable* JSBuiltinReadableStreamBYOBReaderPrivateConstructor::initializeExecutable(JSDOMGlobalObject& globalObject)
{
    return readableByteStreamInternalsPrivateInitializeReadableStreamBYOBReaderCodeGenerator(&globalObject);
}

template<> FunctionExecutable* JSBuiltinReadableStreamBYOBRequestPrivateConstructor::initializeExecutable(JSDOMGlobalObject& globalObject)
{
    return readableByteStreamInternalsPrivateInitializeReadableStreamBYOBRequestCodeGenerator(&globalObject);
}

JSObject* createReadableStreamDefaultReaderPrivateConstructor(VM& vm, JSDOMGlobalObject& globalObject)
{
    return JSBuiltinReadableStreamDefaultReaderPrivateConstructor::create(vm, JSBuiltinReadableStreamDefaultReaderPrivateConstructor::createStructure(vm, &globalObject, globalObject.objectPrototype()), &globalObject);
}

JSObject* createReadableStreamDefaultControllerPrivateConstructor(VM& vm, JSDOMGlobalObject& globalObject)
{
    return JSBuiltinReadableStreamDefaultControllerPrivateConstructor::create(vm, JSBuiltinReadableStreamDefaultControllerPrivateConstructor::createStructure(vm, &globalObject, globalObject.objectPrototype()), &globalObject);
}

JSObject* createReadableByteStreamControllerPrivateConstructor(VM& vm, JSDOMGlobalObject& globalObject)
{
    return JSBuiltinReadableByteStreamControllerPrivateConstructor::create(vm, JSBuiltinReadableByteStreamControllerPrivateConstructor::createStructure(vm, &globalObject, globalObject.objectPrototype()), &globalObject);
}

JSObject* createReadableStreamBYOBReaderPrivateConstructor(VM& vm, JSDOMGlobalObject& globalObject)
{
    return JSBuiltinReadableStreamBYOBReaderPrivateConstructor::create(vm, JSBuiltinReadableStreamBYOBReaderPrivateConstructor::createStructure(vm, &globalObject, globalObject.objectPrototype()), &globalObject);
}

JSObject* createReadableStreamBYOBRequestPrivateConstructor(VM& vm, JSDOMGlobalObject& globalObject)
{
    return JSBuiltinReadableStreamBYOBRequestPrivateConstructor::create(vm, JSBuiltinReadableStreamBYOBRequestPrivateConstructor::createStructure(vm, &globalObject, globalObject.objectPrototype()), &globalObject);
}

} // namespace WebCore

#endif
