/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CANON INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ReadableStream.h"

#include "JSReadableStreamSink.h"
#include "JSReadableStreamSource.h"
#include "WebCoreJSClientData.h"

using namespace JSC;

namespace WebCore {

PassRefPtr<ReadableStream> ReadableStream::create(JSC::ExecState* execState, PassRefPtr<ReadableStreamSource> source)
{
    VM& vm = execState->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    auto& clientData = *static_cast<JSVMClientData*>(vm.clientData);
    auto& globalObject = *JSC::jsCast<JSDOMGlobalObject*>(execState->lexicalGlobalObject());

    auto constructorValue = globalObject.get(vm, execState, clientData.builtinNames().ReadableStreamPrivateName());
    RETURN_IF_EXCEPTION(scope, nullptr);
    auto constructor = JSC::asObject(constructorValue);

    auto constructData = JSC::getConstructData(vm, constructor);
    ASSERT(constructData.type != CallData::Type::None);

    MarkedArgumentBuffer args;
    args.append(source ? toJS(execState, &globalObject, source.get()) : JSC::jsUndefined());

    JSObject* object = JSC::construct(execState, constructor, constructData, args);
    ASSERT(!!scope.exception() == !object);
	RETURN_IF_EXCEPTION(scope, nullptr);

    return create(globalObject, *jsDynamicCast<JSReadableStream*>(object));
}

static inline JSC::JSValue invokeReadableStreamFunction(JSC::ExecState& state, JSC::JSGlobalObject& lexicalGlobalObject, const JSC::Identifier& identifier, JSC::JSValue thisValue, const JSC::ArgList& arguments)
{
    JSC::VM& vm = lexicalGlobalObject.vm();
    JSC::JSLockHolder lock(vm);

    auto function = lexicalGlobalObject.get(vm, &state, identifier);
    ASSERT(function.isCallable(lexicalGlobalObject.vm()));

    auto scope = DECLARE_CATCH_SCOPE(vm);
    auto callData = JSC::getCallData(vm, function);
    ASSERT(callData.type != JSC::CallType::None);
    auto result = call(&state, function, callData, thisValue, arguments);
    EXCEPTION_ASSERT(!scope.exception() || isTerminatedExecutionException(lexicalGlobalObject.vm(), scope.exception()));
	if (scope.exception())
		return JSC::jsUndefined();
    return result;
}

void ReadableStream::pipeTo(ReadableStreamSink& sink)
{
	JSC::ExecState& state = *m_globalObject->globalExec();
    JSVMClientData* clientData = static_cast<JSVMClientData*>(state.vm().clientData);
    const Identifier& privateName = clientData->readableStreamInternalsBuiltinsNames().readableStreamPipeToPrivateName();

    MarkedArgumentBuffer arguments;
    arguments.append(readableStream());
    arguments.append(toJS(&state, m_globalObject.get(), sink));
    invokeReadableStreamFunction(state, *m_globalObject, privateName, JSC::jsUndefined(), arguments);
}

Optional<std::pair<RefPtr<ReadableStream>, RefPtr<ReadableStream>>> ReadableStream::tee()
{
    JSC::ExecState& state = *m_globalObject->globalExec();
    JSVMClientData* clientData = static_cast<JSVMClientData*>(state.vm().clientData);
    const Identifier& privateName = clientData->readableStreamInternalsBuiltinsNames().readableStreamTeePrivateName();

    MarkedArgumentBuffer arguments;
    arguments.append(readableStream());
    arguments.append(JSC::jsBoolean(true));
    JSValue returnedValue = invokeReadableStreamFunction(state, *m_globalObject, privateName, JSC::jsUndefined(), arguments);
    if (m_globalObject->vm().exception())
    	return Nullopt;

    Vector<RefPtr<ReadableStream>> results = toNativeArray<RefPtr<ReadableStream>>(state, returnedValue);

    ASSERT(results.size() == 2);
    return Optional<std::pair<RefPtr<ReadableStream>, RefPtr<ReadableStream>>>(std::make_pair(results[0].release(), results[1].release()));
}

void ReadableStream::lock()
{
	JSC::ExecState& state = *m_globalObject->globalExec();
    VM& vm = m_globalObject->vm();
    auto scope = DECLARE_CATCH_SCOPE(vm);

    auto& clientData = *static_cast<JSVMClientData*>(vm.clientData);

    auto constructorValue = m_globalObject->get(vm, &state, clientData.builtinNames().ReadableStreamDefaultReaderPrivateName());
    EXCEPTION_ASSERT(!scope.exception() || vm.hasPendingTerminationException());
	RETURN_IF_EXCEPTION(scope, void());

    auto constructor = JSC::asObject(constructorValue);

    auto constructData = JSC::getConstructData(vm, constructor);
    ASSERT(constructData.type != CallData::Type::None);

    MarkedArgumentBuffer args;
    args.append(readableStream());

    JSC::construct(&state, constructor, constructData, args);
    //scope.assertNoException();
}

static inline bool checkReadableStream(JSDOMGlobalObject& globalObject, JSReadableStream* readableStream, JSC::JSValue function)
{
    auto& state = *globalObject.globalExec();

    ASSERT(function);
    JSC::MarkedArgumentBuffer arguments;
    arguments.append(readableStream);

    auto& vm = globalObject.vm();
    auto scope = DECLARE_CATCH_SCOPE(vm);
    auto callData = JSC::getCallData(vm, function);
    ASSERT(callData.type != JSC::CallData::Type::None);

    auto result = call(&state, function, callData, JSC::jsUndefined(), arguments);
    EXCEPTION_ASSERT(!scope.exception() || isTerminatedExecutionException(lexicalGlobalObject.vm(), scope.exception()));

    return result.isTrue() || scope.exception();
}

bool ReadableStream::isLocked() const
{
    return checkReadableStream(*globalObject(), readableStream(), globalObject()->builtinInternalFunctions().readableStreamInternals().m_isReadableStreamLockedFunction.get());
}

bool ReadableStream::isDisturbed() const
{
    return checkReadableStream(*globalObject(), readableStream(), globalObject()->builtinInternalFunctions().readableStreamInternals().m_isReadableStreamDisturbedFunction.get());
}

bool ReadableStream::isDisturbed(ExecState& state, JSValue value)
{
    auto& vm = state.vm();
    auto& globalObject = *jsDynamicCast<JSDOMGlobalObject*>(state.lexicalGlobalObject());
    auto* readableStream = jsDynamicCast<JSReadableStream*>(value);

    return checkReadableStream(globalObject, readableStream, globalObject.builtinInternalFunctions().readableStreamInternals().m_isReadableStreamDisturbedFunction.get());
}

}
