/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY CANON INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CANON INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "JSDOMBinding.h"
#include "JSDOMConvert.h"
#include "JSDOMGuardedObject.h"
#include "JSReadableStream.h"
#include <wtf/Optional.h>

namespace WebCore {

class ReadableStreamSink;
class ReadableStreamSource;

class ReadableStream FINAL : public DOMGuarded<JSReadableStream> {
public:
    static Ref<ReadableStream> create(JSDOMGlobalObject& globalObject, JSReadableStream& readableStream) { return adoptRef(*new ReadableStream(globalObject, readableStream)); }

    static PassRefPtr<ReadableStream> create(JSC::ExecState*, PassRefPtr<ReadableStreamSource>);

    static bool isDisturbed(JSC::ExecState&, JSC::JSValue);

    Optional<std::pair<RefPtr<ReadableStream>, RefPtr<ReadableStream>>> tee();

    void lock();
    void pipeTo(ReadableStreamSink&);
    bool isLocked() const;
    bool isDisturbed() const;

    JSReadableStream* readableStream() const { return guarded(); }

protected:
    ReadableStream(JSDOMGlobalObject& globalObject, JSReadableStream& readableStream) : DOMGuarded<JSReadableStream>(globalObject, readableStream) { }
};

inline PassRefPtr<ReadableStream> toReadableStream(JSC::ExecState* state, JSC::JSValue value)
{
    JSC::VM& vm = state->vm();
    auto* globalObject = JSC::jsDynamicCast<JSDOMGlobalObject*>(state->lexicalGlobalObject());
    if (!globalObject)
        return nullptr;

    auto* readableStream = JSC::jsDynamicCast<JSReadableStream*>(value);
    if (!readableStream)
        return nullptr;

    return ReadableStream::create(*globalObject, *readableStream);
}

/*
struct JSReadableStreamWrapperConverter {
    static RefPtr<ReadableStream> toWrapped(JSC::ExecState& state, JSC::JSValue value)
    {
        JSC::VM& vm = state.vm();
        auto* globalObject = jsDynamicDowncast<JSDOMGlobalObject*>(vm, state.lexicalGlobalObject());
        if (!globalObject)
            return nullptr;

        auto* readableStream = jsDynamicDowncast<JSReadableStream*>(vm, value);
        if (!readableStream)
            return nullptr;

        return ReadableStream::create(*globalObject, *readableStream);
    }
};

template<> struct JSDOMWrapperConverterTraits<ReadableStream> {
    using WrapperClass = JSReadableStreamWrapperConverter;
    using ToWrappedReturnType = RefPtr<ReadableStream>;
    static constexpr bool needsState = true;
};
*/

inline JSC::JSValue toJS(JSC::ExecState*, JSC::JSGlobalObject*, ReadableStream* stream)
{
	// Getting an error Response stream in response-consume-stream
    return stream ? stream->readableStream() : JSC::jsNull();
}

template<>
struct NativeValueTraits<RefPtr<ReadableStream>> {
    static inline void nativeValue(JSC::VM&, JSC::ExecState* exec, JSC::JSValue jsValue, RefPtr<ReadableStream>& indexedValue)
    {
    	indexedValue = toReadableStream(exec, jsValue);
    }
};

}
