/*
 * Copyright (C) 2009, 2011 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ScriptState.h"

#include "Frame.h"
#include "JSDOMWindowBase.h"
#include "Node.h"
#include "Page.h"
#include "WorkerGlobalScope.h"
#include "WorkerScriptController.h"
#include <heap/StrongInlines.h>
#include <interpreter/CallFrame.h>
#include <runtime/JSGlobalObject.h>

#if ENABLE(WORKERS)
#include "JSWorkerGlobalScope.h"
#endif

namespace WebCore {

ScriptStateProtectedPtr::~ScriptStateProtectedPtr() = default;

ScriptStateProtectedPtr::ScriptStateProtectedPtr(ScriptState* scriptState)
    : m_globalObject(scriptState->globalData(), scriptState->lexicalGlobalObject())
{
}

ScriptState* ScriptStateProtectedPtr::get() const
{
    if (m_globalObject)
        return const_cast<JSC::JSGlobalObject*>(m_globalObject.get())->globalExec();
    return nullptr;
}

DOMWindow* domWindowFromScriptState(ScriptState* scriptState)
{
    JSC::JSGlobalObject* globalObject = scriptState->lexicalGlobalObject();
    if (!globalObject->inherits(JSDOMWindowBase::info()))
        return nullptr;
    return &JSC::jsCast<JSDOMWindowBase*>(globalObject)->wrapped();
}

ScriptExecutionContext* scriptExecutionContextFromExecState(JSC::JSGlobalObject* globalObject)
{
    if (!globalObject->inherits(JSDOMGlobalObject::info()))
        return nullptr;
    return JSC::jsCast<JSDOMGlobalObject*>(globalObject)->scriptExecutionContext();
}

ScriptExecutionContext* scriptExecutionContextFromScriptState(ScriptState* scriptState)
{
    return scriptExecutionContextFromExecState(scriptState->lexicalGlobalObject());
}

ScriptExecutionContext* scriptExecutionContextFromExecState(JSC::ExecState* scriptState)
{
    return scriptExecutionContextFromExecState(scriptState->lexicalGlobalObject());
}

bool evalEnabled(ScriptState* scriptState)
{
    JSC::JSGlobalObject* globalObject = scriptState->lexicalGlobalObject();
    return globalObject->evalEnabled();
}

void setEvalEnabled(ScriptState* scriptState, bool enabled)
{
    JSC::JSGlobalObject* globalObject = scriptState->lexicalGlobalObject();
    return globalObject->setEvalEnabled(enabled);
}

ScriptState* mainWorldScriptState(Frame* frame)
{
    if (!frame)
        return nullptr;
    JSDOMWindowShell* shell = frame->script()->windowShell(mainThreadNormalWorld());
    return shell->window()->globalExec();
}

JSC::JSGlobalObject* globalObject(DOMWrapperWorld& world, Node* node)
{
    if (!node)
        return nullptr;
    Frame* frame = node->document().frame();
    if (!frame)
        return nullptr;
    if (!frame->script()->canExecuteScripts(NotAboutToExecuteScript))
        return nullptr;
    return frame->script()->globalObject(world);
}

JSC::JSGlobalObject* globalObject(DOMWrapperWorld& world, Page* page)
{
    return page ? page->mainFrame()->script()->globalObject(world) : nullptr;
}

JSC::JSGlobalObject* globalObject(WorkerGlobalScope& workerOrWorkletGlobalScope)
{
    if (auto* scriptController = workerOrWorkletGlobalScope.script())
        return scriptController->globalScopeWrapper();
    return nullptr;
}

ScriptState* scriptStateFromNode(DOMWrapperWorld* world, Node* node)
{
    return globalObject(*world, node)->globalExec();
}

ScriptState* scriptStateFromPage(DOMWrapperWorld* world, Page* page)
{
    return page->mainFrame()->script()->globalObject(world)->globalExec();
}

JSC::ExecState* execStateFromNode(DOMWrapperWorld& world, Node* node)
{
	return scriptStateFromNode(&world, node);
}

JSC::ExecState* execStateFromPage(DOMWrapperWorld& world, Page* page)
{
	return scriptStateFromPage(&world, page);
}

#if ENABLE(WORKERS)
ScriptState* scriptStateFromWorkerGlobalScope(WorkerGlobalScope* workerGlobalScope)
{
    return workerGlobalScope->script()->workerGlobalScopeWrapper()->globalExec();
}

ScriptState* execStateFromWorkerGlobalScope(WorkerGlobalScope* workerGlobalScope)
{
    return scriptStateFromWorkerGlobalScope(workerGlobalScope);
}

#endif

}
