/*
 * Copyright (C) 2016 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"
#include "StructuredClone.h"

#include "ExceptionCode.h"
#include "JSDOMBinding.h"
#include <runtime/JSTypedArrays.h>
#include <runtime/JSGenericTypedArrayViewInlines.h>

using namespace JSC;

namespace WebCore {

enum class CloneMode {
    Full,
    Partial,
};

EncodedJSValue JSC_HOST_CALL cloneArrayBufferImpl(ExecState* state, CloneMode mode)
{
    ASSERT(execState);
    ASSERT(execState->argumentCount());
    ASSERT(execState->lexicalGlobalObject());

    ArrayBuffer* buffer = toArrayBuffer(state->uncheckedArgument(0));
    if (!buffer) {
        setDOMException(state, DATA_CLONE_ERR);
        return JSValue::encode(jsUndefined());
    }
    if (mode == CloneMode::Partial) {
        ASSERT(state->argumentCount() == 3);
        int srcByteOffset = static_cast<int>(state->uncheckedArgument(1).toNumber(state));
        int srcLength = static_cast<int>(state->uncheckedArgument(2).toNumber(state));
        return JSValue::encode(JSArrayBuffer::create(state->vm(), state->lexicalGlobalObject()->arrayBufferStructure(), buffer->slice(srcByteOffset, srcByteOffset + srcLength)));
    }
    return JSValue::encode(JSArrayBuffer::create(state->vm(), state->lexicalGlobalObject()->arrayBufferStructure(), ArrayBuffer::tryCreate(buffer->data(), buffer->byteLength())));
}

EncodedJSValue JSC_HOST_CALL cloneArrayBuffer(ExecState* state)
{
    return cloneArrayBufferImpl(state, CloneMode::Partial);
}

EncodedJSValue JSC_HOST_CALL structuredCloneForStream(ExecState* execState)
{
    ASSERT(execState);
    ASSERT(execState->argumentCount());

    VM& vm = execState->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);

    JSValue value = execState->uncheckedArgument(0);

    if (value.inherits<JSArrayBuffer>(vm))
        RELEASE_AND_RETURN(scope, cloneArrayBufferImpl(execState, CloneMode::Full));

    if (value.inherits<JSArrayBufferView>(vm)) {
        auto* bufferView = jsCast<JSArrayBufferView*>(value);
		ASSERT(bufferView);

		ArrayBuffer* buffer = bufferView->buffer();
		if (!buffer) {
			setDOMException(execState, DATA_CLONE_ERR);
			return JSValue::encode(jsUndefined());
		}
		RefPtr<ArrayBuffer> bufferClone = ArrayBuffer::tryCreate(buffer->data(), buffer->byteLength());
		if (!bufferClone) {
			setDOMException(execState, DATA_CLONE_ERR);
			return JSValue::encode(jsUndefined());
		}
		Structure* structure = bufferView->structure(vm);

#define CLONE_TYPED_ARRAY(name) \
		do { \
			if (bufferView->inherits<JS##name##Array>(vm)) \
				return JSValue::encode(JS##name##Array::create(execState, structure, bufferClone.release(), bufferView->byteOffset(), bufferView->length())); \
		} while (0);

		FOR_EACH_TYPED_ARRAY_TYPE_EXCLUDING_DATA_VIEW(CLONE_TYPED_ARRAY)

#undef CLONE_TYPED_ARRAY

		if (value.inherits<JSDataView>(vm))
			return JSValue::encode(JSDataView::create(execState, structure, bufferClone.release(), bufferView->byteOffset(), bufferView->length()));
	}

	throwTypeError(execState, vm, "structuredClone not implemented for non-ArrayBuffer / non-ArrayBufferView");
    return JSValue::encode(jsUndefined());
}

} // namespace WebCore
