/*
 *  Copyright (C) 1999-2001 Harri Porten (porten@kde.org)
 *  Copyright (C) 2003, 2004, 2005, 2006, 2008, 2009 Apple Inc. All rights reserved.
 *  Copyright (C) 2007 Samuel Weinig <sam@webkit.org>
 *  Copyright (C) 2009 Google, Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#pragma once

#include "DOMWrapperWorld.h"
#include "WebCoreBuiltinNames.h"
#include "WebCoreJSBuiltins.h"
#include "WebCoreTypedArrayController.h"
#include <wtf/HashSet.h>
#include <wtf/RefPtr.h>
#include <wtf/WeakHashSet.h>
#include <wtf/WeakPtr.h>

#if ENABLE(STREAMS_API)
#include "ReadableByteStreamInternalsBuiltinsWrapper.h"
#include "ReadableStreamInternalsBuiltinsWrapper.h"
#include "StreamInternalsBuiltinsWrapper.h"
#include "TransformStreamInternalsBuiltinsWrapper.h"
#include "WritableStreamInternalsBuiltinsWrapper.h"
#endif

#include "JSDOMBindingInternalsBuiltinsWrapper.h"
#if ENABLE(WEB_RTC)
#include "RTCPeerConnectionInternalsBuiltinsWrapper.h"
#endif

namespace WebCore {

class JSVMClientDataClient : public CanMakeWeakPtr<JSVMClientDataClient> {
public:
    virtual ~JSVMClientDataClient() = default;
    virtual void willDestroyVM() = 0;
};

class JSVMClientData : public JSC::JSGlobalData::ClientData {
    WTF_MAKE_NONCOPYABLE(JSVMClientData); WTF_MAKE_FAST_ALLOCATED;
    friend class JSGlobalDataWorldIterator;
    friend void initNormalWorldClientData(JSC::JSGlobalData*);

public:
    explicit JSVMClientData(JSC::JSGlobalData& vm)
	: m_builtinNames(vm)
#if ENABLE(STREAMS_API)
    , m_readableByteStreamInternalsBuiltinsNames(vm)
    , m_readableStreamInternalsBuiltinsNames(vm)
    , m_streamInternalsBuiltinsNames(vm)
    , m_transformStreamInternalsBuiltinsNames(vm)
    , m_writableStreamInternalsBuiltinsNames(vm)
#endif
    , m_jsDOMBindingInternalsBuiltinsNames(vm)
#if ENABLE(WEB_RTC)
    , m_rtcPeerConnectionInternalsBuiltinsNames(vm)
#endif
	{
#if ENABLE(STREAMS_API)
    	m_readableByteStreamInternalsBuiltinsNames.exportNames();
    	m_readableStreamInternalsBuiltinsNames.exportNames();
    	m_streamInternalsBuiltinsNames.exportNames();
    	m_transformStreamInternalsBuiltinsNames.exportNames();
    	m_writableStreamInternalsBuiltinsNames.exportNames();
#endif
    	m_jsDOMBindingInternalsBuiltinsNames.exportNames();
#if ENABLE(WEB_RTC)
    	m_rtcPeerConnectionInternalsBuiltinsNames.exportNames();
#endif
	}

    virtual ~JSVMClientData()
    {
    	for (auto& client : m_clients)
            client.willDestroyVM();

        ASSERT(m_worldSet.contains(m_normalWorld.get()));
        ASSERT(m_worldSet.size() == 1);
        ASSERT(m_normalWorld->hasOneRef());
        m_normalWorld.clear();
        ASSERT(m_worldSet.isEmpty());
    }

    DOMWrapperWorld* normalWorld() { return m_normalWorld.get(); }

    void getAllWorlds(Vector<RefPtr<DOMWrapperWorld> >& worlds)
    {
        copyToVector(m_worldSet, worlds);
    }

    void rememberWorld(DOMWrapperWorld* world)
    {
        ASSERT(!m_worldSet.contains(world));
        m_worldSet.add(world);
    }
    void forgetWorld(DOMWrapperWorld* world)
    {
        ASSERT(m_worldSet.contains(world));
        m_worldSet.remove(world);
    }

    virtual String overrideSourceURL(const JSC::StackFrame&, const String& originalSourceURL) const override;

    WebCoreBuiltinNames& builtinNames() { return m_builtinNames; }

#if ENABLE(STREAMS_API)
    ReadableByteStreamInternalsBuiltinsNames& readableByteStreamInternalsBuiltinsNames() { return m_readableByteStreamInternalsBuiltinsNames; }
    ReadableStreamInternalsBuiltinsNames& readableStreamInternalsBuiltinsNames() { return m_readableStreamInternalsBuiltinsNames; }
    StreamInternalsBuiltinsNames& streamInternalsBuiltinsNames() { return m_streamInternalsBuiltinsNames; }
    TransformStreamInternalsBuiltinsNames& transformStreamInternalsBuiltinsNames() { return m_transformStreamInternalsBuiltinsNames; }
    WritableStreamInternalsBuiltinsNames& writableStreamInternalsBuiltinsNames() { return m_writableStreamInternalsBuiltinsNames; }
#endif

    JSDOMBindingInternalsBuiltinsNames& jsDOMBindingInternalsBuiltinsNames() { return m_jsDOMBindingInternalsBuiltinsNames; }
#if ENABLE(WEB_RTC)
    RTCPeerConnectionInternalsBuiltinsNames& rtcPeerConnectionInternalsBuiltinsNames() { return m_rtcPeerConnectionInternalsBuiltinsNames; }
#endif

    void addClient(JSVMClientDataClient& client) { m_clients.add(client); }

private:
    HashSet<DOMWrapperWorld*> m_worldSet;
    RefPtr<DOMWrapperWorld> m_normalWorld;

    WebCoreBuiltinNames m_builtinNames;

#if ENABLE(STREAMS_API)
    ReadableByteStreamInternalsBuiltinsNames m_readableByteStreamInternalsBuiltinsNames;
    ReadableStreamInternalsBuiltinsNames m_readableStreamInternalsBuiltinsNames;
    StreamInternalsBuiltinsNames m_streamInternalsBuiltinsNames;
    TransformStreamInternalsBuiltinsNames m_transformStreamInternalsBuiltinsNames;
    WritableStreamInternalsBuiltinsNames m_writableStreamInternalsBuiltinsNames;
#endif
    JSDOMBindingInternalsBuiltinsNames m_jsDOMBindingInternalsBuiltinsNames;
#if ENABLE(WEB_RTC)
    RTCPeerConnectionInternalsBuiltinsNames m_rtcPeerConnectionInternalsBuiltinsNames;
#endif

    WeakHashSet<JSVMClientDataClient> m_clients;
};

inline String JSVMClientData::overrideSourceURL(const JSC::StackFrame&, const String& originalSourceURL) const
{
	return originalSourceURL;
}

inline void initNormalWorldClientData(JSC::JSGlobalData* vm)
{
	JSVMClientData* clientData = new JSVMClientData(*vm);
    vm->clientData = clientData; // ~JSGlobalData deletes this pointer.
    clientData->m_normalWorld = DOMWrapperWorld::create(vm, true);
    vm->m_typedArrayController = adoptRef(new WebCoreTypedArrayController());
}

} // namespace WebCore
