/*
 *  Copyright (C) 2010 Espial Group Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "ArgList.h"
#include "Error.h"
#include "ErrorInstance.h"
#include "DateConversion.h"
#include "DateInstance.h"
#include "JSDOMWindow.h"
#include "JSObjectBridge.h"
#include "JSGlobalObject.h"
#include "JSDocument.h"
#include "JSFunction.h"
#include "JSGlobalObject.h"
#include "JSExecState.h"
#include "JSNode.h"
#include "HTMLImageElement.h"
#include "File.h"
#include "Frame.h"
#include "JSFile.h"
#include "JSNode.h"
#include "JSHTMLObjectElement.h"
#include "JSWebGLTexture.h"
#include "DOMWrapperWorld.h"
#include "SlotVisitorInlineMethods.h"
#include "MarkStack.h"
#include "Operations.h"
#include "StructureInlines.h"
#include <runtime/VMEntryScope.h>
#include <wtf/MathExtras.h>
#include <wtf/NakedPtr.h>

using namespace JSC;

namespace
{

class JSCustomFunction : public InternalFunction {
private:
	struct Data
	{
		WTF_MAKE_FAST_ALLOCATED;
	public:
		Data(WebCoreAPIs::JSCustomCall customCall, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
		: m_customCall(customCall)
		, m_customDestroy(customDestroy)
		, m_customContext(customContext)
		{
		}

		EncodedJSValue call(ExecState* exec){
			if(m_customCall) {
				VM& vm = exec->vm();
				return m_customCall(exec, m_customContext);
			}

			return JSValue::encode(jsUndefined());
		}

		~Data()
		{
	    	if(m_customDestroy)
	    		m_customDestroy(m_customContext);
		}

	    WebCoreAPIs::JSCustomCall m_customCall;
	    WebCoreAPIs::JSCustomDestroy m_customDestroy;
	    void* m_customContext;
	};
public:
    typedef InternalFunction Base;

    DECLARE_INFO;

    static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue prototype)
    {
        return Structure::create(globalData, globalObject, prototype, TypeInfo(InternalFunctionType, StructureFlags), info());
    }

	JSCustomFunction(VM& vm, JSC::Structure* structure, WebCoreAPIs::JSCustomCall customCall, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
    : Base(vm, structure, call, nullptr)
	, m_private(adoptPtr(new Data(customCall, customDestroy, customContext)))
    {
    }

    static void destroy(JSCell* cell)
    {
    	static_cast<JSCustomFunction*>(cell)->JSCustomFunction::~JSCustomFunction();
    }

    void finishCreation(JSGlobalData& globalData, const String& name)
    {
    	Base::finishCreation(globalData, name);
    }

    static const unsigned StructureFlags = OverridesGetCallData | Base::StructureFlags;

private:
    static EncodedJSValue JSC_HOST_CALL call(ExecState* exec);

private:
    OwnPtr<Data> m_private;
};

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSCustomFunction);

const ClassInfo JSCustomFunction::s_info = { "Function", &Base::s_info, nullptr, CREATE_METHOD_TABLE(JSCustomFunction) };

EncodedJSValue JSC_HOST_CALL JSCustomFunction::call(ExecState* exec)
{
	if(asObject(exec->callee())->inherits(JSCustomFunction::info())) {
		return jsCast<JSCustomFunction*>(asObject(exec->callee()))->m_private->call(exec);
	}

	return JSValue::encode(jsUndefined());
}

} // anonymous namespace

namespace
{

class JSCustomConstructor : public JSDestructibleObject
{
private:
	struct Data
	{
		WTF_MAKE_FAST_ALLOCATED;
	public:
		Data(WebCoreAPIs::JSCustomConstruct customConstruct, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
		: m_customConstruct(customConstruct)
		, m_customDestroy(customDestroy)
		, m_customContext(customContext)
		{
	    }

		~Data()
		{
	    	if(m_customDestroy)
	    		m_customDestroy(m_customContext);
		}

		EncodedJSValue construct(ExecState* exec)
		{
			if(m_customConstruct)
				return m_customConstruct(exec, m_customContext);
			return JSValue::encode(jsUndefined());
		}

		WebCoreAPIs::JSCustomConstruct m_customConstruct;
		WebCoreAPIs::JSCustomDestroy m_customDestroy;
		void* m_customContext;
	};

public:
    typedef JSDestructibleObject Base;
    static const unsigned StructureFlags = ImplementsHasInstance | Base::StructureFlags;

    static JSCustomConstructor* create(JSGlobalObject* globalObject, WebCoreAPIs::JSCustomConstruct customConstruct, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
    {
    	VM& vm = globalObject->vm();
    	Structure* structure = createStructure(vm, globalObject, globalObject->objectPrototype());
    	JSCustomConstructor* constructor = new (NotNull, allocateCell<JSCustomConstructor>(vm.heap)) JSCustomConstructor(vm, structure, customConstruct, customDestroy, customContext);
    	constructor->Base::finishCreation(vm);
        return constructor;
    }

    DECLARE_INFO;

    static Structure* createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
    {
        return Structure::create(globalData, globalObject, proto, TypeInfo(ObjectType, StructureFlags), info());
    }

    static void destroy(JSCell* cell)
    {
    	static_cast<JSCustomConstructor*>(cell)->JSCustomConstructor::~JSCustomConstructor();
    }

private:
    static EncodedJSValue JSC_HOST_CALL construct(ExecState* exec);

    static CallData getConstructData(JSCell*)
    {
    	CallData constructData;
    	constructData.type = CallData::Type::Native;
        constructData.native.function = construct;
        return constructData;
    }

private:
    JSCustomConstructor(VM& vm, Structure* structure, WebCoreAPIs::JSCustomConstruct customConstruct, WebCoreAPIs::JSCustomDestroy customDestroy, void* customContext)
	: Base(vm, structure)
	, m_private(adoptPtr(new Data(customConstruct, customDestroy, customContext)))
	{
	}

    OwnPtr<Data> m_private;
};

const ClassInfo JSCustomConstructor::s_info = { "Constructor", &Base::s_info, nullptr, CREATE_METHOD_TABLE(JSCustomConstructor) };

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSCustomConstructor);

EncodedJSValue JSC_HOST_CALL JSCustomConstructor::construct(ExecState* exec)
{
	if(asObject(exec->callee())->inherits(JSCustomConstructor::info())){
		return jsCast<JSCustomConstructor*>(asObject(exec->callee()))->m_private->construct(exec);
	}

	return JSValue::encode(jsUndefined());
}

} // // anonymous namespace

namespace WebCoreAPIs
{

JSCustomInstance::~JSCustomInstance()
{
}

Structure* JSCustomInstance::createStructure(JSGlobalData& globalData, JSGlobalObject* globalObject, JSValue proto)
{
	return JSC::Structure::create(globalData, globalObject, proto, JSC::TypeInfo(JSC::ObjectType, StructureFlags), info(), JSC::MayHaveIndexedAccessors);
}

bool JSCustomInstance::put(JSC::JSCell* cell, JSC::ExecState* exec, JSC::PropertyName propertyName, JSC::JSValue value, JSC::PutPropertySlot& slot)
{
	VM& vm = exec->vm();

	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(cell);

	if(customInstance->m_private->put(customInstance, exec, propertyName, value, slot))
		return true;

	return customInstance->Base::putDirect(vm, propertyName, value, slot);
}

bool JSCustomInstance::putByIndex(JSC::JSCell* cell, JSC::ExecState* exec, unsigned propertyName, JSC::JSValue value, bool shouldThrow)
{
	VM& vm = exec->vm();

	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(cell);

	if(customInstance->m_private->putByIndex(customInstance, exec, propertyName, value, shouldThrow))
		return true;

    PutPropertySlot slot(cell, shouldThrow);
    return JSCustomInstance::put(cell, exec, Identifier::from(vm, propertyName), value, slot);
}

bool JSCustomInstance::getOwnPropertySlot(JSObject* object, ExecState* exec, PropertyName propertyName, PropertySlot& slot)
{
	VM& vm = exec->vm();

	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(object);

	if(customInstance->m_private->get(customInstance, exec, propertyName, slot))
		return true;

	return JSObject::getOwnPropertySlot(customInstance, exec, propertyName, slot);
}

bool JSCustomInstance::getOwnPropertySlotByIndex(JSObject* object, ExecState* exec, unsigned propertyName, PropertySlot& slot)
{
	VM& vm = exec->vm();

	JSCustomInstance* customInstance = jsCast<JSCustomInstance*>(object);

	if(customInstance->m_private->getByIndex(customInstance, exec, propertyName, slot))
		return true;

	return JSCustomInstance::getOwnPropertySlot(object, exec, Identifier::from(vm, propertyName), slot);
}

void JSCustomInstance::destroy(JSCell* cell)
{
	static_cast<JSCustomInstance*>(cell)->JSCustomInstance::~JSCustomInstance();
}

void JSCustomInstance::visitChildren(JSCell* cell, SlotVisitor& visitor)
{
	Base::visitChildren(cell, visitor);
	jsCast<JSCustomInstance*>(cell)->m_private->visitChildren(jsCast<JSCustomInstance*>(cell), visitor);
}

DLL_PUBLIC void append(JSC::SlotVisitor& visitor, JSC::JSValue v)
{
	if(v.isUndefinedOrNull())
		return;

	visitor.appendUnbarrieredValue(&v);
}

ALWAYS_INLINE JSCustomInstance::JSCustomInstance(VM& vm, JSC::Structure* structure, const JSCustomObjectConstructorData& constructorData)
: Base(vm, structure)
, m_private(adoptPtr(new Data(constructorData)))
{
}

void JSCustomInstance::finishCreation(JSGlobalData& globalData)
{
	Base::finishCreation(globalData);
}

const ClassInfo JSCustomInstance::s_info = { "Object", &Base::s_info, nullptr, CREATE_METHOD_TABLE(JSCustomInstance) };

ASSERT_HAS_TRIVIAL_DESTRUCTOR(JSCustomInstance);

DLL_PUBLIC JSCustomInstance* asJSCustomInstance(JSC::JSValue value)
{
	if(value.isCell()){
		JSCell* cell = value.asCell();

		if(cell->inherits(JSCustomInstance::info()))
			return jsCast<JSCustomInstance*>(cell);
	}

	return 0;
}

DLL_PUBLIC JSC::PrivateName createPrivateName()
{
	return JSC::PrivateName();
}

DLL_PUBLIC JSC::JSValue get(JSC::JSValue value, JSC::ExecState* exec, JSC::PropertyName propertyName)
{
	VM& vm = exec->vm();
	return value.get(vm, exec, propertyName);
}

DLL_PUBLIC JSC::JSValue get(JSC::JSValue value, JSC::ExecState* exec, const String& propertyName)
{
	VM& vm = exec->vm();
	return value.get(vm, exec, Identifier::fromString(vm, propertyName));
}

DLL_PUBLIC JSC::JSValue get(JSC::JSValue value, JSC::ExecState* exec, unsigned propertyName)
{
	VM& vm = exec->vm();
	return value.get(vm, exec, propertyName);
}

DLL_PUBLIC JSC::JSValue get(JSC::ExecState* exec, JSC::PropertySlot slot, JSC::PropertyName propertyName)
{
	return slot.getValue(exec, propertyName);
}

DLL_PUBLIC bool getDirect(JSC::JSObject* thisObject, JSC::ExecState* exec, const String& propertyName, JSC::PropertySlot& slot)
{
	return JSObject::getOwnPropertySlot(thisObject, exec, Identifier::fromString(exec->vm(), propertyName), slot);
}

DLL_PUBLIC bool getDirect(JSC::JSObject* thisObject, JSC::ExecState* exec, JSC::PropertyName propertyName, JSC::PropertySlot& slot)
{
	return JSObject::getOwnPropertySlot(thisObject, exec, propertyName, slot);
}

DLL_PUBLIC void putDirect(JSC::JSObject* thisObject, JSC::JSGlobalData& globalData, JSC::PropertyName propertyName, JSC::JSValue value, unsigned attributes)
{
	thisObject->JSObject::putDirect(globalData, propertyName, value, attributes);
}

DLL_PUBLIC void putDirect(JSC::JSObject* thisObject, JSC::ExecState* exec, const String& propertyName, JSC::JSValue value, unsigned attributes)
{
	JSC::VM& vm = exec->vm();
	thisObject->putDirect(vm, Identifier::fromString(vm, propertyName), value, attributes);
}

DLL_PUBLIC void putDirect(JSC::JSObject* thisObject, JSC::VM& vm, const String& propertyName, JSC::JSValue value, unsigned attributes)
{
	thisObject->putDirect(vm, Identifier::fromString(vm, propertyName), value, attributes);
}

DLL_PUBLIC WebCore::Node* toNode(JSC::JSValue value)
{
	return WebCore::toNode(value);
}

DLL_PUBLIC WebCore::JSHTMLObjectElement* toJSHTMLObjectElement(JSC::JSValue thisValue)
{
	return JSC::jsDynamicCast<WebCore::JSHTMLObjectElement*>(thisValue);
}

#if ENABLE(WEBGL)
DLL_PUBLIC unsigned int toWebGLTextureObject(JSC::JSValue value)
{
    if(WebCore::WebGLTexture* texture = WebCore::toWebGLTexture(value))
    	return texture->object();
    return 0;
}
#endif

DLL_PUBLIC WebCore::HTMLObjectElement* toHTMLObjectElement(JSC::JSValue thisValue)
{
	if(WebCore::JSHTMLObjectElement* jsElement = JSC::jsDynamicCast<WebCore::JSHTMLObjectElement*>(thisValue))
		return static_cast<WebCore::HTMLObjectElement*>(&jsElement->wrapped());

	return nullptr;
}

DLL_PUBLIC JSC::JSObject* createJSFunction(JSGlobalObject* globalObject, JSCustomCall customCall, JSCustomDestroy customDestroy, void* customContext)
{
	VM& vm = globalObject->vm();
	JSC::Structure* structure = JSCustomFunction::createStructure(vm, globalObject, globalObject->functionPrototype());
	JSCustomFunction* function = new (NotNull, allocateCell<JSCustomFunction>(vm.heap)) JSCustomFunction(vm, structure, customCall, customDestroy, customContext);
    function->finishCreation(vm, emptyString());
    return function;
}

DLL_PUBLIC JSC::JSObject* createJSFunction(JSC::ExecState* exec, JSCustomCall customCall, JSCustomDestroy customDestroy, void* customContext)
{
	JSGlobalObject* globalObject = exec->lexicalGlobalObject();
	return createJSFunction(globalObject, customCall, customDestroy, customContext);
}

DLL_PUBLIC JSC::JSObject* createJSError(JSC::ExecState* exec, const String& message)
{
	return JSC::createError(exec, message, 0);
}

DLL_PUBLIC JSC::JSFinalObject* createJSFinalObject(JSC::ExecState* exec)
{
	JSGlobalObject* globalObject = exec->lexicalGlobalObject();
	VM& vm = globalObject->vm();
	JSC::Structure* structure = JSFinalObject::createStructure(vm, globalObject, globalObject->objectPrototype(), JSFinalObject::defaultInlineCapacity());
	return JSFinalObject::create(vm, structure);
}

DLL_PUBLIC JSC::JSObject* createJSObject(JSC::JSGlobalObject* globalObject, const JSCustomObjectConstructorData& constructorData)
{
	VM& vm = globalObject->vm();
	JSC::Structure* structure = JSCustomInstance::createStructure(vm, globalObject, globalObject->objectPrototype());
	JSCustomInstance* customInstance = new (NotNull, JSC::allocateCell<JSCustomInstance>(vm.heap)) JSCustomInstance(vm, structure, constructorData);
	customInstance->finishCreation(vm);
	return customInstance;
}

DLL_PUBLIC JSC::JSObject* createJSObject(JSC::ExecState* exec, const JSCustomObjectConstructorData& constructorData)
{
	return createJSObject(exec->lexicalGlobalObject(), constructorData);
}

DLL_PUBLIC JSC::JSObject* createJSConstructor(JSGlobalObject* globalObject, JSCustomConstruct customConstruct, JSCustomDestroy customDestroy, void* customContext)
{
	return JSCustomConstructor::create(globalObject, customConstruct, customDestroy, customContext);
}

DLL_PUBLIC bool isJSArray(JSValue v)
{
	return JSC::isJSArray(v);
}

DLL_PUBLIC JSArray* JSArrayBridge::tryCreate(ExecState* exec)
{
	auto* lexicalGlobalObject = exec->lexicalGlobalObject();
	auto& vm = lexicalGlobalObject->vm();
	return JSArray::tryCreate(vm, exec, lexicalGlobalObject->arrayStructureForIndexingTypeDuringAllocation(ArrayWithUndecided));
}

DLL_PUBLIC void JSArrayBridge::push(JSArray* jsArray, ExecState* exec, JSValue value)
{
	jsArray->push(exec, value);
}

DLL_PUBLIC void JSArrayBridge::setLength(JSArray* jsArray, ExecState* exec, unsigned newLength)
{
	jsArray->setLength(exec, newLength);
}

DLL_PUBLIC JSValue newDateInstance(ExecState* exec, double d)
{
    return DateInstance::create(exec->globalData(), exec->lexicalGlobalObject()->dateStructure(), d);
}

DLL_PUBLIC JSValue newDateInstance(
		ExecState* exec,
		int year,	/* specified as an number, i.e. 2010 */
		int month, 	/* specified as a number (0-11), i.e. 5 */
		int day,  	/* day of the month specified as a number (1-31), i.e. 18 */
		int hour, 	/* Hour specified as a number (0-23), i.e. 13 */
		int minute, /* specified as a number (0-59), i.e. 34 */
		int second, /* specified as a number (0-59), i.e. 23 */
		int ms		/* milli-second specified as a number (0-999), i.e. 240 */
		)
{
    GregorianDateTime gregorianDateTime;
    
    gregorianDateTime.setYear(year);
	gregorianDateTime.setMonth(month);
	gregorianDateTime.setMonthDay(day);
	gregorianDateTime.setHour(hour);
	gregorianDateTime.setMinute(minute);
	gregorianDateTime.setSecond(second);

	return newDateInstance(exec, gregorianDateTimeToMS(exec->vm(), gregorianDateTime, 0, WTF::LocalTime));
}

DLL_PUBLIC bool parseDateInstance(
		ExecState* exec,
		JSValue jsValue,
		int& year,	/* specified as an number, i.e. 2010 */
		int& month,	/* specified as a number (0-11), i.e. 5 */
		int& day,  	/* day of the month specified as a number (1-31), i.e. 18 */
		int& hour, 	/* Hour specified as a number (0-23), i.e. 13 */
		int& minute,/* specified as a number (0-59), i.e. 34 */
		int& second,/* specified as a number (0-59), i.e. 23 */
		int& ms		/* milli-second specified as a number (0-999), i.e. 240 */
		)
{
	auto& vm = exec->vm();

	DateInstance* jsDateInstance;

    if (jsValue.inherits(DateInstance::info()))
    	jsDateInstance = asDateInstance(jsValue);
    else {
        double value;
        JSValue primitive = jsValue.toPrimitive(exec);
        RETURN_IF_EXCEPTION(vm, false);
        if (primitive.isString()) {
            String primitiveString = asString(primitive)->value(exec);
            RETURN_IF_EXCEPTION(vm, false);
            value = parseDate(exec, vm, primitiveString);
            RETURN_IF_EXCEPTION(vm, false);
        } else
            value = primitive.toNumber(exec);
        RETURN_IF_EXCEPTION(vm, false);

        jsDateInstance = DateInstance::create(vm, exec->lexicalGlobalObject()->dateStructure(), value);
    }

    if (const GregorianDateTime* gregorianDateTime = jsDateInstance->gregorianDateTime(vm)) {
		year    = gregorianDateTime->year();
		month   = gregorianDateTime->month();
		day     = gregorianDateTime->monthDay();
		hour    = gregorianDateTime->hour();
		minute  = gregorianDateTime->minute();
		second  = gregorianDateTime->second();
		ms      = 0;

		return true;
    }

    String stringValue = jsValue.toWTFString(exec);
    RETURN_IF_EXCEPTION(vm, false);
    if (!stringValue.impl())
    	return false;

    // format 20171227153000000
    auto* stringRep = stringValue.impl();
    if (stringRep->length() == 14 || stringRep->length() == 17) {
    	bool ok = true;

        auto toInt32Strict = [&] (StringImpl& stringImpl, int pos, int len) -> int {
        	if (!ok)
        		return 0;

        	unsigned result = 0;

        	for (int i = 0; i < len; ++i) {
        		UChar ch = stringImpl[pos + i];

        		if (ch >= '0' && ch <= '9') {
        			result = result * 10 + (ch - '0');
        			continue;
        		}

        		ok = false;
    			return 0;
        	}

        	return result;
        };

		year    = toInt32Strict(*stringRep, 0, 4);
		month   = toInt32Strict(*stringRep, 4, 2);
		day     = toInt32Strict(*stringRep, 6, 2);
		hour    = toInt32Strict(*stringRep, 8, 2);
		minute  = toInt32Strict(*stringRep, 10, 2);
		second  = toInt32Strict(*stringRep, 12, 2);
		ms      = stringRep->length() == 17 ? toInt32Strict(*stringRep, 14, 3) : 0;

		return ok;
    }

	return false;
}

DLL_PUBLIC JSValue newJSNaN()
{
	return jsNaN();
}

DLL_PUBLIC double Inf()
{
	return std::numeric_limits<double>::infinity();
}

DLL_PUBLIC JSValue toPrimitive(ExecState* exec, JSValue v)
{
	return v.toPrimitive(exec);
}

DLL_PUBLIC JSValue toJS(ExecState* exec, WebCore::Frame* frame, RefPtr<WebCore::HTMLImageElement> imageElement)
{
	if(!imageElement || !frame)
        return jsUndefined();

	WebCore::JSDOMWindow* globalObject = frame->script()->globalObject(WebCore::mainThreadNormalWorld());
    if (!globalObject)
        return jsUndefined();

	return WebCore::toJS(exec, globalObject, imageElement.get());
}

DLL_PUBLIC JSValue toJS(ExecState* exec, WebCore::Frame* frame, PassRefPtr<WebCore::File> file)
{
	if(!file || !frame)
        return jsUndefined();

	WebCore::JSDOMWindow* globalObject = frame->script()->globalObject(WebCore::mainThreadNormalWorld());
    if (!globalObject)
        return jsUndefined();

	return WebCore::toJS(exec, globalObject, file.get());
}

DLL_PUBLIC JSC::JSValue toJS(JSC::ExecState* exec , WebCore::DOMWindow* domWindow)
{
	return WebCore::toJS(exec, domWindow);
}

DLL_PUBLIC bool toBoolean(ExecState* exec, JSValue v)
{
	return v.toBoolean(exec);
}

DLL_PUBLIC int32_t toInt32(ExecState* exec, JSValue v)
{
	return v.toInt32(exec);
}

DLL_PUBLIC uint32_t toUInt32(ExecState* exec, JSValue v)
{
	return v.toUInt32(exec);
}

DLL_PUBLIC double toIntegerOrInfinity(ExecState* exec, JSValue v)
{
	return v.toIntegerOrInfinity(exec);
}

DLL_PUBLIC double toNumber(ExecState* exec, JSValue v)
{
	return v.toNumber(exec);
}

DLL_PUBLIC JSString* newJSString(ExecState* exec, const char* s)
{
	return jsString(exec->vm(), String::fromLatin1(s));
}

DLL_PUBLIC JSValue newJSString(ExecState* exec, const WTF::String& s)
{
	return jsString(exec->vm(), s);
}

DLL_PUBLIC JSValue newJSSingleCharacterString(ExecState* exec, UChar c)
{
	return jsSingleCharacterString(exec->vm(), c);
}

DLL_PUBLIC JSString* newJSString(VM& vm, const char* s)
{
	return jsString(vm, String::fromLatin1(s));
}

DLL_PUBLIC JSValue newJSString(VM& vm, const WTF::String& s)
{
	return jsString(vm, s);
}

DLL_PUBLIC JSValue newJSSingleCharacterString(VM& vm, UChar c)
{
	return jsSingleCharacterString(vm, c);
}

DLL_PUBLIC void append(JSC::MarkedArgumentBuffer& args, JSC::JSValue v)
{
	args.append(v);
}

DLL_PUBLIC WebCore::DOMWrapperWorld* mainThreadNormalWorld()
{
	return WebCore::mainThreadNormalWorld();
}

DLL_PUBLIC WebCore::JSDOMWindow* globalObject(WebCore::Frame* frame)
{
	return frame->script()->globalObject(WebCore::mainThreadNormalWorld());
}

DLL_PUBLIC JSC::Bindings::RootObject* bindingRootObject(WebCore::Frame* frame)
{
	return frame->script()->bindingRootObject();
}

#if ENABLE(MOTTO_TV)
DLL_PUBLIC Vector<String> getPropertyNames(JSC::ExecState* exec, JSC::JSValue v)
{
	VM& vm = exec->vm();

	Vector<String> names;
	JSObject* object = v.toObject(exec);

	if(object){
		JSPropertyNameIterator* it = JSPropertyNameIterator::create(exec, object);
		for(size_t i = 0; i < it->size(); ++i){
			names.append(it->get(vm, exec, object, i).toWTFString(exec));
		}
	}

	return names;
}
#endif

DLL_PUBLIC JSC::JSValue dispatchEvent(JSC::ExecState* exec, JSC::JSGlobalObject* globalObject, JSC::JSValue listener, const JSC::MarkedArgumentBuffer& args)
{
    auto& vm = exec->vm();
    JSLockHolder lock(vm);

	CallData callData = JSC::getCallData(vm, listener);

	if (callData.type != CallData::Type::None){
		NakedPtr<JSC::Exception> exception;
        VMEntryScope entryScope(vm, vm.entryScope ? vm.entryScope->globalObject() : globalObject);
        return WebCore::JSExecState::call(exec, listener, callData, exec->globalThisValue(), args, exception);
	}

	return JSC::JSValue();
}

/**
 * dispatch event to "obj.a.b" with optional arguments
 */
DLL_PUBLIC JSC::JSValue dispatchEvent(WebCore::Frame* frame, const Vector<String, 2>& objects, const MarkedArgumentBuffer& args)
{
	JSC::JSValue result;

	if(!frame)
		return JSC::JSValue();

	if(objects.size() == 0)
		return JSC::JSValue();

	Ref<WebCore::Frame> protectedFrame(*frame);
	WebCore::JSDOMWindow* globalObject = frame->script()->globalObject(WebCore::mainThreadNormalWorld());

	if(!globalObject)
		return JSC::JSValue();

	ExecState* exec = globalObject->globalExec();
	VM& vm = globalObject->vm();

	JSValue value = globalObject->get(vm, exec, Identifier::fromString(vm, objects[0]));
	if(value.isUndefinedOrNull())
		goto OUT;

	for(int i = 1; i < objects.size(); ++i){
		JSValue property = value.get(vm, exec, Identifier::fromString(vm, objects[i]));
		if(property.isUndefinedOrNull())
			goto OUT;

		value = property;
	}

	result = dispatchEvent(exec, globalObject, value, args);

OUT:
	vm.clearException();

	return result;
}

DLL_PUBLIC Exception* throwException(ExecState* exec, JSObject* error)
{
	return exec->globalData().throwException(exec, error);
}

DLL_PUBLIC JSValue throwException(ExecState* exec, JSValue error)
{
	return exec->globalData().throwException(exec, error);
}

} //namespace WebCoreAPIs


namespace WebCore
{

DLL_PUBLIC Document* getDocument(JSC::ExecState* exec, JSC::JSGlobalObject* globalObject)
{
	VM& vm = globalObject->vm();
	return toDocument(globalObject->get(vm, exec, vm.propertyNames->document));
}

DLL_PUBLIC String toStringWithNullCheck(JSC::ExecState* exec, JSC::JSValue value)
{
    if (value.isNull())
        return String();
    return value.toWTFString(exec);
}

DLL_PUBLIC JSC::JSValue getProperty(JSC::JSObject* object, JSC::ExecState* exec, const String& propertyName)
{
	VM& vm = exec->vm();
	return object->get(vm, exec, JSC::Identifier::fromString(vm, propertyName));
}

} //namespace WebCore

