/*
 * Copyright (C) 2006 Rob Buis <buis@kde.org>
 *           (C) 2008 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "CSSCursorImageValue.h"

#include "CSSImageSetValue.h"
#include "CSSImageValue.h"
#include "TreeScope.h"
#include "PlatformString.h"
#include "StyleBuilderState.h"
#include <wtf/MathExtras.h>
#include <wtf/UnusedParam.h>
#include <wtf/text/StringBuilder.h>

#if ENABLE(SVG)
#include "SVGCursorElement.h"
#include "SVGLengthContext.h"
#include "SVGNames.h"
#include "SVGURIReference.h"
#endif

namespace WebCore {

CSSCursorImageValue::CSSCursorImageValue(Ref<CSSValue>&& imageValue, const Optional<IntPoint>& hotSpot)
    : CSSValue(CursorImageClass)
    , m_imageValue(WTFMove(imageValue))
    , m_hotSpot(hotSpot)
{
    if (m_imageValue->isImageValue())
        m_originalURL = static_cast<CSSImageValue&>(m_imageValue.get()).imageURL();
}

Ref<CSSCursorImageValue> CSSCursorImageValue::create(Ref<CSSValue>&& imageValue, const Optional<IntPoint>& hotSpot)
{
    return adoptRef(*new CSSCursorImageValue(WTFMove(imageValue), hotSpot));
}

CSSCursorImageValue::~CSSCursorImageValue()
{
    for (auto* element : m_cursorElements)
        element->removeClient(*this);
}

String CSSCursorImageValue::customCSSText() const
{
    StringBuilder result;
    result.append(m_imageValue->cssText());
    if (m_hotSpot) {
        result.append(' ');
        result.appendNumber(m_hotSpot->x());
        result.append(' ');
        result.appendNumber(m_hotSpot->y());
    }
    return result.toString();
}

// FIXME: Should this function take a TreeScope instead?
SVGCursorElement* CSSCursorImageValue::updateCursorElement(const Document& document)
{
    RefPtr<Element> element = SVGURIReference::targetElementFromIRIString(m_originalURL, document).element;
    if (!element || !isSVGCursorElement(*element))
        return nullptr;

    // FIXME: Not right to keep old cursor elements as clients. The new one should replace the old, not join it in a set.
	auto& cursorElement = toSVGCursorElement(*element);
    if (m_cursorElements.add(&cursorElement).isNewEntry) {
        cursorElementChanged(cursorElement);
        cursorElement.addClient(*this);
    }
    return &cursorElement;
}

void CSSCursorImageValue::cursorElementRemoved(SVGCursorElement& cursorElement)
{
	// FIXME: Not right to stay a client of a cursor element until the element is destroyed. We'd want to stop being a client once it's no longer a valid target, like when it's disconnected.
	m_cursorElements.remove(&cursorElement);
}

void CSSCursorImageValue::cursorElementChanged(SVGCursorElement& cursorElement)
{
    // FIXME: Seems wrong that changing an old cursor element, one that that is no longer the target, changes the hot spot.
    // FIXME: This will override a hot spot that was specified in CSS, which is probably incorrect.
    // FIXME: Should we clamp from float to int instead of just casting here?
    SVGLengthContext lengthContext(nullptr);
    m_hotSpot = IntPoint(
        static_cast<int>(std::round(cursorElement.x().value(lengthContext))),
        static_cast<int>(std::round(cursorElement.y().value(lengthContext)))
    );

    // FIXME: Why doesn't this funtion check for a change to the href of the cursor element? Why would we dynamically track changes to x/y but not href?
}

ImageWithScale CSSCursorImageValue::selectBestFitImage(const Document& document)
{
    if (m_imageValue->isImageSetValue())
        return static_cast<CSSImageSetValue&>(m_imageValue.get()).selectBestFitImage(document);

	if (auto* cursorElement = updateCursorElement(document)) {
        auto location = document.completeURL(cursorElement->href());
        if (location != static_cast<CSSImageValue&>(m_imageValue.get()).imageURL())
            m_imageValue = CSSImageValue::create(location);
    }

    return ImageWithScale { m_imageValue.ptr() , 1 };
}

bool CSSCursorImageValue::equals(const CSSCursorImageValue& other) const
{
    return m_hotSpot == other.m_hotSpot && compareCSSValue(m_imageValue, other.m_imageValue);
}

Ref<CSSCursorImageValue> CSSCursorImageValue::valueWithStylesResolved(Style::BuilderState& state)
{
    auto imageValue = state.resolveImageStyles(m_imageValue);
    if (imageValue.ptr() == m_imageValue.ptr())
        return *this;
    return create(WTFMove(imageValue), m_hotSpot);
}

} // namespace WebCore
