/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 2004-2005 Allan Sandfeld Jensen (kde@carewolf.com)
 * Copyright (C) 2006, 2007 Nicholas Shanks (webkit@nickshanks.com)
 * Copyright (C) 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Alexey Proskuryakov <ap@webkit.org>
 * Copyright (C) 2007, 2008 Eric Seidel <eric@webkit.org>
 * Copyright (C) 2008, 2009 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)
 * Copyright (c) 2011, Code Aurora Forum. All rights reserved.
 * Copyright (C) Research In Motion Limited 2011. All rights reserved.
 * Copyright (C) 2012 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "CSSDefaultStyleSheets.h"

#include "Chrome.h"
#include "ChromeClient.h"
#include "HTMLEmbedElement.h"
#include "HTMLObjectElement.h"
#include "MediaQueryEvaluator.h"
#include "Page.h"
#include "RenderTheme.h"
#include "RuleSet.h"
#include "StyleSheetContents.h"
#include "StyleResolver.h"
#include "StyleTreeResolver.h"
#include "UserAgentStyleSheets.h"
#include <wtf/NeverDestroyed.h>

namespace WebCore {

using namespace HTMLNames;

RefPtr<RuleSet> CSSDefaultStyleSheets::defaultStyle;
RefPtr<RuleSet> CSSDefaultStyleSheets::defaultQuirksStyle;
RefPtr<RuleSet> CSSDefaultStyleSheets::defaultPrintStyle;
unsigned CSSDefaultStyleSheets::defaultStyleVersion = 0;

//StyleSheetContents* CSSDefaultStyleSheets::simpleDefaultStyleSheet;
RefPtr<StyleSheetContents> CSSDefaultStyleSheets::defaultStyleSheet;
RefPtr<StyleSheetContents> CSSDefaultStyleSheets::quirksStyleSheet;
RefPtr<StyleSheetContents> CSSDefaultStyleSheets::svgStyleSheet;
RefPtr<StyleSheetContents> CSSDefaultStyleSheets::mathMLStyleSheet;
RefPtr<StyleSheetContents> CSSDefaultStyleSheets::mediaControlsStyleSheet;
RefPtr<StyleSheetContents> CSSDefaultStyleSheets::fullscreenStyleSheet;
//StyleSheetContents* CSSDefaultStyleSheets::plugInsStyleSheet;
RefPtr<StyleSheetContents> CSSDefaultStyleSheets::mediaQueryStyleSheet;

DLL_PUBLIC void clearDefaultCSSStyles()
{
	CSSDefaultStyleSheets::defaultStyleSheet.clear();
	CSSDefaultStyleSheets::quirksStyleSheet.clear();
	CSSDefaultStyleSheets::svgStyleSheet.clear();
	CSSDefaultStyleSheets::mathMLStyleSheet.clear();
	CSSDefaultStyleSheets::mediaControlsStyleSheet.clear();
	CSSDefaultStyleSheets::fullscreenStyleSheet.clear();
	CSSDefaultStyleSheets::mediaQueryStyleSheet.clear();

	CSSDefaultStyleSheets::defaultStyle.clear();
	CSSDefaultStyleSheets::defaultQuirksStyle.clear();
	CSSDefaultStyleSheets::defaultPrintStyle.clear();
}

// FIXME: It would be nice to use some mechanism that guarantees this is in sync with the real UA stylesheet.
static const char* simpleUserAgentStyleSheet = "html,body,div{display:block}head{display:none}body{margin:8px}div:focus,span:focus{outline:auto 5px -webkit-focus-ring-color}a:any-link{color:-webkit-link;text-decoration:underline}a:any-link:active{color:-webkit-activelink}";

static inline bool elementCanUseSimpleDefaultStyle(const Element& element)
{
    return element.hasTagName(htmlTag) || element.hasTagName(headTag)
    	|| element.hasTagName(bodyTag) || element.hasTagName(divTag)
		|| element.hasTagName(spanTag) || element.hasTagName(brTag)
		|| element.hasTagName(aTag);
}

static const MediaQueryEvaluator& screenEval()
{
    static NeverDestroyed<const MediaQueryEvaluator> staticScreenEval(String(MAKE_STATIC_STRING_IMPL("screen")));
    return staticScreenEval;
}

static const MediaQueryEvaluator& printEval()
{
    static NeverDestroyed<const MediaQueryEvaluator> staticPrintEval(String(MAKE_STATIC_STRING_IMPL("print")));
    return staticPrintEval;
}

static RefPtr<StyleSheetContents> parseUASheet(const String& str)
{
	RefPtr<StyleSheetContents> sheet = StyleSheetContents::create(CSSParserContext(UASheetMode));
    sheet->parseString(str);
    return sheet;
}

static RefPtr<StyleSheetContents> parseUASheet(const char* characters, unsigned size)
{
    return parseUASheet(String(characters, size));
}

void CSSDefaultStyleSheets::initDefaultStyle()
{
    if (!defaultStyle)
		loadFullDefaultStyle();
}

void CSSDefaultStyleSheets::addToDefaultStyle(StyleSheetContents& sheet)
{
    defaultStyle->addRulesFromSheet(sheet, screenEval());
    defaultPrintStyle->addRulesFromSheet(sheet, printEval());

    // Build a stylesheet consisting of non-trivial media queries seen in default style.
    // Rulesets for these can't be global and need to be built in document context.
    for (auto& rule : sheet.childRules()) {
        if (!(*rule).isMediaRule())
            continue;
        auto& mediaRule = static_cast<StyleRuleMedia&>(*rule);
        auto* mediaQuery = mediaRule.mediaQueries();
        if (!mediaQuery)
            continue;
        if (screenEval().evaluate(*mediaQuery, nullptr))
            continue;
        if (printEval().evaluate(*mediaQuery, nullptr))
            continue;
        mediaQueryStyleSheet->parserAppendRule(mediaRule.copy());
    }

    ++defaultStyleVersion;
}

void CSSDefaultStyleSheets::loadFullDefaultStyle()
{
	ASSERT(!defaultStyle);
	defaultStyle = RuleSet::create();
	defaultPrintStyle = RuleSet::create();
	mediaQueryStyleSheet = StyleSheetContents::create(CSSParserContext(UASheetMode));

	defaultQuirksStyle = RuleSet::create();

    // Strict-mode rules.
    String defaultRules = String(htmlUserAgentStyleSheet, sizeof(htmlUserAgentStyleSheet)) + RenderTheme::defaultTheme()->extraDefaultStyleSheet();
    defaultStyleSheet = parseUASheet(defaultRules);
    addToDefaultStyle(*defaultStyleSheet);

    // Quirks-mode rules.
    String quirksRules = String(quirksUserAgentStyleSheet, sizeof(quirksUserAgentStyleSheet)) + RenderTheme::defaultTheme()->extraQuirksStyleSheet();
    quirksStyleSheet = parseUASheet(quirksRules);
    defaultQuirksStyle->addRulesFromSheet(*quirksStyleSheet, screenEval());
}

#if 0
void CSSDefaultStyleSheets::loadSimpleDefaultStyle()
{
    ASSERT(!defaultStyle);
    ASSERT(!simpleDefaultStyleSheet);

    defaultStyle = RuleSet::create().leakPtr();
    // There are no media-specific rules in the simple default style.
    defaultPrintStyle = defaultStyle;
    defaultQuirksStyle = RuleSet::create().leakPtr();

    simpleDefaultStyleSheet = parseUASheet(simpleUserAgentStyleSheet, strlen(simpleUserAgentStyleSheet));
    defaultStyle->addRulesFromSheet(*simpleDefaultStyleSheet, screenEval());
    ++defaultStyleVersion;
    // No need to initialize quirks sheet yet as there are no quirk rules for elements allowed in simple default style.
}
#endif

void CSSDefaultStyleSheets::ensureDefaultStyleSheetsForElement(const Element& element)
{
#if 0
    if (simpleDefaultStyleSheet && !elementCanUseSimpleDefaultStyle(element)) {
        loadFullDefaultStyle();
        ++defaultStyleVersion;
    }
#endif

	if(element.isHTMLElement()) {
		auto& htmlElement = toHTMLElement(element);
        if (isHTMLObjectElement(htmlElement) || isHTMLEmbedElement(htmlElement)) {
#if 0
            if (!plugInsStyleSheet) {
                String plugInsRules = RenderTheme::themeForPage(element.document().page())->extraPlugInsStyleSheet() + element.document().page()->chrome().client().plugInExtraStyleSheet();
                if (plugInsRules.isEmpty())
                    plugInsRules = String(plugInsUserAgentStyleSheet, sizeof(plugInsUserAgentStyleSheet));
                plugInsStyleSheet = parseUASheet(plugInsRules);
                addToDefaultStyle(*plugInsStyleSheet);
            }
#endif
        }
#if ENABLE(VIDEO)
        else if (htmlElement.hasTagName(videoTag) || htmlElement.hasTagName(audioTag)) {
        	if(UNLIKELY(!mediaControlsStyleSheet)) {
				String mediaRules = String(mediaControlsUserAgentStyleSheet, sizeof(mediaControlsUserAgentStyleSheet)) + RenderTheme::themeForPage(element.document().page())->extraMediaControlsStyleSheet();
				mediaControlsStyleSheet = parseUASheet(mediaRules);
                addToDefaultStyle(*mediaControlsStyleSheet);
        	}
        }
#endif // ENABLE(VIDEO)
#if ENABLE(SERVICE_CONTROLS)
        else if (is<HTMLDivElement>(element) && element.isImageControlsRootElement()) {
            if (!imageControlsStyleSheet) {
                String imageControlsRules = RenderTheme::themeForPage(element.document().page())->imageControlsStyleSheet();
                imageControlsStyleSheet = parseUASheet(imageControlsRules);
                addToDefaultStyle(*imageControlsStyleSheet);
            }
        }
#endif // ENABLE(SERVICE_CONTROLS)
    }
#if ENABLE(SVG)
	else if (element.isSVGElement()) {
		if(UNLIKELY(!svgStyleSheet)) {
			// SVG rules.
			svgStyleSheet = parseUASheet(svgUserAgentStyleSheet, sizeof(svgUserAgentStyleSheet));
			addToDefaultStyle(*svgStyleSheet);
		}
    }
#endif // ENABLE(SVG)
#if ENABLE(MATHML)
	else if (element.isMathMLElement()) {
		if(UNLIKELY(!mathMLStyleSheet)) {
			// MathML rules.
			mathMLStyleSheet = parseUASheet(mathmlUserAgentStyleSheet, sizeof(mathmlUserAgentStyleSheet));
			addToDefaultStyle(*mathMLStyleSheet);
		}
    }
#endif // ENABLE(MATHML)

#if ENABLE(FULLSCREEN_API)
    if (UNLIKELY(!fullscreenStyleSheet) && element.document().webkitIsFullScreen()) {
        String fullscreenRules = String(fullscreenUserAgentStyleSheet, sizeof(fullscreenUserAgentStyleSheet)) + RenderTheme::defaultTheme()->extraFullScreenStyleSheet();
        fullscreenStyleSheet = parseUASheet(fullscreenRules);
        addToDefaultStyle(*fullscreenStyleSheet);
    }
#endif // ENABLE(FULLSCREEN_API)

    ASSERT(defaultStyle->features().idsInRules.isEmpty());
    ASSERT(mathMLStyleSheet || defaultStyle->features().siblingRules.isEmpty());
}

} // namespace WebCore
