/*
 * Copyright (C) 2007, 2008 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "FontFeatureSettings.h"
#include "FontLoadTimingOverride.h"
#include "FontRanges.h"
#include "FontFeatureSettings.h"
#include "FontSelectionAlgorithm.h"
#include "FontSelectionValueInlines.h"
#include "RenderStyleConstants.h"
#include "TextFlags.h"
#include "Timer.h"
#include <wtf/Forward.h>
#include <wtf/HashSet.h>
#include <wtf/Optional.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/Seconds.h>
#include <wtf/Vector.h>
#include <wtf/unicode/Unicode.h>
#include <wtf/WeakPtr.h>
#include <wtf/WeakHashSet.h>

namespace JSC {
class ExecState;
}

namespace WebCore {

class CSSFontFace;
class CSSFontFaceSet;
class CSSFontFaceSource;
class CSSFontSelector;
class CSSSegmentedFontFace;
class CSSValue;
class CSSValueList;
class Document;
class FontDescription;
class Font;
class FontFace;
class ScriptExecutionContext;
class Settings;
class StyleRuleFontFace;

enum class AllowUserInstalledFonts {
    No,
    Yes
};

class CSSFontFace FINAL : public RefCounted<CSSFontFace>, public WTF::CanMakeWeakPtr<CSSFontFace> {
public:
    static Ref<CSSFontFace> create(CSSFontSelector*, StyleRuleFontFace* cssConnection = nullptr, FontFace* wrapper = nullptr, bool isLocalFallback = false);
    virtual ~CSSFontFace();

    // FIXME: These functions don't need to have boolean return values.
    // Callers only call this with known-valid CSS values.
	bool setFamilies(CSSValue&);
	void setStyle(CSSValue&);
	void setWeight(CSSValue&);
    void setStretch(CSSValue&);
	bool setUnicodeRange(CSSValue&);
	bool setVariantLigatures(CSSValue&);
	bool setVariantPosition(CSSValue&);
	bool setVariantCaps(CSSValue&);
	bool setVariantNumeric(CSSValue&);
	bool setVariantAlternates(CSSValue&);
	bool setVariantEastAsian(CSSValue&);
	void setFeatureSettings(CSSValue&);
    void setLoadingBehavior(CSSValue&);

    // Pending => Loading  => TimedOut
    //              ||  \\    //  ||
    //              ||   \\  //   ||
    //              ||    \\//    ||
    //              ||     //     ||
    //              ||    //\\    ||
    //              ||   //  \\   ||
    //              \/  \/    \/  \/
    //             Success    Failure
    enum class Status : uint8_t { Pending, Loading, TimedOut, Success, Failure };

    struct UnicodeRange;

    // Optional return values to represent default string for members of FontFace.h
    Optional<CSSValueList*> families() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<CSSValueList*>(m_families.get()); }
    Optional<FontSelectionRange> weight() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<FontSelectionRange>(m_fontSelectionCapabilities.weight); }
    Optional<FontSelectionRange> stretch() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<FontSelectionRange>(m_fontSelectionCapabilities.width); }
    Optional<FontSelectionRange> italic() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<FontSelectionRange>(m_fontSelectionCapabilities.slope); }
    Optional<FontSelectionCapabilities> fontSelectionCapabilities() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<FontSelectionCapabilities>(m_fontSelectionCapabilities); }
    Optional<Vector<UnicodeRange>> ranges() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<Vector<UnicodeRange>>(m_ranges); }
    Optional<FontFeatureSettings> featureSettings() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<FontFeatureSettings>(m_featureSettings); }
    Optional<FontVariantSettings> variantSettings() const { return m_status == Status::Failure ? WTF::Nullopt : Optional<FontVariantSettings>(m_variantSettings); }
	Optional<FontLoadingBehavior> loadingBehavior() const { return m_status == Status::Failure ? WTF::Nullopt :  Optional<FontLoadingBehavior>(m_loadingBehavior); }
    void setVariantSettings(const FontVariantSettings& variantSettings) { m_variantSettings = variantSettings; }
	void setWeight(FontSelectionRange weight) { m_fontSelectionCapabilities.weight = weight; }
	void setStretch(FontSelectionRange stretch) { m_fontSelectionCapabilities.width = stretch; }
	void setStyle(FontSelectionRange italic) { m_fontSelectionCapabilities.slope = italic; }
	void setFontSelectionCapabilities(FontSelectionCapabilities capabilities) { m_fontSelectionCapabilities = capabilities; }
	bool isLocalFallback() const { return m_isLocalFallback; }
    Status status() const { return m_status; }
    StyleRuleFontFace* cssConnection() const { return m_cssConnection.get(); }

    class Client;
    void addClient(Client&);
    void removeClient(Client&);

    bool computeFailureState() const;

    void adoptSource(PassOwnPtr<CSSFontFaceSource>);
    void sourcesPopulated() { m_sourcesPopulated = true; }

    void fontLoaded(CSSFontFaceSource*);

    void load();

    PassRefPtr<Font> font(const FontDescription&, bool syntheticBold, bool syntheticItalic, ExternalResourceDownloadPolicy);

    static void appendSources(CSSFontFace&, CSSValueList&, Document*, bool isInitiatingElementInUserAgentShadowTree);

    class Client : public CanMakeWeakPtr<Client> {
	public:
		virtual ~Client() = default;
		virtual void fontLoaded(CSSFontFace&) { }
		virtual void fontStateChanged(CSSFontFace&, Status /*oldState*/, Status /*newState*/) { }
		virtual void fontPropertyChanged(CSSFontFace&, CSSValueList* /*oldFamilies*/ = nullptr) { }
		virtual void updateStyleIfNeeded(CSSFontFace&) { }
		virtual void ref() = 0;
		virtual void deref() = 0;
    };

    struct UnicodeRange {
        UnicodeRange(UChar32 from, UChar32 to) : m_from(from), m_to(to) { }
        UChar32 from() const { return m_from; }
        UChar32 to() const { return m_to; }

        bool operator==(const UnicodeRange& other) const { return m_from == other.m_from && m_to == other.m_to; }
        bool operator!=(const UnicodeRange& other) const { return !(*this == other); }
    private:
        UChar32 m_from;
        UChar32 m_to;
    };

    bool rangesMatchCodePoint(UChar32) const;

    // We don't guarantee that the FontFace wrapper will be the same every time you ask for it.
    PassRefPtr<FontFace> wrapper(ScriptExecutionContext*);
    void setWrapper(FontFace&);
    FontFace* existingWrapper() { return m_wrapper.get(); }

    struct FontLoadTiming {
        Seconds blockPeriod;
        Seconds swapPeriod;
    };
    FontLoadTiming fontLoadTiming() const;
    bool shouldIgnoreFontLoadCompletions() const { return m_shouldIgnoreFontLoadCompletions; }

    bool purgeable() const;

    AllowUserInstalledFonts allowUserInstalledFonts() const { return m_allowUserInstalledFonts; }

    void updateStyleIfNeeded();

#if ENABLE(SVG_FONTS)
    bool hasSVGFontFaceSource() const;
#endif
    void setErrorState();

    Document* document() const;

private:
    CSSFontFace(const Settings*, StyleRuleFontFace*, FontFace*, bool isLocalFallback);

    size_t pump(ExternalResourceDownloadPolicy);
    void setStatus(Status);
    void notifyClientsOfFontPropertyChange(CSSValueList* oldFamilies = 0);

    void initializeWrapper();

    void fontLoadEventOccurred();
    void timeoutFired();

    RefPtr<CSSValueList> m_families;
    Vector<UnicodeRange> m_ranges;
    FontFeatureSettings m_featureSettings;
    FontVariantSettings m_variantSettings;
    FontLoadingBehavior m_loadingBehavior { FontLoadingBehavior::Auto };

    Vector<OwnPtr<CSSFontFaceSource>, 0, CrashOnOverflow, 0> m_sources;
    RefPtr<StyleRuleFontFace> m_cssConnection;
	WeakHashSet<Client> m_clients;
    WeakPtr<FontFace> m_wrapper;
    FontSelectionCapabilities m_fontSelectionCapabilities;

    Status m_status { Status::Pending };
    bool m_isLocalFallback { false };
    bool m_sourcesPopulated { false };
    bool m_mayBePurged { true };
    bool m_shouldIgnoreFontLoadCompletions { false };
    FontLoadTimingOverride m_fontLoadTimingOverride { FontLoadTimingOverride::None };
    AllowUserInstalledFonts m_allowUserInstalledFonts { AllowUserInstalledFonts::Yes };

    Timer<CSSFontFace> m_timeoutTimer;
};

}

#include "FontFace.h"
