/*
 * Copyright (C) 2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "CSSFontFaceSet.h"

#include "FontFamily.h"
#include "CSSFontFaceSource.h"
#include "CSSFontSelector.h"
#include "CSSFontStyleValue.h"
#include "CSSParser.h"
#include "CSSPrimitiveValue.h"
#include "CSSSegmentedFontFace.h"
#include "CSSValueList.h"
#include "CSSValuePool.h"
#include "FontCache.h"
#include "FontFace.h"
#include "StyleBuilderConverter.h"
#include "StylePropertySet.h"

#if ENABLE(GLOBAL_FASTMALLOC_NEW_LOG)
ALWAYS_INLINE void logGlobalNew(size_t size) { }
#endif

namespace WebCore {

int gCSSFontFaceSetCount;

CSSFontFaceSet::CSSFontFaceSet(CSSFontSelector* owningFontSelector)
    : m_owningFontSelector(makeWeakPtr(owningFontSelector))
	, m_facesPartitionIndex(0)
	, m_status(Status::Loaded)
	, m_activeCount(0)
{
	++gCSSFontFaceSetCount;
}

CSSFontFaceSet::~CSSFontFaceSet()
{
    for (auto& face : m_faces)
		face->removeClient(*this);

    for (auto& pair : m_locallyInstalledFacesLookupTable) {
        for (auto& face : pair.second)
			face->removeClient(*this);
    }

	--gCSSFontFaceSetCount;
}

void CSSFontFaceSet::addFontModifiedObserver(const CSSFontSelector& fontModifiedObserver)
{
    auto result = m_fontModifiedObservers.add(fontModifiedObserver);
    ASSERT_UNUSED(result, result.isNewEntry);
}

void CSSFontFaceSet::addFontEventClient(const FontEventClient& fontEventClient)
{
    auto result = m_fontEventClients.add(fontEventClient);
    ASSERT_UNUSED(result, result.isNewEntry);
}

void CSSFontFaceSet::incrementActiveCount()
{
    ++m_activeCount;
    if (m_activeCount == 1) {
        m_status = Status::Loading;
        for (auto& client : copyToVector(m_fontEventClients)) {
        	if (client)
        		client->startedLoading();
        }
    }
}

void CSSFontFaceSet::decrementActiveCount()
{
    --m_activeCount;
    if (!m_activeCount) {
        m_status = Status::Loaded;
        for (auto& client : copyToVector(m_fontEventClients)) {
        	if (client)
        		client->completedLoading();
        }
    }
}

bool CSSFontFaceSet::hasFace(const CSSFontFace& face) const
{
    for (auto& myFace : m_faces) {
        if (myFace.ptr() == &face)
            return true;
    }

    return false;
}

// Calling updateStyleIfNeeded() might delete |this|.
void CSSFontFaceSet::updateStyleIfNeeded()
{
    if (m_owningFontSelector)
        m_owningFontSelector->updateStyleIfNeeded();
}

void CSSFontFaceSet::ensureLocalFontFacesForFamilyRegistered(const String& familyName)
{
    ASSERT(m_owningFontSelector);
    if(m_locallyInstalledFacesLookupTable.contains(familyName))
    	return;

    Vector<FontSelectionCapabilities> capabilities = fontCache().getFontSelectionCapabilitiesInFamily(familyName);
    if (capabilities.isEmpty())
        return;

    Vector<Ref<CSSFontFace>> faces;
    for (auto item : capabilities) {
        Ref<CSSFontFace> face = CSSFontFace::create(m_owningFontSelector.get(), nullptr, nullptr, true);

        auto familyList = CSSValueList::createCommaSeparated();
        familyList->append(CSSValuePool::singleton().createFontFamilyValue(familyName));
        face->setFamilies(familyList);
        face->setFontSelectionCapabilities(item);
        face->adoptSource(adoptPtr(new CSSFontFaceSource(face, familyName)));
        ASSERT(!face->computeFailureState());
        faces.append(WTFMove(face));
    }
    m_locallyInstalledFacesLookupTable.add(familyName, WTFMove(faces));
}

String CSSFontFaceSet::familyNameFromPrimitive(const CSSPrimitiveValue& value)
{
    if (value.isFontFamily())
        return value.fontFamily().familyName;
    if (!value.isValueID())
        return String();

    // We need to use the raw text for all the generic family types, since @font-face is a way of actually
    // defining what font to use for those types.
    switch (value.getValueID()) {
    case CSSValueSerif:
        return serifFamily;
    case CSSValueSansSerif:
        return sansSerifFamily;
    case CSSValueCursive:
        return cursiveFamily;
    case CSSValueFantasy:
        return fantasyFamily;
    case CSSValueMonospace:
        return monospaceFamily;
    case CSSValueWebkitPictograph:
        return pictographFamily;
    default:
        return String();
    }
}

void CSSFontFaceSet::addToFacesLookupTable(CSSFontFace& face)
{
    if (!face.families()) {
        // If the font has failed, there's no point in actually adding it to m_facesLookupTable,
        // because no font requests can actually use it for anything. So, let's just ... not add it.
        return;
    }
    auto families = face.families().value();

    for (auto& item : *families) {
        String familyName = CSSFontFaceSet::familyNameFromPrimitive(toCSSPrimitiveValue(item.get()));
        if (familyName.isEmpty())
            continue;

        auto addResult = m_facesLookupTable.add(familyName, Vector<Ref<CSSFontFace>>());
        auto& familyFontFaces = addResult.iterator->second;
        if (addResult.isNewEntry) {
            // m_locallyInstalledFontFaces grows without bound, eventually encorporating every font installed on the system.
            // This is by design.
            if (m_owningFontSelector)
            	ensureLocalFontFacesForFamilyRegistered(familyName);
            familyFontFaces = Vector<Ref<CSSFontFace>>();
        }

        familyFontFaces.append(face);
    }
}

void CSSFontFaceSet::add(CSSFontFace& face)
{
    ASSERT(!hasFace(face));

    for (auto& client : copyToVector(m_fontModifiedObservers)) {
    	if (client)
    		client->fontModified();
    }

    face.addClient(*this);
    m_cache.clear();

    if (face.cssConnection())
        m_faces.insert(m_facesPartitionIndex++, face);
    else
        m_faces.append(face);

    addToFacesLookupTable(face);

    if (face.status() == CSSFontFace::Status::Loading || face.status() == CSSFontFace::Status::TimedOut)
        incrementActiveCount();

    if (face.cssConnection()) {
        ASSERT(!m_constituentCSSConnections.contains(face.cssConnection()));
        m_constituentCSSConnections.add(face.cssConnection(), makeWeakPtr(face));
    }
}

void CSSFontFaceSet::removeFromFacesLookupTable(const CSSFontFace& face, const CSSValueList& familiesToSearchFor)
{
    for (auto& item : familiesToSearchFor) {
        String familyName = CSSFontFaceSet::familyNameFromPrimitive(toCSSPrimitiveValue(item.get()));
        if (familyName.isEmpty())
            continue;

        auto iterator = m_facesLookupTable.find(familyName);
        if (iterator == m_facesLookupTable.end()) {
            // The font may have failed even before addToFacesLookupTable() was called on it,
            // which means we never added it (because there's no point in adding a failed font).
            // So, if it was never added, removing it is free! Woohoo!
            return;
        }
        bool found = false;
        for (size_t i = 0; i < iterator->second.size(); ++i) {
            if (iterator->second[i].ptr() == &face) {
                found = true;
                iterator->second.remove(i);
                break;
            }
        }
        ASSERT_UNUSED(found, found);
        if (!iterator->second.size())
            m_facesLookupTable.remove(iterator);
    }
}

void CSSFontFaceSet::remove(const CSSFontFace& face)
{
	Ref<CSSFontFace> protect(const_cast<CSSFontFace&>(face));

    m_cache.clear();

    for (auto& client : copyToVector(m_fontModifiedObservers)) {
    	if (client)
    		client->fontModified();
    }

    if (face.families() && face.families().hasValue())
        removeFromFacesLookupTable(face, *face.families().value());

    if (face.cssConnection()) {
        ASSERT(m_constituentCSSConnections.get(face.cssConnection()) == &face);
        m_constituentCSSConnections.remove(face.cssConnection());
    }

    for (size_t i = 0; i < m_faces.size(); ++i) {
        if (m_faces[i].ptr() == &face) {
            if (i < m_facesPartitionIndex)
                --m_facesPartitionIndex;
            m_faces[i]->removeClient(*this);
            m_faces.remove(i);
            if (face.status() == CSSFontFace::Status::Loading || face.status() == CSSFontFace::Status::TimedOut)
                decrementActiveCount();
            return;
        }
    }
    ASSERT_NOT_REACHED();
}

CSSFontFace* CSSFontFaceSet::lookupByCSSConnection(StyleRuleFontFace& target)
{
    return m_constituentCSSConnections.get(&target).get();
}

void CSSFontFaceSet::purge()
{
    Vector<Ref<CSSFontFace>> toRemove;
    for (auto& face : m_faces) {
        if (face->purgeable())
            toRemove.append(face);
    }

    for (auto& item : toRemove)
        remove(item.get());
}

void CSSFontFaceSet::emptyCaches()
{
    m_cache.clear();
}

void CSSFontFaceSet::clear()
{
    for (auto& face : m_faces)
        face->removeClient(*this);

    for (auto& pair : m_locallyInstalledFacesLookupTable) {
        for (auto& face : pair.second)
			face->removeClient(*this);
    }

    m_faces.clear();
    m_facesLookupTable.clear();
    m_locallyInstalledFacesLookupTable.clear();
    m_cache.clear();
    m_constituentCSSConnections.clear();
    m_facesPartitionIndex = 0;
    m_status = Status::Loaded;
}

CSSFontFace& CSSFontFaceSet::operator[](size_t i)
{
    ASSERT(i < faceCount());
    return m_faces[i];
}

static FontSelectionRequest computeFontSelectionRequest(MutableStyleProperties& style, ExceptionCode& ec)
{
    RefPtr<CSSValue> weightValue = style.getPropertyCSSValue(CSSPropertyFontWeight).get();
    if (!weightValue || weightValue->isInitialValue())
        weightValue = CSSValuePool::singleton().createIdentifierValue(CSSValueNormal);

    RefPtr<CSSValue> stretchValue = style.getPropertyCSSValue(CSSPropertyFontStretch).get();
    if (!stretchValue || stretchValue->isInitialValue())
        stretchValue = CSSValuePool::singleton().createIdentifierValue(CSSValueNormal);

    RefPtr<CSSValue> styleValue = style.getPropertyCSSValue(CSSPropertyFontStyle).get();
    if (!styleValue || styleValue->isInitialValue())
        styleValue = CSSFontStyleValue::create(CSSValuePool::singleton().createIdentifierValue(CSSValueNormal));

    if (weightValue->isCSSWideKeyword() || stretchValue->isCSSWideKeyword() || styleValue->isCSSWideKeyword()) {
    	ec = SYNTAX_ERR;
    	return FontSelectionRequest();
    }

    FontSelectionValue weightSelectionValue = StyleBuilderConverter::convertFontWeightFromValue(*weightValue);
    FontSelectionValue stretchSelectionValue = StyleBuilderConverter::convertFontStretchFromValue(*stretchValue);
    FontSelectionValue styleSelectionValue = StyleBuilderConverter::convertFontStyleFromValue(*styleValue);

    return FontSelectionRequest(weightSelectionValue, stretchSelectionValue, styleSelectionValue);
}

typedef HashSet<UChar32, DefaultHash<UChar32>::Hash, WTF::UnsignedWithZeroKeyHashTraits<UChar32>> CodePointsMap;
static CodePointsMap codePointsFromString(StringView stringView)
{
    CodePointsMap result;
    auto graphemeClusters = stringView.graphemeClusters();
    for(auto iter = graphemeClusters.begin(); iter != graphemeClusters.end(); ++iter) {
    	auto cluster = *iter;
        ASSERT(cluster.length() > 0);
        UChar32 character = 0;
        if (cluster.is8Bit())
            character = cluster[0];
        else
            U16_GET(cluster.characters16(), 0, 0, cluster.length(), character);
        result.add(character);
    }
    return result;
}

Vector<RefPtr<CSSFontFace>> CSSFontFaceSet::matchingFacesExcludingPreinstalledFonts(const String& font, const String& string, ExceptionCode& ec)
{
    RefPtr<MutableStylePropertySet> style = MutableStylePropertySet::create();
    auto parseResult = CSSParser::parseValue(style.get(), CSSPropertyFont, font, true, CSSParserContext(HTMLStandardMode), nullptr);
    if (parseResult == CSSParser::ParseResult::Error) {
        ec = SYNTAX_ERR;
        return Vector<RefPtr<CSSFontFace>>();
    }

    FontSelectionRequest request = computeFontSelectionRequest(*style, ec);
    if (ec)
        return Vector<RefPtr<CSSFontFace>>();

    RefPtr<CSSValue> family = style->getPropertyCSSValue(CSSPropertyFontFamily);
    if (!family->isValueList()) {
        ec = SYNTAX_ERR;
        return Vector<RefPtr<CSSFontFace>>();
    }
    CSSValueList& familyList = toCSSValueList(*family);

    HashSet<AtomicString> uniqueFamilies;
    Vector<AtomicString> familyOrder;
    for (auto& family : familyList) {
        auto& primitive = toCSSPrimitiveValue(family.get());
        if (!primitive.isFontFamily())
            continue;
        if (uniqueFamilies.add(primitive.fontFamily().familyName).isNewEntry)
            familyOrder.append(primitive.fontFamily().familyName);
    }

    HashSet<CSSFontFace*> resultConstituents;
    for (auto codePoint : codePointsFromString(string)) {
        bool found = false;
        for (auto& family : familyOrder) {
			CSSSegmentedFontFace* faces = getFontFace(request, family);
			if (!faces)
				continue;
            for (auto& constituentFace : faces->constituentFaces()) {
				if (constituentFace->isLocalFallback())
					continue;
                if (constituentFace->rangesMatchCodePoint(codePoint)) {
                    resultConstituents.add(constituentFace.get());
                    found = true;
                    break;
                }
			}
            if (found)
                break;
        }
    }

    Vector<RefPtr<CSSFontFace>> result;
    result.reserveInitialCapacity(resultConstituents.size());
    for (auto* constituent : resultConstituents)
        result.uncheckedAppend(*constituent);
    return result;
}

bool CSSFontFaceSet::check(const String& font, const String& text, ExceptionCode& ec)
{
    auto matchingFaces = this->matchingFacesExcludingPreinstalledFonts(font, text, ec);
    if (ec)
        return false;

    for(int i = 0; i < matchingFaces.size(); ++i) {
        if (matchingFaces[i]->status() == CSSFontFace::Status::Pending)
            return false;
    }
    return true;
}

CSSSegmentedFontFace* CSSFontFaceSet::getFontFace(FontSelectionRequest request, const AtomicString& family)
{
    auto iterator = m_facesLookupTable.find(family);
    if (iterator == m_facesLookupTable.end())
        return 0;
    auto& familyFontFaces = iterator->second;

    auto& segmentedFontFaceCache = m_cache.add(family, FontSelectionHashMap()).iterator->second;

    RefPtr<CSSSegmentedFontFace>& face = segmentedFontFaceCache.add(request, nullptr).iterator->second;
    if (face)
        return face.get();

    face = CSSSegmentedFontFace::create();

    Vector<Ref<CSSFontFace>, 32> candidateFontFaces;
    for (int i = familyFontFaces.size() - 1; i >= 0; --i) {
    	CSSFontFace& candidate = familyFontFaces[i];
        auto capabilitiesWrapped = candidate.fontSelectionCapabilities();
        if (!capabilitiesWrapped.hasValue())
            continue;
        auto capabilities = capabilitiesWrapped.value();
        if (!isItalic(request.slope) && isItalic(capabilities.slope.minimum))
            continue;
        candidateFontFaces.append(candidate);
    }

    auto localIterator = m_locallyInstalledFacesLookupTable.find(family);
    if (localIterator != m_locallyInstalledFacesLookupTable.end()) {
        for (auto& candidate : localIterator->second) {
            auto capabilitiesWrapped = candidate->fontSelectionCapabilities();
            if (!capabilitiesWrapped.hasValue())
                continue;
            auto capabilities = capabilitiesWrapped.value();
            if (!isItalic(request.slope) && isItalic(capabilities.slope.minimum))
                continue;
            candidateFontFaces.append(candidate);
        }
    }

    if (!candidateFontFaces.isEmpty()) {
        Vector<FontSelectionCapabilities> capabilities;
        capabilities.reserveInitialCapacity(candidateFontFaces.size());
        for (auto& face : candidateFontFaces) {
            auto fontSelectionCapabilitiesWrapped = face.get().fontSelectionCapabilities();
            ASSERT(fontSelectionCapabilitiesWrapped.hasValue());
            auto fontSelectionCapabilities = fontSelectionCapabilitiesWrapped.value();
            capabilities.uncheckedAppend(fontSelectionCapabilities);
        }
        FontSelectionAlgorithm fontSelectionAlgorithm(request, capabilities);
        class Functor
        {
        public:
        	Functor(FontSelectionAlgorithm& fontSelectionAlgorithm) : fontSelectionAlgorithm(fontSelectionAlgorithm) { }
        	bool operator()(const CSSFontFace& first, const CSSFontFace& second) {
        		auto firstCapabilitiesWrapped = first.fontSelectionCapabilities();
				auto secondCapabilitiesWrapped = second.fontSelectionCapabilities();
				ASSERT(firstCapabilitiesWrapped.hasValue() && secondCapabilitiesWrapped.hasValue());

				auto firstCapabilities = firstCapabilitiesWrapped.value();
				auto secondCapabilities = secondCapabilitiesWrapped.value();
				auto stretchDistanceFirst = fontSelectionAlgorithm.stretchDistance(firstCapabilities).distance;
				auto stretchDistanceSecond = fontSelectionAlgorithm.stretchDistance(secondCapabilities).distance;
				if (stretchDistanceFirst < stretchDistanceSecond)
					return true;
				if (stretchDistanceFirst > stretchDistanceSecond)
					return false;

				auto styleDistanceFirst = fontSelectionAlgorithm.styleDistance(firstCapabilities).distance;
				auto styleDistanceSecond = fontSelectionAlgorithm.styleDistance(secondCapabilities).distance;
				if (styleDistanceFirst < styleDistanceSecond)
					return true;
				if (styleDistanceFirst > styleDistanceSecond)
					return false;

				auto weightDistanceFirst = fontSelectionAlgorithm.weightDistance(firstCapabilities).distance;
				auto weightDistanceSecond = fontSelectionAlgorithm.weightDistance(secondCapabilities).distance;
				if (weightDistanceFirst < weightDistanceSecond)
					return true;
				return false;
			}
        private:
        	FontSelectionAlgorithm& fontSelectionAlgorithm;
        };
        Functor functor(fontSelectionAlgorithm);
        std::stable_sort(candidateFontFaces.begin(), candidateFontFaces.end(), functor);

        CSSFontFace* previousCandidate = nullptr;
        for (auto& candidate : candidateFontFaces) {
            if (&candidate.get() == previousCandidate)
                continue;
            previousCandidate = &candidate.get();
            face->appendFontFace(candidate.get());
        }
    }

    return face.get();
}

void CSSFontFaceSet::fontStateChanged(CSSFontFace& face, CSSFontFace::Status oldState, CSSFontFace::Status newState)
{
    ASSERT(hasFace(face));
    if (oldState == CSSFontFace::Status::Pending) {
        ASSERT(newState == CSSFontFace::Status::Loading);
        incrementActiveCount();
    }
    if (newState == CSSFontFace::Status::Success || newState == CSSFontFace::Status::Failure) {
        ASSERT(oldState == CSSFontFace::Status::Loading || oldState == CSSFontFace::Status::TimedOut);
        for (auto& client : copyToVector(m_fontEventClients)) {
        	if (client)
        		client->faceFinished(face, newState);
        }
        decrementActiveCount();
    }
}

void CSSFontFaceSet::fontPropertyChanged(CSSFontFace& face, CSSValueList* oldFamilies)
{
    m_cache.clear();

    if (oldFamilies) {
        removeFromFacesLookupTable(face, *oldFamilies);
        addToFacesLookupTable(face);
    }

    for (auto& client : copyToVector(m_fontModifiedObservers)) {
    	if (client)
    		client->fontModified();
    }
}

}
