/*
 * Copyright (C) 2011 Andreas Kling (kling@webkit.org)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"
#include "CSSValue.h"

#include "CSSAspectRatioValue.h"
#include "CSSBorderImageSliceValue.h"
#include "CSSCalculationValue.h"
#include "CSSCanvasValue.h"
#include "CSSContentDistributionValue.h"
#include "CSSCrossfadeValue.h"
#include "CSSCursorImageValue.h"
#include "CSSCustomPropertyValue.h"
#include "CSSFilterImageValue.h"
#include "CSSFontFaceSrcValue.h"
#include "CSSFontStyleRangeValue.h"
#include "CSSFontStyleValue.h"
#include "CSSFunctionValue.h"
#include "CSSGradientValue.h"
#include "CSSGridAutoRepeatValue.h"
#include "CSSGridIntegerRepeatValue.h"
#include "CSSGridLineNamesValue.h"
#include "CSSGridTemplateAreasValue.h"
#include "CSSImageGeneratorValue.h"
#include "CSSImageSetValue.h"
#include "CSSImageValue.h"
#include "CSSLineBoxContainValue.h"
#include "CSSNamedImageValue.h"
#include "CSSPendingSubstitutionValue.h"
#include "CSSPrimitiveValue.h"
#include "CSSProperty.h"
#include "CSSRayValue.h"
#include "CSSReflectValue.h"
#include "CSSTimingFunctionValue.h"
#include "CSSUnicodeRangeValue.h"
#include "CSSValueList.h"
#include "CSSVariableReferenceValue.h"
#include "CSSFontValue.h"
#include "CSSFontFeatureValue.h"
#include "CSSShadowValue.h"

#include "DeprecatedCSSOMPrimitiveValue.h"
#include "DeprecatedCSSOMValueList.h"
#include "SVGFontFaceElement.h"

namespace WebCore {

CSSValue::Type CSSValue::cssValueType() const
{
    if (isInheritValue())
        return CSS_INHERIT;
    if (isPrimitiveValue())
        return CSS_PRIMITIVE_VALUE;
    if (isValueList())
        return CSS_VALUE_LIST;
    if (isInitialValue())
        return CSS_INITIAL;
    if (isUnsetValue())
        return CSS_UNSET;
    if (isRevertValue())
        return CSS_REVERT;
    return CSS_CUSTOM;
}

void CSSValue::collectDirectComputationalDependencies(HashSet<CSSPropertyID>& values) const
{
    if ((*this).isPrimitiveValue())
        toCSSPrimitiveValue(*this).collectDirectComputationalDependencies(values);
}

void CSSValue::collectDirectRootComputationalDependencies(HashSet<CSSPropertyID>& values) const
{
    if ((*this).isPrimitiveValue())
        toCSSPrimitiveValue(*this).collectDirectRootComputationalDependencies(values);
}

template<class ChildClassType>
inline static bool compareCSSValues(const CSSValue& first, const CSSValue& second)
{
    return static_cast<const ChildClassType&>(first).equals(static_cast<const ChildClassType&>(second));
}

bool CSSValue::equals(const CSSValue& other) const
{
    if (m_classType == other.m_classType) {
        switch (m_classType) {
        case AspectRatioClass:
            return compareCSSValues<CSSAspectRatioValue>(*this, other);
        case BorderImageSliceClass:
            return compareCSSValues<CSSBorderImageSliceValue>(*this, other);
        case CanvasClass:
            return compareCSSValues<CSSCanvasValue>(*this, other);
        case NamedImageClass:
            return compareCSSValues<CSSNamedImageValue>(*this, other);
        case CursorImageClass:
            return compareCSSValues<CSSCursorImageValue>(*this, other);
#if ENABLE(CSS_FILTERS)
        case FilterImageClass:
            return compareCSSValues<CSSFilterImageValue>(*this, other);
#endif
        case FontClass:
            return compareCSSValues<CSSFontValue>(*this, other);
        case FontFaceSrcClass:
            return compareCSSValues<CSSFontFaceSrcValue>(*this, other);
        case FontFeatureClass:
            return compareCSSValues<CSSFontFeatureValue>(*this, other);
        case FunctionClass:
            return compareCSSValues<CSSFunctionValue>(*this, other);
        case LinearGradientClass:
            return compareCSSValues<CSSLinearGradientValue>(*this, other);
        case RadialGradientClass:
            return compareCSSValues<CSSRadialGradientValue>(*this, other);
        case ConicGradientClass:
            return compareCSSValues<CSSConicGradientValue>(*this, other);
        case CrossfadeClass:
            return compareCSSValues<CSSCrossfadeValue>(*this, other);
        case ImageClass:
            return compareCSSValues<CSSImageValue>(*this, other);
        case GridAutoRepeatClass:
            return compareCSSValues<CSSGridAutoRepeatValue>(*this, other);
        case GridIntegerRepeatClass:
            return compareCSSValues<CSSGridIntegerRepeatValue>(*this, other);
        case GridLineNamesClass:
            return compareCSSValues<CSSGridLineNamesValue>(*this, other);
        case GridTemplateAreasClass:
            return compareCSSValues<CSSGridTemplateAreasValue>(*this, other);
        case PrimitiveClass:
            return compareCSSValues<CSSPrimitiveValue>(*this, other);
        case ReflectClass:
            return compareCSSValues<CSSReflectValue>(*this, other);
        case ShadowClass:
            return compareCSSValues<CSSShadowValue>(*this, other);
        case LinearTimingFunctionClass:
            return compareCSSValues<CSSLinearTimingFunctionValue>(*this, other);
        case CubicBezierTimingFunctionClass:
            return compareCSSValues<CSSCubicBezierTimingFunctionValue>(*this, other);
        case StepsTimingFunctionClass:
            return compareCSSValues<CSSStepsTimingFunctionValue>(*this, other);
        case SpringTimingFunctionClass:
            return compareCSSValues<CSSSpringTimingFunctionValue>(*this, other);
        case UnicodeRangeClass:
            return compareCSSValues<CSSUnicodeRangeValue>(*this, other);
        case ValueListClass:
            return compareCSSValues<CSSValueList>(*this, other);
        case LineBoxContainClass:
            return compareCSSValues<CSSLineBoxContainValue>(*this, other);
        case CalculationClass:
            return compareCSSValues<CSSCalcValue>(*this, other);
        case ImageSetClass:
            return compareCSSValues<CSSImageSetValue>(*this, other);
        case CSSContentDistributionClass:
            return compareCSSValues<CSSContentDistributionValue>(*this, other);
        case CustomPropertyClass:
            return compareCSSValues<CSSCustomPropertyValue>(*this, other);
        case VariableReferenceClass:
            return compareCSSValues<CSSVariableReferenceValue>(*this, other);
        case PendingSubstitutionValueClass:
            return compareCSSValues<CSSPendingSubstitutionValue>(*this, other);
        case RayClass:
            return compareCSSValues<CSSRayValue>(*this, other);
        case FontStyleClass:
            return compareCSSValues<CSSFontStyleValue>(*this, other);
        case FontStyleRangeClass:
            return compareCSSValues<CSSFontStyleRangeValue>(*this, other);
        default:
            ASSERT_NOT_REACHED();
            return false;
        }
    } else if (m_classType == ValueListClass && other.m_classType != ValueListClass)
        return static_cast<const CSSValueList*>(this)->equals(other);
    else if (m_classType != ValueListClass && other.m_classType == ValueListClass)
        return static_cast<const CSSValueList&>(other).equals(*this);
    return false;
}

bool CSSValue::isCSSLocalURL(StringView relativeURL)
{
    return relativeURL.isEmpty() || relativeURL.startsWith("#");
}

String CSSValue::cssText() const
{
    switch (classType()) {
    case AspectRatioClass:
        return static_cast<const CSSAspectRatioValue*>(this)->customCSSText();
    case BorderImageSliceClass:
        return static_cast<const CSSBorderImageSliceValue*>(this)->customCSSText();
    case CanvasClass:
        return static_cast<const CSSCanvasValue*>(this)->customCSSText();
    case NamedImageClass:
        return toCSSNamedImageValue(*this).customCSSText();
    case CursorImageClass:
        return static_cast<const CSSCursorImageValue*>(this)->customCSSText();
#if ENABLE(CSS_FILTERS)
    case FilterImageClass:
        return static_cast<const CSSFilterImageValue*>(this)->customCSSText();
#endif
    case FontClass:
        return static_cast<const CSSFontValue*>(this)->customCSSText();
    case FontFaceSrcClass:
        return static_cast<const CSSFontFaceSrcValue*>(this)->customCSSText();
    case FontFeatureClass:
        return static_cast<const CSSFontFeatureValue*>(this)->customCSSText();
    case FunctionClass:
        return static_cast<const CSSFunctionValue*>(this)->customCSSText();
    case LinearGradientClass:
        return static_cast<const CSSLinearGradientValue*>(this)->customCSSText();
    case RadialGradientClass:
        return static_cast<const CSSRadialGradientValue*>(this)->customCSSText();
    case ConicGradientClass:
        return static_cast<const CSSConicGradientValue&>(*this).customCSSText();
    case CrossfadeClass:
        return static_cast<const CSSCrossfadeValue*>(this)->customCSSText();
    case ImageClass:
        return static_cast<const CSSImageValue*>(this)->customCSSText();
    case GridAutoRepeatClass:
        return static_cast<const CSSGridAutoRepeatValue*>(this)->customCSSText();
    case GridIntegerRepeatClass:
        return static_cast<const CSSGridIntegerRepeatValue&>(*this).customCSSText();
    case GridLineNamesClass:
        return static_cast<const CSSGridLineNamesValue*>(this)->customCSSText();
    case GridTemplateAreasClass:
        return static_cast<const CSSGridTemplateAreasValue*>(this)->customCSSText();
    case PrimitiveClass:
        return static_cast<const CSSPrimitiveValue*>(this)->customCSSText();
    case ReflectClass:
        return static_cast<const CSSReflectValue*>(this)->customCSSText();
    case ShadowClass:
        return static_cast<const CSSShadowValue*>(this)->customCSSText();
    case LinearTimingFunctionClass:
        return static_cast<const CSSLinearTimingFunctionValue*>(this)->customCSSText();
    case CubicBezierTimingFunctionClass:
        return static_cast<const CSSCubicBezierTimingFunctionValue*>(this)->customCSSText();
    case StepsTimingFunctionClass:
        return static_cast<const CSSStepsTimingFunctionValue*>(this)->customCSSText();
    case SpringTimingFunctionClass:
        return static_cast<const CSSSpringTimingFunctionValue&>(*this).customCSSText();
    case UnicodeRangeClass:
        return static_cast<const CSSUnicodeRangeValue*>(this)->customCSSText();
    case ValueListClass:
        return static_cast<const CSSValueList*>(this)->customCSSText();
    case LineBoxContainClass:
        return static_cast<const CSSLineBoxContainValue*>(this)->customCSSText();
    case CalculationClass:
        return static_cast<const CSSCalcValue*>(this)->customCSSText();
    case ImageSetClass:
        return static_cast<const CSSImageSetValue*>(this)->customCSSText();
    case CSSContentDistributionClass:
        return static_cast<const CSSContentDistributionValue&>(*this).customCSSText();
    case CustomPropertyClass:
        return static_cast<const CSSCustomPropertyValue&>(*this).customCSSText();
    case VariableReferenceClass:
        return toCSSVariableReferenceValue(*this).customCSSText();
    case PendingSubstitutionValueClass:
        return toCSSPendingSubstitutionValue(*this).customCSSText();
    case RayClass:
        return toCSSRayValue(*this).customCSSText();
    case FontStyleClass:
        return toCSSFontStyleValue(*this).customCSSText();
    case FontStyleRangeClass:
        return toCSSFontStyleRangeValue(*this).customCSSText();
    }

    ASSERT_NOT_REACHED();
    return String();
}

void CSSValue::destroy()
{
    switch (classType()) {
    case AspectRatioClass:
        delete static_cast<CSSAspectRatioValue*>(this);
        return;
    case BorderImageSliceClass:
        delete static_cast<CSSBorderImageSliceValue*>(this);
        return;
    case CanvasClass:
        delete static_cast<CSSCanvasValue*>(this);
        return;
    case NamedImageClass:
        delete toCSSNamedImageValue(this);
        return;
    case CursorImageClass:
        delete static_cast<CSSCursorImageValue*>(this);
        return;
    case FontClass:
        delete static_cast<CSSFontValue*>(this);
        return;
    case FontFaceSrcClass:
        delete static_cast<CSSFontFaceSrcValue*>(this);
        return;
    case FontFeatureClass:
        delete static_cast<CSSFontFeatureValue*>(this);
        return;
    case FunctionClass:
        delete static_cast<CSSFunctionValue*>(this);
        return;
    case LinearGradientClass:
        delete static_cast<CSSLinearGradientValue*>(this);
        return;
    case RadialGradientClass:
        delete static_cast<CSSRadialGradientValue*>(this);
        return;
    case ConicGradientClass:
        delete static_cast<CSSConicGradientValue*>(this);
        return;
    case CrossfadeClass:
        delete static_cast<CSSCrossfadeValue*>(this);
        return;
    case ImageClass:
        delete static_cast<CSSImageValue*>(this);
        return;
    case GridAutoRepeatClass:
        delete static_cast<CSSGridAutoRepeatValue*>(this);
        return;
    case GridIntegerRepeatClass:
        delete static_cast<CSSGridIntegerRepeatValue*>(this);
        return;
    case GridLineNamesClass:
        delete static_cast<CSSGridLineNamesValue*>(this);
        return;
    case GridTemplateAreasClass:
        delete static_cast<CSSGridTemplateAreasValue*>(this);
        return;
    case PrimitiveClass:
        delete static_cast<CSSPrimitiveValue*>(this);
        return;
    case ReflectClass:
        delete static_cast<CSSReflectValue*>(this);
        return;
    case ShadowClass:
        delete static_cast<CSSShadowValue*>(this);
        return;
    case LinearTimingFunctionClass:
        delete static_cast<CSSLinearTimingFunctionValue*>(this);
        return;
    case CubicBezierTimingFunctionClass:
        delete static_cast<CSSCubicBezierTimingFunctionValue*>(this);
        return;
    case StepsTimingFunctionClass:
        delete static_cast<CSSStepsTimingFunctionValue*>(this);
        return;
    case SpringTimingFunctionClass:
        delete static_cast<CSSSpringTimingFunctionValue*>(this);
        return;
    case UnicodeRangeClass:
        delete static_cast<CSSUnicodeRangeValue*>(this);
        return;
    case ValueListClass:
        delete static_cast<CSSValueList*>(this);
        return;
    case LineBoxContainClass:
        delete static_cast<CSSLineBoxContainValue*>(this);
        return;
    case CalculationClass:
        delete static_cast<CSSCalcValue*>(this);
        return;
    case ImageSetClass:
        delete static_cast<CSSImageSetValue*>(this);
        return;
    case FilterImageClass:
        delete static_cast<CSSFilterImageValue*>(this);
        return;
    case CSSContentDistributionClass:
        delete static_cast<CSSContentDistributionValue*>(this);
        return;
    case CustomPropertyClass:
        delete static_cast<CSSCustomPropertyValue*>(this);
        return;
    case VariableReferenceClass:
        delete toCSSVariableReferenceValue(this);
        return;
    case PendingSubstitutionValueClass:
        delete toCSSPendingSubstitutionValue(this);
        return;
    case RayClass:
        delete toCSSRayValue(this);
        return;
    case FontStyleClass:
        delete toCSSFontStyleValue(this);
        return;
    case FontStyleRangeClass:
        delete toCSSFontStyleRangeValue(this);
        return;
    }
    ASSERT_NOT_REACHED();
}

Ref<DeprecatedCSSOMValue> CSSValue::createDeprecatedCSSOMWrapper(CSSStyleDeclaration& styleDeclaration) const
{
    if (isImageValue())
        return static_cast<const CSSImageValue*>(this)->createDeprecatedCSSOMWrapper(styleDeclaration);
    if (isPrimitiveValue())
        return DeprecatedCSSOMPrimitiveValue::create(toCSSPrimitiveValue(*this), styleDeclaration);
    if (isValueList())
        return DeprecatedCSSOMValueList::create(toCSSValueList(*this), styleDeclaration);
    return DeprecatedCSSOMComplexValue::create(*this, styleDeclaration);
}

bool CSSValue::treatAsInheritedValue(CSSPropertyID propertyID) const
{
    return isInheritValue() || (isUnsetValue() && CSSProperty::isInheritedProperty(propertyID));
}

bool CSSValue::treatAsInitialValue(CSSPropertyID propertyID) const
{
    return isInitialValue() || (isUnsetValue() && !CSSProperty::isInheritedProperty(propertyID));
}

bool CSSValue::isInitialValue() const
{
    return (*this).isPrimitiveValue() && toCSSPrimitiveValue(*this).isInitialValue();
}

bool CSSValue::isImplicitInitialValue() const
{
    return (*this).isPrimitiveValue() && toCSSPrimitiveValue(*this).isImplicitInitialValue();
}

bool CSSValue::isInheritValue() const
{
    return (*this).isPrimitiveValue() && toCSSPrimitiveValue(*this).isInheritValue();
}

bool CSSValue::isUnsetValue() const
{
    return (*this).isPrimitiveValue() && toCSSPrimitiveValue(*this).isUnsetValue();
}

bool CSSValue::isRevertValue() const
{
    return (*this).isPrimitiveValue() && toCSSPrimitiveValue(*this).isRevertValue();
}

bool CSSValue::isCSSWideKeyword() const
{
    return (*this).isPrimitiveValue() && toCSSPrimitiveValue(*this).isCSSWideKeyword();
}

}
