/*
 * (C) 1999-2003 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#include "CSSPropertyNames.h"
#include "CSSUnits.h"
#include "ExceptionCode.h"
#include <wtf/ListHashSet.h>
#include <wtf/Optional.h>
#include <wtf/Ref.h>
#include <wtf/RefCounted.h>
#include <wtf/RefPtr.h>

namespace WebCore {

class CSSCustomPropertyValue;
class CSSStyleDeclaration;
class CachedResource;
class DeprecatedCSSOMValue;
class StyleSheetContents;

class CSSValue : public WTF::RefCountedBase {
    WTF_MAKE_NONCOPYABLE(CSSValue); WTF_MAKE_FAST_ALLOCATED;
public:
    enum Type {
        CSS_INHERIT = 0,
        CSS_PRIMITIVE_VALUE = 1,
        CSS_VALUE_LIST = 2,
        CSS_CUSTOM = 3,
        CSS_INITIAL = 4,
        CSS_UNSET = 5,
        CSS_REVERT = 6

    };

    // Override RefCounted's deref() to ensure operator delete is called on
    // the appropriate subclass type.
    ALWAYS_INLINE void deref()
    {
        if (derefBase())
            destroy();
    }

    Type cssValueType() const;
    String cssText() const;

    ALWAYS_INLINE bool isPrimitiveValue() const { return m_classType == PrimitiveClass; }
    ALWAYS_INLINE bool isValueList() const { return m_classType >= ValueListClass; }

    bool isBaseValueList() const { return m_classType == ValueListClass; }

    ALWAYS_INLINE bool isAspectRatioValue() const { return m_classType == AspectRatioClass; }
    ALWAYS_INLINE bool isBorderImageSliceValue() const { return m_classType == BorderImageSliceClass; }
    bool isCrossfadeValue() const { return m_classType == CrossfadeClass; }
    ALWAYS_INLINE bool isCursorImageValue() const { return m_classType == CursorImageClass; }
    bool isCustomPropertyValue() const { return m_classType == CustomPropertyClass; }
    ALWAYS_INLINE bool isFontFeatureValue() const { return m_classType == FontFeatureClass; }
    ALWAYS_INLINE bool isFunctionValue() const { return m_classType == FunctionClass; }
    ALWAYS_INLINE bool isFontValue() const { return m_classType == FontClass; }
    bool isFontStyleValue() const { return m_classType == FontStyleClass; }
    bool isFontStyleRangeValue() const { return m_classType == FontStyleRangeClass; }
    bool isImageGeneratorValue() const { return m_classType >= CanvasClass && m_classType <= ConicGradientClass; }
	bool isGradientValue() const { return m_classType >= LinearGradientClass && m_classType <= ConicGradientClass; }
	bool isNamedImageValue() const { return m_classType == NamedImageClass; }
    bool isImageSetValue() const { return m_classType == ImageSetClass; }
    bool isImageValue() const { return m_classType == ImageClass; }
    bool isImplicitInitialValue() const;
    bool isInheritValue() const;
    bool isInitialValue() const;
    bool isUnsetValue() const;
    bool isRevertValue() const;
    bool isCSSWideKeyword() const;
	bool treatAsInitialValue(CSSPropertyID) const;
    bool treatAsInheritedValue(CSSPropertyID) const;
    bool isReflectValue() const { return m_classType == ReflectClass; }
    bool isShadowValue() const { return m_classType == ShadowClass; }
    bool isCubicBezierTimingFunctionValue() const { return m_classType == CubicBezierTimingFunctionClass; }
    bool isLinearTimingFunctionValue() const { return m_classType == LinearTimingFunctionClass; }
    bool isRadialGradientValue() const { return m_classType == RadialGradientClass; }
    bool isConicGradientValue() const { return m_classType == ConicGradientClass; }
    bool isStepsTimingFunctionValue() const { return m_classType == StepsTimingFunctionClass; }
	bool isSpringTimingFunctionValue() const { return m_classType == SpringTimingFunctionClass; }
    bool isCSSLineBoxContainValue() const { return m_classType == LineBoxContainClass; }
    bool isCalculationValue() const {return m_classType == CalculationClass; }
    bool isFilterImageValue() const { return m_classType == FilterImageClass; }
    bool isContentDistributionValue() const { return m_classType == CSSContentDistributionClass; }
    bool isGridAutoRepeatValue() const { return m_classType == GridAutoRepeatClass; }
    bool isGridIntegerRepeatValue() const { return m_classType == GridIntegerRepeatClass; }
	bool isGridTemplateAreasValue() const { return m_classType == GridTemplateAreasClass; }
    bool isGridLineNamesValue() const { return m_classType == GridLineNamesClass; }
    bool isUnicodeRangeValue() const { return m_classType == UnicodeRangeClass; }

    bool isVariableReferenceValue() const { return m_classType == VariableReferenceClass; }
    bool isPendingSubstitutionValue() const { return m_classType == PendingSubstitutionValueClass; }

    bool isRayValue() const { return m_classType == RayClass; }
    
    bool hasVariableReferences() const { return isVariableReferenceValue() || isPendingSubstitutionValue(); }

    Ref<DeprecatedCSSOMValue> createDeprecatedCSSOMWrapper(CSSStyleDeclaration&) const;

    template<class Functor> bool traverseSubresources(const Functor& handler) const;

    // What properties does this value rely on (eg, font-size for em units)
    void collectDirectComputationalDependencies(HashSet<CSSPropertyID>&) const;
    // What properties in the root element does this value rely on (eg. font-size for rem units)
    void collectDirectRootComputationalDependencies(HashSet<CSSPropertyID>&) const;

    bool equals(const CSSValue&) const;
    bool operator==(const CSSValue& other) const { return equals(other); }

    // https://www.w3.org/TR/css-values-4/#local-urls
    // Empty URLs and fragment-only URLs should not be resolved relative to the base URL.
    static bool isCSSLocalURL(StringView relativeURL);

protected:

    static const unsigned ClassTypeBits = 6;
    enum ClassType : uint8_t {
        PrimitiveClass,

        // Image classes.
        ImageClass,
        CursorImageClass,

        // Image generator classes.
        CanvasClass,
        NamedImageClass,
        CrossfadeClass,
        FilterImageClass,
        LinearGradientClass,
        RadialGradientClass,
        ConicGradientClass,

        // Timing function classes.
        CubicBezierTimingFunctionClass,
        LinearTimingFunctionClass,
        StepsTimingFunctionClass,
        SpringTimingFunctionClass,

        // Other class types.
        AspectRatioClass,
        BorderImageSliceClass,
        FontFeatureClass,
        FontClass,
        FontStyleClass,
        FontStyleRangeClass,
        FontFaceSrcClass,
        FunctionClass,

        ReflectClass,
        ShadowClass,
        UnicodeRangeClass,
        LineBoxContainClass,
        CalculationClass,
        GridTemplateAreasClass,

#if ENABLE(CSS_ANIMATIONS_LEVEL_2)
        AnimationTriggerScrollClass,
#endif

        CSSContentDistributionClass,

        CustomPropertyClass,
        VariableReferenceClass,
        PendingSubstitutionValueClass,

        RayClass,

        // List class types must appear after ValueListClass. Note CSSFunctionValue
        // is deliberately excluded, since we don't want it exposed to the CSS OM
        // as a list.
        ValueListClass,
        ImageSetClass,
        GridLineNamesClass,
        GridAutoRepeatClass,
        GridIntegerRepeatClass,
        // Do not append non-list class types here.
    };

public:
    static const unsigned ValueListSeparatorBits = 2;
    enum ValueListSeparator : uint8_t {
        SpaceSeparator,
        CommaSeparator,
        SlashSeparator
    };

protected:
    ALWAYS_INLINE ClassType classType() const { return static_cast<ClassType>(m_classType); }

    explicit ALWAYS_INLINE CSSValue(ClassType classType)
        : m_primitiveUnitType(CSSUnitType::CSS_UNKNOWN)
        , m_hasCachedCSSText(false)
        , m_isQuirkValue(false)
		, m_isImplicit(false)
        , m_valueListSeparator(SpaceSeparator)
        , m_classType(classType)
    {
    }

    // NOTE: This class is non-virtual for memory and performance reasons.
    // Don't go making it virtual again unless you know exactly what you're doing!

    ALWAYS_INLINE ~CSSValue() = default;

private:
    void destroy();

protected:
    // The bits in this section are only used by specific subclasses but kept here
    // to maximize struct packing.

    // CSSPrimitiveValue bits:
    CSSUnitType m_primitiveUnitType;
    mutable unsigned m_hasCachedCSSText : 1;
    unsigned m_isQuirkValue : 1;
    unsigned m_isImplicit : 1;

    ValueListSeparator m_valueListSeparator;

private:
    ClassType m_classType;

    friend class CSSValueList;
};

template<typename CSSValueType>
inline bool compareCSSValueVector(const Vector<RefPtr<CSSValueType> >& firstVector, const Vector<RefPtr<CSSValueType> >& secondVector)
{
    size_t size = firstVector.size();
    if (size != secondVector.size())
        return false;

    for (size_t i = 0; i < size; i++) {
        const RefPtr<CSSValueType>& firstPtr = firstVector[i];
        const RefPtr<CSSValueType>& secondPtr = secondVector[i];
        if (firstPtr == secondPtr || (firstPtr && secondPtr && firstPtr->equals(*secondPtr)))
            continue;
        return false;
    }
    return true;
}

template<typename CSSValueType>
inline bool compareCSSValue(const CSSValueType& first, const CSSValueType& second)
{
    return first.equals(second);
}

template<typename CSSValueType>
inline bool compareCSSValuePtr(const RefPtr<CSSValueType>& first, const RefPtr<CSSValueType>& second)
{
    return first ? second && compareCSSValue(*first, *second) : !second;
}

template<typename CSSValueType>
inline bool compareCSSValue(const Ref<CSSValueType>& first, const Ref<CSSValueType>& second)
{
    return first.get().equals(second);
}

#define CSS_VALUE_TYPE_CASTS(ValueTypeName) \
inline const CSS##ValueTypeName* toCSS##ValueTypeName(const CSSValue* value) \
{ \
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->is##ValueTypeName()); \
    return static_cast<const CSS##ValueTypeName*>(value); \
} \
inline const CSS##ValueTypeName& toCSS##ValueTypeName(const CSSValue& value) \
{ \
    ASSERT_WITH_SECURITY_IMPLICATION(value.is##ValueTypeName()); \
    return static_cast<const CSS##ValueTypeName&>(value); \
} \
inline CSS##ValueTypeName& toCSS##ValueTypeName(CSSValue& value) \
{ \
    ASSERT_WITH_SECURITY_IMPLICATION(value.is##ValueTypeName()); \
    return static_cast<CSS##ValueTypeName&>(value); \
} \
inline CSS##ValueTypeName* toCSS##ValueTypeName(CSSValue* value) \
{ \
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->is##ValueTypeName()); \
    return static_cast<CSS##ValueTypeName*>(value); \
} \
void toCSS##ValueTypeName(const CSS##ValueTypeName*); \
void toCSS##ValueTypeName(const CSS##ValueTypeName&);

} // namespace WebCore
