/*
 * (C) 1999-2003 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#include "CSSValue.h"
#include <wtf/PassRefPtr.h>
#include <wtf/Ref.h>
#include <wtf/Vector.h>

namespace WebCore {

class CSSCustomPropertyValue;
struct CSSParserValue;

class CSSValueList : public CSSValue {
public:
    typedef Vector<Ref<CSSValue>, 4>::iterator iterator;
    typedef Vector<Ref<CSSValue>, 4>::const_iterator const_iterator;

    static ALWAYS_INLINE Ref<CSSValueList> createCommaSeparated()
    {
        return adoptRef(*new CSSValueList(CommaSeparator));
    }
    static ALWAYS_INLINE Ref<CSSValueList> createSpaceSeparated()
    {
        return adoptRef(*new CSSValueList(SpaceSeparator));
    }
    static ALWAYS_INLINE Ref<CSSValueList> createSlashSeparated()
    {
        return adoptRef(*new CSSValueList(SlashSeparator));
    }

    ALWAYS_INLINE size_t size() const { return m_values.size(); }
    ALWAYS_INLINE size_t length() const { return m_values.size(); }
    ALWAYS_INLINE CSSValue* item(size_t index) { return index < m_values.size() ? &m_values[index].get() : nullptr; }
    ALWAYS_INLINE const CSSValue* item(size_t index) const { return index < m_values.size() ? &m_values[index].get() : nullptr; }
    ALWAYS_INLINE CSSValue& itemWithoutBoundsCheck(size_t index) { return m_values[index].get(); }
    ALWAYS_INLINE const CSSValue& itemWithoutBoundsCheck(size_t index) const { ASSERT(index < m_values.size()); return m_values[index].get(); }

    const_iterator begin() const { return m_values.begin(); }
    const_iterator end() const { return m_values.end(); }
    iterator begin() { return m_values.begin(); }
    iterator end() { return m_values.end(); }

    ALWAYS_INLINE void append(Ref<CSSValue>&& value) { m_values.append(WTFMove(value)); }
    ALWAYS_INLINE void prepend(Ref<CSSValue>&& value) { m_values.insert(0, WTFMove(value)); }
    bool removeAll(CSSValue*);
    bool hasValue(CSSValue*) const;
    PassRefPtr<CSSValueList> copy();

    String customCSSText() const;
    bool equals(const CSSValueList&) const;
    bool equals(const CSSValue&) const;

    template<class Functor> bool traverseSubresources(const Functor& handler) const;

    unsigned separator() const { return m_valueListSeparator; }

protected:
    CSSValueList(ClassType, ValueListSeparator);

private:
    explicit CSSValueList(ValueListSeparator);

    Vector<Ref<CSSValue>, 4> m_values;
};

CSS_VALUE_TYPE_CASTS(ValueList)

inline CSSValueList::CSSValueList(ClassType classType, ValueListSeparator listSeparator)
    : CSSValue(classType)
{
    m_valueListSeparator = listSeparator;
}

inline CSSValueList::CSSValueList(ValueListSeparator listSeparator)
    : CSSValue(ValueListClass)
{
    m_valueListSeparator = listSeparator;
}

template<class Functor>
inline bool CSSValueList::traverseSubresources(const Functor& handler) const
{
    for (unsigned i = 0; i < m_values.size(); ++i) {
        if (m_values[i].get().traverseSubresources(handler))
            return true;
    }
    return false;
}

} // namespace WebCore
