/*
 * Copyright (C) 2011, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "CSSValuePool.h"

#include "CSSParser.h"
#include "CSSPrimitiveValueMappings.h"
#include "CSSStyleSheet.h"
#include "CSSValueKeywords.h"
#include "CSSValueList.h"
#include <wtf/HashTraits.h>

namespace WebCore {

CSSValuePool* CSSValuePool::s_instance = nullptr;

NEVER_INLINE CSSValuePool& CSSValuePool::singletonSlow()
{
	s_instance = new CSSValuePool();
    return *s_instance;
}

inline CSSValuePool::CSSValuePool()
    : m_implicitInitialValue(adoptRef(*new CSSPrimitiveValue(CSSPrimitiveValue::ImplicitInitialValue)))
    , m_colorTransparent(CSSPrimitiveValue::createColor(Color(Color::transparent)))
    , m_colorWhite(CSSPrimitiveValue::createColor(Color(Color::white)))
    , m_colorBlack(CSSPrimitiveValue::createColor(Color(Color::black)))
{
}

Ref<CSSPrimitiveValue> CSSValuePool::createIdentifierValue(CSSValueID ident)
{
    if (!ident)
        return CSSPrimitiveValue::createIdentifier(ident);

    RELEASE_ASSERT(ident > 0 && ident < numCSSValueKeywords);
    if (!m_identifierValueCache[ident])
        m_identifierValueCache[ident] = CSSPrimitiveValue::createIdentifier(ident);
    return *m_identifierValueCache[ident];
}

Ref<CSSPrimitiveValue> CSSValuePool::createColorValue(unsigned rgbValue)
{
    // Just because it is common.
    if (rgbValue == Color::black)
        return m_colorBlack.copyRef();
    // These are the empty and deleted values of the hash table.
    else if (rgbValue == Color::white)
        return m_colorWhite.copyRef();
    else if (rgbValue == Color::transparent)
        return m_colorTransparent.copyRef();

    // Remove one entry at random if the cache grows too large.
    // FIXME: Use TinyLRUCache instead?
    const int maximumColorCacheSize = 512;
    if (m_colorValueCache.size() >= maximumColorCacheSize)
        m_colorValueCache.remove(m_colorValueCache.random());

    return *m_colorValueCache.ensure(rgbValue, [&rgbValue] {
        return CSSPrimitiveValue::createColor(Color(rgbValue));
    }).iterator->second;
}

Ref<CSSPrimitiveValue> CSSValuePool::createColorValue(const RGBA& color)
{
	return createColorValue(color.value());
}

Ref<CSSPrimitiveValue> CSSValuePool::createColorValue(const Color& color)
{
	return createColorValue(color.rgb());
}

Ref<CSSPrimitiveValue> CSSValuePool::createValue(double value, CSSUnitType type)
{
    ASSERT(std::isfinite(value));

    if (value < 0 || value > maximumCacheableIntegerValue)
        return CSSPrimitiveValue::create(value, type);

    // Casting to a signed integer first since casting a negative floating point value to an unsigned
    // integer is undefined behavior.
    int intValue = static_cast<int>(value);
    if (value != intValue)
        return CSSPrimitiveValue::create(value, type);

    RefPtr<CSSPrimitiveValue>* cache;
    switch (type) {
    case CSSUnitType::CSS_PX:
        cache = m_pixelValueCache;
        break;
    case CSSUnitType::CSS_PERCENTAGE:
        cache = m_percentValueCache;
        break;
    case CSSUnitType::CSS_NUMBER:
        cache = m_numberValueCache;
        break;
    default:
        return CSSPrimitiveValue::create(value, type);
    }

    if (!cache[intValue])
        cache[intValue] = CSSPrimitiveValue::create(value, type);
    return *cache[intValue];
}

Ref<CSSPrimitiveValue> CSSValuePool::createFontFamilyValue(const String& familyName, FromSystemFontIDOrNot fromSystemFontID)
{
    // Remove one entry at random if the cache grows too large.
    const int maximumFontFamilyCacheSize = 128;
    if (m_fontFamilyValueCache.size() >= maximumFontFamilyCacheSize)
        m_fontFamilyValueCache.remove(m_fontFamilyValueCache.random());

    RefPtr<CSSPrimitiveValue>& value = m_fontFamilyValueCache.add(std::pair<String, int>(familyName, fromSystemFontID), nullptr).iterator->second;
    if (!value)
        value = CSSPrimitiveValue::create(CSSFontFamily { familyName, fromSystemFontID });
    return *value;
}

PassRefPtr<CSSValueList> CSSValuePool::createFontFaceValue(const AtomicString& string)
{
    // Remove one entry at random if the cache grows too large.
    const int maximumFontFaceCacheSize = 128;
    if (m_fontFaceValueCache.size() >= maximumFontFaceCacheSize)
        m_fontFaceValueCache.remove(m_fontFaceValueCache.random());

    auto& value = m_fontFaceValueCache.add(string, nullptr).iterator->second;
    if (value)
		return value;

	if (RefPtr<CSSValue> result = CSSParser::parseSingleValue(CSSPropertyFontFamily, string)) {
		if (result->isValueList()) {
			value = static_pointer_cast<CSSValueList>(result);
		} else if (result->isPrimitiveValue()) {
			auto list = CSSValueList::createCommaSeparated();
			list->append(result.releaseNonNull());
			value = WTFMove(list);
		}
	}
	return value;
}

void CSSValuePool::drain()
{
    m_colorValueCache.clear();
    m_fontFaceValueCache.clear();
    m_fontFamilyValueCache.clear();

    for (int i = 0; i < numCSSValueKeywords; ++i)
        m_identifierValueCache[i] = nullptr;

    for (int i = 0; i < maximumCacheableIntegerValue; ++i) {
        m_pixelValueCache[i] = nullptr;
        m_percentValueCache[i] = nullptr;
        m_numberValueCache[i] = nullptr;
    }
}

void drainCSSValuePool()
{
	CSSValuePool::singleton().drain();
}

}
