/*
 * Copyright (C) 2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "DOMCSSRegisterCustomProperty.h"

#include "CSSCustomPropertyValue.h"
#include "CSSPropertyNames.h"
#include "CSSPropertyParser.h"
#include "CSSRegisteredCustomProperty.h"
#include "CSSTokenizer.h"
#include "DOMCSSNamespace.h"
#include "Document.h"
#include "StyleBuilder.h"
#include "StyleBuilderConverter.h"
#include "StyleResolver.h"
#include <wtf/text/WTFString.h>

namespace WebCore {

void DOMCSSRegisterCustomProperty::registerProperty(Document* document, const DOMCSSCustomPropertyDescriptor& descriptor, ExceptionCode& ec)
{
	ec = 0;

    if (!isCustomPropertyName(descriptor.name)) {
        //return Exception { SyntaxError, "The name of this property is not a custom property name." };
    	ec = SYNTAX_ERR;
		return;
	}

    RefPtr<CSSCustomPropertyValue> initialValue;
    if (!descriptor.initialValue.isEmpty()) {
        CSSTokenizer tokenizer(descriptor.initialValue);
        Ref<StyleResolver> styleResolver = StyleResolver::create(*document);

        // We need to initialize this so that we can successfully parse computationally dependent values (like em units).
        // We don't actually need the values to be accurate, since they will be rejected later anyway
        styleResolver->applyPropertyToStyle(CSSPropertyInvalid, nullptr, styleResolver->defaultStyleForElement());

        HashSet<CSSPropertyID> dependencies;
        CSSPropertyParser::collectParsedCustomPropertyValueDependencies(descriptor.syntax, false, dependencies, tokenizer.tokenRange(), strictCSSParserContext());

        if (!dependencies.isEmpty()) {
            //return Exception { SyntaxError, "The given initial value must be computationally independent." };
            ec = SYNTAX_ERR;
			return;
		}

        MatchResult matchResult;
        Style::Builder dummyBuilder(styleResolver, matchResult, OptionSet<Style::CascadeLevel>());
        initialValue = CSSPropertyParser::parseTypedCustomPropertyValue(descriptor.name, descriptor.syntax, tokenizer.tokenRange(), dummyBuilder.state(), strictCSSParserContext());

        if (!initialValue || !initialValue->isResolved()) {
            //return Exception { SyntaxError, "The given initial value does not parse for the given syntax." };
            ec = SYNTAX_ERR;
            return;
        }

        initialValue->collectDirectComputationalDependencies(dependencies);
        initialValue->collectDirectRootComputationalDependencies(dependencies);

        if (!dependencies.isEmpty()) {
            //return Exception { SyntaxError, "The given initial value must be computationally independent." };
            ec = SYNTAX_ERR;
            return;
        }
    }

    CSSRegisteredCustomProperty property { descriptor.name, descriptor.syntax, descriptor.inherits, WTFMove(initialValue) };
    if (!document->registerCSSProperty(WTFMove(property))) {
        //return Exception { InvalidModificationError, "This property has already been registered." };
    	ec = INVALID_MODIFICATION_ERR;
    	return;
    }

    document->styleScope().didChangeStyleSheetEnvironment();
}

DOMCSSRegisterCustomProperty* DOMCSSRegisterCustomProperty::from(DOMCSSNamespace& css)
{
    auto* supplement = static_cast<DOMCSSRegisterCustomProperty*>(Supplement<DOMCSSNamespace>::from(&css, supplementName()));
    if (!supplement) {
        OwnPtr<DOMCSSRegisterCustomProperty> newSupplement = adoptPtr(new DOMCSSRegisterCustomProperty(css));
        supplement = newSupplement.get();
        provideTo(&css, supplementName(), newSupplement.release());
    }
    return supplement;
}

const char* DOMCSSRegisterCustomProperty::supplementName()
{
    return "DOMCSSRegisterCustomProperty";
}

}
