/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "DOMMatrix.h"

#include "ExceptionCode.h"
#include "ScriptExecutionContext.h"
#include <cmath>
#include <limits>

namespace WebCore {

PassRefPtr<DOMMatrix> DOMMatrix::create(ScriptExecutionContext*, ExceptionCode&)
{
    return adoptRef(new DOMMatrix);
}

PassRefPtr<DOMMatrix> DOMMatrix::create(ScriptExecutionContext* scriptExecutionContext, const String& init, ExceptionCode& ec)
{
	if (!scriptExecutionContext->isDocument()) {
		ec = TypeError;
		return nullptr;
	}

	auto parseResult = parseStringIntoAbstractMatrix(init, ec);
	if (ec)
		return nullptr;

	return adoptRef(new DOMMatrix(parseResult.matrix, parseResult.is2D ? Is2D::Yes : Is2D::No));
}

PassRefPtr<DOMMatrix> DOMMatrix::create(ScriptExecutionContext* scriptExecutionContext, const Vector<double>& init, ExceptionCode& ec)
{
	if (init.size() == 6) {
		return adoptRef(new DOMMatrix(TransformationMatrix(
			init[0], init[1], init[2], init[3], init[4], init[5]
		), Is2D::Yes));
	}
	if (init.size() == 16) {
		return adoptRef(new DOMMatrix(TransformationMatrix(
			init[0], init[1], init[2], init[3],
			init[4], init[5], init[6], init[7],
			init[8], init[9], init[10], init[11],
			init[12], init[13], init[14], init[15]
		), Is2D::No));
	}
	ec = TypeError;
	return nullptr;
}

DOMMatrix::DOMMatrix(const TransformationMatrix& matrix, Is2D is2D)
    : DOMMatrixReadOnly(matrix, is2D)
{
}

// https://drafts.fxtf.org/geometry/#create-a-dommatrix-from-the-dictionary
PassRefPtr<DOMMatrix> DOMMatrix::fromMatrix(DOMMatrixInit& init, ExceptionCode& ec)
{
    return fromMatrixHelper<DOMMatrix>(init, ec);
}

PassRefPtr<DOMMatrix> DOMMatrix::fromFloat32Array(RefPtr<JSC::Float32Array> array32, ExceptionCode& ec)
{
    if (array32->length() == 6)
        return DOMMatrix::create(TransformationMatrix(array32->item(0), array32->item(1), array32->item(2), array32->item(3), array32->item(4), array32->item(5)), Is2D::Yes);

    if (array32->length() == 16) {
        return DOMMatrix::create(TransformationMatrix(
            array32->item(0), array32->item(1), array32->item(2), array32->item(3),
            array32->item(4), array32->item(5), array32->item(6), array32->item(7),
            array32->item(8), array32->item(9), array32->item(10), array32->item(11),
            array32->item(12), array32->item(13), array32->item(14), array32->item(15)
        ), Is2D::No);
    }

    ec = TypeError;
    return nullptr;
}

PassRefPtr<DOMMatrix> DOMMatrix::fromFloat64Array(RefPtr<JSC::Float64Array> array64, ExceptionCode& ec)
{
    if (array64->length() == 6)
        return DOMMatrix::create(TransformationMatrix(array64->item(0), array64->item(1), array64->item(2), array64->item(3), array64->item(4), array64->item(5)), Is2D::Yes);

    if (array64->length() == 16) {
        return DOMMatrix::create(TransformationMatrix(
            array64->item(0), array64->item(1), array64->item(2), array64->item(3),
            array64->item(4), array64->item(5), array64->item(6), array64->item(7),
            array64->item(8), array64->item(9), array64->item(10), array64->item(11),
            array64->item(12), array64->item(13), array64->item(14), array64->item(15)
        ), Is2D::No);
    }

    ec = TypeError;
    return nullptr;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-multiplyself
PassRefPtr<DOMMatrix> DOMMatrix::multiplySelf(DOMMatrixInit& other, ExceptionCode& ec)
{
	RefPtr<DOMMatrix> otherObject = DOMMatrix::fromMatrix(other, ec);
    if (ec)
        return nullptr;
    m_matrix.multiply(otherObject->m_matrix);
    if (!otherObject->is2D())
        m_is2D = false;
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-premultiplyself
PassRefPtr<DOMMatrix> DOMMatrix::preMultiplySelf(DOMMatrixInit& other, ExceptionCode& ec)
{
	RefPtr<DOMMatrix> otherObject = DOMMatrix::fromMatrix(other, ec);
    if (ec)
        return nullptr;
    m_matrix = otherObject->m_matrix * m_matrix;
    if (!otherObject->is2D())
        m_is2D = false;
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-translateself
PassRefPtr<DOMMatrix> DOMMatrix::translateSelf(double tx, double ty, double tz)
{
    m_matrix.translate3d(tx, ty, tz);
    if (tz)
        m_is2D = false;
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-scaleself
PassRefPtr<DOMMatrix> DOMMatrix::scaleSelf(double scaleX, Optional<double> scaleY, double scaleZ, double originX, double originY, double originZ)
{
    if (!scaleY)
        scaleY = scaleX;
    translateSelf(originX, originY, originZ);
    // Post-multiply a non-uniform scale transformation on the current matrix.
    // The 3D scale matrix is described in CSS Transforms with sx = scaleX, sy = scaleY and sz = scaleZ.
    m_matrix.scale3d(scaleX, scaleY.value(), scaleZ);
    translateSelf(-originX, -originY, -originZ);
    if (scaleZ != 1 || originZ)
        m_is2D = false;
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-scale3dself
PassRefPtr<DOMMatrix> DOMMatrix::scale3dSelf(double scale, double originX, double originY, double originZ)
{
    translateSelf(originX, originY, originZ);
    // Post-multiply a uniform 3D scale transformation (m11 = m22 = m33 = scale) on the current matrix.
    // The 3D scale matrix is described in CSS Transforms with sx = sy = sz = scale. [CSS3-TRANSFORMS]
    m_matrix.scale3d(scale, scale, scale);
    translateSelf(-originX, -originY, -originZ);
    if (scale != 1)
        m_is2D = false;
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-rotateself
PassRefPtr<DOMMatrix> DOMMatrix::rotateSelf(double rotX, Optional<double> rotY, Optional<double> rotZ)
{
    if (!rotY && !rotZ) {
        rotZ = rotX;
        rotX = 0;
        rotY = 0;
    }
    m_matrix.rotate3d(rotX, rotY.value_or(0), rotZ.value_or(0));
    if (rotX || rotY.value_or(0))
        m_is2D = false;
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-rotatefromvectorself
PassRefPtr<DOMMatrix> DOMMatrix::rotateFromVectorSelf(double x, double y)
{
    m_matrix.rotateFromVector(x, y);
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-rotateaxisangleself
PassRefPtr<DOMMatrix> DOMMatrix::rotateAxisAngleSelf(double x, double y, double z, double angle)
{
    m_matrix.rotate3d(x, y, z, angle);
    if (x || y)
        m_is2D = false;
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-skewxself
PassRefPtr<DOMMatrix> DOMMatrix::skewXSelf(double sx)
{
    m_matrix.skewX(sx);
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-skewyself
PassRefPtr<DOMMatrix> DOMMatrix::skewYSelf(double sy)
{
    m_matrix.skewY(sy);
    return this;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-invertself
PassRefPtr<DOMMatrix> DOMMatrix::invertSelf()
{
	Optional<TransformationMatrix> inverse = m_matrix.inverse();
    if (inverse)
        m_matrix = *inverse;
    else {
        m_matrix.setMatrix(
            std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(),
            std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(),
            std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(),
            std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN(), std::numeric_limits<double>::quiet_NaN()
        );
        m_is2D = false;
    }
    return this;
}

PassRefPtr<DOMMatrix> DOMMatrix::setMatrixValueForBindings(const String& string, ExceptionCode& ec)
{
    setMatrixValue(string, ec);
    if (ec)
        return nullptr;
    return this;
}

} // namespace WebCore
