/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "DOMMatrixReadOnly.h"

#include "CSSParser.h"
#include "CSSToLengthConversionData.h"
#include "DOMPoint.h"
#include "DOMMatrix.h"
#include "ExceptionCode.h"
#include "ScriptExecutionContext.h"
#include "StylePropertySet.h"
#include "StyleResolver.h"
#include "TransformFunctions.h"
#include "TransformOperations.h"
#include <runtime/GenericTypedArrayViewInlines.h>
#include <runtime/JSGenericTypedArrayViewInlines.h>
#include <heap/Heap.h>
#include <wtf/text/StringBuilder.h>

namespace WebCore {

// https://drafts.fxtf.org/geometry/#dom-dommatrixreadonly-dommatrixreadonly
PassRefPtr<DOMMatrixReadOnly> DOMMatrixReadOnly::create(ScriptExecutionContext*, ExceptionCode&)
{
    return adoptRef(new DOMMatrixReadOnly);
}

PassRefPtr<DOMMatrixReadOnly> DOMMatrixReadOnly::create(ScriptExecutionContext* scriptExecutionContext, const String& init, ExceptionCode& ec)
{
    if (!scriptExecutionContext->isDocument()) {
        ec = TypeError;
        return nullptr;
    }

    AbstractMatrix parseResult = parseStringIntoAbstractMatrix(init, ec);
    if (ec)
        return nullptr;

	return adoptRef(new DOMMatrixReadOnly(parseResult.matrix, parseResult.is2D ? Is2D::Yes : Is2D::No));
}

PassRefPtr<DOMMatrixReadOnly> DOMMatrixReadOnly::create(ScriptExecutionContext* scriptExecutionContext, const Vector<double>& init, ExceptionCode& ec)
{
	if (init.size() == 6) {
		return adoptRef(new DOMMatrixReadOnly(TransformationMatrix(
			init[0], init[1], init[2], init[3], init[4], init[5]
		), Is2D::Yes));
	}
	if (init.size() == 16) {
		return adoptRef(new DOMMatrixReadOnly(TransformationMatrix(
			init[0], init[1], init[2], init[3],
			init[4], init[5], init[6], init[7],
			init[8], init[9], init[10], init[11],
			init[12], init[13], init[14], init[15]
		), Is2D::No));
	}
	ec = TypeError;
	return nullptr;
}

DOMMatrixReadOnly::DOMMatrixReadOnly(const TransformationMatrix& matrix, Is2D is2D)
    : m_matrix(matrix)
    , m_is2D(is2D == Is2D::Yes)
{
    ASSERT(!m_is2D || m_matrix.isAffine());
}

inline PassRefPtr<DOMMatrix> DOMMatrixReadOnly::cloneAsDOMMatrix() const
{
    return DOMMatrix::create(m_matrix, m_is2D ? Is2D::Yes : Is2D::No);
}

// https://tc39.github.io/ecma262/#sec-samevaluezero
static bool sameValueZero(double a, double b)
{
    if (std::isnan(a) && std::isnan(b))
        return true;
    return a == b;
}

// https://drafts.fxtf.org/geometry/#matrix-validate-and-fixup
void DOMMatrixReadOnly::validateAndFixup(DOMMatrix2DInit& init, ExceptionCode& ec)
{
    if (init.a && init.m11 && !sameValueZero(init.a.value(), init.m11.value())) {
        //return Exception { TypeError, ASCIILiteral("init.a and init.m11 do not match") };
    	ec = TypeError;
    	return;
    }
    if (init.b && init.m12 && !sameValueZero(init.b.value(), init.m12.value())) {
        //return Exception { TypeError, ASCIILiteral("init.b and init.m12 do not match") };
    	ec = TypeError;
    	return;
    }
    if (init.c && init.m21 && !sameValueZero(init.c.value(), init.m21.value())) {
        //return Exception { TypeError, ASCIILiteral("init.c and init.m21 do not match") };
    	ec = TypeError;
    	return;
    }
    if (init.d && init.m22 && !sameValueZero(init.d.value(), init.m22.value())) {
        //return Exception { TypeError, ASCIILiteral("init.d and init.m22 do not match") };
    	ec = TypeError;
    	return;
    }
    if (init.e && init.m41 && !sameValueZero(init.e.value(), init.m41.value())) {
        //return Exception { TypeError, ASCIILiteral("init.e and init.m41 do not match") };
    	ec = TypeError;
    	return;
    }
    if (init.f && init.m42 && !sameValueZero(init.f.value(), init.m42.value())) {
        //return Exception { TypeError, ASCIILiteral("init.f and init.m42 do not match") };
    	ec = TypeError;
    	return;
    }

    if (!init.m11)
        init.m11 = init.a.value_or(1);
    if (!init.m12)
        init.m12 = init.b.value_or(0);
    if (!init.m21)
        init.m21 = init.c.value_or(0);
    if (!init.m22)
        init.m22 = init.d.value_or(1);
    if (!init.m41)
        init.m41 = init.e.value_or(0);
    if (!init.m42)
        init.m42 = init.f.value_or(0);

}

void DOMMatrixReadOnly::validateAndFixup(DOMMatrixInit& init, ExceptionCode& ec)
{
    validateAndFixup(static_cast<DOMMatrix2DInit&>(init), ec);
    if (ec)
        return;

    if (init.is2D && init.is2D.value()) {
        if (init.m13) {
            //return Exception { TypeError, ASCIILiteral("m13 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m14) {
            //return Exception { TypeError, ASCIILiteral("m14 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m23) {
            //return Exception { TypeError, ASCIILiteral("m23 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m24) {
            //return Exception { TypeError, ASCIILiteral("m24 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m31) {
            //return Exception { TypeError, ASCIILiteral("m31 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m32) {
            //return Exception { TypeError, ASCIILiteral("m32 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m34) {
            //return Exception { TypeError, ASCIILiteral("m34 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m43) {
            //return Exception { TypeError, ASCIILiteral("m43 should be 0 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m33 != 1) {
            //return Exception { TypeError, ASCIILiteral("m33 should be 1 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
        if (init.m44 != 1) {
            //return Exception { TypeError, ASCIILiteral("m44 should be 1 for a 2D matrix") };
        	ec = TypeError;
        	return;
        }
    }

    if (!init.is2D) {
        if (init.m13 || init.m14 || init.m23 || init.m24 || init.m31 || init.m32 || init.m34 || init.m43 || init.m33 != 1 || init.m44 != 1)
            init.is2D = false;
        else
            init.is2D = true;
    }
}

PassRefPtr<DOMMatrixReadOnly> DOMMatrixReadOnly::fromMatrix(DOMMatrixInit& init, ExceptionCode& ec)
{
    return fromMatrixHelper<DOMMatrixReadOnly>(init, ec);
}

PassRefPtr<DOMMatrixReadOnly> DOMMatrixReadOnly::fromFloat32Array(RefPtr<JSC::Float32Array> array32, ExceptionCode& ec)
{
    if (array32->length() == 6)
        return DOMMatrixReadOnly::create(TransformationMatrix(array32->item(0), array32->item(1), array32->item(2), array32->item(3), array32->item(4), array32->item(5)), Is2D::Yes);

    if (array32->length() == 16) {
        return DOMMatrixReadOnly::create(TransformationMatrix(
            array32->item(0), array32->item(1), array32->item(2), array32->item(3),
            array32->item(4), array32->item(5), array32->item(6), array32->item(7),
            array32->item(8), array32->item(9), array32->item(10), array32->item(11),
            array32->item(12), array32->item(13), array32->item(14), array32->item(15)
        ), Is2D::No);
    }

    ec = TypeError;
    return nullptr;
}

PassRefPtr<DOMMatrixReadOnly> DOMMatrixReadOnly::fromFloat64Array(RefPtr<JSC::Float64Array> array64, ExceptionCode& ec)
{
    if (array64->length() == 6)
        return DOMMatrixReadOnly::create(TransformationMatrix(array64->item(0), array64->item(1), array64->item(2), array64->item(3), array64->item(4), array64->item(5)), Is2D::Yes);

    if (array64->length() == 16) {
        return DOMMatrixReadOnly::create(TransformationMatrix(
            array64->item(0), array64->item(1), array64->item(2), array64->item(3),
            array64->item(4), array64->item(5), array64->item(6), array64->item(7),
            array64->item(8), array64->item(9), array64->item(10), array64->item(11),
            array64->item(12), array64->item(13), array64->item(14), array64->item(15)
        ), Is2D::No);
    }

    ec = TypeError;
    return nullptr;
}

bool DOMMatrixReadOnly::isIdentity() const
{
    return m_matrix.isIdentity();
}

DOMMatrixReadOnly::AbstractMatrix DOMMatrixReadOnly::parseStringIntoAbstractMatrix(const String& string, ExceptionCode& ec)
{
    if (string.isEmpty())
        return AbstractMatrix();

    RefPtr<MutableStylePropertySet> styleDeclaration = MutableStyleProperties::create();
    if (CSSParser::parseValue(styleDeclaration.get(), CSSPropertyTransform, string, true, HTMLStandardMode, 0) == CSSParser::ParseResult::Error) {
        ec = SYNTAX_ERR;
        return AbstractMatrix();
    }

    // Convert to TransformOperations. This can fail if a property requires style (i.e., param uses 'ems' or 'exs')
    RefPtr<CSSValue> value = styleDeclaration->getPropertyCSSValue(CSSPropertyTransform);

    // Check for a "none" or empty transform. In these cases we can use the default identity matrix.
    if (!value || ((*value).isPrimitiveValue() && toCSSPrimitiveValue(*value).valueID() == CSSValueNone))
        return AbstractMatrix();

    TransformOperations operations;
    if (!transformsForValue(*value, CSSToLengthConversionData(), operations)) {
        ec = SYNTAX_ERR;
        return AbstractMatrix();
    }

    AbstractMatrix matrix;

    for(unsigned i = 0; i < operations.operations().size(); ++i) {
    	RefPtr<TransformOperation>& operation = operations.operations()[i];
        if (operation->apply(matrix.matrix, IntSize())) {
            ec = SYNTAX_ERR;
            return AbstractMatrix();
        }
        if (operation->is3DOperation())
            matrix.is2D = false;
    }

    return matrix;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrix-setmatrixvalue
void DOMMatrixReadOnly::setMatrixValue(const String& string, ExceptionCode& ec)
{
    auto parseResult = parseStringIntoAbstractMatrix(string, ec);
    if (ec)
        return;

    m_is2D = parseResult.is2D;
    m_matrix = parseResult.matrix;
    return;
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::translate(double tx, double ty, double tz)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->translateSelf(tx, ty, tz);
}

// https://drafts.fxtf.org/geometry/#dom-dommatrixreadonly-flipx
PassRefPtr<DOMMatrix> DOMMatrixReadOnly::flipX()
{
    auto matrix = cloneAsDOMMatrix();
    matrix->m_matrix.flipX();
    return matrix;
}

// https://drafts.fxtf.org/geometry/#dom-dommatrixreadonly-flipy
PassRefPtr<DOMMatrix> DOMMatrixReadOnly::flipY()
{
    auto matrix = cloneAsDOMMatrix();
    matrix->m_matrix.flipY();
    return matrix;
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::multiply(DOMMatrixInit& other, ExceptionCode& ec) const
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->multiplySelf(other, ec);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::scale(double scaleX, Optional<double> scaleY, double scaleZ, double originX, double originY, double originZ)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->scaleSelf(scaleX, scaleY, scaleZ, originX, originY, originZ);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::scale3d(double scale, double originX, double originY, double originZ)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->scale3dSelf(scale, originX, originY, originZ);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::rotate(double rotX, Optional<double> rotY, Optional<double> rotZ)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->rotateSelf(rotX, rotY, rotZ);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::rotateFromVector(double x, double y)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->rotateFromVectorSelf(x, y);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::rotateAxisAngle(double x, double y, double z, double angle)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->rotateAxisAngleSelf(x, y, z, angle);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::skewX(double sx)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->skewXSelf(sx);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::skewY(double sy)
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->skewYSelf(sy);
}

PassRefPtr<DOMMatrix> DOMMatrixReadOnly::inverse() const
{
    auto matrix = cloneAsDOMMatrix();
    return matrix->invertSelf();
}

// https://drafts.fxtf.org/geometry/#dom-dommatrixreadonly-transformpoint
PassRefPtr<DOMPoint> DOMMatrixReadOnly::transformPoint(DOMPointInit& pointInit)
{
    m_matrix.map4ComponentPoint(pointInit.x, pointInit.y, pointInit.z, pointInit.w);
    return DOMPoint::create(pointInit.x, pointInit.y, pointInit.z, pointInit.w);
}

PassRefPtr<JSC::Float32Array> DOMMatrixReadOnly::toFloat32Array(ExceptionCode& ec) const
{
	RefPtr<JSC::Float32Array> array32 = JSC::Float32Array::createUninitialized(16);
    if (!array32) {
        //return Exception { UnknownError, ASCIILiteral("Out of memory") };
        ec = UnknownError;
        return nullptr;
    }

    unsigned index = 0;
    array32->set(index++, m_matrix.m11());
    array32->set(index++, m_matrix.m12());
    array32->set(index++, m_matrix.m13());
    array32->set(index++, m_matrix.m14());
    array32->set(index++, m_matrix.m21());
    array32->set(index++, m_matrix.m22());
    array32->set(index++, m_matrix.m23());
    array32->set(index++, m_matrix.m24());
    array32->set(index++, m_matrix.m31());
    array32->set(index++, m_matrix.m32());
    array32->set(index++, m_matrix.m33());
    array32->set(index++, m_matrix.m34());
    array32->set(index++, m_matrix.m41());
    array32->set(index++, m_matrix.m42());
    array32->set(index++, m_matrix.m43());
    array32->set(index, m_matrix.m44());
    return array32.releaseNonNull();
}

PassRefPtr<JSC::Float64Array> DOMMatrixReadOnly::toFloat64Array(ExceptionCode& ec) const
{
	RefPtr<JSC::Float64Array> array64 = JSC::Float64Array::createUninitialized(16);
    if (!array64) {
        //return Exception { UnknownError, ASCIILiteral("Out of memory") };
        ec = UnknownError;
        return nullptr;
    }

    unsigned index = 0;
    array64->set(index++, m_matrix.m11());
    array64->set(index++, m_matrix.m12());
    array64->set(index++, m_matrix.m13());
    array64->set(index++, m_matrix.m14());
    array64->set(index++, m_matrix.m21());
    array64->set(index++, m_matrix.m22());
    array64->set(index++, m_matrix.m23());
    array64->set(index++, m_matrix.m24());
    array64->set(index++, m_matrix.m31());
    array64->set(index++, m_matrix.m32());
    array64->set(index++, m_matrix.m33());
    array64->set(index++, m_matrix.m34());
    array64->set(index++, m_matrix.m41());
    array64->set(index++, m_matrix.m42());
    array64->set(index++, m_matrix.m43());
    array64->set(index, m_matrix.m44());
    return array64.releaseNonNull();
}

// https://drafts.fxtf.org/geometry/#dom-dommatrixreadonly-stringifier
String DOMMatrixReadOnly::toString(ExceptionCode& ec) const
{
    if (!m_matrix.containsOnlyFiniteValues()) {
        //return Exception { INVALID_STATE_ERR, ASCIILiteral("Matrix contains non-finite values") };
    	ec = INVALID_STATE_ERR;
    	return String();
    }

    StringBuilder builder;
    if (is2D()) {
        builder.appendLiteral("matrix(");
        builder.append(String::numberToStringECMAScript(m_matrix.a()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.b()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.c()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.d()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.e()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.f()));
    } else {
        builder.appendLiteral("matrix3d(");
        builder.append(String::numberToStringECMAScript(m_matrix.m11()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m12()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m13()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m14()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m21()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m22()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m23()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m24()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m31()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m32()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m33()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m34()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m41()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m42()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m43()));
        builder.appendLiteral(", ");
        builder.append(String::numberToStringECMAScript(m_matrix.m44()));
    }
    builder.append(')');
    return builder.toString();
}

} // namespace WebCore
