/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "MediaQueryEvaluator.h"
#include "RuleSet.h"
#include "SelectorChecker.h"
#include "StyleScope.h"
#include <wtf/PassOwnPtr.h>
#include <wtf/Ref.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>

namespace WebCore {

class DocumentRuleSets;
class MatchRequest;

namespace Style {
struct SelectorMatchingState;
}

class PseudoStyleRequest {
	WTF_MAKE_FAST_ALLOCATED;
public:
    explicit PseudoStyleRequest(PseudoId pseudoId, RenderScrollbar* scrollbar = nullptr, ScrollbarPart scrollbarPart = NoPart)
        : pseudoId(pseudoId)
        , scrollbarPart(scrollbarPart)
        , scrollbar(scrollbar)
    {
    }

    PseudoStyleRequest(PseudoId pseudoId, const AtomicString& highlightName)
        : pseudoId(pseudoId)
        , highlightName(highlightName)
    {
        ASSERT(pseudoId == PseudoId::Highlight);
    }

    PseudoId pseudoId;
    ScrollbarPart scrollbarPart { NoPart };
    RenderScrollbar* scrollbar { nullptr };
    AtomicString highlightName;
};

struct MatchedRule {
	MatchedRule(const RuleData* ruleData, unsigned specificity, Style::ScopeOrdinal styleScopeOrdinal)
	: ruleData(ruleData)
	, specificity(specificity)
	, styleScopeOrdinal(styleScopeOrdinal)
	{ }
    const RuleData* ruleData;
    unsigned specificity;
    Style::ScopeOrdinal styleScopeOrdinal;
};

struct MatchedProperties {
	WTF_MAKE_FAST_ALLOCATED;
public:
	ALWAYS_INLINE MatchedProperties() : linkMatchType(SelectorChecker::MatchAll), whitelistType(PropertyWhitelistNone), styleScopeOrdinal(Style::ScopeOrdinal::Element) { }
	ALWAYS_INLINE MatchedProperties(const StylePropertySet& properties, unsigned linkMatchType = SelectorChecker::MatchAll, PropertyWhitelistType whitelistType = PropertyWhitelistNone, Style::ScopeOrdinal styleScopeOrdinal = Style::ScopeOrdinal::Element)
		: properties(properties)
		, linkMatchType(linkMatchType)
		, whitelistType(whitelistType)
		, styleScopeOrdinal(styleScopeOrdinal) { }

    RefPtr<const StylePropertySet> properties;
    unsigned linkMatchType;
    PropertyWhitelistType whitelistType;
    Style::ScopeOrdinal styleScopeOrdinal;
};

struct MatchResult {
	WTF_MAKE_FAST_ALLOCATED;
public:
    bool isCacheable { true };
    Vector<MatchedProperties> userAgentDeclarations;
    Vector<MatchedProperties> userDeclarations;
    Vector<MatchedProperties> authorDeclarations;

    bool operator==(const MatchResult& other) const
    {
        return isCacheable == other.isCacheable
            && userAgentDeclarations == other.userAgentDeclarations
            && userDeclarations == other.userDeclarations
            && authorDeclarations == other.authorDeclarations;
    }
    bool operator!=(const MatchResult& other) const { return !(*this == other); }

    bool isEmpty() const { return userAgentDeclarations.isEmpty() && userDeclarations.isEmpty() && authorDeclarations.isEmpty(); }
};

class ElementRuleCollector {
	WTF_MAKE_FAST_ALLOCATED;
public:
    ElementRuleCollector(const Element&, const DocumentRuleSets&, Style::SelectorMatchingState*);
    ElementRuleCollector(const Element&, const RuleSet& authorStyle, Style::SelectorMatchingState*);

    void setIncludeEmptyRules(bool value) { m_shouldIncludeEmptyRules = value; }

    void matchAllRules(bool matchAuthorAndUserStyles, bool includeSMILProperties);
    void matchUARules();
    void matchAuthorRules();
    void matchUserRules();

    bool matchesAnyAuthorRules();

    void setMode(SelectorChecker::Mode mode) { m_mode = mode; }
    void setPseudoStyleRequest(const PseudoStyleRequest& request) { m_pseudoStyleRequest = request; }
    void setMedium(const MediaQueryEvaluator* medium) { m_isPrintStyle = medium->mediaTypeMatchSpecific("print"); }

    bool hasAnyMatchingRules(const RuleSet*);

    const MatchResult& matchResult() const;
    const Vector<RefPtr<const StyleRule>>& matchedRuleList() const;

    void clearMatchedRules();

    const PseudoIdSet& matchedPseudoElementIds() const { return m_matchedPseudoElementIds; }
    const Style::Relations& styleRelations() const { return m_styleRelations; }
    bool didMatchUncommonAttributeSelector() const { return m_didMatchUncommonAttributeSelector; }

private:
    ALWAYS_INLINE void addElementStyleProperties(const StylePropertySet*, bool isCacheable = true);

    void matchUARules(const RuleSet&);

    void collectMatchingAuthorRules();
    void addElementInlineStyleProperties(bool includeSMILProperties);

    void matchAuthorShadowPseudoElementRules();
	void matchHostPseudoClassRules();
    void matchSlottedPseudoElementRules();
    void matchPartPseudoElementRules();
    void matchPartPseudoElementRulesForScope(const Element& partMatchingElement);

    void collectMatchingShadowPseudoElementRules(const MatchRequest&);

    void collectMatchingRules(const MatchRequest&);
    void collectMatchingRulesForList(const RuleSet::RuleDataVector&, const MatchRequest&);
    void collectMatchingRulesForList(const RuleSet::RuleDataVector*, const MatchRequest&);
    ALWAYS_INLINE bool ruleMatches(const RuleData&, unsigned &specificity, Style::ScopeOrdinal);

    void sortMatchedRules();

    enum class DeclarationOrigin { UserAgent, User, Author };
    static Vector<MatchedProperties>& declarationsForOrigin(MatchResult&, DeclarationOrigin);
	void sortAndTransferMatchedRules(DeclarationOrigin);
    void transferMatchedRules(DeclarationOrigin, Optional<Style::ScopeOrdinal> forScope = Nullopt);

    ALWAYS_INLINE void addMatchedRule(const RuleData&, unsigned specificity, Style::ScopeOrdinal);
    void addMatchedProperties(MatchedProperties&&, DeclarationOrigin);

    const Element& element() const { return m_element.get(); }

    const Ref<const Element> m_element;
    Ref<const RuleSet> m_authorStyle;
    RefPtr<const RuleSet> m_userStyle;
    RefPtr<const RuleSet> m_userAgentMediaQueryStyle;
    Style::SelectorMatchingState* m_selectorMatchingState;

    bool m_shouldIncludeEmptyRules;
    bool m_isPrintStyle;
    PseudoStyleRequest m_pseudoStyleRequest;
    SelectorChecker::Mode m_mode;
#if ENABLE(SHADOW_DOM)
    bool m_isMatchingHostPseudoClass;
#endif

    Vector<MatchedRule, 64> m_matchedRules;
    size_t m_matchedRuleTransferIndex;

    // Output.
    Vector<RefPtr<const StyleRule>> m_matchedRuleList;
    bool m_didMatchUncommonAttributeSelector;
    MatchResult m_result;
    Style::Relations m_styleRelations;
    PseudoIdSet m_matchedPseudoElementIds;
};

ALWAYS_INLINE bool operator==(const MatchedProperties& a, const MatchedProperties& b)
{
    return a.properties == b.properties && a.linkMatchType == b.linkMatchType;
}

ALWAYS_INLINE bool operator!=(const MatchedProperties& a, const MatchedProperties& b)
{
    return !(a == b);
}

inline const MatchResult& ElementRuleCollector::matchResult() const
{
    ASSERT(m_mode == SelectorChecker::Mode::ResolvingStyle);
    return m_result;
}

inline const Vector<RefPtr<const StyleRule>>& ElementRuleCollector::matchedRuleList() const
{
    ASSERT(m_mode == SelectorChecker::CollectingRules);
    return m_matchedRuleList;
}

} // namespace WebCore
