/*
 * Copyright (C) 2007, 2008, 2011, 2013 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FontFace.h"

#include "CSSComputedStyleDeclaration.h"
#include "CSSFontFace.h"
#include "CSSFontFaceSource.h"
#include "CSSFontFeatureValue.h"
#include "CSSFontSelector.h"
#include "CSSFontStyleValue.h"
#include "CSSParser.h"
#include "CSSPrimitiveValueMappings.h"
#include "CSSFontSelector.h"
#include "CSSUnicodeRangeValue.h"
#include "CSSValue.h"
#include "CSSValueList.h"
#include "CSSValuePool.h"
#include "Dictionary.h"
#include "Document.h"
#include "ExceptionCode.h"
#include "FontVariantBuilder.h"
#include "DOMCoreException.h"
#include "JSFontFace.h"
#include "ScriptExecutionContext.h"
#include "Settings.h"
#include "StylePropertySet.h"
#include <runtime/JSArrayBuffer.h>
#include <wtf/HashSet.h>
#include <wtf/text/StringBuilder.h>

namespace WebCore {

HashSet<FontFace*> gFontFaces;

static bool populateFontFaceWithArrayBuffer(CSSFontFace& fontFace, PassRefPtr<JSC::ArrayBufferView> arrayBufferView)
{
    OwnPtr<CSSFontFaceSource> source = adoptPtr(new CSSFontFaceSource(fontFace, String(), arrayBufferView));
    fontFace.adoptSource(source.release());
    return false;
}

void FontFace::setErrorState()
{
    m_loadedPromise.reject(Exception(SYNTAX_ERR));
    m_backing->setErrorState();
}

PassRefPtr<FontFace> FontFace::create(JSC::ExecState& execState, Document* document, const String& family, JSC::JSValue source, const Descriptors& descriptors)
{
    RefPtr<FontFace> result = adoptRef(new FontFace(document->fontSelector()));
    result->suspendIfNeeded();

    bool dataRequiresAsynchronousLoading = true;

    ExceptionCode ec = 0;
    result->setFamily(family, ec);
    if (ec) {
        result->setErrorState();
        return result;
    }

    if (source.isString()) {
        RefPtr<CSSValue> value = FontFace::parseString(source.getString(&execState), CSSPropertySrc);
        if (!value || !value->isValueList()) {
            result->setErrorState();
            return result;
        }
		CSSFontFace::appendSources(result->backing(), toCSSValueList(*value), document, false);
    } else if (auto arrayBufferView = toArrayBufferView(source))
		dataRequiresAsynchronousLoading = populateFontFaceWithArrayBuffer(result->backing(), arrayBufferView);
	else if (auto arrayBuffer = JSC::toArrayBuffer(source)) {
		auto arrayBufferView = JSC::Uint8Array::create(arrayBuffer, 0, arrayBuffer->byteLength());
		dataRequiresAsynchronousLoading = populateFontFaceWithArrayBuffer(result->backing(), arrayBufferView);
	}

	result->setStyle(descriptors.style.isEmpty() ? ASCIILiteral("normal") : descriptors.style, ec);
	if (ec) {
        result->setErrorState();
        return result;
    }
	result->setWeight(descriptors.weight.isEmpty() ? ASCIILiteral("normal") : descriptors.weight, ec);
	if (ec) {
        result->setErrorState();
        return result;
    }
	result->setStretch(descriptors.stretch.isEmpty() ? ASCIILiteral("normal") : descriptors.stretch, ec);
	if (ec) {
        result->setErrorState();
        return result;
    }
	result->setUnicodeRange(descriptors.unicodeRange.isEmpty() ? ASCIILiteral("U+0-10FFFF") : descriptors.unicodeRange, ec);
	if (ec) {
        result->setErrorState();
        return result;
    }
	result->setVariant(descriptors.variant.isEmpty() ? ASCIILiteral("normal") : descriptors.variant, ec);
	if (ec) {
        result->setErrorState();
        return result;
    }
	result->setFeatureSettings(descriptors.featureSettings.isEmpty() ? ASCIILiteral("normal") : descriptors.featureSettings, ec);
	if (ec) {
        result->setErrorState();
        return result;
    }
    result->setDisplay(descriptors.display.isEmpty() ? ASCIILiteral("auto") : descriptors.display, ec);
    if (ec) {
        result->setErrorState();
        return result;
    }

    if (!dataRequiresAsynchronousLoading) {
        result->backing().load();
        auto status = result->backing().status();
        ASSERT_UNUSED(status, status == CSSFontFace::Status::Success || status == CSSFontFace::Status::Failure);
    }

    return result.release();
}

Ref<FontFace> FontFace::create(ScriptExecutionContext* context, CSSFontFace& face)
{
    auto fontFace = adoptRef(*new FontFace(context, face));
    fontFace->suspendIfNeeded();
    return fontFace;
}

FontFace::FontFace(CSSFontSelector& fontSelector)
	: ActiveDOMObject(fontSelector.document())
	, m_backing(CSSFontFace::create(&fontSelector, nullptr, this))
	, m_loadedPromise(*this, &FontFace::loadedPromiseResolve)
{
    m_backing->addClient(*this);
    gFontFaces.add(this);
}

FontFace::FontFace(ScriptExecutionContext*, CSSFontFace& face)
	: ActiveDOMObject(face.document())
	, m_backing(&face)
	, m_loadedPromise(*this, &FontFace::loadedPromiseResolve)
{
    m_backing->addClient(*this);
    gFontFaces.add(this);
}

FontFace::~FontFace()
{
    m_backing->removeClient(*this);
    gFontFaces.remove(this);
}

PassRefPtr<CSSValue> FontFace::parseString(const String& string, CSSPropertyID propertyID)
{
	// FIXME: Should use the Document to get the right parsing mode.
	return CSSParser::parseFontFaceDescriptor(propertyID, string, HTMLStandardMode);
}

void FontFace::setFamily(const String& family, ExceptionCode& ec)
{
    if (family.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=196381 Don't use a list here.
    // See consumeFontFamilyDescriptor() in CSSPropertyParser.cpp for why we're using it.
    auto list = CSSValueList::createCommaSeparated();
    list->append(CSSValuePool::singleton().createFontFamilyValue(family));
    bool success = m_backing->setFamilies(list);
    if (!success)
        ec = SYNTAX_ERR;
}

void FontFace::setStyle(const String& style, ExceptionCode& ec)
{
    if (style.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    if (auto value = parseString(style, CSSPropertyFontStyle)) {
        m_backing->setStyle(*value);
        return;
    }
    ec = SYNTAX_ERR;
}

void FontFace::setWeight(const String& weight, ExceptionCode& ec)
{
    if (weight.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    if (auto value = parseString(weight, CSSPropertyFontWeight)) {
        m_backing->setWeight(*value);
        return;
    }
	ec = SYNTAX_ERR;
}

void FontFace::setStretch(const String& stretch, ExceptionCode& ec)
{
    if (stretch.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    if (auto value = parseString(stretch, CSSPropertyFontStretch)) {
        m_backing->setStretch(*value);
        return;
    }
	ec = SYNTAX_ERR;
}

void FontFace::setUnicodeRange(const String& unicodeRange, ExceptionCode& ec)
{
    if (unicodeRange.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    bool success = false;
    if (auto value = parseString(unicodeRange, CSSPropertyUnicodeRange))
        success = m_backing->setUnicodeRange(*value);
    if (!success)
        ec = SYNTAX_ERR;
}

void FontFace::setVariant(const String& variant, ExceptionCode& ec)
{
    if (variant.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    Ref<MutableStyleProperties> style = MutableStyleProperties::create();
    auto result = CSSParser::parseValue(style.ptr(), CSSPropertyFontVariant, variant, true, HTMLStandardMode, nullptr);
    if (result == CSSParser::ParseResult::Error) {
		ec = SYNTAX_ERR;
		return;
	}

    // FIXME: Would be much better to stage the new settings and set them all at once
    // instead of this dance where we make a backup and revert to it if something fails.
	auto backup = m_backing->variantSettings();

	auto normal = CSSValuePool::singleton().createIdentifierValue(CSSValueNormal);
	bool success = true;

	if (auto value = style->getPropertyCSSValue(CSSPropertyFontVariantLigatures))
		success &= m_backing->setVariantLigatures(*value);
	else
		m_backing->setVariantLigatures(normal);

	if (auto value = style->getPropertyCSSValue(CSSPropertyFontVariantPosition))
		success &= m_backing->setVariantPosition(*value);
	else
		m_backing->setVariantPosition(normal);

	if (auto value = style->getPropertyCSSValue(CSSPropertyFontVariantCaps))
		success &= m_backing->setVariantCaps(*value);
	else
		m_backing->setVariantCaps(normal);

	if (auto value = style->getPropertyCSSValue(CSSPropertyFontVariantNumeric))
		success &= m_backing->setVariantNumeric(*value);
	else
		m_backing->setVariantNumeric(normal);

	if (auto value = style->getPropertyCSSValue(CSSPropertyFontVariantAlternates))
		success &= m_backing->setVariantAlternates(*value);
	else
		m_backing->setVariantAlternates(normal);

	if (auto value = style->getPropertyCSSValue(CSSPropertyFontVariantEastAsian))
		success &= m_backing->setVariantEastAsian(*value);
	else
		m_backing->setVariantEastAsian(normal);

    if (!success) {
    	if (backup)
    		m_backing->setVariantSettings(*backup);
        ec = SYNTAX_ERR;
    }
}

void FontFace::setFeatureSettings(const String& featureSettings, ExceptionCode& ec)
{
    if (featureSettings.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    RefPtr<CSSValue> value = parseString(featureSettings, CSSPropertyFontFeatureSettings);
    if (!value) {
        ec = SYNTAX_ERR;
        return;
    }
    m_backing->setFeatureSettings(*value);
}

void FontFace::setDisplay(const String& display, ExceptionCode& ec)
{
    if (display.isEmpty()) {
        ec = SYNTAX_ERR;
        return;
    }

    RefPtr<CSSValue> value = parseString(display, CSSPropertyFontDisplay);
    if (!value) {
        ec = SYNTAX_ERR;
        return;
    }
    m_backing->setLoadingBehavior(*value);
}

String FontFace::family() const
{
    const_cast<CSSFontFace&>(*m_backing).updateStyleIfNeeded();

    const auto& families = m_backing->families();
    if (!families.hasValue())
        return "normal";
    auto familiesUnrwapped = families.value();
    // FIXME: https://bugs.webkit.org/show_bug.cgi?id=196381 This is only here because CSSFontFace erroneously uses a list of values instead of a single value.
    // See consumeFontFamilyDescriptor() in CSSPropertyParser.cpp.
    if (familiesUnrwapped->length() == 1) {
        if (familiesUnrwapped->item(0)) {
            const CSSValue& item = *familiesUnrwapped->item(0);
            if (item.isPrimitiveValue()) {
                auto& primitiveValue = toCSSPrimitiveValue(item);
                if (primitiveValue.isFontFamily()) {
                    auto& fontFamily = primitiveValue.fontFamily();
                    return fontFamily.familyName;
                }
            }
        }
    }
    return familiesUnrwapped->cssText();
}

static Ref<CSSPrimitiveValue> fontNonKeywordWeightFromStyleValue(FontSelectionValue weight)
{
    return CSSValuePool::singleton().createValue(static_cast<float>(weight), CSSUnitType::CSS_NUMBER);
}

static Ref<CSSPrimitiveValue> fontWeightFromStyleValue(FontSelectionValue weight)
{
	if (auto value = fontWeightKeyword(weight))
		return CSSValuePool::singleton().createIdentifierValue(value.value());
    return fontNonKeywordWeightFromStyleValue(weight);
}

static Ref<CSSPrimitiveValue> fontNonKeywordStretchFromStyleValue(FontSelectionValue stretch)
{
    return CSSValuePool::singleton().createValue(static_cast<float>(stretch), CSSUnitType::CSS_PERCENTAGE);
}

static Ref<CSSPrimitiveValue> fontStretchFromStyleValue(FontSelectionValue stretch)
{
    if (auto keyword = fontStretchKeyword(stretch))
        return CSSValuePool::singleton().createIdentifierValue(keyword.value());
    return fontNonKeywordStretchFromStyleValue(stretch);
}

static Ref<CSSFontStyleValue> fontNonKeywordStyleFromStyleValue(FontSelectionValue italic)
{
    return CSSFontStyleValue::create(CSSValuePool::singleton().createIdentifierValue(CSSValueOblique), CSSValuePool::singleton().createValue(static_cast<float>(italic), CSSUnitType::CSS_DEG));
}

static Ref<CSSFontStyleValue> fontStyleFromStyleValue(FontSelectionValue italic)
{
    if (auto keyword = fontStyleKeyword(italic))
        return CSSFontStyleValue::create(CSSValuePool::singleton().createIdentifierValue(keyword.value()));
    return fontNonKeywordStyleFromStyleValue(italic);
}

String FontFace::style() const
{
    m_backing->updateStyleIfNeeded();
    const auto& styleWrapped = m_backing->italic();

    if (!styleWrapped.hasValue())
        return "normal";
    auto style = styleWrapped.value();
    RefPtr<CSSFontStyleValue> minimum = fontStyleFromStyleValue(style.minimum);
    RefPtr<CSSFontStyleValue> maximum = fontStyleFromStyleValue(style.maximum);

    if (minimum->equals(*maximum))
        return minimum->cssText();

    RefPtr<CSSFontStyleValue> minimumNonKeyword = fontNonKeywordStyleFromStyleValue(style.minimum);
    RefPtr<CSSFontStyleValue> maximumNonKeyword = fontNonKeywordStyleFromStyleValue(style.maximum);

    ASSERT(minimumNonKeyword->fontStyleValue->valueID() == CSSValueOblique);
    ASSERT(maximumNonKeyword->fontStyleValue->valueID() == CSSValueOblique);

    StringBuilder builder;
    builder.append(minimumNonKeyword->fontStyleValue->cssText());
    builder.append(' ');
    if (minimum->obliqueValue.get() == maximum->obliqueValue.get())
        builder.append(minimumNonKeyword->obliqueValue->cssText());
    else {
        builder.append(minimumNonKeyword->obliqueValue->cssText());
        builder.append(' ');
        builder.append(maximumNonKeyword->obliqueValue->cssText());
    }
    return builder.toString();
}

String FontFace::weight() const
{
    m_backing->updateStyleIfNeeded();
    const auto& weightWrapped = m_backing->weight();
    if (!weightWrapped.hasValue())
        return "normal";
    auto weight = weightWrapped.value();
    RefPtr<CSSPrimitiveValue> minimum = fontWeightFromStyleValue(weight.minimum);
    RefPtr<CSSPrimitiveValue> maximum = fontWeightFromStyleValue(weight.maximum);

    if (minimum->equals(*maximum))
        return minimum->cssText();

    RefPtr<CSSPrimitiveValue> minimumNonKeyword = fontNonKeywordWeightFromStyleValue(weight.minimum);
    RefPtr<CSSPrimitiveValue> maximumNonKeyword = fontNonKeywordWeightFromStyleValue(weight.maximum);

    StringBuilder builder;
    builder.append(minimumNonKeyword->cssText());
    builder.append(' ');
    builder.append(maximumNonKeyword->cssText());
    return builder.toString();
}

String FontFace::stretch() const
{
    m_backing->updateStyleIfNeeded();
    const auto& stretchWrapped = m_backing->stretch();
    if (!stretchWrapped.hasValue())
        return "normal";
    auto stretch = stretchWrapped.value();
    RefPtr<CSSPrimitiveValue> minimum = fontStretchFromStyleValue(stretch.minimum);
    RefPtr<CSSPrimitiveValue> maximum = fontStretchFromStyleValue(stretch.maximum);

    if (minimum->equals(*maximum))
        return minimum->cssText();

    RefPtr<CSSPrimitiveValue> minimumNonKeyword = fontNonKeywordStretchFromStyleValue(stretch.minimum);
    RefPtr<CSSPrimitiveValue> maximumNonKeyword = fontNonKeywordStretchFromStyleValue(stretch.maximum);

    StringBuilder builder;
    builder.append(minimumNonKeyword->cssText());
    builder.append(' ');
    builder.append(maximumNonKeyword->cssText());
    return builder.toString();
}

String FontFace::unicodeRange() const
{
    const_cast<CSSFontFace&>(*m_backing).updateStyleIfNeeded();
    const auto& rangesWrapped = m_backing->ranges();
    if (!rangesWrapped.hasValue())
        return "U+0-10FFFF";
    auto ranges = rangesWrapped.value();
    if (!ranges.size())
        return "U+0-10FFFF";
    auto values = CSSValueList::createCommaSeparated();
    for (auto& range : ranges)
        values->append(CSSUnicodeRangeValue::create(range.from(), range.to()));
    return values->cssText();
}

String FontFace::variant() const
{
    const_cast<CSSFontFace&>(*m_backing).updateStyleIfNeeded();
    const auto& variantSettingsWrapped = m_backing->variantSettings();
    if (!variantSettingsWrapped.hasValue())
        return "normal";
    auto variantSettings = variantSettingsWrapped.value();
    return computeFontVariant(variantSettings)->cssText();
}

String FontFace::featureSettings() const
{
    const_cast<CSSFontFace&>(*m_backing).updateStyleIfNeeded();
    const auto& featureSettingsWrapped = m_backing->featureSettings();
    if (!featureSettingsWrapped.hasValue())
        return "normal";
    auto featureSettings = featureSettingsWrapped.value();
    if (!featureSettings.size())
        return "normal";
    auto list = CSSValueList::createCommaSeparated();
    for (auto& feature : featureSettings)
        list->append(CSSFontFeatureValue::create(FontFeatureTag(feature.tag()), feature.value()));
    return list->cssText();
}

String FontFace::display() const
{
    m_backing->updateStyleIfNeeded();
    const auto& loadingBehaviorWrapped = m_backing->loadingBehavior();
    if (!loadingBehaviorWrapped.hasValue())
        return "auto";
    return CSSValuePool::singleton().createValue(loadingBehaviorWrapped.value())->cssText();
}

FontFace::LoadStatus FontFace::status() const
{
    switch (m_backing->status()) {
    case CSSFontFace::Status::Pending:
        return LoadStatus::Unloaded;
    case CSSFontFace::Status::Loading:
    	return LoadStatus::Loading;
    case CSSFontFace::Status::TimedOut:
        return LoadStatus::Error;
    case CSSFontFace::Status::Success:
        return LoadStatus::Loaded;
    case CSSFontFace::Status::Failure:
        return LoadStatus::Error;
    }
    ASSERT_NOT_REACHED();
    return LoadStatus::Error;
}

void FontFace::adopt(CSSFontFace& newFace)
{
    m_backing->removeClient(*this);
    m_backing = newFace;
    m_backing->addClient(*this);
    newFace.setWrapper(*this);
}

void FontFace::fontStateChanged(CSSFontFace& face, CSSFontFace::Status, CSSFontFace::Status newState)
{
    ASSERT_UNUSED(face, &face == m_backing.ptr());
    switch (newState) {
    case CSSFontFace::Status::Loading:
        break;
    case CSSFontFace::Status::TimedOut:
        if (!m_loadedPromise.isFulfilled())
            m_loadedPromise.reject(Exception(NETWORK_ERR));
		break;
	case CSSFontFace::Status::Success:
        // FIXME: This check should not be needed, but because FontFace's are sometimes adopted after they have already
        // gone through a load cycle, we can sometimes come back through here and try to resolve the promise again.
        if (!m_loadedPromise.isFulfilled())
            m_loadedPromise.resolve<FontFace>(*this);
		return;
	case CSSFontFace::Status::Failure:
        // FIXME: This check should not be needed, but because FontFace's are sometimes adopted after they have already
        // gone through a load cycle, we can sometimes come back through here and try to resolve the promise again.
        if (!m_loadedPromise.isFulfilled())
            m_loadedPromise.reject(Exception(NETWORK_ERR));
		return;
	case CSSFontFace::Status::Pending:
		ASSERT_NOT_REACHED();
		return;
	}
}

auto FontFace::loadForBindings() -> LoadedPromise&
{
    m_mayLoadedPromiseBeScriptObservable = true;
    m_backing->load();
    return m_loadedPromise;
}

auto FontFace::loadedForBindings() -> LoadedPromise&
{
    m_mayLoadedPromiseBeScriptObservable = true;
    return m_loadedPromise;
}

FontFace& FontFace::loadedPromiseResolve()
{
    return *this;
}

const char* FontFace::activeDOMObjectName() const
{
    return "FontFace";
}

bool FontFace::virtualHasPendingActivity() const
{
    return m_mayLoadedPromiseBeScriptObservable && !m_loadedPromise.isFulfilled();
}

}
