/*
 * Copyright (C) 2007, 2008 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ActiveDOMObject.h"
#include "CSSFontFace.h"
#include "CSSPropertyNames.h"
#include "DOMCoreException.h"
#include "DOMPromiseProxy.h"
#include "ExceptionCode.h"
#include "FontFaceLoadStatus.h"
#include "JSDOMPromiseDeferred.h"
#include <wtf/RefCounted.h>
#include <wtf/PassRefPtr.h>
#include <wtf/WeakPtr.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class CSSValue;

class FontFace FINAL : public RefCounted<FontFace>, public ActiveDOMObject, public CSSFontFace::Client {
public:
	struct Descriptors {
		Descriptors()
			: style(ASCIILiteral("normal"))
	    	, weight(ASCIILiteral("normal"))
			, stretch(ASCIILiteral("normal"))
	    	, unicodeRange(ASCIILiteral("U+0-10FFFF"))
	    	, variant(ASCIILiteral("normal"))
	    	, featureSettings(ASCIILiteral("normal"))
			, display(ASCIILiteral("auto"))
		{
		}
		Descriptors(const String& style, const String& weight, const String& stretch, const String& unicodeRange, const String& variant, const String& featureSettings, const String& display)
			: style(style)
			, weight(weight)
			, stretch(stretch)
			, unicodeRange(unicodeRange)
			, variant(variant)
			, featureSettings(featureSettings)
			, display(display)
		{
		}
	    String style;
	    String weight;
	    String stretch;
	    String unicodeRange;
	    String variant;
	    String featureSettings;
	    String display;
	};

    static PassRefPtr<FontFace> create(JSC::ExecState&, Document*, const String& family, JSC::JSValue source, const Descriptors&);
    static Ref<FontFace> create(ScriptExecutionContext*, CSSFontFace&);
	virtual ~FontFace();

    void setFamily(const String&, ExceptionCode&);
    void setStyle(const String&, ExceptionCode&);
    void setWeight(const String&, ExceptionCode&);
    void setStretch(const String&, ExceptionCode&);
    void setUnicodeRange(const String&, ExceptionCode&);
    void setVariant(const String&, ExceptionCode&);
    void setFeatureSettings(const String&, ExceptionCode&);
    void setDisplay(const String&, ExceptionCode&);

    String family() const;
    String style() const;
    String weight() const;
    String stretch() const;
    String unicodeRange() const;
    String variant() const;
    String featureSettings() const;
    String display() const;

    using LoadStatus = FontFaceLoadStatus;
    LoadStatus status() const;

    typedef DOMPromiseProxyWithResolveCallback<FontFace> LoadedPromise;
    LoadedPromise& loadedForBindings();
    LoadedPromise& loadForBindings();

    void adopt(CSSFontFace&);

    CSSFontFace& backing() { return *m_backing; }

    static PassRefPtr<CSSValue> parseString(const String&, CSSPropertyID);

    virtual void fontStateChanged(CSSFontFace&, CSSFontFace::Status oldState, CSSFontFace::Status newState) final;

    virtual void ref() final { RefCounted<FontFace>::ref(); }
    virtual void deref() final { RefCounted<FontFace>::deref(); }

private:
    explicit FontFace(CSSFontSelector&);
    explicit FontFace(ScriptExecutionContext*, CSSFontFace&);

    // ActiveDOMObject.
    virtual const char* activeDOMObjectName() const final;
    virtual bool virtualHasPendingActivity() const final;

    FontFace& loadedPromiseResolve();
    void setErrorState();

    RefPtr<CSSFontFace> m_backing;
    LoadedPromise m_loadedPromise;
    bool m_mayLoadedPromiseBeScriptObservable { false };
};

}
