/*
 * Copyright (C) 2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "FontFaceSet.h"

#include "Document.h"
#include "ExceptionCodeDescription.h"
#include "FontFace.h"
#include "FrameLoader.h"
#include "JSDOMBinding.h"
#include "JSDOMCoreException.h"
#include "JSFontFace.h"
#include "JSFontFaceSet.h"

namespace WebCore {

Ref<FontFaceSet> FontFaceSet::create(Document* document, const Vector<RefPtr<FontFace>>& initialFaces)
{
    Ref<FontFaceSet> result = adoptRef(*new FontFaceSet(*document, initialFaces));
    result->suspendIfNeeded();
    return result;
}

Ref<FontFaceSet> FontFaceSet::create(Document* document, CSSFontFaceSet& backing)
{
	Ref<FontFaceSet> result = adoptRef(*new FontFaceSet(*document, backing));
    result->suspendIfNeeded();
    return result;
}

inline FontFaceSet::FontFaceSet(Document& document, const Vector<RefPtr<FontFace>>& initialFaces)
    : ActiveDOMObject(document)
    , m_backing(CSSFontFaceSet::create())
	, m_readyPromise(*this, &FontFaceSet::readyPromiseResolve)
	, m_isDocumentLoaded(true)
{
    m_backing->addFontEventClient(*this);
    for (auto& face : initialFaces)
        add(*face);
}

inline FontFaceSet::FontFaceSet(Document& document, CSSFontFaceSet& backing)
    : ActiveDOMObject(document)
    , m_backing(backing)
	, m_readyPromise(*this, &FontFaceSet::readyPromiseResolve)
	, m_isDocumentLoaded(true)
{
    if (document.frame())
        m_isDocumentLoaded = document.loadEventFinished() && !document.processingLoadEvent();

    if (m_isDocumentLoaded && !backing.hasActiveFontFaces())
    	m_readyPromise.resolve<FontFaceSet>(*this);

    m_backing->addFontEventClient(*this);
}

FontFaceSet::~FontFaceSet()
{
}

FontFaceSet::Iterator::Iterator(FontFaceSet& set)
    : m_target(&set)
	, m_index(0)
{
}

RefPtr<FontFace> FontFaceSet::Iterator::next()
{
    if (m_index == m_target->size())
        return nullptr;
    return m_target->backing()[m_index++].wrapper(m_target->scriptExecutionContext());
}

FontFaceSet::PendingPromise::PendingPromise(Promise promise)
    : promise(promise)
	, hasReachedTerminalState(false)
{
}

FontFaceSet::PendingPromise::~PendingPromise()
{
}

bool FontFaceSet::has(FontFace& face) const
{
    return m_backing->hasFace(face.backing());
}

size_t FontFaceSet::size()
{
	Ref<CSSFontFaceSet> protect = m_backing;
    protect->updateStyleIfNeeded();
    return protect->faceCount();
}

FontFaceSet* FontFaceSet::add(FontFace& face)
{
    if (!m_backing->hasFace(face.backing()))
        m_backing->add(face.backing());
    return this;
}

bool FontFaceSet::remove(FontFace& face)
{
    bool result = m_backing->hasFace(face.backing());
    if (result)
        m_backing->remove(face.backing());
    return result;
}

void FontFaceSet::clear()
{
    auto facesPartitionIndex = m_backing->facesPartitionIndex();
    while (m_backing->faceCount() > facesPartitionIndex) {
        m_backing->remove(m_backing.get()[m_backing->faceCount() - 1]);
        ASSERT(m_backing->facesPartitionIndex() == facesPartitionIndex);
    }
}

template<>
inline void DeferredPromise::resolve<Vector<RefPtr<FontFace>>>(const Vector<RefPtr<FontFace>>& result)
{
    if (isSuspended())
        return;
    ASSERT(m_deferred);
    ASSERT(m_globalObject);
    auto& vm = m_globalObject->vm();
    JSC::ExecState* exec = m_globalObject->globalExec();
    JSC::JSLockHolder locker(vm);
    resolve(*exec, toJS(exec, m_globalObject.get(), result));
}

void FontFaceSet::load(const String& font, const String& text, RefPtr<DeferredPromise> promise)
{
    ExceptionCode ec = 0;
	Vector<RefPtr<CSSFontFace>> matchingFaces = m_backing->matchingFacesExcludingPreinstalledFonts(font, text, ec);
	if (ec) {
		promise->reject(ec);
		return;
	}

    if (matchingFaces.isEmpty()) {
        promise->resolve(Vector<RefPtr<FontFace>>());
        return;
    }

    for(int i = 0; i < matchingFaces.size(); ++i)
    	matchingFaces[i]->load();

    for(int i = 0; i < matchingFaces.size(); ++i) {
        if (matchingFaces[i]->status() == CSSFontFace::Status::Failure) {
            promise->reject<ExceptionCode>(NETWORK_ERR);
            return;
        }
    }

    RefPtr<PendingPromise> pendingPromise = PendingPromise::create(promise);
    bool waiting = false;

    for(int i = 0; i < matchingFaces.size(); ++i) {
    	RefPtr<CSSFontFace>& face = matchingFaces[i];
        pendingPromise->faces.append(face->wrapper(scriptExecutionContext()));
        if (face->status() == CSSFontFace::Status::Success)
            continue;
        waiting = true;
        ASSERT(face.get().existingWrapper());
        m_pendingPromises.add(face->existingWrapper(), Vector<RefPtr<PendingPromise>>()).iterator->second.append(pendingPromise);
    }

    if (!waiting)
        pendingPromise->promise.resolve(pendingPromise->faces);
}

bool FontFaceSet::check(const String& family, const String& text, ExceptionCode& ec)
{
    return m_backing->check(family, text, ec);
}

String FontFaceSet::status() const
{
    switch (m_backing->status()) {
    case CSSFontFaceSet::Status::Loading:
        return String("loading", String::ConstructFromLiteral);
    case CSSFontFaceSet::Status::Loaded:
        return String("loaded", String::ConstructFromLiteral);
    }
    ASSERT_NOT_REACHED();
    return String("loaded", String::ConstructFromLiteral);
}

bool FontFaceSet::canSuspendForDocumentSuspension() const
{
    return m_backing->status() == CSSFontFaceSet::Status::Loaded;
}

void FontFaceSet::faceFinished(CSSFontFace& face, CSSFontFace::Status newStatus)
{
    if (!face.existingWrapper())
        return;

    auto pendingPromises = m_pendingPromises.take(face.existingWrapper());
    if (pendingPromises.isEmpty())
        return;

    for (auto& pendingPromise : pendingPromises) {
        if (pendingPromise->hasReachedTerminalState)
            continue;
        if (newStatus == CSSFontFace::Status::Success) {
            if (pendingPromise->hasOneRef()) {
                pendingPromise->promise.resolve(pendingPromise->faces);
                pendingPromise->hasReachedTerminalState = true;
            }
        } else {
            ASSERT(newStatus == CSSFontFace::Status::Failure);
			pendingPromise->promise.reject(NETWORK_ERR);
			pendingPromise->hasReachedTerminalState = true;
        }
    }
}

void FontFaceSet::startedLoading()
{
    // FIXME: Fire a "loading" event asynchronously.
}

void FontFaceSet::documentDidFinishLoading()
{
    m_isDocumentLoaded = true;
    if (!m_backing->hasActiveFontFaces() && !m_readyPromise.isFulfilled())
        m_readyPromise.resolve(*this);
}

void FontFaceSet::completedLoading()
{
    if (m_isDocumentLoaded && !m_readyPromise.isFulfilled())
    	m_readyPromise.resolve<FontFaceSet>(*this);
}

FontFaceSet& FontFaceSet::readyPromiseResolve()
{
    return *this;
}

}
