/*
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies)
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "MediaQueryMatcher.h"

#include "Document.h"
#include "Element.h"
#include "EventNames.h"
#include "Frame.h"
#include "FrameView.h"
#include "MediaList.h"
#include "MediaQueryEvaluator.h"
#include "MediaQueryList.h"
#include "MediaQueryParserContext.h"
#include "NodeRenderStyle.h"
#include "StyleResolver.h"
#include "StyleScope.h"

namespace WebCore {

MediaQueryMatcher::MediaQueryMatcher(Document* document)
    : m_document(makeWeakPtr(document))
    , m_evaluationRound(1)
{
    ASSERT(m_document);
}

MediaQueryMatcher::~MediaQueryMatcher() = default;

void MediaQueryMatcher::documentDestroyed()
{
    m_document = nullptr;
    Vector<WeakPtr<MediaQueryList>> mediaQueryLists = WTFMove(m_mediaQueryLists);
    for (auto& mediaQueryList : mediaQueryLists) {
        if (mediaQueryList)
            mediaQueryList->detachFromMatcher();
    }
}

String MediaQueryMatcher::mediaType() const
{
    if (!m_document || !m_document->frame() || !m_document->frame()->view())
        return String();

    return m_document->frame()->view()->mediaType();
}

PassOwnPtr<RenderStyle> MediaQueryMatcher::documentElementUserAgentStyle() const
{
    if (!m_document || !m_document->frame())
        return nullptr;

    Element* documentElement = m_document->documentElement();
    if (!documentElement)
        return nullptr;

    return m_document->styleScope().resolver().styleForElement(*documentElement, Style::ResolutionContext(m_document->renderStyle()), MatchOnlyUserAgentRules).renderStyle.release();
}

bool MediaQueryMatcher::evaluate(const MediaQuerySet* media)
{
    if (!media)
        return false;

    OwnPtr<RenderStyle> style = documentElementUserAgentStyle();
    if (!style)
        return false;
    MediaQueryEvaluator evaluator(mediaType(), *m_document, style.get());
    return evaluator.evaluate(*media);
}

void MediaQueryMatcher::addMediaQueryList(MediaQueryList& list)
{
    ASSERT(!m_mediaQueryLists.contains(&list));
    m_mediaQueryLists.append(makeWeakPtr(&list));
}

void MediaQueryMatcher::removeMediaQueryList(MediaQueryList& list)
{
    m_mediaQueryLists.removeFirst(&list);
}

PassRefPtr<MediaQueryList> MediaQueryMatcher::matchMedia(const String& query)
{
    if (!m_document)
        return nullptr;

    RefPtr<MediaQuerySet> media = MediaQuerySet::create(query, MediaQueryParserContext(*m_document));
    //reportMediaQueryWarningIfNeeded(m_document.get(), media.ptr());
    bool matches = evaluate(media.get());
	return MediaQueryList::create(*m_document, *this, media.release(), matches);
}

void MediaQueryMatcher::evaluateAll(EventMode eventMode)
{
    ASSERT(m_document);

    ++m_evaluationRound;

    OwnPtr<RenderStyle> style = documentElementUserAgentStyle();
    if (!style)
        return;

    MediaQueryEvaluator evaluator(mediaType(), *m_document, style.get());

    Vector<WeakPtr<MediaQueryList>> mediaQueryLists = m_mediaQueryLists;
    for (auto& list : mediaQueryLists) {
    	if (RefPtr<MediaQueryList> protectedList = list.get()) {
    		protectedList->evaluate(evaluator, eventMode);
    		if (!m_document)
                break;
    	}
	}
}

} // namespace WebCore
