/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "CSSSelector.h"
#include <wtf/Forward.h>
#include <wtf/HashMap.h>
#include <wtf/HashSet.h>
#include <wtf/Optional.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/AtomicStringHash.h>

namespace WebCore {

class RuleData;
class StyleRule;

enum class MatchElement : uint8_t {
	Subject,
	Parent,
	Ancestor,
	DirectSibling,
	IndirectSibling,
	AnySibling,
	ParentSibling,
	AncestorSibling,
    ParentAnySibling,
    AncestorAnySibling,
	Host
};
constexpr unsigned matchElementCount = static_cast<unsigned>(MatchElement::Host) + 1;

struct RuleFeature {
	WTF_MAKE_FAST_ALLOCATED;
public:
    explicit RuleFeature(const RuleData&, Optional<MatchElement> matchElement = Nullopt);

    ALWAYS_INLINE static const CSSSelector* invalidationSelector(); // no implementation

    RefPtr<const StyleRule> styleRule;
    uint16_t selectorIndex; // Keep in sync with RuleData's selectorIndex size.
    uint16_t selectorListIndex; // Keep in sync with RuleData's selectorListIndex size.
    Optional<MatchElement> matchElement;
};
static_assert(sizeof(RuleFeature) <= 16, "RuleFeature is a frquently alocated object. Keep it small.");

struct RuleFeatureWithInvalidationSelector : public RuleFeature {
    RuleFeatureWithInvalidationSelector(const RuleData& data, Optional<MatchElement> matchElement = WTF::Nullopt, const CSSSelector* invalidationSelector = nullptr)
        : RuleFeature(data, matchElement)
        , m_invalidationSelector(invalidationSelector)
    { }
    ALWAYS_INLINE const CSSSelector* invalidationSelector() const { return m_invalidationSelector; }
private:
    const CSSSelector* m_invalidationSelector { nullptr };
};

struct RuleFeatureSet {
	WTF_MAKE_FAST_ALLOCATED;
public:
    RuleFeatureSet() = default;

    void add(const RuleFeatureSet&);
    void clear();
    void shrinkToFit();
    void collectFeatures(const RuleData&);
    void registerContentAttribute(const AtomicString&);

    HashSet<AtomicString> idsInRules;
    HashSet<AtomicString> idsMatchingAncestorsInRules;
    HashSet<AtomicString> attributeCanonicalLocalNamesInRules;
    HashSet<AtomicString> attributeLocalNamesInRules;
    HashSet<AtomicString> contentAttributeNamesInRules;
    Vector<RuleFeature> siblingRules;
    Vector<RuleFeature> uncommonAttributeRules;

    HashMap<AtomicString, OwnPtr<Vector<RuleFeature>>> classRules;
    HashMap<AtomicString, OwnPtr<Vector<RuleFeatureWithInvalidationSelector>>> attributeRules;
    HashMap<CSSSelector::PseudoClassType, OwnPtr<Vector<RuleFeature>>, WTF::IntHash<CSSSelector::PseudoClassType>, WTF::StrongEnumHashTraits<CSSSelector::PseudoClassType, unsigned>> pseudoClassRules;
	HashSet<AtomicString> classesAffectingHost;
    HashSet<AtomicString> attributesAffectingHost;
    HashSet<CSSSelector::PseudoClassType, WTF::IntHash<CSSSelector::PseudoClassType>, WTF::StrongEnumHashTraits<CSSSelector::PseudoClassType, unsigned>> pseudoClassesAffectingHost;

    bool usesFirstLineRules { false };
    bool usesFirstLetterRules { false };

private:
    static MatchElement computeNextMatchElement(MatchElement, CSSSelector::RelationType);
    static MatchElement computeSubSelectorMatchElement(MatchElement, const CSSSelector&);

    struct SelectorFeatures {
    	WTF_MAKE_FAST_ALLOCATED;
	public:
    	SelectorFeatures() : hasSiblingSelector(false) { }
        bool hasSiblingSelector;

        Vector<std::pair<AtomicString, MatchElement>, 32> classes;
        Vector<std::pair<const CSSSelector*, MatchElement>, 32> attributes;
        Vector<std::pair<CSSSelector::PseudoClassType, MatchElement>, 32> pseudoClasses;
    };
    void recursivelyCollectFeaturesFromSelector(SelectorFeatures&, const CSSSelector&, MatchElement = MatchElement::Subject);
};

} // namespace WebCore
