/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "MediaList.h"
#include "RuleData.h"
#include "RuleFeature.h"
#include <wtf/Forward.h>
#include <wtf/HashMap.h>
#include <wtf/HashSet.h>
#include <wtf/Optional.h>
#include <wtf/VectorHash.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/AtomicStringHash.h>

namespace WebCore {

class CSSSelector;
class ContainerNode;
class MediaQueryEvaluator;
class Node;
class StyleResolver;
class StyleSheetContents;

class RuleSet;

typedef Vector<RefPtr<const RuleSet>, 1> InvalidationRuleSetVector;

struct DynamicMediaQueryEvaluationChanges {
    enum class Type { InvalidateStyle, ResetStyle };
    Type type;
    InvalidationRuleSetVector invalidationRuleSets;

    explicit DynamicMediaQueryEvaluationChanges(Type type)
    	: type(type)
    {
    }
    explicit DynamicMediaQueryEvaluationChanges(Type type, const RuleSet& ruleSet)
    	: type(type)
    {
    	invalidationRuleSets.append(ruleSet);
    }
    void append(DynamicMediaQueryEvaluationChanges&& other)
    {
        type = std::max(type, other.type);
        if (type == Type::ResetStyle)
            invalidationRuleSets.clear();
        else
            invalidationRuleSets.appendVector(WTFMove(other.invalidationRuleSets));
    };
};

class RuleSet : public RefCounted<RuleSet> {
    WTF_MAKE_NONCOPYABLE(RuleSet);
public:
    static ALWAYS_INLINE PassRefPtr<RuleSet> create() { return adoptRef(new RuleSet); }

    ~RuleSet();

    typedef Vector<RuleData, 1> RuleDataVector;
    typedef HashMap<AtomicString, OwnPtr<RuleDataVector>> AtomRuleMap;

    struct DynamicMediaQueryRules {
    	DynamicMediaQueryRules() : requiresFullReset(false), result(true) { }
        Vector<Ref<const MediaQuerySet>> mediaQuerySets;
        Vector<size_t> affectedRulePositions;
        Vector<RuleFeature> ruleFeatures;
	    bool requiresFullReset;
        bool result;

        void shrinkToFit()
        {
            mediaQuerySets.shrinkToFit();
            affectedRulePositions.shrinkToFit();
            ruleFeatures.shrinkToFit();
        }
    };

    struct MediaQueryCollector {
    	MediaQueryCollector(const MediaQueryEvaluator& evaluator, bool canUseDynamicMediaQueryResolution)
    		: evaluator(evaluator)
    		, collectDynamic(canUseDynamicMediaQueryResolution)
    		, didMutateResolverWithinDynamicMediaQuery(false)
    		, hasViewportDependentMediaQueries(false)
    		{ }
        ~MediaQueryCollector();

        const MediaQueryEvaluator& evaluator;
        const bool collectDynamic;

        struct DynamicContext {
        	DynamicContext(const MediaQuerySet& set) : set(set) { }
            Ref<const MediaQuerySet> set;
            Vector<size_t> affectedRulePositions;
            Vector<RuleFeature> ruleFeatures;
        };
        Vector<DynamicContext> dynamicContextStack;

        Vector<DynamicMediaQueryRules> dynamicMediaQueryRules;
        bool didMutateResolverWithinDynamicMediaQuery;
        bool hasViewportDependentMediaQueries;

        bool pushAndEvaluate(const MediaQuerySet*);
        void pop(const MediaQuerySet*);
        void didMutateResolver();
        void addRuleIfNeeded(const RuleData&);
    };

    void addRulesFromSheet(StyleSheetContents&, const MediaQueryEvaluator&);
    void addRulesFromSheet(StyleSheetContents&, MediaQuerySet* sheetQuery, const MediaQueryEvaluator&, StyleResolver&);

    void addStyleRule(const StyleRule&, MediaQueryCollector&);
    void addRule(const StyleRule&, unsigned selectorIndex, unsigned selectorListIndex, MediaQueryCollector* = nullptr);
    void addPageRule(StyleRulePage&);
    void addToRuleSet(const AtomicString& key, AtomRuleMap&, const RuleData&);
    void shrinkToFit();
    void disableAutoShrinkToFit() { m_autoShrinkToFitEnabled = false; }

    bool hasViewportDependentMediaQueries() const { return m_hasViewportDependentMediaQueries; }

    Optional<DynamicMediaQueryEvaluationChanges> evaluateDynamicMediaQueryRules(const MediaQueryEvaluator&);

    const RuleFeatureSet& features() const { return m_features; }

    const RuleDataVector* idRules(const AtomicString& key) const { return m_idRules.inlineGet(key); }
    const RuleDataVector* classRules(const AtomicString& key) const { return m_classRules.inlineGet(key); }
    const RuleDataVector* tagRules(const AtomicString& key, bool isHTMLName) const;
    const RuleDataVector* shadowPseudoElementRules(const AtomicString& key) const { return m_shadowPseudoElementRules.get(key); }
    const RuleDataVector& linkPseudoClassRules() const { return m_linkPseudoClassRules; }
#if ENABLE(VIDEO_TRACK)
    const RuleDataVector& cuePseudoRules() const { return m_cuePseudoRules; }
#endif
#if ENABLE(SHADOW_DOM)
    const RuleDataVector& hostPseudoClassRules() const { return m_hostPseudoClassRules; }
    const RuleDataVector& slottedPseudoElementRules() const { return m_slottedPseudoElementRules; }
    const RuleDataVector& partPseudoElementRules() const { return m_partPseudoElementRules; }
#endif
    const RuleDataVector& focusPseudoClassRules() const { return m_focusPseudoClassRules; }
    const RuleDataVector& universalRules() const { return m_universalRules; }

    const Vector<StyleRulePage*>& pageRules() const { return m_pageRules; }

    unsigned ruleCount() const { return m_ruleCount; }

    bool hasShadowPseudoElementRules() const;
    bool hasHostPseudoClassRulesMatchingInShadowTree() const { return m_hasHostPseudoClassRulesMatchingInShadowTree; }

private:
    enum class AddRulesMode { Normal, ResolverMutationScan };
    void addRulesFromSheet(StyleSheetContents&, MediaQueryCollector&, StyleResolver*, AddRulesMode);
    void addChildRules(const Vector<RefPtr<StyleRuleBase>>&, MediaQueryCollector&, StyleResolver*, AddRulesMode);
    struct CollectedMediaQueryChanges {
    	CollectedMediaQueryChanges() : requiredFullReset(false) { }
		bool requiredFullReset;
		Vector<size_t> changedQueryIndexes;
		Vector<const Vector<RuleFeature>*> ruleFeatures;
	};
	CollectedMediaQueryChanges evaluateDynamicMediaQueryRules(const MediaQueryEvaluator&, size_t startIndex);

    template<typename Function> void traverseRuleDatas(Function&&);

    RuleSet();

    AtomRuleMap m_idRules;
    AtomRuleMap m_classRules;
    AtomRuleMap m_tagLocalNameRules;
    AtomRuleMap m_tagLowercaseLocalNameRules;
    AtomRuleMap m_shadowPseudoElementRules;
    RuleDataVector m_linkPseudoClassRules;
#if ENABLE(VIDEO_TRACK)
    RuleDataVector m_cuePseudoRules;
#endif
#if ENABLE(SHADOW_DOM)
    RuleDataVector m_hostPseudoClassRules;
    RuleDataVector m_slottedPseudoElementRules;
    RuleDataVector m_partPseudoElementRules;
#endif
    RuleDataVector m_focusPseudoClassRules;
    RuleDataVector m_universalRules;
    Vector<StyleRulePage*> m_pageRules;
    RuleFeatureSet m_features;
    Vector<DynamicMediaQueryRules> m_dynamicMediaQueryRules;
    HashMap<Vector<size_t>, RefPtr<const RuleSet>> m_mediaQueryInvalidationRuleSetCache;
    unsigned m_ruleCount;
    bool m_hasHostPseudoClassRulesMatchingInShadowTree;
    bool m_autoShrinkToFitEnabled;
    bool m_hasViewportDependentMediaQueries;
};

ALWAYS_INLINE const RuleSet::RuleDataVector* RuleSet::tagRules(const AtomicString& key, bool isHTMLName) const
{
    const AtomRuleMap* tagRules;
    if (isHTMLName)
        tagRules = &m_tagLowercaseLocalNameRules;
    else
        tagRules = &m_tagLocalNameRules;
    return tagRules->inlineGet(key);
}

} // namespace WebCore

namespace WTF {
template <> struct VectorTraits<WebCore::RuleData> : SimpleClassVectorTraits { };
}

