/*
 * Copyright (C) 2013 Google Inc. All rights reserved.
 * Copyright (C) 2014 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "BasicShapeFunctions.h"
#include "CSSCalculationValue.h"
#include "CSSContentDistributionValue.h"
#include "CSSFontFeatureValue.h"
#include "CSSFontStyleValue.h"
#include "CSSFunctionValue.h"
#include "CSSGridAutoRepeatValue.h"
#include "CSSGridIntegerRepeatValue.h"
#include "CSSGridLineNamesValue.h"
#include "CSSGridTemplateAreasValue.h"
#include "CSSImageGeneratorValue.h"
#include "CSSImageSetValue.h"
#include "CSSImageValue.h"
#include "CSSPrimitiveValue.h"
#include "CSSPrimitiveValueMappings.h"
#include "CSSRayValue.h"
#include "CSSReflectValue.h"
#include "FontSelectionValueInlines.h"
#include "Frame.h"
#include "LayoutUnit.h"
#include "Length.h"
//#include "LengthRepeat.h"
#include "Pair.h"
#include "QuotesData.h"
#include "SVGPathElement.h"
#include "SVGURIReference.h"
#include "Settings.h"
#include "StyleBuilderState.h"
//#include "StyleScrollSnapPoints.h"
#include "TabSize.h"
#include "TransformFunctions.h"
//#include "WebKitCSSResourceValue.h"
#include <wtf/MathExtras.h>
#include <wtf/Optional.h>

namespace WebCore {

// FIXME: Some of those functions assume the CSS parser only allows valid CSSValue types.
// This might not be true if we pass the CSSValue from js via CSS Typed OM.

class StyleBuilderConverter {
public:
    static Length convertLength(const Style::BuilderState&, const CSSValue&);
    static Length convertLengthOrAuto(const Style::BuilderState&, const CSSValue&);
    static Length convertLengthOrAutoOrContent(const Style::BuilderState&, const CSSValue&);
    static Length convertLengthSizing(const Style::BuilderState&, const CSSValue&);
    static Length convertLengthMaxSizing(const Style::BuilderState&, const CSSValue&);
    static TabSize convertTabSize(const Style::BuilderState&, const CSSValue&);
    template <typename T> static T convertComputedLength(Style::BuilderState&, const CSSValue&);
    template <typename T> static T convertLineWidth(Style::BuilderState&, const CSSValue&);
    static float convertSpacing(Style::BuilderState&, CSSValue&);
    static LengthSize convertRadius(Style::BuilderState&, CSSValue&);
    static LengthPoint convertPosition(Style::BuilderState&, const CSSValue&);
    static LengthPoint convertPositionOrAuto(Style::BuilderState&, const CSSValue&);
    static TextDecoration convertTextDecoration(Style::BuilderState&, CSSValue&);
    template <typename T> static T convertNumber(Style::BuilderState&, const CSSValue&);
    template <typename T> static T convertNumberOrAuto(Style::BuilderState&, const CSSValue&);
    static short convertWebkitHyphenateLimitLines(Style::BuilderState&, CSSValue&);
    template <CSSPropertyID property> static NinePieceImage convertBorderImage(Style::BuilderState&, CSSValue&);
    template <CSSPropertyID property> static NinePieceImage convertBorderMask(Style::BuilderState&, CSSValue&);
    template <CSSPropertyID property> static PassRefPtr<StyleImage> convertStyleImage(Style::BuilderState&, CSSValue&);
    static TransformOperations convertTransform(Style::BuilderState&, CSSValue&);
    static PassRefPtr<RotateTransformOperation> convertRotate(Style::BuilderState&, const CSSValue&);
    static PassRefPtr<ScaleTransformOperation> convertScale(Style::BuilderState&, const CSSValue&);
    static PassRefPtr<TranslateTransformOperation> convertTranslate(Style::BuilderState&, const CSSValue&);
    static String convertString(Style::BuilderState&, CSSValue&);
    static String convertStringOrAuto(Style::BuilderState&, CSSValue&);
    static String convertStringOrNone(Style::BuilderState&, CSSValue&);
    static TextEmphasisPosition convertTextEmphasisPosition(Style::BuilderState&, CSSValue&);
    static TextAlignMode convertTextAlign(Style::BuilderState&, CSSValue&);
    static TextAlignLast convertTextAlignLast(Style::BuilderState&, const CSSValue&);
    static PassRefPtr<ClipPathOperation> convertPathOperation(Style::BuilderState&, const CSSValue&);
    static EResize convertResize(Style::BuilderState&, CSSValue&);
    static int convertMarqueeRepetition(Style::BuilderState&, CSSValue&);
    static int convertMarqueeSpeed(Style::BuilderState&, CSSValue&);
    static PassRefPtr<QuotesData> convertQuotes(Style::BuilderState&, CSSValue&);
    static TextUnderlinePosition convertTextUnderlinePosition(Style::BuilderState&, const CSSValue&);
	static TextUnderlineOffset convertTextUnderlineOffset(Style::BuilderState&, const CSSValue&);
	static TextDecorationThickness convertTextDecorationThickness(Style::BuilderState&, const CSSValue&);
	static PassRefPtr<StyleReflection> convertReflection(Style::BuilderState&, CSSValue&);
    static IntSize convertInitialLetter(Style::BuilderState&, CSSValue&);
    static float convertTextStrokeWidth(Style::BuilderState&, CSSValue&);
    static LineBoxContain convertLineBoxContain(Style::BuilderState&, CSSValue&);
    static TextDecorationSkip convertTextDecorationSkip(Style::BuilderState&, CSSValue&);
    static PassRefPtr<ShapeValue> convertShapeValue(Style::BuilderState&, CSSValue&);
#if ENABLE(CSS_GRID_LAYOUT)
    static GridTrackSize convertGridTrackSize(Style::BuilderState&, const CSSValue&);
    static Vector<GridTrackSize> convertGridTrackSizeList(Style::BuilderState&, CSSValue&);
    static Optional<GridPosition> convertGridPosition(Style::BuilderState&, CSSValue&);
    static GridAutoFlow convertGridAutoFlow(Style::BuilderState&, CSSValue&);
#endif // ENABLE(CSS_GRID_LAYOUT)
    static Optional<Length> convertWordSpacing(Style::BuilderState&, CSSValue&);
    static Optional<float> convertPerspective(Style::BuilderState&, CSSValue&);
    static Optional<Length> convertMarqueeIncrement(Style::BuilderState&, CSSValue&);
    static Optional<FilterOperations> convertFilterOperations(Style::BuilderState&, CSSValue&);
    static FontFeatureSettings convertFontFeatureSettings(Style::BuilderState&, CSSValue&);
    static FontSelectionValue convertFontWeightFromValue(const CSSValue&);
    static FontSelectionValue convertFontStretchFromValue(const CSSValue&);
    static FontSelectionValue convertFontStyleFromValue(const CSSValue&);
    static FontSelectionValue convertFontWeight(Style::BuilderState&, const CSSValue&);
    static FontSelectionValue convertFontStretch(Style::BuilderState&, const CSSValue&);
    static FontSelectionValue convertFontStyle(Style::BuilderState&, const CSSValue&);
    static SVGLengthValue convertSVGLengthValue(Style::BuilderState&, const CSSValue&);
    static Vector<SVGLengthValue> convertSVGLengthVector(Style::BuilderState&, const CSSValue&);
    static Vector<SVGLengthValue> convertStrokeDashArray(Style::BuilderState&, const CSSValue&);
    static PaintOrder convertPaintOrder(Style::BuilderState&, const CSSValue&);
    static float convertOpacity(Style::BuilderState&, CSSValue&);
    static String convertSVGURIReference(Style::BuilderState&, CSSValue&);
    static Color convertSVGColor(Style::BuilderState&, const CSSValue&);
    static EGlyphOrientation convertGlyphOrientation(Style::BuilderState&, CSSValue&);
    static EGlyphOrientation convertGlyphOrientationOrAuto(Style::BuilderState&, CSSValue&);
    static Optional<Length> convertLineHeight(Style::BuilderState&, CSSValue&, float multiplier = 1.f);
    static FontSynthesis convertFontSynthesis(Style::BuilderState&, CSSValue&);

    static StyleSelfAlignmentData convertSelfOrDefaultAlignmentData(Style::BuilderState&, CSSValue&);
    static StyleContentAlignmentData convertContentAlignmentData(Style::BuilderState&, CSSValue&);

    static BreakBetween convertPageBreakBetween(Style::BuilderState&, const CSSValue&);
    static BreakInside convertPageBreakInside(Style::BuilderState&, const CSSValue&);
    static BreakBetween convertColumnBreakBetween(Style::BuilderState&, const CSSValue&);
    static BreakInside convertColumnBreakInside(Style::BuilderState&, const CSSValue&);

    static HangingPunctuation convertHangingPunctuation(Style::BuilderState&, CSSValue&);

    static Length convertPositionComponentX(Style::BuilderState&, const CSSValue&);
    static Length convertPositionComponentY(Style::BuilderState&, const CSSValue&);

    static GapLength convertGapLength(Style::BuilderState&, const CSSValue&);

private:
    friend class StyleBuilderCustom;

    static Length convertToRadiusLength(CSSToLengthConversionData&, CSSPrimitiveValue&);
    static TextEmphasisPosition valueToEmphasisPosition(CSSPrimitiveValue&);
    static TextDecorationSkip valueToDecorationSkip(const CSSPrimitiveValue&);
#if ENABLE(CSS_GRID_LAYOUT)
    static GridLength createGridTrackBreadth(const CSSPrimitiveValue&, Style::BuilderState&);
    static GridTrackSize createGridTrackSize(const CSSValue&, Style::BuilderState&);
    struct TracksData;
    static bool createGridTrackList(CSSValue&, TracksData&, Style::BuilderState&);
    static bool createGridPosition(CSSValue&, GridPosition&);
    static void createImplicitNamedGridLinesFromGridArea(const NamedGridAreaMap&, NamedGridLinesMap&, GridTrackSizingDirection);
#endif // ENABLE(CSS_GRID_LAYOUT)
    static CSSToLengthConversionData csstoLengthConversionDataWithTextZoomFactor(Style::BuilderState&);

    static Length convertTo100PercentMinusLength(const Length&);
    template<CSSValueID, CSSValueID> static Length convertPositionComponent(Style::BuilderState&, const CSSPrimitiveValue&);
};

inline Length StyleBuilderConverter::convertLength(const Style::BuilderState& builderState, const CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    CSSToLengthConversionData conversionData = builderState.useSVGZoomRulesForLength() ?
		builderState.cssToLengthConversionData().copyWithAdjustedZoom(1.0f)
        : builderState.cssToLengthConversionData();

    if (primitiveValue.isLength()) {
        Length length = primitiveValue.computeLength<Length>(conversionData);
        length.setHasQuirk(primitiveValue.isQuirkValue());
        return length;
    }

    if (primitiveValue.isPercentage())
        return Length(primitiveValue.doubleValue(), LengthType::Percent);

    if (primitiveValue.isCalculatedPercentageWithLength())
        return Length(primitiveValue.cssCalcValue()->createCalculationValue(conversionData));

    ASSERT_NOT_REACHED();
    return Length(0, LengthType::Fixed);
}

inline Length StyleBuilderConverter::convertLengthOrAuto(const Style::BuilderState& builderState, const CSSValue& value)
{
    if (toCSSPrimitiveValue(value).getValueID() == CSSValueAuto)
        return Length(LengthType::Auto);
    return convertLength(builderState, value);
}

inline Length StyleBuilderConverter::convertLengthSizing(const Style::BuilderState& builderState, const CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    switch (primitiveValue.getValueID()) {
    case CSSValueInvalid:
        return convertLength(builderState, value);
    case CSSValueIntrinsic:
        return Length(LengthType::Intrinsic);
    case CSSValueMinIntrinsic:
        return Length(LengthType::MinIntrinsic);
    case CSSValueMinContent:
    case CSSValueWebkitMinContent:
        return Length(LengthType::MinContent);
    case CSSValueMaxContent:
    case CSSValueWebkitMaxContent:
        return Length(LengthType::MaxContent);
    case CSSValueWebkitFillAvailable:
        return Length(LengthType::FillAvailable);
    case CSSValueFitContent:
    case CSSValueWebkitFitContent:
        return Length(LengthType::FitContent);
    case CSSValueAuto:
        return Length(LengthType::Auto);
    case CSSValueContent:
        return Length(LengthType::Content);
    default:
        ASSERT_NOT_REACHED();
        return Length();
    }
}

inline Length StyleBuilderConverter::convertLengthMaxSizing(const Style::BuilderState& builderState, const CSSValue& value)
{
    if (toCSSPrimitiveValue(value).getValueID() == CSSValueNone)
        return Length(LengthType::Undefined);
    return convertLengthSizing(builderState, value);
}

inline TabSize StyleBuilderConverter::convertTabSize(const Style::BuilderState& builderState, const CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.isNumber())
        return TabSize(primitiveValue.floatValue(), SpaceValueType);
    return TabSize(primitiveValue.computeLength<float>(builderState.cssToLengthConversionData()), LengthValueType);
}

template <typename T>
inline T StyleBuilderConverter::convertComputedLength(Style::BuilderState& builderState, const CSSValue& value)
{
    return toCSSPrimitiveValue(value).computeLength<T>(builderState.cssToLengthConversionData());
}

template <typename T>
inline T StyleBuilderConverter::convertLineWidth(Style::BuilderState& builderState, const CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    switch (primitiveValue.getValueID()) {
    case CSSValueThin:
        return 1;
    case CSSValueMedium:
        return 3;
    case CSSValueThick:
        return 5;
    case CSSValueInvalid: {
        // Any original result that was >= 1 should not be allowed to fall below 1.
        // This keeps border lines from vanishing.
        T result = convertComputedLength<T>(builderState, value);
        if (builderState.style().effectiveZoom() < 1.0f && result < 1.0) {
            T originalLength = primitiveValue.computeLength<T>(builderState.cssToLengthConversionData().copyWithAdjustedZoom(1.0));
            if (originalLength >= 1.0)
                return 1;
        }
        float minimumLineWidth = 1 / builderState.document().deviceScaleFactor();
        if (result > 0 && result < minimumLineWidth)
            return minimumLineWidth;
        return floorToDevicePixel(result, builderState.document().deviceScaleFactor());
    }
    default:
        ASSERT_NOT_REACHED();
        return 0;
    }
}

inline float StyleBuilderConverter::convertSpacing(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID() == CSSValueNormal)
        return 0.f;

    CSSToLengthConversionData conversionData = builderState.useSVGZoomRulesForLength() ?
		builderState.cssToLengthConversionData().copyWithAdjustedZoom(1.0f)
        : builderState.cssToLengthConversionData();
    return primitiveValue.computeLength<float>(conversionData);
}

inline Length StyleBuilderConverter::convertToRadiusLength(CSSToLengthConversionData& conversionData, CSSPrimitiveValue& value)
{
    if (value.isPercentage())
        return Length(value.doubleValue(), LengthType::Percent);
    if (value.isCalculatedPercentageWithLength())
        return Length(value.cssCalcValue()->createCalculationValue(conversionData));
    Length length = value.computeLength<Length>(conversionData);
	if (length.isNegative())
		return Length(0, LengthType::Fixed);
	return length;
}

inline LengthSize StyleBuilderConverter::convertRadius(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    Pair* pair = primitiveValue.pairValue();
    if (!pair || !pair->first() || !pair->second())
        return LengthSize(Length(0, LengthType::Fixed), Length(0, LengthType::Fixed));

    CSSToLengthConversionData conversionData = builderState.cssToLengthConversionData();
    Length radiusWidth = convertToRadiusLength(conversionData, *pair->first());
    Length radiusHeight = convertToRadiusLength(conversionData, *pair->second());

    ASSERT(!radiusWidth.isNegative());
    ASSERT(!radiusHeight.isNegative());
    if (radiusWidth.isZero() || radiusHeight.isZero())
        return LengthSize(Length(0, LengthType::Fixed), Length(0, LengthType::Fixed));

    return LengthSize(radiusWidth, radiusHeight);
}

inline Length StyleBuilderConverter::convertTo100PercentMinusLength(const Length& length)
{
    if (length.isPercent())
        return Length(100 - length.value(), LengthType::Percent);

    // Turn this into a calc expression: calc(100% - length)
    Vector<OwnPtr<CalcExpressionNode>> lengths;
    lengths.reserveInitialCapacity(2);
    lengths.uncheckedAppend(adoptPtr(new CalcExpressionLength(Length(100, LengthType::Percent))));
    lengths.uncheckedAppend(adoptPtr(new CalcExpressionLength(length)));
    OwnPtr<CalcExpressionOperation> op = adoptPtr(new CalcExpressionOperation(WTFMove(lengths), CalcOperator::Subtract));
    return Length(CalculationValue::create(op.release(), ValueRangeAll));
}

inline Length StyleBuilderConverter::convertPositionComponentX(Style::BuilderState& builderState, const CSSValue& value)
{
    return convertPositionComponent<CSSValueLeft, CSSValueRight>(builderState, toCSSPrimitiveValue(value));
}

inline Length StyleBuilderConverter::convertPositionComponentY(Style::BuilderState& builderState, const CSSValue& value)
{
    return convertPositionComponent<CSSValueTop, CSSValueBottom>(builderState, toCSSPrimitiveValue(value));
}

template <CSSValueID cssValueFor0, CSSValueID cssValueFor100>
inline Length StyleBuilderConverter::convertPositionComponent(Style::BuilderState& builderState, const CSSPrimitiveValue& value)
{
    Length length;

    auto* lengthValue = &value;
    bool relativeToTrailingEdge = false;

    if (value.isPair()) {
        auto& first = *value.pairValue()->first();
        if (first.getValueID() == CSSValueRight || first.getValueID() == CSSValueBottom)
            relativeToTrailingEdge = true;

        lengthValue = value.pairValue()->second();
    }

    if (value.isValueID()) {
        switch (value.valueID()) {
            case cssValueFor0:
                return Length(0, LengthType::Percent);
            case cssValueFor100:
                return Length(100, LengthType::Percent);
            case CSSValueCenter:
                return Length(50, LengthType::Percent);
            default:
                ASSERT_NOT_REACHED();
        }
    }

    length = convertLength(builderState, *lengthValue);

    if (relativeToTrailingEdge)
        length = convertTo100PercentMinusLength(length);

    return length;
}

inline LengthPoint StyleBuilderConverter::convertPosition(Style::BuilderState& builderState, const CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    Pair* pair = primitiveValue.pairValue();
    if (!pair || !pair->first() || !pair->second())
        return RenderStyle::initialObjectPosition();

    Length lengthX = convertPositionComponent<CSSValueLeft, CSSValueRight>(builderState, *pair->first());
    Length lengthY = convertPositionComponent<CSSValueTop, CSSValueBottom>(builderState, *pair->second());

    return LengthPoint(lengthX, lengthY);
}

inline LengthPoint StyleBuilderConverter::convertPositionOrAuto(Style::BuilderState& builderState, const CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    if (primitiveValue.isPair())
        return convertPosition(builderState, value);

    return LengthPoint(Length(LengthType::Auto), Length(LengthType::Auto));
}

inline TextDecoration StyleBuilderConverter::convertTextDecoration(Style::BuilderState&, CSSValue& value)
{
    TextDecoration result = RenderStyle::initialTextDecoration();
    if (value.isValueList()) {
        for (auto& currentValue : toCSSValueList(value))
            result |= toCSSPrimitiveValue(currentValue.get());
    }
    return result;
}

template <typename T>
inline T StyleBuilderConverter::convertNumber(Style::BuilderState&, const CSSValue& value)
{
    return toCSSPrimitiveValue(value).getValue<T>(CSSUnitType::CSS_NUMBER);
}

template <typename T>
inline T StyleBuilderConverter::convertNumberOrAuto(Style::BuilderState& builderState, const CSSValue& value)
{
    if (toCSSPrimitiveValue(value).getValueID() == CSSValueAuto)
        return -1;
    return convertNumber<T>(builderState, value);
}

inline short StyleBuilderConverter::convertWebkitHyphenateLimitLines(Style::BuilderState&, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID() == CSSValueNoLimit)
        return -1;
    return primitiveValue.getValue<short>(CSSUnitType::CSS_NUMBER);
}

template <CSSPropertyID property>
inline NinePieceImage StyleBuilderConverter::convertBorderImage(Style::BuilderState& builderState, CSSValue& value)
{
    NinePieceImage image;
    builderState.styleMap().mapNinePieceImage(property, &value, image);
    return image;
}

template <CSSPropertyID property>
inline NinePieceImage StyleBuilderConverter::convertBorderMask(Style::BuilderState& builderState, CSSValue& value)
{
    NinePieceImage image(NinePieceImage::Type::Mask);
    builderState.styleMap().mapNinePieceImage(property, &value, image);
    return image;
}

template <CSSPropertyID property>
inline PassRefPtr<StyleImage> StyleBuilderConverter::convertStyleImage(Style::BuilderState& builderState, CSSValue& value)
{
    return builderState.createStyleImage(value);
}

inline TransformOperations StyleBuilderConverter::convertTransform(Style::BuilderState& builderState, CSSValue& value)
{
    TransformOperations operations;
    transformsForValue(value, builderState.cssToLengthConversionData(), operations);
    return operations;
}

inline PassRefPtr<TranslateTransformOperation> StyleBuilderConverter::convertTranslate(Style::BuilderState& builderState, const CSSValue& value)
{
    return translateForValue(value, builderState.cssToLengthConversionData());
}

inline PassRefPtr<RotateTransformOperation> StyleBuilderConverter::convertRotate(Style::BuilderState&, const CSSValue& value)
{
    return rotateForValue(value);
}

inline PassRefPtr<ScaleTransformOperation> StyleBuilderConverter::convertScale(Style::BuilderState&, const CSSValue& value)
{
    return scaleForValue(value);
}

inline String StyleBuilderConverter::convertString(Style::BuilderState&, CSSValue& value)
{
    return toCSSPrimitiveValue(value).stringValue();
}

inline String StyleBuilderConverter::convertStringOrAuto(Style::BuilderState& builderState, CSSValue& value)
{
    if (toCSSPrimitiveValue(value).getValueID() == CSSValueAuto)
        return nullAtom;
    return convertString(builderState, value);
}

inline String StyleBuilderConverter::convertStringOrNone(Style::BuilderState& builderState, CSSValue& value)
{
    if (toCSSPrimitiveValue(value).getValueID() == CSSValueNone)
        return nullAtom;
    return convertString(builderState, value);
}

inline TextEmphasisPosition StyleBuilderConverter::valueToEmphasisPosition(CSSPrimitiveValue& primitiveValue)
{
    ASSERT(primitiveValue.isValueID());

    switch (primitiveValue.getValueID()) {
    case CSSValueOver:
        return TextEmphasisPositionOver;
    case CSSValueUnder:
        return TextEmphasisPositionUnder;
    case CSSValueLeft:
        return TextEmphasisPositionLeft;
    case CSSValueRight:
        return TextEmphasisPositionRight;
    default:
        break;
    }

    ASSERT_NOT_REACHED();
    return RenderStyle::initialTextEmphasisPosition();
}

inline TextEmphasisPosition StyleBuilderConverter::convertTextEmphasisPosition(Style::BuilderState&, CSSValue& value)
{
    if (value.isPrimitiveValue())
        return valueToEmphasisPosition(toCSSPrimitiveValue(value));

    TextEmphasisPosition position = 0;
    for (auto& currentValue : toCSSValueList(value))
        position |= valueToEmphasisPosition(toCSSPrimitiveValue(currentValue.get()));
    return position;
}

inline TextAlignMode StyleBuilderConverter::convertTextAlign(Style::BuilderState& builderState, CSSValue& value)
{
    const auto& primitiveValue = toCSSPrimitiveValue(value);
    ASSERT(primitiveValue.isValueID());

    const auto& parentStyle = builderState.parentStyle();

    // User agents are expected to have a rule in their user agent stylesheet that matches th elements that have a parent
    // node whose computed value for the 'text-align' property is its initial value, whose declaration block consists of
    // just a single declaration that sets the 'text-align' property to the value 'center'.
    // https://html.spec.whatwg.org/multipage/rendering.html#rendering
    if (primitiveValue.valueID() == CSSValueInternalThCenter) {
        if (parentStyle.textAlign() == RenderStyle::initialTextAlign())
            return TextAlignMode::Center;
        return parentStyle.textAlign();
    }

    if (primitiveValue.valueID() == CSSValueWebkitMatchParent || primitiveValue.valueID() == CSSValueMatchParent) {
        const auto* element = builderState.element();

        if (element && element == builderState.document().documentElement())
            return TextAlignMode::Start;
		if (parentStyle.textAlign() == TextAlignMode::Start)
			return parentStyle.isLeftToRightDirection() ? TextAlignMode::Left : TextAlignMode::Right;
		if (parentStyle.textAlign() == TextAlignMode::End)
			return parentStyle.isLeftToRightDirection() ? TextAlignMode::Right : TextAlignMode::Left;

		return parentStyle.textAlign();
    }

    return primitiveValue;
}

inline TextAlignLast StyleBuilderConverter::convertTextAlignLast(Style::BuilderState& builderState, const CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    ASSERT(primitiveValue.isValueID());

    if (primitiveValue.valueID() != CSSValueMatchParent)
        return primitiveValue;

    auto& parentStyle = builderState.parentStyle();
    if (parentStyle.textAlignLast() == TextAlignLastStart)
        return parentStyle.isLeftToRightDirection() ? TextAlignLastLeft : TextAlignLastRight;
    if (parentStyle.textAlignLast() == TextAlignLastEnd)
        return parentStyle.isLeftToRightDirection() ? TextAlignLastRight : TextAlignLastLeft;
    return parentStyle.textAlignLast();
}

inline PassRefPtr<ClipPathOperation> StyleBuilderConverter::convertPathOperation(Style::BuilderState& builderState, const CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        auto& primitiveValue = toCSSPrimitiveValue(value);
        if (primitiveValue.primitiveType() == CSSUnitType::CSS_URI) {
            String cssURLValue = primitiveValue.stringValue();
            String fragment = SVGURIReference::fragmentIdentifierFromIRIString(cssURLValue, builderState.document());
			// FIXME: It doesn't work with external SVG references (see https://bugs.webkit.org/show_bug.cgi?id=126133)
            const TreeScope* treeScope = nullptr;
			if (builderState.element())
				treeScope = &builderState.element()->treeScopeForSVGReferences();
			else
				treeScope = &builderState.document();
			auto target = SVGURIReference::targetElementFromIRIString(cssURLValue, *treeScope);
            return ReferencePathOperation::create(cssURLValue, fragment, isSVGElement(target.element.get()) ? toSVGElement(target.element.get()) : nullptr);
        }
        ASSERT(primitiveValue.getValueID() == CSSValueNone);
        return nullptr;
    }

    if (value.isRayValue()) {
        auto& rayValue = toCSSRayValue(value);

        RayPathOperation::Size size = RayPathOperation::Size::ClosestCorner;
        switch (rayValue.size()->valueID()) {
        case CSSValueClosestCorner:
            size = RayPathOperation::Size::ClosestCorner;
            break;
        case CSSValueClosestSide:
            size = RayPathOperation::Size::ClosestSide;
            break;
        case CSSValueFarthestCorner:
            size = RayPathOperation::Size::FarthestCorner;
            break;
        case CSSValueFarthestSide:
            size = RayPathOperation::Size::FarthestSide;
            break;
        case CSSValueSides:
            size = RayPathOperation::Size::Sides;
            break;
        default:
            ASSERT_NOT_REACHED();
            return nullptr;
        }

        return RayPathOperation::create(rayValue.angle()->computeDegrees(), size, rayValue.isContaining());
    }

    CSSBoxType referenceBox = CSSBoxType::BoxMissing;
    RefPtr<ClipPathOperation> operation;

    auto processSingleValue = [&](const CSSValue& singleValue) {
		ASSERT(!is<CSSValueList>(singleValue));
        auto& primitiveValue = toCSSPrimitiveValue(singleValue);
        if (primitiveValue.isShape()) {
            ASSERT(!operation);
            operation = ShapeClipPathOperation::create(basicShapeForValue(builderState.cssToLengthConversionData(),
				*primitiveValue.shapeValue(),
				builderState.style().effectiveZoom()));
        } else {
            ASSERT(primitiveValue.getValueID() == CSSValueContentBox
                   || primitiveValue.getValueID() == CSSValueBorderBox
                   || primitiveValue.getValueID() == CSSValuePaddingBox
                   || primitiveValue.getValueID() == CSSValueMarginBox
                   || primitiveValue.getValueID() == CSSValueFillBox
                   || primitiveValue.getValueID() == CSSValueStrokeBox
                   || primitiveValue.getValueID() == CSSValueViewBox);
            ASSERT(referenceBox == CSSBoxType::BoxMissing);
            referenceBox = primitiveValue;
        }
    };
    if (value.isValueList()) {
        for (auto& currentValue : toCSSValueList(value))
            processSingleValue(currentValue);
    } else
        processSingleValue(value);

    if (operation)
        toShapeClipPathOperation(*operation).setReferenceBox(referenceBox);
    else {
        ASSERT(referenceBox != CSSBoxType::BoxMissing);
        operation = BoxClipPathOperation::create(referenceBox);
    }

    return operation.release();
}

inline EResize StyleBuilderConverter::convertResize(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    EResize resize = RESIZE_NONE;
    if (primitiveValue.getValueID() == CSSValueAuto) {
    	resize = builderState.document().settings().textAreasAreResizable() ? RESIZE_BOTH : RESIZE_NONE;
    } else
        resize = primitiveValue;

    return resize;
}

inline int StyleBuilderConverter::convertMarqueeRepetition(Style::BuilderState&, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID() == CSSValueInfinite)
        return -1; // -1 means repeat forever.

    ASSERT(primitiveValue.isNumber());
    return primitiveValue.intValue();
}

inline int StyleBuilderConverter::convertMarqueeSpeed(Style::BuilderState&, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    int speed = 85;
    if (CSSValueID ident = primitiveValue.getValueID()) {
        switch (ident) {
        case CSSValueSlow:
            speed = 500; // 500 msec.
            break;
        case CSSValueNormal:
            speed = 85; // 85msec. The WinIE default.
            break;
        case CSSValueFast:
            speed = 10; // 10msec. Super fast.
            break;
        default:
            ASSERT_NOT_REACHED();
            break;
        }
    } else if (primitiveValue.isTime())
        speed = primitiveValue.computeTime<int, CSSPrimitiveValue::Milliseconds>();
    else {
        // For scrollamount support.
        ASSERT(primitiveValue.isNumber());
        speed = primitiveValue.intValue();
    }
    return speed;
}

inline PassRefPtr<QuotesData> StyleBuilderConverter::convertQuotes(Style::BuilderState&, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        return QuotesData::create(Vector<std::pair<String, String>>());
    }

    CSSValueList& list = toCSSValueList(value);
    Vector<std::pair<String, String>> quotes;
    quotes.reserveInitialCapacity(list.length() / 2);
    for (unsigned i = 0; i < list.length(); i += 2) {
        auto& first = list.itemWithoutBoundsCheck(i);
        // item() returns null if out of bounds so this is safe.
        auto* second = list.item(i + 1);
        if (!second)
            break;
        String startQuote = toCSSPrimitiveValue(first).stringValue();
        String endQuote = toCSSPrimitiveValue(*second).stringValue();
        quotes.append(std::make_pair(startQuote, endQuote));
    }
    return QuotesData::create(quotes);
}

inline TextUnderlinePosition StyleBuilderConverter::convertTextUnderlinePosition(Style::BuilderState&, const CSSValue& value)
{
    ASSERT(is<CSSPrimitiveValue>(value));
    return toCSSPrimitiveValue(value);
}

inline TextUnderlineOffset StyleBuilderConverter::convertTextUnderlineOffset(Style::BuilderState& builderState, const CSSValue& value)
{
    ASSERT(is<CSSPrimitiveValue>(value));
    auto& primitiveValue = toCSSPrimitiveValue(value);
    switch (primitiveValue.valueID()) {
    case CSSValueAuto:
        return TextUnderlineOffset::createWithAuto();
    default:
        ASSERT(primitiveValue.isLength());
        auto computedLength = convertComputedLength<float>(builderState, primitiveValue);
        return TextUnderlineOffset::createWithLength(computedLength);
    }
}

inline TextDecorationThickness StyleBuilderConverter::convertTextDecorationThickness(Style::BuilderState& builderState, const CSSValue& value)
{
    ASSERT(is<CSSPrimitiveValue>(value));
    auto& primitiveValue = toCSSPrimitiveValue(value);
    switch (primitiveValue.valueID()) {
    case CSSValueAuto:
        return TextDecorationThickness::createWithAuto();
    case CSSValueFromFont:
        return TextDecorationThickness::createFromFont();
    default:
        ASSERT(primitiveValue.isLength());
        auto computedLength = convertComputedLength<float>(builderState, primitiveValue);
        return TextDecorationThickness::createWithLength(computedLength);
    }
}

inline PassRefPtr<StyleReflection> StyleBuilderConverter::convertReflection(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        return nullptr;
    }

    CSSReflectValue& reflectValue = toCSSReflectValue(value);

    RefPtr<StyleReflection> reflection = StyleReflection::create();
    reflection->setDirection(*reflectValue.direction());

    if (reflectValue.offset())
        reflection->setOffset(reflectValue.offset()->convertToLength<FixedIntegerConversion | PercentConversion | CalculatedConversion>(builderState.cssToLengthConversionData()));

    NinePieceImage mask(NinePieceImage::Type::Mask);
    builderState.styleMap().mapNinePieceImage(CSSPropertyWebkitBoxReflect, reflectValue.mask(), mask);
    reflection->setMask(mask);

    return reflection.release();
}

inline IntSize StyleBuilderConverter::convertInitialLetter(Style::BuilderState&, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    if (primitiveValue.getValueID() == CSSValueNormal)
        return IntSize();

    Pair* pair = primitiveValue.pairValue();
    ASSERT(pair);
    ASSERT(pair->first());
    ASSERT(pair->second());

    return IntSize(pair->first()->intValue(), pair->second()->intValue());
}

inline float StyleBuilderConverter::convertTextStrokeWidth(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    float width = 0;
    switch (primitiveValue.getValueID()) {
    case CSSValueThin:
    case CSSValueMedium:
    case CSSValueThick: {
        double result = 1.0 / 48;
        if (primitiveValue.getValueID() == CSSValueMedium)
            result *= 3;
        else if (primitiveValue.getValueID() == CSSValueThick)
            result *= 5;
        RefPtr<CSSPrimitiveValue> emsValue(CSSPrimitiveValue::create(result, CSSUnitType::CSS_EMS));
        width = convertComputedLength<float>(builderState, *emsValue);
        break;
    }
    case CSSValueInvalid: {
        width = convertComputedLength<float>(builderState, primitiveValue);
        break;
    }
    default:
        ASSERT_NOT_REACHED();
        return 0;
    }

    return width;
}

inline LineBoxContain StyleBuilderConverter::convertLineBoxContain(Style::BuilderState&, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        return LineBoxContainNone;
    }

    return toCSSLineBoxContainValue(value).value();
}

inline TextDecorationSkip StyleBuilderConverter::valueToDecorationSkip(const CSSPrimitiveValue& primitiveValue)
{
    ASSERT(primitiveValue.isValueID());

    switch (primitiveValue.getValueID()) {
    case CSSValueAuto:
        return TextDecorationSkipAuto;
    case CSSValueNone:
        return TextDecorationSkipNone;
    case CSSValueInk:
        return TextDecorationSkipInk;
    case CSSValueObjects:
        return TextDecorationSkipObjects;
    default:
        break;
    }

    ASSERT_NOT_REACHED();
    return TextDecorationSkipNone;
}

inline TextDecorationSkip StyleBuilderConverter::convertTextDecorationSkip(Style::BuilderState&, CSSValue& value)
{
    if (value.isPrimitiveValue())
        return valueToDecorationSkip(toCSSPrimitiveValue(value));

    TextDecorationSkip skip = TextDecorationSkipNone;
    for (auto& currentValue : toCSSValueList(value))
        skip |= valueToDecorationSkip(toCSSPrimitiveValue(currentValue.get()));
    return skip;
}

#if ENABLE(CSS_SHAPES)
static inline bool isImageShape(const CSSValue& value)
{
    return value.isImageValue()
        || value.isImageSetValue()
        || value.isImageGeneratorValue();
}

inline PassRefPtr<ShapeValue> StyleBuilderConverter::convertShapeValue(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        return nullptr;
    }

    if (isImageShape(value))
        return ShapeValue::createImageValue(builderState.createStyleImage(value));

    RefPtr<BasicShape> shape;
    CSSBoxType referenceBox = CSSBoxType::BoxMissing;
    auto processSingleValue = [&](const CSSValue& singleValue) {
        auto& primitiveValue = toCSSPrimitiveValue(singleValue);
        if (primitiveValue.isShape())
            shape = basicShapeForValue(builderState.cssToLengthConversionData(), *primitiveValue.shapeValue());
        else if (primitiveValue.getValueID() == CSSValueContentBox
            || primitiveValue.getValueID() == CSSValueBorderBox
            || primitiveValue.getValueID() == CSSValuePaddingBox
            || primitiveValue.getValueID() == CSSValueMarginBox)
            referenceBox = primitiveValue;
        else {
            ASSERT_NOT_REACHED();
        }
    };
    if (value.isValueList()) {
        for (auto& currentValue : toCSSValueList(value))
            processSingleValue(currentValue);
    } else {
        processSingleValue(value);
    }

    if (shape)
        return ShapeValue::createShapeValue(shape.release(), referenceBox);

    if (referenceBox != CSSBoxType::BoxMissing)
        return ShapeValue::createBoxShapeValue(referenceBox);

    ASSERT_NOT_REACHED();
    return nullptr;
}
#endif // ENABLE(CSS_SHAPES)

#if ENABLE(CSS_SCROLL_SNAP)
inline Length StyleBuilderConverter::parseSnapCoordinate(Style::BuilderState& builderState, const CSSValue& value)
{
    return downcast<CSSPrimitiveValue>(value).convertToLength<FixedIntegerConversion | PercentConversion | CalculatedConversion | AutoConversion>(builderState.cssToLengthConversionData());
}

inline std::unique_ptr<ScrollSnapPoints> StyleBuilderConverter::convertScrollSnapPoints(Style::BuilderState& builderState, const CSSValue& value)
{
    // FIXME-NEWPARSER: Old parser supports an identifier value called "elements" when it seems like
    // "none" is what others use. For now, support both in the converter.
    auto points = std::make_unique<ScrollSnapPoints>();

    if (is<CSSPrimitiveValue>(value)) {
        ASSERT(downcast<CSSPrimitiveValue>(value).valueID() == CSSValueElements || downcast<CSSPrimitiveValue>(value).valueID() == CSSValueNone);
        points->usesElements = true;
        return points;
    }

    for (auto& currentValue : downcast<CSSValueList>(value)) {
        if (is<CSSFunctionValue>(currentValue.get())) {
            auto& functionValue = downcast<CSSFunctionValue>(currentValue.get());
            points->repeatOffset = parseSnapCoordinate(builderState, *functionValue.item(0));
            points->hasRepeat = true;
            break;
        }

        auto& itemValue = downcast<CSSPrimitiveValue>(currentValue.get());
        if (auto* lengthRepeat = itemValue.lengthRepeatValue()) {
            // FIXME-NEWPARSER: Old parsing code, which uses a special primitive value.
            // Can remove this once new parser turns on.
            if (auto* interval = lengthRepeat->interval()) {
                points->repeatOffset = parseSnapCoordinate(builderState, *interval);
                points->hasRepeat = true;
                break;
            }
        }
        points->offsets.append(parseSnapCoordinate(builderState, itemValue));
    }

    return points;
}

inline LengthSize StyleBuilderConverter::convertSnapCoordinatePair(Style::BuilderState& builderState, const CSSValue& value, size_t offset)
{
    // FIXME-NEWPARSER: Can make this unconditional once old parser is gone. This is just how
    // we detect that we're dealing with the new parser's pairs.
    if (value.isPrimitiveValue() && downcast<CSSPrimitiveValue>(value).pairValue()) {
        Pair* pair = downcast<CSSPrimitiveValue>(value).pairValue();
        Length lengthX = convertPositionComponent<CSSValueLeft, CSSValueRight>(builderState, *pair->first());
        Length lengthY = convertPositionComponent<CSSValueTop, CSSValueBottom>(builderState, *pair->second());
        return LengthSize(lengthX, lengthY);
    }

    // FIXME-NEWPARSER: Remove once old parser is gone.
    auto& valueList = downcast<CSSValueList>(value);
    return LengthSize(parseSnapCoordinate(builderState, *valueList.item(offset)), parseSnapCoordinate(builderState, *valueList.item(offset + 1)));
}

inline Vector<LengthSize> StyleBuilderConverter::convertScrollSnapCoordinates(Style::BuilderState& builderState, const CSSValue& value)
{
    if (is<CSSPrimitiveValue>(value)) {
        ASSERT(downcast<CSSPrimitiveValue>(value).valueID() == CSSValueNone);
        return Vector<LengthSize>();
    }

    auto& valueList = downcast<CSSValueList>(value);
    if (!valueList.length())
        return Vector<LengthSize>();

    // FIXME-NEWPARSER: Can make this unconditional once old parser is gone. This is just how
    // we detect that we're dealing with the new parser's pairs.
    if (valueList.item(0)->isPrimitiveValue() && downcast<CSSPrimitiveValue>(*valueList.item(0)).pairValue()) {
        Vector<LengthSize> coordinates;
        coordinates.reserveInitialCapacity(valueList.length());
        for (auto& snapCoordinate : valueList) {
            Pair* pair = downcast<CSSPrimitiveValue>(*snapCoordinate.ptr()).pairValue();
            Length lengthX = convertPositionComponent<CSSValueLeft, CSSValueRight>(builderState, *pair->first());
            Length lengthY = convertPositionComponent<CSSValueTop, CSSValueBottom>(builderState, *pair->second());
            coordinates.uncheckedAppend(LengthSize(lengthX, lengthY));
        }
        return coordinates;
    }

    // FIXME-NEWPARSER: Remove all of this once old parser is gone, including all the functions
    // it calls.
    ASSERT(!(valueList.length() % 2));
    size_t pointCount = valueList.length() / 2;
    Vector<LengthSize> coordinates;
    coordinates.reserveInitialCapacity(pointCount);
    for (size_t i = 0; i < pointCount; ++i)
        coordinates.uncheckedAppend(convertSnapCoordinatePair(builderState, valueList, i * 2));
    return coordinates;
}
#endif

#if ENABLE(CSS_GRID_LAYOUT)
inline GridLength StyleBuilderConverter::createGridTrackBreadth(const CSSPrimitiveValue& primitiveValue, Style::BuilderState& builderState)
{
    if (primitiveValue.valueID() == CSSValueMinContent || primitiveValue.valueID() == CSSValueWebkitMinContent)
        return Length(LengthType::MinContent);

    if (primitiveValue.valueID() == CSSValueMaxContent || primitiveValue.valueID() == CSSValueWebkitMaxContent)
        return Length(LengthType::MaxContent);

    // Fractional unit.
    if (primitiveValue.isFlex())
        return GridLength(primitiveValue.doubleValue());

    auto length = primitiveValue.convertToLength<FixedIntegerConversion | PercentConversion | CalculatedConversion | AutoConversion>(builderState.cssToLengthConversionData());
    if (!length.isUndefined())
		return length;
	return Length(0.0, LengthType::Fixed);
}

inline GridTrackSize StyleBuilderConverter::createGridTrackSize(const CSSValue& value, Style::BuilderState& builderState)
{
    if (value.isPrimitiveValue())
        return GridTrackSize(createGridTrackBreadth(toCSSPrimitiveValue(value), builderState));

    if (!isCSSFunctionValue(value))
        return GridTrackSize(GridLength(0));

    ASSERT(is<CSSFunctionValue>(value));
    auto& function = toCSSFunctionValue(value);

    if (function.length() == 1)
        return GridTrackSize(createGridTrackBreadth(toCSSPrimitiveValue(function.itemWithoutBoundsCheck(0)), builderState), FitContentTrackSizing);

    ASSERT_WITH_SECURITY_IMPLICATION(function.length() == 2);
    GridLength minTrackBreadth(createGridTrackBreadth(toCSSPrimitiveValue(function.itemWithoutBoundsCheck(0)), builderState));
    GridLength maxTrackBreadth(createGridTrackBreadth(toCSSPrimitiveValue(function.itemWithoutBoundsCheck(1)), builderState));
    return GridTrackSize(minTrackBreadth, maxTrackBreadth);
}

static void createGridLineNamesList(const CSSValue& value, unsigned currentNamedGridLine, NamedGridLinesMap& namedGridLines, OrderedNamedGridLinesMap& orderedNamedGridLines)
{
    ASSERT(value.isGridLineNamesValue());

    for (auto& namedGridLineValue : toCSSGridLineNamesValue(value)) {
        String namedGridLine = toCSSPrimitiveValue(namedGridLineValue.get()).stringValue();
        auto result = namedGridLines.add(namedGridLine, Vector<unsigned>());
        result.iterator->second.append(currentNamedGridLine);
        auto orderedResult = orderedNamedGridLines.add(currentNamedGridLine, Vector<String>());
        orderedResult.iterator->second.append(namedGridLine);
    }
}

struct StyleBuilderConverter::TracksData {
    WTF_MAKE_NONCOPYABLE(TracksData); WTF_MAKE_FAST_ALLOCATED;
public:
    TracksData()
		: m_autoRepeatInsertionPoint(RenderStyle::initialGridAutoRepeatInsertionPoint())
		, m_autoRepeatType(RenderStyle::initialGridAutoRepeatType())
	{
	}

    Vector<GridTrackSize> m_trackSizes;
    NamedGridLinesMap m_namedGridLines;
    OrderedNamedGridLinesMap m_orderedNamedGridLines;
    Vector<GridTrackSize> m_autoRepeatTrackSizes;
    NamedGridLinesMap m_autoRepeatNamedGridLines;
    OrderedNamedGridLinesMap m_autoRepeatOrderedNamedGridLines;
    unsigned m_autoRepeatInsertionPoint/* { RenderStyle::initialGridAutoRepeatInsertionPoint() }*/;
    AutoRepeatType m_autoRepeatType/* { RenderStyle::initialGridAutoRepeatType() }*/;
};

inline bool StyleBuilderConverter::createGridTrackList(CSSValue& value, TracksData& tracksData, Style::BuilderState& builderState)
{
    // Handle 'none'.
    if (value.isPrimitiveValue())
        return toCSSPrimitiveValue(value).getValueID() == CSSValueNone;

    if (!value.isValueList())
        return false;

    unsigned currentNamedGridLine = 0;
    auto handleLineNameOrTrackSize = [&](const CSSValue& currentValue) {
        if (currentValue.isGridLineNamesValue())
            createGridLineNamesList(currentValue, currentNamedGridLine, tracksData.m_namedGridLines, tracksData.m_orderedNamedGridLines);
        else {
            ++currentNamedGridLine;
            tracksData.m_trackSizes.append(createGridTrackSize(currentValue, builderState));
        }
    };

    for (auto& currentValue : toCSSValueList(value)) {
        if (currentValue.get().isGridAutoRepeatValue()) {
            ASSERT(tracksData.m_autoRepeatTrackSizes.isEmpty());
            unsigned autoRepeatIndex = 0;
            CSSValueID autoRepeatID = toCSSGridAutoRepeatValue(currentValue.get()).autoRepeatID();
            ASSERT(autoRepeatID == CSSValueAutoFill || autoRepeatID == CSSValueAutoFit);
            tracksData.m_autoRepeatType = autoRepeatID == CSSValueAutoFill ? AutoFill : AutoFit;
            for (auto& autoRepeatValue : toCSSValueList(currentValue.get())) {
                if (autoRepeatValue.get().isGridLineNamesValue()) {
                    createGridLineNamesList(autoRepeatValue.get(), autoRepeatIndex, tracksData.m_autoRepeatNamedGridLines, tracksData.m_autoRepeatOrderedNamedGridLines);
                    continue;
                }
                ++autoRepeatIndex;
                tracksData.m_autoRepeatTrackSizes.append(createGridTrackSize(autoRepeatValue.get(), builderState));
            }
            tracksData.m_autoRepeatInsertionPoint = currentNamedGridLine++;
            continue;
        }

        if (currentValue.get().isGridIntegerRepeatValue()) {
            size_t repetitions = static_cast<CSSGridIntegerRepeatValue&>(currentValue.get()).repetitions();
            for (size_t i = 0; i < repetitions; ++i) {
                for (auto& integerRepeatValue : toCSSValueList(currentValue.get()))
                    handleLineNameOrTrackSize(integerRepeatValue);
            }
            continue;
        }

        handleLineNameOrTrackSize(currentValue);
    }

    // The parser should have rejected any <track-list> without any <track-size> as
    // this is not conformant to the syntax.
    ASSERT(!tracksData.m_trackSizes.isEmpty() || !tracksData.m_autoRepeatTrackSizes.isEmpty());
    return true;
}

inline bool StyleBuilderConverter::createGridPosition(CSSValue& value, GridPosition& position)
{
    // We accept the specification's grammar:
    // auto | <custom-ident> | [ <integer> && <custom-ident>? ] | [ span && [ <integer> || <custom-ident> ] ]
    if (value.isPrimitiveValue()) {
        auto& primitiveValue = toCSSPrimitiveValue(value);
        if (primitiveValue.isCustomIdent()) {
            position.setNamedGridArea(primitiveValue.stringValue());
            return true;
        }

        ASSERT(primitiveValue.getValueID() == CSSValueAuto);
        return true;
    }

    auto& values = toCSSValueList(value);
    ASSERT(values.length());

    auto it = values.begin();
    CSSPrimitiveValue* currentValue = &toCSSPrimitiveValue(it->get());
    bool isSpanPosition = false;
    if (currentValue->getValueID() == CSSValueSpan) {
        isSpanPosition = true;
        ++it;
        currentValue = it != values.end() ? &toCSSPrimitiveValue(it->get()) : nullptr;
    }

    int gridLineNumber = 0;
    if (currentValue && currentValue->isNumber()) {
        gridLineNumber = currentValue->intValue();
        ++it;
        currentValue = it != values.end() ? &toCSSPrimitiveValue(it->get()) : nullptr;
    }

    String gridLineName;
    if (currentValue && currentValue->isCustomIdent()) {
        gridLineName = currentValue->stringValue();
        ++it;
    }

    ASSERT(it == values.end());
    if (isSpanPosition)
        position.setSpanPosition(gridLineNumber ? gridLineNumber : 1, gridLineName);
    else
        position.setExplicitPosition(gridLineNumber, gridLineName);

    return true;
}

inline void StyleBuilderConverter::createImplicitNamedGridLinesFromGridArea(const NamedGridAreaMap& namedGridAreas, NamedGridLinesMap& namedGridLines, GridTrackSizingDirection direction)
{
    for (auto& area : namedGridAreas) {
        GridSpan areaSpan = direction == ForRows ? area.second.rows : area.second.columns;
        {
            auto& startVector = namedGridLines.add(makeString(area.first + "-start"), Vector<unsigned>()).iterator->second;
            startVector.append(areaSpan.startLine());
            std::sort(startVector.begin(), startVector.end());
        }
        {
            auto& endVector = namedGridLines.add(makeString(area.first + "-end"), Vector<unsigned>()).iterator->second;
            endVector.append(areaSpan.endLine());
            std::sort(endVector.begin(), endVector.end());
        }
    }
}

inline Vector<GridTrackSize> StyleBuilderConverter::convertGridTrackSizeList(Style::BuilderState& builderState, CSSValue& value)
{
    auto validateValueAndAppend = [&](Style::BuilderState& builderState, Vector<GridTrackSize>& trackSizes, const CSSValue& value) {
        ASSERT(!value.isGridLineNamesValue());
        ASSERT(!value.isGridAutoRepeatValue());
        ASSERT(!value.isGridIntegerRepeatValue());
        trackSizes.uncheckedAppend(convertGridTrackSize(builderState, value));
    };

    if (value.isPrimitiveValue()) {
        auto& primitiveValue = toCSSPrimitiveValue(value);
        if (primitiveValue.isValueID()) {
            ASSERT(primitiveValue.valueID() == CSSValueAuto);
            return RenderStyle::initialGridAutoRows();
        }
        // Values coming from CSS Typed OM may not have been converted to a CSSValueList yet.
        Vector<GridTrackSize> trackSizes;
        trackSizes.reserveInitialCapacity(1);
        validateValueAndAppend(builderState, trackSizes, primitiveValue);
        return trackSizes;
    }

    Vector<GridTrackSize> trackSizes;
    if (value.isValueList())  {
        auto& valueList = toCSSValueList(value);
		trackSizes.reserveInitialCapacity(valueList.length());
		for (auto& currentValue : valueList)
			validateValueAndAppend(builderState, trackSizes, currentValue);
	} else {
		trackSizes.reserveInitialCapacity(1);
		validateValueAndAppend(builderState, trackSizes, value);
	}
    return trackSizes;
}

inline GridTrackSize StyleBuilderConverter::convertGridTrackSize(Style::BuilderState& builderState, const CSSValue& value)
{
    return createGridTrackSize(value, builderState);
}

inline Optional<GridPosition> StyleBuilderConverter::convertGridPosition(Style::BuilderState&, CSSValue& value)
{
    GridPosition gridPosition;
    if (createGridPosition(value, gridPosition))
        return gridPosition;
    return Nullopt;
}

inline GridAutoFlow StyleBuilderConverter::convertGridAutoFlow(Style::BuilderState&, CSSValue& value)
{
    ASSERT(!is<CSSPrimitiveValue>(value) || downcast<CSSPrimitiveValue>(value).isValueID());

    bool isValuelist = value.isValueList();
    if (isValuelist) {
		auto& list = toCSSValueList(value);
		if (!list.length())
			return RenderStyle::initialGridAutoFlow();
    }

    auto& first = toCSSPrimitiveValue(isValuelist ? *(toCSSValueList(value).item(0)) : value);
    auto* second = toCSSPrimitiveValue(isValuelist && toCSSValueList(value).length() == 2 ? toCSSValueList(value).item(1) : nullptr);

    GridAutoFlow autoFlow;
    switch (first.valueID()) {
    case CSSValueRow:
        if (second && second->valueID() == CSSValueDense)
            autoFlow = AutoFlowRowDense;
        else
            autoFlow = AutoFlowRow;
        break;
    case CSSValueColumn:
        if (second && second->valueID() == CSSValueDense)
            autoFlow = AutoFlowColumnDense;
        else
            autoFlow = AutoFlowColumn;
        break;
    case CSSValueDense:
        if (second && second->valueID() == CSSValueColumn)
            autoFlow = AutoFlowColumnDense;
        else
            autoFlow = AutoFlowRowDense;
        break;
    default:
        ASSERT_NOT_REACHED();
        autoFlow = RenderStyle::initialGridAutoFlow();
        break;
    }

    return autoFlow;
}
#endif // ENABLE(CSS_GRID_LAYOUT)

inline float zoomWithTextZoomFactor(Style::BuilderState& builderState)
{
    if (auto* frame = builderState.document().frame()) {
        float textZoomFactor = builderState.style().textZoom() != TextZoomReset ? frame->textZoomFactor() : 1.0f;
        return builderState.style().effectiveZoom() * textZoomFactor;
	}
	return builderState.cssToLengthConversionData().zoom();
}

inline CSSToLengthConversionData StyleBuilderConverter::csstoLengthConversionDataWithTextZoomFactor(Style::BuilderState& builderState)
{
	float zoom = zoomWithTextZoomFactor(builderState);
	if (zoom == builderState.cssToLengthConversionData().zoom())
		return builderState.cssToLengthConversionData();

	return builderState.cssToLengthConversionData().copyWithAdjustedZoom(zoom);
}

inline Optional<Length> StyleBuilderConverter::convertWordSpacing(Style::BuilderState& builderState, CSSValue& value)
{
    Optional<Length> wordSpacing;
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID() == CSSValueNormal)
        wordSpacing = RenderStyle::initialWordSpacing();
    else if (primitiveValue.isLength())
        wordSpacing = primitiveValue.computeLength<Length>(csstoLengthConversionDataWithTextZoomFactor(builderState));
    else if (primitiveValue.isPercentage())
        wordSpacing = Length(clampTo<float>(primitiveValue.doubleValue(), minValueForCssLength, maxValueForCssLength), LengthType::Percent);
    else if (primitiveValue.isNumber())
        wordSpacing = Length(primitiveValue.doubleValue(), LengthType::Fixed);

    return wordSpacing;
}

inline Optional<float> StyleBuilderConverter::convertPerspective(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID() == CSSValueNone)
        return 0.f;

    float perspective = -1;
    if (primitiveValue.isLength())
        perspective = primitiveValue.computeLength<float>(builderState.cssToLengthConversionData());
    else if (primitiveValue.isNumber())
        perspective = primitiveValue.doubleValue() * builderState.cssToLengthConversionData().zoom();
    else
        ASSERT_NOT_REACHED();

    return perspective < 0 ? Nullopt : perspective;
}

inline Optional<Length> StyleBuilderConverter::convertMarqueeIncrement(Style::BuilderState& builderState, CSSValue& value)
{
    Optional<Length> marqueeLength;
    auto& primitiveValue = toCSSPrimitiveValue(value);
    switch (primitiveValue.getValueID()) {
    case CSSValueSmall:
        marqueeLength = Length(1, LengthType::Fixed); // 1px.
        break;
    case CSSValueNormal:
        marqueeLength = Length(6, LengthType::Fixed); // 6px. The WinIE default.
        break;
    case CSSValueLarge:
        marqueeLength = Length(36, LengthType::Fixed); // 36px.
        break;
    case CSSValueInvalid: {
        Length length = primitiveValue.convertToLength<FixedIntegerConversion | PercentConversion | CalculatedConversion>(builderState.cssToLengthConversionData().copyWithAdjustedZoom(1.0f));
        if (!length.isUndefined())
            marqueeLength = length;
        break;
    }
    default:
        break;
    }
    return marqueeLength;
}

inline Optional<FilterOperations> StyleBuilderConverter::convertFilterOperations(Style::BuilderState& builderState, CSSValue& value)
{
    FilterOperations operations;
    if (builderState.createFilterOperations(value, operations))
        return operations;
    return Nullopt;
}

inline FontFeatureSettings StyleBuilderConverter::convertFontFeatureSettings(Style::BuilderState&, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNormal);
        return FontFeatureSettings();
    }

    FontFeatureSettings settings;
    for (auto& item : toCSSValueList(value)) {
        auto& feature = toCSSFontFeatureValue(item.get());
        settings.insert(FontFeature(feature.tag(), feature.value()));
    }
    return settings;
}

inline FontSelectionValue StyleBuilderConverter::convertFontWeightFromValue(const CSSValue& value)
{
    ASSERT(is<CSSPrimitiveValue>(value));
    auto& primitiveValue = toCSSPrimitiveValue(value);

    if (primitiveValue.isNumber())
        return FontSelectionValue::clampFloat(primitiveValue.floatValue());

    ASSERT(primitiveValue.isValueID());
    switch (primitiveValue.valueID()) {
    case CSSValueNormal:
        return normalWeightValue();
    case CSSValueBold:
    case CSSValueBolder:
        return boldWeightValue();
    case CSSValueLighter:
        return lightWeightValue();
    default:
        ASSERT_NOT_REACHED();
        return normalWeightValue();
    }
}

inline FontSelectionValue StyleBuilderConverter::convertFontStretchFromValue(const CSSValue& value)
{
    ASSERT(is<CSSPrimitiveValue>(value));
    const auto& primitiveValue = toCSSPrimitiveValue(value);

    if (primitiveValue.isPercentage())
        return FontSelectionValue::clampFloat(primitiveValue.floatValue());

    ASSERT(primitiveValue.isValueID());
    if (auto value = fontStretchValue(primitiveValue.valueID()))
        return value.value();
    ASSERT_NOT_REACHED();
    return normalStretchValue();
}

inline FontSelectionValue StyleBuilderConverter::convertFontStyleFromValue(const CSSValue& value)
{
    ASSERT(is<CSSFontStyleValue>(value));
    const auto& fontStyleValue = toCSSFontStyleValue(value);

    auto valueID = fontStyleValue.fontStyleValue->valueID();
    if (valueID == CSSValueNormal)
        return normalItalicValue();
    if (valueID == CSSValueItalic)
        return italicValue();
    ASSERT(valueID == CSSValueOblique);
    if (auto* obliqueValue = fontStyleValue.obliqueValue.get())
        return FontSelectionValue(obliqueValue->getValue<float>(CSSUnitType::CSS_DEG));
    return italicValue();
}

inline FontSelectionValue StyleBuilderConverter::convertFontWeight(Style::BuilderState& builderState, const CSSValue& value)
{
    ASSERT(is<CSSPrimitiveValue>(value));
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.isValueID()) {
        auto valueID = primitiveValue.valueID();
        if (valueID == CSSValueBolder)
            return FontDescription::bolderWeight(builderState.parentStyle().fontDescription().weight());
        if (valueID == CSSValueLighter)
            return FontDescription::lighterWeight(builderState.parentStyle().fontDescription().weight());
    }
    return convertFontWeightFromValue(value);
}

inline FontSelectionValue StyleBuilderConverter::convertFontStretch(Style::BuilderState&, const CSSValue& value)
{
    return convertFontStretchFromValue(value);
}

#if 0
inline FontVariationSettings BuilderConverter::convertFontVariationSettings(BuilderState&, const CSSValue& value)
{
    if (is<CSSPrimitiveValue>(value)) {
        ASSERT(value.valueID() == CSSValueNormal || CSSPropertyParserHelpers::isSystemFontShorthand(value.valueID()));
        return { };
    }
    FontVariationSettings settings;
    for (auto& item : downcast<CSSValueList>(value)) {
        auto& feature = downcast<CSSFontVariationValue>(item);
        settings.insert({ feature.tag(), feature.value() });
    }
    return settings;
}
inline FontSizeAdjust BuilderConverter::convertFontSizeAdjust(BuilderState&, const CSSValue& value)
{
    if (is<CSSPrimitiveValue>(value)) {
        auto& primitiveValue = downcast<CSSPrimitiveValue>(value);
        if (primitiveValue.valueID() == CSSValueNone
            || CSSPropertyParserHelpers::isSystemFontShorthand(primitiveValue.valueID()))
            return FontCascadeDescription::initialFontSizeAdjust();
        auto defaultMetric = FontSizeAdjust::Metric::ExHeight;
        if (primitiveValue.isNumber())
            return { defaultMetric, FontSizeAdjust::ValueType::Number, primitiveValue.floatValue() };
        ASSERT(primitiveValue.valueID() == CSSValueFromFont);
        // We cannot determine the primary font here, so we defer resolving the
        // aspect value for from-font to when the primary font is created.
        // See FontCascadeFonts::primaryFont().
        return { defaultMetric, FontSizeAdjust::ValueType::FromFont, std::nullopt };
    }

    if (!is<CSSValuePair>(value))
        return FontCascadeDescription::initialFontSizeAdjust();

    const auto& pair = downcast<CSSValuePair>(value);

    auto metric = fromCSSValueID<FontSizeAdjust::Metric>(downcast<CSSPrimitiveValue>(pair.first()).valueID());
    auto& primitiveValue = downcast<CSSPrimitiveValue>(pair.second());
    if (primitiveValue.isNumber())
        return { metric, FontSizeAdjust::ValueType::Number, primitiveValue.floatValue() };
    ASSERT(primitiveValue.valueID() == CSSValueFromFont);
    return { metric, FontSizeAdjust::ValueType::FromFont, std::nullopt };
}
#endif

inline FontSelectionValue StyleBuilderConverter::convertFontStyle(Style::BuilderState&, const CSSValue& value)
{
    return convertFontStyleFromValue(value);
}

inline SVGLengthValue StyleBuilderConverter::convertSVGLengthValue(Style::BuilderState& builderState, const CSSValue& value)
{
    return SVGLengthValue::fromCSSPrimitiveValue(toCSSPrimitiveValue(value), builderState.cssToLengthConversionData());
}

inline Vector<SVGLengthValue> StyleBuilderConverter::convertSVGLengthVector(Style::BuilderState& builderState, const CSSValue& value)
{
    auto& valueList = toCSSValueList(value);

    Vector<SVGLengthValue> svgLengths;
    svgLengths.reserveInitialCapacity(valueList.length());
    for (auto& item : valueList)
        svgLengths.uncheckedAppend(convertSVGLengthValue(builderState, item));

    return svgLengths;
}

inline Vector<SVGLengthValue> StyleBuilderConverter::convertStrokeDashArray(Style::BuilderState& builderState, const CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        return SVGRenderStyle::initialStrokeDashArray();
    }

    return convertSVGLengthVector(builderState, value);
}

inline PaintOrder StyleBuilderConverter::convertPaintOrder(Style::BuilderState&, const CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNormal);
        return PaintOrder::Normal;
    }

    auto& orderTypeList = toCSSValueList(value);
    switch (toCSSPrimitiveValue(orderTypeList.itemWithoutBoundsCheck(0)).getValueID()) {
    case CSSValueFill:
        return orderTypeList.length() > 1 ? PaintOrder::FillMarkers : PaintOrder::Fill;
    case CSSValueStroke:
        return orderTypeList.length() > 1 ? PaintOrder::StrokeMarkers : PaintOrder::Stroke;
    case CSSValueMarkers:
        return orderTypeList.length() > 1 ? PaintOrder::MarkersStroke : PaintOrder::Markers;
    default:
        ASSERT_NOT_REACHED();
        return PaintOrder::Normal;
    }
}

inline float StyleBuilderConverter::convertOpacity(Style::BuilderState&, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    float opacity = primitiveValue.floatValue();
    if (primitiveValue.isPercentage())
        opacity /= 100.0f;
    return std::max(0.0f, std::min(1.0f, opacity));
}

inline String StyleBuilderConverter::convertSVGURIReference(Style::BuilderState& builderState, CSSValue& value)
{
    String s;
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.isURI())
        s = primitiveValue.stringValue();

    return SVGURIReference::fragmentIdentifierFromIRIString(s, builderState.document());
}

inline Color StyleBuilderConverter::convertSVGColor(Style::BuilderState& builderState, const CSSValue& value)
{
	return builderState.colorFromPrimitiveValue(toCSSPrimitiveValue(value));
}

inline EGlyphOrientation StyleBuilderConverter::convertGlyphOrientation(Style::BuilderState&, CSSValue& value)
{
    float angle = fabsf(fmodf(toCSSPrimitiveValue(value).floatValue(), 360.0f));
    if (angle <= 45.0f || angle > 315.0f)
        return GO_0DEG;
    if (angle > 45.0f && angle <= 135.0f)
        return GO_90DEG;
    if (angle > 135.0f && angle <= 225.0f)
        return GO_180DEG;
    return GO_270DEG;
}

inline EGlyphOrientation StyleBuilderConverter::convertGlyphOrientationOrAuto(Style::BuilderState& builderState, CSSValue& value)
{
    if (toCSSPrimitiveValue(value).getValueID() == CSSValueAuto)
        return GO_AUTO;
    return convertGlyphOrientation(builderState, value);
}

inline Optional<Length> StyleBuilderConverter::convertLineHeight(Style::BuilderState& builderState, CSSValue& value, float multiplier)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID() == CSSValueNormal)
        return RenderStyle::initialLineHeight();

    if (primitiveValue.isLength() || primitiveValue.isCalculatedPercentageWithLength()) {
        auto conversionData = builderState.cssToLengthConversionData().copyWithAdjustedZoomAndPropertyToCompute(zoomWithTextZoomFactor(builderState), CSSPropertyLineHeight);
        Length length;
		if (primitiveValue.isLength())
			length = primitiveValue.computeLength<Length>(conversionData);
		else {
			auto value = primitiveValue.cssCalcValue()->createCalculationValue(conversionData)->evaluate(builderState.style().computedFontSize());
			length = Length(clampTo<float>(value, minValueForCssLength, maxValueForCssLength), LengthType::Fixed);
		}
        if (multiplier != 1.f)
            length = Length(length.value() * multiplier, LengthType::Fixed);
        return length;
    }

    // Line-height percentages need to inherit as if they were Fixed pixel values. In the example:
    // <div style="font-size: 10px; line-height: 150%;"><div style="font-size: 100px;"></div></div>
    // the inner element should have line-height of 15px. However, in this example:
    // <div style="font-size: 10px; line-height: 1.5;"><div style="font-size: 100px;"></div></div>
    // the inner element should have a line-height of 150px. Therefore, we map percentages to Fixed
    // values and raw numbers to percentages.
    if (primitiveValue.isPercentage()) {
        // FIXME: percentage should not be restricted to an integer here.
        return Length((builderState.style().computedFontSize() * primitiveValue.intValue()) / 100, LengthType::Fixed);
    }
    if (primitiveValue.isNumber())
        return Length(primitiveValue.doubleValue() * 100.0, LengthType::Percent);

    // FIXME: The parser should only emit the above types, so this should never be reached. We should change the
    // type of this function to return just a Length (and not an Optional).
    return Nullopt;
}

inline FontSynthesis StyleBuilderConverter::convertFontSynthesis(Style::BuilderState&, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        return FontSynthesisNone;
    }

    FontSynthesis result = FontSynthesisNone;
    ASSERT(is<CSSValueList>(value));
    for (CSSValue& v : toCSSValueList(value)) {
        switch (toCSSPrimitiveValue(v).getValueID()) {
        case CSSValueWeight:
            result |= FontSynthesisWeight;
            break;
        case CSSValueStyle:
            result |= FontSynthesisStyle;
            break;
        case CSSValueSmallCaps:
            result |= FontSynthesisSmallCaps;
            break;
        default:
            ASSERT_NOT_REACHED();
            break;
        }
    }

    return result;
}

inline HangingPunctuation StyleBuilderConverter::convertHangingPunctuation(Style::BuilderState&, CSSValue& value)
{
    HangingPunctuation result = RenderStyle::initialHangingPunctuation();
    if (value.isValueList()) {
        for (auto& currentValue : toCSSValueList(value))
            result |= toCSSPrimitiveValue(currentValue.get());
    }
    return result;
}

inline GapLength StyleBuilderConverter::convertGapLength(Style::BuilderState& builderState, const CSSValue& value)
{
    return (toCSSPrimitiveValue(value).valueID() == CSSValueNormal) ? GapLength() : GapLength(convertLength(builderState, value));
}

inline StyleSelfAlignmentData StyleBuilderConverter::convertSelfOrDefaultAlignmentData(Style::BuilderState&, CSSValue& value)
{
    StyleSelfAlignmentData alignmentData = RenderStyle::initialSelfAlignment();
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (Pair* pairValue = primitiveValue.pairValue()) {
        if (pairValue->first()->getValueID() == CSSValueLegacy) {
            alignmentData.setPositionType(LegacyPosition);
            alignmentData.setPosition(*pairValue->second());
        } else if (pairValue->first()->valueID() == CSSValueFirst) {
            alignmentData.setPosition(ItemPositionBaseline);
        } else if (pairValue->first()->valueID() == CSSValueLast) {
            alignmentData.setPosition(ItemPositionLastBaseline);
        } else {
            alignmentData.setOverflow(*pairValue->first());
            alignmentData.setPosition(*pairValue->second());
        }
    } else
        alignmentData.setPosition(primitiveValue);
    return alignmentData;
}

inline StyleContentAlignmentData StyleBuilderConverter::convertContentAlignmentData(Style::BuilderState&, CSSValue& value)
{
    StyleContentAlignmentData alignmentData = RenderStyle::initialContentAlignment();
    if (!value.isContentDistributionValue())
        return alignmentData;
    auto& contentValue = toCSSContentDistributionValue(value);
    if (contentValue.distribution()->getValueID() != CSSValueInvalid)
        alignmentData.setDistribution(contentValue.distribution().get());
    if (contentValue.position()->getValueID() != CSSValueInvalid)
        alignmentData.setPosition(contentValue.position().get());
    if (contentValue.overflow()->getValueID() != CSSValueInvalid)
        alignmentData.setOverflow(contentValue.overflow().get());
    return alignmentData;
}

} // namespace WebCore
