/*
 * Copyright (C) 2013-2014 Google Inc. All rights reserved.
 * Copyright (C) 2014 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "CSSPrimitiveValue.h"
#include "CSSFontFamily.h"
#include "CSSCursorImageValue.h"
#include "CSSFontFamily.h"
#include "CSSFontFeatureValue.h"
#include "CSSGradientValue.h"
#include "CSSRegisteredCustomProperty.h"
#include "CSSShadowValue.h"
#include "Counter.h"
#include "CounterContent.h"
#include "CursorList.h"
#include "ElementAncestorIterator.h"
#include "FontVariantBuilder.h"
#include "Frame.h"
#include "HTMLElement.h"
#include "Rect.h"
#include "SVGElement.h"
#include "SVGRenderStyle.h"
#include "StyleBuilderConverter.h"
#include "StyleCachedImage.h"
#include "StyleCursorImage.h"
#include "StyleFontSizeFunctions.h"
#include "StyleGeneratedImage.h"
#include "StyleImageSet.h"
#include "StyleResolver.h"
#include "WillChangeData.h"
#include <wtf/SaturatedArithmetic.h>

namespace WebCore {

#define DECLARE_PROPERTY_CUSTOM_HANDLERS(property) \
    static void applyInherit##property(Style::BuilderState&); \
    static void applyInitial##property(Style::BuilderState&); \
    static void applyValue##property(Style::BuilderState&, CSSValue&)

template<typename T> inline T forwardInheritedValue(T&& value) { return std::forward<T>(value); }
inline Length forwardInheritedValue(const Length& value) { auto copy = value; return copy; }
inline LengthSize forwardInheritedValue(const LengthSize& value) { auto copy = value; return copy; }
inline LengthBox forwardInheritedValue(const LengthBox& value) { auto copy = value; return copy; }
inline GapLength forwardInheritedValue(const GapLength& value) { auto copy = value; return copy; }

// Note that we assume the CSS parser only allows valid CSSValue types.
class StyleBuilderCustom {
public:
    // Custom handling of inherit, initial and value setting.
    DECLARE_PROPERTY_CUSTOM_HANDLERS(AspectRatio);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(BorderImageOutset);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(BorderImageRepeat);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(BorderImageSlice);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(BorderImageWidth);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(BoxShadow);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(Clip);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(Content);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(CounterIncrement);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(CounterReset);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(Cursor);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(Fill);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(FontFamily);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(FontSize);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(WebkitHorizontalFontSize);
#if ENABLE(IOS_TEXT_AUTOSIZING)
    DECLARE_PROPERTY_CUSTOM_HANDLERS(LineHeight);
#endif
    DECLARE_PROPERTY_CUSTOM_HANDLERS(OutlineStyle);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(Size);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(Stroke);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(TextIndent);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(TextShadow);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(WebkitBoxShadow);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(FontVariantLigatures);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(FontVariantNumeric);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(FontVariantEastAsian);
#if ENABLE(CSS_GRID_LAYOUT)
    DECLARE_PROPERTY_CUSTOM_HANDLERS(GridTemplateAreas);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(GridTemplateColumns);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(GridTemplateRows);
#endif // ENABLE(CSS_GRID_LAYOUT)
    DECLARE_PROPERTY_CUSTOM_HANDLERS(WebkitMaskBoxImageOutset);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(WebkitMaskBoxImageRepeat);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(WebkitMaskBoxImageSlice);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(WebkitMaskBoxImageWidth);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(WebkitTextEmphasisStyle);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(Zoom);

    DECLARE_PROPERTY_CUSTOM_HANDLERS(NavIndex);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(NavDown);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(NavLeft);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(NavRight);
    DECLARE_PROPERTY_CUSTOM_HANDLERS(NavUp);

    // Custom handling of initial + inherit value setting only.
    static void applyInitialWebkitMaskImage(Style::BuilderState&) { }
    static void applyInheritWebkitMaskImage(Style::BuilderState&) { }
    static void applyInitialFontFeatureSettings(Style::BuilderState&) { }
    static void applyInheritFontFeatureSettings(Style::BuilderState&) { }

    // Custom handling of inherit + value setting only.
    static void applyInheritDisplay(Style::BuilderState&);
    static void applyValueDisplay(Style::BuilderState&, CSSValue&);

    // Custom handling of value setting only.
    static void applyValueBaselineShift(Style::BuilderState&, CSSValue&);
    static void applyValueDirection(Style::BuilderState&, CSSValue&);
    static void applyInheritVerticalAlign(Style::BuilderState&);
    static void applyValueVerticalAlign(Style::BuilderState&, CSSValue&);
    static void applyValueWebkitJustifySelf(Style::BuilderState&, CSSValue&);
    static void applyValueWebkitLocale(Style::BuilderState&, CSSValue&);
    static void applyValueWebkitTextOrientation(Style::BuilderState&, CSSValue&);
    static void applyValueTextOrientation(Style::BuilderState&, CSSValue&);
    static void applyValueWebkitTextSizeAdjust(Style::BuilderState&, CSSValue&);
    static void applyValueWebkitTextZoom(Style::BuilderState&, CSSValue&);
    static void applyValueWritingMode(Style::BuilderState&, CSSValue&);
    static void applyValueAlt(Style::BuilderState&, CSSValue&);

    static void applyValueWillChange(Style::BuilderState&, CSSValue&);

    static void applyValueStrokeWidth(Style::BuilderState&, CSSValue&);
    static void applyValueStrokeColor(Style::BuilderState&, CSSValue&);

    static void applyInitialCustomProperty(Style::BuilderState&, const CSSRegisteredCustomProperty*, const AtomicString& name);
    static void applyInheritCustomProperty(Style::BuilderState&, const CSSRegisteredCustomProperty*, const AtomicString& name);
    static void applyValueCustomProperty(Style::BuilderState&, const CSSRegisteredCustomProperty*, CSSCustomPropertyValue&);

private:
    static void resetEffectiveZoom(Style::BuilderState&);

    static Length mmLength(double mm);
    static Length inchLength(double inch);
    static bool getPageSizeFromName(CSSPrimitiveValue* pageSizeName, CSSPrimitiveValue* pageOrientation, Length& width, Length& height);

    template <CSSPropertyID id>
    static void applyTextOrBoxShadowValue(Style::BuilderState&, CSSValue&);

    enum CounterBehavior {Increment = 0, Reset};
    template <CounterBehavior counterBehavior>
    static void applyInheritCounter(Style::BuilderState&);
    template <CounterBehavior counterBehavior>
    static void applyValueCounter(Style::BuilderState&, CSSValue&);

    static float largerFontSize(float size);
    static float smallerFontSize(float size);
    static float determineRubyTextSizeMultiplier(Style::BuilderState&);
};

inline void StyleBuilderCustom::applyValueDirection(Style::BuilderState& builderState, CSSValue& value)
{
    builderState.style().setDirection(toCSSPrimitiveValue(value));
    builderState.style().setHasExplicitlySetDirection(true);
}

inline void StyleBuilderCustom::resetEffectiveZoom(Style::BuilderState& builderState)
{
    // Reset the zoom in effect. This allows the setZoom method to accurately compute a new zoom in effect.
	builderState.setEffectiveZoom(builderState.parentStyle().effectiveZoom());
}

inline void StyleBuilderCustom::applyInitialZoom(Style::BuilderState& builderState)
{
    resetEffectiveZoom(builderState);
    builderState.setZoom(RenderStyle::initialZoom());
}

inline void StyleBuilderCustom::applyInheritZoom(Style::BuilderState& builderState)
{
    resetEffectiveZoom(builderState);
    builderState.setZoom(builderState.parentStyle().zoom());
}

inline void StyleBuilderCustom::applyValueZoom(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    if (primitiveValue.getValueID() == CSSValueNormal) {
        resetEffectiveZoom(builderState);
        builderState.setZoom(RenderStyle::initialZoom());
    } else if (primitiveValue.getValueID() == CSSValueReset) {
    	builderState.setEffectiveZoom(RenderStyle::initialZoom());
    	builderState.setZoom(RenderStyle::initialZoom());
    } else if (primitiveValue.getValueID() == CSSValueDocument) {
        float docZoom = builderState.rootElementStyle() ? builderState.rootElementStyle()->zoom() : RenderStyle::initialZoom();
        builderState.setEffectiveZoom(docZoom);
        builderState.setZoom(docZoom);
    } else if (primitiveValue.isPercentage()) {
        resetEffectiveZoom(builderState);
        if (float percent = primitiveValue.floatValue())
        	builderState.setZoom(percent / 100.0f);
    } else if (primitiveValue.isNumber()) {
        resetEffectiveZoom(builderState);
        if (float number = primitiveValue.floatValue())
        	builderState.setZoom(number);
    }
}

inline Length StyleBuilderCustom::mmLength(double mm)
{
    RefPtr<CSSPrimitiveValue> value(CSSPrimitiveValue::create(mm, CSSUnitType::CSS_MM));
    return value->computeLength<Length>(CSSToLengthConversionData());
}
inline Length StyleBuilderCustom::inchLength(double inch)
{
    RefPtr<CSSPrimitiveValue> value(CSSPrimitiveValue::create(inch, CSSUnitType::CSS_IN));
    return value->computeLength<Length>(CSSToLengthConversionData());
}
bool StyleBuilderCustom::getPageSizeFromName(CSSPrimitiveValue* pageSizeName, CSSPrimitiveValue* pageOrientation, Length& width, Length& height)
{
    static NeverDestroyed<Length> a5Width(mmLength(148));
    static NeverDestroyed<Length> a5Height(mmLength(210));
    static NeverDestroyed<Length> a4Width(mmLength(210));
    static NeverDestroyed<Length> a4Height(mmLength(297));
    static NeverDestroyed<Length> a3Width(mmLength(297));
    static NeverDestroyed<Length> a3Height(mmLength(420));
    static NeverDestroyed<Length> b5Width(mmLength(176));
    static NeverDestroyed<Length> b5Height(mmLength(250));
    static NeverDestroyed<Length> b4Width(mmLength(250));
    static NeverDestroyed<Length> b4Height(mmLength(353));
    static NeverDestroyed<Length> letterWidth(inchLength(8.5));
    static NeverDestroyed<Length> letterHeight(inchLength(11));
    static NeverDestroyed<Length> legalWidth(inchLength(8.5));
    static NeverDestroyed<Length> legalHeight(inchLength(14));
    static NeverDestroyed<Length> ledgerWidth(inchLength(11));
    static NeverDestroyed<Length> ledgerHeight(inchLength(17));

    if (!pageSizeName)
        return false;

    switch (pageSizeName->getValueID()) {
    case CSSValueA5:
        width = a5Width;
        height = a5Height;
        break;
    case CSSValueA4:
        width = a4Width;
        height = a4Height;
        break;
    case CSSValueA3:
        width = a3Width;
        height = a3Height;
        break;
    case CSSValueB5:
        width = b5Width;
        height = b5Height;
        break;
    case CSSValueB4:
        width = b4Width;
        height = b4Height;
        break;
    case CSSValueLetter:
        width = letterWidth;
        height = letterHeight;
        break;
    case CSSValueLegal:
        width = legalWidth;
        height = legalHeight;
        break;
    case CSSValueLedger:
        width = ledgerWidth;
        height = ledgerHeight;
        break;
    default:
        return false;
    }

    if (pageOrientation) {
        switch (pageOrientation->getValueID()) {
        case CSSValueLandscape:
            std::swap(width, height);
            break;
        case CSSValuePortrait:
            // Nothing to do.
            break;
        default:
            return false;
        }
    }
    return true;
}

inline void StyleBuilderCustom::applyInheritVerticalAlign(Style::BuilderState& builderState)
{
    builderState.style().setVerticalAlignLength(forwardInheritedValue(builderState.parentStyle().verticalAlignLength()));
    builderState.style().setVerticalAlign(forwardInheritedValue(builderState.parentStyle().verticalAlign()));
}

inline void StyleBuilderCustom::applyValueVerticalAlign(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID())
        builderState.style().setVerticalAlign(primitiveValue);
    else
        builderState.style().setVerticalAlignLength(primitiveValue.convertToLength<FixedIntegerConversion | PercentConversion | CalculatedConversion>(builderState.cssToLengthConversionData()));
}

inline void StyleBuilderCustom::applyInheritSize(Style::BuilderState&) { }
inline void StyleBuilderCustom::applyInitialSize(Style::BuilderState&) { }
inline void StyleBuilderCustom::applyValueSize(Style::BuilderState& builderState, CSSValue& value)
{
    builderState.style().resetPageSizeType();
    Length width;
    Length height;
    PageSizeType pageSizeType = PAGE_SIZE_AUTO;
    if (!value.isValueList())
        return;

    auto& valueList = toCSSValueList(value);
    switch (valueList.length()) {
    case 2: {
        auto& firstValue = valueList.itemWithoutBoundsCheck(0);
        auto& secondValue = valueList.itemWithoutBoundsCheck(1);
        // <length>{2} | <page-size> <orientation>
        if (!firstValue.isPrimitiveValue() || !secondValue.isPrimitiveValue())
            return;
        auto& firstPrimitiveValue = toCSSPrimitiveValue(firstValue);
        auto& secondPrimitiveValue = toCSSPrimitiveValue(secondValue);
        if (firstPrimitiveValue.isLength()) {
            // <length>{2}
            if (!secondPrimitiveValue.isLength())
                return;
            CSSToLengthConversionData conversionData = builderState.cssToLengthConversionData().copyWithAdjustedZoom(1.0f);
            width = firstPrimitiveValue.computeLength<Length>(conversionData);
            height = secondPrimitiveValue.computeLength<Length>(conversionData);
        } else {
            // <page-size> <orientation>
            // The value order is guaranteed. See CSSParser::parseSizeParameter.
            if (!getPageSizeFromName(&firstPrimitiveValue, &secondPrimitiveValue, width, height))
                return;
        }
        pageSizeType = PAGE_SIZE_RESOLVED;
        break;
    }
    case 1: {
        auto& value = valueList.itemWithoutBoundsCheck(0);
        // <length> | auto | <page-size> | [ portrait | landscape]
        if (!value.isPrimitiveValue())
            return;
        auto& primitiveValue = toCSSPrimitiveValue(value);
        if (primitiveValue.isLength()) {
            // <length>
            pageSizeType = PAGE_SIZE_RESOLVED;
            width = height = primitiveValue.computeLength<Length>(builderState.cssToLengthConversionData().copyWithAdjustedZoom(1.0f));
        } else {
            switch (primitiveValue.getValueID()) {
            case 0:
                return;
            case CSSValueAuto:
                pageSizeType = PAGE_SIZE_AUTO;
                break;
            case CSSValuePortrait:
                pageSizeType = PAGE_SIZE_AUTO_PORTRAIT;
                break;
            case CSSValueLandscape:
                pageSizeType = PAGE_SIZE_AUTO_LANDSCAPE;
                break;
            default:
                // <page-size>
                pageSizeType = PAGE_SIZE_RESOLVED;
                if (!getPageSizeFromName(&primitiveValue, nullptr, width, height))
                    return;
            }
        }
        break;
    }
    default:
        return;
    }
    builderState.style().setPageSizeType(pageSizeType);
    builderState.style().setPageSize(LengthSize(width, height));
}

inline void StyleBuilderCustom::applyInheritTextIndent(Style::BuilderState& builderState)
{
    builderState.style().setTextIndent(builderState.parentStyle().textIndent());
    builderState.style().setTextIndentLine(builderState.parentStyle().textIndentLine());
    builderState.style().setTextIndentType(builderState.parentStyle().textIndentType());
}

inline void StyleBuilderCustom::applyInitialTextIndent(Style::BuilderState& builderState)
{
    builderState.style().setTextIndent(RenderStyle::initialTextIndent());
    builderState.style().setTextIndentLine(RenderStyle::initialTextIndentLine());
    builderState.style().setTextIndentType(RenderStyle::initialTextIndentType());
}

inline void StyleBuilderCustom::applyValueTextIndent(Style::BuilderState& builderState, CSSValue& value)
{
    Length lengthOrPercentageValue;
    TextIndentLine textIndentLineValue = RenderStyle::initialTextIndentLine();
    TextIndentType textIndentTypeValue = RenderStyle::initialTextIndentType();
    for (auto& item : toCSSValueList(value)) {
        auto& primitiveValue = toCSSPrimitiveValue(item.get());
        if (!primitiveValue.getValueID())
            lengthOrPercentageValue = primitiveValue.convertToLength<FixedIntegerConversion | PercentConversion | CalculatedConversion>(builderState.cssToLengthConversionData());
        else if (primitiveValue.getValueID() == CSSValueWebkitEachLine)
            textIndentLineValue = TextIndentEachLine;
        else if (primitiveValue.getValueID() == CSSValueWebkitHanging)
            textIndentTypeValue = TextIndentHanging;
    }

    if (lengthOrPercentageValue.isUndefined())
        return;

    builderState.style().setTextIndent(lengthOrPercentageValue);
    builderState.style().setTextIndentLine(textIndentLineValue);
    builderState.style().setTextIndentType(textIndentTypeValue);
}

enum BorderImageType { BorderImage, WebkitMaskBoxImage };
enum BorderImageModifierType { Outset, Repeat, Slice, Width };
template <BorderImageType type, BorderImageModifierType modifier>
class ApplyPropertyBorderImageModifier {
public:
    static void applyInheritValue(Style::BuilderState& builderState)
    {
        NinePieceImage image(getValue(builderState.style()));
        switch (modifier) {
        case Outset:
            image.copyOutsetFrom(getValue(builderState.parentStyle()));
            break;
        case Repeat:
            image.copyRepeatFrom(getValue(builderState.parentStyle()));
            break;
        case Slice:
            image.copyImageSlicesFrom(getValue(builderState.parentStyle()));
            break;
        case Width:
            image.copyBorderSlicesFrom(getValue(builderState.parentStyle()));
            break;
        }
        setValue(builderState.style(), image);
    }

    static void applyInitialValue(Style::BuilderState& builderState)
    {
        NinePieceImage image(getValue(builderState.style()));
        switch (modifier) {
        case Outset:
            image.setOutset(LengthBox(LengthType::Relative));
            break;
        case Repeat:
            image.setHorizontalRule(StretchImageRule);
            image.setVerticalRule(StretchImageRule);
            break;
        case Slice:
            // Masks have a different initial value for slices. Preserve the value of 0 for backwards compatibility.
            image.setImageSlices(type == BorderImage ? LengthBox(Length(100, LengthType::Percent), Length(100, LengthType::Percent), Length(100, LengthType::Percent), Length(100, LengthType::Percent)) : LengthBox());
            image.setFill(false);
            break;
        case Width:
            // FIXME: This is a local variable to work around a bug in the GCC 8.1 Address Sanitizer.
            // Might be slightly less efficient when the type is not BorderImage since this is unused in that case.
            // Should be switched back to a temporary when possible. See https://webkit.org/b/186980
            LengthBox lengthBox(Length(1, LengthType::Relative), Length(1, LengthType::Relative), Length(1, LengthType::Relative), Length(1, LengthType::Relative));
            // Masks have a different initial value for widths. They use an 'auto' value rather than trying to fit to the border.
            image.setBorderSlices(type == BorderImage ? lengthBox : LengthBox());
            break;
        }
        setValue(builderState.style(), image);
    }

    static void applyValue(Style::BuilderState& builderState, CSSValue& value)
    {
        NinePieceImage image(getValue(builderState.style()));
        switch (modifier) {
        case Outset:
            image.setOutset(builderState.styleMap().mapNinePieceImageQuad(value));
            break;
        case Repeat:
        	builderState.styleMap().mapNinePieceImageRepeat(value, image);
            break;
        case Slice:
        	builderState.styleMap().mapNinePieceImageSlice(value, image);
            break;
        case Width:
            image.setBorderSlices(builderState.styleMap().mapNinePieceImageQuad(value));
            break;
        }
        setValue(builderState.style(), image);
    }

private:
    static const NinePieceImage& getValue(const RenderStyle& style)
    {
        return type == BorderImage ? style.borderImage() : style.maskBoxImage();
    }

    static void setValue(RenderStyle& style, const NinePieceImage& value)
    {
        return type == BorderImage ? style.setBorderImage(value) : style.setMaskBoxImage(value);
    }
};

#define DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(type, modifier) \
inline void StyleBuilderCustom::applyInherit##type##modifier(Style::BuilderState& builderState) \
{ \
    ApplyPropertyBorderImageModifier<type, modifier>::applyInheritValue(builderState); \
} \
inline void StyleBuilderCustom::applyInitial##type##modifier(Style::BuilderState& builderState) \
{ \
    ApplyPropertyBorderImageModifier<type, modifier>::applyInitialValue(builderState); \
} \
inline void StyleBuilderCustom::applyValue##type##modifier(Style::BuilderState& builderState, CSSValue& value) \
{ \
    ApplyPropertyBorderImageModifier<type, modifier>::applyValue(builderState, value); \
}

DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(BorderImage, Outset)
DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(BorderImage, Repeat)
DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(BorderImage, Slice)
DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(BorderImage, Width)
DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(WebkitMaskBoxImage, Outset)
DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(WebkitMaskBoxImage, Repeat)
DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(WebkitMaskBoxImage, Slice)
DEFINE_BORDER_IMAGE_MODIFIER_HANDLER(WebkitMaskBoxImage, Width)

#if ENABLE(IOS_TEXT_AUTOSIZING)

inline void StyleBuilderCustom::applyInheritLineHeight(Style::BuilderState& builderState)
{
    builderState.style().setLineHeight(builderState.parentStyle().lineHeight());
    builderState.style().setSpecifiedLineHeight(builderState.parentStyle().specifiedLineHeight());
}

inline void StyleBuilderCustom::applyInitialLineHeight(Style::BuilderState& builderState)
{
    builderState.style().setLineHeight(RenderStyle::initialLineHeight());
    builderState.style().setSpecifiedLineHeight(RenderStyle::initialSpecifiedLineHeight());
}

inline void StyleBuilderCustom::applyValueLineHeight(Style::BuilderState& builderState, CSSValue& value)
{
    float multiplier = builderState.style().textSizeAdjust().isPercentage() ? builderState.style().textSizeAdjust().multiplier() : 1.f;
    Optional<Length> lineHeight = StyleBuilderConverter::convertLineHeight(builderState, value, multiplier);
    if (!lineHeight)
        return;

    builderState.style().setLineHeight(lineHeight.value());
    builderState.style().setSpecifiedLineHeight(lineHeight.value());
}
#endif

inline void StyleBuilderCustom::applyInheritOutlineStyle(Style::BuilderState& builderState)
{
    builderState.style().setOutlineStyleIsAuto(builderState.parentStyle().outlineStyleIsAuto());
    builderState.style().setOutlineStyle(builderState.parentStyle().outlineStyle());
}

inline void StyleBuilderCustom::applyInitialOutlineStyle(Style::BuilderState& builderState)
{
    builderState.style().setOutlineStyleIsAuto(RenderStyle::initialOutlineStyleIsAuto());
    builderState.style().setOutlineStyle(RenderStyle::initialBorderStyle());
}

inline void StyleBuilderCustom::applyValueOutlineStyle(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    builderState.style().setOutlineStyleIsAuto(primitiveValue);
    builderState.style().setOutlineStyle(primitiveValue);
}

inline void StyleBuilderCustom::applyInitialClip(Style::BuilderState& builderState)
{
    builderState.style().setClip(Length(), Length(), Length(), Length());
    builderState.style().setHasClip(false);
}

inline void StyleBuilderCustom::applyInheritClip(Style::BuilderState& builderState)
{
    auto& parentStyle = builderState.parentStyle();
    if (!parentStyle.hasClip())
        return applyInitialClip(builderState);
    builderState.style().setClip(parentStyle.clipTop(), parentStyle.clipRight(), parentStyle.clipBottom(), parentStyle.clipLeft());
    builderState.style().setHasClip(true);
}

inline void StyleBuilderCustom::applyValueClip(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    if (Rect* rect = primitiveValue.rectValue()) {
        auto conversionData = builderState.cssToLengthConversionData();
        Length top = rect->top()->convertToLength<FixedIntegerConversion | PercentConversion | AutoConversion>(conversionData);
        Length right = rect->right()->convertToLength<FixedIntegerConversion | PercentConversion | AutoConversion>(conversionData);
        Length bottom = rect->bottom()->convertToLength<FixedIntegerConversion | PercentConversion | AutoConversion>(conversionData);
        Length left = rect->left()->convertToLength<FixedIntegerConversion | PercentConversion | AutoConversion>(conversionData);
        builderState.style().setClip(top, right, bottom, left);
        builderState.style().setHasClip(true);
    } else {
        ASSERT(primitiveValue.getValueID() == CSSValueAuto);
        applyInitialClip(builderState);
    }
}

inline void StyleBuilderCustom::applyValueWebkitLocale(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);

    FontDescription fontDescription = builderState.style().fontDescription();
    if (primitiveValue.getValueID() == CSSValueAuto)
        fontDescription.setLocale(nullAtom);
    else
        fontDescription.setLocale(primitiveValue.stringValue());
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyValueWritingMode(Style::BuilderState& builderState, CSSValue& value)
{
	builderState.setWritingMode(toCSSPrimitiveValue(value));
    builderState.style().setHasExplicitlySetWritingMode(true);
}

void StyleBuilderCustom::applyValueAlt(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.isString())
        builderState.style().setContentAltText(primitiveValue.stringValue());
    else if (primitiveValue.isAttr()) {
        // FIXME: Can a namespace be specified for an attr(foo)?
        if (builderState.style().styleType() == PseudoId::None)
            builderState.style().setUnique();
        else
        	const_cast<RenderStyle&>(builderState.parentStyle()).setUnique();

        QualifiedName attr(nullAtom, primitiveValue.stringValue(), nullAtom);
        const AtomicString& value = builderState.element()->getAttribute(attr);
        builderState.style().setContentAltText(value.isNull() ? emptyAtom : value);

        // Register the fact that the attribute value affects the style.
        builderState.registerContentAttribute(attr.localName());
    } else
        builderState.style().setContentAltText(emptyAtom);
}

inline void StyleBuilderCustom::applyValueWebkitTextOrientation(Style::BuilderState& builderState, CSSValue& value)
{
	builderState.setTextOrientation(toCSSPrimitiveValue(value));
}

inline void StyleBuilderCustom::applyValueTextOrientation(Style::BuilderState& builderState, CSSValue& value)
{
    builderState.setTextOrientation(toCSSPrimitiveValue(value));
}

inline void StyleBuilderCustom::applyValueWebkitTextSizeAdjust(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
#if 0
    if (primitiveValue.getValueID() == CSSValueAuto)
        builderState.style().setTextSizeAdjust(TextSizeAdjustment(AutoTextSizeAdjustment));
    else if (primitiveValue.getValueID() == CSSValueNone)
        builderState.style().setTextSizeAdjust(TextSizeAdjustment(NoTextSizeAdjustment));
    else
        builderState.style().setTextSizeAdjust(TextSizeAdjustment(primitiveValue.floatValue()));
#else
    builderState.style().setTextSizeAdjust(primitiveValue.getValueID() == CSSValueAuto);
#endif
    builderState.setFontDirty();
}

inline void StyleBuilderCustom::applyValueWebkitTextZoom(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.getValueID() == CSSValueNormal)
        builderState.style().setTextZoom(TextZoomNormal);
    else if (primitiveValue.getValueID() == CSSValueReset)
        builderState.style().setTextZoom(TextZoomReset);
    builderState.setFontDirty();
}

template <CSSPropertyID id>
inline void StyleBuilderCustom::applyTextOrBoxShadowValue(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        if (id == CSSPropertyTextShadow)
            builderState.style().setTextShadow(nullptr);
        else
            builderState.style().setBoxShadow(nullptr);
        return;
    }

    bool isFirstEntry = true;
    for (auto& item : toCSSValueList(value)) {
    	CSSShadowValue& shadowValue = toCSSShadowValue(item.get());
        auto& conversionData = builderState.cssToLengthConversionData();
        Length x = shadowValue.x->computeLength<Length>(conversionData);
        Length y = shadowValue.y->computeLength<Length>(conversionData);
        Length blur = shadowValue.blur ? shadowValue.blur->computeLength<Length>(conversionData) : Length(0, LengthType::Fixed);
        Length spread = shadowValue.spread ? shadowValue.spread->computeLength<Length>(conversionData) : Length(0, LengthType::Fixed);
        ShadowStyle shadowStyle = shadowValue.style && shadowValue.style->getValueID() == CSSValueInset ? ShadowStyle::Inset : ShadowStyle::Normal;
        Color color;
        if (shadowValue.color)
            color = builderState.colorFromPrimitiveValueWithResolvedCurrentColor(*shadowValue.color);
        else
            color = builderState.style().color();

        OwnPtr<ShadowData> shadowData = adoptPtr(new ShadowData(LengthPoint(x, y), blur, spread, shadowStyle, id == CSSPropertyWebkitBoxShadow, color.isValid() ? color : Color::transparent));
        if (id == CSSPropertyTextShadow)
            builderState.style().setTextShadow(shadowData.release(), !isFirstEntry); // add to the list if this is not the first entry
        else
            builderState.style().setBoxShadow(shadowData.release(), !isFirstEntry); // add to the list if this is not the first entry
        isFirstEntry = false;
    }
}

inline void StyleBuilderCustom::applyInitialTextShadow(Style::BuilderState& builderState)
{
    builderState.style().setTextShadow(nullptr);
}

inline void StyleBuilderCustom::applyInheritTextShadow(Style::BuilderState& builderState)
{
    builderState.style().setTextShadow(builderState.parentStyle().textShadow() ? adoptPtr(new ShadowData(*builderState.parentStyle().textShadow())) : nullptr);
}

inline void StyleBuilderCustom::applyValueTextShadow(Style::BuilderState& builderState, CSSValue& value)
{
    applyTextOrBoxShadowValue<CSSPropertyTextShadow>(builderState, value);
}

inline void StyleBuilderCustom::applyInitialBoxShadow(Style::BuilderState& builderState)
{
    builderState.style().setBoxShadow(nullptr);
}

inline void StyleBuilderCustom::applyInheritBoxShadow(Style::BuilderState& builderState)
{
    builderState.style().setBoxShadow(builderState.parentStyle().boxShadow() ? adoptPtr(new ShadowData(*builderState.parentStyle().boxShadow())) : nullptr);
}

inline void StyleBuilderCustom::applyValueBoxShadow(Style::BuilderState& builderState, CSSValue& value)
{
    applyTextOrBoxShadowValue<CSSPropertyBoxShadow>(builderState, value);
}

inline void StyleBuilderCustom::applyInitialWebkitBoxShadow(Style::BuilderState& builderState)
{
    applyInitialBoxShadow(builderState);
}

inline void StyleBuilderCustom::applyInheritWebkitBoxShadow(Style::BuilderState& builderState)
{
    applyInheritBoxShadow(builderState);
}

inline void StyleBuilderCustom::applyValueWebkitBoxShadow(Style::BuilderState& builderState, CSSValue& value)
{
    applyTextOrBoxShadowValue<CSSPropertyWebkitBoxShadow>(builderState, value);
}

inline void StyleBuilderCustom::applyInitialFontFamily(Style::BuilderState& builderState)
{
    FontDescription fontDescription = builderState.style().fontDescription();
    FontDescription initialDesc = FontDescription();

    // We need to adjust the size to account for the generic family change from monospace to non-monospace.
    if (fontDescription.useFixedDefaultSize()) {
		if (CSSValueID sizeIdentifier = fontDescription.keywordSizeAsIdentifier())
			builderState.setFontSize(fontDescription, Style::fontSizeForKeyword(sizeIdentifier, false, builderState.document()));
	}
    if (!initialDesc.firstFamily().isEmpty())
        fontDescription.setFamilies(initialDesc.families());

    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInheritFontFamily(Style::BuilderState& builderState)
{
    FontDescription fontDescription = builderState.style().fontDescription();
    FontDescription parentFontDescription = builderState.parentStyle().fontDescription();

    fontDescription.setFamilies(parentFontDescription.families());
    fontDescription.setIsSpecifiedFont(parentFontDescription.isSpecifiedFont());
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyValueFontFamily(Style::BuilderState& builderState, CSSValue& value)
{
    FontDescription fontDescription = builderState.style().fontDescription();
    // Before mapping in a new font-family property, we should reset the generic family.
    bool oldFamilyUsedFixedDefaultSize = fontDescription.useFixedDefaultSize();

    Vector<AtomicString> families;

    if (value.isPrimitiveValue()) {
        auto valueID = toCSSPrimitiveValue(value).getValueID();
#if 0
        if (!CSSPropertyParserHelpers::isSystemFontShorthand(valueID)) {
            // Early return if the invalid CSSValueID is set while using CSSOM API.
            return;
        }
        AtomString family = SystemFontDatabase::singleton().systemFontShorthandFamily(CSSPropertyParserHelpers::lowerFontShorthand(valueID));
        ASSERT(!family.isEmpty());
        fontDescription.setIsSpecifiedFont(false);
        families = Vector<AtomString>::from(WTFMove(family));
#else
        return;
#endif
    } else {
        auto& valueList = toCSSValueList(value);
		for (auto& item : valueList) {
			auto& contentValue = toCSSPrimitiveValue(item.get());
			AtomicString family;
			bool isGenericFamily = false;
			if (contentValue.isFontFamily()) {
				const CSSFontFamily& fontFamily = contentValue.fontFamily();
				family = fontFamily.familyName;
				// If the family name was resolved by the CSS parser from a system font ID, then it is generic.
				isGenericFamily = fontFamily.fromSystemFontID;
			} else {
				switch (contentValue.getValueID()) {
				case CSSValueWebkitBody:
					family = builderState.document().settings().standardFontFamily();
					break;
				case CSSValueSerif:
					family = serifFamily;
					isGenericFamily = true;
					break;
				case CSSValueSansSerif:
					family = sansSerifFamily;
					isGenericFamily = true;
					break;
				case CSSValueCursive:
					family = cursiveFamily;
					isGenericFamily = true;
					break;
				case CSSValueFantasy:
					family = fantasyFamily;
					isGenericFamily = true;
					break;
				case CSSValueMonospace:
					family = monospaceFamily;
					isGenericFamily = true;
					break;
				case CSSValueWebkitPictograph:
					family = pictographFamily;
					isGenericFamily = true;
					break;
				default:
					break;
				}
			}

			if (family.isEmpty())
				continue;
			if (families.isEmpty())
				fontDescription.setIsSpecifiedFont(!isGenericFamily);
			families.append(family);
		}

	    if (families.isEmpty())
	        return;
    }

    fontDescription.setFamilies(families);

    if (fontDescription.useFixedDefaultSize() != oldFamilyUsedFixedDefaultSize) {
        if (CSSValueID sizeIdentifier = fontDescription.keywordSizeAsIdentifier())
        	builderState.setFontSize(fontDescription, Style::fontSizeForKeyword(sizeIdentifier, !oldFamilyUsedFixedDefaultSize, builderState.document()));
    }

    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyValueBaselineShift(Style::BuilderState& builderState, CSSValue& value)
{
    SVGRenderStyle& svgStyle = builderState.style().accessSVGStyle();
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.isValueID()) {
        switch (primitiveValue.getValueID()) {
        case CSSValueBaseline:
            svgStyle.setBaselineShift(BS_BASELINE);
            break;
        case CSSValueSub:
            svgStyle.setBaselineShift(BS_SUB);
            break;
        case CSSValueSuper:
            svgStyle.setBaselineShift(BS_SUPER);
            break;
        default:
            break;
        }
    } else {
        svgStyle.setBaselineShift(BS_LENGTH);
        svgStyle.setBaselineShiftValue(SVGLengthValue::fromCSSPrimitiveValue(primitiveValue, builderState.cssToLengthConversionData()));
    }
}

inline void StyleBuilderCustom::applyInitialAspectRatio(Style::BuilderState& builderState)
{
	builderState.style().setAspectRatioType(RenderStyle::initialAspectRatioType());
	builderState.style().setAspectRatio(RenderStyle::initialAspectRatioWidth(), RenderStyle::initialAspectRatioHeight());
}

inline void StyleBuilderCustom::applyInheritAspectRatio(Style::BuilderState&)
{
}

inline void StyleBuilderCustom::applyValueAspectRatio(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).valueID() == CSSValueAuto);
        return builderState.style().setAspectRatioType(AspectRatioType::Auto);
    }

    if (!value.isValueList())
        return;

    auto& list = toCSSValueList(value);
    if (list.item(1)->isValueList()) {
    	builderState.style().setAspectRatioType(AspectRatioType::AutoAndRatio);
        auto ratioList = toCSSValueList(list.item(1));
        ASSERT(ratioList->length() == 2);
        builderState.style().setAspectRatio(toCSSPrimitiveValue(ratioList->item(0))->doubleValue(), toCSSPrimitiveValue(ratioList->item(1))->doubleValue());
        return;
    }

    ASSERT(list.length() == 2);
    auto width = toCSSPrimitiveValue(list.item(0))->doubleValue();
    auto height = toCSSPrimitiveValue(list.item(1))->doubleValue();
    if (!width || !height)
    	builderState.style().setAspectRatioType(AspectRatioType::AutoZero);
    else
    	builderState.style().setAspectRatioType(AspectRatioType::Ratio);
    builderState.style().setAspectRatio(width, height);
}

inline void StyleBuilderCustom::applyInitialWebkitTextEmphasisStyle(Style::BuilderState& builderState)
{
    builderState.style().setTextEmphasisFill(RenderStyle::initialTextEmphasisFill());
    builderState.style().setTextEmphasisMark(RenderStyle::initialTextEmphasisMark());
    builderState.style().setTextEmphasisCustomMark(RenderStyle::initialTextEmphasisCustomMark());
}

inline void StyleBuilderCustom::applyInheritWebkitTextEmphasisStyle(Style::BuilderState& builderState)
{
    builderState.style().setTextEmphasisFill(builderState.parentStyle().textEmphasisFill());
    builderState.style().setTextEmphasisMark(builderState.parentStyle().textEmphasisMark());
    builderState.style().setTextEmphasisCustomMark(builderState.parentStyle().textEmphasisCustomMark());
}

inline void StyleBuilderCustom::applyValueWebkitTextEmphasisStyle(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isValueList()) {
        auto& list = toCSSValueList(value);
        ASSERT(list.length() == 2);

        for (auto& item : list) {
            CSSPrimitiveValue& value = toCSSPrimitiveValue(item.get());
            if (value.getValueID() == CSSValueFilled || value.getValueID() == CSSValueOpen)
                builderState.style().setTextEmphasisFill(value);
            else
                builderState.style().setTextEmphasisMark(value);
        }
        builderState.style().setTextEmphasisCustomMark(nullAtom);
        return;
    }

    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (primitiveValue.isString()) {
        builderState.style().setTextEmphasisFill(TextEmphasisFillFilled);
        builderState.style().setTextEmphasisMark(TextEmphasisMarkCustom);
        builderState.style().setTextEmphasisCustomMark(primitiveValue.stringValue());
        return;
    }

    builderState.style().setTextEmphasisCustomMark(nullAtom);

    if (primitiveValue.getValueID() == CSSValueFilled || primitiveValue.getValueID() == CSSValueOpen) {
        builderState.style().setTextEmphasisFill(primitiveValue);
        builderState.style().setTextEmphasisMark(TextEmphasisMarkAuto);
    } else {
        builderState.style().setTextEmphasisFill(TextEmphasisFillFilled);
        builderState.style().setTextEmphasisMark(primitiveValue);
    }
}

template <StyleBuilderCustom::CounterBehavior counterBehavior>
inline void StyleBuilderCustom::applyInheritCounter(Style::BuilderState& builderState)
{
    CounterDirectiveMap& map = builderState.style().accessCounterDirectives();
    for (auto& keyValue : const_cast<RenderStyle&>(builderState.parentStyle()).accessCounterDirectives()) {
        CounterDirectives& directives = map.add(keyValue.first, CounterDirectives()).iterator->second;
        if (counterBehavior == Reset)
            directives.resetValue = keyValue.second.resetValue;
        else
            directives.incrementValue = keyValue.second.incrementValue;
    }
}

template <StyleBuilderCustom::CounterBehavior counterBehavior>
inline void StyleBuilderCustom::applyValueCounter(Style::BuilderState& builderState, CSSValue& value)
{
    bool setCounterIncrementToNone = counterBehavior == Increment && value.isPrimitiveValue() && toCSSPrimitiveValue(value).getValueID() == CSSValueNone;

    if (!value.isValueList() && !setCounterIncrementToNone)
        return;

    CounterDirectiveMap& map = builderState.style().accessCounterDirectives();
    for (auto& keyValue : map) {
        if (counterBehavior == Reset)
        	keyValue.second.resetValue = Nullopt;
        else
        	keyValue.second.incrementValue = Nullopt;
    }

    if (setCounterIncrementToNone)
        return;

    for (auto& item : toCSSValueList(value)) {
        Pair* pair = toCSSPrimitiveValue(item.get()).pairValue();
        if (!pair || !pair->first() || !pair->second())
            continue;

        AtomicString identifier = pair->first()->stringValue();
        int value = pair->second()->intValue();
        CounterDirectives& directives = map.add(identifier, CounterDirectives()).iterator->second;
        if (counterBehavior == Reset)
			directives.resetValue = value;
		else
			directives.incrementValue = saturatedAddition(directives.incrementValue.value_or(0), value);
    }
}

inline void StyleBuilderCustom::applyInitialCounterIncrement(Style::BuilderState&) { }

inline void StyleBuilderCustom::applyInheritCounterIncrement(Style::BuilderState& builderState)
{
    applyInheritCounter<Increment>(builderState);
}

inline void StyleBuilderCustom::applyValueCounterIncrement(Style::BuilderState& builderState, CSSValue& value)
{
    applyValueCounter<Increment>(builderState, value);
}

inline void StyleBuilderCustom::applyInitialCounterReset(Style::BuilderState&) { }

inline void StyleBuilderCustom::applyInheritCounterReset(Style::BuilderState& builderState)
{
    applyInheritCounter<Reset>(builderState);
}

inline void StyleBuilderCustom::applyValueCounterReset(Style::BuilderState& builderState, CSSValue& value)
{
    applyValueCounter<Reset>(builderState, value);
}

inline void StyleBuilderCustom::applyInitialCursor(Style::BuilderState& builderState)
{
    builderState.style().clearCursorList();
    builderState.style().setCursor(RenderStyle::initialCursor());
}

inline void StyleBuilderCustom::applyInheritCursor(Style::BuilderState& builderState)
{
    builderState.style().setCursor(builderState.parentStyle().cursor());
    builderState.style().setCursorList(builderState.parentStyle().cursors());
}

inline void StyleBuilderCustom::applyValueCursor(Style::BuilderState& builderState, CSSValue& value)
{
    builderState.style().clearCursorList();
    if (value.isPrimitiveValue()) {
        ECursor cursor = toCSSPrimitiveValue(value);
        if (builderState.style().cursor() != cursor)
            builderState.style().setCursor(cursor);
        return;
    }

    builderState.style().setCursor(CURSOR_AUTO);
    auto& list = toCSSValueList(value);
    for (auto& item : list) {
        if (item.get().isCursorImageValue()) {
            auto& image = toCSSCursorImageValue(item.get());
            builderState.style().addCursor(builderState.createStyleImage(image), image.hotSpot());
            continue;
        }

        builderState.style().setCursor(toCSSPrimitiveValue(item.get()));
        ASSERT_WITH_MESSAGE(item.ptr() == list.item(list.length() - 1), "Cursor ID fallback should always be last in the list");
        return;
    }
}

inline void StyleBuilderCustom::applyInitialFill(Style::BuilderState& builderState)
{
    SVGRenderStyle& svgStyle = builderState.style().accessSVGStyle();
    svgStyle.setFillPaint(SVGRenderStyle::initialFillPaintType(), SVGRenderStyle::initialFillPaintColor(), SVGRenderStyle::initialFillPaintUri(), builderState.applyPropertyToRegularStyle(), builderState.applyPropertyToVisitedLinkStyle());
}

inline void StyleBuilderCustom::applyInheritFill(Style::BuilderState& builderState)
{
    SVGRenderStyle& svgStyle = builderState.style().accessSVGStyle();
    auto& svgParentStyle = builderState.parentStyle().svgStyle();
    svgStyle.setFillPaint(svgParentStyle.fillPaintType(), svgParentStyle.fillPaintColor(), svgParentStyle.fillPaintUri(), builderState.applyPropertyToRegularStyle(), builderState.applyPropertyToVisitedLinkStyle());

}

inline void StyleBuilderCustom::applyValueFill(Style::BuilderState& builderState, CSSValue& value)
{
    SVGRenderStyle& svgStyle = builderState.style().accessSVGStyle();
    const CSSPrimitiveValue* localValue = value.isPrimitiveValue() ? &toCSSPrimitiveValue(value) : 0;
    String url;
    if (value.isValueList()) {
        CSSValueList& list = toCSSValueList(value);
        url = toCSSPrimitiveValue(list.item(0))->stringValue();
        localValue = toCSSPrimitiveValue(list.item(1));
    }

    if (!localValue)
        return;

    Color color;
    SVGPaintType paintType = SVG_PAINTTYPE_RGBCOLOR;
    if (localValue->isURI()) {
        paintType = SVG_PAINTTYPE_URI;
        url = localValue->stringValue();
    } else if (localValue->isValueID() && localValue->getValueID() == CSSValueNone)
        paintType = url.isEmpty() ? SVG_PAINTTYPE_NONE : SVG_PAINTTYPE_URI_NONE;
    else if (localValue->isValueID() && localValue->getValueID() == CSSValueCurrentcolor) {
        color = builderState.style().color();
        paintType = url.isEmpty() ? SVG_PAINTTYPE_CURRENTCOLOR : SVG_PAINTTYPE_URI_CURRENTCOLOR;
        builderState.style().setDisallowsFastPathInheritance();
    } else {
        color = builderState.colorFromPrimitiveValue(*localValue);
        paintType = url.isEmpty() ? SVG_PAINTTYPE_RGBCOLOR : SVG_PAINTTYPE_URI_RGBCOLOR;
    }
    svgStyle.setFillPaint(paintType, color, url, builderState.applyPropertyToRegularStyle(), builderState.applyPropertyToVisitedLinkStyle());
}

inline void StyleBuilderCustom::applyInitialStroke(Style::BuilderState& builderState)
{
    SVGRenderStyle& svgStyle = builderState.style().accessSVGStyle();
    svgStyle.setStrokePaint(SVGRenderStyle::initialStrokePaintType(), SVGRenderStyle::initialStrokePaintColor(), SVGRenderStyle::initialStrokePaintUri(), builderState.applyPropertyToRegularStyle(), builderState.applyPropertyToVisitedLinkStyle());
}

inline void StyleBuilderCustom::applyInheritStroke(Style::BuilderState& builderState)
{
    SVGRenderStyle& svgStyle = builderState.style().accessSVGStyle();
    auto& svgParentStyle = builderState.parentStyle().svgStyle();
    svgStyle.setStrokePaint(svgParentStyle.strokePaintType(), svgParentStyle.strokePaintColor(), svgParentStyle.strokePaintUri(), builderState.applyPropertyToRegularStyle(), builderState.applyPropertyToVisitedLinkStyle());
}

inline void StyleBuilderCustom::applyValueStroke(Style::BuilderState& builderState, CSSValue& value)
{
    SVGRenderStyle& svgStyle = builderState.style().accessSVGStyle();
    const CSSPrimitiveValue* localValue = value.isPrimitiveValue() ? &toCSSPrimitiveValue(value) : 0;
    String url;
    if (value.isValueList()) {
        CSSValueList& list = toCSSValueList(value);
        url = toCSSPrimitiveValue(list.item(0))->stringValue();
        localValue = toCSSPrimitiveValue(list.item(1));
    }

    if (!localValue)
        return;

    Color color;
    SVGPaintType paintType = SVG_PAINTTYPE_RGBCOLOR;
    if (localValue->isURI()) {
        paintType = SVG_PAINTTYPE_URI;
        url = localValue->stringValue();
    } else if (localValue->isValueID() && localValue->getValueID() == CSSValueNone)
        paintType = url.isEmpty() ? SVG_PAINTTYPE_NONE : SVG_PAINTTYPE_URI_NONE;
    else if (localValue->isValueID() && localValue->getValueID() == CSSValueCurrentcolor) {
        color = builderState.style().color();
        paintType = url.isEmpty() ? SVG_PAINTTYPE_CURRENTCOLOR : SVG_PAINTTYPE_URI_CURRENTCOLOR;
        builderState.style().setDisallowsFastPathInheritance();
    } else {
        color = builderState.colorFromPrimitiveValue(*localValue);
        paintType = url.isEmpty() ? SVG_PAINTTYPE_RGBCOLOR : SVG_PAINTTYPE_URI_RGBCOLOR;
    }
    svgStyle.setStrokePaint(paintType, color, url, builderState.applyPropertyToRegularStyle(), builderState.applyPropertyToVisitedLinkStyle());
}

inline void StyleBuilderCustom::applyInitialContent(Style::BuilderState& builderState)
{
    builderState.style().clearContent();
}

inline void StyleBuilderCustom::applyInheritContent(Style::BuilderState&)
{
    // FIXME: In CSS3, it will be possible to inherit content. In CSS2 it is not. This
    // note is a reminder that eventually "inherit" needs to be supported.
}

inline void StyleBuilderCustom::applyValueContent(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        auto& primitiveValue = toCSSPrimitiveValue(value);
        //ASSERT_UNUSED(primitiveValue, primitiveValue.valueID() == CSSValueNormal || primitiveValue.valueID() == CSSValueNone);
        builderState.style().clearContent();
        return;
    }

    bool didSet = false;
    for (auto& item : toCSSValueList(value)) {
        if (item.get().isImageGeneratorValue()) {
            if (item.get().isGradientValue())
                builderState.style().setContent(StyleGeneratedImage::create(static_cast<CSSGradientValue&>(builderState.resolveImageStyles(item.get()).get())), didSet);
            else
                builderState.style().setContent(StyleGeneratedImage::create(static_cast<CSSImageGeneratorValue&>(builderState.resolveImageStyles(item.get()).get())), didSet);
            didSet = true;
        }

        if (item->isImageSetValue()) {
            builderState.style().setContent(StyleImageSet::create(static_cast<CSSImageSetValue&>(builderState.resolveImageStyles(item.get()).get())), didSet);
            didSet = true;
        }

        if (item->isImageValue()) {
            builderState.style().setContent(StyleCachedImage::create(static_cast<CSSImageValue&>(builderState.resolveImageStyles(item.get()).get())), didSet);
            didSet = true;
            continue;
        }

        if (!item.get().isPrimitiveValue())
            continue;

        auto& contentValue = toCSSPrimitiveValue(item.get());

        if (contentValue.isString()) {
        	builderState.style().setContent(contentValue.stringValue().impl(), didSet);
            didSet = true;
        } else if (contentValue.isAttr()) {
            // FIXME: Can a namespace be specified for an attr(foo)?
            if (builderState.style().styleType() == PseudoId::None)
                builderState.style().setHasAttrContent();
            else
            	const_cast<RenderStyle&>(builderState.parentStyle()).setHasAttrContent();
            QualifiedName attr(nullAtom, contentValue.stringValue().impl(), nullAtom);
            const AtomicString& value = builderState.element()->getAttribute(attr);
            builderState.style().setContent(value.isNull() ? emptyAtom : value.impl(), didSet);
            didSet = true;
            // register the fact that the attribute value affects the style
            builderState.registerContentAttribute(attr.localName());
        } else if (contentValue.isCounter()) {
            Counter* counterValue = contentValue.counterValue();
            EListStyleType listStyleType = NoneListStyle;
            CSSValueID listStyleIdent = counterValue->listStyleIdent();
            if (listStyleIdent != CSSValueNone)
                listStyleType = static_cast<EListStyleType>(listStyleIdent - CSSValueDisc);
            OwnPtr<CounterContent> counter = adoptPtr(new CounterContent(counterValue->identifier(), listStyleType, counterValue->separator()));
            builderState.style().setContent(counter.release(), didSet);
            didSet = true;
        } else {
            switch (contentValue.getValueID()) {
            case CSSValueOpenQuote:
                builderState.style().setContent(OPEN_QUOTE, didSet);
                didSet = true;
                break;
            case CSSValueCloseQuote:
                builderState.style().setContent(CLOSE_QUOTE, didSet);
                didSet = true;
                break;
            case CSSValueNoOpenQuote:
                builderState.style().setContent(NO_OPEN_QUOTE, didSet);
                didSet = true;
                break;
            case CSSValueNoCloseQuote:
                builderState.style().setContent(NO_CLOSE_QUOTE, didSet);
                didSet = true;
                break;
            default:
                // normal and none do not have any effect.
                break;
            }
        }
    }
    if (!didSet)
        builderState.style().clearContent();
}

inline void StyleBuilderCustom::applyInheritFontVariantLigatures(Style::BuilderState& builderState)
{
    auto fontDescription = builderState.fontDescription();
    fontDescription.setVariantCommonLigatures(builderState.parentFontDescription().variantCommonLigatures());
    fontDescription.setVariantDiscretionaryLigatures(builderState.parentFontDescription().variantDiscretionaryLigatures());
    fontDescription.setVariantHistoricalLigatures(builderState.parentFontDescription().variantHistoricalLigatures());
    fontDescription.setVariantContextualAlternates(builderState.parentFontDescription().variantContextualAlternates());
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInitialFontVariantLigatures(Style::BuilderState& builderState)
{
    auto fontDescription = builderState.fontDescription();
    fontDescription.setVariantCommonLigatures(FontVariantLigatures::Normal);
    fontDescription.setVariantDiscretionaryLigatures(FontVariantLigatures::Normal);
    fontDescription.setVariantHistoricalLigatures(FontVariantLigatures::Normal);
    fontDescription.setVariantContextualAlternates(FontVariantLigatures::Normal);
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyValueFontVariantLigatures(Style::BuilderState& builderState, CSSValue& value)
{
    auto fontDescription = builderState.fontDescription();
    auto variantLigatures = extractFontVariantLigatures(value);
    fontDescription.setVariantCommonLigatures(variantLigatures.commonLigatures);
    fontDescription.setVariantDiscretionaryLigatures(variantLigatures.discretionaryLigatures);
    fontDescription.setVariantHistoricalLigatures(variantLigatures.historicalLigatures);
    fontDescription.setVariantContextualAlternates(variantLigatures.contextualAlternates);
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInheritFontVariantNumeric(Style::BuilderState& builderState)
{
    auto fontDescription = builderState.fontDescription();
    fontDescription.setVariantNumericFigure(builderState.parentFontDescription().variantNumericFigure());
    fontDescription.setVariantNumericSpacing(builderState.parentFontDescription().variantNumericSpacing());
    fontDescription.setVariantNumericFraction(builderState.parentFontDescription().variantNumericFraction());
    fontDescription.setVariantNumericOrdinal(builderState.parentFontDescription().variantNumericOrdinal());
    fontDescription.setVariantNumericSlashedZero(builderState.parentFontDescription().variantNumericSlashedZero());
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInitialFontVariantNumeric(Style::BuilderState& builderState)
{
    auto fontDescription = builderState.fontDescription();
    fontDescription.setVariantNumericFigure(FontVariantNumericFigure::Normal);
    fontDescription.setVariantNumericSpacing(FontVariantNumericSpacing::Normal);
    fontDescription.setVariantNumericFraction(FontVariantNumericFraction::Normal);
    fontDescription.setVariantNumericOrdinal(FontVariantNumericOrdinal::Normal);
    fontDescription.setVariantNumericSlashedZero(FontVariantNumericSlashedZero::Normal);
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyValueFontVariantNumeric(Style::BuilderState& builderState, CSSValue& value)
{
    auto fontDescription = builderState.fontDescription();
    auto variantNumeric = extractFontVariantNumeric(value);
    fontDescription.setVariantNumericFigure(variantNumeric.figure);
    fontDescription.setVariantNumericSpacing(variantNumeric.spacing);
    fontDescription.setVariantNumericFraction(variantNumeric.fraction);
    fontDescription.setVariantNumericOrdinal(variantNumeric.ordinal);
    fontDescription.setVariantNumericSlashedZero(variantNumeric.slashedZero);
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInheritFontVariantEastAsian(Style::BuilderState& builderState)
{
    auto fontDescription = builderState.fontDescription();
    fontDescription.setVariantEastAsianVariant(builderState.parentFontDescription().variantEastAsianVariant());
    fontDescription.setVariantEastAsianWidth(builderState.parentFontDescription().variantEastAsianWidth());
    fontDescription.setVariantEastAsianRuby(builderState.parentFontDescription().variantEastAsianRuby());
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInitialFontVariantEastAsian(Style::BuilderState& builderState)
{
    auto fontDescription = builderState.fontDescription();
    fontDescription.setVariantEastAsianVariant(FontVariantEastAsianVariant::Normal);
    fontDescription.setVariantEastAsianWidth(FontVariantEastAsianWidth::Normal);
    fontDescription.setVariantEastAsianRuby(FontVariantEastAsianRuby::Normal);
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyValueFontVariantEastAsian(Style::BuilderState& builderState, CSSValue& value)
{
    auto fontDescription = builderState.fontDescription();
    auto variantEastAsian = extractFontVariantEastAsian(value);
    fontDescription.setVariantEastAsianVariant(variantEastAsian.variant);
    fontDescription.setVariantEastAsianWidth(variantEastAsian.width);
    fontDescription.setVariantEastAsianRuby(variantEastAsian.ruby);
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInitialFontSize(Style::BuilderState& builderState)
{
    FontDescription fontDescription = builderState.style().fontDescription();
    float size = Style::fontSizeForKeyword(CSSValueMedium, fontDescription.useFixedDefaultSize(), builderState.document());

    if (size < 0)
        return;

    fontDescription.setKeywordSizeFromIdentifier(CSSValueMedium);
    builderState.setFontSize(fontDescription, size);
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInheritFontSize(Style::BuilderState& builderState)
{
    const FontDescription& parentFontDescription = builderState.parentStyle().fontDescription();
    float size = parentFontDescription.specifiedSize();

    if (size < 0)
        return;

    FontDescription fontDescription = builderState.style().fontDescription();
    fontDescription.setKeywordSize(parentFontDescription.keywordSize());
    builderState.setFontSize(fontDescription, size);
    builderState.setFontDescription(fontDescription);
}

// When the CSS keyword "larger" is used, this function will attempt to match within the keyword
// table, and failing that, will simply multiply by 1.2.
inline float StyleBuilderCustom::largerFontSize(float size)
{
    // FIXME: Figure out where we fall in the size ranges (xx-small to xxx-large) and scale up to
    // the next size level.
    return size * 1.2f;
}

// Like the previous function, but for the keyword "smaller".
inline float StyleBuilderCustom::smallerFontSize(float size)
{
    // FIXME: Figure out where we fall in the size ranges (xx-small to xxx-large) and scale down to
    // the next size level.
    return size / 1.2f;
}

inline float StyleBuilderCustom::determineRubyTextSizeMultiplier(Style::BuilderState& builderState)
{
    if (builderState.style().rubyPosition() != RubyPositionInterCharacter)
        return 0.5f;

    // FIXME: This hack is to ensure tone marks are the same size as
    // the bopomofo. This code will go away if we make a special renderer
    // for the tone marks eventually.
    if (auto* element = builderState.element()) {
        for (auto& ancestor : ancestorsOfType<HTMLElement>(*element)) {
            if (ancestor.hasTagName(HTMLNames::rtTag))
                return 1.0f;
        }
    }
    return 0.25f;
}

inline void StyleBuilderCustom::applyValueFontSize(Style::BuilderState& builderState, CSSValue& value)
{
    FontDescription fontDescription = builderState.style().fontDescription();
    fontDescription.setKeywordSizeFromIdentifier(CSSValueInvalid);

    float parentSize = 0;
    bool parentIsAbsoluteSize = false;
	parentSize = builderState.parentStyle().fontDescription().specifiedSize();
	parentIsAbsoluteSize = builderState.parentStyle().fontDescription().isAbsoluteSize();

    auto& primitiveValue = toCSSPrimitiveValue(value);
    float size;
    if (CSSValueID ident = primitiveValue.getValueID()) {
        fontDescription.setIsAbsoluteSize(parentIsAbsoluteSize && (ident == CSSValueLarger || ident == CSSValueSmaller || ident == CSSValueWebkitRubyText));

        // Keywords are being used.
        switch (ident) {
        case CSSValueXxSmall:
        case CSSValueXSmall:
        case CSSValueSmall:
        case CSSValueMedium:
        case CSSValueLarge:
        case CSSValueXLarge:
        case CSSValueXxLarge:
        case CSSValueWebkitXxxLarge:
            size = Style::fontSizeForKeyword(ident, fontDescription.useFixedDefaultSize(), builderState.document());
            fontDescription.setKeywordSizeFromIdentifier(ident);
            break;
        case CSSValueLarger:
            size = largerFontSize(parentSize);
            break;
        case CSSValueSmaller:
            size = smallerFontSize(parentSize);
            break;
        case CSSValueWebkitRubyText:
            size = determineRubyTextSizeMultiplier(builderState) * parentSize;
            break;
        default:
            return;
        }
    } else {
    	fontDescription.setIsAbsoluteSize(parentIsAbsoluteSize || !primitiveValue.isParentFontRelativeLength());
    	if (primitiveValue.isLength()) {
            CSSToLengthConversionData conversionData(&builderState.parentStyle(), builderState.rootElementStyle(), &builderState.parentStyle(), builderState.document().renderView(), 1.0f, CSSPropertyFontSize, &builderState.style());
            size = primitiveValue.computeLength<float>(conversionData);
        } else if (primitiveValue.isPercentage())
            size = (primitiveValue.floatValue() * parentSize) / 100.0f;
        else if (primitiveValue.isCalculatedPercentageWithLength()) {
        	// FIXME: Why does this need a different root style than the isLength case above?
			CSSToLengthConversionData conversionData(&builderState.parentStyle(), builderState.cssToLengthConversionData().rootStyle(), &builderState.parentStyle(), builderState.document().renderView(), 1.0f, CSSPropertyFontSize, &builderState.style());
			size = primitiveValue.cssCalcValue()->createCalculationValue(conversionData)->evaluate(parentSize);
		} else
			return;
    }

    if (size < 0)
        return;

    builderState.setFontSize(fontDescription, std::min(maximumAllowedFontSize, size));
    builderState.setFontDescription(fontDescription);
}

inline void StyleBuilderCustom::applyInitialWebkitHorizontalFontSize(Style::BuilderState& builderState)
{
    FontDescription fontDescription = builderState.style().fontDescription();
    builderState.setHorizontalFontSize(fontDescription, 0);
    builderState.setFontDescription(fontDescription);
    return;
}

inline void StyleBuilderCustom::applyInheritWebkitHorizontalFontSize(Style::BuilderState& builderState)
{
    float size = builderState.parentStyle().fontDescription().specifiedHorizontalSize();
    if (size < 0)
        return;

    FontDescription fontDescription = builderState.style().fontDescription();
    builderState.setHorizontalFontSize(fontDescription, size);
    builderState.setFontDescription(fontDescription);
    return;
}

inline void StyleBuilderCustom::applyValueWebkitHorizontalFontSize(Style::BuilderState& builderState, CSSValue& value)
{
    if (!value.isPrimitiveValue())
        return;

    CSSPrimitiveValue& primitiveValue = toCSSPrimitiveValue(value);

    FontDescription fontDescription = builderState.style().fontDescription();
    float parentHorizontalSize = 0;
    bool parentIsAbsoluteHorizontalSize = false;
    float size = 0;

	parentHorizontalSize = builderState.parentStyle().fontDescription().specifiedHorizontalSize();
	parentIsAbsoluteHorizontalSize = builderState.parentStyle().fontDescription().isAbsoluteHorizontalSize();

    if (CSSValueID ident = primitiveValue.getValueID()) {
        // Keywords are being used.
		return;
    } else {
    	fontDescription.setIsAbsoluteSize(parentIsAbsoluteHorizontalSize || !primitiveValue.isParentFontRelativeLength());
    	if (primitiveValue.isLength()) {
        	CSSToLengthConversionData conversionData(&builderState.parentStyle(), builderState.rootElementStyle(), &builderState.parentStyle(), builderState.document().renderView(), 1.0f, CSSPropertyFontSize, &builderState.style());
			size = primitiveValue.computeLength<float>(conversionData);
        } else if (primitiveValue.isPercentage())
            size = (primitiveValue.floatValue() * parentHorizontalSize) / 100.0f;
        else if (primitiveValue.isCalculatedPercentageWithLength()) {
        	// FIXME: Why does this need a different root style than the isLength case above?
			CSSToLengthConversionData conversionData(&builderState.parentStyle(), builderState.cssToLengthConversionData().rootStyle(), &builderState.parentStyle(), builderState.document().renderView(), 1.0f, CSSPropertyFontSize, &builderState.style());
			size = primitiveValue.cssCalcValue()->createCalculationValue(conversionData)->evaluate(parentHorizontalSize);
		} else
            return;
    }

    if (size < 0)
        return;

    builderState.setHorizontalFontSize(fontDescription, std::min(maximumAllowedFontSize, size));
    builderState.setFontDescription(fontDescription);
    return;
}

#if ENABLE(CSS_GRID_LAYOUT)
inline void StyleBuilderCustom::applyInitialGridTemplateAreas(Style::BuilderState& builderState)
{
    builderState.style().setImplicitNamedGridColumnLines(RenderStyle::initialNamedGridColumnLines());
    builderState.style().setImplicitNamedGridRowLines(RenderStyle::initialNamedGridRowLines());

    builderState.style().setNamedGridArea(RenderStyle::initialNamedGridArea());
    builderState.style().setNamedGridAreaRowCount(RenderStyle::initialNamedGridAreaCount());
    builderState.style().setNamedGridAreaColumnCount(RenderStyle::initialNamedGridAreaCount());
}

inline void StyleBuilderCustom::applyInheritGridTemplateAreas(Style::BuilderState& builderState)
{
    builderState.style().setImplicitNamedGridColumnLines(builderState.parentStyle().implicitNamedGridColumnLines());
    builderState.style().setImplicitNamedGridRowLines(builderState.parentStyle().implicitNamedGridRowLines());

    builderState.style().setNamedGridArea(builderState.parentStyle().namedGridArea());
    builderState.style().setNamedGridAreaRowCount(builderState.parentStyle().namedGridAreaRowCount());
    builderState.style().setNamedGridAreaColumnCount(builderState.parentStyle().namedGridAreaColumnCount());
}

inline void StyleBuilderCustom::applyValueGridTemplateAreas(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueNone);
        applyInitialGridTemplateAreas(builderState);
        return;
    }

    auto& gridTemplateAreasValue = toCSSGridTemplateAreasValue(value);
    const NamedGridAreaMap& newNamedGridAreas = gridTemplateAreasValue.gridAreaMap();

    NamedGridLinesMap implicitNamedGridColumnLines;
    NamedGridLinesMap implicitNamedGridRowLines;
    StyleBuilderConverter::createImplicitNamedGridLinesFromGridArea(newNamedGridAreas, implicitNamedGridColumnLines, ForColumns);
    StyleBuilderConverter::createImplicitNamedGridLinesFromGridArea(newNamedGridAreas, implicitNamedGridRowLines, ForRows);
    builderState.style().setImplicitNamedGridColumnLines(implicitNamedGridColumnLines);
    builderState.style().setImplicitNamedGridRowLines(implicitNamedGridRowLines);

    builderState.style().setNamedGridArea(gridTemplateAreasValue.gridAreaMap());
    builderState.style().setNamedGridAreaRowCount(gridTemplateAreasValue.rowCount());
    builderState.style().setNamedGridAreaColumnCount(gridTemplateAreasValue.columnCount());
}

inline void StyleBuilderCustom::applyInitialGridTemplateColumns(Style::BuilderState& builderState)
{
    builderState.style().setGridColumns(RenderStyle::initialGridColumns());
    builderState.style().setNamedGridColumnLines(RenderStyle::initialNamedGridColumnLines());
    builderState.style().setOrderedNamedGridColumnLines(RenderStyle::initialOrderedNamedGridColumnLines());
}

inline void StyleBuilderCustom::applyInheritGridTemplateColumns(Style::BuilderState& builderState)
{
    builderState.style().setGridColumns(builderState.parentStyle().gridColumns());
    builderState.style().setNamedGridColumnLines(builderState.parentStyle().namedGridColumnLines());
    builderState.style().setOrderedNamedGridColumnLines(builderState.parentStyle().orderedNamedGridColumnLines());
}

#define SET_TRACKS_DATA(tracksData, style, TrackType) \
    style.setGrid##TrackType##s(tracksData.m_trackSizes); \
    style.setNamedGrid##TrackType##Lines(tracksData.m_namedGridLines); \
    style.setOrderedNamedGrid##TrackType##Lines(tracksData.m_orderedNamedGridLines); \
    style.setGridAutoRepeat##TrackType##s(tracksData.m_autoRepeatTrackSizes); \
    style.setGridAutoRepeat##TrackType##sInsertionPoint(tracksData.m_autoRepeatInsertionPoint); \
    style.setAutoRepeatNamedGrid##TrackType##Lines(tracksData.m_autoRepeatNamedGridLines); \
    style.setAutoRepeatOrderedNamedGrid##TrackType##Lines(tracksData.m_autoRepeatOrderedNamedGridLines); \
    style.setGridAutoRepeat##TrackType##sType(tracksData.m_autoRepeatType);

inline void StyleBuilderCustom::applyValueGridTemplateColumns(Style::BuilderState& builderState, CSSValue& value)
{
    StyleBuilderConverter::TracksData tracksData;
    if (!StyleBuilderConverter::createGridTrackList(value, tracksData, builderState))
        return;
    const NamedGridAreaMap& namedGridAreas = builderState.style().namedGridArea();
    if (!namedGridAreas.isEmpty())
        StyleBuilderConverter::createImplicitNamedGridLinesFromGridArea(namedGridAreas, tracksData.m_namedGridLines, ForColumns);

    SET_TRACKS_DATA(tracksData, builderState.style(), Column);
}

inline void StyleBuilderCustom::applyInitialGridTemplateRows(Style::BuilderState& builderState)
{
    builderState.style().setGridRows(RenderStyle::initialGridRows());
    builderState.style().setNamedGridRowLines(RenderStyle::initialNamedGridRowLines());
    builderState.style().setOrderedNamedGridRowLines(RenderStyle::initialOrderedNamedGridRowLines());
}

inline void StyleBuilderCustom::applyInheritGridTemplateRows(Style::BuilderState& builderState)
{
    builderState.style().setGridRows(builderState.parentStyle().gridRows());
    builderState.style().setNamedGridRowLines(builderState.parentStyle().namedGridRowLines());
    builderState.style().setOrderedNamedGridRowLines(builderState.parentStyle().orderedNamedGridRowLines());
}

inline void StyleBuilderCustom::applyValueGridTemplateRows(Style::BuilderState& builderState, CSSValue& value)
{
    StyleBuilderConverter::TracksData tracksData;
    if (!StyleBuilderConverter::createGridTrackList(value, tracksData, builderState))
        return;
    const NamedGridAreaMap& namedGridAreas = builderState.style().namedGridArea();
    if (!namedGridAreas.isEmpty())
        StyleBuilderConverter::createImplicitNamedGridLinesFromGridArea(namedGridAreas, tracksData.m_namedGridLines, ForRows);

    SET_TRACKS_DATA(tracksData, builderState.style(), Row);
}
#endif // ENABLE(CSS_GRID_LAYOUT)

inline void StyleBuilderCustom::applyValueWillChange(Style::BuilderState& builderState, CSSValue& value)
{
    if (value.isPrimitiveValue()) {
        ASSERT(toCSSPrimitiveValue(value).getValueID() == CSSValueAuto);
        builderState.style().setWillChange(nullptr);
        return;
    }

    RefPtr<WillChangeData> willChange = WillChangeData::create();
    auto processSingleValue = [&](const CSSValue& item) {
        if (!item.isPrimitiveValue())
            return;
        const auto& primitiveValue = toCSSPrimitiveValue(item);
        switch (primitiveValue.getValueID()) {
        case CSSValueScrollPosition:
            willChange->addFeature(WillChangeData::Feature::ScrollPosition);
            break;
        case CSSValueContents:
            willChange->addFeature(WillChangeData::Feature::Contents);
            break;
        default:
            if (primitiveValue.isPropertyID())
                willChange->addFeature(WillChangeData::Feature::Property, primitiveValue.getPropertyID());
            break;
        }
    };
    if (value.isValueList()) {
        for (auto& item : toCSSValueList(value))
            processSingleValue(item);
    } else
        processSingleValue(value);

    builderState.style().setWillChange(willChange.release());
}

inline void StyleBuilderCustom::applyValueStrokeWidth(Style::BuilderState& builderState, CSSValue& value)
{
    builderState.style().setStrokeWidth(StyleBuilderConverter::convertLength(builderState, value));
    builderState.style().setHasExplicitlySetStrokeWidth(true);
}

inline void StyleBuilderCustom::applyValueStrokeColor(Style::BuilderState& builderState, CSSValue& value)
{
    auto& primitiveValue = toCSSPrimitiveValue(value);
    if (builderState.applyPropertyToRegularStyle())
        builderState.style().setStrokeColor(builderState.colorFromPrimitiveValue(primitiveValue, /* forVisitedLink */ false));
    if (builderState.applyPropertyToVisitedLinkStyle())
        builderState.style().setVisitedLinkStrokeColor(builderState.colorFromPrimitiveValue(primitiveValue, /* forVisitedLink */ true));
    builderState.style().setHasExplicitlySetStrokeColor(true);
}

inline void StyleBuilderCustom::applyInitialCustomProperty(Style::BuilderState& builderState, const CSSRegisteredCustomProperty* registered, const AtomicString& name)
{
    if (registered && registered->initialValue()) {
        auto initialValue = registered->initialValueCopy();
        applyValueCustomProperty(builderState, registered, *initialValue);
        return;
    }

    auto invalid = CSSCustomPropertyValue::createUnresolved(name, CSSValueInvalid);
    applyValueCustomProperty(builderState, registered, invalid.get());
}

inline void StyleBuilderCustom::applyInheritCustomProperty(Style::BuilderState& builderState, const CSSRegisteredCustomProperty* registered, const AtomicString& name)
{
	CSSCustomPropertyValue* parentValue = builderState.parentStyle().inheritedCustomProperties().get(name);
    if (parentValue && !(registered && !registered->inherits))
        applyValueCustomProperty(builderState, registered, *parentValue);
    else
        applyInitialCustomProperty(builderState, registered, name);
}

inline void StyleBuilderCustom::applyValueCustomProperty(Style::BuilderState& builderState, const CSSRegisteredCustomProperty* registered, CSSCustomPropertyValue& value)
{
    ASSERT(value.isResolved());
    const auto& name = value.name();

    if (!registered || registered->inherits)
        builderState.style().setInheritedCustomPropertyValue(name, value);
    else
        builderState.style().setNonInheritedCustomPropertyValue(name, value);
}

inline void StyleBuilderCustom::applyInitialNavIndex(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyInheritNavIndex(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyValueNavIndex(Style::BuilderState& builderState, CSSValue& value)
{
	if (value.isPrimitiveValue()) {
		auto& primitiveValue = toCSSPrimitiveValue(value);
		builderState.style().setNavProperty(ENavProperty::Index, primitiveValue.isString() ? primitiveValue.stringValue() : String());
	}
}

inline void StyleBuilderCustom::applyInitialNavDown(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyInheritNavDown(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyValueNavDown(Style::BuilderState& builderState, CSSValue& value)
{
	if (value.isPrimitiveValue()) {
		auto& primitiveValue = toCSSPrimitiveValue(value);
		builderState.style().setNavProperty(ENavProperty::Down, primitiveValue.isString() ? primitiveValue.stringValue() : String());
	}
}

inline void StyleBuilderCustom::applyInitialNavLeft(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyInheritNavLeft(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyValueNavLeft(Style::BuilderState& builderState, CSSValue& value)
{
	if (value.isPrimitiveValue()) {
		auto& primitiveValue = toCSSPrimitiveValue(value);
		builderState.style().setNavProperty(ENavProperty::Left, primitiveValue.isString() ? primitiveValue.stringValue() : String());
	}
}

inline void StyleBuilderCustom::applyInitialNavRight(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyInheritNavRight(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyValueNavRight(Style::BuilderState& builderState, CSSValue& value)
{
	if (value.isPrimitiveValue()) {
		auto& primitiveValue = toCSSPrimitiveValue(value);
		builderState.style().setNavProperty(ENavProperty::Right, primitiveValue.isString() ? primitiveValue.stringValue() : String());
	}
}

inline void StyleBuilderCustom::applyInitialNavUp(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyInheritNavUp(Style::BuilderState& builderState)
{
}
inline void StyleBuilderCustom::applyValueNavUp(Style::BuilderState& builderState, CSSValue& value)
{
	if (value.isPrimitiveValue()) {
		auto& primitiveValue = toCSSPrimitiveValue(value);
		builderState.style().setNavProperty(ENavProperty::Up, primitiveValue.isString() ? primitiveValue.stringValue() : String());
	}
}

} // namespace WebCore
