/*
 * (C) 1999-2003 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2008, 2012 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#include "CSSParserMode.h"
#include "CSSProperty.h"
#include "CSSPropertyNames.h"
#include "CSSValueKeywords.h"
#include <wtf/ListHashSet.h>
#include <wtf/Ref.h>
#include <wtf/Vector.h>
#include <wtf/text/StringBuilder.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class CSSRule;
class CSSStyleDeclaration;
class ComputedStyleExtractor;
class ImmutableStylePropertySet;
class KURL;
class MutableStylePropertySet;
class PropertySetCSSStyleDeclaration;
class StyledElement;
class StylePropertyShorthand;
class StyleSheetContents;

class StylePropertySet : public WTF::RefCountedBase {
    WTF_MAKE_NONCOPYABLE(StylePropertySet); WTF_MAKE_FAST_ALLOCATED;
    friend class PropertyReference;
public:
    // Override RefCounted's deref() to ensure operator delete is called on
    // the appropriate subclass type.
    void deref() const;

    static PassRefPtr<ImmutableStylePropertySet> createImmutable(const CSSProperty* properties, unsigned count, CSSParserMode);

    class PropertyReference {
    public:
        PropertyReference(const StylePropertyMetadata& metadata, const CSSValue* value)
            : m_metadata(metadata)
            , m_value(value)
        { }

        CSSPropertyID id() const { return static_cast<CSSPropertyID>(m_metadata.m_propertyID); }
        CSSPropertyID shorthandID() const { return m_metadata.shorthandID(); }

        bool isImportant() const { return m_metadata.m_important; }
        bool isInherited() const { return m_metadata.m_inherited; }
        bool isImplicit() const { return m_metadata.m_implicit; }

        String cssName() const;
        String cssText() const;

        ALWAYS_INLINE const CSSValue* value() const { return m_value; }
        // FIXME: We should try to remove this mutable overload.
        CSSValue* value() { return const_cast<CSSValue*>(m_value); }

        // FIXME: Remove this.
        CSSProperty toCSSProperty() const { return CSSProperty(id(), const_cast<CSSValue*>(m_value), isImportant(), m_metadata.m_isSetFromShorthand, m_metadata.m_indexInShorthandsVector, isImplicit()); }

    private:
        const StylePropertyMetadata& m_metadata;
        const CSSValue* m_value;
    };

    unsigned propertyCount() const;
    bool isEmpty() const { return !propertyCount(); }
    PropertyReference propertyAt(unsigned) const;

    PassRefPtr<CSSValue> getPropertyCSSValue(CSSPropertyID) const;
    String getPropertyValue(CSSPropertyID) const;
    bool propertyIsImportant(CSSPropertyID) const;
    String getPropertyShorthand(CSSPropertyID) const;
    bool isPropertyImplicit(CSSPropertyID) const;

    PassRefPtr<CSSValue> getCustomPropertyCSSValue(const String& propertyName) const;
    String getCustomPropertyValue(const String& propertyName) const;
    bool customPropertyIsImportant(const String& propertyName) const;

    PassRefPtr<MutableStylePropertySet> copyBlockProperties() const;

    ALWAYS_INLINE CSSParserMode cssParserMode() const { return static_cast<CSSParserMode>(m_cssParserMode); }

    void addSubresourceStyleURLs(ListHashSet<KURL>&) const;

    Ref<MutableStylePropertySet> mutableCopy() const;
    PassRefPtr<StylePropertySet> immutableCopyIfNeeded() const;

    PassRefPtr<MutableStylePropertySet> copyPropertiesInSet(const CSSPropertyID* set, unsigned length) const;
    
    String asText() const;
    AtomString asTextAtom() const;

    ALWAYS_INLINE bool isMutable() const { return m_isMutable; }
    bool hasCSSOMWrapper() const;

    template<class Functor> bool traverseSubresources(const Functor& handler) const;

    static unsigned averageSizeInBytes();

#ifndef NDEBUG
    void showStyle();
#endif

    bool propertyMatches(CSSPropertyID, const CSSValue*) const;

protected:
    StylePropertySet(CSSParserMode cssParserMode)
        : m_cssParserMode(cssParserMode)
        , m_isMutable(true)
        , m_arraySize(0)
    { }

    StylePropertySet(CSSParserMode cssParserMode, unsigned immutableArraySize)
        : m_cssParserMode(cssParserMode)
        , m_isMutable(false)
        , m_arraySize(immutableArraySize)
    { }

    int findPropertyIndex(CSSPropertyID) const;
    int findCustomPropertyIndex(const String& propertyName) const;

    unsigned m_cssParserMode : 3;
    mutable unsigned m_isMutable : 1;
    unsigned m_arraySize : 28;
    
private:
    String getGridShorthandValue(const StylePropertyShorthand&) const;
    String getGridRowColumnShorthandValue(const StylePropertyShorthand&) const;
    String getGridAreaShorthandValue() const;
    String getGridTemplateValue() const;
    String getGridValue() const;
    String getShorthandValue(const StylePropertyShorthand&, const char* separator = " ") const;
	String getCommonValue(const StylePropertyShorthand&) const;
    String getAlignmentShorthandValue(const StylePropertyShorthand&) const;
    String borderImagePropertyValue() const;
    String borderPropertyValue(const StylePropertyShorthand&, const StylePropertyShorthand&, const StylePropertyShorthand&) const;
    String pageBreakPropertyValue(const StylePropertyShorthand&) const;
	String getLayeredShorthandValue(const StylePropertyShorthand&) const;
    String get2Values(const StylePropertyShorthand&) const;
    String get4Values(const StylePropertyShorthand&) const;
    String borderSpacingValue(const StylePropertyShorthand&) const;
    String fontValue() const;
    void appendFontLonghandValueIfExplicit(CSSPropertyID, StringBuilder& result, String& value) const;
    bool shorthandHasVariableReference(CSSPropertyID, String&) const;
    StringBuilder asTextInternal() const;

    friend class PropertySetCSSStyleDeclaration;
};

class ImmutableStylePropertySet : public StylePropertySet {
public:
    ImmutableStylePropertySet(const CSSProperty*, unsigned count, CSSParserMode);
    ~ImmutableStylePropertySet();

    unsigned propertyCount() const { return m_arraySize; }
    bool isEmpty() const { return !propertyCount(); }
    PropertyReference propertyAt(unsigned index) const;

    const CSSValue** valueArray() const;
    const StylePropertyMetadata* metadataArray() const;
    int findPropertyIndex(CSSPropertyID) const;
    int findCustomPropertyIndex(const String& propertyName) const;

    void* m_storage;
};

inline const CSSValue** ImmutableStylePropertySet::valueArray() const
{
    return reinterpret_cast<const CSSValue**>(const_cast<const void**>((&(this->m_storage))));
}

inline const StylePropertyMetadata* ImmutableStylePropertySet::metadataArray() const
{
    return reinterpret_cast<const StylePropertyMetadata*>(&reinterpret_cast<const char*>(&(this->m_storage))[m_arraySize * sizeof(CSSValue*)]);
}

class MutableStylePropertySet : public StylePropertySet {
public:
    static Ref<MutableStylePropertySet> create(CSSParserMode = HTMLQuirksMode);
    static Ref<MutableStylePropertySet> create(Vector<CSSProperty>&&);

    ~MutableStylePropertySet();

    unsigned propertyCount() const { return m_propertyVector.size(); }
    bool isEmpty() const { return !propertyCount(); }
    PropertyReference propertyAt(unsigned index) const;

    PropertySetCSSStyleDeclaration* cssStyleDeclaration();

    // These expand shorthand properties into multiple properties.
    bool setProperty(CSSPropertyID, const String& value, bool important, CSSParserContext, StyleSheetContents* = nullptr);
	bool setProperty(CSSPropertyID, const String& value, bool important = false);
	void setProperty(CSSPropertyID, PassRefPtr<CSSValue>, bool important = false);

    // These do not. FIXME: This is too messy, we can do better.
    bool setProperty(CSSPropertyID, CSSValueID identifier, bool important = false);
    bool setProperty(CSSPropertyID, CSSPropertyID identifier, bool important = false);
    bool setProperty(const CSSProperty&, CSSProperty* slot = 0);

    bool removeProperty(CSSPropertyID, String* returnText = 0);
    void removeBlockProperties();
    bool removePropertiesInSet(const CSSPropertyID* set, unsigned length);

    // FIXME: These two can be moved to EditingStyle.cpp
    void removeEquivalentProperties(const StylePropertySet*);
    void removeEquivalentProperties(ComputedStyleExtractor*);

    void mergeAndOverrideOnConflict(const StylePropertySet&);

    bool addParsedProperties(const ParsedPropertyVector&);
    bool addParsedProperty(const CSSProperty&);

    void clear();
    bool parseDeclaration(const String& styleDeclaration, CSSParserContext);

    CSSStyleDeclaration& ensureCSSStyleDeclaration();
    CSSStyleDeclaration& ensureInlineCSSStyleDeclaration(StyledElement& parentElement);

    int findPropertyIndex(CSSPropertyID) const;
    int findCustomPropertyIndex(const String& propertyName) const;

    Vector<CSSProperty, 4> m_propertyVector;

    // Methods for querying and altering CSS custom properties.
    bool setCustomProperty(const Document*, const String& propertyName, const String& value, bool important, CSSParserContext);
	bool removeCustomProperty(const String& propertyName, String* returnText = 0);

private:
	explicit MutableStylePropertySet(CSSParserMode cssParserMode);
	explicit MutableStylePropertySet(const StylePropertySet&);
    MutableStylePropertySet(Vector<CSSProperty>&&);

    bool removeShorthandProperty(CSSPropertyID);
    CSSProperty* findCSSPropertyWithID(CSSPropertyID);
    CSSProperty* findCustomCSSPropertyWithName(const String&);
    OwnPtr<PropertySetCSSStyleDeclaration> m_cssomWrapper;

    friend class StylePropertySet;
};

// TYPE_CASTS_BASE(MutableStyleProperties, StyleProperties, set, set->isMutable(), set.isMutable());

inline const MutableStylePropertySet* toMutableStyleProperties(const StylePropertySet* set)
{
	return static_cast<const MutableStylePropertySet*>(set);
}

inline const MutableStylePropertySet& toMutableStyleProperties(const StylePropertySet& set)
{
	return static_cast<const MutableStylePropertySet&>(set);
}

// TYPE_CASTS_BASE(ImmutableStyleProperties, StyleProperties, set, !set->isMutable(), !set.isMutable());

inline const ImmutableStylePropertySet* toImmutableStyleProperties(const StylePropertySet* set)
{
	return static_cast<const ImmutableStylePropertySet*>(set);
}

inline const ImmutableStylePropertySet& toImmutableStyleProperties(const StylePropertySet& set)
{
	return static_cast<const ImmutableStylePropertySet&>(set);
}

inline ImmutableStylePropertySet::PropertyReference ImmutableStylePropertySet::propertyAt(unsigned index) const
{
    return PropertyReference(metadataArray()[index], valueArray()[index]);
}

inline MutableStylePropertySet::PropertyReference MutableStylePropertySet::propertyAt(unsigned index) const
{
    const CSSProperty& property = m_propertyVector[index];
    return PropertyReference(property.metadata(), property.value());
}

inline StylePropertySet::PropertyReference StylePropertySet::propertyAt(unsigned index) const
{
    if (m_isMutable)
        return toMutableStyleProperties(*this).propertyAt(index);
    return toImmutableStyleProperties(*this).propertyAt(index);
}

inline unsigned StylePropertySet::propertyCount() const
{
    if (m_isMutable)
        return toMutableStyleProperties(*this).propertyCount();
    return toImmutableStyleProperties(*this).propertyCount();
}

inline void StylePropertySet::deref() const
{
    if (!derefBase())
        return;

    if (m_isMutable)
        delete static_cast<const MutableStylePropertySet*>(this);
    else
        delete static_cast<const ImmutableStylePropertySet*>(this);
}

inline int StylePropertySet::findPropertyIndex(CSSPropertyID propertyID) const
{
    if (m_isMutable)
        return toMutableStyleProperties(*this).findPropertyIndex(propertyID);
    return toImmutableStyleProperties(*this).findPropertyIndex(propertyID);
}

inline int StylePropertySet::findCustomPropertyIndex(const String& propertyName) const
{
    if (m_isMutable)
        return toMutableStyleProperties(*this).findCustomPropertyIndex(propertyName);
    return toImmutableStyleProperties(*this).findCustomPropertyIndex(propertyName);
}

template<class Functor>
inline bool StylePropertySet::traverseSubresources(const Functor& handler) const
{
    unsigned size = propertyCount();
    for (unsigned i = 0; i < size; ++i) {
    	CSSValue* value = propertyAt(i).value();
        if(value && value->traverseSubresources(handler))
            return true;
    }
    return false;
}

typedef StylePropertySet StyleProperties;
typedef MutableStylePropertySet MutableStyleProperties;

} // namespace WebCore
