/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 2004-2005 Allan Sandfeld Jensen (kde@carewolf.com)
 * Copyright (C) 2006, 2007 Nicholas Shanks (webkit@nickshanks.com)
 * Copyright (C) 2005-2019 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Alexey Proskuryakov <ap@webkit.org>
 * Copyright (C) 2007, 2008 Eric Seidel <eric@webkit.org>
 * Copyright (C) 2008, 2009 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)
 * Copyright (c) 2011, Code Aurora Forum. All rights reserved.
 * Copyright (C) Research In Motion Limited 2011. All rights reserved.
 * Copyright (C) 2012, 2013 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "StyleResolver.h"

#include "Attribute.h"
#include "StyleScope.h"
#include "CSSBorderImage.h"
#include "CSSDefaultStyleSheets.h"
#include "CSSFontFaceRule.h"
#include "CSSFontValue.h"
#include "CSSFontSelector.h"
#include "CSSGridAutoRepeatValue.h"
#include "CSSGridLineNamesValue.h"
#include "CSSGridTemplateAreasValue.h"
#include "CSSImportRule.h"
#include "CSSLineBoxContainValue.h"
#include "CSSMediaRule.h"
#include "CSSNamespaceRule.h"
#include "CSSPageRule.h"
#include "CSSParser.h"
#include "CSSPrimitiveValueMappings.h"
#include "CSSPropertyNames.h"
#include "CSSRuleList.h"
#include "CSSSelector.h"
#include "CSSSelectorList.h"
#include "CSSStyleRule.h"
#include "CSSStyleSheet.h"
#include "CSSSupportsRule.h"
#include "CSSTimingFunctionValue.h"
#include "CachedImage.h"
#include "CachedResourceLoader.h"
#if ENABLE(CSS_SHAPES)
#include "CachedResourceLoader.h"
#endif
#include "CalculationValue.h"
#include "ContentData.h"
#include "Counter.h"
#include "CounterContent.h"
#include "CursorList.h"
#include "ElementRuleCollector.h"
#include "ExtensionStyleSheets.h"
#include "CSSFontFeatureValue.h"
#include "CSSFontValue.h"
#include "Frame.h"
#include "FrameLoader.h"
#include "FrameLoaderClient.h"
#include "FrameSelection.h"
#include "FrameView.h"
#include "HTMLDialogElement.h"
#include "HTMLDocument.h"
#include "HTMLIFrameElement.h"
#include "HTMLInputElement.h"
#include "HTMLNames.h"
#include "HTMLOptionElement.h"
#include "HTMLProgressElement.h"
#include "HTMLSlotElement.h"
#include "HTMLStyleElement.h"
#include "HTMLTextAreaElement.h"
#include "InspectorInstrumentation.h"
#include "KeyframeList.h"
#include "LinkHash.h"
#include "LocaleToScriptMapping.h"
#include "MathMLElement.h"
#include "MathMLNames.h"
#include "Matrix3DTransformOperation.h"
#include "MatrixTransformOperation.h"
#include "MediaList.h"
#include "MediaQueryEvaluator.h"
#include "NodeRenderStyle.h"
#include "Page.h"
#include "PageRuleCollector.h"
#include "Pair.h"
#include "PerspectiveTransformOperation.h"
#include "PseudoElement.h"
#include "QuotesData.h"
#include "Rect.h"
#include "RenderGrid.h"
#include "RenderScrollbar.h"
#include "RenderScrollbarTheme.h"
#include "RenderStyleConstants.h"
#include "RenderTheme.h"
#include "RenderView.h"
#include "RotateTransformOperation.h"
#include "RenderView.h"
#include "RuleSet.h"
#include "SVGDocumentExtensions.h"
#include "SVGFontFaceElement.h"
#include "ScaleTransformOperation.h"
#include "SecurityOrigin.h"
#include "Settings.h"
#include "ShadowData.h"
#include "ShadowRoot.h"
#include "StyleAdjuster.h"
#include "SkewTransformOperation.h"
#include "StyleBuilder.h"
#include "StyleFontSizeFunctions.h"
#include "StylePropertyShorthand.h"
//#include "StyleResolveForDocument.h"
#include "StyleRule.h"
#include "StyleRuleImport.h"
#include "StyleSheetContents.h"
#include "StyleSheetList.h"
#include "Text.h"
#include "TransformationMatrix.h"
#include "TranslateTransformOperation.h"
#include "UserAgentStyleSheets.h"
#include "VisitedLinkState.h"
#include "CSSKeyframeRule.h"
#include "CSSKeyframesRule.h"
#include "WebKitFontFamilyNames.h"
#include "WillChangeData.h"
#include "XMLNames.h"
#include <wtf/SetForScope.h>
#include <wtf/StdLibExtras.h>
#include <wtf/Vector.h>

#if ENABLE(DASHBOARD_SUPPORT)
#include "DashboardRegion.h"
#endif

#if ENABLE(SVG)
#include "CachedSVGDocument.h"
#include "CachedSVGDocumentReference.h"
#include "SVGDocument.h"
#include "SVGNames.h"
#include "SVGURIReference.h"
#endif

#if ENABLE(VIDEO_TRACK)
#include "WebVTTElement.h"
#endif

using namespace std;

namespace WebCore {

using namespace HTMLNames;

inline void StyleResolver::clear()
{
    m_element = nullptr;
    m_parentStyle = nullptr;
    m_ownedParentStyle = nullptr;
    m_userAgentAppearanceStyle = nullptr;
}

Ref<StyleResolver> StyleResolver::create(Document& document)
{
    return adoptRef(*new StyleResolver(document));
}

inline StyleResolver::StyleResolver(Document& document)
	: m_ruleSets(*this)
	, m_document(makeWeakPtr(document))
	, m_matchedDeclarationsCache(*this)
	, m_matchAuthorAndUserStyles(document.settings().authorAndUserStylesEnabled())
{
    CSSDefaultStyleSheets::initDefaultStyle();
    
    // construct document root element default style. this is needed
    // to evaluate media queries that contain relative constraints, like "screen and (max-width: 10em)"
    // This is here instead of constructor, because when constructor is run,
    // document doesn't have documentElement
    // NOTE: this assumes that element that gets passed to styleForElement -call
    // is always from the document that owns the style selector
    FrameView* view = document.view();
    if (view)
    	m_mediaQueryEvaluator = MediaQueryEvaluator(view->mediaType());
    else
    	m_mediaQueryEvaluator = MediaQueryEvaluator(String(ASCIILiteral("all")));

    if (auto* documentElement = document.documentElement()) {
        m_rootDefaultStyle = styleForElement(*documentElement, Style::ResolutionContext(document.initialContainingBlockStyle()), MatchOnlyUserAgentRules).renderStyle.release();
        // Turn off assertion against font lookups during style resolver initialization. We may need root style font for media queries.
		m_document->fontSelector().incrementIsComputingRootStyleFont();
		m_rootDefaultStyle->fontCascade().update(&m_document->fontSelector());
		m_rootDefaultStyle->fontCascade().primaryFont();
		m_document->fontSelector().decrementIsComputingRootStyleFont();
	}

    if (m_rootDefaultStyle && view)
    	m_mediaQueryEvaluator = MediaQueryEvaluator(view->mediaType(), document, m_rootDefaultStyle.get());

    m_ruleSets.resetAuthorStyle();
    m_ruleSets.resetUserAgentMediaQueryStyle();
}

void StyleResolver::addCurrentSVGFontFaceRules()
{
#if ENABLE(SVG_FONTS)
	RefPtr<Document> document = m_document.get();
    if (auto* svgExtensions = document->svgExtensionsIfExists()) {
        for (SVGFontFaceElement* svgFontFaceElement : svgExtensions->svgFontFaceElements())
        	document->fontSelector().addFontFaceRule(svgFontFaceElement->fontFaceRule(), svgFontFaceElement->isInUserAgentShadowTree());
    }
#endif
}

void StyleResolver::appendAuthorStyleSheets(const Vector<RefPtr<CSSStyleSheet>>& styleSheets)
{
	m_ruleSets.appendAuthorStyleSheets(styleSheets, m_mediaQueryEvaluator, m_inspectorCSSOMWrappers);

	RefPtr<Document> document = m_document.get();
    if (document->renderView() && document->renderView()->style())
        document->renderView()->style()->font().update(&document->fontSelector());
}

// This is a simplified style setting function for keyframe styles
void StyleResolver::addKeyframeStyle(PassRefPtr<StyleRuleKeyframes> rule)
{
    auto& animationName = rule->name();
    m_keyframesRuleMap.set(animationName, rule);
    m_document->keyframesRuleDidChange(animationName);
}

StyleResolver::~StyleResolver()
{
    RELEASE_ASSERT(!m_document.isResolvingTreeStyle());
    RELEASE_ASSERT(!m_isDeleted);
    m_isDeleted = true;
}

inline void StyleResolver::initElement()
{
	m_element = nullptr;
}

inline void StyleResolver::initElement(const Element& element)
{
    if (m_element != &element) {
        m_element = &element;
    }
}

inline void StyleResolver::initForStyleResolve(const Element* element, const RenderStyle* parentStyle, const RenderStyle* documentElementStyle)
{
    if (element) {
		bool resetStyleInheritance = hasShadowRootParent(element) && toShadowRoot(element->parentNode())->resetStyleInheritance();
		m_parentStyle = resetStyleInheritance ? nullptr : parentStyle;
    } else {
        m_parentStyle = parentStyle;
    }

    Element* documentElement = element ? element->document().documentElement() : nullptr;
    if (!documentElement || documentElement == element)
        m_rootElementStyle = document()->initialContainingBlockStyle();
    else
        m_rootElementStyle = documentElementStyle ? documentElementStyle : documentElement->renderStyle();

    m_style = nullptr;
}

void StyleResolver::setStyle(PassOwnPtr<RenderStyle> s)
{
	m_style = s;
}

void StyleResolver::setParentStyle(PassOwnPtr<RenderStyle> parentStyle)
{
    m_ownedParentStyle = parentStyle;
    m_parentStyle = m_ownedParentStyle.get();
}

static inline bool isAtShadowBoundary(const Element& element)
{
    auto* parentNode = element.parentNode();
    return parentNode && parentNode->isShadowRoot();
}

ElementStyle StyleResolver::styleForElement(const Element& element, const Style::ResolutionContext& context, RuleMatchingBehavior matchingBehavior)
{
    RELEASE_ASSERT(!m_isDeleted);

    initElement(element);
    initForStyleResolve(&element, context.parentStyle, context.documentElementStyle);

    if (m_parentStyle) {
        setStyle(RenderStyle::create());
        m_style->inheritFrom(*m_parentStyle);
    }
    else {
        setStyle(defaultStyleForElement());
        setParentStyle(RenderStyle::clone(*style()));
    }

    auto& style = *this->style();

    if (element.isLink()) {
        style.setIsLink(true);
        EInsideLink linkState = document()->visitedLinkState()->determineLinkState(element);
        style.setInsideLink(linkState);
    }

    CSSDefaultStyleSheets::ensureDefaultStyleSheetsForElement(element);

    OwnPtr<ElementRuleCollector> collector = adoptPtr(new ElementRuleCollector(element, m_ruleSets, context.selectorMatchingState));
    collector->setMedium(&m_mediaQueryEvaluator);

    MatchResult matchResult;
    if (matchingBehavior == MatchOnlyUserAgentRules)
    	collector->matchUARules();
    else
    	collector->matchAllRules(m_matchAuthorAndUserStyles, matchingBehavior != MatchAllRulesExcludingSMIL);

    if (collector->matchedPseudoElementIds())
        style.setHasPseudoStyles(collector->matchedPseudoElementIds());

    // This is required for style sharing.
    if (collector->didMatchUncommonAttributeSelector())
        style.setUnique();

    OwnPtr<Style::Relations> elementStyleRelations = Style::commitRelationsToRenderStyle(style, element, collector->styleRelations());

    applyMatchedProperties(collector->matchResult(), element);

    // Clean up our style object's display and text decorations (among other fixups).
    Style::Adjuster adjuster(*document(), *parentStyle(), context.parentBoxStyle, &element);
	adjuster.adjust(*this->style(), userAgentAppearanceStyle());

    if (style.usesViewportUnits())
        document()->setHasStyleWithViewportUnits();

    clear(); // Clear out for the next resolve.

    return ElementStyle(takeStyle(), elementStyleRelations.release());
}

PassOwnPtr<RenderStyle> StyleResolver::styleForKeyframe(const Element& element, const RenderStyle* elementStyle, const Style::ResolutionContext& context, const StyleRuleKeyframe* keyframe, KeyframeValue& keyframeValue)
{
    RELEASE_ASSERT(!m_isDeleted);

    OwnPtr<MatchResult> result = adoptPtr(new MatchResult());
    result->authorDeclarations.append(MatchedProperties(keyframe->properties()));

    ASSERT(!m_style);

    initElement(element);
    initForStyleResolve(&element, nullptr, context.documentElementStyle);

    // Create the style
    setStyle(RenderStyle::clone(*elementStyle));
    setParentStyle(RenderStyle::clone(context.parentStyle ? *context.parentStyle : *elementStyle));

    Style::Builder builder(*this, *result, OptionSet<Style::CascadeLevel>(Style::CascadeLevel::Author));
    builder.applyAllProperties();

    Style::Adjuster adjuster(*document(), *parentStyle(), nullptr, nullptr);
    adjuster.adjust(*this->style(), userAgentAppearanceStyle());

    // Add all the animating properties to the keyframe.
    unsigned propertyCount = keyframe->properties().propertyCount();
	for (unsigned i = 0; i < propertyCount; ++i) {
		CSSPropertyID property = keyframe->properties().propertyAt(i).id();
		// Timing-function within keyframes is special, because it is not animated; it just
		// describes the timing function between this keyframe and the next.
		if (property != CSSPropertyAnimationTimingFunction)
			keyframeValue.addProperty(property);
	}

    return m_style.release();
}

bool StyleResolver::isAnimationNameValid(const String& name)
{
    return m_keyframesRuleMap.find(AtomicString(name)) != m_keyframesRuleMap.end();
}

const Vector<RefPtr<StyleRuleKeyframe>> StyleResolver::keyframeRulesForName(const AtomString& animationName)
{
    if (animationName.isEmpty())
        return Vector<RefPtr<StyleRuleKeyframe>>();

    m_keyframesRuleMap.checkConsistency();

    auto it = m_keyframesRuleMap.find(animationName.impl());
    if (it == m_keyframesRuleMap.end())
        return Vector<RefPtr<StyleRuleKeyframe>>();

    auto timingFunctionForKeyframe = [](RefPtr<StyleRuleKeyframe> keyframe) -> RefPtr<const TimingFunction> {
        if (auto timingFunctionCSSValue = keyframe->properties().getPropertyCSSValue(CSSPropertyAnimationTimingFunction)) {
            if (auto timingFunction = TimingFunction::createFromCSSValue(*timingFunctionCSSValue))
                return timingFunction;
        }
        return CubicBezierTimingFunction::defaultTimingFunction();
    };

    HashSet<RefPtr<const TimingFunction>> timingFunctions;
    auto uniqueTimingFunctionForKeyframe = [&](RefPtr<StyleRuleKeyframe> keyframe) -> RefPtr<const TimingFunction> {
        auto timingFunction = timingFunctionForKeyframe(keyframe);
        for (auto existingTimingFunction : timingFunctions) {
            if (arePointingToEqualData(timingFunction, existingTimingFunction))
                return existingTimingFunction;
        }
        timingFunctions.add(timingFunction);
        return timingFunction;
    };

    auto* keyframesRule = it->second.get();
    auto* keyframes = &keyframesRule->keyframes();

    typedef std::pair<double, RefPtr<const TimingFunction>> KeyframeUniqueKey;
    auto hasDuplicateKeys = [&]() -> bool {
        HashSet<KeyframeUniqueKey> uniqueKeyframeKeys;
        for (auto& keyframe : *keyframes) {
            auto timingFunction = uniqueTimingFunctionForKeyframe(keyframe);
            for (auto key : keyframe->keys()) {
                if (!uniqueKeyframeKeys.add(std::make_pair(key, timingFunction)))
                    return true;
            }
        }
        return false;
    }();

    if (!hasDuplicateKeys)
        return *keyframes;

    // FIXME: If HashMaps could have Ref<> as value types, we wouldn't need
    // to copy the HashMap into a Vector.
    Vector<RefPtr<StyleRuleKeyframe>> deduplicatedKeyframes;
	// Merge keyframes with a similar offset and timing function.
	HashMap<KeyframeUniqueKey, RefPtr<StyleRuleKeyframe>> keyframesMap;
	for (auto& originalKeyframe : *keyframes) {
		auto timingFunction = uniqueTimingFunctionForKeyframe(originalKeyframe);
		for (auto key : originalKeyframe->keys()) {
			if (RefPtr<StyleRuleKeyframe> keyframe = keyframesMap.get(std::make_pair(key, timingFunction)))
				keyframe->mutableProperties()->mergeAndOverrideOnConflict(originalKeyframe->properties());
			else {
				Ref<StyleRuleKeyframe> styleRuleKeyframe = StyleRuleKeyframe::create(MutableStylePropertySet::create());
				styleRuleKeyframe->setKey(key);
				styleRuleKeyframe->mutableProperties()->mergeAndOverrideOnConflict(originalKeyframe->properties());
				keyframesMap.set(std::make_pair(key, timingFunction), styleRuleKeyframe.ptr());
				deduplicatedKeyframes.append(styleRuleKeyframe);
			}
		}
	}

	return deduplicatedKeyframes;
}

void StyleResolver::keyframeStylesForAnimation(const Element& element, const RenderStyle* elementStyle, const Style::ResolutionContext& context, KeyframeList& list)
{
	list.clear();

	auto keyframeRules = keyframeRulesForName(list.animationName());
	if (keyframeRules.isEmpty())
		return;

    // Construct and populate the style for each keyframe
    for (auto& keyframeRule : keyframeRules) {
        // Add this keyframe style to all the indicated key times
        for (auto key : keyframeRule->keys()) {
        	KeyframeValue keyframeValue(0, nullptr);
        	keyframeValue.setStyle(styleForKeyframe(element, elementStyle, context, keyframeRule.get(), keyframeValue));
            keyframeValue.setKey(key);
            if (RefPtr<CSSValue> timingFunctionCSSValue = keyframeRule->properties().getPropertyCSSValue(CSSPropertyAnimationTimingFunction))
            	keyframeValue.setTimingFunction(TimingFunction::createFromCSSValue(*timingFunctionCSSValue));
			list.insert(WTFMove(keyframeValue));
        }
    }
}

PassOwnPtr<RenderStyle> StyleResolver::pseudoStyleForElement(const Element& element, const PseudoStyleRequest& pseudoStyleRequest, const Style::ResolutionContext& context)
{
    initElement(element);
    initForStyleResolve(&element, context.parentStyle, context.documentElementStyle);

    if(m_parentStyle) {
    	setStyle(RenderStyle::create());
    	m_style->inheritFrom(*m_parentStyle);
    }
    else {
        setStyle(defaultStyleForElement());
        setParentStyle(RenderStyle::clone(*style()));
    }

    // Since we don't use pseudo-elements in any of our quirk/print user agent rules, don't waste time walking
    // those rules.

    // Check UA, user and author rules.
    OwnPtr<ElementRuleCollector> collector = adoptPtr(new ElementRuleCollector(element, m_ruleSets, context.selectorMatchingState));
    collector->setPseudoStyleRequest(pseudoStyleRequest);
    collector->setMedium(&m_mediaQueryEvaluator);
    collector->matchUARules();

    if (m_matchAuthorAndUserStyles) {
        collector->matchUserRules();
        collector->matchAuthorRules();
    }

    ASSERT(!collector->matchedPseudoElementIds());

    if (collector->matchResult().isEmpty())
        return nullptr;

    m_style->setStyleType(pseudoStyleRequest.pseudoId);

    applyMatchedProperties(collector->matchResult(), element);

    // Clean up our style object's display and text decorations (among other fixups).
    Style::Adjuster adjuster(*document(), *this->parentStyle(), context.parentBoxStyle, nullptr);
    adjuster.adjust(*this->style(), userAgentAppearanceStyle());

    if (style()->usesViewportUnits())
        document()->setHasStyleWithViewportUnits();

    // Now return the style.
    return m_style.release();
}

PassOwnPtr<RenderStyle> StyleResolver::styleForPage(int pageIndex)
{
    RELEASE_ASSERT(!m_isDeleted);

    auto* documentElement = m_document->documentElement();
    if (!documentElement || !documentElement->renderStyle())
        return RenderStyle::createPtr();

    initForStyleResolve(documentElement, document()->initialContainingBlockStyle(), nullptr);

    setStyle(RenderStyle::create());
    m_style->inheritFrom(*m_rootElementStyle);

    OwnPtr<PageRuleCollector> collector = adoptPtr(new PageRuleCollector(m_ruleSets, documentElement->renderStyle()->direction()));
    collector->matchAllPageRules(pageIndex);

    auto& result = collector->matchResult();

    Style::Builder builder(*this, result, OptionSet<Style::CascadeLevel>(Style::CascadeLevel::Author));
    builder.applyAllProperties();

    // Now return the style.
    return m_style.release();
}

PassOwnPtr<RenderStyle> StyleResolver::defaultStyleForElement()
{
    setStyle(RenderStyle::create());
    // Make sure our fonts are initialized if we don't inherit them from our parent style.
    const Settings& settings = documentSettings();
	initializeFontStyle(settings);
	m_style->font().update(&m_document->fontSelector());
    return m_style.release();
}

Vector<RefPtr<const StyleRule>> StyleResolver::styleRulesForElement(const Element* element, unsigned rulesToInclude)
{
    return pseudoStyleRulesForElement(element, PseudoId::None, rulesToInclude);
}

Vector<RefPtr<const StyleRule>> StyleResolver::pseudoStyleRulesForElement(const Element* element, PseudoId pseudoId, unsigned rulesToInclude)
{
    if (!element)
        return Vector<RefPtr<const StyleRule>>();

    initElement(*element);
    initForStyleResolve(element, nullptr, nullptr);

    OwnPtr<ElementRuleCollector> collector = adoptPtr(new ElementRuleCollector(*element, m_ruleSets, nullptr));
    collector->setMode(SelectorChecker::Mode::CollectingRules);
    collector->setPseudoStyleRequest(PseudoStyleRequest(pseudoId));
    collector->setMedium(&m_mediaQueryEvaluator);
    collector->setIncludeEmptyRules(rulesToInclude & EmptyCSSRules);

    if (rulesToInclude & UAAndUserCSSRules) {
        // First we match rules from the user agent sheet.
        collector->matchUARules();

        // Now we check user sheet rules.
        if (m_matchAuthorAndUserStyles)
            collector->matchUserRules();
    }

    if (m_matchAuthorAndUserStyles && (rulesToInclude & AuthorCSSRules))
        collector->matchAuthorRules();

    return collector->matchedRuleList();
}

// -------------------------------------------------------------------------------------
// this is mostly boring stuff on how to apply a certain rule to the renderstyle...

Length StyleResolver::convertToIntLength(const CSSPrimitiveValue* primitiveValue, const CSSToLengthConversionData& conversionData)
{
    return primitiveValue ? primitiveValue->convertToLength<FixedIntegerConversion | PercentConversion | CalculatedConversion>(conversionData) : Length(LengthType::Undefined);
}

static bool elementTypeHasAppearanceFromUAStyle(const Element& element)
{
    // NOTE: This is just a hard-coded list of elements that have some -webkit-appearance value in html.css
    const auto& localName = element.localName();
    return localName == HTMLNames::inputTag
        || localName == HTMLNames::textareaTag
        || localName == HTMLNames::buttonTag
        || localName == HTMLNames::progressTag
        || localName == HTMLNames::selectTag
        || localName == HTMLNames::meterTag
        || localName == HTMLNames::isindexTag;
}

void StyleResolver::invalidateMatchedDeclarationsCache()
{
    m_matchedDeclarationsCache.invalidate();
}

void StyleResolver::clearCachedDeclarationsAffectedByViewportUnits()
{
    m_matchedDeclarationsCache.clearEntriesAffectedByViewportUnits();
}

void StyleResolver::applyMatchedProperties(const MatchResult& matchResult, const Element& element)
{
	unsigned cacheHash = Style::MatchedDeclarationsCache::computeHash(matchResult);
	auto includedProperties = Style::PropertyCascade::IncludedProperties::All;

    auto& style = *this->style();
    auto& parentStyle = *this->parentStyle();

    auto* cacheEntry = m_matchedDeclarationsCache.find(cacheHash, matchResult);
    if (cacheEntry && Style::MatchedDeclarationsCache::isCacheable(element, style, parentStyle)) {
        // We can build up the style by copying non-inherited properties from an earlier style object built using the same exact
        // style declarations. We then only need to apply the inherited properties, if any, as their values can depend on the 
        // element context. This is fast and saves memory by reusing the style data structures.
        style.copyNonInheritedFrom(cacheEntry->renderStyle.get());
        if (parentStyle.inheritedEqual(*cacheEntry->parentRenderStyle) && !isAtShadowBoundary(element)) {
            auto linkStatus = style.insideLink();
            // If the cache item parent style has identical inherited properties to the current parent style then the
            // resulting style will be identical too. We copy the inherited properties over from the cache and are done.
            style.inheritFrom(*cacheEntry->renderStyle);

            // Unfortunately the link status is treated like an inherited property. We need to explicitly restore it.
            style.setInsideLink(linkStatus);

            if (cacheEntry->userAgentAppearanceStyle && elementTypeHasAppearanceFromUAStyle(element))
                setUserAgentAppearanceStyle(RenderStyle::clonePtr(*cacheEntry->userAgentAppearanceStyle));

            return;
        }
        includedProperties = Style::PropertyCascade::IncludedProperties::InheritedOnly;
    }

    if (elementTypeHasAppearanceFromUAStyle(element)) {
        // Find out if there's a -webkit-appearance property in effect from the UA sheet.
        // If so, we cache the border and background styles so that RenderTheme::adjustStyle()
        // can look at them later to figure out if this is a styled form control or not.
    	OwnPtr<RenderStyle> userAgentStyle = RenderStyle::clonePtr(style);
    	Style::Builder builder(*userAgentStyle, *this, matchResult, OptionSet<Style::CascadeLevel>(Style::CascadeLevel::UserAgent));
    	builder.applyAllProperties();

        this->setUserAgentAppearanceStyle(userAgentStyle.release());
    }

    Style::Builder builder(*this, matchResult, Style::allCascadeLevels(), includedProperties);

    // High priority properties may affect resolution of other properties (they are mostly font related).
    builder.applyHighPriorityProperties();

    if (cacheEntry && !cacheEntry->isUsableAfterHighPriorityProperties(style)) {
    	// High-priority properties may affect resolution of other properties. Kick out the existing cache entry and try again.
		m_matchedDeclarationsCache.remove(cacheHash);
		applyMatchedProperties(matchResult, element);
        return;
    }

    builder.applyLowPriorityProperties();

    for (auto& contentAttribute : builder.state().registeredContentAttributes())
        ruleSets().mutableFeatures().registerContentAttribute(contentAttribute);

    if (cacheEntry || !cacheHash)
        return;

    if (Style::MatchedDeclarationsCache::isCacheable(element, style, parentStyle))
        m_matchedDeclarationsCache.add(style, parentStyle, userAgentAppearanceStyle(), cacheHash, matchResult);
}

void InspectorCSSOMWrappers::collectFromStyleSheetIfNeeded(CSSStyleSheet* styleSheet)
{
    if (!m_styleRuleToCSSOMWrapperMap.isEmpty())
        collect(styleSheet);
}

template <class ListType>
void InspectorCSSOMWrappers::collect(ListType* listType)
{
    if (!listType)
        return;
    unsigned size = listType->length();
    for (unsigned i = 0; i < size; ++i) {
        CSSRule* cssRule = listType->item(i);
        switch (cssRule->type()) {
        case CSSRule::IMPORT_RULE:
        	collect(static_cast<CSSImportRule*>(cssRule)->styleSheet());
        	break;
		case CSSRule::MEDIA_RULE:
        	collect(static_cast<CSSMediaRule*>(cssRule));
        	break;
#if ENABLE(CSS3_CONDITIONAL_RULES)
        case CSSRule::SUPPORTS_RULE:
        	collect(static_cast<CSSSupportsRule*>(cssRule));
            break;
#endif
        case CSSRule::STYLE_RULE:
            m_styleRuleToCSSOMWrapperMap.add(static_cast<CSSStyleRule*>(cssRule)->styleRule(), static_cast<CSSStyleRule*>(cssRule));
            break;
		default:
			break;
		}
    }
}

void InspectorCSSOMWrappers::collectFromStyleSheetContents(HashSet<RefPtr<CSSStyleSheet> >& sheetWrapperSet, RefPtr<StyleSheetContents> styleSheet)
{
    if (!styleSheet)
        return;
    RefPtr<CSSStyleSheet> styleSheetWrapper = CSSStyleSheet::create(styleSheet);
    sheetWrapperSet.add(styleSheetWrapper);
    collect(styleSheetWrapper.get());
}

void InspectorCSSOMWrappers::collectFromStyleSheets(const Vector<RefPtr<CSSStyleSheet> >& sheets)
{
    for (auto& sheet : sheets)
        collect(sheet.get());
}

CSSStyleRule* InspectorCSSOMWrappers::getWrapperForRuleInSheets(const StyleRule* rule, Style::Scope& authorStyleSheets, ExtensionStyleSheets& extensionStyleSheets)
{
    if (m_styleRuleToCSSOMWrapperMap.isEmpty()) {
    	collectFromStyleSheetContents(m_styleSheetCSSOMWrapperSet, CSSDefaultStyleSheets::defaultStyleSheet);
    	collectFromStyleSheetContents(m_styleSheetCSSOMWrapperSet, CSSDefaultStyleSheets::quirksStyleSheet);
    	collectFromStyleSheetContents(m_styleSheetCSSOMWrapperSet, CSSDefaultStyleSheets::svgStyleSheet);
    	collectFromStyleSheetContents(m_styleSheetCSSOMWrapperSet, CSSDefaultStyleSheets::mathMLStyleSheet);
    	collectFromStyleSheetContents(m_styleSheetCSSOMWrapperSet, CSSDefaultStyleSheets::mediaControlsStyleSheet);
    	collectFromStyleSheetContents(m_styleSheetCSSOMWrapperSet, CSSDefaultStyleSheets::fullscreenStyleSheet);
        collectFromStyleSheetContents(m_styleSheetCSSOMWrapperSet, CSSDefaultStyleSheets::mediaQueryStyleSheet);

        collectFromStyleSheets(authorStyleSheets.activeStyleSheets());
        collect(extensionStyleSheets.pageUserSheet());
        collectFromStyleSheets(extensionStyleSheets.injectedUserStyleSheets());
        collectFromStyleSheets(extensionStyleSheets.documentUserStyleSheets());
    }
    return m_styleRuleToCSSOMWrapperMap.get(rule);
}

void StyleResolver::applyPropertyToStyle(CSSPropertyID id, CSSValue* value, PassOwnPtr<RenderStyle> style)
{
    initElement();
    initForStyleResolve(nullptr, nullptr, nullptr);
    setParentStyle(RenderStyle::clone(*style));
    setStyle(style);
    applyPropertyToCurrentStyle(id, value);
}

void StyleResolver::applyPropertyToCurrentStyle(CSSPropertyID id, CSSValue* value)
{
    if (!value)
        return;

    OwnPtr<MatchResult> matchResult = adoptPtr(new MatchResult());
    Style::Builder builder(*this, *matchResult, OptionSet<Style::CascadeLevel>());
    builder.applyPropertyValue(id, *value);
}

#if ENABLE(CSS_SCROLL_SNAP)

Length StyleResolver::parseSnapCoordinate(CSSPrimitiveValue& value)
{
    return value.convertToLength<FixedIntegerConversion | PercentConversion | AutoConversion>(m_state.cssToLengthConversionData());
}

Length StyleResolver::parseSnapCoordinate(CSSValueList& valueList, unsigned offset)
{
    return parseSnapCoordinate(downcast<CSSPrimitiveValue>(*valueList.item(offset)));
}

LengthSize StyleResolver::parseSnapCoordinatePair(CSSValueList& valueList, unsigned offset)
{
    return LengthSize(parseSnapCoordinate(valueList, offset), parseSnapCoordinate(valueList, offset + 1));
}

ScrollSnapPoints StyleResolver::parseSnapPoints(CSSValue& value)
{
    ScrollSnapPoints points;

    if (is<CSSPrimitiveValue>(value) && downcast<CSSPrimitiveValue>(value).getValueID() == CSSValueElements) {
        points.usesElements = true;
        return points;
    }

    points.hasRepeat = false;
    if (is<CSSValueList>(value)) {
        for (auto& currentValue : downcast<CSSValueList>(value)) {
            auto& itemValue = downcast<CSSPrimitiveValue>(currentValue.get());
            if (auto* lengthRepeat = itemValue.getLengthRepeatValue()) {
                if (auto* interval = lengthRepeat->interval()) {
                    points.repeatOffset = parseSnapCoordinate(*interval);
                    points.hasRepeat = true;
                    break;
                }
            }
            points.offsets.append(parseSnapCoordinate(itemValue));
        }
    }

    return points;
}

#endif

int StyleResolver::defaultFontSize(Document* document, int sizePerSettings)
{
    if(document->frame())
    	return document->frame()->loader()->client()->defaultFontSize(document->url(), sizePerSettings);
    return sizePerSettings;
}

void StyleResolver::initializeFontStyle(const Settings& settings)
{
    FontDescription fontDescription;
    fontDescription.setRenderingMode(settings.fontRenderingMode());
    auto& standardFontFamily = documentSettings().standardFontFamily();
    fontDescription.setOneFamily(standardFontFamily);
    fontDescription.setKeywordSizeFromIdentifier(CSSValueMedium);

    auto size = Style::fontSizeForKeyword(CSSValueMedium, false, *document());
    fontDescription.setSpecifiedSize(size);
    fontDescription.setComputedSize(Style::computedFontSizeFromSpecifiedSize(size, fontDescription.isAbsoluteSize(), this->element() && this->element()->isSVGElement(), this->style(), *document()));

    style()->setFontDescription(fontDescription);
}

bool StyleResolver::hasViewportDependentMediaQueries() const
{
    return m_ruleSets.hasViewportDependentMediaQueries();
}

Optional<DynamicMediaQueryEvaluationChanges> StyleResolver::evaluateDynamicMediaQueries()
{
    return m_ruleSets.evaluateDynamicMediaQueryRules(m_mediaQueryEvaluator);
}

#ifndef NDEBUG
static bool inLoadPendingResources = false;
#endif

} // namespace WebCore
