/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "StyleScope.h"
#include "CSSRule.h"
#include "CSSValueList.h"
#include "DocumentRuleSets.h"
#include "ElementRuleCollector.h"
#include "LinkHash.h"
#include "MatchedDeclarationsCache.h"
#include "MediaQueryEvaluator.h"
#include "RenderStyle.h"
#include "RuleFeature.h"
#include "RuntimeEnabledFeatures.h"
#include "RuleSet.h"
#include "ScrollTypes.h"
#include "SelectorChecker.h"
#include "StylePendingResources.h"
#include "StyleRelations.h"
#include <wtf/BitArray.h>
#include <wtf/Bitmap.h>
#include <wtf/HashMap.h>
#include <wtf/HashSet.h>
#include <wtf/Optional.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>
#include <wtf/text/AtomicStringHash.h>
#include <wtf/text/StringHash.h>

namespace WebCore {

class CSSCursorImageValue;
class CSSFontFace;
class CSSFontFaceRule;
class CSSFontValue;
class CSSImageGeneratorValue;
class CSSImageSetValue;
class CSSImageValue;
class CSSPageRule;
class CSSPrimitiveValue;
class CSSProperty;
class CSSSelector;
class CSSStyleRule;
class CSSStyleSheet;
class CSSValue;
class ContainerNode;
class CustomFilterOperation;
class CustomFilterParameter;
class Document;
class Element;
class ExtensionStyleSheets;
class Frame;
class FrameView;
class KURL;
class KeyframeList;
class KeyframeValue;
class MediaQueryEvaluator;
class Node;
class RuleData;
class RuleSet;
class SelectorFilter;
class Settings;
class StyleCachedImage;
class StyleGeneratedImage;
class StyleImage;
class StyleRuleKeyframe;
class StylePropertySet;
class StyleRule;
class StyleRuleKeyframes;
class StyleRulePage;
class StyleShader;
class StyleSheet;
class StyleSheetContents;
class StyleSheetList;
class StyledElement;
class SVGSVGElement;
class WebKitCSSShaderValue;

// MatchOnlyUserAgentRules is used in media queries, where relative units
// are interpreted according to the document root element style, and styled only
// from the User Agent Stylesheet rules.

enum RuleMatchingBehavior {
    MatchAllRules,
    MatchAllRulesExcludingSMIL,
    MatchOnlyUserAgentRules,
};

class InspectorCSSOMWrappers {
	WTF_MAKE_FAST_ALLOCATED;
public:
    // WARNING. This will construct CSSOM wrappers for all style rules and cache them in a map for significant memory cost.
    // It is here to support inspector. Don't use for any regular engine functions.
    CSSStyleRule* getWrapperForRuleInSheets(const StyleRule*, Style::Scope&, ExtensionStyleSheets&);
    void collectFromStyleSheetIfNeeded(CSSStyleSheet*);

private:
    template <class ListType>
    void collect(ListType*);

    void collectFromStyleSheetContents(HashSet<RefPtr<CSSStyleSheet>>& sheetWrapperSet, RefPtr<StyleSheetContents> styleSheet);
    void collectFromStyleSheets(const Vector<RefPtr<CSSStyleSheet>>&);

    HashMap<const StyleRule*, RefPtr<CSSStyleRule>> m_styleRuleToCSSOMWrapperMap;
    HashSet<RefPtr<CSSStyleSheet>> m_styleSheetCSSOMWrapperSet;
};

struct ElementStyle {
	WTF_MAKE_FAST_ALLOCATED;
public:
	ElementStyle()
	{ }
    ElementStyle(PassOwnPtr<RenderStyle> renderStyle, PassOwnPtr<Style::Relations> relations = nullptr)
        : renderStyle(renderStyle)
        , relations(relations)
    { }
    OwnPtr<RenderStyle> renderStyle;
    OwnPtr<Style::Relations> relations;
};

namespace Style {

struct SelectorMatchingState;

struct ResolutionContext {
	explicit ResolutionContext(const RenderStyle* parentStyle, const RenderStyle* parentBoxStyle = nullptr, const RenderStyle* documentElementStyle = nullptr, SelectorMatchingState* selectorMatchingState = nullptr)
		: parentStyle(parentStyle)
		, parentBoxStyle(parentBoxStyle)
		, documentElementStyle(documentElementStyle)
		, selectorMatchingState(selectorMatchingState)
	{
	}
    const RenderStyle* parentStyle;
    const RenderStyle* parentBoxStyle;
    // This needs to be provided during style resolution when up-to-date document element style is not available via DOM.
    const RenderStyle* documentElementStyle;
    SelectorMatchingState* selectorMatchingState;
};

}

// This class selects a RenderStyle for a given element based on a collection of stylesheets.
class StyleResolver : public RefCounted<StyleResolver>, public CanMakeWeakPtr<StyleResolver> {
public:
    static Ref<StyleResolver> create(Document&);
    ~StyleResolver();

    ElementStyle styleForElement(const Element&, const Style::ResolutionContext&, RuleMatchingBehavior = MatchAllRules);

    void keyframeStylesForAnimation(const Element&, const RenderStyle* elementStyle, const Style::ResolutionContext&, KeyframeList&);

    PassOwnPtr<RenderStyle> pseudoStyleForElement(const Element&, const PseudoStyleRequest&, const Style::ResolutionContext&);

    PassOwnPtr<RenderStyle> styleForPage(int pageIndex);
    PassOwnPtr<RenderStyle> defaultStyleForElement();

    void setStyle(PassOwnPtr<RenderStyle>);
    ALWAYS_INLINE RenderStyle* style() const { return m_style.get(); }
    PassOwnPtr<RenderStyle> takeStyle() { return m_style.release(); }

    void setParentStyle(PassOwnPtr<RenderStyle>);
    ALWAYS_INLINE const RenderStyle* parentStyle() const { return m_parentStyle; }
    ALWAYS_INLINE const RenderStyle* rootElementStyle() const { return m_rootElementStyle; }
    ALWAYS_INLINE const Element* element() const { return m_element; }
    ALWAYS_INLINE Document* document() { return m_document.get(); }
    const Document* document() const { return m_document.get(); }

    void appendAuthorStyleSheets(const Vector<RefPtr<CSSStyleSheet>>&);

    DocumentRuleSets& ruleSets() { return m_ruleSets; }
    const DocumentRuleSets& ruleSets() const { return m_ruleSets; }

    const RenderStyle* userAgentAppearanceStyle() const { return m_userAgentAppearanceStyle.get(); }
    void setUserAgentAppearanceStyle(PassOwnPtr<RenderStyle> style) { m_userAgentAppearanceStyle = style; }

    const MediaQueryEvaluator& mediaQueryEvaluator() const { return m_mediaQueryEvaluator; }

    void addCurrentSVGFontFaceRules();

    PassOwnPtr<RenderStyle> styleForKeyframe(const Element&, const RenderStyle* elementStyle, const Style::ResolutionContext&, const StyleRuleKeyframe*, KeyframeValue&);
    bool isAnimationNameValid(const String&);

    // These methods will give back the set of rules that matched for a given element (or a pseudo-element).
    enum CSSRuleFilter {
        UAAndUserCSSRules   = 1 << 1,
        AuthorCSSRules      = 1 << 2,
        EmptyCSSRules       = 1 << 3,
        AllButEmptyCSSRules = UAAndUserCSSRules | AuthorCSSRules,
        AllCSSRules         = AllButEmptyCSSRules | EmptyCSSRules,
    };
    Vector<RefPtr<const StyleRule>> styleRulesForElement(const Element*, unsigned rulesToInclude = AllButEmptyCSSRules);
    Vector<RefPtr<const StyleRule>> pseudoStyleRulesForElement(const Element*, PseudoId, unsigned rulesToInclude = AllButEmptyCSSRules);

    void applyPropertyToStyle(CSSPropertyID, CSSValue*, PassOwnPtr<RenderStyle>);
    void applyPropertyToCurrentStyle(CSSPropertyID, CSSValue*);

    void initializeFontStyle(const Settings&);

    bool hasSelectorForId(const AtomicString&) const;
    bool hasSelectorForAttribute(const Element&, const AtomicString&) const;

    bool hasViewportDependentMediaQueries() const;
    Optional<DynamicMediaQueryEvaluationChanges> evaluateDynamicMediaQueries();

    void addKeyframeStyle(PassRefPtr<StyleRuleKeyframes>);
    const Vector<RefPtr<StyleRuleKeyframe>> keyframeRulesForName(const AtomString&);

    ALWAYS_INLINE bool usesFirstLineRules() const { return m_ruleSets.features().usesFirstLineRules; }
    bool usesFirstLetterRules() const { return m_ruleSets.features().usesFirstLetterRules; }

    void invalidateMatchedDeclarationsCache();
    void clearCachedDeclarationsAffectedByViewportUnits();

    static Length convertToIntLength(const CSSPrimitiveValue*, const CSSToLengthConversionData&);

    InspectorCSSOMWrappers& inspectorCSSOMWrappers() { return m_inspectorCSSOMWrappers; }

    bool isSharedBetweenShadowTrees() const { return m_isSharedBetweenShadowTrees; }
    void setSharedBetweenShadowTrees() { m_isSharedBetweenShadowTrees = true; }

private:
    StyleResolver(Document&);

    void initForStyleResolve(const Element*, const RenderStyle* parentStyle, const RenderStyle* documentElementStyle);
    void initElement();
    void initElement(const Element&);

    void applyMatchedProperties(const MatchResult&, const Element&);

    ALWAYS_INLINE const Settings& documentSettings() { return m_document->settings(); }

    DocumentRuleSets m_ruleSets;

    OwnPtr<RenderStyle> m_userAgentAppearanceStyle;

    typedef HashMap<AtomicString, RefPtr<StyleRuleKeyframes> > KeyframesRuleMap;
    KeyframesRuleMap m_keyframesRuleMap;

    void clear();

#if ENABLE(BML)
    bool applyBMLProperty(CSSPropertyID id, CSSValue *value, StyledElement* styledElement, RenderStyle* style);
#endif

    static int defaultFontSize(Document*, int sizePerSettings);

    MediaQueryEvaluator m_mediaQueryEvaluator;
    OwnPtr<RenderStyle> m_rootDefaultStyle;

    PseudoId m_dynamicPseudo { PseudoId::None };

    WeakPtr<Document> m_document;

    OwnPtr<RenderStyle> m_style;
    const RenderStyle* m_parentStyle { nullptr };
    OwnPtr<RenderStyle> m_ownedParentStyle;
    const RenderStyle* m_rootElementStyle { nullptr };
    const Element* m_element { nullptr };

    InspectorCSSOMWrappers m_inspectorCSSOMWrappers;

    Style::MatchedDeclarationsCache m_matchedDeclarationsCache;

    bool m_matchAuthorAndUserStyles;

    bool m_isSharedBetweenShadowTrees { false };

    // See if we still have crashes where StyleResolver gets deleted early.
    bool m_isDeleted { false };
};

inline bool StyleResolver::hasSelectorForAttribute(const Element& element, const AtomicString &attributeName) const
{
    ASSERT(!attributeName.isEmpty());
    if (element.isHTMLElement())
        return m_ruleSets.features().attributeCanonicalLocalNamesInRules.contains(attributeName);
    return m_ruleSets.features().attributeLocalNamesInRules.contains(attributeName);
}

inline bool StyleResolver::hasSelectorForId(const AtomicString& idValue) const
{
    ASSERT(!idValue.isEmpty());
    return m_ruleSets.features().idsInRules.contains(idValue);
}

} // namespace WebCore
