/*
 * Copyright (C) 2003 Lars Knoll (knoll@kde.org)
 * Copyright (C) 2005 Allan Sandfeld Jensen (kde@carewolf.com)
 * Copyright (C) 2004-2021 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Nicholas Shanks <webkit@nickshanks.com>
 * Copyright (C) 2008 Eric Seidel <eric@webkit.org>
 * Copyright (C) 2009 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)
 * Copyright (C) 2012 Adobe Systems Incorporated. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "CSSParser.h"

#include "CSSCustomPropertyValue.h"
#include "CSSImageValue.h"
#include "CSSKeyframeRule.h"
#include "CSSParserFastPaths.h"
#include "CSSParserImpl.h"
#include "CSSPendingSubstitutionValue.h"
#include "CSSPropertyParser.h"
#include "CSSSelectorParser.h"
#include "CSSSupportsParser.h"
#include "CSSTokenizer.h"
#include "CSSVariableData.h"
#include "CSSVariableReferenceValue.h"
#include "Document.h"
#include "Element.h"
#include "Page.h"
#include "RenderStyle.h"
#include "RenderTheme.h"
#include "RuntimeEnabledFeatures.h"
#include "Settings.h"
#include "StyleBuilder.h"
#include "StyleColor.h"
#include "StyleResolver.h"
#include "StyleRule.h"
#include "StyleSheetContents.h"
#include <wtf/NeverDestroyed.h>
#include <wtf/text/StringBuilder.h>

using namespace WTF;

namespace WebCore {

const CSSParserContext& strictCSSParserContext()
{
	static NeverDestroyed<CSSParserContext> strictContext(HTMLStandardMode);
    return strictContext;
}

CSSParserContext::CSSParserContext(CSSParserMode mode, const KURL& baseURL)
    : baseURL(baseURL)
    , mode(mode)
{
    // FIXME: We should turn all of the features on from their WebCore Settings defaults.
    if (mode == UASheetMode) {
        individualTransformPropertiesEnabled = true;
        focusVisibleEnabled = true;
#if ENABLE(CSS_TRANSFORM_STYLE_OPTIMIZED_3D)
        transformStyleOptimized3DEnabled = true;
#endif
    }
}

CSSParserContext::CSSParserContext(const Document& document, const KURL& sheetBaseURL, const String& charset)
    : baseURL(sheetBaseURL.isNull() ? document.baseURL() : sheetBaseURL)
    , charset(charset)
    , mode(document.inQuirksMode() ? HTMLQuirksMode : HTMLStandardMode)
    , isHTMLDocument(document.isHTMLDocument())
	, hasDocumentSecurityOrigin(sheetBaseURL.isNull() || document.securityOrigin()->canRequest(baseURL))
	, useSystemAppearance(document.page() ? document.page()->useSystemAppearance() : false)
	, aspectRatioEnabled(document.settings().aspectRatioEnabled())
    , isCSSCustomFilterEnabled(document.settings().isCSSCustomFilterEnabled())
    , isCSSRegionsEnabled(document.cssRegionsEnabled())
	, isCSSCompositingEnabled(document.settings().isCSSCompositingEnabled())
    , needsSiteSpecificQuirks(document.settings().needsSiteSpecificQuirks())
    , enforcesCSSMIMETypeInNoQuirksMode(document.settings().enforceCSSMIMETypeInNoQuirksMode())
	, svgFontsEnabled(document.settings().svgFontsEnabled())
	, focusVisibleEnabled(document.settings().focusVisibleEnabled())
	, hasPseudoClassEnabled(document.settings().hasPseudoClassEnabled())
	, springTimingFunctionEnabled(document.settings().springTimingFunctionEnabled())
	, constantPropertiesEnabled(document.settings().constantPropertiesEnabled())
	, conicGradientsEnabled(document.settings().conicGradientsEnabled())
	, individualTransformPropertiesEnabled(document.settings().cssIndividualTransformPropertiesEnabled())
	, overflowClipEnabled(document.settings().overflowClipEnabled())
{
}

bool operator==(const CSSParserContext& a, const CSSParserContext& b)
{
    return a.baseURL == b.baseURL
        && a.charset == b.charset
        && a.mode == b.mode
        && a.enclosingRuleType == b.enclosingRuleType
        && a.isHTMLDocument == b.isHTMLDocument
        && a.hasDocumentSecurityOrigin == b.hasDocumentSecurityOrigin
        && a.isContentOpaque == b.isContentOpaque
        && a.useSystemAppearance == b.useSystemAppearance
        && a.aspectRatioEnabled == b.aspectRatioEnabled
        && a.isCSSCustomFilterEnabled == b.isCSSCustomFilterEnabled
        && a.isCSSRegionsEnabled == b.isCSSRegionsEnabled
        && a.isCSSCompositingEnabled == b.isCSSCompositingEnabled
        && a.isCSSGridLayoutEnabled == b.isCSSGridLayoutEnabled
        && a.needsSiteSpecificQuirks == b.needsSiteSpecificQuirks
        && a.enforcesCSSMIMETypeInNoQuirksMode == b.enforcesCSSMIMETypeInNoQuirksMode
		&& a.svgFontsEnabled == b.svgFontsEnabled
		&& a.useLegacyBackgroundSizeShorthandBehavior == b.useLegacyBackgroundSizeShorthandBehavior
		&& a.focusVisibleEnabled == b.focusVisibleEnabled
        && a.hasPseudoClassEnabled == b.hasPseudoClassEnabled
		&& a.scrollBehaviorEnabled == b.scrollBehaviorEnabled
		&& a.springTimingFunctionEnabled == b.springTimingFunctionEnabled
        && a.constantPropertiesEnabled == b.constantPropertiesEnabled
        && a.conicGradientsEnabled == b.conicGradientsEnabled
		&& a.individualTransformPropertiesEnabled == b.individualTransformPropertiesEnabled
        && a.overflowClipEnabled == b.overflowClipEnabled;
}

ResolvedURL CSSParserContext::completeURL(const String& string) const
{
    auto result = [&] () -> ResolvedURL {
        // See also Document::completeURL(const String&), but note that CSS always uses UTF-8 for URLs
        if (string.isNull())
            return ResolvedURL();

        if (CSSValue::isCSSLocalURL(string))
            return { string, KURL(KURL(), string) };

        return ResolvedURL { string, KURL(baseURL, string) };
    }();

    if (mode == WebVTTMode && !result.resolvedURL.protocolIsData())
        return ResolvedURL();

    return result;
}

CSSParser::CSSParser(const CSSParserContext& context)
    : m_context(context)
{
}

CSSParser::~CSSParser() = default;

void CSSParser::parseSheet(StyleSheetContents* sheet, const String& string)
{
    return CSSParserImpl::parseStyleSheet(string, m_context, sheet);
}

void CSSParser::parseSheetForInspector(const CSSParserContext& context, StyleSheetContents* sheet, const String& string, CSSParserObserver& observer)
{
    return CSSParserImpl::parseStyleSheetForInspector(string, context, sheet, observer);
}

PassRefPtr<StyleRuleBase> CSSParser::parseRule(const CSSParserContext& context, StyleSheetContents* sheet, const String& string)
{
    return CSSParserImpl::parseRule(string, context, sheet, CSSParserImpl::AllowImportRules);
}

PassRefPtr<StyleRuleKeyframe> CSSParser::parseKeyframeRule(const String& string)
{
    RefPtr<StyleRuleBase> keyframe = CSSParserImpl::parseRule(string, m_context, nullptr, CSSParserImpl::KeyframeRules);
    return static_cast<StyleRuleKeyframe*>(keyframe.get());
}

bool CSSParser::parseSupportsCondition(const String&  condition)
{
	CSSTokenizer tokenizer(condition);
	CSSParserTokenRange range = tokenizer.tokenRange();
	OwnPtr<CSSParserImpl> parser = adoptPtr(new CSSParserImpl(m_context));
	return CSSSupportsParser::supportsCondition(range, *parser, CSSSupportsParser::ForWindowCSS) == CSSSupportsParser::Supported;
}

Color CSSParser::parseColor(const String& string, bool strict)
{
    if (auto color = CSSParserFastPaths::parseSimpleColor(string, strict))
        return *color;
    // FIXME: Unclear why we want to ignore the boolean argument "strict" and always pass strictCSSParserContext here.
    auto value = parseSingleValue(CSSPropertyColor, string, strictCSSParserContext());
    if (!value || !value->isPrimitiveValue())
        return Color();
    const auto& primitiveValue = toCSSPrimitiveValue(*value);
    if (!primitiveValue.isRGBColor())
        return Color();
    return primitiveValue.color();
}

Color CSSParser::parseColor(const String& string, const CSSParserContext& context)
{
	return parseColor(string, isStrictParserMode(context.mode));
}

Color CSSParser::parseColorWithoutContext(const String& string)
{
	return parseColor(string, false);
}

Color CSSParser::parseColorWorkerSafe(StringView string)
{
    if (auto color = CSSParserFastPaths::parseSimpleColor(string))
        return *color;
    // FIXME: This function does not parse many types of valid CSS color syntax, such as color with non-sRGB color spaces.
    return Color();
}

Color CSSParser::parseSystemColor(StringView string, Document* document)
{
    auto keyword = cssValueKeywordID(string);
    if (!StyleColor::isSystemColor(keyword))
        return Color();
    if (document && document->page())
    	return document->page()->theme()->systemColor(keyword);
    return RenderTheme::defaultTheme()->systemColor(keyword);
}

Optional<SimpleColor> CSSParser::parseNamedColor(StringView string)
{
    return CSSParserFastPaths::parseNamedColor(string);
}

Optional<SimpleColor> CSSParser::parseHexColor(StringView string)
{
    return CSSParserFastPaths::parseHexColor(string);
}

PassRefPtr<CSSValue> CSSParser::parseSingleValue(CSSPropertyID propertyID, const String& string, const CSSParserContext& context)
{
    if (string.isEmpty())
        return nullptr;
    if (RefPtr<CSSValue> value = CSSParserFastPaths::maybeParseValue(propertyID, string, context))
        return value.release();
    CSSTokenizer tokenizer(string);
    return CSSPropertyParser::parseSingleValue(propertyID, tokenizer.tokenRange(), context);
}

CSSParser::ParseResult CSSParser::parseValue(MutableStylePropertySet* declaration, CSSPropertyID propertyID, const String& string, bool important, const CSSParserContext& context, StyleSheetContents* contextStyleSheet)
{
    ASSERT(!string.isEmpty());
    if (RefPtr<CSSValue> value = CSSParserFastPaths::maybeParseValue(propertyID, string, context))
        return declaration->addParsedProperty(CSSProperty(propertyID, value.release(), important)) ? CSSParser::ParseResult::Changed : CSSParser::ParseResult::Unchanged;
    CSSParser parser(context);
    return parser.parseValue(declaration, propertyID, string, important);
}

CSSParser::ParseResult CSSParser::parseCustomPropertyValue(MutableStylePropertySet* declaration, const AtomicString& propertyName, const String& string, bool important, const CSSParserContext& context)
{
    return CSSParserImpl::parseCustomPropertyValue(declaration, propertyName, string, important, context);
}

CSSParser::ParseResult CSSParser::parseValue(MutableStylePropertySet* declaration, CSSPropertyID propertyID, const String& string, bool important)
{
    return CSSParserImpl::parseValue(declaration, propertyID, string, important, m_context);
}

Optional<CSSSelectorList> CSSParser::parseSelector(const String& string)
{
    return CSSSelectorParser::parseSelector(CSSTokenizer(string).tokenRange(), m_context, nullptr);
}

PassRefPtr<ImmutableStylePropertySet> CSSParser::parseInlineStyleDeclaration(const String& string, Element* element)
{
    CSSParserContext context(element->document());
    context.mode = strictToCSSParserMode(element->isHTMLElement() && !element->document().inQuirksMode());
    return CSSParserImpl::parseInlineStyleDeclaration(string, element);
}

bool CSSParser::parseDeclaration(MutableStylePropertySet* declaration, const String& string)
{
    return CSSParserImpl::parseDeclarationList(declaration, string, m_context);
}

void CSSParser::parseDeclarationForInspector(const CSSParserContext& context, const String& string, CSSParserObserver& observer)
{
    CSSParserImpl::parseDeclarationListForInspector(string, context, observer);
}

PassRefPtr<CSSValue> CSSParser::parseValueWithVariableReferences(CSSPropertyID propID, const CSSValue& value, Style::BuilderState& builderState)
{
    ASSERT((propID == CSSPropertyCustom && value.isCustomPropertyValue()) || (propID != CSSPropertyCustom && !value.isCustomPropertyValue()));
    auto& style = builderState.style();
    auto direction = style.direction();
    auto writingMode = style.writingMode();

    if (value.isPendingSubstitutionValue()) {
        // FIXME: Should have a resolvedShorthands cache to stop this from being done over and over for each longhand value.
        auto& substitution = toCSSPendingSubstitutionValue(value);

        auto shorthandID = substitution.shorthandPropertyId();
        if (CSSProperty::isDirectionAwareProperty(shorthandID))
            shorthandID = CSSProperty::resolveDirectionAwareProperty(shorthandID, direction, writingMode);

        RefPtr<CSSVariableData> resolvedData = substitution.shorthandValue().resolveVariableReferences(builderState);
        if (!resolvedData)
            return nullptr;

        OwnPtr<ParsedPropertyVector> parsedProperties = adoptPtr(new ParsedPropertyVector());
        if (!CSSPropertyParser::parseValue(shorthandID, false, resolvedData->tokens(), substitution.shorthandValue().context(), *parsedProperties, StyleRuleType::Style))
            return nullptr;

        for (auto& property : *parsedProperties) {
            CSSPropertyID currentId = property.id();
            if (CSSProperty::isDirectionAwareProperty(currentId))
                currentId = CSSProperty::resolveDirectionAwareProperty(currentId, direction, writingMode);
            if (currentId == propID)
                return property.value();
        }

        return nullptr;
    }

    if (value.isVariableReferenceValue()) {
        const CSSVariableReferenceValue& valueWithReferences = toCSSVariableReferenceValue(value);
        RefPtr<CSSVariableData> resolvedData = valueWithReferences.resolveVariableReferences(builderState);
        if (!resolvedData)
            return nullptr;
        return CSSPropertyParser::parseSingleValue(propID, resolvedData->tokens(), valueWithReferences.context());
    }

    const auto& customPropValue = toCSSCustomPropertyValue(value);
    const CSSVariableReferenceValue& valueWithReferences = std::get<Ref<CSSVariableReferenceValue>>(customPropValue.value()).get();

    auto& name = toCSSCustomPropertyValue(value).name();
    auto* registered = builderState.document().getCSSRegisteredCustomPropertySet().get(name);
    String syntax = registered ? registered->syntax : "*";
    RefPtr<CSSVariableData> resolvedData = valueWithReferences.resolveVariableReferences(builderState);

    if (!resolvedData)
        return nullptr;

    // FIXME handle REM cycles.
    HashSet<CSSPropertyID> dependencies;
    CSSPropertyParser::collectParsedCustomPropertyValueDependencies(syntax, false, dependencies, resolvedData->tokens(), valueWithReferences.context());

    for (auto id : dependencies)
        builderState.builder().applyProperty(id);

    return CSSPropertyParser::parseTypedCustomPropertyValue(name, syntax, resolvedData->tokens(), builderState, valueWithReferences.context());
}

Vector<double> CSSParser::parseKeyframeKeyList(const String& selector)
{
    return CSSParserImpl::parseKeyframeKeyList(selector);
}

PassRefPtr<CSSValue> CSSParser::parseFontFaceDescriptor(CSSPropertyID propertyID, const String& propertyValue, const CSSParserContext& context)
{
    StringBuilder builder;
    builder.append("@font-face { ");
    builder.append(getPropertyNameString(propertyID));
    builder.append(" : ");
    builder.append(propertyValue);
    builder.append("; }");
    RefPtr<StyleRuleBase> rule = parseRule(context, nullptr, builder.toString());
    if (!rule || !rule->isFontFaceRule())
        return nullptr;
    return static_cast<StyleRuleFontFace&>(*rule.get()).properties().getPropertyCSSValue(propertyID);
}

}
