/*
 * Copyright (C) 2012 Adobe Systems Incorporated. All rights reserved.
 * Copyright (C) 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AS IS AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#pragma once

#include "CSSPropertyNames.h"
#include "StyleRuleType.h"
#include "TextEncoding.h"
#include "KURL.h"
#include "KURLHash.h"
#include <wtf/HashFunctions.h>
#include <wtf/text/StringHash.h>
#include <wtf/Optional.h>

namespace WebCore {

class Document;

struct ResolvedURL;

// Must not grow beyond 3 bits, due to packing in StyleProperties.
enum CSSParserMode {
    HTMLStandardMode,
    HTMLQuirksMode,
    // HTML attributes are parsed in quirks mode but also allows internal properties and values.
    HTMLAttributeMode,
    // SVG attributes are parsed in quirks mode but rules differ slightly.
    SVGAttributeMode,
    // @viewport rules are parsed in standards mode but CSSOM modifications (via StylePropertySet)
    // must call parseViewportProperties so needs a special mode.
    CSSViewportRuleMode,
    // User agent stylesheets are parsed in standards mode but also allows internal properties and values.
    UASheetMode,
    // WebVTT places limitations on external resources.
    WebVTTMode
};

inline bool isQuirksModeBehavior(CSSParserMode mode)
{
    return mode == HTMLQuirksMode || mode == HTMLAttributeMode;
}

inline bool isUASheetBehavior(CSSParserMode mode)
{
    return mode == UASheetMode;
}

inline bool isUnitLessValueParsingEnabledForMode(CSSParserMode mode)
{
    return mode == HTMLAttributeMode || mode == SVGAttributeMode;
}

inline bool isCSSViewportParsingEnabledForMode(CSSParserMode mode)
{
    return mode == CSSViewportRuleMode;
}

// FIXME-NEWPARSER: Next two functions should be removed eventually.
inline CSSParserMode strictToCSSParserMode(bool inStrictMode)
{
    return inStrictMode ? HTMLStandardMode : HTMLQuirksMode;
}

inline bool isStrictParserMode(CSSParserMode cssParserMode)
{
    return cssParserMode == UASheetMode || cssParserMode == HTMLStandardMode || cssParserMode == SVGAttributeMode;
}

struct CSSParserContext {
    WTF_MAKE_FAST_ALLOCATED;
public:
    CSSParserContext(CSSParserMode, const KURL& baseURL = KURL());
    explicit CSSParserContext(const Document&, const KURL& baseURL = KURL(), const String& charset = emptyString());

    bool isPropertyRuntimeDisabled(CSSPropertyID) const { return false; }

    ResolvedURL completeURL(const String&) const;;

    KURL baseURL;
    String charset;
    CSSParserMode mode { HTMLStandardMode };
    Optional<StyleRuleType> enclosingRuleType;
    bool isHTMLDocument { false };

    // This is only needed to support getMatchedCSSRules.
    bool hasDocumentSecurityOrigin { false };

    bool isContentOpaque { false };
    bool useSystemAppearance { false };

    // Settings.
    bool aspectRatioEnabled { false };
    bool isCSSCustomFilterEnabled { false };
    bool isCSSRegionsEnabled { false };
    bool isCSSCompositingEnabled { false };
    bool isCSSGridLayoutEnabled { true };
    bool needsSiteSpecificQuirks { false };
    bool enforcesCSSMIMETypeInNoQuirksMode { true };
    bool svgFontsEnabled { true };
    bool useLegacyBackgroundSizeShorthandBehavior { false };
    bool focusVisibleEnabled { false };
    bool hasPseudoClassEnabled { false };
    bool scrollBehaviorEnabled { false };
    bool springTimingFunctionEnabled { false };
    bool constantPropertiesEnabled { false };
    bool conicGradientsEnabled { false };
    bool individualTransformPropertiesEnabled { false };
    bool overflowClipEnabled { false };
};

bool operator==(const CSSParserContext&, const CSSParserContext&);
inline bool operator!=(const CSSParserContext& a, const CSSParserContext& b) { return !(a == b); }

const CSSParserContext& strictCSSParserContext();

struct CSSParserContextHash {
    static unsigned hash(const CSSParserContext& context)
    {
        auto hash = KURLHash::hash(context.baseURL);
        if (!context.charset.isEmpty())
            hash ^= StringHash::hash(context.charset);

        unsigned bits = context.isHTMLDocument                  << 0
            | context.hasDocumentSecurityOrigin                 << 1
            | context.isContentOpaque                           << 2
            | context.useSystemAppearance                       << 3
            | context.aspectRatioEnabled                        << 4
            | context.isCSSCustomFilterEnabled                  << 5
            | context.isCSSRegionsEnabled                       << 6
            | context.isCSSCompositingEnabled                   << 7
            | context.isCSSGridLayoutEnabled                    << 8
            | context.needsSiteSpecificQuirks                   << 9
            | context.enforcesCSSMIMETypeInNoQuirksMode         << 10
            | context.svgFontsEnabled                           << 11
            | context.useLegacyBackgroundSizeShorthandBehavior  << 12
#if ENABLE(OVERFLOW_SCROLLING_TOUCH)
            | context.legacyOverflowScrollingTouchEnabled       << 13
#endif
            | context.focusVisibleEnabled                       << 14
            | context.hasPseudoClassEnabled                     << 15
            | context.scrollBehaviorEnabled                     << 16
            | context.springTimingFunctionEnabled               << 17
#if ENABLE(TEXT_AUTOSIZING)
            | context.textAutosizingEnabled                     << 18
#endif
#if ENABLE(CSS_TRANSFORM_STYLE_OPTIMIZED_3D)
            | context.transformStyleOptimized3DEnabled          << 19
#endif
            | context.constantPropertiesEnabled                 << 20
            | context.conicGradientsEnabled                     << 21
#if ENABLE(ATTACHMENT_ELEMENT)
            | context.attachmentEnabled                         << 22
#endif
            | context.individualTransformPropertiesEnabled      << 23
            | context.overflowClipEnabled                       << 24
            | context.mode                                      << 25; // This is multiple bits, so keep it last.

        hash ^= WTF::intHash(bits);
        return hash;
    }
    static bool equal(const CSSParserContext& a, const CSSParserContext& b)
    {
        return a == b;
    }
    static const bool safeToCompareToEmptyOrDeleted = false;
};

};

namespace WTF {
template<> struct HashTraits<WebCore::CSSParserContext> : GenericHashTraits<WebCore::CSSParserContext> {
    static void constructDeletedValue(WebCore::CSSParserContext& slot) { new (NotNull, &slot.baseURL) WebCore::KURL(WTF::HashTableDeletedValue); }
    static bool isDeletedValue(const WebCore::CSSParserContext& value) { return value.baseURL.isHashTableDeletedValue(); }
    static WebCore::CSSParserContext emptyValue() { return WebCore::CSSParserContext(WebCore::HTMLStandardMode); }
};

template<> struct DefaultHash<WebCore::CSSParserContext> {
    typedef WebCore::CSSParserContextHash Hash;
};
}
