/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "AbortSignal.h"

#include "AbortAlgorithm.h"
#include "Event.h"
#include "EventNames.h"
#include "JSDOMBinding.h"
#include "ScriptExecutionContext.h"
#include "runtime/Exception.h"

namespace WebCore {

PassRefPtr<AbortSignal> AbortSignal::create(ScriptExecutionContext* context)
{
    return adoptRef(new AbortSignal(context));
}

// https://dom.spec.whatwg.org/#dom-abortsignal-abort
PassRefPtr<AbortSignal> AbortSignal::abort(ScriptState* state, ScriptExecutionContext* context, JSC::JSValue reason)
{
    ASSERT(reason);
    if (!reason || reason.isUndefined())
        reason = createDOMException(state, ABORT_ERR);
    return adoptRef(new AbortSignal(context, Aborted::Yes, reason));
}

AbortSignal::AbortSignal(ScriptExecutionContext* context, Aborted aborted, JSC::JSValue reason)
    : ContextDestructionObserver(context)
	, m_aborted(aborted == Aborted::Yes)
{
	m_reason.setWithoutWriteBarrier(reason);
}

// https://dom.spec.whatwg.org/#abortsignal-signal-abort
void AbortSignal::signalAbort(JSC::JSValue reason)
{
    // 1. If signal's aborted flag is set, then return.
    if (m_aborted)
        return;
    
    // 2. Set signal’s aborted flag.
    m_aborted = true;

    ASSERT(reason);
	m_reason.setWithoutWriteBarrier(reason);

    RefPtr<AbortSignal> protectedThis(this);
    Vector<OwnPtr<Algorithm>> algorithms;
    algorithms.swap(m_algorithms);
    for (auto& algorithm : algorithms)
        (*algorithm)();
    
    // 5. Fire an event named abort at signal.
    dispatchEvent(Event::create(eventNames().abortEvent, false, false));
}

// https://dom.spec.whatwg.org/#abortsignal-follow
void AbortSignal::signalFollow(AbortSignal& signal)
{
    if (aborted())
        return;

    if (signal.aborted()) {
    	signalAbort(signal.reason());
        return;
    }

    class AlgorithmImpl final : public Algorithm
	{
	public:
    	AlgorithmImpl(AbortSignal& thisRef) : weakThis(makeWeakPtr(thisRef)) { }
    	virtual void operator()() final
    	{
            if (weakThis)
            	weakThis->signalAbort(weakThis->m_followingSignal ? weakThis->m_followingSignal->reason() : JSC::jsUndefined());
    	}
	private:
    	WeakPtr<AbortSignal> weakThis;
	};
    ASSERT(!m_followingSignal);
    m_followingSignal = makeWeakPtr(signal);
    signal.addAlgorithm(adoptPtr(new AlgorithmImpl(*this)));
}

bool AbortSignal::whenSignalAborted(AbortSignal& signal, PassRefPtr<AbortAlgorithm> algorithm)
{
	if (!algorithm)
		return false;

    if (signal.aborted()) {
        algorithm->handleEvent();
        return true;
    }

    class AlgorithmImpl final : public Algorithm
	{
	public:
    	AlgorithmImpl(PassRefPtr<AbortAlgorithm> algorithm) : algorithm(algorithm) { }
    	virtual void operator()() final
    	{
            algorithm->handleEvent();
    	}
	private:
    	RefPtr<AbortAlgorithm> algorithm;
	};
    signal.addAlgorithm(adoptPtr(new AlgorithmImpl(algorithm)));
    return false;
}

void AbortSignal::throwIfAborted(JSC::ExecState* exec)
{
    if (!aborted())
        return;

    auto& vm = exec->lexicalGlobalObject()->vm();
    auto scope = DECLARE_THROW_SCOPE(vm);
    throwException(exec, scope, m_reason.get());
}

JSC::JSValue AbortSignal::reason() const
{
	return m_reason.get();
}

void AbortSignal::visitAdditionalChildren(JSC::SlotVisitor& visitor)
{
    visitor.append(&m_reason);
}

} // namespace WebCore
