/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ContextDestructionObserver.h"
#include "EventTarget.h"
#include "ScriptState.h"
#include <runtime/WriteBarrier.h>
#include <wtf/Ref.h>
#include <wtf/RefCounted.h>
#include <wtf/WeakPtr.h>

namespace JSC
{
class ExecState;
class SlotVisitor;
}

namespace WebCore {

class AbortAlgorithm;
class ScriptExecutionContext;

class AbortSignal FINAL : public RefCounted<AbortSignal>, public EventTargetWithInlineData, private ContextDestructionObserver {
public:
    static PassRefPtr<AbortSignal> create(ScriptExecutionContext*);

    static PassRefPtr<AbortSignal> abort(ScriptState*, ScriptExecutionContext*, JSC::JSValue reason = JSC::JSValue());

    static bool whenSignalAborted(AbortSignal&, PassRefPtr<AbortAlgorithm>);

    void signalAbort(JSC::JSValue reason);
    void signalFollow(AbortSignal&);

    bool aborted() const { return m_aborted; }
    JSC::JSValue reason() const;

    using RefCounted::ref;
    using RefCounted::deref;

    void setOnabort(PassRefPtr<EventListener> listener, DOMWrapperWorld* isolatedWorld)
    {
        setAttributeEventListener(eventNames().abortEvent, listener, isolatedWorld);
    }
    EventListener* onabort(DOMWrapperWorld* isolatedWorld) { return attributeEventListener(eventNames().abortEvent, isolatedWorld); }

    class Algorithm
	{
    	WTF_MAKE_FAST_ALLOCATED;
	public:
    	virtual ~Algorithm() = default;
    	virtual void operator()() = 0;
	};
    void addAlgorithm(PassOwnPtr<Algorithm> algorithm) { m_algorithms.append(algorithm); }

    bool isFollowingSignal() const { return !!m_followingSignal; }

    void throwIfAborted(JSC::ExecState*);

    void visitAdditionalChildren(JSC::SlotVisitor&);

private:
    enum class Aborted : bool { No, Yes };
    explicit AbortSignal(ScriptExecutionContext*, Aborted = Aborted::No, JSC::JSValue reason = JSC::JSValue());

    // EventTarget.
    virtual EventTargetInterface eventTargetInterface() const FINAL { return AbortSignalEventTargetInterfaceType; }
    virtual ScriptExecutionContext* scriptExecutionContext() const FINAL { return ContextDestructionObserver::scriptExecutionContext(); }
    virtual void refEventTarget() FINAL { ref(); }
    virtual void derefEventTarget() FINAL { deref(); }

    bool m_aborted { false };
    Vector<OwnPtr<Algorithm>> m_algorithms;
    WeakPtr<AbortSignal> m_followingSignal;
    JSC::WriteBarrier<JSC::Unknown> m_reason;
};

}

