/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "ContainerNode.h"

#include "AllDescendantsCollection.h"
#include "ChildChangeInvalidation.h"
#include "ChildListMutationScope.h"
#include "ClassCollection.h"
#include "ContainerNodeAlgorithms.h"
#include "LiveNodeList.h"
#include "ElementInlines.h"
#include "ElementTraversal.h"
#include "EventNames.h"
#include "ExceptionCode.h"
#include "FloatRect.h"
#include "Frame.h"
#include "FrameView.h"
#include "GenericCachedHTMLCollection.h"
#include "HTMLFormControlsCollection.h"
#include "HTMLOptionsCollection.h"
#include "HTMLSlotElement.h"
#include "HTMLTableRowsCollection.h"
#include "HTMLNames.h"
#include "InlineTextBox.h"
#include "InspectorInstrumentation.h"
#include "JSNode.h"
#include "JSNodeCustom.h"
#include "LabelsNodeList.h"
#include "MutationEvent.h"
#include "NameNodeList.h"
#include "NodeOrString.h"
#include "NodeRareData.h"
#include "NodeRenderStyle.h"
#include "NodeTraversal.h"
#include "ResourceLoadScheduler.h"
#include "Page.h"
#include "RadioNodeList.h"
#include "RenderBox.h"
#include "RenderTheme.h"
#include "RenderTreeUpdater.h"
#include "RenderWidget.h"
#include "RootInlineBox.h"
#include "RuntimeEnabledFeatures.h"
#include "SelectorQuery.h"
#include "SlotAssignment.h"
#include "SVGDocumentExtensions.h"
#include "SVGElement.h"
#include "SVGNames.h"
#include "SVGUseElement.h"
#include "ScriptDisallowedScope.h"
#include "TagCollection.h"
#include "TemplateContentDocumentFragment.h"
#include "Text.h"
#include <wtf/CurrentTime.h>
#include <wtf/Vector.h>

using namespace std;

namespace WebCore {

static void dispatchChildInsertionEvents(Node&);
static void dispatchChildRemovalEvents(Ref<Node>&);

unsigned ScriptDisallowedScope::s_count = 0;

ALWAYS_INLINE void ContainerNode::removeAllChildrenWithScriptAssertion(ChildChangeSource source, NodeVector& children, DeferChildrenChanged deferChildrenChanged)
{
    ASSERT(children.isEmpty());
    collectChildNodes(*this, children);

    if (UNLIKELY(isDocumentFragmentForInnerOuterHTML())) {
        ScriptDisallowedScope::InMainThread scriptDisallowedScope;
        RELEASE_ASSERT(!connectedSubframeCount() && !hasRareData() && !wrapper());
        ASSERT(!weakPtrFactory().isInitialized());
        while (RefPtr<Node> child = m_firstChild)
            removeBetween(nullptr, child->nextSibling(), child.get());
        document().incDOMTreeVersion();
        return;
    }

    if (source == ContainerNode::ChildChangeSource::API) {
        ChildListMutationScope mutation(this);
        for (auto& child : children) {
            mutation.willRemoveChild(&child.get());
            child->notifyMutationObserversNodeWillDetach();
            dispatchChildRemovalEvents(child);
        }
    } else {
        ASSERT(source == ContainerNode::ChildChangeSource::Parser);
        ScriptDisallowedScope::InMainThread scriptDisallowedScope;
        if (UNLIKELY(document().hasMutationObserversOfType(MutationObserver::ChildList))) {
            ChildListMutationScope mutation(this);
            for (auto& child : children)
                mutation.willRemoveChild(&child.get());
        }
    }

    disconnectSubframesIfNeeded(*this, DescendantsOnly);

    ContainerNode::ChildChange childChange { ChildChangeType::AllChildrenRemoved, nullptr, nullptr, source };

    WidgetHierarchyUpdatesSuspensionScope suspendWidgetHierarchyUpdates;
    ScriptDisallowedScope::InMainThread scriptDisallowedScope;
    {
        Style::ChildChangeInvalidation styleInvalidation(*this, childChange);

		if (UNLIKELY(isShadowRoot() || isInShadowTree()))
			containingShadowRoot()->willRemoveAllChildren(*this);

		document().nodeChildrenWillBeRemoved(*this);

		while (RefPtr<Node> child = m_firstChild) {
			removeBetween(nullptr, child->nextSibling(), child.get());
	        auto subtreeObservability = notifyChildNodeRemoved(*this, *child);
	        if (source == ChildChangeSource::API && subtreeObservability == RemovedSubtreeObservability::MaybeObservableByRefPtr)
	            willCreatePossiblyOrphanedTreeByRemoval(child.get());
	    }
    }

    ASSERT_WITH_SECURITY_IMPLICATION(!document().selection().selection().isOrphan());

    if (deferChildrenChanged == DeferChildrenChanged::No) {
#if ASSERT_ENABLED
        auto treeVersion = document().domTreeVersion();
#endif
        childrenChanged(childChange);
        ASSERT_WITH_SECURITY_IMPLICATION(document().domTreeVersion() > treeVersion);
    }
}

static ContainerNode::ChildChange makeChildChangeForRemoval(Node& childToRemove, ContainerNode::ChildChangeSource source)
{
    auto changeType = [&] {
        if (childToRemove.isElementNode())
            return ContainerNode::ChildChangeType::ElementRemoved;
        if (childToRemove.isTextNode())
            return ContainerNode::ChildChangeType::TextRemoved;
        return ContainerNode::ChildChangeType::NonContentsChildRemoved;
    }();

    return {
        changeType,
        ElementTraversal::previousSibling(childToRemove),
        ElementTraversal::nextSibling(childToRemove),
        source
    };
}

ALWAYS_INLINE bool ContainerNode::removeNodeWithScriptAssertion(Node& childToRemove, ChildChangeSource source)
{
    Ref<Node> protectedChildToRemove(childToRemove);
    ASSERT_WITH_SECURITY_IMPLICATION(childToRemove.parentNode() == this);
    {
        ScriptDisallowedScope::InMainThread scriptDisallowedScope;
        ChildListMutationScope(this).willRemoveChild(&childToRemove);
    }

    ASSERT_WITH_SECURITY_IMPLICATION(ScriptDisallowedScope::InMainThread::isEventDispatchAllowedInSubtree(childToRemove));
	if (source == ContainerNode::ChildChangeSource::API) {
        childToRemove.notifyMutationObserversNodeWillDetach();
        dispatchChildRemovalEvents(protectedChildToRemove);
        if (childToRemove.parentNode() != this)
            return false;
    }

    if (source == ContainerNode::ChildChangeSource::Parser) {
        // FIXME: Merge these two code paths. It's a bug in the parser not to update connectedSubframeCount in time.
        disconnectSubframesIfNeeded(*this, DescendantsOnly);
    } else {
        if (childToRemove.isContainerNode())
            disconnectSubframesIfNeeded(toContainerNode(childToRemove), RootAndDescendants);
    }

    if (childToRemove.parentNode() != this)
        return false;

    auto childChange = makeChildChangeForRemoval(childToRemove, source);

    RemovedSubtreeObservability subtreeObservability;
    {
		WidgetHierarchyUpdatesSuspensionScope suspendWidgetHierarchyUpdates;
		ScriptDisallowedScope::InMainThread scriptDisallowedScope;
        Style::ChildChangeInvalidation styleInvalidation(*this, childChange);

        if (UNLIKELY(isShadowRoot() || isInShadowTree()))
            containingShadowRoot()->resolveSlotsBeforeNodeInsertionOrRemoval();

		document().nodeWillBeRemoved(childToRemove);

		ASSERT_WITH_SECURITY_IMPLICATION(childToRemove.parentNode() == this);
		ASSERT(!childToRemove.isDocumentFragment());

		RefPtr<Node> previousSibling = childToRemove.previousSibling();
		RefPtr<Node> nextSibling = childToRemove.nextSibling();

		removeBetween(previousSibling.get(), nextSibling.get(), &childToRemove);
		subtreeObservability = notifyChildNodeRemoved(*this, childToRemove);
    }

    if (source == ChildChangeSource::API && subtreeObservability == RemovedSubtreeObservability::MaybeObservableByRefPtr)
        willCreatePossiblyOrphanedTreeByRemoval(&childToRemove);

    ASSERT_WITH_SECURITY_IMPLICATION(!document().selection().selection().isOrphan());

    // FIXME: Move childrenChanged into ScriptDisallowedScope block.
    childrenChanged(childChange);

    return true;
}

enum class ReplacedAllChildren { No, Yes };

static ContainerNode::ChildChange makeChildChangeForInsertion(ContainerNode& containerNode, Node& child, Node* beforeChild, ContainerNode::ChildChangeSource source, ReplacedAllChildren replacedAllChildren)
{
    if (replacedAllChildren == ReplacedAllChildren::Yes)
        return { ContainerNode::ChildChangeType::AllChildrenReplaced, nullptr, nullptr, source };

    auto changeType = [&] {
        if (child.isElementNode())
            return ContainerNode::ChildChangeType::ElementInserted;
        if (child.isTextNode())
            return ContainerNode::ChildChangeType::TextInserted;
        return ContainerNode::ChildChangeType::NonContentsChildInserted;
    }();

    return {
        changeType,
        beforeChild ? ElementTraversal::previousSibling(*beforeChild) : ElementTraversal::lastChild(&containerNode),
        !beforeChild || (*beforeChild).isElementNode() ? toElement(beforeChild) : ElementTraversal::nextSibling(*beforeChild),
        source
    };
}

template<typename DOMInsertionWork>
static ALWAYS_INLINE void executeNodeInsertionWithScriptAssertion(ContainerNode& containerNode, Node& child, Node* beforeChild,
    ContainerNode::ChildChangeSource source, ReplacedAllChildren replacedAllChildren, DOMInsertionWork doNodeInsertion)
{
    auto childChange = makeChildChangeForInsertion(containerNode, child, beforeChild, source, replacedAllChildren);

    NodeVector postInsertionNotificationTargets;
    {
        WidgetHierarchyUpdatesSuspensionScope suspendWidgetHierarchyUpdates;
        ScriptDisallowedScope::InMainThread scriptDisallowedScope;
        Style::ChildChangeInvalidation styleInvalidation(containerNode, childChange);

        if (UNLIKELY(containerNode.isShadowRoot() || containerNode.isInShadowTree()))
            containerNode.containingShadowRoot()->resolveSlotsBeforeNodeInsertionOrRemoval();

        doNodeInsertion();
        ChildListMutationScope(&containerNode).childAdded(&child);
        notifyChildNodeInserted(containerNode, child, postInsertionNotificationTargets);
	}

    // FIXME: Move childrenChanged into ScriptDisallowedScope block.
    containerNode.childrenChanged(childChange);

    ASSERT(ScriptDisallowedScope::InMainThread::isEventDispatchAllowedInSubtree(child));
    for (auto& target : postInsertionNotificationTargets)
        target->didFinishInsertingNode();

    if (source == ContainerNode::ChildChangeSource::API)
        dispatchChildInsertionEvents(child);
}

void ContainerNode::removeSelfOrChildNodesForInsertion(Node& child, NodeVector& nodesForInsertion, ExceptionCode& ec)
{
    if (child.isDocumentFragment()) {
        auto& fragment = toDocumentFragment(child);
        if (!fragment.hasChildNodes())
            return;

        ASSERT(nodesForInsertion.isEmpty());
        fragment.removeAllChildrenWithScriptAssertion(ContainerNode::ChildChangeSource::API, nodesForInsertion);

        fragment.rebuildSVGExtensionsElementsIfNecessary();
        fragment.dispatchSubtreeModifiedEvent();

        return;
    }

	nodesForInsertion.append(child);
	RefPtr<ContainerNode> oldParent = child.parentNode();
	if (!oldParent)
		return;
	oldParent->removeChild(child, ec);
}

// FIXME: This function must get a new name.
// It removes all children, not just a category called "detached children".
// So this name is terribly confusing.
void ContainerNode::removeDetachedChildren()
{
    if (connectedSubframeCount()) {
        for (Node* child = firstChild(); child; child = child->nextSibling())
            child->updateAncestorConnectedSubframeCountForRemoval();
    }
    // FIXME: We should be able to ASSERT(!attached()) here: https://bugs.webkit.org/show_bug.cgi?id=107801
    removeDetachedChildrenInContainer(*this);
}

static inline bool mayHaveDisplayContents(Element *element)
{
    return element && (element->hasDisplayContents() || element->displayContentsChanged());
}

static inline void destroyRenderTreeIfNeeded(Node& child)
{
	Element* childAsElement = WebCore::isElement(child) ? &toElement(child) : nullptr;
    if (!child.renderer() && !mayHaveDisplayContents(childAsElement))
        return;
    if (childAsElement)
    	RenderTreeUpdater::tearDownRenderers(*childAsElement);
    else if (child.isTextNode())
    	RenderTreeUpdater::tearDownRenderer(*toText(&child));
}

void ContainerNode::takeAllChildrenFrom(ContainerNode* oldParent)
{
    ASSERT(oldParent);

    NodeVector children;
    oldParent->removeAllChildrenWithScriptAssertion(ChildChangeSource::Parser, children);

    for (auto& child : children) {
    	if (child->parentNode()) // Previous parserAppendChild may have mutated DOM.
			continue;
        ASSERT(!ensurePreInsertionValidity(child, 0).hasException());
        child->setTreeScopeRecursively(treeScope());
        parserAppendChild(child);
    }
}

ContainerNode::~ContainerNode()
{
    if (!isDocumentNode())
    	willBeDeletedFrom(&document());
    removeDetachedChildren();
}

static bool containsIncludingHostElements(const Node& possibleAncestor, const Node& node)
{
    const Node* currentNode = &node;
    do {
        if (currentNode == &possibleAncestor)
            return true;
        const ContainerNode* parent = currentNode->parentNode();
        if (!parent) {
            if (currentNode->isShadowRoot())
                parent = toShadowRoot(currentNode)->host();
            else if ((*currentNode).isDocumentFragment() && toDocumentFragment(*currentNode).isTemplateContent())
                parent = toTemplateContentDocumentFragment(*currentNode).host();
        }
        currentNode = parent;
    } while (currentNode);

    return false;
}

static inline bool isChildTypeAllowed(ContainerNode& newParent, Node& child)
{
    if (!child.isDocumentFragment())
        return newParent.childTypeAllowed(child.nodeType());

    for (Node* node = child.firstChild(); node; node = node->nextSibling()) {
        if (!newParent.childTypeAllowed(node->nodeType()))
            return false;
    }
    return true;
}

enum class ShouldValidateChildParent { No, Yes };
static inline ExceptionCode checkAcceptChild(ContainerNode& newParent, Node& newChild, const Node* refChild, Document::AcceptChildOperation operation, ShouldValidateChildParent shouldValidateChildParent)
{
    if (containsIncludingHostElements(newChild, newParent))
        return HIERARCHY_REQUEST_ERR;

    // Use common case fast path if possible.
    if ((newChild.isElementNode() || newChild.isTextNode()) && newParent.isElementNode()) {
        ASSERT(!newParent->isDocumentTypeNode());
        ASSERT(isChildTypeAllowed(newParent, newChild));
        if (shouldValidateChildParent == ShouldValidateChildParent::Yes && refChild && refChild->parentNode() != &newParent)
			return NOT_FOUND_ERR;
        return 0;
    }

    // This should never happen, but also protect release builds from tree corruption.
    ASSERT(!newChild.isPseudoElement());
    if (newChild.isPseudoElement())
    	return HIERARCHY_REQUEST_ERR;

    if (shouldValidateChildParent == ShouldValidateChildParent::Yes && refChild && refChild->parentNode() != &newParent)
    	return NOT_FOUND_ERR;

    if (newParent.isDocumentNode()) {
        if (!toDocument(newParent).canAcceptChild(newChild, refChild, operation))
            return HIERARCHY_REQUEST_ERR;
    } else if (!isChildTypeAllowed(newParent, newChild))
        return HIERARCHY_REQUEST_ERR;

    return 0;
}

static inline bool checkAcceptChildGuaranteedNodeTypes(ContainerNode& newParent, Node& newChild, ExceptionCode& ec)
{
    ASSERT(!newParent.isDocumentTypeNode());
    ASSERT(isChildTypeAllowed(newParent, newChild));
    if (containsIncludingHostElements(newChild, newParent)) {
        ec = HIERARCHY_REQUEST_ERR;
        return false;
    }

    return true;
}

// https://dom.spec.whatwg.org/#concept-node-ensure-pre-insertion-validity
bool ContainerNode::ensurePreInsertionValidity(Node& newChild, Node* refChild, ExceptionCode& ec)
{
    ec = checkAcceptChild(*this, newChild, refChild, Document::AcceptChildOperation::InsertOrAdd, ShouldValidateChildParent::Yes);
    return !ec;
}

// https://dom.spec.whatwg.org/#concept-node-replace
static inline bool checkPreReplacementValidity(ContainerNode& newParent, Node& newChild, Node& oldChild, ShouldValidateChildParent shouldValidateChildParent, ExceptionCode& ec)
{
    ec = checkAcceptChild(newParent, newChild, &oldChild, Document::AcceptChildOperation::Replace, shouldValidateChildParent);
    return !ec;
}

bool ContainerNode::insertBefore(Node& newChild, Node* refChild, ExceptionCode& ec)
{
    // Check that this node is not "floating".
    // If it is, it can be deleted as a side effect of sending mutation events.
    ASSERT(refCount() || parentOrShadowHostNode());

    ec = 0;

    // Make sure adding the new child is OK.
    if (!ensurePreInsertionValidity(newChild, refChild, ec))
        return false;

    if (refChild == &newChild)
        refChild = newChild.nextSibling();

    // insertBefore(node, null) is equivalent to appendChild(node)
    if (!refChild)
        return appendChildWithoutPreInsertionValidityCheck(newChild, ec);

    Ref<ContainerNode> protectedThis(*this);
    Ref<Node> next(*refChild);

    NodeVector targets;
    removeSelfOrChildNodesForInsertion(newChild, targets, ec);
    if (ec)
        return false;
    if (targets.isEmpty())
        return true;

    // We need this extra check because removeSelfOrChildNodesForInsertion() can fire mutation events.
    for (auto& child : targets) {
    	if (!checkAcceptChildGuaranteedNodeTypes(*this, child, ec))
    		return false;
    }

    InspectorInstrumentation::willInsertDOMNode(&document(), this);

#if ENABLE(MUTATION_OBSERVERS)
    ChildListMutationScope mutation(this);
#endif

    for (auto& child : targets) {
        // Due to arbitrary code running in response to a DOM mutation event it's
        // possible that "next" is no longer a child of "this".
        // It's also possible that "child" has been inserted elsewhere.
        // In either of those cases, we'll just stop.
        if (next->parentNode() != this)
            break;
        if (child->parentNode())
            break;

        executeNodeInsertionWithScriptAssertion(*this, child.get(), next.ptr(), ChildChangeSource::API, ReplacedAllChildren::No, [&] {
            child->setTreeScopeRecursively(treeScope());
            insertBeforeCommon(next, child.get());
        });
    }

    dispatchSubtreeModifiedEvent();
    return true;
}

void ContainerNode::insertBeforeCommon(Node& nextChild, Node& newChild)
{
    ASSERT(!newChild.parentNode()); // Use insertBefore if you need to handle reparenting (and want DOM mutation events).
    ASSERT(!newChild.nextSibling());
    ASSERT(!newChild.previousSibling());
    ASSERT(!newChild.isShadowRoot());

    forbidEventDispatch();
    Node* prev = nextChild.previousSibling();
    ASSERT(m_lastChild != prev);
    nextChild.setPreviousSibling(&newChild);
    if (prev) {
        ASSERT(m_firstChild != nextChild);
        ASSERT(prev->nextSibling() == nextChild);
        prev->setNextSibling(&newChild);
    } else {
        ASSERT(m_firstChild == nextChild);
        m_firstChild = &newChild;
    }
    newChild.setParentNode(this);
    newChild.setPreviousSibling(prev);
    newChild.setNextSibling(&nextChild);
    allowEventDispatch();
}

void ContainerNode::appendChildCommon(Node& child)
{
    child.setParentNode(this);

    if (m_lastChild) {
        child.setPreviousSibling(m_lastChild);
        m_lastChild->setNextSibling(&child);
    } else
        m_firstChild = &child;

    m_lastChild = &child;
}

void ContainerNode::parserInsertBefore(Node& newChild, Node& nextChild)
{
    ASSERT(nextChild.parentNode() == this);
    ASSERT(!newChild.isDocumentFragment());
#if ENABLE(TEMPLATE_ELEMENT)
    ASSERT(!hasTagName(HTMLNames::templateTag));
#endif

    if (nextChild.previousSibling() == &newChild || &nextChild == &newChild) // nothing to do
        return;

    executeNodeInsertionWithScriptAssertion(*this, newChild, &nextChild, ChildChangeSource::Parser, ReplacedAllChildren::No, [&] {
		if (&document() != &newChild.document())
			document().adoptNode(newChild, ASSERT_NO_EXCEPTION);

		insertBeforeCommon(nextChild, newChild);

        newChild.setTreeScopeRecursively(treeScope());
		newChild.updateAncestorConnectedSubframeCountForInsertion();
    });
}

bool ContainerNode::replaceChild(Node& newChild, Node& oldChild, ExceptionCode& ec)
{
    // Check that this node is not "floating".
    // If it is, it can be deleted as a side effect of sending mutation events.
    ASSERT(refCount() || parentOrShadowHostNode());

    Ref<Node> protect(*this);

    ec = 0;

    // Make sure replacing the old child with the new is ok
    if (!checkPreReplacementValidity(*this, newChild, oldChild, ShouldValidateChildParent::Yes, ec))
		return false;

    RefPtr<Node> refChild = oldChild.nextSibling();
    if (refChild.get() == &newChild)
        refChild = refChild->nextSibling();

    NodeVector targets;
    {
    	ChildListMutationScope mutation(this);
    	removeSelfOrChildNodesForInsertion(newChild, targets, ec);
		if (ec)
			return false;
    }

    // Does this one more time because removeSelfOrChildNodesForInsertion() fires a MutationEvent.
    for (auto& child : targets) {
    	if (!checkPreReplacementValidity(*this, child, oldChild, ShouldValidateChildParent::No, ec))
    		return false;
    }

    // Remove the node we're replacing.
    Ref<Node> protectOldChild(oldChild);

#if ENABLE(MUTATION_OBSERVERS)
    ChildListMutationScope mutation(this);
#endif

    // If oldChild == newChild then oldChild no longer has a parent at this point.
    if (oldChild.parentNode()) {
		removeChild(oldChild, ec);
		if (ec)
			return false;

		// Does this one more time because removeChild() fires a MutationEvent.
		for (auto& child : targets) {
			if (!checkPreReplacementValidity(*this, child, oldChild, ShouldValidateChildParent::No, ec))
				return false;
		}
    }

    InspectorInstrumentation::willInsertDOMNode(&document(), this);

    // Add the new child(ren).
    for (auto& child : targets) {
        // Due to arbitrary code running in response to a DOM mutation event it's
        // possible that "refChild" is no longer a child of "this".
        // It's also possible that "child" has been inserted elsewhere.
        // In either of those cases, we'll just stop.
        if (refChild && refChild->parentNode() != this)
            break;
        if (child->parentNode())
            break;

        executeNodeInsertionWithScriptAssertion(*this, child.get(), refChild.get(), ChildChangeSource::API, ReplacedAllChildren::No, [&] {
			child->setTreeScopeRecursively(treeScope());
			if (refChild)
				insertBeforeCommon(*refChild, child);
			else
                appendChildCommon(child);
        });
    }

    dispatchSubtreeModifiedEvent();
    return true;
}

void ContainerNode::disconnectDescendantFrames()
{
    disconnectSubframesIfNeeded(*this, RootAndDescendants);
}

bool ContainerNode::removeChild(Node& oldChild, ExceptionCode& ec)
{
    // Check that this node is not "floating".
    // If it is, it can be deleted as a side effect of sending mutation events.
    ASSERT(refCount() || parentOrShadowHostNode());

    Ref<Node> protect(*this);

    ec = 0;

    // NOT_FOUND_ERR: Raised if oldChild is not a child of this node.
    if (oldChild.parentNode() != this) {
        ec = NOT_FOUND_ERR;
        return false;
    }

    if (!removeNodeWithScriptAssertion(oldChild, ChildChangeSource::API)) {
        ec = NOT_FOUND_ERR;
        return false;
    }

    rebuildSVGExtensionsElementsIfNecessary();
    dispatchSubtreeModifiedEvent();

    return true;
}

void ContainerNode::removeBetween(Node* previousChild, Node* nextChild, Node* oldChild)
{
    ASSERT(oldChild);
    ASSERT(oldChild->parentNode() == this);

	ScriptDisallowedScope::InMainThread assertNoEventDispatch;

    destroyRenderTreeIfNeeded(*oldChild);

    if (UNLIKELY(hasShadowRootContainingSlots()))
        shadowRoot()->willRemoveAssignedNode(*oldChild);

    if (nextChild)
        nextChild->setPreviousSibling(previousChild);
    if (previousChild)
        previousChild->setNextSibling(nextChild);
    if (m_firstChild == oldChild)
        m_firstChild = nextChild;
    if (m_lastChild == oldChild)
        m_lastChild = previousChild;

    oldChild->setPreviousSibling(0);
    oldChild->setNextSibling(0);
    oldChild->setParentNode(0);

    oldChild->setTreeScopeRecursively(document());
}

void ContainerNode::parserRemoveChild(Node& oldChild)
{
    removeNodeWithScriptAssertion(oldChild, ChildChangeSource::Parser);
}

// https://dom.spec.whatwg.org/#concept-node-replace-all
void ContainerNode::replaceAll(Node* node)
{
    if (!node) {
        ChildListMutationScope mutation(this);
        removeChildren();
        return;
    }

    // FIXME: The code below is roughly correct for a new text node with no parent, but needs enhancement to work properly for more complex cases.

    if (!hasChildNodes()) {
        appendChildWithoutPreInsertionValidityCheck(*node, IGNORE_EXCEPTION);
        return;
    }

    Ref<ContainerNode> protectedThis(*this);
    ChildListMutationScope mutation(this);
    NodeVector removedChildren;
    removeAllChildrenWithScriptAssertion(ChildChangeSource::API, removedChildren, DeferChildrenChanged::Yes);

    executeNodeInsertionWithScriptAssertion(*this, *node, nullptr, ChildChangeSource::API, ReplacedAllChildren::Yes, [&] {
        ASSERT(!ensurePreInsertionValidity(node, nullptr).hasException());
        InspectorInstrumentation::willInsertDOMNode(&document(), this);
        node->setTreeScopeRecursively(treeScope());
        appendChildCommon(*node);
    });

    rebuildSVGExtensionsElementsIfNecessary();
    dispatchSubtreeModifiedEvent();
}

// https://dom.spec.whatwg.org/#string-replace-all
void ContainerNode::stringReplaceAll(const String& string)
{
    replaceAll(string.isEmpty() ? nullptr : document().createTextNode(string).ptr());
}

inline void ContainerNode::rebuildSVGExtensionsElementsIfNecessary()
{
    if (auto* svgExtensions = document().svgExtensionsIfExists()) {
        Element* shadowHost = this->shadowHost();
        if (!shadowHost || !isSVGUseElement(*shadowHost))
        	svgExtensions->rebuildElements();
    }
}

// this differs from other remove functions because it forcibly removes all the children,
// regardless of read-only status or event exceptions, e.g.
void ContainerNode::removeChildren()
{
    if (!m_firstChild)
        return;

    RefPtr<ContainerNode> protect(this);
    NodeVector removedChildren;
    removeAllChildrenWithScriptAssertion(ChildChangeSource::API, removedChildren);

    rebuildSVGExtensionsElementsIfNecessary();
    dispatchSubtreeModifiedEvent();
}

bool ContainerNode::appendChild(Node& newChild, ExceptionCode& ec)
{
    // Check that this node is not "floating".
    // If it is, it can be deleted as a side effect of sending mutation events.
    ASSERT(refCount() || parentOrShadowHostNode());

    ec = 0;

    // Make sure adding the new child is ok
    if (!ensurePreInsertionValidity(newChild, nullptr, ec))
		return false;

    return appendChildWithoutPreInsertionValidityCheck(newChild, ec);
}

bool ContainerNode::appendChildWithoutPreInsertionValidityCheck(Node& newChild, ExceptionCode& ec)
{
    Ref<ContainerNode> protectedThis(*this);

    NodeVector targets;
    removeSelfOrChildNodesForInsertion(newChild, targets, ec);
    if (ec)
        return false;

    if (targets.isEmpty())
        return true;

    // We need this extra check because removeSelfOrChildNodesForInsertion() can fire mutation events.
    for (auto& child : targets) {
    	if (!checkAcceptChildGuaranteedNodeTypes(*this, child, ec))
    		return false;
    }

    InspectorInstrumentation::willInsertDOMNode(&document(), this);

#if ENABLE(MUTATION_OBSERVERS)
    ChildListMutationScope mutation(this);
#endif

    // Now actually add the child(ren)
    for (auto& child : targets) {
        // If the child has a parent again, just stop what we're doing, because
        // that means someone is doing something with DOM mutation -- can't re-parent
        // a child that already has a parent.
        if (child->parentNode())
            break;

        // Append child to the end of the list
        executeNodeInsertionWithScriptAssertion(*this, child.get(), nullptr, ChildChangeSource::API, ReplacedAllChildren::No, [&] {
			child->setTreeScopeRecursively(treeScope());
            appendChildCommon(child.get());
        });
    }

    dispatchSubtreeModifiedEvent();
    return true;
}

void ContainerNode::parserAppendChild(Node& newChild)
{
    ASSERT(!newChild.parentNode()); // Use appendChild if you need to handle reparenting (and want DOM mutation events).
    ASSERT(!newChild.isDocumentFragment());
#if ENABLE(TEMPLATE_ELEMENT)
    ASSERT(!hasTagName(HTMLNames::templateTag));
#endif

    executeNodeInsertionWithScriptAssertion(*this, newChild, nullptr, ChildChangeSource::Parser, ReplacedAllChildren::No, [&] {
		if (&document() != &newChild.document())
			document().adoptNode(newChild, ASSERT_NO_EXCEPTION);

        appendChildCommon(newChild);
        newChild.setTreeScopeRecursively(treeScope());
        newChild.updateAncestorConnectedSubframeCountForInsertion();
    });
}

void ContainerNode::childrenChanged(const ChildChange& change)
{
    document().incDOMTreeVersion();
    if (change.source == ChildChangeSource::API && change.type != TextChanged)
        document().updateRangesAfterChildrenChanged(*this);
    invalidateNodeListAndCollectionCachesInAncestors();
}

void ContainerNode::cloneChildNodes(ContainerNode& clone)
{
    ExceptionCode ec = 0;
    Document& targetDocument = clone.document();
    for (Node* child = firstChild(); child && !ec; child = child->nextSibling()) {
        Ref<Node> clonedChild = child->cloneNodeInternal(targetDocument, SelfWithTemplateContent);
        clone.appendChild(clonedChild, ec);

        if (!ec && child->isContainerNode())
           toContainerNode(*child).cloneChildNodes(toContainerNode(clonedChild));
    }
}

unsigned ContainerNode::countChildNodes() const
{
    unsigned count = 0;
    for (Node* child = firstChild(); child; child = child->nextSibling())
        ++count;
    return count;
}

Node* ContainerNode::traverseToChildAt(unsigned index) const
{
    Node* child = firstChild();
    for (; child && index > 0; --index)
        child = child->nextSibling();
    return child;
}

static void dispatchChildInsertionEvents(Node& child)
{
    if (child.isInShadowTree())
        return;

    ASSERT_WITH_SECURITY_IMPLICATION(ScriptDisallowedScope::InMainThread::isEventDispatchAllowedInSubtree(child));

    RefPtr<Node> c = child;
    Ref<Document> document = child.document();

    if (c->parentNode() && document->hasListenerType(Document::DOMNODEINSERTED_LISTENER))
        c->dispatchScopedEvent(MutationEvent::create(eventNames().DOMNodeInsertedEvent, true, c->parentNode()));

    // dispatch the DOMNodeInsertedIntoDocument event to all descendants
    if (c->inDocument() && document->hasListenerType(Document::DOMNODEINSERTEDINTODOCUMENT_LISTENER)) {
        for (; c; c = NodeTraversal::next(c.get(), &child))
            c->dispatchScopedEvent(MutationEvent::create(eventNames().DOMNodeInsertedIntoDocumentEvent, false));
    }
}

static void dispatchChildRemovalEvents(Ref<Node>& child)
{
    ASSERT_WITH_SECURITY_IMPLICATION(NoEventDispatchAssertion::isEventDispatchAllowedInSubtree(child));
    InspectorInstrumentation::willRemoveDOMNode(&child->document(), &child.get());

    if (child->isInShadowTree())
        return;

    Ref<Document> document = child->document();

    // dispatch pre-removal mutation events
    if (child->parentNode() && document->hasListenerType(Document::DOMNODEREMOVED_LISTENER))
        child->dispatchScopedEvent(MutationEvent::create(eventNames().DOMNodeRemovedEvent, true, child->parentNode()));

    // dispatch the DOMNodeRemovedFromDocument event to all descendants
    if (child->isConnected() && document->hasListenerType(Document::DOMNODEREMOVEDFROMDOCUMENT_LISTENER)) {
        for (RefPtr<Node> currentNode = child.copyRef(); currentNode; currentNode = NodeTraversal::next(*currentNode, child.ptr()))
            currentNode->dispatchScopedEvent(MutationEvent::create(eventNames().DOMNodeRemovedFromDocumentEvent, false));
    }
}

PassRefPtr<HTMLCollection> ContainerNode::children()
{
    return ensureRareData().ensureNodeLists()->addCachedCollection<GenericCachedHTMLCollection<CollectionTypeTraits<NodeChildren>::traversalType>>(*this, NodeChildren);
}

Element* ContainerNode::firstElementChild() const
{
    return ElementTraversal::firstChild(this);
}

Element* ContainerNode::lastElementChild() const
{
    return ElementTraversal::lastChild(this);
}

unsigned ContainerNode::childElementCount() const
{
    unsigned count = 0;
    Node* n = firstChild();
    while (n) {
        count += n->isElementNode();
        n = n->nextSibling();
    }
    return count;
}

HTMLCollection* ContainerNode::cachedHTMLCollection(CollectionType type)
{
    return hasRareData() && rareData()->nodeLists() ? rareData()->nodeLists()->cachedCollection<HTMLCollection>(type) : 0;
}

void ContainerNode::append(const Vector<NodeOrString>& nodeOrStringVector, ExceptionCode& ec)
{
    RefPtr<Node> node = convertNodesOrStringsIntoNode(*this, nodeOrStringVector, ec);
    if (ec || !node)
        return;

    appendChild(*node, ec);
}

void ContainerNode::prepend(const Vector<NodeOrString>& nodeOrStringVector, ExceptionCode& ec)
{
    RefPtr<Node> node = convertNodesOrStringsIntoNode(*this, nodeOrStringVector, ec);
    if (ec || !node)
        return;

    insertBefore(*node, firstChild(), ec);
}

// https://dom.spec.whatwg.org/#dom-parentnode-replacechildren
void ContainerNode::replaceChildren(const Vector<NodeOrString>& nodeOrStringVector, ExceptionCode& ec)
{
    // step 1
    RefPtr<Node> node = convertNodesOrStringsIntoNode(*this, nodeOrStringVector, ec);
    if (ec)
		return;

	// step 2
	if (node) {
		auto checkResult = ensurePreInsertionValidity(*node, nullptr, ec);
		if (ec)
			return;
	}

	// step 3
    Ref<ContainerNode> protectedThis(*this);
    ChildListMutationScope mutation(this);
    NodeVector removedChildren;
    removeAllChildrenWithScriptAssertion(ChildChangeSource::API, removedChildren, DeferChildrenChanged::No);

    if (node) {
    	auto appendResult = appendChildWithoutPreInsertionValidityCheck(*node, ec);
        if (ec)
            return;
    }

    rebuildSVGExtensionsElementsIfNecessary();
    dispatchSubtreeModifiedEvent();
}

Element* ContainerNode::querySelector(const String& selectors, ExceptionCode& ec)
{
    if (SelectorQuery* selectorQuery = document().selectorQueryForString(selectors, ec))
        return selectorQuery->queryFirst(*this);
    return nullptr;
}

RefPtr<NodeList> ContainerNode::querySelectorAll(const String& selectors, ExceptionCode& ec)
{
    if (SelectorQuery* selectorQuery = document().selectorQueryForString(selectors, ec))
        return selectorQuery->queryAll(*this);
    return nullptr;
}

PassRefPtr<HTMLCollection> ContainerNode::getElementsByTagName(const AtomicString& qualifiedName)
{
    ASSERT(!qualifiedName.isNull());

    if (qualifiedName == starAtom)
        return ensureRareData().ensureNodeLists()->addCachedCollection<AllDescendantsCollection>(*this, AllDescendants);

    if (document().isHTMLDocument())
        return ensureRareData().ensureNodeLists()->addCachedCollection<HTMLTagCollection>(*this, ByHTMLTag, qualifiedName);
    return ensureRareData().ensureNodeLists()->addCachedCollection<TagCollection>(*this, ByTag, qualifiedName);
}

PassRefPtr<HTMLCollection> ContainerNode::getElementsByTagNameNS(const AtomicString& namespaceURI, const AtomicString& localName)
{
    ASSERT(!localName.isNull());
    return ensureRareData().ensureNodeLists()->addCachedTagCollectionNS(*this, namespaceURI.isEmpty() ? nullAtom : namespaceURI, localName);
}

PassRefPtr<NodeList> ContainerNode::getElementsByName(const AtomicString& elementName)
{
    return ensureRareData().ensureNodeLists()->addCacheWithAtomicName<NameNodeList>(this, elementName);
}

PassRefPtr<HTMLCollection> ContainerNode::getElementsByClassName(const AtomicString& classNames)
{
    return ensureRareData().ensureNodeLists()->addCachedCollection<ClassCollection>(*this, ByClass, classNames);
}

PassRefPtr<RadioNodeList> ContainerNode::radioNodeList(const AtomicString& name)
{
    ASSERT(hasTagName(formTag) || hasTagName(fieldsetTag));
    return ensureRareData().ensureNodeLists()->addCacheWithAtomicName<RadioNodeList>(this, name);
}

} // namespace WebCore
