/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2009, 2010, 2011 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "CollectionType.h"
#include "ExceptionCodePlaceholder.h"
#include "Node.h"

#include <wtf/OwnPtr.h>
#include <wtf/Ref.h>
#include <wtf/Vector.h>

namespace WebCore {

class HTMLCollection;
class FloatPoint;
class NodeOrString;
class RenderElement;

const int initialNodeVectorSize = 11; // Covers 99.5%. See webkit.org/b/80706
typedef Vector<Ref<Node>, initialNodeVectorSize> NodeVector;

class ContainerNode : public Node {
public:
    virtual ~ContainerNode();

    ALWAYS_INLINE Node* firstChild() const { return m_firstChild; }
    ALWAYS_INLINE Node* lastChild() const { return m_lastChild; }
    ALWAYS_INLINE bool hasChildNodes() const { return m_firstChild; }
    bool hasOneChild() const { return m_firstChild && m_firstChild == m_lastChild; }

    bool directChildNeedsStyleRecalc() const { return hasStyleFlag(NodeStyleFlag::DirectChildNeedsStyleResolution); }
    void setDirectChildNeedsStyleRecalc() { setStyleFlag(NodeStyleFlag::DirectChildNeedsStyleResolution); }

    unsigned countChildNodes() const;
    Node* traverseToChildAt(unsigned) const;

    bool insertBefore(Node& newChild, Node* refChild, ExceptionCode& = ASSERT_NO_EXCEPTION);
    bool replaceChild(Node& newChild, Node& oldChild, ExceptionCode& = ASSERT_NO_EXCEPTION);
    bool removeChild(Node& child, ExceptionCode& = ASSERT_NO_EXCEPTION);
    bool appendChild(Node& newChild, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void stringReplaceAll(const String&);
    void replaceAll(Node*);

    // These methods are only used during parsing.
    // They don't send DOM mutation events or handle reparenting.
    // However, arbitrary code may be run by beforeload handlers.
    void parserAppendChild(Node& newChild);
    void parserRemoveChild(Node&);
    void parserInsertBefore(Node& newChild, Node& refChild);

    void removeChildren();

    void takeAllChildrenFrom(ContainerNode*);

    void cloneChildNodes(ContainerNode& clone);

    enum ChildChangeType { ElementInserted, ElementRemoved, TextInserted, TextRemoved, TextChanged, AllChildrenRemoved, NonContentsChildRemoved, NonContentsChildInserted, AllChildrenReplaced };
    enum class ChildChangeSource { Parser, API };
    struct ChildChange {
        ChildChangeType type;
        Element* previousSiblingElement;
        Element* nextSiblingElement;
        ChildChangeSource source;

        bool isInsertion() const
        {
            switch (type) {
            case ElementInserted:
            case TextInserted:
            case NonContentsChildInserted:
            case AllChildrenReplaced:
                return true;
            case ElementRemoved:
            case TextRemoved:
            case TextChanged:
            case AllChildrenRemoved:
            case NonContentsChildRemoved:
                return false;
            }
            ASSERT_NOT_REACHED();
            return false;
        }
    };
    virtual void childrenChanged(const ChildChange&);

    void disconnectDescendantFrames();

    // From the ParentNode interface - https://dom.spec.whatwg.org/#interface-parentnode
    PassRefPtr<HTMLCollection> children();
    Element* firstElementChild() const;
    Element* lastElementChild() const;
    unsigned childElementCount() const;
    void append(const Vector<NodeOrString>&, ExceptionCode&);
    void prepend(const Vector<NodeOrString>&, ExceptionCode&);

    void replaceChildren(const Vector<NodeOrString>&, ExceptionCode&);

    bool ensurePreInsertionValidity(Node& newChild, Node* refChild, ExceptionCode&);

    RenderElement* renderer() const;

    Element* querySelector(const String& selectors, ExceptionCode&);
    RefPtr<NodeList> querySelectorAll(const String& selectors, ExceptionCode&);

    PassRefPtr<HTMLCollection> getElementsByTagName(const AtomicString&);
    PassRefPtr<HTMLCollection> getElementsByTagNameNS(const AtomicString& namespaceURI, const AtomicString& localName);
    PassRefPtr<NodeList> getElementsByName(const AtomicString& elementName);
    PassRefPtr<HTMLCollection> getElementsByClassName(const AtomicString& classNames);
    PassRefPtr<RadioNodeList> radioNodeList(const AtomicString&);

protected:
    ContainerNode(Document*, ConstructionType = CreateContainer);

	friend void removeDetachedChildrenInContainer(ContainerNode& container);

    void removeDetachedChildren();
    void setFirstChild(Node* child) { m_firstChild = child; }
    void setLastChild(Node* child) { m_lastChild = child; }

    HTMLCollection* cachedHTMLCollection(CollectionType);

private:
    void executePreparedChildrenRemoval();
    enum class DeferChildrenChanged { Yes, No };
    void removeAllChildrenWithScriptAssertion(ChildChangeSource, NodeVector& children, DeferChildrenChanged = DeferChildrenChanged::No);
    bool removeNodeWithScriptAssertion(Node&, ChildChangeSource);
    void removeSelfOrChildNodesForInsertion(Node&, NodeVector&, ExceptionCode&);

    void removeBetween(Node* previousChild, Node* nextChild, Node* oldChild);
    bool appendChildWithoutPreInsertionValidityCheck(Node&, ExceptionCode&);
    void insertBeforeCommon(Node& nextChild, Node& oldChild);
    void appendChildCommon(Node&);

    static void dispatchPostAttachCallbacks();
    void suspendPostAttachCallbacks();
    void resumePostAttachCallbacks();

    void rebuildSVGExtensionsElementsIfNecessary();

    Node* m_firstChild { nullptr };
    Node* m_lastChild { nullptr };
};

ALWAYS_INLINE ContainerNode* toContainerNode(Node* node)
{
    ASSERT(!node || node->isContainerNode());
    return static_cast<ContainerNode*>(node);
}

ALWAYS_INLINE const ContainerNode* toContainerNode(const Node* node)
{
    ASSERT(!node || node->isContainerNode());
    return static_cast<const ContainerNode*>(node);
}

ALWAYS_INLINE ContainerNode& toContainerNode(Node& node)
{
    ASSERT(node.isContainerNode());
    return static_cast<ContainerNode&>(node);
}

ALWAYS_INLINE const ContainerNode& toContainerNode(const Node& node)
{
    ASSERT(node.isContainerNode());
    return static_cast<const ContainerNode&>(node);
}

// This will catch anyone doing an unnecessary cast.
void toContainerNode(const ContainerNode*);

ALWAYS_INLINE ContainerNode::ContainerNode(Document* document, ConstructionType type)
    : Node(document, type)
    , m_firstChild(0)
    , m_lastChild(0)
{
}

ALWAYS_INLINE unsigned Node::countChildNodes() const
{
    if (!isContainerNode())
        return 0;
    return toContainerNode(this)->countChildNodes();
}

ALWAYS_INLINE Node* Node::traverseToChildAt(unsigned index) const
{
    if (!isContainerNode())
        return 0;
    return toContainerNode(this)->traverseToChildAt(index);
}

ALWAYS_INLINE Node* Node::firstChild() const
{
    if (!isContainerNode())
        return 0;
    return toContainerNode(this)->firstChild();
}

ALWAYS_INLINE Node* Node::lastChild() const
{
    if (!isContainerNode())
        return 0;
    return toContainerNode(this)->lastChild();
}

inline Node& Node::rootNode() const
{
    if (isInTreeScope())
        return treeScope().rootNode();
    return traverseToRootNode();
}

inline void collectChildNodes(Node& node, NodeVector& children)
{
    for (Node* child = node.firstChild(); child; child = child->nextSibling())
        children.append(*child);
}

} // namespace WebCore
