/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2006 Samuel Weinig (sam@webkit.org)
 * Copyright (C) 2008, 2009 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "DOMImplementation.h"

#include "ContentType.h"
#include "CSSStyleSheet.h"
#include "DocumentType.h"
#include "Element.h"
#include "ExceptionCode.h"
#include "Frame.h"
#include "FrameLoaderClient.h"
#include "FTPDirectoryDocument.h"
#include "HTMLDocument.h"
#include "HTMLHeadElement.h"
#include "HTMLTitleElement.h"
#include "HTMLNames.h"
#include "Image.h"
#include "ImageDocument.h"
#include "MediaDocument.h"
#include "MediaList.h"
#include "MediaPlayer.h"
#include "MIMETypeRegistry.h"
#include "PDFDocument.h"
#include "Page.h"
#include "PluginData.h"
#include "PluginDocument.h"
#include "SecurityOrigin.h"
#include "SecurityOriginPolicy.h"
#include "SegmentedString.h"
#include "Settings.h"
#include "StyleSheetContents.h"
#include "Text.h"
#include "TextDocument.h"
#include "ThreadGlobalData.h"
#include "XMLDocument.h"
#include "XMLNames.h"
#include "DocumentLoader.h"
#include <wtf/StdLibExtras.h>

#if ENABLE(SVG)
#include "SVGNames.h"
#include "SVGDocument.h"
#endif

namespace WebCore {

using namespace HTMLNames;

typedef HashSet<String, ASCIICaseInsensitiveHash> FeatureSet;

static void addString(FeatureSet& set, const String& string)
{
    set.add(string);
}

#if ENABLE(VIDEO)
class DOMImplementationSupportsTypeClient : public MediaPlayerSupportsTypeClient {
public:
    DOMImplementationSupportsTypeClient(bool needsHacks, const String& host)
        : m_needsHacks(needsHacks)
        , m_host(host)
    {
    }

private:
    virtual bool mediaPlayerNeedsSiteSpecificHacks() const OVERRIDE { return m_needsHacks; }
    virtual String mediaPlayerDocumentHost() const OVERRIDE { return m_host; }

    bool m_needsHacks;
    String m_host;
};
#endif

#if ENABLE(SVG)

static bool isSupportedSVG10Feature(const String& feature, const String& version)
{
    if (!version.isEmpty() && version != "1.0")
        return false;

    static NeverDestroyed<FeatureSet> svgFeatures;
    if (svgFeatures->isEmpty()) {
#if ENABLE(FILTERS) && ENABLE(SVG_FONTS)
        addString(svgFeatures.get(), "svg");
        addString(svgFeatures.get(), "svg.static");
#endif
//      addString(svgFeatures.get(), "svg.animation");
//      addString(svgFeatures.get(), "svg.dynamic");
//      addString(svgFeatures.get(), "svg.dom.animation");
//      addString(svgFeatures.get(), "svg.dom.dynamic");
#if ENABLE(FILTERS) && ENABLE(SVG_FONTS)
        addString(svgFeatures.get(), "dom");
        addString(svgFeatures.get(), "dom.svg");
        addString(svgFeatures.get(), "dom.svg.static");
#endif
//      addString(svgFeatures.get(), "svg.all");
//      addString(svgFeatures.get(), "dom.svg.all");
    }
    return feature.startsWith("org.w3c.", false)
        && svgFeatures->contains(feature.right(feature.length() - 8));
}

static bool isSupportedSVG11Feature(const String& feature, const String& version)
{
    if (!version.isEmpty() && version != "1.1")
        return false;

    static NeverDestroyed<FeatureSet> svgFeatures;
    if (svgFeatures->isEmpty()) {
        // Sadly, we cannot claim to implement any of the SVG 1.1 generic feature sets
        // lack of Font and Filter support.
        // http://bugs.webkit.org/show_bug.cgi?id=15480
#if ENABLE(FILTERS) && ENABLE(SVG_FONTS)
        addString(svgFeatures.get(), "SVG");
        addString(svgFeatures.get(), "SVGDOM");
        addString(svgFeatures.get(), "SVG-static");
        addString(svgFeatures.get(), "SVGDOM-static");
#endif
        addString(svgFeatures.get(), "SVG-animation");
        addString(svgFeatures.get(), "SVGDOM-animation");
//      addString(svgFeatures.get(), "SVG-dynamic);
//      addString(svgFeatures.get(), "SVGDOM-dynamic);
        addString(svgFeatures.get(), "CoreAttribute");
        addString(svgFeatures.get(), "Structure");
        addString(svgFeatures.get(), "BasicStructure");
        addString(svgFeatures.get(), "ContainerAttribute");
        addString(svgFeatures.get(), "ConditionalProcessing");
        addString(svgFeatures.get(), "Image");
        addString(svgFeatures.get(), "Style");
        addString(svgFeatures.get(), "ViewportAttribute");
        addString(svgFeatures.get(), "Shape");
        addString(svgFeatures.get(), "Text");
        addString(svgFeatures.get(), "BasicText");
        addString(svgFeatures.get(), "PaintAttribute");
        addString(svgFeatures.get(), "BasicPaintAttribute");
        addString(svgFeatures.get(), "OpacityAttribute");
        addString(svgFeatures.get(), "GraphicsAttribute");
        addString(svgFeatures.get(), "BaseGraphicsAttribute");
        addString(svgFeatures.get(), "Marker");
//      addString(svgFeatures.get(), "ColorProfile"); // requires color-profile, bug 6037
        addString(svgFeatures.get(), "Gradient");
        addString(svgFeatures.get(), "Pattern");
        addString(svgFeatures.get(), "Clip");
        addString(svgFeatures.get(), "BasicClip");
        addString(svgFeatures.get(), "Mask");
#if ENABLE(FILTERS)
        addString(svgFeatures.get(), "Filter");
        addString(svgFeatures.get(), "BasicFilter");
#endif
        addString(svgFeatures.get(), "DocumentEventsAttribute");
        addString(svgFeatures.get(), "GraphicalEventsAttribute");
//      addString(svgFeatures.get(), "AnimationEventsAttribute");
        addString(svgFeatures.get(), "Cursor");
        addString(svgFeatures.get(), "Hyperlinking");
        addString(svgFeatures.get(), "XlinkAttribute");
        addString(svgFeatures.get(), "ExternalResourcesRequired");
        addString(svgFeatures.get(), "View");
        addString(svgFeatures.get(), "Script");
        addString(svgFeatures.get(), "Animation");
#if ENABLE(SVG_FONTS)
        addString(svgFeatures.get(), "Font");
        addString(svgFeatures.get(), "BasicFont");
#endif
        addString(svgFeatures.get(), "Extensibility");
    }
    return feature.startsWith("http://www.w3.org/tr/svg11/feature#", false)
        && svgFeatures->contains(feature.right(feature.length() - 35));
}
#endif

DOMImplementation::DOMImplementation(Document* document)
    : m_document(document)
{
}

bool DOMImplementation::hasFeature(const String& feature, const String& version)
{
    if (feature.startsWith("http://www.w3.org/TR/SVG", false)
        || feature.startsWith("org.w3c.dom.svg", false)
        || feature.startsWith("org.w3c.svg", false)) {
#if ENABLE(SVG)
        // FIXME: SVG 2.0 support?
        return isSupportedSVG10Feature(feature, version) || isSupportedSVG11Feature(feature, version);
#else
        UNUSED_PARAM(version);
        return false;
#endif
    }

    // FIXME: SVG specifications <http://www.w3.org/TR/SVG/script.html#InterfaceSVGZoomEvent>
    // and <http://www.w3.org/TR/SVG2/interact.html#InterfaceSVGZoomEvent>
    // say that we should return true for the feature "SVGZoomEvents".

    return true;
}

PassRefPtr<DocumentType> DOMImplementation::createDocumentType(const String& qualifiedName,
    const String& publicId, const String& systemId, ExceptionCode& ec)
{
    String prefix, localName;
    if (!Document::parseQualifiedName(qualifiedName, prefix, localName, ec))
        return nullptr;

    return DocumentType::create(m_document, qualifiedName, publicId, systemId);
}

DOMImplementation* DOMImplementation::getInterface(const String& /*feature*/)
{
    return 0;
}

PassRefPtr<XMLDocument> DOMImplementation::createDocument(const String& namespaceURI,
    const String& qualifiedName, DocumentType* doctype, ExceptionCode& ec)
{
    RefPtr<XMLDocument> doc;
#if ENABLE(SVG)
    if (namespaceURI == SVGNames::svgNamespaceURI)
        doc = SVGDocument::create(nullptr, KURL());
    else
#endif
    if (namespaceURI == HTMLNames::xhtmlNamespaceURI)
        doc = XMLDocument::createXHTML(nullptr, KURL());
    else
        doc = XMLDocument::create(nullptr, KURL());

    doc->setContextDocument(m_document->contextDocument());
    doc->setSecurityOriginPolicy(m_document->securityOriginPolicy());

    RefPtr<Node> documentElement;
    if (!qualifiedName.isEmpty()) {
        documentElement = doc->createElementNS(namespaceURI, qualifiedName, ec);
        if (ec)
            return nullptr;
    }

    if (doctype)
        doc->appendChild(*doctype);
    if (documentElement)
        doc->appendChild(*documentElement);

    return doc.release();
}

PassRefPtr<CSSStyleSheet> DOMImplementation::createCSSStyleSheet(const String&, const String& media, ExceptionCode&)
{
    // FIXME: Title should be set.
    // FIXME: Media could have wrong syntax, in which case we should generate an exception.
    RefPtr<CSSStyleSheet> sheet = CSSStyleSheet::create(StyleSheetContents::create());
    sheet->setMediaQueries(MediaQuerySet::create(media));
    return sheet;
}

PassRefPtr<HTMLDocument> DOMImplementation::createHTMLDocument(const String& title)
{
    RefPtr<HTMLDocument> document = HTMLDocument::create(0, KURL());
    document->open();
    document->write(0, SegmentedString("<!doctype html><html><head></head><body></body></html>"));
    if (!title.isNull()) {
    	Ref<HTMLTitleElement> titleElement = HTMLTitleElement::create(titleTag, document.get());
		titleElement->appendChild(document->createTextNode(title));
		ASSERT(document->head());
		document->head()->appendChild(titleElement);
	}
    document->setContextDocument(m_document->contextDocument());
    document->setSecurityOriginPolicy(m_document->securityOriginPolicy());
    return document.release();
}

PassRefPtr<Document> DOMImplementation::createDocument(const String& type, Frame* frame, const KURL& url)
{
    // Plugins cannot take HTML and XHTML from us, and we don't even need to initialize the plugin database for those.
    if (type == ASCIILiteral("text/html"))
        return HTMLDocument::create(frame, url);
    if (type == ASCIILiteral("application/xhtml+xml"))
        return XMLDocument::createXHTML(frame, url);

#if ENABLE(FTPDIR)
    // Plugins cannot take FTP from us either
    if (type == ASCIILiteral("application/x-ftp-directory"))
        return FTPDirectoryDocument::create(frame, url);
#endif

#if HAVE(PDFIUM)
    if (frame && MIMETypeRegistry::isPDFMIMEType(type))
        return PDFDocument::create(*frame, url);
#endif

    // If we want to useImageDocumentForSubframePDF, we'll let that override plugin support.
    if (frame && !frame->isMainFrame() && MIMETypeRegistry::isPDFMIMEType(type) && frame->settings().useImageDocumentForSubframePDF())
        return ImageDocument::create(*frame, url);

    PluginData* pluginData = nullptr;
    if (frame && frame->page() && frame->loader()->subframeLoader()->allowPlugins(NotAboutToInstantiatePlugin))
        pluginData = frame->page()->pluginData();

    // PDF is one image type for which a plugin can override built-in support.
    // We do not want QuickTime to take over all image types, obviously.
    if (MIMETypeRegistry::isPDFOrPostScriptMIMEType(type) && pluginData && pluginData->supportsMimeType(type))
        return PluginDocument::create(frame, url);
    if (Image::supportsType(type))
        return ImageDocument::create(*frame, url);

#if ENABLE(VIDEO)
    {
		// Check to see if the type can be played by our MediaPlayer, if so create a MediaDocument
		// Key system is not applicable here.
		DOMImplementationSupportsTypeClient client(frame && frame->settings().needsSiteSpecificQuirks(), url.host().toString());
		MediaEngineSupportParameters parameters;
		ContentType contentType(type);
		parameters.type = contentType.type().lower();
		parameters.codecs = contentType.parameter(ASCIILiteral("codecs"));
		parameters.url = url;
		if(MediaPlayer::supportsType(parameters, &client))
			 return MediaDocument::create(frame, url);
    }
#endif

    // Everything else except text/plain can be overridden by plugins. In particular, Adobe SVG Viewer should be used for SVG, if installed.
    // Disallowing plug-ins to use text/plain prevents plug-ins from hijacking a fundamental type that the browser is expected to handle,
    // and also serves as an optimization to prevent loading the plug-in database in the common case.
    if (type != ASCIILiteral("text/plain") && pluginData && pluginData->supportsMimeType(type))
        return PluginDocument::create(frame, url);
    if (MIMETypeRegistry::isTextMIMEType(type))
        return TextDocument::create(frame, url);

#if ENABLE(SVG)
    if (type == ASCIILiteral("image/svg+xml"))
        return SVGDocument::create(frame, url);
#endif

#if ENABLE(CEHTML)
    if (MIMETypeRegistry::isCEHTMLMIMEType(type)){
    	RefPtr<Document> document = HTMLDocument::create(frame, url);
    	document->setCEHTMLDocument(true);
    	return document.release();
    }
#endif

    if (MIMETypeRegistry::isXMLMIMEType(type))
        return XMLDocument::create(frame, url);

    return HTMLDocument::create(frame, url);
}

}
