/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 *           (C) 2006 Alexey Proskuryakov (ap@webkit.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2012 Apple Inc. All rights reserved.
 * Copyright (C) 2008, 2009 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies)
 * Copyright (C) 2011 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "CSSRegisteredCustomProperty.h"
#include "CollectionType.h"
#include "Color.h"
#include "ContainerNode.h"
#include "DOMHighResTimeStamp.h"
#include "DOMTimeStamp.h"
#include "DocumentIdentifier.h"
#include "DocumentTimelinesController.h"
#include "DocumentTiming.h"
#include "FocusDirection.h"
#include "FocusOptions.h"
#include "FontSelector.h"
#include "FrameDestructionObserver.h"
#include "HitTestRequest.h"
#include "InspectorCounters.h"
#include "IntRect.h"
#include "LayoutPoint.h"
#include "MutationObserver.h"
#include "PageVisibilityState.h"
#include "PlatformScreen.h"
#include "QualifiedName.h"
#include "ReducedResolutionSeconds.h"
#include "ReferrerPolicy.h"
#include "RenderPtr.h"
#include "ScriptExecutionContext.h"
#include "StringWithDirection.h"
#include "StyleUpdate.h"
#include "TextResourceDecoder.h"
#include "Timer.h"
#include "TreeScope.h"
#include "UserActionElementSet.h"
#include "ViewportArguments.h"
#include <wtf/Deque.h>
#include <wtf/FixedArray.h>
#include <wtf/HashMap.h>
#include <wtf/HashSet.h>
#include <wtf/ObjectIdentifier.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/PassRefPtr.h>
#include <wtf/Seconds.h>
#include <wtf/WeakHashSet.h>
#include <wtf/WeakListHashSet.h>

namespace WebCore {

#if USE(REQUEST_ANIMATION_FRAME_DISPLAY_MONITOR)
class DocumentAnimationScheduler;
#endif
class EventLoop;
class AXObjectCache;
class Attr;
class CDATASection;
class CSSCustomPropertyValue;
class CSSFontSelector;
class CSSStyleDeclaration;
class CSSStyleSheet;
class CachedCSSStyleSheet;
class CachedFrameBase;
class CachedResourceLoader;
class CachedScript;
class CanvasRenderingContext;
class CharacterData;
class Comment;
class ConstantPropertyMap;
class DOMImplementation;
class DOMSelection;
class DOMWindow;
class Database;
class DatabaseThread;
class DocumentFragment;
class DocumentLoader;
class DocumentMarkerController;
class DocumentParser;
class DocumentSharedObjectPool;
class DocumentTimeline;
class DocumentType;
class Editor;
class EntityReference;
class EventLoopTaskGroup;
class EventNames;
class ExtensionStyleSheets;
class FloatRect;
class FloatQuad;
class FormController;
class Frame;
class FrameSelection;
class FrameView;
class HTMLAllCollection;
class HTMLBodyElement;
class HTMLCanvasElement;
class HTMLCollection;
class HTMLDialogElement;
class HTMLDocument;
class HTMLElement;
class HTMLFrameOwnerElement;
class HTMLHeadElement;
class HTMLIFrameElement;
class HTMLImageElement;
class HTMLMapElement;
class HTMLMediaElement;
class HTMLScriptElement;
class HitTestLocation;
class HitTestRequest;
class HitTestResult;
class IdleCallbackController;
class IdleRequestCallback;
class ImageBitmapRenderingContext;
class IntPoint;
class DOMWrapperWorld;
class ScriptModuleLoader;
class JSNode;
class Locale;
class LiveNodeList;
class Location;
class MediaCanStartListener;
class MediaQueryList;
class MediaQueryMatcher;
class ScriptModuleLoader;
class MouseEventWithHitTestResults;
class NodeFilter;
class NodeIterator;
class Page;
class PlatformMouseEvent;
class ProcessingInstruction;
class Range;
class RenderView;
class RenderFullScreen;
class RenderTreeBuilder;
class ScriptableDocumentParser;
class ScriptElementData;
class ScriptRunner;
class SecurityOrigin;
class SelectorQuery;
class SelectorQueryCache;
class SerializedScriptValue;
class SegmentedString;
class Settings;
class StyleResolver;
class StyleSheet;
class StyleSheetContents;
class StyleSheetList;
class SVGDocumentExtensions;
class SVGElement;
class SVGUseElement;
class SVGSVGElement;
class Text;
class TextResourceDecoder;
class TreeWalker;
class VisitedLinkState;
class WebAnimation;
class WindowEventLoop;
class UndoManager;
class XMLHttpRequest;
class XPathEvaluator;
class XPathExpression;
class XPathNSResolver;
class XPathResult;

#if ENABLE(XSLT)
class TransformSource;
#endif

#if ENABLE(DASHBOARD_SUPPORT)
struct DashboardRegionValue;
#endif

#if ENABLE(TOUCH_EVENTS)
class Touch;
class TouchList;
#endif

#if ENABLE(REQUEST_ANIMATION_FRAME)
class RequestAnimationFrameCallback;
class ScriptedAnimationController;
#endif

#if ENABLE(MICRODATA)
class MicroDataItemList;
#endif

#if ENABLE(LINK_PRERENDER)
class Prerenderer;
#endif

#if ENABLE(TEXT_AUTOSIZING)
class TextAutosizer;
#endif

class GraphicsContext;
class StyleImage;
class StyledElement;

#if ENABLE(BML)
class HTMLBeitemElement;
#endif
    
#if ENABLE(RMAI)
class GalioWafer;
#endif

class FontFaceSet;

struct BoundaryPoint;
struct IntersectionObserverData;

typedef int ExceptionCode;

#if ENABLE(INTERSECTION_OBSERVER)
class IntersectionObserver;
struct IntersectionObserverData;
#endif

#if ENABLE(RESIZE_OBSERVER)
class ResizeObserver;
#endif

namespace Style {
class CustomPropertyRegistry;
class Scope;
};

enum PageshowEventPersistence {
    PageshowEventNotPersisted = 0,
    PageshowEventPersisted = 1
};

enum NodeListInvalidationType {
    DoNotInvalidateOnAttributeChanges = 0,
    InvalidateOnClassAttrChange,
    InvalidateOnIdNameAttrChange,
    InvalidateOnNameAttrChange,
	InvalidateOnForTypeAttrChange,
    InvalidateForFormControls,
    InvalidateOnHRefAttrChange,
    InvalidateOnItemAttrChange,
    InvalidateOnAnyAttrChange,
};
const int numNodeListInvalidationTypes = InvalidateOnAnyAttrChange + 1;

enum class HttpEquivPolicy {
    Enabled,
    DisabledBySettings,
    DisabledByContentDispositionAttachmentSandbox
};

enum DimensionsCheck { WidthDimensionsCheck = 1 << 0, HeightDimensionsCheck = 1 << 1, AllDimensionsCheck = WidthDimensionsCheck | HeightDimensionsCheck };

enum DocumentClass : uint16_t {
    DefaultDocumentClass = 0,
    HTMLDocumentClass = 1,
    XHTMLDocumentClass = 1 << 1,
    ImageDocumentClass = 1 << 2,
    PluginDocumentClass = 1 << 3,
    MediaDocumentClass = 1 << 4,
    SVGDocumentClass = 1 << 5,
    TextDocumentClass = 1 << 6,
    XMLDocumentClass = 1 << 7,
    NetTVDocumentClass = 1 << 8,
    PDFDocumentClass = 1 << 9,
};

typedef unsigned short DocumentClassFlags;

class Document
	: public ContainerNode
	, public TreeScope
	, public ScriptExecutionContext
	, public FontSelectorClient
	, public FrameDestructionObserver {
public:
    static Ref<Document> create(Frame* frame, const KURL& url);
    static Ref<Document> createNonRenderedPlaceholder(Frame* frame, const KURL& url);
    static Ref<Document> create(Document*);

    virtual ~Document();

    // Nodes belonging to this document increase referencingNodeCount -
    // these are enough to keep the document from being destroyed, but
    // not enough to keep it from removing its children. This allows a
    // node that outlives its document to still have a valid document
    // pointer without introducing reference cycles.
    void incrementReferencingNodeCount()
    {
        ASSERT(!m_deletionHasBegun);
        ++m_referencingNodeCount;
    }

    unsigned referencingNodeCount() const { return m_referencingNodeCount; }

    void decrementReferencingNodeCount()
    {
        ASSERT(!m_deletionHasBegun || !m_referencingNodeCount);
        --m_referencingNodeCount;
        if (!m_referencingNodeCount && !refCount()) {
#if !ASSERT_DISABLED
            m_deletionHasBegun = true;
#endif
            m_refCountAndParentBit = s_refCountIncrement; // Avoid double destruction through use of Ref<T>/RefPtr<T>. (This is a security mitigation in case of programmer error. It will ASSERT in debug builds.)
			delete this;
        }
    }

    void removedLastRef();

    const DocumentIdentifier& identifier() const { return m_identifier; }

    typedef HashMap<DocumentIdentifier, Document*> DocumentsMap;
    static DocumentsMap& allDocumentsMap();

    MediaQueryMatcher* mediaQueryMatcher();

    using ContainerNode::ref;
    using ContainerNode::deref;
    using TreeScope::rootNode;

    virtual bool canContainRangeEndPoint() const FINAL { return true; }

    Element* getElementByAccessKey(const String& key);
    void invalidateAccessKeyMap();

    SelectorQuery* selectorQueryForString(const String&, ExceptionCode&);
    void clearSelectorQueryCache();

    ViewportArguments viewportArguments() const { return m_viewportArguments; }
#ifndef NDEBUG
    bool didDispatchViewportPropertiesChanged() const { return m_didDispatchViewportPropertiesChanged; }
#endif

    void setReferrerPolicy(ReferrerPolicy);
    ReferrerPolicy referrerPolicy() const { return m_referrerPolicy.value_or(ReferrerPolicy::NoReferrerWhenDowngrade); }

    DocumentType* doctype() const;

    DOMImplementation* implementation();
    
    ALWAYS_INLINE Element* documentElement() const
    {
        return m_documentElement.get();
    }

    bool hasManifest() const;
    
    PassRefPtr<Element> createElement(const AtomicString& tagName, ExceptionCode&);
    Ref<DocumentFragment> createDocumentFragment();
    Ref<Text> createTextNode(const String& data);
    Ref<Comment> createComment(const String& data);
    PassRefPtr<CDATASection> createCDATASection(const String& data, ExceptionCode&);
    PassRefPtr<ProcessingInstruction> createProcessingInstruction(const String& target, const String& data, ExceptionCode&);
    PassRefPtr<Attr> createAttribute(const String& name, ExceptionCode&);
    PassRefPtr<Attr> createAttributeNS(const String& namespaceURI, const String& qualifiedName, ExceptionCode&, bool shouldIgnoreNamespaceChecks = false);
    PassRefPtr<EntityReference> createEntityReference(const String& name, ExceptionCode&);
    PassRefPtr<Node> importNode(Node& nodeToImport, bool deep, ExceptionCode&);
    PassRefPtr<Element> createElementNS(const String& namespaceURI, const String& qualifiedName, ExceptionCode&);
    Ref<Element> createElement(const QualifiedName&, bool createdByParser);

    bool cssRegionsEnabled() const;

    PassRefPtr<Range> caretRangeFromPoint(int x, int y);
    Optional<BoundaryPoint> caretPositionFromPoint(const LayoutPoint& clientPoint);

    WEBCORE_EXPORT Element* scrollingElementForAPI();
    Element* scrollingElement();

    String readyState() const;

    String defaultCharsetForBindings() const;

    String charset() const { return Document::encoding(); }
    String characterSetWithUTF8Fallback() const;
    TextEncoding textEncoding() const;

    AtomString encoding() const
    {
        auto encoding = textEncoding().domName();
        return encoding.isNull() ? nullAtom : AtomString(encoding);
    }

    void setCharset(const String&);

    void setContent(const String&);

    String suggestedMIMEType() const;

    void overrideMIMEType(const String&);
    String contentType() const;

    ALWAYS_INLINE const String& contentLanguage() const { return m_contentLanguage; }
    void setContentLanguage(const String&);

    ALWAYS_INLINE const String& xmlEncoding() const { return m_xmlEncoding; }
    ALWAYS_INLINE const String& xmlVersion() const { return m_xmlVersion; }
    enum StandaloneStatus { StandaloneUnspecified, Standalone, NotStandalone };
    ALWAYS_INLINE bool xmlStandalone() const { return m_xmlStandalone == Standalone; }
    ALWAYS_INLINE StandaloneStatus xmlStandaloneStatus() const { return static_cast<StandaloneStatus>(m_xmlStandalone); }
    ALWAYS_INLINE bool hasXMLDeclaration() const { return m_hasXMLDeclaration; }

    ALWAYS_INLINE void setXMLEncoding(const String& encoding) { m_xmlEncoding = encoding; } // read-only property, only to be set from XMLDocumentParser
    void setXMLVersion(const String&, ExceptionCode&);
    void setXMLStandalone(bool, ExceptionCode&);
    ALWAYS_INLINE void setHasXMLDeclaration(bool hasXMLDeclaration) { m_hasXMLDeclaration = hasXMLDeclaration ? 1 : 0; }

    ALWAYS_INLINE const String& documentURI() const { return m_documentURI; }
    void setDocumentURI(const String&);

    virtual KURL baseURI() const;

#if ENABLE(PAGE_VISIBILITY_API)
    void visibilityStateChanged();
    String visibilityState() const;
    bool hidden() const;
    void dispatchVisibilityStateChangeEvent();
#endif

    PassRefPtr<Node> adoptNode(Node& source, ExceptionCode&);

    PassRefPtr<HTMLCollection> images();
    PassRefPtr<HTMLCollection> embeds();
    PassRefPtr<HTMLCollection> plugins(); // an alias for embeds() required for the JS DOM bindings.
    PassRefPtr<HTMLCollection> applets();
    PassRefPtr<HTMLCollection> links();
    PassRefPtr<HTMLCollection> forms();
    PassRefPtr<HTMLCollection> anchors();
    PassRefPtr<HTMLCollection> scripts();
    PassRefPtr<HTMLCollection> all();
    PassRefPtr<HTMLCollection> allFilteredByName(const AtomicString&);

    PassRefPtr<HTMLCollection> windowNamedItems(const AtomicString&);
    PassRefPtr<HTMLCollection> documentNamedItems(const AtomicString&);

    // Other methods (not part of DOM)
    bool isSynthesized() const { return m_isSynthesized; }
    bool isHTMLDocument() const { return m_documentClasses & HTMLDocumentClass; }
    bool isXHTMLDocument() const { return m_documentClasses & XHTMLDocumentClass; }
    bool isXMLDocument() const { return m_documentClasses & XMLDocumentClass; }
    bool isImageDocument() const { return m_documentClasses & ImageDocumentClass; }
    bool isSVGDocument() const { return m_documentClasses & SVGDocumentClass; }
    bool isPluginDocument() const { return m_documentClasses & PluginDocumentClass; }
    bool isMediaDocument() const { return m_documentClasses & MediaDocumentClass; }
    bool isTextDocument() const { return m_documentClasses & TextDocumentClass; }
    bool isNetTVDocument() const { return m_documentClasses & NetTVDocumentClass; }
    bool isPDFDocument() const { return m_documentClasses & PDFDocumentClass; }

    ALWAYS_INLINE bool isCEHTMLDocument() const { return m_isCEHTML; }
    ALWAYS_INLINE void setCEHTMLDocument(bool isCEHTML) { m_isCEHTML = isCEHTML; }

    ALWAYS_INLINE bool isTTMLDocument() const { return m_isTTML; }
    ALWAYS_INLINE void setTTMLDocument(bool isTTML) { m_isTTML = isTTML; }

    ALWAYS_INLINE bool isDirectGLMode() const { return m_isDirectGLMode; }
    ALWAYS_INLINE void setDirectGLMode(bool isDirectGLMode) { m_isDirectGLMode = isDirectGLMode; }

#if ENABLE(SVG)
    bool hasSVGRootNode() const;
#else
    static bool hasSVGRootNode() { return false; }
#endif
    virtual bool isFrameSet() const { return false; }

    ALWAYS_INLINE bool isSrcdocDocument() const { return m_isSrcdocDocument; }

    ALWAYS_INLINE bool sawElementsInKnownNamespaces() const { return m_sawElementsInKnownNamespaces; }

    StyleResolver& userAgentShadowTreeStyleResolver();

    CSSFontSelector& fontSelector() { return *m_fontSelector; }
    const CSSFontSelector& fontSelector() const { return *m_fontSelector; }

    bool haveStylesheetsLoaded() const;
    bool isIgnoringPendingStylesheets() const { return m_ignorePendingStylesheets; }

	// This is a DOM function.
	StyleSheetList* styleSheets();

	Style::Scope& styleScope() { return *m_styleScope; }
    const Style::Scope& styleScope() const { return *m_styleScope; }
    ExtensionStyleSheets& extensionStyleSheets() { return *m_extensionStyleSheets; }
    const ExtensionStyleSheets& extensionStyleSheets() const { return *m_extensionStyleSheets; }

    const Style::CustomPropertyRegistry& customPropertyRegistry() const;

    ALWAYS_INLINE bool gotoAnchorNeededAfterStylesheetsLoad() { return m_gotoAnchorNeededAfterStylesheetsLoad; }
    ALWAYS_INLINE void setGotoAnchorNeededAfterStylesheetsLoad(bool b) { m_gotoAnchorNeededAfterStylesheetsLoad = b; }

    void updateElementsAffectedByMediaQueries();
    void evaluateMediaQueriesAndReportChanges();

    FormController& formController();
    Vector<String> formElementsState() const;
    void setStateForNewFormElements(const Vector<String>&);

    FrameView* view() const; // can be NULL
    Page* page() const; // can be NULL
    const Settings& settings() const { return *m_settings; }

    float deviceScaleFactor() const { return m_deviceScaleFactor; }
    void deviceOrPageScaleFactorChanged();

    Ref<Range> createRange();

    PassRefPtr<NodeIterator> createNodeIterator(Node& root, unsigned long whatToShow, PassRefPtr<NodeFilter>);

    PassRefPtr<TreeWalker> createTreeWalker(Node& root, unsigned long whatToShow, PassRefPtr<NodeFilter>);

    // Special support for editing
    Ref<CSSStyleDeclaration> createCSSStyleDeclaration();
    Ref<Text> createEditingTextNode(const String&);

    enum class ResolveStyleType { Normal, Rebuild };
    void resolveStyle(ResolveStyleType = ResolveStyleType::Normal);
    bool childNeedsAndNotInStyleRecalc();
    bool updateStyleIfNeeded();
    bool needsStyleRecalc() const;
    unsigned lastStyleUpdateSizeForTesting() const { return m_lastStyleUpdateSizeForTesting; }

    void updateLayout();
    enum RunPostLayoutTasks {
        RunPostLayoutTasksAsyhnchronously,
        RunPostLayoutTasksSynchronously,
    };
    void updateLayoutIgnorePendingStylesheets(RunPostLayoutTasks = RunPostLayoutTasksAsyhnchronously);
    PassOwnPtr<RenderStyle> styleForElementIgnoringPendingStylesheets(Element&, const RenderStyle* parentStyle, PseudoId = PseudoId::None);
    PassOwnPtr<RenderStyle> styleForPage(int pageIndex);

    // Returns true if page box (margin boxes and page borders) is visible.
    bool isPageBoxVisible(int pageIndex);

    // Returns the preferred page size and margins in pixels, assuming 96
    // pixels per inch. pageSize, marginTop, marginRight, marginBottom,
    // marginLeft must be initialized to the default values that are used if
    // auto is specified.
    void pageSizeAndMarginsInPixels(int pageIndex, IntSize& pageSize, int& marginTop, int& marginRight, int& marginBottom, int& marginLeft);

    ALWAYS_INLINE CachedResourceLoader* cachedResourceLoader() { return m_cachedResourceLoader.get(); }

    void didBecomeCurrentDocumentInFrame();
    void destroyRenderTree();
    void willBeRemovedFromFrame();

    // Override ScriptExecutionContext methods to do additional work
    virtual bool shouldBypassMainWorldContentSecurityPolicy() const FINAL;
    virtual bool shouldCacheXHRResource() const FINAL;
    virtual void suspendActiveDOMObjects(ActiveDOMObject::ReasonForSuspension) FINAL;
    virtual void resumeActiveDOMObjects(ActiveDOMObject::ReasonForSuspension) FINAL;
    virtual void stopActiveDOMObjects() final;

    RenderView* renderView() const { return m_renderView.get(); }
    const RenderStyle* initialContainingBlockStyle() const { return m_initialContainingBlockStyle.get(); } // This may end up differing from renderView()->style() due to adjustments.

    bool renderTreeBeingDestroyed() const { return m_renderTreeBeingDestroyed; }
    bool hasLivingRenderTree() const { return renderView() && !renderTreeBeingDestroyed(); }
    void updateRenderTree(PassOwnPtr<Style::Update> styleUpdate);

    bool updateLayoutIfDimensionsOutOfDate(Element&, DimensionsCheck = AllDimensionsCheck);

    void clearAXObjectCache();
    AXObjectCache* axObjectCache() const;
    bool axObjectCacheExists() const;
    
    // to get visually ordered hebrew and arabic pages right
    void setVisuallyOrdered();
    ALWAYS_INLINE bool visuallyOrdered() const { return m_visuallyOrdered; }
    
    DocumentLoader* loader() const;

    PassRefPtr<DOMWindow> openForBindings(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& url, const AtomicString& name, const String& features, ExceptionCode&);
    PassRefPtr<Document> openForBindings(Document* responsibleDocument, const String& type, const String& replace, ExceptionCode&);

    // FIXME: We should rename this at some point and give back the name 'open' to the HTML specified ones.
    void open(Document* responsibleDocument = 0);
    void implicitOpen();

    void closeForBindings(ExceptionCode&);

    // FIXME: We should rename this at some point and give back the name 'close' to the HTML specified one.
    void close();
    // In some situations (see the code), we ignore document.close().
    // explicitClose() bypass these checks and actually tries to close the
    // input stream.
    void explicitClose();
    // implicitClose() actually does the work of closing the input stream.
    void implicitClose();

    void cancelParsing();

    void write(Document* responsibleDocument, const SegmentedString&);
    void write(Document* responsibleDocument, const Vector<String>&, ExceptionCode&);
    void writeln(Document* responsibleDocument, const Vector<String>&, ExceptionCode&);

    ALWAYS_INLINE bool wellFormed() const { return m_wellFormed; }

    virtual const KURL& url() const final { return m_url; }
    void setURL(const KURL&);
    const KURL& urlForBindings() const { return m_url.isEmpty() ? blankURL() : m_url; }

    const KURL& creationURL() const { return m_creationURL; }

    // To understand how these concepts relate to one another, please see the
    // comments surrounding their declaration.
    ALWAYS_INLINE const KURL& baseURL() const { return m_baseURL; }
    void setBaseURLOverride(const KURL&);
    ALWAYS_INLINE const KURL& baseURLOverride() const { return m_baseURLOverride; }
    ALWAYS_INLINE const KURL& baseElementURL() const { return m_baseElementURL; }
    ALWAYS_INLINE const String& baseTarget() const { return m_baseTarget; }
    void processBaseElement();

    WEBCORE_EXPORT KURL completeURL(const String&, ForceUTF8 = ForceUTF8::No) const final;
    KURL completeURL(const String&, const KURL& baseURLOverride, ForceUTF8 = ForceUTF8::No) const;

    virtual String userAgent(const KURL&) const FINAL;

    virtual void disableEval(const String& errorMessage) FINAL;

    bool canNavigate(Frame* targetFrame, const KURL& destinationURL = KURL());
    Frame* findUnsafeParentScrollPropagationBoundary();

    bool usesStyleBasedEditability() const;
    void setHasElementUsingStyleBasedEditability();
    
    virtual PassRefPtr<DocumentParser> createParser();
    ALWAYS_INLINE DocumentParser* parser() const { return m_parser.get(); }
    ScriptableDocumentParser* scriptableDocumentParser() const;
    
    ALWAYS_INLINE bool printing() const {
#if ENABLE(PRINTING)
    	return m_printing;
#else
    	return false;
#endif
    }
    ALWAYS_INLINE void setPrinting(bool p) { m_printing = p; }

    ALWAYS_INLINE bool paginatedForScreen() const { return m_paginatedForScreen; }
    ALWAYS_INLINE void setPaginatedForScreen(bool p) { m_paginatedForScreen = p; }
    
    ALWAYS_INLINE bool paginated() const { return printing() || paginatedForScreen(); }

    enum CompatibilityMode { QuirksMode, LimitedQuirksMode, NoQuirksMode };

    void setCompatibilityMode(CompatibilityMode m);
    ALWAYS_INLINE void lockCompatibilityMode() { m_compatibilityModeLocked = true; }
    ALWAYS_INLINE CompatibilityMode compatibilityMode() const { return m_compatibilityMode; }

    String compatMode() const;

    ALWAYS_INLINE bool inQuirksMode() const { return m_compatibilityMode == QuirksMode; }
    ALWAYS_INLINE bool inLimitedQuirksMode() const { return m_compatibilityMode == LimitedQuirksMode; }
    ALWAYS_INLINE bool inNoQuirksMode() const { return m_compatibilityMode == NoQuirksMode; }

    enum ReadyState {
        Loading,
        Interactive,
        Complete
    };
    void setReadyState(ReadyState);
    void setParsing(bool);
    ALWAYS_INLINE bool parsing() const { return m_bParsing; }

    ALWAYS_INLINE void stopLayout() { m_layoutStopped = true; }
    ALWAYS_INLINE bool isLayoutStopped() const { return m_layoutStopped; }
    
    bool shouldScheduleLayout() const;
    bool isLayoutTimerActive() const;
    int elapsedTime() const;
    
    ALWAYS_INLINE void setTextColor(const Color& color) { m_textColor = color; }
    ALWAYS_INLINE Color textColor() const { return m_textColor; }

    ALWAYS_INLINE const Color& linkColor() const { return m_linkColor; }
    ALWAYS_INLINE const Color& visitedLinkColor() const { return m_visitedLinkColor; }
    ALWAYS_INLINE const Color& activeLinkColor() const { return m_activeLinkColor; }
    ALWAYS_INLINE void setLinkColor(const Color& c) { m_linkColor = c; }
    ALWAYS_INLINE void setVisitedLinkColor(const Color& c) { m_visitedLinkColor = c; }
    ALWAYS_INLINE void setActiveLinkColor(const Color& c) { m_activeLinkColor = c; }
    void resetLinkColor();
    void resetVisitedLinkColor();
    void resetActiveLinkColor();
    VisitedLinkState* visitedLinkState() const { return m_visitedLinkState.get(); }
    
    MouseEventWithHitTestResults prepareMouseEvent(const HitTestRequest&, const LayoutPoint&, const PlatformMouseEvent&);

    String preferredStylesheetSet() const;
    String selectedStylesheetSet() const;
    void setSelectedStylesheetSet(const String&);

#if ENABLE(BML)
    virtual
#endif
    bool setFocusedElement(Element*, const FocusOptions& = FocusOptions());
    Element* focusedElement() const { return m_focusedElement.get(); }
    bool wasLastFocusByClick() const { return m_latestFocusTrigger == FocusTrigger::Click; }
    UserActionElementSet& userActionElements()  { return m_userActionElements; }
	const UserActionElementSet& userActionElements() const { return m_userActionElements; }

    virtual bool paint(GraphicsContext*, IntRect&) { return false; }
    
#if ENABLE(BML)
    virtual bool isBMLDocument() const { return false; }
    
    virtual void paintContentBegin(const IntRect&){ }
    virtual void paintContentEnd(const IntRect&){ }
    virtual void didPaintContent() { }

    virtual void queueFocusEvent(Node*){ }
    virtual void queueBlurEvent(Node*){ }
    virtual void queueClickEvent(Node*){ }
    virtual void queueChangeEvent(Node*){ }
    virtual void queueSubmitEvent(Node*){ }
    
    virtual void setClutUrl(const String&) {}
    virtual void setUsedKeyList(const String&) {}
    virtual void setResolution(const String&) {}
    virtual void setDisplayAspectRatio(const String&) {}
    
    virtual Color getClutColor(int) const { return Color(); }
    virtual int colorIndex(Color) const { return -1; }
    virtual String usedKeyList() const { return String(); }
    virtual String resolution() const { return String(); }
    virtual String displayAspectRatio() const { return String(); }
    
    virtual void registerNavElement(int, StyledElement*) {}
    
    virtual void subscribeAttributeChanged(HTMLBeitemElement*, bool){}
    
    virtual Element* currentFocus() { return 0; }
    virtual int bgImageScale(const RenderObject*, const StyleImage*) const { return 1; }
#endif
    
    // The m_ignoreAutofocus flag specifies whether or not the document has been changed by the user enough 
    // for WebCore to ignore the autofocus attribute on any form controls
    ALWAYS_INLINE bool ignoreAutofocus() const { return m_ignoreAutofocus; };
    ALWAYS_INLINE void setIgnoreAutofocus(bool shouldIgnore = true) { m_ignoreAutofocus = shouldIgnore; };

    Element* activeElement();
    bool hasFocus() const;

    void setFocusNavigationStartingNode(Node*);
    Element* focusNavigationStartingNode(FocusDirection) const;

    enum class NodeRemoval { Node, ChildrenOfNode };
    void adjustFocusedNodeOnNodeRemoval(Node&, NodeRemoval = NodeRemoval::Node);
    void adjustFocusNavigationNodeOnNodeRemoval(Node&, NodeRemoval = NodeRemoval::Node);

    bool isAutofocusProcessed() const { return m_isAutofocusProcessed; }
    void setAutofocusProcessed() { m_isAutofocusProcessed = true; }

    void appendAutofocusCandidate(Element&);
    void clearAutofocusCandidates() { m_autofocusCandidates.clear(); }
    void flushAutofocusCandidates();

    void hoveredElementDidDetach(Element*);
    void elementInActiveChainDidDetach(Element*);

    void updateHoverActiveState(const HitTestRequest&, Element*);

    // Updates for :target (CSS3 selector).
    void setCSSTarget(Element*);
    inline Element* cssTarget() const; // Defined in ElementInlines.h.
    
    void scheduleFullStyleRebuild();
    void scheduleStyleRecalc();
    void unscheduleStyleRecalc();
    bool hasPendingStyleRecalc() const;
    bool hasPendingFullStyleRebuild() const;
    void styleRecalcTimerFired();

    void registerNodeListForInvalidation(LiveNodeList&);
    void unregisterNodeListForInvalidation(LiveNodeList&);
    void registerCollection(HTMLCollection&);
    void unregisterCollection(HTMLCollection&);
    void collectionCachedIdNameMap(const HTMLCollection&);
    void collectionWillClearIdNameMap(const HTMLCollection&);
    bool shouldInvalidateNodeListAndCollectionCaches() const;
    bool shouldInvalidateNodeListAndCollectionCachesForAttribute(const QualifiedName& attrName) const;

    template <typename InvalidationFunction>
    void invalidateNodeListAndCollectionCaches(InvalidationFunction);

    void attachNodeIterator(NodeIterator&);
    void detachNodeIterator(NodeIterator&);
    void moveNodeIteratorsToNewDocument(Node&, Document&);

    void attachRange(Range&);
    void detachRange(Range&);

    void updateRangesAfterChildrenChanged(ContainerNode&);
    // nodeChildrenWillBeRemoved is used when removing all node children at once.
    void nodeChildrenWillBeRemoved(ContainerNode&);
    // nodeWillBeRemoved is only safe when removing one node at a time.
    void nodeWillBeRemoved(Node&);
    void parentlessNodeMovedToNewDocument(Node&);

    enum class AcceptChildOperation { Replace, InsertOrAdd };
    bool canAcceptChild(const Node& newChild, const Node* refChild, AcceptChildOperation) const;

    void textInserted(Node&, unsigned offset, unsigned length);
    void textRemoved(Node&, unsigned offset, unsigned length);
    void textNodesMerged(Text& oldNode, unsigned offset);
    void textNodeSplit(Text& oldNode);

    void createDOMWindow();
    void takeDOMWindowFrom(Document*);

    DOMWindow* domWindow() const { return m_domWindow.get(); }
    // In DOM Level 2, the Document's DOMWindow is called the defaultView.
    ALWAYS_INLINE DOMWindow* defaultView() const { return domWindow(); }

    Document& contextDocument() const;
    void setContextDocument(Document& document) { m_contextDocument = makeWeakPtr(document); }

    bool hasBrowsingContext() const { return !!frame(); }

    // Helper functions for forwarding DOMWindow event related tasks to the DOMWindow if it exists.
    void setWindowAttributeEventListener(const AtomicString& eventType, const QualifiedName& attributeName, const AtomicString& value, DOMWrapperWorld*);
    void setWindowAttributeEventListener(const AtomicString& eventType, PassRefPtr<EventListener>, DOMWrapperWorld*);
    EventListener* getWindowAttributeEventListener(const AtomicString& eventType, DOMWrapperWorld*);
    void dispatchWindowEvent(PassRefPtr<Event>, PassRefPtr<EventTarget> = nullptr);
    void dispatchWindowLoadEvent();
    virtual void dispatchWindowUnloadEvent();

    PassRefPtr<Event> createEvent(const String& eventType, ExceptionCode&);

    // keep track of what types of event listeners are registered, so we don't
    // dispatch events unnecessarily
    // FIXME: Consider using OptionSet.
    enum ListenerType {
        DOMSUBTREEMODIFIED_LISTENER          = 1 << 0,
        DOMNODEINSERTED_LISTENER             = 1 << 1,
        DOMNODEREMOVED_LISTENER              = 1 << 2,
        DOMNODEREMOVEDFROMDOCUMENT_LISTENER  = 1 << 3,
        DOMNODEINSERTEDINTODOCUMENT_LISTENER = 1 << 4,
        DOMCHARACTERDATAMODIFIED_LISTENER    = 1 << 5,
        OVERFLOWCHANGED_LISTENER             = 1 << 6,
        TRANSITIONEND_LISTENER               = 1 << 7,
        SCROLL_LISTENER                      = 1 << 8,
        FORCEWILLBEGIN_LISTENER              = 1 << 9,
        FORCECHANGED_LISTENER                = 1 << 10,
        FORCEDOWN_LISTENER                   = 1 << 11,
        FORCEUP_LISTENER                     = 1 << 12,
        FOCUSIN_LISTENER                     = 1 << 13,
        FOCUSOUT_LISTENER                    = 1 << 14,
    };

    ALWAYS_INLINE bool hasListenerType(ListenerType listenerType) const { return (m_listenerTypes & listenerType); }
    ALWAYS_INLINE void addListenerType(ListenerType listenerType) { m_listenerTypes = m_listenerTypes | listenerType; }
    void addListenerTypeIfNeeded(const AtomicString& eventType);

#if ENABLE(MUTATION_OBSERVERS)
    bool hasMutationObserversOfType(MutationObserver::MutationType type) const
    {
        return m_mutationObserverTypes & type;
    }
    bool hasMutationObservers() const { return m_mutationObserverTypes; }
    void addMutationObserverTypes(MutationObserverOptions types) { m_mutationObserverTypes |= types; }
#endif

    CSSStyleDeclaration* getOverrideStyle(Element*, const String& pseudoElt);

    int nodeAbsIndex(Node*);
    Node* nodeWithAbsIndex(int absIndex);

    /**
     * Handles a HTTP header equivalent set by a meta tag using <meta http-equiv="..." content="...">. This is called
     * when a meta tag is encountered during document parsing, and also when a script dynamically changes or adds a meta
     * tag. This enables scripts to use meta tags to perform refreshes and set expiry dates in addition to them being
     * specified in a HTML file.
     *
     * @param equiv The http header name (value of the meta tag's "equiv" attribute)
     * @param content The header value (value of the meta tag's "content" attribute)
     */
    void processHttpEquiv(const String& equiv, const String& content);
    void processViewport(const String& features, ViewportArguments::Type origin);
    void updateViewportArguments();
    void processReferrerPolicy(const String& policy, ReferrerPolicySource);

    // Returns the owning element in the parent document.
    // Returns 0 if this is the top level document.
    HTMLFrameOwnerElement* ownerElement() const;

    // Used by DOM bindings; no direction known.
    const String& title() const { return m_title.string(); }
    void setTitle(const String&);

    const AtomicString& dir() const;
    void setDir(const AtomicString&);

    void titleElementAdded(Element& titleElement);
    void titleElementRemoved(Element& titleElement);
    void titleElementTextChanged(Element& titleElement);

    String cookie(ExceptionCode&);
    void setCookie(const String&, ExceptionCode&);

    String referrer() const;

    String origin() const;

    String domain() const;
    void setDomain(const String& newDomain, ExceptionCode&);

    void overrideLastModified(const Optional<double>&);
    WEBCORE_EXPORT String lastModified() const;

    // The cookieURL is used to query the cookie database for this document's
    // cookies. For example, if the cookie URL is http://example.com, we'll
    // use the non-Secure cookies for example.com when computing
    // document.cookie.
    //
    // Q: How is the cookieURL different from the document's URL?
    // A: The two URLs are the same almost all the time.  However, if one
    //    document inherits the security context of another document, it
    //    inherits its cookieURL but not its URL.
    //
    ALWAYS_INLINE const KURL& cookieURL() const { return m_cookieURL; }
    void setCookieURL(const KURL&);

    // The firstPartyForCookies is used to compute whether this document
    // appears in a "third-party" context for the purpose of third-party
    // cookie blocking.  The document is in a third-party context if the
    // cookieURL and the firstPartyForCookies are from different hosts.
    //
    // Note: Some ports (including possibly Apple's) only consider the
    //       document in a third-party context if the cookieURL and the
    //       firstPartyForCookies have a different registry-controlled
    //       domain.
    //
    ALWAYS_INLINE const KURL& firstPartyForCookies() const { return m_firstPartyForCookies; }
    ALWAYS_INLINE void setFirstPartyForCookies(const KURL& url) { m_firstPartyForCookies = url; }

    bool isFullyActive() const;
    
    // The following implements the rule from HTML 4 for what valid names are.
    // To get this right for all the XML cases, we probably have to improve this or move it
    // and make it sensitive to the type of document.
    static bool isValidName(const String&);

    // The following breaks a qualified name into a prefix and a local name.
    // It also does a validity check, and returns false if the qualified name
    // is invalid.  It also sets ExceptionCode when name is invalid.
    static bool parseQualifiedName(const String& qualifiedName, String& prefix, String& localName, ExceptionCode&);

    // Checks to make sure prefix and namespace do not conflict (per DOM Core 3)
    static bool hasValidNamespaceForElements(const QualifiedName&);
    static bool hasValidNamespaceForAttributes(const QualifiedName&);

    // This is the "HTML body element" as defined by CSSOM View spec, the first body child of the
    // document element. See http://dev.w3.org/csswg/cssom-view/#the-html-body-element.
    HTMLBodyElement* body() const;

    // This is the "body element" as defined by HTML5, the first body or frameset child of the
    // document element. See https://html.spec.whatwg.org/multipage/dom.html#the-body-element-2.
    HTMLElement* bodyOrFrameset() const;
    void setBodyOrFrameset(PassRefPtr<HTMLElement>, ExceptionCode&);

    Location* location() const;

    HTMLHeadElement* head();

    ALWAYS_INLINE DocumentMarkerController* markers() const { return m_markers.get(); }

    bool execCommand(const String& command, bool userInterface = false, const String& value = String());
    bool queryCommandEnabled(const String& command);
    bool queryCommandIndeterm(const String& command);
    bool queryCommandState(const String& command);
    bool queryCommandSupported(const String& command);
    String queryCommandValue(const String& command);

    KURL openSearchDescriptionURL();

    // designMode support
    enum class DesignMode : bool { Off, On };
    void setDesignMode(DesignMode value);
    bool inDesignMode() const { return m_designMode == DesignMode::On; }
    WEBCORE_EXPORT String designMode() const;
    WEBCORE_EXPORT void setDesignMode(const String&);

    Document* parentDocument() const;
    Document* topDocument() const;
    bool isTopDocument() const { return topDocument() == this; }
    
    ScriptRunner* scriptRunner() { return m_scriptRunner.get(); }
    ScriptModuleLoader* moduleLoader() { return m_moduleLoader.get(); }

    HTMLScriptElement* currentScript() const { return !m_currentScriptStack.isEmpty() ? m_currentScriptStack.last().get() : 0; }
    void pushCurrentScript(PassRefPtr<HTMLScriptElement>);
    void popCurrentScript();

#if ENABLE(XSLT)
    void scheduleToApplyXSLTransforms();
    void applyPendingXSLTransformsNowIfScheduled();
    PassRefPtr<Document> transformSourceDocument() { return m_transformSourceDocument; }
    void setTransformSourceDocument(Document* doc) { m_transformSourceDocument = doc; }

    void setTransformSource(PassOwnPtr<TransformSource>);
    TransformSource* transformSource() const { return m_transformSource.get(); }
#endif

    ALWAYS_INLINE void incDOMTreeVersion() { m_domTreeVersion = ++s_globalTreeVersion; }
    ALWAYS_INLINE uint64_t domTreeVersion() const { return m_domTreeVersion; }

    // XPathEvaluator methods
    PassRefPtr<XPathExpression> createExpression(const String& expression,
                                                 XPathNSResolver* resolver,
                                                 ExceptionCode& ec);
    PassRefPtr<XPathNSResolver> createNSResolver(Node *nodeResolver);
    PassRefPtr<XPathResult> evaluate(const String& expression,
                                     Node* contextNode,
                                     XPathNSResolver* resolver,
                                     unsigned short type,
                                     XPathResult* result,
                                     ExceptionCode& ec);
    static PassRefPtr<Node> createNSResolverForBindings(Node& nodeResolver) { return nodeResolver; } // Legacy.

    bool hasNodesWithMissingStyle() const { return m_hasNodesWithMissingStyle; }
    void setHasNodesWithMissingStyle() { m_hasNodesWithMissingStyle = true; }
        
    // Extension for manipulating canvas drawing contexts for use in CSS
    CanvasRenderingContext* getCSSCanvasContext(const String& type, const String& name, int width, int height);
    HTMLCanvasElement* getCSSCanvasElement(const String& name);

    ALWAYS_INLINE bool isDNSPrefetchEnabled() const { return m_isDNSPrefetchEnabled; }
    void parseDNSPrefetchControlHeader(const String&);

    virtual void postTask(PassOwnPtr<Task>) FINAL; // Executes the task on context's thread asynchronously.

    virtual EventLoopTaskGroup& eventLoop() final;
    WindowEventLoop& windowEventLoop();

#if ENABLE(REQUEST_ANIMATION_FRAME)
    ScriptedAnimationController* scriptedAnimationController() { return m_scriptedAnimationController.get(); }
#endif
    void suspendScriptedAnimationControllerCallbacks();
    void resumeScriptedAnimationControllerCallbacks();
    virtual void scriptedAnimationControllerSetThrottled(bool);

    void serviceRequestAnimationFrameCallbacks();

    void windowScreenDidChange(PlatformDisplayID);

    virtual void finishedParsing(); // override in ImageDocument

    enum PageCacheState { NotInPageCache, AboutToEnterPageCache, InPageCache };

    PageCacheState pageCacheState() const { return m_pageCacheState; }
    void setPageCacheState(PageCacheState);

    // FIXME: Update callers to use pageCacheState() instead.
    bool inPageCache() const { return m_pageCacheState != NotInPageCache; }

    // Elements can register themselves for the "documentWillSuspendForPageCache()" and  
    // "documentDidResumeFromPageCache()" callbacks
    void registerForPageCacheSuspensionCallbacks(Element&);
    void unregisterForPageCacheSuspensionCallbacks(Element&);

    void registerForDocumentActivationCallbacks(Element&);
    void unregisterForDocumentActivationCallbacks(Element&);
    void documentWillBecomeInactive();
    void documentWillSuspendForPageCache();
    void documentDidResumeFromPageCache();

    void registerForMediaVolumeCallbacks(HTMLMediaElement&);
    void unregisterForMediaVolumeCallbacks(HTMLMediaElement&);
    void mediaVolumeDidChange();

    void registerForPrivateBrowsingStateChangedCallbacks(HTMLMediaElement&);
    void unregisterForPrivateBrowsingStateChangedCallbacks(HTMLMediaElement&);
    void privateBrowsingStateDidChange();

#if ENABLE(VIDEO_TRACK)
    void registerForCaptionPreferencesChangedCallbacks(HTMLMediaElement&);
    void unregisterForCaptionPreferencesChangedCallbacks(HTMLMediaElement&);
    void captionPreferencesChanged();
#endif

    void setShouldCreateRenderers(bool);
    bool shouldCreateRenderers();

    void setDecoder(PassRefPtr<TextResourceDecoder>);
    ALWAYS_INLINE TextResourceDecoder* decoder() const { return m_decoder.get(); }

    String displayStringModifiedByEncoding(const String&) const;
    PassRefPtr<StringImpl> displayStringModifiedByEncoding(PassRefPtr<StringImpl>) const;
    void displayBufferModifiedByEncoding(LChar* buffer, unsigned len) const
	{
		displayBufferModifiedByEncodingInternal(buffer, len);
	}
	void displayBufferModifiedByEncoding(UChar* buffer, unsigned len) const
	{
		displayBufferModifiedByEncodingInternal(buffer, len);
	}

    // Quirk for the benefit of Apple's Dictionary application.
    ALWAYS_INLINE void setFrameElementsShouldIgnoreScrolling(bool ignore) { m_frameElementsShouldIgnoreScrolling = ignore; }
    ALWAYS_INLINE bool frameElementsShouldIgnoreScrolling() const { return m_frameElementsShouldIgnoreScrolling; }

#if ENABLE(DASHBOARD_SUPPORT)
    void setDashboardRegionsDirty(bool f) { m_dashboardRegionsDirty = f; }
    bool dashboardRegionsDirty() const { return m_dashboardRegionsDirty; }
    bool hasDashboardRegions () const { return m_hasDashboardRegions; }
    void setHasDashboardRegions(bool f) { m_hasDashboardRegions = f; }
    const Vector<DashboardRegionValue>& dashboardRegions() const;
    void setDashboardRegions(const Vector<DashboardRegionValue>&);
#endif

    virtual void removeAllEventListeners() FINAL;

#if ENABLE(SVG)
    SVGDocumentExtensions* svgExtensionsIfExists() { return m_svgExtensions.get(); }
    SVGDocumentExtensions& ensureSVGExtensions();
#endif

    void initSecurityContext();
    void initContentSecurityPolicy();

    void updateURLForPushOrReplaceState(const KURL&);
    void statePopped(PassRefPtr<SerializedScriptValue>);

    ALWAYS_INLINE bool processingLoadEvent() const { return m_processingLoadEvent; }
    ALWAYS_INLINE bool loadEventFinished() const { return m_loadEventFinished; }

    virtual bool isContextThread() const FINAL;
    virtual bool isSecureContext() const FINAL;
    virtual bool isJSExecutionForbidden() const FINAL { return false; }

    void queueTaskToDispatchEvent(TaskSource, PassRefPtr<Event>);
    void queueTaskToDispatchEventOnWindow(TaskSource, PassRefPtr<Event>);
    void enqueueOverflowEvent(PassRefPtr<Event>);
    void dispatchPageshowEvent(PageshowEventPersistence);
    void enqueueHashchangeEvent(const String& oldURL, const String& newURL);
    void dispatchPopstateEvent(PassRefPtr<SerializedScriptValue> stateObject);

    void addMediaCanStartListener(MediaCanStartListener*);
    void removeMediaCanStartListener(MediaCanStartListener*);
    MediaCanStartListener* takeAnyMediaCanStartListener();

    const QualifiedName& idAttributeName() const { return m_idAttributeName; }
    
#if ENABLE(FULLSCREEN_API)
    bool webkitIsFullScreen() const { return m_fullScreenElement.get(); }
    bool webkitFullScreenKeyboardInputAllowed() const { return m_fullScreenElement.get() && m_areKeysEnabledInFullScreen; }
    Element* webkitCurrentFullScreenElement() const { return m_fullScreenElement.get(); }
    Element* webkitCurrentFullScreenElementForBindings() const { return ancestorElementInThisScope(webkitCurrentFullScreenElement()); }

    enum FullScreenCheckType {
        EnforceIFrameAllowFullScreenRequirement,
        ExemptIFrameAllowFullScreenRequirement,
    };

    void requestFullScreenForElement(Element*, unsigned short flags, FullScreenCheckType);
    void webkitCancelFullScreen();
    
    void webkitWillEnterFullScreenForElement(Element*);
    void webkitDidEnterFullScreenForElement(Element*);
    void webkitWillExitFullScreenForElement(Element*);
    void webkitDidExitFullScreenForElement(Element*);
    
    void setFullScreenRenderer(RenderTreeBuilder&, RenderFullScreen&);
    RenderFullScreen* fullScreenRenderer() const;
    
    void setFullScreenRendererSize(const IntSize&);
    void setFullScreenRendererBackgroundColor(Color);
    
    void fullScreenChangeDelayTimerFired();
    bool fullScreenIsAllowedForElement(Element*) const;
    void fullScreenElementRemoved();
    void adjustFullScreenElementOnNodeRemoval(Node&, NodeRemoval = NodeRemoval::Node);

    bool isAnimatingFullScreen() const;
    void setAnimatingFullScreen(bool);

    // W3C API
    bool webkitFullscreenEnabled() const;
    Element* webkitFullscreenElement() const { return !m_fullScreenElementStack.isEmpty() ? m_fullScreenElementStack.first().get() : 0; }
    Element* webkitFullscreenElementForBindings() const { return ancestorElementInThisScope(webkitFullscreenElement()); }
    void webkitExitFullscreen();
#endif

#if ENABLE(POINTER_LOCK)
    void exitPointerLock();
    Element* pointerLockElement() const;
#endif

    // Used to allow element that loads data without going through a FrameLoader to delay the 'load' event.
    void incrementLoadEventDelayCount() { ++m_loadEventDelayCount; }
    void decrementLoadEventDelayCount();
    bool isDelayingLoadEvent() const { return m_loadEventDelayCount; }
    void checkCompleted();

#if ENABLE(TOUCH_EVENTS)
    PassRefPtr<Touch> createTouch(DOMWindow*, EventTarget*, int identifier, int pageX, int pageY, int screenX, int screenY, int radiusX, int radiusY, float rotationAngle, float force, ExceptionCode&) const;
#endif

    ALWAYS_INLINE const DocumentTiming* timing() const { return &m_documentTiming; }

    double monotonicTimestamp() const;

#if ENABLE(REQUEST_ANIMATION_FRAME)
    int requestAnimationFrame(PassRefPtr<RequestAnimationFrameCallback>);
    void cancelAnimationFrame(int id);
    void clearScriptedAnimationController();
#endif

    int requestIdleCallback(Ref<IdleRequestCallback>&&, Seconds timeout);
    void cancelIdleCallback(int id);
    IdleCallbackController* idleCallbackController() { return m_idleCallbackController.get(); }

    virtual EventTarget* errorEventTarget() FINAL;
    virtual void logExceptionToConsole(const String& errorMessage, const String& sourceURL, int lineNumber, int columnNumber, PassRefPtr<ScriptCallStack>) FINAL;

    void initDNSPrefetch();

    ALWAYS_INLINE unsigned wheelEventHandlerCount() const { return m_wheelEventHandlerCount; }
    void didAddWheelEventHandler();
    void didRemoveWheelEventHandler();

#if ENABLE(TOUCH_EVENTS)
    unsigned touchEventHandlerCount() const { return m_touchEventHandlerCount; }
#else
    unsigned touchEventHandlerCount() const { return 0; }
#endif

    void startTrackingStyleRecalcs();
    unsigned styleRecalcCount() const;

    bool mayHaveRenderedSVGForeignObjects() const { return m_mayHaveRenderedSVGForeignObjects; }
    void setMayHaveRenderedSVGForeignObjects() { m_mayHaveRenderedSVGForeignObjects = true; }

    void didAddTouchEventHandler();
    void didRemoveTouchEventHandler();

    bool visualUpdatesAllowed() const { return m_visualUpdatesAllowed; }

#if ENABLE(MICRODATA)
    PassRefPtr<NodeList> getItems(const String& typeNames);
#endif
    
#if ENABLE(UNDO_MANAGER)
    PassRefPtr<UndoManager> undoManager();
#endif
    
    ALWAYS_INLINE bool isInDocumentWrite() { return m_writeRecursionDepth > 0; }

    void suspendScheduledTasks(ActiveDOMObject::ReasonForSuspension);
    void resumeScheduledTasks(ActiveDOMObject::ReasonForSuspension);

#if ENABLE(LINK_PRERENDER)
    Prerenderer* prerenderer() { return m_prerenderer.get(); }
#endif

#if ENABLE(TEXT_AUTOSIZING)
    TextAutosizer* textAutosizer() { return m_textAutosizer.get(); }
#endif

    void convertAbsoluteToClientQuads(Vector<FloatQuad>&, const RenderStyle&);
    void convertAbsoluteToClientRects(Vector<FloatRect>&, const RenderStyle&);
    void convertAbsoluteToClientRect(FloatRect&, const RenderStyle&);

    bool hasActiveParser();
    void incrementActiveParserCount() { ++m_activeParserCount; }
    void decrementActiveParserCount();

    DocumentSharedObjectPool* sharedObjectPool() { return m_sharedObjectPool.get(); }

    void invalidateMatchedPropertiesCacheAndForceStyleRecalc();

    void didRemoveAllPendingStylesheet();

    // Return a Locale for the default locale if the argument is null or empty.
    Locale& getCachedLocale(const AtomicString& locale = nullAtom);

    bool inStyleRecalc() const { return m_inStyleRecalc; }
    bool inRenderTreeUpdate() const { return m_inRenderTreeUpdate; }
    bool isResolvingTreeStyle() const { return m_isResolvingTreeStyle; }
    void setIsResolvingTreeStyle(bool);

    void updateTextRenderer(Text&, unsigned offsetOfReplacedText, unsigned lengthOfReplacedText);
    void updateSVGRenderer(SVGElement&);

#if ENABLE(RMAI)
    GalioWafer* ownerWafer() const;
    bool addEventListener(const AtomicString& eventType, PassRefPtr<EventListener>, const AddEventListenerOptions&) override;
    virtual bool removeEventListener(const AtomicString& eventType, EventListener*, const EventListenerOptions& = EventListenerOptions()) override;
#endif

#if ENABLE(TEMPLATE_ELEMENT)
    const Document* templateDocument() const;
    Document& ensureTemplateDocument();
    void setTemplateDocumentHost(Document* templateDocumentHost) { m_templateDocumentHost = makeWeakPtr(templateDocumentHost); }
    Document* templateDocumentHost() { return m_templateDocumentHost.get(); }
#endif

    Ref<DocumentFragment> documentFragmentForInnerOuterHTML();

    void didAssociateFormControl(Element&);

    bool hasDisabledFieldsetElement() const { return m_disabledFieldsetElementsCount; }
    void addDisabledFieldsetElement() { m_disabledFieldsetElementsCount++; }
    void removeDisabledFieldsetElement() { ASSERT(m_disabledFieldsetElementsCount); m_disabledFieldsetElementsCount--; }

    bool hasDataListElements() const { return m_dataListElementCount; }
    void incrementDataListElementCount() { ++m_dataListElementCount; }
    void decrementDataListElementCount() { ASSERT(m_dataListElementCount); --m_dataListElementCount; }

    void getParserLocation(String& url, unsigned& line, unsigned& column) const;

    virtual void addConsoleMessage(MessageSource, MessageType, MessageLevel, const String& message, unsigned long requestIdentifier = 0) FINAL;

	virtual const SecurityOrigin* topOrigin() const FINAL;

    bool isSameOriginAsTopDocument() const;

    void addTopLayerElement(Element&);
    void removeTopLayerElement(Element&);
    const ListHashSet<Ref<Element>>& topLayerElements() const { return m_topLayerElements; }
    bool hasTopLayerElement() const { return !m_topLayerElements.isEmpty(); }

    HTMLDialogElement* activeModalDialog() const;

    PassRefPtr<FontFaceSet> fonts();

    void setHasStyleWithViewportUnits() { m_hasStyleWithViewportUnits = true; }
    bool hasStyleWithViewportUnits() const { return m_hasStyleWithViewportUnits; }
    void updateViewportUnitsOnResize();

    void setNeedsDOMWindowResizeEvent();
    void setNeedsVisualViewportResize();
    void runResizeSteps();

    void addPendingScrollEventTarget(ContainerNode&);
    void setNeedsVisualViewportScrollEvent();
    void runScrollSteps();

    bool shouldEnforceContentDispositionAttachmentSandbox() const;
    void applyContentDispositionAttachmentSandbox();

#if ENABLE(INTERSECTION_OBSERVER)
    void addIntersectionObserver(IntersectionObserver&);
    void removeIntersectionObserver(IntersectionObserver&);
    unsigned numberOfIntersectionObservers() const { return m_intersectionObservers.size(); }
	void updateIntersectionObservations();
    void scheduleInitialIntersectionObservationUpdate();
    IntersectionObserverData& ensureIntersectionObserverData();
	IntersectionObserverData* intersectionObserverDataIfExists() { return m_intersectionObserverData.get(); }
#endif

#if ENABLE(RESIZE_OBSERVER)
    void addResizeObserver(ResizeObserver&);
    void removeResizeObserver(ResizeObserver&);
    unsigned numberOfResizeObservers() const { return m_resizeObservers.size(); }
    bool hasResizeObservers();
    // Return the minDepth of the active observations.
    size_t gatherResizeObservations(size_t deeperThan);
    void deliverResizeObservations();
    bool hasSkippedResizeObservations() const;
    void setHasSkippedResizeObservations(bool);
    void updateResizeObservations(Page&);
#endif

    void invalidateDOMCookieCache();

    using ContainerNode::setAttributeEventListener;
    void setAttributeEventListener(const AtomicString& eventType, const QualifiedName& attributeName, const AtomicString& value, DOMWrapperWorld* isolatedWorld);

    DOMSelection* getSelection();

    void didInsertInDocumentShadowRoot(ShadowRoot&);
    void didRemoveInDocumentShadowRoot(ShadowRoot&);
    const WeakListHashSet<ShadowRoot>& inDocumentShadowRoots() const { return m_inDocumentShadowRoots; }

    void attachToCachedFrame(CachedFrameBase&);
    void detachFromCachedFrame(CachedFrameBase&);

    void frameWasDisconnectedFromOwner();

    WEBCORE_EXPORT bool hitTest(const HitTestRequest&, HitTestResult&);
    bool hitTest(const HitTestRequest&, const HitTestLocation&, HitTestResult&);
#if !ASSERT_DISABLED
    bool inHitTesting() const { return m_inHitTesting; }
#endif

    void addDynamicMediaQueryDependentImage(HTMLImageElement&);
    void removeDynamicMediaQueryDependentImage(HTMLImageElement&);

    void scheduleRenderingUpdate();

#if USE(REQUEST_ANIMATION_FRAME_DISPLAY_MONITOR)
    DocumentAnimationScheduler& animationScheduler();
#endif

    DocumentTimeline* timeline();
    DocumentTimeline* existingTimeline() const { return m_timeline.get(); }
    Vector<RefPtr<WebAnimation>> getAnimations();
    struct MatchingAnimationsFunction
    {
    	virtual ~MatchingAnimationsFunction() { }
    	virtual bool operator()(Element&) const = 0;
    };
    Vector<RefPtr<WebAnimation>> matchingAnimations(const MatchingAnimationsFunction&);
    DocumentTimelinesController* timelinesController() const { return m_timelinesController.get(); }
	DocumentTimelinesController& ensureTimelinesController();
	void updateAnimationsAndSendEvents();
	void keyframesRuleDidChange(const String& name);

    const AtomicString& bgColor() const;
    void setBgColor(const String&);
    const AtomicString& fgColor() const;
    void setFgColor(const String&);
    const AtomicString& alinkColor() const;
    void setAlinkColor(const String&);
    const AtomicString& linkColorForBindings() const;
    void setLinkColorForBindings(const String&);
    const AtomicString& vlinkColor() const;
    void setVlinkColor(const String&);

    // Per https://html.spec.whatwg.org/multipage/obsolete.html#dom-document-clear, this method does nothing.
    void clear() { }
    // Per https://html.spec.whatwg.org/multipage/obsolete.html#dom-document-captureevents, this method does nothing.
    void captureEvents() { }
    // Per https://html.spec.whatwg.org/multipage/obsolete.html#dom-document-releaseevents, this method does nothing.
    void releaseEvents() { }

    const CSSRegisteredCustomPropertySet& getCSSRegisteredCustomPropertySet() const { return m_CSSRegisteredPropertySet; }
	bool registerCSSProperty(CSSRegisteredCustomProperty&&);

    void registerArticleElement(Element&);
    void unregisterArticleElement(Element&);
    void updateMainArticleElementAfterLayout();
    bool hasMainArticleElement() const { return !!m_mainArticleElement; }

    void addElementWithPendingUserAgentShadowTreeUpdate(Element&);
    void removeElementWithPendingUserAgentShadowTreeUpdate(Element&);

    ConstantPropertyMap& constantProperties() const { return *m_constantPropertyMap; }

    Editor& editor() { return *m_editor; }
    const Editor& editor() const { return *m_editor; }
    FrameSelection& selection() { return *m_selection; }
    const FrameSelection& selection() const { return *m_selection; }

    bool contains(const Node& node) const { return this == &node.treeScope() && node.isConnected(); }
    bool contains(const Node* node) const { return node && contains(*node); }

    KURL fallbackBaseURL() const;

protected:
    enum ConstructionFlags { Synthesized = 1, NonRenderedPlaceholder = 1 << 1 };
    Document(Frame*, const KURL&, unsigned = DefaultDocumentClass, unsigned constructionFlags = 0);

    virtual void didUpdateSecurityOrigin() OVERRIDE;

    ALWAYS_INLINE void clearXMLVersion() { m_xmlVersion = String(); }

    virtual Ref<Document> cloneDocumentWithoutChildren() const;

private:
    friend class Node;
    friend class IgnoreDestructiveWriteCountIncrementer;
    friend class IgnoreOpensDuringUnloadCountIncrementer;

    void detachFromFrame() { observeFrame(nullptr); }

    bool shouldInheritContentSecurityPolicyFromOwner() const;

    void updateTitleElement(Element& changingTitleElement);
    virtual void willDetachPage() FINAL;
    virtual void frameDestroyed() FINAL;

    void commonTeardown();

    RenderObject* renderer() const WTF_DELETED_FUNCTION;
    void setRenderer(RenderObject*) WTF_DELETED_FUNCTION;

    void createRenderTree();
    void detachParser();

    typedef void (*ArgumentsCallback)(const String& keyString, const String& valueString, Document*, void* data);
    void processArguments(const String& features, void* data, ArgumentsCallback);

    // FontSelectorClient
    virtual void fontsNeedUpdate(FontSelector*) FINAL;

    virtual bool isDocument() const FINAL { return true; }

    virtual void childrenChanged(const ChildChange&) FINAL;

    virtual String nodeName() const FINAL;
    virtual NodeType nodeType() const FINAL;
    virtual bool childTypeAllowed(NodeType) const FINAL;
    virtual Ref<Node> cloneNodeInternal(Document&, CloningOperation) FINAL;
    void cloneDataFromDocument(const Document&);

    virtual void refScriptExecutionContext() FINAL { ref(); }
    virtual void derefScriptExecutionContext() FINAL { deref(); }

    virtual void addMessage(MessageSource, MessageType, MessageLevel, const String& message, const String& sourceURL, unsigned lineNumber, unsigned columnNumber, PassRefPtr<ScriptCallStack>, unsigned long requestIdentifier = 0) FINAL;

    virtual double minimumTimerInterval() const FINAL;

    void updateTitleFromTitleElement();
    void updateTitle(const StringWithDirection&);
    void updateBaseURL();

    void buildAccessKeyMap(TreeScope* root);

    void loadEventDelayTimerFired();

    void pendingTasksTimerFired();

    static void didReceiveTask(void*);

    template <typename CharacterType>
    void displayBufferModifiedByEncodingInternal(CharacterType*, unsigned) const;

#if ENABLE(PAGE_VISIBILITY_API)
    PageVisibilityState pageVisibilityState() const;
#endif

    template <CollectionType collectionType>
    PassRefPtr<HTMLCollection> ensureCachedCollection();

#if ENABLE(FULLSCREEN_API)
    void dispatchFullScreenChangeOrErrorEvent(Deque<RefPtr<Node>>&, const AtomicString& eventName, bool shouldNotifyMediaElement);
	void clearFullscreenElementStack();
    void popFullscreenElementStack();
    void pushFullscreenElementStack(Element*);
    void addDocumentToFullScreenChangeEventQueue(Document*);
#endif

    void setVisualUpdatesAllowed(ReadyState);
    void setVisualUpdatesAllowed(bool);
    void visualUpdatesSuppressionTimerFired();

    void didAssociateFormControlsTimerFired();

    // DOM Cookies caching.
    const String& cachedDOMCookies() const { return m_cachedDOMCookies; }
    void setCachedDOMCookies(const String&);
    bool isDOMCookieCacheValid() const { return m_cookieCacheExpiryTimer.isActive(); }
    void domCookieCacheExpiryTimerFired();
    virtual void didLoadResourceSynchronously(const ResourceRequest&) FINAL;

    HttpEquivPolicy httpEquivPolicy() const;

    void platformSuspendOrStopActiveDOMObjects();

    Style::Update& ensurePendingRenderTreeUpdate();

    bool canNavigateInternal(Frame& targetFrame);
    bool isNavigationBlockedByThirdPartyIFrameRedirectBlocking(Frame& targetFrame, const KURL& destinationURL);

    bool isBodyPotentiallyScrollable(HTMLBodyElement&);

    const RefPtr<Settings> m_settings;

    RefPtr<DOMWindow> m_domWindow;
    WeakPtr<Document> m_contextDocument;

    RefPtr<CachedResourceLoader> m_cachedResourceLoader;
    RefPtr<DocumentParser> m_parser;

    // Document URLs.
    KURL m_url; // Document.URL: The URL from which this document was retrieved.
    KURL m_creationURL; // https://html.spec.whatwg.org/multipage/webappapis.html#concept-environment-creation-url.
	KURL m_baseURL; // Node.baseURI: The URL to use when resolving relative URLs.
    KURL m_baseURLOverride; // An alternative base URL that takes precedence over m_baseURL (but not m_baseElementURL).
    KURL m_baseElementURL; // The URL set by the <base> element.
    KURL m_cookieURL; // The URL to use for cookie access.
    KURL m_firstPartyForCookies; // The policy URL for third-party cookie blocking.

    // Document.documentURI:
    // Although URL-like, Document.documentURI can actually be set to any
    // string by content.  Document.documentURI affects m_baseURL unless the
    // document contains a <base> element, in which case the <base> element
    // takes precedence.
    //
    // This property is read-only from JavaScript, but writable from Objective C.
    String m_documentURI;

    String m_baseTarget;

    // MIME type of the document in case it was cloned or created by XHR.
    String m_overriddenMIMEType;

    OwnPtr<DOMImplementation> m_implementation;

    RefPtr<Node> m_focusNavigationStartingNode;
    Deque<WeakPtr<Element>> m_autofocusCandidates;
    RefPtr<Element> m_focusedElement;
    RefPtr<Element> m_hoveredElement;
    RefPtr<Element> m_activeElement;
    RefPtr<Element> m_documentElement;
    UserActionElementSet m_userActionElements;

    uint64_t m_domTreeVersion;
    static uint64_t s_globalTreeVersion;

    WeakHashSet<NodeIterator> m_nodeIterators;
    WeakHashSet<Range> m_ranges;

    OwnPtr<Style::Scope> m_styleScope;
    OwnPtr<ExtensionStyleSheets> m_extensionStyleSheets;
    RefPtr<StyleSheetList> m_styleSheetList;
    
    OwnPtr<FormController> m_formController;

    Color m_textColor { Color::black };
    Color m_linkColor;
    Color m_visitedLinkColor;
    Color m_activeLinkColor;
    OwnPtr<VisitedLinkState> m_visitedLinkState;

    StringWithDirection m_title;
    StringWithDirection m_rawTitle;
    RefPtr<Element> m_titleElement;

    mutable AXObjectCache* m_axObjectCache { nullptr };
    OwnPtr<DocumentMarkerController> m_markers;

    Timer<Document> m_styleRecalcTimer;

    OwnPtr<Style::Update> m_pendingRenderTreeUpdate;

    WeakPtr<Element> m_cssTarget;

    double m_startTime;
    
    OwnPtr<ScriptRunner> m_scriptRunner;
    OwnPtr<ScriptModuleLoader> m_moduleLoader;

    Vector<RefPtr<HTMLScriptElement>> m_currentScriptStack;

#if ENABLE(XSLT)
    void applyPendingXSLTransformsTimerFired();

    OwnPtr<TransformSource> m_transformSource;
    RefPtr<Document> m_transformSourceDocument;
    Timer<Document> m_applyPendingXSLTransformsTimer;
    bool m_hasPendingXSLTransforms { false };
#endif

    String m_xmlEncoding;
    String m_xmlVersion;
    uint8_t m_xmlStandalone { StandaloneUnspecified };
    bool m_hasXMLDeclaration { false };

    String m_contentLanguage;
    
    RefPtr<TextResourceDecoder> m_decoder;

    HashSet<LiveNodeList*> m_listsInvalidatedAtDocument;
    HashSet<HTMLCollection*> m_collectionsInvalidatedAtDocument;
    unsigned m_nodeListAndCollectionCounts[numNodeListInvalidationTypes];

    RefPtr<XPathEvaluator> m_xpathEvaluator;

#if ENABLE(SVG)
    OwnPtr<SVGDocumentExtensions> m_svgExtensions;
#endif

#if ENABLE(DASHBOARD_SUPPORT)
    Vector<DashboardRegionValue> m_dashboardRegions;
    bool m_hasDashboardRegions { false };
    bool m_dashboardRegionsDirty { false };
#endif

    HashMap<String, RefPtr<HTMLCanvasElement>> m_cssCanvasElements;

    PageCacheState m_pageCacheState { NotInPageCache };

    float m_deviceScaleFactor { 1.0f };

    WeakHashSet<Element> m_documentActivationCallbackElements;
    WeakHashSet<Element> m_documentSuspensionCallbackElements;
    WeakHashSet<HTMLMediaElement> m_mediaVolumeCallbackElements;
    WeakHashSet<HTMLMediaElement> m_privateBrowsingStateChangedElements;
#if ENABLE(VIDEO_TRACK)
    WeakHashSet<HTMLMediaElement> m_captionPreferencesChangedElements;
#endif

    WeakPtr<Element> m_mainArticleElement;
    WeakHashSet<Element> m_articleElements;

#if ENABLE(PAGE_VISIBILITY_API)
    WeakHashSet<Element> m_visibilityStateCallbackElements;
#endif

    HashMap<String, WeakPtr<Element>, ASCIICaseInsensitiveHash> m_elementsByAccessKey;
    bool m_accessKeyMapValid { false };

    OwnPtr<ConstantPropertyMap> m_constantPropertyMap;

    OwnPtr<SelectorQueryCache> m_selectorQueryCache;

    DocumentClassFlags m_documentClasses;

    bool m_isCEHTML { false };
    bool m_isTTML { false };

    bool m_layoutStopped { false };
    bool m_isDirectGLMode { false };

    RenderPtr<RenderView> m_renderView;
    OwnPtr<RenderStyle> m_initialContainingBlockStyle;

    HashSet<MediaCanStartListener*> m_mediaCanStartListeners;

    QualifiedName m_idAttributeName;

#if ENABLE(FULLSCREEN_API)
    bool m_areKeysEnabledInFullScreen { false };
    RefPtr<Element> m_fullScreenElement;
    Deque<RefPtr<Element> > m_fullScreenElementStack;
    WeakPtr<RenderFullScreen> m_fullScreenRenderer;
    Timer<Document> m_fullScreenChangeDelayTimer;
    Deque<RefPtr<Node> > m_fullScreenChangeEventTargetQueue;
    Deque<RefPtr<Node> > m_fullScreenErrorEventTargetQueue;
    bool m_isAnimatingFullScreen { false };
    LayoutRect m_savedPlaceholderFrameRect;
    OwnPtr<RenderStyle> m_savedPlaceholderRenderStyle;
#endif

    WeakHashSet<HTMLImageElement> m_dynamicMediaQueryDependentImages;

#if ENABLE(INTERSECTION_OBSERVER)
    Vector<WeakPtr<IntersectionObserver>> m_intersectionObservers;

    void intersectionObserversInitialUpdateTimerFired();
    Timer<Document> m_intersectionObserversInitialUpdateTimer;
    // This is only non-null when this document is an explicit root.
    OwnPtr<IntersectionObserverData> m_intersectionObserverData;
#endif

#if ENABLE(RESIZE_OBSERVER)
    Vector<WeakPtr<ResizeObserver>> m_resizeObservers;
#endif

    Timer<Document> m_loadEventDelayTimer;

    ViewportArguments m_viewportArguments;

    DocumentTiming m_documentTiming;
    RefPtr<MediaQueryMatcher> m_mediaQueryMatcher;
    
    unsigned m_wheelEventHandlerCount { 0 };
#if ENABLE(TOUCH_EVENTS)
    unsigned m_touchEventHandlerCount { 0 };
#endif

    bool m_mayHaveRenderedSVGForeignObjects { false };

#if ENABLE(REQUEST_ANIMATION_FRAME)
    RefPtr<ScriptedAnimationController> m_scriptedAnimationController;
#endif

    OwnPtr<IdleCallbackController> m_idleCallbackController;

    Timer<Document> m_pendingTasksTimer;
    Vector<OwnPtr<Task> > m_pendingTasks;

#if ENABLE(LINK_PRERENDER)
    OwnPtr<Prerenderer> m_prerenderer;
#endif

#if ENABLE(TEXT_AUTOSIZING)
    OwnPtr<TextAutosizer> m_textAutosizer;
#endif

    bool m_scheduledTasksAreSuspended { false };
    
    bool m_visualUpdatesAllowed { true };
    Timer<Document> m_visualUpdatesSuppressionTimer;

    void sharedObjectPoolClearTimerFired();
    Timer<Document> m_sharedObjectPoolClearTimer;

    OwnPtr<DocumentSharedObjectPool> m_sharedObjectPool;

    RefPtr<CSSFontSelector> m_fontSelector;

    bool m_cacheControlContainsNoCache { false };
    bool m_cacheControlContainsNoStore { false };

    WeakListHashSet<ShadowRoot> m_inDocumentShadowRoots;

#if ENABLE(TEMPLATE_ELEMENT)
    RefPtr<Document> m_templateDocument;
    WeakPtr<Document> m_templateDocumentHost; // Manually managed weakref (backpointer from m_templateDocument).
#endif

    RefPtr<DocumentFragment> m_documentFragmentForInnerOuterHTML;

    unsigned m_disabledFieldsetElementsCount { 0 };

    unsigned m_dataListElementCount { 0 };

    Timer<Document> m_cookieCacheExpiryTimer;
    String m_cachedDOMCookies;

    Optional<double> m_overrideLastModified;

#if USE(REQUEST_ANIMATION_FRAME_DISPLAY_MONITOR)
    RefPtr<DocumentAnimationScheduler> m_animationScheduler;
#endif
    DocumentIdentifier m_identifier;

    Timer<Document> m_didAssociateFormControlsTimer;
    WeakHashSet<Element> m_associatedFormControls;

    unsigned m_listenerTypes { 0 };
    unsigned m_referencingNodeCount { 0 };
    int m_loadEventDelayCount { 0 };
    unsigned m_lastStyleUpdateSizeForTesting { 0 };

    // https://html.spec.whatwg.org/multipage/webappapis.html#ignore-opens-during-unload-counter
    unsigned m_ignoreOpensDuringUnloadCount { 0 };

    // https://html.spec.whatwg.org/multipage/webappapis.html#ignore-destructive-writes-counter
    unsigned m_ignoreDestructiveWriteCount { 0 };

    unsigned m_activeParserCount { 0 };
    unsigned m_styleRecalcCount { 0 };

    unsigned m_writeRecursionDepth { 0 };

    DesignMode m_designMode { DesignMode::Off };
    Optional<ReferrerPolicy> m_referrerPolicy;
    ReadyState m_readyState { Complete };

#if ENABLE(MUTATION_OBSERVERS)
    MutationObserverOptions m_mutationObserverTypes { 0 };
#endif

    bool m_ignoreAutofocus { false };

    bool m_writeRecursionIsTooDeep { false };
    bool m_wellFormed { false };
    bool m_createRenderers { true };

    bool m_hasNodesWithMissingStyle { false };
    // But sometimes you need to ignore pending stylesheet count to
    // force an immediate layout when requested by JS.
    bool m_ignorePendingStylesheets { false };

    bool m_hasElementUsingStyleBasedEditability { false };
    bool m_focusNavigationStartingNodeIsRemoved { false };

    bool m_printing { false };
    bool m_paginatedForScreen { false };

    CompatibilityMode m_compatibilityMode { NoQuirksMode };
    bool m_compatibilityModeLocked { false }; // This is cheaper than making setCompatibilityMode virtual.

    // FIXME: Merge these 2 variables into an enum. Also, FrameLoader::m_didCallImplicitClose
    // is almost a duplication of this data, so that should probably get merged in too.
    // FIXME: Document::m_processingLoadEvent and DocumentLoader::m_wasOnloadHandled are roughly the same
    // and should be merged.
    bool m_processingLoadEvent { false };
    bool m_loadEventFinished { false };

    bool m_visuallyOrdered { false };
    bool m_bParsing { false };

    bool m_needsFullStyleRebuild { false };
    bool m_inStyleRecalc { false };
    bool m_closeAfterStyleRecalc { false };
    bool m_inRenderTreeUpdate { false };
    bool m_isResolvingTreeStyle { false };

    bool m_gotoAnchorNeededAfterStylesheetsLoad { false };
    bool m_isDNSPrefetchEnabled { false };
    bool m_haveExplicitlyDisabledDNSPrefetch { false };
    bool m_frameElementsShouldIgnoreScrolling { false };

    bool m_isSynthesized { false };
    bool m_isNonRenderedPlaceholder { false };

    bool m_isAutofocusProcessed { false };

    bool m_sawElementsInKnownNamespaces { false };
    bool m_isSrcdocDocument { false };

    bool m_renderTreeBeingDestroyed { false };
    bool m_hasPreparedForDestruction { false };

    bool m_hasStyleWithViewportUnits { false };
    bool m_needsDOMWindowResizeEvent { false };
    bool m_needsVisualViewportResizeEvent { false };
    bool m_needsVisualViewportScrollEvent { false };

#ifndef NDEBUG
    bool m_didDispatchViewportPropertiesChanged { false };
#endif

#if !ASSERT_DISABLED
    bool m_inHitTesting { false };
#endif

    FocusTrigger m_latestFocusTrigger { FocusTrigger::Other };

    struct PendingScrollEventTargetList;
    OwnPtr<PendingScrollEventTargetList> m_pendingScrollEventTargetList;

    RefPtr<DocumentTimeline> m_timeline;
    OwnPtr<DocumentTimelinesController> m_timelinesController;

    RefPtr<WindowEventLoop> m_eventLoop;
    OwnPtr<EventLoopTaskGroup> m_documentTaskGroup;

    CSSRegisteredCustomPropertySet m_CSSRegisteredPropertySet;

#if ENABLE(UNDO_MANAGER)
    RefPtr<UndoManager> m_undoManager;
#endif

    ListHashSet<Ref<Element>> m_topLayerElements;

    WeakHashSet<Element> m_elementsWithPendingUserAgentShadowTreeUpdates;

    typedef HashMap<AtomicString, OwnPtr<Locale> > LocaleIdentifierToLocaleMap;
    LocaleIdentifierToLocaleMap m_localeCache;

    OwnPtr<Editor> m_editor;
    OwnPtr<FrameSelection> m_selection;
};

#if ENABLE(TEMPLATE_ELEMENT)
inline const Document* Document::templateDocument() const
{
    // If DOCUMENT does not have a browsing context, Let TEMPLATE CONTENTS OWNER be DOCUMENT and abort these steps.
    if (!m_frame)
        return this;

    return m_templateDocument.get();
}
#endif

// Put these methods here, because they require the Document definition, but we really want to inline them.

inline Document *Node::ownerDocument() const
{
    Document& document = this->document();
    return &document == this ? 0 : &document;
}

ALWAYS_INLINE ScriptExecutionContext* Node::scriptExecutionContext() const
{
    return &document().contextDocument();
}

Element* eventTargetElementForDocument(Document*);

inline void Document::registerForDocumentActivationCallbacks(Element& e)
{
    m_documentActivationCallbackElements.add(e);
}

inline void Document::unregisterForDocumentActivationCallbacks(Element& e)
{
    m_documentActivationCallbackElements.remove(e);
}

inline void Document::startTrackingStyleRecalcs()
{
    m_styleRecalcCount = 0;
}

inline unsigned Document::styleRecalcCount() const
{
    return m_styleRecalcCount;
}

inline TextEncoding Document::textEncoding() const
{
    if (auto* decoder = this->decoder())
        return decoder->encoding();
    return TextEncoding();
}

inline bool isDocument(const Node& node) { return node.isDocumentNode(); }
void isDocument(const Document&); // Catch unnecessary runtime check of type known at compile time.

inline bool isDocument(ScriptExecutionContext* scriptExecutionContext)
{
	return scriptExecutionContext && scriptExecutionContext->isDocument();
}

inline Document& toDocument(ScriptExecutionContext& scriptExecutionContext)
{
	return static_cast<Document&>(scriptExecutionContext);
}

inline const Document& toDocument(const ScriptExecutionContext& scriptExecutionContext)
{
	return static_cast<const Document&>(scriptExecutionContext);
}

template<>
ALWAYS_INLINE Document* dynamicDowncast(ScriptExecutionContext* source)
{
    //static_assert(!std::is_same_v<Source, Target>, "Unnecessary cast to same type");
    //static_assert(std::is_base_of_v<Source, Target>, "Should be a downcast");
    return source && source->isDocument() ? static_cast<Document*>(source) : nullptr;
}

NODE_TYPE_CASTS(Document)

#define DOCUMENT_TYPE_CASTS(ToClassName) \
    TYPE_CASTS_BASE(ToClassName, Document, document, WebCore::is##ToClassName(*document), WebCore::is##ToClassName(document))

} // namespace WebCore
