/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 *           (C) 2006 Alexey Proskuryakov (ap@webkit.org)
 * Copyright (C) 2004-2020 Apple Inc. All rights reserved.
 * Copyright (C) 2008, 2009 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)
 * Copyright (C) Research In Motion Limited 2010. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "DocumentMarker.h"
#include "FloatRect.h"
#include "LayoutPoint.h"
#include <wtf/HashMap.h>
#include <wtf/Vector.h>

namespace WebCore {

class Document;
class Node;
class RenderedDocumentMarker;
class Text;

enum class RemovePartiallyOverlappingMarker : bool { No, Yes };

class DocumentMarkerController {
    WTF_MAKE_NONCOPYABLE(DocumentMarkerController); WTF_MAKE_FAST_ALLOCATED;
public:

    DocumentMarkerController(Document&);
    ~DocumentMarkerController();

    void detach();

    void addMarker(const SimpleRange&, DocumentMarker::MarkerType);
    void addMarker(const SimpleRange&, DocumentMarker::MarkerType, const String& description);
    void addMarker(const SimpleRange&, DocumentMarker::MarkerType, PassRefPtr<DocumentMarkerDetails>);
    void addMarker(Text&, unsigned startOffset, unsigned length, DocumentMarker::MarkerType);
    void addMarker(Text&, unsigned startOffset, unsigned length, DocumentMarker::MarkerType, PassRefPtr<DocumentMarkerDetails>);
    void addDraggedContentMarker(const SimpleRange&);

	void copyMarkers(Node& source, OffsetRange, Node& destination);
	bool hasMarkers() const;
	bool hasMarkers(const SimpleRange&, DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers());

    class FilterFunctor
    {
    public:
    	virtual ~FilterFunctor() = default;
    	virtual bool operator()(const DocumentMarker&) = 0;
    protected:
    	FilterFunctor() = default;
    };

    // When a marker partially overlaps with range, if removePartiallyOverlappingMarkers is true, we completely
    // remove the marker. If the argument is false, we will adjust the span of the marker so that it retains
    // the portion that is outside of the range.
    void removeMarkers(const SimpleRange&, DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers(), RemovePartiallyOverlappingMarker = RemovePartiallyOverlappingMarker::No);
    void removeMarkers(Node&, OffsetRange, DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers(), FilterFunctor* filterFunctor = nullptr, RemovePartiallyOverlappingMarker = RemovePartiallyOverlappingMarker::No);

    // Return false from filterFunction to remove the marker.
    WEBCORE_EXPORT void filterMarkers(const SimpleRange&, FilterFunctor* filterFunctor, DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers(), RemovePartiallyOverlappingMarker = RemovePartiallyOverlappingMarker::No);

    void removeMarkers(DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers());
    void removeMarkers(Node&, DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers());
    void repaintMarkers(DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers());
    void shiftMarkers(Node&, unsigned startOffset, int delta);

    void setMarkersActive(Node& node, OffsetRange, bool active);
    void setMarkersActive(const SimpleRange&, bool);

    Vector<RenderedDocumentMarker*> markersFor(Node&, DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers());
    Vector<RenderedDocumentMarker*> markersInRange(const SimpleRange&, DocumentMarker::MarkerTypes);
    void clearDescriptionOnMarkersIntersectingRange(const SimpleRange&, DocumentMarker::MarkerTypes);

    void updateRectsForInvalidatedMarkersOfType(DocumentMarker::MarkerType);

    void invalidateRectsForAllMarkers();
    void invalidateRectsForMarkersInNode(Node&);

    DocumentMarker* markerContainingPoint(const LayoutPoint&, DocumentMarker::MarkerType);
    Vector<FloatRect> renderedRectsForMarkers(DocumentMarker::MarkerType);

#ifndef NDEBUG
    void showMarkers() const;
#endif

private:
    void addMarker(Node&, const DocumentMarker&);

    struct TextRange {
        Ref<Node> node;
        OffsetRange range;
    };
    Vector<TextRange> collectTextRanges(const SimpleRange&);

    template<typename Functor>
    inline void forEach(const SimpleRange&, DocumentMarker::MarkerTypes, Functor&& functor);

    typedef HashMap<RefPtr<Node>, OwnPtr<Vector<RenderedDocumentMarker>>> MarkerMap;
    bool possiblyHasMarkers(DocumentMarker::MarkerTypes);
    void removeMarkersFromList(MarkerMap::iterator, DocumentMarker::MarkerTypes);

    MarkerMap m_markers;
    // Provide a quick way to determine whether a particular marker type is absent without going through the map.
    DocumentMarker::MarkerTypes m_possiblyExistingMarkerTypes;
    Document& m_document;
};

void addMarker(const SimpleRange&, DocumentMarker::MarkerType, const String& description);
void addMarker(const SimpleRange&, DocumentMarker::MarkerType, PassRefPtr<DocumentMarkerDetails> = nullptr);
void addMarker(Text&, unsigned startOffset, unsigned length, DocumentMarker::MarkerType, PassRefPtr<DocumentMarkerDetails> = nullptr);
void removeMarkers(const SimpleRange&, DocumentMarker::MarkerTypes = DocumentMarker::AllMarkers(), RemovePartiallyOverlappingMarker = RemovePartiallyOverlappingMarker::No);

WEBCORE_EXPORT SimpleRange makeSimpleRange(Node&, const DocumentMarker&);

inline bool DocumentMarkerController::hasMarkers() const
{
    ASSERT(m_markers.isEmpty() == !m_possiblyExistingMarkerTypes.containsAny(DocumentMarker::allMarkers()));
    return !m_markers.isEmpty();
}

} // namespace WebCore

#ifndef NDEBUG
void showDocumentMarkers(const WebCore::DocumentMarkerController*);
#endif
