/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Peter Kelly (pmk@post.com)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 *           (C) 2007 David Smith (catfish.man@gmail.com)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010 Apple Inc. All rights reserved.
 *           (C) 2007 Eric Seidel (eric@webkit.org)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "Element.h"

#if HAVE(ACCESSIBILITY)
#include "AXObjectCache.h"
#endif
#include "Attr.h"
#include "AttributeChangeInvalidation.h"
#include "CSSParser.h"
#include "CSSPropertyAnimation.h"
#include "CSSSelectorList.h"
#include "ChildChangeInvalidation.h"
#include "Chrome.h"
#include "ChromeClient.h"
#include "ClassChangeInvalidation.h"
#include "ChildListMutationScope.h"
#include "ClientRect.h"
#include "ClientRectList.h"
#include "ComposedTreeAncestorIterator.h"
#include "ComposedTreeIterator.h"
#include "ContainerNodeAlgorithms.h"
#include "ContentSecurityPolicy.h"
#include "DOMRect.h"
#include "DOMRectList.h"
#include "DOMTokenList.h"
#include "DOMWindow.h"
#include "DatasetDOMStringMap.h"
#include "Dictionary.h"
#include "Document.h"
#include "DocumentFragment.h"
#include "DocumentSharedObjectPool.h"
#include "DocumentTimeline.h"
#include "ElementAnimationRareData.h"
#include "ElementInlines.h"
#include "ElementRareData.h"
#include "ElementIterator.h"
#include "EventDispatcher.h"
#include "ExceptionCode.h"
#include "FocusController.h"
#include "FocusEvent.h"
#include "Frame.h"
#include "FrameView.h"
#include "GetAnimationsOptions.h"
#include "HTMLBodyElement.h"
#include "HTMLCanvasElement.h"
#include "HTMLCollection.h"
#include "HTMLDialogElement.h"
#include "HTMLDocument.h"
#include "HTMLFrameOwnerElement.h"
#include "HTMLHtmlElement.h"
#include "HTMLNameCollection.h"
#include "HTMLNames.h"
#include "HTMLOptGroupElement.h"
#include "HTMLOptionElement.h"
#include "HTMLParserIdioms.h"
#include "HTMLScriptElement.h"
#include "HTMLSelectElement.h"
#include "HTMLTemplateElement.h"
#include "IdChangeInvalidation.h"
#include "IdTargetObserverRegistry.h"
#include "InspectorInstrumentation.h"
#include "JSLazyEventListener.h"
#include "JSShadowRoot.h"
#include "KeyboardEvent.h"
#include "KeyframeEffect.h"
#include "MouseEvent.h"
#include "MutationObserverInterestGroup.h"
#include "MutationRecord.h"
#include "NamedNodeMap.h"
#include "NodeList.h"
#include "NodeRenderStyle.h"
#include "Page.h"
#include "PlatformWheelEvent.h"
#include "PointerLockController.h"
#include "PseudoClassChangeInvalidation.h"
#include "PseudoElement.h"
#include "RenderBlockFlow.h"
#include "RenderLayer.h"
#include "RenderListBox.h"
#include "RenderTheme.h"
#include "RenderTreeUpdater.h"
#include "RenderView.h"
#include "RenderWidget.h"
#include "SelectorQuery.h"
#include "Settings.h"
#include "ShadowRoot.h"
#include "SimulatedClick.h"
#include "SlotAssignment.h"
#include "StyleInvalidator.h"
#include "StyleResolver.h"
#include "StyleScope.h"
#include "StyleTreeResolver.h"
#include "Text.h"
#include "TextIterator.h"
#include "VoidCallback.h"
#include "WebAnimation.h"
#include "XLinkNames.h"
#include "XMLNSNames.h"
#include "XMLNames.h"
#include "htmlediting.h"
#include "markup.h"
#include <wtf/BitVector.h>
#include <wtf/CurrentTime.h>
#include <wtf/Deque.h>
#include <wtf/text/CString.h>

#if ENABLE(SVG)
#include "SVGDocumentExtensions.h"
#include "SVGElement.h"
#include "SVGNames.h"
#include "SVGScriptElement.h"
#endif

namespace WebCore {

using namespace HTMLNames;
using namespace XMLNames;

static inline bool shouldIgnoreAttributeCase(const Element& e)
{
    return e.document().isHTMLDocument() && e.isHTMLElement();
}

typedef Vector<RefPtr<Attr> > AttrNodeList;
typedef HashMap<Element*, OwnPtr<AttrNodeList> > AttrNodeListMap;

static AttrNodeListMap& attrNodeListMap()
{
	static NeverDestroyed<AttrNodeListMap> map;
    return map;
}

static AttrNodeList* attrNodeListForElement(Element* element)
{
    if (!element->hasSyntheticAttrChildNodes())
        return 0;
    ASSERT(attrNodeListMap().contains(element));
    return attrNodeListMap().get(element);
}

static AttrNodeList* ensureAttrNodeListForElement(Element* element)
{
    if (element->hasSyntheticAttrChildNodes()) {
        ASSERT(attrNodeListMap().contains(element));
        return attrNodeListMap().get(element);
    }
    ASSERT(!attrNodeListMap().contains(element));
    element->setHasSyntheticAttrChildNodes(true);
    AttrNodeListMap::AddResult result = attrNodeListMap().add(element, adoptPtr(new AttrNodeList));
    return result.iterator->second.get();
}

static void removeAttrNodeListForElement(Element* element)
{
    ASSERT(element->hasSyntheticAttrChildNodes());
    ASSERT(attrNodeListMap().contains(element));
    attrNodeListMap().remove(element);
    element->setHasSyntheticAttrChildNodes(false);
}

static Attr* findAttrNodeInList(AttrNodeList& attrNodeList, const QualifiedName& name)
{
    for (auto& node : attrNodeList) {
        if (node->qualifiedName().matches(name))
            return node.get();
    }
    return nullptr;
}

PassRefPtr<Element> Element::create(const QualifiedName& tagName, Document* document)
{
    return adoptRef(new Element(tagName, document, CreateElement));
}

Element::~Element()
{
    ASSERT(!beforePseudoElement());
    ASSERT(!afterPseudoElement());

#if ENABLE(INTERSECTION_OBSERVER)
    disconnectFromIntersectionObservers();
#endif

#if ENABLE(RESIZE_OBSERVER)
    disconnectFromResizeObservers();
#endif

	removeShadowRoot();

    if (hasSyntheticAttrChildNodes())
        detachAllAttrNodesFromElement();
}

inline void Node::setTabIndexState(TabIndexState state)
{
    auto bitfields = rareDataBitfields();
    bitfields.tabIndexState = static_cast<uint16_t>(state);
    setRareDataBitfields(bitfields);
}

void Element::setTabIndexExplicitly(Optional<int> tabIndex)
{
    if (!tabIndex) {
        setTabIndexState(TabIndexState::NotSet);
        return;
    }
    switch (tabIndex.value()) {
	case 0:
		setTabIndexState(TabIndexState::Zero);
		break;
	case -1:
		setTabIndexState(TabIndexState::NegativeOne);
		break;
	default:
		ensureElementRareData().setUnusualTabIndex(tabIndex.value());
		setTabIndexState(TabIndexState::InRareData);
		break;
	}
}

Optional<int> Element::tabIndexSetExplicitly() const
{
    switch (tabIndexState()) {
    case TabIndexState::NotSet:
        return WTF::Nullopt;
    case TabIndexState::Zero:
        return 0;
    case TabIndexState::NegativeOne:
        return -1;
    case TabIndexState::InRareData:
        ASSERT(hasRareData());
        return elementRareData()->unusualTabIndex();
    }
    ASSERT_NOT_REACHED();
    return WTF::Nullopt;
}

int Element::defaultTabIndex() const
{
    return -1;
}

bool Element::isNonceable() const
{
    // https://www.w3.org/TR/CSP3/#is-element-nonceable
    if (elementRareData()->nonce().isNull())
        return false;

    if (hasDuplicateAttribute())
        return false;

    if (hasAttributes()
        && (isHTMLScriptElement(*this) || isSVGScriptElement(*this))) {
        static const char scriptString[] = "<script";
        static const char styleString[] = "<style";

        for (const auto& attribute : attributesIterator()) {
            auto name = attribute.localName().convertToASCIILowercase();
            auto value = attribute.value().convertToASCIILowercase();
            if (name.contains(scriptString)
                || name.contains(styleString)
                || value.contains(scriptString)
                || value.contains(styleString))
                return false;
        }
    }

    return true;
}

const AtomString& Element::nonce() const
{
    if (hasRareData() && isNonceable())
        return elementRareData()->nonce();

    return emptyAtom;
}

void Element::setNonce(const AtomString& newValue)
{
    if (newValue == emptyAtom && !hasRareData())
        return;

    ensureElementRareData().setNonce(newValue);
}

void Element::hideNonce()
{
    // https://html.spec.whatwg.org/multipage/urls-and-fetching.html#nonce-attributes
    if (!isConnected())
        return;

    const auto* csp = document().contentSecurityPolicy();
    if (!csp->isHeaderDelivered())
        return;

    // Retain previous IDL nonce.
    AtomString currentNonce = nonce();

    if (!getAttribute(nonceAttr).isEmpty())
        setAttribute(nonceAttr, emptyAtom);

    setNonce(currentNonce);
}

bool Element::supportsFocus() const
{
    return !!tabIndexSetExplicitly();
}

bool Element::isSpatialNavigationFocusable() const
{
	// https://developer.mozilla.org/en-US/docs/Web/HTML/Global_attributes/tabindex
	// A negative value (usually tabindex="-1") means that the element should be focusable,
	// but should not be reachable via sequential keyboard navigation.
	// It's mostly useful to create accessible widgets with JavaScript.
	if (auto tabIndexSetExplicitly = this->tabIndexSetExplicitly())
		return tabIndexSetExplicitly.value() >= 0;
	return true;
}

int Element::tabIndexForBindings() const
{
    //return valueOrCompute(tabIndexSetExplicitly(), [&] { return defaultTabIndex(); });
	if (auto tabIndexSetExplicitly = this->tabIndexSetExplicitly())
		return tabIndexSetExplicitly.value();
	return defaultTabIndex();
}

void Element::setTabIndexForBindings(int value)
{
    setIntegralAttribute(tabindexAttr, value);
}

bool Element::isKeyboardFocusable(KeyboardEvent*) const
{
    if (!(isFocusable() && !shouldBeIgnoredInSequentialFocusNavigation() && tabIndexSetExplicitly().valueOr(0) >= 0))
        return false;
    if (auto* root = shadowRoot()) {
        if (root->delegatesFocus())
            return false;
    }
    return true;
}

bool Element::isMouseFocusable() const
{
    return isFocusable();
}

#if 0
bool Element::shouldUseInputMethod()
{
    return computeEditability(UserSelectAllIsAlwaysNonEditable, ShouldUpdateStyle::Update) != Editability::ReadOnly;
}
#endif

bool Element::dispatchMouseEvent(const PlatformMouseEvent& platformEvent, const AtomicString& eventType, int detail, Element* relatedTarget)
{
    if (isDisabledFormControl())
        return false;

    Ref<MouseEvent> mouseEvent = MouseEvent::create(eventType, document().defaultView(), platformEvent, detail, relatedTarget);

    if (mouseEvent->type().isEmpty())
        return true; // Shouldn't happen.

    Ref<Element> protectedThis(*this);
    bool didNotSwallowEvent = true;

#if ENABLE(POINTER_EVENTS)
    if (RuntimeEnabledFeatures::sharedFeatures().pointerEventsEnabled()) {
        if (auto pointerEvent = PointerEvent::create(mouseEvent)) {
            if (auto* page = document().page())
                page->pointerCaptureController().dispatchEvent(*pointerEvent, this);
            if (pointerEvent->defaultPrevented() || pointerEvent->defaultHandled()) {
                didNotSwallowEvent = false;
                if (pointerEvent->type() == eventNames().pointerdownEvent || pointerEvent->type() == eventNames().pointerupEvent)
                    return false;
            }
        }
    }
#endif

    ASSERT(!mouseEvent->target() || mouseEvent->target() != relatedTarget);
    dispatchEvent(mouseEvent.get());

    if (mouseEvent->defaultPrevented())
    	platformEvent.preventDefault();

    if (mouseEvent->defaultHandled())
    	platformEvent.setDefaultHandled();

    if (mouseEvent->defaultPrevented() || mouseEvent->defaultHandled())
        didNotSwallowEvent = false;

    auto& eventNames = WebCore::eventNames();
    if (mouseEvent->type() == eventNames.clickEvent && mouseEvent->detail() == 2) {
        // Special case: If it's a double click event, we also send the dblclick event. This is not part
        // of the DOM specs, but is used for compatibility with the ondblclick="" attribute. This is treated
        // as a separate event in other DOM-compliant browsers like Firefox, and so we do the same.
        Ref<MouseEvent> doubleClickEvent = MouseEvent::create(eventNames.dblclickEvent,
            mouseEvent->bubbles(),
            mouseEvent->cancelable(),
            Event::IsComposed::Yes,
            mouseEvent->view(), mouseEvent->detail(),
            mouseEvent->screenX(), mouseEvent->screenY(), mouseEvent->clientX(), mouseEvent->clientY(),
            mouseEvent->ctrlKey(), mouseEvent->altKey(), mouseEvent->shiftKey(), mouseEvent->metaKey(),
            mouseEvent->button(), mouseEvent->buttons(), relatedTarget);

        if (mouseEvent->defaultHandled())
            doubleClickEvent->setDefaultHandled();

        dispatchEvent(doubleClickEvent.get());

        if (doubleClickEvent->defaultPrevented())
        	platformEvent.preventDefault();

        if (doubleClickEvent->defaultHandled())
        	platformEvent.setDefaultHandled();

        if (doubleClickEvent->defaultHandled() || doubleClickEvent->defaultPrevented())
            return false;
    }
    return didNotSwallowEvent;
}

bool Element::dispatchSimulatedClick(Event* underlyingEvent, SimulatedClickMouseEventOptions eventOptions, SimulatedClickVisualOptions visualOptions)
{
    return simulateClick(*this, underlyingEvent, eventOptions, visualOptions, SimulatedClickSource::UserAgent);
}

Ref<Node> Element::cloneNodeInternal(Document& targetDocument, CloningOperation type)
{
    switch (type) {
    case OnlySelf:
    case SelfWithTemplateContent:
        return cloneElementWithoutChildren(targetDocument);
    case Everything:
    	break;
    }
	return cloneElementWithChildren(targetDocument);
}

Ref<Element> Element::cloneElementWithChildren(Document& targetDocument)
{
	Ref<Element> clone = cloneElementWithoutChildren(targetDocument);
    cloneChildNodes(clone);
    return clone;
}

Ref<Element> Element::cloneElementWithoutChildren(Document& targetDocument)
{
    Ref<Element> clone = cloneElementWithoutAttributesAndChildren(targetDocument);
    // This will catch HTML elements in the wrong namespace that are not correctly copied.
    // This is a sanity check as HTML overloads some of the DOM methods.
    ASSERT(isHTMLElement() == clone->isHTMLElement());

    clone->cloneDataFromElement(*this);
    return clone;
}

Ref<Element> Element::cloneElementWithoutAttributesAndChildren(Document& targetDocument)
{
    return targetDocument.createElement(tagQName(), false);
}

PassRefPtr<Attr> Element::detachAttribute(unsigned index)
{
    ASSERT(elementData());

    const Attribute& attribute = elementData()->attributeAt(index);

    RefPtr<Attr> attrNode = attrIfExists(attribute.name());
    if (attrNode)
        detachAttrNodeFromElementWithValue(attrNode.get(), attribute.value());
    else
        attrNode = Attr::create(document(), attribute.name(), attribute.value());

    removeAttributeInternal(index, NotInSynchronizationOfLazyAttribute);
    return attrNode.release();
}

bool Element::removeAttribute(const QualifiedName& name)
{
    if (!elementData())
        return false;

    unsigned index = elementData()->findAttributeIndexByName(name);
    if (index == ElementData::attributeNotFound)
		return false;

    removeAttributeInternal(index, NotInSynchronizationOfLazyAttribute);
    return true;
}

void Element::setBooleanAttribute(const QualifiedName& name, bool value)
{
    if (value)
        setAttribute(name, emptyAtom);
    else
        removeAttribute(name);
}

NamedNodeMap* Element::attributes() const
{
    ElementRareData& rareData = const_cast<Element*>(this)->ensureElementRareData();
    if (NamedNodeMap* attributeMap = rareData.attributeMap())
        return attributeMap;

    rareData.setAttributeMap(NamedNodeMap::create(const_cast<Element*>(this)));
    return rareData.attributeMap();
}

Node::NodeType Element::nodeType() const
{
    return ELEMENT_NODE;
}

bool Element::hasAttribute(const QualifiedName& name) const
{
    return hasAttributeNS(name.namespaceURI(), name.localName());
}

void Element::synchronizeAllAttributes() const
{
    if (!elementData())
    	return;
    if (elementData()->m_styleAttributeIsDirty) {
        ASSERT(isStyledElement());
        static_cast<const StyledElement*>(this)->synchronizeStyleAttributeInternal();
    }
#if ENABLE(SVG)
    if (isSVGElement())
        toSVGElement(const_cast<Element&>(*this)).synchronizeAllAttributes();
#endif
}

ALWAYS_INLINE void Element::synchronizeAttribute(const QualifiedName& name) const
{
    if (!elementData())
    	return;
    if (UNLIKELY(name == styleAttr && elementData()->m_styleAttributeIsDirty)) {
    	ASSERT(isStyledElement());
		static_cast<const StyledElement*>(this)->synchronizeStyleAttributeInternal();
		return;
    }
#if ENABLE(SVG)
    if (isSVGElement())
        toSVGElement(const_cast<Element&>(*this)).synchronizeAttribute(name);
#endif
}

static ALWAYS_INLINE bool isStyleAttribute(const Element& element, const AtomicString& attributeLocalName)
{
    if (shouldIgnoreAttributeCase(element))
        return equalLettersIgnoringASCIICase(attributeLocalName, "style");
    return attributeLocalName == styleAttr.localName();
}

ALWAYS_INLINE void Element::synchronizeAttribute(const AtomicString& localName) const
{
    // This version of synchronizeAttribute() is streamlined for the case where you don't have a full QualifiedName,
    // e.g when called from DOM API.
    if (!elementData())
    	return;
    // FIXME: this should be comparing in the ASCII range.
    if (elementData()->m_styleAttributeIsDirty && isStyleAttribute(*this, localName)) {
    	ASSERT(isStyledElement());
		static_cast<const StyledElement*>(this)->synchronizeStyleAttributeInternal();
		return;
    }
#if ENABLE(SVG)
    if (isSVGElement())
        toSVGElement(const_cast<Element&>(*this)).synchronizeAttribute(QualifiedName(nullAtom, localName, nullAtom));
#endif
}

const AtomicString& Element::getAttribute(const QualifiedName& name) const
{
	if (!elementData())
		return nullAtom;
    synchronizeAttribute(name);
    if (const Attribute* attribute = findAttributeByName(name))
		return attribute->value();
    return nullAtom;
}

Vector<String> Element::getAttributeNames() const
{
    Vector<String> attributesVector;
    if (!hasAttributes())
        return attributesVector;

    auto attributes = attributesIterator();
    attributesVector.reserveInitialCapacity(attributes.attributeCount());
    for(auto iter = attributes.begin(); iter != attributes.end(); ++iter)
        attributesVector.uncheckedAppend((*iter).name().toString());
    return attributesVector;
}

bool Element::hasFocusableStyle() const
{
    if (renderer() && renderer()->isSkippedContent())
        return false;

    auto isFocusableStyle = [](const RenderStyle* style) {
        return style && style->display() != DisplayType::None && style->display() != DisplayType::Contents
            && style->visibility() == EVisibility::VISIBLE && !style->effectiveInert();
    };

    if (auto* renderStyle = this->renderStyle())
        return isFocusableStyle(renderStyle);

    // Compute style in yet unstyled subtree without resolving full style.
    auto* style = const_cast<Element&>(*this).resolveComputedStyle(ResolveComputedStyleMode::RenderedOnly);
    return isFocusableStyle(style);
}

bool Element::isFocusable() const
{
    if (!isConnected() || !supportsFocus())
        return false;

    if (!renderer()) {
        // Elements in canvas fallback content are not rendered, but they are allowed to be
        // focusable as long as their canvas is displayed and visible.
    	RefPtr<const HTMLCanvasElement> canvas = ancestorsOfType<HTMLCanvasElement>(*this).first();
		if (canvas && !canvas->hasFocusableStyle())
			return false;
	}

    return hasFocusableStyle();
}

bool Element::isUserActionElementInActiveChain() const
{
    ASSERT(isUserActionElement());
    return document().userActionElements().isInActiveChain(*this);
}

bool Element::isUserActionElementActive() const
{
    ASSERT(isUserActionElement());
    return document().userActionElements().isActive(*this);
}

bool Element::isUserActionElementFocused() const
{
    ASSERT(isUserActionElement());
    return document().userActionElements().isFocused(*this);
}

bool Element::isUserActionElementHovered() const
{
    ASSERT(isUserActionElement());
    return document().userActionElements().isHovered(*this);
}

bool Element::isUserActionElementDragged() const
{
    ASSERT(isUserActionElement());
    return document().userActionElements().isBeingDragged(*this);
}

bool Element::isUserActionElementHasFocusVisible() const
{
    ASSERT(isUserActionElement());
    return document().userActionElements().hasFocusVisible(*this);
}

bool Element::isUserActionElementHasFocusWithin() const
{
    ASSERT(isUserActionElement());
    return document().userActionElements().hasFocusWithin(*this);
}

void Element::setActive(bool flag, bool pause)
{
    if (flag == active())
        return;
    {
        Style::PseudoClassChangeInvalidation styleInvalidation(*this, CSSSelector::PseudoClassActive);
        document().userActionElements().setActive(*this, flag);
    }

    if (!renderer())
        return;

    bool reactsToPress = false;
    if (renderer()->style()->hasEffectiveAppearance() && renderer()->theme()->stateChanged(renderer(), PressedState))
        reactsToPress = true;

    // The rest of this function implements a feature that only works if the
    // platform supports immediate invalidations on the ChromeClient, so bail if
    // that isn't supported.
    if (!document().page()->chrome()->client()->supportsImmediateInvalidation())
        return;

    if (reactsToPress && pause) {
        // The delay here is subtle. It relies on an assumption, namely that the amount of time it takes
        // to repaint the "down" state of the control is about the same time as it would take to repaint the
        // "up" state. Once you assume this, you can just delay for 100ms - that time (assuming that after you
        // leave this method, it will be about that long before the flush of the up state happens again).
#ifdef HAVE_FUNC_USLEEP
        double startTime = currentTime();
#endif

        document().updateStyleIfNeeded();

        // Do an immediate repaint.
        if (renderer())
            renderer()->repaint();

        // FIXME: Come up with a less ridiculous way of doing this.
#ifdef HAVE_FUNC_USLEEP
        // Now pause for a small amount of time (1/10th of a second from before we repainted in the pressed state)
        double remainingTime = 0.1 - (currentTime() - startTime);
        if (remainingTime > 0)
            usleep(static_cast<useconds_t>(remainingTime * 1000000.0));
#endif
    }
}

static bool shouldAlwaysHaveFocusVisibleWhenFocused(const Element& element)
{
    return element.isTextField() || element.isContentEditable() || isHTMLSelectElement(element);
}

void Element::setFocus(bool flag, FocusVisibility visibility)
{
    if (flag == focused())
        return;
    {
        Style::PseudoClassChangeInvalidation focusStyleInvalidation(*this, CSSSelector::PseudoClassFocus);
        Style::PseudoClassChangeInvalidation focusVisibleStyleInvalidation(*this, CSSSelector::PseudoClassFocusVisible);
        document().userActionElements().setFocused(*this, flag);

		// Shadow host with a slot that contain focused element is not considered focused.
		for (auto* root = containingShadowRoot(); root; root = root->host()->containingShadowRoot()) {
			root->setContainsFocusedElement(flag);
			root->host()->invalidateStyle();
		}

		for (Element* element = this; element; element = element->parentElementInComposedTree())
			element->setHasFocusWithin(flag);

		setHasFocusVisible(flag && (visibility == FocusVisibility::Visible || shouldAlwaysHaveFocusVisibleWhenFocused(*this)));
    }
}

void Element::setHasFocusVisible(bool flag)
{
    if (!document().settings().focusVisibleEnabled())
        return;

#if !ASSERT_DISABLED
    ASSERT(!flag || focused());
    ASSERT(!focused() || !shouldAlwaysHaveFocusVisibleWhenFocused(*this) || flag);
#endif

    if (hasFocusVisible() == flag)
        return;

    document().userActionElements().setHasFocusVisible(*this, flag);
}

void Element::setHasFocusWithin(bool flag)
{
    if (hasFocusWithin() == flag)
        return;
    {
        Style::PseudoClassChangeInvalidation styleInvalidation(*this, CSSSelector::PseudoClassFocusWithin);
        document().userActionElements().setHasFocusWithin(*this, flag);
    }
}

void Element::setHasTentativeFocus(bool value)
{
    // Tentative focus is used when trying to set the focus on a new element.
    for (auto& ancestor : composedTreeAncestors(*this)) {
        ASSERT(ancestor.hasFocusWithin() != value);
        document().userActionElements().setHasFocusWithin(ancestor, value);
    }
}

void Element::setHovered(bool flag)
{
    if (flag == hovered())
        return;
    {
        Style::PseudoClassChangeInvalidation styleInvalidation(*this, CSSSelector::PseudoClassHover);
        document().userActionElements().setHovered(*this, flag);
    }

    auto* style = renderStyle();
    if ( style && style->hasEffectiveAppearance())
        renderer()->theme()->stateChanged(renderer(), HoverState);
}

void Element::setBeingDragged(bool flag)
{
    if (flag == isBeingDragged())
        return;

    Style::PseudoClassChangeInvalidation styleInvalidation(*this, CSSSelector::PseudoClassDrag);
    document().userActionElements().setBeingDragged(*this, flag);
}

#if 0
void Element::scrollIntoView(std::optional<std::variant<bool, ScrollIntoViewOptions>>&& arg)
{
    document().updateLayoutIgnorePendingStylesheets();

    if (!renderer())
        return;

    bool insideFixed;
    LayoutRect absoluteBounds = renderer()->absoluteAnchorRectWithScrollMargin(&insideFixed);

    ScrollIntoViewOptions options;
    if (arg) {
        auto value = arg.value();
        if (std::holds_alternative<ScrollIntoViewOptions>(value))
            options = std::get<ScrollIntoViewOptions>(value);
        else if (!std::get<bool>(value))
            options.blockPosition = ScrollLogicalPosition::End;
    }

    auto writingMode = renderer()->style().writingMode();
    ScrollAlignment alignX = toScrollAlignmentForInlineDirection(options.inlinePosition, writingMode, renderer()->style().isLeftToRightDirection());
    ScrollAlignment alignY = toScrollAlignmentForBlockDirection(options.blockPosition, writingMode);
    alignX.disableLegacyHorizontalVisibilityThreshold();

    bool isHorizontal = renderer()->style().isHorizontalWritingMode();
    ScrollRectToVisibleOptions visibleOptions {
        SelectionRevealMode::Reveal,
        isHorizontal ? alignX : alignY,
        isHorizontal ? alignY : alignX,
        ShouldAllowCrossOriginScrolling::No,
        options.behavior.value_or(ScrollBehavior::Auto)
    };
    renderer()->scrollRectToVisible(SelectionRevealMode::Reveal, absoluteBounds, insideFixed, visibleOptions);
}
#endif

void Element::scrollIntoView(bool alignToTop) 
{
    document().updateLayoutIgnorePendingStylesheets();

    if (!renderer())
        return;

    LayoutRect bounds = renderer()->anchorRect();

    // Align to the top / bottom and to the closest edge.
    ScrollAlignment alignY = alignToTop ? ScrollAlignment::alignTopAlways : ScrollAlignment::alignBottomAlways;
    ScrollAlignment alignX = ScrollAlignment::alignToEdgeIfNeeded;
    alignX.disableLegacyHorizontalVisibilityThreshold();

    renderer()->scrollRectToVisible(SelectionRevealMode::Reveal, bounds, alignX, alignY);
}

void Element::scrollIntoViewIfNeeded(bool centerIfNeeded)
{
    document().updateLayoutIgnorePendingStylesheets();

    if (!renderer())
        return;

    LayoutRect bounds = renderer()->anchorRect();

    ScrollAlignment alignY = centerIfNeeded ? ScrollAlignment::alignCenterIfNeeded : ScrollAlignment::alignToEdgeIfNeeded;
    ScrollAlignment alignX = centerIfNeeded ? ScrollAlignment::alignCenterIfNeeded : ScrollAlignment::alignToEdgeIfNeeded;
    alignX.disableLegacyHorizontalVisibilityThreshold();

    renderer()->scrollRectToVisible(SelectionRevealMode::Reveal, bounds, alignX, alignY);
}

void Element::scrollBy(const ScrollToOptions& options)
{
    return scrollBy(options.left.valueOr(0), options.top.valueOr(0));
}

static inline double normalizeNonFiniteValue(double f)
{
    return std::isfinite(f) ? f : 0;
}

void Element::scrollBy(double x, double y)
{
    scrollTo(scrollLeft() + normalizeNonFiniteValue(x), scrollTop() + normalizeNonFiniteValue(y));
}

void Element::scrollTo(const ScrollToOptions& options)
{
    if (!document().settings().CSSOMViewScrollingAPIEnabled()) {
		// If the element is the root element and document is in quirks mode, terminate these steps.
		// Note that WebKit always uses quirks mode document scrolling behavior. See Document::scrollingElement().
		if (this == document().documentElement())
			return;
    }

    document().updateLayoutIgnorePendingStylesheets();

    if (document().scrollingElement() == this) {
        // If the element is the scrolling element and is not potentially scrollable,
        // invoke scroll() on window with options as the only argument, and terminate these steps.
        // FIXME: Scrolling an independently scrollable body is broken: webkit.org/b/161612.
        RefPtr<DOMWindow> window = document().domWindow();
        if (!window)
            return;

        window->scrollTo(options);
        return;
    }

    // If the element does not have any associated CSS layout box, the element has no associated scrolling box,
    // or the element has no overflow, terminate these steps.
    RenderBox* renderer = renderBox();
    if (!renderer || !renderer->hasNonVisibleOverflow())
        return;

    // Normalize non-finite values for left and top dictionary members of options, if present.
    double x = options.left ? normalizeNonFiniteValue(options.left.value()) : adjustForAbsoluteZoom(renderer->scrollLeft(), *renderer);
    double y = options.top ? normalizeNonFiniteValue(options.top.value()) : adjustForAbsoluteZoom(renderer->scrollTop(), *renderer);

    renderer->setScrollLeft(clampToInteger(x * renderer->style()->effectiveZoom()));
    renderer->setScrollTop(clampToInteger(y * renderer->style()->effectiveZoom()));
}

void Element::scrollTo(double x, double y)
{
    scrollTo(ScrollToOptions { x, y });
}

void Element::scrollByUnits(int units, ScrollGranularity granularity)
{
    document().updateLayoutIgnorePendingStylesheets();

    if (!renderer())
        return;

    if (!renderer()->hasNonVisibleOverflow())
        return;

    ScrollDirection direction = ScrollDown;
    if (units < 0) {
        direction = ScrollUp;
        units = -units;
    }
    Node* stopNode = this;
    toRenderBox(renderer())->scroll(direction, granularity, units, &stopNode);
}

void Element::scrollByLines(int lines)
{
    scrollByUnits(lines, ScrollByLine);
}

void Element::scrollByPages(int pages)
{
    scrollByUnits(pages, ScrollByPage);
}

static double localZoomForRenderer(const RenderElement& renderer)
{
    // FIXME: This does the wrong thing if two opposing zooms are in effect and canceled each
    // other out, but the alternative is that we'd have to crawl up the whole render tree every
    // time (or store an additional bit in the RenderStyle to indicate that a zoom was specified).
	double zoomFactor = 1;
    if (renderer.style()->effectiveZoom() != 1) {
        // Need to find the nearest enclosing RenderElement that set up
        // a differing zoom, and then we divide our result by it to eliminate the zoom.
    	const RenderElement* prev = &renderer;
        for (RenderElement* curr = prev->parent(); curr; curr = curr->parent()) {
            if (curr->style()->effectiveZoom() != prev->style()->effectiveZoom()) {
                zoomFactor = prev->style()->zoom();
                break;
            }
            prev = curr;
        }
        if (prev->isRenderView())
            zoomFactor = prev->style()->zoom();
    }
    return zoomFactor;
}

static int adjustContentsScrollPositionOrSizeForZoom(int value, const Frame& frame)
{
    double zoomFactor = frame.pageZoomFactor() * frame.frameScaleFactor();
    if (zoomFactor == 1)
        return value;
    // FIXME (webkit.org/b/189397): Why can't we just ceil/floor?
    // Needed because of truncation (rather than rounding) when scaling up.
    if (zoomFactor > 1)
        value++;
    return static_cast<int>(value / zoomFactor);
}

enum LegacyCSSOMElementMetricsRoundingStrategy { Round, Floor };

static int convertToNonSubpixelValue(double value, const LegacyCSSOMElementMetricsRoundingStrategy roundStrategy = Round)
{
    return roundStrategy == Round ? round(value) : floor(value);
}

static int adjustOffsetForZoomAndSubpixelLayout(RenderBoxModelObject& renderer, const LayoutUnit& offset)
{
    auto offsetLeft = LayoutUnit { roundToInt(offset) };
    double zoomFactor = localZoomForRenderer(renderer);
    if (zoomFactor == 1)
        return convertToNonSubpixelValue(offsetLeft, Floor);
    return convertToNonSubpixelValue(offsetLeft / zoomFactor, Round);
}
static HashSet<TreeScope*> collectAncestorTreeScopeAsHashSet(Node& node)
{
    HashSet<TreeScope*> ancestors;
    for (auto* currentScope = &node.treeScope(); currentScope; currentScope = currentScope->parentTreeScope())
        ancestors.add(currentScope);
    return ancestors;
}

int Element::offsetLeftForBindings()
{
    auto offset = offsetLeft();

    RefPtr<Element> parent = offsetParent();
    if (!parent || !parent->isInShadowTree())
        return offset;

    ASSERT(&parent->document() == &document());
    if (&parent->treeScope() == &treeScope())
        return offset;

    auto ancestorTreeScopes = collectAncestorTreeScopeAsHashSet(*this);
    while (parent && !ancestorTreeScopes.contains(&parent->treeScope())) {
        offset += parent->offsetLeft();
        parent = parent->offsetParent();
    }

    return offset;
}

int Element::offsetLeft()
{
    document().updateLayoutIgnorePendingStylesheets();
    if (RenderBoxModelObject* renderer = renderBoxModelObject())
        return adjustOffsetForZoomAndSubpixelLayout(*renderer, renderer->offsetLeft());
    return 0;
}

int Element::offsetTopForBindings()
{
    auto offset = offsetTop();

    RefPtr<Element> parent = offsetParent();
    if (!parent || !parent->isInShadowTree())
        return offset;

    ASSERT(&parent->document() == &document());
    if (&parent->treeScope() == &treeScope())
        return offset;

    auto ancestorTreeScopes = collectAncestorTreeScopeAsHashSet(*this);
    while (parent && !ancestorTreeScopes.contains(&parent->treeScope())) {
        offset += parent->offsetTop();
        parent = parent->offsetParent();
    }

    return offset;
}

int Element::offsetTop()
{
	document().updateLayoutIgnorePendingStylesheets();
    if (RenderBoxModelObject* renderer = renderBoxModelObject())
        return adjustOffsetForZoomAndSubpixelLayout(*renderer, renderer->offsetTop());
    return 0;
}

int Element::offsetWidth()
{
    document().updateLayoutIfDimensionsOutOfDate(*this, WidthDimensionsCheck);
    if (RenderBoxModelObject* renderer = renderBoxModelObject()) {
    	LayoutUnit offsetWidth = LayoutUnit(roundToInt(renderer->offsetWidth()));
        return convertToNonSubpixelValue(adjustLayoutUnitForAbsoluteZoom(offsetWidth, *renderer).toDouble());
    }
    return 0;
}

int Element::offsetHeight()
{
    document().updateLayoutIfDimensionsOutOfDate(*this, HeightDimensionsCheck);
    if (RenderBoxModelObject* renderer = renderBoxModelObject()) {
    	LayoutUnit offsetHeight = LayoutUnit(roundToInt(renderer->offsetHeight()));
        return convertToNonSubpixelValue(adjustLayoutUnitForAbsoluteZoom(offsetHeight, *renderer).toDouble());
    }
    return 0;
}

Element* Element::offsetParentForBindings()
{
    Element* element = offsetParent();
    if (!element || !element->isInShadowTree())
        return element;
    while (element && !isDescendantOrShadowDescendantOf(element->rootNode()))
        element = element->offsetParent();
    return element;
}

Element* Element::offsetParent()
{
    document().updateLayoutIgnorePendingStylesheets();
    auto* renderer = this->renderer();
    if (!renderer)
        return 0;
    auto* offsetParent = renderer->offsetParent();
    if (!offsetParent)
        return 0;
    return offsetParent->element();
}

int Element::clientLeft()
{
    document().updateLayoutIgnorePendingStylesheets();

    if (auto* renderer = renderBox()) {
    	LayoutUnit clientLeft = LayoutUnit(roundToInt(renderer->clientLeft()));
        return convertToNonSubpixelValue(adjustLayoutUnitForAbsoluteZoom(clientLeft, *renderer).toDouble());
    }
	return 0;
}

int Element::clientTop()
{
    document().updateLayoutIgnorePendingStylesheets();

    if (auto* renderer = renderBox())  {
    	LayoutUnit clientTop = LayoutUnit(roundToInt(renderer->clientTop()));
        return convertToNonSubpixelValue(adjustLayoutUnitForAbsoluteZoom(clientTop, *renderer).toDouble());
    }
    return 0;
}

int Element::clientWidth()
{
    document().updateLayoutIfDimensionsOutOfDate(*this, WidthDimensionsCheck);

    if (!document().hasLivingRenderTree())
        return 0;

    RenderView& renderView = *document().renderView();

    // When in strict mode, clientWidth for the document element should return the width of the containing frame.
    // When in quirks mode, clientWidth for the body element should return the width of the containing frame.
    bool inQuirksMode = document().inQuirksMode();
    if ((!inQuirksMode && document().documentElement() == this) ||
        (inQuirksMode && isHTMLElement() && document().bodyOrFrameset() == this)) {
    	return adjustForAbsoluteZoom(renderView.frameView().layoutWidth(), &renderView);
    }
    
    if (RenderBox* renderer = renderBox()) {
    	LayoutUnit clientWidth = LayoutUnit(roundToInt(renderer->clientWidth()));
        // clientWidth/Height is the visual portion of the box content, not including
        // borders or scroll bars, but includes padding. And per
        // https://www.w3.org/TR/CSS2/tables.html#model,
        // table wrapper box is a principal block box that contains the table box
        // itself and any caption boxes, and table grid box is a block-level box that
        // contains the table's internal table boxes. When table's border is specified
        // in CSS, the border is added to table grid box, not table wrapper box.
        // Currently, WebKit doesn't have table wrapper box, and we are supposed to
        // retrieve clientWidth/Height from table wrapper box, not table grid box. So
        // when we retrieve clientWidth/Height, it includes table's border size.
        if (renderer->isTable())
            clientWidth += renderer->borderLeft() + renderer->borderRight();
        return convertToNonSubpixelValue(adjustLayoutUnitForAbsoluteZoom(clientWidth, *renderer).toDouble());
    }
    return 0;
}

int Element::clientHeight()
{
    document().updateLayoutIfDimensionsOutOfDate(*this, HeightDimensionsCheck);
    if (!document().hasLivingRenderTree())
        return 0;

    RenderView& renderView = *document().renderView();

    // When in strict mode, clientHeight for the document element should return the height of the containing frame.
    // When in quirks mode, clientHeight for the body element should return the height of the containing frame.
    bool inQuirksMode = document().inQuirksMode();

    if ((!inQuirksMode && document().documentElement() == this) ||
        (inQuirksMode && isHTMLElement() && document().bodyOrFrameset() == this)) {
    	return adjustForAbsoluteZoom(renderView.frameView().layoutHeight(), &renderView);
    }
    
    if (RenderBox* renderer = renderBox()) {
    	LayoutUnit clientHeight = LayoutUnit(roundToInt(renderer->clientHeight()));
        // clientWidth/Height is the visual portion of the box content, not including
        // borders or scroll bars, but includes padding. And per
        // https://www.w3.org/TR/CSS2/tables.html#model,
        // table wrapper box is a principal block box that contains the table box
        // itself and any caption boxes, and table grid box is a block-level box that
        // contains the table's internal table boxes. When table's border is specified
        // in CSS, the border is added to table grid box, not table wrapper box.
        // Currently, WebKit doesn't have table wrapper box, and we are supposed to
        // retrieve clientWidth/Height from table wrapper box, not table grid box. So
        // when we retrieve clientWidth/Height, it includes table's border size.
        if (renderer->isTable())
            clientHeight += renderer->borderTop() + renderer->borderBottom();
        return convertToNonSubpixelValue(adjustLayoutUnitForAbsoluteZoom(clientHeight, *renderer).toDouble());
    }
    return 0;
}

ALWAYS_INLINE Frame* Element::documentFrameWithNonNullView() const
{
    auto* frame = document().frame();
    return frame && frame->view() ? frame : nullptr;
}

int Element::scrollLeft()
{
    document().updateLayoutIgnorePendingStylesheets();

    if (document().scrollingElement() == this) {
        if (auto* frame = documentFrameWithNonNullView())
            return adjustContentsScrollPositionOrSizeForZoom(frame->view()->contentsScrollPosition().x(), *frame);
        return 0;
    }

    if (RenderBox* renderer = renderBox())
        return adjustForAbsoluteZoom(renderer->scrollLeft(), renderer);
    return 0;
}

int Element::scrollTop()
{
    document().updateLayoutIgnorePendingStylesheets();

    if (document().scrollingElement() == this) {
        if (auto* frame = documentFrameWithNonNullView())
            return adjustContentsScrollPositionOrSizeForZoom(frame->view()->contentsScrollPosition().y(), *frame);
        return 0;
    }

    if (RenderBox* renderer = renderBox())
        return adjustForAbsoluteZoom(renderer->scrollTop(), renderer);
    return 0;
}

void Element::setScrollLeft(int newLeft)
{
    document().updateLayoutIgnorePendingStylesheets();

    if (document().scrollingElement() == this) {
        if (auto* frame = documentFrameWithNonNullView())
            frame->view()->setScrollPosition(IntPoint(static_cast<int>(newLeft * frame->pageZoomFactor() * frame->frameScaleFactor()), frame->view()->scrollY()));
        return;
    }

    if (auto* renderer = renderBox())
    	renderer->setScrollLeft(static_cast<int>(newLeft * renderer->style()->effectiveZoom()));
}

void Element::setScrollTop(int newTop)
{
    document().updateLayoutIgnorePendingStylesheets();

    if (document().scrollingElement() == this) {
        if (auto* frame = documentFrameWithNonNullView())
            frame->view()->setScrollPosition(IntPoint(frame->view()->scrollX(), static_cast<int>(newTop * frame->pageZoomFactor() * frame->frameScaleFactor())));
        return;
    }

    if (RenderBox* renderer = renderBox())
        renderer->setScrollTop(static_cast<int>(newTop * renderer->style()->effectiveZoom()));
}

int Element::scrollWidth()
{
    document().updateLayoutIfDimensionsOutOfDate(*this, WidthDimensionsCheck);

    if (document().scrollingElement() == this) {
        // FIXME (webkit.org/b/182289): updateLayoutIfDimensionsOutOfDate seems to ignore zoom level change.
        document().updateLayoutIgnorePendingStylesheets();
        if (auto* frame = documentFrameWithNonNullView())
            return adjustContentsScrollPositionOrSizeForZoom(frame->view()->contentsWidth(), *frame);
        return 0;
    }

    if (auto* renderer = renderBox())
        return adjustForAbsoluteZoom(renderer->scrollWidth(), renderer);
    return 0;
}

int Element::scrollHeight()
{
    document().updateLayoutIfDimensionsOutOfDate(*this, HeightDimensionsCheck);

    if (document().scrollingElement() == this) {
        // FIXME (webkit.org/b/182289): updateLayoutIfDimensionsOutOfDate seems to ignore zoom level change.
        document().updateLayoutIgnorePendingStylesheets();
        if (auto* frame = documentFrameWithNonNullView())
            return adjustContentsScrollPositionOrSizeForZoom(frame->view()->contentsHeight(), *frame);
        return 0;
    }

    if (auto* renderer = renderBox())
        return adjustForAbsoluteZoom(renderer->scrollHeight(), renderer);
    return 0;
}

IntRect Element::boundsInRootViewSpace()
{
    document().updateLayoutIgnorePendingStylesheets();

    FrameView* view = document().view();
    if (!view)
        return IntRect();

    Vector<FloatQuad> quads;
    if (isSVGElement() && renderer()) {
        // Get the bounding rectangle from the SVG model.
        SVGElement& svgElement = toSVGElement(*this);
        if (auto localRect = svgElement.getBoundingBox())
            quads.append(renderer()->localToAbsoluteQuad(*localRect));
    } else {
        // Get the bounding rectangle from the box model.
        if (auto* renderBoxModelObject = this->renderBoxModelObject())
            renderBoxModelObject->absoluteQuads(quads);
    }

    return view->contentsToRootView(enclosingIntRect(unitedBoundingBoxes(quads)));
}

static bool layoutOverflowRectContainsAllDescendants(const RenderBox& renderBox)
{
    if (renderBox.isRenderView())
        return true;

    if (!renderBox.element())
        return false;

    // If there are any position:fixed inside of us, game over.
    if (auto* viewPositionedObjects = renderBox.view().positionedObjects()) {
        for (auto* positionedBox : *viewPositionedObjects) {
            if (positionedBox == &renderBox)
                continue;
            if (positionedBox->isFixedPositioned() && renderBox.element()->contains(positionedBox->element()))
                return false;
        }
    }

    if (renderBox.canContainAbsolutelyPositionedObjects()) {
        // Our layout overflow will include all descendant positioned elements.
        return true;
    }

    // This renderer may have positioned descendants whose containing block is some ancestor.
    if (auto* containingBlock = RenderObject::containingBlockForPositionType(PositionType::Absolute, renderBox)) {
		if (auto* positionedObjects = containingBlock->positionedObjects()) {
            for (auto* positionedBox : *positionedObjects) {
                if (positionedBox == &renderBox)
                    continue;
                if (renderBox.element()->contains(positionedBox->element()))
                    return false;
            }
        }
    }
    return false;
}

static Optional<std::pair<RenderElement*, LayoutRect>> listBoxElementBoundingBox(Element& element)
{
    HTMLSelectElement* selectElement;
    bool isGroup;
    if (isHTMLOptionElement(element)) {
        selectElement = toHTMLOptionElement(element).ownerSelectElement();
        isGroup = false;
    } else if (isHTMLOptGroupElement(element)) {
        selectElement = toHTMLOptGroupElement(element).ownerSelectElement();
        isGroup = true;
    } else
        return Nullopt;

    if (!selectElement || !selectElement->renderer() || !selectElement->renderer()->isListBox())
        return Nullopt;

    auto& renderer = *toRenderListBox(selectElement->renderer());
    Optional<LayoutRect> boundingBox;
    int optionIndex = 0;
    for (auto& item : selectElement->listItems()) {
        if (item == &element) {
            LayoutPoint additionOffset;
            boundingBox = renderer.itemBoundingBoxRect(additionOffset, optionIndex);
            if (!isGroup)
                break;
        } else if (isGroup && boundingBox) {
            if (item->parentNode() != &element)
                break;
            LayoutPoint additionOffset;
            boundingBox->setHeight(boundingBox->height() + renderer.itemBoundingBoxRect(additionOffset, optionIndex).height());
        }
        ++optionIndex;
    }

    if (!boundingBox)
        return Nullopt;

    return std::pair<RenderElement*, LayoutRect> { &renderer, boundingBox.value() };
}

PassRefPtr<DOMRectList> Element::getClientRects()
{
    document().updateLayoutIgnorePendingStylesheets();

    auto* renderer = this->renderer();
    Vector<FloatQuad> quads;

    if (auto pair = listBoxElementBoundingBox(*this)) {
        renderer = pair.value().first;
        quads.append(renderer->localToAbsoluteQuad(FloatQuad(pair.value().second)));
    } else if (auto* renderBoxModelObject = this->renderBoxModelObject())
        renderBoxModelObject->absoluteQuads(quads);

    // FIXME: Handle SVG elements.
    // FIXME: Handle table/inline-table with a caption.

    if (quads.isEmpty())
        return DOMRectList::create();

    document().convertAbsoluteToClientQuads(quads, *renderer->style());
    return DOMRectList::create(quads);
}

Optional<std::pair<RenderElement*, FloatRect>> Element::boundingAbsoluteRectWithoutLayout()
{
	auto* renderer = this->renderer();
	Vector<FloatQuad> quads;
	if (isSVGElement() && renderer && !renderer->isSVGRoot()) {
		// Get the bounding rectangle from the SVG model.
		SVGElement& svgElement = toSVGElement(*this);
        if (auto localRect = svgElement.getBoundingBox())
            quads.append(renderer->localToAbsoluteQuad(*localRect));
    } else if (auto pair = listBoxElementBoundingBox(*this)) {
        renderer = pair.value().first;
        quads.append(renderer->localToAbsoluteQuad(FloatQuad(pair.value().second)));
    } else if (auto* renderBoxModelObject = this->renderBoxModelObject())
        renderBoxModelObject->absoluteQuads(quads);

	if (quads.isEmpty())
        return WTF::Nullopt;

    return std::make_pair(renderer, unitedBoundingBoxes(quads));
}

FloatRect Element::boundingClientRect()
{
    document().updateLayoutIgnorePendingStylesheets();
    auto pair = boundingAbsoluteRectWithoutLayout();
    if (!pair)
        return FloatRect();

    RenderElement* renderer = pair->first;
    FloatRect result = pair->second;

    document().convertAbsoluteToClientRect(result, *renderer->style());
    return result;
}

PassRefPtr<DOMRect> Element::getBoundingClientRect()
{
    return DOMRect::create(boundingClientRect());
}

// Note that this is not web-exposed, and does not use the same coordinate system as getBoundingClientRect() and friends.
IntRect Element::clientRect() const
{
    if (auto* renderer = this->renderer())
        return document().view()->contentsToRootView(renderer->absoluteBoundingBoxRect());
    return IntRect();
}

IntRect Element::screenRect() const
{
    if (auto* renderer = this->renderer())
        return document().view()->contentsToScreen(renderer->absoluteBoundingBoxRect());
    return IntRect();
}

DLL_PUBLIC const AtomicString& Element::getAttribute(const AtomicString& localName) const
{
	if (!elementData())
		return nullAtom;
    synchronizeAttribute(localName);
	if (const Attribute* attribute = elementData()->findAttributeByName(localName, shouldIgnoreAttributeCase(*this)))
		return attribute->value();
    return nullAtom;
}

const AtomicString& Element::getAttribute(const char* name) const
{
	return getAttribute(AtomicString(name));
}

const AtomicString& Element::getAttributeNS(const AtomicString& namespaceURI, const AtomicString& localName) const
{
    return getAttribute(QualifiedName(nullAtom, localName, namespaceURI));
}

void Element::setAttribute(const AtomicString& localName, const AtomicString& value, ExceptionCode& ec)
{
    if (!Document::isValidName(localName)) {
        ec = INVALID_CHARACTER_ERR;
        return;
    }

    synchronizeAttribute(localName);
    AtomicString caseAdjustedLocalName = shouldIgnoreAttributeCase(*this) ? localName.convertToASCIILowercase() : localName;

    unsigned index = elementData() ? elementData()->findAttributeIndexByName(caseAdjustedLocalName, false) : ElementData::attributeNotFound;
    const QualifiedName& qName = index != ElementData::attributeNotFound ? attributeAt(index).name() : QualifiedName(nullAtom, caseAdjustedLocalName, nullAtom);
    setAttributeInternal(index, qName, value, NotInSynchronizationOfLazyAttribute);
}

void Element::setAttribute(const QualifiedName& name, const AtomicString& value)
{
    synchronizeAttribute(name);
    unsigned index = elementData() ? elementData()->findAttributeIndexByName(name) : ElementData::attributeNotFound;
    setAttributeInternal(index, name, value, NotInSynchronizationOfLazyAttribute);
}

void Element::setAttributeWithoutSynchronization(const QualifiedName& name, const AtomicString& value)
{
	unsigned index = elementData() ? elementData()->findAttributeIndexByName(name) : ElementData::attributeNotFound;
    setAttributeInternal(index, name, value, NotInSynchronizationOfLazyAttribute);
}

void Element::setSynchronizedLazyAttribute(const QualifiedName& name, const AtomicString& value)
{
    unsigned index = elementData() ? elementData()->findAttributeIndexByName(name) : ElementData::attributeNotFound;
    setAttributeInternal(index, name, value, InSynchronizationOfLazyAttribute);
}

inline void Element::setAttributeInternal(unsigned index, const QualifiedName& name, const AtomicString& newValue, SynchronizationOfLazyAttribute inSynchronizationOfLazyAttribute)
{
    ASSERT_WITH_MESSAGE(refCount() || parentNode(), "Attribute must not be set on an element before adoptRef");

    if (newValue.isNull()) {
        if (index != ElementData::attributeNotFound)
        	removeAttributeInternal(index, inSynchronizationOfLazyAttribute);
        return;
    }

    if (index == ElementData::attributeNotFound) {
        addAttributeInternal(name, newValue, inSynchronizationOfLazyAttribute);
        return;
    }

    if (inSynchronizationOfLazyAttribute) {
        ensureUniqueElementData()->attributeAt(index).setValue(newValue);
        return;
    }

    const Attribute& attribute = attributeAt(index);
    QualifiedName attributeName = attribute.name();
    AtomicString oldValue = attribute.value();

    willModifyAttribute(attributeName, oldValue, newValue);

    if (newValue != oldValue) {
		Style::AttributeChangeInvalidation styleInvalidation(*this, name, oldValue, newValue);
		ensureUniqueElementData()->attributeAt(index).setValue(newValue);
    }

    didModifyAttribute(attributeName, oldValue, newValue);
}

static inline AtomicString makeIdForStyleResolution(const AtomicString& value, bool inQuirksMode)
{
    if (inQuirksMode)
        return value.lower();
    return value;
}

void Element::notifyAttributeChanged(const QualifiedName& name, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason reason)
{
    attributeChanged(name, oldValue, newValue, reason);

    document().incDOMTreeVersion();

#if 0
    if (UNLIKELY(isDefinedCustomElement()))
        CustomElementReactionQueue::enqueueAttributeChangedCallbackIfNeeded(*this, name, oldValue, newValue);
#endif

    if (oldValue != newValue) {
        invalidateNodeListAndCollectionCachesInAncestorsForAttribute(name);

#if HAVE(ACCESSIBILITY)
        if (auto* cache = document().existingAXObjectCache())
            cache->deferAttributeChangeIfNeeded(this, name, oldValue, newValue);
#endif

#if HAVE(ACCESSIBILITY)
		if (!AXObjectCache::accessibilityEnabled())
			return;

		if (name == aria_activedescendantAttr) {
			// any change to aria-activedescendant attribute triggers accessibility focus change, but document focus remains intact
			document().axObjectCache()->handleActiveDescendantChanged(renderer());
		} else if (name == roleAttr) {
			// the role attribute can change at any time, and the AccessibilityObject must pick up these changes
			document().axObjectCache()->handleAriaRoleChanged(renderer());
		} else if (name == aria_valuenowAttr) {
			// If the valuenow attribute changes, AX clients need to be notified.
			document().axObjectCache()->postNotification(renderer(), AXObjectCache::AXValueChanged, true);
		} else if (name == aria_labelAttr || name == aria_labeledbyAttr || name == altAttr || name == titleAttr) {
			// If the content of an element changes due to an attribute change, notify accessibility.
			document().axObjectCache()->contentChanged(renderer());
		} else if (name == aria_checkedAttr)
			document().axObjectCache()->checkedStateChanged(renderer());
		else if (name == aria_selectedAttr)
			document().axObjectCache()->selectedChildrenChanged(renderer());
		else if (name == aria_expandedAttr)
			document().axObjectCache()->handleAriaExpandedChange(renderer());
		else if (name == aria_hiddenAttr)
			document().axObjectCache()->childrenChanged(renderer());
		else if (name == aria_invalidAttr)
			document().axObjectCache()->postNotification(renderer(), AXObjectCache::AXInvalidStatusChanged, true);
#endif
    }
}

void Element::attributeChanged(const QualifiedName& name, const AtomicString& oldValue, const AtomicString& newValue, AttributeModificationReason)
{
	if (oldValue == newValue)
		return;

	if (name == HTMLNames::idAttr) {
		AtomicString oldId = elementData()->idForStyleResolution();
		AtomicString newId = makeIdForStyleResolution(newValue, document().inQuirksMode());
		if (newId != oldId) {
			Style::IdChangeInvalidation styleInvalidation(*this, oldId, newId);
			elementData()->setIdForStyleResolution(newId);
		}

		if (!oldValue.isEmpty())
			treeScope().idTargetObserverRegistry().notifyObservers(*oldValue.impl());
		if (!newValue.isEmpty())
			treeScope().idTargetObserverRegistry().notifyObservers(*newValue.impl());
	} else if (name == classAttr)
			classAttributeChanged(newValue);
	else if (name == HTMLNames::nameAttr)
		elementData()->m_hasNameAttribute = !newValue.isNull();
	else if (name == HTMLNames::nonceAttr) {
		if ((*this).isHTMLElement() || (*this).isSVGElement())
			setNonce(newValue.isNull() ? emptyAtom : newValue);
	} else if (name == HTMLNames::pseudoAttr) {
		if (needsStyleInvalidation() && isInShadowTree())
			invalidateStyleForSubtree();
	}
#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
	else if (name == HTMLNames::slotAttr) {
		if (auto* parent = parentElement()) {
			if (auto* shadowRoot = parent->shadowRoot())
				shadowRoot->hostChildElementDidChangeSlotAttribute(*this, oldValue, newValue);
		}
	} else if (name == HTMLNames::partAttr)
		partAttributeChanged(newValue);
	else if (name == HTMLNames::exportpartsAttr) {
		if (auto* shadowRoot = this->shadowRoot()) {
			shadowRoot->invalidatePartMappings();
			Style::Invalidator::invalidateShadowParts(*shadowRoot);
		}
	}
#endif
}

void Element::classAttributeChanged(const AtomicString& newClassString)
{
    // Note: We'll need ElementData, but it doesn't have to be UniqueElementData.
    if (!elementData())
        ensureUniqueElementData();

    {
        auto shouldFoldCase = document().inQuirksMode();
        SpaceSplitString newClassNames(newClassString, shouldFoldCase);
        Style::ClassChangeInvalidation styleInvalidation(*this, elementData()->classNames(), newClassNames);
        elementData()->setClassNames(newClassNames);
    }

    if (hasRareData()) {
        if (auto* classList = elementRareData()->classList())
            classList->associatedAttributeValueChanged();
    }
}

void Element::partAttributeChanged(const AtomicString& newValue)
{
    SpaceSplitString newParts(newValue, false);
	if (!newParts.isEmpty() || !partNames().isEmpty())
        ensureElementRareData().setPartNames(newParts);

    if (hasRareData()) {
        if (auto* partList = elementRareData()->partList())
            partList->associatedAttributeValueChanged();
    }

    if (needsStyleInvalidation() && isInShadowTree())
        invalidateStyleInternal();
}

KURL Element::absoluteLinkURL() const
{
    if (!isLink())
        return KURL();

    AtomicString linkAttribute;
    if (hasTagName(SVGNames::aTag))
        linkAttribute = getAttribute(SVGNames::hrefAttr, XLinkNames::hrefAttr);
    else
        linkAttribute = getAttribute(HTMLNames::hrefAttr);

    if (linkAttribute.isEmpty())
        return KURL();

    return document().completeURL(stripLeadingAndTrailingHTMLSpaces(linkAttribute));
}

StyleResolver& Element::styleResolver()
{
    if (auto* shadowRoot = containingShadowRoot())
        return shadowRoot->styleScope().resolver();

    return document().styleScope().resolver();
}

ElementStyle Element::resolveStyle(const Style::ResolutionContext& resolutionContext)
{
    return styleResolver().styleForElement(*this, resolutionContext);
}

void invalidateForSiblingCombinators(Element* sibling)
{
    for (; sibling; sibling = sibling->nextElementSibling()) {
        if (sibling->styleIsAffectedByPreviousSibling())
            sibling->invalidateStyleInternal();
        if (sibling->descendantsAffectedByPreviousSibling()) {
            for (auto* siblingChild = sibling->firstElementChild(); siblingChild; siblingChild = siblingChild->nextElementSibling())
                siblingChild->invalidateStyleForSubtreeInternal();
        }
        if (!sibling->affectsNextSiblingElementStyle())
            return;
    }
}

static void invalidateSiblingsIfNeeded(Element& element)
{
    if (!element.affectsNextSiblingElementStyle())
        return;
    auto* parent = element.parentElement();
    if (parent && parent->styleValidity() >= Style::Validity::SubtreeInvalid)
        return;

    invalidateForSiblingCombinators(element.nextElementSibling());
}

void Element::invalidateStyle()
{
    Node::invalidateStyle(Style::Validity::ElementInvalid);
    invalidateSiblingsIfNeeded(*this);
}

void Element::invalidateStyleAndLayerComposition()
{
    Node::invalidateStyle(Style::Validity::ElementInvalid, Style::InvalidationMode::RecompositeLayer);
    invalidateSiblingsIfNeeded(*this);
}

void Element::invalidateStyleForSubtree()
{
    Node::invalidateStyle(Style::Validity::SubtreeInvalid);
    invalidateSiblingsIfNeeded(*this);
}

void Element::invalidateStyleAndRenderersForSubtree()
{
    Node::invalidateStyle(Style::Validity::SubtreeInvalid, Style::InvalidationMode::RebuildRenderer);
}

void Element::invalidateRenderer()
{
    Node::invalidateStyle(Style::Validity::Valid, Style::InvalidationMode::RebuildRenderer);
}

void Element::invalidateStyleInternal()
{
    Node::invalidateStyle(Style::Validity::ElementInvalid);
}

void Element::invalidateStyleForSubtreeInternal()
{
    Node::invalidateStyle(Style::Validity::SubtreeInvalid);
}

bool Element::hasDisplayContents() const
{
    if (!hasRareData())
        return false;

    const RenderStyle* style = elementRareData()->computedStyle();
    return style && style->display() == DisplayType::Contents;
}

void Element::storeDisplayContentsStyle(PassOwnPtr<RenderStyle> style)
{
    ASSERT(style && style->display() == DisplayType::Contents);
    ASSERT(!renderer() || isPseudoElement());
    ensureElementRareData().setComputedStyle(style);
    clearFlag(NodeFlags::IsComputedStyleInvalidFlag);
}

// Returns true is the given attribute is an event handler.
// We consider an event handler any attribute that begins with "on".
// It is a simple solution that has the advantage of not requiring any
// code or configuration change if a new event handler is defined.

static inline bool isEventHandlerAttribute(const Attribute& attribute)
{
    return attribute.name().namespaceURI().isNull() && attribute.name().localName().startsWith("on");
}

bool Element::isJavaScriptURLAttribute(const Attribute& attribute)
{
    if (!isURLAttribute(attribute))
        return false;
    if (!protocolIsJavaScript(stripLeadingAndTrailingHTMLSpaces(attribute.value())))
        return false;
    return true;
}

bool Element::isJavaScriptAttribute(const Attribute& attribute)
{
    if (isEventHandlerAttribute(attribute))
        return true;
    if (isJavaScriptURLAttribute(attribute))
        return true;
    return false;
}

void Element::stripJavaScriptAttributes(Vector<Attribute>& attributeVector)
{
    size_t destination = 0;
    for (size_t source = 0; source < attributeVector.size(); ++source) {
        if (isJavaScriptAttribute(attributeVector[source]))
            continue;

        if (source != destination)
            attributeVector[destination] = attributeVector[source];

        ++destination;
    }
    attributeVector.shrink(destination);
}

void Element::parserSetAttributes(const Vector<Attribute>& attributeVector)
{
    ASSERT(!inDocument());
    ASSERT(!parentNode());
    ASSERT(!m_elementData);

    if (!attributeVector.isEmpty()) {
		if (document().sharedObjectPool())
			m_elementData = document().sharedObjectPool()->cachedShareableElementDataWithAttributes(attributeVector);
		else
			m_elementData = ShareableElementData::createWithAttributes(attributeVector);
    }

    parserDidSetAttributes();

    // Use attributeVector instead of m_elementData because attributeChanged might modify m_elementData.
    for (const auto& attribute : attributeVector)
    	notifyAttributeChanged(attribute.name(), nullAtom, attribute.value(), ModifiedDirectly);
}

void Element::parserDidSetAttributes()
{
}

void Element::didMoveToNewDocument(Document& oldDocument, Document& newDocument)
{
    ASSERT_WITH_SECURITY_IMPLICATION(&document() == &newDocument);

    if (oldDocument.inQuirksMode() != document().inQuirksMode()) {
    	ensureUniqueElementData();
        // ElementData::m_classNames or ElementData::m_idForStyleResolution need to be updated with the right case.
        if (hasID())
        	notifyAttributeChanged(idAttr, nullAtom, getIdAttribute());
        if (hasClass())
        	notifyAttributeChanged(classAttr, nullAtom, getAttribute(classAttr));
    }

#if 0
    if (UNLIKELY(isDefinedCustomElement()))
        CustomElementReactionQueue::enqueueAdoptedCallbackIfNeeded(*this, oldDocument, newDocument);
#endif

#if ENABLE(INTERSECTION_OBSERVER)
    if (auto* observerData = intersectionObserverDataIfExists()) {
        for (auto& observer : observerData->observers) {
            if (observer->hasObservationTargets()) {
                oldDocument.removeIntersectionObserver(*observer);
                newDocument.addIntersectionObserver(*observer);
            }
        }
    }
#endif
}

bool Element::hasAttributes() const
{
    synchronizeAllAttributes();
    return elementData() && elementData()->length();
}

bool Element::hasEquivalentAttributes(const Element& other) const
{
    synchronizeAllAttributes();
    other.synchronizeAllAttributes();
    if (elementData() == other.elementData())
        return true;
    if (elementData())
        return elementData()->isEquivalent(other.elementData());
    if (other.elementData())
        return other.elementData()->isEquivalent(elementData());
    return true;
}

String Element::nodeName() const
{
    return m_tagName.toString();
}

String Element::nodeNamePreservingCase() const
{
    return m_tagName.toString();
}

void Element::setPrefix(const AtomicString& prefix, ExceptionCode& ec)
{
    ec = 0;
    checkSetPrefix(prefix, ec);
    if (ec)
        return;

    m_tagName.setPrefix(prefix.isEmpty() ? nullAtom : prefix);
}

KURL Element::baseURI() const
{
    const AtomicString& baseAttribute = getAttribute(HTMLNames::baseAttr);
    KURL base(KURL(), baseAttribute);
    if (!base.protocol().isEmpty())
        return base;

    ContainerNode* parent = parentNode();
    if (!parent)
        return base;

    const KURL& parentBase = parent->baseURI();
    if (parentBase.isNull())
        return base;

    return KURL(parentBase, baseAttribute);
}

const AtomicString& Element::imageSourceURL() const
{
    return getAttribute(srcAttr);
}

bool Element::rendererIsNeeded(const RenderStyle& style)
{
    return rendererIsEverNeeded() && style.display() != DisplayType::None && style.display() != DisplayType::Contents;
}

RenderPtr<RenderElement> Element::createElementRenderer(PassOwnPtr<RenderStyle> style, const RenderTreePosition&)
{
    return RenderElement::createFor(*this, style);
}

Node::InsertedIntoResult Element::insertedIntoAncestor(const InsertionType& insertionType, ContainerNode& parentOfInsertedTree)
{
    ContainerNode::insertedIntoAncestor(insertionType, parentOfInsertedTree);

#if ENABLE(FULLSCREEN_API)
    if (containsFullScreenElement() && parentElement() && !parentElement()->containsFullScreenElement())
        setContainsFullScreenElementOnAncestorsCrossingFrameBoundaries(true);
#endif

    if (parentNode() == &parentOfInsertedTree) {
        if (auto* shadowRoot = parentNode()->shadowRoot())
            shadowRoot->hostChildElementDidChange(*this);
    }

    if (!parentOfInsertedTree.isInTreeScope())
        return InsertedIntoResult::Done;

    bool becomeConnected = insertionType.connectedToDocument;
    auto* newScope = &parentOfInsertedTree.treeScope();
    HTMLDocument* newDocument = becomeConnected && newScope->documentScope().isHTMLDocument() ? &toHTMLDocument(newScope->documentScope()) : nullptr;
    if (!insertionType.treeScopeChanged)
		newScope = nullptr;

    const AtomicString& idValue = getIdAttribute();
    if (!idValue.isEmpty()) {
        if (newScope)
        	newScope->addElementById(idValue, this);
        if (newDocument)
            updateIdForDocument(newDocument, nullAtom, idValue, AlwaysUpdateHTMLDocumentNamedItemMaps);
    }

    const AtomicString& nameValue = getNameAttribute();
    if (!nameValue.isEmpty()) {
        if (newScope)
        	newScope->addElementByName(nameValue, this);
        if (newDocument)
            updateNameForDocument(newDocument, nullAtom, nameValue);
    }

    if (UNLIKELY(hasTagName(articleTag) && newDocument))
        newDocument->registerArticleElement(*this);

    return InsertedIntoResult::Done;
}

void Element::removedFromAncestor(const RemovalType& removalType, ContainerNode& parentOfRemovedTree)
{
#if ENABLE(FULLSCREEN_API)
    if (containsFullScreenElement())
        setContainsFullScreenElementOnAncestorsCrossingFrameBoundaries(false);
#endif

    if (auto* page = document().page()) {
#if ENABLE(POINTER_LOCK)
        page->pointerLockController()->elementWasRemoved(*this);
#endif
    }

    setSavedLayerScrollPosition(ScrollPosition());

    if (parentOfRemovedTree.isInTreeScope()) {
        TreeScope* oldScope = &parentOfRemovedTree.treeScope();
        Document* oldDocument = removalType.disconnectedFromDocument ? &oldScope->documentScope() : nullptr;
        HTMLDocument* oldHTMLDocument = oldDocument && isHTMLDocument(*oldDocument) ? &toHTMLDocument(*oldDocument) : nullptr;
        if (!removalType.treeScopeChanged)
            oldScope = nullptr;

        const AtomicString& idValue = getIdAttribute();
        if (!idValue.isEmpty()) {
            if (oldScope)
            	oldScope->removeElementById(idValue, this);
            if (oldHTMLDocument)
                updateIdForDocument(oldHTMLDocument, idValue, nullAtom, AlwaysUpdateHTMLDocumentNamedItemMaps);
        }

        const AtomicString& nameValue = getNameAttribute();
        if (!nameValue.isEmpty()) {
            if (oldScope)
            	oldScope->removeElementByName(nameValue, this);
            if (oldHTMLDocument)
                updateNameForDocument(oldHTMLDocument, nameValue, nullAtom);
        }

        if (oldDocument) {
            if (oldDocument->cssTarget() == this)
                oldDocument->setCSSTarget(nullptr);
            if (UNLIKELY(hasTagName(articleTag)))
                oldDocument->unregisterArticleElement(*this);
        }

#if 0
        if (removalType.disconnectedFromDocument && UNLIKELY(isDefinedCustomElement()))
            CustomElementReactionQueue::enqueueDisconnectedCallbackIfNeeded(*this);
#endif
    }

    if (!parentNode()) {
        if (auto* shadowRoot = parentOfRemovedTree.shadowRoot())
            shadowRoot->hostChildElementDidChange(*this);
    }

    clearBeforePseudoElement();
    clearAfterPseudoElement();

    ContainerNode::removedFromAncestor(removalType, parentOfRemovedTree);

    Styleable::fromElement(*this).elementWasRemoved();

#if PLATFORM(MAC)
    if (RefPtr<Frame> frame = document().frame())
        frame->mainFrame().removeLatchingStateForTarget(*this);
#endif

    if (UNLIKELY(isInTopLayer()))
        removeFromTopLayer();

    document().userActionElements().clearAllForElement(*this);

    if (getFlag(HasElementIdentifierFlag)) {
        //document().identifiedElementWasRemovedFromDocument(*this);
        clearFlag(HasElementIdentifierFlag);
    }
}

void Element::addShadowRoot(Ref<ShadowRoot>&& newShadowRoot)
{
    ASSERT(!newShadowRoot->hasChildNodes());
    ASSERT(!shadowRoot());

    ShadowRoot& shadowRoot = newShadowRoot.get();
    {
        WidgetHierarchyUpdatesSuspensionScope suspendWidgetHierarchyUpdates;
        ScriptDisallowedScope::InMainThread scriptDisallowedScope;
        if (renderer() || hasDisplayContents())
            RenderTreeUpdater::tearDownRenderers(*this);

		ensureElementRareData().setShadowRoot(WTFMove(newShadowRoot));

		shadowRoot.setHost(makeWeakPtr(*this));
		shadowRoot.setParentTreeScope(treeScope());

        NodeVector postInsertionNotificationTargets;
        notifyChildNodeInserted(*this, shadowRoot, postInsertionNotificationTargets);
        ASSERT_UNUSED(postInsertionNotificationTargets, postInsertionNotificationTargets.isEmpty());

		invalidateStyleAndRenderersForSubtree();
    }

    if (shadowRoot.mode() == ShadowRootMode::UserAgent)
        didAddUserAgentShadowRoot(&shadowRoot);

    InspectorInstrumentation::didPushShadowRoot(this, &shadowRoot);
}

void Element::removeShadowRoot()
{
    RefPtr<ShadowRoot> oldRoot = shadowRoot();
    if (!oldRoot)
        return;
    InspectorInstrumentation::willPopShadowRoot(this, oldRoot.get());
    document().adjustFocusedNodeOnNodeRemoval(*oldRoot);

    ASSERT(!oldRoot->renderer());

    elementRareData()->clearShadowRoot();

    oldRoot->setHost(nullptr);
    oldRoot->setParentTreeScope(document());
}

PassRefPtr<ShadowRoot> Element::createShadowRoot(ExceptionCode& ec)
{
    if (alwaysCreateUserAgentShadowRoot())
        ensureUserAgentShadowRoot();

    ec = HIERARCHY_REQUEST_ERR;
	return nullptr;
}

bool Element::canHaveUserAgentShadowRoot() const
{
    return false;
}

RefPtr<ShadowRoot> Element::attachShadow(const ShadowRootInit& init, ExceptionCode& ec)
{
    if (canHaveUserAgentShadowRoot()) {
        ec = NOT_SUPPORTED_ERR;
        return nullptr;
    }

    if (shadowRoot()) {
        ec = INVALID_STATE_ERR;
        return nullptr;
    }

    if (init.mode == ShadowRootMode::UserAgent) {
		ec = TypeError;
		return nullptr;
	}

    Ref<ShadowRoot> shadow = ShadowRoot::create(&document(), init.mode, init.delegatesFocus ? ShadowRoot::DelegatesFocus::Yes : ShadowRoot::DelegatesFocus::No);
    auto& result = shadow.get();
    addShadowRoot(WTFMove(shadow));
	return result;
}

ShadowRoot* Element::bindingShadowRoot() const
{
    ShadowRoot* root = shadowRoot();
    if (!root)
        return 0;

    if (root->mode() != ShadowRootMode::Open)
        return 0;

    return root;
}

RefPtr<ShadowRoot> Element::userAgentShadowRoot() const
{
    if (RefPtr<ShadowRoot> shadowRoot = this->shadowRoot()) {
		ASSERT(shadowRoot->mode() == ShadowRootMode::UserAgent);
		return shadowRoot;
    }

    return nullptr;
}

ShadowRoot& Element::ensureUserAgentShadowRoot()
{
	RefPtr<ShadowRoot> shadowRoot = userAgentShadowRoot();
    if (!shadowRoot) {
        addShadowRoot(ShadowRoot::create(&document(), ShadowRootMode::UserAgent));
        shadowRoot = userAgentShadowRoot();
    }
    return *shadowRoot;
}

inline void Node::setCustomElementState(CustomElementState state)
{
    auto bitfields = rareDataBitfields();
    bitfields.customElementState = static_cast<uint16_t>(state);
    setRareDataBitfields(bitfields);
}

#if 0
void Element::setIsDefinedCustomElement(JSCustomElementInterface& elementInterface)
{
    setCustomElementState(CustomElementState::Custom);
    auto& data = ensureElementRareData();
    if (!data.customElementReactionQueue())
        data.setCustomElementReactionQueue(makeUnique<CustomElementReactionQueue>(elementInterface));
    invalidateStyleForSubtree();
    InspectorInstrumentation::didChangeCustomElementState(*this);
}

void Element::setIsFailedCustomElement()
{
    setIsFailedCustomElementWithoutClearingReactionQueue();
    clearReactionQueueFromFailedCustomElement();
}

void Element::setIsFailedCustomElementWithoutClearingReactionQueue()
{
    ASSERT(isUndefinedCustomElement());
    ASSERT(customElementState() == CustomElementState::Undefined);
    setCustomElementState(CustomElementState::Failed);
    InspectorInstrumentation::didChangeCustomElementState(*this);
}

void Element::clearReactionQueueFromFailedCustomElement()
{
    ASSERT(isFailedCustomElement());
    if (hasRareData()) {
        // Clear the queue instead of deleting it since this function can be called inside CustomElementReactionQueue::invokeAll during upgrades.
        if (auto* queue = elementRareData()->customElementReactionQueue())
            queue->clear();
    }
}

void Element::setIsCustomElementUpgradeCandidate()
{
    ASSERT(customElementState() == CustomElementState::Uncustomized);
    setCustomElementState(CustomElementState::Undefined);
    InspectorInstrumentation::didChangeCustomElementState(*this);
}

void Element::enqueueToUpgrade(JSCustomElementInterface& elementInterface)
{
    ASSERT(isCustomElementUpgradeCandidate());
    ASSERT(!isDefinedCustomElement() && !isFailedCustomElement());
    auto& data = ensureElementRareData();
    bool alreadyScheduledToUpgrade = data.customElementReactionQueue();
    if (!alreadyScheduledToUpgrade)
        data.setCustomElementReactionQueue(makeUnique<CustomElementReactionQueue>(elementInterface));
    data.customElementReactionQueue()->enqueueElementUpgrade(*this, alreadyScheduledToUpgrade);
}

CustomElementReactionQueue* Element::reactionQueue() const
{
#if ASSERT_ENABLED
    if (isFailedCustomElement()) {
        auto* queue = elementRareData()->customElementReactionQueue();
        ASSERT(queue);
        ASSERT(queue->isEmpty() || queue->hasJustUpgradeReaction());
    } else
        ASSERT(isDefinedCustomElement() || isCustomElementUpgradeCandidate());
#endif
    if (!hasRareData())
        return nullptr;
    return elementRareData()->customElementReactionQueue();
}
#endif

const AtomicString& Element::shadowPseudoId() const
{
    return pseudo();
}

bool Element::childTypeAllowed(NodeType type) const
{
    switch (type) {
    case ELEMENT_NODE:
    case TEXT_NODE:
    case COMMENT_NODE:
    case PROCESSING_INSTRUCTION_NODE:
    case CDATA_SECTION_NODE:
        return true;
    default:
        break;
    }
    return false;
}

void Element::childrenChanged(const ChildChange& change)
{
    ContainerNode::childrenChanged(change);

    if (auto* shadowRoot = this->shadowRoot()) {
        switch (change.type) {
        case ElementInserted:
        case ElementRemoved:
            // For elements, we notify shadowRoot in Element::insertedIntoAncestor and Element::removedFromAncestor.
            break;
        case AllChildrenRemoved:
        case AllChildrenReplaced:
            shadowRoot->didRemoveAllChildrenOfShadowHost();
            break;
        case TextInserted:
        case TextRemoved:
        case TextChanged:
            shadowRoot->didChangeDefaultSlot();
            break;
        case NonContentsChildInserted:
        case NonContentsChildRemoved:
            break;
        }
    }
}

void Element::setAttributeEventListener(const AtomicString& eventType, const QualifiedName& attributeName, const AtomicString& attributeValue)
{
    setAttributeEventListener(eventType, JSLazyEventListener::create(*this, attributeName, attributeValue), mainThreadNormalWorld());
}

void Element::removeAllEventListeners()
{
    ContainerNode::removeAllEventListeners();
    if (ShadowRoot* shadowRoot = this->shadowRoot())
        shadowRoot->removeAllEventListeners();
}

void Element::beginParsingChildren()
{
    clearIsParsingChildrenFinished();
}

void Element::finishParsingChildren()
{
    ContainerNode::finishParsingChildren();
    setIsParsingChildrenFinished();

    Style::ChildChangeInvalidation::invalidateAfterFinishedParsingChildren(*this);
}

#ifndef NDEBUG
void Element::formatForDebugger(char* buffer, unsigned length) const
{
    String result;
    String s;
    
    s = nodeName();
    if (s.length() > 0) {
        result += s;
    }
          
    s = getIdAttribute();
    if (s.length() > 0) {
        if (result.length() > 0)
            result += "; ";
        result += "id=";
        result += s;
    }
          
    s = getAttribute(classAttr);
    if (s.length() > 0) {
        if (result.length() > 0)
            result += "; ";
        result += "class=";
        result += s;
    }
          
    strncpy(buffer, result.utf8().data(), length - 1);
}
#endif

const Vector<RefPtr<Attr> >& Element::attrNodeList()
{
    ASSERT(hasSyntheticAttrChildNodes());
    return *attrNodeListForElement(this);
}

void Element::attachAttributeNodeIfNeeded(Attr& attrNode)
{
    ASSERT(!attrNode.ownerElement() || attrNode.ownerElement() == this);
    if (attrNode.ownerElement() == this)
        return;

    ScriptDisallowedScope::InMainThread assertNoEventDispatch;

    attrNode.attachToElement(*this);
    ensureAttrNodeListForElement(this)->append(&attrNode);
}

PassRefPtr<Attr> Element::setAttributeNode(Attr& attrNode, ExceptionCode& ec)
{
    RefPtr<Attr> oldAttrNode = attrIfExists(attrNode.qualifiedName());
    if (oldAttrNode.get() == &attrNode)
        return &attrNode; // This Attr is already attached to the element.

    // INUSE_ATTRIBUTE_ERR: Raised if node is an Attr that is already an attribute of another Element object.
    // The DOM user must explicitly clone Attr nodes to re-use them in other elements.
    if (attrNode.ownerElement() && attrNode.ownerElement() != this) {
        ec = INUSE_ATTRIBUTE_ERR;
        return nullptr;
    }

    {
    	ScriptDisallowedScope::InMainThread assertNoEventDispatch;
    	synchronizeAllAttributes();
    }

    UniqueElementData& elementData = *ensureUniqueElementData();

    unsigned existingAttributeIndex = elementData.findAttributeIndexByName(attrNode.qualifiedName());

    // Attr::value() will return its 'm_standaloneValue' member any time its Element is set to nullptr. We need to cache this value
    // before making changes to attrNode's Element connections.
    auto attrNodeValue = attrNode.value();

    if (existingAttributeIndex == ElementData::attributeNotFound) {
        attachAttributeNodeIfNeeded(attrNode);
        setAttributeInternal(elementData.findAttributeIndexByName(attrNode.qualifiedName()), attrNode.qualifiedName(), attrNodeValue, NotInSynchronizationOfLazyAttribute);
    } else {
        const Attribute& attribute = attributeAt(existingAttributeIndex);
        if (oldAttrNode)
            detachAttrNodeFromElementWithValue(oldAttrNode.get(), attribute.value());
        else
            oldAttrNode = Attr::create(document(), attrNode.qualifiedName(), attribute.value());

        attachAttributeNodeIfNeeded(attrNode);

        if (attribute.name().matches(attrNode.qualifiedName()))
            setAttributeInternal(existingAttributeIndex, attrNode.qualifiedName(), attrNodeValue, NotInSynchronizationOfLazyAttribute);
        else {
            removeAttributeInternal(existingAttributeIndex, NotInSynchronizationOfLazyAttribute);
            setAttributeInternal(ensureUniqueElementData()->findAttributeIndexByName(attrNode.qualifiedName()), attrNode.qualifiedName(), attrNodeValue, NotInSynchronizationOfLazyAttribute);
        }
    }

    return oldAttrNode;
}

PassRefPtr<Attr> Element::setAttributeNodeNS(Attr& attrNode, ExceptionCode& ec)
{
    RefPtr<Attr> oldAttrNode = attrIfExists(attrNode.qualifiedName());
    if (oldAttrNode.get() == &attrNode)
        return &attrNode; // This Attr is already attached to the element.

    // INUSE_ATTRIBUTE_ERR: Raised if node is an Attr that is already an attribute of another Element object.
    // The DOM user must explicitly clone Attr nodes to re-use them in other elements.
    if (attrNode.ownerElement() && attrNode.ownerElement() != this) {
        ec = INUSE_ATTRIBUTE_ERR;
        return nullptr;
    }

    // Attr::value() will return its 'm_standaloneValue' member any time its Element is set to nullptr. We need to cache this value
    // before making changes to attrNode's Element connections.
    auto attrNodeValue = attrNode.value();
    unsigned index = 0;
    {
    	ScriptDisallowedScope::InMainThread assertNoEventDispatch;
		synchronizeAllAttributes();
		UniqueElementData& elementData = *ensureUniqueElementData();

		index = elementData.findAttributeIndexByName(attrNode.qualifiedName());

		if (index != ElementData::attributeNotFound) {
			if (oldAttrNode)
				detachAttrNodeFromElementWithValue(oldAttrNode.get(), elementData.attributeAt(index).value());
			else
				oldAttrNode = Attr::create(document(), attrNode.qualifiedName(), elementData.attributeAt(index).value());
		}
    }

    attachAttributeNodeIfNeeded(attrNode);
    setAttributeInternal(index, attrNode.qualifiedName(), attrNodeValue, NotInSynchronizationOfLazyAttribute);

    return oldAttrNode.release();
}

PassRefPtr<Attr> Element::removeAttributeNode(Attr& attr, ExceptionCode& ec)
{
    if (attr.ownerElement() != this) {
        ec = NOT_FOUND_ERR;
        return nullptr;
    }

    ASSERT(&document() == &attr->document());

    synchronizeAllAttributes();

    if (!m_elementData) {
        ec = NOT_FOUND_ERR;
        return nullptr;
    }

    unsigned existingAttributeIndex = m_elementData->findAttributeIndexByName(attr.qualifiedName());

    if (existingAttributeIndex == ElementData::attributeNotFound) {
        ec = NOT_FOUND_ERR;
        return nullptr;
    }

    RefPtr<Attr> attrNode = &attr;
    detachAttrNodeFromElementWithValue(&attr, m_elementData->attributeAt(existingAttributeIndex).value());
    removeAttributeInternal(existingAttributeIndex, NotInSynchronizationOfLazyAttribute);
    return attrNode;
}

bool Element::parseAttributeName(QualifiedName& out, const AtomicString& namespaceURI, const AtomicString& qualifiedName, ExceptionCode& ec)
{
    String prefix, localName;
    if (!Document::parseQualifiedName(qualifiedName, prefix, localName, ec))
        return false;
    ASSERT(!ec);

    QualifiedName qName(prefix, localName, namespaceURI);

    if (!Document::hasValidNamespaceForAttributes(qName)) {
        ec = NAMESPACE_ERR;
        return false;
    }

    out = qName;
    return true;
}

void Element::setAttributeNS(const AtomicString& namespaceURI, const AtomicString& qualifiedName, const AtomicString& value, ExceptionCode& ec)
{
    QualifiedName parsedName = anyName;
    if (!parseAttributeName(parsedName, namespaceURI, qualifiedName, ec))
        return;
    setAttribute(parsedName, value);
}

void Element::removeAttributeInternal(unsigned index, SynchronizationOfLazyAttribute inSynchronizationOfLazyAttribute)
{
    ASSERT(index < attributeCount());

    UniqueElementData& elementData = *ensureUniqueElementData();

    QualifiedName name = elementData.attributeAt(index).name();
    AtomicString valueBeingRemoved = elementData.attributeAt(index).value();

    if (RefPtr<Attr> attrNode = attrIfExists(name))
        detachAttrNodeFromElementWithValue(attrNode.get(), elementData.attributeAt(index).value());

    if (inSynchronizationOfLazyAttribute) {
        elementData.removeAttribute(index);
        return;
    }

    if (!valueBeingRemoved.isNull())
        willModifyAttribute(name, valueBeingRemoved, nullAtom);

    {
        Style::AttributeChangeInvalidation styleInvalidation(*this, name, valueBeingRemoved, nullAtom);
        elementData.removeAttribute(index);
    }

	didRemoveAttribute(name, valueBeingRemoved);
}

void Element::addAttributeInternal(const QualifiedName& name, const AtomicString& value, SynchronizationOfLazyAttribute inSynchronizationOfLazyAttribute)
{
    if (inSynchronizationOfLazyAttribute) {
        ensureUniqueElementData()->addAttribute(name, value);
        return;
    }

    willModifyAttribute(name, nullAtom, value);
    {
		Style::AttributeChangeInvalidation styleInvalidation(*this, name, nullAtom, value);
		ensureUniqueElementData()->addAttribute(name, value);
    }
	didAddAttribute(name, value);
}

bool Element::removeAttribute(const AtomicString& name)
{
    if (!elementData())
        return false;

    String localName = shouldIgnoreAttributeCase(*this) ? name.convertToASCIILowercase() : name;
    unsigned index = elementData()->findAttributeIndexByName(localName, false);
    if (index == ElementData::attributeNotFound) {
        if (UNLIKELY(localName == styleAttr) && elementData()->m_styleAttributeIsDirty && isStyledElement())
            static_cast<StyledElement*>(this)->removeAllInlineStyleProperties();
        return false;
    }

    removeAttributeInternal(index, NotInSynchronizationOfLazyAttribute);
    return true;
}

bool Element::removeAttributeNS(const AtomicString& namespaceURI, const AtomicString& localName)
{
    return removeAttribute(QualifiedName(nullAtom, localName, namespaceURI));
}

PassRefPtr<Attr> Element::getAttributeNode(const AtomicString& localName)
{
    if (!elementData())
    	return nullptr;
    synchronizeAttribute(localName);
    const Attribute* attribute = elementData()->findAttributeByName(localName, shouldIgnoreAttributeCase(*this));
    if (!attribute)
        return nullptr;
    return ensureAttr(attribute->name());
}

PassRefPtr<Attr> Element::getAttributeNodeNS(const AtomicString& namespaceURI, const AtomicString& localName)
{
    if (!elementData())
    	return nullptr;
    QualifiedName qName(nullAtom, localName, namespaceURI);
    synchronizeAttribute(qName);
    const Attribute* attribute = elementData()->findAttributeByName(QualifiedName(nullAtom, localName, namespaceURI));
    if (!attribute)
        return nullptr;
    return ensureAttr(attribute->name());
}

DLL_PUBLIC bool Element::hasAttribute(const AtomicString& localName) const
{
    if (!elementData())
        return false;
    synchronizeAttribute(localName);
    return elementData()->findAttributeByName(localName, shouldIgnoreAttributeCase(*this));
}

bool Element::hasAttribute(const char* name) const
{
    if (!elementData())
        return false;
	return hasAttribute(AtomicString(name));
}

bool Element::hasAttributeNS(const AtomicString& namespaceURI, const AtomicString& localName) const
{
    if (!elementData())
        return false;
    QualifiedName qName(nullAtom, localName, namespaceURI);
    synchronizeAttribute(qName);
    return elementData()->findAttributeByName(qName);
}

static RefPtr<ShadowRoot> shadowRootWithDelegatesFocus(const Element& element)
{
    if (auto* root = element.shadowRoot()) {
        if (root->delegatesFocus())
            return root;
    }
    return nullptr;
}

static bool isProgramaticallyFocusable(Element& element)
{
    ScriptDisallowedScope::InMainThread scriptDisallowedScope;

    if (shadowRootWithDelegatesFocus(element))
        return false;

    // If the stylesheets have already been loaded we can reliably check isFocusable.
    // If not, we continue and set the focused node on the focus controller below so that it can be updated soon after attach.
    if (element.document().haveStylesheetsLoaded()) {
        if (!element.isFocusable() && !element.hasEventListeners(eventNames().clickEvent))
            return false;
    }
    return element.supportsFocus() || element.hasEventListeners(eventNames().clickEvent);
}

static RefPtr<Element> findFirstProgramaticallyFocusableElementInComposedTree(Element& host)
{
    ASSERT(host.shadowRoot());
    for (auto& node : composedTreeDescendants(host)) {
        if (!isElement(node))
            continue;
        auto& element = toElement(node);
        if (isProgramaticallyFocusable(element))
            return &element;
    }
    return nullptr;
}

void Element::focus(const FocusOptions& options)
{
    if (!inDocument())
        return;

    Ref<Document> document = this->document();
    if (document->focusedElement() == this) {
        if (auto* page = document->page())
            page->chrome()->client()->elementDidRefocus(*this);
        return;
    }

    RefPtr<Element> newTarget = this;

    // some contents do not work without layout.
    if (document->haveStylesheetsLoaded())
    	 document->updateLayoutIgnorePendingStylesheets();

    // If we don't have renderer yet, isFocusable will compute it without style update.
    // FIXME: Expand it to avoid style update in all cases.
    if (renderer() && document->haveStylesheetsLoaded())
        document->updateStyleIfNeeded();

    if (&newTarget->document() != document.ptr())
        return;

    if (auto root = shadowRootWithDelegatesFocus(*this)) {
        RefPtr<Element> currentlyFocusedElement = document->focusedElement();
        if (root->containsIncludingShadowDOM(currentlyFocusedElement.get())) {
            if (document->page())
                document->page()->chrome()->client()->elementDidRefocus(*currentlyFocusedElement);
            return;
        }

        newTarget = findFirstProgramaticallyFocusableElementInComposedTree(*this);
        if (!newTarget)
            return;
    } else if (!isProgramaticallyFocusable(*newTarget)) {
    	if(Page* page = document->page())
    		page->chrome()->client()->elementDidFocus(this);
        return;
    }

    if (Page* page = document->page()) {
#if 0
        auto& frame = *document->frame();
        if (!frame.hasHadUserInteraction() && !frame.isMainFrame() && !document->topDocument()->securityOrigin()->canAccess(document->securityOrigin()))
            return;
#endif

        FocusOptions optionsWithVisibility = options;
        if (options.trigger == FocusTrigger::Bindings && document->wasLastFocusByClick())
            optionsWithVisibility.visibility = FocusVisibility::Invisible;
        else if (options.trigger != FocusTrigger::Click)
            optionsWithVisibility.visibility = FocusVisibility::Visible;

        // Focus and change event handlers can cause us to lose our last ref.
        // If a focus event handler changes the focus to a different node it
        // does not make sense to continue and update appearence.
        if (!page->focusController()->setFocusedElement(newTarget.get(), document->frame(), optionsWithVisibility))
            return;
    }

    SelectionRevealMode revealMode = SelectionRevealMode::Reveal;
#if PLATFORM(IOS_FAMILY)
    // Focusing a form element triggers animation in UIKit to scroll to the right position.
    // Calling updateFocusAppearance() would generate an unnecessary call to ScrollView::setScrollPosition(),
    // which would jump us around during this animation. See <rdar://problem/6699741>.
    bool isFormControl = is<HTMLFormControlElement>(newTarget);
    if (isFormControl)
        revealMode = SelectionRevealMode::RevealUpToMainFrame;
#endif

    RefPtr<Element> target = focusAppearanceUpdateTarget();
    if (!target)
        return;

    target->updateFocusAppearance(options.selectionRestorationMode, revealMode);
}

void Element::focusForBindings()
{
	FocusOptions options;
    options.trigger = FocusTrigger::Bindings;
    focus(options);
}

// https://html.spec.whatwg.org/#focus-processing-model
RefPtr<Element> Element::focusAppearanceUpdateTarget()
{
    return this;
}

// https://html.spec.whatwg.org/#focus-processing-model
void Element::updateFocusAppearance(SelectionRestorationMode, SelectionRevealMode revealMode)
{
    if (isRootEditableElement()) {
        // Keep Document alive in this method, since setSelection() may release the last reference to |document|.
        Ref<Document> document = this->document();
        
        // When focusing an editable element in an iframe, don't reset the selection if it already contains a selection.
        if (this == document->selection().selection().rootEditableElement())
            return;

        // FIXME: We should restore the previous selection if there is one.
        VisibleSelection newSelection = VisibleSelection(firstPositionInOrBeforeNode(this), DOWNSTREAM);
        
        if (document->selection().shouldChangeSelection(newSelection)) {
        	document->selection().setSelection(newSelection);
        	document->selection().revealSelection(revealMode);
        }
    } else if (renderer() && !renderer()->isWidget()){
#if ENABLE(BML)
    	if(document().isBMLDocument())
    		return;
#endif
        renderer()->scrollRectToVisible(revealMode, renderer()->anchorRect());
    }
}

void Element::blur(bool byUser)
{
    Document& doc = document();
    if (treeScope().focusedElementInScope() == this) {
    	if (Page* page = doc.page())
    		page->chrome()->client()->elementDidBlur(*this, byUser);
    	
        if (RefPtr<Frame> frame = doc.frame())
        	frame->page()->focusController()->setFocusedElement(nullptr, frame);
        else
            doc.setFocusedElement(nullptr);
    }
}

void Element::runFocusingStepsForAutofocus()
{
    focus();
}

void Element::dispatchFocusInEventIfNeeded(PassRefPtr<Element> oldFocusedElement)
{
    if (!document().hasListenerType(Document::FOCUSIN_LISTENER))
        return;
    ASSERT_WITH_SECURITY_IMPLICATION(ScriptDisallowedScope::InMainThread::isEventAllowed());
	dispatchScopedEvent(FocusEvent::create(eventNames().focusinEvent, true, false, document().defaultView(), 0, oldFocusedElement));
}

void Element::dispatchFocusOutEventIfNeeded(PassRefPtr<Element> newFocusedElement)
{
    if (!document().hasListenerType(Document::FOCUSOUT_LISTENER))
        return;
    ASSERT_WITH_SECURITY_IMPLICATION(ScriptDisallowedScope::InMainThread::isEventAllowed());
	dispatchScopedEvent(FocusEvent::create(eventNames().focusoutEvent, true, false, document().defaultView(), 0, newFocusedElement));
}

void Element::dispatchFocusEvent(PassRefPtr<Element> oldFocusedElement, const FocusOptions&)
{
	// Interfere with 2D navigation
#if 0
    if (auto* page = document().page())
        page->chrome()->client()->elementDidFocus(this);
#endif

#if ENABLE(BML)
	if(document().isBMLDocument()){
		document().queueFocusEvent(this);
		return;
	}
#endif

	EventDispatcher::dispatchEvent(this, FocusEvent::create(eventNames().focusEvent, false, false, document().defaultView(), 0, oldFocusedElement));
}

void Element::dispatchBlurEvent(PassRefPtr<Element> newFocusedElement)
{
	// Interfere with 2D navigation
#if 0
    if (auto* page = document().page())
        page->chrome()->client()->elementDidBlur(*this, false);
#endif

#if ENABLE(BML)
	if(document().isBMLDocument()){
		document().queueBlurEvent(this);
		return;
	}
#endif

	EventDispatcher::dispatchEvent(this, FocusEvent::create(eventNames().blurEvent, false, false, document().defaultView(), 0, newFocusedElement));
}

void Element::mergeWithNextTextNode(PassRefPtr<Text> node, ExceptionCode& ec)
{
    Node* next = node->nextSibling();
    if (!next || !next->isTextNode())
        return;

    RefPtr<Text> textNode = node;
    RefPtr<Text> textNext = toText(next);
    textNode->appendData(textNext->data());
    textNext->remove(ec);
}

String Element::innerHTML() const
{
    return serializeFragment(*this, SerializedNodes::SubtreesOfChildren);
}

String Element::outerHTML() const
{
    return serializeFragment(*this, SerializedNodes::SubtreeIncludingNode);
}

void Element::setOuterHTML(const String& html, ExceptionCode& ec)
{
	// The specification allows setting outerHTML on an Element whose parent is a DocumentFragment and Gecko supports this.
	// https://w3c.github.io/DOM-Parsing/#dom-element-outerhtml
	RefPtr<Element> parent = parentElement();
	if (UNLIKELY(!parent)) {
		if (!parentNode())
			return;
        ec = NO_MODIFICATION_ALLOWED_ERR;
        return;
    }
    RefPtr<Node> prev = previousSibling();
    RefPtr<Node> next = nextSibling();

    RefPtr<DocumentFragment> fragment = createFragmentForInnerOuterHTML(html, parent.get(), AllowScriptingContent, ec);
    if (ec)
        return;

    parent->replaceChild(*fragment, *this, ec);
    if (ec)
        return;

    // The following is not part of the specification but matches Blink's behavior as of June 2021.
    RefPtr<Node> node = next ? next->previousSibling() : 0;
    if (node && node->isTextNode()) {
        mergeWithNextTextNode(toText(node.get()), ec);
        if (ec)
            return;
    }

    if (prev && prev->isTextNode())
        mergeWithNextTextNode(toText(prev.get()), ec);
}

void Element::setInnerHTML(const String& html, ExceptionCode& ec)
{
    ContainerNode* container;
    if (!hasTagName(templateTag))
        container = this;
    else
        container = toHTMLTemplateElement(this)->content();

    // Parsing empty string creates additional elements only inside <html> container
    // https://html.spec.whatwg.org/multipage/parsing.html#parsing-main-inhtml
    if (html.isEmpty() && !isHTMLHtmlElement(*container)) {
        ChildListMutationScope mutation(container);
        container->removeChildren();
        return;
    }

    RefPtr<DocumentFragment> fragment = createFragmentForInnerOuterHTML(html, this, AllowScriptingContent, ec);
    if (ec)
        return;

    return replaceChildrenWithFragment(container, fragment.release(), ec);
}

String Element::innerText()
{
    // We need to update layout, since plainText uses line boxes in the render tree.
    document().updateLayoutIgnorePendingStylesheets();

    if (!renderer())
        return textContent(true);

    return plainText(makeRangeSelectingNodeContents(*this));
}

String Element::outerText()
{
    // Getting outerText is the same as getting innerText, only
    // setting is different. You would think this should get the plain
    // text for the outer range, but this is wrong, <br> for instance
    // would return different values for inner and outer text by such
    // a rule, but it doesn't in WinIE, and we want to match that.
    return innerText();
}

String Element::title() const
{
    return String();
}

const AtomicString& Element::pseudo() const
{
    return getAttribute(pseudoAttr);
}

void Element::setPseudo(const AtomicString& value)
{
	setAttributeWithoutSynchronization(pseudoAttr, value);
}

LayoutSize Element::minimumSizeForResizing() const
{
    return hasRareData() ? elementRareData()->minimumSizeForResizing() : defaultMinimumSizeForResizing();
}

void Element::setMinimumSizeForResizing(const LayoutSize& size)
{
    if (size == defaultMinimumSizeForResizing() && !hasRareData())
        return;
    ensureElementRareData().setMinimumSizeForResizing(size);
}

template<typename Function>
static void forEachRenderLayer(Element& element, const Function& function)
{
    auto* renderer = element.renderer();
    if (!renderer || !renderer->isRenderLayerModelObject())
        return;

    auto& layerModelObject = toRenderLayerModelObject(*renderer);

    if (!layerModelObject.isBoxModelObject()) {
        if (layerModelObject.hasLayer())
            function(*layerModelObject.layer());
        return;
    }

    RenderBoxModelObject::forRendererAndContinuations(toRenderBoxModelObject(*renderer), [&](RenderBoxModelObject& renderer) {
        if (renderer.hasLayer())
            function(*renderer.layer());
    });
}

void Element::addToTopLayer()
{
    RELEASE_ASSERT(!isInTopLayer());
    ScriptDisallowedScope scriptDisallowedScope;

    forEachRenderLayer(*this, [](RenderLayer& layer) {
        layer.establishesTopLayerWillChange();
    });

    document().addTopLayerElement(*this);
    setNodeFlag(NodeFlags::IsInTopLayer);

    // Invalidate inert state
    invalidateStyleInternal();
    if (document().documentElement())
        document().documentElement()->invalidateStyleInternal();

    forEachRenderLayer(*this, [](RenderLayer& layer) {
        layer.establishesTopLayerDidChange();
    });
}

void Element::removeFromTopLayer()
{
    RELEASE_ASSERT(isInTopLayer());
    ScriptDisallowedScope scriptDisallowedScope;

    forEachRenderLayer(*this, [](RenderLayer& layer) {
        layer.establishesTopLayerWillChange();
    });

    // We need to call Styleable::fromRenderer() while this element is still contained in
    // Document::topLayerElements(), since Styleable::fromRenderer() relies on this to
    // find the backdrop's associated element.
    if (auto* renderer = this->renderer()) {
        if (auto backdrop = renderer->backdropRenderer()) {
            if (auto styleable = Styleable::fromRenderer(*backdrop))
                styleable->cancelDeclarativeAnimations();
        }
    }

    document().removeTopLayerElement(*this);
    clearFlag(NodeFlags::IsInTopLayer);

    // Invalidate inert state
    invalidateStyleInternal();
    if (document().documentElement())
        document().documentElement()->invalidateStyleInternal();
    if (auto* modalElement = document().activeModalDialog())
        modalElement->invalidateStyleInternal();

    forEachRenderLayer(*this, [](RenderLayer& layer) {
        layer.establishesTopLayerDidChange();
    });
}

static PseudoElement* beforeOrAfterPseudoElement(const Element& host, PseudoId pseudoElementSpecifier)
{
    switch (pseudoElementSpecifier) {
    case PseudoId::Before:
        return host.beforePseudoElement();
    case PseudoId::After:
        return host.afterPseudoElement();
    default:
        return nullptr;
    }
}

const RenderStyle* Element::existingComputedStyle() const
{
    if (hasRareData()) {
        if (auto* style = elementRareData()->computedStyle())
            return style;
    }

    return renderStyle();
}

const RenderStyle* Element::renderOrDisplayContentsStyle(PseudoId pseudoId) const
{
    if (pseudoId != PseudoId::None) {
        if (auto* pseudoElement = beforeOrAfterPseudoElement(*this, pseudoId))
            return pseudoElement->renderOrDisplayContentsStyle();

        if (auto* computedStyle = existingComputedStyle()) {
            if (auto* cachedPseudoStyle = computedStyle->getCachedPseudoStyle(pseudoId))
                return cachedPseudoStyle;
        }

        return nullptr;
    }

    if (auto* style = renderStyle())
        return style;

    if (!hasRareData())
        return nullptr;
    auto* style = elementRareData()->computedStyle();
    if (style && style->display() == DisplayType::Contents)
        return style;

    return nullptr;
}

const RenderStyle* Element::resolveComputedStyle(ResolveComputedStyleMode mode)
{
    ASSERT(inDocument());
    bool isInDisplayNoneTree = false;

    // Traverse the ancestor chain to find the rootmost element that has invalid computed style.
	auto* rootmostInvalidElement = [&]() -> const Element* {
		if (document().hasPendingFullStyleRebuild())
			return document().documentElement();

		if (!document().documentElement() || document().documentElement()->hasNodeFlag(NodeFlags::IsComputedStyleInvalidFlag))
			return document().documentElement();

		const Element* rootmost = nullptr;

		for (auto* element = this; element; element = element->parentElementInComposedTree()) {
			if (element->hasNodeFlag(NodeFlags::IsComputedStyleInvalidFlag)) {
				rootmost = element;
				continue;
			}
			auto* existing = element->existingComputedStyle();
			if (!existing) {
				rootmost = element;
				continue;
			}
			if (mode == ResolveComputedStyleMode::RenderedOnly && existing->display() == DisplayType::None) {
				isInDisplayNoneTree = true;
				return nullptr;
			}
		}        return rootmost;
	}();

	if (isInDisplayNoneTree)
		return nullptr;

	if (!rootmostInvalidElement)
		return existingComputedStyle();

	auto* ancestorWithValidStyle = rootmostInvalidElement->parentElementInComposedTree();

	Vector<RefPtr<Element>, 32> elementsRequiringComputedStyle;
	for (auto* toResolve = this; toResolve != ancestorWithValidStyle; toResolve = toResolve->parentElementInComposedTree())
		elementsRequiringComputedStyle.append(toResolve);

	const RenderStyle* computedStyle = ancestorWithValidStyle ? ancestorWithValidStyle->existingComputedStyle() : nullptr;

    // On iOS request delegates called during styleForElement may result in re-entering WebKit and killing the style resolver.
    Style::PostResolutionCallbackDisabler disabler(document(), Style::PostResolutionCallbackDisabler::DrainCallbacks::No);

    // Resolve and cache styles starting from the most distant ancestor.
    // FIXME: This is not as efficient as it could be. For example if an ancestor has a non-inherited style change but
	// the styles are otherwise clean we would not need to re-resolve descendants.
	for (RefPtr<Element>& element : makeReversedRange(elementsRequiringComputedStyle)) {
    	bool hadDisplayContents = element->hasDisplayContents();
        OwnPtr<RenderStyle> style = document().styleForElementIgnoringPendingStylesheets(*element, computedStyle);
        computedStyle = style.get();
        ElementRareData& rareData = element->ensureElementRareData();
        rareData.setComputedStyle(style.release());
        element->clearFlag(NodeFlags::IsComputedStyleInvalidFlag);
        if (hadDisplayContents && computedStyle->display() != DisplayType::Contents)
            element->setDisplayContentsChanged();

        if (mode == ResolveComputedStyleMode::RenderedOnly && computedStyle->display() == DisplayType::None)
            return nullptr;
    }

    return computedStyle;
}

const RenderStyle& Element::resolvePseudoElementStyle(PseudoId pseudoElementSpecifier)
{
    ASSERT(!isPseudoElement());

    auto* parentStyle = existingComputedStyle();
    ASSERT(parentStyle);
    ASSERT(!parentStyle->getCachedPseudoStyle(pseudoElementSpecifier));

    Style::PostResolutionCallbackDisabler disabler(document(), Style::PostResolutionCallbackDisabler::DrainCallbacks::No);

    OwnPtr<RenderStyle> style = document().styleForElementIgnoringPendingStylesheets(*this, const_cast<RenderStyle*>(parentStyle), pseudoElementSpecifier);
    if (!style) {
        style = RenderStyle::create();
        style->inheritFrom(*parentStyle);
        style->setStyleType(pseudoElementSpecifier);
    }

    auto* computedStyle = style.get();
    const_cast<RenderStyle*>(parentStyle)->addCachedPseudoStyle(style.release());
    return *computedStyle;
}

RenderStyle* Element::computedStyle(PseudoId pseudoElementSpecifier)
{
    if (!inDocument())
        return nullptr;

    if (PseudoElement* pseudoElement = beforeOrAfterPseudoElement(*this, pseudoElementSpecifier))
        return pseudoElement->computedStyle();

    auto* style = existingComputedStyle();
    if (!style)
        style = resolveComputedStyle();

    if (pseudoElementSpecifier != PseudoId::None) {
        if (auto* cachedPseudoStyle = style->getCachedPseudoStyle(pseudoElementSpecifier))
            return cachedPseudoStyle;
        return const_cast<RenderStyle*>(&resolvePseudoElementStyle(pseudoElementSpecifier));
    }

    return const_cast<RenderStyle*>(style);
}

bool Element::needsStyleInvalidation() const
{
    if (!inRenderedDocument())
        return false;
    if (styleValidity() >= Style::Validity::SubtreeInvalid)
        return false;
    if (document().hasPendingFullStyleRebuild())
        return false;
    return true;
}

void Element::setChildIndex(unsigned index)
{
    ElementRareData& rareData = ensureElementRareData();
    rareData.setChildIndex(index);
}

bool Element::hasFlagsSetDuringStylingOfChildren() const
{
    return childrenAffectedByFirstChildRules()
        || childrenAffectedByLastChildRules()
        || childrenAffectedByForwardPositionalRules()
        || descendantsAffectedByForwardPositionalRules()
        || childrenAffectedByBackwardPositionalRules()
        || descendantsAffectedByBackwardPositionalRules();
}

unsigned Element::rareDataChildIndex() const
{
    ASSERT(hasRareData());
    return elementRareData()->childIndex();
}

AtomicString Element::computeInheritedLanguage() const
{
    const Node* n = this;
    AtomicString value;
    // The language property is inherited, so we iterate over the parents to find the first language.
    while (n && value.isNull()) {
        if (n->isElementNode()) {
        	if (const ElementData* elementData = static_cast<const Element*>(n)->elementData()) {
        		// Spec: xml:lang takes precedence -- http://www.w3.org/TR/xhtml1/#C_7
				if (const Attribute* attribute = elementData->findAttributeByName(XMLNames::langAttr))
					value = attribute->value();
				else if (const Attribute* attribute = elementData->findAttributeByName(HTMLNames::langAttr))
					value = attribute->value();
        	}
        } else if (n->isDocumentNode()) {
            // checking the MIME content-language
            value = static_cast<const Document*>(n)->contentLanguage();
        }

        n = n->parentNode();
    }

    return value;
}

Locale& Element::locale() const
{
    return document().getCachedLocale(computeInheritedLanguage());
}

void Element::normalizeAttributes()
{
    if (!hasAttributes())
        return;

    AttrNodeList* attrNodeList = attrNodeListForElement(this);
    if (!attrNodeList)
        return;

    // Copy the Attr Vector because Node::normalize() can fire synchronous JS
    // events (e.g. DOMSubtreeModified) and a JS listener could add / remove
    // attributes while we are iterating.
    AttrNodeList copyOfAttrNodeList = *attrNodeList;

    for (size_t i = 0; i < copyOfAttrNodeList.size(); ++i) {
        if (RefPtr<Attr> attr = copyOfAttrNodeList[i])
            attr->normalize();
    }
}

PseudoElement& Element::ensurePseudoElement(PseudoId pseudoId)
{
    if (pseudoId == PseudoId::Before) {
        if (!beforePseudoElement())
            ensureElementRareData().setBeforePseudoElement(PseudoElement::create(*this, pseudoId));
        return *beforePseudoElement();
    }

    ASSERT(pseudoId == PseudoId::After);
    if (!afterPseudoElement())
        ensureElementRareData().setAfterPseudoElement(PseudoElement::create(*this, pseudoId));
    return *afterPseudoElement();
}

static void disconnectPseudoElement(PseudoElement* pseudoElement)
{
	if (!pseudoElement)
		return;
    ASSERT(!pseudoElement->renderer());
	ASSERT(pseudoElement->hostElement());
	pseudoElement->clearHostElement();
}

void Element::clearBeforePseudoElementSlow()
{
    ASSERT(hasRareData());
	disconnectPseudoElement(elementRareData()->beforePseudoElement());
	elementRareData()->setBeforePseudoElement(nullptr);
}

void Element::clearAfterPseudoElementSlow()
{
    ASSERT(hasRareData());
	disconnectPseudoElement(elementRareData()->afterPseudoElement());
	elementRareData()->setAfterPseudoElement(nullptr);
}

bool Element::matchesValidPseudoClass() const
{
    return false;
}

bool Element::matchesInvalidPseudoClass() const
{
    return false;
}

bool Element::matchesReadWritePseudoClass() const
{
    return false;
}

bool Element::matchesIndeterminatePseudoClass() const
{
    return shouldAppearIndeterminate();
}

bool Element::matchesDefaultPseudoClass() const
{
    return false;
}

bool Element::matches(const String& selector, ExceptionCode& ec)
{
    SelectorQuery* selectorQuery = document().selectorQueryForString(selector, ec);
    return selectorQuery && selectorQuery->matches(*this);
}

Element* Element::closest(const String& selector, ExceptionCode& ec)
{
    SelectorQuery* selectorQuery = document().selectorQueryForString(selector, ec);
    if (selectorQuery)
        return selectorQuery->closest(*this);
    return 0;
}

bool Element::shouldAppearIndeterminate() const
{
    return false;
}

DOMTokenList* Element::classList()
{
    ElementRareData& data = ensureElementRareData();
    if (!data.classList())
		data.setClassList(adoptPtr(new DOMTokenList(*this, HTMLNames::classAttr)));
	return data.classList();
}

SpaceSplitString Element::partNames() const
{
    return hasRareData() ? elementRareData()->partNames() : SpaceSplitString();
}

DOMTokenList& Element::part()
{
    auto& data = ensureElementRareData();
    if (!data.partList())
        data.setPartList(adoptPtr(new DOMTokenList(*this, HTMLNames::partAttr)));
    return *data.partList();
}

DatasetDOMStringMap* Element::dataset()
{
    ElementRareData& data = ensureElementRareData();
    if (!data.dataset())
        data.setDataset(DatasetDOMStringMap::create(this));
    return data.dataset();
}

KURL Element::getURLAttribute(const QualifiedName& name) const
{
#if !ASSERT_DISABLED
    if (elementData()) {
        if (const Attribute* attribute = findAttributeByName(name))
            ASSERT(isURLAttribute(*attribute));
    }
#endif
    return document().completeURL(getAttribute(name));
}

KURL Element::getNonEmptyURLAttribute(const QualifiedName& name) const
{
#if !ASSERT_DISABLED
    if (elementData()) {
        if (const Attribute* attribute = findAttributeByName(name))
            ASSERT(isURLAttribute(*attribute));
    }
#endif
    String value = stripLeadingAndTrailingHTMLSpaces(getAttribute(name));
    if (value.isEmpty())
        return KURL();
    return document().completeURL(value);
}

int Element::getIntegralAttribute(const QualifiedName& attributeName) const
{
    return parseHTMLInteger(getAttribute(attributeName)).valueOr(0);
}

void Element::setIntegralAttribute(const QualifiedName& attributeName, int value)
{
    setAttribute(attributeName, AtomicString::number(value));
}

unsigned Element::getUnsignedIntegralAttribute(const QualifiedName& attributeName) const
{
    return parseHTMLNonNegativeInteger(getAttribute(attributeName)).valueOr(0);
}

void Element::setUnsignedIntegralAttribute(const QualifiedName& attributeName, unsigned value)
{
    // FIXME: Need an AtomicString version of String::number.
    setAttribute(attributeName, String::number(value));
}

bool Element::childShouldCreateRenderer(const Node* child) const
{
#if ENABLE(SVG)
    // Only create renderers for SVG elements whose parents are SVG elements, or for proper <svg xmlns="svgNS"> subdocuments.
	if (child->isSVGElement()) {
		ASSERT(!isSVGElement());
		return child->hasTagName(SVGNames::svgTag) && toSVGElement(child)->isValid();
	}
#endif

    return true;
}

#if ENABLE(FULLSCREEN_API)
void Element::webkitRequestFullscreen()
{
    document().requestFullScreenForElement(this, ALLOW_KEYBOARD_INPUT, Document::EnforceIFrameAllowFullScreenRequirement);
}

void Element::webkitRequestFullScreen(unsigned short flags)
{
    document().requestFullScreenForElement(this, (flags | LEGACY_MOZILLA_REQUEST), Document::EnforceIFrameAllowFullScreenRequirement);
}

void Element::setContainsFullScreenElement(bool flag)
{
    if (flag)
    	setNodeFlag(ContainsFullScreenElementFlag);
    else
        clearFlag(ContainsFullScreenElementFlag);
    invalidateStyleAndLayerComposition();
}

static Element* parentCrossingFrameBoundaries(Element* element)
{
    ASSERT(element);
    return element->parentElement() ? element->parentElement() : element->document().ownerElement();
}

void Element::setContainsFullScreenElementOnAncestorsCrossingFrameBoundaries(bool flag)
{
    Element* element = this;
    while ((element = parentCrossingFrameBoundaries(element)))
        element->setContainsFullScreenElement(flag);
}
#endif    

#if ENABLE(POINTER_LOCK)
void Element::requestPointerLock()
{
    if (document().page())
        document().page()->pointerLockController()->requestPointerLock(this);
}
#endif

ElementAnimationRareData* Element::animationRareData(PseudoId pseudoId) const
{
    return hasRareData() ? elementRareData()->animationRareData(pseudoId) : nullptr;
}

ElementAnimationRareData& Element::ensureAnimationRareData(PseudoId pseudoId)
{
    return ensureElementRareData().ensureAnimationRareData(pseudoId);
}

KeyframeEffectStack* Element::keyframeEffectStack(PseudoId pseudoId) const
{
    if (auto* animationData = animationRareData(pseudoId))
        return animationData->keyframeEffectStack();
    return nullptr;
}

KeyframeEffectStack& Element::ensureKeyframeEffectStack(PseudoId pseudoId)
{
    return ensureAnimationRareData(pseudoId).ensureKeyframeEffectStack();
}

bool Element::hasKeyframeEffects(PseudoId pseudoId) const
{
    if (auto* animationData = animationRareData(pseudoId)) {
        if (auto* keyframeEffectStack = animationData->keyframeEffectStack())
            return keyframeEffectStack->hasEffects();
    }
    return false;
}

const AnimationCollection* Element::animations(PseudoId pseudoId) const
{
    if (auto* animationData = animationRareData(pseudoId))
        return &animationData->animations();
    return nullptr;
}

bool Element::hasCompletedTransitionsForProperty(PseudoId pseudoId, CSSPropertyID property) const
{
    if (auto* animationData = animationRareData(pseudoId))
        return animationData->completedTransitionsByProperty().contains(property);
    return false;
}

bool Element::hasRunningTransitionsForProperty(PseudoId pseudoId, CSSPropertyID property) const
{
    if (auto* animationData = animationRareData(pseudoId))
        return animationData->runningTransitionsByProperty().contains(property);
    return false;
}

bool Element::hasRunningTransitions(PseudoId pseudoId) const
{
    if (auto* animationData = animationRareData(pseudoId))
        return !animationData->runningTransitionsByProperty().isEmpty();
    return false;
}

AnimationCollection& Element::ensureAnimations(PseudoId pseudoId)
{
    return ensureAnimationRareData(pseudoId).animations();
}

CSSAnimationCollection& Element::animationsCreatedByMarkup(PseudoId pseudoId)
{
    return ensureAnimationRareData(pseudoId).animationsCreatedByMarkup();
}

void Element::setAnimationsCreatedByMarkup(PseudoId pseudoId, CSSAnimationCollection&& animations)
{
    if (animations.isEmpty() && !animationRareData(pseudoId))
        return;
    ensureAnimationRareData(pseudoId).setAnimationsCreatedByMarkup(WTFMove(animations));
}

PropertyToTransitionMap& Element::ensureCompletedTransitionsByProperty(PseudoId pseudoId)
{
    return ensureAnimationRareData(pseudoId).completedTransitionsByProperty();
}

PropertyToTransitionMap& Element::ensureRunningTransitionsByProperty(PseudoId pseudoId)
{
    return ensureAnimationRareData(pseudoId).runningTransitionsByProperty();
}

const RenderStyle* Element::lastStyleChangeEventStyle(PseudoId pseudoId) const
{
    if (auto* animationData = animationRareData(pseudoId))
        return animationData->lastStyleChangeEventStyle();
    return nullptr;
}

void Element::setLastStyleChangeEventStyle(PseudoId pseudoId, PassOwnPtr<const RenderStyle> style)
{
    if (auto* animationData = animationRareData(pseudoId))
        animationData->setLastStyleChangeEventStyle(style);
    else if (style)
    	ensureAnimationRareData(pseudoId).setLastStyleChangeEventStyle(style);
}

void Element::cssAnimationsDidUpdate(PseudoId pseudoId)
{
    ensureAnimationRareData(pseudoId).cssAnimationsDidUpdate();
}

void Element::keyframesRuleDidChange(PseudoId pseudoId)
{
    ensureAnimationRareData(pseudoId).keyframesRuleDidChange();
}

bool Element::hasPendingKeyframesUpdate(PseudoId pseudoId) const
{
    auto* data = animationRareData(pseudoId);
    return data && data->hasPendingKeyframesUpdate();
}

#if ENABLE(INTERSECTION_OBSERVER)
void Element::disconnectFromIntersectionObservers()
{
    auto* observerData = intersectionObserverDataIfExists();
    if (!observerData)
        return;

    for (auto& registration : observerData->registrations) {
        if (registration.observer)
        	registration.observer->targetDestroyed(*this);
    }
    observerData->registrations.clear();

    for (auto& observer : observerData->observers) {
        if (observer)
        	observer->rootDestroyed();
    }
    observerData->observers.clear();
}

IntersectionObserverData& Element::ensureIntersectionObserverData()
{
    auto& rareData = ensureElementRareData();
    if (!rareData.intersectionObserverData())
        rareData.setIntersectionObserverData(adoptPtr(new IntersectionObserverData()));
    return *rareData.intersectionObserverData();
}

IntersectionObserverData* Element::intersectionObserverDataIfExists()
{
    return hasRareData() ? elementRareData()->intersectionObserverData() : nullptr;
}
#endif

#if ENABLE(RESIZE_OBSERVER)
void Element::disconnectFromResizeObservers()
{
    auto* observerData = resizeObserverData();
    if (!observerData)
        return;

    for (const auto& observer : observerData->observers) {
        if (observer)
        	observer->targetDestroyed(*this);
    }
    observerData->observers.clear();
}

ResizeObserverData& Element::ensureResizeObserverData()
{
    auto& rareData = ensureElementRareData();
    if (!rareData.resizeObserverData())
        rareData.setResizeObserverData(adoptPtr(new ResizeObserverData()));
    return *rareData.resizeObserverData();
}

ResizeObserverData* Element::resizeObserverData()
{
    return hasRareData() ? elementRareData()->resizeObserverData() : nullptr;
}
#endif

SpellcheckAttributeState Element::spellcheckAttributeState() const
{
    const AtomicString& value = getAttribute(HTMLNames::spellcheckAttr);
    if (value.isNull())
        return SpellcheckAttributeDefault;
    if (value.isEmpty() || equalLettersIgnoringASCIICase(value, "true"))
        return SpellcheckAttributeTrue;
    if (equalLettersIgnoringASCIICase(value, "false"))
        return SpellcheckAttributeFalse;

    return SpellcheckAttributeDefault;
}

bool Element::isSpellCheckingEnabled() const
{
    const Element* element = this;
    while (element) {
        switch (element->spellcheckAttributeState()) {
        case SpellcheckAttributeTrue:
            return true;
        case SpellcheckAttributeFalse:
            return false;
        case SpellcheckAttributeDefault:
            break;
        }

        ContainerNode* parent = const_cast<Element*>(element)->parentOrShadowHostNode();
        if (parent && parent->isElementNode())
            element = toElement(parent);
        else if (parent && parent->isShadowRoot())
            element = toElement(parent->parentOrShadowHostNode());
        else
            element = 0;
    }

    return true;
}

#ifndef NDEBUG
bool Element::fastAttributeLookupAllowed(const QualifiedName& name) const
{
    if (name == HTMLNames::styleAttr)
        return false;

#if ENABLE(SVG)
    if (isSVGElement())
        return !static_cast<const SVGElement*>(this)->isAnimatedPropertyAttribute(name);
#endif

    return true;
}
#endif

#if defined(DUMP_NODE_STATISTICS) && DUMP_NODE_STATISTICS
bool Element::hasNamedNodeMap() const
{
    return hasRareData() && elementRareData()->attributeMap();
}
#endif

inline void Element::updateName(const AtomicString& oldName, const AtomicString& newName)
{
    if (!isInTreeScope())
        return;

    if (oldName == newName)
        return;

    updateNameForTreeScope(&treeScope(), oldName, newName);

    if (!inDocument())
        return;
    Document& htmlDocument = document();
    if (!htmlDocument.isHTMLDocument())
        return;
    updateNameForDocument(&toHTMLDocument(htmlDocument), oldName, newName);
}

void Element::updateNameForTreeScope(TreeScope* scope, const AtomicString& oldName, const AtomicString& newName)
{
    ASSERT(oldName != newName);

    if (!oldName.isEmpty())
        scope->removeElementByName(oldName, this);
    if (!newName.isEmpty())
        scope->addElementByName(newName, this);
}

void Element::updateNameForDocument(HTMLDocument* document, const AtomicString& oldName, const AtomicString& newName)
{
    ASSERT(oldName != newName);

    if (isInShadowTree())
        return;

    if (WindowNameCollection::nodeMatchesIfNameAttributeMatch(this)) {
    	const AtomicString& id = WindowNameCollection::nodeMatchesIfIdAttributeMatch(this) ? getIdAttribute() : nullAtom;
		if (!oldName.isEmpty() && oldName != id)
            document->removeWindowNamedItem(oldName, this);
		if (!newName.isEmpty() && newName != id)
            document->addWindowNamedItem(newName, this);
    }

    if (DocumentNameCollection::nodeMatchesIfNameAttributeMatch(this)) {
    	const AtomicString& id = DocumentNameCollection::nodeMatchesIfIdAttributeMatch(this) ? getIdAttribute() : nullAtom;
		if (!oldName.isEmpty() && oldName != id)
            document->removeDocumentNamedItem(oldName, this);
		if (!newName.isEmpty() && newName != id)
            document->addDocumentNamedItem(newName, this);
    }
}

inline void Element::updateId(const AtomicString& oldId, const AtomicString& newId, NotifyObservers notifyObservers)
{
    if (!isInTreeScope())
        return;

    if (oldId == newId)
        return;

    updateIdForTreeScope(&treeScope(), oldId, newId, notifyObservers);

    if (!inDocument())
        return;
    Document& htmlDocument = document();
    if (!htmlDocument.isHTMLDocument())
        return;
    updateIdForDocument(&toHTMLDocument(htmlDocument), oldId, newId, UpdateHTMLDocumentNamedItemMapsOnlyIfDiffersFromNameAttribute);
}

void Element::updateIdForTreeScope(TreeScope* scope, const AtomicString& oldId, const AtomicString& newId, NotifyObservers notifyObservers)
{
    ASSERT(isInTreeScope());
    ASSERT(oldId != newId);

    if (!oldId.isEmpty())
        scope->removeElementById(oldId, this, notifyObservers == NotifyObservers::Yes);
    if (!newId.isEmpty())
        scope->addElementById(newId, this, notifyObservers == NotifyObservers::Yes);
}

void Element::updateIdForDocument(HTMLDocument* document, const AtomicString& oldId, const AtomicString& newId, HTMLDocumentNamedItemMapsUpdatingCondition condition)
{
    ASSERT(inDocument());
    ASSERT(oldId != newId);

    if (isInShadowTree())
        return;

    if (WindowNameCollection::nodeMatchesIfIdAttributeMatch(this)) {
    	const AtomicString& name = condition == UpdateHTMLDocumentNamedItemMapsOnlyIfDiffersFromNameAttribute && WindowNameCollection::nodeMatchesIfNameAttributeMatch(this) ? getNameAttribute() : nullAtom;
		if (!oldId.isEmpty() && oldId != name)
            document->removeWindowNamedItem(oldId, this);
		if (!newId.isEmpty() && newId != name)
            document->addWindowNamedItem(newId, this);
    }

    if (DocumentNameCollection::nodeMatchesIfIdAttributeMatch(this)) {
    	const AtomicString& name = condition == UpdateHTMLDocumentNamedItemMapsOnlyIfDiffersFromNameAttribute && DocumentNameCollection::nodeMatchesIfNameAttributeMatch(this) ? getNameAttribute() : nullAtom;
		if (!oldId.isEmpty() && oldId != name)
            document->removeDocumentNamedItem(oldId, this);
		if (!newId.isEmpty() && newId != name)
            document->addDocumentNamedItem(newId, this);
    }
}

void Element::willModifyAttribute(const QualifiedName& name, const AtomicString& oldValue, const AtomicString& newValue)
{
    if (name == HTMLNames::idAttr)
    	updateId(oldValue, newValue, NotifyObservers::No); // Will notify observers after the attribute is actually changed.
	else if (name == HTMLNames::nameAttr)
        updateName(oldValue, newValue);

#if ENABLE(MUTATION_OBSERVERS)
    if (OwnPtr<MutationObserverInterestGroup> recipients = MutationObserverInterestGroup::createForAttributesMutation(this, name))
        recipients->enqueueMutationRecord(MutationRecord::createAttributes(this, name, oldValue));
#endif

#if ENABLE(INSPECTOR)
    InspectorInstrumentation::willModifyDOMAttr(&document(), this, oldValue, newValue);
#endif
}

void Element::didAddAttribute(const QualifiedName& name, const AtomicString& value)
{
	notifyAttributeChanged(name, nullAtom, value);
    InspectorInstrumentation::didModifyDOMAttr(&document(), this, name.localName(), value);
    dispatchSubtreeModifiedEvent();
}

void Element::didModifyAttribute(const QualifiedName& name, const AtomicString& oldValue, const AtomicString& newValue)
{
	notifyAttributeChanged(name, oldValue, newValue);
    InspectorInstrumentation::didModifyDOMAttr(&document(), this, name.localName(), newValue);
    // Do not dispatch a DOMSubtreeModified event here; see bug 81141.
}

void Element::didRemoveAttribute(const QualifiedName& name, const AtomicString& oldValue)
{
	notifyAttributeChanged(name, oldValue, nullAtom);
    InspectorInstrumentation::didRemoveDOMAttr(&document(), this, name.localName());
    dispatchSubtreeModifiedEvent();
}

IntPoint Element::savedLayerScrollPosition() const
{
    return hasRareData() ? elementRareData()->savedLayerScrollPosition() : IntPoint();
}

void Element::setSavedLayerScrollPositionSlow(const IntPoint& position)
{
    ASSERT(!position.isZero() || hasRareData());
    ensureElementRareData().setSavedLayerScrollPosition(position);
}

PassRefPtr<Attr> Element::attrIfExists(const QualifiedName& name)
{
    if (auto* attrNodeList = attrNodeListForElement(this))
        return findAttrNodeInList(*attrNodeList, name);
    return nullptr;
}

Ref<Attr> Element::ensureAttr(const QualifiedName& name)
{
    auto* attrNodeList = ensureAttrNodeListForElement(this);
    RefPtr<Attr> attrNode = findAttrNodeInList(*attrNodeList, name);
    if (!attrNode) {
        attrNode = Attr::create(*this, name);
        attrNode->setTreeScopeRecursively(treeScope());
        attrNodeList->append(attrNode);
    }
    return attrNode.releaseNonNull();
}

void Element::detachAttrNodeFromElementWithValue(Attr* attrNode, const AtomicString& value)
{
    ASSERT(hasSyntheticAttrChildNodes());
    attrNode->detachFromElementWithValue(value);

    AttrNodeList* attrNodeList = attrNodeListForElement(this);
    for (unsigned i = 0; i < attrNodeList->size(); ++i) {
        if (attrNodeList->at(i)->qualifiedName() == attrNode->qualifiedName()) {
            attrNodeList->remove(i);
            if (attrNodeList->isEmpty())
                removeAttrNodeListForElement(this);
            return;
        }
    }
    ASSERT_NOT_REACHED();
}

void Element::detachAllAttrNodesFromElement()
{
    AttrNodeList* attrNodeList = attrNodeListForElement(this);
    ASSERT(attrNodeList);

    auto attributes = attributesIterator();
    for(auto iter = attributes.begin(); iter != attributes.end(); ++iter) {
        const Attribute& attribute = *iter;
        if (RefPtr<Attr> attrNode = findAttrNodeInList(*attrNodeList, attribute.name()))
            attrNode->detachFromElementWithValue(attribute.value());
    }

    removeAttrNodeListForElement(this);
}

void Element::doResetComputedStyle(Element& element)
{
	if (!element.hasRareData() || !element.elementRareData()->computedStyle())
		return;
	if (element.hasCustomStyleResolveCallbacks())
		element.willResetComputedStyle();
	element.elementRareData()->resetComputedStyle();
}

void Element::resetComputedStyle()
{
    if (!hasRareData() || !elementRareData()->computedStyle())
        return;

    doResetComputedStyle(*this);
    //  Take care to reset computed style in all descendants. attachRenderTree no longer does this.
    auto descendantsElement = descendantsOfType<Element>(this);
    for (auto iter = descendantsElement.begin(); iter != descendantsElement.end(); ++iter)
    	doResetComputedStyle(*iter);
}

void Element::resetStyleRelations()
{
    clearStyleFlags(NodeStyleFlag::StyleAffectedByEmpty);
    if (!hasRareData())
        return;
    elementRareData()->setChildIndex(0);
}

void Element::resetChildStyleRelations()
{
	OptionSet<NodeStyleFlag> flags;
	flags.add(NodeStyleFlag::ChildrenAffectedByFirstChildRules);
	flags.add(NodeStyleFlag::ChildrenAffectedByLastChildRules);
	flags.add(NodeStyleFlag::ChildrenAffectedByForwardPositionalRules);
	flags.add(NodeStyleFlag::ChildrenAffectedByBackwardPositionalRules);
	clearStyleFlags(flags);
}

void Element::resetAllDescendantStyleRelations()
{
    resetChildStyleRelations();

	OptionSet<NodeStyleFlag> flags;
	flags.add(NodeStyleFlag::DescendantsAffectedByForwardPositionalRules);
	flags.add(NodeStyleFlag::DescendantsAffectedByBackwardPositionalRules);
	clearStyleFlags(flags);
}

void Element::clearHoverAndActiveStatusBeforeDetachingRenderer()
{
    if (!isUserActionElement())
        return;
    if (hovered())
        document().hoveredElementDidDetach(this);
    if (inActiveChain())
        document().elementInActiveChainDidDetach(this);
    document().userActionElements().clearActiveAndHovered(*this);
}

bool Element::willRecalcStyle(Style::Change)
{
    ASSERT(hasCustomStyleResolveCallbacks());
    return true;
}

void Element::didRecalcStyle(Style::Change)
{
    ASSERT(hasCustomStyleResolveCallbacks());
}

void Element::willAttachRenderers()
{
    ASSERT(hasCustomStyleResolveCallbacks());
}

void Element::didAttachRenderers()
{
    ASSERT(hasCustomStyleResolveCallbacks());
}

void Element::willDetachRenderers()
{
    ASSERT(hasCustomStyleResolveCallbacks());
}

void Element::didDetachRenderers()
{
    ASSERT(hasCustomStyleResolveCallbacks());
}

Optional<ElementStyle> Element::resolveCustomStyle(const Style::ResolutionContext&, const RenderStyle*)
{
    ASSERT(hasCustomStyleResolveCallbacks());
    return Nullopt;
}

void Element::willResetComputedStyle()
{
    ASSERT(hasCustomStyleResolveCallbacks());
}

void Element::cloneAttributesFromElement(const Element& other)
{
    if (hasSyntheticAttrChildNodes())
        detachAllAttrNodesFromElement();

    other.synchronizeAllAttributes();
    if (!other.m_elementData) {
        m_elementData.clear();
        return;
    }

    // We can't update window and document's named item maps since the presence of image and object elements depend on other attributes and children.
    // Fortunately, those named item maps are only updated when this element is in the document, which should never be the case.
    ASSERT(!inDocument());

    const AtomicString& oldID = getIdAttribute();
    const AtomicString& newID = other.getIdAttribute();

    if (!oldID.isNull() || !newID.isNull())
    	updateId(oldID, newID, NotifyObservers::No); // Will notify observers after the attribute is actually changed.

    const AtomicString& oldName = getNameAttribute();
    const AtomicString& newName = other.getNameAttribute();

    if (!oldName.isNull() || !newName.isNull())
        updateName(oldName, newName);

    // If 'other' has a mutable ElementData, convert it to an immutable one so we can share it between both elements.
    // We can only do this if there is no CSSOM wrapper for other's inline style, and there are no presentation attributes.
	if (other.m_elementData->isUnique()
		&& !other.m_elementData->presentationAttributeStyle()
		&& (!other.m_elementData->inlineStyle() || !other.m_elementData->inlineStyle()->hasCSSOMWrapper()))
		const_cast<Element&>(other).m_elementData = static_cast<const UniqueElementData*>(other.m_elementData.get())->makeShareableCopy();

    if (!other.m_elementData->isUnique())
        m_elementData = other.m_elementData;
    else
        m_elementData = other.m_elementData->makeUniqueCopy();

    for (const Attribute& attribute : attributesIterator())
    	notifyAttributeChanged(attribute.name(), nullAtom, attribute.value(), ModifiedByCloning);

    setNonce(other.nonce());
}

void Element::cloneDataFromElement(const Element& other)
{
    cloneAttributesFromElement(other);
    copyNonAttributePropertiesFromElement(other);
}

void Element::createUniqueElementData()
{
    if (!m_elementData)
        m_elementData = UniqueElementData::create();
    else {
        ASSERT(!m_elementData->isUnique());
        m_elementData = static_cast<ShareableElementData*>(m_elementData.get())->makeUniqueCopy();
    }
}

Node* Element::insertAdjacent(const String& where, Ref<Node>&& newChild, ExceptionCode& ec)
{
    // In Internet Explorer if the element has no parent and where is "beforeBegin" or "afterEnd",
    // a document fragment is created and the elements appended in the correct order. This document
    // fragment isn't returned anywhere.
    //
    // This is impossible for us to implement as the DOM tree does not allow for such structures,
    // Opera also appears to disallow such usage.

    if (equalLettersIgnoringASCIICase(where, "beforebegin")) {
        ContainerNode* parent = this->parentNode();
        return (parent && parent->insertBefore(newChild, this, ec)) ? newChild.ptr() : nullptr;
    }

    if (equalLettersIgnoringASCIICase(where, "afterbegin"))
        return insertBefore(newChild, firstChild(), ec) ? newChild.ptr() : nullptr;

    if (equalLettersIgnoringASCIICase(where, "beforeend"))
        return appendChild(newChild, ec) ? newChild.ptr() : nullptr;

    if (equalLettersIgnoringASCIICase(where, "afterend")) {
        ContainerNode* parent = this->parentNode();
        return (parent && parent->insertBefore(newChild, nextSibling(), ec)) ? newChild.ptr() : nullptr;
    }

    ec = SYNTAX_ERR;
    return nullptr;
}

Element* Element::insertAdjacentElement(const String& where, Element& newChild, ExceptionCode& ec)
{
    Node* returnValue = insertAdjacent(where, Ref<Node>(newChild), ec);
    ASSERT_WITH_SECURITY_IMPLICATION(!returnValue || is<Element>(*returnValue));
    return toElement(returnValue);
}

// Step 1 of https://w3c.github.io/DOM-Parsing/#dom-element-insertadjacenthtml.
static ContainerNode* contextNodeForInsertion(const String& where, Element* element, ExceptionCode& ec)
{
    if (equalLettersIgnoringASCIICase(where, "beforebegin") || equalLettersIgnoringASCIICase(where, "afterend")) {
    	auto* parent = element->parentNode();
		if (!parent || isDocument(*parent)) {
            ec = NO_MODIFICATION_ALLOWED_ERR;
            return nullptr;
        }
        return parent;
    }
    if (equalLettersIgnoringASCIICase(where, "afterbegin") || equalLettersIgnoringASCIICase(where, "beforeend"))
        return element;
    ec =  SYNTAX_ERR;
    return nullptr;
}

// https://w3c.github.io/DOM-Parsing/#dom-element-insertadjacenthtml
void Element::insertAdjacentHTML(const String& where, const String& markup, ExceptionCode& ec)
{
    // Step 1.
    auto* contextNode = contextNodeForInsertion(where, this, ec);
    if (!contextNode)
        return;
    RefPtr<Element> contextElement;
    // Step 2.
    if (!isElement(*contextNode)
        || (contextNode->document().isHTMLDocument() && isHTMLHtmlElement(*contextNode))) {
        contextElement = HTMLBodyElement::create(&contextNode->document());
    } else
        contextElement = toElement(contextNode);
    // Step 3.
    RefPtr<DocumentFragment> fragment = createFragmentForInnerOuterHTML(markup, contextElement.get(), AllowScriptingContent, ec);
    if (!fragment)
        return;
    // Step 4.
    insertAdjacent(where, fragment.releaseNonNull(), ec);
}

void Element::insertAdjacentText(const String& where, const String& text, ExceptionCode& ec)
{
    insertAdjacent(where, document().createTextNode(text), ec);
}

PassRefPtr<WebAnimation> Element::animate(JSC::ExecState* state, JSC::JSValue keyframes, ExceptionCode& ec)
{
    return animate(state, keyframes, JSC::JSValue(), ec);
}

PassRefPtr<WebAnimation> Element::animate(JSC::ExecState* state, JSC::JSValue keyframes, JSC::JSValue options, ExceptionCode& ec)
{
    String id = "";
    //Optional<Variant<double, KeyframeEffectOptions>> keyframeEffectOptions;
    JSC::JSValue keyframeEffectOptions;
    if (options) {
        //Variant<double, KeyframeEffectOptions> keyframeEffectOptionsVariant;
        if (options.isNumber())
        	keyframeEffectOptions = JSC::JSValue(options.asDouble());
        else if(Optional<KeyframeAnimationOptions> KeyframeAnimationOptionsResult = convertDictionary<KeyframeAnimationOptions>(*state, options)) {
            id = KeyframeAnimationOptionsResult->id;
            keyframeEffectOptions = options;
        }
    }

    RefPtr<KeyframeEffect> keyframeEffectResult = KeyframeEffect::create(*state, this, keyframes, keyframeEffectOptions, ec);
    if (ec)
        return nullptr;

    RefPtr<WebAnimation> animation = WebAnimation::create(&document(), keyframeEffectResult.get());
    animation->setId(id);

    animation->play(ec);
    if (ec)
        return nullptr;

    return animation.release();
}

Vector<RefPtr<WebAnimation>> Element::getAnimations(Optional<GetAnimationsOptions> options)
{
    // If we are to return animations in the subtree, we can get all of the document's animations and filter
    // animations targeting that are not registered on this element, one of its pseudo elements or a child's
    // pseudo element.
    if (options && options->subtree) {
    	struct MatchingAnimationsFunctionImpl FINAL : public Document::MatchingAnimationsFunction
		{
			MatchingAnimationsFunctionImpl(const Element& thisElement) : thisElement(thisElement) { }
			virtual bool operator()(Element& target) const FINAL {
	            if (target.isPseudoElement())
	                return thisElement.contains(toPseudoElement(target).hostElement());
	            return thisElement.contains(&target);
	        }
			const Element& thisElement;
		};
    	MatchingAnimationsFunctionImpl function(*this);
        return document().matchingAnimations(function);
    }

    // For the list of animations to be current, we need to account for any pending CSS changes,
    // such as updates to CSS Animations and CSS Transitions. This requires updating layout as
    // well since resolving layout-dependent media queries could yield animations.
    // FIXME: We might be able to use ComputedStyleExtractor which is more optimized.
    if (RefPtr<HTMLFrameOwnerElement> owner = document().ownerElement())
        owner->document().updateLayout();
    document().updateStyleIfNeeded();

    Vector<RefPtr<WebAnimation>> animations;
    if (auto* effectStack = keyframeEffectStack(PseudoId::None)) {
        for (auto& effect : effectStack->sortedEffects()) {
            if (effect->animation()->isRelevant())
                animations.append(effect->animation());
        }
    }
    return animations;
}

#if 0
static WeakHashMap<Element, ElementIdentifier>& elementIdentifiersMap()
{
    static MainThreadNeverDestroyed<WeakHashMap<Element, ElementIdentifier>> map;
    return map;
}

ElementIdentifier Element::identifier() const
{
    return elementIdentifiersMap().ensure(*this, [] { return ElementIdentifier::generate(); }).iterator->value;
}

Element* Element::fromIdentifier(ElementIdentifier identifier)
{
    for (auto [element, elementIdentifier] : elementIdentifiersMap()) {
        if (elementIdentifier == identifier)
            return &element;
    }
    return nullptr;
}
#endif

} // namespace WebCore
