/*
 * Copyright (C) 2016 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ErrorEvent.h"

#include "DOMWrapperWorld.h"
#include "EventNames.h"
#include "MessagePort.h"

using namespace JSC;

namespace WebCore {

ErrorEvent::ErrorEvent(const AtomicString& type, const Init& initializer)
    : Event(type, initializer, IsTrusted::No)
    , m_message(initializer.message)
    , m_fileName(initializer.filename)
    , m_lineNumber(initializer.lineno)
    , m_columnNumber(initializer.colno)
    , m_triedToSerialize(false)
{
	m_error.setWithoutWriteBarrier(initializer.error);
}

ErrorEvent::ErrorEvent(const String& message, const String& fileName, unsigned lineNumber, unsigned columnNumber, JSC::JSValue error)
    : Event(eventNames().errorEvent, false, true)
    , m_message(message)
    , m_fileName(fileName)
    , m_lineNumber(lineNumber)
    , m_columnNumber(columnNumber)
    , m_triedToSerialize(false)
{
	m_error.setWithoutWriteBarrier(error);
}

ErrorEvent::~ErrorEvent() = default;

EventInterface ErrorEvent::eventInterface() const
{
    return ErrorEventInterfaceType;
}

JSC::JSValue ErrorEvent::error() const
{
	return m_error.get();
}

JSValue ErrorEvent::sanitizedErrorValue(ExecState& exec, JSGlobalObject& globalObject)
{
    JSValue error = m_error.get();
    if (!error)
        return jsNull();

    if (error.isObject() && worldForDOMObject(error.getObject()) != currentWorld(&exec)) {
        // We need to make sure ErrorEvents do not leak their error property across isolated DOM worlds.
        // Ideally, we would check that the worlds have different privileges but that's not possible yet.
        RefPtr<SerializedScriptValue> serializedError = trySerializeError(exec);
        if (!serializedError)
            return jsNull();
        Vector<RefPtr<MessagePort>> dummyMessagePorts;
        return serializedError->deserialize(&exec, &globalObject, dummyMessagePorts);
    }

    return error;
}

RefPtr<SerializedScriptValue> ErrorEvent::trySerializeError(ExecState& exec)
{
    if (!m_serializedError && !m_triedToSerialize) {
        Vector<RefPtr<MessagePort>> dummyMessagePorts;
        Vector<RefPtr<JSC::ArrayBuffer>> dummyArrayBuffers;
        m_serializedError = SerializedScriptValue::create(&exec, m_error.get(), dummyMessagePorts, dummyArrayBuffers, NonThrowing);
        m_triedToSerialize = true;
    }
    return m_serializedError;
}

bool ErrorEvent::isErrorEvent() const
{
    return true;
}

} // namespace WebCore
