/*
 * Copyright (C) 2016 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "Event.h"
#include "SerializedScriptValue.h"
#include <runtime/JSValue.h>
#include <runtime/WriteBarrier.h>
#include <wtf/text/WTFString.h>

namespace JSC
{
class SlotVisitor;
}

namespace WebCore {

class ErrorEvent FINAL : public Event {
public:
	struct Init : public EventInit {
	    String message;
	    String filename;
	    unsigned lineno { 0 };
	    unsigned colno { 0 };
	    JSC::JSValue error;
	};

    static PassRefPtr<ErrorEvent> create(const String& message, const String& fileName, unsigned lineNumber, unsigned columnNumber, JSC::JSValue error)
    {
        return adoptRef(new ErrorEvent(message, fileName, lineNumber, columnNumber, error));
    }
    static PassRefPtr<ErrorEvent> createForBindings(const AtomicString& type, const Init& initializer)
    {
        return adoptRef(new ErrorEvent(type, initializer));
    }
    virtual ~ErrorEvent();

    const String& message() const { return m_message; }
    const String& filename() const { return m_fileName; }
    unsigned lineno() const { return m_lineNumber; }
    unsigned colno() const { return m_columnNumber; }
    JSC::JSValue error() const;
    JSC::JSValue sanitizedErrorValue(JSC::ExecState&, JSC::JSGlobalObject&);

    SerializedScriptValue* serializedError() const { return m_serializedError.get(); }

    virtual EventInterface eventInterface() const FINAL;

    RefPtr<SerializedScriptValue> trySerializeError(JSC::ExecState&);

    void visitAdditionalChildren(JSC::SlotVisitor&);

private:
    ErrorEvent(const String& message, const String& fileName, unsigned lineNumber, unsigned columnNumber, JSC::JSValue error);
    ErrorEvent(const AtomicString&, const Init&);

    virtual bool isErrorEvent() const override;

    String m_message;
    String m_fileName;
    unsigned m_lineNumber;
    unsigned m_columnNumber;
    JSC::WriteBarrier<JSC::Unknown> m_error;
    RefPtr<SerializedScriptValue> m_serializedError;
    bool m_triedToSerialize;
};

using ErrorEventInit = ErrorEvent::Init;

} // namespace WebCore
