/*
 * Copyright (C) 2001 Peter Kelly (pmk@post.com)
 * Copyright (C) 2001 Tobias Anton (anton@stud.fbi.fh-darmstadt.de)
 * Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
 * Copyright (C) 2003, 2005, 2006, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "Event.h"

#include "DOMWindow.h"
#include "Document.h"
#include "EventNames.h"
#include "EventPath.h"
#include "EventTarget.h"
#include "Performance.h"
#include "UserGestureIndicator.h"
#include "WorkerGlobalScope.h"
#include <wtf/CurrentTime.h>
#include <wtf/HashSet.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/CString.h>

namespace WebCore {

std::atomic<int> gEventCount;

#if ENABLE(DUMP_EVENTS)
static Mutex& mutex()
{
	static Mutex mutex;
	return mutex;
}

static HashSet<Event*>& events()
{
	static HashSet<Event*> evts;
	return evts;
}

DLL_PUBLIC void dumpEvents(Vector<pair<String, int> >& dumpEvts)
{
	MutexLocker locker(mutex());

	HashSet<Event*>& evts = events();

	for(HashSet<Event*>::iterator it = evts.begin(); it != evts.end(); ++it) {
		String name = (*it)->type();
		bool found = false;

		for(int j = 0; j < dumpEvts.size(); ++j) {
			if(name == dumpEvts[j].first) {
				++dumpEvts[j].second;
				found = true;
				break;
			}
		}

		if(!found)
			dumpEvts.append(pair<String, int>(name, 1));
	}
}

#endif

Event::Event(IsTrusted isTrusted)
    : m_isInitialized(false)
	, m_canBubble(false)
    , m_cancelable(false)
	, m_composed(false)
    , m_propagationStopped(false)
    , m_immediatePropagationStopped(false)
    , m_wasCanceled(false)
    , m_defaultHandled(false)
	, m_isDefaultEventHandlerIgnored(false)
	, m_isTrusted(isTrusted == IsTrusted::Yes)
	, m_isExecutingPassiveEventListener(false)
	, m_currentTargetIsInShadowTree(false)
    , m_eventPhase(NONE)
	, m_eventPath(nullptr)
	, m_createTime(MonotonicTime::now())
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().add(this);
#endif

	gEventCount.fetch_add(1, std::memory_order_relaxed);
}

Event::Event(const AtomicString& eventType, bool canBubbleArg, bool cancelableArg, IsComposed composed)
    : m_type(eventType)
	, m_isInitialized(true)
    , m_canBubble(canBubbleArg)
    , m_cancelable(cancelableArg)
	, m_composed(composed == IsComposed::Yes)
    , m_propagationStopped(false)
    , m_immediatePropagationStopped(false)
    , m_wasCanceled(false)
    , m_defaultHandled(false)
	, m_isDefaultEventHandlerIgnored(false)
	, m_isTrusted(true)
	, m_isExecutingPassiveEventListener(false)
	, m_currentTargetIsInShadowTree(false)
    , m_eventPhase(NONE)
	, m_eventPath(nullptr)
	, m_createTime(MonotonicTime::now())
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().add(this);
#endif

	gEventCount.fetch_add(1, std::memory_order_relaxed);
}

Event::Event(const AtomicString& eventType, const EventInit& initializer, IsTrusted isTrusted)
    : m_type(eventType)
	, m_isInitialized(true)
    , m_canBubble(initializer.bubbles)
    , m_cancelable(initializer.cancelable)
	, m_composed(initializer.composed)
    , m_propagationStopped(false)
    , m_immediatePropagationStopped(false)
    , m_wasCanceled(false)
    , m_defaultHandled(false)
	, m_isDefaultEventHandlerIgnored(false)
	, m_isTrusted(isTrusted == IsTrusted::Yes)
	, m_isExecutingPassiveEventListener(false)
	, m_currentTargetIsInShadowTree(false)
    , m_eventPhase(NONE)
	, m_eventPath(nullptr)
	, m_createTime(MonotonicTime::now())
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().add(this);
#endif

	gEventCount.fetch_add(1, std::memory_order_relaxed);
}

Event::~Event()
{
#if ENABLE(DUMP_EVENTS)
	MutexLocker locker(mutex());
	events().remove(this);
#endif

	gEventCount.fetch_sub(1, std::memory_order_relaxed);
}

Ref<Event> Event::create(const AtomString& type, CanBubble canBubble, IsCancelable isCancelable, IsComposed isComposed)
{
    return adoptRef(*new Event(type, canBubble == CanBubble::Yes, isCancelable == IsCancelable::Yes, isComposed));
}

PassRefPtr<Event> Event::create(const AtomicString& type, bool canBubble, bool cancelable, IsComposed isComposed)
{
	return adoptRef(new Event(type, canBubble, cancelable, isComposed));
}

PassRefPtr<Event> Event::createForBindings()
{
    return adoptRef(new Event);
}

PassRefPtr<Event> Event::createForBindings(const AtomicString& type, const EventInit& initializer, IsTrusted isTrusted)
{
	return adoptRef(new Event(type, initializer, isTrusted));
}

void Event::initEvent(const AtomicString& eventTypeArg, bool canBubbleArg, bool cancelableArg)
{
    if (dispatched())
        return;

    m_isInitialized = true;
    m_propagationStopped = false;
    m_immediatePropagationStopped = false;
    m_wasCanceled = false;
    m_isTrusted = false;
    m_type = eventTypeArg;
    m_canBubble = canBubbleArg;
    m_cancelable = cancelableArg;
    m_underlyingEvent = nullptr;
}

void Event::setTarget(PassRefPtr<EventTarget> target)
{
    if (m_target == target)
        return;

    m_target = target;
    if (m_target)
        receivedTarget();
}

void Event::setCurrentTarget(EventTarget* currentTarget, Optional<bool> isInShadowTree)
{
    m_currentTarget = currentTarget;
    m_currentTargetIsInShadowTree = isInShadowTree ? *isInShadowTree : (currentTarget && currentTarget->isNode() && currentTarget->toNode()->isInShadowTree());
}

void Event::setEventPath(const EventPath& path)
{
    m_eventPath = makeWeakPtr(path);
}

Vector<Ref<EventTarget>> Event::composedPath() const
{
    if (!m_eventPath)
        return Vector<Ref<EventTarget>>();
    return m_eventPath->computePathUnclosedToTarget(*m_currentTarget);
}

void Event::setUnderlyingEvent(PassRefPtr<Event> ue)
{
    // Prohibit creation of a cycle by doing nothing if a cycle would be created.
    for (Event* e = ue.get(); e; e = e->underlyingEvent())
        if (e == this)
            return;
    m_underlyingEvent = ue;
}

DOMHighResTimeStamp Event::timeStampForBindings(ScriptExecutionContext* context) const
{
    Performance* performance = nullptr;
    if (context->isWorkerGlobalScope())
        performance = static_cast<WorkerGlobalScope&>(*context).performance();
    else if (auto* window = static_cast<Document&>(*context).domWindow())
        performance = window->performance();

    if (!performance)
        return 0;

    return performance->relativeTimeFromTimeOriginInReducedResolution(m_createTime);
}

void Event::resetBeforeDispatch()
{
    m_defaultHandled = false;
}

void Event::resetAfterDispatch()
{
    m_eventPath = nullptr;
    setCurrentTarget(nullptr);
    m_eventPhase = NONE;
    m_propagationStopped = false;
    m_immediatePropagationStopped = false;
}

} // namespace WebCore
