/*
 * Copyright (C) 2001 Peter Kelly (pmk@post.com)
 * Copyright (C) 2001 Tobias Anton (anton@stud.fbi.fh-darmstadt.de)
 * Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "DOMHighResTimeStamp.h"
#include "EventInit.h"
#include "EventInterfaces.h"
#include "ScriptWrappable.h"
#include <atomic>
#include <wtf/MonotonicTime.h>
#include <wtf/WeakPtr.h>
#include <wtf/text/AtomicString.h>

namespace WebCore {

class EventPath;
class EventTarget;

enum EventInterface {

#define DOM_EVENT_INTERFACE_DECLARE(name) name##InterfaceType,
DOM_EVENT_INTERFACES_FOR_EACH(DOM_EVENT_INTERFACE_DECLARE)
#undef DOM_EVENT_INTERFACE_DECLARE

};

class Event : public ScriptWrappable, public RefCounted<Event> {
public:
    enum class IsTrusted { No, Yes };
    enum class CanBubble : uint8_t { No, Yes };
    enum class IsCancelable : uint8_t { No, Yes };
    enum class IsComposed : uint8_t { No, Yes };

	enum PhaseType : uint8_t {
		NONE                = 0,
		CAPTURING_PHASE     = 1,
		AT_TARGET           = 2,
		BUBBLING_PHASE      = 3
	};

	enum EventType {
		MOUSEDOWN           = 1,
		MOUSEUP             = 2,
		MOUSEOVER           = 4,
		MOUSEOUT            = 8,
		MOUSEMOVE           = 16,
		MOUSEDRAG           = 32,
		CLICK               = 64,
		DBLCLICK            = 128,
		KEYDOWN             = 256,
		KEYUP               = 512,
		KEYPRESS            = 1024,
		DRAGDROP            = 2048,
		FOCUS               = 4096,
		BLUR                = 8192,
		SELECT              = 16384,
		CHANGE              = 32768
	};

    static Ref<Event> create(const AtomString& type, CanBubble, IsCancelable, IsComposed = IsComposed::No);
	static PassRefPtr<Event> create(const AtomicString& type, bool canBubble, bool cancelable, IsComposed = IsComposed::No);
    static PassRefPtr<Event> createForBindings();
	static PassRefPtr<Event> createForBindings(const AtomicString& type, const EventInit& initializer, IsTrusted = IsTrusted::No);

	virtual ~Event();

	void initEvent(const AtomicString& type, bool canBubble, bool cancelable);
    bool isInitialized() const { return m_isInitialized; }

	ALWAYS_INLINE const AtomicString& type() const { return m_type; }
	void setType(const AtomicString& type) { m_type = type; }

	ALWAYS_INLINE EventTarget* target() const { return m_target.get(); }
	void setTarget(PassRefPtr<EventTarget>);

    EventTarget* currentTarget() const { return m_currentTarget.get(); }
    void setCurrentTarget(EventTarget*, Optional<bool> isInShadowTree = WTF::Nullopt);
    bool currentTargetIsInShadowTree() const { return m_currentTargetIsInShadowTree; }

	ALWAYS_INLINE unsigned short eventPhase() const { return m_eventPhase; }
    void setEventPhase(PhaseType phase) { m_eventPhase = phase; }

	ALWAYS_INLINE bool bubbles() const { return m_canBubble; }
	ALWAYS_INLINE bool cancelable() const { return m_cancelable; }
	ALWAYS_INLINE bool composed() const { return m_composed; }

    DOMHighResTimeStamp timeStampForBindings(ScriptExecutionContext*) const;
    MonotonicTime timeStamp() const { return m_createTime; }

    void setEventPath(const EventPath&);
    Vector<Ref<EventTarget>> composedPath() const;

	void stopPropagation() { m_propagationStopped = true; }
	void stopImmediatePropagation() { m_immediatePropagationStopped = true; }

    bool isTrusted() const { return m_isTrusted; }
    void setUntrusted() { m_isTrusted = false; }

    bool legacyReturnValue() const { return !m_wasCanceled; }
    void setLegacyReturnValue(bool returnValue) { m_wasCanceled = !returnValue; }

    virtual EventInterface eventInterface() const { return EventInterfaceType; }

    virtual bool isBeforeTextInsertedEvent() const { return false; }
    virtual bool isBeforeUnloadEvent() const { return false; }
    virtual bool isClipboardEvent() const { return false; }
    virtual bool isCompositionEvent() const { return false; }
	virtual bool isDragEvent() const { return false; }
    virtual bool isErrorEvent() const { return false; }
    virtual bool isFocusEvent() const { return false; }
    virtual bool isInputEvent() const { return false; }
    virtual bool isKeyboardEvent() const { return false; }
    virtual bool isMouseEvent() const { return false; }
    virtual bool isTextEvent() const { return false; }
    virtual bool isTouchEvent() const { return false; }
    virtual bool isUIEvent() const { return false; }
    virtual bool isVersionChangeEvent() const { return false; }
    virtual bool isWheelEvent() const { return false; }

#if ENABLE(CEHTML)
	virtual bool isKeyEvent() const { return false; }
#endif

#if ENABLE(RMAI)
	virtual bool isMOMEvent() const { return false; }
	virtual bool isGalioConfigurationEvent() const { return false; }
	void clearState() {
		m_propagationStopped = false;
		m_immediatePropagationStopped = false;
		m_wasCanceled = false;
		m_defaultHandled = false;
	}
#endif

	ALWAYS_INLINE bool propagationStopped() const { return m_propagationStopped || m_immediatePropagationStopped; }
	ALWAYS_INLINE bool immediatePropagationStopped() const { return m_immediatePropagationStopped; }

    void resetBeforeDispatch();
    void resetAfterDispatch();

	ALWAYS_INLINE bool defaultPrevented() const { return m_wasCanceled; }
	void preventDefault();

	ALWAYS_INLINE bool defaultHandled() const { return m_defaultHandled; }
	void setDefaultHandled() { m_defaultHandled = true; }

    bool isDefaultEventHandlerIgnored() const { return m_isDefaultEventHandlerIgnored; }
    void setIsDefaultEventHandlerIgnored() { m_isDefaultEventHandlerIgnored = true; }

    void setInPassiveListener(bool value) { m_isExecutingPassiveEventListener = value; }

    bool cancelBubble() const { return propagationStopped(); }
    void setCancelBubble(bool);

	ALWAYS_INLINE Event* underlyingEvent() const { return m_underlyingEvent.get(); }
	void setUnderlyingEvent(PassRefPtr<Event>);

	ALWAYS_INLINE bool isBeingDispatched() const { return eventPhase(); }

    virtual EventTarget* relatedTarget() const { return nullptr; }
    virtual void setRelatedTarget(EventTarget*) { }

protected:
    explicit Event(IsTrusted = IsTrusted::No);
	Event(const AtomicString& type, bool canBubble, bool cancelable, IsComposed = IsComposed::No);
	Event(const AtomicString& type, const EventInit&, IsTrusted);

	virtual void receivedTarget() { }
	ALWAYS_INLINE bool dispatched() const { return m_target.get(); }

private:
	AtomicString m_type;

	unsigned m_isInitialized : 1;
	unsigned m_canBubble : 1;
	unsigned m_cancelable : 1;
	unsigned m_composed : 1;

	unsigned m_propagationStopped : 1;
	unsigned m_immediatePropagationStopped : 1;
	unsigned m_wasCanceled : 1;
	unsigned m_defaultHandled : 1;
	unsigned m_isDefaultEventHandlerIgnored : 1;
	unsigned m_isTrusted : 1;
	unsigned m_isExecutingPassiveEventListener : 1;
    unsigned m_currentTargetIsInShadowTree : 1;

    PhaseType m_eventPhase;
    RefPtr<EventTarget> m_currentTarget;
    WeakPtr<const EventPath> m_eventPath;
	RefPtr<EventTarget> m_target;
	MonotonicTime m_createTime;

	RefPtr<Event> m_underlyingEvent;
};

inline void Event::preventDefault()
{
    if (m_cancelable && !m_isExecutingPassiveEventListener)
    	m_wasCanceled = true;
    // FIXME: Specification suggests we log something to the console when preventDefault is called but
    // doesn't do anything because the event is not cancelable or is executing passive event listeners.
}

inline void Event::setCancelBubble(bool cancel)
{
    if (cancel)
        m_propagationStopped = true;
}

} // namespace WebCore
