/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007 Apple Inc. All rights reserved.
 * Copyright (C) 2006 Alexey Proskuryakov (ap@webkit.org)
 *           (C) 2007, 2008 Nikolas Zimmermann <zimmermann@kde.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *
 */

#include "config.h"
#include "EventTarget.h"

#include "AddEventListenerOptions.h"
#include "Event.h"
#include "EventException.h"
#include "InspectorInstrumentation.h"
#include "JSEventListener.h"
#include "ScriptDisallowedScope.h"
#include "WebKitAnimationEvent.h"
#include "WebKitTransitionEvent.h"
#include <wtf/MainThread.h>
#include <wtf/StdLibExtras.h>
#include <wtf/TemporaryChange.h>
#include <wtf/Vector.h>

using namespace WTF;

namespace WebCore {

#ifndef NDEBUG
static int gEventDispatchForbidden = 0;

void forbidEventDispatch()
{
    if (!isMainThread())
        return;
    ++gEventDispatchForbidden;
}

void allowEventDispatch()
{
    if (!isMainThread())
        return;
    if (gEventDispatchForbidden > 0)
        --gEventDispatchForbidden;
}

bool eventDispatchForbidden()
{
    if (!isMainThread())
        return false;
    return gEventDispatchForbidden > 0;
}
#endif // NDEBUG

EventTargetData::EventTargetData() = default;
EventTargetData::~EventTargetData() = default;

DLL_PUBLIC RefPtr<Node> EventTarget::toNode()
{
    return nullptr;
}

DLL_PUBLIC bool EventTarget::addEventListener(const AtomicString& eventType, PassRefPtr<EventListener> listener, const AddEventListenerOptions& options)
{
    if (options.signal && options.signal->aborted())
        return false;

    if (!ensureEventTargetData()->eventListenerMap.add(eventType, listener, options.capture))
    	return false;

    if (options.signal) {
        class AlgorithmImpl final : public AbortSignal::Algorithm
    	{
    	public:
        	AlgorithmImpl(EventTarget& thisRef, const AtomicString& eventType, EventListener* listener, bool capture)
    			: weakThis(makeWeakPtr(thisRef))
    			, listener(makeWeakPtr(listener))
    			, eventType(eventType)
    			, capture(capture)
    		{ }
        	virtual void operator()() final
        	{
        		if (weakThis && listener)
        			weakThis->removeEventListener(eventType, listener.get(), capture);
        	}
    	private:
        	WeakPtr<EventTarget> weakThis;
        	WeakPtr<EventListener> listener;
        	const AtomicString& eventType;
        	bool capture;
    	};
        options.signal->addAlgorithm(adoptPtr(new AlgorithmImpl(*this, eventType, listener.get(), options.capture)));
    }

    eventListenersDidChange();
    return true;
}

void EventTarget::addEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener> listener, bool useCapture)
{
    if (!listener)
        return;
	addEventListener(eventType, listener, useCapture);
}

void EventTarget::removeEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener> listener, bool useCapture)
{
    if (!listener)
        return;
	removeEventListener(eventType, listener.get(), useCapture);
}

void EventTarget::addEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener> listener, const AddEventListenerOptions& options)
{
    if (!listener)
        return;
	addEventListener(eventType, listener, options);
}

void EventTarget::removeEventListenerForBindings(const AtomicString& eventType, PassRefPtr<EventListener> listener, const EventListenerOptions& options)
{
    if (!listener)
        return;
	removeEventListener(eventType, listener.get(), options.capture);
}

DLL_PUBLIC bool EventTarget::removeEventListener(const AtomicString& eventType, EventListener* listener, const EventListenerOptions& options)
{
    EventTargetData* d = eventTargetData();
    if (!d)
        return false;

    if (d->eventListenerMap.remove(eventType, listener, options.capture)) {
        eventListenersDidChange();
        return true;
    }
    return false;
}

bool EventTarget::setAttributeEventListener(const AtomicString& eventType, PassRefPtr<EventListener> listener, DOMWrapperWorld* isolatedWorld)
{
    EventListener* existingListener = attributeEventListener(eventType, isolatedWorld);
    if (!listener) {
        if (existingListener)
            removeEventListener(eventType, existingListener, false);
        return false;
    }
    if (existingListener) {
        eventTargetData()->eventListenerMap.replace(eventType, *existingListener, listener, RegisteredEventListener::Options());
        return true;
    }
    return addEventListener(eventType, listener, false);
}

EventListener* EventTarget::attributeEventListener(const AtomicString& eventType, DOMWrapperWorld* isolatedWorld)
{
    const EventListenerVector& entry = getEventListeners(eventType);
    for (size_t i = 0; i < entry.size(); ++i) {
    	EventListener& listener = entry[i]->listener();
        if (!listener.isAttribute())
            continue;

        auto* listenerWorld = static_cast<JSEventListener&>(listener).isolatedWorld();
        if (listenerWorld == isolatedWorld)
            return &listener;
    }

    return 0;
}

bool EventTarget::dispatchEventForBindings(PassRefPtr<Event> prpEvent, ExceptionCode& ec)
{
	RefPtr<Event> event(prpEvent);

    if (!event) {
        ec = TypeError;
        return false;
    }

    if (!event->isInitialized() || event->isBeingDispatched()) {
        ec = INVALID_STATE_ERR;
        return false;
    }

    event->setUntrusted();

    if (event->isBeingDispatched()) {
        ec = EventException::DISPATCH_REQUEST_ERR;
        return false;
    }

    if (!scriptExecutionContext())
        return false;

    dispatchEvent(event);
    return event->legacyReturnValue();
}

DLL_PUBLIC void EventTarget::dispatchEvent(PassRefPtr<Event> event)
{
#if ENABLE(JS_EXECUTION_TIMING)
    double fireTime = WTF::monotonicallyIncreasingTime();
#endif
    // FIXME: We should always use EventDispatcher.

    EventPath eventPath(*this);
    event->setTarget(this);
    event->setCurrentTarget(this);
    event->setEventPhase(Event::AT_TARGET);
    event->resetBeforeDispatch();
    event->setEventPath(eventPath);
    fireEventListeners(*event, EventInvokePhase::Capturing);
    fireEventListeners(*event, EventInvokePhase::Bubbling);
    event->resetAfterDispatch();
#if ENABLE(JS_EXECUTION_TIMING)
    double now = WTF::monotonicallyIncreasingTime();
    WTF::logMessage(tryMakeString("EventTarget::dispatchEvent(", event->type(), ") takes ", String::number((int)((now - fireTime) * 1000)), "ms"));
#endif
}

static const AtomicString& legacyType(const Event* event)
{
	EventNames& eventNames = WebCore::eventNames();
    if (event->type() == eventNames.animationendEvent)
        return eventNames.webkitAnimationEndEvent;

    if (event->type() == eventNames.animationstartEvent)
        return eventNames.webkitAnimationStartEvent;

    if (event->type() == eventNames.animationiterationEvent)
        return eventNames.webkitAnimationIterationEvent;

    if (event->type() == eventNames.transitionendEvent)
        return eventNames.webkitTransitionEndEvent;

    if (event->type() == eventNames.wheelEvent)
        return eventNames.mousewheelEvent;

    return emptyAtom;
}

// https://dom.spec.whatwg.org/#concept-event-listener-invoke
void EventTarget::fireEventListeners(Event& event, EventInvokePhase phase)
{
    ASSERT_WITH_SECURITY_IMPLICATION(ScriptDisallowedScope::isScriptAllowedInMainThread());
	ASSERT(event && event->isInitialized());

    EventTargetData* data = eventTargetData();
    if (!data)
        return;

    TemporaryChange<bool> firingEventListenersScope(data->isFiringEventListeners, true);

    if (auto* listenersVector = data->eventListenerMap.find(event.type())) {
        innerInvokeEventListeners(event, *listenersVector, phase);
        return;
    }

    const AtomicString& legacyTypeName = legacyType(&event);
    if (!legacyTypeName.isEmpty()) {
        if (auto* legacyListenersVector = data->eventListenerMap.find(legacyTypeName)) {
			AtomicString typeName = event.type();
			event.setType(legacyTypeName);
            innerInvokeEventListeners(event, *legacyListenersVector, phase);
			event.setType(typeName);
        }
    }
}

// Intentionally creates a copy of the listeners vector to avoid event listeners added after this point from being run.
// Note that removal still has an effect due to the removed field in RegisteredEventListener.
// https://dom.spec.whatwg.org/#concept-event-listener-inner-invoke
void EventTarget::innerInvokeEventListeners(Event& event, EventListenerVector listeners, EventInvokePhase phase)
{
    RefPtr<EventTarget> protect = this;
    ASSERT(!listeners.isEmpty());

    ScriptExecutionContext* context = scriptExecutionContext();

    for (auto& registeredListener : listeners) {
        if (UNLIKELY(registeredListener->wasRemoved()))
            continue;

        if (phase == EventInvokePhase::Capturing && !registeredListener->useCapture())
            continue;
        if (phase == EventInvokePhase::Bubbling && registeredListener->useCapture())
            continue;

        // If stopImmediatePropagation has been called, we just break out immediately, without
        // handling any more events on this target.
        if (event.immediatePropagationStopped())
            break;

        // Make sure the JS wrapper and function stay alive until the end of this scope. Otherwise,
        // event listeners with 'once' flag may get collected as soon as they get unregistered below,
        // before we call the js function.
        auto* wrapper = registeredListener->callback().wrapper();
        auto* jsFunction = registeredListener->callback().jsFunction();
        JSC::EnsureStillAliveScope wrapperProtector(wrapper ? JSValue(wrapper) : JSValue());
        JSC::EnsureStillAliveScope jsFunctionProtector(jsFunction ? JSValue(jsFunction) : JSValue());

        // Do this before invocation to avoid reentrancy issues.
        if (registeredListener->isOnce())
            removeEventListener(event.type(), &registeredListener->callback(), registeredListener->useCapture());

        if (registeredListener->isPassive())
            event.setInPassiveListener(true);

        InspectorInstrumentationCookie cookie = InspectorInstrumentation::willHandleEvent(context, &event);
        // To match Mozilla, the AT_TARGET phase fires both capturing and bubbling
        // event listeners, even though that violates some versions of the DOM spec.
        registeredListener->listener().handleEvent(context, &event);
        InspectorInstrumentation::didHandleEvent(cookie);

        if (registeredListener->isPassive())
            event.setInPassiveListener(false);
    }
}

const EventListenerVector& EventTarget::getEventListeners(const AtomicString& eventType)
{
	static NeverDestroyed<EventListenerVector> emptyVector;

    EventTargetData* d = eventTargetData();
    if (!d)
        return emptyVector;

    EventListenerVector* listenerVector = d->eventListenerMap.find(eventType);
    if (!listenerVector)
        return emptyVector;

    return *listenerVector;
}

DLL_PUBLIC void EventTarget::removeAllEventListeners()
{
    auto* data = eventTargetData();
    if (data && !data->eventListenerMap.isEmpty()) {
        data->eventListenerMap.clear();
        eventListenersDidChange();
    }
}

} // namespace WebCore
