/*
 * Copyright (C) 2001 Peter Kelly (pmk@post.com)
 * Copyright (C) 2001 Tobias Anton (anton@stud.fbi.fh-darmstadt.de)
 * Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "EventModifierInit.h"
#include "UIEventWithKeyState.h"
#include <wtf/Optional.h>
#include <wtf/Vector.h>

namespace WebCore {

class Node;
class PlatformKeyboardEvent;

#if PLATFORM(MAC)
struct KeypressCommand {
	KeypressCommand() { }
	explicit KeypressCommand(const String& commandName) : commandName(commandName) { ASSERT(isASCIILower(commandName[0U])); }
	KeypressCommand(const String& commandName, const String& text) : commandName(commandName), text(text) { ASSERT(commandName == "insertText:"); }

	String commandName; // Actually, a selector name - it may have a trailing colon, and a name that can be different from an editor command name.
	String text;
};
#endif

// Introduced in DOM Level 3
class KeyboardEvent FINAL : public UIEventWithKeyState {
public:
	struct Init : public EventModifierInit {
		Init()
			: location(0)
			, repeat(false)
			, isComposing(false)
			, charCode(0)
	    	, keyCode(0)
	    	, which(0)
		{
		}
		Init(bool bubbles, bool cancelable, bool composed, PassRefPtr<DOMWindow> view, int detail,
	        bool ctrlKey, bool shiftKey, bool altKey, bool metaKey, bool modifierAltGraph, bool modifierCapsLock,
	        String key, String code, unsigned location, bool repeat, bool isComposing, String keyIdentifier,
	        Optional<unsigned> keyLocation, unsigned charCode, unsigned keyCode, unsigned which);

	    String key;
	    String code;
	    unsigned location;
	    bool repeat;
	    bool isComposing;

	    // Legacy.
	    String keyIdentifier;
	    Optional<unsigned> keyLocation;
	    unsigned charCode;
	    unsigned keyCode;
	    unsigned which;
	};

	enum KeyLocationCode {
        DOM_KEY_LOCATION_STANDARD   = 0x00,
        DOM_KEY_LOCATION_LEFT       = 0x01,
        DOM_KEY_LOCATION_RIGHT      = 0x02,
        DOM_KEY_LOCATION_NUMPAD     = 0x03
	};

	static PassRefPtr<KeyboardEvent> create(const PlatformKeyboardEvent& platformEvent, AbstractView* view)
	{
		return adoptRef(new KeyboardEvent(platformEvent, view));
	}

	static PassRefPtr<KeyboardEvent> createForBindings()
	{
		return adoptRef(new KeyboardEvent);
	}

	static PassRefPtr<KeyboardEvent> create(const AtomicString& type, const Init& initializer)
	{
		return adoptRef(new KeyboardEvent(type, initializer));
	}

	virtual ~KeyboardEvent();

	void initKeyboardEvent(const AtomicString& type, bool canBubble, bool cancelable, AbstractView*,
						   const String& keyIdentifier, unsigned location,
						   bool ctrlKey, bool altKey, bool shiftKey, bool metaKey, bool altGraphKey = false);

#if ENABLE(KEYBOARD_KEY_ATTRIBUTE)
    const String& key() const { return m_key; }
#endif
#if ENABLE(KEYBOARD_CODE_ATTRIBUTE)
    const String& code() const { return m_code; }
#endif

	const String& keyIdentifier() const { return m_keyIdentifier; }
	unsigned location() const { return m_location; }
    bool repeat() const { return m_repeat; }

	bool getModifierState(const String& keyIdentifier) const;

    const PlatformKeyboardEvent* underlyingPlatformEvent() const { return m_underlyingPlatformEvent.get(); }
    PlatformKeyboardEvent* underlyingPlatformEvent() { return m_underlyingPlatformEvent.get(); }

	int keyCode() const; // key code for keydown and keyup, character for keypress
	int charCode() const; // character code for keypress, 0 for keydown and keyup

	virtual EventInterface eventInterface() const FINAL;
	virtual bool isKeyboardEvent() const FINAL;
	virtual int which() const FINAL;

    bool isComposing() const { return m_isComposing; }

#if PLATFORM(MAC)
	// We only have this need to store keypress command info on the Mac.
	Vector<KeypressCommand>& keypressCommands() { return m_keypressCommands; }
#endif

private:
	KeyboardEvent();
	KeyboardEvent(const PlatformKeyboardEvent&, AbstractView*);
	KeyboardEvent(const AtomicString&, const Init&);
	KeyboardEvent(const AtomicString& type, bool canBubble, bool cancelable, AbstractView*,
				  const String& keyIdentifier, unsigned keyLocation,
				  bool ctrlKey, bool altKey, bool shiftKey, bool metaKey, bool altGraphKey);

	OwnPtr<PlatformKeyboardEvent> m_underlyingPlatformEvent;
#if ENABLE(KEYBOARD_KEY_ATTRIBUTE)
    String m_key;
#endif
#if ENABLE(KEYBOARD_CODE_ATTRIBUTE)
    String m_code;
#endif
	String m_keyIdentifier;
	unsigned m_location;
	bool m_repeat;
	bool m_isComposing;
    Optional<unsigned> m_charCode;
    Optional<unsigned> m_keyCode;
    Optional<unsigned> m_which;

#if PLATFORM(MAC)
	// Commands that were sent by AppKit when interpreting the event. Doesn't include input method commands.
	Vector<KeypressCommand> m_keypressCommands;
#endif
};

inline KeyboardEvent::Init::Init(bool bubbles, bool cancelable, bool composed, PassRefPtr<DOMWindow> view, int detail,
	bool ctrlKey, bool shiftKey, bool altKey, bool metaKey, bool modifierAltGraph, bool modifierCapsLock,
	String key, String code, unsigned location, bool repeat, bool isComposing, String keyIdentifier,
	Optional<unsigned> keyLocation, unsigned charCode, unsigned keyCode, unsigned which)
	: EventModifierInit(bubbles, cancelable, composed, view, detail, ctrlKey, shiftKey, altKey, metaKey, modifierAltGraph, modifierCapsLock)
	, key(key)
	, code(code)
	, location(location)
	, repeat(repeat)
	, isComposing(isComposing)
	, keyIdentifier(keyIdentifier)
	, keyLocation(keyLocation)
	, charCode(charCode)
	, keyCode(keyCode)
	, which(which)
{ }

} // namespace WebCore
