/*
 * Copyright (C) 2007 Henry Mason (hmason@mac.com)
 * Copyright (C) 2003, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *
 */

#include "config.h"
#include "MessageEvent.h"

#include "DOMWindow.h"
#include "EventNames.h"

namespace WebCore {

using namespace JSC;

static inline bool isValidSource(EventTarget* source)
{
    return !source || source->toDOMWindow() || source->isMessagePort();
}

MessageEvent::MessageEvent()
    : m_dataType(DataTypeScriptValue)
	, m_triedToSerialize(false)
{
}

MessageEvent::MessageEvent(const AtomicString& type, const Init& initializer, IsTrusted isTrusted)
    : Event(type, initializer, isTrusted)
    , m_dataType(DataTypeScriptValue)
	, m_triedToSerialize(false)
    , m_origin(initializer.origin)
    , m_lastEventId(initializer.lastEventId)
	, m_source(isValidSource(initializer.source.get()) ? initializer.source : nullptr)
    , m_ports(adoptPtr(new Vector<RefPtr<MessagePort>>(initializer.ports)))
{
	m_dataAsScriptValue.setWithoutWriteBarrier(initializer.data);
}

MessageEvent::MessageEvent(JSC::JSValue data, const String& origin, const String& lastEventId, PassRefPtr<EventTarget> source, PassOwnPtr<Vector<RefPtr<MessagePort>>> ports)
    : Event(eventNames().messageEvent, false, false)
    , m_dataType(DataTypeScriptValue)
	, m_triedToSerialize(false)
    , m_origin(origin)
    , m_lastEventId(lastEventId)
    , m_source(source)
    , m_ports(ports)
{
	m_dataAsScriptValue.setWithoutWriteBarrier(data);
    ASSERT(isValidSource(m_source.get()));
}

MessageEvent::MessageEvent(PassRefPtr<SerializedScriptValue> data, const String& origin, const String& lastEventId, PassRefPtr<EventTarget> source, PassOwnPtr<Vector<RefPtr<MessagePort>>> ports)
    : Event(eventNames().messageEvent, false, false)
    , m_dataType(DataTypeSerializedScriptValue)
    , m_dataAsSerializedScriptValue(data)
	, m_triedToSerialize(false)
    , m_origin(origin)
    , m_lastEventId(lastEventId)
    , m_source(source)
    , m_ports(ports)
{
#if USE(V8)
    if (m_dataAsSerializedScriptValue)
        m_dataAsSerializedScriptValue->registerMemoryAllocatedWithCurrentScriptContext();
#endif
    ASSERT(isValidSource(m_source.get()));
}

MessageEvent::MessageEvent(const AtomicString& type, bool canBubble, bool cancelable, PassRefPtr<SerializedScriptValue> data, const String& origin, const String& lastEventId)
    : Event(type, canBubble, cancelable)
    , m_dataType(DataTypeSerializedScriptValue)
    , m_dataAsSerializedScriptValue(data)
	, m_triedToSerialize(false)
    , m_origin(origin)
    , m_lastEventId(lastEventId)
{
}

MessageEvent::MessageEvent(const String& data, const String& origin)
    : Event(eventNames().messageEvent, false, false)
    , m_dataType(DataTypeString)
	, m_triedToSerialize(false)
    , m_dataAsString(data)
    , m_origin(origin)
{
}

MessageEvent::MessageEvent(PassRefPtr<Blob> data, const String& origin)
    : Event(eventNames().messageEvent, false, false)
    , m_dataType(DataTypeBlob)
	, m_triedToSerialize(false)
    , m_dataAsBlob(data)
    , m_origin(origin)
{
}

MessageEvent::MessageEvent(PassRefPtr<ArrayBuffer> data, const String& origin)
    : Event(eventNames().messageEvent, false, false)
    , m_dataType(DataTypeArrayBuffer)
	, m_triedToSerialize(false)
    , m_dataAsArrayBuffer(data)
    , m_origin(origin)
{
}

MessageEvent::~MessageEvent()
{
}

void MessageEvent::initMessageEvent(const AtomicString& type, bool canBubble, bool cancelable, JSC::JSValue data, const String& origin, const String& lastEventId, EventTarget* source, PassOwnPtr<Vector<RefPtr<MessagePort>>> ports)
{
    if (dispatched())
        return;

    initEvent(type, canBubble, cancelable);

    m_dataType = DataTypeScriptValue;
    m_dataAsScriptValue.setWithoutWriteBarrier(data);
    m_dataAsSerializedScriptValue = 0;
    m_triedToSerialize = false;
    m_origin = origin;
    m_lastEventId = lastEventId;
    m_source = isValidSource(source) ? source : nullptr;
    m_ports = ports;
}

void MessageEvent::initMessageEvent(const AtomicString& type, bool canBubble, bool cancelable, PassRefPtr<SerializedScriptValue> data, const String& origin, const String& lastEventId, EventTarget* source, PassOwnPtr<Vector<RefPtr<MessagePort>>> ports)
{
    if (dispatched())
        return;

    initEvent(type, canBubble, cancelable);

    m_dataType = DataTypeSerializedScriptValue;
    m_dataAsSerializedScriptValue = data;
    m_origin = origin;
    m_lastEventId = lastEventId;
    m_source = isValidSource(source) ? source : nullptr;
    m_ports = ports;

#if USE(V8)
    if (m_dataAsSerializedScriptValue)
        m_dataAsSerializedScriptValue->registerMemoryAllocatedWithCurrentScriptContext();
#endif
}

JSC::JSValue MessageEvent::dataAsScriptValue() const
{
	ASSERT(m_dataType == DataTypeScriptValue);
	return m_dataAsScriptValue.get();
}

RefPtr<SerializedScriptValue> MessageEvent::trySerializeData(JSC::ExecState* exec)
{
    ASSERT(!m_dataAsScriptValue.hasNoValue());

    if (!m_dataAsSerializedScriptValue && !m_triedToSerialize) {
        Vector<RefPtr<MessagePort>> dummyMessagePorts;
        Vector<RefPtr<JSC::ArrayBuffer>> dummyArrayBuffers;
        m_dataAsSerializedScriptValue = SerializedScriptValue::create(exec, m_dataAsScriptValue.get(), dummyMessagePorts, dummyArrayBuffers, NonThrowing);
        m_triedToSerialize = true;
    }

    return m_dataAsSerializedScriptValue;
}

EventInterface MessageEvent::eventInterface() const
{
    return MessageEventInterfaceType;
}

} // namespace WebCore
