/*
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *
 */

#pragma once

#include "ActiveDOMObject.h"
#include "EventListener.h"
#include "EventTarget.h"
#include "MessagePortChannel.h"
#include <wtf/Forward.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>
#include <wtf/text/AtomicStringHash.h>

namespace WebCore {

class Event;
class Frame;
class MessagePort;
class ScriptExecutionContext;

class MessagePort FINAL : public RefCounted<MessagePort>, public ActiveDOMObject, public EventTargetWithInlineData {
public:
	static PassRefPtr<MessagePort> create(ScriptExecutionContext& scriptExecutionContext) { return adoptRef(new MessagePort(scriptExecutionContext)); }
	virtual ~MessagePort();

	void postMessage(PassRefPtr<SerializedScriptValue> message, ExceptionCode&);
	void postMessage(PassRefPtr<SerializedScriptValue> message, const Vector<RefPtr<MessagePort>>*, ExceptionCode&);

	void start();
	void close();

	void entangle(PassOwnPtr<MessagePortChannel>);
	PassOwnPtr<MessagePortChannel> disentangle();

	// Returns 0 if there is an exception, or if the passed-in array is 0/empty.
	static PassOwnPtr<MessagePortChannelArray> disentanglePorts(const Vector<RefPtr<MessagePort>>*, ExceptionCode&);

	// Returns 0 if the passed array is 0/empty.
	static PassOwnPtr<Vector<RefPtr<MessagePort>>> entanglePorts(ScriptExecutionContext&, PassOwnPtr<MessagePortChannelArray>);

	void messageAvailable();

	bool started() const { return m_started; }
	bool isDetached() const { return m_isDetached; }

	void dispatchMessages();

	using RefCounted<MessagePort>::ref;
	using RefCounted<MessagePort>::deref;

    // Returns null if there is no entangled port, or if the entangled port is run by a different thread.
    // This is used solely to enable a GC optimization. Some platforms may not be able to determine ownership
    // of the remote port (since it may live cross-process) - those platforms may always return null.
    MessagePort* locallyEntangledPort() const;

    // A port starts out its life entangled, and remains entangled until it is closed or is cloned.
    bool isEntangled() const { return !m_isDetached && !isNeutered(); }

    // A port gets neutered when it is transferred to a new owner via postMessage().
    bool isNeutered() const { return !m_entangledChannel; }

	// EventTargetWithInlineData.
	virtual bool isMessagePort() const OVERRIDE FINAL { return true; }
	virtual EventTargetInterface eventTargetInterface() const override;
	virtual ScriptExecutionContext* scriptExecutionContext() const FINAL { return ActiveDOMObject::scriptExecutionContext(); }

	virtual void refEventTarget() override { ref(); }
	virtual void derefEventTarget() override { deref(); }

	virtual bool addEventListener(const AtomicString& eventType, PassRefPtr<EventListener>, const AddEventListenerOptions&) override;

private:
	explicit MessagePort(ScriptExecutionContext&);

	// ActiveDOMObject
	virtual const char* activeDOMObjectName() const FINAL;
	virtual bool canSuspendForDocumentSuspension() const FINAL;
	virtual void contextDestroyed() FINAL;
	virtual void stop() FINAL { close(); }
	virtual bool virtualHasPendingActivity() const FINAL;

	OwnPtr<MessagePortChannel> m_entangledChannel;

	bool m_started { false };
	bool m_isDetached { false };
};

} // namespace WebCore
