/*
 * Copyright (C) 2001 Peter Kelly (pmk@post.com)
 * Copyright (C) 2001 Tobias Anton (anton@stud.fbi.fh-darmstadt.de)
 * Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
 * Copyright (C) 2003, 2004, 2005, 2006, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "DataTransfer.h"
#include "MouseEventInit.h"
#include "MouseRelatedEvent.h"

namespace WebCore {

class PlatformMouseEvent;

// Introduced in DOM Level 2
class MouseEvent : public MouseRelatedEvent {
public:
	static Ref<MouseEvent> create(const AtomicString& type, bool canBubble, bool cancelable, IsComposed isComposed, PassRefPtr<AbstractView> view,
		int detail, int screenX, int screenY, int pageX, int pageY,
#if ENABLE(POINTER_LOCK)
		int movementX, int movementY,
#endif
		bool ctrlKey, bool altKey, bool shiftKey, bool metaKey, unsigned short button, unsigned short buttons,
		PassRefPtr<EventTarget> relatedTarget, PassRefPtr<DataTransfer> clipboard = nullptr, bool isSimulated = false)
	{
		return adoptRef(*new MouseEvent(type, canBubble, cancelable, isComposed, view, detail, screenX, screenY, pageX, pageY,
#if ENABLE(POINTER_LOCK)
			movementX, movementY,
#endif
			ctrlKey, altKey, shiftKey, metaKey, button, buttons, relatedTarget, clipboard, isSimulated));
	}
	static Ref<MouseEvent> create(const AtomicString& eventType, PassRefPtr<AbstractView>, const PlatformMouseEvent&, int detail, PassRefPtr<Node> relatedTarget);

    static Ref<MouseEvent> create(const AtomicString& eventType, bool canBubble, bool cancelable, IsComposed, AbstractView*,
        int detail, int screenX, int screenY, int clientX, int clientY,
        bool ctrlKey, bool altKey, bool shiftKey, bool metaKey,
        unsigned short button, unsigned short buttons, PassRefPtr<EventTarget> relatedTarget);

	static Ref<MouseEvent> createForBindings()
	{
		return adoptRef(*new MouseEvent);
	}

    static Ref<MouseEvent> createForBindings(const AtomicString& eventType, const MouseEventInit&);

	virtual ~MouseEvent();

	void initMouseEvent(const AtomicString& type, bool canBubble, bool cancelable, PassRefPtr<AbstractView>,
						int detail, int screenX, int screenY, int clientX, int clientY,
						bool ctrlKey, bool altKey, bool shiftKey, bool metaKey,
						unsigned short button, PassRefPtr<EventTarget> relatedTarget);

	unsigned short button() const { return m_button; }
    unsigned short buttons() const { return m_buttons; }
    unsigned short syntheticClickType() const { return m_syntheticClickType; }
	bool buttonDown() const { return m_buttonDown; }
	virtual EventTarget* relatedTarget() const FINAL { return m_relatedTarget.get(); }

	RefPtr<Node> toElement() const;
	RefPtr<Node> fromElement() const;

    static bool canTriggerActivationBehavior(const Event&);

	DataTransfer* dataTransfer() const { return isDragEvent() ? m_dataTransfer.get() : nullptr; }

	virtual EventInterface eventInterface() const override;

	virtual bool isMouseEvent() const final;
	virtual bool isDragEvent() const final;
	virtual int which() const final;

protected:
	MouseEvent(const AtomicString& type, bool canBubble, bool cancelable, IsComposed, PassRefPtr<AbstractView>,
	   int detail, int screenX, int screenY, int pageX, int pageY,
#if ENABLE(POINTER_LOCK)
	   int movementX, int movementY,
#endif
	   bool ctrlKey, bool altKey, bool shiftKey, bool metaKey, unsigned short button, unsigned short buttons,
	   PassRefPtr<EventTarget> relatedTarget, PassRefPtr<DataTransfer>, bool isSimulated);

    MouseEvent(const AtomicString& type, bool canBubble, bool cancelable, IsComposed, AbstractView*,
        int detail, int screenX, int screenY, int clientX, int clientY,
        bool ctrlKey, bool altKey, bool shiftKey, bool metaKey,
        unsigned short button, unsigned short buttons, PassRefPtr<EventTarget> relatedTarget);

    MouseEvent(const AtomicString& type, const MouseEventInit&);

	MouseEvent();

private:

	virtual void setRelatedTarget(EventTarget* relatedTarget) final { m_relatedTarget = relatedTarget; }

	unsigned short m_button { 0 };
    unsigned short m_buttons { 0 };
    unsigned short m_syntheticClickType { 0 };
	bool m_buttonDown { false };
	RefPtr<EventTarget> m_relatedTarget;
	RefPtr<DataTransfer> m_dataTransfer;
};

inline MouseEvent* toMouseEvent(Event* event)
{
    ASSERT(event && event->isMouseEvent());
    return static_cast<MouseEvent*>(event);
}

inline MouseEvent& toMouseEvent(Event& event)
{
    ASSERT(event.isMouseEvent());
    return static_cast<MouseEvent&>(event);
}

} // namespace WebCore
