/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.
 * Copyright (C) 2008, 2009 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "Event.h"
#include "EventDispatcher.h"
#include "EventTarget.h"
#include "GCReachableRef.h"
#include "KURLHash.h"
#include "LayoutRect.h"
#include "MutationObserver.h"
#include "RenderStyleConstants.h"
#include "SimulatedClickOptions.h"
#include "StyleValidity.h"
#include "TreeScope.h"
#include <wtf/CompactPointerTuple.h>
//#include <wtf/CompactUniquePtrTuple.h>
#include <wtf/Forward.h>
#include <wtf/ListHashSet.h>
#include <wtf/MainThread.h>
#include <wtf/OptionSet.h>
#include <wtf/WeakPtr.h>
#include <wtf/text/AtomicString.h>

namespace WebCore {

class Attribute;
class ClassCollection;
class ContainerNode;
class DOMTokenList ;
class Document;
class Element;
class EventListener;
class FloatPoint;
class Frame;
class HTMLInputElement;
class HTMLQualifiedName;
class HTMLSlotElement;
class IntRect;
class KeyboardEvent;
class MathMLQualifiedName;
class NSResolver;
class NameNodeList;
class NamedNodeMap;
class NodeList;
class NodeListsNodeData;
class NodeOrString;
class NodeRareData;
class PlatformKeyboardEvent;
class PlatformMouseEvent;
class PlatformWheelEvent;
class QualifiedName;
class RadioNodeList;
class RegisteredEventListener;
class RenderBox;
class RenderBoxModelObject;
class RenderObject;
class RenderStyle;
class SVGQualifiedName;
class ShadowRoot;
class TagCollection;

#if ENABLE(MICRODATA)
class HTMLPropertiesCollection;
class PropertyNodeList;
#endif

typedef int ExceptionCode;

class Node : public EventTarget {
	WTF_MAKE_FAST_ALLOCATED;

    friend class Document;
    friend class TreeScope;

public:
    enum NodeType {
        ELEMENT_NODE = 1,
        ATTRIBUTE_NODE = 2,
        TEXT_NODE = 3,
        CDATA_SECTION_NODE = 4,
        PROCESSING_INSTRUCTION_NODE = 7,
        COMMENT_NODE = 8,
        DOCUMENT_NODE = 9,
        DOCUMENT_TYPE_NODE = 10,
        DOCUMENT_FRAGMENT_NODE = 11,
#if ENABLE(RMAI)
        WAFER_NODE = 99
#endif
    };
    enum DeprecatedNodeType {
        ENTITY_REFERENCE_NODE = 5,
        ENTITY_NODE = 6,
        NOTATION_NODE = 12,
    };
    enum DocumentPosition {
        DOCUMENT_POSITION_EQUIVALENT = 0x00,
        DOCUMENT_POSITION_DISCONNECTED = 0x01,
        DOCUMENT_POSITION_PRECEDING = 0x02,
        DOCUMENT_POSITION_FOLLOWING = 0x04,
        DOCUMENT_POSITION_CONTAINS = 0x08,
        DOCUMENT_POSITION_CONTAINED_BY = 0x10,
        DOCUMENT_POSITION_IMPLEMENTATION_SPECIFIC = 0x20,
    };

    static bool isSupported(const String& feature, const String& version);

    static void startIgnoringLeaks();
    static void stopIgnoringLeaks();

    static void dumpStatistics();

    virtual ~Node();
    void willBeDeletedFrom(Document*);

    // DOM methods & attributes for Node

    bool hasTagName(const HTMLQualifiedName&) const;
    bool hasTagName(const MathMLQualifiedName&) const;
    bool hasTagName(const SVGQualifiedName&) const;
    virtual String nodeName() const = 0;
    virtual String nodeValue() const;
    virtual void setNodeValue(const String&);
    virtual NodeType nodeType() const = 0;
    ContainerNode* parentNode() const;
    Element* parentElement() const;
    Node* previousSibling() const { return m_previous; }
    Node* nextSibling() const { return m_next; }
    PassRefPtr<NodeList> childNodes();
    Node* firstChild() const;
    Node* lastChild() const;
    bool hasAttributes() const;
    NamedNodeMap* attributes() const;
    Node* pseudoAwareNextSibling() const;
    Node* pseudoAwarePreviousSibling() const;
    Node* pseudoAwareFirstChild() const;
    Node* pseudoAwareLastChild() const;

    virtual KURL baseURI() const;
    
    void getSubresourceURLs(ListHashSet<KURL>&) const;

    // These should all actually return a node, but this is only important for language bindings,
    // which will already know and hold a ref on the right node to return. Returning bool allows
    // these methods to be more efficient since they don't need to return a ref
    bool insertBefore(Node& newChild, Node* refChild, ExceptionCode&);
    bool replaceChild(Node& newChild, Node& oldChild, ExceptionCode&);
    bool removeChild(Node& child, ExceptionCode&);
    bool appendChild(Node& newChild, ExceptionCode&);

    ALWAYS_INLINE bool hasChildNodes() const { return firstChild(); }

    enum CloningOperation {
        OnlySelf,
        SelfWithTemplateContent,
        Everything,
    };
    virtual Ref<Node> cloneNodeInternal(Document&, CloningOperation) = 0;
    ALWAYS_INLINE Ref<Node> cloneNode(bool deep) { return cloneNodeInternal(document(), deep ? Everything : OnlySelf); }
    RefPtr<Node> cloneNodeForBindings(bool deep, ExceptionCode&);

    virtual const AtomicString& localName() const;
    virtual const AtomicString& namespaceURI() const;
    virtual const AtomicString& prefix() const;
    virtual void setPrefix(const AtomicString&, ExceptionCode&);
    void normalize();

    ALWAYS_INLINE bool isSameNode(Node* other) const { return this == other; }
    bool isEqualNode(Node*) const;
    bool isDefaultNamespace(const AtomicString& namespaceURI) const;
    const AtomicString& lookupPrefix(const AtomicString& namespaceURI) const;
    const AtomicString& lookupNamespaceURI(const AtomicString& prefix) const;
    
    String textContent(bool convertBRsToNewlines = false) const;
    void setTextContent(const String&);
    
    Node* lastDescendant() const;
    Node* firstDescendant() const;

    // From the NonDocumentTypeChildNode - https://dom.spec.whatwg.org/#nondocumenttypechildnode
    Element* previousElementSibling() const;
    Element* nextElementSibling() const;

    // From the ChildNode - https://dom.spec.whatwg.org/#childnode
    void before(const Vector<NodeOrString>&, ExceptionCode&);
    void after(const Vector<NodeOrString>&, ExceptionCode&);
    void replaceWith(const Vector<NodeOrString>&, ExceptionCode&);
    void remove(ExceptionCode&);

    virtual bool isActiveNode() const { return false; }
    
    // Other methods (not part of DOM)

    ALWAYS_INLINE bool isElementNode() const { return getFlag(IsElementFlag); }
    ALWAYS_INLINE bool isContainerNode() const { return getFlag(IsContainerFlag); }
    bool isCharacterData() const { return getFlag(IsCharacterDataFlag); }
    ALWAYS_INLINE bool isTextNode() const { return getFlag(IsTextFlag); }
    ALWAYS_INLINE bool isHTMLElement() const { return getFlag(IsHTMLFlag); }
    ALWAYS_INLINE bool isSVGElement() const { return getFlag(IsSVGFlag); }
    ALWAYS_INLINE bool isMathMLElement() const { return getFlag(IsMathMLFlag); }

    bool isUnknownElement() const { return getFlag(IsUnknownElementFlag); }
    bool isHTMLUnknownElement() const { return isHTMLElement() && isUnknownElement(); }
    bool isSVGUnknownElement() const { return isSVGElement() && isUnknownElement(); }
    bool isMathMLUnknownElement() const { return isMathMLElement() && isUnknownElement(); }

    ALWAYS_INLINE bool isPseudoElement() const { return pseudoId() != PseudoId::None; }
    ALWAYS_INLINE bool isBeforePseudoElement() const { return pseudoId() == PseudoId::Before; }
    ALWAYS_INLINE bool isAfterPseudoElement() const { return pseudoId() == PseudoId::After; }
    PseudoId pseudoId() const { return (isElementNode() && hasCustomStyleResolveCallbacks()) ? customPseudoId() : PseudoId::None; }

    virtual bool isMediaControlElement() const { return false; }
    virtual bool isMediaControls() const { return false; }
#if ENABLE(VIDEO_TRACK)
    virtual bool isWebVTTElement() const { return false; }
#endif
    ALWAYS_INLINE bool isStyledElement() const { return getFlag(IsHTMLFlag) || getFlag(IsSVGFlag) || getFlag(IsMathMLFlag); }
	virtual bool isAttributeNode() const { return false; }
	bool isCharacterDataNode() const { return getFlag(IsCharacterDataFlag); }
	virtual bool isFrameOwnerElement() const { return false; }
    virtual bool isPluginElement() const { return false; }

    ALWAYS_INLINE bool isDocumentNode() const { return getFlag(IsDocumentNodeFlag); }
    ALWAYS_INLINE bool isTreeScope() const { return getFlag(IsDocumentNodeFlag) || getFlag(IsShadowRootFlag); }
    ALWAYS_INLINE bool isDocumentFragment() const { return getFlag(IsDocumentFragmentFlag); }
    ALWAYS_INLINE bool isShadowRoot() const { return getFlag(IsShadowRootFlag); }
    bool isUserAgentShadowRoot() const; // Defined in ShadowRoot.h

    ALWAYS_INLINE static bool isShadowRoot(const Node& node) { return node.isShadowRoot(); }

    ALWAYS_INLINE bool hasCustomStyleResolveCallbacks() const { return getFlag(HasCustomStyleResolveCallbacksFlag); }

    ALWAYS_INLINE bool hasSyntheticAttrChildNodes() const { return getFlag(HasSyntheticAttrChildNodesFlag); }
    ALWAYS_INLINE void setHasSyntheticAttrChildNodes(bool flag) { setNodeFlag(HasSyntheticAttrChildNodesFlag, flag); }

    bool hasShadowRootContainingSlots() const { return getFlag(HasShadowRootContainingSlots); }
    void setHasShadowRootContainingSlots(bool flag) { setNodeFlag(HasShadowRootContainingSlots, flag); }

    bool needsSVGRendererUpdate() const { return hasNodeFlag(NeedsSVGRendererUpdate); }
    void setNeedsSVGRendererUpdate(bool flag) { setNodeFlag(NeedsSVGRendererUpdate, flag); }

    // If this node is in a shadow tree, returns its shadow host. Otherwise, returns 0.
    Element* shadowHost() const;
    // If this node is in a shadow tree, returns its shadow host. Otherwise, returns this.
    // Deprecated. Should use shadowHost() and check the return value.
    ShadowRoot* containingShadowRoot() const;
    ShadowRoot* shadowRoot() const; // Defined in ShadowRoot.h
    bool isClosedShadowHidden(const Node&) const;

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    HTMLSlotElement* assignedSlot() const;
    HTMLSlotElement* assignedSlotForBindings() const;
#endif

    bool isUndefinedCustomElement() const { return customElementState() == CustomElementState::Undefined || customElementState() == CustomElementState::Failed; }
    bool isCustomElementUpgradeCandidate() const { return customElementState() == CustomElementState::Undefined; }
    bool isDefinedCustomElement() const { return customElementState() == CustomElementState::Custom; }
    bool isFailedCustomElement() const { return customElementState() == CustomElementState::Failed; }

    // Returns 0, a child of ShadowRoot, or a legacy shadow root.
    Node* nonBoundaryShadowTreeRootNode();

    bool isInUserAgentShadowTree() const;

    // Node's parent or shadow tree host.
    ContainerNode* parentOrShadowHostNode() const; // Defined in ShadowRoot.h
    ContainerNode* parentInComposedTree() const;
    Element* parentElementInComposedTree() const;
    Element* parentOrShadowHostElement() const;
    void setParentNode(ContainerNode*);
    Node& rootNode() const;
    Node& traverseToRootNode() const;
    Node& shadowIncludingRoot() const;

    struct GetRootNodeOptions {
        bool composed { false };
    };
    Node& getRootNode(const GetRootNodeOptions&) const;

    void* opaqueRoot() const;

    // Use when it's guaranteed to that shadowHost is 0.
    ContainerNode* parentNodeGuaranteedHostFree() const;
    // Returns the parent node, but 0 if the parent node is a ShadowRoot.
    ContainerNode* nonShadowBoundaryParentNode() const;

    ALWAYS_INLINE bool selfOrAncestorHasDirAutoAttribute() const { return getFlag(SelfOrAncestorHasDirAutoFlag); }
    ALWAYS_INLINE void setSelfOrAncestorHasDirAutoAttribute(bool value) { setNodeFlag(SelfOrAncestorHasDirAutoFlag, value); }

    // Returns the enclosing event parent node (or self) that, when clicked, would trigger a navigation.
    Node* enclosingLinkEventParentOrSelf();
    
    // These low-level calls give the caller responsibility for maintaining the integrity of the tree.
    ALWAYS_INLINE void setPreviousSibling(Node* previous) { m_previous = previous; }
    ALWAYS_INLINE void setNextSibling(Node* next) { m_next = next; }

    virtual bool canContainRangeEndPoint() const { return false; }

    // FIXME: These two functions belong in editing -- "atomic node" is an editing concept.
    Node* previousNodeConsideringAtomicNodes() const;
    Node* nextNodeConsideringAtomicNodes() const;
    
    // Returns the next leaf node or 0 if there are no more.
    // Delivers leaf nodes as if the whole DOM tree were a linear chain of its leaf nodes.
    // Uses an editing-specific concept of what a leaf node is, and should probably be moved
    // out of the Node class into an editing-specific source file.
    Node* nextLeafNode() const;

    // Returns the previous leaf node or 0 if there are no more.
    // Delivers leaf nodes as if the whole DOM tree were a linear chain of its leaf nodes.
    // Uses an editing-specific concept of what a leaf node is, and should probably be moved
    // out of the Node class into an editing-specific source file.
    Node* previousLeafNode() const;

    bool isRootEditableElement() const;
    Element* rootEditableElement() const;

    // Called by the parser when this element's close tag is reached,
    // signaling that all child tags have been parsed and added.
    // This is needed for <applet> and <object> elements, which can't lay themselves out
    // until they know all of their nested <param>s. [Radar 3603191, 4040848].
    // Also used for script elements and some SVG elements for similar purposes,
    // but making parsing a special case in this respect should be avoided if possible.
    virtual void finishParsingChildren() { }
    virtual void beginParsingChildren() { }

    // For <link> and <style> elements.
    virtual bool sheetLoaded() { return true; }
    virtual void notifyLoadedSheetAndAllCriticalSubresources(bool /* error loading subresource */) { }
    virtual void startLoadingDynamicSheet() { ASSERT_NOT_REACHED(); }

    bool hasID() const;
    bool hasClass() const;

    bool isUserActionElement() const { return getFlag(IsUserActionElement); }
    void setUserActionElement(bool value) { setNodeFlag(IsUserActionElement, value); }

    bool inRenderedDocument() const;
    bool needsStyleRecalc() const { return styleValidity() != Style::Validity::Valid || hasInvalidRenderer(); }
    Style::Validity styleValidity() const { return styleBitfields().styleValidity(); }
    bool hasInvalidRenderer() const { return hasStyleFlag(NodeStyleFlag::HasInvalidRenderer); }
    bool styleResolutionShouldRecompositeLayer() const { return hasStyleFlag(NodeStyleFlag::StyleResolutionShouldRecompositeLayer); }
	bool childNeedsStyleRecalc() const { return hasStyleFlag(NodeStyleFlag::DescendantNeedsStyleResolution); }
	bool isEditingText() const { return getFlag(IsTextFlag) && getFlag(IsEditingText); }

    bool isDocumentFragmentForInnerOuterHTML() const { return getFlag(IsDocumentFragmentForInnerOuterHTML); }

    void setChildNeedsStyleRecalc() { setStyleFlag(NodeStyleFlag::DescendantNeedsStyleResolution); }
    void clearChildNeedsStyleRecalc();

    void setHasValidStyle();

	ALWAYS_INLINE bool isLink() const { return getFlag(IsLinkFlag); }
    ALWAYS_INLINE void setIsLink(bool value = true) { setNodeFlag(IsLinkFlag, value); }
    ALWAYS_INLINE void clearIsLink() { clearFlag(IsLinkFlag); }

    ALWAYS_INLINE bool hasEventTargetData() const { return getFlag(HasEventTargetDataFlag); }
    ALWAYS_INLINE void setHasEventTargetData(bool value) { setNodeFlag(HasEventTargetDataFlag, value); }

    bool isInGCReacheableRefMap() const { return hasNodeFlag(NodeFlags::IsInGCReachableRefMap); }
    void setIsInGCReacheableRefMap(bool flag) { setNodeFlag(NodeFlags::IsInGCReachableRefMap, flag); }

    virtual bool isSpatialNavigationFocusable() const { return true; }

    bool isContentEditable() const;
    bool isContentRichlyEditable() const;

    void inspect();

    enum UserSelectAllTreatment {
        UserSelectAllDoesNotAffectEditability,
        UserSelectAllIsAlwaysNonEditable
    };
    enum class Editability { ReadOnly, CanEditPlainText, CanEditRichly };
    enum class ShouldUpdateStyle { Update, DoNotUpdate };
    bool hasEditableStyle(UserSelectAllTreatment treatment = UserSelectAllIsAlwaysNonEditable) const
    {
        return computeEditability(treatment, ShouldUpdateStyle::DoNotUpdate) != Editability::ReadOnly;
    }

    // FIXME: Replace every use of this function by helpers in htmlediting.h
    bool hasRichlyEditableStyle() const
    {
        return computeEditability(UserSelectAllIsAlwaysNonEditable, ShouldUpdateStyle::DoNotUpdate) == Editability::CanEditRichly;
    }

    Editability computeEditability(UserSelectAllTreatment, ShouldUpdateStyle) const;

    virtual bool shouldUseInputMethod();
    LayoutRect renderRect(bool* isReplaced);
    IntRect pixelSnappedRenderRect(bool* isReplaced) { return snappedIntRect(renderRect(isReplaced)); }

    WEBCORE_EXPORT unsigned computeNodeIndex() const;
    unsigned nodeIndex() const { return computeNodeIndex(); }

    // Returns the DOM ownerDocument attribute. This method never returns 0, except in the case
    // of a Document node.
    Document* ownerDocument() const;

    // Returns the document associated with this node. This method never returns 0.
    // A Document node returns itself.
    ALWAYS_INLINE Document& document() const
    {
        ASSERT(this);
        return treeScope().documentScope();
    }

    TreeScope& treeScope() const
    {
        ASSERT(m_treeScope);
        return *m_treeScope;
    }
    void setTreeScopeRecursively(TreeScope&);

    TreeScope& treeScopeForSVGReferences() const;

    // Returns true if this node is associated with a document and is in its associated document's
    // node tree, false otherwise.
    ALWAYS_INLINE bool inDocument() const { return getFlag(IsConnectedFlag); }
    ALWAYS_INLINE bool isConnected() const { return getFlag(IsConnectedFlag); }
    bool isInShadowTree() const { return getFlag(IsInShadowTreeFlag); }
    bool isInTreeScope() const { return getFlag(static_cast<NodeFlags>(IsConnectedFlag | IsInShadowTreeFlag)); }
    bool hasBeenInUserAgentShadowTree() const { return getFlag(HasBeenInUserAgentShadowTree); }

    bool isDocumentTypeNode() const { return nodeType() == DOCUMENT_TYPE_NODE; }
    virtual bool childTypeAllowed(NodeType) const { return false; }
    unsigned countChildNodes() const;
    unsigned length() const;
    Node* traverseToChildAt(unsigned) const;

    void checkSetPrefix(const AtomicString& prefix, ExceptionCode&);

    bool isDescendantOf(const Node&) const;
    inline bool isDescendantOf(const Node* other) const { return other && isDescendantOf(*other); }
    WEBCORE_EXPORT bool contains(const Node&) const;
	bool contains(const Node* other) const { return other && contains(*other); }

    WEBCORE_EXPORT bool isDescendantOrShadowDescendantOf(const Node&) const;
    bool isDescendantOrShadowDescendantOf(const Node* other) const { return other && isDescendantOrShadowDescendantOf(*other); }
    bool containsIncludingShadowDOM(const Node*) const;

    // Whether or not a selection can be started in this object
    virtual bool canStartSelection() const;

    // Getting points into and out of screen space
    FloatPoint convertToPage(const FloatPoint&) const;
    FloatPoint convertFromPage(const FloatPoint&) const;

    // -----------------------------------------------------------------------------
    // Integration with rendering tree

    // As renderer() includes a branch you should avoid calling it repeatedly in hot code paths.
    RenderObject* renderer() const { return m_rendererWithStyleFlags.pointer(); }
	void setRenderer(RenderObject*); // Defined in RenderObject.h
    
    // Use these two methods with caution.
    RenderBox* renderBox() const;
    RenderBoxModelObject* renderBoxModelObject() const;
    
    // Wrapper for nodes that don't have a renderer, but still cache the style (like HTMLOptionElement).
    const RenderStyle* renderStyle() const;

    virtual RenderStyle* computedStyle(PseudoId pseudoElementSpecifier = PseudoId::None);

    enum class InsertedIntoResult {
        Done,
        NeedsPostInsertionCallback,
    };

    struct InsertionType {
    	InsertionType() : connectedToDocument(false), treeScopeChanged(false) { }
        InsertionType(bool connectedToDocument, bool treeScopeChanged)
            : connectedToDocument(connectedToDocument)
            , treeScopeChanged(treeScopeChanged)
        { }
        bool connectedToDocument;
        bool treeScopeChanged;
    };
    // Called *after* this node or its ancestor is inserted into a new parent (may or may not be a part of document) by scripts or parser.
    // insertedIntoAncestor **MUST NOT** invoke scripts. Return NeedsPostInsertionCallback and implement didFinishInsertingNode instead to run scripts.
    virtual InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode& parentOfInsertedTree);
    virtual void didFinishInsertingNode() { }

    struct RemovalType {
    	RemovalType() : disconnectedFromDocument(false), treeScopeChanged(false) { }
        RemovalType(bool disconnectedFromDocument, bool treeScopeChanged)
            : disconnectedFromDocument(disconnectedFromDocument)
            , treeScopeChanged(treeScopeChanged)
        { }
        bool disconnectedFromDocument;
        bool treeScopeChanged;
    };
    virtual void removedFromAncestor(const RemovalType&, ContainerNode& parentOfRemovedTree);

#ifndef NDEBUG
    virtual void formatForDebugger(char* buffer, unsigned length) const;

    void showNode(const char* prefix = "") const;
    void showTreeForThis() const;
    void showNodePathForThis() const;
    void showTreeAndMark(const Node* markedNode1, const char* markedLabel1, const Node* markedNode2 = 0, const char* markedLabel2 = 0) const;
    void showTreeForThisAcrossFrame() const;
#endif

    void invalidateNodeListAndCollectionCachesInAncestors();
	void invalidateNodeListAndCollectionCachesInAncestorsForAttribute(const QualifiedName& attrName);
	NodeListsNodeData* nodeLists();
    void clearNodeLists();

    virtual bool willRespondToMouseMoveEvents();
    virtual bool willRespondToMouseClickEvents();

    unsigned short compareDocumentPosition(Node&);

    virtual bool isNode() const FINAL;
    virtual RefPtr<Node> toNode() FINAL;
    virtual HTMLInputElement* toInputElement();
    const HTMLInputElement* toInputElement() const { return const_cast<Node*>(this)->toInputElement(); }

    virtual EventTargetInterface eventTargetInterface() const OVERRIDE;
    virtual ScriptExecutionContext* scriptExecutionContext() const final;

    virtual bool addEventListener(const AtomicString& eventType, PassRefPtr<EventListener>, const AddEventListenerOptions&) override;
    virtual bool removeEventListener(const AtomicString& eventType, EventListener*, const EventListenerOptions& = EventListenerOptions()) OVERRIDE;

    using EventTarget::dispatchEvent;
    virtual void dispatchEvent(PassRefPtr<Event>) OVERRIDE;

    void dispatchScopedEvent(PassRefPtr<Event>);

    void dispatchSubtreeModifiedEvent();
    void dispatchDOMActivateEvent(PassRefPtr<Event> underlyingClickEvent);

    bool dispatchKeyEvent(const PlatformKeyboardEvent&);
    bool dispatchWheelEvent(const PlatformWheelEvent&);

    void dispatchInputEvent();

    // Perform the default action for an event.
    virtual void defaultEventHandler(Event*);

    void ref() const;
    void deref() const;
    bool hasOneRef() const;
    unsigned refCount() const;

#ifndef NDEBUG
    bool m_deletionHasBegun { false };
    mutable bool m_inRemovedLastRefFunction { false };
    bool m_adoptionIsRequired { true };
#endif

    virtual EventTargetData* eventTargetData() final;
    virtual EventTargetData* ensureEventTargetData() final;

#if ENABLE(MICRODATA)
    DOMTokenList * itemProp();
    DOMTokenList * itemRef();
    DOMTokenList * itemType();
    PassRefPtr<PropertyNodeList> propertyNodeList(const String&);
#endif

#if ENABLE(MUTATION_OBSERVERS)
    void getRegisteredMutationObserversOfType(HashMap<MutationObserver*, MutationRecordDeliveryOptions>&, MutationObserver::MutationType, const QualifiedName* attributeName);
    MutationObserverRegistration* registerMutationObserver(PassRefPtr<MutationObserver>);
    void unregisterMutationObserver(MutationObserverRegistration&);
    void registerTransientMutationObserver(MutationObserverRegistration&);
    void unregisterTransientMutationObserver(MutationObserverRegistration&);
    void notifyMutationObserversNodeWillDetach();
#endif // ENABLE(MUTATION_OBSERVERS)

    unsigned connectedSubframeCount() const { return rareDataBitfields().connectedSubframeCount; }
    void incrementConnectedSubframeCount(unsigned amount = 1);
    void decrementConnectedSubframeCount(unsigned amount = 1);
    void updateAncestorConnectedSubframeCountForRemoval() const;
    void updateAncestorConnectedSubframeCountForInsertion() const;

#if ENABLE(CSS_SELECTOR_JIT)
    static ptrdiff_t nodeFlagsMemoryOffset() { return OBJECT_OFFSETOF(Node, m_nodeFlags); }
    static ptrdiff_t rareDataMemoryOffset() { return OBJECT_OFFSETOF(Node, m_rareDataWithBitfields); }
#if CPU(ADDRESS64)
    static uint64_t rareDataPointerMask() { return CompactPointerTuple<NodeRareData*, uint16_t>::pointerMask; }
#else
    static uint32_t rareDataPointerMask() { return -1; }
#endif
    static int32_t flagIsText() { return IsTextFlag; }
    static int32_t flagIsElement() { return IsElementFlag; }
    static int32_t flagIsShadowRoot() { return IsShadowRootFlag; }
    static int32_t flagIsHTML() { return IsHTMLFlag; }
    static int32_t flagIsLink() { return IsLinkFlag; }
    static int32_t flagIsParsingChildrenFinished() { return IsParsingChildrenFinishedFlag; }
    static int32_t flagChildrenAffectedByFirstChildRulesFlag() { return ChildrenAffectedByFirstChildRulesFlag; }
    static int32_t flagChildrenAffectedByLastChildRulesFlag() { return ChildrenAffectedByLastChildRulesFlag; }

    static int32_t flagAffectsNextSiblingElementStyle() { return AffectsNextSiblingElementStyle; }
    static int32_t flagStyleIsAffectedByPreviousSibling() { return StyleIsAffectedByPreviousSibling; }
#endif // ENABLE(CSS_SELECTOR_JIT)

protected:
    enum NodeFlags {
        IsCharacterDataFlag = 1 << 0,
        IsTextFlag = 1 << 1,
        IsContainerFlag = 1 << 2,
        IsElementFlag = 1 << 3,
        IsHTMLFlag = 1 << 4,
        IsSVGFlag = 1 << 5,
        IsMathMLFlag = 1 << 6,
        IsDocumentNodeFlag = 1 << 7,
        IsDocumentFragmentFlag = 1 << 8,
        IsShadowRootFlag = 1 << 9,
        IsConnectedFlag = 1 << 10,
        IsInShadowTreeFlag = 1 << 11,
        IsUnknownElementFlag = 1 << 12,
        HasEventTargetDataFlag = 1 << 13,

        // These bits are used by derived classes, pulled up here so they can
        // be stored in the same memory word as the Node bits above.
        IsDocumentFragmentForInnerOuterHTML = 1 << 14, // DocumentFragment
        IsEditingText = 1 << 15, // Text
        IsLinkFlag = 1 << 16, // Element
        IsUserActionElement = 1 << 17,
        IsParsingChildrenFinishedFlag = 1 << 18,
        HasSyntheticAttrChildNodesFlag = 1 << 19,
        SelfOrAncestorHasDirAutoFlag = 1 << 20,

        HasCustomStyleResolveCallbacksFlag = 1 << 21,

        HasPendingResourcesFlag = 1 << 22,
        IsInGCReachableRefMap = 1 << 23,
#if ENABLE(FULLSCREEN_API)
        ContainsFullScreenElementFlag = 1 << 24,
#endif
        IsComputedStyleInvalidFlag = 1 << 25,
        HasShadowRootContainingSlots = 1 << 26,
        IsInTopLayer = 1 << 27,
        NeedsSVGRendererUpdate = 1 << 28,

        HasElementIdentifierFlag = 1 << 29,
        HasBeenInUserAgentShadowTree = 1 << 30,

        // Bit 31 is free.

        DefaultNodeFlags = IsParsingChildrenFinishedFlag,
    };

    enum class TabIndexState : uint8_t {
        NotSet = 0,
        Zero = 1,
        NegativeOne = 2,
        InRareData = 3,
    };

    enum class CustomElementState : uint8_t {
        Uncustomized = 0,
        Undefined = 1,
        Custom = 2,
        Failed = 3,
    };

    struct RareDataBitFields {
        uint16_t connectedSubframeCount : 10;
        uint16_t tabIndexState : 2;
        uint16_t customElementState : 2;
    };

    ALWAYS_INLINE bool hasNodeFlag(NodeFlags flag) const { return m_nodeFlags.contains(flag); }
    ALWAYS_INLINE void setNodeFlag(NodeFlags flag, bool value = true) const { m_nodeFlags.set(flag, value); }
    ALWAYS_INLINE void clearNodeFlag(NodeFlags flag) const { m_nodeFlags.remove(flag); }

    ALWAYS_INLINE bool getFlag(NodeFlags flag) const { return hasNodeFlag(flag); }
    ALWAYS_INLINE void clearFlag(NodeFlags flag) const { clearNodeFlag(flag); }

    RareDataBitFields rareDataBitfields() const { return bitwise_cast<RareDataBitFields>(m_rareDataWithBitfields.type()); }
    void setRareDataBitfields(RareDataBitFields bitfields) { m_rareDataWithBitfields.setType(bitwise_cast<uint16_t>(bitfields)); }

    TabIndexState tabIndexState() const { return static_cast<TabIndexState>(rareDataBitfields().tabIndexState); }
    void setTabIndexState(TabIndexState);

    CustomElementState customElementState() const { return static_cast<CustomElementState>(rareDataBitfields().customElementState); }
    void setCustomElementState(CustomElementState);

    bool isParsingChildrenFinished() const { return getFlag(IsParsingChildrenFinishedFlag); }
    void setIsParsingChildrenFinished() { setNodeFlag(IsParsingChildrenFinishedFlag); }
    void clearIsParsingChildrenFinished() { clearFlag(IsParsingChildrenFinishedFlag); }

    enum ConstructionType {
        CreateOther = DefaultNodeFlags,
        CreateCharacterData = DefaultNodeFlags | IsCharacterDataFlag,
        CreateText = CreateCharacterData | IsTextFlag,
        CreateContainer = DefaultNodeFlags | IsContainerFlag,
        CreateElement = CreateContainer | IsElementFlag,
        CreatePseudoElement =  CreateElement | IsConnectedFlag,
        CreateShadowRoot = CreateContainer | IsShadowRootFlag | IsInShadowTreeFlag,
        CreateDocumentFragment = CreateContainer | IsDocumentFragmentFlag,
        CreateHTMLElement = CreateElement | IsHTMLFlag,
        CreateSVGElement = CreateElement | IsSVGFlag | HasCustomStyleResolveCallbacksFlag,
        CreateMathMLElement = CreateElement | IsMathMLFlag,
        CreateDocument = CreateContainer | IsDocumentNodeFlag | IsConnectedFlag,
        CreateEditingText = CreateText | IsEditingText,
    };
    Node(Document*, ConstructionType);

    static constexpr uint32_t s_refCountIncrement = 2;
    static constexpr uint32_t s_refCountMask = ~static_cast<uint32_t>(1);

    enum class NodeStyleFlag : uint16_t {
        DescendantNeedsStyleResolution                          = 1 << 0,
        DirectChildNeedsStyleResolution                         = 1 << 1,
        StyleResolutionShouldRecompositeLayer                   = 1 << 2,
        HasInvalidRenderer                                      = 1 << 3,

        ChildrenAffectedByFirstChildRules                       = 1 << 4,
        ChildrenAffectedByLastChildRules                        = 1 << 5,
        AffectsNextSiblingElementStyle                          = 1 << 6,
        StyleIsAffectedByPreviousSibling                        = 1 << 7,
        DescendantsAffectedByPreviousSibling                    = 1 << 8,
        StyleAffectedByEmpty                                    = 1 << 9,
        // We optimize for :first-child and :last-child. The other positional child selectors like nth-child or
        // *-child-of-type, we will just give up and re-evaluate whenever children change at all.
        ChildrenAffectedByForwardPositionalRules                = 1 << 10,
        DescendantsAffectedByForwardPositionalRules             = 1 << 11,
        ChildrenAffectedByBackwardPositionalRules               = 1 << 12,
        DescendantsAffectedByBackwardPositionalRules            = 1 << 13,
    };

    struct StyleBitfields {
    public:
        static StyleBitfields fromRaw(uint16_t packed) { return bitwise_cast<StyleBitfields>(packed); }
        uint16_t toRaw() const { return bitwise_cast<uint16_t>(*this); }

        Style::Validity styleValidity() const { return static_cast<Style::Validity>(m_styleValidity); }
        void setStyleValidity(Style::Validity validity) { m_styleValidity = static_cast<uint8_t>(validity); }

        OptionSet<NodeStyleFlag> flags() const { return OptionSet<NodeStyleFlag>::fromRaw(m_flags); }
        void setFlag(NodeStyleFlag flag) { m_flags = (flags() | flag).toRaw(); }
        void clearFlag(NodeStyleFlag flag) { m_flags = (flags() - flag).toRaw(); }
        void clearFlags(OptionSet<NodeStyleFlag> flagsToClear) { m_flags = (flags() - flagsToClear).toRaw(); }
		void clearDescendantsNeedStyleResolution() { m_flags = (flags() - NodeStyleFlag::DescendantNeedsStyleResolution - NodeStyleFlag::DirectChildNeedsStyleResolution).toRaw(); }

    private:
        uint16_t m_styleValidity : 2;
        uint16_t m_flags : 14;
    };

    StyleBitfields styleBitfields() const { return StyleBitfields::fromRaw(m_rendererWithStyleFlags.type()); }
    void setStyleBitfields(StyleBitfields bitfields) { m_rendererWithStyleFlags.setType(bitfields.toRaw()); }
    ALWAYS_INLINE bool hasStyleFlag(NodeStyleFlag flag) const { return styleBitfields().flags().contains(flag); }
    ALWAYS_INLINE void setStyleFlag(NodeStyleFlag);
    ALWAYS_INLINE void clearStyleFlags(OptionSet<NodeStyleFlag>);

    virtual PseudoId customPseudoId() const
    {
        ASSERT(hasCustomStyleResolveCallbacks());
        return PseudoId::None;
    }

    virtual void addSubresourceAttributeURLs(ListHashSet<KURL>&) const { }

    bool hasRareData() const { return !!m_rareDataWithBitfields.pointer(); }
    NodeRareData* rareData() const { return m_rareDataWithBitfields.pointer(); }
    NodeRareData& ensureRareData();
    void clearRareData();

    void clearEventTargetData();

    void setHasCustomStyleResolveCallbacks() { setNodeFlag(HasCustomStyleResolveCallbacksFlag); }

    void setTreeScope(TreeScope& scope) { m_treeScope = &scope; }

    void invalidateStyle(Style::Validity, Style::InvalidationMode = Style::InvalidationMode::Normal);
    void updateAncestorsForStyleRecalc();

private:
    void removedLastRef();

    virtual void refEventTarget() final;
    virtual void derefEventTarget() final;

    void trackForDebugging();
    void materializeRareData();

#if ENABLE(MUTATION_OBSERVERS)
    Vector<OwnPtr<MutationObserverRegistration> >* mutationObserverRegistry();
    WeakHashSet<MutationObserverRegistration>* transientMutationObserverRegistry();
    void collectMatchingObserversForMutation(HashMap<MutationObserver*, MutationRecordDeliveryOptions>&, Node* fromNode, MutationObserver::MutationType, const QualifiedName* attributeName);
#endif

    void adjustStyleValidity(Style::Validity, Style::InvalidationMode);

    static void moveShadowTreeToNewDocument(ShadowRoot&, Document& oldDocument, Document& newDocument);
    static void moveTreeToNewScope(Node&, TreeScope& oldScope, TreeScope& newScope);
    void moveNodeToNewDocument(Document& oldDocument, Document& newDocument);

    void* opaqueRootSlow() const;

    mutable uint32_t m_refCountAndParentBit;
    mutable OptionSet<NodeFlags> m_nodeFlags;

    ContainerNode* m_parentNode;
    TreeScope* m_treeScope;
    Node* m_previous;
    Node* m_next;
    CompactPointerTuple<RenderObject*, uint16_t, uint16_t> m_rendererWithStyleFlags;
    CompactPointerTuple<NodeRareData*, uint16_t, uint16_t> m_rareDataWithBitfields;

protected:
#if ENABLE(MICRODATA)
    void setItemProp(const String&);
    void setItemRef(const String&);
    void setItemType(const String&);
#endif
};

bool connectedInSameTreeScope(const Node*, const Node*);

// Designed to be used the same way as C++20 std::partial_ordering class.
// FIXME: Consider putting this in a separate header.
// FIXME: Once we can require C++20, replace with std::partial_ordering.
class PartialOrdering {
public:
    static constexpr PartialOrdering less();
    static constexpr PartialOrdering equivalent();
    static constexpr PartialOrdering greater();
    static constexpr PartialOrdering unordered();

    friend constexpr bool is_eq(PartialOrdering);
    friend constexpr bool is_lt(PartialOrdering);
    friend constexpr bool is_gt(PartialOrdering);

private:
    enum class Type : uint8_t { Less, Equivalent, Greater, Unordered };
    constexpr PartialOrdering(Type type) : m_type { type } { }
    Type m_type;
};
constexpr bool is_eq(PartialOrdering);
constexpr bool is_lt(PartialOrdering);
constexpr bool is_gt(PartialOrdering);
constexpr bool is_neq(PartialOrdering);
constexpr bool is_lteq(PartialOrdering);
constexpr bool is_gteq(PartialOrdering);

enum TreeType { Tree, ShadowIncludingTree, ComposedTree };
template<TreeType = Tree> ContainerNode* parent(const Node&);
template<TreeType = Tree> Node* commonInclusiveAncestor(const Node&, const Node&);
template<TreeType = Tree> PartialOrdering treeOrder(const Node&, const Node&);

WEBCORE_EXPORT PartialOrdering treeOrderForTesting(TreeType, const Node&, const Node&);

#ifndef NDEBUG
inline void adopted(Node* node)
{
    if (!node)
        return;
    ASSERT(!node->m_deletionHasBegun);
    ASSERT(!node->m_inRemovedLastRefFunction);
    node->m_adoptionIsRequired = false;
}
#endif

#define DEBUG_DOCUMENT_LEAK 0

#if !DEBUG_DOCUMENT_LEAK
ALWAYS_INLINE void Node::ref() const
{
    ASSERT(isMainThread());
    ASSERT(!m_deletionHasBegun);
    ASSERT(!m_inRemovedLastRefFunction);
    ASSERT(!m_adoptionIsRequired);
    m_refCountAndParentBit += s_refCountIncrement;
}

ALWAYS_INLINE void Node::deref() const
{
    ASSERT(isMainThread());
    ASSERT(refCount());
    ASSERT(!m_deletionHasBegun);
    ASSERT(!m_inRemovedLastRefFunction);
    ASSERT(!m_adoptionIsRequired);
    auto updatedRefCount = m_refCountAndParentBit - s_refCountIncrement;
    if (!updatedRefCount) {
        // Don't update m_refCountAndParentBit to avoid double destruction through use of Ref<T>/RefPtr<T>.
        // (This is a security mitigation in case of programmer error. It will ASSERT in debug builds.)
#ifndef NDEBUG
        m_inRemovedLastRefFunction = true;
#endif
        const_cast<Node&>(*this).removedLastRef();
        return;
    }
    m_refCountAndParentBit = updatedRefCount;
}
#endif

ALWAYS_INLINE bool Node::hasOneRef() const
{
    ASSERT(!m_deletionHasBegun);
    ASSERT(!m_inRemovedLastRefFunction);
    return refCount() == 1;
}

ALWAYS_INLINE unsigned Node::refCount() const
{
    return m_refCountAndParentBit / s_refCountIncrement;
}

// Used in Node::addSubresourceAttributeURLs() and in addSubresourceStyleURLs()
inline void addSubresourceURL(ListHashSet<KURL>& urls, const KURL& url)
{
    if (!url.isNull())
        urls.add(url);
}

inline void Node::setParentNode(ContainerNode* parent)
{
    ASSERT(isMainThread());
    m_parentNode = parent;
    m_refCountAndParentBit = (m_refCountAndParentBit & s_refCountMask) | !!parent;
}

inline ContainerNode* Node::parentNode() const
{
    ASSERT(isMainThreadOrGCThread());
    return m_parentNode;
}

inline void* Node::opaqueRoot() const
{
    // FIXME: Possible race?
    // https://bugs.webkit.org/show_bug.cgi?id=165713
    if (inDocument())
        return &document();
    return opaqueRootSlow();
}

inline ContainerNode* Node::parentNodeGuaranteedHostFree() const
{
    ASSERT(!isShadowRoot());
    return parentNode();
}

ALWAYS_INLINE void Node::setStyleFlag(NodeStyleFlag flag)
{
    auto bitfields = styleBitfields();
    bitfields.setFlag(flag);
    setStyleBitfields(bitfields);
}

ALWAYS_INLINE void Node::clearStyleFlags(OptionSet<NodeStyleFlag> flags)
{
    auto bitfields = styleBitfields();
    bitfields.clearFlags(flags);
    setStyleBitfields(bitfields);
}

inline void Node::clearChildNeedsStyleRecalc()
{
    auto bitfields = styleBitfields();
    bitfields.clearDescendantsNeedStyleResolution();
    setStyleBitfields(bitfields);
}

inline void Node::setHasValidStyle()
{
    auto bitfields = styleBitfields();
    bitfields.setStyleValidity(Style::Validity::Valid);
    bitfields.clearFlags({ NodeStyleFlag::HasInvalidRenderer, NodeStyleFlag::StyleResolutionShouldRecompositeLayer });
    setStyleBitfields(bitfields);
    clearFlag(NodeFlags::IsComputedStyleInvalidFlag);
}

inline void Node::setTreeScopeRecursively(TreeScope& newTreeScope)
{
    ASSERT(!isDocumentNode());
    ASSERT(!m_deletionHasBegun);
    if (m_treeScope != &newTreeScope)
        moveTreeToNewScope(*this, *m_treeScope, newTreeScope);
}

inline bool Node::isNode() const
{
    return true;
}

inline RefPtr<Node> Node::toNode()
{
    return this;
}

inline bool Node::shouldUseInputMethod()
{
    return isContentEditable();
}

inline void Node::refEventTarget()
{
    ref();
}

inline void Node::derefEventTarget()
{
    deref();
}

inline void Node::dispatchScopedEvent(PassRefPtr<Event> event)
{
    EventDispatcher::dispatchScopedEvent(*this, event);
}

inline void Node::dispatchEvent(PassRefPtr<Event> event)
{
    EventDispatcher::dispatchEvent(this, event);
}

inline bool GCReachableRefMap::contains(Node& node)
{
	return node.isInGCReacheableRefMap();
}

inline void GCReachableRefMap::add(Node& node)
{
    if (map().add(&node).isNewEntry)
        node.setIsInGCReacheableRefMap(true);
}

inline void GCReachableRefMap::remove(Node& node)
{
    if (map().remove(&node))
        node.setIsInGCReacheableRefMap(false);
}

#define NODE_TYPE_CASTS(ToClassName) \
    TYPE_CASTS_BASE(ToClassName, Node, node, WebCore::is##ToClassName(*node), WebCore::is##ToClassName(node))

inline Node& toNode(EventTarget& eventTarget)
{
	return static_cast<Node&>(eventTarget);
}

inline const Node& toNode(const EventTarget& eventTarget)
{
	return static_cast<const Node&>(eventTarget);
}

inline constexpr PartialOrdering PartialOrdering::less() { return PartialOrdering(Type::Less); }
inline constexpr PartialOrdering PartialOrdering::equivalent() { return PartialOrdering(Type::Equivalent); }
inline constexpr PartialOrdering PartialOrdering::greater() { return PartialOrdering(Type::Greater); }
inline constexpr PartialOrdering PartialOrdering::unordered() { return PartialOrdering(Type::Unordered); }

constexpr bool is_eq(PartialOrdering ordering)
{
    return ordering.m_type == PartialOrdering::Type::Equivalent;
}

constexpr bool is_lt(PartialOrdering ordering)
{
    return ordering.m_type == PartialOrdering::Type::Less;
}

constexpr bool is_gt(PartialOrdering ordering)
{
    return ordering.m_type == PartialOrdering::Type::Greater;
}

constexpr bool is_neq(PartialOrdering ordering)
{
    return is_lt(ordering) || is_gt(ordering);
}

constexpr bool is_lteq(PartialOrdering ordering)
{
    return is_lt(ordering) || is_eq(ordering);
}

constexpr bool is_gteq(PartialOrdering ordering)
{
    return is_gt(ordering) || is_eq(ordering);
}

} //namespace

#ifndef NDEBUG
// Outside the WebCore namespace for ease of invocation from gdb.
void showTree(const WebCore::Node*);
void showNodePath(const WebCore::Node*);
#endif
