/*
 * Copyright (C) 2008, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2008 David Smith <catfish.man@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "ChildNodeList.h"
#include "ClassCollection.h"
#include "DOMTokenList.h"
#include "HTMLCollection.h"
#include "LiveNodeList.h"
#include "MutationObserver.h"
#include "MutationObserverRegistration.h"
#include "Page.h"
#include "QualifiedName.h"
#include "TagCollection.h"
#if ENABLE(VIDEO_TRACK)
#include "TextTrack.h"
#endif
#include <wtf/HashSet.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/WeakHashSet.h>
#include <wtf/text/AtomicString.h>

#if ENABLE(MICRODATA)
#include "HTMLPropertiesCollection.h"
#include "MicroDataItemList.h"
#endif

namespace WebCore {

class LabelsNodeList;
class NameNodeList;
class RadioNodeList;
class TreeScope;

template <class ListType> struct NodeListTypeIdentifier;
template <> struct NodeListTypeIdentifier<NameNodeList> { static int value() { return 0; } };
template <> struct NodeListTypeIdentifier<RadioNodeList> { static int value() { return 1; } };
template <> struct NodeListTypeIdentifier<LabelsNodeList> { static int value() { return 2; } };

class NodeListsNodeData {
    WTF_MAKE_NONCOPYABLE(NodeListsNodeData); WTF_MAKE_FAST_ALLOCATED;
public:
    void clearChildNodeListCache()
    {
        if (m_childNodeList)
            m_childNodeList->invalidateCache();
    }

    PassRefPtr<ChildNodeList> ensureChildNodeList(ContainerNode* node)
    {
        ASSERT(!m_emptyChildNodeList);
        if (m_childNodeList)
            return *m_childNodeList;
        auto list = ChildNodeList::create(node);
        m_childNodeList = makeWeakPtr(list.get());
        return list;
    }

    void removeChildNodeList(ChildNodeList* list)
    {
        ASSERT(m_childNodeList == list);
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(list->ownerNode()))
            return;
        m_childNodeList = nullptr;
    }

    PassRefPtr<EmptyNodeList> ensureEmptyChildNodeList(Node* node)
    {
        ASSERT(!m_childNodeList);
        if (m_emptyChildNodeList)
            return *m_emptyChildNodeList;
        auto list = EmptyNodeList::create(node);
        m_emptyChildNodeList = makeWeakPtr(list.get());
        return list;
    }

    void removeEmptyChildNodeList(EmptyNodeList* list)
    {
        ASSERT(m_emptyChildNodeList == list);
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(list->ownerNode()))
            return;
        m_emptyChildNodeList = nullptr;
    }

    struct NodeListCacheMapEntryHash {
        static unsigned hash(const std::pair<unsigned char, AtomicString>& entry)
        {
            return DefaultHash<AtomicString>::Hash::hash(entry.second) + entry.first;
        }
        static bool equal(const std::pair<unsigned char, AtomicString>& a, const std::pair<unsigned char, AtomicString>& b) { return a.first == b.first && DefaultHash<AtomicString>::Hash::equal(a.second, b.second); }
		static const bool safeToCompareToEmptyOrDeleted = DefaultHash<AtomicString>::Hash::safeToCompareToEmptyOrDeleted;
    };

    typedef HashMap<std::pair<unsigned char, AtomicString>, LiveNodeList*, NodeListCacheMapEntryHash> NodeListAtomicNameCacheMap;
    typedef HashMap<std::pair<unsigned char, AtomicString>, HTMLCollection*, NodeListCacheMapEntryHash> CollectionCacheMap;
    typedef HashMap<QualifiedName, TagCollectionNS*> TagCollectionNSCache;

    template<typename T, typename ContainerType>
    PassRefPtr<T> addCacheWithAtomicName(ContainerType* container, const AtomicString& name)
    {
        NodeListAtomicNameCacheMap::AddResult result = m_atomicNameCaches.add(namedNodeListKey<T>(name), reinterpret_cast<LiveNodeList*>(0));
        if (!result.isNewEntry)
            return static_cast<T*>(result.iterator->second);

        RefPtr<T> list = T::create(container, name);
        result.iterator->second = list.get();
        return list.release();
    }

    PassRefPtr<TagCollectionNS> addCachedTagCollectionNS(ContainerNode& node, const AtomicString& namespaceURI, const AtomicString& localName)
    {
        QualifiedName name(nullAtom, localName, namespaceURI);
        TagCollectionNSCache::AddResult result = m_tagCollectionNSCache.add(name, reinterpret_cast<TagCollectionNS*>(0));
        if (!result.isNewEntry)
            return result.iterator->second;

        RefPtr<TagCollectionNS> list = TagCollectionNS::create(&node, namespaceURI, localName);
        result.iterator->second = list.get();
        return list.release();
    }

    template<typename T, typename ContainerType>
    PassRefPtr<T> addCachedCollection(ContainerType& container, CollectionType collectionType, const AtomicString& name)
    {
        CollectionCacheMap::AddResult result = m_cachedCollections.add(namedCollectionKey(collectionType, name), reinterpret_cast<HTMLCollection*>(0));
        if (!result.isNewEntry)
            return static_cast<T*>(result.iterator->second);

        RefPtr<T> list = T::create(&container, collectionType, name);
        result.iterator->second = list.get();
        return list.release();
    }

    template<typename T, typename ContainerType>
    PassRefPtr<T> addCachedCollection(ContainerType& container, CollectionType collectionType)
    {
        CollectionCacheMap::AddResult result = m_cachedCollections.add(namedCollectionKey(collectionType, starAtom), reinterpret_cast<HTMLCollection*>(0));
        if (!result.isNewEntry)
            return static_cast<T*>(result.iterator->second);

        RefPtr<T> list = T::create(&container, collectionType);
        result.iterator->second = list.get();
        return list.release();
    }

    template<typename T>
    T* cachedCollection(CollectionType collectionType)
    {
        return static_cast<T*>(m_cachedCollections.get(namedCollectionKey(collectionType, starAtom)));
    }

    template <class NodeListType>
    void removeCacheWithAtomicName(NodeListType* list, const AtomicString& name = starAtom)
    {
        ASSERT(list == m_atomicNameCaches.get(namedNodeListKey<NodeListType>(name)));
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(list->ownerNode()))
            return;
        m_atomicNameCaches.remove(namedNodeListKey<NodeListType>(name));
    }

    void removeCachedTagCollectionNS(HTMLCollection& collection, const AtomicString& namespaceURI, const AtomicString& localName)
    {
        QualifiedName name(nullAtom, localName, namespaceURI);
        ASSERT(&collection == m_tagCollectionNSCache.get(name));
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(collection.ownerNode()))
            return;
        m_tagCollectionNSCache.remove(name);
    }

    void removeCachedCollection(HTMLCollection* collection, const AtomicString& name = starAtom)
    {
        ASSERT(collection == m_cachedCollections.get(namedCollectionKey(collection->type(), name)));
        if (deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(collection->ownerNode()))
            return;
        m_cachedCollections.remove(namedCollectionKey(collection->type(), name));
    }

    static ALWAYS_INLINE PassOwnPtr<NodeListsNodeData> create()
    {
        return adoptPtr(new NodeListsNodeData);
    }

    void invalidateCaches();
    void invalidateCachesForAttribute(const QualifiedName& attrName);

    bool isEmpty() const
	{
		return m_atomicNameCaches.isEmpty() && m_cachedCollections.isEmpty() && m_tagCollectionNSCache.isEmpty();
	}

    void adoptTreeScope()
    {
        invalidateCaches();
    }

    void adoptDocument(Document& oldDocument, Document& newDocument)
    {
        if (&oldDocument == &newDocument) {
            invalidateCaches();
            return;
        }

		for (auto it = m_atomicNameCaches.begin(), end = m_atomicNameCaches.end(); it != end; ++it)
			it->second->invalidateCacheForDocument(oldDocument);

		for (auto it = m_tagCollectionNSCache.begin(), end = m_tagCollectionNSCache.end(); it != end; ++it) {
			TagCollectionNS& list = *it->second;
            ASSERT(!list.isRootedAtDocument());
            list.invalidateCacheForDocument(oldDocument);
		}

		for (auto it = m_cachedCollections.begin(), end = m_cachedCollections.end(); it != end; ++it)
			it->second->invalidateCacheForDocument(oldDocument);
    }

private:
    NodeListsNodeData() = default;

    std::pair<unsigned char, AtomicString> namedCollectionKey(CollectionType type, const AtomicString& name)
    {
        return std::pair<unsigned char, AtomicString>(type, name);
    }

    template <class NodeListType>
    std::pair<unsigned char, AtomicString> namedNodeListKey(const AtomicString& name)
    {
        return std::pair<unsigned char, AtomicString>(NodeListTypeIdentifier<NodeListType>::value(), name);
    }

    bool deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(Node&);

    // These two are currently mutually exclusive and could be unioned. Not very important as this class is large anyway.
    WeakPtr<ChildNodeList> m_childNodeList;
    WeakPtr<EmptyNodeList> m_emptyChildNodeList;

    NodeListAtomicNameCacheMap m_atomicNameCaches;
    TagCollectionNSCache m_tagCollectionNSCache;
    CollectionCacheMap m_cachedCollections;
};

#if ENABLE(MUTATION_OBSERVERS)
class NodeMutationObserverData {
    WTF_MAKE_NONCOPYABLE(NodeMutationObserverData); WTF_MAKE_FAST_ALLOCATED;
public:
    Vector<OwnPtr<MutationObserverRegistration> > registry;
    WeakHashSet<MutationObserverRegistration> transientRegistry;

    static PassOwnPtr<NodeMutationObserverData> create() { return adoptPtr(new NodeMutationObserverData); }

private:
    NodeMutationObserverData() { }
};
#endif

#if ENABLE(MICRODATA)
class NodeMicroDataTokenLists {
    WTF_MAKE_NONCOPYABLE(NodeMicroDataTokenLists); WTF_MAKE_FAST_ALLOCATED;
public:
    static PassOwnPtr<NodeMicroDataTokenLists> create() { return adoptPtr(new NodeMicroDataTokenLists); }

    DOMTokenList* itemProp() const
    {
        if (!m_itemProp)
            m_itemProp = DOMTokenList::create();
        return m_itemProp.get();
    }

    DOMTokenList* itemRef() const
    {
        if (!m_itemRef)
            m_itemRef = DOMTokenList::create();
        return m_itemRef.get();
    }

    DOMTokenList* itemType() const
    {
        if (!m_itemType)
            m_itemType = DOMTokenList::create();
        return m_itemType.get();
    }

private:
    NodeMicroDataTokenLists() { }

    mutable RefPtr<DOMTokenList> m_itemProp;
    mutable RefPtr<DOMTokenList> m_itemRef;
    mutable RefPtr<DOMTokenList> m_itemType;
};
#endif

class NodeRareData {
    WTF_MAKE_NONCOPYABLE(NodeRareData); WTF_MAKE_FAST_ALLOCATED;

public:
#if defined(DUMP_NODE_STATISTICS) && DUMP_NODE_STATISTICS
    enum class UseType : uint32_t {
        NodeList = 1 << 0,
        MutationObserver = 1 << 1,
        TabIndex = 1 << 2,
        MinimumSize = 1 << 3,
        ScrollingPosition = 1 << 4,
        ComputedStyle = 1 << 5,
        Dataset = 1 << 6,
        ClassList = 1 << 7,
        ShadowRoot = 1 << 8,
        CustomElementQueue = 1 << 9,
        AttributeMap = 1 << 10,
        InteractionObserver = 1 << 11,
        ResizeObserver = 1 << 12,
        Animations = 1 << 13,
        PseudoElements = 1 << 14,
        StyleMap = 1 << 15,
        PartList = 1 << 16,
        PartNames = 1 << 17,
        Nonce = 1 << 18,
    };
#endif

    enum class Type { Element, Node };

    explicit NodeRareData(Type type = Type::Node)
        : m_unusualTabIndex(0)
    	, m_childIndex(0)
        , m_isElementRareData(type == Type::Element)
    {
    }

    bool isElementRareData() const { return m_isElementRareData; }

    ALWAYS_INLINE void clearNodeLists() { m_nodeLists.clear(); }
    ALWAYS_INLINE NodeListsNodeData* nodeLists() const { return m_nodeLists.get(); }
    NodeListsNodeData* ensureNodeLists()
    {
        if (!m_nodeLists)
        	m_nodeLists = NodeListsNodeData::create();
        return m_nodeLists.get();
    }

#if ENABLE(MUTATION_OBSERVERS)
    NodeMutationObserverData* mutationObserverDataIfExists() { return m_mutationObserverData.get(); }
    NodeMutationObserverData& mutationObserverData()
    {
        if (!m_mutationObserverData)
            m_mutationObserverData = NodeMutationObserverData::create();
        return *m_mutationObserverData;
    }
#endif

#if ENABLE(MICRODATA)
    NodeMicroDataTokenLists* ensureMicroDataTokenLists() const
    {
        if (!m_microDataTokenLists)
            m_microDataTokenLists = NodeMicroDataTokenLists::create();
        return m_microDataTokenLists.get();
    }
#endif

#if defined(DUMP_NODE_STATISTICS) && DUMP_NODE_STATISTICS
    OptionSet<UseType> useTypes() const
    {
        OptionSet<UseType> result;
        if (m_nodeLists)
            result.add(UseType::NodeList);
        if (m_mutationObserverData)
            result.add(UseType::MutationObserver);
        return result;
    }
#endif

protected:
    // Used by ElementRareData. Defined here for better packing in 64-bit.
    int m_unusualTabIndex;
    unsigned short m_childIndex;

private:
    bool m_isElementRareData;

    OwnPtr<NodeListsNodeData> m_nodeLists;

#if ENABLE(MUTATION_OBSERVERS)
    OwnPtr<NodeMutationObserverData> m_mutationObserverData;
#endif

#if ENABLE(MICRODATA)
    mutable OwnPtr<NodeMicroDataTokenLists> m_microDataTokenLists;
#endif
};

inline bool NodeListsNodeData::deleteThisAndUpdateNodeRareDataIfAboutToRemoveLastList(Node& ownerNode)
{
    ASSERT(ownerNode.nodeLists() == this);
    if ((m_childNodeList ? 1 : 0) + (m_emptyChildNodeList ? 1 : 0) + m_atomicNameCaches.size()
        + m_tagCollectionNSCache.size() + m_cachedCollections.size() != 1)
        return false;
    ownerNode.clearNodeLists();
    return true;
}

inline NodeRareData& Node::ensureRareData()
{
    if (!hasRareData())
        materializeRareData();
    return *rareData();
}

// Ensure the 10 bits reserved for the m_connectedFrameCount cannot overflow
COMPILE_ASSERT(Page::maxNumberOfFrames < 1024, Frame_limit_should_fit_in_rare_data_count);

} // namespace WebCore
