/*
 * Copyright (C) 2005, 2006, 2009 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include <wtf/HashTraits.h>
#include <wtf/Hasher.h>
#include <wtf/NeverDestroyed.h>
#include <wtf/RefCounted.h>
#include <wtf/text/AtomicString.h>

namespace WebCore {

struct QualifiedNameComponents {
	AtomicStringImpl* m_prefix;
	AtomicStringImpl* m_localName;
	AtomicStringImpl* m_namespace;
};

inline void add(Hasher& hasher, const QualifiedNameComponents& components)
{
    add(hasher, components.m_prefix, components.m_localName, components.m_namespace);
}

class QualifiedName {
    WTF_MAKE_FAST_ALLOCATED;
public:
    class QualifiedNameImpl : public RefCounted<QualifiedNameImpl> {
    public:
        static ALWAYS_INLINE Ref<QualifiedNameImpl> create(const AtomicString& prefix, const AtomicString& localName, const AtomicString& namespaceURI)
        {
            return adoptRef(*new QualifiedNameImpl(prefix, localName, namespaceURI));
        }

        ~QualifiedNameImpl();

        unsigned computeHash() const;

        mutable unsigned m_existingHash;
        const AtomicString m_prefix;
        const AtomicString m_localName;
        const AtomicString m_namespace;
        mutable AtomicString m_localNameUpper;

    private:
        QualifiedNameImpl(const AtomicString& prefix, const AtomicString& localName, const AtomicString& namespaceURI)
            : m_existingHash(0)
        	, m_prefix(prefix)
            , m_localName(localName)
            , m_namespace(namespaceURI)
        {
            ASSERT(!namespaceURI.isEmpty() || namespaceURI.isNull());
        }        
    };

    QualifiedName(const AtomicString& prefix, const AtomicString& localName, const AtomicString& namespaceURI);
    QualifiedName(QualifiedNameImpl& impl) : m_impl(impl) { }
    explicit QualifiedName(WTF::HashTableDeletedValueType) : m_impl(WTF::HashTableDeletedValue) { }
    bool isHashTableDeletedValue() const { return m_impl.isHashTableDeletedValue(); }

#ifdef QNAME_DEFAULT_CONSTRUCTOR
    QualifiedName() : m_impl(0) { }
#endif

    ALWAYS_INLINE bool operator==(const QualifiedName& other) const { return m_impl == other.m_impl; }
    ALWAYS_INLINE bool operator!=(const QualifiedName& other) const { return !(*this == other); }

    ALWAYS_INLINE bool matches(const QualifiedName& other) const { return m_impl == other.m_impl || (localName() == other.localName() && namespaceURI() == other.namespaceURI()); }

    ALWAYS_INLINE bool hasPrefix() const { return m_impl->m_prefix != nullAtom; }
    void setPrefix(const AtomicString& prefix) { *this = QualifiedName(prefix, localName(), namespaceURI()); }

    ALWAYS_INLINE const AtomicString& prefix() const { return m_impl->m_prefix; }
    ALWAYS_INLINE const AtomicString& localName() const { return m_impl->m_localName; }
    ALWAYS_INLINE const AtomicString& namespaceURI() const { return m_impl->m_namespace; }

    // Uppercased localName, cached for efficiency
    ALWAYS_INLINE const AtomicString& localNameUpper() const {
        if (LIKELY(m_impl->m_localNameUpper))
        	return m_impl->m_localNameUpper;
        return localNameUpperSlow();
    }

    String toString() const;

    QualifiedNameImpl* impl() const { return m_impl.get(); }
    
    // Init routine for globals
    static void init();
    
private:
    const AtomicString& localNameUpperSlow() const;
    String toStringSlow() const;

    static ALWAYS_INLINE QualifiedNameImpl* hashTableDeletedValue() { return RefPtr<QualifiedNameImpl>::hashTableDeletedValue(); }

    RefPtr<QualifiedNameImpl> m_impl;
};

inline void add(Hasher& hasher, const QualifiedName::QualifiedNameImpl& impl)
{
    add(hasher, impl.m_prefix, impl.m_localName, impl.m_namespace);
}

inline void add(Hasher& hasher, const QualifiedName& name)
{
    add(hasher, *name.impl());
}

extern LazyNeverDestroyed<const QualifiedName> anyName;
inline const QualifiedName& anyQName() { return anyName; }

extern LazyNeverDestroyed<const QualifiedName> nullName;
inline const QualifiedName& nullQName() { return nullName; }

ALWAYS_INLINE bool operator==(const AtomicString& a, const QualifiedName& q) { return a == q.localName(); }
ALWAYS_INLINE bool operator!=(const AtomicString& a, const QualifiedName& q) { return a != q.localName(); }
ALWAYS_INLINE bool operator==(const QualifiedName& q, const AtomicString& a) { return a == q.localName(); }
ALWAYS_INLINE bool operator!=(const QualifiedName& q, const AtomicString& a) { return a != q.localName(); }

struct QualifiedNameHash {
    static ALWAYS_INLINE unsigned hash(const QualifiedName& name) { return hash(name.impl()); }

    static ALWAYS_INLINE unsigned hash(const QualifiedName::QualifiedNameImpl* name)
    {
        if (UNLIKELY(!name->m_existingHash))
            name->m_existingHash = name->computeHash();
        return name->m_existingHash;
    }

    static ALWAYS_INLINE bool equal(const QualifiedName& a, const QualifiedName& b) { return a == b; }
    static ALWAYS_INLINE bool equal(const QualifiedName::QualifiedNameImpl* a, const QualifiedName::QualifiedNameImpl* b) { return a == b; }

    static const bool safeToCompareToEmptyOrDeleted = false;
};

void createQualifiedName(void* targetAddress, const char* name, unsigned nameLength);
void createQualifiedName(void* targetAddress, const char* name, unsigned nameLength, const AtomicString& nameNamespace);

inline String QualifiedName::toString() const
{
    if (!hasPrefix())
        return localName();

    return toStringSlow();
}

}

namespace WTF {
    
    template<typename T> struct DefaultHash;

    template<> struct DefaultHash<WebCore::QualifiedName> {
        typedef WebCore::QualifiedNameHash Hash;
    };
    
    template<> struct HashTraits<WebCore::QualifiedName> : SimpleClassHashTraits<WebCore::QualifiedName> {
        static const bool emptyValueIsZero = false;
        static ALWAYS_INLINE WebCore::QualifiedName emptyValue() { return WebCore::nullQName(); }
    };
}
