/*
 * (C) 1999 Lars Knoll (knoll@kde.org)
 * (C) 2000 Gunnstein Lye (gunnstein@netcom.no)
 * (C) 2000 Frederik Holljen (frederik.holljen@hig.no)
 * (C) 2001 Peter Kelly (pmk@post.com)
 * Copyright (C) 2004-2020 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "ExceptionCodePlaceholder.h"
#include "ParserContentPolicy.h"
#include "Node.h"
#include "RangeBoundaryPoint.h"

namespace JSC {
class SlotVisitor;
}

namespace WebCore {

class ContainerNode;
class DOMRect;
class DOMRectList;
class DOMWindow;
class Document;
class DocumentFragment;
class NodeWithIndex;
class Text;

struct SimpleRange;

class Range : public RefCounted<Range>, public CanMakeWeakPtr<Range> {
public:
    static Ref<Range> create(Document&);
    static RefPtr<Range> create(Document*);
    ~Range();

    Node& startContainer() const { return m_start.container(); }
    unsigned startOffset() const { return m_start.offset(); }
    Node& endContainer() const { ASSERT(m_end.container()); return m_end.container(); }
    unsigned endOffset() const { return m_end.offset(); }
    bool collapsed() const { return m_start == m_end; }
    WEBCORE_EXPORT Node* commonAncestorContainer() const;

    void setStart(Node& container, unsigned offset, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void setEnd(Node& container, unsigned offset, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void setStartBefore(Node&, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void setStartAfter(Node&, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void setEndBefore(Node&, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void setEndAfter(Node&, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void collapse(bool toStart);
    void selectNode(Node&, ExceptionCode& = ASSERT_NO_EXCEPTION);
    void selectNodeContents(Node&, ExceptionCode& = ASSERT_NO_EXCEPTION);

    enum CompareHow : unsigned short { START_TO_START, START_TO_END, END_TO_END, END_TO_START };
    short compareBoundaryPoints(unsigned short compareHow, const Range& sourceRange, ExceptionCode&) const;

    void deleteContents(ExceptionCode& = ASSERT_NO_EXCEPTION);
    PassRefPtr<DocumentFragment> extractContents(ExceptionCode&);
    PassRefPtr<DocumentFragment> cloneContents(ExceptionCode&);
    void insertNode(Node&, ExceptionCode&);
    void surroundContents(Node&, ExceptionCode& = ASSERT_NO_EXCEPTION);

    Ref<Range> cloneRange() const;
    static void detach() { }

    bool isPointInRange(Node&, unsigned offset, ExceptionCode&);
    short comparePoint(Node&, unsigned offset, ExceptionCode&) const;
    bool intersectsNode(Node&) const;

    String toString() const;

    PassRefPtr<DOMRectList> getClientRects() const;
    PassRefPtr<DOMRect> getBoundingClientRect() const;

    PassRefPtr<DocumentFragment> createContextualFragment(const String& fragment, ExceptionCode&);

    // Expand range to a unit (word or sentence or block or document) boundary.
    // Please refer to https://bugs.webkit.org/show_bug.cgi?id=27632 comment #5 for details.
    WEBCORE_EXPORT void expand(const String&, ExceptionCode&);

    enum CompareResults { NODE_BEFORE, NODE_AFTER, NODE_BEFORE_AND_AFTER, NODE_INSIDE };
    CompareResults compareNode(Node&, ExceptionCode&) const;

    void nodeChildrenChanged(ContainerNode&);
    void nodeChildrenWillBeRemoved(ContainerNode&);
    void nodeWillBeRemoved(Node&);
    bool parentlessNodeMovedToNewDocumentAffectsRange(Node&);
    void updateRangeForParentlessNodeMovedToNewDocument(Node&);

    void textInserted(Node&, unsigned offset, unsigned length);
    void textRemoved(Node&, unsigned offset, unsigned length);
    void textNodesMerged(NodeWithIndex& oldNode, unsigned offset);
    void textNodeSplit(Text& oldNode);

    void didAssociateWithSelection() { m_isAssociatedWithSelection = true; }
    void didDisassociateFromSelection() { m_isAssociatedWithSelection = false; }
    void updateFromSelection(const SimpleRange&);

    // For use by garbage collection. Returns nullptr for ranges not assocated with selection.
    DOMWindow* window() const;

    static Node* checkNodeOffsetPair(Node&, unsigned offset, ExceptionCode&);

#ifndef NDEBUG
    void formatForDebugger(char* buffer, unsigned length) const;
#endif

    void visitNodesConcurrently(JSC::SlotVisitor&) const;

    enum ActionType : uint8_t { DELETE_CONTENTS, EXTRACT_CONTENTS, CLONE_CONTENTS };

private:
    explicit Range(Document&);

    void updateDocument();
    void updateAssociatedSelection();
    PassRefPtr<DocumentFragment> processContents(ActionType, ExceptionCode&);

    Ref<Document> m_ownerDocument;
    RangeBoundaryPoint m_start;
    RangeBoundaryPoint m_end;
    bool m_isAssociatedWithSelection { false };
};

SimpleRange makeSimpleRange(const Range&);
SimpleRange makeSimpleRange(const Ref<Range>&);
Optional<SimpleRange> makeSimpleRange(const Range*);
Optional<SimpleRange> makeSimpleRange(const RefPtr<Range>&);

WEBCORE_EXPORT Ref<Range> createLiveRange(const SimpleRange&);
WEBCORE_EXPORT RefPtr<Range> createLiveRange(const Optional<SimpleRange>&);

} // namespace

#ifndef NDEBUG
// Outside the WebCore namespace for ease of invocation from gdb.
void showTree(const WebCore::Range*);
#endif
