/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ShadowRoot.h"

#include "CSSStyleSheet.h"
#include "ChildListMutationScope.h"
#include "ElementTraversal.h"
#include "HTMLParserIdioms.h"
#include "HTMLSlotElement.h"
#include "RenderElement.h"
#include "RuntimeEnabledFeatures.h"
#include "SlotAssignment.h"
#include "StyleResolver.h"
#include "StyleScope.h"
#include "Text.h"
#include "WebAnimation.h"
#include "markup.h"

namespace WebCore {

#if 0
struct SameSizeAsShadowRoot : public DocumentFragment, public TreeScope {
    unsigned countersAndFlags[1];
    void* styleScope;
    void* host;
#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    void* slotAssignment;
    Optional<HashMap<AtomString, AtomString>> partMappings;
#endif
};

COMPILE_ASSERT(sizeof(ShadowRoot) == sizeof(SameSizeAsShadowRoot), shadowroot_should_stay_small);
#endif

inline ShadowRoot::ShadowRoot(Document* document, ShadowRootMode type, DelegatesFocus delegatesFocus)
	: DocumentFragment(document, CreateShadowRoot)
	, TreeScope(*this, *document)
	, m_resetStyleInheritance(false)
	, m_hasBegunDeletingDetachedChildren(false)
	, m_delegatesFocus(delegatesFocus == DelegatesFocus::Yes)
	, m_containsFocusedElement(false)
	, m_type(type)
	, m_styleScope(adoptPtr(new Style::Scope(*this)))
{
    if (type == ShadowRootMode::UserAgent)
        setNodeFlag(HasBeenInUserAgentShadowTree);
}

Ref<ShadowRoot> ShadowRoot::create(Document* document, ShadowRootMode type, DelegatesFocus delegatesFocus)
{
    return adoptRef(*new ShadowRoot(document, type, delegatesFocus));
}

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)

inline ShadowRoot::ShadowRoot(Document* document, PassOwnPtr<SlotAssignment> slotAssignment)
    : DocumentFragment(document, CreateShadowRoot)
    , TreeScope(*this, *document)
	, m_resetStyleInheritance(false)
	, m_hasBegunDeletingDetachedChildren(false)
	, m_delegatesFocus(false)
	, m_containsFocusedElement(false)
    , m_type(ShadowRootMode::UserAgent)
	, m_styleScope(adoptPtr(new Style::Scope(*this)))
    , m_slotAssignment(slotAssignment)
{
    setNodeFlag(HasBeenInUserAgentShadowTree);
}

Ref<ShadowRoot> ShadowRoot::create(Document* document, PassOwnPtr<SlotAssignment> assignment)
{
    return adoptRef(*new ShadowRoot(document, assignment));
}

#endif

ShadowRoot::~ShadowRoot()
{
    if (inDocument())
        document().didRemoveInDocumentShadowRoot(*this);

    // We cannot let ContainerNode destructor call willBeDeletedFrom()
    // for this ShadowRoot instance because TreeScope destructor
    // clears Node::m_treeScope thus ContainerNode is no longer able
    // to access it Document reference after that.
    willBeDeletedFrom(&document());

    ASSERT(!m_hasBegunDeletingDetachedChildren);
    m_hasBegunDeletingDetachedChildren = true;

    // We must remove all of our children first before the TreeScope destructor
    // runs so we don't go through Node::setTreeScopeRecursively for each child with a
    // destructed tree scope in each descendant.
    removeDetachedChildren();
}

void ShadowRoot::childrenChanged(const ChildChange& childChange)
{
    DocumentFragment::childrenChanged(childChange);

    if (!m_host || m_type == ShadowRootMode::UserAgent)
        return; // Don't support first-child, nth-of-type, etc... in UA shadow roots as an optimization.

    // FIXME: Avoid always invalidating style just for first-child, etc... as done in Element::childrenChanged.
    switch (childChange.type) {
    case ElementInserted:
    case ElementRemoved:
        m_host->invalidateStyleForSubtreeInternal();
        break;
    case TextInserted:
    case TextRemoved:
    case TextChanged:
    case AllChildrenRemoved:
    case NonContentsChildRemoved:
    case NonContentsChildInserted:
    case AllChildrenReplaced:
        break;
    }
}

Node::InsertedIntoResult ShadowRoot::insertedIntoAncestor(const InsertionType& insertionType, ContainerNode& parentOfInsertedTree)
{
    DocumentFragment::insertedIntoAncestor(insertionType, parentOfInsertedTree);
    if (insertionType.connectedToDocument)
        document().didInsertInDocumentShadowRoot(*this);
    return InsertedIntoResult::Done;
}

void ShadowRoot::removedFromAncestor(const RemovalType& removalType, ContainerNode& parentOfRemovedTree)
{
    DocumentFragment::removedFromAncestor(removalType, parentOfRemovedTree);
    if (removalType.disconnectedFromDocument)
        document().didRemoveInDocumentShadowRoot(*this);
}

void ShadowRoot::moveShadowRootToNewParentScope(TreeScope& newScope, Document& newDocument)
{
    setParentTreeScope(newScope);
    moveShadowRootToNewDocument(newDocument);
}

void ShadowRoot::moveShadowRootToNewDocument(Document& newDocument)
{
    setDocumentScope(newDocument);
    ASSERT_WITH_SECURITY_IMPLICATION(!parentTreeScope() || &parentTreeScope()->documentScope() == &newDocument);

    // Style scopes are document specific.
    m_styleScope = adoptPtr(new Style::Scope(*this));
    ASSERT_WITH_SECURITY_IMPLICATION(&m_styleScope->document() == &newDocument);
}

String ShadowRoot::innerHTML() const
{
    return serializeFragment(*this, SerializedNodes::SubtreesOfChildren);
}

void ShadowRoot::setInnerHTML(const String& markup, ExceptionCode& ec)
{
    if (markup.isEmpty()) {
        ChildListMutationScope mutation(this);
        removeChildren();
        return;
    }

    RefPtr<DocumentFragment> fragment = createFragmentForInnerOuterHTML(markup, host(), AllowScriptingContent, ec);
    if (ec)
    	return;
    return replaceChildrenWithFragment(this, fragment.release(), ec);
}

bool ShadowRoot::childTypeAllowed(NodeType type) const
{
    switch (type) {
    case ELEMENT_NODE:
    case PROCESSING_INSTRUCTION_NODE:
    case COMMENT_NODE:
    case TEXT_NODE:
    case CDATA_SECTION_NODE:
        return true;
    default:
        return false;
    }
}

Ref<Node> ShadowRoot::cloneNodeInternal(Document&, CloningOperation)
{
    RELEASE_ASSERT_NOT_REACHED();
    return *static_cast<Node*>(nullptr); // ShadowRoots should never be cloned.
}

void ShadowRoot::removeAllEventListeners()
{
    DocumentFragment::removeAllEventListeners();
    for (Node* node = firstChild(); node; node = NodeTraversal::next(node))
        node->removeAllEventListeners();
}

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)

HTMLSlotElement* ShadowRoot::findAssignedSlot(const Node& node)
{
    ASSERT(node.parentNode() == host());
    if (!m_slotAssignment)
        return nullptr;
    return m_slotAssignment->findAssignedSlot(node);
}

void ShadowRoot::renameSlotElement(HTMLSlotElement& slot, const AtomicString& oldName, const AtomicString& newName)
{
    ASSERT(m_slotAssignment);
    return m_slotAssignment->renameSlotElement(slot, oldName, newName, *this);
}

void ShadowRoot::addSlotElementByName(const AtomicString& name, HTMLSlotElement& slot)
{
    ASSERT(&slot.rootNode() == this);
    if (!m_slotAssignment)
        m_slotAssignment = adoptPtr(new SlotAssignment());

    return m_slotAssignment->addSlotElementByName(name, slot, *this);
}

void ShadowRoot::removeSlotElementByName(const AtomicString& name, HTMLSlotElement& slot, ContainerNode& oldParentOfRemovedTree)
{
    ASSERT(m_slotAssignment);
    return m_slotAssignment->removeSlotElementByName(name, slot, &oldParentOfRemovedTree, *this);
}

void ShadowRoot::slotFallbackDidChange(HTMLSlotElement& slot)
{
    ASSERT(&slot.rootNode() == this);
    return m_slotAssignment->slotFallbackDidChange(slot, *this);
}

const Vector<WeakPtr<Node>>* ShadowRoot::assignedNodesForSlot(const HTMLSlotElement& slot)
{
    if (!m_slotAssignment)
        return nullptr;
    return m_slotAssignment->assignedNodesForSlot(slot, *this);
}

static Optional<std::pair<AtomicString, AtomicString>> parsePartMapping(StringView mappingString)
{
    const auto end = mappingString.length();

    auto skipWhitespace = [&] (size_t position) {
        while (position < end && isHTMLSpace(mappingString[position]))
            ++position;
        return position;
    };

    auto collectValue = [&] (size_t position) {
        while (position < end && (!isHTMLSpace(mappingString[position]) && mappingString[position] != ':'))
            ++position;
        return position;
    };

    size_t begin = 0;
    begin = skipWhitespace(begin);

    size_t firstPartEnd = collectValue(begin);
    if (firstPartEnd == begin)
        return Nullopt;

    auto firstPart = mappingString.substring(begin, firstPartEnd - begin).toAtomicString();

    begin = skipWhitespace(firstPartEnd);
    if (begin == end)
        return std::make_pair(firstPart, firstPart);

    if (mappingString[begin] != ':')
        return Nullopt;

    begin = skipWhitespace(begin + 1);

    size_t secondPartEnd = collectValue(begin);
    if (secondPartEnd == begin)
        return Nullopt;

    auto secondPart = mappingString.substring(begin, secondPartEnd - begin).toAtomicString();

    begin = skipWhitespace(secondPartEnd);
    if (begin != end)
        return Nullopt;

    return std::make_pair(firstPart, secondPart);
}

static void parsePartMappingsList(ShadowRoot::PartMappings& mappings, StringView mappingsListString)
{
    const auto end = mappingsListString.length();

    size_t begin = 0;
    while (begin < end) {
        size_t mappingEnd = begin;
        while (mappingEnd < end && mappingsListString[mappingEnd] != ',')
            ++mappingEnd;

        auto result = parsePartMapping(mappingsListString.substring(begin, mappingEnd - begin));
        if (result)
            mappings.add(result->first, Vector<AtomicString, 1>()).iterator->second.append(result->second);

        if (mappingEnd == end)
            break;

        begin = mappingEnd + 1;
    }
}

const ShadowRoot::PartMappings& ShadowRoot::partMappings() const
{
    if (!m_partMappings) {
        m_partMappings = PartMappings();

        AtomicString exportpartsValue = host()->attributeWithoutSynchronization(HTMLNames::exportpartsAttr);
        if (!exportpartsValue.isEmpty()/* && RuntimeEnabledFeatures::sharedFeatures().cssShadowPartsEnabled()*/)
            parsePartMappingsList(*m_partMappings, exportpartsValue);
    }

    return *m_partMappings;
}

void ShadowRoot::invalidatePartMappings()
{
    m_partMappings = Nullopt;
}

Vector<ShadowRoot*> assignedShadowRootsIfSlotted(const Node& node)
{
    Vector<ShadowRoot*> result;
    for (auto* slot = node.assignedSlot(); slot; slot = slot->assignedSlot()) {
        ASSERT(slot->containingShadowRoot());
        result.append(slot->containingShadowRoot());
    }
    return result;
}

#endif

Vector<RefPtr<WebAnimation>> ShadowRoot::getAnimations()
{
	struct MatchingAnimationsFunctionImpl FINAL : public Document::MatchingAnimationsFunction
	{
		MatchingAnimationsFunctionImpl(const ShadowRoot* thisShadowRoot) : thisShadowRoot(thisShadowRoot) { }
		virtual bool operator()(Element& target) const FINAL {
	        return target.containingShadowRoot() == thisShadowRoot;
        }
		const ShadowRoot* thisShadowRoot;
	};
	MatchingAnimationsFunctionImpl function(this);
    return document().matchingAnimations(function);
}

}
