/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ContainerNode.h"
#include "Document.h"
#include "DocumentFragment.h"
#include "Element.h"
#include "ElementInlines.h"
#include "ExceptionCode.h"
#include "ShadowRootMode.h"
#include "StyleScopeOrdinal.h"
#include "TreeScope.h"
#include <wtf/HashMap.h>
#include <wtf/Optional.h>
#include <wtf/Ref.h>

namespace WebCore {

class HTMLSlotElement;
class SlotAssignment;
class StyleResolver;
class WebAnimation;

class ShadowRoot FINAL : public DocumentFragment, public TreeScope {
	WTF_MAKE_FAST_ALLOCATED;
public:

    enum class DelegatesFocus : uint8_t { Yes, No };

    static Ref<ShadowRoot> create(Document*, ShadowRootMode, DelegatesFocus delegatesFocus = DelegatesFocus::No);

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    static Ref<ShadowRoot> create(Document*, PassOwnPtr<SlotAssignment>);
#endif

    virtual ~ShadowRoot();

    using TreeScope::rootNode;

    Style::Scope& styleScope();

    bool resetStyleInheritance() const;
    void setResetStyleInheritance(bool);

    bool delegatesFocus() const { return m_delegatesFocus; }
    bool containsFocusedElement() const { return m_containsFocusedElement; }
    void setContainsFocusedElement(bool flag) { m_containsFocusedElement = flag; }

    Element* host() const { return m_host.get(); }
    void setHost(WeakPtr<Element>&& host) { m_host = WTFMove(host); }

    String innerHTML() const;
    void setInnerHTML(const String&, ExceptionCode&);

    Element* activeElement() const;

    ShadowRootMode mode() const { return m_type; }
    bool shouldFireSlotchangeEvent() const { return m_type != ShadowRootMode::UserAgent && !m_hasBegunDeletingDetachedChildren; }

    virtual void removeAllEventListeners() OVERRIDE;

    using TreeScope::getElementById;

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    HTMLSlotElement* findAssignedSlot(const Node&);

    void renameSlotElement(HTMLSlotElement&, const AtomicString& oldName, const AtomicString& newName);
    void addSlotElementByName(const AtomicString&, HTMLSlotElement&);
    void removeSlotElementByName(const AtomicString&, HTMLSlotElement&, ContainerNode& oldParentOfRemovedTree);
    void slotFallbackDidChange(HTMLSlotElement&);
    void resolveSlotsBeforeNodeInsertionOrRemoval();
    void willRemoveAllChildren(ContainerNode&);
    void willRemoveAssignedNode(const Node&);

    void didRemoveAllChildrenOfShadowHost();
    void didChangeDefaultSlot();
    void hostChildElementDidChange(const Element&);
    void hostChildElementDidChangeSlotAttribute(Element&, const AtomicString& oldValue, const AtomicString& newValue);

    const Vector<WeakPtr<Node>>* assignedNodesForSlot(const HTMLSlotElement&);
#endif

    void moveShadowRootToNewParentScope(TreeScope&, Document&);
    void moveShadowRootToNewDocument(Document&);

    typedef HashMap<AtomicString, Vector<AtomicString, 1>> PartMappings;
    const PartMappings& partMappings() const;
    void invalidatePartMappings();

    Vector<RefPtr<WebAnimation>> getAnimations();

private:
    explicit ShadowRoot(Document*, ShadowRootMode, DelegatesFocus);

#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    ShadowRoot(Document*, PassOwnPtr<SlotAssignment>);
#endif

    virtual bool childTypeAllowed(NodeType) const OVERRIDE;

    virtual Ref<Node> cloneNodeInternal(Document&, CloningOperation) OVERRIDE;

    virtual Node::InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode&) OVERRIDE;
    virtual void removedFromAncestor(const RemovalType&, ContainerNode&) OVERRIDE;
    virtual void childrenChanged(const ChildChange&) OVERRIDE;

    bool m_resetStyleInheritance;
    bool m_hasBegunDeletingDetachedChildren;
    bool m_delegatesFocus;
    bool m_containsFocusedElement;
    ShadowRootMode m_type;

    WeakPtr<Element> m_host;

    OwnPtr<Style::Scope> m_styleScope;
#if ENABLE(SHADOW_DOM) || ENABLE(DETAILS_ELEMENT)
    OwnPtr<SlotAssignment> m_slotAssignment;
#endif
    mutable Optional<PartMappings> m_partMappings;
};

inline Element* ShadowRoot::activeElement() const
{
    return treeScope().focusedElementInScope();
}

inline bool isShadowRoot(const Node* node)
{
	return node && node->isShadowRoot();
}

inline bool isShadowRoot(const Node& node)
{
	return node.isShadowRoot();
}

inline ShadowRoot* toShadowRoot(Node* node)
{
    ASSERT(!node || node->isShadowRoot());
    return static_cast<ShadowRoot*>(node);
}

inline const ShadowRoot* toShadowRoot(const Node* node)
{
    ASSERT(!node || node->isShadowRoot());
    return static_cast<const ShadowRoot*>(node);
}

inline ShadowRoot& toShadowRoot(Node& node)
{
    ASSERT(node.isShadowRoot());
    return static_cast<ShadowRoot&>(node);
}

inline const ShadowRoot& toShadowRoot(const Node& node)
{
    ASSERT(node.isShadowRoot());
    return static_cast<const ShadowRoot&>(node);
}

inline ShadowRoot* Node::shadowRoot() const
{
    if (!isElementNode())
        return nullptr;
    return toElement(this)->shadowRoot();
}

inline bool Node::isUserAgentShadowRoot() const
{
    return isShadowRoot() && toShadowRoot(*this).mode() == ShadowRootMode::UserAgent;
}

inline ContainerNode* Node::parentOrShadowHostNode() const
{
    ASSERT(isMainThreadOrGCThread());
    if (isShadowRoot())
        return toShadowRoot(this)->host();
    return parentNode();
}

inline bool hasShadowRootParent(const Node* node)
{
    return node->parentNode() && node->parentNode()->isShadowRoot();
}

inline ShadowRoot* Node::containingShadowRoot() const
{
    auto& root = treeScope().rootNode();
    return root.isShadowRoot() ? &toShadowRoot(root) : nullptr;
}

inline Style::Scope& ShadowRoot::styleScope()
{
    return *m_styleScope;
}

inline bool ShadowRoot::resetStyleInheritance() const
{
    return m_resetStyleInheritance;
}

inline void ShadowRoot::setResetStyleInheritance(bool value)
{
    // If this was ever changed after initialization, child styles would need to be invalidated here.
    m_resetStyleInheritance = value;
}

Vector<ShadowRoot*> assignedShadowRootsIfSlotted(const Node&);

} // namespace
