/*
 * Copyright (C) 2011 Google Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "TreeScope.h"

#include "Attr.h"
#include "ContainerNode.h"
#include "DOMWindow.h"
#include "Document.h"
#include "Element.h"
#include "ElementTraversal.h"
#include "FocusController.h"
#include "Frame.h"
#include "FrameView.h"
#include "HTMLAnchorElement.h"
#include "HTMLFrameOwnerElement.h"
#include "HTMLImageElement.h"
#include "HTMLMapElement.h"
#include "HTMLNames.h"
#include "HitTestResult.h"
#include "IdTargetObserverRegistry.h"
#include "NodeRareData.h"
#include "Page.h"
#include "PseudoElement.h"
#include "RadioButtonGroups.h"
#include "RenderView.h"
#include "Settings.h"
#include "ShadowRoot.h"
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/CString.h>

namespace WebCore {

#if 0
struct SameSizeAsTreeScope {
    void* pointers[10];
};

COMPILE_ASSERT(sizeof(TreeScope) == sizeof(SameSizeAsTreeScope), treescope_should_stay_small);
#endif

using namespace HTMLNames;

TreeScope::TreeScope(ShadowRoot& shadowRoot, Document& document)
    : m_rootNode(shadowRoot)
    , m_documentScope(&document)
    , m_parentTreeScope(&document)
	, m_idTargetObserverRegistry(IdTargetObserverRegistry::create())
{
    shadowRoot.setTreeScope(*this);
}

TreeScope::TreeScope(Document& document)
    : m_rootNode(document)
    , m_documentScope(&document)
    , m_parentTreeScope(0)
	, m_idTargetObserverRegistry(IdTargetObserverRegistry::create())
{
    document.setTreeScope(*this);
}

TreeScope::~TreeScope() = default;

void TreeScope::destroyTreeScopeData()
{
    m_elementsById = nullptr;
    m_elementsByName = nullptr;
    m_imageMapsByName = nullptr;
    m_imagesByUsemap = nullptr;
}

void TreeScope::setParentTreeScope(TreeScope& newParentScope)
{
    // A document node cannot be re-parented.
    ASSERT(!rootNode().isDocumentNode());
    // Every scope other than document needs a parent scope.

    m_parentTreeScope = &newParentScope;
    setDocumentScope(newParentScope.documentScope());
}

Element* TreeScope::getElementById(const AtomicString& elementId) const
{
    if (elementId.isNull())
        return nullptr;
    if (!m_elementsById)
        return nullptr;
    return m_elementsById->getElementById(elementId.impl(), this);
}

Element* TreeScope::getElementById(const String& elementId) const
{
    if (!m_elementsById)
        return nullptr;

    if (RefPtr<AtomicStringImpl> atomicElementId = AtomicString::find(elementId.impl()))
        return m_elementsById->getElementById(atomicElementId.get(), this);

    return nullptr;
}

const Vector<Element*>* TreeScope::getAllElementsById(const AtomicString& elementId) const
{
    if (elementId.isEmpty())
        return nullptr;
    if (!m_elementsById)
        return nullptr;
    return m_elementsById->getAllElementsById(elementId.impl(), this);
}

void TreeScope::addElementById(const AtomicString& elementId, Element* element, bool notifyObservers)
{
    if (!m_elementsById)
        m_elementsById = adoptPtr(new DocumentOrderedMap);
    m_elementsById->add(elementId.impl(), element, *this);
    if (notifyObservers)
    	m_idTargetObserverRegistry->notifyObservers(elementId);
}

void TreeScope::removeElementById(const AtomicString& elementId, Element* element, bool notifyObservers)
{
    if (!m_elementsById)
        return;
    m_elementsById->remove(elementId.impl(), element);
    if (notifyObservers)
    	m_idTargetObserverRegistry->notifyObservers(elementId);
}

Element* TreeScope::getElementByName(const AtomicString& name) const
{
    if (name.isEmpty())
        return nullptr;
    if (!m_elementsByName)
        return nullptr;
    return m_elementsByName->getElementByName(name.impl(), this);
}

void TreeScope::addElementByName(const AtomicString& name, Element* element)
{
    if (!m_elementsByName)
        m_elementsByName = adoptPtr(new DocumentOrderedMap);
    m_elementsByName->add(name.impl(), element, *this);
}

void TreeScope::removeElementByName(const AtomicString& name, Element* element)
{
    if (!m_elementsByName)
        return;
    m_elementsByName->remove(name.impl(), element);
}

Node& TreeScope::retargetToScope(Node& node) const
{
    auto& scope = node.treeScope();
    if (LIKELY(this == &scope || !node.isInShadowTree()))
        return node;
    ASSERT(is<ShadowRoot>(scope.rootNode()));

    Vector<TreeScope*, 8> nodeTreeScopes;
    for (auto* currentScope = &scope; currentScope; currentScope = currentScope->parentTreeScope())
        nodeTreeScopes.append(currentScope);
    ASSERT(nodeTreeScopes.size() >= 2);

    Vector<const TreeScope*, 8> ancestorScopes;
    for (auto* currentScope = this; currentScope; currentScope = currentScope->parentTreeScope())
        ancestorScopes.append(currentScope);

    size_t i = nodeTreeScopes.size();
    size_t j = ancestorScopes.size();
    while (i > 0 && j > 0 && nodeTreeScopes[i - 1] == ancestorScopes[j - 1]) {
        --i;
        --j;
    }

    bool nodeIsInOuterTreeScope = !i;
    if (nodeIsInOuterTreeScope)
        return node;

    ShadowRoot& shadowRootInLowestCommonTreeScope = toShadowRoot(nodeTreeScopes[i - 1]->rootNode());
    return *shadowRootInLowestCommonTreeScope.host();
}

Node* TreeScope::ancestorNodeInThisScope(Node* node) const
{
	for(; node; node = node->shadowHost()) {
        if (&node->treeScope() == this)
            return node;
        if (!node->isInShadowTree())
            return nullptr;
    }

    return nullptr;
}

Element* TreeScope::ancestorElementInThisScope(Element* element) const
{
    for (; element; element = element->shadowHost()) {
        if (&element->treeScope() == this)
            return element;
        if (!element->isInShadowTree())
            return nullptr;
    }
    return nullptr;
}

void TreeScope::addImageMap(HTMLMapElement* imageMap)
{
    AtomicStringImpl* name = imageMap->getName().impl();
    if (!name)
        return;
    if (!m_imageMapsByName)
        m_imageMapsByName = adoptPtr(new DocumentOrderedMap);
    m_imageMapsByName->add(name, imageMap, *this);
}

void TreeScope::removeImageMap(HTMLMapElement* imageMap)
{
    if (!m_imageMapsByName)
        return;
    AtomicStringImpl* name = imageMap->getName().impl();
    if (!name)
        return;
    m_imageMapsByName->remove(name, imageMap);
}

HTMLMapElement* TreeScope::getImageMap(const AtomicString& name) const
{
    if (!m_imageMapsByName || !name.impl())
        return nullptr;
    return m_imageMapsByName->getElementByMapName(name.impl(), this);
}

void TreeScope::addImageElementByUsemap(const AtomicStringImpl& name, HTMLImageElement& element)
{
    if (!m_imagesByUsemap)
        m_imagesByUsemap = adoptPtr(new DocumentOrderedMap());
    m_imagesByUsemap->add(&name, &element, *this);
}

void TreeScope::removeImageElementByUsemap(const AtomicStringImpl& name, HTMLImageElement& element)
{
    if (!m_imagesByUsemap)
        return;
    m_imagesByUsemap->remove(&name, &element);
}

HTMLImageElement* TreeScope::imageElementByUsemap(const AtomicStringImpl& name) const
{
    if (!m_imagesByUsemap)
        return nullptr;
    return m_imagesByUsemap->getElementByUsemap(name, *this);
}

#if 0
void TreeScope::addLabel(const AtomicString& forAttributeValue, HTMLLabelElement* element)
{
    ASSERT(m_labelsByForAttribute);
    m_labelsByForAttribute->add(forAttributeValue.impl(), element);
}

void TreeScope::removeLabel(const AtomicString& forAttributeValue, HTMLLabelElement* element)
{
    ASSERT(m_labelsByForAttribute);
    m_labelsByForAttribute->remove(forAttributeValue.impl(), element);
}

HTMLLabelElement* TreeScope::labelElementForId(const AtomicString& forAttributeValue)
{
    if (forAttributeValue.isEmpty())
        return 0;

    if (!m_labelsByForAttribute) {
        // Populate the map on first access.
        m_labelsByForAttribute = adoptPtr(new DocumentOrderedMap);
        for (HTMLLabelElement* label = Traversal<HTMLLabelElement>::firstWithin(&rootNode()); label; label = Traversal<HTMLLabelElement>::next(label)) {
            const AtomicString& forValue = label->attributeWithoutSynchronization(forAttr);
            if (!forValue.isEmpty())
                addLabel(forValue, label);
        }
    }

    return toHTMLLabelElement(m_labelsByForAttribute->getElementByLabelForAttribute(forAttributeValue.impl(), this));
}
#endif

static Optional<LayoutPoint> absolutePointIfNotClipped(Document& document, const LayoutPoint& clientPoint)
{
    if (!document.frame() || !document.view())
        return Nullopt;

    const auto& settings = document.frame()->settings();
	if (settings.visualViewportEnabled() && settings.clientCoordinatesRelativeToLayoutViewport()) {
        document.updateLayout();
        if (!document.view() || !document.hasLivingRenderTree())
            return Nullopt;
        auto* view = document.view();
        FloatPoint layoutViewportPoint = view->clientToLayoutViewportPoint(clientPoint);
        FloatRect layoutViewportBounds({ }, view->layoutViewportRect().size());
        if (!layoutViewportBounds.contains(layoutViewportPoint))
            return Nullopt;
        return LayoutPoint(view->layoutViewportToAbsolutePoint(layoutViewportPoint));
    }

    auto* frame = document.frame();
    auto* view = document.view();
    float scaleFactor = frame->pageZoomFactor() * frame->frameScaleFactor();

    LayoutPoint absolutePoint = clientPoint;
    absolutePoint.scale(scaleFactor, scaleFactor);
    absolutePoint.moveBy(view->contentsScrollPosition());

    LayoutRect visibleRect;
#if PLATFORM(IOS)
    visibleRect = view->unobscuredContentRect();
#else
    visibleRect = view->visibleContentRect();
#endif
    if (visibleRect.contains(absolutePoint))
        return absolutePoint;
    return Nullopt;
}

Node* TreeScope::nodeFromPoint(const LayoutPoint& clientPoint, LayoutPoint* localPoint)
{
    auto absolutePoint = absolutePointIfNotClipped(documentScope(), clientPoint);
    if (!absolutePoint)
        return nullptr;

    HitTestResult result(absolutePoint.value());
    documentScope().hitTest(HitTestRequest(), result);
    if (localPoint)
        *localPoint = result.localPoint();
    return result.innerNode();
}

PassRefPtr<Element> TreeScope::elementFromPoint(double clientX, double clientY)
{
    Document& document = documentScope();
    if (!document.hasLivingRenderTree())
        return nullptr;

    Node* node = nodeFromPoint(LayoutPoint(clientX, clientY), nullptr);
    if (!node)
        return nullptr;

    node = &retargetToScope(*node);
    while (!isElement(*node)) {
        node = node->parentInComposedTree();
        if (!node)
            break;
        node = &retargetToScope(*node);
    }

    return toElement(node);
}

Vector<RefPtr<Element>> TreeScope::elementsFromPoint(double clientX, double clientY)
{
    Vector<RefPtr<Element>> elements;

    Document& document = documentScope();
    if (!document.hasLivingRenderTree())
        return elements;

    auto absolutePoint = absolutePointIfNotClipped(document, LayoutPoint(clientX, clientY));
    if (!absolutePoint)
        return elements;

    HitTestRequest request(HitTestRequest::ReadOnly
        | HitTestRequest::Active
        | HitTestRequest::DisallowUserAgentShadowContent
        | HitTestRequest::CollectMultipleElements
        | HitTestRequest::IncludeAllElementsUnderPoint);
    HitTestResult result(absolutePoint.value());
    documentScope().hitTest(request, result);

    Node* lastNode = nullptr;
    for (auto listBasedNode : result.listBasedTestResult()) {
        Node* node = listBasedNode.get();
        node = &retargetToScope(*node);
        while (!(*node).isElementNode()) {
            node = node->parentInComposedTree();
            if (!node)
                break;
            node = &retargetToScope(*node);
        }

        if (!node)
            continue;

        if ((*node).isPseudoElement())
            node = toPseudoElement(*node).hostElement();

        // Prune duplicate entries. A pseudo ::before content above its parent
        // node should only result in one entry.
        if (node == lastNode)
            continue;

        elements.append(toElement(node));
        lastNode = node;
    }

    if (m_rootNode.isDocumentNode()) {
        if (Element* rootElement = toDocument(m_rootNode).documentElement()) {
            if (elements.isEmpty() || elements.last() != rootElement)
                elements.append(rootElement);
        }
    }

    return elements;
}

Vector<RefPtr<Element>> TreeScope::elementsFromPoint(const FloatPoint& p)
{
    return elementsFromPoint(p.x(), p.y());
}

Element* TreeScope::findAnchor(const String& name)
{
    if (name.isEmpty())
        return nullptr;
    if (Element* element = getElementById(name))
        return element;
    for (HTMLAnchorElement* anchor = Traversal<HTMLAnchorElement>::firstWithin(&rootNode()); anchor; anchor = Traversal<HTMLAnchorElement>::next(anchor)) {
		if (rootNode().document().inQuirksMode()) {
			// Quirks mode, case insensitive comparison of names.
			if (equalIgnoringASCIICase(anchor->name(), name))
				return anchor;
		} else {
			// Strict mode, names need to match exactly.
			if (anchor->name() == name)
				return anchor;
		}
    }
    return nullptr;
}

static Element* focusedFrameOwnerElement(Frame* focusedFrame, Frame* currentFrame)
{
    for (; focusedFrame; focusedFrame = focusedFrame->tree()->parent()) {
        if (focusedFrame->tree()->parent() == currentFrame)
            return focusedFrame->ownerElement();
    }
    return nullptr;
}

Element* TreeScope::focusedElementInScope()
{
    Document& document = rootNode().document();
    Element* element = document.focusedElement();

    if (!element && document.page())
        element = focusedFrameOwnerElement(document.page()->focusController()->focusedFrame(), document.frame());

    return ancestorElementInThisScope(element);
}

static void listTreeScopes(Node* node, Vector<TreeScope*, 5>& treeScopes)
{
    while (true) {
        treeScopes.append(&node->treeScope());
        Element* ancestor = node->shadowHost();
        if (!ancestor)
            break;
        node = ancestor;
    }
}

TreeScope* commonTreeScope(Node* nodeA, Node* nodeB)
{
    if (!nodeA || !nodeB)
        return 0;

    if (&nodeA->treeScope() == &nodeB->treeScope())
        return &nodeA->treeScope();

    Vector<TreeScope*, 5> treeScopesA;
    listTreeScopes(nodeA, treeScopesA);

    Vector<TreeScope*, 5> treeScopesB;
    listTreeScopes(nodeB, treeScopesB);

    size_t indexA = treeScopesA.size();
    size_t indexB = treeScopesB.size();

    for (; indexA > 0 && indexB > 0 && treeScopesA[indexA - 1] == treeScopesB[indexB - 1]; --indexA, --indexB) { }

    return treeScopesA[indexA] == treeScopesB[indexB] ? treeScopesA[indexA] : 0;
}

RadioButtonGroups& TreeScope::radioButtonGroups()
{
    if (!m_radioButtonGroups)
        m_radioButtonGroups = adoptPtr(new RadioButtonGroups());
    return *m_radioButtonGroups;
}

struct SVGResourcesMap {
    WTF_MAKE_NONCOPYABLE(SVGResourcesMap); WTF_MAKE_FAST_ALLOCATED;
public:
    SVGResourcesMap() = default;

    HashMap<AtomicString, WeakHashSet<SVGElement>> pendingResources;
    HashMap<AtomicString, WeakHashSet<SVGElement>> pendingResourcesForRemoval;
    HashMap<AtomicString, RenderSVGResourceContainer*> resources;
};

SVGResourcesMap& TreeScope::svgResourcesMap() const
{
    if (!m_svgResourcesMap)
        const_cast<TreeScope&>(*this).m_svgResourcesMap = adoptPtr(new SVGResourcesMap());
    return *m_svgResourcesMap;
}

void TreeScope::addSVGResource(const AtomicString& id, RenderSVGResourceContainer& resource)
{
    if (id.isEmpty())
        return;

    // Replaces resource if already present, to handle potential id changes
    svgResourcesMap().resources.set(id, &resource);
}

void TreeScope::removeSVGResource(const AtomicString& id)
{
    if (id.isEmpty())
        return;

    svgResourcesMap().resources.remove(id);
}

RenderSVGResourceContainer* TreeScope::svgResourceById(const AtomicString& id) const
{
    if (id.isEmpty())
        return nullptr;

    return svgResourcesMap().resources.get(id);
}

void TreeScope::addPendingSVGResource(const AtomicString& id, SVGElement& element)
{
    if (id.isEmpty())
        return;

    auto result = svgResourcesMap().pendingResources.add(id, WeakHashSet<SVGElement>());
    result.iterator->second.add(element);

    element.setHasPendingResources();
}

bool TreeScope::isIdOfPendingSVGResource(const AtomicString& id) const
{
    if (id.isEmpty())
        return false;

    return svgResourcesMap().pendingResources.contains(id);
}

bool TreeScope::isElementWithPendingSVGResources(SVGElement& element) const
{
    // This algorithm takes time proportional to the number of pending resources and need not.
    // If performance becomes an issue we can keep a counted set of elements and answer the question efficiently.
    return WTF::anyOf(svgResourcesMap().pendingResources.values(), [&] (const WeakHashSet<SVGElement>& elements) {
        return elements.contains(element);
    });
}

bool TreeScope::isPendingSVGResource(SVGElement& element, const AtomicString& id) const
{
    if (id.isEmpty())
        return false;

    auto& pendingResources = svgResourcesMap().pendingResources;
    auto it = pendingResources.find(id);
    if (it == pendingResources.end())
        return false;

    return it->second.contains(element);
}

void TreeScope::clearHasPendingSVGResourcesIfPossible(SVGElement& element)
{
    if (!isElementWithPendingSVGResources(element))
        element.clearHasPendingResources();
}

void TreeScope::removeElementFromPendingSVGResources(SVGElement& element)
{
    if (!svgResourcesMap().pendingResources.isEmpty() && element.hasPendingResources()) {
        Vector<AtomicString> toBeRemoved;
        for (auto& resource : svgResourcesMap().pendingResources) {
            auto& elements = resource.second;
            elements.remove(element);
            if (elements.isEmptyIgnoringNullReferences())
                toBeRemoved.append(resource.first);
        }

        clearHasPendingSVGResourcesIfPossible(element);

        // We use the removePendingResource function here because it deals with set lifetime correctly.
        for (auto& resource : toBeRemoved)
            removePendingSVGResource(resource);
    }

    if (!svgResourcesMap().pendingResourcesForRemoval.isEmpty()) {
        Vector<AtomicString> toBeRemoved;
        for (auto& resource : svgResourcesMap().pendingResourcesForRemoval) {
            auto& elements = resource.second;
            elements.remove(element);
            if (elements.isEmptyIgnoringNullReferences())
                toBeRemoved.append(resource.first);
        }

        // We use m_pendingResourcesForRemoval here because it deals with set lifetime correctly.
        for (auto& resource : toBeRemoved)
            svgResourcesMap().pendingResourcesForRemoval.remove(resource);
    }
}

WeakHashSet<SVGElement> TreeScope::removePendingSVGResource(const AtomicString& id)
{
    return svgResourcesMap().pendingResources.take(id);
}

void TreeScope::markPendingSVGResourcesForRemoval(const AtomicString& id)
{
    if (id.isEmpty())
        return;

    ASSERT(!svgResourcesMap().pendingResourcesForRemoval.contains(id));

    auto existing = svgResourcesMap().pendingResources.take(id);
    if (!existing.isEmptyIgnoringNullReferences())
        svgResourcesMap().pendingResourcesForRemoval.add(id, WTFMove(existing));
}

RefPtr<SVGElement> TreeScope::takeElementFromPendingSVGResourcesForRemovalMap(const AtomicString& id)
{
    if (id.isEmpty())
        return nullptr;

    auto it = svgResourcesMap().pendingResourcesForRemoval.find(id);
    if (it == svgResourcesMap().pendingResourcesForRemoval.end())
        return nullptr;

    auto& resourceSet = it->second;
    RefPtr<SVGElement> firstElement = resourceSet.begin().get();
    if (!firstElement)
        return nullptr;

    resourceSet.remove(*firstElement);

    if (resourceSet.isEmptyIgnoringNullReferences())
        svgResourcesMap().pendingResourcesForRemoval.remove(id);

    return firstElement;
}

} // namespace WebCore
