/*
 * Copyright (C) 2011 Google Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "DocumentOrderedMap.h"
#include <wtf/Forward.h>
#include <wtf/Vector.h>
#include <wtf/WeakHashSet.h>
#include <wtf/text/AtomicString.h>

namespace WebCore {

class ContainerNode;
class Document;
class DOMSelection;
class Element;
class FloatPoint;
class HTMLImageElement;
class HTMLMapElement;
class IdTargetObserverRegistry;
class LayoutPoint;
class Node;
class RadioButtonGroups;
class RenderSVGResourceContainer;
class SVGElement;
class ShadowRoot;
struct SVGResourcesMap;

class TreeScope {
    friend class Document;
public:
    TreeScope* parentTreeScope() const { return m_parentTreeScope; }
    void setParentTreeScope(TreeScope&);

    Element* focusedElementInScope();
    Element* getElementById(const AtomicString&) const;
    Element* getElementById(const String&) const;
    const Vector<Element*>* getAllElementsById(const AtomicString&) const;
    bool hasElementWithId(AtomicStringImpl* id) const;
    bool containsMultipleElementsWithId(const AtomicString& id) const;
    void addElementById(const AtomicString& elementId, Element*, bool notifyObservers = true);
    void removeElementById(const AtomicString& elementId, Element*, bool notifyObservers = true);

    Element* getElementByName(const AtomicString&) const;
    bool hasElementWithName(AtomicStringImpl*) const;
    bool containsMultipleElementsWithName(const AtomicString&) const;
    void addElementByName(const AtomicString&, Element*);
    void removeElementByName(const AtomicString&, Element*);

    Document& documentScope() const { return *m_documentScope; }

    // https://dom.spec.whatwg.org/#retarget
    Node& retargetToScope(Node&) const;

    Node* ancestorNodeInThisScope(Node*) const;
    Element* ancestorElementInThisScope(Element*) const;

    void addImageMap(HTMLMapElement*);
    void removeImageMap(HTMLMapElement*);
    HTMLMapElement* getImageMap(const AtomicString&) const;

    void addImageElementByUsemap(const AtomicStringImpl&, HTMLImageElement&);
    void removeImageElementByUsemap(const AtomicStringImpl&, HTMLImageElement&);
    HTMLImageElement* imageElementByUsemap(const AtomicStringImpl&) const;

    WEBCORE_EXPORT PassRefPtr<Element> elementFromPoint(double clientX, double clientY);
    WEBCORE_EXPORT Vector<RefPtr<Element>> elementsFromPoint(double clientX, double clientY);
    WEBCORE_EXPORT Vector<RefPtr<Element>> elementsFromPoint(const FloatPoint&);

    // Find first anchor with the given name.
    // First searches for an element with the given ID, but if that fails, then looks
    // for an anchor with the given name. ID matching is always case sensitive, but
    // Anchor name matching is case sensitive in strict mode and not case sensitive in
    // quirks mode for historical compatibility reasons.
    Element* findAnchor(const String& name);

    ContainerNode& rootNode() const { return m_rootNode; }

    IdTargetObserverRegistry& idTargetObserverRegistry() const { return *m_idTargetObserverRegistry.get(); }

    RadioButtonGroups& radioButtonGroups();

    void addSVGResource(const AtomicString& id, RenderSVGResourceContainer&);
    void removeSVGResource(const AtomicString& id);
    RenderSVGResourceContainer* svgResourceById(const AtomicString& id) const;

    void addPendingSVGResource(const AtomicString& id, SVGElement&);
    bool isIdOfPendingSVGResource(const AtomicString& id) const;
    bool isPendingSVGResource(SVGElement&, const AtomicString& id) const;
    void clearHasPendingSVGResourcesIfPossible(SVGElement&);
    void removeElementFromPendingSVGResources(SVGElement&);
    WeakHashSet<SVGElement> removePendingSVGResource(const AtomicString&);
    void markPendingSVGResourcesForRemoval(const AtomicString&);
    RefPtr<SVGElement> takeElementFromPendingSVGResourcesForRemovalMap(const AtomicString&);

protected:
    TreeScope(ShadowRoot&, Document&);
    explicit TreeScope(Document&);
    ~TreeScope();

    void destroyTreeScopeData();
    void setDocumentScope(Document& document)
    {
        m_documentScope = &document;
    }

    Node* nodeFromPoint(const LayoutPoint& clientPoint, LayoutPoint* localPoint);

private:
    SVGResourcesMap& svgResourcesMap() const;
    bool isElementWithPendingSVGResources(SVGElement&) const;

    ContainerNode& m_rootNode;
    Document* m_documentScope;
    TreeScope* m_parentTreeScope;

    OwnPtr<DocumentOrderedMap> m_elementsById;
    OwnPtr<DocumentOrderedMap> m_elementsByName;
    OwnPtr<DocumentOrderedMap> m_imageMapsByName;
    OwnPtr<DocumentOrderedMap> m_imagesByUsemap;

    OwnPtr<IdTargetObserverRegistry> m_idTargetObserverRegistry;

    OwnPtr<RadioButtonGroups> m_radioButtonGroups;

    OwnPtr<SVGResourcesMap> m_svgResourcesMap;
};

ALWAYS_INLINE bool TreeScope::hasElementWithId(AtomicStringImpl* id) const
{
    ASSERT(id);
    return m_elementsById && m_elementsById->contains(id);
}

ALWAYS_INLINE bool TreeScope::containsMultipleElementsWithId(const AtomicString& id) const
{
    return m_elementsById && m_elementsById->containsMultiple(id.impl());
}

inline bool TreeScope::hasElementWithName(AtomicStringImpl* id) const
{
    ASSERT(id);
    return m_elementsByName && m_elementsByName->contains(id);
}

inline bool TreeScope::containsMultipleElementsWithName(const AtomicString& name) const
{
    return m_elementsByName && m_elementsByName->containsMultiple(name.impl());
}

TreeScope* commonTreeScope(Node*, Node*);

} // namespace WebCore
