/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "Blob.h"

#include "BlobLoader.h"
#include "BlobURL.h"
#include "File.h"
#include "HistogramSupport.h"
#include "JSDOMPromiseDeferred.h"
#include "ScriptExecutionContext.h"
#include "ThreadableBlobRegistry.h"
#include "WebKitBlobBuilder.h"
#include <wtf/NeverDestroyed.h>

namespace WebCore {

namespace {

// Used in histograms to see when we can actually deprecate the prefixed slice.
enum SliceHistogramEnum {
    SliceWithoutPrefix,
    SliceWithPrefix,
    SliceHistogramEnumMax,
};

} // namespace

class BlobURLRegistry FINAL : public URLRegistry {
public:
    virtual void registerURL(SecurityOrigin*, const KURL&, URLRegistrable*) OVERRIDE;
    virtual void unregisterURL(const KURL&) OVERRIDE;

    static URLRegistry& registry();
};


void BlobURLRegistry::registerURL(SecurityOrigin* origin, const KURL& publicURL, URLRegistrable* blob)
{
    ASSERT(&blob->registry() == this);
    ThreadableBlobRegistry::registerBlobURL(origin, publicURL, static_cast<Blob*>(blob)->url());
}

void BlobURLRegistry::unregisterURL(const KURL& url)
{
    ThreadableBlobRegistry::unregisterBlobURL(url);
}

URLRegistry& BlobURLRegistry::registry()
{
    static NeverDestroyed<BlobURLRegistry> instance;
    return instance;
}

Blob::Blob(ScriptExecutionContext* context)
    : ActiveDOMObject(context)
	, m_size(0)
{
    OwnPtr<BlobData> blobData = BlobData::create();

    // Create a new internal URL and register it with the provided blob data.
    m_internalURL = BlobURL::createInternalURL();
    ThreadableBlobRegistry::registerBlobURL(m_internalURL, blobData.release());
}

Blob::Blob(ScriptExecutionContext* context, PassOwnPtr<BlobData> blobData, long long size)
    : ActiveDOMObject(context)
	, m_type(blobData->contentType())
    , m_size(size)
{
    ASSERT(blobData);

    // Create a new internal URL and register it with the provided blob data.
    m_internalURL = BlobURL::createInternalURL();
    ThreadableBlobRegistry::registerBlobURL(m_internalURL, blobData);
}

Ref<Blob> Blob::create(ScriptExecutionContext* context, const void* data, unsigned length, const String& contentType)
{
	RefPtr<WebKitBlobBuilder> builder = WebKitBlobBuilder::create();
    builder->append(data, length);

    Ref<Blob> blob = builder->getBlob(context, contentType);
    blob->suspendIfNeeded();
    return blob;
}

Ref<Blob> Blob::create(ScriptExecutionContext* context, const SharedBuffer& buffer, const String& contentType)
{
	return create(context, buffer.data(), buffer.size(), contentType);
}

Ref<Blob> Blob::create(ScriptExecutionContext* context, const Vector<char>& data, const String& contentType)
{
	return create(context, data.data(), data.size(), contentType);
}

Ref<Blob> Blob::create(ScriptExecutionContext* context, const Vector<uint8_t>& data, const String& contentType)
{
	return create(context, data.data(), data.size(), contentType);
}

Blob::Blob(ScriptExecutionContext* context, const KURL& srcURL, const String& type, long long size)
	: ActiveDOMObject(context)
	, m_type(Blob::normalizedContentType(type))
    , m_size(size)
{
    // Create a new internal URL and register it with the same blob data as the source URL.
    m_internalURL = BlobURL::createInternalURL();
    ThreadableBlobRegistry::registerBlobURL(nullptr, m_internalURL, srcURL);
}

Blob::Blob(ReferencingExistingBlobConstructor, const Blob& blob)
	: ActiveDOMObject(blob.scriptExecutionContext())
	, m_internalURL(BlobURL::createInternalURL())
    , m_type(blob.type())
    , m_size(blob.size())
{
    ThreadableBlobRegistry::registerBlobURL(nullptr, m_internalURL, blob.url(), m_type);
}

#if ENABLE(BLOB)
Blob::Blob(ScriptExecutionContext* context, const KURL& srcURL, long long start, long long end, const String& type)
    : ActiveDOMObject(context)
	, m_type(Blob::normalizedContentType(type))
	, m_size(-1) // size is not necessarily equal to end - start.
{
    // Create a new internal URL and register it with the same blob data as the source URL.
    m_internalURL = BlobURL::createInternalURL();
    ThreadableBlobRegistry::registerBlobURLForSlice(m_internalURL, srcURL, start, end);
}
#endif

Blob::~Blob()
{
    ThreadableBlobRegistry::unregisterBlobURL(m_internalURL);
    while (!m_blobLoaders.isEmpty())
        (*m_blobLoaders.begin())->cancel();
}

unsigned long long Blob::size() const
{
    if (m_size < 0) {
        // FIXME: JavaScript cannot represent sizes as large as unsigned long long, we need to
        // come up with an exception to throw if file size is not representable.
        unsigned long long actualSize = ThreadableBlobRegistry::blobSize(m_internalURL);
        m_size = (actualSize <= std::numeric_limits<long long>::max()) ? static_cast<long long>(actualSize) : 0;
    }

    return static_cast<unsigned long long>(m_size);
}

bool Blob::isValidContentType(const String& contentType)
{
    if (contentType.isNull())
        return true;

    size_t length = contentType.length();
    if (contentType.is8Bit()) {
        const LChar* characters = contentType.characters8();
        for (size_t i = 0; i < length; ++i) {
            if (characters[i] < 0x20 || characters[i] > 0x7e)
                return false;
        }
    } else {
        const UChar* characters = contentType.characters16();
        for (size_t i = 0; i < length; ++i) {
            if (characters[i] < 0x20 || characters[i] > 0x7e)
                return false;
        }
    }
    return true;
}

String Blob::normalizedContentType(const String& contentType)
{
    if (Blob::isValidContentType(contentType))
        return contentType.lower();
    return emptyString();
}

void Blob::loadBlob(ScriptExecutionContext& context, FileReaderLoader::ReadType readType, PassOwnPtr<CompletionHandler> completionHandler)
{
    class CompletionHandlerImpl final : public BlobLoader::CompletionHandler
    {
    public:
    	CompletionHandlerImpl(Blob& blob, PassOwnPtr<Blob::CompletionHandler> completionHandler)
    		: blob(blob)
    		, pendingActivity(blob.makePendingActivity(blob))
    		, completionHandler(completionHandler)
    	{
    	}
    	virtual void operator()(BlobLoader& blobLoader) final
    	{
    		auto& blobLoaders = blob->blobLoaders();
    		for (auto it = blobLoaders.begin(); it != blobLoaders.end(); ++it) {
    			if (*it == &blobLoader) {
    				(*completionHandler)(blobLoaders.take(it));
    				return;
    			}
    		}
    	}
    private:
    	Ref<Blob> blob;
    	Ref<PendingActivity<Blob>> pendingActivity;
    	OwnPtr<Blob::CompletionHandler> completionHandler;
    };
    OwnPtr<BlobLoader> blobLoader = adoptPtr(new BlobLoader(adoptPtr(new CompletionHandlerImpl(*this, completionHandler))));
    auto* blobLoaderPtr = blobLoader.get();
    m_blobLoaders.add(blobLoader.release());
    blobLoaderPtr->start(*this, &context, readType);
}

void Blob::text(ScriptExecutionContext* context, RefPtr<DeferredPromise> promise)
{
    class TextHandler final : public CompletionHandler
    {
    public:
    	TextHandler(PassRefPtr<DeferredPromise> promise) : promise(promise) { }
    	virtual void operator()(OwnPtr<BlobLoader> blobLoader) final
		{
    		if (auto optionalErrorCode = blobLoader->errorCode()) {
				promise->reject(Exception(*optionalErrorCode));
				return;
			}
			promise->resolve(blobLoader->stringResult());
		}
    private:
    	RefPtr<DeferredPromise> promise;
    };
    loadBlob(*context, FileReaderLoader::ReadAsText, adoptPtr(new TextHandler(promise.release())));
}

void Blob::arrayBuffer(ScriptExecutionContext* context, RefPtr<DeferredPromise> promise)
{
    class ArrayBufferHandler final : public CompletionHandler
    {
    public:
    	ArrayBufferHandler(PassRefPtr<DeferredPromise> promise) : promise(promise) { }
    	virtual void operator()(OwnPtr<BlobLoader> blobLoader) final
		{
    		if (auto optionalErrorCode = blobLoader->errorCode()) {
				promise->reject(Exception(*optionalErrorCode));
				return;
			}
			auto arrayBuffer = blobLoader->arrayBufferResult();
			if (!arrayBuffer) {
				promise->reject(Exception(INVALID_STATE_ERR));
				return;
			}
			promise->resolve(arrayBuffer.get());
		}
    private:
    	RefPtr<DeferredPromise> promise;
    };
    loadBlob(*context, FileReaderLoader::ReadAsArrayBuffer, adoptPtr(new ArrayBufferHandler(promise.release())));
}

bool Blob::isNormalizedContentType(const String& contentType)
{
    if (contentType.isNull())
        return true;

    size_t length = contentType.length();
    if (contentType.is8Bit()) {
        const LChar* characters = contentType.characters8();
        for (size_t i = 0; i < length; ++i) {
            if (characters[i] < 0x20 || characters[i] > 0x7e)
                return false;
            if (characters[i] >= 'A' && characters[i] <= 'Z')
                return false;
        }
    } else {
        const UChar* characters = contentType.characters16();
        for (size_t i = 0; i < length; ++i) {
            if (characters[i] < 0x20 || characters[i] > 0x7e)
                return false;
            if (characters[i] >= 'A' && characters[i] <= 'Z')
                return false;
        }
    }
    return true;
}

bool Blob::isNormalizedContentType(const CString& contentType)
{
    size_t length = contentType.length();
    const char* characters = contentType.data();
    for (size_t i = 0; i < length; ++i) {
        if (characters[i] < 0x20 || characters[i] > 0x7e)
            return false;
        if (characters[i] >= 'A' && characters[i] <= 'Z')
            return false;
    }
    return true;
}

URLRegistry& Blob::registry() const
{
    return BlobURLRegistry::registry();
}

const char* Blob::activeDOMObjectName() const
{
    return "Blob";
}

BlobURLHandle Blob::handle() const
{
    return BlobURLHandle(m_internalURL);
}

} // namespace WebCore
