/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ActiveDOMObject.h"
#include "BlobData.h"
#include "BlobURL.h"
#include "FileReaderLoader.h"
#include "ScriptWrappable.h"
#include "PlatformString.h"
#include "URLRegistry.h"
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class Blob;
class BlobLoader;
class DeferredPromise;
class ScriptExecutionContext;
class SharedBuffer;

class Blob : public ScriptWrappable, public URLRegistrable, public RefCounted<Blob>, public ActiveDOMObject {
public:
    static Ref<Blob> create(ScriptExecutionContext* context)
    {
    	Ref<Blob> blob = adoptRef(*new Blob(context));
        blob->suspendIfNeeded();
        return blob;
    }

    static Ref<Blob> create(ScriptExecutionContext* context, PassOwnPtr<BlobData> blobData, long long size)
    {
    	Ref<Blob> blob = adoptRef(*new Blob(context, blobData, size));
        blob->suspendIfNeeded();
        return blob;
    }

    static Ref<Blob> create(ScriptExecutionContext* context, const void* data, unsigned length, const String& contentType);
    static Ref<Blob> create(ScriptExecutionContext* context, const SharedBuffer& buffer, const String& contentType);
    static Ref<Blob> create(ScriptExecutionContext* context, const Vector<char>& data, const String& contentType);
    static Ref<Blob> create(ScriptExecutionContext* context, const Vector<uint8_t>& data, const String& contentType);

    // For deserialization.
    static Ref<Blob> create(ScriptExecutionContext* context, const KURL& srcURL, const String& type, long long size)
    {
        ASSERT(Blob::isNormalizedContentType(type));
        Ref<Blob> blob = adoptRef(*new Blob(context, srcURL, type, size));
        blob->suspendIfNeeded();
        return blob;
    }

    virtual ~Blob();

    const KURL& url() const { return m_internalURL; }
    const String& type() const { return m_type; }

    unsigned long long size() const;
    virtual bool isFile() const { return false; }

    // The checks described in the File API spec.
    static bool isValidContentType(const String&);
    // The normalization procedure described in the File API spec.
    static String normalizedContentType(const String&);
    // Intended for use in ASSERT statements.
    static bool isNormalizedContentType(const String&);
    static bool isNormalizedContentType(const CString&);

    // URLRegistrable
    virtual URLRegistry& registry() const OVERRIDE;

#if ENABLE(BLOB)
    PassRefPtr<Blob> slice(ScriptExecutionContext* context, long long start = 0, long long end = std::numeric_limits<long long>::max(), const String& contentType = String()) const
    {
    	RefPtr<Blob> blob = adoptRef(new Blob(context, m_internalURL, start, end, contentType));
    	blob->suspendIfNeeded();
		return blob;
	}
#endif

    void text(ScriptExecutionContext*, RefPtr<DeferredPromise>);
    void arrayBuffer(ScriptExecutionContext*, RefPtr<DeferredPromise>);

    // Keeping the handle alive will keep the Blob data alive (but not the Blob object).
    BlobURLHandle handle() const;

    HashSet<OwnPtr<BlobLoader>>& blobLoaders() { return m_blobLoaders; }

protected:
    WEBCORE_EXPORT explicit Blob(ScriptExecutionContext*);
    Blob(ScriptExecutionContext*, PassOwnPtr<BlobData>, long long size);

    enum ReferencingExistingBlobConstructor { referencingExistingBlobConstructor };
    Blob(ReferencingExistingBlobConstructor, const Blob&);

    // For deserialization.
    Blob(ScriptExecutionContext*, const KURL& srcURL, const String& type, long long size);

#if ENABLE(BLOB)
    // For slicing.
    Blob(ScriptExecutionContext*, const KURL& srcURL, long long start, long long end, const String& contentType);
#endif

    class CompletionHandler
    {
    public:
    	virtual ~CompletionHandler() = default;
    	virtual void operator()(OwnPtr<BlobLoader>) = 0;
    };
    void loadBlob(ScriptExecutionContext&, FileReaderLoader::ReadType, PassOwnPtr<CompletionHandler>);

    // ActiveDOMObject.
    const char* activeDOMObjectName() const override;

    // This is an internal URL referring to the blob data associated with this object. It serves
    // as an identifier for this blob. The internal URL is never used to source the blob's content
    // into an HTML or for FileRead'ing, public blob URLs must be used for those purposes.
    KURL m_internalURL;

    String m_type;
    mutable long long m_size;
    HashSet<OwnPtr<BlobLoader>> m_blobLoaders;
};

} // namespace WebCore
