/*
 * Copyright (C) 2008 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "Blob.h"
#include "PlatformString.h"
#include <wtf/Optional.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>

namespace WebCore {

struct FileMetadata;
class KURL;

class File final : public Blob {
public:
    // AllContentTypes should only be used when the full path/name are trusted; otherwise, it could
    // allow arbitrary pages to determine what applications an user has installed.
    enum ContentTypeLookupPolicy {
        WellKnownContentTypes,
        AllContentTypes,
    };

    static Ref<File> create(ScriptExecutionContext* context, const String& path, ContentTypeLookupPolicy policy = WellKnownContentTypes)
    {
    	auto file = adoptRef(*new File(context, path, policy));
        file->suspendIfNeeded();
        return file;
    }

    // For deserialization.
    static Ref<File> create(ScriptExecutionContext* context, const String& path, const KURL& srcURL, const String& type)
    {
    	Ref<File> file = adoptRef(*new File(context, path, srcURL, type));
        file->suspendIfNeeded();
        return file;
    }

#if ENABLE(DIRECTORY_UPLOAD)
    static Ref<File> createWithRelativePath(ScriptExecutionContext* context, const String& path, const String& relativePath);
#endif

#if ENABLE(FILE_SYSTEM)
    // If filesystem files live in the remote filesystem, the port might pass the valid metadata (whose length field is non-negative) and cache in the File object.
    //
    // Otherwise calling size(), lastModifiedTime() and slice() will synchronously query the file metadata.
    static Ref<File> createForFileSystemFile(ScriptExecutionContext* context, const String& name, const FileMetadata& metadata)
    {
    	Ref<File> file = adoptRef(*new File(name, metadata));
        file->suspendIfNeeded();
        return file;
    }
#endif

    // Create a file with a name exposed to the author (via File.name and associated DOM properties) that differs from the one provided in the path.
    static Ref<File> createWithName(ScriptExecutionContext* context, const String& path, const String& name, ContentTypeLookupPolicy policy = WellKnownContentTypes)
    {
        if (name.isEmpty()) {
        	Ref<File> file = adoptRef(*new File(context, path, policy));
            file->suspendIfNeeded();
            return file;
        } else {
        	Ref<File> file = adoptRef(*new File(context, path, name, policy));
            file->suspendIfNeeded();
            return file;
        }
    }

    static Ref<File> create(const Blob& blob, const String& name)
    {
        return adoptRef(*new File(blob, name));
    }

    static Ref<File> create(const File& file, const String& name)
    {
        return adoptRef(*new File(file, name));
    }

    virtual bool isFile() const override { return true; }

    const String& path() const { return m_path; }
    const String& relativePath() const { return m_relativePath; }
    void setRelativePath(const String& relativePath) { m_relativePath = relativePath; }
    const String& name() const { return m_name; }

    // This may return NaN (which is converted to null Date in javascript layer) if getFileModificationTime() platform call has failed or the information is not available.
    double lastModifiedDate() const;

#if ENABLE(DIRECTORY_UPLOAD)
    // Returns the relative path of this file in the context of a directory selection.
    const String& webkitRelativePath() const { return m_relativePath; }
#endif

    // Note that this involves synchronous file operation. Think twice before calling this function.
    void captureSnapshot(long long& snapshotSize, double& snapshotModificationTime) const;

    static String contentTypeFromFilePath(const String&, ContentTypeLookupPolicy);

    bool isDirectory() const;

private:
    File(ScriptExecutionContext* context, const String& path, ContentTypeLookupPolicy);

    // For deserialization.
    File(ScriptExecutionContext* context, const String& path, const KURL& srcURL, const String& type);
    File(ScriptExecutionContext* context, const String& path, const String& name, ContentTypeLookupPolicy);
    File(const Blob&, const String& name);
	File(const File&, const String& name);

# if ENABLE(FILE_SYSTEM)
    File(const String& name, const FileMetadata&);

    // Returns true if this has a valid snapshot metadata (i.e. m_snapshotSize >= 0).
    bool hasValidSnapshotMetadata() const { return m_snapshotSize >= 0; }
#endif

    // ActiveDOMObject.
    const char* activeDOMObjectName() const final;

    String m_path;
    String m_relativePath;
    String m_name;

    Optional<int64_t> m_overrideLastModifiedDate;
    mutable Optional<bool> m_isDirectory;

#if ENABLE(FILE_SYSTEM)
    // If m_snapshotSize is negative (initialized to -1 by default), the snapshot metadata is invalid and we retrieve the latest metadata synchronously in size(), lastModifiedTime() and slice().
    // Otherwise, the snapshot metadata are used directly in those methods.
    const long long m_snapshotSize;
    const double m_snapshotModificationTime;
#endif

};

inline File* toFile(Blob* blob)
{
    ASSERT(!blob || blob->isFile());
    return static_cast<File*>(blob);
}

inline const File* toFile(const Blob* blob)
{
    ASSERT(!blob || blob->isFile());
    return static_cast<const File*>(blob);
}

} // namespace WebCore
