/*
 * Copyright (C) 2004, 2006, 2007 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2010 Torch Mobile (Beijing) Co. Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "HTMLCanvasElement.h"

#include "Attribute.h"
#include "Blob.h"
#include "BlobCallback.h"
#include "CanvasContextAttributes.h"
#include "CanvasGradient.h"
#include "CanvasPattern.h"
#include "CanvasRenderingContext2D.h"
#include "CanvasRenderingContext2DSettings.h"
#include "Chrome.h"
#include "ChromeClient.h"
#include "Document.h"
#include "ExceptionCode.h"
#include "Frame.h"
#include "FrameLoader.h"
#include "FrameLoaderClient.h"
#include "GeometryUtilities.h"
#include "GraphicsContext.h"
#include "HTMLNames.h"
#include "ImageBitmapRenderingContext.h"
#include "ImageBitmapRenderingContextSettings.h"
#include "ImageBuffer.h"
#include "ImageData.h"
#include "JSCanvasRenderingContext2D.h"
#include "JSCanvasRenderingContext2DSettings.h"
#include "JSHTMLCanvasElement.h"
#include "JSImageBitmapRenderingContext.h"
#include "MIMETypeRegistry.h"
#include "Page.h"
#include "RenderHTMLCanvas.h"
#include "Settings.h"

#if ENABLE(WEBGL)
#include "JSDictionary.h"
#include "JSWebGLRenderingContext.h"
#include "ScriptObject.h"
#include "WebGLContextAttributes.h"
#include "WebGLRenderingContext.h"
#if USE(ACCELERATED_COMPOSITING)
#include "PlatformLayer.h"
#endif
#endif
#include <wtf/GetPtr.h>

#include <math.h>
#include <stdio.h>

#if USE(CAIRO)
#define GL_GLEXT_PROTOTYPES 1
#include "PlatformContextCairo.h"
#if (CAIRO_HAS_GLESV2_SURFACE || CAIRO_HAS_GL_SURFACE)
#include "cairo/cairo-gl-ext.h"
#endif
#endif

#include <runtime/JSLock.h>
#include <runtime/Operations.h>

namespace WebCore {

using namespace HTMLNames;

// These values come from the WhatWG spec.
static const int DefaultWidth = 300;
static const int DefaultHeight = 150;

// Firefox limits width/height to 32767 pixels, but slows down dramatically before it
// reaches that limit. We limit by area instead, giving us larger maximum dimensions,
// in exchange for a smaller maximum canvas size.
static const float MaxCanvasArea = 32768 * 8192; // Maximum canvas area in CSS pixels

//In Skia, we will also limit width/height to 32767.
static const float MaxSkiaDim = 32767.0F; // Maximum width/height in CSS pixels.

int gHTMLCanvasElementCount;

inline HTMLCanvasElement::HTMLCanvasElement(const QualifiedName& tagName, Document* document)
    : HTMLElement(tagName, document)
	, ActiveDOMObject(document)
    , m_size(DefaultWidth, DefaultHeight)
    , m_rendererIsCanvas(false)
#if ENABLE(HIGH_DPI_CANVAS)
      // NOTE: High-DPI canvas requires the platform-specific ImageBuffer implementation to respect
      // the resolutionScale parameter.
    , m_deviceScaleFactor(document->frame() ? document->frame()->page()->deviceScaleFactor() : 1)
#else
    , m_deviceScaleFactor(1)
#endif
    , m_originClean(true)
{
    ASSERT(hasTagName(canvasTag));

    document->registerForDocumentActivationCallbacks(*this);

    ++gHTMLCanvasElementCount;
}

Ref<HTMLCanvasElement> HTMLCanvasElement::create(Document* document)
{
    auto canvas = adoptRef(*new HTMLCanvasElement(canvasTag, document));
    canvas->suspendIfNeeded();
    return canvas;
}

Ref<HTMLCanvasElement> HTMLCanvasElement::create(const QualifiedName& tagName, Document* document)
{
    auto canvas = adoptRef(*new HTMLCanvasElement(tagName, document));
    canvas->suspendIfNeeded();
    return canvas;
}

HTMLCanvasElement::~HTMLCanvasElement()
{
	--gHTMLCanvasElementCount;

    document().unregisterForDocumentActivationCallbacks(*this);

    for (auto* observer : m_observers)
    	observer->canvasDestroyed(this);

    m_context = nullptr; // Ensure this goes away before the ImageBuffer.

    releaseImageBufferAndContext();
}

void HTMLCanvasElement::attributeChanged(const QualifiedName& name, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason attributeModificationReason)
{
    if (name == widthAttr || name == heightAttr)
        reset();
    HTMLElement::attributeChanged(name, oldValue, newValue, attributeModificationReason);
}

RenderPtr<RenderElement> HTMLCanvasElement::createElementRenderer(PassOwnPtr<RenderStyle> style, const RenderTreePosition& insertionPosition)
{
    Frame* frame = document().frame();
    if (frame && frame->script()->canExecuteScripts(NotAboutToExecuteScript)) {
        m_rendererIsCanvas = true;
        return createRenderer<RenderHTMLCanvas>(*this, style);
    }

    m_rendererIsCanvas = false;
    return HTMLElement::createElementRenderer(style, insertionPosition);
}

void HTMLCanvasElement::addObserver(CanvasObserver* observer)
{
    m_observers.add(observer);
}

void HTMLCanvasElement::removeObserver(CanvasObserver* observer)
{
    m_observers.remove(observer);
}

void HTMLCanvasElement::setHeight(int value, ExceptionCode& ec)
{
    if (m_context && m_context->isPlaceholder()) {
    	ec = INVALID_STATE_ERR;
        return;
    }
    setAttribute(heightAttr, AtomicString::number(value));
}

void HTMLCanvasElement::setWidth(int value, ExceptionCode& ec)
{
    if (m_context && m_context->isPlaceholder()) {
    	ec = INVALID_STATE_ERR;
        return;
    }
    setAttribute(widthAttr, AtomicString::number(value));
}

CanvasRenderingContext* HTMLCanvasElement::getContext(JSC::ExecState* exec, const String& contextId, ExceptionCode& ec)
{
    if (m_context) {
        if (m_context->isPlaceholder()) {
        	ec = INVALID_STATE_ERR;
            return nullptr;
        }

        if (m_context->is2d()) {
            if (!is2dType(contextId))
                return nullptr;
            return m_context.get();
        }

        if (m_context->isBitmapRenderer()) {
            if (!isBitmapRendererType(contextId))
                return nullptr;
            return m_context.get();
        }

#if ENABLE(WEBGL)
        if (m_context->isWebGL()) {
            if (!isWebGLType(contextId))
                return nullptr;
            if ((*m_context).isWebGL1())
                return m_context.get();
#if ENABLE(WEBGL2)
            ASSERT(is<WebGL2RenderingContext>(*m_context));
            return std::optional<RenderingContext> { RefPtr<WebGL2RenderingContext> { &downcast<WebGL2RenderingContext>(*m_context) } };
#endif
        }
#endif

#if ENABLE(WEBGPU)
        if (m_context->isWebGPU()) {
            if (!isWebGPUType(contextId))
                return std::optional<RenderingContext> { std::nullopt };
            return std::optional<RenderingContext> { RefPtr<WebGPURenderingContext> { &downcast<WebGPURenderingContext>(*m_context) } };
        }
#endif

        ASSERT_NOT_REACHED();
        return nullptr;
    }

    if (is2dType(contextId)) {
    	CanvasRenderingContext2DSettings settings = convertDictionary<CanvasRenderingContext2DSettings>(*exec, exec->argument(1));
		if (UNLIKELY(exec->exception()))
            return nullptr;
        auto* context = createContext2d(contextId, settings);
        if (!context)
            return nullptr;
		return context;
    }

    if (isBitmapRendererType(contextId)) {
        auto context = createContextBitmapRenderer(contextId);
        if (!context)
            return nullptr;
        return context;
    }

#if ENABLE(WEBGL)
    if (isWebGLType(contextId)) {
		auto get3DContextAttributes = [&](JSC::ExecState* exec, RefPtr<WebGLContextAttributes>& attrs)
		{
			JSC::JSValue initializerValue = exec->argument(1);
			if (initializerValue.isUndefinedOrNull())
				return;

			auto* initializerObject = initializerValue.toObject(exec);
			JSDictionary dictionary(exec, initializerObject);

			GraphicsContext3D::Attributes graphicsAttrs;

			dictionary.tryGetProperty(ASCIILiteral("alpha"), graphicsAttrs.alpha);
			dictionary.tryGetProperty(ASCIILiteral("depth"), graphicsAttrs.depth);
			dictionary.tryGetProperty(ASCIILiteral("stencil"), graphicsAttrs.stencil);
			dictionary.tryGetProperty(ASCIILiteral("antialias"), graphicsAttrs.antialias);
			dictionary.tryGetProperty(ASCIILiteral("premultipliedAlpha"), graphicsAttrs.premultipliedAlpha);
			dictionary.tryGetProperty(ASCIILiteral("preserveDrawingBuffer"), graphicsAttrs.preserveDrawingBuffer);

			attrs = WebGLContextAttributes::create(graphicsAttrs);
		};

	    RefPtr<WebGLContextAttributes> attrs;
		get3DContextAttributes(exec, attrs);
		if (UNLIKELY(exec->exception()))
            return nullptr;
    	auto* context = createContextWebGL(contextId, attrs.get());
        if (!context)
            return nullptr;
		return context;
    }
#endif

#if ENABLE(WEBGPU)
    if (isWebGPUType(contextId)) {
        auto context = createContextWebGPU(contextId);
        if (!context)
            return std::optional<RenderingContext> { std::nullopt };
        return std::optional<RenderingContext> { RefPtr<WebGPURenderingContext> { context } };
    }
#endif

    return nullptr;
}

CanvasRenderingContext* HTMLCanvasElement::getContext(const String& type)
{
    if (HTMLCanvasElement::is2dType(type))
        return getContext2d(type, CanvasRenderingContext2DSettings());

    if (HTMLCanvasElement::isBitmapRendererType(type))
        return getContextBitmapRenderer(type);

#if ENABLE(WEBGPU)
    if (HTMLCanvasElement::isWebGPUType(type) && RuntimeEnabledFeatures::sharedFeatures().webGPUEnabled())
        return getContextWebGPU(type);
#endif

#if ENABLE(WEBGL)
    if (HTMLCanvasElement::isWebGLType(type))
        return getContextWebGL(type);
#endif

    return nullptr;
}

bool HTMLCanvasElement::is2dType(const String& type)
{
    return type == "2d";
}

CanvasRenderingContext2D* HTMLCanvasElement::createContext2d(const String& type, const CanvasRenderingContext2DSettings& settings)
{
    ASSERT_UNUSED(HTMLCanvasElement::is2dType(type), type);
    ASSERT(!m_context);
	bool usesDashbardCompatibilityMode = false;
#if ENABLE(DASHBOARD_SUPPORT)
	usesDashbardCompatibilityMode = document().settings().usesDashboardBackwardCompatibilityMode();
#endif
	m_context = CanvasRenderingContext2D::create(*this, settings, document().inQuirksMode(), usesDashbardCompatibilityMode);
#if USE(IOSURFACE_CANVAS_BACKING_STORE) || (ENABLE(ACCELERATED_2D_CANVAS) && USE(ACCELERATED_COMPOSITING))
	// Need to make sure a RenderLayer and compositing layer get created for the Canvas
	invalidateStyleAndLayerComposition();
	if (auto* renderBox = this->renderBox())
		renderBox->contentChanged(CanvasChanged);
#endif

	return static_cast<CanvasRenderingContext2D*>(m_context.get());
}

CanvasRenderingContext2D* HTMLCanvasElement::getContext2d(const String& type, const CanvasRenderingContext2DSettings& settings)
{
    ASSERT_UNUSED(HTMLCanvasElement::is2dType(type), type);

    if (m_context && !m_context->is2d())
        return nullptr;

    if (!m_context)
        return createContext2d(type, settings);
    return static_cast<CanvasRenderingContext2D*>(m_context.get());
}

#if ENABLE(WEBGL)
bool HTMLCanvasElement::isWebGLType(const String& type)
{
    // Retain support for the legacy "webkit-3d" name.
    return type == "webgl" || type == "experimental-webgl"
#if ENABLE(WEBGL2)
        || type == "webgl2"
#endif
        || type == "webkit-3d";
}

static bool shouldEnableWebGL(const Document& document)
{
    const Settings& settings = document.settings();
    if (!settings.webGLEnabled())
        return false;

    if (document.frame() && document.frame()->loader()->client()->allowWebGL(settings.webGLEnabled(), document.url())
#if !PLATFORM(CHROMIUM) && !PLATFORM(GTK) && !PLATFORM(EFL)
        && settings.acceleratedCompositingEnabled()
#endif
        ) {
    	return true;
    }

    return false;
}

WebGLRenderingContext* HTMLCanvasElement::createContextWebGL(const String& type, WebGLContextAttributes* attrs)
{
    ASSERT(HTMLCanvasElement::isWebGLType(type));
    ASSERT(!m_context);

    if (!shouldEnableWebGL(document()))
        return nullptr;

    m_context = WebGLRenderingContext::create(*this, attrs);
    if (m_context) {
        // Need to make sure a RenderLayer and compositing layer get created for the Canvas
    	invalidateStyleAndLayerComposition();
    	if (auto* renderBox = this->renderBox())
            renderBox->contentChanged(CanvasChanged);
    }
    return static_cast<WebGLRenderingContext*>(m_context.get());
}

WebGLRenderingContext* HTMLCanvasElement::getContextWebGL(const String& type, WebGLContextAttributes* attrs)
{
    ASSERT(HTMLCanvasElement::isWebGLType(type));

    if (!shouldEnableWebGL(document()))
        return nullptr;

    if (m_context && !m_context->isWebGL())
        return nullptr;

    if (!m_context)
        return createContextWebGL(type, attrs);
    return static_cast<WebGLRenderingContext*>(m_context.get());
}

#endif

#if ENABLE(WEBGPU)
bool HTMLCanvasElement::isWebGPUType(const String& type)
{
    return type == "webgpu";
}

WebGPURenderingContext* HTMLCanvasElement::createContextWebGPU(const String& type)
{
    ASSERT_UNUSED(type, HTMLCanvasElement::isWebGPUType(type));
    ASSERT(!m_context);

    if (!RuntimeEnabledFeatures::sharedFeatures().webGPUEnabled())
        return nullptr;

    m_context = WebGPURenderingContext::create(*this);
    if (m_context) {
        // Need to make sure a RenderLayer and compositing layer get created for the Canvas
        invalidateStyleAndLayerComposition();

        InspectorInstrumentation::didCreateCanvasRenderingContext(*this);
    }

    return static_cast<WebGPURenderingContext*>(m_context.get());
}

WebGPURenderingContext* HTMLCanvasElement::getContextWebGPU(const String& type)
{
    ASSERT_UNUSED(type, HTMLCanvasElement::isWebGPUType(type));

    if (!RuntimeEnabledFeatures::sharedFeatures().webGPUEnabled())
        return nullptr;

    if (m_context && !m_context->isWebGPU())
        return nullptr;

    if (!m_context)
        return createContextWebGPU(type);
    return static_cast<WebGPURenderingContext*>(m_context.get());
}
#endif

bool HTMLCanvasElement::isBitmapRendererType(const String& type)
{
    return type == "bitmaprenderer";
}

// FIXME: Needs to accept ImageBitmapRenderingContext::Settings.
ImageBitmapRenderingContext* HTMLCanvasElement::createContextBitmapRenderer(const String& type, const ImageBitmapRenderingContextSettings& settings)
{
    ASSERT_UNUSED(type, HTMLCanvasElement::isBitmapRendererType(type));
    ASSERT(!m_context);

    m_context = adoptPtr(new ImageBitmapRenderingContext(*this, settings));

    // Need to make sure a RenderLayer and compositing layer get created for the Canvas
	invalidateStyleAndLayerComposition();
	if (auto* renderBox = this->renderBox())
        renderBox->contentChanged(CanvasChanged);

    return static_cast<ImageBitmapRenderingContext*>(m_context.get());
}

ImageBitmapRenderingContext* HTMLCanvasElement::getContextBitmapRenderer(const String& type, const ImageBitmapRenderingContextSettings& settings)
{
    ASSERT_UNUSED(type, HTMLCanvasElement::isBitmapRendererType(type));
    if (!m_context)
        return createContextBitmapRenderer(type, settings);
    return static_cast<ImageBitmapRenderingContext*>(m_context.get());
}

void HTMLCanvasElement::didDraw(const FloatRect& rect)
{
	if(m_imageBuffer)
		m_imageBuffer->didDraw();

    clearCopiedImage();

    FloatRect dirtyRect = rect;
    if (auto* renderer = renderBox()) {
        FloatRect destRect;
        if ((*renderer).isRenderReplaced())
            destRect = toRenderReplaced(*renderer).replacedContentRect();
        else
            destRect = renderer->contentBoxRect();

        // Inflate dirty rect to cover antialiasing on image buffers.
		if (drawingContext() && drawingContext()->shouldAntialias())
			dirtyRect.inflate(1);
		FloatRect r = mapRect(dirtyRect, FloatRect(0, 0, size().width(), size().height()), destRect);
		r.intersect(destRect);

        if (!r.isEmpty() && !m_dirtyRect.contains(r)) {
        	m_dirtyRect.unite(r);
        	renderer->repaintRectangle(enclosingIntRect(m_dirtyRect));
        }
    }
    notifyObserversCanvasChanged(dirtyRect);
}

void HTMLCanvasElement::notifyObserversCanvasChanged(const FloatRect& rect)
{
    for (auto it = m_observers.begin(), end = m_observers.end(); it != end; ++it)
        (*it)->canvasChanged(this, rect);
}

void HTMLCanvasElement::reset()
{
    if (m_ignoreReset)
        return;

    bool ok;
    bool hadImageBuffer = hasCreatedImageBuffer();
    int w = getAttribute(widthAttr).toInt(&ok);
    if (!ok || w < 0)
        w = DefaultWidth;
    int h = getAttribute(heightAttr).toInt(&ok);
    if (!ok || h < 0)
        h = DefaultHeight;

    if (m_contextStateSaver) {
        // Reset to the initial graphics context state.
        m_contextStateSaver->restore();
        m_contextStateSaver->save();
    }

    if (m_context && m_context->is2d()) {
        CanvasRenderingContext2D* context2D = static_cast<CanvasRenderingContext2D*>(m_context.get());
        context2D->reset();
    }

    IntSize oldSize = size();
    // If the size of an existing buffer matches, we can just clear it instead of reallocating.
    // This optimization is only done for 2D canvases for now.
    if (m_hasCreatedImageBuffer && oldSize == IntSize(w, h) && m_context && m_context->is2d()) {
        if (!m_didClearImageBuffer)
            clearImageBuffer();
        return;
    }
    setSurfaceSize(IntSize(w, h));

#if ENABLE(WEBGL)
    if (m_context && m_context->is3d() && oldSize != size())
        static_cast<WebGLRenderingContext*>(m_context.get())->reshape(width(), height());
#endif

    if (auto renderer = this->renderer()) {
        if (m_rendererIsCanvas) {
            if (oldSize != size()) {
                toRenderHTMLCanvas(renderer)->canvasSizeChanged();
#if USE(ACCELERATED_COMPOSITING)
                if (renderBox() && renderBox()->hasAcceleratedCompositing())
                    renderBox()->contentChanged(CanvasChanged);
#endif
            }
            if (hadImageBuffer)
                renderer->repaint();
        }
    }

    HashSet<CanvasObserver*>::iterator end = m_observers.end();
    for (HashSet<CanvasObserver*>::iterator it = m_observers.begin(); it != end; ++it)
        (*it)->canvasResized(this);
}

bool HTMLCanvasElement::paintsIntoCanvasBuffer() const
{
    ASSERT(m_context);
#if USE(IOSURFACE_CANVAS_BACKING_STORE)
    if (m_context->is2d() || m_context->isBitmapRenderer())
        return true;
#endif

#if USE(ACCELERATED_COMPOSITING)
    if (!m_context->isAccelerated())
        return true;

    if (renderBox() && renderBox()->hasAcceleratedCompositing())
        return false;
#endif
    return true;
}


void HTMLCanvasElement::paint(GraphicsContext* context, const LayoutRect& r, bool useLowQualityScale)
{
	if(!m_dirtyRect.isEmpty()){
		if (hasCreatedImageBuffer())
			if(ImageBuffer* imageBuffer = buffer())
				imageBuffer->flush();

		// Clear the dirty rect
		m_dirtyRect = FloatRect();
	}

    if (context->paintingDisabled())
        return;
    
    if (m_context) {
        if (!paintsIntoCanvasBuffer() && !document().printing())
            return;
        m_context->paintRenderingResultsToCanvas();
    }

    if (hasCreatedImageBuffer()) {
        ImageBuffer* imageBuffer = buffer();
        if (imageBuffer) {
            if (m_presentedImage)
                context->drawImage(m_presentedImage.get(), snappedIntRect(r), CompositeSourceOver, DoNotRespectImageOrientation, useLowQualityScale);
            else
                context->drawImageBuffer(imageBuffer, snappedIntRect(r), CompositeSourceOver, BlendModeNormal, useLowQualityScale);
        }
    }

#if ENABLE(WEBGL)    
    if (is3D())
        static_cast<WebGLRenderingContext*>(m_context.get())->markLayerComposited();
#endif
}

#if ENABLE(WEBGL)
bool HTMLCanvasElement::is3D() const
{
    return m_context && m_context->is3d();
}

PassRefPtr<WebGLTexture> HTMLCanvasElement::texture()
{
#if USE(ACCELERATED_COMPOSITING)
    if(m_context && document().isDirectGLMode())
    	if(PlatformLayer* platformLayer = m_context->platformLayer()){
			if (hasCreatedImageBuffer())
				if(ImageBuffer* imageBuffer = buffer())
					imageBuffer->flush();

    		if(uint32_t webglTexture = platformLayer->webglTexture())
    			return WebGLTexture::create(webglTexture);
    	}
#endif

	return nullptr;
}
#endif

void HTMLCanvasElement::makeRenderingResultsAvailable()
{
    if (m_context)
        m_context->paintRenderingResultsToCanvas();
}

void HTMLCanvasElement::makePresentationCopy()
{
    if (!m_presentedImage) {
        // The buffer contains the last presented data, so save a copy of it.
		if(ImageBuffer* imageBuffer = buffer())
			m_presentedImage = imageBuffer->copyImage(CopyBackingStore);
    }
}

void HTMLCanvasElement::clearPresentationCopy()
{
    m_presentedImage = nullptr;
}

void HTMLCanvasElement::releaseImageBufferAndContext()
{
    m_contextStateSaver = nullptr;
    setImageBuffer(nullptr);
}

void HTMLCanvasElement::setSurfaceSize(const IntSize& size)
{
    m_size = size;
    m_hasCreatedImageBuffer = false;
    releaseImageBufferAndContext();
    clearCopiedImage();
}

String HTMLCanvasElement::toEncodingMimeType(const String& mimeType)
{
    String lowercaseMimeType = mimeType.lower();

    // FIXME: Make isSupportedImageMIMETypeForEncoding threadsafe (to allow this method to be used on a worker thread).
    if (mimeType.isNull() || !MIMETypeRegistry::isSupportedImageMIMETypeForEncoding(lowercaseMimeType))
        lowercaseMimeType = "image/png";

    return lowercaseMimeType;
}

String HTMLCanvasElement::toDataURL(const String& mimeType, Optional<double> quality, ExceptionCode& ec)
{
    if (!m_originClean) {
    	document().addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement::toDataURL security error."));
        ec = SECURITY_ERR;
        return String();
    }

    if (m_size.isEmpty() || !buffer())
        return String("data:,");

    String encodingMimeType = toEncodingMimeType(mimeType);

#if USE(CG) || USE(SKIA)
    // Try to get ImageData first, as that may avoid lossy conversions.
    RefPtr<ImageData> imageData = getImageData();

    if (imageData)
        return ImageDataToDataURL(*imageData, encodingMimeType, quality);
#endif

    makeRenderingResultsAvailable();

    return buffer()->toDataURL(encodingMimeType, quality);
}

void HTMLCanvasElement::toBlob(ScriptExecutionContext* context, RefPtr<BlobCallback> callback, const String& mimeType, JSC::JSValue qualityValue, ExceptionCode& ec)
{
    if (!m_originClean) {
    	document().addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement::toBlob security error."));
		ec = SECURITY_ERR;
    	return;
    }

    if (m_size.isEmpty() || !buffer()) {
        callback->scheduleCallback(context, nullptr);
        return;
    }

    String encodingMIMEType = toEncodingMimeType(mimeType);
    Optional<double> quality;
    if (qualityValue.isNumber())
        quality = qualityValue.toNumber(context->execState());

#if USE(CG)
    if (auto imageData = getImageData()) {
        RefPtr<Blob> blob;
        Vector<uint8_t> blobData = data(*imageData, encodingMIMEType, quality);
        if (!blobData.isEmpty())
            blob = Blob::create(&document(), WTFMove(blobData), encodingMIMEType);
        callback->scheduleCallback(context, WTFMove(blob));
        return { };
    }
#endif

    makeRenderingResultsAvailable();

    RefPtr<Blob> blob;
    Vector<uint8_t> blobData = buffer()->toData(encodingMIMEType, quality);
    if (!blobData.isEmpty()) {
        blob = Blob::create(&document(), blobData, encodingMIMEType);
    } else {
    	String message = tryMakeString("HTMLCanvasElement::toBlob(", encodingMIMEType, ") no image data.");
    	document().addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, message);
    }
    callback->scheduleCallback(context, blob.release());
    return;
}

void HTMLCanvasElement::toBlob(ScriptExecutionContext* context, RefPtr<BlobCallback> callback, const String& mimeType, ExceptionCode& ec)
{
	toBlob(context, callback, mimeType, JSC::JSValue(), ec);
}
PassRefPtr<ImageData> HTMLCanvasElement::getImageData()
{
    if (!m_context || !m_context->is3d())
       return nullptr;

#if ENABLE(WEBGL)    
    WebGLRenderingContext* ctx = static_cast<WebGLRenderingContext*>(m_context.get());

    return ctx->paintRenderingResultsToImageData();
#else
    return nullptr;
#endif
}

FloatRect HTMLCanvasElement::convertLogicalToDevice(const FloatRect& logicalRect) const
{
    FloatRect deviceRect(logicalRect);
    deviceRect.scale(m_deviceScaleFactor);

    float x = floorf(deviceRect.x());
    float y = floorf(deviceRect.y());
    float w = ceilf(deviceRect.maxX() - x);
    float h = ceilf(deviceRect.maxY() - y);
    deviceRect.setX(x);
    deviceRect.setY(y);
    deviceRect.setWidth(w);
    deviceRect.setHeight(h);

    return deviceRect;
}

FloatSize HTMLCanvasElement::convertLogicalToDevice(const FloatSize& logicalSize) const
{
    float width = ceilf(logicalSize.width() * m_deviceScaleFactor);
    float height = ceilf(logicalSize.height() * m_deviceScaleFactor);
    return FloatSize(width, height);
}

FloatSize HTMLCanvasElement::convertDeviceToLogical(const FloatSize& deviceSize) const
{
    float width = ceilf(deviceSize.width() / m_deviceScaleFactor);
    float height = ceilf(deviceSize.height() / m_deviceScaleFactor);
    return FloatSize(width, height);
}

SecurityOrigin* HTMLCanvasElement::securityOrigin() const
{
    return document().securityOrigin();
}

bool HTMLCanvasElement::shouldAccelerate(const IntSize& size) const
{
#if USE(IOSURFACE_CANVAS_BACKING_STORE)
    UNUSED_PARAM(size);
    return document().settings().canvasUsesAcceleratedDrawing();
#elif ENABLE(ACCELERATED_2D_CANVAS)
    if (m_context && !m_context->is2d())
        return false;

    auto& document = this->document();

    if (Page* page = document.page()) {
    	ChromeClient::CompositingTriggerFlags compositingTriggers = page->chrome()->client()->allowedCompositingTriggers();
    	if (!(compositingTriggers & ChromeClient::CanvasTrigger)) {
    		//document.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement can not use accelerated surface (disabled by compositing triggers)"));
			return false;
    	}
    }

    if (!document.settings().accelerated2dCanvasEnabled()) {
		//document.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement can not use accelerated surface (disabled by settings)"));
        return false;
    }

    // Do not use acceleration for small canvas.
    if (size.width() * size.height() < document.settings().minimumAccelerated2dCanvasSize()) {
		//document.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement can not use accelerated surface (small size)"));
        return false;
    }

#if PLATFORM(CHROMIUM)
    if (!WebKit::Platform::current()->canAccelerate2dCanvas())
        return false;
#endif

    const RenderStyle* style = nullptr;

    if (auto* renderer = this->renderer())
    	style = renderer->style();
    else
    	style = const_cast<HTMLCanvasElement&>(*this).computedStyle(PseudoId::None);

    if (style) {
    	if(style->hasBackground()) {
    		//document.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement can not use accelerated surface (has background)"));
    		return false;
    	}
    	if(style->hasVisibleBorderDecoration()) {
    		//document.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement can not use accelerated surface (has visible border decoration)"));
    		return false;
    	}
    }

	AtomicString xAcceleratedStr("X-Accelerated", WTF::AtomicString::ConstructFromLiteral);
	if(hasAttribute(xAcceleratedStr)) {
		if(getAttribute(xAcceleratedStr) == "0") {
			//document.addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("HTMLCanvasElement can not use accelerated surface (disabled by attribute 'X-Accelerated')"));
			return false;
		}
		return true;
	}

    return true;
#else
    UNUSED_PARAM(size);
    return false;
#endif
}

size_t HTMLCanvasElement::memoryCost() const
{
    if (!m_imageBuffer)
        return 0;
    return 4 * m_imageBuffer->internalSize().width() * m_imageBuffer->internalSize().height();
}

void HTMLCanvasElement::createImageBuffer() const
{
    ASSERT(!m_imageBuffer);

    m_hasCreatedImageBuffer = true;
    m_didClearImageBuffer = true;

    FloatSize logicalSize = size();
    FloatSize deviceSize = convertLogicalToDevice(logicalSize);
    if (!deviceSize.isExpressibleAsIntSize())
        return;

    if (deviceSize.width() * deviceSize.height() > MaxCanvasArea)
        return;
#if USE(SKIA)
    if (deviceSize.width() > MaxSkiaDim || deviceSize.height() > MaxSkiaDim)
        return;
#endif

    IntSize bufferSize(deviceSize.width(), deviceSize.height());
    if (!bufferSize.width() || !bufferSize.height())
        return;

    OwnPtr<ImageBuffer> imageBuffer;

    // force image surface when canvas size is small.
    if (size().width() * size().height() < document().settings().forceMemorySurface2dCanvasSize()) {
    	imageBuffer = ImageBuffer::createWithImageSurface(size());
    } else {
        RenderingMode renderingMode = shouldAccelerate(bufferSize) ? RenderingMode::Accelerated : RenderingMode::Unaccelerated;
    	imageBuffer = ImageBuffer::create(size(), m_deviceScaleFactor, ColorSpaceSRGB, renderingMode, LogGLMemoryError::No);

    	// try again with image surface
    	if (!imageBuffer && renderingMode == RenderingMode::Accelerated) {
    		renderingMode = RenderingMode::Unaccelerated;
        	imageBuffer = ImageBuffer::createWithImageSurface(size());
    	}
    }

    if (!imageBuffer) {
    	document().addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("Canvas paint error (image buffer is null).\n"));
        return;
    }

#if ENABLE(3D_RENDERING) && (CAIRO_HAS_GLESV2_SURFACE || CAIRO_HAS_GL_SURFACE)
    cairo_surface_t* cairoSurface = cairo_get_target(imageBuffer->context()->platformContext()->cr());
	if (cairo_surface_get_type(cairoSurface) == CAIRO_SURFACE_TYPE_GL) {
		if (cairo_gl_surface_framebuffer_complete(cairoSurface) != CAIRO_STATUS_SUCCESS) {
			document().addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, ASCIILiteral("Canvas paint error (GL frame buffer is not complete, use image surface).\n"));
        	imageBuffer = ImageBuffer::createWithImageSurface(size());
		}
	}
#endif

	setImageBuffer(imageBuffer.release());
    m_imageBuffer->context()->setShadowsIgnoreTransforms(true);
    m_imageBuffer->context()->setImageInterpolationQuality(DefaultInterpolationQuality);
    m_imageBuffer->context()->setStrokeThickness(1);
    m_contextStateSaver = adoptPtr(new GraphicsContextStateSaver(*m_imageBuffer->context()));

#if USE(JSC)
    {
    	auto& vm = *Node::scriptExecutionContext()->globalData();
    	JSC::JSLockHolder lock(vm);
    	vm.heap.reportExtraMemoryAllocated(memoryCost());
    }
#endif

#if USE(IOSURFACE_CANVAS_BACKING_STORE) || (ENABLE(ACCELERATED_2D_CANVAS) && USE(ACCELERATED_COMPOSITING))
    if (m_context && m_context->is2d()) {
        // Recalculate compositing requirements if acceleration state changed.
        const_cast<HTMLCanvasElement*>(this)->invalidateStyleAndLayerComposition();
    }
#endif
}

void HTMLCanvasElement::setImageBuffer(PassOwnPtr<ImageBuffer> buffer) const
{
	m_imageBuffer = buffer;
}

void HTMLCanvasElement::setImageBufferAndMarkDirty(PassOwnPtr<ImageBuffer> buffer)
{
    m_hasCreatedImageBuffer = true;
    setImageBuffer(buffer);
    didDraw(FloatRect(FloatPoint(), m_size));
}

GraphicsContext* HTMLCanvasElement::drawingContext() const
{
    if (m_context && !m_context->is2d())
        return nullptr;

    return buffer() ? m_imageBuffer->context() : nullptr;
}

Image* HTMLCanvasElement::copiedImage() const
{
    if (!m_copiedImage && buffer()) {
        if (m_context)
            m_context->paintRenderingResultsToCanvas();
        m_copiedImage = buffer()->copyImage(CopyBackingStore);
    }
    return m_copiedImage.get();
}

void HTMLCanvasElement::clearImageBuffer() const
{
    ASSERT(m_hasCreatedImageBuffer);
    ASSERT(!m_didClearImageBuffer);
    ASSERT(m_context);

    m_didClearImageBuffer = true;

    if (m_context->is2d()) {
        CanvasRenderingContext2D* context2D = static_cast<CanvasRenderingContext2D*>(m_context.get());
        // No need to undo transforms/clip/etc. because we are called right after the context is reset.
        context2D->clearRect(0, 0, width(), height());
    }
}

void HTMLCanvasElement::clearCopiedImage()
{
    m_copiedImage.clear();
    m_didClearImageBuffer = false;
}

AffineTransform HTMLCanvasElement::baseTransform() const
{
    ASSERT(m_hasCreatedImageBuffer);
    FloatSize unscaledSize = size();
    FloatSize deviceSize = convertLogicalToDevice(unscaledSize);
    IntSize size(deviceSize.width(), deviceSize.height());
    AffineTransform transform;
    if (size.width() && size.height())
        transform.scaleNonUniform(size.width() / unscaledSize.width(), size.height() / unscaledSize.height());
    if (m_imageBuffer)
    	return m_imageBuffer->baseTransform() * transform;
    return transform;
}

void HTMLCanvasElement::documentWillBecomeInactive()
{
    m_hasCreatedImageBuffer = false;
    m_contextStateSaver.clear();
    m_imageBuffer.clear();
    clearCopiedImage();
}

void HTMLCanvasElement::didMoveToNewDocument(Document& oldDocument, Document& newDocument)
{
    oldDocument.unregisterForDocumentActivationCallbacks(*this);

    newDocument.registerForDocumentActivationCallbacks(*this);

    HTMLElement::didMoveToNewDocument(oldDocument, newDocument);
}

void HTMLCanvasElement::removedFromAncestor(const RemovalType& removalType, ContainerNode& oldParentOfRemovedTree)
{
    if (m_imageBuffer) {
    	auto& vm = *Node::scriptExecutionContext()->globalData();
    	JSC::JSLockHolder lock(vm);
    	vm.heap.reportExtraMemoryAllocated(memoryCost());
    }

    HTMLElement::removedFromAncestor(removalType, oldParentOfRemovedTree);
}

const char* HTMLCanvasElement::activeDOMObjectName() const
{
    return "HTMLCanvasElement";
}

bool HTMLCanvasElement::virtualHasPendingActivity() const
{
    if (isContextStopped())
        return false;

#if ENABLE(WEBGL)
    if (m_context && m_context->isWebGL()) {
        // WebGL rendering context may fire contextlost / contextchange / contextrestored events at any point.
        return m_hasRelevantWebGLEventListener && !static_cast<WebGLRenderingContext&>(*m_context).isContextUnrecoverablyLost();
    }
#endif

    return false;
}

void HTMLCanvasElement::eventListenersDidChange()
{
#if ENABLE(WEBGL)
    m_hasRelevantWebGLEventListener = hasEventListeners(eventNames().webglcontextchangedEvent)
        || hasEventListeners(eventNames().webglcontextlostEvent)
        || hasEventListeners(eventNames().webglcontextrestoredEvent);
#endif
}

}
