/*
 * Copyright (C) 2004, 2006, 2009, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2010 Torch Mobile (Beijing) Co. Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "ActiveDOMObject.h"
#include "ExceptionCodePlaceholder.h"
#include "FloatRect.h"
#include "HTMLElement.h"
#include "ImageBitmapRenderingContextSettings.h"
#include "ImageBuffer.h"
#include "IntSize.h"

#if ENABLE(WEBGL)
#include "WebGLTexture.h"
#endif

#if USE(CG)
#define DefaultInterpolationQuality InterpolationLow
#else
#define DefaultInterpolationQuality InterpolationDefault
#endif

namespace JSC {
class ExecState;
}

namespace WebCore {

class BlobCallback;
class CanvasContextAttributes;
class CanvasRenderingContext;
class CanvasRenderingContext2D;
class GraphicsContext;
class GraphicsContextStateSaver;
class HTMLCanvasElement;
class Image;
class ImageData;
class IntSize;
struct CanvasRenderingContext2DSettings;
struct ImageBitmapRenderingContextSettings;
#if ENABLE(WEBGL)
class WebGLContextAttributes;
class WebGLRenderingContext;
#endif

class CanvasObserver {
public:
    virtual ~CanvasObserver() = default;

    virtual void canvasChanged(HTMLCanvasElement*, const FloatRect& changedRect) = 0;
    virtual void canvasResized(HTMLCanvasElement*) = 0;
    virtual void canvasDestroyed(HTMLCanvasElement*) = 0;
};

class HTMLCanvasElement FINAL : public HTMLElement, public ActiveDOMObject {
public:
    static Ref<HTMLCanvasElement> create(Document*);
    static Ref<HTMLCanvasElement> create(const QualifiedName&, Document*);
    virtual ~HTMLCanvasElement();

    void addObserver(CanvasObserver*);
    void removeObserver(CanvasObserver*);

    // Attributes and functions exposed to script
    ALWAYS_INLINE int width() const { return size().width(); }
    ALWAYS_INLINE int height() const { return size().height(); }

    void setWidth(int, ExceptionCode& ec = IGNORE_EXCEPTION);
    void setHeight(int, ExceptionCode& ec = IGNORE_EXCEPTION);

    const IntSize& size() const { return m_size; }

    void setSize(const IntSize& newSize)
    { 
        if (newSize == size())
            return;
        m_ignoreReset = true; 
        setWidth(newSize.width());
        setHeight(newSize.height());
        m_ignoreReset = false;
        reset();
    }

    ScriptExecutionContext* canvasBaseScriptExecutionContext() const { return HTMLElement::scriptExecutionContext(); }

    CanvasRenderingContext* getContext(JSC::ExecState*, const String&, ExceptionCode& ec);

    CanvasRenderingContext* getContext(const String&);

    static bool is2dType(const String&);
    CanvasRenderingContext2D* createContext2d(const String&, const CanvasRenderingContext2DSettings&);
    CanvasRenderingContext2D* getContext2d(const String&, const CanvasRenderingContext2DSettings&);

#if ENABLE(WEBGL)
    static bool isWebGLType(const String&);
    static bool is3dType(const String& type) { return isWebGLType(type); }
    WebGLRenderingContext* createContextWebGL(const String&, WebGLContextAttributes* = nullptr);
    WebGLRenderingContext* getContextWebGL(const String&, WebGLContextAttributes* = nullptr);
#endif
#if ENABLE(WEBGPU)
    static bool isWebGPUType(const String&);
    WebGPURenderingContext* createContextWebGPU(const String&);
    WebGPURenderingContext* getContextWebGPU(const String&);
#endif

    static bool isBitmapRendererType(const String&);
    ImageBitmapRenderingContext* createContextBitmapRenderer(const String&, const ImageBitmapRenderingContextSettings& = { });
    ImageBitmapRenderingContext* getContextBitmapRenderer(const String&, const ImageBitmapRenderingContextSettings& = { });

    static String toEncodingMimeType(const String& mimeType);
    String toDataURL(const String& mimeType, Optional<double> quality, ExceptionCode&);
    ALWAYS_INLINE String toDataURL(const String& mimeType, ExceptionCode& ec) { return toDataURL(mimeType, Nullopt, ec); }
    void toBlob(ScriptExecutionContext*, RefPtr<BlobCallback>, const String& mimeType, JSC::JSValue qualityValue, ExceptionCode& ec);
    void toBlob(ScriptExecutionContext*, RefPtr<BlobCallback>, const String& mimeType, ExceptionCode& ec);

    // Used for rendering
    void didDraw(const FloatRect&);
    void notifyObserversCanvasChanged(const FloatRect&);

    void paint(GraphicsContext*, const LayoutRect&, bool useLowQualityScale = false);

    ALWAYS_INLINE GraphicsContext* existingDrawingContext() const
    {
        if (m_hasCreatedImageBuffer)
        	return drawingContext();
        return nullptr;
    }

    GraphicsContext* drawingContext() const;

    ALWAYS_INLINE CanvasRenderingContext* renderingContext() const { return m_context.get(); }

    ALWAYS_INLINE ImageBuffer* buffer() const
    {
        if (UNLIKELY(!m_hasCreatedImageBuffer))
            createImageBuffer();
        return m_imageBuffer.get();
    }

    ALWAYS_INLINE ImageBuffer* existingBuffer() const
    {
        return m_imageBuffer.get();
    }

    Image* copiedImage() const;
    void clearCopiedImage();
    PassRefPtr<ImageData> getImageData();
    void makePresentationCopy();
    void clearPresentationCopy();

    FloatRect convertLogicalToDevice(const FloatRect&) const;
    FloatSize convertLogicalToDevice(const FloatSize&) const;

    FloatSize convertDeviceToLogical(const FloatSize&) const;

    SecurityOrigin* securityOrigin() const;
    void setOriginClean() { m_originClean = true; }
    void setOriginTainted() { m_originClean = false; }
    bool originClean() const { return m_originClean; }

    AffineTransform baseTransform() const;

#if ENABLE(WEBGL)    
    bool is3D() const;
    PassRefPtr<WebGLTexture> texture();
#endif

    void makeRenderingResultsAvailable();
    ALWAYS_INLINE bool hasCreatedImageBuffer() const { return m_hasCreatedImageBuffer; }

    bool shouldAccelerate(const IntSize&) const;

    float deviceScaleFactor() const { return m_deviceScaleFactor; }

    size_t memoryCost() const;

    // FIXME: Only some canvas rendering contexts need an ImageBuffer.
    // It would be better to have the contexts own the buffers.
    void setImageBufferAndMarkDirty(PassOwnPtr<ImageBuffer>);

private:
    HTMLCanvasElement(const QualifiedName&, Document*);

    // ActiveDOMObject.
    virtual const char* activeDOMObjectName() const final;
    virtual bool virtualHasPendingActivity() const final;

    // EventTarget.
    virtual void eventListenersDidChange() final;

    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason) final;
    virtual RenderPtr<RenderElement> createElementRenderer(PassOwnPtr<RenderStyle>, const RenderTreePosition&) final;

    void reset();

    void createImageBuffer() const;
    void clearImageBuffer() const;

    void setSurfaceSize(const IntSize&);
    void setImageBuffer(PassOwnPtr<ImageBuffer>) const;
    void releaseImageBufferAndContext();

    bool paintsIntoCanvasBuffer() const;

    virtual void documentWillBecomeInactive() override;

    virtual void didMoveToNewDocument(Document& oldDocument, Document& newDocument) final;
    virtual void removedFromAncestor(const RemovalType&, ContainerNode&) final;

    HashSet<CanvasObserver*> m_observers;

    IntSize m_size;

    OwnPtr<CanvasRenderingContext> m_context;

    bool m_rendererIsCanvas;

    bool m_ignoreReset { false };
    FloatRect m_dirtyRect;

    float m_deviceScaleFactor;
    bool m_originClean;

    // m_createdImageBuffer means we tried to malloc the buffer.  We didn't necessarily get it.
    mutable bool m_hasCreatedImageBuffer { false };
    mutable bool m_didClearImageBuffer { false };
#if ENABLE(WEBGL)
    bool m_hasRelevantWebGLEventListener { false };
#endif
    mutable OwnPtr<ImageBuffer> m_imageBuffer;
    mutable OwnPtr<GraphicsContextStateSaver> m_contextStateSaver;
    
    mutable RefPtr<Image> m_presentedImage;
    mutable RefPtr<Image> m_copiedImage; // FIXME: This is temporary for platforms that have to copy the image buffer to render (and for CSSCanvasValue).
};

inline bool isHTMLCanvasElement(const Node& node)
{
    return node.hasTagName(HTMLNames::canvasTag);
}

inline bool isHTMLCanvasElement(const Element& element)
{
    return element.hasTagName(HTMLNames::canvasTag);
}

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLCanvasElement, ArgType> {
    static bool is(ArgType& node) { return isHTMLCanvasElement(node); }
};

NODE_TYPE_CASTS(HTMLCanvasElement)

} //namespace
