/*
 * Copyright (C) 2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "HTMLDialogElement.h"

#include "CSSSelector.h"
#include "ElementTraversal.h"
#include "EventNames.h"
#include "EventSender.h"
#include "FocusOptions.h"
#include "PseudoClassChangeInvalidation.h"
#include "RenderElement.h"
#include "ScopedEventQueue.h"
#include "TypedElementDescendantIterator.h"

#include "HTMLNames.h"

namespace WebCore {

//WTF_MAKE_ISO_ALLOCATED_IMPL(HTMLDialogElement);

using namespace HTMLNames;

static DialogEventSender& dialogEventSender()
{
    static NeverDestroyed<DialogEventSender> sharedCloseEventSender;
    return sharedCloseEventSender;
}

HTMLDialogElement::HTMLDialogElement(const QualifiedName& tagName, Document* document)
    : HTMLElement(tagName, document)
	, m_isModal(false)
{
}

HTMLDialogElement::~HTMLDialogElement()
{
    dialogEventSender().cancelEvent(*this);
}

void HTMLDialogElement::show(ExceptionCode& ec)
{
	ec = 0;

    // If the element already has an open attribute, then return.
    if (isOpen()) {
        if (!isModal())
        	return;
    	ec = INVALID_STATE_ERR;
    	return;
    }

    setBooleanAttribute(openAttr, true);

    m_previouslyFocusedElement = makeWeakPtr(document().focusedElement());

    runFocusingSteps();
}

void HTMLDialogElement::showModal(ExceptionCode& ec)
{
	ec = 0;

    // If subject already has an open attribute, then throw an "InvalidStateError" DOMException.
    if (isOpen()) {
        if (isModal())
            return;
    	ec = INVALID_STATE_ERR;
    	return;
    }

    // If subject is not connected, then throw an "InvalidStateError" DOMException.
    if (!isConnected()) {
    	ec = INVALID_STATE_ERR;
    	return;
    }

    // setBooleanAttribute will dispatch a DOMSubtreeModified event.
    // Postpone callback execution that can potentially make the dialog disconnected.
    EventQueueScope scope;
    setBooleanAttribute(openAttr, true);

    setIsModal(true);

    WeakPtr<RenderBlock> containingBlockBeforeStyleResolution;
	if (auto* renderer = this->renderer())
		containingBlockBeforeStyleResolution = makeWeakPtr(renderer->containingBlock());

    if (!isInTopLayer())
        addToTopLayer();

    RenderElement::markRendererDirtyAfterTopLayerChange(this->renderer(), containingBlockBeforeStyleResolution.get());

    m_previouslyFocusedElement = makeWeakPtr(document().focusedElement());

    runFocusingSteps();
}

void HTMLDialogElement::close(const String& result)
{
    if (!isOpen())
        return;

    setBooleanAttribute(openAttr, false);

    setIsModal(false);

    if (!result.isNull())
        m_returnValue = result;

    if (isInTopLayer())
        removeFromTopLayer();

    RefPtr<Element> element = m_previouslyFocusedElement.get();
    m_previouslyFocusedElement = nullptr;
    if (element) {
        FocusOptions options;
        options.preventScroll = true;
        element->focus(options);
    }

    dialogEventSender().dispatchEventSoon(*this, eventNames().closeEvent);
}

void HTMLDialogElement::dispatchPendingEvent(DialogEventSender* eventSender, const AtomString& eventType)
{
	ASSERT_UNUSED(eventSender, eventSender == &dialogEventSender());
	dispatchEvent(Event::create(eventType, Event::CanBubble::No, Event::IsCancelable::No));
}

void HTMLDialogElement::queueCancelTask()
{
    dialogEventSender().dispatchEventSoon(*this, eventNames().cancelEvent);
}

// https://html.spec.whatwg.org/multipage/interactive-elements.html#dialog-focusing-steps
void HTMLDialogElement::runFocusingSteps()
{
    if (!isConnected())
        return;

    // Ensure renderer and inert style are up-to-date.
    document().updateStyleIfNeeded();
    if (renderer() && renderer()->style()->effectiveInert())
        return;

    RefPtr<Element> control;
    for (auto& element : descendantsOfType<Element>(*this)) {
        if (!element.isFocusable())
            continue;

        if (element.hasAttribute(autofocusAttr)) {
            control = &element;
            break;
        }

        // FIXME: Potentially remove this and adjust related WPTs after https://github.com/whatwg/html/pull/4184.
        if (!control)
            control = &element;
    }

    if (!control)
        control = this;

    if (control->isFocusable())
        control->runFocusingStepsForAutofocus();
    else
        document().setFocusedElement(nullptr);

    if (!control->document().isSameOriginAsTopDocument())
        return;

    RefPtr<Document> topDocument = control->document().topDocument();
    topDocument->clearAutofocusCandidates();
    topDocument->setAutofocusProcessed();
}

void HTMLDialogElement::removedFromAncestor(const RemovalType& removalType, ContainerNode& oldParentOfRemovedTree)
{
    HTMLElement::removedFromAncestor(removalType, oldParentOfRemovedTree);
    setIsModal(false);
}

void HTMLDialogElement::setIsModal(bool newValue)
{
	if (m_isModal == newValue)
		return;
	Style::PseudoClassChangeInvalidation styleInvalidation(*this, CSSSelector::PseudoClassModal/*, newValue*/);
	m_isModal = newValue;
}

}
