/*
 * Copyright (C) 2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ExceptionCode.h"
#include "HTMLElement.h"

namespace WebCore {

class HTMLDialogElement;

template<typename T> class EventSender;
typedef EventSender<HTMLDialogElement> DialogEventSender;

class HTMLDialogElement FINAL : public HTMLElement {
    //WTF_MAKE_ISO_ALLOCATED(HTMLDialogElement);
public:
    static Ref<HTMLDialogElement> create(const QualifiedName& tagName, Document* document) { return adoptRef(*new HTMLDialogElement(tagName, document)); }
    virtual ~HTMLDialogElement();

    bool isOpen() const { return hasAttribute(HTMLNames::openAttr); }

    const String& returnValue() const { return m_returnValue; }
    void setReturnValue(const String& value) { m_returnValue = value; }

    void show(ExceptionCode&);
    void showModal(ExceptionCode&);
    void close(const String&);

    bool isModal() const { return m_isModal; };

    void queueCancelTask();

    void runFocusingSteps();

    void dispatchPendingEvent(DialogEventSender*, const AtomString& eventType);

private:
    HTMLDialogElement(const QualifiedName&, Document*);

    virtual void removedFromAncestor(const RemovalType&, ContainerNode& oldParentOfRemovedTree) FINAL;
    void setIsModal(bool newValue);

    String m_returnValue;
    bool m_isModal;
    WeakPtr<Element> m_previouslyFocusedElement;
};

inline bool isHTMLDialogElement(const Node& node)
{
    return node.hasTagName(HTMLNames::dialogTag);
}

inline bool isHTMLDialogElement(const Element& element)
{
    return element.hasTagName(HTMLNames::dialogTag);
}

inline bool isHTMLDialogElement(const HTMLElement& element)
{
    return element.hasTagName(HTMLNames::dialogTag);
}

NODE_TYPE_CASTS(HTMLDialogElement)

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLDialogElement, ArgType> {
    static bool is(ArgType& node) { return isHTMLDialogElement(node); }
};

} // namespace WebCore
