/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004-2009, 2015 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "StyledElement.h"
#include <wtf/HashMap.h>

namespace WebCore {

class DocumentFragment;
class FormAssociatedElement;
class FormNamedItem;
class HTMLCollection;
class HTMLFormElement;

#if ENABLE(MICRODATA)
class HTMLPropertiesCollection;
class MicroDataItemValue;
#endif

enum TranslateAttributeMode {
    TranslateAttributeYes,
    TranslateAttributeNo,
    TranslateAttributeInherit
};

class HTMLElement : public StyledElement {
public:
    static PassRefPtr<HTMLElement> create(const QualifiedName& tagName, Document*);

    virtual String title() const OVERRIDE;

    void setInnerText(const String&, ExceptionCode&);
    void setOuterText(const String&, ExceptionCode&);

    virtual bool hasCustomFocusLogic() const;
    virtual bool supportsFocus() const OVERRIDE;

    String contentEditable() const;
    void setContentEditable(const String&, ExceptionCode&);

    static Editability editabilityFromContentEditableAttr(const Node&);

    virtual bool draggable() const;
    void setDraggable(bool);

    bool spellcheck() const;
    void setSpellcheck(bool);

    bool translate() const;
    void setTranslate(bool);

    void click();

    virtual bool accessKeyAction(bool sendMouseEvents) OVERRIDE;

    virtual bool rendererIsEverNeeded() FINAL;

    virtual HTMLFormElement* form() const;

    const AtomicString& dir() const;
    void setDir(const AtomicString&);

    TextDirection computeDirectionality() const;
    bool hasDirectionAuto() const;
    Optional<TextDirection> directionalityIfDirIsAuto() const;

#if ENABLE(MICRODATA)
    void setItemValue(const String&, ExceptionCode&);
    PassRefPtr<MicroDataItemValue> itemValue() const;
    PassRefPtr<HTMLPropertiesCollection> properties();
    void getItemRefElements(Vector<HTMLElement*>&);
#endif

    virtual bool isTextControlInnerTextElement() const { return false; }
    virtual bool isDateTimeNumericFieldElement() const { return false; }
    virtual bool isDateTimeSymbolicFieldElement() const { return false; }

    virtual bool isLabelable() const { return false; }
    virtual FormNamedItem* asFormNamedItem();
    virtual FormAssociatedElement* asFormAssociatedElement();

    virtual bool isInteractiveContent() const { return false; }

    bool hasTagName(const HTMLQualifiedName& name) const { return hasLocalName(name.localName()); }

    static const AtomicString& eventNameForEventHandlerAttribute(const QualifiedName& attributeName);

    // Only some element types can be disabled: https://html.spec.whatwg.org/multipage/scripting.html#concept-element-disabled
    bool canBeActuallyDisabled() const;
    virtual bool isActuallyDisabled() const;

protected:
    HTMLElement(const QualifiedName& tagName, Document*, ConstructionType);

    void addHTMLLengthToStyle(MutableStyleProperties&, CSSPropertyID, StringView value);
    void addHTMLMultiLengthToStyle(MutableStyleProperties&, CSSPropertyID, StringView value);
    void addHTMLPixelsToStyle(MutableStyleProperties&, CSSPropertyID, StringView value);
    void addHTMLNumberToStyle(MutableStyleProperties&, CSSPropertyID, StringView value);

    void addHTMLLengthToStyle(MutableStyleProperties* style, CSSPropertyID propertyID, StringView value)
    {
    	addHTMLLengthToStyle(*style, propertyID, value);
    }

    void addHTMLColorToStyle(MutableStylePropertySet*, CSSPropertyID, const String& color);

    void applyAspectRatioFromWidthAndHeightAttributesToStyle(StringView widthAttribute, StringView heightAttribute, MutableStyleProperties&);
    void applyAlignmentAttributeToStyle(const AtomicString&, MutableStylePropertySet*);
    void applyBorderAttributeToStyle(const AtomicString&, MutableStylePropertySet*);

    virtual bool matchesReadWritePseudoClass() const OVERRIDE;
    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason) override;
    Node::InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode& parentOfInsertedTree) OVERRIDE;
	virtual bool isPresentationAttribute(const QualifiedName&) const OVERRIDE;
	virtual void collectStyleForPresentationAttribute(const QualifiedName&, const AtomicString&, MutableStylePropertySet*) override;
	unsigned parseBorderWidthAttribute(const AtomicString& value) const;

    virtual void childrenChanged(const ChildChange&) OVERRIDE;
	void calculateAndAdjustDirectionality();

    typedef HashMap<AtomicStringImpl*, AtomicString> EventHandlerNameMap;
    template<size_t tableSize> static void populateEventHandlerNameMap(EventHandlerNameMap&, const QualifiedName* const (&table)[tableSize]);
    static const AtomicString& eventNameForEventHandlerAttribute(const QualifiedName& attributeName, const EventHandlerNameMap&);

    virtual bool isURLAttribute(const Attribute&) const OVERRIDE;

private:
    virtual String nodeName() const OVERRIDE;

    void mapLanguageAttributeToLocale(const AtomicString&, MutableStylePropertySet*);

    void dirAttributeChanged(const AtomicString&);
    void adjustDirectionalityIfNeededAfterChildAttributeChanged(Element* child);
    void adjustDirectionalityIfNeededAfterChildrenChanged(Element* beforeChange, ChildChangeType);

    struct TextDirectionWithStrongDirectionalityNode {
		TextDirection direction;
		RefPtr<Node> strongDirectionalityNode;
	};
	TextDirectionWithStrongDirectionalityNode directionality() const;

    TranslateAttributeMode translateAttributeMode() const;

    static void populateEventHandlerNameMap(EventHandlerNameMap&, const QualifiedName* const table[], size_t tableSize);
    static EventHandlerNameMap createEventHandlerNameMap();

#if ENABLE(MICRODATA)
    virtual String itemValueText() const;
    virtual void setItemValueText(const String&, ExceptionCode&);
#endif

    enum class AllowPercentage : bool { No, Yes };
    enum class UseCSSPXAsUnitType : bool { No, Yes };
    enum class IsMultiLength : bool { No, Yes };
    void addHTMLLengthToStyle(MutableStyleProperties&, CSSPropertyID, StringView value, AllowPercentage, UseCSSPXAsUnitType, IsMultiLength);
};

inline bool isHTMLElement(const Node& node) { return node.isHTMLElement(); }

NODE_TYPE_CASTS(HTMLElement)

ALWAYS_INLINE HTMLElement::HTMLElement(const QualifiedName& tagName, Document* document, ConstructionType type = CreateHTMLElement)
    : StyledElement(tagName, document, type)
{
    ASSERT(tagName.localName().impl());
}

template<size_t tableSize> inline void HTMLElement::populateEventHandlerNameMap(EventHandlerNameMap& map, const QualifiedName* const (&table)[tableSize])
{
    populateEventHandlerNameMap(map, table, tableSize);
}

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLElement, ArgType> {
    static bool is(ArgType& node) { return node.isHTMLElement(); }
};

inline bool Node::hasTagName(const HTMLQualifiedName& name) const
{
    return isHTMLElement() && toHTMLElement(*this).hasTagName(name);
}

} // namespace WebCore
