/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004-2023 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "FormAssociatedElement.h"
#include "LabelableElement.h"

namespace WebCore {

class FormDataList;
class HTMLFieldSetElement;
class HTMLFormElement;
class HTMLLegendElement;
class ValidationMessage;

// HTMLFormControlElement is the default implementation of FormAssociatedElement,
// and form-associated element implementations should use HTMLFormControlElement
// unless there is a special reason.
class HTMLFormControlElement : public LabelableElement, public FormAssociatedElement {
public:
    virtual ~HTMLFormControlElement();

    virtual HTMLFormElement* form() const OVERRIDE { return FormAssociatedElement::form(); }

    String formEnctype() const;
    void setFormEnctype(const String&);
    String formMethod() const;
    void setFormMethod(const String&);
    bool formNoValidate() const;

    void setAncestorDisabled(bool isDisabled);

    virtual void reset() { }

    virtual bool formControlValueMatchesRenderer() const { return m_valueMatchesRenderer; }
    virtual void setFormControlValueMatchesRenderer(bool b) { m_valueMatchesRenderer = b; }

    bool wasChangedSinceLastFormControlChangeEvent() const { return m_wasChangedSinceLastFormControlChangeEvent; }
    void setChangedSinceLastFormControlChangeEvent(bool);

    virtual void dispatchFormControlChangeEvent();
    void dispatchChangeEvent();
    void dispatchFormControlInputEvent();

    virtual bool isDisabledFormControl() const override;
    void setDisabled(bool);

    virtual bool isEnumeratable() const OVERRIDE { return false; }

    // Determines whether or not a control will be automatically focused.
    virtual bool autofocus() const;

    bool required() const;
    bool isRequired() const;

    const AtomicString& type() const { return formControlType(); }

    virtual const AtomicString& formControlType() const = 0;

    virtual bool canTriggerImplicitSubmission() const { return false; }

    // Override in derived classes to get the encoded name=value pair for submitting.
    // Return true for a successful control (see HTML4-17.13.2).
    virtual bool appendFormData(FormDataList&, bool) OVERRIDE { return false; }

    virtual bool isSuccessfulSubmitButton() const { return false; }
    virtual bool isActivatedSubmit() const { return false; }
    virtual void setActivatedSubmit(bool) { }

    // "willValidate" means "is a candidate for constraint validation".
    virtual bool willValidate() const FINAL;
    void updateVisibleValidationMessage();
    void hideVisibleValidationMessage();
    bool checkValidity(Vector<RefPtr<HTMLFormControlElement>>* unhandledInvalidControls = nullptr);
    bool reportValidity();
	void focusAndShowValidationMessage();
    // This must be called when a validation constraint or control value is changed.
    void updateValidity();
    virtual void setCustomValidity(const String&) OVERRIDE;

    bool isReadOnly() const { return m_isReadOnly; }
    bool isDisabledOrReadOnly() const { return isDisabledFormControl() || m_isReadOnly; }

    static HTMLFormControlElement* enclosingFormControlElement(Node*);

    virtual bool isSubmitButton() const { return false; }

    virtual String resultForDialogSubmit() const { return attributeWithoutSynchronization(HTMLNames::valueAttr); }

    using Node::ref;
    using Node::deref;

protected:
    HTMLFormControlElement(const QualifiedName& tagName, Document*, HTMLFormElement*);

    bool disabledByAncestorFieldset() const { return m_disabledByAncestorFieldset; }

    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason) override;
    virtual void disabledAttributeChanged();
    virtual void disabledStateChanged();
    virtual void readOnlyStateChanged();
    virtual void requiredStateChanged();
    virtual void didAttachRenderers() OVERRIDE;
    virtual InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode&) OVERRIDE;
    void didFinishInsertingNode() OVERRIDE;
    virtual void removedFromAncestor(const RemovalType&, ContainerNode&) OVERRIDE;
    virtual void didMoveToNewDocument(Document& oldDocument, Document& newDocument) OVERRIDE;

    virtual bool supportsFocus() const OVERRIDE;
    virtual bool isKeyboardFocusable(KeyboardEvent*) const OVERRIDE;
    virtual bool isMouseFocusable() const OVERRIDE;

    virtual void didRecalcStyle(Style::Change) OVERRIDE;

    virtual void dispatchBlurEvent(PassRefPtr<Element> newFocusedElement) OVERRIDE;

    // This must be called any time the result of willValidate() has changed.
    void setNeedsWillValidateCheck();
    virtual bool computeWillValidate() const;

    bool validationMessageShadowTreeContains(const Node*) const;

    virtual void willChangeForm() OVERRIDE;
    virtual void didChangeForm() OVERRIDE;
    bool belongsToFormThatIsBeingDestroyed() const { return m_belongsToFormThatIsBeingDestroyed; }

private:
    virtual void refFormAssociatedElement() final { ref(); }
    virtual void derefFormAssociatedElement() final { deref(); }

    virtual void runFocusingStepsForAutofocus() FINAL;

    virtual bool matchesValidPseudoClass() const OVERRIDE;
    virtual bool matchesInvalidPseudoClass() const OVERRIDE;

    virtual bool isFormControlElement() const FINAL { return true; }
    virtual bool alwaysCreateUserAgentShadowRoot() const OVERRIDE { return true; }

    virtual int defaultTabIndex() const FINAL;

    bool isValidFormControlElement() const;

    bool computeIsDisabledByFieldsetAncestor() const;

    virtual HTMLElement& asHTMLElement() FINAL { return *this; }
    virtual const HTMLFormControlElement& asHTMLElement() const FINAL { return *this; }
    virtual FormNamedItem* asFormNamedItem() FINAL { return this; }
    virtual FormAssociatedElement* asFormAssociatedElement() FINAL { return this; }
    virtual void formWillBeDestroyed() final;

    OwnPtr<ValidationMessage> m_validationMessage;
    bool m_disabled : 1;
    bool m_isReadOnly : 1;
    bool m_isRequired : 1;
    bool m_valueMatchesRenderer : 1;
    bool m_disabledByAncestorFieldset : 1;
    bool m_belongsToFormThatIsBeingDestroyed : 1;

    enum DataListAncestorState : uint8_t { Unknown, InsideDataList, NotInsideDataList };
    mutable enum DataListAncestorState m_dataListAncestorState;

    // The initial value of m_willValidate depends on the derived class. We can't
    // initialize it with a virtual function in the constructor. m_willValidate
    // is not deterministic as long as m_willValidateInitialized is false.
    mutable bool m_willValidateInitialized: 1;
    mutable bool m_willValidate : 1;

    // Cache of validity()->valid().
    // But "candidate for constraint validation" doesn't affect m_isValid.
    bool m_isValid : 1;

    bool m_wasChangedSinceLastFormControlChangeEvent : 1;
};

inline bool isHTMLFormControlElement(const Node& node)
{
    return node.isElementNode() && toElement(node).isFormControlElement();
}

inline bool isHTMLFormControlElement(const Element* element)
{
    return element && element->isFormControlElement();
}

inline bool isHTMLFormControlElement(const Element& element)
{
    return element.isFormControlElement();
}

inline bool isHTMLFormControlElement(const FormAssociatedElement& element)
{
    return element.isFormControlElement();
}

inline bool HTMLFormControlElement::required() const
{
    return m_isRequired;
}

inline bool HTMLFormControlElement::isRequired() const
{
    return m_isRequired;
}

NODE_TYPE_CASTS(HTMLFormControlElement)

inline HTMLFormControlElement& toHTMLFormControlElement(FormAssociatedElement& element)
{
    ASSERT_WITH_SECURITY_IMPLICATION(isHTMLFormControlElement(element));
    return static_cast<HTMLFormControlElement&>(element);
}

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLFormControlElement, ArgType> {
    static bool is(ArgType& node) { return isHTMLFormControlElement(node); }
};

} // namespace
