/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2008, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "FormNamedItem.h"
#include "GraphicsTypes.h"
#include "HTMLElement.h"
#include "HTMLImageLoader.h"
#include "MediaQueryEvaluator.h"
#if ENABLE(WEBGL)
#include "WebGLTexture.h"
#endif
#include <wtf/WeakPtr.h>

namespace WebCore {

class HTMLFormElement;
class HTMLMapElement;
class HTMLPictureElement;
class HTMLSourceElement;
struct ImageCandidate;

class HTMLImageElement : public HTMLElement, public FormNamedItem {
    friend class HTMLFormElement;
public:
    static Ref<HTMLImageElement> create(Document*);
    static Ref<HTMLImageElement> create(const QualifiedName&, Document*, HTMLFormElement*);
    static Ref<HTMLImageElement> createForJSConstructor(Document*, Optional<unsigned> width, Optional<unsigned> height);

    virtual ~HTMLImageElement();

    unsigned width();
    unsigned height();

    int naturalWidth() const;
    int naturalHeight() const;
    const KURL& currentURL() const { return m_currentURL; }
    const AtomicString& currentSrc() const { return m_currentSrc; }

    bool isServerMap() const;

    bool isAccelerated();

#if ENABLE(WEBGL)
    PassRefPtr<WebGLTexture> webglTexture();
    ALWAYS_INLINE PassRefPtr<WebGLTexture> texture() { return webglTexture(); }
    void clearTexture();
#endif

    String altText() const;

    CompositeOperator compositeOperator() const { return m_compositeOperator; }

    ALWAYS_INLINE CachedImage* cachedImage() const { return m_imageLoader.image(); }
    ALWAYS_INLINE void setCachedImage(CachedImage* i) { m_imageLoader.setImage(i); };

    ALWAYS_INLINE void setLoadManually(bool loadManually) { m_imageLoader.setLoadManually(loadManually); }

    bool matchesUsemap(const AtomicStringImpl&) const;
    HTMLMapElement* associatedMapElement() const;

    const AtomicString& alt() const;

    void setHeight(unsigned);

    KURL src() const;
    void setSrc(const String&);

    void setCrossOrigin(const AtomicString&);
    String crossOrigin() const;

    void setWidth(unsigned);

    int x() const;
    int y() const;

    bool complete() const;

    ALWAYS_INLINE bool hasPendingActivity() const { return m_imageLoader.hasPendingActivity(); }

    virtual bool canContainRangeEndPoint() const { return false; }

    virtual const AtomicString& imageSourceURL() const OVERRIDE;

    bool hasShadowControls() const { return m_experimentalImageMenuEnabled; }

    HTMLPictureElement* pictureElement() const;
    void setPictureElement(HTMLPictureElement*);

    void evaluateDynamicMediaQueryDependencies();

protected:
    HTMLImageElement(const QualifiedName&, Document*, HTMLFormElement* = nullptr);

    virtual void didMoveToNewDocument(Document& oldDocument, Document& newDocument) OVERRIDE;

    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason) override;

private:
    virtual bool isPresentationAttribute(const QualifiedName&) const OVERRIDE;
    virtual void collectStyleForPresentationAttribute(const QualifiedName&, const AtomicString&, MutableStylePropertySet*) override;
    virtual void collectExtraStyleForPresentationalHints(MutableStyleProperties&) OVERRIDE;
	void invalidateAttributeMapping();

    virtual void didAttachRenderers() OVERRIDE;
    virtual RenderPtr<RenderElement> createElementRenderer(PassOwnPtr<RenderStyle>, const RenderTreePosition&);
    void setBestFitURLAndDPRFromImageCandidate(const ImageCandidate&);

    virtual bool canStartSelection() const;

    virtual bool isURLAttribute(const Attribute&) const OVERRIDE;

    virtual bool draggable() const;

    virtual void addSubresourceAttributeURLs(ListHashSet<KURL>&) const;

    virtual InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode&) OVERRIDE;
    virtual void removedFromAncestor(const RemovalType&, ContainerNode&) OVERRIDE;

    virtual bool isFormAssociatedElement() const OVERRIDE FINAL { return false; }
    virtual FormNamedItem* asFormNamedItem() OVERRIDE FINAL { return this; }
    virtual HTMLImageElement& asHTMLElement() OVERRIDE FINAL { return *this; }
    virtual const HTMLImageElement& asHTMLElement() const OVERRIDE FINAL { return *this; }

#if ENABLE(MICRODATA)
    virtual String itemValueText() const OVERRIDE;
    virtual void setItemValueText(const String&, ExceptionCode&) OVERRIDE;
#endif

    bool isInteractiveContent() const final;

    void selectImageSource(RelevantMutation);

    ImageCandidate bestFitSourceFromPictureElement();

    HTMLSourceElement* sourceElement() const;
    void setSourceElement(HTMLSourceElement*);

    float effectiveImageDevicePixelRatio() const;

    HTMLImageLoader m_imageLoader;
    WeakPtr<HTMLFormElement> m_form;
    WeakPtr<HTMLFormElement> m_formSetByParser;

    CompositeOperator m_compositeOperator;
    AtomicString m_bestFitImageURL;
    AtomicString m_currentSrc;
    KURL m_currentURL;
    float m_imageDevicePixelRatio;
    AtomicString m_parsedUsemap;
    bool m_experimentalImageMenuEnabled;
    bool m_hadNameBeforeAttributeChanged; // FIXME: We only need this because parseAttribute() can't see the old value.

#if ENABLE(IMAGE_CONTROLS)
    void updateImageControls();
    void createImageControls();
    void destroyImageControls();
    bool hasImageControls() const;
    virtual bool childShouldCreateRenderer(const Node&) const override;
#endif

    WeakPtr<HTMLPictureElement> m_pictureElement;
    // The source element that was selected to provide the source URL.
    WeakPtr<HTMLSourceElement> m_sourceElement;
    MediaQueryDynamicResults m_mediaQueryDynamicResults;

    friend class HTMLPictureElement;
};

inline bool isHTMLImageElement(const Node* node)
{
    return node && node->hasTagName(HTMLNames::imgTag);
}

inline bool isHTMLImageElement(const Node& node)
{
    return node.hasTagName(HTMLNames::imgTag);
}

inline bool isHTMLImageElement(const Element& element)
{
    return element.hasTagName(HTMLNames::imgTag);
}

inline HTMLImageElement* toHTMLImageElement(Node* node)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!node || isHTMLImageElement(node));
    return static_cast<HTMLImageElement*>(node);
}

inline HTMLImageElement& toHTMLImageElement(Node& node)
{
    ASSERT_WITH_SECURITY_IMPLICATION(isHTMLImageElement(node));
    return static_cast<HTMLImageElement&>(node);
}

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLImageElement, ArgType> {
    static bool is(ArgType& node) { return isHTMLImageElement(node); }
};

} //namespace
