/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2012 Samsung Electronics. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "DateComponents.h"
#include "HTMLFormElement.h"
#include "HTMLTextFormControlElement.h"
#include "StepRange.h"
#include <wtf/WeakPtr.h>

namespace WebCore {

class RadioButtonGroups;
class DragData;
class FileList;
class HTMLDataListElement;
class HTMLImageLoader;
class HTMLOptionElement;
class Icon;
class InputType;
class ListAttributeTargetObserver;
class TextControlInnerTextElement;
class KURL;
struct DateTimeChooserParameters;

struct InputElementClickState {
    bool stateful { false };
    bool checked { false };
    bool indeterminate { false };
    RefPtr<HTMLInputElement> checkedRadioButton;
};

class HTMLInputElement : public HTMLTextFormControlElement {
public:
    static PassRefPtr<HTMLInputElement> create(const QualifiedName&, Document*, HTMLFormElement*, bool createdByParser);
    virtual ~HTMLInputElement();

    virtual HTMLInputElement* toInputElement() { return this; }

    virtual bool shouldAutocomplete() const FINAL;

    // For ValidityState
    virtual bool hasBadInput() const OVERRIDE;
    virtual bool patternMismatch() const FINAL;
    virtual bool rangeUnderflow() const FINAL;
    virtual bool rangeOverflow() const FINAL;
    virtual bool stepMismatch() const FINAL;
    virtual bool tooShort() const FINAL;
    virtual bool tooLong() const FINAL;
    virtual bool typeMismatch() const FINAL;
    virtual bool valueMissing() const FINAL;
    bool isValid() const FINAL;
    virtual String validationMessage() const FINAL;

    // Returns the minimum value for type=date, number, or range.  Don't call this for other types.
    double minimum() const;
    // Returns the maximum value for type=date, number, or range.  Don't call this for other types.
    // This always returns a value which is >= minimum().
    double maximum() const;
    // Sets the "allowed value step" defined in the HTML spec to the specified double pointer.
    // Returns false if there is no "allowed value step."
    bool getAllowedValueStep(Decimal*) const;
    StepRange createStepRange(AnyStepHandling) const;

    // Implementations of HTMLInputElement::stepUp() and stepDown().
    void stepUp(int, ExceptionCode&);
    void stepDown(int, ExceptionCode&);
    void stepUp(ExceptionCode& ec) { stepUp(1, ec); }
    void stepDown(ExceptionCode& ec) { stepDown(1, ec); }
    // stepUp()/stepDown() for user-interaction.
    bool isSteppable() const;

    bool isTextButton() const;

    bool isRadioButton() const;
    virtual bool isTextField() const FINAL;
    bool isSearchField() const;
    bool isInputTypeHidden() const;
    bool isPasswordField() const;
    bool isCheckbox() const;
    bool isRangeControl() const;

    // FIXME: It's highly likely that any call site calling this function should instead
    // be using a different one. Many input elements behave like text fields, and in addition
    // any unknown input type is treated as text. Consider, for example, isTextField or
    // isTextField && !isPasswordField.
    bool isText() const;
    bool isTextType() const;
    bool isEmailField() const;
    bool isFileUpload() const;
    bool isImageButton() const;
    bool isNumberField() const;
    virtual bool isSubmitButton() const final;
    bool isTelephoneField() const;
    bool isURLField() const;
    bool isDateField() const;
    bool isDateTimeLocalField() const;
    bool isMonthField() const;
    bool isTimeField() const;
    bool isWeekField() const;

    DateComponents::Type dateType() const;

#if ENABLE(INPUT_SPEECH)
    bool isSpeechEnabled() const;
#endif

    HTMLElement* containerElement() const;

    virtual RefPtr<TextControlInnerTextElement> innerTextElement() const FINAL;
    virtual PassOwnPtr<RenderStyle> createInnerTextStyle(const RenderStyle&) OVERRIDE;

    HTMLElement* innerBlockElement() const;
    HTMLElement* innerSpinButtonElement() const;
    HTMLElement* resultsButtonElement() const;
    HTMLElement* cancelButtonElement() const;
#if ENABLE(INPUT_SPEECH)
    HTMLElement* speechButtonElement() const;
#endif
    HTMLElement* sliderThumbElement() const;
    HTMLElement* sliderTrackElement() const;
    virtual HTMLElement* placeholderElement() const FINAL;
    WEBCORE_EXPORT HTMLElement* autoFillButtonElement() const;
#if ENABLE(DATALIST_ELEMENT)
    HTMLElement* dataListButtonElement() const;
#endif

    bool checked() const { return m_isChecked; }
    void setChecked(bool);

    // 'indeterminate' is a state independent of the checked state that causes the control to draw in a way that hides the actual state.
    bool indeterminate() const { return m_isIndeterminate; }
    void setIndeterminate(bool);
    // shouldAppearChecked is used by the rendering tree/CSS while checked() is used by JS to determine checked state
    bool shouldAppearChecked() const;
    virtual bool matchesIndeterminatePseudoClass() const FINAL;
    virtual bool shouldAppearIndeterminate() const FINAL;

    int size() const;
    bool sizeShouldIncludeDecoration(int& preferredSize) const;
    float decorationWidth() const;

    void setType(const String&);

    virtual String value() const FINAL;
    void setValue(const String&, ExceptionCode&, TextFieldEventBehavior = DispatchNoEvent);
    void setValue(const String&, TextFieldEventBehavior = DispatchNoEvent);
    void setValueForUser(const String& value) { setValue(value, DispatchInputAndChangeEvent); }
    // Checks if the specified string would be a valid value.
    // We should not call this for types with no string value such as CHECKBOX and RADIO.
    bool isValidValue(const String&) const;
    bool hasDirtyValue() const { return !m_valueIfDirty.isNull(); };

    String placeholder() const;

    String sanitizeValue(const String&) const;

    String localizeValue(const String&) const;

    // The value which is drawn by a renderer.
    String visibleValue() const;

    const String& suggestedValue() const;
    void setSuggestedValue(const String&);

    void setEditingValue(const String&);

    double valueAsDate() const;
    void setValueAsDate(double, ExceptionCode&);

    double valueAsNumber() const;
    void setValueAsNumber(double, ExceptionCode&, TextFieldEventBehavior = DispatchNoEvent);

    String valueWithDefault() const;

    void setValueFromRenderer(const String&);

    bool canHaveSelection() const;

    virtual bool rendererIsNeeded(const RenderStyle&) FINAL;
    virtual RenderPtr<RenderElement> createElementRenderer(PassOwnPtr<RenderStyle>, const RenderTreePosition&) FINAL;
    virtual void willAttachRenderers() FINAL;
    virtual void didAttachRenderers() FINAL;
    virtual void didDetachRenderers() FINAL;

    // FIXME: For isActivatedSubmit and setActivatedSubmit, we should use the NVI-idiom here by making
    // it private virtual in all classes and expose a public method in HTMLFormControlElement to call
    // the private virtual method.
    virtual bool isActivatedSubmit() const FINAL;
    virtual void setActivatedSubmit(bool flag) FINAL;

    String altText() const;

    void willDispatchEvent(Event&, InputElementClickState&);
    void didDispatchClickEvent(Event&, const InputElementClickState&);

    void didBlur();

    int maxResults() const { return m_maxResults; }

    String defaultValue() const;
    void setDefaultValue(const String&);

    Vector<String> acceptMIMETypes();
    Vector<String> acceptFileExtensions();
    String accept() const;
    String alt() const;

    void setSize(unsigned);

    KURL src() const;

    unsigned effectiveMaxLength() const;

    String characterType() const;
    
    bool multiple() const;

    bool isAutoFilled() const { return m_isAutoFilled; }
    WEBCORE_EXPORT void setAutoFilled(bool = true);

    AutoFillButtonType autoFillButtonType() const { return static_cast<AutoFillButtonType>(m_autoFillButtonType); }
    WEBCORE_EXPORT void setShowAutoFillButton(AutoFillButtonType);

    bool hasAutoFillStrongPasswordButton() const  { return autoFillButtonType() == AutoFillButtonType::StrongPassword || autoFillButtonType() == AutoFillButtonType::StrongConfirmationPassword; }

    FileList* files();
    void setFiles(PassRefPtr<FileList>);

    // Returns true if the given DragData has more than one dropped files.
    bool receiveDroppedFiles(const DragData*);

#if ENABLE(FILE_SYSTEM)
    String droppedFileSystemId();
#endif

    Icon* icon() const;
    // These functions are used for rendering the input active during a
    // drag-and-drop operation.
    bool canReceiveDroppedFiles() const;
    void setCanReceiveDroppedFiles(bool);

    void addSearchResult();
    void onSearch();

    virtual bool willRespondToMouseClickEvents() OVERRIDE;

#if ENABLE(DATALIST_ELEMENT)
    HTMLElement* list() const;
    HTMLDataListElement* dataList() const;
    void dataListMayHaveChanged();
#endif

    Vector<Ref<HTMLInputElement>> radioButtonGroup() const;
    PassRefPtr<HTMLInputElement> checkedRadioButtonForGroup() const;
    bool isInRequiredRadioButtonGroup();
    // Returns null if this isn't associated with any radio button group.
    RadioButtonGroups* radioButtonGroups() const;

    // Functions for InputType classes.
    void setValueInternal(const String&, TextFieldEventBehavior);
    bool isTextFormControlFocusable() const;
    bool isTextFormControlKeyboardFocusable(KeyboardEvent*) const;
    bool isTextFormControlMouseFocusable() const;
    bool valueAttributeWasUpdatedAfterParsing() const { return m_valueAttributeWasUpdatedAfterParsing; }

    void cacheSelectionInResponseToSetValue(int caretOffset) { cacheSelection(caretOffset, caretOffset, SelectionHasNoDirection); }

#if ENABLE(INPUT_TYPE_COLOR)
    // For test purposes.
    void selectColorInColorChooser(const Color&);
#endif

    String defaultToolTip() const;

#if ENABLE(MEDIA_CAPTURE)
    String capture() const;
    void setCapture(const String& value);
#endif

    static const unsigned maxEffectiveLength;

    unsigned height() const;
    unsigned width() const;
    void setHeight(unsigned);
    void setWidth(unsigned);

    virtual void blur(bool byUser = false) OVERRIDE;
    void defaultBlur(bool byUser);

    virtual const AtomicString& name() const FINAL;

    virtual void setRangeText(const String& replacement, ExceptionCode&) FINAL;
    virtual void setRangeText(const String& replacement, unsigned start, unsigned end, const String& selectionMode, ExceptionCode&) FINAL;

#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
    bool setupDateTimeChooserParameters(DateTimeChooserParameters&);
#endif

    void endEditing();

    bool hasImageLoader() const { return m_imageLoader.get(); }
    HTMLImageLoader* imageLoader();

    bool shouldTruncateText(const RenderStyle&) const;

    int selectionStartForBindings(ExceptionCode&) const;
    void setSelectionStartForBindings(int, ExceptionCode&);

    int selectionEndForBindings(ExceptionCode&) const;
    void setSelectionEndForBindings(int, ExceptionCode&);

    String selectionDirectionForBindings(ExceptionCode&) const;
    void setSelectionDirectionForBindings(const String&, ExceptionCode&);

    void setSelectionRangeForBindings(int start, int end, const String& direction, ExceptionCode&);

    virtual bool matchesReadWritePseudoClass() const OVERRIDE;

    virtual String resultForDialogSubmit() const final;

protected:
    HTMLInputElement(const QualifiedName&, Document*, HTMLFormElement*, bool createdByParser);

    virtual void defaultEventHandler(Event*) OVERRIDE;

private:
    enum AutoCompleteSetting { Uninitialized, On, Off };

    virtual void didAddUserAgentShadowRoot(ShadowRoot*) FINAL;
    virtual bool canHaveUserAgentShadowRoot() const FINAL { return true; }

    virtual void willChangeForm() FINAL;
    virtual void didChangeForm() FINAL;
    virtual InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode&) FINAL;
    virtual void didFinishInsertingNode() FINAL;
    virtual void removedFromAncestor(const RemovalType&, ContainerNode&) FINAL;
    virtual void didMoveToNewDocument(Document& oldDocument, Document& newDocument) FINAL;

    virtual bool hasCustomFocusLogic() const FINAL;
    virtual bool isKeyboardFocusable(KeyboardEvent*) const FINAL;
    virtual bool isMouseFocusable() const FINAL;
    virtual bool isEnumeratable() const FINAL;
    virtual bool supportLabels() const FINAL;
    virtual void updateFocusAppearance(SelectionRestorationMode, SelectionRevealMode) FINAL;
    virtual bool shouldUseInputMethod() FINAL;

    bool isInteractiveContent() const final;

    virtual bool isInnerTextElementEditable() const final { return !hasAutoFillStrongPasswordButton() && HTMLTextFormControlElement::isInnerTextElementEditable(); }

    virtual bool canTriggerImplicitSubmission() const FINAL { return isTextField(); }

    virtual const AtomicString& formControlType() const FINAL;

    virtual bool shouldSaveAndRestoreFormControlState() const FINAL;
    virtual FormControlState saveFormControlState() const FINAL;
    virtual void restoreFormControlState(const FormControlState&) FINAL;

    virtual bool canStartSelection() const FINAL;

    virtual bool accessKeyAction(bool sendMouseEvents) FINAL;

    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason = ModifiedDirectly) final;
    virtual bool isPresentationAttribute(const QualifiedName&) const FINAL;
    virtual void collectStyleForPresentationAttribute(const QualifiedName&, const AtomicString&, MutableStylePropertySet*) final;
    virtual void finishParsingChildren() FINAL;
    virtual void parserDidSetAttributes() FINAL;

    virtual void copyNonAttributePropertiesFromElement(const Element&) FINAL;

    virtual bool appendFormData(FormDataList&, bool) FINAL;

    virtual bool isSuccessfulSubmitButton() const FINAL;
    virtual bool matchesDefaultPseudoClass() const FINAL;

    virtual void reset() FINAL;

    virtual bool isURLAttribute(const Attribute&) const FINAL;
    virtual bool isInRange() const FINAL;
    virtual bool isOutOfRange() const FINAL;

    virtual void documentDidResumeFromPageCache() FINAL;

    virtual void addSubresourceAttributeURLs(ListHashSet<KURL>&) const FINAL;

    bool needsSuspensionCallback();
    void registerForSuspensionCallbackIfNeeded();
    void unregisterForSuspensionCallbackIfNeeded();

    bool supportsMinLength() const { return isTextType(); }
    bool supportsMaxLength() const { return isTextType(); }
    bool tooShort(StringView, NeedsToCheckDirtyFlag) const;
    bool tooLong(StringView, NeedsToCheckDirtyFlag) const;

    virtual bool supportsPlaceholder() const FINAL;
    virtual bool isPlaceholderEmpty() const FINAL;
    virtual void updatePlaceholderText() FINAL;
    virtual bool isEmptyValue() const FINAL { return innerTextValue().isEmpty(); }
    virtual bool isEmptySuggestedValue() const FINAL { return suggestedValue().isEmpty(); }
    virtual void handleFocusEvent(Node*, FocusDirection) FINAL;
    virtual void handleBlurEvent() FINAL;

    virtual bool isOptionalFormControl() const FINAL { return !isRequiredFormControl(); }
    virtual bool isRequiredFormControl() const FINAL;
    virtual bool computeWillValidate() const FINAL;
    virtual void requiredStateChanged() FINAL;

    void initializeInputType();
    void updateType();
    void runPostTypeUpdateTasks();
    
    virtual void subtreeHasChanged() FINAL;
    virtual void disabledStateChanged() final;
    virtual void readOnlyStateChanged() final;

#if ENABLE(DATALIST_ELEMENT)
    void resetListAttributeTargetObserver();
#endif
    void maxLengthAttributeChanged(const AtomicString& newValue);
    void minLengthAttributeChanged(const AtomicString& newValue);
    void updateValueIfNeeded();

    void addToRadioButtonGroup();
    void removeFromRadioButtonGroup();

    AtomicString m_name;
    String m_valueIfDirty;
    String m_suggestedValue;
    int m_size;
    short m_maxResults { -1 };
    bool m_isChecked : 1;
    bool m_dirtyCheckednessFlag : 1;
    bool m_isIndeterminate : 1;
    bool m_hasType : 1;
    bool m_isActivatedSubmit : 1;
    unsigned m_autocomplete : 2; // AutoCompleteSetting
    bool m_isAutoFilled : 1;
    unsigned m_autoFillButtonType : 3; // AutoFillButtonType;
#if ENABLE(DATALIST_ELEMENT)
    bool m_hasNonEmptyList : 1;
#endif
    bool m_stateRestored : 1;
    bool m_parsingInProgress : 1;
    bool m_valueAttributeWasUpdatedAfterParsing : 1;
    bool m_wasModifiedByUser : 1;
    bool m_canReceiveDroppedFiles : 1;
    bool m_hasTouchEventHandler: 1;
    RefPtr<InputType> m_inputType;
    // The ImageLoader must be owned by this element because the loader code assumes
    // that it lives as long as its owning element lives. If we move the loader into
    // the ImageInput object we may delete the loader while this element lives on.
    OwnPtr<HTMLImageLoader> m_imageLoader;
#if ENABLE(DATALIST_ELEMENT)
    OwnPtr<ListAttributeTargetObserver> m_listAttributeTargetObserver;
#endif
};

inline bool isHTMLInputElement(const Node& node)
{
    return node.hasTagName(HTMLNames::inputTag);
}

inline bool isHTMLInputElement(const Element& element)
{
    return element.hasTagName(HTMLNames::inputTag);
}

inline bool isHTMLInputElement(const HTMLElement& element)
{
    return element.hasTagName(HTMLNames::inputTag);
}

template<>
ALWAYS_INLINE HTMLInputElement* dynamicDowncast(HTMLElement* source)
{
    //static_assert(!std::is_same_v<Source, Target>, "Unnecessary cast to same type");
    //static_assert(std::is_base_of_v<Source, Target>, "Should be a downcast");
    return source && isHTMLInputElement(*source) ? static_cast<HTMLInputElement*>(source) : nullptr;
}

template<>
ALWAYS_INLINE HTMLInputElement* dynamicDowncast(HTMLTextFormControlElement* source)
{
    //static_assert(!std::is_same_v<Source, Target>, "Unnecessary cast to same type");
    //static_assert(std::is_base_of_v<Source, Target>, "Should be a downcast");
    return source && isHTMLInputElement(*source) ? static_cast<HTMLInputElement*>(source) : nullptr;
}

NODE_TYPE_CASTS(HTMLInputElement)

} //namespace
