/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2010 Apple Inc. All rights reserved.
 *           (C) 2006 Alexey Proskuryakov (ap@nypop.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "HTMLLabelElement.h"

#include "Document.h"
#include "ElementIterator.h"
#include "ElementTraversal.h"
#include "Event.h"
#include "EventNames.h"
#include "FormAssociatedElement.h"
#include "Frame.h"
#include "HTMLFormElement.h"
#include "HTMLNames.h"
#include "VisibleSelection.h"
#include <wtf/SetForScope.h>

namespace WebCore {

using namespace HTMLNames;

static LabelableElement* firstElementWithIdIfLabelable(TreeScope& treeScope, const AtomicString& id)
{
    Element* e = treeScope.getElementById(id);
    if (!e || !e->isHTMLElement())
        return nullptr;

    HTMLElement* element = toHTMLElement(e);
    if (!element->isLabelable())
        return nullptr;

    LabelableElement* labelableElement = static_cast<LabelableElement*>(element);
    return labelableElement->supportLabels() ? labelableElement : nullptr;
}

inline HTMLLabelElement::HTMLLabelElement(const QualifiedName& tagName, Document* document)
    : HTMLElement(tagName, document)
{
    ASSERT(hasTagName(labelTag));
}

PassRefPtr<HTMLLabelElement> HTMLLabelElement::create(const QualifiedName& tagName, Document* document)
{
    return adoptRef(new HTMLLabelElement(tagName, document));
}

LabelableElement* HTMLLabelElement::control()
{
    const AtomicString& controlId = getAttribute(forAttr);
    if (controlId.isNull()) {
        // Search the children and descendants of the label element for a form element.
        // per http://dev.w3.org/html5/spec/Overview.html#the-label-element
        // the form element must be "labelable form-associated element".
        for (auto& labelableElement : descendantsOfType<LabelableElement>(*this)) {
            if (labelableElement.supportLabels())
				return &labelableElement;
		}
        return nullptr;
    }

    return inDocument() ? firstElementWithIdIfLabelable(treeScope(), controlId) : nullptr;
}

HTMLFormElement* HTMLLabelElement::form() const
{
    return FormAssociatedElement::findAssociatedForm(this, 0);
}

void HTMLLabelElement::setActive(bool down, bool pause)
{
    if (down == active())
        return;

    // Update our status first.
    HTMLElement::setActive(down, pause);

    // Also update our corresponding control.
    if (HTMLElement* element = control())
        element->setActive(down, pause);
}

void HTMLLabelElement::setHovered(bool over)
{
    if (over == hovered())
        return;
        
    // Update our status first.
    HTMLElement::setHovered(over);

    // Also update our corresponding control.
    if (HTMLElement* element = control())
        element->setHovered(over);
}

bool HTMLLabelElement::isEventTargetedAtInteractiveDescendants(Event& event) const
{
    if (!event.target() || !event.target()->isNode())
        return false;

    RefPtr<Node> node = event.target()->toNode();
    if (!containsIncludingShadowDOM(node.get()))
        return false;

    for (const Node* it = node.get(); it && it != this; it = it->parentElementInComposedTree()) {
        if (it->isHTMLElement() && toHTMLElement(*it).isInteractiveContent())
            return true;
    }

    return false;
}

void HTMLLabelElement::defaultEventHandler(Event* event)
{
    if (event->type() == eventNames().clickEvent && !m_processingClick) {
        RefPtr<LabelableElement> control = this->control();

        // If we can't find a control or if the control received the click
        // event, then there's no need for us to do anything.
        if (!control || (event->target() && control->containsIncludingShadowDOM(event->target()->toNode().get()))) {
        	HTMLElement::defaultEventHandler(event);
			return;
		}

        // The activation behavior of a label element for events targeted at interactive
        // content descendants of a label element, and any descendants of those interactive
        // content descendants, must be to do nothing.
        // https://html.spec.whatwg.org/#the-label-element
        if (isEventTargetedAtInteractiveDescendants(*event)) {
            HTMLElement::defaultEventHandler(event);
            return;
        }

        SetForScope<bool> processingClick(m_processingClick, true);

        control->dispatchSimulatedClick(event);

        document().updateLayoutIgnorePendingStylesheets();
        if (control->isMouseFocusable()) {
        	FocusOptions options;
        	options.trigger = FocusTrigger::Click;
        	control->focus(options);
        }
        
        event->setDefaultHandled();
    }
    
    HTMLElement::defaultEventHandler(event);
}

bool HTMLLabelElement::willRespondToMouseClickEvents()
{
    if (control() && control()->willRespondToMouseClickEvents())
        return true;

    return HTMLElement::willRespondToMouseClickEvents();
}

void HTMLLabelElement::focus(const FocusOptions& options)
{
    Ref<HTMLLabelElement> protectedThis(*this);
    if (document().haveStylesheetsLoaded()) {
        document().updateLayout();
        if (isFocusable()) {
            // The value of restorationMode is not used for label elements as it doesn't override updateFocusAppearance.
            Element::focus(options);
            return;
        }
    }

    // to match other browsers, always restore previous selection
    if (auto* element = control()) {
    	FocusOptions newOptions;
    	newOptions.selectionRestorationMode = SelectionRestorationMode::RestoreOrSelectAll;
    	newOptions.direction = options.direction;
        element->focus(newOptions);
    }
}

bool HTMLLabelElement::accessKeyAction(bool sendMouseEvents)
{
    if (HTMLElement* element = control())
        return element->accessKeyAction(sendMouseEvents);

    return HTMLElement::accessKeyAction(sendMouseEvents);
}

} // namespace
