/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "FormAssociatedElement.h"
#include "HTMLFormElement.h"
#include "HTMLPlugInImageElement.h"

namespace JSC
{
class ExecState;
}

namespace WebCore {

class HTMLFormElement;
class PluginView;
class JSHTMLObjectElement;

class JSPropertyDelegate
{
public:
	virtual ~JSPropertyDelegate() = default;

	virtual bool jsGet(JSC::ExecState*, JSC::PropertyName, JSC::PropertySlot&, JSHTMLObjectElement*) = 0;
	virtual bool jsPut(JSC::ExecState*, JSC::PropertyName, JSC::JSValue, JSC::PutPropertySlot&, JSHTMLObjectElement*) = 0;

	virtual void updateNativePlayerID() { }
	virtual void resetEndedEventDispatched() { }

protected:
	JSPropertyDelegate() = default;
};

class HTMLObjectElement : public HTMLPlugInImageElement, public FormAssociatedElement {
public:
    static PassRefPtr<HTMLObjectElement> create(const QualifiedName&, Document*, HTMLFormElement*);

    bool isDocNamedItem() const { return m_docNamedItem; }

    bool containsJavaApplet() const;

    bool jsGet(JSC::ExecState*, JSC::PropertyName, JSC::PropertySlot&, JSHTMLObjectElement*);
    bool jsPut(JSC::ExecState*, JSC::PropertyName, JSC::JSValue, JSC::PutPropertySlot&, JSHTMLObjectElement*);

    bool hasFallbackContent() const;
    virtual bool useFallbackContent() const { return m_useFallbackContent; }
    virtual void renderFallbackContent();

    virtual bool willValidate() const OVERRIDE { return false; }

    // Implementation of constraint validation API.
    // Note that the object elements are always barred from constraint validation.
    static bool checkValidity() { return true; }
    static bool reportValidity() { return true; }

    using HTMLPlugInImageElement::ref;
    using HTMLPlugInImageElement::deref;

    using FormAssociatedElement::form;

    // FIXME: This function should not deal with url or serviceType
    // so that we can better share code between <object> and <embed>.
    void parametersForPlugin(Vector<String>& paramNames, Vector<String>& paramValues, String& url, String& serviceType);

protected:
    HTMLObjectElement(const QualifiedName&, Document*, HTMLFormElement*);
    virtual ~HTMLObjectElement();

    virtual void updateWidget(PluginCreationOption);

    virtual void removedFromAncestor(const RemovalType&, ContainerNode&) OVERRIDE;

    virtual void attributeChanged(const QualifiedName&, const AtomicString& oldValue, const AtomicString& newValue, AttributeModificationReason = ModifiedDirectly) override;

    virtual void createJSPropertyDelegate() {}

    virtual void didMoveToNewDocument(Document& oldDocument, Document& newDocument) OVERRIDE;

private:
    virtual RenderWidget* renderWidgetLoadingPlugin() const FINAL;

    virtual int defaultTabIndex() const FINAL;

    virtual bool isPresentationAttribute(const QualifiedName&) const OVERRIDE;
    virtual void collectStyleForPresentationAttribute(const QualifiedName&, const AtomicString&, MutableStylePropertySet*) override;

    virtual InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode&) OVERRIDE;
    virtual void didFinishInsertingNode() FINAL;

    virtual bool rendererIsNeeded(const RenderStyle&) OVERRIDE;

    virtual void childrenChanged(const ChildChange&) OVERRIDE;

    virtual bool isURLAttribute(const Attribute&) const final;
    virtual const AtomicString& imageSourceURL() const final;

    virtual void addSubresourceAttributeURLs(ListHashSet<KURL>&) const final;

    void updateDocNamedItem();
    
    bool shouldAllowQuickTimeClassIdQuirk();
    bool hasValidClassId();

    virtual void refFormAssociatedElement() { ref(); }
    virtual void derefFormAssociatedElement() { deref(); }

    virtual FormNamedItem* asFormNamedItem() OVERRIDE FINAL { return this; }
    virtual FormAssociatedElement* asFormAssociatedElement() FINAL { return this; }
    virtual HTMLObjectElement& asHTMLElement() OVERRIDE FINAL { return *this; }
    virtual const HTMLObjectElement& asHTMLElement() const OVERRIDE FINAL { return *this; }

    bool isInteractiveContent() const final;

    virtual bool isFormControlElement() const final { return false; }

    virtual bool isEnumeratable() const final { return true; }
    virtual bool appendFormData(FormDataList&, bool) final;
    virtual bool canContainRangeEndPoint() const FINAL;

#if ENABLE(MICRODATA)
    virtual String itemValueText() const OVERRIDE;
    virtual void setItemValueText(const String&, ExceptionCode&) OVERRIDE;
#endif

    bool m_docNamedItem;
    bool m_useFallbackContent;

protected:
    OwnPtr<JSPropertyDelegate> m_jsPropertyDelegate;
};

inline bool isHTMLObjectElement(const Node& node)
{
    return node.hasTagName(HTMLNames::objectTag);
}

inline bool isHTMLObjectElement(const Element& element)
{
    return element.hasTagName(HTMLNames::objectTag);
}

inline bool isHTMLObjectElement(const HTMLElement& element)
{
    return element.hasTagName(HTMLNames::objectTag);
}

NODE_TYPE_CASTS(HTMLObjectElement)

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLObjectElement, ArgType> {
    static bool is(ArgType& node) { return isHTMLObjectElement(node); }
};

}
