/*
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2009, 2010, 2011 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "HTMLFormControlElementWithState.h"
#include "HTMLFormElement.h"
#include "HTMLOptionElement.h"
#include "HTMLOptionsCollection.h"
#include "TypeAhead.h"
#include <wtf/Vector.h>

namespace WebCore {

class HTMLSelectElement : public HTMLFormControlElementWithState, private TypeAheadDataSource {
public:
    static PassRefPtr<HTMLSelectElement> create(const QualifiedName&, Document*, HTMLFormElement*);

    int selectedIndex() const;
    void setSelectedIndex(int);

    void optionSelectedByUser(int index, bool dispatchChangeEvent, bool allowMultipleSelection = false);

    virtual String validationMessage() const FINAL;
    virtual bool valueMissing() const FINAL;

    unsigned length() const;

    int size() const { return m_size; }
    bool multiple() const { return m_multiple; }

    bool usesMenuList() const;

    void add(HTMLElement&, HTMLElement* beforeElement, ExceptionCode&);
    void addByIndex(HTMLElement&, int beforeIndex, ExceptionCode&);

    using Node::remove;
    void removeByIndex(int); // Should be remove(int) but that conflicts with Node::remove(ExceptionCode&).

    String value() const;
    void setValue(const String&);

    PassRefPtr<HTMLOptionsCollection> options();
    PassRefPtr<HTMLCollection> selectedOptions();

    void optionElementChildrenChanged();

    void setRecalcListItems();
    void invalidateSelectedItems();
    void updateListItemSelectedStates(AllowStyleInvalidation = AllowStyleInvalidation::Yes);

    const Vector<WeakPtr<HTMLElement>>& listItems() const;

    void accessKeySetSelectedIndex(int);

    void setSize(int);

    // Called by the bindings for the unnamed index-setter.
    void setItem(unsigned index, HTMLOptionElement*, ExceptionCode&);
    void setLength(unsigned, ExceptionCode&);

    HTMLOptionElement* namedItem(const AtomicString& name);
    HTMLOptionElement* item(unsigned index);

    void scrollToSelection();

    void listBoxSelectItem(int listIndex, bool allowMultiplySelections, bool shift, bool fireOnChangeNow = true);

    bool canSelectAll() const;
    void selectAll();
    int listToOptionIndex(int listIndex) const;
    void listBoxOnChange();
    int optionToListIndex(int optionIndex) const;
    int activeSelectionStartListIndex() const;
    int activeSelectionEndListIndex() const;
    void setActiveSelectionAnchorIndex(int);
    void setActiveSelectionEndIndex(int);
    void updateListBoxSelection(bool deselectOtherOptions);
    
    // For use in the implementation of HTMLOptionElement.
    void optionSelectionStateChanged(HTMLOptionElement&, bool optionIsSelected);
    
protected:
    HTMLSelectElement(const QualifiedName&, Document*, HTMLFormElement*);

private:
    virtual const AtomicString& formControlType() const FINAL;
    
    virtual bool isKeyboardFocusable(KeyboardEvent*) const FINAL;
    virtual bool isMouseFocusable() const FINAL;

    virtual void dispatchFocusEvent(PassRefPtr<Element> oldFocusedElement, const FocusOptions&) FINAL;
    virtual void dispatchBlurEvent(PassRefPtr<Element> newFocusedElement) FINAL;
    
    virtual bool canStartSelection() const FINAL { return false; }

    virtual bool canHaveUserAgentShadowRoot() const FINAL { return true; }

    virtual bool isEnumeratable() const FINAL { return true; }
    virtual bool supportLabels() const FINAL { return true; }

    bool isInteractiveContent() const final { return true; }

    virtual FormControlState saveFormControlState() const FINAL;
    virtual void restoreFormControlState(const FormControlState&) FINAL;

    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason = ModifiedDirectly) final;
    virtual bool isPresentationAttribute(const QualifiedName&) const FINAL;

    virtual bool childShouldCreateRenderer(const Node*) const FINAL;
    virtual RenderPtr<RenderElement> createElementRenderer(PassOwnPtr<RenderStyle>, const RenderTreePosition&) FINAL;
    virtual bool appendFormData(FormDataList&, bool) FINAL;

    virtual void reset() FINAL;

    virtual void defaultEventHandler(Event*) FINAL;
    virtual bool willRespondToMouseClickEvents() FINAL;

    void dispatchChangeEventForMenuList();
    
    void recalcListItems(bool updateSelectedStates = true, AllowStyleInvalidation = AllowStyleInvalidation::Yes) const;

    void deselectItems(HTMLOptionElement* excludeElement = nullptr);
    void typeAheadFind(KeyboardEvent*);
    void saveLastSelection();

    virtual InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode&) FINAL;

    virtual bool isOptionalFormControl() const FINAL { return !isRequiredFormControl(); }
    virtual bool isRequiredFormControl() const FINAL;

    bool hasPlaceholderLabelOption() const;

    enum SelectOptionFlag {
        DeselectOtherOptions = 1 << 0,
        DispatchChangeEvent = 1 << 1,
        UserDriven = 1 << 2,
    };
    typedef unsigned SelectOptionFlags;
    void selectOption(int optionIndex, SelectOptionFlags = 0);
    void deselectItemsWithoutValidation(HTMLElement* elementToExclude = nullptr);
    void parseMultipleAttribute(const AtomicString&);
    int lastSelectedListIndex() const;
    void updateSelectedState(int listIndex, bool multi, bool shift);
    void menuListDefaultEventHandler(Event*);
    bool platformHandleKeydownEvent(KeyboardEvent*);
    void listBoxDefaultEventHandler(Event*);
    void setOptionsChangedOnRenderer();
    size_t searchOptionsForValue(const String&, size_t listIndexStart, size_t listIndexEnd) const;

    enum SkipDirection { SkipBackwards = -1, SkipForwards = 1 };
    int nextValidIndex(int listIndex, SkipDirection, int skip) const;
    int nextSelectableListIndex(int startIndex) const;
    int previousSelectableListIndex(int startIndex) const;
    int firstSelectableListIndex() const;
    int lastSelectableListIndex() const;
    int nextSelectableListIndexPageAway(int startIndex, SkipDirection) const;

    virtual void childrenChanged(const ChildChange&) FINAL;

    // TypeAheadDataSource functions.
    virtual int indexOfSelectedOption() const FINAL;
    virtual int optionCount() const FINAL;
    virtual String optionAtIndex(int index) const FINAL;

    // m_listItems contains HTMLOptionElement, HTMLOptGroupElement, and HTMLHRElement objects.
    mutable Vector<WeakPtr<HTMLElement>> m_listItems;
    Vector<bool> m_lastOnChangeSelection;
    Vector<bool> m_cachedStateForActiveSelection;
    TypeAhead m_typeAhead;
    int m_size;
    int m_lastOnChangeIndex;
    int m_activeSelectionAnchorIndex;
    int m_activeSelectionEndIndex;
    bool m_isProcessingUserDrivenChange;
    bool m_multiple;
    bool m_activeSelectionState;
    mutable bool m_shouldRecalcListItems;
};

inline bool isHTMLSelectElement(const Node& node)
{
    return node.hasTagName(HTMLNames::selectTag);
}

inline bool isHTMLSelectElement(const Element& element)
{
    return element.hasTagName(HTMLNames::selectTag);
}

inline bool isHTMLSelectElement(const HTMLElement& element)
{
    return element.hasTagName(HTMLNames::selectTag);
}

NODE_TYPE_CASTS(HTMLSelectElement)

template <typename ArgType>
struct ElementTypeCastTraits<const HTMLSelectElement, ArgType> {
    static bool is(ArgType& node) { return isHTMLSelectElement(node); }
};

} // namespace
