/*
 * Copyright (C) 2007, 2008, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"

#include "HTMLSourceElement.h"

#include "Event.h"
#include "EventNames.h"
#include "HTMLImageElement.h"
#include "HTMLDocument.h"
#include "HTMLMediaElement.h"
#include "HTMLNames.h"
#include "HTMLPictureElement.h"
#include "Logging.h"
#include "MediaQueryParser.h"

using namespace std;

namespace WebCore {

using namespace HTMLNames;

inline HTMLSourceElement::HTMLSourceElement(const QualifiedName& tagName, Document* document)
    : HTMLElement(tagName, document)
	, ActiveDOMObject(*document)
    , m_errorEventTimer(this, &HTMLSourceElement::errorEventTimerFired)
{
    LOG(Media, "HTMLSourceElement::HTMLSourceElement - %p", this);
    ASSERT(hasTagName(sourceTag));
}

PassRefPtr<HTMLSourceElement> HTMLSourceElement::create(const QualifiedName& tagName, Document* document)
{
	RefPtr<HTMLSourceElement> sourceElement = adoptRef(new HTMLSourceElement(tagName, document));
	sourceElement->suspendIfNeeded();
	return sourceElement.release();
}

Node::InsertedIntoResult HTMLSourceElement::insertedIntoAncestor(const InsertionType& insertionType, ContainerNode& parentOfInsertedTree)
{
    HTMLElement::insertedIntoAncestor(insertionType, parentOfInsertedTree);
    RefPtr<Element> parent = parentElement();
    if (parent == &parentOfInsertedTree) {
#if ENABLE(VIDEO)
    	if(parent->isMediaElement())
    		static_cast<HTMLMediaElement*>(parentNode())->sourceWasAdded(this);
    	else
#endif
		if (isHTMLPictureElement(*parent)) {
			// The new source element only is a relevant mutation if it precedes any img element.
			m_shouldCallSourcesChanged = true;
			for (const Node* node = previousSibling(); node; node = node->previousSibling()) {
				if (isHTMLImageElement(*node))
					m_shouldCallSourcesChanged = false;
			}
			if (m_shouldCallSourcesChanged)
				toHTMLPictureElement(*parent).sourcesChanged();
		}
	}
    return InsertedIntoResult::Done;
}

void HTMLSourceElement::removedFromAncestor(const RemovalType& removalType, ContainerNode& parentOfRemovedTree)
{
    HTMLElement::removedFromAncestor(removalType, parentOfRemovedTree);
    if (!parentNode() && isElement(parentOfRemovedTree)) {
#if ENABLE(VIDEO)
        if (isHTMLMediaElement(parentOfRemovedTree))
    		toHTMLMediaElement(parentOfRemovedTree).sourceWasRemoved(this);
    	else
#endif
		if (m_shouldCallSourcesChanged) {
			toHTMLPictureElement(parentOfRemovedTree).sourcesChanged();
			m_shouldCallSourcesChanged = false;
        }
    }
}

void HTMLSourceElement::scheduleErrorEvent()
{
    LOG(Media, "HTMLSourceElement::scheduleErrorEvent - %p", this);
    if (m_errorEventTimer.isActive())
        return;

    m_errorEventTimer.startOneShot(0);
}

void HTMLSourceElement::cancelPendingErrorEvent()
{
    LOG(Media, "HTMLSourceElement::cancelPendingErrorEvent - %p", this);
    m_errorEventTimer.stop();
}

void HTMLSourceElement::errorEventTimerFired()
{
    LOG(Media, "HTMLSourceElement::errorEventTimerFired - %p", this);
    dispatchEvent(Event::create(eventNames().errorEvent, false, true));
}

bool HTMLSourceElement::isURLAttribute(const Attribute& attribute) const
{
    return attribute.name() == srcAttr || HTMLElement::isURLAttribute(attribute);
}

#if ENABLE(MICRODATA)
String HTMLSourceElement::itemValueText() const
{
    return getURLAttribute(srcAttr);
}

void HTMLSourceElement::setItemValueText(const String& value, ExceptionCode&)
{
    setAttribute(srcAttr, value);
}
#endif

const char* HTMLSourceElement::activeDOMObjectName() const
{
    return "HTMLSourceElement";
}

bool HTMLSourceElement::canSuspendForDocumentSuspension() const
{
    return true;
}

void HTMLSourceElement::suspend(ReasonForSuspension why)
{
    if (why == PageCache) {
        m_shouldRescheduleErrorEventOnResume = m_errorEventTimer.isActive();
        m_errorEventTimer.stop();
    }
}

void HTMLSourceElement::resume()
{
    if (m_shouldRescheduleErrorEventOnResume) {
        m_errorEventTimer.startOneShot(0);
        m_shouldRescheduleErrorEventOnResume = false;
    }
}

void HTMLSourceElement::stop()
{
    cancelPendingErrorEvent();
}

void HTMLSourceElement::attributeChanged(const QualifiedName& name, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason attributeModificationReason)
{
    HTMLElement::attributeChanged(name, oldValue, newValue, attributeModificationReason);
    if (name == srcsetAttr || name == sizesAttr || name == mediaAttr || name == typeAttr) {
        if (name == mediaAttr)
            m_mediaQuerySet = MediaQuerySet::create(newValue, MediaQueryParserContext(document()));
    	RefPtr<ContainerNode> parent = parentNode();
        if (m_shouldCallSourcesChanged && parent)
        	toHTMLPictureElement(*parent).sourcesChanged();
    } else if (name == widthAttr || name == heightAttr) {
		RefPtr<ContainerNode> parent = parentNode();
		if (parent && isHTMLPictureElement(*parent))
			toHTMLPictureElement(*parent).sourceDimensionAttributesChanged(*this);
	}
}

}
