/*
 * Copyright (C) 2011 Nokia Corporation and/or its subsidiary(-ies)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "HTMLSummaryElement.h"

#if ENABLE(DETAILS_ELEMENT)
#include "DetailsMarkerControl.h"
#include "HTMLDetailsElement.h"
#include "HTMLNames.h"
#include "HTMLSlotElement.h"
#include "KeyboardEvent.h"
#include "MouseEvent.h"
#include "PlatformMouseEvent.h"
#include "RenderBlockFlow.h"
#include "SVGAElement.h"
#include "ShadowRoot.h"
#include "SlotAssignment.h"

namespace WebCore {

using namespace HTMLNames;

class SummarySlotElement FINAL : public SlotAssignment {
private:
    virtual void hostChildElementDidChange(const Element&, ShadowRoot& shadowRoot) OVERRIDE
    {
        didChangeSlot(SlotAssignment::defaultSlotName(), shadowRoot);
    }

    virtual const AtomicString& slotNameForHostChild(const Node&) const OVERRIDE { return SlotAssignment::defaultSlotName(); }
};

Ref<HTMLSummaryElement> HTMLSummaryElement::create(const QualifiedName& tagName, Document* document)
{
    Ref<HTMLSummaryElement> summary = adoptRef(*new HTMLSummaryElement(tagName, document));
    summary->addShadowRoot(ShadowRoot::create(document, adoptPtr(new SummarySlotElement())));
	return summary;
}

inline HTMLSummaryElement::HTMLSummaryElement(const QualifiedName& tagName, Document* document)
    : HTMLElement(tagName, document)
{
    ASSERT(hasTagName(summaryTag));
}

RenderPtr<RenderElement> HTMLSummaryElement::createElementRenderer(PassOwnPtr<RenderStyle> style, const RenderTreePosition&)
{
    return RenderElement::createFor(*this, style, RenderElement::OnlyCreateBlockAndFlexboxRenderers);
}

void HTMLSummaryElement::didAddUserAgentShadowRoot(ShadowRoot* root)
{
    root->appendChild(DetailsMarkerControl::create(&document()), ASSERT_NO_EXCEPTION);
    root->appendChild(HTMLSlotElement::create(slotTag, &document()), ASSERT_NO_EXCEPTION);
}

HTMLDetailsElement* HTMLSummaryElement::detailsElement() const
{
    auto* parent = parentElement();
    if (parent && isHTMLDetailsElement(*parent))
        return toHTMLDetailsElement(parent);
    // Fallback summary element is in the shadow tree.
    auto* host = shadowHost();
    if (host && isHTMLDetailsElement(*host))
        return toHTMLDetailsElement(host);
    return nullptr;
}

bool HTMLSummaryElement::isActiveSummary() const
{
    HTMLDetailsElement* details = detailsElement();
    if (!details)
        return false;
    return details->isActiveSummary(*this);
}

static bool isInSummaryInteractiveContent(Node* target)
{
    if (!isElement(target))
        return false;
    for (RefPtr<Element> element = toElement(*target); element && !isHTMLSummaryElement(*element); element = element->parentOrShadowHostElement()) {
        if ((isHTMLElement(*element) && toHTMLElement(*element).isInteractiveContent()) || isSVGAElement(*element))
            return true;
    }
    return false;
}

int HTMLSummaryElement::defaultTabIndex() const
{
    return isActiveSummary() ? 0 : -1;
}

bool HTMLSummaryElement::supportsFocus() const
{
    return isActiveSummary() || HTMLElement::supportsFocus();
}

void HTMLSummaryElement::defaultEventHandler(Event* event)
{
    if (isActiveSummary()) {
        if (event->type() == eventNames().DOMActivateEvent && !isInSummaryInteractiveContent(event->target()->toNode().get())) {
            if (RefPtr<HTMLDetailsElement> details = detailsElement())
                details->toggleOpen();
            event->setDefaultHandled();
            return;
        }

        if (event->isKeyboardEvent()) {
            if (event->type() == eventNames().keydownEvent && static_cast<KeyboardEvent*>(event)->keyIdentifier() == "U+0020") {
                setActive(true, true);
                // No setDefaultHandled() - IE dispatches a keypress in this case.
                return;
            }
            if (event->type() == eventNames().keypressEvent) {
                switch (static_cast<KeyboardEvent*>(event)->charCode()) {
                case '\r':
                    dispatchSimulatedClick(event);
                    event->setDefaultHandled();
                    return;
                case ' ':
                    // Prevent scrolling down the page.
                    event->setDefaultHandled();
                    return;
                }
            }
            if (event->type() == eventNames().keyupEvent && static_cast<KeyboardEvent*>(event)->keyIdentifier() == "U+0020") {
                if (active())
                    dispatchSimulatedClick(event);
                event->setDefaultHandled();
                return;
            }
        }
    }

    HTMLElement::defaultEventHandler(event);
}

bool HTMLSummaryElement::willRespondToMouseClickEvents()
{
    return isActiveSummary() || HTMLElement::willRespondToMouseClickEvents();
}

}

#endif
