/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2010 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "HTMLTextFormControlElement.h"

namespace WebCore {

class BeforeTextInsertedEvent;
class VisibleSelection;

class HTMLTextAreaElement FINAL : public HTMLTextFormControlElement {
public:
    WEBCORE_EXPORT static Ref<HTMLTextAreaElement> create(Document&);
    static Ref<HTMLTextAreaElement> create(const QualifiedName&, Document*, HTMLFormElement*);

    int cols() const { return m_cols; }
    int rows() const { return m_rows; }

    bool shouldWrapText() const { return m_wrap != NoWrap; }

    virtual String value() const FINAL;
    void setValue(const String&);
    String defaultValue() const;
    void setDefaultValue(const String&);
    int textLength() const { return value().length(); }
    int effectiveMaxLength() const { return maxLength(); }
    // For ValidityState
    virtual String validationMessage() const FINAL;
    virtual bool valueMissing() const FINAL;
    virtual bool tooShort() const FINAL;
    virtual bool tooLong() const FINAL;
    bool isValidValue(const String&) const;
    
    virtual RefPtr<TextControlInnerTextElement> innerTextElement() const FINAL;
    virtual PassOwnPtr<RenderStyle> createInnerTextStyle(const RenderStyle&) FINAL;

    void setCols(int);
    void setRows(int);

    String characterType() const;
    
private:
    HTMLTextAreaElement(const QualifiedName&, Document*, HTMLFormElement*);

    enum WrapMethod { NoWrap, SoftWrap, HardWrap };

    virtual void didAddUserAgentShadowRoot(ShadowRoot*) FINAL;
    virtual bool canHaveUserAgentShadowRoot() const FINAL { return true; }

    void maxLengthAttributeChanged(const AtomicString& newValue);
    void minLengthAttributeChanged(const AtomicString& newValue);

    void handleBeforeTextInsertedEvent(BeforeTextInsertedEvent*) const;
    static String sanitizeUserInputValue(const String&, unsigned maxLength);
    void updateValue() const;
    void setNonDirtyValue(const String&);
    void setValueCommon(const String&);

    virtual bool supportsPlaceholder() const FINAL { return true; }
    virtual HTMLElement* placeholderElement() const FINAL;
    virtual void updatePlaceholderText() FINAL;
    virtual bool isEmptyValue() const FINAL { return value().isEmpty(); }

    virtual bool isOptionalFormControl() const FINAL { return !isRequiredFormControl(); }
    virtual bool isRequiredFormControl() const FINAL { return required(); }

    virtual void defaultEventHandler(Event*) FINAL;
    
    virtual void subtreeHasChanged() FINAL;

    virtual bool isEnumeratable() const FINAL { return true; }
    virtual bool supportLabels() const FINAL { return true; }

    bool isInteractiveContent() const final { return true; }

    virtual const AtomicString& formControlType() const FINAL;

    virtual FormControlState saveFormControlState() const FINAL;
    virtual void restoreFormControlState(const FormControlState&) FINAL;

    virtual bool isTextField() const FINAL { return true; }

    virtual void childrenChanged(const ChildChange&) FINAL;
    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason = ModifiedDirectly) final;
    virtual bool isPresentationAttribute(const QualifiedName&) const FINAL;
    virtual void collectStyleForPresentationAttribute(const QualifiedName&, const AtomicString&, MutableStylePropertySet*) override;
    virtual RenderPtr<RenderElement> createElementRenderer(PassOwnPtr<RenderStyle>, const RenderTreePosition&) FINAL;
    virtual bool appendFormData(FormDataList&, bool) FINAL;
    virtual void reset() FINAL;
    virtual bool hasCustomFocusLogic() const FINAL;
    virtual bool isMouseFocusable() const FINAL;
    virtual bool isKeyboardFocusable(KeyboardEvent*) const FINAL;
    virtual void updateFocusAppearance(SelectionRestorationMode, SelectionRevealMode) FINAL;

    virtual bool accessKeyAction(bool sendMouseEvents) final;

    virtual bool shouldUseInputMethod() FINAL;
    virtual bool matchesReadWritePseudoClass() const FINAL;

    bool valueMissing(const String& value) const { return isRequiredFormControl() && !isDisabledOrReadOnly() && value.isEmpty(); }
    bool tooShort(StringView, NeedsToCheckDirtyFlag) const;
	bool tooLong(StringView, NeedsToCheckDirtyFlag) const;

    int m_rows;
    int m_cols;
    WrapMethod m_wrap;
    RefPtr<HTMLElement> m_placeholder;
    mutable String m_value;
    mutable bool m_isDirty;
    mutable bool m_wasModifiedByUser;
};

inline bool isHTMLTextAreaElement(const Node& node)
{
    return node.hasTagName(HTMLNames::textareaTag);
}

inline bool isHTMLTextAreaElement(const Element& element)
{
    return element.hasTagName(HTMLNames::textareaTag);
}

inline HTMLTextAreaElement* toHTMLTextAreaElement(Node* node)
{
    return static_cast<HTMLTextAreaElement*>(node);
}

inline HTMLTextAreaElement& toHTMLTextAreaElement(Node& node)
{
    return static_cast<HTMLTextAreaElement&>(node);
}

} //namespace
