/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2009, 2010, 2011 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "HTMLFormControlElementWithState.h"

namespace WebCore {

class Position;
class RenderTextControl;
class TextControlInnerTextElement;
class VisiblePosition;

struct SimpleRange;

enum class AutoFillButtonType : uint8_t { None, Credentials, Contacts, StrongConfirmationPassword, StrongPassword, CreditCard };
enum TextFieldSelectionDirection { SelectionHasNoDirection, SelectionHasForwardDirection, SelectionHasBackwardDirection };
enum TextFieldEventBehavior { DispatchNoEvent, DispatchChangeEvent, DispatchInputAndChangeEvent };

class HTMLTextFormControlElement : public HTMLFormControlElementWithState {
public:
    // Common flag for HTMLInputElement::tooLong() / tooShort() and HTMLTextAreaElement::tooLong().
    enum NeedsToCheckDirtyFlag {CheckDirtyFlag, IgnoreDirtyFlag};

    virtual ~HTMLTextFormControlElement();

    void didEditInnerTextValue();
    void forwardEvent(Event*);

    long maxLength() const { return m_maxLength; }
    WEBCORE_EXPORT void setMaxLength(int, ExceptionCode&);
    void setMaxLength(int maxLength) { m_maxLength = maxLength; }
    long minLength() const { return m_minLength; }
    void setMinLength(int, ExceptionCode&);
    void setMinLength(int minLength) { m_minLength = minLength; }

    virtual InsertedIntoResult insertedIntoAncestor(const InsertionType&, ContainerNode&) OVERRIDE;

    // The derived class should return true if placeholder processing is needed.
    bool isPlaceholderVisible() const { return m_isPlaceholderVisible; }
    virtual bool supportsPlaceholder() const = 0;
    virtual HTMLElement* placeholderElement() const = 0;
    void updatePlaceholderVisibility();

    int indexForVisiblePosition(const VisiblePosition&) const;
    VisiblePosition visiblePositionForIndex(int index) const;
    int selectionStart() const;
    int selectionEnd() const;
    const AtomicString& selectionDirection() const;
    void setSelectionStart(int);
    void setSelectionEnd(int);
    void setSelectionDirection(const String&);
    void select(SelectionRevealMode = SelectionRevealMode::DoNotReveal);
    virtual void setRangeText(const String& replacement, ExceptionCode&);
	virtual void setRangeText(const String& replacement, unsigned start, unsigned end, const String& selectionMode, ExceptionCode&);
	void setSelectionRange(int start, int end, const String& direction);
    void setSelectionRange(int start, int end, TextFieldSelectionDirection = SelectionHasNoDirection, SelectionRevealMode = SelectionRevealMode::DoNotReveal);

    Optional<SimpleRange> selection() const;
    String selectedText() const;

    virtual void dispatchFormControlChangeEvent() final;

    virtual String value() const = 0;

    virtual RefPtr<TextControlInnerTextElement> innerTextElement() const = 0;
    virtual PassOwnPtr<RenderStyle> createInnerTextStyle(const RenderStyle&) = 0;

    void selectionChanged(bool shouldFireSelectEvent);
    bool lastChangeWasUserEdit() const;
    void setInnerTextValue(const String&);
    String innerTextValue() const;

    String directionForFormData() const;

    void setTextAsOfLastFormControlChangeEvent(const String& text) { m_textAsOfLastFormControlChangeEvent = text; }

protected:
    HTMLTextFormControlElement(const QualifiedName&, Document*, HTMLFormElement*);
    virtual bool isPlaceholderEmpty() const;
    virtual void updatePlaceholderText() = 0;

    virtual void attributeChanged(const QualifiedName&, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason) override;

    virtual void disabledStateChanged() OVERRIDE;
    virtual void readOnlyStateChanged() OVERRIDE;
    virtual bool isInnerTextElementEditable() const;
    void updateInnerTextElementEditability();

    void cacheSelection(int start, int end, TextFieldSelectionDirection direction)
    {
        m_cachedSelectionStart = start;
        m_cachedSelectionEnd = end;
        m_cachedSelectionDirection = direction;
    }

    void restoreCachedSelection(SelectionRevealMode);
    bool hasCachedSelection() const { return m_cachedSelectionStart >= 0; }

    virtual void subtreeHasChanged() = 0;

    void setLastChangeWasNotUserEdit() { m_lastChangeWasUserEdit = false; }

    String valueWithHardLineBreaks() const;

    void adjustInnerTextStyle(const RenderStyle& parentStyle, RenderStyle& textBlockStyle) const;

private:
    TextFieldSelectionDirection cachedSelectionDirection() const { return static_cast<TextFieldSelectionDirection>(m_cachedSelectionDirection); }

    virtual bool isTextFormControlElement() const FINAL { return true; }

    int computeSelectionStart() const;
    int computeSelectionEnd() const;
    TextFieldSelectionDirection computeSelectionDirection() const;

    virtual void dispatchFocusEvent(PassRefPtr<Element> oldFocusedElement, const FocusOptions&) override;
    virtual void dispatchBlurEvent(PassRefPtr<Element> newFocusedElement) override;
    virtual bool childShouldCreateRenderer(const Node*) const OVERRIDE;

    unsigned indexForPosition(const Position&) const;

    // Returns true if user-editable value is empty. Used to check placeholder visibility.
    virtual bool isEmptyValue() const = 0;
    // Returns true if suggested value is empty. Used to check placeholder visibility.
    virtual bool isEmptySuggestedValue() const { return true; }
    // Called in dispatchFocusEvent(), after placeholder process, before calling parent's dispatchFocusEvent().
    virtual void handleFocusEvent(Node* /* oldFocusedNode */, FocusDirection) { }
    // Called in dispatchBlurEvent(), after placeholder process, before calling parent's dispatchBlurEvent().
    virtual void handleBlurEvent() { }

    bool placeholderShouldBeVisible() const;

    RenderTextControl* textRendererAfterUpdateLayout();

    String m_textAsOfLastFormControlChangeEvent;
    
    int m_cachedSelectionStart;
    int m_cachedSelectionEnd;

    int m_maxLength;
    int m_minLength;

    unsigned char m_cachedSelectionDirection;
    unsigned char m_lastChangeWasUserEdit;
    unsigned char m_isPlaceholderVisible;
};

inline bool isHTMLTextFormControlElement(const Node& node)
{
    return node.isElementNode() && toElement(node).isTextFormControlElement();
}

inline bool isHTMLTextFormControlElement(const Element& element)
{
    return element.isTextFormControlElement();
}

ALWAYS_INLINE HTMLTextFormControlElement& toHTMLTextFormControlElement(EventTarget& target)
{
    ASSERT(isHTMLTextFormControlElement(target));
    return static_cast<HTMLTextFormControlElement&>(target);
}

ALWAYS_INLINE const HTMLTextFormControlElement& toHTMLTextFormControlElement(const EventTarget& target)
{
    ASSERT(isHTMLTextFormControlElement(target));
    return static_cast<const HTMLTextFormControlElement&>(target);
}

template<>
ALWAYS_INLINE HTMLTextFormControlElement* dynamicDowncast(HTMLElement* source)
{
    //static_assert(!std::is_same_v<Source, Target>, "Unnecessary cast to same type");
    //static_assert(std::is_base_of_v<Source, Target>, "Should be a downcast");
    return source && isHTMLTextFormControlElement(*source) ? static_cast<HTMLTextFormControlElement*>(source) : nullptr;
}

NODE_TYPE_CASTS(HTMLTextFormControlElement)

HTMLTextFormControlElement* enclosingTextFormControl(const Position&);

} // namespace
