/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"
#include "LinkRelAttribute.h"
#include <wtf/text/WTFString.h>

#include "RuntimeEnabledFeatures.h"

namespace WebCore {

LinkRelAttribute::LinkRelAttribute()
    : m_isStyleSheet(false)
    , m_isAlternate(false)
    , m_isDNSPrefetch(false)
	, isLinkPreload(false)
#if ENABLE(LINK_PREFETCH)
    , m_isLinkPrefetch(false)
    , m_isLinkSubresource(false)
#endif
#if ENABLE(LINK_PRERENDER)
    , m_isLinkPrerender(false)
#endif
{
}

// Keep LinkRelAttribute::isSupported() in sync when updating this constructor.
LinkRelAttribute::LinkRelAttribute(const String& rel)
    : m_isStyleSheet(false)
    , m_isAlternate(false)
    , m_isDNSPrefetch(false)
	, isLinkPreload(false)
#if ENABLE(LINK_PREFETCH)
    , m_isLinkPrefetch(false)
    , m_isLinkSubresource(false)
#endif
#if ENABLE(LINK_PRERENDER)
    , m_isLinkPrerender(false)
#endif
{
    if (equalLettersIgnoringASCIICase(rel, "stylesheet"))
        m_isStyleSheet = true;
    else if (equalLettersIgnoringASCIICase(rel, "icon") || equalLettersIgnoringASCIICase(rel, "shortcut icon"))
        iconType = LinkIconType::Favicon;
#if ENABLE(TOUCH_ICON_LOADING)
    else if (equalLettersIgnoringASCIICase(rel, "apple-touch-icon"))
        iconType = LinkIconType::TouchIcon;
    else if (equalLettersIgnoringASCIICase(rel, "apple-touch-icon-precomposed"))
        iconType = LinkIconType::TouchPrecomposedIcon;
#endif
    else if (equalLettersIgnoringASCIICase(rel, "dns-prefetch"))
        m_isDNSPrefetch = true;
    else if (RuntimeEnabledFeatures::linkPreloadEnabled() && equalLettersIgnoringASCIICase(rel, "preload"))
        isLinkPreload = true;
    else if (equalLettersIgnoringASCIICase(rel, "alternate stylesheet") || equalLettersIgnoringASCIICase(rel, "stylesheet alternate")) {
        m_isStyleSheet = true;
        m_isAlternate = true;
    } else {
        // Tokenize the rel attribute and set bits based on specific keywords that we find.
        String relCopy = rel;
        relCopy.replace('\n', ' ');
        Vector<String> list;
        relCopy.split(' ', list);
        Vector<String>::const_iterator end = list.end();
        for (Vector<String>::const_iterator it = list.begin(); it != end; ++it) {
            if (equalLettersIgnoringASCIICase(*it, "stylesheet"))
                m_isStyleSheet = true;
            else if (equalLettersIgnoringASCIICase(*it, "alternate"))
                m_isAlternate = true;
            else if (equalLettersIgnoringASCIICase(*it, "icon"))
                iconType = LinkIconType::Favicon;
#if ENABLE(TOUCH_ICON_LOADING)
            else if (equalLettersIgnoringASCIICase(*it, "apple-touch-icon"))
                iconType = LinkIconType::TouchIcon;
            else if (equalLettersIgnoringASCIICase(*it, "apple-touch-icon-precomposed"))
                iconType = LinkIconType::TouchPrecomposedIcon;
#endif
#if ENABLE(LINK_PREFETCH)
            else if (equalLettersIgnoringASCIICase(*it, "prefetch"))
              m_isLinkPrefetch = true;
            else if (equalLettersIgnoringASCIICase(*it, "subresource"))
              m_isLinkSubresource = true;
#endif
#if ENABLE(LINK_PRERENDER)
            else if (equalLettersIgnoringASCIICase(*it, "prerender"))
              m_isLinkPrerender = true;
#endif
        }
    }
}

// https://html.spec.whatwg.org/#linkTypes
bool LinkRelAttribute::isSupported(StringView attribute)
{
    static const char* supportedAttributes[] = {
        "alternate", "dns-prefetch", "icon", "stylesheet", "apple-touch-icon", "apple-touch-icon-precomposed",
#if ENABLE(LINK_PREFETCH)
        "prefetch", "subresource",
#endif
    };

    for (auto* supportedAttribute : supportedAttributes) {
        if (equalIgnoringASCIICase(attribute, supportedAttribute))
            return true;
    }

    if (RuntimeEnabledFeatures::linkPreloadEnabled() && equalLettersIgnoringASCIICase(attribute, "preload"))
        return true;

    return false;
}

}
