/*
 * Copyright (C) 2006, 2007, 2008, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "PDFDocument.h"

#if HAVE(PDFIUM)
#include "Chrome.h"
#include "DocumentLoader.h"
#include "EventListener.h"
#include "EventNames.h"
#include "Frame.h"
#include "FrameLoaderClient.h"
#include "FrameView.h"
#include "HTMLBodyElement.h"
#include "HTMLHeadElement.h"
#include "HTMLHtmlElement.h"
#include "HTMLObjectElement.h"
#include "HTMLNames.h"
#include "LocalizedStrings.h"
#include "MIMETypeRegistry.h"
#include "MouseEvent.h"
#include "NotImplemented.h"
#include "Page.h"
#include "RawDataDocumentParser.h"
#include "RenderElement.h"
#include "SharedBuffer.h"
#include "Settings.h"

namespace WebCore {

using namespace HTMLNames;

namespace {
class PDFDocumentParser FINAL : public RawDataDocumentParser {
public:
    static PassRefPtr<PDFDocumentParser> create(PDFDocument* document)
    {
        return adoptRef(new PDFDocumentParser(document));
    }
    
private:
    PDFDocumentParser(PDFDocument* document)
        : RawDataDocumentParser(document)
    {
    }

    PDFDocument& document() const;

    virtual void appendBytes(DocumentWriter*, const char*, size_t) OVERRIDE;
    virtual void finish() OVERRIDE;
};

inline PDFDocument& PDFDocumentParser::document() const
{
    // Only used during parsing, so document is guaranteed to be non-null.
    ASSERT(RawDataDocumentParser::document());
    return *toPDFDocument(RawDataDocumentParser::document());
}

void PDFDocumentParser::appendBytes(DocumentWriter*, const char*, size_t)
{
}

void PDFDocumentParser::finish()
{
    document().finishedParsing();
}

}

void PDFDocument::finishedParsing()
{
    HTMLDocument::finishedParsing();

    createDocumentStructure();
}

// --------

inline PDFDocument::PDFDocument(Frame& frame, const KURL& url)
    : HTMLDocument(&frame, url, PDFDocumentClass)
{
    setCompatibilityMode(QuirksMode);
    lockCompatibilityMode();
}

Ref<PDFDocument> PDFDocument::create(Frame& frame, const KURL& url)
{
    return adoptRef(*new PDFDocument(frame, url));
}

PassRefPtr<DocumentParser> PDFDocument::createParser()
{
    return PDFDocumentParser::create(this);
}

void PDFDocument::createDocumentStructure()
{
    auto rootElement = HTMLHtmlElement::create(this);
    appendChild(rootElement);
    rootElement->insertedByParser();

    if (frame() && frame()->loader())
        frame()->loader()->dispatchDocumentElementAvailable();

    // We need a <head> so that the call to setTitle() later on actually has an <head> to append to <title> to.
    auto head = HTMLHeadElement::create(this);
    rootElement->appendChild(head);

    auto body = HTMLBodyElement::create(this);
    body->setAttribute(styleAttr, "margin: 0px");
    if (MIMETypeRegistry::isPDFMIMEType(document().loader()->responseMIMEType()))
        body->setInlineStyleProperty(CSSPropertyBackgroundColor, "white");
    rootElement->appendChild(body);
    
    Ref<Element> pluginElement = createElement(objectTag, false);
    pluginElement->setAttribute(typeAttr, "application/pdf");
    pluginElement->setAttribute(dataAttr, url().string());
    pluginElement->setAttribute(styleAttr, "width:100%; height:100%;");
    body->appendChild(pluginElement);

    m_objectElement = makeWeakPtr(pluginElement.get());
}

}

#endif // HAVE(PDFIUM)
