/*
 * Copyright (C) 2006, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2008, 2010 Nokia Corporation and/or its subsidiary(-ies)
 * Copyright (C) 2007 Alp Toker <alp@atoker.com>
 * Copyright (C) 2008 Eric Seidel <eric@webkit.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "CanvasStyle.h"

#include "CSSParser.h"
#include "CSSPropertyNames.h"
#include "CanvasGradient.h"
#include "CanvasPattern.h"
#include "GraphicsContext.h"
#include "HTMLCanvasElement.h"
#include <wtf/Assertions.h>
#include <wtf/PassRefPtr.h>

#if USE(CG)
#include <CoreGraphics/CGContext.h>
#endif

#if PLATFORM(QT)
#include <QPainter>
#include <QBrush>
#include <QPen>
#include <QColor>
#endif

namespace WebCore {

enum ColorParseResult { ParsedRGBA, ParsedCurrentColor, ParsedSystemColor, ParseFailed };

bool isCurrentColorString(const String& colorString)
{
    return equalLettersIgnoringASCIICase(colorString, "currentcolor");
}

Color parseColor(const String& colorString, HTMLCanvasElement& canvasBase)
{
#if ENABLE(OFFSCREEN_CANVAS)
    if (is<OffscreenCanvas>(canvasBase))
        return CSSPropertyParserWorkerSafe::parseColor(colorString);
#endif
    Color color;
    if (isHTMLCanvasElement(canvasBase))
        color = CSSParser::parseColor(colorString, CSSParserContext(canvasBase.document()));
    else
        color = CSSParser::parseColorWithoutContext(colorString);
    if (color.isValid())
        return color;
    return CSSParser::parseSystemColor(colorString);
}

Color parseColor(const String& colorString)
{
    Color color = CSSParser::parseColorWithoutContext(colorString);
    if (color.isValid())
        return color;
    return CSSParser::parseSystemColor(colorString);
}

Color currentColor(HTMLCanvasElement& canvasBase)
{
    if (!isHTMLCanvasElement(canvasBase))
        return Color::black;
    auto& canvas = canvasBase;
    if (!canvas.isConnected() || !canvas.inlineStyle())
        return Color::black;
    Color color = CSSParser::parseColorWithoutContext(canvas.inlineStyle()->getPropertyValue(CSSPropertyColor));
    if (!color.isValid())
        return Color::black;
    return color;
}
Color parseColorOrCurrentColor(const String& colorString, HTMLCanvasElement& canvasBase)
{
    if (isCurrentColorString(colorString))
        return currentColor(canvasBase);
    return parseColor(colorString, canvasBase);
}

CanvasStyle::CanvasStyle(const Color& color)
    : m_rgba(color.rgba())
    , m_type(RGBA)
{
}

CanvasStyle::CanvasStyle(float grayLevel, float alpha)
    : m_rgba(makeRGBA32FromFloats(grayLevel, grayLevel, grayLevel, alpha))
    , m_type(RGBA)
{
}

CanvasStyle::CanvasStyle(float r, float g, float b, float a)
    : m_rgba(makeRGBA32FromFloats(r, g, b, a))
    , m_type(RGBA)
{
}

CanvasStyle::CanvasStyle(float c, float m, float y, float k, float a)
    : m_cmyka(new CMYKAValues(makeRGBAFromCMYKA(c, m, y, k, a), c, m, y, k, a))
    , m_type(CMYKA)
{
}

CanvasStyle::CanvasStyle(PassRefPtr<CanvasGradient> gradient)
    : m_gradient(gradient.leakRef())
    , m_type(Gradient)
{
    if (!m_gradient)
        m_type = Invalid;
}

CanvasStyle::CanvasStyle(PassRefPtr<CanvasPattern> pattern)
    : m_pattern(pattern.leakRef())
    , m_type(ImagePattern)
{
    if (!m_pattern)
        m_type = Invalid;
}

CanvasStyle::~CanvasStyle()
{
    if (m_type == Gradient)
        m_gradient->deref();
    else if (m_type == ImagePattern)
        m_pattern->deref();
    else if (m_type == CMYKA)
        delete m_cmyka;
}

CanvasStyle CanvasStyle::createFromString(const String& colorString, HTMLCanvasElement& canvasBase)
{
    if (isCurrentColorString(colorString))
        return CanvasStyle(ConstructCurrentColor);

    Color color = parseColor(colorString, canvasBase);
    if (!color.isValid())
        return CanvasStyle();

	return CanvasStyle(color.rgb());
}

CanvasStyle CanvasStyle::createFromStringWithOverrideAlpha(const String& colorString, float alpha, HTMLCanvasElement& canvasBase)
{
    if (isCurrentColorString(colorString))
        return CanvasStyle(CurrentColorWithOverrideAlpha, alpha);

    Color color = parseColor(colorString, canvasBase);
    if (!color.isValid())
        return CanvasStyle();

	return CanvasStyle(colorWithOverrideAlpha(color, alpha));
}

bool CanvasStyle::isEquivalentColor(const CanvasStyle& other) const
{
    if (m_type != other.m_type)
        return false;

    switch (m_type) {
    case RGBA:
        return m_rgba == other.m_rgba;
    case CMYKA:
        return m_cmyka->c == other.m_cmyka->c
            && m_cmyka->m == other.m_cmyka->m
            && m_cmyka->y == other.m_cmyka->y
            && m_cmyka->k == other.m_cmyka->k
            && m_cmyka->a == other.m_cmyka->a;
    case Gradient:
    case ImagePattern:
    case CurrentColor:
    case CurrentColorWithOverrideAlpha:
        return false;
    case Invalid:
        break;
    }

    ASSERT_NOT_REACHED();
    return false;
}

bool CanvasStyle::isEquivalentRGBA(float r, float g, float b, float a) const
{
    if (m_type != RGBA)
        return false;

    return m_rgba == makeRGBA32FromFloats(r, g, b, a);
}

bool CanvasStyle::isEquivalentCMYKA(float c, float m, float y, float k, float a) const
{
    if (m_type != CMYKA)
        return false;

    return c == m_cmyka->c
        && m == m_cmyka->m
        && y == m_cmyka->y
        && k == m_cmyka->k
        && a == m_cmyka->a;
}

CanvasStyle::CanvasStyle(const CanvasStyle& other)
{
    memcpy(this, &other, sizeof(CanvasStyle));
    if (m_type == Gradient)
        m_gradient->ref();
    else if (m_type == ImagePattern)
        m_pattern->ref();
    else if (m_type == CMYKA)
        m_cmyka = new CMYKAValues(other.m_cmyka->rgba, other.m_cmyka->c, other.m_cmyka->m, other.m_cmyka->y, other.m_cmyka->k, other.m_cmyka->a);
}

CanvasStyle& CanvasStyle::operator=(const CanvasStyle& other)
{
    if (this != &other) {
        this->~CanvasStyle();
        new (this) CanvasStyle(other);
    }
    return *this;
}

void CanvasStyle::applyStrokeColor(GraphicsContext* context) const
{
    if (!context)
        return;
    switch (m_type) {
    case RGBA:
        context->setStrokeColor(m_rgba);
        break;
    case CMYKA: {
        // FIXME: Do this through platform-independent GraphicsContext API.
        // We'll need a fancier Color abstraction to support CMYKA correctly
#if USE(CG)
        CGContextSetCMYKStrokeColor(context->platformContext(), m_cmyka->c, m_cmyka->m, m_cmyka->y, m_cmyka->k, m_cmyka->a);
#elif PLATFORM(QT)
        QPen currentPen = context->platformContext()->pen();
        QColor clr;
        clr.setCmykF(m_cmyka->c, m_cmyka->m, m_cmyka->y, m_cmyka->k, m_cmyka->a);
	   currentPen.setColor(clr);
        context->platformContext()->setPen(currentPen);
#else
        context->setStrokeColor(m_cmyka->rgba);
#endif
        break;
    }
    case Gradient:
        context->setStrokeGradient(canvasGradient()->gradient(), AffineTransform());
        break;
    case ImagePattern:
        context->setStrokePattern(canvasPattern()->pattern());
        break;
    case CurrentColor:
    case CurrentColorWithOverrideAlpha:
    case Invalid:
        ASSERT_NOT_REACHED();
        break;
    }
}

void CanvasStyle::applyFillColor(GraphicsContext* context) const
{
    if (!context)
        return;
    switch (m_type) {
    case RGBA:
        context->setFillColor(m_rgba);
        break;
    case CMYKA: {
        // FIXME: Do this through platform-independent GraphicsContext API.
        // We'll need a fancier Color abstraction to support CMYKA correctly
#if USE(CG)
        CGContextSetCMYKFillColor(context->platformContext(), m_cmyka->c, m_cmyka->m, m_cmyka->y, m_cmyka->k, m_cmyka->a);
#elif PLATFORM(QT)
        QBrush currentBrush = context->platformContext()->brush();
        QColor clr;
        clr.setCmykF(m_cmyka->c, m_cmyka->m, m_cmyka->y, m_cmyka->k, m_cmyka->a);
        currentBrush.setColor(clr);
        context->platformContext()->setBrush(currentBrush);
#else
        context->setFillColor(m_cmyka->rgba);
#endif
        break;
    }
    case Gradient:
        context->setFillGradient(canvasGradient()->gradient(), AffineTransform());
        break;
    case ImagePattern:
        context->setFillPattern(canvasPattern()->pattern());
        break;
    case CurrentColor:
    case CurrentColorWithOverrideAlpha:
    case Invalid:
        ASSERT_NOT_REACHED();
        break;
    }
}

}
