/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "NavigationAction.h"

#include "Document.h"
#include "Element.h"
#include "FrameLoader.h"
#include "FrameLoaderClient.h"
#include "HistoryItem.h"
#include "MouseEvent.h"

namespace WebCore {

NavigationAction::UIEventWithKeyStateData::UIEventWithKeyStateData(const UIEventWithKeyState& uiEvent)
    : isTrusted(uiEvent.isTrusted())
    , shiftKey(uiEvent.shiftKey())
    , ctrlKey(uiEvent.ctrlKey())
    , altKey(uiEvent.altKey())
    , metaKey(uiEvent.metaKey())
{
}

NavigationAction::MouseEventData::MouseEventData(const MouseEvent& mouseEvent)
    : UIEventWithKeyStateData(mouseEvent)
    , absoluteLocation(mouseEvent.absoluteLocation())
    , locationInRootViewCoordinates(mouseEvent.locationInRootViewCoordinates())
    , button(mouseEvent.button())
    , syntheticClickType(mouseEvent.syntheticClickType())
    , buttonDown(mouseEvent.buttonDown())
{
}

NavigationAction::NavigationAction() = default;
NavigationAction::~NavigationAction() = default;

static bool shouldTreatAsSameOriginNavigation(const Document& document, const KURL& url)
{
    return url.isAboutBlank() || url.protocolIsData() || (url.protocolIsBlob() && document.securityOrigin()->canRequest(url));
}

static Optional<NavigationAction::UIEventWithKeyStateData> keyStateDataForFirstEventWithKeyState(Event* event)
{
    if (UIEventWithKeyState* uiEvent = findEventWithKeyState(event))
        return NavigationAction::UIEventWithKeyStateData(*uiEvent);
    return Nullopt;
}

static Optional<NavigationAction::MouseEventData> mouseEventDataForFirstMouseEvent(Event* event)
{
    for (Event* e = event; e; e = e->underlyingEvent()) {
        if (e->isMouseEvent())
            return NavigationAction::MouseEventData(static_cast<const MouseEvent&>(*e));
    }
    return Nullopt;
}

NavigationAction::NavigationAction(Document& requester, const ResourceRequest& resourceRequest, NavigationType type, PassRefPtr<Event> event, const AtomicString& downloadAttribute)
	: m_requester(NavigationRequester::from(requester))
	, m_resourceRequest(resourceRequest)
    , m_type(type)
	, m_keyStateEventData(keyStateDataForFirstEventWithKeyState(event.get()))
	, m_mouseEventData(mouseEventDataForFirstMouseEvent(event.get()))
	, m_downloadAttribute(downloadAttribute)
{
}

static NavigationType navigationType(FrameLoadType frameLoadType, bool isFormSubmission, bool haveEvent)
{
    if (isFormSubmission)
        return NavigationTypeFormSubmitted;
    if (haveEvent)
        return NavigationTypeLinkClicked;
    if (frameLoadType == FrameLoadTypeReload || frameLoadType == FrameLoadTypeReloadFromOrigin)
        return NavigationTypeReload;
    if (isBackForwardLoadType(frameLoadType))
        return NavigationTypeBackForward;
    return NavigationTypeOther;
}

NavigationAction::NavigationAction(Document& requester, const ResourceRequest& resourceRequest, FrameLoadType frameLoadType,
	bool isFormSubmission, PassRefPtr<Event> event, const AtomicString& downloadAttribute)
    : m_requester(NavigationRequester::from(requester))
	, m_resourceRequest(resourceRequest)
    , m_type(navigationType(frameLoadType, isFormSubmission, !!event))
	, m_keyStateEventData(keyStateDataForFirstEventWithKeyState(event.get()))
	, m_mouseEventData(mouseEventDataForFirstMouseEvent(event.get()))
	, m_downloadAttribute(downloadAttribute)
{
}

}
