/*
 * Copyright (C) 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2009 Torch Mobile, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "CachedFont.h"

#include "CachedResourceClient.h"
#include "CachedResourceClientWalker.h"
#include "CachedResourceLoader.h"
#include "FontDescription.h"
#include "FontPlatformData.h"
#include "MemoryCache.h"
#include "SharedBuffer.h"
#include "TextResourceDecoder.h"
#include "WOFFFileFormat.h"
#include "WOFF2FileFormat.h"
#include <wtf/Vector.h>
#include "FontCustomPlatformData.h"

namespace WebCore {

CachedFont::CachedFont(CachedResourceRequest& request, Type type)
    : CachedResource(request, type)
    , m_loadInitiated(false)
	, m_hasCreatedFontDataWrappingResource(false)
	, m_fontPlatformDataDeletionTimer(this, &CachedFont::fontPlatformDataDeletionTimerCallback)
{
}

CachedFont::~CachedFont()
{
}

ResourceError CachedFont::load(CachedResourceLoader*)
{
    // Don't load the file yet.  Wait for an access before triggering the load.
    setLoading(true);
    return ResourceError();
}

void CachedFont::didAddClient(CachedResourceClient& client)
{
	if(m_fontPlatformDataDeletionTimer.isActive())
		m_fontPlatformDataDeletionTimer.stop();

    ASSERT(client.resourceClientType() == CachedFontClient::expectedType());

	CachedResource::didAddClient(client);

    if (!isLoading())
        static_cast<CachedFontClient&>(client).fontLoaded(this);
}

void CachedFont::finishLoading(SharedBuffer* data)
{
    m_data = data;
    setEncodedSize(m_data.get() ? m_data->size() : 0);
    setLoading(false);
    checkNotify();
}

void CachedFont::beginLoadIfNeeded(CachedResourceLoader* dl)
{
    if (!m_loadInitiated) {
        m_loadInitiated = true;
        CachedResource::load(dl);
    }
}

bool CachedFont::ensureCustomFontData(const AtomicString&)
{
	if(m_data)
		return ensureCustomFontData(m_data.get());
	return false;
}

bool CachedFont::ensureCustomFontData(SharedBuffer* buffer)
{
    if (!m_fontCustomPlatformData && !errorOccurred() && !isLoading() && buffer) {
		bool wrapping;
		m_fontCustomPlatformData = createCustomFontData(*buffer, wrapping);
		m_hasCreatedFontDataWrappingResource = m_fontCustomPlatformData && wrapping;
		if (!m_fontCustomPlatformData)
			setStatus(DecodeError);
	}

    return m_fontCustomPlatformData.get();
}

PassRefPtr<FontCustomPlatformData> CachedFont::createCustomFontData(SharedBuffer& bytes, bool& wrapping)
{
	RefPtr<SharedBuffer> buffer = &bytes;
	wrapping = true;

	if (isWOFF(*buffer)) {
		Vector<char> convertedFont;
		if (!convertWOFFToSfnt(*buffer, convertedFont))
			buffer = nullptr;
		else
			buffer = SharedBuffer::adoptVector(convertedFont);
		wrapping = false;
	}
	else if (isWOFF2(buffer.get())) {
		Vector<char> convertedFont;
		if (!convertWOFF2ToSfnt(buffer.get(), convertedFont))
			buffer = nullptr;
		else
			buffer = SharedBuffer::adoptVector(convertedFont);
		wrapping = false;
	}

	return buffer ? createFontCustomPlatformData(*buffer) : nullptr;
}

PassRefPtr<Font> CachedFont::createFont(const FontDescription& fontDescription, const AtomicString&, bool syntheticBold, bool syntheticItalic, const FontFeatureSettings& fontFaceFeatures, const FontVariantSettings& fontFaceVariantSettings)
{
    return platformDataFromCustomData(fontDescription, syntheticBold, syntheticItalic, fontFaceFeatures, fontFaceVariantSettings);
}

PassRefPtr<Font> CachedFont::platformDataFromCustomData(const FontDescription& fontDescription, bool bold, bool italic, const FontFeatureSettings& fontFaceFeatures, const FontVariantSettings& fontFaceVariantSettings)
{
    ASSERT(m_fontCustomPlatformData);
    return platformDataFromCustomData(*m_fontCustomPlatformData, fontDescription, bold, italic, fontFaceFeatures, fontFaceVariantSettings);
}

PassRefPtr<Font> CachedFont::platformDataFromCustomData(FontCustomPlatformData& fontCustomPlatformData, const FontDescription& fontDescription, bool bold, bool italic, const FontFeatureSettings& fontFaceFeatures, const FontVariantSettings& fontFaceVariantSettings)
{
#if PLATFORM(COCOA)
    return m_fontCustomPlatformData->fontPlatformData(fontDescription, bold, italic, fontFaceFeatures, fontFaceVariantSettings);
#else
    UNUSED_PARAM(fontFaceFeatures);
    UNUSED_PARAM(fontFaceVariantSettings);
    return fontCustomPlatformData.fontPlatformData(fontDescription, bold, italic, fontDescription.orientation());
#endif
}

void CachedFont::fontPlatformDataDeletionTimerCallback()
{
	m_fontCustomPlatformData = nullptr;
}

void CachedFont::purge()
{
	if(m_fontCustomPlatformData)
		m_fontCustomPlatformData->purge();
}

void CachedFont::allClientsRemoved()
{
    if (double interval = MemoryCache::singleton().deadFontPlatformDataDeletionInterval())
		m_fontPlatformDataDeletionTimer.startOneShot(interval);

    CachedResource::allClientsRemoved();
}

void CachedFont::checkNotify()
{
    if (isLoading())
        return;
    
    CachedResourceClientWalker<CachedFontClient> walker(*this);
    while (CachedFontClient* client = walker.next())
    	client->fontLoaded(this);
}

bool CachedFont::mayTryReplaceEncodedData() const
{
    // If a FontCustomPlatformData has ever been constructed to wrap the internal resource buffer then it still might be in use somewhere.
    // That platform font object might directly reference the encoded data buffer behind this CachedFont,
    // so replacing it is unsafe.

    return !m_hasCreatedFontDataWrappingResource;
}

}
