/*
    Copyright (C) 1998 Lars Knoll (knoll@mpi-hd.mpg.de)
    Copyright (C) 2001 Dirk Mueller (mueller@kde.org)
    Copyright (C) 2002 Waldo Bastian (bastian@kde.org)
    Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
    Copyright (C) 2004, 2005, 2006, 2007 Apple Inc. All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include "config.h"
#include "CachedImage.h"

#include "BitmapImage.h"
#include "MemoryCache.h"
#include "CachedResourceClient.h"
#include "CachedResourceClientWalker.h"
#include "CachedResourceLoader.h"
#include "Frame.h"
#include "FrameLoaderClient.h"
#include "FrameLoaderTypes.h"
#include "FrameView.h"
#include "Page.h"
#include "RenderElement.h"
#include "SharedBuffer.h"
#include "SecurityOrigin.h"
#include "Settings.h"
#include "SubresourceLoader.h"
#include <wtf/CurrentTime.h>
#include <wtf/StdLibExtras.h>
#include <wtf/Vector.h>

#if USE(CG)
#include "PDFDocumentImage.h"
#endif

#if ENABLE(SVG)
#include "SVGImage.h"
#endif

using std::max;

namespace WebCore {

CachedImage::CachedImage(CachedResourceRequest& request)
	: CachedResource(request, ImageResource)
	, m_isManuallyCached(false)
	, m_shouldPaintBrokenImage(true)
	, m_acceleratedSurfaceDeletionTimer(this, &CachedImage::acceleratedSurfaceDeletionTimerCallback)
{
    setStatus(Unknown);
}

inline CachedImage::CachedImage(Image* image)
    : CachedResource(KURL(), ImageResource)
    , m_image(image)
	, m_isManuallyCached(false)
    , m_shouldPaintBrokenImage(true)
	, m_acceleratedSurfaceDeletionTimer(this, &CachedImage::acceleratedSurfaceDeletionTimerCallback)
{
}

DLL_PUBLIC CachedImage* CachedImage::create(Image* image)
{
	return new CachedImage(image);
}

CachedImage::~CachedImage()
{
    clearImage();
}

ResourceError CachedImage::load(CachedResourceLoader* loader)
{
    m_skippingRevalidationDocument = makeWeakPtr(loader->document());
    if (loader->shouldPerformImageLoad(url()))
        return CachedResource::load(loader);
    else {
        setLoading(false);
        return ResourceError();
    }
}

void CachedImage::setBodyDataFrom(const CachedResource& resource)
{
    ASSERT(resource.type() == type());
    const CachedImage& image = static_cast<const CachedImage&>(resource);

    CachedResource::setBodyDataFrom(resource);

    m_image = image.m_image;
    m_imageObserver = image.m_imageObserver;
    if (m_imageObserver)
        m_imageObserver->cachedImages().add(*this);

    if (m_image && m_image->isSVGImage())
        m_svgImageCache = adoptPtr(new SVGImageCache(&static_cast<SVGImage&>(*m_image)));
}

void CachedImage::didAddClient(CachedResourceClient& c)
{    
    if (m_data && !m_image && !errorOccurred()) {
        createImage();
        m_image->setData(m_data.get(), true, mimeType());
    }
    
    ASSERT(c->resourceClientType() == CachedImageClient::expectedType());
    if (m_image && !m_image->isNull())
        static_cast<CachedImageClient&>(c).imageChanged(this);

    if (m_image)
        m_image->startAnimationAsynchronously();

    CachedResource::didAddClient(c);
}

void CachedImage::didRemoveClient(CachedResourceClient& client)
{
    ASSERT(client);
    ASSERT(client->resourceClientType() == CachedImageClient::expectedType());

    m_pendingContainerContextRequests.remove(&static_cast<CachedImageClient&>(client));
#if ENABLE(SVG)
    if (m_svgImageCache)
        m_svgImageCache->removeClientFromCache(&static_cast<CachedImageClient&>(client));
#endif

    CachedResource::didRemoveClient(client);

    static_cast<CachedImageClient&>(client).didRemoveCachedImageClient(*this);
}

void CachedImage::switchClientsToRevalidatedResource()
{
    ASSERT(resourceToRevalidate());
    ASSERT(resourceToRevalidate()->isImage());
    // Pending container size requests need to be transferred to the revalidated resource.
    if (!m_pendingContainerContextRequests.isEmpty()) {
        // A copy of pending size requests is needed as they are deleted during CachedResource::switchClientsToRevalidateResouce().
    	ContainerContextRequests switchContainerContextRequests;
        for (auto it = m_pendingContainerContextRequests.begin(); it != m_pendingContainerContextRequests.end(); ++it)
        	switchContainerContextRequests.set(it->first, it->second);
        CachedResource::switchClientsToRevalidatedResource();
        CachedImage* revalidatedCachedImage = static_cast<CachedImage*>(resourceToRevalidate());
        for (auto it = switchContainerContextRequests.begin(); it != switchContainerContextRequests.end(); ++it)
            revalidatedCachedImage->setContainerContextForClient(*it->first, it->second.containerSize, it->second.containerZoom, it->second.imageURL);
        return;
    }

    CachedResource::switchClientsToRevalidatedResource();
}

void CachedImage::allClientsRemoved()
{
	m_pendingContainerContextRequests.clear();
    if (m_image && !errorOccurred())
        m_image->resetAnimation();

    if(!inCache())
    	destroyDecodedData();

    CachedResource::allClientsRemoved();
}

void CachedImage::acceleratedSurfaceDeletionTimerCallback()
{
	if(m_image)
		m_image->destroyAllAcceleratedSurfaces();
}

pair<Image*, float> CachedImage::brokenImage(float deviceScaleFactor) const
{
    if (deviceScaleFactor >= 2) {
        static Image* brokenImageHiRes = Image::loadPlatformResource("missingImage@2x").leakRef();
        if(brokenImageHiRes)
        	return std::make_pair(brokenImageHiRes, 2);
    }
    else{
    	static Image* brokenImageHiRes = Image::loadPlatformResource("missingImage").leakRef();
    	if(brokenImageHiRes)
    		return std::make_pair(brokenImageHiRes, 1);
    }

    return std::make_pair(Image::nullImage(), 1);
}

bool CachedImage::willPaintBrokenImage() const
{
    return errorOccurred() && m_shouldPaintBrokenImage;
}

Image* CachedImage::image() const
{
    if (errorOccurred() && m_shouldPaintBrokenImage) {
        // Returning the 1x broken image is non-ideal, but we cannot reliably access the appropriate
        // deviceScaleFactor from here. It is critical that callers use CachedImage::brokenImage() 
        // when they need the real, deviceScaleFactor-appropriate broken image icon. 
        return brokenImage(1).first;
    }

    if (m_image)
        return m_image.get();

    return Image::nullImage();
}

Image* CachedImage::imageForRenderer(const RenderObject* renderer)
{
    if (errorOccurred() && m_shouldPaintBrokenImage) {
        // Returning the 1x broken image is non-ideal, but we cannot reliably access the appropriate
        // deviceScaleFactor from here. It is critical that callers use CachedImage::brokenImage() 
        // when they need the real, deviceScaleFactor-appropriate broken image icon. 
    	if(Image* image = brokenImage(1).first)
    		return image;
        return Image::nullImage();
    }

    if (!m_image)
        return Image::nullImage();

#if ENABLE(SVG)
    if (m_image->drawsSVGImage()) {
        Image* image = m_svgImageCache->imageForRenderer(renderer);
        if (image != Image::nullImage())
            return image;
    }
#else
    UNUSED_PARAM(renderer);
#endif
    return m_image.get();
}

bool CachedImage::hasSVGImage() const
{
    return image() && image()->drawsSVGImage();
}

void CachedImage::setContainerContextForClient(const CachedImageClient& client, const LayoutSize& containerSize, float containerZoom, const URL& imageURL)
{
    if (containerSize.isEmpty())
        return;
    ASSERT(renderer);
    ASSERT(containerZoom);
    if (!m_image) {
    	m_pendingContainerContextRequests.set(&client, ContainerContext(containerSize, containerZoom, imageURL));
		return;
    }
#if ENABLE(SVG)
    if (!m_image->drawsSVGImage()) {
        m_image->setContainerSize(containerSize);
        return;
    }

    m_svgImageCache->setContainerContextForClient(client, containerSize, containerZoom, imageURL);
#else
    UNUSED_PARAM(containerZoom);
    m_image->setContainerSize(containerSize);
#endif
}

bool CachedImage::usesImageContainerSize() const
{
    if (m_image)
        return m_image->usesContainerSize();

    return false;
}

bool CachedImage::imageHasRelativeWidth() const
{
    if (m_image)
        return m_image->hasRelativeWidth();

    return false;
}

bool CachedImage::imageHasRelativeHeight() const
{
    if (m_image)
        return m_image->hasRelativeHeight();

    return false;
}

FloatSize CachedImage::imageSizeForRenderer(const RenderElement* renderer, SizeType sizeType) const
{
    if (!m_image)
        return LayoutSize();

    if (m_image->isBitmapImage() && (renderer && renderer->shouldRespectImageOrientation() == RespectImageOrientation))
        return LayoutSize(static_cast<BitmapImage*>(m_image.get())->sizeRespectingOrientation());

#if ENABLE(SVG)
    if (m_image->drawsSVGImage() && sizeType == UsedSize)
        return LayoutSize(m_svgImageCache->imageSizeForRenderer(renderer));
#endif

	return LayoutSize(m_image->size());
}

LayoutSize CachedImage::unclampedImageSizeForRenderer(const RenderElement* renderer, float multiplier, SizeType sizeType) const
{
    LayoutSize imageSize = LayoutSize(imageSizeForRenderer(renderer, sizeType));
    if (imageSize.isEmpty() || multiplier == 1.0f)
        return imageSize;

    float widthScale = m_image->hasRelativeWidth() ? 1.0f : multiplier;
    float heightScale = m_image->hasRelativeHeight() ? 1.0f : multiplier;
    imageSize.scale(widthScale, heightScale);
	return imageSize;
}

LayoutSize CachedImage::imageSizeForRenderer(const RenderElement* renderer, float multiplier, SizeType sizeType) const
{
	auto imageSize = unclampedImageSizeForRenderer(renderer, multiplier, sizeType);
	if (imageSize.isEmpty() || multiplier == 1.0f)
		return imageSize;

	// Don't let images that have a width/height >= 1 shrink below 1 when zoomed.
	LayoutSize minimumSize(imageSize.width() > 0 ? 1 : 0, imageSize.height() > 0 ? 1 : 0);
    imageSize.clampToMinimumSize(minimumSize);

    ASSERT(multiplier != 1.0f || (imageSize.width().fraction() == 0.0f && imageSize.height().fraction() == 0.0f));
	return imageSize;
}

void CachedImage::computeIntrinsicDimensions(Length& intrinsicWidth, Length& intrinsicHeight, FloatSize& intrinsicRatio)
{
    if (m_image)
        m_image->computeIntrinsicDimensions(intrinsicWidth, intrinsicHeight, intrinsicRatio);
}

void CachedImage::notifyObservers(const IntRect* changeRect, bool allDataReceived)
{
    CachedResourceClientWalker<CachedImageClient> w(*this);
    while (CachedImageClient* client = w.next())
        client->imageChanged(this, changeRect, allDataReceived);
}

void CachedImage::checkShouldPaintBrokenImage()
{
    if (!m_loader || m_loader->reachedTerminalState())
        return;

    m_shouldPaintBrokenImage = m_loader->frameLoader()->client()->shouldPaintBrokenImage(url());
}

bool CachedImage::isPDFResource() const
{
    return Image::isPDFResource(m_response.mimeType(), url());
}

bool CachedImage::isPostScriptResource() const
{
    return Image::isPostScriptResource(m_response.mimeType(), url());
}

void CachedImage::clear()
{
    destroyDecodedData();
    clearImage();
    m_pendingContainerContextRequests.clear();
    setEncodedSize(0);
}

inline void CachedImage::createImage()
{
    // Create the image if it doesn't yet exist.
    if (m_image)
        return;

    m_imageObserver = CachedImageObserver::create(*this);

    m_image = Image::create(*m_imageObserver);

    if (m_image) {
        if ((*m_image).isSVGImage())
            m_svgImageCache = adoptPtr(new SVGImageCache(&static_cast<SVGImage&>(*m_image)));

		// Send queued container size requests.
		if (m_image->usesContainerSize()) {
            for (auto& request : m_pendingContainerContextRequests)
				setContainerContextForClient(*request.first, request.second.containerSize, request.second.containerZoom, request.second.imageURL);
		}
		m_pendingContainerContextRequests.clear();
    }
}

CachedImage::CachedImageObserver::CachedImageObserver(CachedImage& image)
{
    m_cachedImages.add(image);
}

void CachedImage::CachedImageObserver::decodedSizeChanged(const Image* image, int delta)
{
    for (auto& cachedImage : m_cachedImages)
        cachedImage.decodedSizeChanged(image, delta);
}

void CachedImage::CachedImageObserver::didDraw(const Image* image)
{
    for (auto& cachedImage : m_cachedImages)
        cachedImage.didDraw(image);
}

void CachedImage::CachedImageObserver::animationAdvanced(const Image* image)
{
    for (auto& cachedImage : m_cachedImages)
        cachedImage.animationAdvanced(image);
}

void CachedImage::CachedImageObserver::changedInRect(const Image* image, const IntRect& rect)
{
    for (auto& cachedImage : m_cachedImages)
        cachedImage.changedInRect(image, rect);
}

void CachedImage::CachedImageObserver::scheduleRenderingUpdate(const Image& image)
{
    for (auto& cachedImage : m_cachedImages)
        cachedImage.scheduleRenderingUpdate(image);
}

inline void CachedImage::clearImage()
{
    if (!m_image)
        return;

    if (m_imageObserver) {
        m_imageObserver->cachedImages().remove(*this);

        if (m_imageObserver->cachedImages().isEmptyIgnoringNullReferences()) {
            ASSERT(m_imageObserver->hasOneRef());
            m_image->setImageObserver(nullptr);
        }

        m_imageObserver = nullptr;
    }

    m_image = nullptr;
}

bool CachedImage::canBeDrawn() const
{
    if (!m_image || m_image->isNull())
        return false;

    if (!m_loader || m_loader->reachedTerminalState())
        return true;

    Settings& settings = m_loader->frameLoader()->frame()->settings();
    size_t estimatedDecodedImageSize = m_image->width() * m_image->height() * 4; // no overflow check
    return estimatedDecodedImageSize <= settings.maximumDecodedImageSize();
}

void CachedImage::addIncrementalDataBuffer(SharedBuffer* data)
{
    m_data = data;
    if (!data)
        return;

    createImage();

    if(!m_image)
    	return;

    // Have the image update its data from its internal buffer.
    // It will not do anything now, but will delay decoding until 
    // queried for info (like size or specific image frames).
    bool sizeAvailable = m_image->setData(m_data.get(), false, mimeType());
    if (!sizeAvailable)
		return;

    if (!canBeDrawn()) {
        // There's no image to draw or its decoded size is bigger than the maximum allowed.
		error(errorOccurred() ? status() : DecodeError);
		if (inCache()){
			auto& memoryCache = MemoryCache::singleton();
			memoryCache.revalidationFailed(*this);
			memoryCache.remove(*this);
		}
		return;
	}

    // Go ahead and tell our observers to try to draw.
    // Each chunk from the network causes observers to repaint, which will
    // force that chunk to decode.
	// It would be nice to only redraw the decoded band of the image, but with the current design
	// (decoding delayed until painting) that seems hard.
	notifyObservers(0, false);

	setEncodedSize(m_image->data() ? m_image->data()->size() : 0);
}

void CachedImage::addDataBuffer(SharedBuffer* data)
{
    ASSERT(m_options.dataBufferingPolicy == BufferData);
    addIncrementalDataBuffer(data);
}

void CachedImage::addData(const char* data, unsigned length)
{
    ASSERT(m_options.dataBufferingPolicy == DoNotBufferData);
    addIncrementalDataBuffer(SharedBuffer::create(data, length).ptr());
}

void CachedImage::finishLoading(SharedBuffer* data)
{
    m_data = data;
    if (!m_image && data)
        createImage();

    if (m_image) {
#if ENABLE(SVG)
    	if (m_loader && m_image->isSVGImage())
			static_cast<SVGImage*>(m_image.get())->setDataProtocolLoader(&m_loader->dataProtocolFrameLoader());
#endif
        m_image->setData(m_data ? m_data.get() : nullptr, true);
    }

    if (!canBeDrawn()) {
        // There's no image to draw or its decoded size is bigger than the maximum allowed.
        error(errorOccurred() ? status() : DecodeError);
        if (inCache()){
			MemoryCache& memoryCache = MemoryCache::singleton();
			memoryCache.revalidationFailed(*this);
			memoryCache.remove(*this);
		}
        return;
    }

    notifyObservers(0, true);
    if (m_image)
        setEncodedSize(m_image->data() ? m_image->data()->size() : 0);
    CachedResource::finishLoading(data);
}

void CachedImage::error(CachedResource::Status status)
{
    checkShouldPaintBrokenImage();
    clear();
    CachedResource::error(status);
    notifyObservers(0, false);
}

void CachedImage::responseReceived(const ResourceResponse& response)
{
    if (!m_response.isNull())
        clear();
    CachedResource::responseReceived(response);
}

void CachedImage::handleOutOfMemory(bool clearRawData)
{
	if (m_image && !errorOccurred())
        m_image->handleOutOfMemory();
}

void CachedImage::destroyDecodedData()
{
    bool canDeleteImage = !m_image || (m_image->hasOneRef() && m_image->isBitmapImage());
    if (canDeleteImage && !isLoading() && !hasClients()) {
        m_image = nullptr;
        setDecodedSize(0);
    } else if (m_image && !errorOccurred())
        m_image->destroyDecodedData();
}

void CachedImage::decodedSizeChanged(const Image* image, int delta)
{
    if (!image || image != m_image)
        return;
    
    setDecodedSize(decodedSize() + delta);
}

void CachedImage::didDraw(const Image* image)
{
    if (!image || image != m_image)
        return;
    
    CachedResource::didAccessDecodedData(WTF::monotonicallyIncreasingTime());

	double interval = MemoryCache::singleton().acceleratedSurfaceDeletionInterval();
	if(interval > 0){
		if(m_acceleratedSurfaceDeletionTimer.isActive())
			m_acceleratedSurfaceDeletionTimer.stop();

		m_acceleratedSurfaceDeletionTimer.startOneShot(interval);
	}
}

void CachedImage::animationAdvanced(const Image* image)
{
    if (!image || image != m_image)
        return;

    bool shouldPauseAnimation = true;

    CachedResourceClientWalker<CachedImageClient> clientWalker(*this);
    while (CachedImageClient* client = clientWalker.next()) {
        bool canPause = false;
        client->newImageAnimationFrameAvailable(*this, canPause);
        if (!canPause)
            shouldPauseAnimation = false;
    }

    if (shouldPauseAnimation)
        m_image->stopAnimation();
}

void CachedImage::changedInRect(const Image* image, const IntRect& rect)
{
    if (!image || image != m_image)
        return;
    notifyObservers(&rect, isLoaded());
}

void CachedImage::scheduleRenderingUpdate(const Image& image)
{
    if (&image != m_image)
        return;

    CachedResourceClientWalker<CachedImageClient> walker(*this);
    while (auto* client = walker.next())
        client->scheduleRenderingUpdate();
}

bool CachedImage::isOriginClean(SecurityOrigin* origin)
{
    ASSERT_UNUSED(origin, origin);
    ASSERT(this->origin());
    ASSERT(origin->toString() == this->origin()->toString());
    return !loadFailedOrCanceled() && isCORSSameOrigin();
}

CachedResource::RevalidationDecision CachedImage::makeRevalidationDecision(CachePolicy cachePolicy) const
{
    if (UNLIKELY(isManuallyCached())) {
        // Do not revalidate manually cached images. This mechanism is used as a
        // way to efficiently share an image from the client to content and
        // the URL for that image may not represent a resource that can be
        // retrieved by standard means. If the manual caching SPI is used, it is
        // incumbent on the client to only use valid resources.
        return RevalidationDecision::No;
    }
    return CachedResource::makeRevalidationDecision(cachePolicy);
}

bool CachedImage::canSkipRevalidation(const CachedResourceLoader& loader, const CachedResourceRequest& request) const
{
    if (options().mode != request.options().mode || options().credentials != request.options().credentials || resourceRequest().allowCookies() != request.resourceRequest().allowCookies())
        return false;

    // Skip revalidation as per https://html.spec.whatwg.org/#ignore-higher-layer-caching which defines a per-document image list.
    // This rule is loosely implemented by other browsers, we could relax it and should update it once memory cache is properly specified.
    return m_skippingRevalidationDocument && loader.document() == m_skippingRevalidationDocument;
}

bool CachedImage::currentFrameKnownToBeOpaque(const RenderElement* renderer)
{
    Image* image = imageForRenderer(renderer);
	if (image->isBitmapImage())
		image->nativeImageForCurrentFrame(); // force decode
    return image->currentFrameKnownToBeOpaque();
}

} // namespace WebCore
