/*
    Copyright (C) 1998 Lars Knoll (knoll@mpi-hd.mpg.de)
    Copyright (C) 2001 Dirk Mueller (mueller@kde.org)
    Copyright (C) 2002 Waldo Bastian (bastian@kde.org)
    Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include "config.h"
#include "CachedResource.h"

#include "CachedMetadata.h"
#include "CachedResourceClient.h"
#include "CachedResourceClientWalker.h"
#include "CachedResourceHandle.h"
#include "CachedResourceLoader.h"
#include "CrossOriginAccessControl.h"
#include "Document.h"
#include "DocumentLoader.h"
#include "Frame.h"
#include "FrameLoaderClient.h"
#include "KURL.h"
#include "Logging.h"
#include "MemoryCache.h"
#include "SharedBuffer.h"
#include "ResourceHandle.h"
#include "ResourceLoadScheduler.h"
#include "SchemeRegistry.h"
#include "SecurityOrigin.h"
#include "SubresourceLoader.h"
#include <wtf/CurrentTime.h>
#include <wtf/MathExtras.h>
#include <wtf/RefCountedLeakCounter.h>
#include <wtf/StdLibExtras.h>
#include <wtf/text/CString.h>
#include <wtf/Vector.h>

using namespace WTF;

namespace WebCore {

DLL_PUBLIC HashSet<CachedResource*> gCachedResources;

ResourceLoadPriority CachedResource::defaultPriorityForResourceType(Type type)
{
    switch (type) {
    	case CachedResource::MainResource:
    		return ResourceLoadPriorityVeryHigh;
        case CachedResource::CSSStyleSheet:
        case CachedResource::Script:
            return ResourceLoadPriorityHigh;
#if ENABLE(SVG_FONTS)
        case CachedResource::SVGFontResource:
#endif
        case CachedResource::MediaResource:
        case CachedResource::FontResource:
        case CachedResource::RawResource:
        case CachedResource::Favicon:
            return ResourceLoadPriorityMedium;
        case CachedResource::ImageResource:
            return ResourceLoadPriorityLow;
#if ENABLE(XSLT)
        case CachedResource::XSLStyleSheet:
            return ResourceLoadPriorityHigh;
#endif
#if ENABLE(SVG)
        case CachedResource::SVGDocumentResource:
            return ResourceLoadPriorityLow;
#endif
#if ENABLE(LINK_PREFETCH)
        case CachedResource::LinkPrefetch:
            return ResourceLoadPriorityVeryLow;
        case CachedResource::LinkSubresource:
            return ResourceLoadPriorityVeryLow;
#endif
#if ENABLE(VIDEO_TRACK)
        case CachedResource::TextTrackResource:
            return ResourceLoadPriorityLow;
#endif
    }
    ASSERT_NOT_REACHED();
    return ResourceLoadPriorityLow;
}

#if PLATFORM(CHROMIUM) || PLATFORM(BLACKBERRY)
static ResourceRequest::TargetType cachedResourceTypeToTargetType(CachedResource::Type type)
{
    switch (type) {
    case CachedResource::MainResource:
        return ResourceRequest::TargetIsMainFrame;
    case CachedResource::CSSStyleSheet:
#if ENABLE(XSLT)
    case CachedResource::XSLStyleSheet:
#endif
        return ResourceRequest::TargetIsStyleSheet;
    case CachedResource::Script: 
        return ResourceRequest::TargetIsScript;
    case CachedResource::FontResource:
        return ResourceRequest::TargetIsFontResource;
    case CachedResource::ImageResource:
        return ResourceRequest::TargetIsImage;
    case CachedResource::RawResource:
        return ResourceRequest::TargetIsSubresource;    
#if ENABLE(LINK_PREFETCH)
    case CachedResource::LinkPrefetch:
        return ResourceRequest::TargetIsPrefetch;
    case CachedResource::LinkSubresource:
        return ResourceRequest::TargetIsSubresource;
#endif
#if ENABLE(VIDEO_TRACK)
    case CachedResource::TextTrackResource:
        return ResourceRequest::TargetIsTextTrack;
#endif
#if ENABLE(SVG)
    case CachedResource::SVGDocumentResource:
        return ResourceRequest::TargetIsImage;
#endif
    }
    ASSERT_NOT_REACHED();
    return ResourceRequest::TargetIsSubresource;
}
#endif

DEFINE_DEBUG_ONLY_GLOBAL(RefCountedLeakCounter, cachedResourceLeakCounter, ("CachedResource"));

CachedResource::CachedResource(CachedResourceRequest& request, Type type, SessionID sessionID)
    : m_resourceRequest(request.releaseResourceRequest())
	, m_options(request.options())
	, m_sessionID(sessionID)
    , m_loadPriority(defaultPriorityForResourceType(type))
	, m_responseTainting(ResourceResponse::Tainting::Basic)
    , m_responseTimestamp(currentTime())
	, m_decodedDataDeletionTimer(this, &CachedResource::decodedDataDeletionTimerFired)
	, m_fragmentIdentifierForRequest(request.releaseFragmentIdentifier())
	, m_origin(request.releaseOrigin())
    , m_lastDecodedAccessTime(0)
    , m_loadFinishTime(0)
	, m_lastAccessTime(0)
    , m_encodedSize(0)
    , m_decodedSize(0)
    , m_accessCount(0)
    , m_handleCount(0)
    , m_preloadCount(0)
    , m_preloadResult(PreloadNotReferenced)
    , m_requestedFromNetworkingLayer(false)
    , m_inCache(false)
    , m_loading(false)
	, m_isLinkPreload(request.isLinkPreload())
	, m_canReuse(true)
    , m_switchingClientsToRevalidatedResource(false)
	, m_ignoreForRequestCount(request.ignoreForRequestCount())
    , m_type(type)
    , m_status(Pending)
#ifndef NDEBUG
    , m_deleted(false)
    , m_lruIndex(0)
#endif
{
    ASSERT(m_type == unsigned(type)); // m_type is a bitfield, so this tests careless updates of the enum.
    ASSERT(sessionID.isValid());
#ifndef NDEBUG
    cachedResourceLeakCounter.increment();
#endif

    gCachedResources.add(this);

    // FIXME: We should have a better way of checking for Navigation loads, maybe FetchMode::Options::Navigate.
    ASSERT(m_origin || m_type == CachedResource::MainResource);

    if (isRequestCrossOrigin(m_origin.get(), m_resourceRequest.url(), m_options))
        setCrossOrigin();
}

// FIXME: For this constructor, we should probably mandate that the URL has no fragment identifier.
CachedResource::CachedResource(const KURL& url, Type type, SessionID sessionID)
    : m_resourceRequest(url)
	, m_sessionID(sessionID)
    , m_loadPriority(defaultPriorityForResourceType(type))
	, m_responseTainting(ResourceResponse::Tainting::Basic)
    , m_responseTimestamp(currentTime())
	, m_decodedDataDeletionTimer(this, &CachedResource::decodedDataDeletionTimerFired)
	, m_fragmentIdentifierForRequest(CachedResourceRequest::splitFragmentIdentifierFromRequestURL(m_resourceRequest))
    , m_lastDecodedAccessTime(0)
    , m_loadFinishTime(0)
	, m_lastAccessTime(0)
    , m_encodedSize(0)
    , m_decodedSize(0)
    , m_accessCount(0)
    , m_handleCount(0)
    , m_preloadCount(0)
    , m_preloadResult(PreloadNotReferenced)
    , m_requestedFromNetworkingLayer(false)
    , m_inCache(false)
    , m_loading(false)
	, m_isLinkPreload(false)
	, m_canReuse(true)
    , m_switchingClientsToRevalidatedResource(false)
	, m_ignoreForRequestCount(false)
    , m_type(type)
    , m_status(Pending)
#ifndef NDEBUG
    , m_deleted(false)
    , m_lruIndex(0)
#endif
{
    ASSERT(m_type == unsigned(type)); // m_type is a bitfield, so this tests careless updates of the enum.
    ASSERT(sessionID.isValid());
#ifndef NDEBUG
    cachedResourceLeakCounter.increment();
#endif

    gCachedResources.add(this);
}

CachedResource::~CachedResource()
{
    ASSERT(!m_resourceToRevalidate); // Should be true because canDelete() checks this.
    ASSERT(canDelete());
    ASSERT(!inCache());
    ASSERT(!m_deleted);
    ASSERT(url().isNull() || !allowsCaching() || MemoryCache::singleton().resourceForURL(KURL(ParsedURLString, url())) != this);
    
#ifndef NDEBUG
    m_deleted = true;
    cachedResourceLeakCounter.decrement();
#endif

    gCachedResources.remove(this);
}

void CachedResource::failBeforeStarting()
{
    // FIXME: What if resources in other frames were waiting for this revalidation?
    LOG(ResourceLoading, "Cannot start loading '%s'", url().string().latin1().data());
    if (allowsCaching() && m_resourceToRevalidate)
    	MemoryCache::singleton().revalidationFailed(*this);
    error(CachedResource::LoadError);
}

ResourceError CachedResource::load(CachedResourceLoader* cachedResourceLoader)
{
    if (!cachedResourceLoader->frame()) {
    	WTF::logMessage(ASCIILiteral("CachedResource::load frame is null."));
        failBeforeStarting();
        return ResourceError();
    }

    FrameLoader* frameLoader = cachedResourceLoader->frame()->loader();
    if (m_options.securityCheck == DoSecurityCheck && (frameLoader->state() == FrameStateProvisional || !frameLoader->activeDocumentLoader() || frameLoader->activeDocumentLoader()->isStopping())) {
    	if (frameLoader->state() == FrameStateProvisional)
    		WTF::logMessage(ASCIILiteral("CachedResource::load: Failed security check -- state is provisional."));
    	else if (!frameLoader->activeDocumentLoader())
    		WTF::logMessage(ASCIILiteral("CachedResource::load: Failed security check -- not active document."));
    	else if (frameLoader->activeDocumentLoader()->isStopping())
    		WTF::logMessage(ASCIILiteral("CachedResource::load: Failed security check -- active loader is stopping."));
    	failBeforeStarting();
        return ResourceError();
    }

    m_loading = true;

#if PLATFORM(CHROMIUM) || PLATFORM(BLACKBERRY)
    if (m_resourceRequest.targetType() == ResourceRequest::TargetIsUnspecified)
        m_resourceRequest.setTargetType(cachedResourceTypeToTargetType(type()));
#endif

    if (isCacheValidator()) {
        CachedResource* resourceToRevalidate = m_resourceToRevalidate.get();
        ASSERT(resourceToRevalidate->canUseCacheValidator());
        ASSERT(resourceToRevalidate->isLoaded());
        const String lastModified = resourceToRevalidate->response().httpHeaderField(ASCIILiteral("Last-Modified"));
        const String eTag = resourceToRevalidate->response().httpHeaderField(ASCIILiteral("ETag"));
        if (!lastModified.isEmpty() || !eTag.isEmpty()) {
            ASSERT(cachedResourceLoader->cachePolicy(type()) != CachePolicyReload);
            if (cachedResourceLoader->cachePolicy(type()) == CachePolicyRevalidate)
                m_resourceRequest.setHTTPHeaderField(ASCIILiteral("Cache-Control"), ASCIILiteral("max-age=0"));
            if (!lastModified.isEmpty())
                m_resourceRequest.setHTTPHeaderField(HTTPHeaderName::IfModifiedSince, lastModified);
            if (!eTag.isEmpty())
                m_resourceRequest.setHTTPHeaderField(HTTPHeaderName::IfNoneMatch, eTag);
        }
    }

#if ENABLE(LINK_PREFETCH)
    if (type() == CachedResource::LinkPrefetch || type() == CachedResource::LinkSubresource)
        m_resourceRequest.setHTTPHeaderField("Purpose", "prefetch");
#endif
    m_resourceRequest.setPriority(loadPriority());

    // Navigation algorithm is setting up the request before sending it to CachedResourceLoader?CachedResource.
    // So no need for extra fields for MainResource.
    if (type() != CachedResource::MainResource)
        frameLoader->addExtraFieldsToSubresourceRequest(m_resourceRequest);

    // FIXME: It's unfortunate that the cache layer and below get to know anything about fragment identifiers.
    // We should look into removing the expectation of that knowledge from the platform network stacks.
    ResourceRequest request(m_resourceRequest);
    if (!m_fragmentIdentifierForRequest.isNull()) {
        KURL url = request.url();
        url.setFragmentIdentifier(m_fragmentIdentifierForRequest);
        request.setURL(url);
        m_fragmentIdentifierForRequest = String();
    }

    if (m_options.keepAlive) {
#if 0
        if (!cachedResourceLoader.keepaliveRequestTracker().tryRegisterRequest(*this)) {
            setResourceError({ errorDomainWebKitInternal, 0, request.url(), ASCIILiteral("Reached maximum amount of queued data of 64Kb for keepalive requests"), ResourceError::Type::AccessControl });
            failBeforeStarting();
            return;
        }
        // FIXME: We should not special-case Beacon here.
        if (shouldUsePingLoad(type())) {
            ASSERT(m_originalRequest);
            CachedResourceHandle<CachedResource> protectedThis(this);

            // FIXME: Move beacon loads to normal subresource loading to get normal inspector request instrumentation hooks.
            unsigned long identifier = frame.page()->progress().createUniqueIdentifier();
            InspectorInstrumentation::willSendRequestOfType(&frame, identifier, frameLoader.activeDocumentLoader(), request, InspectorInstrumentation::LoadType::Beacon);

            platformStrategies()->loaderStrategy()->startPingLoad(frame, request, m_originalRequest->httpHeaderFields(), m_options, [this, protectedThis = WTFMove(protectedThis), protectedFrame = makeRef(frame), identifier] (const ResourceError& error, const ResourceResponse& response) {
                if (!response.isNull())
                    InspectorInstrumentation::didReceiveResourceResponse(protectedFrame, identifier, protectedFrame->loader().activeDocumentLoader(), response, nullptr);
                if (error.isNull()) {
                    finishLoading(nullptr);
                    NetworkLoadMetrics emptyMetrics;
                    InspectorInstrumentation::didFinishLoading(protectedFrame.ptr(), protectedFrame->loader().activeDocumentLoader(), identifier, emptyMetrics, nullptr);
                } else {
                    setResourceError(error);
                    this->error(LoadError);
                    InspectorInstrumentation::didFailLoading(protectedFrame.ptr(), protectedFrame->loader().activeDocumentLoader(), identifier, error);
                }
            });
            return;
        }
#endif
    }

    m_loader = resourceLoadScheduler()->scheduleSubresourceLoad(cachedResourceLoader->frame(), this, request, request.priority(), m_options);
    if (!m_loader) {
    	WTF::logMessage(ASCIILiteral("CachedResource::load subresource loader is null."));
        failBeforeStarting();
        return ResourceError();
    }

    m_status = Pending;
    return ResourceError();
}

void CachedResource::loadFrom(const CachedResource& resource)
{
    ASSERT(url() == resource.url());
    ASSERT(type() == resource.type());
    ASSERT(resource.status() == Status::Cached);

    if (isCrossOrigin() && m_options.mode == FetchOptions::Mode::Cors) {
        ASSERT(m_origin);
        String errorMessage;
        if (!WebCore::passesAccessControlCheck(resource.response(), m_options.storedCredentialsPolicy, *m_origin, errorMessage)) {
            setResourceError(ResourceError(String(), 0, url(), errorMessage, ResourceError::Type::AccessControl));
            return;
        }
    }

    setBodyDataFrom(resource);
    setStatus(Status::Cached);
    setLoading(false);
}

void CachedResource::setBodyDataFrom(const CachedResource& resource)
{
    m_data = resource.m_data;
    m_response = resource.m_response;
    m_response.setTainting(m_responseTainting);
    setDecodedSize(resource.decodedSize());
    setEncodedSize(resource.encodedSize());
}

void CachedResource::checkNotify()
{
    if (isLoading() || stillNeedsLoad())
        return;

    CachedResourceClientWalker<CachedResourceClient> w(*this);
    while (CachedResourceClient* c = w.next())
        c->notifyFinished(this);
}

void CachedResource::addDataBuffer(SharedBuffer*)
{
    ASSERT(m_options.dataBufferingPolicy == BufferData);
}

void CachedResource::addData(const char*, unsigned)
{
    ASSERT(m_options.dataBufferingPolicy == DoNotBufferData);
}

void CachedResource::finishLoading(SharedBuffer*)
{
    setLoading(false);
    checkNotify();
}

void CachedResource::error(CachedResource::Status status)
{
    setStatus(status);
    ASSERT(errorOccurred());
    m_data.clear();

    setLoading(false);
    checkNotify();
}

void CachedResource::cancelLoad()
{
    if (!isLoading() && !stillNeedsLoad())
        return;

    setStatus(LoadError);
    setLoading(false);
    checkNotify();
}

void CachedResource::finish()
{
    if (!errorOccurred())
        m_status = Cached;
}

void CachedResource::setCrossOrigin()
{
    ASSERT(m_options.mode != FetchOptions::Mode::SameOrigin);
    m_responseTainting = (m_options.mode == FetchOptions::Mode::Cors) ? ResourceResponse::Tainting::Cors : ResourceResponse::Tainting::Opaque;
}

bool CachedResource::isCrossOrigin() const
{
    return m_responseTainting != ResourceResponse::Tainting::Basic;
}

bool CachedResource::isCORSSameOrigin() const
{
    // Following resource types do not use CORS
    ASSERT(type() != CachedResource::Type::FontResource);
    ASSERT(type() != CachedResource::Type::SVGFontResource);
    ASSERT(type() != CachedResource::XSLStyleSheet);

    // https://html.spec.whatwg.org/multipage/infrastructure.html#cors-same-origin
    return !loadFailedOrCanceled() && m_responseTainting != ResourceResponse::Tainting::Opaque;
}

bool CachedResource::isExpired() const
{
    if (m_response.isNull())
        return false;

    return computeCurrentAge(m_response, m_responseTimestamp) > freshnessLifetime(m_response);
}

static inline bool shouldCacheSchemeIndefinitely(StringView scheme)
{
#if PLATFORM(COCOA)
    if (equalLettersIgnoringASCIICase(scheme, "applewebdata"))
        return true;
#endif
#if USE(SOUP)
    if (equalLettersIgnoringASCIICase(scheme, "resource"))
        return true;
#endif
    return equalLettersIgnoringASCIICase(scheme, "data");
}

double CachedResource::freshnessLifetime(const ResourceResponse& response) const
{
    if (!response.url().protocolIsInHTTPFamily()) {
        StringView protocol = response.url().protocol();
        // Don't cache non-HTTP main resources since we can't check for freshness.
        // FIXME: We should not cache subresources either, but when we tried this
        // it caused performance and flakiness issues in our test infrastructure.
        if (m_type == MainResource && !SchemeRegistry::shouldCacheResponsesFromURLSchemeIndefinitely(protocol))
            return 0;

        return std::numeric_limits<double>::max();
    }

    return computeFreshnessLifetimeForHTTPFamily(response, m_responseTimestamp);
}

const ResourceResponse& CachedResource::responseForSameOriginPolicyChecks() const
{
    return m_redirectResponseForSameOriginPolicyChecks.isNull() ? m_response : m_redirectResponseForSameOriginPolicyChecks;
}

void CachedResource::redirectReceived(ResourceRequest& request, const ResourceResponse& response)
{
    m_requestedFromNetworkingLayer = true;
    if (response.isNull())
        return;

    // Redirect to data: URL uses the last HTTP response for SOP.
    if (response.isHTTP() && request.url().protocolIsData())
        m_redirectResponseForSameOriginPolicyChecks = response;

    updateRedirectChainStatus(m_redirectChainCacheStatus, response);
}

void CachedResource::setResponse(const ResourceResponse& response)
{
    ASSERT(m_response.type() == ResourceResponse::Type::Default);
    m_response = response;
    m_response.setRedirected(m_redirectChainCacheStatus.status != RedirectChainCacheStatus::NoRedirection);
    if (m_response.tainting() == ResourceResponse::Tainting::Basic || m_response.tainting() == ResourceResponse::Tainting::Cors)
		m_response.setTainting(m_responseTainting);

    m_varyingHeaderValues = collectVaryingRequestHeaders(m_resourceRequest, m_response, m_sessionID);
}

void CachedResource::responseReceived(const ResourceResponse& response)
{
    setResponse(response);
    m_responseTimestamp = currentTime();
    String encoding = response.textEncodingName();
    if (!encoding.isNull())
        setEncoding(encoding);
}

void CachedResource::setSerializedCachedMetadata(const char* data, size_t size)
{
    // We only expect to receive cached metadata from the platform once.
    // If this triggers, it indicates an efficiency problem which is most
    // likely unexpected in code designed to improve performance.
    ASSERT(!m_cachedMetadata);

    m_cachedMetadata = CachedMetadata::deserialize(data, size);
}

void CachedResource::setCachedMetadata(unsigned dataTypeID, const char* data, size_t size)
{
    // Currently, only one type of cached metadata per resource is supported.
    // If the need arises for multiple types of metadata per resource this could
    // be enhanced to store types of metadata in a map.
    ASSERT(!m_cachedMetadata);

    m_cachedMetadata = CachedMetadata::create(dataTypeID, data, size);
    ResourceHandle::cacheMetadata(m_response, m_cachedMetadata->serialize());
}

CachedMetadata* CachedResource::cachedMetadata(unsigned dataTypeID) const
{
    if (!m_cachedMetadata || m_cachedMetadata->dataTypeID() != dataTypeID)
        return 0;
    return m_cachedMetadata.get();
}

void CachedResource::clearLoader()
{
    ASSERT(m_loader);            
    m_loader = 0;
    deleteIfPossible();
}

DLL_PUBLIC void CachedResource::addClient(CachedResourceClient& client)
{
    if (addClientToSet(client))
        didAddClient(client);
}

void CachedResource::didAddClient(CachedResourceClient& client)
{
    if (m_decodedDataDeletionTimer.isActive())
        m_decodedDataDeletionTimer.stop();

    if (m_clientsAwaitingCallback.remove(client)) {
        m_clients.add(client);
#if ASSERT_ENABLED
        client.addAssociatedResource(*this);
#endif
    }

    // FIXME: Make calls to notifyFinished async
    if (!isLoading() && !stillNeedsLoad())
    	client.notifyFinished(this);
}

bool CachedResource::addClientToSet(CachedResourceClient& client)
{
    if (m_preloadResult == PreloadNotReferenced) {
        if (isLoaded())
            m_preloadResult = PreloadReferencedWhileComplete;
        else if (m_requestedFromNetworkingLayer)
            m_preloadResult = PreloadReferencedWhileLoading;
        else
            m_preloadResult = PreloadReferenced;
    }
    if (allowsCaching() && !hasClients() && inCache())
    	MemoryCache::singleton().addToLiveResourcesSize(this);

    if ((m_type == RawResource || m_type == MainResource) && !m_response.isNull() && !m_proxyResource) {
		// Certain resources (especially XHRs and main resources) do crazy things if an asynchronous load returns
		// synchronously (e.g., scripts may not have set all the state they need to handle the load).
        // Therefore, rather than immediately sending callbacks on a cache hit like other CachedResources,
        // we schedule the callbacks and ensure we never finish synchronously.
        ASSERT(!m_clientsAwaitingCallback.contains(client));
        m_clientsAwaitingCallback.add(client, CachedResourceCallback::schedule(this, client));
        return false;
    }

    m_clients.add(client);
    return true;
}

DLL_PUBLIC void CachedResource::removeClient(CachedResourceClient& client)
{
    OwnPtr<CachedResourceCallback> callback = m_clientsAwaitingCallback.take(client);
    if (callback) {
        ASSERT(!m_clients.contains(client));
        callback->cancel();
        callback.clear();
    } else {
        ASSERT(m_clients.contains(client));
        m_clients.remove(client);
        didRemoveClient(client);
    }
    if (deleteIfPossible()) {
        // `this` object is dead here.
        return;
    }

    if (hasClients())
        return;

	MemoryCache& memoryCache = MemoryCache::singleton();
	if (allowsCaching() && inCache()) {
		memoryCache.removeFromLiveResourcesSize(this);
	}

	if (!m_switchingClientsToRevalidatedResource)
		allClientsRemoved();
	destroyDecodedDataIfNeeded();

    if (!allowsCaching())
        return;

	if (response().cacheControlContainsNoStore() && url().protocolIs("https")) {
		memoryCache.revalidationFailed(*this);

		// RFC2616 14.9.2:
		// "no-store: ... MUST make a best-effort attempt to remove the information from volatile storage as promptly as possible"
		// "... History buffers MAY store such responses as part of their normal operation."
		// We allow non-secure content to be reused in history, but we do not allow secure content to be reused.
		memoryCache.remove(*this);
	}
    memoryCache.pruneSoon();
}

void CachedResource::destroyDecodedDataIfNeeded()
{
    if (!m_decodedSize)
        return;

    if (m_decodedDataDeletionTimer.isActive())
    	m_decodedDataDeletionTimer.stop();

    MemoryCache& memoryCache = MemoryCache::singleton();

    double interval = hasClients() ? memoryCache.decodedDataDeletionInterval() : memoryCache.deadDecodedDataDeletionInterval();
    if(interval)
        m_decodedDataDeletionTimer.startOneShot(interval);
}

void CachedResource::decodedDataDeletionTimerFired()
{
    destroyDecodedData();
}

bool CachedResource::deleteIfPossible()
{
    if (canDelete()) {
    	if(!inCache()) {
    		delete this;
    		return true;
    	}
    }
    return false;
}

void CachedResource::setDecodedSize(unsigned size)
{
    if (size == m_decodedSize)
        return;

    int delta = size - m_decodedSize;

    MemoryCache& memoryCache = MemoryCache::singleton();

    // The object must now be moved to a different queue, since its size has been changed.
    // We have to remove explicitly before updating m_decodedSize, so that we find the correct previous
    // queue.
    if (allowsCaching() && inCache() && !memoryCache.inHandleOutOfMemory())
    	memoryCache.removeFromAccessOrderedResourcesList(*this);
    
    m_decodedSize = size;
   
    if (allowsCaching() && inCache()) {
        // Now insert into the new LRU list.
    	if(!memoryCache.inHandleOutOfMemory())
    		memoryCache.insertInAccessOrderedResourcesList(*this);
        
        // Update the cache's size totals.
    	memoryCache.adjustSize(hasClients(), delta);
    }
}

void CachedResource::setEncodedSize(unsigned size)
{
    if (size == m_encodedSize)
        return;

    // The size cannot ever shrink (unless it is being nulled out because of an error).  If it ever does, assert.
    ASSERT(size == 0 || size >= m_encodedSize);
    
    int delta = size - m_encodedSize;

    // The object must now be moved to a different queue, since its size has been changed.
    // We have to remove explicitly before updating m_encodedSize, so that we find the correct previous
    // queue.
    if (allowsCaching() && inCache())
        MemoryCache::singleton().removeFromAccessOrderedResourcesList(*this);
    
    m_encodedSize = size;
   
    if (allowsCaching() && inCache()) {
    	MemoryCache& memoryCache = MemoryCache::singleton();

        // Now insert into the new LRU list.
    	memoryCache.insertInAccessOrderedResourcesList(*this);
        
        // Update the cache's size totals.
    	memoryCache.adjustSize(hasClients(), delta);
    }
}

void CachedResource::didAccessDecodedData(double timeStamp)
{
    m_lastDecodedAccessTime = timeStamp;
    
    if (allowsCaching() && inCache()) {
    	MemoryCache& memoryCache = MemoryCache::singleton();

    	memoryCache.didAccessDecodedData(*this);
    	memoryCache.pruneSoon();
    }

    destroyDecodedDataIfNeeded();
}
    
void CachedResource::allClientsRemoved()
{
	m_lastAccessTime = WTF::monotonicallyIncreasingTime();
}

void CachedResource::setResourceToRevalidate(CachedResource& resource)
{ 
    ASSERT(!m_resourceToRevalidate);
    ASSERT(&resource != this);
    ASSERT(m_handlesToRevalidate.isEmpty());
    ASSERT(resource.type() == type());
    ASSERT(!resource.m_proxyResource);

    LOG(ResourceLoading, "CachedResource %p setResourceToRevalidate %p", this, &resource);

    resource.m_proxyResource = makeWeakPtr(*this);
    m_resourceToRevalidate = makeWeakPtr(resource);
}

void CachedResource::clearResourceToRevalidate() 
{ 
    ASSERT(m_resourceToRevalidate);
    ASSERT(m_resourceToRevalidate->m_proxyResource == this);

    if (m_switchingClientsToRevalidatedResource)
        return;

	m_resourceToRevalidate->m_proxyResource = nullptr;
	m_resourceToRevalidate->deleteIfPossible();

    m_handlesToRevalidate.clear();
    m_resourceToRevalidate = nullptr;
    deleteIfPossible();
}
    
void CachedResource::switchClientsToRevalidatedResource()
{
    ASSERT(m_resourceToRevalidate);
    ASSERT(m_resourceToRevalidate->inCache());
    ASSERT(!inCache());

    LOG(ResourceLoading, "CachedResource %p switchClientsToRevalidatedResource %p", this, m_resourceToRevalidate);

    m_switchingClientsToRevalidatedResource = true;
    for (CachedResourceHandleBase* handle : m_handlesToRevalidate) {
        handle->m_resource = m_resourceToRevalidate.get();
        m_resourceToRevalidate->registerHandle(handle);
        --m_handleCount;
    }
    ASSERT(!m_handleCount);
    m_handlesToRevalidate.clear();

    Vector<WeakPtr<CachedResourceClient>> clientsToMove;
    for (auto entry : m_clients) {
        auto& client = entry.key;
        unsigned count = entry.value;
        while (count) {
            clientsToMove.append(makeWeakPtr(client));
            --count;
        }
    }

    for (auto& client : clientsToMove) {
        if (client)
            removeClient(*client);
    }
    ASSERT(!m_clients.computeSize());

    for (auto& client : clientsToMove) {
        if (client)
            m_resourceToRevalidate->addClientToSet(*client);
    }
    for (auto& client : clientsToMove) {
        // Calling didAddClient may do anything, including trying to cancel revalidation.
        // Assert that it didn't succeed.
        ASSERT(m_resourceToRevalidate);
        // Calling didAddClient for a client may end up removing another client. In that case it won't be in the set anymore.
        if (client && m_resourceToRevalidate->m_clients.contains(*client))
            m_resourceToRevalidate->didAddClient(*client);
    }
    m_switchingClientsToRevalidatedResource = false;
}
    
void CachedResource::updateResponseAfterRevalidation(const ResourceResponse& validatingResponse)
{
    m_responseTimestamp = currentTime();

    updateResponseHeadersAfterRevalidation(m_response, validatingResponse);
}

void CachedResource::registerHandle(CachedResourceHandleBase* h)
{
    ++m_handleCount;
    if (m_resourceToRevalidate)
        m_handlesToRevalidate.add(h);
}

DLL_PUBLIC void CachedResource::unregisterHandle(CachedResourceHandleBase* h)
{
    ASSERT(m_handleCount > 0);
    --m_handleCount;

    if (m_resourceToRevalidate)
         m_handlesToRevalidate.remove(h);

    if (!m_handleCount)
        deleteIfPossible();
}

bool CachedResource::canUseCacheValidator() const
{
    if (m_loading || errorOccurred())
        return false;

    if (m_response.cacheControlContainsNoStore())
        return false;
    return m_response.hasCacheValidatorFields();
}

CachedResource::RevalidationDecision CachedResource::makeRevalidationDecision(CachePolicy cachePolicy) const
{    
    switch (cachePolicy) {
    case CachePolicyHistoryBuffer:
        return RevalidationDecision::No;

    case CachePolicyReload:
    case CachePolicyRevalidate:
        return RevalidationDecision::YesDueToCachePolicy;

    case CachePolicyVerify:
        if (m_response.cacheControlContainsNoCache())
            return RevalidationDecision::YesDueToNoCache;
        // FIXME: Cache-Control:no-store should prevent storing, not reuse.
        if (m_response.cacheControlContainsNoStore())
            return RevalidationDecision::YesDueToNoStore;

        if (isExpired())
            return RevalidationDecision::YesDueToExpired;

        return RevalidationDecision::No;
    };
    ASSERT_NOT_REACHED();
    return RevalidationDecision::No;
}

bool CachedResource::redirectChainAllowsReuse(ReuseExpiredRedirectionOrNot reuseExpiredRedirection) const
{
    return WebCore::redirectChainAllowsReuse(m_redirectChainCacheStatus, reuseExpiredRedirection);
}

bool CachedResource::varyHeaderValuesMatch(const ResourceRequest& request)
{
    if (m_varyingHeaderValues.isEmpty())
        return true;

    return verifyVaryingRequestHeaders(m_varyingHeaderValues, request, m_sessionID);
}

unsigned CachedResource::overheadSize() const
{
    static const int kAverageClientsHashMapSize = 384;
    return sizeof(CachedResource) + m_response.memoryUsage() + kAverageClientsHashMapSize + m_resourceRequest.url().string().length() * 2;
}

bool CachedResource::areAllClientsXMLHttpRequests() const
{
    if (type() != RawResource)
        return false;

    for (auto& client : m_clients) {
        if (!client.key.isXMLHttpRequest())
            return false;
    }
    return true;
}

void CachedResource::setLoadPriority(const Optional<ResourceLoadPriority>& loadPriority)
{
    if (loadPriority)
        m_loadPriority = loadPriority.value();
    else
        m_loadPriority = defaultPriorityForResourceType(type());
}

CachedResource::CachedResourceCallback::CachedResourceCallback(CachedResource* resource, CachedResourceClient& client)
    : m_resource(resource)
    , m_client(client)
    , m_callbackTimer(this, &CachedResourceCallback::timerFired)
{
    m_callbackTimer.startOneShot(0);
}

void CachedResource::CachedResourceCallback::cancel()
{
    if (m_callbackTimer.isActive())
        m_callbackTimer.stop();
}

void CachedResource::CachedResourceCallback::timerFired()
{
    m_resource->didAddClient(m_client);
}

#if USE(FOUNDATION) || USE(SOUP)

void CachedResource::tryReplaceEncodedData(PassRefPtr<SharedBuffer> newBuffer)
{
    if (!m_data)
        return;

    if (!mayTryReplaceEncodedData())
        return;

    // Because the disk cache is asynchronous and racey with regards to the data we might be asked to replace,
    // we need to verify that the new buffer has the same contents as our old buffer.
    if (m_data->size() != newBuffer->size() || memcmp(m_data->data(), newBuffer->data(), m_data->size()))
        return;

    if (m_data->tryReplaceContentsWithPlatformBuffer(newBuffer)) {
        // FIXME: Should we call checkNotify() here to move already-decoded images to the new data source?
    }
}
#endif

}
