/*
    Copyright (C) 1998 Lars Knoll (knoll@mpi-hd.mpg.de)
    Copyright (C) 2001 Dirk Mueller <mueller@kde.org>
    Copyright (C) 2006 Samuel Weinig (sam.weinig@gmail.com)
    Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011 Apple Inc. All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#pragma once

#include "CachePolicy.h"
#include "CacheValidation.h"
#include "CachedResourceClient.h"
#include "FrameLoaderTypes.h"
#include "ResourceHandle.h"
#include "ResourceLoaderOptions.h"
#include "ResourceLoadPriority.h"
#include "ResourceRequest.h"
#include "ResourceResponse.h"
#include "SessionID.h"
#include "Timer.h"
#include <wtf/WeakHashCountedSet.h>
#include <wtf/WeakHashMap.h>
#include <wtf/OwnPtr.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>
#include <time.h>

namespace WebCore {

class MemoryCache;
class CachedMetadata;
class CachedResourceClient;
class CachedResourceHandleBase;
class CachedResourceLoader;
class CachedResourceRequest;
class InspectorResource;
class SharedBuffer;
class SecurityOrigin;
class SubresourceLoader;
class TextResourceDecoder;

// A resource that is held in the cache. Classes who want to use this object should derive
// from CachedResourceClient, to get the function calls in case the requested data has arrived.
// This class also does the actual communication with the loader to obtain the resource from the network.
class CachedResource : public CanMakeWeakPtr<CachedResource> {
    WTF_MAKE_NONCOPYABLE(CachedResource); WTF_MAKE_FAST_ALLOCATED;
    friend class MemoryCache;
    friend class InspectorResource;
    template<typename T> friend class CachedResourceClientWalker;
    
public:
    enum Type : uint8_t {
        MainResource,
        ImageResource,
        CSSStyleSheet,
        Script,
        FontResource,
        SVGFontResource,
        MediaResource,
        RawResource,
        Favicon,
        SVGDocumentResource,
#if ENABLE(XSLT)
        XSLStyleSheet,
#endif
#if ENABLE(LINK_PREFETCH)
        LinkPrefetch,
        LinkSubresource,
#endif
#if ENABLE(VIDEO_TRACK)
        TextTrackResource,
#endif
#if ENABLE(BML)
        ClutResource,
#endif
    };

    enum Status : uint8_t {
        Unknown,      // let cache decide what to do with it
        Pending,      // only partially loaded
        Cached,       // regular case
        LoadError,
        DecodeError
    };

    CachedResource(CachedResourceRequest&, Type, SessionID = SessionID::defaultSessionID());
    virtual ~CachedResource();

    virtual ResourceError load(CachedResourceLoader*);
    
    virtual void setEncoding(const String&) { }
    virtual String encoding() const { return String(); }
    virtual const TextResourceDecoder* textResourceDecoder() const { return nullptr; }
    virtual void addDataBuffer(SharedBuffer*);
    virtual void addData(const char* data, unsigned length);
    virtual void finishLoading(SharedBuffer*);
    virtual void error(CachedResource::Status);

    virtual void purge() { }

    void setResourceError(const ResourceError& error) { m_error = error; }
    const ResourceError& resourceError() const { return m_error; }

    virtual bool shouldIgnoreHTTPStatusCodeErrors() const { return false; }

    const ResourceRequest& resourceRequest() const { return m_resourceRequest; }
    const KURL& url() const { return m_resourceRequest.url();}
    SessionID sessionID() const { return m_sessionID; }
    Type type() const { return static_cast<Type>(m_type); }
    String mimeType() const { return m_response.mimeType(); }
    long long expectedContentLength() const { return m_response.expectedContentLength(); }

    ALWAYS_INLINE ResourceLoadPriority loadPriority() const { return m_loadPriority; }
    void setLoadPriority(const Optional<ResourceLoadPriority>&);

    void addClient(CachedResourceClient&);
    void removeClient(CachedResourceClient&);
    bool hasClients() const { return m_clients.computeSize() || m_clientsAwaitingCallback.computeSize(); }
	bool hasClient(const CachedResourceClient& client) { return m_clients.contains(client) || m_clientsAwaitingCallback.contains(client); }
	bool deleteIfPossible();

    enum PreloadResult : uint8_t {
        PreloadNotReferenced,
        PreloadReferenced,
        PreloadReferencedWhileLoading,
        PreloadReferencedWhileComplete
    };
    ALWAYS_INLINE PreloadResult preloadResult() const { return static_cast<PreloadResult>(m_preloadResult); }

    virtual void didAddClient(CachedResourceClient&);
    virtual void didRemoveClient(CachedResourceClient&) { }
    virtual void allClientsRemoved();
    void destroyDecodedDataIfNeeded();

    unsigned numberOfClients() const { return m_clients.computeSize(); }
    unsigned count() const { return numberOfClients(); }

    ALWAYS_INLINE Status status() const { return static_cast<Status>(m_status); }
    ALWAYS_INLINE void setStatus(Status status) { m_status = status; }

    ALWAYS_INLINE unsigned size() const { return encodedSize() + decodedSize() + overheadSize(); }
    unsigned encodedSize() const { return m_encodedSize; }
    unsigned decodedSize() const { return m_decodedSize; }
    unsigned overheadSize() const;
    
    ALWAYS_INLINE bool isLoaded() const { return !m_loading; } // FIXME. Method name is inaccurate. Loading might not have started yet.

    ALWAYS_INLINE bool isLoading() const { return m_loading; }
    ALWAYS_INLINE void setLoading(bool b) { m_loading = b; }
    virtual bool stillNeedsLoad() const { return false; }

    SubresourceLoader* loader() { return m_loader.get(); }

    bool areAllClientsXMLHttpRequests() const;

    bool isImage() const { return type() == ImageResource; }
    // FIXME: CachedRawResource could be a main resource, an audio/video resource, or a raw XHR/icon resource.
	bool isMainOrMediaOrFaviconOrRawResource() const { return type() == MainResource || type() == MediaResource || type() == Favicon || type() == RawResource; }

    // Whether this request should impact request counting and delay window.onload.
    bool ignoreForRequestCount() const
    {
        return m_ignoreForRequestCount
            || type() == MainResource
#if ENABLE(LINK_PREFETCH)
            || type() == LinkPrefetch
            || type() == LinkSubresource
#endif
            || type() == Favicon
            || type() == RawResource;
    }

    void setIgnoreForRequestCount(bool ignoreForRequestCount) { m_ignoreForRequestCount = ignoreForRequestCount; }

    ALWAYS_INLINE unsigned accessCount() const { return m_accessCount; }
    ALWAYS_INLINE void increaseAccessCount() { m_accessCount++; }

    // Computes the status of an object after loading.  
    // Updates the expire date on the cache entry file
    void finish();

    // Called by the cache if the object has been removed from the cache
    // while still being referenced. This means the object should delete itself
    // if the number of clients observing it ever drops to 0.
    // The resource can be brought back to cache after successful revalidation.
    ALWAYS_INLINE void setInCache(bool inCache) { m_inCache = inCache; }
    ALWAYS_INLINE bool inCache() const { return m_inCache; }

    void clearLoader();

    ALWAYS_INLINE SharedBuffer* resourceBuffer() const { return m_data.get(); }

    virtual void redirectReceived(ResourceRequest&, const ResourceResponse&);
    virtual void responseReceived(const ResourceResponse&);
    void setResponse(const ResourceResponse&);
	ALWAYS_INLINE const ResourceResponse& response() const { return m_response; }
    // This is the same as response() except after HTTP redirect to data: URL.
    const ResourceResponse& responseForSameOriginPolicyChecks() const;

    // Sets the serialized metadata retrieved from the platform's cache.
    void setSerializedCachedMetadata(const char*, size_t);

    // Caches the given metadata in association with this resource and suggests
    // that the platform persist it. The dataTypeID is a pseudo-randomly chosen
    // identifier that is used to distinguish data generated by the caller.
    void setCachedMetadata(unsigned dataTypeID, const char*, size_t);

    // Returns cached metadata of the given type associated with this resource.
    CachedMetadata* cachedMetadata(unsigned dataTypeID) const;

    void setCrossOrigin();
    bool isCrossOrigin() const;
    bool isCORSSameOrigin() const;
    ResourceResponse::Tainting responseTainting() const { return m_responseTainting; }

    void loadFrom(const CachedResource&);

    SecurityOrigin* origin() const { return m_origin.get(); }

    ALWAYS_INLINE bool canDelete() const { return !hasClients() && !m_loader && !m_preloadCount && !m_handleCount && !m_resourceToRevalidate && !m_proxyResource; }
    ALWAYS_INLINE bool hasOneHandle() const { return m_handleCount == 1; }

    bool isExpired() const;

    void cancelLoad();
    bool wasCanceled() const { return m_error.isCancellation(); }
    bool errorOccurred() const { return m_status == LoadError || m_status == DecodeError; }
    bool loadFailedOrCanceled() const { return !m_error.isNull(); }

    ALWAYS_INLINE bool sendResourceLoadCallbacks() const { return m_options.sendLoadCallbacks == SendCallbacks; }

    ALWAYS_INLINE bool shouldSendResourceLoadCallbacks() const { return m_options.sendLoadCallbacks == SendCallbacks; }
    ALWAYS_INLINE DataBufferingPolicy dataBufferingPolicy() const { return m_options.dataBufferingPolicy; }

    bool allowsCaching() const { return m_options.cachingPolicy == CachingPolicy::AllowCaching; }

    const FetchOptions& options() const { return m_options; }

    virtual void handleOutOfMemory(bool clearRawData) { }
    virtual void destroyDecodedData() { }

    ALWAYS_INLINE bool isPreloaded() const { return m_preloadCount; }
    ALWAYS_INLINE void increasePreloadCount() { ++m_preloadCount; }
    ALWAYS_INLINE void decreasePreloadCount() { ASSERT(m_preloadCount); --m_preloadCount; }
    bool isLinkPreload() const { return m_isLinkPreload; }
    void setLinkPreload() { m_isLinkPreload = true; }
    
    void registerHandle(CachedResourceHandleBase*);
    void unregisterHandle(CachedResourceHandleBase*);
    
    bool canUseCacheValidator() const;

    enum class RevalidationDecision { No, YesDueToCachePolicy, YesDueToNoStore, YesDueToNoCache, YesDueToExpired };
    virtual RevalidationDecision makeRevalidationDecision(CachePolicy) const;
    bool redirectChainAllowsReuse(ReuseExpiredRedirectionOrNot) const;
    bool hasRedirections() const { return m_redirectChainCacheStatus.status != RedirectChainCacheStatus::Status::NoRedirection;  }

    bool varyHeaderValuesMatch(const ResourceRequest&);

    ALWAYS_INLINE bool isCacheValidator() const { return m_resourceToRevalidate.get(); }
    ALWAYS_INLINE CachedResource* resourceToRevalidate() const { return m_resourceToRevalidate.get(); }
    
    // HTTP revalidation support methods for CachedResourceLoader.
    void setResourceToRevalidate(CachedResource&);
    virtual void switchClientsToRevalidatedResource();
    void clearResourceToRevalidate();
    void updateResponseAfterRevalidation(const ResourceResponse& validatingResponse);
    bool validationInProgress() const { return m_proxyResource.get(); }
	bool validationCompleting() const { return m_proxyResource && m_proxyResource->m_switchingClientsToRevalidatedResource; }

    virtual void didSendData(unsigned long long /* bytesSent */, unsigned long long /* totalBytesToBeSent */) { }
#if PLATFORM(CHROMIUM)
    virtual void didDownloadData(int) { }
#endif

    ALWAYS_INLINE void setLoadFinishTime(double finishTime) { m_loadFinishTime = finishTime; }
    ALWAYS_INLINE double loadFinishTime() const { return m_loadFinishTime; }

    void disableReuse() { m_canReuse = false; }
    virtual bool canReuse(const ResourceRequest&) const { return m_canReuse; }
    
    static ResourceLoadPriority defaultPriorityForResourceType(Type);

protected:
    virtual void checkNotify();

    // CachedResource constructor that may be used when the CachedResource can already be filled with response data.
    CachedResource(const KURL&, Type, SessionID = SessionID::defaultSessionID());

    void setEncodedSize(unsigned);
    void setDecodedSize(unsigned);
    void didAccessDecodedData(double timeStamp);
    
    WeakHashCountedSet<CachedResourceClient> m_clients;

    class CachedResourceCallback {
    	WTF_MAKE_FAST_ALLOCATED;
    public:
        static PassOwnPtr<CachedResourceCallback> schedule(CachedResource* resource, CachedResourceClient& client) { return adoptPtr(new CachedResourceCallback(resource, client)); }
        void cancel();
    private:
        CachedResourceCallback(CachedResource*, CachedResourceClient&);
        void timerFired();

        CachedResource* m_resource;
        CachedResourceClient& m_client;
        Timer<CachedResourceCallback> m_callbackTimer;
    };
    WeakHashMap<CachedResourceClient, OwnPtr<CachedResourceCallback> > m_clientsAwaitingCallback;

    ResourceRequest m_resourceRequest;
    SessionID m_sessionID;
    RefPtr<SubresourceLoader> m_loader;
    ResourceLoaderOptions m_options;
    ResourceLoadPriority m_loadPriority;

    ResourceResponse m_response;
    ResourceResponse m_redirectResponseForSameOriginPolicyChecks;
    ResourceResponse::Tainting m_responseTainting;
    double m_responseTimestamp;

    RefPtr<SharedBuffer> m_data;
    Timer<CachedResource> m_decodedDataDeletionTimer;

protected:
    bool addClientToSet(CachedResourceClient&);
    void decodedDataDeletionTimerFired();

    virtual void setBodyDataFrom(const CachedResource&);

    virtual bool mayTryReplaceEncodedData() const { return false; }

    double freshnessLifetime(const ResourceResponse&) const;

    void failBeforeStarting();

    String m_fragmentIdentifierForRequest;

    RefPtr<CachedMetadata> m_cachedMetadata;

    ResourceError m_error;
    RefPtr<SecurityOrigin> m_origin;

    double m_lastDecodedAccessTime; // Used as a "thrash guard" in the cache
    double m_loadFinishTime;
    double m_lastAccessTime;

    unsigned m_encodedSize;
    unsigned m_decodedSize;
    unsigned m_accessCount;
    unsigned m_handleCount;
    unsigned m_preloadCount;

    PreloadResult m_preloadResult;

    bool m_requestedFromNetworkingLayer : 1;

    bool m_inCache : 1;
    bool m_loading : 1;
    bool m_isLinkPreload : 1;
    bool m_canReuse : 1;

    bool m_switchingClientsToRevalidatedResource : 1;

    bool m_ignoreForRequestCount : 1;

    Type m_type;
    Status m_status;

#ifndef NDEBUG
    bool m_deleted;
    unsigned m_lruIndex;
#endif

    // If this field is non-null we are using the resource as a proxy for checking whether an existing resource is still up to date
    // using HTTP If-Modified-Since/If-None-Match headers. If the response is 304 all clients of this resource are moved
    // to to be clients of m_resourceToRevalidate and the resource is deleted. If not, the field is zeroed and this
    // resources becomes normal resource load.
    WeakPtr<CachedResource> m_resourceToRevalidate;

    // If this field is non-null, the resource has a proxy for checking whether it is still up to date (see m_resourceToRevalidate).
    WeakPtr<CachedResource> m_proxyResource;

    // These handles will need to be updated to point to the m_resourceToRevalidate in case we get 304 response.
    HashSet<CachedResourceHandleBase*> m_handlesToRevalidate;

    RedirectChainCacheStatus m_redirectChainCacheStatus;

    Vector<std::pair<String, String>> m_varyingHeaderValues;
};

}
