/*
    Copyright (C) 1998 Lars Knoll (knoll@mpi-hd.mpg.de)
    Copyright (C) 2001 Dirk Mueller (mueller@kde.org)
    Copyright (C) 2002 Waldo Bastian (bastian@kde.org)
    Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1301, USA.
*/

#include "config.h"
#include "MemoryCache.h"

#if ENABLE(BML)
#include "CachedClut.h"
#endif
#include "CachedCSSStyleSheet.h"
#include "CachedFont.h"
#include "CachedImage.h"
#include "CachedScript.h"
#include "CachedXSLStyleSheet.h"
#include "CachedResourceLoader.h"
#include "CrossThreadTask.h"
#include "Document.h"
#include "FrameLoader.h"
#include "FrameLoaderTypes.h"
#include "FrameView.h"
#include "Image.h"
#include "Logging.h"
#include "ResourceHandle.h"
#include "SecurityOrigin.h"
#include "SecurityOriginHash.h"
#include "WorkerGlobalScope.h"
#include "WorkerLoaderProxy.h"
#include "WorkerThread.h"
#include <stdio.h>
#include <wtf/CurrentTime.h>
#include <wtf/SetForScope.h>
#include <wtf/TemporaryChange.h>
#include <wtf/text/CString.h>

using namespace std;

namespace WebCore {

static const int cDefaultCacheCapacity = 16 * 1024 * 1024;
static const double cMinDelayBeforeLiveDecodedPrune = 1; // Seconds.
static const float cTargetPrunePercentage = .75f; // Percentage of capacity toward which we prune, to avoid immediately pruning again.
static const double cDefaultDecodedDataDeletionInterval = 0;

DLL_PUBLIC MemoryCache& MemoryCache::singleton()
{
    ASSERT(WTF::isMainThread());
    static NeverDestroyed<MemoryCache> memoryCache;
    return memoryCache;
}

MemoryCache::MemoryCache()
: m_disabled(false)
, m_pruneEnabled(true)
, m_inPruneResources(false)
, m_inHandleOutOfMemory(false)
, m_capacity(cDefaultCacheCapacity)
, m_minDeadCapacity(0)
, m_maxDeadCapacity(cDefaultCacheCapacity)
, m_decodedDataDeletionInterval(0)
, m_deadDecodedDataDeletionInterval(cDefaultDecodedDataDeletionInterval)
, m_deadFontPlatformDataDeletionInterval(8)
, m_acceleratedSurfaceDeletionInterval(0)
, m_simpleFontDataDeletionInterval(8)
, m_liveSize(0)
, m_deadSize(0)
, m_resourcesInUpdating(false)
, m_pruneTimer(this, &MemoryCache::pruneTimerFired)
{
}

bool MemoryCache::shouldRemoveFragmentIdentifier(const KURL& originalURL)
{
    if (!originalURL.hasFragmentIdentifier())
        return false;
    // Strip away fragment identifier from HTTP URLs.
    // Data URLs must be unmodified. For file and custom URLs clients may expect resources
    // to be unique even when they differ by the fragment identifier only.
    return originalURL.protocolIsInHTTPFamily();
}

KURL MemoryCache::removeFragmentIdentifierIfNeeded(const KURL& originalURL)
{
    if (!shouldRemoveFragmentIdentifier(originalURL))
        return originalURL;
    KURL url = originalURL;
    url.removeFragmentIdentifier();
    return url;
}

bool MemoryCache::add(CachedResource& resource)
{
    if (disabled())
        return false;

    ASSERT(WTF::isMainThread());

	m_resourcesInUpdating = true;
    m_resources.set(resource.url(), &resource);
	m_resourcesInUpdating = false;
	
    resource.setInCache(true);
    
    resourceAccessed(resource);
    
    LOG(ResourceLoading, "MemoryCache::add Added '%s', resource %p\n", resource.url().string().latin1().data(), resource);
    return true;
}

void MemoryCache::revalidationSucceeded(CachedResource& revalidatingResource, const ResourceResponse& response)
{
    CachedResourceHandle<CachedResource> protectedRevalidatingResource(revalidatingResource);
    CachedResource* resource = revalidatingResource.resourceToRevalidate();
    ASSERT(resource);
    ASSERT(!resource->inCache());
    ASSERT(resource->isLoaded());

    // Calling evict() can potentially delete revalidatingResource, which we use
    // below. This mustn't be the case since revalidation means it is loaded
    // and so canDelete() is false.
    ASSERT(!revalidatingResource->canDelete());

    remove(revalidatingResource);

    // A resource with the same URL may have been added back in the cache during revalidation.
    // In this case, we remove the cached resource and replace it with our freshly revalidated
    // one.
	if (auto* existingResource = m_resources.get(resource->url()))
		remove(*existingResource);

    ASSERT(!m_resources.get(resource->url()));
	m_resourcesInUpdating = true;
    m_resources.add(resource->url(), resource);
	m_resourcesInUpdating = false;
	
    resource->setInCache(true);
    resource->updateResponseAfterRevalidation(response);

	removeFromAccessOrderedResourcesList(*resource);
    insertInAccessOrderedResourcesList(*resource);

    int delta = resource->size();
    if (delta)
        adjustSize(resource->hasClients(), delta);
    
    revalidatingResource.switchClientsToRevalidatedResource();
    ASSERT(!revalidatingResource.m_deleted);
    // this deletes the revalidating resource
    if(revalidatingResource.resourceToRevalidate())
    	revalidatingResource.clearResourceToRevalidate();
}

void MemoryCache::revalidationFailed(CachedResource& revalidatingResource)
{
    ASSERT(WTF::isMainThread());
    LOG(ResourceLoading, "Revalidation failed for %p", revalidatingResource);
    // this deletes the revalidating resource
    if(revalidatingResource.resourceToRevalidate())
    	revalidatingResource.clearResourceToRevalidate();
}

CachedResource* MemoryCache::resourceForURL(const KURL& originalURL)
{
    ASSERT(WTF::isMainThread());

    KURL url = removeFragmentIdentifierIfNeeded(originalURL);

    CachedResource* resource = m_resources.get(url);
    if(resource && resource->errorOccurred()) {    	  
        revalidationFailed(*resource); // this deletes the revalidating resource
        evict(*resource);
        return nullptr;
    }
    
    return resource;
}

CachedResource* MemoryCache::resourceForRequest(const CachedResourceRequest& cachedResourceRequest)
{
    // FIXME: Change all clients to make sure HTTP(s) URLs have no fragment identifiers before calling here.
    // CachedResourceLoader is now doing this. Add an assertion once all other clients are doing it too.
	return resourceForURL(cachedResourceRequest.resourceRequest().url());
}

DLL_PUBLIC int MemoryCache::prune(int requiredSize, const Image* forImage, bool canReuse)
{
	int oldSize = (m_liveSize + m_deadSize);
	
	pruneDeadResources(true);
	
	if((oldSize - m_liveSize - m_deadSize) < requiredSize){
        // Make a copy of the LRUList first (and ref the resources) as calling
        // destroyDecodedData() can alter the LRUList.
        auto lruList = copyToVector(m_accessOrderedResources);

        for (auto& resource : lruList) {
			if (!resource)
				continue;

			if (!resource->inCache())
				continue;

			if(resource->isImage() && (static_cast<CachedImage&>(*resource).image() != forImage))
				resource->destroyDecodedData();

			if((oldSize - m_liveSize - m_deadSize) >= requiredSize)
				break;

			if(!canReuse)
				resource->disableReuse(); // do not reuse it, while delete when possible
		}
	}

    return (oldSize - (m_liveSize + m_deadSize));
}

DLL_PUBLIC int MemoryCache::handleOutOfMemory(int requiredSize, bool& noMemoryToFree)
{
	noMemoryToFree = false;

	int oldSize = (m_liveSize + m_deadSize);

	SetForScope<bool> inHandleOutOfMemory(m_inHandleOutOfMemory, true);

    // Make a copy of the LRUList first (and ref the resources) as calling
    // destroyDecodedData() can alter the LRUList.
    auto lruList = copyToVector(m_accessOrderedResources);

    int index = 0;
    for ( ; index < lruList.size(); ++index) {
    	auto& resource = lruList[index];

		if (!resource)
			continue;

		if (!resource->inCache())
			continue;

		resource->handleOutOfMemory(false);

        if((oldSize - m_liveSize - m_deadSize) >= requiredSize)
        	break;
	}

    if (index == lruList.size())
    	noMemoryToFree = true;

	return (oldSize - (m_liveSize + m_deadSize));
}

DLL_PUBLIC void MemoryCache::purge()
{
    // Make a copy of the LRUList first (and ref the resources) as calling
    // destroyDecodedData() can alter the LRUList.
    auto lruList = copyToVector(m_accessOrderedResources);

    for (auto& resource : lruList) {
		if (!resource)
			continue;

		if (!resource->inCache())
			continue;

		resource->purge();
	}
}

void MemoryCache::pruneLiveResources()
{
    if (!m_pruneEnabled)
        return;

    unsigned capacity = liveCapacity();
    if (capacity && m_liveSize <= capacity)
        return;

    unsigned targetSize = static_cast<unsigned>(capacity * cTargetPrunePercentage); // Cut by a percentage to avoid immediately pruning again.

    pruneLiveResourcesToSize(targetSize);
}

void MemoryCache::pruneLiveResourcesToPercentage(float prunePercentage)
{
    if (!m_pruneEnabled)
        return;

    if (prunePercentage < 0.0f  || prunePercentage > 0.95f)
        return;

    unsigned currentSize = m_liveSize + m_deadSize;
    unsigned targetSize = static_cast<unsigned>(currentSize * prunePercentage);

    pruneLiveResourcesToSize(targetSize);
}

void MemoryCache::pruneLiveResourcesToSize(unsigned targetSize)
{
    if (m_inPruneResources)
        return;

    SetForScope<bool> reentrancyProtector(m_inPruneResources, true);

    // Make a copy of the LRUList first (and ref the resources) as calling
    // destroyDecodedData() can alter the LRUList.
    auto lruList = copyToVector(m_accessOrderedResources);

    // Now evict objects from this list.
    // Remove from the head, since this is the least frequently accessed of the objects.
    for (auto& resource : lruList) {
        if (!resource)
            continue;

        if (!resource->inCache())
			continue;

        ASSERT(current->hasClients());
        if (resource->isLoaded() && resource->decodedSize()) {
        	resource->destroyDecodedData();

            if (targetSize && m_liveSize <= targetSize)
                return;
        }
    }
}

DLL_PUBLIC void MemoryCache::pruneDeadResources(bool all)
{
    if (!m_pruneEnabled)
        return;

    if(all){
    	pruneDeadResourcesToSize(0);
    	return;
    }
    
    unsigned capacity = deadCapacity();
    if (capacity && m_deadSize <= capacity)
        return;

    unsigned targetSize = static_cast<unsigned>(capacity * cTargetPrunePercentage); // Cut by a percentage to avoid immediately pruning again.
    pruneDeadResourcesToSize(targetSize);
}

void MemoryCache::pruneDeadResourcesToPercentage(float prunePercentage)
{
    if (!m_pruneEnabled)
        return;

    if (prunePercentage < 0.0f  || prunePercentage > 0.95f)
        return;

    unsigned currentSize = m_liveSize + m_deadSize;
    unsigned targetSize = static_cast<unsigned>(currentSize * prunePercentage);

    pruneDeadResourcesToSize(targetSize);
}

void MemoryCache::pruneDeadResourcesToSize(unsigned targetSize)
{
    if (m_inPruneResources)
        return;

    TemporaryChange<bool> reentrancyProtector(m_inPruneResources, true);

    // Make a copy of the LRUList first (and ref the resources) as calling
    // destroyDecodedData() can alter the LRUList.
    auto lruList = copyToVector(m_accessOrderedResources);

    // Now evict objects from this list.
    // Remove from the head, since this is the least frequently accessed of the objects.
    for (auto& resource : lruList) {
        if (!resource)
            continue;

        LOG(ResourceLoading, " lru resource %p - in cache %d, has clients %d, preloaded %d, loaded %d", resource.get(), resource->inCache(), resource->hasClients(), resource->isPreloaded(), resource->isLoaded());
        if (!resource->inCache())
            continue;

        if (!resource->hasClients() && !resource->isPreloaded() && !resource->isCacheValidator()) {
			evict(*resource);
			if (targetSize && m_deadSize <= targetSize)
				return;
		}
	}
}

DLL_PUBLIC void MemoryCache::pruneDeadResourcesSinceLastAccess(double interval)
{
    if (m_inPruneResources)
        return;

    if(m_deadSize == 0)
    	return;

    SetForScope<bool> reentrancyProtector(m_inPruneResources, true);

    double lastAccessTime = WTF::monotonicallyIncreasingTime() - interval;

    // Make a copy of the LRUList first (and ref the resources) as calling
    // destroyDecodedData() can alter the LRUList.
    auto lruList = copyToVector(m_accessOrderedResources);

    // Now evict objects from this list.
    // Remove from the head, since this is the least frequently accessed of the objects.
    for (auto& resource : lruList) {
        if (!resource)
            continue;

        LOG(ResourceLoading, " lru resource %p - in cache %d, has clients %d, preloaded %d, loaded %d", resource.get(), resource->inCache(), resource->hasClients(), resource->isPreloaded(), resource->isLoaded());
        if (!resource->inCache())
            continue;

		if (!resource->hasClients() && !resource->isPreloaded() && !resource->isCacheValidator()) {
			if(resource->m_lastAccessTime < lastAccessTime)
				evict(*resource);
		}
	}
}

DLL_PUBLIC void MemoryCache::setCapacities(unsigned minDeadBytes, unsigned maxDeadBytes, unsigned totalBytes)
{
    ASSERT(minDeadBytes <= maxDeadBytes);
    ASSERT(maxDeadBytes <= totalBytes);
    m_minDeadCapacity = minDeadBytes;
    m_maxDeadCapacity = maxDeadBytes;
    m_capacity = totalBytes;
    prune();
}

DLL_PUBLIC int MemoryCache::evict(const String& urlPrefix)
{
	int total = 0;

	Vector<WeakPtr<CachedResource>> resourcesToEvict;
	for (auto& resource : m_resources.values()) {
		if(resource->url().string().startsWith(urlPrefix))
			resourcesToEvict.append(makeWeakPtr(*resource));
	}

	for (auto& resource : resourcesToEvict) {
		if (!resource)
			continue;

		//fprintf(stderr, ">>>>Cache::evict(%s)\n", resource->url().latin1().data());

		++total;

		// this deletes the revalidating resource
		revalidationFailed(*resource);
		evict(*resource);
	}
    
    return total;
}

bool MemoryCache::evict(CachedResource& resource)
{
    ASSERT(WTF::isMainThread());
    LOG(ResourceLoading, "Evicting resource %p for '%s' from cache", resource, resource->url().string().latin1().data());
    // The resource may have already been removed by someone other than our caller,
    // who needed a fresh copy for a reload. See <http://bugs.webkit.org/show_bug.cgi?id=12479#c6>.
    if (resource.inCache()) {
        // Remove from the resource map.

    	m_resourcesInUpdating = true;
        m_resources.remove(resource.url());
    	m_resourcesInUpdating = false;
    	
        resource.setInCache(false);

        // Remove from the appropriate LRU list.
        removeFromAccessOrderedResourcesList(resource);
        adjustSize(resource.hasClients(), -static_cast<int>(resource.size()));
    } else
        ASSERT(m_resources.get(resource->url()) != resource);

    if (resource.canDelete()) {
        delete &resource;
        return true;
    }

    if(!resource.hasClients())
    	resource.destroyDecodedDataIfNeeded();

    return false;
}

void MemoryCache::insertInAccessOrderedResourcesList(CachedResource& resource)
{
    ASSERT(resource.inCache());
    ASSERT(resource.accessCount() > 0);

    auto addResult = m_accessOrderedResources.add(resource);
    ASSERT_UNUSED(addResult, addResult.isNewEntry);
}

void MemoryCache::removeFromAccessOrderedResourcesList(CachedResource& resource)
{
	bool removed = m_accessOrderedResources.remove(resource);
    ASSERT_UNUSED(removed, removed);
}

void MemoryCache::didAccessDecodedData(CachedResource& resource)
{
	removeFromAccessOrderedResourcesList(resource);
	insertInAccessOrderedResourcesList(resource);
}

void MemoryCache::resourceAccessed(CachedResource& resource)
{
    ASSERT(resource.inCache());
    
    // Need to make sure to remove before we increase the access count, since
    // the queue will possibly change.
    removeFromAccessOrderedResourcesList(resource);
    
    // If this is the first time the resource has been accessed, adjust the size of the cache to account for its initial size.
    if (!resource.accessCount())
        adjustSize(resource.hasClients(), resource.size());
    
    // Add to our access count.
    resource.increaseAccessCount();
    
    // Now insert into the new queue.
    insertInAccessOrderedResourcesList(resource);
}

void MemoryCache::removeResourcesWithOrigin(SecurityOrigin* origin)
{
    Vector<WeakPtr<CachedResource>> resourcesWithOrigin;

    for (auto* resource : m_resources.values()) {
        RefPtr<SecurityOrigin> resourceOrigin = SecurityOrigin::createFromString(resource->url());
        if (!resourceOrigin)
            continue;
        if (resourceOrigin->equal(origin))
            resourcesWithOrigin.append(makeWeakPtr(*resource));
    }

    for (auto& resource : resourcesWithOrigin) {
    	if (!resource)
    		continue;
        remove(*resource);
    }
}

void MemoryCache::getOriginsWithCache(SecurityOriginSet& origins)
{
    for (auto* resource : m_resources.values())
        origins.add(SecurityOrigin::createFromString(resource->url()));
}

void MemoryCache::addToLiveResourcesSize(CachedResource* resource)
{
    m_liveSize += resource->size();
    m_deadSize -= resource->size();
}

void MemoryCache::removeFromLiveResourcesSize(CachedResource* resource)
{
    m_liveSize -= resource->size();
    m_deadSize += resource->size();
}

void MemoryCache::adjustSize(bool live, int delta)
{
    if (live) {
        ASSERT(delta >= 0 || ((int)m_liveSize + delta >= 0));
        m_liveSize += delta;
    } else {
        ASSERT(delta >= 0 || ((int)m_deadSize + delta >= 0));
        m_deadSize += delta;
    }
}


void MemoryCache::removeUrlFromCache(ScriptExecutionContext* context, const String& urlString) 
{
#if ENABLE(WORKERS)
    if (context->isWorkerGlobalScope()) {
      WorkerGlobalScope* workerGlobalScope = static_cast<WorkerGlobalScope*>(context);
      workerGlobalScope->thread()->workerLoaderProxy().postTaskToLoader(
          createCallbackTask(&removeUrlFromCacheImpl, urlString));
      return;
    }
#endif
    removeUrlFromCacheImpl(context, urlString);
}

void MemoryCache::removeUrlFromCacheImpl(ScriptExecutionContext*, const String& urlString)
{
    KURL url(KURL(), urlString);

    auto& memoryCache = MemoryCache::singleton();
    if (CachedResource* resource = memoryCache.resourceForURL(url))
    	memoryCache.remove(*resource);
}

void MemoryCache::TypeStatistic::addResource(CachedResource* o)
{
    count++;
    size += o->size();
    liveSize += o->hasClients() ? o->size() : 0;
    decodedSize += o->decodedSize();
}

MemoryCache::Statistics MemoryCache::getStatistics()
{
    Statistics stats;
    CachedResourceMap::iterator end = m_resources.end();
    for (CachedResourceMap::iterator i = m_resources.begin(); i != end; ++i) {
        CachedResource* resource = i->second;
        switch (resource->type()) {
        case CachedResource::ImageResource:
            stats.images.addResource(resource);
            break;
        case CachedResource::CSSStyleSheet:
            stats.cssStyleSheets.addResource(resource);
            break;
        case CachedResource::Script:
            stats.scripts.addResource(resource);
            break;
#if ENABLE(XSLT)
        case CachedResource::XSLStyleSheet:
            stats.xslStyleSheets.addResource(resource);
            break;
#endif
        case CachedResource::FontResource:
#if ENABLE(SVG_FONTS)
		case CachedResource::SVGFontResource:
#endif
            stats.fonts.addResource(resource);
            break;
        case CachedResource::RawResource:
            stats.rawResources.addResource(resource);
            break;
        default:
        	stats.others.addResource(resource);
            break;
        }
    }
    return stats;
}

void MemoryCache::setDisabled(bool disabled)
{
    m_disabled = disabled;
    if (!m_disabled)
        return;

	Vector<WeakPtr<CachedResource>> resourcesToEvict;
	resourcesToEvict.reserveInitialCapacity(m_resources.size());
	for (auto& resource : m_resources.values())
		resourcesToEvict.append(makeWeakPtr(*resource));

	for (auto& resource : resourcesToEvict) {
		if (!resource)
			continue;
    	        
        revalidationFailed(*resource); // this deletes the revalidating resource
        evict(*resource);
    }
}

void MemoryCache::evictResources()
{
    if (disabled())
        return;

    setDisabled(true);
    setDisabled(false);
}

void MemoryCache::pruneTimerFired()
{
    prune();
}

void MemoryCache::pruneSoon()
{
     if (m_pruneTimer.isActive())
        return;
     if (!needsPruning())
         return;
     m_pruneTimer.startOneShot(0);
}

void MemoryCache::pruneToPercentage(float targetPercentLive)
{
    pruneDeadResourcesToPercentage(targetPercentLive); // Prune dead first, in case it was "borrowing" capacity from live.
    pruneLiveResourcesToPercentage(targetPercentLive);
}

static void dumpStat(const char* name, MemoryCache::TypeStatistic& statistics, int (*logFunction)(const char * format, ... ))
{
	if(statistics.count == 0)
		return;
	logFunction("%-13s %13d %13d %13d %13d\n", name, statistics.count, statistics.size, statistics.liveSize, statistics.decodedSize);
}

DLL_PUBLIC void MemoryCache::dumpStats(int (*logFunction)(const char * format, ... ))
{
	if(m_resourcesInUpdating){
		logFunction("    >Updating resources, forbidden statistics\n");		    
		return;
	}
	
    logFunction("%-13s %-13s %-13s %-13s %-13s\n", "", "        Count", "         Size", "     LiveSize", "  DecodedSize");
    logFunction("%-13s %-13s %-13s %-13s %-13s\n", "-------------", "-------------", "-------------", "-------------", "-------------");

	if(m_resources.size() == 0)
		return;

    Statistics s = getStatistics();
    dumpStat("Images", s.images, logFunction);
    dumpStat("CSS", s.cssStyleSheets, logFunction);
#if ENABLE(XSLT)
    dumpStat("XSL", s.xslStyleSheets, logFunction);
#endif
    dumpStat("JavaScript", s.scripts, logFunction);
    dumpStat("Fonts", s.fonts, logFunction);
    dumpStat("Raw", s.rawResources, logFunction);
    dumpStat("Others", s.others, logFunction);
    logFunction("%-13s %-13s %-13s %-13s %-13s\n", "-------------", "-------------", "-------------", "-------------", "-------------");
}

extern HashSet<CachedResource*> gCachedResources;

DLL_PUBLIC void MemoryCache::dumpLeakedResources(int (*logFunction)(const char * format, ... ))
{
	for(auto iter = gCachedResources.begin(); iter != gCachedResources.end(); ++iter) {
		CachedResource& resource = **iter;

		if(resource.inCache())
			continue;

		CString url = resource.url().string().utf8();
		if(resource.hasClients())
			logFunction("    >hasClients             %s\n", url.data());
		else if(resource.m_loader)
			logFunction("    >m_loader               %s\n", url.data());
		else if(resource.m_preloadCount)
			logFunction("    >m_preloadCount         %s\n", url.data());
		else if(resource.m_handleCount)
			logFunction("    >m_handleCount          %s %i\n", url.data(), resource.m_handleCount);
		else if(resource.m_resourceToRevalidate)
			logFunction("    >m_resourceToRevalidate %s\n", url.data());
		else if(resource.m_proxyResource)
			logFunction("    >m_proxyResource        %s\n", url.data());
		else
			logFunction("    >unknown                %s\n", url.data());
	}
}

#ifndef NDEBUG
void MemoryCache::dumpLRULists(bool includeLive) const
{
    printf("LRU-SP lists in eviction order (Kilobytes decoded, Kilobytes encoded, Access count, Referenced):\n");

    int size = m_allResources.size();
    for (int i = size - 1; i >= 0; i--) {
        printf("\n\nList %d: ", i);
        CachedResource* current = m_allResources[i].m_tail;
        while (current) {
            CachedResource* prev = current->m_prevInAllResourcesList;
            if (includeLive || !current->hasClients())
                printf("(%.1fK, %.1fK, %uA, %dR); ", current->decodedSize() / 1024.0f, (current->encodedSize() + current->overheadSize()) / 1024.0f, current->accessCount(), current->hasClients());

            current = prev;
        }
    }
}
#endif

} // namespace WebCore
