/*
 * Copyright (C) 2007, 2009 Apple Inc. All rights reserved.
 * Copyright (C) 2012 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "DOMSelection.h"

#include "Document.h"
#include "ExceptionCode.h"
#include "Frame.h"
#include "FrameSelection.h"
#include "Node.h"
#include "PlatformString.h"
#include "Range.h"
#include "Settings.h"
#include "SimpleRange.h"
#include "TextIterator.h"
#include "TreeScope.h"
#include "htmlediting.h"

namespace WebCore {

static RefPtr<Node> selectionShadowAncestor(Document& document)
{
    ASSERT(!document.settings().liveRangeSelectionEnabled());
    auto* node = document.selection().selection().base().anchorNode();
    if (!node || !node->isInShadowTree())
        return nullptr;
    return node->document().ancestorNodeInThisScope(node);
}

inline DOMSelection::DOMSelection(DOMWindow& window)
    : DOMWindowProperty(&window)
{
}

Ref<DOMSelection> DOMSelection::create(DOMWindow& window)
{
	return adoptRef(*new DOMSelection(window));
}

RefPtr<Frame> DOMSelection::frame() const
{
    return DOMWindowProperty::frame();
}

Optional<SimpleRange> DOMSelection::range() const
{
    auto frame = this->frame();
    if (!frame)
        return WTF::Nullopt;
    auto* document = frame->document();
    if (!document)
    	return Nullopt;
    auto& selection = document->selection().selection();
    auto range = document->settings().liveRangeSelectionEnabled()
        ? selection.range()
        : selection.firstRange();
    if (!range || range->start.container->isInShadowTree())
        return WTF::Nullopt;
    return range;
}

Position DOMSelection::anchorPosition() const
{
    auto frame = this->frame();
    if (!frame)
        return { };
    auto* document = frame->document();
    if (!document)
    	return { };
    auto& selection = document->selection().selection();
    if (frame->settings().liveRangeSelectionEnabled())
        return selection.anchor();
    return (selection.isBaseFirst() ? selection.start() : selection.end()).parentAnchoredEquivalent();
}

Position DOMSelection::focusPosition() const
{
    auto frame = this->frame();
    if (!frame)
        return { };
    auto* document = frame->document();
    if (!document)
    	return { };
    auto& selection = document->selection().selection();
    if (frame->settings().liveRangeSelectionEnabled())
        return selection.focus();
    return (selection.isBaseFirst() ? selection.end() : selection.start()).parentAnchoredEquivalent();
}

Position DOMSelection::basePosition() const
{
	// FIXME: Remove this once liveRangeSelectionEnabled is always on, since base and anchor should be the same thing.
	auto frame = this->frame();
    if (!frame)
        return { };
    auto* document = frame->document();
    if (!document)
    	return { };
    auto& selection = document->selection().selection();
    if (frame->settings().liveRangeSelectionEnabled())
        return selection.anchor();
    return selection.base().parentAnchoredEquivalent();
}

Position DOMSelection::extentPosition() const
{
	// FIXME: Remove this once liveRangeSelectionEnabled is always on, since extent and focus should be the same thing.
	auto frame = this->frame();
    if (!frame)
        return { };
    auto* document = frame->document();
    if (!document)
    	return { };
    auto& selection = document->selection().selection();
    if (frame->settings().liveRangeSelectionEnabled())
        return selection.focus();
    return selection.extent().parentAnchoredEquivalent();
}

RefPtr<Node> DOMSelection::anchorNode() const
{
    return shadowAdjustedNode(anchorPosition());
}

unsigned DOMSelection::anchorOffset() const
{
    return shadowAdjustedOffset(anchorPosition());
}

RefPtr<Node> DOMSelection::focusNode() const
{
    return shadowAdjustedNode(focusPosition());
}

unsigned DOMSelection::focusOffset() const
{
    return shadowAdjustedOffset(focusPosition());
}

RefPtr<Node> DOMSelection::baseNode() const
{
    return shadowAdjustedNode(basePosition());
}

unsigned DOMSelection::baseOffset() const
{
    return shadowAdjustedOffset(basePosition());
}

RefPtr<Node> DOMSelection::extentNode() const
{
    return shadowAdjustedNode(extentPosition());
}

unsigned DOMSelection::extentOffset() const
{
    return shadowAdjustedOffset(extentPosition());
}

bool DOMSelection::isCollapsed() const
{
    auto frame = this->frame();
    if (!frame)
        return true;
    auto range = this->range();
    return !range || range->collapsed();
}

String DOMSelection::type() const
{
    auto frame = this->frame();
    if (!frame)
        return ASCIILiteral("None");
    auto* document = frame->document();
    if (!document)
        return ASCIILiteral("None");
    auto& selection = document->selection();
    if (frame->settings().liveRangeSelectionEnabled())
    	return !selection.isInDocumentTree() ? ASCIILiteral("None") : range()->collapsed() ? ASCIILiteral("Caret") : ASCIILiteral("Range");
	if (selection.isNone())
        return "None";
    if (selection.isCaret())
        return "Caret";
    return "Range";
}

unsigned DOMSelection::rangeCount() const
{
    auto frame = this->frame();
    if (!frame)
        return 0;
    auto* document = frame->document();
    if (!document)
        return 0;
    if (frame->settings().liveRangeSelectionEnabled())
        return document->selection().isInDocumentTree();
    return document->selection().isNone() ? 0 : 1;
}

void DOMSelection::collapse(Node* node, unsigned offset, ExceptionCode& ec)
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
    if (frame->settings().liveRangeSelectionEnabled()) {
        if (!node) {
            removeAllRanges();
            return;
        }
        auto* result = Range::checkNodeOffsetPair(*node, offset, ec);
        if (ec)
            return;
        if (!document->contains(*node))
            return;
    } else {
        if (!isValidForPosition(node))
            return;
    }
    auto& selection = document->selection();
    selection.disassociateLiveRange();
    selection.moveTo(makeContainerOffsetPosition(node, offset), Affinity::Downstream);
}

void DOMSelection::collapseToEnd(ExceptionCode& ec)
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;

    auto& selection = document->selection();
    if (selection.isNone()) {
        ec = INVALID_STATE_ERR;
        return;
    }
    if (frame->settings().liveRangeSelectionEnabled()) {
        selection.disassociateLiveRange();
        selection.moveTo(selection.selection().uncanonicalizedEnd(), Affinity::Downstream);
    } else
    	selection.moveTo(VisiblePosition(selection.selection().end(), Affinity::Downstream));
}

void DOMSelection::collapseToStart(ExceptionCode& ec)
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;

    auto& selection = document->selection();
    if (selection.isNone()) {
        ec = INVALID_STATE_ERR;
        return;
    }
    if (frame->settings().liveRangeSelectionEnabled()) {
        selection.disassociateLiveRange();
        selection.moveTo(selection.selection().uncanonicalizedStart(), Affinity::Downstream);
    } else
    	selection.moveTo(VisiblePosition(selection.selection().start(), Affinity::Downstream));
}

void DOMSelection::empty()
{
    removeAllRanges();
}

void DOMSelection::setBaseAndExtent(Node* baseNode, unsigned baseOffset, Node* extentNode, unsigned extentOffset, ExceptionCode& ec)
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
    if (frame->settings().liveRangeSelectionEnabled()) {
        // FIXME: We should do this by making the arguments non-nullable in the IDL file, once liveRangeSelectionEnabled is always true.
        if (!baseNode || !extentNode) {
            ec = TypeError;
            return;
        }
        Range::checkNodeOffsetPair(*baseNode, baseOffset, ec);
        if (ec)
            return;
        Range::checkNodeOffsetPair(*extentNode, extentOffset, ec);
        if (ec)
            return;
        if (!document->contains(*baseNode) || !document->contains(*extentNode))
            return;
    } else {
        if (!isValidForPosition(baseNode) || !isValidForPosition(extentNode))
            return;
    }
    auto& selection = document->selection();
    selection.disassociateLiveRange();
    selection.moveTo(makeContainerOffsetPosition(baseNode, baseOffset), makeContainerOffsetPosition(extentNode, extentOffset), Affinity::Downstream);
}

void DOMSelection::setPosition(Node* node, unsigned offset, ExceptionCode& ec)
{
    return collapse(node, offset, ec);
}

void DOMSelection::modify(const String& alterString, const String& directionString, const String& granularityString)
{
    FrameSelection::EAlteration alter;
    if (equalLettersIgnoringASCIICase(alterString, "extend"))
        alter = FrameSelection::AlterationExtend;
    else if (equalLettersIgnoringASCIICase(alterString, "move"))
        alter = FrameSelection::AlterationMove;
    else
        return;

    SelectionDirection direction;
    if (equalLettersIgnoringASCIICase(directionString, "forward"))
        direction = DirectionForward;
    else if (equalLettersIgnoringASCIICase(directionString, "backward"))
        direction = DirectionBackward;
    else if (equalLettersIgnoringASCIICase(directionString, "left"))
        direction = DirectionLeft;
    else if (equalLettersIgnoringASCIICase(directionString, "right"))
        direction = DirectionRight;
    else
        return;

    TextGranularity granularity;
    if (equalLettersIgnoringASCIICase(granularityString, "character"))
        granularity = CharacterGranularity;
    else if (equalLettersIgnoringASCIICase(granularityString, "word"))
        granularity = WordGranularity;
    else if (equalLettersIgnoringASCIICase(granularityString, "sentence"))
        granularity = SentenceGranularity;
    else if (equalLettersIgnoringASCIICase(granularityString, "line"))
        granularity = LineGranularity;
    else if (equalLettersIgnoringASCIICase(granularityString, "paragraph"))
        granularity = ParagraphGranularity;
    else if (equalLettersIgnoringASCIICase(granularityString, "lineboundary"))
        granularity = LineBoundary;
    else if (equalLettersIgnoringASCIICase(granularityString, "sentenceboundary"))
        granularity = SentenceBoundary;
    else if (equalLettersIgnoringASCIICase(granularityString, "paragraphboundary"))
        granularity = ParagraphBoundary;
    else if (equalLettersIgnoringASCIICase(granularityString, "documentboundary"))
        granularity = DocumentBoundary;
    else
        return;

    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
	document->selection().modify(alter, direction, granularity);
}

void DOMSelection::extend(Node& node, unsigned offset, ExceptionCode& ec)
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
    if (frame->settings().liveRangeSelectionEnabled()) {
        if (!document->contains(node))
            return;
        Range::checkNodeOffsetPair(node, offset, ec);
        if (ec)
            return;
        auto& selection = document->selection();
        auto newSelection = selection.selection();
        newSelection.setExtent(makeContainerOffsetPosition(&node, offset));
        selection.disassociateLiveRange();
        selection.setSelection(newSelection);
    } else {
		if (offset > node.length()) {
			ec = INDEX_SIZE_ERR;
			return;
		}
		if (!isValidForPosition(&node))
			return;
        document->selection().setExtent(makeContainerOffsetPosition(&node, offset), Affinity::Downstream);
    }
}

PassRefPtr<Range> DOMSelection::getRangeAt(unsigned index, ExceptionCode& ec)
{
    if (index >= rangeCount()) {
        ec = INDEX_SIZE_ERR;
        return nullptr;
    }
    auto frame = this->frame();
    if (!frame)
        return nullptr;
    auto* document = frame->document();
    if (!document)
        return nullptr;
    if (frame->settings().liveRangeSelectionEnabled())
        return document->selection().associatedLiveRange();
    if (auto shadowAncestor = selectionShadowAncestor(*document))
        return createLiveRange(makeSimpleRange(makeBoundaryPointBeforeNode(*shadowAncestor)));
    return createLiveRange(document->selection().selection().firstRange());
}

void DOMSelection::removeAllRanges()
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
    document->selection().clear();
}

void DOMSelection::addRange(Range& liveRange)
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
    auto& selection = document->selection();
    if (frame->settings().liveRangeSelectionEnabled()) {
        if (selection.isNone())
            selection.associateLiveRange(liveRange);
        return;
    }
    auto range = makeSimpleRange(liveRange);
    if (auto selectedRange = selection.selection().toNormalizedRange()) {
        if (!selectedRange->start.container->containingShadowRoot() && intersects<Tree>(*selectedRange, range))
            selection.setSelection(unionRange(*selectedRange, range));
        return;
    }
    selection.setSelection(range);
}

void DOMSelection::removeRange(Range& liveRange, ExceptionCode& ec)
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
    ASSERT(frame->settings().liveRangeSelectionEnabled());
    if (&liveRange != document->selection().associatedLiveRange()){
    	ec = NOT_FOUND_ERR;
    	return;
    }
    removeAllRanges();
    return;
}

void DOMSelection::deleteFromDocument()
{
    auto frame = this->frame();
    if (!frame)
        return;
    auto* document = frame->document();
    if (!document)
        return;
    if (frame->settings().liveRangeSelectionEnabled()) {
        if (auto range = document->selection().associatedLiveRange())
            range->deleteContents();
        return;
    }
    auto selectedRange = document->selection().selection().toNormalizedRange();
    if (!selectedRange || selectedRange->start.container->containingShadowRoot())
        return;
    createLiveRange(*selectedRange)->deleteContents();
    auto container = selectedRange->start.container.ptr();
    auto offset = selectedRange->start.offset;
    setBaseAndExtent(container, offset, container, offset);
}

bool DOMSelection::containsNode(Node& node, bool allowPartial) const
{
	// FIXME: The rule implemented here for text nodes is wrong, and was added to work around anomalies caused when we canonicalize selection endpoints.
	if (node.isTextNode() && !node.document().settings().liveRangeSelectionEnabled())
        allowPartial = true;
    auto range = this->range();
    return range && (allowPartial ? intersects<Tree>(*range, node) : contains<Tree>(*range, node));
}

void DOMSelection::selectAllChildren(Node& node, ExceptionCode& ec)
{
    // This doesn't (and shouldn't) select text node characters.
    setBaseAndExtent(&node, 0, &node, node.countChildNodes(), ec);
}

String DOMSelection::toString() const
{
    auto frame = this->frame();
    if (!frame)
        return String();
    auto* document = frame->document();
    if (!document)
        return String();
    if (frame->settings().liveRangeSelectionEnabled()) {
        auto range = this->range();
        return range ? plainText(*range) : emptyString();
    }
    auto range = document->selection().selection().firstRange();
    return range ? plainText(*range) : emptyString();
}

RefPtr<Node> DOMSelection::shadowAdjustedNode(const Position& position) const
{
    if (position.isNull())
        return nullptr;

    auto frame = this->frame();
    if (!frame)
    	return nullptr;

    if (frame->settings().liveRangeSelectionEnabled()) {
        auto node = position.containerNode();
        return !node || node->isInShadowTree() ? nullptr : node;
    }
    auto* document = frame->document();
    if (!document)
        return nullptr;
    Node* containerNode = position.containerNode();
    Node* adjustedNode = document->ancestorNodeInThisScope(containerNode);

    if (!adjustedNode)
        return nullptr;

    if (containerNode == adjustedNode)
        return containerNode;

    return adjustedNode->parentNodeGuaranteedHostFree();
}

int DOMSelection::shadowAdjustedOffset(const Position& position) const
{
    if (position.isNull())
        return 0;

    auto frame = this->frame();
    if (!frame)
    	return 0;
    if (frame->settings().liveRangeSelectionEnabled())
        return shadowAdjustedNode(position) ? position.computeOffsetInContainerNode() : 0;

    if (!frame->document())
    	return 0;
    Node* containerNode = position.containerNode();
    Node* adjustedNode = frame->document()->ancestorNodeInThisScope(containerNode);

    if (!adjustedNode)
        return 0;

    if (containerNode == adjustedNode)
        return position.computeOffsetInContainerNode();

    return adjustedNode->nodeIndex();
}

bool DOMSelection::isValidForPosition(Node* node) const
{
    auto frame = this->frame();
    ASSERT(!frame->settings().liveRangeSelectionEnabled());
    return frame && (!node || &node->document() == frame->document());
}

} // namespace WebCore
