/*
 * Copyright (C) 2006, 2007, 2008, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "DOMWindow.h"

#include "BackForwardController.h"
#include "BarProp.h"
#include "BeforeUnloadEvent.h"
#include "ContentSecurityPolicy.h"
#include "CSSComputedStyleDeclaration.h"
#include "CSSRule.h"
#include "CSSRuleList.h"
#include "Chrome.h"
#include "ChromeClient.h"
#include "ConsoleTypes.h"
#include "Crypto.h"
#include "DOMApplicationCache.h"
#include "DOMSelection.h"
#include "DOMStringList.h"
#include "DOMTimer.h"
#include "DOMTokenList.h"
#include "DOMURL.h"
#include "DOMWindowExtension.h"
#include "DOMWindowNotifications.h"
#include "DeviceMotionController.h"
#include "DeviceOrientationController.h"
#include "Document.h"
#include "DocumentLoader.h"
#include "Editor.h"
#include "Element.h"
#include "EventException.h"
#include "EventListener.h"
#include "EventNames.h"
#include "EventPath.h"
#include "ExceptionCode.h"
#include "FloatRect.h"
#include "FocusController.h"
#include "Frame.h"
#include "FrameLoadRequest.h"
#include "FrameLoader.h"
#include "FrameLoaderClient.h"
#include "FrameTree.h"
#include "FrameView.h"
#include "HTMLCanvasElement.h"
#include "HTMLFrameOwnerElement.h"
#include "HTMLImageElement.h"
#include "HTMLVideoElement.h"
#include "History.h"
#include "IdleRequestCallback.h"
#include "IdleRequestOptions.h"
#include "ImageData.h"
#include "InspectorInstrumentation.h"
#include "JSDOMPromiseDeferred.h"
#include "JSExecState.h"
#include "KURL.h"
#include "Location.h"
#include "MediaQueryList.h"
#include "MediaQueryMatcher.h"
#include "MessageEvent.h"
#include "Navigator.h"
#include "NotificationCenter.h"
#include "NotificationController.h"
#include "Page.h"
#include "PageConsoleClient.h"
#include "PageGroup.h"
#include "PageTransitionEvent.h"
#include "Performance.h"
#include "PlatformScreen.h"
#include "RuntimeEnabledFeatures.h"
#include "ScheduledAction.h"
#include "Screen.h"
#include "ScriptCallStack.h"
#include "ScriptCallStackFactory.h"
#include "SecurityOrigin.h"
#include "SerializedScriptValue.h"
#include "Settings.h"
#include "Storage.h"
#include "StorageArea.h"
#include "StorageNamespace.h"
#include "StyleMedia.h"
#include "StyleResolver.h"
#include "StyleScope.h"
#include "SuddenTermination.h"
#include "WebKitPoint.h"
#include "WindowFeatures.h"
#include "WindowFocusAllowedIndicator.h"
#include <algorithm>
#include <wtf/CurrentTime.h>
#include <wtf/MainThread.h>
#include <wtf/MathExtras.h>
#include <wtf/text/WTFString.h>
#include "OfflineCache.h"

#if ENABLE(REQUEST_ANIMATION_FRAME)
#include "RequestAnimationFrameCallback.h"
#endif

#if ENABLE(RMAI)
#include "GalioDebug.h"
#include "GalioWafer.h"
#include "MOM.h"
#endif

using std::min;
using std::max;

namespace WebCore {

class PostMessageTimer FINAL : public TimerBase {
public:
    PostMessageTimer(DOMWindow* window, PassRefPtr<SerializedScriptValue> message, const String& sourceOrigin, PassRefPtr<DOMWindow> source, PassOwnPtr<MessagePortChannelArray> channels, SecurityOrigin* targetOrigin, PassRefPtr<ScriptCallStack> stackTrace)
        : m_window(window)
        , m_message(message)
        , m_origin(sourceOrigin)
        , m_source(source)
        , m_channels(channels)
        , m_targetOrigin(targetOrigin)
        , m_stackTrace(stackTrace)
    {
    }

    PassRefPtr<MessageEvent> event(ScriptExecutionContext* context)
    {
        OwnPtr<Vector<RefPtr<MessagePort>>> messagePorts = MessagePort::entanglePorts(*context, m_channels.release());
        return MessageEvent::create(messagePorts.release(), m_message, m_origin, String(), m_source);
    }
    SecurityOrigin* targetOrigin() const { return m_targetOrigin.get(); }
    ScriptCallStack* stackTrace() const { return m_stackTrace.get(); }

private:
    virtual void fired()
    {
        m_window->postMessageTimerFired(adoptPtr(this));
        // This object is deleted now.
    }

    RefPtr<DOMWindow> m_window;
    RefPtr<SerializedScriptValue> m_message;
    String m_origin;
    RefPtr<DOMWindow> m_source;
    OwnPtr<MessagePortChannelArray> m_channels;
    RefPtr<SecurityOrigin> m_targetOrigin;
    RefPtr<ScriptCallStack> m_stackTrace;
};

typedef HashCountedSet<DOMWindow*> DOMWindowSet;

static DOMWindowSet& windowsWithUnloadEventListeners()
{
	static NeverDestroyed<DOMWindowSet> windowsWithUnloadEventListeners;
    return windowsWithUnloadEventListeners;
}

static DOMWindowSet& windowsWithBeforeUnloadEventListeners()
{
	static NeverDestroyed<DOMWindowSet> windowsWithBeforeUnloadEventListeners;
    return windowsWithBeforeUnloadEventListeners;
}

static void addUnloadEventListener(DOMWindow* domWindow)
{
    if (windowsWithUnloadEventListeners().add(domWindow).isNewEntry)
        domWindow->disableSuddenTermination();
}

static void removeUnloadEventListener(DOMWindow* domWindow)
{
    if (windowsWithUnloadEventListeners().remove(domWindow))
        domWindow->enableSuddenTermination();
}

static void removeAllUnloadEventListeners(DOMWindow* domWindow)
{
    if (windowsWithUnloadEventListeners().removeAll(domWindow))
        domWindow->enableSuddenTermination();
}

static void addBeforeUnloadEventListener(DOMWindow* domWindow)
{
    if (windowsWithBeforeUnloadEventListeners().add(domWindow).isNewEntry)
        domWindow->disableSuddenTermination();
}

static void removeBeforeUnloadEventListener(DOMWindow* domWindow)
{
    if (windowsWithBeforeUnloadEventListeners().remove(domWindow))
        domWindow->enableSuddenTermination();
}

static void removeAllBeforeUnloadEventListeners(DOMWindow* domWindow)
{
    if (windowsWithBeforeUnloadEventListeners().removeAll(domWindow))
        domWindow->enableSuddenTermination();
}

static bool allowsBeforeUnloadListeners(DOMWindow* window)
{
    ASSERT_ARG(window, window);
    Frame* frame = window->frame();
    if (!frame)
        return false;
    Page* page = frame->page();
    if (!page)
        return false;
    return frame == page->mainFrame();
}

bool DOMWindow::dispatchAllPendingBeforeUnloadEvents()
{
    DOMWindowSet& set = windowsWithBeforeUnloadEventListeners();
    if (set.isEmpty())
        return true;

    static bool alreadyDispatched = false;
    ASSERT(!alreadyDispatched);
    if (alreadyDispatched)
        return true;

    Vector<RefPtr<DOMWindow> > windows;
    for (auto it = set.begin(); it != set.end(); ++it)
        windows.append(it->first);

    size_t size = windows.size();
    for (size_t i = 0; i < size; ++i) {
        DOMWindow* window = windows[i].get();
        if (!set.contains(window))
            continue;

        Frame* frame = window->frame();
        if (!frame)
            continue;

        if (!frame->loader()->shouldClose())
            return false;

        window->enableSuddenTermination();
    }

    alreadyDispatched = true;
    return true;
}

unsigned DOMWindow::pendingUnloadEventListeners() const
{
    return windowsWithUnloadEventListeners().count(const_cast<DOMWindow*>(this));
}

void DOMWindow::dispatchAllPendingUnloadEvents()
{
    DOMWindowSet& set = windowsWithUnloadEventListeners();
    if (set.isEmpty())
        return;

    static bool alreadyDispatched = false;
    ASSERT(!alreadyDispatched);
    if (alreadyDispatched)
        return;

    Vector<RefPtr<DOMWindow> > windows;
    DOMWindowSet::iterator end = set.end();
    for (DOMWindowSet::iterator it = set.begin(); it != end; ++it)
        windows.append(it->first);

    size_t size = windows.size();
    for (size_t i = 0; i < size; ++i) {
        DOMWindow* window = windows[i].get();
        if (!set.contains(window))
            continue;

        window->dispatchEvent(PageTransitionEvent::create(eventNames().pagehideEvent, false), window->document());
        window->dispatchEvent(Event::create(eventNames().unloadEvent, false, false), window->document());

        window->enableSuddenTermination();
    }

    alreadyDispatched = true;
}

// This function:
// 1) Validates the pending changes are not changing to NaN
// 2) Constrains the window rect to the minimum window size and no bigger than
//    the the float rect's dimensions.
// 3) Constrain window rect to within the top and left boundaries of the screen rect
// 4) Constraint the window rect to within the bottom and right boundaries of the
//    screen rect.
// 5) Translate the window rect coordinates to be within the coordinate space of
//    the screen rect.
void DOMWindow::adjustWindowRect(const FloatRect& screen, FloatRect& window, const FloatRect& pendingChanges) const
{
    // Make sure we're in a valid state before adjusting dimensions.
    ASSERT(isfinite(screen.x()));
    ASSERT(isfinite(screen.y()));
    ASSERT(isfinite(screen.width()));
    ASSERT(isfinite(screen.height()));
    ASSERT(isfinite(window.x()));
    ASSERT(isfinite(window.y()));
    ASSERT(isfinite(window.width()));
    ASSERT(isfinite(window.height()));
    
    // Update window values if new requested values are not NaN.
    if (!isnan(pendingChanges.x()))
        window.setX(pendingChanges.x());
    if (!isnan(pendingChanges.y()))
        window.setY(pendingChanges.y());
    if (!isnan(pendingChanges.width()))
        window.setWidth(pendingChanges.width());
    if (!isnan(pendingChanges.height()))
        window.setHeight(pendingChanges.height());

    auto* frame = this->frame();
    ASSERT(frame);
    Page* page = frame->page();
    FloatSize minimumSize = page ? frame->page()->chrome()->client()->minimumWindowSize() : FloatSize(100, 100);
    window.setWidth(min(max(minimumSize.width(), window.width()), screen.width()));
    window.setHeight(min(max(minimumSize.height(), window.height()), screen.height()));

    // Constrain the window position to the screen.
    window.setX(max(screen.x(), min(window.x(), screen.maxX() - window.width())));
    window.setY(max(screen.y(), min(window.y(), screen.maxY() - window.height())));
}

bool DOMWindow::allowPopUp(Frame* firstFrame)
{
    ASSERT(firstFrame);

    if (ScriptController::processingUserGesture())
        return true;

    return firstFrame->settings().javaScriptCanOpenWindowsAutomatically();
}

bool DOMWindow::allowPopUp()
{
    auto* frame = this->frame();
    return frame && allowPopUp(frame);
}

bool DOMWindow::canShowModalDialog(const Frame* frame)
{
    if (!frame)
        return false;

    // Override support for layout testing purposes.
    if (auto* document = frame->document()) {
        if (auto* window = document->domWindow()) {
            if (window->m_canShowModalDialogOverride)
                return window->m_canShowModalDialogOverride.value();
        }
    }

    auto* page = frame->page();
    return page ? page->chrome()->canRunModal() : false;
}

void DOMWindow::setCanShowModalDialogOverride(bool allow)
{
    m_canShowModalDialogOverride = allow;
}

DOMWindow::DOMWindow(Document* document)
    : AbstractDOMWindow()
	, ContextDestructionObserver(document)
{
    ASSERT(frame());
    ASSERT(DOMWindow::document());
}

void DOMWindow::didSecureTransitionTo(Document* document)
{
    observeContext(document);
}

DLL_PUBLIC DOMWindow::~DOMWindow()
{
    if (m_suspendedForPageCache)
        willDestroyCachedFrame();
    else
        willDestroyDocumentInFrame();

    removeAllUnloadEventListeners(this);
    removeAllBeforeUnloadEventListeners(this);
}

ScriptExecutionContext* DOMWindow::scriptExecutionContext() const
{
    return ContextDestructionObserver::scriptExecutionContext();
}

DOMWindow* DOMWindow::toDOMWindow()
{
    return this;
}

PassRefPtr<MediaQueryList> DOMWindow::matchMedia(const String& media)
{
    return document() ? document()->mediaQueryMatcher()->matchMedia(media) : nullptr;
}

Page* DOMWindow::page() const
{
    auto* frame = this->frame();
    return frame ? frame->page() : nullptr;
}

void DOMWindow::frameDestroyed()
{
    RefPtr<DOMWindow> protectedThis(this);

    willDestroyDocumentInFrame();
    JSDOMWindowBase::fireFrameClearedWatchpointsForWindow(this);
}

void DOMWindow::willDestroyCachedFrame()
{
    // It is necessary to copy m_properties to a separate vector because the DOMWindowProperties may
    // unregister themselves from the DOMWindow as a result of the call to willDestroyGlobalObjectInCachedFrame.
    Vector<DOMWindowProperty*> properties = copyToVector(m_properties);
    for (size_t i = 0; i < properties.size(); ++i) {
    	auto* property = properties[i];
        if (m_properties.contains(property))
        	property->willDestroyGlobalObjectInCachedFrame();
    }
}

void DOMWindow::willDestroyDocumentInFrame()
{
    // It is necessary to copy m_properties to a separate vector because the DOMWindowProperties may
    // unregister themselves from the DOMWindow as a result of the call to willDestroyGlobalObjectInFrame.
    Vector<DOMWindowProperty*> properties = copyToVector(m_properties);
    for (size_t i = 0; i < properties.size(); ++i) {
    	auto* property = properties[i];
        if (m_properties.contains(property))
        	property->willDestroyGlobalObjectInFrame();
    }
}

void DOMWindow::willDetachDocumentFromFrame()
{
    if (!frame())
        return;

    // It is necessary to copy m_properties to a separate vector because the DOMWindowProperties may
    // unregister themselves from the DOMWindow as a result of the call to willDetachGlobalObjectFromFrame.
    Vector<DOMWindowProperty*> properties = copyToVector(m_properties);
    for (size_t i = 0; i < properties.size(); ++i) {
    	auto* property = properties[i];
        if (m_properties.contains(property))
        	property->willDetachGlobalObjectFromFrame();
    }

#if ENABLE(RESOURCE_TIMING)
    if (m_performance)
        m_performance->clearResourceTimings();
#endif

    JSDOMWindowBase::fireFrameClearedWatchpointsForWindow(this);
    InspectorInstrumentation::frameWindowDiscarded(frame(), this);
}

void DOMWindow::registerProperty(DOMWindowProperty* property)
{
    m_properties.add(property);
}

void DOMWindow::unregisterProperty(DOMWindowProperty* property)
{
    m_properties.remove(property);
}

void DOMWindow::clear()
{
    // The main frame will always try to clear its DOMWindow when a new load is committed, even if that
    // DOMWindow is suspended in the page cache.
    // In those cases we need to make sure we don't actually clear it.
    if (m_suspendedForPageCache)
        return;
    
    willDestroyDocumentInFrame();
    resetDOMWindowProperties();
}

void DOMWindow::suspendForPageCache()
{
    disconnectDOMWindowProperties();
    m_suspendedForPageCache = true;
}

void DOMWindow::resumeFromPageCache()
{
    reconnectDOMWindowProperties();
    m_suspendedForPageCache = false;
}

void DOMWindow::disconnectDOMWindowProperties()
{
    // It is necessary to copy m_properties to a separate vector because the DOMWindowProperties may
    // unregister themselves from the DOMWindow as a result of the call to disconnectFrameForPageCache.
    Vector<DOMWindowProperty*> properties;
    copyToVector(m_properties, properties);
    for (size_t i = 0; i < properties.size(); ++i)
        properties[i]->disconnectFrameForPageCache();
}

void DOMWindow::reconnectDOMWindowProperties()
{
    ASSERT(m_suspendedForPageCache);
    // It is necessary to copy m_properties to a separate vector because the DOMWindowProperties may
    // unregister themselves from the DOMWindow as a result of the call to reconnectFromPageCache.
    Vector<DOMWindowProperty*> properties;
    copyToVector(m_properties, properties);
    for (size_t i = 0; i < properties.size(); ++i)
        properties[i]->reconnectFrameFromPageCache(frame());
}

void DOMWindow::resetDOMWindowProperties()
{
    m_properties.clear();

    m_applicationCache = nullptr;
    m_crypto = nullptr;
    m_history = nullptr;
    m_localStorage = nullptr;
    m_location = nullptr;
    m_locationbar = nullptr;
    m_media = nullptr;
    m_menubar = nullptr;
    m_navigator = nullptr;
    m_offlineCache = nullptr;
    m_personalbar = nullptr;
    m_screen = nullptr;
    m_scrollbars = nullptr;
    m_selection = nullptr;
    m_sessionStorage = nullptr;
    m_statusbar = nullptr;
    m_toolbar = nullptr;

#if ENABLE(LEGACY_NOTIFICATIONS)
	m_notifications = nullptr;
#endif

#if ENABLE(WEB_TIMING)
    m_performance = nullptr;
#endif
}

bool DOMWindow::isCurrentlyDisplayedInFrame() const
{
    auto* frame = this->frame();
    return frame && frame->document()->domWindow() == this;
}

#if ENABLE(ORIENTATION_EVENTS)
int DOMWindow::orientation() const
{
    auto* frame = this->frame();
    if (!frame)
        return 0;

    return frame->orientation();
}
#endif

Screen* DOMWindow::screen()
{
    if (!m_screen)
        m_screen = Screen::create(*this);
    return m_screen.get();
}

History* DOMWindow::history()
{
    if (!m_history)
        m_history = History::create(*this);
    return m_history.get();
}

Crypto* DOMWindow::crypto()
{
    if (!m_crypto)
        m_crypto = Crypto::create(document());
    return m_crypto.get();
}

BarProp* DOMWindow::locationbar()
{
    if (!m_locationbar)
        m_locationbar = BarProp::create(*this, BarProp::Locationbar);
    return m_locationbar.get();
}

BarProp* DOMWindow::menubar()
{
    if (!m_menubar)
        m_menubar = BarProp::create(*this, BarProp::Menubar);
    return m_menubar.get();
}

BarProp* DOMWindow::personalbar()
{
    if (!m_personalbar)
        m_personalbar = BarProp::create(*this, BarProp::Personalbar);
    return m_personalbar.get();
}

BarProp* DOMWindow::scrollbars()
{
    if (!m_scrollbars)
        m_scrollbars = BarProp::create(*this, BarProp::Scrollbars);
    return m_scrollbars.get();
}

BarProp* DOMWindow::statusbar()
{
    if (!m_statusbar)
        m_statusbar = BarProp::create(*this, BarProp::Statusbar);
    return m_statusbar.get();
}

BarProp* DOMWindow::toolbar()
{
    if (!m_toolbar)
        m_toolbar = BarProp::create(*this, BarProp::Toolbar);
    return m_toolbar.get();
}

PageConsoleClient* DOMWindow::pageConsole() const
{
    if (!isCurrentlyDisplayedInFrame())
        return nullptr;
    Frame* frame = this->frame();
    return frame->page() ? frame->page()->console() : nullptr;
}

DOMApplicationCache* DOMWindow::applicationCache()
{
    if (!m_applicationCache)
        m_applicationCache = DOMApplicationCache::create(*this);
    return m_applicationCache.get();
}

Navigator* DOMWindow::navigator()
{
    if (!m_navigator)
        m_navigator = Navigator::create(this);
    return m_navigator.get();
}

#if ENABLE(WEB_TIMING)
Performance* DOMWindow::performance() const
{
    if (!m_performance) {
    	MonotonicTime timeOrigin = document()->loader() ? document()->loader()->timing()->referenceMonotonicTime() : MonotonicTime::now();
        m_performance = Performance::create(document(), timeOrigin);
    }
    ASSERT(m_performance->scriptExecutionContext() == document());
    return m_performance.get();
}
#endif

ReducedResolutionSeconds DOMWindow::nowTimestamp() const
{
#if ENABLE(WEB_TIMING)
    if (auto* performance = this->performance())
    	return performance->nowInReducedResolutionSeconds();
#endif
    if (document())
    	return Seconds(document()->monotonicTimestamp());
    return Seconds();
}

void DOMWindow::freezeNowTimestamp()
{
    m_frozenNowTimestamp = nowTimestamp();
}

void DOMWindow::unfreezeNowTimestamp()
{
    m_frozenNowTimestamp = WTF::Nullopt;
}

ReducedResolutionSeconds DOMWindow::frozenNowTimestamp() const
{
    return m_frozenNowTimestamp.valueOr(nowTimestamp());
}

Location* DOMWindow::location()
{
    if (!m_location)
        m_location = Location::create(*this);
    return m_location.get();
}

#if 0
VisualViewport* DOMWindow::visualViewport()
{
    if (!m_visualViewport)
        m_visualViewport = VisualViewport::create(*this);
    return m_visualViewport.get();
}
#endif

Storage* DOMWindow::sessionStorage(ExceptionCode& ec)
{
    if (!isCurrentlyDisplayedInFrame())
        return 0;

    Document* document = this->document();
    if (!document)
        return 0;

    if (!document->securityOrigin()->canAccessSessionStorage(document->topDocument()->securityOrigin())) {
        ec = SECURITY_ERR;
        return 0;
    }

    Frame* frame = this->frame();
    if (m_sessionStorage) {
        if (!m_sessionStorage->area()->canAccessStorage(frame)) {
            ec = SECURITY_ERR;
            return 0;
        }
        return m_sessionStorage.get();
    }

    Page* page = document->page();
    if (!page)
        return 0;

    RefPtr<StorageArea> storageArea = page->sessionStorage()->storageArea(document->securityOrigin());
    if (!storageArea->canAccessStorage(frame)) {
		ec = SECURITY_ERR;
		return 0;
	}
    InspectorInstrumentation::didUseDOMStorage(page, storageArea.get(), false, frame);

    m_sessionStorage = Storage::create(*this, storageArea.release());
    return m_sessionStorage.get();
}

Storage* DOMWindow::localStorage(ExceptionCode& ec)
{
    if (!isCurrentlyDisplayedInFrame())
        return 0;

    Document* document = this->document();
    if (!document)
        return 0;

    if (!document->securityOrigin()->canAccessLocalStorage(0)) {
        ec = SECURITY_ERR;
        return 0;
    }

    Frame* frame = this->frame();
    if (m_localStorage) {
        if (!m_localStorage->area()->canAccessStorage(frame)) {
            ec = SECURITY_ERR;
            return 0;
        }
        return m_localStorage.get();
    }

    Page* page = document->page();
    if (!page)
        return 0;

    if (!page->settings().localStorageEnabled())
        return 0;

    PageGroup& pageGroup = page->group();
    RefPtr<StorageArea> storageArea;
    if (!document->securityOrigin()->canAccessLocalStorage(document->topOrigin()))
        storageArea = pageGroup.transientLocalStorage(document->topOrigin())->storageArea(document->securityOrigin());
    else {
    	KURL localStorageURL = page->chrome()->client()->localStorageDomain(document->url());

    	if(localStorageURL.host() == document->url().host())
    		storageArea = pageGroup.localStorage(page->settings(), document->url().host().toString())->storageArea(document->securityOrigin());
    	else
    		storageArea = pageGroup.localStorage(page->settings(), localStorageURL.host().toString())->storageArea(SecurityOrigin::create(localStorageURL));
    }

    if (!storageArea->canAccessStorage(frame)) {
		ec = SECURITY_ERR;
		return 0;
	}
    InspectorInstrumentation::didUseDOMStorage(page, storageArea.get(), true, frame);

    m_localStorage = Storage::create(*this, storageArea.release());
    return m_localStorage.get();
}

OfflineCache* DOMWindow::offlineCache(ExceptionCode& ec)
{
    if (!isCurrentlyDisplayedInFrame())
        return 0;

    m_offlineCache = OfflineCache::create(*this);
    return m_offlineCache.get();
}

#if ENABLE(LEGACY_NOTIFICATIONS)
NotificationCenter* DOMWindow::webkitNotifications() const
{
    if (!isCurrentlyDisplayedInFrame())
        return 0;
    if (m_notifications)
        return m_notifications.get();

    Document* document = this->document();
    if (!document)
        return 0;

    Page* page = document->page();
    if (!page)
        return 0;

    NotificationClient* provider = NotificationController::clientFrom(page);
    if (!provider)
    	provider = page->chrome()->client()->notificationClient();

    if (provider)
        m_notifications = NotificationCenter::create(document, provider);

    return m_notifications.get();
}

void DOMWindow::resetNotifications()
{
    if (!m_notifications)
        return;
    //m_notifications->disconnectFrame();
    m_notifications = 0;
}
#endif

#if ENABLE(RMAI)
GalioDebug* DOMWindow::debug()
{
	if (!m_galioDebug) {
		RefPtr<Console> c = console();
		m_galioDebug = GalioDebug::create(c);
	}
	return m_galioDebug.get();
}

MOM* DOMWindow::mom()
{
    if (!MOM::initialised())
        return 0;
	if (!m_mom)
		m_mom = MOM::create(scriptExecutionContext());
	return m_mom.get();
}
#endif

void DOMWindow::postMessage(JSC::ExecState*, DOMWindow* incumbentWindow, PassRefPtr<SerializedScriptValue> message, const Vector<RefPtr<MessagePort>>* ports, const String& targetOrigin, ExceptionCode& ec)
{
    if (!isCurrentlyDisplayedInFrame())
        return;

    Document* sourceDocument = incumbentWindow->document();

    // Compute the target origin.  We need to do this synchronously in order
    // to generate the SYNTAX_ERR exception correctly.
    RefPtr<SecurityOrigin> target;
    if (targetOrigin == "/") {
        if (!sourceDocument)
            return;
        target = sourceDocument->securityOrigin();
    } else if (targetOrigin != "*") {
        target = SecurityOrigin::createFromString(targetOrigin);
        // It doesn't make sense target a postMessage at a unique origin
        // because there's no way to represent a unique origin in a string.
        if (target->isUnique()) {
            ec = SYNTAX_ERR;
            return;
        }
    }

    OwnPtr<MessagePortChannelArray> channels = MessagePort::disentanglePorts(ports, ec);
    if (ec)
        return;

    // Capture the source of the message.  We need to do this synchronously
    // in order to capture the source of the message correctly.
    if (!sourceDocument)
        return;
    String sourceOrigin = sourceDocument->securityOrigin()->toString();

    // Capture stack trace only when inspector front-end is loaded as it may be time consuming.
    RefPtr<ScriptCallStack> stackTrace;
    if (InspectorInstrumentation::consoleAgentEnabled(sourceDocument))
        stackTrace = createScriptCallStack(JSExecState::currentState(), ScriptCallStack::maxCallStackSizeToCapture);

    // Schedule the message.
    PostMessageTimer* timer = new PostMessageTimer(this, message, sourceOrigin, incumbentWindow, channels.release(), target.get(), stackTrace.release());
    timer->startOneShot(0);
}

void DOMWindow::postMessageTimerFired(PassOwnPtr<PostMessageTimer> t)
{
    OwnPtr<PostMessageTimer> timer(t);

    if (!document() || !isCurrentlyDisplayedInFrame())
        return;

    RefPtr<MessageEvent> event = timer->event(document());

    // Give the embedder a chance to intercept this postMessage because this
    // DOMWindow might be a proxy for another in browsers that support
    // postMessage calls across WebKit instances.
    if (frame()->loader()->client()->willCheckAndDispatchMessageEvent(timer->targetOrigin(), event.get()))
        return;

    dispatchMessageEventWithOriginCheck(timer->targetOrigin(), event, timer->stackTrace());
}

void DOMWindow::dispatchMessageEventWithOriginCheck(SecurityOrigin* intendedTargetOrigin, PassRefPtr<Event> event, PassRefPtr<ScriptCallStack> stackTrace)
{
    if (intendedTargetOrigin) {
        // Check target origin now since the target document may have changed since the timer was scheduled.
        if (!intendedTargetOrigin->isSameSchemeHostPort(document()->securityOrigin())) {
        	if (auto* pageConsole = this->pageConsole()) {
        		String message = tryMakeString("Unable to post message to ", intendedTargetOrigin->toString(), ". Recipient has origin ", document()->securityOrigin()->toString(), ".\n");
        		pageConsole->addMessage(MessageSource::JS, MessageLevel::Error, message, stackTrace);
        	}
            return;
        }
    }

    dispatchEvent(event);
}

DOMSelection* DOMWindow::getSelection()
{
    if (!isCurrentlyDisplayedInFrame())
        return nullptr;
    if (!m_selection)
        m_selection = DOMSelection::create(*this);
    return m_selection.get();
}

Element* DOMWindow::frameElement() const
{
    Frame* frame = this->frame();
    if (!frame)
        return nullptr;

    return frame->ownerElement();
}

void DOMWindow::focus(DOMWindow* incumbentWindow)
{
    focus(opener() && opener() != this && incumbentWindow == opener());
}

void DOMWindow::focus(bool allowFocus)
{
    if (!frame())
        return;

    RefPtr<Frame> protectedFrame = frame();

    Page* page = protectedFrame->page();
    if (!page)
        return;

    allowFocus = allowFocus || WindowFocusAllowedIndicator::windowFocusAllowed() || !protectedFrame->settings().windowFocusRestricted();

    // If we're a top level window, bring the window to the front.
    if (protectedFrame->isMainFrame() && allowFocus)
        page->chrome()->focus();

    // Clear the current frame's focused node if a new frame is about to be focused.
    RefPtr<Frame> focusedFrame = page->focusController()->focusedFrame();
    if (focusedFrame && focusedFrame != protectedFrame)
        focusedFrame->document()->setFocusedElement(0);

    protectedFrame->eventHandler()->focusDocumentView();
}

void DOMWindow::blur()
{
    Frame* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    if (frame->settings().windowFocusRestricted())
        return;

    if (frame != page->mainFrame())
        return;

    page->chrome()->unfocus();
}

void DOMWindow::close(Document* document)
{
    if (!document->canNavigate(frame()))
        return;
    close();
}

void DOMWindow::close()
{
    Frame* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    if (frame != page->mainFrame())
        return;

    bool allowScriptsToCloseWindows = frame->settings().allowScriptsToCloseWindows();

    if (!(page->openedByDOM() || page->backForward()->count() <= 1 || allowScriptsToCloseWindows))
        return;

    if (!frame->loader()->shouldClose())
        return;

    page->chrome()->closeWindowSoon();
}

void DOMWindow::print()
{
    Frame* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    if (!page->arePromptsAllowed()) {
        printErrorMessage("Use of window.print is not allowed while unloading a page.");
        return;
    }

    if (frame->loader()->activeDocumentLoader()->isLoading()) {
        m_shouldPrintWhenFinishedLoading = true;
        return;
    }
    m_shouldPrintWhenFinishedLoading = false;
    page->chrome()->print(frame);
}

void DOMWindow::stop()
{
    Frame* frame = this->frame();
    if (!frame)
        return;

    // We must check whether the load is complete asynchronously, because we might still be parsing
    // the document until the callstack unwinds.
    frame->loader()->stopForUserCancel(true);
}

#if ENABLE(HBBTV)
void DOMWindow::debug(const String& arg)
{
    if (arg.isEmpty())
        return;

    RefPtr<ScriptCallStack> stackTrace;
    stackTrace = createScriptCallStack(JSExecState::currentState(), ScriptCallStack::maxCallStackSizeToCapture);

    if (auto* pageConsole = this->pageConsole()) {
        pageConsole->addMessage(MessageSource::JS, MessageLevel::Debug, arg, stackTrace.release());
    }
}
#endif

void DOMWindow::alert(const String& message)
{
    Frame* frame = this->frame();
    if (!frame)
        return;

    auto* page = frame->page();
    if (!page)
        return;

    if (!page->arePromptsAllowed()) {
        printErrorMessage("Use of window.alert is not allowed while unloading a page.");
        return;
    }

    frame->document()->updateStyleIfNeeded();

    page->chrome()->runJavaScriptAlert(frame, message);
}

bool DOMWindow::confirm(const String& message)
{
    Frame* frame = this->frame();
    if (!frame)
        return false;

    auto* page = frame->page();
    if (!page)
        return false;

    if (!page->arePromptsAllowed()) {
        printErrorMessage("Use of window.confirm is not allowed while unloading a page.");
        return false;
    }

    frame->document()->updateStyleIfNeeded();

    return page->chrome()->runJavaScriptConfirm(frame, message);
}

String DOMWindow::prompt(const String& message, const String& defaultValue)
{
    Frame* frame = this->frame();
    if (!frame)
        return String();

    auto* page = frame->page();
    if (!page)
        return String();

    if (!page->arePromptsAllowed()) {
        printErrorMessage("Use of window.prompt is not allowed while unloading a page.");
        return String();
    }

    frame->document()->updateStyleIfNeeded();

    String returnValue;
    if (page->chrome()->runJavaScriptPrompt(frame, message, defaultValue, returnValue))
        return returnValue;

    return String();
}

bool DOMWindow::find(const String& string, bool caseSensitive, bool backwards, bool wrap, bool /*wholeWord*/, bool /*searchInFrames*/, bool /*showDialog*/) const
{
    if (!isCurrentlyDisplayedInFrame())
        return false;

    // FIXME (13016): Support wholeWord, searchInFrames and showDialog
    FindOptions options = (backwards ? Backwards : 0) | (caseSensitive ? 0 : CaseInsensitive) | (wrap ? WrapAround : 0);
	return frame()->editor().findString(string, options | DoNotTraverseFlatTree);
}

bool DOMWindow::offscreenBuffering() const
{
    return true;
}

int DOMWindow::outerHeight() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    Page* page = frame->page();
    if (!page)
        return 0;

    return static_cast<int>(page->chrome()->windowRect().height());
}

int DOMWindow::outerWidth() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    Page* page = frame->page();
    if (!page)
        return 0;

    return static_cast<int>(page->chrome()->windowRect().width());
}

int DOMWindow::innerHeight() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    // Force enough layout in the parent document to ensure that the FrameView has been resized.
    if (auto* frameElement = this->frameElement())
        frameElement->document().updateLayoutIfDimensionsOutOfDate(*frameElement, HeightDimensionsCheck);

    FrameView* view = frame->view();
    if (!view)
        return 0;

    return view->mapFromLayoutToCSSUnits(static_cast<int>(view->unobscuredContentRectIncludingScrollbars().height()));
}

int DOMWindow::innerWidth() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    // Force enough layout in the parent document to ensure that the FrameView has been resized.
    if (auto* frameElement = this->frameElement())
        frameElement->document().updateLayoutIfDimensionsOutOfDate(*frameElement, WidthDimensionsCheck);

    FrameView* view = frame->view();
    if (!view)
        return 0;

    return view->mapFromLayoutToCSSUnits(static_cast<int>(view->unobscuredContentRectIncludingScrollbars().width()));
}

int DOMWindow::screenX() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    Page* page = frame->page();
    if (!page)
        return 0;

    return static_cast<int>(page->chrome()->windowRect().x());
}

int DOMWindow::screenY() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    Page* page = frame->page();
    if (!page)
        return 0;

    return static_cast<int>(page->chrome()->windowRect().y());
}

int DOMWindow::scrollX() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    FrameView* view = frame->view();
    if (!view)
        return 0;

    int scrollX = view->contentsScrollPosition().x();
    if (!scrollX)
        return 0;

    frame->document()->updateLayoutIgnorePendingStylesheets();

    return view->mapFromLayoutToCSSUnits(view->contentsScrollPosition().x());
}

int DOMWindow::scrollY() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    FrameView* view = frame->view();
    if (!view)
        return 0;

    int scrollY = view->contentsScrollPosition().y();
    if (!scrollY)
        return 0;

    frame->document()->updateLayoutIgnorePendingStylesheets();

    return view->mapFromLayoutToCSSUnits(view->contentsScrollPosition().y());
}

bool DOMWindow::closed() const
{
    return !frame();
}

unsigned DOMWindow::length() const
{
    if (!isCurrentlyDisplayedInFrame())
        return 0;

    return frame()->tree()->scopedChildCount();
}

String DOMWindow::name() const
{
    Frame* frame = this->frame();
    if (!frame)
        return String();

    return frame->tree()->name();
}

void DOMWindow::setName(const String& string)
{
    Frame* frame = this->frame();
    if (!frame)
        return;

    frame->tree()->setName(string);
}

void DOMWindow::setStatus(const String& string) 
{
    m_status = string;

    Frame* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    ASSERT(frame->document()); // Client calls shouldn't be made when the frame is in inconsistent state.
    page->chrome()->setStatusbarText(frame, m_status);
} 
    
void DOMWindow::setDefaultStatus(const String& string) 
{
    m_defaultStatus = string;

    Frame* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    ASSERT(frame->document()); // Client calls shouldn't be made when the frame is in inconsistent state.
    page->chrome()->setStatusbarText(frame, m_defaultStatus);
}

DOMWindow* DOMWindow::self() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    return frame->document()->domWindow();
}

DOMWindow* DOMWindow::opener() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    Frame* opener = frame->loader()->opener();
    if (!opener)
        return 0;

    return opener->document()->domWindow();
}

void DOMWindow::disownOpener()
{
    if(Frame* frame = this->frame())
        frame->loader()->setOpener(0);
}

DOMWindow* DOMWindow::parent() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    Frame* parent = frame->tree()->parent();
    if (parent)
        return parent->document()->domWindow();

    return frame->document()->domWindow();
}

DOMWindow* DOMWindow::top() const
{
    Frame* frame = this->frame();
    if (!frame)
        return 0;

    Page* page = frame->page();
    if (!page)
        return 0;

    return frame->tree()->top()->document()->domWindow();
}

String DOMWindow::origin() const
{
    auto document = this->document();
    return document ? document->securityOrigin()->toString() : emptyString();
}

#if ENABLE(CEHTML)
// SHALL always return false and have the same effect as calling method top.replace(url).
bool DOMWindow::escapeBeyondTopMost(const String& url, DOMWindow* activeWindow, DOMWindow* firstWindow)
{
    if (!url.isNull()) {
        DOMWindow* domWindow = top();
        domWindow->setLocation(activeWindow, firstWindow, url, LockHistoryAndBackForwardList);
    }
    return false;
}

// Reloads the currently loaded URL inside this window/frame as identified by property â€œlocationâ€�.
// This method is equivalent to calling â€œlocation.reload()â€�.
void DOMWindow::reload(DOMWindow* activeWindow)
{
   location()->reload(activeWindow);
}

// Loads the given URL into this â€œwindowâ€� object. This method is equivalent to calling â€œlocation.replace(url).
// Loading a new url into a window/frame SHOULD not automatically trigger a focus change to the â€œwindowâ€� object
// that loads the new document. 
// The method window.focus() SHOULD be used to achieve that.
void DOMWindow::replace(const String& url, DOMWindow* activeWindow, DOMWindow* firstWindow)
{
    location()->replace(activeWindow, firstWindow, url);
}

// Requests the browser to set the focus to the window/frame that corresponds to this "windowâ€� object.
// Returns a Boolean that indicates whether the browser has granted such a request or not.
bool DOMWindow::requestFocus(ScriptExecutionContext* context)
{
    Frame* frame = this->frame();
    if (!frame)
        return false;

    Page* page = frame->page();
    if (!page)
        return false;

    bool allowFocus = WindowFocusAllowedIndicator::windowFocusAllowed() || !frame->settings().windowFocusRestricted();
    if (context) {
        ASSERT(isMainThread());
        Document* activeDocument = static_cast<Document*>(context);
        if (opener() && activeDocument->domWindow() == opener())
            allowFocus = true;
    }

    // Bring the window to the front.
    if (allowFocus)
        page->chrome()->focus();

    if (frame)
        frame->eventHandler()->focusDocumentView();

    return allowFocus;
}

// Equivalent to calling â€œsetTimeout(function, delayInSeconds*1000)â€�.
int DOMWindow::setTimer(PassRefPtr<ScheduledAction> action, int timeout, ExceptionCode& ec)
{
    return setTimeout(action, timeout*1000, ec);
}

// Convenience method that returns a sub-frame/window with the given ID, 
// that is a direct descendant of the current frame/window. The type of the 
// object that is returned is of type â€œwindowâ€�. This is the same as searching
// in the frames array for â€œwindowâ€� objects with the given ID.
PassRefPtr<DOMWindow> DOMWindow::getFrame(const AtomicString& id)
{
    if (!isCurrentlyDisplayedInFrame())
        return nullptr;

    if (id.isEmpty())
        return nullptr;

    Frame* thisFrame = this->frame();
    if(Frame* targetFrame = thisFrame->tree()->find(id, *thisFrame))
    	return targetFrame->document()->domWindow();

    return nullptr;
}

// Read-write property that provides access to the HTTP-connection timeout for all connections 
// resulting from interaction with the current frame, or from a script inside the current 
// window/frame, as an integer value representing a number of seconds. Changing this value 
// is equivalent to calling method setHttpTimeout() for all possible URLs that can result from 
// interacting with the current frame or that can be called inside a script running in the current frame.
unsigned DOMWindow::httptimeout() const
{
    if (!isCurrentlyDisplayedInFrame())
        return 0;

    return frame()->httptimeout();
}

// This is the setter of attribute httptimeout.
// Not to be confused with the CE-HTML setHttpTimeout(url, seconds) method
void DOMWindow::setHttptimeout(const unsigned timeout)
{
    if (!isCurrentlyDisplayedInFrame())
        return;

    frame()->setHttptimeout(timeout);
}

// Notifies the browser that the user has stopped using the user interface of the Remote UI Server 
// that provided the XHTML contents of the window/frame indicated by property â€œtopâ€�. As a result of
// this, the Remote UI Client MAY for example decide to go back to its UI shell.
void DOMWindow::exitUnit()
{
    // we do nothing
	WTF::logMessage(ASCIILiteral("DOMWindow::exitUnit()"));
    return;
}

// If mode equals value â€œdelayedâ€�, delays the actual screen rendering of (changes to) the XHTML page,
// until the rendering mode has been set back to â€œdefaultâ€�. However, scripts continue executing, e.g.
// to update the DOM tree. If mode equals value â€�defaultâ€�, resumes the default screen rendering behavior
// as defined by the Remote UI Client (which would occur if this method is not called). If the rendering 
// mode is set to "default" after being set to â€œdelayedâ€�, it shows the result of all delayed screen updates
// that have occurred while the â€œdelayedâ€� rendering mode was in effect.
// Unknown values given to mode should be ignored.
void DOMWindow::setRenderMode(const String& mode)
{
    if(mode == "delayed" && m_renderMode == RenderModeDefault) {
        m_renderMode = RenderModeDelayed;
    } else if (mode =="default" && m_renderMode == RenderModeDelayed) {
         m_renderMode = RenderModeDefault;
    }
}

// Sets the HTTP-connection timeout for connections to the given URL, resulting from interaction inside 
// the current window/frame, or called within a script running in the current window/frame, to the given 
// number of seconds. A value of zero indicates that the clientâ€™s default timeout SHALL be restored.
void DOMWindow::setHttpTimeout(const String& url, unsigned seconds)
{
    if (!isCurrentlyDisplayedInFrame())
        return;

    if(url.isEmpty())
        return;

    auto* frame = this->frame();
    frame->setTimeoutForURL(url, seconds);
}

// Synchrously loads the document indicated by the given URL. The method returns an object of type 
// "Document" as defined in the ECMAScript binding Section of [11]. It returns null if it fails.
PassRefPtr<Document> DOMWindow::openURL(const String& url) const
{
    if (!isCurrentlyDisplayedInFrame())
        return nullptr;

    if(url.isEmpty())
        return nullptr;

    return Document::create(frame(), KURL(ParsedURLString, url));
}
#endif // ENABLE(CEHTML)

Document* DOMWindow::document() const
{
    ScriptExecutionContext* context = ContextDestructionObserver::scriptExecutionContext();
    ASSERT(!context || context->isDocument());
    return static_cast<Document*>(context);
}

SecurityOrigin* DOMWindow::securityOrigin() const
{
    return document() ? document()->securityOrigin() : 0;
}

PassRefPtr<StyleMedia> DOMWindow::styleMedia()
{
    if (!m_media)
        m_media = StyleMedia::create(*this);
    return m_media.get();
}

PassRefPtr<CSSStyleDeclaration> DOMWindow::getComputedStyle(Element* elt, const String& pseudoElt) const
{
    if (!elt)
        return nullptr;

    return CSSComputedStyleDeclaration::create(*elt, false, pseudoElt);
}

PassRefPtr<CSSRuleList> DOMWindow::getMatchedCSSRules(Element* element, const String& pseudoElement, bool authorOnly) const
{
    if (!isCurrentlyDisplayedInFrame())
        return nullptr;

    unsigned colonStart = pseudoElement[0] == ':' ? (pseudoElement[1] == ':' ? 2 : 1) : 0;
    CSSSelector::PseudoElementType pseudoType = CSSSelector::parsePseudoElementType(pseudoElement.substringSharingImpl(colonStart));
    if (pseudoType == CSSSelector::PseudoElementUnknown && !pseudoElement.isEmpty())
        return nullptr;

    auto* frame = this->frame();
    frame->document()->styleScope().flushPendingUpdate();

    unsigned rulesToInclude = StyleResolver::AuthorCSSRules;
    if (!authorOnly)
        rulesToInclude |= StyleResolver::UAAndUserCSSRules;

    PseudoId pseudoId = CSSSelector::pseudoId(pseudoType);

    Vector<RefPtr<const StyleRule>> matchedRules = frame->document()->styleScope().resolver().pseudoStyleRulesForElement(element, pseudoId, rulesToInclude);
    if (matchedRules.isEmpty())
        return nullptr;

    bool allowCrossOrigin = frame->settings().crossOriginCheckInGetMatchedCSSRulesDisabled();

    RefPtr<StaticCSSRuleList> ruleList = StaticCSSRuleList::create();
    for (auto& rule : matchedRules) {
        if (!allowCrossOrigin && !rule->hasDocumentSecurityOrigin())
            continue;
        ruleList->rules().append(rule->createCSSOMWrapper());
    }

    if (ruleList->rules().isEmpty())
        return nullptr;

    return ruleList.release();
}

PassRefPtr<WebKitPoint> DOMWindow::webkitConvertPointFromNodeToPage(Node* node, const WebKitPoint* p) const
{
    if (!node || !p)
        return nullptr;

    if (!document())
        return nullptr;

    document()->updateLayoutIgnorePendingStylesheets();

    FloatPoint pagePoint(p->x(), p->y());
    pagePoint = node->convertToPage(pagePoint);
    return WebKitPoint::create(pagePoint.x(), pagePoint.y());
}

PassRefPtr<WebKitPoint> DOMWindow::webkitConvertPointFromPageToNode(Node* node, const WebKitPoint* p) const
{
    if (!node || !p)
        return nullptr;

    if (!document())
        return nullptr;

    document()->updateLayoutIgnorePendingStylesheets();

    FloatPoint nodePoint(p->x(), p->y());
    nodePoint = node->convertFromPage(nodePoint);
    return WebKitPoint::create(nodePoint.x(), nodePoint.y());
}

double DOMWindow::devicePixelRatio() const
{
    auto* frame = this->frame();
    if (!frame)
        return 0.0;

    Page* page = frame->page();
    if (!page)
        return 0.0;

    if(auto ratio = page->chrome()->client()->devicePixelRatio())
    	return ratio.value();

    return page->deviceScaleFactor();
}

void DOMWindow::scrollBy(const ScrollToOptions& options) const
{
    return scrollBy(options.left.valueOr(0), options.top.valueOr(0));
}

void DOMWindow::scrollBy(double x, double y) const
{
    if (!isCurrentlyDisplayedInFrame())
        return;

    document()->updateLayoutIgnorePendingStylesheets();

    auto* frame = this->frame();
    FrameView* view = frame->view();
    if (!view)
        return;

    // Normalize non-finite values (https://drafts.csswg.org/cssom-view/#normalize-non-finite-values).
    x = std::isfinite(x) ? x : 0;
    y = std::isfinite(y) ? y : 0;

    IntSize scaledOffset(view->mapFromCSSToLayoutUnits(x), view->mapFromCSSToLayoutUnits(y));
    view->setContentsScrollPosition(view->contentsScrollPosition() + scaledOffset);
}

void DOMWindow::scrollTo(const ScrollToOptions& options) const
{
    auto* frame = this->frame();
    RefPtr<FrameView> view = frame->view();
    if (!view)
        return;

    double x = options.left ? options.left.value() : view->contentsScrollPosition().x();
    double y = options.top ? options.top.value() : view->contentsScrollPosition().y();
    return scrollTo(x, y);
}

void DOMWindow::scrollTo(double x, double y) const
{
    if (!isCurrentlyDisplayedInFrame())
        return;

    auto* frame = this->frame();
    RefPtr<FrameView> view = frame->view();
    if (!view)
        return;

    // Normalize non-finite values (https://drafts.csswg.org/cssom-view/#normalize-non-finite-values).
    x = std::isfinite(x) ? x : 0;
    y = std::isfinite(y) ? y : 0;

    if (!x && !y && view->contentsScrollPosition() == IntPoint(0, 0))
        return;

    document()->updateLayoutIgnorePendingStylesheets();

    IntPoint layoutPos(view->mapFromCSSToLayoutUnits(x), view->mapFromCSSToLayoutUnits(y));
    view->setContentsScrollPosition(layoutPos);
}

void DOMWindow::moveBy(float x, float y) const
{
    auto* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    if (frame != page->mainFrame()) {
#if ENABLE(RMAI)
        if(Document* doc = document()){
            if(GalioWafer* wafer = doc->ownerWafer())
                wafer->moveBy(x, y);
        }
#endif
        return;
    }

    FloatRect fr = page->chrome()->windowRect();
    FloatRect update = fr;
    update.move(x, y);
    // Security check (the spec talks about UniversalBrowserWrite to disable this check...)
    adjustWindowRect(screenAvailableRect(page->mainFrame()->view()), fr, update);
    page->chrome()->setWindowRect(fr);
}

void DOMWindow::moveTo(float x, float y) const
{
    auto* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    if (frame != page->mainFrame()) {
#if ENABLE(RMAI)
        if(Document* doc = document()){
            if(GalioWafer* wafer = doc->ownerWafer())
                wafer->moveTo(x, y);
        }
#endif
        return;
    }

    FloatRect fr = page->chrome()->windowRect();
    FloatRect sr = screenAvailableRect(page->mainFrame()->view());
    fr.setLocation(sr.location());
    FloatRect update = fr;
    update.move(x, y);     
    // Security check (the spec talks about UniversalBrowserWrite to disable this check...)
    adjustWindowRect(sr, fr, update);
    page->chrome()->setWindowRect(fr);
}

void DOMWindow::resizeBy(float x, float y) const
{
    auto* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    if (frame != page->mainFrame()) {
#if ENABLE(RMAI)
        if(Document* doc = document()){
            if(GalioWafer* wafer = doc->ownerWafer())
                wafer->resizeBy(x, y);
        }
#endif
        return;
    }

    FloatRect fr = page->chrome()->windowRect();
    FloatSize dest = fr.size() + FloatSize(x, y);
    FloatRect update(fr.location(), dest);
    adjustWindowRect(screenAvailableRect(page->mainFrame()->view()), fr, update);
    page->chrome()->setWindowRect(fr);
}

void DOMWindow::resizeTo(float width, float height) const
{
    auto* frame = this->frame();
    if (!frame)
        return;

    Page* page = frame->page();
    if (!page)
        return;

    if (frame != page->mainFrame()) {
#if ENABLE(RMAI)
        if(Document* doc = document()){
            if(GalioWafer* wafer = doc->ownerWafer())
                wafer->resizeTo(width, height);
        }
#endif
        return;
    }

    FloatRect fr = page->chrome()->windowRect();
    FloatSize dest = FloatSize(width, height);
    FloatRect update(fr.location(), dest);
    adjustWindowRect(screenAvailableRect(page->mainFrame()->view()), fr, update);
    page->chrome()->setWindowRect(fr);
}

int DOMWindow::setTimeout(PassRefPtr<ScheduledAction> action, int timeout, ExceptionCode& ec)
{
    ScriptExecutionContext* context = scriptExecutionContext();
    if (!context) {
        ec = INVALID_ACCESS_ERR;
        return -1;
    }
    return DOMTimer::install(context, action, timeout, true);
}

void DOMWindow::clearTimeout(int timeoutId)
{
    ScriptExecutionContext* context = scriptExecutionContext();
    if (!context)
        return;
    DOMTimer::removeById(context, timeoutId);
}

int DOMWindow::setInterval(PassRefPtr<ScheduledAction> action, int timeout, ExceptionCode& ec)
{
    ScriptExecutionContext* context = scriptExecutionContext();
    if (!context) {
        ec = INVALID_ACCESS_ERR;
        return -1;
    }
    return DOMTimer::install(context, action, timeout, false);
}

void DOMWindow::clearInterval(int timeoutId)
{
    ScriptExecutionContext* context = scriptExecutionContext();
    if (!context)
        return;
    DOMTimer::removeById(context, timeoutId);
}

#if ENABLE(REQUEST_ANIMATION_FRAME)
int DOMWindow::requestAnimationFrame(PassRefPtr<RequestAnimationFrameCallback> callback)
{
    if (Document* d = document())
        return d->requestAnimationFrame(callback);
    return 0;
}

int DOMWindow::webkitRequestAnimationFrame(PassRefPtr<RequestAnimationFrameCallback> callback)
{
    static bool firstTime = true;
    if (firstTime && document()) {
        document()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Warning, ASCIILiteral("webkitRequestAnimationFrame() is deprecated and will be removed. Please use requestAnimationFrame() instead."));
        firstTime = false;
    }
    return requestAnimationFrame(callback);
}

void DOMWindow::cancelAnimationFrame(int id)
{
    if (Document* d = document())
        d->cancelAnimationFrame(id);
}
#endif

template<typename T>
inline void DOMWindow::createImageBitmapInternal(PassRefPtr<T> source, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
    auto* document = this->document();
    if (!document) {
        promise->reject<ExceptionCode>(INVALID_STATE_ERR);
        return;
    }
    ImageBitmap::createPromise(*document, source, options, promise);
}

template<typename T>
inline void DOMWindow::createImageBitmapInternal(PassRefPtr<T> source, int sx, int sy, int sw, int sh, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
    auto* document = this->document();
    if (!document) {
        promise->reject<ExceptionCode>(INVALID_STATE_ERR);
        return;
    }
    ImageBitmap::createPromise(*document, source, options, sx, sy, sw, sh, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<HTMLImageElement> source, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<HTMLCanvasElement> source, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<HTMLVideoElement> source, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<ImageBitmap> source, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<Blob> source, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<ImageData> source, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<HTMLImageElement> source, int sx, int sy, int sw, int sh, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, sx, sy, sw, sh, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<HTMLCanvasElement> source, int sx, int sy, int sw, int sh, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, sx, sy, sw, sh, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<HTMLVideoElement> source, int sx, int sy, int sw, int sh, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, sx, sy, sw, sh, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<ImageBitmap> source, int sx, int sy, int sw, int sh, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, sx, sy, sw, sh, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<Blob> source, int sx, int sy, int sw, int sh, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, sx, sy, sw, sh, options, promise);
}

void DOMWindow::createImageBitmap(PassRefPtr<ImageData> source, int sx, int sy, int sw, int sh, const ImageBitmapOptions& options, PassRefPtr<DeferredPromise> promise)
{
	createImageBitmapInternal(source, sx, sy, sw, sh, options, promise);
}

int DOMWindow::requestIdleCallback(PassRefPtr<IdleRequestCallback> callback, const IdleRequestOptions& options)
{
    RefPtr<Document> document(this->document());
    if (!document)
        return 0;
    ASSERT(callback);
    return document->requestIdleCallback(Ref<IdleRequestCallback>(*callback), Seconds::fromMilliseconds(options.timeout));
}

void DOMWindow::cancelIdleCallback(int id)
{
    RefPtr<Document> document(this->document());
    if (!document)
        return;
    return document->cancelIdleCallback(id);
}

bool DOMWindow::isSecureContext() const
{
    auto* document = this->document();
    if (!document)
        return false;
    return document->isSecureContext();
}

static void didAddStorageEventListener(DOMWindow* window)
{
    // Creating these WebCore::Storage objects informs the system that we'd like to receive
    // notifications about storage events that might be triggered in other processes. Rather
    // than subscribe to these notifications explicitly, we subscribe to them implicitly to
    // simplify the work done by the system. 
    ExceptionCode unused;
    window->localStorage(unused);
    window->sessionStorage(unused);
}

bool DOMWindow::addEventListener(const AtomicString& eventType, PassRefPtr<EventListener> listener, const AddEventListenerOptions& options)
{
    if (!EventTarget::addEventListener(eventType, listener, options))
        return false;

    auto& eventNames = WebCore::eventNames();
    if (Document* document = this->document()) {
        document->addListenerTypeIfNeeded(eventType);
        if (eventType == eventNames.wheelEvent || eventType == eventNames.mousewheelEvent)
			document->didAddWheelEventHandler();
        else if (eventNames.isTouchEventType(eventType))
            document->didAddTouchEventHandler();
        else if (eventType == eventNames.storageEvent)
            didAddStorageEventListener(this);
    }

    if (eventType == eventNames.unloadEvent)
        addUnloadEventListener(this);
    else if (eventType == eventNames.beforeunloadEvent && allowsBeforeUnloadListeners(this))
        addBeforeUnloadEventListener(this);
#if ENABLE(DEVICE_ORIENTATION)
    else if (eventType == eventNames.devicemotionEvent) {
        if (DeviceMotionController* controller = DeviceMotionController::from(page()))
            controller->addListener(this);
    } else if (eventType == eventNames.deviceorientationEvent) {
        if (DeviceOrientationController* controller = DeviceOrientationController::from(page()))
            controller->addListener(this);
    }
#endif

    return true;
}

bool DOMWindow::removeEventListener(const AtomicString& eventType, EventListener* listener, const EventListenerOptions& options)
{
    if (!EventTarget::removeEventListener(eventType, listener, options))
        return false;

    auto& eventNames = WebCore::eventNames();
    if (Document* document = this->document()) {
    	if (eventType == eventNames.wheelEvent || eventType == eventNames.mousewheelEvent)
			document->didRemoveWheelEventHandler();
        else if (eventNames.isTouchEventType(eventType))
            document->didRemoveTouchEventHandler();
    }

    if (eventType == eventNames.unloadEvent)
        removeUnloadEventListener(this);
    else if (eventType == eventNames.beforeunloadEvent && allowsBeforeUnloadListeners(this))
        removeBeforeUnloadEventListener(this);
#if ENABLE(DEVICE_ORIENTATION)
    else if (eventType == eventNames.devicemotionEvent) {
        if (DeviceMotionController* controller = DeviceMotionController::from(page()))
            controller->removeListener(this);
    } else if (eventType == eventNames.deviceorientationEvent) {
        if (DeviceOrientationController* controller = DeviceOrientationController::from(page()))
            controller->removeListener(this);
    }
#endif

    return true;
}

#if 0
void DOMWindow::languagesChanged()
{
    // https://html.spec.whatwg.org/multipage/system-state.html#dom-navigator-languages
    if (auto document = makeRefPtr(this->document()))
        document->queueTaskToDispatchEventOnWindow(TaskSource::DOMManipulation, Event::create(eventNames().languagechangeEvent, Event::CanBubble::No, Event::IsCancelable::No));
}
#endif

void DOMWindow::dispatchLoadEvent()
{
#if ENABLE(RMAI)
    if(Document* doc = document()){
        if(GalioWafer* wafer = doc->ownerWafer())
            wafer->documentLoaded(doc);
    }
#endif

    // If we did not protect it, the document loader and its timing subobject might get destroyed
    // as a side effect of what event handling code does.
    RefPtr<DOMWindow> protectedThis(this);
    RefPtr<DocumentLoader> protectedLoader(frame() ? frame()->loader()->documentLoader() : static_cast<DocumentLoader*>(0));
    bool shouldMarkLoadEventTimes = protectedLoader && !protectedLoader->timing()->loadEventStart();

    if (shouldMarkLoadEventTimes)
        protectedLoader->timing()->markLoadEventStart();

    auto& eventNames = WebCore::eventNames();
    dispatchEvent(Event::create(eventNames.loadEvent, false, false), document());

    if (shouldMarkLoadEventTimes)
        protectedLoader->timing()->markLoadEventEnd();


    // Send a separate load event to the element that owns this frame.
    if (Frame* frame = this->frame()) {
        if (auto* owner = frame->ownerElement())
            owner->dispatchEvent(Event::create(eventNames.loadEvent, false, false));

        InspectorInstrumentation::loadEventFired(frame);
    }

#if ENABLE(RMAI)
    if(Document* doc = document()){
		// Also send load event to the ownerWafer, if any
		if(GalioWafer* wafer = doc->ownerWafer()){
			wafer->dispatchWaferEvent(eventNames.loadEvent);
		}
    }
#endif
}

void DOMWindow::dispatchEvent(PassRefPtr<Event> prpEvent, PassRefPtr<EventTarget> prpTarget)
{
    // FIXME: It's confusing to have both the inherited EventTarget::dispatchEvent function
    // and this function, which does something nearly identical but subtly different if
    // called with a target of null. Most callers pass the document as the target, though.
    // Fixing this could allow us to remove the special case in DocumentEventQueue::dispatchEvent.

    RefPtr<EventTarget> protect(this);
    RefPtr<Event> event(prpEvent);

    // Pausing a page may trigger pagehide and pageshow events. WebCore also implicitly fires these
    // events when closing a WebView. Here we keep track of the state of the page to prevent duplicate,
    // unbalanced events per the definition of the pageshow event:
    // <http://www.whatwg.org/specs/web-apps/current-work/multipage/history.html#event-pageshow>.
    // FIXME: This code should go at call sites where pageshowEvent and pagehideEvents are
    // generated, not here inside the event dispatching process.

    event->setTarget(prpTarget ? prpTarget : this);

    EventPath eventPath(*this);
    event->setCurrentTarget(this);
    event->setEventPhase(Event::AT_TARGET);
    event->resetBeforeDispatch();
    event->setEventPath(eventPath);

    // FIXME: It doesn't seem right to have the inspector instrumentation here since not all
	// events dispatched to the window object are guaranteed to flow through this function.
	// But the instrumentation prevents us from calling EventDispatcher::dispatchEvent here.
    auto cookie = InspectorInstrumentation::willDispatchEventOnWindow(frame(), *event, this);

    // FIXME: We should use EventDispatcher everywhere.
	fireEventListeners(*event, EventInvokePhase::Capturing);
	fireEventListeners(*event, EventInvokePhase::Bubbling);

    InspectorInstrumentation::didDispatchEventOnWindow(cookie);

    event->resetAfterDispatch();
}

void DOMWindow::removeAllEventListeners()
{
    EventTarget::removeAllEventListeners();

#if ENABLE(DEVICE_ORIENTATION)
    if (DeviceMotionController* controller = DeviceMotionController::from(page()))
        controller->removeAllListeners(this);
    if (DeviceOrientationController* controller = DeviceOrientationController::from(page()))
        controller->removeAllListeners(this);
#endif

#if ENABLE(WEB_TIMING)
    if (m_performance)
        m_performance->removeAllEventListeners();
#endif

    removeAllUnloadEventListeners(this);
    removeAllBeforeUnloadEventListeners(this);
}

void DOMWindow::captureEvents()
{
    // Not implemented.
}

void DOMWindow::releaseEvents()
{
    // Not implemented.
}

void DOMWindow::finishedLoading()
{
    if (m_shouldPrintWhenFinishedLoading) {
        m_shouldPrintWhenFinishedLoading = false;
        print();
    }
}

void DOMWindow::setLocation(DOMWindow& activeWindow, const KURL& completedURL, SetLocationLocking locking)
{
    if (!isCurrentlyDisplayedInFrame())
        return;

    Document* activeDocument = activeWindow.document();
    if (!activeDocument)
        return;

    auto* frame = this->frame();
    if (!activeDocument->canNavigate(frame, completedURL))
        return;

    if (isInsecureScriptAccess(&activeWindow, completedURL))
        return;

    // Check the CSP of the embedder to determine if we allow execution of javascript: URLs via child frame navigation.
    if (completedURL.protocolIsJavaScript() && frameElement() && !frameElement()->document().contentSecurityPolicy()->allowJavaScriptURLs(blankURL().string(), OrdinalNumber(), completedURL.string(), frameElement()))
        return;

    RefPtr<Frame> localParent = frame->tree()->parent();
    // If the loader for activeWindow's frame (browsing context) has no outgoing referrer, set its outgoing referrer
    // to the URL of its parent frame's Document.
    RefPtr<Frame> activeFrame = activeWindow.frame();
    if (activeFrame && activeFrame->loader()->outgoingReferrer().isEmpty() && localParent)
        activeFrame->loader()->setOutgoingReferrer(document()->completeURL(localParent->document()->url().strippedForUseAsReferrer()));

    // We want a new history item if we are processing a user gesture.
    LockHistory lockHistory = (locking != LockHistoryBasedOnGestureState || !ScriptController::processingUserGesture()) ? LockHistory::Yes : LockHistory::No;
	LockBackForwardList lockBackForwardList = (locking != LockHistoryBasedOnGestureState) ? LockBackForwardList::Yes : LockBackForwardList::No;
	frame->navigationScheduler()->scheduleLocationChange(*activeDocument, *activeDocument->securityOrigin(),
        // FIXME: What if activeDocument()->frame() is 0?
        completedURL, activeDocument->frame()->loader()->outgoingReferrer(),
        lockHistory, lockBackForwardList);
}

void DOMWindow::printErrorMessage(const String& message)
{
    if (message.isEmpty())
        return;

    auto* frame = this->frame();
    if (frame && frame->settings().privateBrowsingEnabled())
        return;

    if (auto* pageConsole = this->pageConsole())
    	pageConsole->addMessage(MessageSource::JS, MessageLevel::Error, message);
}

String DOMWindow::crossDomainAccessErrorMessage(DOMWindow* activeWindow)
{
    const KURL& activeWindowURL = activeWindow->url();
    if (activeWindowURL.isNull())
        return String();

    ASSERT(!activeWindow->document()->securityOrigin()->canAccess(document()->securityOrigin()));

    // FIXME: This message, and other console messages, have extra newlines. Should remove them.
    String message = tryMakeString("Unsafe JavaScript attempt to access frame with URL ", document()->url().string(), " from frame with URL ", activeWindowURL.string(), ".");

    // Sandbox errors.
    if (document()->isSandboxed(SandboxOrigin) || activeWindow->document()->isSandboxed(SandboxOrigin)) {
        if (document()->isSandboxed(SandboxOrigin) && activeWindow->document()->isSandboxed(SandboxOrigin))
            return tryMakeString("Sandbox access violation: ", message, " Both frames are sandboxed into unique origins.\n");
        if (document()->isSandboxed(SandboxOrigin))
            return tryMakeString("Sandbox access violation: ", message, " The frame being accessed is sandboxed into a unique origin.\n");
        return tryMakeString("Sandbox access violation: ", message, " The frame requesting access is sandboxed into a unique origin.\n");
    }

    SecurityOrigin* activeOrigin = activeWindow->document()->securityOrigin();
    SecurityOrigin* targetOrigin = document()->securityOrigin();
    if (targetOrigin->protocol() != activeOrigin->protocol())
        return tryMakeString(message, " The frame requesting access has a protocol of '", activeOrigin->protocol(), "', the frame being accessed has a protocol of '", targetOrigin->protocol(), "'. Protocols must match.\n");

    // 'document.domain' errors.
    if (targetOrigin->domainWasSetInDOM() && activeOrigin->domainWasSetInDOM())
        return tryMakeString(message, " The frame requesting access set 'document.domain' to '", activeOrigin->domain(), "', the frame being accessed set it to '", targetOrigin->domain(), "'. Both must set 'document.domain' to the same value to allow access.\n");
    if (activeOrigin->domainWasSetInDOM())
        return tryMakeString(message, " The frame requesting access set 'document.domain' to '", activeOrigin->domain(), "', but the frame being accessed did not. Both must set 'document.domain' to the same value to allow access.\n");
    if (targetOrigin->domainWasSetInDOM())
        return tryMakeString(message, " The frame being accessed set 'document.domain' to '", targetOrigin->domain(), "', but the frame requesting access did not. Both must set 'document.domain' to the same value to allow access.\n");

    // Default.
    return tryMakeString("Blocked a frame with origin \"", activeOrigin->toString(), "\" from accessing a frame with origin \"", targetOrigin->toString(), "\". Protocols, domains, and ports must match.");
}

bool DOMWindow::isInsecureScriptAccess(DOMWindow* activeWindow, const String& urlString)
{
    if (!protocolIsJavaScript(urlString))
        return false;

    // If this DOMWindow isn't currently active in the Frame, then there's no
    // way we should allow the access.
    // FIXME: Remove this check if we're able to disconnect DOMWindow from
    // Frame on navigation: https://bugs.webkit.org/show_bug.cgi?id=62054
    if (isCurrentlyDisplayedInFrame()) {
        // FIXME: Is there some way to eliminate the need for a separate "activeWindow == this" check?
        if (activeWindow == this)
            return false;

        // FIXME: The name canAccess seems to be a roundabout way to ask "can execute script".
        // Can we name the SecurityOrigin function better to make this more clear?
        if (activeWindow->securityOrigin()->canAccess(securityOrigin()))
            return false;
    }

    printErrorMessage(crossDomainAccessErrorMessage(activeWindow));
    return true;
}

PassRefPtr<Frame> DOMWindow::createWindow(const String& urlString, const AtomicString& frameName, const WindowFeatures& windowFeatures,
    DOMWindow* activeWindow, Frame* firstFrame, Frame* openerFrame, PrepareDialogFunction function, void* functionContext)
{
    Frame* activeFrame = activeWindow->frame();
    if (!activeFrame)
        return nullptr;

    RefPtr<Document> activeDocument = activeWindow->document();
    if (!activeDocument)
        return nullptr;

    // For whatever reason, Firefox uses the first frame to determine the outgoingReferrer. We replicate that behavior here.
    String referrer = windowFeatures.noreferrer ? String() : firstFrame->loader()->outgoingReferrer();

    KURL completedURL = urlString.isEmpty() ? KURL(ParsedURLString, emptyString()) : firstFrame->document()->completeURL(urlString);
    if (!completedURL.isEmpty() && !completedURL.isValid()) {
        // Don't expose client code to invalid URLs.
        activeWindow->printErrorMessage(tryMakeString("Unable to open a window with invalid URL '", completedURL.string(), "'.\n"));
        return nullptr;
    }

    ResourceRequest request(completedURL, referrer);
    FrameLoader::addHTTPOriginIfNeeded(request, firstFrame->loader()->outgoingOrigin());
    FrameLoadRequest frameRequest(*activeDocument, *activeDocument->securityOrigin(), request, frameName, LockHistory::No, LockBackForwardList::No, ReferrerPolicy::EmptyString, AllowNavigationToInvalidURL::Yes, NewFrameOpenerPolicy::Allow, ReplaceDocumentIfJavaScriptURL);

    // We pass the opener frame for the lookupFrame in case the active frame is different from
    // the opener frame, and the name references a frame relative to the opener frame.
    bool created;
    RefPtr<Frame> newFrame = WebCore::createWindow(*activeFrame, openerFrame, frameRequest, windowFeatures, created);
    if (!newFrame)
        return nullptr;

    bool noopener = windowFeatures.noopener || windowFeatures.noreferrer;
    if (!noopener)
    	newFrame->loader()->setOpener(openerFrame);

    if (created)
    	newFrame->page()->setOpenedByDOM();

    if (newFrame->document()->domWindow()->isInsecureScriptAccess(activeWindow, completedURL))
    	return noopener ? PassRefPtr<Frame>(nullptr) : newFrame.release();

    if (function)
        function(newFrame->document()->domWindow(), functionContext);

    if (created) {
        ResourceRequest resourceRequest(completedURL, referrer, UseProtocolCachePolicy);
        FrameLoadRequest frameRequest(*activeDocument, *activeDocument->securityOrigin(), resourceRequest, "_self", LockHistory::No, LockBackForwardList::No, ReferrerPolicy::EmptyString, AllowNavigationToInvalidURL::Yes, NewFrameOpenerPolicy::Allow);
        newFrame->loader()->changeLocation(frameRequest);
    } else if (!urlString.isEmpty()) {
    	LockHistory lockHistory = ScriptController::processingUserGesture() ? LockHistory::No : LockHistory::Yes;
		newFrame->navigationScheduler()->scheduleLocationChange(*activeDocument, *activeDocument->securityOrigin(), completedURL, referrer, lockHistory, LockBackForwardList::No);
    }

    // Navigating the new frame could result in it being detached from its page by a navigation policy delegate.
    if (!newFrame->page())
        return nullptr;

    return noopener ? PassRefPtr<Frame>(nullptr) : newFrame.release();
}

PassRefPtr<DOMWindow> DOMWindow::open(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& urlString, const AtomicString& frameName, const String& windowFeaturesString)
{
    if (!isCurrentlyDisplayedInFrame())
        return nullptr;

    RefPtr<Document> activeDocument = activeWindow->document();
    if (!activeDocument)
        return nullptr;

    Frame* firstFrame = firstWindow->frame();
    if (!firstFrame)
        return nullptr;

    auto* frame = this->frame();
    if (!firstWindow->allowPopUp()) {
        // Because FrameTree::findFrameForNavigation() returns true for empty strings, we must check for empty frame names.
        // Otherwise, illegitimate window.open() calls with no name will pass right through the popup blocker.
        if (frameName.isEmpty() || !frame->loader()->findFrameForNavigation(frameName, activeDocument.get()))
            return nullptr;
    }

    // Get the target frame for the special cases of _top and _parent.
    // In those cases, we schedule a location change right now and return early.
    Frame* targetFrame = nullptr;
    if (equalIgnoringASCIICase(frameName, "_top"))
        targetFrame = frame->tree()->top();
    else if (equalIgnoringASCIICase(frameName, "_parent")) {
        if (Frame* parent = frame->tree()->parent())
            targetFrame = parent;
        else
            targetFrame = frame;
    }
    if (targetFrame) {
        if (!activeDocument->canNavigate(targetFrame))
            return nullptr;

        KURL completedURL = firstFrame->document()->completeURL(urlString);

        if (targetFrame->document()->domWindow()->isInsecureScriptAccess(activeWindow, completedURL))
            return targetFrame->document()->domWindow();

        if (urlString.isEmpty())
            return targetFrame->document()->domWindow();

        // For whatever reason, Firefox uses the first window rather than the active window to
        // determine the outgoing referrer. We replicate that behavior here.
        LockHistory lockHistory = ScriptController::processingUserGesture() ? LockHistory::No : LockHistory::Yes;
		targetFrame->navigationScheduler()->scheduleLocationChange(*activeDocument, *activeDocument->securityOrigin(), completedURL, firstFrame->loader()->outgoingReferrer(),
            lockHistory, LockBackForwardList::No);
        return targetFrame->document()->domWindow();
    }

    WindowFeatures windowFeatures(windowFeaturesString);
    FloatRect windowRect(windowFeatures.x.value_or(0), windowFeatures.y.value_or(0), windowFeatures.width.value_or(0), windowFeatures.height.value_or(0));
    Page* page = frame->page();
    DOMWindow::adjustWindowRect(screenAvailableRect(page ? page->mainFrame()->view() : 0), windowRect, windowRect);
    windowFeatures.x = windowRect.x();
    windowFeatures.y = windowRect.y();
    windowFeatures.height = windowRect.height();
    windowFeatures.width = windowRect.width();

    RefPtr<Frame> result = createWindow(urlString, frameName, windowFeatures, activeWindow, firstFrame, frame);
    return result ? result->document()->domWindow() : nullptr;
}

void DOMWindow::showModalDialog(const String& urlString, const String& dialogFeaturesString,
    DOMWindow* activeWindow, DOMWindow* firstWindow, PrepareDialogFunction function, void* functionContext)
{
    if (!isCurrentlyDisplayedInFrame())
        return;
    Frame* activeFrame = activeWindow->frame();
    if (!activeFrame)
        return;
    Frame* firstFrame = firstWindow->frame();
    if (!firstFrame)
        return;

    auto* frame = this->frame();
    auto* page = frame->page();
    if (!page)
        return;

    if (!page->arePromptsAllowed()) {
        printErrorMessage("Use of window.showModalDialog is not allowed while unloading a page.");
        return;
    }

    if (!canShowModalDialog(frame) || !firstWindow->allowPopUp())
        return;

    RefPtr<Frame> dialogFrame = createWindow(urlString, emptyAtom, WindowFeatures(dialogFeaturesString, screenAvailableRect(frame->view())),
        activeWindow, firstFrame, frame, function, functionContext);
    if (!dialogFrame)
        return;

    dialogFrame->page()->chrome()->runModal();
}

void DOMWindow::enableSuddenTermination()
{
    if (Page* page = this->page())
        page->chrome()->enableSuddenTermination();
}

void DOMWindow::disableSuddenTermination()
{
    if (Page* page = this->page())
        page->chrome()->disableSuddenTermination();
}

Frame* DOMWindow::frame() const
{
    auto* document = this->document();
    return document ? document->frame() : reinterpret_cast<Frame*>(0);
}

} // namespace WebCore
