/*
 * Copyright (C) 2006, 2007, 2009, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "AbstractDOMWindow.h"
#include "Base64Utilities.h"
#include "ContextDestructionObserver.h"
#include "EventTarget.h"
#include "FrameDestructionObserver.h"
#include "Handle.h"
#include "ImageBitmap.h"
#include "KURL.h"
#include "ReducedResolutionSeconds.h"
#include "ScrollToOptions.h"
#include "Supplementable.h"
#include <wtf/Forward.h>
#include <wtf/WeakPtr.h>

namespace JSC {
class ExecState;
class JSObject;
class JSValue;
template<typename T, ShouldStrongDestructorGrabLock shouldStrongDestructorGrabLock> class Strong;
}

namespace WebCore {

class BarProp;
class CSSRuleList;
class CSSStyleDeclaration;
class Crypto;
class DOMApplicationCache;
class DOMSelection;
class DOMURL;
class DOMWindowProperty;
class Database;
class DatabaseCallback;
class Document;
class Element;
class EventListener;
class FloatRect;
class Frame;
class History;
class IdleRequestCallback;
class IDBFactory;
class Location;
class MediaQueryList;
class MessageEvent;
class Navigator;
class Node;
class NotificationCenter;
class Page;
class PageConsoleClient;
class Performance;
class PostMessageTimer;
class RequestAnimationFrameCallback;
class RequestIdleCallback;
class ScheduledAction;
class Screen;
class ScriptCallStack;
class SecurityOrigin;
class SerializedScriptValue;
class Storage;
class StyleMedia;
class WebKitPoint;
class OfflineCache;

#if ENABLE(RMAI)
class GalioDebug;
class MOM;
#endif

struct IdleRequestOptions;
struct ImageBitmapOptions;
struct WindowFeatures;

typedef int ExceptionCode;

enum SetLocationLocking { LockHistoryBasedOnGestureState, LockHistoryAndBackForwardList };

#if ENABLE(CEHTML)
enum RenderMode { RenderModeDelayed, RenderModeDefault };
#endif

// FIXME: Rename DOMWindow to LocalWindow and AbstractDOMWindow to DOMWindow.
class DOMWindow FINAL
	: public AbstractDOMWindow
	, public ContextDestructionObserver
    , public Base64Utilities
	, public Supplementable<DOMWindow> {
public:
	static ALWAYS_INLINE PassRefPtr<DOMWindow> create(Document* document) { return adoptRef(new DOMWindow(document)); }
	virtual ~DOMWindow();

    // In some rare cases, we'll re-used a DOMWindow for a new Document. For example,
    // when a script calls window.open("..."), the browser gives JavaScript a window
    // synchronously but kicks off the load in the window asynchronously. Web sites
    // expect that modifications that they make to the window object synchronously
    // won't be blown away when the network load commits. To make that happen, we
    // "securely transition" the existing DOMWindow to the Document that results from
    // the network load. See also SecurityContext::isSecureTransitionTo.
    void didSecureTransitionTo(Document*);

	virtual ScriptExecutionContext* scriptExecutionContext() const;

	virtual bool isLocalDOMWindow() const FINAL { return true; }
	virtual bool isRemoteDOMWindow() const FINAL { return false; }

	virtual DOMWindow* toDOMWindow();

	void registerProperty(DOMWindowProperty*);
	void unregisterProperty(DOMWindowProperty*);

	void clear();
	void suspendForPageCache();
	void resumeFromPageCache();

    virtual Frame* frame() const FINAL;

	Page* page() const;

	PassRefPtr<MediaQueryList> matchMedia(const String&);

    // FIXME: Callers should use document()->securityOrigin() directly.
    SecurityOrigin* securityOrigin() const;

    // FIXME: We should get the URL from document()->url() directly.
	void setURL(const KURL& url) { m_url = url; }
	const KURL& url() const { return m_url; }

	unsigned pendingUnloadEventListeners() const;

	static bool dispatchAllPendingBeforeUnloadEvents();
	static void dispatchAllPendingUnloadEvents();

	void adjustWindowRect(const FloatRect& screen, FloatRect& window, const FloatRect& pendingChanges) const;

	bool allowPopUp(); // Call on first window, not target window.
	static bool allowPopUp(Frame* firstFrame);
	static bool canShowModalDialog(const Frame*);
    void setCanShowModalDialogOverride(bool);

	Screen* screen();
	History* history();
	Crypto* crypto();
	BarProp* locationbar();
	BarProp* menubar();
	BarProp* personalbar();
	BarProp* scrollbars();
	BarProp* statusbar();
	BarProp* toolbar();
	Navigator* navigator();
	ALWAYS_INLINE Navigator* clientInformation() { return navigator(); }

	Location* location();
	void setLocation(DOMWindow& activeWindow, const KURL& completedURL, SetLocationLocking = LockHistoryBasedOnGestureState);

	DOMSelection* getSelection();

	Element* frameElement() const;

    void focus(bool allowFocus = false);
	void focus(DOMWindow* incumbentWindow);
	void blur();
    void close();
	void close(Document*);
	void print();
	void stop();

#if ENABLE(HBBTV)
	void debug(const String& arg);
#endif

	PassRefPtr<DOMWindow> open(DOMWindow* activeWindow, DOMWindow* firstWindowconst, const String& urlString, const AtomicString& frameName, const String& windowFeaturesString);

	typedef void (*PrepareDialogFunction)(DOMWindow*, void* context);
	void showModalDialog(const String& urlString, const String& dialogFeaturesString,
		DOMWindow* activeWindow, DOMWindow* firstWindow, PrepareDialogFunction, void* functionContext);

	void alert(const String& message);
	bool confirm(const String& message);
	String prompt(const String& message, const String& defaultValue);

	bool find(const String&, bool caseSensitive, bool backwards, bool wrap, bool wholeWord, bool searchInFrames, bool showDialog) const;

	bool offscreenBuffering() const;

	int outerHeight() const;
	int outerWidth() const;
	int innerHeight() const;
	int innerWidth() const;
	int screenX() const;
	int screenY() const;
	ALWAYS_INLINE int screenLeft() const { return screenX(); }
	ALWAYS_INLINE int screenTop() const { return screenY(); }
	int scrollX() const;
	int scrollY() const;

	bool closed() const;

	unsigned length() const;

	String name() const;
	void setName(const String&);

	String status() const;
	void setStatus(const String&);
	String defaultStatus() const;
	void setDefaultStatus(const String&);

	// This attribute is an alias of defaultStatus and is necessary for legacy uses.
	ALWAYS_INLINE String defaultstatus() const { return defaultStatus(); }
	ALWAYS_INLINE void setDefaultstatus(const String& status) { setDefaultStatus(status); }

	// Self-referential attributes

	DOMWindow* self() const;
	ALWAYS_INLINE DOMWindow* window() const { return self(); }
	ALWAYS_INLINE DOMWindow* frames() const { return self(); }

	DOMWindow* opener() const;
    void disownOpener();
	DOMWindow* parent() const;
	DOMWindow* top() const;

    String origin() const;

#if ENABLE(CEHTML)
	bool escapeBeyondTopMost(const String& url, DOMWindow* activeWindow, DOMWindow* firstWindow);
	const String cea2014_protocol_subversionNr() const { return "1.0"; }
	const String cea2014_protocol_version() const { return "CE_HTML"; }
	void reload(DOMWindow* activeWindow);
	void replace(const String& url, DOMWindow* activeWindow, DOMWindow* firstWindow);
	bool requestFocus(ScriptExecutionContext* = 0);
	int setTimer(PassRefPtr<ScheduledAction>, int timeout, ExceptionCode&);
	PassRefPtr<DOMWindow> getFrame(const AtomicString& id);
	unsigned httptimeout() const;
	void setHttptimeout(const unsigned timeout);
	void exitUnit();
	void setRenderMode(const String& mode);
	void setHttpTimeout(const String& url, unsigned seconds);
	PassRefPtr<Document> openURL(const String& url) const;

	RenderMode m_renderMode { RenderModeDefault };
#endif

	// DOM Level 2 AbstractView Interface

	Document* document() const;

	// CSSOM View Module

	PassRefPtr<StyleMedia> styleMedia();

	// DOM Level 2 Style Interface

	PassRefPtr<CSSStyleDeclaration> getComputedStyle(Element*, const String& pseudoElt) const;

	// WebKit extensions

	PassRefPtr<CSSRuleList> getMatchedCSSRules(Element*, const String& pseudoElt, bool authorOnly = true) const;
	double devicePixelRatio() const;

	PassRefPtr<WebKitPoint> webkitConvertPointFromPageToNode(Node*, const WebKitPoint*) const;
	PassRefPtr<WebKitPoint> webkitConvertPointFromNodeToPage(Node*, const WebKitPoint*) const;

	PageConsoleClient* pageConsole() const;

	void printErrorMessage(const String&);
	String crossDomainAccessErrorMessage(DOMWindow* activeWindow);

#if ENABLE(RMAI)
	MOM* mom();
	GalioDebug* debug();
#endif

	void postMessage(JSC::ExecState*, DOMWindow* incumbentWindow, PassRefPtr<SerializedScriptValue> message, const Vector<RefPtr<MessagePort>>*, const String& targetOrigin, ExceptionCode&);
	void postMessageTimerFired(PassOwnPtr<PostMessageTimer>);
	void dispatchMessageEventWithOriginCheck(SecurityOrigin* intendedTargetOrigin, PassRefPtr<Event>, PassRefPtr<ScriptCallStack>);

    void scrollBy(const ScrollToOptions&) const;
    void scrollBy(double x, double y) const;
    void scrollTo(const ScrollToOptions&) const;
    void scrollTo(double x, double y) const;

	void moveBy(float x, float y) const;
	void moveTo(float x, float y) const;

	void resizeBy(float x, float y) const;
	void resizeTo(float width, float height) const;

	// Timers
	int setTimeout(PassRefPtr<ScheduledAction>, int timeout, ExceptionCode&);
	void clearTimeout(int timeoutId);
	int setInterval(PassRefPtr<ScheduledAction>, int timeout, ExceptionCode&);
	void clearInterval(int timeoutId);

	// WebKit animation extensions
#if ENABLE(REQUEST_ANIMATION_FRAME)
	int requestAnimationFrame(PassRefPtr<RequestAnimationFrameCallback>);
    int webkitRequestAnimationFrame(PassRefPtr<RequestAnimationFrameCallback>);
	void cancelAnimationFrame(int id);
#endif

    // ImageBitmap
    void createImageBitmap(PassRefPtr<HTMLImageElement>, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<HTMLCanvasElement>, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<HTMLVideoElement>, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<ImageBitmap>, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<Blob>, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<ImageData>, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<HTMLImageElement>, int sx, int sy, int sw, int sh, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<HTMLCanvasElement>, int sx, int sy, int sw, int sh, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<HTMLVideoElement>, int sx, int sy, int sw, int sh, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<ImageBitmap>, int sx, int sy, int sw, int sh, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<Blob>, int sx, int sy, int sw, int sh, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
    void createImageBitmap(PassRefPtr<ImageData>, int sx, int sy, int sw, int sh, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);

    int requestIdleCallback(PassRefPtr<IdleRequestCallback>, const IdleRequestOptions&);
    void cancelIdleCallback(int id);

    // Secure Contexts
    bool isSecureContext() const;

	// Events
	// EventTarget API
	virtual bool addEventListener(const AtomicString& eventType, PassRefPtr<EventListener>, const AddEventListenerOptions&) override;
	virtual bool removeEventListener(const AtomicString& eventType, EventListener*, const EventListenerOptions& = EventListenerOptions()) override;
	virtual void removeAllEventListeners() override;

	using EventTarget::dispatchEvent;
	void dispatchEvent(PassRefPtr<Event> prpEvent, PassRefPtr<EventTarget> prpTarget);
	void dispatchLoadEvent();

	void captureEvents();
	void releaseEvents();

	void finishedLoading();

	// HTML 5 key/value storage
	Storage* sessionStorage(ExceptionCode&);
	Storage* localStorage(ExceptionCode&);
	ALWAYS_INLINE Storage* optionalSessionStorage() const { return m_sessionStorage.get(); }
	ALWAYS_INLINE Storage* optionalLocalStorage() const { return m_localStorage.get(); }
	OfflineCache* offlineCache(ExceptionCode&);

#if ENABLE(LEGACY_NOTIFICATIONS)
	NotificationCenter* webkitNotifications() const;

	// Renders webkitNotifications object safely inoperable, disconnects
	// if from embedder-provided NotificationClient.
	void resetNotifications();
#endif

	DOMApplicationCache* applicationCache();
	ALWAYS_INLINE DOMApplicationCache* optionalApplicationCache() const { return m_applicationCache.get(); }

#if ENABLE(ORIENTATION_EVENTS)
	// This is the interface orientation in degrees. Some examples are:
	//  0 is straight up; -90 is when the device is rotated 90 clockwise;
	//  90 is when rotated counter clockwise.
	int orientation() const;
#endif

#if ENABLE(WEB_TIMING)
	Performance* performance() const;
#endif
	ReducedResolutionSeconds nowTimestamp() const;

    void freezeNowTimestamp();
    void unfreezeNowTimestamp();
    ReducedResolutionSeconds frozenNowTimestamp() const;

#if ENABLE(RMAI)
	mutable RefPtr<MOM> m_mom;
	mutable RefPtr<GalioDebug> m_galioDebug;
#endif

#if ENABLE(USER_MESSAGE_HANDLERS)
    bool shouldHaveWebKitNamespaceForWorld(DOMWrapperWorld&);
    WebKitNamespace* webkitNamespace();
#endif

	// FIXME: When this DOMWindow is no longer the active DOMWindow (i.e.,
	// when its document is no longer the document that is displayed in its
	// frame), we would like to zero out m_frame to avoid being confused
	// by the document that is currently active in m_frame.
	bool isCurrentlyDisplayedInFrame() const;

	void willDetachDocumentFromFrame();
	void willDestroyCachedFrame();

    void enableSuddenTermination();
    void disableSuddenTermination();

    void willDestroyDocumentInFrame();
    void frameDestroyed();

private:
	explicit DOMWindow(Document*);

	static PassRefPtr<Frame> createWindow(const String& urlString, const AtomicString& frameName, const WindowFeatures&,
		DOMWindow* activeWindow, Frame* firstFrame, Frame* openerFrame,
		PrepareDialogFunction = 0, void* functionContext = 0);
	bool isInsecureScriptAccess(DOMWindow* activeWindow, const String& urlString);

	void resetDOMWindowProperties();
	void disconnectDOMWindowProperties();
	void reconnectDOMWindowProperties();

	template<typename T> void createImageBitmapInternal(PassRefPtr<T>, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);
	template<typename T> void createImageBitmapInternal(PassRefPtr<T>, int sx, int sy, int sw, int sh, const ImageBitmapOptions&, PassRefPtr<DeferredPromise>);

	KURL m_url;

	bool m_shouldPrintWhenFinishedLoading { false };
	bool m_suspendedForPageCache { false };
    Optional<bool> m_canShowModalDialogOverride;

	HashSet<DOMWindowProperty*> m_properties;

	mutable RefPtr<Crypto> m_crypto;
	mutable RefPtr<History> m_history;
	mutable RefPtr<BarProp> m_locationbar;
	mutable RefPtr<StyleMedia> m_media;
	mutable RefPtr<BarProp> m_menubar;
	mutable RefPtr<Navigator> m_navigator;
	mutable RefPtr<BarProp> m_personalbar;
	mutable RefPtr<Screen> m_screen;
	mutable RefPtr<BarProp> m_scrollbars;
    mutable RefPtr<DOMSelection> m_selection;
	mutable RefPtr<BarProp> m_statusbar;
	mutable RefPtr<BarProp> m_toolbar;
	mutable RefPtr<Location> m_location;

	String m_status;
	String m_defaultStatus;

	mutable RefPtr<Storage> m_sessionStorage;
	mutable RefPtr<Storage> m_localStorage;
	mutable RefPtr<DOMApplicationCache> m_applicationCache;
	mutable RefPtr<OfflineCache> m_offlineCache;

#if ENABLE(LEGACY_NOTIFICATIONS)
	mutable RefPtr<NotificationCenter> m_notifications;
#endif

#if ENABLE(WEB_TIMING)
	mutable RefPtr<Performance> m_performance;
#endif

    Optional<ReducedResolutionSeconds> m_frozenNowTimestamp;
};

ALWAYS_INLINE String DOMWindow::status() const
{
	return m_status;
}

ALWAYS_INLINE String DOMWindow::defaultStatus() const
{
	return m_defaultStatus;
}

} // namespace WebCore

#if 0
SPECIALIZE_TYPE_TRAITS_BEGIN(WebCore::DOMWindow)
    static bool isType(const WebCore::AbstractDOMWindow& window) { return window.isLocalDOMWindow(); }
    static bool isType(const WebCore::EventTarget& target) { return target.eventTargetInterface() == WebCore::DOMWindowEventTargetInterfaceType; }
SPECIALIZE_TYPE_TRAITS_END()
#endif
