/*
 * Copyright (C) 2006 Apple Computer, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#include <wtf/NotFound.h>
#include <wtf/WeakPtr.h>
#include <wtf/text/AtomicString.h>

namespace WebCore {

class Frame;
class TreeScope;

class FrameTree {
	WTF_MAKE_NONCOPYABLE(FrameTree);
public:
	const static unsigned invalidCount = static_cast<unsigned>(WTF::notFound);

	FrameTree(Frame* thisFrame, Frame* parentFrame);

	~FrameTree();

	const AtomicString& name() const { return m_name; }
	const AtomicString& uniqueName() const { return m_uniqueName; }
	void setName(const AtomicString&);
	void clearName();
	Frame* parent() const;

	Frame* nextSibling() const { return m_nextSibling.get(); }
	Frame* previousSibling() const { return m_previousSibling.get(); }
	Frame* firstChild() const { return m_firstChild.get(); }
	Frame* lastChild() const { return m_lastChild.get(); }

	Frame* firstRenderedChild() const;
	Frame* nextRenderedSibling() const;

	bool isDescendantOf(const Frame* ancestor) const;
	Frame* traverseNext(const Frame* stayWithin = 0) const;
	// Rendered means being the main frame or having an ownerRenderer. It may not have been parented in the Widget tree yet (see WidgetHierarchyUpdatesSuspensionScope).
	Frame* traverseNextRendered(const Frame* stayWithin = 0) const;
	Frame* traverseNextWithWrap(bool) const;
	Frame* traversePreviousWithWrap(bool) const;

    Frame* traverseNextInPostOrderWithWrap(bool) const;

	void appendChild(PassRefPtr<Frame>);
	void detachFromParent() { m_parent = nullptr; }
	void removeChild(Frame*);

	Frame* child(unsigned index) const;
	Frame* child(const AtomicString& name) const;
	Frame* find(const AtomicString& name, Frame& activeFrame) const;
	unsigned childCount() const;

	Frame* top() const;

	Frame* scopedChild(unsigned index) const;
	Frame* scopedChild(const AtomicString& name) const;
	unsigned scopedChildCount() const;

	void resetFrameIdentifiers() { m_frameIDGenerator = 0; }

private:
    Frame* deepFirstChild() const;
	Frame* deepLastChild() const;
	void actuallyAppendChild(PassRefPtr<Frame>);

	bool scopedBy(TreeScope*) const;
	Frame* scopedChild(unsigned index, TreeScope*) const;
	Frame* scopedChild(const AtomicString& name, TreeScope*) const;
	unsigned scopedChildCount(TreeScope*) const;

	AtomicString uniqueChildName(const AtomicString& requestedName) const;
	AtomicString generateUniqueName() const;

	Frame* m_thisFrame;

	WeakPtr<Frame> m_parent;
	AtomicString m_name; // The actual frame name (may be empty).
	AtomicString m_uniqueName;

	// FIXME: use ListRefPtr?
	RefPtr<Frame> m_nextSibling;
	WeakPtr<Frame> m_previousSibling;
	RefPtr<Frame> m_firstChild;
	WeakPtr<Frame> m_lastChild;
	mutable unsigned m_scopedChildCount;
	mutable uint64_t m_frameIDGenerator;
};

inline Frame* FrameTree::parent() const
{
	return m_parent.get();
}

inline void FrameTree::clearName()
{
	m_name = AtomicString();
	m_uniqueName = AtomicString();
}

ASCIILiteral blankTargetFrameName();
ASCIILiteral selfTargetFrameName();

bool isBlankTargetFrameName(StringView);
bool isParentTargetFrameName(StringView);
bool isSelfTargetFrameName(StringView);
bool isTopTargetFrameName(StringView);

} // namespace WebCore

#ifndef NDEBUG
// Outside the WebCore namespace for ease of invocation from gdb.
void showFrameTree(const WebCore::Frame*);
#endif
