/*
 * Copyright (C) 2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if ENABLE(INTERSECTION_OBSERVER)
#include "IntersectionObserver.h"

#include "CSSParserTokenRange.h"
#include "CSSPropertyParserHelpers.h"
#include "CSSTokenizer.h"
#include "Element.h"
#include "Frame.h"
#include "IntersectionObserverCallback.h"
#include "IntersectionObserverEntry.h"
#include "JSDOMConvert.h"
#include "JSIntersectionObserverCallback.h"
#include "Performance.h"
#include "heap/SlotVisitorInlines.h"
#include <wtf/Vector.h>

namespace WebCore {

unsigned gIntersectionObserverCount;

static LengthBox parseRootMargin(const String& rootMargin, ExceptionCode& ec)
{
    CSSTokenizer tokenizer(rootMargin);
    auto tokenRange = tokenizer.tokenRange();
    tokenRange.consumeWhitespace();
    Vector<Length, 4> margins;
    while (!tokenRange.atEnd()) {
        if (margins.size() == 4) {
            //return Exception { SyntaxError, "Failed to construct 'IntersectionObserver': Extra text found at the end of rootMargin." };
        	ec = SYNTAX_ERR;
        	return LengthBox();
        }
        RefPtr<CSSPrimitiveValue> parsedValue = CSSPropertyParserHelpers::consumeLengthOrPercent(tokenRange, HTMLStandardMode, ValueRangeAll);
        if (!parsedValue || parsedValue->isCalculated()) {
            //return Exception { SyntaxError, "Failed to construct 'IntersectionObserver': rootMargin must be specified in pixels or percent." };
        	ec = SYNTAX_ERR;
			return LengthBox();
		}
		if (parsedValue->isPercentage())
            margins.append(Length(parsedValue->doubleValue(), LengthType::Percent));
        else if (parsedValue->isPx())
            margins.append(Length(parsedValue->intValue(), LengthType::Fixed));
        else {
            //return Exception { SyntaxError, "Failed to construct 'IntersectionObserver': rootMargin must be specified in pixels or percent." };
        	ec = SYNTAX_ERR;
			return LengthBox();
		}
    }
    switch (margins.size()) {
    case 0:
        for (unsigned i = 0; i < 4; ++i)
            margins.append(Length(0, LengthType::Fixed));
        break;
    case 1:
        for (unsigned i = 0; i < 3; ++i)
            margins.append(margins[0]);
        break;
    case 2:
        margins.append(margins[0]);
        margins.append(margins[1]);
        break;
    case 3:
        margins.append(margins[1]);
        break;
    case 4:
        break;
    default:
        ASSERT_NOT_REACHED();
    }

    return LengthBox(margins[0], margins[1], margins[2], margins[3]);
}

PassRefPtr<IntersectionObserver> IntersectionObserver::create(JSC::ExecState& state, Document* document, PassRefPtr<IntersectionObserverCallback> callback, const Init& init, ExceptionCode& ec)
{
    RefPtr<ContainerNode> root;
    if (init.root) {
    	if (init.root->isElementNode())
    		root = toElement(init.root.get());
    	else if (init.root->isDocumentNode())
            root = toDocument(init.root.get());
    	else {
    		ec = TypeError;
    		return nullptr;
    	}
    }

	LengthBox rootMargin = parseRootMargin(init.rootMargin, ec);
    if (ec)
        return nullptr;

    Vector<double> thresholds;
    if (init.threshold && !init.threshold.isUndefined()) {
		if(init.threshold.isObject())
			thresholds = toNativeArray<double>(state, asObject(init.threshold));
		else {
			thresholds.reserveInitialCapacity(1);
			thresholds.uncheckedAppend(convert<double>(state, init.threshold, ShouldAllowNonFiniteNo));
		}
		if (state.exception()) {
    		ec = TypeError;
			return nullptr;
		}
    } else {
		thresholds.reserveInitialCapacity(1);
		thresholds.uncheckedAppend(0.0);
    }

    if (thresholds.isEmpty())
        thresholds.append(0.f);

    for (auto threshold : thresholds) {
        if (!(threshold >= 0 && threshold <= 1)) {
            //return Exception { RangeError, "Failed to construct 'IntersectionObserver': all thresholds must lie in the range [0.0, 1.0]." };
        	ec = RangeError;
        	return nullptr;
        }
    }

    return adoptRef(new IntersectionObserver(document, static_cast<JSIntersectionObserverCallback*>(callback.get()), root.get(), rootMargin, WTFMove(thresholds)));
}

IntersectionObserver::IntersectionObserver(Document* document, PassRefPtr<JSIntersectionObserverCallback> callback, ContainerNode* root, const LengthBox& parsedRootMargin, Vector<double>&& thresholds)
    : m_root(makeWeakPtr(root))
    , m_rootMargin(parsedRootMargin)
	, m_thresholds(WTFMove(thresholds))
    , m_callback(callback)
{
    if (root && root->isDocumentNode()) {
        auto& observerData = toDocument(*root).ensureIntersectionObserverData();
        observerData.observers.append(makeWeakPtr(*this));
    } else if (root) {
        auto& observerData = toElement(*root).ensureIntersectionObserverData();
        observerData.observers.append(makeWeakPtr(this));
    } else if (auto* frame = document->frame())
        m_implicitRootDocument = makeWeakPtr(frame->mainFrame()->document());

    std::sort(m_thresholds.begin(), m_thresholds.end());

    ++gIntersectionObserverCount;
}

IntersectionObserver::~IntersectionObserver()
{
    RefPtr<ContainerNode> root = m_root.get();
    if (root && root->isDocumentNode()) {
    	toDocument(*root).intersectionObserverDataIfExists()->observers.removeFirst(this);
    } else if (root)
		toElement(*root).intersectionObserverDataIfExists()->observers.removeFirst(this);
    disconnect();

    --gIntersectionObserverCount;
}

String IntersectionObserver::rootMargin() const
{
    StringBuilder stringBuilder;
    PhysicalBoxSide sides[4] = { PhysicalBoxSide::TopSide, PhysicalBoxSide::RightSide, PhysicalBoxSide::BottomSide, PhysicalBoxSide::LeftSide };
    for (auto side : sides) {
        auto& length = m_rootMargin.at(side);
        stringBuilder.appendNumber(length.intValue());
        if (length.type() == LengthType::Percent)
            stringBuilder.append('%');
        else
            stringBuilder.appendLiteral("px");
        if (side != PhysicalBoxSide::LeftSide)
            stringBuilder.append(' ');
    }
    return stringBuilder.toString();
}

bool IntersectionObserver::isObserving(const Element& element) const
{
    return m_observationTargets.findMatching([&](const WeakPtr<Element>& target) {
        return target.get() == &element;
    }) != notFound;
}

void IntersectionObserver::observe(Element& target)
{
    if (!trackingDocument() || !m_callback || isObserving(target))
        return;

    target.ensureIntersectionObserverData().registrations.append(IntersectionObserverRegistration { makeWeakPtr(*this), Nullopt });
    bool hadObservationTargets = hasObservationTargets();
    m_observationTargets.append(makeWeakPtr(target));

    // Per the specification, we should dispatch at least one observation for the target. For this reason, we make sure to keep the
    // target alive until this first observation. This, in turn, will keep the IntersectionObserver's JS wrapper alive via
    // isReachableFromOpaqueRoots(), so the callback stays alive.
    m_targetsWaitingForFirstObservation.append(target);

    auto* document = trackingDocument();
    if (!hadObservationTargets)
        document->addIntersectionObserver(*this);
    document->scheduleInitialIntersectionObservationUpdate();
}

void IntersectionObserver::unobserve(Element& target)
{
    if (!removeTargetRegistration(target))
        return;

    bool removed = m_observationTargets.removeFirst(&target);
    ASSERT_UNUSED(removed, removed);
    m_targetsWaitingForFirstObservation.removeFirstMatching([&](const GCReachableRef<Element>& pendingTarget) { return pendingTarget.ptr() == &target; });

    if (!hasObservationTargets()) {
        if (auto* document = trackingDocument())
            document->removeIntersectionObserver(*this);
    }
}

void IntersectionObserver::disconnect()
{
    if (!hasObservationTargets()) {
        ASSERT(m_targetsWaitingForFirstObservation.isEmpty());
        return;
    }

    removeAllTargets();
    if (auto* document = trackingDocument())
        document->removeIntersectionObserver(*this);
}

auto IntersectionObserver::takeRecords() -> TakenRecords
{
    return TakenRecords { WTFMove(m_queuedEntries), WTFMove(m_pendingTargets) };
}

void IntersectionObserver::targetDestroyed(Element& target)
{
    m_observationTargets.removeFirst(&target);
    m_targetsWaitingForFirstObservation.removeFirstMatching([&](const GCReachableRef<Element>& pendingTarget) { return pendingTarget.ptr() == &target; });
	if (!hasObservationTargets()) {
        if (auto* document = trackingDocument())
            document->removeIntersectionObserver(*this);
    }
}

bool IntersectionObserver::removeTargetRegistration(Element& target)
{
    auto* observerData = target.intersectionObserverDataIfExists();
    if (!observerData)
        return false;

    auto& registrations = observerData->registrations;
    return registrations.removeFirstMatching([this](IntersectionObserverRegistration& registration) {
        return registration.observer.get() == this;
    });
}

void IntersectionObserver::removeAllTargets()
{
    for (auto& target : m_observationTargets) {
        bool removed = removeTargetRegistration(*target);
        ASSERT_UNUSED(removed, removed);
    }
    m_observationTargets.clear();
    m_targetsWaitingForFirstObservation.clear();
}

void IntersectionObserver::rootDestroyed()
{
    ASSERT(m_root);
    disconnect();
    m_root = nullptr;
}

Optional<ReducedResolutionSeconds> IntersectionObserver::nowTimestamp() const
{
    if (!m_callback)
        return WTF::Nullopt;

    auto* context = m_callback->scriptExecutionContext();
    if (!context)
        return WTF::Nullopt;

    ASSERT(context->isDocument());
    auto& document = static_cast<const Document&>(*context);
    if (auto* window = document.domWindow())
        return window->frozenNowTimestamp();

    return WTF::Nullopt;
}

void IntersectionObserver::appendQueuedEntry(Ref<IntersectionObserverEntry>&& entry)
{
    ASSERT(entry->target());
    m_pendingTargets.append(*entry->target());
    m_queuedEntries.append(WTFMove(entry));
}

void IntersectionObserver::notify()
{
    if (m_queuedEntries.isEmpty()) {
        ASSERT(m_pendingTargets.isEmpty());
        return;
    }

    auto takenRecords = takeRecords();
    Vector<GCReachableRef<Element>> targetsWaitingForFirstObservation = WTFMove(m_targetsWaitingForFirstObservation);

    // FIXME: The JSIntersectionObserver wrapper should be kept alive as long as the intersection observer can fire events.
    ASSERT(m_callback->hasCallback());
    if (!m_callback->callbackData()->callback())
        return;

    auto* context = m_callback->scriptExecutionContext();
    if (!context)
        return;

    m_callback->handleEvent(takenRecords.records, this);
}

bool IntersectionObserver::isReachableFromOpaqueRoots(JSC::SlotVisitor& visitor) const
{
    for (auto& target : m_observationTargets) {
    	auto* element = target.get();
        if (element && visitor.containsOpaqueRoot(element->opaqueRoot()))
            return true;
    }
    for (auto& target : m_pendingTargets) {
        if (visitor.containsOpaqueRoot(target->opaqueRoot()))
            return true;
    }
    return !m_targetsWaitingForFirstObservation.isEmpty();
}

} // namespace WebCore

#endif // ENABLE(INTERSECTION_OBSERVER)
