/*
 * Copyright (C) 2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#if ENABLE(INTERSECTION_OBSERVER)

#include "ExceptionCode.h"
#include "GCReachableRef.h"
#include "IntersectionObserverCallback.h"
#include "IntersectionObserverEntry.h"
#include "LengthBox.h"
#include "JSValue.h"
#include "ReducedResolutionSeconds.h"
#include <wtf/Optional.h>
#include <wtf/RefCounted.h>
#include <wtf/WeakPtr.h>
//#include <wtf/Variant.h>
#include <wtf/text/WTFString.h>

namespace JSC {

class SlotVisitor;

}

namespace WebCore {

class Document;
class Element;
class ContainerNode;
class JSIntersectionObserverCallback;

struct IntersectionObserverRegistration {
    WeakPtr<IntersectionObserver> observer;
    Optional<size_t> previousThresholdIndex;
};

struct IntersectionObserverData {
    WTF_MAKE_FAST_ALLOCATED;
public:
    // IntersectionObservers for which the node that owns this IntersectionObserverData is the root.
	// An IntersectionObserver without any targets is only owned by JavaScript wrappers. An
	// IntersectionObserver with at least one target is also owned by its trackingDocument.
	Vector<WeakPtr<IntersectionObserver>> observers;

	// IntersectionObserverRegistrations for which the node that owns this IntersectionObserverData is the target.
	Vector<IntersectionObserverRegistration> registrations;
};

struct IntersectionObserverInit {
	RefPtr<Node> /*Optional<Variant<RefPtr<Element>, RefPtr<Document>>>*/ root;
    String rootMargin;
    JSC::JSValue /*Variant<double, Vector<double>>*/ threshold;
};

class IntersectionObserver : public RefCounted<IntersectionObserver>, public CanMakeWeakPtr<IntersectionObserver> {
public:
	typedef IntersectionObserverInit Init;

    static PassRefPtr<IntersectionObserver> create(JSC::ExecState&, Document*, PassRefPtr<IntersectionObserverCallback>, const Init&, ExceptionCode&);

    ~IntersectionObserver();

    Document* trackingDocument() const { return m_root ? &m_root->document() : m_implicitRootDocument.get(); }

    ContainerNode* root() const { return m_root.get(); }
    String rootMargin() const;
    const LengthBox& rootMarginBox() const { return m_rootMargin; }
    const Vector<double>& thresholds() const { return m_thresholds; }
    const Vector<WeakPtr<Element>>& observationTargets() const { return m_observationTargets; }
    bool hasObservationTargets() const { return m_observationTargets.size(); }
	bool isObserving(const Element&) const;

    void observe(Element&);
    void unobserve(Element&);
    void disconnect();

    struct TakenRecords {
        Vector<RefPtr<IntersectionObserverEntry>> records;
        Vector<GCReachableRef<Element>> pendingTargets;
    };
    TakenRecords takeRecords();

    void targetDestroyed(Element&);
    void rootDestroyed();

    Optional<ReducedResolutionSeconds> nowTimestamp() const;

    void appendQueuedEntry(Ref<IntersectionObserverEntry>&&);
    void notify();

    JSIntersectionObserverCallback* callbackConcurrently() { return m_callback.get(); }
    bool isReachableFromOpaqueRoots(JSC::SlotVisitor&) const;

private:
    IntersectionObserver(Document*, PassRefPtr<JSIntersectionObserverCallback>, ContainerNode* root, const LengthBox&, Vector<double>&& thresholds);

    bool removeTargetRegistration(Element&);
    void removeAllTargets();

    WeakPtr<Document> m_implicitRootDocument;
    WeakPtr<ContainerNode> m_root;
    LengthBox m_rootMargin;
    Vector<double> m_thresholds;
    RefPtr<JSIntersectionObserverCallback> m_callback;
    Vector<WeakPtr<Element>> m_observationTargets;
    Vector<GCReachableRef<Element>> m_pendingTargets;
    Vector<RefPtr<IntersectionObserverEntry>> m_queuedEntries;
    Vector<GCReachableRef<Element>> m_targetsWaitingForFirstObservation;
};

} // namespace WebCore

#endif // ENABLE(INTERSECTION_OBSERVER)
