/*
 * Copyright (C) 2008, 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "Location.h"

#include "DOMWindow.h"
#include "Document.h"
#include "ExceptionCode.h"
#include "Frame.h"
#include "FrameLoader.h"
#include "KURL.h"
#include "SecurityOrigin.h"

namespace WebCore {

Location::Location(DOMWindow& window)
    : DOMWindowProperty(&window)
{
}

inline const KURL& Location::url() const
{
    ASSERT(frame());

    const KURL& url = frame()->document()->url();
    if (!url.isValid())
        return blankURL(); // Use "about:blank" while the page is still loading (before we have a frame).

    return url;
}

String Location::href() const
{
    if (!frame())
        return String();

    return url().string();
}

String Location::protocol() const
{
    if (!frame())
        return String();

    return makeString(url().protocol(), ":");
}

String Location::host() const
{
    if (!frame())
        return String();

    // Note: this is the IE spec. The NS spec swaps the two, it says
    // "The hostname property is the concatenation of the host and port properties, separated by a colon."
    return url().hostAndPort();
}

String Location::hostname() const
{
    if (!frame())
        return String();

    return url().host().toString();
}

String Location::port() const
{
    if (!frame())
        return String();

    const KURL& url = this->url();
    return url.port() ? String::number(url.port().value()) : emptyString();
}

String Location::pathname() const
{
    if (!frame())
        return String();

    const KURL& url = this->url();
    return url.path().isEmpty() ? "/" : url.path();
}

String Location::search() const
{
    if (!frame())
        return String();

    const KURL& url = this->url();
    return url.query().isEmpty() ? emptyString() : "?" + url.query();
}

String Location::origin() const
{
    if (!frame())
        return String();
    return SecurityOrigin::create(url())->toString();
}

PassRefPtr<DOMStringList> Location::ancestorOrigins() const
{
    RefPtr<DOMStringList> origins = DOMStringList::create();
    auto* frame = this->frame();
    if (!frame)
        return origins;
    for (auto* ancestor = frame->tree()->parent(); ancestor; ancestor = ancestor->tree()->parent())
        origins->append(ancestor->document()->securityOrigin()->toString());
    return origins;
}

String Location::hash() const
{
    if (!frame())
        return String();

    String fragmentIdentifier = url().fragmentIdentifier();
    return fragmentIdentifier.isEmpty() ? emptyString() : "#" + fragmentIdentifier;
}

void Location::setHref(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& url, ExceptionCode& ec)
{
    if (!frame())
        return;
    setLocation(activeWindow, firstWindow, url, ec);
}

void Location::setProtocol(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& protocol, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    KURL url = frame->document()->url();
    if (!url.setProtocol(protocol)) {
        ec = SYNTAX_ERR;
        return;
    }
    setLocation(activeWindow, firstWindow, url.string(), ec);
}

void Location::setHost(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& host, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    KURL url = frame->document()->url();
    url.setHostAndPort(host);
    setLocation(activeWindow, firstWindow, url.string(), ec);
}

void Location::setHostname(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& hostname, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    KURL url = frame->document()->url();
    url.setHost(hostname);
    setLocation(activeWindow, firstWindow, url.string(), ec);
}

void Location::setPort(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& portString, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    KURL url = frame->document()->url();
    int port = portString.toInt();
    if (port < 0 || port > 0xFFFF || portString.isEmpty())
        url.removePort();
    else
        url.setPort(port);
    setLocation(activeWindow, firstWindow, url.string(), ec);
}

void Location::setPathname(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& pathname, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    KURL url = frame->document()->url();
    url.setPath(pathname);
    setLocation(activeWindow, firstWindow, url.string(), ec);
}

void Location::setSearch(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& search, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    KURL url = frame->document()->url();
    url.setQuery(search);
    setLocation(activeWindow, firstWindow, url.string(), ec);
}

void Location::setHash(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& hash, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    KURL url = frame->document()->url();
    String oldFragmentIdentifier = url.fragmentIdentifier();
    String newFragmentIdentifier = hash;
    if (hash[0] == '#')
        newFragmentIdentifier = hash.substring(1);
    url.setFragmentIdentifier(newFragmentIdentifier);
    // Note that by parsing the URL and *then* comparing fragments, we are 
    // comparing fragments post-canonicalization, and so this handles the 
    // cases where fragment identifiers are ignored or invalid. 
    if (equalIgnoringNullity(oldFragmentIdentifier, url.fragmentIdentifier()))
        return;
    setLocation(activeWindow, firstWindow, url.string(), ec);
}

void Location::assign(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& url, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    setLocation(activeWindow, firstWindow, url, ec);
}

void Location::replace(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& urlString, ExceptionCode& ec)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    ASSERT(frame->document());
    ASSERT(frame->document()->domWindow());

    Frame* firstFrame = firstWindow->frame();
    if (!firstFrame || !firstFrame->document())
        return;

    URL completedURL = firstFrame->document()->completeURL(urlString);
    if (!completedURL.isValid()) {
    	ec = SYNTAX_ERR;
    	return;
    }

    if (!activeWindow->document()->canNavigate(frame, completedURL)) {
    	ec = SECURITY_ERR;
        return;
    }

    // Note: We call DOMWindow::setLocation directly here because replace() always operates on the current frame.
    frame->document()->domWindow()->setLocation(*activeWindow, completedURL, LockHistoryAndBackForwardList);
}

void Location::reload(DOMWindow* activeWindow)
{
    auto* frame = this->frame();
    if (!frame)
        return;
    // FIXME: It's not clear this cross-origin security check is valuable.
    // We allow one page to change the location of another. Why block attempts to reload?
    // Other location operations simply block use of JavaScript URLs cross origin.
    DOMWindow* targetWindow = frame->document()->domWindow();
    if (!activeWindow->securityOrigin()->canAccess(targetWindow->securityOrigin())) {
        targetWindow->printErrorMessage(targetWindow->crossDomainAccessErrorMessage(activeWindow));
        return;
    }
    if (protocolIsJavaScript(frame->document()->url()))
        return;
    frame->navigationScheduler()->scheduleRefresh(*activeWindow->document());
}

void Location::setLocation(DOMWindow* activeWindow, DOMWindow* firstWindow, const String& urlString, ExceptionCode& ec)
{
    auto* frame = this->frame();
    ASSERT(frame);

    Frame* firstFrame = firstWindow->frame();
    if (!firstFrame || !firstFrame->document())
        return;

    KURL completedURL = firstFrame->document()->completeURL(urlString);

    if (!completedURL.isValid()) {
        ec = TypeError;
        return;
    }

    if (!activeWindow->document()->canNavigate(frame, completedURL)) {
    	ec = SECURITY_ERR;
        return;
    }

    ASSERT(frame->document());
    ASSERT(frame->document()->domWindow());
    frame->document()->domWindow()->setLocation(*activeWindow, completedURL);
}

} // namespace WebCore
