/*
 * Copyright (C) 2003, 2006, 2007, 2008, 2009, 2011 Apple Inc. All rights reserved.
 *           (C) 2006 Graham Dennis (graham.dennis@gmail.com)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "EditingBehaviorTypes.h"
#include "FontLoadTimingOverride.h"
#include "TextFlags.h"
#include "KURL.h"
#include "SecurityOrigin.h"
#include "Timer.h"
#include <wtf/HashMap.h>
#include <wtf/RefCounted.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/AtomicStringHash.h>
#include <wtf/unicode/Unicode.h>
#include <wtf/text/StringHash.h>

#if ENABLE(TEXT_AUTOSIZING)
#include "IntSize.h"
#endif

namespace WebCore {

class FontGenericFamilies;
class Page;

enum EditableLinkBehavior : uint8_t {
	EditableLinkDefaultBehavior,
	EditableLinkAlwaysLive,
	EditableLinkOnlyLiveWithShiftKey,
	EditableLinkLiveWhenNotFocused,
	EditableLinkNeverLive
};

enum TextDirectionSubmenuInclusionBehavior : uint8_t {
	TextDirectionSubmenuNeverIncluded,
	TextDirectionSubmenuAutomaticallyIncluded,
	TextDirectionSubmenuAlwaysIncluded
};

class Settings final : public RefCounted<Settings> {
	WTF_MAKE_NONCOPYABLE(Settings); WTF_MAKE_FAST_ALLOCATED;
public:
	static PassRefPtr<Settings> create(Page*);

    ~Settings();

    void pageDestroyed() { m_page = nullptr; }

	void setStandardFontFamily(const AtomicString&, UScriptCode = USCRIPT_COMMON);
	const AtomicString& standardFontFamily(UScriptCode = USCRIPT_COMMON) const;

	void setFixedFontFamily(const AtomicString&, UScriptCode = USCRIPT_COMMON);
	const AtomicString& fixedFontFamily(UScriptCode = USCRIPT_COMMON) const;

	void setSerifFontFamily(const AtomicString&, UScriptCode = USCRIPT_COMMON);
	const AtomicString& serifFontFamily(UScriptCode = USCRIPT_COMMON) const;

	void setSansSerifFontFamily(const AtomicString&, UScriptCode = USCRIPT_COMMON);
	const AtomicString& sansSerifFontFamily(UScriptCode = USCRIPT_COMMON) const;

	void setCursiveFontFamily(const AtomicString&, UScriptCode = USCRIPT_COMMON);
	const AtomicString& cursiveFontFamily(UScriptCode = USCRIPT_COMMON) const;

	void setFantasyFontFamily(const AtomicString&, UScriptCode = USCRIPT_COMMON);
	const AtomicString& fantasyFontFamily(UScriptCode = USCRIPT_COMMON) const;

	void setPictographFontFamily(const AtomicString&, UScriptCode = USCRIPT_COMMON);
	const AtomicString& pictographFontFamily(UScriptCode = USCRIPT_COMMON) const;

	void setMinimumFontSize(int);
	ALWAYS_INLINE int minimumFontSize() const { return m_minimumFontSize; }

	void setMinimumLogicalFontSize(int);
	ALWAYS_INLINE int minimumLogicalFontSize() const { return m_minimumLogicalFontSize; }

	void setDefaultFontSize(int);
	ALWAYS_INLINE int defaultFontSize() const { return m_defaultFontSize; }

	void setDefaultFixedFontSize(int);
	ALWAYS_INLINE int defaultFixedFontSize() const { return m_defaultFixedFontSize; }

#if ENABLE(TEXT_AUTOSIZING)
	void setTextAutosizingEnabled(bool);
	bool textAutosizingEnabled() const { return m_textAutosizingEnabled; }

	// Only set by Layout Tests, and only used if textAutosizingEnabled() returns true.
	void setTextAutosizingWindowSizeOverride(const IntSize&);
	const IntSize& textAutosizingWindowSizeOverride() const { return m_textAutosizingWindowSizeOverride; }
#endif

	// Unlike areImagesEnabled, this only suppresses the network load of
	// the image URL.  A cached image will still be rendered if requested.
	void setLoadsImagesAutomatically(bool);
	ALWAYS_INLINE bool loadsImagesAutomatically() const { return m_loadsImagesAutomatically; }

	// This setting only affects site icon image loading if loadsImagesAutomatically setting is false and this setting is true.
	// All other permutations still heed loadsImagesAutomatically setting.
	void setLoadsSiteIconsIgnoringImageLoadingSetting(bool);
	ALWAYS_INLINE bool loadsSiteIconsIgnoringImageLoadingSetting() const { return m_loadsSiteIconsIgnoringImageLoadingSetting; }

	ALWAYS_INLINE void setScriptEnabled(bool isScriptEnabled) { m_isScriptEnabled = isScriptEnabled; }
	// Instead of calling isScriptEnabled directly, please consider calling
	// ScriptController::canExecuteScripts, which takes things like the
	// HTML sandbox attribute into account.
	ALWAYS_INLINE bool isScriptEnabled() const { return m_isScriptEnabled; }

	void setWebSecurityEnabled(bool);
	ALWAYS_INLINE bool isWebSecurityEnabled() const { return m_isWebSecurityEnabled; }

	void setAllowUniversalAccessFromFileURLs(bool);
	ALWAYS_INLINE bool allowUniversalAccessFromFileURLs() const { return m_allowUniversalAccessFromFileURLs; }

	void setAllowFileAccessFromFileURLs(bool);
	ALWAYS_INLINE bool allowFileAccessFromFileURLs() const { return m_allowFileAccessFromFileURLs; }

	ALWAYS_INLINE void setJavaScriptCanOpenWindowsAutomatically(bool javaScriptCanOpenWindowsAutomatically)
	{
		m_javaScriptCanOpenWindowsAutomatically = javaScriptCanOpenWindowsAutomatically;
	}
	ALWAYS_INLINE bool javaScriptCanOpenWindowsAutomatically() const { return m_javaScriptCanOpenWindowsAutomatically; }

	void setJavaScriptCanAccessClipboard(bool);
	ALWAYS_INLINE bool javaScriptCanAccessClipboard() const { return m_javaScriptCanAccessClipboard; }

	void setSpatialNavigationEnabled(bool);
	ALWAYS_INLINE bool isSpatialNavigationEnabled() const { return m_isSpatialNavigationEnabled; }

	void setJavaEnabled(bool);
	ALWAYS_INLINE bool isJavaEnabled() const { return m_isJavaEnabled; }

	void setInspectorEnabled(bool enabled) { m_inspectorEnabled = enabled; }
	bool inspectorEnabled() const { return m_inspectorEnabled; }

	// This settings is only consulted if isJavaEnabled() returns true;
	void setJavaEnabledForLocalFiles(bool);
	ALWAYS_INLINE bool isJavaEnabledForLocalFiles() const { return m_isJavaEnabledForLocalFiles; }

	void setImagesEnabled(bool);
	ALWAYS_INLINE bool areImagesEnabled() const { return m_areImagesEnabled; }

	void setMediaEnabled(bool);
	ALWAYS_INLINE bool isMediaEnabled() const { return m_isMediaEnabled; }

	ALWAYS_INLINE void setPluginsEnabled(bool arePluginsEnabled) { m_arePluginsEnabled = arePluginsEnabled; }
	ALWAYS_INLINE bool arePluginsEnabled() const { return m_arePluginsEnabled; }

	ALWAYS_INLINE void setLocalStorageEnabled(bool localStorageEnabled) { m_localStorageEnabled = localStorageEnabled; }
	ALWAYS_INLINE bool localStorageEnabled() const { return m_localStorageEnabled; }

	ALWAYS_INLINE void setLocalStorageUntrustedDomainsQuotaPerDomain(unsigned i) { m_localStorageUntrustedDomainsQuotaPerDomain = i; }
	ALWAYS_INLINE unsigned localStorageUntrustedDomainsQuotaPerDomain() const { return m_localStorageUntrustedDomainsQuotaPerDomain; }
	ALWAYS_INLINE void setLocalStorageUntrustedDomainsTotalQuota(unsigned i) { m_localStorageUntrustedDomainsTotalQuota = i; }
	ALWAYS_INLINE unsigned localStorageUntrustedDomainsTotalQuota() const { return m_localStorageUntrustedDomainsTotalQuota; }
	ALWAYS_INLINE void setLocalStorageTrustedDomainsTotalQuota(unsigned i) { m_localStorageTrustedDomainsTotalQuota = i; }
	ALWAYS_INLINE unsigned localStorageTrustedDomainsTotalQuota() const { return m_localStorageTrustedDomainsTotalQuota; }

	ALWAYS_INLINE void clearAllStorageQuota(void) {
		m_localStorageQuota.clear();
	}

	ALWAYS_INLINE void setLocalStorageQuota(const String& url, unsigned localStorageQuota)
	{
		m_localStorageQuota.set(url, localStorageQuota);
	}

	bool localStorageQuota(const String& url, pair<String, unsigned>& res) const;

	// Allow clients concerned with memory consumption to set a quota on session storage
	// since the memory used won't be released until the Page is destroyed.
	// Default is noQuota.
	void setSessionStorageQuota(unsigned);
	ALWAYS_INLINE unsigned sessionStorageQuota() const { return m_sessionStorageQuota; }

	// When this option is set, WebCore will avoid storing any record of browsing activity
	// that may persist on disk or remain displayed when the option is reset.
	// This option does not affect the storage of such information in RAM.
	// The following functions respect this setting:
	//  - HTML5/DOM Storage
	//  - Icon Database
	//  - Console Messages
	//  - MemoryCache
	//  - Application Cache
	//  - Back/Forward Page History
	//  - Page Search Results
	//  - HTTP Cookies
	//  - Plug-ins (that support NPNVprivateModeBool)
	void setPrivateBrowsingEnabled(bool);
	ALWAYS_INLINE bool privateBrowsingEnabled() const { return m_privateBrowsingEnabled; }

	ALWAYS_INLINE void setCaretBrowsingEnabled(bool caretBrowsingEnabled) { m_caretBrowsingEnabled = caretBrowsingEnabled; }
	ALWAYS_INLINE bool caretBrowsingEnabled() const { return m_caretBrowsingEnabled; }

	void setDefaultTextEncodingName(const String&);
	ALWAYS_INLINE const String& defaultTextEncodingName() const { return m_defaultTextEncodingName; }

	void setUsesEncodingDetector(bool);
	ALWAYS_INLINE bool usesEncodingDetector() const { return m_usesEncodingDetector; }

	void setDNSPrefetchingEnabled(bool);
	ALWAYS_INLINE bool dnsPrefetchingEnabled() const { return m_dnsPrefetchingEnabled; }

	void setUserStyleSheetLocation(const KURL&);
	ALWAYS_INLINE const KURL& userStyleSheetLocation() const { return m_userStyleSheetLocation; }

	void setFixedElementsLayoutRelativeToFrame(bool);
	ALWAYS_INLINE bool fixedElementsLayoutRelativeToFrame() const { return m_fixedElementsLayoutRelativeToFrame; }

	void setShouldPrintBackgrounds(bool);
	ALWAYS_INLINE bool shouldPrintBackgrounds() const { return m_shouldPrintBackgrounds; }

	void setTextAreasAreResizable(bool);
	ALWAYS_INLINE bool textAreasAreResizable() const { return m_textAreasAreResizable; }

	void setEditableLinkBehavior(EditableLinkBehavior);
	ALWAYS_INLINE EditableLinkBehavior editableLinkBehavior() const { return m_editableLinkBehavior; }

	void setTextDirectionSubmenuInclusionBehavior(TextDirectionSubmenuInclusionBehavior);
	ALWAYS_INLINE TextDirectionSubmenuInclusionBehavior textDirectionSubmenuInclusionBehavior() const { return m_textDirectionSubmenuInclusionBehavior; }

#if ENABLE(DASHBOARD_SUPPORT)
	void setUsesDashboardBackwardCompatibilityMode(bool);
	bool usesDashboardBackwardCompatibilityMode() const { return m_usesDashboardBackwardCompatibilityMode; }
#endif

	void setNeedsAdobeFrameReloadingQuirk(bool);
	ALWAYS_INLINE bool needsAcrobatFrameReloadingQuirk() const { return m_needsAdobeFrameReloadingQuirk; }

	void setNeedsKeyboardEventDisambiguationQuirks(bool);
	ALWAYS_INLINE bool needsKeyboardEventDisambiguationQuirks() const { return m_needsKeyboardEventDisambiguationQuirks; }

	void setTreatsAnyTextCSSLinkAsStylesheet(bool);
	ALWAYS_INLINE bool treatsAnyTextCSSLinkAsStylesheet() const { return m_treatsAnyTextCSSLinkAsStylesheet; }

	void setNeedsLeopardMailQuirks(bool);
	ALWAYS_INLINE bool needsLeopardMailQuirks() const { return m_needsLeopardMailQuirks; }

	void setDOMPasteAllowed(bool);
	ALWAYS_INLINE bool isDOMPasteAllowed() const { return m_isDOMPasteAllowed; }

	static void setDefaultMinDOMTimerInterval(double); // Interval specified in seconds.
	static double defaultMinDOMTimerInterval() { return gDefaultMinDOMTimerInterval; }

	void setMinDOMTimerInterval(double); // Per-page; initialized to default value.
	double minDOMTimerInterval() const;

	void setUsesPageCache(bool);
	ALWAYS_INLINE bool usesPageCache() const { return m_usesPageCache; }

	ALWAYS_INLINE void setPageCacheSupportsPlugins(bool pageCacheSupportsPlugins) { m_pageCacheSupportsPlugins = pageCacheSupportsPlugins; }
	ALWAYS_INLINE bool pageCacheSupportsPlugins() const { return m_pageCacheSupportsPlugins; }

	void setShouldCacheXHRResource(bool shouldCacheXHRResource) { m_shouldCacheXHRResource = shouldCacheXHRResource; }
	bool shouldCacheXHRResource() const { return m_shouldCacheXHRResource; }

	void setShrinksStandaloneImagesToFit(bool);
	ALWAYS_INLINE bool shrinksStandaloneImagesToFit() const { return m_shrinksStandaloneImagesToFit; }

	void setShowsURLsInToolTips(bool);
	ALWAYS_INLINE bool showsURLsInToolTips() const { return m_showsURLsInToolTips; }

	void setShowsToolTipOverTruncatedText(bool);
	ALWAYS_INLINE bool showsToolTipOverTruncatedText() const { return m_showsToolTipOverTruncatedText; }

	void setFTPDirectoryTemplatePath(const String&);
	ALWAYS_INLINE const String& ftpDirectoryTemplatePath() const { return m_ftpDirectoryTemplatePath; }

	void setForceFTPDirectoryListings(bool);
	ALWAYS_INLINE bool forceFTPDirectoryListings() const { return m_forceFTPDirectoryListings; }

	void setDeveloperExtrasEnabled(bool);
	ALWAYS_INLINE bool developerExtrasEnabled() const { return m_developerExtrasEnabled; }

	void setJavaScriptExperimentsEnabled(bool);
	ALWAYS_INLINE bool javaScriptExperimentsEnabled() const { return m_javaScriptExperimentsEnabled; }

	void setFrameFlatteningEnabled(bool);
	ALWAYS_INLINE bool frameFlatteningEnabled() const { return m_frameFlatteningEnabled; }

	void setAuthorAndUserStylesEnabled(bool);
	ALWAYS_INLINE bool authorAndUserStylesEnabled() const { return m_authorAndUserStylesEnabled; }

	void setFontRenderingMode(FontRenderingMode mode);
	FontRenderingMode fontRenderingMode() const;

    static bool globalConstRedeclarationShouldThrow()
    {
#if PLATFORM(MAC)
        return !MacApplication::isIBooks();
#elif PLATFORM(IOS)
        return !IOSApplication::isIBooks();
#else
        return true;
#endif
    }

	void setNeedsSiteSpecificQuirks(bool);
	ALWAYS_INLINE bool needsSiteSpecificQuirks() const { return m_needsSiteSpecificQuirks; }

#if ENABLE(WEB_ARCHIVE)
	void setWebArchiveDebugModeEnabled(bool);
	bool webArchiveDebugModeEnabled() const { return m_webArchiveDebugModeEnabled; }
#endif

	void setLocalFileContentSniffingEnabled(bool);
	ALWAYS_INLINE bool localFileContentSniffingEnabled() const { return m_localFileContentSniffingEnabled; }

	void setLocalStorageUntrustedDomainsPath(const String&);
	ALWAYS_INLINE const String& localStorageUntrustedDomainsPath() const { return m_localStorageUntrustedDomainsPath; }

	void setLocalStorageTrustedDomainsPath(const String&);
	ALWAYS_INLINE const String& localStorageTrustedDomainsPath() const { return m_localStorageTrustedDomainsPath; }

	void setApplicationChromeMode(bool);
	ALWAYS_INLINE bool inApplicationChromeMode() const { return m_inApplicationChromeMode; }

	void setOfflineWebApplicationCacheEnabled(bool);
	ALWAYS_INLINE bool offlineWebApplicationCacheEnabled() const { return m_offlineWebApplicationCacheEnabled; }

	void setEnforceCSSMIMETypeInNoQuirksMode(bool);
	ALWAYS_INLINE bool enforceCSSMIMETypeInNoQuirksMode() const { return m_enforceCSSMIMETypeInNoQuirksMode; }

	ALWAYS_INLINE void setMaximumDecodedImageSize(size_t size) { m_maximumDecodedImageSize = size; }
	ALWAYS_INLINE size_t maximumDecodedImageSize() const { return m_maximumDecodedImageSize; }

	ALWAYS_INLINE void setAllowScriptsToCloseWindows(bool allowScriptsToCloseWindows) { m_allowScriptsToCloseWindows = allowScriptsToCloseWindows; }
	ALWAYS_INLINE bool allowScriptsToCloseWindows() const { return m_allowScriptsToCloseWindows; }

	ALWAYS_INLINE void setEditingBehaviorType(EditingBehaviorType behavior) { m_editingBehaviorType = behavior; }
	ALWAYS_INLINE EditingBehaviorType editingBehaviorType() const { return static_cast<EditingBehaviorType>(m_editingBehaviorType); }

	void setDownloadableBinaryFontsEnabled(bool);
	ALWAYS_INLINE bool downloadableBinaryFontsEnabled() const { return m_downloadableBinaryFontsEnabled; }

	void setXSSAuditorEnabled(bool);
	ALWAYS_INLINE bool xssAuditorEnabled() const { return m_xssAuditorEnabled; }

	void setCanvasUsesAcceleratedDrawing(bool);
	ALWAYS_INLINE bool canvasUsesAcceleratedDrawing() const { return m_canvasUsesAcceleratedDrawing; }

	ALWAYS_INLINE void setAcceleratedDrawingEnabled(bool enabled) { m_acceleratedDrawingEnabled = enabled; }
	ALWAYS_INLINE bool acceleratedDrawingEnabled() const { return m_acceleratedDrawingEnabled; }

	ALWAYS_INLINE void setAcceleratedFiltersEnabled(bool enabled) { m_acceleratedFiltersEnabled = enabled; }
	ALWAYS_INLINE bool acceleratedFiltersEnabled() const { return m_acceleratedFiltersEnabled; }

	ALWAYS_INLINE void setCSSCustomFilterEnabled(bool enabled) { m_isCSSCustomFilterEnabled = enabled; }
	ALWAYS_INLINE bool isCSSCustomFilterEnabled() const { return m_isCSSCustomFilterEnabled; }

	ALWAYS_INLINE void setCSSFilterEnabled(bool enabled) { m_isCSSFilterEnabled = enabled; }
	ALWAYS_INLINE bool isCSSFilterEnabled() const { return m_isCSSFilterEnabled; }

	ALWAYS_INLINE void setCSSCompositingEnabled(bool enabled) { m_isCSSCompositingEnabled = enabled; }
	ALWAYS_INLINE bool isCSSCompositingEnabled() const { return m_isCSSCompositingEnabled; }

#if ENABLE(CSS_REGIONS)
	void setCSSRegionsEnabled(bool enabled) { m_cssRegionsEnabled = enabled; }
	bool cssRegionsEnabled() const { return m_cssRegionsEnabled; }
#else
	void setCSSRegionsEnabled(bool) { }
	bool cssRegionsEnabled() const { return false; }
#endif

	ALWAYS_INLINE void setRegionBasedColumnsEnabled(bool enabled) { m_regionBasedColumnsEnabled = enabled; }
	ALWAYS_INLINE bool regionBasedColumnsEnabled() const { return m_regionBasedColumnsEnabled; }

	void setAcceleratedCompositingEnabled(bool);
	ALWAYS_INLINE bool acceleratedCompositingEnabled() const { return m_acceleratedCompositingEnabled; }

	void setAcceleratedCompositingFor3DTransformsEnabled(bool);
	ALWAYS_INLINE bool acceleratedCompositingFor3DTransformsEnabled() const { return m_acceleratedCompositingFor3DTransformsEnabled; }

	void setAcceleratedCompositingForVideoEnabled(bool);
	ALWAYS_INLINE bool acceleratedCompositingForVideoEnabled() const { return m_acceleratedCompositingForVideoEnabled; }

	void setAcceleratedCompositingForPluginsEnabled(bool);
	ALWAYS_INLINE bool acceleratedCompositingForPluginsEnabled() const { return m_acceleratedCompositingForPluginsEnabled; }

	void setAcceleratedCompositingForCanvasEnabled(bool);
	ALWAYS_INLINE bool acceleratedCompositingForCanvasEnabled() const { return m_acceleratedCompositingForCanvasEnabled; }

	void setAcceleratedCompositingForAnimationEnabled(bool);
	ALWAYS_INLINE bool acceleratedCompositingForAnimationEnabled() const { return m_acceleratedCompositingForAnimationEnabled; }

	ALWAYS_INLINE void setAcceleratedCompositingForFixedPositionEnabled(bool enabled) { m_acceleratedCompositingForFixedPositionEnabled = enabled; }
	ALWAYS_INLINE bool acceleratedCompositingForFixedPositionEnabled() const { return m_acceleratedCompositingForFixedPositionEnabled; }

	ALWAYS_INLINE void setAcceleratedCompositingForScrollableFramesEnabled(bool enabled) { m_acceleratedCompositingForScrollableFramesEnabled = enabled; }
	ALWAYS_INLINE bool acceleratedCompositingForScrollableFramesEnabled() const { return m_acceleratedCompositingForScrollableFramesEnabled; }

	void setShowDebugBorders(bool);
	ALWAYS_INLINE bool showDebugBorders() const { return m_showDebugBorders; }

	void setShowRepaintCounter(bool);
	ALWAYS_INLINE bool showRepaintCounter() const { return m_showRepaintCounter; }

	void setExperimentalNotificationsEnabled(bool);
	ALWAYS_INLINE bool experimentalNotificationsEnabled() const { return m_experimentalNotificationsEnabled; }

#if PLATFORM(WIN) || (OS(WINDOWS) && PLATFORM(WX))
	static void setShouldUseHighResolutionTimers(bool);
	static bool shouldUseHighResolutionTimers() { return gShouldUseHighResolutionTimers; }
#endif

	ALWAYS_INLINE void setWebAudioEnabled(bool enabled) { m_webAudioEnabled = enabled; }
	ALWAYS_INLINE bool webAudioEnabled() const { return m_webAudioEnabled; }

	ALWAYS_INLINE void setWorkerEnabled(bool enabled) { m_workerEnabled = enabled; }
	ALWAYS_INLINE bool workerEnabled() const { return m_workerEnabled; }

	ALWAYS_INLINE void setMediaKeysEnabled(bool enabled) { m_mediaKeysEnabled = enabled; }
	ALWAYS_INLINE bool mediaKeysEnabled() const { return m_mediaKeysEnabled; }

	ALWAYS_INLINE void setWebKitMediaKeysEnabled(bool enabled) { m_webkitMediaKeysEnabled = enabled; }
	ALWAYS_INLINE bool webKitMediaKeysEnabled() const { return m_webkitMediaKeysEnabled; }

	ALWAYS_INLINE void setWebGLEnabled(bool enabled) { m_webGLEnabled = enabled; }
	ALWAYS_INLINE bool webGLEnabled() const { return m_webGLEnabled; }

	void setWebGLErrorsToConsoleEnabled(bool);
	ALWAYS_INLINE bool webGLErrorsToConsoleEnabled() const { return m_webGLErrorsToConsoleEnabled; }

	ALWAYS_INLINE void setOpenGLMultisamplingEnabled(bool enabled) { m_openGLMultisamplingEnabled = enabled; }
	ALWAYS_INLINE bool openGLMultisamplingEnabled() const { return m_openGLMultisamplingEnabled; }

	void setPrivilegedWebGLExtensionsEnabled(bool);
	ALWAYS_INLINE bool privilegedWebGLExtensionsEnabled() const { return m_privilegedWebGLExtensionsEnabled; }

	ALWAYS_INLINE void setAccelerated2dCanvasEnabled(bool enabled) { m_acceleratedCanvas2dEnabled = enabled; }
	ALWAYS_INLINE bool accelerated2dCanvasEnabled() const { return m_acceleratedCanvas2dEnabled; }

	void setDeferred2dCanvasEnabled(bool);
	ALWAYS_INLINE bool deferred2dCanvasEnabled() const { return m_deferredCanvas2dEnabled; }

	// Number of pixels below which 2D canvas is rendered in software
	// even if hardware acceleration is enabled.
	// Hardware acceleration is useful for large canvases where it can avoid the
	// pixel bandwidth between the CPU and GPU. But GPU acceleration comes at
	// a price - extra back-buffer and texture copy. Small canvases are also
	// widely used for stylized fonts. Anti-aliasing text in hardware at that
	// scale is generally slower. So below a certain size it is better to
	// draw canvas in software.
	void setMinimumAccelerated2dCanvasSize(int);
	ALWAYS_INLINE int minimumAccelerated2dCanvasSize() const { return m_minimumAccelerated2dCanvasSize; }

	void setForceMemorySurface2dCanvasSize(int size) { m_forceMemorySurface2dCanvasSize = size; }
	int forceMemorySurface2dCanvasSize() const { return m_forceMemorySurface2dCanvasSize; }

	void setLoadDeferringEnabled(bool);
	ALWAYS_INLINE bool loadDeferringEnabled() const { return m_loadDeferringEnabled; }

	void setTiledBackingStoreEnabled(bool);
	ALWAYS_INLINE bool tiledBackingStoreEnabled() const { return m_tiledBackingStoreEnabled; }

	void setPaginateDuringLayoutEnabled(bool flag) { m_paginateDuringLayoutEnabled = flag; }
	ALWAYS_INLINE bool paginateDuringLayoutEnabled() const { return m_paginateDuringLayoutEnabled; }

#if ENABLE(FULLSCREEN_API)
	ALWAYS_INLINE void setFullScreenEnabled(bool flag) { m_fullScreenAPIEnabled = flag; }
	ALWAYS_INLINE bool fullScreenEnabled() const  { return m_fullScreenAPIEnabled; }
#endif

#if USE(AVFOUNDATION)
	static void setAVFoundationEnabled(bool flag) { gAVFoundationEnabled = flag; }
	static bool isAVFoundationEnabled() { return gAVFoundationEnabled; }
#endif

	ALWAYS_INLINE void setAsynchronousSpellCheckingEnabled(bool flag) { m_asynchronousSpellCheckingEnabled = flag; }
	ALWAYS_INLINE bool asynchronousSpellCheckingEnabled() const  { return m_asynchronousSpellCheckingEnabled; }

	ALWAYS_INLINE void setUnifiedTextCheckerEnabled(bool flag) { m_unifiedTextCheckerEnabled = flag; }
	ALWAYS_INLINE bool unifiedTextCheckerEnabled() const { return m_unifiedTextCheckerEnabled; }

	ALWAYS_INLINE void setMemoryInfoEnabled(bool flag) { m_memoryInfoEnabled = flag; }
	ALWAYS_INLINE bool memoryInfoEnabled() const { return m_memoryInfoEnabled; }

    // This feature requires an implementation of ValidationMessageClient.
	ALWAYS_INLINE void setInteractiveFormValidationEnabled(bool flag) { m_interactiveFormValidation = flag; }
	ALWAYS_INLINE bool interactiveFormValidationEnabled() const { return m_interactiveFormValidation; }

	// Sets the maginication value for validation message timer.
	// If the maginication value is N, a validation message disappears
	// automatically after <message length> * N / 1000 seconds. If N is
	// equal to or less than 0, a validation message doesn't disappears
	// automaticaly. The default value is 50.
	ALWAYS_INLINE void setValidationMessageTimerMagnification(int newValue) { m_validationMessageTimerMagnification = newValue; }
	ALWAYS_INLINE int validationMessageTimerMaginification() const { return m_validationMessageTimerMagnification; }

	ALWAYS_INLINE void setUsePreHTML5ParserQuirks(bool flag) { m_usePreHTML5ParserQuirks = flag; }
	ALWAYS_INLINE bool usePreHTML5ParserQuirks() const { return m_usePreHTML5ParserQuirks; }

	static const unsigned defaultMaximumHTMLParserDOMTreeDepth = 512;
	ALWAYS_INLINE void setMaximumHTMLParserDOMTreeDepth(unsigned maximumHTMLParserDOMTreeDepth) { m_maximumHTMLParserDOMTreeDepth = maximumHTMLParserDOMTreeDepth; }
	ALWAYS_INLINE unsigned maximumHTMLParserDOMTreeDepth() const { return m_maximumHTMLParserDOMTreeDepth; }

	ALWAYS_INLINE void setHyperlinkAuditingEnabled(bool flag) { m_hyperlinkAuditingEnabled = flag; }
	ALWAYS_INLINE bool hyperlinkAuditingEnabled() const { return m_hyperlinkAuditingEnabled; }

	ALWAYS_INLINE void setCrossOriginCheckInGetMatchedCSSRulesDisabled(bool flag) { m_crossOriginCheckInGetMatchedCSSRulesDisabled = flag; }
	ALWAYS_INLINE bool crossOriginCheckInGetMatchedCSSRulesDisabled() const { return m_crossOriginCheckInGetMatchedCSSRulesDisabled; }

	ALWAYS_INLINE void setLayoutFallbackWidth(int width) { m_layoutFallbackWidth = width; }
	ALWAYS_INLINE int layoutFallbackWidth() const { return m_layoutFallbackWidth; }

	ALWAYS_INLINE void setDeviceWidth(int width) { m_deviceWidth = width; }
	ALWAYS_INLINE int deviceWidth() const { return m_deviceWidth; }

	ALWAYS_INLINE void setDeviceHeight(int height) { m_deviceHeight = height; }
	ALWAYS_INLINE int deviceHeight() const { return m_deviceHeight; }

	ALWAYS_INLINE void setForceCompositingMode(bool flag) { m_forceCompositingMode = flag; }
	ALWAYS_INLINE bool forceCompositingMode() const { return m_forceCompositingMode; }

	ALWAYS_INLINE void setShouldInjectUserScriptsInInitialEmptyDocument(bool flag) { m_shouldInjectUserScriptsInInitialEmptyDocument = flag; }
	ALWAYS_INLINE bool shouldInjectUserScriptsInInitialEmptyDocument() { return m_shouldInjectUserScriptsInInitialEmptyDocument; }

	ALWAYS_INLINE void setAllowDisplayOfInsecureContent(bool flag) { m_allowDisplayOfInsecureContent = flag; }
	ALWAYS_INLINE bool allowDisplayOfInsecureContent() const { return m_allowDisplayOfInsecureContent; }
	ALWAYS_INLINE void setAllowRunningOfInsecureContent(bool flag) { m_allowRunningOfInsecureContent = flag; }
	ALWAYS_INLINE bool allowRunningOfInsecureContent() const { return m_allowRunningOfInsecureContent; }

#if ENABLE(SMOOTH_SCROLLING)
	void setEnableScrollAnimator(bool flag) { m_scrollAnimatorEnabled = flag; }
	bool scrollAnimatorEnabled() const { return m_scrollAnimatorEnabled; }
#endif

	ALWAYS_INLINE void setMediaPlaybackRequiresUserGesture(bool flag) { m_mediaPlaybackRequiresUserGesture = flag; };
	ALWAYS_INLINE bool mediaPlaybackRequiresUserGesture() const { return m_mediaPlaybackRequiresUserGesture; }

	ALWAYS_INLINE void setMediaControllerEnabled(bool flag) { m_mediaControllerEnabled = flag; };
	ALWAYS_INLINE bool mediaControllerEnabled() const { return m_mediaControllerEnabled; }

	ALWAYS_INLINE void setMediaPlaybackAllowsInline(bool flag) { m_mediaPlaybackAllowsInline = flag; };
	ALWAYS_INLINE bool mediaPlaybackAllowsInline() const { return m_mediaPlaybackAllowsInline; }

	ALWAYS_INLINE void setPasswordEchoEnabled(bool flag) { m_passwordEchoEnabled = flag; }
	ALWAYS_INLINE bool passwordEchoEnabled() const { return m_passwordEchoEnabled; }

	ALWAYS_INLINE void setSVGFontsEnabled(bool flag) { m_svgFontsEnabled = flag; }
	ALWAYS_INLINE bool svgFontsEnabled() const { return m_svgFontsEnabled; }

	ALWAYS_INLINE void setSuppressesIncrementalRendering(bool flag) { m_suppressesIncrementalRendering = flag; }
	ALWAYS_INLINE bool suppressesIncrementalRendering() const { return m_suppressesIncrementalRendering; }

	ALWAYS_INLINE void setBackspaceKeyNavigationEnabled(bool flag) { m_backspaceKeyNavigationEnabled = flag; }
	ALWAYS_INLINE bool backspaceKeyNavigationEnabled() const { return m_backspaceKeyNavigationEnabled; }

	ALWAYS_INLINE void setPasswordEchoDurationInSeconds(double durationInSeconds) { m_passwordEchoDurationInSeconds = durationInSeconds; }
	ALWAYS_INLINE double passwordEchoDurationInSeconds() const { return m_passwordEchoDurationInSeconds; }

#if USE(SAFARI_THEME)
	// Windows debugging pref (global) for switching between the Aqua look and a native windows look.
	static void setShouldPaintNativeControls(bool);
	static bool shouldPaintNativeControls() { return gShouldPaintNativeControls; }
#endif

	static void setMockScrollbarsEnabled(bool flag);
	static bool mockScrollbarsEnabled();

#if ENABLE(VIDEO_TRACK)
	void setShouldDisplaySubtitles(bool flag) { m_shouldDisplaySubtitles = flag; }
	bool shouldDisplaySubtitles() const { return m_shouldDisplaySubtitles; }

	void setShouldDisplayCaptions(bool flag) { m_shouldDisplayCaptions = flag; }
	bool shouldDisplayCaptions() const { return m_shouldDisplayCaptions; }

	void setShouldDisplayTextDescriptions(bool flag) { m_shouldDisplayTextDescriptions = flag; }
	bool shouldDisplayTextDescriptions() const { return m_shouldDisplayTextDescriptions; }
#endif

	ALWAYS_INLINE void setScrollingCoordinatorEnabled(bool enabled) { m_scrollingCoordinatorEnabled = enabled; }
	ALWAYS_INLINE bool scrollingCoordinatorEnabled() const { return m_scrollingCoordinatorEnabled; }

	ALWAYS_INLINE void setNotificationsEnabled(bool enabled) { m_notificationsEnabled = enabled; }
	ALWAYS_INLINE bool notificationsEnabled() const { return m_notificationsEnabled; }

	// Some apps needs isLoadingInAPISense to account for active subresource loaders.
	ALWAYS_INLINE void setNeedsIsLoadingInAPISenseQuirk(bool enabled) { m_needsIsLoadingInAPISenseQuirk = enabled; }
	ALWAYS_INLINE bool needsIsLoadingInAPISenseQuirk() const { return m_needsIsLoadingInAPISenseQuirk; }

#if ENABLE(TOUCH_EVENTS)
	void setTouchEventEmulationEnabled(bool enabled) { m_touchEventEmulationEnabled = enabled; }
	bool isTouchEventEmulationEnabled() const { return m_touchEventEmulationEnabled; }
#endif

	ALWAYS_INLINE void setShouldRespectImageOrientation(bool enabled) { m_shouldRespectImageOrientation = enabled; }
	ALWAYS_INLINE bool shouldRespectImageOrientation() const { return m_shouldRespectImageOrientation; }

	ALWAYS_INLINE void setWantsBalancedSetDefersLoadingBehavior(bool flag) { m_wantsBalancedSetDefersLoadingBehavior = flag; }
	ALWAYS_INLINE bool wantsBalancedSetDefersLoadingBehavior() const { return m_wantsBalancedSetDefersLoadingBehavior; }

	ALWAYS_INLINE void setIncrementalRenderingSuppressionTimeoutInSeconds(double timeout) { m_incrementalRenderingSuppressionTimeoutInSeconds = timeout; }
	ALWAYS_INLINE double incrementalRenderingSuppressionTimeoutInSeconds() const { return m_incrementalRenderingSuppressionTimeoutInSeconds; }

	ALWAYS_INLINE void setRequestAnimationFrameEnabled(bool enabled) { m_requestAnimationFrameEnabled = enabled; }
	ALWAYS_INLINE bool requestAnimationFrameEnabled() const { return m_requestAnimationFrameEnabled; }

	ALWAYS_INLINE void setDeviceSupportsTouch(bool enabled) { m_deviceSupportsTouch = enabled; }
	ALWAYS_INLINE bool deviceSupportsTouch() const { return m_deviceSupportsTouch; }

	ALWAYS_INLINE void setDeviceSupportsMouse(bool enabled) { m_deviceSupportsMouse = enabled; }
	ALWAYS_INLINE bool deviceSupportsMouse() const { return m_deviceSupportsMouse; }

	ALWAYS_INLINE void setNeedsDidFinishLoadOrderQuirk(bool needsQuirk) { m_needsDidFinishLoadOrderQuirk = needsQuirk; }
	ALWAYS_INLINE bool needsDidFinishLoadOrderQuirk() const { return m_needsDidFinishLoadOrderQuirk; }

	ALWAYS_INLINE void setFixedPositionCreatesStackingContext(bool creates) { m_fixedPositionCreatesStackingContext = creates; }
	ALWAYS_INLINE bool fixedPositionCreatesStackingContext() const { return m_fixedPositionCreatesStackingContext; }

	ALWAYS_INLINE void setSyncXHRInDocumentsEnabled(bool enabled) { m_syncXHRInDocumentsEnabled = enabled; }
	ALWAYS_INLINE bool syncXHRInDocumentsEnabled() const { return m_syncXHRInDocumentsEnabled; }

	// When enabled, window.blur() does not change focus, and
	// window.focus() only changes focus when invoked from the context that
	// created the window.
	ALWAYS_INLINE void setWindowFocusRestricted(bool restricted) { m_windowFocusRestricted = restricted; }
	ALWAYS_INLINE bool windowFocusRestricted() const { return m_windowFocusRestricted; }

    void setStorageBlockingPolicy(SecurityOrigin::StorageBlockingPolicy policy) { m_storageBlockingPolicy = policy; }
    SecurityOrigin::StorageBlockingPolicy storageBlockingPolicy() const { return m_storageBlockingPolicy; }

#if USE(JSC)
	static void setShouldRespectPriorityInCSSAttributeSetters(bool);
	static bool shouldRespectPriorityInCSSAttributeSetters();
#endif

	static const String& platformResourcePath() { return m_platformResourcePath; }
	static void setPlatformResourcePath(const String&);

	ALWAYS_INLINE void setCookieEnabled(bool enabled) { m_cookieEnabled = enabled; }
	ALWAYS_INLINE bool cookieEnabled() const { return m_cookieEnabled; }

	void setThirdPartyIframeRedirectBlockingEnabled(bool enabled) { m_thirdPartyIframeRedirectBlockingEnabled = enabled; }
	bool thirdPartyIframeRedirectBlockingEnabled() const { return m_thirdPartyIframeRedirectBlockingEnabled; }

	ALWAYS_INLINE void setDrawImageRectEnabled(bool enabled) { m_drawImageRectEnabled = enabled; }
	ALWAYS_INLINE bool drawImageRectEnabled() const { return m_drawImageRectEnabled; }

	ALWAYS_INLINE void setDrawBrokenImageEnabled(bool enabled) { m_drawBrokenImageEnabled = enabled; }
	ALWAYS_INLINE bool drawBrokenImageEnabled() const { return m_drawBrokenImageEnabled; }

	ALWAYS_INLINE void setDiagnosticLoggingEnabled(bool enabled) { m_diagnosticLoggingEnabled = enabled; }
	ALWAYS_INLINE bool diagnosticLoggingEnabled() const { return m_diagnosticLoggingEnabled; }

    void setApplyPageScaleFactorInCompositor(bool enabled) { m_applyPageScaleFactorInCompositor = enabled; }
    bool applyPageScaleFactorInCompositor() const { return m_applyPageScaleFactorInCompositor; }

	const String& defaultVideoPosterURL() const { return m_defaultVideoPosterURL; }
	void setDefaultVideoPosterURL(const String& url) { m_defaultVideoPosterURL = url; }

	// Some ports (e.g. iOS) might choose to display attachments inline, regardless of whether the response includes the
	// HTTP header "Content-Disposition: attachment". This setting enables a sandbox around these attachments. The sandbox
	// enforces all frame sandbox flags (see enum SandboxFlag in SecurityContext.h), and also disables <meta http-equiv>
	// processing and subframe loading.
	ALWAYS_INLINE void setContentDispositionAttachmentSandboxEnabled(bool enabled) { m_contentDispositionAttachmentSandboxEnabled = enabled; }
	ALWAYS_INLINE bool contentDispositionAttachmentSandboxEnabled() const { return m_contentDispositionAttachmentSandboxEnabled; }

	ALWAYS_INLINE void setHttpEquivEnabled(bool enabled) { m_httpEquivEnabled = enabled; }
	ALWAYS_INLINE bool httpEquivEnabled() const { return m_httpEquivEnabled; }

	bool debugTiming() const { return m_debugTiming; }
	void setDebugTiming(bool debug) { m_debugTiming = debug; }

	bool fixedBackgroundsPaintRelativeToDocument() const { return m_fixedBackgroundsPaintRelativeToDocument; }
	void setFixedBackgroundsPaintRelativeToDocument(bool b) { m_fixedBackgroundsPaintRelativeToDocument = b; }

    void setFontFallbackPrefersPictographs(bool);
    bool fontFallbackPrefersPictographs() const { return m_fontFallbackPrefersPictographs; }

    bool subresourceIntegrityEnabled() const { return m_subresourceIntegrityEnabled; }
    void setSubresourceIntegrityEnabled(bool enabled) { m_subresourceIntegrityEnabled = enabled; }

    bool displayContentsEnabled() const { return m_displayContentsEnabled; }
    void setDisplayContentsEnabled(bool enabled) { m_displayContentsEnabled = enabled; }

    bool needsFrameNameFallbackToIdQuirk() const { return m_needsFrameNameFallbackToIdQuirk; }
    void setNeedsFrameNameFallbackToIdQuirk(bool b) { m_needsFrameNameFallbackToIdQuirk = b; }

    bool springTimingFunctionEnabled() const { return m_springTimingFunctionEnabled; }
    void setSpringTimingFunctionEnabled(bool b) { m_springTimingFunctionEnabled = b; }

    bool conicGradientsEnabled() const { return m_conicGradientsEnabled; }
    void setConicGradientsEnabled(bool b) { m_conicGradientsEnabled = b; }

    bool hasPseudoClassEnabled() const { return m_hasPseudoClassEnabled; }
    void setHasPseudoClassEnabled(bool b) { m_hasPseudoClassEnabled = b; }

    bool mediaCaptureRequiresSecureConnection() const { return m_mediaCaptureRequiresSecureConnection; }
    void setMediaCaptureRequiresSecureConnection(bool b) { m_mediaCaptureRequiresSecureConnection = b; }

    bool focusVisibleEnabled() const { return m_focusVisibleEnabled; }
    void setFocusVisibleEnabled(bool b) { m_focusVisibleEnabled = b; }

    bool aspectRatioEnabled() const { return m_aspectRatioEnabled; }
    void setAspectRatioEnabled(bool b) { m_aspectRatioEnabled = b; }

    bool shouldIgnoreFontLoadCompletions() const { return m_shouldIgnoreFontLoadCompletions; }
    void setShouldIgnoreFontLoadCompletions(bool b) { m_shouldIgnoreFontLoadCompletions = b; }

    FontLoadTimingOverride fontLoadTimingOverride() const { return m_fontLoadTimingOverride; }
    void setFontLoadTimingOverride(FontLoadTimingOverride fontLoadTimingOverride) { m_fontLoadTimingOverride = fontLoadTimingOverride; }

    bool inertAttributeEnabled() const { return m_inertAttributeEnabled; }
    void setInertAttributeEnabled(bool enabled) { m_inertAttributeEnabled = enabled; }

    bool cssIndividualTransformPropertiesEnabled() const { return m_cssIndividualTransformPropertiesEnabled; }
    void setCSSIndividualTransformPropertiesEnabled(bool enabled) { m_cssIndividualTransformPropertiesEnabled = enabled; }

    bool aspectRatioOfImgFromWidthAndHeightEnabled() const { return m_aspectRatioOfImgFromWidthAndHeightEnabled; }
    void setAspectRatioOfImgFromWidthAndHeightEnabled(bool enabled) { m_aspectRatioOfImgFromWidthAndHeightEnabled = enabled; }

    bool useImageDocumentForSubframePDF() const { return m_useImageDocumentForSubframePDF; }
    void setUseImageDocumentForSubframePDF(bool enabled) { m_useImageDocumentForSubframePDF = enabled; }

    bool unsafePluginPastingEnabled() const { return m_unsafePluginPastingEnabled; }
    void setUnsafePluginPastingEnabled(bool enabled) { m_unsafePluginPastingEnabled = enabled; }

    const Vector<String>& arphicMobileFontNames() const { return m_arphicMobileFontNames; }
    void setArphicMobileFontNames(const Vector<String>& names) { m_arphicMobileFontNames = names; }

    bool dirPseudoEnabled() const { return m_dirPseudoEnabled; }
    void setDirPseudoEnabled(bool enabled) { m_dirPseudoEnabled = enabled; }

    bool css3DTransformInteroperabilityEnabled() const { return m_css3DTransformInteroperabilityEnabled; }
    void setCSS3DTransformInteroperabilityEnabled(bool enabled) { m_css3DTransformInteroperabilityEnabled = enabled; }

    bool webAnimationsCompositeOperationsEnabled() const { return m_webAnimationsCompositeOperationsEnabled; }
    void setWebAnimationsCompositeOperationsEnabled(bool enabled) { m_webAnimationsCompositeOperationsEnabled = enabled; }

    void setBlankAnchorTargetImpliesNoOpenerEnabled(bool isEnabled) { m_blankAnchorTargetImpliesNoOpenerEnabled = isEnabled; }
    bool blankAnchorTargetImpliesNoOpenerEnabled() const { return m_blankAnchorTargetImpliesNoOpenerEnabled; }

    void setReferrerPolicyAttributeEnabled(bool enabled) { m_referrerPolicyAttributeEnabled = enabled; }
    bool referrerPolicyAttributeEnabled() const { return m_referrerPolicyAttributeEnabled; }

    bool linkPreloadResponsiveImagesEnabled() const { return m_linkPreloadResponsiveImagesEnabled; }
    void setLinkPreloadResponsiveImagesEnabled(bool isEnabled) { m_linkPreloadResponsiveImagesEnabled = isEnabled; }

    void setConstantPropertiesEnabled(bool enabled) { m_constantPropertiesEnabled = enabled; }
    bool constantPropertiesEnabled() const { return m_constantPropertiesEnabled; }

    void setShouldAllowUserInstalledFonts(bool b) { m_shouldAllowUserInstalledFonts = b; }
    bool shouldAllowUserInstalledFonts() const { return m_shouldAllowUserInstalledFonts; }

    void setCSSOMViewScrollingAPIEnabled(bool b) { m_CSSOMViewScrollingAPIEnabled = b; }
    bool CSSOMViewScrollingAPIEnabled() const { return m_CSSOMViewScrollingAPIEnabled; }

    void setOverflowClipEnabled(bool b) { m_overflowClipEnabled = b; }
    bool overflowClipEnabled() const { return m_overflowClipEnabled; }

    void setIntersectionObserverEnabled(bool b) { m_intersectionObserverEnabled = b; }
    bool intersectionObserverEnabled() const { return m_intersectionObserverEnabled; }

    void setResizeObserverEnabled(bool b) { m_resizeObserverEnabled = b; }
    bool resizeObserverEnabled() const { return m_resizeObserverEnabled; }

    void setInputEventsEnabled(bool b) { m_inputEventsEnabled = b; }
    bool inputEventsEnabled() const { return m_inputEventsEnabled; }

    void setVisualViewportEnabled(bool b) { m_visualViewportEnabled = b; }
    bool visualViewportEnabled() const { return m_visualViewportEnabled; }

    void setClientCoordinatesRelativeToLayoutViewport(bool b) { m_clientCoordinatesRelativeToLayoutViewport = b; }
    bool clientCoordinatesRelativeToLayoutViewport() const { return m_clientCoordinatesRelativeToLayoutViewport; }

    void setDateTimeInputsEditableComponentsEnabled(bool b) { m_dateTimeInputsEditableComponentsEnabled = b; }
    bool dateTimeInputsEditableComponentsEnabled() const { return m_dateTimeInputsEditableComponentsEnabled; }

    void setLiveRangeSelectionEnabled(bool b) { m_liveRangeSelectionEnabled = b; }
    bool liveRangeSelectionEnabled() const { return m_liveRangeSelectionEnabled; }

    void setSelectionAcrossShadowBoundariesEnabled(bool b) { m_selectionAcrossShadowBoundariesEnabled = b; }
    bool selectionAcrossShadowBoundariesEnabled() const { return m_selectionAcrossShadowBoundariesEnabled; }

    void setSelectionIncludesAltImageText(bool b) { m_selectionIncludesAltImageText = b; }
    bool selectionIncludesAltImageText() const { return m_selectionIncludesAltImageText; }

    void setLogsPageMessagesToSystemConsoleEnabled(bool b) { m_logsPageMessagesToSystemConsoleEnabled = b; }
    bool logsPageMessagesToSystemConsoleEnabled() const { return m_logsPageMessagesToSystemConsoleEnabled; }

    void setMultipartXMixedReplaceImageEnabled(bool b) { m_multipartXMixedReplaceImageEnabled = b; }
    bool multipartXMixedReplaceImageEnabled() const { return m_multipartXMixedReplaceImageEnabled; }

    void setSendMouseEventsToDisabledFormControlsEnabled(bool b) { m_sendMouseEventsToDisabledFormControlsEnabled = b; }
    bool sendMouseEventsToDisabledFormControlsEnabled() const { return m_sendMouseEventsToDisabledFormControlsEnabled; }

    void setImageBitmapEnabled(bool b) { m_imageBitmapEnabled = b; }
    bool imageBitmapEnabled() const { return m_imageBitmapEnabled; }

    void setShouldRestrictBaseURLSchemes(bool b) { m_shouldRestrictBaseURLSchemes = b; }
    bool shouldRestrictBaseURLSchemes() const { return m_shouldRestrictBaseURLSchemes; }

    void setDefaultMaximumRenderTreeDepth(unsigned depth) { m_defaultMaximumRenderTreeDepth = depth; }
    unsigned defaultMaximumRenderTreeDepth() const { return m_defaultMaximumRenderTreeDepth; }

    void setUseAnonymousModeWhenFetchingMaskImages(bool use) { m_useAnonymousModeWhenFetchingMaskImages = use; }
    bool useAnonymousModeWhenFetchingMaskImages() const { return m_useAnonymousModeWhenFetchingMaskImages; }

private:
	explicit Settings(Page*);

	void initializeDefaultFontFamilies();

	Page* m_page { nullptr };

	String m_defaultTextEncodingName;
	String m_ftpDirectoryTemplatePath;
	String m_localStorageUntrustedDomainsPath;
	String m_localStorageTrustedDomainsPath;

	static String m_platformResourcePath;

	KURL m_userStyleSheetLocation;
    OwnPtr<FontGenericFamilies> m_fontGenericFamilies;
	EditableLinkBehavior m_editableLinkBehavior;
	TextDirectionSubmenuInclusionBehavior m_textDirectionSubmenuInclusionBehavior;
	double m_passwordEchoDurationInSeconds;
	int m_minimumFontSize;
	int m_minimumLogicalFontSize;
	int m_defaultFontSize;
	int m_defaultFixedFontSize;
	int m_validationMessageTimerMagnification;
	int m_minimumAccelerated2dCanvasSize;
	int m_forceMemorySurface2dCanvasSize;
	int m_layoutFallbackWidth;
	size_t m_maximumDecodedImageSize;
	int m_deviceWidth;
	int m_deviceHeight;
	unsigned m_localStorageUntrustedDomainsQuotaPerDomain;
	unsigned m_localStorageUntrustedDomainsTotalQuota;
	unsigned m_localStorageTrustedDomainsTotalQuota;
	HashMap<String, unsigned> m_localStorageQuota;
	unsigned m_sessionStorageQuota;
	unsigned m_editingBehaviorType;
	unsigned m_maximumHTMLParserDOMTreeDepth;
    SecurityOrigin::StorageBlockingPolicy m_storageBlockingPolicy;
#if ENABLE(TEXT_AUTOSIZING)
	IntSize m_textAutosizingWindowSizeOverride;
	bool m_textAutosizingEnabled : 1;
#endif
	bool m_isSpatialNavigationEnabled : 1;
	bool m_isJavaEnabled : 1;
	bool m_isJavaEnabledForLocalFiles : 1;
	bool m_loadsImagesAutomatically : 1;
	bool m_loadsSiteIconsIgnoringImageLoadingSetting : 1;
	bool m_privateBrowsingEnabled : 1;
	bool m_caretBrowsingEnabled : 1;
	bool m_areImagesEnabled : 1;
	bool m_isMediaEnabled : 1;
	bool m_arePluginsEnabled : 1;
	bool m_localStorageEnabled : 1;
	bool m_isScriptEnabled : 1;
	bool m_isWebSecurityEnabled : 1;
	bool m_allowUniversalAccessFromFileURLs: 1;
	bool m_allowFileAccessFromFileURLs: 1;
	bool m_javaScriptCanOpenWindowsAutomatically : 1;
	bool m_javaScriptCanAccessClipboard : 1;
	bool m_inspectorEnabled : 1;
	bool m_shouldPrintBackgrounds : 1;
	bool m_textAreasAreResizable : 1;
#if ENABLE(DASHBOARD_SUPPORT)
	bool m_usesDashboardBackwardCompatibilityMode : 1;
#endif
	bool m_needsAdobeFrameReloadingQuirk : 1;
	bool m_needsKeyboardEventDisambiguationQuirks : 1;
	bool m_treatsAnyTextCSSLinkAsStylesheet : 1;
	bool m_needsLeopardMailQuirks : 1;
	bool m_isDOMPasteAllowed : 1;
	bool m_shrinksStandaloneImagesToFit : 1;
	bool m_usesPageCache : 1;
	bool m_pageCacheSupportsPlugins : 1;
	bool m_shouldCacheXHRResource : 1;
	bool m_showsURLsInToolTips : 1;
	bool m_showsToolTipOverTruncatedText : 1;
	bool m_forceFTPDirectoryListings : 1;
	bool m_developerExtrasEnabled : 1;
	bool m_javaScriptExperimentsEnabled : 1;
	bool m_authorAndUserStylesEnabled : 1;
	bool m_needsSiteSpecificQuirks : 1;
	unsigned m_fontRenderingMode : 1;
	bool m_frameFlatteningEnabled : 1;
#if ENABLE(WEB_ARCHIVE)
	bool m_webArchiveDebugModeEnabled : 1;
#endif
	bool m_localFileContentSniffingEnabled : 1;
	bool m_inApplicationChromeMode : 1;
	bool m_offlineWebApplicationCacheEnabled : 1;
	bool m_enforceCSSMIMETypeInNoQuirksMode : 1;
	bool m_usesEncodingDetector : 1;
	bool m_allowScriptsToCloseWindows : 1;
	bool m_canvasUsesAcceleratedDrawing : 1;
	bool m_acceleratedDrawingEnabled : 1;
	bool m_acceleratedFiltersEnabled : 1;
	bool m_isCSSCustomFilterEnabled : 1;
	bool m_isCSSFilterEnabled : 1;
	bool m_isCSSCompositingEnabled : 1;
#if ENABLE(CSS_REGIONS)
	bool m_cssRegionsEnabled : 1;
#endif
	bool m_regionBasedColumnsEnabled : 1;
	bool m_downloadableBinaryFontsEnabled : 1;
	bool m_xssAuditorEnabled : 1;
	bool m_acceleratedCompositingEnabled : 1;
	bool m_acceleratedCompositingFor3DTransformsEnabled : 1;
	bool m_acceleratedCompositingForVideoEnabled : 1;
	bool m_acceleratedCompositingForPluginsEnabled : 1;
	bool m_acceleratedCompositingForCanvasEnabled : 1;
	bool m_acceleratedCompositingForAnimationEnabled : 1;
	bool m_acceleratedCompositingForFixedPositionEnabled : 1;
	bool m_acceleratedCompositingForScrollableFramesEnabled : 1; // Works only in conjunction with forceCompositingMode
	bool m_showDebugBorders : 1;
	bool m_showRepaintCounter : 1;
	bool m_experimentalNotificationsEnabled : 1;
	bool m_webGLEnabled : 1;
	bool m_webGLErrorsToConsoleEnabled : 1;
	bool m_openGLMultisamplingEnabled : 1;
	bool m_privilegedWebGLExtensionsEnabled : 1;
	bool m_webAudioEnabled : 1;
	bool m_workerEnabled : 1;
	bool m_mediaKeysEnabled : 1;
	bool m_webkitMediaKeysEnabled : 1;
	bool m_acceleratedCanvas2dEnabled : 1;
	bool m_deferredCanvas2dEnabled : 1;
	bool m_loadDeferringEnabled : 1;
	bool m_tiledBackingStoreEnabled : 1;
	bool m_paginateDuringLayoutEnabled : 1;
	bool m_dnsPrefetchingEnabled : 1;
#if ENABLE(FULLSCREEN_API)
	bool m_fullScreenAPIEnabled : 1;
#endif
	bool m_asynchronousSpellCheckingEnabled: 1;
	bool m_unifiedTextCheckerEnabled: 1;
	bool m_memoryInfoEnabled: 1;
	bool m_interactiveFormValidation: 1;
	bool m_usePreHTML5ParserQuirks: 1;
	bool m_hyperlinkAuditingEnabled : 1;
	bool m_crossOriginCheckInGetMatchedCSSRulesDisabled : 1;
	bool m_forceCompositingMode : 1;
	bool m_shouldInjectUserScriptsInInitialEmptyDocument : 1;
	bool m_fixedElementsLayoutRelativeToFrame : 1;
	bool m_allowDisplayOfInsecureContent : 1;
	bool m_allowRunningOfInsecureContent : 1;
#if ENABLE(SMOOTH_SCROLLING)
	bool m_scrollAnimatorEnabled : 1;
#endif
	bool m_mediaControllerEnabled : 1;
	bool m_mediaPlaybackRequiresUserGesture : 1;
	bool m_mediaPlaybackAllowsInline : 1;
	bool m_passwordEchoEnabled : 1;
	bool m_svgFontsEnabled : 1;
	bool m_suppressesIncrementalRendering : 1;
	bool m_backspaceKeyNavigationEnabled : 1;

#if ENABLE(VIDEO_TRACK)
	bool m_shouldDisplaySubtitles : 1;
	bool m_shouldDisplayCaptions : 1;
	bool m_shouldDisplayTextDescriptions : 1;
#endif
	bool m_scrollingCoordinatorEnabled : 1;

	bool m_notificationsEnabled : 1;
	bool m_needsIsLoadingInAPISenseQuirk : 1;

#if ENABLE(TOUCH_EVENTS)
	bool m_touchEventEmulationEnabled : 1;
#endif
	bool m_shouldRespectImageOrientation : 1;
	bool m_wantsBalancedSetDefersLoadingBehavior : 1;
	bool m_requestAnimationFrameEnabled : 1;
	bool m_deviceSupportsTouch : 1;
	bool m_deviceSupportsMouse : 1;
	bool m_needsDidFinishLoadOrderQuirk : 1;

	bool m_fixedPositionCreatesStackingContext : 1;
	bool m_syncXHRInDocumentsEnabled : 1;
	bool m_cookieEnabled : 1;
	bool m_thirdPartyIframeRedirectBlockingEnabled : 1;

	bool m_windowFocusRestricted : 1;

    bool m_diagnosticLoggingEnabled : 1;

    bool m_applyPageScaleFactorInCompositor : 1;

	bool m_drawImageRectEnabled : 1;
	bool m_drawBrokenImageEnabled : 1;

	bool m_httpEquivEnabled : 1;
	bool m_contentDispositionAttachmentSandboxEnabled : 1;

	bool m_debugTiming : 1;

	bool m_fixedBackgroundsPaintRelativeToDocument : 1;

    bool m_fontFallbackPrefersPictographs : 1;

    bool m_subresourceIntegrityEnabled : 1;

    bool m_displayContentsEnabled : 1;

    bool m_needsFrameNameFallbackToIdQuirk : 1;

    bool m_springTimingFunctionEnabled : 1;

    bool m_conicGradientsEnabled : 1;

    bool m_hasPseudoClassEnabled : 1;

    bool m_mediaCaptureRequiresSecureConnection : 1;

    bool m_focusVisibleEnabled : 1;

    bool m_aspectRatioEnabled : 1;

    bool m_aspectRatioOfImgFromWidthAndHeightEnabled : 1;

    bool m_cssIndividualTransformPropertiesEnabled : 1;

    bool m_shouldIgnoreFontLoadCompletions : 1;

    bool m_inertAttributeEnabled : 1;

    bool m_useImageDocumentForSubframePDF : 1;

    bool m_unsafePluginPastingEnabled : 1;

    bool m_dirPseudoEnabled : 1;

    bool m_css3DTransformInteroperabilityEnabled : 1;

    bool m_webAnimationsCompositeOperationsEnabled : 1;

    bool m_blankAnchorTargetImpliesNoOpenerEnabled : 1;

    bool m_referrerPolicyAttributeEnabled : 1;

    bool m_linkPreloadResponsiveImagesEnabled : 1;

    bool m_constantPropertiesEnabled : 1;

    bool m_shouldAllowUserInstalledFonts : 1;

    bool m_CSSOMViewScrollingAPIEnabled : 1;

    bool m_overflowClipEnabled : 1;

    bool m_intersectionObserverEnabled : 1;

    bool m_resizeObserverEnabled : 1;

    bool m_inputEventsEnabled : 1;

    bool m_visualViewportEnabled : 1;

    bool m_sendMouseEventsToDisabledFormControlsEnabled : 1;

    bool m_clientCoordinatesRelativeToLayoutViewport : 1;
    bool m_dateTimeInputsEditableComponentsEnabled : 1;
    bool m_liveRangeSelectionEnabled : 1;
    bool m_selectionAcrossShadowBoundariesEnabled : 1;
    bool m_selectionIncludesAltImageText : 1;
    bool m_logsPageMessagesToSystemConsoleEnabled : 1;

    bool m_multipartXMixedReplaceImageEnabled : 1;

    bool m_imageBitmapEnabled : 1;

    bool m_shouldRestrictBaseURLSchemes : 1;

    bool m_useAnonymousModeWhenFetchingMaskImages : 1;

    FontLoadTimingOverride m_fontLoadTimingOverride { FontLoadTimingOverride::None };

    Timer<Settings> m_setImageLoadingSettingsTimer;
    void imageLoadingSettingsTimerFired();

	double m_incrementalRenderingSuppressionTimeoutInSeconds;

	unsigned m_defaultMaximumRenderTreeDepth;

	static double gDefaultMinDOMTimerInterval;

#if USE(AVFOUNDATION)
	static bool gAVFoundationEnabled;
#endif
	static bool gMockScrollbarsEnabled;

#if USE(SAFARI_THEME)
	static bool gShouldPaintNativeControls;
#endif
#if PLATFORM(WIN) || (OS(WINDOWS) && PLATFORM(WX))
	static bool gShouldUseHighResolutionTimers;
#endif
#if USE(JSC)
	static bool gShouldRespectPriorityInCSSAttributeSetters;
#endif

	String m_defaultVideoPosterURL;

	Vector<String> m_arphicMobileFontNames;
};

ALWAYS_INLINE void Settings::setSpatialNavigationEnabled(bool isSpatialNavigationEnabled)
{
	m_isSpatialNavigationEnabled = isSpatialNavigationEnabled;
}

ALWAYS_INLINE void Settings::setJavaEnabled(bool isJavaEnabled)
{
	m_isJavaEnabled = isJavaEnabled;
}

ALWAYS_INLINE void Settings::setJavaScriptCanAccessClipboard(bool javaScriptCanAccessClipboard)
{
	m_javaScriptCanAccessClipboard = javaScriptCanAccessClipboard;
}

ALWAYS_INLINE void Settings::setDefaultTextEncodingName(const String& defaultTextEncodingName)
{
	m_defaultTextEncodingName = defaultTextEncodingName;
}

ALWAYS_INLINE void Settings::setLocalFileContentSniffingEnabled(bool enabled)
{
	m_localFileContentSniffingEnabled = enabled;
}

ALWAYS_INLINE void Settings::setApplicationChromeMode(bool mode)
{
	m_inApplicationChromeMode = mode;
}

ALWAYS_INLINE void Settings::setOfflineWebApplicationCacheEnabled(bool enabled)
{
	m_offlineWebApplicationCacheEnabled = enabled;
}

ALWAYS_INLINE void Settings::setEnforceCSSMIMETypeInNoQuirksMode(bool enforceCSSMIMETypeInNoQuirksMode)
{
	m_enforceCSSMIMETypeInNoQuirksMode = enforceCSSMIMETypeInNoQuirksMode;
}

ALWAYS_INLINE void Settings::setLoadsSiteIconsIgnoringImageLoadingSetting(bool loadsSiteIcons)
{
	m_loadsSiteIconsIgnoringImageLoadingSetting = loadsSiteIcons;
}

ALWAYS_INLINE void Settings::setWebSecurityEnabled(bool isWebSecurityEnabled)
{
	m_isWebSecurityEnabled = isWebSecurityEnabled;
}

ALWAYS_INLINE void Settings::setAllowUniversalAccessFromFileURLs(bool allowUniversalAccessFromFileURLs)
{
	m_allowUniversalAccessFromFileURLs = allowUniversalAccessFromFileURLs;
}

ALWAYS_INLINE void Settings::setAllowFileAccessFromFileURLs(bool allowFileAccessFromFileURLs)
{
	m_allowFileAccessFromFileURLs = allowFileAccessFromFileURLs;
}

ALWAYS_INLINE void Settings::setJavaEnabledForLocalFiles(bool isJavaEnabledForLocalFiles)
{
	m_isJavaEnabledForLocalFiles = isJavaEnabledForLocalFiles;
}

ALWAYS_INLINE void Settings::setImagesEnabled(bool areImagesEnabled)
{
	m_areImagesEnabled = areImagesEnabled;

    // See comment in setLoadsImagesAutomatically.
    m_setImageLoadingSettingsTimer.startOneShot(0);
}

ALWAYS_INLINE void Settings::setMediaEnabled(bool isMediaEnabled)
{
	m_isMediaEnabled = isMediaEnabled;
}

ALWAYS_INLINE void Settings::setSessionStorageQuota(unsigned sessionStorageQuota)
{
	m_sessionStorageQuota = sessionStorageQuota;
}

ALWAYS_INLINE void Settings::setLoadsImagesAutomatically(bool loadsImagesAutomatically)
{
	m_loadsImagesAutomatically = loadsImagesAutomatically;

	// Changing this setting to true might immediately start new loads for images that had previously had loading disabled.
	// If this happens while a WebView is being dealloc'ed, and we don't know the WebView is being dealloc'ed, these new loads
	// can cause crashes downstream when the WebView memory has actually been free'd.
	// One example where this can happen is in Mac apps that subclass WebView then do work in their overridden dealloc methods.
	// Starting these loads synchronously is not important.  By putting it on a 0-delay, properly closing the Page cancels them
	// before they have a chance to really start.
	// See http://webkit.org/b/60572 for more discussion.
	m_setImageLoadingSettingsTimer.startOneShot(0);
}

ALWAYS_INLINE void Settings::setLocalStorageUntrustedDomainsPath(const String& path)
{
	m_localStorageUntrustedDomainsPath = path;
}

ALWAYS_INLINE void Settings::setLocalStorageTrustedDomainsPath(const String& path)
{
	m_localStorageTrustedDomainsPath = path;
}

ALWAYS_INLINE void Settings::setCanvasUsesAcceleratedDrawing(bool enabled)
{
    m_canvasUsesAcceleratedDrawing = enabled;
}

ALWAYS_INLINE void Settings::setAcceleratedCompositingFor3DTransformsEnabled(bool enabled)
{
    m_acceleratedCompositingFor3DTransformsEnabled = enabled;
}

ALWAYS_INLINE void Settings::setAcceleratedCompositingForVideoEnabled(bool enabled)
{
    m_acceleratedCompositingForVideoEnabled = enabled;
}

ALWAYS_INLINE void Settings::setAcceleratedCompositingForPluginsEnabled(bool enabled)
{
    m_acceleratedCompositingForPluginsEnabled = enabled;
}

ALWAYS_INLINE void Settings::setAcceleratedCompositingForCanvasEnabled(bool enabled)
{
    m_acceleratedCompositingForCanvasEnabled = enabled;
}

ALWAYS_INLINE void Settings::setMinimumAccelerated2dCanvasSize(int numPixels)
{
    m_minimumAccelerated2dCanvasSize = numPixels;
}

ALWAYS_INLINE void Settings::setWebGLErrorsToConsoleEnabled(bool enabled)
{
    m_webGLErrorsToConsoleEnabled = enabled;
}

ALWAYS_INLINE void Settings::setPrivilegedWebGLExtensionsEnabled(bool enabled)
{
    m_privilegedWebGLExtensionsEnabled = enabled;
}

ALWAYS_INLINE void Settings::setDeferred2dCanvasEnabled(bool enabled)
{
    m_deferredCanvas2dEnabled = enabled;
}

ALWAYS_INLINE void Settings::setLoadDeferringEnabled(bool enabled)
{
    m_loadDeferringEnabled = enabled;
}

ALWAYS_INLINE void Settings::setShrinksStandaloneImagesToFit(bool shrinksStandaloneImagesToFit)
{
    m_shrinksStandaloneImagesToFit = shrinksStandaloneImagesToFit;
}

ALWAYS_INLINE void Settings::setShowsURLsInToolTips(bool showsURLsInToolTips)
{
    m_showsURLsInToolTips = showsURLsInToolTips;
}

ALWAYS_INLINE void Settings::setShowsToolTipOverTruncatedText(bool showsToolTipForTruncatedText)
{
    m_showsToolTipOverTruncatedText = showsToolTipForTruncatedText;
}

ALWAYS_INLINE void Settings::setFTPDirectoryTemplatePath(const String& path)
{
    m_ftpDirectoryTemplatePath = path;
}

ALWAYS_INLINE void Settings::setForceFTPDirectoryListings(bool force)
{
    m_forceFTPDirectoryListings = force;
}

ALWAYS_INLINE void Settings::setDeveloperExtrasEnabled(bool developerExtrasEnabled)
{
    m_developerExtrasEnabled = developerExtrasEnabled;
}

ALWAYS_INLINE void Settings::setJavaScriptExperimentsEnabled(bool javaScriptExperimentsEnabled)
{
    m_javaScriptExperimentsEnabled = javaScriptExperimentsEnabled;
}

ALWAYS_INLINE void Settings::setNeedsSiteSpecificQuirks(bool needsQuirks)
{
    m_needsSiteSpecificQuirks = needsQuirks;
}

ALWAYS_INLINE void Settings::setFrameFlatteningEnabled(bool frameFlatteningEnabled)
{
    m_frameFlatteningEnabled = frameFlatteningEnabled;
}

#if ENABLE(WEB_ARCHIVE)
ALWAYS_INLINE void Settings::setWebArchiveDebugModeEnabled(bool enabled)
{
    m_webArchiveDebugModeEnabled = enabled;
}
#endif


#if USE(SAFARI_THEME)
ALWAYS_INLINE void Settings::setShouldPaintNativeControls(bool shouldPaintNativeControls)
{
    gShouldPaintNativeControls = shouldPaintNativeControls;
}
#endif

ALWAYS_INLINE void Settings::setUsesEncodingDetector(bool usesEncodingDetector)
{
    m_usesEncodingDetector = usesEncodingDetector;
}

ALWAYS_INLINE void Settings::setDownloadableBinaryFontsEnabled(bool downloadableBinaryFontsEnabled)
{
    m_downloadableBinaryFontsEnabled = downloadableBinaryFontsEnabled;
}

ALWAYS_INLINE void Settings::setXSSAuditorEnabled(bool xssAuditorEnabled)
{
    m_xssAuditorEnabled = xssAuditorEnabled;
}

ALWAYS_INLINE void Settings::setAcceleratedCompositingForAnimationEnabled(bool enabled)
{
    m_acceleratedCompositingForAnimationEnabled = enabled;
}

ALWAYS_INLINE void Settings::setExperimentalNotificationsEnabled(bool enabled)
{
    m_experimentalNotificationsEnabled = enabled;
}

} // namespace WebCore
