/*
 * Copyright (C) 2011 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "Length.h"
#include "LengthFunctions.h"
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/Vector.h>

namespace WebCore {

class TextStream;

// FIXME: Find a way to unify this with CSSPrimitiveValue::UnitCategory?
enum class CalculationCategory : uint8_t {
    Number,
    Length,
    Percent,
    PercentNumber,
    PercentLength,
    Angle,
    Time,
    Frequency,
    // TODO:
    // Flex,
    // Resolution
    Other
};

enum class CalcOperator : uint8_t {
    Add = '+',
    Subtract = '-',
    Multiply = '*',
    Divide = '/',
    Min = 0,
    Max = 1,
};

enum class CalcExpressionNodeType : uint8_t {
    Undefined,
    Number,
    Length,
    Operation,
    Negation,
    Inversion,
    BlendLength,
};
        
class CalcExpressionNode {
    WTF_MAKE_FAST_ALLOCATED;
public:
    explicit CalcExpressionNode(CalcExpressionNodeType = CalcExpressionNodeType::Undefined);
    virtual ~CalcExpressionNode() { }

    CalcExpressionNodeType type() const { return m_type; }
    
    virtual float evaluate(float maxValue) const = 0;
    virtual bool operator==(const CalcExpressionNode&) const = 0;
    virtual void dump(TextStream&) const = 0;
    
protected:
    CalcExpressionNodeType m_type;
};

class CalcExpressionNumber FINAL : public CalcExpressionNode {
public:
    explicit CalcExpressionNumber(float);

    float value() const { return m_value; }

private:
    virtual float evaluate(float) const OVERRIDE;
    virtual bool operator==(const CalcExpressionNode&) const OVERRIDE;
    void dump(TextStream&) const OVERRIDE;

    float m_value;
};

class CalcExpressionLength FINAL : public CalcExpressionNode {
public:
    explicit CalcExpressionLength(const Length&);

    const Length& length() const { return m_length; }

private:
    virtual float evaluate(float maxValue) const OVERRIDE;
    virtual bool operator==(const CalcExpressionNode&) const OVERRIDE;
    void dump(TextStream&) const OVERRIDE;

    Length m_length;
};

class CalcExpressionNegation FINAL : public CalcExpressionNode {
public:
    CalcExpressionNegation(PassOwnPtr<CalcExpressionNode> node)
        : CalcExpressionNode(CalcExpressionNodeType::Negation)
        , m_child(node)
    {
        ASSERT(m_child);
    }

    float evaluate(float maxValue) const FINAL;
    bool operator==(const CalcExpressionNode&) const FINAL;
    void dump(TextStream&) const FINAL;

    const CalcExpressionNode* child() const { return m_child.get(); }

private:
    OwnPtr<CalcExpressionNode> m_child;
};

bool operator==(const CalcExpressionNegation&, const CalcExpressionNegation&);

inline const CalcExpressionNegation* toCalcExpressionNegation(const CalcExpressionNode* value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->type() == CalcExpressionNodeType::Negation);
    return static_cast<const CalcExpressionNegation*>(value);
}

inline const CalcExpressionNegation& toCalcExpressionNegation(const CalcExpressionNode& value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(value.type() == CalcExpressionNodeType::Negation);
    return static_cast<const CalcExpressionNegation&>(value);
}

class CalcExpressionInversion FINAL : public CalcExpressionNode {
public:
    CalcExpressionInversion(PassOwnPtr<CalcExpressionNode> node)
        : CalcExpressionNode(CalcExpressionNodeType::Inversion)
        , m_child(node)
    {
        ASSERT(m_child);
    }

    float evaluate(float maxValue) const FINAL;
    bool operator==(const CalcExpressionNode&) const FINAL;
    void dump(TextStream&) const FINAL;

    const CalcExpressionNode* child() const { return m_child.get(); }

private:
    OwnPtr<CalcExpressionNode> m_child;
};

bool operator==(const CalcExpressionInversion&, const CalcExpressionInversion&);

inline const CalcExpressionInversion* toCalcExpressionInversion(const CalcExpressionNode* value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->type() == CalcExpressionNodeType::Inversion);
    return static_cast<const CalcExpressionInversion*>(value);
}

inline const CalcExpressionInversion& toCalcExpressionInversion(const CalcExpressionNode& value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(value.type() == CalcExpressionNodeType::Inversion);
    return static_cast<const CalcExpressionInversion&>(value);
}

class CalcExpressionOperation FINAL : public CalcExpressionNode {
public:
	CalcExpressionOperation(Vector<OwnPtr<CalcExpressionNode>>&& children, CalcOperator, CalculationCategory destinationCategory = CalculationCategory::Other);

    CalcOperator getOperator() const { return m_operator; }
    CalculationCategory destinationCategory() const { return m_destinationCategory; }

    const Vector<OwnPtr<CalcExpressionNode>>& children() const { return m_children; }

private:
    virtual float evaluate(float maxValue) const OVERRIDE;
    virtual bool operator==(const CalcExpressionNode&) const OVERRIDE;
    void dump(TextStream&) const OVERRIDE;

    Vector<OwnPtr<CalcExpressionNode>> m_children;
    CalcOperator m_operator;
    CalculationCategory m_destinationCategory;
};

class CalcExpressionBlendLength FINAL : public CalcExpressionNode {
public:
    CalcExpressionBlendLength(const Length& from, const Length& to, double progress);

    const Length& from() const { return m_from; }
    const Length& to() const { return m_to; }
    double progress() const { return m_progress; }

private:
    virtual float evaluate(float maxValue) const OVERRIDE;
    virtual bool operator==(const CalcExpressionNode&) const OVERRIDE;
    void dump(TextStream&) const OVERRIDE;

    Length m_from;
    Length m_to;
    double m_progress;
};

class CalculationValue : public RefCounted<CalculationValue> {
public:
    static PassRefPtr<CalculationValue> create(PassOwnPtr<CalcExpressionNode>, ValueRange);
    float evaluate(float maxValue) const;

    bool shouldClampToNonNegative() const { return m_shouldClampToNonNegative; }
    const CalcExpressionNode& expression() const { return *m_expression; }

private:
    CalculationValue(PassOwnPtr<CalcExpressionNode>, ValueRange);

    PassOwnPtr<CalcExpressionNode> m_expression;
    bool m_shouldClampToNonNegative;
};

inline CalcExpressionNode::CalcExpressionNode(CalcExpressionNodeType type)
    : m_type(type)
{
}

inline CalculationValue::CalculationValue(PassOwnPtr<CalcExpressionNode> expression, ValueRange range)
    : m_expression(expression)
    , m_shouldClampToNonNegative(range == ValueRangeNonNegative)
{
}

inline bool operator==(const CalculationValue& a, const CalculationValue& b)
{
    return a.expression() == b.expression();
}

inline CalcExpressionNumber::CalcExpressionNumber(float value)
    : CalcExpressionNode(CalcExpressionNodeType::Number)
    , m_value(value)
{
}

inline bool operator==(const CalcExpressionNumber& a, const CalcExpressionNumber& b)
{
    return a.value() == b.value();
}

inline const CalcExpressionNumber* toCalcExpressionNumber(const CalcExpressionNode* value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->type() == CalcExpressionNodeType::Number);
    return static_cast<const CalcExpressionNumber*>(value);
}

inline const CalcExpressionNumber& toCalcExpressionNumber(const CalcExpressionNode& value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(value.type() == CalcExpressionNodeType::Number);
    return static_cast<const CalcExpressionNumber&>(value);
}

inline CalcExpressionLength::CalcExpressionLength(const Length& length)
    : CalcExpressionNode(CalcExpressionNodeType::Length)
    , m_length(length)
{
}

inline bool operator==(const CalcExpressionLength& a, const CalcExpressionLength& b)
{
    return a.length() == b.length();
}

inline const CalcExpressionLength* toCalcExpressionLength(const CalcExpressionNode* value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->type() == CalcExpressionNodeType::Length);
    return static_cast<const CalcExpressionLength*>(value);
}

inline const CalcExpressionLength& toCalcExpressionLength(const CalcExpressionNode& value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(value.type() == CalcExpressionNodeType::Length);
    return static_cast<const CalcExpressionLength&>(value);
}

inline CalcExpressionOperation::CalcExpressionOperation(Vector<OwnPtr<CalcExpressionNode>>&& children, CalcOperator op, CalculationCategory destinationCategory)
    : CalcExpressionNode(CalcExpressionNodeType::Operation)
    , m_children(WTFMove(children))
    , m_operator(op)
	, m_destinationCategory(destinationCategory)
{
}

bool operator==(const CalcExpressionOperation&, const CalcExpressionOperation&);

inline const CalcExpressionOperation* toCalcExpressionOperation(const CalcExpressionNode* value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->type() == CalcExpressionNodeType::Operation);
    return static_cast<const CalcExpressionOperation*>(value);
}

inline const CalcExpressionOperation& toCalcExpressionOperation(const CalcExpressionNode& value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(value.type() == CalcExpressionNodeType::Operation);
    return static_cast<const CalcExpressionOperation&>(value);
}

inline CalcExpressionBlendLength::CalcExpressionBlendLength(const Length& from, const Length& to, double progress)
    : CalcExpressionNode(CalcExpressionNodeType::BlendLength)
    , m_from(from)
    , m_to(to)
    , m_progress(progress)
{
}

inline bool operator==(const CalcExpressionBlendLength& a, const CalcExpressionBlendLength& b)
{
    return a.progress() == b.progress() && a.from() == b.from() && a.to() == b.to();
}

inline const CalcExpressionBlendLength* toCalcExpressionBlendLength(const CalcExpressionNode* value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(!value || value->type() == CalcExpressionNodeType::BlendLength);
    return static_cast<const CalcExpressionBlendLength*>(value);
}

inline const CalcExpressionBlendLength& toCalcExpressionBlendLength(const CalcExpressionNode& value)
{
    ASSERT_WITH_SECURITY_IMPLICATION(value.type() == CalcExpressionNodeType::BlendLength);
    return static_cast<const CalcExpressionBlendLength&>(value);
}

TextStream& operator<<(TextStream&, const CalculationValue&);
TextStream& operator<<(TextStream&, const CalcExpressionNode&);
TextStream& operator<<(TextStream&, CalcOperator);

} // namespace WebCore

#if 0

#define SPECIALIZE_TYPE_TRAITS_CALCEXPRESSION_NODE(ToValueTypeName, predicate) \
SPECIALIZE_TYPE_TRAITS_BEGIN(WebCore::ToValueTypeName) \
    static bool isType(const WebCore::CalcExpressionNode& node) { return node.predicate; } \
SPECIALIZE_TYPE_TRAITS_END()

SPECIALIZE_TYPE_TRAITS_CALCEXPRESSION_NODE(CalcExpressionNumber, type() == WebCore::CalcExpressionNodeType::Number)
SPECIALIZE_TYPE_TRAITS_CALCEXPRESSION_NODE(CalcExpressionLength, type() == WebCore::CalcExpressionNodeType::Length)
SPECIALIZE_TYPE_TRAITS_CALCEXPRESSION_NODE(CalcExpressionOperation, type() == WebCore::CalcExpressionNodeType::Operation)
SPECIALIZE_TYPE_TRAITS_CALCEXPRESSION_NODE(CalcExpressionNegation, type() == WebCore::CalcExpressionNodeType::Negation)
SPECIALIZE_TYPE_TRAITS_CALCEXPRESSION_NODE(CalcExpressionInversion, type() == WebCore::CalcExpressionNodeType::Inversion)
SPECIALIZE_TYPE_TRAITS_CALCEXPRESSION_NODE(CalcExpressionBlendLength, type() == WebCore::CalcExpressionNodeType::BlendLength)

#endif
