/*
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2011, 2012 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "KURLWTFURLImpl.h"
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>
#include <wtf/text/CString.h>
#include <wtf/text/StringView.h>
#include <wtf/text/WTFString.h>

#if USE(CF)
typedef const struct __CFURL* CFURLRef;
#endif

#if PLATFORM(MAC) || (PLATFORM(QT) && USE(QTKIT))
OBJC_CLASS NSURL;
#endif

#if PLATFORM(QT)
QT_BEGIN_NAMESPACE
class QUrl;
QT_END_NAMESPACE
#endif

#if USE(GOOGLEURL)
#include "KURLGooglePrivate.h"
#endif

namespace WebCore {

class URLTextEncoding {
public:
    virtual Vector<uint8_t> encodeForURLParsing(StringView) const = 0;
    virtual ~URLTextEncoding() = default;
};

enum ParsedURLStringTag { ParsedURLString };

class KURL final {
	WTF_MAKE_FAST_ALLOCATED;
public:
    // Generates a URL which contains a null string.
	ALWAYS_INLINE KURL() { invalidate(); }

    // The argument is an absolute URL string. The string is assumed to be output of KURL::string() called on a valid
    // KURL object, or indiscernible from such.
    // It is usually best to avoid repeatedly parsing a string, unless memory saving outweigh the possible slow-downs.
    KURL(ParsedURLStringTag, const String& absoluteURL);

    explicit ALWAYS_INLINE KURL(WTF::HashTableDeletedValueType) : m_string(WTF::HashTableDeletedValue) { }
    ALWAYS_INLINE bool isHashTableDeletedValue() const { return string().isHashTableDeletedValue(); }

    // Resolves the relative URL with the given base URL. If provided, the
    // URLTextEncoding is used to encode non-ASCII characers. The base URL can be
    // null or empty, in which case the relative URL will be interpreted as
    // absolute.
    // FIXME: If the base URL is invalid, this always creates an invalid
    // URL. Instead I think it would be better to treat all invalid base URLs
    // the same way we treate null and empty base URLs.
    KURL(const KURL& base, const String& relative, const URLTextEncoding* = nullptr);

    WTF_EXPORT_PRIVATE static KURL fakeURLWithRelativePart(StringView);
    WTF_EXPORT_PRIVATE static KURL fileURLWithFileSystemPath(StringView);

    String strippedForUseAsReferrer() const;

    // FIXME: The above functions should be harmonized so that passing a
    // base of null or the empty string gives the same result as the
    // standard String constructor.

    // Makes a deep copy. Helpful only if you need to use a KURL on another
    // thread.  Since the underlying StringImpl objects are immutable, there's
    // no other reason to ever prefer copy() over plain old assignment.
    KURL isolatedCopy() const;

    bool isNull() const;
    bool isEmpty() const;
    bool isValid() const;

    // Returns true if this URL has a path. Note that "http://foo.com/" has a
    // path of "/", so this function will return true. Only invalid or
    // non-hierarchical (like "javascript:") URLs will have no path.
    bool hasPath() const;

    bool hasCredentials() const;

    ALWAYS_INLINE const String& string() const { return m_string; }

    String stringCenterEllipsizedToLength(unsigned length = 1024) const;

    //! Returns the URL with the host and path in unicode.
    String unicode() const;
    StringView protocol() const;
    StringView host() const;
    Optional<uint16_t> port() const;

	String stringWithoutFragmentIdentifier() const;

    StringView queryWithLeadingQuestionMark() const;
	StringView fragmentIdentifierWithLeadingNumberSign() const;
	StringView stringWithoutQueryOrFragmentIdentifier() const;
	StringView viewWithoutFragmentIdentifier() const;

    String hostAndPort() const;
    String protocolHostAndPort() const;
    String user() const;
    String password() const;
    String pass() const { return password(); }
    String path() const;
    String lastPathComponent() const;
    String query() const;
    String fragmentIdentifier() const;
    bool hasFragmentIdentifier() const;

    WTF_EXPORT_PRIVATE KURL truncatedForUseAsBase() const;

    bool hasUsername() const;
    bool hasPassword() const;
    bool hasQuery() const;
    bool hasFragment() const;

    // Unlike user() and pass(), these functions don't decode escape sequences.
    // This is necessary for accurate round-tripping, because encoding doesn't encode '%' characters.
    String encodedUser() const;
    String encodedPassword() const;
    String encodedPass() const { return encodedPassword(); }

    String baseAsString() const;

    String fileSystemPath() const;

    // Returns true if the current URL's protocol is the same as the null-
    // terminated ASCII argument. The argument must be lower-case.
    bool protocolIs(const char*) const;
    bool protocolIs(StringView) const;
    bool protocolIsAbout() const;
    ALWAYS_INLINE bool protocolIsBlob() const { return protocolIs(StringView("blob")); }
    ALWAYS_INLINE bool protocolIsData() const { return protocolIs(StringView("data")); }
    bool protocolIsFile() const { return protocolIs("file"); }
    bool protocolIsJavaScript() const;
    bool protocolIsInHTTPFamily() const;
    bool isLocalFile() const;
    bool hasOpaquePath() const { return m_hasOpaquePath; }

    bool isAboutBlank() const;
    bool isAboutSrcDoc() const;

    bool isMatchingDomain(const String&) const;

    bool setProtocol(StringView);
    void setHost(StringView);

    void removePort();
    void setPort(Optional<uint16_t>);

    // Input is like "foo.com" or "foo.com:8000".
    void setHostAndPort(StringView);
    WTF_EXPORT_PRIVATE void removeHostAndPort();

    WTF_EXPORT_PRIVATE static bool hostIsIPAddress(StringView);

    void setUser(StringView);
    void setPassword(StringView);
    void setPass(StringView password) { return setPassword(password); }
    WTF_EXPORT_PRIVATE void removeCredentials();

    // If you pass an empty path for HTTP or HTTPS URLs, the resulting path
    // will be "/".
    void setPath(StringView);

    // The query may begin with a question mark, or, if not, one will be added
    // for you. Setting the query to the empty string will leave a "?" in the
    // URL (with nothing after it). To clear the query, pass a null string.
    void setQuery(StringView);

    void setFragmentIdentifier(StringView);
    void removeFragmentIdentifier();
    WTF_EXPORT_PRIVATE void removeQueryAndFragmentIdentifier();

    friend bool equalIgnoringFragmentIdentifier(const KURL&, const KURL&);

    friend bool protocolHostAndPortAreEqual(const KURL&, const KURL&);

    unsigned hostStart() const;
    unsigned hostEnd() const;

    unsigned pathStart() const;
    unsigned pathEnd() const;
    unsigned pathAfterLastSlash() const;

    ALWAYS_INLINE operator const String&() const { return string(); }

#if USE(CF)
    explicit KURL(CFURLRef);
    CFURLRef createCFURL() const;
#endif

#if PLATFORM(MAC) || (PLATFORM(QT) && USE(QTKIT))
    KURL(NSURL*);
    operator NSURL*() const;
#endif
#ifdef __OBJC__
    operator NSString*() const { return string(); }
#endif

#ifndef NDEBUG
    void print() const;
#endif

    bool hasSpecialScheme() const;

private:
    friend class URLParser;
    void invalidate();
    unsigned credentialsEnd() const;
    static bool protocolIs(const String&, const char*);
    void copyToBuffer(Vector<char, 512>& buffer) const;
    void remove(unsigned start, unsigned length);
    void parse(const String&);
    void parseAllowingC0AtEnd(const String&);

    void maybeTrimTrailingSpacesFromOpaquePath();

    String m_string;

    unsigned m_isValid : 1;
    unsigned m_protocolIsInHTTPFamily : 1;
    unsigned m_hasOpaquePath : 1;

    // This is out of order to align the bits better. The port is after the host.
    unsigned m_portLength : 3;
    static constexpr unsigned maxPortLength = (1 << 3) - 1;

    static constexpr unsigned maxSchemeLength = (1 << 26) - 1;
    unsigned m_schemeEnd : 26;
    unsigned m_userStart;
    unsigned m_userEnd;
    unsigned m_passwordEnd;
    unsigned m_hostEnd;
    unsigned m_pathAfterLastSlash;
    unsigned m_pathEnd;
    unsigned m_queryEnd;
};

//static_assert(sizeof(URL) == sizeof(String) + 8 * sizeof(unsigned), "URL should stay small");

ALWAYS_INLINE void KURL::invalidate()
{
    m_isValid = false;
    m_protocolIsInHTTPFamily = false;
    m_hasOpaquePath = false;
    m_schemeEnd = 0;
    m_userStart = 0;
    m_userEnd = 0;
    m_passwordEnd = 0;
    m_hostEnd = 0;
    m_portLength = 0;
    m_pathEnd = 0;
    m_pathAfterLastSlash = 0;
    m_queryEnd = 0;
}

bool operator==(const KURL&, const KURL&);
bool operator==(const KURL&, const String&);
bool operator==(const String&, const KURL&);
bool operator!=(const KURL&, const KURL&);
bool operator!=(const KURL&, const String&);
bool operator!=(const String&, const KURL&);

bool equalIgnoringFragmentIdentifier(const KURL&, const KURL&);
bool protocolHostAndPortAreEqual(const KURL&, const KURL&);
WTF_EXPORT_PRIVATE Vector<std::pair<String, String>> differingQueryParameters(const KURL&, const KURL&);
WTF_EXPORT_PRIVATE bool isEqualIgnoringQueryAndFragments(const KURL&, const KURL&);

const KURL& blankURL();
const KURL& aboutSrcDocURL();

// Functions to do URL operations on strings.
// These are operations that aren't faster on a parsed URL.
// These are also different from the KURL functions in that they don't require the string to be a valid and parsable URL.
// This is especially important because valid javascript URLs are not necessarily considered valid by KURL.

bool protocolIs(const String& url, const char* protocol);
bool protocolIsJavaScript(const String& url);
bool protocolIsJavaScript(StringView url);
bool protocolIsInHTTPFamily(const String& url);

Optional<uint16_t> defaultPortForProtocol(StringView protocol);
bool isDefaultPortForProtocol(uint16_t port, StringView protocol);
bool portAllowed(const KURL&); // Blacklist ports that should never be used for Web resources.

String mimeTypeFromDataURL(const String& url);

// Unescapes the given string using URL escaping rules, given an optional
// encoding (defaulting to UTF-8 otherwise). DANGER: If the URL has "%00"
// in it, the resulting string will have embedded null characters!
String decodeURLEscapeSequences(const String&);
class TextEncoding;
String decodeURLEscapeSequences(const String&, const TextEncoding&);

// FIXME: This is a wrong concept to expose, different parts of a URL need different escaping per the URL Standard.
String encodeWithURLEscapeSequences(const String&);

// Inlines.

ALWAYS_INLINE bool operator==(const KURL& a, const KURL& b)
{
    return a.string() == b.string();
}

ALWAYS_INLINE bool operator==(const KURL& a, const String& b)
{
    return a.string() == b;
}

ALWAYS_INLINE bool operator==(const String& a, const KURL& b)
{
    return a == b.string();
}

ALWAYS_INLINE bool operator!=(const KURL& a, const KURL& b)
{
    return a.string() != b.string();
}

ALWAYS_INLINE bool operator!=(const KURL& a, const String& b)
{
    return a.string() != b;
}

ALWAYS_INLINE bool operator!=(const String& a, const KURL& b)
{
    return a != b.string();
}

ALWAYS_INLINE StringView KURL::protocol() const
{
    if (!m_isValid)
        return StringView();

    return StringView(m_string).substring(0, m_schemeEnd);
}

ALWAYS_INLINE String KURL::query() const
{
    if (m_queryEnd == m_pathEnd)
        return String();

    return m_string.substring(m_pathEnd + 1, m_queryEnd - (m_pathEnd + 1));
}

ALWAYS_INLINE String KURL::path() const
{
    if (!m_isValid)
        return String();

    return m_string.substring(pathStart(), m_pathEnd - pathStart());
}

ALWAYS_INLINE String KURL::fragmentIdentifier() const
{
    if (!hasFragmentIdentifier())
        return String();

    return m_string.substring(m_queryEnd + 1);
}

ALWAYS_INLINE bool KURL::hasFragmentIdentifier() const
{
    return m_isValid && m_string.length() != m_queryEnd;
}

ALWAYS_INLINE String KURL::baseAsString() const
{
    return m_string.left(m_pathAfterLastSlash);
}

inline KURL KURL::truncatedForUseAsBase() const
{
    return KURL(KURL(), m_string.left(m_pathAfterLastSlash));
}

ALWAYS_INLINE bool KURL::isLocalFile() const
{
    // Including feed here might be a bad idea since drag and drop uses this check
    // and including feed would allow feeds to potentially let someone's blog
    // read the contents of the clipboard on a drag, even without a drop.
    // Likewise with using the FrameLoader::shouldTreatURLAsLocal() function.
    return protocolIs("file");
}

#if !USE(GOOGLEURL) && !USE(WTFURL)

// Inline versions of some non-GoogleURL functions so we can get inlining
// without having to have a lot of ugly ifdefs in the class definition.

ALWAYS_INLINE bool KURL::isNull() const
{
    return m_string.isNull();
}

ALWAYS_INLINE bool KURL::isEmpty() const
{
    return m_string.isEmpty();
}

ALWAYS_INLINE bool KURL::isValid() const
{
    return m_isValid;
}

inline bool KURL::hasPath() const
{
    return m_pathEnd != m_hostEnd + m_portLength;
}

inline bool KURL::hasCredentials() const
{
    return m_passwordEnd > m_userStart;
}

inline bool KURL::hasUsername() const
{
    return m_userEnd > m_userStart;
}

inline bool KURL::hasPassword() const
{
    return m_passwordEnd > (m_userEnd + 1);
}

inline bool KURL::hasQuery() const
{
    return m_queryEnd > m_pathEnd;
}

inline bool KURL::hasFragment() const
{
    return m_isValid && m_string.length() > m_queryEnd;
}

ALWAYS_INLINE bool KURL::protocolIsInHTTPFamily() const
{
    return m_protocolIsInHTTPFamily;
}

ALWAYS_INLINE bool KURL::protocolIsJavaScript() const
{
    return WebCore::protocolIsJavaScript(string());
}

ALWAYS_INLINE unsigned KURL::hostStart() const
{
    return (m_passwordEnd == m_userStart) ? m_passwordEnd : m_passwordEnd + 1;
}

ALWAYS_INLINE unsigned KURL::hostEnd() const
{
    return m_hostEnd;
}

ALWAYS_INLINE unsigned KURL::pathEnd() const
{
    return m_pathEnd;
}

ALWAYS_INLINE unsigned KURL::pathAfterLastSlash() const
{
    return m_pathAfterLastSlash;
}

typedef KURL URL;

#endif // !USE(GOOGLEURL) && !USE(WTFURL)

} // namespace WebCore
