/*
 * Copyright (C) 2010 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "PlatformDelegate.h"

#include "PlatformEventLoop.h"
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <wtf/CurrentTime.h>
#include <wtf/MainThread.h>
#include <wtf/PlatformUserPreferredLanguages.h>
#include <wtf/text/WTFString.h>

namespace WebCore
{

static PlatformDelegate* s_platformDelegate;

PlatformDelegate& platformDelegate()
{
	ASSERT(s_platformDelegate);
	return *s_platformDelegate;
}

DLL_PUBLIC void setPlatformDelegate(PlatformDelegate& platformDelegate)
{
	s_platformDelegate = &platformDelegate;
}

#if ENABLE(INSPECTOR)
void PlatformEventLoop::cycle()
{
	platformDelegate().eventLoopCycle(*this);
}
#endif

}

namespace WTF {

String platformLanguage()
{
	return WebCore::platformDelegate().platformLanguage();
}

void logMessage(const String& message)
{
	WebCore::platformDelegate().logMessage(message);
}

void scheduleDispatchFunctionsOnMainThread()
{
	WebCore::platformDelegate().scheduleDispatchFunctionsOnMainThread();
}

LocalTimeOffset calculateLocalTimeOffset(time_t localTime)
{
	return WebCore::platformDelegate().calculateLocalTimeOffset(localTime);
}

double currentTime()
{
	if (LIKELY(WebCore::s_platformDelegate))
		return WebCore::s_platformDelegate->currentTime();

#if CPU(X86_64)
    struct timeval now;
    gettimeofday(&now, 0);
    return now.tv_sec + now.tv_usec / 1000000.0;
#else
	struct timespec tp;
	clock_gettime(CLOCK_REALTIME, &tp);
	return (tp.tv_sec + ((double)tp.tv_nsec / (double)1000000000.0));
#endif
}

double monotonicallyIncreasingTime()
{
#if ENABLE(LINUX_MONOTONIC_TIME)
	struct timespec tp;
	if(LIKELY(clock_gettime(CLOCK_MONOTONIC, &tp) == 0))
		return (tp.tv_sec + ((double)tp.tv_nsec / (double)1000000000.0));

	if(LIKELY(clock_gettime(CLOCK_REALTIME, &tp) == 0))
		return (tp.tv_sec + ((double)tp.tv_nsec / (double)1000000000.0));

	return 0;
#else
	if (LIKELY(WebCore::s_platformDelegate))
		return WebCore::s_platformDelegate->monotonicallyIncreasingTime();

	return 0;
#endif
}

int getTimeZoneName(char* buf, int buf_len)
{
	return WebCore::platformDelegate().getTimeZoneName(buf, buf_len);
}

}
