/*
 * Copyright (C) 2010 Apple Inc. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"
#include "SchemeRegistry.h"
#include <wtf/HashSet.h>
#include <wtf/MainThread.h>
#include <wtf/NeverDestroyed.h>
#include <wtf/TCSpinLock.h>
#include <wtf/ThreadingPrimitives.h>
#include <mutex>

namespace WebCore {

static SpinLock& schemeRegistryLock()
{
	static LazyNeverDestroyed<SpinLock> lock;
	static std::once_flag onceFlag;
	std::call_once(
		onceFlag,
		[] {
			lock.construct();
		});

	return lock;
}

static URLSchemesMap& localURLSchemes()
{
    ASSERT(schemeRegistryLock().isHeld());
    static NeverDestroyed<URLSchemesMap> localSchemes;

    if (localSchemes->isEmpty()) {
        localSchemes->add(ASCIILiteral("file"));
#if PLATFORM(MAC)
        localSchemes->add(ASCIILiteral("applewebdata"));
#endif
#if PLATFORM(QT)
        localSchemes->add(ASCIILiteral("qrc"));
#endif
    }

    return localSchemes;
}

static URLSchemesMap& displayIsolatedURLSchemes()
{
    ASSERT(schemeRegistryLock().isHeld());
    static NeverDestroyed<URLSchemesMap> displayIsolatedSchemes;
    return displayIsolatedSchemes;
}

static URLSchemesMap& secureSchemes()
{
    ASSERT(schemeRegistryLock().isHeld());
    static NeverDestroyed<URLSchemesMap> secureSchemes;

    if (secureSchemes->isEmpty()) {
        secureSchemes->add(ASCIILiteral("https"));
        secureSchemes->add(ASCIILiteral("about"));
        secureSchemes->add(ASCIILiteral("data"));
        secureSchemes->add(ASCIILiteral("wss"));
    }

    return secureSchemes;
}

static URLSchemesMap& schemesWithUniqueOrigins()
{
    ASSERT(schemeRegistryLock().isHeld());
    static NeverDestroyed<URLSchemesMap> schemesWithUniqueOrigins;

    if (schemesWithUniqueOrigins->isEmpty()) {
        schemesWithUniqueOrigins->add(ASCIILiteral("about"));
        schemesWithUniqueOrigins->add(ASCIILiteral("javascript"));
        // This is a willful violation of HTML5.
        // See https://bugs.webkit.org/show_bug.cgi?id=11885
        schemesWithUniqueOrigins->add(ASCIILiteral("data"));
    }

    return schemesWithUniqueOrigins;
}

static URLSchemesMap& emptyDocumentSchemes()
{
    ASSERT(isMainThread());
    static NeverDestroyed<URLSchemesMap> emptyDocumentSchemes;

    if (emptyDocumentSchemes->isEmpty())
        emptyDocumentSchemes->add(ASCIILiteral("about"));

    return emptyDocumentSchemes;
}

static HashSet<String>& schemesForbiddenFromDomainRelaxation()
{
    ASSERT(isMainThread());
    static NeverDestroyed<HashSet<String>> schemes;
    return schemes;
}

static URLSchemesMap& canDisplayOnlyIfCanRequestSchemes()
{
    ASSERT(schemeRegistryLock().isHeld());
    static NeverDestroyed<URLSchemesMap> canDisplayOnlyIfCanRequestSchemes;

#if ENABLE(BLOB)
    if (canDisplayOnlyIfCanRequestSchemes->isEmpty()) {
        canDisplayOnlyIfCanRequestSchemes->add(ASCIILiteral("blob"));
#if ENABLE(FILE_SYSTEM)
        canDisplayOnlyIfCanRequestSchemes->add(ASCIILiteral("filesystem"));
#endif
    }
#endif // ENABLE(BLOB)

    return canDisplayOnlyIfCanRequestSchemes;
}

static URLSchemesMap& notAllowingJavascriptURLsSchemes()
{
    ASSERT(isMainThread());
    static NeverDestroyed<URLSchemesMap> notAllowingJavascriptURLsSchemes;
    return notAllowingJavascriptURLsSchemes;
}

void SchemeRegistry::registerURLSchemeAsLocal(const String& scheme)
{
    if (scheme.isNull())
        return;

	SpinLockHolder holder(schemeRegistryLock());
    localURLSchemes().add(scheme);
}

void SchemeRegistry::removeURLSchemeRegisteredAsLocal(const String& scheme)
{
    if (scheme == ASCIILiteral("file"))
        return;
#if PLATFORM(MAC)
    if (scheme == ASCIILiteral("applewebdata"))
        return;
#endif
	SpinLockHolder holder(schemeRegistryLock());
    localURLSchemes().remove(scheme);
}

static HashSet<String>& schemesHandledBySchemeHandler()
{
    ASSERT(schemeRegistryLock.isHeld());
    static NeverDestroyed<HashSet<String>> set;
    return set.get();
}

void SchemeRegistry::registerURLSchemeAsHandledBySchemeHandler(const String& scheme)
{
	SpinLockHolder holder(schemeRegistryLock());
    schemesHandledBySchemeHandler().add(scheme);
}

bool SchemeRegistry::schemeIsHandledBySchemeHandler(StringView scheme)
{
	SpinLockHolder holder(schemeRegistryLock());
    return schemesHandledBySchemeHandler().contains(scheme.toString());
}

static URLSchemesMap& schemesAllowingLocalStorageAccessInPrivateBrowsing()
{
    ASSERT(isMainThread());
    static NeverDestroyed<URLSchemesMap> schemesAllowingLocalStorageAccessInPrivateBrowsing;
    return schemesAllowingLocalStorageAccessInPrivateBrowsing;
}

static URLSchemesMap& schemesAllowingDatabaseAccessInPrivateBrowsing()
{
    ASSERT(isMainThread());
    static NeverDestroyed<URLSchemesMap> schemesAllowingDatabaseAccessInPrivateBrowsing;
    return schemesAllowingDatabaseAccessInPrivateBrowsing;
}

static URLSchemesMap& CORSEnabledSchemes()
{
    ASSERT(isMainThread());
    // FIXME: http://bugs.webkit.org/show_bug.cgi?id=77160
    static NeverDestroyed<URLSchemesMap> CORSEnabledSchemes;

    if (CORSEnabledSchemes->isEmpty()) {
        CORSEnabledSchemes->add(ASCIILiteral("http"));
        CORSEnabledSchemes->add(ASCIILiteral("https"));
    }

    return CORSEnabledSchemes;
}

static URLSchemesMap& ContentSecurityPolicyBypassingSchemes()
{
    ASSERT(schemeRegistryLock().isHeld());
    static NeverDestroyed<URLSchemesMap> schemes;
    return schemes;
}

bool SchemeRegistry::shouldTreatURLSchemeAsLocal(const String& scheme)
{
    if (scheme.isNull())
        return false;

	SpinLockHolder holder(schemeRegistryLock());
    return localURLSchemes().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsNoAccess(const String& scheme)
{
    if (scheme.isNull())
        return;

	SpinLockHolder holder(schemeRegistryLock());
    schemesWithUniqueOrigins().add(scheme);
}

bool SchemeRegistry::shouldTreatURLSchemeAsNoAccess(const String& scheme)
{
    if (scheme.isNull())
        return false;

	SpinLockHolder holder(schemeRegistryLock());
    return schemesWithUniqueOrigins().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsDisplayIsolated(const String& scheme)
{
    if (scheme.isNull())
        return;

	SpinLockHolder holder(schemeRegistryLock());
    displayIsolatedURLSchemes().add(scheme);
}

bool SchemeRegistry::shouldTreatURLSchemeAsDisplayIsolated(const String& scheme)
{
    if (scheme.isNull())
        return false;

	SpinLockHolder holder(schemeRegistryLock());
    return displayIsolatedURLSchemes().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsSecure(const String& scheme)
{
    if (scheme.isNull())
        return;

	SpinLockHolder holder(schemeRegistryLock());
    secureSchemes().add(scheme);
}

bool SchemeRegistry::shouldTreatURLSchemeAsSecure(const String& scheme)
{
    if (scheme.isNull())
        return false;

	SpinLockHolder holder(schemeRegistryLock());
    return secureSchemes().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsEmptyDocument(const String& scheme)
{
    if (scheme.isNull())
        return;
    emptyDocumentSchemes().add(scheme);
}

bool SchemeRegistry::shouldLoadURLSchemeAsEmptyDocument(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
    return emptyDocumentSchemes().contains(scheme);
}

void SchemeRegistry::setDomainRelaxationForbiddenForURLScheme(bool forbidden, const String& scheme)
{
    if (scheme.isEmpty())
        return;

    if (forbidden)
        schemesForbiddenFromDomainRelaxation().add(scheme);
    else
        schemesForbiddenFromDomainRelaxation().remove(scheme);
}

bool SchemeRegistry::isDomainRelaxationForbiddenForURLScheme(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
    return schemesForbiddenFromDomainRelaxation().contains(scheme);
}

bool SchemeRegistry::canDisplayOnlyIfCanRequest(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
	SpinLockHolder holder(schemeRegistryLock());
    return canDisplayOnlyIfCanRequestSchemes().contains(scheme);
}

void SchemeRegistry::registerAsCanDisplayOnlyIfCanRequest(const String& scheme)
{
    if (scheme.isNull())
        return;

	SpinLockHolder holder(schemeRegistryLock());
    canDisplayOnlyIfCanRequestSchemes().add(scheme);
}

void SchemeRegistry::registerURLSchemeAsNotAllowingJavascriptURLs(const String& scheme)
{
    if (scheme.isNull())
        return;
    notAllowingJavascriptURLsSchemes().add(scheme);
}

bool SchemeRegistry::shouldTreatURLSchemeAsNotAllowingJavascriptURLs(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
    return notAllowingJavascriptURLsSchemes().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsAllowingLocalStorageAccessInPrivateBrowsing(const String& scheme)
{
    if (scheme.isNull())
        return;
    schemesAllowingLocalStorageAccessInPrivateBrowsing().add(scheme);
}

bool SchemeRegistry::allowsLocalStorageAccessInPrivateBrowsing(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
    return schemesAllowingLocalStorageAccessInPrivateBrowsing().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsAllowingDatabaseAccessInPrivateBrowsing(const String& scheme)
{
    if (scheme.isNull())
        return;
    schemesAllowingDatabaseAccessInPrivateBrowsing().add(scheme);
}

bool SchemeRegistry::allowsDatabaseAccessInPrivateBrowsing(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
    return schemesAllowingDatabaseAccessInPrivateBrowsing().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsCORSEnabled(const String& scheme)
{
    if (scheme.isNull())
        return;
    CORSEnabledSchemes().add(scheme);
}

bool SchemeRegistry::shouldTreatURLSchemeAsCORSEnabled(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
    return CORSEnabledSchemes().contains(scheme);
}

void SchemeRegistry::registerURLSchemeAsBypassingContentSecurityPolicy(const String& scheme)
{
    if (scheme.isNull())
        return;
	SpinLockHolder holder(schemeRegistryLock());
    ContentSecurityPolicyBypassingSchemes().add(scheme);
}

void SchemeRegistry::removeURLSchemeRegisteredAsBypassingContentSecurityPolicy(const String& scheme)
{
    if (scheme.isNull())
        return;
	SpinLockHolder holder(schemeRegistryLock());
    ContentSecurityPolicyBypassingSchemes().remove(scheme);
}

bool SchemeRegistry::schemeShouldBypassContentSecurityPolicy(const String& scheme)
{
    if (scheme.isEmpty())
        return false;
	SpinLockHolder holder(schemeRegistryLock());
    return ContentSecurityPolicyBypassingSchemes().contains(scheme);
}

bool SchemeRegistry::shouldCacheResponsesFromURLSchemeIndefinitely(StringView scheme)
{
#if PLATFORM(MAC)
    if (equalLettersIgnoringASCIICase(scheme, "applewebdata"))
        return true;
#endif
    return equalLettersIgnoringASCIICase(scheme, "blob") || equalLettersIgnoringASCIICase(scheme, "data");
}

} // namespace WebCore
