/*
 * Copyright (C) 2000 Lars Knoll (knoll@kde.org)
 *           (C) 2000 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2003, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 * Copyright (C) 2006 Graham Dennis (graham.dennis@gmail.com)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "CSSValue.h"
#include "ExceptionCode.h"
#include <wtf/PassRefPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class TimingFunction : public RefCounted<TimingFunction> {
public:

    virtual PassRefPtr<TimingFunction> clone() const = 0;

    enum TimingFunctionType {
        LinearFunction, CubicBezierFunction, StepsFunction, SpringFunction
    };
    
    virtual ~TimingFunction() { }

    TimingFunctionType type() const { return m_type; }
    
    bool isLinearTimingFunction() const { return m_type == LinearFunction; }
    bool isCubicBezierTimingFunction() const { return m_type == CubicBezierFunction; }
    bool isStepsTimingFunction() const { return m_type == StepsFunction; }
    bool isSpringTimingFunction() const { return m_type == SpringFunction; }
    
    virtual bool operator==(const TimingFunction& other) const = 0;
    bool operator!=(const TimingFunction& other) const { return !(*this == other); }

    static PassRefPtr<TimingFunction> createFromCSSText(const String&, ExceptionCode&);
    static PassRefPtr<TimingFunction> createFromCSSValue(const CSSValue&);
    double transformTime(double, double, bool before = false) const;
    String cssText() const;

protected:
    TimingFunction(TimingFunctionType type)
        : m_type(type)
    {
    }
    
    TimingFunctionType m_type;
};

class LinearTimingFunction FINAL : public TimingFunction {
public:
    static PassRefPtr<LinearTimingFunction> create()
    {
        return adoptRef(new LinearTimingFunction);
    }
    
    ~LinearTimingFunction() { }
    
    virtual bool operator==(const TimingFunction& other) const FINAL
    {
        return other.isLinearTimingFunction();
    }
    
private:
    LinearTimingFunction()
        : TimingFunction(LinearFunction)
    {
    }

    virtual PassRefPtr<TimingFunction> clone() const OVERRIDE
    {
        return adoptRef(new LinearTimingFunction);
    }
};
    
inline bool isLinearTimingFunction(const TimingFunction* timingFunction)
{
	return timingFunction && timingFunction->type() == TimingFunction::LinearFunction;
}

class CubicBezierTimingFunction FINAL : public TimingFunction {
public:
    enum TimingFunctionPreset {
        Ease,
        EaseIn,
        EaseOut,
        EaseInOut,
        Custom
    };

    static PassRefPtr<CubicBezierTimingFunction> create(double x1, double y1, double x2, double y2)
    {
        return adoptRef(new CubicBezierTimingFunction(Custom, x1, y1, x2, y2));
    }

    static PassRefPtr<CubicBezierTimingFunction> create()
    {
        return adoptRef(new CubicBezierTimingFunction());
    }

    static PassRefPtr<CubicBezierTimingFunction> create(TimingFunctionPreset preset)
    {
        switch (preset) {
        case Ease:
            return adoptRef(new CubicBezierTimingFunction());
        case EaseIn:
            return adoptRef(new CubicBezierTimingFunction(EaseIn, 0.42, 0.0, 1.0, 1.0));
        case EaseOut:
            return adoptRef(new CubicBezierTimingFunction(EaseOut, 0.0, 0.0, 0.58, 1.0));
        case EaseInOut:
            return adoptRef(new CubicBezierTimingFunction(EaseInOut, 0.42, 0.0, 0.58, 1.0));
        default:
            ASSERT_NOT_REACHED();
            return nullptr;
        }
    }

    ~CubicBezierTimingFunction() { }
    
    virtual bool operator==(const TimingFunction& other) const FINAL
    {
        if (other.isCubicBezierTimingFunction()) {
            const CubicBezierTimingFunction* ctf = static_cast<const CubicBezierTimingFunction*>(&other);

            if (m_timingFunctionPreset != Custom)
				return m_timingFunctionPreset == ctf->m_timingFunctionPreset;

            return m_x1 == ctf->m_x1 && m_y1 == ctf->m_y1 && m_x2 == ctf->m_x2 && m_y2 == ctf->m_y2;
        }
        return false;
    }

    double x1() const { return m_x1; }
    double y1() const { return m_y1; }
    double x2() const { return m_x2; }
    double y2() const { return m_y2; }

    void setValues(double x1, double y1, double x2, double y2)
    {
        m_x1 = x1;
        m_y1 = y1;
        m_x2 = x2;
        m_y2 = y2;
    }

    TimingFunctionPreset timingFunctionPreset() const { return m_timingFunctionPreset; }
    void setTimingFunctionPreset(TimingFunctionPreset preset) { m_timingFunctionPreset = preset; }

    static const CubicBezierTimingFunction* defaultTimingFunction()
    {
        static const CubicBezierTimingFunction* dtf = create().leakRef();
        return dtf;
    }

    PassRefPtr<CubicBezierTimingFunction> createReversed() const
    {
        return create(1.0 - m_x2, 1.0 - m_y2, 1.0 - m_x1, 1.0 - m_y1);
    }

    bool isLinear() const
    {
        return (!m_x1 && !m_y1 && !m_x2 && !m_y2) || (m_x1 == 1.0 && m_y1 == 1.0 && m_x2 == 1.0 && m_y2 == 1.0) || (m_x1 == 0.0 && m_y1 == 0.0 && m_x2 == 1.0 && m_y2 == 1.0);
    }
    
private:
    explicit CubicBezierTimingFunction(TimingFunctionPreset preset = Ease, double x1 = 0.25, double y1 = 0.1, double x2 = 0.25, double y2 = 1.0)
        : TimingFunction(CubicBezierFunction)
        , m_x1(x1)
        , m_y1(y1)
        , m_x2(x2)
        , m_y2(y2)
    	, m_timingFunctionPreset(preset)
    {
    }

    virtual PassRefPtr<TimingFunction> clone() const OVERRIDE
    {
        return adoptRef(new CubicBezierTimingFunction(m_timingFunctionPreset, m_x1, m_y1, m_x2, m_y2));
    }

    double m_x1;
    double m_y1;
    double m_x2;
    double m_y2;
    TimingFunctionPreset m_timingFunctionPreset;
};

class StepsTimingFunction FINAL : public TimingFunction {
public:
    enum class StepPosition {
        JumpStart,
        JumpEnd,
        JumpNone,
        JumpBoth,
        Start,
        End,
    };

    static PassRefPtr<StepsTimingFunction> create(int steps, Optional<StepPosition> stepPosition)
    {
        return adoptRef(new StepsTimingFunction(steps, stepPosition));
    }
    static PassRefPtr<StepsTimingFunction> create()
    {
        return adoptRef(new StepsTimingFunction(1, StepPosition::End));
    }
    
    ~StepsTimingFunction() { }
    
    virtual bool operator==(const TimingFunction& other) const FINAL
    {
        if (!other.isStepsTimingFunction())
        	return false;

		auto& otherSteps = static_cast<const StepsTimingFunction&>(other);

        if (m_steps != otherSteps.m_steps)
            return false;

        if (m_stepPosition == otherSteps.m_stepPosition)
            return true;

        if (!m_stepPosition && *otherSteps.m_stepPosition == StepPosition::End)
            return true;

        if (!otherSteps.m_stepPosition && *m_stepPosition == StepPosition::End)
			return true;

        return false;
    }
    
    int numberOfSteps() const { return m_steps; }
    void setNumberOfSteps(int steps) { m_steps = steps; }

    Optional<StepPosition> stepPosition() const { return m_stepPosition; }
    void setStepPosition(Optional<StepPosition> stepPosition) { m_stepPosition = stepPosition; }

private:
    StepsTimingFunction(int steps, Optional<StepPosition> stepPosition)
        : TimingFunction(StepsFunction)
        , m_steps(steps)
		, m_stepPosition(stepPosition)
    {
    }

    virtual PassRefPtr<TimingFunction> clone() const OVERRIDE
    {
        return adoptRef(new StepsTimingFunction(m_steps, m_stepPosition));
    }
    
    int m_steps;
    Optional<StepPosition> m_stepPosition;
};

inline bool isStepsTimingFunction(const TimingFunction* timingFunction)
{
	return timingFunction && timingFunction->type() == TimingFunction::StepsFunction;
}

class SpringTimingFunction FINAL : public TimingFunction {
public:
    static PassRefPtr<SpringTimingFunction> create(double mass, double stiffness, double damping, double initialVelocity)
    {
        return adoptRef(new SpringTimingFunction(mass, stiffness, damping, initialVelocity));
    }

    static PassRefPtr<SpringTimingFunction> create()
    {
        // This create() function should only be used by the argument decoders, and it is expected that
        // real values will be filled in using setValues().
        return create(0, 0, 0, 0);
    }
    
    virtual bool operator==(const TimingFunction& other) const OVERRIDE
    {
        if (other.isSpringTimingFunction()) {
            const SpringTimingFunction& otherString = *static_cast<const SpringTimingFunction*>(&other);
            return m_mass == otherString.m_mass && m_stiffness == otherString.m_stiffness && m_damping == otherString.m_damping && m_initialVelocity == otherString.m_initialVelocity;
        }
        return false;
    }

    double mass() const { return m_mass; }
    double stiffness() const { return m_stiffness; }
    double damping() const { return m_damping; }
    double initialVelocity() const { return m_initialVelocity; }

    void setValues(double mass, double stiffness, double damping, double initialVelocity)
    {
        m_mass = mass;
        m_stiffness = stiffness;
        m_damping = damping;
        m_initialVelocity = initialVelocity;
    }

private:
    explicit SpringTimingFunction(double mass, double stiffness, double damping, double initialVelocity)
        : TimingFunction(SpringFunction)
        , m_mass(mass)
        , m_stiffness(stiffness)
        , m_damping(damping)
        , m_initialVelocity(initialVelocity)
    {
    }

    PassRefPtr<TimingFunction> clone() const OVERRIDE
    {
        return adoptRef(new SpringTimingFunction(m_mass, m_stiffness, m_damping, m_initialVelocity));
    }

    double m_mass;
    double m_stiffness;
    double m_damping;
    double m_initialVelocity;
};

} // namespace WebCore

#if 0

#define SPECIALIZE_TYPE_TRAITS_TIMINGFUNCTION(ToValueTypeName, predicate) \
SPECIALIZE_TYPE_TRAITS_BEGIN(ToValueTypeName) \
static bool isType(const WebCore::TimingFunction& function) { return function.predicate; } \
SPECIALIZE_TYPE_TRAITS_END()

SPECIALIZE_TYPE_TRAITS_TIMINGFUNCTION(WebCore::LinearTimingFunction, isLinearTimingFunction())
SPECIALIZE_TYPE_TRAITS_TIMINGFUNCTION(WebCore::CubicBezierTimingFunction, isCubicBezierTimingFunction())
SPECIALIZE_TYPE_TRAITS_TIMINGFUNCTION(WebCore::StepsTimingFunction, isStepsTimingFunction())
SPECIALIZE_TYPE_TRAITS_TIMINGFUNCTION(WebCore::SpringTimingFunction, isSpringTimingFunction())

namespace WTF {

template<> struct EnumTraits<WebCore::StepsTimingFunction::StepPosition> {
    using values = EnumValues<
        WebCore::StepsTimingFunction::StepPosition,
        WebCore::StepsTimingFunction::StepPosition::JumpStart,
        WebCore::StepsTimingFunction::StepPosition::JumpEnd,
        WebCore::StepsTimingFunction::StepPosition::JumpNone,
        WebCore::StepsTimingFunction::StepPosition::JumpBoth,
        WebCore::StepsTimingFunction::StepPosition::Start,
        WebCore::StepsTimingFunction::StepPosition::End
    >;
};

} // namespace WTF
#endif
