/*
 * Copyright (C) 2006, 2008 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2007-2008 Torch Mobile, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer. 
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "CSSFontSelector.h"
#include "FontDescription.h"
#include <limits.h>
#include <wtf/FixedArray.h>
#include <wtf/Forward.h>
#include <wtf/HashSet.h>
#include <wtf/PassRefPtr.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>
#include <wtf/WeakHashSet.h>
#include <wtf/text/AtomicStringHash.h>
#include <wtf/unicode/Unicode.h>

#if OS(WINDOWS)
#include <windows.h>
#include <objidl.h>
#include <mlang.h>
#endif

#include "FontPlatformData.h"
#include "RefPtrCairo.h"

namespace WebCore
{

class FontCascade;
class FontPlatformData;
class FontSelector;
class OpenTypeVerticalData;
class Font;

// This key contains the FontDescription fields other than family that matter when fetching FontDatas (platform fonts).
struct FontDescriptionKey {
    explicit FontDescriptionKey(unsigned size = 0)
        : m_isDeletedValue(false)
    	, m_size(size)
    	, m_horizontalSize(0)
    	, m_flags(0)
    {
    }
    FontDescriptionKey(const FontDescription& description)
        : m_isDeletedValue(false)
		, m_size(description.computedSize())
    	, m_horizontalSize(description.computedHorizontalSize())
    	, m_fontSelectionRequest(description.fontSelectionRequest())
        , m_flags(makeFlagsKey(description))
    	, m_locale(description.locale())
    	, m_featureSettings(description.featureSettings())
    { }

    explicit FontDescriptionKey(WTF::HashTableDeletedValueType)
        : m_isDeletedValue(true)
		, m_size(0)
		, m_horizontalSize(0)
    {
    }

    bool operator==(const FontDescriptionKey& other) const
    {
        return m_isDeletedValue == other.m_isDeletedValue
			&& m_size == other.m_size
			&& m_horizontalSize == other.m_horizontalSize
			&& m_fontSelectionRequest == other.m_fontSelectionRequest
			&& m_flags == other.m_flags
			&& m_locale == other.m_locale
            && m_featureSettings == other.m_featureSettings;
    }
    bool operator!=(const FontDescriptionKey& other) const
    {
        return !(*this == other);
    }

    float size() const { return m_size; }
    const FontSelectionRequest& fontSelectionRequest() const { return m_fontSelectionRequest; }
    const FixedArray<unsigned, 2>& flags() const { return m_flags; }

    bool isHashTableDeletedValue() const { return m_isDeletedValue; }

    inline unsigned computeHash() const
    {
        Hasher hasher;
        add(hasher, m_size, m_horizontalSize);
        add(hasher, m_fontSelectionRequest.weight);
        add(hasher, m_fontSelectionRequest.width);
        add(hasher, m_fontSelectionRequest.slope);
        add(hasher, m_locale.existingHash());
        for (unsigned i = 0; i < m_flags.size(); ++i)
            add(hasher, m_flags[i]);
        add(hasher, m_featureSettings.hash());
        return hasher.hash();
    }

private:
    static FixedArray<unsigned, 2> makeFlagsKey(const FontDescription& description)
    {
    	//static_assert(USCRIPT_CODE_LIMIT < 0x1000, "Script code must fit in an unsigned along with the other flags");
    	unsigned first = static_cast<unsigned>(description.script()) << 11
			| static_cast<unsigned>(description.textRenderingMode()) << 9
			| static_cast<unsigned>(description.fontSynthesis()) << 6
			| static_cast<unsigned>(description.widthVariant()) << 4
            | static_cast<unsigned>(description.nonCJKGlyphOrientation()) << 3
            | static_cast<unsigned>(description.orientation()) << 2
            | static_cast<unsigned>(description.renderingMode());
        unsigned second = static_cast<unsigned>(description.variantEastAsianRuby()) << 27
            | static_cast<unsigned>(description.variantEastAsianWidth()) << 25
            | static_cast<unsigned>(description.variantEastAsianVariant()) << 22
            | static_cast<unsigned>(description.variantAlternates()) << 21
            | static_cast<unsigned>(description.variantNumericSlashedZero()) << 20
            | static_cast<unsigned>(description.variantNumericOrdinal()) << 19
            | static_cast<unsigned>(description.variantNumericFraction()) << 17
            | static_cast<unsigned>(description.variantNumericSpacing()) << 15
            | static_cast<unsigned>(description.variantNumericFigure()) << 13
            | static_cast<unsigned>(description.variantCaps()) << 10
            | static_cast<unsigned>(description.variantPosition()) << 8
            | static_cast<unsigned>(description.variantContextualAlternates()) << 6
            | static_cast<unsigned>(description.variantHistoricalLigatures()) << 4
            | static_cast<unsigned>(description.variantDiscretionaryLigatures()) << 2
            | static_cast<unsigned>(description.variantCommonLigatures());
        FixedArray<unsigned, 2> result;
        result[0] = first;
        result[1] = second;
        return result;
    }


    bool m_isDeletedValue;
    float m_size;
    float m_horizontalSize;
    FontSelectionRequest m_fontSelectionRequest;
    FixedArray<unsigned, 2> m_flags;
    AtomicString m_locale;
    FontFeatureSettings m_featureSettings;
};

struct FontDescriptionKeyHash {
    static unsigned hash(const FontDescriptionKey& key)
    {
        return key.computeHash();
    }

    static bool equal(const FontDescriptionKey& a, const FontDescriptionKey& b)
    {
        return a == b;
    }

    static const bool safeToCompareToEmptyOrDeleted = true;
};

class FontCache final {
    friend class FontCachePurgePreventer;
    friend class NeverDestroyed<FontCache>;

    WTF_MAKE_NONCOPYABLE(FontCache); WTF_MAKE_FAST_ALLOCATED;
public:
    friend FontCache& fontCache();

    // This method is implemented by the platform.
    PassRefPtr<Font> systemFallbackForCharacters(const FontDescription&, const Font* originalFontData, bool isPlatformFont, const UChar* characters, int length);

    // Also implemented by the platform.
    bool platformInit();

#if OS(WINCE) && !PLATFORM(QT)
#if defined(IMLANG_FONT_LINK) && (IMLANG_FONT_LINK == 2)
    IMLangFontLink2* getFontLinkInterface();
#else
    IMLangFontLink* getFontLinkInterface();
#endif
    static void comInitialize();
    static void comUninitialize();
    static IMultiLanguage* getMultiLanguageInterface();
#elif PLATFORM(WIN)
    IMLangFontLink2* getFontLinkInterface();
#endif
    Vector<FontSelectionCapabilities> getFontSelectionCapabilitiesInFamily(const AtomicString&);

    PassRefPtr<Font> fontForFamily(const FontDescription&, const AtomicString&, const FontFeatureSettings* fontFaceFeatures = nullptr, const FontVariantSettings* fontFaceVariantSettings = nullptr, bool checkingAlternateName = false);
    PassRefPtr<Font> lastResortFallbackFont(const FontDescription&);
    Font* getNonRetainedLastResortFallbackFont(const FontDescription&);
    PassRefPtr<Font> fontForPlatformData(const FontPlatformData&);
    PassRefPtr<Font> similarFont(const FontDescription&);

    UChar32 missingCharacter(UChar32 c) const { return c == m_fontMissingCharacter ? 0 : m_fontMissingCharacter; }
    void setMissingCharacter(UChar32 c) { m_fontMissingCharacter = c; }
    
    void addClient(CSSFontSelector&);
    void removeClient(CSSFontSelector&);

    unsigned generation() const { return m_generation; }
    void invalidate();

    const FontCharReplaceMap& similarCharacters() const { return m_similarCharacters; }

    void addSimilarCharacter(UChar32 c, UChar32 replacement)
    {
    	m_similarCharacters.add(c, replacement);
    }

    void clearSimilarCharacters()
    {
    	m_similarCharacters.clear();
    }

    const Vector<String>& systemFontFamilyNames() const { return m_systemFontFamilyNames; }
    void scanSystemFontFamilyNames();

    size_t fontCount();
    size_t inactiveFontCount();
    void purgeInactiveFont(int count = INT_MAX);
    void purgeInactiveFontIfNeeded();

    void schedulePurgeInactiveFont();

#if PLATFORM(WIN)
    PassRefPtr<Font> fontFromDescriptionAndLogFont(const FontDescription&, ShouldRetain, const LOGFONT& font, AtomicString& outFontFamilyName);
#elif PLATFORM(CHROMIUM) && OS(WINDOWS)
    PassRefPtr<Font> fontFromDescriptionAndLogFont(const FontDescription&, ShouldRetain, const LOGFONT& font, wchar_t* outFontFamilyName);
#endif

#if ENABLE(OPENTYPE_VERTICAL)
#if USE(SKIA)
    typedef uint32_t FontFileKey;
#else
    typedef AtomicString FontFileKey;
#endif
    PassRefPtr<OpenTypeVerticalData> getVerticalData(const FontFileKey&, const FontPlatformData&);
#endif

private:
    FontCache();
    ~FontCache();

    void disablePurging() { m_purgePreventCount++; }
    void enablePurging()
    {
        ASSERT(m_purgePreventCount);
        if (!--m_purgePreventCount)
            purgeInactiveFontIfNeeded();
    }

    // FIXME: This method should eventually be removed.
    FontPlatformData* getCachedFontPlatformData(const FontDescription&, const AtomicString& family, bool checkingAlternateName = false);

    // These methods are implemented by each platform.
    FontPlatformData* createFontPlatformData(const FontDescription&, const AtomicString& family, const FontFeatureSettings* fontFaceFeatures, const FontVariantSettings* fontFaceVariantSettings);

    WeakHashSet<CSSFontSelector> m_clients;

    unsigned m_generation { 0 };

    // Don't purge if this count is > 0;
    int m_purgePreventCount { 0 };

    UChar32 m_fontMissingCharacter { 0 };

    FontCharReplaceMap m_similarCharacters;

    Vector<String> m_systemFontFamilyNames;
    RefPtr<FcPattern> m_perCharactersPattern;

    bool m_platformInited { false };

#if PLATFORM(MAC) || (PLATFORM(CHROMIUM) && OS(DARWIN)) || OS(ANDROID)
    friend class ComplexTextController;
#endif
    friend class Font; // For getCachedFontData(const FontPlatformData*)
};

// Get the global fontCache.
FontCache& fontCache();

inline void FontCache::addClient(CSSFontSelector& client)
{
    ASSERT(!m_clients.contains(client));
    m_clients.add(client);
}

inline void FontCache::removeClient(CSSFontSelector& client)
{
    ASSERT(m_clients.contains(client));
    m_clients.remove(client);
}

class FontCachePurgePreventer {
public:
    FontCachePurgePreventer() { fontCache().disablePurging(); }
    ~FontCachePurgePreventer() { fontCache().enablePurging(); }
};

#if !PLATFORM(COCOA)
inline PassRefPtr<Font> FontCache::similarFont(const FontDescription&)
{
    return nullptr;
}
#endif

#if PLATFORM(ESPIAL)
inline Vector<FontSelectionCapabilities> FontCache::getFontSelectionCapabilitiesInFamily(const AtomicString& familyName)
{
	return Vector<FontSelectionCapabilities>();
}
#endif

}
