/*
 * Copyright (C) 2006, 2010 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#ifndef FontCascadeFonts_h
#define FontCascadeFonts_h

#include "FontRanges.h"
#include "FontSelector.h"
#include "GlyphPage.h"
#include "Font.h"
#include "WidthCache.h"
#include <wtf/Forward.h>
#include <wtf/MainThread.h>
#include <wtf/PassOwnPtr.h>

namespace WebCore {

class FontCache;
class FontDescription;
class FontPlatformData;
class FontSelector;
class GraphicsContext;
class IntRect;
class MixedFontGlyphPage;

class FontCascadeFonts : public RefCounted<FontCascadeFonts> {
    WTF_MAKE_NONCOPYABLE(FontCascadeFonts);
public:
    static ALWAYS_INLINE PassRefPtr<FontCascadeFonts> create(PassRefPtr<FontSelector> fontSelector) { return adoptRef(new FontCascadeFonts(fontSelector)); }
    static PassRefPtr<FontCascadeFonts> createForPlatformFont(const FontPlatformData& platformData) { return adoptRef(new FontCascadeFonts(platformData)); }

    ~FontCascadeFonts();

    bool isForPlatformFont() const { return m_isForPlatformFont; }

    GlyphData glyphDataForCharacter(UChar32, const FontDescription&, FontVariant, bool useSimilarChar);

    bool isFixedPitch(const FontDescription&);
    void determinePitch(const FontDescription&);

    bool isLoadingCustomFonts() const;

    ALWAYS_INLINE FontSelector* fontSelector() { return m_fontSelector.get(); }
    // FIXME: It should be possible to combine fontSelectorVersion and generation.
    ALWAYS_INLINE unsigned fontSelectorVersion() const { return m_fontSelectorVersion; }
    ALWAYS_INLINE unsigned generation() const { return m_generation; }

    WidthCache& widthCache() { return m_widthCache; }
    const WidthCache& widthCache() const { return m_widthCache; }

    const Font* primaryFont(const FontDescription&);
    const FontRanges& realizeFallbackRangesAt(const FontDescription& description, unsigned fallbackIndex);

    HashSet<UChar32>& knownMissingGlyphMap() { return m_knownMissingGlyphMap; }

    void pruneSystemFallbacks();

private:
    FontCascadeFonts(PassRefPtr<FontSelector>);
    FontCascadeFonts(const FontPlatformData&);

    GlyphData glyphDataForSystemFallback(UChar32, const FontDescription&, FontVariant, bool useSimilarChar, bool systemFallbackShouldBeInvisible);
    GlyphData glyphDataForNormalVariant(UChar32, const FontDescription&, bool useSimilarChar);
	GlyphData glyphDataForVariant(UChar32, const FontDescription&, FontVariant, bool useSimilarChar, unsigned fallbackIndex);

    FontCache& m_fontCache;

    Vector<FontRanges, 1> m_realizedFallbackRanges;
    unsigned m_lastRealizedFallbackIndex;

    class GlyphPageCacheEntry {
    public:
        GlyphData glyphDataForCharacter(UChar32);

        void setSingleFontPage(PassRefPtr<GlyphPage>);
        void setGlyphDataForCharacter(UChar32, GlyphData);

        bool isNull() const { return !m_singleFont && !m_mixedFont; }
        bool isMixedFont() const { return !!m_mixedFont; }

    private:
        // Only one of these is non-null.
        RefPtr<GlyphPage> m_singleFont;
        RefPtr<MixedFontGlyphPage> m_mixedFont;
    };

    GlyphPageCacheEntry m_cachedPageZero;
    HashMap<int, GlyphPageCacheEntry> m_cachedPages;

    HashSet<RefPtr<Font>> m_systemFallbackFontSet;
    HashSet<UChar32> m_knownMissingGlyphMap;

    const Font* m_cachedPrimaryFont;
    RefPtr<FontSelector> m_fontSelector;

    WidthCache m_widthCache;

    unsigned m_fontSelectorVersion;
    unsigned m_generation;
    Pitch m_pitch;
    bool m_isForPlatformFont;
};

inline bool FontCascadeFonts::isFixedPitch(const FontDescription& description)
{
    if (m_pitch == UnknownPitch)
        determinePitch(description);
    return m_pitch == FixedPitch;
};

inline const Font* FontCascadeFonts::primaryFont(const FontDescription& description)
{
    ASSERT(isMainThread());
    if (!m_cachedPrimaryFont) {
        auto& primaryRanges = realizeFallbackRangesAt(description, 0);
        m_cachedPrimaryFont = primaryRanges.glyphDataForCharacter(' ', ExternalResourceDownloadPolicy::Allow).font;
        if (!m_cachedPrimaryFont)
            m_cachedPrimaryFont = primaryRanges.rangeAt(0).font(ExternalResourceDownloadPolicy::Allow);
        else if (m_cachedPrimaryFont->isInterstitial()) {
			for (unsigned index = 1; ; ++index) {
				auto& localRanges = realizeFallbackRangesAt(description, index);
				if (localRanges.isNull())
					break;
				auto* font = localRanges.glyphDataForCharacter(' ', ExternalResourceDownloadPolicy::Forbid).font;
				if (font && !font->isInterstitial()) {
					m_cachedPrimaryFont = font;
					break;
				}
			}
		}
    }
    return m_cachedPrimaryFont;
}

}

#endif // FontCascadeFonts_h
