/*
 * Copyright (C) 2003, 2006, 2009 Apple Inc. All rights reserved.
 *               2006 Rob Buis <buis@kde.org>
 * Copyright (C) 2007-2008 Torch Mobile, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "RoundedRect.h"
#include "WindRule.h"
#include <wtf/FastAllocBase.h>
#include <wtf/Forward.h>
#include <wtf/Optional.h>
#include <wtf/Vector.h>

#if USE(CG)
#include <wtf/RetainPtr.h>
#include <CoreGraphics/CGPath.h>
typedef struct CGPath PlatformPath;
#elif USE(OPENVG)
namespace WebCore {
class PlatformPathOpenVG;
}
typedef WebCore::PlatformPathOpenVG PlatformPath;
#elif PLATFORM(WX) && USE(WXGC)
class wxGraphicsPath;
typedef wxGraphicsPath PlatformPath;
#elif USE(SKIA)
class SkPath;
typedef SkPath PlatformPath;
#elif OS(WINCE)
namespace WebCore {
    class PlatformPath;
}
typedef WebCore::PlatformPath PlatformPath;
#else
typedef void PlatformPath;
#endif

#if USE(CAIRO)
#include "RefPtrCairo.h"
#else
typedef PlatformPath* PlatformPathPtr;
#endif

namespace WebCore {

class AffineTransform;
class FloatPoint;
class FloatRect;
class FloatRoundedRect;
class FloatSize;
class GraphicsContext;
class PathTraversalState;
class StrokeStyleApplier;
class TextStream;

// The points in the sturcture are the same as those that would be used with the
// add... method. For example, a line returns the endpoint, while a cubic returns
// two tangent points and the endpoint.
struct PathElement {
    enum class Type : uint8_t {
        MoveToPoint, // The points member will contain 1 value.
        AddLineToPoint, // The points member will contain 1 value.
        AddQuadCurveToPoint, // The points member will contain 2 values.
        AddCurveToPoint, // The points member will contain 3 values.
        CloseSubpath // The points member will contain no values.
    };
	FloatPoint points[3];
    Type type;
};

typedef void (*PathApplierFunction)(void* info, const PathElement&);

class Path {
	WTF_MAKE_FAST_ALLOCATED;
public:
	Path();
#if USE(CG)
	Path(RetainPtr<CGMutablePathRef>);
#endif
#if USE(CAIRO)
    explicit Path(PassRefPtr<cairo_t>);
#endif
	~Path();

	Path(const Path&);
	Path(Path&&);
	Path& operator=(const Path&);
	Path& operator=(Path&&);

	static Path polygonPathFromPoints(const Vector<FloatPoint>&);

	bool contains(const FloatPoint&, WindRule rule = RULE_NONZERO) const;
	bool strokeContains(StrokeStyleApplier*, const FloatPoint&) const;
	// fastBoundingRect() should equal or contain boundingRect(); boundingRect()
	// should perfectly bound the points within the path.
	FloatRect boundingRect() const;
	FloatRect fastBoundingRect() const;
	FloatRect strokeBoundingRect(StrokeStyleApplier* = 0) const;

	float length() const;
	PathTraversalState traversalStateAtLength(float length) const;
	FloatPoint pointAtLength(float length) const;

	void clear();
#if PLATFORM(QT)
	bool isNull() const { return false; }
#else
	bool isNull() const { return !m_path; }
#endif
	bool isEmpty() const;
    bool definitelySingleLine() const;

	// Gets the current point of the current path, which is conceptually the final point reached by the path so far.
	// Note the Path can be empty (isEmpty() == true) and still have a current point.
	bool hasCurrentPoint() const;
	FloatPoint currentPoint() const;

	void moveTo(const FloatPoint&);
	void addLineTo(const FloatPoint&);
	void addQuadCurveTo(const FloatPoint& controlPoint, const FloatPoint& endPoint);
	void addBezierCurveTo(const FloatPoint& controlPoint1, const FloatPoint& controlPoint2, const FloatPoint& endPoint);
	void addArcTo(const FloatPoint&, const FloatPoint&, float radius);
	void closeSubpath();

	void addArc(const FloatPoint&, float radius, float startAngle, float endAngle, bool anticlockwise);
	void addRect(const FloatRect&);
	void addEllipse(const FloatPoint&, float radiusX, float radiusY, float rotation, float startAngle, float endAngle, bool anticlockwise);
	void addEllipse(const FloatRect&);

	enum RoundedRectStrategy {
		PreferNativeRoundedRect,
		PreferBezierRoundedRect
	};

	void addRoundedRect(const FloatRect&, const FloatSize& roundingRadii, RoundedRectStrategy = PreferNativeRoundedRect);
	void addRoundedRect(const FloatRoundedRect&, RoundedRectStrategy = PreferNativeRoundedRect);
	void addRoundedRect(const RoundedRect&);

	void addPath(const Path&, const AffineTransform&);

	void translate(const FloatSize&);

    // To keep Path() cheap, it does not allocate a PlatformPath immediately
    // meaning Path::platformPath() can return null.
#if USE(DIRECT2D)
    FloatRect fastBoundingRectForStroke(const PlatformContextDirect2D&) const;
    PlatformPathPtr platformPath() const { return m_path.get(); }
#elif USE(CG)
    WEBCORE_EXPORT PlatformPathPtr platformPath() const;
#elif USE(CAIRO)
    cairo_t* cairoPath() const { return m_path.get(); }
#else
    PlatformPathPtr platformPath() const { return m_path; }
#endif

#if !USE(CAIRO)
    // ensurePlatformPath() will allocate a PlatformPath if it has not yet been and will never return null.
    WEBCORE_EXPORT PlatformPathPtr ensurePlatformPath();
#endif

	void apply(void* info, PathApplierFunction) const;
	void transform(const AffineTransform&);

	void addBeziersForRoundedRect(const FloatRect&, const FloatSize& topLeftRadius, const FloatSize& topRightRadius, const FloatSize& bottomLeftRadius, const FloatSize& bottomRightRadius);

#if USE(CG)
	void platformAddPathForRoundedRect(const FloatRect&, const FloatSize& topLeftRadius, const FloatSize& topRightRadius, const FloatSize& bottomLeftRadius, const FloatSize& bottomRightRadius);
#endif

#ifndef NDEBUG
	void dump() const;
#endif

private:
#if USE(CAIRO)
    cairo_t* ensureCairoPath();
    void appendElement(PathElement::Type, Vector<FloatPoint, 3>&&);
#endif

#if USE(CAIRO)
    RefPtr<cairo_t> m_path;
#else
    mutable PlatformPathStorageType m_path;
#endif
#if USE(CAIRO)
    Optional<Vector<PathElement>> m_elements;
#endif
};

TextStream& operator<<(TextStream&, const Path&);

}
