/*
 * Copyright (C) 2011 Apple Inc. All rights reserved.
 * Copyright (C) 2010 Sencha, Inc. All rights reserved.
 * Copyright (C) 2010 Igalia S.L. All rights reserved.
 * Copyright (C) Research In Motion Limited 2011. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#if USE(CAIRO)
#define GL_GLES_PROTOTYPES
#define GL_GLEXT_PROTOTYPES
#endif
#include "ShadowBlur.h"

#include "AffineTransform.h"
#include "FloatQuad.h"
#include "GraphicsContext.h"
#include "ImageBuffer.h"
#include "PlatformDelegate.h"
#include "Timer.h"
#include <wtf/HashSet.h>
#include <wtf/MathExtras.h>
#include <wtf/NeverDestroyed.h>
#include <wtf/Noncopyable.h>
#include <wtf/Scope.h>
#include <wtf/UnusedParam.h>


#if USE(CAIRO)
#define GL_GLES_PROTOTYPES
#include "CairoUtilities.h"
#if (defined(CAIRO_HAS_GL_SURFACE) || defined(CAIRO_HAS_GLESV2_SURFACE))
#include <cairo-gl-ext.h>
#endif
#endif

using namespace std;

namespace WebCore {

namespace
{
enum {
    leftLobe = 0,
    rightLobe = 1
};

static inline int roundUpToMultipleOf32(int d)
{
    return (1 + (d >> 5)) << 5;
}

class ScratchBuffer;

static const unsigned gMaxCachedScratchBuffers = 16;
static Vector<ScratchBuffer*> gCachedScratchBuffers;

static int cachedScratchBufferCount;

// ShadowBlur needs a scratch image as the buffer for the blur filter.
// Instead of creating and destroying the buffer for every operation,
// we create a buffer which will be automatically purged via a timer.
class ScratchBuffer {
    WTF_MAKE_FAST_ALLOCATED;
public:
    ScratchBuffer(bool canDelete)
	: m_purgeTimer(this, &ScratchBuffer::timerFired)
	, m_lastWasInset(false)
	, m_forceImageSurface(false)
	, m_canDelete(canDelete)
	, m_bufferInUse(false)
    {
    	++cachedScratchBufferCount;

    	//fprintf(stderr, "cachedScratchBufferCount %i\n", cachedScratchBufferCount);
    	//fflush(stderr);
    }

    ~ScratchBuffer()
    {
    	for (int i = 0; i < gCachedScratchBuffers.size(); ++i) {
    		if (gCachedScratchBuffers[i] == this) {
    			gCachedScratchBuffers.remove(i);
    			break;
    		}
    	}

    	--cachedScratchBufferCount;

    	//fprintf(stderr, "cachedScratchBufferCount %i\n", cachedScratchBufferCount);
    	//fflush(stderr);
    }
    
    ImageBuffer* getScratchBuffer(const IntSize& size, bool forceImageSurface)
    {
        ASSERT(!m_bufferInUse);

        m_bufferInUse = true;

        // We do not need to recreate the buffer if the current buffer is large enough.
        if (m_imageBuffer && m_imageBuffer->logicalSize().width() >= size.width() && m_imageBuffer->logicalSize().height() >= size.height())
            return m_imageBuffer.get();

        // Round to the nearest 32 pixels so we do not grow the buffer for similar sized requests.
        IntSize roundedSize(roundUpToMultipleOf32(size.width()), roundUpToMultipleOf32(size.height()));

        clearScratchBuffer();

        if(forceImageSurface)
        	m_imageBuffer = ImageBuffer::createWithImageSurface(roundedSize);
        else
            m_imageBuffer = ImageBuffer::create(roundedSize, 1, ColorSpaceSRGB, RenderingMode::Unaccelerated, LogGLMemoryError::No);

        return m_imageBuffer.get();
    }

    ImageBuffer* getAcceleratedImageBuffer()
    {
		if (m_acceleratedImageBuffer)
			return m_acceleratedImageBuffer.get();

#if (defined(CAIRO_HAS_GL_SURFACE) || defined(CAIRO_HAS_GLESV2_SURFACE))
		if (m_imageBuffer && cairo_surface_get_type(m_imageBuffer->nativeSurface()) == CAIRO_SURFACE_TYPE_IMAGE) {
			if(OwnPtr<ImageBuffer> glImageBuffer = ImageBuffer::create(*m_imageBuffer, RenderingMode::Accelerated, LogGLMemoryError::No))
				m_acceleratedImageBuffer = glImageBuffer.release();
		}
#endif

		return m_acceleratedImageBuffer ? m_acceleratedImageBuffer.get() : m_imageBuffer.get();
    }

    bool canUse(const IntSize& size)
	{
		// We do not need to recreate the buffer if the current buffer is large enough.
		if (m_imageBuffer && m_imageBuffer->logicalSize().width() >= size.width() && m_imageBuffer->logicalSize().height() >= size.height())
			return true;

		return false;
	}

    bool matchCachedShadowValues(const FloatSize& radius, const Color& color, const FloatRect& shadowRect, const FloatRoundedRect::Radii& radii, const FloatSize& layerSize, bool forceImageSurface)
    {
        if (!m_lastWasInset && m_lastRadius == radius && m_lastColor == color && m_lastShadowRect == shadowRect &&  m_lastRadii == radii && m_lastLayerSize == layerSize && m_forceImageSurface == forceImageSurface)
            return true;
    	return false;
    }

    bool setCachedShadowValues(const FloatSize& radius, const Color& color, const FloatRect& shadowRect, const FloatRoundedRect::Radii& radii, const FloatSize& layerSize, bool forceImageSurface)
    {
        if(matchCachedShadowValues(radius, color, shadowRect, radii, layerSize, forceImageSurface))
            return false;

        m_lastWasInset = false;
        m_forceImageSurface = forceImageSurface;
        m_lastRadius = radius;
        m_lastColor = color;
        m_lastShadowRect = shadowRect;
        m_lastRadii = radii;
        m_lastLayerSize = layerSize;
        m_acceleratedImageBuffer = nullptr;

        return true;
    }

    bool matchCachedInsetShadowValues(const FloatSize& radius, const Color& color, const FloatRect& bounds, const FloatRect& shadowRect, const FloatRoundedRect::Radii& radii, bool forceImageSurface)
	{
		if (m_lastWasInset && m_lastRadius == radius && m_lastColor == color && m_lastInsetBounds == bounds && shadowRect == m_lastShadowRect && radii == m_lastRadii && forceImageSurface == m_forceImageSurface)
			return true;
		return false;
	}

    bool setCachedInsetShadowValues(const FloatSize& radius, const Color& color, const FloatRect& bounds, const FloatRect& shadowRect, const FloatRoundedRect::Radii& radii, bool forceImageSurface)
    {
        if (matchCachedInsetShadowValues(radius, color, bounds, shadowRect, radii, forceImageSurface))
            return false;

        m_lastWasInset = true;
        m_forceImageSurface = forceImageSurface;
        m_lastInsetBounds = bounds;
        m_lastRadius = radius;
        m_lastColor = color;
        m_lastShadowRect = shadowRect;
        m_lastRadii = radii;
        m_acceleratedImageBuffer = nullptr;

        return true;
    }

    void scheduleScratchBufferPurge()
    {
        m_bufferInUse = false;

        if (m_purgeTimer.isActive())
            m_purgeTimer.stop();

        const double scratchBufferPurgeInterval = 5;
        m_purgeTimer.startOneShot(scratchBufferPurgeInterval);
    }

    static ScratchBuffer& sharedOther();
    
    void clearScratchBuffer()
    {
        m_imageBuffer = nullptr;
        m_acceleratedImageBuffer = nullptr;
        m_lastRadius = FloatSize();
        m_lastLayerSize = FloatSize();
    }

    bool bufferInUse() const { return m_bufferInUse; }

private:
    void timerFired()
    {
        clearScratchBuffer();

        if(m_canDelete)
        	delete this;
    }
    
    OwnPtr<ImageBuffer> m_imageBuffer;
    Timer<ScratchBuffer> m_purgeTimer;

    OwnPtr<ImageBuffer> m_acceleratedImageBuffer;
    
    FloatRect m_lastInsetBounds;
    FloatRect m_lastShadowRect;
    FloatRoundedRect::Radii m_lastRadii;
    Color m_lastColor;
    FloatSize m_lastRadius;
    bool m_lastWasInset;
    FloatSize m_lastLayerSize;
    bool m_forceImageSurface;
    
    const bool m_canDelete;
    bool m_bufferInUse;
};

ScratchBuffer& ScratchBuffer::sharedOther()
{
    static NeverDestroyed<ScratchBuffer> scratchBuffer(false);
    return scratchBuffer.get();
}
}

static const int templateSideLength = 1;

void clearCachedShadowBlurBuffer()
{
	if (!ScratchBuffer::sharedOther().bufferInUse())
		ScratchBuffer::sharedOther().clearScratchBuffer();

	for (int i = gCachedScratchBuffers.size() - 1; i >= 0; --i) {
		auto* scratchBuffer = gCachedScratchBuffers[i];
		if (!scratchBuffer->bufferInUse())
			delete scratchBuffer; // destructor removes it from container.
	}
}

static ScratchBuffer* getCachedScratchBuffer(const IntSize& size, const FloatSize& radius, const Color& color, const FloatRect& shadowRect, const FloatRoundedRect::Radii& radii, const FloatSize& layerSize, bool forceImageSurface)
{
	for (auto* ScratchBuffer : gCachedScratchBuffers) {
		if(ScratchBuffer->canUse(size) && ScratchBuffer->matchCachedShadowValues(radius, color, shadowRect, radii, layerSize, forceImageSurface))
			return ScratchBuffer;
	}

	ScratchBuffer* scratchBuffer = new ScratchBuffer(true);
	gCachedScratchBuffers.append(scratchBuffer);
	return scratchBuffer;
}

static ScratchBuffer* getCachedInsetScratchBuffer(const IntSize& size, const FloatSize& radius, const Color& color, const FloatRect& bounds, const FloatRect& shadowRect, const FloatRoundedRect::Radii& radii, bool forceImageSurface)
{
	for (auto* ScratchBuffer : gCachedScratchBuffers) {
		if(ScratchBuffer->canUse(size) && ScratchBuffer->matchCachedInsetShadowValues(radius, color, bounds, shadowRect, radii, forceImageSurface))
			return ScratchBuffer;
	}

	ScratchBuffer* scratchBuffer = new ScratchBuffer(true);
	gCachedScratchBuffers.append(scratchBuffer);
	return scratchBuffer;
}

ShadowBlur::ShadowBlur(const FloatSize& radius, const FloatSize& offset, const Color& color, bool shadowsIgnoreTransforms)
: m_color(color)
, m_blurRadius(radius)
, m_offset(offset)
, m_shadowsIgnoreTransforms(shadowsIgnoreTransforms)
{
      updateShadowBlurValues();
}

ShadowBlur::ShadowBlur(const GraphicsContextState& state)
    : m_color(state.shadowColor)
    , m_blurRadius(state.shadowBlur, state.shadowBlur)
    , m_offset(state.shadowOffset)
    , m_shadowsIgnoreTransforms(state.shadowsIgnoreTransforms)
{
#if USE(CG)
    if (state.shadowsUseLegacyRadius) {
        float shadowBlur = radiusToLegacyRadius(state.shadowBlur);
        m_blurRadius = FloatSize(shadowBlur, shadowBlur);
    }
#endif
    updateShadowBlurValues();
}

ShadowBlur::~ShadowBlur() = default;


void ShadowBlur::setShadowValues(const FloatSize& radius, const FloatSize& offset, const Color& color, bool ignoreTransforms)
{
    m_blurRadius = radius;
    m_offset = offset;
    m_color = color;
    m_shadowsIgnoreTransforms = ignoreTransforms;

    updateShadowBlurValues();
}

void ShadowBlur::updateShadowBlurValues()
{
    // Limit blur radius to 128 to avoid lots of very expensive blurring.
    m_blurRadius = m_blurRadius.shrunkTo(FloatSize(128, 128));

    // The type of shadow is decided by the blur radius, shadow offset, and shadow color.
    if (!m_color.isValid() || !m_color.alpha()) {
        // Can't paint the shadow with invalid or invisible color.
        m_type = NoShadow;
    } else if (m_blurRadius.width() > 0 || m_blurRadius.height() > 0) {
        // Shadow is always blurred, even the offset is zero.
        m_type = BlurShadow;
    } else if (!m_offset.width() && !m_offset.height()) {
        // Without blur and zero offset means the shadow is fully hidden.
        m_type = NoShadow;
    } else
        m_type = SolidShadow;
}

// Instead of integer division, we use 17.15 for fixed-point division.
static const int blurSumShift = 15;

// Takes a two dimensional array with three rows and two columns for the lobes.
static void calculateLobes(int lobes[][2], float blurRadius, bool shadowsIgnoreTransforms)
{
    int diameter;
    if (shadowsIgnoreTransforms)
        diameter = max(2, static_cast<int>(floorf((2 / 3.f) * blurRadius))); // Canvas shadow. FIXME: we should adjust the blur radius higher up.
    else {
        // http://dev.w3.org/csswg/css3-background/#box-shadow
        // Approximate a Gaussian blur with a standard deviation equal to half the blur radius,
        // which http://www.w3.org/TR/SVG/filters.html#feGaussianBlurElement tell us how to do.
        // However, shadows rendered according to that spec will extend a little further than m_blurRadius,
        // so we apply a fudge factor to bring the radius down slightly.
        float stdDev = blurRadius / 2;
        const float gaussianKernelFactor = 3 / 4.f * sqrtf(2 * piFloat);
        const float fudgeFactor = 0.88f;
        diameter = max(2, static_cast<int>(floorf(stdDev * gaussianKernelFactor * fudgeFactor + 0.5f)));
    }

    if (diameter & 1) {
        // if d is odd, use three box-blurs of size 'd', centered on the output pixel.
        int lobeSize = (diameter - 1) / 2;
        lobes[0][leftLobe] = lobeSize;
        lobes[0][rightLobe] = lobeSize;
        lobes[1][leftLobe] = lobeSize;
        lobes[1][rightLobe] = lobeSize;
        lobes[2][leftLobe] = lobeSize;
        lobes[2][rightLobe] = lobeSize;
    } else {
        // if d is even, two box-blurs of size 'd' (the first one centered on the pixel boundary
        // between the output pixel and the one to the left, the second one centered on the pixel
        // boundary between the output pixel and the one to the right) and one box blur of size 'd+1' centered on the output pixel
        int lobeSize = diameter / 2;
        lobes[0][leftLobe] = lobeSize;
        lobes[0][rightLobe] = lobeSize - 1;
        lobes[1][leftLobe] = lobeSize - 1;
        lobes[1][rightLobe] = lobeSize;
        lobes[2][leftLobe] = lobeSize;
        lobes[2][rightLobe] = lobeSize;
    }
}

void ShadowBlur::clear()
{
    m_type = NoShadow;
    m_color = Color();
    m_blurRadius = FloatSize();
    m_offset = FloatSize();
}

void ShadowBlur::blurLayerImage(unsigned char* imageData, const IntSize& size, int rowStride)
{
    const int channels[4] = { 3, 0, 1, 3 };

    int lobes[3][2]; // indexed by pass, and left/right lobe
    calculateLobes(lobes, m_blurRadius.width(), m_shadowsIgnoreTransforms);

    // First pass is horizontal.
    int stride = 4;
    int delta = rowStride;
    int final = size.height();
    int dim = size.width();

    // Two stages: horizontal and vertical
    for (int pass = 0; pass < 2; ++pass) {
        unsigned char* pixels = imageData;
        
        if (!pass && !m_blurRadius.width())
            final = 0; // Do no work if horizonal blur is zero.

        for (int j = 0; j < final; ++j, pixels += delta) {
            // For each step, we blur the alpha in a channel and store the result
            // in another channel for the subsequent step.
            // We use sliding window algorithm to accumulate the alpha values.
            // This is much more efficient than computing the sum of each pixels
            // covered by the box kernel size for each x.
            for (int step = 0; step < 3; ++step) {
                int side1 = lobes[step][leftLobe];
                int side2 = lobes[step][rightLobe];
                int pixelCount = side1 + 1 + side2;
                int invCount = ((1 << blurSumShift) + pixelCount - 1) / pixelCount;
                int ofs = 1 + side2;
                int alpha1 = pixels[channels[step]];
                int alpha2 = pixels[(dim - 1) * stride + channels[step]];

                unsigned char* ptr = pixels + channels[step + 1];
                unsigned char* prev = pixels + stride + channels[step];
                unsigned char* next = pixels + ofs * stride + channels[step];

                int i;
                int sum = side1 * alpha1 + alpha1;
                int limit = (dim < side2 + 1) ? dim : side2 + 1;

                for (i = 1; i < limit; ++i, prev += stride)
                    sum += *prev;

                if (limit <= side2)
                    sum += (side2 - limit + 1) * alpha2;

                limit = (side1 < dim) ? side1 : dim;
                for (i = 0; i < limit; ptr += stride, next += stride, ++i, ++ofs) {
                    *ptr = (sum * invCount) >> blurSumShift;
                    sum += ((ofs < dim) ? *next : alpha2) - alpha1;
                }
                
                prev = pixels + channels[step];
                for (; ofs < dim; ptr += stride, prev += stride, next += stride, ++i, ++ofs) {
                    *ptr = (sum * invCount) >> blurSumShift;
                    sum += (*next) - (*prev);
                }
                
                for (; i < dim; ptr += stride, prev += stride, ++i) {
                    *ptr = (sum * invCount) >> blurSumShift;
                    sum += alpha2 - (*prev);
                }
            }
        }

        // Last pass is vertical.
        stride = rowStride;
        delta = 4;
        final = size.width();
        dim = size.height();

        if (!m_blurRadius.height())
            break;

        if (m_blurRadius.width() != m_blurRadius.height())
            calculateLobes(lobes, m_blurRadius.height(), m_shadowsIgnoreTransforms);
    }
}

void ShadowBlur::adjustBlurRadius(GraphicsContext* context)
{
    if (!m_shadowsIgnoreTransforms)
        return;

    AffineTransform transform = context->getCTM();
    m_blurRadius.scale(1 / static_cast<float>(transform.xScale()), 1 / static_cast<float>(transform.yScale()));
}

IntSize ShadowBlur::blurredEdgeSize() const
{
    IntSize edgeSize = expandedIntSize(m_blurRadius);

    // To avoid slowing down blurLayerImage() for radius == 1, we give it two empty pixels on each side.
    if (edgeSize.width() == 1)
        edgeSize.setWidth(2);

    if (edgeSize.height() == 1)
        edgeSize.setHeight(2);

    return edgeSize;
}

IntRect ShadowBlur::calculateLayerBoundingRect(GraphicsContext* context, const FloatRect& shadowedRect, const IntRect& clipRect, bool shouldClipLayerRect)
{
    IntSize edgeSize = blurredEdgeSize();

    // Calculate the destination of the blurred and/or transformed layer.
    FloatRect layerRect;
    IntSize inflation;

    const AffineTransform transform = context->getCTM();
    if (m_shadowsIgnoreTransforms && !transform.isIdentity()) {
        FloatQuad transformedPolygon = transform.mapQuad(FloatQuad(shadowedRect));
        transformedPolygon.move(m_offset);
        layerRect = transform.inverse().valueOr(AffineTransform()).mapQuad(transformedPolygon).boundingBox();
    } else {
        layerRect = shadowedRect;
        layerRect.move(m_offset);
    }

    // We expand the area by the blur radius to give extra space for the blur transition.
    if (m_type == BlurShadow) {
        layerRect.inflateX(edgeSize.width());
        layerRect.inflateY(edgeSize.height());
        inflation = edgeSize;
    }

    FloatRect unclippedLayerRect = layerRect;

    if (!clipRect.contains(enclosingIntRect(layerRect))) {
        // If we are totally outside the clip region, we aren't painting at all.
        if (intersection(layerRect, clipRect).isEmpty())
            return IntRect();

        if(shouldClipLayerRect) {
			IntRect inflatedClip = clipRect;
			// Pixels at the edges can be affected by pixels outside the buffer,
			// so intersect with the clip inflated by the blur.
			if (m_type == BlurShadow) {
				inflatedClip.inflateX(edgeSize.width());
				inflatedClip.inflateY(edgeSize.height());
			} else {
				// Enlarge the clipping area 1 pixel so that the fill does not
				// bleed (due to antialiasing) even if the unaligned clip rect occurred
				inflatedClip.inflateX(1);
				inflatedClip.inflateY(1);
			}

			layerRect.intersect(inflatedClip);
        }
    }

    IntSize frameSize = inflation;
    frameSize.scale(2);
    m_sourceRect = FloatRect(0, 0, shadowedRect.width() + frameSize.width(), shadowedRect.height() + frameSize.height());
    m_layerOrigin = FloatPoint(layerRect.x(), layerRect.y());
    m_layerSize = layerRect.size();

    const FloatPoint unclippedLayerOrigin = FloatPoint(unclippedLayerRect.x(), unclippedLayerRect.y());
    const FloatSize clippedOut = unclippedLayerOrigin - m_layerOrigin;

    // Set the origin as the top left corner of the scratch image, or, in case there's a clipped
    // out region, set the origin accordingly to the full bounding rect's top-left corner.
    float translationX = -shadowedRect.x() + inflation.width() - fabsf(clippedOut.width());
    float translationY = -shadowedRect.y() + inflation.height() - fabsf(clippedOut.height());
    m_layerContextTranslation = FloatSize(translationX, translationY);

    return enclosingIntRect(layerRect);
}

void ShadowBlur::drawShadowBuffer(GraphicsContext* graphicsContext, ImageBuffer* layerImage)
{
    if (!layerImage)
        return;

    GraphicsContextStateSaver stateSaver(*graphicsContext);

    IntSize bufferSize = layerImage->internalSize();
    if (bufferSize != m_layerSize) {
        // The rect passed to clipToImageBuffer() has to be the size of the entire buffer,
        // but we may not have cleared it all, so clip to the filled part first.
        graphicsContext->clip(FloatRect(m_layerOrigin, m_layerSize));
    }
    graphicsContext->clipToImageBuffer(layerImage, FloatRect(m_layerOrigin, bufferSize));
    graphicsContext->setFillColor(m_color);

    graphicsContext->clearShadow();
    graphicsContext->fillRect(FloatRect(m_layerOrigin, m_sourceRect.size()));
}

static void computeSliceSizesFromRadii(const IntSize& twiceRadius, const FloatRoundedRect::Radii& radii, int& leftSlice, int& rightSlice, int& topSlice, int& bottomSlice)
{
    leftSlice = twiceRadius.width() + max(radii.topLeft().width(), radii.bottomLeft().width()); 
    rightSlice = twiceRadius.width() + max(radii.topRight().width(), radii.bottomRight().width());

    topSlice = twiceRadius.height() + max(radii.topLeft().height(), radii.topRight().height());
    bottomSlice = twiceRadius.height() + max(radii.bottomLeft().height(), radii.bottomRight().height());
}

IntSize ShadowBlur::templateSize(const IntSize& radiusPadding, const FloatRoundedRect::Radii& radii)
{
    const int templateSideLength = 1;

    int leftSlice;
    int rightSlice;
    int topSlice;
    int bottomSlice;
    
    IntSize blurExpansion = radiusPadding;
    blurExpansion.scale(2);

    computeSliceSizesFromRadii(blurExpansion, radii, leftSlice, rightSlice, topSlice, bottomSlice);
    
    return IntSize(templateSideLength + leftSlice + rightSlice,
                   templateSideLength + topSlice + bottomSlice);
}

void ShadowBlur::drawRectShadow(GraphicsContext* graphicsContext, const FloatRoundedRect& shadowedRect)
{
    IntRect layerRect = calculateLayerBoundingRect(graphicsContext, shadowedRect.rect(), graphicsContext->clipBounds(), false);
    if (layerRect.isEmpty())
        return;

    adjustBlurRadius(graphicsContext);

    bool canUseTilingTechnique = true;

    // drawRectShadowWithTiling does not work with rotations.
    // https://bugs.webkit.org/show_bug.cgi?id=45042
    if (!graphicsContext->getCTM().preservesAxisAlignment() || m_type != BlurShadow)
        canUseTilingTechnique = false;

    IntSize edgeSize = blurredEdgeSize();
    IntSize templateSize = this->templateSize(edgeSize, shadowedRect.radii());
    const FloatRect& rect = shadowedRect.rect();

    if (templateSize.width() > rect.width() || templateSize.height() > rect.height()
        || (templateSize.unclampedArea() > m_sourceRect.area()))
        canUseTilingTechnique = false;

	if (canUseTilingTechnique)
		drawRectShadowWithTiling(graphicsContext, shadowedRect, templateSize, edgeSize);
	else
		drawRectShadowWithoutTiling(graphicsContext, shadowedRect, layerRect);
}

void ShadowBlur::drawInsetShadow(GraphicsContext* graphicsContext, const FloatRect& rect, const FloatRoundedRect& holeRect)
{
    IntRect layerRect = calculateLayerBoundingRect(graphicsContext, rect, graphicsContext->clipBounds(), false);
    if (layerRect.isEmpty())
        return;

    adjustBlurRadius(graphicsContext);

    // drawInsetShadowWithTiling does not work with rotations.
    // https://bugs.webkit.org/show_bug.cgi?id=45042
    if (!graphicsContext->getCTM().preservesAxisAlignment() || m_type != BlurShadow) {
        drawInsetShadowWithoutTiling(graphicsContext, rect, holeRect, layerRect);
        return;
    }

    IntSize edgeSize = blurredEdgeSize();
    IntSize templateSize = this->templateSize(edgeSize, holeRect.radii());
    const FloatRect& hRect = holeRect.rect();

    if (templateSize.width() > hRect.width() || templateSize.height() > hRect.height()
        || (templateSize.width() * templateSize.height() > hRect.width() * hRect.height())) {
        drawInsetShadowWithoutTiling(graphicsContext, rect, holeRect, layerRect);
        return;
    }

    drawInsetShadowWithTiling(graphicsContext, rect, holeRect, templateSize, edgeSize);
}

void ShadowBlur::drawRectShadowWithoutTiling(GraphicsContext* graphicsContext, const FloatRoundedRect& shadowedRect, const IntRect& layerRect)
{
    FloatRect bufferRelativeShadowedRect = shadowedRect.rect();
    bufferRelativeShadowedRect.move(m_layerContextTranslation);

	ScratchBuffer* scratchBuffer = getCachedScratchBuffer(layerRect.size(), m_blurRadius, Color::black, bufferRelativeShadowedRect, shadowedRect.radii(), m_layerSize, s_forceImageSurfaceForBoxShadow);

	ImageBuffer* layerImage = scratchBuffer->getScratchBuffer(layerRect.size(), s_forceImageSurfaceForBoxShadow);
    if (!layerImage){
    	delete scratchBuffer;
        return;
    }

    // Only redraw in the scratch buffer if its cached contents don't match our needs
    bool redrawNeeded = scratchBuffer->setCachedShadowValues(m_blurRadius, Color::black, bufferRelativeShadowedRect, shadowedRect.radii(), m_layerSize, s_forceImageSurfaceForBoxShadow);
    if (redrawNeeded) {
        GraphicsContext* shadowContext = layerImage->context();
        GraphicsContextStateSaver stateSaver(*shadowContext);

        // Add a pixel to avoid later edge aliasing when rotated.
        shadowContext->clearRect(FloatRect(0, 0, m_layerSize.width() + 1, m_layerSize.height() + 1));
        shadowContext->translate(m_layerContextTranslation);
        shadowContext->setFillColor(Color::black);
        if (shadowedRect.radii().isZero())
            shadowContext->fillRect(shadowedRect.rect());
        else {
            Path path;
            path.addRoundedRect(shadowedRect);
            shadowContext->fillPath(path);
        }

        blurShadowBuffer(layerImage, expandedIntSize(m_layerSize));
    }

    layerImage = scratchBuffer->getAcceleratedImageBuffer();
    drawShadowBuffer(graphicsContext, layerImage);
    scratchBuffer->scheduleScratchBufferPurge();

	if (gCachedScratchBuffers.size() > gMaxCachedScratchBuffers)
		delete gCachedScratchBuffers[0];  // destructor removes it from container.
}

void ShadowBlur::drawInsetShadowWithoutTiling(GraphicsContext* graphicsContext, const FloatRect& rect, const FloatRoundedRect& holeRect, const IntRect& layerRect)
{
    FloatRect bufferRelativeRect = rect;
    bufferRelativeRect.move(m_layerContextTranslation);

    FloatRect bufferRelativeHoleRect = holeRect.rect();
    bufferRelativeHoleRect.move(m_layerContextTranslation);

	ScratchBuffer* scratchBuffer = getCachedInsetScratchBuffer(layerRect.size(), m_blurRadius, Color::black, bufferRelativeRect, bufferRelativeHoleRect, holeRect.radii(), s_forceImageSurfaceForBoxShadow);

	ImageBuffer* layerImage = scratchBuffer->getScratchBuffer(layerRect.size(), s_forceImageSurfaceForBoxShadow);
    if (!layerImage){
    	delete scratchBuffer;
        return;
    }

    // Only redraw in the scratch buffer if its cached contents don't match our needs
    bool redrawNeeded = scratchBuffer->setCachedInsetShadowValues(m_blurRadius, Color::black, bufferRelativeRect, bufferRelativeHoleRect, holeRect.radii(), s_forceImageSurfaceForBoxShadow);
    if (redrawNeeded) {
        GraphicsContext* shadowContext = layerImage->context();
        GraphicsContextStateSaver stateSaver(*shadowContext);

        // Add a pixel to avoid later edge aliasing when rotated.
        shadowContext->clearRect(FloatRect(0, 0, m_layerSize.width() + 1, m_layerSize.height() + 1));
        shadowContext->translate(m_layerContextTranslation);

        Path path;
        path.addRect(rect);
        if (holeRect.radii().isZero())
            path.addRect(holeRect.rect());
        else
            path.addRoundedRect(holeRect);

        shadowContext->setFillRule(RULE_EVENODD);
        shadowContext->setFillColor(Color::black);
        shadowContext->fillPath(path);

        blurShadowBuffer(layerImage, expandedIntSize(m_layerSize));
    }

    layerImage = scratchBuffer->getAcceleratedImageBuffer();
    drawShadowBuffer(graphicsContext, layerImage);
    scratchBuffer->scheduleScratchBufferPurge();

	if (gCachedScratchBuffers.size() > gMaxCachedScratchBuffers)
		delete gCachedScratchBuffers[0];  // destructor removes it from container.
}

/*
  These functions use tiling to improve the performance of the shadow
  drawing of rounded rectangles. The code basically does the following
  steps:

     1. Calculate the size of the shadow template, a rectangle that
     contains all the necessary tiles to draw the complete shadow.

     2. If that size is smaller than the real rectangle render the new
     template rectangle and its shadow in a new surface, in other case
     render the shadow of the real rectangle in the destination
     surface.

     3. Calculate the sizes and positions of the tiles and their
     destinations and use drawPattern to render the final shadow. The
     code divides the rendering in 8 tiles:

        1 | 2 | 3
       -----------
        4 |   | 5
       -----------
        6 | 7 | 8

     The corners are directly copied from the template rectangle to the
     real one and the side tiles are 1 pixel width, we use them as
     tiles to cover the destination side. The corner tiles are bigger
     than just the side of the rounded corner, we need to increase it
     because the modifications caused by the corner over the blur
     effect. We fill the central or outer part with solid color to complete
     the shadow.
 */

void ShadowBlur::drawInsetShadowWithTiling(GraphicsContext* graphicsContext, const FloatRect& rect, const FloatRoundedRect& holeRect, const IntSize& templateSize, const IntSize& edgeSize)
{
    // Draw the rectangle with hole.
    FloatRect templateBounds(0, 0, templateSize.width(), templateSize.height());
    FloatRect templateHole = FloatRect(edgeSize.width(), edgeSize.height(), templateSize.width() - 2 * edgeSize.width(), templateSize.height() - 2 * edgeSize.height());

	ScratchBuffer* scratchBuffer = getCachedInsetScratchBuffer(templateSize, m_blurRadius, m_color, templateBounds, templateHole, holeRect.radii(), s_forceImageSurfaceForBoxShadow);

	ImageBuffer* layerImage = scratchBuffer->getScratchBuffer(templateSize, s_forceImageSurfaceForBoxShadow);
    if (!layerImage){
    	delete scratchBuffer;
        return;
    }

    // Only redraw in the scratch buffer if its cached contents don't match our needs
    bool redrawNeeded = scratchBuffer->setCachedInsetShadowValues(m_blurRadius, m_color, templateBounds, templateHole, holeRect.radii(), s_forceImageSurfaceForBoxShadow);
    if (redrawNeeded) {
        // Draw shadow into a new ImageBuffer.
        GraphicsContext* shadowContext = layerImage->context();
        GraphicsContextStateSaver shadowStateSaver(*shadowContext);
        shadowContext->clearRect(templateBounds);
        shadowContext->setFillRule(RULE_EVENODD);
        shadowContext->setFillColor(Color::black);

        Path path;
        path.addRect(templateBounds);
        if (holeRect.radii().isZero())
            path.addRect(templateHole);
        else
            path.addRoundedRect(FloatRoundedRect(templateHole, holeRect.radii()));

        shadowContext->fillPath(path);

        blurAndColorShadowBuffer(layerImage, templateSize);
    }
    FloatSize offset = m_offset;
    if (shadowsIgnoreTransforms()) {
        AffineTransform transform = graphicsContext->getCTM();
        offset.scale(1 / transform.xScale(), 1 / transform.yScale());
    }

    FloatRect boundingRect = rect;
    boundingRect.move(offset);

    FloatRect destHoleRect = holeRect.rect();
    destHoleRect.move(offset);
    FloatRect destHoleBounds = destHoleRect;
    destHoleBounds.inflateX(edgeSize.width());
    destHoleBounds.inflateY(edgeSize.height());

    // Fill the external part of the shadow (which may be visible because of offset).
    Path exteriorPath;
    exteriorPath.addRect(boundingRect);
    exteriorPath.addRect(destHoleBounds);

    {
        GraphicsContextStateSaver fillStateSaver(*graphicsContext);
        graphicsContext->setFillRule(RULE_EVENODD);
        graphicsContext->setFillColor(m_color);
        graphicsContext->clearShadow();
        graphicsContext->fillPath(exteriorPath);
    }

    layerImage = scratchBuffer->getAcceleratedImageBuffer();
    drawLayerPieces(graphicsContext, layerImage, destHoleBounds, holeRect.radii(), edgeSize, templateSize, InnerShadow);
    scratchBuffer->scheduleScratchBufferPurge();

	if (gCachedScratchBuffers.size() > gMaxCachedScratchBuffers)
		delete gCachedScratchBuffers[0];  // destructor removes it from container.
}

void ShadowBlur::drawRectShadowWithTiling(GraphicsContext* graphicsContext, const FloatRoundedRect& shadowedRect, const IntSize& templateSize, const IntSize& edgeSize)
{
    FloatRect templateShadow = FloatRect(edgeSize.width(), edgeSize.height(), templateSize.width() - 2 * edgeSize.width(), templateSize.height() - 2 * edgeSize.height());

	ScratchBuffer* scratchBuffer = getCachedScratchBuffer(templateSize, m_blurRadius, m_color, templateShadow, shadowedRect.radii(), m_layerSize, s_forceImageSurfaceForBoxShadow);

	ImageBuffer* layerImage = scratchBuffer->getScratchBuffer(templateSize, s_forceImageSurfaceForBoxShadow);
    if (!layerImage){
    	delete scratchBuffer;
        return;
    }

    // Only redraw in the scratch buffer if its cached contents don't match our needs
    bool redrawNeeded = scratchBuffer->setCachedShadowValues(m_blurRadius, m_color, templateShadow, shadowedRect.radii(), m_layerSize, s_forceImageSurfaceForBoxShadow);
    if (redrawNeeded) {
        // Draw shadow into the ImageBuffer.
        GraphicsContext* shadowContext = layerImage->context();
        GraphicsContextStateSaver shadowStateSaver(*shadowContext);

        shadowContext->clearRect(FloatRect(0, 0, templateSize.width(), templateSize.height()));
        shadowContext->setFillColor(Color::black);
        
        if (shadowedRect.radii().isZero())
            shadowContext->fillRect(templateShadow);
        else {
            Path path;
            path.addRoundedRect(FloatRoundedRect(templateShadow, shadowedRect.radii()));
			shadowContext->fillPath(path);
        }

        blurAndColorShadowBuffer(layerImage, templateSize);
    }
    FloatSize offset = m_offset;
    if (shadowsIgnoreTransforms()) {
        AffineTransform transform = graphicsContext->getCTM();
        offset.scale(1 / transform.xScale(), 1 / transform.yScale());
    }

    FloatRect shadowBounds = shadowedRect.rect();
    shadowBounds.move(offset);
    shadowBounds.inflateX(edgeSize.width());
    shadowBounds.inflateY(edgeSize.height());

    layerImage = scratchBuffer->getAcceleratedImageBuffer();
    drawLayerPieces(graphicsContext, layerImage, shadowBounds, shadowedRect.radii(), edgeSize, templateSize, OuterShadow);
    scratchBuffer->scheduleScratchBufferPurge();

	if (gCachedScratchBuffers.size() > gMaxCachedScratchBuffers)
		delete gCachedScratchBuffers[0];  // destructor removes it from container.
}

void ShadowBlur::drawLayerPieces(GraphicsContext* graphicsContext, ImageBuffer* layerImage, const FloatRect& shadowBounds, const FloatRoundedRect::Radii& radii, const IntSize& bufferPadding, const IntSize& templateSize, ShadowDirection direction)
{
    const IntSize twiceRadius = IntSize(bufferPadding.width() * 2, bufferPadding.height() * 2);

    int leftSlice;
    int rightSlice;
    int topSlice;
    int bottomSlice;
    computeSliceSizesFromRadii(twiceRadius, radii, leftSlice, rightSlice, topSlice, bottomSlice);

    int centerWidth = shadowBounds.width() - leftSlice - rightSlice;
    int centerHeight = shadowBounds.height() - topSlice - bottomSlice;

    if (direction == OuterShadow) {
        FloatRect shadowInterior(shadowBounds.x() + leftSlice, shadowBounds.y() + topSlice, centerWidth, centerHeight);
        if (!shadowInterior.isEmpty()) {
            GraphicsContextStateSaver stateSaver(*graphicsContext);
            graphicsContext->setFillColor(m_color);
            graphicsContext->clearShadow();
            graphicsContext->fillRect(shadowInterior);
        }
    }

    GraphicsContextStateSaver stateSaver(*graphicsContext);
    graphicsContext->setFillColor(m_color);
    graphicsContext->clearShadow();

    // Note that drawing the ImageBuffer is faster than creating a Image and drawing that,
    // because ImageBuffer::draw() knows that it doesn't have to copy the image bits.
    FloatRect centerRect(shadowBounds.x() + leftSlice, shadowBounds.y() + topSlice, centerWidth, centerHeight);
    centerRect = graphicsContext->roundToDevicePixels(centerRect);
    
    // Top side.
    FloatRect tileRect = FloatRect(leftSlice, 0, templateSideLength, topSlice);
    FloatRect destRect = FloatRect(centerRect.x(), centerRect.y() - topSlice, centerRect.width(), topSlice);
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);

    // Draw the bottom side.
    tileRect.setY(templateSize.height() - bottomSlice);
    tileRect.setHeight(bottomSlice);
    destRect.setY(centerRect.maxY());
    destRect.setHeight(bottomSlice);
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);

    // Left side.
    tileRect = FloatRect(0, topSlice, leftSlice, templateSideLength);
    destRect = FloatRect(centerRect.x() - leftSlice, centerRect.y(), leftSlice, centerRect.height());
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);

    // Right side.
    tileRect.setX(templateSize.width() - rightSlice);
    tileRect.setWidth(rightSlice);
    destRect.setX(centerRect.maxX());
    destRect.setWidth(rightSlice);
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);

    // Top left corner.
    tileRect = FloatRect(0, 0, leftSlice, topSlice);
    destRect = FloatRect(centerRect.x() - leftSlice, centerRect.y() - topSlice, leftSlice, topSlice);
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);

    // Top right corner.
    tileRect = FloatRect(templateSize.width() - rightSlice, 0, rightSlice, topSlice);
    destRect = FloatRect(centerRect.maxX(), centerRect.y() - topSlice, rightSlice, topSlice);
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);

    // Bottom right corner.
    tileRect = FloatRect(templateSize.width() - rightSlice, templateSize.height() - bottomSlice, rightSlice, bottomSlice);
    destRect = FloatRect(centerRect.maxX(), centerRect.maxY(), rightSlice, bottomSlice);
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);

    // Bottom left corner.
    tileRect = FloatRect(0, templateSize.height() - bottomSlice, leftSlice, bottomSlice);
    destRect = FloatRect(centerRect.x() - leftSlice, centerRect.maxY(), leftSlice, bottomSlice);
    graphicsContext->drawImageBuffer(layerImage, destRect, tileRect);
}

void ShadowBlur::blurShadowBuffer(ImageBuffer* layerImage, const IntSize& templateSize)
{
    if (m_type != BlurShadow)
        return;

    const IntRect blurRect(IntPoint(), templateSize);

#if USE(CAIRO)

    if(!layerImage->nativeSurface())
    	return;
#if (defined(CAIRO_HAS_GL_SURFACE) || defined(CAIRO_HAS_GLESV2_SURFACE))
    if (isGPUAcceleratedBlur() && cairo_surface_get_type(layerImage->nativeSurface()) == CAIRO_SURFACE_TYPE_GL)
    {
        GLuint texture, fb;
        cairo_gl_surface_get_texture(layerImage->nativeSurface(), &texture, &fb);
        cairo_surface_flush(layerImage->nativeSurface());
        IntSize intBlurRadius(m_blurRadius.width(), m_blurRadius.height());

        if (glBlurShadowBuffer(texture, layerImage->internalSize(), templateSize, intBlurRadius))
        	return;
    }
#endif
    //fprintf(stderr, "%s -> image\n", __func__);
    //fflush(stderr);


    int width = templateSize.width();
    int height = templateSize.height();
    unsigned char* dataSrc = NULL;
    int stride = 0;

    if(!platformDelegate().lockSurface(layerImage->nativeSurface(), &dataSrc, &stride, false))
    	return;

    for (int y = 0; y < height; ++y) {
    	unsigned char* destRows = dataSrc + stride * y;
        unsigned* row = reinterpret_cast<unsigned*>(dataSrc + stride * y);
        for (int x = 0; x < width; x++) {
            int basex = x * 4;
            unsigned* pixel = row + x;
            Color pixelColor = colorFromPremultipliedARGB(*pixel);
#if CAIRO_USE_ARGB
            destRows[basex]     = pixelColor.red();
            destRows[basex + 1] = pixelColor.green();
            destRows[basex + 2] = pixelColor.blue();
#else
            destRows[basex]     = pixelColor.blue();
            destRows[basex + 1] = pixelColor.green();
            destRows[basex + 2] = pixelColor.red();
#endif
            destRows[basex + 3] = pixelColor.alpha();

        }
    }

	blurLayerImage(dataSrc, blurRect.size(), stride);

    for (int y = 0; y < height; ++y) {
    	unsigned char* srcRows = dataSrc + stride * y;
        unsigned* row = reinterpret_cast<unsigned*>(dataSrc + stride * y);
        for (int x = 0; x < width; x++) {
            int basex = x * 4;
            unsigned* pixel = row + x;
            Color pixelColor = Color::createUnCheked(srcRows[basex],
                                                     srcRows[basex + 1],
                                                     srcRows[basex + 2],
                                                     srcRows[basex + 3]);
            *pixel = premultipliedARGBFromColor(pixelColor);
        }
    }

    cairo_surface_mark_dirty_rectangle(layerImage->nativeSurface(), 0, 0, width, height);
    cairo_surface_flush (layerImage->nativeSurface());

    platformDelegate().unlockSurface(layerImage->nativeSurface(), true);
#else
    RefPtr<Uint8ClampedArray> layerData = layerImage->getUnmultipliedImageData(blurRect);
    if(layerData){
    	blurLayerImage(layerData->data(), blurRect.size(), blurRect.width() * 4);
    	layerImage->putByteArray(Unmultiplied, layerData.get(), blurRect.size(), blurRect, IntPoint());
    }
#endif
}

void ShadowBlur::blurAndColorShadowBuffer(ImageBuffer* layerImage, const IntSize& templateSize)
{
    blurShadowBuffer(layerImage, templateSize);

    // Mask the image with the shadow color.
    GraphicsContext* shadowContext = layerImage->context();
    GraphicsContextStateSaver stateSaver(*shadowContext);
    shadowContext->setCompositeOperation(CompositeSourceIn);
    shadowContext->setFillColor(m_color);
    shadowContext->fillRect(FloatRect(0, 0, templateSize.width(), templateSize.height()));
}

ImageBuffer* ShadowBlur::beginShadowLayer(GraphicsContext *context, const FloatRect& layerArea)
{
    if (type() == NoShadow)
    	return nullptr;

    adjustBlurRadius(context);

    IntRect layerRect = calculateLayerBoundingRect(context, layerArea, context->clipBounds(), true);

    if (layerRect.isEmpty())
        return nullptr;

    // We reset the scratch buffer values here, because the buffer will no longer contain
    // data from any previous rectangle or inset shadows drawn via the tiling path.
    ScratchBuffer::sharedOther().setCachedShadowValues(FloatSize(), Color::black, IntRect(), FloatRoundedRect::Radii(), m_layerSize, s_forceImageSurfaceForTextShadow);
    ImageBuffer* layerImage = ScratchBuffer::sharedOther().getScratchBuffer(layerRect.size(), s_forceImageSurfaceForTextShadow);
    if(!layerImage)
    	return nullptr;

    GraphicsContext* shadowContext = layerImage->context();
    shadowContext->save();

    // Add a pixel to avoid later edge aliasing when rotated.
    shadowContext->clearRect(FloatRect(0, 0, m_layerSize.width() + 1, m_layerSize.height() + 1));

    shadowContext->translate(m_layerContextTranslation);
    return layerImage;
}

void ShadowBlur::endShadowLayer(GraphicsContext* context, ImageBuffer* layerImage)
{
    layerImage->context()->restore();

    blurAndColorShadowBuffer(layerImage, expandedIntSize(m_layerSize));
    GraphicsContextStateSaver stateSave(*context);

    context->clearShadow();
    context->drawImageBuffer(layerImage, roundedIntPoint(m_layerOrigin), IntRect(0, 0, m_layerSize.width(), m_layerSize.height()), context->compositeOperation());

    ScratchBuffer::sharedOther().scheduleScratchBufferPurge();
}

bool ShadowBlur::s_forceImageSurfaceForBoxShadow = false;
bool ShadowBlur::s_forceImageSurfaceForTextShadow = false;
bool ShadowBlur::s_isGPUAcceleratedBlur = false;
size_t ShadowBlur::s_gpuAcceleratedBlurMaxRadius = 0;

#if USE(CAIRO) && (defined(CAIRO_HAS_GL_SURFACE) || defined(CAIRO_HAS_GLESV2_SURFACE))
static bool g_glBlurInitialized;
static unsigned g_glBlurFrameBuffer, g_glBlurTexture;
static unsigned g_glBlurArrayBuffer;
static unsigned g_glBlurVSH;
static WTF::Vector<unsigned> g_glBlurProgram;
static WTF::Vector<unsigned> g_glBlurTemplateSizeLoc, g_glBlurTextureSizeLoc, g_glBlurDirectionLoc;
static unsigned g_glBlurPosLoc = 1, g_glBlurTexLoc = 2;
#endif

DLL_PUBLIC void ShadowBlur::setIsGPUAcceleratedBlur(bool forceImageSurfaceForBoxShadow, bool forceImageSurfaceForTextShadow, bool isGPUAcceleratedBlur, size_t maxRadius)
{
	s_forceImageSurfaceForBoxShadow = forceImageSurfaceForBoxShadow;
	s_forceImageSurfaceForTextShadow = forceImageSurfaceForTextShadow;
    s_isGPUAcceleratedBlur = isGPUAcceleratedBlur;

#if USE(CAIRO) && (defined(CAIRO_HAS_GL_SURFACE) || defined(CAIRO_HAS_GLESV2_SURFACE))
    if (g_glBlurProgram.size() < maxRadius)
    {
        size_t oldRadius = g_glBlurProgram.size();
        g_glBlurProgram.grow(maxRadius);
        for (size_t i = oldRadius; i < maxRadius; i++)
            g_glBlurProgram[i] = 0;
    }

    if (g_glBlurProgram.size() > maxRadius)
    {
        size_t oldRadius = g_glBlurProgram.size();

        for (size_t i = maxRadius; i < oldRadius; i++)
            if (g_glBlurProgram[i] != 0)
                glDeleteProgram(g_glBlurProgram[i]);

        g_glBlurProgram.shrink(maxRadius);
    }

    g_glBlurTemplateSizeLoc.resize(maxRadius);
    g_glBlurTextureSizeLoc.resize(maxRadius);
    g_glBlurDirectionLoc.resize(maxRadius);
#endif

    s_gpuAcceleratedBlurMaxRadius = maxRadius;
}

#if USE(CAIRO) && (defined(CAIRO_HAS_GL_SURFACE) || defined(CAIRO_HAS_GLESV2_SURFACE))


bool ShadowBlur::glBlurCompileShader(unsigned blurSize)
{
	if(blurSize >= g_glBlurProgram.size())
		return false;

	if (g_glBlurProgram[blurSize] != 0)
		return true;

	constexpr const char * shader_blur_fsh_tempalte[] = {
		"#version 100",
		R"(
	#ifdef GL_ES
		precision highp float;
	#endif
	uniform sampler2D tex0;
	uniform vec2 u_direction;
	uniform vec2 u_textureSize;
	uniform vec2 u_templateSize;
	varying vec2 v_tex;
	vec4 blur(float dist, float weight)
	{
		vec4 res = vec4(0.0);
		vec2 coord1 = v_tex - dist * u_direction;
		if (coord1.x > 0.0 && coord1.y > 0.0)
			res += texture2D(tex0, coord1 / u_textureSize) * weight;
		else
			res += texture2D(tex0, vec2(0.0)) * weight;
		vec2 coord2 = v_tex + dist * u_direction;
		if (coord2.x <= u_templateSize.x && coord2.y <= u_templateSize.y)
			res += texture2D(tex0, coord2 / u_textureSize) * weight;
		else
			res += texture2D(tex0, (u_templateSize - vec2(1.0))/ u_textureSize) * weight;
		return res;
	}

	void  main()
	{
		vec4 sum = vec4(0.0); )"

		,
			R"(
		gl_FragColor  = sum;
	}
			)"};

	constexpr const char * shader_blur_vsh[] = {
		"#version 100",
		R"(
	#ifdef GL_ES
		precision highp float;
	#endif
	attribute vec4 a_pos;
	attribute vec2 a_tex;
	varying   vec2 v_tex;

	void  main()
	{
			gl_Position = a_pos;
			v_tex = a_tex ;
	}
			)"
	};


	auto print_info_log = [](
			GLuint object,// handle to the shader,
			GLenum check,
			void (*getIv)(GLuint, GLenum, GLint *),
			void (*getInfoLog)(GLuint, GLsizei, GLsizei*, GLchar*)
	) -> bool
	{
		GLint  length;

		GLint success;
		getIv( object, check, &success );
		if ( success != GL_TRUE )
		{
			getIv( object, GL_INFO_LOG_LENGTH , &length );

			if ( length ) {
				char* buffer  =  new char [ length ];
				getInfoLog ( object, length , NULL , buffer );
				fprintf(stderr, "object info: %s\n", buffer);
				fflush(stderr);
				delete [] buffer;

			}
			return false;
		}
		return true;
	};



	auto loadShader = [&print_info_log]( const char* const*shader_source, int source_sections,	GLenum type) -> GLuint
	{
		GLuint  shader = glCreateShader( type );
		if (shader == 0)
			return 0;
		unsigned ES_2 = 0;

#if USE(OPENGL_ES_2)
		ES_2 = 1;
#endif

		glShaderSource  ( shader , source_sections - ES_2, (const char **)(shader_source + ES_2), NULL );
		glCompileShader ( shader );

		if (print_info_log( shader, GL_COMPILE_STATUS, glGetShaderiv, glGetShaderInfoLog ))
			return shader;
		else
			glDeleteShader(type);

		return 0;
	};

	if (g_glBlurVSH == 0)
	{
		g_glBlurVSH = loadShader(shader_blur_vsh, 2, GL_VERTEX_SHADER);
		if (g_glBlurVSH == 0)
		{
			//fprintf(stderr, "failed to initialize glBlurShader\n");
			//fflush(stderr);
			return false;
		}
	}

	auto createProgram = [&print_info_log](GLuint vsh, GLuint fsh) -> GLuint
	{
		GLuint program = glCreateProgram();
		glAttachShader(program, vsh);
		glAttachShader(program, fsh);

		glBindAttribLocation(program, g_glBlurPosLoc, "a_pos");
		glBindAttribLocation(program, g_glBlurTexLoc, "a_tex");

		glLinkProgram(program);
		if (!print_info_log(program, GL_LINK_STATUS, glGetProgramiv, glGetProgramInfoLog))
		{
			glDeleteProgram(program);
			program = 0;
		}

		return program;
	};

	auto pdf = [](float x, float stdDev) -> float{
		// https://drafts.fxtf.org/filter-effects/#feGaussianBlurElement
		// H(x) = exp(-x^2 / (2s^2)) / sqrt(2π * s^2)
		static float inv_sqrt_2pi = 0.3989422804014327;
		float a = (x) / stdDev;

		return std::exp(-0.5f * a * a) * inv_sqrt_2pi / stdDev;
	};

	float density[255] = {0};
	float distance[255] = {0};
	float sumDen = 0;
	float stddev = blurSize / 3.7;
	distance[0] = blurSize % 2 ? 0 : 0.5;
	for (size_t i = 1 ; i < (blurSize + 1) / 2; i++)
		distance[i] = distance [i-1] + 1;
	for (size_t i=0; i < (blurSize + 1) / 2; i++)
	{
		density[i] += pdf(distance[i], stddev);
		sumDen += density[i];
	}
	sumDen *= 2;
	if (blurSize % 2)
		sumDen -= density[0];
	for (size_t i=0; i <= (blurSize + 1) / 2; i++)
		density[i] /= sumDen;

	size_t total = 1;
	unsigned updateBlurSize = blurSize;

	if (updateBlurSize % 2 == 0)
	{ // correction required for even sized kernels
		// we require kernel sized to always be odd, so we simulate an odd sized kernel with the following

		for (size_t i = (blurSize + 1) / 2; i > 0; i--)
		{
			density[i] = (density[i] + density[i - 1]) / 2;
		}
		//density[0] = (density[0] + density[-1]) / 2

		distance[0] = 0;
		updateBlurSize ++;
	}

	for (size_t i=1; i * 2 - 1 < (updateBlurSize + 1) / 2; i++)
	{
		distance[i] = (
					(i * 2 - 1) * density[i * 2 - 1] +
					(i * 2) * density[i * 2]
				) / (density[i * 2] + density[i * 2 - 1]);
		density[i] = density[i * 2] + density[i * 2 - 1];
		total ++;
	}

	{
		Vector<const char*> shaderSource;
		shaderSource.resize(total + 3);
		shaderSource[0] = shader_blur_fsh_tempalte[0];
		shaderSource[1] = shader_blur_fsh_tempalte[1];
		char tmp[255];
		snprintf(tmp, 255, "sum += texture2D(tex0, v_tex / u_textureSize) * %.6f; // u_distance[0] = 0\n", density[0]);
		shaderSource[2] = strdup(tmp);

		for (size_t i=1; i < total; i++)
		{

			snprintf(tmp, 255, "sum += blur(%.6f, %.6f);\n", distance[i], density[i]);
			shaderSource[i + 2] = strdup(tmp);
		}
		shaderSource.last() = shader_blur_fsh_tempalte[2];
		GLuint fsh = loadShader(shaderSource.data(), shaderSource.size(), GL_FRAGMENT_SHADER);
		if (!fsh)

		{
			//fprintf(stderr, "failed to initialize glBlurShader\n");
			//fflush(stderr);
			return false;
		}

		g_glBlurProgram[blurSize] = createProgram(g_glBlurVSH, fsh);
		glDeleteShader(fsh);
		for (size_t i=0; i < total; i++)
		{
			free(const_cast<char*>(shaderSource[i + 2]));
		}
	}

	if (g_glBlurProgram[blurSize] == 0)
	{
		//fprintf(stderr, "failed to initialize glBlurShader\n");
		//fflush(stderr);
		return false;
	}

	g_glBlurTemplateSizeLoc[blurSize] = glGetUniformLocation(g_glBlurProgram[blurSize], "u_templateSize");
	g_glBlurTextureSizeLoc[blurSize] = glGetUniformLocation(g_glBlurProgram[blurSize], "u_textureSize");
	g_glBlurDirectionLoc[blurSize] = glGetUniformLocation(g_glBlurProgram[blurSize], "u_direction");
	return true;
}

bool ShadowBlur::glBlurInitialize()
{
	//fprintf(stderr, "%s %d\n", __func__, g_glBlurInitialized);
	//fflush(stderr);
	if (g_glBlurInitialized)
		return true;

	int textureRestore;
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &textureRestore);
	glGenBuffers(1, &g_glBlurArrayBuffer);
	glGenTextures(1, &g_glBlurTexture);
	glBindTexture(GL_TEXTURE_2D, g_glBlurTexture);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glBindTexture(GL_TEXTURE_2D, textureRestore);
	glGenFramebuffers(1, &g_glBlurFrameBuffer);
	g_glBlurInitialized = true;

	return true;
}


bool ShadowBlur::glBlurShadowBuffer(GLuint texture, const IntSize &textureSize, const IntSize &templateSize, const IntSize& blurSize)
{

	constexpr float downsample = 1;

	unsigned int blurWidth = blurSize.width() * 1.5 * downsample;
	unsigned int blurHeight = blurSize.height() * 1.5;

	if (blurWidth >= g_glBlurProgram.size())
		blurWidth = g_glBlurProgram.size() - 1;
	if (blurHeight >= g_glBlurProgram.size())
		blurHeight = g_glBlurProgram.size() - 1;
	if (blurWidth == 0 && blurHeight == 0)
		return false;

	int arrayBufferRestore, frameBufferRestore, textureRestore, activeTextureRestore, programRestore, blendRestore;
	int viewportRestore[4];

	if (!g_glBlurInitialized && !glBlurInitialize())
		return false;

	glGetIntegerv(GL_CURRENT_PROGRAM, &programRestore);
	if (!glBlurCompileShader(blurWidth) || !glBlurCompileShader(blurHeight))
	{
	    glUseProgram(programRestore);
		return false;
	}

	glGetIntegerv(GL_VIEWPORT, viewportRestore);
	glViewport(0, 0, textureSize.width(), textureSize.height());
	blendRestore = glIsEnabled(GL_BLEND);
	glDisable(GL_BLEND);
	glUseProgram(g_glBlurProgram[blurWidth]);

	glGetIntegerv(GL_ARRAY_BUFFER_BINDING, &arrayBufferRestore);

	
	// We know cairo does not use array_buffers, 
	// also on sonybdr GL_ARRAY_BUFFER_BINDING does not return 0 so we hack it to 0;
	
	arrayBufferRestore = 0; 
	
	glBindBuffer(GL_ARRAY_BUFFER, g_glBlurArrayBuffer);
	float rectData[] = {
		-1, -1, 0, 0,
		-1, 1, 0, static_cast<float>(templateSize.height()),
		1, -1, static_cast<float>(templateSize.width()), 0,
		1, 1, static_cast<float>(templateSize.width()), static_cast<float>(templateSize.height()),
	};
	glBufferData(GL_ARRAY_BUFFER, sizeof(rectData), rectData, GL_STATIC_DRAW);
	glGetIntegerv(GL_FRAMEBUFFER_BINDING, &frameBufferRestore);
	glBindFramebuffer(GL_FRAMEBUFFER, g_glBlurFrameBuffer);
	glGetIntegerv(GL_ACTIVE_TEXTURE, &activeTextureRestore);

	glActiveTexture(GL_TEXTURE0);
	glGetIntegerv(GL_TEXTURE_BINDING_2D, &textureRestore);
	glBindTexture(GL_TEXTURE_2D, g_glBlurTexture);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, textureSize.width() * downsample, textureSize.height() * downsample, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
	glViewport(0, 0, templateSize.width() * downsample, templateSize.height() * downsample);
	glBindTexture(GL_TEXTURE_2D, texture);

	int texParamerRestoreWrapS, texParamerRestoreWrapT, texParamerRestoreMinFilter, texParamerRestoreMagFilter;
	glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, &texParamerRestoreMagFilter);
	glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, &texParamerRestoreMinFilter);
	glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, &texParamerRestoreWrapS);
	glGetTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, &texParamerRestoreWrapT);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, g_glBlurTexture, 0);
	//horizontal pass

	glUniform2f(g_glBlurDirectionLoc[blurWidth], 1.0, 0);
	glUniform2f(g_glBlurTemplateSizeLoc[blurWidth], templateSize.width(), templateSize.height());
	glUniform2f(g_glBlurTextureSizeLoc[blurWidth], textureSize.width(), textureSize.height());
	unsigned int maxAttribCount;
	glGetIntegerv(GL_MAX_VERTEX_ATTRIBS, (int*)&maxAttribCount);
	struct AttribInfo{
		int enabled;
		int size;
		int type;
		int normalized;
		int stride;
		int bufferBinding;
		void* pointer;
	};
	WTF::Vector<AttribInfo> vertexAttribState(maxAttribCount);

	GLuint edits[] = {g_glBlurPosLoc, g_glBlurTexLoc};
	for (size_t i=0 ; i < sizeof(edits) / sizeof(edits[0]); i++)
	{
		int attrib = edits[i];
		glGetVertexAttribiv(attrib, GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING, &vertexAttribState[attrib].bufferBinding);
		glGetVertexAttribiv(attrib, GL_VERTEX_ATTRIB_ARRAY_SIZE, &vertexAttribState[attrib].size);
		glGetVertexAttribiv(attrib, GL_VERTEX_ATTRIB_ARRAY_STRIDE, &vertexAttribState[attrib].stride);
		glGetVertexAttribiv(attrib, GL_VERTEX_ATTRIB_ARRAY_TYPE, &vertexAttribState[attrib].type);
		glGetVertexAttribiv(attrib, GL_VERTEX_ATTRIB_ARRAY_NORMALIZED, &vertexAttribState[attrib].normalized);
		glGetVertexAttribPointerv(attrib, GL_VERTEX_ATTRIB_ARRAY_POINTER, &vertexAttribState[attrib].pointer);
	}
	for (unsigned i=0; i < maxAttribCount; i++) {
		glGetVertexAttribiv(i, GL_VERTEX_ATTRIB_ARRAY_ENABLED, &vertexAttribState[i].enabled);
		glDisableVertexAttribArray(i);
	}

	glEnableVertexAttribArray(g_glBlurPosLoc);
	glEnableVertexAttribArray(g_glBlurTexLoc);
	glVertexAttribPointer(g_glBlurPosLoc, 2, GL_FLOAT, 0, sizeof(float) * 4, 0);
	glVertexAttribPointer(g_glBlurTexLoc, 2, GL_FLOAT, 0, sizeof(float) * 4, (void*)(sizeof(float) * 2));
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	// vertical pass
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, texture, 0);
	glBindTexture(GL_TEXTURE_2D, g_glBlurTexture);
	glViewport(0, 0, templateSize.width(), templateSize.height());

	if (blurWidth != blurHeight)
	{
		glUseProgram(g_glBlurProgram[blurHeight]);
		glUniform2f(g_glBlurTemplateSizeLoc[blurWidth], templateSize.width(), templateSize.height());
		glUniform2f(g_glBlurTextureSizeLoc[blurWidth], textureSize.width(), textureSize.height());
	}

	glUniform2f(g_glBlurDirectionLoc[blurHeight], 0, 1.0);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	//restoring state:

	for (unsigned i=0 ; i < sizeof(edits) / sizeof(edits[0]); i++)
	{
		int attrib = edits[i];
		glBindBuffer(GL_ARRAY_BUFFER, vertexAttribState[attrib].bufferBinding);
		glVertexAttribPointer(attrib,
							 vertexAttribState[attrib].size,
							 vertexAttribState[attrib].type,
							 vertexAttribState[attrib].normalized,
							 vertexAttribState[attrib].stride,
							 vertexAttribState[attrib].pointer);
	}

	for (unsigned i=0; i < maxAttribCount; i++) {
		if (vertexAttribState[i].enabled)
			glEnableVertexAttribArray(i);
		else
			glDisableVertexAttribArray(i);
	}

	glBindTexture(GL_TEXTURE_2D, texture);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, texParamerRestoreMinFilter);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, texParamerRestoreMagFilter);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, texParamerRestoreWrapS);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, texParamerRestoreWrapT);
	glBindTexture(GL_TEXTURE_2D, textureRestore);

	glActiveTexture(activeTextureRestore);

	glBindBuffer(GL_ARRAY_BUFFER, arrayBufferRestore);

	glBindFramebuffer(GL_FRAMEBUFFER, frameBufferRestore);

	glUseProgram(programRestore);
	if (blendRestore)
		glEnable(GL_BLEND);
	glViewport(viewportRestore[0], viewportRestore[1], viewportRestore[2], viewportRestore[3]);

	return true;
}

#endif

} // namespace WebCore
