/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2009 Dirk Schulze <krit@webkit.org>
 * Copyright (C) 2010 Holger Hans Peter Freyther
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "FontCascade.h"
#include "AffineTransform.h"
#include "GlyphBuffer.h"
#include "Gradient.h"
#include "GraphicsContext.h"
#include "PlatformContextCairo.h"
#include "ImageBuffer.h"
#include "Pattern.h"
#include "PlatformDelegate.h"
#include "FontPlatformData.h"
#include "Font.h"
#include "CairoUtilities.h"
#include "FileSystem.h"
#include "FontCache.h"
#include "Path.h"
#include "RefPtrCairo.h"
#include "SharedBuffer.h"
#include "TextEncoding.h"
#include <cairo-ft.h>
#include <cairo.h>
#include <stdio.h>
#include <wtf/HashMap.h>
#include <wtf/text/CString.h>
    
#define SYNTHETIC_OBLIQUE_ANGLE 14

namespace WebCore {

#if ENABLE(BML_DRCS)
static void drawDRCSText(GraphicsContext*, UChar32 c, const FloatPoint& startp, float fontComputedSize, const Font&);
static void addDRCSToMap(/*BMLDocument* doc, */const char* bytes, int data_size);
#endif

// cairo_save is called before, will call cairo_restore after.
static NEVER_INLINE void drawGlyphsToContext(cairo_t* cr, FontPlatformData* font, cairo_glyph_t* glyphs, unsigned numGlyphs, float syntheticBoldOffset)
{
    cairo_matrix_t originalTransform;
	cairo_get_matrix(cr, &originalTransform);

    bool didSetScaledFontToCairo = false;
    Vector<cairo_glyph_t> cairo_glyphs_copy;

    if (originalTransform.xy == 0 && originalTransform.yx == 0) {
		double max_scale = std::max(fabs(originalTransform.xx) + fabs(originalTransform.xy), fabs(originalTransform.yx) + fabs(originalTransform.yy));

		if (max_scale != 1.0 && !font->localFamilyName().isEmpty() && cairo_scaled_font_get_type(font->scaledFont()) == CAIRO_FONT_TYPE_USER) {
			FontDescription scaledFontDescription = font->fontDescription();
			scaledFontDescription.setComputedSize(max_scale * font->fontDescription().computedSize());

			if (RefPtr<Font> newFont = fontCache().fontForFamily(scaledFontDescription, font->localFamilyName())) {
				if (newFont->platformData().scaledFont()) {
					cairo_glyphs_copy.resize(numGlyphs);

					double yOffset = glyphs[0].y * max_scale;
					for (unsigned i = 0; i < numGlyphs; i++) {
						cairo_glyph_t& cairo_glyph = cairo_glyphs_copy[i];
						cairo_glyph.index = glyphs[i].index;
						cairo_glyph.x = glyphs[i].x * max_scale;
						cairo_glyph.y = yOffset;
					}

					cairo_matrix_t newTransform = { 1, 0, 0, 1, originalTransform.x0, originalTransform.y0 };
					cairo_set_matrix(cr, &newTransform);

					cairo_set_scaled_font(cr, newFont->platformData().scaledFont());

					didSetScaledFontToCairo = true;
				}
			}
		}
    }

    if (!didSetScaledFontToCairo) {
		cairo_set_scaled_font(cr, font->scaledFont());

		cairo_scaled_font_t* used_scaled_font = cairo_get_scaled_font(cr);

		if(used_scaled_font != font->scaledFont())
			font->cacheScaledFont(used_scaled_font);
    }

	cairo_glyph_t* used_glyphs = cairo_glyphs_copy.isEmpty() ? glyphs : cairo_glyphs_copy.data();

    cairo_show_glyphs(cr, used_glyphs, numGlyphs);

    if (syntheticBoldOffset) {
		cairo_translate(cr, syntheticBoldOffset, 0);
		cairo_show_glyphs(cr, used_glyphs, numGlyphs);
    }
}

static void drawGlyphsShadow(GraphicsContext* graphicsContext, const FloatPoint& point, FontPlatformData* font, cairo_glyph_t* glyphs, unsigned numGlyphs, float syntheticBoldOffset)
{
	ShadowState shadowState(graphicsContext->state());
	ShadowBlur shadow(FloatSize(shadowState.blur, shadowState.blur), shadowState.offset, shadowState.color, shadowState.ignoreTransforms);

    if (shadow.type() == ShadowBlur::NoShadow)
        return;

    if (!(graphicsContext->textDrawingMode() & TextModeFill))
        return;

    if (!shadowState.isRequired(*graphicsContext->platformContext())) {
        // Optimize non-blurry shadows, by just drawing text without the ShadowBlur.
        cairo_t* context = graphicsContext->platformContext()->cr();
        cairo_save(context);

        FloatSize shadowOffset(graphicsContext->state().shadowOffset);
        cairo_translate(context, shadowOffset.width(), shadowOffset.height());

		float red, green, blue, alpha;
        graphicsContext->state().shadowColor.getRGBA(red, green, blue, alpha);
        cairo_set_source_rgba(context, red, green, blue, alpha);

        drawGlyphsToContext(context, font, glyphs, numGlyphs, syntheticBoldOffset);

        cairo_restore(context);
        return;
    }

    cairo_text_extents_t extents;
    cairo_scaled_font_glyph_extents(font->scaledFont(), glyphs, numGlyphs, &extents);
    FloatRect fontExtentsRect(point.x() + extents.x_bearing, point.y() + extents.y_bearing, extents.width, extents.height);

    if (ImageBuffer* shadowLayerImage = shadow.beginShadowLayer(graphicsContext, fontExtentsRect)) {
        cairo_t* context = shadowLayerImage->context()->platformContext()->cr();
        cairo_save(context);
        drawGlyphsToContext(context, font, glyphs, numGlyphs, syntheticBoldOffset);
        cairo_restore(context);

		shadow.endShadowLayer(graphicsContext, shadowLayerImage);
    }
}

void FontPlatformData::drawGlyphs(GraphicsContext* context, const GlyphBuffer& glyphBuffer, unsigned from, unsigned numGlyphs, const FloatPoint& point, float syntheticBoldOffset)
{
	if (numGlyphs == 0)
		return;

    cairo_t* cr = context->platformContext()->cr();
	if(UNLIKELY(!m_scaledFont))
		return;

    Vector<cairo_glyph_t> cairo_glyphs(numGlyphs);

    auto xOffset = point.x();
    {
        auto* glyphsData = glyphBuffer.glyphs(from);
        auto* advances = glyphBuffer.advances(from);

        auto yOffset = point.y();
		for (unsigned i = 0; i < numGlyphs; i++) {
			cairo_glyph_t& cairo_glyph = cairo_glyphs[i];
			cairo_glyph.index = glyphsData[i];
			cairo_glyph.x = xOffset;
			cairo_glyph.y = yOffset;

			xOffset += advances[i];
		}
    }
    
    drawGlyphsShadow(context, point, this, cairo_glyphs.data(), numGlyphs, syntheticBoldOffset);

    if (context->textDrawingMode() & TextModeFill) {
        cairo_save(cr);
    	context->platformContext()->prepareForFilling(context, PlatformContextCairo::AdjustPatternForGlobalAlpha);
        drawGlyphsToContext(cr, this, cairo_glyphs.data(), numGlyphs, syntheticBoldOffset);
        cairo_restore(cr);
    }

    // Prevent running into a long computation within cairo. If the stroke width is
    // twice the size of the width of the text we will not ask cairo to stroke
    // the text as even one single stroke would cover the full wdth of the text.
    //  See https://bugs.webkit.org/show_bug.cgi?id=33759.
    if ((context->textDrawingMode() & TextModeStroke) && context->strokeThickness() < 2 * xOffset) {
        cairo_save(cr);
    	context->platformContext()->prepareForStroking(context);
        cairo_set_line_width(cr, context->strokeThickness());

        // This may disturb the CTM, but we are going to call cairo_restore soon after.
        cairo_set_scaled_font(cr, m_scaledFont.get());
        cairo_glyph_path(cr, cairo_glyphs.data(), numGlyphs);
        cairo_stroke(cr);
        cairo_restore(cr);
    }
}

#if ENABLE(BML_DRCS)
namespace {
struct DRCSImage {
	int fontId;
	int width;
	int height;
	Vector<unsigned char> pixels;
};

struct DRCSCharactor : public RefCounted<DRCSCharactor>
{
	static PassRefPtr<DRCSCharactor> create() { return adoptRef(new DRCSCharactor()); }

	Vector<DRCSImage> DRCSImages;

private:
	DRCSCharactor() { }
};

typedef HashMap<int32_t, RefPtr<DRCSCharactor> > DRCSMap;

static DRCSMap gDRCSImages;
}
#endif

void FontCascade::drawGlyphs(GraphicsContext* context, const TextRun& textRun, const Font* font, const GlyphBuffer& glyphBuffer, unsigned from, unsigned numGlyphs, const FloatPoint& point) const
{
#if ENABLE(BML_DRCS)
	if (gDRCSImages.isEmpty()) {
		if (RefPtr<SharedBuffer> sharedBuffer = SharedBuffer::createWithContentsOfFile("./drcs188.drc")) {
			addDRCSToMap(sharedBuffer->data(), sharedBuffer->size());
		}
	}

	const Glyph* glyphs = glyphBuffer.glyphs(from);

	float offsetX = point.x();
	float offsetY = point.y();

	float size = this->size();

	int sindex = 0;
	FloatPoint startp;
	while (sindex < numGlyphs) {
		while (sindex < numGlyphs) {
			UChar32 c = textRun[from + sindex];

			fprintf(stderr, "---->drawGlyphs %x (font size %f)\n", c, size);

			if (!c || !gDRCSImages.contains(c))
				break;

			startp.setX(offsetX);
			startp.setY(offsetY);
			drawDRCSText(context, c, startp, size, *font);
			offsetX += std::max(size, glyphBuffer.advanceAt(from + sindex));
			sindex++;
		}
		startp.setX(offsetX);

		int eindex, tempoffsetX = 0;
		for (eindex = sindex; eindex < numGlyphs; eindex++) {
			UChar32 c = textRun[eindex];
			if (!gDRCSImages.contains(c))
				tempoffsetX += glyphBuffer.advanceAt(from + eindex);
			else
				break;
		}

		if (eindex - sindex > 0) {
			startp.setX(offsetX);
			startp.setY(offsetY);
			const_cast<FontPlatformData&>(font->platformData()).drawGlyphs(context, glyphBuffer, from + sindex, eindex - sindex, startp, font->syntheticBoldOffset());
			offsetX += tempoffsetX;
		}

		sindex = eindex;
	}
#else
	const_cast<FontPlatformData&>(font->platformData()).drawGlyphs(context, glyphBuffer, from, numGlyphs, point, font->syntheticBoldOffset());
#endif
}

namespace {
struct GlyphIterationState {
    GlyphIterationState(FloatPoint startingPoint, FloatPoint currentPoint, float centerOfLine, float minX, float maxX)
        : startingPoint(startingPoint)
        , currentPoint(currentPoint)
        , centerOfLine(centerOfLine)
        , minX(minX)
        , maxX(maxX)
    {
    }
    FloatPoint startingPoint;
    FloatPoint currentPoint;
    float centerOfLine;
    float minX;
    float maxX;
};

static bool findIntersectionPoint(float y, FloatPoint p1, FloatPoint p2, float& x)
{
    x = p1.x() + (y - p1.y()) * (p2.x() - p1.x()) / (p2.y() - p1.y());
    return (p1.y() < y && p2.y() > y) || (p1.y() > y && p2.y() < y);
}

static void updateX(GlyphIterationState& state, float x)
{
    state.minX = std::min(state.minX, x);
    state.maxX = std::max(state.maxX, x);
}

// This function is called by Path::apply and is therefore invoked for each contour in a glyph. This
// function models each contours as a straight line and calculates the intersections between each
// pseudo-contour and the vertical center of the underline found in GlyphIterationState::centerOfLine.
// It keeps track of the leftmost and rightmost intersection in  GlyphIterationState::minX and
// GlyphIterationState::maxX.
static void findPathIntersections(void* stateAsVoidPointer, const PathElement& element)
{
    auto& state = *static_cast<GlyphIterationState*>(stateAsVoidPointer);
    bool doIntersection = false;
    FloatPoint point = FloatPoint();
    switch (element.type) {
    case PathElement::Type::MoveToPoint:
        state.startingPoint = element.points[0];
        state.currentPoint = element.points[0];
        break;
    case PathElement::Type::AddLineToPoint:
        doIntersection = true;
        point = element.points[0];
        break;
    case PathElement::Type::AddQuadCurveToPoint:
        doIntersection = true;
        point = element.points[1];
        break;
    case PathElement::Type::AddCurveToPoint:
        doIntersection = true;
        point = element.points[2];
        break;
    case PathElement::Type::CloseSubpath:
        doIntersection = true;
        point = state.startingPoint;
        break;
    }

    if (!doIntersection)
        return;

    float x;
    if (findIntersectionPoint(state.centerOfLine, state.currentPoint, point, x))
        updateX(state, x);

    state.currentPoint = point;
}

class CairoGlyphToPathTranslator FINAL : public GlyphToPathTranslator {
public:
    CairoGlyphToPathTranslator(const TextRun& textRun, const GlyphBuffer& glyphBuffer, const FloatPoint& textOrigin)
        : m_index(0)
        , m_textRun(textRun)
        , m_glyphBuffer(glyphBuffer)
        , m_font(glyphBuffer.fontAt(m_index))
        , m_translation(AffineTransform().translate(textOrigin.x(), textOrigin.y()))
    {
    }

    virtual bool containsMorePaths() OVERRIDE { return m_index != m_glyphBuffer.size(); }
    virtual Path path() OVERRIDE;
    virtual std::pair<float, float> extents() OVERRIDE;
    virtual GlyphUnderlineType underlineType() OVERRIDE;
    virtual void advance() OVERRIDE;

private:
    unsigned m_index;
    const TextRun& m_textRun;
    const GlyphBuffer& m_glyphBuffer;
    const Font* m_font;
    AffineTransform m_translation;
};

Path CairoGlyphToPathTranslator::path()
{
    RefPtr<cairo_t> cr = adoptRef(cairo_create(adoptRef(cairo_image_surface_create(CAIRO_FORMAT_A8, 1, 1)).get()));

    cairo_glyph_t cairoGlyph;
    cairoGlyph.index = m_glyphBuffer.glyphAt(m_index);
    cairo_set_scaled_font(cr.get(), m_font->platformData().scaledFont());
    cairo_glyph_path(cr.get(), &cairoGlyph, 1);

    float syntheticBoldOffset = m_font->syntheticBoldOffset();
    if (syntheticBoldOffset) {
        cairo_translate(cr.get(), syntheticBoldOffset, 0);
        cairo_show_glyphs(cr.get(), &cairoGlyph, 1);
    }

    Path path(cr.release());
    path.transform(m_translation);
    return path;
}

std::pair<float, float> CairoGlyphToPathTranslator::extents()
{
    FloatPoint beginning = m_translation.mapPoint(FloatPoint());
    GlyphBufferAdvance advance = m_glyphBuffer.advanceAt(m_index);
    FloatSize end = m_translation.mapSize(FloatSize(advance, advance));
    return std::make_pair(static_cast<float>(beginning.x()), static_cast<float>(beginning.x() + end.width()));
}

GlyphToPathTranslator::GlyphUnderlineType CairoGlyphToPathTranslator::underlineType()
{
    return computeUnderlineType(m_textRun, m_glyphBuffer, m_index);
}

void CairoGlyphToPathTranslator::advance()
{
	GlyphBufferAdvance advance = m_glyphBuffer.advanceAt(m_index);
	m_translation = m_translation.translate(advance, advance);
	++m_index;
	if (m_index < m_glyphBuffer.size())
		m_font = m_glyphBuffer.fontAt(m_index);
}
}

DashArray FontCascade::dashesForIntersectionsWithRect(const TextRun& run, const FloatPoint& textOrigin, const FloatRect& lineExtents) const
{
    if (isLoadingCustomFonts())
        return DashArray();

    OwnPtr<GlyphBuffer> glyphBuffer = adoptPtr(new GlyphBuffer());
    glyphBuffer->saveOffsetsInString();
    float deltaX;
    if (codePath(run) != FontCascade::Complex)
        deltaX = getGlyphsAndAdvancesForSimpleText(run, 0, run.length(), *glyphBuffer);
    else
        deltaX = getGlyphsAndAdvancesForComplexText(run, 0, run.length(), *glyphBuffer);

    if (!glyphBuffer->size())
        return DashArray();

    // FIXME: Handle SVG + non-SVG interleaved runs. https://bugs.webkit.org/show_bug.cgi?id=133778
    const Font* fontData = glyphBuffer->fontAt(0);
    FloatPoint origin = FloatPoint(textOrigin.x() + deltaX, textOrigin.y());
    CairoGlyphToPathTranslator translator(run, *glyphBuffer, origin);
    DashArray result;
    for (int index = 0; translator.containsMorePaths(); ++index, translator.advance()) {
        float centerOfLine = lineExtents.y() + (lineExtents.height() / 2);
        GlyphIterationState info = GlyphIterationState(FloatPoint(), FloatPoint(), centerOfLine, lineExtents.x() + lineExtents.width(), lineExtents.x());
        const Font* localFontData = glyphBuffer->fontAt(index);
        if (!localFontData) {
            // The advances will get all messed up if we do anything other than bail here.
            result.clear();
            break;
        }
        switch (translator.underlineType()) {
        case GlyphToPathTranslator::GlyphUnderlineType::SkipDescenders: {
            Path path = translator.path();
            path.apply(&info, &findPathIntersections);
            if (info.minX < info.maxX) {
                result.append(info.minX - lineExtents.x());
                result.append(info.maxX - lineExtents.x());
            }
            break;
        }
        case GlyphToPathTranslator::GlyphUnderlineType::SkipGlyph: {
            std::pair<float, float> extents = translator.extents();
            result.append(extents.first - lineExtents.x());
            result.append(extents.second - lineExtents.x());
            break;
        }
        case GlyphToPathTranslator::GlyphUnderlineType::DrawOverGlyph:
            // Nothing to do
            break;
        }
    }
    return result;
}

#if ENABLE(BML_DRCS)

static void addDRCSToMap(/*BMLDocument* doc, */const char* bytes, int data_size)
{
	if (data_size == 0)
		return;

	String charset; /* = doc->characterSet()*/;

	if(charset.length() == 0)
		charset = "EUC-JP";

	TextEncoding textEncoding(charset);

	int offset = 0;
	unsigned int numberOfCodes = static_cast<unsigned char>(bytes[offset++]);

	//if(WebEspial::gDebugMask & WebEspial::Debug_Mask_Loading)
	fprintf(stderr, "---->addDRCSToMap numberOfCodes %u, characterSet %s\n", numberOfCodes, charset.utf8().data());

	for (int codeIndex = 0; codeIndex < numberOfCodes; ++codeIndex) {
		if (data_size < offset + 3)
			break;

		int characterCode = static_cast<int>(bytes[offset] << 8 | bytes[offset + 1]) + 0x8080;
		int NumberOfFont = static_cast<int>(bytes[offset + 2]);
		offset += 3;

		RefPtr<DRCSCharactor> DRCSChars = DRCSCharactor::create();
		DRCSChars->DRCSImages.reserveCapacity(NumberOfFont);

		for (int fontIndex = 0; fontIndex < NumberOfFont; ++fontIndex){
			if (data_size < offset + 1)
				break;
			int fontId = static_cast<int>(bytes[offset] & 0xF0);
			int mode = static_cast<int>(bytes[offset] & 0x0F);
			++offset;
			if (data_size < offset + 3)
				break;
			if (mode == 0 || mode == 1){
				const int depth = static_cast<int>(bytes[offset]);
				const int width = static_cast<int>(bytes[offset + 1]);
				const int height = static_cast<int>(bytes[offset + 2]);
				offset += 3;
				if(depth != 2 || width == 0 || height == 0)
					break;

				DRCSImage drcsimage;
				drcsimage.fontId = fontId;
				drcsimage.height = height;
				drcsimage.width = width;
				drcsimage.pixels.resize(width * height);

				const int N = width * height >> 2;
				if (data_size < offset + N)
					break;

				int wIndex = 0;
				int hIndex = 0;

				for (int byteIndex = 0; byteIndex < N; ++byteIndex) {
					const char pixel = bytes[offset];
					for (int m = 6; m >= 0; m -=2) {
					   drcsimage.pixels[hIndex * width + wIndex] = (pixel >> m) & 0x03;

					   ++hIndex;
					   if (hIndex == height) {
						   hIndex = 0;
						   ++wIndex;
					   }
					}

					++offset;
				}

				DRCSChars->DRCSImages.append(drcsimage);
			}
		}

		//if(WebEspial::gDebugMask & WebEspial::Debug_Mask_Font)
		fprintf(stderr, "---->addDRCSToMap %4x\n", characterCode);

		gDRCSImages.set((int32_t)characterCode, DRCSChars);
	}
}

static void drawDRCSText(GraphicsContext* context, UChar32 c, const FloatPoint& startp, float fontComputedSize, const Font& font)
{
	RefPtr<DRCSCharactor> DRCSCharactor = gDRCSImages.get(c);
	if(!DRCSCharactor) {
		//if(WebEspial::gDebugMask & WebEspial::Debug_Mask_Font)
		fprintf(stderr, "---->drawDRCSText %x not found\n", c);
		return;
	}

	const auto& DRCSImages = DRCSCharactor->DRCSImages;

	Color fillColor = context->fillColor();

	float min = std::abs(DRCSImages[0].width - fontComputedSize);
	int min_index = 0;

	for (unsigned DRCSindex = 0; DRCSindex < DRCSImages.size(); DRCSindex++) {
		if (min >= std::abs(DRCSImages[DRCSindex].width - fontComputedSize)) {
			min = std::abs(DRCSImages[DRCSindex].width - fontComputedSize);
			min_index = DRCSindex;
		}
	}

	auto* pixels = DRCSImages[min_index].pixels.data();
	int width = DRCSImages[min_index].width;
	int height = DRCSImages[min_index].height;

	cairo_t* cr = context->platformContext()->cr();
	cairo_save(cr);

	float startX = startp.x();
	float startY = startp.y();
	startY = startY - height;

	float red, green, blue, alpha;
	fillColor.getRGBA(red, green, blue, alpha);

	//if(WebEspial::gDebugMask & WebEspial::Debug_Mask_Font)
	fprintf(stderr, "---->drawDRCSText %x with color %x (font size %f DRCS image %ix%i)\n", c, fillColor.rgb(), fontComputedSize, width, height);

    float syntheticBoldOffset = font.syntheticBoldOffset();

	double lineWidth = (double)fontComputedSize / width;

	cairo_set_line_width(cr, lineWidth);

	for (int i = 0; i < width; ++i) {
		for (int j = 0; j < height; ++j) {
			float adjustedAlpha = alpha * pixels[i * width + j] / 3.0;
			if (adjustedAlpha == 0)
				continue;
			cairo_set_source_rgba(cr, red, green, blue, 1.0);

			float x = startX + i * lineWidth;
			float y = startY + j * lineWidth;

			cairo_move_to(cr, x, y);
			cairo_line_to(cr, x, y + lineWidth);
			cairo_stroke(cr);

			if (syntheticBoldOffset) {
				cairo_move_to(cr, x + lineWidth, y);
				cairo_line_to(cr, x + lineWidth, y + lineWidth);
				cairo_stroke(cr);
			}
		}
	}

	cairo_restore(cr);
}

#endif

}
