/*
 * Copyright (C) 2006 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2007 Holger Hans Peter Freyther <zecke@selfish.org>
 * Copyright (C) 2008, 2009 Dirk Schulze <krit@webkit.org>
 * Copyright (C) 2010 Torch Mobile (Beijing) Co. Ltd. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "ImageBuffer.h"

#include "BitmapImage.h"
#include "CairoUtilities.h"
#include "Color.h"
#include "GraphicsContext.h"
#include "ImageData.h"
#include "MIMETypeRegistry.h"
#include "NotImplemented.h"
#include "Pattern.h"
#include "PlatformContextCairo.h"
#include "PlatformDelegate.h"
#include "PlatformString.h"
#include "RefPtrCairo.h"
#include <cairo.h>
#include <jpeglib.h>
#include <runtime/Operations.h>
#include <runtime/TypedArrayInlines.h>
#include <wtf/Vector.h>
#include <wtf/text/Base64.h>

using namespace std;

namespace WebCore {

inline PassOwnPtr<ImageBufferData> ImageBufferData::create(const IntSize& size, RenderingMode renderingMode, LogGLMemoryError logGLMemoryError)
{
	return platformDelegate().createImageBufferData(size, renderingMode, logGLMemoryError);
}

inline PassOwnPtr<ImageBufferData> ImageBufferData::create(const IntSize& size, bool forceImageSurface)
{
	return platformDelegate().createImageBufferData(size, forceImageSurface);
}

inline PassOwnPtr<ImageBufferData> ImageBufferData::create(const ImageBufferData& other, const IntSize& size, LogGLMemoryError logGLMemoryError)
{
	return platformDelegate().createImageBufferData(other, size, logGLMemoryError);
}

ImageBuffer::ImageBuffer(const FloatSize& size, float /* resolutionScale */, ColorSpace, RenderingMode renderingMode, LogGLMemoryError logGLMemoryError,  bool& success)
    : m_data(ImageBufferData::create(IntSize(size), renderingMode, logGLMemoryError))
    , m_size(size)
    , m_logicalSize(size)
	, m_resolutionScale(1.0f)
	, m_hasAlpha(true)
{
    success = false;  // Make early return mean error.

    if(!m_data->m_surface)
		return;

    if (cairo_surface_status(m_data->m_surface) != CAIRO_STATUS_SUCCESS)
        return;  // create will notice we didn't set m_initialized and fail.

    RefPtr<cairo_t> cr = adoptRef(cairo_create(m_data->m_surface));
    m_data->m_platformContext.setCr(cr.get());
    m_context = adoptPtr(new GraphicsContext(&m_data->m_platformContext));
    m_context->setIsAcceleratedContext(renderingMode == RenderingMode::Accelerated);
    platformDelegate().setupGraphicsContextProperties(*m_context);
    success = true;
}

ImageBuffer::ImageBuffer(const FloatSize& size, bool forceImageSurface, bool& success)
    : m_data(ImageBufferData::create(IntSize(size), forceImageSurface))
    , m_size(size)
    , m_logicalSize(size)
	, m_resolutionScale(1.0f)
	, m_hasAlpha(true)
{
    success = false;  // Make early return mean error.

    if(!m_data->m_surface)
		return;

    if (cairo_surface_status(m_data->m_surface) != CAIRO_STATUS_SUCCESS)
        return;  // create will notice we didn't set m_initialized and fail.

    RefPtr<cairo_t> cr = adoptRef(cairo_create(m_data->m_surface));
    m_data->m_platformContext.setCr(cr.get());
    m_context = adoptPtr(new GraphicsContext(&m_data->m_platformContext));
    m_context->setIsAcceleratedContext(false);
    platformDelegate().setupGraphicsContextProperties(*m_context);
    success = true;
}

ImageBuffer::ImageBuffer(const ImageBuffer& other, RenderingMode renderingMode, LogGLMemoryError logGLMemoryError, bool& success)
    : m_data(ImageBufferData::create(*other.m_data, IntSize(other.m_size), logGLMemoryError))
    , m_size(other.m_size)
    , m_logicalSize(other.m_size)
	, m_resolutionScale(1.0f)
	, m_hasAlpha(true)
{
    success = false;  // Make early return mean error.

    if(!m_data->m_surface)
		return;

    if (cairo_surface_status(m_data->m_surface) != CAIRO_STATUS_SUCCESS)
        return;  // create will notice we didn't set m_initialized and fail.

    RefPtr<cairo_t> cr = adoptRef(cairo_create(m_data->m_surface));
    m_data->m_platformContext.setCr(cr.get());
    m_context = adoptPtr(new GraphicsContext(&m_data->m_platformContext));
    m_context->setIsAcceleratedContext(renderingMode == RenderingMode::Accelerated);
    platformDelegate().setupGraphicsContextProperties(*m_context);
    success = true;
}

ImageBuffer::~ImageBuffer()
{
}

void ImageBuffer::didDraw()
{
	m_data->m_needFlush = true;
}

void ImageBuffer::flush()
{
	if(m_data->m_needFlush){
		if(m_data->m_surface)
			cairo_surface_flush(m_data->m_surface);
		m_data->m_needFlush = false;
	}
}

PassRefPtr<Image> ImageBuffer::copyImage(BackingStoreCopy copyBehavior) const
{
    if (copyBehavior == CopyBackingStore) {
    	if (auto* cairoSurface = platformDelegate().copySurface(m_data->m_surface))
    		return BitmapImage::create(adoptRef(cairoSurface), nullptr, m_hasAlpha);
    	return nullptr;
    }

    // BitmapImage will release the passed in surface on destruction
    return BitmapImage::create(adoptRef(cairo_surface_reference(m_data->m_surface)), 0, m_hasAlpha);
}

void ImageBuffer::clip(GraphicsContext* context, const FloatRect& maskRect) const
{
    context->platformContext()->pushImageMask(m_data->m_surface, maskRect);
}

void ImageBuffer::draw(GraphicsContext* destinationContext, const FloatRect& destRect, const FloatRect& srcRect,
                       CompositeOperator op, BlendMode blendMode, bool useLowQualityScale)
{
    BackingStoreCopy copyMode = destinationContext == context() ? CopyBackingStore : DontCopyBackingStore;
    if (RefPtr<Image> image = copyImage(copyMode))
    	destinationContext->drawImage(image.get(), destRect, srcRect, op, blendMode, DoNotRespectImageOrientation, useLowQualityScale);
}

void ImageBuffer::drawPattern(GraphicsContext* context, const FloatRect& srcRect, const AffineTransform& patternTransform,
	  const FloatPoint& phase, const FloatSize& spacing, CompositeOperator op, const FloatRect& destRect, BlendMode)
{
    // BitmapImage will release the passed in surface on destruction
    if (RefPtr<Image> image = copyImage(DontCopyBackingStore))
    	image->drawPattern(context, srcRect, patternTransform, phase, spacing, op, destRect);
}

void ImageBuffer::platformTransformColorSpace(const Vector<int>& lookUpTable)
{
    unsigned char* dataSrc = NULL;
    int stride = 0;

    if(!platformDelegate().lockSurface(m_data->m_surface, &dataSrc, &stride, false))
    	return;

	if(!dataSrc)
		return;

    for (int y = 0; y < m_size.height(); ++y) {
        unsigned* row = reinterpret_cast<unsigned*>(dataSrc + stride * y);
        for (int x = 0; x < m_size.width(); x++) {
            unsigned* pixel = row + x;
            Color pixelColor = colorFromPremultipliedARGB(*pixel);
            pixelColor = Color(lookUpTable[pixelColor.red()],
                               lookUpTable[pixelColor.green()],
                               lookUpTable[pixelColor.blue()],
                               pixelColor.alpha());
            *pixel = premultipliedARGBFromColor(pixelColor);
        }
    }

    cairo_surface_mark_dirty_rectangle(m_data->m_surface, 0, 0, m_size.width(), m_size.height());
    cairo_surface_flush(m_data->m_surface);

    platformDelegate().unlockSurface(m_data->m_surface, true);
}

template <Multiply multiplied>
PassRefPtr<Uint8ClampedArray> getImageData(const IntRect& rect, const ImageBufferData& data, const IntSize& size)
{
    RefPtr<Uint8ClampedArray> result = Uint8ClampedArray::createUninitialized(rect.width() * rect.height() * 4);

    if(!result)
    	return nullptr;

    unsigned char* dataDst = result->data();

    if (rect.x() < 0 || rect.y() < 0 || (rect.x() + rect.width()) > size.width() || (rect.y() + rect.height()) > size.height())
        result->zeroFill();

    int originx = rect.x();
    int destx = 0;
    if (originx < 0) {
        destx = -originx;
        originx = 0;
    }
    int endx = rect.maxX();
    if (endx > size.width())
        endx = size.width();
    int numColumns = endx - originx;

    int originy = rect.y();
    int desty = 0;
    if (originy < 0) {
        desty = -originy;
        originy = 0;
    }
    int endy = rect.maxY();
    if (endy > size.height())
        endy = size.height();
    int numRows = endy - originy;

    unsigned char* dataSrc = NULL;
    int stride = 0;

    if(!platformDelegate().lockSurface(data.m_surface, &dataSrc, &stride, false))
    	return nullptr;

	if(!dataSrc)
		return nullptr;

    unsigned destBytesPerRow = 4 * rect.width();

    unsigned char* destRows = dataDst + desty * destBytesPerRow + destx * 4;
    for (int y = 0; y < numRows; ++y) {
        unsigned* row = reinterpret_cast<unsigned*>(dataSrc + stride * (y + originy));
        for (int x = 0; x < numColumns; x++) {
            int basex = x * 4;
            unsigned* pixel = row + x + originx;
            Color pixelColor;
            if (multiplied == Unmultiplied)
                pixelColor = colorFromPremultipliedARGB(*pixel);
            else
                pixelColor = Color(*pixel);
#if CAIRO_USE_ARGB
            destRows[basex]     = pixelColor.red();
            destRows[basex + 1] = pixelColor.green();
            destRows[basex + 2] = pixelColor.blue();
#else
            destRows[basex]     = pixelColor.blue();
            destRows[basex + 1] = pixelColor.green();
            destRows[basex + 2] = pixelColor.red();
#endif
            destRows[basex + 3] = pixelColor.alpha();
        }
        destRows += destBytesPerRow;
    }

    platformDelegate().unlockSurface(data.m_surface, false);

    return result.release();
}

PassRefPtr<Uint8ClampedArray> ImageBuffer::getUnmultipliedImageData(const IntRect& rect, CoordinateSystem) const
{
    return getImageData<Unmultiplied>(rect, *m_data, m_size);
}

PassRefPtr<Uint8ClampedArray> ImageBuffer::getPremultipliedImageData(const IntRect& rect, CoordinateSystem) const
{
    return getImageData<Premultiplied>(rect, *m_data, m_size);
}

void ImageBuffer::putByteArray(Multiply multiplied, Uint8ClampedArray* source, const IntSize& sourceSize, const IntRect& sourceRect, const IntPoint& destPoint, CoordinateSystem)
{
    ASSERT(sourceRect.width() > 0);
    ASSERT(sourceRect.height() > 0);

    int originx = sourceRect.x();
    int destx = destPoint.x() + sourceRect.x();
    ASSERT(destx >= 0);
    ASSERT(destx < m_size.width());
    ASSERT(originx >= 0);
    ASSERT(originx <= sourceRect.maxX());

    int endx = destPoint.x() + sourceRect.maxX();
    ASSERT(endx <= m_size.width());

    int numColumns = endx - destx;

    int originy = sourceRect.y();
    int desty = destPoint.y() + sourceRect.y();
    ASSERT(desty >= 0);
    ASSERT(desty < m_size.height());
    ASSERT(originy >= 0);
    ASSERT(originy <= sourceRect.maxY());

    int endy = destPoint.y() + sourceRect.maxY();
    ASSERT(endy <= m_size.height());
    int numRows = endy - desty;

    unsigned srcBytesPerRow = 4 * sourceSize.width();

    unsigned char* dataDst = NULL;
    int stride = 0;

    if(!platformDelegate().lockSurface(m_data->m_surface, &dataDst, &stride, true))
    	return;

	if(!dataDst)
		return;

    unsigned char* srcRows = source->data() + originy * srcBytesPerRow + originx * 4;
    for (int y = 0; y < numRows; ++y) {
        unsigned* row = reinterpret_cast<unsigned*>(dataDst + stride * (y + desty));
        for (int x = 0; x < numColumns; x++) {
            int basex = x * 4;
            unsigned* pixel = row + x + destx;
#if CAIRO_USE_ARGB
            Color pixelColor = Color::createUnCheked(srcRows[basex],
                                                     srcRows[basex + 1],
                                                     srcRows[basex + 2],
                                                     srcRows[basex + 3]);
#else
            Color pixelColor = Color::createUnCheked(srcRows[basex + 2],
                                                     srcRows[basex + 1],
                                                     srcRows[basex],
                                                     srcRows[basex + 3]);
#endif
            if (multiplied == Unmultiplied)
                *pixel = premultipliedARGBFromColor(pixelColor);
            else
                *pixel = pixelColor.rgb();
        }
        srcRows += srcBytesPerRow;
    }

    cairo_surface_mark_dirty_rectangle (m_data->m_surface, destx, desty, numColumns, numRows);
    cairo_surface_flush (m_data->m_surface);

    platformDelegate().unlockSurface(m_data->m_surface, true);
}

static cairo_status_t writeFunction(void* output, const unsigned char* data, unsigned int length)
{
    if (!reinterpret_cast<Vector<uint8_t>*>(output)->tryAppend(data, length))
        return CAIRO_STATUS_WRITE_ERROR;
    return CAIRO_STATUS_SUCCESS;
}

#if JPEG_SUPPORT_MEM_DST && defined LIBJPEG_TURBO_VERSION_NUMBER
/*! This function creates a JPEG file in memory from a Cairo image surface.
 * @param sfc Pointer to a Cairo surface. It should be an image surface of
 * either CAIRO_FORMAT_ARGB32 or CAIRO_FORMAT_RGB24 or CAIRO_FORMAT_ABGR32. Other formats are
 * converted to CAIRO_FORMAT_RGB24 before compression.
 * Please note that this may give unexpected results because JPEG does not
 * support transparency. Thus, default background color is used to replace
 * transparent regions. The default background color is black if not specified
 * explicitly. Thus converting e.g. PDF surfaces without having any specific
 * background color set will apear with black background and not white as you
 * might expect. In such cases it is suggested to manually convert the surface
 * to RGB24 before calling this function.
 * @param data Pointer to a memory pointer. This parameter receives a pointer
 * to the memory area where the final JPEG data is found in memory. This
 * function reservers the memory properly and it has to be freed by the caller
 * with free(3).
 * @param len Pointer to a variable of type size_t which will receive the final
 * lenght of the memory buffer.
 * @param quality Compression quality, 0-100.
 * @return On success the function returns CAIRO_STATUS_SUCCESS. In case of
 * error CAIRO_STATUS_INVALID_FORMAT is returned.
 */
static cairo_status_t cairo_image_surface_write_to_jpeg_mem(cairo_surface_t *sfc, unsigned char **data, unsigned long *len, Optional<double> quality)
{
   struct jpeg_compress_struct cinfo;
   struct jpeg_error_mgr jerr;
   JSAMPROW row_pointer[1];
   cairo_surface_t *other = NULL;

   cairo_format_t src_cairo_format = cairo_image_surface_get_format(sfc);

   // check valid input format (must be IMAGE_SURFACE && (ARGB32 || RGB24 || ABGR32))
   if (cairo_surface_get_type(sfc) != CAIRO_SURFACE_TYPE_IMAGE ||
	 (src_cairo_format != CAIRO_FORMAT_ARGB32 && src_cairo_format != CAIRO_FORMAT_RGB24 && src_cairo_format != CAIRO_FORMAT_ABGR32))
   {
      // create a similar surface with a proper format if supplied input format
      // does not fulfill the requirements
      double x1, y1, x2, y2;
      other = sfc;
      cairo_t *ctx = cairo_create(other);
      // get extents of original surface
      cairo_clip_extents(ctx, &x1, &y1, &x2, &y2);
      cairo_destroy(ctx);

      // create new image surface
      sfc = cairo_surface_create_similar_image(other, CAIRO_FORMAT_RGB24, x2 - x1, y2 - y1);
      if (cairo_surface_status(sfc) != CAIRO_STATUS_SUCCESS)
         return CAIRO_STATUS_INVALID_FORMAT;

      // paint original surface to new surface
      ctx = cairo_create(sfc);
      cairo_set_source_surface(ctx, other, 0, 0);
      cairo_paint(ctx);
      cairo_destroy(ctx);
   }

   // finish queued drawing operations
   cairo_surface_flush(sfc);

   // init jpeg compression structures
   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_compress(&cinfo);

   // set compression parameters
   jpeg_mem_dest(&cinfo, data, len);
   cinfo.image_width = cairo_image_surface_get_width(sfc);
   cinfo.image_height = cairo_image_surface_get_height(sfc);
#if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
   cinfo.in_color_space = src_cairo_format != CAIRO_FORMAT_ABGR32 ? JCS_EXT_BGRX : JCS_EXT_RGBX;
#else
   cinfo.in_color_space = JCS_EXT_XRGB;
#endif
   cinfo.input_components = 4;
   jpeg_set_defaults(&cinfo);
   if(quality)
	   jpeg_set_quality(&cinfo, quality.value() * 100, 1 /*TRUE*/);

   // start compressor
   jpeg_start_compress(&cinfo, 1 /*TRUE*/);

   // loop over all lines and compress
   while (cinfo.next_scanline < cinfo.image_height)
   {
      row_pointer[0] = cairo_image_surface_get_data(sfc) + (cinfo.next_scanline
            * cairo_image_surface_get_stride(sfc));
      (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
   }

   // finalize and close everything
   jpeg_finish_compress(&cinfo);
   jpeg_destroy_compress(&cinfo);

   // destroy temporary image surface (if available)
   if (other != NULL)
      cairo_surface_destroy(sfc);

   return CAIRO_STATUS_SUCCESS;
}
#endif

static bool encodeImage(cairo_surface_t* image, const String& mimeType, Optional<double> quality, Vector<uint8_t>* output)
{
#if JPEG_SUPPORT_MEM_DST && defined LIBJPEG_TURBO_VERSION_NUMBER
    if (mimeType == "image/jpeg") {
    	unsigned char *data = 0;
    	unsigned long len = 0;

    	if(quality && ( *quality < 0.0 || *quality > 1.0))
    		quality = Nullopt;

    	cairo_status_t status = cairo_image_surface_write_to_jpeg_mem(image, &data, &len, quality);
    	if(status == CAIRO_STATUS_SUCCESS) {
    	    bool success = output->tryAppend(data, len);
    		::free(data);
    		return success;
    	}
    }
#endif

    ASSERT(mimeType == "image/png"); // Only PNG output is supported for now.
    return cairo_surface_write_to_png_stream(image, writeFunction, output) == CAIRO_STATUS_SUCCESS;
}

String ImageBuffer::toDataURL(const String& mimeType, Optional<double> quality, CoordinateSystem) const
{
    Vector<uint8_t> encodedImage = toData(mimeType, quality);
    if (encodedImage.isEmpty())
        return "data:,";

    Vector<char> base64Data;
    base64Encode(encodedImage.data(), encodedImage.size(), base64Data);

    return "data:" + mimeType + ";base64," + base64Data;
}

Vector<uint8_t> ImageBuffer::toData(const String& mimeType, Optional<double> quality) const
{
    ASSERT(MIMETypeRegistry::isSupportedImageMIMETypeForEncoding(mimeType));

    cairo_surface_t* image = cairo_get_target(context()->platformContext()->cr());

    Vector<uint8_t> encodedImage;
    if (!image || !encodeImage(image, mimeType, quality, &encodedImage))
        return Vector<uint8_t>();

    return encodedImage;
}

} // namespace WebCore
