/*
 * Copyright (C) 2011 Igalia S.L.
 * Copyright (c) 2008, Google Inc. All rights reserved.
 * Copyright (c) 2012, Intel Corporation
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "CairoUtilities.h"
#include "Gradient.h"
#include "GraphicsContext.h"
#include "ImageBuffer.h"
#include "PlatformContextCairo.h"
#include "PlatformDelegate.h"
#include "OwnPtrCairo.h"
#include "Pattern.h"
#include <cairo.h>

#if ENABLE(WEBGL)
#include "ImageDecoder.h"
#include "GraphicsContext3D.h"
#endif

namespace WebCore {

void PlatformContextCairo::restore()
{
    if (m_state->m_mask.pattern) {
        cairo_pop_group_to_source(m_cr.get());

        cairo_matrix_t matrix;
        cairo_get_matrix(m_cr.get(), &matrix);
        cairo_set_matrix(m_cr.get(), &m_state->m_mask.matrix);
        cairo_mask(m_cr.get(), m_state->m_mask.pattern.get());
        cairo_set_matrix(m_cr.get(), &matrix);
    }

    m_stateStack.removeLast();
    ASSERT(!m_stateStack.isEmpty());
    m_state = &m_stateStack.last();

    cairo_restore(m_cr.get());
}

void PlatformContextCairo::save()
{
    m_stateStack.append(State(*m_state));
    m_state = &m_stateStack.last();

    cairo_save(m_cr.get());
}

void PlatformContextCairo::pushImageMask(cairo_surface_t* surface, const FloatRect& rect)
{
    // We must call savePlatformState at least once before we can use image masking,
    // since we actually apply the mask in restorePlatformState.
    ASSERT(!m_stateStack.isEmpty());
    m_state->m_mask.pattern = adoptRef(cairo_pattern_create_for_surface(surface));
    cairo_get_matrix(m_cr.get(), &m_state->m_mask.matrix);

    cairo_matrix_t matrix;
    cairo_matrix_init_translate(&matrix, -rect.x(), -rect.y());
    cairo_pattern_set_matrix(m_state->m_mask.pattern.get(), &matrix);

    cairo_push_group(m_cr.get());
}

static void drawPatternToCairoContext(cairo_t* cr, cairo_pattern_t* pattern, const FloatRect& destRect, float alpha)
{
    cairo_translate(cr, destRect.x(), destRect.y());
    cairo_set_source(cr, pattern);
    cairo_rectangle(cr, 0, 0, destRect.width(), destRect.height());
    cairo_clip(cr);
    cairo_paint_with_alpha(cr, std::max<float>(0, std::min<float>(1.0, alpha)));
}

void PlatformContextCairo::drawSurfaceToContext(cairo_surface_t* surface, const FloatRect& destRect, const FloatRect& originalSrcRect, GraphicsContext* context)
{
    // Avoid invalid cairo matrix with small values.
    if (std::fabs(destRect.width()) < 0.5f || std::fabs(destRect.height()) < 0.5f)
        return;

    FloatRect srcRect = originalSrcRect;

    // We need to account for negative source dimensions by flipping the rectangle.
    if (originalSrcRect.width() < 0) {
        srcRect.setX(originalSrcRect.x() + originalSrcRect.width());
        srcRect.setWidth(std::fabs(originalSrcRect.width()));
    }
    if (originalSrcRect.height() < 0) {
        srcRect.setY(originalSrcRect.y() + originalSrcRect.height());
        srcRect.setHeight(std::fabs(originalSrcRect.height()));
    }

    RefPtr<cairo_surface_t> patternSurface = surface;
    float leftPadding = 0;
    float topPadding = 0;
    if (srcRect.x() || srcRect.y() || srcRect.size() != cairoSurfaceSize(surface)) {
        // Cairo subsurfaces don't support floating point boundaries well, so we expand the rectangle.
        IntRect expandedSrcRect(enclosingIntRect(srcRect));

        // We use a subsurface here so that we don't end up sampling outside the originalSrcRect rectangle.
        // See https://bugs.webkit.org/show_bug.cgi?id=58309
        patternSurface = adoptRef(cairo_surface_create_for_rectangle(surface, expandedSrcRect.x(),
            expandedSrcRect.y(), expandedSrcRect.width(), expandedSrcRect.height()));

        leftPadding = static_cast<float>(expandedSrcRect.x()) - floorf(srcRect.x());
        topPadding = static_cast<float>(expandedSrcRect.y()) - floorf(srcRect.y());
    }

    RefPtr<cairo_pattern_t> pattern = adoptRef(cairo_pattern_create_for_surface(patternSurface.get()));

    ASSERT(m_state);
	WebCore::setImageInterpolationQuality(pattern.get(), m_state->m_imageInterpolationQuality);
    cairo_pattern_set_extend(pattern.get(), CAIRO_EXTEND_PAD);

    // The pattern transformation properly scales the pattern for when the source rectangle is a
    // different size than the destination rectangle. We also account for any offset we introduced
    // by expanding floating point source rectangle sizes. It's important to take the absolute value
    // of the scale since the original width and height might be negative.
    float scaleX = std::fabs(srcRect.width() / destRect.width());
    float scaleY = std::fabs(srcRect.height() / destRect.height());
    cairo_matrix_t matrix = { scaleX, 0, 0, scaleY, leftPadding, topPadding };
    cairo_pattern_set_matrix(pattern.get(), &matrix);

	ShadowState shadowState(context->state());
	ShadowBlur shadow(FloatSize(shadowState.blur, shadowState.blur), shadowState.offset, shadowState.color, shadowState.ignoreTransforms);
    if (shadow.type() != ShadowBlur::NoShadow) {
        if (ImageBuffer* shadowLayerImage = shadow.beginShadowLayer(context, destRect)) {
            drawPatternToCairoContext(shadowLayerImage->context()->platformContext()->cr(), pattern.get(), destRect, 1);
            shadow.endShadowLayer(context, shadowLayerImage);
        }
    }

    cairo_save(m_cr.get());
    drawPatternToCairoContext(m_cr.get(), pattern.get(), destRect, globalAlpha());
    cairo_restore(m_cr.get());
}

void PlatformContextCairo::setImageInterpolationQuality(InterpolationQuality quality)
{
    ASSERT(m_state);
    m_state->m_imageInterpolationQuality = quality;
}

InterpolationQuality PlatformContextCairo::imageInterpolationQuality() const
{
    ASSERT(m_state);
    return m_state->m_imageInterpolationQuality;
}

static inline void reduceSourceByAlpha(cairo_t* cr, float alpha)
{
    if (alpha >= 1)
        return;
    cairo_push_group(cr);
    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
    cairo_paint_with_alpha(cr, alpha);
    cairo_pop_group_to_source(cr);
}

static void prepareCairoContextSource(GraphicsContext* context, cairo_t* cr, Pattern* pattern, Gradient* gradient, const Color& color, float globalAlpha, const AffineTransform& gradientSpaceTransform)
{
    if (pattern) {
        if (RefPtr<cairo_pattern_t> cairoPattern = adoptRef(pattern->createPlatformPattern(AffineTransform()))) {
        	setImageInterpolationQuality(cairoPattern.get(), context->imageInterpolationQuality());
        	cairo_set_source(cr, cairoPattern.get());
        	reduceSourceByAlpha(cr, globalAlpha);
        } else
        	setSourceRGBAFromColor(cr, Color::transparent);
    } else if (gradient) {
        RefPtr<cairo_pattern_t> cairoPattern = adoptRef(gradient->createPlatformGradient(globalAlpha, gradientSpaceTransform));
    	setImageInterpolationQuality(cairoPattern.get(), context->imageInterpolationQuality());
        cairo_set_source(cr, cairoPattern.get());
    } else { // Solid color source.
        if (globalAlpha < 1)
            setSourceRGBAFromColor(cr, colorWithOverrideAlpha(color.rgb(), color.alpha() / 255.f * globalAlpha));
        else
            setSourceRGBAFromColor(cr, color);
    }
}

void PlatformContextCairo::prepareForFilling(GraphicsContext* context, PatternAdjustment patternAdjustment)
{
	auto& state = context->state();
    cairo_set_fill_rule(m_cr.get(), state.fillRule == RULE_EVENODD ?  CAIRO_FILL_RULE_EVEN_ODD : CAIRO_FILL_RULE_WINDING);
    prepareCairoContextSource(context, m_cr.get(),
		state.fillPattern.get(),
		state.fillGradient.get(),
		state.fillColor,
		patternAdjustment == AdjustPatternForGlobalAlpha ? globalAlpha() : 1,
		state.fillGradientSpaceTransform);
    
    if (state.fillPattern)
        clipForPatternFilling(state);
}

void PlatformContextCairo::prepareForStroking(GraphicsContext* context, AlphaPreservation alphaPreservation)
{
	auto& state = context->state();
    prepareCairoContextSource(context, m_cr.get(),
		state.strokePattern.get(),
		state.strokeGradient.get(),
		state.strokeColor,
		alphaPreservation == PreserveAlpha ? globalAlpha() : 1,
		state.strokeGradientSpaceTransform);
}

void PlatformContextCairo::clipForPatternFilling(const GraphicsContextState& state)
{
    ASSERT(state.fillPattern);

    // Hold current cairo path in a variable for restoring it after configuring the pattern clip rectangle.
    OwnPtr<cairo_path_t> currentPath = adoptPtr(cairo_copy_path(m_cr.get()));
    cairo_new_path(m_cr.get());

    // Initialize clipping extent from current cairo clip extents, then shrink if needed according to pattern.
    // Inspired by GraphicsContextQt::drawRepeatPattern.
    double x1, y1, x2, y2;
    cairo_clip_extents(m_cr.get(), &x1, &y1, &x2, &y2);
    FloatRect clipRect(x1, y1, x2 - x1, y2 - y1);

    Image* patternImage = state.fillPattern->tileImage();
    ASSERT(patternImage);
    const AffineTransform& patternTransform = state.fillPattern->patternSpaceTransform();
    FloatRect patternRect = patternTransform.mapRect(FloatRect(0, 0, patternImage->width(), patternImage->height()));

    bool repeatX = state.fillPattern->repeatX();
    bool repeatY = state.fillPattern->repeatY();

    if (!repeatX) {
        clipRect.setX(patternRect.x());
        clipRect.setWidth(patternRect.width());
    }
    if (!repeatY) {
        clipRect.setY(patternRect.y());
        clipRect.setHeight(patternRect.height());
    }
    if (!repeatX || !repeatY) {
        cairo_rectangle(m_cr.get(), clipRect.x(), clipRect.y(), clipRect.width(), clipRect.height());
        cairo_clip(m_cr.get());
    }

    // Restoring cairo path.
    cairo_append_path(m_cr.get(), currentPath.get());
}

#if ENABLE(WEBGL)
bool GraphicsContext3D::getImageData(Image* image, unsigned int format, unsigned int type, bool premultiplyAlpha, bool ignoreGammaAndColorProfile, Vector<uint8_t>& outputVector)
{
    if (!image)
        return false;

    Image::Locker locker(image);

    AlphaOp alphaOp = AlphaDoNothing;
    RefPtr<cairo_surface_t> imageSurface;

    int width = 0;
    int height = 0;
    int bytesPerRow = 0;
    unsigned char* data_ptr = nullptr;
#if CAIRO_USE_ARGB
    SourceDataFormat sourceDataFormat = SourceFormatBGRA8;
#else
    SourceDataFormat sourceDataFormat = SourceFormatRGBA8;
#endif

    if(ImageFrame* imageFrame = image->frameBufferForCurrentFrame()){
        width = imageFrame->width();
        height = imageFrame->height();
        bytesPerRow = width * 4;
        data_ptr = reinterpret_cast<unsigned char* >(imageFrame->getAddr(0, 0));
    }
    else if(auto imageSurface = image->nativeImageForCurrentFrame()){
        if (!premultiplyAlpha)
            alphaOp = AlphaDoUnmultiply;

		if(!platformDelegate().getSurfaceSize(imageSurface.get(), width, height) || !width || !height)
			return false;

		cairo_format_t imageFormat = CAIRO_FORMAT_INVALID;
		if(!platformDelegate().getSurfaceFormat(imageSurface.get(), imageFormat))
			return false;

		if (imageFormat == CAIRO_FORMAT_ARGB32)
			sourceDataFormat = SourceFormatBGRA8;
		else if (imageFormat == CAIRO_FORMAT_ABGR32)
			sourceDataFormat = SourceFormatRGBA8;
		else
			return false;

		if(!platformDelegate().lockSurface(imageSurface.get(), &data_ptr, &bytesPerRow, false))
			return false;
    }
    else
    	return false;

    unsigned int srcUnpackAlignment = 1;
    unsigned int padding = bytesPerRow - 4 * width;
    if (padding) {
        srcUnpackAlignment = padding + 1;
        while (bytesPerRow % srcUnpackAlignment)
            ++srcUnpackAlignment;
    }

    outputVector.resize(width * height * 4);

    bool result = false;
    result = packPixels((uint8_t*)data_ptr, sourceDataFormat, width, height, srcUnpackAlignment, format, type, alphaOp, outputVector.data());

    if(imageSurface)
    	platformDelegate().unlockSurface(imageSurface.get(), false);

    return result;
}

bool GraphicsContext3D::getImageData(ImageBuffer* imageBuffer, unsigned int format, unsigned int type, bool premultiplyAlpha, bool ignoreGammaAndColorProfile, Vector<uint8_t>& outputVector)
{
    if (!imageBuffer)
        return false;

    cairo_surface_t* imageSurface = imageBuffer->nativeSurface();

    if(!imageSurface)
    	return false;

    int width;
    int height;
    AlphaOp alphaOp = AlphaDoNothing;
    unsigned int srcUnpackAlignment = 1;
#if CAIRO_USE_ARGB
    SourceDataFormat sourceDataFormat = SourceFormatBGRA8;
#else
    SourceDataFormat sourceDataFormat = SourceFormatRGBA8;
#endif

	if (!premultiplyAlpha)
		alphaOp = AlphaDoUnmultiply;

	if(!platformDelegate().getSurfaceSize(imageSurface, width, height) || !width || !height)
		return false;

	cairo_format_t imageFormat = CAIRO_FORMAT_INVALID;
	if(!platformDelegate().getSurfaceFormat(imageSurface, imageFormat))
		return false;

	if (imageFormat == CAIRO_FORMAT_ARGB32)
		sourceDataFormat = SourceFormatBGRA8;
	else if (imageFormat == CAIRO_FORMAT_ABGR32)
		sourceDataFormat = SourceFormatRGBA8;
	else
		return false;

    outputVector.resize(width * height * 4);

    if(type == UNSIGNED_BYTE && format == RGBA && srcUnpackAlignment <= 4 && (alphaOp == AlphaDoNothing || alphaOp == AlphaDoUnmultiply)) {
		// No conversion necessary.
		if(platformDelegate().extractSurfaceData(imageSurface, outputVector.data(), outputVector.size())){
			if(alphaOp == AlphaDoUnmultiply)
				packPixels((uint8_t*)outputVector.data(), sourceDataFormat, width, height, srcUnpackAlignment, format, type, alphaOp, outputVector.data());
			return true;
		}
	}

    int bytesPerRow;
    unsigned char* data_ptr;

	if(!platformDelegate().lockSurface(imageSurface, &data_ptr, &bytesPerRow, false))
		return false;

    unsigned int padding = bytesPerRow - 4 * width;
    if (padding) {
        srcUnpackAlignment = padding + 1;
        while (bytesPerRow % srcUnpackAlignment)
            ++srcUnpackAlignment;
    }

    bool result = packPixels((uint8_t*)data_ptr, sourceDataFormat, width, height, srcUnpackAlignment, format, type, alphaOp, outputVector.data());

    platformDelegate().unlockSurface(imageSurface, false);

    return result;
}

#endif

} // namespace WebCore
