/*
 * Copyright (C) 2009 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "Length.h"
#include "LengthFunctions.h"
#include "TransformOperation.h"
#include <wtf/Optional.h>

namespace WebCore {

struct BlendingContext;

class PerspectiveTransformOperation FINAL : public TransformOperation {
public:
    static Ref<PerspectiveTransformOperation> create(const Optional<Length>& p)
    {
        return adoptRef(*new PerspectiveTransformOperation(p));
    }

    const Optional<Length>& perspective() const { return m_p; }
    
private:
    virtual bool isIdentity() const FINAL { return !m_p; }
    virtual bool isAffectedByTransformOrigin() const FINAL { return !isIdentity(); }
    virtual bool isRepresentableIn2D() const final { return false; }

    virtual bool operator==(const TransformOperation& o) const FINAL
    {
        if (!isSameType(o))
            return false;
        const PerspectiveTransformOperation* p = static_cast<const PerspectiveTransformOperation*>(&o);
        return m_p == p->m_p;
    }

    Optional<float> floatValue() const
    {
        if (!m_p)
            return Nullopt;

        // From https://www.w3.org/TR/css-transforms-2/#perspective-property:
        // "As very small <length> values can produce bizarre rendering results and stress the numerical accuracy of
        // transform calculations, values less than 1px must be treated as 1px for rendering purposes. (This clamping
        // does not affect the underlying value, so perspective: 0; in a stylesheet will still serialize back as 0.)"
        return std::max(1.0f, floatValueForLength(*m_p, 1.0));
    }

    virtual bool apply(TransformationMatrix& transform, const FloatSize&) const FINAL
    {
        if (auto value = floatValue())
            transform.applyPerspective(*value);
        return false;
    }

    virtual PassRefPtr<TransformOperation> blend(const TransformOperation* from, const BlendingContext&, bool blendToIdentity = false) FINAL;

    PerspectiveTransformOperation(const Optional<Length>& p)
        : TransformOperation(PERSPECTIVE)
		, m_p(p)
    {
        ASSERT(!p || (*p).isFixed());
    }

    Optional<Length> m_p;
};

} // namespace WebCore
