/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2009, 2010, 2011 Espial Group Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#include "ImageDecoder.h"
#include <png.h>
#include <wtf/OwnPtr.h>

namespace WebCore
{

class PNGImageReader;

// This class decodes the PNG image format.
class PNGImageDecoder final : public ImageDecoder {
public:
	PNGImageDecoder(ImageSource::AlphaOption, ImageSource::GammaAndColorProfileOption);
    virtual ~PNGImageDecoder();

    // ImageDecoder
    virtual String filenameExtension() const override { return "png"; }
    virtual size_t frameCount() override { return m_frameCount; }
    virtual RepetitionCount repetitionCount() const override;
    virtual bool isSizeAvailable() override;
    virtual bool setSize(unsigned width, unsigned height) override;
    virtual ImageFrame* frameBufferAtIndex(size_t index) override;
	// CAUTION: setFailed() deletes |m_reader|.  Be careful to avoid
    // accessing deleted memory, especially when calling this from inside
    // PNGImageReader!
   	virtual bool setFailed() override;
        
    // Callbacks from libpng
    void headerAvailable();
    void rowAvailable(unsigned char* rowBuffer, unsigned rowIndex, int interlacePass);
    void pngComplete();
	void readChunks(png_unknown_chunkp);
	void frameHeader();

	void init();
	virtual void clearFrameBufferCache(size_t clearBeforeFrame) override;

    bool isComplete() const
    {
        if (m_frameBufferCache.isEmpty())
            return false;

        for (auto& imageFrame : m_frameBufferCache) {
            if (imageFrame.status() != ImageFrame::FrameComplete)
                return false;
        }

        return true;
    }

    bool isCompleteAtIndex(size_t index)
    {
        return (index < m_frameBufferCache.size() && m_frameBufferCache[index].status() == ImageFrame::FrameComplete);
    }

protected:
    // Decodes the image.  If |onlySize| is true, stops decoding after
    // calculating the image size.  If decoding fails but there is no more
    // data coming, sets the "decode failure" flag.
    void decode(bool onlySize, unsigned haltAtFrame);
    void initFrameBuffer(size_t frameIndex);
    void frameComplete();
    int processingStart(png_unknown_chunkp);
    int processingFinish();
    void fallbackNotAnimated();

    OwnPtr<PNGImageReader> m_reader;
	bool m_doNothingOnFailure { false };
    unsigned m_currentFrame { 0 };
    png_structp m_png { nullptr };
    png_infop m_info { nullptr };
    bool m_isAnimated { false };
    bool m_frameInfo { false };
    bool m_frameIsHidden { false };
    bool m_hasInfo { false };
    int m_gamma { 45455 };
    size_t m_frameCount { 1 };
    unsigned m_playCount { 0 };
    unsigned m_totalFrames { 0 };
    unsigned m_sizePLTE { 0 };
    unsigned m_sizetRNS { 0 };
    unsigned m_sequenceNumber { 0 };
    unsigned m_width { 0 };
    unsigned m_height { 0 };
    unsigned m_xOffset { 0 };
    unsigned m_yOffset { 0 };
    unsigned m_delayNumerator { 1 };
    unsigned m_delayDenominator { 1 };
    unsigned m_dispose { 0 };
    unsigned m_blend { 0 };
    png_byte m_dataIHDR[12 + 13];
    png_byte m_dataPLTE[12 + 256 * 3];
    png_byte m_datatRNS[12 + 256];
};

} // namespace WebCore
