/*
 * Copyright (C) 2007 Apple Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
#pragma once

#include "ProtectionSpace.h"
#include <wtf/Optional.h>
#include <wtf/Ref.h>
#include <wtf/RefCounted.h>
#include <wtf/text/CString.h>
#include <wtf/text/WTFString.h>

#define CERTIFICATE_CREDENTIALS_SUPPORTED (PLATFORM(MAC) && (PLATFORM(IOS) || __MAC_OS_X_VERSION_MIN_REQUIRED >= 1060))

#if CERTIFICATE_CREDENTIALS_SUPPORTED
#include <Security/SecBase.h>
#include <wtf/RetainPtr.h>
#endif

namespace WebCore {

enum CredentialPersistence {
    CredentialPersistenceNone,
    CredentialPersistenceForSession,
    CredentialPersistencePermanent
};

#if CERTIFICATE_CREDENTIALS_SUPPORTED
enum CredentialType {
    CredentialTypePassword,
    CredentialTypeClientCertificate
};
#endif

struct DigestData final : public RefCounted<DigestData> {
public:
	static Ref<DigestData> create() { return adoptRef(*new DigestData()); }
	static Ref<DigestData> create(const DigestData& other)
	{
		Ref<DigestData> copy = create();

		copy->nonce = other.nonce;
		copy->cnonce = other.cnonce;
		copy->realm = other.realm;
		copy->algo = other.algo;
		copy->stale = other.stale;
		copy->opaque = other.opaque;
		copy->qop = other.qop;
		copy->algorithm = other.algorithm;
		copy->nc = other.nc;

		return copy;
	}

	CString nonce;
	CString cnonce;
	CString realm;
	int algo { 0 };
	bool stale { false }; /* set true for re-negotiation */
	CString opaque;
	CString qop;
	CString algorithm;
	int nc { 0 }; /* nounce count */

private:
	DigestData() = default;
};

class Credential final {
public:
    Credential();
    Credential(const String& user, const String& password, CredentialPersistence, const ProtectionSpace& protectionSpace);
    Credential(const Credential& original, CredentialPersistence);
#if CERTIFICATE_CREDENTIALS_SUPPORTED
    Credential(SecIdentityRef identity, CFArrayRef certificates, CredentialPersistence);
#endif
    
    bool isEmpty() const;

    const String& user() const { return m_user; }
    const String& password() const { return m_password; }
    bool hasPassword() const;
    CredentialPersistence persistence() const;
    const ProtectionSpace& protectionSpace() const { return m_protectionSpace; }
        
#if CERTIFICATE_CREDENTIALS_SUPPORTED
    SecIdentityRef identity() const;
    CFArrayRef certificates() const;
    CredentialType type() const;
#endif

    RefPtr<DigestData> httpDigestData() const { return m_httpDigestData; }
    void SetHttpDigestData(RefPtr<DigestData> httpDigestData) { m_httpDigestData = httpDigestData; }
    
private:
    String m_user;
    String m_password;
    CredentialPersistence m_persistence;
    ProtectionSpace m_protectionSpace;
#if CERTIFICATE_CREDENTIALS_SUPPORTED
    RetainPtr<SecIdentityRef> m_identity;
    RetainPtr<CFArrayRef> m_certificates;
    CredentialType m_type;
#endif
    mutable RefPtr<DigestData> m_httpDigestData;
};

// Need to enforce empty, non-null strings due to the pickiness of the String == String operator
// combined with the semantics of the String(NSString*) constructor
inline Credential::Credential()
    : m_user(emptyString())
    , m_password(emptyString())
    , m_persistence(CredentialPersistenceNone)
#if CERTIFICATE_CREDENTIALS_SUPPORTED
    , m_type(CredentialTypePassword)
#endif
{
}

// Need to enforce empty, non-null strings due to the pickiness of the String == String operator
// combined with the semantics of the String(NSString*) constructor
inline Credential::Credential(const String& user, const String& password, CredentialPersistence persistence, const ProtectionSpace& protectionSpace)
    : m_user(user.length() ? user : emptyString())
    , m_password(password.length() ? password : emptyString())
    , m_persistence(persistence)
    , m_protectionSpace(protectionSpace)
#if CERTIFICATE_CREDENTIALS_SUPPORTED
    , m_type(CredentialTypePassword)
#endif
{
}

inline Credential::Credential(const Credential& original, CredentialPersistence persistence)
    : m_user(original.user())
    , m_password(original.password())
    , m_persistence(persistence)
#if CERTIFICATE_CREDENTIALS_SUPPORTED
    , m_identity(original.identity())
    , m_certificates(original.certificates())
    , m_type(original.type())
#endif
{
}

inline bool Credential::isEmpty() const
{
#if CERTIFICATE_CREDENTIALS_SUPPORTED
    if (m_type == CredentialTypeClientCertificate && (m_identity || m_certificates))
        return false;
#endif

    return m_user.isEmpty() && m_password.isEmpty();
}

inline bool Credential::hasPassword() const
{
    return !m_password.isEmpty();
}

inline CredentialPersistence Credential::persistence() const
{
    return m_persistence;
}

bool operator==(const Credential& a, const Credential& b);
inline bool operator!=(const Credential& a, const Credential& b) { return !(a == b); }
    
};
