/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (C) 2009 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#ifndef HTTPHeaderMap_h
#define HTTPHeaderMap_h

#include <utility>
#include "HTTPHeaderNames.h"
#include <wtf/HashMap.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/Vector.h>
#include <wtf/text/AtomicString.h>
#include <wtf/text/AtomicStringHash.h>
#include <wtf/text/StringHash.h>
#include <wtf/text/StringView.h>

namespace WebCore {

typedef Vector<std::pair<String, String> > CrossThreadHTTPHeaderMapData;

struct HTTPHeaderHashTraits : HashTraits<String> {
	static const int minimumTableSize = 8;
};

class HTTPHeaderMap {
	WTF_MAKE_FAST_ALLOCATED;
private:
	// Adapter that allows the HashMap to take C strings as keys.
	struct CaseFoldingASCIILiteralTranslator {
		static ALWAYS_INLINE unsigned hash(const ASCIILiteral& literal)
		{
			return ASCIICaseInsensitiveHash::hash(literal.characters(), literal.length());
		}

		static ALWAYS_INLINE bool equal(const String& key, const ASCIILiteral& literal)
		{
			return equalIgnoringASCIICase(key, literal);
		}

		static ALWAYS_INLINE void translate(String& location, const ASCIILiteral& literal, unsigned /*hash*/)
		{
			location = String::fromLiteral(literal.characters(), literal.length());
		}
	};

public:
	typedef HashMap<String, String, ASCIICaseInsensitiveHash, HTTPHeaderHashTraits> HashMapType;

    typedef HashMapType::const_iterator const_iterator;
    typedef HashMapType::iterator iterator;

	// Gets a copy of the data suitable for passing to another thread.
	PassOwnPtr<CrossThreadHTTPHeaderMapData> copyData() const;

	void adopt(PassOwnPtr<CrossThreadHTTPHeaderMapData>);

    bool isEmpty() const { return m_headers.isEmpty(); }
    int size() const { return m_headers.size(); }

    void clear() { m_headers.clear(); }

	String get(const String& name) const
	{
		return m_headers.get(name);
	}

    String get(HTTPHeaderName headerName) const
    {
    	StringView name = httpHeaderNameString(headerName);
    	return m_headers.get(name.toStringWithoutCopying());
    }

	void add(const String& name, const String& value)
	{
	    auto result = m_headers.add(name, value);
	    if (!result.isNewEntry)
	        result.iterator->second = result.iterator->second + ',' + value;
	}

	void add(HTTPHeaderName headerName, const String& value)
	{
    	StringView name = httpHeaderNameString(headerName);
		add(name.toStringWithoutCopying(), value);
	}

	bool addIfNotPresent(HTTPHeaderName headerName, const String& value)
	{
    	StringView name = httpHeaderNameString(headerName);
    	return m_headers.add(name.toStringWithoutCopying(), value).isNewEntry;
	}

	String get(const ASCIILiteral& name) const
	{
		auto it = m_headers.find<CaseFoldingASCIILiteralTranslator>(name);
		if (it == m_headers.end())
			return String();
		return it->second;
	}

	bool contains(const ASCIILiteral& name) const
	{
	    return m_headers.find<CaseFoldingASCIILiteralTranslator>(name) != m_headers.end();
	}

	bool contains(const String& name) const
	{
		return m_headers.contains(name);
	}

	bool contains(HTTPHeaderName headerName) const
	{
    	StringView name = httpHeaderNameString(headerName);
		return m_headers.contains(name.toStringWithoutCopying());
	}

	void remove(const ASCIILiteral& name)
	{
		m_headers.remove(m_headers.find<CaseFoldingASCIILiteralTranslator>(name));
	}

	void remove(const String& name)
	{
		m_headers.remove(name);
	}

	void remove(HTTPHeaderName headerName)
	{
    	StringView name = httpHeaderNameString(headerName);
    	m_headers.remove(name.toStringWithoutCopying());
	}

	void set(const String& name, const String& value)
	{
		m_headers.set(name, value);
	}

	void set(HTTPHeaderName headerName, const String& value)
	{
    	StringView name = httpHeaderNameString(headerName);
    	m_headers.set(name.toStringWithoutCopying(), value);
	}

    const HashMapType& headers() const { return m_headers; }
    HashMapType& headers() { return m_headers; }

    const_iterator begin() const { return m_headers.begin(); }
    const_iterator end() const { return m_headers.end(); }

    friend bool operator==(const HTTPHeaderMap& a, const HTTPHeaderMap& b)
    {
        return a.m_headers == b.m_headers;
    }

    friend bool operator!=(const HTTPHeaderMap& a, const HTTPHeaderMap& b)
    {
        return !(a == b);
    }

private:
	HashMapType m_headers;
};

} // namespace WebCore

#endif // HTTPHeaderMap_h
