/*
 * Copyright (C) 2012 Adobe Systems Incorporated. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#pragma once

#include "BasicShapes.h"
#include "OffsetRotation.h"
#include "Path.h"
#include "RenderStyleConstants.h"
#include <wtf/Optional.h>
#include <wtf/OwnPtr.h>
#include <wtf/PassOwnPtr.h>
#include <wtf/RefCounted.h>
#include <wtf/text/WTFString.h>

namespace WebCore {

class SVGElement;

class ClipPathOperation : public RefCounted<ClipPathOperation> {
public:
    enum OperationType {
        Reference,
        Shape,
        Box,
        Ray
    };

    virtual ~ClipPathOperation() { }

    virtual bool operator==(const ClipPathOperation&) const = 0;
    bool operator!=(const ClipPathOperation& o) const { return !(*this == o); }

    OperationType type() const { return m_type; }
    bool isSameType(const ClipPathOperation& o) const { return o.type() == m_type; }
    virtual const Optional<Path> getPath(const FloatRect& reference = FloatRect(), FloatPoint anchor = FloatPoint(), OffsetRotation rotation = OffsetRotation()) const = 0;
protected:
    ClipPathOperation(OperationType type)
        : m_type(type)
    {
    }

    OperationType m_type;
};

class ReferencePathOperation final : public ClipPathOperation {
public:
    static Ref<ReferencePathOperation> create(const String& url, const String& fragment, const RefPtr<SVGElement>);

    const String& url() const { return m_url; }
    const String& fragment() const { return m_fragment; }
    const Optional<Path> getPath(const FloatRect&, FloatPoint, OffsetRotation) const final { return m_path; }
    const Optional<Path> path() const { return m_path; }
private:
    virtual bool operator==(const ClipPathOperation& other) const override
    {
        if (!isSameType(other))
            return false;
        const auto& referenceClip = static_cast<const ReferencePathOperation&>(other);
        return m_url == referenceClip.m_url;
    }

    ReferencePathOperation(const String& url, const String& fragment, const RefPtr<SVGElement>);

    String m_url;
    String m_fragment;
    Optional<Path> m_path;
};

class ShapeClipPathOperation final : public ClipPathOperation {
public:
    static Ref<ShapeClipPathOperation> create(Ref<BasicShape>&& shape)
    {
        return adoptRef(*new ShapeClipPathOperation(WTFMove(shape)));
    }

    const BasicShape& basicShape() const { return m_shape.get(); }
    WindRule windRule() const { return m_shape->windRule(); }
    const Path& pathForReferenceRect(const FloatRect& boundingRect) const { return m_shape->path(boundingRect); }

    void setReferenceBox(CSSBoxType referenceBox) { m_referenceBox = referenceBox; }
    CSSBoxType referenceBox() const { return m_referenceBox; }
    const Optional<Path> getPath(const FloatRect& reference, FloatPoint, OffsetRotation) const final { return pathForReferenceRect(reference); }

private:
    virtual bool operator==(const ClipPathOperation& other) const override
    {
        if (!isSameType(other))
            return false;
        const ShapeClipPathOperation& shapeClip = static_cast<const ShapeClipPathOperation&>(other);
        return m_referenceBox == shapeClip.referenceBox()
			&& (m_shape.ptr() == shapeClip.m_shape.ptr() || m_shape.get() == shapeClip.m_shape.get());
	}

    ShapeClipPathOperation(Ref<BasicShape>&& shape)
        : ClipPathOperation(Shape)
    	, m_shape(WTFMove(shape))
		, m_referenceBox(CSSBoxType::BoxMissing)
	{
	}

	Ref<BasicShape> m_shape;
	CSSBoxType m_referenceBox;
};

class BoxClipPathOperation final : public ClipPathOperation {
public:
    static PassRefPtr<BoxClipPathOperation> create(CSSBoxType referenceBox)
    {
        return adoptRef(new BoxClipPathOperation(referenceBox));
    }

    const Path pathForReferenceRect(const FloatRoundedRect& boundingRect) const
    {
        Path path;
        path.addRoundedRect(boundingRect);
        return path;
    }

    void setPathForReferenceRect(const FloatRoundedRect& boundingRect)
    {
        m_path.clear();
        m_path.addRoundedRect(boundingRect);
    }

    const Optional<Path> getPath(const FloatRect&, FloatPoint, OffsetRotation) const final { return m_path; }
	CSSBoxType referenceBox() const { return m_referenceBox; }

private:
    virtual bool operator==(const ClipPathOperation& other) const OVERRIDE
    {
        if (!isSameType(other))
            return false;
        const BoxClipPathOperation& boxClip = static_cast<const BoxClipPathOperation&>(other);
        return m_referenceBox == boxClip.m_referenceBox;
    }

    explicit BoxClipPathOperation(CSSBoxType referenceBox)
        : ClipPathOperation(Box)
        , m_referenceBox(referenceBox)
    {
    }

    Path m_path;
    CSSBoxType m_referenceBox;
};

class RayPathOperation final : public ClipPathOperation {
public:
    enum class Size {
        ClosestSide,
        ClosestCorner,
        FarthestSide,
        FarthestCorner,
        Sides
    };

    static Ref<RayPathOperation> create(float angle, Size size, bool isContaining)
    {
        return adoptRef(*new RayPathOperation(angle, size, isContaining));
    }

    float angle() const { return m_angle; }
    Size size() const { return m_size; }
    bool isContaining() const { return m_isContaining; }

    bool canBlend(const RayPathOperation& other) const
    {
        // Two rays can only be blended if they have the same size and are both containing.
        return m_size == other.m_size && m_isContaining == other.m_isContaining;
    }

    Ref<RayPathOperation> blend(const RayPathOperation& to, const BlendingContext& context) const
    {
        return RayPathOperation::create(WebCore::blend(m_angle, to.m_angle, context), m_size, m_isContaining);
    }

    double lengthForPath() const;
    double lengthForContainPath(const FloatRect& elementRect, double computedPathLength, const FloatPoint& anchor, const OffsetRotation rotation) const;

    void setContainingBlockReferenceRect(const FloatRect& boundingRect)
    {
        m_containingBlockBoundingRect = boundingRect;
    }
    void setStartingPosition(const FloatPoint& position)
    {
        m_position = position;
    }
    const Optional<Path> getPath(const FloatRect& reference = FloatRect(), FloatPoint anchor = FloatPoint(), OffsetRotation rotation = OffsetRotation()) const final;
private:
    bool operator==(const ClipPathOperation& other) const override
    {
        if (!isSameType(other))
            return false;

        auto& otherCasted = static_cast<const RayPathOperation&>(other);
        return m_angle == otherCasted.m_angle
            && m_size == otherCasted.m_size
            && m_isContaining == otherCasted.m_isContaining;
    }

    RayPathOperation(float angle, Size size, bool isContaining)
        : ClipPathOperation(Ray)
        , m_angle(angle)
        , m_size(size)
        , m_isContaining(isContaining)
    {
    }

    float m_angle { 0 };
    Size m_size;
    bool m_isContaining { false };
    FloatRect m_containingBlockBoundingRect;
    FloatPoint m_position;
};

//#define CLIP_PATH_OPERATION_CASTS(ToValueTypeName, predicate) \
//    TYPE_CASTS_BASE(ToValueTypeName, ClipPathOperation, operation, operation->type() == ClipPathOperation::predicate, operation.type() == ClipPathOperation::predicate)

//CLIP_PATH_OPERATION_CASTS(ReferencePathOperation, Reference)
//CLIP_PATH_OPERATION_CASTS(ShapeClipPathOperation, Shape)
//CLIP_PATH_OPERATION_CASTS(BoxClipPathOperation, Box)
//SPECIALIZE_TYPE_TRAITS_CLIP_PATH_OPERATION(RayPathOperation, PathOperation::Ray)

inline bool isShapeClipPathOperation(const ClipPathOperation& operation)
{
	return operation.type() == ClipPathOperation::Shape;
}

inline bool isShapeClipPathOperation(const ClipPathOperation* operation)
{
	return operation && isShapeClipPathOperation(*operation);
}

inline bool isBoxClipPathOperation(const ClipPathOperation& operation)
{
	return operation.type() == ClipPathOperation::Box;
}

inline bool isReferenceClipPathOperation(const ClipPathOperation& operation)
{
	return operation.type() == ClipPathOperation::Reference;
}

inline bool isRayPathOperation(const ClipPathOperation& operation)
{
	return operation.type() == ClipPathOperation::Ray;
}

inline bool isRayPathOperation(const ClipPathOperation* operation)
{
	return operation && isRayPathOperation(*operation);
}

inline ReferencePathOperation& toReferenceClipPathOperation(ClipPathOperation& operation)
{
	ASSERT(operation.type() == Reference);
	return static_cast<ReferencePathOperation&>(operation);
}

inline const ReferencePathOperation& toReferenceClipPathOperation(const ClipPathOperation& operation)
{
	ASSERT(operation.type() == Reference);
	return static_cast<const ReferencePathOperation&>(operation);
}

inline ShapeClipPathOperation& toShapeClipPathOperation(ClipPathOperation& operation)
{
	ASSERT(operation.type() == Shape);
	return static_cast<ShapeClipPathOperation&>(operation);
}

inline const ShapeClipPathOperation& toShapeClipPathOperation(const ClipPathOperation& operation)
{
	ASSERT(operation.type() == Shape);
	return static_cast<const ShapeClipPathOperation&>(operation);
}

inline BoxClipPathOperation& toBoxClipPathOperation(ClipPathOperation& operation)
{
	ASSERT(operation.type() == Box);
	return static_cast<BoxClipPathOperation&>(operation);
}

inline const BoxClipPathOperation& toBoxClipPathOperation(const ClipPathOperation& operation)
{
	ASSERT(operation.type() == Box);
	return static_cast<const BoxClipPathOperation&>(operation);
}

inline RayPathOperation& toRayPathOperation(ClipPathOperation& operation)
{
	ASSERT(operation.type() == Ray);
	return static_cast<RayPathOperation&>(operation);
}

inline const RayPathOperation& toRayPathOperation(const ClipPathOperation& operation)
{
	ASSERT(operation.type() == Ray);
	return static_cast<const RayPathOperation&>(operation);
}

} // namespace WebCore
