/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2003, 2006, 2007, 2009, 2013 Apple Inc. All rights reserved.
 * Copyright (C) 2010, 2012 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#include "RenderObject.h"
#include "RenderPtr.h"

namespace WebCore {

class ContentData;
class KeyframeList;
class ReferencedSVGResources;
class RenderBlock;
class RenderStyle;
class RenderTreeBuilder;

class RenderElement : public RenderObject {
public:
    virtual ~RenderElement();

    static bool isContentDataSupported(const ContentData&);

    enum RendererCreationType { CreateAllRenderers, OnlyCreateBlockAndFlexboxRenderers };
    static RenderPtr<RenderElement> createFor(Element&, PassOwnPtr<RenderStyle>, RendererCreationType = CreateAllRenderers);

    bool hasInitializedStyle() const { return m_hasInitializedStyle; }

    const RenderStyle* style() const { return m_style.get(); }
    const RenderStyle* parentStyle() const { return !m_parent ? nullptr : m_parent->style(); }
	const RenderStyle* firstLineStyle() const;

    virtual const RenderStyle* outlineStyleForRepaint() const OVERRIDE { return style(); }

    // FIXME: Style shouldn't be mutated.
    RenderStyle& mutableStyle() { return *m_style; }

    void initializeStyle();

    // Calling with minimalStyleDifference > StyleDifferenceEqual indicates that
    // out-of-band state (e.g. animations) requires that styleDidChange processing
    // continue even if the style isn't different from the current style.
    void setStyle(PassOwnPtr<RenderStyle>, StyleDifference minimalStyleDifference = StyleDifferenceEqual);

    // The pseudo element style can be cached or uncached.  Use the cached method if the pseudo element doesn't respect
    // any pseudo classes (and therefore has no concept of changing state).
    RenderStyle* getCachedPseudoStyle(PseudoId, const RenderStyle* parentStyle = nullptr) const;
    PassOwnPtr<RenderStyle> getUncachedPseudoStyle(const PseudoStyleRequest&, const RenderStyle* parentStyle = nullptr, const RenderStyle* ownStyle = nullptr) const;

    // This is null for anonymous renderers.
    Element* element() const { return toElement(RenderObject::node()); }
    Element* nonPseudoElement() const { return toElement(RenderObject::nonPseudoNode()); }
    Element* generatingElement() const;

    RenderObject* firstChild() const { return m_firstChild; }
    RenderObject* lastChild() const { return m_lastChild; }
    RenderObject* firstInFlowChild() const;
    RenderObject* lastInFlowChild() const;

    bool shouldApplyLayoutContainment() const;

    bool isAnonymousBlock() const
	{
		// This function is kept in sync with anonymous block creation conditions in
		// RenderBlock::createAnonymousBlock(). This includes creating an anonymous
		// RenderBlock having a DisplayType::Block or DisplayType::Box display. Other classes such as RenderTextFragment
		// are not RenderBlocks and will return false. See https://bugs.webkit.org/show_bug.cgi?id=56709.
		return isAnonymous()
			&& isRenderBlock()
			&& (style()->display() == DisplayType::Block || style()->display() == DisplayType::Box)
			&& style()->styleType() == PseudoId::None
#if ENABLE(FULLSCREEN_API)
			&& !isRenderFullScreen()
			&& !isRenderFullScreenPlaceholder()
#endif
#if ENABLE(MATHML)
			&& !isRenderMathMLBlock()
#endif
			&& !isListMarker()
			&& !isRenderFragmentedFlow()
			&& !isRenderMultiColumnSet()
			&& !isRenderView();
	}

    // Note that even if these 2 "canContain" functions return true for a particular renderer, it does not necessarily mean the renderer is the containing block (see containingBlockForAbsolute(Fixed)Position).
	bool canContainFixedPositionObjects() const;
    bool canContainAbsolutelyPositionedObjects() const;
    bool canEstablishContainingBlockWithTransform() const;

    Color selectionColor(CSSPropertyID colorProperty) const;
    PassOwnPtr<RenderStyle> selectionPseudoStyle() const;

    // Obtains the selection colors that should be used when painting a selection.
    Color selectionBackgroundColor() const;
    Color selectionForegroundColor() const;
    Color selectionEmphasisMarkColor() const;

    // FIXME: Make these standalone and move to relevant files.
    bool isRenderLayerModelObject() const;
    bool isBoxModelObject() const;
    bool isRenderBlock() const;
    bool isRenderBlockFlow() const;
    bool isRenderReplaced() const;
    bool isRenderInline() const;
    bool isSVGRoot() const;
    bool isSVGForeignObject() const;
    bool isRenderFragmentedFlow() const;

    static bool isRenderBlockFlow(const RenderObject*);
    static bool isRenderBlockFlow(const RenderElement*);

    ALWAYS_INLINE bool isFixedPositioned() const { return isOutOfFlowPositioned() && style()->position() == PositionType::Fixed; }
	ALWAYS_INLINE bool isAbsolutelyPositioned() const { return isOutOfFlowPositioned() && style()->position() == PositionType::Absolute; }

    bool isTransformed() const { return hasTransformRelatedProperty() && (style()->affectsTransform() || hasSVGTransform()); }
    bool hasTransformOrPerspective() const { return hasTransformRelatedProperty() && (isTransformed() || style()->hasPerspective()); }

    bool shouldUseTransformFromContainer(const RenderElement* container) const;

    virtual bool isChildAllowed(RenderObject&, const RenderStyle&) const { return true; }
    void didAttachChild(RenderObject& child, RenderObject* beforeChild);

    // Return the renderer whose background style is used to paint the root background. Should only be called on the renderer for which isDocumentElementRenderer() is true.
    RenderElement& rendererForRootBackground();

    // The following functions are used when the render tree hierarchy changes to make sure layers get
    // properly added and removed. Since containership can be implemented by any subclass, and since a hierarchy
    // can contain a mixture of boxes and other object types, these functions need to be in the base class.
    RenderLayer* layerParent() const;
    RenderLayer* layerNextSibling(RenderLayer& parentLayer) const;
    void removeLayers();
    void moveLayers(RenderLayer& newParent);

    virtual void dirtyLinesFromChangedChild(RenderObject*) { }

    bool ancestorLineBoxDirty() const { return m_ancestorLineBoxDirty; }
    void setAncestorLineBoxDirty(bool f = true);

    void setChildNeedsLayout(MarkingBehavior = MarkContainingBlockChain);
    void setOutOfFlowChildNeedsStaticPositionLayout();
    void clearChildNeedsLayout();
    void setNeedsPositionedMovementLayout(const RenderStyle* oldStyle);
    void setNeedsSimplifiedNormalFlowLayout();

    // inline-block elements paint all phases atomically. This function ensures that. Certain other elements
    // (grid items, flex items) require this behavior as well, and this function exists as a helper for them.
    // It is expected that the caller will call this function independent of the value of paintInfo.phase.
    void paintAsInlineBlock(PaintInfo&, const LayoutPoint&);

    // Updates only the local style ptr of the object. Does not update the state of the object,
    // and so only should be called when the style is known not to have changed (or from setStyle).
    void setStyleInternal(PassOwnPtr<RenderStyle> style) { m_style = style; }

    // Repaint only if our old bounds and new bounds are different. The caller may pass in newBounds and newOutlineBox if they are known.
    bool repaintAfterLayoutIfNeeded(RenderLayerModelObject* repaintContainer, const LayoutRect& oldBounds, const LayoutRect& oldOutlineBox, const LayoutRect* newBoundsPtr = 0, const LayoutRect* newOutlineBoxPtr = 0);

    bool borderImageIsLoadedAndCanBeRendered() const;
    bool mayCauseRepaintInsideViewport(const IntRect* visibleRect = 0) const;

    // Returns true if this renderer requires a new stacking context.
    static bool createsGroupForStyle(const RenderStyle&);
	bool createsGroup() const { return createsGroupForStyle(*style()); }

    bool isTransparent() const { return style()->opacity() < 1.0f; }
    float opacity() const { return style()->opacity(); }

    bool visibleToHitTesting() const
    {
        if (style()->effectiveInert())
            return false;

    	return style()->visibility() == VISIBLE && style()->pointerEvents() != PE_NONE;
    }

    bool hasBackground() const { return style()->hasBackground(); }
    ALWAYS_INLINE bool hasMask() const { return style()->hasMask(); }
    bool hasClip() const { return isOutOfFlowPositioned() && style()->hasClip(); }
	bool hasClipOrNonVisibleOverflow() const { return hasClip() || hasNonVisibleOverflow(); }
	bool hasClipPath() const { return style()->clipPath(); }
    bool hasHiddenBackface() const { return style()->backfaceVisibility() == BackfaceVisibilityHidden; }
    bool hasOutlineAnnotation() const;
	bool hasOutline() const { return style()->hasOutline() || hasOutlineAnnotation(); }
	bool hasSelfPaintingLayer() const;

	bool checkForRepaintDuringLayout() const;

    // anchorRect() is conceptually similar to absoluteBoundingBoxRect(), but is intended for scrolling to an anchor.
	// For inline renderers, this gets the logical top left of the first leaf child and the logical bottom right of the
	// last leaf child, converts them to absolute coordinates, and makes a box out of them.
	LayoutRect anchorRect() const;

#if ENABLE(CSS_FILTERS)
    bool hasFilter() const { return style()->hasFilter(); }
#else
    bool hasFilter() const { return false; }
#endif

#if ENABLE(CSS_COMPOSITING)
    bool hasBlendMode() const { return style()->hasBlendMode(); }
#else
    bool hasBlendMode() const { return false; }
#endif

#if ENABLE(CSS_SHAPES)
    bool hasShapeOutside() const { return style()->shapeOutside(); }
#else
    bool hasShapeOutside() const { return false; }
#endif

    bool repaintForPausedImageAnimationsIfNeeded(const IntRect& visibleRect, CachedImage&);
    bool hasPausedImageAnimations() const { return m_hasPausedImageAnimations; }
    void setHasPausedImageAnimations(bool b) { m_hasPausedImageAnimations = b; }

    bool hasBackdropFilter() const
	{
#if ENABLE(FILTERS_LEVEL_2)
		return style()->hasBackdropFilter();
#else
		return false;
#endif
	}

    const RenderElement* enclosingRendererWithTextDecoration(TextDecoration, bool firstLine) const;

    bool isFirstLetter() const { return m_isFirstLetter; }
	void setIsFirstLetter() { m_isFirstLetter = true; }

#if !ASSERT_DISABLED
    bool m_reparentingChild { false };
#endif

    RenderObject* attachRendererInternal(RenderPtr<RenderObject> child, RenderObject* beforeChild);
    RenderPtr<RenderObject> detachRendererInternal(RenderObject&);

    void removeFromRenderFlowThread();
    virtual void resetEnclosingFragmentedFlowAndChildInfoIncludingDescendants(RenderFragmentedFlow*);

    bool hasCounterNodeMap() const { return m_hasCounterNodeMap; }
    void setHasCounterNodeMap(bool f) { m_hasCounterNodeMap = f; }

    // Called before anonymousChild.setStyle(). Override to set custom styles for
    // the child.
    virtual void updateAnonymousChildStyle(RenderStyle&) const { };

    bool hasContinuationChainNode() const { return m_hasContinuationChainNode; }
    bool isContinuation() const { return m_isContinuation; }
    void setIsContinuation() { m_isContinuation = true; }

    virtual bool startAnimation(double /* timeOffset */, const Animation&, const KeyframeList&) { return false; }
    virtual void animationPaused(double /* timeOffset */, const String& /* name */) { }
    virtual void animationFinished(const String& /* name */) { }
    virtual void transformRelatedPropertyDidChange() { }

    // https://www.w3.org/TR/css-transforms-1/#transform-box
    FloatRect transformReferenceBoxRect(const RenderStyle& style) const { return referenceBoxRect(transformBoxToCSSBoxType(style.transformBox())); }
    FloatRect transformReferenceBoxRect() const { return transformReferenceBoxRect(*style()); }

    // https://www.w3.org/TR/css-transforms-1/#reference-box
    virtual FloatRect referenceBoxRect(CSSBoxType) const;

    OwnPtr<RenderStyle> animatedStyle();
    virtual void suspendAnimations(double time = 0) { }

    WeakPtr<RenderBlockFlow> backdropRenderer() const;
    void setBackdropRenderer(RenderBlockFlow&);

    ReferencedSVGResources& ensureReferencedSVGResources();

    Overflow effectiveOverflowX() const;
    Overflow effectiveOverflowY() const;
    Overflow effectiveOverflowInlineDirection() const { return style()->isHorizontalWritingMode() ? effectiveOverflowX() : effectiveOverflowY(); }
    Overflow effectiveOverflowBlockDirection() const { return style()->isHorizontalWritingMode() ? effectiveOverflowY() : effectiveOverflowX(); }

    bool isWritingModeRoot() const { return !parent() || parent()->style()->writingMode() != style()->writingMode(); }

    bool isDeprecatedFlexItem() const { return !isInline() && !isFloatingOrOutOfFlowPositioned() && parent() && parent()->isDeprecatedFlexibleBox(); }
    bool isFlexItemIncludingDeprecated() const { return !isInline() && !isFloatingOrOutOfFlowPositioned() && parent() && parent()->isFlexibleBoxIncludingDeprecated(); }

    static void markRendererDirtyAfterTopLayerChange(RenderElement* renderer, RenderBlock* containingBlockBeforeStyleResolution);

protected:
    enum BaseTypeFlags : uint16_t {
        RenderLayerModelObjectFlag = 1 << 0,
        RenderBoxModelObjectFlag   = 1 << 1,
        RenderInlineFlag           = 1 << 2,
        RenderReplacedFlag         = 1 << 3,
        RenderBlockFlag            = 1 << 4,
        RenderBlockFlowFlag        = 1 << 5,
		RenderSVGRootFlag          = 1 << 6,
		RenderSVGForeignObjectFlag = 1 << 7,

		RenderFragmentedFlowFlag   = 1 << 8,
    };

    RenderElement(Element&, PassOwnPtr<RenderStyle>, unsigned baseTypeFlags);
    RenderElement(Document&, PassOwnPtr<RenderStyle>, unsigned baseTypeFlags);

    enum StylePropagationType { PropagateToAllChildren, PropagateToBlockChildrenOnly };
    void propagateStyleToAnonymousChildren(StylePropagationType);

    bool layerCreationAllowedForSubtree() const;

    void setFirstChild(RenderObject* child) { m_firstChild = child; }
    void setLastChild(RenderObject* child) { m_lastChild = child; }

    bool repaintBeforeStyleChange(StyleDifference, const RenderStyle& oldStyle, const RenderStyle& newStyle);

    virtual void styleWillChange(StyleDifference, const RenderStyle*);
    virtual void styleDidChange(StyleDifference, const RenderStyle*);

    virtual void insertedIntoTree(IsInternalMove = IsInternalMove::No) OVERRIDE;
    virtual void willBeRemovedFromTree(IsInternalMove = IsInternalMove::No) OVERRIDE;
    virtual void willBeDestroyed() OVERRIDE;
    virtual void notifyFinished(CachedResource*) OVERRIDE;

    void setHasContinuationChainNode(bool b) { m_hasContinuationChainNode = b; }

    void paintOutline(PaintInfo&, const LayoutRect&);
    void updateOutlineAutoAncestor(bool hasOutlineAuto) const;

    void removeFromRenderFlowThreadIncludingDescendants(bool shouldUpdateState);
    void adjustFragmentedFlowStateOnContainingBlockChangeIfNeeded(const RenderStyle& oldStyle, const RenderStyle& newStyle);

private:
    void node() const WTF_DELETED_FUNCTION;
    void nonPseudoNode() const WTF_DELETED_FUNCTION;
    void generatingNode() const WTF_DELETED_FUNCTION;
    void isText() const WTF_DELETED_FUNCTION;
    void isRenderElement() const WTF_DELETED_FUNCTION;

    virtual RenderObject* firstChildSlow() const FINAL { return firstChild(); }
    virtual RenderObject* lastChildSlow() const FINAL { return lastChild(); }

    bool shouldRepaintForStyleDifference(StyleDifference) const;

    void updateFillImages(const FillLayer*, const FillLayer*);
    void updateImage(StyleImage*, StyleImage*);
#if ENABLE(CSS_SHAPES)
    void updateShapeImage(const ShapeValue*, const ShapeValue*);
#endif

    StyleDifference adjustStyleDifference(StyleDifference, OptionSet<StyleDifferenceContextSensitiveProperty>) const;

    virtual void newImageAnimationFrameAvailable(CachedImage&, bool& canPause) FINAL;
    virtual void didRemoveCachedImageClient(CachedImage&) FINAL;
    virtual void scheduleRenderingUpdate() FINAL;

    void issueRepaintForOutlineAuto(float outlineSize);

    void clearSubtreeLayoutRootIfNeeded() const;

    bool getLeadingCorner(FloatPoint& output) const;
    bool getTrailingCorner(FloatPoint& output) const;

    void updateReferencedSVGResources();
    void clearReferencedSVGResources();

protected:
    unsigned m_baseTypeFlags : 16;

private:
    unsigned m_ancestorLineBoxDirty : 1;
    unsigned m_hasInitializedStyle : 1;

    unsigned m_hasPausedImageAnimations : 1;
    unsigned m_hasCounterNodeMap : 1;
    unsigned m_hasContinuationChainNode : 1;
    unsigned m_isContinuation : 1;
    unsigned m_isFirstLetter : 1;

    RenderObject* m_firstChild;
    RenderObject* m_lastChild;

    OwnPtr<RenderStyle> m_style;
};

template <> inline bool isRendererOfType<RenderElement>(const RenderObject& renderer) { return renderer.isRenderElement(); }
template <> inline bool isRendererOfType<const RenderElement>(const RenderObject& renderer) { return renderer.isRenderElement(); }
RENDER_OBJECT_TYPE_CASTS(RenderElement, isRenderElement())

inline void RenderElement::setAncestorLineBoxDirty(bool f)
{
    m_ancestorLineBoxDirty = f;
    if (m_ancestorLineBoxDirty)
        setNeedsLayout();
}

inline void RenderElement::setChildNeedsLayout(MarkingBehavior markParents)
{
    ASSERT(!isSetNeedsLayoutForbidden());
    if (normalChildNeedsLayout())
        return;
    setNormalChildNeedsLayoutBit(true);
    if (markParents == MarkContainingBlockChain)
        markContainingBlocksForLayout();
}

inline bool RenderElement::isRenderLayerModelObject() const
{
    return m_baseTypeFlags & RenderLayerModelObjectFlag;
}

inline bool RenderElement::isBoxModelObject() const
{
    return m_baseTypeFlags & RenderBoxModelObjectFlag;
}

inline bool RenderElement::isRenderBlock() const
{
    return m_baseTypeFlags & RenderBlockFlag;
}

inline bool RenderElement::isRenderBlockFlow() const
{
    return m_baseTypeFlags & RenderBlockFlowFlag;
}

inline bool RenderElement::isRenderReplaced() const
{
    return m_baseTypeFlags & RenderReplacedFlag;
}

inline bool RenderElement::isRenderInline() const
{
    return m_baseTypeFlags & RenderInlineFlag;
}

inline bool RenderElement::isSVGRoot() const
{
    return m_baseTypeFlags & RenderSVGRootFlag;
}

inline bool RenderElement::isSVGForeignObject() const
{
    return m_baseTypeFlags & RenderSVGForeignObjectFlag;
}

inline bool RenderElement::isRenderFragmentedFlow() const
{
    return m_baseTypeFlags & RenderFragmentedFlowFlag;
}

inline Element* RenderElement::generatingElement() const
{
    return toElement(RenderObject::generatingNode());
}

inline bool RenderElement::canContainAbsolutelyPositionedObjects() const
{
    return isRenderView()
		|| style()->position() != PositionType::Static
		|| (canEstablishContainingBlockWithTransform() && hasTransformRelatedProperty())
		|| (hasBackdropFilter() && !isDocumentElementRenderer())
		|| (isRenderBlock() && style()->willChange() && style()->willChange()->createsContainingBlockForAbsolutelyPositioned(isDocumentElementRenderer()))
		|| isSVGForeignObject();
}

inline bool RenderElement::canContainFixedPositionObjects() const
{
    return isRenderView()
		|| (canEstablishContainingBlockWithTransform() && hasTransformRelatedProperty())
		|| (hasBackdropFilter() && !isDocumentElementRenderer())
		|| (isRenderBlock() && style()->willChange() && style()->willChange()->createsContainingBlockForOutOfFlowPositioned(isDocumentElementRenderer()))
		|| isSVGForeignObject();
}

inline bool RenderElement::canEstablishContainingBlockWithTransform() const
{
    return isRenderBlock() || (isTablePart() && !isRenderTableCol());
}

inline bool RenderObject::isRenderLayerModelObject() const
{
    return isRenderElement() && toRenderElement(*this).isRenderLayerModelObject();
}

inline bool RenderObject::isBoxModelObject() const
{
    return isRenderElement() && toRenderElement(*this).isBoxModelObject();
}

inline bool RenderObject::isRenderBlock() const
{
    return isRenderElement() && toRenderElement(*this).isRenderBlock();
}

inline bool RenderObject::isRenderBlockFlow() const
{
    return isRenderElement() && toRenderElement(*this).isRenderBlockFlow();
}

inline bool RenderObject::isRenderReplaced() const
{
    return isRenderElement() && toRenderElement(*this).isRenderReplaced();
}

inline bool RenderObject::isRenderInline() const
{
    return isRenderElement() && toRenderElement(*this).isRenderInline();
}

inline bool RenderObject::isSVGRoot() const
{
    return isRenderElement() && toRenderElement(*this).isSVGRoot();
}

inline bool RenderObject::isSVGForeignObject() const
{
    return isRenderElement() && toRenderElement(*this).isSVGForeignObject();
}

inline bool RenderObject::isRenderFragmentedFlow() const
{
    return isRenderElement() && toRenderElement(*this).isRenderFragmentedFlow();
}

inline const RenderStyle* RenderObject::style() const
{
    if (isText())
        return m_parent->style();
    return toRenderElement(*this).style();
}

inline const RenderStyle* RenderObject::firstLineStyle() const
{
    if (isText())
        return m_parent->firstLineStyle();
    return toRenderElement(*this).firstLineStyle();
}

ALWAYS_INLINE bool RenderObject::isAnonymousBlock() const
{
	return isRenderElement() && toRenderElement(*this).isAnonymousBlock();
}

inline RenderElement* ContainerNode::renderer() const
{
    return toRenderElement(Node::renderer());
}

inline bool RenderElement::createsGroupForStyle(const RenderStyle& style)
{
    return style.hasOpacity() || style.hasMask() || style.clipPath() || style.hasFilter() || style.hasBackdropFilter() || style.hasBlendMode();
}

inline LayoutUnit adjustLayoutUnitForAbsoluteZoom(LayoutUnit value, RenderElement& renderer)
{
    return adjustLayoutUnitForAbsoluteZoom(value, *renderer.style());
}

inline LayoutUnit adjustLayoutUnitForAbsoluteZoom(LayoutUnit value, RenderElement* renderer)
{
    return adjustLayoutUnitForAbsoluteZoom(value, *renderer);
}

inline RenderObject* RenderElement::firstInFlowChild() const
{
    if (auto* firstChild = this->firstChild()) {
        if (firstChild->isInFlow())
            return firstChild;
        return firstChild->nextInFlowSibling();
    }
    return nullptr;
}

inline RenderObject* RenderElement::lastInFlowChild() const
{
    if (auto* lastChild = this->lastChild()) {
        if (lastChild->isInFlow())
            return lastChild;
        return lastChild->previousInFlowSibling();
    }
    return nullptr;
}

inline bool RenderElement::isRenderBlockFlow(const RenderObject* renderer)
{
	return renderer && renderer->isRenderBlockFlow();
}

inline bool RenderElement::isRenderBlockFlow(const RenderElement* renderer)
{
	return renderer && renderer->isRenderBlockFlow();
}

inline bool RenderElement::shouldApplyLayoutContainment() const
{
#if 0
    return shouldApplyLayoutOrPaintContainment(Containment::Layout);
#else
    return false;
#endif
}

inline bool isRenderBlock(const RenderElement& renderer)
{
	return renderer.isRenderBlock();
}

} // namespace WebCore
