/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include "InlineFlowBox.h"
#include "RenderBoxModelObject.h"
#include "RenderLineBoxList.h"
#include <wtf/PassOwnPtr.h>

namespace WebCore {

class Position;
class RenderFragmentContainer;

class RenderInline : public RenderBoxModelObject {
public:
    RenderInline(Element&, PassOwnPtr<RenderStyle>);
    RenderInline(Document&, PassOwnPtr<RenderStyle>);

    virtual LayoutUnit marginLeft() const FINAL;
    virtual LayoutUnit marginRight() const FINAL;
    virtual LayoutUnit marginTop() const FINAL;
    virtual LayoutUnit marginBottom() const FINAL;
    virtual LayoutUnit marginBefore(const RenderStyle* otherStyle = 0) const FINAL;
    virtual LayoutUnit marginAfter(const RenderStyle* otherStyle = 0) const FINAL;
    virtual LayoutUnit marginStart(const RenderStyle* otherStyle = 0) const FINAL;
    virtual LayoutUnit marginEnd(const RenderStyle* otherStyle = 0) const FINAL;

    virtual void absoluteRects(Vector<IntRect>&, const LayoutPoint& accumulatedOffset) const FINAL;
    virtual void absoluteQuads(Vector<FloatQuad>&, bool* wasFixed) const OVERRIDE;

    virtual LayoutSize offsetFromContainer(RenderElement&, const LayoutPoint&, bool* offsetDependsOnPoint = nullptr) const FINAL;

    IntRect linesBoundingBox() const;
    LayoutRect linesVisualOverflowBoundingBox() const;
    LayoutRect linesVisualOverflowBoundingBoxInFragment(const RenderFragmentContainer*) const;

    InlineFlowBox* createAndAppendInlineFlowBox();

    void dirtyLineBoxes(bool fullLayout);
    void deleteLines();

    RenderLineBoxList& lineBoxes() { return m_lineBoxes; }
    const RenderLineBoxList& lineBoxes() const { return m_lineBoxes; }

    InlineFlowBox* firstLineBox() const { return m_lineBoxes.firstLineBox(); }
    InlineFlowBox* lastLineBox() const { return m_lineBoxes.lastLineBox(); }

#if PLATFORM(IOS)
    virtual void absoluteQuadsForSelection(Vector<FloatQuad>& quads) const OVERRIDE;
#endif
    
    LayoutSize offsetForInFlowPositionedInline(const RenderBox& child) const;

    virtual void addFocusRingRects(Vector<IntRect>&, const LayoutPoint& additionalOffset, const RenderLayerModelObject* paintContainer = 0) FINAL;
    void paintOutline(PaintInfo&, const LayoutPoint&);

    bool mayAffectLayout() const;

protected:
    virtual void willBeDestroyed() OVERRIDE;

    virtual void styleWillChange(StyleDifference, const RenderStyle* newStyle) OVERRIDE;
    virtual void styleDidChange(StyleDifference, const RenderStyle* oldStyle) OVERRIDE;

    virtual void updateFromStyle() OVERRIDE;

private:
	virtual const char* renderName() const OVERRIDE;

    virtual bool canHaveChildren() const FINAL { return true; }

    template<typename GeneratorContext>
    void generateLineBoxRects(GeneratorContext& yield) const;

    virtual void layout() FINAL { ASSERT_NOT_REACHED(); } // Do nothing for layout()

    virtual void paint(PaintInfo&, const LayoutPoint&) FINAL;

    virtual bool nodeAtPoint(const HitTestRequest&, HitTestResult&, const HitTestLocation& locationInContainer, const LayoutPoint& accumulatedOffset, HitTestAction) FINAL;

    virtual bool requiresLayer() const OVERRIDE { return isInFlowPositioned() || createsGroup() || hasClipPath() || willChangeCreatesStackingContext(); }

    virtual LayoutUnit offsetLeft() const FINAL;
    virtual LayoutUnit offsetTop() const FINAL;
    virtual LayoutUnit offsetWidth() const FINAL { return linesBoundingBox().width(); }
    virtual LayoutUnit offsetHeight() const FINAL { return linesBoundingBox().height(); }
    LayoutPoint firstInlineBoxTopLeft() const;

    virtual LayoutRect clippedOverflowRectForRepaint(RenderLayerModelObject* repaintContainer) const OVERRIDE;
    virtual LayoutRect rectWithOutlineForRepaint(RenderLayerModelObject* repaintContainer, LayoutUnit outlineWidth) const FINAL;

    virtual Optional<LayoutRect> computeVisibleRectInContainer(const LayoutRect&, const RenderLayerModelObject* container, VisibleRectContext) const final;
    LayoutRect computeVisibleRectUsingPaintOffset(const LayoutRect&) const;

    virtual void mapLocalToContainer(RenderLayerModelObject* repaintContainer, TransformState&, OptionSet<MapCoordinatesMode>, bool* wasFixed) const OVERRIDE;
	virtual const RenderElement* pushMappingToContainer(const RenderLayerModelObject* ancestorToStopAt, RenderGeometryMap&) const OVERRIDE;

    virtual VisiblePosition positionForPoint(const LayoutPoint&, const RenderFragmentContainer*) FINAL;

    virtual LayoutRect frameRectForStickyPositioning() const FINAL { return linesBoundingBox(); }

    virtual LayoutRect borderBoundingBox() const final
	{
		return LayoutRect(LayoutPoint(), linesBoundingBox().size());
	}

    virtual PassOwnPtr<InlineFlowBox> createInlineFlowBox(); // Subclassed by SVG and Ruby

    virtual void dirtyLinesFromChangedChild(RenderObject* child) FINAL { m_lineBoxes.dirtyLinesFromChangedChild(this, child); }

    virtual LayoutUnit lineHeight(bool firstLine, LineDirectionMode, LinePositionMode = PositionOnContainingLine) const FINAL;
    virtual LayoutUnit baselinePosition(FontBaseline, bool firstLine, LineDirectionMode, LinePositionMode = PositionOnContainingLine) const FINAL;
    
    virtual void updateHitTestResult(HitTestResult&, const LayoutPoint&) FINAL;

    virtual void imageChanged(WrappedImagePtr, const IntRect* = 0) FINAL;

#if ENABLE(DASHBOARD_SUPPORT)
    virtual void addDashboardRegions(Vector<DashboardRegionValue>&) FINAL;
#endif

    void paintOutlineForLine(GraphicsContext*, const LayoutPoint&, const LayoutRect& prevLine, const LayoutRect& thisLine, const LayoutRect& nextLine, const Color);

    bool willChangeCreatesStackingContext() const
    {
        return style()->willChange() && style()->willChange()->canCreateStackingContext();
    }

    RenderLineBoxList m_lineBoxes;   // All of the line boxes created for this inline flow.  For example, <i>Hello<br>world.</i> will have two <i> line boxes.
};

template<> inline bool isRendererOfType<const RenderInline>(const RenderObject& renderer) { return renderer.isRenderInline(); }
RENDER_OBJECT_TYPE_CASTS(RenderInline, isRenderInline())

inline bool isRenderInline(const RenderObject* renderer)
{
	return renderer && renderer->isRenderInline();
}

inline bool isRenderInline(const RenderObject& renderer)
{
	return renderer.isRenderInline();
}

inline bool isRenderInline(const RenderElement* renderer)
{
	return renderer && renderer->isRenderInline();
}

inline bool isRenderInline(const RenderElement& renderer)
{
	return renderer.isRenderInline();
}

template<>
ALWAYS_INLINE const RenderInline* dynamicDowncast(const RenderBoxModelObject& renderer)
{
    //static_assert(!std::is_same_v<Source, Target>, "Unnecessary cast to same type");
    //static_assert(std::is_base_of_v<Source, Target>, "Should be a downcast");
    return isRenderInline(renderer) ? &toRenderInline(renderer) : nullptr;
}

} // namespace WebCore
