/**
 * Copyright (C) 2006, 2007 Apple Inc. All rights reserved.
 *           (C) 2008 Torch Mobile Inc. All rights reserved. (http://www.torchmobile.com/)  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "RenderTextControl.h"

#include "HTMLTextFormControlElement.h"
#include "HitTestResult.h"
#include "RenderText.h"
#include "RenderTextControlSingleLine.h"
#include "RenderTheme.h"
#include "ScrollbarTheme.h"
#include "TextControlInnerElements.h"
#include "VisiblePosition.h"
#include <wtf/NeverDestroyed.h>
#include <wtf/unicode/CharacterNames.h>

namespace WebCore {

RenderTextControl::RenderTextControl(HTMLTextFormControlElement& element, PassOwnPtr<RenderStyle> style)
    : RenderBlockFlow(element, style)
{
}

RenderTextControl::~RenderTextControl() = default;

HTMLTextFormControlElement& RenderTextControl::textFormControlElement() const
{
    return toHTMLTextFormControlElement(nodeForNonAnonymous());
}

RefPtr<TextControlInnerTextElement> RenderTextControl::innerTextElement() const
{
    return textFormControlElement().innerTextElement();
}

void RenderTextControl::styleDidChange(StyleDifference diff, const RenderStyle* oldStyle)
{
	RenderBlockFlow::styleDidChange(diff, oldStyle);
	RefPtr<TextControlInnerTextElement> innerText = innerTextElement();
    if (!innerText)
        return;
    RenderTextControlInnerBlock* innerTextRenderer = innerText->renderer();
    if (innerTextRenderer && oldStyle) {
		// FIXME: The height property of the inner text block style may be mutated by RenderTextControlSingleLine::layout.
		// See if the original has changed before setting it and triggering a layout.
    	OwnPtr<RenderStyle> newInnerTextStyle = textFormControlElement().createInnerTextStyle(*style());
    	OwnPtr<RenderStyle> oldInnerTextStyle = textFormControlElement().createInnerTextStyle(*oldStyle);
		if (*newInnerTextStyle != *oldInnerTextStyle)
			innerTextRenderer->setStyle(newInnerTextStyle.release());
		else if (diff == StyleDifference::StyleDifferenceRepaintIfText || diff == StyleDifference::StyleDifferenceRepaint) {
			// Repaint is expected to be propagated down to the shadow tree when non-inherited style property changes
			// (e.g. text-decoration-color) since that's where the value actually takes effect.
			innerTextRenderer->repaint();
		}
	}
    textFormControlElement().updatePlaceholderVisibility();
}

int RenderTextControl::scrollbarThickness() const
{
    // FIXME: We should get the size of the scrollbar from the RenderTheme instead.
    return ScrollbarTheme::theme()->scrollbarThickness();
}

RenderBox::LogicalExtentComputedValues RenderTextControl::computeLogicalHeight(LayoutUnit logicalHeight, LayoutUnit logicalTop) const
{
	RefPtr<TextControlInnerTextElement> innerText = innerTextElement();
    if (!innerText)
        return RenderBox::computeLogicalHeight(LayoutUnit(), LayoutUnit());

    if (RenderBox* innerTextBox = innerText->renderBox()) {
		LayoutUnit nonContentHeight = innerTextBox->verticalBorderAndPaddingExtent() + innerTextBox->verticalMarginExtent();
		logicalHeight = computeControlLogicalHeight(innerTextBox->lineHeight(true, HorizontalLine, PositionOfInteriorLineBoxes), nonContentHeight);

		// We are able to have a horizontal scrollbar if the overflow style is scroll, or if its auto and there's no word wrap.
		auto shouldIncludeScrollbarHeight = [&] {
			auto& style = *this->style();
			auto isHorizontalWritingMode = this->isHorizontalWritingMode();
			return (isHorizontalWritingMode && style.overflowX() == Overflow::Scroll) || (!isHorizontalWritingMode && style.overflowY() == Overflow::Scroll);
		};
		if (shouldIncludeScrollbarHeight())
			logicalHeight += scrollbarThickness();

        // FIXME: The logical height of the inner text box should have been added
        // before calling computeLogicalHeight to avoid this hack.
        cacheIntrinsicContentLogicalHeightForFlexItem(logicalHeight);

        logicalHeight += verticalBorderAndPaddingExtent();
    }

    return RenderBox::computeLogicalHeight(logicalHeight, logicalTop);
}

void RenderTextControl::hitInnerTextElement(HitTestResult& result, const LayoutPoint& pointInContainer, const LayoutPoint& accumulatedOffset)
{
	RefPtr<TextControlInnerTextElement> innerText = innerTextElement();
    if (!innerText)
        return;
    if (!innerText->renderer())
        return;

    LayoutPoint adjustedLocation = accumulatedOffset + location();
    LayoutPoint localPoint = pointInContainer - toLayoutSize(adjustedLocation + innerText->renderBox()->location()) + toLayoutSize(scrollPosition());
    result.setInnerNode(innerText.get());
    result.setInnerNonSharedNode(innerText.get());
    result.setLocalPoint(localPoint);
}

static ASCIILiteral fontFamiliesWithInvalidCharWidth[] = {
    ASCIILiteral("American Typewriter"),
    ASCIILiteral("Arial Hebrew"),
    ASCIILiteral("Chalkboard"),
    ASCIILiteral("Cochin"),
    ASCIILiteral("Corsiva Hebrew"),
    ASCIILiteral("Courier"),
    ASCIILiteral("Euphemia UCAS"),
    ASCIILiteral("Geneva"),
    ASCIILiteral("Gill Sans"),
    ASCIILiteral("Hei"),
    ASCIILiteral("Helvetica"),
    ASCIILiteral("Hoefler Text"),
    ASCIILiteral("InaiMathi"),
    ASCIILiteral("Kai"),
    ASCIILiteral("Lucida Grande"),
    ASCIILiteral("Marker Felt"),
    ASCIILiteral("Monaco"),
    ASCIILiteral("Mshtakan"),
    ASCIILiteral("New Peninim MT"),
    ASCIILiteral("Osaka"),
    ASCIILiteral("Raanana"),
    ASCIILiteral("STHeiti"),
    ASCIILiteral("Symbol"),
    ASCIILiteral("Times"),
    ASCIILiteral("Apple Braille"),
    ASCIILiteral("Apple LiGothic"),
    ASCIILiteral("Apple LiSung"),
    ASCIILiteral("Apple Symbols"),
    ASCIILiteral("AppleGothic"),
    ASCIILiteral("AppleMyungjo"),
    ASCIILiteral("#GungSeo"),
    ASCIILiteral("#HeadLineA"),
    ASCIILiteral("#PCMyungjo"),
    ASCIILiteral("#PilGi"),
};

// For font families where any of the fonts don't have a valid entry in the OS/2 table
// for avgCharWidth, fallback to the legacy webkit behavior of getting the avgCharWidth
// from the width of a '0'. This only seems to apply to a fixed number of Mac fonts,
// but, in order to get similar rendering across platforms, we do this check for
// all platforms.
bool RenderTextControl::hasValidAvgCharWidth(const AtomicString& family)
{
    static NeverDestroyed<HashSet<AtomicString>> fontFamiliesWithInvalidCharWidthMap;

    if (family.isEmpty())
        return false;

    if (fontFamiliesWithInvalidCharWidthMap->isEmpty()) {
        for (size_t i = 0; i < WTF_ARRAY_LENGTH(fontFamiliesWithInvalidCharWidth); ++i)
            fontFamiliesWithInvalidCharWidthMap->add(AtomicString(fontFamiliesWithInvalidCharWidth[i]));
    }

    return !fontFamiliesWithInvalidCharWidthMap->contains(family);
}

float RenderTextControl::getAvgCharWidth(const AtomicString& family)
{
    if (hasValidAvgCharWidth(family))
        return roundf(style()->font().primaryFont()->avgCharWidth());

    const UChar ch = '0';
    const String str = String(&ch, 1);
    const FontCascade& font = style()->font();
    TextRun textRun = constructTextRun(str, *style(), AllowTrailingExpansion);
    return font.width(textRun);
}

float RenderTextControl::scaleEmToUnits(int x) const
{
    // This matches the unitsPerEm value for MS Shell Dlg and Courier New from the "head" font table.
    float unitsPerEm = 2048.0f;
    return roundf(style()->font().size() * x / unitsPerEm);
}

void RenderTextControl::computeIntrinsicLogicalWidths(LayoutUnit& minLogicalWidth, LayoutUnit& maxLogicalWidth) const
{
    // Use average character width. Matches IE.
    const AtomicString& family = style()->font().firstFamily();
    maxLogicalWidth = preferredContentLogicalWidth(const_cast<RenderTextControl*>(this)->getAvgCharWidth(family));
	RefPtr<TextControlInnerTextElement> innerText = innerTextElement();
    if (RenderBox* innerTextRenderBox = innerText ? innerText->renderBox() : nullptr)
        maxLogicalWidth += innerTextRenderBox->paddingStart() + innerTextRenderBox->paddingEnd();
    if (!style()->logicalWidth().isPercentOrCalculated())
        minLogicalWidth = maxLogicalWidth;
}

void RenderTextControl::computePreferredLogicalWidths()
{
    ASSERT(preferredLogicalWidthsDirty());

    m_minPreferredLogicalWidth = 0;
    m_maxPreferredLogicalWidth = 0;

    if (style()->logicalWidth().isFixed() && style()->logicalWidth().value() >= 0)
        m_minPreferredLogicalWidth = m_maxPreferredLogicalWidth = adjustContentBoxLogicalWidthForBoxSizing(style()->logicalWidth());
    else
		computeIntrinsicLogicalWidths(m_minPreferredLogicalWidth, m_maxPreferredLogicalWidth);

    RenderBox::computePreferredLogicalWidths(style()->logicalMinWidth(), style()->logicalMaxWidth(), borderAndPaddingLogicalWidth());

    setPreferredLogicalWidthsDirty(false);
}

void RenderTextControl::addFocusRingRects(Vector<IntRect>& rects, const LayoutPoint& additionalOffset, const RenderLayerModelObject*)
{
    if (!size().isEmpty())
        rects.append(snappedIntRect(additionalOffset, size()));
}

void RenderTextControl::layoutExcludedChildren(bool relayoutChildren)
{
    RenderBlockFlow::layoutExcludedChildren(relayoutChildren);

    HTMLElement* placeholder = textFormControlElement().placeholderElement();
    RenderElement* placeholderRenderer = placeholder ? placeholder->renderer() : 0;
    if (!placeholderRenderer)
        return;
    placeholderRenderer->setIsExcludedFromNormalLayout(true);

    if (relayoutChildren) {
        // The markParents arguments should be false because this function is
        // called from layout() of the parent and the placeholder layout doesn't
        // affect the parent layout.
        placeholderRenderer->setChildNeedsLayout(MarkOnlyThis);
    }
}

#if PLATFORM(IOS)
bool RenderTextControl::canScroll() const
{
	RefPtr<TextControlInnerTextElement> innerText = innerTextElement();
    return innerText && innerText->renderer() && innerText->renderer()->hasNonVisibleOverflow();
}

LayoutUnit RenderTextControl::innerLineHeight() const
{
	RefPtr<TextControlInnerTextElement> innerText = innerTextElement();
    if (innerText && innerText->renderer())
        return innerText->renderer()->style().computedLineHeight();
    return style().computedLineHeight();
}
#endif

} // namespace WebCore
