/*
 * Copyright (C) 2019 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "LineLayoutTraversal.h"

#include "InlineTextBox.h"
#include "RenderBlockFlow.h"
#include "RenderText.h"

namespace WebCore {
namespace LineLayoutTraversal {

FloatRect TextBox::rect() const
{
	auto& path = iterator().m_complexPath;
	return path.inlineTextBox->frameRect();
}

FloatRect TextBox::logicalRect() const
{
	auto& path = iterator().m_complexPath;
    return path.inlineTextBox->logicalFrameRect();
}

bool TextBox::hasHyphen() const
{
	auto& path = iterator().m_complexPath;
	return path.inlineTextBox->hasHyphen();
}

bool TextBox::isLeftToRightDirection() const
{
	auto& path = iterator().m_complexPath;
	return path.inlineTextBox->isLeftToRightDirection();
}

bool TextBox::dirOverride() const
{
	auto& path = iterator().m_complexPath;
	return path.inlineTextBox->dirOverride();
}

StringView TextBox::text() const
{
	auto& path = iterator().m_complexPath;
	return StringView(path.inlineTextBox->renderer().text()).substring(path.inlineTextBox->start(), path.inlineTextBox->len());
}

bool TextBox::isLineBreak() const
{
	auto& path = iterator().m_complexPath;
    return path.inlineTextBox->isLineBreak();
}

unsigned TextBox::localStartOffset() const
{
	auto& path = iterator().m_complexPath;
	return path.inlineTextBox->start();
}

unsigned TextBox::localEndOffset() const
{
	auto& path = iterator().m_complexPath;
	return path.inlineTextBox->end();
}

unsigned TextBox::length() const
{
	auto& path = iterator().m_complexPath;
	return path.inlineTextBox->len();
}

bool TextBox::isLastOnLine() const
{
	auto& path = iterator().m_complexPath;
	auto* next = path.nextInlineTextBoxInTextOrder();
	return !next || path.inlineTextBox->root() != next->root();
}

bool TextBox::isLast() const
{
	auto& path = iterator().m_complexPath;
	return !path.nextInlineTextBoxInTextOrder();
}

inline const TextBoxIterator& TextBox::iterator() const
{
    return static_cast<const TextBoxIterator&>(*this);
}

TextBoxIterator::TextBoxIterator(const InlineTextBox* inlineTextBox, Vector<const InlineTextBox*>&& sorted, size_t index)
    : m_complexPath(inlineTextBox, WTFMove(sorted), index)
{
}

TextBoxIterator& TextBoxIterator::traverseNextInVisualOrder()
{
	auto& path = m_complexPath;
	path.inlineTextBox = path.inlineTextBox->nextTextBox();
    return *this;
}

const InlineTextBox* TextBoxIterator::ComplexPath::nextInlineTextBoxInTextOrder() const
{
    if (!sortedInlineTextBoxes.isEmpty()) {
        if (sortedInlineTextBoxIndex + 1 < sortedInlineTextBoxes.size())
            return sortedInlineTextBoxes[sortedInlineTextBoxIndex + 1];
        return nullptr;
    }
    return inlineTextBox->nextTextBox();
}

TextBoxIterator& TextBoxIterator::traverseNextInTextOrder()
{
	auto& path = m_complexPath;
	path.inlineTextBox = path.nextInlineTextBoxInTextOrder();
	if (!path.sortedInlineTextBoxes.isEmpty())
		++path.sortedInlineTextBoxIndex;

    return *this;
}

bool TextBoxIterator::operator==(const TextBoxIterator& other) const
{
	auto& path = m_complexPath;
	return path.inlineTextBox == other.m_complexPath.inlineTextBox;
}

bool TextBoxIterator::atEnd() const
{
	auto& path = m_complexPath;
	return !path.inlineTextBox;
}

static const RenderBlockFlow& flowForText(const RenderText& text)
{
    return toRenderBlockFlow(*RenderObject::containingBlockForPositionType(PositionType::Static, text));
}

TextBoxIterator firstTextBoxFor(const RenderText& text)
{
    return TextBoxIterator(text.firstTextBox(), Vector<const InlineTextBox*>(), 0);
}

TextBoxIterator firstTextBoxInTextOrderFor(const RenderText& text)
{
    auto& flow = flowForText(text);

    if (flow.complexLineLayout() && text.containsReversedText() && text.firstTextBox()) {
        Vector<const InlineTextBox*> sortedTextBoxes;
        for (auto* textBox = text.firstTextBox(); textBox; textBox = textBox->nextTextBox())
            sortedTextBoxes.append(textBox);
        std::sort(sortedTextBoxes.begin(), sortedTextBoxes.end(), InlineTextBox::compareByStart);
        auto* first = sortedTextBoxes[0];
        return TextBoxIterator(first, WTFMove(sortedTextBoxes), 0);
    }

    return firstTextBoxFor(text);
}

TextBoxRange textBoxesFor(const RenderText& text)
{
    return TextBoxRange(firstTextBoxFor(text));
}

}
}
