/*
 * Copyright (C) 2000 Lars Knoll (knoll@kde.org)
 *           (C) 2000 Antti Koivisto (koivisto@kde.org)
 *           (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2003, 2005, 2006, 2007, 2008, 2013 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#pragma once

#include <wtf/DataRef.h>
#include "LayoutRect.h"
#include "LayoutSize.h"
#include "LengthBox.h"
#include "StyleImage.h"
#include <wtf/Optional.h>
#include <wtf/Vector.h>

namespace WebCore {

enum ENinePieceImageRule : uint8_t {
    StretchImageRule, RoundImageRule, SpaceImageRule, RepeatImageRule
};

enum ImagePiece {
    MinPiece = 0,
    TopLeftPiece = MinPiece,
    LeftPiece,
    BottomLeftPiece,
    TopRightPiece,
    RightPiece,
    BottomRightPiece,
    TopPiece,
    BottomPiece,
    MiddlePiece,
    MaxPiece
};

inline ImagePiece& operator++(ImagePiece& piece)
{
    piece = static_cast<ImagePiece>(static_cast<int>(piece) + 1);
    return piece;
}

inline bool isCornerPiece(ImagePiece piece)
{
    return piece == TopLeftPiece || piece == TopRightPiece || piece == BottomLeftPiece || piece == BottomRightPiece;
}

inline bool isMiddlePiece(ImagePiece piece)
{
    return piece == MiddlePiece;
}

inline bool isHorizontalPiece(ImagePiece piece)
{
    return piece == TopPiece || piece == BottomPiece || piece == MiddlePiece;
}

inline bool isVerticalPiece(ImagePiece piece)
{
    return piece == LeftPiece || piece == RightPiece || piece == MiddlePiece;
}

inline Optional<PhysicalBoxSide> imagePieceHorizontalSide(ImagePiece piece)
{
    if (piece == TopLeftPiece || piece == TopPiece || piece == TopRightPiece)
        return TopSide;

    if (piece == BottomLeftPiece || piece == BottomPiece || piece == BottomRightPiece)
        return BottomSide;

    return Nullopt;
}

inline Optional<PhysicalBoxSide> imagePieceVerticalSide(ImagePiece piece)
{
    if (piece == TopLeftPiece || piece == LeftPiece || piece == BottomLeftPiece)
        return LeftSide;

    if (piece == TopRightPiece || piece == RightPiece || piece == BottomRightPiece)
        return RightSide;

    return Nullopt;
}

class RenderStyle;

class NinePieceImageData : public RefCounted<NinePieceImageData> {
public:
    static Ref<NinePieceImageData> create() { return adoptRef(*new NinePieceImageData); }
    Ref<NinePieceImageData> copy() const { return adoptRef(*new NinePieceImageData(*this)); }

    bool operator==(const NinePieceImageData&) const;
    bool operator!=(const NinePieceImageData& o) const { return !(*this == o); }

    bool fill { false };
    ENinePieceImageRule horizontalRule { StretchImageRule };
    ENinePieceImageRule verticalRule { StretchImageRule };
    RefPtr<StyleImage> image;
    LengthBox imageSlices { Length(100, LengthType::Percent), Length(100, LengthType::Percent), Length(100, LengthType::Percent), Length(100, LengthType::Percent) };
    LengthBox borderSlices { Length(1, LengthType::Relative), Length(1, LengthType::Relative), Length(1, LengthType::Relative), Length(1, LengthType::Relative) };
    LengthBox outset { LengthType::Relative };

private:
    NinePieceImageData();
    NinePieceImageData(const NinePieceImageData&);
};

class NinePieceImage {
	WTF_MAKE_FAST_ALLOCATED;
public:
    enum class Type {
        Normal,
        Mask
    };

    NinePieceImage(Type = Type::Normal);
    NinePieceImage(PassRefPtr<StyleImage>, const LengthBox& imageSlices, bool fill, const LengthBox& borderSlices, const LengthBox& outset, ENinePieceImageRule horizontalRule, ENinePieceImageRule verticalRule);

    ALWAYS_INLINE bool operator==(const NinePieceImage& other) const { return m_data == other.m_data; }
    ALWAYS_INLINE bool operator!=(const NinePieceImage& other) const { return m_data != other.m_data; }

    ALWAYS_INLINE bool hasImage() const { return m_data->image.get(); }
    ALWAYS_INLINE StyleImage* image() const { return m_data->image.get(); }
    void setImage(PassRefPtr<StyleImage> image) { m_data.access().image = image; }
    
    ALWAYS_INLINE const LengthBox& imageSlices() const { return m_data->imageSlices; }
    void setImageSlices(const LengthBox& slices) { m_data.access().imageSlices = slices; }

    ALWAYS_INLINE bool fill() const { return m_data->fill; }
    void setFill(bool fill) { m_data.access().fill = fill; }

    ALWAYS_INLINE const LengthBox& borderSlices() const { return m_data->borderSlices; }
    void setBorderSlices(const LengthBox& slices) { m_data.access().borderSlices = slices; }

    ALWAYS_INLINE const LengthBox& outset() const { return m_data->outset; }
    void setOutset(const LengthBox& outset) { m_data.access().outset = outset; }

    ALWAYS_INLINE ENinePieceImageRule horizontalRule() const { return static_cast<ENinePieceImageRule>(m_data->horizontalRule); }
    void setHorizontalRule(ENinePieceImageRule rule) { m_data.access().horizontalRule = rule; }
    
    ALWAYS_INLINE ENinePieceImageRule verticalRule() const { return static_cast<ENinePieceImageRule>(m_data->verticalRule); }
    void setVerticalRule(ENinePieceImageRule rule) { m_data.access().verticalRule = rule; }

    void copyImageSlicesFrom(const NinePieceImage& other)
    {
        m_data.access().imageSlices = other.m_data->imageSlices;
        m_data.access().fill = other.m_data->fill;
    }

    void copyBorderSlicesFrom(const NinePieceImage& other)
    {
        m_data.access().borderSlices = other.m_data->borderSlices;
    }
    
    void copyOutsetFrom(const NinePieceImage& other)
    {
        m_data.access().outset = other.m_data->outset;
    }

    void copyRepeatFrom(const NinePieceImage& other)
    {
        m_data.access().horizontalRule = other.m_data->horizontalRule;
        m_data.access().verticalRule = other.m_data->verticalRule;
    }

    static LayoutUnit computeOutset(const Length& outset, LayoutUnit borderWidth)
    {
        if (outset.isRelative())
            return LayoutUnit(outset.value() * borderWidth);
        return LayoutUnit(outset.value());
    }

    static LayoutUnit computeSlice(Length, LayoutUnit width, LayoutUnit slice, LayoutUnit extent);
    static LayoutBoxExtent computeSlices(const LayoutSize&, const LengthBox& lengths, int scaleFactor);
    static LayoutBoxExtent computeSlices(const LayoutSize&, const LengthBox& lengths, const FloatBoxExtent& widths, const LayoutBoxExtent& slices);

    static bool isEmptyPieceRect(ImagePiece, const LayoutBoxExtent& slices);
    static bool isEmptyPieceRect(ImagePiece, const Vector<FloatRect>& destinationRects, const Vector<FloatRect>& sourceRects);

    static Vector<FloatRect> computeNineRects(const FloatRect& outer, const LayoutBoxExtent& slices, float deviceScaleFactor);

    static void scaleSlicesIfNeeded(const LayoutSize&, LayoutBoxExtent& slices, float deviceScaleFactor);

    static FloatSize computeSideTileScale(ImagePiece, const Vector<FloatRect>& destinationRects, const Vector<FloatRect>& sourceRects);
    static FloatSize computeMiddleTileScale(const Vector<FloatSize>& scales, const Vector<FloatRect>& destinationRects, const Vector<FloatRect>& sourceRects, ENinePieceImageRule hRule, ENinePieceImageRule vRule);
    static Vector<FloatSize> computeTileScales(const Vector<FloatRect>& destinationRects, const Vector<FloatRect>& sourceRects, ENinePieceImageRule hRule, ENinePieceImageRule vRule);

    void paint(GraphicsContext&, RenderElement*, const RenderStyle&, const LayoutRect& destination, const LayoutSize& source, float deviceScaleFactor, CompositeOperator) const;

private:

    static DataRef<NinePieceImageData>& defaultMaskData();

    DataRef<NinePieceImageData> m_data;
};

} // namespace WebCore
