/*
 * Copyright (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008, 2009, 2010 Apple Inc. All rights reserved.
 * Copyright (C) 2011 Adobe Systems Incorporated. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "RenderStyle.h"

#include "ContentData.h"
#include "CSSCustomPropertyValue.h"
#include "CSSPropertyNames.h"
#include "CursorList.h"
#include "CSSWrapShapes.h"
#include "FloatRoundedRect.h"
#include "FontSelector.h"
#include "InlineTextBoxStyle.h"
#include "Pagination.h"
#include "QuotesData.h"
#include "RenderBlock.h"
#include "RenderObject.h"
#include "RenderTheme.h"
#include "SVGPathData.h"
#include "ScaleTransformOperation.h"
#include "ShadowData.h"
#include "StyleBuilderConverter.h"
#include "StyleImage.h"
#include "StyleResolver.h"
#include "StyleSelfAlignmentData.h"
#include "StyleTreeResolver.h"
#include "WillChangeData.h"
#include <wtf/NeverDestroyed.h>
#include <wtf/MathExtras.h>
#include <wtf/OwnPtrComparison.h>
#include <wtf/PointerComparison.h>
#include <wtf/StdLibExtras.h>
#include <algorithm>

namespace WebCore {

namespace {
struct SameSizeAsBorderValue {
    Color m_color;
    float m_width;
    int m_restBits;
};

COMPILE_ASSERT(sizeof(BorderValue) == sizeof(SameSizeAsBorderValue), BorderValue_should_not_grow);

struct SameSizeAsRenderStyle {
    void* dataRefs1[5];
    struct InheritedFlags {
    	uint64_t m_bitfields;
    } m_inheritedFlags;

    void* dataRefs2[2];
    struct NonInheritedFlags {
    	uint64_t m_bitfields;
    } m_nonInheritedFlags;
    void* ownPtrs[1];
#if ENABLE(SVG)
    void* dataRefSvgStyle;
#endif
#if !ASSERT_DISABLED || ENABLE(SECURITY_ASSERTIONS)
    bool deletionCheck;
#endif
};

static_assert(sizeof(RenderStyle) <= sizeof(SameSizeAsRenderStyle), "RenderStyle should stay small");
}

static_assert(!((static_cast<unsigned>(PseudoId::AfterLastInternalPseudoId) - 1) >> StyleTypeBits), "StyleTypeBits is small");

static_assert(sizeof(RenderStyle::NonInheritedFlags) == sizeof(uint64_t), "NonInheritedFlags is too big");
static_assert(sizeof(RenderStyle::InheritedFlags) == sizeof(uint64_t), "InheritedFlags is too big");

NEVER_INLINE RenderStyle& RenderStyle::defaultStyle()
{
    static RenderStyle& style = *RenderStyle::createDefaultStyle().leakPtr();
    return style;
}

PassOwnPtr<RenderStyle> RenderStyle::create()
{
    return clone(defaultStyle());
}

PassOwnPtr<RenderStyle> RenderStyle::createPtr()
{
    return clone(defaultStyle());
}

PassOwnPtr<RenderStyle> RenderStyle::createDefaultStyle()
{
    return adoptPtr(new RenderStyle(CreateDefaultStyle));
}

PassOwnPtr<RenderStyle> RenderStyle::createAnonymousStyleWithDisplay(const RenderStyle& parentStyle, DisplayType display)
{
    OwnPtr<RenderStyle> newStyle = RenderStyle::create();
    newStyle->inheritFrom(parentStyle);
    newStyle->inheritUnicodeBidiFrom(&parentStyle);
    newStyle->setDisplay(display);
    return newStyle.release();
}

PassOwnPtr<RenderStyle> RenderStyle::clone(const RenderStyle& other)
{
    return adoptPtr(new RenderStyle(other, Clone));
}

PassOwnPtr<RenderStyle> RenderStyle::cloneIncludingPseudoElements(const RenderStyle& style)
{
	OwnPtr<RenderStyle> newStyle = adoptPtr(new RenderStyle(style, Clone));
    newStyle->copyPseudoElementsFrom(style);
    return newStyle.release();
}

PassOwnPtr<RenderStyle> RenderStyle::clonePtr(const RenderStyle& other)
{
    return adoptPtr(new RenderStyle(other, Clone));
}

PassOwnPtr<RenderStyle> RenderStyle::createStyleInheritingFromPseudoStyle(const RenderStyle& pseudoStyle)
{
    ASSERT(pseudoStyle.styleType() == PseudoId::Before || pseudoStyle.styleType() == AFTER);

    // Images are special and must inherit the pseudoStyle so the width and height of
    // the pseudo element doesn't change the size of the image. In all other cases we
    // can just share the style.
    OwnPtr<RenderStyle> style = RenderStyle::create();
    style->inheritFrom(pseudoStyle);
    return style.release();
}

inline RenderStyle::RenderStyle(CreateDefaultStyleTag)
: m_boxData(StyleBoxData::create())
, visual(StyleVisualData::create())
, m_background(StyleBackgroundData::create())
, m_surroundData(StyleSurroundData::create())
, m_rareNonInheritedData(StyleRareNonInheritedData::create())
, m_rareInheritedData(StyleRareInheritedData::create())
, m_inheritedData(StyleInheritedData::create())
#if ENABLE(SVG)
, m_svgStyle(SVGRenderStyle::create())
#endif
{
    m_inheritedFlags.u._empty_cells = initialEmptyCells();
    m_inheritedFlags.u._caption_side = initialCaptionSide();
    m_inheritedFlags.u._list_style_type = initialListStyleType();
    m_inheritedFlags.u._list_style_position = initialListStylePosition();
    m_inheritedFlags.u._visibility = initialVisibility();
    m_inheritedFlags.u._text_align = static_cast<unsigned>(initialTextAlign());
    m_inheritedFlags.u._text_transform = initialTextTransform();
    m_inheritedFlags.u._text_decorations = initialTextDecoration();
    m_inheritedFlags.u._cursor_style = initialCursor();
#if ENABLE(CURSOR_VISIBILITY)
    m_inheritedFlags.u.m_cursorVisibility = initialCursorVisibility();
#endif
    m_inheritedFlags.u._direction = initialDirection();
    m_inheritedFlags.u._white_space = static_cast<unsigned>(initialWhiteSpace());
    m_inheritedFlags.u._border_collapse = initialBorderCollapse();
    m_inheritedFlags.u.m_rtlOrdering = initialRTLOrdering();
    m_inheritedFlags.u._box_direction = initialBoxDirection();
    m_inheritedFlags.u.m_printColorAdjust = initialPrintColorAdjust();
    m_inheritedFlags.u._pointerEvents = initialPointerEvents();
    m_inheritedFlags.u._insideLink = NotInsideLink;
    m_inheritedFlags.u.m_writingMode = initialWritingMode();

    m_nonInheritedFlags.u.effectiveDisplay = static_cast<unsigned>(initialDisplay());
    m_nonInheritedFlags.u.originalDisplay = static_cast<unsigned>(initialDisplay());
    m_nonInheritedFlags.u.overflowX = static_cast<unsigned>(initialOverflowX());
    m_nonInheritedFlags.u.overflowY = static_cast<unsigned>(initialOverflowY());
    m_nonInheritedFlags.u.verticalAlign = static_cast<unsigned>(initialVerticalAlign());
    m_nonInheritedFlags.u.clear = static_cast<unsigned>(initialClear());
    m_nonInheritedFlags.u.position = static_cast<unsigned>(initialPosition());
    m_nonInheritedFlags.u.unicodeBidi = static_cast<unsigned>(initialUnicodeBidi());
    m_nonInheritedFlags.u.floating = static_cast<unsigned>(initialFloating());
    m_nonInheritedFlags.u.tableLayout = static_cast<unsigned>(initialTableLayout());
    m_nonInheritedFlags.u.hasExplicitlySetDirection = false;
    m_nonInheritedFlags.u.hasExplicitlySetWritingMode = false;
    m_nonInheritedFlags.u.usesViewportUnits = false;
    m_nonInheritedFlags.u.usesContainerUnits = false;
    m_nonInheritedFlags.u.hasExplicitlyInheritedProperties = false;
    m_nonInheritedFlags.u.disallowsFastPathInheritance = false;
    m_nonInheritedFlags.u.isUnique = false;
    m_nonInheritedFlags.u.emptyState = false;
    m_nonInheritedFlags.u.firstChildState = false;
    m_nonInheritedFlags.u.lastChildState = false;
    m_nonInheritedFlags.u.isLink = false;
    m_nonInheritedFlags.u.styleType = static_cast<unsigned>(PseudoId::None);
    m_nonInheritedFlags.u.pseudoBits = static_cast<unsigned>(PseudoId::None);

    COMPILE_ASSERT((sizeof(InheritedFlags) <= 8), InheritedFlags_does_not_grow);
    COMPILE_ASSERT((sizeof(NonInheritedFlags) <= 8), NonInheritedFlags_does_not_grow);
}

inline RenderStyle::RenderStyle(const RenderStyle& o, CloneTag)
    : m_boxData(o.m_boxData)
    , visual(o.visual)
    , m_background(o.m_background)
    , m_surroundData(o.m_surroundData)
    , m_rareNonInheritedData(o.m_rareNonInheritedData)
    , m_rareInheritedData(o.m_rareInheritedData)
    , m_inheritedData(o.m_inheritedData)
#if ENABLE(SVG)
    , m_svgStyle(o.m_svgStyle)
#endif
    , m_inheritedFlags(o.m_inheritedFlags)
    , m_nonInheritedFlags(o.m_nonInheritedFlags)
{
}

RenderStyle::~RenderStyle()
{
#if !ASSERT_DISABLED || ENABLE(SECURITY_ASSERTIONS)
    ASSERT_WITH_SECURITY_IMPLICATION(!m_deletionHasBegun);
    m_deletionHasBegun = true;
#endif
}

static StyleSelfAlignmentData resolvedSelfAlignment(const StyleSelfAlignmentData& value, ItemPosition normalValueBehavior)
{
	if (value.position() == ItemPositionLegacy || value.position() == ItemPositionNormal || value.position() == ItemPositionAuto)
		return StyleSelfAlignmentData(normalValueBehavior, OverflowAlignmentDefault);
    return value;
}

StyleSelfAlignmentData RenderStyle::resolvedAlignItems(ItemPosition normalValueBehaviour) const
{
    return resolvedSelfAlignment(alignItems(), normalValueBehaviour);
}

StyleSelfAlignmentData RenderStyle::resolvedAlignSelf(const RenderStyle* parentStyle, ItemPosition normalValueBehaviour) const
{
    // The auto keyword computes to the parent's align-items computed value.
    // We will return the behaviour of 'normal' value if needed, which is specific of each layout model.
	if (!parentStyle || alignSelf().position() != ItemPositionAuto)
		return resolvedSelfAlignment(alignSelf(), normalValueBehaviour);
	return parentStyle->resolvedAlignItems(normalValueBehaviour);
}

StyleSelfAlignmentData RenderStyle::resolvedJustifyItems(ItemPosition normalValueBehaviour) const
{
    return resolvedSelfAlignment(justifyItems(), normalValueBehaviour);
}

StyleSelfAlignmentData RenderStyle::resolvedJustifySelf(const RenderStyle* parentStyle, ItemPosition normalValueBehaviour) const
{
    // The auto keyword computes to the parent's justify-items computed value.
    // We will return the behaviour of 'normal' value if needed, which is specific of each layout model.
	 if (!parentStyle || justifySelf().position() != ItemPositionAuto)
		return resolvedSelfAlignment(justifySelf(), normalValueBehaviour);
	return parentStyle->resolvedJustifyItems(normalValueBehaviour);
}

static inline StyleContentAlignmentData resolvedContentAlignment(const StyleContentAlignmentData& value, const StyleContentAlignmentData& normalValueBehavior)
{
    return (value.position() == ContentPositionNormal && value.distribution() == ContentDistributionDefault) ? normalValueBehavior : value;
}

StyleContentAlignmentData RenderStyle::resolvedAlignContent(const StyleContentAlignmentData& normalValueBehavior) const
{
    // We will return the behaviour of 'normal' value if needed, which is specific of each layout model.
    return resolvedContentAlignment(alignContent(), normalValueBehavior);
}

StyleContentAlignmentData RenderStyle::resolvedJustifyContent(const StyleContentAlignmentData& normalValueBehavior) const
{
    // We will return the behaviour of 'normal' value if needed, which is specific of each layout model.
    return resolvedContentAlignment(justifyContent(), normalValueBehavior);
}

static inline ContentPosition resolvedContentAlignmentPosition(const StyleContentAlignmentData& value, const StyleContentAlignmentData& normalValueBehavior)
{
    return (value.position() == ContentPositionNormal && value.distribution() == ContentDistributionDefault) ? normalValueBehavior.position() : value.position();
}

static inline ContentDistributionType resolvedContentAlignmentDistribution(const StyleContentAlignmentData& value, const StyleContentAlignmentData& normalValueBehavior)
{
    return (value.position() == ContentPositionNormal && value.distribution() == ContentDistributionDefault) ? normalValueBehavior.distribution() : value.distribution();
}

ContentPosition RenderStyle::resolvedJustifyContentPosition(const StyleContentAlignmentData& normalValueBehavior) const
{
    return resolvedContentAlignmentPosition(justifyContent(), normalValueBehavior);
}

ContentDistributionType RenderStyle::resolvedJustifyContentDistribution(const StyleContentAlignmentData& normalValueBehavior) const
{
    return resolvedContentAlignmentDistribution(justifyContent(), normalValueBehavior);
}

ContentPosition RenderStyle::resolvedAlignContentPosition(const StyleContentAlignmentData& normalValueBehavior) const
{
    return resolvedContentAlignmentPosition(alignContent(), normalValueBehavior);
}

ContentDistributionType RenderStyle::resolvedAlignContentDistribution(const StyleContentAlignmentData& normalValueBehavior) const
{
    return resolvedContentAlignmentDistribution(alignContent(), normalValueBehavior);
}

void RenderStyle::inheritFrom(const RenderStyle& inheritParent)
{
	m_rareInheritedData = inheritParent.m_rareInheritedData;
    m_inheritedData = inheritParent.m_inheritedData;
    m_inheritedFlags = inheritParent.m_inheritedFlags;

#if ENABLE(SVG)
    if (m_svgStyle != inheritParent.m_svgStyle)
        m_svgStyle.access().inheritFrom(inheritParent.m_svgStyle);
#endif
}

void RenderStyle::fastPathInheritFrom(const RenderStyle& inheritParent)
{
    ASSERT(!disallowsFastPathInheritance());

    if (m_inheritedData.ptr() == inheritParent.m_inheritedData.ptr())
        return;

    // FIXME: Use this mechanism for other properties too, like variables.
    if (m_inheritedData->nonFastPathInheritedEqual(*inheritParent.m_inheritedData)) {
        m_inheritedData = inheritParent.m_inheritedData;
        return;
    }
    m_inheritedData.access().fastPathInheritFrom(*inheritParent.m_inheritedData);
}

void RenderStyle::copyNonInheritedFrom(const RenderStyle* other)
{
    m_boxData = other->m_boxData;
    visual = other->visual;
    m_background = other->m_background;
    m_surroundData = other->m_surroundData;
    m_rareNonInheritedData = other->m_rareNonInheritedData;
    m_nonInheritedFlags.copyNonInheritedFrom(other->m_nonInheritedFlags);

#if ENABLE(SVG)
    if (m_svgStyle != other->m_svgStyle)
        m_svgStyle.access().copyNonInheritedFrom(other->m_svgStyle);
#endif

    ASSERT(zoom() == initialZoom());
}

void RenderStyle::copyContentFrom(const RenderStyle& other)
{
    if (!other.m_rareNonInheritedData->m_content)
        return;
    m_rareNonInheritedData.access().m_content = other.m_rareNonInheritedData->m_content->clone();
}

void RenderStyle::copyPseudoElementsFrom(const RenderStyle& other)
{
    if (!other.m_cachedPseudoStyles)
        return;

    for (auto& pseudoElementStyle : *other.m_cachedPseudoStyles)
        addCachedPseudoStyle(cloneIncludingPseudoElements(*pseudoElementStyle));
}

bool RenderStyle::operator==(const RenderStyle& o) const
{
    // compare everything except the pseudoStyle pointer
    return m_inheritedFlags == o.m_inheritedFlags
        && m_nonInheritedFlags == o.m_nonInheritedFlags
        && m_boxData == o.m_boxData
        && visual == o.visual
        && m_background == o.m_background
        && m_surroundData == o.m_surroundData
        && m_rareNonInheritedData == o.m_rareNonInheritedData
        && m_rareInheritedData == o.m_rareInheritedData
        && m_inheritedData == o.m_inheritedData
#if ENABLE(SVG)
        && m_svgStyle == o.m_svgStyle
#endif
            ;
}

bool RenderStyle::hasUniquePseudoStyle() const
{
    if (!m_cachedPseudoStyles || styleType() != PseudoId::None)
        return false;

    for (size_t i = 0; i < m_cachedPseudoStyles->size(); ++i) {
        RenderStyle* pseudoStyle = m_cachedPseudoStyles->at(i).get();
        if (pseudoStyle->unique())
            return true;
    }

    return false;
}

RenderStyle* RenderStyle::getCachedPseudoStyle(PseudoId pid) const
{
    if (!m_cachedPseudoStyles || !m_cachedPseudoStyles->size())
        return nullptr;

    for (auto& pseudoStyle : *m_cachedPseudoStyles) {
        if (pseudoStyle->styleType() == pid)
            return pseudoStyle.get();
    }

    return nullptr;
}

RenderStyle* RenderStyle::addCachedPseudoStyle(PassOwnPtr<RenderStyle> pseudo)
{
    if (UNLIKELY(!pseudo))
        return nullptr;

    ASSERT(pseudo->styleType() > PseudoId::None);

    RenderStyle* result = pseudo.get();

    if (!m_cachedPseudoStyles)
        m_cachedPseudoStyles = adoptPtr(new PseudoStyleCache);

    m_cachedPseudoStyles->append(WTFMove(pseudo));

    return result;
}

#if ENABLE(IOS_TEXT_AUTOSIZING)

static inline unsigned computeFontHash(const FontCascade& font)
{
    Hasher hasher;
    hasher.add(ASCIICaseInsensitiveHash::hash(font.fontDescription().firstFamily()));
    hasher.add(font.fontDescription().specifiedSize());
    return hasher.hash();
}

unsigned RenderStyle::hashForTextAutosizing() const
{
    // FIXME: Not a very smart hash. Could be improved upon. See <https://bugs.webkit.org/show_bug.cgi?id=121131>.
    unsigned hash = rareNonInheritedData->m_appearance;
    hash ^= rareNonInheritedData->lineClamp.value();
    hash ^= rareInheritedData->overflowWrap;
    hash ^= rareInheritedData->nbspMode;
    hash ^= rareInheritedData->lineBreak;
    hash ^= WTF::FloatHash<float>::hash(inherited->specifiedLineHeight.value());
    hash ^= computeFontHash(inherited->fontCascade);
    hash ^= WTF::FloatHash<float>::hash(inherited->horizontal_border_spacing);
    hash ^= WTF::FloatHash<float>::hash(inherited->vertical_border_spacing);
    hash ^= inherited_flags._box_direction;
    hash ^= inherited_flags.m_rtlOrdering;
    hash ^= noninherited_flags.position();
    hash ^= noninherited_flags.floating();
    hash ^= rareNonInheritedData->textOverflow;
    hash ^= rareInheritedData->textSecurity;
    return hash;
}

bool RenderStyle::equalForTextAutosizing(const RenderStyle* other) const
{
    return rareNonInheritedData->m_appearance == other->rareNonInheritedData->m_appearance
        && rareNonInheritedData->lineClamp == other->rareNonInheritedData->lineClamp
        && rareInheritedData->textSizeAdjust == other->rareInheritedData->textSizeAdjust
        && rareInheritedData->overflowWrap == other->rareInheritedData->overflowWrap
        && rareInheritedData->nbspMode == other->rareInheritedData->nbspMode
        && rareInheritedData->lineBreak == other->rareInheritedData->lineBreak
        && rareInheritedData->textSecurity == other->rareInheritedData->textSecurity
        && inherited->specifiedLineHeight == other->inherited->specifiedLineHeight
        && inherited->fontCascade.equalForTextAutoSizing(other->inherited->fontCascade)
        && inherited->horizontal_border_spacing == other->inherited->horizontal_border_spacing
        && inherited->vertical_border_spacing == other->inherited->vertical_border_spacing
        && inherited_flags._box_direction == other->inherited_flags._box_direction
        && inherited_flags.m_rtlOrdering == other->inherited_flags.m_rtlOrdering
        && noninherited_flags.position() == other->noninherited_flags.position()
        && noninherited_flags.floating() == other->noninherited_flags.floating()
        && rareNonInheritedData->textOverflow == other->rareNonInheritedData->textOverflow;
}

#endif // ENABLE(IOS_TEXT_AUTOSIZING)

bool RenderStyle::inheritedEqual(const RenderStyle& other) const
{
    return m_inheritedFlags == other.m_inheritedFlags
        && m_inheritedData == other.m_inheritedData
        && (m_svgStyle.ptr() == other.m_svgStyle.ptr() || m_svgStyle->inheritedEqual(other.m_svgStyle))
        && m_rareInheritedData == other.m_rareInheritedData;
}

bool RenderStyle::fastPathInheritedEqual(const RenderStyle& other) const
{
    if (m_inheritedData.ptr() == other.m_inheritedData.ptr())
        return true;
    return m_inheritedData->fastPathInheritedEqual(*other.m_inheritedData);
}

bool RenderStyle::nonFastPathInheritedEqual(const RenderStyle& other) const
{
    if (m_inheritedFlags != other.m_inheritedFlags)
        return false;
    if (m_inheritedData.ptr() != other.m_inheritedData.ptr() && !m_inheritedData->nonFastPathInheritedEqual(*other.m_inheritedData))
        return false;
    if (m_rareInheritedData != other.m_rareInheritedData)
        return false;
    if (m_svgStyle.ptr() != other.m_svgStyle.ptr() && !m_svgStyle->inheritedEqual(other.m_svgStyle))
        return false;
    return true;
}

bool RenderStyle::descendantAffectingNonInheritedPropertiesEqual(const RenderStyle& other) const
{
    if (m_rareNonInheritedData.ptr() == other.m_rareNonInheritedData.ptr())
        return true;

    return m_rareNonInheritedData->alignItems == other.m_rareNonInheritedData->alignItems
        && m_rareNonInheritedData->justifyItems == other.m_rareNonInheritedData->justifyItems;
}

#if ENABLE(TEXT_AUTOSIZING)

static inline unsigned computeFontHash(const FontCascade& font)
{
    // FIXME: Would be better to hash the family name rather than hashing a hash of the family name. Also, should this use FontCascadeDescription::familyNameHash?
    return computeHash(ASCIICaseInsensitiveHash::hash(font.fontDescription().firstFamily()), font.fontDescription().specifiedSize());
}

unsigned RenderStyle::hashForTextAutosizing() const
{
    // FIXME: Not a very smart hash. Could be improved upon. See <https://bugs.webkit.org/show_bug.cgi?id=121131>.
    unsigned hash = m_rareNonInheritedData->effectiveAppearance;
    hash ^= m_rareNonInheritedData->lineClamp.value();
    hash ^= m_rareInheritedData->overflowWrap;
    hash ^= m_rareInheritedData->nbspMode;
    hash ^= m_rareInheritedData->lineBreak;
    hash ^= WTF::FloatHash<float>::hash(m_inheritedData->specifiedLineHeight.value());
    hash ^= computeFontHash(m_inheritedData->fontCascade);
    hash ^= WTF::FloatHash<float>::hash(m_inheritedData->horizontalBorderSpacing);
    hash ^= WTF::FloatHash<float>::hash(m_inheritedData->verticalBorderSpacing);
    hash ^= m_inheritedFlags.boxDirection;
    hash ^= m_inheritedFlags.rtlOrdering;
    hash ^= m_nonInheritedFlags.position;
    hash ^= m_nonInheritedFlags.floating;
    hash ^= m_rareNonInheritedData->textOverflow;
    hash ^= m_rareInheritedData->textSecurity;
    return hash;
}

bool RenderStyle::equalForTextAutosizing(const RenderStyle& other) const
{
    return m_rareNonInheritedData->effectiveAppearance == other.m_rareNonInheritedData->effectiveAppearance
        && m_rareNonInheritedData->lineClamp == other.m_rareNonInheritedData->lineClamp
        && m_rareInheritedData->textSizeAdjust == other.m_rareInheritedData->textSizeAdjust
        && m_rareInheritedData->overflowWrap == other.m_rareInheritedData->overflowWrap
        && m_rareInheritedData->nbspMode == other.m_rareInheritedData->nbspMode
        && m_rareInheritedData->lineBreak == other.m_rareInheritedData->lineBreak
        && m_rareInheritedData->textSecurity == other.m_rareInheritedData->textSecurity
        && m_inheritedData->specifiedLineHeight == other.m_inheritedData->specifiedLineHeight
        && m_inheritedData->fontCascade.equalForTextAutoSizing(other.m_inheritedData->fontCascade)
        && m_inheritedData->horizontalBorderSpacing == other.m_inheritedData->horizontalBorderSpacing
        && m_inheritedData->verticalBorderSpacing == other.m_inheritedData->verticalBorderSpacing
        && m_inheritedFlags.boxDirection == other.m_inheritedFlags.boxDirection
        && m_inheritedFlags.rtlOrdering == other.m_inheritedFlags.rtlOrdering
        && m_nonInheritedFlags.position == other.m_nonInheritedFlags.position
        && m_nonInheritedFlags.floating == other.m_nonInheritedFlags.floating
        && m_rareNonInheritedData->textOverflow == other.m_rareNonInheritedData->textOverflow;
}

bool RenderStyle::isIdempotentTextAutosizingCandidate(std::optional<AutosizeStatus> overrideStatus) const
{
    // Refer to <rdar://problem/51826266> for more information regarding how this function was generated.
    auto fields = OptionSet<AutosizeStatus::Fields>::fromRaw(m_inheritedFlags.autosizeStatus);
    if (overrideStatus)
        fields = overrideStatus->fields();

    if (fields.contains(AutosizeStatus::Fields::AvoidSubtree))
        return false;

    const float smallMinimumDifferenceThresholdBetweenLineHeightAndSpecifiedFontSizeForBoostingText = 5;
    const float largeMinimumDifferenceThresholdBetweenLineHeightAndSpecifiedFontSizeForBoostingText = 25;

    if (fields.contains(AutosizeStatus::Fields::FixedHeight)) {
        if (fields.contains(AutosizeStatus::Fields::FixedWidth)) {
            if (whiteSpace() == WhiteSpace::NoWrap) {
                if (width().isFixed())
                    return false;

                if (height().isFixed() && specifiedLineHeight().isFixed()) {
                    float specifiedSize = specifiedFontSize();
                    if (height().value() == specifiedSize && specifiedLineHeight().value() == specifiedSize)
                        return false;
                }

                return true;
            }

            if (fields.contains(AutosizeStatus::Fields::Floating)) {
                if (specifiedLineHeight().isFixed() && height().isFixed()) {
                    float specifiedSize = specifiedFontSize();
                    if (specifiedLineHeight().value() - specifiedSize > smallMinimumDifferenceThresholdBetweenLineHeightAndSpecifiedFontSizeForBoostingText
                        && height().value() - specifiedSize > smallMinimumDifferenceThresholdBetweenLineHeightAndSpecifiedFontSizeForBoostingText) {
                        return true;
                    }
                }
                return false;
            }

            if (fields.contains(AutosizeStatus::Fields::OverflowXHidden))
                return false;

            return true;
        }

        if (fields.contains(AutosizeStatus::Fields::OverflowXHidden)) {
            if (fields.contains(AutosizeStatus::Fields::Floating))
                return false;

            return true;
        }

        return true;
    }

    if (width().isFixed()) {
        if (breakWords())
            return true;

        return false;
    }

    if (textSizeAdjust().isPercentage() && textSizeAdjust().percentage() == 100) {
        if (fields.contains(AutosizeStatus::Fields::Floating))
            return true;

        if (fields.contains(AutosizeStatus::Fields::FixedWidth))
            return true;

        if (specifiedLineHeight().isFixed() && specifiedLineHeight().value() - specifiedFontSize() > largeMinimumDifferenceThresholdBetweenLineHeightAndSpecifiedFontSizeForBoostingText)
            return true;

        return false;
    }

    if (hasBackgroundImage() && backgroundRepeatX() == FillRepeat::NoRepeat && backgroundRepeatY() == FillRepeat::NoRepeat)
        return false;

    return true;
}

AutosizeStatus RenderStyle::autosizeStatus() const
{
    return OptionSet<AutosizeStatus::Fields>::fromRaw(m_inheritedFlags.autosizeStatus);
}

void RenderStyle::setAutosizeStatus(AutosizeStatus autosizeStatus)
{
    m_inheritedFlags.autosizeStatus = autosizeStatus.fields().toRaw();
}

#endif // ENABLE(TEXT_AUTOSIZING)

static bool positionChangeIsMovementOnly(const LengthBox& a, const LengthBox& b, const Length& width)
{
    // If any unit types are different, then we can't guarantee
    // that this was just a movement.
    if (a.left().type() != b.left().type()
        || a.right().type() != b.right().type()
        || a.top().type() != b.top().type()
        || a.bottom().type() != b.bottom().type())
        return false;

    // Only one unit can be non-auto in the horizontal direction and
    // in the vertical direction.  Otherwise the adjustment of values
    // is changing the size of the box.
    if (!a.left().isIntrinsicOrAuto() && !a.right().isIntrinsicOrAuto())
        return false;
    if (!a.top().isIntrinsicOrAuto() && !a.bottom().isIntrinsicOrAuto())
        return false;
    // If our width is auto and left or right is specified then this
    // is not just a movement - we need to resize to our container.
    if ((!a.left().isIntrinsicOrAuto() || !a.right().isIntrinsicOrAuto()) && width.isIntrinsicOrAuto())
        return false;

    // One of the units is fixed or percent in both directions and stayed
    // that way in the new style.  Therefore all we are doing is moving.
    return true;
}

inline bool RenderStyle::changeAffectsVisualOverflow(const RenderStyle& other) const
{
    if (m_rareNonInheritedData.ptr() != other.m_rareNonInheritedData.ptr()
        && !arePointingToEqualData(m_rareNonInheritedData->m_boxShadow, other.m_rareNonInheritedData->m_boxShadow))
        return true;

    if (m_rareInheritedData.ptr() != other.m_rareInheritedData.ptr()
        && !arePointingToEqualData(m_rareInheritedData->textShadow, other.m_rareInheritedData->textShadow))
        return true;

    if (m_inheritedFlags.u._text_decorations != other.m_inheritedFlags.u._text_decorations
        || m_rareNonInheritedData->textDecorationStyle != other.m_rareNonInheritedData->textDecorationStyle
        || m_rareNonInheritedData->textDecorationThickness != other.m_rareNonInheritedData->textDecorationThickness
        || m_rareInheritedData->textUnderlineOffset != other.m_rareInheritedData->textUnderlineOffset
        || m_rareInheritedData->textUnderlinePosition != other.m_rareInheritedData->textUnderlinePosition) {
    	// Underlines are always drawn outside of their textbox bounds when text-underline-position: under;
		// is specified. We can take an early out here.
		if (textUnderlinePosition() == TextUnderlinePosition::Under || other.textUnderlinePosition() == TextUnderlinePosition::Under)
			return true;
		return visualOverflowForDecorations(*this, 0) != visualOverflowForDecorations(other, 0);
	}

    auto hasOutlineInVisualOverflow = this->hasOutlineInVisualOverflow();
    auto otherHasOutlineInVisualOverflow = other.hasOutlineInVisualOverflow();
    if (hasOutlineInVisualOverflow != otherHasOutlineInVisualOverflow
        || (hasOutlineInVisualOverflow && otherHasOutlineInVisualOverflow && outlineSize() != other.outlineSize()))
        return true;

    return false;
}

static bool rareNonInheritedDataChangeRequiresLayout(const StyleRareNonInheritedData& first, const StyleRareNonInheritedData& second, OptionSet<StyleDifferenceContextSensitiveProperty>& changedContextSensitiveProperties)
{
    ASSERT(&first != &second);

    if (first.effectiveAppearance != second.effectiveAppearance
        || first.lineClamp != second.lineClamp
        || first.initialLetter != second.initialLetter
        || first.textOverflow != second.textOverflow)
        return true;

    if (first.shapeMargin != second.shapeMargin)
        return true;

    if (first.deprecatedFlexibleBox != second.deprecatedFlexibleBox)
        return true;

    if (first.flexibleBox != second.flexibleBox)
        return true;

    if (first.columnGap != second.columnGap || first.rowGap != second.rowGap)
        return true;

    if (first.order != second.order
        || first.alignContent != second.alignContent
        || first.alignItems != second.alignItems
        || first.alignSelf != second.alignSelf
        || first.justifyContent != second.justifyContent
        || first.justifyItems != second.justifyItems
        || first.justifySelf != second.justifySelf)
        return true;

    if (!arePointingToEqualData(first.boxReflect, second.boxReflect))
        return true;

    // If the counter directives change, trigger a relayout to re-calculate counter values and rebuild the counter node tree.
    if (!arePointingToEqualData(first.m_counterDirectives, second.m_counterDirectives))
        return true;

    if (first.multiCol != second.multiCol)
        return true;

    if (first.transform.ptr() != second.transform.ptr()) {
        if (first.transform->hasTransform() != second.transform->hasTransform())
            return true;
        if (*first.transform != *second.transform) {
#if USE(ACCELERATED_COMPOSITING)
            changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::Transform);
            // Don't return; keep looking for another change
#else
            return true;
#endif
        }
    }

    if (!arePointingToEqualData(first.scale, second.scale)
        || !arePointingToEqualData(first.rotate, second.rotate)
        || !arePointingToEqualData(first.translate, second.translate)) {
#if USE(ACCELERATED_COMPOSITING)
        changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::Transform);
        // Don't return; keep looking for another change
#else
        return true;
#endif
    }

    if (first.grid != second.grid
        || first.gridItem != second.gridItem)
        return true;

#if !USE(ACCELERATED_COMPOSITING)
	if (first.m_transformStyle3D != second.m_transformStyle3D
		|| first.m_backfaceVisibility != second.m_backfaceVisibility
		|| first.m_perspective != second.m_perspective
		|| first.m_perspectiveOriginX != second.m_perspectiveOriginX
		|| first.m_perspectiveOriginY != second.m_perspectiveOriginY)
		return true;
#endif

#if ENABLE(DASHBOARD_SUPPORT)
    // If regions change, trigger a relayout to re-calc regions.
    if (first.dashboardRegions != second.dashboardRegions)
        return true;
#endif

    if (!arePointingToEqualData(first.willChange, second.willChange)) {
        changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::WillChange);
        // Don't return; keep looking for another change
    }

    if (first.textCombine != second.textCombine)
        return true;

    if (first.breakBefore != second.breakBefore
        || first.breakAfter != second.breakAfter
        || first.breakInside != second.breakInside)
        return true;

    if (first.hasOpacity() != second.hasOpacity()) {
        // FIXME: We would like to use SimplifiedLayout here, but we can't quite do that yet.
        // We need to make sure SimplifiedLayout can operate correctly on RenderInlines (we will need
        // to add a selfNeedsSimplifiedLayout bit in order to not get confused and taint every line).
        // In addition we need to solve the floating object issue when layers come and go. Right now
        // a full layout is necessary to keep floating object lists sane.
        return true;
    }

#if ENABLE(CSS_COMPOSITING)
    if (first.isolation != second.isolation) {
        // Ideally this would trigger a cheaper layout that just updates layer z-order trees (webit.org/b/190088).
        return true;
    }
#endif

    if (first.hasFilters() != second.hasFilters())
        return true;

#if ENABLE(FILTERS_LEVEL_2)
    if (first.hasBackdropFilters() != second.hasBackdropFilters())
        return true;
#endif

    if (first.m_aspectRatioType != second.m_aspectRatioType
        || first.aspectRatioWidth != second.aspectRatioWidth
        || first.aspectRatioHeight != second.aspectRatioHeight) {
        return true;
    }

    // content-visibiliy:hidden turns on contain:size which requires relayout.
    if ((static_cast<ContentVisibility>(first.contentVisibility) == ContentVisibility::Hidden) != (static_cast<ContentVisibility>(second.contentVisibility) == ContentVisibility::Hidden))
        return true;

#if ENABLE(CSS_SCROLL_SNAP)
    if (first.scrollPadding != second.scrollPadding)
        return true;

    if (first.scrollSnapType != second.scrollSnapType)
        return true;
#endif

    return false;
}

static bool rareInheritedDataChangeRequiresLayout(const StyleRareInheritedData& first, const StyleRareInheritedData& second)
{
    ASSERT(&first != &second);

    if (first.indent != second.indent
        || first.textAlignLast != second.textAlignLast
        || first.textJustify != second.textJustify
        || first.textIndentLine != second.textIndentLine
        || first.effectiveZoom != second.effectiveZoom
        || first.textZoom != second.textZoom
#if ENABLE(TEXT_AUTOSIZING)
        || first.textSizeAdjust != second.textSizeAdjust
#endif
        || first.wordBreak != second.wordBreak
        || first.overflowWrap != second.overflowWrap
        || first.nbspMode != second.nbspMode
        || first.lineBreak != second.lineBreak
        || first.textSecurity != second.textSecurity
        || first.hyphens != second.hyphens
        || first.hyphenationLimitBefore != second.hyphenationLimitBefore
        || first.hyphenationLimitAfter != second.hyphenationLimitAfter
        || first.hyphenationString != second.hyphenationString
        || first.rubyPosition != second.rubyPosition
        || first.textEmphasisMark != second.textEmphasisMark
        || first.textEmphasisPosition != second.textEmphasisPosition
        || first.textEmphasisCustomMark != second.textEmphasisCustomMark
        || first.textOrientation != second.textOrientation
        || first.tabSize != second.tabSize
        || first.lineBoxContain != second.lineBoxContain
        || first.lineGrid != second.lineGrid
#if ENABLE(CSS_IMAGE_ORIENTATION)
        || first.imageOrientation != second.imageOrientation
#endif
        || first.lineSnap != second.lineSnap
        || first.lineAlign != second.lineAlign
        || first.hangingPunctuation != second.hangingPunctuation
        || first.effectiveContentVisibility != second.effectiveContentVisibility
#if ENABLE(ACCELERATED_OVERFLOW_SCROLLING)
        || first.useTouchOverflowScrolling != second.useTouchOverflowScrolling
#endif
        || first.listStyleImage != second.listStyleImage) // FIXME: needs arePointingToEqualData()?
        return true;

    if (first.textStrokeWidth != second.textStrokeWidth)
        return true;

    // These properties affect the cached stroke bounding box rects.
    if (first.capStyle != second.capStyle
        || first.joinStyle != second.joinStyle
        || first.strokeWidth != second.strokeWidth
        || first.miterLimit != second.miterLimit)
        return true;

    if (!arePointingToEqualData(first.quotes, second.quotes))
        return true;

    return false;
}

bool RenderStyle::changeRequiresLayout(const RenderStyle& other, OptionSet<StyleDifferenceContextSensitiveProperty>& changedContextSensitiveProperties) const
{
    if (m_svgStyle.ptr() != other.m_svgStyle.ptr() && m_svgStyle->changeRequiresLayout(other.m_svgStyle))
        return true;

    if (m_boxData.ptr() != other.m_boxData.ptr()) {
        if (m_boxData->width() != other.m_boxData->width()
            || m_boxData->minWidth() != other.m_boxData->minWidth()
            || m_boxData->maxWidth() != other.m_boxData->maxWidth()
            || m_boxData->height() != other.m_boxData->height()
            || m_boxData->minHeight() != other.m_boxData->minHeight()
            || m_boxData->maxHeight() != other.m_boxData->maxHeight())
            return true;

        if (m_boxData->verticalAlign() != other.m_boxData->verticalAlign())
            return true;

        if (m_boxData->boxSizing() != other.m_boxData->boxSizing())
            return true;

        if (m_boxData->hasAutoUsedZIndex() != other.m_boxData->hasAutoUsedZIndex())
            return true;
    }

    if (m_surroundData.ptr() != other.m_surroundData.ptr()) {
		if (m_surroundData->margin != other.m_surroundData->margin)
			return true;

		if (m_surroundData->padding != other.m_surroundData->padding)
			return true;

	    // If our border widths change, then we need to layout.  Other changes to borders
	    // only necessitate a repaint.
	    if (borderLeftWidth() != other.borderLeftWidth()
	        || borderTopWidth() != other.borderTopWidth()
	        || borderBottomWidth() != other.borderBottomWidth()
	        || borderRightWidth() != other.borderRightWidth())
	        return true;

	    if (position() != PositionType::Static) {
			if (m_surroundData->offset != other.m_surroundData->offset) {
				// FIXME: We would like to use SimplifiedLayout for relative positioning, but we can't quite do that yet.
				// We need to make sure SimplifiedLayout can operate correctly on RenderInlines (we will need
				// to add a selfNeedsSimplifiedLayout bit in order to not get confused and taint every line).
				if (position() != PositionType::Absolute)
					return true;

				// Optimize for the case where a positioned layer is moving but not changing size.
				if (!positionChangeIsMovementOnly(m_surroundData->offset, other.m_surroundData->offset, m_boxData->width()))
					return true;
			}
		}
    }

    // FIXME: We should add an optimized form of layout that just recomputes visual overflow.
    if (changeAffectsVisualOverflow(other))
        return true;

    if (m_rareNonInheritedData.ptr() != other.m_rareNonInheritedData.ptr()
        && rareNonInheritedDataChangeRequiresLayout(*m_rareNonInheritedData, *other.m_rareNonInheritedData, changedContextSensitiveProperties))
        return true;

    if (m_rareInheritedData.ptr() != other.m_rareInheritedData.ptr()
        && rareInheritedDataChangeRequiresLayout(*m_rareInheritedData, *other.m_rareInheritedData))
        return true;

    if (m_inheritedData.ptr() != other.m_inheritedData.ptr()) {
		if (m_inheritedData->line_height != other.m_inheritedData->line_height
			|| m_inheritedData->font != other.m_inheritedData->font
			|| m_inheritedData->horizontal_border_spacing != other.m_inheritedData->horizontal_border_spacing
			|| m_inheritedData->vertical_border_spacing != other.m_inheritedData->vertical_border_spacing)
			return true;
    }

    if (m_inheritedFlags.u._box_direction != other.m_inheritedFlags.u._box_direction
        || m_inheritedFlags.u.m_rtlOrdering != other.m_inheritedFlags.u.m_rtlOrdering
        || m_nonInheritedFlags.u.position != other.m_nonInheritedFlags.u.position
        || m_nonInheritedFlags.u.floating != other.m_nonInheritedFlags.u.floating
        || m_nonInheritedFlags.u.originalDisplay != other.m_nonInheritedFlags.u.originalDisplay
        || m_nonInheritedFlags.u.verticalAlign != other.m_nonInheritedFlags.u.verticalAlign)
        return true;

    if (static_cast<DisplayType>(m_nonInheritedFlags.u.effectiveDisplay) >= DisplayType::Table) {
        if (m_inheritedFlags.u._border_collapse != other.m_inheritedFlags.u._border_collapse
            || m_inheritedFlags.u._empty_cells != other.m_inheritedFlags.u._empty_cells
            || m_inheritedFlags.u._caption_side != other.m_inheritedFlags.u._caption_side
            || m_nonInheritedFlags.u.tableLayout != other.m_nonInheritedFlags.u.tableLayout)
            return true;

        // In the collapsing border model, 'hidden' suppresses other borders, while 'none'
        // does not, so these style differences can be width differences.
        if (m_inheritedFlags.u._border_collapse
            && ((borderTopStyle() == BHIDDEN && other.borderTopStyle() == BNONE)
                || (borderTopStyle() == BNONE && other.borderTopStyle() == BHIDDEN)
                || (borderBottomStyle() == BHIDDEN && other.borderBottomStyle() == BNONE)
                || (borderBottomStyle() == BNONE && other.borderBottomStyle() == BHIDDEN)
                || (borderLeftStyle() == BHIDDEN && other.borderLeftStyle() == BNONE)
                || (borderLeftStyle() == BNONE && other.borderLeftStyle() == BHIDDEN)
                || (borderRightStyle() == BHIDDEN && other.borderRightStyle() == BNONE)
                || (borderRightStyle() == BNONE && other.borderRightStyle() == BHIDDEN)))
            return true;
    }

    if (static_cast<DisplayType>(m_nonInheritedFlags.u.effectiveDisplay) == DisplayType::ListItem) {
        if (m_inheritedFlags.u._list_style_type != other.m_inheritedFlags.u._list_style_type
            || m_inheritedFlags.u._list_style_position != other.m_inheritedFlags.u._list_style_position)
            return true;
    }

    if (m_inheritedFlags.u._text_align != other.m_inheritedFlags.u._text_align
        || m_inheritedFlags.u._text_transform != other.m_inheritedFlags.u._text_transform
        || m_inheritedFlags.u._direction != other.m_inheritedFlags.u._direction
        || m_inheritedFlags.u._white_space != other.m_inheritedFlags.u._white_space
        || m_nonInheritedFlags.u.clear != other.m_nonInheritedFlags.u.clear
        || m_nonInheritedFlags.u.unicodeBidi != other.m_nonInheritedFlags.u.unicodeBidi)
        return true;

    // Check block flow direction.
    if (m_inheritedFlags.u.m_writingMode != other.m_inheritedFlags.u.m_writingMode)
        return true;

    // Overflow returns a layout hint.
    if (m_nonInheritedFlags.u.overflowX != other.m_nonInheritedFlags.u.overflowX
        || m_nonInheritedFlags.u.overflowY != other.m_nonInheritedFlags.u.overflowY)
        return true;

    if ((visibility() == COLLAPSE) != (other.visibility() == COLLAPSE))
        return true;

    bool hasFirstLineStyle = hasPseudoStyle(PseudoId::FirstLine);
    if (hasFirstLineStyle != other.hasPseudoStyle(PseudoId::FirstLine))
        return true;

    if (hasFirstLineStyle) {
        auto* firstLineStyle = getCachedPseudoStyle(PseudoId::FirstLine);
        if (!firstLineStyle)
            return true;
        auto* otherFirstLineStyle = other.getCachedPseudoStyle(PseudoId::FirstLine);
        if (!otherFirstLineStyle)
            return true;
        // FIXME: Not all first line style changes actually need layout.
        if (*firstLineStyle != *otherFirstLineStyle)
            return true;
    }

	return false;
}

bool RenderStyle::changeRequiresPositionedLayoutOnly(const RenderStyle* other, OptionSet<StyleDifferenceContextSensitiveProperty>&) const
{
    if (position() == PositionType::Static)
        return false;

    if (m_surroundData->offset != other->m_surroundData->offset) {
        // Optimize for the case where a positioned layer is moving but not changing size.
        if (position() == PositionType::Absolute && positionChangeIsMovementOnly(m_surroundData->offset, other->m_surroundData->offset, m_boxData->width()))
            return true;
    }

    return false;
}

static bool rareNonInheritedDataChangeRequiresLayerRepaint(const StyleRareNonInheritedData& first, const StyleRareNonInheritedData& second, OptionSet<StyleDifferenceContextSensitiveProperty>& changedContextSensitiveProperties)
{
#if ENABLE(CSS_COMPOSITING)
    if (first.m_effectiveBlendMode != second.m_effectiveBlendMode)
        return true;
#endif

    if (first.opacity != second.opacity) {
#if USE(ACCELERATED_COMPOSITING)
        changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::Opacity);
        // Don't return; keep looking for another change.
#else
        return true;
#endif
    }

#if ENABLE(CSS_FILTERS)
    if (first.m_filter != second.m_filter) {
#if USE(ACCELERATED_COMPOSITING)
        changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::Filter);
        // Don't return; keep looking for another change.
#else
        return true;
#endif
    }
#endif

#if ENABLE(FILTERS_LEVEL_2)
    if (first.m_backdropFilter != second.m_backdropFilter) {
#if USE(ACCELERATED_COMPOSITING)
        changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::Filter);
        // Don't return; keep looking for another change.
#else
        return true;
#endif
    }
#endif

    // FIXME: In SVG this needs to trigger a layout.
    if (first.mask != second.mask || first.m_maskBoxImage != second.m_maskBoxImage)
        return true;

    return false;
}

bool RenderStyle::changeRequiresLayerRepaint(const RenderStyle& other, OptionSet<StyleDifferenceContextSensitiveProperty>& changedContextSensitiveProperties) const
{
    // StyleResolver has ensured that zIndex is non-auto only if it's applicable.
    if (m_boxData.ptr() != other.m_boxData.ptr()) {
		if (m_boxData->usedZIndex() != other.m_boxData->usedZIndex() || m_boxData->hasAutoUsedZIndex() != other.m_boxData->hasAutoUsedZIndex())
			return true;
    }

    if (position() != PositionType::Static) {
        if (visual.ptr() != other.visual.ptr()) {
			if (visual->clip != other.visual->clip || visual->hasClip != other.visual->hasClip) {
				changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::ClipRect);
				return true;
			}
        }
    }

    if (m_rareNonInheritedData.ptr() != other.m_rareNonInheritedData.ptr()
        && rareNonInheritedDataChangeRequiresLayerRepaint(*m_rareNonInheritedData, *other.m_rareNonInheritedData, changedContextSensitiveProperties))
        return true;

    return false;
}

static bool requiresPainting(const RenderStyle& style)
{
    if (style.visibility() == HIDDEN)
        return false;
    if (!style.opacity())
        return false;
    return true;
}

static bool rareNonInheritedDataChangeRequiresRepaint(const StyleRareNonInheritedData& first, const StyleRareNonInheritedData& second, OptionSet<StyleDifferenceContextSensitiveProperty>& changedContextSensitiveProperties)
{
    if (first.userDrag != second.userDrag
        || first.m_objectFit != second.m_objectFit
        || first.m_objectPosition != second.m_objectPosition)
        return true;

    if (first.m_shapeOutside != second.m_shapeOutside)
        return true;

    // FIXME: this should probably be moved to changeRequiresLayerRepaint().
    if (first.m_clipPath != second.m_clipPath) {
#if USE(ACCELERATED_COMPOSITING)
        changedContextSensitiveProperties.add(StyleDifferenceContextSensitiveProperty::ClipPath);
        // Don't return; keep looking for another change.
#else
        return true;
#endif
    }

    return false;
}

static bool rareInheritedDataChangeRequiresRepaint(const StyleRareInheritedData& first, const StyleRareInheritedData& second)
{
    return first.userModify != second.userModify
        || first.userSelect != second.userSelect
        //|| first.appleColorFilter != second.appleColorFilter
        || first.imageRendering != second.imageRendering;
}

bool RenderStyle::changeRequiresRepaint(const RenderStyle& other, OptionSet<StyleDifferenceContextSensitiveProperty>& changedContextSensitiveProperties) const
{
    bool currentColorDiffers = m_inheritedData->color != other.m_inheritedData->color;

    if (currentColorDiffers || m_svgStyle.ptr() != other.m_svgStyle.ptr()) {
    	if (m_svgStyle->changeRequiresRepaint(other.m_svgStyle, currentColorDiffers))
    		return true;
    }

    if (!requiresPainting(*this) && !requiresPainting(other))
        return false;

    if (m_inheritedFlags.u._visibility != other.m_inheritedFlags.u._visibility
        || m_inheritedFlags.u.m_printColorAdjust != other.m_inheritedFlags.u.m_printColorAdjust
        || m_inheritedFlags.u._insideLink != other.m_inheritedFlags.u._insideLink)
    	return true;

    if (currentColorDiffers || m_background.ptr() != other.m_background.ptr()) {
        if (!m_background->isEquivalentForPainting(*other.m_background, currentColorDiffers))
            return true;
    }

    if (currentColorDiffers || m_surroundData.ptr() != other.m_surroundData.ptr()) {
        if (!m_surroundData->border.isEquivalentForPainting(other.m_surroundData->border, currentColorDiffers))
            return true;
    }

    if (m_rareNonInheritedData.ptr() != other.m_rareNonInheritedData.ptr()
        && rareNonInheritedDataChangeRequiresRepaint(*m_rareNonInheritedData, *other.m_rareNonInheritedData, changedContextSensitiveProperties))
        return true;

    if (m_rareInheritedData.ptr() != other.m_rareInheritedData.ptr()
        && rareInheritedDataChangeRequiresRepaint(*m_rareInheritedData, *other.m_rareInheritedData))
        return true;

    return false;
}

bool RenderStyle::changeRequiresRepaintIfText(const RenderStyle& other, OptionSet<StyleDifferenceContextSensitiveProperty>&) const
{
    // FIXME: Does this code need to consider currentColorDiffers? webkit.org/b/266833
    if (m_inheritedData->color != other.m_inheritedData->color)
        return true;

    if (m_inheritedFlags.u._text_decorations != other.m_inheritedFlags.u._text_decorations
        || visual->textDecoration != other.visual->textDecoration)
        return true;

    if (m_rareNonInheritedData.ptr() != other.m_rareNonInheritedData.ptr()) {
        if (m_rareNonInheritedData->textDecorationStyle != other.m_rareNonInheritedData->textDecorationStyle
			|| m_rareNonInheritedData->m_textDecorationColor != other.m_rareNonInheritedData->m_textDecorationColor
			|| m_rareNonInheritedData->textDecorationThickness != other.m_rareNonInheritedData->textDecorationThickness)
        	return true;
    }

    if (m_rareInheritedData.ptr() != other.m_rareInheritedData.ptr()) {
        if (m_rareInheritedData->textDecorationSkip != other.m_rareInheritedData->textDecorationSkip
            || m_rareInheritedData->textFillColor != other.m_rareInheritedData->textFillColor
            || m_rareInheritedData->textStrokeColor != other.m_rareInheritedData->textStrokeColor
            || m_rareInheritedData->textEmphasisColor != other.m_rareInheritedData->textEmphasisColor
            || m_rareInheritedData->textEmphasisFill != other.m_rareInheritedData->textEmphasisFill
            || m_rareInheritedData->strokeColor != other.m_rareInheritedData->strokeColor
            || m_rareInheritedData->caretColor != other.m_rareInheritedData->caretColor)
            return true;
    }

    return false;
}

bool RenderStyle::changeRequiresRecompositeLayer(const RenderStyle* other, OptionSet<StyleDifferenceContextSensitiveProperty>&) const
{
#if USE(ACCELERATED_COMPOSITING)
    if (m_inheritedFlags.u._pointerEvents != other->m_inheritedFlags.u._pointerEvents)
        return true;

    if (m_rareNonInheritedData.ptr() != other->m_rareNonInheritedData.ptr()) {
        if (usedTransformStyle3D() != other->usedTransformStyle3D()
            || m_rareNonInheritedData->m_backfaceVisibility != other->m_rareNonInheritedData->m_backfaceVisibility
            || m_rareNonInheritedData->m_perspective != other->m_rareNonInheritedData->m_perspective
            || m_rareNonInheritedData->m_perspectiveOriginX != other->m_rareNonInheritedData->m_perspectiveOriginX
            || m_rareNonInheritedData->m_perspectiveOriginY != other->m_rareNonInheritedData->m_perspectiveOriginY)
            return true;
    }
#endif
    return false;
}

StyleDifference RenderStyle::diff(const RenderStyle* other, OptionSet<StyleDifferenceContextSensitiveProperty>& changedContextSensitiveProperties) const
{
    changedContextSensitiveProperties = OptionSet<StyleDifferenceContextSensitiveProperty>();

    if (changeRequiresLayout(*other, changedContextSensitiveProperties))
        return StyleDifferenceLayout;

    // FIXME: we need to also check for repaint changes after a PositionedMovementOnly change
    // on a composited layer: https://bugs.webkit.org/show_bug.cgi?id=116319
    if (changeRequiresPositionedLayoutOnly(other, changedContextSensitiveProperties))
        return StyleDifferenceLayoutPositionedMovementOnly;

    if (changeRequiresLayerRepaint(*other, changedContextSensitiveProperties))
        return StyleDifferenceRepaintLayer;

    if (changeRequiresRepaint(*other, changedContextSensitiveProperties))
        return StyleDifferenceRepaint;

    if (changeRequiresRepaintIfText(*other, changedContextSensitiveProperties))
        return StyleDifferenceRepaintIfText;

    // FIXME: RecompositeLayer should also behave as a priority bit (e.g when the style change requires layout, we know that
    // the content also needs repaint and it will eventually get repainted,
    // but a repaint type of change (e.g. color change) does not necessarily trigger recomposition).
#if USE(ACCELERATED_COMPOSITING)
    if (changeRequiresRecompositeLayer(other, changedContextSensitiveProperties))
        return StyleDifferenceRecompositeLayer;
#endif

    // Cursors are not checked, since they will be set appropriately in response to mouse events,
    // so they don't need to cause any repaint or layout.

    // Animations don't need to be checked either.  We always set the new style on the RenderObject, so we will get a chance to fire off
    // the resulting transition properly.
    return StyleDifferenceEqual;
}

bool RenderStyle::diffRequiresLayerRepaint(const RenderStyle& style, bool isComposited) const
{
	OptionSet<StyleDifferenceContextSensitiveProperty> changedContextSensitiveProperties;

    if (changeRequiresRepaint(style, changedContextSensitiveProperties))
        return true;

    if (isComposited && changeRequiresLayerRepaint(style, changedContextSensitiveProperties))
    	return changedContextSensitiveProperties.contains(StyleDifferenceContextSensitiveProperty::ClipRect);

    return false;
}

void RenderStyle::setClip(const Length& top, const Length& right, const Length& bottom, const Length& left)
{
    auto& data = visual.access();
    data.clip.top() = top;
    data.clip.right() = right;
    data.clip.bottom() = bottom;
    data.clip.left() = left;
}

void RenderStyle::addCursor(PassRefPtr<StyleImage> image, const Optional<IntPoint>& hotSpot)
{
    auto& cursorData = m_rareInheritedData.access().cursorData;
    if (!cursorData)
        cursorData = CursorList::create();
    // Point outside the image is how we tell the cursor machinery there is no hot spot, and it should generate one (done in the Cursor class).
    // FIXME: Would it be better to extend the concept of "no hot spot" deeper, into CursorData and beyond, rather than using -1/-1 for it?
    cursorData->append(CursorData(image, hotSpot.value_or(IntPoint(-1, -1))));
}

void RenderStyle::setCursorList(PassRefPtr<CursorList> other)
{
    m_rareInheritedData.access().cursorData = other;
}

void RenderStyle::setQuotes(PassRefPtr<QuotesData> q)
{
    if (arePointingToEqualData(m_rareInheritedData->quotes.get(), q.get()))
        return;

    m_rareInheritedData.access().quotes = q;
}

void RenderStyle::setWillChange(PassRefPtr<WillChangeData> willChangeData)
{
    if (arePointingToEqualData(m_rareNonInheritedData->willChange.get(), willChangeData.get()))
        return;

    m_rareNonInheritedData.access().willChange = willChangeData;
}

void RenderStyle::setScale(PassRefPtr<ScaleTransformOperation> t)
{
    m_rareNonInheritedData.access().scale = t;
}

void RenderStyle::setRotate(PassRefPtr<RotateTransformOperation> t)
{
    m_rareNonInheritedData.access().rotate = t;
}

void RenderStyle::setTranslate(PassRefPtr<TranslateTransformOperation> t)
{
    m_rareNonInheritedData.access().translate = t;
}

void RenderStyle::clearCursorList()
{
    if (m_rareInheritedData->cursorData)
        m_rareInheritedData.access().cursorData = nullptr;
}

void RenderStyle::clearContent()
{
    if (m_rareNonInheritedData->m_content)
        m_rareNonInheritedData.access().m_content = nullptr;
}

void RenderStyle::appendContent(PassOwnPtr<ContentData> contentData)
{
    OwnPtr<ContentData>& content = m_rareNonInheritedData.access().m_content;
    ContentData* lastContent = content.get();
    while (lastContent && lastContent->next())
        lastContent = lastContent->next();

    if (lastContent)
        lastContent->setNext(contentData);
    else
        content = contentData;
}

void RenderStyle::setContent(PassRefPtr<StyleImage> image, bool add)
{
    if (!image)
        return;
        
    if (add) {
        appendContent(ContentData::create(image));
        return;
    }

    m_rareNonInheritedData.access().m_content = ContentData::create(image);
    if (!m_rareNonInheritedData.access().m_altText.isNull())
		m_rareNonInheritedData.access().m_content->setAltText(m_rareNonInheritedData.access().m_altText);
}

void RenderStyle::setContent(const String& string, bool add)
{
    OwnPtr<ContentData>& content = m_rareNonInheritedData.access().m_content;
    if (add) {
        ContentData* lastContent = content.get();
        while (lastContent && lastContent->next())
            lastContent = lastContent->next();

        if (lastContent) {
            // We attempt to merge with the last ContentData if possible.
            if (lastContent->isText()) {
                TextContentData* textContent = static_cast<TextContentData*>(lastContent);
                textContent->setText(textContent->text() + string);
            } else
                lastContent->setNext(ContentData::create(string));

            if (!m_rareNonInheritedData.access().m_altText.isNull())
				lastContent->setAltText(m_rareNonInheritedData.access().m_altText);
			return;
        }
    }

    content = ContentData::create(string);
    if (!m_rareNonInheritedData.access().m_altText.isNull())
		content->setAltText(m_rareNonInheritedData.access().m_altText);
}

void RenderStyle::setContent(PassOwnPtr<CounterContent> counter, bool add)
{
    if (!counter)
        return;

    if (add) {
        appendContent(ContentData::create(counter));
        return;
    }

    m_rareNonInheritedData.access().m_content = ContentData::create(counter);
}

void RenderStyle::setContent(QuoteType quote, bool add)
{
    if (add) {
        appendContent(ContentData::create(quote));
        return;
    }

    m_rareNonInheritedData.access().m_content = ContentData::create(quote);
}

void RenderStyle::setContentAltText(const String& string)
{
    m_rareNonInheritedData.access().m_altText = string;
    
    if (m_rareNonInheritedData.access().m_content)
        m_rareNonInheritedData.access().m_content->setAltText(string);
}

void RenderStyle::setHasAttrContent()
{
    setUnique();
    SET_VAR(m_rareNonInheritedData, m_hasAttrContent, true);
}

void RenderStyle::setNavProperty(ENavProperty index, const String& value)
{
	if(!m_rareNonInheritedData->m_navPropertyData)
		m_rareNonInheritedData.access().m_navPropertyData = adoptPtr(new StyleNavPropertyData());

	m_rareNonInheritedData->m_navPropertyData->values[static_cast<int>(index)] = value;
}

bool RenderStyle::affectedByTransformOrigin() const
{
    if (m_rareNonInheritedData->rotate && !m_rareNonInheritedData->rotate->isIdentity())
        return true;

    if (m_rareNonInheritedData->scale && !m_rareNonInheritedData->scale->isIdentity())
        return true;

    auto& transformOperations = m_rareNonInheritedData->transform->m_operations;
    if (transformOperations.affectedByTransformOrigin())
        return true;

#if 0
    if (offsetPath())
        return true;
#endif

    return false;
}

FloatPoint RenderStyle::computePerspectiveOrigin(const FloatRect& boundingBox) const
{
    return boundingBox.location() + floatPointForLengthPoint(perspectiveOrigin(), boundingBox.size());
}

void RenderStyle::applyPerspective(TransformationMatrix& transform, const RenderObject& renderer, const FloatPoint& originTranslate) const
{
    // https://www.w3.org/TR/css-transforms-2/#perspective
    // The perspective matrix is computed as follows:
    // 1. Start with the identity matrix.

    // 2. Translate by the computed X and Y values of perspective-origin
    transform.translate(originTranslate.x(), originTranslate.y());

    // 3. Multiply by the matrix that would be obtained from the perspective() transform function, where the length is provided by the value of the perspective property
    transform.applyPerspective(usedPerspective(renderer));

    // 4. Translate by the negated computed X and Y values of perspective-origin
    transform.translate(-originTranslate.x(), -originTranslate.y());
}

FloatPoint3D RenderStyle::computeTransformOrigin(const FloatRect& boundingBox) const
{
    FloatPoint3D originTranslate;
    originTranslate.setXY(boundingBox.location() + floatPointForLengthPoint(transformOriginXY(), boundingBox.size()));
    originTranslate.setZ(transformOriginZ());
    return originTranslate;
}

void RenderStyle::applyTransformOrigin(TransformationMatrix& transform, const FloatPoint3D& originTranslate) const
{
    if (!originTranslate.isZero())
        transform.translate3d(originTranslate.x(), originTranslate.y(), originTranslate.z());
}

void RenderStyle::unapplyTransformOrigin(TransformationMatrix& transform, const FloatPoint3D& originTranslate) const
{
    if (!originTranslate.isZero())
        transform.translate3d(-originTranslate.x(), -originTranslate.y(), -originTranslate.z());
}

void RenderStyle::applyTransform(TransformationMatrix& transform, const FloatRect& boundingBox, OptionSet<RenderStyle::TransformOperationOption> options) const
{
	if (!options.contains(RenderStyle::TransformOperationOption::TransformOrigin) || !affectedByTransformOrigin()) {
		applyCSSTransform(transform, boundingBox, options);
		return;
	}

    auto originTranslate = computeTransformOrigin(boundingBox);
    applyTransformOrigin(transform, originTranslate);
	applyCSSTransform(transform, boundingBox, options);
	unapplyTransformOrigin(transform, originTranslate);
}

void RenderStyle::applyCSSTransform(TransformationMatrix& transform, const FloatRect& boundingBox, OptionSet<RenderStyle::TransformOperationOption> options) const
{
	// https://www.w3.org/TR/css-transforms-2/#ctm
	// The transformation matrix is computed from the transform, transform-origin, translate, rotate, scale, and offset properties as follows:
	// 1. Start with the identity matrix.

	// 2. Translate by the computed X, Y, and Z values of transform-origin.
	// (implemented in applyTransformOrigin)

    // 3. Translate by the computed X, Y, and Z values of translate.
    if (options.contains(RenderStyle::TransformOperationOption::Translate)) {
    	if (auto* translate = m_rareNonInheritedData->translate.get())
    		translate->apply(transform, boundingBox.size());
    }

    // 4. Rotate by the computed <angle> about the specified axis of rotate.
    if (options.contains(RenderStyle::TransformOperationOption::Rotate)) {
    	if (auto* rotate = m_rareNonInheritedData->rotate.get())
    		rotate->apply(transform, boundingBox.size());
    }

    // 5. Scale by the computed X, Y, and Z values of scale.
    if (options.contains(RenderStyle::TransformOperationOption::Scale)) {
    	if (auto* scale = m_rareNonInheritedData->scale.get())
    		scale->apply(transform, boundingBox.size());
    }

	// 6. Translate and rotate by the transform specified by offset. (FIXME: we don't support the offset property)

    // 7. Multiply by each of the transform functions in transform from left to right.
    auto& transformOperations = m_rareNonInheritedData->transform->m_operations;
    for (auto& operation : transformOperations.operations())
		operation->apply(transform, boundingBox.size());

    // 8. Translate by the negated computed X, Y and Z values of transform-origin.
    // (implemented in unapplyTransformOrigin)
}

#if 0
static PathTraversalState getTraversalStateAtDistance(const Path& path, const Length& distance)
{
    auto pathLength = path.length();
    auto distanceValue = floatValueForLength(distance, pathLength);

    float resolvedLength = 0;
    if (path.isClosed()) {
        if (pathLength) {
            resolvedLength = fmod(distanceValue, pathLength);
            if (resolvedLength < 0)
                resolvedLength += pathLength;
        }
    } else
        resolvedLength = clampTo<float>(distanceValue, 0, pathLength);

    ASSERT(resolvedLength >= 0);
    return path.traversalStateAtLength(resolvedLength);
}

void RenderStyle::applyMotionPathTransform(TransformationMatrix& transform, const FloatRect& boundingBox) const
{
    if (!offsetPath())
        return;

    auto transformOrigin = computeTransformOrigin(boundingBox).xy();
    auto anchor = transformOrigin;
    if (!offsetAnchor().x().isAuto())
        anchor = floatPointForLengthPoint(offsetAnchor(), boundingBox.size()) + boundingBox.location();

    // Shift element to the point on path specified by offset-path and offset-distance.
    auto path = offsetPath()->getPath(boundingBox, anchor, offsetRotate());
	if (!path)
		return;
    auto traversalState = getTraversalStateAtDistance(*path, offsetDistance());
    transform.translate(traversalState.current().x(), traversalState.current().y());

    // Shift element to the anchor specified by offset-anchor.
    transform.translate(-anchor.x(), -anchor.y());

    auto shiftToOrigin = anchor - transformOrigin;
    transform.translate(shiftToOrigin.width(), shiftToOrigin.height());

    // Apply rotation.
    auto rotation = offsetRotate();
    if (rotation.hasAuto())
        transform.rotate(traversalState.normalAngle() + rotation.angle());
    else
        transform.rotate(rotation.angle());

    transform.translate(-shiftToOrigin.width(), -shiftToOrigin.height());
}
#endif

void RenderStyle::setPageScaleTransform(float scale)
{
    if (scale == 1)
        return;
    TransformOperations transform;
    transform.operations().append(ScaleTransformOperation::create(scale, scale, ScaleTransformOperation::SCALE));
    setTransform(transform);
    setTransformOriginX(Length(0, LengthType::Fixed));
    setTransformOriginY(Length(0, LengthType::Fixed));
}

void RenderStyle::setTextShadow(PassOwnPtr<ShadowData> shadowData, bool add)
{
    ASSERT(!shadowData || (shadowData->spread().isZero() && shadowData->style() == ShadowStyle::Normal));

    auto& rareData = m_rareInheritedData.access();
    if (!add) {
        rareData.textShadow = shadowData;
        return;
    }

    shadowData->setNext(rareData.textShadow.release());
    rareData.textShadow = shadowData;
}

void RenderStyle::setBoxShadow(PassOwnPtr<ShadowData> shadowData, bool add)
{
    auto& rareData = m_rareNonInheritedData.access();
    if (!add) {
        rareData.m_boxShadow = shadowData;
        return;
    }

    shadowData->setNext(rareData.m_boxShadow.release());
    rareData.m_boxShadow = shadowData;
}

static RoundedRect::Radii calcRadiiFor(const BorderData& border, const LayoutSize& size)
{
    return RoundedRect::Radii(
		LayoutSize(valueForLength(border.topLeft().width(), size.width()),
			valueForLength(border.topLeft().height(), size.height())),
		LayoutSize(valueForLength(border.topRight().width(), size.width()),
			valueForLength(border.topRight().height(), size.height())),
		LayoutSize(valueForLength(border.bottomLeft().width(), size.width()),
			valueForLength(border.bottomLeft().height(), size.height())),
		LayoutSize(valueForLength(border.bottomRight().width(), size.width()),
			valueForLength(border.bottomRight().height(), size.height())));
}

RoundedRect RenderStyle::getRoundedBorderFor(const LayoutRect& borderRect, bool includeLogicalLeftEdge, bool includeLogicalRightEdge) const
{
    RoundedRect roundedRect(borderRect);
    if (hasBorderRadius()) {
        RoundedRect::Radii radii = calcRadiiFor(m_surroundData->border, borderRect.size());
        radii.scale(calcBorderRadiiConstraintScaleFor(borderRect, radii));
        roundedRect.includeLogicalEdges(radii, isHorizontalWritingMode(), includeLogicalLeftEdge, includeLogicalRightEdge);
    }
    return roundedRect;
}

RoundedRect RenderStyle::getRoundedInnerBorderFor(const LayoutRect& borderRect, bool includeLogicalLeftEdge, bool includeLogicalRightEdge) const
{
    bool horizontal = isHorizontalWritingMode();

    LayoutUnit leftWidth = (!horizontal || includeLogicalLeftEdge) ? borderLeftWidth() : 0;
    LayoutUnit rightWidth = (!horizontal || includeLogicalRightEdge) ? borderRightWidth() : 0;
    LayoutUnit topWidth = (horizontal || includeLogicalLeftEdge) ? borderTopWidth() : 0;
    LayoutUnit bottomWidth = (horizontal || includeLogicalRightEdge) ? borderBottomWidth() : 0;

    return getRoundedInnerBorderFor(borderRect, topWidth, bottomWidth, leftWidth, rightWidth, includeLogicalLeftEdge, includeLogicalRightEdge);
}

RoundedRect RenderStyle::getRoundedInnerBorderFor(const LayoutRect& borderRect, LayoutUnit topWidth, LayoutUnit bottomWidth,
	LayoutUnit leftWidth, LayoutUnit rightWidth, bool includeLogicalLeftEdge, bool includeLogicalRightEdge) const
{
    auto width = std::max(LayoutUnit(0), borderRect.width() - leftWidth - rightWidth);
    auto height = std::max(LayoutUnit(0), borderRect.height() - topWidth - bottomWidth);
    auto roundedRect = RoundedRect {
        borderRect.x() + leftWidth,
        borderRect.y() + topWidth,
        width,
        height
    };
    if (hasBorderRadius()) {
        RoundedRect::Radii radii = getRoundedBorderFor(borderRect).radii();
        radii.shrink(topWidth, bottomWidth, leftWidth, rightWidth);
        roundedRect.includeLogicalEdges(radii, isHorizontalWritingMode(), includeLogicalLeftEdge, includeLogicalRightEdge);
    }
    if (!roundedRect.isRenderable())
		roundedRect.adjustRadii();
    return roundedRect;
}

static bool allLayersAreFixed(const FillLayer& layers)
{
    for (auto* layer = &layers; layer; layer = layer->next()) {
        if (!(layer->image() && layer->attachment() == FixedBackgroundAttachment))
            return false;
    }
    return true;
}

bool RenderStyle::hasEntirelyFixedBackground() const
{
    return allLayersAreFixed(backgroundLayers());
}

const CounterDirectiveMap* RenderStyle::counterDirectives() const
{
    return m_rareNonInheritedData->m_counterDirectives.get();
}

CounterDirectiveMap& RenderStyle::accessCounterDirectives()
{
    OwnPtr<CounterDirectiveMap>& map = m_rareNonInheritedData.access().m_counterDirectives;
    if (!map)
        map = adoptPtr(new CounterDirectiveMap);
    return *map;
}

const AtomicString& RenderStyle::hyphenString() const
{
    ASSERT(hyphens() != HyphensNone);

    const AtomicString& hyphenationString = m_rareInheritedData->hyphenationString;
    if (!hyphenationString.isNull())
        return hyphenationString;

    // FIXME: This should depend on locale.
    DEFINE_STATIC_LOCAL_STRING(AtomicString, hyphenMinusString, (&hyphenMinus, 1));
    DEFINE_STATIC_LOCAL_STRING(AtomicString, hyphenString, (&hyphen, 1));
    return font().primaryFontHasGlyphForCharacter(hyphen) ? hyphenString : hyphenMinusString;
}

const AtomicString& RenderStyle::textEmphasisMarkString() const
{
    switch (textEmphasisMark()) {
    case TextEmphasisMarkNone:
        return nullAtom;
    case TextEmphasisMarkCustom:
        return textEmphasisCustomMark();
    case TextEmphasisMarkDot: {
        DEFINE_STATIC_LOCAL_STRING(AtomicString, filledDotString, (&bullet, 1));
        DEFINE_STATIC_LOCAL_STRING(AtomicString, openDotString, (&whiteBullet, 1));
        return textEmphasisFill() == TextEmphasisFillFilled ? filledDotString : openDotString;
    }
    case TextEmphasisMarkCircle: {
        DEFINE_STATIC_LOCAL_STRING(AtomicString, filledCircleString, (&blackCircle, 1));
        DEFINE_STATIC_LOCAL_STRING(AtomicString, openCircleString, (&whiteCircle, 1));
        return textEmphasisFill() == TextEmphasisFillFilled ? filledCircleString : openCircleString;
    }
    case TextEmphasisMarkDoubleCircle: {
        DEFINE_STATIC_LOCAL_STRING(AtomicString, filledDoubleCircleString, (&fisheye, 1));
        DEFINE_STATIC_LOCAL_STRING(AtomicString, openDoubleCircleString, (&bullseye, 1));
        return textEmphasisFill() == TextEmphasisFillFilled ? filledDoubleCircleString : openDoubleCircleString;
    }
    case TextEmphasisMarkTriangle: {
        DEFINE_STATIC_LOCAL_STRING(AtomicString, filledTriangleString, (&blackUpPointingTriangle, 1));
        DEFINE_STATIC_LOCAL_STRING(AtomicString, openTriangleString, (&whiteUpPointingTriangle, 1));
        return textEmphasisFill() == TextEmphasisFillFilled ? filledTriangleString : openTriangleString;
    }
    case TextEmphasisMarkSesame: {
        DEFINE_STATIC_LOCAL_STRING(AtomicString, filledSesameString, (&sesameDot, 1));
        DEFINE_STATIC_LOCAL_STRING(AtomicString, openSesameString, (&whiteSesameDot, 1));
        return textEmphasisFill() == TextEmphasisFillFilled ? filledSesameString : openSesameString;
    }
    case TextEmphasisMarkAuto:
        ASSERT_NOT_REACHED();
        return nullAtom;
    }

    ASSERT_NOT_REACHED();
    return nullAtom;
}

#if ENABLE(DASHBOARD_SUPPORT)
const Vector<StyleDashboardRegion>& RenderStyle::initialDashboardRegions()
{
	static NeverDestroyed<Vector<StyleDashboardRegion>> emptyList;
    return emptyList;
}

const Vector<StyleDashboardRegion>& RenderStyle::noneDashboardRegions()
{
	static NeverDestroyed<Vector<StyleDashboardRegion>> noneList;
    static bool noneListInitialized = false;

    if (!noneListInitialized) {
        StyleDashboardRegion region;
        region.label = "";
        region.offset.top()  = Length();
        region.offset.right() = Length();
        region.offset.bottom() = Length();
        region.offset.left() = Length();
        region.type = StyleDashboardRegion::None;
        noneList.append(region);
        noneListInitialized = true;
    }
    return noneList;
}
#endif

void RenderStyle::adjustAnimations()
{
    AnimationList* animationList = m_rareNonInheritedData->m_animations.get();
    if (!animationList)
        return;

    // Get rid of empty animations and anything beyond them
    for (size_t i = 0; i < animationList->size(); ++i) {
        if (animationList->animation(i).isEmpty()) {
            animationList->resize(i);
            break;
        }
    }

    if (animationList->isEmpty()) {
        clearAnimations();
        return;
    }

    // Repeat patterns into layers that don't have some properties set.
    animationList->fillUnsetProperties();
}

void RenderStyle::adjustTransitions()
{
    AnimationList* transitionList = m_rareNonInheritedData->m_transitions.get();
    if (!transitionList)
        return;

    // Get rid of empty transitions and anything beyond them
    for (size_t i = 0; i < transitionList->size(); ++i) {
        if (transitionList->animation(i).isEmpty()) {
            transitionList->resize(i);
            break;
        }
    }

    if (transitionList->isEmpty()) {
        clearTransitions();
        return;
    }

    // Repeat patterns into layers that don't have some properties set.
    transitionList->fillUnsetProperties();
}

AnimationList& RenderStyle::ensureAnimations()
{
    if (!m_rareNonInheritedData.access().m_animations)
        m_rareNonInheritedData.access().m_animations = AnimationList::create();
    return *m_rareNonInheritedData->m_animations;
}

AnimationList& RenderStyle::ensureTransitions()
{
    if (!m_rareNonInheritedData.access().m_transitions)
        m_rareNonInheritedData.access().m_transitions = AnimationList::create();
    return *m_rareNonInheritedData->m_transitions;
}

float RenderStyle::usedPerspective(const RenderObject& object) const
{
    return object.document().settings().css3DTransformInteroperabilityEnabled() ? std::max(1.0f, perspective()) : perspective();
}

const Animation* RenderStyle::transitionForProperty(CSSPropertyID property) const
{
    auto* transitions = this->transitions();
    if (!transitions)
        return nullptr;
	for (size_t i = 0; i < transitions->size(); ++i) {
		auto& animation = transitions->animation(i);
		if (animation.property().mode == Animation::TransitionMode::All || animation.property().id == property)
			return &animation;
	}
    return nullptr;
}

LayoutUnit RenderStyle::computedLineHeight() const
{
    return computeLineHeight(lineHeight());
}

LayoutUnit RenderStyle::computeLineHeight(const Length& lineHeightLength) const
{
    // Negative value means the line height is not set.  Use the font's built-in spacing.
    if (lineHeightLength.isNegative())
        return fontMetrics().lineSpacing();

    if (lineHeightLength.isPercentOrCalculated())
        return minimumValueForLength(lineHeightLength, computedFontSize());

    return clampTo<int>(lineHeightLength.value());
}

void RenderStyle::setWordSpacing(const Length& value)
{
    float fontWordSpacing;
    switch (value.type()) {
    case LengthType::Auto:
        fontWordSpacing = 0;
        break;
    case LengthType::Percent:
        fontWordSpacing = value.percent() * font().widthOfSpaceString() / 100;
        break;
    case LengthType::Fixed:
        fontWordSpacing = value.value();
        break;
    case LengthType::Calculated:
        fontWordSpacing = value.nonNanCalculatedValue(static_cast<float>(maxValueForCssLength));
        break;
    default:
        ASSERT_NOT_REACHED();
        fontWordSpacing = 0;
        break;
    }
    m_inheritedData.access().font.setWordSpacing(fontWordSpacing);
    m_rareInheritedData.access().wordSpacing = value;
}

void RenderStyle::setFontSize(float size)
{
    // size must be specifiedSize if Text Autosizing is enabled, but computedSize if text
    // zoom is enabled (if neither is enabled it's irrelevant as they're probably the same).

    ASSERT(isfinite(size));
    if (!isfinite(size) || size < 0)
        size = 0;
    else
        size = std::min(maximumAllowedFontSize, size);

    FontSelector* currentFontSelector = font().fontSelector();
    FontDescription desc(fontDescription());
    desc.setSpecifiedSize(size);
    desc.setComputedSize(size);

#if ENABLE(TEXT_AUTOSIZING)
    float multiplier = textAutosizingMultiplier();
    if (multiplier > 1) {
        float autosizedFontSize = TextAutosizer::computeAutosizedFontSize(size, multiplier);
        desc.setComputedSize(min(maximumAllowedFontSize, autosizedFontSize));
    }
#endif

    setFontDescription(desc);
    font().update(currentFontSelector);
}

LayoutBoxExtent RenderStyle::shadowExtent(const ShadowData* shadow)
{
    LayoutUnit top;
    LayoutUnit right;
    LayoutUnit bottom;
    LayoutUnit left;

    for ( ; shadow; shadow = shadow->next()) {
        if (shadow->style() == ShadowStyle::Inset)
            continue;

        LayoutUnit extentAndSpread = shadow->paintingExtent() + LayoutUnit(shadow->spread().value());
        top = std::min<LayoutUnit>(top, LayoutUnit(shadow->y().value()) - extentAndSpread);
        right = std::max<LayoutUnit>(right, LayoutUnit(shadow->x().value()) + extentAndSpread);
        bottom = std::max<LayoutUnit>(bottom, LayoutUnit(shadow->y().value()) + extentAndSpread);
        left = std::min<LayoutUnit>(left, LayoutUnit(shadow->x().value()) - extentAndSpread);
    }

	return LayoutBoxExtent(top, right, bottom, left);
}

LayoutBoxExtent RenderStyle::shadowInsetExtent(const ShadowData* shadow)
{
    LayoutUnit top;
    LayoutUnit right;
    LayoutUnit bottom;
    LayoutUnit left;

    for ( ; shadow; shadow = shadow->next()) {
        if (shadow->style() == ShadowStyle::Normal)
            continue;

        LayoutUnit extentAndSpread = shadow->paintingExtent() + LayoutUnit(shadow->spread().value());
        top = std::max<LayoutUnit>(top, LayoutUnit(shadow->y().value()) + extentAndSpread);
        right = std::min<LayoutUnit>(right, LayoutUnit(shadow->x().value()) - extentAndSpread);
        bottom = std::min<LayoutUnit>(bottom, LayoutUnit(shadow->y().value()) - extentAndSpread);
        left = std::max<LayoutUnit>(left, LayoutUnit(shadow->x().value()) + extentAndSpread);
    }

    return LayoutBoxExtent(top, right, bottom, left);
}

void RenderStyle::getShadowHorizontalExtent(const ShadowData* shadow, LayoutUnit &left, LayoutUnit &right)
{
    left = 0;
    right = 0;

    for ( ; shadow; shadow = shadow->next()) {
        if (shadow->style() == ShadowStyle::Inset)
            continue;

        LayoutUnit extentAndSpread = shadow->paintingExtent() + LayoutUnit(shadow->spread().value());
        left = std::min<LayoutUnit>(left, LayoutUnit(shadow->x().value()) - extentAndSpread);
        right = std::max<LayoutUnit>(right, LayoutUnit(shadow->x().value()) + extentAndSpread);
    }
}

void RenderStyle::getShadowVerticalExtent(const ShadowData* shadow, LayoutUnit &top, LayoutUnit &bottom)
{
    top = 0;
    bottom = 0;

    for ( ; shadow; shadow = shadow->next()) {
        if (shadow->style() == ShadowStyle::Inset)
            continue;

        LayoutUnit extentAndSpread = shadow->paintingExtent() + LayoutUnit(shadow->spread().value());
        top = std::min<LayoutUnit>(top, LayoutUnit(shadow->y().intValue()) - extentAndSpread);
        bottom = std::max<LayoutUnit>(bottom, LayoutUnit(shadow->y().intValue()) + extentAndSpread);
    }
}

Color RenderStyle::unresolvedColorForProperty(CSSPropertyID colorProperty, bool visitedLink) const
{
    switch (colorProperty) {
    case CSSPropertyColor:
        return visitedLink ? visitedLinkColor() : color();
    case CSSPropertyBackgroundColor:
        return visitedLink ? visitedLinkBackgroundColor() : backgroundColor();
    case CSSPropertyBorderBottomColor:
        return visitedLink ? visitedLinkBorderBottomColor() : borderBottomColor();
    case CSSPropertyBorderLeftColor:
        return visitedLink ? visitedLinkBorderLeftColor() : borderLeftColor();
    case CSSPropertyBorderRightColor:
        return visitedLink ? visitedLinkBorderRightColor() : borderRightColor();
    case CSSPropertyBorderTopColor:
        return visitedLink ? visitedLinkBorderTopColor() : borderTopColor();
    case CSSPropertyFill:
        return fillPaintColor();
    case CSSPropertyFloodColor:
        return floodColor();
    case CSSPropertyLightingColor:
        return lightingColor();
    case CSSPropertyOutlineColor:
        return visitedLink ? visitedLinkOutlineColor() : outlineColor();
    case CSSPropertyStopColor:
        return stopColor();
    case CSSPropertyStroke:
        return strokePaintColor();
    case CSSPropertyStrokeColor:
        return visitedLink ? visitedLinkStrokeColor() : strokeColor();
    case CSSPropertyBorderBlockEndColor:
    case CSSPropertyBorderBlockStartColor:
    case CSSPropertyBorderInlineEndColor:
    case CSSPropertyBorderInlineStartColor:
        return unresolvedColorForProperty(CSSProperty::resolveDirectionAwareProperty(colorProperty, direction(), writingMode()));
    case CSSPropertyColumnRuleColor:
        return visitedLink ? visitedLinkColumnRuleColor() : columnRuleColor();
    case CSSPropertyWebkitTextEmphasisColor:
        return visitedLink ? visitedLinkTextEmphasisColor() : textEmphasisColor();
    case CSSPropertyWebkitTextFillColor:
        return visitedLink ? visitedLinkTextFillColor() : textFillColor();
    case CSSPropertyWebkitTextStrokeColor:
        return visitedLink ? visitedLinkTextStrokeColor() : textStrokeColor();
    case CSSPropertyTextDecorationColor:
        return visitedLink ? visitedLinkTextDecorationColor() : textDecorationColor();
    case CSSPropertyCaretColor:
        return visitedLink ? visitedLinkCaretColor() : caretColor();
    default:
        ASSERT_NOT_REACHED();
        break;
    }

    return Color();
}

Color RenderStyle::colorResolvingCurrentColor(CSSPropertyID colorProperty, bool visitedLink) const
{
    auto result = unresolvedColorForProperty(colorProperty, visitedLink);

    if (isCurrentColor(result)) {
        if (colorProperty == CSSPropertyTextDecorationColor) {
            if (hasPositiveStrokeWidth()) {
                // Prefer stroke color if possible but not if it's fully transparent.
                auto strokeColor = colorResolvingCurrentColor(effectiveStrokeColorProperty(), visitedLink);
                if (strokeColor.isVisible())
                    return strokeColor;
            }

            return colorResolvingCurrentColor(CSSPropertyWebkitTextFillColor, visitedLink);
        }

        return visitedLink ? visitedLinkColor() : color();
    }

    return result;
}

Color RenderStyle::colorResolvingCurrentColor(const Color& color) const
{
    if (isCurrentColor(color))
        return this->color();

    return color;
}

DLL_PUBLIC Color RenderStyle::visitedDependentColor(CSSPropertyID colorProperty) const
{
    Color unvisitedColor = colorResolvingCurrentColor(colorProperty, false);
    if (insideLink() != InsideVisitedLink)
        return unvisitedColor;

    Color visitedColor = colorResolvingCurrentColor(colorProperty, true);

    // FIXME: Technically someone could explicitly specify the color transparent, but for now we'll just
    // assume that if the background color is transparent that it wasn't set. Note that it's weird that
    // we're returning unvisited info for a visited link, but given our restriction that the alpha values
    // have to match, it makes more sense to return the unvisited background color if specified than it
    // does to return black. This behavior matches what Firefox 4 does as well.
    if (colorProperty == CSSPropertyBackgroundColor && visitedColor == Color::transparent)
        return unvisitedColor;

    // Take the alpha from the unvisited color, but get the RGB values from the visited color.
    return Color::createUnCheked(visitedColor.red(), visitedColor.green(), visitedColor.blue(), unvisitedColor.alpha());
}

Color RenderStyle::visitedDependentColorWithColorFilter(CSSPropertyID colorProperty) const
{
    if (!hasAppleColorFilter())
        return visitedDependentColor(colorProperty);

    return colorByApplyingColorFilter(visitedDependentColor(colorProperty));
}

Color RenderStyle::colorByApplyingColorFilter(const Color& color) const
{
    Color transformedColor = color;
    appleColorFilter().transformColor(transformedColor);
    return transformedColor;
}

const BorderValue& RenderStyle::borderBefore() const
{
    switch (writingMode()) {
    case TopToBottomWritingMode:
        return borderTop();
    case BottomToTopWritingMode:
        return borderBottom();
    case LeftToRightWritingMode:
        return borderLeft();
    case RightToLeftWritingMode:
        return borderRight();
    }
    ASSERT_NOT_REACHED();
    return borderTop();
}

const BorderValue& RenderStyle::borderAfter() const
{
    switch (writingMode()) {
    case TopToBottomWritingMode:
        return borderBottom();
    case BottomToTopWritingMode:
        return borderTop();
    case LeftToRightWritingMode:
        return borderRight();
    case RightToLeftWritingMode:
        return borderLeft();
    }
    ASSERT_NOT_REACHED();
    return borderBottom();
}

const BorderValue& RenderStyle::borderStart() const
{
    if (isHorizontalWritingMode())
        return isLeftToRightDirection() ? borderLeft() : borderRight();
    return isLeftToRightDirection() ? borderTop() : borderBottom();
}

const BorderValue& RenderStyle::borderEnd() const
{
    if (isHorizontalWritingMode())
        return isLeftToRightDirection() ? borderRight() : borderLeft();
    return isLeftToRightDirection() ? borderBottom() : borderTop();
}

float RenderStyle::borderBeforeWidth() const
{
    switch (writingMode()) {
    case TopToBottomWritingMode:
        return borderTopWidth();
    case BottomToTopWritingMode:
        return borderBottomWidth();
    case LeftToRightWritingMode:
        return borderLeftWidth();
    case RightToLeftWritingMode:
        return borderRightWidth();
    }
    ASSERT_NOT_REACHED();
    return borderTopWidth();
}

float RenderStyle::borderAfterWidth() const
{
    switch (writingMode()) {
    case TopToBottomWritingMode:
        return borderBottomWidth();
    case BottomToTopWritingMode:
        return borderTopWidth();
    case LeftToRightWritingMode:
        return borderRightWidth();
    case RightToLeftWritingMode:
        return borderLeftWidth();
    }
    ASSERT_NOT_REACHED();
    return borderBottomWidth();
}

float RenderStyle::borderStartWidth() const
{
    if (isHorizontalWritingMode())
        return isLeftToRightDirection() ? borderLeftWidth() : borderRightWidth();
    return isLeftToRightDirection() ? borderTopWidth() : borderBottomWidth();
}

float RenderStyle::borderEndWidth() const
{
    if (isHorizontalWritingMode())
        return isLeftToRightDirection() ? borderRightWidth() : borderLeftWidth();
    return isLeftToRightDirection() ? borderBottomWidth() : borderTopWidth();
}

void RenderStyle::setMarginStart(Length margin)
{
    if (isHorizontalWritingMode()) {
        if (isLeftToRightDirection())
            setMarginLeft(margin);
        else
            setMarginRight(margin);
    } else {
        if (isLeftToRightDirection())
            setMarginTop(margin);
        else
            setMarginBottom(margin);
    }
}

void RenderStyle::setMarginEnd(Length margin)
{
    if (isHorizontalWritingMode()) {
        if (isLeftToRightDirection())
            setMarginRight(margin);
        else
            setMarginLeft(margin);
    } else {
        if (isLeftToRightDirection())
            setMarginBottom(margin);
        else
            setMarginTop(margin);
    }
}

TextEmphasisMark RenderStyle::textEmphasisMark() const
{
    TextEmphasisMark mark = static_cast<TextEmphasisMark>(m_rareInheritedData->textEmphasisMark);
    if (mark != TextEmphasisMarkAuto)
        return mark;

    if (isHorizontalWritingMode())
        return TextEmphasisMarkDot;

    return TextEmphasisMarkSesame;
}

#if ENABLE(TOUCH_EVENTS)
Color RenderStyle::initialTapHighlightColor()
{
    return RenderTheme::tapHighlightColor();
}
#endif

LayoutBoxExtent RenderStyle::imageOutsets(const NinePieceImage& image) const
{
    return LayoutBoxExtent(NinePieceImage::computeOutset(image.outset().top(), borderTopWidth()),
                           NinePieceImage::computeOutset(image.outset().right(), borderRightWidth()),
                           NinePieceImage::computeOutset(image.outset().bottom(), borderBottomWidth()),
                           NinePieceImage::computeOutset(image.outset().left(), borderLeftWidth()));
}

std::pair<FontOrientation, NonCJKGlyphOrientation> RenderStyle::fontAndGlyphOrientation()
{
    // FIXME: TextOrientationSideways should map to sideways-left in vertical-lr, which is not supported yet.

    if (isHorizontalWritingMode())
        return { FontOrientation::Horizontal, NonCJKGlyphOrientation::Mixed };

    switch (textOrientation()) {
    case TextOrientation::Mixed:
        return { FontOrientation::Vertical, NonCJKGlyphOrientation::Mixed };
    case TextOrientation::Upright:
        return { FontOrientation::Vertical, NonCJKGlyphOrientation::Upright };
    case TextOrientation::Sideways:
        return { FontOrientation::Horizontal, NonCJKGlyphOrientation::Mixed };
    case TextOrientation::SidewaysRight:
        return { FontOrientation::Horizontal, NonCJKGlyphOrientation::Mixed };
    default:
        ASSERT_NOT_REACHED();
        return { FontOrientation::Horizontal, NonCJKGlyphOrientation::Mixed };
    }
}

void RenderStyle::setBorderImageSource(PassRefPtr<StyleImage> image)
{
    if (m_surroundData->border.m_image.image() == image.get())
        return;
    m_surroundData.access().border.m_image.setImage(image);
}

void RenderStyle::setBorderImageSliceFill(bool fill)
{
    if (m_surroundData->border.m_image.fill() == fill)
        return;
    m_surroundData.access().border.m_image.setFill(fill);
}

void RenderStyle::setBorderImageSlices(const LengthBox& slices)
{
    if (m_surroundData->border.m_image.imageSlices() == slices)
        return;
    m_surroundData.access().border.m_image.setImageSlices(slices);
}

void RenderStyle::setBorderImageWidth(const LengthBox& slices)
{
    if (m_surroundData->border.m_image.borderSlices() == slices)
        return;
    m_surroundData.access().border.m_image.setBorderSlices(slices);
}

void RenderStyle::setBorderImageOutset(const LengthBox& outset)
{
    if (m_surroundData->border.m_image.outset() == outset)
        return;
    m_surroundData.access().border.m_image.setOutset(outset);
}

Vector<PaintType, 3> RenderStyle::paintTypesForPaintOrder(PaintOrder order)
{
    Vector<PaintType, 3> paintOrder;
    switch (order) {
    case PaintOrder::Normal:
        FALLTHROUGH;
    case PaintOrder::Fill:
        paintOrder.append(PaintType::Fill);
        paintOrder.append(PaintType::Stroke);
        paintOrder.append(PaintType::Markers);
        break;
    case PaintOrder::FillMarkers:
        paintOrder.append(PaintType::Fill);
        paintOrder.append(PaintType::Markers);
        paintOrder.append(PaintType::Stroke);
        break;
    case PaintOrder::Stroke:
        paintOrder.append(PaintType::Stroke);
        paintOrder.append(PaintType::Fill);
        paintOrder.append(PaintType::Markers);
        break;
    case PaintOrder::StrokeMarkers:
        paintOrder.append(PaintType::Stroke);
        paintOrder.append(PaintType::Markers);
        paintOrder.append(PaintType::Fill);
        break;
    case PaintOrder::Markers:
        paintOrder.append(PaintType::Markers);
        paintOrder.append(PaintType::Fill);
        paintOrder.append(PaintType::Stroke);
        break;
    case PaintOrder::MarkersStroke:
        paintOrder.append(PaintType::Markers);
        paintOrder.append(PaintType::Stroke);
        paintOrder.append(PaintType::Fill);
        break;
    };
    return paintOrder;
}

void RenderStyle::setColumnStylesFromPaginationMode(const Pagination::Mode& paginationMode)
{
    if (paginationMode == Pagination::Unpaginated)
        return;

    setColumnFill(ColumnFillAuto);

    switch (paginationMode) {
    case Pagination::LeftToRightPaginated:
        setColumnAxis(HorizontalColumnAxis);
        if (isHorizontalWritingMode())
            setColumnProgression(isLeftToRightDirection() ? NormalColumnProgression : ReverseColumnProgression);
        else
            setColumnProgression(isFlippedBlocksWritingMode() ? ReverseColumnProgression : NormalColumnProgression);
        break;
    case Pagination::RightToLeftPaginated:
        setColumnAxis(HorizontalColumnAxis);
        if (isHorizontalWritingMode())
            setColumnProgression(isLeftToRightDirection() ? ReverseColumnProgression : NormalColumnProgression);
        else
            setColumnProgression(isFlippedBlocksWritingMode() ? NormalColumnProgression : ReverseColumnProgression);
        break;
    case Pagination::TopToBottomPaginated:
        setColumnAxis(VerticalColumnAxis);
        if (isHorizontalWritingMode())
            setColumnProgression(isFlippedBlocksWritingMode() ? ReverseColumnProgression : NormalColumnProgression);
        else
            setColumnProgression(isLeftToRightDirection() ? NormalColumnProgression : ReverseColumnProgression);
        break;
    case Pagination::BottomToTopPaginated:
        setColumnAxis(VerticalColumnAxis);
        if (isHorizontalWritingMode())
            setColumnProgression(isFlippedBlocksWritingMode() ? NormalColumnProgression : ReverseColumnProgression);
        else
            setColumnProgression(isLeftToRightDirection() ? ReverseColumnProgression : NormalColumnProgression);
        break;
    case Pagination::Unpaginated:
        ASSERT_NOT_REACHED();
        break;
    }
}

bool RenderStyle::hasReferenceFilterOnly() const
{
    if (!hasFilter())
        return false;

    const auto& filterOperations = m_rareNonInheritedData->m_filter->m_operations;
    if (filterOperations.size() != 1)
        return false;

    const auto& filterOperation = *filterOperations.at(0);
    if (filterOperation.type() != FilterOperation::REFERENCE)
        return false;

    return true;
}

void RenderStyle::deduplicateInheritedCustomProperties(const RenderStyle& other)
{
    auto& properties = const_cast<DataRef<StyleCustomPropertyData>&>(m_rareInheritedData->customProperties);
    auto& otherProperties = other.m_rareInheritedData->customProperties;
    if (properties.ptr() != otherProperties.ptr() && *properties == *otherProperties)
        properties = otherProperties;
}

void RenderStyle::setInheritedCustomPropertyValue(const AtomicString& name, PassRefPtr<CSSCustomPropertyValue> value)
{
    auto* existingValue = m_rareInheritedData->customProperties->values.get(name);
    if (existingValue && existingValue->equals(*value.get()))
        return;
    m_rareInheritedData.access().customProperties.access().setCustomPropertyValue(name, value);
}

void RenderStyle::setNonInheritedCustomPropertyValue(const AtomicString& name, PassRefPtr<CSSCustomPropertyValue> value)
{
    auto* existingValue = m_rareNonInheritedData->customProperties->values.get(name);
    if (existingValue && existingValue->equals(*value.get()))
        return;
    m_rareNonInheritedData.access().customProperties.access().setCustomPropertyValue(name, value);
}

float RenderStyle::outlineWidth() const
{
    if (m_background->outline.style() == BNONE)
        return 0;
    if (outlineStyleIsAuto())
        return std::max(m_background->outline.width(), RenderTheme::platformFocusRingWidth());
    return m_background->outline.width();
}

float RenderStyle::outlineOffset() const
{
    if (outlineStyleIsAuto())
        return (m_background->outline.offset() + RenderTheme::platformFocusRingOffset(outlineWidth()));
    return m_background->outline.offset();
}

bool RenderStyle::shouldPlaceVerticalScrollbarOnLeft() const
{
    return (!isLeftToRightDirection() && isHorizontalWritingMode()) || writingMode() == RightToLeftWritingMode;
}

float RenderStyle::computedStrokeWidth(const IntSize& viewportSize) const
{
    // Use the stroke-width and stroke-color value combination only if stroke-color has been explicitly specified.
    // Since there will be no visible stroke when stroke-color is not specified (transparent by default), we fall
    // back to the legacy Webkit text stroke combination in that case.
    if (!hasExplicitlySetStrokeColor())
        return textStrokeWidth();

    auto& length = strokeWidth();

    if (length.isPercent()) {
        // According to the spec, https://drafts.fxtf.org/paint/#stroke-width, the percentage is relative to the scaled viewport size.
        // The scaled viewport size is the geometric mean of the viewport width and height.
    	float result = length.value() * (viewportSize.width() + viewportSize.height()) / 200.0f;
        return result;
    }

    if (length.isAuto() || !length.isSpecified())
        return 0;

    return floatValueForLength(length, viewportSize.width());
}

bool RenderStyle::hasPositiveStrokeWidth() const
{
    if (!hasExplicitlySetStrokeWidth())
        return textStrokeWidth() > 0;

    return strokeWidth().isPositive();
}

Color RenderStyle::computedStrokeColor() const
{
    return visitedDependentColor(effectiveStrokeColorProperty());
}

UsedClear RenderStyle::usedClear(const RenderObject& renderer)
{
    auto computedValue = renderer.style()->clear();
    switch (computedValue) {
    case Clear::None:
        return UsedClear::None;
    case Clear::Left:
        return UsedClear::Left;
    case Clear::Right:
        return UsedClear::Right;
    case Clear::Both:
        return UsedClear::Both;
    case Clear::InlineStart:
    case Clear::InlineEnd: {
        auto containingBlockDirection = renderer.containingBlock()->style()->direction();
        if (containingBlockDirection == TextDirection::RTL)
            return computedValue == Clear::InlineStart ? UsedClear::Right : UsedClear::Left;
        return computedValue == Clear::InlineStart ? UsedClear::Left : UsedClear::Right;
    }
    default:
        return UsedClear::None;
    }
}

UsedFloat RenderStyle::usedFloat(const RenderObject& renderer)
{
    auto computedValue = renderer.style()->floating();
    switch (computedValue) {
    case Float::None:
        return UsedFloat::None;
    case Float::Left:
        return UsedFloat::Left;
    case Float::Right:
        return UsedFloat::Right;
    case Float::InlineStart:
    case Float::InlineEnd: {
        auto containingBlockDirection = renderer.containingBlock()->style()->direction();
        if (containingBlockDirection == TextDirection::RTL)
            return computedValue == Float::InlineStart ? UsedFloat::Right : UsedFloat::Left;
        return computedValue == Float::InlineStart ? UsedFloat::Left : UsedFloat::Right;
    }
    default:
        return UsedFloat::None;
    }
}

EUserSelect RenderStyle::effectiveUserSelect() const
{
    if (effectiveInert())
        return EUserSelect::SELECT_NONE;

    auto value = userSelect();
    if (userModify() != EUserModify::READ_ONLY && userDrag() != EUserDrag::DRAG_ELEMENT)
        return value == EUserSelect::SELECT_NONE ? EUserSelect::SELECT_TEXT : value;

    return value;
}

} // namespace WebCore
