/*
 * Copyright (C) 2015 Apple Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "RenderStyleConstants.h"

#include "TextStream.h"

namespace WebCore {

TextStream& operator<<(TextStream& ts, AspectRatioType aspectRatioType)
{
    switch (aspectRatioType) {
    case AspectRatioType::Auto: ts << "auto"; break;
    case AspectRatioType::Ratio: ts << "ratio"; break;
    case AspectRatioType::AutoAndRatio: ts << "autoandratio"; break;
    }
    return ts;
}

TextStream& operator<<(TextStream& ts, EFillSizeType sizeType)
{
    switch (sizeType) {
    case Contain: ts << "contain"; break;
    case Cover: ts << "cover"; break;
    case SizeLength: ts << "size-length"; break;
    case SizeNone: ts << "size-none"; break;
    }
    
    return ts;
}

TextStream& operator<<(TextStream& ts, EFillAttachment attachment)
{
    switch (attachment) {
    case ScrollBackgroundAttachment: ts << "scroll"; break;
    case LocalBackgroundAttachment: ts << "local"; break;
    case FixedBackgroundAttachment: ts << "fixed"; break;
    }
    return ts;
}

TextStream& operator<<(TextStream& ts, EFillBox fill)
{
    switch (fill) {
    case BorderFillBox: ts << "border"; break;
    case PaddingFillBox: ts << "padding"; break;
    case ContentFillBox: ts << "content"; break;
    case TextFillBox: ts << "text"; break;
    case NoClipFillBox: ts << "no-clip"; break;
    }
    return ts;
}

TextStream& operator<<(TextStream& ts, EFillRepeat repeat)
{
    switch (repeat) {
    case RepeatFill: ts << "repeat"; break;
    case NoRepeatFill: ts << "no-repeat"; break;
    case RoundFill: ts << "round"; break;
    case SpaceFill: ts << "space"; break;
    }

    return ts;
}

TextStream& operator<<(TextStream& ts, MaskMode maskMode)
{
    switch (maskMode) {
    case MaskMode::Alpha: ts << "alpha"; break;
    case MaskMode::Luminance: ts << "luminance"; break;
    case MaskMode::MatchSource: ts << "match-source"; break;
    }

    return ts;
}

TextStream& operator<<(TextStream& ts, Edge edge)
{
    switch (edge) {
    case Edge::Top: ts << "top"; break;
    case Edge::Right: ts << "right"; break;
    case Edge::Bottom: ts << "bottom"; break;
    case Edge::Left: ts << "left"; break;
    }
    return ts;
}

bool alwaysPageBreak(BreakBetween between)
{
    return between >= PageBreakBetween;
}

const float defaultMiterLimit = 4;

CSSBoxType transformBoxToCSSBoxType(TransformBox transformBox)
{
    switch (transformBox) {
    case TransformBox::StrokeBox:
        return CSSBoxType::StrokeBox;
    case TransformBox::ContentBox:
        return CSSBoxType::ContentBox;
    case TransformBox::BorderBox:
        return CSSBoxType::BorderBox;
    case TransformBox::FillBox:
        return CSSBoxType::FillBox;
    case TransformBox::ViewBox:
        return CSSBoxType::ViewBox;
    default:
        ASSERT_NOT_REACHED();
        return CSSBoxType::BorderBox;
    }
}

TextStream& operator<<(TextStream& ts, PseudoId pseudoId)
{
    switch (pseudoId) {
    case PseudoId::None: ts << "none"; break;
    case PseudoId::FirstLine: ts << "first-line"; break;
    case PseudoId::FirstLetter: ts << "first-letter"; break;
    case PseudoId::Highlight: ts << "highlight"; break;
    case PseudoId::Marker: ts << "marker"; break;
    case PseudoId::Backdrop: ts << "backdrop"; break;
    case PseudoId::Before: ts << "before"; break;
    case PseudoId::After: ts << "after"; break;
    case PseudoId::Selection: ts << "selection"; break;
    case PseudoId::Scrollbar: ts << "scrollbar"; break;
    case PseudoId::ScrollbarThumb: ts << "scrollbar-thumb"; break;
    case PseudoId::ScrollbarButton: ts << "scrollbar-button"; break;
    case PseudoId::ScrollbarTrack: ts << "scrollbar-track"; break;
    case PseudoId::ScrollbarTrackPiece: ts << "scrollbar-trackpiece"; break;
    case PseudoId::ScrollbarCorner: ts << "scrollbar-corner"; break;
    case PseudoId::Resizer: ts << "resizer"; break;
    default:
        ts << "other";
        break;
    }
    return ts;
}

TextStream& operator<<(TextStream& ts, Overflow overflow)
{
#if 0
    switch (overflow) {
    case Overflow::Visible: ts << "visible"; break;
    case Overflow::Hidden: ts << "hidden"; break;
    case Overflow::Scroll: ts << "scroll"; break;
    case Overflow::Auto: ts << "auto"; break;
    case Overflow::PagedX: ts << "paged-x"; break;
    case Overflow::PagedY: ts << "paged-y"; break;
    case Overflow::Clip: ts << "clip"; break;
    }
#endif
    return ts;
}

} // namespace WebCore
