/*
 * Copyright (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "StyleRareNonInheritedData.h"

#include "ContentData.h"
#include "RenderCounter.h"
#include "RenderStyle.h"
#include "ShadowData.h"
#include "StyleFilterData.h"
#include "StyleTransformData.h"
#include "StyleImage.h"
#include "StyleResolver.h"
#include <wtf/OwnPtrComparison.h>
#include <wtf/PointerComparison.h>
#include <wtf/RefPtr.h>

namespace WebCore {

StyleRareNonInheritedData::StyleRareNonInheritedData()
    : opacity(RenderStyle::initialOpacity())
	, aspectRatioWidth(RenderStyle::initialAspectRatioWidth())
	, aspectRatioHeight(RenderStyle::initialAspectRatioHeight())
    , m_perspective(RenderStyle::initialPerspective())
    , m_perspectiveOriginX(RenderStyle::initialPerspectiveOriginX())
    , m_perspectiveOriginY(RenderStyle::initialPerspectiveOriginY())
    , lineClamp(RenderStyle::initialLineClamp())
	, initialLetter(RenderStyle::initialInitialLetter())
	, deprecatedFlexibleBox(StyleDeprecatedFlexibleBoxData::create())
	, flexibleBox(StyleFlexibleBoxData::create())
	, m_marquee(StyleMarqueeData::create())
	, multiCol(StyleMultiColData::create())
	, transform(StyleTransformData::create())
#if ENABLE(CSS_FILTERS)
	, m_filter(StyleFilterData::create())
#endif
#if ENABLE(FILTERS_LEVEL_2)
	, m_backdropFilter(StyleFilterData::create())
#endif
	, grid(StyleGridData::create())
	, gridItem(StyleGridItemData::create())
	, willChange(RenderStyle::initialWillChange())
    , mask(FillLayer::create(MaskFillLayer))
	, m_maskBoxImage(NinePieceImage::Type::Mask)
	, m_objectPosition(RenderStyle::initialObjectPosition())
    , m_shapeOutside(RenderStyle::initialShapeOutside())
    , shapeMargin(RenderStyle::initialShapeMargin())
	, m_shapeImageThreshold(RenderStyle::initialShapeImageThreshold())
	, m_clipPath(RenderStyle::initialClipPath())
    , m_visitedLinkBackgroundColor(RenderStyle::initialBackgroundColor())
    , order(RenderStyle::initialOrder())
	, alignContent(RenderStyle::initialContentAlignment())
	, alignItems(RenderStyle::initialDefaultAlignment())
	, alignSelf(RenderStyle::initialSelfAlignment())
	, justifyContent(RenderStyle::initialContentAlignment())
	, justifyItems(RenderStyle::initialJustifyItems())
	, justifySelf(RenderStyle::initialSelfAlignment())
	, customProperties(StyleCustomPropertyData::create())
	, rotate(RenderStyle::initialRotate())
	, scale(RenderStyle::initialScale())
	, translate(RenderStyle::initialTranslate())
	, offsetPath(RenderStyle::initialOffsetPath())
    , m_pageSizeType(PAGE_SIZE_AUTO)
    , m_transformStyle3D(RenderStyle::initialTransformStyle3D())
	, transformStyleForcedToFlat(false)
    , m_backfaceVisibility(RenderStyle::initialBackfaceVisibility())
    , userDrag(RenderStyle::initialUserDrag())
    , textOverflow(RenderStyle::initialTextOverflow())
	, appearance(static_cast<unsigned>(RenderStyle::initialAppearance()))
	, effectiveAppearance(static_cast<unsigned>(RenderStyle::initialAppearance()))
    , textCombine(RenderStyle::initialTextCombine())
    , textDecorationStyle(RenderStyle::initialTextDecorationStyle())
	, textDecorationThickness(RenderStyle::initialTextDecorationThickness())
#if USE(ACCELERATED_COMPOSITING)
    , m_runningAcceleratedAnimation(false)
#endif
	, m_aspectRatioType(static_cast<unsigned>(RenderStyle::initialAspectRatioType()))
	, contentVisibility(static_cast<unsigned>(RenderStyle::initialContentVisibility()))
#if ENABLE(CSS_COMPOSITING)
    , m_effectiveBlendMode(RenderStyle::initialBlendMode())
	, isolation(RenderStyle::initialIsolation())
#endif
	, m_objectFit(RenderStyle::initialObjectFit())
	, breakBefore(RenderStyle::initialBreakBetween())
	, breakAfter(RenderStyle::initialBreakBetween())
	, breakInside(RenderStyle::initialBreakInside())
	, m_hasAttrContent(false)
	, columnGap(RenderStyle::initialColumnGap())
	, rowGap(RenderStyle::initialRowGap())
	, offsetDistance(RenderStyle::initialOffsetDistance())
	, offsetPosition(RenderStyle::initialOffsetPosition())
	, offsetAnchor(RenderStyle::initialOffsetAnchor())
{
}

StyleRareNonInheritedData::StyleRareNonInheritedData(const StyleRareNonInheritedData& o)
    : RefCounted<StyleRareNonInheritedData>()
    , opacity(o.opacity)
    , aspectRatioWidth(o.aspectRatioWidth)
    , aspectRatioHeight(o.aspectRatioHeight)
    , m_perspective(o.m_perspective)
    , m_perspectiveOriginX(o.m_perspectiveOriginX)
    , m_perspectiveOriginY(o.m_perspectiveOriginY)
    , lineClamp(o.lineClamp)
    , initialLetter(o.initialLetter)
    , deprecatedFlexibleBox(o.deprecatedFlexibleBox)
    , flexibleBox(o.flexibleBox)
    , m_marquee(o.m_marquee)
    , multiCol(o.multiCol)
    , transform(o.transform)
#if ENABLE(CSS_FILTERS)
    , m_filter(o.m_filter)
#endif
#if ENABLE(FILTERS_LEVEL_2)
    , m_backdropFilter(o.m_backdropFilter)
#endif
    , grid(o.grid)
    , gridItem(o.gridItem)
    , m_content(o.m_content ? o.m_content->clone() : nullptr)
    , m_counterDirectives(o.m_counterDirectives ? adoptPtr(new CounterDirectiveMap(*o.m_counterDirectives)) : nullptr)
	, m_altText(o.m_altText)
	, m_boxShadow(o.m_boxShadow ? adoptPtr(new ShadowData(*o.m_boxShadow)) : nullptr)
	, willChange(o.willChange)
    , boxReflect(o.boxReflect)
    , m_animations(o.m_animations ? o.m_animations->copy() : o.m_animations)
    , m_transitions(o.m_transitions ? o.m_transitions->copy() : o.m_transitions)
    , mask(o.mask)
    , m_maskBoxImage(o.m_maskBoxImage)
    , m_pageSize(o.m_pageSize)
    , m_objectPosition(o.m_objectPosition)
    , m_shapeOutside(o.m_shapeOutside)
    , shapeMargin(o.shapeMargin)
    , m_shapeImageThreshold(o.m_shapeImageThreshold)
    , m_clipPath(o.m_clipPath)
    , m_textDecorationColor(o.m_textDecorationColor)
    , m_visitedLinkTextDecorationColor(o.m_visitedLinkTextDecorationColor)
    , m_visitedLinkBackgroundColor(o.m_visitedLinkBackgroundColor)
    , m_visitedLinkOutlineColor(o.m_visitedLinkOutlineColor)
    , m_visitedLinkBorderLeftColor(o.m_visitedLinkBorderLeftColor)
    , m_visitedLinkBorderRightColor(o.m_visitedLinkBorderRightColor)
    , m_visitedLinkBorderTopColor(o.m_visitedLinkBorderTopColor)
    , m_visitedLinkBorderBottomColor(o.m_visitedLinkBorderBottomColor)
    , order(o.order)
    , alignContent(o.alignContent)
    , alignItems(o.alignItems)
    , alignSelf(o.alignSelf)
    , justifyContent(o.justifyContent)
    , justifyItems(o.justifyItems)
    , justifySelf(o.justifySelf)
    , customProperties(o.customProperties)
    , rotate(o.rotate)
	, scale(o.scale)
	, translate(o.translate)
    , offsetPath(o.offsetPath)
    , m_pageSizeType(o.m_pageSizeType)
    , m_transformStyle3D(o.m_transformStyle3D)
    , transformStyleForcedToFlat(o.transformStyleForcedToFlat)
    , m_backfaceVisibility(o.m_backfaceVisibility)
    , userDrag(o.userDrag)
    , textOverflow(o.textOverflow)
    , appearance(o.appearance)
    , effectiveAppearance(o.effectiveAppearance)
    , textCombine(o.textCombine)
    , textDecorationStyle(o.textDecorationStyle)
    , textDecorationThickness(o.textDecorationThickness)
#if USE(ACCELERATED_COMPOSITING)
    , m_runningAcceleratedAnimation(o.m_runningAcceleratedAnimation)
#endif
    , m_aspectRatioType(o.m_aspectRatioType)
    , contentVisibility(o.contentVisibility)
#if ENABLE(CSS_COMPOSITING)
    , m_effectiveBlendMode(o.m_effectiveBlendMode)
    , isolation(o.isolation)
#endif
    , m_objectFit(o.m_objectFit)
    , breakBefore(o.breakBefore)
    , breakAfter(o.breakAfter)
    , breakInside(o.breakInside)
    , m_hasAttrContent(o.m_hasAttrContent)
    , columnGap(o.columnGap)
    , rowGap(o.rowGap)
    , offsetDistance(o.offsetDistance)
    , offsetPosition(o.offsetPosition)
    , offsetAnchor(o.offsetAnchor)
	, m_navPropertyData(o.m_navPropertyData ? adoptPtr(new StyleNavPropertyData(*o.m_navPropertyData)) : nullptr)
    , m_characterType(o.m_characterType)
{
}

StyleRareNonInheritedData::~StyleRareNonInheritedData()
{
}

bool StyleRareNonInheritedData::operator==(const StyleRareNonInheritedData& o) const
{
    return opacity == o.opacity
		&& aspectRatioWidth == o.aspectRatioWidth
		&& aspectRatioHeight == o.aspectRatioHeight
        && m_perspective == o.m_perspective
        && m_perspectiveOriginX == o.m_perspectiveOriginX
        && m_perspectiveOriginY == o.m_perspectiveOriginY
        && lineClamp == o.lineClamp
        && initialLetter == o.initialLetter
#if ENABLE(DASHBOARD_SUPPORT)
        && m_dashboardRegions == o.m_dashboardRegions
#endif
        && deprecatedFlexibleBox == o.deprecatedFlexibleBox
        && flexibleBox == o.flexibleBox
        && m_marquee == o.m_marquee
        && multiCol == o.multiCol
        && transform == o.transform
#if ENABLE(CSS_FILTERS)
        && m_filter == o.m_filter
#endif
#if ENABLE(FILTERS_LEVEL_2)
        && m_backdropFilter == o.m_backdropFilter
#endif
        && grid == o.grid
        && gridItem == o.gridItem
        && contentDataEquivalent(o)
		&& arePointingToEqualData(m_counterDirectives, o.m_counterDirectives)
        && m_altText == o.m_altText
		&& arePointingToEqualData(m_boxShadow, o.m_boxShadow)
		&& arePointingToEqualData(willChange, o.willChange)
		&& arePointingToEqualData(boxReflect, o.boxReflect)
		&& arePointingToEqualData(m_animations, o.m_animations)
		&& arePointingToEqualData(m_transitions, o.m_transitions)
        && mask == o.mask
        && m_maskBoxImage == o.m_maskBoxImage
        && m_pageSize == o.m_pageSize
        && m_objectPosition == o.m_objectPosition
    	&& arePointingToEqualData(m_shapeOutside, o.m_shapeOutside)
        && shapeMargin == o.shapeMargin
        && m_shapeImageThreshold == o.m_shapeImageThreshold
        && arePointingToEqualData(m_clipPath, o.m_clipPath)
        && m_textDecorationColor == o.m_textDecorationColor
        && m_visitedLinkTextDecorationColor == o.m_visitedLinkTextDecorationColor
        && m_visitedLinkBackgroundColor == o.m_visitedLinkBackgroundColor
        && m_visitedLinkOutlineColor == o.m_visitedLinkOutlineColor
        && m_visitedLinkBorderLeftColor == o.m_visitedLinkBorderLeftColor
        && m_visitedLinkBorderRightColor == o.m_visitedLinkBorderRightColor
        && m_visitedLinkBorderTopColor == o.m_visitedLinkBorderTopColor
        && m_visitedLinkBorderBottomColor == o.m_visitedLinkBorderBottomColor
        && order == o.order
        && alignContent == o.alignContent
        && alignItems == o.alignItems
        && alignSelf == o.alignSelf
        && justifyContent == o.justifyContent
        && justifyItems == o.justifyItems
        && justifySelf == o.justifySelf
        && customProperties == o.customProperties
        && m_pageSizeType == o.m_pageSizeType
        && m_transformStyle3D == o.m_transformStyle3D
        && transformStyleForcedToFlat == o.transformStyleForcedToFlat
        && m_backfaceVisibility == o.m_backfaceVisibility
        && userDrag == o.userDrag
        && textOverflow == o.textOverflow
        && appearance == o.appearance
        && effectiveAppearance == o.effectiveAppearance
        && textCombine == o.textCombine
        && textDecorationStyle == o.textDecorationStyle
        && textDecorationThickness == o.textDecorationThickness
        && arePointingToEqualData(rotate, o.rotate)
        && arePointingToEqualData(scale, o.scale)
        && arePointingToEqualData(translate, o.translate)
        && arePointingToEqualData(offsetPath, o.offsetPath)
#if USE(ACCELERATED_COMPOSITING)
        && !m_runningAcceleratedAnimation && !o.m_runningAcceleratedAnimation
#endif
        && m_aspectRatioType == o.m_aspectRatioType
        && contentVisibility == o.contentVisibility
#if ENABLE(CSS_COMPOSITING)
        && m_effectiveBlendMode == o.m_effectiveBlendMode
        && isolation == o.isolation
#endif
        && m_objectFit == o.m_objectFit
        && breakAfter == o.breakAfter
        && breakBefore == o.breakBefore
        && breakInside == o.breakInside
        && m_hasAttrContent == o.m_hasAttrContent
        && columnGap == o.columnGap
        && rowGap == o.rowGap
        && offsetDistance == o.offsetDistance
        && offsetPosition == o.offsetPosition
        && offsetAnchor == o.offsetAnchor
		&& arePointingToEqualData(m_navPropertyData, o.m_navPropertyData)
		&& m_characterType == o.m_characterType;
}

bool StyleRareNonInheritedData::contentDataEquivalent(const StyleRareNonInheritedData& o) const
{
    ContentData* a = m_content.get();
    ContentData* b = o.m_content.get();

    while (a && b && *a == *b) {
        a = a->next();
        b = b->next();
    }

    return !a && !b;
}

#if ENABLE(CSS_FILTERS)
bool StyleRareNonInheritedData::hasFilters() const
{
    return !m_filter->m_operations.isEmpty();
}
#endif

#if ENABLE(FILTERS_LEVEL_2)
bool StyleRareNonInheritedData::hasBackdropFilters() const
{
    return !m_backdropFilter->m_operations.isEmpty();
}
#endif

} // namespace WebCore
