/*
 * Copyright (C) 2012 Adobe Systems Incorporated. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials
 *    provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef WrapShapes_h
#define WrapShapes_h

#include "Length.h"
#include "WindRule.h"
#include <wtf/RefCounted.h>
#include <wtf/RefPtr.h>
#include <wtf/Vector.h>

namespace WebCore {

class WrapShape : public WTF::RefCountedBase {
    WTF_MAKE_NONCOPYABLE(WrapShape); WTF_MAKE_FAST_ALLOCATED;
public:
    enum Type {
        WRAP_SHAPE_RECTANGLE = 1,
        WRAP_SHAPE_CIRCLE = 2,
        WRAP_SHAPE_ELLIPSE = 3,
        WRAP_SHAPE_POLYGON = 4
    };
    
    void deref() 
    {
        if (derefBase())
            destroy();
    }

    Type type() const { return m_type; }

protected:
    WrapShape(Type type)
        : m_type(type)
    { }
    
private:
    void destroy();
    Type m_type;
};

class WrapShapeRectangle : public WrapShape {
public:
    static PassRefPtr<WrapShapeRectangle> create() { return adoptRef(new WrapShapeRectangle); }

    const Length& x() const { return m_x; }
    const Length& y() const { return m_y; }
    const Length& width() const { return m_width; }
    const Length& height() const { return m_height; }
    const Length& cornerRadiusX() const { return m_cornerRadiusX; }
    const Length& cornerRadiusY() const { return m_cornerRadiusY; }

    void setX(const Length& x) { m_x = x; }
    void setY(const Length& y) { m_y = y; }
    void setWidth(const Length& width) { m_width = width; }
    void setHeight(const Length& height) { m_height = height; }
    void setCornerRadiusX(const Length& radiusX) { m_cornerRadiusX = radiusX; }
    void setCornerRadiusY(const Length& radiusY) { m_cornerRadiusY = radiusY; }
    
private:
    WrapShapeRectangle()
        : WrapShape(WRAP_SHAPE_RECTANGLE)
        , m_cornerRadiusX(LengthType::Undefined)
        , m_cornerRadiusY(LengthType::Undefined)
    { }

    Length m_y;
    Length m_x;
    Length m_width;
    Length m_height;
    Length m_cornerRadiusX;
    Length m_cornerRadiusY;
};

class WrapShapeCircle : public WrapShape {
public:
    static PassRefPtr<WrapShapeCircle> create() { return adoptRef(new WrapShapeCircle); }

    const Length& centerX() const { return m_centerX; }
    const Length& centerY() const { return m_centerY; }
    const Length& radius() const { return m_radius; }

    void setCenterX(const Length& centerX) { m_centerX = centerX; }
    void setCenterY(const Length& centerY) { m_centerY = centerY; }
    void setRadius(const Length& radius) { m_radius = radius; }

private:
    WrapShapeCircle() 
        : WrapShape(WRAP_SHAPE_CIRCLE)
    { }

    Length m_centerX;
    Length m_centerY;
    Length m_radius;
};

class WrapShapeEllipse : public WrapShape {
public:
    static PassRefPtr<WrapShapeEllipse> create() { return adoptRef(new WrapShapeEllipse); }

    const Length& centerX() const { return m_centerX; }
    const Length& centerY() const { return m_centerY; }
    const Length& radiusX() const { return m_radiusX; }
    const Length& radiusY() const { return m_radiusY; }

    void setCenterX(const Length& centerX) { m_centerX = centerX; }
    void setCenterY(const Length& centerY) { m_centerY = centerY; }
    void setRadiusX(const Length& radiusX) { m_radiusX = radiusX; }
    void setRadiusY(const Length& radiusY) { m_radiusY = radiusY; }

private:
    WrapShapeEllipse() 
        : WrapShape(WRAP_SHAPE_ELLIPSE)
    { }

    Length m_centerX;
    Length m_centerY;
    Length m_radiusX;
    Length m_radiusY;
};

class WrapShapePolygon : public WrapShape {
public:
    static PassRefPtr<WrapShapePolygon> create() { return adoptRef(new WrapShapePolygon); }

    WindRule windRule() const { return m_windRule; }
    const Vector<Length>& values() const { return m_values; }
    const Length& getXAt(unsigned i) const { return m_values.at(2 * i); }
    const Length& getYAt(unsigned i) const { return m_values.at(2 * i + 1); }

    void setWindRule(WindRule windRule) { m_windRule = windRule; }
    void appendPoint(const Length& x, const Length& y) { m_values.append(x); m_values.append(y); }

private:
    WrapShapePolygon()
        : WrapShape(WRAP_SHAPE_POLYGON)
        , m_windRule(RULE_NONZERO)
    { }

    WindRule m_windRule;
    Vector<Length> m_values;
};
}
#endif
