/*
 * Copyright (C) 2006-2023 Apple Inc.
 * Copyright (C) 2007 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) Research In Motion Limited 2010-2012. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#pragma once

#if ENABLE(SVG)
#include "AffineTransform.h"
#include "RenderSVGBlock.h"
#include "SVGTextLayoutAttributesBuilder.h"

namespace WebCore {

class RenderSVGInlineText;
class SVGTextElement;

class RenderSVGText FINAL : public RenderSVGBlock {
public:
    RenderSVGText(SVGTextElement&, PassOwnPtr<RenderStyle>);
    virtual ~RenderSVGText();

    virtual bool isChildAllowed(RenderObject&, const RenderStyle&) const FINAL;

    void setNeedsPositioningValuesUpdate() { m_needsPositioningValuesUpdate = true; }
    virtual void setNeedsTransformUpdate()  FINAL { m_needsTransformUpdate = true; }
    void setNeedsTextMetricsUpdate() { m_needsTextMetricsUpdate = true; }
    virtual FloatRect repaintRectInLocalCoordinates() const FINAL;

    static RenderSVGText* locateRenderSVGTextAncestor(RenderObject&);
    static const RenderSVGText* locateRenderSVGTextAncestor(const RenderObject&);

    bool needsReordering() const { return m_needsReordering; }
    Vector<SVGTextLayoutAttributes*>& layoutAttributes() { return m_layoutAttributes; }

    void subtreeChildWasAdded(RenderObject*);
    void subtreeChildWillBeRemoved(RenderObject*, Vector<SVGTextLayoutAttributes*, 2>& affectedAttributes);
    void subtreeChildWasRemoved(const Vector<SVGTextLayoutAttributes*, 2>& affectedAttributes);
    void willLayout();
    void subtreeTextDidChange(RenderSVGInlineText&);

    virtual FloatRect objectBoundingBox() const FINAL { return frameRect(); }
    virtual FloatRect strokeBoundingBox() const FINAL;

private:
    virtual const char* renderName() const  FINAL{ return "RenderSVGText"; }
    virtual bool isSVGText() const FINAL { return true; }

    virtual void paint(PaintInfo&, const LayoutPoint&) FINAL;
    virtual bool nodeAtPoint(const HitTestRequest&, HitTestResult&, const HitTestLocation& locationInContainer, const LayoutPoint& accumulatedOffset, HitTestAction) FINAL;
    virtual bool nodeAtFloatPoint(const HitTestRequest&, HitTestResult&, const FloatPoint& pointInParent, HitTestAction) FINAL;
    virtual VisiblePosition positionForPoint(const LayoutPoint&, const RenderFragmentContainer*) FINAL;

    virtual bool requiresLayer() const FINAL { return false; }
    virtual void layout() FINAL;

    virtual void absoluteQuads(Vector<FloatQuad>&, bool* wasFixed) const FINAL;

    virtual LayoutRect clippedOverflowRectForRepaint(RenderLayerModelObject* repaintContainer) const FINAL;
    virtual Optional<LayoutRect> computeVisibleRectInContainer(const LayoutRect&, const RenderLayerModelObject* container, VisibleRectContext) const override;
    virtual Optional<FloatRect> computeFloatVisibleRectInContainer(const FloatRect&, const RenderLayerModelObject* container, VisibleRectContext) const override;

    virtual void mapLocalToContainer(RenderLayerModelObject* repaintContainer, TransformState&, OptionSet<MapCoordinatesMode>, bool* wasFixed) const FINAL;
	virtual const RenderElement* pushMappingToContainer(const RenderLayerModelObject* ancestorToStopAt, RenderGeometryMap&) const FINAL;
    virtual void willBeDestroyed() FINAL;

    virtual void styleDidChange(StyleDifference, const RenderStyle* oldStyle) final;

    virtual const AffineTransform& localToParentTransform() const FINAL { return m_localTransform; }
    virtual AffineTransform localTransform() const FINAL { return m_localTransform; }

    bool shouldHandleSubtreeMutations() const;

    bool m_needsReordering;
    bool m_needsPositioningValuesUpdate;
    bool m_needsTransformUpdate;
    bool m_needsTextMetricsUpdate;
    AffineTransform m_localTransform;
    SVGTextLayoutAttributesBuilder m_layoutAttributesBuilder;
    Vector<SVGTextLayoutAttributes*> m_layoutAttributes;
};

inline bool isRenderSVGText(const RenderObject& renderer)
{
	return renderer.isSVGText();
}

template<> inline bool isRendererOfType<const RenderSVGText>(const RenderObject& renderer) { return renderer.isSVGText(); }
RENDER_OBJECT_TYPE_CASTS(RenderSVGText, isSVGText())

}

#endif // ENABLE(SVG)
