/*
 * Copyright (C) 2017 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "RenderTreeBuilder.h"

#if HAVE(ACCESSIBILITY)
#include "AXObjectCache.h"
#endif
#include "Frame.h"
#include "FrameSelection.h"
#include "RenderButton.h"
#include "RenderCounter.h"
#include "RenderDescendantIterator.h"
#include "RenderElement.h"
#include "RenderFullScreen.h"
#include "RenderGrid.h"
#include "RenderBR.h"
#include "RenderMathMLFenced.h"
#include "RenderMenuList.h"
#include "RenderMultiColumnFlow.h"
#include "RenderMultiColumnSpannerPlaceholder.h"
#include "RenderRuby.h"
#include "RenderRubyBase.h"
#include "RenderRubyRun.h"
#include "RenderSVGContainer.h"
#include "RenderSVGInline.h"
#include "LegacyRenderSVGRoot.h"
#include "RenderSVGText.h"
#include "RenderTable.h"
#include "RenderTableCell.h"
#include "RenderTableRow.h"
#include "RenderTableSection.h"
#include "RenderText.h"
#include "RenderTextFragment.h"
#include "RenderTreeBuilderBlock.h"
#include "RenderTreeBuilderBlockFlow.h"
#include "RenderTreeBuilderContinuation.h"
#include "RenderTreeBuilderFirstLetter.h"
#include "RenderTreeBuilderFormControls.h"
#include "RenderTreeBuilderFullScreen.h"
#include "RenderTreeBuilderInline.h"
#include "RenderTreeBuilderList.h"
#include "RenderTreeBuilderMathML.h"
#include "RenderTreeBuilderMultiColumn.h"
#include "RenderTreeBuilderRuby.h"
#include "RenderTreeBuilderSVG.h"
#include "RenderTreeBuilderTable.h"
#include <wtf/SetForScope.h>

namespace WebCore {

RenderTreeBuilder* RenderTreeBuilder::s_current;

static void getInlineRun(RenderObject* start, RenderObject* boundary, RenderObject*& inlineRunStart, RenderObject*& inlineRunEnd)
{
    // Beginning at |start| we find the largest contiguous run of inlines that
    // we can. We denote the run with start and end points, |inlineRunStart|
    // and |inlineRunEnd|. Note that these two values may be the same if
    // we encounter only one inline.
    //
    // We skip any non-inlines we encounter as long as we haven't found any
    // inlines yet.
    //
    // |boundary| indicates a non-inclusive boundary point. Regardless of whether |boundary|
    // is inline or not, we will not include it in a run with inlines before it. It's as though we encountered
    // a non-inline.

    // Start by skipping as many non-inlines as we can.
    auto* curr = start;
    bool sawInline;
    do {
        while (curr && !(curr->isInline() || curr->isFloatingOrOutOfFlowPositioned()))
            curr = curr->nextSibling();

        inlineRunStart = inlineRunEnd = curr;

        if (!curr)
            return; // No more inline children to be found.

        sawInline = curr->isInline();

        curr = curr->nextSibling();
        while (curr && (curr->isInline() || curr->isFloatingOrOutOfFlowPositioned()) && (curr != boundary)) {
            inlineRunEnd = curr;
            if (curr->isInline())
                sawInline = true;
            curr = curr->nextSibling();
        }
    } while (!sawInline);
}

RenderTreeBuilder::RenderTreeBuilder(RenderView& view)
    : m_view(view)
	, m_previous(nullptr)
	, m_firstLetterBuilder(adoptPtr(new FirstLetter(*this)))
	, m_listBuilder(adoptPtr(new List(*this)))
	, m_multiColumnBuilder(adoptPtr(new MultiColumn(*this)))
	, m_tableBuilder(adoptPtr(new Table(*this)))
	, m_rubyBuilder(adoptPtr(new Ruby(*this)))
	, m_formControlsBuilder(adoptPtr(new FormControls(*this)))
	, m_blockBuilder(adoptPtr(new Block(*this)))
	, m_blockFlowBuilder(adoptPtr(new BlockFlow(*this)))
	, m_inlineBuilder(adoptPtr(new Inline(*this)))
	, m_svgBuilder(adoptPtr(new SVG(*this)))
#if ENABLE(MATHML)
	, m_mathMLBuilder(adoptPtr(new MathML(*this)))
#endif
	, m_continuationBuilder(adoptPtr(new Continuation(*this)))
#if ENABLE(FULLSCREEN_API)
    , m_fullScreenBuilder(adoptPtr(new FullScreen(*this)))
#endif
	, m_hasBrokenContinuation(false)
	, m_internalMovesType(RenderObject::IsInternalMove::No)
{
    RELEASE_ASSERT(!s_current || &m_view != &s_current->m_view);
    m_previous = s_current;
    s_current = this;
}

RenderTreeBuilder::~RenderTreeBuilder()
{
    s_current = m_previous;
}

bool RenderTreeBuilder::isRebuildRootForChildren(const RenderElement& renderer)
{
    // If a (non-anonymous) child is added or removed to a rebuild root then we'll rebuild the full
    // subtree instead of trying to maintain the correct anonymous box structure on per-child basis.
    // This can greatly simplify the code needed to maintain the correct structure.

    auto display = renderer.style()->display();
    if (display == DisplayType::Ruby || display == DisplayType::RubyBlock)
        return true;

    return false;
}

void RenderTreeBuilder::destroy(RenderObject& renderer, CanCollapseAnonymousBlock canCollapseAnonymousBlock)
{
    ASSERT(renderer.parent());
    auto toDestroy = detach(*renderer.parent(), renderer, canCollapseAnonymousBlock);

#if ENABLE(FULLSCREEN_API)
    if (renderer.isRenderFullScreen())
        fullScreenBuilder().cleanupOnDestroy(toRenderFullScreen(renderer));
#endif

    if (isRenderTextFragment(renderer))
        firstLetterBuilder().cleanupOnDestroy(toRenderTextFragment(renderer));

    if (renderer.isBoxModelObject())
        continuationBuilder().cleanupOnDestroy(toRenderBoxModelObject(renderer));

    // We need to detach the subtree first so that the descendants don't have
    // access to previous/next sublings at detach().
    // FIXME: webkit.org/b/182909.
    if (!toDestroy->isRenderElement())
        return;

    auto& childToDestroy = toRenderElement(*toDestroy.get());
    while (childToDestroy.firstChild()) {
        auto& firstChild = *childToDestroy.firstChild();
        if (auto* node = firstChild.node())
            node->setRenderer(nullptr);
        destroy(firstChild);
    }
}

void RenderTreeBuilder::attach(RenderElement& parent, RenderPtr<RenderObject> child, RenderObject* beforeChild)
{
    auto insertRecursiveIfNeeded = [&](RenderElement& parentCandidate) {
        if (&parent == &parentCandidate) {
            // Parents inside multicols can't call internal attach directly.
            if (parent.isRenderBlockFlow() && toRenderBlockFlow(parent).multiColumnFlow()) {
                blockFlowBuilder().attach(toRenderBlockFlow(parent), WTFMove(child), beforeChild);
                return;
            }
            attachToRenderElement(parent, WTFMove(child), beforeChild);
            return;
        }
        attach(parentCandidate, WTFMove(child), beforeChild);
    };

    ASSERT(&parent.view() == &m_view);

    if (beforeChild && beforeChild->isText()) {
        if (auto* wrapperInline = toRenderText(*beforeChild).inlineWrapperForDisplayContents())
            beforeChild = wrapperInline;
    } else if (beforeChild && beforeChild->isBox()) {
        // Adjust the beforeChild if it happens to be a spanner and the its actual location is inside the fragmented flow.
        auto& beforeChildBox = toRenderBox(*beforeChild);
        if (auto* enclosingFragmentedFlow = parent.enclosingFragmentedFlow()) {
            auto columnSpannerPlaceholderForBeforeChild = [&]() -> RenderMultiColumnSpannerPlaceholder* {
                if (!(*enclosingFragmentedFlow).isRenderMultiColumnFlow())
                    return nullptr;
                auto& multiColumnFlow = toRenderMultiColumnFlow(*enclosingFragmentedFlow);
                return multiColumnFlow.findColumnSpannerPlaceholder(&beforeChildBox);
            };

            if (auto* spannerPlaceholder = columnSpannerPlaceholderForBeforeChild())
                beforeChild = spannerPlaceholder;
        }
    }

    if (parent.isTableRow()) {
        auto& parentCandidate = tableBuilder().findOrCreateParentForChild(toRenderTableRow(parent), *child, beforeChild);
        if (&parentCandidate == &parent) {
            tableBuilder().attach(toRenderTableRow(parentCandidate), WTFMove(child), beforeChild);
            return;
        }
        insertRecursiveIfNeeded(parentCandidate);
        return;
    }

    if (parent.isTableSection()) {
        auto& parentCandidate = tableBuilder().findOrCreateParentForChild(toRenderTableSection(parent), *child, beforeChild);
        if (&parent == &parentCandidate) {
            tableBuilder().attach(toRenderTableSection(parent), WTFMove(child), beforeChild);
            return;
        }
        insertRecursiveIfNeeded(parentCandidate);
        return;
    }

    if (parent.isTable()) {
        auto& parentCandidate = tableBuilder().findOrCreateParentForChild(toRenderTable(parent), *child, beforeChild);
        if (&parentCandidate == &parent) {
            tableBuilder().attach(toRenderTable(parentCandidate), WTFMove(child), beforeChild);
            return;
        }
        insertRecursiveIfNeeded(parentCandidate);
        return;
    }

    if (parent.isRubyBlock()) {
        insertRecursiveIfNeeded(rubyBuilder().findOrCreateParentForChild(toRenderRubyAsBlock(parent), *child, beforeChild));
        return;
    }

    if (parent.isRubyInline()) {
        insertRecursiveIfNeeded(rubyBuilder().findOrCreateParentForChild(toRenderRubyAsInline(parent), *child, beforeChild));
        return;
    }

    if (parent.isRubyRun()) {
    	rubyBuilder().attach(toRenderRubyRun(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isRenderButton()) {
        formControlsBuilder().attach(toRenderButton(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isMenuList()) {
        formControlsBuilder().attach(toRenderMenuList(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isSVGContainer()) {
        svgBuilder().attach(toRenderSVGContainer(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isSVGInline()) {
        svgBuilder().attach(toRenderSVGInline(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isSVGRoot()) {
        svgBuilder().attach(toRenderSVGRoot(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isSVGText()) {
        svgBuilder().attach(toRenderSVGText(parent), WTFMove(child), beforeChild);
        return;
    }
#if ENABLE(MATHML)
    if (parent.isRenderMathMLFenced()) {
        mathMLBuilder().attach(toRenderMathMLFenced(parent), WTFMove(child), beforeChild);
        return;
    }
#endif
    if (parent.isRenderGrid()) {
        attachToRenderGrid(toRenderGrid(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isRenderBlockFlow()) {
        blockFlowBuilder().attach(toRenderBlockFlow(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isRenderBlock()) {
        blockBuilder().attach(toRenderBlock(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isRenderInline()) {
        inlineBuilder().attach(toRenderInline(parent), WTFMove(child), beforeChild);
        return;
    }

    attachToRenderElement(parent, WTFMove(child), beforeChild);
}

void RenderTreeBuilder::attachIgnoringContinuation(RenderElement& parent, RenderPtr<RenderObject> child, RenderObject* beforeChild)
{
    if (parent.isRenderInline()) {
        inlineBuilder().attachIgnoringContinuation(toRenderInline(parent), WTFMove(child), beforeChild);
        return;
    }

    if (parent.isRenderBlock()) {
        blockBuilder().attachIgnoringContinuation(toRenderBlock(parent), WTFMove(child), beforeChild);
        return;
    }

    attach(parent, WTFMove(child), beforeChild);
}

RenderPtr<RenderObject> RenderTreeBuilder::detach(RenderElement& parent, RenderObject& child, CanCollapseAnonymousBlock canCollapseAnonymousBlock)
{
    if (parent.isRubyInline())
        return rubyBuilder().detach(toRenderRubyAsInline(parent), child);

    if (parent.isRubyBlock())
        return rubyBuilder().detach(toRenderRubyAsBlock(parent), child);

    if (parent.isRubyRun())
        return rubyBuilder().detach(toRenderRubyRun(parent), child);

    if (parent.isMenuList())
        return formControlsBuilder().detach(toRenderMenuList(parent), child);

    if (parent.isRenderButton())
        return formControlsBuilder().detach(toRenderButton(parent), child);

    if (parent.isRenderGrid())
        return detachFromRenderGrid(toRenderGrid(parent), child);

    if (parent.isSVGText())
        return svgBuilder().detach(toRenderSVGText(parent), child);

    if (parent.isSVGInline())
        return svgBuilder().detach(toRenderSVGInline(parent), child);

    if (parent.isSVGContainer())
        return svgBuilder().detach(toRenderSVGContainer(parent), child);

    if (parent.isSVGRoot())
        return svgBuilder().detach(toRenderSVGRoot(parent), child);

    if (parent.isRenderBlockFlow())
        return blockBuilder().detach(toRenderBlockFlow(parent), child, canCollapseAnonymousBlock);

    if (parent.isRenderBlock())
        return blockBuilder().detach(toRenderBlock(parent), child, canCollapseAnonymousBlock);

    return detachFromRenderElement(parent, child);
}

void RenderTreeBuilder::attach(RenderTreePosition& position, RenderPtr<RenderObject> child)
{
    attach(position.parent(), WTFMove(child), position.nextSibling());
}

#if ENABLE(FULLSCREEN_API)
void RenderTreeBuilder::createPlaceholderForFullScreen(RenderFullScreen& renderer, PassOwnPtr<RenderStyle> style, const LayoutRect& frameRect)
{
    fullScreenBuilder().createPlaceholder(renderer, style, frameRect);
}
#endif

void RenderTreeBuilder::attachToRenderElement(RenderElement& parent, RenderPtr<RenderObject> child, RenderObject* beforeChild)
{
    if (tableBuilder().childRequiresTable(parent, *child)) {
        RenderTable* table;
        RenderObject* afterChild = beforeChild ? beforeChild->previousSibling() : parent.lastChild();
        if (afterChild && afterChild->isAnonymous() && (*afterChild).isTable() && !afterChild->isBeforeContent())
            table = toRenderTable(afterChild);
        else {
            auto newTable = RenderTable::createAnonymousWithParentRenderer(parent);
            table = newTable.get();
            attach(parent, WTFMove(newTable), beforeChild);
        }

        attach(*table, WTFMove(child));
        return;
    }
    auto& newChild = *child.get();
    attachToRenderElementInternal(parent, WTFMove(child), beforeChild);
    parent.didAttachChild(newChild, beforeChild);
}

void RenderTreeBuilder::attachToRenderElementInternal(RenderElement& parent, RenderPtr<RenderObject> child, RenderObject* beforeChild)
{
    RELEASE_ASSERT_WITH_MESSAGE(!parent.view().frameView().layoutContext().layoutState(), "Layout must not mutate render tree");
    ASSERT(parent.canHaveChildren() || parent.canHaveGeneratedChildren());
    ASSERT(!child->parent());
    ASSERT(!parent.isRenderBlockFlow() || (!child->isTableSection() && !child->isTableRow() && !child->isTableCell()));

    while (beforeChild && beforeChild->parent() && beforeChild->parent() != &parent)
    beforeChild = beforeChild->parent();

    ASSERT(!beforeChild || beforeChild->parent() == &parent);
    ASSERT(!is<RenderText>(beforeChild) || !downcast<RenderText>(*beforeChild).inlineWrapperForDisplayContents());

    // Take the ownership.
    RenderObject* newChild = parent.attachRendererInternal(WTFMove(child), beforeChild);
    if (parent.renderTreeBeingDestroyed()) {
        ASSERT_NOT_REACHED();
        return;
    }

    newChild->insertedIntoTree(m_internalMovesType);
    if (m_internalMovesType == RenderObject::IsInternalMove::No) {
        newChild->initializeFragmentedFlowStateOnInsertion();
		auto* fragmentedFlow = newChild->enclosingFragmentedFlow();
		if (fragmentedFlow && (*fragmentedFlow).isRenderMultiColumnFlow())
			multiColumnBuilder().multiColumnDescendantInserted(toRenderMultiColumnFlow(*fragmentedFlow), *newChild);

		if ((*newChild).isRenderElement())
			RenderCounter::rendererSubtreeAttached(toRenderElement(*newChild));
    }

    newChild->setNeedsLayoutAndPrefWidthsRecalc();
    bool isOutOfFlowBox = newChild->style()->hasOutOfFlowPosition();
    if (!isOutOfFlowBox)
    	parent.setPreferredLogicalWidthsDirty(true);

    if (!parent.normalChildNeedsLayout()) {
        if (isOutOfFlowBox) {
            // setNeedsLayoutAndPrefWidthsRecalc above already takes care of propagating dirty bits on the ancestor chain, but
            // in order to compute static position for out of flow boxes, the parent has to run normal flow layout as well (as opposed to simplified)
            // FIXME: Introduce a dirty bit to bridge the gap between parent and containing block which would
            // not trigger layout but a simple traversal all the way to the direct parent and also expand it non-direct parent cases.
            if (newChild->containingBlock() == &parent)
                parent.setOutOfFlowChildNeedsStaticPositionLayout();
            else
                parent.setChildNeedsLayout();
        } else
            parent.setChildNeedsLayout();
    }

#if HAVE(ACCESSIBILITY)
    if (AXObjectCache* cache = parent.document().axObjectCache())
    	cache->childrenChanged(&parent, newChild);
#endif

    if (parent.isRenderBlockFlow())
    	toRenderBlockFlow(parent).invalidateLineLayoutPath();
    if (parent.hasOutlineAutoAncestor() || parent.outlineStyleForRepaint()->outlineStyleIsAuto())
    	newChild->setHasOutlineAutoAncestor();
}

void RenderTreeBuilder::move(RenderBoxModelObject& from, RenderBoxModelObject& to, RenderObject& child, RenderObject* beforeChild, NormalizeAfterInsertion normalizeAfterInsertion)
{
    // We assume that callers have cleared their positioned objects list for child moves so the
    // positioned renderer maps don't become stale. It would be too slow to do the map lookup on each call.
    ASSERT(normalizeAfterInsertion == NormalizeAfterInsertion::No || !is<RenderBlock>(from) || !downcast<RenderBlock>(from).hasPositionedObjects());

    ASSERT(&from == child.parent());
    ASSERT(!beforeChild || &to == beforeChild->parent());
    if (normalizeAfterInsertion == NormalizeAfterInsertion::Yes && (to.isRenderBlock() || to.isRenderInline())) {
        // Takes care of adding the new child correctly if toBlock and fromBlock
        // have different kind of children (block vs inline).
        auto childToMove = detachFromRenderElement(from, child);
        attach(to, WTFMove(childToMove), beforeChild);
    } else {
    	SetForScope<RenderObject::IsInternalMove> internalMoveScope(m_internalMovesType, RenderObject::IsInternalMove::Yes);
        auto childToMove = detachFromRenderElement(from, child, WillBeDestroyed::No);
        attachToRenderElementInternal(to, WTFMove(childToMove), beforeChild);
    }

    auto findBFCRootAndDestroyInlineTree = [&] {
        auto* containingBlock = &from;
        while (containingBlock) {
            containingBlock->setNeedsLayout();
            if ((*containingBlock).isRenderBlockFlow()) {
                toRenderBlockFlow(*containingBlock).deleteLines();
                break;
            }
            containingBlock = containingBlock->containingBlock();
        }
    };
    // When moving a subtree out of a BFC we need to make sure that the line boxes generated for the inline tree are not accessible anymore from the renderers.
    // Let's find the BFC root and nuke the inline tree (At some point we are going to destroy the subtree instead of moving these renderers around.)
    if (child.isRenderInline())
        findBFCRootAndDestroyInlineTree();
}

void RenderTreeBuilder::move(RenderBoxModelObject& from, RenderBoxModelObject& to, RenderObject& child, NormalizeAfterInsertion normalizeAfterInsertion)
{
    move(from, to, child, nullptr, normalizeAfterInsertion);
}

void RenderTreeBuilder::moveAllChildren(RenderBoxModelObject& from, RenderBoxModelObject& to, NormalizeAfterInsertion normalizeAfterInsertion)
{
    moveAllChildren(from, to, nullptr, normalizeAfterInsertion);
}

void RenderTreeBuilder::moveAllChildren(RenderBoxModelObject& from, RenderBoxModelObject& to, RenderObject* beforeChild, NormalizeAfterInsertion normalizeAfterInsertion)
{
    moveChildren(from, to, from.firstChild(), nullptr, beforeChild, normalizeAfterInsertion);
}

void RenderTreeBuilder::moveChildren(RenderBoxModelObject& from, RenderBoxModelObject& to, RenderObject* startChild, RenderObject* endChild, NormalizeAfterInsertion normalizeAfterInsertion)
{
    moveChildren(from, to, startChild, endChild, nullptr, normalizeAfterInsertion);
}

void RenderTreeBuilder::moveChildren(RenderBoxModelObject& from, RenderBoxModelObject& to, RenderObject* startChild, RenderObject* endChild, RenderObject* beforeChild, NormalizeAfterInsertion normalizeAfterInsertion)
{
    // This condition is rarely hit since this function is usually called on
    // anonymous blocks which can no longer carry positioned objects (see r120761)
    // or when fullRemoveInsert is false.
    if (normalizeAfterInsertion == NormalizeAfterInsertion::Yes && from.isRenderBlock()) {
    	auto& blockFlow = toRenderBlock(from);
        blockFlow.removePositionedObjects(nullptr);
        removeFloatingObjects(blockFlow);
    }

    ASSERT(!beforeChild || &to == beforeChild->parent());
    for (RenderObject* child = startChild; child && child != endChild; ) {
        // Save our next sibling as move will clear it.
        RenderObject* nextSibling = child->nextSibling();

        // FIXME: This logic here fails to detect the first letter in certain cases
        // and skips a valid sibling renderer (see webkit.org/b/163737).
        // Check to make sure we're not saving the firstLetter as the nextSibling.
        // When the |child| object will be moved, its firstLetter will be recreated,
        // so saving it now in nextSibling would leave us with a stale object.
        if (isRenderTextFragment(*child) && nextSibling && (*nextSibling).isText()) {
            RenderObject* firstLetterObj = nullptr;
            if (RenderBlock* block = toRenderTextFragment(*child).blockForAccompanyingFirstLetter()) {
                RenderElement* firstLetterContainer = nullptr;
                block->getFirstLetter(firstLetterObj, firstLetterContainer, child);
            }

            // This is the first letter, skip it.
            if (firstLetterObj == nextSibling)
            	nextSibling = nextSibling->nextSibling();
        }

        move(from, to, *child, beforeChild, normalizeAfterInsertion);
        child = nextSibling;
    }
}

void RenderTreeBuilder::moveAllChildrenIncludingFloats(RenderBlock& from, RenderBlock& to, RenderTreeBuilder::NormalizeAfterInsertion normalizeAfterInsertion)
{
    if (from.isRenderBlockFlow()) {
        blockFlowBuilder().moveAllChildrenIncludingFloats(toRenderBlockFlow(from), to, normalizeAfterInsertion);
        return;
    }
    moveAllChildren(from, to, normalizeAfterInsertion);
}

void RenderTreeBuilder::normalizeTreeAfterStyleChange(RenderElement& renderer, RenderStyle& oldStyle)
{
    if (!renderer.parent())
        return;

    bool wasFloating = oldStyle.isFloating();
    bool wasOutOfFlowPositioned = oldStyle.hasOutOfFlowPosition();
    bool isFloating = renderer.style()->isFloating();
    bool isOutOfFlowPositioned = renderer.style()->hasOutOfFlowPosition();
    bool startsAffectingParent = false;
    bool noLongerAffectsParent = false;
    auto handleFragmentedFlowStateChange = [&] {
        if (!renderer.parent())
            return;

		// Out of flow children of RenderMultiColumnFlow are not really part of the multicolumn flow. We need to ensure that changes in positioning like this
		// trigger insertions into the multicolumn flow.
		auto* enclosingFragmentedFlow = renderer.parent()->enclosingFragmentedFlow();
		if (enclosingFragmentedFlow && (*enclosingFragmentedFlow).isRenderMultiColumnFlow()) {
			auto movingIntoMulticolumn = [&] {
				if (wasOutOfFlowPositioned && !isOutOfFlowPositioned)
					return true;
				auto* containingBlock = renderer.containingBlock();
				if (containingBlock && isOutOfFlowPositioned) {
					// Sometimes the flow state could change even when the renderer stays out-of-flow (e.g when going from fixed to absolute and
					// the containing block is inside a multi-column flow).
					return containingBlock->fragmentedFlowState() == RenderObject::InsideInFragmentedFlow
						&& renderer.fragmentedFlowState() == RenderObject::NotInsideFragmentedFlow;
				}
				return false;
			}();
			if (movingIntoMulticolumn) {
				renderer.initializeFragmentedFlowStateOnInsertion();
				multiColumnBuilder().multiColumnDescendantInserted(toRenderMultiColumnFlow(*enclosingFragmentedFlow), renderer);
				return;
			}
			auto movingOutOfMulticolumn = !wasOutOfFlowPositioned && isOutOfFlowPositioned;
			if (movingOutOfMulticolumn) {
				multiColumnBuilder().restoreColumnSpannersForContainer(renderer, toRenderMultiColumnFlow(*enclosingFragmentedFlow));
				return;
			}

			// Style change may have moved some subtree out of the fragmented flow. Their flow states have already been updated (see adjustFragmentedFlowStateOnContainingBlockChangeIfNeeded)
			// and here is where we take care of the remaining, spanner tree mutation.
			WeakHashSet<RenderElement> spannerContainingBlockSet;
			for (auto& descendant : descendantsOfType<RenderMultiColumnSpannerPlaceholder>(renderer)) {
				auto* containingBlock = descendant.containingBlock();
				if (containingBlock && containingBlock->enclosingFragmentedFlow() != enclosingFragmentedFlow)
					spannerContainingBlockSet.add(*containingBlock);
			}
			WeakPtr<RenderFragmentedFlow> oldEnclosingFragmentedFlow = makeWeakPtr(*enclosingFragmentedFlow);
			for (auto& containingBlock : spannerContainingBlockSet) {
				if (!oldEnclosingFragmentedFlow)
					break;
				multiColumnBuilder().restoreColumnSpannersForContainer(containingBlock, toRenderMultiColumnFlow(*oldEnclosingFragmentedFlow));
			}
		}
    };

    auto& parent = *renderer.parent();
    if (parent.isRenderBlock())
        noLongerAffectsParent = (!wasFloating && isFloating) || (!wasOutOfFlowPositioned && isOutOfFlowPositioned);

    if (parent.isRenderBlockFlow() || parent.isRenderInline()) {
        startsAffectingParent = (wasFloating || wasOutOfFlowPositioned) && !isFloating && !isOutOfFlowPositioned;
        ASSERT(!startsAffectingParent || !noLongerAffectsParent);
    }

    if (startsAffectingParent) {
        // We have gone from not affecting the inline status of the parent flow to suddenly
        // having an impact. See if there is a mismatch between the parent flow's
        // childrenInline() state and our state.
        if (renderer.isInline() != renderer.parent()->childrenInline())
            childFlowStateChangesAndAffectsParentBlock(renderer);
        // WARNING: original parent might be deleted at this point.
        handleFragmentedFlowStateChange();
        return;
    }

    if (noLongerAffectsParent) {
        childFlowStateChangesAndNoLongerAffectsParentBlock(renderer);

        if (isFloating && renderer.isRenderBlockFlow()) {
            auto clearDescendantFloats = [&] {
                // These descendent floats can not intrude other, sibling block containers anymore.
                for (auto& descendant : descendantsOfType<RenderBox>(renderer)) {
                    if (descendant.isFloatingOrOutOfFlowPositioned())
                        descendant.removeFloatingOrPositionedChildFromBlockLists();
                }
            };
            clearDescendantFloats();
            removeFloatingObjects(toRenderBlock(renderer));
            // Fresh floats need to be reparented if they actually belong to the previous anonymous block.
            // It copies the logic of RenderBlock::addChildIgnoringContinuation
            if (renderer.previousSibling() && renderer.previousSibling()->isAnonymousBlock())
                move(toRenderBoxModelObject(parent), toRenderBoxModelObject(*renderer.previousSibling()), renderer, RenderTreeBuilder::NormalizeAfterInsertion::No);
        }
    }

    handleFragmentedFlowStateChange();
}

void RenderTreeBuilder::makeChildrenNonInline(RenderBlock& parent, RenderObject* insertionPoint)
{
    // makeChildrenNonInline takes a block whose children are *all* inline and it
    // makes sure that inline children are coalesced under anonymous
    // blocks. If |insertionPoint| is defined, then it represents the insertion point for
    // the new block child that is causing us to have to wrap all the inlines. This
    // means that we cannot coalesce inlines before |insertionPoint| with inlines following
    // |insertionPoint|, because the new child is going to be inserted in between the inlines,
    // splitting them.
    ASSERT(parent.isInlineBlockOrInlineTable() || !parent.isInline());
    ASSERT(!insertionPoint || insertionPoint->parent() == &parent);

    parent.setChildrenInline(false);

    auto* child = parent.firstChild();
    if (!child)
        return;

    parent.deleteLines();

    while (child) {
        RenderObject* inlineRunStart = nullptr;
        RenderObject* inlineRunEnd = nullptr;
        getInlineRun(child, insertionPoint, inlineRunStart, inlineRunEnd);

        if (!inlineRunStart)
            break;

        child = inlineRunEnd->nextSibling();

        auto newBlock = parent.createAnonymousBlock();
        auto& block = *newBlock;
        attachToRenderElementInternal(parent, WTFMove(newBlock), inlineRunStart);
        moveChildren(parent, block, inlineRunStart, child, RenderTreeBuilder::NormalizeAfterInsertion::No);
	}
#ifndef NDEBUG
    for (RenderObject* c = parent.firstChild(); c; c = c->nextSibling())
        ASSERT(!c->isInline());
#endif
    parent.repaint();
}

RenderObject* RenderTreeBuilder::splitAnonymousBoxesAroundChild(RenderBox& parent, RenderObject& originalBeforeChild)
{
    // Adjust beforeChild if it is a column spanner and has been moved out of its original position.
    auto* beforeChild = RenderTreeBuilder::MultiColumn::adjustBeforeChildForMultiColumnSpannerIfNeeded(originalBeforeChild);
    bool didSplitParentAnonymousBoxes = false;

    while (beforeChild->parent() != &parent) {
        auto& boxToSplit = toRenderBox(*beforeChild->parent());
        if (boxToSplit.firstChild() != beforeChild && boxToSplit.isAnonymous()) {
            didSplitParentAnonymousBoxes = true;

            // We have to split the parent box into two boxes and move children
            // from |beforeChild| to end into the new post box.
            auto newPostBox = boxToSplit.createAnonymousBoxWithSameTypeAs(parent);
            auto& postBox = *newPostBox;
            postBox.setChildrenInline(boxToSplit.childrenInline());
            RenderBox* parentBox = toRenderBox(boxToSplit.parent());
            // We need to invalidate the |parentBox| before inserting the new node
            // so that the table repainting logic knows the structure is dirty.
            // See for example RenderTableCell:clippedOverflowRectForRepaint.
            markBoxForRelayoutAfterSplit(*parentBox);
            attachToRenderElementInternal(*parentBox, WTFMove(newPostBox), boxToSplit.nextSibling());
            moveChildren(boxToSplit, postBox, beforeChild, nullptr, RenderTreeBuilder::NormalizeAfterInsertion::Yes);

            markBoxForRelayoutAfterSplit(boxToSplit);
            markBoxForRelayoutAfterSplit(postBox);

            beforeChild = &postBox;
        } else
            beforeChild = &boxToSplit;
    }

    if (didSplitParentAnonymousBoxes)
        markBoxForRelayoutAfterSplit(parent);

    ASSERT(beforeChild->parent() == &parent);
    return beforeChild;
}

void RenderTreeBuilder::childFlowStateChangesAndAffectsParentBlock(RenderElement& child)
{
	if (!child.isInline()) {
		auto parent = makeWeakPtr(child.parent());
		if ((*parent).isRenderBlock())
			blockBuilder().childBecameNonInline(toRenderBlock(*parent), child);
		else if ((*parent).isRenderInline())
			inlineBuilder().childBecameNonInline(toRenderInline(*parent), child);
		// WARNING: original parent might be deleted at this point.
		auto* newParent = child.parent();
		if (newParent != parent && newParent && newParent->isRenderGrid()) {
			// We need to re-run the grid items placement if it had gained a new item.
			toRenderGrid(*newParent).dirtyGrid();
		}
        return;
    }
	// An anonymous block must be made to wrap this inline.
	auto* parent = child.parent();
	auto newBlock = toRenderBlock(*parent).createAnonymousBlock();
	auto& block = *newBlock;
	attachToRenderElementInternal(*parent, WTFMove(newBlock), &child);
	auto thisToMove = detachFromRenderElement(*parent, child);
	attachToRenderElementInternal(block, WTFMove(thisToMove));
}

void RenderTreeBuilder::removeAnonymousWrappersForInlineChildrenIfNeeded(RenderElement& parent)
{
    if (!parent.isRenderBlock())
        return;
    auto& blockParent = toRenderBlock(parent);
    if (!blockParent.canDropAnonymousBlockChild())
        return;

    // We have changed to floated or out-of-flow positioning so maybe all our parent's
    // children can be inline now. Bail if there are any block children left on the line,
    // otherwise we can proceed to stripping solitary anonymous wrappers from the inlines.
    // FIXME: We should also handle split inlines here - we exclude them at the moment by returning
    // if we find a continuation.
    Optional<bool> shouldAllChildrenBeInline;
    for (auto* current = blockParent.firstChild(); current; current = current->nextSibling()) {
        if (current->style()->isFloating() || current->style()->hasOutOfFlowPosition())
            continue;
        if (!current->isAnonymousBlock() || toRenderBlock(*current).isContinuation())
            return;
        // Anonymous block not in continuation. Check if it holds a set of inline or block children and try not to mix them.
        auto* firstChild = current->firstChildSlow();
        if (!firstChild)
            continue;
        bool isInlineLevelBox = firstChild->isInline();
        if (!shouldAllChildrenBeInline.hasValue()) {
            shouldAllChildrenBeInline = isInlineLevelBox;
            continue;
        }
        // Mixing inline and block level boxes?
        if (*shouldAllChildrenBeInline != isInlineLevelBox)
            return;
    }

    RenderObject* next = nullptr;
    for (auto* current = blockParent.firstChild(); current; current = next) {
        next = current->nextSibling();
        if (current->isAnonymousBlock())
        	blockBuilder().dropAnonymousBoxChild(blockParent, toRenderBlock(*current));
	}
}

void RenderTreeBuilder::childFlowStateChangesAndNoLongerAffectsParentBlock(RenderElement& child)
{
    ASSERT(child.parent());
    removeAnonymousWrappersForInlineChildrenIfNeeded(*child.parent());
}

void RenderTreeBuilder::destroyAndCleanUpAnonymousWrappers(RenderObject& rendererToDestroy)
{
    // If the tree is destroyed, there is no need for a clean-up phase.
    if (rendererToDestroy.renderTreeBeingDestroyed()) {
        destroy(rendererToDestroy);
        return;
    }

    // Also destroy ::backdrop along with element if there is one
    if (rendererToDestroy.isRenderElement()) {
        if (auto backdropRenderer = toRenderElement(rendererToDestroy).backdropRenderer())
            destroy(*backdropRenderer);
    }

    auto isAnonymousAndSafeToDelete = [] (const RenderElement& renderer) {
    	return renderer.isAnonymous() && !isRenderRubyBase(renderer) && !renderer.isRenderView() && !renderer.isRenderFragmentedFlow();
	};

    auto destroyRootIncludingAnonymous = [&] () -> RenderObject& {
        auto* destroyRoot = &rendererToDestroy;
        while (!(*destroyRoot).isRenderView()) {
            auto& destroyRootParent = *destroyRoot->parent();
            if (!isAnonymousAndSafeToDelete(destroyRootParent))
                break;
            bool destroyingOnlyChild = destroyRootParent.firstChild() == destroyRoot && destroyRootParent.lastChild() == destroyRoot;
            if (!destroyingOnlyChild)
                break;
            destroyRoot = &destroyRootParent;
        }
        return *destroyRoot;
    };

    WeakPtr<RenderObject> destroyRoot = makeWeakPtr(destroyRootIncludingAnonymous());

    auto clearFloatsAndOutOfFlowPositionedObjects = [&] {
        // Remove floats and out-of-flow positioned objects from their containing block before detaching
        // the renderer from the tree. It includes all the anonymous block descendants that we are about
        // to destroy as well as part of the cleanup process below.
        if (!destroyRoot->isRenderElement())
            return;
        for (auto& descendant : descendantsOfType<RenderBox>(toRenderElement(*destroyRoot))) {
            if (descendant.isFloatingOrOutOfFlowPositioned())
                descendant.removeFloatingOrPositionedChildFromBlockLists();
        }
        if (destroyRoot->isBox() && destroyRoot->isFloatingOrOutOfFlowPositioned())
            toRenderBox(*destroyRoot).removeFloatingOrPositionedChildFromBlockLists();
    };
    clearFloatsAndOutOfFlowPositionedObjects();

    auto collapseAndDestroyAnonymousSiblings = [&] {
        // FIXME: Probably need to handle other table parts here as well.
        if (isRenderTableCell(*destroyRoot)) {
            tableBuilder().collapseAndDestroyAnonymousSiblingCells(toRenderTableCell(*destroyRoot));
            return;
        }

        if (isRenderTableRow(*destroyRoot)) {
            tableBuilder().collapseAndDestroyAnonymousSiblingRows(toRenderTableRow(*destroyRoot));
            return;
        }
    };
    collapseAndDestroyAnonymousSiblings();

    // FIXME: Do not try to collapse/cleanup the anonymous wrappers inside destroy (see webkit.org/b/186746).
    auto destroyRootParent = makeWeakPtr(*destroyRoot->parent());
    if (&rendererToDestroy != destroyRoot.get()) {
        // Destroy the child renderer first, before we start tearing down the anonymous wrapper ancestor chain.
    	auto anonymousDestroyRoot = SetForScope<WeakPtr<RenderObject>>(m_anonymousDestroyRoot, destroyRoot);
		destroy(rendererToDestroy);
    }

    if (destroyRoot) {
        auto anonymousDestroyRoot = SetForScope<WeakPtr<RenderObject>>(m_anonymousDestroyRoot, destroyRoot.get() != &rendererToDestroy ? destroyRoot : nullptr);
        destroy(*destroyRoot);
    }

    if (!destroyRootParent)
        return;

    auto anonymousDestroyRoot = SetForScope<WeakPtr<RenderObject>>(m_anonymousDestroyRoot, destroyRootParent);
	removeAnonymousWrappersForInlineChildrenIfNeeded(*destroyRootParent);

    // Anonymous parent might have become empty, try to delete it too.
    if (isAnonymousAndSafeToDelete(*destroyRootParent) && !destroyRootParent->firstChild())
        destroyAndCleanUpAnonymousWrappers(*destroyRootParent);
    // WARNING: rendererToDestroy is deleted here.
}

void RenderTreeBuilder::updateAfterDescendants(RenderElement& renderer)
{
    if (renderer.isRenderBlock())
        firstLetterBuilder().updateAfterDescendants(toRenderBlock(renderer));
    if (renderer.isListItem())
		listBuilder().updateItemMarker(toRenderListItem(renderer));
	if (renderer.isRenderBlockFlow())
		multiColumnBuilder().updateAfterDescendants(toRenderBlockFlow(renderer));
}

RenderPtr<RenderObject> RenderTreeBuilder::detachFromRenderGrid(RenderGrid& parent, RenderObject& child)
{
    auto takenChild = blockBuilder().detach(parent, child);
    // Positioned grid items do not take up space or otherwise participate in the layout of the grid,
    // for that reason we don't need to mark the grid as dirty when they are removed.
    if (child.isOutOfFlowPositioned())
        return takenChild;

    // The grid needs to be recomputed as it might contain auto-placed items that will change their position.
    parent.dirtyGrid();
    return takenChild;
}

RenderPtr<RenderObject> RenderTreeBuilder::detachFromRenderElement(RenderElement& parent, RenderObject& child, WillBeDestroyed willBeDestroyed)
{
    RELEASE_ASSERT_WITH_MESSAGE(!parent.view().frameView().layoutContext().layoutState(), "Layout must not mutate render tree");

    ASSERT(parent.canHaveChildren() || parent.canHaveGeneratedChildren());
    ASSERT(child.parent() == &parent);

    if (child.isFloatingOrOutOfFlowPositioned())
        toRenderBox(child).removeFloatingOrPositionedChildFromBlockLists();

    // So that we'll get the appropriate dirty bit set (either that a normal flow child got yanked or
    // that a positioned child got yanked). We also repaint, so that the area exposed when the child
    // disappears gets repainted properly.
    if (!parent.renderTreeBeingDestroyed() && child.everHadLayout()) {
        if (child.isBody())
            parent.view().repaintRootContents();
        else
            child.repaint();
        child.setNeedsLayoutAndPrefWidthsRecalc();
    }

    // If we have a line box wrapper, delete it.
    if (child.isBox())
        toRenderBox(child).deleteLineBoxWrapper();
    else if (child.isLineBreak())
        toRenderLineBreak(child).deleteInlineBoxWrapper();
    else if (child.isText())
        toRenderText(child).removeAndDestroyTextBoxes();

    if (!parent.renderTreeBeingDestroyed() && parent.isFlexibleBox() && !child.isFloatingOrOutOfFlowPositioned() && child.isBox()) {
        toRenderFlexibleBox(parent).clearCachedChildIntrinsicContentLogicalHeight(toRenderBox(child));
        toRenderFlexibleBox(parent).clearCachedMainSizeForChild(toRenderBox(child));
    }

    // If child is the start or end of the selection, then clear the selection to
    // avoid problems of invalid pointers.
    if (!parent.renderTreeBeingDestroyed() && willBeDestroyed == WillBeDestroyed::Yes && child.isSelectionBorder())
		parent.frame().selection().setNeedsSelectionUpdate();

    if (!parent.renderTreeBeingDestroyed() && m_internalMovesType == RenderObject::IsInternalMove::No)
        child.resetFragmentedFlowStateOnRemoval();

    if (!parent.renderTreeBeingDestroyed())
        child.willBeRemovedFromTree(m_internalMovesType);

    // WARNING: There should be no code running between willBeRemovedFromTree() and the actual removal below.
    // This is needed to avoid race conditions where willBeRemovedFromTree() would dirty the tree's structure
    // and the code running here would force an untimely rebuilding, leaving |child| dangling.
    auto childToTake = parent.detachRendererInternal(child);

    // rendererRemovedFromTree() walks the whole subtree. We can improve performance
    // by skipping this step when destroying the entire tree.
    if (!parent.renderTreeBeingDestroyed() && (*childToTake).isRenderElement())
        RenderCounter::rendererRemovedFromTree(toRenderElement(*childToTake));

    if (!parent.renderTreeBeingDestroyed()) {
#if HAVE(ACCESSIBILITY)
        if (AXObjectCache* cache = parent.document().existingAXObjectCache())
            cache->childrenChanged(&parent);
#endif
    }

    return childToTake;
}

void RenderTreeBuilder::attachToRenderGrid(RenderGrid& parent, RenderPtr<RenderObject> child, RenderObject* beforeChild)
{
    auto& newChild = *child;
    blockBuilder().attach(parent, WTFMove(child), beforeChild);

    // Positioned grid items do not take up space or otherwise participate in the layout of the grid,
    // for that reason we don't need to mark the grid as dirty when they are added.
    if (newChild.isOutOfFlowPositioned())
        return;

    // The grid needs to be recomputed as it might contain auto-placed items that
    // will change their position.
    parent.dirtyGrid();
}

void RenderTreeBuilder::markBoxForRelayoutAfterSplit(RenderBox& box)
{
    // FIXME: The table code should handle that automatically. If not,
    // we should fix it and remove the table part checks.
    if (box.isTable()) {
        // Because we may have added some sections with already computed column structures, we need to
        // sync the table structure with them now. This avoids crashes when adding new cells to the table.
        toRenderTable(box).forceSectionsRecalc();
    } else if (box.isTableSection())
        toRenderTableSection(box).setNeedsCellRecalc();

    box.setNeedsLayoutAndPrefWidthsRecalc();
}

void RenderTreeBuilder::removeFloatingObjects(RenderBlock& renderer)
{
	if (renderer.renderTreeBeingDestroyed())
		return;
	if (!renderer.isRenderBlockFlow())
		return;
	auto* floatingObjects = toRenderBlockFlow(renderer).floatingObjectSet();
	if (!floatingObjects)
		return;
	// Here we remove the floating objects from the descendants as well.
	Vector<FloatingObject*> copyOfFloatingObjects(floatingObjects->size());
	for (auto& floatingObject : *floatingObjects)
		copyOfFloatingObjects.uncheckedAppend(floatingObject.get());
	for (auto* floatingObject : copyOfFloatingObjects)
		floatingObject->renderer().removeFloatingOrPositionedChildFromBlockLists();
};

}
