/*
 * Copyright (C) 2016 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#pragma once

#include "ExceptionCode.h"
#include "SVGAngleValue.h"
#include "SVGValueProperty.h"

namespace WebCore {

class SVGAngle : public SVGValueProperty<SVGAngleValue> {
    using Base = SVGValueProperty<SVGAngleValue>;
    using Base::Base;
    using Base::m_value;

public:
    static Ref<SVGAngle> create(const SVGAngleValue& value = SVGAngleValue())
    {
        return adoptRef(*new SVGAngle(value));
    }

    static Ref<SVGAngle> create(SVGPropertyOwner* owner, SVGPropertyAccess access, const SVGAngleValue& value = SVGAngleValue())
    {
        return adoptRef(*new SVGAngle(owner, access, value));
    }

#if 0
    template<typename T>
    static ExceptionOr<Ref<SVGAngle>> create(ExceptionOr<T>&& value)
    {
        if (value.hasException())
            return value.releaseException();
        return adoptRef(*new SVGAngle(value.releaseReturnValue()));
    }
#endif

    SVGAngleValue::Type unitType() const
    {
        return m_value.unitType();
    }

    void setValueForBindings(float value, ExceptionCode& ec)
    {
        if (isReadOnly()) {
            ec = NO_MODIFICATION_ALLOWED_ERR;
            return;
        }

        m_value.setValue(value);
        commitChange();
    }
    
    float valueForBindings() const
    {
        return m_value.value();
    }

    void setValueInSpecifiedUnits(float valueInSpecifiedUnits, ExceptionCode& ec)
    {
        if (isReadOnly()) {
            ec = NO_MODIFICATION_ALLOWED_ERR;
            return;
        }

        m_value.setValueInSpecifiedUnits(valueInSpecifiedUnits);
        commitChange();
    }
    
    float valueInSpecifiedUnits() const
    {
        return m_value.valueInSpecifiedUnits();
    }

    void setValueAsString(const String& value, ExceptionCode& ec)
    {
        if (isReadOnly()) {
            ec = NO_MODIFICATION_ALLOWED_ERR;
            return;
        }

        m_value.setValueAsString(value, ec);
        if (ec)
            return;
        
        commitChange();
    }

    String valueAsString() const override
    {
        return m_value.valueAsString();
    }

    void newValueSpecifiedUnits(unsigned short unitType, float valueInSpecifiedUnits, ExceptionCode& ec)
    {
        if (isReadOnly()) {
            ec = NO_MODIFICATION_ALLOWED_ERR;
            return;
        }

        m_value.newValueSpecifiedUnits(unitType, valueInSpecifiedUnits, ec);
        if (ec)
            return;
        
        commitChange();
    }
    
    void convertToSpecifiedUnits(unsigned short unitType, ExceptionCode& ec)
    {
        if (isReadOnly()) {
            ec = NO_MODIFICATION_ALLOWED_ERR;
            return;
        }

        m_value.convertToSpecifiedUnits(unitType, ec);
        if (ec)
            return;
        
        commitChange();
    }
};

} // namespace WebCore
