/*
 * Copyright (C) 2004, 2005, 2007 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2004, 2005, 2006 Rob Buis <buis@kde.org>
 * Copyright (C) 2005 Oliver Hunt <oliver@nerget.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"

#if ENABLE(SVG) && ENABLE(FILTERS)
#include "SVGFELightElement.h"

#include "Attribute.h"
#include "ElementIterator.h"
#include "RenderObject.h"
#include "RenderSVGResource.h"
#include "SVGFEDiffuseLightingElement.h"
#include "SVGFESpecularLightingElement.h"
#include "SVGFilterElement.h"
#include "SVGFilterPrimitiveStandardAttributes.h"
#include "SVGNames.h"

namespace WebCore {

SVGFELightElement::SVGFELightElement(const QualifiedName& tagName, Document* document)
    : SVGElement(tagName, document, adoptPtr(new PropertyRegistry(*this)))
{
    static std::once_flag onceFlag;
    std::call_once(onceFlag, [] {
        PropertyRegistry::registerProperty<SVGNames::azimuthAttr, &SVGFELightElement::m_azimuth>();
        PropertyRegistry::registerProperty<SVGNames::elevationAttr, &SVGFELightElement::m_elevation>();
        PropertyRegistry::registerProperty<SVGNames::xAttr, &SVGFELightElement::m_x>();
        PropertyRegistry::registerProperty<SVGNames::yAttr, &SVGFELightElement::m_y>();
        PropertyRegistry::registerProperty<SVGNames::zAttr, &SVGFELightElement::m_z>();
        PropertyRegistry::registerProperty<SVGNames::pointsAtXAttr, &SVGFELightElement::m_pointsAtX>();
        PropertyRegistry::registerProperty<SVGNames::pointsAtYAttr, &SVGFELightElement::m_pointsAtY>();
        PropertyRegistry::registerProperty<SVGNames::pointsAtZAttr, &SVGFELightElement::m_pointsAtZ>();
        PropertyRegistry::registerProperty<SVGNames::specularExponentAttr, &SVGFELightElement::m_specularExponent>();
        PropertyRegistry::registerProperty<SVGNames::limitingConeAngleAttr, &SVGFELightElement::m_limitingConeAngle>();
    });
}

SVGFELightElement* SVGFELightElement::findLightElement(const SVGElement* svgElement)
{
    for (auto& child : childrenOfType<SVGElement>(*svgElement)) {
        if (child.hasTagName(SVGNames::feDistantLightTag) || child.hasTagName(SVGNames::fePointLightTag) || child.hasTagName(SVGNames::feSpotLightTag)) {
        	return static_cast<SVGFELightElement*>(const_cast<SVGElement*>(&child));
        }
    }
    return nullptr;
}

void SVGFELightElement::attributeChanged(const QualifiedName& name, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason attributeModificationReason)
{
    if (name == SVGNames::azimuthAttr)
        m_azimuth->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::elevationAttr)
        m_elevation->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::xAttr)
        m_x->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::yAttr)
        m_y->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::zAttr)
        m_z->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::pointsAtXAttr)
        m_pointsAtX->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::pointsAtYAttr)
        m_pointsAtY->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::pointsAtZAttr)
        m_pointsAtZ->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::specularExponentAttr)
        m_specularExponent->setBaseValInternal(newValue.toFloat());
    else if (name == SVGNames::limitingConeAngleAttr)
        m_limitingConeAngle->setBaseValInternal(newValue.toFloat());

    SVGElement::attributeChanged(name, oldValue, newValue, attributeModificationReason);
}

void SVGFELightElement::svgAttributeChanged(const QualifiedName& attrName)
{
    if (PropertyRegistry::isKnownAttribute(attrName)) {
    	ASSERT(attrName == SVGNames::azimuthAttr || attrName == SVGNames::elevationAttr || attrName == SVGNames::xAttr || attrName == SVGNames::yAttr
			|| attrName == SVGNames::zAttr || attrName == SVGNames::pointsAtXAttr || attrName == SVGNames::pointsAtYAttr || attrName == SVGNames::pointsAtZAttr
			|| attrName == SVGNames::specularExponentAttr || attrName == SVGNames::limitingConeAngleAttr);
		RefPtr<Element> parent = parentElement();
        if (!parent)
            return;

        auto* renderer = parent->renderer();
        if (!renderer || !renderer->isSVGResourceFilterPrimitive())
            return;

        if (parent->hasTagName(SVGNames::feDiffuseLightingTag)) {
            InstanceInvalidationGuard guard(*this);
            static_cast<SVGFEDiffuseLightingElement&>(*parent).lightElementAttributeChanged(this, attrName);
        } else if (parent->hasTagName(SVGNames::feSpecularLightingTag)) {
            InstanceInvalidationGuard guard(*this);
            static_cast<SVGFESpecularLightingElement&>(*parent).lightElementAttributeChanged(this, attrName);
        }

        return;
    }

    SVGElement::svgAttributeChanged(attrName);
}

void SVGFELightElement::childrenChanged(const ChildChange& change)
{
    SVGElement::childrenChanged(change);

    if (change.source == ChildChangeSource::Parser)
        return;

    SVGFilterPrimitiveStandardAttributes::invalidateFilterPrimitiveParent(this);
}

}

#endif // ENABLE(SVG)
