/*
 * Copyright (C) 2004, 2005, 2006, 2007 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2004, 2005, 2006 Rob Buis <buis@kde.org>
 * Copyright (C) 2006 Samuel Weinig <sam.weinig@gmail.com>
 * Copyright (C) 2009 Dirk Schulze <krit@webkit.org>
 * Copyright (C) Research In Motion Limited 2010. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"

#if ENABLE(SVG) && ENABLE(FILTERS)
#include "SVGFilterElement.h"

#include "Attr.h"
#include "RenderSVGResourceFilter.h"
#include "SVGFilterBuilder.h"
#include "SVGFilterPrimitiveStandardAttributes.h"
#include "SVGNames.h"
#include "SVGParserUtilities.h"
#include "XLinkNames.h"

namespace WebCore {

inline SVGFilterElement::SVGFilterElement(const QualifiedName& tagName, Document* document)
    : SVGElement(tagName, document, adoptPtr(new PropertyRegistry(*this)))
	, SVGURIReference(this)
{
	// Spec: If the x/y attribute is not specified, the effect is as if a value of "-10%" were specified.
	// Spec: If the width/height attribute is not specified, the effect is as if a value of "120%" were specified.
	ASSERT(hasTagName(SVGNames::filterTag));

    static std::once_flag onceFlag;
    std::call_once(onceFlag, [] {
        PropertyRegistry::registerProperty<SVGNames::filterUnitsAttr, SVGUnitTypes::SVGUnitType, &SVGFilterElement::m_filterUnits>();
        PropertyRegistry::registerProperty<SVGNames::primitiveUnitsAttr, SVGUnitTypes::SVGUnitType, &SVGFilterElement::m_primitiveUnits>();
        PropertyRegistry::registerProperty<SVGNames::xAttr, &SVGFilterElement::m_x>();
        PropertyRegistry::registerProperty<SVGNames::yAttr, &SVGFilterElement::m_y>();
        PropertyRegistry::registerProperty<SVGNames::widthAttr, &SVGFilterElement::m_width>();
        PropertyRegistry::registerProperty<SVGNames::heightAttr, &SVGFilterElement::m_height>();
    });
}

PassRefPtr<SVGFilterElement> SVGFilterElement::create(const QualifiedName& tagName, Document* document)
{
    return adoptRef(new SVGFilterElement(tagName, document));
}

void SVGFilterElement::attributeChanged(const QualifiedName& name, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason attributeModificationReason)
{
    SVGParsingError parseError = NoError;

    if (name == SVGNames::filterUnitsAttr) {
        SVGUnitTypes::SVGUnitType propertyValue = SVGPropertyTraits<SVGUnitTypes::SVGUnitType>::fromString(newValue);
        if (propertyValue > 0)
            m_filterUnits->setBaseValInternal<SVGUnitTypes::SVGUnitType>(propertyValue);
    } else if (name == SVGNames::primitiveUnitsAttr) {
        SVGUnitTypes::SVGUnitType propertyValue = SVGPropertyTraits<SVGUnitTypes::SVGUnitType>::fromString(newValue);
        if (propertyValue > 0)
            m_primitiveUnits->setBaseValInternal<SVGUnitTypes::SVGUnitType>(propertyValue);
    } else if (name == SVGNames::xAttr)
    	m_x->setBaseValInternal(SVGLengthValue::construct(SVGLengthMode::Width, newValue, parseError));
    else if (name == SVGNames::yAttr)
    	m_y->setBaseValInternal(SVGLengthValue::construct(SVGLengthMode::Height, newValue, parseError));
    else if (name == SVGNames::widthAttr)
    	m_width->setBaseValInternal(SVGLengthValue::construct(SVGLengthMode::Width, newValue, parseError));
    else if (name == SVGNames::heightAttr)
    	m_height->setBaseValInternal(SVGLengthValue::construct(SVGLengthMode::Height, newValue, parseError));

    reportAttributeParsingError(parseError, name, newValue);

    SVGURIReference::parseAttribute(name, newValue);
    SVGElement::attributeChanged(name, oldValue, newValue, attributeModificationReason);
}

void SVGFilterElement::svgAttributeChanged(const QualifiedName& attrName)
{
    if (PropertyRegistry::isAnimatedLengthAttribute(attrName)) {
        InstanceInvalidationGuard guard(*this);
        setPresentationalHintStyleIsDirty();
        return;
    }

    if (PropertyRegistry::isKnownAttribute(attrName) || SVGURIReference::isKnownAttribute(attrName)) {
    	updateSVGRendererForElementChange();
		return;
	}

    SVGElement::svgAttributeChanged(attrName);
}

void SVGFilterElement::childrenChanged(const ChildChange& change)
{
    SVGElement::childrenChanged(change);

    if (change.source == ChildChangeSource::Parser)
        return;

    updateSVGRendererForElementChange();
}

RenderPtr<RenderElement> SVGFilterElement::createElementRenderer(PassOwnPtr<RenderStyle> style, const RenderTreePosition&)
{
    return createRenderer<RenderSVGResourceFilter>(*this, style);
}

bool SVGFilterElement::childShouldCreateRenderer(const Node* child) const
{
    if (!child->isSVGElement())
        return false;

    const SVGElement* element = toSVGElement(child);

    static NeverDestroyed<HashSet<QualifiedName>> allowedChildElementTags;
    if (allowedChildElementTags->isEmpty()) {
        allowedChildElementTags->add(SVGNames::feBlendTag);
        allowedChildElementTags->add(SVGNames::feColorMatrixTag);
        allowedChildElementTags->add(SVGNames::feComponentTransferTag);
        allowedChildElementTags->add(SVGNames::feCompositeTag);
        allowedChildElementTags->add(SVGNames::feConvolveMatrixTag);
        allowedChildElementTags->add(SVGNames::feDiffuseLightingTag);
        allowedChildElementTags->add(SVGNames::feDisplacementMapTag);
        allowedChildElementTags->add(SVGNames::feDistantLightTag);
        allowedChildElementTags->add(SVGNames::feDropShadowTag);
        allowedChildElementTags->add(SVGNames::feFloodTag);
        allowedChildElementTags->add(SVGNames::feFuncATag);
        allowedChildElementTags->add(SVGNames::feFuncBTag);
        allowedChildElementTags->add(SVGNames::feFuncGTag);
        allowedChildElementTags->add(SVGNames::feFuncRTag);
        allowedChildElementTags->add(SVGNames::feGaussianBlurTag);
        allowedChildElementTags->add(SVGNames::feImageTag);
        allowedChildElementTags->add(SVGNames::feMergeTag);
        allowedChildElementTags->add(SVGNames::feMergeNodeTag);
        allowedChildElementTags->add(SVGNames::feMorphologyTag);
        allowedChildElementTags->add(SVGNames::feOffsetTag);
        allowedChildElementTags->add(SVGNames::fePointLightTag);
        allowedChildElementTags->add(SVGNames::feSpecularLightingTag);
        allowedChildElementTags->add(SVGNames::feSpotLightTag);
        allowedChildElementTags->add(SVGNames::feTileTag);
        allowedChildElementTags->add(SVGNames::feTurbulenceTag);
    }

    return allowedChildElementTags->contains<SVGAttributeHashTranslator>(element->tagQName());
}

}

#endif
