/*
 * Copyright (C) 2004, 2005, 2006, 2008 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2004, 2005, 2006, 2007 Rob Buis <buis@kde.org>
 * Copyright (C) 2018 Apple Inc. All rights reserved.
 * Copyright (C) 2018 Adobe Systems Incorporated. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "SVGGeometryElement.h"

#include "DOMPoint.h"
#include "RenderSVGResource.h"
#include "LegacyRenderSVGShape.h"
#include "SVGDocumentExtensions.h"
#include "SVGPathUtilities.h"
#include "SVGPoint.h"
//#include <wtf/IsoMallocInlines.h>
#include <wtf/HashSet.h>

namespace WebCore {

//WTF_MAKE_ISO_ALLOCATED_IMPL(SVGGeometryElement);

SVGGeometryElement::SVGGeometryElement(const QualifiedName& tagName, Document* document, PassOwnPtr<SVGPropertyRegistry> propertyRegistry)
    : SVGGraphicsElement(tagName, document, propertyRegistry)
{
    static std::once_flag onceFlag;
    std::call_once(onceFlag, [] {
        PropertyRegistry::registerProperty<SVGNames::pathLengthAttr, &SVGGeometryElement::m_pathLength>();
    });
}

float SVGGeometryElement::getTotalLength() const
{
    document().updateLayoutIgnorePendingStylesheets();

    auto* renderer = toRenderSVGShape(this->renderer());
    if (!renderer)
        return 0;

    return renderer->getTotalLength();
}

PassRefPtr<SVGPoint> SVGGeometryElement::getPointAtLength(float distance, ExceptionCode& ec) const
{
    document().updateLayoutIgnorePendingStylesheets();

    auto* renderer = toRenderSVGShape(this->renderer());

    // Spec: If current element is a non-rendered element, throw an InvalidStateError.
    if (!renderer) {
    	ec = INVALID_STATE_ERR;
    	return nullptr;
    }

    // Spec: Clamp distance to [0, length].
    distance = clampTo<float>(distance, 0, getTotalLength());

    // Spec: Return a newly created, detached SVGPoint object.
    return SVGPoint::create(renderer->getPointAtLength(distance));
}

bool SVGGeometryElement::isPointInFill(const DOMPointInit& pointInit)
{
    document().updateLayoutIgnorePendingStylesheets();

    auto* renderer = toRenderSVGShape(this->renderer());
    if (!renderer)
        return false;

    FloatPoint point(static_cast<float>(pointInit.x), static_cast<float>(pointInit.y));
    return renderer->isPointInFill(point);
}

bool SVGGeometryElement::isPointInStroke(const DOMPointInit& pointInit)
{
    document().updateLayoutIgnorePendingStylesheets();

    auto* renderer = toRenderSVGShape(this->renderer());
    if (!renderer)
        return false;

    FloatPoint point(static_cast<float>(pointInit.x), static_cast<float>(pointInit.y));
    return renderer->isPointInStroke(point);
}

void SVGGeometryElement::attributeChanged(const QualifiedName& name, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason attributeModificationReason)
{
    if (name == SVGNames::pathLengthAttr) {
        m_pathLength->setBaseValInternal(newValue.toFloat());
        if (m_pathLength->baseVal() < 0)
        	SVGDocumentExtensions::reportError(document(), "A negative value for path attribute <pathLength> is not allowed");
    }

    SVGGraphicsElement::attributeChanged(name, oldValue, newValue, attributeModificationReason);
}

void SVGGeometryElement::svgAttributeChanged(const QualifiedName& attrName)
{
    if (PropertyRegistry::isKnownAttribute(attrName)) {
        ASSERT(attrName == SVGNames::pathLengthAttr);
        InstanceInvalidationGuard guard(*this);
        updateSVGRendererForElementChange();
        return;
    }

    SVGGraphicsElement::svgAttributeChanged(attrName);
}

}
