/*
 * Copyright (C) 2004, 2005, 2006 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2004, 2005, 2006, 2007 Rob Buis <buis@kde.org>
 * Copyright (C) 2007 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "SVGLengthValue.h"

#include "AnimationUtilities.h"
#include "CSSHelper.h"
#include "CSSPrimitiveValue.h"
#include "ExceptionCode.h"
#include "ExceptionCodePlaceholder.h"
#include "FloatConversion.h"
#include "SVGElement.h"
#include "SVGLengthContext.h"
#include "SVGNames.h"
#include "SVGParserUtilities.h"
#include "TextStream.h"
#include <wtf/MathExtras.h>
#include <wtf/NeverDestroyed.h>
#include <wtf/text/StringView.h>

namespace WebCore {

// Helper functions
static const char* lengthTypeToString(SVGLengthType type)
{
    switch (type) {
    case SVGLengthType::Unknown:
    case SVGLengthType::Number:
        return "";
    case SVGLengthType::Percentage:
        return "%";
    case SVGLengthType::Ems:
        return "em";
    case SVGLengthType::Exs:
        return "ex";
    case SVGLengthType::Pixels:
        return "px";
    case SVGLengthType::Centimeters:
        return "cm";
    case SVGLengthType::Millimeters:
        return "mm";
    case SVGLengthType::Inches:
        return "in";
    case SVGLengthType::Points:
        return "pt";
    case SVGLengthType::Picas:
        return "pc";
    }

    ASSERT_NOT_REACHED();
    return "";
}

inline SVGLengthType parseLengthType(const UChar* ptr, const UChar* end)
{
    if (ptr == end)
        return SVGLengthType::Number;

    const UChar firstChar = *ptr;

    if (++ptr == end)
        return firstChar == '%' ? SVGLengthType::Percentage : SVGLengthType::Unknown;

    const UChar secondChar = *ptr;

    if (++ptr != end)
        return SVGLengthType::Unknown;

    if (firstChar == 'e' && secondChar == 'm')
        return SVGLengthType::Ems;
    if (firstChar == 'e' && secondChar == 'x')
        return SVGLengthType::Exs;
    if (firstChar == 'p' && secondChar == 'x')
        return SVGLengthType::Pixels;
    if (firstChar == 'c' && secondChar == 'm')
        return SVGLengthType::Centimeters;
    if (firstChar == 'm' && secondChar == 'm')
        return SVGLengthType::Millimeters;
    if (firstChar == 'i' && secondChar == 'n')
        return SVGLengthType::Inches;
    if (firstChar == 'p' && secondChar == 't')
        return SVGLengthType::Points;
    if (firstChar == 'p' && secondChar == 'c')
        return SVGLengthType::Picas;

    return SVGLengthType::Unknown;
}

static inline SVGLengthType primitiveTypeToLengthType(CSSUnitType primitiveType)
{
    switch (primitiveType) {
    case CSSUnitType::CSS_UNKNOWN:
        return SVGLengthType::Unknown;
    case CSSUnitType::CSS_NUMBER:
        return SVGLengthType::Number;
    case CSSUnitType::CSS_PERCENTAGE:
        return SVGLengthType::Percentage;
    case CSSUnitType::CSS_EMS:
        return SVGLengthType::Ems;
    case CSSUnitType::CSS_EXS:
        return SVGLengthType::Exs;
    case CSSUnitType::CSS_PX:
        return SVGLengthType::Pixels;
    case CSSUnitType::CSS_CM:
        return SVGLengthType::Centimeters;
    case CSSUnitType::CSS_MM:
        return SVGLengthType::Millimeters;
    case CSSUnitType::CSS_IN:
        return SVGLengthType::Inches;
    case CSSUnitType::CSS_PT:
        return SVGLengthType::Points;
    case CSSUnitType::CSS_PC:
        return SVGLengthType::Picas;
    }

    return SVGLengthType::Unknown;
}

static inline CSSUnitType lengthTypeToPrimitiveType(SVGLengthType lengthType)
{
    switch (lengthType) {
    case SVGLengthType::Unknown:
        return CSSUnitType::CSS_UNKNOWN;
    case SVGLengthType::Number:
        return CSSUnitType::CSS_NUMBER;
    case SVGLengthType::Percentage:
        return CSSUnitType::CSS_PERCENTAGE;
    case SVGLengthType::Ems:
        return CSSUnitType::CSS_EMS;
    case SVGLengthType::Exs:
        return CSSUnitType::CSS_EXS;
    case SVGLengthType::Pixels:
        return CSSUnitType::CSS_PX;
    case SVGLengthType::Centimeters:
        return CSSUnitType::CSS_CM;
    case SVGLengthType::Millimeters:
        return CSSUnitType::CSS_MM;
    case SVGLengthType::Inches:
        return CSSUnitType::CSS_IN;
    case SVGLengthType::Points:
        return CSSUnitType::CSS_PT;
    case SVGLengthType::Picas:
        return CSSUnitType::CSS_PC;
    }

    ASSERT_NOT_REACHED();
    return CSSUnitType::CSS_UNKNOWN;
}

SVGLengthValue::SVGLengthValue(SVGLengthMode mode, const String& valueAsString)
    : m_valueInSpecifiedUnits(0)
	, m_lengthType(SVGLengthType::Number)
	, m_lengthMode(mode)
{
    setValueAsString(valueAsString, IGNORE_EXCEPTION);
}

SVGLengthValue::SVGLengthValue(float valueInSpecifiedUnits, SVGLengthType lengthType, SVGLengthMode lengthMode)
    : m_valueInSpecifiedUnits(valueInSpecifiedUnits)
    , m_lengthType(lengthType)
    , m_lengthMode(lengthMode)
{
    ASSERT(m_lengthType != SVGLengthType::Unknown);
}

SVGLengthValue::SVGLengthValue(const SVGLengthContext& context, float value, SVGLengthMode mode, SVGLengthType unitType)
    : m_valueInSpecifiedUnits(0)
	, m_lengthType(unitType)
	, m_lengthMode(mode)
{
    setValue(value, context, IGNORE_EXCEPTION);
}

void SVGLengthValue::setValueAsString(const String& valueAsString, SVGLengthMode mode, ExceptionCode& ec)
{
    m_valueInSpecifiedUnits = 0;
    m_lengthType = SVGLengthType::Number;
    m_lengthMode = mode;
    setValueAsString(valueAsString, ec);
}

SVGLengthValue SVGLengthValue::construct(SVGLengthMode mode, const String& valueAsString, SVGParsingError& parseError, SVGLengthNegativeValuesMode negativeValuesMode)
{
    SVGLengthValue length(mode);

    ExceptionCode ec = 0;
    length.setValueAsString(valueAsString, ec);
    if (ec)
        parseError = ParsingAttributeFailedError;
    else if (negativeValuesMode == ForbidNegativeLengths && length.valueInSpecifiedUnits() < 0)
        parseError = NegativeValueForbiddenError;

    return length;
}

SVGLengthValue SVGLengthValue::blend(const SVGLengthValue& from, const SVGLengthValue& to, float progress)
{
    if ((from.isZero() && to.isZero())
        || from.lengthType() == SVGLengthType::Unknown
        || to.lengthType() == SVGLengthType::Unknown
        || (!from.isZero() && from.lengthType() != SVGLengthType::Percentage && to.lengthType() == SVGLengthType::Percentage)
        || (!to.isZero() && from.lengthType() == SVGLengthType::Percentage && to.lengthType() != SVGLengthType::Percentage)
        || (!from.isZero() && !to.isZero() && (from.lengthType() == SVGLengthType::Ems || from.lengthType() == SVGLengthType::Exs) && from.lengthType() != to.lengthType()))
        return to;

    if (from.lengthType() == SVGLengthType::Percentage || to.lengthType() == SVGLengthType::Percentage) {
        auto fromPercent = from.valueAsPercentage() * 100;
        auto toPercent = to.valueAsPercentage() * 100;
        return SVGLengthValue(WebCore::blend(fromPercent, toPercent, BlendingContext(progress)), SVGLengthType::Percentage);
    }

    if (from.lengthType() == to.lengthType() || from.isZero() || to.isZero() || from.isRelative()) {
        auto fromValue = from.valueInSpecifiedUnits();
        auto toValue = to.valueInSpecifiedUnits();
        return SVGLengthValue(WebCore::blend(fromValue, toValue, BlendingContext(progress)), to.isZero() ? from.lengthType() : to.lengthType());
    }

    SVGLengthContext nonRelativeLengthContext(nullptr);
    ExceptionCode ec = 0;
    auto fromValueInUserUnits = nonRelativeLengthContext.convertValueToUserUnits(from.valueInSpecifiedUnits(), from.lengthMode(), from.lengthType(), ec);
    if (ec)
        return SVGLengthValue();

    auto fromValue = nonRelativeLengthContext.convertValueFromUserUnits(fromValueInUserUnits, to.lengthMode(), to.lengthType(), ec);
    if (ec)
        return SVGLengthValue();

    float toValue = to.valueInSpecifiedUnits();
    return SVGLengthValue(WebCore::blend(fromValue, toValue, BlendingContext(progress)), to.lengthType());
}

float SVGLengthValue::value(const SVGLengthContext& context) const
{
    ExceptionCode ec = 0;
    auto result = valueForBindings(context, ec);
    if (ec)
        return 0;
    return result;
}

float SVGLengthValue::valueForBindings(const SVGLengthContext& context, ExceptionCode& ec) const
{
    return context.convertValueToUserUnits(m_valueInSpecifiedUnits, m_lengthMode, m_lengthType, ec);
}

void SVGLengthValue::setValue(const SVGLengthContext& context, float value, SVGLengthMode mode, SVGLengthType unitType, ExceptionCode& ec)
{
    // FIXME: Seems like a bug that we change the value of m_unit even if setValue throws an exception.
	m_lengthType = unitType;
	m_lengthMode = mode;
    setValue(value, context, ec);
}

void SVGLengthValue::setValue(float value, const SVGLengthContext& context, ExceptionCode& ec)
{
    // 100% = 100.0 instead of 1.0 for historical reasons, this could eventually be changed
    if (m_lengthType == SVGLengthType::Percentage)
        value = value / 100;

    auto convertedValue = context.convertValueFromUserUnits(value, m_lengthMode, m_lengthType, ec);
    if (ec)
        return;
    m_valueInSpecifiedUnits = convertedValue;
}
float SVGLengthValue::valueAsPercentage() const
{
    // 100% = 100.0 instead of 1.0 for historical reasons, this could eventually be changed
    if (m_lengthType == SVGLengthType::Percentage)
        return m_valueInSpecifiedUnits / 100;

    return m_valueInSpecifiedUnits;
}

void SVGLengthValue::setValueAsString(const String& string, ExceptionCode& ec)
{
    if (string.isEmpty())
        return;

    float convertedNumber = 0;
    auto upconvertedCharacters = StringView(string).upconvertedCharacters();
    const UChar* ptr = upconvertedCharacters;
    const UChar* end = ptr + string.length();

    if (!parseNumber(ptr, end, convertedNumber, false)) {
        ec = SYNTAX_ERR;
        return;
    }

    auto lengthType = parseLengthType(ptr, end);
    if (lengthType == SVGLengthType::Unknown) {
        ec = SYNTAX_ERR;
        return;
    }

    m_lengthType = lengthType;
    m_valueInSpecifiedUnits = convertedNumber;
}

String SVGLengthValue::valueAsString() const
{
    return makeString(String::number(m_valueInSpecifiedUnits), lengthTypeToString(m_lengthType));
}

AtomString SVGLengthValue::valueAsAtomicString() const
{
    StringBuilder builder;
    builder.appendNumber(m_valueInSpecifiedUnits);
    builder.append(lengthTypeToString(m_lengthType));
    return builder.toAtomicString();
}

void SVGLengthValue::convertToSpecifiedUnits(const SVGLengthContext& context, SVGLengthType lengthType, ExceptionCode& ec)
{
    auto valueInUserUnits = valueForBindings(context, ec);
    if (ec)
        return;

    auto originalLengthType = m_lengthType;
    m_lengthType = lengthType;
    setValue(valueInUserUnits, context, ec);
    if (!ec)
    	return;

	m_lengthType = originalLengthType;
}

SVGLengthValue SVGLengthValue::fromCSSPrimitiveValue(const CSSPrimitiveValue& value, const CSSToLengthConversionData& conversionData, ShouldConvertNumberToPxLength shouldConvertNumberToPxLength)
{
	auto primitiveType = value.primitiveType();
	if (primitiveType == CSSUnitType::CSS_NUMBER && shouldConvertNumberToPxLength == ShouldConvertNumberToPxLength::Yes)
		 return SVGLengthValue(value.floatValue(), SVGLengthType::Pixels);

	auto lengthType = primitiveTypeToLengthType(primitiveType);
	switch (lengthType) {
	case SVGLengthType::Unknown:
		return SVGLengthValue();
	case SVGLengthType::Number:
	case SVGLengthType::Percentage:
		return SVGLengthValue(value.floatValue(), lengthType);
	default:
		return SVGLengthValue(value.computeLength<float>(conversionData), SVGLengthType::Pixels);
	}

	ASSERT_NOT_REACHED();
	return SVGLengthValue();
}

Ref<CSSPrimitiveValue> SVGLengthValue::toCSSPrimitiveValue(const Element* element) const
{
    if (isSVGElement(element)) {
        SVGLengthContext context(toSVGElement(element));
        ExceptionCode ec = 0;
        auto computedValue = context.convertValueToUserUnits(valueInSpecifiedUnits(), lengthMode(), lengthType(), ec);
        if (!ec)
            return CSSPrimitiveValue::create(computedValue, CSSUnitType::CSS_PX);
    }

    return CSSPrimitiveValue::create(valueInSpecifiedUnits(), lengthTypeToPrimitiveType(lengthType()));
}

SVGLengthMode SVGLengthValue::lengthModeForAnimatedLengthAttribute(const QualifiedName& attrName)
{
    static NeverDestroyed<HashMap<QualifiedName, SVGLengthMode>> map;
    if (map.get().isEmpty()) {
    	map.get().add(SVGNames::xAttr, SVGLengthMode::Width);
    	map.get().add(SVGNames::yAttr, SVGLengthMode::Height);
    	map.get().add(SVGNames::cxAttr, SVGLengthMode::Width);
    	map.get().add(SVGNames::cyAttr, SVGLengthMode::Height);
		map.get().add(SVGNames::dxAttr, SVGLengthMode::Width);
		map.get().add(SVGNames::dyAttr, SVGLengthMode::Height);
		map.get().add(SVGNames::fxAttr, SVGLengthMode::Width);
		map.get().add(SVGNames::fyAttr, SVGLengthMode::Height);
		map.get().add(SVGNames::widthAttr, SVGLengthMode::Width);
		map.get().add(SVGNames::heightAttr, SVGLengthMode::Height);
		map.get().add(SVGNames::x1Attr, SVGLengthMode::Width);
		map.get().add(SVGNames::x2Attr, SVGLengthMode::Width);
		map.get().add(SVGNames::y1Attr, SVGLengthMode::Height);
		map.get().add(SVGNames::y2Attr, SVGLengthMode::Height);
		map.get().add(SVGNames::refXAttr, SVGLengthMode::Width);
		map.get().add(SVGNames::refYAttr, SVGLengthMode::Height);
		map.get().add(SVGNames::markerWidthAttr, SVGLengthMode::Width);
		map.get().add(SVGNames::markerHeightAttr, SVGLengthMode::Height);
		map.get().add(SVGNames::textLengthAttr, SVGLengthMode::Width);
		map.get().add(SVGNames::startOffsetAttr, SVGLengthMode::Width);
    }
    
    auto result = map.get().find(attrName);
    if (result == map.get().end())
        return SVGLengthMode::Other;
    return result->second;
}

TextStream& operator<<(TextStream& ts, const SVGLengthValue& length)
{
    ts << length.valueAsString();
    return ts;
}

}
