/*
 * Copyright (C) 2004, 2005, 2006, 2008 Nikolas Zimmermann <zimmermann@kde.org>
 * Copyright (C) 2004, 2005, 2006, 2007 Rob Buis <buis@kde.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"

#if ENABLE(SVG)
#include "SVGPathElement.h"

#include "Attribute.h"
#include "RenderSVGPath.h"
#include "RenderSVGResource.h"
#include "SVGMPathElement.h"
#include "SVGNames.h"
#include "SVGPathSegList.h"
#include "SVGPathUtilities.h"
#include "SVGPoint.h"
#include "SVGSVGElement.h"
#include <wtf/NeverDestroyed.h>

namespace WebCore {

namespace {
class PathSegListCache {
public:
    static PathSegListCache& singleton();

    Optional<DataRef<SVGPathByteStream::Data>> get(const AtomString& attributeValue) const;
	void add(const AtomString& attributeValue, DataRef<SVGPathByteStream::Data>);
	void clear();

private:
	friend class NeverDestroyed<PathSegListCache>;
	PathSegListCache() = default;

    HashMap<AtomString, DataRef<SVGPathByteStream::Data>> m_cache;
	uint64_t m_sizeInBytes { 0 };
    static constexpr uint64_t maxItemSizeInBytes = 5 * 1024; // 5 Kb.
    static constexpr uint64_t maxCacheSizeInBytes = 150 * 1024; // 150 Kb.
};

PathSegListCache& PathSegListCache::singleton()
{
	static NeverDestroyed<PathSegListCache> cache;
    return cache;
}

Optional<DataRef<SVGPathByteStream::Data>> PathSegListCache::get(const AtomString& attributeValue) const
{
    auto iterator = m_cache.find(attributeValue);
    return iterator != m_cache.end() ? Optional<DataRef<SVGPathByteStream::Data>>(iterator->second) : Optional<DataRef<SVGPathByteStream::Data>>();
}

void PathSegListCache::add(const AtomString& attributeValue, DataRef<SVGPathByteStream::Data> data)
{
    size_t newDataSize = data->size();
    if (UNLIKELY(newDataSize > maxItemSizeInBytes))
        return;

    m_sizeInBytes += newDataSize;
    while (m_sizeInBytes > maxCacheSizeInBytes) {
        ASSERT(!m_cache.isEmpty());
        auto iteratorToRemove = m_cache.random();
        ASSERT(iteratorToRemove != m_cache.end());
        ASSERT(m_sizeInBytes >= iteratorToRemove->value->size());
        m_sizeInBytes -= iteratorToRemove->second->size();
        m_cache.remove(iteratorToRemove);
    }
    m_cache.add(attributeValue, WTFMove(data));
}

void PathSegListCache::clear()
{
    m_cache.clear();
    m_sizeInBytes = 0;
}
}

ALWAYS_INLINE SVGPathElement::SVGPathElement(const QualifiedName& tagName, Document* document)
    : SVGGeometryElement(tagName, document, adoptPtr(new PropertyRegistry(*this)))
{
	ASSERT(hasTagName(SVGNames::pathTag));

    static std::once_flag onceFlag;
    std::call_once(onceFlag, [] {
        PropertyRegistry::registerProperty<SVGNames::dAttr, &SVGPathElement::m_pathSegList>();
    });
}

Ref<SVGPathElement> SVGPathElement::create(const QualifiedName& tagName, Document* document)
{
    return adoptRef(*new SVGPathElement(tagName, document));
}

SVGPathElement::~SVGPathElement() = default;

void SVGPathElement::attributeChanged(const QualifiedName& name, const AtomString& oldValue, const AtomString& newValue, AttributeModificationReason attributeModificationReason)
{
    if (name == SVGNames::dAttr) {
		auto& cache = PathSegListCache::singleton();
        if (newValue.isEmpty())
            m_pathSegList->baseVal()->clearByteStreamData();
        else if (auto data = cache.get(newValue))
			m_pathSegList->baseVal()->updateByteStreamData(WTFMove(data.value()));
		else if (m_pathSegList->baseVal()->parse(newValue))
			cache.add(newValue, m_pathSegList->baseVal()->existingPathByteStream().data());
		else
			SVGDocumentExtensions::reportError(document(), tryMakeString("Problem parsing d=\"", newValue, "\""));
    }

    SVGGeometryElement::attributeChanged(name, oldValue, newValue, attributeModificationReason);
}

void SVGPathElement::clearCache()
{
    PathSegListCache::singleton().clear();
}

void SVGPathElement::svgAttributeChanged(const QualifiedName& attrName)
{
    if (PropertyRegistry::isKnownAttribute(attrName)) {
        ASSERT(attrName == SVGNames::dAttr);
        InstanceInvalidationGuard guard(*this);
        invalidateMPathDependencies();

        if (auto* renderer = this->renderer())
            static_cast<RenderSVGPath*>(renderer)->setNeedsShapeUpdate();
        updateSVGRendererForElementChange();

        return;
    }

	SVGGeometryElement::svgAttributeChanged(attrName);
}

void SVGPathElement::invalidateMPathDependencies()
{
    // <mpath> can only reference <path> but this dependency is not handled in
    // markForLayoutAndParentResourceInvalidation so we update any mpath dependencies manually.
	for (auto& element : referencingElements()) {
		if (isSVGMPathElement(element))
			toSVGMPathElement(element.get()).targetPathChanged();
	}
}

Node::InsertedIntoResult SVGPathElement::insertedIntoAncestor(const InsertionType& insertionType, ContainerNode& parentOfInsertedTree)
{
	auto result = SVGGeometryElement::insertedIntoAncestor(insertionType, parentOfInsertedTree);
    invalidateMPathDependencies();
    return result;
}

void SVGPathElement::removedFromAncestor(const RemovalType& removalType, ContainerNode& parentOfRemovedTree)
{
	SVGGeometryElement::removedFromAncestor(removalType, parentOfRemovedTree);
    invalidateMPathDependencies();
}

float SVGPathElement::getTotalLength() const
{
    return getTotalLengthOfSVGPathByteStream(pathByteStream());
}

PassRefPtr<SVGPoint> SVGPathElement::getPointAtLength(float length, ExceptionCode&) const
{
    FloatPoint point = getPointAtLengthOfSVGPathByteStream(pathByteStream(), length);
    return SVGPoint::create(point);
}

unsigned SVGPathElement::getPathSegAtLength(float length) const
{
    unsigned pathSeg = getSVGPathSegAtLengthFromSVGPathByteStream(pathByteStream(), length);
    return pathSeg;
}

FloatRect SVGPathElement::getBBox(StyleUpdateStrategy styleUpdateStrategy)
{
    if (styleUpdateStrategy == AllowStyleUpdate)
        this->document().updateLayoutIgnorePendingStylesheets();

    RenderSVGPath* renderer = static_cast<RenderSVGPath*>(this->renderer());

    // FIXME: Eventually we should support getBBox for detached elements.
    // FIXME: If the path is null it means we're calling getBBox() before laying out this element,
    // which is an error.
    if (!renderer || !renderer->hasPath())
        return FloatRect();

    return renderer->path().boundingRect();
}

RenderPtr<RenderElement> SVGPathElement::createElementRenderer(PassOwnPtr<RenderStyle> style, const RenderTreePosition&)
{
    // By default, any subclass is expected to do path-based drawing
    return createRenderer<RenderSVGPath>(*this, style);
}

}

#endif // ENABLE(SVG)
