/*
 * Copyright (C) 2006 Eric Seidel <eric@webkit.org>
 * Copyright (C) 2009 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#pragma once

#if ENABLE(SVG)

#include "Image.h"
#include "KURL.h"

namespace WebCore {

class Element;
class FrameLoader;
class FrameView;
class ImageBuffer;
class Page;
class RenderBox;
class SVGFrameLoaderClient;
class SVGImageChromeClient;
class SVGImageForContainer;
class SVGSVGElement;

class SVGImage FINAL : public Image {
public:
    static PassRefPtr<SVGImage> create(ImageObserver& observer) { return adoptRef(new SVGImage(observer)); }

    RenderBox* embeddedContentBox() const;
    FrameView* frameView() const;

    virtual bool isSVGImage() const FINAL { return true; }
    virtual FloatSize size() const FINAL { return m_intrinsicSize; }

    virtual bool hasSingleSecurityOrigin() const OVERRIDE;

    void setDataProtocolLoader(FrameLoader* dataProtocolLoader) { m_dataProtocolLoader = dataProtocolLoader; }

    virtual bool hasRelativeWidth() const FINAL;
    virtual bool hasRelativeHeight() const FINAL;

    virtual void startAnimation(bool /*catchUpIfNecessary*/ = true) FINAL;
    virtual void stopAnimation() FINAL;
    virtual void resetAnimation() FINAL;
    bool isAnimating() const FINAL;

    void scheduleStartAnimation();

    Page* internalPage() { return m_page.get(); }
    WEBCORE_EXPORT RefPtr<SVGSVGElement> rootElement() const;

    virtual NativeImagePtr nativeImageForCurrentFrame() override;

private:
    friend class SVGImageChromeClient;
    friend class SVGImageForContainer;

    virtual ~SVGImage();

    virtual String filenameExtension() const OVERRIDE;

    virtual void setContainerSize(const FloatSize&) OVERRIDE;
    IntSize containerSize() const;
    virtual bool usesContainerSize() const OVERRIDE { return true; }
    virtual void computeIntrinsicDimensions(Length& intrinsicWidth, Length& intrinsicHeight, FloatSize& intrinsicRatio) OVERRIDE;

    virtual bool dataChanged(bool allDataReceived) OVERRIDE;

    // FIXME: SVGImages are underreporting decoded sizes and will be unable
    // to prune because these functions are not implemented yet.
    virtual void destroyDecodedData(bool) OVERRIDE { }
    virtual unsigned decodedSize() const OVERRIDE { return 0; }

    // FIXME: Implement this to be less conservative.
    virtual bool currentFrameKnownToBeOpaque() OVERRIDE { return false; }

    void startAnimationTimerFired();

    explicit SVGImage(ImageObserver&);
    virtual void draw(GraphicsContext*, const FloatRect& fromRect, const FloatRect& toRect, CompositeOperator, BlendMode) OVERRIDE;
    void drawForContainer(GraphicsContext*, const FloatSize containerSize, float containerZoom, const KURL& initialFragmentURL, const FloatRect& dstRect, const FloatRect& srcRect, CompositeOperator, BlendMode);
	void drawPatternForContainer(GraphicsContext*, const FloatSize& containerSize, float containerZoom, const KURL& initialFragmentURL, const FloatRect& srcRect, const AffineTransform&, const FloatPoint& phase, const FloatSize& spacing,
		CompositeOperator, const FloatRect&, BlendMode);

    OwnPtr<SVGFrameLoaderClient> m_loaderClient;
    OwnPtr<SVGImageChromeClient> m_chromeClient;
    OwnPtr<Page> m_page;
    FloatSize m_intrinsicSize;
    FrameLoader* m_dataProtocolLoader;

    Timer<SVGImage> m_startAnimationTimer;
};

bool isInSVGImage(const Element&);

inline bool isSVGImage(const Image& image)
{
	return image.isSVGImage();
}

template<>
ALWAYS_INLINE SVGImage* dynamicDowncast(Image* source)
{
    //static_assert(!std::is_same_v<Source, Target>, "Unnecessary cast to same type");
    //static_assert(std::is_base_of_v<Source, Target>, "Should be a downcast");
    return source && isSVGImage(*source) ? static_cast<SVGImage*>(source) : nullptr;
}

}

#endif // ENABLE(SVG)
