/*
 *  Copyright (C) 2004, 2006, 2008 Apple Inc. All rights reserved.
 *  Copyright (C) 2005-2007 Alexey Proskuryakov <ap@webkit.org>
 *  Copyright (C) 2007, 2008 Julien Chaffraix <jchaffraix@webkit.org>
 *  Copyright (C) 2008, 2011 Google Inc. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "config.h"
#include "XMLHttpRequest.h"

#include "Blob.h"
#include "BlobData.h"
#include "CachedResourceRequestInitiators.h"
#include "ConsoleTypes.h"
#include "ContentSecurityPolicy.h"
#include "CrossOriginAccessControl.h"
#include "DOMFormData.h"
#include "DOMImplementation.h"
#include "Event.h"
#include "EventException.h"
#include "EventListener.h"
#include "EventNames.h"
#include "ExceptionCode.h"
#include "File.h"
#include "Frame.h"
#include "FrameLoader.h"
#include "FrameLoaderClient.h"
#include "HTMLDocument.h"
#include "HTTPHeaderNames.h"
#include "HTTPHeaderValues.h"
#include "HTTPParsers.h"
#include "HistogramSupport.h"
#include "InspectorInstrumentation.h"
#include "MemoryCache.h"
#include "MIMETypeRegistry.h"
#include "ParsedContentType.h"
#include "ResourceError.h"
#include "ResourceRequest.h"
#include "ScriptCallStack.h"
#include "ScriptProfile.h"
#include "SecurityOrigin.h"
#include "SecurityOriginPolicy.h"
#include "Settings.h"
#include "SharedBuffer.h"
#include "TextResourceDecoder.h"
#include "ThreadableLoader.h"
#include "URLSearchParams.h"
#include "XMLDocument.h"
#include "XMLHttpRequestProgressEvent.h"
#include "XMLHttpRequestUpload.h"
#include "markup.h"
#include <runtime/ArrayBuffer.h>
#include <runtime/ArrayBufferView.h>
#include <wtf/HexNumber.h>
#include <wtf/RefCountedLeakCounter.h>
#include <wtf/StdLibExtras.h>
#include <wtf/UnusedParam.h>
#include <wtf/text/CString.h>

#if USE(JSC)
#include "JSDOMBinding.h"
#include "JSDOMWindow.h"
#include <heap/Strong.h>
#include <runtime/JSLock.h>
#include <runtime/Operations.h>
#endif

namespace WebCore {

DEFINE_DEBUG_ONLY_GLOBAL(WTF::RefCountedLeakCounter, xmlHttpRequestCounter, ("XMLHttpRequest"));

// Histogram enum to see when we can deprecate xhr.send(ArrayBuffer).
enum XMLHttpRequestSendArrayBufferOrView {
    XMLHttpRequestSendArrayBuffer,
    XMLHttpRequestSendArrayBufferView,
    XMLHttpRequestSendArrayBufferOrViewMax,
};

static void replaceCharsetInMediaTypeIfNeeded(String& mediaType)
{
    auto parsedContentType = ParsedContentType::create(mediaType);
    if (!parsedContentType || parsedContentType->charset().isEmpty() || equalIgnoringASCIICase(parsedContentType->charset(), "UTF-8"))
        return;

    parsedContentType->setCharset(ASCIILiteral("UTF-8"));
    mediaType = parsedContentType->serialize();
}

static void logConsoleError(ScriptExecutionContext* context, const String& message)
{
    if (!context)
        return;
    // FIXME: It's not good to report the bad usage without indicating what source line it came from.
    // We should pass additional parameters so we can tell the console where the mistake occurred.
    context->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Error, message);
}

PassRefPtr<XMLHttpRequest> XMLHttpRequest::create(ScriptExecutionContext* context)
{
    RefPtr<XMLHttpRequest> xmlHttpRequest(adoptRef(new XMLHttpRequest(context)));
    xmlHttpRequest->suspendIfNeeded();

    return xmlHttpRequest.release();
}

XMLHttpRequest::XMLHttpRequest(ScriptExecutionContext* context)
	: ActiveDOMObject(context)
	, m_async(true)
	, m_includeCredentials(false)
	, m_state(UNSENT)
	, m_sendFlag(false)
	, m_createdDocument(false)
	, m_error(false)
	, m_uploadListenerFlag(false)
	, m_uploadComplete(false)
	, m_wasAbortedByClient(false)
	, m_receivedLength(0)
	, m_lastSendLineNumber(0)
	, m_progressEventThrottle(*this)
	, m_responseCacheIsValid(false)
	, m_timeoutMilliseconds(0)
	, m_sendingTime(0.0)
	, m_timeoutTimer(this, &XMLHttpRequest::timeoutTimerFired)
{
#ifndef NDEBUG
    xmlHttpRequestCounter.increment();
#endif
}

XMLHttpRequest::~XMLHttpRequest()
{
#ifndef NDEBUG
    xmlHttpRequestCounter.decrement();
#endif
}

Document* XMLHttpRequest::document() const
{
    if(scriptExecutionContext()->isDocument())
    	return static_cast<Document*>(scriptExecutionContext());
    return nullptr;
}

#if ENABLE(DASHBOARD_SUPPORT)
bool XMLHttpRequest::usesDashboardBackwardCompatibilityMode() const
{
    if (scriptExecutionContext()->isWorkerGlobalScope())
        return false;
    return document()->settings().usesDashboardBackwardCompatibilityMode();
}
#endif

String XMLHttpRequest::responseText(ExceptionCode& ec)
{
    if (m_responseType != ResponseType::EmptyString && m_responseType != ResponseType::Text) {
        ec = INVALID_STATE_ERR;
        return String();
    }
    return responseTextIgnoringResponseType();
}

void XMLHttpRequest::didCacheResponseJSON()
{
    ASSERT(m_responseType == ResponseType::Json);
    ASSERT(doneWithoutErrors());
    m_responseCacheIsValid = true;
    m_responseBuilder.clear();
}

Document* XMLHttpRequest::responseXML(ExceptionCode& ec)
{
    ASSERT(scriptExecutionContext()->isDocument());

	if (m_responseType != ResponseType::EmptyString && m_responseType != ResponseType::Document) {
		ec = INVALID_STATE_ERR;
        return nullptr;
    }

    if (!doneWithoutErrors())
        return nullptr;

    if (!m_createdDocument) {
        String mimeType = responseMIMEType();
        bool isHTML = equalLettersIgnoringASCIICase(mimeType, "text/html");
        bool isXML = MIMETypeRegistry::isXMLMIMEType(mimeType);

        // The W3C spec requires the final MIME type to be some valid XML type, or text/html.
        // If it is text/html, then the responseType of "document" must have been supplied explicitly.
        if ((m_response.isHTTP() && !isXML && !isHTML)
            || (isHTML && m_responseType == ResponseType::EmptyString)) {
            m_responseDocument = nullptr;
        } else {
            if (isHTML)
                m_responseDocument = HTMLDocument::create(nullptr, m_response.url());
            else
                m_responseDocument = XMLDocument::create(nullptr, m_response.url());
            m_responseDocument->overrideLastModified(m_response.lastModified());
            m_responseDocument->setContextDocument(static_cast<Document&>(*scriptExecutionContext()));
			m_responseDocument->setSecurityOriginPolicy(scriptExecutionContext()->securityOriginPolicy());
            m_responseDocument->overrideMIMEType(mimeType);
            m_responseDocument->setContent(m_responseBuilder.toStringPreserveCapacity());
            if (m_decoder)
                m_responseDocument->setDecoder(m_decoder.copyRef());

            if (!m_responseDocument->wellFormed())
                m_responseDocument = nullptr;
        }
        m_createdDocument = true;
    }

    return m_responseDocument.get();
}

Blob* XMLHttpRequest::responseBlob()
{
    ASSERT(m_responseType == ResponseType::Blob);
    ASSERT(doneWithoutErrors());

    if (!m_responseBlob) {
        // FIXME: This causes two (or more) unnecessary copies of the data.
        // Chromium stores blob data in the browser process, so we're pulling the data
        // from the network only to copy it into the renderer to copy it back to the browser.
        // Ideally we'd get the blob/file-handle from the ResourceResponse directly
        // instead of copying the bytes. Embedders who store blob data in the
        // same process as WebCore would at least to teach BlobData to take
        // a SharedBuffer, even if they don't get the Blob from the network layer directly.
        OwnPtr<BlobData> blobData = BlobData::create();
        // If we errored out or got no data, we still return a blob, just an empty one.
        size_t size = 0;
		if (m_binaryResponseBuilder) {
			RefPtr<RawData> rawData = RawData::create();
			size = m_binaryResponseBuilder->size();
			rawData->mutableData()->append(m_binaryResponseBuilder->data(), size);
			blobData->appendData(rawData, 0, BlobDataItem::toEndOfFile);
		}
		blobData->setContentType(responseMIMEType(FinalMIMEType::Yes)); // responseMIMEType defaults to text/xml which may be incorrect.
		m_binaryResponseBuilder = nullptr;
        m_responseBlob = Blob::create(scriptExecutionContext(), blobData.release(), size);
    }

    return m_responseBlob.get();
}

JSC::ArrayBuffer* XMLHttpRequest::responseArrayBuffer()
{
    ASSERT(m_responseType == ResponseType::Arraybuffer);
    ASSERT(doneWithoutErrors());

    if (!m_responseArrayBuffer) {
        if (m_binaryResponseBuilder) {
        	int length = m_binaryResponseBuilder->size();
        	char* data = m_binaryResponseBuilder->releaseBuffer();

        	m_binaryResponseBuilder.clear();

        	m_responseArrayBuffer = JSC::ArrayBuffer::createAdopted(data, length * sizeof(char));
        }
        else
            m_responseArrayBuffer = JSC::ArrayBuffer::tryCreate(nullptr, 0);
    }

    return m_responseArrayBuffer.get();
}

void XMLHttpRequest::setTimeout(unsigned timeout, ExceptionCode& ec)
{
	if (scriptExecutionContext()->isDocument() && !m_async) {
        logConsoleError(scriptExecutionContext(), ASCIILiteral("XMLHttpRequest.timeout cannot be set for synchronous HTTP(S) requests made from the window context."));
        ec = INVALID_ACCESS_ERR;
        return;
    }
    m_timeoutMilliseconds = timeout;
    if (!m_timeoutTimer.isActive())
        return;

    // If timeout is zero, we should use the default network timeout. But we disabled it so let's mimic it with a 60 seconds timeout value.
    double interval = ((m_timeoutMilliseconds ? m_timeoutMilliseconds : 60000) / 1000.0) - (monotonicallyIncreasingTime() - m_sendingTime);
    m_timeoutTimer.startOneShot(std::max(0.0, interval));
}

void XMLHttpRequest::setResponseType(ResponseType type, ExceptionCode& ec)
{
    if (!scriptExecutionContext()->isDocument() && type == ResponseType::Document)
        return;

    if (m_state >= LOADING) {
        ec = INVALID_STATE_ERR;
        return;
    }

    // Newer functionality is not available to synchronous requests in window contexts, as a spec-mandated 
    // attempt to discourage synchronous XHR use. responseType is one such piece of functionality.
    // We'll only disable this functionality for HTTP(S) requests since sync requests for local protocols
    // such as file: and data: still make sense to allow.
    if (!m_async && scriptExecutionContext()->isDocument() && m_url.protocolIsInHTTPFamily()) {
        logConsoleError(scriptExecutionContext(), ASCIILiteral("XMLHttpRequest.responseType cannot be changed for synchronous HTTP(S) requests made from the window context."));
        ec = INVALID_ACCESS_ERR;
        return;
    }

    m_responseType = type;
}

String XMLHttpRequest::responseURL() const
{
    KURL responseURL(m_response.url());
    responseURL.removeFragmentIdentifier();

    return responseURL.string();
}

XMLHttpRequestUpload* XMLHttpRequest::upload()
{
    if (!m_upload)
        m_upload = XMLHttpRequestUpload::create(this);
    return m_upload.get();
}

void XMLHttpRequest::changeState(State newState)
{
    if (m_state != newState) {
        // Setting the readyState to DONE could get the wrapper collected before we get a chance to fire the JS
        // events in callReadyStateChangeListener() below so we extend the lifetime of the JS wrapper until the
        // of this scope.
        auto eventFiringActivity = makePendingActivity(*this);

        m_state = newState;
        if (readyState() == DONE) {
            // The XHR object itself holds on to the responseText, and
            // thus has extra cost even independent of any
            // responseText or responseXML objects it has handed
            // out. But it is protected from GC while loading, so this
            // can't be recouped until the load is done, so only
            // report the extra cost at that point.
            if (auto* context = scriptExecutionContext()) {
                JSC::VM& vm = *context->vm();
                JSC::JSLockHolder lock(vm);
                vm.heap.reportExtraMemoryAllocated(memoryCost());
            }
        }
        callReadyStateChangeListener();
    }
}

void XMLHttpRequest::callReadyStateChangeListener()
{
    if (!scriptExecutionContext())
        return;

    // Check whether sending load and loadend events before sending readystatechange event, as it may change m_error/m_state values.
    bool shouldSendLoadEvent = (m_state == DONE && !m_error);

    InspectorInstrumentationCookie cookie = InspectorInstrumentation::willChangeXHRReadyState(scriptExecutionContext(), this);

    if (m_async || (m_state <= OPENED || m_state == DONE))
        m_progressEventThrottle.dispatchReadyStateChangeEvent(Event::create(eventNames().readystatechangeEvent, false, false), m_state == DONE ? FlushProgressEvent : DoNotFlushProgressEvent);

    InspectorInstrumentation::didChangeXHRReadyState(cookie);

    if (shouldSendLoadEvent) {
        auto& eventNames = WebCore::eventNames();
        InspectorInstrumentationCookie cookie = InspectorInstrumentation::willLoadXHR(scriptExecutionContext(), this);
        m_progressEventThrottle.dispatchProgressEvent(eventNames.loadEvent);
        InspectorInstrumentation::didLoadXHR(cookie);
        m_progressEventThrottle.dispatchProgressEvent(eventNames.loadendEvent);
    }
}

void XMLHttpRequest::setWithCredentials(bool value, ExceptionCode& ec)
{
    if (m_state > OPENED || m_sendFlag) {
        ec = INVALID_STATE_ERR;
        return;
    }

    m_includeCredentials = value;
}

bool XMLHttpRequest::isAllowedHTTPMethod(const String& method)
{
    return !equalLettersIgnoringASCIICase(method, "trace")
        && !equalLettersIgnoringASCIICase(method, "track")
        && !equalLettersIgnoringASCIICase(method, "connect");
}

String XMLHttpRequest::uppercaseKnownHTTPMethod(const String& method)
{
    const char* const methods[] = { "COPY", "DELETE", "GET", "HEAD", "INDEX", "LOCK", "M-POST", "MKCOL", "MOVE", "OPTIONS", "POST", "PROPFIND", "PROPPATCH", "PUT", "UNLOCK" };
    for (unsigned i = 0; i < WTF_ARRAY_LENGTH(methods); ++i) {
        if (equalIgnoringASCIICase(method, methods[i])) {
            // Don't bother allocating a new string if it's already all uppercase.
            if (method == methods[i])
                break;
            return String::fromLiteral(methods[i]);
        }
    }
    return method;
}

void XMLHttpRequest::open(const String& method, const String& url, ExceptionCode& ec)
{
    // If the async argument is omitted, set async to true.
    return open(method, scriptExecutionContext()->completeURL(url), true, ec);
}

void XMLHttpRequest::open(const String& method, const KURL& url, bool async, ExceptionCode& ec)
{
    auto* context = scriptExecutionContext();
    bool contextIsDocument = (*context).isDocument();
    if (contextIsDocument && !static_cast<Document&>(*context).isFullyActive()) {
        //return Exception { InvalidStateError, "Document is not fully active"_s };
    	ec = INVALID_STATE_ERR;
    	return;
    }

    if (!isValidHTTPToken(method)) {
        ec = SYNTAX_ERR;
        return;
    }

    if (!isAllowedHTTPMethod(method)) {
        logConsoleError(context, tryMakeString("XMLHttpRequest::open(", method, ") not allowed."));
        ec = SECURITY_ERR;
        return;
    }

    if (!async && contextIsDocument) {
        // Newer functionality is not available to synchronous requests in window contexts, as a spec-mandated
        // attempt to discourage synchronous XHR use. responseType is one such piece of functionality.
        // We'll only disable this functionality for HTTP(S) requests since sync requests for local protocols
        // such as file: and data: still make sense to allow.
        if (url.protocolIsInHTTPFamily() && m_responseType != ResponseType::EmptyString) {
            logConsoleError(context, ASCIILiteral("Synchronous HTTP(S) requests made from the window context cannot have XMLHttpRequest.responseType set."));
            ec = INVALID_ACCESS_ERR;
            return;
        }

        // Similarly, timeouts are disabled for synchronous requests as well.
        if (m_timeoutMilliseconds > 0) {
            logConsoleError(context, ASCIILiteral("Synchronous XMLHttpRequests must not have a timeout value set."));
            ec = INVALID_ACCESS_ERR;
            return;
        }
    }

    if (!internalAbort())
        return;

    m_sendFlag = false;
    m_uploadListenerFlag = false;
    m_method = uppercaseKnownHTTPMethod(method);
    m_error = false;
    m_uploadComplete = false;
    m_wasAbortedByClient = false;

    // clear stuff from possible previous load
    clearResponse();
    clearRequest();

    m_url = url;
    context->contentSecurityPolicy()->upgradeInsecureRequestIfNeeded(m_url, ContentSecurityPolicy::InsecureRequestType::Load);
    if (m_url.protocolIsBlob())
		m_blobURLLifetimeExtension = m_url;

    m_async = async;

    ASSERT(!m_loadingActivity);

	changeState(OPENED);
}

void XMLHttpRequest::open(const String& method, const String& url, bool async, const String& user, const String& password, ExceptionCode& ec)
{
    KURL urlWithCredentials = scriptExecutionContext()->completeURL(url);
    if (!user.isNull()) {
        urlWithCredentials.setUser(user);
        if (!password.isNull())
            urlWithCredentials.setPass(password);
    }

    open(method, urlWithCredentials, async, ec);
}

bool XMLHttpRequest::initSend(ExceptionCode& ec)
{
    if (!scriptExecutionContext())
        return false;

    if (m_state != OPENED || m_sendFlag) {
        ec = INVALID_STATE_ERR;
        return false;
    }
    ASSERT(!m_loadingActivity);

    // FIXME: Convert this to check the isolated world's Content Security Policy once webkit.org/b/104520 is solved.
    if (!scriptExecutionContext()->contentSecurityPolicy()->allowConnectToSource(m_url, scriptExecutionContext()->shouldBypassMainWorldContentSecurityPolicy())) {
        if (m_async) {
            m_timeoutTimer.stop();
        	class NetworkErrorTask final : public ActiveDOMObject::KeepingObjectAliveTask<XMLHttpRequest>
        	{
        	public:
        		NetworkErrorTask(XMLHttpRequest& object) : ActiveDOMObject::KeepingObjectAliveTask<XMLHttpRequest>(object) { }

            	virtual void operator()() override
            	{
            		m_protectedObject->networkError();
            	}
        	};
            queueTaskKeepingObjectAlive(*this, TaskSource::Networking, adoptPtr(new NetworkErrorTask(*this)));
        } else
            ec = NETWORK_ERR;
        return false;
    }

    m_error = false;
    return true;
}

void XMLHttpRequest::send(ExceptionCode& ec)
{
    send(String(), ec);
}

void XMLHttpRequest::send(Document& document, ExceptionCode& ec)
{
    if (!initSend(ec))
        return;

    if (m_method != "GET" && m_method != "HEAD" && m_url.protocolIsInHTTPFamily()) {
    	if (!m_requestHeaders.contains(HTTPHeaderName::ContentType)) {
#if ENABLE(DASHBOARD_SUPPORT)
            if (usesDashboardBackwardCompatibilityMode())
                setRequestHeaderInternal(HTTPHeaderName::ContentType, "application/x-www-form-urlencoded");
            else
#endif
                // FIXME: this should include the charset used for encoding.
                setRequestHeaderInternal(HTTPHeaderName::ContentType, document.isHTMLDocument() ? ASCIILiteral("text/html;charset=UTF-8") : ASCIILiteral("application/xml;charset=UTF-8"));
        } else {
            String contentType = m_requestHeaders.get(HTTPHeaderName::ContentType);
            replaceCharsetInMediaTypeIfNeeded(contentType);
            m_requestHeaders.set(HTTPHeaderName::ContentType, contentType);
        }

        // FIXME: According to XMLHttpRequest Level 2, this should use the Document.innerHTML algorithm
        // from the HTML5 specification to serialize the document.
        String body = serializeFragment(document, SerializedNodes::SubtreeIncludingNode);

        m_requestEntityBody = FormData::create(UTF8Encoding().encode(body, EntitiesForUnencodables));
        if (m_upload)
            m_requestEntityBody->setAlwaysStream(true);
    }

    createRequest(ec);
}

void XMLHttpRequest::send(const String& body, ExceptionCode& ec)
{
    if (!initSend(ec))
        return;

    if (!body.isNull() && m_method != "GET" && m_method != "HEAD" && m_url.protocolIsInHTTPFamily()) {
        String contentType = getRequestHeader(HTTPHeaderName::ContentType);
        if (contentType.isNull()) {
#if ENABLE(DASHBOARD_SUPPORT)
            if (usesDashboardBackwardCompatibilityMode())
                setRequestHeaderInternal(HTTPHeaderName::ContentType, "application/x-www-form-urlencoded");
            else
#endif
                setRequestHeaderInternal(HTTPHeaderName::ContentType, HTTPHeaderValues::textPlainContentType());
        } else {
            replaceCharsetInMediaTypeIfNeeded(contentType);
            m_requestHeaders.set(HTTPHeaderName::ContentType, contentType);
        }

        m_requestEntityBody = FormData::create(UTF8Encoding().encode(body, EntitiesForUnencodables));
        if (m_upload)
            m_requestEntityBody->setAlwaysStream(true);
    }

    createRequest(ec);
}

void XMLHttpRequest::send(Blob& body, ExceptionCode& ec)
{
    if (!initSend(ec))
        return;

    if (m_method != "GET" && m_method != "HEAD" && m_url.protocolIsInHTTPFamily()) {
        if (!m_requestHeaders.contains(HTTPHeaderName::ContentType)) {
			const String& blobType = body.type();
			if (!blobType.isEmpty() && isValidContentType(blobType))
				setRequestHeaderInternal(HTTPHeaderName::ContentType, blobType);
			else {
				// From FileAPI spec, whenever media type cannot be determined, empty string must be returned.
				setRequestHeaderInternal(HTTPHeaderName::ContentType, emptyString());
			}
		}

        // FIXME: add support for uploading bundles.
        m_requestEntityBody = FormData::create();
        if (body.isFile())
            m_requestEntityBody->appendFile(toFile(&body)->path());
#if ENABLE(BLOB)
        else
            m_requestEntityBody->appendBlob(body.url());
#endif
    }

    createRequest(ec);
}

void XMLHttpRequest::send(const URLSearchParams& params, ExceptionCode& ec)
{
    if (!m_requestHeaders.contains(HTTPHeaderName::ContentType))
        m_requestHeaders.set(HTTPHeaderName::ContentType, "application/x-www-form-urlencoded;charset=UTF-8");
    send(params.toString(), ec);
}

void XMLHttpRequest::send(DOMFormData& body, ExceptionCode& ec)
{
    if (!initSend(ec))
        return;

    if (m_method != "GET" && m_method != "HEAD" && m_url.protocolIsInHTTPFamily()) {
        m_requestEntityBody = FormData::createMultiPart(static_cast<FormDataList&>(body), body.encoding(), document());

        // We need to ask the client to provide the generated file names if needed. When FormData fills the element
        // for the file, it could set a flag to use the generated file name, i.e. a package file on Mac.
        m_requestEntityBody->generateFiles(document());

        if (!m_requestHeaders.contains(HTTPHeaderName::ContentType))
			setRequestHeaderInternal(ASCIILiteral("Content-Type"), makeString("multipart/form-data; boundary=", m_requestEntityBody->boundary().data()));
	}

    createRequest(ec);
}

void XMLHttpRequest::send(JSC::ArrayBuffer& body, ExceptionCode& ec)
{
    String consoleMessage(ASCIILiteral("ArrayBuffer is deprecated in XMLHttpRequest.send(). Use ArrayBufferView instead."));
    scriptExecutionContext()->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Warning, consoleMessage);

    HistogramSupport::histogramEnumeration("WebCore.XHR.send.ArrayBufferOrView", XMLHttpRequestSendArrayBuffer, XMLHttpRequestSendArrayBufferOrViewMax);

    sendBytesData(body.data(), body.byteLength(), ec);
}

void XMLHttpRequest::send(JSC::ArrayBufferView* body, ExceptionCode& ec)
{
    HistogramSupport::histogramEnumeration("WebCore.XHR.send.ArrayBufferOrView", XMLHttpRequestSendArrayBufferView, XMLHttpRequestSendArrayBufferOrViewMax);

    sendBytesData(body->baseAddress(), body->byteLength(), ec);
}

void XMLHttpRequest::sendBytesData(const void* data, size_t length, ExceptionCode& ec)
{
    if (!initSend(ec))
        return;

    if (m_method != "GET" && m_method != "HEAD" && m_url.protocolIsInHTTPFamily()) {
        m_requestEntityBody = FormData::create(data, length);
        if (m_upload)
            m_requestEntityBody->setAlwaysStream(true);
    }

    createRequest(ec, true);
}

void XMLHttpRequest::createRequest(ExceptionCode& ec, bool isBytesData)
{
#if ENABLE(BLOB)
    // Only GET request is supported for blob URL.
	if (!m_async && m_url.protocolIs("blob") && m_method != "GET") {
        m_blobURLLifetimeExtension.clear();
        ec = NETWORK_ERR;
        return;
    }
#endif

    bool forceAllowed = false;
    ScriptExecutionContext* context = scriptExecutionContext();

    if (context && context->isDocument()){
    	Document* document = static_cast<Document*>(context);
		if(Frame* frame = document->frame()){
			if(!frame->loader()->client()->canLoadXHRRequest(document->url(), m_url, forceAllowed)){
				ec = NETWORK_ERR;
				return;
			}
		}
    }

    if (m_async && m_upload && m_upload->hasEventListeners())
        m_uploadListenerFlag = true;

    ResourceRequest request(m_url);
    request.setHTTPMethod(m_method);
#if PLATFORM(CHROMIUM) || PLATFORM(BLACKBERRY)
    request.setTargetType(ResourceRequest::TargetIsXHR);
#endif

    if (m_requestEntityBody) {
        ASSERT(m_method != "GET");
        ASSERT(m_method != "HEAD");
        request.setHTTPBody(m_requestEntityBody.release());
    }

    if (m_requestHeaders.size() > 0)
        request.addHTTPHeaderFields(m_requestHeaders);

    ThreadableLoaderOptions options;
    options.cachingPolicy = context->shouldCacheXHRResource() ? CachingPolicy::AllowCaching : CachingPolicy::DisallowCaching;
    options.sendLoadCallbacks = SendCallbacks;
    // The presence of upload event listeners forces us to use preflighting because POSTing to an URL that does not
	// permit cross origin requests should look exactly like POSTing to an URL that does not respond at all.
	options.preflightPolicy = m_uploadListenerFlag ? ForcePreflight : ConsiderPreflight;
	options.credentials = m_includeCredentials ? FetchOptions::Credentials::Include : FetchOptions::Credentials::SameOrigin;
	options.mode = forceAllowed ? FetchOptions::Mode::NoCors : FetchOptions::Mode::Cors;
    options.contentSecurityPolicyEnforcement = context->shouldBypassMainWorldContentSecurityPolicy() ? ContentSecurityPolicyEnforcement::DoNotEnforce : ContentSecurityPolicyEnforcement::EnforceConnectSrcDirective;
    options.sameOriginDataURLFlag = SameOriginDataURLFlag::Set;
    options.filteringPolicy = ResponseFilteringPolicy::Enable;
    options.initiator = cachedResourceRequestInitiators().xmlhttprequest;
    options.forceAllowed = forceAllowed;

    if(isBytesData && m_method != "GET" && m_method != "HEAD" && m_url.protocolIsInHTTPFamily()) {
    	if(request.httpContentType().isEmpty()){
    		// check simple cross origin access before adding content-type header
    		if( options.preflightPolicy == ConsiderPreflight && isSimpleCrossOriginAccessRequest(m_method, request.httpHeaderFields()))
    			options.preflightPolicy = PreventPreflight;

    		request.setHTTPContentType("application/octet-stream"); // prevent server parsing form data

    	    if (context){
    	    	String message = tryMakeString("XMLHttpRequest(", m_url.string(), ") is simple cross origin access before adding \"content-type: application/octet-stream\".");
    	    	context->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Warning, message);
    	    }
    	}
    }

    if (m_timeoutMilliseconds) {
        if (!m_async)
            request.setTimeoutInterval(m_timeoutMilliseconds / 1000.0);
        else {
        	request.setTimeoutInterval(std::numeric_limits<double>::infinity());
            m_sendingTime = WTF::monotonicallyIncreasingTime();
            m_timeoutTimer.startOneShot(m_timeoutMilliseconds / 1000.0);
        }
    }

    m_exceptionCode = 0;
    m_error = false;
    m_uploadComplete = !request.httpBody();
    m_sendFlag = true;

    if (m_async) {
        m_progressEventThrottle.dispatchProgressEvent(eventNames().loadstartEvent);
        if (!m_uploadComplete && m_uploadListenerFlag)
            m_upload->dispatchProgressEvent(eventNames().loadstartEvent, 0, request.httpBody()->lengthInBytes());

        if (m_state != OPENED || !m_sendFlag || m_loadingActivity)
            return;

    	request.setPriority(ResourceLoadPriorityHigh);

        if (m_upload)
            request.setReportUploadProgress(true);

        // ThreadableLoader::create can return null here, for example if we're no longer attached to a page or if a content blocker blocks the load.
        // This is true while running onunload handlers.
        // FIXME: Maybe we need to be able to send XMLHttpRequests from onunload, <http://bugs.webkit.org/show_bug.cgi?id=10904>.
        RefPtr<ThreadableLoader> loader = ThreadableLoader::create(context, this, request, options);
        if (loader)
			m_loadingActivity = LoadingActivity { Ref<XMLHttpRequest>(*this), loader.releaseNonNull() };

		// Either loader is null or some error was synchronously sent to us.
		ASSERT(m_loadingActivity || !m_sendFlag);
    } else {
        InspectorInstrumentation::willLoadXHRSynchronously(context);
        ThreadableLoader::loadResourceSynchronously(context, request, *this, options);
        InspectorInstrumentation::didLoadXHRSynchronously(context);
    }

    if (!m_exceptionCode && m_error)
        m_exceptionCode = NETWORK_ERR;
    ec = m_exceptionCode;
}

void XMLHttpRequest::abort()
{
    RefPtr<XMLHttpRequest> protect(this);

    m_wasAbortedByClient = true;
    if (!internalAbort())
        return;

    clearResponseBuffers();

    m_requestHeaders.clear();

    if ((m_state == OPENED && m_sendFlag) || m_state == HEADERS_RECEIVED || m_state == LOADING) {
		ASSERT(!m_loadingActivity);
		m_sendFlag = false;
		changeState(DONE);
		dispatchErrorEvents(eventNames().abortEvent);
	}
    if (m_state == DONE)
    	m_state = UNSENT;
}

bool XMLHttpRequest::internalAbort()
{
    m_error = true;

    // FIXME: when we add the support for multi-part XHR, we will have to think be careful with this initialization.
    m_receivedLength = 0;

    m_decoder = nullptr;

    m_timeoutTimer.stop();

    if (!m_loadingActivity)
        return true;

    // Cancelling m_loadingActivity may trigger a window.onload callback which can call open() on the same xhr.
    // This would create internalAbort reentrant call.
    // m_loadingActivity is set to WTF::nullopt before being cancelled to exit early in any reentrant internalAbort() call.
    Optional<LoadingActivity> loadingActivity = std::exchange(m_loadingActivity, Nullopt);
    loadingActivity->loader->cancel();

    // If window.onload callback calls open() and send() on the same xhr, m_loadingActivity is now set to a new value.
    // The function calling internalAbort() should abort to let the open() and send() calls continue properly.
    // We ask the function calling internalAbort() to exit by returning false.
    // Save this information to a local variable since we are going to drop protection.
    bool newLoadStarted = !!m_loadingActivity;

    return !newLoadStarted;
}

void XMLHttpRequest::clearResponse()
{
    m_response = ResourceResponse();
    clearResponseBuffers();
}

void XMLHttpRequest::clearResponseBuffers()
{
    m_responseBuilder.clear();
    m_responseEncoding = String();
    m_createdDocument = false;
    m_responseDocument = nullptr;
    m_responseBlob = nullptr;
    m_binaryResponseBuilder.clear();
    m_responseArrayBuffer.clear();
    m_responseCacheIsValid = false;
}

void XMLHttpRequest::clearRequest()
{
    m_requestHeaders.clear();
    m_requestEntityBody = nullptr;
    m_url = KURL();
    m_blobURLLifetimeExtension.clear();
}

void XMLHttpRequest::genericError()
{
    clearResponse();
    clearRequest();
    m_sendFlag = false;
    m_error = true;

    changeState(DONE);
}

void XMLHttpRequest::networkError()
{
    genericError();
    dispatchErrorEvents(eventNames().errorEvent);
    internalAbort();
}

void XMLHttpRequest::abortError()
{
    ASSERT(m_wasAbortedByClient);
    genericError();
    dispatchErrorEvents(eventNames().abortEvent);
}

size_t XMLHttpRequest::memoryCost() const
{
    if (readyState() == DONE)
        return m_responseBuilder.length() * 2;
    return 0;
}

void XMLHttpRequest::overrideMimeType(const String& mimeType, ExceptionCode& ec)
{
    if (readyState() == LOADING || readyState() == DONE) {
    	ec = INVALID_STATE_ERR;
    	return;
    }

    m_mimeTypeOverride = "application/octet-stream";
    if (isValidContentType(mimeType))
        m_mimeTypeOverride = mimeType;
}

void XMLHttpRequest::setRequestHeader(const String& name, const String& value, ExceptionCode& ec)
{
    if (m_state != OPENED || m_sendFlag) {
#if ENABLE(DASHBOARD_SUPPORT)
        if (usesDashboardBackwardCompatibilityMode())
            return;
#endif

        ec = INVALID_STATE_ERR;
        return;
    }

    String normalizedValue = stripLeadingAndTrailingHTTPSpaces(value);
    if (!isValidHTTPToken(name) || !isValidHTTPHeaderValue(normalizedValue)) {
        ec = SYNTAX_ERR;
        return;
    }

    bool allowUnsafeHeaderField = false;
#if ENABLE(DASHBOARD_SUPPORT)
    allowUnsafeHeaderField = usesDashboardBackwardCompatibilityMode();
#endif
    if (!allowUnsafeHeaderField && isForbiddenHeaderName(name)) {
        logConsoleError(scriptExecutionContext(), tryMakeString("Refused to set unsafe header \"", name, "\""));
        return;
    }

    setRequestHeaderInternal(name, normalizedValue);
}

void XMLHttpRequest::setRequestHeaderInternal(const String& name, const String& value)
{
    m_requestHeaders.add(name, value);
}

void XMLHttpRequest::setRequestHeaderInternal(HTTPHeaderName name, const String& value)
{
    m_requestHeaders.add(name, value);
}

String XMLHttpRequest::getRequestHeader(const String& name) const
{
    return m_requestHeaders.get(name);
}

String XMLHttpRequest::getRequestHeader(HTTPHeaderName name) const
{
    return m_requestHeaders.get(name);
}

String XMLHttpRequest::getAllResponseHeaders() const
{
    if (m_state < HEADERS_RECEIVED || m_error)
        return emptyString();

    if (!m_allResponseHeaders) {
        Vector<std::pair<String, String>> headers;
        headers.reserveInitialCapacity(m_response.httpHeaderFields().size());

        for (auto& header : m_response.httpHeaderFields())
            headers.uncheckedAppend(std::make_pair(header.first, header.second));

        std::sort(headers.begin(), headers.end(), [] (const std::pair<String, String>& x, const std::pair<String, String>& y) {
            unsigned xLength = x.first.length();
            unsigned yLength = y.first.length();
            unsigned commonLength = std::min(xLength, yLength);
            for (unsigned i = 0; i < commonLength; ++i) {
                auto xCharacter = toASCIIUpper(x.first[i]);
                auto yCharacter = toASCIIUpper(y.first[i]);
                if (xCharacter != yCharacter)
                    return xCharacter < yCharacter;
            }
            return xLength < yLength;
        });

        StringBuilder stringBuilder;
        for (auto& header : headers) {
            stringBuilder.append(header.first.lower());
            stringBuilder.append(": ");
            stringBuilder.append(header.second);
            stringBuilder.append("\r\n");
        }

        m_allResponseHeaders = stringBuilder.toString();
    }

    return m_allResponseHeaders;
}

String XMLHttpRequest::getResponseHeader(const String& name) const
{
    if (m_state < HEADERS_RECEIVED || m_error)
        return String();

    return m_response.httpHeaderField(name);
}

String XMLHttpRequest::responseMIMEType(FinalMIMEType finalMIMEType) const
{
    String contentType = m_mimeTypeOverride;
    if (contentType.isEmpty()) {
        // Same logic as externalEntityMimeTypeAllowed() in XMLDocumentParserLibxml2.cpp. Keep them in sync.
        if (m_response.isInHTTPFamily())
            contentType = m_response.httpHeaderField(HTTPHeaderName::ContentType);
        else
            contentType = m_response.mimeType();
    }
    if (auto parsedContentType = ParsedContentType::create(contentType))
        return finalMIMEType == FinalMIMEType::Yes ? parsedContentType->serialize() : parsedContentType->mimeType();
    return "text/xml";
}

int XMLHttpRequest::status() const
{
    if (m_state == UNSENT || m_state == OPENED || m_error)
        return 0;

    return m_response.httpStatusCode();
}

String XMLHttpRequest::statusText() const
{
    if (m_state == UNSENT || m_state == OPENED || m_error)
        return String();

    return m_response.httpStatusText();
}

static String resourceErrorTypeString(ResourceErrorBase::Type type)
{
	switch(type) {
	case ResourceErrorBase::Type::Null:
		return emptyString();
	case ResourceErrorBase::Type::General:
		return "General";
	case ResourceErrorBase::Type::AccessControl:
		return "AccessControl";
	case ResourceErrorBase::Type::Cancellation:
		return "Cancellation";
	case ResourceErrorBase::Type::Timeout:
		return "Timeout";
	case ResourceErrorBase::Type::Http:
		return "Http";
	default:
		return emptyString();
	}
}
void XMLHttpRequest::didFail(const ResourceError& error)
{
    RefPtr<XMLHttpRequest> protect(this);

    // If we are already in an error state, for instance we called abort(), bail out early.
    if (m_error)
        return;

    // The XHR specification says we should only fire an abort event if the cancelation was requested by the client.
    if (m_wasAbortedByClient && error.isCancellation()) {
		String message = tryMakeString("XMLHttpRequest cannot load ", error.failingURL(), " due to abort.");
		logConsoleError(scriptExecutionContext(), message);
        m_exceptionCode = ABORT_ERR;
        abortError();
        return;
    }

    // In case of worker sync timeouts.
    if (error.isTimeout()) {
		String message = tryMakeString("XMLHttpRequest cannot load ", error.failingURL(), " due to timeout.");
		logConsoleError(scriptExecutionContext(), message);
    	didReachTimeout();
        return;
    }

    // In case didFail is called synchronously on an asynchronous XHR call, let's dispatch network error asynchronously
    if (m_async && m_sendFlag && !m_loadingActivity) {
        m_sendFlag = false;
        m_timeoutTimer.stop();
    	class NetworkErrorTask final : public ActiveDOMObject::KeepingObjectAliveTask<XMLHttpRequest>
    	{
    	public:
    		NetworkErrorTask(XMLHttpRequest& object) : ActiveDOMObject::KeepingObjectAliveTask<XMLHttpRequest>(object) { }

        	virtual void operator()() override
        	{
        		m_protectedObject->networkError();
        	}
    	};
        queueTaskKeepingObjectAlive(*this, TaskSource::Networking, adoptPtr(new NetworkErrorTask(*this)));
        return;
    }
    m_exceptionCode = NETWORK_ERR;
    networkError();
}

void XMLHttpRequest::didFinishLoading(unsigned long identifier, double)
{
    RefPtr<XMLHttpRequest> protect(this);

    if (m_error)
        return;

    // Make sure that didSendData() was called at least one before marking the load as complete
    // so that a progress events get fired on m_upload.
    if (m_uploadListenerFlag && m_requestEntityBody && !m_wasDidSendDataCalledForTotalBytes) {
        auto bodyLength = m_requestEntityBody->lengthInBytes();
        didSendData(bodyLength, bodyLength);
    }

    if (m_state < HEADERS_RECEIVED)
        changeState(HEADERS_RECEIVED);

    if (m_decoder)
        m_responseBuilder.append(m_decoder->flush());

    m_responseBuilder.shrinkToFit();

    InspectorInstrumentation::resourceRetrievedByXMLHttpRequest(scriptExecutionContext(), identifier, m_responseBuilder.toStringPreserveCapacity(), m_url, m_lastSendURL, m_lastSendLineNumber);

    m_loadingActivity = WTF::Nullopt;
    m_url = KURL();
    m_blobURLLifetimeExtension.clear();

    m_sendFlag = false;
    changeState(DONE);
    m_responseEncoding = String();
    m_decoder = 0;

    m_timeoutTimer.stop();
}

void XMLHttpRequest::didSendData(unsigned long long bytesSent, unsigned long long totalBytesToBeSent)
{
    if (!m_upload)
        return;

    if (m_uploadListenerFlag)
        m_upload->dispatchProgressEvent(eventNames().progressEvent, bytesSent, totalBytesToBeSent);

    if (bytesSent == totalBytesToBeSent && !m_uploadComplete) {
    	m_wasDidSendDataCalledForTotalBytes = true;
        m_uploadComplete = true;
        if (m_uploadListenerFlag) {
            m_upload->dispatchProgressEvent(eventNames().loadEvent, bytesSent, totalBytesToBeSent);
            m_upload->dispatchProgressEvent(eventNames().loadendEvent, bytesSent, totalBytesToBeSent);
        }
    }
}

void XMLHttpRequest::didReceiveResponse(unsigned long identifier, const ResourceResponse& response)
{
    InspectorInstrumentation::didReceiveXHRResponse(scriptExecutionContext(), identifier);

    m_response = response;
}

static inline bool shouldDecodeResponse(XMLHttpRequest::ResponseType type)
{
    switch (type) {
    case XMLHttpRequest::ResponseType::EmptyString:
    case XMLHttpRequest::ResponseType::Document:
    case XMLHttpRequest::ResponseType::Json:
    case XMLHttpRequest::ResponseType::Text:
        return true;
    case XMLHttpRequest::ResponseType::Arraybuffer:
    case XMLHttpRequest::ResponseType::Blob:
        return false;
    }
    ASSERT_NOT_REACHED();
    return true;
}

// https://xhr.spec.whatwg.org/#final-charset
TextEncoding XMLHttpRequest::finalResponseCharset() const
{
    String label = m_responseEncoding;

    String overrideResponseCharset = extractCharsetFromMediaType(m_mimeTypeOverride);
    if (!overrideResponseCharset.isEmpty())
        label = overrideResponseCharset;

    return TextEncoding(label);
}

PassRefPtr<TextResourceDecoder> XMLHttpRequest::createDecoder() const
{
    TextEncoding finalResponseCharset = this->finalResponseCharset();
    if (finalResponseCharset.isValid())
        return TextResourceDecoder::create("text/plain", finalResponseCharset);

    switch (m_responseType) {
    case ResponseType::EmptyString:
        if (MIMETypeRegistry::isXMLMIMEType(responseMIMEType())) {
            auto decoder = TextResourceDecoder::create("application/xml");
            // Don't stop on encoding errors, unlike it is done for other kinds of XML resources. This matches the behavior of previous WebKit versions, Firefox and Opera.
            decoder->useLenientXMLDecoding();
            return decoder;
        }
        FALLTHROUGH;
    case ResponseType::Text:
    case ResponseType::Json: {
        auto decoder = TextResourceDecoder::create("text/plain", "UTF-8");
        if (responseType() == ResponseType::Json)
            decoder->setAlwaysUseUTF8();
        return decoder;
    }
    case ResponseType::Document: {
        if (equalLettersIgnoringASCIICase(responseMIMEType(), "text/html"))
            return TextResourceDecoder::create("text/html", "UTF-8");
        auto decoder = TextResourceDecoder::create("application/xml");
        // Don't stop on encoding errors, unlike it is done for other kinds of XML resources. This matches the behavior of previous WebKit versions, Firefox and Opera.
        decoder->useLenientXMLDecoding();
        return decoder;
    }
    case ResponseType::Arraybuffer:
    case ResponseType::Blob:
        ASSERT_NOT_REACHED();
        break;
    }
    return TextResourceDecoder::create("text/plain", "UTF-8");
}

void XMLHttpRequest::didReceiveData(const char* data, int len)
{
    if (m_error)
        return;

    if (m_state < HEADERS_RECEIVED)
        changeState(HEADERS_RECEIVED);

    if (!m_mimeTypeOverride.isEmpty())
		m_responseEncoding = extractCharsetFromMediaType(m_mimeTypeOverride);
	if (m_responseEncoding.isEmpty())
		m_responseEncoding = m_response.textEncodingName();

    bool useDecoder = shouldDecodeResponse(m_responseType);

    if (useDecoder && !m_decoder)
        m_decoder = createDecoder();

    if (!len)
        return;

    if (len == -1)
        len = strlen(data);

    if (useDecoder)
        m_responseBuilder.append(m_decoder->decode(data, len));
    else {
        // Buffer binary data.
        if (!m_binaryResponseBuilder){
            m_binaryResponseBuilder = SharedBuffer::create();
            m_binaryResponseBuilder->setExpectedLength(m_response.expectedContentLength());

            if(m_response.expectedContentLength() >= (8 * 1024 * 1024)) {
            	m_scriptExecutionContext->addConsoleMessage(MessageSource::JS, MessageType::Log, MessageLevel::Debug,
					tryMakeString("XMLHttpRequest created SharedBuffer with expected length ", String::number(m_response.expectedContentLength()), " ", m_url.string()));
            }
        }

        if(!m_binaryResponseBuilder->tryAppend(data, len)){
            m_exceptionCode = NETWORK_ERR;
            networkError();
            return;
        }
    }

    if (!m_error) {
        m_receivedLength += len;

        if (readyState() != LOADING)
            changeState(LOADING);
        else {
            // Firefox calls readyStateChanged every time it receives data, 4449442
            callReadyStateChangeListener();
        }

		long long expectedLength = m_response.expectedContentLength();
		bool lengthComputable = expectedLength > 0 && m_receivedLength <= expectedLength;
		unsigned long long total = lengthComputable ? expectedLength : 0;
		m_progressEventThrottle.updateProgress(m_async, lengthComputable, m_receivedLength, total);
	}
}

void XMLHttpRequest::dispatchErrorEvents(const AtomicString& type)
{
    auto& eventNames = WebCore::eventNames();
    if (!m_uploadComplete) {
        m_uploadComplete = true;
        if (m_upload && m_uploadListenerFlag) {
            m_upload->dispatchProgressEvent(type, 0, 0);
            m_upload->dispatchProgressEvent(eventNames.loadendEvent, 0, 0);
        }
    }
    m_progressEventThrottle.dispatchProgressEvent(type);
    m_progressEventThrottle.dispatchProgressEvent(eventNames.loadendEvent);
}

void XMLHttpRequest::didReachTimeout()
{
    RefPtr<XMLHttpRequest> protect(this);
    if (!internalAbort())
    	return;

    clearResponse();
    clearRequest();

    m_sendFlag = false;
    m_error = true;
    m_exceptionCode = TIMEOUT_ERR;

    if (!m_async) {
        m_state = DONE;
        m_exceptionCode = TIMEOUT_ERR;
        return;
    }

    changeState(DONE);

    dispatchErrorEvents(eventNames().timeoutEvent);
}

void XMLHttpRequest::timeoutTimerFired()
{
	didReachTimeout();
}

void XMLHttpRequest::suspend(ReasonForSuspension)
{
    m_progressEventThrottle.suspend();
}

void XMLHttpRequest::resume()
{
    m_progressEventThrottle.resume();
}

void XMLHttpRequest::stop()
{
    internalAbort();
}

void XMLHttpRequest::contextDestroyed()
{
    ASSERT(!m_loadingActivity);
    ActiveDOMObject::contextDestroyed();
}

void XMLHttpRequest::updateHasRelevantEventListener()
{
	auto& eventNames = WebCore::eventNames();
    m_hasRelevantEventListener = hasEventListeners(eventNames.abortEvent)
        || hasEventListeners(eventNames.errorEvent)
        || hasEventListeners(eventNames.loadEvent)
        || hasEventListeners(eventNames.loadendEvent)
        || hasEventListeners(eventNames.progressEvent)
        || hasEventListeners(eventNames.readystatechangeEvent)
        || hasEventListeners(eventNames.timeoutEvent)
        || (m_upload && m_upload->hasRelevantEventListener());
}

void XMLHttpRequest::eventListenersDidChange()
{
    updateHasRelevantEventListener();
}

// An XMLHttpRequest object must not be garbage collected if its state is either opened with the send() flag set, headers received, or loading, and
// it has one or more event listeners registered whose type is one of readystatechange, progress, abort, error, load, timeout, and loadend.
bool XMLHttpRequest::virtualHasPendingActivity() const
{
    if (!m_hasRelevantEventListener)
        return false;

    switch (readyState()) {
    case OPENED:
        return m_sendFlag;
    case HEADERS_RECEIVED:
    case LOADING:
        return true;
    case UNSENT:
    case DONE:
        break;
    }
    return false;
}

EventTargetInterface XMLHttpRequest::eventTargetInterface() const
{
    return XMLHttpRequestEventTargetInterfaceType;
}

ScriptExecutionContext* XMLHttpRequest::scriptExecutionContext() const
{
    return ActiveDOMObject::scriptExecutionContext();
}

} // namespace WebCore
