/*
 * This file is part of the XSL implementation.
 *
 * Copyright (C) 2004, 2005, 2006, 2008 Apple Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "XSLImportRule.h"

#if ENABLE(XSLT)

#include "CachedResourceLoader.h"
#include "Document.h"
#include "Element.h"
#include "TextResourceDecoder.h"
#include "ThreadableLoader.h"
#include "XSLStyleSheet.h"

namespace WebCore {

XSLImportRule::XSLImportRule(XSLStyleSheet& parent, const String& href)
    : m_parentStyleSheet(makeWeakPtr(parent))
    , m_strHref(href)
	, m_decoder(TextResourceDecoder::create("text/xsl"))
{
}

XSLImportRule::~XSLImportRule()
{
    if (m_styleSheet)
        m_styleSheet->setParentStyleSheet(nullptr);
}

void XSLImportRule::setXSLStyleSheet(const String& href, const KURL& baseURL, const String& sheet)
{
    if (m_styleSheet)
        m_styleSheet->setParentStyleSheet(nullptr);

    // FIXME: parentStyleSheet() should never be null here.
    RefPtr<XSLStyleSheet> parent = parentStyleSheet();
    m_styleSheet = XSLStyleSheet::create(parent.get(), href, baseURL);

    m_styleSheet->parseString(sheet);
    m_loading = false;
    
    if (parent)
        parent->checkLoaded();
}

bool XSLImportRule::isLoading()
{
    return (m_loading || (m_styleSheet && m_styleSheet->isLoading()));
}

void XSLImportRule::didReceiveResponse(unsigned long identifier, const ResourceResponse& resourceResponse)
{
	m_resourceResponse = resourceResponse;
    String encoding = resourceResponse.textEncodingName();
    if (!encoding.isNull())
        m_decoder->setEncoding(encoding, TextResourceDecoder::EncodingFromHTTPHeader);
}

void XSLImportRule::didReceiveData(const char* data, int dataLength)
{
    m_responseBuilder.append(m_decoder->decode(data, dataLength));
}

void XSLImportRule::didFinishLoading(unsigned long identifier, double finishTime)
{
    m_responseBuilder.append(m_decoder->flush());
    m_responseBuilder.shrinkToFit();
}

void XSLImportRule::didFail(const ResourceError& resourceError)
{
	m_responseBuilder.clear();
	m_resourceError = resourceError;
}

void XSLImportRule::loadSheet()
{
	RefPtr<XSLStyleSheet> rootSheet = parentStyleSheet();
	while (auto* parentSheet = rootSheet->parentStyleSheet())
		rootSheet = parentSheet;

	RefPtr<Document> document = rootSheet->ownerDocument();
	if (!document)
		return;
    
    String absHref = m_strHref;
    RefPtr<XSLStyleSheet> parentSheet = parentStyleSheet();
    if (!parentSheet->baseURL().isNull()) {
        // use parent styleheet's URL as the base URL
        absHref = KURL(parentSheet->baseURL(), m_strHref).string();
    }
    
    // Check for a cycle in our import chain.  If we encounter a stylesheet
    // in our parent chain with the same URL, then just bail.
    for (RefPtr<XSLStyleSheet> parentSheet = parentStyleSheet(); parentSheet; parentSheet = parentSheet->parentStyleSheet()) {
        if (absHref == parentSheet->baseURL())
            return;
    }

	m_responseBuilder.clear();
    m_resourceError = ResourceError();

    ThreadableLoaderOptions options(CachedResourceLoader::defaultCachedResourceOptions(), ConsiderPreflight, ContentSecurityPolicyEnforcement::EnforceConnectSrcDirective, nullAtom, ResponseFilteringPolicy::Disable);
    options.mode = FetchOptions::Mode::SameOrigin;
    ResourceRequest resourceRequest(document->completeURL(absHref));

    // XSLStyleSheet::locateStylesheetSubResource needs it.
    ThreadableLoader::loadResourceSynchronously(document->scriptExecutionContext(), resourceRequest, *this, options);
    
    if (m_resourceError.isNull()) {
        setXSLStyleSheet(resourceRequest.url(), m_resourceResponse.url(), m_responseBuilder.toString());
        
        // If the imported sheet is in the cache, then setXSLStyleSheet gets called,
        // and the sheet even gets parsed (via parseString).  In this case we have
        // loaded (even if our subresources haven't), so if we have a stylesheet after
        // checking the cache, then we've clearly loaded.
        if (!m_styleSheet)
            m_loading = true;
    }
}

} // namespace WebCore

#endif // ENABLE(XSLT)
