/*
 * Copyright (C) 2008, 2009, 2012 Apple Inc. All rights reserved.
 * Copyright (C) 2008 Cameron Zwarich <cwzwarich@uwaterloo.ca>
 * Copyright (C) 2012 Igalia, S.L.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "BytecodeGenerator.h"

#include "ArithProfile.h"
#include "BatchedTransitionOptimizer.h"
#include "BuiltinExecutables.h"
#include "BuiltinNames.h"
#include "BytecodeGeneratorification.h"
#include "BytecodeUseDef.h"
#include "CatchScope.h"
#include "ExceptionHelpers.h"
#include "DefinePropertyAttributes.h"
#include "Interpreter.h"
#include "JSAsyncGeneratorFunction.h"
#include "JSBigInt.h"
#include "JSFixedArray.h"
#include "JSLexicalEnvironment.h"
#include "JSFunction.h"
#include "JSGeneratorFunction.h"
#include "JSTemplateObjectDescriptor.h"
#include "LowLevelInterpreter.h"
#include "Operations.h"
#include "Options.h"
#include "PrivateFieldPutKind.h"
#include "StackAlignment.h"
#include "StrongInlines.h"
#include <wtf/BitVector.h>
#include <wtf/CommaPrinter.h>
#include <wtf/HashSet.h>
#include <wtf/MemoryMonitor.h>
#include <wtf/SmallPtrSet.h>
#include <wtf/text/WTFString.h>

using namespace std;

namespace JSC {

template<typename T>
static inline void shrinkToFit(T& segmentedVector)
{
    while (segmentedVector.size() && !segmentedVector.last().refCount())
        segmentedVector.removeLast();
}

void Label::setLocation(BytecodeGenerator& generator, unsigned location)
{
    m_location = location;
    
    unsigned size = m_unresolvedJumps.size();
    for (unsigned i = 0; i < size; ++i)
        generator.m_instructions[m_unresolvedJumps[i].second].u.operand = m_location - m_unresolvedJumps[i].first;
}

static bool s_dumpsGeneratedCode = false;

void BytecodeGenerator::setDumpsGeneratedCode(bool dumpsGeneratedCode)
{
    s_dumpsGeneratedCode = dumpsGeneratedCode;
}

bool BytecodeGenerator::dumpsGeneratedCode()
{
    return s_dumpsGeneratedCode;
}

FinallyContext::FinallyContext(BytecodeGenerator& generator, Label* finallyLabel)
    : m_outerContext(generator.m_currentFinallyContext)
    , m_finallyLabel(finallyLabel)
	, m_handlesReturns(false)
{
    ASSERT(m_jumps.isEmpty());
    m_completionRecord.typeRegister = generator.newTemporary();
    m_completionRecord.valueRegister = generator.newTemporary();
    generator.emitLoad(completionTypeRegister(), CompletionType::Normal);
    generator.moveEmptyValue(completionValueRegister());
}

ParserError BytecodeGenerator::generate(ExecState* exec, JSScope* scope)
{
    SamplingRegion samplingRegion("Bytecode Generation");
    
    m_codeBlock->setThisRegister(m_thisRegister.index());

    // If we have declared a variable named "arguments" and we are using arguments then we should
    // perform that assignment now.
    if (m_needToInitializeArguments)
        initializeVariable(variable(propertyNames().arguments), m_argumentsRegister);

    if(m_functionsToInitialize.size()) {
        bool shouldHoistInEval = m_codeType == EvalCode && !m_ecmaMode.isStrict();
        RefPtr<RegisterID> topLevelScope;
        for (auto functionPair : m_functionsToInitialize) {
        	FunctionMetadataNode* metadata = functionPair.first;
            FunctionVariableType functionType = functionPair.second;
            if (functionType == NormalFunctionVariable) {
                Variable var = variable(metadata->ident());
                if (RegisterID* local = var.local())
                    emitNewFunction(local, metadata);
                else {
                    RefPtr<RegisterID> temp = newTemporary();
                    emitNewFunction(temp.get(), metadata);
                    initializeVariable(var, temp.get());
                }
            } else if (functionType == TopLevelFunctionVariable) {
                RefPtr<RegisterID> temp = newTemporary();
                emitNewFunction(temp.get(), metadata);
                if (!topLevelScope) {
                    // We know this will resolve to the top level scope or global object because our parser/global initialization code
                    // doesn't allow let/const/class variables to have the same names as functions.
                    // This is a top level function, and it's an error to ever create a top level function
                    // name that would resolve to a lexical variable. E.g:
                    // ```
                    //     function f() {
                    //         {
                    //             let x;
                    //             {
                    //             //// error thrown here
                    //                  eval("function x(){}");
                    //             }
                    //         }
                    //     }
                    // ```
                    // Therefore, we're guaranteed to have this resolve to a top level variable.
                	RefPtr<RegisterID> topLevelObjectScope = shouldHoistInEval
						? emitResolveScopeForHoistingFuncDeclInEval(nullptr, metadata->ident())
						: emitResolveScope(nullptr, Variable(metadata->ident()));

					topLevelScope = newBlockScopeVariable();
					move(topLevelScope.get(), topLevelObjectScope.get());
				}
                if (shouldHoistInEval)
					emitPutToScopeDynamic(topLevelScope.get(), metadata->ident(), temp.get(), ThrowIfNotFound, InitializationMode::NotInitialization);
				else
					emitPutToScope(topLevelScope.get(), Variable(metadata->ident()), temp.get(), ThrowIfNotFound, InitializationMode::NotInitialization);
            } else
                RELEASE_ASSERT_NOT_REACHED();
        }
    }

    bool callingClassConstructor = constructorKind() != ConstructorKind::None && !isConstructor();
    if (!callingClassConstructor)
    	m_scopeNode->emitBytecode(*this);
    else {
        // At this point we would have emitted an unconditional throw followed by some nonsense that's
        // just an artifact of how this generator is structured. That code never runs, but it confuses
        // bytecode analyses because it constitutes an unterminated basic block. So, we terminate the
        // basic block the strongest way possible.
        emitUnreachable();
    }

    for (auto& handler : m_exceptionHandlersToEmit) {
        RefPtr<Label> realCatchTarget = newEmittedLabel();
        TryData* tryData = handler.tryData;

        emitOpcode(op_catch);
        instructions().uncheckedAppend(handler.exceptionRegister.offset());
        instructions().uncheckedAppend(handler.thrownValueRegister.offset());
        instructions().uncheckedAppend(0);
        if (handler.completionTypeRegister.isValid()) {
            RegisterID completionTypeRegister(handler.completionTypeRegister.offset());
            CompletionType completionType =
                tryData->handlerType == HandlerType::Finally || tryData->handlerType == HandlerType::SynthesizedFinally
                ? CompletionType::Throw
                : CompletionType::Normal;
            emitLoad(&completionTypeRegister, completionType);
        }

        emitJump(tryData->target.get());
        tryData->target = realCatchTarget.release();
    }

    m_staticPropertyAnalyzer.kill();

    for (auto& range : m_tryRanges) {
        int start = range.start->bind();
        int end = range.end->bind();

        // This will happen for empty try blocks and for some cases of finally blocks:
        //
        // try {
        //    try {
        //    } finally {
        //        return 42;
        //        // *HERE*
        //    }
        // } finally {
        //    print("things");
        // }
        //
        // The return will pop scopes to execute the outer finally block. But this includes
        // popping the try context for the inner try. The try context is live in the fall-through
        // part of the finally block not because we will emit a handler that overlaps the finally,
        // but because we haven't yet had a chance to plant the catch target. Then when we finish
        // emitting code for the outer finally block, we repush the try contex, this time with a
        // new start index. But that means that the start index for the try range corresponding
        // to the inner-finally-following-the-return (marked as "*HERE*" above) will be greater
        // than the end index of the try block. This is harmless since end < start handlers will
        // never get matched in our logic, but we do the runtime a favor and choose to not emit
        // such handlers at all.
        if (end <= start)
            continue;

        ASSERT(range.tryData->targetScopeDepth != UINT_MAX);
        HandlerInfo info = {
			static_cast<uint32_t>(start), static_cast<uint32_t>(end),
			static_cast<uint32_t>(range.tryData->target->bind()), range.tryData->handlerType,
            CodeLocationLabel()
        };
        m_codeBlock->addExceptionHandler(info);
    }

    if (isGeneratorOrAsyncFunctionBodyParseMode(m_codeBlock->parseMode()))
        performGeneratorification(*this, m_codeBlock, m_instructions, m_generatorFrameSymbolTable.get(), m_generatorFrameSymbolTableIndex);

	if(UNLIKELY(lowMemory()))
        return ParserError(ParserError::OutOfMemory);

    m_codeBlock->instructions() = RefCountedArray<Instruction>(m_instructions);

	if(UNLIKELY(lowMemory()))
        return ParserError(ParserError::OutOfMemory);

    if(!m_codeBlock->handlePostGenerate(scope, m_linkTimeConstantCodeBlockConstantIndexes, m_arrayAllocationProfileCount, m_objectAllocationProfileCount))
        return ParserError(ParserError::OutOfMemory);

    m_codeBlock->shrinkToFit(CodeBlock::EarlyShrink);

    if (m_expressionTooDeep)
        return ParserError(ParserError::OutOfMemory);
    return ParserError(ParserError::ErrorNone);
}

BytecodeGenerator::BytecodeGenerator(ProgramNode* programNode, JSGlobalObject* globalObject, ProgramCodeBlock* codeBlock, const RefPtr<TDZEnvironmentLink>& parentScopeTDZVariables, const PrivateNameEnvironment*, const Vector<Identifier>*, ECMAMode ecmaMode)
: m_shouldEmitDebugHooks(globalObject->debugger() || Options::shouldEmitDebugHooks())
, m_shouldEmitProfileHooks(globalObject->globalObjectMethodTable()->supportsProfiling(globalObject))
, m_shouldEmitRichSourceInfo(globalObject->globalObjectMethodTable()->supportsRichSourceInfo(globalObject))
, m_scopeNode(programNode)
, m_globalObject(globalObject)
, m_codeBlock(codeBlock)
, m_thisRegister(CallFrame::thisArgumentOffset())
, m_codeType(GlobalCode)
, m_staticPropertyAnalyzer(&m_instructions)
, m_vm(&globalObject->vm())
, m_stack(wtfThreadData().stack())
, m_needsToUpdateArrowFunctionContext(programNode->usesArrowFunction() || programNode->usesEval())
, m_ecmaMode(ecmaMode)
{
    ASSERT_UNUSED(parentScopeTDZVariables, !parentScopeTDZVariables);

    m_linkTimeConstantRegisters.fill(nullptr);

    m_codeBlock->setNumParameters(1); // Allocate space for "this"

    emitOpcode(op_enter);

    allocateAndEmitScope();
    
    for (auto* function : programNode->functionStack())
        m_functionsToInitialize.append(std::make_pair(function, TopLevelFunctionVariable));

    //codeBlock->setVariableDeclarations(programNode->varDeclarations());
    //codeBlock->setLexicalDeclarations(programNode->lexicalVariables());
    // Even though this program may have lexical variables that go under TDZ, when linking the get_from_scope/put_to_scope
    // operations we emit we will have ResolveTypes that implictly do TDZ checks. Therefore, we don't need
    // additional TDZ checks on top of those. This is why we can omit pushing programNode->lexicalVariables()
    // to the TDZ stack.

    if (needsToUpdateArrowFunctionContext()) {
        initializeArrowFunctionContextScopeIfNeeded();
        emitPutThisToArrowFunctionContextScope();
    }
}

BytecodeGenerator::BytecodeGenerator(FunctionNode* functionNode, JSGlobalObject* globalObject, CodeBlock* codeBlock, const RefPtr<TDZEnvironmentLink>& parentScopeTDZVariables, const PrivateNameEnvironment* parentPrivateNameEnvironment, const Vector<Identifier>* generatorOrAsyncWrapperFunctionParameterNames, ECMAMode ecmaMode)
	: m_shouldEmitDebugHooks(globalObject->debugger() || Options::shouldEmitDebugHooks())
	, m_shouldEmitProfileHooks(globalObject->globalObjectMethodTable()->supportsProfiling(globalObject))
	, m_shouldEmitRichSourceInfo(globalObject->globalObjectMethodTable()->supportsRichSourceInfo(globalObject))
	, m_scopeNode(functionNode)
	, m_functionNode(functionNode)
	, m_globalObject(globalObject)
	, m_codeBlock(codeBlock)
	, m_codeType(FunctionCode)
	, m_staticPropertyAnalyzer(&m_instructions)
	, m_vm(&globalObject->vm())
	, m_stack(wtfThreadData().stack())
	// FIXME: This should be a flag
	, m_usesExceptions(false)
	, m_expressionTooDeep(false)
	, m_isBuiltinFunction(codeBlock->isBuiltinFunction())
	, m_usesNonStrictEval(codeBlock->usesEval() && !ecmaMode.isStrict())
	, m_inTailPosition(Options::enableTailCalls() && constructorKind() == ConstructorKind::None && ecmaMode.isStrict())
	, m_needsToUpdateArrowFunctionContext(functionNode->usesArrowFunction() || functionNode->usesEval())
	, m_ecmaMode(ecmaMode)
	, m_derivedContextType(codeBlock->derivedContextType())
{
    pushPrivateAccessNames(parentPrivateNameEnvironment);

	VM& vm = this->vm();

    m_linkTimeConstantRegisters.fill(nullptr);

    if (m_isBuiltinFunction)
        m_shouldEmitDebugHooks = false;

    SymbolTable* functionSymbolTable = SymbolTable::create(*m_vm);
    functionSymbolTable->setUsesNonStrictEval(m_usesNonStrictEval);
    int symbolTableConstantIndex = 0;

    m_cachedParentTDZ = parentScopeTDZVariables;

    m_generatorOrAsyncWrapperFunctionParameterNames = generatorOrAsyncWrapperFunctionParameterNames;

	FunctionParameters& parameters = *functionNode->parameters();
	// http://www.ecma-international.org/ecma-262/6.0/index.html#sec-functiondeclarationinstantiation
	// This implements IsSimpleParameterList in the Ecma 2015 spec.
	// If IsSimpleParameterList is false, we will create a strict-mode like arguments object.
	// IsSimpleParameterList is false if the argument list contains any default parameter values,
	// a rest parameter, or any destructuring patterns.
	// If we do have default parameters, destructuring parameters, or a rest parameter, our parameters will be allocated in a different scope.
	bool isSimpleParameterList = parameters.isSimpleParameterList();

    SourceParseMode parseMode = codeBlock->parseMode();

    bool containsArrowOrEvalButNotInArrowBlock = ((functionNode->usesArrowFunction() && functionNode->doAnyInnerArrowFunctionsUseAnyFeature()) || functionNode->usesEval()) && !m_codeBlock->isArrowFunction();
	bool shouldCaptureSomeOfTheThings = m_shouldEmitDebugHooks || functionNode->needsActivation() || containsArrowOrEvalButNotInArrowBlock;

    bool shouldCaptureAllOfTheThings = m_shouldEmitDebugHooks || codeBlock->usesEval();
    bool needsArguments = ((functionNode->usesArguments() && !codeBlock->isArrowFunction()) || codeBlock->usesEval() || (functionNode->usesArrowFunction() && !codeBlock->isArrowFunction() && isArgumentsUsedInInnerArrowFunction())) && parseMode != SourceParseMode::ClassFieldInitializerMode;

    if (isGeneratorOrAsyncFunctionBodyParseMode(parseMode)) {
        // Generator and AsyncFunction never provides "arguments". "arguments" reference will be resolved in an upper generator function scope.
        needsArguments = false;
    } else if (isGeneratorOrAsyncFunctionWrapperParseMode(parseMode)) {
        // Generator does not provide "arguments". Instead, wrapping GeneratorFunction provides "arguments".
        // This is because arguments of a generator should be evaluated before starting it.
        // To workaround it, we evaluate these arguments as arguments of a wrapping generator function, and reference it from a generator.
        //
        //    function *gen(a, b = hello())
        //    {
        //        return {
        //            @generatorNext: function (@generator, @generatorState, @generatorValue, @generatorResumeMode, @generatorFrame)
        //            {
        //                arguments;  // This `arguments` should reference to the gen's arguments.
        //                ...
        //            }
        //        }
        //    }
        if (needsArguments)
            shouldCaptureSomeOfTheThings = true;
        if (parameters.size()) {
            shouldCaptureSomeOfTheThings = true;
            shouldCaptureAllOfTheThings = true;
        }
    }

    if (shouldCaptureAllOfTheThings)
		functionNode->varDeclarations().markAllVariablesAsCaptured();

    class CapturesFunctor {
    public:
    	CapturesFunctor(VM& vm, FunctionNode* functionNode, bool shouldCaptureSomeOfTheThings, bool shouldCaptureAllOfTheThings, bool needsArguments)
    	: vm(vm)
    	, functionNode(functionNode)
    	, shouldCaptureSomeOfTheThings(shouldCaptureSomeOfTheThings)
    	, shouldCaptureAllOfTheThings(shouldCaptureAllOfTheThings)
    	, needsArguments(needsArguments)
    	{
    	}

    	bool operator()(UniquedStringImpl* uid) {
    		if (!shouldCaptureSomeOfTheThings)
    			return false;
    		if (needsArguments && uid == vm.propertyNames->arguments.impl()) {
    			// Actually, we only need to capture the arguments object when we "need full activation"
    			// because of name scopes. But historically we did it this way, so for now we just preserve
    			// the old behavior.
    			// FIXME: https://bugs.webkit.org/show_bug.cgi?id=143072
    			return true;
    		}
    		return functionNode->captures(uid);
    	}
    private:
    	FunctionNode* functionNode;
    	bool shouldCaptureSomeOfTheThings;
    	bool shouldCaptureAllOfTheThings;
    	bool needsArguments;
    	VM& vm;
    };
    CapturesFunctor captures(vm, functionNode, shouldCaptureSomeOfTheThings, shouldCaptureAllOfTheThings, needsArguments);

    class VarKindFunctor {
    public:
    	VarKindFunctor(CapturesFunctor& captures)
    	: captures(captures)
    	{
    	}

    	VarKind operator()(UniquedStringImpl* uid) {
            return captures(uid) ? VarKind::Scope : VarKind::Stack;
    	}
    private:
    	CapturesFunctor& captures;
    };
    VarKindFunctor varKind(captures);

    m_calleeRegister.setIndex(JSStack::Callee);

    initializeParameters(parameters);
    ASSERT(!(isSimpleParameterList && m_restParameter));

    emitOpcode(op_enter);

    if (isGeneratorOrAsyncFunctionBodyParseMode(parseMode))
        m_generatorRegister = &m_parameters[1];

    allocateAndEmitScope();

    if (functionNameIsInScope(functionNode->ident(), functionNode->functionMode())) {
    	ASSERT(parseMode != SourceParseMode::GeneratorBodyMode);
        ASSERT(!isAsyncFunctionBodyParseMode(parseMode));
        bool isDynamicScope = functionNameScopeIsDynamic(codeBlock->usesEval(), ecmaMode.isStrict());
        bool isFunctionNameCaptured = captures(functionNode->ident().impl());
        bool markAsCaptured = isDynamicScope || isFunctionNameCaptured;
        emitPushFunctionNameScope(functionNode->ident(), &m_calleeRegister, markAsCaptured);
    }

	if (shouldCaptureSomeOfTheThings)
        m_lexicalEnvironmentRegister = addVar();

    if (isGeneratorOrAsyncFunctionBodyParseMode(parseMode) || shouldCaptureSomeOfTheThings/* || vm.typeProfiler()*/)
        symbolTableConstantIndex = addConstantValue(functionSymbolTable)->index();

	// We can allocate the "var" environment if we don't have default parameter expressions. If we have
	// default parameter expressions, we have to hold off on allocating the "var" environment because
	// the parent scope of the "var" environment is the parameter environment.
	if (isSimpleParameterList)
		initializeVarLexicalEnvironment(symbolTableConstantIndex, functionSymbolTable, shouldCaptureSomeOfTheThings);

    // Need to know what our functions are called. Parameters have some goofy behaviors when it
    // comes to functions of the same name.
    for (FunctionMetadataNode* function : functionNode->functionStack())
        m_functions.add(function->ident().impl());

    if (needsArguments) {
        // Create the arguments object now. We may put the arguments object into the activation if
        // it is captured. Either way, we create two arguments object variables: one is our
        // private variable that is immutable, and another that is the user-visible variable. The
        // immutable one is only used here, or during formal parameter resolutions if we opt for
        // DirectArguments.

        m_argumentsRegister = addVar();
        m_argumentsRegister->ref();
    }

    if (needsArguments && !ecmaMode.isStrict() && isSimpleParameterList) {
        // If we captured any formal parameter by name, then we use ScopedArguments. Otherwise we
        // use DirectArguments. With ScopedArguments, we lift all of our arguments into the
        // activation.
		bool capturesAnyParameterByName = false;
		if (functionNode->hasCapturedVariables()) {
			auto& parameters = *functionNode->parameters();
			for (size_t i = 0; i < parameters.size(); ++i) {
				auto* pattern = parameters.at(i).first;
				if (!pattern->isBindingNode())
					continue;
				const Identifier& ident = static_cast<const BindingNode*>(pattern)->boundProperty();
				if (captures(ident.impl())) {
					capturesAnyParameterByName = true;
					break;
				}
			}
		}

        if (capturesAnyParameterByName) {
            ASSERT(m_lexicalEnvironmentRegister);
            functionSymbolTable->setArgumentsLength(vm, parameters.size());

            // For each parameter, we have two possibilities:
            // Either it's a binding node with no function overlap, in which case it gets a name
            // in the symbol table - or it just gets space reserved in the symbol table. Either
            // way we lift the value into the scope.
            for (unsigned i = 0; i < parameters.size(); ++i) {
                ScopeOffset offset = functionSymbolTable->takeNextScopeOffset();
                functionSymbolTable->setArgumentOffset(vm, i, offset);

                unsigned varOrAnonymous = UINT_MAX;

                if (UniquedStringImpl* name = visibleNameForParameter(parameters.at(i).first)) {
                    VarOffset varOffset(offset);
                    SymbolTableEntry entry(varOffset);
                    functionSymbolTable->set(name, entry);

                    //IGNORE_GCC_WARNINGS_BEGIN("dangling-reference")
					const Identifier& ident =
						static_cast<const BindingNode*>(parameters.at(i).first)->boundProperty();
                    //IGNORE_GCC_WARNINGS_END

					varOrAnonymous = addConstant(ident);
				}

                emitOpcode(op_put_to_scope);
                instructions().uncheckedAppend(m_lexicalEnvironmentRegister->index());
                instructions().uncheckedAppend(varOrAnonymous);
                instructions().uncheckedAppend(virtualRegisterForArgument(1 + i));
                instructions().uncheckedAppend(GetPutInfo(ThrowIfNotFound, ResolvedClosureVar, InitializationMode::ScopedArgumentInitialization, ecmaMode).operand());
                instructions().uncheckedAppend(symbolTableConstantIndex);
                instructions().uncheckedAppend(offset.offset());
            }

            // This creates a scoped arguments object and copies the overflow arguments into the
            // scope. It's the equivalent of calling ScopedArguments::createByCopying().
            emitOpcode(op_create_scoped_arguments);
            instructions().uncheckedAppend(m_argumentsRegister->index());
            instructions().uncheckedAppend(m_lexicalEnvironmentRegister->index());
        } else {
            // We're going to put all parameters into the DirectArguments object. First ensure
            // that the symbol table knows that this is happening.
            for (unsigned i = 0; i < parameters.size(); ++i) {
                if (UniquedStringImpl* name = visibleNameForParameter(parameters.at(i).first))
                    functionSymbolTable->set(name, SymbolTableEntry(VarOffset(DirectArgumentsOffset(i))));
            }

            emitOpcode(op_create_direct_arguments);
            instructions().uncheckedAppend(m_argumentsRegister->index());
        }
    } else if (isSimpleParameterList) {
        // Create the formal parameters the normal way. Any of them could be captured, or not. If
        // captured, lift them into the scope. We can not do this if we have default parameter expressions
        // because when default parameter expressions exist, they belong in their own lexical environment
        // separate from the "var" lexical environment.
        for (unsigned i = 0; i < parameters.size(); ++i) {
            UniquedStringImpl* name = visibleNameForParameter(parameters.at(i).first);
            if (!name)
                continue;

            if (!captures(name)) {
                // This is the easy case - just tell the symbol table about the argument. It will
                // be accessed directly.
                functionSymbolTable->set(name, SymbolTableEntry(VarOffset(VirtualRegister_(virtualRegisterForArgument(1 + i)))));
                continue;
            }

            ScopeOffset offset = functionSymbolTable->takeNextScopeOffset();
            const Identifier& ident =
                static_cast<const BindingNode*>(parameters.at(i).first)->boundProperty();
            functionSymbolTable->set(name, SymbolTableEntry(VarOffset(offset)));

            emitOpcode(op_put_to_scope);
            instructions().uncheckedAppend(m_lexicalEnvironmentRegister->index());
            instructions().uncheckedAppend(addConstant(ident));
            instructions().uncheckedAppend(virtualRegisterForArgument(1 + i));
            instructions().uncheckedAppend(GetPutInfo(ThrowIfNotFound, ResolvedClosureVar, InitializationMode::NotInitialization, ecmaMode).operand());
            instructions().uncheckedAppend(symbolTableConstantIndex);
            instructions().uncheckedAppend(offset.offset());
        }
    }

    if (needsArguments && (ecmaMode.isStrict() || !isSimpleParameterList)) {
    	 // Allocate a cloned arguments object.
		emitOpcode(op_create_cloned_arguments);
        instructions().uncheckedAppend(m_argumentsRegister->index());
    }

    // There are some variables that need to be preinitialized to something other than Undefined:
    //
    // - "arguments": unless it's used as a function or parameter, this should refer to the
    //   arguments object.
    //
    // - functions: these always override everything else.
    //
    // The most logical way to do all of this is to initialize none of the variables until now,
    // and then initialize them in BytecodeGenerator::generate() in such an order that the rules
    // for how these things override each other end up holding. We would initialize "arguments" first,
    // then all arguments, then the functions.
    //
    // But some arguments are already initialized by default, since if they aren't captured and we
    // don't have "arguments" then we just point the symbol table at the stack slot of those
    // arguments. We end up initializing the rest of the arguments that have an uncomplicated
    // binding (i.e. don't involve destructuring) above when figuring out how to lay them out,
    // because that's just the simplest thing. This means that when we initialize them, we have to
    // watch out for the things that override arguments (namely, functions).
    //
    // We also initialize callee here as well, just because it's so weird. We know whether we want
    // to do this because we can just check if it's in the symbol table.

    // This is our final act of weirdness. "arguments" is overridden by everything except the
    // callee. We add it to the symbol table if it's not already there and it's not an argument.
    bool shouldCreateArgumentsVariableInParameterScope = false;
    if (needsArguments) {
        // If "arguments" is overridden by a function or destructuring parameter name, then it's
        // OK for us to call createVariable() because it won't change anything. It's also OK for
        // us to them tell BytecodeGenerator::generate() to write to it because it will do so
        // before it initializes functions and destructuring parameters. But if "arguments" is
        // overridden by a "simple" function parameter, then we have to bail: createVariable()
        // would assert and BytecodeGenerator::generate() would write the "arguments" after the
        // argument value had already been properly initialized.

        bool haveParameterNamedArguments = false;
        for (unsigned i = 0; i < parameters.size(); ++i) {
            UniquedStringImpl* name = visibleNameForParameter(parameters.at(i).first);
            if (name == propertyNames().arguments.impl()) {
                haveParameterNamedArguments = true;
                break;
            }
        }

        bool shouldCreateArgumensVariable = !haveParameterNamedArguments
			&& !SourceParseModeSet(SourceParseMode::ArrowFunctionMode, SourceParseMode::AsyncArrowFunctionMode, SourceParseMode::ClassFieldInitializerMode).contains(m_codeBlock->parseMode());
		shouldCreateArgumentsVariableInParameterScope = shouldCreateArgumensVariable && !isSimpleParameterList;
        // Do not create arguments variable in case of Arrow function. Value will be loaded from parent scope
        if (shouldCreateArgumensVariable && !shouldCreateArgumentsVariableInParameterScope) {
            createVariable(
                propertyNames().arguments, varKind(propertyNames().arguments.impl()), functionSymbolTable);

            m_needToInitializeArguments = true;
        }
    }

    for (FunctionMetadataNode* function : functionNode->functionStack()) {
        const Identifier& ident = function->ident();
        createVariable(ident, varKind(ident.impl()), functionSymbolTable);
        m_functionsToInitialize.append(std::make_pair(function, NormalFunctionVariable));
    }
    for (auto& entry : functionNode->varDeclarations()) {
    	ASSERT(!entry.second.isLet() && !entry.second.isConst());
        if (!entry.second.isVar()) // This is either a parameter or callee.
            continue;
        if (shouldCreateArgumentsVariableInParameterScope && entry.first.get() == propertyNames().arguments.impl())
            continue;
        if (isGeneratorOrAsyncFunctionBodyParseMode(parseMode) && generatorOrAsyncWrapperFunctionParameterNames->contains(entry.first.get()))
            continue;
        createVariable(Identifier::fromUid(*m_vm, entry.first.get()), varKind(entry.first.get()), functionSymbolTable, IgnoreExisting);
    }

    if (functionNode->needsNewTargetRegisterForThisScope() || isNewTargetUsedInInnerArrowFunction() || codeBlock->usesEval())
        m_newTargetRegister = addVar();

    switch (parseMode) {
    case SourceParseMode::GeneratorWrapperFunctionMode:
    case SourceParseMode::GeneratorWrapperMethodMode:
    case SourceParseMode::AsyncGeneratorWrapperMethodMode:
    case SourceParseMode::AsyncGeneratorWrapperFunctionMode: {
        m_generatorRegister = addVar();

        // FIXME: Emit to_this only when Generator uses it.
        // https://bugs.webkit.org/show_bug.cgi?id=151586
        emitToThis();

        move(m_generatorRegister, &m_calleeRegister);
        emitCreateThis(m_generatorRegister);
        break;
    }

    case SourceParseMode::AsyncArrowFunctionMode:
    case SourceParseMode::AsyncMethodMode:
    case SourceParseMode::AsyncFunctionMode: {
        ASSERT(!isConstructor());
        ASSERT(constructorKind() == ConstructorKind::None);
        m_generatorRegister = addVar();
        m_promiseCapabilityRegister = addVar();

        if (parseMode != SourceParseMode::AsyncArrowFunctionMode) {
            // FIXME: Emit to_this only when AsyncFunctionBody uses it.
            // https://bugs.webkit.org/show_bug.cgi?id=151586
            emitToThis();
        }

        emitNewObject(m_generatorRegister);

        // let promiseCapability be @newPromiseCapability(@Promise)
        auto varNewPromiseCapability = variable(propertyNames().builtinNames().newPromiseCapabilityPrivateName());
        RefPtr<RegisterID> scope = newTemporary();
        move(scope.get(), emitResolveScope(scope.get(), varNewPromiseCapability));
        RefPtr<RegisterID> newPromiseCapability = emitGetFromScope(newTemporary(), scope.get(), varNewPromiseCapability, ThrowIfNotFound);

        CallArguments args(*this, 0, 1);
        emitLoad(args.thisRegister(), jsUndefined());

        auto& builtinNames = propertyNames().builtinNames();
		auto varPromiseConstructor = variable(m_isBuiltinFunction ? builtinNames.InternalPromisePrivateName() : builtinNames.PromisePrivateName());
		move(scope.get(), emitResolveScope(scope.get(), varPromiseConstructor));
        emitGetFromScope(args.argumentRegister(0), scope.get(), varPromiseConstructor, ThrowIfNotFound);

        // JSTextPosition(int _line, int _offset, int _lineStartOffset)
        JSTextPosition divot(m_scopeNode->firstLine(), m_scopeNode->startOffset(), m_scopeNode->lineStartOffset());
        emitCall(promiseCapabilityRegister(), newPromiseCapability.get(), NoExpectedFunction, args, divot, divot, divot);
        break;
    }

    case SourceParseMode::AsyncGeneratorBodyMode:
    case SourceParseMode::AsyncFunctionBodyMode:
    case SourceParseMode::AsyncArrowFunctionBodyMode:
    case SourceParseMode::GeneratorBodyMode: {
        // |this| is already filled correctly before here.
        if (m_newTargetRegister)
        	emitLoad(m_newTargetRegister, jsUndefined());
        break;
    }

    default: {
        if (SourceParseMode::ArrowFunctionMode != parseMode) {
			if (isConstructor()) {
		        if (m_newTargetRegister)
		        	move(m_newTargetRegister, &m_thisRegister);
                switch (constructorKind()) {
                case ConstructorKind::Naked:
                    // Naked constructor not create |this| automatically.
                    break;
                case ConstructorKind::None:
                case ConstructorKind::Base:
                    emitCreateThis(&m_thisRegister);
                    if (Options::usePrivateMethods() && privateBrandRequirement() == PrivateBrandRequirement::Needed)
                        emitInstallPrivateBrand(&m_thisRegister);

                    emitInstanceFieldInitializationIfNeeded(&m_thisRegister, &m_calleeRegister, m_scopeNode->position(), m_scopeNode->position(), m_scopeNode->position());
                    break;
                case ConstructorKind::Extends:
                    moveEmptyValue(&m_thisRegister);
                    break;
                }
			} else if (constructorKind() != ConstructorKind::None)
				emitThrowTypeError(ASCIILiteral("Cannot call a class constructor without |new|"));
			else {
                bool shouldEmitToThis = false;
                if (functionNode->usesThis() || codeBlock->usesEval() || m_scopeNode->doAnyInnerArrowFunctionsUseThis() || m_scopeNode->doAnyInnerArrowFunctionsUseEval())
                    shouldEmitToThis = true;
                else if ((functionNode->usesSuperProperty() || m_scopeNode->doAnyInnerArrowFunctionsUseSuperProperty()) && !ecmaMode.isStrict()) {
                    // We must emit to_this when we're not in strict mode because we
                    // will convert |this| to an object, and that object may be passed
                    // to a strict function as |this|. This is observable because that
                    // strict function's to_this will just return the object.
                    //
                    // We don't need to emit this for strict-mode code because
                    // strict-mode code may call another strict function, which will
                    // to_this if it directly uses this; this is OK, because we defer
                    // to_this until |this| is used directly. Strict-mode code might
                    // also call a sloppy mode function, and that will to_this, which
                    // will defer the conversion, again, until necessary.
                    shouldEmitToThis = true;
                }

                if (shouldEmitToThis)
                    emitToThis();
			}
        }
		break;
    }
    }

    // We need load |super| & |this| for arrow function before initializeDefaultParameterValuesAndSetupFunctionScopeStack
    // if we have default parameter expression. Because |super| & |this| values can be used there
    if ((SourceParseModeSet(SourceParseMode::ArrowFunctionMode, SourceParseMode::AsyncArrowFunctionMode).contains(parseMode) && !isSimpleParameterList) || parseMode == SourceParseMode::AsyncArrowFunctionBodyMode) {
        if (functionNode->usesThis() || functionNode->usesSuperProperty())
            emitLoadThisFromArrowFunctionLexicalEnvironment();

        if (m_scopeNode->needsNewTargetRegisterForThisScope())
            emitLoadNewTargetFromArrowFunctionLexicalEnvironment();
    }

    if (needsToUpdateArrowFunctionContext() && !codeBlock->isArrowFunction()) {
    	bool canReuseLexicalEnvironment = isSimpleParameterList;
		initializeArrowFunctionContextScopeIfNeeded(functionSymbolTable, canReuseLexicalEnvironment);
        emitPutThisToArrowFunctionContextScope();
        emitPutNewTargetToArrowFunctionContextScope();
        emitPutDerivedConstructorToArrowFunctionContextScope();
    }

    RefPtr<Label> catchLabel = newLabel();
    TryData* tryFormalParametersData = 0;
    bool needTryCatch = isAsyncFunctionWrapperParseMode(parseMode) && !isSimpleParameterList;
    if (needTryCatch) {
        RefPtr<Label> tryFormalParametersStart = newEmittedLabel();
        tryFormalParametersData = pushTry(tryFormalParametersStart.get(), catchLabel.get(), HandlerType::SynthesizedCatch);
    }

    // All "addVar()"s needs to happen before "initializeDefaultParameterValuesAndSetupFunctionScopeStack()" is called
    // because a function's default parameter ExpressionNodes will use temporary registers.
    initializeDefaultParameterValuesAndSetupFunctionScopeStack(parameters, isSimpleParameterList, functionNode, functionSymbolTable, symbolTableConstantIndex, captures, shouldCreateArgumentsVariableInParameterScope);

    if (needTryCatch) {
        RefPtr<Label> didNotThrow = newLabel();
        emitJump(didNotThrow.get());
        emitLabel(catchLabel.get());
        popTry(tryFormalParametersData, catchLabel.get());

        RefPtr<RegisterID> thrownValue = newTemporary();
        emitOutOfLineCatchHandler(thrownValue.get(), 0, tryFormalParametersData);

        // return promiseCapability.@reject(thrownValue)
        RefPtr<RegisterID> reject = emitGetById(newTemporary(), promiseCapabilityRegister(), m_vm->propertyNames->builtinNames().rejectPrivateName());

        CallArguments args(*this, 0, 1);
        emitLoad(args.thisRegister(), jsUndefined());
        move(args.argumentRegister(0), thrownValue.get());

        JSTextPosition divot(functionNode->firstLine(), functionNode->startOffset(), functionNode->lineStartOffset());

        RefPtr<RegisterID> result = emitCall(newTemporary(), reject.get(), NoExpectedFunction, args, divot, divot, divot);
        emitReturn(emitGetById(newTemporary(), promiseCapabilityRegister(), m_vm->propertyNames->builtinNames().promisePrivateName()));

        emitLabel(didNotThrow.get());
    }

    // If we don't have  default parameter expression, then loading |this| inside an arrow function must be done
    // after initializeDefaultParameterValuesAndSetupFunctionScopeStack() because that function sets up the
    // SymbolTable stack and emitLoadThisFromArrowFunctionLexicalEnvironment() consults the SymbolTable stack
    if (SourceParseModeSet(SourceParseMode::ArrowFunctionMode, SourceParseMode::AsyncArrowFunctionMode).contains(parseMode) && isSimpleParameterList) {
		if (functionNode->usesThis() || functionNode->usesSuperProperty())
            emitLoadThisFromArrowFunctionLexicalEnvironment();

        if (m_scopeNode->needsNewTargetRegisterForThisScope())
            emitLoadNewTargetFromArrowFunctionLexicalEnvironment();
    }

    // Set up the lexical environment scope as the generator frame. We store the saved and resumed generator registers into this scope with the symbol keys.
    // Since they are symbol keyed, these variables cannot be reached from the usual code.
    if (isGeneratorOrAsyncFunctionBodyParseMode(parseMode)) {
        m_generatorFrameSymbolTable.set(*m_vm, functionSymbolTable);
        m_generatorFrameSymbolTableIndex = symbolTableConstantIndex;
        if (m_lexicalEnvironmentRegister)
        	move(generatorFrameRegister(), m_lexicalEnvironmentRegister);
        else {
            // It would be possible that generator does not need to suspend and resume any registers.
            // In this case, we would like to avoid creating a lexical environment as much as possible.
            // op_create_generator_frame_environment is a marker, which is similar to op_yield.
            // Generatorification inserts lexical environment creation if necessary. Otherwise, we convert it to op_mov frame, `undefined`.
            //OpCreateGeneratorFrameEnvironment::emit(this, generatorFrameRegister(), scopeRegister(), VirtualRegister { symbolTableConstantIndex }, addConstantValue(jsUndefined()));
            emitOpcode(op_create_generator_frame_environment);
            instructions().uncheckedAppend(generatorFrameRegister()->index());
            instructions().uncheckedAppend(scopeRegister()->index());
            instructions().uncheckedAppend(symbolTableConstantIndex);
            instructions().uncheckedAppend(addConstantValue(jsUndefined())->index());
        }
        emitPutById(generatorRegister(), propertyNames().generatorFramePrivateName, generatorFrameRegister());
    }

    bool shouldInitializeBlockScopedFunctions = false; // We generate top-level function declarations in ::generate().
	pushLexicalScope(m_scopeNode, ScopeType::LetConstScope, TDZCheckOptimization::Optimize, NestedScopeType::IsNotNested, nullptr, shouldInitializeBlockScopedFunctions);
}

BytecodeGenerator::BytecodeGenerator(EvalNode* evalNode, JSGlobalObject* globalObject, EvalCodeBlock* codeBlock, const RefPtr<TDZEnvironmentLink>& parentScopeTDZVariables, const PrivateNameEnvironment* parentPrivateNameEnvironment, const Vector<Identifier>*, ECMAMode ecmaMode)
: m_shouldEmitDebugHooks(globalObject->debugger() || Options::shouldEmitDebugHooks())
, m_shouldEmitProfileHooks(globalObject->globalObjectMethodTable()->supportsProfiling(globalObject))
, m_shouldEmitRichSourceInfo(globalObject->globalObjectMethodTable()->supportsRichSourceInfo(globalObject))
, m_scopeNode(evalNode)
, m_globalObject(globalObject)
, m_codeBlock(codeBlock)
, m_thisRegister(CallFrame::thisArgumentOffset())
, m_codeType(EvalCode)
, m_staticPropertyAnalyzer(&m_instructions)
, m_vm(&globalObject->vm())
, m_stack(wtfThreadData().stack())
, m_usesNonStrictEval(codeBlock->usesEval() && !ecmaMode.isStrict())
, m_needsToUpdateArrowFunctionContext(evalNode->usesArrowFunction() || evalNode->usesEval())
, m_ecmaMode(ecmaMode)
, m_derivedContextType(codeBlock->derivedContextType())
{
    m_linkTimeConstantRegisters.fill(nullptr);

    m_codeBlock->setNumParameters(1);

    pushPrivateAccessNames(parentPrivateNameEnvironment);

    m_cachedParentTDZ = parentScopeTDZVariables;

    emitOpcode(op_enter);

    allocateAndEmitScope();

    for (FunctionMetadataNode* function : evalNode->functionStack()) {
		m_codeBlock->addFunctionDecl(makeFunction(function));
		m_functionsToInitialize.append(std::make_pair(function, TopLevelFunctionVariable));
	}

    const VariableEnvironment& varDeclarations = evalNode->varDeclarations();
    Vector<Identifier> variables;
    Vector<Identifier> hoistedFunctions;
    for (auto& entry : varDeclarations) {
        ASSERT(entry.second.isVar());
        ASSERT(entry.first->isAtomic() || it->first->isSymbol());
        if (entry.second.isSloppyModeHoistedFunction())
            hoistedFunctions.append(Identifier::fromUid(*m_vm, entry.first.get()));
        else if (!entry.second.isFunction())
        	variables.append(Identifier::fromUid(*m_vm, entry.first.get()));
    }
    codeBlock->adoptVariables(variables);
    codeBlock->adoptFunctionHoistingCandidates(hoistedFunctions);

    if (evalNode->needsNewTargetRegisterForThisScope())
        m_newTargetRegister = addVar();

    if (codeBlock->isArrowFunctionContext() && (evalNode->usesThis() || evalNode->usesSuperProperty()))
		emitLoadThisFromArrowFunctionLexicalEnvironment();

    if (evalNode->needsNewTargetRegisterForThisScope())
        emitLoadNewTargetFromArrowFunctionLexicalEnvironment();

    if (needsToUpdateArrowFunctionContext() && !codeBlock->isArrowFunctionContext() && !isDerivedConstructorContext()) {
        initializeArrowFunctionContextScopeIfNeeded();
        emitPutThisToArrowFunctionContextScope();
    }

    bool shouldInitializeBlockScopedFunctions = false; // We generate top-level function declarations in ::generate().
    pushLexicalScope(m_scopeNode, ScopeType::LetConstScope, TDZCheckOptimization::Optimize, NestedScopeType::IsNotNested, nullptr, shouldInitializeBlockScopedFunctions);
}

BytecodeGenerator::BytecodeGenerator(ModuleProgramNode* moduleProgramNode, JSGlobalObject* globalObject, ModuleProgramCodeBlock* codeBlock, const RefPtr<TDZEnvironmentLink>& parentScopeTDZVariables, const PrivateNameEnvironment*, const Vector<Identifier>*, ECMAMode ecmaMode)
: m_shouldEmitDebugHooks(globalObject->debugger() || Options::shouldEmitDebugHooks())
, m_shouldEmitProfileHooks(globalObject->globalObjectMethodTable()->supportsProfiling(globalObject))
, m_shouldEmitRichSourceInfo(globalObject->globalObjectMethodTable()->supportsRichSourceInfo(globalObject))
, m_scopeNode(moduleProgramNode)
, m_globalObject(globalObject)
, m_codeBlock(codeBlock)
, m_thisRegister(CallFrame::thisArgumentOffset())
, m_codeType(GlobalCode)
, m_staticPropertyAnalyzer(&m_instructions)
, m_vm(&globalObject->vm())
, m_stack(wtfThreadData().stack())
, m_needsToUpdateArrowFunctionContext(moduleProgramNode->usesArrowFunction() || moduleProgramNode->usesEval())
, m_ecmaMode(ecmaMode)
{
    ASSERT_UNUSED(parentScopeTDZVariables, !parentScopeTDZVariables);

    m_linkTimeConstantRegisters.fill(nullptr);

    if (m_isBuiltinFunction)
        m_shouldEmitDebugHooks = false;

    SymbolTable* moduleEnvironmentSymbolTable = SymbolTable::create(*m_vm);
    moduleEnvironmentSymbolTable->setUsesNonStrictEval(m_usesNonStrictEval);
    moduleEnvironmentSymbolTable->setScopeType(SymbolTable::ScopeType::LexicalScope);

    bool shouldCaptureAllOfTheThings = m_shouldEmitDebugHooks || codeBlock->usesEval();
    if (shouldCaptureAllOfTheThings)
        moduleProgramNode->varDeclarations().markAllVariablesAsCaptured();

    class Captures
	{
    public:
    	Captures(ModuleProgramNode* moduleProgramNode)
    	: moduleProgramNode(moduleProgramNode)
    	{
    	}
    	bool operator()(UniquedStringImpl* uid) {
    		return moduleProgramNode->captures(uid);
    	}
    private:
    	ModuleProgramNode* moduleProgramNode;
    };
    Captures captures(moduleProgramNode);
    class LookUpVarKind
    {
    public:
    	LookUpVarKind(Captures& captures)
    	: captures(captures)
		{
		}
    	VarKind operator()(UniquedStringImpl* uid, const VariableEnvironmentEntry& entry){
			// Allocate the exported variables in the module environment.
			if (entry.isExported())
				return VarKind::Scope;

			// Allocate the namespace variables in the module environment to instantiate
			// it from the outside of the module code.
			if (entry.isImportedNamespace())
				return VarKind::Scope;

			if (entry.isCaptured())
				return VarKind::Scope;
			return captures(uid) ? VarKind::Scope : VarKind::Stack;
		}
    private:
    	 Captures& captures;
    };
    LookUpVarKind lookUpVarKind(captures);

    emitOpcode(op_enter);

    allocateAndEmitScope();
    RegisterID* moduleScope = addVar();
    move(moduleScope, scopeRegister());

    m_calleeRegister.setIndex(JSStack::Callee);

    m_codeBlock->setNumParameters(1); // Allocate space for "this"

    // Now declare all variables.

    createVariable(m_vm->propertyNames->starNamespacePrivateName, VarKind::Scope, moduleEnvironmentSymbolTable, VerifyExisting);
	createVariable(m_vm->propertyNames->metaPrivateName, VarKind::Scope, moduleEnvironmentSymbolTable, VerifyExisting);

    for (auto iter = moduleProgramNode->varDeclarations().begin(); iter != moduleProgramNode->varDeclarations().end(); ++iter) {
        ASSERT(!iter->second.isLet() && !iter->second.isConst());
        if (!iter->second.isVar()) // This is either a parameter or callee.
            continue;
        // Imported bindings are not allocated in the module environment as usual variables' way.
        // These references remain the "Dynamic" in the unlinked code block. Later, when linking
        // the code block, we resolve the reference to the "ModuleVar".
        if (iter->second.isImported() && !iter->second.isImportedNamespace())
            continue;
        createVariable(Identifier::fromUid(*m_vm, iter->first.get()), lookUpVarKind(iter->first.get(), iter->second), moduleEnvironmentSymbolTable, IgnoreExisting);
    }

    VariableEnvironment& lexicalVariables = moduleProgramNode->lexicalVariables();
    instantiateLexicalVariables(lexicalVariables, ScopeType::LetConstScope, moduleEnvironmentSymbolTable, ScopeRegisterType::Block, lookUpVarKind);

    // We keep the symbol table in the constant pool.
    RegisterID* constantSymbolTable = nullptr;
    constantSymbolTable = addConstantValue(moduleEnvironmentSymbolTable->cloneScopePart(*m_vm));

    pushTDZVariables(lexicalVariables, TDZCheckOptimization::Optimize, TDZRequirement::UnderTDZ);
	bool isWithScope = false;

	m_lexicalScopeStack.append(LexicalScopeStackEntry { moduleEnvironmentSymbolTable, moduleScope, isWithScope, constantSymbolTable->index() });
    emitPrefillStackTDZVariables(lexicalVariables, moduleEnvironmentSymbolTable);

    // makeFunction assumes that there's correct TDZ stack entries.
    // So it should be called after putting our lexical environment to the TDZ stack correctly.

    for (FunctionMetadataNode* function : moduleProgramNode->functionStack()) {
        const auto& iterator = moduleProgramNode->lexicalVariables().find(function->ident().impl());
        RELEASE_ASSERT(iterator != moduleProgramNode->lexicalVariables().end());
        RELEASE_ASSERT(!iterator->value.isImported());

        VarKind varKind = lookUpVarKind(iterator->first.get(), iterator->second);
        if (varKind == VarKind::Scope) {
            // http://www.ecma-international.org/ecma-262/6.0/#sec-moduledeclarationinstantiation
            // Section 15.2.1.16.4, step 16-a-iv-1.
            // All heap allocated function declarations should be instantiated when the module environment
            // is created. They include the exported function declarations and not-exported-but-heap-allocated
            // function declarations. This is required because exported function should be instantiated before
            // executing the any module in the dependency graph. This enables the modules to link the imported
            // bindings before executing the any module code.
            //
            // And since function declarations are instantiated before executing the module body code, the spec
            // allows the functions inside the module to be executed before its module body is executed under
            // the circular dependencies. The following is the example.
            //
            // Module A (executed first):
            //    import { b } from "B";
            //    // Here, the module "B" is not executed yet, but the function declaration is already instantiated.
            //    // So we can call the function exported from "B".
            //    b();
            //
            //    export function a() {
            //    }
            //
            // Module B (executed second):
            //    import { a } from "A";
            //
            //    export function b() {
            //        c();
            //    }
            //
            //    // c is not exported, but since it is referenced from the b, we should instantiate it before
            //    // executing the "B" module code.
            //    function c() {
            //        a();
            //    }
            //
            // Module EntryPoint (executed last):
            //    import "B";
            //    import "A";
            //
            m_codeBlock->addFunctionDecl(makeFunction(function));
        } else {
            // Stack allocated functions can be allocated when executing the module's body.
            m_functionsToInitialize.append(std::make_pair(function, NormalFunctionVariable));
        }
    }

    // Remember the constant register offset to the top-most symbol table. This symbol table will be
    // cloned in the code block linking. After that, to create the module environment, we retrieve
    // the cloned symbol table from the linked code block by using this offset.
    codeBlock->setModuleEnvironmentSymbolTableConstantRegisterOffset(constantSymbolTable->index());
}

BytecodeGenerator::~BytecodeGenerator() = default;

template<typename CapturesFunctor>
void BytecodeGenerator::initializeDefaultParameterValuesAndSetupFunctionScopeStack(
    FunctionParameters& parameters, bool isSimpleParameterList, FunctionNode* functionNode, SymbolTable* functionSymbolTable,
    int symbolTableConstantIndex, CapturesFunctor& captures, bool shouldCreateArgumentsVariableInParameterScope)
{
    Vector<std::pair<Identifier, RefPtr<RegisterID>>> valuesToMoveIntoVars;
    ASSERT(!(isSimpleParameterList && shouldCreateArgumentsVariableInParameterScope));
	if (!isSimpleParameterList) {
        // Refer to the ES6 spec section 9.2.12: http://www.ecma-international.org/ecma-262/6.0/index.html#sec-functiondeclarationinstantiation
        // This implements step 21.
        VariableEnvironment environment;
        Vector<Identifier, 16> allParameterNames;
        for (unsigned i = 0; i < parameters.size(); i++)
            parameters.at(i).first->collectBoundIdentifiers(allParameterNames);
        if (shouldCreateArgumentsVariableInParameterScope)
            allParameterNames.append(propertyNames().arguments);
        IdentifierSet parameterSet;
        for (auto& ident : allParameterNames) {
            parameterSet.add(ident.impl());
            auto addResult = environment.add(ident);
            addResult.iterator->second.setIsLet(); // When we have default parameter expressions, parameters act like "let" variables.
            if (captures(ident.impl()))
                addResult.iterator->second.setIsCaptured();
        }
        // This implements step 25 of section 9.2.12.
        pushLexicalScopeInternal(environment, TDZCheckOptimization::Optimize, NestedScopeType::IsNotNested, 0, TDZRequirement::UnderTDZ, ScopeType::LetConstScope, ScopeRegisterType::Block);

        if (shouldCreateArgumentsVariableInParameterScope) {
            Variable argumentsVariable = variable(propertyNames().arguments);
            initializeVariable(argumentsVariable, m_argumentsRegister);
            liftTDZCheckIfPossible(argumentsVariable);
        }

        RefPtr<RegisterID> temp = newTemporary();
        for (unsigned i = 0; i < parameters.size(); i++) {
            std::pair<DestructuringPatternNode*, ExpressionNode*> parameter = parameters.at(i);
            if (parameter.first->isRestParameter())
				continue;
            if ((i + 1) < m_parameters.size())
                move(temp.get(), &m_parameters[i + 1]);
            else
                emitGetArgument(temp.get(), i);
            if (parameter.second) {
                RefPtr<RegisterID> condition = emitIsUndefined(newTemporary(), temp.get());
                RefPtr<Label> skipDefaultParameterBecauseNotUndefined = newLabel();
                emitJumpIfFalse(condition.get(), skipDefaultParameterBecauseNotUndefined.get());
                emitNode(temp.get(), parameter.second);
                emitLabel(skipDefaultParameterBecauseNotUndefined.get());
            }

            parameter.first->bindValue(*this, temp.get());
        }

        if (m_restParameter)
            m_restParameter->emit(*this);

        // Final act of weirdness for default parameters. If a "var" also
        // has the same name as a parameter, it should start out as the
        // value of that parameter. Note, though, that they will be distinct
        // bindings.
        // This is step 28 of section 9.2.12.
        for (auto& entry : functionNode->varDeclarations()) {
            if (!entry.second.isVar()) // This is either a parameter or callee.
                continue;

            if (parameterSet.contains(entry.first)) {
                Identifier ident = Identifier::fromUid(*m_vm, entry.first.get());
                Variable var = variable(ident);
                RegisterID* scope = emitResolveScope(nullptr, var);
                RefPtr<RegisterID> value = emitGetFromScope(newTemporary(), scope, var, DoNotThrowIfNotFound);
                valuesToMoveIntoVars.append(std::make_pair(ident, value));
            }
        }

        // Functions with default parameter expressions must have a separate environment
        // record for parameters and "var"s. The "var" environment record must have the
        // parameter environment record as its parent.
        // See step 28 of section 9.2.12.
        bool hasCapturedVariables = !!m_lexicalEnvironmentRegister;
		initializeVarLexicalEnvironment(symbolTableConstantIndex, functionSymbolTable, hasCapturedVariables);
	}

    // This completes step 28 of section 9.2.12.
    for (unsigned i = 0; i < valuesToMoveIntoVars.size(); i++) {
        ASSERT(!isSimpleParameterList);
        Variable var = variable(valuesToMoveIntoVars[i].first);
        RegisterID* scope = emitResolveScope(0, var);
        emitPutToScope(scope, var, valuesToMoveIntoVars[i].second.get(), DoNotThrowIfNotFound, InitializationMode::NotInitialization);
    }
}

bool BytecodeGenerator::needsDerivedConstructorInArrowFunctionLexicalEnvironment()
{
	ASSERT(m_codeBlock->isClassContext() || !(isConstructor() && constructorKind() == ConstructorKind::Extends));
	return m_codeBlock->isClassContext() && isSuperUsedInInnerArrowFunction();
}

void BytecodeGenerator::initializeArrowFunctionContextScopeIfNeeded(SymbolTable* functionSymbolTable, bool canReuseLexicalEnvironment)
{
    ASSERT(!m_arrowFunctionContextLexicalEnvironmentRegister);

    if (canReuseLexicalEnvironment && m_lexicalEnvironmentRegister) {
        RELEASE_ASSERT(!m_codeBlock->isArrowFunction());
        RELEASE_ASSERT(functionSymbolTable);

        m_arrowFunctionContextLexicalEnvironmentRegister = m_lexicalEnvironmentRegister;

        ScopeOffset offset;

        //ConcurrentJITLocker locker(ConcurrentJITLocker::NoLockingNecessary);
        if (isThisUsedInInnerArrowFunction()) {
            offset = functionSymbolTable->takeNextScopeOffset();
            functionSymbolTable->set(propertyNames().builtinNames().thisPrivateName().impl(), SymbolTableEntry(VarOffset(offset)));
        }

        if (m_codeType == FunctionCode && isNewTargetUsedInInnerArrowFunction()) {
            offset = functionSymbolTable->takeNextScopeOffset();
            functionSymbolTable->set(propertyNames().builtinNames().newTargetLocalPrivateName().impl(), SymbolTableEntry(VarOffset(offset)));
        }

        if (needsDerivedConstructorInArrowFunctionLexicalEnvironment()) {
            offset = functionSymbolTable->takeNextScopeOffset();
            functionSymbolTable->set(propertyNames().builtinNames().derivedConstructorPrivateName().impl(), SymbolTableEntry(VarOffset(offset)));
        }

        return;
    }

    VariableEnvironment environment;

    if (isThisUsedInInnerArrowFunction()) {
		auto addResult = environment.add(propertyNames().builtinNames().thisPrivateName());
		addResult.iterator->second.setIsCaptured();
		addResult.iterator->second.setIsLet();
    }

    if (m_codeType == FunctionCode && isNewTargetUsedInInnerArrowFunction()) {
        auto addTarget = environment.add(propertyNames().builtinNames().newTargetLocalPrivateName());
        addTarget.iterator->second.setIsCaptured();
        addTarget.iterator->second.setIsLet();
    }

    if (needsDerivedConstructorInArrowFunctionLexicalEnvironment()) {
        auto derivedConstructor = environment.add(propertyNames().builtinNames().derivedConstructorPrivateName());
        derivedConstructor.iterator->second.setIsCaptured();
        derivedConstructor.iterator->second.setIsLet();
    }

    if (environment.totalSize() > 0) {
		size_t size = m_lexicalScopeStack.size();
		pushLexicalScopeInternal(environment, TDZCheckOptimization::Optimize, NestedScopeType::IsNotNested, 0, TDZRequirement::UnderTDZ, ScopeType::LetConstScope, ScopeRegisterType::Block);

		ASSERT_UNUSED(size, m_lexicalScopeStack.size() == size + 1);

		m_arrowFunctionContextLexicalEnvironmentRegister = m_lexicalScopeStack.last().m_scope;
    }
}

RegisterID* BytecodeGenerator::initializeNextParameter()
{
    VirtualRegister_ reg = VirtualRegister_(virtualRegisterForArgument(m_codeBlock->numParameters()));
    m_parameters.grow(m_parameters.size() + 1);
	auto& parameter = registerFor(reg);
    parameter.setIndex(reg.offset());
    m_codeBlock->addParameter();
    return &parameter;
}

void BytecodeGenerator::initializeParameters(FunctionParameters& parameters)
{
    // Make sure the code block knows about all of our parameters, and make sure that parameters
    // needing destructuring are noted.
    m_thisRegister.setIndex(initializeNextParameter()->index()); // this

    bool nonSimpleArguments = false;
    for (unsigned i = 0; i < parameters.size(); ++i) {
        auto parameter = parameters.at(i);
        DestructuringPatternNode* pattern = parameter.first;
        if (pattern->isRestParameter()) {
            RELEASE_ASSERT(!m_restParameter);
            m_restParameter = static_cast<RestParameterNode*>(pattern);
            nonSimpleArguments = true;
            continue;
        }
        if (parameter.second) {
            nonSimpleArguments = true;
            continue;
        }
        if (!nonSimpleArguments)
            initializeNextParameter();
    }
}

void BytecodeGenerator::initializeVarLexicalEnvironment(int symbolTableConstantIndex, SymbolTable* functionSymbolTable, bool hasCapturedVariables)
{
    if (hasCapturedVariables) {
		RELEASE_ASSERT(m_lexicalEnvironmentRegister);
		emitOpcode(op_create_lexical_environment);
		instructions().uncheckedAppend(m_lexicalEnvironmentRegister->index());
		instructions().uncheckedAppend(scopeRegister()->index());
		instructions().uncheckedAppend(symbolTableConstantIndex);
		instructions().uncheckedAppend(addConstantValue(jsUndefined())->index());

		emitOpcode(op_mov);
		instructions().uncheckedAppend(scopeRegister()->index());
		instructions().uncheckedAppend(m_lexicalEnvironmentRegister->index());

		pushLocalControlFlowScope();
	}
	bool isWithScope = false;
	m_lexicalScopeStack.append(LexicalScopeStackEntry{ functionSymbolTable, m_lexicalEnvironmentRegister, isWithScope, symbolTableConstantIndex });
	m_varScopeLexicalScopeStackIndex = m_lexicalScopeStack.size() - 1;
}

UniquedStringImpl* BytecodeGenerator::visibleNameForParameter(DestructuringPatternNode* pattern)
{
    if (pattern->isBindingNode()) {
        const Identifier& ident = static_cast<const BindingNode*>(pattern)->boundProperty();
        if (!m_functions.contains(ident.impl()))
            return ident.impl();
    }
    return 0;
}

RegisterID* BytecodeGenerator::newRegister()
{
    m_calleeLocals.append(virtualRegisterForLocal(m_calleeLocals.size()));
    int numCalleeRegisters = std::max<int>(m_codeBlock->m_numCalleeLocals, m_calleeLocals.size());
    numCalleeRegisters = WTF::roundUpToMultipleOf(stackAlignmentRegisters(), numCalleeRegisters);
	m_codeBlock->m_numCalleeLocals = numCalleeRegisters;
    return &m_calleeLocals.last();
}

ALWAYS_INLINE void BytecodeGenerator::reclaimFreeRegisters()
{
    shrinkToFit(m_calleeLocals);
}

RegisterID* BytecodeGenerator::newBlockScopeVariable()
{
    reclaimFreeRegisters();

    return newRegister();
}

RegisterID* BytecodeGenerator::newTemporary()
{
    reclaimFreeRegisters();

    RegisterID* result = newRegister();
    result->setTemporary();
    return result;
}

PassRefPtr<LabelScope> BytecodeGenerator::newLabelScope(LabelScope::Type type, const Identifier* name)
{
    shrinkToFit(m_labelScopes);

    // Allocate new label scope.
    m_labelScopes.append(type, name, labelScopeDepth(), newLabel(), type == LabelScope::Loop ? newLabel() : PassRefPtr<Label>()); // Only loops have continue targets.
    return &m_labelScopes.last();
}

Ref<Label> BytecodeGenerator::newLabel()
{
    shrinkToFit(m_labels);

    // Allocate new label ID.
    m_labels.append(Label());
    return m_labels.last();
}

Ref<Label> BytecodeGenerator::newEmittedLabel()
{
	Ref<Label> label = newLabel();
    emitLabel(label.get());
    return label;
}

void BytecodeGenerator::emitLabel(Label& l0)
{
    unsigned newLabelIndex = instructions().size();
    l0.setLocation(*this, newLabelIndex);

    if (m_codeBlock->numberOfJumpTargets()) {
        unsigned lastLabelIndex = m_codeBlock->lastJumpTarget();
        ASSERT(lastLabelIndex <= newLabelIndex);
        if (newLabelIndex == lastLabelIndex) {
            // Peephole optimizations have already been disabled by emitting the last label
            return;
        }
    }

    m_codeBlock->addJumpTarget(newLabelIndex);

    // This disables peephole optimizations when an instruction is a jump target
    disablePeepholeOptimization();
}

ALWAYS_INLINE ArrayProfile* BytecodeGenerator::newArrayProfile()
{
    return m_codeBlock->addArrayProfile(instructions().size());
}

UnlinkedArrayAllocationProfile BytecodeGenerator::newArrayAllocationProfile()
{
    return m_arrayAllocationProfileCount++;
}

UnlinkedObjectAllocationProfile BytecodeGenerator::newObjectAllocationProfile()
{
    return m_objectAllocationProfileCount++;
}

ALWAYS_INLINE UnlinkedValueProfile BytecodeGenerator::emitProfiledOpcode(OpcodeID opcodeID)
{
    UnlinkedValueProfile result = m_codeBlock->addValueProfile();
    emitOpcode(opcodeID);
    return result;
}

void BytecodeGenerator::emitLoopHint()
{
    emitOpcode(op_loop_hint);
}

void BytecodeGenerator::retrieveLastBinaryOp(int& dstIndex, int& src1Index, int& src2Index)
{
    ASSERT(instructions().size() >= 4);
    size_t size = instructions().size();
    dstIndex = instructions().at(size - 3).u.operand;
    src1Index = instructions().at(size - 2).u.operand;
    src2Index = instructions().at(size - 1).u.operand;
}

void BytecodeGenerator::retrieveLastUnaryOp(int& dstIndex, int& srcIndex)
{
    ASSERT(instructions().size() >= 3);
    size_t size = instructions().size();
    dstIndex = instructions().at(size - 2).u.operand;
    srcIndex = instructions().at(size - 1).u.operand;
}

void ALWAYS_INLINE BytecodeGenerator::rewindBinaryOp()
{
    ASSERT(instructions().size() >= 4);
    instructions().shrink(instructions().size() - 4);
    disablePeepholeOptimization();
}

void ALWAYS_INLINE BytecodeGenerator::rewindUnaryOp()
{
    ASSERT(instructions().size() >= 3);
    instructions().shrink(instructions().size() - 3);
    disablePeepholeOptimization();
}

PassRefPtr<Label> BytecodeGenerator::emitJump(Label* target)
{
    size_t begin = instructions().size();
    emitOpcode(op_jmp);
    instructions().uncheckedAppend(target->bind(begin, instructions().size()));
    return target;
}

void BytecodeGenerator::emitJumpIfTrue(RegisterID* cond, Label* target)
{
    if (canDoPeepholeOptimization()) {
        auto fuseCompareAndJump = [&] (OpcodeID jumpID) {
            int dstIndex;
            int src1Index;
            int src2Index;

            retrieveLastBinaryOp(dstIndex, src1Index, src2Index);

            if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
                rewindBinaryOp();

                size_t begin = instructions().size();
                emitOpcode(jumpID);
                instructions().uncheckedAppend(src1Index);
                instructions().uncheckedAppend(src2Index);
                instructions().uncheckedAppend(target->bind(begin, instructions().size()));
                return true;
            }
            return false;
        };

        if (m_lastOpcodeID == op_less) {
            if (fuseCompareAndJump(op_jless))
                return;
        } else if (m_lastOpcodeID == op_lesseq) {
            if (fuseCompareAndJump(op_jlesseq))
                return;
        } else if (m_lastOpcodeID == op_greater) {
            if (fuseCompareAndJump(op_jgreater))
                return;
        } else if (m_lastOpcodeID == op_greatereq) {
            if (fuseCompareAndJump(op_jgreatereq))
                return;
        } else if (m_lastOpcodeID == op_eq) {
            if (fuseCompareAndJump(op_jeq))
                return;
        } else if (m_lastOpcodeID == op_stricteq) {
            if (fuseCompareAndJump(op_jstricteq))
                return;
        } else if (m_lastOpcodeID == op_neq) {
            if (fuseCompareAndJump(op_jneq))
                return;
        } else if (m_lastOpcodeID == op_nstricteq) {
            if (fuseCompareAndJump(op_jnstricteq))
                return;
        } else if (m_lastOpcodeID == op_below) {
            if (fuseCompareAndJump(op_jbelow))
                return;
        } else if (m_lastOpcodeID == op_beloweq) {
            if (fuseCompareAndJump(op_jbeloweq))
                return;
		} else if (m_lastOpcodeID == op_eq_null && target->isForward()) {
			int dstIndex;
			int srcIndex;

			retrieveLastUnaryOp(dstIndex, srcIndex);

			if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
				rewindUnaryOp();

				size_t begin = instructions().size();
				emitOpcode(op_jeq_null);
				instructions().uncheckedAppend(srcIndex);
				instructions().uncheckedAppend(target->bind(begin, instructions().size()));
				return;
			}
		} else if (m_lastOpcodeID == op_neq_null && target->isForward()) {
			int dstIndex;
			int srcIndex;

			retrieveLastUnaryOp(dstIndex, srcIndex);

			if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
				rewindUnaryOp();

				size_t begin = instructions().size();
				emitOpcode(op_jneq_null);
				instructions().uncheckedAppend(srcIndex);
				instructions().uncheckedAppend(target->bind(begin, instructions().size()));
				return;
			}
		} else if (m_lastOpcodeID == op_is_undefined_or_null && target->isForward()) {
			int dstIndex;
			int srcIndex;

			retrieveLastUnaryOp(dstIndex, srcIndex);

			if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
				rewindUnaryOp();

				size_t begin = instructions().size();
				emitOpcode(op_jundefined_or_null);
				instructions().uncheckedAppend(srcIndex);
				instructions().uncheckedAppend(target->bind(begin, instructions().size()));
				return;
			}
		}
    }

    size_t begin = instructions().size();

    emitOpcode(op_jtrue);
    instructions().uncheckedAppend(cond->index());
    instructions().uncheckedAppend(target->bind(begin, instructions().size()));
}

void BytecodeGenerator::emitJumpIfFalse(RegisterID* cond, Label* target)
{
    auto fuseCompareAndJump = [&] (OpcodeID jumpID, bool replaceOperands) {
        int dstIndex;
        int src1Index;
        int src2Index;

        retrieveLastBinaryOp(dstIndex, src1Index, src2Index);

        if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
            rewindBinaryOp();

            size_t begin = instructions().size();
            emitOpcode(jumpID);
            // Since op_below and op_beloweq only accepts Int32, replacing operands is not observable to users.
            if (replaceOperands)
                std::swap(src1Index, src2Index);
            instructions().uncheckedAppend(src1Index);
            instructions().uncheckedAppend(src2Index);
            instructions().uncheckedAppend(target->bind(begin, instructions().size()));
            return true;
        }
        return false;
    };

    if (m_lastOpcodeID == op_less && target->isForward()) {
        if (fuseCompareAndJump(op_jnless, false))
            return;
    } else if (m_lastOpcodeID == op_lesseq && target->isForward()) {
        if (fuseCompareAndJump(op_jnlesseq, false))
            return;
    } else if (m_lastOpcodeID == op_greater && target->isForward()) {
        if (fuseCompareAndJump(op_jngreater, false))
            return;
    } else if (m_lastOpcodeID == op_greatereq && target->isForward()) {
        if (fuseCompareAndJump(op_jngreatereq, false))
            return;
    } else if (m_lastOpcodeID == op_eq && target->isForward()) {
        if (fuseCompareAndJump(op_jneq, false))
            return;
    } else if (m_lastOpcodeID == op_stricteq && target->isForward()) {
        if (fuseCompareAndJump(op_jnstricteq, false))
            return;
    } else if (m_lastOpcodeID == op_neq && target->isForward()) {
        if (fuseCompareAndJump(op_jeq, false))
            return;
    } else if (m_lastOpcodeID == op_nstricteq && target->isForward()) {
        if (fuseCompareAndJump(op_jstricteq, false))
            return;
    } else if (m_lastOpcodeID == op_below && target->isForward()) {
        if (fuseCompareAndJump(op_jbeloweq, true))
            return;
    } else if (m_lastOpcodeID == op_beloweq && target->isForward()) {
        if (fuseCompareAndJump(op_jbelow, true))
            return;
    } else if (m_lastOpcodeID == op_not) {
        int dstIndex;
        int srcIndex;

        retrieveLastUnaryOp(dstIndex, srcIndex);

        if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
            rewindUnaryOp();

            size_t begin = instructions().size();
            emitOpcode(op_jtrue);
            instructions().uncheckedAppend(srcIndex);
            instructions().uncheckedAppend(target->bind(begin, instructions().size()));
            return;
        }
    } else if (m_lastOpcodeID == op_eq_null && target->isForward()) {
        int dstIndex;
        int srcIndex;

        retrieveLastUnaryOp(dstIndex, srcIndex);

        if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
            rewindUnaryOp();

            size_t begin = instructions().size();
            emitOpcode(op_jneq_null);
            instructions().uncheckedAppend(srcIndex);
            instructions().uncheckedAppend(target->bind(begin, instructions().size()));
            return;
        }
    } else if (m_lastOpcodeID == op_neq_null && target->isForward()) {
        int dstIndex;
        int srcIndex;

        retrieveLastUnaryOp(dstIndex, srcIndex);

        if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
            rewindUnaryOp();

            size_t begin = instructions().size();
            emitOpcode(op_jeq_null);
            instructions().uncheckedAppend(srcIndex);
            instructions().uncheckedAppend(target->bind(begin, instructions().size()));
            return;
        }
    } else if (m_lastOpcodeID == op_is_undefined_or_null && target->isForward()) {
        int dstIndex;
        int srcIndex;

        retrieveLastUnaryOp(dstIndex, srcIndex);

        if (cond->index() == dstIndex && cond->isTemporary() && !cond->refCount()) {
            rewindUnaryOp();

            size_t begin = instructions().size();
            emitOpcode(op_jnundefined_or_null);
            instructions().uncheckedAppend(srcIndex);
            instructions().uncheckedAppend(target->bind(begin, instructions().size()));
            return;
        }
    }

    size_t begin = instructions().size();
    emitOpcode(op_jfalse);
    instructions().uncheckedAppend(cond->index());
    instructions().uncheckedAppend(target->bind(begin, instructions().size()));
}

PassRefPtr<Label> BytecodeGenerator::emitJumpIfNotFunctionCall(RegisterID* cond, Label* target)
{
    size_t begin = instructions().size();

    emitOpcode(op_jneq_ptr);
    instructions().uncheckedAppend(cond->index());
    instructions().uncheckedAppend(Special::CallFunction);
    instructions().uncheckedAppend(target->bind(begin, instructions().size()));
    instructions().uncheckedAppend(0);
    return target;
}

PassRefPtr<Label> BytecodeGenerator::emitJumpIfNotFunctionApply(RegisterID* cond, Label* target)
{
    size_t begin = instructions().size();

    emitOpcode(op_jneq_ptr);
    instructions().uncheckedAppend(cond->index());
    instructions().uncheckedAppend(Special::ApplyFunction);
    instructions().uncheckedAppend(target->bind(begin, instructions().size()));
    instructions().uncheckedAppend(0);
    return target;
}

ALWAYS_INLINE unsigned BytecodeGenerator::addConstant(const Identifier& ident)
{
    auto* rep = ident.impl();
    auto result = m_identifierMap.fastAdd(rep, m_codeBlock->numberOfIdentifiers());
    if (result.isNewEntry)
        m_codeBlock->addIdentifier(ident);

    return result.iterator->second;
}

// We can't hash JSValue(), so we use a dedicated data member to cache it.
RegisterID* BytecodeGenerator::addConstantEmptyValue()
{
    if (!m_emptyValueRegister) {
        int index = addConstantIndex();
        m_codeBlock->addConstant(JSValue());
        m_emptyValueRegister = &m_constantPoolRegisters[index];
    }

    return m_emptyValueRegister;
}

RegisterID* BytecodeGenerator::addConstantValue(JSValue v, SourceCodeRepresentation sourceCodeRepresentation)
{
    if (!v)
        return addConstantEmptyValue();

    int index = m_nextConstantOffset;

    if (sourceCodeRepresentation == SourceCodeRepresentation::Double && v.isInt32())
        v = jsDoubleNumber(v.asNumber());
    EncodedJSValueWithRepresentation valueMapKey = EncodedJSValueWithRepresentation(JSValue::encode(v), sourceCodeRepresentation);
	auto result = m_jsValueMap.fastAdd(valueMapKey, m_nextConstantOffset);
    if (result.isNewEntry) {
        addConstantIndex();
        m_codeBlock->addConstant(v, sourceCodeRepresentation);
    } else
        index = result.iterator->second;

    return &m_constantPoolRegisters[index];
}

RegisterID* BytecodeGenerator::emitMove(RegisterID* dst, RegisterID* src)
{
    m_staticPropertyAnalyzer.mov(dst->index(), src->index());
    emitOpcode(op_mov);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitMoveLinkTimeConstant(RegisterID* dst, LinkTimeConstant type)
{
    unsigned constantIndex = static_cast<unsigned>(type);
    if (!m_linkTimeConstantRegisters[constantIndex]) {
        int index = addConstantIndex();
        m_linkTimeConstantCodeBlockConstantIndexes[constantIndex] = m_codeBlock->addConstant(type);
        m_linkTimeConstantRegisters[constantIndex] = &m_constantPoolRegisters[index];
    }

    if (!dst)
        return m_linkTimeConstantRegisters[constantIndex];

    emitOpcode(op_mov);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(m_linkTimeConstantRegisters[constantIndex]->index());

    return dst;
}

RegisterID* BytecodeGenerator::moveEmptyValue(RegisterID* dst)
{
    RefPtr<RegisterID> emptyValue = addConstantEmptyValue();

    emitOpcode(op_mov);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(emptyValue->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitUnaryOp(OpcodeID opcodeID, RegisterID* dst, RegisterID* src, OperandTypes types)
{
	// has a value profile
	if (opcodeID == op_to_number || opcodeID == op_bitnot || opcodeID == op_to_numeric)
		return emitUnaryOpProfiled(opcodeID, dst, src);

    emitOpcode(opcodeID);
    instructions().append(dst->index());
    instructions().append(src->index());

    // has an Arith Profile
    if (opcodeID == op_negate)
        instructions().append(ArithProfile(types.first()).bits());
    return dst;
}

RegisterID* BytecodeGenerator::emitUnaryOpProfiled(OpcodeID opcodeID, RegisterID* dst, RegisterID* src)
{
    UnlinkedValueProfile profile = emitProfiledOpcode(opcodeID);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitToObject(RegisterID* dst, RegisterID* src, const Identifier& message)
{
    UnlinkedValueProfile profile = emitProfiledOpcode(op_to_object);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    instructions().uncheckedAppend(addConstant(message));
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitInc(RegisterID* srcDst)
{
    emitOpcode(op_inc);
    instructions().uncheckedAppend(srcDst->index());
    instructions().uncheckedAppend(ArithProfile().bits());
    return srcDst;
}

RegisterID* BytecodeGenerator::emitDec(RegisterID* srcDst)
{
    emitOpcode(op_dec);
    instructions().uncheckedAppend(srcDst->index());
    instructions().uncheckedAppend(ArithProfile().bits());
    return srcDst;
}

RegisterID* BytecodeGenerator::emitBinaryOp(OpcodeID opcodeID, RegisterID* dst, RegisterID* src1, RegisterID* src2, OperandTypes types)
{
    if (opcodeID == op_bitand || opcodeID == op_bitor || opcodeID == op_bitxor || opcodeID == op_lshift || opcodeID == op_rshift) {
        UnlinkedValueProfile profile = emitProfiledOpcode(opcodeID);
        instructions().append(dst->index());
        instructions().append(src1->index());
        instructions().append(src2->index());
        instructions().append(profile);
        return dst;
    }

    emitOpcode(opcodeID);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src1->index());
    instructions().uncheckedAppend(src2->index());

    if (opcodeID == op_add || opcodeID == op_mul || opcodeID == op_sub || opcodeID == op_div)
        instructions().uncheckedAppend(ArithProfile(types.first(), types.second()).bits());

    return dst;
}

bool BytecodeGenerator::emitEqualityOpImpl(RegisterID* dst, RegisterID* src1, RegisterID* src2)
{
    if (!canDoPeepholeOptimization())
        return false;

    if (m_lastOpcodeID == op_typeof) {
        int dstIndex;
        int srcIndex;

        retrieveLastUnaryOp(dstIndex, srcIndex);

        if (src1->index() == dstIndex
            && src1->isTemporary()
            && CodeBlock::isConstantRegisterIndex(src2->index())
            && m_codeBlock->constantRegister(src2->index()).get().isString()) {
            const String& value = asString(m_codeBlock->constantRegister(src2->index()).get())->tryGetValue();
            if (value == "undefined") {
                rewindUnaryOp();
                emitOpcode(op_typeof_is_undefined);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(srcIndex);
                return true;
            }
            if (value == "boolean") {
                rewindUnaryOp();
                emitOpcode(op_is_boolean);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(srcIndex);
                return true;
            }
            if (value == "number") {
                rewindUnaryOp();
                emitOpcode(op_is_number);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(srcIndex);
                return true;
            }
            if (value == "string") {
                rewindUnaryOp();
                emitOpcode(op_is_cell_with_type);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(srcIndex);
                instructions().uncheckedAppend(StringType);
                return true;
            }
            if (value == "symbol") {
                rewindUnaryOp();
                emitOpcode(op_is_cell_with_type);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(srcIndex);
                instructions().uncheckedAppend(SymbolType);
                return true;
            }
            if (Options::useBigInt() && value == "bigint") {
                rewindUnaryOp();
#if USE(BIGINT32)
                OpIsBigInt::emit(this, dst, op.m_value);
#else
                emitOpcode(op_is_cell_with_type);
                instructions().append(dst->index());
                instructions().append(srcIndex);
                instructions().append(HeapBigIntType);
#endif
                return true;
            }
            if (value == "object") {
                rewindUnaryOp();
                emitOpcode(op_typeof_is_object);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(srcIndex);
                return true;
            }
            if (value == "function") {
                rewindUnaryOp();
                emitOpcode(op_typeof_is_function);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(srcIndex);
                return true;
            }
        }
    }

    return false;
}

RegisterID* BytecodeGenerator::emitEqualityOp(OpcodeID opcodeID, RegisterID* dst, RegisterID* src1, RegisterID* src2)
{
	ASSERT(opcodeID == op_eq || opcodeID == op_stricteq);
    if (!emitEqualityOpImpl(dst, src1, src2)) {
        //EqOp::emit(this, dst, src1, src2);
        emitOpcode(opcodeID);
        instructions().uncheckedAppend(dst->index());
        instructions().uncheckedAppend(src1->index());
        instructions().uncheckedAppend(src2->index());
    }
    return dst;
}

#if 0
void BytecodeGenerator::emitTypeProfilerExpressionInfo(const JSTextPosition& startDivot, const JSTextPosition& endDivot)
{
    unsigned start = startDivot.offset; // Ranges are inclusive of their endpoints, AND 0 indexed.
    unsigned end = endDivot.offset - 1; // End Ranges already go one past the inclusive range, so subtract 1.
    unsigned instructionOffset = instructions().size() - 1;
    m_codeBlock->addTypeProfilerExpressionInfo(instructionOffset, start, end);
}

void BytecodeGenerator::emitProfileType(RegisterID* registerToProfile, ProfileTypeBytecodeFlag flag, const Identifier* identifier)
{
    if (flag == ProfileTypeBytecodeGetFromScope || flag == ProfileTypeBytecodePutToScope)
        RELEASE_ASSERT(identifier);

    // The format of this instruction is: op_profile_type regToProfile, TypeLocation*, flag, identifier?, resolveType?
    emitOpcode(op_profile_type);
    instructions().uncheckedAppend(registerToProfile->index());
    instructions().uncheckedAppend(labelScopeDepth());
    instructions().uncheckedAppend(flag);
    instructions().uncheckedAppend(identifier ? addConstant(*identifier) : 0);
    instructions().uncheckedAppend(resolveType());
}

void BytecodeGenerator::emitProfileControlFlow(int textOffset)
{
    if (vm().controlFlowProfiler()) {
        RELEASE_ASSERT(textOffset >= 0);
        size_t bytecodeOffset = instructions().size();
        m_codeBlock->addOpProfileControlFlowBytecodeOffset(bytecodeOffset);

        emitOpcode(op_profile_control_flow);
        instructions().uncheckedAppend(textOffset);
    }
}
#endif

unsigned BytecodeGenerator::addConstantIndex()
{
    unsigned index = m_nextConstantOffset;
    m_constantPoolRegisters.append(FirstConstantRegisterIndex + m_nextConstantOffset);
    ++m_nextConstantOffset;
    return index;
}

RegisterID* BytecodeGenerator::emitLoad(RegisterID* dst, bool b)
{
    return emitLoad(dst, jsBoolean(b));
}

RegisterID* BytecodeGenerator::emitLoad(RegisterID* dst, const Identifier& identifier)
{
    ASSERT(!identifier.isSymbol());
    JSString*& stringInMap = m_stringMap.add(identifier.impl(), reinterpret_cast<JSString*>(0)).iterator->second;
    if (!stringInMap)
        stringInMap = jsOwnedString(vm(), identifier.string());

    return emitLoad(dst, JSValue(stringInMap));
}

RegisterID* BytecodeGenerator::emitLoad(RegisterID* dst, JSValue v, SourceCodeRepresentation sourceCodeRepresentation)
{
    RegisterID* constantID = addConstantValue(v, sourceCodeRepresentation);
    if (dst)
        return move(dst, constantID);
    return constantID;
}

RegisterID* BytecodeGenerator::emitLoad(RegisterID* dst, IdentifierSet&& set)
{
    unsigned setIndex = m_codeBlock->addSetConstant(WTFMove(set));
    return emitLoad(dst, jsNumber(setIndex));
}

RegisterID* BytecodeGenerator::emitLoadGlobalObject(RegisterID* dst)
{
    if (!m_globalObjectRegister) {
        int index = addConstantIndex();
        m_codeBlock->addConstant(JSValue());
        m_globalObjectRegister = &m_constantPoolRegisters[index];
        m_codeBlock->setGlobalObjectRegister(index);
    }
    if (dst)
        move(dst, m_globalObjectRegister);
    return m_globalObjectRegister;
}

template<typename LookUpVarKindFunctor>
bool BytecodeGenerator::instantiateLexicalVariables(const VariableEnvironment& lexicalVariables, ScopeType scopeType, SymbolTable* symbolTable, ScopeRegisterType scopeRegisterType, LookUpVarKindFunctor lookUpVarKind)
{
    bool hasCapturedVariables = false;
    {
        bool hasPrivateNames = lexicalVariables.privateNamesSize();
        // We need to ensure that @privateClassBrand and @privateBrand offsets are 0 and 1 respectively.
        // To ensure that, we first define them, and later, we filter them out from lexicalVariables.
        if (scopeType == ScopeType::ClassScope) {
            if (hasPrivateNames) {
                hasCapturedVariables = true;
                VarOffset privateClassBrandOffset = VarOffset(symbolTable->takeNextScopeOffset());
                ASSERT(privateClassBrandOffset.rawOffset() == PrivateNameEntry::privateClassBrandOffset);
                symbolTable->add(propertyNames().builtinNames().privateClassBrandPrivateName().impl(), SymbolTableEntry(privateClassBrandOffset, static_cast<unsigned>(PropertyAttribute::ReadOnly)));

                VarOffset privateBrandOffset = VarOffset(symbolTable->takeNextScopeOffset());
                ASSERT(privateBrandOffset.rawOffset() == PrivateNameEntry::privateBrandOffset);
                symbolTable->add(propertyNames().builtinNames().privateBrandPrivateName().impl(), SymbolTableEntry(privateBrandOffset, static_cast<unsigned>(PropertyAttribute::ReadOnly)));
            }
        }

        //ConcurrentJITLocker locker(symbolTable->m_lock);
        for (auto& entry : lexicalVariables) {
            ASSERT(entry.value.isLet() || entry.value.isConst() || entry.value.isFunction());
            ASSERT(!entry.value.isVar());

            auto* key = entry.first.get();
            if (scopeType == ScopeType::ClassScope) {
                if (hasPrivateNames) {
                    if (key == propertyNames().builtinNames().privateClassBrandPrivateName()) {
                        ASSERT(entry.value.isConst());
                        continue;
                    }
                    if (key == propertyNames().builtinNames().privateBrandPrivateName()) {
                        ASSERT(entry.value.isConst());
                        continue;
                    }
                } else {
                    ASSERT(key != propertyNames().builtinNames().privateClassBrandPrivateName());
                    ASSERT(key != propertyNames().builtinNames().privateBrandPrivateName());
                }
            }

#if ASSERT_ENABLED
            SymbolTableEntry symbolTableEntry = symbolTable->get(NoLockingNecessary, key);
            ASSERT(symbolTableEntry.isNull());
#endif

            // Imported bindings which are not the namespace bindings are not allocated
            // in the module environment as usual variables' way.
            // And since these types of the variables only seen in the module environment,
            // other lexical environment need not to take care this.
            if (entry.second.isImported() && !entry.second.isImportedNamespace())
                continue;

            VarKind varKind = lookUpVarKind(key, entry.second);
            VarOffset varOffset;
            if (varKind == VarKind::Scope) {
                varOffset = VarOffset(symbolTable->takeNextScopeOffset());
                hasCapturedVariables = true;
            } else {
                ASSERT(varKind == VarKind::Stack);
                RegisterID* local;
                if (scopeRegisterType == ScopeRegisterType::Block) {
                    local = newBlockScopeVariable();
                    local->ref();
                } else
                    local = addVar();
                varOffset = VarOffset(local->virtualRegister());
            }

            SymbolTableEntry newEntry(varOffset, entry.second.isConst() ? ReadOnly : 0);
            symbolTable->add(key, newEntry);

            // FIXME: only do this if there is an eval() within a nested scope --- otherwise it isn't needed.
            // https://bugs.webkit.org/show_bug.cgi?id=206663

            const PrivateNameEnvironment* privateEnvironment = lexicalVariables.privateNameEnvironment();
            if (!privateEnvironment)
                continue;

            auto findResult = privateEnvironment->find(key);

            if (findResult == privateEnvironment->end())
                continue;

            symbolTable->addPrivateName(findResult->first.get(), findResult->second);
        }
    }
    return hasCapturedVariables;
}

void BytecodeGenerator::emitPrefillStackTDZVariables(const VariableEnvironment& lexicalVariables, SymbolTable* symbolTable)
{
    // Prefill stack variables with the TDZ empty value.
    // Scope variables will be initialized to the TDZ empty value when JSLexicalEnvironment is allocated.
	for (auto& entry : lexicalVariables) {
        // Imported bindings which are not the namespace bindings are not allocated
        // in the module environment as usual variables' way.
        // And since these types of the variables only seen in the module environment,
        // other lexical environment need not to take care this.
        if (entry.second.isImported() && !entry.second.isImportedNamespace())
            continue;

        if (entry.second.isFunction())
            continue;

        SymbolTableEntry symbolTableEntry = symbolTable->get(entry.first.get());
        ASSERT(!symbolTableEntry.isNull());
        VarOffset offset = symbolTableEntry.varOffset();
        if (offset.isScope())
            continue;

        ASSERT(offset.isStack());
        moveEmptyValue(&registerFor(offset.stackOffset()));
    }
}

void BytecodeGenerator::pushLexicalScope(VariableEnvironmentNode* node, ScopeType scopeType, TDZCheckOptimization tdzCheckOptimization, NestedScopeType nestedScopeType, RegisterID** constantSymbolTableResult, bool shouldInitializeBlockScopedFunctions)
{
    VariableEnvironment& environment = node->lexicalVariables();
    RegisterID* constantSymbolTableResultTemp = 0;
    pushLexicalScopeInternal(environment, tdzCheckOptimization, nestedScopeType, &constantSymbolTableResultTemp, TDZRequirement::UnderTDZ, scopeType, ScopeRegisterType::Block);

    if (shouldInitializeBlockScopedFunctions)
        initializeBlockScopedFunctions(environment, node->functionStack(), constantSymbolTableResultTemp);

    if (constantSymbolTableResult && constantSymbolTableResultTemp)
        *constantSymbolTableResult = constantSymbolTableResultTemp;
}

void BytecodeGenerator::pushClassHeadLexicalScope(VariableEnvironment& environment)
{
    pushLexicalScopeInternal(environment, TDZCheckOptimization::Optimize, NestedScopeType::IsNested, nullptr, TDZRequirement::UnderTDZ, ScopeType::LetConstScope, ScopeRegisterType::Block);
}

void BytecodeGenerator::popClassHeadLexicalScope(VariableEnvironment& environment)
{
    popLexicalScopeInternal(environment);
}

void BytecodeGenerator::pushLexicalScopeInternal(VariableEnvironment& environment, TDZCheckOptimization tdzCheckOptimization, NestedScopeType nestedScopeType,
    RegisterID** constantSymbolTableResult, TDZRequirement tdzRequirement, ScopeType scopeType, ScopeRegisterType scopeRegisterType)
{
    if (!environment.totalSize())
        return;

    if (m_shouldEmitDebugHooks)
        environment.markAllVariablesAsCaptured();

    SymbolTable* symbolTable = SymbolTable::create(*m_vm);
    switch (scopeType) {
	case ScopeType::CatchScope:
		symbolTable->setScopeType(SymbolTable::ScopeType::CatchScope);
		break;
    case ScopeType::CatchScopeWithSimpleParameter:
        symbolTable->setScopeType(SymbolTable::ScopeType::CatchScopeWithSimpleParameter);
        break;
	case ScopeType::LetConstScope:
    case ScopeType::ClassScope:
		symbolTable->setScopeType(SymbolTable::ScopeType::LexicalScope);
		break;
    case ScopeType::FunctionNameScope:
        symbolTable->setScopeType(SymbolTable::ScopeType::FunctionNameScope);
        break;
	}

    if (nestedScopeType == NestedScopeType::IsNested)
        symbolTable->markIsNestedLexicalScope();

    class LookUpVarKindFunctor
	{
	public:
    	VarKind operator()(UniquedStringImpl*, const VariableEnvironmentEntry& entry) {
    		return entry.isCaptured() ? VarKind::Scope : VarKind::Stack;
    	}
    };

    bool hasCapturedVariables = instantiateLexicalVariables(environment, scopeType, symbolTable, scopeRegisterType, LookUpVarKindFunctor());

    RegisterID* newScope = nullptr;
    RegisterID* constantSymbolTable = nullptr;
    int symbolTableConstantIndex = 0;
    if (hasCapturedVariables) {
        if (scopeRegisterType == ScopeRegisterType::Block) {
        	newScope = newBlockScopeVariable();
        	newScope->ref();
        } else
            newScope = addVar();
        if (!constantSymbolTable) {
            ASSERT(!vm()->typeProfiler());
            constantSymbolTable = addConstantValue(symbolTable->cloneScopePart(*m_vm));
            symbolTableConstantIndex = constantSymbolTable->index();
        }
        if (constantSymbolTableResult)
            *constantSymbolTableResult = constantSymbolTable;

        emitOpcode(op_create_lexical_environment);
        instructions().uncheckedAppend(newScope->index());
        instructions().uncheckedAppend(scopeRegister()->index());
        instructions().uncheckedAppend(constantSymbolTable->index());
        instructions().uncheckedAppend(addConstantValue(tdzRequirement == TDZRequirement::UnderTDZ ? jsTDZValue() : jsUndefined())->index());

        move(scopeRegister(), newScope);

        pushLocalControlFlowScope();
    }

    bool isWithScope = false;
	m_lexicalScopeStack.append(LexicalScopeStackEntry{ symbolTable, newScope, isWithScope, symbolTableConstantIndex });
	pushTDZVariables(environment, tdzCheckOptimization, tdzRequirement);

    if (tdzRequirement == TDZRequirement::UnderTDZ)
        emitPrefillStackTDZVariables(environment, symbolTable);
}

void BytecodeGenerator::initializeBlockScopedFunctions(VariableEnvironment& environment, FunctionStack& functionStack, RegisterID* constantSymbolTable)
{
    /*
     * We must transform block scoped function declarations in strict mode like so:
     *
     * function foo() {
     *     if (c) {
     *           function foo() { ... }
     *           if (bar) { ... }
     *           else { ... }
     *           function baz() { ... }
     *     }
     * }
     *
     * to:
     *
     * function foo() {
     *     if (c) {
     *         let foo = function foo() { ... }
     *         let baz = function baz() { ... }
     *         if (bar) { ... }
     *         else { ... }
     *     }
     * }
     *
     * But without the TDZ checks.
    */

    if (!environment.totalSize()) {
        RELEASE_ASSERT(!functionStack.size());
        return;
    }

    if (!functionStack.size())
        return;

    SymbolTable* symbolTable = m_lexicalScopeStack.last().m_symbolTable;
    RegisterID* scope = m_lexicalScopeStack.last().m_scope;
    RefPtr<RegisterID> temp = newTemporary();
    int symbolTableIndex = constantSymbolTable ? constantSymbolTable->index() : 0;
    for (int i = 0; i < functionStack.size(); ++i) {
    	FunctionMetadataNode* function = functionStack[i];
        const Identifier& name = function->ident();
        auto iter = environment.find(name.impl());
        RELEASE_ASSERT(iter != environment.end());
        RELEASE_ASSERT(iter->value.isFunction());
        // We purposefully don't hold the symbol table lock around this loop because emitNewFunctionExpressionCommon may GC.
        SymbolTableEntry entry = symbolTable->get(name.impl());
        RELEASE_ASSERT(!entry.isNull());
        emitNewFunctionExpressionCommon(temp.get(), function);
        bool isLexicallyScoped = true;
        emitPutToScope(scope, variableForLocalEntry(name, entry, symbolTableIndex, isLexicallyScoped), temp.get(), DoNotThrowIfNotFound, InitializationMode::Initialization);
    }
}

void BytecodeGenerator::hoistSloppyModeFunctionIfNecessary(FunctionMetadataNode* metadata)
{
    if (metadata->isSloppyModeHoistedFunction()) {
        const Identifier& functionName = metadata->ident();
        if (isGeneratorOrAsyncFunctionBodyParseMode(parseMode())) {
            RELEASE_ASSERT(m_generatorOrAsyncWrapperFunctionParameterNames);
            if (m_generatorOrAsyncWrapperFunctionParameterNames->contains(functionName))
                return;
        }

		Variable currentFunctionVariable = variable(functionName);
		RefPtr<RegisterID> currentValue;
		if (RegisterID* local = currentFunctionVariable.local())
			currentValue = local;
		else {
			RefPtr<RegisterID> scope = emitResolveScope(0, currentFunctionVariable);
			currentValue = emitGetFromScope(newTemporary(), scope.get(), currentFunctionVariable, DoNotThrowIfNotFound);
		}

        switch (codeType()) {
        case FunctionCode: {
			ASSERT(m_varScopeLexicalScopeStackIndex);
			ASSERT(*m_varScopeLexicalScopeStackIndex < m_lexicalScopeStack.size());
			LexicalScopeStackEntry varScope = m_lexicalScopeStack[m_varScopeLexicalScopeStackIndex.value()];
			SymbolTable* varSymbolTable = varScope.m_symbolTable;
			ASSERT(varSymbolTable->scopeType() == SymbolTable::ScopeType::VarScope);
			SymbolTableEntry entry = varSymbolTable->get(functionName.impl());
			if (functionName == propertyNames().arguments && entry.isNull()) {
				// "arguments" might be put in the parameter scope when we have a non-simple
				// parameter list since "arguments" is visible to expressions inside the
				// parameter evaluation list.
				// e.g:
				// function foo(x = arguments) { { function arguments() { } } }
				RELEASE_ASSERT(*m_varScopeLexicalScopeStackIndex > 0);
				varScope = m_lexicalScopeStack[m_varScopeLexicalScopeStackIndex.value() - 1];
				SymbolTable* parameterSymbolTable = varScope.m_symbolTable;
				entry = parameterSymbolTable->get(functionName.impl());
			}
			RELEASE_ASSERT(!entry.isNull());
			bool isLexicallyScoped = false;
			emitPutToScope(varScope.m_scope, variableForLocalEntry(functionName, entry, varScope.m_symbolTableConstantIndex, isLexicallyScoped), currentValue.get(), DoNotThrowIfNotFound, InitializationMode::NotInitialization);
			break;
		}
		case GlobalCode:
		case EvalCode: {
			RefPtr<RegisterID> scopeId = emitResolveScopeForHoistingFuncDeclInEval(0, functionName);
			RefPtr<RegisterID> checkResult = emitIsUndefined(newTemporary(), scopeId.get());

			RefPtr<Label> isNotVarScopeLabel = newLabel();
			emitJumpIfTrue(checkResult.get(), isNotVarScopeLabel.get());

			// Put to outer scope
			emitPutToScopeDynamic(scopeId.get(), functionName, currentValue.get(), DoNotThrowIfNotFound, InitializationMode::NotInitialization);
			emitLabel(isNotVarScopeLabel.get());
            break;
        }
        case ModuleCode:
            RELEASE_ASSERT_NOT_REACHED();
        }
    }
}

RegisterID* BytecodeGenerator::emitResolveScopeForHoistingFuncDeclInEval(RegisterID* dst, const Identifier& property)
{
	RefPtr<RegisterID> result = finalDestination(dst);
	RefPtr<RegisterID> scope = newTemporary();

    emitOpcode(op_get_scope);
    instructions().uncheckedAppend(scope->index());

	emitOpcode(op_resolve_scope_for_hoisting_func_decl_in_eval);
	instructions().uncheckedAppend(kill(result.get()));
	instructions().uncheckedAppend(scope->index());
	instructions().uncheckedAppend(addConstant(property));
    return result.get();
}

void BytecodeGenerator::popLexicalScope(VariableEnvironmentNode* node)
{
    VariableEnvironment& environment = node->lexicalVariables();
    popLexicalScopeInternal(environment);
}

void BytecodeGenerator::popLexicalScopeInternal(VariableEnvironment& environment)
{
	// NOTE: This function only makes sense for scopes that aren't ScopeRegisterType::Var (only function name scope right now is ScopeRegisterType::Var).
    // This doesn't make sense for ScopeRegisterType::Var because we deref RegisterIDs here.
    if (!environment.totalSize())
        return;

    if (m_shouldEmitDebugHooks)
        environment.markAllVariablesAsCaptured();

    LexicalScopeStackEntry stackEntry = m_lexicalScopeStack.takeLast();
    SymbolTable* symbolTable = stackEntry.m_symbolTable;
    //ConcurrentJITLocker locker(symbolTable->m_lock);
    bool hasCapturedVariables = false;
    for (auto it = environment.begin(); it != environment.end(); ++it) {
        if (it->second.isCaptured()) {
            hasCapturedVariables = true;
            continue;
        }
        SymbolTableEntry symbolTableEntry = symbolTable->get(it->first.get());
        ASSERT(!symbolTableEntry.isNull());
        VarOffset offset = symbolTableEntry.varOffset();
        ASSERT(offset.isStack());
        RegisterID* local = &registerFor(offset.stackOffset());
        local->deref();
    }

    if (hasCapturedVariables) {
        RELEASE_ASSERT(stackEntry.m_scope);
        emitGetParentScope(scopeRegister(), stackEntry.m_scope);
        popLocalControlFlowScope();
        stackEntry.m_scope->deref();
    }

	m_TDZStack.removeLast();
}

void BytecodeGenerator::prepareLexicalScopeForNextForLoopIteration(VariableEnvironmentNode* node, RegisterID* loopSymbolTable)
{
    auto& environment = node->lexicalVariables();
    if (!environment.totalSize())
        return;
    if (m_shouldEmitDebugHooks)
        environment.markAllVariablesAsCaptured();
    if (!environment.hasCapturedVariables())
        return;

    RELEASE_ASSERT(loopSymbolTable);

    // This function needs to do setup for a for loop's activation if any of
    // the for loop's lexically declared variables are captured (that is, variables
    // declared in the loop header, not the loop body). This function needs to
    // make a copy of the current activation and copy the values from the previous
    // activation into the new activation because each iteration of a for loop
    // gets a new activation.

    LexicalScopeStackEntry stackEntry = m_lexicalScopeStack.last();
    SymbolTable* symbolTable = stackEntry.m_symbolTable;
    RegisterID* loopScope = stackEntry.m_scope;
    ASSERT(symbolTable->scopeSize());
    ASSERT(loopScope);
    Vector<std::pair<RegisterID*, Identifier>> activationValuesToCopyOver;

    {
        //ConcurrentJITLocker locker(symbolTable->m_lock);
        activationValuesToCopyOver.reserveInitialCapacity(symbolTable->scopeSize());

        for (auto end = symbolTable->end(), ptr = symbolTable->begin(); ptr != end; ++ptr) {
            if (!ptr->second.varOffset().isScope())
                continue;

            RefPtr<UniquedStringImpl> ident = ptr->first;
            Identifier identifier = Identifier::fromUid(*m_vm, ident.get());

            RegisterID* transitionValue = newBlockScopeVariable();
            transitionValue->ref();
            emitGetFromScope(transitionValue, loopScope, variableForLocalEntry(identifier, ptr->second, loopSymbolTable->index(), true), DoNotThrowIfNotFound);
            activationValuesToCopyOver.uncheckedAppend(std::make_pair(transitionValue, identifier));
        }
    }

    // We need this dynamic behavior of the executing code to ensure
    // each loop iteration has a new activation object. (It's pretty ugly).
    // Also, this new activation needs to be assigned to the same register
    // as the previous scope because the loop body is compiled under
    // the assumption that the scope's register index is constant even
    // though the value in that register will change on each loop iteration.
    emitGetParentScope(scopeRegister(), loopScope);

    emitOpcode(op_create_lexical_environment);
    instructions().uncheckedAppend(loopScope->index());
    instructions().uncheckedAppend(scopeRegister()->index());
    instructions().uncheckedAppend(loopSymbolTable->index());
    instructions().uncheckedAppend(addConstantValue(jsTDZValue())->index());

    move(scopeRegister(), loopScope);

    {
        //ConcurrentJITLocker locker(symbolTable->m_lock);
    	for(int i = 0; i < activationValuesToCopyOver.size(); ++i) {
    		auto& pair = activationValuesToCopyOver[i];
            const Identifier& identifier = pair.second;
            SymbolTableEntry entry = symbolTable->get(identifier.impl());
            RELEASE_ASSERT(!entry.isNull());
            RegisterID* transitionValue = pair.first;
            emitPutToScope(loopScope, variableForLocalEntry(identifier, entry, loopSymbolTable->index(), true), transitionValue, DoNotThrowIfNotFound, InitializationMode::NotInitialization);
            transitionValue->deref();
        }
    }
}

Variable BytecodeGenerator::variable(const Identifier& property, ThisResolutionType thisResolutionType)
{
	if (property == propertyNames().builtinNames().thisPrivateName() && thisResolutionType == ThisResolutionType::Local) {
        return Variable(property, VarOffset(thisRegister()->virtualRegister()), thisRegister(),
            ReadOnly, Variable::SpecialVariable, 0, false);
    }

    // We can optimize lookups if the lexical variable is found before a "with" or "catch"
    // scope because we're guaranteed static resolution. If we have to pass through
    // a "with" or "catch" scope we loose this guarantee.
    // We can't optimize cases like this:
    // {
    //     let x = ...;
    //     with (o) {
    //         doSomethingWith(x);
    //     }
    // }
    // Because we can't gaurantee static resolution on x.
    // But, in this case, we are guaranteed static resolution:
    // {
    //     let x = ...;
    //     with (o) {
    //         let x = ...;
    //         doSomethingWith(x);
    //     }
    // }
    for (unsigned i = m_lexicalScopeStack.size(); i--; ) {
    	LexicalScopeStackEntry& stackEntry = m_lexicalScopeStack[i];
        if (stackEntry.m_isWithScope)
            return Variable(property);
        SymbolTable* symbolTable = stackEntry.m_symbolTable;
        SymbolTableEntry symbolTableEntry = symbolTable->get(property.impl());
        if (symbolTableEntry.isNull())
            continue;
        bool resultIsCallee = false;
        if (symbolTable->scopeType() == SymbolTable::ScopeType::FunctionNameScope) {
            if (m_usesNonStrictEval) {
                // We don't know if an eval has introduced a "var" named the same thing as the function name scope variable name.
                // We resort to dynamic lookup to answer this question.
                Variable result = Variable(property);
                return result;
            }
            resultIsCallee = true;
        }
        Variable result = variableForLocalEntry(property, symbolTableEntry, stackEntry.m_symbolTableConstantIndex, symbolTable->scopeType() == SymbolTable::ScopeType::LexicalScope);
        if (resultIsCallee)
            result.setIsReadOnly();
        return result;
    }

	return Variable(property);
}

Variable BytecodeGenerator::variableForLocalEntry(
    const Identifier& property, const SymbolTableEntry& entry, int symbolTableConstantIndex, bool isLexicallyScoped)
{
    VarOffset offset = entry.varOffset();

    RegisterID* local;
    if (offset.isStack())
        local = &registerFor(offset.stackOffset());
    else
        local = 0;
	return Variable(property, offset, local, entry.getAttributes(), Variable::NormalVariable, symbolTableConstantIndex, isLexicallyScoped);
}

void BytecodeGenerator::createVariable(
    const Identifier& property, VarKind varKind, SymbolTable* symbolTable, ExistingVariableMode existingVariableMode)
{
    ASSERT(property != propertyNames().builtinNames().thisPrivateName());
    //ConcurrentJITLocker locker(symbolTable.m_lock);
    SymbolTableEntry entry = symbolTable->get(property.impl());

    if (!entry.isNull()) {
        if (existingVariableMode == IgnoreExisting)
            return;

        // Do some checks to ensure that the variable we're being asked to create is sufficiently
        // compatible with the one we have already created.

        VarOffset offset = entry.varOffset();

        // We can't change our minds about whether it's captured.
        if (offset.kind() != varKind) {
#if 0
            dataLog(
				"Trying to add variable called ", property, " as ", varKind,
				" but it was already added as ", offset, ".\n");
#endif
            RELEASE_ASSERT_NOT_REACHED();
        }

        return;
    }

    VarOffset varOffset;
    if (varKind == VarKind::Scope)
        varOffset = VarOffset(symbolTable->takeNextScopeOffset());
    else {
        ASSERT(varKind == VarKind::Stack);
        varOffset = VarOffset(VirtualRegister_(virtualRegisterForLocal(m_calleeLocals.size())));
    }
    SymbolTableEntry newEntry(varOffset, 0);
    symbolTable->add(property.impl(), newEntry);

    if (varKind == VarKind::Stack) {
        RegisterID* local = addVar();
        RELEASE_ASSERT(local->index() == varOffset.stackOffset().offset());
    }
}

RegisterID* BytecodeGenerator::emitOverridesHasInstance(RegisterID* dst, RegisterID* constructor, RegisterID* hasInstanceValue)
{
    emitOpcode(op_overrides_has_instance);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(constructor->index());
    instructions().uncheckedAppend(hasInstanceValue->index());
    return dst;
}

// Indicates the least upper bound of resolve type based on local scope. The bytecode linker
// will start with this ResolveType and compute the least upper bound including intercepting scopes.
ResolveType BytecodeGenerator::resolveType()
{
    for (unsigned i = m_lexicalScopeStack.size(); i--; ) {
        if (m_lexicalScopeStack[i].m_isWithScope)
            return Dynamic;
        if (m_usesNonStrictEval && m_lexicalScopeStack[i].m_symbolTable->scopeType() == SymbolTable::ScopeType::FunctionNameScope) {
        	// We never want to assign to a FunctionNameScope. Returning Dynamic here achieves this goal.
			// If we aren't in non-strict eval mode, then NodesCodeGen needs to take care not to emit
			// a put_to_scope with the destination being the function name scope variable.
			return Dynamic;
        }
    }

    if (m_usesNonStrictEval)
        return GlobalPropertyWithVarInjectionChecks;
    return GlobalProperty;
}

RegisterID* BytecodeGenerator::emitResolveScope(RegisterID* dst, const Variable& variable)
{
    switch (variable.offset().kind()) {
    case VarKind::Stack:
        return 0;

    case VarKind::DirectArgument:
        return argumentsRegister();

    case VarKind::Scope:
        // This always refers to the activation that *we* allocated, and not the current scope that code
        // lives in. Note that this will change once we have proper support for block scoping. Once that
        // changes, it will be correct for this code to return scopeRegister(). The only reason why we
        // don't do that already is that m_lexicalEnvironment is required by ConstDeclNode. ConstDeclNode
        // requires weird things because it is a shameful pile of nonsense, but block scoping would make
        // that code sensible and obviate the need for us to do bad things.
    	for (unsigned i = m_lexicalScopeStack.size(); i--; ) {
    		LexicalScopeStackEntry& stackEntry = m_lexicalScopeStack[i];
			// We should not resolve a variable to VarKind::Scope if a "with" scope lies in between the current
			// scope and the resolved scope.
			// We'd like to say: RELEASE_ASSERT(!stackEntry.m_isWithOrCatch);
			// But, our current implementation 'const' is a pile of crap and uses variablePerSymbolTable
			// which recklessly ignores the scope stack.
            RELEASE_ASSERT(!stackEntry.m_isWithScope);

			if (stackEntry.m_symbolTable->inlineGet(variable.ident().impl()).isNull())
				continue;

			RegisterID* scope = stackEntry.m_scope;
			RELEASE_ASSERT(scope);
			return scope;
		}

		RELEASE_ASSERT_NOT_REACHED();
		return 0;

    case VarKind::Invalid:
        // Indicates non-local resolution.

        // resolve_scope dst, id, ResolveType, depth
        dst = tempDestination(dst);
        emitOpcode(op_resolve_scope);
        instructions().uncheckedAppend(kill(dst));
        instructions().uncheckedAppend(scopeRegister()->index());
        instructions().uncheckedAppend(addConstant(variable.ident()));
        instructions().uncheckedAppend(resolveType());
        instructions().uncheckedAppend(localScopeDepth());
        instructions().uncheckedAppend(0); // symbolTable
        return dst;
    }

    RELEASE_ASSERT_NOT_REACHED();
    return 0;
}

RegisterID* BytecodeGenerator::emitGetFromScope(RegisterID* dst, RegisterID* scope, const Variable& variable, ResolveMode resolveMode)
{
    switch (variable.offset().kind()) {
    case VarKind::Stack:
        return move(dst, variable.local());

    case VarKind::DirectArgument: {
        UnlinkedValueProfile profile = emitProfiledOpcode(op_get_from_arguments);
        instructions().uncheckedAppend(kill(dst));
        instructions().uncheckedAppend(scope->index());
        instructions().uncheckedAppend(variable.offset().capturedArgumentsOffset().offset());
        instructions().uncheckedAppend(profile);
        return dst;
    }

    case VarKind::Scope:
    case VarKind::Invalid: {
        // get_from_scope dst, scope, id, GetPutInfo, Structure, Operand
        UnlinkedValueProfile profile = emitProfiledOpcode(op_get_from_scope);
        instructions().uncheckedAppend(kill(dst));
        instructions().uncheckedAppend(scope->index());
        instructions().uncheckedAppend(addConstant(variable.ident()));
        instructions().uncheckedAppend(GetPutInfo(resolveMode, variable.offset().isScope() ? ResolvedClosureVar : resolveType(), InitializationMode::NotInitialization, ecmaMode()).operand());
        instructions().uncheckedAppend(localScopeDepth()); // or structure
        instructions().uncheckedAppend(variable.offset().isScope() ? variable.offset().scopeOffset().offset() : 0); // or uintptr_t
        instructions().uncheckedAppend(profile);
        return dst;
    } }

    RELEASE_ASSERT_NOT_REACHED();
    return 0;
}

RegisterID* BytecodeGenerator::emitPutToScope(RegisterID* scope, const Variable& variable, RegisterID* value, ResolveMode resolveMode, InitializationMode initializationMode)
{
    switch (variable.offset().kind()) {
    case VarKind::Stack:
        move(variable.local(), value);
        return value;

    case VarKind::DirectArgument:
        emitOpcode(op_put_to_arguments);
        instructions().uncheckedAppend(scope->index());
        instructions().uncheckedAppend(variable.offset().capturedArgumentsOffset().offset());
        instructions().uncheckedAppend(value->index());
        return value;

    case VarKind::Scope:
    case VarKind::Invalid: {
        // put_to_scope scope, id, value, GetPutInfo, Structure, Operand
        emitOpcode(op_put_to_scope);
        instructions().uncheckedAppend(scope->index());
        instructions().uncheckedAppend(addConstant(variable.ident()));
        instructions().uncheckedAppend(value->index());
        ScopeOffset offset;
        if (variable.offset().isScope()) {
            offset = variable.offset().scopeOffset();
            instructions().uncheckedAppend(GetPutInfo(resolveMode, ResolvedClosureVar, initializationMode, ecmaMode()).operand());
            instructions().uncheckedAppend(variable.symbolTableConstantIndex());
        } else {
            ASSERT(resolveType() != ResolvedClosureVar);
            instructions().uncheckedAppend(GetPutInfo(resolveMode, resolveType(), initializationMode, ecmaMode()).operand());
            instructions().uncheckedAppend(localScopeDepth());
        }
        instructions().uncheckedAppend(!!offset ? offset.offset() : 0);
        return value;
    } }

    RELEASE_ASSERT_NOT_REACHED();
    return 0;
}

RegisterID* BytecodeGenerator::emitPutToScopeDynamic(RegisterID* scope, const Identifier& ident, RegisterID* value, ResolveMode resolveMode, InitializationMode initializationMode)
{
    GetPutInfo getPutInfo(resolveMode, Dynamic, initializationMode, ecmaMode());
    unsigned scopeOffset = 0;
    // put_to_scope scope, id, value, GetPutInfo, Structure, Operand
	emitOpcode(op_put_to_scope);
	instructions().uncheckedAppend(scope->index());
	instructions().uncheckedAppend(addConstant(ident));
	instructions().uncheckedAppend(value->index());
	instructions().uncheckedAppend(getPutInfo.operand());
	instructions().uncheckedAppend(localScopeDepth());
	instructions().uncheckedAppend(scopeOffset);
    return value;
}

RegisterID* BytecodeGenerator::initializeVariable(const Variable& variable, RegisterID* value)
{
    RELEASE_ASSERT(variable.offset().kind() != VarKind::Invalid);
    RegisterID* scope = emitResolveScope(0, variable);
    return emitPutToScope(scope, variable, value, ThrowIfNotFound, InitializationMode::NotInitialization);
}

RegisterID* BytecodeGenerator::emitInstanceOf(RegisterID* dst, RegisterID* value, RegisterID* basePrototype)
{ 
    emitOpcode(op_instanceof);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(basePrototype->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitInstanceOfCustom(RegisterID* dst, RegisterID* value, RegisterID* constructor, RegisterID* hasInstanceValue)
{
    emitOpcode(op_instanceof_custom);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(constructor->index());
    instructions().uncheckedAppend(hasInstanceValue->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitIn(RegisterID* dst, RegisterID* property, RegisterID* base)
{
    ArrayProfile* arrayProfile = newArrayProfile();
    emitOpcode(op_in);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(arrayProfile);
    return dst;
}

RegisterID* BytecodeGenerator::emitTryGetById(RegisterID* dst, RegisterID* base, const Identifier& property)
{
    ASSERT_WITH_MESSAGE(!parseIndex(property), "Indexed properties are not supported with tryGetById.");

    UnlinkedValueProfile profile = emitProfiledOpcode(op_try_get_by_id);
    instructions().uncheckedAppend(kill(dst));
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(addConstant(property));
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitGetById(RegisterID* dst, RegisterID* base, const Identifier& property)
{
    ASSERT_WITH_MESSAGE(!parseIndex(property), "Indexed properties should be handled with get_by_val.");

    UnlinkedValueProfile profile = emitProfiledOpcode(op_get_by_id);
    instructions().uncheckedAppend(kill(dst));
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(addConstant(property));
#if ENABLE(LLINT)
    instructions().uncheckedAppend(0); // LLINT structure
    instructions().uncheckedAppend(0); // LLINT cachedOffset
    instructions().uncheckedAppend(0); // LLINT cachedSlot
    instructions().uncheckedAppend(0); // LLINT GetByIdMode
#endif
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitGetById(RegisterID* dst, RegisterID* base, RegisterID* thisVal, const Identifier& property)
{
    ASSERT_WITH_MESSAGE(!parseIndex(property), "Indexed properties should be handled with get_by_val.");

    UnlinkedValueProfile profile = emitProfiledOpcode(op_get_by_id_with_this);
    instructions().uncheckedAppend(kill(dst));
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(thisVal->index());
    instructions().uncheckedAppend(addConstant(property));
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitDirectGetById(RegisterID* dst, RegisterID* base, const Identifier& property)
{
    ASSERT_WITH_MESSAGE(!parseIndex(property), "Indexed properties should be handled with get_by_val_direct.");

    UnlinkedValueProfile profile = emitProfiledOpcode(op_get_by_id_direct);
    instructions().uncheckedAppend(kill(dst));
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(addConstant(property));
#if ENABLE(LLINT)
    instructions().uncheckedAppend(0); // old structure
    instructions().uncheckedAppend(0); // offset
#endif
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitPutById(RegisterID* base, const Identifier& property, RegisterID* value)
{
    ASSERT_WITH_MESSAGE(!parseIndex(property), "Indexed properties should be handled with put_by_val.");

    unsigned propertyIndex = addConstant(property);

    m_staticPropertyAnalyzer.putById(base->index(), propertyIndex);

    emitOpcode(op_put_by_id);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyIndex);
    instructions().uncheckedAppend(value->index());
#if ENABLE(LLINT)
    instructions().uncheckedAppend(0); // old structure
    instructions().uncheckedAppend(0); // offset
    instructions().uncheckedAppend(0); // new structure
    instructions().uncheckedAppend(0); // structure chain
#endif
    instructions().uncheckedAppend(PutByIdFlags::create(ecmaMode())); // is not direct
    return value;
}

RegisterID* BytecodeGenerator::emitPutById(RegisterID* base, RegisterID* thisValue, const Identifier& property, RegisterID* value)
{
    ASSERT_WITH_MESSAGE(!parseIndex(property), "Indexed properties should be handled with put_by_val.");

    unsigned propertyIndex = addConstant(property);

    emitOpcode(op_put_by_id_with_this);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(thisValue->index());
    instructions().uncheckedAppend(propertyIndex);
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(ecmaMode().value());

    return value;
}

RegisterID* BytecodeGenerator::emitDirectPutById(RegisterID* base, const Identifier& property, RegisterID* value)
{
    ASSERT_WITH_MESSAGE(!parseIndex(property), "Indexed properties should be handled with put_by_val(direct).");

    unsigned propertyIndex = addConstant(property);

    m_staticPropertyAnalyzer.putById(base->index(), propertyIndex);
    
    emitOpcode(op_put_by_id);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyIndex);
    instructions().uncheckedAppend(value->index());
#if ENABLE(LLINT)
    instructions().uncheckedAppend(0); // old structure
    instructions().uncheckedAppend(0); // offset
    instructions().uncheckedAppend(0); // new structure
    instructions().uncheckedAppend(0); // structure chain (unused if direct)
#endif
    PutByIdFlags type = PutByIdFlags::createDirect(ecmaMode());
	instructions().uncheckedAppend(type);
    return value;
}

void BytecodeGenerator::emitPutGetterById(RegisterID* base, const Identifier& property, unsigned attributes, RegisterID* getter)
{
    unsigned propertyIndex = addConstant(property);
    m_staticPropertyAnalyzer.putById(base->index(), propertyIndex);

    emitOpcode(op_put_getter_by_id);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyIndex);
    instructions().uncheckedAppend(attributes);
    instructions().uncheckedAppend(getter->index());
}

void BytecodeGenerator::emitPutSetterById(RegisterID* base, const Identifier& property, unsigned attributes, RegisterID* setter)
{
    unsigned propertyIndex = addConstant(property);
    m_staticPropertyAnalyzer.putById(base->index(), propertyIndex);

    emitOpcode(op_put_setter_by_id);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyIndex);
    instructions().uncheckedAppend(attributes);
    instructions().uncheckedAppend(setter->index());
}

void BytecodeGenerator::emitPutGetterSetter(RegisterID* base, const Identifier& property, unsigned attributes, RegisterID* getter, RegisterID* setter)
{
    unsigned propertyIndex = addConstant(property);

    m_staticPropertyAnalyzer.putById(base->index(), propertyIndex);

    emitOpcode(op_put_getter_setter);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyIndex);
    instructions().uncheckedAppend(attributes);
    instructions().uncheckedAppend(getter->index());
    instructions().uncheckedAppend(setter->index());
}

void BytecodeGenerator::emitPutGetterByVal(RegisterID* base, RegisterID* property, unsigned attributes, RegisterID* getter)
{
    emitOpcode(op_put_getter_by_val);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(attributes);
    instructions().uncheckedAppend(getter->index());
}

void BytecodeGenerator::emitPutSetterByVal(RegisterID* base, RegisterID* property, unsigned attributes, RegisterID* setter)
{
    emitOpcode(op_put_setter_by_val);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(attributes);
    instructions().uncheckedAppend(setter->index());
}

void BytecodeGenerator::emitPutGeneratorFields(RegisterID* nextFunction)
{
    // FIXME: Currently, we just create an object and store generator related fields as its properties for ease.
    // But to make it efficient, we will introduce JSGenerator class, add opcode new_generator and use its C++ fields instead of these private properties.
    // https://bugs.webkit.org/show_bug.cgi?id=151545

    emitDirectPutById(m_generatorRegister, propertyNames().builtinNames().generatorNextPrivateName(), nextFunction);

    // We do not store 'this' in arrow function within constructor,
    // because it might be not initialized, if super is called later.
    if (!(isDerivedConstructorContext() && m_codeBlock->parseMode() == SourceParseMode::AsyncArrowFunctionMode))
        emitDirectPutById(m_generatorRegister, propertyNames().builtinNames().generatorThisPrivateName(), &m_thisRegister);

    emitDirectPutById(m_generatorRegister, propertyNames().generatorStatePrivateName, emitLoad(0, jsNumber(0)));

    emitDirectPutById(m_generatorRegister, propertyNames().generatorFramePrivateName, emitLoad(0, jsNull()));
}

void BytecodeGenerator::emitPutAsyncGeneratorFields(RegisterID* nextFunction)
{
    ASSERT(isAsyncGeneratorWrapperParseMode(parseMode()));

    emitDirectPutById(m_generatorRegister, propertyNames().builtinNames().generatorNextPrivateName(), nextFunction);

    emitDirectPutById(m_generatorRegister, propertyNames().builtinNames().generatorThisPrivateName(), &m_thisRegister);

    emitDirectPutById(m_generatorRegister, propertyNames().generatorStatePrivateName, emitLoad(0, jsNumber(static_cast<int32_t>(JSAsyncGeneratorFunction::AsyncGeneratorState::SuspendedStart))));

    emitDirectPutById(m_generatorRegister, propertyNames().generatorFramePrivateName, emitLoad(0, jsNull()));

    emitDirectPutById(m_generatorRegister, propertyNames().builtinNames().asyncGeneratorSuspendReasonPrivateName(), emitLoad(0, jsNumber(static_cast<int32_t>(JSAsyncGeneratorFunction::AsyncGeneratorSuspendReason::None))));

    emitDirectPutById(m_generatorRegister, propertyNames().builtinNames().asyncGeneratorQueueFirstPrivateName(), emitLoad(0, jsNull()));
    emitDirectPutById(m_generatorRegister, propertyNames().builtinNames().asyncGeneratorQueueLastPrivateName(), emitLoad(0, jsNull()));
}

RegisterID* BytecodeGenerator::emitDeleteById(RegisterID* dst, RegisterID* base, const Identifier& property)
{
    emitOpcode(op_del_by_id);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(addConstant(property));
    instructions().uncheckedAppend(ecmaMode().value());
    return dst;
}

RegisterID* BytecodeGenerator::emitGetByVal(RegisterID* dst, RegisterID* base, RegisterID* property)
{
    for (size_t i = m_forInContextStack.size(); i--; ) {
        ForInContext* context = m_forInContextStack[i].get();
        if (context->local() != property)
            continue;

        unsigned instIndex = instructions().size();

        if (context->isIndexedForInContext()) {
            auto& indexedContext = context->asIndexedForInContext();
            indexedContext.addGetInst(instIndex, property->index());
            property = indexedContext.index();
            break;
        }

        StructureForInContext& structureContext = context->asStructureForInContext();
        ArrayProfile* arrayProfile = newArrayProfile();
        UnlinkedValueProfile profile = emitProfiledOpcode(op_get_direct_pname);
        instructions().uncheckedAppend(kill(dst));
        instructions().uncheckedAppend(base->index());
        instructions().uncheckedAppend(property->index());
        instructions().uncheckedAppend(structureContext.index()->index());
        instructions().uncheckedAppend(structureContext.enumerator()->index());
        instructions().uncheckedAppend(profile);

        structureContext.addGetInst(instIndex, property->index(), arrayProfile, profile);
		return dst;
    }

    ArrayProfile* arrayProfile = newArrayProfile();
    UnlinkedValueProfile profile = emitProfiledOpcode(op_get_by_val);
    instructions().uncheckedAppend(kill(dst));
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(arrayProfile);
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitGetByVal(RegisterID* dst, RegisterID* base, RegisterID* thisValue, RegisterID* property)
{
	UnlinkedValueProfile profile = emitProfiledOpcode(op_get_by_val_with_this);
    instructions().uncheckedAppend(kill(dst));
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(thisValue->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitGetPrototypeOf(RegisterID* dst, RegisterID* value)
{
    //OpGetPrototypeOf::emit(this, dst, value);
	UnlinkedValueProfile profile = emitProfiledOpcode(op_get_prototype_of);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitDirectSetPrototypeOf(RegisterID* base, RegisterID* prototype)
{
    RefPtr<RegisterID> setPrototypeDirect = addConstantValue(m_globalObject->setPrototypeDirectFunction());

    CallArguments args(*this, nullptr, 1);
    move(args.thisRegister(), base);
    move(args.argumentRegister(0), prototype);

    JSTextPosition position;
    emitCall(newTemporary(), setPrototypeDirect.get(), NoExpectedFunction, args, position, position, position/*, DebuggableCall::No*/);
    return base;
}

RegisterID* BytecodeGenerator::emitPutByVal(RegisterID* base, RegisterID* property, RegisterID* value)
{
    ArrayProfile* arrayProfile = newArrayProfile();
    emitOpcode(op_put_by_val);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(ecmaMode().value());
    instructions().uncheckedAppend(arrayProfile);
    return value;
}

RegisterID* BytecodeGenerator::emitPutByVal(RegisterID* base, RegisterID* thisValue, RegisterID* property, RegisterID* value)
{
    emitOpcode(op_put_by_val_with_this);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(thisValue->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(ecmaMode().value());

    return value;
}

RegisterID* BytecodeGenerator::emitGetPrivateName(RegisterID* dst, RegisterID* base, RegisterID* property)
{
	UnlinkedValueProfile profile = emitProfiledOpcode(op_get_private_name);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
#if ENABLE(LLINT)
    instructions().uncheckedAppend(0); // property
    instructions().uncheckedAppend(0); // offset
    instructions().uncheckedAppend(0); // structure
#endif
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitHasPrivateName(RegisterID* dst, RegisterID* base, RegisterID* property)
{
    emitOpcode(op_has_private_name);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitDirectPutByVal(RegisterID* base, RegisterID* property, RegisterID* value)
{
    ArrayProfile* arrayProfile = newArrayProfile();
    emitOpcode(op_put_by_val_direct);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(ecmaMode());
    instructions().uncheckedAppend(arrayProfile);
    return value;
}

RegisterID* BytecodeGenerator::emitDefinePrivateField(RegisterID* base, RegisterID* property, RegisterID* value)
{
    emitOpcode(op_put_private_name);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(PrivateFieldPutKind::define());
#if ENABLE(LLINT)
    instructions().uncheckedAppend(0); // oldStructureID
    instructions().uncheckedAppend(0); // property WriteBarrier[JSCell]
    instructions().uncheckedAppend(0); // offset
    instructions().uncheckedAppend(0); // newStructureID
#endif
    return value;
}

void BytecodeGenerator::emitCreatePrivateBrand(const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    RefPtr<RegisterID> createPrivateSymbol = emitMoveLinkTimeConstant(nullptr, LinkTimeConstant::CreatePrivateSymbol);

    CallArguments arguments(*this, nullptr, 0);
    emitLoad(arguments.thisRegister(), jsUndefined());
    RegisterID* newSymbol = emitCall(finalDestination(nullptr, createPrivateSymbol.get()), createPrivateSymbol.get(), NoExpectedFunction, arguments, divot, divotStart, divotEnd/*, DebuggableCall::No*/);

    Variable privateBrandVar = variable(propertyNames().builtinNames().privateBrandPrivateName());

    emitPutToScope(scopeRegister(), privateBrandVar, newSymbol, DoNotThrowIfNotFound, InitializationMode::ConstInitialization);
}

void BytecodeGenerator::emitInstallPrivateBrand(RegisterID* target)
{
    Variable privateBrandVar = variable(propertyNames().builtinNames().privateBrandPrivateName());
    RefPtr<RegisterID> privateBrandVarScope = emitResolveScope(nullptr, privateBrandVar);
    bool isStatic = false;
	RegisterID* privateBrandSymbol = emitGetPrivateBrand(newTemporary(), privateBrandVarScope.get(), isStatic);
	emitOpcode(op_set_private_brand);
    instructions().uncheckedAppend(target->index());
    instructions().uncheckedAppend(privateBrandSymbol->index());
}

void BytecodeGenerator::emitInstallPrivateClassBrand(RegisterID* target)
{
    Variable privateBrandVar = variable(propertyNames().builtinNames().privateClassBrandPrivateName());
    emitPutToScope(scopeRegister(), privateBrandVar, target, DoNotThrowIfNotFound, InitializationMode::ConstInitialization);
}

RegisterID* BytecodeGenerator::emitGetPrivateBrand(RegisterID* dst, RegisterID* scope, bool isStatic)
{
    ASSERT(scope);

    // get_from_scope dst, scope, id, GetPutInfo, Structure, Operand
    UnlinkedValueProfile profile = emitProfiledOpcode(op_get_from_scope);
    instructions().uncheckedAppend(kill(dst));
    instructions().uncheckedAppend(scope->index());
    instructions().uncheckedAppend(addConstant(isStatic ? propertyNames().builtinNames().privateClassBrandPrivateName() : propertyNames().builtinNames().privateBrandPrivateName()));
    instructions().uncheckedAppend(GetPutInfo(ThrowIfNotFound, ResolvedClosureVar, InitializationMode::NotInitialization, ecmaMode()).operand());
    instructions().uncheckedAppend(0); // or structure
    instructions().uncheckedAppend(Instruction(isStatic ? PrivateNameEntry::privateClassBrandOffset : PrivateNameEntry::privateBrandOffset)); // or uintptr_t
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitPrivateFieldPut(RegisterID* base, RegisterID* property, RegisterID* value)
{
    emitOpcode(op_put_private_name);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(PrivateFieldPutKind::set());
#if ENABLE(LLINT)
    instructions().uncheckedAppend(0); // oldStructureID
    instructions().uncheckedAppend(0); // property WriteBarrier[JSCell]
    instructions().uncheckedAppend(0); // offset
    instructions().uncheckedAppend(0); // newStructureID
#endif
    return value;
}

RegisterID* BytecodeGenerator::emitHasPrivateBrand(RegisterID* dst, RegisterID* base, RegisterID* brand, bool isStatic)
{
    if (isStatic) {
        Ref<Label> isObjectLabel = newLabel();
        emitJumpIfTrue(emitIsObject(newTemporary(), base), isObjectLabel.ptr());
        emitThrowTypeError(ASCIILiteral("Cannot access static private method or accessor of a non-Object"));
        emitLabel(isObjectLabel.get());
        emitEqualityOp(op_stricteq, dst, base, brand);
    } else {
        emitOpcode(op_has_private_brand);
        instructions().uncheckedAppend(dst->index());
        instructions().uncheckedAppend(base->index());
        instructions().uncheckedAppend(brand->index());
    }
    return dst;
}

void BytecodeGenerator::emitCheckPrivateBrand(RegisterID* base, RegisterID* brand, bool isStatic)
{
    if (isStatic) {
        Ref<Label> brandCheckOkLabel = newLabel();
        emitTDZCheck(brand);
        emitJumpIfTrue(emitEqualityOp(op_stricteq, newTemporary(), base, brand), brandCheckOkLabel.ptr());
        emitThrowTypeError(ASCIILiteral("Cannot access static private method or accessor"));
        emitLabel(brandCheckOkLabel.get());
        return;

    }
    emitOpcode(op_check_private_brand);
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(brand->index());
}

RegisterID* BytecodeGenerator::emitDeleteByVal(RegisterID* dst, RegisterID* base, RegisterID* property)
{
    emitOpcode(op_del_by_val);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(property->index());
    instructions().uncheckedAppend(ecmaMode().value());
    return dst;
}

RegisterID* BytecodeGenerator::emitIdWithProfile(RegisterID* src, SpeculatedType profile)
{
    emitOpcode(op_identity_with_profile);
    instructions().uncheckedAppend(src->index());
    instructions().uncheckedAppend(static_cast<uint32_t>(profile >> 32));
    instructions().uncheckedAppend(static_cast<uint32_t>(profile));
    return src;
}

void BytecodeGenerator::emitUnreachable()
{
    emitOpcode(op_unreachable);
}

RegisterID* BytecodeGenerator::emitGetArgument(RegisterID* dst, int32_t index)
{
    UnlinkedValueProfile profile = emitProfiledOpcode(op_get_argument);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(index + 1); // Including |this|.
    instructions().uncheckedAppend(profile);
    return dst;
}

RegisterID* BytecodeGenerator::emitCreateThis(RegisterID* dst)
{
    size_t begin = instructions().size();
    m_staticPropertyAnalyzer.createThis(dst->index(), begin + 3);

    emitOpcode(op_create_this);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(0); // inlineCapacity
    instructions().uncheckedAppend(0); // cachedFunction
    return dst;
}

RegisterID* BytecodeGenerator::emitInstanceFieldInitializationIfNeeded(RegisterID* dst, RegisterID* constructor, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    if (!(isConstructor() || isDerivedConstructorContext()) || needsClassFieldInitializer() == NeedsClassFieldInitializer::No)
        return dst;

    RefPtr<RegisterID> initializer = emitDirectGetById(newTemporary(), constructor, propertyNames().builtinNames().instanceFieldInitializerPrivateName());
    CallArguments args(*this, nullptr);
    emitMove(args.thisRegister(), dst);
    emitCall(newTemporary(), initializer.get(), NoExpectedFunction, args, divot, divotStart, divotEnd/*, DebuggableCall::No*/);

    return dst;
}

void BytecodeGenerator::emitTDZCheck(RegisterID* target)
{
    emitOpcode(op_check_tdz);
    instructions().uncheckedAppend(target->index());
}

bool BytecodeGenerator::needsTDZCheck(const Variable& variable)
{
    for (unsigned i = m_TDZStack.size(); i--;) {
        auto iter = m_TDZStack[i].first.find(variable.ident().impl());
        if (iter == m_TDZStack[i].first.end())
            continue;
        return iter->second != TDZNecessityLevel::NotNeeded;
    }

    {
        TDZEnvironmentLink* environment = m_cachedParentTDZ.get();
        while (environment) {
            if (environment->contains(variable.ident().impl()))
                return true;
            environment = environment->parent();
        }
    }

    return false;
}

void BytecodeGenerator::emitTDZCheckIfNecessary(const Variable& variable, RegisterID* target, RegisterID* scope)
{
    if (needsTDZCheck(variable)) {
        if (target)
            emitTDZCheck(target);
        else {
            RELEASE_ASSERT(!variable.isLocal() && scope);
            RefPtr<RegisterID> result = emitGetFromScope(newTemporary(), scope, variable, DoNotThrowIfNotFound);
            emitTDZCheck(result.get());
        }
    }
}

void BytecodeGenerator::liftTDZCheckIfPossible(const Variable& variable)
{
    RefPtr<UniquedStringImpl> identifier(variable.ident().impl());
    for (unsigned i = m_TDZStack.size(); i--;) {
        auto iter = m_TDZStack[i].first.find(identifier);
        if (iter != m_TDZStack[i].first.end()) {
            if (iter->second == TDZNecessityLevel::Optimize)
                iter->second = TDZNecessityLevel::NotNeeded;
            break;
        }
    }
}

// This should be called only with PrivateNames available.
PrivateNameEntry BytecodeGenerator::getPrivateTraits(const Identifier& ident)
{
    for (unsigned i = m_privateNamesStack.size(); i--; ) {
        auto& map = m_privateNamesStack[i];
        auto it = map.find(ident.impl());
        if (it != map.end())
            return it->second;
    }

    RELEASE_ASSERT_NOT_REACHED();
    return PrivateNameEntry();
}

void BytecodeGenerator::pushPrivateAccessNames(const PrivateNameEnvironment* environment)
{
    if (!environment || !environment->size())
        return;

    m_privateNamesStack.append(*environment);
}

void BytecodeGenerator::popPrivateAccessNames()
{
    ASSERT(m_privateNamesStack.size());
    m_privateNamesStack.removeLast();
}

void BytecodeGenerator::pushTDZVariables(const VariableEnvironment& environment, TDZCheckOptimization optimization, TDZRequirement requirement)
{
    if (!environment.totalSize())
        return;

    TDZNecessityLevel level;
    if (requirement == TDZRequirement::UnderTDZ) {
        if (optimization == TDZCheckOptimization::Optimize)
            level = TDZNecessityLevel::Optimize;
        else
            level = TDZNecessityLevel::DoNotOptimize;
    } else
        level = TDZNecessityLevel::NotNeeded;

    TDZMap map;
    for (auto& entry : environment)
        map.add(entry.first, entry.second.isFunction() ? TDZNecessityLevel::NotNeeded : level);

    m_TDZStack.append(TDZStackEntry { WTFMove(map), nullptr });
}

PassOwnPtr<Vector<Identifier>> BytecodeGenerator::getParameterNames() const
{
    RELEASE_ASSERT(m_scopeNode->isFunctionNode());
    FunctionParameters& parameters = *static_cast<FunctionNode*>(m_scopeNode)->parameters();
    Vector<Identifier, 16> parameterNames;
    for (unsigned i = 0; i < parameters.size(); i++)
        parameters.at(i).first->collectBoundIdentifiers(parameterNames);
    return adoptPtr(new Vector<Identifier>(parameterNames));
}

Optional<PrivateNameEnvironment> BytecodeGenerator::getAvailablePrivateAccessNames()
{
    PrivateNameEnvironment result;
    HashSet<UniquedStringImpl*> excludedNames;
    for (unsigned i = m_privateNamesStack.size(); i--; ) {
        auto& map = m_privateNamesStack[i];
        for (auto& entry : map)  {
            auto addResult = excludedNames.add(entry.first.get());
            if (addResult.isNewEntry)
                result.add(entry.first, entry.second);
        }
    }

    if (!result.size())
        return WTF::Nullopt;
    return result;
}

RefPtr<TDZEnvironmentLink> BytecodeGenerator::getVariablesUnderTDZ()
{
    RefPtr<TDZEnvironmentLink> parent = m_cachedParentTDZ;
    if (!m_TDZStack.size())
        return parent;

    auto assertCacheIsCoherent = [&] {
#if ASSERT_ENABLED
        TDZEnvironmentLink* parent = m_cachedParentTDZ.get();
        for (auto& entry : m_TDZStack) {
            ASSERT(entry.second);
            ASSERT(entry.second->parent() == parent);
            parent = entry.second.get();
        }
#endif
    };

    if (m_TDZStack.last().second) {
        assertCacheIsCoherent();
        return m_TDZStack.last().second;
    }

    for (auto& entry : m_TDZStack) {
        if (!entry.second) {
        	TDZMap& map = entry.first;
            TDZEnvironment environment;
            for (auto& entry : map) {
                if (entry.second != TDZNecessityLevel::NotNeeded)
                    environment.add(entry.first.get());
            }
            entry.second = TDZEnvironmentLink::create(m_vm->m_compactVariableMap->get(environment), parent);
        }
        parent = entry.second;
    }

    assertCacheIsCoherent();

    return parent;
}

void BytecodeGenerator::preserveTDZStack(BytecodeGenerator::PreservedTDZStack& preservedStack)
{
    preservedStack.m_preservedTDZStack = m_TDZStack;
}

void BytecodeGenerator::restoreTDZStack(const BytecodeGenerator::PreservedTDZStack& preservedStack)
{
    m_TDZStack = preservedStack.m_preservedTDZStack;
}

RegisterID* BytecodeGenerator::emitNewObject(RegisterID* dst)
{
    size_t begin = instructions().size();
    m_staticPropertyAnalyzer.newObject(dst->index(), begin + 2);

    emitOpcode(op_new_object);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(0);
    instructions().uncheckedAppend(newObjectAllocationProfile());
    return dst;
}

JSValue BytecodeGenerator::addBigIntConstant(const Identifier& identifier, uint8_t radix, bool sign)
{
    return m_bigIntMap.ensure(BigIntMapEntry(identifier.impl(), radix, sign), [&] {
        auto scope = DECLARE_CATCH_SCOPE(vm());
        auto parseIntSign = sign ? JSBigInt::ParseIntSign::Signed : JSBigInt::ParseIntSign::Unsigned;
        JSValue bigIntInMap = JSBigInt::parseInt(nullptr, vm(), identifier.string(), radix, JSBigInt::ErrorParseMode::ThrowExceptions, parseIntSign);
		scope.assertNoException();
        addConstantValue(bigIntInMap);

        return bigIntInMap;
    }).iterator->second;
}

JSString* BytecodeGenerator::addStringConstant(const Identifier& identifier)
{
    JSString*& stringInMap = m_stringMap.add(identifier.impl(), reinterpret_cast<JSString*>(0)).iterator->second;
    if (!stringInMap) {
        stringInMap = jsString(vm(), identifier.ustring());
        addConstantValue(stringInMap);
    }
    return stringInMap;
}

RegisterID* BytecodeGenerator::addTemplateObjectConstant(RefPtr<TemplateObjectDescriptor> descriptor, int endOffset)
{
    auto result = m_templateObjectDescriptorSet.add(descriptor);
    JSTemplateObjectDescriptor* descriptorValue = m_templateDescriptorMap.ensure(endOffset, [&] {
        return JSTemplateObjectDescriptor::create(vm(), result.iterator->get(), endOffset);
    }).iterator->second;
    int index = addConstantIndex();
    m_codeBlock->addConstant(descriptorValue);
    return &m_constantPoolRegisters[index];
}

RegisterID* BytecodeGenerator::emitNewArrayBuffer(RegisterID* dst, JSFixedArray* array)
{
	emitOpcode(op_new_array_buffer);
	instructions().uncheckedAppend(dst->index());
	instructions().uncheckedAppend(addConstantValue(array)->index());
	instructions().uncheckedAppend(newArrayAllocationProfile());
	return dst;
}

RegisterID* BytecodeGenerator::emitNewArray(RegisterID* dst, ElementNode* elements, unsigned length)
{
    Vector<RefPtr<RegisterID>, 16> argv;
    for (ElementNode* n = elements; n; n = n->next()) {
    	if (!length)
			break;
		length--;
		ASSERT(!n->value()->isSpreadExpression());
        argv.append(newTemporary());
        // op_new_array requires the initial values to be a sequential range of registers
        ASSERT(argv.size() == 1 || argv[argv.size() - 1]->index() == argv[argv.size() - 2]->index() - 1);
        emitNode(argv.last().get(), n->value());
    }
    ASSERT(!length);
    emitOpcode(op_new_array);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(argv.size() ? argv[0]->index() : 0); // argv
    instructions().uncheckedAppend(static_cast<int>(argv.size())); // argc
    instructions().uncheckedAppend(newArrayAllocationProfile());
    return dst;
}

RegisterID* BytecodeGenerator::emitNewArrayWithSpread(RegisterID* dst, ElementNode* elements)
{
    BitVector bitVector;
    Vector<RefPtr<RegisterID>, 16> argv;
    for (ElementNode* node = elements; node; node = node->next()) {
        bitVector.set(argv.size(), node->value()->isSpreadExpression());

        argv.append(newTemporary());
        // op_new_array_with_spread requires the initial values to be a sequential range of registers.
        RELEASE_ASSERT(argv.size() == 1 || argv[argv.size() - 1]->index() == argv[argv.size() - 2]->index() - 1);
    }

    RELEASE_ASSERT(argv.size());

    {
        unsigned i = 0;
        for (ElementNode* node = elements; node; node = node->next()) {
            if (node->value()->isSpreadExpression()) {
                ExpressionNode* expression = static_cast<SpreadExpressionNode*>(node->value())->expression();
                RefPtr<RegisterID> tmp = newTemporary();
                emitNode(tmp.get(), expression);

                emitOpcode(op_spread);
                instructions().uncheckedAppend(argv[i].get()->index());
                instructions().uncheckedAppend(tmp.get()->index());
            } else {
                ExpressionNode* expression = node->value();
                emitNode(argv[i].get(), expression);
            }
            i++;
        }
    }

    unsigned bitVectorIndex = m_codeBlock->addBitVector(WTFMove(bitVector));
    emitOpcode(op_new_array_with_spread);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(argv[0]->index()); // argv
    instructions().uncheckedAppend(static_cast<unsigned>(argv.size())); // argc
    instructions().uncheckedAppend(bitVectorIndex);

    return dst;
}

RegisterID* BytecodeGenerator::emitNewArrayWithSize(RegisterID* dst, RegisterID* length)
{
    emitOpcode(op_new_array_with_size);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(length->index());
    instructions().uncheckedAppend(newArrayAllocationProfile());

    return dst;
}

RegisterID* BytecodeGenerator::emitNewRegExp(RegisterID* dst, RegExp* regExp)
{
    emitOpcode(op_new_regexp);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(addConstantValue(regExp)->index());
    return dst;
}

void BytecodeGenerator::emitNewFunctionExpressionCommon(RegisterID* dst, FunctionMetadataNode* function)
{
	if(UNLIKELY(lowMemory()))
		return;

    unsigned index = m_codeBlock->addFunctionExpr(makeFunction(function));

    OpcodeID opcodeID = op_new_func_exp;
    switch (function->parseMode()) {
    case SourceParseMode::GeneratorWrapperFunctionMode:
    case SourceParseMode::GeneratorWrapperMethodMode:
        opcodeID = op_new_generator_func_exp;
        break;
    case SourceParseMode::AsyncFunctionMode:
    case SourceParseMode::AsyncMethodMode:
    case SourceParseMode::AsyncArrowFunctionMode:
        opcodeID = op_new_async_func_exp;
        break;
    case SourceParseMode::AsyncGeneratorWrapperFunctionMode:
    case SourceParseMode::AsyncGeneratorWrapperMethodMode:
        ASSERT(Options::useAsyncIterator());
        opcodeID = op_new_async_generator_func_exp;
        break;
    default:
        break;
    }

    emitOpcode(opcodeID);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(scopeRegister()->index());
    instructions().uncheckedAppend(index);
}

RegisterID* BytecodeGenerator::emitNewFunctionExpression(RegisterID* dst, FuncExprNode* func)
{
    emitNewFunctionExpressionCommon(dst, func->metadata());
    return dst;
}

RegisterID* BytecodeGenerator::emitNewArrowFunctionExpression(RegisterID* dst, ArrowFuncExprNode* func)
{
	ASSERT(SourceParseModeSet(SourceParseMode::ArrowFunctionMode, SourceParseMode::AsyncArrowFunctionMode).contains(func->metadata()->parseMode()));
	emitNewFunctionExpressionCommon(dst, func->metadata());
    return dst;
}

RegisterID* BytecodeGenerator::emitNewMethodDefinition(RegisterID* dst, MethodDefinitionNode* func)
{
    ASSERT(isMethodParseMode(func->metadata()->parseMode()));
    emitNewFunctionExpressionCommon(dst, func->metadata());
    return dst;
}

RegisterID* BytecodeGenerator::emitNewDefaultConstructor(RegisterID* dst, ConstructorKind constructorKind, const Identifier& name,
	const Identifier& ecmaName, const SourceCode& classSource, NeedsClassFieldInitializer needsClassFieldInitializer, PrivateBrandRequirement privateBrandRequirement)
{
    FunctionExecutable* executable = m_globalObject->builtinExecutables()->createDefaultConstructor(constructorKind, name, needsClassFieldInitializer, privateBrandRequirement);
    executable->setEcmaName(ecmaName);
    executable->setClassSource(classSource);

    unsigned index = m_codeBlock->addFunctionExpr(executable);

    emitOpcode(op_new_func_exp);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(scopeRegister()->index());
    instructions().uncheckedAppend(index);
    return dst;
}

RegisterID* BytecodeGenerator::emitNewClassFieldInitializerFunction(RegisterID* dst, PassOwnPtr<Vector<JSTextPosition>> classFieldLocations, bool isDerived)
{
    DerivedContextType newDerivedContextType;
    SuperBinding superBinding;
    if (!isDerived) {
        newDerivedContextType = DerivedContextType::None;
        superBinding = SuperBinding::NotNeeded;
    } else {
        newDerivedContextType = DerivedContextType::DerivedMethodContext;
        superBinding = SuperBinding::Needed;
    }

    auto variablesUnderTDZ = getVariablesUnderTDZ();
    Optional<PrivateNameEnvironment> parentPrivateNameEnvironment = getAvailablePrivateAccessNames();
	SourceParseMode parseMode = SourceParseMode::ClassFieldInitializerMode;
    ConstructAbility constructAbility = ConstructAbility::CannotConstruct;

    const bool alwaysStrictInClass = true;
    FunctionMetadataNode metadata(parserArena(), JSTokenLocation(), JSTokenLocation(), 0, 0, 0, 0, 0, alwaysStrictInClass, ConstructorKind::None, superBinding, 0, parseMode, false);
    metadata.finishParsing(m_scopeNode->source(), Identifier(), FunctionMode::MethodDefinition);
    ScriptExecutable* topLevelExecutable = m_codeBlock->ownerExecutable()->topLevelExecutable();
    FunctionExecutable* initializer = FunctionExecutable::create(*m_vm, topLevelExecutable, &metadata, metadata.startColumn(), isBuiltinFunction(), constructAbility, scriptMode(), WTFMove(variablesUnderTDZ), WTFMove(parentPrivateNameEnvironment), nullptr, newDerivedContextType, NeedsClassFieldInitializer::No, PrivateBrandRequirement::None, NoIntrinsic);
    initializer->setClassFieldLocations(classFieldLocations);

    unsigned index = m_codeBlock->addFunctionExpr(initializer);
    //OpNewFuncExp::emit(this, dst, scopeRegister(), index);
    emitOpcode(op_new_func_exp);
	instructions().uncheckedAppend(dst->index());
	instructions().uncheckedAppend(scopeRegister()->index());
	instructions().uncheckedAppend(index);
    return dst;
}

RegisterID* BytecodeGenerator::emitNewFunction(RegisterID* dst, FunctionMetadataNode* function)
{
	if(UNLIKELY(lowMemory()))
		return dst;

    unsigned index = m_codeBlock->addFunctionDecl(makeFunction(function));
    if (isGeneratorWrapperParseMode(function->parseMode()))
        emitOpcode(op_new_generator_func);
    else if (function->parseMode() == SourceParseMode::AsyncFunctionMode)
        emitOpcode(op_new_async_func);
    else if (isAsyncGeneratorWrapperParseMode(function->parseMode())) {
        ASSERT(Options::useAsyncIterator());
        emitOpcode(op_new_async_generator_func);
    } else
        emitOpcode(op_new_func);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(scopeRegister()->index());
    instructions().uncheckedAppend(index);
    return dst;
}

bool BytecodeGenerator::shouldSetFunctionName(ExpressionNode* node)
{
    if (node->isBaseFuncExprNode()) {
        FunctionMetadataNode* metadata = static_cast<BaseFuncExprNode*>(node)->metadata();
        if (!metadata->ecmaName().isNull())
            return false;
    } else if (node->isClassExprNode()) {
        ClassExprNode* classExprNode = static_cast<ClassExprNode*>(node);
        if (!classExprNode->ecmaName().isNull())
            return false;
        if (classExprNode->hasStaticProperty(m_vm->propertyNames->name))
            return false;
    } else
        return false;

    return true;
}

void BytecodeGenerator::emitSetFunctionName(RegisterID* value, const Identifier& ident)
{
    RefPtr<RegisterID> name = emitLoad(newTemporary(), ident);

    // FIXME: We should use an op_call to an internal function here instead.
    // https://bugs.webkit.org/show_bug.cgi?id=155547
    emitSetFunctionName(value, name.get());
}

void BytecodeGenerator::emitSetFunctionName(RegisterID* value, RegisterID* name)
{
    // FIXME: We should use an op_call to an internal function here instead.
    // https://bugs.webkit.org/show_bug.cgi?id=155547
    emitOpcode(op_set_function_name);
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(name->index());
}

RegisterID* BytecodeGenerator::emitCall(RegisterID* dst, RegisterID* func, ExpectedFunction expectedFunction, CallArguments& callArguments, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    return emitCall(op_call, dst, func, expectedFunction, callArguments, divot, divotStart, divotEnd);
}

RegisterID* BytecodeGenerator::emitCallInTailPosition(RegisterID* dst, RegisterID* func, ExpectedFunction expectedFunction, CallArguments& callArguments, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    // FIXME: We should be emitting a new op_tail_call here when
    // m_inTailPosition is false
    // https://bugs.webkit.org/show_bug.cgi?id=148661
    return emitCall(op_call, dst, func, expectedFunction, callArguments, divot, divotStart, divotEnd);
}

RegisterID* BytecodeGenerator::emitCallEval(RegisterID* dst, RegisterID* func, CallArguments& callArguments, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    return emitCall(op_call_eval, dst, func, NoExpectedFunction, callArguments, divot, divotStart, divotEnd);
}

ExpectedFunction BytecodeGenerator::expectedFunctionForIdentifier(const Identifier& identifier)
{
    if (identifier == m_vm->propertyNames->Object || identifier == m_vm->propertyNames->builtinNames().ObjectPrivateName())
        return ExpectObjectConstructor;
    if (identifier == m_vm->propertyNames->Array || identifier == m_vm->propertyNames->builtinNames().ArrayPrivateName())
        return ExpectArrayConstructor;
    return NoExpectedFunction;
}

ExpectedFunction BytecodeGenerator::emitExpectedFunctionSnippet(RegisterID* dst, RegisterID* func, ExpectedFunction expectedFunction, CallArguments& callArguments, Label* done)
{
    RefPtr<Label> realCall = newLabel();
    switch (expectedFunction) {
    case ExpectObjectConstructor: {
        // If the number of arguments is non-zero, then we can't do anything interesting.
        if (callArguments.argumentCountIncludingThis() >= 2)
            return NoExpectedFunction;

        size_t begin = instructions().size();
        emitOpcode(op_jneq_ptr);
        instructions().uncheckedAppend(func->index());
        instructions().uncheckedAppend(Special::ObjectConstructor);
        instructions().uncheckedAppend(realCall->bind(begin, instructions().size()));
        instructions().uncheckedAppend(0);

        if (dst != ignoredResult())
			emitNewObject(dst);
        break;
    }

    case ExpectArrayConstructor: {
        // If you're doing anything other than "new Array()" or "new Array(foo)" then we
        // don't do inline it, for now. The only reason is that call arguments are in
        // the opposite order of what op_new_array expects, so we'd either need to change
        // how op_new_array works or we'd need an op_new_array_reverse. Neither of these
        // things sounds like it's worth it.
        if (callArguments.argumentCountIncludingThis() > 2)
            return NoExpectedFunction;

        size_t begin = instructions().size();
        emitOpcode(op_jneq_ptr);
        instructions().uncheckedAppend(func->index());
        instructions().uncheckedAppend(Special::ArrayConstructor);
        instructions().uncheckedAppend(realCall->bind(begin, instructions().size()));
        instructions().uncheckedAppend(0);

        if (dst != ignoredResult()) {
            if (callArguments.argumentCountIncludingThis() == 2)
                emitNewArrayWithSize(dst, callArguments.argumentRegister(0));
            else {
                ASSERT(callArguments.argumentCountIncludingThis() == 1);
                emitOpcode(op_new_array);
                instructions().uncheckedAppend(dst->index());
                instructions().uncheckedAppend(0);
                instructions().uncheckedAppend(0);
                instructions().uncheckedAppend(newArrayAllocationProfile());
            }
        }
        break;
    }

    default:
        ASSERT(expectedFunction == NoExpectedFunction);
        return NoExpectedFunction;
    }

    size_t begin = instructions().size();
    emitOpcode(op_jmp);
    instructions().uncheckedAppend(done->bind(begin, instructions().size()));
    emitLabel(realCall.get());

    return expectedFunction;
}

RegisterID* BytecodeGenerator::emitCall(OpcodeID opcodeID, RegisterID* dst, RegisterID* func, ExpectedFunction expectedFunction, CallArguments& callArguments, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    ASSERT(opcodeID == op_call || opcodeID == op_call_eval);
    ASSERT(func->refCount());

    if (m_shouldEmitProfileHooks)
        move(callArguments.profileHookRegister(), func);

    // Generate code for arguments.
    unsigned argument = 0;
    if (callArguments.argumentsNode()) {
        ArgumentListNode* n = callArguments.argumentsNode()->m_listNode;
        if (n && n->m_expr->isSpreadExpression()) {
            RELEASE_ASSERT(!n->m_next);
            ExpressionNode* expression = static_cast<SpreadExpressionNode*>(n->m_expr)->expression();
            if (expression->isArrayLiteral()) {
                auto* elements = static_cast<ArrayNode*>(expression)->elements();
                if (elements && !elements->next() && elements->value()->isSpreadExpression()) {
                    ExpressionNode* expression = static_cast<SpreadExpressionNode*>(elements->value())->expression();
                    RefPtr<RegisterID> argumentRegister = tempDestination(emitNode(callArguments.argumentRegister(0), expression));
                    emitOpcode(op_spread);
                    instructions().append(argumentRegister.get()->index());
                    instructions().append(argumentRegister.get()->index());

                    RefPtr<RegisterID> thisRegister = emitMove(newTemporary(), callArguments.thisRegister());
                    return emitCallVarargs(dst, func, callArguments.thisRegister(), argumentRegister.get(), newTemporary(), 0, callArguments.profileHookRegister(), divot, divotStart, divotEnd);
                }
            }
            RefPtr<RegisterID> argumentRegister;
			argumentRegister = expression->emitBytecode(*this, callArguments.argumentRegister(0));
			RefPtr<RegisterID> thisRegister = move(newTemporary(), callArguments.thisRegister());
			return emitCallVarargs(dst, func, callArguments.thisRegister(), argumentRegister.get(), newTemporary(), 0, callArguments.profileHookRegister(), divot, divotStart, divotEnd);
		}
        for (; n; n = n->m_next)
    		emitNode(callArguments.argumentRegister(argument++), n);
    }

    // Reserve space for call frame.
    Vector<RefPtr<RegisterID>, JSStack::CallFrameHeaderSize> callFrame;
    for (int i = 0; i < JSStack::CallFrameHeaderSize; ++i)
        callFrame.append(newTemporary());

    if (m_shouldEmitProfileHooks) {
        emitOpcode(op_profile_will_call);
        instructions().uncheckedAppend(callArguments.profileHookRegister()->index());
    }

    emitExpressionInfo(divot, divotStart, divotEnd);

    RefPtr<Label> done = newLabel();
    expectedFunction = emitExpectedFunctionSnippet(dst, func, expectedFunction, callArguments, done.get());

    // Emit call.
    ArrayProfile* arrayProfile = newArrayProfile();
    UnlinkedValueProfile profile = emitProfiledOpcode(opcodeID);
    ASSERT(dst);
    ASSERT(dst != ignoredResult());
    instructions().uncheckedAppend(dst->index()); // result
    instructions().uncheckedAppend(func->index()); // func
    instructions().uncheckedAppend(callArguments.argumentCountIncludingThis()); // argCount
    instructions().uncheckedAppend(callArguments.registerOffset()); // registerOffset
#if ENABLE(LLINT)
    instructions().uncheckedAppend(m_codeBlock->addLLIntCallLinkInfo());
#else
    instructions().uncheckedAppend(0);
#endif
    instructions().uncheckedAppend(ecmaMode().value());
    instructions().uncheckedAppend(arrayProfile);
	instructions().uncheckedAppend(profile);

    if (expectedFunction != NoExpectedFunction)
        emitLabel(done.get());

    if (m_shouldEmitProfileHooks) {
        emitOpcode(op_profile_did_call);
        instructions().uncheckedAppend(callArguments.profileHookRegister()->index());
    }

    return dst;
}

RegisterID* BytecodeGenerator::emitCallVarargs(RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    return emitCallVarargs(op_call_varargs, dst, func, thisRegister, arguments, firstFreeRegister, firstVarArgOffset, profileHookRegister, divot, divotStart, divotEnd);
}


RegisterID* BytecodeGenerator::emitCallVarargsInTailPosition(RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    // FIXME: We should be emitting a new op_tail_call here when
    // m_inTailPosition is false
    // https://bugs.webkit.org/show_bug.cgi?id=148661
    return emitCallVarargs(op_call_varargs, dst, func, thisRegister, arguments, firstFreeRegister, firstVarArgOffset, profileHookRegister, divot, divotStart, divotEnd);
}

RegisterID* BytecodeGenerator::emitConstructVarargs(RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    return emitCallVarargs(op_construct_varargs, dst, func, thisRegister, arguments, firstFreeRegister, firstVarArgOffset, profileHookRegister, divot, divotStart, divotEnd);
}

RegisterID* BytecodeGenerator::emitCallVarargs(OpcodeID opcode, RegisterID* dst, RegisterID* func, RegisterID* thisRegister, RegisterID* arguments, RegisterID* firstFreeRegister, int32_t firstVarArgOffset, RegisterID* profileHookRegister, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    if (m_shouldEmitProfileHooks) {
        move(profileHookRegister, func);
        emitOpcode(op_profile_will_call);
        instructions().uncheckedAppend(profileHookRegister->index());
    }
    
    emitExpressionInfo(divot, divotStart, divotEnd);

    // Emit call.
    ArrayProfile* arrayProfile = newArrayProfile();
    UnlinkedValueProfile profile = emitProfiledOpcode(opcode);
    ASSERT(dst != ignoredResult());
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(func->index());
    instructions().uncheckedAppend(thisRegister ? thisRegister->index() : 0);
    instructions().uncheckedAppend(arguments->index());
    instructions().uncheckedAppend(firstFreeRegister->index());
    instructions().uncheckedAppend(firstVarArgOffset);
    instructions().uncheckedAppend(arrayProfile);
	instructions().uncheckedAppend(profile);
    if (m_shouldEmitProfileHooks) {
        emitOpcode(op_profile_did_call);
        instructions().uncheckedAppend(profileHookRegister->index());
    }
    return dst;
}

void BytecodeGenerator::emitCallDefineProperty(RegisterID* newObj, RegisterID* propertyNameRegister,
    RegisterID* valueRegister, RegisterID* getterRegister, RegisterID* setterRegister, unsigned options, const JSTextPosition& position)
{
    DefinePropertyAttributes attributes;
    if (options & PropertyConfigurable)
        attributes.setConfigurable(true);

    if (options & PropertyWritable)
        attributes.setWritable(true);
    else if (valueRegister)
        attributes.setWritable(false);

    if (options & PropertyEnumerable)
        attributes.setEnumerable(true);

    if (valueRegister)
        attributes.setValue();
    if (getterRegister)
        attributes.setGet();
    if (setterRegister)
        attributes.setSet();

    ASSERT(!valueRegister || (!getterRegister && !setterRegister));

    emitExpressionInfo(position, position, position);

    if (attributes.hasGet() || attributes.hasSet()) {
        RefPtr<RegisterID> throwTypeErrorFunction;
        if (!attributes.hasGet() || !attributes.hasSet())
            throwTypeErrorFunction = emitMoveLinkTimeConstant(nullptr, LinkTimeConstant::ThrowTypeErrorFunction);

        RefPtr<RegisterID> getter;
        if (attributes.hasGet())
            getter = getterRegister;
        else
            getter = throwTypeErrorFunction;

        RefPtr<RegisterID> setter;
        if (attributes.hasSet())
            setter = setterRegister;
        else
            setter = throwTypeErrorFunction;

        emitOpcode(op_define_accessor_property);
        instructions().uncheckedAppend(newObj->index());
        instructions().uncheckedAppend(propertyNameRegister->index());
        instructions().uncheckedAppend(getter->index());
        instructions().uncheckedAppend(setter->index());
        instructions().uncheckedAppend(emitLoad(0, jsNumber(attributes.rawRepresentation()))->index());
    } else {
        emitOpcode(op_define_data_property);
        instructions().uncheckedAppend(newObj->index());
        instructions().uncheckedAppend(propertyNameRegister->index());
        instructions().uncheckedAppend(valueRegister->index());
        instructions().uncheckedAppend(emitLoad(0, jsNumber(attributes.rawRepresentation()))->index());
    }
}

RegisterID* BytecodeGenerator::emitReturn(RegisterID* src)
{
    if (isConstructor()) {
        bool mightBeDerived = constructorKind() == ConstructorKind::Extends;
        bool srcIsThis = src->index() == m_thisRegister.index();

        if (!srcIsThis) {
			RefPtr<Label> isObjectLabel = newLabel();
			emitJumpIfTrue(emitIsObject(newTemporary(), src), isObjectLabel.get());

			if (mightBeDerived) {
				RefPtr<Label> isUndefinedLabel = newLabel();
				emitJumpIfTrue(emitIsUndefined(newTemporary(), src), isUndefinedLabel.get());
				emitThrowTypeError(ASCIILiteral("Cannot return a non-object type in the constructor of a derived class."));
				emitLabel(isUndefinedLabel.get());
			}

			emitUnaryNoDstOp(op_ret, ensureThis());
			emitLabel(isObjectLabel.get());
        }
    }

    return emitUnaryNoDstOp(op_ret, src);
}

RegisterID* BytecodeGenerator::emitConstruct(RegisterID* dst, RegisterID* func, RegisterID* lazyThis, ExpectedFunction expectedFunction, CallArguments& callArguments, const JSTextPosition& divot, const JSTextPosition& divotStart, const JSTextPosition& divotEnd)
{
    ASSERT(func->refCount());

    if (m_shouldEmitProfileHooks)
        move(callArguments.profileHookRegister(), func);

    // Generate code for arguments.
    unsigned argument = 0;
    if (ArgumentsNode* argumentsNode = callArguments.argumentsNode()) {
        ArgumentListNode* n = callArguments.argumentsNode()->m_listNode;
        if (n && n->m_expr->isSpreadExpression()) {
            RELEASE_ASSERT(!n->m_next);
            auto expression = static_cast<SpreadExpressionNode*>(n->m_expr)->expression();
            if (expression->isArrayLiteral()) {
				auto* elements = static_cast<ArrayNode*>(expression)->elements();
				if (elements && !elements->next() && elements->value()->isSpreadExpression()) {
					ExpressionNode* expression = static_cast<SpreadExpressionNode*>(elements->value())->expression();
					RefPtr<RegisterID> argumentRegister = tempDestination(emitNode(callArguments.argumentRegister(0), expression));
					emitOpcode(op_spread);
					instructions().uncheckedAppend(argumentRegister.get()->index());
					instructions().uncheckedAppend(argumentRegister.get()->index());

					move(callArguments.thisRegister(), lazyThis);
					RefPtr<RegisterID> thisRegister = move(newTemporary(), callArguments.thisRegister());
					return emitConstructVarargs(dst, func, callArguments.thisRegister(), argumentRegister.get(), newTemporary(), 0, callArguments.profileHookRegister(), divot, divotStart, divotEnd);
				}
			}
            RefPtr<RegisterID> argumentRegister;
            argumentRegister = expression->emitBytecode(*this, callArguments.argumentRegister(0));
            move(callArguments.thisRegister(), lazyThis);
            return emitConstructVarargs(dst, func, callArguments.thisRegister(), argumentRegister.get(), newTemporary(), 0, callArguments.profileHookRegister(), divot, divotStart, divotEnd);
        }

        for (ArgumentListNode* n = argumentsNode->m_listNode; n; n = n->m_next)
            emitNode(callArguments.argumentRegister(argument++), n);
    }

    if (m_shouldEmitProfileHooks) {
        emitOpcode(op_profile_will_call);
        instructions().uncheckedAppend(callArguments.profileHookRegister()->index());
    }

    move(callArguments.thisRegister(), lazyThis);

    // Reserve space for call frame.
    Vector<RefPtr<RegisterID>, JSStack::CallFrameHeaderSize> callFrame;
    for (int i = 0; i < JSStack::CallFrameHeaderSize; ++i)
        callFrame.append(newTemporary());

    emitExpressionInfo(divot, divotStart, divotEnd);

    RefPtr<Label> done = newLabel();
    expectedFunction = emitExpectedFunctionSnippet(dst, func, expectedFunction, callArguments, done.get());

    UnlinkedValueProfile profile = emitProfiledOpcode(op_construct);
    ASSERT(dst != ignoredResult());
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(func->index()); // func
    instructions().uncheckedAppend(callArguments.argumentCountIncludingThis()); // argCount
    instructions().uncheckedAppend(callArguments.registerOffset()); // registerOffset
#if ENABLE(LLINT)
    instructions().uncheckedAppend(m_codeBlock->addLLIntCallLinkInfo());
#else
    instructions().uncheckedAppend(0);
#endif
    instructions().uncheckedAppend(0);
    instructions().uncheckedAppend(0);
    instructions().uncheckedAppend(profile);

    if (expectedFunction != NoExpectedFunction)
        emitLabel(done.get());

    if (m_shouldEmitProfileHooks) {
        emitOpcode(op_profile_did_call);
        instructions().uncheckedAppend(callArguments.profileHookRegister()->index());
    }

    return dst;
}

RegisterID* BytecodeGenerator::emitStrcat(RegisterID* dst, RegisterID* src, int count)
{
    emitOpcode(op_strcat);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    instructions().uncheckedAppend(count);

    return dst;
}

void BytecodeGenerator::emitToPrimitive(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_to_primitive);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
}

RegisterID* BytecodeGenerator::emitToPropertyKey(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_to_property_key);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

void BytecodeGenerator::emitGetScope()
{
    emitOpcode(op_get_scope);
    instructions().uncheckedAppend(scopeRegister()->index());
}

RegisterID* BytecodeGenerator::emitPushWithScope(RegisterID* objectScope)
{
	pushLocalControlFlowScope();
    RegisterID* newScope = newBlockScopeVariable();
    newScope->ref();

    emitOpcode(op_push_with_scope);
    instructions().uncheckedAppend(newScope->index());
    instructions().uncheckedAppend(scopeRegister()->index());
    instructions().uncheckedAppend(objectScope->index());

    move(scopeRegister(), newScope);
    m_lexicalScopeStack.append(LexicalScopeStackEntry{ 0, newScope, true, 0 });

    return newScope;
}

RegisterID* BytecodeGenerator::emitGetParentScope(RegisterID* dst, RegisterID* scope)
{
    emitOpcode(op_get_parent_scope);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(scope->index());
    return dst;
}

void BytecodeGenerator::emitPopWithScope()
{
	emitGetParentScope(scopeRegister(), scopeRegister());
    popLocalControlFlowScope();
    LexicalScopeStackEntry stackEntry = m_lexicalScopeStack.takeLast();
    stackEntry.m_scope->deref();
    RELEASE_ASSERT(stackEntry.m_isWithScope);
}

void BytecodeGenerator::emitDebugHook(DebugHookID debugHookID, const JSTextPosition& divot)
{
    if (!m_shouldEmitDebugHooks)
        return;

    emitExpressionInfo(divot, divot, divot);
    emitOpcode(op_debug);
    instructions().uncheckedAppend(debugHookID);
}

void BytecodeGenerator::emitDebugHook(DebugHookID debugHookID, unsigned line, unsigned charOffset, unsigned lineStart)
{
    emitDebugHook(debugHookID, JSTextPosition(line, charOffset, lineStart));
}

void BytecodeGenerator::emitDebugHook(StatementNode* statement)
{
    RELEASE_ASSERT(statement->needsDebugHook());
    emitDebugHook(WillExecuteStatement, statement->position());
}

void BytecodeGenerator::emitDebugHook(ExpressionNode* expr)
{
    RELEASE_ASSERT(expr->needsDebugHook());
    emitDebugHook(WillExecuteStatement, expr->position());
}

void BytecodeGenerator::emitWillLeaveCallFrameDebugHook()
{
    RELEASE_ASSERT(m_scopeNode->isFunctionNode());
    emitDebugHook(WillLeaveCallFrame, m_scopeNode->lastLine(), m_scopeNode->startOffset(), m_scopeNode->lineStartOffset());
}

void BytecodeGenerator::pushFinallyControlFlowScope(FinallyContext& finallyContext)
{
    ControlFlowScope scope(ControlFlowScope::Finally, currentLexicalScopeIndex(), &finallyContext);
    m_controlFlowScopeStack.append(scope);

    m_finallyDepth++;
    m_currentFinallyContext = &finallyContext;
}

void BytecodeGenerator::popFinallyControlFlowScope()
{
    ASSERT(m_controlFlowScopeStack.size());
    ASSERT(m_controlFlowScopeStack.last().isFinallyScope());
    ASSERT(m_finallyDepth > 0);
    ASSERT(m_currentFinallyContext);
    m_currentFinallyContext = m_currentFinallyContext->outerContext();
    m_finallyDepth--;
    m_controlFlowScopeStack.removeLast();
}

LabelScope* BytecodeGenerator::breakTarget(const Identifier& name)
{
    shrinkToFit(m_labelScopes);

    if (!m_labelScopes.size())
        return 0;

    // We special-case the following, which is a syntax error in Firefox:
    // label:
    //     break;
    if (name.isEmpty()) {
        for (int i = m_labelScopes.size() - 1; i >= 0; --i) {
            LabelScope& scope = m_labelScopes[i];
            if (scope.type() != LabelScope::NamedLabel)
                return &scope;
        }
        return 0;
    }

    for (int i = m_labelScopes.size() - 1; i >= 0; --i) {
        LabelScope& scope = m_labelScopes[i];
        if (scope.name() && *scope.name() == name)
            return &scope;
    }
    return 0;
}

LabelScope* BytecodeGenerator::continueTarget(const Identifier& name)
{
    shrinkToFit(m_labelScopes);

    if (!m_labelScopes.size())
        return 0;

    if (name.isEmpty()) {
        for (int i = m_labelScopes.size() - 1; i >= 0; --i) {
            LabelScope& scope = m_labelScopes[i];
            if (scope.type() == LabelScope::Loop) {
                ASSERT(scope.continueTarget());
                return &scope;
            }
        }
        return 0;
    }

    // Continue to the loop nested nearest to the label scope that matches
    // 'name'.
    LabelScope* result = 0;
    for (int i = m_labelScopes.size() - 1; i >= 0; --i) {
        LabelScope& scope = m_labelScopes[i];
        if (scope.type() == LabelScope::Loop) {
            ASSERT(scope.continueTarget());
            result = &scope;
        }
        if (scope.name() && *scope.name() == name)
            return result; // may be null.
    }
    return 0;
}

void BytecodeGenerator::allocateAndEmitScope()
{
    m_scopeRegister = addVar();
    m_scopeRegister->ref();
    m_codeBlock->setScopeRegister(scopeRegister()->virtualRegister());
    emitGetScope();
}

TryData* BytecodeGenerator::pushTry(Label* start, Label* handlerLabel, HandlerType handlerType)
{
    TryData tryData;
    tryData.target = handlerLabel;
    tryData.handlerType = handlerType;
    m_tryData.append(tryData);
    TryData* result = &m_tryData.last();

    TryContext tryContext;
    tryContext.start = start;
    tryContext.tryData = result;

    m_tryContextStack.append(tryContext);

    return result;
}

void BytecodeGenerator::popTry(TryData* tryData, Label* end)
{
    m_usesExceptions = true;

    ASSERT_UNUSED(tryData, m_tryContextStack.last().tryData == tryData);

    TryRange tryRange;
    tryRange.start = m_tryContextStack.last().start;
    tryRange.end = end;
    tryRange.tryData = m_tryContextStack.last().tryData;
    m_tryRanges.append(tryRange);
    m_tryContextStack.removeLast();
}

void BytecodeGenerator::emitOutOfLineCatchHandler(RegisterID* thrownValueRegister, RegisterID* completionTypeRegister, TryData* data)
{
    RegisterID* unused = newTemporary();
    emitOutOfLineExceptionHandler(unused, thrownValueRegister, completionTypeRegister, data);
}

void BytecodeGenerator::emitOutOfLineFinallyHandler(RegisterID* exceptionRegister, RegisterID* completionTypeRegister, TryData* data)
{
    RegisterID* unused = newTemporary();
    ASSERT(completionTypeRegister);
    emitOutOfLineExceptionHandler(exceptionRegister, unused, completionTypeRegister, data);
}

void BytecodeGenerator::emitOutOfLineExceptionHandler(RegisterID* exceptionRegister, RegisterID* thrownValueRegister, RegisterID* completionTypeRegister, TryData* data)
{
    VirtualRegister_ completionTypeVirtualRegister = completionTypeRegister ? completionTypeRegister : VirtualRegister_();
    m_exceptionHandlersToEmit.append(CatchEntry { data, exceptionRegister, thrownValueRegister, completionTypeVirtualRegister });
}

void BytecodeGenerator::restoreScopeRegister(int lexicalScopeIndex)
{
    if (lexicalScopeIndex == CurrentLexicalScopeIndex)
        return; // No change needed.

    if (lexicalScopeIndex != OutermostLexicalScopeIndex) {
        ASSERT(lexicalScopeIndex < static_cast<int>(m_lexicalScopeStack.size()));
        int endIndex = lexicalScopeIndex + 1;
        for (size_t i = endIndex; i--; ) {
            if (m_lexicalScopeStack[i].m_scope) {
                move(scopeRegister(), m_lexicalScopeStack[i].m_scope);
                return;
            }
        }
    }
    // Note that if we don't find a local scope in the current function/program,
    // we must grab the outer-most scope of this bytecode generation.
    emitGetScope();
}

void BytecodeGenerator::restoreScopeRegister()
{
    restoreScopeRegister(currentLexicalScopeIndex());
}

int BytecodeGenerator::labelScopeDepthToLexicalScopeIndex(int targetLabelScopeDepth)
{
    ASSERT(labelScopeDepth() - targetLabelScopeDepth >= 0);
    size_t scopeDelta = labelScopeDepth() - targetLabelScopeDepth;
    ASSERT(scopeDelta <= m_controlFlowScopeStack.size());
    if (!scopeDelta)
        return CurrentLexicalScopeIndex;

    ControlFlowScope& targetScope = m_controlFlowScopeStack[targetLabelScopeDepth];
    return targetScope.lexicalScopeIndex;
}

int BytecodeGenerator::localScopeDepth() const
{
    return m_localScopeDepth;
}

int BytecodeGenerator::labelScopeDepth() const
{
	unsigned depth = localScopeDepth() + m_finallyDepth;
    ASSERT(depth == m_controlFlowScopeStack.size());
    return depth;
}

void BytecodeGenerator::emitThrowStaticError(ErrorType errorType, RegisterID* raw)
{
    RefPtr<RegisterID> message = newTemporary();
    emitToString(message.get(), raw);
    emitOpcode(op_throw_static_error);
    instructions().uncheckedAppend(message->index());
    instructions().uncheckedAppend(static_cast<unsigned>(errorType));
}

void BytecodeGenerator::emitThrowStaticError(ErrorType errorType, const Identifier& message)
{
    emitOpcode(op_throw_static_error);
    instructions().uncheckedAppend(addConstantValue(addStringConstant(message))->index());
    instructions().uncheckedAppend(static_cast<unsigned>(errorType));
}

void BytecodeGenerator::emitThrowReferenceError(const String& message)
{
    emitThrowStaticError(ErrorType::ReferenceError, Identifier::fromString(*m_vm, message));
}

void BytecodeGenerator::emitThrowTypeError(const String& message)
{
    emitThrowStaticError(ErrorType::TypeError, Identifier::fromString(*m_vm, message));
}

void BytecodeGenerator::emitThrowTypeError(const Identifier& message)
{
    emitThrowStaticError(ErrorType::TypeError, message);
}

void BytecodeGenerator::emitThrowRangeError(const Identifier& message)
{
    emitThrowStaticError(ErrorType::RangeError, message);
}

void BytecodeGenerator::emitThrowOutOfMemoryError()
{
    emitThrowStaticError(ErrorType::Error, Identifier::fromString(*m_vm, "Out of memory"));
}

void BytecodeGenerator::emitPushFunctionNameScope(const Identifier& property, RegisterID* callee, bool isCaptured)
{
    // There is some nuance here:
    // If we're in strict mode code, the function name scope variable acts exactly like a "const" variable.
    // If we're not in strict mode code, we want to allow bogus assignments to the name scoped variable.
    // This means any assignment to the variable won't throw, but it won't actually assign a new value to it.
    // To accomplish this, we don't report that this scope is a lexical scope. This will prevent
    // any throws when trying to assign to the variable (while still ensuring it keeps its original
    // value). There is some ugliness and exploitation of a leaky abstraction here, but it's better than
    // having a completely new op code and a class to handle name scopes which are so close in functionality
    // to lexical environments.
    VariableEnvironment nameScopeEnvironment;
    auto addResult = nameScopeEnvironment.add(property);
    if (isCaptured)
    	addResult.iterator->second.setIsCaptured();
    addResult.iterator->second.setIsConst(); // The function name scope name acts like a const variable.
    unsigned numVars = m_codeBlock->m_numVars;
    pushLexicalScopeInternal(nameScopeEnvironment, TDZCheckOptimization::Optimize, NestedScopeType::IsNotNested, 0, TDZRequirement::NotUnderTDZ, ScopeType::FunctionNameScope, ScopeRegisterType::Var);
    ASSERT_UNUSED(numVars, m_codeBlock->m_numVars == static_cast<int>(numVars + 1)); // Should have only created one new "var" for the function name scope.
    bool shouldTreatAsLexicalVariable = ecmaMode().isStrict();
    Variable functionVar = variableForLocalEntry(property, m_lexicalScopeStack.last().m_symbolTable->get(property.impl()), m_lexicalScopeStack.last().m_symbolTableConstantIndex, shouldTreatAsLexicalVariable);
    emitPutToScope(m_lexicalScopeStack.last().m_scope, functionVar, callee, ThrowIfNotFound, InitializationMode::NotInitialization);
}

void BytecodeGenerator::pushLocalControlFlowScope()
{
    ControlFlowScope scope(ControlFlowScope::Label, currentLexicalScopeIndex());
    m_controlFlowScopeStack.append(scope);
    m_localScopeDepth++;
}

void BytecodeGenerator::popLocalControlFlowScope()
{
    ASSERT(m_controlFlowScopeStack.size());
    ASSERT(!m_controlFlowScopeStack.last().isFinallyScope());
    m_controlFlowScopeStack.removeLast();
    m_localScopeDepth--;
}

void BytecodeGenerator::emitPushCatchScope(VariableEnvironment& environment, ScopeType scopeType)
{
    pushLexicalScopeInternal(environment, TDZCheckOptimization::Optimize, NestedScopeType::IsNotNested, nullptr, TDZRequirement::UnderTDZ, scopeType, ScopeRegisterType::Block);
}

void BytecodeGenerator::emitPopCatchScope(VariableEnvironment& environment)
{
    popLexicalScopeInternal(environment);
}

void BytecodeGenerator::beginSwitch(RegisterID* scrutineeRegister, SwitchInfo::SwitchType type)
{
    SwitchInfo info = { static_cast<uint32_t>(instructions().size()), type };
    switch (type) {
        case SwitchInfo::SwitchImmediate:
            emitOpcode(op_switch_imm);
            break;
        case SwitchInfo::SwitchCharacter:
            emitOpcode(op_switch_char);
            break;
        case SwitchInfo::SwitchString:
            emitOpcode(op_switch_string);
            break;
        default:
            ASSERT_NOT_REACHED();
    }

    instructions().append(0); // place holder for table index
    instructions().append(0); // place holder for default target
    instructions().append(scrutineeRegister->index());
    m_switchContextStack.append(info);
}

static int32_t keyForImmediateSwitch(ExpressionNode* node, int32_t min, int32_t max)
{
    UNUSED_PARAM(max);
    ASSERT(node->isNumber());
    double value = static_cast<NumberNode*>(node)->value();
    int32_t key = static_cast<int32_t>(value);
    ASSERT(key == value);
    ASSERT(key >= min);
    ASSERT(key <= max);
    return key - min;
}

static int32_t keyForCharacterSwitch(ExpressionNode* node, int32_t min, int32_t max)
{
    UNUSED_PARAM(max);
    ASSERT(node->isString());
    auto* clause = static_cast<StringNode*>(node)->value().impl();
    ASSERT(clause->length() == 1);
    
    int32_t key = (*clause)[0];
    ASSERT(key >= min);
    ASSERT(key <= max);
    return key - min;
}

static void prepareJumpTableForSwitch(
    SimpleJumpTable& jumpTable, int32_t switchAddress, uint32_t clauseCount,
    RefPtr<Label>* labels, ExpressionNode** nodes, int32_t min, int32_t max,
    int32_t (*keyGetter)(ExpressionNode*, int32_t min, int32_t max))
{
    jumpTable.min = min;
    jumpTable.branchOffsets.resize(max - min + 1);
    jumpTable.branchOffsets.fill(0);
    for (uint32_t i = 0; i < clauseCount; ++i) {
        // We're emitting this after the clause labels should have been fixed, so 
        // the labels should not be "forward" references
        ASSERT(!labels[i]->isForward());
        jumpTable.add(keyGetter(nodes[i], min, max), labels[i]->bind(switchAddress, switchAddress + 3));
    }
}

static void prepareJumpTableForStringSwitch(StringJumpTable& jumpTable, int32_t switchAddress, uint32_t clauseCount, RefPtr<Label>* labels, ExpressionNode** nodes)
{
    for (uint32_t i = 0; i < clauseCount; ++i) {
        // We're emitting this after the clause labels should have been fixed, so 
        // the labels should not be "forward" references
        ASSERT(!labels[i]->isForward());
        
        ASSERT(nodes[i]->isString());
        auto* clause = static_cast<StringNode*>(nodes[i])->value().impl();
        OffsetLocation location;
        location.branchOffset = labels[i]->bind(switchAddress, switchAddress + 3);
        jumpTable.offsetTable.add(clause, location);
    }
}

void BytecodeGenerator::endSwitch(uint32_t clauseCount, RefPtr<Label>* labels, ExpressionNode** nodes, Label* defaultLabel, int32_t min, int32_t max)
{
    SwitchInfo switchInfo = m_switchContextStack.last();
    m_switchContextStack.removeLast();

    switch (switchInfo.switchType) {
    case SwitchInfo::SwitchImmediate:
    case SwitchInfo::SwitchCharacter: {
        instructions()[switchInfo.bytecodeOffset + 1] = static_cast<int>(m_codeBlock->numberOfSwitchJumpTables());
        instructions()[switchInfo.bytecodeOffset + 2] = defaultLabel->bind(switchInfo.bytecodeOffset, switchInfo.bytecodeOffset + 3);

        SimpleJumpTable& jumpTable = m_codeBlock->addSwitchJumpTable();
        prepareJumpTableForSwitch(
            jumpTable, switchInfo.bytecodeOffset, clauseCount, labels, nodes, min, max,
            switchInfo.switchType == SwitchInfo::SwitchImmediate
                ? keyForImmediateSwitch
                : keyForCharacterSwitch);
        break;
    }
        
    case SwitchInfo::SwitchString: {
        ASSERT(switchInfo.switchType == SwitchInfo::SwitchString);
        instructions()[switchInfo.bytecodeOffset + 1] = static_cast<int>(m_codeBlock->numberOfStringSwitchJumpTables());
        instructions()[switchInfo.bytecodeOffset + 2] = defaultLabel->bind(switchInfo.bytecodeOffset, switchInfo.bytecodeOffset + 3);

        StringJumpTable& jumpTable = m_codeBlock->addStringSwitchJumpTable();
        prepareJumpTableForStringSwitch(jumpTable, switchInfo.bytecodeOffset, clauseCount, labels, nodes);
        break;
	}

	default:
		RELEASE_ASSERT_NOT_REACHED();
		break;
	}
}

RegisterID* BytecodeGenerator::emitThrowExpressionTooDeepException()
{
    // It would be nice to do an even better job of identifying exactly where the expression is.
    // And we could make the caller pass the node pointer in, if there was some way of getting
    // that from an arbitrary node. However, calling emitExpressionInfo without any useful data
    // is still good enough to get us an accurate line number.
    m_expressionTooDeep = true;
    return newTemporary();
}

bool BytecodeGenerator::isArgumentNumber(const Identifier& ident, int argumentNumber)
{
    RegisterID* registerID = variable(ident).local();
    if (!registerID)
        return false;
    return registerID->index() == CallFrame::argumentOffset(argumentNumber);
}

bool BytecodeGenerator::emitReadOnlyExceptionIfNeeded(const Variable& variable)
{
	// If we're in strict mode, we always throw.
    // If we're not in strict mode, we throw for "const" variables but not the function callee.
	if (ecmaMode().isStrict() || variable.isConst()) {
    	emitThrowTypeError(Identifier::fromString(*m_vm, ReadonlyPropertyWriteError));
		return true;
	}
	return false;
}

void BytecodeGenerator::emitEnumeration(ThrowableExpressionData* node, ExpressionNode* subjectNode, EnumerationFunctor& callBack, ForOfNode* forLoopNode, RegisterID* forLoopSymbolTable)
{
    bool isForAwait = forLoopNode ? forLoopNode->isForAwait() : false;
    auto shouldEmitAwait = isForAwait ? EmitAwait::Yes : EmitAwait::No;
    ASSERT(!isForAwait || (isForAwait && isAsyncFunctionParseMode(parseMode())));

    RefPtr<RegisterID> subject = newTemporary();
    emitNode(subject.get(), subjectNode);
    RefPtr<RegisterID> iterator = isForAwait ? emitGetAsyncIterator(subject.get(), node) : emitGetIterator(subject.get(), node);
    RefPtr<RegisterID> nextMethod = emitGetById(newTemporary(), iterator.get(), propertyNames().next);

    RefPtr<Label> loopDone = newLabel();
    RefPtr<Label> tryStartLabel = newLabel();
    RefPtr<Label> finallyViaThrowLabel = newLabel();
    RefPtr<Label> finallyLabel = newLabel();
    RefPtr<Label> catchLabel = newLabel();
    RefPtr<Label> endCatchLabel = newLabel();

    // RefPtr<Register> iterator's lifetime must be longer than IteratorCloseContext.
    FinallyContext finallyContext(*this, finallyLabel.get());
	pushFinallyControlFlowScope(finallyContext);

    {
    	RefPtr<LabelScope> scope = newLabelScope(LabelScope::Loop);
		RefPtr<RegisterID> value = newTemporary();
		emitLoad(value.get(), jsUndefined());

		emitJump(scope->continueTarget());

		RefPtr<Label> loopStart = newLabel();
		emitLabel(loopStart.get());
		emitLoopHint();

        emitLabel(tryStartLabel.get());
        TryData* tryData = pushTry(tryStartLabel.get(), finallyViaThrowLabel.get(), HandlerType::SynthesizedFinally);
        callBack(*this, value.get());
        emitJump(scope->continueTarget());

        // IteratorClose sequence for abrupt completions.
        {
            // Finally block for the enumeration.
            emitLabel(finallyViaThrowLabel.get());
            popTry(tryData, finallyViaThrowLabel.get());

            RefPtr<Label> finallyBodyLabel = newLabel();
            RefPtr<RegisterID> finallyExceptionRegister = newTemporary();

            emitOutOfLineFinallyHandler(finallyContext.completionValueRegister(), finallyContext.completionTypeRegister(), tryData);
            move(finallyExceptionRegister.get(), finallyContext.completionValueRegister());
            emitJump(finallyBodyLabel.get());

            emitLabel(finallyLabel.get());
            moveEmptyValue(finallyExceptionRegister.get());

            // Finally fall through case.
            emitLabel(finallyBodyLabel.get());
            restoreScopeRegister();

            RefPtr<Label> returnCallTryStart = newLabel();
            emitLabel(returnCallTryStart.get());
            TryData* returnCallTryData = pushTry(returnCallTryStart.get(), catchLabel.get(), HandlerType::SynthesizedCatch);

            emitIteratorClose(iterator.get(), node, shouldEmitAwait);
            RefPtr<Label> finallyDone = newEmittedLabel();
            emitFinallyCompletion(finallyContext, endCatchLabel.get());

            popTry(returnCallTryData, finallyDone.get());

            // Catch block for exceptions that may be thrown while calling the return
            // handler in the enumeration finally block. The only reason we need this
            // catch block is because if entered the above finally block due to a thrown
            // exception, then we want to re-throw the original exception on exiting
            // the finally block. Otherwise, we'll let any new exception pass through.
            {
                emitLabel(catchLabel.get());

                RefPtr<RegisterID> exceptionRegister = newTemporary();
                emitOutOfLineFinallyHandler(exceptionRegister.get(), finallyContext.completionTypeRegister(), returnCallTryData);
                // Since this is a synthesized catch block and we're guaranteed to never need
                // to resolve any symbols from the scope, we can skip restoring the scope
                // register here.

                RefPtr<Label> throwLabel = newLabel();
                emitJumpIfTrue(emitIsEmpty(newTemporary(), finallyExceptionRegister.get()), throwLabel.get());
                move(exceptionRegister.get(), finallyExceptionRegister.get());

                emitLabel(throwLabel.get());
                emitThrow(exceptionRegister.get());

                emitLabel(endCatchLabel.get());
            }
        }

		emitLabel(scope->continueTarget());
        if (forLoopNode)
            prepareLexicalScopeForNextForLoopIteration(forLoopNode, forLoopSymbolTable);

		{
            emitIteratorNext(value.get(), nextMethod.get(), iterator.get(), node, shouldEmitAwait);

			emitJumpIfTrue(emitGetById(newTemporary(), value.get(), propertyNames().done), loopDone.get());
			emitGetById(value.get(), value.get(), propertyNames().value);
			emitJump(loopStart.get());
		}

        bool breakLabelIsBound = scope->breakTargetMayBeBound();
        if (breakLabelIsBound)
            emitLabel(scope->breakTarget());
        popFinallyControlFlowScope();
        if (breakLabelIsBound) {
            // IteratorClose sequence for break-ed control flow.
            emitIteratorClose(iterator.get(), node, shouldEmitAwait);
        }
    }
    emitLabel(loopDone.get());
}

RegisterID* BytecodeGenerator::emitGetEnumerableLength(RegisterID* dst, RegisterID* base)
{
    emitOpcode(op_get_enumerable_length);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitHasEnumerableIndexedProperty(RegisterID* dst, RegisterID* base, RegisterID* propertyName)
{
    ArrayProfile* arrayProfile = newArrayProfile();
    emitOpcode(op_has_enumerable_indexed_property);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyName->index());
    instructions().uncheckedAppend(arrayProfile);
    return dst;
}

RegisterID* BytecodeGenerator::emitHasEnumerableStructureProperty(RegisterID* dst, RegisterID* base, RegisterID* propertyName, RegisterID* enumerator)
{
    emitOpcode(op_has_enumerable_structure_property);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyName->index());
    instructions().uncheckedAppend(enumerator->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitHasEnumerableProperty(RegisterID* dst, RegisterID* base, RegisterID* propertyName)
{
    emitOpcode(op_has_enumerable_property);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    instructions().uncheckedAppend(propertyName->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitGetPropertyEnumerator(RegisterID* dst, RegisterID* base)
{
    emitOpcode(op_get_property_enumerator);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(base->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitEnumeratorStructurePropertyName(RegisterID* dst, RegisterID* enumerator, RegisterID* index)
{
    emitOpcode(op_enumerator_structure_pname);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(enumerator->index());
    instructions().uncheckedAppend(index->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitEnumeratorGenericPropertyName(RegisterID* dst, RegisterID* enumerator, RegisterID* index)
{
    emitOpcode(op_enumerator_generic_pname);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(enumerator->index());
    instructions().uncheckedAppend(index->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitToIndexString(RegisterID* dst, RegisterID* index)
{
    emitOpcode(op_to_index_string);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(index->index());
    return dst;
}

void BytecodeGenerator::pushIndexedForInScope(RegisterID* localRegister, RegisterID* indexRegister)
{
    if (!localRegister)
        return;
    unsigned bodyBytecodeStartOffset = instructions().size();
    m_forInContextStack.append(adoptRef(new IndexedForInContext(localRegister, indexRegister, bodyBytecodeStartOffset)));
}

void BytecodeGenerator::popIndexedForInScope(RegisterID* localRegister)
{
    if (!localRegister)
        return;
    unsigned bodyBytecodeEndOffset = instructions().size();
    m_forInContextStack.last()->asIndexedForInContext().finalize(*this, m_codeBlock, bodyBytecodeEndOffset);
    m_forInContextStack.removeLast();
}

RegisterID* BytecodeGenerator::emitLoadArrowFunctionLexicalEnvironment(const Identifier& identifier)
{
    ASSERT(m_codeBlock->isArrowFunction() || m_codeBlock->isArrowFunctionContext() || constructorKind() == ConstructorKind::Extends || m_codeType == EvalCode || m_codeBlock->parseMode() == SourceParseMode::ClassFieldInitializerMode);

    return emitResolveScope(0, variable(identifier, ThisResolutionType::Scoped));
}

void BytecodeGenerator::emitLoadThisFromArrowFunctionLexicalEnvironment()
{
    emitGetFromScope(thisRegister(), emitLoadArrowFunctionLexicalEnvironment(propertyNames().builtinNames().thisPrivateName()), variable(propertyNames().builtinNames().thisPrivateName(), ThisResolutionType::Scoped), DoNotThrowIfNotFound);
}

RegisterID* BytecodeGenerator::emitLoadNewTargetFromArrowFunctionLexicalEnvironment()
{
	Variable newTargetVar = variable(propertyNames().builtinNames().newTargetLocalPrivateName());

	return emitGetFromScope(m_newTargetRegister, emitLoadArrowFunctionLexicalEnvironment(propertyNames().builtinNames().newTargetLocalPrivateName()), newTargetVar, ThrowIfNotFound);
}

RegisterID* BytecodeGenerator::emitLoadDerivedConstructorFromArrowFunctionLexicalEnvironment()
{
    Variable protoScopeVar = variable(propertyNames().builtinNames().derivedConstructorPrivateName());
    return emitGetFromScope(newTemporary(), emitLoadArrowFunctionLexicalEnvironment(propertyNames().builtinNames().derivedConstructorPrivateName()), protoScopeVar, ThrowIfNotFound);
}

RegisterID* BytecodeGenerator::emitLoadDerivedConstructor()
{
    ASSERT(constructorKind() == ConstructorKind::Extends || isDerivedConstructorContext());
    if (constructorKind() == ConstructorKind::Extends)
        return &m_calleeRegister;
    return emitLoadDerivedConstructorFromArrowFunctionLexicalEnvironment();
}

RegisterID* BytecodeGenerator::ensureThis()
{
    if (constructorKind() == ConstructorKind::Extends || isDerivedConstructorContext()) {
        if ((needsToUpdateArrowFunctionContext() && isSuperCallUsedInInnerArrowFunction()) || m_codeBlock->parseMode() == SourceParseMode::AsyncArrowFunctionBodyMode)
            emitLoadThisFromArrowFunctionLexicalEnvironment();

        emitTDZCheck(thisRegister());
    }

    return thisRegister();
}

bool BytecodeGenerator::isThisUsedInInnerArrowFunction()
{
    return m_scopeNode->doAnyInnerArrowFunctionsUseThis() || m_scopeNode->doAnyInnerArrowFunctionsUseSuperProperty() || m_scopeNode->doAnyInnerArrowFunctionsUseSuperCall() || m_scopeNode->doAnyInnerArrowFunctionsUseEval() || m_codeBlock->usesEval();
}

bool BytecodeGenerator::isArgumentsUsedInInnerArrowFunction()
{
    return m_scopeNode->doAnyInnerArrowFunctionsUseArguments() || m_scopeNode->doAnyInnerArrowFunctionsUseEval();
}

bool BytecodeGenerator::isNewTargetUsedInInnerArrowFunction()
{
    return m_scopeNode->doAnyInnerArrowFunctionsUseNewTarget() || m_scopeNode->doAnyInnerArrowFunctionsUseSuperCall() || m_scopeNode->doAnyInnerArrowFunctionsUseEval() || m_codeBlock->usesEval();
}

bool BytecodeGenerator::isSuperUsedInInnerArrowFunction()
{
    return m_scopeNode->doAnyInnerArrowFunctionsUseSuperCall() || m_scopeNode->doAnyInnerArrowFunctionsUseSuperProperty() || m_scopeNode->doAnyInnerArrowFunctionsUseEval() || m_codeBlock->usesEval();
}

bool BytecodeGenerator::isSuperCallUsedInInnerArrowFunction()
{
    return m_scopeNode->doAnyInnerArrowFunctionsUseSuperCall() || m_scopeNode->doAnyInnerArrowFunctionsUseEval() || m_codeBlock->usesEval();
}

void BytecodeGenerator::emitPutNewTargetToArrowFunctionContextScope()
{
    if (isNewTargetUsedInInnerArrowFunction()) {
        ASSERT(m_arrowFunctionContextLexicalEnvironmentRegister);

        Variable newTargetVar = variable(propertyNames().builtinNames().newTargetLocalPrivateName());
        emitPutToScope(m_arrowFunctionContextLexicalEnvironmentRegister, newTargetVar, newTarget(), DoNotThrowIfNotFound, InitializationMode::Initialization);
    }
}

void BytecodeGenerator::emitPutDerivedConstructorToArrowFunctionContextScope()
{
    if (needsDerivedConstructorInArrowFunctionLexicalEnvironment()) {
		ASSERT(m_arrowFunctionContextLexicalEnvironmentRegister);

		Variable protoScope = variable(propertyNames().builtinNames().derivedConstructorPrivateName());
		emitPutToScope(m_arrowFunctionContextLexicalEnvironmentRegister, protoScope, &m_calleeRegister, DoNotThrowIfNotFound, InitializationMode::Initialization);
	}
}

void BytecodeGenerator::emitPutThisToArrowFunctionContextScope()
{
	if (isThisUsedInInnerArrowFunction() || (m_scopeNode->usesSuperCall() && m_codeType == EvalCode)) {
		ASSERT(isDerivedConstructorContext() || m_arrowFunctionContextLexicalEnvironmentRegister != 0);

		Variable thisVar = variable(propertyNames().builtinNames().thisPrivateName(), ThisResolutionType::Scoped);
		RegisterID* scope = isDerivedConstructorContext() ? emitLoadArrowFunctionLexicalEnvironment(propertyNames().builtinNames().thisPrivateName()) : m_arrowFunctionContextLexicalEnvironmentRegister;

		emitPutToScope(scope, thisVar, thisRegister(), ThrowIfNotFound, InitializationMode::NotInitialization);
    }
}

void BytecodeGenerator::pushStructureForInScope(RegisterID* localRegister, RegisterID* indexRegister, RegisterID* propertyRegister, RegisterID* enumeratorRegister)
{
    if (!localRegister)
        return;
    unsigned bodyBytecodeStartOffset = instructions().size();
    m_forInContextStack.append(adoptRef(new StructureForInContext(localRegister, indexRegister, propertyRegister, enumeratorRegister, bodyBytecodeStartOffset)));
}

void BytecodeGenerator::popStructureForInScope(RegisterID* localRegister)
{
    if (!localRegister)
        return;
    unsigned bodyBytecodeEndOffset = instructions().size();
    m_forInContextStack.last()->asStructureForInContext().finalize(*this, m_codeBlock, bodyBytecodeEndOffset);
	m_forInContextStack.removeLast();
}

RegisterID* BytecodeGenerator::emitGetTemplateObject(RegisterID* dst, TaggedTemplateNode* taggedTemplate)
{
	TemplateObjectDescriptor::StringVector rawStrings;
	TemplateObjectDescriptor::OptionalStringVector cookedStrings;

    TemplateStringListNode* templateString = taggedTemplate->templateLiteral()->templateStrings();
    for (; templateString; templateString = templateString->next()) {
    	auto* string = templateString->value();
		ASSERT(string->raw());
		rawStrings.append(string->raw()->impl());
		if (!string->cooked())
			cookedStrings.append(Nullopt);
		else
			cookedStrings.append(string->cooked()->impl());
    }

    RefPtr<RegisterID> constant = addTemplateObjectConstant(TemplateObjectDescriptor::create(WTFMove(rawStrings), WTFMove(cookedStrings)), taggedTemplate->endOffset());
    if (!dst)
        return constant.get();
    return move(dst, constant.get());
}

RegisterID* BytecodeGenerator::emitGetGlobalPrivate(RegisterID* dst, const Identifier& property)
{
    dst = tempDestination(dst);
    Variable var = variable(property);
    if (RegisterID* local = var.local())
        return move(dst, local);

    RefPtr<RegisterID> scope = newTemporary();
    move(scope.get(), emitResolveScope(scope.get(), var));
    return emitGetFromScope(dst, scope.get(), var, ThrowIfNotFound);
}

RegisterID* BytecodeGenerator::emitIsCellWithType(RegisterID* dst, RegisterID* src, JSType type)
{
    emitOpcode(op_is_cell_with_type);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    instructions().uncheckedAppend(type);
    return dst;
}

RegisterID* BytecodeGenerator::emitIsObject(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_is_object);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitIsConstructor(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_is_constructor);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitIsCallable(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_is_callable);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitIsNumber(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_is_number);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitIsUndefinedOrNull(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_is_undefined_or_null);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitIsEmpty(RegisterID* dst, RegisterID* src)
{
    emitOpcode(op_is_empty);
    instructions().uncheckedAppend(dst->index());
    instructions().uncheckedAppend(src->index());
    return dst;
}

RegisterID* BytecodeGenerator::emitIteratorNext(RegisterID* dst, RegisterID* nextMethod, RegisterID* iterator, const ThrowableExpressionData* node, EmitAwait doEmitAwait)
{
    {
        CallArguments nextArguments(*this, 0);
        move(nextArguments.thisRegister(), iterator);
        emitCall(dst, nextMethod, NoExpectedFunction, nextArguments, node->divot(), node->divotStart(), node->divotEnd());

        if (doEmitAwait == EmitAwait::Yes)
            emitAwait(dst);
    }
    {
        RefPtr<Label> typeIsObject = newLabel();
        emitJumpIfTrue(emitIsObject(newTemporary(), dst), typeIsObject.get());
        emitThrowTypeError(ASCIILiteral("Iterator result interface is not an object."));
        emitLabel(typeIsObject.get());
    }
    return dst;
}

RegisterID* BytecodeGenerator::emitIteratorNextWithValue(RegisterID* dst, RegisterID* nextMethod, RegisterID* iterator, RegisterID* value, const ThrowableExpressionData* node)
{
    {
        CallArguments nextArguments(*this, 0, 1);
        move(nextArguments.thisRegister(), iterator);
        move(nextArguments.argumentRegister(0), value);
        emitCall(dst, nextMethod, NoExpectedFunction, nextArguments, node->divot(), node->divotStart(), node->divotEnd());
    }
    return dst;
}

void BytecodeGenerator::emitIteratorClose(RegisterID* iterator, const ThrowableExpressionData* node, EmitAwait doEmitAwait)
{
    RefPtr<Label> done = newLabel();
    RefPtr<RegisterID> returnMethod = emitGetById(newTemporary(), iterator, propertyNames().returnKeyword);
    emitJumpIfTrue(emitIsUndefinedOrNull(newTemporary(), returnMethod.get()), done.get());

    RefPtr<RegisterID> value = newTemporary();
    CallArguments returnArguments(*this, 0);
    move(returnArguments.thisRegister(), iterator);
    emitCall(value.get(), returnMethod.get(), NoExpectedFunction, returnArguments, node->divot(), node->divotStart(), node->divotEnd());

    if (doEmitAwait == EmitAwait::Yes)
        emitAwait(value.get());

    emitJumpIfTrue(emitIsObject(newTemporary(), value.get()), done.get());
    emitThrowTypeError(ASCIILiteral("Iterator result interface is not an object."));
    emitLabel(done.get());
}

void ForInContext::finalize(BytecodeGenerator& generator, CodeBlock* codeBlock, unsigned bodyBytecodeEndOffset)
{
    // Lexically invalidating ForInContexts is kind of weak sauce, but it only occurs if
    // either of the following conditions is true:
    //
    // (1) The loop iteration variable is re-assigned within the body of the loop.
    // (2) The loop iteration variable is captured in the lexical scope of the function.
    //
    // These two situations occur sufficiently rarely that it's okay to use this style of
    // "analysis" to make iteration faster. If we didn't want to do this, we would either have
    // to perform some flow-sensitive analysis to see if/when the loop iteration variable was
    // reassigned, or we'd have to resort to runtime checks to see if the variable had been
    // reassigned from its original value.

    for (unsigned offset = bodyBytecodeStartOffset(); isValid() && offset < bodyBytecodeEndOffset;) {
        Instruction* instruction = &generator.instructions()[offset];
        OpcodeID opcodeID = instruction->u.opcode;
        unsigned opcodeLength = opcodeLengths[opcodeID];

        class Functor
        {
        public:
        	Functor(ForInContext& forInContext)
        	: forInContext(forInContext)
			{
			}
        	void operator()(CodeBlock*, Instruction*, OpcodeID, int operand) const {
                if (forInContext.local()->index() == operand)
                	forInContext.invalidate();
            }
        private:
        	ForInContext& forInContext;
        };
        ASSERT(opcodeID != op_enter);
        Functor functor(*this);
        computeDefsForBytecodeOffset(codeBlock, opcodeID, instruction, functor);
        offset += opcodeLength;
    }
}

void StructureForInContext::finalize(BytecodeGenerator& generator, CodeBlock* codeBlock, unsigned bodyBytecodeEndOffset)
{
    Base::finalize(generator, codeBlock, bodyBytecodeEndOffset);
    if (isValid())
        return;

    for(int i = 0; i < m_getInsts.size(); ++i) {
        unsigned instIndex = m_getInsts[i].instIndex;
        int propertyRegIndex = m_getInsts[i].propertyRegIndex;
        ArrayProfile* arrayProfile = m_getInsts[i].arrayProfile;
        UnlinkedValueProfile valueProfile = m_getInsts[i].valueProfile;
        OpcodeID op = generator.instructions()[instIndex].u.opcode;
        RELEASE_ASSERT(op == op_get_direct_pname);
        ASSERT(opcodeLengths[op_get_direct_pname] == 7);
        ASSERT(opcodeLengths[op_get_by_val] == 6);

        // 0. Change the opcode to get_by_val.
        generator.instructions()[instIndex].u.opcode = op_get_by_val;
        // 1. dst stays the same.
        // 2. base stays the same.
        // 3. property gets switched to the original property.
        generator.instructions()[instIndex + 3].u.operand = propertyRegIndex;
        // 4. set the result array profile.
        generator.instructions()[instIndex + 4].u.arrayProfile = arrayProfile;
        // 5. set the result value profile.
        generator.instructions()[instIndex + 5].u.unsignedValue = valueProfile;
        // 6. nop out the last instruction word.
        generator.instructions()[instIndex + 6].u.opcode = op_nop;
    }
}

void IndexedForInContext::finalize(BytecodeGenerator& generator, CodeBlock* codeBlock, unsigned bodyBytecodeEndOffset)
{
    Base::finalize(generator, codeBlock, bodyBytecodeEndOffset);
    if (isValid())
        return;

    for(int i = 0; i < m_getInsts.size(); ++i) {
        unsigned instIndex = m_getInsts[i].first;
        int propertyRegIndex = m_getInsts[i].second;
        OpcodeID op = generator.instructions()[instIndex].u.opcode;
        RELEASE_ASSERT(op == op_get_by_val);
        // We just need to perform the get_by_val with the original property here,
        // not the indexed one.
        generator.instructions()[instIndex + 3].u.operand = propertyRegIndex;
    }
}

RegisterID* BytecodeGenerator::emitRestParameter(RegisterID* result, unsigned numParametersToSkip)
{
    RefPtr<RegisterID> restArrayLength = newTemporary();
    emitOpcode(op_get_rest_length);
    instructions().uncheckedAppend(restArrayLength->index());
    instructions().uncheckedAppend(numParametersToSkip);

    emitNewArrayWithSize(result, restArrayLength.get());

    emitOpcode(op_copy_rest);
    instructions().uncheckedAppend(result->index());
    instructions().uncheckedAppend(restArrayLength->index());
    instructions().uncheckedAppend(numParametersToSkip);

    return result;
}

void BytecodeGenerator::emitRequireObjectCoercible(RegisterID* value, const String& error)
{
    Ref<Label> target = newLabel();
    size_t begin = instructions().size();
    emitOpcode(op_jnundefined_or_null);
    instructions().uncheckedAppend(value->index());
    instructions().uncheckedAppend(target->bind(begin, instructions().size()));
    emitThrowTypeError(error);
    emitLabel(target.get());
}

void BytecodeGenerator::emitYieldPoint(RegisterID* argument, JSAsyncGeneratorFunction::AsyncGeneratorSuspendReason result)
{
    RefPtr<Label> mergePoint = newLabel();
    unsigned yieldPointIndex = m_yieldPoints++;
    emitGeneratorStateChange(yieldPointIndex + 1);

    if (parseMode() == SourceParseMode::AsyncGeneratorBodyMode) {
        int suspendReason = static_cast<int32_t>(result);
        emitPutById(generatorRegister(), propertyNames().builtinNames().asyncGeneratorSuspendReasonPrivateName(), emitLoad(0, jsNumber(suspendReason)));
    }

    // Split the try range here.
    RefPtr<Label> savePoint = newEmittedLabel();
    for (unsigned i = m_tryContextStack.size(); i--;) {
        TryContext& context = m_tryContextStack[i];
        TryRange range;
        range.start = context.start;
        range.end = savePoint;
        range.tryData = context.tryData;
        m_tryRanges.append(range);

        // Try range will be restared at the merge point.
        context.start = mergePoint;
    }
    Vector<TryContext> savedTryContextStack;
    m_tryContextStack.swap(savedTryContextStack);

    emitOpcode(op_yield);
    instructions().uncheckedAppend(generatorFrameRegister()->index());
    instructions().uncheckedAppend(yieldPointIndex);
    instructions().uncheckedAppend(argument->index());

    // Restore the try contexts, which start offset is updated to the merge point.
    m_tryContextStack.swap(savedTryContextStack);
    emitLabel(mergePoint.get());
}

RegisterID* BytecodeGenerator::emitYield(RegisterID* argument)
{
    emitYieldPoint(argument, JSAsyncGeneratorFunction::AsyncGeneratorSuspendReason::Yield);

    RefPtr<Label> normalLabel = newLabel();
    emitJumpIfTrue(emitEqualityOp(op_stricteq, newTemporary(), generatorResumeModeRegister(), emitLoad(nullptr, JSGeneratorFunction::GeneratorResumeMode::NormalMode)), normalLabel.get());

    RefPtr<Label> throwLabel = newLabel();
    emitJumpIfTrue(emitEqualityOp(op_stricteq, newTemporary(), generatorResumeModeRegister(), emitLoad(nullptr, JSGeneratorFunction::GeneratorResumeMode::ThrowMode)), throwLabel.get());
    // Return.
    {
        RefPtr<RegisterID> returnRegister = generatorValueRegister();
        bool hasFinally = emitReturnViaFinallyIfNeeded(returnRegister.get());
        if (!hasFinally)
        	emitReturn(returnRegister.get());
    }

    // Throw.
    emitLabel(throwLabel.get());
    emitThrow(generatorValueRegister());

    // Normal.
    emitLabel(normalLabel.get());
    return generatorValueRegister();
}

RegisterID* BytecodeGenerator::emitAwait(RegisterID* dst, RegisterID* src)
{
    emitYieldPoint(src, JSAsyncGeneratorFunction::AsyncGeneratorSuspendReason::Await);

    Ref<Label> normalLabel = newLabel();
    emitJumpIfTrue(emitEqualityOp(op_stricteq, newTemporary(), generatorResumeModeRegister(), emitLoad(nullptr, JSGeneratorFunction::GeneratorResumeMode::NormalMode)), normalLabel.ptr());

    emitThrow(generatorValueRegister());

    emitLabel(normalLabel.get());
    return move(dst, generatorValueRegister());
}

RegisterID* BytecodeGenerator::emitCallIterator(RegisterID* iterator, RegisterID* argument, ThrowableExpressionData* node)
{
    CallArguments args(*this, 0);
    move(args.thisRegister(), argument);
    emitCall(iterator, iterator, NoExpectedFunction, args, node->divot(), node->divotStart(), node->divotEnd()/*, DebuggableCall::No*/);

    return iterator;
}

RegisterID* BytecodeGenerator::emitGetIterator(RegisterID* argument, ThrowableExpressionData* node)
{
    RefPtr<RegisterID> iterator = emitGetById(newTemporary(), argument, propertyNames().iteratorSymbol);
    emitCallIterator(iterator.get(), argument, node);

    return iterator.get();
}

RegisterID* BytecodeGenerator::emitGetAsyncIterator(RegisterID* argument, ThrowableExpressionData* node)
{
    RefPtr<RegisterID> iterator = emitGetById(newTemporary(), argument, propertyNames().asyncIteratorSymbol);
    RefPtr<Label> asyncIteratorNotFound = newLabel();
    RefPtr<Label> asyncIteratorFound = newLabel();
    RefPtr<Label> iteratorReceived = newLabel();

    emitJumpIfTrue(emitIsUndefinedOrNull(newTemporary(), iterator.get()), asyncIteratorNotFound.get());

    emitJump(asyncIteratorFound.get());
    emitLabel(asyncIteratorNotFound.get());

    RefPtr<RegisterID> commonIterator = emitGetIterator(argument, node);
    move(iterator.get(), commonIterator.get());

    RefPtr<RegisterID> nextMethod = emitGetById(newTemporary(), iterator.get(), propertyNames().next);

    Variable varCreateAsyncFromSyncIterator = variable(propertyNames().builtinNames().createAsyncFromSyncIteratorPrivateName());
    RefPtr<RegisterID> scope = newTemporary();
    move(scope.get(), emitResolveScope(scope.get(), varCreateAsyncFromSyncIterator));
    RefPtr<RegisterID> createAsyncFromSyncIterator = emitGetFromScope(newTemporary(), scope.get(), varCreateAsyncFromSyncIterator, ThrowIfNotFound);

    CallArguments args(*this, 0, 2);
    emitLoad(args.thisRegister(), jsUndefined());

    move(args.argumentRegister(0), iterator.get());
    move(args.argumentRegister(1), nextMethod.get());

    JSTextPosition divot(m_scopeNode->firstLine(), m_scopeNode->startOffset(), m_scopeNode->lineStartOffset());
    emitCall(iterator.get(), createAsyncFromSyncIterator.get(), NoExpectedFunction, args, divot, divot, divot/*, DebuggableCall::No*/);

    emitJump(iteratorReceived.get());

    emitLabel(asyncIteratorFound.get());
    emitCallIterator(iterator.get(), argument, node);
    emitLabel(iteratorReceived.get());

    return iterator.get();
}

RegisterID* BytecodeGenerator::emitDelegateYield(RegisterID* argument, ThrowableExpressionData* node)
{
    RefPtr<RegisterID> value = newTemporary();
    {
        RefPtr<RegisterID> iterator = parseMode() == SourceParseMode::AsyncGeneratorBodyMode ? emitGetAsyncIterator(argument, node) : emitGetIterator(argument, node);
        RefPtr<RegisterID> nextMethod = emitGetById(newTemporary(), iterator.get(), propertyNames().next);

        RefPtr<Label> loopDone = newLabel();
        {
            RefPtr<Label> nextElement = newLabel();
            emitLoad(value.get(), jsUndefined());

            emitJump(nextElement.get());

            RefPtr<Label> loopStart = newLabel();
            emitLabel(loopStart.get());
            emitLoopHint();

            RefPtr<Label> branchOnResult = newLabel();
            {
                emitYieldPoint(value.get(), JSAsyncGeneratorFunction::AsyncGeneratorSuspendReason::Yield);
                move(value.get(), generatorValueRegister());

                RefPtr<Label> normalLabel = newLabel();
                emitJumpIfTrue(emitEqualityOp(op_stricteq, newTemporary(), generatorResumeModeRegister(), emitLoad(nullptr, JSGeneratorFunction::GeneratorResumeMode::NormalMode)), normalLabel.get());

                RefPtr<Label> returnLabel = newLabel();
                emitJumpIfTrue(emitEqualityOp(op_stricteq, newTemporary(), generatorResumeModeRegister(), emitLoad(nullptr, JSGeneratorFunction::GeneratorResumeMode::ReturnMode)), returnLabel.get());

                // Throw.
                {
                    RefPtr<Label> throwMethodFound = newLabel();
                    RefPtr<RegisterID> throwMethod = emitGetById(newTemporary(), iterator.get(), propertyNames().throwKeyword);
                    emitJumpIfFalse(emitIsUndefinedOrNull(newTemporary(), throwMethod.get()), throwMethodFound.get());

                    EmitAwait emitAwaitInIteratorClose = parseMode() == SourceParseMode::AsyncGeneratorBodyMode ? EmitAwait::Yes : EmitAwait::No;
                    emitIteratorClose(iterator.get(), node, emitAwaitInIteratorClose);

                    emitThrowTypeError(ASCIILiteral("Delegated generator does not have a 'throw' method."));

                    emitLabel(throwMethodFound.get());
                    CallArguments throwArguments(*this, 0, 1);
                    move(throwArguments.thisRegister(), iterator.get());
                    move(throwArguments.argumentRegister(0), value.get());
                    emitCall(value.get(), throwMethod.get(), NoExpectedFunction, throwArguments, node->divot(), node->divotStart(), node->divotEnd());

                    emitJump(branchOnResult.get());
				}

                // Return.
                emitLabel(returnLabel.get());
                {
                    RefPtr<Label> returnMethodFound = newLabel();
                    RefPtr<RegisterID> returnMethod = emitGetById(newTemporary(), iterator.get(), propertyNames().returnKeyword);
                    emitJumpIfFalse(emitIsUndefinedOrNull(newTemporary(), returnMethod.get()), returnMethodFound.get());

                    if (parseMode() == SourceParseMode::AsyncGeneratorBodyMode)
                        emitAwait(value.get());

                    RefPtr<Label> returnSequence = newLabel();
                    emitJump(returnSequence.get());

                    emitLabel(returnMethodFound.get());
                    CallArguments returnArguments(*this, 0, 1);
                    move(returnArguments.thisRegister(), iterator.get());
                    move(returnArguments.argumentRegister(0), value.get());
                    emitCall(value.get(), returnMethod.get(), NoExpectedFunction, returnArguments, node->divot(), node->divotStart(), node->divotEnd());

                    if (parseMode() == SourceParseMode::AsyncGeneratorBodyMode)
                        emitAwait(value.get());

                    RefPtr<Label> returnIteratorResultIsObject = newLabel();
                    emitJumpIfTrue(emitIsObject(newTemporary(), value.get()), returnIteratorResultIsObject.get());
                    emitThrowTypeError(ASCIILiteral("Iterator result interface is not an object."));

                    emitLabel(returnIteratorResultIsObject.get());

                    RefPtr<Label> returnFromGenerator = newLabel();
                    emitJumpIfTrue(emitGetById(newTemporary(), value.get(), propertyNames().done), returnFromGenerator.get());

                    emitGetById(value.get(), value.get(), propertyNames().value);
                    emitJump(loopStart.get());

                    emitLabel(returnFromGenerator.get());
                    emitGetById(value.get(), value.get(), propertyNames().value);

                    emitLabel(returnSequence.get());
                    bool hasFinally = emitReturnViaFinallyIfNeeded(value.get());
                    if (!hasFinally)
                    	emitReturn(value.get());
                }

                // Normal.
                emitLabel(normalLabel.get());
            }

            emitLabel(nextElement.get());
			emitIteratorNextWithValue(value.get(), nextMethod.get(), iterator.get(), value.get(), node);

            emitLabel(branchOnResult.get());

            if (parseMode() == SourceParseMode::AsyncGeneratorBodyMode)
                emitAwait(value.get());

            RefPtr<Label> iteratorValueIsObject = newLabel();
            emitJumpIfTrue(emitIsObject(newTemporary(), value.get()), iteratorValueIsObject.get());
            emitThrowTypeError(ASCIILiteral("Iterator result interface is not an object."));
            emitLabel(iteratorValueIsObject.get());

			emitJumpIfTrue(emitGetById(newTemporary(), value.get(), propertyNames().done), loopDone.get());
			emitGetById(value.get(), value.get(), propertyNames().value);

			emitJump(loopStart.get());
        }
        emitLabel(loopDone.get());
    }

    emitGetById(value.get(), value.get(), propertyNames().value);
    return value.get();
}


void BytecodeGenerator::emitGeneratorStateChange(int32_t state)
{
    RegisterID* completedState = emitLoad(0, jsNumber(state));
    emitPutById(generatorRegister(), propertyNames().generatorStatePrivateName, completedState);
}

bool BytecodeGenerator::emitJumpViaFinallyIfNeeded(int targetLabelScopeDepth, Label* jumpTarget)
{
    ASSERT(labelScopeDepth() - targetLabelScopeDepth >= 0);
    size_t numberOfScopesToCheckForFinally = labelScopeDepth() - targetLabelScopeDepth;
    ASSERT(numberOfScopesToCheckForFinally <= m_controlFlowScopeStack.size());
    if (!numberOfScopesToCheckForFinally)
        return false;

    FinallyContext* innermostFinallyContext = 0;
    FinallyContext* outermostFinallyContext = 0;
    size_t scopeIndex = m_controlFlowScopeStack.size() - 1;
    while (numberOfScopesToCheckForFinally--) {
        ControlFlowScope* scope = &m_controlFlowScopeStack[scopeIndex--];
        if (scope->isFinallyScope()) {
            FinallyContext* finallyContext = scope->finallyContext;
            if (!innermostFinallyContext)
                innermostFinallyContext = finallyContext;
            outermostFinallyContext = finallyContext;
            finallyContext->incNumberOfBreaksOrContinues();
        }
    }
    if (!outermostFinallyContext)
        return false; // No finallys to thread through.

    CompletionType jumpID = bytecodeOffsetToJumpID(instructions().size());
    int lexicalScopeIndex = labelScopeDepthToLexicalScopeIndex(targetLabelScopeDepth);
    outermostFinallyContext->registerJump(jumpID, lexicalScopeIndex, jumpTarget);

    emitLoad(innermostFinallyContext->completionTypeRegister(), jumpID);
    emitJump(innermostFinallyContext->finallyLabel());
    return true; // We'll be jumping to a finally block.
}

bool BytecodeGenerator::emitReturnViaFinallyIfNeeded(RegisterID* returnRegister)
{
    size_t numberOfScopesToCheckForFinally = m_controlFlowScopeStack.size();
    if (!numberOfScopesToCheckForFinally)
        return false;

    FinallyContext* innermostFinallyContext = 0;
    while (numberOfScopesToCheckForFinally) {
        size_t scopeIndex = --numberOfScopesToCheckForFinally;
        ControlFlowScope* scope = &m_controlFlowScopeStack[scopeIndex];
        if (scope->isFinallyScope()) {
            FinallyContext* finallyContext = scope->finallyContext;
            if (!innermostFinallyContext)
                innermostFinallyContext = finallyContext;
            finallyContext->setHandlesReturns();
        }
    }
    if (!innermostFinallyContext)
        return false; // No finallys to thread through.

    emitLoad(innermostFinallyContext->completionTypeRegister(), CompletionType::Return);
    move(innermostFinallyContext->completionValueRegister(), returnRegister);
    emitJump(innermostFinallyContext->finallyLabel());
    return true; // We'll be jumping to a finally block.
}

void BytecodeGenerator::emitFinallyCompletion(FinallyContext& context, Label* normalCompletionLabel)
{
    if (context.numberOfBreaksOrContinues() || context.handlesReturns()) {
        emitJumpIfTrue(emitEqualityOp(op_stricteq, newTemporary(), context.completionTypeRegister(), emitLoad(nullptr, CompletionType::Normal)), normalCompletionLabel);

        FinallyContext* outerContext = context.outerContext();

        size_t numberOfJumps = context.numberOfJumps();
        ASSERT(outerContext || numberOfJumps == context.numberOfBreaksOrContinues());

        // Handle Break or Continue completions that jumps into this FinallyContext.
        for (size_t i = 0; i < numberOfJumps; i++) {
            RefPtr<Label> nextLabel = newLabel();
            FinallyJump& jump = context.jumps(i);
            emitJumpIfFalse(emitEqualityOp(op_stricteq, newTemporary(), context.completionTypeRegister(), emitLoad(nullptr, jump.jumpID)), nextLabel.get());

            // This case is for Break / Continue completions from an inner finally context
            // with a jump target that is not beyond the next outer finally context:
            //
            //     try {
            //         for (... stuff ...) {
            //             try {
            //                 continue; // Sets completionType to jumpID of top of the for loop.
            //             } finally {
            //             } // Jump to top of the for loop on completion.
            //         }
            //     } finally {
            //     }
            //
            // Since the jumpID is targetting a label that is inside the outer finally context,
            // we can jump to it directly on completion of this finally context: there is no intermediate
            // finally blocks to run. After the Break / Continue, we will contnue execution as normal.
            // So, we'll set the completionType to Normal (on behalf of the target) before we jump.
            // We can also set the completion value to undefined, but it will never be used for normal
            // completion anyway. So, we'll skip setting it.

            restoreScopeRegister(jump.targetLexicalScopeIndex);
            emitLoad(context.completionTypeRegister(), CompletionType::Normal);
            emitJump(jump.targetLabel.get());

            emitLabel(nextLabel.get());
        }

        // Handle completions that take us out of this FinallyContext.
        if (outerContext) {
            if (context.handlesReturns()) {
                RefPtr<Label> isNotReturnLabel = newLabel();
                emitJumpIfFalse(emitEqualityOp(op_stricteq, newTemporary(), context.completionTypeRegister(), emitLoad(nullptr, CompletionType::Return)), isNotReturnLabel.get());

                // This case is for Return completion from an inner finally context:
                //
                //     try {
                //         try {
                //             return result; // Sets completionType to Return, and completionValue to result.
                //         } finally {
                //         } // Jump to outer finally on completion.
                //     } finally {
                //     }
                //
                // Since we know there's at least one outer finally context (beyond the current context),
                // we cannot actually return from here. Instead, we pass the completionType and completionValue
                // on to the next outer finally, and let it decide what to do next on its completion. The
                // outer finally may or may not actual return depending on whether it encounters an abrupt
                // completion in its body that overrrides this Return completion.

                move(outerContext->completionTypeRegister(), context.completionTypeRegister());
                move(outerContext->completionValueRegister(), context.completionValueRegister());
                emitJump(outerContext->finallyLabel());

                emitLabel(isNotReturnLabel.get());
            }

            bool hasBreaksOrContinuesThatEscapeCurrentFinally = context.numberOfBreaksOrContinues() > numberOfJumps;
            if (hasBreaksOrContinuesThatEscapeCurrentFinally) {
                RefPtr<Label> isThrowOrNormalLabel = newLabel();
                emitJumpIfTrue(emitBinaryOp(op_beloweq, newTemporary(), context.completionTypeRegister(), emitLoad(nullptr, CompletionType::Throw)), isThrowOrNormalLabel.get());

                // A completionType above Throw means we have a Break or Continue encoded as a jumpID.
                // We already ruled out Return above.
                static_assert(CompletionType::Throw < CompletionType::Return && CompletionType::Throw < CompletionType::Return, "jumpIDs are above CompletionType::Return");

                // This case is for Break / Continue completions in an inner finally context:
                //
                // 10: label:
                // 11: try {
                // 12:     try {
                // 13:         for (... stuff ...)
                // 14:             break label; // Sets completionType to jumpID of label.
                // 15:     } finally {
                // 16:     } // Jumps to outer finally on completion.
                // 17:  } finally {
                // 18:  }
                //
                // The break (line 14) says to continue execution at the label at line 10. Before we can
                // goto line 10, the inner context's finally (line 15) needs to be run, followed by the
                // outer context's finally (line 17). 'outerContext' being non-null above tells us that
                // there is at least one outer finally context that we need to run after we complete the
                // current finally. Note that unless the body of the outer finally abruptly completes in a
                // different way, that outer finally also needs to complete with a Break / Continue to
                // the same target label. Hence, we need to pass the jumpID in this finally's completionTypeRegister
                // to the outer finally. The completion value for Break and Continue according to the spec
                // is undefined, but it won't ever be used. So, we'll skip setting it.
                //
                // Note that all we're doing here is passing the Break / Continue completion to the next
                // outer finally context. We don't worry about finally contexts beyond that. It is the
                // responsibility of the next outer finally to determine what to do next at its completion,
                // and pass on to the next outer context if present and needed.

                move(outerContext->completionTypeRegister(), context.completionTypeRegister());
                emitJump(outerContext->finallyLabel());

                emitLabel(isThrowOrNormalLabel.get());
            }

        } else {
            // We are the outermost finally.
            if (context.handlesReturns()) {
                RefPtr<Label> notReturnLabel = newLabel();
                emitJumpIfFalse(emitEqualityOp(op_stricteq, newTemporary(), context.completionTypeRegister(), emitLoad(nullptr, CompletionType::Return)), notReturnLabel.get());

                // This case is for Return completion from the outermost finally context:
                //
                //     try {
                //         return result; // Sets completionType to Return, and completionValue to result.
                //     } finally {
                //     } // Executes the return of the completionValue.
                //
                // Since we know there's no outer finally context (beyond the current context) to run,
                // we can actually execute a return for this Return completion. The value to return
                // is whatever is in the completionValueRegister.

                emitWillLeaveCallFrameDebugHook();
                emitReturn(context.completionValueRegister());

                emitLabel(notReturnLabel.get());
            }
        }
    }

    // By now, we've rule out all Break / Continue / Return completions above. The only remaining
    // possibilities are Normal or Throw.

    emitJumpIfFalse(emitEqualityOp(op_stricteq, newTemporary(), context.completionTypeRegister(), emitLoad(nullptr, CompletionType::Throw)), normalCompletionLabel);

    // We get here because we entered this finally context with Throw completionType (i.e. we have
    // an exception that we need to rethrow), and we didn't encounter a different abrupt completion
    // that overrides that incoming completionType. All we have to do here is re-throw the exception
    // captured in the completionValue.
    //
    // Note that unlike for Break / Continue / Return, we don't need to worry about outer finally
    // contexts. This is because any outer finally context (if present) will have its own exception
    // handler, which will take care of receiving the Throw completion, and re-capturing the exception
    // in its completionValue.

    emitThrow(context.completionValueRegister());
}

void BytecodeGenerator::pushOptionalChainTarget()
{
    m_optionalChainTargetStack.append(newLabel());
}

void BytecodeGenerator::popOptionalChainTarget()
{
    ASSERT(m_optionalChainTargetStack.size());
    emitLabel(m_optionalChainTargetStack.takeLast().get());
}

void BytecodeGenerator::popOptionalChainTarget(RegisterID* dst, bool isDelete)
{
    RefPtr<Label> endLabel = newLabel();
    emitJump(endLabel.get());

    popOptionalChainTarget();
    emitLoad(dst, isDelete ? jsBoolean(true) : jsUndefined());

    emitLabel(endLabel.get());
}

void BytecodeGenerator::emitOptionalCheck(RegisterID* src)
{
    ASSERT(m_optionalChainTargetStack.size());
    emitJumpIfTrue(emitIsUndefinedOrNull(newTemporary(), src), m_optionalChainTargetStack.last().get());
}

RegisterID* BytecodeGenerator::emitToThis(RegisterID* srcDst)
{
    UnlinkedValueProfile profile = emitProfiledOpcode(op_to_this);
    instructions().uncheckedAppend(kill(srcDst));
    instructions().uncheckedAppend(ecmaMode().value());
    instructions().uncheckedAppend(0); // cachedStructure
    instructions().uncheckedAppend(0); // toThisStatus
    instructions().uncheckedAppend(profile);
    return srcDst;
}

} // namespace JSC

namespace WTF {

void printInternal(PrintStream& out, JSC::Variable::VariableKind kind)
{
#ifndef NDEBUG
    switch (kind) {
    case JSC::Variable::NormalVariable:
        out.print("Normal");
        return;
    case JSC::Variable::SpecialVariable:
        out.print("Special");
        return;
    }
    RELEASE_ASSERT_NOT_REACHED();
#endif
}

} // namespace WTF
